/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.awt.image.*;
import java.io.*;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME.ExhibitTypeParameters;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;


/**Base interface to compute the score and confidence for a 2D still image.
 * All methods in this interface are guaranteed to be "safe"
 * insofaras they will complete in "reasonable" time
 * with reasonable heap memory (and other resource, eg stack)
 * and without doing anything that wouldn't be allowed in
 * a minimal Applet/JWS sandbox.
 * <p>
 * All classes implementing this interface should be completely thread-safe
 * in their implementation of the computeScoreAndConfidence() method,
 * and preferrably purely functional (no visible side-effects),
 * with as many concurrent threads as required safely doing separate computations
 * in any one instance.
 * <p>
 * Classes implementing this interface should, where possible,
 * do their calculations using integer arithmetic,
 * since FPUs to support float/double calaculations may be a scarce resource
 * on newer highly-threaded CPUs such as Sun's Niagara.
 */
public abstract class AbstractImgScorer extends AbstractScorer implements ScorerIF
    {
    /**Create simple non-parameterised instance. */
    public AbstractImgScorer() { }

    /**Create parameterised version. */
    public AbstractImgScorer(final String nameAndParameters)
        { super(nameAndParameters); }

    /**Compute score [-1,+1] and confidence[0,+1] for given image; never null.
     *
     * @param stillImage  non-null still 2D image;
     *     if greater than 256x256 pixels then will be internally scaled
     */
    public abstract ScoreAndConf computeScoreAndConfidence(RenderedImage stillImage);

    /**Implement core/generic compute method specially for image types.
     * For any suitable 2D still image type that has a standard thumbnail,
     * this uses calls the computeScoreAndConfidence(RenderedImage) method
     * with that thumbnail.
     * <p>
     * We use the standard thumbnail as being representative of the full image
     * (and usually good enough for a human to make a judgement from)
     * with out requiring all the system resources of the full image.
     * <p>
     * For non-image types (or images with no thumbnails possible)
     * this returns (0,0) to indicate that it does not have a view.
     *
     * @throws IOException in case of I/O difficulties,
     *     or when a thumbnail is not currently available but may be later (upon retry)
     */
    public ScoreAndConf computeScoreAndConfidence(final SimpleExhibitPipelineIF dataSource,
                                                  final String exhibitName)
        throws IOException
        {
        final ExhibitTypeParameters exhibitType = ExhibitMIME.getInputFileType(exhibitName);
        // If this does look like an image type from which we can make thumbnails,
        // then return the (0,0) "no-idea" value.
        if((exhibitType == null) ||
           !exhibitType.canPossiblyCreateThumbnailOfSameMIMEType() ||
           !(exhibitType.handler instanceof AbstractImageHandler))
            { return(ScoreAndConf.NO_OPINION); }
        final ExhibitThumbnails tns = dataSource.getThumbnails(exhibitName, true);

        // Where thumbnails are not currently available,
        // but may become so in future,
        // we throw an exception.
        if(tns == null)  { throw new InterruptedIOException("thumbnail not currently available: may retry: " + exhibitName); }

        // If no standard thumbnail is available (or if it cannot be decoded)
        // then return the (0,0) "no-idea" value.
        if(tns.getStandard() == null) { return(ScoreAndConf.NO_OPINION); }
        final BufferedImage tnbi = exhibitType.handler.decodeImage(new ByteArrayInputStream(tns.getStandard().toByteArrray()));
        if(tnbi == null) { return(ScoreAndConf.NO_OPINION); }

        // Score the thumbnail image.
        return(computeScoreAndConfidence(tnbi));
        }
    }
