package org.hd.d.pg2k.ai.scorer;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.Tuple.Pair;

import sun.security.action.GetBooleanAction;

/**Class to implement some common Scorer features in one place.
 * 
 * @author dhd
 */
public abstract class AbstractScorer implements ScorerIF
    {
    /**Default constructor uses class-based name and allows no parameterisation. */
    protected AbstractScorer()
        {
        nameAndParameters = getDefaultName();
        paramValueMap = Collections.emptyMap(); // No parameters.
        }

    /**Construct instance with name and optional parameters.
     * The name-and-parameters value must be synatically valid,
     * and the Scorer name must match the "default" name (at least for now),
     * but the parameters are not interpretted here.
     * 
     * @throws IllegalArgumentException  if parameter syntax or Scorer bad name is wrong
     */
    protected AbstractScorer(final String nameAndParameters)
        throws IllegalArgumentException
        {
        final Pair<String, Map<String, String>> nap = parseNameAndParameters(nameAndParameters);
        if(!getDefaultName().equals(nap.first))
            { throw new IllegalArgumentException("Scorer base name mismatch"); }
        this.nameAndParameters = nameAndParameters;
        paramValueMap = nap.second; // Capture the parameters.
        }

    /**Get default name for this Scorer; always a valid ScorerName.
     * Must not be overrideable else it will not be safe to use in constructors.
     */
    protected final String getDefaultName()
        {
        final String fullName = this.getClass().getName();
        final String[] tokens = fullName.split("[.$]"); // Allow for inner classes too.
        final String name = tokens[tokens.length-1];  // Use final class-name component.
        assert(isValidScorerName(name));
        return(name);
        }

    /**Name and parameters of this scorer always starting with our  and syntactically valid; never null.
     * By default this has no parameters, and is based on the implementing class' name.
     */
    protected final String nameAndParameters;

    /**Immutable parameter map; never null, may be empty. */
    protected final Map<String, String> paramValueMap;

    /**Return the base name (the last component of the class name unless overridden); never null. */
    public String getName()
        {
        final int sep = nameAndParameters.indexOf(SEPARATOR);
        if(sep == -1) { return(nameAndParameters); /* No parameters to remove. */ }
        return(nameAndParameters.substring(0, sep));
        }

    /**Return the name with any parameters for this instance; never null. */
    public String getNameAndParameters() { return(nameAndParameters); }

    /**Computes randomly perturbed parameters (if any); never null.
     * This returns the name and parameters (if any of the current Scorer)
     * with zero or more of any parameters altered by a small amount
     * and guaranteed to be valid (eg in range for ant int value).
     * <p>
     * Usually at least one parameter will be perturbed.
     */
    public String getPerturbedNameAndParameters()
        {
        final Collection<ScorerParam> params = getParameterDefsAndValues();
        // If no parameters then just return the base name.
        final int pSize = params.size();
        if(pSize == 0) { return(nameAndParameters); }

        // Prepare a new perturbed value.
        final StringBuilder sb = new StringBuilder(2 * nameAndParameters.length());
        sb.append(getName());
        // Attempt to force a mutation in the i-th parameter.
        int i = (pSize < 2) ? 0 : Rnd.goodRnd.nextInt(pSize);
        for(final ScorerParam p : params)
            {
            sb.append(SEPARATOR);
            if((i-- == 0) || Rnd.goodRnd.nextBoolean())
                { sb.append(p.perturb().toNameValueString()); /* Mutate. */ }
            else
                { sb.append(p.toNameValueString()); /* Leave as-is. */ }
            }
        return(sb.toString());
        }

    /**Get (immutable, ordered) parameter definitions and current values (immutable) for this Scorer, empty by default; never null. */
    public Collection<ScorerParam> getParameterDefsAndValues() { return(Collections.emptyList()); }

    /**Regex pattern to match for Scorer parameter-name validity; never null. */
    public static final Pattern parameterNameRegex = Pattern.compile("[a-zA-Z_][0-9a-zA-Z_]*");

    /**Regex pattern to match for Scorer name validity; never null.
     * Currently the same as the parameter name regex for simplicity.
     */
    public static final Pattern scorerNameRegex = parameterNameRegex;

    /**Returns true iff the parameter name supplied is non-null and is valid. */
    public static boolean isValidParameterName(final String name)
        {
        if(null == name) { return(false); }
        return(parameterNameRegex.matcher(name).matches());
        }

    /**Returns true iff the Scorer name supplied is non-null and is valid. */
    public static boolean isValidScorerName(final String name)
        {
        if(null == name) { return(false); }
        return(scorerNameRegex.matcher(name).matches());
        }

    /**The character used to tokenise the scorer name-plus-parameters form <i>ScorerName{:name=value}*</i>.
     * A printable non-whitespace 7-bit ASCII character.
     */
    public static final char SEPARATOR = ':';

    /**Double separator, not permitted in Scorer name-and-parameters. */
    private static final String DOUBLE_SEP = ("" + AbstractScorer.SEPARATOR) + AbstractScorer.SEPARATOR;

    /**Regex pattern to separate parameters; never null. */
    public static final Pattern parameterSepRegex = Pattern.compile("["+AbstractScorer.SEPARATOR+"]");

    /**Parse name-and-parameters into name and name-value immutable Map; never null.
     * 
     * @param nameAndParameters  Scorer name-and-parameters; never null
     *
     * @return non-null non-empty Scorer base name and non-null possibly empty parameter set
     * 
     * @throws IllegalArgmentException  for null/invalid argument
     */
    public static Pair<String, Map<String,String>> parseNameAndParameters(final String nameAndParameters)
        throws IllegalArgumentException
        {
        // Reject common syntax/validity problems quickly.
        if((nameAndParameters == null) ||
           (nameAndParameters.length() == 0) ||
           (nameAndParameters.charAt(0) == AbstractScorer.SEPARATOR) ||
           (nameAndParameters.indexOf(DOUBLE_SEP) != -1))
            { throw new IllegalArgumentException(); }

        // Fast-track the no-parameters case...
        if(nameAndParameters.indexOf(AbstractScorer.SEPARATOR) == -1)
            {
            if(!isValidScorerName(nameAndParameters)) { throw new IllegalArgumentException("bad Scorer simple name"); }
            return(new Pair<String, Map<String, String>>(nameAndParameters, Collections.<String, String>emptyMap()));
            }

        final String tokens[] = parameterSepRegex.split(nameAndParameters);
        if(tokens.length < 2) { throw new IllegalArgumentException(); }

        final String name = tokens[0];
        if(!isValidScorerName(name)) { throw new IllegalArgumentException("bad Scorer name with parameters"); }

        // Parse the parameters.
        final HashMap<String, String> m = new HashMap<String, String>(2 * tokens.length);
        for(int i = tokens.length; --i > 0; )
            {
            final String token = tokens[i];
            final int ei = token.indexOf('=');
            if(ei <= 0) { throw new IllegalArgumentException("missing/initial '=' in parameter"); }
            if(ei != token.lastIndexOf('=')) { throw new IllegalArgumentException("multiple '=' in parameter"); }
            final String paramName = token.substring(0, ei);
            if(!isValidParameterName(paramName)) { throw new IllegalArgumentException("invalid parameter name"); }
            // We intern the parameter name as likely common to all instances.
            m.put(MemoryTools.intern(paramName), token.substring(ei+1));
            }
 
        return(new Pair<String, Map<String, String>>(name, Collections.unmodifiableMap(m)));
        }
    }