/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.util.Comparator;

import org.hd.d.pg2k.svrCore.ExhibitPropsComputableMutable;

/**Immutable return type from scorer method.
 * Allowed values run from -MAX (bad) through 0 (neutral) to +MAX (good) for score,
 * and from 0 (no confidence) to +MAX (full) for confidence.
 * <p>
 * Note that the score values are symmetrical about zero.
 * <p>
 * The low-ish precision of the elements is deliberate
 * to reflect the fact that scoring is bound to be fairly imprecise.
 */
public final class ScoreAndConf
    {
    /**Maximum positive or negative value of any element; strictly positive. */
    public static final short MAX = Short.MAX_VALUE;
    /**Score: -MAX (bad) through 0 (neutral) to +MAX (good). */
    public final short score;
    /**Confidence: 0 (no confidence) to +MAX (full) for confidence. */
    public final short confidence;

    /**Construct an instance. */
    public ScoreAndConf(final short score, final short confidence)
        {
        // We only need limited validation since we hit representation limits on +ve side.
        if((score < -MAX) || (confidence < 0)) { throw new IllegalArgumentException(); }
        this.score = score;
        this.confidence = confidence;
        }
    /**Construct an instance using int values; requires extra checking. */
    public ScoreAndConf(final int score, final int confidence)
        {
        this((short) score, (short) confidence);
        // Validate post-hoc in case there were invalid msbits set.
        if((score < -MAX) || (confidence < 0)) { throw new IllegalArgumentException(); }
        if((score > +MAX) || (confidence > +MAX)) { throw new IllegalArgumentException(); }
        }

    /**The hash depends on all elements. */
    @Override public int hashCode() { return((score * 32797) + confidence); }
    /**Equality depends on all elements. */
    @Override public boolean equals(Object obj)
        {
        if(this == obj) return true;
        if(!(obj instanceof ScoreAndConf)) return false;
        final ScoreAndConf other = (ScoreAndConf) obj;
        if(confidence != other.confidence) return false;
        if(score != other.score) return false;
        return true;
        }

    /**Compute human-readable summary; never null.
     * This is also designed to be parsable to (re)construct an instance
     * as an alternative to serialisation.
     * <p>
     * This includes the scale (maximum) for non-zero values
     * to allow for robust reconstruction in future versions.
     */
    @Override public String toString()
        {
        final StringBuilder sb = new StringBuilder(32);
        sb.append("score=").append(score);
        if(score != 0) { sb.append('/').append(MAX); }
        sb.append(":confidence=").append(confidence);
        if(confidence != 0) { sb.append('/').append(MAX); }
        return(sb.toString());
        }

    /**Find out if this is notably good or bad using the same thresholds as for voting (etc) in the rest of the Gallery.
     * Returns TRUE if rated good, FALSE if bad,
     * null if not significantly either (ie too close to neutral) or if unknown.
     * <p>
     * This is based on the product of score and confidence.
     */
    public Boolean isGood()
        {
        final int goodness = (score * confidence * 2); // Normalise for compare against int threshold.
//System.err.println("goodness = " + goodness);
        if((goodness < ExhibitPropsComputableMutable.GOODBAD_LIMIT_INT) &&
           (goodness > -ExhibitPropsComputableMutable.GOODBAD_LIMIT_INT)) { return(null); }
        return((goodness >= 0) ? Boolean.TRUE : Boolean.FALSE);
        }

    /**All-zeros value used to mean "no idea" / "no opinion" when returned by a Scorer. */
    public static final ScoreAndConf NO_OPINION = new ScoreAndConf(0, 0);


    /**A static Comparator that sorts by confidence. */
    public static final Comparator<ScoreAndConf> ByConfidence = (new Comparator<ScoreAndConf>()
        {
        public int compare(final ScoreAndConf o1, final ScoreAndConf o2)
            { return(o1.confidence - (int) o2.confidence); }
        });

    /**A static Comparator that sorts by confidence, ignoring sign. */
    public static final Comparator<ScoreAndConf> ByAbsConfidence = (new Comparator<ScoreAndConf>()
        {
        public int compare(final ScoreAndConf o1, final ScoreAndConf o2)
            { return(Math.abs(o1.confidence) - (int)Math.abs(o2.confidence)); }
        });
    }