/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.io.IOException;
import java.util.*;


/**Base interface to compute (and cache) the score and confidence for exhibits.
 * Note: since the result of this computation may be used in computing
 * (EPCM) the ExhibitPropsComputableMutable value for an exhibit,
 * then any implementation of this must avoid forcing recaclulation
 * of any EPCM value to avoid danger of infinite recursion
 * (other than the static calcVoteFactor() method).
 * Ideally the value computed will not depend on any EPCM value.
 * <p>
 * Methods that take an allowStale parameter
 * will generally try to reduce compute and I/O effort and increase robustness
 * at a smallish cost in currency.
 * Typically such methods will return data up to about one day old if present in cache,
 * else the newest available from cache however old it is if there is a problem
 * computing an up-to-date value (eg because of connectivity issues).
 */
public interface ScorerCacheIF
    {
    /**Computes a weighted composite score [-1,+1] and confidence [0,+1] for the specified exhibit with the best available scorers/parameters; never null but may be (0,0).
     * This is the highest-level available entry to the cache.
     *
     * @param exhibitName  valid full exhibit name
     * @param allowStale  if true then allow a stale value from cache,
     *     else throw an exception if nothing is currently available
     *
     * @return (0,0) if named scorer is not available,
     *         cannot be used with the specified parameters,
     *         or cannot be applied to the indicated exhibit
     *         (eg because of the exhibit type or the exhibit does not exist);
     *         else returns a non-null ScoreAndConf value
     */
    ScoreAndConf computeCompositeScoreAndConfidence(final String exhibitName,
                                                    final boolean allowStale)
        throws IOException;

    /**Computes the raw score [-1,+1] and confidence [0,+1] for the specified exhibit with the specified scorer and parameters; never null but may be (0,0).
     * This is <em>NOT</em> moderated/weighted in the light of how well the scorer's predictions
     * match reality; that requires an extra step.
     *
     * @param exhibitName  valid full exhibit name
     * @param scorerNameAndParameters  the name and parameters of the scorer; never null
     * @param allowStale  if true then allow a stale value from cache,
     *     else throw an exception if nothing is currently available
     *
     * @return (0,0) if named scorer is not available,
     *         cannot be used with the specified parameters,
     *         or cannot be applied to the indicated exhibit
     *         (eg because of the exhibit type or the exhibit does not exist);
     *         else returns a non-null ScoreAndConf value
     */
    ScoreAndConf computeUnweightedScoreAndConfidence(final String exhibitName,
                                                     final String scorerNameAndParameters,
                                                     final boolean allowStale)
        throws IOException;

    /**ScoreAndConfidence for the given scorer itself over all exhibit types; never null but may be (0,0) where the scorer is unknown or untested.
     * Essentially the result of this should be multiplied by the result for each exhibit
     * (for the same scorer and parameters)
     * to normalise the predicted score and confidence for the exhibit.
     * <p>
     * Some results will be cached (typically those from the base/current Scorer lists)
     * while other may have to be computed each time, which may be slow.
     *
     * @param scorerNameAndParameters  the name and parameters of the scorer; never null
     * @param allowStale  if true then allow a stale or low-confidence value from cache,
     *     else throw an exception if nothing is currently available
     *     and we cannot quickly compute enough points to increase our confidence
     *
     * @return  the score represents the correlation with the underlying votes
     *     (and whatever the scoring is measured against)
     *     with MAX meaning perfect correlation, 0 meaning no correlation,
     *     and -MAX meaning perfectly wrong answers all the time,
     *     and the confidence 0 if we have no (or very few) data points yet
     *     and approaching MAX as we have a large (enough) number of data points
     */
    ScoreAndConf computeScorerWeighting(final String scorerNameAndParameters,
                                        final boolean allowStale)
        throws IOException;

    /**Base set of available scorers name (no parameters); never null but may be empty.
     * The values returned are of the form <i>ScorerName</i>.
     */
    Set<String> getBaseScorersWithoutParameters();

    /**Current set of available scorers name and parameters (where applicable); never null but may be empty.
     * The values returned are of the form <i>ScorerName{:name=value}*</i>.
     * <p>
     * The scorers returned by this will generally be the best available,
     * usually the best one or two per base-Scorer type,
     * allowing in particular that different Scorer types may have different domains.
     */
    Set<String> getCurrentScorersWithParameters(final boolean allowStale);
    
    /**Get current population size; non-negative. */
    int size();
    

    /**Trivial implementation that always returns "no comment"/empty values; non-null. */
    public static final ScorerCacheIF TRIVIAL = new ScorerCacheIF() {
        public ScoreAndConf computeCompositeScoreAndConfidence(String exhibitName, boolean allowStale) { return(ScoreAndConf.NO_OPINION); }
        public ScoreAndConf computeScorerWeighting(String scorerNameAndParameters, boolean allowStale)  { return(ScoreAndConf.NO_OPINION); }
        public ScoreAndConf computeUnweightedScoreAndConfidence(String exhibitName, String scorerNameAndParameters, boolean allowStale) { return(ScoreAndConf.NO_OPINION); }
        public Set<String> getBaseScorersWithoutParameters() { return(Collections.emptySet()); }
        public Set<String> getCurrentScorersWithParameters(boolean allowStale) { return(Collections.emptySet()); }
        public int size() { return(0); }
        };
    }
