/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.*;

import org.hd.d.pg2k.ai.scorer.fixed.*;
import org.hd.d.pg2k.ai.scorer.parameterised.LocalSampler;
import org.hd.d.pg2k.ai.scorer.parameterised.SimpleExposure;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.ExhibitPropsComputableMutable.Factor;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.vars.*;

import ORG.hd.d.IsDebug;


/**Simple/default implementation to compute (and cache) the score and confidence for exhibits.
 * Note: since the result of this computation may be used in computing
 * (EPCM) the ExhibitPropsComputableMutable value for an exhibit,
 * then any implementation of this must avoid forcing recaclulation
 * of any EPCM value to avoid danger of infinite recursion
 * (other than the static calcVoteFactor() method).
 * Ideally the value computed will not depend on any EPCM value.
 */
public final class ScorerCacheImpl implements ScorerCacheIF
    {
    /**Construct an instance attached to the supplied data source.
     * 
     * @param dataSource  live data source; must not be null
     */
    public ScorerCacheImpl(final SimpleExhibitPipelineIF dataSource,
                           final SimpleLoggerIF log)
        {
        if((dataSource == null) || (log == null))
            { throw new IllegalArgumentException(); }
        this.dataSource = dataSource;
        this.log = log;

        // Create a new population of Scorers.
        population = new ScorerPopulation(fixedSimpleScorers.keySet(),
                                          MAX_SCORER_SCORES_RETAINED);
        }

    /**Live data source for variables, exhibit data and metadata, etc; never null. */
    private final SimpleExhibitPipelineIF dataSource;

    /**Logger; never null. */
    private final SimpleLoggerIF log;

    /**Computes a weighted composite score [-1,+1] and confidence [0,+1] for the specified exhibit with the best available scorers/parameters; never null but may be (0,0).
     * This uses the Scorers available from getCurrentScorersWithParameters(),
     * but skipping any marked as BadScorer that have made it through filtering.
     *
     * @param exhibitName  valid full exhibit name
     * @param allowStale  if true then allow a stale value from cache,
     *     else throw an exception if nothing is currently available
     *
     * @return (0,0) if named scorer is not available,
     *         cannot be used with the specified parameters,
     *         or cannot be applied to the indicated exhibit
     *         (eg because of the exhibit type or the exhibit does not exist);
     *         else returns a non-null ScoreAndConf value
     */
    public ScoreAndConf computeCompositeScoreAndConfidence(final String exhibitName,
                                                           final boolean allowStale)
        throws IOException
        {
        // We apply each available non-zero-confidence Scorer to the exhibit,
        // and compute a single overall score.
        // We don't cache this result;
        // we recompute it each time as it should be quick to so do
        // as all of its components are cached.
        // Note that this may be VERY slow if it forces all Scorers to be (re)evaluated.

        // Compute results from individual eligible Scorers.
        // Because of a potential lack of symmetry in scoring
        // (ie reversing the value of a bad Scorer does not necessarily make it a good one)
        // we are only interested in the highest-confidence, positive-correlation Scorers.
        SortedMap<ScoreAndConf, ScorerIF> results = new TreeMap<ScoreAndConf, ScorerIF>(ScoreAndConf.ByConfidence);
        for(final String snp : getCurrentScorersWithParameters(allowStale))
            {
            // Compute the weighting over all available exhibits.
            final ScoreAndConf scorerWeighting = computeScorerWeighting(snp, allowStale);
            // Quickly reject those Scorers with a non-positive score or confidence.
            if(scorerWeighting.score <= 0) { continue; }
            if(scorerWeighting.confidence == 0) { continue; }
            // Now "score" the exhibit with this Scorer.
            final ScoreAndConf score = computeUnweightedScoreAndConfidence(exhibitName, snp, allowStale);
            // Apply the weighting to the score, rejecting any whose confidence is now zero.
            // Note that we only use the sign of the Scorer correlation (score) for the score,
            // but we multiply both the Scorer correlation and confidence into the final confidence.
            final ScoreAndConf weighted = new ScoreAndConf(score.score,
                                                           (int) ((score.confidence * scorerWeighting.confidence * (long)scorerWeighting.score) / (ScoreAndConf.MAX * (long) ScoreAndConf.MAX)));
            if(weighted.confidence == 0) { continue; }
            assert(weighted.confidence > 0);
            final ScorerIF sc = getScorerInstance(snp);
            assert(sc != null);
            results.put(weighted, sc);
            }

        final int rSize = results.size();
        // Handle trival cases quickly...
        // If no applicable results then return "no opinion".
        if(rSize == 0) { return(ScoreAndConf.NO_OPINION); }

        // Get the keys (S&C values) in rising order by confidence (ignoring sign).
        List<ScoreAndConf> sorted = new ArrayList<ScoreAndConf>(results.keySet());
        assert(sorted.get(0).confidence <= sorted.get(sorted.size()-1).confidence);

        // Discard the least-confident results if we have more than a few.
        // The aim is to reduce noise and improve "strength" of answer.
        // We do this BEFORE weeding out any "straw man" dummy/fake/test Scorers
        // so as to help push out any very weak real Scorers (weaker than chance/faking).
        // We don't discard too many Scorers so as to allow for multi-modal ranking
        // by a reasonable diversity of Scorer types.
        final int toDiscard = (rSize >>> 3); // FIXME: Optimise/document this fraction.
        if(toDiscard > 0) { sorted = sorted.subList(toDiscard, rSize); }
        // Now weed out any remaining dummy/stooge/test/benchmark Scorers.
        for(int i = sorted.size(); --i >= 0; )
            { if(results.get(sorted.get(i)) instanceof BadScorer) { sorted.remove(i); } }
        // Recompute the size.
        final int finalSize = sorted.size();

        // Handle trival cases quickly...
        // If no applicable results then return "no opinion".
        if(finalSize == 0) { return(ScoreAndConf.NO_OPINION); }
        // If exactly one result then return it "as is".
        if(finalSize == 1) { return(sorted.get(0)); }

        // Get the highest confidence value (ignoring sign).
        final int highestConfidence = sorted.get(finalSize-1).confidence;
        assert(sorted.get(0).confidence <= highestConfidence);
        // Compute the result score as the mean score*confidence
        // (thus contributing to the result in proportion to confidence)
        // normalised to be score*1 for the highest confidence,
        // and the confidence being the RMS confidence of all contributors
        // (to be biased towards greater confidence in the result
        // when a number of low-confidence Scorers are included).
        long totalSC = 0;
        long totalConfSq = 0;
        for(final ScoreAndConf sc : sorted)
            {
            totalSC += (((sc.score * (int)sc.confidence) * (long)ScoreAndConf.MAX) / highestConfidence);
            totalConfSq += (sc.confidence * (int)sc.confidence);
            }
        // Normalise the score prediction to be able to run the full gamut [-MAX,+MAX].
        final long rawCompositeScore = totalSC / (ScoreAndConf.MAX * finalSize);
//if(IsDebug.isDebug) { System.out.println("rSize = " + finalSize); }
//if(IsDebug.isDebug) { System.out.println("rawCompositeScore = " + rawCompositeScore); }
        final long rawCompositeConf = Math.round(Math.sqrt(totalConfSq / finalSize));
//if(IsDebug.isDebug) { System.out.println("rawCompositeConf = " + rawCompositeConf); }
        return(new ScoreAndConf((short) Math.min(ScoreAndConf.MAX, Math.max(-ScoreAndConf.MAX, rawCompositeScore)),
                                (short) Math.min(ScoreAndConf.MAX, Math.max(0, rawCompositeConf))));
        }


    /**Size-limited cache of raw exhibit scores (and when computed); never null.
     * Map from scorer name-and-parameters and exhibit name to ScoreAndConfidence and time computed.
     * <p>
     * This is a thread-safe map, with automatic LRU discarding of stale/unused values.
     * <p>
     * A lock can be held on this map to make compound operations atomic,
     * though the lock should be held for as little time as possible
     * so as to maximise available concurrency.
     */
    private final MemoryTools.SimpleLRUMapAutoSizeForHitRate<Pair<String,String>, Pair<Long, ScoreAndConf>> _rawScores =
        new MemoryTools.SimpleLRUMapAutoSizeForHitRate<Pair<String,String>, Pair<Long, ScoreAndConf>>(0.01f, 2*MIN_VOTED_FOR_SET_SIZE, 100*MIN_VOTED_FOR_SET_SIZE, 0.7f);

    /**Computes the raw score [-1,+1] and confidence [0,+1] for the specified exhibit with the specified scorer and parameters; never null but may be (0,0).
     * This is <em>NOT</em> moderated/weighted in the light of how well the scorer's predictions
     * match reality; that requires an extra step.
     *
     * @param exhibitName  valid full exhibit name
     * @param scorerNameAndParameters  the name and parameters of the scorer; never null
     * @param allowStale  if true then allow a stale value from cache,
     *     else throw an exception if nothing is currently available
     *
     * @return (0,0) if named scorer is not available,
     *         cannot be used with the specified parameters,
     *         or cannot be applied to the indicated exhibit
     *         (eg because of the exhibit type or the exhibit does not exist);
     *         else returns a non-null ScoreAndConf value
     */
    public ScoreAndConf computeUnweightedScoreAndConfidence(final String exhibitName,
                                                            final String scorerNameAndParameters,
                                                            final boolean allowStale)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();

        // Get vote events for all retained full/complete (VLONG) slots.
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(startTime);

        // Check that the scorer requested is available/valid.
        // If not then return the "no opinion" result.
        final ScorerIF scorer = getScorerInstance(scorerNameAndParameters);
        if(scorer == null) { return(ScoreAndConf.NO_OPINION); }

        // If exhibit is not valid/current then return the "no opinion" result.
        final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);
        final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);
        if(esa == null) { return(ScoreAndConf.NO_OPINION); }

        // We can use a value from cache if the following conditions hold:
        //  1) The cached value is from the current period
        //     (or the previous period if allowStale is true).
        //  2) The cached value is newer than the exhibit file date
        //     (and any thumbnails that might be used).
        // Since we expect thumbnails only to change rarely for a given exhibit,
        // ie after a thumbnail-generation algorithm change,
        // we certainly do not need to force thumbnail creation just to get a timestamp.
        final Pair<String,String> cacheKey = new Pair<String, String>(scorerNameAndParameters, exhibitName);
        // See what, if anything, is currently in cache.
        final Pair<Long, ScoreAndConf> cachedValue = _rawScores.get(cacheKey);
        final ExhibitThumbnails tns;
        if((cachedValue != null) &&
           (EventPeriod.VLONG.getIntervalNumber(cachedValue.first) >= (allowStale ? currentPeriod-1 : currentPeriod) &&
           (cachedValue.first >= esa.timestamp) &&
           (allowStale || (null == (tns = dataSource.getThumbnails(exhibitName, false))) || (cachedValue.first >= tns.created))))
            { return(cachedValue.second); }

        // Compute value from scratch.
        final ScoreAndConf result = scorer.computeScoreAndConfidence(dataSource, exhibitName);

        // Atomically replace the value in cache
        // making sure that we never replace a newer value.
        final Pair<Long, ScoreAndConf> newValue = new Pair<Long, ScoreAndConf>(startTime, result);
        synchronized(_rawScores)
            {
            final Pair<Long, ScoreAndConf> curValue = _rawScores.get(cacheKey);
            if((curValue == null) || (curValue.first < newValue.first))
                { _rawScores.put(cacheKey, newValue); }
            }
//        if(cachedValue == null) { _rawScores.putIfAbsent(cacheKey, newValue); }
//        else { _rawScores.replace(cacheKey, cachedValue, newValue); }

        return(result);
        }


    /**If true then we reject any Scorer that does not give a reasonable range of predicted scores. */
    private static final boolean REQUIRE_RESULT_VARIATION = false;

    /**Maximum number of Scorers for which retain/cache scores; strictly positive.
     * The actual number of Scorers (including fixed Scorers)
     * that we attempt to use at any one time should be (much) less than this.
     * <p>
     * This limits the amount of memory dedicated to cache, etc.
     */
    private static final int MAX_SCORER_SCORES_RETAINED = Math.max(1 << 16,
            4*SystemVariables.AI_SCORER_STRING_LOCAL_EVENT.getMaxDiffEventCount());

    /**The minimum recent period for which we collect all available votes, in ms; strictly positive.
     * This is usually at least a year to smooth out for seasonal patterns.
     */
    private static final long MIN_VOTE_SAMPLE_NEAR_TERM_TIME_MS = 365242L * 24 * 3600; // 365.242 days in a year.

    /**Scorer population (and cache of Scorer "goodness"); never null. */
    private final ScorerPopulation population;

    /**ScoreAndConfidence for the given scorer itself over all exhibit types; never null but may be (0,0) where the scorer is unknown or untested.
     * Essentially the result of this should be multiplied by the result for each exhibit
     * (for the same scorer and parameters)
     * to normalise the predicted score and confidence for the exhibit,
     * though usually the sign of the correlation (score) is multiplied into the raw score,
     * while the value of correlation (score) is multiplied with the confidence
     * into the final confidence.
     * <p>
     * We attempt to ignore small numbers of errors for robustness.
     * @param scorerNameAndParameters  the name and parameters of the scorer; never null
     * @param allowStale  if true then allow a stale or low-confidence value from cache,
     *     else throw an exception if nothing is currently available
     *     and we cannot quickly compute enough points to increase our confidence
     *
     * @return  the score represents the correlation with the underlying votes
     *     (and whatever the scoring is measured against)
     *     with MAX meaning perfect correlation, 0 meaning no correlation,
     *     and -MAX meaning perfectly wrong answers all the time,
     *     and the confidence 0 if we have no data points yet
     *     and approaching MAX as we get a large number of data points
     */
    public ScoreAndConf computeScorerWeighting(final String scorerNameAndParameters,
                                               final boolean allowStale)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();

        // Check that the scorer requested is available/valid.
        // If not, return the "no opinion" result.
        final ScorerIF scorer = getScorerInstance(scorerNameAndParameters);
        if(scorer == null) { return(ScoreAndConf.NO_OPINION); }

        // Return what we have in cache, if anything.
        final ScoreAndConf cachedValue = population.getScorerWeighting(scorerNameAndParameters, allowStale);
        if(cachedValue != null) { return(cachedValue); }

        final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);
        // If there are no live exhibits then we can't compute (nor cache) a result.
        // This is especially important to avoid cacheing spurious false results at start-up.
        if(aep.aeid.length == 0) { return(ScoreAndConf.NO_OPINION); }

        // Compute weighting (effectively the correlation) from scratch.
        final Map<String, Factor> votedForExhibits = getVotedForExhibitsAndVoteFactors(dataSource, allowStale);
        // If there are no voted-for exhibits then we cannot compute (nor cache) a result.
        if(votedForExhibits.isEmpty()) { return(ScoreAndConf.NO_OPINION); }

        // We capture the range of scores for a given Scorer
        // to allow us to eliminate those Scorers that output only in a very narrow range
        // or simply too few results to calibrate.
        final ArrayList<Short> scoreValues = new ArrayList<Short>(votedForExhibits.size());

        // Exhibits which were usable to calibrate this Scorer.
        int usableExhibits = 0;
        // The total squared error (and confidence) in the range [0, (2*MAX)^2].
        // Each error term is multiplied by (moderated downwards by)
        // the product of the Scorer and vote-Factor confidence/correlation.
        assert((long) ScoreAndConf.MAX * (long) ScoreAndConf.MAX <= Integer.MAX_VALUE);
        long cumError = 0;
        // The dot product of the Scorer confidence and the vote Factor correlation,
        // with theoretical range [0,Y] where Y = ScoreAndConf.MAX * usableExhibits.
        // Rounding errors may take the value just outside of this range.
        long confDP = 0;
        // Last captured error and cout, if any.
        IOException err = null;
        int errorCount = 0;
        // If we didn't finish then we should not cache the result.
        boolean stoppedPrematurely = false;
        // Randomise the order in which we do the computation
        // to better queue background work, avoid systemic errors, etc.
        final List<String> names = new ArrayList<String>(votedForExhibits.keySet());
        if(Rnd.goodRnd.nextBoolean()) { Collections.shuffle(names, Rnd.fastRnd); }
        for(final String shortName : names)
            {
            // If the exhibit is missing (eg has been deleted or renamed) then skip it.
            final String fullName = aep.aeid.getFullName(shortName);
            if(fullName == null) { continue; }

            try
                {
                // Get vote-based rating (cannot be null or have non-positive correlation).
                final Factor voteFactor = votedForExhibits.get(shortName);
                // Vote correlation must be strictly positive at this point.
                assert(voteFactor.correlation > 0);

                // Compute raw Scorer result for this exhibit.
                // Note that this may quickly return zero confidence
                // therefore meaning we can skip further work for this exhibit
                // since this Scorer may not be suitable for this exhibit type.
                final ScoreAndConf sac = computeUnweightedScoreAndConfidence(fullName, scorerNameAndParameters, allowStale);
                // If confidence is zero then skip this exhibit.
                if(sac.confidence == 0) { continue; }

                final int conf = Math.round(sac.confidence * voteFactor.correlation);
                assert(conf >= 0);
                // Skip zero or noisy (very-low-confidence) correlations.
                if(conf < 1) { continue; }
                confDP += conf;
                // Compute raw error (difference) in range [0,2*MAX].
                // The Scorer result is an actual predicted score for the exhibit.
                // The smaller the difference between this predicted score
                // and the human-vote-derived score the better.
                final int rawError = sac.score - Math.round(ScoreAndConf.MAX*voteFactor.goodness);
                // Compute squared error times the (normalised to [0.0,1.0]) confidence.
                final long sqErrConf = ((rawError * (long)rawError) * conf) / ScoreAndConf.MAX;
//if(IsDebug.isDebug) { System.out.println("[Error/diff="+error+" @ conf="+conf+" for "+scorerNameAndParameters+" on "+shortName+"; sac="+sac+"; voteFactor="+voteFactor+".]"); }
                assert(sqErrConf >= 0);
                cumError += sqErrConf;
                assert(cumError >= 0); // FIXME: Check that we never overflow.
                ++usableExhibits;

                // Record this predicted exhibit score value.
                scoreValues.add(sac.score);
                }
            catch(final IOException e)
                {
                // Capture the error.
                ++errorCount;
                err = e;

                // We are prepared to skip over a few missing data points
                // (ie ignore a small number of possibly-transient errors)
                // though then do not subsequently cache the result
                // unless there has been only a very small fraction of errors.
                // This may also allow requests (eg for thumbnails/events)
                // to be queued in the background ready for next time
                // while making computation generally more robust.
                if((errorCount <= usableExhibits) ||
                   (System.currentTimeMillis() - startTime < CoreConsts.MAX_INTERACTIVE_DELAY_MS))
                    { continue; }

                // We've spent long enough: do calculations based on what we have so far...
                // If we've had to give up and we don't allow stale/partial results then abort...
                if(!allowStale) { throw e; }
                // ...else continue with what we have but don't cache it.
                stoppedPrematurely = true;
                break;
                }
            }
if(IsDebug.isDebug) { log.log("[ScorerCacheImpl: " + scorerNameAndParameters + ": usableExhibits = " + usableExhibits + ", errorCount = " + errorCount + "]"); }

        // We will accept the result if there were far more data points
        // where we had a usable vote factor AND a non-zero Scorer confidence
        // than the count of errors that we hit (and points we should have included)
        // OR if there were some exhibits but none at all that we could use with this Scorer.
        final boolean enoughDataPoints = (errorCount <= usableExhibits/16) && (aep.aeid.length > 0);

        // If not allowing stale/partial results then we must abort after too many errors.
        if(!allowStale && !enoughDataPoints) { throw err; }

        ScoreAndConf result = ScoreAndConf.NO_OPINION; // Scorer unusable unless proven otherwise.
        // We can't compute a result with no (or very few) data points
        // nor any/sufficient cumulative confidence.
        // We can cache a valid negative result though.
        final int svSize = scoreValues.size();
        if((usableExhibits > 0) && (confDP > 0) && (svSize >= 4))
            {
            final double svRangeConf;
            if(REQUIRE_RESULT_VARIATION)
                {
                // Check that the range of score values is reasonable
                // by checking that the central range (discarding outliers) is healthy,
                // ie at least (say) 25% of the available [-MAX,+MAX] range.
                // If not so, then scale down our confidence appropriately.
                Collections.sort(scoreValues);
                final int outlierIndex = Math.max(1, svSize/16);
                final short svLow = scoreValues.get(outlierIndex);
                final short svHigh = scoreValues.get(svSize-1 - outlierIndex);
                // Compute range-based confidence [0.0,1.0].
                svRangeConf = Math.min(1.0, (svHigh-svLow) / (double) (ScoreAndConf.MAX/2));
//System.out.println("svRangeConf = "+svRangeConf+" for "+scorerNameAndParameters+" with median range high/low "+svHigh+'/'+svLow);
                }
            else
                { svRangeConf = 1.0; }
            
            // Compute RMS error in range [0,2*MAX].
            final double errorRMSRaw = Math.sqrt(cumError / (confDP / (double)ScoreAndConf.MAX));
            assert((errorRMSRaw > -1) && (errorRMSRaw < 1 + 2*ScoreAndConf.MAX)); // We'll allow a little wiggle room for rounding errors...
            // Compute score in range +/- ScoreAndConf.MAX.
            final long score = ScoreAndConf.MAX - Math.round(errorRMSRaw);
//if(IsDebug.isDebug) { System.out.println("ScorerCacheImpl: score = " + score); }
            // Compute confidence in range [0,ScoreAndConf.MAX] (plus possible rounding errors).
            // We allow extra samples over our minimum desired number
            // to compensate for poor per-exhibit confidence.
            final long confidence = Math.min(ScoreAndConf.MAX,
                Math.round(svRangeConf * (confDP / usableExhibits) *
                    Math.sqrt(usableExhibits / (double) MIN_VOTED_FOR_SET_SIZE)));
//System.out.println("confidence = " + confidence);
            result = new ScoreAndConf((short) Math.max(-ScoreAndConf.MAX, Math.min(ScoreAndConf.MAX, score)),
                                      (short) Math.max(0, confidence));
            }

        // Atomically store/replace the result in cache
        // iff enough data points were available and we hadn't run out of time
        // making sure that we never replace a newer value (discarding this if necessary)
        // that was concurrently computed while we were working.
        // We allow values computed with potentially-stale inputs to be cached
        // as probably reasonably accurate if other conditions are met.
        // Note that if the Scorer that we have just generated is amongst the best
        // then it may be persisted/propagated.
        if(enoughDataPoints && !stoppedPrematurely)
            {
            population.putScorerWeighting(scorerNameAndParameters, result, dataSource, log);
log.log("[ScorerCacheImpl: computed/cached for Scorer "+scorerNameAndParameters+" "+result+".]");
            }

        return(result);
        }


    /**Compute voted-for-exhibit set; never null but may be empty.
     * This attempts to retrieve the latest vote event sets for all or most time slots
     * (concentrating on more recent ones if necessary)
     * and computes the set of all exhibits that have a vote.
     * <p>
     * This does <em>NOT</em> attempt to filter the result in any way,
     * ie by which exhibits are still live or even syntactically-valid.
     * 
     * @return  the (non-null, possibly empty) set of voted-for-exhibits
     *     up to the last full VLONG slot
     *     paired with the time at which we started to compute/fetch it
     * @throws IOException  if the data is not available or there is a timeout
     */
    private Pair<Long, Set<String>> computeRawVotedForExhibits(final SimpleVariablePipelineIF vars)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();
        final Set<String> exhibits = new HashSet<String>(2*MIN_VOTED_FOR_SET_SIZE);

        // Get vote events for all retained full/complete (VLONG) slots.
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(startTime);
        final BitSet whichValues = new BitSet(SystemVariables.EVENT_SAMPLES_RETAINED);
        whichValues.set(0, SystemVariables.EVENT_SAMPLES_RETAINED);
        final SimpleVariableDefinition[] defs = new SimpleVariableDefinition[]{ SystemVariables.VOTE_PRO, SystemVariables.VOTE_CON };
        for(final SimpleVariableDefinition def : defs)
            {
            final EventVariableValue evvsAll[] = vars.getEventValues(
                def, EventPeriod.VLONG, currentPeriod, whichValues);
            for(final EventVariableValue evv : evvsAll)
                {
                if((evv != null) && (evv.getTotalEventCount() != 0))
                    {
                    // We expect the keys to be String short exhibit names.
                    for(final Object o : evv.getInfo().keySet())
                        { exhibits.add((String) o); }
                    }
                }
            }

        return(new Pair<Long, Set<String>>(startTime, exhibits));
        }

    /**Get all voted-for exhibits (by short name) and their vote factors; never null.
     * The exhibit set is those that have been voted for or against,
     * possibly including no-longer valid/extant exhibit names,
     * because we will be using this for calibration.
     * <p>
     * Private to getVotedForExhibits().
     * <p>
     * We index from the time at which we gathered the set
     * so that we can regenerate or trim this list if the aep or event slot changes.
     * <p>
     * Because the (first) map is sorted it is quick to find the latest/oldest entries.
     * <p>
     * Usually we only keep one value (the latest one),
     * and any older values are purged only AFTER a new value has been inserted,
     * ie this map never becomes transiently empty nor more stale than necessary once non-empty.
     * <p>
     * We can keep older information around to allow some (stale) results to be computed
     * even if we cannot immediately fetch new information.
     * <p>
     * The outer and inner maps are thread-safe.
     * <p>
     * Do not assume that other activity in this collection can be locked out
     * by holding a lock on this object.  DO NOT hold a lock on this object!
     * <p>
     * The values must be immutable to enable them to be shared safely without copying.
     * <p>
     * TODO: upgrade to ConcurrentNavigableMap under JDK 1.6 to improve concurrency.
     */
    private final SortedMap<Long, Map<String, Factor>> _votedForExhibits =
        Collections.synchronizedSortedMap(new TreeMap<Long, Map<String, Factor>>());

    /**Get latest (immutable) Map of all voted-for exhibits (by short name) to vote factor; never null.
     * Returns a set of extant exhibits that have been voted for or against.
     * <p>
     * Uses a cache to avoid wasted computation from one event tick to the next,
     * and to allow this mechanism to work even in the face of brief outages
     * in data connectivity (eg when trying to fetch new values of the vote events).
     * 
     * @param dataSource  event source for fetching event set if necessary; never null
     * @param allowStale  if true, allow older data to be returned to save time,
     *     or if new event data should be used but cannot be fetched
     *     (ie allow working from stale event data for speed and/or robustness)
     */
    private Map<String, Factor> getVotedForExhibitsAndVoteFactors(final SimpleExhibitPipelineIF dataSource,
                                                                  final boolean allowStale)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();

        // What is the current period (ie for non-stale data)?
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(startTime);

        // Spin to retrieve latest cached result that is current (up-to-date).
        // (If allowStale is true, we allow a previous-tick result.)
        // This relies on:
        //  1) the map never going empty once non-empty
        //  2) the map never getting more stale
        //     (refetching the last key can only get the same or a newer value)
        // This can only return with a "current" (enough) item.
        // We fall through if an appropriate value is not available.
        if(!_votedForExhibits.isEmpty())
            {
            while(EventPeriod.VLONG.getIntervalNumber(_votedForExhibits.lastKey()) >= (allowStale ? currentPeriod-1 : currentPeriod))
                {
                final Map<String, Factor> result = _votedForExhibits.get(_votedForExhibits.lastKey());
                if(null != result) { return(result); }
                }
            }

        // No suitable cached up-to-date value,
        // so try to compute one (and cache it) now.
        // If this cannot complete (generates an exception) and allowStale is false
        // then we propagate that exception upwards.
        Pair<Long, Set<String>> rawSet = null;
        IOException err = null;
        if(!allowStale)
            { rawSet = computeRawVotedForExhibits(dataSource); }
        else
            { try { rawSet = computeRawVotedForExhibits(dataSource); } catch(final IOException e) { err = e; /* Record for later. */ } }

        // If we got something, then cache it, even though it *may* now not be the newest.
        if(rawSet != null)
            {
            final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);
            // We refuse to cache the result from an empty AEP as possibly erroneous.
            if(aep.aeid.length == 0) { return(Collections.emptyMap()); }

            final Set<String> vfe = rawSet.second;
            final Map<String, Factor> m = new HashMap<String, Factor>(2 * vfe.size());
            for(final String shortName : vfe)
                {
                // If the exhibit is missing (eg has been deleted or renamed) then skip it.
                final String fullName = aep.aeid.getFullName(shortName);
                if(fullName == null) { continue; }

                // Compute vote-based rating.
                final Factor voteFactor = ExhibitPropsComputableMutable.calcVoteFactor(fullName, aep, dataSource, MIN_VOTE_SAMPLE_NEAR_TERM_TIME_MS);
                // If vote correlation for this exhibit is too low to use then skip it.
                // "Too low" is where the effective correlation is zero for a ScoreAndConf representation.
                if(voteFactor.correlation < 0.5f/ScoreAndConf.MAX) { continue; }

                // Store a useful vote factor in the map.
                m.put(shortName, voteFactor);
//if(IsDebug.isDebug && (voteFactor.goodness > -1) && (voteFactor.goodness < +1)) { System.out.println("Middling vote Factor = "+voteFactor+" for "+fullName); }
                }

            final Map<String, Factor> result = Collections.unmodifiableMap(m);
            _votedForExhibits.put(rawSet.first, result);
            // Now zap anything older (ie stale).
            // This does not remove our entry, and therefore guarantees not to empty the map.
            // Concurrent updates may leave multiple entries in the map.
            Long oldest;
            while((oldest = _votedForExhibits.firstKey()).longValue() < rawSet.first.longValue())
                { _votedForExhibits.remove(oldest); }

            // Return this known-good value.
            return(result);
            }

        // We didn't manage to compute a new/current/up-to-date result,
        // but the caller has allowed us to return a stale result,
        // so return the latest value from the cache, if any.
        assert(allowStale);
        if(!_votedForExhibits.isEmpty())
            {
            while(true)
                {
                final Map<String, Factor> result = _votedForExhibits.get(_votedForExhibits.lastKey());
                if(null != result) { return(result); }
                }
            }

        // The cache was empty so rethrow the error that we got earlier
        // when trying to compute a new/current/up-to-date value.
        assert(err != null);
        throw err;
        }

    /**The (small) immutable current fixed set of parameterless base Scorer instances.
     * Note that since these are fixed and immutable and have no parameters,
     * we can safely create and store instances here.
     * <p>
     * These exist partly for test/calibration purposes.
     */
    public static final Map<String, ScorerIF> fixedSimpleScorers;
    /**Initialise fixedSimpleScorers. */
    static
        {
        final ScorerIF simpleScorers[] =
            {
            // Bogus/fake/test Scorers.
            new NoConfidence(),
            new RandomScore(),
            new RandomPositiveScore(),
            new ZeroScore(),
            new MinScore(),
            new MaxScore(),
            new FixedScore(),

            // True Scorers.
            new SimpleExposure(),
            new LocalSampler(),
            };
        // We use a HashMap for speed of lookup.
        final HashMap<String, ScorerIF> m = new HashMap<String, ScorerIF>(2 * simpleScorers.length);
        for(final ScorerIF sc : simpleScorers)
            { m.put(sc.getName(), sc); }
        fixedSimpleScorers = Collections.unmodifiableMap(m); // Make read-only.
        }

    /**Minimum desireable site of set of voted-for exhibits for calibration to be fully confident of the result; strictly positive. */
    private static final int MIN_VOTED_FOR_SET_SIZE = 1024;

    /**Base set of available scorers name (no parameters); never null but may be empty.
     * The values returned are of the form <i>ScorerName</i>.
     */
    public Set<String> getBaseScorersWithoutParameters()
        { return(fixedSimpleScorers.keySet()); }

    /**Current set of best-available scorers with their parameters; never null but may be empty.
     * This operation should always be relatively fast.
     */
    public Set<String> getCurrentScorersWithParameters(final boolean allowStale)
        { return(population.getCurrentScorersWithParameters(allowStale)); }

    /**Get Scorer instance given the Scorer{:value=name}* format; null if no such Scorer available.
     * Any instance returned is likely to be a shared/cached instance
     * rather than a new instance.
     */
    private ScorerIF getScorerInstance(final String nameAndParameters)
        {
        final Pair<String, Map<String, String>> nap = AbstractScorer.parseNameAndParameters(nameAndParameters);

        // Get a base instance of the Scorer, if any.
        final ScorerIF sc = fixedSimpleScorers.get(nap.first);
        if(sc == null) { return(null); }

        // Fast-track the simple no-parameters case.
        if(nap.second.isEmpty()) { return(sc); }

        try
            {
            // Get the String constructor for name-and-parameters.
            final Constructor<? extends ScorerIF> constructor = sc.getClass().getConstructor(new Class[] { String.class });
            // Attempt to create a new instance with the specified parameters.
            return(constructor.newInstance(new Object[] { nameAndParameters } ));
            }
        // Return null if no String constructor (or no access to it) or any other error.
        catch(final Exception e) { return(null); }        
        }

    /**Get current population size; non-negative. */
    public int size() { return(population.size()); }


    /**Expected approximate poll cycle time (ms); strictly positive. */
    private static final int APPROX_POLL_CYCLE_MS = 1000;

    /**Poll periodically (of the order of a second) to do background tasks.
     * This launches any significant work in a low-proprity daemon thread,
     * and limits the number of such concurrent work threads globally.
     */
    public final void poll()
        throws IOException
        {
        ThreadUtils.lowPriorityThreadPoolDiscardable.submit(new Runnable(){
            public final void run()
                {
                // Set a time to aim to stop processing by on this run.
                //
                // This is usually greater than the expected cycle time
                // to try to enable multiple threads to run in parallel on a multi-CPU server
                // (at minimum priority to be friendly to other processing)
                // up to the limit imposed by the thread pool.
                // However, on a single-CPU box (TODO: and/or if system load is high)
                // then we run for much less than the expected cycle time
                // to try to avoid hogging all the CPU, especially on a single-CPU server.
                final long endTime = System.currentTimeMillis() + 1 +
                    ((Runtime.getRuntime().availableProcessors() > 1) ?
                            (APPROX_POLL_CYCLE_MS*2) : (APPROX_POLL_CYCLE_MS/2));

                // Make sure that we get base Scorers up to date and keep them so.
                // This will also ensure that all sorts of supporting data is kept fresh too.
                final Set<String> scorerBaseNames = getBaseScorersWithoutParameters();
                final ArrayList<String> shuffledBaseNames = new ArrayList<String>(scorerBaseNames);
                Collections.shuffle(shuffledBaseNames, Rnd.fastRnd);
                for(final String scorerName : shuffledBaseNames)
                    {
                    // Attempt to bring the selected Scorer weighting fully up-to-date
                    // if not already so.
                    try { computeScorerWeighting(scorerName, false); }
                    catch(final IOException e) { /* Ignore benign/expected reports. */ }
                    catch(final Exception e)
                        {
                        // Shouldn't get anything other than IOExceptions normally...
                        e.printStackTrace();
                        System.err.println("Problem with computeScorerWeighting("+scorerName+", false): " + e.getMessage());
                        }
                    // Stop and exit this run immediately if we've exhausted our time chunk...
                    // This forces us to get the basics in place and up-to-date first.
                    if(System.currentTimeMillis() > endTime) { return; }
                    }

                // Try to import/create at least one good/saved Scorer from local or global state.
                // Try more imports if we've nowhere-near filled up our population/gene pool yet
                // or computation is going very quickly.
                final int rounds = Math.min(1+SystemVariables.EVENT_SAMPLES_RETAINED,
                        Math.max(2, (MAX_SCORER_SCORES_RETAINED / (1+population.size()))));
                int roundsLeft = rounds; // Decremented on successful creation/import of new variant.
                while((roundsLeft > 0) && (System.currentTimeMillis() <= endTime))
                    {
                    // Attempt to create new Scorer variations GA-style
                    // or import extant Scorers from local/global event histories.
                    final boolean create = Rnd.fastRnd.nextBoolean(); // Create new Scorer (else import).
                    if(create)
                        {
                        // Using the shuffled base-name list,
                        // find the first (parameterisable) Scorer that has some variants
                        // (including the base/default variant)
                        // and try to create new variants from them by perturbing the parameters.
                        for(final String scorerName : shuffledBaseNames)
                            {
                            final ScorerIF sc = fixedSimpleScorers.get(scorerName);
                            final Collection<ScorerParam> parameterDefs = sc.getParameterDefsAndValues();
                            if(parameterDefs.isEmpty()) { continue; } // No variation possible.
                            final List<String> breedingSet = population.getBreedingSet(scorerName);
                            if(breedingSet.isEmpty()) { continue; }

                            // Randomly select one or more different Scorers to mutate and combine.
                            final Set<String> parents = new HashSet<String>();
                            do
                                {
                                // Randomly select a Scorer, skewed towards "better" Scorers.
                                final String extant = breedingSet.get(
                                        Rnd.goodRnd.nextInt(1 + Rnd.goodRnd.nextInt(1 + breedingSet.size())));
                                // Add this mutant if different to the other parents (eg if none so far).
                                parents.add(extant);
                                // Stop if this parent has zero or one parameters,
                                // since multiple parents cannot be usefully "combined" sexually.
                                if(extant.indexOf(AbstractScorer.SEPARATOR) == extant.lastIndexOf(AbstractScorer.SEPARATOR))
                                    { break; }
                                // We try not to draw too many samples from a small breeding set
                                // and we keep low the probability of using many parents.
                                } while((parents.size() < breedingSet.size()/4) && Rnd.goodRnd.nextBoolean());
                            assert(!parents.isEmpty());

                            // Now make the new new variant/child by combining the parents if more than one
                            // and mutating the result.
                            // Else if there is exactly one parent then mutate it to get the new child.
                            final String newVariant;
                            if(parents.size() == 1)
                                { newVariant = getScorerInstance(parents.iterator().next()).getPerturbedNameAndParameters(); }
                            else
                                {
                                // Combine the parents using simple String splicing
                                // and mutate the result/offspring.
                                final ArrayList<String[]> pt = new ArrayList<String[]>(parents.size());
                                for(final String snp : parents)
                                    { pt.add(AbstractScorer.parameterSepRegex.split(snp)); }
                                // Shuffle once to get a completely random order.
                                Collections.shuffle(pt, Rnd.goodRnd);
                                // Now, add one token at a time from the first parent,
                                // occasionally shuffling to switch the source of tokens
                                // to be from a different parent.
                                // We stop when the current/first parent is out of tokens.
                                // We force at least one crossover.
                                final StringBuilder result = new StringBuilder();
                                final int pt0len = pt.get(0).length;
                                final int forcedCrossoverAfter = (pt0len <= 3) ? 1 : (1+Rnd.goodRnd.nextInt(pt0len-2));
                                for(int tokenNumber = 0; tokenNumber < pt0len; ++tokenNumber)
                                    {
                                    if(result.length() != 0) { result.append(AbstractScorer.SEPARATOR); }
                                    result.append(pt.get(0)[tokenNumber]);
                                    // Occasionally switch to another parent.
                                    // We make it quite unlikely to switch/crossover at any one point
                                    // so as to give some notion of locality/clustering.
                                    // FIXME: The frequency/probability of crossover may be very important.
                                    if((forcedCrossoverAfter == tokenNumber) ||
                                       (Rnd.fastRnd.nextBoolean() && Rnd.goodRnd.nextBoolean()))
                                        {
                                        final int ptSize = pt.size();
                                        // Swap the current 0 item (which is the source of "genes")
                                        // randomly with one of the others.
                                        // This is analogous to a DNA crossover during meiosis.
                                        final int swapIndex = (ptSize == 2) ? 1 : 1+Rnd.goodRnd.nextInt(ptSize-1);
                                        final String[] old0 = pt.get(0);
                                        pt.set(0, pt.get(swapIndex));
                                        pt.set(swapIndex, old0);
                                        }
                                    }
                                final String childSNP = result.toString();
                                // Get the un-mutated child.
                                final ScorerIF child = getScorerInstance(childSNP);
                                // Always mutate the child if identical to one parent
                                // or already exists in the population.
                                newVariant = (!parents.contains(childSNP) && (population.getScorerWeighting(childSNP, true) == null)) ?
                                    child.getNameAndParameters() /* Already new. */ :
                                    child.getPerturbedNameAndParameters();
                                }

                            // Check that this is new (or at least not rescored recently).
                            if(population.getScorerWeighting(newVariant, true) != null) { continue; }

if(IsDebug.isDebug) { log.log("[ScorerCacheImpl.poll(): trying new variant: " + newVariant + " of " + (new ArrayList<String>(parents)) + "...]"); }

                            // Evaluate new variant, automatically recording/sharing if good.
                            try
                                {
                                final ScoreAndConf sac = computeScorerWeighting(newVariant, false);
                                --roundsLeft; // Successfully created and evaluated new creature.
if(IsDebug.isDebug) { log.log("[ScorerCacheImpl.poll(): ...variant ScoreAndConfidence: " + sac + ".]"); }
                                }
                            catch(final IOException e) { break; /* Stop this run if we see an error. */ }
                            }           
                        }
                    else
                        {
                        // Choose between the longest and a less long period
                        // to try to pick up more variety of result.
                        final EventPeriod intervalSelector = Rnd.fastRnd.nextBoolean() ?
                                EventPeriod.LONG : EventPeriod.VLONG;
                        // Choose a period from the previous full period to back as far as possible
                        // (for the given period/interval length)
                        // to improve long-term memory of Scorer values tests
                        // and to increase the chance of finding something rare,
                        // but with the choice heavily skewed to more recent events
                        // as more likely to be "better" or even present at all.
                        final long whichPeriod = intervalSelector.getIntervalNumber(System.currentTimeMillis()) -
                            Rnd.goodRnd.nextInt(1+Rnd.goodRnd.nextInt(1+Rnd.goodRnd.nextInt(SystemVariables.EVENT_SAMPLES_RETAINED)));
                        // Select from local (larger volume) and global (broadcast) results.
                        final SimpleVariableDefinition def = Rnd.fastRnd.nextBoolean() ?
                                SystemVariables.AI_SCORER_STRING_LOCAL_EVENT : SystemVariables.AI_SCORER_STRING_GLOBAL_EVENT;
                        try
                            {
                            final EventVariableValue[] evvs = dataSource.getEventValues(def, intervalSelector, whichPeriod, null);
                            if(evvs.length == 0) { continue; }
                            EventVariableValue evv = evvs[0];
                            // If the selected (full, historical) period has no events
                            // then try the current (partial) period instead,
                            // which has the advantage of getting us bang up-to-date.
                            if((evv == null) || (evv.getTotalDistinctValues() == 0))
                                { evv = dataSource.getEventValue(def, intervalSelector, true); }
                            if((evv == null) || (evv.getTotalDistinctValues() == 0)) { continue; }
                            final ArrayList<Object> vals = new ArrayList<Object>(evv.getInfo().keySet());
                            if(vals.isEmpty()) { continue; }
                            // Select an entry at random.
                            final String scorerNameAndStringParameters = (String) vals.get(Rnd.goodRnd.nextInt(vals.size()));
                            // Check that it is new...
                            if(population.getScorerWeighting(scorerNameAndStringParameters, true) != null) { continue; }
                            // Evaluate/record it if possible.
                            final ScoreAndConf sac = computeScorerWeighting(scorerNameAndStringParameters, false);
                            --roundsLeft; // Successfully imported and evaluated new creature.
if(IsDebug.isDebug) { log.log("[ScorerCacheImpl.poll(): imported "+scorerNameAndStringParameters+" ScoreAndConfidence: "+sac+".]"); }
                            }
                        catch(final IOException e) { break; /* Stop this run after an error. */ }
                        catch(final Exception e) { e.printStackTrace(); break; /* Shouldn't happen. */ }
                        }
                    }
                }
            });
        }

    /**Save work-in-progress if possible, and free up resources, ASAP.
     * This may enable us to reduce work lost during a graceful system shutdown,
     * but many shutdowns may not be graceful and so we should incrementally save/checkpoint too.
     */
    public final void destroy()
        {
        }
    }
