/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.awt.image.RenderedImage;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 18-Apr-2006
 * Time: 20:57:40
 */

/**Base interface to compute the score and confidence for an exhibit.
 * All methods in this interface are guaranteed to be "safe"
 * insofaras they will complete in "reasonable" time
 * with reasonable heap memory (and other resource, eg stack)
 * and without doing anything that wouldn't be allowed in
 * a minimal Applet/JWS sandbox.
 * <p>
 * All classes implementing this interface should be completely thread-safe
 * in their implementation of the computeScoreAndConfidence() method,
 * and preferrably purely functional (no visible side-effects),
 * with as many concurrent threads as required safely doing separate computations
 * in any one instance.
 * <p>
 * Classes implementing this interface should, where possible,
 * do their calculations using integer arithmetic,
 * since FPUs to support float/double calaculations may be a scarce resource
 * on newer highly-threaded CPUs such as Sun's Niagara.
 */
public interface ScorerIF
    {
    /**Compute score [-1,+1] and confidence[0,+1] for given exhibit; never null.
     *
     * @param dataSource  souce of exhibit data and metadata; never null
     * @param exhibitName  name of exhibit; must be syntactically valid and not null
     * 
     * @throws IOException in case of difficulty fetching the exhibit details/data
     *     or in computing the result
     */
    ScoreAndConf computeScoreAndConfidence(SimpleExhibitPipelineIF dataSource,
                                           String exhibitName)
        throws IOException;

    /**Get name of this Scorer (excluding any parameters); never null.
     * Pure alphanumeric, essentially same rules as for a Java Class identifier
     * (but with '$' and '_' and '.' excluded), ie <code>[a-zA-Z][a-zA-Z0-9]*</code>.
     */
    String getName();

    /**Get name of this Scorer and any parameters; never null.
     * If there are any parameters then they appended with the form
     * <code>(:name=value)*</code>.
     */
    String getNameAndParameters();

    /**Computes randomly perturbed parameters (if any); never null.
     * This returns the name and parameters (if any of the current Scorer)
     * with zero or more of any parameters altered by a small amount
     * and guaranteed to be valid (in range).
     */
    String getPerturbedNameAndParameters();

    /**Get (immutable, scorers) parameter definitions and current values for this Scorer; never null. */
    Collection<ScorerParam> getParameterDefsAndValues();
    }
