/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import org.hd.d.pg2k.svrCore.Rnd;

/**Immutable (enum) Scorer parameter.
 * The immutablity of this class/instance depends on that of the enum.
 */
public final class ScorerParamEnum<E extends Enum<E>> implements ScorerParam
    {
    /**Construct with all components, validated and stored as-is. */
    public ScorerParamEnum(final E def,
                           final String name, final E value)
        {
        if(!AbstractScorer.isValidParameterName(name)) { throw new IllegalArgumentException(); }
        if((def == null) || (value == null)) { throw new IllegalArgumentException(); }
        if(def.getClass() != value.getClass()) { throw new IllegalArgumentException(); }
        this.name = name;
        this.value = value;
        this.def = def;
        }

    /**Construct with the default value. */
    public ScorerParamEnum(final E def,
                           final String name)
        { this(def, name, def); }

    /**Parameter name with allowed syntax as for Java variable name; never null. */
    public final String name;
    /**Parameter value from the enumeration; non-null. */
    public final E value;

    /**Default value for this parameter from the enumeration; non-null. */
    public final E def;

    /**For an enumerated parameter, "perturbing" involves selecting a new value at random.
     */
    @SuppressWarnings("unchecked")
    public ScorerParam perturb()
        {
        final Enum[] enumConstants = def.getClass().getEnumConstants();
        assert((enumConstants != null) && (enumConstants.length > 0));
        final E en = (E) enumConstants[Rnd.goodRnd.nextInt(enumConstants.length)];
        if(en == value) { return(this); } // Return unchanged if same value.
        return(new ScorerParamEnum<E>(def, name, en));
        }


    /**Parse the String representation of a parameter returning a value of the same type; never null.
     * If the argument is null or unparsable or invalid,
     * then the result has the default value,
     * else an instance with the requested parameter value is returned.
     * @return legal value of the same type as this, with the parsed input String's value if possible
     */
    @SuppressWarnings("unchecked")
    public ScorerParam parse(final String v)
        {
        E parsedValue = def; // Use default value unless the parse is successful.
        if(v != null)
            {
            try { parsedValue = E.valueOf((Class<E>) def.getClass(), v); }
            catch(final IllegalArgumentException e) { /* Ignore. */ }
            }

        // Memory optimisation where value is the same as for the current instance.
        if(parsedValue == value) { return(this); }

        // Create new instance containing the parsed value.
        return(new ScorerParamEnum<E>(def, name, parsedValue));
        }


    /**Generate the "name=value" text; never null or empty, though the value part may be empty. */
    public String toNameValueString()
        {
        final StringBuilder sb = new StringBuilder(16 + name.length());
        sb.append(name).append('=').append(value);
        return(sb.toString());
        }

    /**Human-readable representation. */
    public String toString() { return(toNameValueString()); }
    }
