/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import org.hd.d.pg2k.svrCore.Rnd;

/**Immutable (int) Scorer parameter.
 */
public final class ScorerParamInteger implements ScorerParam
    {
    /**Construct with all components, validated and stored as-is. */
    public ScorerParamInteger(final int min, final int def, final int max,
                              final int delta, final boolean biasedLow,
                              final String name, final int value)
        {
        if(!AbstractScorer.isValidParameterName(name)) { throw new IllegalArgumentException(); }
        if((min > def) || (def > max)) { throw new IllegalArgumentException(); }
        if((value < min) || (value > max)) { throw new IllegalArgumentException(); }
        if(delta <= 0) { throw new IllegalArgumentException(); }
        this.name = name;
        this.value = value;
        this.min = min;
        this.def = def;
        this.max = max;
        this.delta = delta;
        this.biasedLow = biasedLow;
        }

    /**Construct with the default value. */
    public ScorerParamInteger(final int min, final int def, final int max,
                              final int delta, final boolean biasedLow,
                              final String name)
        { this(min, def, max, delta, biasedLow, name, def); }

    /**Construct with the default value, no bias and a unit delta. */
    public ScorerParamInteger(final int min, final int def, final int max,
                              final String name)
        { this(min, def, max, 1, false, name, def); }

    /**Parameter name with allowed syntax as for Java variable name; never null. */
    public final String name;
    /**Parameter value in inclusive/closed range [min,max]. */
    public final int value;

    /**Minimum legal value for this parameter; no larger than max. */
    public final int min;
    /**Default value for this parameter; no smaller than min, no larger than max. */
    public final int def;
    /**Maximum legal value for this parameter; no smaller than min. */
    public final int max;

    /**The size of a small change during mutation; strictly positive. */
    public final int delta;
    /**If true, mutations are biased towards reducing the parameter value. */
    public final boolean biasedLow;

    /* (non-Javadoc)
     * @see org.hd.d.pg2k.ai.scorer.ScorerParam#perturb()
     */
    public ScorerParam perturb()
        {
        final boolean down = (!biasedLow) ? Rnd.goodRnd.nextBoolean() :
            (Rnd.goodRnd.nextInt(3) != 0); // Apply bias.
        // For delta > 1, pick random value up to given delta.
        final int deltaToUse = (delta == 1) ? delta : (Rnd.goodRnd.nextInt(delta)+1);
        final int newValue = down ? (value-deltaToUse) : (value+deltaToUse);
        // Return existing value if not actually different to save a little heap churn.
        if(newValue == value) { return(this); }
        return(new ScorerParamInteger(min, def, max, delta, biasedLow, name,
                Math.max(min, Math.min(max, newValue))));
        }


    /**Parse the String representation of a parameter returning a value of the same type; never null.
     * If the argument is null or unparsable or out of range,
     * then the result has the default value,
     * else an instance with the requested parameter value is returned.
     * @return legal value of the same type as this, with the parsed input String's value if possible
     */
    public ScorerParam parse(final String v)
        {
        int parsedValue = def; // Use default value unless the parse is successful.
        if(v != null)
            {
            try
                {
                final int p = Integer.parseInt(v);
                if((p >= min) && (p <= max)) { parsedValue = p; }
                }
            catch(final NumberFormatException e) { /* Ignore. */ }
            }

        // Memory optimisation where value is the same as for the current instance.
        if(parsedValue == value) { return(this); }

        // Create new instance containing the parsed value.
        return(new ScorerParamInteger(min, def, max, delta, biasedLow, name, parsedValue));
        }


    /**Generate the "name=value" text; never null or empty, though the value part may be empty. */
    public String toNameValueString()
        {
        final StringBuilder sb = new StringBuilder(12 + name.length());
        sb.append(name).append('=').append(value);
        return(sb.toString());
        }

    /**Human-readable representation. */
    public String toString() { return(toNameValueString()); }
    }
