/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer;

import java.io.IOException;
import java.util.*;

import org.hd.d.pg2k.svrCore.GenUtils;
import org.hd.d.pg2k.svrCore.SimpleLoggerIF;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.vars.*;

import ORG.hd.d.IsDebug;


/**Thread-safe container for Scorer-and-parameter information to maintain an evolving population.
 * This separates Scorers by basic (parameterless) name,
 * on the grounds that they are not directly comparable.
 * <p>
 * This has limited capacity (so as to cap resource use),
 * and discards the least good parameterised Scorers as needed
 * to make way for new (better) instances.
 * <p>
 * This has the automatially persists/announces/shares new/good instances
 * using the system variables events mechanism, local and global.
 * <p>
 * TODO: in case of accute memory shortage abandon least-good Scorers
 */
public final class ScorerPopulation
    {
    /**Construct with a list of the base Scorers (names, no parameters) to keep track of.
     * Note that we may temporarily exceed the maximum population size for expediency,
     * and we may dump some (less good) part of the population under memory stress.
     *
     * @param baseScorers  non-null, non-empty, pure-base-name seed set
     * @param maxPopulationSize  maximum number of Scorer details to retain in memory,
     *     possibly excluding the fixed/base Scorers, strictly positive
     */
    public ScorerPopulation(final Collection<String> baseScorers,
                            final int maxPopulationSize)
        {
        if(maxPopulationSize <= 0) { throw new IllegalArgumentException(); }
        this.maxPopulationSize = maxPopulationSize;

//        if(dataSource == null) { throw new IllegalArgumentException(); }
//        this.dataSource = dataSource;

        if(baseScorers == null) { throw new IllegalArgumentException(); }
        // Make immutable (de-duped) set.
        baseNames = Collections.unmodifiableSet(new HashSet<String>(baseScorers));
        if(baseNames.isEmpty()) { throw new IllegalArgumentException(); }
        // Validate the Scorer base names.
        for(final String n : baseScorers)
            {
            // Validate the name, and make sure that there are no parameters.
            if(!AbstractScorer.parseNameAndParameters(n).second.isEmpty())
                { throw new IllegalArgumentException("no parameters allowed on seed set"); }
            }

        // Create to accomodate the specified capacity without further expansion.
        _scorerScores = new HashMap<String, Pair<Long, ScoreAndConf>>(2*maxPopulationSize);

        // Create the (immutable) outer map with empty sub-maps for _scorerByGoodness.
        // The comparator arranges for the items be be sorted by increasing "goodness".
        // Because of a potential lack of symmetry in scoring
        // (ie reversing the value of a bad Scorer does not necessarily make it a good one)
        // we are most interested in the highest-confidence, positive-correlation Scorers.
        final Comparator<String> sbgComp = new Comparator<String>()
            {
            // This will fail if the items are not both present in _scorerScores.
            public final int compare(final String a, final String b)
                {
                final int goodnessA = computeScorerGoodness(_scorerScores.get(a).second);
                final int goodnessB = computeScorerGoodness(_scorerScores.get(b).second);
                if(goodnessA < goodnessB) { return(-1); }
                if(goodnessA > goodnessB) { return(+1); }
                return(a.compareTo(b)); // Break ties by name for total ordering.
                }
            };
        final Map<String, NavigableSet<String>> sbg = new HashMap<String, NavigableSet<String>>(2 * baseScorers.size());
        for(final String n : baseScorers)
            { sbg.put(n, new TreeSet<String>(sbgComp)); }
        _scorerByGoodness = Collections.unmodifiableMap(sbg);
        }

    /**Maximum number of Scorer details to retain, possibly excluding the fixed/base Scorers, strictly positive. */
    private final int maxPopulationSize;

//    /**Live data source for variables, exhibit data and metadata, etc; never null. */
//    private final SimpleExhibitPipelineIF dataSource;

    /**Immutable Set of base Scorer names (no parameters) to monitor/maintain; non-null, non-empty. */
    private final Set<String> baseNames;

    /**Non-user-visible lock to serialise activity on _scorerScores and _scorerByGoodness; never null.
     * No blocking/external calls are made in the scope of this lock
     * so as to ensure that it is held only for very short intervals.
     */
    private final Object _lock = new Object();

    /**Cache of Scorer scores (and time cached), indiectly size-limited; never null.
     * Map from scorer name-and-parameters to ScoreAndConfidence and time computed.
     * <p>
     * Not thread-safe; access only under _lock.
     */
    private final Map<String, Pair<Long, ScoreAndConf>> _scorerScores;

    /**Map from base Scorer name to a NavigableSet of mappings from full nameAndParameters to goodness; never null.
     * We pre-populate this with empty maps under each base name that we handle
     * (so it is never necessary to test for presence in, or ever alter, the outer map; hence it is immutable).
     * <p>
     * The top-level Comparator is selected to sort by goodness derived from the _scorerScores,
     * so any extant entry in the nameAndParameters should be removed from the _scorerByGoodness map
     * before removal from or replacement in the _scorerScores map,
     * and re-inserted afterwards as necessary.
     * <p>
     * Since the different base Scorer names/types are held separately,
     * this helps keep the probably-distinct populations from competing with one another,
     * though the overall population size cap is still present.
     * <p>
     * Because of a potential lack of symmetry in scoring
     * (ie reversing the value of a bad Scorer does not necessarily make it a good one)
     * we are interested in preserving the highest-confidence, positive-correlation Scorers.
     * <p>
     * Not thread-safe; access only under _lock.
     */
    private final Map<String, NavigableSet<String>> _scorerByGoodness;

    /**ScoreAndConfidence for the given scorer itself over all exhibit types; null if not known.
     * Never recomputes, always fetches from its cache if present else returns null.
     * <p>
     * Older results (typically up to 1 day older) may be returned if allowStale is true;
     * this may help the system overall ride out temporary I/O problems and disconnections.
     * <p>
     * These get operations are generally fast,
     * though some cycle-stealing incremental housekeeping might be performed on this call.
     *
     * @param scorerNameAndParameters  the name and parameters of the scorer; never null
     * @param allowStale  if true then allow an older cache response
     *     (up to VLONG period older)
     *
     * @return  the score represents the correlation with the underlying votes
     *     (and whatever the scoring is measured against)
     *     with MAX meaning perfect correlation, 0 meaning no correlation,
     *     and -MAX meaning perfectly wrong answers all the time,
     *     and the confidence 0 if we have no (or very few) data points yet
     *     and approaching MAX as we have a large (enough) number of data points
     */
    public ScoreAndConf getScorerWeighting(final String scorerNameAndParameters,
                                           final boolean allowStale)
        {
        if(scorerNameAndParameters == null) { throw new IllegalArgumentException(); }
 
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());

        // Use lock for thread-safety.
        synchronized(_lock)
            {
            // We don't need to validate the requested name (other than as being non-null)
            // since an invalid value will correctly return null.
            final Pair<Long, ScoreAndConf> cachedValue = _scorerScores.get(scorerNameAndParameters);
            // We can use a value from cache if the cached value is from the current period
            //     (or the previous period if allowStale is true).
            if((cachedValue != null) &&
               (EventPeriod.VLONG.getIntervalNumber(cachedValue.first) >= (allowStale ? currentPeriod-1 : currentPeriod)))
                { return(cachedValue.second); }
            }

        return(null); // Nothing in cache.
        }

    /**GenProps 'generic' parameter that must be set for us to post to the global event store. */
    public static final String GEN_PROPS_GLOBAL_POST_ENABLE_SUFFIX = "ai.scorerGlobalPostEnable";

    /**Save the ScoreAndConfidence for the given scorer itself over all exhibit types.
     * Saves the given Scorer and "goodness",
     * if necessary evicting a poorly-performing Scorer to make room
     * ie to limit the population size.
     * <p>
     * If the "put" Scorer is very good compared to the existing population members
     * then it may get persisted/shared via the event mechanism.
     * <p>
     * This routine will only consider storing entries for Scorers whose base name
     * is on the list we were constructed with.
     * @param nameAndParameters  the name and parameters of the scorer; never null
     * @param sac  rating of this Scorer across all exhibits; never null
     * @param dataSource  to record/persist/share good Scorers through; never null
     * @param log logger; never null
     *
     * @return  true if the Scorer result was stored internally,
     *     false if it was not appropriate or good enough to retain
     */
    public boolean putScorerWeighting(final String nameAndParameters,
                                      final ScoreAndConf sac,
                                      final SimpleExhibitPipelineIF dataSource,
                                      final SimpleLoggerIF log)
        {
        if(sac == null) { throw new IllegalArgumentException(); }
        if(nameAndParameters == null) { throw new IllegalArgumentException(); }

        final Pair<String, Map<String, String>> nap = AbstractScorer.parseNameAndParameters(nameAndParameters);
        final String baseName = nap.first;
        // Don't store if not one of our recognised base Scorers.
        if(!baseNames.contains(baseName)) { return(false); }
        final int goodness = computeScorerGoodness(sac);
        final boolean parameterless = nap.second.isEmpty();

        final long startTime = System.currentTimeMillis();
//        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(startTime);

        // Set true if we discover that this is the best Scorer of its type so far
        // so we should save/share it.
        boolean isBest = false;

        // We set this false if we discover that we could not retain it.
        boolean isStored = true;
 
        // Still a candidate for storing, so grab a lock...
        synchronized(_lock)
            {
            // See if we have an existing entry for this nameAndParameters.
            final Pair<Long, ScoreAndConf> oldVal = _scorerScores.get(nameAndParameters);
            final SortedSet<String> insertedScorerSubPopulation = _scorerByGoodness.get(baseName);
            // If we do, then remove the corresponding entry from _scorerByGoodness first.
            if(oldVal != null)
                {
                assert(insertedScorerSubPopulation.contains(nameAndParameters));
                insertedScorerSubPopulation.remove(nameAndParameters);
                assert(!insertedScorerSubPopulation.contains(nameAndParameters));
                }

            // Insert the value: first in _scorerScores, then in the dependent _scorerByGoodness.
            _scorerScores.put(nameAndParameters, new Pair<Long, ScoreAndConf>(startTime, sac));
            insertedScorerSubPopulation.add(nameAndParameters);

            final int bestInSubPop = computeScorerGoodness(_scorerScores.get(insertedScorerSubPopulation.last()).second);
            // Test if this is the best Scorer in its sub-population
            // and has a non-zero goodness.
            // Note that this has already been inserted into that sub-population,
            // so cannot be "better".
            assert(goodness <= bestInSubPop);
            if((goodness > 0) && (goodness == bestInSubPop))
                { isBest = true; }

            // If the population is now oversize
            // then we'll cull the worst Scorer from the largest population if possible.
            if(_scorerScores.size() > maxPopulationSize)
                {
                // Find the largest sub-population.
                int largestSubPopulation = -1;
                String largestSubPopulationBasename = null;
                for(final String subPopBaseName : _scorerByGoodness.keySet())
                    {
                    final int subPopSize = _scorerByGoodness.get(subPopBaseName).size();
                    if(subPopSize > largestSubPopulation)
                        {
                        largestSubPopulation = subPopSize;
                        largestSubPopulationBasename = subPopBaseName;
                        }
                    }
                assert(largestSubPopulation > 0);
                assert(largestSubPopulationBasename != null);

                // Iterate through the largest population
                // culling the first (least good) parameterised Scorer found, if any.
                // We may not be able to remove anything, leaving the population slightly oversized.
                final SortedSet<String> victimSubPop = _scorerByGoodness.get(largestSubPopulationBasename);
                for(final String putativeVictim : victimSubPop)
                    {
                    // We never remove the parameterless base entry.
                    if(putativeVictim.indexOf(AbstractScorer.SEPARATOR) == -1) { continue; }
                    // Zap our victim first from the dependent _scorerByGoodness first.
                    final boolean removedVictimFromSBG = victimSubPop.remove(putativeVictim);
                    assert(removedVictimFromSBG);
                    assert(_scorerScores.containsKey(putativeVictim));
                    _scorerScores.remove(putativeVictim);
                    assert(!_scorerScores.containsKey(putativeVictim));
                    // Change the method return value if we expelled the new item.
                    if(putativeVictim.equals(nameAndParameters))
                        { isStored = false; }
                    break;
                    }
                }
            }

        // If this was the best Scorer (in its sub-population) so far then post it locally
        // [TODO and globally once all servers are upgraded]
        // if not zero-goodness or parameterless.
        // This is done outside of the lock scope
        // to avoid blocking other get/put operations during I/O.
        // We absorb (and report) any I/O error, but continue.
        if(isBest & (goodness > 0) && !parameterless)
            {
log.log("[ScorerPopulation: new best Scorer: "+nameAndParameters+" "+sac+", 'goodness'="+goodness+"...]");
            final SimpleVariableValue svvL = new SimpleVariableValue(SystemVariables.AI_SCORER_STRING_LOCAL_EVENT,
                                                                    nameAndParameters);
            try { dataSource.setVariable(svvL); }
            catch(final IOException e) { e.printStackTrace(); /* Complain but continue after error. */ }
            try
                {
                if(USE_GLOBAL_EVENTS_ALSO &&
                        "true".equalsIgnoreCase(dataSource.getGenProps(-1).getGen().get(GEN_PROPS_GLOBAL_POST_ENABLE_SUFFIX)))
                    {
                    final SimpleVariableValue svvG = new SimpleVariableValue(SystemVariables.AI_SCORER_STRING_GLOBAL_EVENT,
                            nameAndParameters);
                    try { dataSource.setVariable(svvG); }
                    catch(final IOException e) { e.printStackTrace(); /* Complain but continue after error. */ }        
                    }
                }
            catch(final IOException e) { e.printStackTrace(); /* Complain but continue after error. */ }
            }

        return(isStored);
        }

    /**If true, allow posting the global event for exchange of the best/new Scorers.
     * The mechanism must also be enabled in the GenProps for this to happen.
     */
    private static final boolean USE_GLOBAL_EVENTS_ALSO = true;

    /**Get current population size; non-negative.
     * This value may slightly exceed requested maximum.
     */
    public int size() { synchronized(_lock) { return(_scorerScores.size()); } }

    /**Current set of available scorers name and parameters (where applicable); never null but may be empty.
     * The values returned are of the form <i>ScorerName{:name=value}*</i>.
     * <p>
     * The scorers returned by this will generally be the best available,
     * usually the best one or few per base-Scorer type,
     * allowing that different Scorer types may have different "skill" domains,
     * and excluding any with zero goodness.
     */
    public Set<String> getCurrentScorersWithParameters(final boolean allowStale)
        {
        // Build up initial dup-free list.
        // Weed out zero-goodness items before returning.
        final Set<String> r = new HashSet<String>(4 * baseNames.size());

        // Have the number of samples from each base name/type to include
        // be at least 2 if possible,
        // and growing (very) slowly with the number of different types.
        final int maxToInclude = Math.max(2, GenUtils.log2Approx(baseNames.size()));

        // Hold a lock on the data structures while working.
        synchronized(_lock)
            {
            // Attempt to a candidate from each base type (sub population).
            for(final NavigableSet<String> subPop : _scorerByGoodness.values())
                {
                if(!subPop.isEmpty())
                    {
                    // Get the best Scorer of this type and its "goodness".
                    final String best = subPop.last();
                    final int bGoodness = computeScorerGoodness(_scorerScores.get(best).second);
                    if(bGoodness <= 0) { continue; /* All bad! */ }
                    // We'll accept other Scorers nearly as good as the best.
                    final int threshold = bGoodness - (bGoodness>>>3);
                    assert(threshold >= 0);
                    // Work backwards from the best...
                    int remaining = maxToInclude;
                    for(final String snp : subPop.descendingSet())
                        {
                        final int g = computeScorerGoodness(_scorerScores.get(snp).second);
                        // Stop if the goodness falls too far (or to/below zero).
                        if(g <= threshold) { break; }
                        assert(g > 0); // Only keeping Scorers with merit.
                        // This Scorer is good enough.
                        r.add(snp);
                        // We're one closer to our target.
                        if(--remaining <= 0) { break; }
                        }
                    }
                }
//            // Add in all the base (parameterless) Scorers.
//            r.addAll(baseNames);
//
//            // Remove all Scorers without merit or recent cached scores.
//            for(final Iterator<String> it = r.iterator(); it.hasNext(); )
//                {
//                final String scorerNameAndParameters = it.next();
//                final ScoreAndConf sac = getScorerWeighting(scorerNameAndParameters, allowStale);
//                if((null == sac) || (computeScorerGoodness(sac) <= 0))
//                    { it.remove(); }
//                }
            }

        return(r);
        }

    /**Get random-access list of Scorers to "breed" from, best first; never null, possibly empty.
     * This returns all of (or some of the best of) the Scorers with the given base name.
     */
    public List<String> getBreedingSet(final String baseName)
        {
        if(baseName == null) { throw new IllegalArgumentException(); }
        if(!baseNames.contains(baseName)) { throw new IllegalArgumentException(); }

        // Hold lock while accessing mutable data.
        synchronized(_lock)
            {
            // Get in "worst-first" order.
            final ArrayList<String> result = new ArrayList<String>(_scorerByGoodness.get(baseName));
            // Convert to "best-first" order.
            Collections.reverse(result);
            // Eliminate all those with zero or negative goodness.
            for(int i = result.size(); --i >= 0; )
                {
                final int goodness = computeScorerGoodness(_scorerScores.get(result.get(i)).second);
                if(goodness > 0) { break; }
                // Remove this trailing zero-goodness item.
                result.remove(i);
                }
            // Return it to the caller to use.
            return(result);
            }
        }

    /**Compute Scorer "goodness" from its ScoreAndConfidence value.
     * We want Scorers with a good correlation (+ve) and good confidence.
     * <p>
     * This is an internal unitless relative value
     * where -ve is "wrongheaded", 0 is "useless", and +ve is "good".
     */
    private int computeScorerGoodness(final ScoreAndConf sac)
        {
        if(sac.confidence == 0) { return(0); } // Relatively common case.
        return(sac.score * sac.confidence);
        }
    }
