/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer.parameterised;

import java.awt.Color;
import java.awt.Point;
import java.awt.image.*;
import java.util.*;

import org.hd.d.pg2k.ai.scorer.*;
import org.hd.d.pg2k.svrCore.*;

import com.sun.tools.javac.util.List;

/**Simple but configurable Scoring based on local pixel sampling; never null.
 * Since this does not involve any global image operations,
 * running time is nominally independent of image dimensions,
 * ie in all case local image properties are sampled.
 * <p>
 * This can sample various pixel parameters such as R, G and B components,
 * and can use the absolute values or available ranges.
 * <p>
 * The set of sample points is a grid spaced in a regular (though not a simple grid)
 * pattern across the image.  The layout of the sample points depends on the image
 * size and aspect ratio, and is generated from a pseudo-randon generator,
 * to help avoid poor sampling of images with regular patterns in them,
 * <p>
 * The confidence depends on the number of selected sample points that are available.
 * <p>
 * Very small images will have all their pixels sampled,
 * and so can in principle still achieve a MAX confidence.
 * <p>
 * Parameters include:
 * <ul>
 * <li>Pixel sample-set size.
 * <li>Pixel component to sample: A (alpha), R, G, B, Y (luminance), H, S, B.
 * <li>Pixel sampling method: value or median range (on grid centred on specified pixel).
 * <li>Pixel sample combination method: mean or median range.
 * <li>Fraction of outliers to discard for sample-combination median range computations.
 * <li>Mapping interpolation vector from sample combination value to score.
 * <li>TODO: Mapping interpolation vector from sample combination value to confidence.
 * </ul>
 * <p>
 * This may be most useful for continuous-tone true-colour/greyscale (eg JPEG) images.
 * 
 * @author dhd
 */
public final class LocalSampler extends AbstractImgScorer
    {
    /**Create simple non-parameterised instance. */
    public LocalSampler()
        {
        // All parameters get default values.
        sampleSizeParam = sampleSizeParamBounds;
        outlierParam = outlierParamBounds;
        componentParam = componentParamBounds;
        sampleCombinationParam = sampleCombinationParamBounds;
        pixelSampleMethodParam = pixelSampleMethodParamBounds;
        scoreMappingParams = scoreMappingParamsBounds;
        confidenceMappingParams = confidenceMappingParamsBounds;
        }

    /**Create parameterised version. */
    public LocalSampler(final String nameAndParameters)
        {
        super(nameAndParameters);

        // Assign parameters from the captured map where possible, using default values otherwise.
        sampleSizeParam = (ScorerParamInteger) sampleSizeParamBounds.parse(paramValueMap.get(sampleSizeParamBounds.name));
        outlierParam = (ScorerParamInteger) outlierParamBounds.parse(paramValueMap.get(outlierParamBounds.name));
        componentParam = (ScorerParamEnum<PixelComponent>) componentParamBounds.parse(paramValueMap.get(componentParamBounds.name));
        sampleCombinationParam = (ScorerParamEnum<SampleType>) sampleCombinationParamBounds.parse(paramValueMap.get(sampleCombinationParamBounds.name));
        pixelSampleMethodParam = (ScorerParamEnum<SampleType>) pixelSampleMethodParamBounds.parse(paramValueMap.get(pixelSampleMethodParamBounds.name));
        final ArrayList<ScorerParamInteger> s = new ArrayList<ScorerParamInteger>(MAPPING_INTERVALS+1);
        for(final ScorerParamInteger p : scoreMappingParamsBounds)
            { s.add((ScorerParamInteger) p.parse(paramValueMap.get(p.name))); }
        scoreMappingParams = Collections.unmodifiableList(s);
        final ArrayList<ScorerParamInteger> c = new ArrayList<ScorerParamInteger>(MAPPING_INTERVALS+1);
        for(final ScorerParamInteger p : confidenceMappingParamsBounds)
            { c.add((ScorerParamInteger) p.parse(paramValueMap.get(p.name))); }
        confidenceMappingParams = Collections.unmodifiableList(c);
        }

    /**Score the image for "exposure"; never null nor negative in any component.
     * @throws IllegalArgumentException for a null image
     */
    public ScoreAndConf computeScoreAndConfidence(final RenderedImage stillImage)
        {
        if(stillImage == null) { throw new IllegalArgumentException(); }

        final int h = stillImage.getHeight();
        final int w = stillImage.getWidth();

        // True if we will be sampling local pixel "spread".
        final boolean pixelSpread = (pixelSampleMethodParam.value == SampleType.RANGE);
        // We need slightly larger minimal image size to sample pixel spread.
        if(pixelSpread && ((h < 3) || (w < 3))) { return(ScoreAndConf.NO_OPINION); }

        // Don't try to score strange (eg zero-sized) images.
        if((h < 1) || (w < 1)) { return(ScoreAndConf.NO_OPINION); }

        final long availablePixels = h * w;
        assert(availablePixels > 0);

        // If the image is small we may simply examine every pixel.
        final int target_sample_points = (1 << sampleSizeParam.value);
        final boolean smallImage = (availablePixels <= target_sample_points);
        final int maxSampleablePoints = smallImage ? (int) availablePixels : target_sample_points;

        final int minX = stillImage.getMinX();
        final int minY = stillImage.getMinY();

        // For sampling we'll need a buffered image;
        // so we copy the input image if need be.
        final BufferedImage bi = ImageUtils.convertRenderedImageToBufferedImage(stillImage);

        // Set of values sampled from the image (when used for "range" results.
        // Assumed to be sortable, ie all values finite.
        final ArrayList<Byte> pointsSampled = new ArrayList<Byte>(maxSampleablePoints);

        // For small images sample all pixels.
        if(smallImage)
            {
            for(int x = w; --x >= 0; )
                {
                for(int y = h; --y >= 0; )
                    {
                    final Byte b = pixelSpread ? samplePointRange(bi, x + minX, y + minY) : samplePoint(bi, x + minX, y + minY);
                    if(b == null) { continue; }
                    pointsSampled.add(b);
                    }
                }
            }
        // For reasonable-size (not tiny) images,
        // we sample pixels in a pseudo-random pattern based on image size/shape.
        else
            {
            // Make the sample pattern (random seen) dependent only on image size/shape
            // to help circumvent some systematic sample pattern problems.
            // We do not make the sample pattern/seed dependent on sample size
            // so that as we take more samples all previous values are included.
            // We create a different seed to that used in SimpleExposure.
            final Random r = new Random(37 + h + (((long)w) << 17));

            // To allow in the main for non-square partially-transparent images
            // but also for images close in pixels to our threshold size
            // (so that we capture as many of the samplable points as reasonable possible)
            // we are prepared to attempt the sampling operation
            // more than the target number of samples
            // (though we stop as soon as our target sample count is reached).
            // We also reject would-be duplicate sampling of the same point.
            assert((new Point(0, 0)).equals(new Point(0, 0)));
            final Set<Point> alreadySampled = new HashSet<Point>(1 + 2*target_sample_points);
            for(int i = 2*target_sample_points; (--i >= 0) && (alreadySampled.size() < target_sample_points); )
                {
                // Choose a (new) random point and take a sample.
                final int x = r.nextInt(w);
                final int y = r.nextInt(h);
                final Point p = new Point(x, y);
                if(!alreadySampled.add(p)) { continue; } // Skip if point not new...
//System.out.println("[Sample point: "+p+".]");
                final Byte b = pixelSpread ? samplePointRange(bi, x + minX, y + minY) : samplePoint(bi, x + minX, y + minY);
                if(b == null) { continue; }
//System.out.println("[Point accepted.]");
                pointsSampled.add(b);
                }
            }

        final int nSamples = pointsSampled.size();
        // If we are to discard any outliers
        // and have at least two points left to compute the range with
        // then we must have collected at least 4 samples.
        // If we don't have at least this many, then return a "no confidence" result.
        if(nSamples < 4) { return(ScoreAndConf.NO_OPINION); }
        // Compute number of outliers to discard from either extreme.
        final int nOutliers = Math.max(1, nSamples / (1 << outlierParam.value));
        final int medianSamples = nSamples - 2*nOutliers;
        assert(medianSamples >= 2);

        // Sort the sampled points, which allows us to
        // discard/ignore outliers from either end of the range, likely to be noise.
        // Compute the relative interval between the remaining extremes to compute the score.
        // The confidence is based on the sample size and the score.
        Collections.sort(pointsSampled);

        // Compute the raw value [0,MAX_COMPONENT_VALUE] before mapping, eg mean or range.
        final int value;
        if(sampleCombinationParam.value == SampleType.RANGE)
            {
            final int lowest = pointsSampled.get(nOutliers).intValue();
            final int highest = pointsSampled.get(nSamples-1 - nOutliers).intValue();
//    System.out.println("SimpleExposure: low/high = "+lowest+"/"+highest + "; lowest/highest = "+pointsSampled.get(0)+"/"+pointsSampled.get(nSamples-1));
            value = highest - lowest;
            }
        else
            {
            int sum = 0;
            for(int i = nSamples - nOutliers; --i >= nOutliers; )
                { sum += pointsSampled.get(i); }
            value = sum / medianSamples;
            }
        assert((value >= 0) && (value <= MAX_COMPONENT_VALUE));

        // Generate final score.
        // Note that value is in range [0,MAX_COMPONENT_VALUE] so does needs normalisation.
        final int score = mapRawValue(value, scoreMappingParams);

        // The confidence value depends on the number of samples that we were able to take
        // as well as the mapping from raw value.
        final int confidence = Math.max(0, Math.min(ScoreAndConf.MAX,
            (int) Math.round(mapRawValue(value, confidenceMappingParams) * Math.sqrt(nSamples / (double) maxSampleablePoints))));

        return(new ScoreAndConf(score, confidence));
        }

    /**Max brightness returned by samplePoint (minimum is 0); strictly positive. */
    private static final byte MAX_COMPONENT_VALUE = 127;

    /**Compute (interpolated) mapping from raw value to score/confidence.
     * The result is in the range permitted by the input mapping's min/max bounds.
     */
    private static final int mapRawValue(final int rawVal, java.util.List<ScorerParamInteger> mapping)
        {
        assert((rawVal >= 0) && (rawVal <= MAX_COMPONENT_VALUE));
        assert((mapping != null) && (mapping.size() == MAPPING_INTERVALS+1));

        // Deal with both extremes with a special fast path.
        if(rawVal == 0) { return(mapping.get(0).value); }
        if(rawVal == MAX_COMPONENT_VALUE) { return(mapping.get(MAPPING_INTERVALS).value); }

        // Compute the interval number n,
        // so we can interpolate between mapping value n and n+1.
        final int intervalNumber = (rawVal * MAPPING_INTERVALS) / MAX_COMPONENT_VALUE;
        assert((intervalNumber >= 0) && (intervalNumber < MAPPING_INTERVALS));
        final int low = mapping.get(intervalNumber).value;
        final int high = mapping.get(intervalNumber+1).value;

        final int maxResidual = (MAX_COMPONENT_VALUE+1) / MAPPING_INTERVALS;
        final int residual = rawVal % maxResidual;

        final int interpolatedValue = low + (((high-low) * residual) / maxResidual);
        assert(interpolatedValue >= mapping.get(intervalNumber).min);
        assert(interpolatedValue <= mapping.get(intervalNumber).max);

        return(interpolatedValue);
        }

    /**Take sample of area centred on specified point in image; null means point cannot be sampled, else [0,127].
     * We compute the median range of the appropriate component of the specified pixel
     * and the 8 around it, skipping any that would be outside the image boundary
     * (the specified point is assumed to be within the image)
     * or that are returned as unsamplable by the the samplePoint() method.
     * <p>
     * We discard the highest and lowest values (as possible noise)
     * and return the span of the remaining values.
     * <p>
     * We must be able to find at least 4 (four) samplable values
     * else we must return null.
     *
     * @return null for an unsamplable point, else value in range [0,127] inclusive
     */
    private Byte samplePointRange(final BufferedImage stillImage, int x, int y)
        {
        final int h = stillImage.getHeight();
        final int w = stillImage.getWidth();
        final int minX = stillImage.getMinX();
        final int minY = stillImage.getMinY();
        assert((x >= minX) && (x < minX+w));
        assert((y >= minY) && (y < minY+h));

        final ArrayList<Byte> values = new ArrayList<Byte>(9);

        final int xl = Math.max(minX, x-1);
        final int xh = Math.min(minX+w-1, x+1);
        final int yl = Math.max(minY, y-1);
        final int yh = Math.min(minY+h-1, y+1);
        for(int xi = xl; xi <= xh; ++xi)
            {
            for(int yi = yl; yi <= yh; ++yi)
                {
                final Byte value = samplePoint(stillImage, xi, yi);
                if(value != null) { values.add(value); }
                }
            }

        // Must have enough samples to extract median range,
        // ie 4 if we are to discard highest and lowest values.
        // Note that 4 is few enough to be able to sample at an image corner.
        final int sampleCount = values.size();
        if(sampleCount < 4) { return(null); }

        // Sort the values into order,
        // and ignore the top and bottom points,
        // taking the result as the range of the remainder.
        Collections.sort(values);
        final int low = values.get(1);
        final int high = values.get(sampleCount-1);
        assert(low <= high);
        return(Byte.valueOf((byte) (high - low))); // Does not create a new instance.
        }

    /**Take sample at given point in image; null means point cannot be sampled, else [0,127].
     * We regard points that are less than 50% opaque as being unsamplable
     * except where we are sampling the alpha value itself.
     *
     * @return null for an unsamplable point, else value in range [0,127] inclusive
     */
    private Byte samplePoint(final BufferedImage stillImage, int x, int y)
        {
        final int argb = stillImage.getRGB(x, y);

        final int alpha = (argb >>> 24);
        if(componentParam.value == PixelComponent.ALPHA) { return(Byte.valueOf((byte) (alpha >>> 1))); }

        // Reject mainly-transparent pixels.
        // alpha == 0xff ==> opaque (we can use it), alpha == 0 ==> transparent/unusable.
        if(alpha < 0x80) { return(null); }

        // Extract r g b components in the range [0,255].
        final int r = (argb >>> 16) & 0xff;
        if(componentParam.value == PixelComponent.R) { return(Byte.valueOf((byte) (r >>> 1))); }
        final int g = (argb >>>  8) & 0xff;
        if(componentParam.value == PixelComponent.G) { return(Byte.valueOf((byte) (g >>> 1))); }
        final int b =  argb         & 0xff;
        if(componentParam.value == PixelComponent.B) { return(Byte.valueOf((byte) (b >>> 1))); }

        // Compute B of HSB model efficiently.
        if(componentParam.value == PixelComponent.BRI)
            {
            final int cmax = Math.max(r, Math.max(g, b));
            final int brByte = cmax/2;
            assert((brByte >= 0) && (brByte <= MAX_COMPONENT_VALUE));
            return(Byte.valueOf((byte) brByte)); // Does not create new instances.
            }

        // Compute Y value efficiently.
        if(componentParam.value == PixelComponent.Y)
            {
            // Compute Y of YUV model (as used in PAL TV encodings).
            // Approx .299r + .587g + .114b, but mapping from [0-255] inputs to [0,127] result.
            final int Y = (37 * r + 74 * g + 14 * b + 128) >> 8;
            assert((Y >= 0) && (Y <= MAX_COMPONENT_VALUE));
            return(Byte.valueOf((byte) Y)); // Does not create new instances.
            }
 
        // Hue or saturation; make it default to hue for expansion robustness.
        // Slow computation involving floating-point.
        final boolean hueRequired = (componentParam.value != PixelComponent.SAT);
        return(Byte.valueOf((byte) Math.max(0, Math.min(MAX_COMPONENT_VALUE, Math.round(
            Color.RGBtoHSB(r, g, b, null)[hueRequired ? 0 : 1] * MAX_COMPONENT_VALUE)))));
        }

    /**Number of intervals in which to map score and confidence; strictly positive.
     * Note that there will be one more parameters that this for each of
     * the score and confidence mappings
     * since the parameters mark the end-points of each interval,
     * with linear interpolation between them.
     * <p>
     * Too small a value hinders a good fit to multi-modal distributions,
     * too large a value makes the state space to explore unnecessarily large.
     * <p>
     * There is no point in having more intervals than distinct component values
     * (from which this is driven), indeed there should be many fewer.
     * <p>
     * A power of two may help speed up some computations.
     */
    private static final int MAPPING_INTERVALS = (1 << 4);

    /**Default base score mapping output value; never null.
     * Because there are many parameters based on this, with this name as a prefix,
     * we keep the name as short as possible.
     * <p>
     * We ensure that this prefix uniquely distinguishes this from all other parameters
     * to give us the maximum possible choice of suffixes.
     * <p>
     * The perturbation value allows a reasonable state-space exploration speed.
     * <p>
     * There is no bias: there is no preferred/cheap perturbation direction.
     */
    private static final ScorerParamInteger scoreMappingParamBoundsBase =
        new ScorerParamInteger(-ScoreAndConf.MAX, 0, +ScoreAndConf.MAX, ScoreAndConf.MAX/64, false, "S");

    /**Immutable List of default mapping factors for the score from least to highest input values; never null.
     * Always MAPPING_INTERVALS+1 entries.
     */
    private static final java.util.List<ScorerParamInteger> scoreMappingParamsBounds;
    /**Initialise scoreMappingParamsBounds. */
    static
        {
        final ArrayList<ScorerParamInteger> l = new ArrayList<ScorerParamInteger>(MAPPING_INTERVALS+1);
        for(int i = 0; i <= MAPPING_INTERVALS; ++i)
            {
            // Set the values to run from 0 to MAX.
            final int value = (scoreMappingParamBoundsBase.max * i) / MAPPING_INTERVALS;
            // Name the encoding of the parameter number as short as possible.
            final String n = Integer.toString(i, Character.MAX_RADIX);
            l.add(new ScorerParamInteger(scoreMappingParamBoundsBase.min,
                                         scoreMappingParamBoundsBase.def,
                                         scoreMappingParamBoundsBase.max,
                                         scoreMappingParamBoundsBase.delta,
                                         scoreMappingParamBoundsBase.biasedLow,
                                         scoreMappingParamBoundsBase.name + n,
                                         value));
            }
        // Make result immutable for safety/sanity!
        scoreMappingParamsBounds = Collections.<ScorerParamInteger>unmodifiableList(l);
        }

    /**Mapping factors for the score from least to highest input values; never null.
     * Always MAPPING_INTERVALS+1 entries.
     */
    private final java.util.List<ScorerParamInteger> scoreMappingParams;

    /**Default base confidence mapping output value; never null.
     * Because there are many parameters based on this, with this name as a prefix,
     * we keep the name as short as possible.
     * <p>
     * We ensure that this prefix uniquely distinguishes this from all other parameters
     * to give us the maximum possible choice of suffixes.
     * <p>
     * The perturbation value allows a reasonable state-space exploration speed.
     * <p>
     * There is no bias: there is no preferred/cheap perturbation direction.
     */
    private static final ScorerParamInteger confidenceMappingParamBoundsBase =
        new ScorerParamInteger(0, 0, +ScoreAndConf.MAX, ScoreAndConf.MAX/128, false, "C");

    /**Immutable List of default mapping factors for the confidence from least to highest input values; never null.
     * Always MAPPING_INTERVALS+1 entries.
     */
    private static final java.util.List<ScorerParamInteger> confidenceMappingParamsBounds;
    /**Initialise confidenceMappingParamsBounds. */
    static
        {
        final ArrayList<ScorerParamInteger> l = new ArrayList<ScorerParamInteger>(MAPPING_INTERVALS+1);
        for(int i = 0; i <= MAPPING_INTERVALS; ++i)
            {
            // Set the values to run from 0 to MAX.
            final int value = (confidenceMappingParamBoundsBase.max * i) / MAPPING_INTERVALS;
            // Name the encoding of the parameter number as short as possible.
            final String n = Integer.toString(i, Character.MAX_RADIX);
            l.add(new ScorerParamInteger(confidenceMappingParamBoundsBase.min,
                                         confidenceMappingParamBoundsBase.def,
                                         confidenceMappingParamBoundsBase.max,
                                         confidenceMappingParamBoundsBase.delta,
                                         confidenceMappingParamBoundsBase.biasedLow,
                                         confidenceMappingParamBoundsBase.name + n,
                                         value));
            }
        // Make result immutable for safety/sanity!
        confidenceMappingParamsBounds = Collections.<ScorerParamInteger>unmodifiableList(l);
        }

    /**Mapping factors for the score from least to highest input values; never null.
     * Always MAPPING_INTERVALS+1 entries.
     */
    private final java.util.List<ScorerParamInteger> confidenceMappingParams;

    /**Default pixel sample method; never null. */
    private static final ScorerParamEnum<SampleType> pixelSampleMethodParamBounds =
        new ScorerParamEnum<SampleType>(SampleType.VALUE, "pixelSampleMethod");

    /**Pixel sample method; never null. */
    private final ScorerParamEnum<SampleType> pixelSampleMethodParam;

    /**Default sample combination method; never null. */
    private static final ScorerParamEnum<SampleType> sampleCombinationParamBounds =
        new ScorerParamEnum<SampleType>(SampleType.RANGE, "sampleCombination");

    /**Pixel-value constituent to sample; never null. */
    private final ScorerParamEnum<SampleType> sampleCombinationParam;

    /**Default pixel-value constituent to sample; never null. */
    private static final ScorerParamEnum<PixelComponent> componentParamBounds =
        new ScorerParamEnum<PixelComponent>(PixelComponent.Y, "component");

    /**Pixel-value constituent to sample; never null. */
    private final ScorerParamEnum<PixelComponent> componentParam;

    /**Min, default and max shift (power-of-two) for the target sample size; never null.
     * This has a bias to lower values since this results in less work done.
     */
    private static final ScorerParamInteger sampleSizeParamBounds =
        new ScorerParamInteger(2, 7, 12, 1, true, "sampleSizePower");

    /**Target sample size for this instance; never null.
     * Target number of sample points; strictly positive.
     * If fewer sample points than this are available,
     * then the result's confidence will drop
     * in accordance with expected sample-variance noise reduction.
     * <p>
     * Reducing the sample size by a factor of x is assumed to increase the noise,
     * and thus reduce the confidence, by a factor of sqrt(x).
     * <p>
     * Experience suggests that upwards of a thousand samples gives good consistency.
     */
    private final ScorerParamInteger sampleSizeParam;

    /**Min, default and max shift for the outlier fraction to remove; never null.
     * Shift to get fraction/portion of the samples that are discarded from either end as potential noise; 3 or greater.
     * A larger value discards fewer samples/values,
     * but makes the measurement more sensitive to noise/grain and small highlights (etc).
     * We always trim at least one outlier at each end.
     * <p>
     * A good shift value may line in the region of 3 to 7
     * thus discarding ~13% to ~1% of the outliers from either extreme.
     */
    private static final ScorerParamInteger outlierParamBounds =
        new ScorerParamInteger(2, 6, 10, "outlierFractionPower");

    /**Outlier parameter value for this instance; never null. */
    private final ScorerParamInteger outlierParam;

    /**Get ordered parameter definitions and values (immutable) for this Scorer; never null. */
    public Collection<ScorerParam> getParameterDefsAndValues()
        {
        final ArrayList<ScorerParam> result = new ArrayList<ScorerParam>();
        result.addAll(Arrays.asList(new ScorerParam[] {
                sampleSizeParam, outlierParam, componentParam, sampleCombinationParam, pixelSampleMethodParam,
            }));
        result.addAll(scoreMappingParams);
        result.addAll(confidenceMappingParams);
        return(Collections.unmodifiableList(result));
        }
    }
