/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.ai.scorer.parameterised;

import java.awt.*;
import java.awt.image.*;
import java.util.*;

import org.hd.d.pg2k.ai.scorer.*;
import org.hd.d.pg2k.svrCore.*;

/**Simple score/measure of image "exposure" over random sample points; never null.
 * This ignores outlier values of luminance (Y/brightness),
 * and the wider the spread of the remainder normalised to the theoretical range the better
 * and the more confident our prediction.
 * This does not generate negative (bad) results,
 * just good results for those images that seem to have good exposure by this simple metric.
 * <p>
 * The set of sample points is a grid spaced in a regular (though not a simple grid)
 * pattern across the image.  The layout of the sample points depends on the image
 * size and aspect ratio, and is generated from a pseudo-randon generator,
 * to help avoid poor sampling of images with regular patterns in them,
 * <p>
 * For example, if having discarded the few top and bottom Y values the remainder
 * cover 90% of the available Y range, this returns a score of 90% of MAX.
 * <p>
 * The confidence depends on the number of selected sample points that are available;
 * if all those chosen are available and fully opaque (no alpha) then the confidence is 1.
 * If no chosen sample points are available and fully opaque then the confidence is 0.
 * Partially-opaque points may be ignored or contribute pro-rata by opacity/alpha.
 * <p>
 * A single-pixel transparent image should result in a result of (0, 0)
 * ie no range of luminance (Y), and no fully-opaque sample points.
 * <p>
 * Very small images will have all their pixels sampled,
 * and so can in principle still achieve a MAX confidence.
 * <p>
 * This scoring method should run in constant time regardless of image size
 * since it samples at most a fixed ceiling number of image points.
 * <p>
 * This may be most useful for continuous-tone true-colour/greyscale (eg JPEG) images.
 * 
 * @author dhd
 */
public final class SimpleExposure extends AbstractImgScorer
    {
    /**Create simple non-parameterised instance. */
    public SimpleExposure()
        {
        // All parameters get default values.
        sampleSizeParam = sampleSizeParamBounds;
        brightnessParam = brightnessParamBounds;
        outlierParam = outlierParamBounds;
        }

    /**Create parameterised version. */
    public SimpleExposure(final String nameAndParameters)
        {
        super(nameAndParameters);

        // Assign parameters from the captured map where possible, using default values otherwise.
        sampleSizeParam = (ScorerParamInteger) sampleSizeParamBounds.parse(paramValueMap.get(sampleSizeParamBounds.name));
        brightnessParam = (ScorerParamInteger) brightnessParamBounds.parse(paramValueMap.get(brightnessParamBounds.name));
        outlierParam = (ScorerParamInteger) outlierParamBounds.parse(paramValueMap.get(outlierParamBounds.name));
        }

    /**Score the image for "exposure"; never null nor negative in any component.
     * @throws IllegalArgumentException for a null image
     */
    public ScoreAndConf computeScoreAndConfidence(final RenderedImage stillImage)
        {
        if(stillImage == null) { throw new IllegalArgumentException(); }

        final int h = stillImage.getHeight();
        final int w = stillImage.getWidth();

        // Don't try to score strange (eg zero-sized) images.
        if((h < 1) || (w < 1)) { return(ScoreAndConf.NO_OPINION); }

        final long availablePixels = h * w;
        assert(availablePixels > 0);

        // If the image is small we may simply examine every pixel.
        final int target_sample_points = (1 << sampleSizeParam.value);
        final boolean smallImage = (availablePixels <= target_sample_points);
        final int maxSampleablePoints = smallImage ? (int) availablePixels : target_sample_points;

        final int minX = stillImage.getMinX();
        final int minY = stillImage.getMinY();

        // For sampling we'll need a buffered image;
        // so we copy the input image if need be.
        final BufferedImage bi = ImageUtils.convertRenderedImageToBufferedImage(stillImage);

//System.out.println("Brightness at min X,Y (width,height = "+w+","+h+"): " + samplePoint(bi, minX, minY));

        // Set of brightness values sampled from the image.
        // Assumed to be sortable, ie all values finite.
        final ArrayList<Byte> pointsSampled = new ArrayList<Byte>(maxSampleablePoints);

        // For small images sample all pixels.
        if(smallImage)
            {
            for(int x = w; --x >= 0; )
                {
                for(int y = h; --y >= 0; )
                    {
                    final Byte b = samplePoint(bi, x + minX, y + minY);
                    if(b == null) { continue; }
                    pointsSampled.add(b);
                    }
                }
            }
        // For reasonable-size (not tiny) images,
        // we sample pixels in a pseudo-random pattern based on image size/shape.
        else
            {
            // Make the sample pattern (random seen) dependent only on image size/shape
            // to help circumvent some systematic sample pattern problems.
            // We do not make the sample pattern/seed dependent on sample size
            // so that as we take more samples all previous values are included.
            final Random r = new Random(w + (((long)h) << 16));

            // To allow in the main for non-square partially-transparent images
            // but also for images close in pixels to our threshold size
            // (so that we capture as many of the samplable points as reasonable possible)
            // we are prepared to attempt the sampling operation
            // more than the target number of samples
            // (though we stop as soon as our target sample count is reached).
            // We also reject would-be duplicate sampling of the same point.
            assert((new Point(0, 0)).equals(new Point(0, 0)));
            final Set<Point> alreadySampled = new HashSet<Point>(1 + 2*target_sample_points);
            for(int i = 2*target_sample_points; (--i >= 0) && (alreadySampled.size() < target_sample_points); )
                {
                // Choose a (new) random point and take a sample.
                final int x = r.nextInt(w);
                final int y = r.nextInt(h);
                final Point p = new Point(x, y);
                if(!alreadySampled.add(p)) { continue; } // Skip if point not new...
//System.out.println("[Sample point: "+p+".]");
                final Byte b = samplePoint(bi, x + minX, y + minY);
                if(b == null) { continue; }
//System.out.println("[Point accepted.]");
                pointsSampled.add(b);
                } 
            }

        final int nSamples = pointsSampled.size();
        // If we are to discard any outliers
        // and have at least two points left to compute the range with
        // then we must have collected at least 4 samples.
        // If we don't have at least this many, return a "no confidence" result.
        if(nSamples < 4) { return(ScoreAndConf.NO_OPINION); }
        // Compute number of outliers to discard from either extreme.
        final int nOutliers = Math.max(1, nSamples / (1 << outlierParam.value));
        assert(nSamples - 2*nOutliers >= 2);

        // Sort the sampled points.
        // Discard/ignore outliers from either end of the range, likely to be noise.
        // Compute the relative interval between the remaining extremes to compute the score.
        // The confidence is based on the sample size and the score.
        Collections.sort(pointsSampled);
        final int lowest = pointsSampled.get(nOutliers).intValue();
        final int highest = pointsSampled.get(nSamples-1 - nOutliers).intValue();
//System.out.println("SimpleExposure: low/high = "+lowest+"/"+highest + "; lowest/highest = "+pointsSampled.get(0)+"/"+pointsSampled.get(nSamples-1));
        // Note that brightness is in range [0,MAX_BRIGHTNESS] so does needs normalisation.
        // We regard any image which covers a reasonable chunk of the luminance range to be fine.
        final int score = Math.max(0, Math.min(ScoreAndConf.MAX, Math.round((ScoreAndConf.MAX * (highest-lowest) * 100) / (MAX_BRIGHTNESS * brightnessParam.value))));
        final int confidence = Math.max(0, Math.min(ScoreAndConf.MAX, (int) Math.round(score * Math.sqrt(nSamples / (double) maxSampleablePoints))));

        return(new ScoreAndConf(score, confidence));
        }

    /**Max brightness returned by samplePoint (minimum is 0); strictly positive. */
    private static final byte MAX_BRIGHTNESS = 127;

    /**If true then use the B of the HSB model (max of R, G and B), else use the weighted Y value from YUV. */
    private static final boolean USE_B_NOT_Y = false;

    /**Take sample at given point in image; null means point cannot be sampled, else [0,127].
     * We regard points that are less than 50% opaque as being unsamplable.
     *
     * @return null for an unsamplable point, else brightness in range [0,127] inclusive
     */
    private Byte samplePoint(final BufferedImage stillImage, int x, int y)
        {
        final int argb = stillImage.getRGB(x, y);

        // Reject mainly-transparent pixels.
        // alpha == 0xff ==> opaque (we can use it), alpha == 0 ==> transparent/unusable.
        if((argb >>> 24) < 0x80) { return(null); }

        // Convert from sRGB to brightness/luminance/Y value in range [0,127].

        // Extract r g b components in the range [0,255].
        final int r = (argb >>> 16) & 0xff;
        final int g = (argb >>>  8) & 0xff;
        final int b =  argb         & 0xff;

        // Compute B of HSB model.
        if(USE_B_NOT_Y)
            {
            final int cmax = Math.max(r, Math.max(g, b));
            final int brByte = cmax/2;
            assert((brByte >= 0) && (brByte <= MAX_BRIGHTNESS));
            return(Byte.valueOf((byte) brByte)); // Does not create new instances.
            }

        // Compute Y of YUV model (as used in PAL TV encodings).
        // Approx .299r + .587g + .114b, but mapping from [0-255] inputs to [0,127] result.
        final int Y = (37 * r + 74 * g + 14 * b + 128) >> 8;
        assert((Y >= 0) && (Y <= MAX_BRIGHTNESS));
        return(Byte.valueOf((byte) Y)); // Does not create new instances.
        }

    /**Min, default and max shift (power-of-two) for the target sample size; never null.
     * This has a bias to lower values since this results in less work done.
     */
    private static final ScorerParamInteger sampleSizeParamBounds =
        new ScorerParamInteger(2, 7, 12, 1, true, "sampleSizePower");

    /**Target sample size for this instance; never null.
     * Target number of sample points; strictly positive.
     * If fewer sample points than this are available,
     * then the result's confidence will drop
     * in accordance with expected sample-variance noise reduction.
     * <p>
     * Reducing the sample size by a factor of x is assumed to increase the noise,
     * and thus reduce the confidence, by a factor of sqrt(x).
     * <p>
     * Experience suggests that upwards of a thousand samples gives good consistency.
     */
    private final ScorerParamInteger sampleSizeParam;

    /**Min, default and max percentage of Y range that should be spanned for MAX score; never null. */
    private static final ScorerParamInteger brightnessParamBounds =
        new ScorerParamInteger(3, 65, 100, 5, false, "brightnessRangePercent");

    /**Percentage brightness range that should be spanned for MAX score for this instance; never null. */
    private final ScorerParamInteger brightnessParam;

    /**Min, default and max shift for the outlier fraction to remove; never null.
     * Shift to get fraction/portion of the samples that are discarded from either end as potential noise; 3 or greater.
     * A larger value discards fewer samples/values,
     * but makes the measurement more sensitive to noise/grain and small highlights (etc).
     * We always trim at least one outlier at each end.
     * <p>
     * A good shift value may line in the region of 3 to 7
     * thus discarding ~13% to ~1% of the outliers from either extreme.
     */
    private static final ScorerParamInteger outlierParamBounds =
        new ScorerParamInteger(2, 6, 10, "outlierFractionPower");

    /**Outlier parameter value for this instance; never null. */
    private final ScorerParamInteger outlierParam;

    /**Get parameter definitions and values (immutable) for this Scorer; never null. */
    public Collection<ScorerParam> getParameterDefsAndValues()
        {
        return(Collections.unmodifiableList(Arrays.asList(new ScorerParam[] {
                sampleSizeParam, brightnessParam, outlierParam
            })));
        }
    }
