/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.offline;

import org.hd.d.pg2k.svrCore.AddrTools;
import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.location.GeoProximity;
import org.hd.d.pg2k.svrCore.location.GeoUtils;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 01-May-2006
 * Time: 12:02:55
 */

/**Off-line utility to create properties file for IPv4 to ccTLD mapping.
 */
public class MakeCCTLDFromIPPrefixProperties
    {
    /**Prevent instances from being constructed. */
    private MakeCCTLDFromIPPrefixProperties() { }

    /**Input data format name for MySQL-dump download from www.hostIP.info (circa 2006/04). */
    public static final String INPUT_FORMAT_HOSTIP = "hostIPSQL";

    /**All allowed input formats; immutable and never null. */
    public static final SortedSet<String> INPUT_FORMATS =
        Collections.unmodifiableSortedSet(new TreeSet<String>(Arrays.asList(new String[]{
            INPUT_FORMAT_HOSTIP,
            })));

    /**Threshold (out of 256) at which we will discard minority values when doing lossy cc-prefix-table compression; strictly positive.
     * If all the values in a given block share the same value except a small minority,
     * and the count of minority items is less than or equal to this value
     * then we pretend that the exception did not exist so we can trim down the tree at this point.
     * <p>
     * This is computed so that given the rough expected maximum cost of a worst-possible routing error
     * (paying transit and getting ropey connectivity rather than a fast free local connection)
     * we will not on average pay more than about twice the optimum routing costs
     * because of our lossy encoding.
     * This assumes that we'd never do much better than "country" level routing.
     */
    private static final int UNCONDITIONAL_PREFIX_LOSSY_COMPRESS_THRESHOLD = 1 +
            256/GeoProximity.COUNTRY.getCloseness();

    /**Threshold (out of 256) at which we will lump countries into a region when doing lossy cc-prefix-table compression; strictly positive.
     * Provided that all but this number of entries in a given block
     * are in the same region as the dominant ccTLD in that block
     * then during lossy compression we can replace them with that dominant ccTLD.
     * <p>
     * The cost of this is assumed to be mainly
     * routing cost for the wrong country within a region when claiming the dominant ccTLD.
     * (We currently neglect the possible cost of completely-wrong routing out-of-region.)
     */
    private static final int REGION_PREFIX_LOSSY_COMPRESS_THRESHOLD = 1 +
            256/GeoProximity.COUNTRYGROUP.getCloseness();

    /**Threshold (out of 256) at which we will lump countries/regions into a continent when doing lossy cc-prefix-table compression; strictly positive.
     * Provided that all but this number of entries in a given block
     * are in the same region as the dominant ccTLD or region in that block
     * then during lossy compression we can replace them with that dominant ccTLD/region.
     * <p>
     * The cost of this is assumed to be mainly
     * routing cost for the wrong country within a continent when claiming the dominant ccTLD/region.
     * (We currently neglect the possible cost of completely-wrong routing out-of-contient.)
     */
    private static final int CONTINENT_PREFIX_LOSSY_COMPRESS_THRESHOLD = 1 +
            256/GeoProximity.CONTINENT.getCloseness();


    /**Dump an IP-prefix map as-is to the given Writer.
     * This does no "optimisation" or other transformation.
     * <p>
     * The output is printed in sorted order by the address-prefix key.
     * <p>
     * The output should be directly suitable to form all or part of a
     * ccTLDFromIPPrefix.properties set.
     */
    public static final void dumpPrefixMap(final SortedMap<AddrTools.AddrPrefix,String> map, final PrintWriter pw)
        {
        for(final AddrTools.AddrPrefix ap : map.keySet())
            { pw.println(ap.toPaddedDottedPrefix() + ' ' + map.get(ap)); }
        }

    /**Perform (lossy) compression on an IP-prefix-to-ccTLD map.
     * This may use lossy compression of the loaded values
     * where the saving is large and the implied increase in routing/performance cost is low.
     * <p>
     * We never delete "" entries to avoid semantic changes,
     * nor top-level (1-octet) entries for clarity.
     */
    private static SortedMap<AddrTools.AddrPrefix,String> lossyCompressCcTLDFromIPPrefixMap(final SortedMap<AddrTools.AddrPrefix,String> mapIn)
        {
        // We take a working copy of the map.
        final SortedMap<AddrTools.AddrPrefix,String> work = new TreeMap<AddrTools.AddrPrefix, String>(mapIn);

        // We find the length of the longest prefix in the map.
        int longestPrefix = 0;
        for(final AddrTools.AddrPrefix prefix : work.keySet())
            { if(prefix.length() > longestPrefix) { longestPrefix = prefix.length(); } }

        // We work backwards in length,
        // trying to trim away large numbers of leaves that carry little information...
        // We never try to remove 1-octect entries.
        for(int i = longestPrefix; i > 1; --i)
            { _lossyTrimPrefixesAtLength(work, i); }

        return(work);
        }

    /**Try to trim away leaves at the given prefix length.
     * We never remove "" entries.
     * <p>
     * We won't remove entries at the stated length
     * that have longer sub-entries.
     * This implies that trimming should be performed from the longest prefixes
     * down to the shortest.
     * <p>
     * The map is updated in place.
     */
    private static void _lossyTrimPrefixesAtLength(final SortedMap<AddrTools.AddrPrefix,String> map,
                                                   final int lengthToTrim)
        {
        if((map == null) || (lengthToTrim < 1))
            { throw new IllegalArgumentException(); }

System.out.println("_lossyTrimPrefixesAtLength(): STARTING trim at length: " + lengthToTrim);

        // First we copy all the keys into a List for easier manipulation.
        // We assume that this is always in sorted order.
        final AddrTools.AddrPrefix keys[] = map.keySet().toArray(new AddrTools.AddrPrefix[map.size()]);

        // We will search for entries of the stated length
        // working on blocks of those that differ only in their final byte
        // and refusing to trim if there are longer sub-entries.
        nextBlock: for(int i = 0; i < keys.length; ++i)
            {
            final AddrTools.AddrPrefix ap = keys[i];
            if(ap.length() != lengthToTrim) { continue; }

            // Make the prefix that is one octet shorter.
            final AddrTools.AddrPrefix prefix = new AddrTools.AddrPrefix(ap, ap.length() - 1);

            // Note if we find longer sub-entries.
            boolean foundLongerSubEntries = false;

            // Find how big this block is, and possibly trim it.
            // Remember all non-"" entries, ie that we might zap.
            final List<AddrTools.AddrPrefix> potentiallyZappable = new ArrayList<AddrTools.AddrPrefix>();
            final int blockStart = i;
            int blockEnd = blockStart;
            for(int j = blockStart; j < keys.length; ++j)
                {
                final AddrTools.AddrPrefix keyInBlock = keys[j];
                if(!keyInBlock.isStrictPrefix(prefix))
                    {
                    // Left the block of candidate entries.
                    break;
                    }

                // If not a "" entry, then it is potentially zappable.
                final boolean isEmptyString = "".equals(map.get(keyInBlock));
                if(!isEmptyString)
                    { potentiallyZappable.add(keyInBlock); }

                // Still part of the block...
                blockEnd = j;

                // We will not be able to trim if we find longer sub-entries
                // other than "" entries which we always keep.
                if((keyInBlock.length() > lengthToTrim) && !isEmptyString)
                    { foundLongerSubEntries = true; }
                }

            // Whether we trim or not,
            // the outer loop can start no skip this block
            // (note that it will get incremented at the loop re-initialisation).
            i = blockEnd;

//            final int blockLength = blockEnd - blockStart + 1;

//System.out.println("_lossyTrimPrefixesAtLength(): found candidate block to trim: " + prefix + ".*; blockLength=" + blockLength +
//                   ", potentiallyZappable.size()=" + potentiallyZappable.size() +
//                   (!foundLongerSubEntries ? "" : " (untrimmable because of longer non-\"\" sub-entries)"));

            // If we found longer sub-entries then we can't trim this block,
            // so move to the next one.
            if(foundLongerSubEntries)
                { continue nextBlock; /* Start on next block... */ }

            // Collect counts of different values at this level of the tree.
            // Note that we do this by trying to look up every possible value
            // within this block, even for those values not explicitly present,
            // to correctly account for propagation upwards of more general entries.
            final Map<String,Integer> valueCounts = new HashMap<String, Integer>();
            // Count of null values, ie where the lookup could not find anything.
            int nullCount = 0;

            // Do IPv4 address lookup,
            // so use a 4-byte address (leaving trailing octets at zero).
            final byte addr[] = new byte[4];
            // Copy in the correct prefix...
            final byte rawPrefix[] = prefix.toByteArray();
            for(int j = rawPrefix.length; --j >= 0; ) { addr[j] = rawPrefix[j]; }

            try
                {
                for(int b = 256; --b >= 0; )
                    {
                    // Adjust the variable byte for this block.
                    addr[rawPrefix.length] = (byte) b;

                    // Create an IPv4 InetAddress.
                    final InetAddress synthInetAddress = InetAddress.getByAddress(addr);
                    final String v = GeoUtils.lookupAddrInIPToCcTLDMap(synthInetAddress, map, ap.length());

                    // Count all successful lookups.
                    if(v != null)
                        {
                        final Integer count = valueCounts.get(v);
                        if(count == null) { valueCounts.put(v, 1); }
                        else { valueCounts.put(v, 1 + count); }
                        }
                    else { ++nullCount; }
                    }
                }
            catch(final UnknownHostException e) // Should never happen.
                {
                e.printStackTrace();
                throw new Error("Internal error");
                }

//System.out.println("_lossyTrimPrefixesAtLength(): block="+prefix+".*, nulls="+nullCount+", valueCounts=" + valueCounts);

            // OK, decide if we can compress this block.
            // If we *do*, then we remove all non-"" nodes from the block
            // and write store at the prefix level the chosen representative value
            // (if that is null, it means that we remove the prefix node too).

            // Simplest case:
            // One overwhelming majority value, whatever it is,
            // if in a huge majority, is used to replace this whole block.
            final int overwhelmThreshold = 256 - UNCONDITIONAL_PREFIX_LOSSY_COMPRESS_THRESHOLD;
            assert(overwhelmThreshold > 128); // Must be a clear majority to avoid ambiguity...

            // If we know almost nothing about this block
            // then we simply remove the noise within it.
            if(nullCount >= overwhelmThreshold)
                {
                // Zap all the (non-"") nodes from this block and possibly the prefix node.
System.out.println("_lossyTrimPrefixesAtLength(): removing block below "+prefix+" (and parent node if > 1 octet); potentiallyZappable.size()=" + potentiallyZappable.size() + ", valueCounts="+valueCounts);
                if(prefix.length() > 1) { map.remove(prefix); }
                map.keySet().removeAll(potentiallyZappable);
                continue nextBlock; // Start on next block...
                }

//            // If we have too many unknowns in a block to be confortable about
//            // making any specific routing judgement for the whole block
//            // then give up now.
//            if(nullCount > UNCONDITIONAL_PREFIX_LOSSY_COMPRESS_THRESHOLD)
//                { continue nextBlock; /* Start on next block... */ }

            // Find the value with the highest count...
            String highestCountValue = null;
            int highestCount = 0;
            for(final String v : valueCounts.keySet())
                {
                final int c = valueCounts.get(v);
                if(c > highestCount)
                    {
                    highestCount = c;
                    highestCountValue = v;
                    }
                }
            assert((highestCountValue != null) && (highestCount > 0));
//System.out.println("_lossyTrimPrefixesAtLength(): dominant="+ highestCountValue + ", count="+highestCount);

            // If there is one overwhelmingly common entry in a block
            // then rewrite the block to that one value.
            if(highestCount >= overwhelmThreshold)
                {
                // Zap all the (non-"") nodes from this block and replace the prefix node.
System.out.println("_lossyTrimPrefixesAtLength(): replacing block with new parent node: `" + prefix + ' ' + highestCountValue + "'; potentiallyZappable.size()=" + potentiallyZappable.size() + ", valueCounts="+valueCounts);
                map.put(prefix, highestCountValue);
                map.keySet().removeAll(potentiallyZappable);
                continue nextBlock; // Start on next block...
                }

            final int regionThreshold = 256 - REGION_PREFIX_LOSSY_COMPRESS_THRESHOLD;

            // If there is one very common ccTLD in a block
            // and all but a very few other entries are countries in the same region,
            // then replace with the dominant ccTLD.
            if(GeoUtils.CCTLD.isSyntaticallyValidCcTLD(highestCountValue) &&
                    (highestCount >= regionThreshold))
                {
                // Check that enough other entries are countries in the same region.
                int outOfRegion = nullCount; // Include the nulls in the not-same-region list.
                final Set<GeoUtils.CCTLD> closeCcTLDs = GeoUtils.getCloseCCTLDs(new GeoUtils.CCTLD(highestCountValue));
                for(final String v : valueCounts.keySet())
                    {
                    if(v.equals(highestCountValue)) { continue; }

                    if(!GeoUtils.CCTLD.isSyntaticallyValidCcTLD(v) || !closeCcTLDs.contains(new GeoUtils.CCTLD(v)))
                        {
                        // Bad luck; this isn't a close country...
                        outOfRegion += valueCounts.get(v);
//System.out.println("_lossyTrimPrefixesAtLength(): outOfRegion="+outOfRegion+" after "+v +" dominant="+highestCountValue);
                        }
                    }
                if(outOfRegion <= REGION_PREFIX_LOSSY_COMPRESS_THRESHOLD)
                    {
                    // Zap all the (non-"") nodes from this block and replace the prefix node.
System.out.println("_lossyTrimPrefixesAtLength(): replacing block with new parent node: `" + prefix + ' ' + highestCountValue + "'; potentiallyZappable.size()=" + potentiallyZappable.size() + ", valueCounts="+valueCounts);
                    map.put(prefix, highestCountValue);
                    map.keySet().removeAll(potentiallyZappable);
                    continue nextBlock; // Start on next block...
                    }
                }


            final int continentThreshold = 256 - CONTINENT_PREFIX_LOSSY_COMPRESS_THRESHOLD;

            // If there is one very common ccTLD or region/registry/continent in a block
            // and all but a very few other entries are in the same region,
            // then replace with the dominant ccTLD/region.
            if((GeoUtils.CCTLD.isSyntaticallyValidCcTLD(highestCountValue) ||
                    GeoUtils.isSyntaticallyValidRegistryName(highestCountValue)) &&
                        (highestCount >= continentThreshold))
                {
                // Check that enough other entries are countries in the same region.
                int outOfContinent = nullCount; // Include the nulls in the not-same-continent list.
                final Set<GeoUtils.CCTLD> closeCcTLDs = GeoUtils.CCTLD.isSyntaticallyValidCcTLD(highestCountValue) ?
                                                        GeoUtils.getCloseCCTLDs(new GeoUtils.CCTLD(highestCountValue)) :
                                                        GeoUtils.getCountriesInRegion(highestCountValue);
                for(final String v : valueCounts.keySet())
                    {
                    if(v.equals(highestCountValue)) { continue; }

                    // Make sure that this is a country close to the dominant country,
                    // OR that this is a country in the dominant region,
                    // else it is "out of area".
                    if((!GeoUtils.CCTLD.isSyntaticallyValidCcTLD(v) || !closeCcTLDs.contains(new GeoUtils.CCTLD(v))) &&
                            (!GeoUtils.isSyntaticallyValidRegistryName(v) ||
                            !GeoUtils.CCTLD.isSyntaticallyValidCcTLD(highestCountValue) ||
                            !GeoUtils.getCountriesInRegion(v).contains(new GeoUtils.CCTLD(highestCountValue))))
                        {
                        // Bad luck; this isn't a close country/region...
                        outOfContinent += valueCounts.get(v);
//System.out.println("_lossyTrimPrefixesAtLength(): outOfContinent="+outOfContinent+" after "+v +" dominant="+highestCountValue);
                        }
                    }
                if(outOfContinent <= CONTINENT_PREFIX_LOSSY_COMPRESS_THRESHOLD)
                    {
                    // Zap all the (non-"") nodes from this block and replace the prefix node.
System.out.println("_lossyTrimPrefixesAtLength(): replacing block with new parent node: `" + prefix + ' ' + highestCountValue + "'; potentiallyZappable.size()=" + potentiallyZappable.size() + ", valueCounts="+valueCounts);
                    map.put(prefix, highestCountValue);
                    map.keySet().removeAll(potentiallyZappable);
                    continue nextBlock; // Start on next block...
                    }
                }








            // TODO: more subtle cases...






            // If no lossy compression scheme has applied
            // and if there are no nulls at this level
            // then attempt some simple lossless compression to save some space.
            // Set the node above to the most common value,
            // and remove all instances of that value from the block.
            // Only worth doing when we will remove at least two sub-nodes.
            if((nullCount == 0) && (highestCount > 1))
                {
                // Zap all the (non-"") nodes from this block and replace the prefix node.
System.out.println("_lossyTrimPrefixesAtLength(): removing dups at this level having added new parent node: `" + prefix + ' ' + highestCountValue + "', valueCounts="+valueCounts);
                map.put(prefix, highestCountValue);
                for(final AddrTools.AddrPrefix pzap : potentiallyZappable)
                    {
                    if(highestCountValue.equals(map.get(pzap)))
                        { map.remove(pzap); }
                    }
                continue nextBlock; // Start on next block...
                }
            }
        }

    /**Load external IP/location data into internal-style table; never null.
     * This data can potentially be in one of several formats,
     * each of which will require some massaging to get into our model.
     * <p>
     * Only input data records corresponding to ccTLDs explicitly listed in
     * the "geo-proximity" data, ie whose getCloseCCTLDs() result is non-empty,
     * will be retained.
     * <p>
     * We construct and return an unsorted map for speed.
     *
     * @param inputData  the (readable) input data file; never null
     * @param inputFormat  the format (one of INPUT_FORMATS); never null
     *
     * @return input data in our format, for "interesting" ccTLDs
     */
    private static Map<AddrTools.AddrPrefix, String> loadData(final File inputData,
                                                              final String inputFormat)
        throws IOException
        {
        final Map<AddrTools.AddrPrefix, String> result = new HashMap<AddrTools.AddrPrefix, String>();

        if(INPUT_FORMAT_HOSTIP.equals(inputFormat))
            {
            // Parse MySQL dump format!

            // Get a buffered handle on the country table data.
            // This will be several DB rows as one or more lines
            // near the start of the table...
            final BufferedReader r1 = new BufferedReader(new FileReader(inputData));
            // Prepare map of countries that we are interested in...
            final Map<Integer, GeoUtils.CCTLD> countryNumberToCCTLD = new HashMap<Integer, GeoUtils.CCTLD>();
            String countriesRecord;
            try
                {
                while((countriesRecord = getNextRecordStarting(r1, "INSERT INTO countries ")) != null)
                    {
                    // Parse the countries data.
                    // Be forgiving of records that we can't parse easily,
                    // since there is some free text in country names.
                    //
                    // Expected row format (3 fields):
                    //   * Internal country number (int)
                    //   * Country name (String)
                    //   * Upper-case CCTLD (String)
                    final String[][] countriesData =
                        parseINSERTRecord(countriesRecord, 3, true);

                    for(int i = 0; i < countriesData.length; ++i)
                        {
                        final String c = countriesData[i][2].toLowerCase();
                        if(!GeoUtils.CCTLD.isSyntaticallyValidCcTLD(c)) { continue; }
                        final GeoUtils.CCTLD ccTLD = new GeoUtils.CCTLD(c);
                        if(GeoUtils.getCloseCCTLDs(ccTLD).isEmpty()) { continue; }
                        // Interesting country, so file it...
System.out.println(" Keeping data on country: " + Arrays.asList(countriesData[i]));
                        countryNumberToCCTLD.put(Integer.valueOf(countriesData[i][0]),
                                                 MemoryTools.intern(new GeoUtils.CCTLD(MemoryTools.intern(c))));
                        }

                    }
                }
            finally { r1.close(); }

            if(countryNumberToCCTLD.isEmpty())
                { throw new EOFException("Could not find `countries` table/data."); }

            // Now parse the main data records...
            // Get a new buffered handle on the country table data.
            final BufferedReader r2 = new BufferedReader(new FileReader(inputData));
            try
                {
                // Get the IP records in whatever order they are available...
                String ipRecord;
                final String prefix = "INSERT INTO ip4_";
                final int prefixLen = prefix.length();
//                int rowNum = 0;
                while((ipRecord = getNextRecordStarting(r2, prefix)) != null)
                    {
                    // Extract first byte of IP address (part of table name).
                    final int q = ipRecord.indexOf(' ', prefixLen);
                    final int byte1 = Integer.parseInt(ipRecord.substring(prefixLen, q), 10);
//System.out.println("Parsing entry "+(++rowNum)+" for addresses "+byte1+".X.X...  (Entries so far: "+result.size()+"...)");

                    final byte addrPrefix[] = { (byte) byte1, 0, 0 };

                    // Now find and store all the entries
                    // for countries that we are interested in.
                    //
                    // Expected row format (5 fields):
                    //   * Second byte of IPv4 address (int)
                    //   * Third byte of IPv4 address (int)
                    //   * Internal country number (int)
                    //   * Internal city number (int)
                    //   * Time/date of record.
                    final String[][] addrData =
                        parseINSERTRecord(ipRecord, 5, false);
                    for(int i = addrData.length; --i >= 0; )
                        {
                        final String[] row = addrData[i];
                        final GeoUtils.CCTLD cctld = countryNumberToCCTLD.get(Integer.valueOf(row[2]));
                        if(cctld == null) { continue; }

                        addrPrefix[1] = (byte) Integer.parseInt(row[0], 10);
                        addrPrefix[2] = (byte) Integer.parseInt(row[1], 10);
                        result.put(new AddrTools.AddrPrefix(addrPrefix),
                                   cctld.code);
                        }
                    }

System.out.println("Finished reading data from: " + inputData);
                }
            finally { r2.close(); }
            }
        else
            { throw new IllegalArgumentException("unrecognised input format"); }

        return(result);
        }

    /**Parse a MySQL insert record from a data dump.
     * This parses a MySQL record of the format:
     * <pre>
INSERT INTO `<i>table name</i>` VALUES (<i>field1,field2,...</i>),<i>...</i>(<i>...</i>);
     * </pre>
     *
     * @param record  the full line; never null
     * @param expectedFields  the expected number of expectedFields per record; non-negative
     * @param ignoreBadRecords  silently skip bad records that we cannot parse
     *     (for example because they contain our separator characters)
     *
     * @return zero or more rows each of the specified number of expectedFields;
     *     no null records or fields
     */
    private static String[][] parseINSERTRecord(final String record,
                                                final int expectedFields,
                                                final boolean ignoreBadRecords)
        {
        if((record == null) || (expectedFields < 0))
            { throw new IllegalArgumentException(); }

        if(!record.startsWith("INSERT INTO ") || (!record.endsWith(";")))
            { throw new IllegalArgumentException("malformatted line start/end"); }

        final int valuesStart = record.indexOf(" VALUES ");
        if(valuesStart == -1)
            { throw new IllegalArgumentException("malformatted line VALUES"); }

        final String core = record.substring(valuesStart + 8, record.length()-1).trim();
//System.out.println("core: " + core);

        // Chop into table rows (when multiple rows are on one line)...
        final StringTokenizer st = new StringTokenizer(core, ")(");
        final List<String[]> allRows = new ArrayList<String[]>(1+(st.countTokens()/2));
        for(int rowNum = 0; st.hasMoreTokens(); ++rowNum)
            {
            final String row = st.nextToken();
            if(",".equals(row)) { continue; /* Row separator. */ }
            final StringTokenizer stRow = new StringTokenizer(row, ",");
//System.out.println("ROW: `"+row+"'.");
            final int nFields = stRow.countTokens();
            if(nFields != expectedFields)
                {
                if(ignoreBadRecords) { continue; }
                throw new IllegalArgumentException("Wrong number of fields (expected "+expectedFields+", got "+nFields+") in record "+rowNum+": " + row);
                }
            final String rowResult[] = new String[nFields];
            for(int j = 0; j < nFields; ++j)
                {
                String s = stRow.nextToken();
                // Strip any surrounding quotes...
                if((s.length() > 2) && s.startsWith("'") && s.endsWith("'"))
                    { s = s.substring(1, s.length()-1); }
                rowResult[j] = s;
                }
//System.out.println("PARSED ROW: " + Arrays.asList(rowResult));
            allRows.add(rowResult);
            }

        final String[][] result = new String[allRows.size()][];
        allRows.toArray(result);
        return(result);
        }

    /**Returns the next line/record starting with the specified String, else null at EOF. */
    private static String getNextRecordStarting(final BufferedReader r,
                                                final String lineStart)
        throws IOException
        {
        String record;
        while((record = r.readLine()) != null)
            {
            if(record.startsWith(lineStart))
                { return(record); }
            }
        return(null);
        }

    /**An entry point to load the prefix map and write it out in a more compact format.
     * This may use lossy compression of the loaded values
     * where the saving is large
     * and the implied increase in routing/performance cost is low.
     * <p>
     * The new map is dumped to the file named as the first argument.
     * <p>
     * If a source of new input data is supplied as an optional second argument,
     * then it is merged into the existing data in memory,
     * and the dump will be the lossily-compressed result.
     * Only new input data corresponding to ccTLDs explicitly listed in
     * the "geo-proximity" data, ie whose getCloseCCTLDs() result is non-empty,
     * will be retained.
     * This input data file may need to be re-read in multiple passes.
     */
    public static final void main(final String args[])
        {
        if((args.length != 1) && (args.length != 3))
            {
            System.err.println("Usage: main filenameToDumpTo [newDataFileName FORMAT]");
            System.err.println("  Allowed values of FORMAT are: " + new ArrayList<String>(INPUT_FORMATS));
            Runtime.getRuntime().exit(1);
            return;
            }

        final String filename = args[0];

        System.out.println("Will dump result to file: " + filename);

        final SortedMap<AddrTools.AddrPrefix,String> ccTLDFromIPPrefix = GeoUtils.getCcTLDFromIPPrefix();
        System.out.println("Loaded static map size: " + ccTLDFromIPPrefix.size());
//        System.out.println("Loaded static map maximum key length: " + GeoUtils.ccTLDFromIPPrefixLongestKey);

        // The current static data...
        SortedMap<AddrTools.AddrPrefix,String> ccTLDFromIPPrefixWorking = ccTLDFromIPPrefix;

        // If new data has been supplied,
        // then load it now,
        // overriding extant static data in case of conflict.
        if(args.length == 3)
            {
            final File inputData = new File(args[1]);
            final String inputFormat = args[2];

            if(!INPUT_FORMATS.contains(inputFormat))
                {
                System.err.println("I do not understand input format: " + inputFormat);
                System.err.println("Valid formats are: " + new ArrayList<String>(INPUT_FORMATS));
                System.exit(1);
                }

            if(!inputData.exists() || !inputData.canRead() || !inputData.isFile())
                {
                System.err.println("Cannot open input file: " + inputData);
                System.exit(1);
                }

            try
                {
                // Copy existing static data...
                ccTLDFromIPPrefixWorking = new TreeMap<AddrTools.AddrPrefix, String>(ccTLDFromIPPrefix);
                // Override with new data...
                ccTLDFromIPPrefixWorking.putAll(loadData(inputData, inputFormat));
                }
            catch(final IOException e)
                {
                e.printStackTrace();
                System.err.println("Problem reading input data file: " + inputData + ": " + e.getMessage());
                System.exit(1);
                }
            catch(final Exception e)
                {
                e.printStackTrace();
                System.err.println("Problem with input data file: " + inputData + ": " + e.getMessage());
                System.exit(1);
                }
            }

        // Now transform/compress the map.
        System.out.println("Starting lossy compression on map of size: " + ccTLDFromIPPrefixWorking.size());
        final SortedMap<AddrTools.AddrPrefix,String> c1 = lossyCompressCcTLDFromIPPrefixMap(ccTLDFromIPPrefixWorking);
        System.out.println("Map size after lossy compression: " + c1.size());
        System.out.println("Entries removed: " + (ccTLDFromIPPrefixWorking.size() - c1.size()));

        // Now dump the new map...
        try
            {
            final PrintWriter pw = new PrintWriter(new FileWriter(filename));
            dumpPrefixMap(c1, pw);
            pw.flush();
            pw.close();
            }
        catch(final IOException e)
            {
            System.err.println("FAILED to write output file with IOException");
            e.printStackTrace();
            Runtime.getRuntime().exit(1);
            return;
            }
        }
    }
