/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.uploader;

import org.hd.d.pg2k.svrCore.*;

import javax.jnlp.FileContents;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.Map;
import java.util.WeakHashMap;

/**Immutable class to hold details of one file selected.
 * This is used during the selection/naming and upload phase.
 * <p>
 * All package-visible for access from the GUI classes.
 * <p>
 * Thread-safe.
 * <p>
 * The get/set status methods are dynamic and do not operate on the instance;
 * the class holds a weak-ref map from selected file details to an optional
 * String status/error, which "shadows" any computed error.
 */
final class SelectedFileDetails
    {
    /**Column for the (potentially mutable) proposed exhibit name. */
    public static final int COLNUM_EXHIBIT_NAME = 0;

    /**Column for the (immutable) local filename. */
    public static final int COLNUM_LOCAL_FILE_NAME = 1;

    /**Column for the (immutable) length in bytes. */
    public static final int COLNUM_DESCRIPTION = 2;

    /**Column for the (immutable) length in bytes. */
    public static final int COLNUM_BYTESLEN = 3;

    /**Column for the (dynamic) status string. */
    public static final int COLNUM_STATUS = 4;


    /**Column names for table, in order. */
    private static final String[] fileLoadTableColNames = {
        "Proposed Exhibit Name", "Filename", "Description", "Bytes", "Status/Error"
        };

    /**Return the number of columns; fixed for this model. */
    static int getColumnCount() { return(fileLoadTableColNames.length); }

    /**Return a column name, indexed from 0. */
    static String getColumnName(final int col) { return(fileLoadTableColNames[col]); }

    /**Get the value at the given column; null if none. */
    Object getValueAt(final UploaderLogic l, final int col)
        {
        switch(col)
            {
            case COLNUM_LOCAL_FILE_NAME:
                {
                try { return(fc.getName()); }
                catch(final IOException e) { throw new Error(e); } // Should not happen.
                }

            case COLNUM_EXHIBIT_NAME:
                { return(esa.filePath); }

            case COLNUM_BYTESLEN:
                { return(esa.length); }

            case COLNUM_DESCRIPTION:
                { return(getDescription()); }

            case COLNUM_STATUS:
                { return(getStatus(l)); }

            default: return(null); // Unexpected/unsupported column.
            }
        }

    /**Construct a new instance.
     * @param userFile underlying (validated) user file; never null
     * @param putativeExhibitName  syntatically-valid proposed exhibit name;
     *     never null
     * @param description  description text; may be "" but not null
     * @param hashMD5  MD5 hash of exhibit content;
     *     if null then it is computed here
     */
    SelectedFileDetails(final FileContents userFile,
                        final String putativeExhibitName,
                        final String description,
                        ROByteArray hashMD5)
        throws IOException
        {
        if((userFile == null) ||
           !ExhibitName.validNameSyntax(putativeExhibitName) ||
           (description == null))
            { throw new IllegalArgumentException(); }
        // TODO: more checking of args and file data?
        this.fc = userFile;
        this.localFilename = fc.getName();
        if((localFilename == null) || (localFilename.length() == 0))
            { throw new IllegalArgumentException(); }
        this.esa = new ExhibitStaticAttr(putativeExhibitName, userFile.getLength(), System.currentTimeMillis());
        this.description = description;

        // If need be, compute the MD5 hash here.
        if(hashMD5 == null)
            {
            final InputStream is = fc.getInputStream();
            try
                {
                final Tuple.Pair<Integer,ROByteArray> hashes = AccessionData.computeFullFileHashes(is);
                hashMD5 = hashes.second;
                }
            finally { is.close(); }
            }

        this.hashMD5 = hashMD5;
        }

    /**Get the FileContents; never null. */
    FileContents getFc() { return(fc); }

    /**Get the local filename; never null (nor ""). */
    String getLocalFilename() { return(localFilename); }

    /**Get the static exhibit details; never null. */
    ExhibitStaticAttr getEsa() { return(esa); }

    /**Get the description; never null but may be "". */
    String getDescription() { return(description); }

    /**Get the MD5 hash; never null. */
    ROByteArray getHashMD5() { return(hashMD5); }


    /**Set a status note for this item; use "" or null to clear.
     * Does not change the instance itself,
     * but stores any note in a private global map.
     */
    void setStatus(final String n)
        {
        if((n == null) || (n.length() == 0))
            { notes.remove(this); }
        else
            { notes.put(this, n); }
        }

    /**Get status/error: if this returns null all is OK, else the error is described.
     * The routine must be quick to run, eg it cannot recompute a file hash.
     * <p>
     * If a note has been set it takes precedence over any computed value.
     *
     * @return null if all seems OK, non-null error message otherwise
     */
    String getStatus(final UploaderLogic l)
        {
        if(l == null) { throw new IllegalArgumentException(); }

        final String note = notes.get(this);
        if(note != null) { return(note); }

        if(esa.length < 1)
            { return("file is zero-length"); }

        final int websvr_max_ex_bytes = l.getGenProps().getWEBSVR_MAX_EX_BYTES();
        if(esa.length > websvr_max_ex_bytes)
            { return("file is too long: limit (bytes): " + websvr_max_ex_bytes); }

        final AllExhibitProperties aep = l.getAep();

        // Check for hash clash with uploaded exhibit.
        if(l.checkIfAlreadyUploadedByHashMD5(hashMD5)) { return("identical to recently-uploaded exhibit"); }

        // Check for name clash with extant Gallery exhibit.
        final String clashingFullName = aep.aeid.getFullName(ExhibitName.getFileComponent(esa.filePath));
        assert((clashingFullName == null) || ExhibitName.validNameSyntax(clashingFullName));
        if(clashingFullName != null) { return("name clashes with extant exhibit: `" + clashingFullName + "'"); }

        // Check for hash clash with extant Gallery exhibit.
        final String clashingHashName = aep.getHashMD5ToName().get(hashMD5);
        assert((clashingHashName == null) || ExhibitName.validNameSyntax(clashingHashName));
        if(clashingHashName != null) { return("identical content to extant exhibit: `" + clashingHashName + "'"); }

        try
            {
            if(!fc.canRead()) { return("cannot read file"); }
            if(esa.length != fc.getLength()) { return("file length changed"); }
            }
        catch(final Exception e)
            {
            e.printStackTrace();
            return("Problem inspecting file: " + e.getMessage());
            }

        return(null); // All OK.
        }

    /**Local filename; never null (nor ""). */
    private final String localFilename;

    /**File on user side (should already be checked for validity); never null. */
    private final FileContents fc;

    /**Proposed upload name/details (valid unique new exhibit name); never null.
     * Volatile for thread-safety without locks.
     */
    private final ExhibitStaticAttr esa;

    /**Description; never null though may be "". */
    private final String description;

    /**MD5 hash of exhibit content; never null. */
    private final ROByteArray hashMD5;


    /**Equality is based on the (monocased) local filename and exhibit name.
     * Both names are used to help avoid surprising ambiguities.
     * <p>
     * The MD5 hash is considered a feature of the content
     * rather than of this object,
     * and is not used in the object hash or equality.
     */
    public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof SelectedFileDetails)) { return(false); }
        final SelectedFileDetails other = (SelectedFileDetails) obj;
        return(localFilename.equalsIgnoreCase(other.localFilename) &&
               esa.filePath.equals(other.esa.filePath));
        }

    /**The cached instance hash value; the hash may be expensive to compute. */
    private transient int hash;

    /**The hash code is based on the (monocased) local filename and exhibit name.
     * Both names are used to help avoid surprising ambiguities.
     * <p>
     * The MD5 hash is considered a feature of the content
     * rather than of this object,
     * and is not used in the object hash or equality.
     */
    public int hashCode()
        {
        int h = hash;
        if(h == 0)
            {
            // Compute and cache.
            hash = h = localFilename.toLowerCase().hashCode() ^ (37 * esa.filePath.hashCode());
            }
        return(h);
        }

    /**Global "error/status note" on instances of this class.
     * Uses a WeakHashMap so as not to retain a note unnecessarily
     * once the associated file has been dealt with and instance GCed.
     * <p>
     * Thread-safe.
     */
    private static final Map<SelectedFileDetails,String> notes =
        Collections.synchronizedMap(new WeakHashMap<SelectedFileDetails, String>());
    }
