/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.uploader;

import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.ROByteArray;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

/**Database of selected files.
 * Package-visible for direct manipulation by GUI classes.
 * <p>
 * Thread-safe.
 * <p>
 * A lock can be held on an instance to make compound atomic operations.
 */
final class SelectedFilesDB
    {
    /**Ordered random-access list of proposed exhibits to be uploaded; never null.
     */
    private final ArrayList<SelectedFileDetails> files = new ArrayList<SelectedFileDetails>();

    /**Set of all local file names; never null. */
    private final Set<String> localFilenames = new HashSet<String>();

    /**Set of all putative short exhibit names; never null. */
    private final Set<String> shortNames = new HashSet<String>();

    /**Set of all MD5 hashes; never null. */
    private final Set<ROByteArray> hashesMD5 = new HashSet<ROByteArray>();


    /**Add item, silently dropping duplicate file entries if possible.
     * Duplicate entries for a given local file or exhibit (short) name
     * will be ignored.
     *
     * @param newFile  new file to add; must be null
     */
    synchronized void add(final SelectedFileDetails newFile)
        {
        if(newFile == null) { throw new IllegalArgumentException(); }

        if(localFilenames.contains(newFile.getLocalFilename()) ||
           shortNames.contains(ExhibitName.getFileComponent(newFile.getEsa().filePath)) ||
           hashesMD5.contains(newFile.getHashMD5()))
            { return; /* Already present. */ }

        // Add to all internal tables...
        localFilenames.add(newFile.getLocalFilename());
        shortNames.add(ExhibitName.getFileComponent(newFile.getEsa().filePath));
        hashesMD5.add(newFile.getHashMD5());
        files.add(newFile);

        changed = true;
        }

    /**Remove the indicated item (shuffling any following elements down).
     * Silently ignores out-of-range requests.
     */
    synchronized void remove(final int index)
        {
        if((index < 0) || (index >= files.size())) { return; }

        // Remove from all internal tables.
        final SelectedFileDetails selectedFileDetails = files.get(index);
        localFilenames.remove(selectedFileDetails.getLocalFilename());
        shortNames.remove(ExhibitName.getFileComponent(selectedFileDetails.getEsa().filePath));
        hashesMD5.remove(selectedFileDetails.getHashMD5());
        files.remove(index);

        changed = true;
        }

    /**Remove the indicated item (shuffling any following elements down).
     * Silently ignores requests where the items is not present.
     */
    synchronized boolean remove(final SelectedFileDetails selectedFileDetails)
        {
        if(selectedFileDetails == null) { throw new IllegalArgumentException(); }

        // Remove from all internal tables.
        localFilenames.remove(selectedFileDetails.getLocalFilename());
        shortNames.remove(ExhibitName.getFileComponent(selectedFileDetails.getEsa().filePath));
        hashesMD5.remove(selectedFileDetails.getHashMD5());
        final boolean removed = files.remove(selectedFileDetails);

        if(removed) { changed = true; }
        return(removed);
        }

    /**Change the putative exhibit name to another new syntatically-valid one.
     * Attempting to set it to an invalid or existing (in this DB) name
     * (ignoring the directory component) will be rejected with an exception.
     * @param index  entry to be replaced
     * @param newFullName  new unique (in this DB) valid exhibit name
     */
    synchronized void setExhibitName(final int index, final String newFullName)
        {
        if(!ExhibitName.validNameSyntax(newFullName)) { throw new IllegalArgumentException(); }

        final String newShortName = ExhibitName.getFileComponent(newFullName);

        // Atomically replace old details with new.
        if(shortNames.contains(newShortName))
            { throw new IllegalArgumentException("duplicate name"); }

        try
            {
            final SelectedFileDetails oldSFD = files.get(index);
            final SelectedFileDetails newSFD =
                new SelectedFileDetails(oldSFD.getFc(), newFullName, oldSFD.getDescription(), oldSFD.getHashMD5());
            files.set(index, newSFD);
            final String oldShortName = ExhibitName.getFileComponent(oldSFD.getEsa().filePath);
            shortNames.remove(oldShortName);
            shortNames.add(newShortName);

            changed = true;
            }
        catch(final IOException e) { throw new Error(e); } // Should not happen.
        }

    /**Change the putative exhibit description.
     * @param index  entry to be replaced
     * @param newDescription  new description; never null
     */
    synchronized void setDescription(final int index, final String newDescription)
        {
        if(newDescription == null) { throw new IllegalArgumentException(); }

        try
            {
            final SelectedFileDetails oldSFD = files.get(index);
            if(oldSFD.getDescription().equals(newDescription))
                { return; /* No change; leave untouched. */ }
            final SelectedFileDetails newSFD =
                new SelectedFileDetails(oldSFD.getFc(), oldSFD.getEsa().filePath, newDescription, oldSFD.getHashMD5());
            files.set(index, newSFD);

            changed = true;
            }
        catch(final IOException e) { throw new Error(e); } // Should not happen.
        }

    /**Clear the whole DB. */
    synchronized void clear()
        {
        localFilenames.clear();
        shortNames.clear();
        hashesMD5.clear();
        files.clear();

        changed = true;
        }

    /**Get number of entries in DB; non-negative. */
    synchronized int size()
        {
        return(files.size());
        }

    /**Get given numbered entry by index; never null. */
    synchronized SelectedFileDetails get(final int index)
        {
        return(files.get(index));
        }

    /**Returns true if an exhibit in this DB already has this short name.
     * @param shortName  never null
     */
    synchronized boolean shortNameInUse(final String shortName)
        {
        return(shortNames.contains(shortName));
        }

    /**Returns true if this local name is already in this DB.
     * @param localFilename  never null
     */
    synchronized boolean localFilenameInUse(final String localFilename)
        {
        return(localFilenames.contains(localFilename));
        }


    /**Flag set true when the table is changed. */
    private boolean changed;

    /**Returns the "changed" flag value and clears it.
     * Can be used to arrange redrawing of a view of this data after an update.
     */
    synchronized boolean testAndClearChangedFlag()
        {
        final boolean result = changed;
        changed = false;
        return(result);
        }
    }
