/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.uploader;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataHTTPTunnelSource;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.uploader.UploadInfoBean;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;

import javax.jnlp.*;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;


/**Uploader "business-logic" holder.
 * Is designed to be GUI-free and just contain the logic.
 * <p>
 * Package visible since need be seen only by the main GUI class.
 */
final class UploaderLogic
    {
    /**Package-visible constructor since need be seen only by the main GUI class.
     * @param logger  reference to central logger; never null
     */
    UploaderLogic(final SimpleLoggerIF logger)
        {
        assert(logger != null);
        this.logger = logger;

        // Get access to basic services.
        BasicService basicService = null;
        try { basicService = (BasicService) ServiceManager.lookup("javax.jnlp.BasicService"); }
        catch(final UnavailableServiceException e) { }
        bs = basicService;

        // Get access to persistence management.
        PersistenceService persistenceService = null;
        try { persistenceService = (PersistenceService) ServiceManager.lookup("javax.jnlp.PersistenceService"); }
        catch(final UnavailableServiceException e) { }
        ps = persistenceService;

        // Get access to file open services.
        FileOpenService fileOpenService = null;
        try { fileOpenService = (FileOpenService) ServiceManager.lookup("javax.jnlp.FileOpenService"); }
        catch(final UnavailableServiceException e) { }
        fos = fileOpenService;

        // Get access to single-instance services.
        SingleInstanceService singleInstanceService = null;
        try { singleInstanceService = (SingleInstanceService) ServiceManager.lookup("javax.jnlp.SingleInstanceService"); }
        catch(final UnavailableServiceException e) { }
        sis = singleInstanceService;

        // Get access to extended services.
        ExtendedService extendedService = null;
        try { extendedService = (ExtendedService) ServiceManager.lookup("javax.jnlp.ExtendedService"); }
        catch(final UnavailableServiceException e) { }
        exs = extendedService;

        if((ps != null) && (bs != null))
            {
            // Try to reload the properties/preferences...
            // This should be quick,
            // and doing it here should avoid races with the UI.
            try
                {
                final FileContents fc = ps.get(makePropsURL(bs));
                final InputStream inputStream = fc.getInputStream();
                try { props.loadFromPersistentData(inputStream); }
                finally { inputStream.close(); }
                }
            catch(final IOException e)
                {
                logger.log("No saved properties/preferences to reload (yet).");
                }
            }
        else
            {
            // If there is no persistence,
            // then create a default AEP,
            // for testing, and to get users started.
            final Map<String, ExhibitStaticAttr> infoMap = new HashMap<String, ExhibitStaticAttr>();
            infoMap.put(DEFAULT_EXHIBIT_NAME, new ExhibitStaticAttr(DEFAULT_EXHIBIT_NAME, 1, System.currentTimeMillis()));
            final AllExhibitProperties defaultAEP =
                new AllExhibitProperties(new ExhibitPropsGlobalImmutable(),
                                         new AllExhibitImmutableData(infoMap, 1),
                                         new HashMap<String, ExhibitPropsLoadable>(),
                                         new HashMap<String, ExhibitPropsComputable>());

            aep = defaultAEP;
            logger.log("Created default AEP.");

            // Set ID to match for now.
            props.setUserID(ExhibitName.getAuthorComponent(DEFAULT_EXHIBIT_NAME));
            }

        // Create our tunnel endpoint if at least basic service is available.
        // Create the tunnel before any async threads are running.
        URL masterURL = null;
        if(bs != null)
            {
            try { masterURL = new URL(bs.getCodeBase(), CoreConsts.TUNNEL_URI); }
            catch(final Exception e) { }
            }
        tunnel = ((bs == null) || (masterURL == null)) ? null :
            new ExhibitDataHTTPTunnelSource(masterURL.toString(), logger);

        // Create our worker threads for uploads.
        // They get start()ed after the constructor has completed, by startup().
        final List<Thread> workers = new ArrayList<Thread>(MAX_CONC_UPLOADS);
        for(int i = 0; i < MAX_CONC_UPLOADS; ++i)
            { workers.add(new UploaderWorkerThread(i)); }
        // Store immutable List view of unstarted uploader threads.
        uploaderWorkerThreads = Collections.unmodifiableList(workers);
        }

    /**Reference to central logger; never null.
     * We make this package-visible,
     * though the preferred access is via the Main class.
     */
    final SimpleLoggerIF logger;

    /**Persistence properties; never null.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final UploaderProps props = new UploaderProps();


    /**Handle on JWS basic service; null if none.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final BasicService bs;

    /**Handle on JWS persistence service; null if none.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final PersistenceService ps;

    /**Handle on JWS file-open service; null if none.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final FileOpenService fos;

    /**Handle on JWS singleton service; null if none.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final SingleInstanceService sis;

    /**Handle on JWS extended service; null if none.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final ExtendedService exs;


    /**Our end of main tunnel to server; null if no tunnel possible.
     * Package-visible so as to be directly usable by GUI classes.
     */
    final ExhibitDataHTTPTunnelSource tunnel;


    /**Uploader worker thread.
     * Private to this class.
     */
    private final class UploaderWorkerThread extends Thread
        {
        UploaderWorkerThread(final int index)
            {
            super("upload worker thread " + index);

            if((index < 0) || (index > MAX_CONC_UPLOADS))
                { throw new IllegalArgumentException(); }

            setDaemon(true); // Daemon by default.

            this.index = index;
            }

        /**Index of this worker thread. */
        final int index;

        /**Time an idle worker sleeps for in ms; strictly positive. */
        private static final int WORKER_SLEEP_MS = 1011;


        /**Do any uploading; sleep when nothing to do.
         */
        public void run()
            {
            // Run forever...
            for( ; ; )
                {
                // At the start of each loop our slot must be empty.
                assert(inProgressSlots[index] == null);

                try
                    {
                    // Short-cut when no work to be done...
                    // If nothing at all to upload
                    // or no basic service
                    // or invalid user credentials
                    // or broken connection to the server
                    // then sleep awhile...
                    if((uploadingFiles.size() == 0) ||
                       (bs == null) ||
                       (!ExhibitName.validAuthorSyntax(userID)) ||
                       (tunnel.isBroken()))
                        {
                        // Sleep then go back and check again.
                        // Use a randomised time to help avoid collisions.
                        Thread.sleep(WORKER_SLEEP_MS + Rnd.fastRnd.nextInt(WORKER_SLEEP_MS));
                        continue;
                        }

                    // Find the first available item in the upload queue
                    // with no error,
                    // and not in any other upload slot (ie not being uploaded).
                    //
                    // Use a lock on inProgressSlots[]
                    // to avoid selection races with more than one worker
                    // starting to upload the same file.
                    //
                    // Grab the uploadingFiles lock inside it.
                    AtomicLong bytesUploaded = null;
                    UploadStatus stats = null;
                    SelectedFileDetails selectedFileDetails = null;
                    synchronized(inProgressSlots)
                        {
                        synchronized(uploadingFiles)
                            {
                            selectLoop: for(int i = uploadingFiles.size(); --i >= 0; )
                                {
                                selectedFileDetails = uploadingFiles.get(i);

                                // This file is not a candidate
                                // if it has any error or other non-null status.
                                if(selectedFileDetails.getStatus(UploaderLogic.this) != null)
                                    { continue; }

                                // This file is not a candidate
                                // if any (other) worker is already uploading it.
                                for(int j = MAX_CONC_UPLOADS; --j >= 0; )
                                    {
                                    if(selectedFileDetails.equals(inProgressSlots[j]))
                                        { continue selectLoop; }
                                    }

                                // OK, mark this as uploading
                                // and copy it to our slot.
                                selectedFileDetails.setStatus("Uploading... (worker "+index+")");
                                bytesUploaded = new AtomicLong(-1); // Not started data yet...
                                inProgressSlots[index] = stats = new UploadStatus(selectedFileDetails, bytesUploaded);
                                break;
                                }
                            }
                        }

                    // If we didn't find anything to do, then sleep.
                    // This might happen in a race for work with other threads.
                    if(stats == null)
                        {
                        // Use a randomised time to help avoid collisions.
                        Thread.sleep(WORKER_SLEEP_MS + Rnd.fastRnd.nextInt(WORKER_SLEEP_MS));
                        continue;
                        }

                    // We expect all the refs to be filled in (non-null).
                    assert(selectedFileDetails != null);
                    assert(bytesUploaded != null);

                    // Note time before starting this upload.
                    final long startTime = System.currentTimeMillis();

                    URLConnection uc = null; // openURLConnection;
                    HttpURLConnection huc = null;
                    try
                        {
                        // Compute path at which batch-upload JSP is mounted...
                        final URL u = (new URL(bs.getCodeBase(), "/_upload/batchupload"));
                        uc = u.openConnection();

                        if(!(uc instanceof HttpURLConnection))
                            { throw new IOException("incorrect connection type from URL: " + u); }
                        huc = (HttpURLConnection) uc;

                        // Set properties of connection.
                        uc.setDoInput(true);
                        uc.setDoOutput(true);
                        uc.setUseCaches(false);
                        // Wait only a short while for a connection.
                        // If we can't even get a connection quickly
                        // then we might well have problems with uploads anyway.
                        uc.setConnectTimeout(2001 + (UploaderConsts.BATCH_UPLOAD_MAX_IDLE_MS / 2));
                        // Make read timeout MUCH longer than others
                        // so as to try hard not to lose uploads in progress.
                        uc.setReadTimeout(12001 + (UploaderConsts.BATCH_UPLOAD_MAX_IDLE_MS * 2));
                        // Don't bother the user for any interaction.
                        uc.setAllowUserInteraction(false);

                        // Set auth credentials...
                        uc.setRequestProperty("X-" + UploaderConsts.USERID_FIELD_NAME, String.valueOf(userID));
                        uc.setRequestProperty("X-" + UploaderConsts.PASSWD_FIELD_NAME, String.valueOf(passwd));

                        // We know what the content length will be
                        // so set it here to allow some HTTP optimisations.
                        final int contentLength = UploaderUtils.computeLengthOfRequestFromBatchUploadClient(
                            selectedFileDetails.getEsa(),
                            selectedFileDetails.getHashMD5(),
                            selectedFileDetails.getDescription());
                        huc.setRequestProperty("Content-Length", String.valueOf(contentLength));
                        huc.setFixedLengthStreamingMode(contentLength);

                        // This must be a POST request.
                        huc.setRequestMethod("POST");

                        // No redirects should be needed.
                        huc.setInstanceFollowRedirects(false);

                        // Force connection now...
                        uc.connect();
                        }
                    catch(final Exception e)
                        {
                        e.printStackTrace();
                        logger.log("CANNOT CONNECT TO SERVER: " + e.getMessage());

                        // Check for a 403 (forbidden) error code.
                        if(huc != null)
                            {
                            if(huc.getResponseCode() == 403)
                                {
                                logger.log("CANNOT CONNECT TO SERVER: uploads disabled, or user name or password wrong?: " + e.getMessage());
                                }
                            }

                        // Whinge but don't penalise the file itself
                        // if we could not even get a server connection.
                        inProgressSlots[index] = null;
                        selectedFileDetails.setStatus(null); // Not the file's problem...
                        // Use a randomised time to help avoid collisions.
                        // A longer-than-usual time for the server to recover.
                        Thread.sleep(3 * WORKER_SLEEP_MS + Rnd.fastRnd.nextInt(5 * WORKER_SLEEP_MS));
                        continue;
                        }

                    IOException writeIOException = null;

                    // Send request and data up the wire...
                    final OutputStream os = uc.getOutputStream();
                    try
                        {
                        final InputStream exhibitDataStream =
                            new BufferedInputStream(selectedFileDetails.getFc().getInputStream());
                        try
                            {
                            // Not within any locks
                            // (to allow good concurrency),
                            // upload the new exhibit...
                            UploaderUtils.sendRequestFromBatchUploadClient(
                                os,
                                selectedFileDetails.getEsa(),
                                selectedFileDetails.getHashMD5(),
                                selectedFileDetails.getDescription(),
                                exhibitDataStream,
                                bytesUploaded);
                            os.flush(); // Force all data upstream.
                            }
                        finally { exhibitDataStream.close(); /* Release resources. */ }
                        }
                    catch(final IOException e)
                        {
                        // Note error in upload...
                        selectedFileDetails.setStatus("ERROR: " + e.getMessage());
                        writeIOException = e;
                        }
                    finally { os.close(); }

                    final InputStream is = uc.getInputStream();
                    try
                        {
                        // Collect OK/FAIL response from server...
                        if(!selectedFileDetails.getEsa().filePath.equals(UploaderUtils.decodeResponseToBatchUploadClient(is)))
                            { throw new IOException("got invalid response to upload from server"); }
                        }
                    finally { is.close(); }

                    // If we got an exception sending then rethrow it now.
                    if(writeIOException != null) { throw writeIOException; }

                    // If OK then remove the file entirely
                    // (though record the hash to avoid uploading a dup later).
                    selectedFileDetails.setStatus(null); // Done!
                    uploadedHashesMD5.add(selectedFileDetails.getHashMD5()); // Avoid dup upload.
                    uploadingFiles.remove(selectedFileDetails); // Uploaded!

                    // Clear slot...
                    inProgressSlots[index] = null;

                    // Record finish time including most of the overheads
                    // to give a conservative Bps throughput measure.
                    final long endTime = System.currentTimeMillis();
                    final long durationMS = endTime - startTime;

                    final long length = selectedFileDetails.getEsa().length;
                    final long Bps = (1000*length + 500) / Math.max(1, durationMS);
                    updateSmoothedUploadSpeedBps(Bps);

                    logger.log("Finished uploading exhibit #"+getCompletedUploadsThisSession()+" at "+Bps+"Bps:  "+length+" bytes in "+durationMS+"ms.");

                    // If no files left to upload then we could disconnect...
                    if((uploadingFiles.size() == 0) && (huc != null))
                        { huc.disconnect(); }
                    }
                catch(final Throwable e)
                    {
                    // Log the problem.
                    e.printStackTrace();
                    logger.log("UPLOAD ERROR: " + e.getMessage());

                    // Mark any in-progress item as in error.
                    final UploadStatus inProgress = inProgressSlots[index];
                    if(inProgress != null)
                        { inProgress.file.setStatus("ERROR: " + e.getMessage()); }

                    // Kill the content of our slot.
                    inProgressSlots[index] = null;

                    // Sleep a bit in case the error condition keeps occurring.
                    try { Thread.sleep(2*WORKER_SLEEP_MS + Rnd.fastRnd.nextInt(WORKER_SLEEP_MS)); }
                    catch(final InterruptedException ignored) { }
                    }
                }
            }
        }


    /**Time-constant for updating smoothedUploadSpeedBps; strictly positive.
     * The larger this is, the more smoothed is the computed value.
     * <p>
     * A value of 1 gives no smoothing at all.
     * <p>
     * A power of two may result in very slightly faster computations.
     * <p>
     * A value in the range 2 to 64 is probably good.
     */
    private static final int TC_SUSB = 16;

    /**Current smoothed upload performance in bytes-per-second, initially -1; never null.
     * Initially -1 before any exhibit is uploaded.
     * Thereafter updated via a smoothing (low-pass-filter) algorithm
     * to reflect recent history.
     * <p>
     * This represents a conservative performance measure,
     * ie including overheads such as connection set-up time,
     * but not dead time between uploads.
     * <p>
     * AtomicLong allows thread-safe lockless access for read and write.
     */
    private final AtomicLong smoothedUploadSpeedBps = new AtomicLong(-1);

    /**Get current smoothed upload performance in bytes-per-second, initially -1.
     * Initially -1 before any exhibit is uploaded.
     * Thereafter updated via a smoothing (low-pass-filter) algorithm
     * to reflect recent history.
     * <p>
     * This represents a conservative performance measure,
     * ie including overheads such as connection set-up time,
     * but not dead time between uploads.
     * <p>
     * This is thread-safe and lock-free.
     * <p>
     * Package-visible to allow access by GUI classes.
     */
    final long getSmoothedUploadSpeedBps()
        { return(smoothedUploadSpeedBps.get()); }

    /**Update our notion of mean upload speed (in bytes per second).
     * The argument must be non-negative,
     * and should be conservative,
     * ie including overheads such as connection set-up time,
     * but not dead time between uploads.
     * <p>
     * This will be used to update a smoothed performance measure.
     * <p>
     * This tries to allow for concurrent uploads,
     * ie where the total throughput is shared by several workers.
     * <p>
     * This is thread-safe and lock-free.
     *
     * @param Bps  non-negative speed of upload in bytes per second
     */
    private void updateSmoothedUploadSpeedBps(long Bps)
        {
        if(Bps < 0) { throw new IllegalArgumentException(); }

        // We may adjust Bps to allow for multiple concurrent workers.
        final int filesStillQueuedToUpload = uploadingFiles.size();
        if((MAX_CONC_UPLOADS > 1) && (filesStillQueuedToUpload > 0))
            {
            // We probably just finished uploading a file,
            // and were probably uploading up to the maximum concurrency
            // that that previous queue length implies.
            Bps *= Math.max(1, Math.min(MAX_CONC_UPLOADS, filesStillQueuedToUpload+1));
            }

        // Retry until we manage to set the value...
        for( ; ; )
            {
            final long current = smoothedUploadSpeedBps.get();
            // Compute a smoothed updated value,
            // though in the first instance (value is -1)
            // set the value to the one passed in with no smoothing.
            final long updated = (current < 0) ? Bps :
                                    ((current + (Bps * (TC_SUSB-1))) / TC_SUSB);
            if(smoothedUploadSpeedBps.compareAndSet(current, updated))
                {
                return; /* Update succeeded, so return! */
                }
            }
        }


    /**Set of MD5 hashes of exhibits already uploaded (or rejected as duplicate content); never null.
     * This information currently does not persist from one run to the next,
     * so can be cleared by restarting the app.
     * <p>
     * Private so as to prevent accidental deletion of entries!
     * <p>
     * Thread-safe.
     */
    private final Set<ROByteArray> uploadedHashesMD5 =
        Collections.synchronizedSet(new HashSet<ROByteArray>());

    /**Check (using the content MD5 hash) if a putative exhibit has already been uploaded (or rejected as duplicate content).
     * Package-visible so as to be directly usable by GUI classes.
     */
    boolean checkIfAlreadyUploadedByHashMD5(final ROByteArray hashMD5)
        {
        if((hashMD5 == null) || (hashMD5.length() != 16))
            { throw new IllegalArgumentException(); }
        return(uploadedHashesMD5.contains(hashMD5));
        }

    /**Get number of exhibits successfully uploaded this session; non-negative. */
    int getCompletedUploadsThisSession()
        { return(uploadedHashesMD5.size()); }


    /**Details of one selected file being uploaded; immutable except for progress value. */
    static final class UploadStatus
        {
        /**Selected file; never null. */
        final SelectedFileDetails file;

        /**Current upload progress (bytes uploaded); never null.
         * Private so that only get/read access is possible.
         */
        private final AtomicLong progress;

        /**Get bytes of this exhibit uploaded so far.
         * Thread-safe.
         *
         * @return -1 if upload of data not yet started,
         *     Long.MAX_VALUE once all data has been sent,
         *     else the number of bytes sent/uploaded
         */
        long getBytesUploaded() { return(progress.get()); }

        /**Create an instance.
         * All args must be non-null and unique to this instance.
         */
        UploadStatus(final SelectedFileDetails selectedFile,
                     final AtomicLong bytesUploaded)
            {
            if((selectedFile == null) || (bytesUploaded == null))
                { throw new IllegalArgumentException(); }
            file = selectedFile;
            progress = bytesUploaded;
            }
        }


    /**Maximum simultaneous uploads (to overcome network latency, etc); strictly positive.
     * A value of much higher than 2 is more likely to cause network congestion
     * than actually fix anything, so a value of 1 or 2 is probably optimal.
     */
    static final int MAX_CONC_UPLOADS = 2;

    /**Our (daemon) uploader worker threads started by startup(); never null. */
    private final List<Thread> uploaderWorkerThreads;

    /**Our current uploads-in-progress; never null and always MAX_CONC_UPLOADS long.
     * Each uploader worker thread controls the content of the slot/index
     * indicated by its own index.
     * <p>
     * A slot is empty when not in use; all slots are initially empty.
     * <p>
     * Private so as to prevent accidental deletion of entries!
     * <p>
     * All access to the contents of this array should grab a lock
     * on this array to ensure that an up-to-date/valid view is seen.
     * <p>
     * Worker threads keep a lock on this while looking for a
     * file to start uploading,
     * and thus the uploadingFiles instance lock may be grabbed inside this one;
     * the reverse is not allowed so as to avoid deadlocks.
     */
    private final UploadStatus inProgressSlots[] = new UploadStatus[MAX_CONC_UPLOADS];

    /**Get the current contents of an in-progress slot; may return null.
     * Package-visible so as to be directly usable by GUI classes.
     */
    UploadStatus getInProgressSlot(final int index)
        { synchronized(inProgressSlots) { return(inProgressSlots[index]); } }


    /**Name under the codebase where we perist the properties/preferences. */
    private static final String FNAME_MAIN_PROPS = "main.properties";

    /**Name under the codebase where we perist the AEP. */
    private static final String FNAME_AEP = "aep.ser.gz";


    /**Database of selected files; never null.
     * Package-visible for direct manipulation by GUI classes.
     * <p>
     * Thread-safe.
     */
    final SelectedFilesDB selectedFiles = new SelectedFilesDB();

    /**Database of files queued for upload; never null.
     * Package-visible for direct manipulation by GUI classes.
     * <p>
     * Thread-safe.
     */
    final SelectedFilesDB uploadingFiles = new SelectedFilesDB();


    /**Last time the user was active in this run, eg in the UI; initially zero.
     * Volatile to allow access without a lock.
     */
    private volatile long userLastActive;

    /**Get the last time the user was active in this run, eg in the UI; initially zero.
     * No harm in letting everyone see this, so is public.
     * <p>
     * When the user has not been active for a long time,
     * some activities, such as polling the server, may halt or slow down
     * to conserve resources.
     */
    public long getUserLastActive() { return(userLastActive); }

    /**Note user as active.
     * Made package-visible (ie not public) for security.
     * <p>
     * This may be triggered/called by a number of events
     * that indicate that the user is active.
     * <p>
     * When the user has not been active for a long time,
     * some activities, such as polling the server, may halt or slow down
     * to conserve resources.
     */
    void setUserLastActive(final String doingWhat)
        {
        userLastActive = System.currentTimeMillis();
//logger.log("UI active: " + doingWhat + ": " + (new Date())); }
        }

    /**Default exhibit name when there is an empty AEP. */
    final String DEFAULT_EXHIBIT_NAME = "places-and-sights/example-ANON.jpg";

    /**Load any (large) persisted data from a previous execution and start a background worker thread.
     * Do this as early as possible after construction,
     * and preferably before allowing (much) user interaction.
     * <p>
     * We spin this off into a separate thread to avoid blocking startup.
     * <p>
     * This should be called at most once.
     * <p>
     * Package-visible so as to be directly usable by GUI classes.
     */
    void startup()
        {
        final Thread th = new Thread("UploaderLogic startup()"){
            /**Do load work as background thread. */
            public final void run()
                {
                if((ps != null) && (bs != null))
                    {
                    // Try to reload the AEP...
                    // This may take a while.
                    logger.log("Reloading saved AEP...");
                    try
                        {
                        final FileContents fc = ps.get(makeAEPURL(bs));
                        final ObjectInputStream ois = new ObjectInputStream(new GZIPInputStream(fc.getInputStream()));
                        try
                            {
                            final AllExhibitProperties allExhibitProperties = (AllExhibitProperties) ois.readObject();
                            if(null != allExhibitProperties)
                                {
                                aep = allExhibitProperties;
                                logger.log("Reloaded saved AEP.");
                                }
                            }
                        finally { ois.close(); }
                        }
                    catch(final Exception e)
                        {
                        logger.log("No valid saved AEP to reload.");
                        }
                    }

                // Once we have finished trying to load any saved AEP,
                // start a (daemon) poller thread for non-UI async activity
                // such as fetching any updated AEP over the tunnel.
                // After a short delay, run approximately every second or so.
                // This is created and started AFTER the UI object construction
                // is complete to avoid any unpleasant races.
                final java.util.Timer timer = new java.util.Timer(true);
                timer.schedule(new TimerTask(){
                    /**The action to be performed by this timerUI task. */
                    public final void run() { poll(); }
                    }, 3000, 2000 + Rnd.fastRnd.nextInt(1000));

                // Finally start our uploader worker threads...
                for(final Thread t : uploaderWorkerThreads)
                    { t.start(); }
                }
            };
        th.start();
        }

    /**Make the props persistence URL; only viable when in JWS/JNLP. */
    private static URL makePropsURL(final BasicService bs) throws IOException
        {
        final URL codebase = bs.getCodeBase();
        final URL urlProps = new URL(codebase, FNAME_MAIN_PROPS);
        return(urlProps);
        }

    /**Make the AEP persistence URL; only viable when in JWS/JNLP. */
    private static URL makeAEPURL(final BasicService bs) throws IOException
        {
        final URL codebase = bs.getCodeBase();
        final URL urlAEP = new URL(codebase, FNAME_AEP);
        return(urlAEP);
        }

    /**Our private copy of the (less sensitive) valid user ID, or null if none or invalid ID/pass presented. */
    private volatile String userID;

    /**Our private note of the password supplied. */
    private volatile String passwd;

    /**Set whenever the userID changes; cleared when we have attempted to test the connection.
     * This is <em>not</em> critical, and may suffer races.
     * <p>
     * This is simply a device to try to get a new user ID/passwd tested
     * quietly in the background.
     */
    private volatile boolean userIDChanged;

    /**Set authentication information for this user.
     * To set authentication arguments must be non-null and non-empty,
     * else all arguments must be null to clear authentication data.
     * <p>
     * Package-visible so as to be directly usable by the main GUI class.
     * <p>
     * Ignored if there is no tunnel.
     */
    void setAuthenticationInfo(final String userID, final char[] passwd)
        {
        if(tunnel == null)
            { return; }

        // Any null/empty argument implies unsetting the authentication data.
        if((userID == null) || (userID.length() == 0) ||
           (passwd == null) || (passwd.length == 0))
            {
            this.userID = null;
            this.passwd = null;
            tunnel.setAuthenticationInfo(null, null);
            return;
            }

        // If the arguments passed are syntactically invalid,
        // then ignore them and clear any extant authentication data instead.
        if(!ExhibitName.validAuthorSyntax(userID) ||
           (passwd.length < CoreConsts.MIN_PASSWORD_LEN) ||
           (passwd.length > CoreConsts.MAX_PASSWORD_LEN))
            {
            this.userID = null;
            this.passwd = null;
            tunnel.setAuthenticationInfo(null, null);
            return;
            }

        // The authentication data provided is syntactically OK.
        // Set it for the tunnel.
        // Also make it available for the upload worker threads.
        final String s = new String(passwd);
        final boolean changed =
            !userID.equals(this.userID) ||
            !this.passwd.equals(s);
        this.userID = userID;
        this.passwd = s;
        tunnel.setAuthenticationInfo(userID, passwd);

        if(changed)
            {
            // We take an ID change to be indicative of user acitivity.
            setUserLastActive("user ID changed");

            // Set this flag *after* setting the value on the tunnel, etc.
            userIDChanged = true;
            }
        }

    /**Exhibit properties; never null.
     * Set by poll().
     * <p>
     * Volatile so as to be safe to access without a lock if need be.
     */
    private volatile AllExhibitProperties aep = new AllExhibitProperties();

    /**Get the current AEP (exhibit properties); never null. */
    AllExhibitProperties getAep() { return(aep); }

    /**Generic properties; never null.
     * Set and used by poll().
     * <p>
     * Volatile so as to be safe to access without a lock if need be.
     */
    private volatile GenProps gp = new GenProps();

    /**Get the current set of GenProps (generic system properties); never null. */
    GenProps getGenProps() { return(gp); }

    /**Note of polled connection state; TRUE is good, FALSE is bad, null is unknown.
     * Set and used by poll().
     * <p>
     * Volatile so as to be safe to access without a lock if need be.
     */
    private volatile Boolean connectionGood;

    /**Time we last polled for GP/AEP updates.
     * Set and used by poll().
     * <p>
     * Volatile so as to be safe to access without a lock if need be.
     */
    private volatile long lastAEPUpdatePoll;

    /**Maximum interval between polls for new values of AEP (and GP); strictly positive.
     * Relatively short for this interactive application.
     * <p>
     * We may suspend polling entirely if the UI/app appears to be idle.
     */
    private static final int MAX_AEP_POLL_MS = 10 * 60 * 1000;

    /**Called (by a daemon thread) to perform async activity.
     * This is not Swing-safe, and does not block Swing if blocked.
     * <p>
     * This is not called until construction is complete.
     * <p>
     * Package-visible so as to be directly usable by the main GUI class.
     */
    void poll()
        {
        try
            {
            // Perform tunnel processing, if any.
            final ExhibitDataHTTPTunnelSource t = tunnel;
            if((t != null) && (userID != null))
                {
                try
                    {
                    // Perform background activity for the tunnel, if extant.
                    t.poll(gp);

                    // Try a new/updated user/password combo in the background.
                    // Force recheck if state is not known to be good.
                    if((userID != null) &&
                       (userIDChanged || !Boolean.TRUE.equals(connectionGood)))
                        {
                        // Don't try again unless ID/passwd is changed again...
                        userIDChanged = false;

                        try
                            {
                            // Try a real connection if we're not seeing errors.
                            t.doNOOP(false);

                            // Poll was successful.
                            connectionGood = Boolean.TRUE;
                            logger.log("Connection to server established.");
                            }
                        catch(final IOException e)
                            {
                            connectionGood = Boolean.FALSE;
                            logger.log("Connection to server bad/unauthorised: " + e.getMessage());
                            }
                        }

                    // If we have a good connection
                    // and we have no AEP (or GP)
                    // or it is a while since we last polled for them
                    // then try to fetch updates/changes to them.
                    // We use the same short-ish refetch interval
                    // for both GenProps and AEP
                    // for this very interactive app.
                    //
                    // We don't do this if the user hasn't been active
                    // since the last poll.
                    // This conserves resources and enhances security.
                    if(Boolean.TRUE.equals(connectionGood) &&
                       (getUserLastActive() > lastAEPUpdatePoll) &&
                       ((aep.aeid.length == 0) || (gp.timestamp == 0) || // Missing vital data...
                            (System.currentTimeMillis() > lastAEPUpdatePoll +
                                  Math.min(MAX_AEP_POLL_MS, gp.getWEBSVR_MIN_EX_IMATTR_RECHECK_MS()))))
                        {
                        logger.log("PLEASE WAIT: fetching generic system properties from server...");
                        final GenProps gpNew = t.getGenProps(gp.timestamp);
                        if(gpNew != null)
                            {
                            gp = gpNew;
                            logger.log("Updated GenProps as of: " + new Date(gpNew.timestamp) + ", exhibit max size: " + gpNew.getWEBSVR_MAX_EX_BYTES());
                            assert(gpNew.timestamp != 0);
                            }

                        logger.log("PLEASE WAIT: fetching AllExhibitProperties from server... (may take several of your Earth minutes)");
                        final long fetchStart = System.currentTimeMillis();
                        final AllExhibitProperties aepNew = t.getAllExhibitProperties(aep, true);
                        final long fetchEnd = System.currentTimeMillis();
                        logger.log("Done fetching AEP... ("+((fetchEnd-fetchStart+500)/1000)+"s)");
                        if(aepNew != null)
                            {
                            // Save new AEP *before* we try to cache it,
                            // since cacheing may fail due to JNLP restrictions.
                            aep = aepNew;

                            // Now attempt to cache the AEP locally.
                            logger.log("Cacheing AEP locally...");
                            try
                                {
                                cacheAEP(aepNew);
                                logger.log("Cached AEP successfully.");
                                }
                            catch(final Exception e)
                                { logger.log("Failed to cache AEP locally: " + e.getMessage()); }
                            }

                        // Polled successfully for AEP.
                        // Postpone next poll...
                        lastAEPUpdatePoll = System.currentTimeMillis();
                        }
                    }
                catch(final InterruptedIOException e)
                    {
                    // The status remains unchanged because
                    // InterruptedIOException is an invitation to retry soon.
                    logger.log("RPC interrupted, will retry: " + e.getMessage());
                    }
                catch(final IOException e)
                    {
                    // This status is unknown rather than bad,
                    // as other things than a bad authorisation
                    // such as a temporary server overload
                    // may cause an IOException.
                    connectionGood = null;
                    logger.log("Connection to server status unknown: " + e.getMessage());
                    e.printStackTrace();
                    }
                }

            // Save any internal state if it has changed...
            final byte[] persistentDataIfChanged = props.getPersistentDataIfChanged();
            if(null != persistentDataIfChanged)
                {
if(IsDebug.isDebug) { logger.log("Properties data to save (bytes): " + persistentDataIfChanged.length); }
                saveProperties();
                }
            }
        catch(final Exception e)
            {
            // Absorb any exceptions to avoid killing the timer.
            logger.log("Unexpected exception: " + e.getMessage());
            e.printStackTrace();
            }
        }

    /**Number of bytes of expansion room we allow for one growth in the AEP; non-negative.
     * We can do this so that we do not ask the JNLP/user too often
     * for expanded storage space.
     * <p>
     * Should typically allow for a few days' growth as new exhibits are added.
     */
    private static final int AEP_EXPANSION_BYTES = 30001;

    /**Save/cache local copy of AEP in persistent store.
     * This is only possible if the PersistenceService is available,
     * else this call is ignored.
     * <p>
     * This save may be prevented on grounds of size, etc, by the JNLP runtime.
     * <p>
     * This should NOT be called by more than one thread at once.
     *
     * @param aepToSave  AEP to save; usually not null
     */
    private void cacheAEP(final AllExhibitProperties aepToSave)
        throws IOException
        {
        // Do nothing if not running in JWS/JNLP environment.
        if((bs == null) || (ps == null)) { return; }

        final URL aepURL = makeAEPURL(bs);

        // Start with a reasonable size buffer to write into...
        int initialBufSize = 65536;
        // Try to guess a new larger buffer size from any existing store size...
        try { initialBufSize = Math.max(initialBufSize, (int) (AEP_EXPANSION_BYTES + ps.get(aepURL).getLength())); }
        catch(final IOException e) { /* Ignore missing store error. */ }

        final ByteArrayOutputStream baos = new ByteArrayOutputStream(initialBufSize);
        final ObjectOutputStream ois = new ObjectOutputStream(new GZIPOutputStream(baos));
        ois.writeObject(aepToSave);
        ois.close();

        // OK, how big is the compressed serialised AEP?
        final int len = baos.size();
if(IsDebug.isDebug) { logger.log("AEP save size (bytes) is: " + len); }

        // Create entry for AEP if not yet extant.
        try { ps.create(aepURL, len + AEP_EXPANSION_BYTES); /* Enough for current AEP plus a margin for growth... */ }
        catch(final IOException e) { } // Ignore errors.

        // Now try to actually persist the data
        // by making sure that the entry has enough space granted
        // and then overwriting the entry.
        final FileContents fc = ps.get(aepURL);
        final long available = fc.getMaxLength();
        if(available < len)
            {
            // Currently not enough space available for current AEP...
            // so ask for more, in fact lots more than we currently need
            // to allow room for expansion without pestering the user too much.
            final long granted = fc.setMaxLength(AEP_EXPANSION_BYTES + 2 * len);
            if(len > granted)
                { throw new IOException("could not get enough space to cache AEP ("+granted+" vs "+len+")"); }
            }
        final OutputStream os = fc.getOutputStream(true);
        try { baos.writeTo(os); }
        finally { os.close(); }
        }

    /**Returns true if this (short) exhibit name is valid and unique.
     * The name has to be syntatically valid,
     * and the short name (and thus the whole name thaty contains it)
     * not currently present in:
     * </ul>
     * <li>The Gallery exhibit set (the main AEP).
     * <li>The uploading-files list.
     * <li>The selected-files list.
     * </ul>
     */
    boolean shortNameValidAndUnique(final String newShortName)
        {
        if(!ExhibitName.validNameFinalComponentSyntax(newShortName)) { return(false); }
        if(selectedFiles.shortNameInUse(newShortName)) { return(false); }
        if(uploadingFiles.shortNameInUse(newShortName)) { return(false); }
        if(aep.aeid.getFullName(newShortName) != null) { return(false); }

        // Seems OK.
        return(true);
        }

    /**If our name is not unique then increment the number-in-series (if non-zero) until we get a unique name.
     * This checks the (short) name against the AEP and other databases.
     */
    void uibIncNumberIfNotZeroWhileNotUnique(final UploadInfoBean uib)
        {
        while(uib.getNumber() != 0)
            {
            final String shortName = uib.getName(false);
            if(shortNameValidAndUnique(shortName))
                { break; /* Seems that we now have a unique name! */ }
            uib.incNumberIfNonZero();
            }
        }


    /**Propose/select/add selected files for upload.
     * Takes one or more files somehow selected by the user
     * (eg with some sort of GUI front end)
     * and a base exhibit name:
     * <ul>
     * <li>validates/guesses the files' types
     * <li>sorts the files into name order
     * <li>adds them to the table of names where the user can adjust details
     * </ul>
     * before asking for the files to be uploaded.
     * <p>
     * Duplicate entries may be silently discarded.
     * <p>
     * Package-visible to be callable from GUI classes.
     *
     * @param autoGuessFormat  if true then the routine may try to guess the
     *     type/format of files that do not match the type implied by the name
     * @param files  one (or more if the number-in-series is not zero) files
     *     that the user has selected for upload;
     *     never null not zero-size no containing nulls
     * @param uib  initialised with the first/desired name;
     *     never null and must be in a state where enoughValidUniqueInfo()==true
     *
     * @return null if all is well, else an error message (and no files added)
     */
    String addSelectedFiles(final boolean autoGuessFormat,
                            final FileContents files[],
                            final UploadInfoBean uib)
        {
        if((files == null) || (files.length < 1) ||
           (uib == null) || !uib.enoughValidUniqueInfo())
            { throw new IllegalArgumentException(); }

        final String baseExhibitName = uib.getFullName();
        final String extensionComponent = ExhibitName.getExtensionComponent(baseExhibitName);
        final ExhibitMIME.ExhibitTypeParameters etp = ExhibitMIME.getExhibitType(extensionComponent);

        try
            {
            // Sort files into order.
            // We use a case-insensitive ordering by name.
            Arrays.sort(files, new Comparator<FileContents>(){
                /**Compares its two arguments for order. */
                public final int compare(final FileContents fc1, final FileContents fc2)
                    {
                    try { return(fc1.getName().compareToIgnoreCase(fc2.getName())); }
                    catch(final IOException e)
                        {
                        final String message = "INTERNAL ERROR: failed to sort file(s) for upload: " + e.getMessage();
                        logger.log(message);
                        throw new Error(message, e);
                        }
                    }
                });

            // Validate all the files offered.
            // They must be readable and have a correct magic number
            // and file-extension if not auto-guessing...
            // Deal with them is file-name order.
            for(int i = 0; i < files.length; ++i)
                {
                final FileContents fc = files[i];

                // Skip this local file if already in selected-files list.
                final String name = fc.getName();
                if(selectedFiles.localFilenameInUse(name))
                    { continue; }

                // Is the file readable?
                if(!fc.canRead())
                    { return("Cannot read file: " + name); }

                // Is the file a sensible length?
                if(fc.getLength() < 1)
                    { return("File is zero-length: " + name); }

                // Has file got correct extension (ignoring case).
                final boolean correctExt = extensionComponent.equalsIgnoreCase(
                    FileTools.getExtension(name));

                // Has the file got the correct extension AND magic number?
                // It is an error not to if "auto"/guess mode is not selected.
                final int longestMagicBytes = ExhibitMIME.getLongestMagicBytes();
                final InputStream is = new BufferedInputStream(fc.getInputStream(),
                                                               longestMagicBytes);
                try
                    {
                    uibIncNumberIfNotZeroWhileNotUnique(uib); // Move to next unique exhibit name if need be.

                    is.mark(longestMagicBytes); // Allow rewind to start of file.
                    final boolean isFileTypeOK =
                        correctExt &&
                        ExhibitMIME.magicOK(etp, is);
                    if(isFileTypeOK)
                        {
                        // Magic is fine (and there *is* a magic number)
                        // and file extension is fine (ignoring case);
                        // thus file can be added with suggested name
                        // and the default/suggested extension.
                        uib.setSuffix(etp.dotSuffixForInputFile);
                        selectedFiles.add(new SelectedFileDetails(fc, uib.getFullName(), uib.getDescription(), null));
                        }
                    else if(!autoGuessFormat)
                        { return("File may be corrupt or wrong type (try 'auto' mode): " + name); }
                    else
                        {
                        // First try to guess by file extension...
                        ExhibitMIME.ExhibitTypeParameters etpGuessed =
                            ExhibitMIME.getInputFileType(name.toLowerCase());
                        is.reset(); // Rewind ready to check magic...
                        final boolean guessedOKFromExtension =
                            (etpGuessed != null) &&
                            ExhibitMIME.magicOK(etpGuessed, is);

                        // If the file extension does not seem to be right
                        // then try to guess just from the file magic number,
                        // ie from the file content rather than its name.
                        if(!guessedOKFromExtension)
                            {
                            is.reset(); // Rewind ready to check magic...
                            etpGuessed = ExhibitMIME.guessTypeFromMagic(is);
                            }

                        // Failed...
                        if(etpGuessed == null)
                            { return("File type is unrecognised: " + name); }

                        // Need to adjust the exhibit type for this file.
                        uib.setSuffix("." + etpGuessed.suffixForInputFile);
                        selectedFiles.add(new SelectedFileDetails(fc, uib.getFullName(), uib.getDescription(), null));
                        }
                    }
                finally { is.close(); }
                }
            }
        catch(final Exception e)
            {
            e.printStackTrace();
            logger.log("Failed to open file(s) for upload: " + e.getMessage());

            // Reset the name to the original in case of error.
            uib.setFullName(baseExhibitName);

            return("Cannot upload: " + e.getMessage());
            }

        return(null); // OK!
        }




    /**Save persistent properties (if possible).
     * If we have PersistenceService then
     * try to create/allocate space to save (ignoring errors if already there)
     * and save the data.
     */
    private void saveProperties()
        {
        if((ps != null) && (bs != null))
            {
            try
                {
                final URL propsURL = makePropsURL(bs);

                // Try to create a space for persisting the data
                // if not already present (ignoring error if it is).
                try { ps.create(propsURL, UploaderProps.MAX_PERS_BYTES); }
                catch(final IOException e) { }

                // Now try to save our properties data.
                final FileContents fc = ps.get(propsURL);
                final OutputStream os = fc.getOutputStream(true);
                try { os.write(props.getPersistentData()); }
                finally { os.close(); }

if(IsDebug.isDebug) { logger.log("Saved properties..."); }
                }
            catch(final IOException e)
                {
                e.printStackTrace();
                logger.log("Cannot save properties: " + e.getMessage());
                }
            }
        }


    /**Perform any activity required to shut down cleanly, eg save state.
     * This should try to avoid taking a long time.
     * <p>
     * Package-visible so as to be directly usable by the main GUI class.
     */
    void shutdown()
        {
        // Save any state...
        saveProperties();

        // Shut down the tunnel last.
        if(tunnel != null) { tunnel.destroy(); }
        }
    }
