/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.uploader;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.uploader.ExhibitHandlerBeanBase;
import org.hd.d.pg2k.svrCore.uploader.UploadInfoBean;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;

import javax.jnlp.BasicService;
import javax.jnlp.FileContents;
import javax.jnlp.SingleInstanceListener;
import javax.swing.*;
import javax.swing.Timer;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.AbstractTableModel;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;


/**Main (UI) class of JWS-based exhibit uploader.
 * Runs as a Swing App using JNLP resource-access facilities
 * such as persistence and file read/write.
 * <p>
 * This class/file contains as little non-UI code as is reasonably practical,
 * so that if we change the UI details then other classes should be unaffected.
 * <p>
 * Derived initially from Sun's "WebPad" sample JWS/JNLP code.
 * <p>
 * We don't really want/need this class to be Serializable,
 * but this class inherits Serializable from JFrame.
 */
public final class UploaderMain extends JFrame implements ActionListener
    {
    /**Central logger instance for uploader; never null.
     * This instance may log to the status bar and elsewhere.
     */
    private final SimpleLoggerIF logger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            {
            // Log to the Java console.
            System.out.println(message);

            // Schedule a job for the event-dispatching thread:
            // logging this message in the status bar.
            // We hope that these events do not get re-ordered too often.
            javax.swing.SwingUtilities.invokeLater(new Runnable(){
                public final void run()
                    { status.setText(message); }
                });
            }
        };

    /**If true, only allow upload in a JWS environment. */
    static final boolean JWS_UPLOAD_ONLY = false;

    /**Our companion "business-logic" class; never null. */
    private final UploaderLogic logic = new UploaderLogic(logger);

    /**The action performed by the "Select" button. */
    private final SelectAction selectAction = new SelectAction();
    /**The action performed by the "About" menu entry. */
    private final AboutAction aboutAction = new AboutAction();
    /**The action performed by the "Exit" menu entry. */
    private final ExitAction exitAction = new ExitAction();

//    /**Thread-safe List of all the actions; never null. */
//    private final java.util.List<JLFAbstractAction> actions = new Vector<JLFAbstractAction>();

    /**Select our locale; never null. */
    private final LocaleBeanBase lbb = new LocaleBeanBase();

    /**Status bar; never null. */
    private final JLabel status = createStatusBar();

    /**Single listener instance. */
    private final SISListener sisListener = new SISListener();

    /**Bean to hold our current view of the exhibit we are naming; never null. */
    final UploadInfoBean uib = new UploadInfoBean();

    /**If nothing else resets it, start with a more friendly number-in-series of 1 rather than 0. */
    { uib.setNumber(1); }

    /**Contains user-selected auto/manual file-type extraction; never null. */
    private final JCheckBox autoSuffixCheckBox = new JCheckBox();

    /**Handles Mouse over messages on toolbar buttons and menu items; never null. */
    private final MouseHandler mouseHandler = new MouseHandler(status);

    /**Background colour we set for read-only informational areas.
     * We use this to avoid confusion with data-entry areas.
     * <p>
     * We may co-ordinate this with the Web site uploader colours.
     */
    private static final Color BG_RO_COL = Color.YELLOW;


    /**Row marked for editing in uploadTable; -1 if none; never null.
     * Should be ignored if out of range for the current table content.
     * <p>
     * We use AtomicInteger for lockless thread-safety.
     */
    private final AtomicInteger uploadTableEditRow = new AtomicInteger(-1);

    ;


    /**The table model instance for the files selected for upload; never null.
     * This doesn't hold any data itself,
     * simply interfaces to the data in logic.selectedFiles.
     * <p>
     * Call fireTableDataChanged() or something more fine-grained
     * when the data has changed to allow proper redrawing.
     */
    private final AbstractTableModel uploadTableModel =
        new UploadFileTableModel(logic, logic.selectedFiles, uploadTableEditRow);

    /**The table model for the files queued for upload; never null.
     * This doesn't hold any data itself,
     * simply interfaces to the data in logic.uploadingFiles.
     * <p>
     * Call fireTableDataChanged() or something more fine-grained
     * when the data has changed to allow proper redrawing.
     */
    private final AbstractTableModel queueTableModel =
        new UploadFileTableModel(logic, logic.uploadingFiles, null);

    /**Title shown for application. */
    private static final String APPLICATION_WINDOW_TITLE = "Gallery Uploader";

    /**Idle time in text entry box before we take opportunity to show updates, ms; strictly positive. */
    private static final int TEXT_ENTRY_TIMEOUT_MS = 6001;


    /**Create an instance of the Uploader app main window.
     * Designed to be called by main().
     */
    private UploaderMain()
        {
        super(APPLICATION_WINDOW_TITLE);

        // Set up user actions.
        initActions();

        // ASAP, try to avoid multiple instances being started.
        if(logic.sis != null)
            { logic.sis.addSingleInstanceListener(sisListener); }

        // Create the contents of this frame...
        setJMenuBar(createMenu());
        final Tuple.Pair<JComponent, ActionListener> mainPane = createMainPane();
        pollAL = mainPane.second;
        assert(mainPane != null);
        assert(mainPane.first != null);
        assert(mainPane.second != null);
        getContentPane().add(mainPane.first, BorderLayout.CENTER);
        getContentPane().add(status, BorderLayout.SOUTH);

        // Arrange to exit the VM if the window is closed...
        addWindowListener(new WindowAdapter()
            {
            public final void windowClosing(final WindowEvent evt)
                {
                // Shut down gracefully (and exit()) if allowed to,
                // else attempt to veto with an exception.
                try { shutdown(); }
                catch(final UnsupportedOperationException e) { }
                }
            });
        }

    /**Perform any activity required to shut down cleanly, eg save state, then exit.
     * This should try to avoid taking a long time.
     * <p>
     * This may throw an UnsupportedOperationException to try to veto
     * an exit that the user changes their mind about,
     * eg because we have work in progress!
     *
     * @throws UnsupportedOperationException  if the user vetoes the shut-down
     */
    private void shutdown()
        throws UnsupportedOperationException
        {
        // If there is any work in progress
        // then double-check before allowing shutdown.
        if((logic.selectedFiles.size() > 0) ||
           (logic.uploadingFiles.size() > 0))
            {
            if(JOptionPane.OK_OPTION != JOptionPane.showConfirmDialog(this, "Sure you want to discard current selected files and uploads?", "Quit with work in progess...?", JOptionPane.OK_CANCEL_OPTION))
                {
                logger.log("Shutdown vetoed...");
                throw new UnsupportedOperationException("User vetoed shutdown...");
                }
            }

        // Do any shutdown required by the non-GUI components...
        logic.shutdown();

        // Now allow other instances to be started...
        if(logic.sis != null)
            { logic.sis.removeSingleInstanceListener(sisListener); }

        // Now allow a gracefull exit!
        System.exit(0);
        }

    /**This method should be called before creating the UI to create all the Actions. */
    private void initActions()
        {
//        actions.clear();
        registerAction(selectAction);
        registerAction(aboutAction);
        registerAction(exitAction);
        }

    private void registerAction(final JLFAbstractAction action)
        {
        action.addActionListener(this);
//        actions.addElement(action);
        }

    /**Create the application menu bar. */
    private JMenuBar createMenu()
        {
        final JMenuBar menuBar = new JMenuBar();

        // Build the File menu
        final JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic('F');
        fileMenu.add(exitAction).addMouseListener(mouseHandler);
//        fileMenu.setMnemonic('U');
//        fileMenu.add(selectAction).addMouseListener(mouseHandler);

        // Build the help menu
        final JMenu helpMenu = new JMenu("Help");
        helpMenu.setMnemonic('H');
        helpMenu.add(aboutAction).addMouseListener(mouseHandler);

        menuBar.add(fileMenu);
        menuBar.add(helpMenu);

        return menuBar;
        }

    /**Create the (main) tabbed pane component and a lister for UI polling. */
    private Tuple.Pair<JComponent, ActionListener> createMainPane()
        {
        final JTabbedPane tabbedPane = new JTabbedPane();
        final ImageIcon icon = null;

        int panel = 0;
        final Tuple.Pair<JComponent, ActionListener> statusComponent = createStatusPanel();
        tabbedPane.addTab("Status", icon, statusComponent.first,
                          "Connection and system status");
        statusComponent.first.setPreferredSize(new Dimension(600, 250));
        tabbedPane.setMnemonicAt(panel++, KeyEvent.VK_S);

        final Tuple.Pair<JComponent, ActionListener> chooserComponent = createChooserPanel();
        tabbedPane.addTab("Choose", icon, chooserComponent.first,
                          "Choose files/exhibits for upload");
        tabbedPane.setMnemonicAt(panel++, KeyEvent.VK_C);

        final Tuple.Pair<JComponent, ActionListener> progressComponent = createProgressPanel();
        tabbedPane.addTab("Progress", icon, progressComponent.first,
                          "Progress of uploads to server");
        tabbedPane.setMnemonicAt(panel++, KeyEvent.VK_P);

        return(new Tuple.Pair<JComponent,ActionListener>(tabbedPane, new ActionListener(){
            /**Call the internal listeners of each panel. */
            public void actionPerformed(final ActionEvent e)
                {
                statusComponent.second.actionPerformed(e);
                chooserComponent.second.actionPerformed(e);
                progressComponent.second.actionPerformed(e);
                }
            }));
        }

    /**Make the login/status component; never null.
     * Also creates a listener suitable to poll regularly
     * from the Swing thread to do async UI updates.
     */
    private Tuple.Pair<JComponent, ActionListener> createStatusPanel()
        {
        final JPanel pane = new JPanel(new GridBagLayout());
        final GridBagConstraints c = new GridBagConstraints();

        // Wrap insets round everything.
        c.insets = makeStandardInsets();

        // Heading top (left).
        c.gridy = 0;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all columns.
        final JLabel heading = new JLabel("<html><big><strong>STATUS</strong></big></html>");
        pane.add(heading, c);
        c.gridwidth = 1; // Back to default.

        // Vertical spacer...
        ++c.gridy;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        pane.add(new JLabel(" "), c);
        c.gridwidth = 1; // Back to default.

        // Left-align everything for neatness.
        c.anchor = GridBagConstraints.WEST;

        ++c.gridy;
        c.gridx = 0;
        final JLabel authInitialsLabel = new JLabel("Your author initials:");
        pane.add(authInitialsLabel, c);
        ++c.gridx;
        // Initialise auth initial field from persistent properties, if any.
        final JTextField authInitials = new JTextField(logic.props.getUserID(), ExhibitName.MAX_AUTH_INITIALS_LENGTH);
        authInitials.setMinimumSize(authInitials.getPreferredSize()); // Prevent squeezing!
        authInitials.setText(logic.props.getUserID()); // Set up with any extant value...
        pane.add(authInitials, c);
        authInitialsLabel.setLabelFor(authInitials);

        // Password below (left).
        c.gridx = 0;
        ++c.gridy;
        final JLabel passwordFieldLabel = new JLabel("Your password:");
        pane.add(passwordFieldLabel, c);
        ++c.gridx;
        final JPasswordField passwordField = new JPasswordField(CoreConsts.MAX_PASSWORD_LEN);
        passwordField.setMinimumSize(passwordField.getPreferredSize()); // Prevent squeezing!
        pane.add(passwordField, c);
        passwordFieldLabel.setLabelFor(passwordField);

        // Status (on-line or not) below (left).
        c.gridx = 0;
        ++c.gridy;
        final JLabel serverConnectionStatusLabel = new JLabel("Connection status:");
        pane.add(serverConnectionStatusLabel, c);
        ++c.gridx;
        final JLabel serverConnectionStatus = new JLabel("?");
        pane.add(serverConnectionStatus, c);
        serverConnectionStatusLabel.setLabelFor(serverConnectionStatus);

        // Hostname below (left).
        c.gridx = 0;
        ++c.gridy;
        final JLabel serverURLLabel = new JLabel("Server URL:");
        pane.add(serverURLLabel, c);
        ++c.gridx;
        final JLabel serverURL = new JLabel((logic.bs == null) ? "?" :
            logic.bs.getCodeBase().toString());
        pane.add(serverURL, c);
        serverURLLabel.setLabelFor(serverURL);

        // Time of last successful contact with serverURL.
        c.gridx = 0;
        ++c.gridy;
        final JLabel lastServerContactTimeLabel = new JLabel("Last server contact:");
        pane.add(lastServerContactTimeLabel, c);
        ++c.gridx;
        final JLabel lastServerContactTime = new JLabel("?");
        pane.add(lastServerContactTime, c);
        lastServerContactTimeLabel.setLabelFor(lastServerContactTime);

        // Cached exhibit count in Gallery.
        c.gridx = 0;
        ++c.gridy;
        final JLabel exhibitCountLabel = new JLabel("Gallery exhibit count / update:");
        pane.add(exhibitCountLabel, c);
        ++c.gridx;
        final JLabel exhibitCount = new JLabel("?");
        pane.add(exhibitCount, c);
        exhibitCountLabel.setLabelFor(exhibitCount);

//        // Count of exhibits in the upload area.
//        c.gridx = 0;
//        ++c.gridy;
//        pane.add(new JLabel("Uploaded exhibit count:"), c);
//        ++c.gridx;
//        pane.add(new JLabel("?"), c);

        // Memory usage and total available...
        c.gridx = 0;
        ++c.gridy;
        final JLabel memoryUsageLabel = new JLabel("Memory use/total:");
        pane.add(memoryUsageLabel, c);
        ++c.gridx;
        final JLabel memoryUsage = new JLabel("?");
        pane.add(memoryUsage, c);
        memoryUsageLabel.setLabelFor(memoryUsage);

        // Last dir visited for files (eg to upload).
        c.gridx = 0;
        ++c.gridy;
        final JLabel lastDirLabel = new JLabel("Last directory visited:");
        pane.add(lastDirLabel, c);
        ++c.gridx;
        final JLabel lastDir = new JLabel("?");
        pane.add(lastDir, c);
        lastDirLabel.setLabelFor(lastDir);

        // Application UI last active...
        c.gridx = 0;
        ++c.gridy;
        final JLabel appUILastActiveLabel = new JLabel("UI/app last active:");
        pane.add(appUILastActiveLabel, c);
        ++c.gridx;
        final JLabel appUILastActive = new JLabel("?");
        pane.add(appUILastActive, c);
        appUILastActiveLabel.setLabelFor(appUILastActive);

        // Number of items queued for upload...
        c.gridx = 0;
        ++c.gridy;
        final JLabel queuedCountLabel = new JLabel("Files queued for upload:");
        pane.add(queuedCountLabel, c);
        ++c.gridx;
        final JLabel queuedCount = new JLabel("?");
        pane.add(queuedCount, c);
        queuedCountLabel.setLabelFor(queuedCount);

        // Number of items already uploaded...
        c.gridx = 0;
        ++c.gridy;
        final JLabel uploadedCountLabel = new JLabel("Files uploaded this session:");
        pane.add(uploadedCountLabel, c);
        ++c.gridx;
        final JLabel uploadedCount = new JLabel("?");
        pane.add(uploadedCount, c);
        uploadedCountLabel.setLabelFor(uploadedCount);

        // Transfer speed of items already uploaded...
        c.gridx = 0;
        ++c.gridy;
        final JLabel uploadedSpeedLabel = new JLabel("File upload speed (Bps) this session:");
        pane.add(uploadedSpeedLabel, c);
        ++c.gridx;
        final JLabel uploadedSpeed = new JLabel("?");
        pane.add(uploadedSpeed, c);
        uploadedSpeedLabel.setLabelFor(uploadedSpeed);

        // Vertical spacer...
        c.gridx = 0;
        ++c.gridy;
        pane.add(new JLabel(" "), c);

        // Start of advanced status...
        c.gridx = 0;
        ++c.gridy;
        pane.add(new JLabel("Advanced..."), c);

        // JWS status...
        c.gridx = 0;
        ++c.gridy;
        pane.add(new JLabel("JWS Services Available: "), c);
        ++c.gridx;
        final StringBuilder sbSrv = new StringBuilder();
        if(logic.bs != null) { sbSrv.append("Basic "); }
        if(logic.ps != null) { sbSrv.append("Peristence "); }
        if(logic.fos != null) { sbSrv.append("FileOpen "); }
        if(logic.sis != null) { sbSrv.append("SingleInstance "); }
        if(logic.exs != null) { sbSrv.append("Extended "); }
        if(sbSrv.length() == 0) { sbSrv.append("-none-"); }
        pane.add(new JLabel(sbSrv.toString()), c);


        // Disable the ID/passwd fields until we are sure that we are online.
        authInitials.setEnabled(false);
        passwordField.setEnabled(false);

        // Create a listener to handle async UI-safe updates.
        final ActionListener al = new ActionListener(){
            public void actionPerformed(final ActionEvent evt)
                {
                try
                    {
                    final BasicService bs = logic.bs;
                    if(bs == null)
                        {
                        serverConnectionStatus.setText("(not in JWS)");
                        authInitials.setEnabled(false);
                        passwordField.setEnabled(false);
                        }
                    else
                        {
                        // Check if offline.
                        serverConnectionStatus.setText("checking...");
                        final boolean offline = bs.isOffline();
                        serverConnectionStatus.setText(offline ? "off-line" : "on-line");
                        if(offline)
                            {
                            authInitials.setEnabled(false);
                            passwordField.setEnabled(false);
                            }
                        else // We appear to be online.
                            {
                            // We can allow the user to enter/alter ID,
                            // and do running updates of some status fields.
                            authInitials.setEnabled(true);
                            passwordField.setEnabled(true);

                            final char[] pass = passwordField.getPassword();
                            final String userID = authInitials.getText();

                            // Save the current ID value in the properties...
                            logic.props.setUserID(userID);

                            if(!ExhibitName.validAuthorSyntax(userID))
                                {
                                logic.setAuthenticationInfo(null, null);
                                serverConnectionStatus.setText("To connect, please enter your author initials...");
                                }
                            else if((pass == null) || (pass.length == 0))
                                {
                                logic.setAuthenticationInfo(null, null);
                                serverConnectionStatus.setText("To connect, please enter your password ("+CoreConsts.MIN_PASSWORD_LEN+" to "+CoreConsts.MAX_PASSWORD_LEN+" chars)...");
                                }
                            else if(pass.length < CoreConsts.MIN_PASSWORD_LEN)
                                {
                                logic.setAuthenticationInfo(null, null);
                                serverConnectionStatus.setText("Password too short (min "+CoreConsts.MIN_PASSWORD_LEN+" chars)");
                                }
                            else if(pass.length > CoreConsts.MAX_PASSWORD_LEN)
                                {
                                logic.setAuthenticationInfo(null, null);
                                serverConnectionStatus.setText("Password too long (max "+CoreConsts.MAX_PASSWORD_LEN+" chars)");
                                }
                            else
                                {
                                // These may be usable authentication values.
                                logic.setAuthenticationInfo(userID, pass);
                                }
                            // Clear our password copy to zero.
                            if(pass != null)
                                { for(int i = pass.length; --i >= 0; ) { pass[i] = 0; } }

                            // Update the last-server-contact time displayed.
                            if(logic.tunnel != null)
                                {
                                final long lSCT = logic.tunnel.getLastSuccessfulConnectionTime();
                                lastServerContactTime.setText((lSCT == 0) ? "?" :
                                                              (new Date(lSCT)).toString());
                                if(logic.tunnel.isBroken())
                                    { serverConnectionStatus.setText("server connection is broken (bad password?)"); }
                                }
                            }
                        }
                    }
                catch(final Exception e)
                    {
                    e.printStackTrace();
                    serverConnectionStatus.setText("unexpected exception: " + e.getMessage() + ((e.getStackTrace().length > 0) ? (" @ " + e.getStackTrace()[0]) : ""));
                    }

                // Update displayed AEP data.
                final AllExhibitProperties aep = logic.getAep();
                exhibitCount.setText(((aep.aeid.length == 0) ? "?" :
                                        String.valueOf(aep.aeid.length)) + " / " +
                                     ((aep.hashNotChangedSince == 0) ? "?" :
                                          ((new Date(aep.hashNotChangedSince)).toString())));

                // Update memory stats.
                final Runtime r = Runtime.getRuntime();
                memoryUsage.setText(TextUtils.sizeAsText(r.totalMemory() - r.freeMemory(), true) + " / " +
                                    TextUtils.sizeAsText(r.totalMemory(), true));

                // Update UI activity display.
                final long lastActive = logic.getUserLastActive();
                if(lastActive == 0)
                    { appUILastActive.setText("never"); }
                else
                    {
                    final long t = (System.currentTimeMillis()-lastActive)/1000;
                    appUILastActive.setText(String.valueOf(t) + "s ago");
                    if(t < 10) { appUILastActive.setForeground(Color.RED); }
                    else if(t < 30) { appUILastActive.setForeground(Color.MAGENTA); }
                    else if(t < 90) { appUILastActive.setForeground(Color.BLUE); }
                    else { appUILastActive.setForeground(Color.BLACK); }
                    }

                // Update queued-files count.
                final int qc = logic.uploadingFiles.size();
                queuedCount.setText(String.valueOf(qc));
                queuedCount.setForeground((qc == 0) ? Color.BLACK : Color.RED);

                // Update uploaded-files count and speed.
                uploadedCount.setText(String.valueOf(logic.getCompletedUploadsThisSession()));
                uploadedSpeed.setText(String.valueOf(logic.getSmoothedUploadSpeedBps()));

                // Update last-directory-visited value.
                final File lDir = logic.props.getFileChooserPathHint();
                lastDir.setText((lDir == null) ? null : lDir.getPath());
                }
            };

        return(new Tuple.Pair<JComponent, ActionListener>(pane, al));
        }


    /**If true, try to enable file drag-and-drop into the table.
     * This may not be allowed, eg in the JNLP/JWS sandbox.
     */
    private static final boolean ENABLE_FILE_DROP = true;

    /**Make the chooser/naming component; never null.
     * Also creates a listener suitable to poll regularly
     * from the Swing thread to do async UI updates.
     */
    private Tuple.Pair<JComponent, ActionListener> createChooserPanel()
        {
        // Set it up with any initial AEP state...
        uib.setAep(logic.getAep());
        // Initially pretend that there is no "uploader area" set.
        uib.setUploadAeid(new AllExhibitImmutableData());

        // Fill in "popular"/likely values for some blank fields.
        uib.setCommonValuesForUnsetFields();

        // Create our new pane.
        final JPanel pane = new JPanel();
        pane.setLayout(new GridBagLayout());
        final GridBagConstraints c = new GridBagConstraints();

        // Wrap insets round everything.
        c.insets = makeStandardInsets();

        // Header...
        c.gridx = 0;
        c.gridy = 0;
        pane.add(new JLabel(""), c);
        ++c.gridx;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all columns.
        final JLabel heading = new JLabel("<html><big><strong>EXHIBIT CHOOSER/NAMER</strong></big></html>");
        pane.add(heading, c);
        c.gridwidth = 1; // Back to default.

        // Status (below)...
        // Generally we expect to set/update the status asynchronously.
        ++c.gridy;
        c.gridx = 0;
        final JLabel statusLabel = new JLabel("<html><strong>Status:</strong></html>");
        pane.add(statusLabel, c);
        statusLabel.setForeground(Color.RED);
        ++c.gridx;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        final JLabel statusText = new JLabel("?");
        statusText.setForeground(Color.RED);
        statusText.setBackground(BG_RO_COL);
        statusText.setOpaque(true);
        pane.add(statusText, c);
        c.gridwidth = 1; // Back to default.

        // Vertical spacer...
        ++c.gridy;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        pane.add(new JLabel(" "), c);
        c.gridwidth = 1; // Back to default.

        // FileChooser/dropzone.
        ++c.gridy;
        c.gridx = 0;
        final JLabel fileSelectButtonLabel = new JLabel("File(s):");
        pane.add(fileSelectButtonLabel, c);
        ++c.gridx;
        final String SELECT_FILE = "Select File...";
        final String SELECT_FILES = "Select File(s)...";
        final JButton fileSelectButton = new JButton(SELECT_FILE);
        fileSelectButton.setEnabled(false); // Initially false
        fileSelectButton.setActionCommand(selectAction.getActionCommand());
        fileSelectButton.addActionListener(this);
        fileSelectButton.setToolTipText("Press to choose files to name for upload.");
        fileSelectButtonLabel.setLabelFor(fileSelectButton);
        pane.add(fileSelectButton, c);
        ++c.gridx;
        final JTable uploadTable = new JTable(uploadTableModel);
        uploadTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION); // Select just one row at once.
//        uploadTable.setToolTipText("List of files you are naming for upload.");
        uploadTable.setBackground(BG_RO_COL);
        uploadTable.setOpaque(true);
        final JScrollPane uploadTableScroller = new JScrollPane(uploadTable);
        uploadTableScroller.setMinimumSize(new Dimension(2 * fileSelectButtonLabel.getWidth(), 5 * fileSelectButton.getHeight()));
        uploadTableScroller.setToolTipText("List of files that you are naming for upload.");
        c.weightx = 1.0; // Expand to take available space.
        c.weighty = 1.0; // Expand to take available space.
        c.fill = GridBagConstraints.BOTH; // Expand to take available space.
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        c.gridheight = 5; // Sit beside 5 items to left.
        pane.add(uploadTableScroller, c);
        c.fill = GridBagConstraints.NONE; // Back to default.
        c.weightx = 0; // Back to default.
        c.weighty = 0; // Back to default.
        c.gridwidth = 1; // Back to default.
        c.gridheight = 1; // Back to default.
        ++c.gridy;
        c.gridx = 1;
        final JButton fileClearAllButton = new JButton("Clear all");
        fileClearAllButton.setEnabled(false); // Initially false
        fileClearAllButton.addActionListener(new ActionListener()
            {
            /**Clear the "selected files" table if the button is pressed. */
            public void actionPerformed(final ActionEvent e)
                {
                logic.selectedFiles.clear();
                uploadTableEditRow.set(-1); // Nothing in the table being edited now.
                uploadTableModel.fireTableDataChanged();
                }
            });
        fileClearAllButton.setToolTipText("Clear all selected files from view (no files are deleted).");
        pane.add(fileClearAllButton, c);
        ++c.gridy;
        c.gridx = 1;
        final JButton fileClearButton = new JButton("Remove");
        fileClearButton.setEnabled(false); // Initially false
        fileClearButton.addActionListener(new ActionListener()
            {
            /**Clear/remove the current selected row if the button is pressed. */
            public void actionPerformed(final ActionEvent e)
                {
                final int selectedRow = uploadTable.getSelectedRow();
                if(selectedRow == -1) { return; /* Nothing to do... */ }
                uploadTableEditRow.set(-1); // Nothing in the table being edited now.
                logic.selectedFiles.remove(selectedRow);
                uploadTableModel.fireTableDataChanged();
                }
            });
        fileClearButton.setToolTipText("Clear the current selected file from view.");
        pane.add(fileClearButton, c);
        ++c.gridy;
        c.gridx = 1;
        final JButton fileUploadButton = new JButton("Start upload...");
        fileUploadButton.setEnabled(false); // Initially false
        fileUploadButton.setToolTipText("Press to start upload...");
        fileUploadButton.addActionListener(new ActionListener()
            {
            public final void actionPerformed(final ActionEvent e)
                {
                // If any selected file has an error status,
                // scroll to it and pop up an error,
                // else move all selected files to the upload queue!
                for(int i = 0; i < logic.selectedFiles.size(); ++i)
                    {
                    final SelectedFileDetails selectedFileDetails = logic.selectedFiles.get(i);
                    final String errMsg = selectedFileDetails.getStatus(logic);
                    if(errMsg != null)
                        {
                        uploadTable.changeSelection(i, SelectedFileDetails.COLNUM_EXHIBIT_NAME, false, false);
                        JOptionPane.showMessageDialog(pane, "Please fix: " + errMsg, "problem with selected file", JOptionPane.ERROR_MESSAGE);
                        return;
                        }
                    }

                // OK, move all the selected files across to the upload queue!
                logger.log("Queueing files for upload: " + logic.selectedFiles.size());
                while(logic.selectedFiles.size() > 0)
                    {
                    final int last = logic.selectedFiles.size() - 1;
                    final SelectedFileDetails selectedFileDetails = logic.selectedFiles.get(last);
                    logic.uploadingFiles.add(selectedFileDetails);
                    logic.selectedFiles.remove(last);
                    }

                // Redraw the now-empty table.
                uploadTableEditRow.set(-1); // Nothing in the table being edited now.
                uploadTableModel.fireTableDataChanged();
                }
            });
        pane.add(fileUploadButton, c);
        ++c.gridy;
        c.gridx = 1;
        final JPanel selectedFilesCountPanel = new JPanel();
        selectedFilesCountPanel.add(new JLabel("Selected files:"));
        final JLabel selectedFilesCount = new JLabel("");
        selectedFilesCountPanel.add(selectedFilesCount);
        pane.add(selectedFilesCountPanel, c);

        // Vertical spacer...
        ++c.gridy;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        pane.add(new JLabel(" "), c);
        c.gridwidth = 1; // Back to default.

        // Column headings (below)...
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel(" Field "), c);
        ++c.gridx;
        ++c.gridx;
        pane.add(new JLabel(" Value "), c);

        // Section/Category.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Category:"), c);
        ++c.gridx;
        ++c.gridx;
        // When there is an empty AEP, fill in a default.
        // This helps with debugging and gets the user started...
        final JComboBox categoryCombo = new JComboBox(uib.getAllCategories());
        categoryCombo.setSelectedItem(uib.getCategory());
        pane.add(categoryCombo, c);
        // Special rendering of category combobox elements...
        final class CatCellRenderer extends JLabel implements ListCellRenderer
            {
            private static final long serialVersionUID = -5061083427831496368L;
            CatCellRenderer() { setOpaque(true); }
            /**Render extension in decorated form using live lookup into DB. */
            public Component getListCellRendererComponent(final JList list,
                                                          final Object value,
                                                          final int index,
                                                          final boolean isSelected,
                                                          final boolean cellHasFocus)
                {
                final String val = (String) value;
                final StringBuilder sb = new StringBuilder(64);
                if(val != null)
                    {
                    sb.append(val);
                    sb.append(" [").append(GenUtils.computeSectionTitle(val, lbb)).append(']');
                    final Integer c = logic.getAep().getCategoryExhibitCounts().get(val);
                    if(c != null)
                        { sb.append(" (").append(c).append(')'); }
                    sb.append("      "); // FIXME: Hack to stop RHS being chopped off...
                    }
                setText(sb.toString());
                setBackground(isSelected ? Color.blue : Color.white);
                setForeground(isSelected ? Color.white : Color.black);

                // Note user activity...
                logic.setUserLastActive("category rendered");

                return(this);
                }
            }
        categoryCombo.setRenderer(new CatCellRenderer());
        categoryCombo.setToolTipText("Choose which section/category your exhibit belongs in.");
        pane.add(categoryCombo, c);
        ++c.gridx;
        final JLabel categoryDesc = new JLabel();
        categoryDesc.setToolTipText("Full category/section title.");
        categoryDesc.setOpaque(true);
        categoryDesc.setBackground(BG_RO_COL);
        pane.add(categoryDesc, c);
        // Update the uib and full title each time the user selects something.
        final ActionListener cBAL = new ActionListener()
            {
            public final void actionPerformed(ActionEvent e)
                {
                final Object catItem = categoryCombo.getSelectedItem();
                if(catItem instanceof String)
                    {
                    final String cat = (String) catItem;
                    uib.setCategory(cat); // Note new selection...
                    final String nCat = uib.getCategory();
                    if(!"".equals(nCat))
                        { categoryDesc.setText("<html>"+GenUtils.computeSectionTitle(nCat, lbb)+"</html>"); }
                    else
                        { categoryDesc.setText(null); }
                    }
                else
                    { categoryDesc.setText(null); }

                // Note user activity...
                logic.setUserLastActive("category action performed");
                }
            };
        // Listen for updates.
        categoryCombo.addActionListener(cBAL);
        // Show initial value, if any.
        final String nCat = uib.getCategory();
        categoryCombo.setSelectedItem(nCat);
        if(!"".equals(nCat))
            { categoryDesc.setText(GenUtils.computeSectionTitle(nCat, lbb)); }

        // Main words (stem).
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Main Words:"), c);
        ++c.gridx;
        ++c.gridx;
        final int textCols = 48;
        final JTextArea mainWords = new JTextArea(1+(ExhibitName.MAX_STEM_LENGTH/(textCols*4)), textCols);
        mainWords.setLineWrap(true);
        mainWords.setToolTipText("One or more words describing the exhibit, most important first.");
        c.weighty = 0.5; // Take up some slack vertically...
//        c.weightx = 0.5;
        c.fill = GridBagConstraints.BOTH;
        pane.add(mainWords, c);
        c.weighty = 0; // Back to normal.
        c.weightx = 0;
        c.fill = GridBagConstraints.NONE;
        ++c.gridx;
        final JEditorPane mainWordsResponse = new JEditorPane();
        final StringBuilder mainWordsResponseText = new StringBuilder(); // Current text displyed.
        mainWordsResponse.setEditable(false); // Read-only.
        mainWordsResponse.setContentType("text/html"); // HTML formatted.
        // Make this the same size as the box to its left by default.
        mainWordsResponse.setPreferredSize(mainWords.getPreferredSize());
        mainWordsResponse.setBackground(BG_RO_COL);
        final JScrollPane mWRSP = new JScrollPane(mainWordsResponse);
        mWRSP.setPreferredSize(mainWords.getPreferredSize());
        c.weighty = 0.5; // Take up some slack vertically...
        c.fill = GridBagConstraints.VERTICAL;
        pane.add(mWRSP, c);
        c.weighty = 0; // Back to normal.
        c.fill = GridBagConstraints.NONE;
        ++c.gridx;
        // Help appropriately size the upload item list.
        uploadTable.setPreferredScrollableViewportSize(new Dimension(2 * mainWords.getPreferredSize().width, Math.max(mainWords.getPreferredSize().height/2, 6 * fileSelectButtonLabel.getPreferredSize().height)));

        // Attributes.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Attributes:"), c);
        ++c.gridx;
        final Vector<String> attrWords = new Vector<String>();
        attrWords.add(ExhibitHandlerBeanBase.SETTER_ALL); // Indicates no attribute.
        attrWords.addAll(ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
        ++c.gridx;
        final JPanel attrPane = new JPanel();
        c.gridwidth = 2; // Span multiple columns...
        pane.add(attrPane, c);
        c.gridwidth = 1; // Back to normal...
        final int maxAttrs = 4;
        final JComboBox attrs[] = new JComboBox[maxAttrs];
        final List<String> initialAttrWords = uib.getAttributeWordsAsList();
        for(int i = 0; i < maxAttrs; ++i)
            {
            final boolean isFirst = (i == 0);
            // Special rendering of attribute combobox elements...
            // Though maybe only for first/last attr combo to save space,
            // or only for the selected cell?
            final class AttrCellRenderer extends JLabel implements ListCellRenderer
                {
                private static final long serialVersionUID = -4204688614542667708L;
                AttrCellRenderer() { setOpaque(true); }
                /**Render extension in decorated form using live lookup into DB. */
                public Component getListCellRendererComponent(final JList list,
                                                              final Object value,
                                                              final int index,
                                                              final boolean isSelected,
                                                              final boolean cellHasFocus)
                    {
                    final String val = (String) value;
                    if(val != null)
                        {
                        final StringBuilder sb = new StringBuilder(64);
                        sb.append(val);

                        final Map<String,Set<String>> attrCount = logic.getAep().getExhibitsByAttribute();

                        if(isFirst)
                            {
                            // Insert a localised description if possible.
                            final String i18nName = UploaderConsts.ATTR_I18N_DESC_PREFIX + val;
                            final String msg = lbb.getLocalisedMessage(i18nName);
                            if(!msg.equals(i18nName))
                                {
                                sb.append(" [").append(msg).append(']');
                                }

                            // Insert a count if possible.
                            final Set<String> names = (attrCount.get(val));
                            if((names != null) && (names.size() > 0))
                                {
                                final int count = names.size();
                                sb.append(" (").append(count).append(')');
                                }

                            sb.append("      "); // FIXME: Hack to stop RHS being chopped off...
                            setText(sb.toString());
                            }
                        else
                            {
                            setText(val); // Very simple rendering...
                            }
                        }
                    else { setText(null); } // No value, so no text.
                    setBackground(isSelected ? Color.blue : Color.white);
                    setForeground(isSelected ? Color.white : Color.black);

                    // Note user activity...
                    logic.setUserLastActive("attributes rendered");

                    return(this);
                    }
                }
            final JComboBox attrWordsCombo = new JComboBox(attrWords);
            attrWordsCombo.setToolTipText("Select an attribute for your exhibit.");
            attrWordsCombo.setRenderer(new AttrCellRenderer());
            // Capture any existing value.
            if(initialAttrWords.size() > i)
                { attrWordsCombo.setSelectedItem(initialAttrWords.get(i)); }
            attrs[i] = attrWordsCombo;
            attrPane.add(attrWordsCombo, c);
            }

        // Number in series.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Number in Series:"), c);
        ++c.gridx;
        ++c.gridx;
        final SpinnerModel numberInSeriesModel = new SpinnerNumberModel(
                                       uib.getNumber(), // initial value
                                       0,    // min
                                       9999, // max
                                       1);   // step
        final JSpinner nos = new JSpinner(numberInSeriesModel);
        pane.add(nos, c);
        nos.setEditor(new JSpinner.NumberEditor(nos, "#"));
        nos.setToolTipText("Number-in-series value for exhibit, or 0 if none.");

        // Type/extension.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Type/suffix:"), c);
        ++c.gridx;
        final JPanel ap = new JPanel();
        final JLabel apl = new JLabel("Auto:");
        ap.add(apl, c);
        ap.add(autoSuffixCheckBox, c);
        autoSuffixCheckBox.setSelected(true); // Default to "auto" mode.
        autoSuffixCheckBox.setToolTipText("In 'auto' mode, the uploader can try to guess the type of each exhibit that you select.");
        apl.setLabelFor(autoSuffixCheckBox);
        pane.add(ap, c);
        ++c.gridx;
        final Vector<String> extensions = new Vector<String>();
        for(final ExhibitMIME.ExhibitTypeParameters etp : ExhibitMIME.getAllValidExhibitTypes())
            { extensions.add(etp.dotSuffixForInputFile); }
        Collections.sort(extensions);
        final JComboBox fileType = new JComboBox(extensions);
        // Special rendering of combobox elements...
        final class TypeCellRenderer extends JLabel implements ListCellRenderer
            {
            private static final long serialVersionUID = 5537423282374537074L;
            TypeCellRenderer() { setOpaque(true); }
            /**Render extension in decorated form using live lookup into DB. */
            public Component getListCellRendererComponent(final JList list,
                                                          final Object value,
                                                          final int index,
                                                          final boolean isSelected,
                                                          final boolean cellHasFocus)
                {
                final String val = (String) value;
                final StringBuilder sb = new StringBuilder(64);
                if(val != null)
                    {
                    sb.append(val);
                    final ExhibitMIME.ExhibitTypeParameters etp = ExhibitMIME.getExhibitType(val.substring(1));
                    if(etp != null)
                        {
                        sb.append(" [" + etp.mimeType + ": " + etp.description + "]");
                        final Integer c = logic.getAep().getDottedExtensionExhibitCounts().get(val);
                        if(c != null)
                            { sb.append(" (").append(c).append(')'); }
                        }
                    sb.append("      "); // FIXME: Hack to stop RHS being chopped off...
                    }
                setText(sb.toString());
                setBackground(isSelected ? Color.blue : Color.white);
                setForeground(isSelected ? Color.white : Color.black);

                // Note user activity...
                logic.setUserLastActive("exhibit type rendered");

                return(this);
                }
            }
        fileType.setRenderer(new TypeCellRenderer());
        fileType.setSelectedItem(uib.getSuffix()); // Select existing value, if any.
        fileType.setToolTipText("Type/extension of your exhibit, eg .jpg for a JPEG image.");
        pane.add(fileType, c);
        ++c.gridx;
        final JLabel fileExt = new JLabel(uib.getSuffix());
        fileExt.setBackground(BG_RO_COL);
        fileExt.setOpaque(true);
        pane.add(fileExt, c);

        // Author.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Author:"), c);
        ++c.gridx;
        ++c.gridx;
        final JLabel auth = new JLabel(logic.props.getUserID());
        auth.setBackground(BG_RO_COL);
        auth.setOpaque(true);
        pane.add(auth, c);
        ++c.gridx;
        final JLabel authDesc = new JLabel();
        authDesc.setBackground(BG_RO_COL);
        authDesc.setOpaque(true);
        pane.add(authDesc, c);

        // Description.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("Description:"), c);
        ++c.gridx;
        ++c.gridx;
        final JTextArea description = new JTextArea(1+(Math.min(CoreConsts.DESCRIPTION_MAX_CHARS/4, 256)/textCols), textCols);
        description.setLineWrap(true);
        description.setWrapStyleWord(true);
        description.setToolTipText("Optional plain ASCII description of the exhibit, possibly with HTML markup, maximum " + CoreConsts.DESCRIPTION_MAX_CHARS + " characters.");
        c.weighty = 0.5; // Take up some slack vertically...
//        c.weightx = 0.5;
        c.fill = GridBagConstraints.BOTH;
        pane.add(description, c);
        c.weighty = 0; // Back to normal.
        c.weightx = 0;
        c.fill = GridBagConstraints.NONE;
        ++c.gridx;
        final JEditorPane descriptionResponse = new JEditorPane();
        final StringBuilder descriptionText = new StringBuilder(); // Current text displyed.
        descriptionResponse.setEditable(false); // Read-only.
        descriptionResponse.setContentType("text/html"); // HTML formatted.
        descriptionResponse.setBackground(BG_RO_COL);
        // Make this the same size as the box to its left by default.
        descriptionResponse.setPreferredSize(description.getPreferredSize());
        final JScrollPane dSP = new JScrollPane(descriptionResponse);
        dSP.setPreferredSize(description.getPreferredSize());
        c.weighty = 0.5; // Take up some slack vertically...
        c.fill = GridBagConstraints.VERTICAL;
        pane.add(dSP, c);
        c.weighty = 0; // Back to normal.
        c.fill = GridBagConstraints.NONE;


        // Ask to be notified of selection changes in the table...
        final ListSelectionModel rowSM = uploadTable.getSelectionModel();
        rowSM.addListSelectionListener(new ListSelectionListener()
            {
            public final void valueChanged(final ListSelectionEvent e)
                {
                final ListSelectionModel lsm =
                    (ListSelectionModel) e.getSource();
                if(!lsm.isSelectionEmpty())
                    {
                    final int selectedRow = lsm.getMinSelectionIndex();
                    // Make this edit row available to the rest of the UI.
                    uploadTableEditRow.set(selectedRow);
                    // Make sure that this row is visible.
                    uploadTable.changeSelection(selectedRow, SelectedFileDetails.COLNUM_EXHIBIT_NAME, false, false);

                    // Attempt to set editor fields from selected entry...
                    String name = null;
                    String prevName = null, nextName = null;
                    String desc = null;
                    synchronized(logic.selectedFiles) // Avoid races...
                        {
                        if((selectedRow >= 0) && (selectedRow < logic.selectedFiles.size()))
                            {
                            name = logic.selectedFiles.get(selectedRow).getEsa().filePath;
                            desc = logic.selectedFiles.get(selectedRow).getDescription();
                            }
                        // Get the names on either side.
                        if((selectedRow >= +1) && (selectedRow < logic.selectedFiles.size()+1))
                            { prevName = logic.selectedFiles.get(selectedRow-1).getEsa().filePath; }
                        if((selectedRow >= -1) && (selectedRow < logic.selectedFiles.size()-1))
                            { nextName = logic.selectedFiles.get(selectedRow+1).getEsa().filePath; }
                        }
                    // Don't set the name to one of the values on either side.
                    // This should avoid the most obvious finger trouble.
                    if((name != null) &&
                       !name.equals(prevName) && !name.equals(nextName) &&
                       (desc != null))
                        {
                        // Set the name as best we can.
                        uib.setFullName(name);
                        uib.setDescription(desc);

                        // Set UI editor fields individually...
                        categoryCombo.setSelectedItem(ExhibitName.getCategoryComponent(name));
                        final SortedSet<String> legalAttrs = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();
                        mainWords.setText(ExhibitName.getMainWordsComponent(name, legalAttrs));
                        final Enumeration attrEn = ExhibitName.getAttributeWordsComponentEnumeration(name, legalAttrs);
                        for(int i = 0; i < maxAttrs; ++i)
                            {
                            if((attrEn != null) && attrEn.hasMoreElements())
                                { attrs[i].setSelectedItem(attrEn.nextElement()); }
                            else
                                { attrs[i].setSelectedItem(ExhibitHandlerBeanBase.SETTER_ALL); }
                            }
                        numberInSeriesModel.setValue(Integer.valueOf(ExhibitName.getNumberInSeriesComponent(name)));
                        auth.setText(ExhibitName.getAuthorComponent(name));
                        fileType.setSelectedItem("." + ExhibitName.getExtensionComponent(name));
                        description.setText(desc);
                        }
//logger.log("SELECTION INDEX/NAME: " + selectedRow + '/' + name);
                    }
                }
            });


        // Attach file-drop handler to uploadTable / button / pane if possible.
        if(ENABLE_FILE_DROP)
            {
            final TransferHandler uTTH =
                new FileTransferHandler(autoSuffixCheckBox,
                                        uploadTableModel,
                                        logic,
                                        uib);
//            uploadTable.setDragEnabled(false); // Drop target only...
            uploadTable.setTransferHandler(uTTH);
            pane.setTransferHandler(uTTH);
            pane.setToolTipText("Drop your new exhibit files on this panel...");
            fileSelectButton.setTransferHandler(uTTH);
            fileSelectButton.setToolTipText("Press to choose files to name for upload, or drag the files here.");
            }


        // Create a listener to handle async UI-safe updates.
        final ActionListener al = new ActionListener(){
            public final void actionPerformed(final ActionEvent evt)
                {
                // If the AEP has changed
                // then reset the category combo box
                // and fill in any blank fields we can guess.
                final AllExhibitProperties aep = logic.getAep();
                if(aep.longHash != uib.getAep().longHash)
                    {
                    uib.setAep(aep);
                    uib.setCommonValuesForUnsetFields();

                    categoryCombo.setModel(new DefaultComboBoxModel(uib.getAllCategories()));
                    categoryCombo.setSelectedItem(uib.getCategory());
                    fileType.setSelectedItem(uib.getSuffix());
                    }

                // Allow table clear iff something in the table to be cleared!
                final int numberOfSelectedFiles = logic.selectedFiles.size();
                fileClearAllButton.setEnabled(numberOfSelectedFiles != 0);

                // Allow row clear iff something selected in the table!
                fileClearButton.setEnabled(uploadTable.getSelectedRow() != -1);

                // Allow "upload" if there is something available.
                // However, a dialogue may get popped up
                // if an error is encountered elsewhere.
                fileUploadButton.setEnabled(numberOfSelectedFiles != 0);

                // Show correct file count...
                selectedFilesCount.setText(String.valueOf(numberOfSelectedFiles));

                // Renormalise main-words component and show any warnings.
                // To avoid some ugly things,
                // such as partial words being taken as attributes,
                // and to save a bit of CPU and time,
                // only do an update when the user enters a separator
                // or the text area looses focus
                // or the user has been paused a long time.
                final String currentMainWords = mainWords.getText();
                final int mainWordsCaretPosition = mainWords.getCaretPosition();
                final boolean mWHF = mainWords.hasFocus();
                final boolean trailingSep =
                    currentMainWords.endsWith(" ") ||
                    currentMainWords.endsWith(ExhibitName.WORD_SEPS);
                if(/* trailingSep || */ !mWHF ||
                   (System.currentTimeMillis() - logic.getUserLastActive() > TEXT_ENTRY_TIMEOUT_MS)) // User paused a long time!
                    {
                    if(!currentMainWords.equals(uib.getMainWords()))
                        { uib.setMainWords(currentMainWords); }
                    final String normMainWords = uib.getMainWords();
                    // If completed text is definitely not normalised
                    // then replace it with a normalised version.
                    // Make sure that we do not zap trailing space/separator
                    // while user is typing in a name.
                    final String currentWordsTrimmed = currentMainWords.trim();
                    if(!currentWordsTrimmed.equals(normMainWords))
                        {
                        mainWords.setText(trailingSep ?
                            (normMainWords + ExhibitName.WORD_SEP) : normMainWords);
                        // Try to restore caret position...
                        // Ignore complaints about invalid position...
                        try { mainWords.setCaretPosition(mainWordsCaretPosition); }
                        catch(final IllegalArgumentException e) { }

                        // Note user activity...
                        logic.setUserLastActive("main words updated/normalised");
                        }
                    // Generate warning message for new/bad words,
                    // info about information available, location, etc.
                    final StringBuilder mainWordsMsg = new StringBuilder();
                    final String newWords = uib.getNewMainWords();
                    if(newWords.length() != 0)
                        {
                        // Do warning about unrecognised words...
                        final String localisedMessage = lbb.getLocalisedMessage("uploader.newWordsWarning", newWords);
                        mainWordsMsg.append(localisedMessage);
                        }
                    // Add any available section and tree-structured description.
                    // We can do this only if we have the category and some words.
                    // Do this mostly as would be shown in a catalogue page.
                    final String cat = uib.getCategory();
                    if(!"".equals(cat))
                        {
                        // We'll see if there is descriptive text for this category.
                        final String i18nName = UploaderConsts.SECTION_DESC_PROPNAME_PREFIX + cat;
                        final String catDesc = lbb.getLocalisedMessage(i18nName);
                        // We show the localised version if it exists and isn't just in a different case.
                        if(!catDesc.equals(i18nName))
                            {
                            mainWordsMsg.append("<p><strong>");
                            mainWordsMsg.append(lbb.getLocalisedMessage("common.cat.sectionDescription"));
                            mainWordsMsg.append("</strong></p>");
                            mainWordsMsg.append("<p>");
                            mainWordsMsg.append(catDesc);
                            mainWordsMsg.append("</p><hr>");
                            }

                        if(!"".equals(normMainWords))
                            {
                            // Synthesise a fake exhibit name for tree-data lookup.
                            final String synthExhibitName = cat + '/' + normMainWords +
                                                               "-A.a";
                            final String treeDesc = GenUtils.getLocalisedTreeDesc(aep,
                                                                                  synthExhibitName,
                                                                                  lbb, true, true, true, false);
                            if(!"".equals(treeDesc))
                                {
                                mainWordsMsg.append("<p><ul>");
                                mainWordsMsg.append(treeDesc);
                                mainWordsMsg.append("</ul></p>");
                                }
                            // Show any new-words warning at the end (again)
                            // so that it is hard to miss.
                            if(newWords.length() != 0)
                                {
                                // Do warning about unrecognised words...
                                final String localisedMessage = lbb.getLocalisedMessage("uploader.newWordsWarning", newWords);
                                mainWordsMsg.append(localisedMessage);
                                }
                            }
                        }
                    // Now display the main-words message (if any)...
                    final String mWPText = mainWordsMsg.toString();
                    if((mainWordsResponseText.length() != mWPText.length()) ||
                       !mWPText.equals(mainWordsResponseText.toString()))
                        {
                        mainWordsResponseText.setLength(0);
                        mainWordsResponseText.append(mWPText);
                        mainWordsResponse.setDocument(mainWordsResponse.getEditorKit().createDefaultDocument());
                        mainWordsResponse.setText(mWPText);

                        // Note user activity...
                        logic.setUserLastActive("main words updated");
                        }
                    }

                // Keep the attributes in sync.
                final List<String> attrWordsInUI = new ArrayList<String>(maxAttrs);
                for(int i = 0; i < maxAttrs; ++i)
                    {
                    final Object selectedItem = attrs[i].getSelectedItem();
                    if((selectedItem instanceof String) &&
                       !selectedItem.equals(ExhibitHandlerBeanBase.SETTER_ALL))
                        { attrWordsInUI.add((String) selectedItem); }
                    }
                final List<String> a1 = uib.getAttributeWordsAsList();
                if(!attrWordsInUI.equals(a1))
                    {
                    // Copy new attrs into uib...
                    uib.setAttributeWords(attrWordsInUI);

                    // Note user activity...
                    logic.setUserLastActive("attributes updated");
                    }
                // Copy normalised (uib) value back to UI.
                final List<String> attributeWordsAsList = uib.getAttributeWordsAsList();
                for(int i = 0; i < maxAttrs; ++i)
                    {
                    // Clear any trailing values completely...
                    if(i >= attributeWordsAsList.size())
                        { attrs[i].setSelectedItem(ExhibitHandlerBeanBase.SETTER_ALL); }
                    else
                        { attrs[i].setSelectedItem(attributeWordsAsList.get(i)); }
                    }

                // Keep the number-in-series field in sync.
                final int nosGUIValue = ((Number) nos.getValue()).intValue();
                if(nosGUIValue != uib.getNumber())
                    {
                    // Copy UI value into uib.
                    uib.setNumber(nosGUIValue);

                    // Note user activity...
                    logic.setUserLastActive("number-in-series updated");
                    }
                // Copy normalised (uib) value back to UI.
                nos.setValue(Integer.valueOf(uib.getNumber()));
                // Indicate selection of "file" or "files" as appropriate...
                fileSelectButton.setText((uib.getNumber() == 0) ? SELECT_FILE : SELECT_FILES);

                // Keep the author field in sync.
                uib.setAuthor(logic.props.getUserID());
                final String authorNorm = uib.getAuthor();
                auth.setText(authorNorm);
                final GenProps.AuthData authData = logic.getGenProps().getAuthData(authorNorm);
                authDesc.setText((authData == null) ? null : "<html>"+authData.name+"</html>");

                // Keep the extension/type field in sync.
                uib.setSuffix((String) fileType.getSelectedItem());
                final String suffixNorm = uib.getSuffix();
                fileType.setSelectedItem(suffixNorm);
                fileExt.setText(suffixNorm);

                // Render any description text as HTML.
                final String desc = description.getText();
                final int descCaretPosition = description.getCaretPosition();
                final boolean trailingS = desc.endsWith(" ");
                uib.setDescription(desc);
                final String descNorm = uib.getDescription();
                if(!descNorm.equals(desc))
                    {
                    // Normalise input text...
                    // (IFF user is not still typing!)
                    if(!description.hasFocus() ||
                       (System.currentTimeMillis() - logic.getUserLastActive() > TEXT_ENTRY_TIMEOUT_MS))
                        {
                        description.setText(trailingS ?
                            (descNorm + ' ') : descNorm);
                        // Try to put the cursor back where it was...
                        try { description.setCaretPosition(descCaretPosition); }
                        catch(final IllegalArgumentException e) { } // Ignore out-of-range complaints.

                        // Note user activity...
                        logic.setUserLastActive("description updated/normalised");
                        }
                    }
                if((descriptionText.length() != descNorm.length()) ||
                   !descNorm.equals(descriptionText.toString()))
                    {
                    descriptionResponse.setDocument(descriptionResponse.getEditorKit().createDefaultDocument());
                    descriptionResponse.setText(descNorm);

                    descriptionText.setLength(0);
                    descriptionText.append(descNorm);
                    descriptionResponse.setDocument(descriptionResponse.getEditorKit().createDefaultDocument());
                    descriptionResponse.setText(descNorm);

                    // Note user activity...
                    logic.setUserLastActive("description updated");
                    }


                // Can we allow selection of new files to upload?
                if(JWS_UPLOAD_ONLY && (logic.fos == null))
                    {
                    statusLabel.setForeground(Color.RED);
                    statusText.setForeground(Color.RED);
                    fileSelectButton.setEnabled(false); // Disable upload.
                    statusText.setText("No JWS FileOpenService available; cannot upload, sorry.");
                    }
                // Check the status of the uib;
                // have we enough info to allow an upload to proceed
                // and is the name unique in all our databases?
                else if(uib.enoughValidUniqueInfo() &&
                        !logic.uploadingFiles.shortNameInUse(uib.getName(false)))
                    {
                    // All OK!
                    // Potentially enable the upload stage.
                    statusLabel.setForeground(Color.GREEN);
                    statusText.setForeground(Color.GREEN);
                    statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.nameOK", uib.getFullName()) + "</html>");
                    fileSelectButton.setEnabled(true); // Enable upload.
                    }
                else
                    {
                    // Not enough info yet, or a conflict.
                    // Show the error and disable upload stage.
                    statusLabel.setForeground(Color.RED);
                    statusText.setForeground(Color.RED);
                    fileSelectButton.setEnabled(false); // Disable file selection.

                    // Try to provide specific error info/clues.
                    if(!uib.enoughInfo())
                        {
                        statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.notEnoughInfo") + "</html>");
                        }
                    else if(!uib.canMakeValidName())
                        {
                        statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.notValidName") + "</html>");
                        }
                    else if(!uib.nameUniqueInMainDatabase())
                        {
                        statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.nameExistsLive", uib.getFullName()) + "</html>");
                        }
                    else if(!uib.nameUniqueInAuxDatabase() || logic.uploadingFiles.shortNameInUse(uib.getName(false)))
                        {
                        statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.nameExistsUpload", uib.getFullName()) + "</html>");
                        }
                    else
                        {
                        // Generic error message...
                        statusText.setText("<html>" + lbb.getLocalisedMessage("uploader.nameCannotBeUsed", uib.getFullName()) + "</html>");
                        }
                    }


                // If an entry in the table is marked for editing,
                // then copy any updated name back into it
                // provided that the putative new name is valid and unique.
                // Also deal with any updated description.
                final int editRow = uploadTableEditRow.get();
                final String fullName = uib.getFullName();
                synchronized(logic.selectedFiles) // Avoid races...
                    {
                    if(ExhibitName.validNameSyntax(fullName) &&
                       (editRow >= 0) && (editRow < logic.selectedFiles.size()))
                        {
                        final SelectedFileDetails selectedFileDetails = logic.selectedFiles.get(editRow);

                        final String shortName = ExhibitName.getFileComponent(fullName);

                        // If this short name is valid and unique
                        // then it also implies that it has changed...
                        if(logic.shortNameValidAndUnique(shortName))
                            {
                            try
                                {
                                // Something changed; do an update...
                                logic.selectedFiles.setExhibitName(editRow, fullName);
                                // ...make sure the changed item is visible...
                                uploadTable.changeSelection(editRow, SelectedFileDetails.COLNUM_EXHIBIT_NAME, false, false);
                                // ...and get the display redrawn.
                                uploadTableModel.fireTableDataChanged();
                                }
                            catch(final IllegalArgumentException e)
                                {
                                // Silently ignore attempts to set duplicate names...
                                }
                            }

                        // Keep the description in sync too...
                        if(!descNorm.equals(selectedFileDetails.getDescription()))
                            {
                            // Something changed; do an update...
                            logic.selectedFiles.setDescription(editRow, descNorm);
                            // ...make sure the changed item is visible...
                            uploadTable.changeSelection(editRow, SelectedFileDetails.COLNUM_DESCRIPTION, false, false);
                            // ...and get the display redrawn.
                            uploadTableModel.fireTableDataChanged();
                            }
                        }
                    }
                }
            };

        return(new Tuple.Pair<JComponent, ActionListener>(
            new JScrollPane(pane), al));
        }


    /**Make the progress component; never null.
     * Also creates a listener suitable to poll regularly
     * from the Swing thread to do async UI updates.
     */
    private Tuple.Pair<JComponent, ActionListener> createProgressPanel()
        {
        // Create our new pane.
        final JPanel pane = new JPanel();
        pane.setLayout(new GridBagLayout());
        final GridBagConstraints c = new GridBagConstraints();

        // Wrap insets round everything.
        c.insets = makeStandardInsets();

        // Header...
        c.gridx = 0;
        c.gridy = 0;
        pane.add(new JLabel(""), c);
        ++c.gridx;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all columns.
        final JLabel heading = new JLabel("<html><big><strong>UPLOAD PROGRESS</strong></big></html>");
        pane.add(heading, c);
        c.gridwidth = 1; // Back to default.

        // Vertical spacer...
        ++c.gridy;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        pane.add(new JLabel(" "), c);
        c.gridwidth = 1; // Back to default.

        // Status line...
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel("So far:"), c);
        ++c.gridx;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        final JPanel statusLine = new JPanel();
        statusLine.setOpaque(true);
        statusLine.setBackground(Color.YELLOW);
        statusLine.setMinimumSize(heading.getPreferredSize());
        final JLabel uploadedLabel = new JLabel("Uploaded:");
        final JLabel uploadedCount = new JLabel("0");
        uploadedLabel.setLabelFor(uploadedCount);
        statusLine.add(uploadedLabel);
        statusLine.add(uploadedCount);
        final JLabel uploadSpeedLabel = new JLabel("  Upload speed:");
        final JLabel uploadSpeedValue = new JLabel("?");
        uploadSpeedLabel.setLabelFor(uploadSpeedValue);
        statusLine.add(uploadSpeedLabel);
        statusLine.add(uploadSpeedValue);
        final JLabel uploadCompleteLabel = new JLabel("  Time to finish:");
        final String TEXT_NA = "N/A";
        final JLabel uploadCompleteValue = new JLabel(TEXT_NA);
        uploadCompleteLabel.setLabelFor(uploadCompleteValue);
        statusLine.add(uploadCompleteLabel);
        statusLine.add(uploadCompleteValue);
        pane.add(statusLine, c);
        c.gridwidth = 1; // Back to default.

        // "Uploading now" slots...
        ++c.gridy;
        c.gridx = 0;
//        pane.add(new JLabel("Uploading now:"), c);
        ++c.gridx;
        pane.add(new JLabel("Name"), c);
        ++c.gridx;
        pane.add(new JLabel("Size"), c);
        ++c.gridx;
        pane.add(new JLabel("Progress"), c);
        final class UploadingNowItem
            {
            final JTextField nameBox = new JTextField("", 20);
            final JTextField sizeBox = new JTextField("", 10);
            final JProgressBar progressBar = new JProgressBar();
            }
        // Implicitly each index in uploadingNow[]
        // is associated with the same index in logic.inProgressSlots[].
        final UploadingNowItem uploadingNow[] = new UploadingNowItem[UploaderLogic.MAX_CONC_UPLOADS];
        for(int i = 0; i < UploaderLogic.MAX_CONC_UPLOADS; ++i)
            {
            ++c.gridy;
            c.gridx = 0;
            pane.add(new JLabel("Uploading now:"), c);
            ++c.gridx;
            final UploadingNowItem uni = new UploadingNowItem();
            uploadingNow[i] = uni;
            uni.nameBox.setEditable(false); // Display only.
            uni.nameBox.setMinimumSize(heading.getPreferredSize());
            pane.add(uni.nameBox, c);
            ++c.gridx;
            uni.sizeBox.setEditable(false); // Display only.
            uni.sizeBox.setMinimumSize(uni.sizeBox.getPreferredSize());
            pane.add(uni.sizeBox, c);
            ++c.gridx;
            pane.add(uni.progressBar, c);
            uni.progressBar.setStringPainted(true); // Show a string.
            uni.progressBar.setString(""); // Hide % string.
            uni.progressBar.setMinimumSize(heading.getPreferredSize());
            }

        // Vertical spacer...
        ++c.gridy;
        c.gridx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        pane.add(new JLabel(" "), c);
        c.gridwidth = 1; // Back to default.

        // Upload queue
        ++c.gridy;
        c.gridx = 0;
        final JLabel queueLabel = new JLabel("Upload queue:");
        pane.add(queueLabel, c);
        ++c.gridx;
        final JTable queueTable = new JTable(queueTableModel);
        queueTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION); // Select just one row at once.
//        queueTable.setToolTipText("List of files queued for upload.");
        queueTable.setBackground(BG_RO_COL);
        queueTable.setOpaque(true);
        final JScrollPane queueTableScroller = new JScrollPane(queueTable);
        queueTableScroller.setMinimumSize(new Dimension(2 * heading.getWidth(), 10 * heading.getHeight()));
        queueTableScroller.setToolTipText("List of files queued for upload.");
        c.weightx = 1.0; // Expand to take available space.
        c.weighty = 1.0; // Expand to take available space.
        c.fill = GridBagConstraints.BOTH; // Expand to take available space.
        c.gridwidth = GridBagConstraints.REMAINDER; // Span all remaining columns.
        c.gridheight = 5; // Sit beside various items to left.
        pane.add(queueTableScroller, c);
        c.fill = GridBagConstraints.NONE; // Back to default.
        c.weightx = 0; // Back to default.
        c.weighty = 0; // Back to default.
        c.gridwidth = 1; // Back to default.
        c.gridheight = 1; // Back to default.
        ++c.gridy;
        c.gridx = 0;
        pane.add(new JLabel(" "), c); // Spacer...
        ++c.gridy;
        c.gridx = 0;
        final JButton fileClearButton = new JButton("Discard");
        fileClearButton.setEnabled(false); // Initially false.
        fileClearButton.addActionListener(new ActionListener()
            {
            /**Clear/remove the current selected row if the button is pressed. */
            public void actionPerformed(final ActionEvent e)
                {
                final int selectedRow = queueTable.getSelectedRow();
                if(selectedRow == -1) { return; /* Nothing to do... */ }
                final SelectedFileDetails selectedFileDetails = logic.uploadingFiles.get(selectedRow);
                selectedFileDetails.setStatus(null); // Clear any transient error.
                // If there is a permanent error
                // then remove without asking further
                // else ask for confirmation.
                if(selectedFileDetails.getStatus(logic) != null)
                    { logic.uploadingFiles.remove(selectedRow); }
                else
                    {
                    // Double-check before zapping possibly-OK entry...
                    if(JOptionPane.OK_OPTION == JOptionPane.showConfirmDialog(pane, "Sure you want to stop trying to upload " + selectedFileDetails.getLocalFilename() + "?", fileClearButton.getText(), JOptionPane.OK_CANCEL_OPTION))
                        { logic.uploadingFiles.remove(selectedRow); }
                    }
                queueTableModel.fireTableDataChanged();
                }
            });
        fileClearButton.setToolTipText("Give up trying to upload this file.");
        pane.add(fileClearButton, c);
        ++c.gridy;
        c.gridx = 0;
        final JButton fileRetryButton = new JButton("Retry");
        fileRetryButton.setEnabled(false); // Initially false.
        fileRetryButton.addActionListener(new ActionListener()
            {
            /**Clear errors from the current selected row if the button is pressed. */
            public void actionPerformed(final ActionEvent e)
                {
                final int selectedRow = queueTable.getSelectedRow();
                if(selectedRow == -1) { return; /* Nothing to do... */ }
                logic.uploadingFiles.get(selectedRow).setStatus(null); // Clear any transient error.
                queueTableModel.fireTableDataChanged();
                }
            });
        fileRetryButton.setToolTipText("Try to clear any error on this file to retry upload.");
        pane.add(fileRetryButton, c);
        ++c.gridy;
        c.gridx = 0;
        final JButton fileRetryAllButton = new JButton("Retry All");
        fileRetryAllButton.setEnabled(false); // Initially false.
        fileRetryAllButton.addActionListener(new ActionListener()
            {
            /**Clear all errors if the button is pressed. */
            public void actionPerformed(final ActionEvent e)
                {
                synchronized(logic.uploadingFiles)
                    {
                    for(int i = logic.uploadingFiles.size(); --i >= 0; )
                        { logic.uploadingFiles.get(i).setStatus(null); } // Clear any transient errors.
                    }
                queueTableModel.fireTableDataChanged();
                }
            });
        fileRetryAllButton.setToolTipText("Try to clear errors on all files to retry upload.");
        pane.add(fileRetryAllButton, c);





        // The listener that will be polled regularly.
        final ActionListener al = new ActionListener()
            {
            public void actionPerformed(final ActionEvent e)
                {
                // Redraw the table if necessary.
                if(logic.uploadingFiles.testAndClearChangedFlag())
                    {
                    // Get table redrawn.
                    queueTableModel.fireTableDataChanged();

                    // Update metrics/status.
                    uploadedCount.setText(String.valueOf(logic.getCompletedUploadsThisSession()));
                    final long smoothedUploadSpeedBps = logic.getSmoothedUploadSpeedBps();
                    uploadSpeedValue.setText((smoothedUploadSpeedBps <= 0) ? "?" : (TextUtils.sizeAsText(smoothedUploadSpeedBps, true) + "/s"));

                    if(logic.uploadingFiles.size() == 0)
                        {
                        uploadCompleteValue.setText(TEXT_NA);
                        }
                    else
                        {
                        // Conversative estimate of bytes left to transfer.
                        long bytesLeft = 0;
                        // Grab a long on the upload table
                        // so that it does not change under our feet...
                        synchronized(logic.uploadingFiles)
                            {
                            for(int i = logic.uploadingFiles.size(); --i >= 0; )
                                {
                                final SelectedFileDetails selectedFileDetails = logic.uploadingFiles.get(i);
                                bytesLeft += selectedFileDetails.getEsa().length;
                                }
                            }
                        // Guess 10kBytes/s transfer speed until first measurements are available.
                        final long timeEst = bytesLeft / ((smoothedUploadSpeedBps <= 0) ? 10000 : smoothedUploadSpeedBps);
                        uploadCompleteValue.setText(String.valueOf(timeEst) + "s");
                        }
                    }

                // Only enable the buttons if the table is not empty!
                final boolean tableIsNotEmpty = (logic.uploadingFiles.size() > 0);
                fileClearButton.setEnabled(tableIsNotEmpty);
                fileRetryButton.setEnabled(tableIsNotEmpty);
                fileRetryAllButton.setEnabled(tableIsNotEmpty);

                // Highlight items being uploaded.
                for(int i = UploaderLogic.MAX_CONC_UPLOADS; --i >= 0; )
                    {
                    final UploadingNowItem uploadingNowItem = uploadingNow[i];
                    final UploaderLogic.UploadStatus status = logic.getInProgressSlot(i);
                    final JProgressBar progressBar = uploadingNowItem.progressBar;
                    if(status == null) // Not in use.
                        {
                        // Did anything change?
                        // Save lots of useless update CPU if not...
                        final String idleMsg = "idle";
                        if(!idleMsg.equals(progressBar.getString()))
                            {
                            progressBar.setString(idleMsg); // Hide % string.
                            uploadingNowItem.sizeBox.setText(null);
                            uploadingNowItem.nameBox.setText(null);
                            progressBar.setMinimum(0);
                            progressBar.setMaximum(0);
                            progressBar.setValue(0);
                            progressBar.setIndeterminate(false);

                            // Change main app title back to normal to show no uploads in progress...
                            UploaderMain.this.setTitle(APPLICATION_WINDOW_TITLE);

                            // Show any changes to the in-progress table...
                            queueTableModel.fireTableDataChanged();

                            // Log when all uploads have finished.
                            // This may log up to once per worker thread
                            // if all threads stop uploads very close together.
                            if((logic.uploadingFiles.size() == 0) && // None left to upload.
                               (logic.getCompletedUploadsThisSession() > 0)) // Not at start-up!
                                { logger.log("Uploads finished..."); }
                            }
                        }
                    else // In progress...
                        {
                        progressBar.setString(null); // Show % string.
                        final long fileLength = status.file.getEsa().length;
                        // Show (short) exhibit name and length.
                        final String newFileComponent = ExhibitName.getFileComponent(status.file.getEsa().filePath);
                        if(!newFileComponent.equals(uploadingNowItem.nameBox.getText()))
                            {
                            // If name for this upload has changed...
                            uploadingNowItem.nameBox.setText(newFileComponent);
                            // then show any changes to the in-progress table...
                            queueTableModel.fireTableDataChanged();
                            }
                        uploadingNowItem.sizeBox.setText(String.valueOf(fileLength));
                        // Make an int-safe clamped version of fileLength.
                        final int clampedLen = (int) Math.min(fileLength, Integer.MAX_VALUE);
                        // Maximum is "safe" clamped length of exhibit.
                        progressBar.setMinimum(0); // Minimum is 0 bytes uploaded.
                        progressBar.setMaximum(clampedLen);

                        final long bytesUploaded = status.getBytesUploaded();
                        // Value is bytes-sent clamped into range...
                        final int newProgressValue = Math.max(0, (int) Math.min(clampedLen, bytesUploaded));
                        if(newProgressValue != progressBar.getValue())
                            {
                            // If progress for this upload has changed...
                            progressBar.setValue(newProgressValue);
                            // then show any changes to the in-progress table...
                            queueTableModel.fireTableDataChanged();
                            }
                        // Indeterminate iff an out-of-range value,
                        // used to indicate not started or waiting for response...
                        progressBar.setIndeterminate((bytesUploaded < 0) || (bytesUploaded > fileLength));

                        // Change main app title to show upload(s) in progress...
                        final StringBuilder sb = new StringBuilder(APPLICATION_WINDOW_TITLE.length() + 32);
                        sb.append(APPLICATION_WINDOW_TITLE);
                        sb.append(": UPLOADING...");
                        final boolean oddSecond = (((System.currentTimeMillis() / 1000) & 1) != 0);
                        // Flashing constant-width activity indicator!
                        sb.append(oddSecond ? " .*" : " *.");
                        final String title = sb.toString();
                        if(!title.equals(UploaderMain.this.getTitle()))
                            { UploaderMain.this.setTitle(title); }
                        }
                    }
                }
            };

        return(new Tuple.Pair<JComponent, ActionListener>(pane, al));
        }



    /**Make the standard insets to wrap round most components.
     * Note that the returned Insets() object is mutable.
     */
    private static Insets makeStandardInsets()
        { return(new Insets(2, 2, 2, 2)); }

    /**Creates and initialises a status bar. */
    private JLabel createStatusBar()
        {
        final JLabel sbar = new JLabel("Ready");
        sbar.setBorder(BorderFactory.createEtchedBorder());
        return(sbar);
        }

    /**This method acts as the Action handler delegate for all the actions. */
    public void actionPerformed(final ActionEvent evt)
        {
        final String command = evt.getActionCommand();

        // Compare the action command to the known actions,
        // most-frequent first for efficiency.
        if(command.equals(selectAction.getActionCommand()))
            {
            // The "Select/Name for upload" action was invoked...
            uploadTableEditRow.set(-1); // Nothing in the table being edited now.
            selectFilesForUpload();
            }
        else if(command.equals(aboutAction.getActionCommand()))
            {
            // The "About" action was invoked...
            JOptionPane.showMessageDialog(this, aboutAction.getLongDescription(), aboutAction.getShortDescription(), JOptionPane.INFORMATION_MESSAGE);
            }
        else if(command.equals(exitAction.getActionCommand()))
            {
            // The "Exit" action was invoked...
            // Shut down unless vetoed...
            try { shutdown(); }
            catch(final UnsupportedOperationException e) { }
            }
        else
            {
            logger.log("Unexpected command: " + command);
            }

        // Note user activity...
        logic.setUserLastActive("actionPeformed: " + command);
        }

    /**This adapter is constructed to handle mouse-over component events. */
    private static final class MouseHandler extends MouseAdapter
        {
        private JLabel label;
        private String oldMsg;

        /**Adaptor constructor.
         * @param label the JLabel which will recieve value of the
         *              Action.LONG_DESCRIPTION key
         */
        public MouseHandler(final JLabel label)
            {
            setLabel(label);
            oldMsg = label.getText();
            }

        public void setLabel(final JLabel label)
            {
            this.label = label;
            }

        public void mouseEntered(final MouseEvent evt)
            {
            if(evt.getSource() instanceof AbstractButton)
                {
                AbstractButton button = (AbstractButton) evt.getSource();
                Action action = button.getAction(); // getAction is new in JDK 1.3
                if(action != null)
                    {
                    oldMsg = label.getText();
                    String message = (String) action.getValue(Action.LONG_DESCRIPTION);
                    label.setText(message);
                    }
                }
            }

        public void mouseExited(final MouseEvent evt)
            {
            label.setText(oldMsg);
            }
        }

    /**Select (in a file-chooser) some files for upload.
     * Since we will have to interact with disc, eg to compute hashes,
     * this may take a while for many and/or large files.
     */
    private void selectFilesForUpload()
        {
        // Selected files before we start...
        final int filesBefore = logic.selectedFiles.size();

        // Make sure initial name is unique if not already so.
        logic.uibIncNumberIfNotZeroWhileNotUnique(uib);

        // Putative complete name.
        final String putativeName = uib.getFullName();

        // Can only load a single file if number-in-series is 0.
        final boolean singleFile = (uib.getNumber() == 0);

        // We double-check the validity and uniqueness of putativeName
        // to avoid most obvious race conditions...
        final boolean nameOK = ExhibitName.validNameSyntax(putativeName) &&
            (logic.getAep().aeid.getStaticAttr(putativeName) == null) &&
            (logic.getAep().aeid.getFullName(ExhibitName.getFileComponent(putativeName)) == null);

        // Suggested extension(s).
        final String extensionComponent = ExhibitName.getExtensionComponent(putativeName);
        final String extensions[] = { extensionComponent };

        // Try to open/check the suggested file(s).
        final FileContents[] files;
        try
            {
            if(!nameOK)
                {
                JOptionPane.showMessageDialog(this, "Cannot upload with name: " + putativeName, selectAction.getShortDescription(), JOptionPane.ERROR_MESSAGE);
                return;
                }
            else if(logic.fos == null)
                {
                if(JWS_UPLOAD_ONLY)
                    {
                    JOptionPane.showMessageDialog(this, "Cannot upload: no JWS FileOpenService", selectAction.getShortDescription(), JOptionPane.ERROR_MESSAGE);
                    return;
                    }

                final JFileChooser chooser = new JFileChooser(logic.props.getFileChooserPathHint());
                chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
                chooser.setMultiSelectionEnabled(!singleFile);
                chooser.addChoosableFileFilter(new FileFilter()
                    {
                    /**Accept files with the correct extension(s). */
                    public final boolean accept(final File f)
                        {
                        // Allow directory traversal.
                        if(f.isDirectory()) { return(true); }

                        // Allow files with correct extension.
                        for(final String ext : extensions)
                            {
                            if(f.getPath().endsWith(ext) &&
                               f.getPath().endsWith("." + ext))
                                { return(true); }
                            }
                        return(false);
                        }

                    /**Get the description of each extension type accepted. */
                    public final String getDescription()
                        {
                        final StringBuilder sb = new StringBuilder();
                        for(final String ext : extensions)
                            {
                            if(sb.length() != 0) { sb.append(", "); }
                            final ExhibitMIME.ExhibitTypeParameters exhibitType = ExhibitMIME.getExhibitType(ext);
                            if(null != exhibitType)
                                {
                                final String description = exhibitType.description;
                                if(description.length() > 0)
                                    {
                                    sb.append(description);
                                    sb.append(' ');
                                    }
                                }
                            sb.append("*.").append(ext);
                            }
                        return(sb.toString());
                        }
                    });
                if(JFileChooser.APPROVE_OPTION != chooser.showDialog(this, "Select"))
                    { files = new FileContents[0]; }
                else
                    {
                    File[] selectedFiles = chooser.getSelectedFiles();
                    if(((selectedFiles == null) || (selectedFiles.length == 0)) &&
                       (chooser.getSelectedFile() != null))
                        { selectedFiles = new File[]{chooser.getSelectedFile()}; }
                    files = new FileContents[selectedFiles.length];
                    for(int i = selectedFiles.length; --i >= 0; )
                        {
                        final File f = selectedFiles[i];
                        assert(f != null);
                        files[i] = new NormalFileContent(f);
                        }
                    }
                // Try to remember where we were...
                logic.props.setFileChooserPathHint(chooser.getCurrentDirectory());
                }
            else
                {
                if(singleFile)
                    {
                    final File dir = logic.props.getFileChooserPathHint();
                    final FileContents fileContents = logic.fos.openFileDialog(
                        (dir == null) ? null : dir.getPath(), extensions);
                    if(fileContents == null)
                        { files = new FileContents[0]; }
                    else
                        { files = new FileContents[]{fileContents}; }
                    }
                else
                    {
                    files = logic.fos.openMultiFileDialog(null, extensions);
                    }
                }

            // Use the directory (if any) of the last file chosen
            // to start the next search.
            for(int i = files.length; --i >= 0; )
                {
                if(files[i] == null) { continue; }
                final File parentFile = (new File(files[i].getName())).getParentFile();
                if(parentFile == null) { continue; }
                logic.props.setFileChooserPathHint(parentFile);
                break;
                }

            if(files.length > 0)
                {
                final boolean autoMode = autoSuffixCheckBox.isSelected();

                // THIS CALL may be slow if there are many/large files.
                final String err = logic.addSelectedFiles(autoMode, files, uib); // FIXME

                if(err != null)
                    {
                    JOptionPane.showMessageDialog(this, err, "File problem", JOptionPane.ERROR_MESSAGE);
                    return;
                    }
                }
            }
        catch(final IOException e)
            {
            logger.log("Failed to open file(s) for upload: " + e.getMessage());
            JOptionPane.showMessageDialog(this, "Cannot upload: IOException: " + e.getMessage(), selectAction.getShortDescription(), JOptionPane.ERROR_MESSAGE);
            return;
            }
        finally
            {
            // Make sure that the table display is up-to-date...
            // ...and that we are not "editing" any row in the table.
            uploadTableEditRow.set(-1); // Nothing in the table being edited now.
            uploadTableModel.fireTableDataChanged();
            }

        final int filesAfter = logic.selectedFiles.size();
        JOptionPane.showMessageDialog(this, "Selected new files for upload: " + (filesAfter-filesBefore), selectAction.getShortDescription(), JOptionPane.INFORMATION_MESSAGE);
        }




    /**ActionListener invoked by pollUI; never null. */
    private final ActionListener pollAL;

    /**Poll UI (in Swing thread, ie Swing-safe). */
    private void pollUI(final ActionEvent e)
        {
        pollAL.actionPerformed(e);
        }


    /**Listener class used to veto attempts to start another app instance.
     * A (modal) dialog[ue] is shown instead.
     */
    private static final class SISListener implements SingleInstanceListener
        {
        public final void newActivation(final String[] params)
            {
            System.err.println("Attempted to launch another instance, args: " + Arrays.asList(params));

            // Schedule a job for the event-dispatching thread:
            // creating and a blocking dialogue.
            javax.swing.SwingUtilities.invokeLater(new Runnable(){
                public final void run()
                    {
                    JOptionPane.showMessageDialog(null, "Uploader already running...", "Already running", JOptionPane.ERROR_MESSAGE);
                    }
                });
            }
        }

    /**Main method invoked from JWS.
     */
    public static void main(final String[] args)
        {
        // Try to set a local look-and-feel...
        try { UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); }
        catch(final Exception e) { }

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        // Also start a daemon thread to drive async non-GUI activity.
        javax.swing.SwingUtilities.invokeLater(new Runnable(){
            public final void run()
                {
                final UploaderMain mainFrame = new UploaderMain();
                mainFrame.pack();
                mainFrame.setVisible(true);

                // After displaying the UI,
                // but before allowing much else to happen,
                // reload any persisted data
                // and start any background processing...
                mainFrame.logic.startup();

                // Start a (daemon) poller thread for UI async activity.
                // This is called from a Swing timerUI, so is Swing-safe.
                // This is created and started AFTER the UI object construction
                // is complete to avoid any unpleasant races.
                // We call this every 100ms or so in order to feel responsive.
                final Timer timerUI = new javax.swing.Timer(70 + Rnd.fastRnd.nextInt(30),
                                                            (new ActionListener(){
                    /**Poll the UI logic. */
                    public void actionPerformed(final ActionEvent e)
                        { mainFrame.pollUI(e); }
                    }));
                // Delay first poll a little to allow system to start up...
//                timerUI.setInitialDelay(301);
                timerUI.start();
                }
            });
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -8885213711913005041L;
    }
