/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.clApp.uploader;

import org.hd.d.pg2k.svrCore.ExhibitName;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;


/**Holder for uploader mutable persistent properties, preferences, etc.
 * Holds a small amount of structured information
 * that can be persisted from one uploader run to the next.
 * <p>
 * Is thread-safe, though using multiple threads at once,
 * especially mutators, may be confusing.
 * <p>
 * Not designed to be Serialisable,
 * but instead can be stored in the plain-text Java properties format.
 * <p>
 * Is designed to be GUI-free.
 * <p>
 * Package visible since need be seen only by the main GUI class.
 * <p>
 * TODO: ADD locale
 * <p>
 * TODO: ADD upload category / main words / number-in-series / type / description
 */
final class UploaderProps
    {
    /**Maximum length (in bytes) expected necessary to persist this data; strictly positive. */
    public static final int MAX_PERS_BYTES = 8192;


    /**Internal flag to note if any flags have changed.
     * Volatile for thread-safe access without a lock.
     * <p>
     * Because of possible races this cannot be guaranteed to catch all changes,
     * so a final "save-on-exit" and/or periodic save would be valuable.
     */
    private volatile boolean changed;


    /**User ID (author initials), initially null; null or valid author initials.
     * Would usually be the last set of initials (ie user ID) used.
     * <p>
     * Volatile for thread-safe access without a lock.
     */
    private volatile String userID;

    /**Get the user ID (author initials), initially null; null or valid author initials.
     * Usually be the last set of initials (ie user ID) used
     * to attempt a server connection.
     */
    public String getUserID() { return(userID); }

    /**Set the user ID (author initials), null or valid author initials.
     * An attempt to set with an invalid value is treated as if null.
     */
    public void setUserID(final String authInitials)
        {
        if(!ExhibitName.validAuthorSyntax(authInitials))
            {
            if(userID == null) { return; } // Nothing to do...
            changed = true;
            userID = null;
            }
        else
            {
            if(authInitials.equals(userID)) { return; } // Nothing to do...
            changed = true;
            userID = authInitials;
            }
        }



    /**File chooser directory hint, or null.
     * The last directory that a user looked in for files, eg to upload.
     * <p>
     * Volatile for thread-safe access without a lock.
     */
    private volatile File fileChooserPathHint;

    /**Get the file chooser directory hint, or null.
     * The last directory that a user looked in for files, eg to upload.
     */
    public File getFileChooserPathHint() { return(fileChooserPathHint); }

    /**Set the file chooser directory hint, or null.
     */
    public void setFileChooserPathHint(final File pathHint)
        {
        if(pathHint == null)
            {
            if(fileChooserPathHint == null) { return; } // Nothing to do...
            changed = true;
            fileChooserPathHint = null;
            }
        else
            {
            if(pathHint.equals(fileChooserPathHint)) { return; } // Nothing to do...
            changed = true;
            fileChooserPathHint = pathHint;
            }
        }




    /**Get data in properties-file format byte[]; never null. */
    public byte[] getPersistentData()
        {
        final Properties props = new Properties();

        final String u = userID;
        if(u != null) { props.put("userID", u); }
        final File ph = fileChooserPathHint;
        if(ph != null) { props.put("fileChooserPathHint", ph.getPath()); }

        final ByteArrayOutputStream baos = new ByteArrayOutputStream(256);
        try { props.store(baos, null); }
        catch(final IOException e) { throw new Error("INTERNAL ERROR"); }

        return(baos.toByteArray());
        }

    /**Get data in properties file if properties have changes, else null.
     * The changed flag is cleared if this routine returns a non-null value,
     * thus it can be used to save incremental updates asynchronously.
     */
    public byte[] getPersistentDataIfChanged()
        {
        if(!changed) { return(null); }

        // Clear changed flag *before* getting data to help reduce race effects.
        changed = false;
        return(getPersistentData());
        }

    /**Load from previously-generated persistent data format.
     * Any invalid data is simply ignored.
     * <p>
     * All updates are done as if through the setXXX() members
     * to validate/coerce them, concurrently with other use of this instance.
     */
    public void loadFromPersistentData(final InputStream is)
        throws IOException
        {
        if(is == null) { throw new IllegalArgumentException(); }
        final Properties props = new Properties();
        props.load(is);

        // Retrieve the values from the properties set.
        setUserID(props.getProperty("userID"));
        final String ph = props.getProperty("fileChooserPathHint");
        setFileChooserPathHint((ph == null) ? null : new File(ph));
        }
    }
