/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**Immutable, Serializable collection of all the accession data for an exhibit.
 * This contains all of the available accession data for an exhibit,
 * ie data collected at or after the exhibit was added to the library,
 * such as checksums to verify that the exhibit has not become corrupted.
 * <p>
 * In some cases not all fields will be complete.
 * <p>
 * The core fields are made available as public final fields
 * for speed of access.
 * <p>
 * We do not ever expect two instances for different exhibits
 * to be equal, so the only reason to intern() the whole object
 * is to reduce heap churn between an extant and new/updated exhibit set.
 * We don't do this automatically during deserialisation,
 * and leave it up to a user of this type to do so if they wish.
 * We <em>do not</em> intern() the fields of this object individually,
 * to help avoid pointless space and time being taken in intern().
 */
public final class AccessionData implements Serializable,
                                            ObjectInputValidation,
                                            MemoryTools.Internable
    {
    /**Top-level node name for XML representation, ie tag name. */
    public static final String NAME_TOP_NODE = "accessionData";

    /**Name of the accession date sub-node, if any. */
    public static final String NAME_ACCESSION_DATE = "date";

    /**Name of the accession-time size sub-node, if any. */
    public static final String NAME_ACCESSION_SIZE = "size";

    /**Name of the accession-time whole-exhibit CRC32 hash sub-node, if any. */
    public static final String NAME_ACCESSION_CRC32 = "hash-CRC32";

    /**Name of the accession-time whole-exhibit MD5 hash sub-node, if any. */
    public static final String NAME_ACCESSION_MD5 = "hash-MD5";


    /**Block size (in bytes) used for partial hashes; strictly positive power of two.
     * This is can be used while reading in a file in a block at a time
     * to avoid holding a lock too long,
     * and as the size of each block of a per-block hash
     * such as can be used to check validity of the start of
     * a partly-cached exhibit.
     */
    public static final int HASH_BLOCK_SIZE_BYTES = 1024 * 1024;


    /**Nominal time/data at which exhibit was added to the Gallery; null or strictly positive.
     * This is null if not available,
     * ie the accession time/date was not recorded,
     * else it is the strictly positive Java time in milliseconds
     * (since 19700101 00:00 UTC).
     * <p>
     * This can the same as, or later than, the exhibit's timestamp.
     * If earlier than the exhibit's timestamp it may mean that the exhibit
     * has been altered in some way.
     */
    public final Long date;

    /**Size of exhibit file at time of accession; null or strictly positive.
     * This is null if not available,
     * ie the accession time/date was not recorded,
     * else it is the strictly positive number of bytes in the file.
     * <p>
     * This should be identical to the current exhibit length
     * else it may mean that the exhibit has been altered in some way.
     */
    public final Long size;

    /**Accession-time CRC32 whole-exhibit hash or null if not available.
     * We use CRC32 since it is commonly used elsewhere as a file checksum,
     * for example in ZIP files.
     */
    public final Integer hashCRC32;

    /**Accession-time MD5 whole-exhibit hash or null if not available.
     * If non-null the array is always exactly 16 bytes long.
     * <p>
     * We use MD5 since it is commonly used elsewhere as a file checksum,
     * and will in practice detect any conceivable corruption in exhibits
     * especially given that we separately record the length.
     */
    public final ROByteArray hashMD5;


    /**If true, some accession data is missing (eg null). */
    public final boolean isIncomplete()
        {
        // Check all fields are non-null...
        if(date == null) { return(true); }
        if(size == null) { return(true); }
        if(hashCRC32 == null) { return(true); }
        if(hashMD5 == null) { return(true); }

        return(false); // Everything is present that should be...
        }

    /**Private shared factory_pFX for getAsDOM(); never null. */
    private static final DocumentBuilderFactory dbf_gAD = DocumentBuilderFactory.newInstance();

    /**Get exhibit accession metadata as DOM, never null.
     * This data includes such items as:
     * <ul>
     * <li>When the exhibit was loaded into to the Gallery.
     * <li>Its checksums, timestamp, length, etc when loaded into the Gallery.
     * </ul>
     * <p>
     * All this data is also accessible by (probably cheaper)
     * direct accessor methods or fields if required.
     */
    public synchronized Node getAsDOM()
        {
        final DocumentBuilder db;
        try { db = dbf_gAD.newDocumentBuilder(); }
        catch(final ParserConfigurationException e)
            {
            e.printStackTrace();
            throw new Error("Cannot create DOM form of accession data", e);
            }
        final Document doc = db.newDocument();

        // Create a provisional root node,
        // but only assign it to the result (ie to get a non-null result)
        // when we know that there is some real data in there.
        final Element root = doc.createElement(NAME_TOP_NODE);
        doc.appendChild(root);

        if(date != null)
            {
            final Element e = doc.createElement(NAME_ACCESSION_DATE);
            e.setAttribute("value", date.toString());
            root.appendChild(e);
            }

        if(size != null)
            {
            final Element e = doc.createElement(NAME_ACCESSION_SIZE);
            e.setAttribute("value", size.toString());
            root.appendChild(e);
            }

        if(hashCRC32 != null)
            {
            final Element e = doc.createElement(NAME_ACCESSION_CRC32);
            e.setAttribute("value", Long.toHexString(hashCRC32.intValue() & 0xffffffffL));
            root.appendChild(e);
            }

        if(hashMD5 != null)
            {
            final Element e = doc.createElement(NAME_ACCESSION_MD5);
            e.setAttribute("value", hashMD5.toHexString());
            root.appendChild(e);
            }

        // Insert other elements here...

        return(root);
        }


    /**The hashCode is VM-independent.
     * The size and CRC32 fields are used in the computation.
     */
    @Override
	public final int hashCode()
        {
        int result = 539235362;
        if(size != null) { result ^= (int) size.longValue(); }
        if(hashCRC32 != null) { result ^= hashCRC32.intValue(); }
        return(result);
        }

    /**All fields are compared. */
    @Override
	public final boolean equals(final Object o)
        {
        if(this == o) { return(true); }
        if(!(o instanceof AccessionData)) { return(false); }
        final AccessionData other = (AccessionData) o;

        if(date == null) { if(other.date != null) { return(false); } }
        else if(!date.equals(other.date)) { return(false); }

        if(size == null) { if(other.size != null) { return(false); } }
        else if(!size.equals(other.size)) { return(false); }

        if(hashCRC32 == null) { if(other.hashCRC32 != null) { return(false); } }
        else if(!hashCRC32.equals(other.hashCRC32)) { return(false); }

        // Quite a lot of data to compare, so may be slow.
        if(hashMD5 == null) { if(other.hashMD5 != null) { return(false); } }
        else if(!hashMD5.equals(other.hashMD5)) { return(false); }

        return(true); // Equal!
        }




    /**Construct new immutable instance with no data.
     * Is private to help with instance control.
     */
    private AccessionData()
        { this(null, null, null, null); }

    /**Construct new immutable instance with the given data.
     * @throws IllegalArgumentException  in the case of bad data
     */
    public AccessionData(final Long accessionDate,
                         final Long accessionSize,
                         final Integer accessionCRC32,
                         final ROByteArray accessionMD5)
        throws IllegalArgumentException
        {
        date = accessionDate;
        size = accessionSize;
        hashCRC32 = accessionCRC32;
        hashMD5 = accessionMD5;

        // Verify what we've been given.
        try { validateObject(); }
        catch(final InvalidObjectException e)
            { throw new IllegalArgumentException(e); }
        }

    /**Public empty instance. */
    public static final AccessionData EMPTY = new AccessionData();

    /**Private shared factory_pFX for parseFromXML(); never null. */
    private static final DocumentBuilderFactory factory_pFX = DocumentBuilderFactory.newInstance();

    /**Configure factory_pFX. */
    static
        {
        // Configure the factory_pFX object settings.
        factory_pFX.setValidating(false); // No validating (yet)...
        factory_pFX.setNamespaceAware(false); // No namespaces (yet)...
        factory_pFX.setIgnoringElementContentWhitespace(true); // Trim unnecessary whitespace.
        }

    /**Parse from XML format UNICODE text.
     * We will attempt to be generous in our parsing,
     * but must be able to extract the top-level node
     * and at least one sub-ordinate node
     * else we will assume that we have been given a bogus file.
     *
     * @param xml  must be correct XML format and start with '&lt;',
     *     not null, not zero-length
     *
     * @throws IOException  in case of difficulty parsing the text
     */
    public static AccessionData parseFromXML(final String xml)
        throws IOException
        {
        if(xml == null) { throw new IllegalArgumentException(); }

        if(!xml.startsWith("<"))
            { throw new IOException("text must start with '<'"); }

        try
            {
            final DocumentBuilder builder = factory_pFX.newDocumentBuilder();

            // Parse the XML input String...
            final Document document = builder.parse(new InputSource(new StringReader(xml)));
            final Node root = document.getFirstChild();

            if(!NAME_TOP_NODE.equals(root.getNodeName()))
                { throw new IOException("invalid top-level node: expected " + NAME_TOP_NODE); }

            // Prepare to load the individual fields.
            Long date = null;
            Long size = null;
            Integer hashCRC32 = null;
            ROByteArray hashMD5 = null;

            for(Node child = root.getFirstChild(); child != null; child = child.getNextSibling())
                {
                // Extract the "value" attribute if present.
                String value = null;
                final NamedNodeMap attributes = child.getAttributes();
                if(attributes != null)
                    {
                    final Node namedItem = attributes.getNamedItem("value");
                    if(namedItem != null)
                        { value = namedItem.getNodeValue(); }
                    }

                final String name = child.getNodeName();
                if(NAME_ACCESSION_DATE.equals(name))
                    { if(value != null) { date = new Long(value); } }
                else if(NAME_ACCESSION_SIZE.equals(name))
                    { if(value != null) { size = new Long(value); } }
                else if(NAME_ACCESSION_CRC32.equals(name))
                    { if(value != null) { hashCRC32 = new Integer((int) Long.parseLong(value, 16)); } }
                else if(NAME_ACCESSION_MD5.equals(name))
                    { if(value != null) { hashMD5 = ROByteArray.fromHexString(value); } }
                else
                    { throw new IOException("unexpected node: " + name); }
                }

            final AccessionData result = new AccessionData(date, size, hashCRC32, hashMD5);
            if(result.equals(EMPTY)) { return(EMPTY); }
            return(result);
            }
        catch(final ParserConfigurationException e)
            {
            throw new Error("unable to configure XML parser", e);
            }
        catch(final SAXException e)
            {
            throw new IOException("problem parsing XML: " + e.getMessage());
            }
        }

    /**Construct accession data purely from the exhibit file.
     * This does not look at any extant or old-style accession data.
     * <p>
     * This does not check that the exhibit is valid except that it is
     * a plain file and not zero-length.
     */
    public static AccessionData fromExhibitFile(final File f)
        throws IOException
        {
        if(f == null) { throw new IllegalArgumentException(); }

        final long size = f.length();
        if(!f.exists() || !f.isFile() || !f.canRead() || (size < 1))
            { throw new IOException("file not readable or is zero-size"); }

        final long date = f.lastModified();
        if(date <= 0)
            { throw new IOException("invalid timestamp on file"); }

        final InputStream is = new FileInputStream(f);
        try
            {
            final Tuple.Pair<Integer,ROByteArray> hashes =
                computeFullFileHashes(is);

            return(new AccessionData(new Long(date),
                                     new Long(size),
                                     hashes.first,
                                     hashes.second));
            }
        finally { is.close(); }
        }

    /**Compute the full-file hashes over an exhibit; never null and neither element of the Pair null.
     * This reads the entire InputStream and computes all the full-stream hashes
     * at once for efficiency.
     * <p>
     * This does not close the InputStream.
     * <p>
     * The InputStream need not be buffered since we read in large efficient
     * blocks anyway if possible.
     *
     * @return the CRC32 checksum as an Integer in the first element,
     *         and the MD5 hash as an ROByteArray in the second element
     */
    public static Tuple.Pair<Integer,ROByteArray> computeFullFileHashes(final InputStream is)
        throws IOException
        {
        final java.util.zip.Checksum hCRC32 = new java.util.zip.CRC32();
        final MessageDigest hMD5;
        try { hMD5 = MessageDigest.getInstance(CoreConsts.HASH_MD5); }
        catch(final NoSuchAlgorithmException e) // Should never happen...
            { throw new Error("could not find "+CoreConsts.HASH_MD5+" digester!"); }

        // Use a fairly big buffer for read efficiency.
        final byte buf[] = new byte[1 << 16];
        int n;
        while((n = is.read(buf)) > 0)
            {
            // Update the hashes with the next block of data.
            hCRC32.update(buf, 0, n);
            hMD5.update(buf, 0, n);
            }

        assert(is.read() == -1); // Should be at EOF...

        // Return the hashes.
        return(new Tuple.Pair<Integer,ROByteArray>(
            new Integer((int) hCRC32.getValue()),
            new ROByteArray(hMD5.digest())
            ));
        }



    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -2699150966581906068L;

    /**Deserialise: use constructor for validation, defensive copying, etc. */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Eliminate duplicates of the empty instance.
        if(equals(EMPTY)) { return(EMPTY); }

        // Construct new instance of object in normal defensive way.
        // This also ensures that we get a chance to intern() stuff, etc.
        // However, we do not ever expect two instances for different exhibits
        // to be equal, so the only reason to intern() the whole object
        // is to reduce heap churn between an old and new/updated exhibit set.
        return(new AccessionData(date, size, hashCRC32, hashMD5));
        }


    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        if((date != null) && !GenUtils.isValidGalleryTimestamp(date.longValue()))
            { throw new InvalidObjectException("bad object: invalid date/timestamp "+date+ " ("+(new java.util.Date(date.longValue()))+")"); }
        if((size != null) && (size.longValue() <= 0))
            { throw new InvalidObjectException("bad object: non-positive size"); }
        if((hashMD5 != null) && (hashMD5.length() != 16))
            { throw new InvalidObjectException("bad object: incorrect size hashMD5"); }
        }
    }
