/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the
distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.xbill.DNS.*;

import java.lang.ref.SoftReference;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.*;

/**This class has tools for IP-address and DNS manipulation.
 *
 * @author Damon Hart-Davis
 */
public final class AddrTools
    {
    /**Prevent creation of instances of this class. */
    private AddrTools() { }

    /**If true, use DNSJava to do DNS lookups where possible, else use Java's built-in DNS resolver.
     * DNSJava does not retain all results indefinitely unlike Java's DNS resolver,
     * and allows us to use separate caches with different characteristics for different query types
     * which can make it much more robust for our purposes.
     */
    public static final boolean USE_DNSJAVA = true;

    /**Fixed empty search path for lookups. */
    private static final Name[] EMPTY_SEARCH_PATH = new Name[0];

    /**Approximate total maximum timeout for DNS resolution in seconds; strictly positive.
     * Quite short for the the geo-lookup (etc)
     * as we usually only get limited value from a small number of lookups,
     * and we do not want to block interactive activity for an extended period.
     * <p>
     * A value of a few seconds to a few tens of seconds should normally suffice,
     * and puts a reasonable cap on the cost of a "full" lookup.
     */
    private static final int RESOLVER_QUICK_TIMEOUT_S = 5 + Rnd.fastRnd.nextInt(2);

    /**Maximum retries of resolution against each DNS server/resolver; strictly positive.
     * A value greater than 1 is more robust,
     * ie allows for the odd lost packet,
     * and may get good results because the retry may gain the benefit
     * of async cacheing upstream between calls on one particular resolver.
     * <p>
     * A value of 1 or 2 is typical.
     */
    private static final int RESOLVER_QUICK_RETRIES = 2;

    /**SoftReference to our shared DNS cache (for IN records); never null though referent may be.
     * This may be created and set up on first use,
     * and is tuned for our purposes
     * (for example shorter timeouts and longer negative cacheing than default).
     * It may be discarded and recreated if we are short of memory;
     * hopefully caching by boundary DNS resolvers should minimise
     * the impact if we do this.
     * <p>
     * We don't hand this out to other users,
     * but they can give us a Lookup object and we will set the cache
     * and other relevant parameters,
     * and return the result to them.
     * <p>
     * This is intended to be highly threadable for many concurrent lookups.
     * <p>
     * Get/set access to this value is synchronised on the _cacheIN_SR_lock.
     */
    private static SoftReference<Cache> _cacheIN_SR = new SoftReference<Cache>(null);

    /**Lock for get/set access to _cacheIN_SR SoftReference value. */
    private static final Object _cacheIN_SR_lock = new Object();

    /**Default maximum time to positively cache a record for (seconds); strictly positive.
     * Limiting this to a few hours at most can limit memory load
     * when there are many users.
     */
    private static final int DEFAULT_DNS_CACHE_S = 3601 + Rnd.fastRnd.nextInt(3600);

    /**Default maximum time to negatively cache a record for (seconds); strictly positive.
     * Setting this to a few tens of seconds
     * (plus a few interleaved "failed lookup" intervals)
     * can improve performance.
     */
    private static final int DEFAULT_DNS_NCACHE_S = 31 + Rnd.fastRnd.nextInt(23) +
            (7 * RESOLVER_QUICK_TIMEOUT_S);

    /**Our ExtendedResolver with the normal servers but a short timeout.
     * The DNS servers will be deduced once at initialisation.
     */
    private static final ExtendedResolver RESOLVER_QUICK;

    /**Initialise our resolver. */
    static
        {
        try {
            RESOLVER_QUICK = new ExtendedResolver();
            RESOLVER_QUICK.setLoadBalance(true); // Be nice to upstream resolvers.
            RESOLVER_QUICK.setRetries(RESOLVER_QUICK_RETRIES);
            final int resolverCount = RESOLVER_QUICK.getResolvers().length; // Resolvers found.
            // Set the timeout per resolver to get the appoximate maximum time specified.
            RESOLVER_QUICK.setTimeout(Math.max(1, RESOLVER_QUICK_TIMEOUT_S /
                    (RESOLVER_QUICK_RETRIES * Math.max(1, resolverCount))));
            }
        catch(final UnknownHostException e)
            {
            e.printStackTrace();
            throw new Error(e);
            }
        }

    /**If true then we randomise the order in which we do RBL lookups.
     * We do this to better spread the load if several of them could
     * answer the question.
     * <p>
     * This does not have to be "very" random,
     * indeed, round-robin would probably do just as well.
     */
    private static final boolean RANDOMISE_RBL_LOOKUP_ORDER = true;


    /**Return access to the DNS cache, creating and initialising it if necessary; never null.
     * If we are short of memory we may automatically discard this entire cache.
     * <p>
     * This cache may have several parameters tuned for geo-lookup purposes,
     * such as negative/positive cache time-limits.
     */
    private static Cache _getDNSCache()
        {
        synchronized(_cacheIN_SR_lock)
            {
            Cache result = _cacheIN_SR.get();
            if(result == null)
                {
                result = new Cache(DClass.IN);
                result.setMaxCache(DEFAULT_DNS_CACHE_S);
                result.setMaxNCache(DEFAULT_DNS_NCACHE_S);
                result.setMaxEntries(101010); // Limit the cache size to something bearable.
//                result.setCleanInterval(Math.max(5, DEFAULT_DNS_CACHE_S / 60)); // Set to match +ve cacheing time...
                _cacheIN_SR = new SoftReference<Cache>(result);
                }
            return(result);
            }
        }

    /**Do (quick) blocking lookup in our shared and tuned cache and resolver.
     * This should work better for us than the built-in Java implementation
     * and we should build a useful but size-bounded cache unless
     * the JVM gets very low on memory, in which case we may discard it and start again.
     * <p>
     * The caller should <em>not</em> extract/sequester/inspect
     * the cache/resolver/etc properties that we set in the Lookup parameter,
     * so this routine is only package-visible.
     * <p>
     * FIXME: As of 20060703: most expensive consumer of time for doHTML.jsp.
     */
    static Record[] doQuickLookupWithTunedCache(final Lookup lookup)
        {
//final long startTime = System.currentTimeMillis();

        lookup.setCache(_getDNSCache());
        lookup.setSearchPath(EMPTY_SEARCH_PATH);
        lookup.setResolver(RESOLVER_QUICK);
        final Record[] result = lookup.run();

//final long time = System.currentTimeMillis() - startTime;
//if(time > 1) /* Ignore noise. */ { (new Throwable("time: "+time+"ms")).printStackTrace(); }

        return(result);
        }

//    /**Do local lookup in our shared and tuned cache only; never null.
//     * Do not do DNS resolution; return nothing if there is nothing in cache.
//     * <p>
//     * The Name supplied is looked up exactly as-is,
//     * so had better be absolute for example.
//     * <p>
//     * We only make this package-visible in case the DNS java classes
//     * can leak state.
//     */
//    static SetResponse doLookupInTunedCacheOnly(final Name name,
//                                                final int type,
//                                                final byte minCred)
//        {
//        return(_getDNSCache().lookupRecords(name, type, minCred));
//        }

    /**Return true if we can find an A record for the given name quickly.
     * Uses our shared/tuned cache and does a quick lookup
     * (ie times out quickly).
     *
     * @throws IllegalArgumentException  if argument is null or unparsable.
     */
    public static boolean hasARecordQuick(final String name)
        {
        if(name == null)
            { throw new IllegalArgumentException(); }

        try
            {
            final Lookup lookup = new Lookup(name, Type.A);
            final Record[] result = doQuickLookupWithTunedCache(lookup);

            // If we got some results, there is an A record..
            return((result != null) && (result.length > 0));
            }
        catch(final TextParseException e)
            {
            throw new IllegalArgumentException("bad DNS name: " + name);
            }
        }

    /**Convert an (IPv4) InetAddress to dotted-quad form.
     * Avoid doing any DNS lookup if at all possible.
     *
     * @param addr  parameter to convert to dotted-quad;
     *     must be non-null IPv4 address
     *
     * @throws IllegalArgumentException  if addr is null or not an IPv4 address
     */
    public static String toStringDottedQuad(final InetAddress addr)
        throws IllegalArgumentException
        {
        if(addr == null)
            { throw new IllegalArgumentException(); }

        final byte raw[] = addr.getAddress();

        if(raw.length != 4)
            { throw new IllegalArgumentException("not an IPv4 address"); }

        final StringBuilder sb = new StringBuilder(15);
        for(int i = 0; i < 4; ++i)
            {
            if(sb.length() != 0) { sb.append('.'); }
            sb.append(raw[i] & 0xff);
            }

        return(sb.toString());
        }

    /**Convert an (IPv4) InetAddress to reverse dotted-quad form with a trailing dot.
     * This can be useful for RBL and reverse lookups.
     * <p>
     * Avoid doing any DNS lookup if at all possible.
     *
     * @param addr  parameter to convert to reverse dotted-quad;
     *     must be non-null IPv4 address
     *
     * @throws IllegalArgumentException  if addr is null or not an IPv4 address
     */
    public static String toStringReverseDottedQuad(final InetAddress addr)
        throws IllegalArgumentException
        {
        if(addr == null)
            { throw new IllegalArgumentException(); }

        final byte raw[] = addr.getAddress();

        if(raw.length != 4)
            { throw new IllegalArgumentException("not an IPv4 address"); }

        final StringBuilder sb = new StringBuilder(16);
        for(int i = raw.length; --i >= 0; )
            {
            sb.append(raw[i] & 0xff).append('.');
            }

        return(sb.toString());
        }

    /**Do reverse lookup on IP address to get the name, returns null if lookup fails.
     * This can optionally verify the name by trying to look up the name it
     * found and making sure that at least one of the IPs returned matches.
     * <p>
     * This avoids the built-in JVM lookup routines to avoid the default
     * infinite cacheing that it does for class-loading safety.
     * <p>
     * Blocks until it has an answer or times out.
     *
     * @param addr  IP address to look up as numeric address; never null
     * @param verify  if true, attempt to verify with forward lookup of
     *     the putative result, returning null if no suitable match is found
     */
    public static String doReverseLookup(final String addr,
                                         final boolean verify)
        {
        if(addr == null)
            { throw new IllegalArgumentException(); }

        try { return(doReverseLookup(InetAddress.getByName(addr), verify)); }
        // Return null for an unparsable address.
        catch(final UnknownHostException e) { return(null); }
        }

    /**Do reverse lookup on IP address to get the name, returns null if lookup fails.
     * This can optionally verify the name by trying to look up the name it
     * found and making sure that at least one of the IPs returned matches.
     * <p>
     * This avoids the built-in JVM lookup routines to avoid the default
     * infinite cacheing that it does for class-loading safety.
     * <p>
     * Blocks until it has an answer or times out.
     *
     * @param addr  IP address to look up; never null
     * @param verify  if true, attempt to verify with forward lookup of
     *     the putative result, returning null if no suitable match is found
     */
    public static String doReverseLookup(final InetAddress addr,
                                         final boolean verify)
        {
        if(addr == null)
            { throw new IllegalArgumentException(); }

//        // Special case for loopback address
//        // to avoid going to DNS
//        // which should enhance speed and robustness.
//        //
//        // This does not need verification.
//        if(isLoopbackAddress(addr))
//            { return(DEFAULT_IPV4_LOOPBACK_CANON_NAME); }

//        Options.set("verbose"); // Make DNS java talkative.

//System.err.println("Doing reverse lookup for address: " + toStringDottedQuad(addr));
        try
            {
            // Do the reverse lookupRev...
//            final String result = Address.getHostName(addr);
            // Use our tuned cache to do it.
            final Name name = ReverseMap.fromAddress(addr);
            final Lookup lookupRev = new Lookup(name, Type.PTR);
            final Record recordsRev[] = doQuickLookupWithTunedCache(lookupRev);
            if((recordsRev == null) || (recordsRev.length < 1))
                { throw new UnknownHostException("unknown address: " + addr.getHostAddress()); }
            final PTRRecord ptr = (PTRRecord) recordsRev[0];
            final String result = ptr.getTarget().toString();

//System.err.println("  Reverse lookupRev result was: " + result);

            // If not verifying, return the result immediately.
            if(!verify) { return(result); }

            // If verifying, check against all addresses for the forward lookup.
//            final InetAddress addrs[] = Address.getAllByName(result);
            // Use our tuned cache.
            final Lookup lookupFwd = new Lookup(result, Type.A);
            final Record recordsFwd[] = doQuickLookupWithTunedCache(lookupFwd);
            if((recordsFwd == null) || (recordsFwd.length < 1))
                { throw new UnknownHostException("unknown host: " + result); }
            final InetAddress[] addrs = new InetAddress[recordsFwd.length];
            for(int i = addrs.length; --i >= 0; )
                {
                final ARecord a = (ARecord) recordsFwd[i];
                addrs[i] = a.getAddress();
                }
//System.err.println("  Verifying, forward lookupRev answers: " + addrs.length);
            for(int i = addrs.length; --i >= 0; )
                {
                // If an address matched then the address is verified.
                if(addr.equals(addrs[i])) { return(result); }
                }

            return(null); // Failed to verify the address.
            }
        catch(UnknownHostException e)
            {
            // Lookup failed.
//System.err.println("  Lookup failed: " + e.getMessage());
            return(null);
            }
        catch(TextParseException e)
            {
//System.err.println("  Lookup failed: " + e.getMessage());
            return(null);
            }
        }

    /**Check an (IPv4) IP address with the RBL servers for being "blackholed".
     * For an IPv4 address of the numerical form a.b.c.d,
     * list does a lookup for an A record for d.c.b.a.RBL.domain,
     * and if it finds one, the client address is "bad".
     * <p>
     * We may do these checks serially, or in parallel for speed.
     * We may pick one RBL server at random each time for speed.
     * The first suitable A record returned stops the search.
     * <p>
     * We may randomise the order in which we do lookup to spread
     * the load amongst the RBL servers more fairly.
     * <p>
     * TODO: For now, always returns null for non-IPv4 addresses.
     *
     * @param quick  if true, check just one of the BLs at random
     *
     * @return name of one BL server that blackholed address if so,
     *     null if address is OK
     *
     * @throws IllegalArgumentException  if given a null address
     */
    public static final String isAddrBlackholed(final Collection<String> extantRBLServers,
                                                final InetAddress addr,
                                                final boolean quick)
        throws IllegalArgumentException
        {
        if((addr == null) || (extantRBLServers == null))
            { throw new IllegalArgumentException(); }
        final byte rawAddr[] = addr.getAddress();
        if(rawAddr.length != 4)
            { return(null); }

        // Return null immediately if no servers available.
        final int nServers = extantRBLServers.size();
        if(nServers == 0) { return(null); }

        // Create the reverse dotted-quad address plus trailing dot.
        final String prefix = toStringReverseDottedQuad(addr);

        // If there is exactly one DNS BL server then check it immediately.
        if(nServers == 1) { return(_checkOneAddrInDNSBL(prefix, extantRBLServers.iterator().next())); }

        // The iterator for the RBL servers to test.
        final Iterator<String> serverIt;

        if(quick)
            {
            // Try just one of the RBL servers at random for speed.
            serverIt = Collections.singletonList(
                ((extantRBLServers instanceof List) ? (List<String>)extantRBLServers : new ArrayList<String>(extantRBLServers)).
                    get(Rnd.fastRnd.nextInt(nServers))).iterator();
            return(_checkOneAddrInDNSBL(prefix, serverIt.next()));
            }
        else if(RANDOMISE_RBL_LOOKUP_ORDER)
            {
            // Try all RBL servers, but in a random order.
            final List<String> l = new ArrayList<String>(extantRBLServers);
            Collections.shuffle(l, Rnd.fastRnd);
            serverIt = l.iterator();
            }
        else
            {
            // Try all RBL servers in the order given.
            serverIt = extantRBLServers.iterator();
            }

        // Check the DNS BL servers concurrently using our I/O-bound thread pool.
        final List<Future<String>> pending = new ArrayList<Future<String>>(nServers);
        while(serverIt.hasNext())
            {
            final String server = serverIt.next();
            ThreadUtils.nonCPUThreadPool.submit(new Callable<String>(){
                public final String call() throws Exception
                    { return(_checkOneAddrInDNSBL(prefix, server)); }
                });
            }

        // Collect the results.
        for(final Future<String> f : pending)
            {
            try
                {
                final String server = f.get();
                // As soon as we find a DNS BL that lists this client,
                // stop and return the DNS BL server name,
                // since the client is blacklisted!
                if(null != server) { return(server); }
                }
            catch(final Exception e)
                {
                // Log the error, but absorb/ignore it.
                e.printStackTrace();
                }
            }

        return(null); // No, this address is fine.
        }

    /**Look up one address in DNSBL.
     */
    private static String _checkOneAddrInDNSBL(final String prefix, final String server)
        {
        final String fqdn = prefix + server + '.'; // Make absolute.

        // If we can find an A record,
        // then this address is blackholed.
        return(hasARecordQuick(fqdn) ? server : null);
        }


    /**Immutable store of a non-null, non-zero-length unsigned-byte IP(v4) address prefix.
     */
    public static final class AddrPrefix implements Comparable<AddrPrefix>,
                                                    MemoryTools.Internable
        {
        /**Create an instance from a non-null, non-zero-length byte array.
         * Takes a copy of the prefix data.
         * @param addrPrefix  prefix of address; non-null
         */
        public AddrPrefix(final byte addrPrefix[])
            {
            if((addrPrefix == null) || (addrPrefix.length < 1))
                { throw new IllegalArgumentException(); }
            prefix = addrPrefix.clone();
            }

        /**Create an instance from the prefix of a non-null, non-zero-length byte array. */
        public AddrPrefix(final byte addrPrefix[], final int len)
            {
            if((addrPrefix == null) ||
                    (len < 1) || (len > addrPrefix.length))
                { throw new IllegalArgumentException(); }
            prefix = new byte[len];
            for(int i = len; --i >= 0; ) { prefix[i] = addrPrefix[i]; }
            }

        /**Create a shorter (but non-zero-length) instance from an existing prefix. */
        public AddrPrefix(final AddrPrefix ap, final int len)
            {
            if((ap == null) ||
                    (len < 1) || (len >= ap.prefix.length))
                { throw new IllegalArgumentException(); }
            prefix = new byte[len];
            for(int i = len; --i >= 0; ) { prefix[i] = ap.prefix[i]; }
            }

        /**Create an instance by parsing a padded dotted prefix format.
         * This is tolerant and does not insist that the octets
         * are padded with zeros on the left.
         * <p>
         * Can be performance-critical, especially at start-up,
         * so hand-crafted for speed and to avoid creating unnecessary objects.
         *
         * @param s  dotted prefix of form n(.n)* eg 127.0.0 or 10.0 or 10.000;
         *     there must be at least one octet,
         *     and each octet must be between 1 and 3 digits long
         *
         * @throws IllegalArgumentException  if the value is unparsable
         *     or not at least one octet
         */
        public AddrPrefix(final String s)
            throws IllegalArgumentException
            {
            if(s == null)
                { throw new IllegalArgumentException(); }
            final int sl = s.length();
            if(sl < 1)
                { throw new IllegalArgumentException(); }

            // Count the internal delimiter dots.
            // (There should not be dots in the first or last positions,
            // nor adjacent to one another.)
            int nDots = 0;
            for(int i = sl-1; --i > 0; )
                { if('.' == s.charAt(i)) { ++nDots; --i; } }
            final int prefixLength = 1 + nDots;
            final byte[] d = new byte[prefixLength];

            // Work backwards through the array
            // collecting one octet per loop.
            int index = prefixLength;
            for(int i = sl; --i >= 0; )
                {
                final char lastDigit = s.charAt(i--);
                if((lastDigit < '0') || (lastDigit > '9'))
                    { throw new IllegalArgumentException("last octet digit invalid: '" + lastDigit + "' at index " +(i+1) + " in \"" + s + "\" with prefixLength="+prefixLength); }
                int octet = lastDigit - '0';

                // We're done if we run out of input String
                // (and this must be the first octet).
                if(i < 0) { d[0] = (byte)octet; break; }
                final char digit2 = s.charAt(i);
                // We're done for this this octet if we hit a dot.
                if('.' == digit2) { d[--index] = (byte)octet; continue; }
                if((digit2 < '0') || (digit2 > '9'))
                    { throw new IllegalArgumentException("digit2 invalid"); }
                octet += 10 * (digit2 - '0');

                // We're done if we run out of input String
                // (and this must be the first octet).
                if(i <= 0) { d[0] = (byte)octet; break; }
                final char digit3 = s.charAt(--i);
                // We're done for this this octet if we hit a dot.
                if('.' == digit3) { d[--index] = (byte)octet; continue; }
                if((digit3 < '0') || (digit3 > '2'))
                    { throw new IllegalArgumentException("digit3 invalid"); }
                octet += 100 * (digit3 - '0');
                if(octet > 255)
                    { throw new IllegalArgumentException("octet > 255"); }
                d[--index] = (byte)octet;

                // Check for presence of delimiter dot if input not finished.
                if((i > 1) && ('.' != s.charAt(--i)))
                    { throw new IllegalArgumentException("missing ."); }
                }

            prefix = d;
            }

        /**Address prefix, non-zero-length, non-null. */
        private final byte prefix[];

        /**Get the prefix length, in bytes; strictly positive. */
        public int length()
            { return(prefix.length); }

        /**Returns true if the argument is a strict prefix of this value. */
        public boolean isStrictPrefix(final AddrPrefix other)
            {
            final int apLen = other.prefix.length;
            if(apLen >= prefix.length) { return(false); }
            for(int i = apLen; --i >= 0; )
                { if(other.prefix[i] != prefix[i]) { return(false); } }
            return(true);
            }

        /**Sorted as if in a dictionary with the bytes treated as unsigned. */
        public int compareTo(final AddrPrefix other)
            {
            final int minLen = Math.min(prefix.length, other.prefix.length);
            for(int i = 0; i < minLen; ++i)
                {
                final int diff = (prefix[i] & 0xff) - (other.prefix[i] & 0xff);
                if(diff != 0) { return(diff); }
                }

            // Equal in their common portion, so the shorter one sorts first.
            return(prefix.length - other.prefix.length);
            }

        /**Compute hash based on the assumption that most prefixes are 1--3 bytes long.
         * All bytes should be involved in the hash for it to be effective
         * since prefixes are likely to be dense in maps, etc.
         */
        public int hashCode()
            {
            final int length = prefix.length;
            // For CPU efficiency, and for the quality of the generated hash,
            // deal with the common short prefixes specially.
            // In each case, we assume that the last byte is the most "noisy".
            switch(length)
                {
                case 1: { return(prefix[0]); }
                case 2: { return(prefix[1] ^ ((131+prefix[0]) * 271)); }
                case 3: { return(prefix[2] + ((prefix[0]-129) * 82193) + (prefix[1] * 263)); }
                }
            // Default (general) case, using all prefix bytes.
            return(Arrays.hashCode(prefix));
            }

        /**Equal if the prefixes are the same length and have the same content. */
        public boolean equals(final Object obj)
            {
            if(this == obj) { return(true); }
            if(!(obj instanceof AddrPrefix)) { return(false); }
            final AddrPrefix other = ((AddrPrefix)obj);
            final int length = prefix.length;
            if(length != other.prefix.length) { return(false); }
            for(int i = length; --i >= 0; )
                { if(prefix[i] != other.prefix[i]) { return(false); } }
            return(true); // Identical.
            }

        /**Get a copy of the byte[] prefix data; never null. */
        public byte[] toByteArray()
            { return(prefix.clone()); }

        /**Print as padded dotted prefix; never null. */
        public String toPaddedDottedPrefix()
            {
            final StringBuilder sb = new StringBuilder(prefix.length * 4);
            final StringBuilder octet = new StringBuilder(3);
            for(int i = 0; i < prefix.length; ++i)
                {
                octet.setLength(0);
                octet.append(prefix[i] & 0xff);
                while(octet.length() < 3) { octet.insert(0, '0'); }
                if(sb.length() > 0) { sb.append('.'); }
                sb.append(octet);
                }
            return(sb.toString());
            }

        /**Human-readable representation; currently equivalent to toPaddedDottedPrefix(); never null. */
        public String toString()
            { return(toPaddedDottedPrefix()); }
        }
    }
