/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.*;
import java.util.*;

/**Immutable set of names and ExhibitStaticAttr for all exhibits.
 * This provides a set of all exhibits, mapped from the full name
 * to the static/immutable attributes for an exhibit (ie the parts
 * that will not change unless the exhibit itself is changed).
 * <p>
 * Also provides a mapping from the final file component of an exhibit to its
 * full name within the collection.  All exhibits should be unique in this
 * last component; if not then only the alphabetically first full-name value
 * will be kept.
 * <p>
 * Also contains the time of last change to exhibit set when this
 * snapshot was taken; this is verified simply to be non-negative
 * and may for example be some combination of timestamp and hash over
 * the exhibit content to best catch changes in the exhibit-set
 * composition or it may be a pure hash.  If a combination then
 * usually the most significant bits would be those of the newest
 * exhibit and the least significant would be the hash.  If the
 * exhibit set changes then this value should change, however
 * the timestamp may not monotonically increase as new exhibits
 * are added (or existing ones are altered).
 * <p>
 * Designed to be efficient `on-the-wire'; for serialisation
 * we send a count and the ExhibitStaticAttr objects in sorted
 * order by name instead of the map, which should be much less bulky than
 * our internal representation (which is designed for lookup speed),
 * and contains all the required information.  On deserialising
 * we simply reconstruct the map.
 * We still let default serialisation deal with any other fields
 * (the timestamp).
 * <p>
 * (It looks like HashMap's serialisation must be quite good since the
 * saving in uncompressed size is only about 6%.)
 * <p>
 * The hashCode() is based on a hash of the timestamp,
 * and equality is based on the sets of exhibits
 * and global and per-exhibit attributes being equal.
 */
public final class AllExhibitImmutableData implements Serializable,
                                                      ObjectInputValidation
    {
    /**Construct an empty, zero-timestamp snapshot.
     */
    public AllExhibitImmutableData()
        { this(new TreeMap<String,ExhibitStaticAttr>(), 0); }

    /**Construct a new snapshot.
     * The Map must consist purely of mappings from valid String
     * exhibit names to ExhibitStaticAttr items; the Map is copied
     * and this is checked.
     * <p>
     * The timestamp must be zero if the map is empty, else it
     * must be strictly positive.
     *
     * @throws IllegalArgumentException  if the timestamp or Map are
     *     found to be malformed.
     */
    public AllExhibitImmutableData(final Map<String,ExhibitStaticAttr> nameToStaticAttr,
                                   final long lastExhibitChangeTimestamp)
        throws IllegalArgumentException
        {
        if((nameToStaticAttr == null) ||
           (lastExhibitChangeTimestamp < 0))
            { throw new IllegalArgumentException(); }
        m.putAll(nameToStaticAttr);
        timestamp = lastExhibitChangeTimestamp;
        length = m.size();

        // Validate map contents.
        for(final Iterator it = m.keySet().iterator(); it.hasNext(); )
            {
            final Object n = it.next();
            if(!(n instanceof String))
                { throw new IllegalArgumentException("non-String key in map"); }
            final String name = (String) n;
            final Object a = m.get(name);
            if(!(a instanceof ExhibitStaticAttr))
                { throw new IllegalArgumentException("non-ExhibitStaticAttr value in map"); }
            final ExhibitStaticAttr sa = (ExhibitStaticAttr) a;
            if(!ExhibitName.validNameSyntax(name) ||
               !name.equals(sa.filePath))
                { throw new IllegalArgumentException("invalid name or mismatched key"); }
            }

        // Build our short-to-full name map now if eager...
        _buildShortToFull();

        // Verify what we've been given.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Timestamp of last exhibit update when this snapshot was taken.
     * Strictly positive unless map is empty, in which case it will be zero.
     * <p>
     * This information is accessable without holding any locks.
     */
    public final long timestamp;

    /**Exhibit count, with lock-free access; non-negative.
     * The information in this is redundant
     * (it is available in m.size() for example),
     * but is small and useful for speed and an integrity check, etc.
     * <p>
     * This emulates the length field of an array while size()
     * emulates the behaviour of collections...
     * <p>
     * This information is accessable without holding any locks.
     */
    public final int length;

    /**Map from name to static attr; never null once construction/deserialisation is complete.
     * Never altered once installed so is thread-safe.
     * <p>
     * Custom serialisation means that this map has to be mutable,
     * so cannot be given out to a caller as-is.
     */
    private transient /* final */ Map<String,ExhibitStaticAttr> m = new HashMap<String, ExhibitStaticAttr>(1001);

    /**Immutable Map from final file component (ie short name) to full name; never null. */
    private transient /* final */ Map<String,String> shortToFull;

    /**List of duplicates found by _buildShortToFull; never null. */
    private transient /* final */ String[] shortNameDups;

    /**Builds shortToFull (and shortNameDups) from m if required.
     * We ensure that at least the short names are intern()ed to avoid
     * duplication in memory.
     */
    private void _buildShortToFull()
        {
        // Working map from short to full names.
        final Map<String,String> sToF = new HashMap<String,String>(m.size() * 2 + 1);

        // Set of full names that contain duplicate short (file) components.
        final SortedSet<String> dups = new TreeSet<String>();

        for(final Iterator<String> it = m.keySet().iterator(); it.hasNext(); )
            {
            final String fullName = it.next();
            // We avoid wasting memory with redundant copies of the short names,
            // as we pass these around lots, and they can hang around a while.
            final String shortName = MemoryTools.intern(
                ExhibitName.getFileComponent(getPooledExhibitName(fullName)));

            final String dup = sToF.get(shortName);
            if(dup == null)
                {
                // No duplicate; the simple case!
                sToF.put(shortName, fullName);
                continue;
                }

            // OK, we've got to deal with a duplicate...

            // Put the dup and the new value in the dups list.
            dups.add(dup);
            dups.add(fullName);

            // If the new full name is lexically before the old one,
            // replace the mapping to keep the lexically-earlier full name.
            if(dup.compareTo(fullName) > 0)
                { sToF.put(shortName, fullName); }
            }

        // Store (sorted) list of duplicates.
        final String sND[] = new String[dups.size()];
        dups.toArray(sND);
        shortNameDups = sND;

        // Atomically store mapping when everything else is complete.
        shortToFull = Collections.unmodifiableMap(sToF);
        }

    /**Gets (sorted) list of full names that have duplicate short names; zero-length if none.
     * This returns a copy that can be safely tampered with by the caller.
     */
    public String[] getFullNamesWithDuplicateShortNames()
        {
//        if(!EAGER_BUILDSHORTTOFULL && (null == shortNameDups)) { _buildShortToFull(); }
        return(shortNameDups.clone());
        }

    /**Gets a Set of all short exhibit names.
     */
    public Set<String> getAllExhibitShortNamesSet()
        {
//        if(!EAGER_BUILDSHORTTOFULL && (null == shortToFull)) { _buildShortToFull(); }
        return(shortToFull.keySet());
        }

    /**Gets the full name of an exhibit from its (unique) file component; null if no exhibit has the specified short name.
     */
    public String getFullName(final String shortName)
        {
//        if(!EAGER_BUILDSHORTTOFULL && (null == shortToFull)) { _buildShortToFull(); }
        return(shortToFull.get(shortName));
        }

    /**Gets the immutable Map from short to full name; null if no exhibit has the short name.
     */
    public Map<String,String> getShortToFullMap()
        {
//        if(!EAGER_BUILDSHORTTOFULL && (null == shortToFull)) { _buildShortToFull(); }
        return(shortToFull);
        }

    /**Get exhibit attributes; if result non-null the exhibit is valid.
     */
    public ExhibitStaticAttr getStaticAttr(final String name)
        {
        return(m.get(name));
        }

    /**Get exhibit count; 0 if no exhibits.
     */
    public int size()
        {
        return(length);
        }

    /**Gets a Set of all (String) raw (full) exhibit names.
     * The result is immutable.
     * <p>
     * Should be relatively efficient.
     */
    public Set<String> getAllExhibitNamesSet()
        {
        return(Collections.unmodifiableSet(m.keySet()));
        }
    
    /**Cache for getAllExhibitNamesSorted().
     * Initially null, but never non-null once set.
     * <p>
     * Marked volatile for lock-free thread-safe access.
     * <p>
     * Marked transient as this is recomputable redundant state.
     */
    private transient volatile String[] _cache_getAllExhibitNamesSorted;

    /**Gets an array of all raw exhibit names; sorted by raw name.
     * This may be slow, as it may have to do the sort in place.
     * <p>
     * This returns a private array.
     */
    public String[] getAllExhibitNamesSorted()
        {
        String result[] = _cache_getAllExhibitNamesSorted;
        if(null == result)
            {
            // Compute and cache result.
            final Set<String> s = m.keySet(); // We assume that this is unsorted.
            result = new String[s.size()];
            s.toArray(result);
            Arrays.sort(result);
            _cache_getAllExhibitNamesSorted = result;
            }
        // Return a private copy to avoid tampering with our value.
        return(result.clone());
        }

    /**Gets Set of all static attrs unsorted.
     * Could be very large and slow to get; use with caution.
     */
    public Set<ExhibitStaticAttr> getAllStaticAttrs()
        {
        return(new HashSet<ExhibitStaticAttr>(m.values()));
        }

    /**Gets SortedSet of all static attrs; sorted by raw name.
     * Could be very large and slow to get; use with caution.
     */
    public SortedSet<ExhibitStaticAttr> getAllStaticAttrsSorted()
        {
        return(new TreeSet<ExhibitStaticAttr>(m.values()));
        }

    /**Canonicalise exhibit name String.
     * This looks up a full String name in the list of exhibits.
     * If not found it returns the original String.
     * If found it returns the copy of the String in the exhibit.
     */
    public String getPooledExhibitName(final String name)
        {
        final ExhibitStaticAttr esa = m.get(name);
        if(esa == null) { return(name); } // Can't pool it.
        return(esa.filePath); // Return pooled copy.
        }


    /**Conservatively estimates filespace bytes required to store all exhibits.
     * This makes some assumptions about filespace overhead and so on.
     * <p>
     * This is for the specific named author, unless author is null in which
     * case it for all authors.
     * <p>
     * This separately rounds up each exhibit size and name size to an
     * assume filesystem block size as per FileTools.roundUpToFSBlockSize().
     *
     * @throws IllegalArgumentException  if author is not null and is not a
     *     syntactically-valid author name (initials)
     */
    public final long computeFileSpaceBytes(final String author)
        {
        if((author != null) &&
           !ExhibitName.validAuthorSyntax(author))
            { throw new IllegalArgumentException(); }

        long result = 0;
        for(final Iterator<ExhibitStaticAttr> it = m.values().iterator(); it.hasNext(); )
            {
            final ExhibitStaticAttr esa = it.next();

            // Skip this if only adding code for specific author other than
            // this one.
            if((author != null) &&
               !author.equals(ExhibitName.getAuthorComponent(esa.filePath)))
                { continue; } // Not interested in this exhibit.

            result += FileTools.roundUpToFSBlockSize(esa.filePath.length());
            result += FileTools.roundUpToFSBlockSize(esa.length);
            }

        return(result);
        }


    /**Returns a hash code value for the object derived from the timestamp.
     *
     * @return  a hash code value for this object.
     * @see     Object#equals(Object)
     * @see     Hashtable
     */
    public int hashCode()
        {
        return(((int) (timestamp >>> 32)) ^ (int) timestamp);
        }

    /**Returns true when the underlying set of exhibits and their attributes is the same.
     *
     * @param   obj   the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     * @see     Boolean#hashCode()
     * @see     Hashtable
     */
    public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof AllExhibitImmutableData)) { return(false); }
        final AllExhibitImmutableData other = (AllExhibitImmutableData) obj;
        return(m.equals(other.m));
        }


    /**Write out a less-redundant (and more compressable) form of our internal information.
     */
    private void writeObject(final ObjectOutputStream oos)
        throws IOException
        {
        // Write the fields that we are not trying to optimise.
        // Note that this includes our length field.
        oos.defaultWriteObject();

        // Send the attr data as a count and then the static attr objects
        // in sorted order to enhance compression
        // (by improving locality, with similar items close to one another).
        final int size = m.size();
        final ExhibitStaticAttr outValues[] = new ExhibitStaticAttr[size];
        m.values().toArray(outValues);
        Arrays.sort(outValues);
        // oos.writeInt(size); // This is in our length field...
        for(int i = size; --i >= 0; )
            { oos.writeObject(outValues[i]); }
        }

    /**Deserialise.
     */
    private void readObject(final ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        // Read the fields that we are not trying to optimise.
        // This includes our length field...
        ois.defaultReadObject();

        // Read the map as a count followed by that many static attr objects
        // which we post into a new Map.
        // We replace the old map atomically.
        final int newSize = length;
        // Create a new, appropriately-sized map for fast load and access.
        final Map<String,ExhibitStaticAttr> newMap = new HashMap<String,ExhibitStaticAttr>(2 * newSize + 1);
        // Add the objects to the map, indexed by their own keys.
        for(int i = newSize; --i >= 0; )
            {
            final ExhibitStaticAttr esa = (ExhibitStaticAttr) ois.readObject();
            newMap.put(esa.filePath, esa);
            }
        assert(newMap.size() == length) : "found duplicate name(s) in serial data";
        // Install the new map atomically.
        m = newMap;

        // Build the short-to-full-name map if eager.
        _buildShortToFull();

        validateObject();
        }

//    /**Deserialise: use constructor for validation, defensive copying, etc. */
//    protected Object readResolve()
//        // throws ObjectStreamException
//        {
//        // Construct new instance of object in normal defensive way.
//        // This also ensures that we get a chance to intern() stuff, etc,
//        // and reduce old-heap churn in generational GC systems.
//        return(new AllExhibitImmutableData(newMap, timestamp));
//        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        if(timestamp < 0)
            { throw new InvalidObjectException("bad object: timestamp < 0"); }
        if(length < 0)
            { throw new InvalidObjectException("bad object: length < 0"); }
        if((length > 0) && (timestamp <= 0))
            { throw new InvalidObjectException("bad object: timestamp must be positive if there are exhibits"); }
        if((length == 0) && (timestamp != 0))
            { throw new InvalidObjectException("bad object: timestamp must be zero if no exhibits"); }
        // The two size measures must match.
        if(length != size())
            { throw new InvalidObjectException("bad object: length != size()"); }
        }

    /**Our serial version... */
    private static final long serialVersionUID = 1227490240470952655L;
    }
