/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.ai.scorer.ScorerCacheIF;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.BasicVarMgrInterface;
import org.hd.d.pg2k.svrCore.vars.SimpleVariablePipelineIF;

import java.io.*;
import java.nio.ByteBuffer;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**Set of all exhibit names and properties.
 * This includes AllExhibitImmutableData and all other loadable and
 * calculatable exhibit properties, including emergent statistical
 * properties computed across the entire exhibit set.
 * <p>
 * This has a hash over all the loaded/static properties and the
 * exhibit set so that changes should be easily caught.
 * Computed state is excluded since it is computed on demand,
 * and adds no extra information or includes continigent information and noise
 * such as hints from system settings.
 * <p>
 * Designed to be efficient `on-the-wire' and in memory.
 * <p>
 * This object is logically immutable except that:
 * <ul>
 * <li>Some information is computed on demand and may be discarded if
 *     out-of-date or in low-memory conditions or for serialisation.
 * <li>The not-yet-computed state of some expensive-to-compute items
 *     is made visible for more efficient use elsewhere in the system
 *     and this state may change according to the previous point.
 * <li>Some expensive computable information may be retained across
 *     serialiation having been computed in a different environment
 *     to the current one.
 * </ul>
 * <p>
 * This holds per-exhibit information such as timestamp, x/y dimensions, etc,
 * and whole-collection data such as hot/cold lists, location maps, etc.
 * <p>
 * The hashCode() and equals() methods are implemented,
 * based on the underlying data.
 */
public final class AllExhibitProperties implements Serializable, ObjectInputValidation,
                                                   MemoryTools.Compactable
    {
    /**Construct a default (empty) AllExhibitProperties object with zero timestamp.
     * This is operation is fast, and produces a small object,
     * with no data but entirely consistent.
     * <p>
     * The hash of this object is 0, as is the timestamp.
     */
    public AllExhibitProperties()
        {
        this(null,
             new ExhibitPropsGlobalImmutable(),
             new AllExhibitImmutableData(),
             new TreeMap<String,ExhibitPropsLoadable>(),
             new HashMap<String,ExhibitPropsComputable>(),
             0);
        }

    /**Construct a new AllExhibitProperties object from scratch, checking for consistency and validity.
     */
    public AllExhibitProperties(final ExhibitPropsGlobalImmutable _epgi,
                                final AllExhibitImmutableData _aeid,
                                final Map<String, ExhibitPropsLoadable> _loadedProps,
                                final Map<String, ExhibitPropsComputable> _computedProps)
        {
        this(null, _epgi, _aeid, _loadedProps, _computedProps, 0);
        }
    
    /**Canonicalise AEP, ie reconstruct new instance from old data; null only if input is.
     * Can be useful for recomputing hashes, etc,
     * from AEP instances created and serialised before an minor implementation change.
     */
    public static AllExhibitProperties canonicalise(final AllExhibitProperties in)
        {
        if(in == null) { return(null); }
        final AllExhibitProperties result = new AllExhibitProperties(in, in.epgi, in.aeid, in.getExhibitPropsLoadableMap(), in.getExhibitPropsComputableMap(), in.hashNotChangedSince);
        assert(in.equals(result)); // This operation should preserve equality!
        return(result);
        }

    /**Build a new AllExhibitProperties object, preserving any computed state possible from the old one.
     * Because computing some information is very expensive,
     * this can save significant time.
     * <p>
     * If old is null, the object is built from scratch.
     * @param _aeid immutable core exhibit data
     * @param _loadedProps  map from exhibit name (String) to loaded properties
     *     (ExhibitPropsLoadable);
     *     not all exhibits need have a mapping
     *     (in which case they are assumed no have no properties)
     * @param unchangedSince  if non-zero, the hashNotChangedSince value is set from this
     */
    public AllExhibitProperties(final AllExhibitProperties old,
                                final ExhibitPropsGlobalImmutable _epgi,
                                final AllExhibitImmutableData _aeid,
                                final Map<String, ExhibitPropsLoadable> _loadedProps,
                                final Map<String, ExhibitPropsComputable> _computedProps,
                                final long unchangedSince)
        {
        // Capture ExhibitPropsGlobalImmutable.
        epgi = _epgi;

//        // Capture LocationMap.
//        lm = _lm;

        // Capture AllExhibitImmutableData.
        aeid = _aeid;

        final boolean isEmpty = (aeid.size() == 0); // No exhibits.

        // Set our construction timestamp; zero if no exhibits.
        constructedAt = isEmpty ? 0L : System.currentTimeMillis();

        // Make our loadedProps value.
        // Is read-only so does not need to be synchronized.
        final Map<String,ExhibitPropsLoadable> lP = new HashMap<String,ExhibitPropsLoadable>(_loadedProps.size() * 2 + 1);
        // Now work through our input data...
        for(final Iterator<String> it = _loadedProps.keySet().iterator(); it.hasNext(); )
            {
            final String key = aeid.getPooledExhibitName(it.next());
            final ExhibitPropsLoadable epl = _loadedProps.get(key);
            if((epl != null) && !epl.equals(ExhibitPropsLoadable.EMPTY))
                { lP.put(key, epl); }
            }
        loadedProps = lP; // Assign atomically.
//        _optimiseLP(); // Optimise memory footprint.

        // Copy in (valid) computed properties.
        // Our new map will be logically immutable
        // so an ordinary HashMap is fine and allows concurrency.
        computedProps = new HashMap<String, ExhibitPropsComputable>(2*_computedProps.size());
        if(!isEmpty)
            {
            for(final Iterator<String> it = _computedProps.keySet().iterator(); it.hasNext(); )
                {
                final String key = it.next();
                final ExhibitStaticAttr newesa = aeid.getStaticAttr(key);
                // If named exhibit does not exist, abort.
                if(newesa == null) { throw new IllegalArgumentException("EPC value specified for non-existant exhibit"); }

                final ExhibitPropsComputable epc = _computedProps.get(key);
                // Don't store null/EMPTY instances.
                if((null == epc) || epc.equals(ExhibitPropsComputable.EMPTY)) { continue; }
                // Seems OK; use the intern()ed newesa filePath as the key for memory efficiency.
                computedProps.put(newesa.filePath, epc);
                }

            // If we have mutable properties,
            // then copy across the entries corresponding to exhibits in this instance.
            // We remove entries corresponding to non-existant exhibits.
            // Even stale entries are copied since they may still be pretty accurate
            // and expensive to recompute, though trivially-stale values are discarded.
            // Note that it is safe to share entries with an old instance
            // since they are immutable.
            // We drop entries where the exhibit itself appears to have changed,
            // but we don't test things such as description text and other aux data.
            if((old != null) && (old.epcmMap != null))
                {
                final Map<String, ExhibitPropsComputableMutable> tmpMap =
                        new HashMap<String, ExhibitPropsComputableMutable>(old.epcmMap);
                epcmMap.clear(); // Ensure that we have no deserialised weird state!
                for(final Iterator<String> it = tmpMap.keySet().iterator(); it.hasNext(); )
                    {
                    final String key = it.next();
                    final ExhibitStaticAttr newesa = aeid.getStaticAttr(key);
                    // If exhibit no longer exists or has changed, remove this, don't keep it.
                    if((newesa == null) || (!newesa.isIdentical(old.aeid.getStaticAttr(key))))
                        {
//                        it.remove();
                        continue;
                        }

                    // Remove null entries and trivially-stale values
                    // that we can quickly recompute if required.
                    final ExhibitPropsComputableMutable epcm = tmpMap.get(key);
                    if((epcm == null) || (epcm.isTriviallyStale()))
                        {
//                        it.remove();
                        continue;
                        }

                    // Seems OK; use the newesa filePath as the key
                    // for memory efficiency (ie as the canonical key).
                    epcmMap.put(newesa.filePath, epcm);
                    }
                }
            }

        // Compute our long hash.
        // This should cover all the elements that equals() does.
        if(isEmpty) // Guaranteed zero (and fast to compute!) if no exhibits.
            { longHash = 0; }
        else
            {
            try {
                // Start a message digest (SHA1).
                final MessageDigest md = GenUtils.getStandardDigest();
                final DataOutputStream dos =
                    new DataOutputStream(
                        new DigestOutputStream(
                            (new OutputStream(){ // Null output stream...
                                public final void write(int b) { }
                                public final void write(byte[] b, int off, int len) { }
                                }),
                            md));

                // Get the full, sorted list of aeid keys.
                final String[] allKeys = aeid.getAllExhibitNamesSorted();

                // Put in the names and the loadable/computable property hashes,
                // in (reverse) sorted key order.
                for(int i = allKeys.length; --i >= 0; )
                    {
                    final String name = allKeys[i];
                    dos.writeUTF(name); // Capture the full names...
                    final ExhibitPropsLoadable epl = getExhibitPropsLoadable(name);
                    dos.writeInt(epl.hashCode());
                    // We include the EPC hash explictly,
                    // even though EPC values are nominally derived values,
                    // because their representation (eg format) may change/improve.
                    final ExhibitPropsComputable epc = getExhibitPropsComputable(name);
                    if(null != epc) { dos.writeInt(epc.hashCode()); }
                    else { dos.writeByte(0); }
                    }

                // Contribute hashes from embedded global properties.
                dos.writeLong(epgi.longHash);

                // Fold in the aeid contribution,
                // trusting the timestamp as representative.
                dos.writeLong(aeid.timestamp);

                // Create the basic hash from the first 8 digest bytes.
                dos.flush();
                final byte[] digest = md.digest();
                long lHash = 0xDeadBeefDeadBeefL;
                for(int i = 8; --i >= 0; )
                    { lHash ^= (((long)(digest[i] & 0xff)) << (8*i)); }

                // We guarantee that the hash is strictly positive
                // if there are exhibits
                // (and that the hash is zero if there are no exhibits).
                lHash = (lHash >>> 1) ^ (lHash & 1);
                longHash = (lHash == 0) ? 1 : lHash;
                }
            catch(final IOException e) { throw new Error("internal error"); } // Should never happen.
            }

        // See if the hash has changed and set a timestamp if so.
        // If an explict not-changed-since timestamp has been supplied
        // then use the supplied explicit timestamp.
        // (This is to support explicit reconstruction of AEP state, eg from a diff.)
        // Else if there is no old object or the old object's hash is different,
        // then use the constructedAt time.
        // Else carry forward the old hashNotChangedSince time,
        // but force it to be no older than the aeid.timestamp
        // as a belt-and-braces measure in case the hash misses something.
        if(unchangedSince != 0)
            { hashNotChangedSince = unchangedSince; }
        else if((old == null) || (old.longHash != longHash))
            { hashNotChangedSince = constructedAt; }
        else
            { hashNotChangedSince = Math.max(old.hashNotChangedSince, aeid.timestamp); }

        // Verify object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**The immutable global exhibit properties; never null.
     * This is made available as a field for maximum performance.
     */
    public final ExhibitPropsGlobalImmutable epgi;

    /**The immutable exhibit data; never null.
     * This is made available as a field for maximum performance.
     */
    public final AllExhibitImmutableData aeid;

    /**Logically immutable map from exhibit name (String) to loadable props (ExhibitPropsLoadable); never null.
     * If an entry for an exhibit is null then we return ExhibitPropsLoadable.EMPTY.
     * <p>
     * The keys here are canonicalised against aeid so that we don't
     * have duplicate String values knocking about.
     */
    private final Map<String,ExhibitPropsLoadable> loadedProps;

    /**Get loadable props for exhibit; never null.
     * This is completely thread-safe as loadedProps is read-only/immutable
     * once construction/deserialisation is complete.
     *
     * @return non-null instance; ExhibitPropsLoadable.EMPTY if none
     */
    public ExhibitPropsLoadable getExhibitPropsLoadable(final String exhibitName)
        {
        final ExhibitPropsLoadable result = loadedProps.get(exhibitName);
        if(result == null) { return(ExhibitPropsLoadable.EMPTY); } // Never return null.
        return(result);
        }

    /**Get immutable map from full exhibit name to ExhibitPropsLoadable; never null.
     * May omit entries that map to ExhibitPropsLoadable.EMPTY.
     */
    public Map<String,ExhibitPropsLoadable> getExhibitPropsLoadableMap()
        { return(Collections.unmodifiableMap(loadedProps)); }

    /**Class that provides data source to get at exhibit data semi-random access or as stream.
     * An implementor must override getRawFile() and may also override
     * getInputStream() with a markable stream.
     * <p>
     * This does not hold a lock while calling back to the underlying data source
     * in order to help avoid deadlocks.
     * <p>
     * The state of an InputStream returned by this is undefined after
     * an IOException from the stream.
     */
    public static abstract class ExhibitDataSource
        {
        /**Check if the given exhibit is fully loaded; true if so.
         * This means that the whole exhibit is quickly available,
         * eg an a local file system, for reading.
         * <p>
         * This is only a hint; an immediately following call
         * to read the exhibit might find the exhibit no longer local.
         * <p>
         * The result is undefined for a non-existent exhibit.
         */
        public abstract boolean isExhibitFullyLoaded(ExhibitStaticAttr esa)
            throws IOException;

        /**Read a chunk of the raw exhibit binary into the given buffer.
         * As for SimpleExhibitPipelineIF.getRawFile().
         */
        public abstract void getRawFile(ByteBuffer buf, String exhibitName, int position)
            throws IOException;

        /**Get a the raw exhibit binary as a (markable) stream to read sequentially.
         * The default implementation uses getRawFile()
         * and does no buffering.
         * <p>
         * The returned stream is thread-safe, but may not behave as expected
         * is used by more than one thread at once,
         * ie concurrent reads are not supported.
         */
        public InputStream getInputStream(final ExhibitStaticAttr esa)
            {
            final InputStream result = new InputStream(){
                /**Current file position (0 at start, esa.length at EOF); non-negative. */
                private int pos;

                /**Return exact number of bytes remaining, ie available. */
                public final synchronized int available()
                    { return((int) Math.min(Integer.MAX_VALUE, esa.length - pos)); }

                /**Read a single byte; highly inefficient but functionality must be supplied.
                 * This <strong>does not</strong> hold a lock during the call to getRawFile().
                 */
                public final int read()
                    throws IOException
                    {
                    final byte data[] = new byte[1];
                    final int n = read(data, 0, 1);
                    if(n == -1) { return(-1); } // EOF.
                    assert(n == 1); // We block to read at least one byte...
                    return(data[0] & 0xff);
                    }

                /**Read a block of bytes directly from the raw data source.
                 * Always does a full read unless at EOF when remaining
                 * bytes will be returned.
                 * <p>
                 * This is thread-safe, but may not behave as expected
                 * is used by more than one thread at once,
                 * ie concurrent reads are not supported.
                 * <p>
                 * This <strong>does not</strong> hold a lock during the call to getRawFile().
                 * <p>
                 * After an IOException the stream is left positioned at EOF.
                 */
                public final int read(final byte buf[], final int off, int len)
                    throws IOException
                    {
                    final int avail = available();
                    if(avail <= 0) { return(-1); } // EOF.
                    if(len > avail) { len = avail; } // Limit to remaining data.
                    if(len < 1) { return(0); }

                    try
                        {
                        final ByteBuffer byteBuffer = ByteBuffer.wrap(buf, off, len);
                        assert(byteBuffer.position() == off);
                        assert(byteBuffer.remaining() == len);
                        getRawFile(byteBuffer, esa.filePath, pos);
                        final int bytesRead = len - byteBuffer.remaining();
                        assert(bytesRead >= 0);

                        // Update file position (thread-) safely.
                        if(bytesRead > 0)
                            {
                            synchronized(this)
                                {
                                if((pos += bytesRead) > esa.length)
                                    { throw new IOException("concurrent reads took position beyond EOF"); }
                                }
                            }
                        }
                    catch(final IOException e)
                        {
                        // In case of error move position to EOF and rethrow.
                        pos = (int) esa.length;
                        throw e;
                        }

                    return(len);
                    }

                /**Marked position (ie store for pos when mark() is called); [0, esa.length]. */
                private int markedPos;

                /**Marks the current position in this input stream.
                 * A subsequent call to
                 * the <code>reset</code> method repositions this stream at the last marked
                 * position so that subsequent reads re-read the same bytes.
                 * <p/>
                 * <p> The <code>readlimit</code> arguments tells this input stream to
                 * allow that many bytes to be read before the mark position gets
                 * invalidated.
                 * <p/>
                 * <p>This implementation records the current stream position.</p>
                 *
                 * @param readlimit the maximum limit of bytes that can be read before
                 *                  the mark position becomes invalid.
                 * @see java.io.InputStream#reset()
                 */
                public synchronized void mark(final int readlimit)
                    {
                    // Notes the current position.
                    markedPos = pos;
                    }

                /**Tests if this input stream supports the <code>mark</code> and <code>reset</code> methods.
                 * Always true for this implementation.
                 * @return <code>true</code>
                 * @see java.io.InputStream#mark(int)
                 * @see java.io.InputStream#reset()
                 */
                public boolean markSupported()
                    { return(true); }

                /**Repositions this stream to the position at the time the <code>mark</code> method was last called on this input stream.
                 * <p>This implementation always restores the stream to the marked position
                 * and never needs to throw an IOException.</p>
                 *
                 * @see java.io.InputStream#mark(int)
                 * @see java.io.IOException
                 */
                public synchronized void reset()
                    // throws IOException
                    {
                    pos = markedPos;
                    }

                /**Skips over and discards <code>n</code> bytes of data from this input stream.
                 * The <code>skip</code> method may, for a variety of reasons, end
                 * up skipping over some smaller number of bytes, possibly <code>0</code>.
                 * This may result from any of a number of conditions; reaching end of file
                 * before <code>n</code> bytes have been skipped is only one possibility.
                 * The actual number of bytes skipped is returned.  If <code>n</code> is
                 * negative, no bytes are skipped.
                 * <p/>
                 * <p>This implementation never needs to throw an IOException.</p>
                 *
                 * @param n the number of bytes to be skipped
                 * @return the actual number of bytes skipped
                 */
                public synchronized long skip(final long n)
                    // throws IOException
                    {
                    if((n < 0) || (n >= Integer.MAX_VALUE))
                        { throw new IllegalArgumentException(); }

                    // Compute number of bytes before EOF.
                    final int av = available();

                    // If the skip request does not take us beyond EOF
                    // then do it as requested.
                    if(n <= av)
                        {
                        pos += n;
                        return(n);
                        }

                    // Else skip to EOF and return the number of bytes skipped.
                    pos += av;
                    return(av);
                    }
                };
            return(result);
            }
        }

    /**Map from exhibit name (String) to ExhibitPropsComputable properties; never null.
     * This is a Hashtable to be completely thread-safe, even during
     * serialisation where Hashtable.writeObject() is synchronized.
     * <p>
     * An instance of ExhibitPropsComputable is created on first demand
     * and posted to the map.  No lock is held while the computation
     * is done, so redundant computations are possible, but a high
     * degree of concurrency is available in return.
     * <p>
     * Entries are never removed from this map,
     * and null is never posted as a value.
     * <p>
     * Defensively copied during deserialisation.
     */
    private final Map<String,ExhibitPropsComputable> computedProps;

    /**Get computable (immutable) props for exhibit; if none available then returns null.
     */
    public ExhibitPropsComputable getExhibitPropsComputable(final String exhibitName)
        { return(computedProps.get(exhibitName)); }

    /**Get immutable map from full exhibit name to ExhibitPropsComputable; never null.
     */
    public Map<String,ExhibitPropsComputable> getExhibitPropsComputableMap()
        { return(Collections.unmodifiableMap(computedProps)); }

    /**The hash of all the data held; guaranteed non-negative.
     * Depends on the timestamp of all the exhibit (aeid) and
     * other loadable information held (eg map data).
     * <p>
     * Is guaranteed to be zero if the exhibit set is empty,
     * and strictly positive otherwise.
     * <p>
     * Computable data (and hashNotChangedSince) is excluded from the hash
     * since it is derived from other data that is included in the hash.
     */
    public final long longHash;

    /**The time that this object was constructed.
     * Strictly positive unless there are no exhibits,
     * in which case this is zero.
     * <p>
     * We keep this private.
     */
    private final long constructedAt;

    /**Time since which we claim that the exhibit set/data has not changed; guaranteed non-negative.
     * If a new AllExhibitProperties object is constructed
     * with the aid of an old one, and the longHash has not changed
     * between the two, then this is copied from the old
     * object (except that as a belt and braces measure it is
     * ensured to be no older than the aeid.timestamp),
     * else it is the same as the constructedAt time.
     * <p>
     * If a new AllExhibitProperties object is constructed from
     * scratch each time then this will be the time that the
     * object was constructed, and so we can regard this as a
     * an `exhibits have not changed since' time.
     * <p>
     * Strictly positive unless there are no exhibits,
     * in which case this is zero (to prompt early recomputation!).
     */
    public final long hashNotChangedSince;


//    /**Optimise loadedProps by canonicalising keys and eliminating duplicate/EMPTY values.
//     * Must only be called during construction or deserialisation
//     * when no other access to loadedProps is possible.
//     */
//    private synchronized void _optimiseLP()
//        {
//        final HashMap<ExhibitPropsLoadable,ExhibitPropsLoadable> eplDup = new HashMap<ExhibitPropsLoadable, ExhibitPropsLoadable>(loadedProps.size() * 2 + 1);
//        for(final Iterator<String> it = (new HashSet<String>(loadedProps.keySet())).iterator(); it.hasNext(); )
//            {
//            // Pool exhibit name.
//            final String key = aeid.getPooledExhibitName(it.next());
//            // Zap any entry not corresponding to a valid exhibit.
//            if(aeid.getStaticAttr(key) == null)
//                { loadedProps.remove(key); }
//            // Examine the value.
//            final ExhibitPropsLoadable epl = loadedProps.get(key);
//            // Zap anything equivalent to EMPTY (or null).
//            if((epl == null) || epl.equals(ExhibitPropsLoadable.EMPTY))
//                { loadedProps.remove(key); }
//            // If this is value is equal to an extant one, use the extant one.
//            final ExhibitPropsLoadable epl2 = eplDup.get(epl);
//            if(epl2 == null)
//                {
//                // See if we can pool this VM-wide.
//                final ExhibitPropsLoadable ePooled = MemoryTools.intern(epl);
//                eplDup.put(ePooled, ePooled); // New value...
//                // If global pooling did not return our original value, then use the new global value.
//                if(ePooled != epl)
//                    { loadedProps.put(key, ePooled); }
//                }
//            else { loadedProps.put(key, epl2); } // Replace duplicate!
//            }
////if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[AllExhibitProperties._optimiseLP(): #exhibits="+aeid.length+", #EPL="+loadedProps.size()+", #distinctEPL="+eplDup.size()+".]"); }
//        }


    /**Frequently-used value instance Integer(1). */
    private static final Integer ONE = Integer.valueOf(1);

    /**Map from author initials (String) to exhibit count for that author (Integer).
     * Only authors that have have one or more exhibits are stored in the map.
     * <p>
     * This map is immutable, and is computed on first use.
     * <p>
     * This map is transient (ie not stored in the persistent state).
     */
    private transient volatile Map<String,Integer> authorExhibitCounts;

    /**Get a map from author initials (String) to exhibit count for that author (Integer); never null.
     * Only authors that have have one or more exhibits are stored in the map.
     * <p>
     * This map is transient (ie not stored in the persistent state),
     * and volatile for thread-safe lockless access.
     */
    public final Map<String,Integer> getAuthorExhibitCounts()
        {
        Map<String,Integer> result = authorExhibitCounts;
        if(result == null)
            {
            // Needs (re)computing.
            final Set<String> allExhibits = aeid.getAllExhibitNamesSet(); // Fast.
            final HashMap<String,Integer> rawResult = new HashMap<String,Integer>(); // For speed.
            for(final Iterator<String> it = allExhibits.iterator(); it.hasNext(); )
                {
                final String fullName = it.next();
                final String author = ExhibitName.getAuthorComponent(fullName);
                final Integer oldCount = rawResult.get(author);
                if(oldCount == null) { rawResult.put(author, ONE); }
                else { rawResult.put(author, Integer.valueOf(oldCount.intValue() + 1)); }
                }
            authorExhibitCounts = result = Collections.unmodifiableMap(rawResult);
            }
        return(result);
        }

    /**Map from category [top-level directory] (String) to exhibit count for that category (Integer).
     * Only categories that have have one or more exhibits are stored in the map.
     * <p>
     * This map is immutable, and is computed on first use.
     * <p>
     * This map is transient (ie not stored in the persistent state),
     * and volatile for thread-safe lockless access.
     */
    private transient volatile Map<String,Integer> categoryExhibitCounts;

    /**Get a map from category [top-level directory] (String) to exhibit count for that category (Integer); never null.
     * Only categories that have have one or more exhibits are stored in the map.
     * <p>
     * This map is immutable, and is computed on first use.
     */
    public final Map<String,Integer> getCategoryExhibitCounts()
        {
        Map<String,Integer> result = categoryExhibitCounts;
        if(result == null)
            {
            // Needs (re)computing.
            final Set<String> allExhibits = aeid.getAllExhibitNamesSet(); // Fast.
            final HashMap<String,Integer> rawResult = new HashMap<String,Integer>(32); // For speed.
            for(final Iterator<String> it = allExhibits.iterator(); it.hasNext(); )
                {
                final String fullName = it.next();
                final String category = ExhibitName.getCategoryComponent(fullName);
                final Integer oldCount = rawResult.get(category);
                if(oldCount == null) { rawResult.put(category, ONE); }
                else { rawResult.put(category, Integer.valueOf(oldCount.intValue() + 1)); }
                }
            categoryExhibitCounts = result = Collections.unmodifiableMap(rawResult);
            }
        return(result);
        }

    /**Limit of goodness*correlation to consider something (eg category) significantly good or bad. */
    public static final float GOODBAD_LIMIT = 0.4f;


    /**Find out if a category is rated "good"/popular or not.
     * Returns TRUE if rated good, FALSE if bad,
     * null if not significantly either or if not known.
     *
     * @param categoryDir  the initial directory component of an extant exhibit
     * @param force  if true may force (expensive) computation to give
     *     a more accurate answer,
     *     else may return a more approximate or stale answer,
     *     or none at all (null)
     */
    public Boolean isCategoryGood(final String categoryDir,
                                  final BasicVarMgrInterface vars,
                                  final boolean force)
        throws IOException
        { return(_vcCache.isCategoryGood(categoryDir, this, vars, force)); }

    /**Map from dotted-file-extension (String) to exhibit count for that extension (Integer).
     * Only extensions that have have one or more exhibits are stored in the map.
     * <p>
     * This map is immutable, and is computed on first use.
     * <p>
     * This map is transient (ie not stored in the persistent state),
     * and volatile for thread-safe lockless access.
     */
    private transient volatile Map<String,Integer> dottedExtensionExhibitCounts;

    /**Get a map from dotted-file-extension (String) to exhibit count for that author (Integer); never null.
     * Only extensions that have have one or more exhibits are stored in the map.
     * <p>
     * This map is immutable, and is computed on first use.
     */
    public final Map<String,Integer> getDottedExtensionExhibitCounts()
        {
        Map<String,Integer> result = dottedExtensionExhibitCounts;
        if(result == null)
            {
            // Needs (re)computing.
            final Set<String> allExhibits = aeid.getAllExhibitNamesSet(); // Fast.
            final HashMap<String,Integer>  rawResult = new HashMap<String,Integer> (); // For speed.
            for(final Iterator<String> it = allExhibits.iterator(); it.hasNext(); )
                {
                final String fullName = it.next();
                final String dExt = "." + ExhibitName.getExtensionComponent(fullName);
                final Integer oldCount = rawResult.get(dExt);
                if(oldCount == null) { rawResult.put(dExt, ONE); }
                else { rawResult.put(dExt, Integer.valueOf(oldCount.intValue() + 1)); }
                }
            dottedExtensionExhibitCounts = result = Collections.unmodifiableMap(rawResult);
            }
        return(result);
        }




    /**Returns set of available extensions, sorted case insensitively.
     * This may be more than actually appear in the data set
     * (and this result does not depend on the data set, in fact).
     * This result contains no duplicates.
     * <p>
     * These are suffixes starting with `.' eg ``.jpg''.
     * <p>
     * Never returns null.
     */
    public static String[] getLegalSuffixes()
        {
        final ExhibitMIME.ExhibitTypeParameters types[] = ExhibitMIME.getAllValidExhibitTypes();
        final Set<String> suffixes = new HashSet<String>(types.length * 2 + 1);
        for(int i = types.length; --i >= 0; )
            { suffixes.add(types[i].dotSuffixForInputFile); }
        final String result[] = new String[suffixes.size()];
        suffixes.toArray(result);
        Arrays.sort(result, String.CASE_INSENSITIVE_ORDER);
        return(result);
        }




    /**Get total size in bytes of all exhibits.
     * Computed on first use.
     */
    public long getTotalExhibitBytes()
        {
        long totalBytes = exhibitBytes; // One read access in normal case.
        if(totalBytes == 0) // Recompute.
            {
            final Set<String> allExhibits = aeid.getAllExhibitNamesSet();
            for(final Iterator<String> it = allExhibits.iterator(); it.hasNext(); )
                {
                final String name = it.next();
                totalBytes += aeid.getStaticAttr(name).length;
                }
            exhibitBytes = totalBytes; // Cache result.
            }
        return(totalBytes);
        }

    /**Total size in bytes of all exhibits.
     * A zero value is taken to mean "not computed", though in the case where
     * there are no exhibits the resulting redundant recomputations will be cheap.
     * <p>
     * This is transient since it can be recalculated on demand.
     * <p>
     * This is volatile to allow safe lock-free access.
     */
    private transient volatile long exhibitBytes;



    /**Returns a Map from attribute (String word) to a Set of full names of all exhibits with each attribute; never null.
     * An exhibit can have a particular attribute either
     * by having the attribute in the name (in the usual place at the end),
     * or in some circumstances by the exhibit type or content being examined.
     * <p>
     * The returned value is immutable (and the keys and values are immutable)
     * and is never null.
     * <p>
     * Attributes which no exhibits have are present in the Map
     * with an empty Set as the mapped-to value.
     */
    public Map<String,Set<String>> getExhibitsByAttribute()
        {
        synchronized(_exhibitsByAttribute_lock)
            {
            if(exhibitsByAttribute == null)
                { // Needs recomputing.
                final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

                // We use a HashMap for lookup speed.
                final Map<String,Set<String>> result = new HashMap<String,Set<String>>(attrWords.size() * 2 + 1);
                // Make empty Set entries for each attribute word.
                for(final Iterator<String> it = attrWords.iterator(); it.hasNext(); )
                    {
                    // We use HashSet for eventual lookup speed.
                    result.put(it.next(), new HashSet<String>());
                    }

                // Now go through all the exhibits,
                // adding them all by their explicit attribute words.
                for(final Iterator<String> it = aeid.getAllExhibitNamesSet().iterator(); it.hasNext(); )
                    {
                    final String fullName = it.next();
                    // Get the attribute words for this to be filed under.
                    final Enumeration en = ExhibitName.getAttributeWordsComponentEnumeration(
                            fullName, attrWords);
                    if(en != null)
                        {
                        while(en.hasMoreElements())
                            {
                            final String attrWord = (String) en.nextElement();
                            result.get(attrWord).add(fullName);
                            }
                        }
                    }

                // Now replace all the Sets in the Map with unmodifiable ones,
                // In particular replacing empty sets with the static singleton.
                final Set<String> noWords = Collections.emptySet();
                for(final Iterator<String> it = result.keySet().iterator(); it.hasNext(); )
                    {
                    final String attrWord = it.next();
                    final Set<String> s = result.get(attrWord);
                    if(s.size() == 0)
                        { result.put(attrWord, noWords); }
                    else
                        { result.put(attrWord, Collections.unmodifiableSet(s)); }
                    }

                // Now make the whole map unmodifiable.
                exhibitsByAttribute = Collections.unmodifiableMap(result);
                }

            return(exhibitsByAttribute);
            }
        }

    /**A Map from attribute (String word) to a Set of full names of all exhibits with each attribute.
     * An exhibit can have a particular attribute either
     * by having the attribute in the name (in the usual place at the end),
     * or in some circumstances by the exhibit type or content being examined.
     * <p>
     * This is immutable, and is transient and recomputed on first use,
     * being entirely derived from other information in object.
     * <p>
     * We hold a private lock while recomputing this information
     * to allow other processing to continue in parallel with this
     * possibly slow task.
     */
    private transient Map<String,Set<String>> exhibitsByAttribute; // Initially null.

    /**Private lock for manipulating exhibitsByAttribute; must use construction to create correctly. */
    private final transient Object _exhibitsByAttribute_lock = new Object();


    /**Get case-sensitive Set of all main words (String values) in extant exhibits; never null.
     * Result will be empty if there are no exhibits.
     * <p>
     * Computed on first use under a private lock as this may be slow
     * and we do not want to waste CPU time on redundant computations.
     */
    public Set<String> getMainWords()
        {
        synchronized(_mainWords_lock)
            {
            if(mainWords == null)
                {
                final Set<String> allNames = aeid.getAllExhibitNamesSet();
                final Set<String> allAttrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

                // Create the result making an initial conservative guess at size.
                final Set<String> result = new HashSet<String>(3 * allNames.size()); // HashSet for lookup speed.

                for(final Iterator<String> it = allNames.iterator(); it.hasNext(); )
                    {
                    final String fullName = it.next();
                    final Enumeration words = ExhibitName.getMainWords(fullName, allAttrWords);
                    while(words.hasMoreElements())
                        { result.add((String) words.nextElement()); }
                    }

                // Cache as an immutable value.
                mainWords = Collections.unmodifiableSet(result);
                }

            return(mainWords); // Can safely hand out immutable.
            }
        }

    /**Case-sensitve immutable cache (transient, possibly memory-sensitive) of Set of all exhibit main words.
     * Computed on first use.
     * <p>
     * Accessed under the _mainWords_lock.
     * <p>
     * Used, for example, by the upload routines to warn of possibly-misspelt
     * words.
     */
    private transient Set<String> mainWords;

    /**Private lock for manipulating mainWords; must use constructor to get set up correctly. */
    private final transient Object _mainWords_lock = new Object();


    /**Private cache; Map from full exhibit name to ExhibitPropsComputableMutable.
     * May or may not be serialised (may be transient to economise on space);
     * never null if a constructor has been called.
     * <p>
     * A thread-safe Map implementation is used;
     * one that allows decent concurrency, especially without locking.
     */
    private transient final ConcurrentMap<String,ExhibitPropsComputableMutable> epcmMap =
        _createNewEpcmMap(1001);

    /**Recover EPCM (and possibly other details) from old AEP.
     * This can save a great deal of recomputation after an AEP update.
     * <p>
     * Will only use details from old AEP:
     * <ul>
     * <li>That match current exhibits.
     * <li>That are not stale.
     * <li>If we have no such data at all
     *     (to prevent later overwriting of newly-computed data).
     * </ul>
     * <p>
     * This should not be called once threads other than the caller have access
     * to this AEP since after performing our initial checks there is a danger
     * of races that might leave old data in place.
     * <p>
     * This does not compute any new entries.
     */
    public void recoverOldExhibitPropsComputableMutableData(final AllExhibitProperties oldAEP,
                                                            final SimpleLoggerIF logger)
        {
        if(oldAEP == null) { throw new IllegalArgumentException(); }

        final ConcurrentMap<String,ExhibitPropsComputableMutable> m = epcmMap;
        assert(m != null);

        // If map is non-empty then don't accept data from old AEP.
        // This avoids accidentally overwriting fresh values with stale ones
        // or adding stale/dubious values to a live AEP instance.
        if(!m.isEmpty()) { return; }

        // If old map is empty then there is nothing to recover.
        final ConcurrentMap<String,ExhibitPropsComputableMutable> oldM = oldAEP.epcmMap;
        if((oldM == null) || (oldM.isEmpty())) { return; }

if(logger != null) { logger.log("[AllExhibitProperties: recovering EPCM entries: old map size: "+oldM.size()+".]"); }

        for(final String exhibitName : aeid.getAllExhibitNamesSet())
            {
            final ExhibitPropsComputableMutable oldECPM =
                oldM.get(exhibitName);
            if((oldECPM == null) || oldECPM.isStale())
                { continue; }
            // Double-check that we have not computed own result in interim.
            if(m.get(exhibitName) != null)
                { continue; }
            // OK, save non-stale value.
            m.put(exhibitName, oldECPM);
            }

if(logger != null) { logger.log("[AllExhibitProperties: finished recovering EPCM entries: new map size: "+m.size()+".]"); }
        }

    /**Get ExhibitPropsComputableMutable for specified exhibit; null if no such exhibit else never null.
     * If the "allowStale" parameter is false,
     * the system will make best efforts to compute an accurate, non-stale value.
     * However, this may not be possible for a number of contingent reasons,
     * such as transient I/O errors.
     * <p>
     * Equally, if the "allowStale" flag is true,
     * there may be a non-stale value cached that can be returned.
     * <p>
     * So the "allowStale" parameter should be regarded as a strong hint at most.
     *
     * @param exhibitName  full name of exhibit for which data is desired; never null
     * @param allowStale  allows "approximate" or "stale" version to be returned
     *     if need be, ie does not force re-computation of new value,
     *     useful if site is busy or time is short
     * @param gp  system properties;
     *     never null and not default/empty for non-stale result
     * @param dataSource  source of event and exhibit (meta) data;
     *     must not be null if a non-stale result is desired
     * @param scorers  source of exhibit-content based scoring;
     *     must not be null if a non-stale result is desired
     * @return non-null result if exhibit exists;
     *     possibly stale/approximate if allowStale==true or gp/dsb are empty/null
     */
    public ExhibitPropsComputableMutable getExhibitPropsComputableMutable(final String exhibitName,
                                                                          final boolean allowStale,
                                                                          final GenProps gp,
                                                                          final SimpleExhibitPipelineIF dataSource,
                                                                          final ScorerCacheIF scorers)
        {
        // Get internal representation of name to save memory if creating new entry,
        // or reject request if exhibit name is not valid/current.
        final ExhibitStaticAttr esa = aeid.getStaticAttr(exhibitName);
        if(esa == null) { return(null); /* No such exhibit. */ }
        final String ex = esa.filePath; // Shared representation of name to save heap.

        // Get current map.
        final ConcurrentMap<String,ExhibitPropsComputableMutable> m = epcmMap;
        assert(m != null) : "epcmMap cannot be null now";

        final ExhibitPropsComputableMutable cachedValue = epcmMap.get(ex);
        final boolean cachedValueIsNull = (cachedValue == null);
        if(allowStale)
            {
            if(cachedValueIsNull)
                {
                // Quickly create an approximate and (probably) stale entry.
                final ExhibitPropsComputableMutable result = ExhibitPropsComputableMutable.generateFastApproximation(esa, gp);
                epcmMap.putIfAbsent(ex, result); // Avoid overwriting a newer value.

                // If we superficially have all the required data sources available
                // (we leave fine sufficiency judgements up to the EPCM calculation routine)
                // then attempt to create fully up-to-date value in the background.
                // We severely limit the resources consumed by this speculative computation,
                // and never block waiting for the computation since it may take a LONG time.
                if(result.isStale() &&
                   (aeid.length != 0) &&
                   (gp != null) && (gp.timestamp != 0) &&
                   (dataSource != null) &&
                   (scorers != null))
                    {
                    // Recursively call to try to compute and cache an up-to-date value.
                    // The allowStale flag is flipped thus preventing further recursion.
                    ThreadUtils.lowPriorityThreadPoolDiscardable.submit(new Runnable(){
                        public final void run()
                            { getExhibitPropsComputableMutable(exhibitName, false, gp, dataSource, scorers); }
                        });
                    }

                return(result);
                }

            // Fall through to return non-null cached value.
            }
        else
            {
            // Force our cached entry up-to-date if absent/stale,
            // given that stale is not good enough for the caller.
            if(cachedValueIsNull || cachedValue.isStale())
                {
                // Compute full result if possible,
                // but fall back to "stale" result if necessary so as to avoid returning null.
                try
                    {
                    final ExhibitPropsComputableMutable result = ExhibitPropsComputableMutable.compute(
                        esa, gp, this, dataSource, _vcCache, scorers);
                    // Avoid race causing older value to overwrite a newer concurrently-computed one.
                    if(cachedValueIsNull) { epcmMap.putIfAbsent(ex, result); }
                    else { epcmMap.replace(ex, cachedValue, result); }
                    return(result);
                    }
                catch(final Exception e)
                    {
                    // If full calculation failed, then try for a fast approximation.
                    final ExhibitPropsComputableMutable result = ExhibitPropsComputableMutable.generateFastApproximation(esa, gp);
                    // Avoid race causing older value to overwrite a newer concurrently-computed one.
                    if(cachedValueIsNull) { epcmMap.putIfAbsent(ex, result); }
                    else { epcmMap.replace(ex, cachedValue, result); }
                    return(result);
                    }
                }

            // Fall through to return non-null non-stale cached value.
            }

        assert(cachedValue != null) : "result cannot be null";
        return(cachedValue);
        }

    /**Private vote/correlation cache; never null after construction/deserialisation.
     * This state could be serialised to save some recomputation on re-use.
     * <p>
     * This is thread-safe.
     */
    private final transient ExhibitPropsComputableMutableVoteCacheIF _vcCache =
        new ExhibitPropsComputableMutableVoteCache();

    /**Bring up-to-date any vote/correlation data.
     * Useful to call from a background thread from time to time
     * in order that all correlations will be precomputed when required.
     *
     * @param vars  non-null handle on system variables
     * @param noTimeLimit  if true, this will try to complete the work in one go
     */
    public void updateVoteCache(final SimpleVariablePipelineIF vars,
                                final boolean noTimeLimit)
        throws IOException
        { _vcCache.update(this, vars, noTimeLimit); }

    /**Create new epcmMap.
     * This is where we set policy on map segments, load factor, etc.
     * <p>
     * We use a concurrency-friendly, lock-free Map for lookup and update speed.
     * We don't expect lots of concurrent writes,
     * so a few segments (fewer than say 10) should be OK.
     */
    private static ConcurrentHashMap<String, ExhibitPropsComputableMutable> _createNewEpcmMap(final int capacity)
        {
        return(new ConcurrentHashMap<String, ExhibitPropsComputableMutable>(
                            capacity, 0.75f, 4));
        }


    /**Get exhibit Location information, specific or generic, if any; Location.NONE if none, never null.
     * This tries first for a "specific" location for this exhibit,
     * then tries for a generic location from the pattern-matching mechanism.
     *
     * @param exhibitName  the full name of the exhibit; must not be null
     * @return the location information for this exhibit; NONE if none
     */
    public Location.Base getLocation(final String exhibitName)
        {
        // Try for "specific" location information first.
        final Location.Base ls =
            getExhibitPropsLoadable(exhibitName).getLocation();
        if((ls != null) &&
            (!ls.equals(Location.NONE))) // Possibly redundant.
            { return(ls); }

        // OK: sould find any specific information, so try for generic info.
        final Location.Base lg = epgi.getLocationMap().
            locLookup(exhibitName, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
        if(lg != null)
            { return(lg); }

        // No location found.
        return(Location.NONE);
        }



    /**Immutable map from MD5 hash to exhibit full name; null until computed on first use.
     * Volatile for lock-free thread-safe access.
     * <p>
     * Transient; (re)computed on demand after deserialisation.
     */
    private volatile transient Map<ROByteArray,String> hashMD5ToName;

    /**Immutable Set of all exhibits with duplicate or missing MD5 hashes; null until computed on first use.
     * Ideally this is empty.
     * <p>
     * Volatile for lock-free thread-safe access.
     * <p>
     * Transient; (re)computed on demand after deserialisation.
     */
    private volatile transient Set<String> hashMD5Err;

    /**Get immutable map from MD5 hash to valid exhibit full name; never null.
     * Computed on first use.
     * <p>
     * Sets internal hashMD5ToName and hashMD5Err on first use as a side-effect.
     */
    public Map<ROByteArray,String> getHashMD5ToName()
        {
        if(hashMD5ToName == null)
            {
            // Compute on first use...

            // We try not to waste lots of space in this map.
            // We use a HashMap for lookup speed.
            final HashMap<ROByteArray, String> m = new HashMap<ROByteArray, String>(aeid.length + (aeid.length/4) + 1, 0.81f);

            // Collect the "error" list too.
            final HashSet<String> e = new HashSet<String>();

            for(final String fullName : aeid.getAllExhibitNamesSet())
                {
                assert(ExhibitName.validNameSyntax(fullName));
                final ExhibitPropsLoadable exhibitPropsLoadable = getExhibitPropsLoadable(fullName);
                final AccessionData accessionMetadata = exhibitPropsLoadable.getAccessionMetadata();
                if(accessionMetadata == null)
                    {
                    // No meta-data, so this goes in the error list...
                    e.add(fullName);
                    continue;
                    }
                final ROByteArray hashMD5 = accessionMetadata.hashMD5;
                if(hashMD5 == null)
                    {
                    // No MD5 hash, so this goes in the error list...
                    e.add(fullName);
                    continue;
                    }
                // Insert into table.
                final String prev = m.put(hashMD5, fullName);
                if(prev != null)
                    {
                    assert(ExhibitName.validNameSyntax(prev));
                    // Previous value falls into error bucket...
                    e.add(prev);
                    }
                }

            // OK, save results, errors first to avoid races in getHashMD5Err().
            if(e.isEmpty()) { hashMD5Err = Collections.emptySet(); }
            else { hashMD5Err = Collections.unmodifiableSet(e); }
            hashMD5ToName = Collections.unmodifiableMap(m);
            }

        final Map<ROByteArray,String> result = hashMD5ToName;
        assert(result != null); // Result is never null.
        return(result);
        }

    /**Get immutable Set of all exhibits with duplicate or missing MD5 hashes; never null.
     * Ideally this is empty.
     */
    public Set<String> getHashMD5Err()
        {
        if(hashMD5Err == null)
            {
            getHashMD5ToName(); // Get our value recomputed as a side-effect.
            }
        final Set<String> result = hashMD5Err;
        assert(result != null); // Result is never null.
        return(result);
        }


    /**Filter class for exhibits in AEP.
     * If the accept() method returns false then the given exhibit
     * is excluded from the results of select().
     */
    public static interface AEPFilter
        {
        public boolean accept(AllExhibitProperties aep, String fullExhibitName);
        }

    /**Filter results optionally starting with an input set of exhibit names; never null.
     * This returns a String[] of full exhibit name results,
     * starting with the given input set of names,
     * or all exhibits (set up in smart-sorted order) if null is passed in.
     * <p>
     * This search may be parallelised.
     * <p>
     * Note that is its up to the filters to decide whether to
     * call getExhibitPropsComputable() to guarantee correct answers
     * or getExhibitPropsComputableIfPresent() to guarantee fast answers.
     * <p>
     * If the stopAfter parameter is greater than zero,
     * then this routine will stop as soon as possible after stopAfter results,
     * and will then truncate to exactly stopAfter results.
     * <p>
     * The results array is sorted in a `smart'ish way.
     * <p>
     * The filter argument must not be null.
     *
     * @return  non-null, possibly empty, smart-sorted full exhibit names
     */
    public String[] select(final AEPFilter f,
                           String initialNames[],
                           final int stopAfter)
//        throws IOException
        {
        assert(f != null);
        if(initialNames == null) { initialNames = aeid.getAllExhibitNamesSorted(); }

        // Accumulate results here.
        List<String> result = new ArrayList<String>();

        // Do the filtering...
        for(int i = initialNames.length; --i >= 0; )
            {
            final String thisExhibit = initialNames[i];
            if(!f.accept(this, thisExhibit)) { continue; }
            result.add(thisExhibit);

            if((stopAfter > 0) && (result.size() >= stopAfter)) { break; }
            }

        // Truncate results to size of need be.
        if((stopAfter > 0) && (result.size() > stopAfter))
            { result = result.subList(0, stopAfter); }

        // Generate (sorted) array form and return...
        final String r[] = new String[result.size()];
        result.toArray(r);
        Arrays.sort(r, ExhibitName.SIMPLE_SMART_ORDER);
        return(r);
        }
    
    /**Generate human-readable summary of this instance's internal state. */
    @Override
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(128);
        sb.append("aeid.length=").append(aeid.length);
//        sb.append(",epl.size=").append(loadedProps.size());
        sb.append(",epgi=").append(epgi);
        return(sb.toString());
        }

    /**Returns a hash code value for the object; derived from the longHash.
     * Guaranteed zero if there are no exhibits, non-zero otherwise.
     *
     * @return  a hash code value for this object.
     * @see     Object#equals(Object)
     * @see     Hashtable
     */
    public int hashCode()
        {
        return(((int) (longHash >>> 32)) ^ (int) longHash);
        }

    /**Indicates whether some other object is "equal to" this one; the underlying data is the same if true.
     * The hashNotChangedSince value is excluded from the comparison.
     * <p>
     * Loaded data is included in this notion of equality;
     * computed/derived data is excluded.
     * The construction time and "unchanged since" values are also excluded.
     * <p>
     * This routine is likely to be very slow,
     * especially to return a "true" result,
     * ie fully verify that two instances are identical in all significant aspects.
     *
     * @param   obj   the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     * @see     Boolean#hashCode()
     * @see     Hashtable
     */
    public synchronized boolean equals(final Object obj)
        {
    	if(this == obj) { return(true); }
        // Must be of the same type to be equal.
        if(!(obj instanceof AllExhibitProperties)) { return(false); }
        final AllExhibitProperties other = (AllExhibitProperties) obj;

        // We can quickly reject items that don't have the same hash...
        // (Else hashCode() may be inconsistent with equals().)
        if(longHash != other.longHash) { return(false); }

        // Check the underlying immutable and other source data.
        if(!aeid.equals(other.aeid)) { return(false); }
        if(!epgi.equals(other.epgi)) { return(false); }
        if(!loadedProps.equals(other.loadedProps)) { return(false); }
        if(!computedProps.equals(other.computedProps)) { return(false); }

        // OK, seems to actually be equal.
        return(true);
        }

    /**Compact the internal representation of this instance (and its sub-objects) if possible.
     * This has no effect on the logical content of this instance in-memory or serialised,
     * is guaranteed to be safe to run concurrently with other uses of this instance
     * (and will take any locks as needed to work incrementally),
     * and may do nothing but consume some CPU cycles.
     * <p>
     * This may be able to convert some state to a more memory-efficient representation
     * after construction or deserialisation,
     * and is suitable to call in a background thread.
     * <p>
     * We don't prevent multiple concurrent calls to this routine,
     * since they are at worst wasteful of CPU but not unsafe.
     */
    public void compact()
        {
        // Iterate over all compact()able per-exhibit data...
        for(final String exhibitName : aeid.getAllExhibitNamesSet())
            {
            try
                {
                final ExhibitPropsComputable epc = getExhibitPropsComputable(exhibitName);
                if(epc != null) { epc.compact(); }
                final ExhibitPropsLoadable epl = getExhibitPropsLoadable(exhibitName);
                if(epl != null) { epl.compact(); }
                Thread.yield(); // Be nice to other activity in the JVM.
                }
            catch(final RuntimeException e)
                {
                // Ignore (but log) any transient problems.
                e.printStackTrace();
                }
            }
        }

//    /**Deserialise. */
//    private synchronized void readObject(ObjectInputStream in)
//        throws IOException, ClassNotFoundException
//        {
//        in.defaultReadObject();
//
//        // Create any private lock objects needed.
////        _exhibitsByAttribute_lock = new Object();
////        _mainWords_lock = new Object();
//
//        // Defensive copy of mutable state.
//        computedProps = new Hashtable<String,ExhibitPropsComputable>(computedProps);
//        if(epcmMap != null)
//            {
//            final ConcurrentHashMap<String,ExhibitPropsComputableMutable> newEpcmMap =
//                    _createNewEpcmMap(epcmMap.size());
//            newEpcmMap.putAll(epcmMap);
//            epcmMap = newEpcmMap;
//            }
//
//        _optimiseLP(); // Optimise memory footprint.
//
//        // Reset transient values.
//        computedProps_count = computedProps.size();
//
//        validateObject(); // Validate state immediately.
//        }


    /**Deserialise: use constructor for validation, defensive copying, renormalisation, etc.
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Construct new instance of object in normal defensive way.
    	// Note: allow the hash to be recomputed to suit any algorithm changes;
    	// we keep the hashUnchangedSince value iff the hash doesn't change on recomputation.
        return(new AllExhibitProperties(this, epgi, aeid, loadedProps, computedProps, 0));
        }


    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        if(longHash < 0)
            { throw new InvalidObjectException("bad object: longHash < 0"); }
        if(constructedAt < 0) // We could have a tighter check, eg > 2001/01/01.
            { throw new InvalidObjectException("bad object: constructedAt < 0"); }
        if(hashNotChangedSince < 0) // We could have a tighter check, eg > 2001/01/01.
            { throw new InvalidObjectException("bad object: hashNotChangedSince < 0"); }
        if(aeid == null)
            { throw new InvalidObjectException("bad object: aeid == null"); }
        if(aeid.timestamp < 0) // Sheer nosiness!
            { throw new InvalidObjectException("bad object: aeid.timestamp < 0"); }
        if(epgi == null)
            { throw new InvalidObjectException("bad object: epgi == null"); }
//        if(lm == null)
//            { throw new InvalidObjectException("bad object: lm == null"); }
        // Make sure that loadedProps maps valid exhibits to ExhibitPropsLoadable items.
        if(loadedProps == null)
            { throw new InvalidObjectException("bad object: loadedProps == null"); }
        for(final Iterator it = loadedProps.keySet().iterator(); it.hasNext(); )
            {
            final Object key = it.next();
            if(!(key instanceof String))
                { throw new InvalidObjectException("bad object: loadedProps: non-String key"); }
            if(aeid.getStaticAttr((String) key) == null)
                { throw new InvalidObjectException("bad object: loadedProps: non-exhibit key"); }
            if(!(loadedProps.get(key) instanceof ExhibitPropsLoadable))
                { throw new InvalidObjectException("bad object: loadedProps: non-ExhibitPropsLoadable value"); }
            }
        // Make sure that computedProps maps valid exhibits to ExhibitPropsComputable items.
        if(computedProps == null)
            { throw new InvalidObjectException("bad object: computedProps == null"); }
        synchronized(computedProps)
            {
            for(final Iterator it = computedProps.keySet().iterator(); it.hasNext(); )
                {
                final Object key = it.next();
                if(!(key instanceof String))
                    { throw new InvalidObjectException("bad object: computedProps: non-String key"); }
                if(aeid.getStaticAttr((String) key) == null)
                    { throw new InvalidObjectException("bad object: computedProps: non-exhibit key"); }
                if(!(computedProps.get(key) instanceof ExhibitPropsComputable))
                    { throw new InvalidObjectException("bad object: computedProps: non-ExhibitPropsComputable value"); }
                }
            }
        // Make sure that any epcmMap maps valid exhibits to ExhibitPropsComputableMutable items.
        if(epcmMap != null)
            {
            for(final Iterator it = epcmMap.keySet().iterator(); it.hasNext(); )
                {
                final Object key = it.next();
                if(!(key instanceof String))
                    { throw new InvalidObjectException("bad object: epcmMap: non-String key"); }
                if(aeid.getStaticAttr((String) key) == null)
                    { throw new InvalidObjectException("bad object: epcmMap: non-exhibit key"); }
                if(!(epcmMap.get(key) instanceof ExhibitPropsComputableMutable))
                    { throw new InvalidObjectException("bad object: epcmMap: non-ExhibitPropsComputableMutable value"); }
                }
            }
        }

    /**Our serial version... */
    private static final long serialVersionUID = -2563720785711714753L;
    }
