/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import java.io.*;
import java.util.*;

import ORG.hd.d.IsDebug;

/**Class to compute/represent changes from one AEP instance to another.
 * Where a small number of changes has occurred between one AEP and another,
 * especially where those changes are renamings or minor textual changes,
 * it may often be vastly more efficient to send a diff from an extant
 * AEP instance than send the new instance in toto.
 * <p>
 * There may be circumstances in which this class cannot compute a diff,
 * or where the diff is unlikely to be worthwhile (eg after huge changes),
 * in which case the diff computation routine indicates failure/unwillingness
 * and the caller may have to arrange for a full AEP to be transmitted.
 * <p>
 * This class is Serializable, and is reasonably efficient on-the-wire,
 * but is not necessarily intended for long-term persistence;
 * store the AEP instances themselves for that by preference.
 * <p>
 * We are cautious in that we store several explict before-and-after hashes
 * to be clear that we are applying diffs to the right AEP instance
 * and so as to be sure that we end up with the correct value.
 * 
 * @author DHD
 */
public final class AllExhibitPropertiesDelta implements Serializable
    {
    /**The long hash for the AEP to which the diff is to be applied; guaranteed non-negative. */
    public final long longHashAEPBefore;

    /**The long hash for the AEP result; guaranteed non-negative. */
    public final long longHashAEPAfter;

    /**The timestamp from the AEP result; guaranteed non-negative. */
    public final long hashNotChangedSinceAfter;

    /**The timestamp/hash for the AEIP in the AEP to which the diff is to be applied; guaranteed non-negative. */
    public final long timestampAEIDBefore;

    /**The timestamp/hash for the AEIP in the AEP result; guaranteed non-negative. */
    public final long timestampAEIDAfter;

    /**The exhibit count for the AEIP in the AEP to which the diff is to be applied; guaranteed non-negative. */
    public final int lengthAEIDBefore;

    /**The exhibit count for the AEIP in the AEP result; guaranteed non-negative. */
    public final int lengthAEIDAfter;

    /**The diff from the previous EPGI; may be null to represent no changes to EPGI value. */ 
    private final ExhibitPropsGlobalImmutable.EPGIDiff epgiDiff;

    /**The set of exhibits (full exhibit names) removed (or changed); may be null if no deletions/changes.
     * Not serialised as-is.
     * <p>
     * Does not contain duplicates, nulls or invalid-syntax values.
     * <p>
     * This is all exhibits whose old definitions/data must be
     * logically removed from the input AEP first.
     */
    private /* final */ transient Set<String> exhibitsDeleted;

    /**The set of exhibits (full exhibit names) added (or changed); may be null if no additions/changes.
     * Not serialised as-is.
     * <p>
     * Does not contain duplicates, nulls or invalid-syntax values.
     * <p>
     * This is all exhibits whose old definitions/data must be
     * logically added to the input AEP after the exhibitsDeleted set has been removed.
     */
    private /* final */ transient Set<Change> exhibitsAdded;


    /**Serialisation ID. */
    private static final long serialVersionUID = 8608225153676490831L;


    /**Minimal checked exception thrown to indicate that diff could not be generated/applied. */
    public static final class DiffException extends Exception
        {
        private static final long serialVersionUID = -330320938545639614L;
        public DiffException() { super(); }
        public DiffException(final String message) { super(message); }
        public DiffException(final String message, final Throwable cause) { super(message, cause); } 
        }
    
    /**Data for new/changed exhibit; immutable and serialisable. */
    private static final class Change implements Serializable, Comparable<Change>
        {
        /**Serialisation ID. */
        private static final long serialVersionUID = 1511290927439367951L;

        public Change(final ExhibitStaticAttr esa,
                      final ExhibitPropsLoadable epl,
                      final ExhibitPropsComputable epc)
            {
            if(null == esa)
                { throw new IllegalArgumentException(); }
            this.esa = esa;
            this.epl = epl;
            this.epc = epc;
            }

        /**ESA; never null. */
        public final ExhibitStaticAttr esa;
        /**EPL; may be null. */
        public final ExhibitPropsLoadable epl;
        /**EPC; may be null. */
        public final ExhibitPropsComputable epc;

        /**Equal iff all fields are identical. */
        @Override public boolean equals(final Object obj)
            {
            if(this == obj) { return(true); }
            if(!(obj instanceof Change)) { return(false); }
            final Change other = (Change) obj;

            if(null == epl) { if(null != other.epl) { return(false); } }
            else if(!epl.equals(other.epl)) { return(false); }
            if(null == epc) { if(null != other.epc) { return(false); } }
            else if(!epc.equals(other.epc)) { return(false); }

            final boolean esasIdentical = esa.isIdentical(other.esa);
//if(!esasIdentical && esa.equals(other.esa)) { System.err.println("Mismatch? "+esa+" vs "+other.esa); }
            return(esasIdentical);
            }
        /**The hash is based on the the exhibit name. */
        @Override public int hashCode()
            { return(esa.filePath.hashCode()); }
        /**Provide a total ordering, using name as the primary sort field. */
        public int compareTo(final Change other)
            {
            // Deal with esa fields first.
            final int nameCompare = esa.filePath.compareTo(other.esa.filePath);
            if(nameCompare != 0) { return(nameCompare); }
            if(esa.length < other.esa.length) { return(-1); }
            if(esa.length > other.esa.length) { return(+1); }
            if(esa.timestamp < other.esa.timestamp) { return(-1); }
            if(esa.timestamp > other.esa.timestamp) { return(+1); }

            // Deal with epl field.
            // Fake a total ordering using the hash.
            final int eplh = ((null == epl) ? 0 : epl.hashCode());
            final int oeplh = ((null == other.epl) ? 0 : other.epl.hashCode());
            if(eplh < oeplh) { return(-1); }
            if(eplh > oeplh) { return(+1); }

            // Deal with epc field.
            // Fake a total ordering using the hash.
            final int eplc = ((null == epc) ? 0 : epc.hashCode());
            final int oeplc = ((null == other.epc) ? 0 : other.epc.hashCode());
            if(eplc < oeplc) { return(-1); }
            if(eplc > oeplc) { return(+1); }

            // Items seem to be identical.
            return(0);
            }

        /**Human-readable summary. */
        @Override public String toString()
            {
            final StringBuilder sb = new StringBuilder();
            sb.append("<Change:");
            sb.append(esa);
            if(null != epl) { sb.append(',').append(epl); }
            if(null != epc) { sb.append(',').append(epc); }
            sb.append(">");
            return(sb.toString());
            }
        }


    /**Construct an empty instance.
     * This nominally represents the null diff/change
     * from one empty AEP instance to another empty AEP instance.
     */
    public AllExhibitPropertiesDelta()
        {
        this(0, 0, 0, 0, 0, 0, 0, null, null, null);
        }
    
    /**Construct an instance.
     * The Collection arguments are defensively copied.
     * <p>
     * The String and other complex values are assumed to already have been intern()ed
     * as far as needed/possible, given that any diff instance may be short-lived.
     */
    public AllExhibitPropertiesDelta(final long longHashAEPBefore,
                                     final long longHashAEPAfter,
                                     final long timestampAEIDBefore,
                                     final long timestampAEIDAfter,
                                     final int lengthAEIDBefore,
                                     final int lengthAEIDAfter,
                                     final long hashNotChangedSinceAfter,
                                     final ExhibitPropsGlobalImmutable.EPGIDiff epgiDiff,
                                     final Set<String> exhibitsDeleted,
                                     final Set<Change> exhibitsAdded)
        {
        this.longHashAEPBefore = longHashAEPBefore;
        this.longHashAEPAfter = longHashAEPAfter;
        this.hashNotChangedSinceAfter = hashNotChangedSinceAfter;
        this.timestampAEIDBefore = timestampAEIDBefore;
        this.timestampAEIDAfter = timestampAEIDAfter;
        this.lengthAEIDBefore = lengthAEIDBefore;
        this.lengthAEIDAfter = lengthAEIDAfter;
        this.epgiDiff = epgiDiff;

        // Take defensive copies of mutable collection arguments.
        this.exhibitsDeleted = ((exhibitsDeleted == null) || exhibitsDeleted.isEmpty()) ?
            null : new HashSet<String>(exhibitsDeleted);
        this.exhibitsAdded = ((exhibitsAdded == null) || exhibitsAdded.isEmpty()) ?
            null : new HashSet<Change>(exhibitsAdded);

        // Verify object state.
        try { validateObject(); }
        catch(final InvalidObjectException e)
            {
            final IllegalArgumentException err = new IllegalArgumentException(e.getMessage());
            err.initCause(e);
            throw err;
            }
        }
                                     

    /**Write out a less-redundant form of our internal information.
     * In particular, since deltas will not usually involve deletion of any exhibits,
     * we make the no-deletions case take no space at all in the serialised form.
     */
    private void writeObject(final ObjectOutputStream oos)
        throws IOException
        {
        // Write the fields that we are not trying to optimise.
        // Note that this includes our length field.
        oos.defaultWriteObject();

        // Send the deleted names as bare Strings in sorted order
        // to help with compression on the wire.
        final int sizeD = (null == exhibitsDeleted) ? 0 : exhibitsDeleted.size();
        final String outD[] = new String[sizeD];
        if(sizeD > 0) { exhibitsDeleted.toArray(outD); }
        Arrays.sort(outD);
        // We don't even write a length.
        // The presence of the first non-String value is sufficient delimiter.
        for(final String dn : outD)
            { oos.writeObject(dn); }

        // Send the added/changed exhibits in sorted order
        // to help with compression on the wire.
        final int sizeA = (null == exhibitsAdded) ? 0 : exhibitsAdded.size();
        final Change outA[] = new Change[sizeA];
        if(sizeA > 0) { exhibitsAdded.toArray(outA); }
        Arrays.sort(outA);
        oos.writeObject(outA);
        }

    /**Deserialise.
     */
    private void readObject(final ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        // Read the fields that we are not trying to optimise.
        ois.defaultReadObject();

        // Get the deleted exhibit names
        // reading until we hit the Change/additions array.
        Object objectIn = null;
        final HashSet<String> deletedNames = new HashSet<String>();
        while((objectIn = ois.readObject()) instanceof String)
            {
            if(!deletedNames.add((String) objectIn))
                { throw new InvalidObjectException("duplicate 'deleted' exhibit names"); }
            }
        if(!deletedNames.isEmpty())
            { exhibitsDeleted = new HashSet<String>(deletedNames); }

        // Get the added/changed exhibit details.
        // Must be in the non-String item that we just read.
        final Change inA[] = (Change[]) objectIn;
        if(inA.length > 0)
            {
            exhibitsAdded = new HashSet<Change>(Arrays.asList(inA));
            if(exhibitsAdded.size() != inA.length)
                { throw new InvalidObjectException("duplicate 'added' exhibit details"); }
            }

        validateObject();
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Basic validation of hashes/timestamps.
        if(longHashAEPBefore < 0)
            { throw new InvalidObjectException("bad object: longHashAEPBefore < 0"); }
        if(longHashAEPAfter < 0)
            { throw new InvalidObjectException("bad object: longHashAEPAfter < 0"); }
        if(timestampAEIDBefore < 0)
            { throw new InvalidObjectException("bad object: timestampAEIDBefore < 0"); }
        if(timestampAEIDAfter < 0)
            { throw new InvalidObjectException("bad object: timestampAEIDAfter < 0"); }
        if(lengthAEIDBefore < 0)
            { throw new InvalidObjectException("bad object: lengthAEIDBefore < 0"); }
        if(lengthAEIDAfter < 0)
            { throw new InvalidObjectException("bad object: lengthAEIDAfter < 0"); }
        if(hashNotChangedSinceAfter < 0)
            { throw new InvalidObjectException("bad object: hashNotChangedSinceAfter < 0"); }


        // Check that all "deleted" names are valid.
        // (Other values will have already been validated.)
        if(exhibitsDeleted != null)
            {
            for(final String en : exhibitsDeleted)
                {
                if(!ExhibitName.validNameSyntax(en))
                    { throw new InvalidObjectException("bad object: illegal 'deleted' exhibit name"); }
                }
            }

        
        // Check some basic relationships between fields.
        // (We expect this to be done anyway in the AEP/AEID instances themselves.)
        if((lengthAEIDBefore == 0) != (timestampAEIDBefore == 0))
            { throw new InvalidObjectException("bad object: zero AEID (before) length and timestamp should go together"); }
        if((lengthAEIDAfter == 0) != (timestampAEIDAfter == 0))
            { throw new InvalidObjectException("bad object: zero AEID (after) length and timestamp should go together"); }
        if((lengthAEIDBefore == 0) != (longHashAEPBefore == 0))
            { throw new InvalidObjectException("bad object: zero AEID (before) length and AEP longHash should go together"); }
        if((lengthAEIDAfter == 0) != (longHashAEPAfter == 0))
            { throw new InvalidObjectException("bad object: zero AEID (after) length and AEP longHash should go together"); }
        // TODO: add more such checks if it seems worthwhile to so do.
        }

    /**Create "change" values from AEP; never null.
     * This is the set of values to be added to an empty AEP to regenerate this one
     * (ignoring the EPGI).
     */
    private static Set<Change> _computeChangeValues(final AllExhibitProperties aep)
        {
        final Set<Change> result = new HashSet<Change>(2*aep.aeid.length);
        for(final String exhibitFullName : aep.aeid.getAllExhibitNamesSet())
            {
            result.add(new Change(aep.aeid.getStaticAttr(exhibitFullName),
                                  aep.getExhibitPropsLoadable(exhibitFullName),
                                  aep.getExhibitPropsComputable(exhibitFullName)));
            }
        assert(result.size() == aep.aeid.length);
        return(result);
        }


    /**Create diff between two AEP instances.
     * This creates a diff to be applied to an instance of the first argument
     * to recreate the second argument.
     * <p>
     * This will refuse to create a diff if it seems that
     * the diff is unlikely to be useful, for example:
     * <ul>
     * <li>If one AEP has more than twice the exhibits of the other
     *     (the diff may not be much smaller than sending the second AEP whole).
     * </ul>
     * 
     * @param force  if true then force a diff to be produced
     *     even if this routine would normally refuse to do so on efficiency grounds
     * 
     * @throws DiffException  if no diff can be generated
     *     or is unlikely to be worthwhile to use
     *     (eg would be more than a fraction of the size of the second AEP)
     */
    public static AllExhibitPropertiesDelta createDiff(final AllExhibitProperties aep1,
                                                       final AllExhibitProperties aep2,
                                                       final boolean force)
        throws DiffException
        {
        if((null == aep1) || (null == aep2))
            { throw new IllegalArgumentException(); }

        if(!force)
            {
            if((aep1.aeid.length > 2*aep2.aeid.length) ||
               (aep2.aeid.length > 2*aep1.aeid.length))
                { throw new DiffException("sizes too different (force==false)"); }
            }

        final Set<Change> aep1c = _computeChangeValues(aep1);
        final Set<Change> aep2c = _computeChangeValues(aep2);

        // The items that we will need to have in our "added" list.
        final Set<Change> added = new HashSet<Change>(aep2c);
        added.removeAll(aep1c);

        // The items that we will need to have in our deleted list
        // (we'll only need names in the end for this).
        final Set<Change> deleted = new HashSet<Change>(aep1c);
        deleted.removeAll(aep2c);

        // If there is probably not enough in common between the AEP instances
        // for a diff to be likely to be efficient, then abort.
        if(!force && (added.size() + deleted.size() > aep2.aeid.length/2))
            { throw new DiffException("exhibit sets too different (force==false): "+(added.size() + deleted.size())); }

        // If the EPGI has changed then we will need to include the diff.
        final ExhibitPropsGlobalImmutable.EPGIDiff diffEPGI = aep1.epgi.equals(aep2.epgi) ? null :
            ExhibitPropsGlobalImmutable.EPGIDiff.createDiff(aep1.epgi, aep2.epgi);

        // Create the "deleted" list as a Set of the names.
        // (Note that these names have already been intern()ed by ExhibitStaticAttr.)
        final Set<String> deletedNames = new HashSet<String>(2*deleted.size());
        for(final Change d : deleted)
            { deletedNames.add(d.esa.filePath); }

        // Create the diff object.
        final AllExhibitPropertiesDelta result = new AllExhibitPropertiesDelta(
                        aep1.longHash,
                        aep2.longHash,
                        aep1.aeid.timestamp,
                        aep2.aeid.timestamp,
                        aep1.aeid.length,
                        aep2.aeid.length,
                        aep2.hashNotChangedSince,
                        diffEPGI,
                        deletedNames,
                        added);

//        System.out.println(" *** AEP CREATE: DIFF="+result+"; AEP OUT="+aep2+"; AEP IN="+aep1);
        if(IsDebug.isDebug)
	        {
            assert(aep1.aeid.getAllExhibitNamesSet().containsAll(deletedNames));
	        final AllExhibitProperties aepOut;
			try { aepOut = applyDiff(aep1, result); }
			catch (final Exception e) { throw new Error(e); }			
			if(!aep2.equals(aepOut)) { throw new Error("Internal Error"); }
	        }

        return(result);
        }

    /**Applies diff to an extant AEP instance to generate a new AEP instance; never null.
     * The diff must only be applied to the same AEP instance value
     * that the diff was generated from (ie an equal one).
     * 
     * @throws DiffException  if the diff cannot be applied
     */
    public static AllExhibitProperties applyDiff(final AllExhibitProperties aep1,
                                                 final AllExhibitPropertiesDelta diff)
        throws DiffException
        {
        if((null == aep1) || (null == diff))
            { throw new IllegalArgumentException(); }

        // Make sure that the diff "before" values match those from the AEP.
        if((diff.lengthAEIDBefore != aep1.aeid.length) ||
           (diff.longHashAEPBefore != aep1.longHash) ||
           (diff.timestampAEIDBefore != aep1.aeid.timestamp))
            { throw new DiffException("diff cannot be applied to this AEP"); }

        // Compute mutable Set of the Change values for the input AEP.
        // We will mutate this to become the full exhibit data set for the result.
        final Set<Change> aep1c = new HashSet<Change>(_computeChangeValues(aep1));

        // If there were some deleted items,
        // then zap them here.
        final Set<String> del = diff.exhibitsDeleted;
        if(del != null)
            {
            final int initialSize = aep1c.size();
            for(final Iterator<Change> cit = aep1c.iterator(); cit.hasNext(); )
                {
                final Change c = cit.next();
                // Remove this exhibit if noted as deleted.
                if(del.contains(c.esa.filePath))
                    { cit.remove(); }
                }
            // All "deleted" exhibits must exist in the initial AEP.
            if(initialSize - del.size() != aep1c.size())
                { throw new DiffException("invalid disjunction/gap between initial and deleted items: initialSize="+initialSize+", delSize="+(del.size())+", trimmedSize="+aep1c.size()); }
            }

        // If there were some added items,
        // then add them here.
        final Set<Change> add = diff.exhibitsAdded;
        if(add != null)
            {
            // There should be no overlap between the remaining and added items.
            final int remainingSize = aep1c.size();
            aep1c.addAll(add);
            if(remainingSize + add.size() != aep1c.size())
                { throw new DiffException("invalid overlap between remaining and added items"); }
            }

        // Now synthesise a new AEID, etc, and build a new AEP.
        final Map<String, ExhibitStaticAttr> nameToStaticAttr = new HashMap<String, ExhibitStaticAttr>(2*aep1c.size());
        final Map<String, ExhibitPropsLoadable> loadedProps = new HashMap<String, ExhibitPropsLoadable>(2*aep1c.size());
        final Map<String, ExhibitPropsComputable> computedProps = new HashMap<String, ExhibitPropsComputable>(2*aep1c.size());
        for(final Change c : aep1c)
            {
            final String key = c.esa.filePath;
            nameToStaticAttr.put(key, c.esa);
            if((c.epl != null) && !c.epl.equals(ExhibitPropsLoadable.EMPTY))
                { loadedProps.put(key, c.epl); }
            if((c.epc != null) && !c.epc.equals(ExhibitPropsComputable.EMPTY))
                { computedProps.put(key, c.epc); }
            }
        final AllExhibitImmutableData aeid = new AllExhibitImmutableData(nameToStaticAttr, diff.timestampAEIDAfter);

        // Either epgi is unchanged so store the original in the result,
        // or else epgi has changed so apply the diff to get the new value.
        final ExhibitPropsGlobalImmutable newEpgi = ((diff.epgiDiff == null) ? aep1.epgi :
            (ExhibitPropsGlobalImmutable.EPGIDiff.applyDiff(aep1.epgi, diff.epgiDiff)));

        final AllExhibitProperties result = new AllExhibitProperties(
                        null,
                        newEpgi,
                        aeid,
                        loadedProps,
                        computedProps,
                        diff.hashNotChangedSinceAfter);

//        System.out.println(" *** AEP APPLY: DIFF="+diff+"; AEP OUT="+result);

        if((result.aeid.length != diff.lengthAEIDAfter) ||
           (result.aeid.timestamp != diff.timestampAEIDAfter))
            { throw new DiffException("unable to reconstruct aeid correctly"); }
        if(result.hashNotChangedSince != diff.hashNotChangedSinceAfter)
            { throw new DiffException("unable to reconstruct aep.hashNotChangedSince correctly"); }
        // We may not be able to reconstruct the AEP hash exactly (unless we force it)
        // when restoring data from an old AEP version since, for example,
        // the hash algorithm may have changed.
        if(result.longHash != diff.longHashAEPAfter)
            { throw new DiffException("unable to reconstruct aep.longHash correctly: diff="+diff+", putative aep="+result); }

        return(result);
        }
    
    /**Human-readable summary. */
    public final String toString()
        {
        final StringBuilder sb = new StringBuilder();
        sb.append("AEPDelta");
        if(exhibitsDeleted != null)
            { sb.append(":deleted=").append(exhibitsDeleted.size()); /* .append(new ArrayList<String>(exhibitsDeleted)); */ }
        if(exhibitsAdded != null)
            { sb.append(":added=").append(exhibitsAdded.size()); /* .append(new ArrayList<Change>(exhibitsAdded)); */ }
        return(sb.toString());
        }
    }
