/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;

import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.Tuple.Triple;

import ORG.hd.d.IsDebug;

/**Compact in-memory and on-the-wire representation of a (short) 7-bit ASCII String.
 * Converting to and from this form should be relatively resource-light,
 * eg low CPU and memory requirements.
 * <p>
 * This class is instance-controlled,
 * and uses intern() to avoid duplicate in-memory copies.
 * <p>
 * This cannot hold text longer than Short.MAX_VALUE characters long.
 * <p>
 * Internally this uses a byte[] to store the data
 * (a basic 2-fold space reduction of in-memory size in most JVMs)
 * with some tokenising for further compaction.
 * No attempt is made to share state with other instances in memory
 * (unless a shared StaticDictionary is used).
 * <p>
 * An instance of Compact7BitString will serialise itself as a String
 * rather than as the usual Compact7BitString
 * if it seems likely to be more efficent/compressable/safe to do so
 * (a) given the special handling of String and (b) for global compressability
 * or (c) if the value cannot be encoded other than as a String instance.
 * <p> 
 * So a coder had better make any serialisable field with a Compact7BitString value
 * be of type Object or Serializable to allow it to be read back as either.
 */
public final class Compact7BitString implements Serializable,
                                                ObjectInputValidation,
                                                MemoryTools.Internable
    {
    /**If true, log our encode/decode steps to help debug problems. */
    private static final boolean VERBOSE_DEBUG = false;

    /**If true, then log some global stats for tuning.
     * Should generally be false so as to avoid wasting CPU time and memory.
     */
    private static final boolean LOG_STATS = false;

    /**Immutable static dictionary to improve in-memory compression.
     * It is not permitted to serialise a Compact7BitString
     * that is using one of these so as to prevent decompression errors/ambiguitites.
     */
    public static final class StaticDictionary
        {
        /**Create an instance of a static dictionary.
         * 
         * @param name  dictionary name; must not be null
         * @param tokens  unique tokens, non-null, best first, each of length &gt; 1,
         *     non-empty, no longer than MAX_TOKENS
         */
        public StaticDictionary(final String name,
                                final List<String> tokens)
            {
            if(name == null) { throw new IllegalArgumentException(); }
            this.name = name;
            if(tokens == null) { throw new IllegalArgumentException(); }
            // Take a defensive copy.
            this.tokens = Collections.unmodifiableList(new ArrayList<String>(tokens));
            // Validate the token set.
            if(this.tokens.isEmpty() || (this.tokens.size() > MAX_TOKENS))
                { throw new IllegalArgumentException("too many/few tokens"); }
            if(this.tokens.size() != (new HashSet<String>(this.tokens)).size())
                { throw new IllegalArgumentException("duplicate token"); }
            for(final String s : this.tokens)
                {
                if((s == null) || (s.length() < 2))
                    { throw new IllegalArgumentException("invalid token (null or too short): " + s); }
                if(tokenise(s).size() != 1)
                    { throw new IllegalArgumentException("invalid token (mixed word and non-word chars): " + s); }
                }

            // Make lookup table from token to virtual "-ve" offsets.
            // Note that "best" tokens have the smallest offset
            // so as to be reachable from as far into the input stream as possible.
            final HashMap<String,Integer> virtualOffsets = new HashMap<String,Integer>(2 * this.tokens.size());
            // We always start at offset -2 in case use of offset -1 is not possible.
            int offset = FIRST_VIRTUAL_OFFSET;
            for(final String t : this.tokens)
                {
                assert(offset >= Byte.MIN_VALUE);
                final Integer oI = Integer.valueOf(offset--);
                virtualOffsets.put(t, oI);
                }
            assert(!virtualOffsets.isEmpty());
            this.virtualOffsets = virtualOffsets;
            }

        /**The maximum number of simple tokens for the static dictionary; strictly positive. */
        public static final int MAX_TOKENS = -Byte.MIN_VALUE - 2;

        /**First (numerically most positive) "virtual" offset used by the static dictionary; strictly negative.
         * We start at -2 to avoid conflict with -1 used to escape/start an MTM sequence.
         */
        private static final int FIRST_VIRTUAL_OFFSET = -2;

        /**The name of this dictionary, never null. */
        public final String name;

        /**The immutable (non-empty) list of unique, best-first tokens; never null nor empty. */
        public final List<String> tokens;

        /**Virtual offsets of tokens in static dictionary; never null nor empty.
         * Logically immutable.
         * <p>
         * Kept private because it is faster to keep it unwrapped and mutable.
         */
        private final HashMap<String,Integer> virtualOffsets;

        /**Human-readable summary: including the dictionary name; never null. */
        public String toString() { return(name); }

        /**Thread-safe map from text of uncompressed token (close to start of input) to count and summed-position of the token; non-null iff LOG_STATS is true.
         * Updated when attempting to compress String data.
         * <p>
         * This map is for early tokens of length 2 or more characters,
         * ie that we might be able to compress better eusing a static dictionary.
         * <p>
         * Only the first occurrence of each token in each compress/input is counted.
         */
        private final ConcurrentMap<String,Tuple.Pair<AtomicInteger,AtomicInteger>> _uncompEarlyTokens = !LOG_STATS ? null :
            new ConcurrentHashMap<String,Tuple.Pair<AtomicInteger,AtomicInteger>>(1<<16);

        /**Returns true if stats-collecting is enabled. */
        public boolean isStatsEnabled() { return(LOG_STATS); }

        /**Resets the stats (if any) being recorded for this dictionary. */
        public void resetStats()
            {
            if(LOG_STATS) { _uncompEarlyTokens.clear(); }
            }

        /**Dump stats (if any) to the given (non-null) stream.
         * This dumps the stats of compressions against this dictionary instance.
         * 
         * @param out  never null
         */
        public void dumpStats(final PrintStream out)
            {
            if(!LOG_STATS) { return; /* No stats to show. */ }

            // Generate a sorted set of early uncompressable tokens,
            // sorted by the product of count and token length.
            // In order these would then be good candidates
            // for a static compression dictionary for this class.
            // The set is of <token,count,rounded-up-mean-first-position>.
            final SortedSet<Tuple.Triple<String,Integer,Integer>> tokens = new TreeSet<Tuple.Triple<String,Integer,Integer>>(new Comparator<Tuple.Triple<String,Integer,Integer>>(){
                public int compare(final Triple<String,Integer,Integer> o1, final Triple<String,Integer,Integer> o2)
                    {
                    // Sort with higher count/length product first.
                    // (We use length-1 to reflect number of bytes saved after backref.)
                    final long l1 = o1.second.longValue() * (o1.first.length() - 1);
                    final long l2 = o2.second.longValue() * (o2.first.length() - 1);
                    if(l1 != l2)
                        {
                        if(l1 > l2) { return(-1); /* Correct order. */ }
                        return(1); /* Wrong order. */
                        }

                    // We'll break ties on the token value itself
                    // only so as to guarantee a stable total ordering.
                    return(o1.first.compareTo(o2.first));
                    }
                });

            // We rely on the thread-safety
            // (and immunity from ConcurrentModificationException)
            // of our iterator over _uncompEarlyTokens.
            for(final String token : _uncompEarlyTokens.keySet())
                {
                final Pair<AtomicInteger, AtomicInteger> pair = _uncompEarlyTokens.get(token);
                final int count = pair.first.get();
                assert(count > 0);
                final int summedFirstPos = pair.second.get();
                final int rUMFP = (summedFirstPos + count-1) / count;
                tokens.add(new Tuple.Triple<String,Integer,Integer>(token,
                                Integer.valueOf(count),
                                Integer.valueOf(rUMFP)));
                }

            out.println();
            out.println("DICTIONARY NAME: "+name);

            // Dump the tokens in order, best first.
            // We limit this to some initial plausible set of candidates.
            int shownSoFar = 0;
            for(final Tuple.Triple<String,Integer,Integer> t : tokens)
                {
                // Skip rare entries, noting that it takes > 2 to save memory overall.
                if(t.second.intValue() < 3) { continue; }

                // Skip entries that would be out of reach of their target tokens.
                if(t.third.intValue() + shownSoFar > StaticDictionary.FIRST_VIRTUAL_OFFSET + -Byte.MIN_VALUE) { continue; }

                // Compute an "escaped" version of the token
                // that can go in a String literal/constant.
                String s = t.first;
                s = s.replace("\\", "\\\\"); // Escape all \ as \\.
                s = s.replace("\"", "\\\""); // Replace " with \".
                s = s.replace("\r", "\\r");  // Relace CR with \r.
                s = s.replace("\n", "\\n");  // Relace LF with \n.
                s = s.replace("\t", "\\t");  // Relace TAB with \t.
                // Replace all other control characters with octal escapes.
                // Note that we cope gracefully with the value expanding while we work.
                for(int i = 0; i < s.length(); ++i)
                    {
                    final char c = s.charAt(i);
                    if((c < 32) || (c > 126))
                        { s.replace(""+c, "\\" + Integer.toOctalString(c)); /* Replace all occurences of this char at once. */ }
                    }

                out.println("\"" + s + "\",    /* count="+t.second+", saving="+(t.second.intValue()*(t.first.length()-1))+", meanFirstPos="+t.third+" */");
                if(++shownSoFar == StaticDictionary.MAX_TOKENS) { break; }
                }
            }
        }

    /**Convert to full String form. */
    public String toString()
        {
        // Simple empty case.
        if(text == null) { return(""); }
        
        // We accumulate the latest literal bytes of input here
        // until we can unambiguously extract one or more full tokens from it.
        // A surprising number of tokens are > 16 chars (the default SB size)
        // and accumulating untokenised input may thus often force a (slow) resize.
        final StringBuilder latestInput = new StringBuilder(32);

        final int len = text.length;
        final StringBuilder result = new StringBuilder(32 + (len<<3)); // Assume usually <8x compression.
        final ArrayList<String> tokens = new ArrayList<String>(16 + (len>>1)); // Assume ~2x fewer output tokens than compressed bytes.
        for(int i = 0; i < len; ++i)
            {
            final int ic = text[i];

            // A positive (7-bit) value is a literal/plain character.
            // It may or may not be part of a multi-byte token.
            if(ic >= 0)
                {
if(VERBOSE_DEBUG) { System.out.println("LITERAL "+ic+((ic >= 32) ? (" '"+((char)ic)+"'") : "")); }
                final char c = (char) ic;
                result.append(c);
                latestInput.append(c);
                continue;
                }

            // Tokenise any accumulated literal characters...
            if(latestInput.length() > 0)
                {
                final List<String> newTokens = tokenise(latestInput);
                // Handle the common case of a single new token efficiently.
                if(newTokens.size() == 1) { tokens.add(newTokens.get(0)); }
                else { tokens.addAll(newTokens); }
                // ...and clear the accumulated input now that it has been tokenised.
                latestInput.setLength(0);
                }

            // We never normally expect a direct reference to the previous token,
            // so this should indicate a multi-token match code.
            if(ALLOW_MULTI_TOKEN_MATCHING && (ic == -1))
                {
                if(i == len-1)
                    { throw new IllegalStateException("illegal multi-token escape at end of data"); }

                // We are decoding multiple tokens at once.
                // Hold on to your hats...
                final int ic2 = text[++i] | ~0xff; // Extend sign from logical 9th bit upwards...

                // Extract the encoded quantities...
                final int count = (ic2 & ~(~0 << TOKEN_SEQ_LEN_BITS)) + TOKEN_SEQ_LEN_MIN;
                final int offset = (ic2 >> TOKEN_SEQ_LEN_BITS) - 1;
if(VERBOSE_DEBUG) { System.out.println("DECODED SEQ COUNT, OFFSET, OPC: " + count + ", " + offset); }

                // Expand the token sequence...
                int index = tokens.size() + offset;
                for(int j = count; --j >= 0; )
                    {
                    final String token = tokens.get(index++);
                    result.append(token);
                    // And note that this is an input token too.
                    tokens.add(token);
                    }
                continue;
                }

            // Now replace the back-reference with the single token that it indicates...
            final int index = tokens.size() + ic;
            final String token;
            if(index < 0)
                {
                if((dict != null) && (index <= StaticDictionary.FIRST_VIRTUAL_OFFSET))
                    {
                    // This may be a reference to a token from the static dictionary.
                    final int vIndex = StaticDictionary.FIRST_VIRTUAL_OFFSET - index;
                    if(vIndex >= dict.tokens.size())
                        { throw new IllegalStateException("illegal back reference: ic="+ic+", dict.tokens.size()="+dict.tokens.size()); } 
                    token = dict.tokens.get(vIndex);
                    }
                else
                    { throw new IllegalStateException("illegal back reference: ic="+ic+", tokens.size()="+tokens.size()); }
                }
            else
                { token = tokens.get(index); }
if(VERBOSE_DEBUG) { System.out.println("BACK REF OFFSET "+ic+" = " + token); }
            result.append(token);
            // And note that this is an input token too.
            tokens.add(token);
            }

if(VERBOSE_DEBUG) { System.out.println("--- DECODE DONE"); }

        return(result.toString());
        }

    /**Basic (unshared) representation of 1 byte per character; never empty (is null instead).
     * Negative values are backward references by the specified number of tokens,
     * eg -2 means "insert a copy of the token two places back".
     * <p>
     * Written out in a customised low-overhead form.
     */
    private transient /* final */ byte[] text;
    
    /**Static dictionary (if any) used to aid compression; null if none.
     * It is forbidden to serialise an instance where this is non-null
     * so as to prevent decompression errors/ambiguitites.
     */
    private transient final StaticDictionary dict;
    
    /**Copy of the raw bytes in the internal representation; null if none.
     * The internal representation is subject to change,
     * but, with care, this can be used for extra-compact temporary storage.
     */
    public byte[] getInternalBytes() { return((text == null) ? null : text.clone()); }
    
    /**Make a Compact7BitString from the raw getInternalBytes() value; never null.
     * This is not guaranteed to work if getInternalBytes() is called in one release
     * and serialised in another, though this will not be broken casually.
     * <p>
     * The result is EMPTY or is intern()ed so as to avoid duplicates.
     * @param dict TODO
     */
    public static Compact7BitString makeFromInternalBytes(final byte raw[], final StaticDictionary dict)
        { return((raw == null) ? EMPTY : MemoryTools.intern(new Compact7BitString(raw, dict))); }

    /**Make simple un-tokenised byte[] representation.
     * The byte array must be pure 7-bit ASCII and
     * it must be guaranteed unshared so that we need not copy it for safety.
     * <p>
     * The empty string should be represented with a null array to save space,
     * not an empty array.
     */
    private Compact7BitString(final byte text[], final StaticDictionary dict)
        {
        this.text = text;
        this.dict = dict;

        // Verify object state.
        try { validateObject(); }
        catch(final InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage(), e); }
        }

    /**Equal iff the static dictionaries (if any) and compressed texts are equal. */
    @Override public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof Compact7BitString)) { return(false); }
        final Compact7BitString other = (Compact7BitString) obj;
        // Instances with different dictionaries are considered unequal.
        if(dict == null) { if(other.dict != null) { return(false); } }
        else if(!dict.equals(other.dict)) { return(false); }
        // Equal if the compressed format is identical.
        return(Arrays.equals(text, other.text));
        }

    /**Hash is zero for a zero-length (empty) text string.
     * This includes all bytes of the compressed text in the hash,
     * since that gives best overall performance in hash-based collections.
     */
    @Override public int hashCode()
        { return(Arrays.hashCode(text)); }

    /**Value representing an empty String. */
    public static final Compact7BitString EMPTY = new Compact7BitString(null, null);

    /**Deserialise. */
    private void readObject(final ObjectInputStream in)
        throws IOException, ClassNotFoundException
        {
        final byte l0 = in.readByte();
        final int length;
        if(l0 >= 0) { length = l0; }
        else
            {
            final byte l1 = in.readByte();
            length = -((l0 << 8) + (l1 & 0xff));
            }
        // Create and read the text field.
        if(length == 0) { text = null; }
        else
            {
            text = new byte[length];
            in.readFully(text);
            }

        // IMPORTANT:
        // We assume that the constructor will eventually be called
        // to do validation and any other setup required.
        assert(dict == null); // We must never deserialise with a dictionary.
        }


    /**Write out a minimally-redundant form of our internal information.
     * The more-efficient on-the-wire format also makes defensive
     * reading easier.
     * <p>
     * We don't write *any* default state.
     * <p>
     * We must never be asked to serialise if we have a (non-null) dictionary.
     * <p>
     * We must never be asked to serialise if the text is too long for us to encode.
     */
    private synchronized void writeObject(final ObjectOutputStream oos)
        throws IOException
        {
        assert (dict == null) : "we must not try to serialise if we are using a dictionary";
        assert ((text == null) || (text.length <= Short.MAX_VALUE)) : "we must not try to serialise a text too long for us to encode";

        // Write the length in a variable-length format.
        //
        // A null text is a zero-length text.
        if(null == text) { oos.writeByte(0); }
        // For very small strings use a byte-value length.            
        else if(text.length <= Byte.MAX_VALUE) { oos.writeByte(text.length); }
        // Write larger text sizes using a negative length
        // so that reading the first byte will give a negative value.
        else { oos.writeShort(-text.length); }

        // Now write the text bytes directly, unwrapped.
        if(null != text) { oos.write(text); }
        }

    /**Deserialise: use constructor for validation, defensive copying, etc. */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // As a minor optimisation, use EMPTY for any EMPTY value.
        if(equals(EMPTY)) { return(EMPTY); }

        // The text field is assumed guaranteed unshared already.
        return(MemoryTools.intern(new Compact7BitString(text, dict)));
        }
    
    /**Margin that byte[] must be smaller than String.length() for this to be smaller on the wire.
     * If we get this slightly too large then we will write more Strings than necessary
     * but may improve compression of multiple instances a little.
     * <p>
     * This value represents the overhead of serialising this instance
     * compared to the special-case String representation
     * plus some other overheads and potential loss of global compressability.
     */
    private static final int _FUDGE_FACTOR = 64;

    /**Serialise: write in the best format for the wire.
     * We may write ourself out as a java.lang.String
     * if we belive that would be more efficient.
     * <p>
     * We almost always write ourself out as a String
     * assuming that a good general-purpose compressor such as ZIP
     * will do better if we feed it the original text than our memory-efficient form,
     * but if we manage to find a lot of compressability
     * then we'll save ourself directly.
     * <p>
     * If a static dictionary is in use then we force serialisation as a String
     * for safety.
     */
    protected Object writeReplace()
        throws ObjectStreamException
        {
        // A zero-length text is best represented as an empty String.
        if(text == null) { return(""); }

        // If a static dictionary has been set, or
        // if we didn't manage to super-compress the input String
        // or if the compressed text is too long to represent on the wire
        // then serialise this instance as a String again rather than itself
        // since that is probably optimal (in stream UTF-8 format) already
        // especially when multiple instances are compressed in one stream
        // with a good general-purpose compressor.
        final String s = toString();
        final int sLen = s.length();
        if((null != dict) ||
           (text.length > Short.MAX_VALUE) ||
           (text.length + _FUDGE_FACTOR >= sLen) ||
           (text.length > (sLen>>>4)))
            { return(MemoryTools.intern(s)); }

if(IsDebug.isDebug) { System.out.println("[C7BS: serialising as self: text.length="+text.length+", toString().length()="+toString().length()+".]"); }

        // Write this instance out as itself.
        return(this);
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // We should use a null array rather than an empty one.
        if((text != null) && (text.length == 0))
            { throw new InvalidObjectException("bad object: text.length == 0"); }
        if((text != null) && (text.length > Short.MAX_VALUE))
            { throw new InvalidObjectException("bad object: text.length > Short.MAX_VALUE"); }

        // TODO: validate content
        }

    /**Unique serialisation ID. */
    private static final long serialVersionUID = -5836076660507008468L;

    /**Our tokenising pattern; never null.
     * This accepts either a "word" or a non-word,
     * the idea is to compress an alternating sequence of such non-empty tokens.
     * <p>
     * The word characters are [a-zA-Z_0-9].
     * The non-word characters are everything else in the [0,127] 7-bit-ASCII alphabet.
     * <p>
     * These tokens need not be strictly alternating
     * provided thst ambiguity can be avoided.
     * <p>
     * These tokens are currently not limited in length.
     * <p>
     * Inspired by the HuffWord compression tokeniser from "Managing Gigabytes".
     */
//    private static final Pattern tokeniser = Pattern.compile("([\\w]+)|([\\W]+)");
    
//    /**If true, use regex-based tokenisation, else use hand-crafted solution. */
//    private static final boolean USE_REGEX_TOKENISER = false;

    /**Boolean true/false flags for what are considered "word" chars; initialised on first use.
     * Also holds intern()ed one-byte Strings for each token
     * to avoid run-time allocation similar in intent to Integer.valueOf(int).
     * <p>
     * The word characters are [a-zA-Z_0-9].
     * The non-word characters are everything else in the [0,127] 7-bit-ASCII alphabet.
     * <p>
     * These tokens need not be strictly alternating
     * provided thst ambiguity can be avoided.
     * <p>
     * These tokens are currently not limited in length.
     * <p>
     * Inspired by the HuffWord compression tokeniser from "Managing Gigabytes".
     */
    private static final class WordChars
        {
        private WordChars(){} // Prevent construction.

        /**An entry is true if the corresponding ASCII char of the index is a "word" char. */
        static final boolean isWordChar[] = new boolean[Byte.MAX_VALUE+1];
        static
            {
            Arrays.fill(isWordChar, '0', '9'+1, true);
            Arrays.fill(isWordChar, 'A', 'Z'+1, true);
            Arrays.fill(isWordChar, 'a', 'z'+1, true);
            isWordChar['_'] = true;
            }

        /**One-byte intern()ed String for this char. */
        static final String s[] = new String[Byte.MAX_VALUE+1];
        static
            {
            for(int i = Byte.MAX_VALUE+1; --i >= 0; )
                { s[i] = String.valueOf((char) i).intern(); }
            }
        }

    /**Routine to chop the (non-empty) input CharSequence as an in-order List of String tokens; result may be empty but is never null.
     * The toString() method may be used to convert all if the input (or subsequences) to a String,
     * so must behave well.
     * <p>
     * The returned list has fast random access, ie with get(index).
     */
    private static List<String> tokenise(final CharSequence in)
        {
        final int len = in.length();
        // Handle small inputs specially:
        //   * Empty input is (would be) zero tokens.
        //   * Single byte input is a single token.
        if(len < 2)
            {
            assert(len == 1);
//            if(len == 0)
//                {
//                final List<String> result = Collections.emptyList();
//                return(result);               
//                }
            final List<String> result = Collections.singletonList(in.toString());
            return(result);
            }

        final ArrayList<String> tokens = new ArrayList<String>(2 + (len>>2));

//        if(USE_REGEX_TOKENISER)
//            {
//            // Use the simple but slow regex mechanism...
//            final Matcher matcher = tokeniser.matcher(in);
//            while(matcher.find())
//                { tokens.add(matcher.group()); }
//            return(tokens);
//            }

        int currentTokenStart = 0; // Start of current token...
        boolean currentTokenIsWord = WordChars.isWordChar[in.charAt(0)];
        for(int i = 0; ++i < len; )
            {
            final boolean thisIsWordChar = WordChars.isWordChar[in.charAt(i)];
            if(thisIsWordChar != currentTokenIsWord)
                {
                // We just started a new token.
                // Handle common single-character token case efficiently (no allocation).
                if(i == currentTokenStart+1)
                    { tokens.add(WordChars.s[in.charAt(currentTokenStart)]); }
                else // Generic (length != 1) case.
                    { tokens.add(in.subSequence(currentTokenStart, i).toString()); }
                // Note details of new token.
                currentTokenStart = i;
                currentTokenIsWord = thisIsWordChar;
                }
            }
        // Create/add the trailing token,
        // making sure that the common single-char case is efficient.
        if(len == currentTokenStart+1)
            { tokens.add(WordChars.s[in.charAt(currentTokenStart)]); }
        else // Generic (length != 1) case.
            { tokens.add(in.subSequence(currentTokenStart, len).toString()); }
        return(tokens);
        }
    
    /**If true, then don't try to match immediate previous input token.
     * This is usually not going to be effective
     * as unless we split tokensw we can never match the previous value
     * as we alternate between word and non-word.
     * So not trying to match saves a little bit of time.
     * <p>
     * This also frees up the "-1" offset value as an escape,
     * eg for multi-token-match expansion.
     */
    private static final boolean DONT_TRY_TO_MATCH_PREV_TOKEN = true;

    /**If true then allow multi-token matches (MTM) for better compression of very repetitive text.
     * This is for where a number of tokens in sequence are identical
     * to some previous (near) token sequence.
     */
    private static final boolean ALLOW_MULTI_TOKEN_MATCHING = DONT_TRY_TO_MATCH_PREV_TOKEN;

    /**If true, allow us to do include 1-char tokens in multi-token matches (MTMs).
     * This should help maximise the compression from multi-token matches,
     * since many possible multi-token sequences will start with
     * a single-character token, such as " ".
     */
    private static final boolean ALLOW_1_CHAR_MTM = true;

    /**If true, allow partial (prefix) token matching for long (semi-)unique tokens.
     * Can be slow, but allows a little more compression.
     */
    private static final boolean ALLOW_PARTIAL_TOKEN_MATCHING = true;

    /**Bits that we use to represent a token sequence length; strictly positive and less than 8. */
    private static final int TOKEN_SEQ_LEN_BITS = 3;

    /**Minimum number of sequential tokens that we can encode as a sequence; strictly positive.
     * Since the minimum encoding is usually 2 bytes
     * then we set a minimum sequence length longer than this
     * to guarantee to do better than just a sequence of separate back-references.
     */
    private static final int TOKEN_SEQ_LEN_MIN = 3;

    /**Maximum number of sequential tokens that we can encode as a sequence; strictly positive.
     * Note that we store the sequence length - 2,
     * since we would use the normal back-reference for a single-token match,
     * so we use a zero value to indicate a length-of-two sequence.
     */
    private static final int TOKEN_SEQ_LEN_MAX = (1 << TOKEN_SEQ_LEN_BITS) - 1 + TOKEN_SEQ_LEN_MIN;

    /**Maximally-negative offset that we can hold above the sequence length in one byte; strictly negative.
     * We have an implict logical 9th sign bit, always 1 (ie negative).
     */
    private static final int TOKEN_SEQ_OFFSET_MIN = (~0xff) >> TOKEN_SEQ_LEN_BITS; // Use sign-preserving right shift.

    /**Losslessly convert a String entirely consisting of 7-bit ASCII text to a tokenised form for compression/serialisation; not null unless the input is null.
     * This should be especially effective for XML, but also OK for general text.
     * <p>
     * This is designed to be relatively fast and consume little memory.
     * <p>
     * Use of a good shared static dictionary between multiple instances
     * may significantly reduce memory consumption,
     * but it is not permitted to serialised instance using such a dictionary.
     * 
     * @param s  the 7-bit ASCII non-empty value to encode 
     * @param dict  static compression dictionary to use, or null if none
     *
     * @throws  IllegalArgumentException  if the argument is not 7-bit ASCII
     *     ie contains characters outside the range 0--127
     */
    public static final Compact7BitString convertToCompact7BitString(final CharSequence s,
                                                                     final StaticDictionary dict)
        {
        if(s == null) { return(null); }

        // In the simple, zero-length, case we return the EMPTY value.
        final int length = s.length();
        if(length == 0) { return(EMPTY); }

        // Verify that the input is pure 7-bit.
        // Veto conversion with IllegalArgumentException if not.
        for(int i = length; --i >= 0; )
            {
            final char c = s.charAt(i);
            if(c > 127) { throw new IllegalArgumentException("not pure 7-bit ASCII text"); }
            }

        // Tokenise the input.
        final List<String> tokensRaw = tokenise(s);
        assert(tokensRaw instanceof RandomAccess); // We need fast random access into the list.
        final List<String> tokens;
        // Without a dictionary we use the tokens list from the tokeniser as-is.
        if(dict == null)
            { tokens = tokensRaw; }
        // With a dictionary, we allow -ve get()s into the static tokens.
        else
            {
            final class VirtualStringList extends AbstractList<String> implements RandomAccess
                {
                @Override public final String get(final int index)
                    {
                    // In the simple +ve index case, fetch from the underlying List.
                    if(index >= 0) { return(tokensRaw.get(index)); }
                    // Deal with -ve references into the static dictionary.
                    if(index <= StaticDictionary.FIRST_VIRTUAL_OFFSET)
                        {
                        final int vIndex = StaticDictionary.FIRST_VIRTUAL_OFFSET - index;
                        if(vIndex < dict.tokens.size())
                            { return(dict.tokens.get(vIndex)); }
                        throw new IndexOutOfBoundsException("illegal back reference: " + index);
                        }
                    return(""); // Deal with "gap" at -1 by returning an impossible token...
                    }

                @Override public final void add(int index, String element)
                    { tokensRaw.add(index, element); }

                @Override public final String set(int index, String element)
                    { return(tokensRaw.set(index, element)); }

                /**Size is reported as that of the underlying "raw" List; -ve indexed items are not counted. */
                @Override public final int size() { return(tokensRaw.size()); }
                }
            tokens = new VirtualStringList();
            }

        // Map of all tokens seen so far to last position they were seen in.
        // All tokens here will be >1 char, unless ALLOW_1_CHAR_MTM is true.
        final HashMap<String,Integer> tokensSeenSoFar = new HashMap<String,Integer>(256 + 2*tokens.size());

        // If we have a static dictionary then insert its tokens now.
        if(dict != null) { tokensSeenSoFar.putAll(dict.virtualOffsets); }

        // Generate the output.
        final StringBuilder sb = new StringBuilder(s.length());
        main: for(int i = 0; i < tokens.size(); ++i)
            {
            final String token = tokens.get(i);
            final int tokenLength = token.length();
            assert(tokenLength > 0);

            // We keep this single-character token as-is
            // (ie as an uncompressed literal byte)
            // unless we are potentially allowing this as
            // the start of a multi-token match.
            if(!ALLOW_1_CHAR_MTM && (tokenLength == 1))
                {
if(VERBOSE_DEBUG) { System.out.println("LITERAL "+((int) token.charAt(0))); }
                tokensSeenSoFar.put(token, Integer.valueOf(i));
                sb.append(token);
                continue;                
                }

            // Current input token-list size.
            final int nTokensNow = tokens.size();

            // Look for the nearest match backwards in the valid match window.
            // But only look if we know that the token has already been seen so far.
            final Integer lastPos = tokensSeenSoFar.get(token);
            if(lastPos != null)
                {
                final int lp = lastPos.intValue();
                assert(token.equals(tokens.get(lp)));
                final int offset = lp - i;
                // Ignore matches that we cannot use.
                if((offset >= Byte.MIN_VALUE) &&
                   (!(DONT_TRY_TO_MATCH_PREV_TOKEN && (offset == -1))))
                    {
                    assert(token.equals(tokens.get(i+offset)));

                    if(ALLOW_MULTI_TOKEN_MATCHING)
                        {
                        // Look FORWARD and count how many sequential tokens match from here
                        // (from index j inclusive up to index i exclusive)
                        // which represent tokens already encoded/sent.
                        // We may then be able to encode them all in one fell swoop...
                        int count = 1;
                        final int endStop = nTokensNow-1 + offset; // Can't search beyond end of the input!
                        for(int k = lp; ++k <= endStop; )
                            {
                            // Stop if the sequence stops matching...
                            if(!tokens.get(k).equals(tokens.get(k-offset)))
                                { break; }
                            // Stop if we reach the maximum sequence length that we can handle.
                            if(++count == TOKEN_SEQ_LEN_MAX)
                                { break; }
                            }
                        assert((count > 0) && (count <= TOKEN_SEQ_LEN_MAX));
                        // Handle a long-enough-to-be-worthwhile sequence,
                        // but only if it is within range...
                        if(count >= TOKEN_SEQ_LEN_MIN)
                            {
                            final int normalisedOffset = offset + 1; // Offset of -1 is never used.
                            if(normalisedOffset >= TOKEN_SEQ_OFFSET_MIN)
                                {
if(VERBOSE_DEBUG) { System.out.println("ENCODABLE SEQ COUNT, OFFSET, SEQ: " + count + ", " + offset + ", " + tokens.subList(i, i+count)); }
                                sb.append((char) -1); // Escape code.
                                final int normalisedCount = (count - TOKEN_SEQ_LEN_MIN);
                                assert(normalisedCount < (1 << TOKEN_SEQ_LEN_BITS));
                                final int encoded = (normalisedOffset << TOKEN_SEQ_LEN_BITS) + normalisedCount;
                                sb.append((char) encoded); // Only the bottom byte is important.

                                // Note the first token processed.
                                tokensSeenSoFar.put(token, Integer.valueOf(i));

                                // Skip an extra count-1 tokens that we're encoding...
                                // Note all the (> length 1) tokens that we're skipping...
                                for(int k = count; --k > 0; )
                                    {
                                    final String skippedToken = tokens.get(++i);
                                    if(ALLOW_1_CHAR_MTM || (skippedToken.length() > 1))
                                        { tokensSeenSoFar.put(skippedToken, Integer.valueOf(i)); }
                                    }

                                continue main;
                                }
                           }
                        // Fall through to write normal single-token back-reference.

                        // If the token length is 1
                        // then we prefer to write the literal over the back-ref
                        // though it does not matter in fact.
                        if(tokenLength == 1)
                            {
if(VERBOSE_DEBUG) { System.out.println("LITERAL "+((int) token.charAt(0))); }
                            tokensSeenSoFar.put(token, Integer.valueOf(i));
                            sb.append(token);
                            continue;      
                            }

                        // We have a match to a single previous nearby token...
                        // So write a single-byte relative back-reference rather than the token.
if(VERBOSE_DEBUG) { System.out.println("BACK REF OFFSET "+offset+" = " + token); }
                        tokensSeenSoFar.put(token, Integer.valueOf(i));
                        sb.append((char) offset); // Only care about bottom byte...
                        continue main;
                        }
                    }
                }

            // If we could not find a suitable full-token match
            // then we may be able to find a partial match instead
            // (providing that this token is long enough to split in two).
            if(ALLOW_PARTIAL_TOKEN_MATCHING && (tokenLength > 2))
                {
                // We prefer to match the longest-possible prefix
                // to try to maximise compression.
                for(int prefixLength = tokenLength; --prefixLength > 1; )
                    {
                    final String prefix = token.substring(0, prefixLength);
                    final Integer lastPrefixPos = tokensSeenSoFar.get(prefix);
                    if(lastPrefixPos == null) { continue; }
                    final int lpp = lastPrefixPos.intValue();
                    assert(prefix.equals(tokens.get(lpp)));
                    final int offset = lpp - i;
                    // Skip matches that we cannot use.
                    if(offset < Byte.MIN_VALUE) { continue; }
                    if(DONT_TRY_TO_MATCH_PREV_TOKEN && (offset == -1)) { continue; }
                    assert(prefix.equals(tokens.get(i+offset)));
                    // We veto this prefix match (and of all other prefixes of this token)
                    // if the whole token may be the target of a back-reference later.
                    if(tokens.subList(i+1, Math.min(nTokensNow, i-Byte.MIN_VALUE)).contains(token)) { break; }
//System.out.println("Prefix match allowed, length: " + prefix.length());
                    // Adjust tokens list with new split token.
                    tokens.set(i, prefix);
                    tokens.add(i+1, token.substring(prefixLength));
if(VERBOSE_DEBUG) { System.out.println("BACK (PREFIX) REF OFFSET "+offset+" = " + token); }
                    tokensSeenSoFar.put(prefix, Integer.valueOf(i));
                    sb.append((char) offset); // Only care about bottom byte...
                    continue main;
                    }
                }

            // Log this uncompressed >1-char early token
            // if the first occurrence for this compression/input
            // if compiling global stats
            // and if this is close to the start of the input.
            if(LOG_STATS && (dict != null) &&
                (tokenLength > 1) && (i < -Byte.MIN_VALUE) &&
                (null == tokensSeenSoFar.get(token)))
                {
                // Create a counter for this token if need be.
                dict._uncompEarlyTokens.putIfAbsent(token, new Tuple.Pair<AtomicInteger,AtomicInteger>(new AtomicInteger(),new AtomicInteger()));
                // Atomically increment the counter for this token.
                final Pair<AtomicInteger, AtomicInteger> pair = dict._uncompEarlyTokens.get(token);
                final int tc = pair.first.incrementAndGet();
                // Update the summed earliest-position value...
                pair.second.addAndGet(i);
if(VERBOSE_DEBUG && ((tc % 1000) == 0)) { System.out.println("[Token count for '"+token+"': "+tc+".]"); }
                }

            // Now now update tokensSeenSoFar with this token.
            tokensSeenSoFar.put(token, Integer.valueOf(i));
            // Did not find a (near enough) match, so write token as-is.
            sb.append(token);
            }
//System.out.println("[Length in (chars) / out (bytes): "+s.length()+"/"+sb.length()+".]");

        // Convert back to byte[].
        final byte[] text = new byte[sb.length()];
        for(int i = text.length; --i >= 0; )
            { text[i] = (byte) sb.charAt(i); }

if(VERBOSE_DEBUG) { System.out.println("--- ENCODE DONE"); }
        final Compact7BitString result = new Compact7BitString(text, dict);
//        assert(s.equals(result.toString())) : "Generated bad compressed form: '"+result+"' from input: '"+s+"'";
        // De-dup this instance with intern().
        return(MemoryTools.intern(result));
        }
    }