/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

/**This holds fundamental compile-time constants for all parts of PG2K.
 */
public final class CoreConsts
    {
    /**Default name by which whole app is known internally. */
    public static final String APP_DEFAULT_NAME = "PG2K";

    /**Role name (user/group) to for most priv actions. */
    public static final String PRIV_ROLE_NAME = "admin";

    /**8-bit ISO-8859-1 file encoding; our default throughout PG2K. */
    public static final String FILE_ENCODING_8859_1 = "ISO-8859-1";

    /**8-bit UTF-8 file encoding. */
    public static final String FILE_ENCODING_UTF_8 = "UTF-8";

    /**Name of main exhibit and catalogue-page server host(s); generally sessionless.
     * This may be an alias to one or more hosts.
     * <p>
     * This is rendered all lower-case to slightly improve compressability
     * in certain contexts, and is also prefixless (eg no "www.")
     * to be in canonical form.
     */
    public static final String MAIN_DATA_HOST = "gallery.hd.org";

    /**Default temporal slackness in seconds; strictly positive.
     * This is the default slackness or inconsistency that we will
     * allow in the system in order to gain efficiency.
     * <p>
     * This can be used, for example, as a default time for
     * which exhibits and catalogue pages may be cached by a
     * client, and that we might internally cache exhibit
     * properties, etc.
     * <p>
     * Note that it may sometimes be necessary to ensure that
     * multiples of this interval cannot be catenated in bad
     * cases.
     */
    public static final int DEFAULT_TEMPORAL_SLACKNESS_S = 3600;

    /**Approximate start date of the Gallery, primarily for validating timestamps, etc; strictly positive.
     * No exhibit timestamp or accession date should be older than this
     * (nor newer than "now" allowing for clock skew).
     * <p>
     * Other dates may be validated with reference this value.
     * <p>
     * It is probably accurate enough to use in synthetic copyright notices too.
     * <p>
     * The Gallery started in 1996/1997.
     * <p>
     * For reference:
     * <ul>
     * <li>1996/01/01 00:00 UTC = 820454400000.  Calculated with (new Date(96,0,1,0,0)).getTime().
     * <li>1997/01/01 00:00 UTC = 852076800000.  Calculated with (new Date(97,0,1,0,0)).getTime().
     * </ul>
     */
    public static final long GALLERY_EPOC_START = 820454400000L; // 1996/01/01 00:00 UTC
    //static { System.err.println("GALLERY_EPOC_START = " + GALLERY_EPOC_START); }

    /**java:comp/env usual JNDI namespace prefix for EJBs... */
    public static final String JNDI_COMPENV_PREFIX = "java:comp/env/";

    /**Short JNDI name by which ExhibitFacade front-end (session) bean is accessed. */
    public static final String JNDI_SHORT_NAME_ExhibitFacade = "ejb/ExhibitFacade";

    /**Short JNDI name by which PropsFacade front-end (session) bean is accessed. */
    public static final String JNDI_SHORT_NAME_PropsFacade = "ejb/PropsFacade";



    /**Default dase dir for exhibit data load from filesystem. */
    public static final String FS_DATA_ROOT = "fs_data";

    /**Default base dir for meta-data load from filesystem. */
    public static final String FS_CONF_ROOT = "fs_conf";

    /**Filesystem path below conf root for load/save of global system properties. */
    public static final String FS_CONF_SYSPROPS = "sys.props";

    /**Filesystem path below conf root for load/save of private security properties. */
    public static final String FS_CONF_SECPROPS = "security.props";

//    /**FIXME: REMOVE: Name of pre-PG2K location database (properties) file; only file component. */
//    public static final String FS_CONF_PREPG2K_LOCDB_FILENAME = "auxInfo.txt";

    /**Name of static properties cache file usually within data directory.
     * For most purposes this is read-only,
     * but some applications are able to write to it to atomically update it.
     */
    public static final String FS_DATA_CACHE_FILENAME = "_expropsCache.dat";

    /**Immutable Set of all configuration filenames (as String values). */
    public static final Set<String> ALL_FS_CONF_FILENAMES;
    static
        {
        // Provide a small but perfectly formed HashSet for quick access.
        final Set<String> afcf = new HashSet<String>(7);
        // Manually add in all the names here.
        afcf.add(FS_CONF_SYSPROPS);
        afcf.add(FS_CONF_SECPROPS);
        // Make an unmodifiable version available...
        ALL_FS_CONF_FILENAMES = Collections.unmodifiableSet(afcf);
        }

    /**Name of properties class/file on classpath containing simple global properties.
     * This does not include a file suffix.
     */
    public static final String GLOBAL_PROPS_NAME = "Global";

    /**Name of properties class/file on classpath or in exhibits area containing built-in location-map data.
     * This does not include a file suffix.
     */
    public static final String LOCDB_PROPS_NAME = "locationDB";

    /**Name of WAR context-param boolean flag to set cache aggressiveness. */
    public static final String WAR_CTXTPARAM_AGGRESSIVE_CACHE = "aggressiveCacheing";

    /**Name of WAR context-param `master/slave/etc' flag. */
    public static final String WAR_CTXTPARAM_DISPOSITION = "disposition";

    // Valid disposition names.
    /**Master (J2EE) taking data from EJBs (and probably not with / context root). */
    public static final String WAR_CTXTPARAM_DISPOSITION_MASTER = "master";
    /**Master (pure WAR) taking data from filesystem, with / context root. */
    public static final String WAR_CTXTPARAM_DISPOSITION_WARONLY = "waronly";
    /**Slave (pure WAR) taking data over HTTP tunnel, with / context root. */
    public static final String WAR_CTXTPARAM_DISPOSITION_SLAVE = "slave";

//    /**Name of system property for ``waronly'' indicating root of (read-only) config info. */
//    public static final String WAR_SYSPROPNAME_WARONLY_CONFIGROOT = "org.hd.d.pg2k.WAR.waronly.configroot";
//    /**Name of system property for ``waronly'' indicating root of (read-only) exhibit data. */
//    public static final String WAR_SYSPROPNAME_WARONLY_EXHIBITROOT = "org.hd.d.pg2k.WAR.waronly.exhibitdataroot";
//    /**Name of system property for ``waronly'' indicating location of master-readable serialised, compressed, exhibit properties object. */
//    public static final String WAR_SYSPROPNAME_WARONLY_STATICCACHEFILE = "org.hd.d.pg2k.WAR.waronly.staticcachefile";
//    /**Name of system property for ``waronly'' indicating root of (read-write) meta-data cache. */
//    public static final String WAR_SYSPROPNAME_WARONLY_RWMETADATAROOT = "org.hd.d.pg2k.WAR.waronly.metadatacacheroot";
//    /**Name of system property for ``waronly'' indicating root of (read-write) log area. */
//    public static final String WAR_SYSPROPNAME_WARONLY_LOGROOT = "org.hd.d.pg2k.WAR.waronly.logroot";

//    /**Name of system property for ``waronly'' indicating the comma-separated list of properties files from which global exhibit data may be extracted. */
//    public static final String WAR_SYSPROPNAME_WARONLY_EXGLOBDATA = "org.hd.d.pg2k.WAR.waronly.exglobdata";


    /**The WAR context-param `master/slave' flag can be set early with this named system property.
     * This allows the WAR front-end to start-up faster in some circumstances,
     * but if this is not set, or not set to a usable value, the WAR front-end
     * will resort to its default method of determining master/slave disposition.
     * <p>
     * The property value must not <em>conflict</em> with that deduced my the
     * default methods, if this property is defined at all.
     */
    public static final String WAR_SYSPROPNAME_DISPOSITION_PRESET = "org.hd.d.pg2k.WAR.disposition";


    /**Name of WAR context-param boot-server URL. */
    public static final String WAR_CTXTPARAM_BOOTURL = "bootServerURL";



    /**Our preferred simple date formater for all locales; yyyy/MM/dd Japanese style.
     * FIXME: note that SimpleDateFormat is not immutable.
     */
    public static final SimpleDateFormat DEFAULT_DATE_FORMAT = new SimpleDateFormat("yyyy/MM/dd");

    /**Maximum length of description entry (or entries) in characters. */
    public static final int DESCRIPTION_MAX_CHARS = 2048;

    /**Description suffix on full path name of exhibit. */
    public static final String DESCRIPTION_FILE_SUFFIX = ".txt";

    /**Minimum password length for all (eg Web) logins; strictly positive. */
    public static final int MIN_PASSWORD_LEN = 4;
    /**Maximum password length for all (eg Web) logins; strictly positive. */
    public static final int MAX_PASSWORD_LEN = 16;

    /**Time short enough not to be noticable to users in GUI interactions.
     * Users are said not to notice delays of 100ms or less.
     */
    public static final int MAX_INTERACTIVE_DELAY_MS = 100;


    /**Checksum/hash for data to correspond to use in tools such as [G]ZIP; not "secure".
     * Not a secure hash,
     * but used "as if" in some non-critical places as a quick cross-check,
     * and used in ZIP archive so quick to compare against ZIP entries.
     */
    public static final String HASH_CRC32 = "CRC32";

    /**Preferred hash for data integrity; reasonably "secure". */
    public static final String HASH_SHA1 = "SHA1";

    /**Alternative fast hash for data integrity; reasonably "secure".
     * Not as secure as SHA-1, but possibly a good alternative/cross-check,
     * and commonly used for simple integrity checking,
     * eg with the HTTP Content-MD5 header.
     */
    public static final String HASH_MD5 = "MD5";

    /**HMAC authentication hash algorithm to protect data in transit.
     * Although we use a MAC primarily to guard against data corruption on the wire,
     * we also want to provide some protection against tampering and replay attacks,
     * eg someone trying to slip p0rn or warez or trojans into the datastream to a slave.
     * <p>
     * Thus we want a strong-ish hash that is relatively collision-free,
     * though hopefully not too expensive to compute
     * nor contributing too much message overhead (ie too large a hash/MAC).
     * <p>
     * Note that our key generation and distribution is not necessarily very robust,
     * ie anyone with access to any of the servers as root could steal the key,
     * and thus tamper/insert to their heart's content, so we should not go overboard.
     * This mechanism should prevent corruption and casual tampering by ISPs,
     * transparent proxies, broken routers, and others that can play "man-in-the-middle",
     * but without admin access to any servers.
     * Note that this does not prevent any "man-in-the-middle" from
     * <em>observing</em> all our traffic; we only need authetication, not encryption.
     * <p>
     * Note that not all communications need to be guarded by this MAC;
     * in particular read-only operations by lightweight clients may not be
     * (at least in part so that we do not need to distribute the secret key to them)
     * which may save some CPU cycles and a little bandwidth.
     * In particular we may respond without a MAC to a request without a (valid) MAC,
     * which may also make it harder for an imposter to probe for our key.
     * <p>
     * The algorithm must be supported by the JDK or a plugged-in provider.
     * <p>
     * MD5 and SHA-1 are showing signs of age and weakness,
     * so we need to use something of at least SHA-256 (256-bit) strength.
     * All implementations wishing to communicate must agree on this algorithm.
     */
    public static final String HMAC_ALG = "HmacSHA256";


    /**Max age of an exhibit to be considered "new", in seconds; strictly positive.
     * A value of a few weeks to a few months is probably appropriate
     * given observed visitor frequency and upload rate of new exhibits.
     */
    public static final int MAX_EXHIBIT_AGE_NEW_S = 183 * 24 * 3600; // 6 months.

    /**Max age of an exhibit to be considered "new", in milliseconds; strictly positive.
     * A value of a few weeks to a few months is probably appropriate
     * given observed visitor frequency and upload rate of new exhibits.
     */
    public static final long MAX_EXHIBIT_AGE_NEW_MS = MAX_EXHIBIT_AGE_NEW_S * 1000L;


    /**If true, avoid using connection and read timeouts on sockets.
     * Under JDK 1.5.0 (and probably 1.4.x), using read and connection timeouts
     * forces the VM to use non-blocking I/O which in turn can lead to fd leaks
     * and other problems.
     */
    public static final boolean AVOID_UNSAFE_TCP_TIMEOUTS = false;

    /**Suggested transfer size of exhibit data (and some other bulk data).
     * We will usually copy (exhibit) data around in chunks of about this size
     * (or a small power-of-2 multiple of this size).
     * This applies to transfer to/from local media,
     * such as disc, and over slow (eg network) media.
     * <p>
     * This is large enough to be efficient and to allow the OS
     * to be buffering up data for good throughput over TCP/IP for example,
     * but small enough not to require stupid amounts of memory or time
     * (and in particular to retain reasonable interactivity).
     * <p>
     * A power of two for efficiency in various places.
     * <p>
     * A size of about 1kByte to 64kBytes is probably reasonable;
     * 16kB or less seems good to retain interactivity.
     */
    public static final int BULK_DATA_TRANSFER_SIZE = (1<<14);

    /**Do-nothing page root-relative URL that will accept any parameters and does no work. */
    public static final String DO_NOTHING_PAGE_RRURL = "/null.jsp";

    /**Tunnel URL on server. */
    public static final String TUNNEL_URI = "/_tunnel";


    /**GenProps debug flag tree root, including trailing dot. */
    public static final String GEN_PREFIX_debugFlag = "debugFlag.";

    /**Maximum permissable clock skew between any two PG2K server instances, ms; strictly positive.
     * Of the order of a few minutes is good.
     */
    public static final int MAX_PEER_CLOCK_SKEW_MS = 3 * 60 * 1000;
    }
