/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.SequenceInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

/**Like GZIPInputStream but no GZIP or ZLIB header or checksum.
 * This is meant to give maximum compression and assumes error
 * checking/recovery is done elsewhere.
 */
public final class DefInputStream extends InflaterInputStream
    {
    /**Indicates end of input stream. */
    private boolean eos;

    /**Is the stream already closed? */
    private boolean closed = false;

    /**Check to make sure that this stream has not been closed.
     */
    private synchronized void ensureOpen()
        throws IOException
        { if(closed) { throw new IOException("already closed"); } }

    /**Creates a new input stream with the specified buffer size.
     * In cases where we know that the input stream has one or more
     * trailing bytes that we don't care about
     * then the hasDummyTrailingData arg can be true
     * to save us having to wrap the input stream.
     *
     * @param in     the input stream
     * @param hasDummyTrailingByte  if true then the input stream has
     *     one or more bytes of trailing data that ZLIB can eat harmlessly
     * @exception    IOException if an I/O error has occurred
     */
    public DefInputStream(final InputStream in,
                          final boolean hasDummyTrailingByte)
        throws IOException
        {
        super(hasDummyTrailingByte ? in : new SequenceInputStream(in, new ByteArrayInputStream(new byte[1])),
              new Inflater(true),
              512); // Standard deflater buffer size is fine...
        }

    /**Creates a new input stream with a default buffer size.
     * @param in   the input stream
     * @exception  IOException if an I/O error has occurred
     */
    public DefInputStream(InputStream in)
        throws IOException
        { this(in, false); }

    /**Reads uncompressed data into an array of bytes.
     * Blocks until enough input is available for decompression.
     *
     * @return  the actual number of bytes read, or -1 if the end of the
     *          compressed input stream is reached
     * @exception IOException  if an I/O error has occurred or the compressed
     *     input data is corrupt
     */
    public synchronized int read(byte[] buf, int off, int len)
        throws IOException
        {
        ensureOpen();
        if(eos) { return(-1); }
        len = super.read(buf, off, len);
        if(len == -1) { eos = true; }
        return(len);
        }

    /**Closes the input stream and releases non-Java resources.
     * @exception IOException  if an I/O error has occurred
     */
    public synchronized void close()
        throws IOException
        {
        inf.end();
        in.close();
        eos = true;
        closed = true;
        }
    }
