/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

/**Like GZIPOutputStream but no GZIP header or checksum.
 * This is meant to give maximum compression and assumes error
 * checking/recovery is done elsewhere.
 */
public final class DefOutputStream extends DeflaterOutputStream
    {
    /**Creates a new output stream with the specified buffer size.
     * Forces use of the best possible compression at the possible
     * expense of CPU time.
     * <p>
     * This compressed stream is wrapped with neither GZIP nor ZLIB
     * headers/checksums to minimise overheads; we'd better be doing
     * error detection and correction elsewhere!
     *
     * @param out the output stream
     * @param size the output buffer size;
     *     changing this from the default of 512 bytes is unlikely
     *     to improve performance
     */
    private DefOutputStream(final OutputStream out, final int size)
        // throws IOException
        { super(out, new Deflater(Deflater.BEST_COMPRESSION, true), size); }

    /**Creates a new output stream with a default buffer size.
     * @param out the output stream
     */
    public DefOutputStream(final OutputStream out)
        // throws IOException
        { this(out, 512); }

    /**Writes array of bytes to the compressed output stream; blocking.
     * @exception IOException   if an I/O error has occurred
     */
    public synchronized void write(byte[] buf, int off, int len)
        throws IOException
        { super.write(buf, off, len); }

    /**Finishes writing compressed data to the output stream without closing it.
     * @exception IOException   if an I/O error has occurred
     */
    public synchronized void finish()
        throws IOException
        {
        if(!def.finished())
            {
            def.finish();
            while(!def.finished()) { deflate(); }
            }
        }

    /**Writes remaining compressed data to the output stream and closes it.
     * @exception IOException   if an I/O error has occurred
     */
    public synchronized void close()
        throws IOException
        {
        finish();
        out.close();
        }
    }
