/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import ORG.hd.d.IsDebug;

import java.io.*;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.util.*;


/**Basic utilities for retrieving and handling exhibit attribute/discardable words.
 */
public final class ExhibitAttrUtils
    {
    /**Prevent instantiation. */
    private ExhibitAttrUtils() { }

    /**Name of attribute words list in global properties. */
    public static final String ATTR_WORD_LIST_NAME = "org.hd.d.pg.nameDiscardableWords";

    /**Gets unsorted list of attribute words from global properties; never null.
     * This returns a list of valid unduplicated attribute words,
     * with any problems found being reported on System.err.
     * <p>
     * This may return an empty list but not a null.
     */
    private static String[] _getAttributeWords()
        {
        try
            {
            final ResourceBundle gp = ResourceBundle.getBundle(CoreConsts.GLOBAL_PROPS_NAME);
            final String words = gp.getString(ATTR_WORD_LIST_NAME);

            final StringTokenizer wordsList = new StringTokenizer(words);
            final HashSet<String> theseWords = new HashSet<String>(wordsList.countTokens() * 2 + 1);
            while(wordsList.hasMoreTokens())
                {
                final String w = wordsList.nextToken();
                if(!ExhibitName.validAttributeWord(w))
                    {
                    System.err.println("WARNING: ExhibitAttrUtils: illegal attribute word \"" + w + "\" in properties " + CoreConsts.GLOBAL_PROPS_NAME + ", property name " + ATTR_WORD_LIST_NAME + ".");
                    continue; // Do not include the word.
                    }
                // Eliminate duplicates.
                if(!theseWords.contains(w))
                    { theseWords.add(MemoryTools.intern(w)); }
                }

            // Extract and return the (unsorted) set/list of attribute words.
            final String result[] = new String[theseWords.size()];
            theseWords.toArray(result);
if(IsDebug.isDebug) { System.out.println("[ExhibitAttrUtils: attribute words loaded: " + result.length + ".]"); }
            return(result);
            }
        catch(final Exception e)
            {
            System.err.println("ERROR: ExhibitAttrUtils: did not find attribute words in properties " + CoreConsts.GLOBAL_PROPS_NAME + ", property name " + ATTR_WORD_LIST_NAME + ".");
            e.printStackTrace();
            return(new String[0]);
            }
        }

    /**This class contains exhibit attribute words.
     * This data is derived from global config data.
     */
    public static final class ExhibitAttrWords implements Serializable
        {
        /**Construct a set of global (immutable) properties.
         *
         * @param attributeWords  set of attribute words for exhibit names;
         *     need not be sorted,
         *     may be null or zero length,
         *     but should not contain any zero-length words
         *     nor words consisting purely of digits
         *     nor that contain anything other than digits or lower-case letters
         */
        private ExhibitAttrWords(final String attributeWords[])
            {
            // Make private copy of array and ensure that words are sorted.
            // We will check that entries are valid and unduplicated in
            // validateObject();
            attrWordsSortedSet = Collections.unmodifiableSortedSet(
                new TreeSet<String>(Arrays.asList(attributeWords)));
            attrWords = new String[attrWordsSortedSet.size()];
            attrWordsSortedSet.toArray(attrWords);

            // Now compute our hash over the entries.
            longHash = computeHash();

            // Set up our comparators and stubs.
            _setComparators(attrWordsSortedSet);
            SUBJECT_ORDER =
                (new IndirectComparatorStub(longHash){ // Trivial stub.
                    protected final Comparator<String> getComp() { return(_SUBJECT_ORDER); }
                    private static final long serialVersionUID = -7219799635852371814L;
                    });
            SMART_ORDER =
                (new IndirectComparatorStub(longHash){ // Trivial stub.
                    protected final Comparator<String> getComp() { return(_SMART_ORDER); }
                    private static final long serialVersionUID = -6487931467358267993L;
                    });

            // Verify object state.
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**The hash of all the data held; guaranteed non-negative.
         * Depends all the information held in this object.
         */
        public final long longHash;

        /**The sorted and duplicate-free list of attribute words.
         * All of these should meet the rules for ExhibitNames.validAttributeWord().
         * <p>
         * We are careful never to share a reference to this outside
         * the class, eg upon deserialisation.
         * <p>
         * This is the primary store of attribute words information.
         */
        private /* final */ String attrWords[];

        /**The immutable sorted set of attribute words.
         * A duplicate of the info in attrWords[].
         */
        private transient SortedSet<String> attrWordsSortedSet;

        /**The sorted and duplicate-free non-null list of attribute words.
         * This array contains no nulls nor duplicates
         * and all entries meet the rules for
         * ExhibitNames.validAttributeWord().
         * <p>
         * Creating the return value may be expensive,
         * requiring extensive copying.
         *
         * @return  private copy of attribute-words list
         */
        public String[] getAttrWords()
            {
            return(((String[]) attrWords.clone()));
            }

        /**The set of attribute words as a view in an immutable SortedSet.
         * This set contains no nulls
         * and all entries meet the rules for
         * ExhibitNames.validAttributeWord().
         * <p>
         * Preparing this return value should take a constant, small, time.
         *
         * @return  immutable view of attribute-words sorted set
         */
        public SortedSet<String> getAttrWordsSortedSet()
            {
            return(attrWordsSortedSet);
            }

        /**Device to allow serialisation of trivial stubs and do proper quality test.
         * This allows serialisation to take place,
         * and also allows equals() to be implemented by comparing the
         * implementation class and the underlying longHash.
         */
        private static abstract class IndirectComparatorStub
            implements Serializable, Comparator<String>
            {
            /**Construct an instance with the hash embedded. */
            IndirectComparatorStub(final long longHash)
                { hash = (int) longHash; }

            /**Local version of main hash. */
            private final int hash;

            /**Get the comparator that we indirectly call. */
            protected abstract Comparator<String> getComp();

            /**Give this a fixed ID. */
            private static final long serialVersionUID = 0x5e0f51a9b80d99ddL;

            /**Apply the underlying Comparator */
            public final int compare(String ex1, String ex2)
                { return(getComp().compare(ex1, ex2)); }

            /**Compute a condensed version of the main hash. */
            public int hashCode()
                { return(hash); }

            /**Is comparator equal (ie has some data changed)? */
            public boolean equals(final Object obj)
                {
                if(obj == this) { return(true); } // Trival-but-common case.

                // Must be a stub comparator...
                if(!(obj instanceof IndirectComparatorStub)) { return(false); }
                final IndirectComparatorStub other = (IndirectComparatorStub) obj;

                // Must be same final (most-derived) anonymous Comparator class.
                if(getComp().getClass() != other.getComp().getClass()) { return(false); }

                // Must have the same underlying hash
                // and therefore the same attribute words.
                if(hashCode() != obj.hashCode()) { return(false); }

                return(true);
                }
            }

        /**Internal copy of SUBJECT_ORDER; not serialised. */
        private transient Comparator<String> _SUBJECT_ORDER;

        /**Internal copy of SMART_ORDER; not serialised. */
        private transient Comparator<String> _SMART_ORDER;

        /**Class to implement our "subject" order Comparator. */
        private static final class SubjOrderComp implements Comparator<String>
            {
            SubjOrderComp(final Set<String> attrWordsSortedSet)
                { this.attrWordsSortedSet = attrWordsSortedSet; }
            private final Set<String> attrWordsSortedSet;

            public final int compare(String s1, String s2)
                {
                // Comparison by main-words (stem) portion, case-insensitive.
                final int mainResult =
                    String.CASE_INSENSITIVE_ORDER.compare(
                        ExhibitName.getMainWordsComponent(s1, attrWordsSortedSet),
                        ExhibitName.getMainWordsComponent(s2, attrWordsSortedSet));
                if(mainResult != 0) { return(mainResult); }

                // Comparison by author (initials) portion.
                final int authResult =
                    ExhibitName.getAuthorComponent(s1).compareTo(
                    ExhibitName.getAuthorComponent(s2));
                if(authResult != 0) { return(authResult); }

                // Comparison by number-in-series portion.
                final int nisResult =
                    ExhibitName.getNumberInSeriesComponent(s1) -
                    ExhibitName.getNumberInSeriesComponent(s2);
                if(nisResult != 0) { return(nisResult); }

                return(0); // May differ only in attribute words.
                }
            };

        /**Class to implement our "smart" order Comparator. */
        private static final class SmartOrderComp implements Comparator<String>
            {
            SmartOrderComp(final SubjOrderComp subjOC, final Set<String> attrWordsSortedSet)
                {
                this.subjOC = subjOC;
                this.attrWordsSortedSet = attrWordsSortedSet;
                }
            private final SubjOrderComp subjOC;
            private final Set<String> attrWordsSortedSet;

            public final int compare(String s1, String s2)
                {
                // Comparison/sort on exhibit subject.
                final int subjectResult =  subjOC.compare(s1, s2);
                if(subjectResult != 0) { return(subjectResult); }

                // Comparison by attribute-words (stem) portion.
                String a1 = ExhibitName.getAttributeWordsComponent(s1, attrWordsSortedSet);
                if(a1 == null) { a1 = ""; }
                String a2 = ExhibitName.getAttributeWordsComponent(s2, attrWordsSortedSet);
                if(a2 == null) { a2 = ""; }
                final int attrResult =
                    a1.compareTo(a2);
                if(attrResult != 0) { return(attrResult); }

                // Needs a tie-break to achieve a total order.
                return(ExhibitName.SIMPLE_SMART_ORDER.compare(s1, s2));
                }
            };

        /**Routine to set _SUBJECT_ORDER and _SMART_ORDER.
         * Called from constructor and readObject().
         */
        private void _setComparators(final Set<String> attrWordsSortedSet)
            {
            final SubjOrderComp subjO = new SubjOrderComp(attrWordsSortedSet);
            _SUBJECT_ORDER = subjO;
            _SMART_ORDER = new SmartOrderComp(subjO, attrWordsSortedSet);
            }

        /**A Comparator that sorts exhibits by name comparing as equal those that differ only in attribute words.
         * This sorts, in order by:
         * <ol>
         * <li>main stem of the name case-insensitively
         * <li>author
         * <li>number-in-series
         * </ol>
         */
        public final Comparator<String> SUBJECT_ORDER;

        /**A Comparator that sorts exhibits by name taking account of attribute words.
         * This sorts, in order by:
         * <ol>
         * <li>SUBJECT_ORDER to sort into order by exhibit subject
         * <li>attribute words to sort within an exhibit subject
         * <li>ExhibitName.SIMPLE_SMART_ORDER to break ties
         * </ol>
         * <p>
         * equals() is true iff referring underlying comparator's backing AEP
         * has an unchanged longHash.
         */
        public final Comparator<String> SMART_ORDER;

        /**Compute a hash over all the data in the object; guaranteed non-negative.
         * Aims to be quick and comprehensive and portable between VMs,
         * and should not miss even quite subtle changes in the stored data.
         */
        private long computeHash()
            {
            // Running computed hash value.
            long result = 0;

            // Start a message digest (SHA1).
            final MessageDigest md = GenUtils.getStandardDigest();

            try {
                final DataOutputStream dos = new DataOutputStream(
                                new DigestOutputStream(
                                    (new OutputStream(){ // Null output stream...
                                        public final void write(int b) { }
                                        public final void write(byte[] b, int off, int len) { }
                                        }),
                                    md));

                // Put in attr words if any at all.
                for(int i = attrWords.length; --i >= 0; )
                    {
                    dos.writeUTF(attrWords[i]);
                    }

                dos.flush();
                }
            catch(IOException e) // Should never happen.
                { throw new Error("unexpected internal error"); }

            // Create the basic hash from the first 8 digest bytes.
            final byte[] digest = md.digest();
            long lHash = 0x9fe93d5e4820804bL; // Unique random base value.
            for(int i = 8; --i >= 0; )
                { lHash ^= (((long) (digest[i] & 0xff)) << (8*i)); }

            // We guarantee that the hash is non-negative.
            result ^= ((lHash >>> 1) | 1);

            // Return our computed hash.
            return(result);
            }

         /**Returns a hash code value for the object; derived from the longHash.
          * Guaranteed zero if there are no exhibits, non-zero otherwise.
          *
          * @return  a hash code value for this object.
          * @see     Object#equals(Object)
          * @see     Hashtable
          */
         public int hashCode()
             {
             return(((int) (longHash >>> 32)) ^ (int) longHash);
             }

         /**Indicates whether some other object is "equal to" this one; the underlying data is the same if true.
          * The hashNotChangedSince value is excluded from the comnparison.
          *
          * @param   obj   the reference object with which to compare.
          * @return  <code>true</code> if this object is the same as the obj
          *          argument; <code>false</code> otherwise.
          * @see     Boolean#hashCode()
          * @see     Hashtable
          */
         public synchronized boolean equals(Object obj)
             {
             // Must be of the same type to be equal.
             if(!(obj instanceof ExhibitAttrWords)) { return(false); }
             final ExhibitAttrWords eaw = (ExhibitAttrWords) obj;

             // We can quickly reject items that don't have the same hash...
             if(longHash != eaw.longHash) { return(false); }

             // Check the underlying immutable data.
             if(!Arrays.equals(attrWords, eaw.attrWords)) { return(false); }

             // OK, seems to actually be equal.
             return(true);
             }


        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         */
        public void validateObject()
            throws InvalidObjectException
            {
            // Check that all components are sane and safe.
            if(longHash < 0)
                { throw new InvalidObjectException("bad object: longHash < 0"); }
            // Check that we can correctly recompute the hash.
            if(longHash != computeHash())
                { throw new InvalidObjectException("bad object: hash wrong"); }
            if(attrWords == null)
                { throw new InvalidObjectException("bad object: attrWords == null"); }
            for(int i = attrWords.length; --i >= 0; )
                {
                final String aW = attrWords[i];
                if(!ExhibitName.validAttributeWord(aW))
                    { throw new InvalidObjectException("bad object: bad attrWord " + aW); }
                /*
                Don't need to check for duplicates or ordering
                because we process loaded words via a sorted set
                on construction and deserialisation.
                if((i > 0) && (aW.compareTo(attrWords[i-1]) <= 0))
                    { throw new InvalidObjectException("bad object: duplicate or out-of-order attrWord " + aW); }
                */
                }
            if(!(SMART_ORDER instanceof IndirectComparatorStub) ||
               !(SUBJECT_ORDER instanceof IndirectComparatorStub))
                { throw new InvalidObjectException("bad object: comparator wrong"); }
            if(!(_SMART_ORDER instanceof Comparator) ||
               !(_SUBJECT_ORDER instanceof Comparator))
                { throw new InvalidObjectException("bad object: internal comparator wrong"); }
            }

        /**Deserialise. */
        private synchronized void readObject(ObjectInputStream in)
            throws IOException, ClassNotFoundException
            {
            in.defaultReadObject();

            // Make private copy of attrWords[] and ensure that words are sorted.
            // We will check that entries are valid in validateObject().
            attrWordsSortedSet = Collections.unmodifiableSortedSet(
                new TreeSet<String>(Arrays.asList(attrWords)));
            attrWords = new String[attrWordsSortedSet.size()];
            attrWordsSortedSet.toArray(attrWords);

            // Set up our comparators.
            _setComparators(attrWordsSortedSet);

            validateObject(); // Validate state immediately.
            }

        /**Our serial version... */
        private static final long serialVersionUID = 0xf130cee4509ca259L;
        }

    /**Immutable set of exhibit attribute words, comparators, etc; never null. */
    private static final ExhibitAttrWords attrWords = new ExhibitAttrWords(_getAttributeWords());

    /**Gets (immutable) set of exhibit attribute words, comparators, etc; never null. */
    public static final ExhibitAttrWords getAttrWords()
        { return(attrWords); }
    }
