/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.par.ParTasks;
import org.hd.d.par.TaskProc;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;

/**Routines to handle the filesystem representation of an exhibit.
 * This absorbs some functionality buried in the old ImageInfoCache
 * classes and methods.
 */
public final class ExhibitFile
    {
    /**Collect a set of names of exhibits in the filesystem.
     * Given the base directory, this collects a set of
     * files that appear syntactically valid as exhibit
     * names.  The result is a Set of String valid exhibit names.
     * <p>
     * For system (such ans Windows) where the directory separator
     * is not '/' it is converted to this internally to provide
     * the canonical name format.
     * <p>
     * These are not checked to exist or contain valid data unless
     * checkFileContent is true though the directories encountered are
     * always checked to be directories.
     * <p>
     * The names returned omit the base directory.
     * <p>
     * The names are collected in parallel if possible.
     * <p>
     * The returned Set is mutable and synchronized.
     * <p>
     * Files are always checked for a recognized file extension.
     *
     * @param checkFileContent  if true, any putative exhibit is
     *     checked for readability and for a correct magic number
     */
    public static Set<String> getFilesystemBasedExhibitNames(final File baseDir,
                                                    final boolean checkFileContent)
        {
        final Set<String> result = Collections.synchronizedSet(new HashSet<String>());

        // First we get a set of top-level collection dirs.
        final String colDirs[] = baseDir.list(new java.io.FilenameFilter(){
            /**Check for a valid top-level gallery subdirectory.
             * The name must have the correct syntax and
             * actually be a directory that we can read.
             */
            public boolean accept(java.io.File dir, String name)
                {
                if(!ExhibitName.validNameInitialComponentSyntax(name))
                    { return(false); }
                final java.io.File f = new java.io.File(dir, name);
                return(f.isDirectory() && f.canRead());
                }
            });

        // We can return our empty result set if our dir list is empty.
        // (We can get a null result if our dir does not exist at all.)
        // (A minor optimisation.)
        if((colDirs == null) || (colDirs.length == 0))
            { return(result); }

if(ORG.hd.d.IsDebug.isDebug) { System.err.println("[ExhibitFile.getFilesystemBasedExhibitNames(): directories found: "+(colDirs.length)+".]"); }

        // We process the (top-level) directories in parallel,
        // recursing into extension directories if necessary.
        // (Currently we don't process the extension directories in parallel.)
        // The results are added to our result Set.
        ParTasks.apply(Arrays.asList(colDirs),
            (new TaskProc(){
                /**Scan the directory and its subdirectories.
                 * The argument is the (String) name of one top-level directory.
                 */
                public final void fn(Object o)
                    {
                    try {
                        // Stack of outstanding directories that we will work through...
                        final Stack<String> dirsToCheck = new Stack<String>();

                        // Get new top-level directory to recurse into...
                        final String topDir = (String) o;
    //System.err.println(" Examining top-level dir: " + topDir);

                        // Directories to search; initially `seeded'
                        // with the start directory, but new subdirectories
                        // can be added if suitably named.
                        //
                        // We loop until this is empty.
                        //
                        // The dir names are relative to the cwd, which
                        // is the Gallery root.
                        //
                        // We use a LIFO structure for simplicity,
                        // though we should probably use a FIFO one
                        // for a better (faster) disc-access pattern.
                        dirsToCheck.push(topDir);

                        while(!dirsToCheck.empty())
                            {
                            final String thisDir = (String) dirsToCheck.pop();
                            // Actually search the directory for candidate
                            // items (and subdirectories) and add them to
                            // the cache or worklist respectively.
                            final File d = new File(baseDir, thisDir); // Physical name.
                            final String entries[] = d.list(); // Get all dir entries in one pass.
                            if(entries == null)
                                {
                                System.err.println("ExhibitFile.getFilesystemBasedExhibitNames(): WARNING: bogus exhibit non-directory or non-reachable directory: " + d);
                                continue;
                                }
                            final String ds = thisDir + File.separatorChar;
                            for(int i = entries.length; --i >= 0; )
                                {
                                final String sName = entries[i];
                                // Does this look like a valid exhibit name?
                                if(ExhibitName.validNameFinalComponentSyntax(sName))
                                    {
                                    // This appears to be an item for the
                                    // cache, so attempt to add it...
                                    // If the directory separator is not
                                    // UNIX/URL style then replace it here...
                                    final String relativeFilePath = ds + sName;
                                    // Pure UNIX/URL style name.
                                    final String fullURLStyleName =
                                        (File.separatorChar == '/') ?
                                            relativeFilePath : (relativeFilePath.replace(File.separatorChar, '/'));
    //System.err.println(" Found item: " + relativeFilePath);
    //                                final String clashingFile =
    //                                    (String) shortNames.put(sName, relativeFilePath);
    //                                if(clashingFile != null)
    //                                    { System.err.println(" WARNING: CLASHING ITEMS with short name ``" + sName + "'': " + clashingFile + " and " + relativeFilePath); }
                                    // Now check the while name and
                                    // add to the final set if good...
                                    // (else silently ignore)
                                    if(!ExhibitName.validNameSyntax(fullURLStyleName)) { continue; }
    //System.err.println("  Syntax of full name is valid: " + relativeFilePath);

                                    // Reject if not a recognised (input)
                                    // file extension.
                                    final ExhibitMIME.ExhibitTypeParameters etp =
                                        ExhibitMIME.getInputFileType(fullURLStyleName);
                                    if(etp == null) { continue; }

                                    if(checkFileContent)
                                        {
                                        // Do extra (expensive) checking of
                                        // file names and file content.
    //System.err.println("  Extension is valid: " + relativeFilePath);

                                        final File f = new File(baseDir, relativeFilePath);

                                        // Veto anything unreadable or zero-length,
                                        // or has an unusable timestamp.
                                        if(!f.isFile() ||
                                           !f.canRead() ||
                                           (f.length() < 1) ||
                                           (f.lastModified() < 1))
                                            { continue; }
    //System.err.println("  File exists, is readable, and not zero-length and has a sensible timestamp: " + relativeFilePath);

                                        // Now attempt to read/check the magic number...
                                        InputStream is = null;
                                        try {
                                            is = new FileInputStream(f);
                                            // Veto if bad magic number...
                                            if(!ExhibitMIME.magicOK(etp, is))
                                                {
                                                System.err.println("ERROR: invalid magic number for file: " + relativeFilePath);
                                                continue;
                                                }
    //System.err.println("  File magic is valid: " + relativeFilePath);
                                            }
                                        catch(IOException e) { continue; } // Veto if any file error...
                                        finally
                                            {
                                            if(is != null) { try { is.close(); } catch(IOException e) { } }
                                            is = null; // Help GC...
                                            }
                                        }

                                    synchronized(result) { result.add(fullURLStyleName); }
                                    }
                                // Check if this might be a continuation subdir.
                                else if(sName.startsWith(ExhibitName.intermediateDirPrefix))
                                    {
    // We may accept and search more sub-dirs than
    // we will ever accept exhibits from as this stands...
                                    // Might be a continuation subdirectory;
                                    // if not really a directory reject it.
                                    final String relativeFilePath = ds + sName;
                                    if(!(new File(baseDir, relativeFilePath)).isDirectory())
                                        { continue; }
    //System.err.println(" Found subdir: " + relativeFilePath);
                                    // Looks fine; add it to the dirs to search...
                                    dirsToCheck.push(relativeFilePath);
                                    }
                                }
                            }
                        }
                    catch(Throwable t)
                        {
                        System.err.println("ExhibitFile.getFilesystemBasedExhibitNames(): ERROR: unexpected exception in task.");
                        t.printStackTrace();
                        }
                    }
                }),
            Math.min(_gFBEN_concurrencyLimit, colDirs.length),
            _estimatedDirScanUtil);

        return(result);
        }

    /**Limit on number of concurrent threads in getFilesystemBasedExhibitNames().
     * Used to limit number of filehandles we simultaneously need
     * while still allowing decent concurrency to overcome I/O
     * delays (will be most effective with native threads).
     * <p>
     * Values of 2--8 are usually good.
     */
    private static final int _gFBEN_concurrencyLimit = 8;

    /**Estimated CPU utilisation per getFilesystemBasedExhibitNames() filter operation/thread.
     * This assumes that some filtering is I/O bound and some CPU bound,
     * but is at best a guess...
     * <p>
     * ParTasks.UTILISATION_MAX is taken to represent 100% CPU-bound.
     * <p>
     * We make sure that this value is at least UTILISATION_MIN.
     */
    private static final int _estimatedDirScanUtil =
        Math.max(ParTasks.UTILISATION_MAX / _gFBEN_concurrencyLimit, ParTasks.UTILISATION_MIN);
    }
