/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.util.*;

/**Routines to validate/parse an exhibit name.
 * An exhibit name is a (relative) path name in a filesystem representation,
 * and a (relative) URL in a Web presentation.
 * <p>
 * This is based on the assumption that we use the String name
 * everywhere as a sort of universal currency, rather than pre-parsing
 * everything as in the old Attributes.ItemName.
 * <p>
 * The syntax of the file name is:
 * <ol>
 * <li>A category directory name (consisting of characters from the
 *     range [a-z-_] starting with a character in the range [a-z]
 * <li>Zero or more directory components of the form _more[0-9A-Z]*
 * <li>A final unique `file' name consisting of hyphen-separated
 *     words ending with a recognised extension that indicates the
 *     MIME type of the underlying file; full details of the syntax
 *     of this part are given below.
 * </ol>
 * <p>
 * The syntax of a Gallery image name is as follows:
 *     {word-}+{discardableword-}*[number-]AUTH.ext
 * where word is an alphanumeric sequence
 * containing at least one letter (there must
 * be at least one such word in the name),
 * discardableword is like word in syntax but
 * comes from a small list of words that can
 * trail the main image description and indicate
 * some gross features of the image, eg `bg'
 * and `mono' (these words are optional and
 * will only be recognised as discardable
 * if after all non-discardable words),
 * number is a optional decimal number consisting
 * purely of the digits 0-9 (with any leading
 * zeros ignored rather than indicating octal),
 * AUTH being the all-upper-case all-alpha
 * author's initials (this is compulsory), and
 * ext being the extension indicating the image
 * type(this is compulsory).
 * <p>
 * Note that other than the extension, all components
 * are delimited by dashes, and there must be no spaces
 * in the name.  The extension may contain dots, and
 * dots are not allowed elsewhere in the name.
 * <p>
 * For the purposes of sorting, the sort order
 * is first by the [word-]+ portion (ASCII order),
 * then by the author (ASCII order),
 * then by the number portion (numerically),
 * then by the {discardableword-}* portion (ASCII order),
 * then by the extension (ASCII order).
 * <p>
 * It is possible to build a ParsedName which is split into
 * its consitituent parts, and is Comparable with the sort
 * order given above.
 */
public final class ExhibitName
    {
    /**The character used to separate words. */
    public static final char WORD_SEP = '-';
    /**The character used to separate words as a string. */
    public static final String WORD_SEPS = "-";

    /**Maximum valid name length.
     * This is basically limited by maximum URL and UNIX-filename length,
     * and the fact that we need some "decoration" overhead of exhibit names
     * when enbedded in filenames, URLs, etc.
     */
    public static final int MAX_NAME_LENGTH = 1023 - 128;

    /**Minimum length of file component of any valid name.
     * Such a minimal name must be of the form ``a-A.a''.
     * In practice any real name must be longer since we will
     * not have one-character author initials nor file extensions,
     * but this will help us quickly discard directory entries
     * such as ``.'' and ``..'' for example.
     */
    private static final int _minFileLen = 5;

    /**Minimum valid name length.
     * Of form "a/a-A.a".
     */
    public static final int MIN_NAME_LENGTH = 2 + _minFileLen;


    /**Maximum allowable length of any single word.
     * This is where the name is of the form a/word-A.a.
     */
    public static final int MAX_WORD_LENGTH = MAX_NAME_LENGTH - 6;

    /**Maximum allowable length of any name stem (ie just main words).
     * This is where the name is of the form a/stem-A.a
     * where the stem is one or more words.
     */
    public static final int MAX_STEM_LENGTH = MAX_WORD_LENGTH;

    /* Maximum allowable attribute word length.
     * Maximum attribute word must be at most long enough to allow for
     * a single-letter main word and dash,
     * so is two less than the longest allowable main word.
     * This is where the name is of the form ``a/a-attrword-A.a''.
     */
    public static final int MAX_ATTR_WORD_LENGTH = MAX_WORD_LENGTH - 2;

    /**Prefix of intermediate directory components in name. */
    public static final String intermediateDirPrefix = "_more";
    /**Length of prefix of intermediate directory components in name. */
    private static final int _iDirPrefixLen = 5;

    /**Minimum length of author-initial component; strictly positive. */
    public static final int MIN_AUTH_INITIALS_LENGTH = 1;

    /**Maximum length of author-initial component. */
    public static final int MAX_AUTH_INITIALS_LENGTH = 8;

    /**Validates the syntax of the first component of a name; returns true if valid.
     * This is helpful to identify the roots of a directory scan, for example.
     * <p>
     * This ignores any characters from len onwards, so the portion from
     * 0--len-1 must be a complete valid initial directory, and the string
     * passed must be at least len characters long.
     * <p>
     * Designed to be efficiently callable from validNameSyntax() without
     * requiring creation of any objects.
     */
    public static boolean validNameInitialComponentSyntax(final String initialNameComponent,
                                                          final int len)
        {
        // DO INITIAL SET OF VERY QUICK TESTS.
        //Assert(initialNameComponent != null);
        //Assert(len <= initialNameComponent.length());
        if(len < 1) // Shortest possible name ``a''.
            { return(false); }
        // Allow room for a minimal name tail.
        if(len > MAX_NAME_LENGTH - (_minFileLen-1))
            { return(false); }

//        // Make sure that this is a pure single dir component; no path.
//        // Note that this is implict in the following tests.
//        if(-1 != initialNameComponent.lastIndexOf('/', len-1)) { return(false); }

        // First character of (directory component of) name
        // must be lower-case (ASCII) letter.
        final char firstChar = initialNameComponent.charAt(0);
        if((firstChar < 'a') || (firstChar > 'z'))
            { return(false); }

        // Now check the first dir/category component fully,
        // from after the first char.
        // Just lower-case letters and dash or underscore are permitted.
        for(int i = len; --i > 0; )
            {
            final char c = initialNameComponent.charAt(i);
            if(((c < 'a') || (c > 'z')) &&
               (c != '-') &&
               (c != '_'))
                { return(false); }
            }

        return(true); // Initial dir component OK!
        }

    /**Validates a set of author's initials for syntax; returns true if valid.
     * This version examines part of a string.
     * <p>
     * Characters from start to end-1 are checked.
     * <p>
     * The start and end positions must be valid in the String passed and start
     * start must come before end.
     * <p>
     * Returns false if the string is null.
     *
     * @param s  the value to be examined
     * @param start  the starting position of the author;
     *     must be valid offset within string
     * @param end  just after the end of the author;
     *     must be greater than start and no greater than the string length
     */
    public static boolean validAuthorSyntax(final String s,
                                            final int start,
                                            final int end)
        {
        if(s == null)
            { return(false); }

        if((start < 0) || (start > end) || (end > s.length()))
            { throw new IllegalArgumentException(); }

        final int len = end - start;
        if(len < MIN_AUTH_INITIALS_LENGTH) { return(false); }
        if(len > MAX_AUTH_INITIALS_LENGTH) { return(false); }

        // The auth initials must run from start up to just before end
        // and must be upper-case letters.
        for(int i = end; --i >= start; )
            {
            final char c = s.charAt(i);
            if((c < 'A') || (c > 'Z'))
                { return(false); } // Invalid author initial.
            }

        // Seems OK.
        return(true);
        }

     /**Validates a set of author's initials for syntax; returns true if valid.
      * The whole string must be a valid author ID (and not null).
      *
      * @return true iff s is not null and is a valid set of author initials
      */
     public static boolean validAuthorSyntax(final String s)
         { return((s != null) && validAuthorSyntax(s, 0, s.length())); }


    /**Checks that the String passed to it is a valid word (main or attribute).
     * This means it must be non-zero length (and non-null),
     * and consist only of letters and digits.
     */
    public static boolean validWord(final String s)
        {
        if((s == null) ||
            (s.length() == 0) || (s.length() > MAX_WORD_LENGTH))
            { return(false); }

        // Make sure that all characters are individually acceptable...
        for(int i = s.length(); --i >= 0; )
            {
            if(!validWordCharacter(s.charAt(i)))
                { return(false); }
            }

        return(true); // Seems OK...
        }


    /**Checks that the String passed to it is a valid attribute word.
     * This means it must be non-zero length (and non-null),
     * and consist only of letters and digits
     * and must not consist entirely of digits or upper-case letters
     * (to avoid ambiguity with the number-in-series value and author).
     * <p>
     * Maximum attribute word must be at most long enough to allow for
     * a single-letter main word and dash,
     * so is two less than the longest allowable main word.
     */
    public static boolean validAttributeWord(final String s)
        {
        if((s == null) ||
            (s.length() == 0) || (s.length() > MAX_ATTR_WORD_LENGTH))
            { return(false); }

        // Make sure that the word is valid in itself...
        if(!validWord(s)) { return(false); }

        final char firstChar = s.charAt(0);
        if((firstChar < '0') || (firstChar > '9'))
            {
            // Make sure that word is not entirely digits.
            for(int i = s.length(); --i >= 0; )
                {
                final char c = s.charAt(i);
                if((c < '0') || (c > '9'))
                    { return(true); } // Not purely numerical...
                }
            return(false); // Whoops, was only digits.
            }
        else if((firstChar < 'A') || (firstChar > 'Z'))
            {
            // Make sure that word is not upper-case.
            for(int i = s.length(); --i >= 0; )
                {
                final char c = s.charAt(i);
                if((c < 'A') || (c > 'Z'))
                    { return(true); } // Not purely numerical...
                }
            return(false); // Whoops, was only digits.
            }

        return(true); // Seems OK...
        }

    /**Validates the syntax of the first component of a name; returns true if valid.
     * This is helpful to identify the roots of a directory scan, for example.
     */
    public static boolean validNameInitialComponentSyntax(final String initialNameComponent)
        {
        if(initialNameComponent == null)
            { return(false); }
        return(validNameInitialComponentSyntax(initialNameComponent, initialNameComponent.length()));
        }

    /**Returns true if the character passed is a valid word character.
     * A valid word character is an ASCII digit or letter (either case).
     * <p>
     * We test the most common cases first for speed.
     */
    public static boolean validWordCharacter(final char c)
        {
        return(((c >= 'a') && (c <= 'z')) || // Most commonly lower-case.
               ((c >= 'A') && (c <= 'Z')) ||
               ((c >= '0') && (c <= '9')));
        }

    /**Validates the syntax of the last component of a name; returns true if valid.
     * This is helpful when running a directory scan, for example.
     * <p>
     * This does not check that the author's initials or the extension
     * are actually acceptable other than that they are syntactically valid.
     */
    public static boolean validNameFinalComponentSyntax(final String finalNameComponent)
        {
        // DO INITIAL SET OF VERY QUICK TESTS.
        if(finalNameComponent == null)
            { return(false); }
        final int len = finalNameComponent.length();
        // Reject filenames such as ``.'' and ``..'' quickly.
        if(len < _minFileLen) // Shortest possible name ``a-A.a''.
            { return(false); }
        // Allow room for a minimum-length directory component.
        if(len > MAX_NAME_LENGTH - 2)
            { return(false); }
        // Make sure that this is a pure file component; no path.
        if(finalNameComponent.indexOf('/') != -1)
            { return(false); } // Contains a directory component.

        final int lastDot = finalNameComponent.lastIndexOf('.');
        if((lastDot < _minFileLen - 2) || // Extension missing (-1), or this is hidden file (0) or initial part of name too short.
           (lastDot == len-1)) // Bogus empty extension...
            { return(false); }
        if(finalNameComponent.indexOf('.') != lastDot)
            { return(false); } // More than one dot; ambiguous extension.

        // Start looking for the author initials.
        final int lastDash = finalNameComponent.lastIndexOf(WORD_SEP);
        // Dismiss some simple malformations quickly...
        if((lastDash > lastDot) || // Bogus dash in extension.
           (lastDash == lastDot-1) || // Zero-length author initials
           (lastDash < 1)) // No dash at all (-1) or zero-length first word (0).
            { return(false); }
        // The auth initials must run from lastDash+1 upto just before lastDot
        // and must be upper-case letters.
        if(!validAuthorSyntax(finalNameComponent, lastDash+1, lastDot))
            { return(false); }

        // Check that the text up to the last dash consists of
        // one or more dash-terminated words each of which consists
        // of one or more ASCII letters and digits.
        // (Two consecutive dashes are not allowed.)

        // nextWord is positioned successively at the start of
        // each word.
        doNextWord: for(int nextWord = 0; nextWord <= lastDash; )
            {
            // Gobble up one word checking that it is not zero-length
            // and contains only valid characters.
            // We will stop at lastDash at worst.
            for(int i = nextWord; ; ++i)
                {
                final char c = finalNameComponent.charAt(i);
                if(c == WORD_SEP)
                    {
                    if(i == nextWord) { return(false); } // Empty word.
                    // OK, one valid word seen...
                    nextWord = i + 1; // Skip to start of next word.
                    continue doNextWord; // Process next word.
                    }

                // Check that the current word char is valid.
                if(!validWordCharacter(c))
                    { return(false); } // Not a valid word character.
                }
            }

        return(true); // Seems OK!
        }

    /**Very quick basic set of name validity checks; returns true if valid.
     * Very quick <em>constant-time</em> checks that the name is not null
     * and is of a legitimate length.
     */
    public static boolean validNameSyntaxBasic(final String name)
        {
        if(name == null)
            { return(false); }
        final int len = name.length();
        if(len < 2 + _minFileLen) // Shortest possible name a/a-A.a.
            { return(false); }
        if(len > MAX_NAME_LENGTH)
            { return(false); }
        return(true);
        }

    /**Fully validates the syntax of a name; returns true if valid.
     * It attempts to be fast and to not create too many intermediate/temporary
     * objects.
     * <p>
     * This does not attempt to check a name against current databases
     * nor return any parsed components.
     */
    public static boolean validNameSyntax(final String name)
        {
//System.err.println("  [validNameSyntax("+name+")]");
        // DO INITIAL SET OF VERY QUICK TESTS.
        if(!validNameSyntaxBasic(name))
            { return(false); }
        final int len = name.length();
        // Must contain at least one of each . / and -.
        final int firstSlash = name.indexOf('/');
        final int lastSlash = name.lastIndexOf('/');
        if(lastSlash >= len - _minFileLen) // Final component too short.
            { return(false); }
        if((firstSlash < 1) || // No slash at all (-1) or first dir component absent (0) eg absolute path.
           (name.lastIndexOf('-') < lastSlash)) // Must be at least one word and author in file component...
            { return(false); }

//System.err.println("  [validNameSyntax("+name+"): quick tests passed]");

        // Invalid initial component?
        if(!validNameInitialComponentSyntax(name, firstSlash))
            { return(false); }
//System.err.println("  [validNameSyntax("+name+"): validNameInitialComponentSyntax]");

        // Check intermediate components (if any).
        // Must be of the form _more[0-9A-Z]*
        if(lastSlash != firstSlash)
            {
//System.err.println("  [validNameSyntax("+name+"): checking intermediate dir names...]");
            // Next part points to the _more prefix of the next
            // intermediate directory.
            nextDirComponent: for(int nextPart = firstSlash + 1; nextPart <= lastSlash; )
                {
//System.err.println("   [validNameSyntax("+name+"): looking at "+name.substring(nextPart)+"]");
                if(!name.regionMatches(nextPart, intermediateDirPrefix, 0, _iDirPrefixLen))
                    { return(false); }
//System.err.println("   [validNameSyntax("+name+"): intermediate dir prefix OK]");

                for(int tail = nextPart + _iDirPrefixLen; ; ++tail)
                    {
                    final char c = name.charAt(tail);
                    if(c == '/')
                        {
                        // No problems found in dir name component.
                        nextPart = tail + 1; // Skip over the next slash.
//System.err.println("   [validNameSyntax("+name+"): found end of dir component]");
                        continue nextDirComponent; // Start on next component.
                        }
                    if(!(((c >= '0') && (c <= '9')) ||
                        ((c >= 'A') && (c <= 'Z'))))
                        { return(false); } // Whoops; invalid character.
//System.err.println("   [validNameSyntax("+name+"): dir char "+c+" OK]");
                    }
                }
            }
//System.err.println("  [validNameSyntax("+name+"): intermediate components OK]");

        // Now thoroughly check the final component.
        // Oh dear; this makes a temporary (String) object.
        if(!validNameFinalComponentSyntax(name.substring(lastSlash + 1)))
            { return(false); }
//System.err.println("  [validNameSyntax("+name+"): validNameFinalComponentSyntax]");
        return(true);
        }

    /**A simple invariant comparator that sorts (String) full exhibit names in a human-friendly order.
     * This is essentially a case-insensitive sort on the file component.
     * <p>
     * Ties are broken by a normal String lexical ordering on the full
     * names.
     * <p>
     * Discardable/attribute words are not discarded
     * nor otherwise treated specially for this comparison.
     */
    public static final Comparator<String> SIMPLE_SMART_ORDER = (new Comparator<String>(){
        public final int compare(final String s1, final String s2)
            {
            // Case-insensitive comparison first by filename portion.
            final int ciResult = String.CASE_INSENSITIVE_ORDER.compare(
                getFileComponent(s1),
                getFileComponent(s2));
            if(ciResult != 0) { return(ciResult); }

            // Needs a tie-break.
            return(s1.compareTo(s2));
            }
        });



    /**Extract the file component (short name) of a full exhibit name, assuming the name is valid.
     * Two exhibits should always be distinguishable by this component,
     * also known as the "short" name.
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     */
    public static String getFileComponent(final String fullExhibitName)
        {
        assert(validNameSyntax(fullExhibitName));
        return(fullExhibitName.substring(fullExhibitName.lastIndexOf('/') + 1));
        }

    /**Extract the category component (top directory) of a full exhibit name, assuming the name is valid.
     * If the argument is not a valid full exhibit name, the result is undefined.
     */
    public static String getCategoryComponent(final String fullExhibitName)
        {
        assert(validNameSyntax(fullExhibitName));
        return(fullExhibitName.substring(0, fullExhibitName.indexOf('/')));
        }

    /**Extract the full directory component of a full exhibit name, assuming the name is valid.
     * This does not include the trailing directory separator.
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     */
    public static String getDirComponent(final String fullExhibitName)
        {
        assert(validNameSyntax(fullExhibitName));
        return(fullExhibitName.substring(0, fullExhibitName.lastIndexOf('/')));
        }

    /**Get the "virtual" name consisting of the category directory and then the file component.
     * The results is a syntactically-valid full exhibit name,
     * but omits the internal directory "noise".
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     */
    public static String getVirtualExhibitName(final String fullExhibitName)
        {
        assert(validNameSyntax(fullExhibitName));

        // Optimisation:
        // if there is only one slash, the name is already virtual,
        // so the name can be returned untouched.
        final int firstSlash = fullExhibitName.indexOf('/');
        final int lastSlash = fullExhibitName.lastIndexOf('/');
        if(firstSlash == lastSlash) { return(fullExhibitName); }

        // Do this with only one append operation for efficiency.
        return(fullExhibitName.substring(0, firstSlash+1) +
               fullExhibitName.substring(lastSlash+1));
        }

    /**Find end of main stem and of attribute words of exhibit name.
     * This returns a three-element array, of which:
     * <ul>
     * <li>Element 0 is the position of the separator immediately
     *     before the first main word (a '/').
     * <li>Element 1 is the position (index in the input String)
     * of the separator immediately following the final word of
     * the main stem of the exhibit name.
     * All the words of the main stem precede it,
     * and any attribute words and number-in-series value follow it.
     * <li>Element 2 of the result is the position (index in the input String)
     * of the separator immediately following the final attribute word of
     * the exhibit name.
     * (If there are no attribute words this will be the same value
     * as element 1.)
     * Any number-in-series value follows it.
     * </ul>
     * <p>
     * This has to be passed a set of all valid attribute words
     * (as Strings which meet the requirements of validAttributeWord())
     * to be able to compute this boundary.
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     *
     * @param allAttrWords  a Set of all legal attribute words (String values);
     *     can be empty but not null
     */
    public static int[] getMainAndAttrWordComponentBoundaries(
                                                final String fullExhibitName,
                                                final Set<String> allAttrWords)
        {
        assert(validNameSyntax(fullExhibitName)) : ("not valid full name: " + fullExhibitName);

        // Find the end of the attribute words.
        final int lastDash = fullExhibitName.lastIndexOf('-'); // Before author.
        // Skip back one word to omit number-in-series value if present.
        final int endOfAttrWords =
            (getNumberInSeriesComponentAsString(fullExhibitName) == null) ?
            lastDash : fullExhibitName.lastIndexOf('-', lastDash-1);

        // End of main word; same as end of attr words if no attr words.
        int endOfMainWords = endOfAttrWords; // Always pointing to a dash.

        // The last slash is the character before the first main word.
        final int lastSlash = fullExhibitName.lastIndexOf('/');

        // If there are no attribute words then we can skip some processing...
        if(!allAttrWords.isEmpty())
            {
            // Find end of first main word,
            // given that there must be at least one (of at least length 1).
            final int endOfFirstMainWord = fullExhibitName.indexOf('-', lastSlash+2);

            // Work back a word at a time leaving the boundary after the
            // first non-attribute word.
            while(endOfMainWords > endOfFirstMainWord)
                {
                //assert(fullExhibitName.charAt(endOfMainWords, '-'));

                // We can skip back 2 each time since each word must be at least
                // a single letter long preceded by a dash.
                final int previousDash = fullExhibitName.lastIndexOf('-', endOfMainWords-2);

                final String putativeAttrWord =
                    fullExhibitName.substring(previousDash + 1, endOfMainWords);
                if(!allAttrWords.contains(putativeAttrWord)) { break; }
                // If the word was an attribute word then move the boundary back.
                endOfMainWords = previousDash;
                }
            }

//        assert (lastSlash>0) && (lastSlash < endOfMainWords) && (endOfMainWords <= endOfAttrWords) && (endOfAttrWords <= lastDash);
//System.out.println("lastSlash, endOfMainWords, endOfAttrWords: " + lastSlash + ", " + endOfMainWords + ", " + endOfAttrWords);

        // Return offsets...
        return(new int[]{ lastSlash, endOfMainWords, endOfAttrWords });
        }


    /**Extract the main words (stem) component of a full exhibit name, assuming the name is valid; never null nor empty.
     * There is always at least one main word; so the result is always
     * non-null and non-empty.
     * <p>
     * (This does not end nor start with a separator.)
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     *
     * @param allAttrWords  a Set of all legal attribute words (String values)
     */
    public static String getMainWordsComponent(final String fullExhibitName, final Set<String> allAttrWords)
        {
//        assert(validNameSyntax(fullExhibitName)) : ("not valid full name: " + fullExhibitName);
        final int[] offsets = getMainAndAttrWordComponentBoundaries(fullExhibitName, allAttrWords);
        return(fullExhibitName.substring(offsets[0]+1, offsets[1]));
        }

    /**Return Enumeration over main words; never null, never empty if the name is well-formed.
     * Uses StringTokenizer.
     */
    public static Enumeration getMainWords(final String fullExhibitName, final Set<String> allAttrWords)
        {
        return(new StringTokenizer(getMainWordsComponent(fullExhibitName, allAttrWords), WORD_SEPS));
        }

    /**Extract the attribute words component of a full exhibit name, assuming the name is valid.
     * If there are no attribute words this returns null,
     * else the result is non-empty
     * and is the fragment of the full name containing the attribute word
     * with the words separated as usual.
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     *
     * @param allAttrWords  a Set of all legal attribute words (String values)
     */
    public static String getAttributeWordsComponent(final String fullExhibitName, final Set<String> allAttrWords)
        {
        assert(validNameSyntax(fullExhibitName));
        final int[] offsets = getMainAndAttrWordComponentBoundaries(fullExhibitName, allAttrWords);
        if(offsets[1] == offsets[2]) { return(null); } // No attr words.
        return(fullExhibitName.substring(offsets[1]+1, offsets[2]));
        }

    /**Extract the attribute words component of a full exhibit name as an Enumeration of String, assuming the name is valid.
     * If there are no attribute words this returns null,
     * else the result is non-empty
     * and is an Enumeration of String values of the attributes in order.
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     *
     * @param allAttrWords  a Set of all legal attribute words (String values)
     */
    public static Enumeration getAttributeWordsComponentEnumeration(final String fullExhibitName, final Set<String> allAttrWords)
        {
        assert(validNameSyntax(fullExhibitName));
        final String s = getAttributeWordsComponent(fullExhibitName, allAttrWords);
        if(s == null) { return(null); }
        return(new StringTokenizer(s, WORD_SEPS));
        }

    /**Extract the attribute words component of a full exhibit name as a SortedSet of String, assuming the name is valid; never null.
     * If there are no attribute words this returns
     * a (fixed, immutable) empty set.
     * <p>
     * Duplicates attribute words are automatically eliminated
     * <p>
     * If the argument is not a valid full exhibit name, the result is undefined.
     *
     * @param allAttrWords  a Set of all legal attribute words (String values)
     *
     * @return non-null, de-duped, alpha-sorted attribute words from the name
     */
    public static SortedSet<String> getAttributeWordsComponentSortedSet(final String fullExhibitName,
                                                                        final Set<String> allAttrWords)
        {
        assert(validNameSyntax(fullExhibitName));
        final String s = getAttributeWordsComponent(fullExhibitName, allAttrWords);
        if(s == null) { return(NO_ATTR_WORDS); }

        final SortedSet<String> result = new TreeSet<String>();
        final StringTokenizer st = new StringTokenizer(s, WORD_SEPS);
        while(st.hasMoreElements()) { result.add(st.nextToken()); }

        return(result);
        }

    /**Immutable empty attribute word set. */
    private static final SortedSet<String> NO_ATTR_WORDS =
        Collections.unmodifiableSortedSet(new TreeSet<String>());

    /**Extract the number-in-series component of a full exhibit name as a String, assuming the name is valid.
     * If the argument is not a valid full exhibit name, the result is undefined
     * <em>unless</em> the name is the final (file) component of a valid name.
     * <p>
     * A missing number-in-series value causes us to return null.
     * <p>
     * This is:
     * <ul>
     * <li>the final word before the author initials,
     * <li>must consist entirely of digits,
     * <li>is always non-negative,
     * <li>is always to be interpretted in radix 10,
     *     so "010" = "10" = 10 (ie ten),
     * <li>and is not the only word.
     * </ul>
     */
    public static String getNumberInSeriesComponentAsString(final String exhibitName)
        {
        assert(validNameSyntax(exhibitName) || validNameFinalComponentSyntax(exhibitName));
        // The number-in-series value, if present,
        // is the final word and contains only digits
        // and must not be the only word.
        final int lastDash = exhibitName.lastIndexOf('-');
        final int previousDash = exhibitName.lastIndexOf('-', lastDash-1);
        if(previousDash <= 0) { return(null); } // Only word?
        final String putativeNiS = exhibitName.substring(previousDash+1, lastDash);
        for(int i = putativeNiS.length(); --i >= 0; )
            {
            final char c = putativeNiS.charAt(i);
            if((c < '0') || (c > '9')) { return(null); } // Not pure number.
            }
        // Looks OK!
        return(putativeNiS);
        }

    /**Extract the number-in-series component of a full exhibit name as an int, assuming the name is valid.
     * If the argument is not a valid full exhibit name, the result is undefined.
     * <p>
     * A missing number-in-series value causes us to return zero.
     * <p>
     * This is:
     * <ul>
     * <li>the final word before the author initials,
     * <li>must consist entirely of digits,
     * <li>is always non-negative,
     * <li>is always to be interpretted in radix 10,
     *     so "010" = "10" = 10 (ie ten),
     * <li>and is not the only word.
     * </ul>
     */
    public static int getNumberInSeriesComponent(final String fullExhibitName)
        {
        final String nis = getNumberInSeriesComponentAsString(fullExhibitName);
        if(nis == null) { return(0); }
        return(Integer.parseInt(nis, 10));
        }

    /**Extract the author component of a full exhibit name, assuming the name is valid.
     * If the argument is not a valid full exhibit name, the result is undefined
     * <em>unless</em> the name is the final (file) component of a valid name.
     */
    public static String getAuthorComponent(final String exhibitName)
        {
        assert(validNameSyntax(exhibitName) || validNameFinalComponentSyntax(exhibitName));
        // The author initials lie between the last dash and the last dot.
        final int lastDash = exhibitName.lastIndexOf('-');
        final int lastDot = exhibitName.lastIndexOf('.');
        return(exhibitName.substring(lastDash+1, lastDot));
        }

    /**Extract the extension (without dot) of a full exhibit name, assuming the name is valid.
     * If the argument is not a valid full exhibit name, the result is undefined
     * <em>unless</em> the name is the final (file) component of a valid name.
     */
    public static String getExtensionComponent(final String exhibitName)
        {
        assert(validNameSyntax(exhibitName) || validNameFinalComponentSyntax(exhibitName));
        final int lastDot = exhibitName.lastIndexOf('.');
        return(exhibitName.substring(lastDot+1));
        }
    }
