/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.AllExhibitProperties.ExhibitDataSource;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.MIME.Handler;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.awt.*;
import java.io.*;
import java.lang.ref.SoftReference;
import java.util.Arrays;

/**Immutable (and serialisable) store of all immutable computable auxiliary properties of a single exhibit.
 * These are properties that can be computed from the basic exhibit binary
 * data and that generally do not need to be recomputed
 * (unless that exhibit changes,
 * or if we upgrade the code and can get "better" values).
 * <p>
 * Some of the attributes contained in this object may be expensive to
 * compute (and may require reading the entire exhibit)
 * and thus this object is precomputed on the master server (or offline)
 * where the exhibit is available from fast local storage in its entirety.
 * <p>
 * This is designed to be efficient on the wire and in memory, since these
 * details will be held for each and every exhibit.
 * <p>
 * Some data in this object is transient and recomputable from other fields
 * and is recomputed if this object is serialised and deserialised
 * or is kicked out of memory due a shortage.
 * <p>
 * The equals() and hashCode() of this class can be very expensive,
 * so intern()ing it can be too.  However, we do intern() some of its elements.
 * <p>
 * The construction of this item is brittle, so that
 * if we cannot construct it all, correctly, then we veto construction entirely,
 * so as to avoid results from transient failures persisting indefinitely.
 */
public final class ExhibitPropsComputable implements Serializable,
                                                     ObjectInputValidation,
                                                     MemoryTools.Internable,
                                                     MemoryTools.Compactable
    {
    /**Shared empty instance. */
    public static final ExhibitPropsComputable EMPTY = new ExhibitPropsComputable();

    /**Make an empty instance. */
    private ExhibitPropsComputable() { this(null, null); }

    /**Make non-empty instance.
     * This does not intern() any of its arguments
     * (so any that can usefully be intern()ed should already have been).
     * <p>
     * This may defensively copy mutable argument values.
     * 
     * @param xyDim  xy (positive) dimensions in pixels or null if none
     * @param metadata  normalised metadata else null if none.
     */
    private ExhibitPropsComputable(final Dimension xyDim,
                                   final Object metadata)
        {
        // Take defensive copy if input if need be.
        xyDimensions = (xyDim == null) ? null : new Dimension(xyDim);
        this.metadata = metadata;

        // Verify object state.
        try { validateObject(); }
        catch(final InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage(), e); }
        }


    /**Common "assumed" prefix on XML metadata. */
    private static final String MDPREFIX = "<" + Handler.TAG_NAME_METADATA_TOP + ">";
    private static final int MDPREFIX_LENGTH = MDPREFIX.length();
    /**Common "assumed" suffix on XML metadata. */
    private static final String MDSUFFIX = "</" + Handler.TAG_NAME_METADATA_TOP + ">";
    private static final int MDSUFFIX_LENGTH = MDSUFFIX.length();

    /**X,Y dimensions in pixels for images (moving or still) that have a fixed rectangular dimension, else null.
     * If this is non-null then x and y are strictly positive.
     * <p>
     * Because Dimension is mutable, we hand out copies of it to callers.
     */
    private final Dimension xyDimensions;

    /**Get x,y dimensions in pixels for images (moving or still) that have a fixed rectangular dimension, else null.
     * If this is non-null then x and y are strictly positive.
     * <p>
     * We return a copy of our internal data.
     */
    public java.awt.Dimension getXyDimensions()
        {
        if(xyDimensions == null) { return(null); }
        return(new Dimension(xyDimensions));
        }

    /**Shared static dictionary for use with in-memory Compact7BitString metadata.
     * We expose this to help generate tuned dictionary values;
     * this is safe to do since this dictionary is immutable.
     */
    public static final Compact7BitString.StaticDictionary sDict = new Compact7BitString.StaticDictionary("EPC",
        // Most-common and largest early tokens first in list for maximum savings.
        Arrays.asList(new String[]{
            "NumProgressiveScans",    /* count=8471, saving=152478, meanFirstPos=39 */
            "ImageOrientation",    /* count=9021, saving=135315, meanFirstPos=51 */
            "javax_imageio_1",    /* count=9058, saving=126812, meanFirstPos=3 */
            "ColorSpaceType",    /* count=9054, saving=117702, meanFirstPos=9 */
            "NumChannels",    /* count=9054, saving=90540, meanFirstPos=22 */
            "Compression",    /* count=9049, saving=90490, meanFirstPos=24 */
            "Orientation",    /* count=7724, saving=77240, meanFirstPos=96 */
            "Dimension",    /* count=9045, saving=72360, meanFirstPos=48 */
            "Resolution",    /* count=7783, saving=70047, meanFirstPos=108 */
            "Lossless",    /* count=9045, saving=63315, meanFirstPos=33 */
            "TypeName",    /* count=8477, saving=59339, meanFirstPos=27 */
            "Chroma",    /* count=9056, saving=45280, meanFirstPos=7 */
            "normal",    /* count=8433, saving=42165, meanFirstPos=53 */
            "native",    /* count=8232, saving=41160, meanFirstPos=64 */
            "value",    /* count=9206, saving=36824, meanFirstPos=17 */
            "\"/></",    /* count=9194, saving=36776, meanFirstPos=21 */
            "image",    /* count=9058, saving=36232, meanFirstPos=8 */
            "false",    /* count=8480, saving=33920, meanFirstPos=37 */
            "YCbCr",    /* count=8367, saving=33468, meanFirstPos=13 */
            "Model",    /* count=7697, saving=30788, meanFirstPos=83 */
            "\"/><",    /* count=9205, saving=27615, meanFirstPos=15 */
            "name",    /* count=9053, saving=27159, meanFirstPos=11 */
            "JPEG",    /* count=8433, saving=25299, meanFirstPos=30 */
            "EXIF",    /* count=8232, saving=24696, meanFirstPos=66 */
            "Make",    /* count=7437, saving=22311, meanFirstPos=72 */
            "left",    /* count=6913, saving=20739, meanFirstPos=101 */
            "PixelAspectRatio",    /* count=1236, saving=18540, meanFirstPos=56 */
            "></",    /* count=9019, saving=18038, meanFirstPos=61 */
            "tag",    /* count=8232, saving=16464, meanFirstPos=68 */
            "DIGITAL",    /* count=2422, saving=14532, meanFirstPos=86 */
            "SONY",    /* count=4230, saving=12690, meanFirstPos=75 */
            "HorizontalScreenSize",    /* count=574, saving=10906, meanFirstPos=77 */
            "CompressionTypeName",    /* count=572, saving=10296, meanFirstPos=31 */
            "VerticalScreenSize",    /* count=574, saving=9758, meanFirstPos=83 */
            "F828",    /* count=3215, saving=9645, meanFirstPos=87 */
            "Canon",    /* count=2383, saving=9532, meanFirstPos=75 */
            "=\"",    /* count=9206, saving=9206, meanFirstPos=12 */
            "><",    /* count=9206, saving=9206, meanFirstPos=3 */
            "\" ",    /* count=8545, saving=8545, meanFirstPos=72 */
            "BackgroundIndex",    /* count=572, saving=8008, meanFirstPos=21 */
            "BitsPerSample",    /* count=616, saving=7392, meanFirstPos=55 */
            "SampleFormat",    /* count=613, saving=6743, meanFirstPos=49 */
            "HorizontalPixelSize",    /* count=346, saving=6228, meanFirstPos=66 */
            "BlackIsZero",    /* count=620, saving=6200, meanFirstPos=16 */
            "VerticalPixelSize",    /* count=346, saving=5536, meanFirstPos=74 */
            "stream",    /* count=718, saving=3590, meanFirstPos=1 */
            "Normal",    /* count=588, saving=2940, meanFirstPos=76 */
            "keyword",    /* count=416, saving=2496, meanFirstPos=65 */
            "comment",    /* count=413, saving=2478, meanFirstPos=67 */
            "Index",    /* count=578, saving=2312, meanFirstPos=52 */
            "SAMSUNG",    /* count=337, saving=2022, meanFirstPos=70 */
            "Data",    /* count=616, saving=1848, meanFirstPos=46 */
            "true",    /* count=599, saving=1797, meanFirstPos=41 */
            "TRUE",    /* count=598, saving=1794, meanFirstPos=20 */
            "Entry",    /* count=415, saving=1660, meanFirstPos=63 */
            "TECHWIN",    /* count=260, saving=1560, meanFirstPos=69 */
            "sampleRate",    /* count=144, saving=1296, meanFirstPos=19 */
            "sampleSize",    /* count=144, saving=1296, meanFirstPos=31 */
            "Text",    /* count=417, saving=1251, meanFirstPos=61 */
            "RGB",    /* count=607, saving=1214, meanFirstPos=13 */
            "frameSize",    /* count=144, saving=1152, meanFirstPos=49 */
            "lzw",    /* count=574, saving=1148, meanFirstPos=35 */
            "encoding",    /* count=146, saving=1022, meanFirstPos=15 */
            "channels",    /* count=144, saving=1008, meanFirstPos=43 */
            "sample",    /* count=144, saving=720, meanFirstPos=37 */
            "frames",    /* count=143, saving=715, meanFirstPos=3 */
            "audio",    /* count=148, saving=592, meanFirstPos=11 */
            "bytes",    /* count=144, saving=576, meanFirstPos=53 */
            "frame",    /* count=144, saving=576, meanFirstPos=55 */
            "unit",    /* count=146, saving=438, meanFirstPos=21 */
            "PCM_SIGNED",    /* count=48, saving=432, meanFirstPos=17 */
            "bits",    /* count=144, saving=432, meanFirstPos=35 */
            "8000",    /* count=124, saving=372, meanFirstPos=27 */
            "ULAW",    /* count=91, saving=273, meanFirstPos=17 */
            "GRAY",    /* count=78, saving=234, meanFirstPos=13 */
            "PaletteEntry",    /* count=14, saving=154, meanFirstPos=33 */
            "Hz",    /* count=144, saving=144, meanFirstPos=23 */
            "Palette",    /* count=14, saving=84, meanFirstPos=31 */
            "Gamma",    /* count=19, saving=76, meanFirstPos=21 */
            "green",    /* count=14, saving=56, meanFirstPos=40 */
            "index",    /* count=14, saving=56, meanFirstPos=44 */
            "PCM_UNSIGNED",    /* count=5, saving=55, meanFirstPos=17 */
            "22050",    /* count=11, saving=44, meanFirstPos=27 */
            "45453998",    /* count=6, saving=42, meanFirstPos=27 */
            "None",    /* count=14, saving=42, meanFirstPos=36 */
            "blue",    /* count=14, saving=42, meanFirstPos=36 */
            "39999998",    /* count=5, saving=35, meanFirstPos=27 */
            "44100",    /* count=8, saving=32, meanFirstPos=27 */
            "alpha",    /* count=6, saving=24, meanFirstPos=35 */
            "FALSE",    /* count=5, saving=20, meanFirstPos=19 */
            "CCITT",    /* count=4, saving=16, meanFirstPos=36 */
            "BI_RGB",    /* count=3, saving=15, meanFirstPos=32 */
            "45455",    /* count=3, saving=12, meanFirstPos=27 */
            "/><",    /* count=4, saving=8, meanFirstPos=15 */     
        }));

    /**Metadata for the exhibit; null if none available.
     * This may exist in a number of formats for efficiency "on the wire"
     * and to save space in memory.
     * <p>
     * <ul>
     * <li>null: no metadata for this exhibit
     * <li>String: the simplest format, containing well-formed XML-format metadata
     * <li>ROByteArray: deflated UTF-8 version of String format (LEGACY)
     * <li>Compact7BitString: compact byte representation (MOST COMMON)
     * <li>byte[]: raw internal Compact7BitString data (NOT SUPPORTED)
     * </ul>
     * <p>
     * Note that this is stored without the redundant top-level tag.
     * <p>
     * Marked volatile to allow safe lockless update by compact().
     */
    private volatile Object metadata;
    
    /**Get metadata as full XML string surrounded with top-level tags; null if no metadata.
     * @return null if no metadata
     */
    public String getMetadataAsXML()
        {
        // Capture local/fast snapshot.
        final Object md = metadata;

        // Handle most common formats first.

//        if(metadata instanceof byte[]) // Unwrapped compressed XML String.
//            { return(MDPREFIX+(Compact7BitString.makeFromInternalBytes((byte[]) metadata, null))+MDSUFFIX); }

        if(md instanceof Compact7BitString) // Compressed XML String.
            { return(MDPREFIX+md+MDSUFFIX); }

        if(md instanceof String) // Uncompressed XML String.
            { return(MDPREFIX+md+MDSUFFIX); }

        if(md instanceof ROByteArray) // Deflated XML String.
            {
            try { return(MDPREFIX+ROByteArray.uncompressToString((ROByteArray) md)+MDSUFFIX); }
            catch(final IOException e) { throw new Error("invalid metadata", e); }
            }

        if(null == metadata) // No metadata (rare, but legal).
            { return(null); }

        // Format unknown...
        throw new Error("invalid metadata");
        }

    /**If true then cache the decoded metadata via a SoftReference. */
    private static final boolean CACHE_DECODED_METADATA = false;

    /**Decoded form of metadata; null if not decoded or no metadata.
     * Held via a SoftReference to allow this (redundant and bulky) decoded data's memory to be freed.
     * <p>
     * Transient since this contains no unique data to serialise.
     * <p>
     * Accessed (and set upon demand) by getMetadata().
     */
    private transient SoftReference<Node> metadataDecoded;

    /**Get exhibit metadata; null if none available.
     */
    public synchronized Node getMetadata()
        {
        // If no metadata at all, then immediately return null.
        if(metadata == null) { return(null); }

        Node result = null;
        if(!CACHE_DECODED_METADATA ||
           (metadataDecoded == null) || ((result = metadataDecoded.get()) == null))
            {
            try { result = parseMetadata(getMetadataAsXML()); }
            catch(final Exception e)
                {
                e.printStackTrace();
                return(null); // Parse failed.
                }

            // Cache the result.
            if(CACHE_DECODED_METADATA)
                {
                if(result == null) { metadataDecoded = null; }
                else { metadataDecoded = new SoftReference<Node>(result); }
                }
            }

        return(result);
        }

    /**Create a factory to re-use for each parse; never null.
     * Hold a lock on this object while using it
     * in case the factory is not thread-safe.
     * It is assumed that multiple document builders created by this factory
     * are safe with respect to one another providing each is single-threaded.
     */
    private static final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

    /**Configure the DocumentBuilderFactory. */
    static
        {
        // Configure the factory object settings.
        factory.setValidating(false); // No validating (yet)...
        factory.setNamespaceAware(false); // No namespaces (yet)...
        factory.setIgnoringElementContentWhitespace(true); // Trim unnecessary whitespace.
        }

    /**Parse metadata from XML String format; never null.
     *
     * @param xml  well-formed non-null non-empty XML metadata
     * @return  non-null DOM node
     */
    private static Node parseMetadata(final String xml)
        throws IOException, SAXException, ParserConfigurationException
        {
        if((xml == null) || (xml.length() == 0))
            { throw new IllegalArgumentException(); }

        final DocumentBuilder builder;
        synchronized(factory) { builder = factory.newDocumentBuilder(); }

        // Parse the XML input String...
        final Document document = builder.parse(new InputSource(new StringReader(xml)));
        final Node result = document.getFirstChild();

        if(!Handler.TAG_NAME_METADATA_TOP.equals(result.getNodeName()))
            { throw new IllegalArgumentException("invalid top-level node"); }

        return(result);
        }

    /**If true then attempt in-memory ZIP compression of metadata to save space. */
    private static final boolean ATTEMPT_ROBA_COMPRESSION = false;
    
    /**If true then attempt in-memory compression of matadata to save space with Compact7BitString. */
    private static final boolean ATTEMPT_C7BS_COMPRESSION = true;    

    /**Attempt to compress supplied already-trimmed metadata; returns null if input is null.
     * Given the metadata in full XML form,
     * returns it trimmed, and compressed if possible.
     */
    private static Object _compressTrimmedMetadata(final String mdTrimmed)
        {
        if(mdTrimmed == null) { return(null); }

        // We never expect to be handed an empty String.
        assert(mdTrimmed.length() > 0);

        // Compact7BitString pretty-much guarantees at least a x2 memory saving
        // and is pretty light-weight on resources, so try it first.
        //
        // We use a private static dictionary to better compress metadata in memory.
        // We choose this dictionary to work well for typical data in this class,
        // but even a completely inappropriate dictionary does no harm
        // relative to having no dictionary at all.
        if(ATTEMPT_C7BS_COMPRESSION)
            {
            // If we encounter an error (eg non-ASCII data) just fall through.
            try { return(Compact7BitString.convertToCompact7BitString(mdTrimmed, sDict)); }
            catch(final Exception e) { /* Fall through if we cannot use this scheme. */ }
            }

        // Note that heavy use of ZIP/deflate really hurts the JVM, etc.
        if(ATTEMPT_ROBA_COMPRESSION)
            {
            // Don't even bother trying to compress small chunks of metadata
            // so as to save CPU cycles on every access
            // and probably some space in aggregated data on the wire...
            if(mdTrimmed.length() < FileTools.TYPICAL_DEFLATE_MIN_TEXT_SIZE_COMPRESSABLE)
                {
//System.out.println("[_compressTrimmedMetadata(): TOO SMALL.]");
                return(mdTrimmed);
                }
            else
                {
                // Try compressing the (trimmed) metadata,
                // and if it seems to be smaller (assuming 1 char == 2 bytes)
                // then store the compressed representation instead.
                final ROByteArray comp = ROByteArray.compressFromString(mdTrimmed);
//if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitPropsComparable: compressed metadata XML bytes "+comp.length()+" for exhibit "+esa+".]"); }
    
                // Use MemoryTools.intern() to eliminate duplicates.
                if(comp.length() < mdTrimmed.length() * 2)
                    {
//System.out.println("[_compressTrimmedMetadata(): compressed to "+comp.length()+" bytes from "+mdTrimmed.length()+" chars.]");
                    return(comp);
                    }
                else
                    {
//System.out.println("[_compressTrimmedMetadata(): UNCOMPRESSABLE.]");
                    return(mdTrimmed);
                    }
                }
            }

        // No compression applicable: return as-is.
        return(mdTrimmed);
        }
    
    /**Factory method to create a fully populated ExhibitPropsComputable object.
     * This is given the a data source from which it can fetch
     * the exhibit data one or more times to do its computations.
     * <p>
     * The data source object is not stored in the object.
     * <p>
     * This will veto any attempt at construction
     * if the data object is not fully loaded,
     * ie all available immediately with high bandwidth.
     * <p>
     * Underlying methods, such as getMetadata(), can throw all sorts of horrors,
     * such as an Error, so we try very hard to fail gracefully if this happens
     * and only throw an IOException.
     */
     public static ExhibitPropsComputable createExhibitPropsComputable(ExhibitStaticAttr esa, ExhibitDataSource ds) throws IOException
        {
        if((esa == null) || (ds == null)) { throw new IllegalArgumentException(); }

        // Compute image X,Y dimensions, if applicable.
        final ExhibitMIME.ExhibitTypeParameters et = (ExhibitMIME.getInputFileType(esa.filePath));

        // If there is no handler for whatever reason,
        // the give up trying to compute any details.
        // (Though this should not really happen.)
        if((et == null) || (et.handler == null)) { return(EMPTY); }

        // Don't attempt to get any stream access if the exhibit is not fully loaded.
        if(!ds.isExhibitFullyLoaded(esa))
            { throw new IOException("exhibit not fully loaded"); }

        final Dimension xyDimensions;
        final InputStream is1 = ds.getInputStream(esa);
        try { xyDimensions = et.handler.get2DImageDimensions(is1); }
        finally { is1.close(); }

        Object metadata = null;
        final InputStream is2 = ds.getInputStream(esa);
        try
            {
            // If we get no metadata back,
            // or the top node has no children,
            // then do not store any metadata.
            final Node n = et.handler.getMetadata(is2, esa.filePath);
            if((n != null) && n.hasChildNodes())
                {
                // We assume the top-level tag is the correct one.
                assert(Handler.TAG_NAME_METADATA_TOP.equals(n.getNodeName()));

                final String mdString = TextUtils.toXML(n, false, true);
                // Double-check that we can parse what we just got/generated...
                assert(null != parseMetadata(mdString));

                // Double-check that XML starts and ends with the top-level tag boilerplate.
                assert(mdString.startsWith(MDPREFIX));
                assert(mdString.endsWith(MDSUFFIX));

//if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitPropsComparable: raw metadata XML length "+mdString.length()+" for exhibit "+esa+".]"); }
                
                // Create a version of this the metadata String with the boilerplate removed.
                final String mdTrimmed = _trimStringFormOfOuterMetadataTag(mdString);

                // If deferring compaction then don't attempt it here.
                metadata = DEFER_COMPACTION ? mdTrimmed : _compressTrimmedMetadata(mdTrimmed);
                }
            }
        catch(final SAXException e)
            {
            // A SAX error suggests that we are having difficulty (such as bad characters or quoting)
            // with our XML text and thus should just pretend that we don't have any
            // until we can fix the code and try again.
            System.err.println("Trouble handling XML in extracting metadata for: " + esa.filePath);
            System.err.println("We regard this item as having no metadata for now, but correctly and completely computed.");
            System.err.println("Try again when the code has been fixed!");
            e.printStackTrace();
            }
        catch(final Throwable e)
            {
            // Note that we may get all manner of nasties thrown by metadata extractor routines...
            // This includes things derived from Error and RuntimeException.
            System.err.println("Trouble extracting metadata for: " + esa.filePath);
            e.printStackTrace();
            final IOException e2;
            if(e instanceof IOException) { e2 = (IOException) e; }
            else
                {
                e2 = new IOException("Trouble extracting metadata for: " + esa.filePath + ": " + e.getMessage());
                e2.initCause(e);
                }
            throw e2; // Veto construction; this item is not usable.
            }
        finally { is2.close(); }

        // DHD20060815: We observe about half the metadata instances to be duplicates
        // so we always try to intern() matadata for new instances.
        return(new ExhibitPropsComputable(xyDimensions, MemoryTools.intern(metadata)));
        }

    /**Remove explicit fixed outer tag to save a little space, or null if input is null.
     * Not for casual use, and not part of the supported API of this class.
     */
    public static String _trimStringFormOfOuterMetadataTag(final String mdString)
        {
        if(mdString == null) { return(null); }
        return mdString.substring(MDPREFIX_LENGTH, mdString.length() - MDSUFFIX_LENGTH);
        }

    /**Our serial version... */
    private static final long serialVersionUID = -8518817446313821597L;

    /**Deserialise: use constructor for validation, defensive copying, etc.
     * Also resolve all empty instances to a singleton as a minor optimisation,
     * and immediately intern() new values so as to immediately discard
     * duplicates (eg of exhibits already known) ASAP to minimise heap churn.
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Avoid duplicates of the empty instance.
        if(equals(EMPTY)) { return(EMPTY); }

        // Construct new instance of object in normal defensive way.
        // We try to (re)normalise/compress the metadata on loading if in String form.
        // This on-the-fly compression can be quite expensive.
        if(!DEFER_COMPACTION &&
           (ATTEMPT_C7BS_COMPRESSION  || ATTEMPT_ROBA_COMPRESSION) && 
           (metadata instanceof String))
            {
            // DHD20060815: We observe about half the metadata instances to be duplicates.
            return(new ExhibitPropsComputable(xyDimensions,
                            MemoryTools.intern(_compressTrimmedMetadata((String) metadata))));
            }

//        // If the data is a raw (and dangerously shareable/mutable) byte[],
//        // then wrap it for safety (it is intern()ed to avoid duplicates).
//        if(metadata instanceof byte[])
//            { return(new ExhibitPropsComputable(xyDimensions, Compact7BitString.makeFromInternalBytes((byte[]) metadata, null))); }

        // If the metadata is non-null, then make sure that we intern() it
        // unless deferring all compaction
        // on the assumption that the serialised form is reasonably duplicate-free.
        // DHD20060815: We observe about half the metadata instances to be duplicates.
        if(!DEFER_COMPACTION && (null != metadata))
            { return(new ExhibitPropsComputable(xyDimensions, MemoryTools.intern(metadata))); }

        // Return as-is.
        return(this);
        }

    /**Serialise: write in the best format for the wire.
     * To get best aggregate compressed size on the wire,
     * eg where the compressed stream contains many similar non-identical instances,
     * we always write out the metadata in its String format (minus the outer tags)
     * regardless of how it is actually held in memory.
     * This also makes us immune to changes in the the internals of the other formats
     * and allows use of a static dictionary with Compact7BitString for better in-memory compression
     * (ie effectively cross-instance compression).
     * <p>
     * This allows a stream compressor to effectively compress the redundancy between
     * instances of this class on the wire as well as internal redundancies.
     * <p>
     * We assume that there will almost never be identical instances on one stream
     * so we don't mind writing new copies each time where it does happen.
     */
    protected Object writeReplace()
        // throws ObjectStreamException
        {
        // Don't write multiple EMPTY instances.
        if(this.equals(EMPTY)) { return(EMPTY); }

        // If the metadata is null or String format then write this instance as-is.
        if((null == metadata) || (metadata instanceof String)) { return(this); }

        // Return a String-based instance for better inter-instance stream compression,
        // and for a more stable serialised representation.
        return(new ExhibitPropsComputable(xyDimensions,
                       MemoryTools.intern(_trimStringFormOfOuterMetadataTag(getMetadataAsXML()))));
        }

    /**If true, test semantics of metadata somewhat during object validation; can be very expensive. */
    private static final boolean VAL_MD_SEMANTICS = false;

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        // If xyDimensions is set make sure that its values are OK.
        if((xyDimensions != null) &&
           ((xyDimensions.width <= 0) || (xyDimensions.height <= 0)))
            { throw new InvalidObjectException("bad object: xyDimensions non-positive x,y"); }

        if(metadata != null)
            {
            try
                {
                String s = null;
                // Validate metadata in each legal representation.
                if(metadata instanceof Compact7BitString) // Deflated XML String
                    { s = metadata.toString(); }
                else if(metadata instanceof ROByteArray) // Deflated XML String
                    { s = ROByteArray.uncompressToString((ROByteArray) metadata); }
                else if(metadata instanceof String)
                    { s = (String) metadata; }

                if(s == null)
                    { throw new InvalidObjectException("bad object: metadata invalid format/type"); }
                if(!s.startsWith("<") || !s.endsWith(">"))
                    { throw new InvalidObjectException("bad object: metadata not expected XML format"); }
                if(s.startsWith(MDPREFIX) || s.endsWith(MDSUFFIX))
                    { throw new InvalidObjectException("bad object: invalid duplicate "+MDPREFIX+" outer tag"); }
                if(VAL_MD_SEMANTICS)
                    {
                    final Node n = getMetadata();
                    if(!Handler.TAG_NAME_METADATA_TOP.equals(n.getNodeName()))
                        { throw new InvalidObjectException("bad object: metadata invalid top level tag in String format"); }
                    }
                }
            catch(final Exception e)
                {
                e.printStackTrace();
                throw new InvalidObjectException("bad object: cannot parse XML metadata");
                }
            }
        }

    /**Equal if all the members are. */
    @Override public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof ExhibitPropsComputable)) { return(false); }
        final ExhibitPropsComputable other = (ExhibitPropsComputable) obj;
        if(null == xyDimensions)
            { if(null != other.xyDimensions) { return(false); } }
        else if(!xyDimensions.equals(other.xyDimensions)) { return(false); }

        if(null == metadata) { return(null == other.metadata); }
        if(null == other.metadata) { return(false); }

        // Before testing the full metadata XML, which is likely to be expensive,
        // we can reject these instance as not equal if their hashes are not.
        // Note that our computed hashes are cached for speed.
        if(hashCode() != other.hashCode()) { return(false); }

        // Do quick test on the internal format for equality if the types are the same...
        if(metadata.getClass().equals(other.metadata.getClass())) { return(metadata.equals(other.metadata)); }      
        // ...else test the full XML form (expensive, but canonical).
        // With luck we will only have to do this relatively rarely.
        return(getMetadataAsXML().equals(other.getMetadataAsXML()));
        }

    /**Human-readable summary. */
    @Override public String toString()
        {
        final StringBuilder sb = new StringBuilder();
        sb.append("<ExhibitPropsComputable");
        final String md = getMetadataAsXML();
        if(md != null) { sb.append(":metadata=").append(md); }
        if(xyDimensions != null) { sb.append(":xyDimensions=").append(xyDimensions); }
        sb.append(">");
        return(sb.toString());
        }

    /**Private cache of the hash value; initially zero, and zero for an EMPTY instance.
     * Thread-safe as int access guaranteed atomic.
     */
    private transient int _hash;

    /**Compute the hash value.
     * As well as being used for hash tables,
     * this will be used to detect changes in the representation of computed data,
     * ie for the overall AEP hash,
     * so this should include all significant elements.
     * <p>
     * Note that this also means that the AEP hash
     * contains a semantically-significant sampling of the exhibit data
     * as well as such things as length and timestamp from the AEID.
     */
    private int _computeHash()
        {
        // Compute the x,y dimension component of the hash...
        // Popular formats such as GIF and JPEG are limited to 2^16 pixels per side,
        // and indeed most such images are well below this limit,
        // so using a prime multiplier a little below this for one dimension
        // and adding to the other should achieve a reasonable hash spread.
        final int h1;
        if(null != xyDimensions)
            { h1 = xyDimensions.width ^ (xyDimensions.height * 65521); }
        else
            { h1 = 0; }

        // Compute the metadata component of the hash...
        if(null == metadata) { return(h1); }
        // If in the (trimmed) String format internally then use its hash directly.
        if(metadata instanceof String) { return(h1 ^ metadata.hashCode()); }
        // Else compute the hash of the synthesised trimmed XML String.
        return(h1 ^ _trimStringFormOfOuterMetadataTag(getMetadataAsXML()).hashCode());        
        }

    /**Hash based on the dimensions, else on the metadata. */
    @Override public int hashCode()
        {
        if(_hash == 0)
            { _hash = _computeHash(); /* Cache the hash value for speed. */ }
        return(_hash);
        }

    /**If true then defer compaction of metadata.
     * If true, then (any) compaction is <em>NOT</em> done
     * during construction or deserialisation,
     * but until some later point such as a call to compact().
     * <p>
     * This should save a lot of time constructing/deserialising data
     * before the first operations can be performed on it
     * and where memory space is not the primary constraint.
     */
    private static final boolean DEFER_COMPACTION = true;

    /**Compact the internal representation of this instance (and its sub-objects) if possible.
     * This has no effect on the logical content of this instance in-memory or serialised,
     * is guaranteed to be safe to run concurrently with other uses of this instance
     * (and will take any locks as needed to work incrementally),
     * and may do nothing but consume some CPU cycles.
     * <p>
     * This may be able to convert some state to a more memory-efficient representation
     * after construction or deserialisation,
     * and is suitable to call in a background thread.
     * <p>
     * We don't prevent multiple concurrent calls to this routine,
     * since they are at worst wasteful of CPU but not unsafe.
     */
    public void compact()
        {
        if(!DEFER_COMPACTION) { return; /* Nothing to do. */ }

        final Object md = metadata; // Capture snapshot.

        // If already in compact (non-String) form (or null) then intern/return.
        if(!(md instanceof String))
            {
            metadata = MemoryTools.intern(md);
            return;
            }

        // Replace metadata with intern()ed compact form if possible.
        // DHD20060815: We observe about half the metadata instances to be duplicates.
        metadata = MemoryTools.intern(_compressTrimmedMetadata((String) md));
        }
    }
