/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.ai.scorer.ScoreAndConf;
import org.hd.d.pg2k.ai.scorer.ScorerCacheIF;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.*;

import ORG.hd.d.IsDebug;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.Serializable;
import java.util.*;


/**Immutable (and serialisable) store of all ephemeral computable auxiliary properties of a single exhibit.
 * These are properties that are computed from exhibit and non-exhibit
 * data and that need to be periodically recomputed,
 * such as popularity score.
 * <p>
 * This is is called "mutable" because although any one instance is immutable,
 * successive values retrieved for the same exhibit may be different.
 * <p>
 * This data has limited lifetime,
 * but may be somewhat useful even when stale,
 * so can be persisted if desired.
 * <p>
 * This is designed to be efficient and robust on the wire and in memory,
 * since these details will be held for each and every exhibit.
 */
public final class ExhibitPropsComputableMutable implements Serializable,
                                                            ObjectInputValidation
    {
    /**Limit of goodness*correlation to consider something (eg category) significantly good or bad. */
    public static final float GOODBAD_LIMIT = 0.4f;

    /**Limit of goodness*correlation to consider something (eg category) significantly good or bad, is int value. */
    public static final int GOODBAD_LIMIT_INT = Math.round(GOODBAD_LIMIT * Integer.MAX_VALUE);

    /**Compute a "quick"/approximate value for a specified exhibit; never null.
     * This uses only values from the exhibit static data
     * plus the system properties object
     * to compute a fast approximation to a full value.
     * <p>
     * The value is immediately marked as stale
     * (unless it is possible to compute the exact value fast),
     * but is useful to generate something quickly if the system is busy,
     * or at start-up for example.
     *
     * @param esa  the exhibit to compute properties for; never null
     * @param gp   the system properties; never null
     *
     * @return non-null quick approximation
     */
    public static ExhibitPropsComputableMutable generateFastApproximation(
                                        final ExhibitStaticAttr esa,
                                        final GenProps gp)
        {
        return(compute(esa, gp, null, null, null, null));
        }

    /**Compute (accurate) value for a specified exhibit; never null.
     * This uses values from the exhibit static data
     * plus the system properties object
     * plus the data source (and thus might take a long time and be expensive)
     * to compute an accurate/full value.
     * <p>
     * Providing the supplied data source is non-null and functional,
     * and the GenProps does not appear to be empty (zero timestamp)
     * then the value will not become stale for at least half of
     * MAX_AGE_MS, and will become stale in no more than MAX_AGE_MS ms.
     * <p>
     * If the data source is not available/functional then
     * an approximation will be returned instead.
     *
     * @param esa  the exhibit to compute properties for; never null
     * @param gp  the system properties; never null
     * @param aep  the exhibit properties; can be null
     * @param vars  source of event data; can be null
     * @param voteCache  vote cache; can be null
     *
     * @return non-null full value or approximation
     *     depending on the available data sources
     */
    public static ExhibitPropsComputableMutable compute(final ExhibitStaticAttr esa,
                                                        final GenProps gp,
                                                        final AllExhibitProperties aep,
                                                        final BasicVarMgrInterface vars,
                                                        final ExhibitPropsComputableMutableVoteCacheIF voteCache,
                                                        final ScorerCacheIF scorers)
        {
        return(new ExhibitPropsComputableMutable(esa, gp, aep, vars, voteCache, scorers));
        }

    /**Create an instance with the desired properties. */
    private ExhibitPropsComputableMutable(final long staleAfter, final int goodness)
        {
        this.staleAfter = staleAfter;
        this.goodness = goodness;

        // Verify object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Trivially-stale entirely-neutral instance. */
    public static final ExhibitPropsComputableMutable TRIVIAL_NEUTRAL = new ExhibitPropsComputableMutable(0, 0);

    /**Create a fully populated ExhibitPropsComputableMutable object.
     * This is given the data sources from which it can fetch
     * the exhibit data one or more times to do its computations.
     * <p>
     * The data source references are not stored in the object.
     *
     * @param esa  the exhibit to compute properties for; never null
     * @param gp  the system properties; never null
     * @param aep  the exhibit properties; can be null
     * @param vars  source of event data; can be null
     * @param scorers  source of content-based sceoring; can be null
     * @param voteCache  vote cache; can be null
     */
    private ExhibitPropsComputableMutable(final ExhibitStaticAttr esa,
                                          final GenProps gp,
                                          final AllExhibitProperties aep,
                                          final BasicVarMgrInterface vars,
                                          final ExhibitPropsComputableMutableVoteCacheIF voteCache,
                                          final ScorerCacheIF scorers)
        {
        if((esa == null) || (gp == null))
            { throw new IllegalArgumentException(); }

        // Initial set of data-dependent Factor items.
        final List<Factor> components = new ArrayList<Factor>();


        // Calculate factors that depend on access to event histories.
        boolean dataAccessTrouble = true; // Note if 'stale'/partial because data not complete...
        if((aep != null) && (vars != null) && (voteCache != null) && (scorers != null))
            {
            try
                {
                // Compute synthetic factors based on access stats.
                components.addAll(calcAccessFactors(esa.filePath, aep, vars));

                // Compute factor based on explicit voting for this exhibit.
                final Factor vote = voteCache.calcVoteFactor(esa.filePath, aep, vars);
                // We use the minimum of the weighting and the confidence
                // to realistally boost the vote contribution as much as possible.
                components.add(new Factor(vote.goodness, Math.min(POP_VOTES, vote.correlation), vote.comment));

                // Add in factor, if any, from AI Scorers, weighting by confidence.
                // We allow possibly-slightly-stale Scorer values (assumed still to be useful)
                // so as not to be pole-axed when the date/period rolls.
                // At start-up we may simply not have all Scorers loaded and assessed.
                final ScoreAndConf compositeScoreAndConf = scorers.computeCompositeScoreAndConfidence(esa.filePath, true);
                // Only bother even computing the Factor if we have a non-zero confidence...
                if(compositeScoreAndConf.confidence != 0)
                    {
                    // We use the minimum of the weighting and the confidence
                    // to realistally boost the Scorer contribution as much as possible.
                    final Factor scorerFactor = new Factor(compositeScoreAndConf.score / (float) ScoreAndConf.MAX,
                                                           Math.min(POP_AI_SCORER, compositeScoreAndConf.confidence / (float) ScoreAndConf.MAX));
                    components.add(scorerFactor);
if(IsDebug.isDebug) { System.out.println("[EPCM: scorerFactor=" + scorerFactor + " for " +esa.filePath + ".]"); }
                    }

                // Compute factors based on vote correlations, etc.
                components.addAll(calcCorrelationFactors(esa, aep, vars, voteCache));

                // Managed to gather all the data required.
                dataAccessTrouble = false;
                }
            catch(final IOException e) { /* Absorb error. */ }
            }


        // Compute the full goodness value
        // and scale it to the form we hold it internally.
        final Factor gf = _computeTotalGoodnessFactor(esa, gp, aep, components);
        goodness = Math.max(-Integer.MAX_VALUE, Math.min(Integer.MAX_VALUE,
            Math.round(Integer.MAX_VALUE * gf.goodness)));

        // Data is marked stale if any data source is null/empty/default
        // or if we had trouble fetching some data dynamically,
        // else data is non-stale and we pick a random expiry time between half and maximum.
        staleAfter = (dataAccessTrouble ||
                (aep == null) || (aep.longHash == 0) ||
                (gp.timestamp == 0)) ? 0 :
            ((System.currentTimeMillis() + (MAX_AGE_BEFORE_STALE_MS/2)
                                         + Rnd.fastRnd.nextInt(MAX_AGE_BEFORE_STALE_MS/2)));

        // Verify object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**If true then force correlation data up to date.
     * If false then accept somewhat stale or missing data,
     * since this only changes slowly and a complete absence is neutral.
     */
    private static final boolean FORCE_CORR_RECOMP = true;

    /**Compute factors based on correlations.
     * These have a confidence pre-scaled in the range 0 to POP_VOTE_CORR.
     */
    private static List<Factor> calcCorrelationFactors(final ExhibitStaticAttr esa,
                    final AllExhibitProperties aep,
                    final BasicVarMgrInterface vars,
                    final ExhibitPropsComputableMutableVoteCacheIF voteCache)
        throws IOException
        {
        // Get the raw factors; will need scaling.
        // Force retrieved values to be (reasonably) up-to-date.
        final Factor[] rawFactors = voteCache.getCorrelates(esa, aep, vars, FORCE_CORR_RECOMP);

        // Full set of scaled values.
        // We scale the certainty down in relation to other factors.
        final List<Factor> result = new ArrayList<Factor>(rawFactors.length);
        for(final Factor f : rawFactors)
            { result.add(new Factor(f.goodness, f.correlation * POP_VOTE_CORR, f.comment)); }

        return(result);
        }

    /**Time until which this data is considered valid; beyond this time is stale.
     * If zero, or before the current time of day, the data is stale.
     * <p>
     * If the stale date is too far in the future for us to believe,
     * ie much more than would be allowed with this version of the object,
     * then we treat the data as stale in case it is from a version of this
     * object with very different staleness limits or there was a persistence error.
     */
    private final long staleAfter;

    /**If true, the data in this object is stale.
     * The data may still be correct,
     * but should be treated as an approximation,
     * and recomputed as soon as possible.
     */
    public boolean isStale()
        {
        // Object was immediately stale.
        if(staleAfter == 0) { return(true); }

        final long now = System.currentTimeMillis();

        // Object expired.
        if(staleAfter < now) { return(true); }

        // Suspicious stale date much too far in future, so treat data as stale.
        if(staleAfter >= now + 5*MAX_AGE_BEFORE_STALE_MS) { return(true); }

        // Looks OK!
        return(false);
        }


    /**If true, the data in this object is not only stale; it is only even a fast approximation.
     * The data is probably not worth persisting, for example,
     * since it was not fully computed and could probably be very quickly recomputed.
     */
    public boolean isTriviallyStale()
        {
        // Object was immediately stale.
        return(staleAfter == 0);
        }

    /**Maximum age (in ms) of instances before considering them stale and recomputing them.
     * <em>Slightly</em> (much less than a factor of two) different limit value
     * between <em>class</em>/JVM instances.
     * <p>
     * Each instance of this object should pick a random time
     * between about half of this value and the full value before going stale
     * in order to spread out refresh computations over a reasonable interval.
     * Recomputing these values may prove quite expensive.
     * <p>
     * This value is selected so that most exhibits will not have changed
     * significantly in the given interval;
     * something of the order of tens of hours to tens of days is good.
     */
    public static final int MAX_AGE_BEFORE_STALE_MS = 46 * 3600 * 1000 +
        Rnd.fastRnd.nextInt(4 * 3600 * 1000);


    /**The "goodness" score, -MAX_VALUE is maximally bad and MAX_VALUE maximally good, zero is neutral.
     * This is a composite of many factors,
     * including a small random element.
     * <p>
     * Has enough significant digits to allow a total ordering
     * even over a very large number of exhibits.
     * <p>
     * This is nominally logically thought of in the range [-1.0, +1.0].
     */
    private final int goodness;

    /**Get the "goodness" score, -MAX_VALUE is maximally bad and MAX_VALUE maximally good, zero is neutral.
     * This gets the value as a raw int value;
     * useful for fast compares, sorts, etc.
     */
    public int getGoodness()
        { return(goodness); }

    /**Get the "goodness" score as a normalised float in the range -1 (bad) to +1 (good), 0 is neutral.
     * This is represented as a float which can lose precision/information,
     * but may be easier to work with for display and some computations.
     */
    public float getGoodnessAsFloat()
        {
        return((float) (((double) goodness) / Integer.MAX_VALUE));
        }

    /**Find out if this exhibit is rated "good"/popular or not.
     * Returns TRUE if rated good, FALSE if bad,
     * null if not significantly either (ie too close to neutral) or if unknown.
     * <p>
     * This is based on the computed "goodness" score.
     */
    public Boolean isGood()
        {
        if((goodness < GOODBAD_LIMIT_INT) && (goodness > -GOODBAD_LIMIT_INT)) { return(null); }
        return((goodness >= 0) ? Boolean.TRUE : Boolean.FALSE);
        }

    /**Maximum weighting for vox pop user votes.
     * Should be large but possibly not overwhelming,
     * eg because of pranks to attempt to "fix" the votes.
     * <p>
     * The vote element can count for or against an exhibit (ie be +ve or -ve),
     * and an absence of any votes may tend to dilute goodness back towards neutral.
     */
    public static final float POP_VOTES = 0.7f;

    /**Maximum weighting from correlated voting on related exhibits.
     * Possibly quite a crude measure,
     * though will be more valuable with more data available,
     * and so if all our required samples are in place,
     * then we'll give it a rating that should be visible to users,
     * and more significant than just newness, for example.
     */
    public static final float POP_VOTE_CORR = 0.1331f;

    /**Maximum weighting from Scorer judgement of exhibit content.
     * This AI-based mechanism inspects the content of each exhibit,
     * and is driven to get the best possible fit with voting data.
     * <p>
     * We give this equal weighting with the metadata-driven statistical methods.
     */
    public static final float POP_AI_SCORER = POP_VOTE_CORR;

    /**Maximum weighting for exhibit recently viewed or downloaded.
     * Possibly quite a crude measure,
     * though will be more valuable with more data available,
     * and so if all our required samples are in place,
     * we'll give it a rating that should be visible to users,
     * and more significant than just newness, for example.
     * <p>
     * We make this a little less important than vote correlations.
     */
    public static final float POP_RECENT_ACCESS = POP_VOTE_CORR * 0.73f;

    /**Weighting/correlation of "newness" component of popularity.
     * This is only applied as a positive (proportional) factor
     * for exhibits considered "new",
     * since newer items are potentially more interesting.
     * <p>
     * This factor should be relatively small,
     * not overwhelming manually-set weightings for example,
     * and mainly intended as a tie-breaker.
     */
    public static final float POP_CORR_NEWNESS = 0.001f;

    /**Weighting/correlation of "random" component of popularity.
     * Exists to help break ties and perturb the popularity ordering a little.
     * Should be very small so as not to outweigh any genuine factor,
     * but more-or-less force a total ordering.
     * <p>
     * This factor falls equally either side of zero, is a +/- limit.
     */
    public static final float POP_CORR_RANDOM = 0.00001f;

    /**Weighting/correlation of has-description component of popularity.
     * Applied to items that have a description
     * as slightly more interesting than those without.
     * <p>
     * Should generally be greater than the randomness weighting,
     * and probably less than the newness rating.
     */
    public static final float POP_CORR_HASDESC = Math.min(POP_CORR_RANDOM * 10,
                                                          POP_CORR_NEWNESS / 3);

    /**Weighting factor for an item with a specific description. */
    private static final Factor FACTOR_HASDESC =
            new Factor(1, POP_CORR_HASDESC);

    /**Weighting factor for an item with a generic description.
     * Carries less weight than having a specific description.
     * <p>
     * May apply to having location or AKA or description text.
     */
    private static final Factor FACTOR_HASGENERICDESC =
            new Factor(1, POP_CORR_HASDESC/2);

    /**Single (immutable) component of multi-factor value.
     * Consists of:
     * <ul>
     * <li>a goodness value [-1,+1]
     * <li>a correlation/confidence rating [-1,+1]
     * </ul>
     * <p>
     * The value added into the final value is proportional to:
     * <samp>(goodness * correlation) / |correlation|</samp>.
     * <p>
     * Uses float values since there is probably not even that much
     * real precision available, and doing so probably saves space and time.
     * Double values can be passed to the constructors,
     * but the extra precision is discarded.
     */
    static public final class Factor
        {
        public Factor(final float goodness, final float correlation)
            { this(goodness, correlation, null); }
        public Factor(final float goodness, final float correlation, final String comment)
            {
            if(!(goodness >= -1) && (goodness <= +1) ||
               !(correlation >= -1) && (correlation <= +1))
                { throw new IllegalArgumentException(); }
            this.goodness = goodness;
            this.correlation = correlation;
            this.comment = comment;
            }
        /**Goodness contributed by this factor: range [-1,+1]. */
        public final float goodness;
        /**Correlation of this value with overall result: range [-1,+1]. */
        public final float correlation;
        /**Optional comment fields; can be null. */
        public final String comment;

        /**We will accept double values, but discard the extra precision. */
        public Factor(final double goodness, final double correlation)
            { this((float) goodness, (float) correlation, null); }
        /**We will accept double values, but discard the extra precision. */
        public Factor(final double goodness, final double correlation, final String comment)
            { this((float) goodness, (float) correlation, comment); }

        /**Convert to human-readable form. */
        public String toString()
            {
            final StringBuilder sb = new StringBuilder();
            sb.append("Factor");
            sb.append(":goodness=").append(goodness);
            sb.append(":correlation=").append(correlation);
            if(comment != null)
                { sb.append(":comment=").append(comment); }
            return(sb.toString());
            }
        }

    /**Zero factor; zero weight, zero goodness. */
    public static final Factor FACTOR_ZERO = new Factor(0, 0);

    /**Compute random "goodness" factor symmetrical about zero; never null. */
    private static Factor _computeRandomGoodnessFactor()
        {
        return new Factor((Rnd.fastRnd.nextBoolean() ?
                              Rnd.fastRnd.nextDouble() : -Rnd.fastRnd.nextDouble()),
                          POP_CORR_RANDOM);
        }

    /**Compute newness bonus (if any); never null.
     * If not new enough to be considered "new" then a zero factor is returned.
     * <p>
     * This is not necessarily precisely aligned with the "isExhibitNew()" result,
     * though except in race conditions they should be aligned.
     *
     * @param esa  exhibit data; never null
     * @return factor, positive but low weight if exhibit "new",
     *     else zero with zero weighting
     */
    private static Factor _computeNewnessBonusFactor(final ExhibitStaticAttr esa)
        {
        // Also for some wobble is server time vs new-exhibit timestamps.
        final long age = Math.max(0, System.currentTimeMillis() - esa.timestamp);
        if(age <= CoreConsts.MAX_EXHIBIT_AGE_NEW_MS)
            {
            // A very new exhibit has a goodness of nearly 1.0.
            // An exhibit only just "new" has a goodness of just over 0.0.
            final double goodness = (CoreConsts.MAX_EXHIBIT_AGE_NEW_MS - age) /
                    (double) CoreConsts.MAX_EXHIBIT_AGE_NEW_MS;
            return(new Factor(goodness, POP_CORR_NEWNESS));
            }

        // Exhibit not "new", so zero goodness but the same weight.
        return(_NOT_NEW);
        }

    /**Contribution for goodness of non-new exhibits; zero goodness but the same "newness" weight. */
    private static final Factor _NOT_NEW = new Factor(0, POP_CORR_NEWNESS);

    /**Compute full goodness factor, using full data source if available; never null.
     *
     * @param esa  exhibit data; never null
     * @param initialComponents  initial set of (data-dependent) Factor elements;
     *     possibly empty but never null
     *
     * @return factor, positive but low weight if exhibit "new", else zero with zero weighting
     */
    private static Factor _computeTotalGoodnessFactor(final ExhibitStaticAttr esa,
                                                      final GenProps gp,
                                                      final AllExhibitProperties aep,
                                                      final List<Factor> initialComponents)
        {
        // Compute a list of components of goodness Factors.
        final List<Factor> components = new ArrayList<Factor>(initialComponents);

        // Compute the fast components that do not require a
        // (potentially slow) data source.
        components.add(_computeRandomGoodnessFactor());
        components.add(_computeNewnessBonusFactor(esa));
        // Factor in the author weight, if any.
        final String exhibitName = esa.filePath;
        final Byte aW = gp.getPopWeightForAuth(ExhibitName.getAuthorComponent(exhibitName));
        if(aW != null)
            { components.add(new Factor(1, aW.doubleValue() / GenProps.MAX_POPWT_VAL)); }
        // Factor in the file-type weight, if any.
        final Byte tW = gp.getPopWeightForType(ExhibitName.getExtensionComponent(exhibitName));
        if(tW != null)
            { components.add(new Factor(1, tW.doubleValue() / GenProps.MAX_POPWT_VAL)); }
        // Penalise one-word exhibits (ignoring number-in-series) for not being descriptive enough.
        final Set<String> noAttrWords = Collections.emptySet();
        if(ExhibitName.getMainWordsComponent(exhibitName, noAttrWords).indexOf(ExhibitName.WORD_SEP) == -1)
            { components.add(new Factor(-1, POP_CORR_HASDESC)); }
        // Slight penalty to exhibits that *cannot* show thumbnails, since images are most accessible.
        if(!(ExhibitMIME.getInputFileType(esa.filePath)).canPossiblyCreateThumbnailOfSameMIMEType())
            { components.add(new Factor(-1, POP_CORR_HASDESC/3)); }


        // Do the full (but potentially-expensive) goodness calcs
        // iff we have an apparently-working data source.
        if(aep != null)
            {
            // Factor in weight from attribute words.
            final Enumeration aWEn =
                ExhibitName.getAttributeWordsComponentEnumeration(exhibitName, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
            if(aWEn != null)
                {
                while(aWEn.hasMoreElements())
                    {
                    final String attrWord = (String) aWEn.nextElement();
                    final Byte attrW = gp.getPopWeightForAttr(attrWord);
                    if(attrW != null)
                        { components.add(new Factor(1, attrW.doubleValue() / GenProps.MAX_POPWT_VAL)); }
                    }
                }

            // Factor in weight from description, if any.
            final ExhibitPropsLoadable epl = aep.getExhibitPropsLoadable(exhibitName);
            if(epl.hasDescription())
                { components.add(FACTOR_HASDESC); }

            // Having a location is a generic description...
            if(Location.NONE != aep.getLocation(exhibitName))
                { components.add(FACTOR_HASGENERICDESC); }

//            // Look for tree-based generic description...
//            final String treeDesc = WebUtils.getLocalisedTreeDesc(exhibitName,
//                                                                  aep,
//                                                                  localeBean, true, false);
//            if(!"".equals(treeDesc))
//                { components.add(FACTOR_HASGENERICDESC); }


            // TODO: factor in computed weights from correlations.

            }

        // Now compute the weighted goodness value...
        double g = 0;
        double w = 0;
        for(final Iterator<Factor> it = components.iterator(); it.hasNext(); )
            {
            final Factor f = it.next();
            w += Math.abs(f.correlation);
            g += f.goodness * f.correlation;
            }

        // Adjust so final weight is 1.0,
        // and the cumulative value is divided by at least that.
        final double wPrime = Math.max(1.0, w);
        return(new Factor(g / wPrime, 1.0));
        }


    /**Minimum sample period to even out short cycles in Web access patterns, in ms; strictly positive.
     * Must usually be a multiple (and minimum) of a week to allow for most common
     * daily and weekly access patterns to be smoothed as well as possible.
     */
    private static final long _SAMPLE_CYCLE_PERIOD_MINOR_MS = 7 * 24 * 3600 * 1000L;

    /**The longest cycle time we will look for in historical data, in ms; strictly positive.
     * This is usually a year to look for seasonal patterns,
     * eg what is popular at Christmas or in the summer.
     */
    private static final long _SAMPLE_CYCLE_PERIOD_MAJOR_MS = 365242L * 24 * 3600; // 365.242 days in a year.

    /**Number of VLONG samples corresponding to annual cycles in data. */
    private static final int _SEASONAL_VLONG_CYCLE_PERIODS = Math.max(1, (int)
                Math.round(_SAMPLE_CYCLE_PERIOD_MAJOR_MS / (double) EventPeriod.VLONG.getIntervalMs()));

    /**Minimum number of VLONG samples to smooth out daily/weekly cycles in data. */
    private static final int _MIN_VLONG_SAMPLE_PERIODS = Math.max(1, (int)
                Math.round(_SAMPLE_CYCLE_PERIOD_MINOR_MS / (double) EventPeriod.VLONG.getIntervalMs()));


    /**Compute Factor(s) that depend on access data.
     * If this cannot fetch some of the data it requires due to an IOException
     * (not the same as fetching data successfully that has no entries)
     * then the exception is propagated to the caller.
     * (Normally this results in the EPCM item for which this is called being marked stale
     * so that we can try again when the data is available.)
     * <p>
     * The comment fields of the factors produced are
     * the event names from which they were generated.
     * <p>
     * This is public to assist testing.
     *
     * @throws IOException  if this was unable to fetch some data required
     */
    public static Collection<Factor> calcAccessFactors(final String exhibitName,
                                                       final AllExhibitProperties aep,
                                                       final BasicVarMgrInterface vars)
        throws IOException
        {
        if((exhibitName == null) || (aep == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        // Get short name of exhibit,
        // by which most stats are recorded.
        final String shortName = ExhibitName.getFileComponent(exhibitName);

        // Factor in weight from access stats...
        final int nExhibits = aep.aeid.size();

        // Threshold at which usage is considered significantly "good".
        final int rankLimit = nExhibits / 3;

        // Set of slots/intervals in which we will sample access-pattern stats.
        // We let the routine force some solid coverage at the near end.
        final BitSet whichIntervals = _computeSampleBitSet(0);

        // Count the full set of samples that we are looking for.
        final int desiredSamples = whichIntervals.cardinality();

        final long prevInterval = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis()) - 1;



        // Access stats event definitions which we want and that record count by short exhibit name.
        final List<SimpleVariableDefinition> eventVarNames = new ArrayList<SimpleVariableDefinition>(3);
        eventVarNames.add(SystemVariables.ACCESSPATTERN_CAT_PAGE_VIEW);
        eventVarNames.add(SystemVariables.ACCESSPATTERN_CLICKTHROUGH);
        // Only look for downloads if not automatically excluded from beng counted.
        // (Note that this definition may vary between data being collected and being used.)
        final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);
        if((esa != null) && !ImageUtils.canBeOwnThumbnail(esa))
            { eventVarNames.add(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD); }

        // Create/size the results container...
        final List<Factor> components = new ArrayList<Factor>(eventVarNames.size());

        // Retrieve data for each event...
        // Currently we we give all intervals the same weighting.
        // This is especially dodgy for the current interval, which may be very short and unrepresentative.
        // We give total weighting in proportion to the square root of the number of samples obtained.
        // Shuffle the order in which we fetch the data to improve concurrency a little.
        Collections.shuffle(eventVarNames, Rnd.fastRnd);
        for(final SimpleVariableDefinition def : eventVarNames)
            {
            assert(def.isEvent() && (def.getType() == SimpleVariableDefinition.TYPE_STRING));
            final EventVariableValue evvs[] = vars.getEventValues(def,
                                                                  EventPeriod.VLONG,
                                                                  prevInterval,
                                                                  whichIntervals);
            assert(evvs != null); // May be empty, but never null.
            // Count the number of results we actually retrieved.
            int resultsFound = 0;
            double totalScore = 0; // Cumulative score so far...
            for(int i = whichIntervals.nextSetBit(0); i >= 0; i = whichIntervals.nextSetBit(i+1))
                {
                if(i >= evvs.length) { break; /* Run out of answers... */ }

                final EventVariableValue evv = evvs[i];

                // Ignore null values, and those with *no* events at all for any value.
                if(evv == null) { continue; }
                if(evv.getTotalEventCount() < 1) { continue; }

                // OK, have a real result; we can see if this exhibit ranks well or not.
                ++resultsFound;

                // If the event bucket may have "overflowed" (too many distinct event values)
                // then we should probably ignore count==1 entries as unreliable/incomplete.
                final boolean possibleOverflow =
                        (evv.getTotalDistinctValues() >= evv.getDef().getMaxDiffEventCount());

                final int rank = evv.getRank(shortName);
                if((rank < rankLimit) &&
                   (!possibleOverflow || (evv.getCount(shortName) > 1)))
                    { totalScore += (rankLimit - rank) / (double) rankLimit; }
                }

            // We should always get at least one response.
            // Maybe we should abort with an error if not.
//            if(resultsFound < 1) { throw new IOException("not even current results found"); }

            // Compute goodness score.
            final double g = totalScore / Math.max(1, resultsFound);
            // Compute confidence weighting based on number of samples.
            final double confidence = Math.sqrt(resultsFound / (double) desiredSamples);
//if(IsDebug.isDebug) { System.out.println("***EPCM._calcAccessFactors(): score/resultsFound/confidence/exhibit/event = " +g+", " +resultsFound+", " +confidence+", " +exhibitName+", " +def.getName()); }
            components.add(new Factor(g, POP_RECENT_ACCESS * confidence, def.getName()));
            }

        return(components);
        }

    /**If true then include the "all" buckets in our calculations. */
    private static final boolean USE_ALL_BUCKET = false;

    /**Compute exactly one Factor that depends on explicit user votes for specified exhibit; never null.
     * Because we expect voting to be sparse,
     * we total up votes across all sample intervals to reduce noise,
     * rather than taking each sample separately,
     * and do various other calculations a little differently to dense measures.
     * <p>
     * If this cannot fetch some of the data it requires due to an IOException
     * (not the same as fetching data successfully that has no entries)
     * then the exception is propagated to the caller.
     * (Normally this results in the EPCM item for which this is called being
     * marked stale so that we can try again when the data is available.)
     * <p>
     * This is public to assist testing and computation of correlations.
     * <p>
     * The returned factor's goodness can range from -1 to +1,
     * and confidence from 0 to +1;
     * any scaling required will have to be applied elsewhere.
     * 
     * @param minNearMs minimum interval (ms) at near end
     *     to include all available samples for
     *
     * @throws IOException  if this was unable to fetch some data required
     */
    public static Factor calcVoteFactor(final String exhibitName,
                                        final AllExhibitProperties aep,
                                        final BasicVarMgrInterface vars,
                                        final long minNearMs)
        throws IOException
        {
        if((exhibitName == null) || (aep == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        if(null == aep.aeid.getStaticAttr(exhibitName))
            { throw new IllegalArgumentException("no such exhibit"); }

        // Get short name of exhibit,
        // by which most stats are recorded.
        final String shortName = ExhibitName.getFileComponent(exhibitName);

        // Set of slots/intervals in which we will sample voting behaviour.
        // Insist on over a month's worth of near-term votes if possible.
        final BitSet whichIntervals = _computeSampleBitSet(Math.max(minNearMs,
                                                                    33 * 24 * 3600 * 1000L));

        // Count the full set of samples that we are looking for
        // optionally including the "all" set assumed always available.
        final int desiredSamples = whichIntervals.cardinality() + (USE_ALL_BUCKET?1:0);

        // Last available complete interval/period.
        final long prevInterval = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis()) - 1;

        // Get votes "for" this exhibit.
        final EventVariableValue vPro[] = vars.getEventValues(SystemVariables.VOTE_PRO,
                                                              EventPeriod.VLONG,
                                                              prevInterval,
                                                              whichIntervals);
        // Get votes "against" this exhibit.
        final EventVariableValue vCon[] = vars.getEventValues(SystemVariables.VOTE_CON,
                                                              EventPeriod.VLONG,
                                                              prevInterval,
                                                              whichIntervals);

        // Set of all event values seen in all sampled slots.
        // We use a HashSet for speed.
        final HashSet<Object> allVoteValues = new HashSet<Object>();
        // Count of all votes in all slots.
        int allVotesCount = 0;

        int resultsFound = 0; // Total results usable including "all" slot.
        int totalProVotesThisEx = 0; // Total votes for this exhibit in the sample slots.
        int totalConVotesThisEx = 0; // Total votes against this exhibit in the sample slots.

        if(USE_ALL_BUCKET)
            {
            // Get the "all" slot votes "for"...
            final EventVariableValue[] aPro = vars.getEventValues(SystemVariables.VOTE_PRO,
                                        EventPeriod.VLONG,
                                        0,
                                        null);

            // Get the "all" slot votes "against"...
            final EventVariableValue[] aCon = vars.getEventValues(SystemVariables.VOTE_CON,
                                        EventPeriod.VLONG,
                                        0,
                                        null);
            // Regard the "all" slot as always present if we're using it
            // (which they logically are, even if not necessarily programmatically).
            ++resultsFound;
            if((aPro.length > 0) && (aPro[0] != null))
                {
                allVoteValues.addAll(aPro[0].getInfo().keySet());
                allVotesCount += aPro[0].getTotalEventCount();
                totalProVotesThisEx += aPro[0].getCount(shortName);
                }
            if((aCon.length > 0) && (aCon[0] != null))
                {
                allVoteValues.addAll(aCon[0].getInfo().keySet());
                allVotesCount += aCon[0].getTotalEventCount();
                totalConVotesThisEx += aCon[0].getCount(shortName);
                }
            }

        for(int i = whichIntervals.nextSetBit(0); i >= 0; i = whichIntervals.nextSetBit(i+1))
            {
            final EventVariableValue pro = (i < vPro.length) ? vPro[i] : null;
            final EventVariableValue con = (i < vCon.length) ? vCon[i] : null;

            // Ignore slot where no votes were collected either way for any exhibit at all.
            final int proTEC = (pro == null) ? 0 : pro.getTotalEventCount();
            final int conTEC = (con == null) ? 0 : con.getTotalEventCount();
            final int tEC = proTEC + conTEC;
            if(tEC < 1) { continue; }

            // OK, have a real result; we can see if this exhibit has attracted any votes.
            ++resultsFound;

            // Capture all the distinct event values in this slot
            // so that we know all the exhibits that have attracted votes
            // in the slots that we have sampled.
            if(pro != null) { allVoteValues.addAll(pro.getInfo().keySet()); }
            if(con != null) { allVoteValues.addAll(con.getInfo().keySet()); }
            allVotesCount += tEC;

            // Sum up the total votes either way for this exhibit.
            if(pro != null) { totalProVotesThisEx += pro.getCount(shortName); }
            if(con != null) { totalConVotesThisEx += con.getCount(shortName); }
            }

        // Take a simplistic view that the overall vote
        // is simply the mean of all those cast.
        final int totalScoreThisEx = totalProVotesThisEx - totalConVotesThisEx;
        final int totalVotesCastThisEx = totalProVotesThisEx + totalConVotesThisEx;

        // Compute goodness score.
        final double g = Math.min(1, Math.max(-1, totalScoreThisEx / (double) Math.max(1, totalVotesCastThisEx)));

        // Compute average number of votes per voted-for-exhibit
        // over all the sampled slots with at least one vote...
        // FIXME: should re-do this to allow for "lost" vote events,
        // else we end up overestimating the votes-per-voted-for-exhibit value
        // and this underrating even the most-voted-for exhibit(s).
        final double meanVotesPerEx = allVotesCount / Math.max(1.0, allVoteValues.size());

//System.out.println("***EPCM.calcVoteFactor(): allVotesCount = " + allVotesCount);
//System.out.println("***EPCM.calcVoteFactor(): allVoteValues.size() = " + allVoteValues.size());
//System.out.println("***EPCM.calcVoteFactor(): totalVotesCastThisEx = " + totalVotesCastThisEx);
//System.out.println("***EPCM.calcVoteFactor(): totalScoreThisEx = " + totalScoreThisEx);
//System.out.println("***EPCM.calcVoteFactor(): meanVotesPerEx = " + meanVotesPerEx);
//System.out.println("***EPCM.calcVoteFactor(): desiredSamples = " + desiredSamples + " : " + whichIntervals);

        // Confidence based on number of votes received for this exhibit
        // compared to the mean number of votes for each voted-for-exhibit,
        // using normal variance-reduction (sqrt) policy.
        // This value is may exceed 1.0 to boost overall result confidence
        // where this exhibit has received an unusually high number of votes.
        final boolean enoughVotesCast = totalVotesCastThisEx >= meanVotesPerEx;
        final double confidenceThisExhibit = enoughVotesCast ? 1 :
            Math.sqrt(totalVotesCastThisEx / meanVotesPerEx);
        // Compute confidence weighting based on number of usable samples
        // and the total number of votes we got compared to the mean vote count per exhibit in sampled slots.
        final double confidence = Math.min(1.0, Math.sqrt(resultsFound / (double) desiredSamples) * confidenceThisExhibit);
//if(IsDebug.isDebug) { System.out.println("***EPCM.calcVoteFactor(): score/resultsFound/confidence/exhibit = " +g+", " +resultsFound+", " +confidence+", " +exhibitName); }
        // For low-confidence (low-vote) exhibits, tone down the goodness too.
        return(new Factor(enoughVotesCast ? g : (g * confidenceThisExhibit), confidence, "Votes: " + shortName));
        }


    /**If true, use ALL available votes rather than a sampling in _computeSampleBitSet().
     * If this is false then a caller of _computeSampleBitSet() can get a similar effect
     * with a large minNearMs value.
     */
    private static final boolean USE_ALL_VOTES = false;

    /**Compute BitSet of samples of event history to take; never null nor empty.
     * Computes a BitSet of length up to SystemVariables.EVENT_SAMPLES_RETAINED slots,
     * intended to be used with the EventPeriod.VLONG interval size,
     * with index 0 indicating the <em>previous</em> slot (ie no use of the current period).
     * <p>
     * This includes:
     * <ul>
     * <li>near-end samples designed to smooth out short (daily/weekly) cycles,
     * <li>annual/seasonal look-ahead to predict what may be of interest,
     * <li>a random sampling of points throughout the event history for robustness.
     * </ul>
     *
     * @param minNearMs  if positive, the minimum number of recent ms that
     *     the sampling should cover (else an internal minimum will be imposed)
     *
     * @return non-null, non-empty BitSet
     */
    private static BitSet _computeSampleBitSet(final long minNearMs)
        {
        if(USE_ALL_VOTES)
            {
            // Sample every available vote history point for maximum accuracy
            // at a possible performance cost.
            final BitSet whichIntervals = new BitSet(SystemVariables.EVENT_SAMPLES_RETAINED);
            whichIntervals.set(0, SystemVariables.EVENT_SAMPLES_RETAINED);
            return(whichIntervals);
            }

        // Number of our longest access periods needed to cover
        // "the minimum cycle" to smooth out short-term access patterns.
        // Typically at least 7 days to smooth out daily and weekly cycles.
        // We assume that our short cycle *does* fit in our data window.
        assert(_MIN_VLONG_SAMPLE_PERIODS < SystemVariables.EVENT_SAMPLES_RETAINED);

        // Number of intervals in our long cycle,
        // typically looking for seasonal patterns.
        // Typically 365 days to pick up yearly festivals, seasonal patterns, etc.
        // We assume that our long cycle *does* fit in our data window.
        assert(_SEASONAL_VLONG_CYCLE_PERIODS < SystemVariables.EVENT_SAMPLES_RETAINED);

        // Now calculate a set of sample periods to use; index 0 is the previous interval.
        // We factor in:
        //   * a window at the short end to catch recent activity.
        //   * a window around our long cycle to catch seasonal patterns.
        //   * a large random sampling of the full window to be robust.
        // We take a little more than the minimum window at the short end to help predictive ability.
        // We DO NOT include the current interval,
        // since it is of variable quality and may be expensive to fetch.
        final BitSet whichIntervals = new BitSet(SystemVariables.EVENT_SAMPLES_RETAINED);
        // Near-end samples, NOT including the current period.
        // The caller can force this to be extended to cover the period they are interested in.
        whichIntervals.set(0, Math.min(SystemVariables.EVENT_SAMPLES_RETAINED-1,
                Math.max(4*_MIN_VLONG_SAMPLE_PERIODS,
                         1+(int)(minNearMs/SystemVariables.EVENT_INTERVAL_VLONG_TERM_MS)) +
                1+(MAX_AGE_BEFORE_STALE_MS/SystemVariables.EVENT_INTERVAL_VLONG_TERM_MS)));
        // Predictive (we hope) seasonal samples; skew this forward for predictive value.
        final int seasonOffset = _SEASONAL_VLONG_CYCLE_PERIODS - _MIN_VLONG_SAMPLE_PERIODS;
        whichIntervals.set(seasonOffset, seasonOffset + 3*_MIN_VLONG_SAMPLE_PERIODS);
        // Random samples; take an extensive sampling of points otherwise not covered.
        for(int i = 1+Math.max(4*_MIN_VLONG_SAMPLE_PERIODS, SystemVariables.EVENT_SAMPLES_RETAINED/4); --i >= 0; )
            { _chooseRandomSlot(Rnd.fastRnd, whichIntervals, true); }
        // Chose a new one using a good generator with some real entropy in, thus hard to guess (but expensive).
        _chooseRandomSlot(Rnd.goodRnd, whichIntervals, true);

        return(whichIntervals);
        }

    /**Set a random bit in the BitSet to take a random sample.
     * Sets a bit at random.
     * <p>
     * It is possible (though may be expensive) to insist that a new (previously false)
     * bit is set, but only if (much) less than SystemVariables.EVENT_SAMPLES_RETAINED bits
     * are already true.
     *
     * @param rnd  random number source; never null
     * @param whichIntervals  the BitSet to set a random bit in; never null
     * @param forceNew  if true, ensure that we set a new bit ie one that was false
     *     unless at least half of all possible bits have already been set
     */
    private static void _chooseRandomSlot(final Random rnd,
                                          final BitSet whichIntervals,
                                          final boolean forceNew)
        {
        do
            {
            final int bitIndex = rnd.nextInt(SystemVariables.EVENT_SAMPLES_RETAINED);
            if(forceNew && whichIntervals.get(bitIndex) &&
                    (whichIntervals.cardinality() < (SystemVariables.EVENT_SAMPLES_RETAINED/2)))
                { continue; /* Selected bit already set; try again... */ }

            // OK, actually set the bit and stop!
            whichIntervals.set(bitIndex);
            break;

            } while(forceNew);
        }

    /**Our serial version... */
    private static final long serialVersionUID = 8518817446315221597L;

//    /**Deserialise. */
//    private void readObject(ObjectInputStream in)
//        throws IOException, ClassNotFoundException
//        {
//        in.defaultReadObject();
//        validateObject(); // Validate state immediately.
//        }

    /**Deserialise: use constructor for validation, defensive copying, etc.
     * Replace all trivially-stale "empty" values with a single shared value.
     * <p>
     * <strong>NOTE: this may not always preserve all values as expected in future.</strong>
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Eliminate redundant deserialised copies of TRIVIAL_NEUTRAL.
        if(isTriviallyStale() && (goodness == 0)) { return(TRIVIAL_NEUTRAL); }

        // Construct new instance of object in normal defensive way.
        return(new ExhibitPropsComputableMutable(staleAfter, goodness));
        }

    /**Serialise: replace all trivially-stale values being serialised with TRIVIAL_NEUTRAL.
     * We save all other (non-trivially-stale) values as-is, even if stale,
     * because they may have taken significant resources to compute
     * and may still be much more accurate than any fast approximation we could compute.
     * <p>
     * <strong>NOTE: this is effectively a LOSSY compression mechanism.</strong>
     */
    protected Object writeReplace()
        // throws ObjectStreamException
        {
        // Don't bother saving stale values to the stream
        // and replace them all with a single value.
        if(isTriviallyStale()) { return(TRIVIAL_NEUTRAL); }

        // Save this non-trivial value as-is.
        return(this);
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        if(staleAfter < 0)
            { throw new InvalidObjectException("bad object: invalid 'staleAfter' time"); }
        if(goodness < -Integer.MAX_VALUE)
            { throw new InvalidObjectException("bad object: invalid 'goodness' value"); }
        }
    }
