/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 02-Jul-2006
 * Time: 16:40:48
 */

/**Class to cache vote computations and correlated values.
 * In particular this aims to avoid repeated expensive recomputation
 * of underlying vote values so as global correlations can be
 * efficiently computed.
 */
public final class ExhibitPropsComputableMutableVoteCache implements ExhibitPropsComputableMutableVoteCacheIF,
                                                                     Serializable,
                                                                     ObjectInputValidation
    {
    /**Construct default empty cache. */
    public ExhibitPropsComputableMutableVoteCache() { }

    /**Reconstruct object, eg during deserialisation.
     * We take a defensive copy of mutable state, etc.
     * <p>
     * We also take the opportunity to strip out any invalid/stale state.
     * <p>
     * We hold a lock on the map parameter while copying its state.
     */
    private ExhibitPropsComputableMutableVoteCache(final Hashtable<String, Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor>> map)
        throws InvalidObjectException
        {
        assert(map != null);

        // Copy in legitimate state.
        synchronized(map) { voteCorrCacheMap.putAll(map); }
        // Clear stale entries from those supplied.
        _clearStaleState(voteCorrCacheMap);

        validateObject();
        }

    /**Key-separator character; not a valid exhibit-name character. */
    private static final char KEY_SEPARATOR_CHAR = ':';

    /**Correlation types.
     * By default most correlation types/data is "dense",
     * ie there should be many examples for and against each value.
     * <p>
     * Some types are "sparse",
     * ie where we usually expect only one or two examples for a given value,
     * and so we want to filter these more gently than usual.
     */
    public static enum CorrType
        {
        /**By variation on same exhibit (sparse). */    VARIANT(true),
        /**By similar exhibit (sparse). */              SIMILAR(true),
        /**By exhibit size (bytes). */                  SIZE,
        /**By exhibit attribute word(s). */             ATTRIBUTE,
        /**By exhibit resolution (eg pixels). */        RESOLUTION,
        /**By exhibit main word. */                     MAIN_WORD,
        /**By exhibit category. */                      CATEGORY,
        /**By exhibit type. */                          TYPE,
        /**By author. */                                AUTHOR,
        /**By MIME primary type component. */           MIME1,
        /**By prefixes. */                              PREFIX,
        /**By all stem words. */                        STEM_WORDS;

        /**By default correlations are not expected to be sparse. */
        private CorrType() { this.sparse = false; }
        /*Construct correlation with specified sparseness. */
        private CorrType(final boolean isSparse) { this.sparse = isSparse; }

        /**Is this correlation "sparse"?. */
        private final boolean sparse;

        /**Get the "sparseness" flag. */
        public final boolean isSparse() { return(sparse); }

        /**Make the appropriate cache lookup key(s) for a given value and full valid exhibit name; never null.
         * Add zero-or-more, non-null, non-empty cache key(s) of the form:
         *     <samp>ENUM : value</samp> eg <samp>SIZE:16</samp>
         * @param corrType  the correlation type; never null
         * @param esa  exhibit details; never null
         * @param aep  never null
         * @param result  non-null, mutable Set to which result is added
         *
         * @throws IOException  if key cannot be constructed
         */
        static void makeCacheKey(final CorrType corrType,
                                 final ExhibitStaticAttr esa,
                                 final AllExhibitProperties aep,
                                 final Set<String> result)
            throws IOException
            {
            assert(corrType != null);
            assert(esa != null);
            assert(aep != null);

            switch(corrType)
                {
                case AUTHOR:
                    { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + ExhibitName.getAuthorComponent(esa.filePath)); return; }

                case TYPE:
                    { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + ExhibitName.getExtensionComponent(esa.filePath)); return; }

                case CATEGORY:
                    { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + ExhibitName.getCategoryComponent(esa.filePath)); return; }

                case MAIN_WORD:
                    {
                    // Just the main word matches, case-sensitively.
                    final String f = ExhibitName.getFileComponent(esa.filePath);
                    final int dash = f.indexOf(ExhibitName.WORD_SEP);
                    final String mainWord = f.substring(0, dash);
                    result.add(corrType.toString() + KEY_SEPARATOR_CHAR + mainWord);
                    return;
                    }

                case MIME1:
                    {
                    final ExhibitMIME.ExhibitTypeParameters type = (ExhibitMIME.getInputFileType(esa.filePath));
                    if(type == null) { return; }
                    final int slash = type.mimeType.indexOf('/');
                    if(slash < 1) { return; }
                    final String primaryComponent = type.mimeType.substring(0, slash);
                    result.add(corrType.toString() + KEY_SEPARATOR_CHAR + primaryComponent);
                    return;
                    }

                case SIMILAR:
                    {
                    // "Similar" means the main-words component matches
                    // case-insensitively.
                    result.add(corrType.toString() + KEY_SEPARATOR_CHAR + ExhibitName.getMainWordsComponent(
                        esa.filePath,
                        ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet()).toLowerCase());
                    return;
                    }

                case VARIANT:
                    {
                    // "Variant" means all components match except:
                    //   * the category
                    //   * the attribute words
                    //   * the extension
                    // All matching is case-sensitive.
                    final String mainWords = ExhibitName.getMainWordsComponent(
                        esa.filePath,
                        ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
                    final int num = ExhibitName.getNumberInSeriesComponent(esa.filePath);
                    final String auth = ExhibitName.getAuthorComponent(esa.filePath);
                    final StringBuilder sb = new StringBuilder(29 + mainWords.length());
                    sb.append(corrType.toString()).append(KEY_SEPARATOR_CHAR);
                    sb.append(mainWords).append('-');
                    sb.append(num).append('-');
                    sb.append(auth);
                    result.add(sb.toString());
                    return;
                    }

                case SIZE: // Size in bytes grouped into buckets by exponent.
                    { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + GenUtils.log2Approx(esa.length)); return; }

                case RESOLUTION:
                    {
                    final StringBuilder sb = new StringBuilder();
                    sb.append(corrType.toString()).append(KEY_SEPARATOR_CHAR);

                    // Some notion of the raw information content of the exhibit.
                    // Zero-or-more items extracted from meta-data.
                    final ExhibitPropsComputable ePC =
                        aep.getExhibitPropsComputable(esa.filePath);
                    if(ePC != null)
                        {
                        final java.awt.Dimension xyDim = ePC.getXyDimensions();
                        if(xyDim != null)
                            {
                            // Add pixel resolution for images.
                            sb.append("px");
                            sb.append(GenUtils.log2Approx(xyDim.width * (long) xyDim.height));
                            result.add(sb.toString());
                            }

                        // TODO: should have sample rate for audio, etc.
//                        final Node md = ePC.getMetadata();
                        }

                    // We will not have created any key
                    // if no useful metadata was available.
                    return;
                    }

                case ATTRIBUTE:
                    {
                    // Get the exhibit's attribute words, if any.
                    final Set<String> aws = ExhibitName.getAttributeWordsComponentSortedSet(
                        esa.filePath, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
                    // Prefix them and add them to the result.
                    for(final String aw : aws)
                        { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + aw); }
                    return;
                    }

                case STEM_WORDS:
                    {
                    // By all (mono-cased) stem words.
                    final Enumeration mainWords = ExhibitName.getMainWords(
                        esa.filePath, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
                    while(mainWords.hasMoreElements())
                        { result.add(corrType.toString() + KEY_SEPARATOR_CHAR + ((String) mainWords.nextElement()).toLowerCase()); }
                    return;
                    }

                case PREFIX:
                    {
                    // By all stem prefixes of the "virtual" name.
                    final Enumeration mainWords = ExhibitName.getMainWords(
                        esa.filePath, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
                    final StringBuilder sb = new StringBuilder(esa.filePath.length() - 4);
                    sb.append(ExhibitName.getCategoryComponent(esa.filePath));
                    sb.append('/');
                    while(mainWords.hasMoreElements())
                        {
                        final String word = ((String) mainWords.nextElement());
                        sb.append(word).append(ExhibitName.WORD_SEP);
                        result.add(corrType.toString() + KEY_SEPARATOR_CHAR + sb);
                        }
                    return;
                    }

                default:
                    { throw new Error("Cannot generate key for correlation type " + corrType); }
                }
            }

        /**Make all cache keys for a given exhibit; never null.
         * This may return a variable number of keys,
         * and may ignore those that cannot currently be made and
         * are unlikely ever to be made,
         * but is as consistent as possible for any one exhibit.
         *
         * @return non-null, possibly-empty, set of cache keys
         */
        static Set<String> makeCacheKeys(final ExhibitStaticAttr esa,
                                         final AllExhibitProperties aep)
            {
            final HashSet<String> result = new HashSet<String>();
            for(final CorrType ct : CorrType.values())
                {
                try { makeCacheKey(ct, esa, aep, result); }
                catch(final IOException e) { /* Ignore ungeneratable keys. */ }
                }
            return(result);
            }
        }


    /**Map from vote/correlation key to compute Factor and period for which it was computed; never null after construction/deserialisation.
     * Since all our calculations depend on VLONG events,
     * cache entries need only be invalidated when the period number
     * changes.
     * <p>
     * One valid key format is a full, valid exhibit name.
     * <p>
     * Other keys are of the form of a CorrType name followed by a colon
     * followed by the rest of the key.
     * <p>
     * This state can be serialised; it is also recomputed on demand.
     * <p>
     * We use a Hashtable to be thread-safe.
     */
    private final Hashtable<String, Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor>> voteCorrCacheMap =
        new Hashtable<String, Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor>>();

    /**Compute the vote or retrieve from cache; never null.
     * A computed/cached value is valid until the VLONG period changes.
     * <p>
     * The returned factor's goodness can range from -1 to +1,
     * and confidence from 0 to +1;
     * any scaling required will have to be applied elsewhere.
     *
     * @param exhibitName  full, valid exhibit name; never null
     * @param aep  never null
     * @param vars  never null
     * @return  vote Factor for the given exhibit; never null.
     * @throws java.io.IOException
     */
    public final ExhibitPropsComputableMutable.Factor calcVoteFactor(
            final String exhibitName,
            final AllExhibitProperties aep,
            final BasicVarMgrInterface vars)
        throws IOException
        {
        if(!ExhibitName.validNameSyntaxBasic(exhibitName))
            { throw new IllegalArgumentException(); }
        if((aep == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        // Avoid clogging up our cache for non-existant exhibits...
        if(null == aep.aeid.getStaticAttr(exhibitName))
            { throw new IllegalArgumentException("no such exhibit"); }

        // Compute the current period.
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());

        // Get the current cached value, if any.
        final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> cached =
            voteCorrCacheMap.get(exhibitName);
        if(cached != null)
            {
            // If it is not null, but is stale,
            // then clear the entry (and go on to recompute it in a moment).
            // (There is a small race risk here,
            // but the worst that can happen is some wasted
            // recomputations, not a wrong result.)
            final long l = cached.first.longValue();
            // If the value is still current then return it!
            if(l == currentPeriod)
                { return(cached.second); }
//            // Clear stale cached value.
//            else
//                { voteCorrCacheMap.remove(exhibitName); }
            }

        // OK no cached value, or stale, so (re)compute it,
        // and store the new value in the cache.
        final ExhibitPropsComputableMutable.Factor result =
            ExhibitPropsComputableMutable.calcVoteFactor(exhibitName, aep, vars, 0);
        voteCorrCacheMap.put(MemoryTools.intern(exhibitName),
            new Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor>(
                MemoryTools.intern(new Long(currentPeriod)), result));

        // Return our result...
        return(result);
        }

    /**Get correlates for specified exhibit; never null but result may be empty.
     * This is based on votes of other "related" exhibits.
     * <p>
     * The goodness of each Factor is either -1 or +1,
     * with the correlation/confidence ranging between 0 and 1.
     * <p>
     * This will return values computed up to one period ago if need be,
     * to help avoid a sudden splurge of CPU effort as we tick from one period
     * to the next.
     * <p>
     * No particular ordering of the results is guaranteed,
     * but there will be no duplicates and no nulls.
     *
     * @param force  if true, force complete computation if need be,
     *     else we will just return what we have in cache;
     *     complete recomputation may be expensive but should last a long time
     *     and we will abort with an IOException if we cannot complete
     *     the recomputation in a reasonable time
     *
     * @throws IOException cannot extract required correlates
     */
    public ExhibitPropsComputableMutable.Factor[] getCorrelates(
            final ExhibitStaticAttr esa,
            final AllExhibitProperties aep,
            final BasicVarMgrInterface vars,
            final boolean force)
        throws IOException
        {
        if((esa == null) || (aep == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        // Avoid clogging up our cache for non-existant exhibits...
        if(null == aep.aeid.getStaticAttr(esa.filePath))
            { return(NO_CORRELATES); }

        // Compute the current period.
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());

        // Non-stale cached values available.
        final List<ExhibitPropsComputableMutable.Factor> result =
            new ArrayList<ExhibitPropsComputableMutable.Factor>();

        // Check all correlation types.
        // Accept any cached value however ancient, unless "forcing".
        _findCorrelates(esa, aep, currentPeriod, result, force);

        // If we got no valid correlates for this exhibit at all,
        // then try for a full recomputation and re-extract our factors.
        if(force && (result.size() == 0))
            {
            // Update/recompute values if needed.
            update(aep, vars, false);

            // Re-do our factor extraction...
            _findCorrelates(esa, aep, currentPeriod, result, force);
            }

        if(result.size() == 0)
            { return(NO_CORRELATES); /* Efficient way to return empty list. */ }

        final ExhibitPropsComputableMutable.Factor[] r =
            new ExhibitPropsComputableMutable.Factor[result.size()];
        result.toArray(r);
        return(r);
        }

    /**Find out if a category is rated "good"/popular or not.
     * Returns TRUE if rated good, FALSE if bad,
     * null if not significantly either or if not known.
     *
     * @param categoryDir  the initial directory component of an extant exhibit
     * @param force  if true may force (expensive) computation to give
     *     a more accurate answer,
     *     else may return a more approximate or stale answer,
     *     or none at all (null)
     */
    public Boolean isCategoryGood(final String categoryDir,
                                  final AllExhibitProperties aep,
                                  final BasicVarMgrInterface vars,
                                  final boolean force)
        throws IOException
        {
        if(!ExhibitName.validNameInitialComponentSyntax(categoryDir))
            { throw new IllegalArgumentException(); }

//System.out.println("***Looking for category goodness: " + categoryDir);

        // Make a syntactically-correct fake name/attrs from the category.
        final ExhibitStaticAttr esa = new ExhibitStaticAttr(
            categoryDir + "/a-A.a", 1, System.currentTimeMillis());

        // Compute the correlate name for this category.
        final SortedSet<String> keys = new TreeSet<String>();
        ExhibitPropsComputableMutableVoteCache.CorrType.makeCacheKey(
            ExhibitPropsComputableMutableVoteCache.CorrType.CATEGORY,
            esa, aep, keys);
        assert(keys.size() == 1); /* We expect to get exactly one key. */
        final String key = keys.first();

//System.out.println("***Looking for key: " + key);

        // Ensure data is current, if so requested.
        if(force) { update(aep, vars, true); }

        final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> f =
            voteCorrCacheMap.get(key);
        // If no Factor or too old then return null immediately.
        if(f == null)
            { return(null); }
        if(force &&
           (f.first.longValue() < EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis())-1))
            { return(null); }

//System.out.println("***Got: " + f.second);

        // If Factor not definite/strong enough then return null...
        if(Math.abs(f.second.goodness * f.second.correlation) < ExhibitPropsComputableMutable.GOODBAD_LIMIT)
            { return(null); }

        // Return a definite good/bad result.
        return((f.second.goodness >= 0) ? Boolean.TRUE : Boolean.FALSE);
        }

    /**Append to the result List any factors we find for this exhibit.
     * This ignores any keys that cannot be constructed.
     *
     * @param esa  the exhibit name/attr; never null
     * @param currentPeriod  current VLONG period; strictly positive
     * @param result  found correlates are appended to this value
     * @param force  if false then accept any extant value;
     *     if true then only accept values up to one period old
     */
    private void _findCorrelates(final ExhibitStaticAttr esa,
                                 final AllExhibitProperties aep,
                                 final long currentPeriod,
                                 final List<ExhibitPropsComputableMutable.Factor> result,
                                 final boolean force)
        {
        // Make all cache keys for this exhibit...
        final Set<String> keys = CorrType.makeCacheKeys(esa, aep);
        // and update all keyed entries all with this exhibit's rating.
        for(final String key : keys)
            {
            final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> f =
                voteCorrCacheMap.get(key);
            if(f == null) { continue; }
            if(!force)
                { result.add(f.second); }
            else if(f.first.longValue() >= currentPeriod-1)
                { result.add(f.second); }
//            else
//                { result.remove(key); /* Remove stale value. */ }
            }
        }

    /**Immutable empty correlates list. */
    private static final ExhibitPropsComputableMutable.Factor[] NO_CORRELATES =
        new ExhibitPropsComputableMutable.Factor[0];

    /**Significance confidence threshold; non-negative.
     * Computed correlations of confidence less than +/- this
     * are discarded to save space.
     * <p>
     * A value between 0.001 (0.1%) and 0.1 (10%) is probably reasonable.
     */
    private static final float SIGNIFICANCE_CONF_THRESHOLD = 0.003f;

    /**Significance count threshold for non-sparse correlations; non-negative.
     * Computed non-sparse correlations with counts less than this
     * are discarded to save space and time.
     * <p>
     * A value between 2 and 10 is probably reasonable;
     * an odd value avoids recording a "neutral" score at minimum count.
     */
    private static final int SIGNIFICANCE_COUNT_THRESHOLD = 3;

    /**The class in which we accumulate stats while recomputing correlations. */
    private static final class Accum
        {
        /**Number of samples that have contributed to this value; strictly positive. */
        final AtomicInteger count = new AtomicInteger();

        /**Accumulated (sum) confidence value with negative goodness being treated as a negative confidence.
         * All access protected by the instance lock.
         */
        private double totalConfidence;

        synchronized double getTotalConfidence()
            { return(totalConfidence); }

        synchronized void addToTotalConfidence(final double delta)
            { this.totalConfidence += delta; }
        }

    /**Time limit for update() in ms; strictly positive.
     * A value of the order of a second or so is probably right.
     * This allows some significant work to get done and cached,
     * but should not block anything (such as page generation)
     * for an unbearable amount of time even to a visitor.
     * <p>
     * This should still let us get a reasonable amount of work done
     * so that we can rebuild the data incrementally if necessary.
     */
    private final int UPDATE_TIME_LIMIT_MS = 521 +
                                             Rnd.fastRnd.nextInt(521);

    /**Private lock to prevent more than one thread at once expending effort in the core part of update(). */
    private final ReentrantLock _uCoreLock = new ReentrantLock();

    /**Maximum number of threads to allow in update(); strictly positive.
     * It is probably worth allowing into update() a few threads
     * to try to parallelise probably-I/O-bound sections at the start
     * and to do CPU-bound computations concurrently on multi-CPU systems,
     * but too many concurrent threads will possibly just waste resources
     * and overload I/O subsystems (eg back-end HTTP connections).
     */
    private static final int _computeMaxConcurrentUpdateThreads = Math.max(3, Math.min(7,
                1 + Runtime.getRuntime().availableProcessors()));

    /**Counting semaphore to limit update() first-phase concurrency; never null. */
    private final Semaphore _uStartSem = new Semaphore(_computeMaxConcurrentUpdateThreads);

    /**If true then try to load/compute early the scores of voted-for-exhibits.
     * An eager set-up attempts to start its computations
     * while part of its data (one of the "all" buckets) is still being fetched.
     * <p>
     * An eager setup should be able to compute the correlations more quickly,
     * but at increased risk of using part/all somewhat/very stale values,
     * and thus generating a less good result.
     */
    private static final boolean EAGER_VOTE_LOAD = true;

    /**If true then try to use the "all" votes buckets rather than summing the other buckets.
     * Using the "all" bucket may less accurate/comprehensive,
     * and may delay initialisation if it has to be fetched from upstream.
     */
    private static final boolean USE_ALL_BUCKET = false;

    /**Bring correlations data up to date.
     * This may also perform housekeeping such as clearing stale state.
     * <p>
     * This may be a relatively expensive call,
     * though if values are up-to-date then this should be quite cheap.
     * <p>
     * This is thread-safe and will use multiple caller threads to help fetch
     * vote data and concurrently recompute exhibit scores,
     * though second and subsequent concurrent calling threads
     * will return immediately upon reaching the core correlation computations
     * in order to prevent wasted duplicate effort.
     * <p>
     * If this routine cannot get values it needs then it throws an IOException.
     * <p>
     * No work will be done for an empty aep.
     * <p>
     * This will abort with an IOException if taking too long
     * (of the order of many seconds) or cannot otherwise ensure that
     * the correlations data is up-to-date.
     * Thus if this returns without throwing an IOException then
     * the correlations data can be assumed to be up-to-date.
     *
     * @param aep  current exhibit properties; never null
     * @param vars  handle on system variables; never null
     * @param noTimeLimit  if true, this runs until complete if possible
     *
     * @throws IOException  if the correlations data cannot be guaranteed
     *     to be up-to-date
     *     and/or a problem was found fetching required data
     *     and/or the routine was taking to long to complete the calculations
     */
    public void update(final AllExhibitProperties aep,
                       final BasicVarMgrInterface vars,
                       final boolean noTimeLimit)
        throws IOException
        {
        if((aep == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        // Nothing to be done for an empty set of exhibits.
        if(aep.aeid.length == 0) { return; }

        // If we can find our marker key,
        // and it is for the current period,
        // then we can assume that computations are up-to-date.
        final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());
        final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> entry =
                            voteCorrCacheMap.get(MARKER_KEY);
        if((entry != null) && (entry.first.longValue() == currentPeriod))
            { return; /* Apparently completely up-to-date! */ }

        if(_uCoreLock.isLocked())
            { throw new IOException("update() already busy in core phase; vote cache data is stale"); }


        // The complete set of voted-for exhibits (pro and con).
        // Guess size based on a few votes per sample period (ie per day).
        // Note that these values are seen only by this thread.
        final Set<String> votedForNames = new HashSet<String>(8 * SystemVariables.EVENT_SAMPLES_RETAINED);
        final long currentPeriodL;

        // If lots of external threads are needing the update() to be done
        // then allow a few of them in to help with the work.
        if(!_uStartSem.tryAcquire())
            { throw new IOException("update() already busy in first phase; vote cache data is stale"); }
        try
            {
            // Are we probably the first thread into this section?
            // (It doesn't matter if we get this wrong occasionally.)
            final boolean firstThreadIn = _uStartSem.availablePermits() == (_computeMaxConcurrentUpdateThreads-1);
            
            // We are now really getting started...
            final long startTime = System.currentTimeMillis();
            // Note our limit time;
            // if really out of date then we allow MUCH longer before timing out.
            // But in any case we don't quit until we have made some progress
            // to guarantee progress, ie that we do eventually complete.
            final boolean reasonablyUpToDate =
                ((entry != null) && (entry.first.longValue() == currentPeriod-1));
            final long stopBy = startTime +
                                Rnd.fastRnd.nextInt(UPDATE_TIME_LIMIT_MS) +
                (reasonablyUpToDate ?  UPDATE_TIME_LIMIT_MS
                                    : (3 * UPDATE_TIME_LIMIT_MS));

if(IsDebug.isDebug) { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): cache size: " + voteCorrCacheMap.size() + ".]"); }


            // Request/fetch the current-period event sets now,
            // so that they may be ready later to be added into the mix
            // if not immediately available.
            final EventVariableValue pcur = vars.getEventValue(SystemVariables.VOTE_PRO, EventPeriod.VLONG, true);
            final EventVariableValue ccur = vars.getEventValue(SystemVariables.VOTE_CON, EventPeriod.VLONG, true);

            final List<EventVariableValue> availableALLBuckets = new ArrayList<EventVariableValue>(2);
            if(USE_ALL_BUCKET)
                {
                // Get pro and con votes' "all" period event set.
                // This is a heuristic to abort if we cannot get vote "all" buckets,
                // and to try to encourage fetching of such data by the cache,
                // or to continue quietly if what we currently have is OK,
                // even empty if there seems to have been no recent voting.
                //
                // If we get null/stale/empty "all" buckets
                // then abort and hope that they will be present/fresh
                // the next time that we are called
                // (eg allowing for asynchronous fetch/cacheing of these values).
                final EventVariableValue pevvs[] = vars.getEventValues(
                    SystemVariables.VOTE_PRO, EventPeriod.VLONG, 0, null);
                final EventVariableValue cevvs[] = vars.getEventValues(
                    SystemVariables.VOTE_CON, EventPeriod.VLONG, 0, null);
                if((pevvs.length != 0) && (pevvs[0] != null)) { availableALLBuckets.add(pevvs[0]); }
                if((cevvs.length != 0) && (cevvs[0] != null)) { availableALLBuckets.add(cevvs[0]); }
                // If not eager and we don't have both "all" buckets right now
                // then give up quickly, without burning much CPU time.
                if(availableALLBuckets.size() < (EAGER_VOTE_LOAD ? 1 : 2))
                    { throw new IOException("missing vote one or both all buckets (early)"); }
                }

            // List of buckets from which we will extract (short) exhibit names.
            // Starts with the pro/con 'all' buckets.
            final List<EventVariableValue> buckets = new ArrayList<EventVariableValue>(availableALLBuckets);

            // If not using the "all" vote buckets or
            // if we are using "all" buckets but either is completely empty
            // then we may wish to resort
            // to explicitly fetching all/most available recent event periods
            // and computing the union of their values.
            if(!USE_ALL_BUCKET ||
               ((availableALLBuckets.size() == 2) &&
                ((availableALLBuckets.get(0).getTotalEventCount() == 0) ||
                 (availableALLBuckets.get(1).getTotalEventCount() == 0))))
                {
if(IsDebug.isDebug && USE_ALL_BUCKET) { System.err.println("[ExhibitPropsComputableMutableVoteCache.update(): WARNING: at least one 'all' bucket is empty: collecting vote data the slow way...]"); }

                // Remove extant dubious 'all' values.
                if(USE_ALL_BUCKET) { buckets.clear(); }

                // Forcing alternative (slow) mechanism...
                // Get events for all retained slots.
                final BitSet whichValues = new BitSet(SystemVariables.EVENT_SAMPLES_RETAINED);
                whichValues.set(0, SystemVariables.EVENT_SAMPLES_RETAINED);
                final SimpleVariableDefinition[] defs = new SimpleVariableDefinition[]{ SystemVariables.VOTE_PRO, SystemVariables.VOTE_CON };
                // Have extra threads try for values in a different order for concurrency...
                if(!firstThreadIn) { Collections.shuffle(Arrays.asList(defs), Rnd.goodRnd); }
                for(final SimpleVariableDefinition def : defs)
                    {
                    final EventVariableValue evvsAll[] = vars.getEventValues(
                        def, EventPeriod.VLONG, currentPeriod, whichValues);
                    for(final EventVariableValue evv : evvsAll)
                        {
                        if((evv != null) && (evv.getTotalEventCount() != 0))
                            { buckets.add(evv); }
                        }
                    }
                }

            // Throw in current-slot votes at the last moment.
            // Try and fetch them again if they were empty when we last tried.
            // These data are nice-to-have if we can get them,
            // since they make the correlation values as fresh as possible.
            buckets.add((pcur.getTotalEventCount() != 0) ? pcur :
                        vars.getEventValue(SystemVariables.VOTE_PRO, EventPeriod.VLONG, true));
            buckets.add((ccur.getTotalEventCount() != 0) ? ccur :
                        vars.getEventValue(SystemVariables.VOTE_CON, EventPeriod.VLONG, true));


            // If eager then get/compute votes/scores ASAP,
            // even when we don't have both "all" buckets.
            //
            // We convert from short name to long/full form
            // dropping any exhibit (name) no longer extant.
            final long calcVoteFactorStart = System.currentTimeMillis();
            currentPeriodL = MemoryTools.intern(new Long(currentPeriod));
            // Scramble order of buckets to help ensure that
            // parallel threads don't waste too much time redoing the same work.
            if(!firstThreadIn) { Collections.shuffle(buckets); }
            for(final EventVariableValue evv : buckets)
                {
                // Bucket entries should never be null.
                assert(evv != null);

                final List so = new ArrayList<Object>(evv.getInfo().keySet());
                // Again, shuffle work order to help avoid threads colliding.
                if(!firstThreadIn) { Collections.shuffle(so, Rnd.fastRnd); }
                for(final Object key : so)
                    {
                    if(!(key instanceof String)) { continue; }
                    final String shortName = (String) key;
                    final String exhibitName = aep.aeid.getFullName(shortName);
                    if(exhibitName == null) { continue; }
                    // Skip if we've already handled/added this exhibitName.
                    if(!votedForNames.add(exhibitName)) { continue; }

                    // If eagerly fetching/computing vote values then do it now,
                    // possibly in parallel with fetching one of the "all" buckets.
                    if(EAGER_VOTE_LOAD)
                        {
                        // ----------------------------------------------------
                        // Bring at least one vote up to date
                        // if not all yet up-to-date.

                        // Get any extant cached vote value for this exhibit.
                        final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> oldVVB =
                            voteCorrCacheMap.get(exhibitName);
                        if((oldVVB != null) && (oldVVB.first.longValue() == currentPeriod))
                            { continue; /* Already up-to-date. */ }

                        // Compute a new/current vote value for this exhibit.
                        final ExhibitPropsComputableMutable.Factor result =
                            ExhibitPropsComputableMutable.calcVoteFactor(exhibitName, aep, vars, 0);
                        voteCorrCacheMap.put(MemoryTools.intern(exhibitName),
                            new Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor>(
                                currentPeriodL, result));

                        // Abort if we've run out of time.
                        if(!noTimeLimit && (System.currentTimeMillis() > stopBy))
                            {
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): ABORTED calcVoteFactor time "+(System.currentTimeMillis()-calcVoteFactorStart)+"ms]"); }
                            throw new IOException("ran out of time in update(); data not brought up-to-date yet because not all voted-for-exhibits up-to-date yet");
                            }

                        // Abort if another thread has already made it as far as
                        // the core section.
                        if(_uCoreLock.isLocked())
                            {
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): ABORTED (another thread already in update() core) calcVoteFactor time "+(System.currentTimeMillis()-calcVoteFactorStart)+"ms]"); }
                            throw new IOException("another thread already in update() core");
                            }
                        }
                    }
                }
            final long calcVoteFactorEnd = System.currentTimeMillis();
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): calcVoteFactor time "+(calcVoteFactorEnd-calcVoteFactorStart)+"ms]"); }

            // If we have not yet managed to get both buckets
            // and thus all the exhibit names that they contain
            // then we cannot go any further for the moment.
            if(EAGER_VOTE_LOAD && USE_ALL_BUCKET && (availableALLBuckets.size() < 2))
                { throw new IOException("too few all vote buckets (EAGER)"); }
            if(USE_ALL_BUCKET) { assert(availableALLBuckets.size() == 2) : "must have at least both pro and con `all' buckets or full set of pro/con buckets by now"; }

if(IsDebug.isDebug) { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): votedForNames size: " + votedForNames.size() + ".]"); }
            }
        finally
            { _uStartSem.release(); }

        // Only allow at most one (external) thread at once
        // to consume CPU in the core part of update()
        // so as to avoid wasting resources
        // though we will inject local parallelism directly for speed.
        if(!_uCoreLock.tryLock()) { throw new IOException("update() already busy in core; vote cache data is stale"); }
        try
            {
            // Double-check that the work was not completed by another thread
            // while we were in an earlier section!
            final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> e2 =
                                voteCorrCacheMap.get(MARKER_KEY);
            if((e2 != null) && (e2.first.longValue() == currentPeriod))
                { return; /* Apparently completely up-to-date! */ }

            // For every voted-for exhibit,
            // and for every correlation type that we can compute,
            // accumulate correlation stats.
            // We then discard any that are insignificant,
            // and insert the remaining significant values into the cache.
            //
            // Also compute the mean count per non-sparse value
            // as a baseline against which non-sparse values with lower counts
            // can be rated for confidence/certainty.
            //
            // We optimistically assume that in the typical case
            // we can get through this entire operation quickly
            // (especially since we parallelise it)
            // so we don't bother splitting it into phases.
            //
            // We expect this work to be mainly compute-bound.
            final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

            final AtomicInteger totalNonSparseEntries = new AtomicInteger();
            final AtomicInteger totalNonSparseCount = new AtomicInteger();
            final ConcurrentHashMap<String,Accum> working = new ConcurrentHashMap<String, Accum>(17 + 19*votedForNames.size());
            final long beforeCompute = System.currentTimeMillis();
  
            // List of results that we can wait for completion of.
            final List<Future<?>> tasks = new ArrayList<Future<?>>(votedForNames.size());

            // Post tasks for each of the exhibits that has been voted for.
            for(final String exhibitName : votedForNames)
                {
                tasks.add(ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
                    public final void run()
                        {
                        final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);
                        // If exhibit appears not to exist any more then we're done...
                        // (This is in practice very rare.)
                        if(esa == null) { return; }

                        // Get/compute the current/new vote value for this exhibit.
                        final ExhibitPropsComputableMutable.Factor vote;
                        try { vote = calcVoteFactor(exhibitName, aep, vars); }
                        catch(final IOException e)
                            {
                            // If we get an IOException here
                            // then we probably have to abort the entire calculation
                            // and restart later.
                            e.printStackTrace();
                            throw new IllegalStateException("unexpected IOException: will have to restart", e);
                            }
                        assert(vote != null);

                        // Make all cache keys for this exhibit.
                        final Set<String> keys = CorrType.makeCacheKeys(esa, aep);
                        // Count the words in this exhibit for scaling...
                        final int nWords = ExhibitName.getMainWordsComponent(esa.filePath, attrWords).split(ExhibitName.WORD_SEPS).length;
                        // Update all keyed entries with this exhibit's rating.
                        for(final String key : keys)
                            {
                            // Is this a "sparse" correlation type?
                            final CorrType corrType = CorrType.valueOf(key.substring(0, key.indexOf(KEY_SEPARATOR_CHAR)));
                            final boolean isSparse = corrType.isSparse();

                            // Is this an especially "dense" term that needs scaling
                            // so as not to swamp other results.
                            final boolean needsScalingByWordCount =
                                (corrType == CorrType.STEM_WORDS) ||
                                (corrType == CorrType.PREFIX);

                            // Retrieve existing accumulator, or create a new one.
                            Accum ac;
                            while(null == (ac = working.get(key)))
                                {
                                // Create new all-zeros accumulator
                                // if one not already inserted.
                                // (This is race-free c/o putIfAbsent().)
                                working.putIfAbsent(key, new Accum());
                                }
                            // Update the accumulator.
                            final int count = ac.count.incrementAndGet();
                            final float corr = (!needsScalingByWordCount) ?
                                vote.correlation : (vote.correlation / nWords);
                            if(vote.goodness < 0) { ac.addToTotalConfidence(-corr); }
                            else if(vote.goodness > 0) { ac.addToTotalConfidence(+corr); }

                            // Note non-sparse entries that have a significant count.
                            if(!isSparse && (count >= SIGNIFICANCE_COUNT_THRESHOLD))
                                {
                                if(count == SIGNIFICANCE_COUNT_THRESHOLD)
                                    { totalNonSparseEntries.incrementAndGet(); /* Note new (non-sparse) entry. */ }
                                totalNonSparseCount.incrementAndGet();
                                }
                            }
                        }
                    }));
                }

            // Wail for all the tasks to complete...
            assert(tasks.size() == votedForNames.size());
            for(final Future<?> task : tasks)
                {
                try { task.get(); }
                catch(final InterruptedException e)
                    {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                    final IOException et = new InterruptedIOException("unexpected interrupt during calculation");
                    et.initCause(e);
                    throw et;
                   }
                catch(final ExecutionException e)
                    {
                    e.printStackTrace();
                    final IOException et = new IOException("unexpected exception during calculation");
                    et.initCause(e);
                    throw et;
                    }
                }

            final long afterCompute = System.currentTimeMillis();
/*if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): *** correlation factors compute time "+(afterCompute-beforeCompute)+"ms.]"); }

            final int tNSE = totalNonSparseEntries.get();
            final double meanNonSparseCount = (tNSE < 1) ? 0.0
                : (totalNonSparseCount.get() / (double) tNSE);
            final double mNSC_min = Math.max(meanNonSparseCount,
                                             SIGNIFICANCE_COUNT_THRESHOLD);

/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): votedForNames.size()="+votedForNames.size()+", working.size()="+working.size()+", meanNonSparseCount="+meanNonSparseCount + ".]"); }

            // Use the time when we started this whole update
            // to take a conservative view of the age of our source data,
            // eg in case the period ticked over while we were working.
            final Long cp = currentPeriodL;

            // Go through the working set,
            // converting significant results into main cache entries.
            final long finalComputePhaseStart = System.currentTimeMillis();
            for(final String key : working.keySet())
                {
                // Is this a "sparse" correlation type?
                final boolean isSparse =
                    CorrType.valueOf(key.substring(0, key.indexOf(KEY_SEPARATOR_CHAR))).isSparse();

                final Accum ac = working.get(key);
                assert(ac != null);

                // Ignore values with very small counts
                // (except for explicit "sparse" types)
                // as potentially unreliable (or at least very noisy).
                final int count = ac.count.get();
                if(!isSparse && (count < SIGNIFICANCE_COUNT_THRESHOLD))
                    { continue; }

                assert(ac.count.get() > 0);
                assert(!Double.isInfinite(ac.getTotalConfidence()));
                assert(!Double.isNaN(ac.getTotalConfidence()));
                // Compute mean/normalised confidence
                // scaling as appropriate for below-mean counts.
                // We scale by the expected extra variance of a low sample size.
                final double meanConf = Math.abs(ac.getTotalConfidence() / count);
                final double scaledConf =
                    (/* isSparse || */ (count >= mNSC_min)) ? meanConf :
                    (meanConf * Math.sqrt(count / mNSC_min));

                // If too small to be significant, then don't use it.
                // Test in such a way as to reject NaN values too,
                // ie unless a test succeeds, since all tests against NaN fail.
                if(!(scaledConf > SIGNIFICANCE_CONF_THRESHOLD)) { continue; }

                // Make our factor with a helpful comment.
                final String comment = key + " count=" + count;
                final ExhibitPropsComputableMutable.Factor f = new ExhibitPropsComputableMutable.Factor(
                    (ac.getTotalConfidence() < 0) ? -1 : +1,
                    scaledConf,
                    comment);

                // Insert the new Factor into the cache.
                voteCorrCacheMap.put(MemoryTools.intern(key), // Reduce old-heap churn.
                                     new Tuple.Pair<Long,ExhibitPropsComputableMutable.Factor>(cp, f));
                }
            final long finalComputePhaseEnd = System.currentTimeMillis();
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): final phase compute time "+(finalComputePhaseEnd-finalComputePhaseStart)+"ms.]"); }

            // Insert a distinguished marker key
            // so that it is quick to check when we last recomputed values here.
            // We do this only when all updates are complete.
            voteCorrCacheMap.put(MARKER_KEY,
                                 new Tuple.Pair<Long,ExhibitPropsComputableMutable.Factor>(cp, ExhibitPropsComputableMutable.FACTOR_ZERO));

            // Remove any stale state since we now have up-to-date values.
            _clearStaleState(voteCorrCacheMap);

/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitPropsComputableMutableVoteCache.update(): final cache size: " + voteCorrCacheMap.size() + ".]"); }
            }
        finally { _uCoreLock.unlock(); }

        // Dump the correlations table for debug purposes,
        // but outside the lock so as not to hold things up unduly...
        // Only show the non-sparse, non-exhibit entries.
        if(IsDebug.isDebug)
            {
            // Nicely dump the table in sorted order...
            final List<String> keys = new ArrayList<String>(voteCorrCacheMap.size());
            // Capture the key set atomically...
            synchronized(voteCorrCacheMap) { keys.addAll(voteCorrCacheMap.keySet()); }
            // ... and sort our copy to iterate over.
            Collections.sort(keys);
            // Make a re-usable working buffer.
            final StringBuilder sb = new StringBuilder(255);

            for(final String key : keys)
                {
                // Skip the raw exhibit votes.
                if(key.indexOf(KEY_SEPARATOR_CHAR) == -1) { continue; }
                // Skip the sparse entries (there may be *lots* of them)...
                if(CorrType.valueOf(key.substring(0, key.indexOf(KEY_SEPARATOR_CHAR))).isSparse()) { continue; }

                sb.setLength(0); // Clear the buffer.
                sb.append("    "); // Indent...
                sb.append(key).append(" = ");
                final Tuple.Pair<Long, ExhibitPropsComputableMutable.Factor> p = voteCorrCacheMap.get(key);
                if(p != null)
                    {
                    sb.append((p.second.goodness < 0) ? '-' : '+');
                    sb.append(p.second.correlation);
                    }
                System.out.println(sb);
                }
            }
        }

    /**Special marker key value to indicate when we last computed correlations.
     * Legal key, but never conflicts with any normal usage.
     */
    private static final String MARKER_KEY = CorrType.AUTHOR.toString() + KEY_SEPARATOR_CHAR;


    /**Serialise: strip out stale information before serialising. */
    protected Object writeReplace()
        // throws ObjectStreamException
        {
        // Clear stale entries before saving.
        _clearStaleState(voteCorrCacheMap);

        // Save the object as-is.
        return(this);
        }

    /**Remove stale state from the map passed in.
     * A lock is held on the map while this is done
     * to prevent it being altered unexpectedly.
     * <p>
     * We only remove things more than one period old
     * so that we can gently recompute values in the background
     * rather than being bounced into it as we roll into a new period.
     *
     * @param map  non-null map.
     */
    static private void _clearStaleState(final Hashtable<String,Tuple.Pair<Long,ExhibitPropsComputableMutable.Factor>> map)
        {
        synchronized(map)
            {
            // Compute the current period.
            final long currentPeriod = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());

            for(final Iterator<Tuple.Pair<Long,ExhibitPropsComputableMutable.Factor>> it = map.values().iterator(); it.hasNext(); )
                {
                final Tuple.Pair<Long,ExhibitPropsComputableMutable.Factor> tp = it.next();
                // Simply weed out stale entries,
                // ie anything not in the current VLONG period or the one before.
                if(tp.first.longValue() < currentPeriod-1)
                    { it.remove(); continue; }
                }
            }
        }


    /**Deserialise: use constructor for validation, defensive copying, etc.
     */
    protected Object readResolve()
        throws ObjectStreamException
        {
        // Construct new instance of object in normal defensive way.
        return(new ExhibitPropsComputableMutableVoteCache(voteCorrCacheMap));
        }


    /**Validates the object.
     *
     * @throws java.io.InvalidObjectException If the object cannot validate itself
     */
    public void validateObject() throws InvalidObjectException
        {
        if(voteCorrCacheMap == null)
            { throw new InvalidObjectException("bad object: null voteCorrCacheMap"); }

        synchronized(voteCorrCacheMap)
            {
            for(final Iterator it = voteCorrCacheMap.keySet().iterator(); it.hasNext(); )
                {
                // Weed out any invalid entries...
                final Object ok = it.next();
                if(!(ok instanceof String))
                    { throw new InvalidObjectException("bad object: invalid voteCorrCacheMap key type"); }

                // A key can be a valid exhibit name...
                // or a CorrType + ':' + rest of key.
                final String sk = (String) ok;
                if(!ExhibitName.validNameSyntax(sk))
                    {
                    final int cPos = sk.indexOf(KEY_SEPARATOR_CHAR);
                    if(cPos < 1)
                        { throw new InvalidObjectException("bad object: invalid voteCorrCacheMap key syntax"); }

                    // This first part must be a CorrType name.
                    final String firstPart = sk.substring(0, cPos);
                    if(null == CorrType.valueOf(firstPart))
                        { throw new InvalidObjectException("bad object: invalid voteCorrCacheMap composite key syntax"); }
                   }

                final Object ov = voteCorrCacheMap.get(ok);
                if(!(ok instanceof Tuple.Pair))
                    { throw new InvalidObjectException("bad object: invalid voteCorrCacheMap value type"); }
                final Tuple.Pair tp = (Tuple.Pair) ov;
                if(!(tp.first instanceof Long) ||
                   !(tp.second instanceof ExhibitPropsComputableMutable.Factor))
                    { throw new InvalidObjectException("bad object: invalid voteCorrCacheMap value sub-type"); }
                }
            }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 7048481653530545311L;
    }
