/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.vars.BasicVarMgrInterface;

import java.io.IOException;


/**Interface for retrieving/calculating votes and correlations.
 */
public interface ExhibitPropsComputableMutableVoteCacheIF
    {
    /**Compute exactly one Factor that depends on explicit user votes for specified exhibit; never null.
     * Because we expect voting to be sparse,
     * we total up votes across all sample intervals to reduce noise,
     * rather than taking each sample separately,
     * and do various other calculations a little differently to dense measures.
     * <p>
     * If this cannot fetch some of the data it requires due to an IOException
     * (not the same as fetching data successfully that has no entries)
     * then the exception is propagated to the caller.
     * (Normally this results in the EPCM item for which this is called
     * being marked stale so that we can try again when the data is available.)
     * <p>
     * This may cache values once computed for efficiency.
     * <p>
     * The returned factor's goodness can range from -1 to +1,
     * and confidence from 0 to +1;
     * any scaling required will have to be applied elsewhere.
     *
     * @throws java.io.IOException  if this was unable to fetch some data required
     */
    public ExhibitPropsComputableMutable.Factor calcVoteFactor(
                                        final String exhibitName,
                                        final AllExhibitProperties aep,
                                        final BasicVarMgrInterface vars)
        throws IOException;

    /**Get correlates for specified exhibit; never null but result may be empty.
     * This is based on votes of other "related" exhibits and exhibit metadata.
     * <p>
     * The goodness of each Factor is either -1 or +1,
     * with the correlation/confidence ranging between 0 and 1.
     * <p>
     * This will return values computed up to one period ago if need be,
     * to help avoid a sudden splurge of CPU effort as we tick from one period
     * to the next.
     * <p>
     * No particular ordering of the results is guaranteed,
     * but there will be no duplicates and no nulls.
     *
     * @param force  if true, force complete computation if need be,
     *     else we will just return what we have in cache;
     *     complete recomputation may be expensive but should last a long time
     *     and we will abort with an IOException if we cannot complete
     *     the recomputation in a reasonable time
     *
     * @throws IOException cannot extract required correlates
     */
    public ExhibitPropsComputableMutable.Factor[] getCorrelates(
            final ExhibitStaticAttr esa,
            final AllExhibitProperties aep,
            final BasicVarMgrInterface vars,
            final boolean force)
        throws IOException;

    /**Find out if a category is rated "good"/popular or not.
     * Returns TRUE if rated good, FALSE if bad,
     * null if not significantly either or if not known.
     *
     * @param categoryDir  the initial directory component of an extant exhibit
     * @param force  if true may force (expensive) computation to give
     *     a more accurate answer,
     *     else may return a more approximate or stale answer,
     *     or none at all (null)
     */
    public Boolean isCategoryGood(final String categoryDir,
                                  final AllExhibitProperties aep,
                                  final BasicVarMgrInterface vars,
                                  final boolean force)
        throws IOException;


    /**Bring correlations data up to date.
     * This may also perform housekeeping such as clearing stale state.
     * <p>
     * This may be a relatively expensive call.
     * <p>
     * This is thread-safe, though calling from more than one thread at once
     * may result in redundant work being performed.
     *
     * @param aep  current exhibit properties; never null
     * @param vars  handle on system variables; never null
     * @param noTimeLimit  if true, this runs until complete if possible
     */
    public void update(final AllExhibitProperties aep,
                       final BasicVarMgrInterface vars,
                       final boolean noTimeLimit)
        throws IOException;




    /**Trivial cache that doesn't actually cache anything, nor compute anything except the basic vote factor. */
    public static final ExhibitPropsComputableMutableVoteCacheIF TRIVIAL =
        new ExhibitPropsComputableMutableVoteCacheIF()
            {
            /**Fall through to directly compute the vote. */
            public final ExhibitPropsComputableMutable.Factor calcVoteFactor(
                    final String exhibitName,
                    final AllExhibitProperties aep,
                    final BasicVarMgrInterface vars)
                throws IOException
                { return(ExhibitPropsComputableMutable.calcVoteFactor(exhibitName, aep, vars, 0)); }

            /**Return empty correlates list. */
            public ExhibitPropsComputableMutable.Factor[] getCorrelates(ExhibitStaticAttr esa, AllExhibitProperties aep, BasicVarMgrInterface vars, boolean force)
                { return(new ExhibitPropsComputableMutable.Factor[0]); }

            /**Do nothing. */
            public void update(AllExhibitProperties aep, BasicVarMgrInterface vars, boolean noTimeLimit)
                { }

            /**Find out if a category is rated "good"/popular or not; always returns null. */
            public Boolean isCategoryGood(String categoryDir, AllExhibitProperties aep, BasicVarMgrInterface vars, boolean force)
                { return(null); }
            };
    }
