/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 09-Sep-02
 * Time: 18:10:39

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectInputValidation;
import java.io.Serializable;
import java.util.Date;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;

import org.hd.d.pg2k.svrCore.AllExhibitPropertiesDelta.DiffException;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.PropertiesBundleDiff;
import org.hd.d.pg2k.svrCore.props.PropertiesDiff;

/**This contains global immutable exhibit data over all exhibits that must live with the exhibits.
 * This is data that does not apply to any one exhibit but is small
 * and critical to handling of all exhibits.
 * <p>
 * Currently this contains:
 * <ul>
 * <li>The AEP-attached location DB properties.
 * </ul>
 * <p>
 * This might include:
 * <ul>
 * <li>tree-structured (i18n-ed) info
 * <li>hot/cold lists
 * <li>see-also lists (links)
 * <li>prefix translations and comments and aliases
 * <li>vote history
 * </ul>
 * <p>
 * This has a hash computed over its contents to make it easy for users
 * of this object to tell if it has changed since previous incarnations.
 * <p>
 * This object is immutable and Serializable, and tries to validate its
 * contents upon deserialisation.
 * <p>
 * TODO: need to be able to compute/apply a diff.
 */
public final class ExhibitPropsGlobalImmutable implements Serializable
    {
    /**Create an empty instance. */
    public ExhibitPropsGlobalImmutable()
        { this((PropertiesDiff) null, 0, (PropertiesBundleDiff) null, 0); }

    /**Create an instance. */
    public ExhibitPropsGlobalImmutable(final Properties locationDB,
                                       final long locationDBTimestamp,
                                       final Map<String, Properties> treedesc,
                                       final long treedescTimestamp)
        {
        this(PropertiesDiff.createAsStandAloneDiff(locationDB), locationDBTimestamp,
             PropertiesBundleDiff.createAsStandAloneDiff(treedesc), treedescTimestamp);
        }

    /**Create an instance. */
    public ExhibitPropsGlobalImmutable(final PropertiesDiff locationDB,
                                       final long locationDBTimestamp,
                                       final PropertiesBundleDiff treedesc,
                                       final long treedescTimestamp)
        {
        this.locationDB = ((locationDB == null) || !locationDB.isEmpty()) ? locationDB : null;
        this.locationDBTimestamp = locationDBTimestamp;
        this.treedesc = ((treedesc == null) || !treedesc.isEmpty()) ? treedesc : null;
        this.treedescTimestamp = treedescTimestamp;

        // Compute a +ve composite hash.
        // Zero if this instance contains no data.
        longHash =
            (((treedesc == null) ? 0 : (treedesc.hashCode() << 16)) ^
             ((locationDB == null) ? 0 : 3*locationDB.hashCode())) >>> 1;

        try { validateObject(); }
        catch(final InvalidObjectException e) { throw new IllegalArgumentException(e); }
        }

    /**Load global data from filesystem.
     */
    public static ExhibitPropsGlobalImmutable loadFromDataDir(final File exhibitDataDir)
        throws IOException
        {
        if(exhibitDataDir == null) { throw new IllegalArgumentException(); }

        // Load locationDB.
        final File locDBF = new File(exhibitDataDir, CoreConsts.LOCDB_PROPS_NAME + ".properties");
        final Tuple.Pair<Properties, Long> locDB = PropertiesDiff.loadFromFile(locDBF);
        // Load treedesc tree-structured i18n AKA/description info.
        final Pair<Map<String, Properties>, Long> treedescBundle = PropertiesBundleDiff.loadBundle(new File(exhibitDataDir, "_i18n"), "treedesc");
        return(new ExhibitPropsGlobalImmutable(
                        locDB.first, locDB.second.longValue(),
                        treedescBundle.first, treedescBundle.second.longValue()));
        }

    /**The location DB timestamp; strictly positive, or zero if no/empty locationDB. */
    public final long locationDBTimestamp;

    /**The (immutable) location DB in properties format; null if absent.
     * We store this in PropertiesDiff format for efficiency on the wire,
     * but convert it to LocationMap before use.
     * <p>
     * May be absent if there was no location data available at AEP creation
     * or when deserialising an older AEP that has no location data,
     * or as a more efficient way to represent an empty value
     * (we always store an empty value as null for efficiency).
     */
    private final PropertiesDiff locationDB;

    /**The treedesc bundle timestamp; non-negative, or may be zero if locationDBis empty/absent. */
    public final long treedescTimestamp;

    /**The (immutable) treedesc bundle; null if absent.
     * We store this in PropertiesBundleDiff format for efficiency on the wire.
     * <p>
     * May be absent if there was no location data available at AEP creation
     * or when deserialising an older AEP that has no location data,
     * or as a more efficient way to represent an empty value
     * (we always store an empty value as null for efficiency).
     */
    private final PropertiesBundleDiff treedesc;
    
    /**Get the hash of the treedesc; non-negative, or may be zero if treedesc is empty/absent. */
    public final int getTreedescHash()
        { return((treedesc == null) ? 0 : treedesc.hashCode()); }

    /**The (immutable) LocationMap derived from the locationDB; never null.
     * Created on or before first use.
     * <p>
     * Derived from the locationDB info, empty if no locationDB.
     * <p>
     * Not part of the serialised state.
     * <p>
     * Marked volatile for lock-free thread-safe access.
     */
    private volatile transient LocationMap locMap;
    
    /**Get the LocationMap; never null. */
    public LocationMap getLocationMap()
        {
        // Once the value is computed, reads do not require a lock.
        LocationMap result = locMap;
        if(locMap != null) { return(locMap); }

        // We prevent one than one thread creating the LocationMap instance.
        synchronized(this)
            {
            // If another thread got in there first and did the work
            // then we don't need to do it again.
            result = locMap;
            if(locMap != null) { return(locMap); }

            try
                {
                // Compute and cache locMap for first access.
                result = (locationDB == null) ? (new LocationMap()) :
                    (new LocationMap(PropertiesDiff.createFromStandAloneDiff(locationDB), locationDBTimestamp));
                }
            catch(final DiffException e)
                {
                // Should not happen.
                // Data errors should have been caught at construction/deserialisation.
                throw new Error(e);
                }
            locMap = result;
            return(result);
            }
        }

    /**Gets the appropriate localised tree-description message; never null unless the key is null.
     * In case of a missing message, or unavailable treedesc data,
     * this returns the key itself.
     * <p>
     * The key is converted to lower-case (if necessary) before lookup.
     * <p>
     * If the name is null, the result is null.
     * 
     * @param locale  desired locale, or null
     */
    public String getLocalisedTreeDescMessage(final String msgName,
                                              final Locale locale)
        {
        if(msgName == null) { return(null); }

        // If no treedec available at all, return original message.
        if(treedesc == null) { return(msgName); }
        // Always fetch from primary properties set for now.
        final PropertiesDiff pd = treedesc.getNewValues().get("");
        // If appopriate treedec not available, return original message.
        if(pd == null) { return(msgName); }
        // Extract correct message.
        final String s = pd.getNewValues().get(msgName.toLowerCase());
        // If message not available, return original message.
        if(s == null) { return(msgName); }

        return(s);
        }

    /**The hash of all the data held; guaranteed non-negative.
     * Depends all the information held in this object.
     * <p>
     * Will be zero if this contains no data.
     */
    public final long longHash;

    /**Returns a hash code value for the object; derived from the longHash.
     * Guaranteed zero if there are no exhibits.
     *
     * @return a hash code value for this object.
     */
    public int hashCode()
        { return((int) longHash); }

    /**Indicates whether some other object is "equal to" this one; the underlying data is the same if true.
     *
     * @param obj the reference object with which to compare.
     * @return <code>true</code> if this object is the same as the obj
     *         argument; <code>false</code> otherwise.
     */
    public boolean equals(final Object obj)
        {
    	if(this == obj) { return(true); }
        // Must be of the same type to be equal.
        if(!(obj instanceof ExhibitPropsGlobalImmutable)) { return(false); }
        final ExhibitPropsGlobalImmutable other = (ExhibitPropsGlobalImmutable) obj;

        // Hash calaculation may change wih newer EPGI implementations,
        // so we can't use the longHash as a proxy for the content,
        // unless we force its recomputation with readResolve().
        if(longHash != other.longHash) { return(false); }

        // Check the underlying immutable data.

        // Location DB timestamps must match.
        if(locationDBTimestamp != other.locationDBTimestamp) { return(false); } 

        // We check the locationDB value directly.
        if(locationDB == null) { if(other.locationDB != null) { return(false); } }
        else if(!locationDB.equals(other.locationDB)) { return(false); }

        // Treedesc timestamps must match.
        if(treedescTimestamp != other.treedescTimestamp) { return(false); } 

        // We check the locationDB value directly.
        if(treedesc == null) { if(other.treedesc != null) { return(false); } }
        else if(!treedesc.equals(other.treedesc)) { return(false); }

        return(true); // Equal.
        }

    /**Human-readable summary of state. */
    @Override public String toString()
        {
        final StringBuilder result = new StringBuilder(128);
        result.append("EPGI");
        result.append(":locationDB[").append(locationDB). /* append('|').append(new Date(locationDBTimestamp)). */ append(']');
        result.append(":treedesc[").append(treedesc). /* append('|').append(new Date(treedescTimestamp)). */ append(']');
        return(result.toString());
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        if(longHash < 0)
            { throw new InvalidObjectException("bad object: longHash < 0"); }
        if(locationDBTimestamp < 0)
            { throw new InvalidObjectException("bad object: locationDBTimestamp < 0"); }
        if((locationDBTimestamp == 0) && (locationDB != null))
            { throw new InvalidObjectException("bad object: locationDBTimestamp does not match locationDB presence: "+(new Date(locationDBTimestamp))+" vs "+locationDB); }
        if(treedescTimestamp < 0)
            { throw new InvalidObjectException("bad object: treedescTimestamp < 0"); }
        if((treedescTimestamp == 0) && (treedesc != null))
            { throw new InvalidObjectException("bad object: treedescTimestamp does not match treedesc presence: "+(new Date(treedescTimestamp))+" vs "+treedesc); }
        }

    /**Deserialise: use constructor for validation, defensive copying, conversion from old formats, etc.
     * Also allows us to (re)normalise the data in this instance,
     * eg to recompute the longHash in case our hash algorithm changes.
     */
    protected Object readResolve()
        throws java.io.ObjectStreamException
        {
        return(new ExhibitPropsGlobalImmutable(locationDB,
                                               locationDBTimestamp,
                                               treedesc,
                                               treedescTimestamp));
        }

//    /**Deserialise. */
//    private synchronized void readObject(ObjectInputStream in)
//        throws IOException, ClassNotFoundException
//        {
//        in.defaultReadObject();
//        validateObject(); // Validate state immediately.
//        }

    /**Our serial version... */
    private static final long serialVersionUID = 0xf130cee4509ca258L;


    /**Immutable record of the difference between two EPGI instances.
     * This contains the "diffable" items.
     */
    public static final class EPGIDiff implements Serializable, ObjectInputValidation
        {
        /**Create an empty instance. */
        public EPGIDiff()
            { this(0, null, 0, null); }

        /**Create an instance. */
        public EPGIDiff(final long locationDBTimestampNew,
                        final PropertiesDiff locationDBDiff,
                        final long treedescTimestampNew,
                        final PropertiesBundleDiff treedescDiff)
            {
            this.locationDBTimestampNew = locationDBTimestampNew;
            this.locationDBDiff = ((locationDBDiff == null) || !locationDBDiff.isEmpty()) ? locationDBDiff : null;
            this.treedescTimestampNew = treedescTimestampNew;
            this.treedescDiff = ((treedescDiff == null) || !treedescDiff.isEmpty()) ? treedescDiff : null;
            try { validateObject(); }
            catch(final InvalidObjectException e) { throw new IllegalArgumentException(e); }
            }

        /**The new location DB timestamp; strictly positive, or zero for empty diff. */
        public final long locationDBTimestampNew;

        /**The diffed location DB in properties format; non-empty else null. */
        private final PropertiesDiff locationDBDiff;

        /**The new treedesc timestamp; strictly positive, or zero for empty diff. */
        public final long treedescTimestampNew;

        /**The diffed treedesc in properties format; non-empty else null. */
        private final PropertiesBundleDiff treedescDiff;

        /**Create a diff from the old EPGI value to the new one.
         * If the new locationDB is absent then we fake a timestamp of now.
         */
        public static EPGIDiff createDiff(final ExhibitPropsGlobalImmutable e1,
                                          final ExhibitPropsGlobalImmutable e2)
            throws DiffException
            {
            if((null == e1) || (null == e2))
                { throw new IllegalArgumentException(); }
            return(new EPGIDiff(
                e2.locationDBTimestamp,
                PropertiesDiff.createDiff(
                    PropertiesDiff.createFromStandAloneDiff(e1.locationDB != null ? e1.locationDB : PropertiesDiff.EMPTY_DIFF),
                    PropertiesDiff.createFromStandAloneDiff(e2.locationDB != null ? e2.locationDB : PropertiesDiff.EMPTY_DIFF),
                    true, // We always need a diff.
                    false), // Should already be intern()ed by now.
                e2.treedescTimestamp,
                PropertiesBundleDiff.createDiff(
                    PropertiesBundleDiff.createFromStandAloneDiff(e1.treedesc != null ? e1.treedesc : PropertiesBundleDiff.EMPTY_DIFF),
                    PropertiesBundleDiff.createFromStandAloneDiff(e2.treedesc != null ? e2.treedesc : PropertiesBundleDiff.EMPTY_DIFF),
                    true, // We always need a diff.
                    false))); // Should already be intern()ed by now.
            }

        /**Apply diff to derive new EPGI; never null.
         * @param oldEpgi  null is treated as if an empty EPGI
         * @param diff  diff to apply; never null
         * @return  new non-null EPGI value from old one (null implies empty) and diff
         */
        public static ExhibitPropsGlobalImmutable applyDiff(ExhibitPropsGlobalImmutable oldEpgi,
                                                            final EPGIDiff diff)
            {
            if(oldEpgi == null) { oldEpgi = new ExhibitPropsGlobalImmutable(); }
            if(diff == null) { throw new IllegalArgumentException(); }
            try
                {
                return(new ExhibitPropsGlobalImmutable(
                    PropertiesDiff.createAsStandAloneDiff(
                        PropertiesDiff.applyDiff((oldEpgi.locationDB == null) ? new Properties() : PropertiesDiff.createFromStandAloneDiff(oldEpgi.locationDB),
                                        (diff.locationDBDiff != null) ? diff.locationDBDiff : PropertiesDiff.EMPTY_DIFF)),
                    diff.locationDBTimestampNew,
                    PropertiesBundleDiff.createAsStandAloneDiff(
                        PropertiesBundleDiff.applyDiff((oldEpgi.treedesc == null) ? (new HashMap<String,Properties>()) : PropertiesBundleDiff.createFromStandAloneDiff(oldEpgi.treedesc),
                                        (diff.treedescDiff != null) ? diff.treedescDiff : PropertiesBundleDiff.EMPTY_DIFF)),
                    diff.treedescTimestampNew));
                }
            catch(final DiffException e)
                { throw new Error(e); /* Should never happen. */ }
            }

        /**Deserialise: use constructor for validation, defensive copying, conversion from old formats, etc.
         * Also allows us to (re)normalise the data in this instance.
         */
        protected Object readResolve()
            throws java.io.ObjectStreamException
            {
            return(new EPGIDiff(locationDBTimestampNew,
                                locationDBDiff,
                                treedescTimestampNew,
                                treedescDiff));
            }

        /**Checks that the object is internally consistent. */
        public void validateObject() throws InvalidObjectException
            {
            if((locationDBTimestampNew == 0) != (locationDBDiff == null))
                { throw new InvalidObjectException("bad object: zero locationDBTimestampNew <=> empty diff"); }
            if(locationDBTimestampNew < 0)
                { throw new InvalidObjectException("bad object: -ve locationDBTimestampNew"); }
            if((locationDBDiff != null) && locationDBDiff.isEmpty())
                { throw new InvalidObjectException("bad object: non-null empty locationDBDiff"); }

            if((treedescTimestampNew == 0) != (treedescDiff == null))
                { throw new InvalidObjectException("bad object: zero treedescTimestampNew <=> empty diff"); }
            if(treedescTimestampNew < 0)
                { throw new InvalidObjectException("bad object: -ve treedescTimestampNew"); }
            if((treedescDiff != null) && treedescDiff.isEmpty())
                { throw new InvalidObjectException("bad object: non-null empty treedescDiff"); }
            }

        /**Serial UID.*/
        private static final long serialVersionUID = -3566302089418451668L;
        }
    }
