/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.location.Location;

import java.io.*;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

/**Immutable (and serialisable) store of all loadable auxiliary properties of a single exhibit.
 * These are properties (such as descriptive comment) that must be fetched
 * from the exhibit database and that cannot be computed.
 * <p>
 * This is designed to be efficient on the wire and in memory, since these
 * details will be held for each and every exhibit.
 * <p>
 * Note that a factory method is used to generate instances of this class
 * and where there is no loadable data for an exhibit (which will be rare)
 * null is returned.
 * <p>
 * Equals is implemented so that duplicates (particularly of EMPTY) can
 * easily be discarded.  hashCode() is implemented to support the semantics
 * of equals() and as a hash over all the contents of the object.
 * <p>
 * By default this looks for data in a new-style properties file first
 * (fileBase + '/' + exhibitName + ".props"),
 * and then, iff READ_OLD_FILES == true, in the old-style individual files.
 */
public final class ExhibitPropsLoadable implements Serializable, ObjectInputValidation,
                                                   MemoryTools.Internable,
                                                   MemoryTools.Compactable
    {
    /**Shared EMPTY instance (no auxiliary data on exhibit). */
    public static final ExhibitPropsLoadable EMPTY = new ExhibitPropsLoadable(
        null,    // No description.
        null,    // No location.
        null     // No accession data.
        );

    /**Private constructor so creation is by factory method.
     * Parameters should already have been checked before we are called.
     * <p>
     * We may intern() components that are likely to be duplicated
     * and that may consume significant memory.
     *
     * @param _description no more than MAX_DESCRIPTION_CHARS or null or "" if none
     */
    private ExhibitPropsLoadable(final Object _description,
                                 final Location.Base _location,
                                 final AccessionData _accession)
        {
        // Store the accession data.
        // Replace empty instances with null
        // and intern() the rest to reduce (tenured) heap churn during updates.
        accession = ((_accession == null) || (_accession.equals(AccessionData.EMPTY))) ?
                    null : MemoryTools.intern(_accession);

        // If the description is null or empty, store a null.
        if((_description == null) ||
           ((_description instanceof String) && (((String) _description).length() == 0)))
            { description = null; }
        else
            {
            // If not deferring compaction then
            // intern() whatever we were given.
            if(DEFER_COMPACTION)
                { description = _description; }
            else if(_description instanceof String)
                {
                Object d = _description;
                try { d = Compact7BitString.convertToCompact7BitString((String) _description, sDict); }
                catch(final Exception e) { }
                description = d;
                }
            }
//        // If the description is very short, store it as-is.
//        else if(_description.length() < FileTools.TYPICAL_DEFLATE_MIN_TEXT_SIZE_COMPRESSABLE)
//            { description = MemoryTools.intern(_description); } // Store as-is.
//        // Else consider compressing it or storing it as-is if compression appears ineffective.
//        else
//            {
//            assert(_description.length() > 0) && (_description.length() <= CoreConsts.DESCRIPTION_MAX_CHARS);
//
//            final ROByteArray compressed = ROByteArray.compressFromString(_description);
//            // We keep the compressed form if it looks likely to be
//            // smaller in memory and on the wire,
//            // assuming 1 char == 2 bytes,
//            // and a little extra overhead for the ROByteArray wrapper.
//            if(compressed.length() + 16 < _description.length() * 2)
//                { description = MemoryTools.intern(compressed); }
//            else
//                { description = MemoryTools.intern(_description); }
//            }

        // Store any location information supplied.
        if(Location.NONE.equals(_location))
            { location = null; } // More efficient representation for "NONE".
        else
            { location = MemoryTools.intern(_location); }


        // Verify object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Make relative path to new-style properties file. */
    private static String relPathToNewProperties(String exhibitName)
        {
        return(exhibitName + EXHIBIT_PROPS_FILENAME_SUFFIX);
        }

    /**Make relative path to old-style en-GB HTML generic description file. */
    private static String relPathToOldDescription(String exhibitName)
        {
        return(exhibitName + CoreConsts.DESCRIPTION_FILE_SUFFIX);
        }

    /**Make relative path to new accession-date file; never null.
     * This is an XML-format file.
     * <p>
     * Public so that other routines can find and read this file directly.
     *
     * @param exhibitName  non-null relative path to exhibit
     */
    public static String relPathToNewAccession(String exhibitName)
        {
        final String exhibitDir = (new File(exhibitName)).getParent();
        final String exhibitFile = (new File(exhibitName)).getName();
        return((new File(exhibitDir, ".accession." + exhibitFile + ".xml")).getPath());
        }

    /**Make relative path to old accession-date file; never null.
     * Historically the format of this file is a single line of the format:
     * <pre>
     *     YYYY[/MM[/DD]]
     * </pre>
     * where YYYY is the year and MM is the month 01--12 and DD is the 01-31,
     * all relative to UTC/GMT, and represent the approximate date that the
     * exhibits where incorporated in the library.
     * <p>
     * Public so that other routines can find and read this file directly.
     *
     * @param exhibitName  non-null relative path to exhibit
     */
    public static String relPathToOldAccession(String exhibitName)
        {
        final String exhibitDir = (new File(exhibitName)).getParent();
        final String exhibitFile = (new File(exhibitName)).getName();
        return((new File(exhibitDir, ".accession." + exhibitFile + ".txt")).getPath());
        }

    /**Make relative path to old-style accession-date YYYY[/MM[/DD]] file. */
    private static String relPathToOldLocation(String exhibitName)
        {
        final String exhibitDir = (new File(exhibitName)).getParent();
        final String exhibitFile = (new File(exhibitName)).getName();
        return((new File(exhibitDir, ".location." + exhibitFile + ".txt")).getPath());
        }


    /**Make a new instance with data loaded from the filesystem, given the exhibit name and some filesystem base directories.
     * In case of severe unexpected problems with retrieving information
     * an IOException is thrown, but minor problems are resolved
     * quietly and a message is reported on System.err.
     *
     * @throws IOException  in case of severe, unrecoverable difficulty with this exhibit
     * @return null or non-null, non-EMPTY result
     */
    public static ExhibitPropsLoadable getLoadableProperties(final String exhibitName,
                                                             final File exhibitBaseDir)
        throws IOException
        {
//        final File fullExhibitPath = new File(exhibitBaseDir, exhibitName);
//        if(!fullExhibitPath.exists())
//            { throw new FileNotFoundException("exhibit missing"); }

        // Attempt to load the new-style properties file,
        // if extant.
        Properties p = null;
        final File newPropsPath = new File(exhibitBaseDir, relPathToNewProperties(exhibitName));
        if(newPropsPath.exists())
            {
            // Attempt to load from file and allow any errors to
            // propagate to caller.
            final FileInputStream fis = new FileInputStream(newPropsPath);
            try { p = FileTools.loadProperties(new BufferedInputStream(fis)); }
            finally { fis.close(); }
            }

        Object _description = null;
        // Try to retrieve description from new properties.
        if(p != null) { _description = p.getProperty(PNAME_DESCRIPTION); }
        // If not in new properties, and in right mode, look for old file.
        if(READ_OLD_FILES && (_description == null))
            {
            final File oldDescriptionPath = new File(exhibitBaseDir, relPathToOldDescription(exhibitName));
            // If old file exists, try to read it, propagating any IOException.
            if(oldDescriptionPath.exists())
                {
                _description = FileTools.readTextFile(oldDescriptionPath);
                }
            }
        // Trim the description, and if zero length, set it back to null.
        if(_description instanceof String)
            {
            if(_description != null) { _description = ((String) _description).trim(); }
            if((_description != null) && (((String) _description).length() == 0)) { _description = null; }
            // If it is over-length, complain on System.err and set it back to null.
            if((_description != null) && (((String) _description).length() > CoreConsts.DESCRIPTION_MAX_CHARS))
                {
                System.err.println("ERROR: description too long (max "+CoreConsts.DESCRIPTION_MAX_CHARS+" chars) for exhibit " + exhibitName);
                _description = null;
                }
            // If not deferring compaction and we have been passed a String
            // attempt to compact and intern() it now.
            if(!DEFER_COMPACTION)
                {
                try { _description = MemoryTools.intern(Compact7BitString.convertToCompact7BitString((String) _description, sDict)); }
                catch(final Exception e) { }
                }
            }

        Location.Base _location = null;
        // Try to retrieve description from new properties.
        if((p != null) &&
            // Check that at least minimal location field is present
            // before attempting full parse.
            (p.getProperty(PNAME_LOCATION_PREFIX + '.' + Location.Base.typeKey) != null))
            {
            _location = Location.Base.buildFromProperties(
                            true, // Mark as "specific".
                            PNAME_LOCATION_PREFIX,
                            p);
            }
        // If not in new properties, and in right mode, look for old file.
        if(READ_OLD_FILES && (_location == null))
            {
            final File oldLocationPath = new File(exhibitBaseDir, relPathToOldLocation(exhibitName));
            // If old file exists, try to read it, propagating any IOException.
            if(oldLocationPath.exists())
                { _location = Location.Base.buildFromFile(true, oldLocationPath); }
            }

        // Try to load new accession data, if present.
        // There is no point trying to intern() this to save memory
        // since it should be different for every exhibit,
        // though doing so may help reduce tenured heap churn during updates.
        AccessionData ad = null;
        final File newAccessionDataPath = new File(exhibitBaseDir, relPathToNewAccession(exhibitName));
        if(newAccessionDataPath.exists())
            {
            try
                {
                // Get the XML accession data from (UTF-8 format) file...
                final Reader r = new InputStreamReader(
                    new BufferedInputStream(
                        new FileInputStream(newAccessionDataPath)),
                                                CoreConsts.FILE_ENCODING_UTF_8);
                try
                    {
                    final StringBuilder sb = new StringBuilder((int) newAccessionDataPath.length());
                    int ic;
                    while((ic = r.read()) != -1)
                        { sb.append((char) ic); }
                    ad = AccessionData.parseFromXML(sb.toString());
                    }
                finally
                    { r.close(); }
                }
            catch(final Exception e)
                {
                e.printStackTrace();
                throw new IOException("unable to parse accession file: " + newAccessionDataPath + ": " + e.getMessage());
                }
            }


        // Try to create a new set of loadable properties.
        final ExhibitPropsLoadable result = new ExhibitPropsLoadable(_description,
                                                                     _location,
                                                                     ad);

        // If it is equivalent to the EMPTY one then return null instead.
        if(EMPTY.equals(result)) { return(null); }

        // Got some loadable properties!
        // Make sure that we eliminate any easy duplicates.
        return(MemoryTools.intern(result));
        }


    /**Based on entire content of object; if equivalent content then they are equal.
     * Can be used to eliminate duplicates, eg after deserialising.
     */
    public boolean equals(final Object o)
        {
        if(this == o) { return(true); } // True if same object!
        if(!(o instanceof ExhibitPropsLoadable)) { return(false); }
        final ExhibitPropsLoadable other = (ExhibitPropsLoadable) o;

        // Compare location for equality.
        // Handle null case specially.
        // This should be fast to compare, so do it early.
        if(location == null) { if(other.location != null) { return(false); } }
        else if(!location.equals(other.location)) { return(false); }

        // This may be somewhat slow to compare.
        if(accession == null) { if(other.accession != null) { return(false); } }
        else if(!accession.equals(other.accession)) { return(false); }

        // Compare description by external form (likely to be very slow).
        // Use quick internal instance compare first to see
        // if we really need to decode the description...
        if(description != other.description)
            {
            final String desc = getDescription();
            final String descO = other.getDescription();
            // Handle null case specially.
            if(desc == null) { if(descO != null) { return(false); } }
            else if(!desc.equals(descO)) { return(false); }
            }

        assert(hashCode() == other.hashCode());

        return(true); // Yes, they are the same!
        }

    /**Human-readable summary. */
    @Override public String toString()
        {
        final StringBuilder sb = new StringBuilder();
        sb.append("<ExhibitPropsLoadable");
        final String desc = getDescription();
        if(desc != null) { sb.append(":description=").append(description); }
        if(location != null) { sb.append(":location=").append(location); }
        if(accession != null) { sb.append(":accession=").append(accession); }
        sb.append(">");
        return(sb.toString());
        }

    /**Private to hashCode() lock-free cache of computed hash value; initially 0 indicating not computed. */
    private transient int _hash;

    /**Based on entire content of object; if equivalent content then they have same hash.
     * Can be used for tracking if content has changed from one scan to another,
     * ie is optimised for spotting changes rather than key distribution for use
     * in a Hashtable, for example, though it should be serviceable for
     * that purpose too as it is computed once and cached (transiently).
     * <p>
     * The cache relies on atomic assignment of int and is lock-free.
     * <p>
     * This may be slow to compute.
     * <p>
     * The hash value is never zero.
     */
    public int hashCode()
        {
        int result = _hash;
        if(result == 0) // Result needs recomputing.
            {
            // Recompute and cache.
            try {
                // Make a CRC-32 `digest' or checksum...
                // We use CRC32 rather than Adler32 because we are
                // likely to be operating on small amounts of data
                // or small changes where CRC32 may prove much superior.
                final java.util.zip.Checksum check = new java.util.zip.CRC32();
                final DataOutputStream dos =
                    new DataOutputStream(
                        (new OutputStream(){
                            public final void write(int b)
                                { check.update(b); }
                            public final void write(byte[] b, int off, int len)
                                { check.update(b, off, len); }
                            })
                        );

                // Insert data/property values into the checksum...
                // We must always insert the same value for a field
                // while its external representation remains unchanged,
                // and insert a different value if the external value changes,
                // so as to make the hash sensitive to what a user of the
                // class would see.

                // Hash the description as a UTF8 String in its external representation...
                final String desc = getDescription();
                if(desc == null) { dos.writeByte(255); } else { dos.writeUTF(desc); }

                // We write out the hashCode() to avoid toString() formatting funnies.
                final Location.Base loc = getLocation();
                if(loc == null) { dos.writeByte(1); } else { dos.writeInt(loc.hashCode()); }

                // Capture the accession data as part of the hash/checksum...
                final AccessionData accMD = getAccessionMetadata();
                if(accMD == null) { dos.writeByte(2); } else { dos.writeInt(accMD.hashCode()); }

                dos.flush(); // Make sure that all the data is in the checksum.
                result = (int) check.getValue();
                if(result == 0) { result = 1; } // Ensure computed hash is non-zero.
                _hash = result; // Cache the computed hash.
                }
            catch(IOException e) { throw new Error("internal error"); } // Should not happen.
            }
        return(result); // Return the hash...
        }



    /**My serial ID. */
    private static final long serialVersionUID = 8659120481539311156L;

    /**Deserialise: use constructor for validation, defensive copying, etc.
     * Also resolve all empty instances to a singleton as a minor optimisation,
     * and immediately intern() new values so as to immediately discard
     * duplicates (eg of exhibits already known) ASAP to minimise heap churn.
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Avoid duplicates of empty case.
        if(equals(EMPTY)) { return(EMPTY); }

        // Construct new instance of object in normal defensive way.
        // Try to intern/compact if not deferring this.
        if(!DEFER_COMPACTION)
            {
            Object d = description;
            try { d = Compact7BitString.convertToCompact7BitString(getDescription(), sDict); }
            catch(final Exception e) { }
            return(new ExhibitPropsLoadable(d, location, accession));
            }
        return(this);
        }

    /**Serialise: write in the best format for the wire.
     * To get best aggregate compressed size on the wire,
     * eg where the compressed stream contains many similar non-identical instances,
     * we always write out the metadata in its String format (minus the outer tags)
     * regardless of how it is actually held in memory.
     * This also makes us immune to changes in the the internals of the other formats
     * and allows use of a static dictionary with Compact7BitString for better in-memory compression
     * (ie effectively cross-instance compression).
     * <p>
     * This allows a stream compressor to effectively compress the redundancy between
     * instances of this class on the wire as well as internal redundancies.
     * <p>
     * We assume that there will almost never be identical instances on one stream
     * so we don't mind writing new copies each time where it does happen.
     */
    protected Object writeReplace()
        // throws ObjectStreamException
        {
        // Don't write multiple EMPTY instances.
        if(this.equals(EMPTY)) { return(EMPTY); }

        // If the description is null or String format then write this instance as-is.
        if((null == description) || (description instanceof String)) { return(this); }

        // Return a String-based instance for better inter-instance stream compression,
        // and for a more stable serialised representation.
        // We make sure that the String description is de-duped.
        return(new ExhibitPropsLoadable(MemoryTools.intern(getDescription()),
                        location,
                        accession));
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        // Check that the description is in an allowed form and can be decompressed.
        if(description != null)
            {
            if(description instanceof String)
                {
                final int len = ((String) description).length();
                if((len < 1) || (len > CoreConsts.DESCRIPTION_MAX_CHARS))
                    { throw new InvalidObjectException("bad object: bad description String length"); }
                }
            else if(description instanceof Compact7BitString)
                {
                final String s = description.toString();
                final int len = s.length();
                if((len < 1) || (len > CoreConsts.DESCRIPTION_MAX_CHARS))
                    { throw new InvalidObjectException("bad object: bad description Compact7BitString length"); }
                }
            else if(description instanceof ROByteArray)
                {
                final int rawLen = ((ROByteArray) description).length();
                if((rawLen < 1) || (rawLen > CoreConsts.DESCRIPTION_MAX_CHARS))
                    { throw new InvalidObjectException("bad object: bad description ROByteArray length"); }
                try {
                    final int len = (ROByteArray.uncompressToString((ROByteArray) description)).length();
                    if((len < 1) || (len > CoreConsts.DESCRIPTION_MAX_CHARS))
                        { throw new InvalidObjectException("bad object: bad description ROByteArray uncompressed length"); }
                    }
                catch(final Exception e)
                    { throw new InvalidObjectException("bad object: bad description ROByteArray: " + e.getMessage()); }
                }
            else if(description instanceof byte[])
                {
                final int rawLen = ((byte[]) description).length;
                if((rawLen < 1) || (rawLen > CoreConsts.DESCRIPTION_MAX_CHARS))
                    { throw new InvalidObjectException("bad object: bad description byte[] length"); }
                try {
                    final int len = _decompressASCII7Text((byte[]) description).length();
                    if((len < 1) || (len > CoreConsts.DESCRIPTION_MAX_CHARS))
                        { throw new InvalidObjectException("bad object: bad description byte[] uncompressed length"); }
                    }
                catch(final Exception e)
                    { throw new InvalidObjectException("bad object: bad description byte[]: " + e.getMessage()); }
                }
            else
                { throw new InvalidObjectException("bad object: bad description type"); }
            }

        // If location data is not null, it must be marked "specific".
        if(location != null)
            {
            if(Location.NONE.equals(location))
                { throw new InvalidObjectException("bad object: location must be null rather than NONE"); }
            if(location.specific == false)
                { throw new InvalidObjectException("bad object: location data must be \"specific\""); }
            }

        if(AccessionData.EMPTY.equals(accession))
            { throw new InvalidObjectException("bad object: accession must be null rather than EMPTY"); }
        }

    /**Suffix for new-format exhibit-properties file, if extant. */
    public static final String EXHIBIT_PROPS_FILENAME_SUFFIX = ".props";

    /**Allow backwards compatability if true, ie look for old-style files for info not found in new props file. */
    public static final boolean READ_OLD_FILES = true;

    /**Name of property for (default English text/HTML) description in new properties file. */
    public static final String PNAME_DESCRIPTION = "description";

    /**Prefix of property names for location information in new properties file. */
    public static final String PNAME_LOCATION_PREFIX = "location";

    /**Shared static dictionary for use with in-memory Compact7BitString metadata.
     * We expose this to help generate tuned dictionary values;
     * this is safe to do since this dictionary is immutable.
     */
    public static final Compact7BitString.StaticDictionary sDict = new Compact7BitString.StaticDictionary("EPL",
        // Most-common and largest early tokens first in list for maximum savings.
        Arrays.asList(new String[]{
            "the",    /* count=804, saving=1608, meanFirstPos=18 */
            "Taken",    /* count=285, saving=1140, meanFirstPos=4 */
            "and",    /* count=462, saving=924, meanFirstPos=30 */
            "from",    /* count=211, saving=633, meanFirstPos=27 */
            "with",    /* count=208, saving=624, meanFirstPos=26 */
            ", ",    /* count=605, saving=605, meanFirstPos=25 */
            "of",    /* count=541, saving=541, meanFirstPos=25 */
            "The",    /* count=260, saving=520, meanFirstPos=20 */
            "target",    /* count=102, saving=510, meanFirstPos=41 */
            "Arrived",    /* count=82, saving=492, meanFirstPos=3 */
            "in",    /* count=448, saving=448, meanFirstPos=27 */
            "this",    /* count=146, saving=438, meanFirstPos=32 */
            "http",    /* count=141, saving=423, meanFirstPos=29 */
            "that",    /* count=136, saving=408, meanFirstPos=42 */
            "blockquote",    /* count=45, saving=405, meanFirstPos=44 */
            "called",    /* count=78, saving=390, meanFirstPos=17 */
            "href",    /* count=130, saving=390, meanFirstPos=27 */
            "This",    /* count=125, saving=375, meanFirstPos=17 */
            "to",    /* count=375, saving=375, meanFirstPos=33 */
            "for",    /* count=168, saving=336, meanFirstPos=36 */
            "_top",    /* count=111, saving=333, meanFirstPos=43 */
            "over",    /* count=111, saving=333, meanFirstPos=14 */
            "is",    /* count=312, saving=312, meanFirstPos=28 */
            "was",    /* count=149, saving=298, meanFirstPos=33 */
            "on",    /* count=283, saving=283, meanFirstPos=27 */
            "://",    /* count=141, saving=282, meanFirstPos=30 */
            "afternoon",    /* count=33, saving=264, meanFirstPos=13 */
            "distribution",    /* count=24, saving=264, meanFirstPos=12 */
            " (",    /* count=252, saving=252, meanFirstPos=30 */
            ". ",    /* count=240, saving=240, meanFirstPos=34 */
            "www",    /* count=120, saving=240, meanFirstPos=32 */
            "taken",    /* count=57, saving=228, meanFirstPos=22 */
            "London",    /* count=45, saving=225, meanFirstPos=27 */
            "are",    /* count=112, saving=224, meanFirstPos=38 */
            "picture",    /* count=37, saving=222, meanFirstPos=29 */
            "about",    /* count=55, saving=220, meanFirstPos=33 */
            "Originally",    /* count=24, saving=216, meanFirstPos=0 */
            "background",    /* count=24, saving=216, meanFirstPos=26 */
            "which",    /* count=54, saving=216, meanFirstPos=45 */
            "at",    /* count=212, saving=212, meanFirstPos=28 */
            "email",    /* count=52, saving=208, meanFirstPos=16 */
            "resolution",    /* count=23, saving=207, meanFirstPos=20 */
            "Scanned",    /* count=34, saving=204, meanFirstPos=9 */
            "2000",    /* count=67, saving=201, meanFirstPos=15 */
            "scanned",    /* count=32, saving=192, meanFirstPos=12 */
            "building",    /* count=27, saving=189, meanFirstPos=25 */
            "have",    /* count=62, saving=186, meanFirstPos=50 */
            "100dpi",    /* count=35, saving=175, meanFirstPos=18 */
            "by",    /* count=175, saving=175, meanFirstPos=30 */
            "people",    /* count=35, saving=175, meanFirstPos=43 */
            "...",    /* count=85, saving=170, meanFirstPos=34 */
            "colour",    /* count=34, saving=170, meanFirstPos=27 */
            "compression",    /* count=17, saving=170, meanFirstPos=47 */
            "England",    /* count=28, saving=168, meanFirstPos=31 */
            "local",    /* count=42, saving=168, meanFirstPos=26 */
            "\" ",    /* count=167, saving=167, meanFirstPos=34 */
            ") ",    /* count=165, saving=165, meanFirstPos=35 */
            "very",    /* count=55, saving=165, meanFirstPos=30 */
            ".  ",    /* count=82, saving=164, meanFirstPos=31 */
            "some",    /* count=54, saving=162, meanFirstPos=32 */
            "Net",    /* count=78, saving=156, meanFirstPos=10 */
            "more",    /* count=52, saving=156, meanFirstPos=34 */
            "after",    /* count=38, saving=152, meanFirstPos=32 */
            "2004",    /* MANUALLY ADDED: prefix count ~ 70 */
            "</",    /* count=147, saving=147, meanFirstPos=48 */
            "Possibly",    /* count=21, saving=147, meanFirstPos=1 */
            "See",    /* count=73, saving=146, meanFirstPos=16 */
            "can",    /* count=73, saving=146, meanFirstPos=39 */
            " \"",    /* count=144, saving=144, meanFirstPos=24 */
            "2003",    /* count=48, saving=144, meanFirstPos=17 */
            "apparently",    /* count=16, saving=144, meanFirstPos=26 */
            "=\"",    /* count=142, saving=142, meanFirstPos=28 */
            "but",    /* count=71, saving=142, meanFirstPos=45 */
            "says",    /* count=47, saving=141, meanFirstPos=16 */
            "image",    /* count=35, saving=140, meanFirstPos=37 */
            "original",    /* count=20, saving=140, meanFirstPos=40 */
            "small",    /* count=34, saving=136, meanFirstPos=27 */
            "like",    /* count=45, saving=135, meanFirstPos=28 */
            "think",    /* count=33, saving=132, meanFirstPos=36 */
            "through",    /* count=22, saving=132, meanFirstPos=42 */
            "as",    /* count=130, saving=130, meanFirstPos=37 */
            "rather",    /* count=26, saving=130, meanFirstPos=38 */
            "html",    /* count=43, saving=129, meanFirstPos=41 */
            "GMT",    /* count=64, saving=128, meanFirstPos=11 */
            "; ",    /* count=127, saving=127, meanFirstPos=31 */
            "one",    /* count=63, saving=126, meanFirstPos=37 */
            "behind",    /* count=25, saving=125, meanFirstPos=38 */
            " <",    /* count=123, saving=123, meanFirstPos=25 */
            "images",    /* count=24, saving=120, meanFirstPos=33 */
            "foreground",    /* count=13, saving=117, meanFirstPos=23 */
            "camera",    /* count=22, saving=110, meanFirstPos=30 */
            "University",    /* count=12, saving=108, meanFirstPos=13 */
            "between",    /* count=18, saving=108, meanFirstPos=31 */
            "just",    /* count=36, saving=108, meanFirstPos=27 */
            "outside",    /* count=18, saving=108, meanFirstPos=18 */
            "subject",    /* count=18, saving=108, meanFirstPos=17 */
            "little",    /* count=21, saving=105, meanFirstPos=29 */
            "buildings",    /* count=13, saving=104, meanFirstPos=25 */
            "From",    /* count=34, saving=102, meanFirstPos=17 */
            "station",    /* count=17, saving=102, meanFirstPos=18 */
            "white",    /* count=25, saving=100, meanFirstPos=24 */
            "yellow",    /* count=19, saving=95, meanFirstPos=22 */
            "midnight",    /* count=13, saving=91, meanFirstPos=9 */
            "June",    /* count=30, saving=90, meanFirstPos=18 */
            "Routemaster",    /* count=9, saving=90, meanFirstPos=8 */
            "blue",    /* count=30, saving=90, meanFirstPos=14 */
            "green",    /* count=22, saving=88, meanFirstPos=16 */
            "thanks",    /* count=17, saving=85, meanFirstPos=18 */
            "Note",    /* count=28, saving=84, meanFirstPos=10 */
            "Probably",    /* count=12, saving=84, meanFirstPos=13 */
            "Canon",    /* count=18, saving=72, meanFirstPos=5 */
            "Hattusha",    /* count=10, saving=70, meanFirstPos=0 */
            "Bruno",    /* count=17, saving=68, meanFirstPos=7 */
            "Brighton",    /* count=9, saving=63, meanFirstPos=9 */
            "Courtesy",    /* count=8, saving=56, meanFirstPos=0 */
            "Looking",    /* count=9, saving=54, meanFirstPos=2 */
            "Cappadocia",    /* count=5, saving=45, meanFirstPos=0 */
            "Southbound",    /* count=5, saving=45, meanFirstPos=0 */
            "Guarameranga",    /* count=4, saving=44, meanFirstPos=0 */
            "View",    /* count=14, saving=42, meanFirstPos=1 */
            "Three",    /* count=8, saving=32, meanFirstPos=0 */
        }));

    /**The description.
     * This can be null, meaning no description,
     * or a non-empty String for the literal description,
     * or a non-empty byte[] containing a maximally deflated
     * copy of the string without zlib or gzip headers and trailers
     * or a ROByteArray created with compressFromString()
     * or a Compact7BitString.
     * <p>
     * The byte[] format exists mainly for backwards compatibility.
     * <p>
     * In the byte[] case, all bytes in the source string were of value
     * 0--127 (ie ASCII 7-bit), and the byte[] is estimated to take up
     * less space than a plain String representation.
     * <p>
     * Marked volatile to allow safe lockless update by compact().
     */
    private volatile Object description;

    /**Get the description; returns null if none, else non-empty String.
     * This text may contain non-XML-safe and non-ASCII 7-bit characters.
     */
    public String getDescription()
        {
        // Capture local/fast snapshot.
        final Object d = description;

        if(d == null) { return(null); }
        if(d instanceof String) { return((String) d); }
        if(d instanceof Compact7BitString) { return(d.toString()); }

        // OK, prepare to uncompress the data.
        try
            {
            if(d instanceof ROByteArray)
                { return(ROByteArray.uncompressToString((ROByteArray) d)); }

            // Assume byte[] format.
            return(_decompressASCII7Text((byte[]) d));
            }
        catch(final Exception e) { return(null); } // Return null in case of problem!
        }

    /**Reports if there is a description; returns true if so.
     * This is relatively inexpensive as it does not require
     * inspecting or decompressing (etc) any message text.
     */
    public boolean hasDescription()
        { return(description != null); }


    /**Attempt to decompress some deflated bytes to an ASCII 7-bit String.
     * (Actually, we won't mind if it happens to be 8-bit.)
     */
    private static String _decompressASCII7Text(byte compressed[])
        throws IOException
        {
        final byte raw7bitText[] = FileTools.decompressDeflatedData(compressed);
        // Turn this into a String.
        final StringBuilder sb = new StringBuilder(raw7bitText.length);
        for(int i = 0; i < raw7bitText.length; ++i)
            { sb.append((char) ((raw7bitText[i]) & 0xff)); }
        return(sb.toString());
        }

//    /**Attempt to compress an ASCII 7-bit String to a maximally-deflated (no-header/no-trailer) byte[].
//     * The string must be pure ASCII 7-bit, ie all characters 0--127.
//     */
//    private static byte[] _compressASCII7Text(String in)
//        {
//        final byte raw7bitText[] = new byte[in.length()];
//        for(int i = raw7bitText.length; --i >= 0; )
//            {
//            final char c = in.charAt(i);
//            assert (c & ~0x7f) == 0;
//            raw7bitText[i] = (byte) c;
//            }
//        return(FileTools.compressDeflatableData(raw7bitText, 0, raw7bitText.length));
//        }

    /**Location information; null if none.
     * Information here must always be marked "specific".
     */
    private final Location.Base location;

    /**Return location information; null if none.
     * Always marked as "specific".
     */
    public Location.Base getLocation()
        {
        return(location);
        }

    /**Accession data; null if none. */
    private final AccessionData accession;

    /**Get exhibit accession metadata; null if none available.
     * This data includes such items as:
     * <ul>
     * <li>When the exhibit was loaded into to the Gallery.
     * <li>Its checksums, timestamp, length, etc, when added to the Gallery.
     * </ul>
     */
    public AccessionData getAccessionMetadata()
        {
        return(accession);
        }

    /**If true then defer compaction of metadata.
     * If true, then (any) compaction is <em>NOT</em> done
     * during construction or deserialisation,
     * but until some later point such as a call to compact().
     * <p>
     * This should save a lot of time constructing/deserialising data
     * before the first operations can be performed on it
     * and where memory space is not the primary constraint.
     */
    private static final boolean DEFER_COMPACTION = true;

    /**Compact the internal representation of this instance (and its sub-objects) if possible.
     * This has no effect on the logical content of this instance in-memory or serialised,
     * is guaranteed to be safe to run concurrently with other uses of this instance
     * (and will take any locks as needed to work incrementally),
     * and may do nothing but consume some CPU cycles.
     * <p>
     * This may be able to convert some state to a more memory-efficient representation
     * after construction or deserialisation,
     * and is suitable to call in a background thread.
     * <p>
     * We don't prevent multiple concurrent calls to this routine,
     * since they are at worst wasteful of CPU but not unsafe.
     */
    public void compact()
        {
        if(!DEFER_COMPACTION) { return; /* Nothing to do. */ }

        // Ensure that the the description is compacted and intern()ed.

        final Object d = description; // Capture snapshot.

        // If already in compact (non-String) form (or null) then intern/return.
        if(!(d instanceof String))
            {
            description = MemoryTools.intern(d);
            return;
            }

        // Replace metadata with intern()ed compact form if possible.
        // DHD20060815: We observe about half the metadata instances to be duplicates.
        try { description = Compact7BitString.convertToCompact7BitString((String) d, sDict); }
        // Cope with non-7-bit-pure text by leaving as-is, though intern()ed.
        catch(final IllegalArgumentException e) { description = MemoryTools.intern((String) d); }
        }


    /**Method that lists exhibit-specific files containing loadable properties of the given exhibit.
     * This takes the same arguments as getLoadableProperties() and returns
     * a Set of the String names of all files (relative to exhibitBaseDir)
     * that contain loadable-properties for this exhibit.
     * <p>
     * This may name more files than are currently used by the system,
     * eg including legacy files.
     * <p>
     * This never returns null, though may return an empty set.
     * <p>
     * When backing up exhibit data these files should be backed up
     * with the relevant exhibit.
     */
    public static Set<String> listPropsFiles(final String exhibitName,
                                             final File exhibitBaseDir)
        {
//        final File fullExhibitPath = new File(exhibitBaseDir, exhibitName);
//        if(!fullExhibitPath.exists())
//            { throw new FileNotFoundException("exhibit missing"); }

        // Generate a Set of String candidate names,
        // and eliminate those that don't actually exist.
        final Set<String> result = new HashSet<String>(11);
        // New properties file.
        result.add(relPathToNewProperties(exhibitName));
        // Old description file.
        result.add(relPathToOldDescription(exhibitName));
        // New accession file.
        result.add(relPathToNewAccession(exhibitName));
        // Old accession file.
        result.add(relPathToOldAccession(exhibitName));
        // Old location file.
        result.add(relPathToOldLocation(exhibitName));

        for(final Iterator<String> it = result.iterator(); it.hasNext(); )
            {
            final String f = it.next();
            if(!(new File(exhibitBaseDir, f)).exists())
                {
                // We should remove this file as it doesn't exist.
                it.remove();
                }
            }

        return(result);
        }
    }

