/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.Serializable;

/**Immutable, Serializable collection of all the static attributes of an exhibit.
 * This is designed to be fast to do comparisons against, and to be
 * transportable across a network, persistable to disc, and easy to run
 * filtering queries against.
 * <p>
 * This verifies that its contents make sense at construction and
 * after deserialisation.
 * <p>
 * This is keyed on the exhibit underlying name, but should be easy to
 * use with other comparators.
 * <p>
 * The timestamp can be used to compare that the version of the object
 * held is up-to-date.
 * <p>
 * The core fields are made available as public final fields
 * for speed of access.
 */
public final class ExhibitStaticAttr implements Comparable,
                                        Serializable, ObjectInputValidation
    {
    /**The full (relative) path/name of the exhibit; never null.
     * Must be unique and acceptable to ExhibitName.
     */
    public final String filePath;

    /**The length of the exhibit in bytes; is strictly positive.
     */
    public final long length;

    /**The Java-style timestamp (last-modified-time) of the exhibit.
     * Strictly positive.
     */
    public final long timestamp;


    /**All comparison is done on the name. */
    public final int hashCode()
        { return(filePath.hashCode()); }

    /**All comparison is done on the name, NOT on all fields. */
    public final boolean equals(final Object o)
        {
        if(!(o instanceof ExhibitStaticAttr)) { return(false); }
        return(filePath.equals(((ExhibitStaticAttr) o).filePath));
        }

    /**Human-readable summary contain at least the name. */
    public String toString()
        {
        return(filePath + "["+length+"b]");
        }

    /**All comparison is done on the name. */
    public final int compareTo(final Object o)
        {
        return(filePath.compareTo(((ExhibitStaticAttr) o).filePath));
        }

    /**Tests that instances are identical in all parameters. */
    public final boolean isIdentical(final ExhibitStaticAttr other)
        {
        if(other == this) { return(true); }
        // We compare fields in the order that will probably
        // fail fastest for non-identical exhibits.
        return((timestamp == other.timestamp) &&
               (length == other.length) &&
               filePath.equals(other.filePath));
        }

    /**Construct new immutable instance.
     * Try to force sharing of the potentially large and numerous paths
     * to avoid memory bloat, even though this costs some time.
     *
     * @throws IllegalArgumentException  in the case of bad data
     */
    public ExhibitStaticAttr(final String _filePath,
                             final long _length,
                             final long _timestamp)
        throws IllegalArgumentException
        {
        filePath = MemoryTools.intern(_filePath); // Eliminate duplicates.
        length = _length;
        timestamp = _timestamp;

        // Verify what we've been given.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e); }
        }


    /**My initial version number. */
    private static final long serialVersionUID = -8233812711746526095L;

    /**Deserialise: use constructor for validation, defensive copying, etc. */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Construct new instance of object in normal defensive way.
        // This also ensures that we get a chance to intern() stuff, etc,
        // and reduce old-heap churn in generational GC systems.
        return(new ExhibitStaticAttr(filePath, length, timestamp));
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe
        // (eg strings not dangerously long).
        if(!ExhibitName.validNameSyntax(filePath))
            { throw new InvalidObjectException("bad object: name = ``" + filePath + "''"); }
        if(length <= 0)
            { throw new InvalidObjectException("bad object: length = " + length); }
        if(!GenUtils.isValidGalleryTimestamp(timestamp))
            { throw new InvalidObjectException("bad object: timestamp = " + timestamp + " ("+(new java.util.Date(timestamp))+") for exhibit " + filePath); }
        }
    }
