/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.awt.*;
import java.io.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Date;

/**Immutable, Serializable thumbnails/samples of an exhibit.
 * This is designed to be transportable across a network if need be,
 * and persistable to disc, and reasonably efficient when serialised.
 * <p>
 * This holds a standard and a small thumbnail (of a static image)
 * or equivalent sample of an exhibit for other exhibit types,
 * or an indication that such a sample is not needed because the original can be used,
 * or an indication that such a sample could not be made (eg failed and should not be rebuilt).
 * <p>
 * Exhibit thumbnails/samples are strictly limited in size to minimise
 * the space they consume in toto and the time that they take to send across
 * the network to clients.  The small thumbnails are meant to be small
 * enough that they can be permanently held in memory at the servers
 * and are lightning-fast to transfer to client (while not so small
 * as to simple dwarfed by network overheads).
 * <p>
 * (All thumbnails/samples must be smaller than the original exhibit.)
 * <p>
 * Any sample is of the same MIME type as the original.
 * <p>
 * An example of when an exhibit could be used as its own thumbnail is
 * when its size in bytes is less than the appropriate limit, and the (x,y)
 * dimensions are no greater than the appropriate limits, then no thumbnail
 * is required and the image can be used directly.
 * <p>
 * This is made to implement equals() and hashCode() so that
 * comparisons for equality, eg in a Set, can be performed.
 * Two objects are equal if their constituent thumbnails are equal.
 * The hash is computed from those of the thumbnails.
 */
public final class ExhibitThumbnails implements Serializable, ObjectInputValidation
    {
    /**Value indicating that no thumbnails could be made for an exhibit.
     * This non-null value positively indicates that there are no thumbnails.
     * <p>
     * This value is a singleton.
     */
    public static final ExhibitThumbnails NO_THUMBNAILS =
        new ExhibitThumbnails(null, null);

    /**Returns true if its argument is null or contains no thumbnails.
     */
    public static boolean isEmpty(final ExhibitThumbnails tns)
        {
        if(tns == null) { return(true); }
        return((tns.getSmall() == null) && (tns.getStandard() == null));
        }

    /**Small thumbnail; null means none can be constructed. */
    private final Thumbnail sml;

    /**Get small thumbnail; null means none can be constructed. */
    public Thumbnail getSmall() { return(sml); }

    /**Standard thumbnail; null means none can be constructed. */
    private final Thumbnail std;

    /**Get standard thumbnail; null means none can be constructed. */
    public Thumbnail getStandard() { return(std); }

    /**Creation timestamp (positive), or zero if none. */
    public final long created;

    /**Construct from a pair of thumbnails.
     * Either can be null or non-null, but if both are non-null
     * small.size() must not be greater than standard.size().
     * <p>
     * The timestamp is always zero if both thumbnails are.
     */
    private ExhibitThumbnails(final Thumbnail small, final Thumbnail standard)
        {
        sml = small;
        std = standard;
        created = ((small == null) && (standard == null)) ? 0 : System.currentTimeMillis();

        // Verify what we've been given.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Absolute maximum size of any standard thumbnail/sample binary in bytes (32kB-1).
     * Should be small enough to download reasonably quickly,
     * but may be too large to cache in memory.
     * <p>
     * This value is small enough to store in a short.
     */
    public static final short STD_ABS_MAX_BYTES = 0x7fff;

    /**Absolute maximum size of any small thumbnail/sample binary in bytes (4kB-1).
     * Should be very fast to download, even over a WAN connection,
     * and small enough to cache all exhibits' small thumbnails in memory if required.
     * <p>
     * This value is small enough to store in a short.
     */
    public static final short SML_ABS_MAX_BYTES = 0xfff;


    /**For a standard static image thumbnail, its longest dimension in pixels.
     * Normally the longest dimension will be exactly this in pixels,
     * but it may be smaller.
     */
    public static final int STD_STATIC_IMAGE_TN_LDIM_PX = 256;

    /**For a small static image thumbnail, its longest dimension in pixels.
     * Normally the longest dimension will be exactly this in pixels,
     * but it may be smaller.
     */
    public static final int SML_STATIC_IMAGE_TN_LDIM_PX = 64;

    /**Crude estimate of maximum bytes consumed in memory and on disc by thumbnails; ignores VM and serialisation details, etc. */
    public static final int MAX_BYTES_EST = STD_ABS_MAX_BYTES + SML_ABS_MAX_BYTES + 1024;

    /**Returns a hash code value for the object.
     * This is based on the hashes of the constituent thumbnails.
     */
    public int hashCode()
        {
        // Uses unique (different, small-prime) values for missing thumbnails.
        final int smlHC = ((sml == null) ? 3 : sml.hashCode());
        // We work a bit harder on the std thumbnail value as we assume it
        // will be missing more often and so we will pay a slightly lower
        // cost than it would otherwise appear.
        final int stdHC = ((std == null) ? 5 : (17 * std.hashCode()));
        return(smlHC ^ stdHC);
        }

    /**Indicates whether some other object is "equal to" this one.
     * For this the constituent thumbnails must be "equal",
     * and it means that iff one of the thumbnails is missing in one
     * of the objects then it must be missing in the other too.
     * <p>
     * Equality ignores the creation timestamp.
     */
    public boolean equals(final Object obj)
        {
        if(obj == this) { return(true); } // Will be reasonably common case.
        if(!(obj instanceof ExhibitThumbnails)) { return(false); }
        final ExhibitThumbnails other = (ExhibitThumbnails) obj;

        // Possibly quicker to test equality on smaller thumbnail first.
        final boolean smlIsSame = (sml == null) ?
            (other.sml == null) :
            (sml.equals(other.sml));
        if(!smlIsSame) { return(false); }

        final boolean stdIsSame = (std == null) ?
            (other.std == null) :
            (std.equals(other.std));
        if(!stdIsSame) { return(false); }

        return(true); // Seem to be the same!
        }

    /**Given the original dimensions of an image, compute the size of the standard/small thumbnail.
     * The argument must be non-null and the width and height strictly positive.
     * <p>
     * If the longest dimension of the original image is no more than
     * {STD|SML}_STATIC_IMAGE_TN_LDIM_PX pixels then the input dimensions will
     * be returned unaltered, else the values are scaled so that the
     * longest dimension is exactly {STD|SML}_STATIC_IMAGE_TN_LDIM_PX pixels.
     *
     * @param std  if true, compute for standard, else compute for small
     */
    public static Dimension computeThumbnailDimensions(final Dimension original,
                                                       final boolean std)
        {
        if(original == null) { throw new IllegalArgumentException(); }
        final int w = original.width;
        final int h = original.height;
        if((w <= 0) || (h <= 0)) { throw new IllegalArgumentException(); }

        final int limit = std ? STD_STATIC_IMAGE_TN_LDIM_PX : SML_STATIC_IMAGE_TN_LDIM_PX;

        final int longestDim = Math.max(w, h);
        if(longestDim <= limit) { return(original); }

        final Dimension result = new Dimension();
        result.width  = (w >= h) ? limit : Math.min(limit, (w * limit + longestDim/2) / longestDim);
        result.height = (h >= w) ? limit : Math.min(limit, (h * limit + longestDim/2) / longestDim);

        assert Math.max(result.width, result.height) <= limit;

        return(result);
        }


    /**My initial version number. */
    private static final long serialVersionUID = 2686345754458313314L;

    /**Deserialise. */
    private void readObject(final ObjectInputStream in)
        throws IOException, ClassNotFoundException
        {
        in.defaultReadObject();
        validateObject(); // Validate state immediately.
        }

    /**Eliminate duplicate empty objects after deserialisation. */
    protected Object readResolve()
        {
        if(NO_THUMBNAILS.equals(this))
            { return(NO_THUMBNAILS); }
        return(this);
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        // We check that the small thumbnail is no larger than the standard one.
        // We should probably also check thumbnail image dimensions...
        if((std != null) && (sml != null) && (sml.size() > std.size()))
            { throw new InvalidObjectException("bad object: sml is larger than std"); }

        // Check the harsher constraints on the small thumbnail.
        if(sml != null)
            {
            // The data must not be too large.
            if(sml.size() > SML_ABS_MAX_BYTES)
                { throw new InvalidObjectException("bad object: small thumbnail data too large"); }

            // xyDim, if present, must have strictly positive components.
            final Dimension xyDim = sml.getXyDim();
            if(xyDim != null)
                {
                if((xyDim.width > SML_STATIC_IMAGE_TN_LDIM_PX) || (xyDim.height > SML_STATIC_IMAGE_TN_LDIM_PX))
                    { throw new InvalidObjectException("bad object: small thumbnail xyDim width or height too large"); }
                }
            }

        // If the timestamp is not absent (zero) then it must be positive.
        if(created < 0)
            { throw new InvalidObjectException("bad object: invalid timestamp"); }
        if((sml == null) && (std == null) && (created != 0))
            { throw new InvalidObjectException("bad object: invalid non-zero timestamp with no thumbnail data"); }
        }

    /**Make a pair of thumbnails.
     * Either can be null or non-null, but if both are non-null
     * small.size() must not be greater than standard.size(),
     * and both must be within their respective size limits.
     */
    public static ExhibitThumbnails createExhibitThumbnails(final Thumbnail small, final Thumbnail standard)
        {
        // Avoid redundant creation of multiple "empty" placeholders.
        if((small == null) && (standard == null)) { return(NO_THUMBNAILS); }

        return(new ExhibitThumbnails(small, standard));
        }


    /**Immutable, Serializable single thumbnail (standard or small).
     * Implements hashCode() and equals(),
     * with equals() meaning dimensional and byte-for-byte equality.
     */
    public static final class Thumbnail implements Serializable, ObjectInputValidation
        {
        /**Construct an image thumbnail with x,y dimensions and a byte array.
         * The byte array must not be null nor zero-length.
         * The Dimension object must be non-null and the width and height
         * must be strictly positive.
         * <p>
         * Both arguments are copied and are not altered.
         */
        public Thumbnail(final byte _data[], final Dimension _xyDim)
            {
            data = _data.clone(); // Defensive copy.
            xyDim = (_xyDim == null) ? null : new Dimension(_xyDim); // Defensive copy.
            hashMD5 = _computeMD5Hash(data); // Compute new copy.

            // Verify what we've been given.
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Mark all mutable members as unshared for safety. */
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField("data", (new byte[0]).getClass(), true),
            new ObjectStreamField("xyDim", Dimension.class, true),
            new ObjectStreamField("hashMD5", (new byte[0]).getClass(), true),
            };

        /**Thumbnail raw data; non-null, non-zero-length. */
        private final byte data[];

        /**Image width and height for image thumbnail, both strictly positive or whole object null. */
        private final Dimension xyDim;

        /**The MD5 hash of the thumbnail data (16 bytes), or null.
         * This is only present (non-null) if computed at source/construction.
         * <p>
         * This may also be null when deserialised from an old instance
         * before we computed and stored this value.
         */
        private final byte hashMD5[];

        /**Return size of data array. */
        public int size() { return(data.length); }

        /**Get data as a (copied) byte[]. */
        public byte[] toByteArrray() { return(data.clone()); }

        /**Write data to a stream, avoiding an unnecessary copy in many cases.
         * We trust the os.write() method(s) not to alter
         * the data that we pass to them.
         *
         * @param os  (safe) OutputStream on which to write the binary data;
         *     never null
         * @throws IOException  if thrown by os
         */
        public void writeData(final OutputStream os)
            throws IOException
            { os.write(data); }

        /**Get XxY dimension of thumbnail, or null if not appropriate. */
        public Dimension getXyDim()
            {
            if(xyDim == null) { return(null); }
            return(new Dimension(xyDim));
            }

        /**The hash code is based on things we can compute very quickly.
         * For this we use:
         * <ul>
         * <li>The byte-for-byte size of the thumbnail data.
         * <li>The first and mid-point bytes of the thumbnail data
         *     (we know there must be at least one byte,
         *     and they may contain a magic number and something interesting).
         * </ul>
         * <p>
         * We could use the height and width, but we assume that they may
         * not be very effective in distinguishing between thumbnails as
         * they may be constrained to be fixed in one dimension or the other.
         */
        public int hashCode()
            {
            return((data.length * 65327) +
                   (((int) data[0]) * 257) + (data[data.length / 2]));
            }

        /**For equality the thumbnails must be indistinguishable.
         * That is, they must be the same dimensions and their
         * data length must be the same and the data must match byte for byte.
         */
        public boolean equals(final Object obj)
            {
            if(this == obj) { return(true); }
            if(!(obj instanceof Thumbnail)) { return(false); }
            final Thumbnail other = (Thumbnail) obj;

            // Check data length first,
            // as this is quick to do and should have a good spread of values,
            // ie should distingish most non-identical thumbnails.
            if(data.length != other.data.length) { return(false); }

            // Check for different dimensions.
            if(xyDim != null)
                { if(!xyDim.equals(other.xyDim)) { return(false); } }
            else
                { if(other.xyDim != null) { return(false); } }

            // Now try a byte-for-byte comparison, which may be very expensive.
            return(Arrays.equals(data, other.data));
            }

        /**Human-readable summary. */
        @Override
        public String toString()
            {
            final StringBuilder sb = new StringBuilder(64);
            sb.append("tn:data.length=").append(data.length);
            if(xyDim != null) { sb.append(':').append(xyDim); }
            return(sb.toString());
            }

        /**True iff this thumbnail has a precomputed MD5hash stored.
         * If true then getMD5Hash() does not actually have to calculate the hash.
         * <p>
         * If false then this instance has reduced protection against data corruption.
         */
        public boolean hasMD5Hash()
            {
            return(hashMD5 != null);
            }

        /**Get the MD5 hash of the thumbnail data; never null.
         * If this is stored in the object then we return a copy to the caller,
         * else we return the result of computing it afresh on the data.
         *
         * @return a private copy of the MD5 hash bytes; never null
         */
        public byte[] getMD5Hash()
            {
            if(hashMD5 != null) { return(hashMD5.clone()); }
            return(_computeMD5Hash(data));
            }

        /**Compute the MD5 hash of the supplied data; never null.
         * Computed on demand every time.
         */
        private byte[] _computeMD5Hash(final byte[] d)
            {
            final MessageDigest hMD5;
            try { hMD5 = MessageDigest.getInstance(CoreConsts.HASH_MD5); }
            catch(final NoSuchAlgorithmException e) // Should never happen...
                { throw new Error("could not find "+CoreConsts.HASH_MD5+" digester!"); }
            hMD5.update(d); // Digest data all in one go.
            return(hMD5.digest());
            }

        /**Serialization version number. */
        private static final long serialVersionUID =  -1279061480258712366L;

//        /**Deserialise: use constructor for validation, defensive copying, etc. */
//        protected Object readResolve()
//            // throws ObjectStreamException
//            {
//            // Construct new instance of object in normal defensive way.
//            return(new Thumbnail(data, xyDim));
//            }

        /**Deserialise. */
        private void readObject(final ObjectInputStream in)
            throws IOException, ClassNotFoundException
            {
            in.defaultReadObject();
            validateObject(); // Validate state immediately.
            }

        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * This only checks that this is good as a standard thumbnail:
         * extra checking may have to be imposed elsewhere for a small
         * thumbnail.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         */
        public void validateObject()
            throws InvalidObjectException
            {
            // Check that all components are sane and safe.

            // The data must be non-zero length.
            if((data == null) || (data.length == 0))
                { throw new InvalidObjectException("bad object: data null or zero length"); }
            // The data must not be too large.
            if(data.length > STD_ABS_MAX_BYTES)
                { throw new InvalidObjectException("bad object: data too large"); }

            // xyDim, if present, must have strictly positive components.
            if(xyDim != null)
                {
                if((xyDim.width <= 0) || (xyDim.height <= 0))
                    { throw new InvalidObjectException("bad object: xyDim width or height non-positive"); }
                if((xyDim.width > STD_STATIC_IMAGE_TN_LDIM_PX) || (xyDim.height > STD_STATIC_IMAGE_TN_LDIM_PX))
                    { throw new InvalidObjectException("bad object: xyDim width or height too large"); }
                }
            // Check combinations of values.
            // For the moment, xyDim must always be non-null.
            else
                { throw new InvalidObjectException("bad object: xyDim is null"); }

            // If the hash is present, it must match our data.
            if(hashMD5 != null)
                {
                if(!Arrays.equals(hashMD5, _computeMD5Hash(data)))
                    { throw new InvalidObjectException("bad object: data corrupt; does not match hash"); }
                }
            }
        }

    /**Human-readable summary. */
    @Override
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(128);
        sb.append("thumbnails");
        if(created != 0) { sb.append(":created=").append(new Date(created)); }
        if(sml != null) { sb.append(":sml:").append(sml); }
        if(std != null) { sb.append(":std:").append(std); }
        return(sb.toString());
        }
    }
