/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.FEC;

/**This holds compile-time constants for forward-error-correction (FEC) of exhibit data.
 */
public final class FECConsts
    {
    /**Prevent construction of an instance. */
    private FECConsts() { }


    /**Disc block size; postitive power of two larger than a disc sector.
     * We assume that an especially common failure mode
     * is for an entire sector to become unrecoverably unreadable,
     * and thus our repair algorithm should work when a block
     * at least this large (and on this alignment) fails.
     * <p>
     * Current common (2006) sector sizes are 512 bytes for magnetic media
     * and 1024 bytes for optical media,
     * with 4096 bytes being proposed as a new standard circa 2006.
     * <p>
     * We use the largest of these as our fundamental FEC block size.
     */
    public static final int BASIC_FEC_BLOCK_BYTES = (1<<12);

    /**FEC protection unit/block size; a positive power of two much larger than BASIC_FEC_BLOCK_BYTES.
     * Each block of this size of a raw exhibit file (and on this alignment)
     * is treated independently by the FEC algorithm.
     * <p>
     * Any one exhibit will have a single fixable-unit block size,
     * though some exhibits may transiently have a different-to-this size
     * after a change in this value until the FEC data can be recomputed.
     * <p>
     * Its is useful be able to manipulate (multiple variants of)
     * a whole such block in memory without straining the JVM/OS,
     * and so a value from hundreds of kB to tens of MB is probably reasonable.
     */
    public static final int FIXABLE_UNIT_FEC_BLOCK_SIZE = (1<<20);

    /**The "fixable unit" block-checksum cryptographic checksum algorithm name.
     * We want a modern, reliable, reasonably-fast, JVM-supported algorithm,
     * different to (and stronger/better than) the MD5 algorithm
     * used to checksum entire exhibit files.
     * <p>
     * Ideally it should also be different to any underlying filesystem checksum
     * (such as hard-disc CRCs/ECCs and the ZFS SHA-256 block-check),
     * so as to maximise the chance of this FEC algorithm caching something
     * that the others miss.
     */
    public static final String FIXABLE_UNIT_FEC_CHECKSUM_ALG = "SHA-512";

    /**Minimum number of "basic FEC block" failures per "fixable unit block" to be able to recover from; strictly positive.
     * For RS-style FEC the extra file space overhead is directly proportional to this.
     * <p>
     * Since we are expecting the underlying filesystem to have a very low
     * undetected/unrecoverable error rate then this can reasonably be as low as one.
     * <p>
     * We may be prepared to provide extra recovery/protection
     * for popular and/or heavily-used and/or designated exhibits.
     */
    public static final int MIN_BASIC_BLOCKS_REPAIRABLE = 1;
    }
