/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
    package org.hd.d.pg2k.svrCore;

import java.io.*;
import java.util.Properties;

    /**This class has tools for UNIX-like file features.
     * Things like the atomic `mv' operation, and `touch' and `chmod'
     * are provided here.  If we need to run code on more than a
     * UNIX platform these may prove challenging...
     *
     * @author Damon Hart-Davis
     * @version  1.17 02/01/04; derived from old FileTools 1.100 01/02/03
     */
    public final class FileTools
        {
        /**Prevent creation of instances of this class. */
        private FileTools() { }

        /**Prefix used on temporary files, eg while doing atomic replacements.
         * This used to be in GlobalParams but we may even need it
         * while loading GlobalParams.
         */
        public static final String F_tmpPrefix = ".tmp.pg.";

// --Recycle Bin START (13/07/03 16:43):
//        /**Prefix used on more permanent Gallery-generated files, eg saved thumbnail files.
//         * These are files that are meant to remain hidden (not readable)
//         * to most users.
//         */
//        public static final String F_permPrefix = ".pg.";
// --Recycle Bin STOP (13/07/03 16:43)

        /**Rough estimate of file-system allocation size for usage estimate (bytes).
         * We should separately round up both the file size and the filename size to
         * a multiple of this in order to make a conservative estimate of
         * filesystem usage when we are creating files, and especially if we
         * cannot measure our actual usage.
         * <p>
         * An estimate in the region of 512--1024 bytes is reasonable for most
         * filesystems.
         */
        public static final int FS_EST_BLOCK_SIZE_BYTES = 1024;

// --Recycle Bin START (13/07/03 16:43):
//        /**Checks that a filename is `safe', returning true if so.
//         * By `safe' is meant:
//         * <UL>
//         * <LI>Not containing shell metacharacters or separators.
//         * <LI>Generally containing only a small subset of characters
//         *     also allowable in URLs.
//         * </UL>
//         */
//        public static boolean isSafeFileName(final String fn)
//            {
//            final int l = fn.length();
//            if(l < 1) { return(false); } // Zero length filename?
//
//            for(int i = l; --i >= 0; )
//                {
//                final int ic = (int) fn.charAt(i);
//                if((ic >= sizeofFCO) || !fileCharOK[ic])
//                    { return(false); } // Bad char, abandon ship.
//                }
//
//            return(true); // All seems OK.
//            }
// --Recycle Bin STOP (13/07/03 16:43)
// --Recycle Bin START (13/07/03 16:43):
//        /**Used by isSafeFileName() to check each filename character.
//         * Any character that would be OK is set to true.
//         * Note that only 7-bit ASCII characters are even considered.
//         */
//        private static final boolean fileCharOK[] = makeFCO();
// --Recycle Bin STOP (13/07/03 16:43)
// --Recycle Bin START (13/07/03 16:44):
//        /**Size of fileCharOK[]. */
//        private static final int sizeofFCO = 128;
// --Recycle Bin STOP (13/07/03 16:44)
// --Recycle Bin START (13/07/03 16:44):
//        /**Initialiser code for fileCharOK[]. */
//        private static boolean[] makeFCO()
//            {
//            boolean tmpFCO[] = new boolean[sizeofFCO];
//            for(int i = tmpFCO.length; --i >= 0; )
//                {
//                switch(i)
//                    {
//                    // Misc characters...
//                    case '.': case '-': case '_': case '/':
//                    // Digits.
//                    case '0': case '1': case '2': case '3': case '4':
//                    case '5': case '6': case '7': case '8': case '9':
//                    // Upper-case ASCII letters.
//                    case 'A': case 'B': case 'C': case 'D': case 'E':
//                    case 'F': case 'G': case 'H': case 'I': case 'J':
//                    case 'K': case 'L': case 'M': case 'N': case 'O':
//                    case 'P': case 'Q': case 'R': case 'S': case 'T':
//                    case 'U': case 'V': case 'W': case 'X': case 'Y':
//                    case 'Z':
//                    // Lower-case ASCII letters.
//                    case 'a': case 'b': case 'c': case 'd': case 'e':
//                    case 'f': case 'g': case 'h': case 'i': case 'j':
//                    case 'k': case 'l': case 'm': case 'n': case 'o':
//                    case 'p': case 'q': case 'r': case 's': case 't':
//                    case 'u': case 'v': case 'w': case 'x': case 'y':
//                    case 'z':
//                        {
//                        tmpFCO[i] = true;
//                        }
//                    }
//                }
//            return(tmpFCO); // One and only assignment.
//            }
// --Recycle Bin STOP (13/07/03 16:44)


        /**Get the extension of a file, not including the leading dot.
         * Returns null if no extension present.
         */
        public static final String getExtension(String name)
            {
            final int pos = name.lastIndexOf('.');
            if(pos == -1) { return(null); }
            final String extension = name.substring(pos + 1);
            if(extension.indexOf(File.separatorChar) != -1)
                { return(null); } // Whoops, was not really extension.
            return(extension);
            }

// --Recycle Bin START (13/07/03 16:44):
//        /**How old a file is in milliseconds.
//         * This accommodates any system-specific filestamp types
//         * that are not in the same format as currentTimeMillis().
//         * (Though from JDK 1.2 onwards they are all in this form.)
//         * <p>
//         * If the file does not exist, -1 is returned, else the number
//         * returned is zero or positive (ie any timestamps that would
//         * appear to make a file of negative age will appear as zero age).
//         */
//        public static final long howOld(File f)
//            {
//            final long timestamp = f.lastModified();
//            if(timestamp == 0) { return(-1); } // File does not exist.
//            // Assume that timestamps are all in Java format for now.
//            final long age = System.currentTimeMillis() - timestamp;
//            if(age < 0) { return(0); }
//            return(age);
//            }
// --Recycle Bin STOP (13/07/03 16:44)
// --Recycle Bin START (13/07/03 16:43):
//        /**How old a file is in milliseconds.
//         * String interface; more convenient for some purposes.
//         */
//        public static final long howOld(String file)
//            { return(howOld(new File(file))); }
// --Recycle Bin STOP (13/07/03 16:43)


        /**Private lock for runCmd. */
        private static final Object _runCmdLock = new Object();


        /**If _runCmdFix is true, we are working round a JDK 1.2 bugs in Process.
         * In the JDK 1.2 bug, if some more than one thread runs a Process,
         * even if not simultaneously, there is a small change of one
         * spawned by a thread that is not the main thread hanging.
         */
        private static final boolean _runCmdFix = false;

        /**Private object used for signalling within the _runCmdLock.
         * Used as a signal between the runner thread and the runCmd() code
         * that a new command is being posted or the result is
         * ready.
         */
        private static final Object _runCmdSignal = new Object();
        /**Private location for runCmd() where a new command's args are put.
         * Null when no command currently running.
         * <p>
         * When this is set non-null by runCmd() it does a notifyAll() on
         * _runCmdSignal, then waits on the same signal until
         * _runCmdResult becomes non-null.  This is cleared by
         * the runner thread when it accepts and starts to run the
         * new command.
         * <p>
         * Change under the _runCmdSignal lock.
         */
        private static String[] _runCmdArgs;
        /**The result of running a command.
         * Thgis can be Integer for an exit value,
         * or IOException or a RuntimeException.
         * <p>
         * Set non-null by the runner thread (and a signal sent on
         * _runCmdSignal) when the command terminates; cleared
         * by runCmd() on accepting the result.
         * <p>
         * Change under the _runCmdSignal lock.
         */
        private static Object _runCmdResult;
        /**The runner thread for commands.
         * Assumed immortal.
         */
        private static final Thread _cmdRunner = !_runCmdFix ? null :
            (new RunnerThread());
        /**Initialise (as a daemon) and start the runner thread. */
        static
            {
            if(_runCmdFix)
                {
                _cmdRunner.setDaemon(true); // Don't prevent process exiting.
                _cmdRunner.start();
                }
            }

        /**Runs a command, and returns the exit status.
         * <EM>Does not expect the process to use stdin/stdout/stderr;
         * the process may hang if it tries to use them.</EM>
         * <p>
         * Because of possible deadlock/hang problems in JDK1.2,
         * I am serialising all sub-processing runs at the cost of
         * some parallelism.
         */
        public static int runCmd(String cmd[])
            throws IOException
            {
            if((cmd == null) || (cmd.length < 1))
                { throw new IllegalArgumentException(); }
            synchronized(_runCmdLock)
                {
                if(_runCmdFix)
                    {
                    // Use elaborate work-round for JDK 1.2 bug.
                    //
                    // The args/result values should always be null,
                    // and the system quiescent when we get here, but the
                    // runner should be alive.
                    if((_runCmdArgs != null) || (_runCmdResult != null))
                        { throw new Error("INTERNAL ERROR: messed-up state"); }
                    if((_cmdRunner == null) || !_cmdRunner.isAlive())
                        { throw new Error("INTERNAL ERROR: runner missing or dead: " + _cmdRunner); }

                    // Set the args and signal the runner to wake it up.
                    synchronized(_runCmdSignal)
                        {
                        _runCmdArgs = cmd;
                        _runCmdSignal.notifyAll();
                        }

                    // Now sleep until signalled to collect the results
                    // (and set them to null when we get them).
                    Object result = null;
                    synchronized(_runCmdSignal)
                        {
                        while((result = _runCmdResult) == null)
                            {
                            try { _runCmdSignal.wait(10000); }
                            catch(InterruptedException e) { } // Ignore.
                            if(!_cmdRunner.isAlive())
                                { throw new RuntimeException("INTERNAL ERROR: runner has died: " + _cmdRunner); }
                            }
                        _runCmdResult = null;
                        }

                    // Return our (non-null) result.
                    if(result instanceof Integer)
                        {
                        // Exit code.
                        return(((Integer) result).intValue());
                        }
                    if(result instanceof IOException)
                        {
                        // Expected IOException.
                        throw ((IOException) result);
                        }
                    throw new Error("unexpected error: " + result);
                    }

                // The simple route...
                final Runtime r = Runtime.getRuntime();
                final Process p = r.exec(cmd);
                p.getInputStream().close(); // Make sure it does not wait for stdin.
                p.getOutputStream().close();
                p.getErrorStream().close();

                // We should really ensure we read and stdout/stderr generated.

                for( ; ; ) // Loop until we have reaped the child...
                    {
                    try { return(p.waitFor()); }
                    catch(InterruptedException e) { }
                    }
                }
            }

// --Recycle Bin START (13/07/03 16:43):
//        /**Hard-links second file to first if possible.
//         * Throws an exception in case of error.
//         * You might call this from a routine that can fall back to
//         * making a copy.
//         * <P>
//         * At the moment stderr (and stdout) is not captured, but
//         * could be folded into the description of any exception thrown.
//         */
//        public static final void ln(String existingFile, String newFile)
//            throws IOException
//            {
//            final String cmd[] = {"/bin/ln", "-f", existingFile, newFile};
//            if(runCmd(cmd) != 0)
//                { throw new IOException("ln failed"); }
//            }
// --Recycle Bin STOP (13/07/03 16:43)


        /**Moves first file to second, deleting destination if writable.
         * Throws an exception in case of error.
         * <p>
         * Seems to have been causing hangs occasionally; more often
         * when multiple threads are invoking this simultaneously.
         * 2000/01/06 added "-f" to try to ensure no output generated
         * that might induce a hang.
         */
        public static final void mv(final String oldName, final String newName)
            throws IOException
            {
            final String cmd[] = {"/bin/mv", "-f", oldName, newName};
            if(runCmd(cmd) != 0)
                { throw new IOException("mv failed"); }
            }

        /**Recursively removes specified file/directory.
         * Equivalent of UNIX "rm -rf file".
         * <p>
         * Fails if it cannot remove the target,
         * but does not complain if the target does not exist.
         */
        public static void rmRecursively(final File fileOrDir)
            throws IOException
            { rmRecursively(fileOrDir, null); }

        /**Recursively removes specified file/directory.
         * Equivalent of UNIX "rm -rf file".
         * <p>
         * Fails if it cannot remove the target,
         * but does not complain if the target does not exist.
         *
         * @param filter  matches files/dirs to remove;
         *     null to remove all files/dirs from that specified
         */
        public static void rmRecursively(final File fileOrDir,
                                         final FileFilter filter)
            throws IOException
            {
            if(fileOrDir == null)
                { throw new IllegalArgumentException(); }

            // If target does not exist, return quietly.
            if(!fileOrDir.exists())
                { return; }

            // If a directory then recursively deal with any contents first.
            if(fileOrDir.isDirectory())
                {
                final String files[] = fileOrDir.list();
                if(files != null)
                    {
                    for(int i = files.length; --i >= 0; )
                        {
                        final String file = files[i];
                        // Skip UNIX-style "." and ".." entries...
                        if(".".equals(file) || "..".equals(file))
                            { continue; }

                        // Remove/investigate any other entry...
                        rmRecursively(new File(fileOrDir, file), filter);
                        }
                    }
                }

            // Attempt to remove immediate target directly
            // (unless a filter is supplied and this file is not accepted).
            if((filter != null) && !filter.accept(fileOrDir))
                { return; }

            // Delete since no filter,
            // or the file is accepted by the filter supplied...
            fileOrDir.delete();

            // Make sure the target has been removed.
            if(fileOrDir.exists())
                { throw new IOException("cannot remove: " + fileOrDir); }
            }


        /**Makes a file readable and writable only by us (ie private). */
        public static final void chmod600(final String file)
            throws IOException
            {
            final String cmd[] = {"/bin/chmod", "-f", "600", file};
            if(runCmd(cmd) != 0)
                { throw new IOException("chmod failed"); }
            }

        /**Makes a file globally-readable (and writable by us).
         * This makes it readable to a Web server and thus potentially
         * to the world, and makes the image easy to replace atomically
         * with a `mv' operation, or to delete.
         */
        public static final void chmod644(String file)
            throws IOException
            {
            final String cmd[] = {"/bin/chmod", "-f", "644", file};
            if(runCmd(cmd) != 0)
                { throw new IOException("chmod failed"); }
            }

        /**Makes a directory globally-readable/searchable (and writable by us).
         * This makes it readable to a Web server and thus potentially
         * to the world, and makes the image easy to replace atomically
         * with a `mv' operation, or to delete.
         */
        public static final void chmod755(String file)
            throws IOException
            {
            final String cmd[] = {"/bin/chmod", "-f", "755", file};
            if(runCmd(cmd) != 0)
                { throw new IOException("chmod failed"); }
            }

// --Recycle Bin START (13/07/03 16:44):
//        /**Touches a file (brings at least its last-modified stamp up-to-date).
//         * Throws an exception in case of error.
//         * <P>
//         * At the moment stderr (and stdout) is not captured, but
//         * could be folded into the description of any exception thrown.
//         * <P>
//         * The -f flag is needed on SunOS for r/o files, and is ignored
//         * on Solaris, so defensively I include it.
//         */
//        public static final void touch(String file)
//            throws IOException
//            {
//            final String cmd[] = {"/bin/touch", "-f", file};
//            if(runCmd(cmd) != 0)
//                { throw new IOException("touch failed"); }
//            }
// --Recycle Bin STOP (13/07/03 16:44)


// --Recycle Bin START (13/07/03 16:43):
//        /**OutputStream that consumes its output indefinitely.
//         * Use the no-argument constructor.
//         */
//        public static class NullOutputStream extends OutputStream
//            {
//            public void write(int b) { }
//            public void write(byte b[]) { }
//            public void write(byte b[], int o, int l) { } // Absorb, fast...
//            }
// --Recycle Bin STOP (13/07/03 16:43)




//        /**Make timestamp filename from filename it is stamp for.
//         * Usually such a file will contain a hash code for the file
//         * for file(s) being update, represented by this main file,
//         * which is often only updated when all the others have been
//         * successfully built.
//         * <p>
//         * It is assumed that the last component of the existing
//         * file name is a plain file, not directory, component.
//         */
//        public static File makeTimestampFilename(File baseFilename)
//            {
//            return(new File(baseFilename.getParent(),
//                            GlobalParams.F_timestamp_prefix +
//                                baseFilename.getName()));
//            }

        /**Replaces an existing published file with a new one (see 3-arg version).
         * This uses the "pure java" mechanism to do the file replacement.
         *
         * @param quiet  if true then only error messages will be output
         */
        public static boolean replacePublishedFile(String name, byte data[], boolean quiet)
            throws IOException
            { return(replacePublishedFile(name, data, quiet, true)); }

        /**Replaces an existing published file with a new one (see 3-arg version).
         * Is verbose when it replaces the file.
         */
        public static boolean replacePublishedFile(String name, byte data[])
            throws IOException
            { return(replacePublishedFile(name, data, false)); }

        /**Replaces an existing published file with a new one.
         * This replaces (atomically if possible) the existing file (if any)
         * of the given name, ensuring the correct permissions for
         * a file to be published with a Web server (ie basically
         * global read permissions), provided the following
         * conditions are met:
         * <p>
         * <ul>
         * <li>The filename extension is acceptable (not checked yet).
         * <li>The data array is non-null and not zero-length.
         * <li>The content of the data array is different to the file.
         * <li>All the required permissions are available.
         * </ul>
         * <p>
         * If the file is successfully replaced, true is returned.
         * <p>
         * If the file does not need replacing, false is returned.
         * <p>
         * If an error occurs, eg in the input data or during file
         * operations, an IOException is thrown.
         * <p>
         * Even if the file is not replaced, if it is older than
         * GlobalParams.T_genFileTouchTime_ms then we touch it to
         * mark that it is not stale and does not need to be deleted.
         * NOTE: NO LONGER DONE.
         * <p>
         * This routine enforces locking so that only one such
         * operation may be performed at any one time.  This does
         * not avoid the possiblity of externally generated races.
         * <p>
         * The final file, once replaced, will be mode 644, ie globally
         * readable, and writable by us.  (If the final component
         * of the file starts with ".", the mode will be 600.)
         *
         * @param quiet     if true then only error messages will be output
         * @param pureJava  if true this avoids running subprocesses
         *     and does the best it can through the Java File APIs
         */
        public static boolean replacePublishedFile(String name, byte data[],
                                            boolean quiet,
                                            boolean pureJava)
            throws IOException
            {
            if((name == null) || (name.length() == 0))
                { throw new IOException("inappropriate file name"); }
            if((data == null) || (data.length == 0))
                { throw new IOException("inappropriate file content"); }

            final File extant = new File(name);

            // Lock the critical external bits.
            synchronized(rPF_lock)
                {
                // Use a temporary file in the same directory (and thus
                // filesystem) to avoid unexpectedly truncating the file
                // when moving it.
                File tempFile;
                for( ; ; )
                    {
                    tempFile = new File(extant.getParent(),
                        F_tmpPrefix +
                        Long.toString((Rnd.fastRnd.nextLong() >>> 1),
                            Character.MAX_RADIX) /* +
                        "." +
                        extant.getName() */ ); // Avoid making very long names...
                    if(tempFile.exists())
                        {
                        System.err.println("WARNING: FileTools.replacePublishedFile(): "+
                            "temporary file " + tempFile.getPath() +
                            " exists, looping...");
                        continue;
                        }
                    break;
                    }

                // Get extant file's length.
                final long oldLength = extant.length();
                // Should we overwrite it?
                boolean overwrite = (oldLength < 1); // Missing or zero length.

                // If length has changed, we should overwrite.
                if(data.length != oldLength) { overwrite = true; }

//                // If we haven't already decided to overwrite the file,
//                // check if it needs touching anyway to mark it as
//                // not stale.
//                if(!overwrite)
//                    {
//                    final long ageIn_ms = howOld(name);
//                    if(shouldTouch(ageIn_ms, name))
//                        { overwrite = true; }
//                    }

                // Now, if we haven't already decided to overwrite the file,
                // check the content.
                if(!overwrite)
                    {
                    try
                        {
                        final InputStream is = new BufferedInputStream(
                            new FileInputStream(extant));
                        try
                            {
                            final int l = data.length;
                            for(int i = 0; i < l; ++i)
                                {
                                if(data[i] != is.read())
                                    { overwrite = true; break; }
                                }
                            }
                        finally
                            { is.close(); }
                        }
                    catch(FileNotFoundException e) { overwrite = true; }
                    }

                // OK, we don't want to overwrite, so return.
                if(!overwrite) { return(false); }


                // OVERWRITE OLD FILE WITH NEW...

                try {
                    // Write new temp file...
                    // (Allow any IOException to terminate the function.)
                    OutputStream os = new FileOutputStream(tempFile);
                    os.write(data);
                    os.flush();
                    os.close();
                    os = null; // Help GC.
                    if(tempFile.length() != data.length)
                        { new IOException("temp file not written correctly"); }

                    final boolean globalRead = !extant.getName().startsWith(".");

                    if(!pureJava)
                        {
                        // Ensure it has the correct permissions.
                        if(globalRead) { chmod644(tempFile.getPath()); }
                        else           { chmod600(tempFile.getPath()); }
                        }

                    // Warn if target does not have write perms, and
                    // try to add them.
                    final boolean alreadyExists = extant.exists();
                    if(alreadyExists && !extant.canWrite())
                        {
                        System.err.println("FileTools.replacePublishedFile(): "+
                            "WARNING: " + name + " not writable.");
                        if(!pureJava)
                            {
                            if(globalRead) { chmod644(name); }
                            else           { chmod600(name); }
                            }
                        if(!extant.canWrite())
                            {
                            throw new IOException("can't make target writable");
                            }
                        }

                    // (Atomically) move tempFile to extant file.
                    if(!pureJava) { mv(tempFile.getPath(), name); }
                    else /* { tempFile.renameTo(extant); } */
                        {
                        // renameTo() may not be atomic
                        // and we may have to remove the target file first.
                        if(!tempFile.renameTo(extant))
                            {
                            // If the target already exists,
                            // then be prepared to explicitly delete it.
                            if(!alreadyExists || !extant.delete() || !tempFile.renameTo(extant))
                                { throw new IOException("renameTo/update of "+name+" failed"); }
                            if(!quiet) { System.err.println("[WARNING: atomic replacement not possible for: " + name + ": used explicit delete.]"); }
                            }
                        }
                    if(extant.length() != data.length)
                        { new IOException("update of "+name+" failed"); }
                    if(!pureJava)
                        {
                        if(globalRead) { chmod644(name); } // Belt and braces...
                        else           { chmod600(name); }
                        }
                    if(!quiet) { System.err.println("["+(alreadyExists?"Updated":"Created")+" " + name + "]"); }
                    return(true); // All seems OK.
                    }
                finally // Tidy up...
                    {
                    tempFile.delete(); // Remove the temp file.
//                    Thread.yield(); // That was probably expensive; give up the CPU...
                    }
                }

            // Can't get here...
            }
        /**Private lock for replacePublishedFile().
         */
        private static final Object rPF_lock = new Object();

        /**Makes a publicly-readable directory path if not already present.
         * Like File.mkdirs(), but attempts to ensure that any directory
         * component created by this routine is publically readable
         * and searchable, ie at least permissions read and execute
         * for all.  Final permissions will usually be 0755,.
         * <p>
         * Optionally, a new empty index.html file can be created
         * inside any directory that is created as a simple Web security
         * precaution, at least until something is put in its place.
         * <p>
         * If this routine fails it may have succeeded in creating some of the necessary
         * parent directories.
         * <p>
         * This shares a lock with replacePublishedFile().
         * <p>
         * (This routine should maybe be merged with makeHTMLSubDirs(),
         * though the reationship is not trivial.)
         *
         * @return  <code>true</code> if and only if the directory was created,
         *          along with all necessary parent directories; <code>false</code>
         *          otherwise
         */
        public static boolean makePublishingDir(File path)
            throws IOException
            {
            synchronized(rPF_lock)
                {
                if(path.exists()) { return(false); } // Nothing we can do...
                if(path.mkdir())
                    {
                    chmod755(path.getPath());
                    return(true);
                    }
                final String parent = path.getParent();
                if(parent == null) { return(false); }
                if(!makePublishingDir(new File(parent))) { return(false); }
                if(!path.mkdir()) { return(false); }
                chmod755(path.getPath());
                return(true);
                }
            }

        /**Read text file into a String.
         * Reads a line at a time, trimming whitespace off either
         * end and putting in a single new line at the end instead.
         * <p>
         * This treats the file as ISO-8859-1 8-bit data.
         *
         * @exception IOException  in case of trouble
         */
        public static String readTextFile(final File f)
            throws IOException
            {
            final StringBuilder sb = new StringBuilder((int) f.length());
            BufferedReader br = null;
            try {
                br = new BufferedReader(
                        new InputStreamReader(
                            new FileInputStream(f), CoreConsts.FILE_ENCODING_8859_1));
                String inputLine;
                while((inputLine = br.readLine()) != null)
                    { sb.append(inputLine.trim()); sb.append('\n'); }
                }
            finally
                {
                // Try to close the file (possibly provoking an exception)...
                if(br != null) { br.close(); br = null; }
                }
            return(sb.toString());
            }

        /**Given a file, serialises an object to it.
         * This atomically replaces the target file if possible.
         *
         * @param gzipped  if true, the file is written GZIP-compressed
         *     to (usually) save signifcant space
         * @param quiet  if true, only outputs errors
         *
         * @throws IOException  if something bad happens
         */
        public static void serialiseToFile(final Object o,
                                           File filename, boolean gzipped,
                                           boolean quiet)
            throws IOException
            {
            ObjectOutputStream oos = null;
            try {
                // Serialise to a byte[].
                final ByteArrayOutputStream baos = new ByteArrayOutputStream();
                // We may gzip the stream to remove some of the redundancy.
                final java.util.zip.GZIPOutputStream gos = gzipped ?
                    new java.util.zip.GZIPOutputStream(baos) : null;
                oos = new ObjectOutputStream(gzipped ?
                    (OutputStream)gos : (OutputStream)baos);
                oos.writeObject(o);
                oos.flush();
                if(gos != null) { gos.finish(); }
                final byte data[] = baos.toByteArray();

                // Now save to file using pure Java APIs.
                FileTools.replacePublishedFile(filename.getPath(), data, quiet, true);
                }
            finally
                {
                if(oos != null) { oos.close(); } // Free up OS resources.
                }
            }

        /**Given a file, deserialises an object from it.
         * This buffers the input for efficiency.
         *
         * @param gzipped  if true, the file is assumed to be in GZIP
         *     format and the stream is decompressed on the fly
         * @throws IOException  if something bad happens
         */
        public static Object deserialiseFromFile(final File filename, final boolean gzipped)
            throws IOException
            {
            ObjectInputStream ois = null;
            try {
                // Need to reload from disc.
                InputStream is = new BufferedInputStream(new FileInputStream(
                    filename), 8192);
                if(gzipped) { is = new java.util.zip.GZIPInputStream(is, 8192); }
                ois = new ObjectInputStream(is);
                return(ois.readObject());
                }
            catch(final ClassNotFoundException e)
                { throw new IOException("ClassNotFoundException deserialising file ``"+filename+"'': " + e.getMessage()); }
            catch(final Exception e)
                {
                final IOException err = new IOException("unexpected exception deserialising file ``"+filename+"'': " + e.getMessage());
                err.initCause(e);
                throw err;
                }
            finally
                {
                // Ensure that we release the OS file handling resources.
                if(ois != null) { ois.close(); }
                }
            }

        /**Load properties for the given InputStream.
         * In case of any problem reading the stream for properties
         * an IOException is thrown.
         * <p>
         * Also, if the property ``end'' is not defined with the
         * value OK, an EOFException is thrown to indicate the the
         * file was not read completely (eg if we catch it in the
         * middle of a save).  So make sure that files are defined
         * with this property at the end.
         *
         * @throws IOException  in case of general I/O problems
         * @throws EOFException  in case of missing <sample>end=OK</sample> value
         */
        public static Properties loadProperties(final InputStream is)
            throws IOException
            {
            final Properties props = new Properties();
            props.load(is);
            if(!"OK".equals(props.getProperty("end")))
                { throw new EOFException("missing end=OK at end of data: possibly corrupt or not fully saved?"); }
            return(props);
            }


        /**Typical size in bytes/characters of (English, ASCII) text below which deflate/zlib compression unlikely to be very effective.
         * This is for typical (English, ASCII) text given the overheads of
         * the compression methods etc.
         * <p>
         * Avoiding trying to compress things smaller than this
         * may save lots of CPU time without too much memory/space cost,
         * and may allow better compression of aggregates.
         */
        public static final int TYPICAL_DEFLATE_MIN_TEXT_SIZE_COMPRESSABLE = 256;

        /**Compress data (maximally deflate without zlib/gzip headers and footers) from byte[] to byte[]; never null.
         * This means that there is no checksum,
         * so any integrity checking required had better be done elsewhere.
         * <p>
         * This may increase the size of the data in some cases,
         * especially where the input array is short.
         * <p>
         * This routine will not alter the content of the input array.
         */
        public static byte[] compressDeflatableData(byte in[], int off, int len)
            {
            try
                {
                // Data expansion is unusual except for short inputs.
                // We assume that normally we get some compression
                // and size the working buffer accordingly.
                final ByteArrayOutputStream baos =
                    new ByteArrayOutputStream((3*in.length)/4 + 32);
                final DefOutputStream cos =
                    new DefOutputStream(baos);
                try
                    {
                    // Write uncompressed data to stream.
                    cos.write(in, off, len);
                    // Force everything out...
                    cos.finish();
                    cos.flush();
                    return(baos.toByteArray());
                    }
                finally
                    {
                    cos.close(); // Free resources, including non-Java resources.
                    }
                }
            catch(final IOException e)
                {
                // Should never happen...
                throw new Error("unexpected internal error", e);
                }
            }

        /**Decompress deflated data (without zlib/gzip headers and footers) from byte[] to byte[].
         * This means that there is no checksum,
         * so any integrity checking required had better be done elsewhere.
         * <p>
         * This routine will not alter the content of the input array.
         *
         * @throws IOException  in case of difficulty, such as corrupt data
         */
        public static byte[] decompressDeflatedData(final byte in[])
            throws IOException
            {
            // Uncompress from the input stream...
            final InputStream cis =
                new DefInputStream(
                    new ByteArrayInputStream(in));
            try
                {
                // Write to a new output stream in memory.
                // Assume that a reasonable compression factor
                // will have been obtained when sizing the
                // internal buffer.
                final int guessedExpandedSize = 1 + in.length * 2; // Always > 0.
                final ByteArrayOutputStream boas = new ByteArrayOutputStream(guessedExpandedSize);
                // Copy from input to output until EOF,
                // using a small buffer for efficiency.
                final byte buf[] = new byte[Math.min(512, guessedExpandedSize)];
                for( ; ; )
                    {
                    final int n = cis.read(buf);
                    if(n == -1) { break; } // EOF.
                    boas.write(buf, 0, n);
                    }
                // Use the uncompressed data!
                final byte uncompressedData[] = boas.toByteArray();
                boas.close();
                return(uncompressedData);
                }
            finally { cis.close(); /* Release Java and non-Java resources. */ }
            }

        /**Rounds up a byte length to the next block size.
         * Return is no smaller than input value.
         * <p>
         * Input value must be non-negative.
         * <p>
         * An input value of 0 results in a return value of 0.
         */
        public static long roundUpToFSBlockSize(long length)
            {
            assert length >= 0;
            return(((length + FS_EST_BLOCK_SIZE_BYTES - 1) / FS_EST_BLOCK_SIZE_BYTES) * FS_EST_BLOCK_SIZE_BYTES);
            }

        /**Thread to encapsulate execution of command. */
        private static final class RunnerThread extends Thread {
            public RunnerThread() {
                super("FileTools runCmd runner");
            }

            public final void run()
                {
//System.err.println("FileTools runCmd runner starting...");
                for( ; ; )
                    {
                    try {
                        // Get new command...
                        String cmd[] = null;
                        synchronized(_runCmdSignal)
                            {
                            while((cmd = _runCmdArgs) == null)
                                {
                                try { _runCmdSignal.wait(); }
                                catch(InterruptedException e) { } // Ignore.
                                }
                            _runCmdArgs = null;
                            }

//System.err.println("FileTools runner go... " + cmd[0]);

                        // Now run the command...
                        Object result = null;
                        try {
                            final Runtime r = Runtime.getRuntime();
                            final Process p = r.exec(cmd);
                            p.getInputStream().close(); // Make sure it does not wait for stdin.
                            p.getOutputStream().close();
                            p.getErrorStream().close();

                            // We should really ensure we read and stdout/stderr generated.

                            for( ; ; ) // Loop until we have reaped the child...
                                {
                                try {
                                    result = new Integer(p.waitFor());
                                    break;
                                    }
                                catch(InterruptedException e) { }
                                }
                            }
                        catch(IOException e) { result = e; }
                        catch(RuntimeException e) { result = e; }

                        // Now return the result, which we ensure is non-null.
                        if(result == null) { result = new Error("INTERNAL ERROR"); }
                        synchronized(_runCmdSignal)
                            {
                            _runCmdResult = result;
                            _runCmdSignal.notifyAll();
                            }
//System.err.println("FileTools runner done.");
                        }
                    catch(Throwable t)
                        {
                        System.err.println("FileTools cmd runner caught: " + t.getMessage());
                        t.printStackTrace();
                        }
                    }
                }
        }
    }
