/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import LZMA.LzmaInputStream;
import ORG.hd.d.IsDebug;
import SevenZip.CoderPropID;
import SevenZip.InvalidParamException;

import org.apache.tools.bzip2.CBZip2InputStream;
import org.apache.tools.bzip2.CBZip2OutputStream;
import org.hd.d.pg2k.svrCore.MemoryTools.Internable;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;

import java.io.*;
import java.lang.ref.WeakReference;
import java.security.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;

/**Some general utility methods of use throughout the application.
 */
public final class GenUtils
    {
    /**Prevent construction of an instance. */
    private GenUtils() { }


    /**Simple logger instance to log to System.out. */
    public static final SimpleLoggerIF systemOutLogger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            { System.out.println(message); }
        };

    /**Simple logger instance to log to System.err. */
    public static final SimpleLoggerIF systemErrLogger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            { System.err.println(message); }
        };

    /**Simple logger instance to throw away log output. */
    public static final SimpleLoggerIF nullLogger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            { /* Discard output. */ }
        };

    /**Get instance of our standard good/secure digester (a new instance each time).
     * This is the SHA1 digest.
     * <p>
     * Instances returned by this call must not be shared between threads.
     */
    public static MessageDigest getStandardDigest()
        {
        try { return(MessageDigest.getInstance(CoreConsts.HASH_SHA1 /*"SHA"*/)); }
        catch(final NoSuchAlgorithmException e) // Should never happen...
            { throw new Error("could not find "+CoreConsts.HASH_SHA1+" digester!", e); }
        }

    /**Compute the approximate rounded-down log-base-2 of the input.
     * The input must be non-negative; input zero returns -1.
     * <p>
     * In effect this returns a 1-significant-bit answer,
     * which is useful for "clumping" values that have a large dynamic range.
     * <p>
     * @return floor(log<sub>2<sub>(v))
     */
    public static int log2Approx(long v)
        {
        if(v < 0) { throw new IllegalArgumentException(); }

        int l;
        for(l = -1; v != 0; ++l)
            { v >>>= 1; }

        return(l);
        }

    /**Get section title (as HTML).
     * This attempts to get or compute a section title given the
     * top-level section directory name from an exhibit (such as "brain_scan").
     * <p>
     * The fall-back algorithm is for this to convert all non-alphanumeric
     * characters in the name to spaces and capitalise the first letter of
     * the title.
     * <p>
     * However, if this routine is passed a valid locale bean, this attempts to
     * look up the directory name (suitably prefixed and suffixed)
     * in the catalogue to find a sugared and/or internationalised version.
     */
    public static String computeSectionTitle(final String sectionDir,
                                             final LocaleBeanBase lb)
        {
        if(lb != null)
            {
            final String key = UploaderConsts.SECTION_TITLE_PROPNAME_PREFIX + sectionDir;
            final String i18nedTitle = lb.getLocalisedMessage(key);
            if((i18nedTitle != null) &&
               (i18nedTitle.length() != 0) &&
               !i18nedTitle.equals(key)) // Usual graceful failure mode of LocaleBean.
                { return(i18nedTitle); } // Got i18n version!
            // Failed: fall through to resort to backup algorith...
            }

        // Fall-back algorithm: compute name from section title...
        final StringBuilder simpleResult = new StringBuilder(sectionDir.length());
        for(int i = 0; i < sectionDir.length(); ++i)
            {
            final char ch = sectionDir.charAt(i);
            if(!Character.isLetterOrDigit(ch)) // All non-alphanumeric to space.
                { simpleResult.append(' '); }
            else if(i == 0) // Capitalise first character.
                { simpleResult.append(Character.toTitleCase(ch)); }
            else // Leave as is.
                { simpleResult.append(ch); }
            }
        return(simpleResult.toString());
        }


    /**Maximum capacity for _gLTD_cache; strictly positive.
     * The larger this capacity the more memory that we may consume
     * but the greater the chance we have of making a cache hit.
     * <p>
     * If this cache is to help significantly with building the by-word index
     * then this should be easily large enough to handle the fact
     * that keys will at best be looked up in sorted order on part of the
     * exhibit name, and that may not correspond exactly with the way
     * that AKA/desc data is organised.
     * <p>
     * If this is to help significantly with normal catalogue-page lookup
     * then this should allow for the typical distribution of page views
     * and of locales used by visitors.
     * <p>
     * This suggests a size of at least several hundred entries.
     */
    private static final int MIN_gLTD_cache_SIZE = 1001 +
                                               Rnd.fastRnd.nextInt(300);
    /**An LRU cache for lookups in getLocalisedTreeDesc() and private to it; never null.
     * This is a map from a composite key based on the data hash,
     * exhibit name, lookup type and locale to the result of the lookup.
     * <p>
     * The result text can be extracted with toString().
     * <p>
     * (The data hash that we use from the AEP is from just the treedesc
     * data so that changes in other parts of the AEP do not invalidate
     * cache entries.  This is useful because the treedesc data
     * usually changes much more slowly than the entire AEP
     * (and the AEP longHash does not always capture the treedesc state)
     * and because we can use a shorter hash for quicker key construction.)
     * <p>
     * The keys are essentially unique and this not intern()ed,
     * but the result texts may not be unique,
     * and thus are intern()ed to economise on memory.
     * <p>
     * This cache can in principle support multiple AEPs simultaneously
     * (assuming that their hashes do not collide)
     * and thus will cope with an update to a new AEP in the Web server.
     */
    private static final MemoryTools.SimpleLRUMapAutoSizeForHitRate<String,Object> _gLTD_cache =
        new MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Object>(0.01f, MIN_gLTD_cache_SIZE, 101*MIN_gLTD_cache_SIZE, 0.7f);
    /**Prefix of descriptive-text keys in tree-text i18n bundles. */
    public static final String TREE_TEXT_PREFIX_DESCRIPTION = "desc.";
    /**Prefix of Also-Known-As-keyword-list keys in tree-text i18n bundles. */
    public static final String TREE_TEXT_PREFIX_AKA = "aka.";
    /**Default maximum acceptable gap between aka/desc entries for search to continue; non-negative. */
    public static final int DEFAULT_MAX_DESCAKA_GAP = 4;

    /**Extract description tree information as a list; returns "" if none; never null.
     * Returns an HTML fragment,
     * zero or more <li>key (aka keywords) = text</li> values in order,
     * with localised i18n text extracted about the exhibit.
     * <p>
     * This will stop searching after a certain number of failures,
     * ie nodes where no description/AKA text is available.
     * This is to reduce search times on long exhibit names
     * or where there is no suitable text at all.
     * <p>
     * Behaviour is undefined if the exhibitName is not syntactically valid.
     * @param aep complete exhibit metadata/properties; never null
     * @param exhibitName  syntactically-valid case-insensitive-main-component
     *     (not necessarily extant) exhibit name;
     *     never null
     * @param getDesc  if true, include descriptive text from the tree
     * @param getAKA  if true, include "Also-Known-As" sysnonymn text from the tree
     * @param akaLinks  if true, insert hyperlinks from synonyms
     *     to appropriate places in the catalogue tree where possible
     * @param mostSpecificOnly  if true, show only the most-specific text,
     */
    public static final String getLocalisedTreeDesc(final AllExhibitProperties aep,
                                                    final String exhibitName,
                                                    final LocaleBeanBase localeBean,
                                                    final boolean getDesc,
                                                    final boolean getAKA,
                                                    final boolean akaLinks,
                                                    final boolean mostSpecificOnly)
        {
        assert(ExhibitName.validNameSyntaxBasic(exhibitName));

        if((aep == null) || (localeBean == null))
            { throw new IllegalArgumentException(); }

        if(!getDesc && !getAKA)
            { throw new IllegalArgumentException("no point in asking for neither desc nor AKA"); }

        if(!getAKA && akaLinks)
            { throw new IllegalArgumentException("no point in asking for AKA links with no AKA"); }

        // All attribute words (to be stripped from names before lookup).
        final SortedSet<String> attrWordsSortedSet = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();
        // We lower-case the main component to:
        //   * help reduce the cache key space and increase the cache hit rate,
        //   * reduce work later in converting keys to lower-case.
        final String mainWordsComponentLC = ExhibitName.getMainWordsComponent(exhibitName, attrWordsSortedSet).toLowerCase();
        // The category component is expected always to be lower-case.
//        final String catLC = ExhibitName.getCategoryComponent(exhibitName).toLowerCase();
        final String catLC = ExhibitName.getCategoryComponent(exhibitName);
if(IsDebug.isDebug) { assert(catLC.equals(catLC.toLowerCase())); }

//final long kts = System.nanoTime();
        // Compute cache lookup key.
        // Try to do the least work (fewest appends) for the most common cases.
        final StringBuilder cacheKeySB = new StringBuilder(exhibitName.length());
        // We include a hash over just our treedesc keys and values,
        // so we should remain immune to changes elsewhere in the AEP.
        // Fixed-char-count to avoid ambiguity without needing a trailing separator.
        final int dataHash = aep.epgi.getTreedescHash();
        cacheKeySB.append((char) dataHash).append((char) (dataHash >>> 16));
        // Append markers for "uncommon" flag values.
        if(!getAKA) { cacheKeySB.append('A'); }
        if(getDesc) { cacheKeySB.append('d'); }
        if(akaLinks) { cacheKeySB.append('l'); }
        if(mostSpecificOnly) { cacheKeySB.append('s'); }
        cacheKeySB.append('|');
        // Don't include the default locale in the key so as to save time...
        final Locale locale = localeBean.getLocale();
        if(!locale.equals(I18NTools.DEFAULT_SYSTEM_LOCALE))
            {
            cacheKeySB.append(locale.toString());
            cacheKeySB.append('|');
            }
        cacheKeySB.append(catLC);
        cacheKeySB.append('/');
        cacheKeySB.append(mainWordsComponentLC);
        final String cacheKey = cacheKeySB.toString();

        // If we have an entry in the cache then return it immediately...
        final Object fromCache = _gLTD_cache.get(cacheKey);
//final long kte = System.nanoTime(); System.out.println("[AKA key gen/use time: "+(kte-kts)+"ns.]");
        if(fromCache != null) { return(fromCache.toString()); }

        // Compute the result...

//final long rts = System.nanoTime();
        // Limits maximum (potentially fruitless) depth of search.
        final int maxGap = DEFAULT_MAX_DESCAKA_GAP;
        int remainingGap = maxGap - 1; // Allow for section-dir component.

        // Result is collected in this.
        final StringBuilder result = new StringBuilder();

        // Build up basic common key word by word...
        final StringBuilder basicKey = new StringBuilder(exhibitName.length());
        basicKey.append(catLC);
        final int afterSection = basicKey.length() + 1;

        final ExhibitPropsGlobalImmutable epgi = aep.epgi;
        final Locale l = localeBean.getLocale();

        final Enumeration wEn = (new StringTokenizer(mainWordsComponentLC, ExhibitName.WORD_SEPS));
        while(wEn.hasMoreElements() && (--remainingGap >= 0))
            {
            // Capture the "old"/previous basic key length.
            // We do this so that we can easily construct the previous key
            // or derivatives of it if we want to later.
            final int oldBkLen = basicKey.length();

            // Extend basic key with new word.
            final String word = (String) wEn.nextElement();
            basicKey.append('.').append(word);
            final String bk = basicKey.toString();

            final String descKeyS = TREE_TEXT_PREFIX_DESCRIPTION + bk;
            final String desc = epgi.getLocalisedTreeDescMessage(descKeyS, l);
            final String akaKeyS = TREE_TEXT_PREFIX_AKA + bk;
            final String aka = epgi.getLocalisedTreeDescMessage(akaKeyS, l);
            // If we found some real description or "AKA", then capture it.
            final boolean gotDescText = getDesc && !descKeyS.equals(desc);
            final boolean gotAKAText = getAKA && !akaKeyS.equals(aka);
            if(gotDescText || gotAKAText)
                {
                // Clear any earlier text we had collected
                // if we want to keep only the most-specific text available.
                if(mostSpecificOnly)
                    { result.setLength(0); }

                // We found something, so can reset our counter.
                remainingGap = maxGap;

                // Make a printable version of the key.
                // Drop the section category and
                // show the words of the partial key in order,
                // space-separated.
                final String keyP =
                    bk.substring(afterSection).replace('.', ' ');
                if(!mostSpecificOnly)
                    { result.append("<li><i>").append(keyP); }
                if(gotAKAText)
                    {
                    if(!mostSpecificOnly)
                        { result.append(" = "); }

                    // If not doing links then put in text "as-is".
                    if(!akaLinks)
                        { result.append(aka); }
                    else
                        {
                        // If doing AKA hypertext links
                        // then go through the synonyms one-by-one
                        // looking for related information
                        // and linking to it if any found.
                        final StringTokenizer st = new StringTokenizer(aka, ",");
                        while(st.hasMoreTokens())
                            {
                            final String s = st.nextToken().trim();

                            // Could the current synonym (AKA)
                            // be a single valid "peer" word
                            // that has its own (interesting) AKA/desc text
                            // and related exhibits?
                            final boolean validWord = ExhibitName.validWord(s);
                            boolean doneLink = false; // Hyperlinked if true.
                            if(validWord)
                                {
                                // Look for peer use of AKA term...
                                // (We don't check for peer description here,
                                // as we assume that possible reciprocal
                                // AKA "referencing" may be a better filter.)
                                final String peerBk =
                                    basicKey.substring(0, oldBkLen) +
                                    '.' + s.toLowerCase();
                                final String peerAKAKey =
                                    TREE_TEXT_PREFIX_AKA + peerBk;
                                final String peerAka = epgi.getLocalisedTreeDescMessage(peerAKAKey, l);
                                if(!peerAKAKey.equals(peerAka))
                                    {
                                    // OK, have found a peer key, so link.
                                    result.append("<a href=\"/").
                                        append(peerBk.replace('.', '/')).
                                        append("/\">");
                                    doneLink = true;
if(IsDebug.isDebug) { System.err.println("[Inserted AKA hyperlink from "+bk+" to "+peerBk+".]"); }
                                    }
                                }

                            result.append(s);

                            // Close hyperlink if one was made/opened.
                            if(doneLink) { result.append("</a>"); }

                            // If not last on the list
                            // then reinsert the usual separator.
                            if(st.hasMoreTokens())
                                { result.append(", "); }
                            }
                        }
                    }
                if(!mostSpecificOnly)
                    { result.append("</i>"); }
                if(gotDescText)
                    { result.append(" = ").append(desc); }
                if(!mostSpecificOnly)
                    { result.append("</li>"); }
                }
            }

        // Cache and return a negative ("") result.
        if(result.length() == 0)
            {
            _gLTD_cache.put(cacheKey, "");
            return("");
            }

        // Cache the positive result, compressed and intern()ed if possible.
        String s = result.toString();
//final long rte = System.nanoTime(); System.out.println("[AKA compute time: "+(rte-rts)+"ns.]");

        try { _gLTD_cache.put(cacheKey, MemoryTools.intern(Compact7BitString.convertToCompact7BitString(s, null))); }
        catch(final IllegalArgumentException e) { _gLTD_cache.put(cacheKey, s = MemoryTools.intern(s)); }

        // Return the (non-null) result.
        return(s);
        }

    /**GenProps generic property name used to flag sensitive name substrings.
     * This value, if set, is treated as a space-separated list
     * of case-sensitive substrings of "sensitive" exhibit names.
     * <p>
     * This is the lookup key, ie excluding the "gen." prefix.
     */
    public static final String GP_SENSITIVE_NAME_SUBSTRS_KEY = "sensitiveNames";

    /**Returns true iff the full/short exhibit name or URI supplied is "sensitive" in some way.
     * Can be used to suppress "promotion" of (or advertising against) an exhibit,
     * or other accidentally-tasteless activity.
     * <p>
     * Simply matches (case-insensitively) potentially-problematic substrings,
     * so has to be used with care to avoid false/silly matches,
     * the main defense against which is to use sufficiently long and specific substrings.
     * <p>
     * TODO: consider switching to regex expression and/or cached lookup
     */
    public static boolean isSensitive(final String nameOrURI,
                                      final AllExhibitProperties aep,
                                      final GenProps gp)
        {
//if(IsDebug.isDebug) { System.err.println("isSensitive("+nameOrURI+")"); }
        if(nameOrURI == null) { return(false); }

        final String substrs = gp.getGen().get(GP_SENSITIVE_NAME_SUBSTRS_KEY);
        // Return immediately if no sensitive strings flagged.
        if(substrs == null) { return(false); }

        final StringTokenizer st = new StringTokenizer(substrs);
        while(st.hasMoreTokens())
            {
            // If this substring matches then this name is sensitive.
            if(nameOrURI.indexOf(st.nextToken()) != -1)
                { return(true); }
            }

        // All clear.
        return(false);
        }

    /**Create a perf monitor attached to the given thread, filling in the given map.
     * This adds to an existing map,
     * and perf multiple monitors can update the same map.
     * <p>
     * This thread can be interrupt()ed to stop it collecting,
     * and it will also stop when the thread being monitored exits.
     * <p>
     * This runs as a high-priority thread sampling at a high rate (~1ms).
     * 
     * @param toObserve  the thread to be sampled; never null
     * @param counts  the map from code sample site to sample count; never null
     * @param prefix  if non-null then the lowest method on the stack
     *     whose class starts with this prefix is logged,
     *     else the lowest method is logged
     * @param stopBy  the time by which we should stop sampling
     * @param sampleTimeMS TODO
     */
    public static Thread startThreadPerfMonitor(final Thread toObserve,
                                                final ConcurrentHashMap<StackTraceElement, AtomicInteger> counts,
                                                final String prefix,
                                                final long stopBy,
                                                final int sampleTimeMS)
        {
        if((null == toObserve) || (null == counts) || (sampleTimeMS < 1))
            { throw new IllegalArgumentException(); }
        
        final Thread t = new Thread("perf monitor thread for " + toObserve){
            @Override public final void run()
                { 
                while(toObserve.isAlive() &&
                      !isInterrupted() &&
                      (System.currentTimeMillis() < stopBy))
                    {
                    final Thread.State state = toObserve.getState();
                    // Omit sample for threads in probably-uninteresting states...
                    if((state == Thread.State.NEW) ||
                       (state == Thread.State.TERMINATED))
                        { continue; }
                    // Get stack trace.
                    final StackTraceElement[] stackTraceElements = toObserve.getStackTrace();
                    // Skip threads with an empty or strange stack.
                    StackTraceElement top;
                    if((stackTraceElements == null) ||
                       (stackTraceElements.length < 1) ||
                       (null == (top = stackTraceElements[0])))
                        { continue; }
                    
                    if(null != prefix)
                        {
                        // Look for a better match than top-of-stack.
                        for(int i = 0; i < stackTraceElements.length; ++i)
                            {
                            final StackTraceElement ste = stackTraceElements[i];
                            if(null == ste) { continue; }
                            final String cn = ste.getClassName();
                            final boolean ourCode = cn.startsWith(prefix);
                            if(ourCode)
                                {
                                // We've found a better match,
                                // so use use it and stop searching.
                                top = ste;
                                break;
                                }
                            }
                        }

                    // Get existing counter for this execution site,
                    // or (atomically) create a new zero one.
                    AtomicInteger count;
                    while(null == (count = counts.get(top)))
                        { counts.putIfAbsent(top, new AtomicInteger()); }

                    // Increment counter.
                    count.incrementAndGet();
                    
                    try { Thread.sleep(sampleTimeMS); }
                    catch(final InterruptedException e) { return; /* Terminate on interrupt. */ }
                    }
                }
            };
        t.setDaemon(true);
        t.setPriority(Thread.MAX_PRIORITY);
        t.start();
        return(t);
        }
    
    /**Stop a monitor thread and dump its samples.
     * 
     * @param monitorThread  thread being profiled; never null
     * @param title   title to display; can be null
     * @param counts  map from sample site to counts at that site; never null
     * @param maxToShow  max (top) entries to display; strictly positive
     * @param logger  if null no dump is generated, else is log sink
     * 
     * @throws InterruptedException  if this thread is interrupted
     */
    public static void stopPerfMonitorandDumpSamples(
                    final Thread monitorThread,
                    final String title,
                    final Map<StackTraceElement, AtomicInteger> counts,
                    final int maxToShow,
                    final SimpleLoggerIF logger)
        throws InterruptedException
        {
        // Stop the monitor thread and wait a little while for it to exit.
        while(monitorThread.isAlive())
            {
            monitorThread.interrupt();
            monitorThread.join(100);
            }
        
        // If no dump is required then return now.
        if(null == logger) { return; }
 
        // Do the dump...
        dumpPerfSamples(title, counts, maxToShow, logger);
        }

    /**Dump a set of performance-monitoring site samples.
     * We take a snapshot of the current profile stats when we start,
     * so that it is safe to dump from a map that is still being updated
     * (though keys/entries must not be removed while we are working)
     * and an iterator over the counts map and get() must be thread-safe.
     * 
     * @param title   title to display; can be null
     * @param counts  map from sample site to counts at that site; never null
     * @param maxToShow  max (top) entries to display; strictly positive
     * @param logger  destination of profile results; never null
     */
    public static void dumpPerfSamples(final String title,
                                       final Map<StackTraceElement, AtomicInteger> counts,
                                       final int maxToShow,
                                       final SimpleLoggerIF logger)
        {
        // Take a snapshot of the data.
        final HashMap<StackTraceElement, Integer> snapshot = new HashMap<StackTraceElement, Integer>(counts.size());
        
        // We have to assume that the iterator will be safe.
        int totalHits = 0;
        for(final StackTraceElement ste : counts.keySet())
            {
            final Integer count = new Integer(counts.get(ste).get());
            totalHits += count;
            snapshot.put(ste, count);
            }
        
        // Prepare a set of samples sorted by count (then sample).
        // We only even insert the best samples.
        final SortedSet<StackTraceElement> best = new TreeSet<StackTraceElement>(new Comparator<StackTraceElement>()
            {
            /**We sort highest-first by the count,
             * then the StackTraceElement content to break ties.
             */
            public int compare(final StackTraceElement se0,
                               final StackTraceElement se1)
                {
                final int c0 = snapshot.get(se0).intValue();
                final int c1 = snapshot.get(se1).intValue();
                if(c0 > c1) { return(-1); /* Correct order. */ }
                if(c0 < c1) { return(+1); /* Wrong order. */ }
        
                // Break ties (expensively)
                // with the full String form of the sample data.
                return(se0.toString().compareTo(se1.toString()));
                }
            });
        
        // Crudely add samples.
        // TODO: make this more efficient.
        best.addAll(snapshot.keySet());

        if(title != null) { logger.log(title); }
        logger.log("Sampled sites: "+(counts.size())+"; max dumped: "+maxToShow+"");
        logger.log("COUNT\tHit%\tSite");

        // We build each result line in here.
        final StringBuilder line = new StringBuilder(256);

        // We build the percentage figure in here.
        final StringBuilder p = new StringBuilder(3);

        // Dump top (highest-count) samples.
        final Iterator<StackTraceElement> it = best.iterator();
        for(int i = maxToShow; (--i >= 0) && it.hasNext(); )
            {
            final StackTraceElement ste = it.next();
            line.setLength(0);
            final int count = snapshot.get(ste).intValue();
            line.append(count);

            line.append('\t');
            p.setLength(0);
            p.append((100 * count) / totalHits);
            while(p.length() < 3) { p.insert(0, ' '); }
            line.append(p).append('%');

            line.append('\t');
            line.append(ste.toString());
            logger.log(line.toString());
            }
        }

    /**Validate internal Gallery timestamp; returns true if sensible, false otherwise.
     * Such timestamps should not be before the Gallery was created,
     * nor much in the future (allowing for some clock skew between consenting participants).
     */
    public static final boolean isValidGalleryTimestamp(final long timestamp)
        {
        if(timestamp < CoreConsts.GALLERY_EPOC_START) { return(false); }
        if(timestamp > System.currentTimeMillis() + 2*CoreConsts.MAX_PEER_CLOCK_SKEW_MS) { return(false); }
        return(true); // Seems OK.
        }


    /**Maximum compression level currently supported by compressData(); never null. */
    public static final CompressionLevel MAX_SUPPORTED_COMPRESSION_LEVEL = CompressionLevel.BZIP2;
    
    /**Compress the supplied binary data as well as possible up to the given compression level; never null.
     * Tries all the supported compression methods up to the specified level
     * returning whichever results in the smallest result.
     * <p>
     * (Resource constraints may prevent some of the compression types being applied,
     * usually the highest levels.)
     * <p>
     * This does NOT assume that higher compression levels always compress given data better,
     * not does it assume a lower bound on data size worth compressing;
     * this tries all the possibilities exhaustively if it can.
     * <p>
     * This is likely to be very memory- and CPU- intensive; use with care.
     * <p>
     * This may attempt compressions in parallel on a multi-CPU machine,
     * though each compression is likely to thrash the cache
     * so this may or may not be good for compression and system performance.
     * <p>
     * This routine does not alter the content of the input array.
     * <p>
     * TODO: WRAP AS MEMORY-INTENSIVE OP FOR MORE EXTREME COMPRESSION METHODS
     * 
     * @param in  the data to be compressed; never null
     * @param maxLevel  the maximum level at which to attempt to compress data,
     *     using a value greater than MAX_SUPPORTED_COMPRESSION_LEVEL is allowed but ineffective;
     *     never null
     *     
     * @return non-null data compressed as well as possible;
     *     if compression was not possible the level is NONE
     *     and the input array is returned in the result as-is
     */
    public static Tuple.Pair<CompressionLevel,byte[]> compressData(final byte[] in, CompressionLevel maxLevel)
        throws InterruptedException
        {
        if((null == in) || (null == maxLevel))
            { throw new IllegalArgumentException(); }

        // Coerce maxLevel to reflect what we actually have implemented.
        if(maxLevel.getLevel() > MAX_SUPPORTED_COMPRESSION_LEVEL.getLevel())
            { maxLevel = MAX_SUPPORTED_COMPRESSION_LEVEL; }

        // Default is not to compress at all, returning the input data as-is.
        final AtomicReference<Tuple.Pair<CompressionLevel,byte[]>> result =
            new AtomicReference<Tuple.Pair<CompressionLevel,byte[]>>(new Tuple.Pair<CompressionLevel,byte[]>(CompressionLevel.NONE, in));

        // Handle zero-length (trivial) input specially.
        if(in.length == 0) { return(result.get()); }
        
        // Set of tasks/results.
        final Queue<Future<?>> tasks = new ArrayBlockingQueue<Future<?>>(CompressionLevel.values().length);

        // Iterate over all known compression levels,
        // possibly processing them concurrently,
        // trying compression at the given level.
        for(final CompressionLevel cl : CompressionLevel.values())
            {
            // Skip unusable compression levels.
            if((cl == CompressionLevel.NONE) || (cl.getLevel() > maxLevel.getLevel()))
                { continue; }

            // Submit this compression attempt to run as a CPU-intensive task.
            tasks.add(ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    final byte[] r; // Result for this level. 
                    switch(cl)
                        {
                        case ZIP:
                            {
                            r = FileTools.compressDeflatableData(in, 0, in.length);
                            break;
                            }

                        case BZIP2:
                            {
                            try
                                {
                                // Assume that we get *some* compression in sizing the output buffer...
                                final ByteArrayOutputStream baos = new ByteArrayOutputStream(512+(in.length/2));
                                // Use maximum BZIP2 compression.
                                final CBZip2OutputStream bz2os = new CBZip2OutputStream(baos, 9);
                                bz2os.write(in);
                                bz2os.flush();
                                bz2os.close();
                                r = baos.toByteArray();
                                break;
                                }
                            catch(final IOException e)
                                {
                                e.printStackTrace();
                                throw new RuntimeException(e);
                                }
                            }

                        case LZMA:
                            {
                            // We do not pick the most extreme settings
                            // since the compression gain is likely to be small
                            // and we may not have the CPU time and memory to spare.

                            // Setting (as per defaults).
                            final int dictionary = 1 << 22; // 4MB decode dictionary.
                            final int posStateBits = 2;
                            final int litContextBits = 3; // for normal files
                            // int litContextBits = 0; // for 32-bit data
                            final int litPosBits = 0;
                            // int litPosBits = 2; // for 32-bit data
                            final int algorithm = 2; // 2 ==> "max" mode.
                            final int numFastBytes = 128;
//                            final String smf ="bt4";
                            final int mf = SevenZip.Compression.LZMA.Encoder.EMatchFinderType_BT4;

                            final int [] propIDs =
                                {
                                CoderPropID.DictionarySize,
                                CoderPropID.PosStateBits,
                                CoderPropID.LitContextBits,
                                CoderPropID.LitPosBits,
                                CoderPropID.Algorithm,
                                CoderPropID.NumFastBytes,
                                CoderPropID.MatchFinder,
                                CoderPropID.EndMarker
                                };

                            final int[] properties =
                                {
                                dictionary,
                                posStateBits,
                                litContextBits,
                                litPosBits,
                                algorithm,
                                numFastBytes,
                                mf,
                                0 // No end-of-stream marker.
                                };

                            try
                                {
                                final ByteArrayOutputStream baos = new ByteArrayOutputStream(256 + in.length/4);
                                SevenZip.Compression.LZMA.Encoder encoder = new SevenZip.Compression.LZMA.Encoder();
                                encoder.SetCoderProperties(propIDs, properties);
                                encoder.WriteCoderProperties(baos);
                                for(int i = 0; i < 8; i++)
                                    { baos.write((byte)(in.length >> (8 * i))); }
                                encoder.Code(new ByteArrayInputStream(in), baos, null);
                                r = baos.toByteArray();
                                }
                            catch(final InvalidParamException e)
                                {
                                e.printStackTrace();
                                throw new Error(e); // Should not happen.
                                }
                            catch(final IOException e)
                                {
                                e.printStackTrace();
                                throw new RuntimeException(e);
                                }
                            break;
                            }

                        default: return; // Skip unsupported level.
                        }

                    assert(r != null); // Compression should not generate null result.
                    assert(r.length > 0); // Compression should not generate zero-length result.

                    // If what we have computed looks to be the best result so far
                    // then atomically replace the result.
                    // This frees up resources ASAP (eg from sub-optimal results).
                    final Tuple.Pair<CompressionLevel, byte[]> putativeResult = new Tuple.Pair<CompressionLevel,byte[]>(cl, r);
                    for( ; ; )
                        {
                        final Tuple.Pair<CompressionLevel,byte[]> bestSoFar = result.get();
                        if(r.length < bestSoFar.second.length)
                            {
                            if(!result.compareAndSet(bestSoFar, putativeResult))
                                { continue; /* Failed to set result yet. */ }
                            }
                        break; // This is not the best result, so quit.
                        }

                    return;
                    }
                }));

            // Because some of the compressors are very resource-intensive,
            // we'll wait for earlier tasks to complete
            // to ensure that we don't have more tasks running
            // than we have CPUs, regardless of the details of the thread pool.
            while(!tasks.isEmpty() && (tasks.size() >= Runtime.getRuntime().availableProcessors()))
                {
                try { tasks.remove().get(); }
                catch(final ExecutionException e)
                    {
                    // Ignore (though log) error in this particular compression attempt.
                    e.printStackTrace();
                    }
                }
            }

        // Wait for all remaining tasks to complete.
        for(final Future<?> task : tasks)
            {
            try { task.get(); }
            catch(final ExecutionException e)
                {
                // Ignore (though log) error in this particular compression attempt.
                e.printStackTrace();
                }
            }

        return(result.get()); // Return best result.
        }

    /**Serialise the given object and compress the result as well as possible up to the given compression level; never null.
     * Tries all the supported compression methods up to the specified level
     * returning whichever results in the smallest result.
     * <p>
     * (Resource constraints may prevent some of the compression types being applied,
     * usually the highest levels.)
     * <p>
     * This does NOT assume that higher compression levels always compress given data better,
     * not does it assume a lower bound on data size worth compressing;
     * this tries all the possibilities exhaustively if it can.
     * <p>
     * This is likely to be very memory- and CPU- intensive; use with care.
     * <p>
     * This may attempt compressions in parallel on a multi-CPU machine,
     * though each compression is likely to thrash the cache
     * so this may or may not be good for compression and system performance.
     * <p>
     * This routine does not alter the content of the input array.
     * 
     * @param in  the object to be compressed; never null
     * @param maxLevel  the maximum level at which to attempt to compress data,
     *     using a value greater than MAX_SUPPORTED_COMPRESSION_LEVEL is allowed but ineffective;
     *     never null
     *     
     * @return non-null data compressed as well as possible;
     *     if compression was not possible the level is NONE
     *     and the input array is returned
     */
    public static Tuple.Pair<CompressionLevel,byte[]> compressObject(final Serializable in, CompressionLevel maxLevel)
        throws IOException, InterruptedException
        {
        if((null == in) || (null == maxLevel))
            { throw new IllegalArgumentException(); }

        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(in);
        oos.close();
        return(compressData(baos.toByteArray(), maxLevel));
        }

    /**Wrap input stream for decompression according to the compression level; never null.
     * If an unsupported compression level is selected then an exception is thrown.
     * <p>
     * If NONE is chosen as the compression level then the input stream is returned as-is.
     * <p>
     * Can decompress data compressed with compressData()/compressObject().
     */
    public static final InputStream wrapForDecompression(final InputStream in,
                                                         final CompressionLevel cl)
        throws IOException
        {
        if((null == in) || (null == cl))
            { throw new IllegalArgumentException(); }

        switch(cl)
            {
            // NONE means no compression, so return the stream unwrapped.
            case NONE:  { return(in); }

            // ZIP is so CPU-efficient that a data pump is probably not very useful
            //     on single-CPU machines, but may be helpful beyond that.
            // When we do use a data-pump, we allow for good (10x) compression
            // given the ZIP input buffer of 32kB,
            // implying ~320kB+ of buffering is probably good.
            // We assume that ZIP does a reasonable job of buffering its input.
            case ZIP:
                {
                // Only start a new thread if upstream data is not flowing well.
                return(dataPump(new DefInputStream(in), 512 * 1024, false));
                }

            // BZIP2 needs some wrapping for robustness and performance.
            // BZIP2 reads one byte at a time, so some sort of input buffering is vital.
            // BZIP2 can probably benefit from a data pump to multi-thread work and to keep data flow smooth.
            case BZIP2:
                {
                // We wrap this decompressor in two data pumps with output buffering
                // much larger than the maximum input/uncompressed 900kB block size
                // (to optimistically allow for good (~4x) compression!)
                // and input buffering much larger than the 900kB block size
                // which should help to keep the inbound compressed bytes flowing smoothly
                // in the face of BZIP2's very lumpy progress.
                // We assume that the dataPump() will also support BZIP2's
                // inefficient byte-at-a-time read()s reasonably efficiently.
                final InputStream cis;
                try { cis = new CBZip2InputStream(dataPump(in, 2*1024*1024, true)); }
                catch(final NullPointerException e)
                    {
                    // NPE is actually caused by a corrupt input stream.
                    throw new IOException("corrupt/invalid input to BZIP2 decompressor");
                    }
                if(Runtime.getRuntime().availableProcessors() < 2)
                    { return(cis); }
                // Only put a data pump on the output of the compressor
                // if we have multiple CPUs and thus may improve throughput.
                return(dataPump(cis, 4*1024*1024, false));
                }

            // LZMA uses lots of memory but may not be desperately heavy on CPU.
            case LZMA:
                {
                final InputStream cis = new LzmaInputStream(dataPump(in, 1024*1024, true));
                return(cis);
                }
            }

        throw new IOException("unsupported compression level");
        }

    /**Basic dataPump() read size (bytes) for efficient operations; strictly positive. */
    private static final int DATA_PUMP_BLOCK_SIZE = 8192;

    /**Wraps InputStream with a data pump (if possible); never null.
     * This can be used to keep data flowing in a pipeline with "lumpy" CPU use,
     * such as a stream decompressor with large input blocks.
     * <p>
     * This also allows stages on either side of the pump to run in separate threads
     * which can improve throughput further on a multi-CPU JVM.
     * <p>
     * In general this should be transparent to the user,
     * with close()s, exceptions, etc, being propagated much as usual.
     * <p>
     * In particular, a close() on the returned stream will cause a close()
     * on the argument input stream.
     * <p>
     * This routine may return the input stream if it cannot create a pump.
     * <p>
     * Instances of the returned class may not be thread-safe.
     * This is so that we can maximise performance, especially for small reads.
     * <p>
     * If flagged as aggressive then this routine always forces
     * the creation of an extra thread to do read-ahead,
     * else this only does so when an attempted read from down-stream
     * cannot be immediately (partly or fully) satisfied.
     * Thus a "non-aggressive" data pump may save some overhead
     * where is it not needed, eg on short/fast streams.
     *
     * @param is  input data stream; never null
     * @param bufferSize  approximate maximum amount of data to read-ahead;
     *     strictly positive
     * @param aggressive  if true then always start another thread,
     *     else possibly do so only to avoid blocking
     */
    public static final InputStream dataPump(final InputStream is,
                                             final int bufferSize,
                                             final boolean aggressive)
        {
        if((is == null) || (bufferSize <= 0))
            { throw new IllegalArgumentException(); }

//if(IsDebug.isDebug) { System.out.println("[dataPump("+is+","+bufferSize+") starting...]"); }

        // We must override at least the read() and close() methods.
        // Instances of this class are not thread-safe.
        return(new InputStream(){
            /**If true then close() has already been called. */
            private boolean closed;

            /**Close the stream(s), release resources.
             * Is idempotent, ie can be called more than once without ill effect.
             * <p>
             * This is not synchronized to allow an asynchronous call.
             */
            @Override public void close() throws IOException
                {
                if(!closed)
                    {
                    // Mark stream as closed immediately.
                    closed = true;

                    // Ensure that the underlying input is closed directly,
                    // possibly asynchronously...
                    try { is.close(); }
                    catch(final Exception e) { /* Ignore close() errors. */ }

                    // Explictly interrupt the readerThread, if any (and alive),
                    // until it goes away.
                    final Thread t = readerThread;
                    if(t != null)
                        {
                        while(t.isAlive())
                            {
                            t.interrupt();
                            try { Thread.sleep(100); }
                            catch(final InterruptedException e)
                                { throw new InterruptedIOException("could not kill reader thread"); }
                            }
                        }
                    }

                // Release any buffers held.
                readAhead.clear();
                }

            /**Private buffer used by read(void) to avoid lots of heap churn; never null. */
            private final byte[] readBuf = new byte[1];

            /**This is potentially very inefficient but must be provided.
             * Unfortunately it is also the only read method called by
             * (for example) CBZip2InputStream,
             * so we cannot have it be massively slow.
             */
            @Override public int read() throws IOException
                {
//if(IsDebug.isDebug) { System.out.println("[dataPump() read(1).]"); }
                // Fast-path for most reads...
                // ie when we have data immediately available on the head.
                final Pair<Integer, byte[]> head = readAhead.peek();
                if(head != null)
                    {
                    final int available = head.first - readAheadHeadOffset;
                    assert(available > 0) : "we must never have an empty head of queue";
                    final int result = head.second[readAheadHeadOffset++] & 0xff;
                    assert(readAheadHeadOffset <= head.first.intValue());
                    // Remove any now-empty head item.
                    if(readAheadHeadOffset == head.first.intValue())
                        {
                        readAhead.remove();
                        readAheadHeadOffset = 0;
                        }
                    return(result);
                    }

                // Fall back to main generic routine for tricky cases,
                // eg starting the reader thread, blocking for data, and detecting EOF.
                final int n = read(readBuf, 0, 1);
                if(n != 1) { return(-1); /* EOF */ }
                return(readBuf[0] & 0xff);
                }
            
            /**Reader thread to suck data up the pipe.
             * The thread is started only once we have determined that is needed
             * (eg to avoid unnecessary overhead on small/fast transfers)
             * usually at the first read that would have blocked.
             * <p>
             * Once non-null this is not set null again.
             */
            private Thread readerThread;

            /**The maximum size of the read-ahead queue; strictly positive.
             * Determined by a block size and the requested read-ahead buffering.
             * <p>
             * If we have lots of mostly-empty buffers
             * then much less actual data will be queued then the user requested.
             */
            private final int maxQueueLength = 1 + (bufferSize/DATA_PUMP_BLOCK_SIZE);

            /**Queue of data read by the readerThread; never null.
             * This is a queue of (block-size,data) tuples.
             */
            private final BlockingQueue<Tuple.Pair<Integer,byte[]>> readAhead =
                new LinkedBlockingQueue<Tuple.Pair<Integer,byte[]>>(maxQueueLength);

            /**Any (IO) Exception caught by the reader thread, or null if none. */
            private final AtomicReference<IOException> readerEx = new AtomicReference<IOException>();

            /**Offset into current head item of readAhead queue; usually zero. */
            private int readAheadHeadOffset;
            
            /**Object notified when new data is queued. */
            private final Object signalObject = new Object();

            /**This is the generic efficient bulk-data read method.
             * If asked for one or more bytes of input
             * will block until at least one byte is ready to return or EOF is reached,
             * but will otherwise avoid blocking by returning what is immediate to hand.
             */
            @Override public int read(byte[] b, int off, int len) throws IOException
                {
//if(IsDebug.isDebug) { System.out.println("[dataPump() read(len="+len+").]"); }
                // If already closed then quit immediately.
                if(closed) { return(-1); }

                // Handle trivial zero-length read immediately.
                if(len == 0) { return(0); }

                // If we are not aggressive and have not yet started read-ahead
                // then attempt to read without blocking from the underlying stream.
                if(!aggressive && (readerThread == null))
                    {
                    final int avail = is.available();
                    // We can continue to eschew a read-ahead thread for now
                    // if there is some data ready.
                    if(avail > 0)
                        {
                        // Read what we can without blocking.
                        final int n = is.read(b, off, Math.min(avail, len));
                        if(n < 1) { close(); return(-1); /* EOF */ }
                        return(n);
                        }
                    // Fall through to start read-ahead thread.
                    }

                try
                    {
                    // Create the readerThread on the first call.
                    if(readerThread == null)
                        {
                        readerThread = new Thread("GenUtils.dataPump() reader thread"){
                            @Override public final void run()
                                {
                                try
                                    {
//if(IsDebug.isDebug) { System.out.println("[dataPump() started reader thread...]"); }
                                    for( ; ; )
                                        {
                                        final int available = is.available();
                                        // Read what's immediately available,
                                        // but if none then block for a full chunk.
                                        final int bufSize = (available < 1) ? DATA_PUMP_BLOCK_SIZE :
                                            Math.min(available, DATA_PUMP_BLOCK_SIZE);
                                        // TODO: get buffer from a pool if available...
                                        byte buf[] = new byte[bufSize];
                                        final int n = is.read(buf, 0, bufSize);
//if(IsDebug.isDebug) { System.out.println("[dataPump() read "+n+" bytes in buffer of size "+bufSize+".]"); }
                                        // Quit gracefully at EOF.
                                        if(n <= 0) { return; }
                                        // To avoid wasting memory in the queue,
                                        // if we only did a tiny read
                                        // and the buffer is unlikely to be consumed/freed quickly
                                        // then copy the data into a smaller buffer.
                                        if((n != bufSize) && (n < (bufSize>>1)) && !readAhead.isEmpty())
                                            {
                                            final byte newBuf[] = new byte[n];
                                            System.arraycopy(buf, 0, newBuf, 0, n);
                                            buf = newBuf;
//if(IsDebug.isDebug) { System.out.println("[dataPump() reallocated buffer.]"); }
                                            }
                                        // Else if we got some data, try to queue it.
                                        readAhead.put(new Tuple.Pair<Integer,byte[]>(Integer.valueOf(n), buf));
                                        // Wake/notify the consumer thread.
                                        synchronized(signalObject) { signalObject.notifyAll(); }
                                        }
                                    }
                                catch(final IOException e)
                                    {
                                    // Capture the error to re-throw for the caller
                                    // on its next read() operation.
                                    readerEx.set(e);
                                    // And then exit immediately.
                                    return;
                                    }
                                catch(final InterruptedException e)
                                    {
                                    final IOException err = new InterruptedIOException("interrupted");
                                    err.initCause(e);
                                    // Capture the error to re-throw for the caller
                                    // on its next read() operation.
                                    readerEx.set(err);
                                    // And then exit immediately.
                                    return;
                                    }
                                }
                            };
                        readerThread.setDaemon(true); // Don't stop the JVM exiting.
                        readerThread.start();
                        }

                    // Run until there seems to be nothing left to do...
                    for( ; ; )
                        {
                        // While there is queued input,
                        // try to return it to the caller.
                        // We do not block.
                        if(!readAhead.isEmpty())
                            {
                            int bytesRead = 0;
                            do
                                {
                                // Compute what is immediately available on the head of the queue
                                // minus any already returned from this value already.
                                final Pair<Integer, byte[]> head = readAhead.peek();
                                final int available = head.first - readAheadHeadOffset;
                                assert (available > 0) : "we must never have an empty head of queue";
                                final int toCopy = Math.min(len, available);
                                // Copy data to the caller's buffer.
                                System.arraycopy(head.second, readAheadHeadOffset, b, off, toCopy);
                                // Adjust our notion of the amount read from this item.
                                readAheadHeadOffset += toCopy;
                                off += toCopy;
                                len -= toCopy;
                                bytesRead += toCopy;
                                assert(readAheadHeadOffset <= head.first.intValue());
                                // If we've exhausted data from the HEAD then remove() it.
                                if(readAheadHeadOffset == head.first.intValue())
                                    {
                                    readAhead.remove();
                                    readAheadHeadOffset = 0;
                                    }
                                } while(!readAhead.isEmpty() && (len > 0));

                            // Return the data to the caller.
//if(IsDebug.isDebug) { System.out.println("[dataPump() read()="+bytesRead+".]"); }
                            assert(bytesRead > 0);
                            return(bytesRead);
                            }
                        // Quit if the reader thread has died
                        // (with assumed memory-release semantics)
                        // and the read-ahead queue is still empty.
                        else if(!readerThread.isAlive() && readAhead.isEmpty())
                            { break; }

                        final IOException err = readerEx.get();
                        // Quit (and close()) if the reader thread encountered an error.
                        // (Once we've dealt with any pending/queued input...)
                        if(err != null)
                            {
                            close();
                            throw err;
                            }

                        // Wait a little while for some more data and try reading again.
                        // We should end up waking immediately when there is something to do.
                        // We don't block for ever so that a missed notification is not fatal.
                        synchronized(signalObject)
                            {
                            while(readerThread.isAlive() && readAhead.isEmpty())
                                { signalObject.wait(101); }
                            }
                        }

                    // Assume that we have now reached EOF...
//if(IsDebug.isDebug) { System.out.println("[dataPump().read() hit EOF...]"); /* Most expected uses will know they are at EOF without getting this explicit return. */ }
                    close(); // Release resources...
                    return(-1); // Indicate EOF to the caller.
                    }
                catch(final IOException e)
                    {
                    // In case of error try to ensure that resources are released.
                    close(); // Try to free upstream resources ASAP.
                    throw e; // Rethrow the original exception.
                    }
                catch(final InterruptedException e)
                    {
                    // In case of interruption try to ensure that resources are released.
                    close(); // Try to free upstream resources ASAP.
                    final IOException err = new InterruptedIOException("interrupted");
                    err.initCause(e);
                    throw err;
                    }
                }

            /**Indicate what should be immediately available without blocking; non-negative.
             */
            @Override public int available() throws IOException
                {
                // If the reader thread is not yet started
                // then report what upstream says is available.
                if(readerThread == null)
                    { return(is.available()); }

                // Else conservatively report what we have available
                // on the head of the queue right now
                // (always at least 1 if the queue not empty)
                // plus at least one byte for each queued block behind!
                if(!readAhead.isEmpty())
                    {
                    final int onHead = readAhead.peek().first - readAheadHeadOffset;
                    assert(onHead > 0);
                    return(onHead + (readAhead.size()-1));
                    }

                // Nothing obviously immediately available.
                return(0);
                }
            });
        }
    }
