/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.location.GeoUtils.CCTLD;
import org.hd.d.pg2k.svrCore.props.LocalProps;

/**Some simple common host-name utilities.
 */
public final class HostUtils
    {
    /**Prevent construction of an instance. */
    private HostUtils() { }

    /**Normalise host (Web server) name (ours or a third-party one).
     * This converts the host name to lower-case and
     * strips any "www." or "mirror-CC-name." prefix and/or
     * "." suffix.
     * <p>
     * If the name is already normalised,
     * it is returned unchanged.
     *
     * @throws IllegalArgumentException  if arguments are malformed
     */
    public static String normaliseVirtualHostName(String name)
        {
        if(name == null) { throw new IllegalArgumentException(); }

        // Must be at least of form "a.b" and is limited by RFC rules not to be huge.
        if((name.length() < 3) || (name.length() > 255))
            { throw new IllegalArgumentException(); }

        // Force names to lower-case.
        name = name.toLowerCase();

        final int firstDot = name.indexOf('.');

        // Strip off standard prefixes...
        if(name.startsWith("www."))
            { name = name.substring(4); }
        else if(isSpecialName(name))
            { name = name.substring(firstDot+1); }
        else if(name.startsWith("mmm.")) // "Mobile" prefix.
            { name = name.substring(4); }

        // Strip off redundant trailing dot(s)...
        while(name.endsWith("."))
            { name = name.substring(0, name.length() - 1); }

        // Normalised...
        return(name);
        }

    /**Returns true if this looks like a mirror name.
     * This is a canonical name prefixed with something of the form:
     * <samp>mirror-cc-x."</samp>
     * where "cc" is the two-letter country code where the mirror is hosted
     * (eg uk or us or fr)
     * and "x" is a descriptive short tag to indicate which mirror instance.
     *
     * @param name  host name, must be lower-case and non-null
     */
    public static boolean isMirrorName(final String name)
        {
        if(!name.startsWith("mirror-")) { return(false); }
        final int firstDot = name.indexOf('.');
        if(firstDot <= 10) { return(false); } /* At least "mirror-cc-x." */
        return(true);
        }

    /**Extract mirror tag from mirror URL/prefix; null if none.
     * The mirror tag extracted should be of the form
     * returned by LocalProps.getMirrorTag() for a well-formed server name.
     */
    public static String getMirrorTag(final String name)
        {
        if(!isMirrorName(name)) { return(null); }
        final String cc = name.substring(7, 9);
        if(!CCTLD.isSyntaticallyValidCcTLD(cc)) { return(null); }
        return(name.substring(7, name.indexOf('.')));
        }

    /**Extract country code from mirror URL/prefix; null if none. */
    public static CCTLD getMirrorCC(final String name)
        {
        if(!name.startsWith("mirror-")) { return(null); }
        if(name.length() < 11) { return(null); }
        final String cc = name.substring(7, 9);
        if(!CCTLD.isSyntaticallyValidCcTLD(cc)) { return(null); }
        return(new CCTLD(cc));
        }

    /**Returns true if this looks like a "master" server name.
     * This is a canoical name prefixed with "master."
     * and indicates a central server, eg on which uploads are permittted.
     *
     * @param name  host name, must be lower-case and non-null
     */
    public static boolean isMasterName(final String name)
        { return(name.startsWith("master.")); }

    /**Returns true if this looks like a special (mirror or master) name.
     * @param name  host name, must be lower-case and non-null
     */
    public static boolean isSpecialName(final String name)
        { return(isMirrorName(name) || isMasterName(name)); }

    /**Make mirror version of standard "generic" hostname; never null.
     * Generates hostname of form "mirror-tag.gallery.hd.org".
     *
     * @param tag  valid non-null mirror tag
     */
    public static String makeMirrorNameGeneric(final String tag)
        { return("mirror-" + tag + "." + CoreConsts.MAIN_DATA_HOST); }
    }
