/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.location.GeoUtils;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 14-Sep-2003
 * Time: 20:02:02
 */

/**Holds constants and some utility methods for internationalisation (i18n).
 * These are features common to the core system, Web presentation, etc.
 */
public final class I18NTools
    {
    /**Name of common resource bundle for core i18n messages. */
    public static final String BUNDLE_COMMON = "common";

    /**Name of tree-description i18n resource bundle. */
    public static final String BUNDLE_TREEDESC = "treedesc";

    /**Default locale to use in the system and on the Web site.
     * This is British English, and reflects what we expect to
     * find in the base messages (files) that the Web sites
     * will use.
     * <p>
     * This locale is one that we should always be able to support.
     */
    public static final Locale DEFAULT_SYSTEM_LOCALE = new Locale("en", "GB");

    /**Main locales that we support i18n for in an immutable List.
     * Earlier items are preferred (eg more commonly used by visitors);
     * the site default is the first value.
     * <p>
     * Every item in this list is a Locale whose String version is of the form
     * <tt>ll</tt> or <tt>ll_CC</tt>
     * where "ll" is the language and "CC" the optional country code,
     * eg <tt>fr</tt> for French and <tt>en_GB</tt> for UK English.
     * <p>
     * Every item that appears in this list has at least:
     * <ul>
     * <li>Most or all of the common message catalogue translated.
     * <li>A standard flag icon.
     * </ul>
     * <p>
     * There may be partial support for languages/locales other than those
     * listed here.
     */
    public static final List<Locale> MAIN_LOCALES;

    /**Initialise MAIN_LOCALES. */
    static
        {
        final List<Locale> l = new ArrayList<Locale>();

        // Default system value is first.
        l.add(DEFAULT_SYSTEM_LOCALE); // Some variety of English (en).

        // Then other (hopefully) well-supported locales
        // in approximate descending order of use by visitors.
        // List should be kept as small as possible since
        // each extra choice consumes bandwidth and space on most pages,
        // and, for example, we have very limited space on handheld/mobile devices.
        l.add(new Locale("de", ""));
        l.add(new Locale("es", ""));
        l.add(new Locale("zh", "CN"));

        MAIN_LOCALES = Collections.unmodifiableList(l);
        }

    /**Immutable map from (a small set of common) top-level ccTLDs to best overall locale in our supported set for that location.
     * This is not comprehensive
     * (though we try to cover common cases where we have any explicit support),
     * and is mainly to help when a browser (or search engine) does not indicate
     * a preferred locale explicitly.
     */
    public static final Map<GeoUtils.CCTLD, Locale> LOCALE_BY_CCTLD;

    /**Initialise LOCALE_BY_CCTLD. */
    static
        {
        // We use a HashMap for lookup speed.
        final Map<GeoUtils.CCTLD, Locale> m = new HashMap<GeoUtils.CCTLD, Locale>(31);

        m.put(MemoryTools.intern(new GeoUtils.CCTLD("uk")), new Locale("en", "GB")); // Default.
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("us")), new Locale("en", "US")); // Where most users with no language set in their browsers are from.
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("ca")), new Locale("en", "CA"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("au")), new Locale("en", "AU"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("nz")), new Locale("en", "NZ"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("za")), new Locale("en", "ZA"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("in")), new Locale("en", "IN"));

        m.put(MemoryTools.intern(new GeoUtils.CCTLD("cn")), new Locale("zh", "CN"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("tw")), new Locale("zh", "TW"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("hk")), new Locale("zh", "HK"));

        m.put(MemoryTools.intern(new GeoUtils.CCTLD("nl")), new Locale("nl", ""));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("fr")), new Locale("fr", ""));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("de")), new Locale("de", ""));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("at")), new Locale("de", "AT"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("es")), new Locale("es", ""));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("mx")), new Locale("es", "MX"));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("it")), new Locale("it", ""));
        m.put(MemoryTools.intern(new GeoUtils.CCTLD("jp")), new Locale("ja", ""));

        m.put(MemoryTools.intern(new GeoUtils.CCTLD("mt")), new Locale("mt", ""));

        LOCALE_BY_CCTLD = Collections.unmodifiableMap(m);
        }


    /**Immutable Set of most locales supported by the system.
     * Designed for relatively fast lookup.
     */
    public static final Set<Locale> LOCALES;

    /**Initialise LOCALES. */
    static
        {
//        final TreeSet<Locale> s = new TreeSet<Locale>(new Comparator<Locale>(){
//            /**Compares its two arguments for order.
//             * Sorts in toString() order, but is designed to be fairly fast.
//             */
//            public int compare(final Locale locale1, final Locale locale2)
//                {
//                // For speed, do initial compare on language component
//                // which avoids contructing new String values dynamically.
//                final int llComp = locale1.getLanguage().compareTo(locale2.getLanguage());
//                if(llComp != 0) { return(llComp); }
//                // Total ordering on full String representation.
//                return(locale1.toString().compareTo(locale2.toString()));
//                }
//            });

        final HashSet<Locale> s = new HashSet<Locale>(2 * (MAIN_LOCALES.size() + LOCALE_BY_CCTLD.size()));
        s.addAll(MAIN_LOCALES);
        s.addAll(LOCALE_BY_CCTLD.values());

        LOCALES = Collections.unmodifiableSet(s);
        }
    }
