/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.MIME.Quantize;

import java.awt.image.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 30-Aug-2003
 * Time: 23:46:46
 */

/**Basic image-handling utilities.
 */
public final class ImageUtils
    {
    /**Prevent instantiation. */
    private ImageUtils() { }


    /**Count distinct colours in 24-bit RGB colour space.
     * Counts the number of distinct colours in the sRGB colour model
     * in the given image, ignoring alpha, with each the depth of each
     * band being 8 bits, for a total of 24 bits for each pixel.
     * <p>
     * If the source image is not in RGB format then
     * colour conversion will take place.
     * <p>
     * Is slow but tries to be reasonably memory efficient.
     * <p>
     * The input image is not altered.
     */
    public static int countDistinct24BitRGBColours(final BufferedImage im)
        {
        final int height = im.getHeight();
        final int width = im.getWidth();
        final Set<Integer> distinctColours = new HashSet<Integer>();

        for(int y = height; --y >= 0; )
            {
            for(int x = width; --x >= 0; )
                {
                distinctColours.add(new Integer(im.getRGB(x, y) & 0xffffff));
                }
            }

        return(distinctColours.size());
        }


    /**Method to convert a RenderedImage to a BufferedImage.
     * Thanks to <a href="http://www.magelang.com/faq/view.jsp?EID=114602">Jim Moore (06 Apr 2003)</a>
     * for his routine to "do it right"!  Very slightly adjusted by DHD.
     *
     * @param ri  the image to be converted
     *     (if already a BufferedImage then it is returned intact);
     *     must not be null
     *
     * @return the original if already a BufferedImage,
     *     else a copy of the data
     */
    public static BufferedImage convertRenderedImageToBufferedImage(final RenderedImage ri)
        {
        // Return as-is if already a BufferedImage.
        if(ri instanceof BufferedImage)
            { return (BufferedImage) ri; }

        // Extract basic metadata.
        final ColorModel cm = ri.getColorModel();
        final int width = ri.getWidth();
        final int height = ri.getHeight();
        final boolean isAlphaPremultiplied = cm.isAlphaPremultiplied();

        // Extract properties if any.
        final String[] keys = ri.getPropertyNames();
        Hashtable<String, Object> properties = null;
        if(keys != null)
            {
            properties = new Hashtable<String, Object>(keys.length * 2 + 1);
            for(int i = 0; i < keys.length; i++)
                { properties.put(keys[i], ri.getProperty(keys[i])); }
            }

        // Construct BufferedImage and copy (raster) data in.
        final WritableRaster raster = cm.createCompatibleWritableRaster(width, height);
        final BufferedImage result = new BufferedImage(cm, raster, isAlphaPremultiplied, properties);
        ri.copyData(raster);

        return(result);
        }

    /**Create colour reduced image; never null
     * A new image is created that is potentially colour-reduced from the
     * original; no new colours should be introduced unless during the
     * conversion to or from the RGB colour space in which reduction is done.
     * The R, G and B quantities are each 8 bits.
     * <p>
     * By default the new image is returned in the same ColorModel as
     * the original, but forcing the result to an indexed
     * (palette-based) model such as used by GIF can requested,
     * in which case the palette size is capped appropriately,
     * and the bits-per-pixel may be chosen to be 1, 2, 4, or 8
     * depending on the palette size.
     * <p>
     * Any input image that can be converted to RGB should be acceptable.
     * <p>
     * If no colour reduction is achieved and no ColorModel change is requested
     * then the original image is returned;
     * in no case is the original image altered.
     * <p>
     * This may return a single-colour image if the input image is too tricky
     * and the maximum number of colours it can return is too small,
     * but that will be avoided if possible.
     *
     * @param in  the original image; must not be null
     * @param maxColours  maximum number of distinct colours allowed in the
     *     result image (as measured in a 24-bit (8,8,8) sRGB colour space);
     *     must be at least 2 and will be capped to 256 if an indexed model
     *     is forced for the result
     * @param forceByteIndexModel  if true, force the result to be a
     *     byte-indexed palette-based color model (maximum 256 colours)
     *     that may be (a) suitable for formats such as GIF and
     *     (b) more compressable
     *
     * @return image same dimensions as the input image,
     *     possibly with fewer colours; never null though might be single colour
     */
    public static BufferedImage makeColourReducedBufferedImage(
                                                    final BufferedImage in,
                                                    int maxColours,
                                                    final boolean forceByteIndexModel)
        {
//System.out.println("makeColourReducedBufferedImage("+maxColours+"): START *****************");

        if(in == null)
            { throw new IllegalArgumentException("input image cannot be null"); }

        if(maxColours < 2)
            { throw new IllegalArgumentException("at least two colours must be allowed in the result"); }

        if(forceByteIndexModel && (maxColours > 256))
            { maxColours = 256; }

        // Extract copy of pixels from source image in TYPE_INT_ARGB format.
        // Set scansize to be width to efficiently pack the result.
        final int width = in.getWidth();
        final int height = in.getHeight();
        final ColorModel cm = ImageUtils.extractColorModelOrRGB(in);
        //final boolean isAlphaPremultiplied = in.isAlphaPremultiplied();

        // If the input already has a byte-indexed palette,
        // and maxColours is no smaller than the palette already in the image,
        // then return the original.
        if((cm instanceof IndexColorModel) &&
           (((IndexColorModel) cm).getMapSize() <= maxColours) &&
           (cm.getTransferType() == DataBuffer.TYPE_BYTE))
            {
//System.out.println("makeColourReducedBufferedImage(): input no more colours than max and already byte indexed: returning original");
            return(in); // Input already a byte-indexed colour model.
            }

        // Extract RGB pixels from source image.
        final int pixels[] = in.getRGB(0, 0, width, height, null, 0, width);

//System.out.print("makeColourReducedBufferedImage(): input pixels = ");
//            for(int i = Math.min(20, pixels.length); --i >= 0; ) { System.out.print(Integer.toHexString(pixels[i]) + ", "); }
//System.out.println();

        // Count the colours in the original image (ignore any alpha).
        final Set<Integer> originalDistinctColours = new TreeSet<Integer>();
        for(int i = pixels.length; --i >= 0; )
            { originalDistinctColours.add(new Integer(pixels[i] & 0xffffff)); }
//System.out.println("makeColourReducedBufferedImage(): original colour count: " + originalDistinctColours.size());

        // Now do the colour reduction...
        // Note that each RGB pixel value is replaced with a palete *index*.
        final int[] palette = Quantize.quantizeImage(
            new int[][]{pixels}, maxColours);
//System.out.println("makeColourReducedBufferedImage(): new colour count: " + palette.length);

        // If we did not reduce the colours, consider returning the original.
        final int origCols = originalDistinctColours.size();
        if(palette.length >= origCols)
            {
            // If did not try to force an indexed output,
            // then return the original.
            if(!forceByteIndexModel)
                {
//System.out.println("makeColourReducedBufferedImage(): did not manage to reduce number of colors: returning original");
                return(in); // Didn't reduce the image at all.
                }

            // If the input already had a byte-indexed palette,
            // and the new palette is no smaller than the old
            //    (ie there were no redundant/duplicate colours in the original)
            // then return the original.
            if((cm instanceof IndexColorModel) &&
               (((IndexColorModel) cm).getMapSize() <= palette.length) &&
               (cm.getTransferType() == DataBuffer.TYPE_BYTE))
                {
//System.out.println("makeColourReducedBufferedImage(): did not manage to reduce number of colors and already byte indexed: returning original");
                return(in); // Input already a byte-indexed colour model.
                }

            // Need to generate byte-indexed form, so fall through...
            }

//System.out.print("makeColourReducedBufferedImage(): palette = ");
//            for(int i = palette.length; --i >= 0; ) { System.out.print(Integer.toHexString(palette[i]) + ", "); }
//System.out.println();
//System.out.print("makeColourReducedBufferedImage(): pixels = ");
//            for(int i = Math.min(20, pixels.length); --i >= 0; ) { System.out.print(Integer.toHexString(pixels[i]) + ", "); }
//System.out.println();


        // Replace palette index values with RGB values.
        // Force all pixels to be opaque before setting them in the image.
        for(int i = pixels.length; --i >= 0; )
            { pixels[i] = palette[pixels[i]] | 0xff000000; }

        // Handle creation of a new byte-indexed colour map specially,
        // though we don't really need to.
        BufferedImage result;
        if(forceByteIndexModel)
            {
            // Force all palette entries to be opaque.
            for(int i = palette.length; --i >= 0; )
                { palette[i] |= 0xff000000; }

            final IndexColorModel outputColourModel =
                new IndexColorModel(computeIndexBitsForByteIndexColorMap(palette.length),
                                    palette.length,
                                    palette,
                                    0, // Start at offset zero.
                                    false, // No alpha.
                                    -1, // No transparent colour.
                                    DataBuffer.TYPE_BYTE); // 8-bit palette

            result = new BufferedImage(width, height,
                                       BufferedImage.TYPE_BYTE_INDEXED,
                                       outputColourModel);
            }
        else
            {
            // Coerce data into original colour model.
            // Discard any properties of the original.
            final ColorModel outputColourModel = cm;

            // Handle other images in default way.
            final WritableRaster raster =
                outputColourModel.createCompatibleWritableRaster(width, height);
            result = new BufferedImage(outputColourModel, raster, false, null);
            }

        // Actually store the pixels...
        result.setRGB(0, 0, width, height, pixels, 0, width);

//System.err.println("makeColourReducedBufferedImage(): palette.length = " + palette.length + ", number of colours in result = " + countDistinct24BitRGBColours(result));

        return(result);
        }

    /**Compute the number of bits to store the pixel of an indexed image given the size of the palette; strictly positive power of two up to 8.
     * The default behaviour is to return 8, for a byte-indexed palette,
     * but this may return 4 if the palette has no more than 16 entries,
     * 2 if the palette has no more than 4 entries, or
     * 1 if the palette has no more than 2 entries,
     * ie widths that will pack neatly into a byte.
     *
     * @return  number of pits needed to store palette,
     *     as submultiple of 8
     */
    private static int computeIndexBitsForByteIndexColorMap(int paletteSize)
        {
        if((paletteSize < 0) || (paletteSize > 256))
            { throw new IllegalArgumentException(); }

        if(paletteSize <= 2) { return(1); }
        if(paletteSize <= 4) { return(2); }
        if(paletteSize <= 16) { return(4); }

        // Default is to use a whole byte for each pixel.
        return(8);
        }

    /**Return true-colour ARGB format image if src is indexed, else return original.
     * The returned image should be easy to draw on.
     * <p>
     * The src image is not altered.
     *
     * @param src  input image, is not altered; must not be null
     * @return non-indexed image
     */
    public static BufferedImage convertToTrueColourARGB(final BufferedImage src,
                                                        final boolean forceCopy)
        {
        if(src == null)
            { throw new IllegalArgumentException(); }

        // If image is indexed, expand to full true-colour format.
        // Else if copy is forced, do it anyway to get a separate copy.
        if(forceCopy ||
           (src.getColorModel() instanceof IndexColorModel))
            {
            final int width = src.getWidth();
            final int height = src.getHeight();

            final BufferedImage rgbBi = new BufferedImage(
                width, height, BufferedImage.TYPE_INT_ARGB);
            rgbBi.setRGB(0, 0, width, height,
                         src.getRGB(0, 0, width, height, null, 0, width),
                         0, width);

            return(rgbBi);
            }

        // Return original.
        return(src);
        }

    /**Extract the ColorModel of the inpuit image, or return the default RGB model; never null.
     * This is needed because BufferedImage.getColorModel() can return null.
     *
     * @param imageIn  input image from which to extract the ColorModel;
     *     must not be null
     * @return a non-null ColorModel
     */
    public static ColorModel extractColorModelOrRGB(final BufferedImage imageIn)
        {
        // Extract just the map fragment that we need.
        // Coerce data into original colour model.
        // Discard any properties of the original.
        final ColorModel rcm = imageIn.getColorModel();

        // If the source image does not have a colour model,
        // create an appropriate default one (RGB).
        final ColorModel cm = (rcm != null) ?
            rcm : DirectColorModel.getRGBdefault();

        return(cm);
        }

    /**Returns true if exhibit might be used as own thumbnail (in an HTML page for example).
     * If true then we don't count downloads of this exhibit
     * (and may adjust computation of "goodness" to take account of this).
     * <p>
     * An exhibit is potentially inlinable if:
     * <ul>
     * <li>It is small enough (smaller than the maximum size of a thumbnail).
     * <li>It is a type which can typically be inlined, eg a GIF image.
     * </ul>
     *
     * @param esa  the exhibit details; never null
     *
     * @return true  if the exhibit is potentially inlinable
     */
    public static boolean canBeOwnThumbnail(final ExhibitStaticAttr esa)
        {
        return((esa.length <= ExhibitThumbnails.STD_ABS_MAX_BYTES) &&
               canInlineInHTMLPageSimple((ExhibitMIME.getInputFileType(esa.filePath))));
        }

    /**Returns true if the given MIME-type can always be inlined in an HTML page.
     * This only tests for very common and widely-supported image types.
     * <p>
     * This does not dynamically test the capabilities of the user agent
     * that content is to be shown to.
     * <p>
     * If the exhibitType argument is null, this returns false.
     */
    public static boolean canInlineInHTMLPageSimple(final ExhibitMIME.ExhibitTypeParameters exhibitType)
        {
        if(exhibitType == null) { return(false); }
        switch(exhibitType.type)
            {
            case ExhibitMIME.ET_JPEG:
            case ExhibitMIME.ET_GIF:
            case ExhibitMIME.ET_PNG: // Most HTML browsers will accept PNG now.
                { return(true); }
            }
        return(false);
        }
    }
