package org.hd.d.pg2k.svrCore;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 04-Sep-2006
 * Time: 16:34:48
 * To change this template use File | Settings | File Templates.
 */
public class LocaleBeanBase implements Serializable
    {
    /**``Safe'' locale is the one we want all activity to default to.
     * This locale is one that we should always be able to support.
     */
    protected static final Locale SAFE_LOCALE = I18NTools.DEFAULT_SYSTEM_LOCALE;

    /**Locale retrieved by setRequest, or a safe locale value; never null.
     * Volatile to allow access without a lock.
     */
    protected volatile Locale userLocale;

    /**If true then monitor missing i18n support.
     * This helps us prioritise translation work!
     * <p>
     * Note that this may be relatively expensive to enable.
     */
    private static final boolean MONITOR_MISSING_I18N = true;

    /**If true then count missing message translations in supported languages.
     * This helps us prioritise translation work!
     * <p>
     * Note that this may be relatively expensive to enable.
     */
    private static final boolean COUNT_MISSING_TRANSLATIONS = true;

    /**Public no-arg constructor for ease of use as a JavaBean.
     * Sets to a safe locale.
     */
    public LocaleBeanBase()
        { this(SAFE_LOCALE); }

    /**Public no-arg constructor for ease of use as a JavaBean.
     * This defers as much work as it reasonably can.
     *
     * @param l  initial locale; never null
     */
    public LocaleBeanBase(final Locale l)
        {
        this.userLocale = l;

        // Verify object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Retrieve the current locale.
     * The value returned is never null.
     */
    public Locale getLocale()
        { return(userLocale); }
    
    /**Cache from real bundle locale to the common bundle.
     * Assumed to be bounded in size by the actual concrete bundles available.
     */
    private static final ConcurrentHashMap<Locale, ResourceBundle> _cachedCommonBundles =
        new ConcurrentHashMap<Locale, ResourceBundle>(101); 

    /**Private LRU cache from requested common locale to actual locale.
     * Kept large enough to keep the hit rate acceptable,
     * but otherwise effectively size-bounded with unused items purged.
     */
    private static final MemoryTools.SimpleLRUMapAutoSizeForHitRate<Locale, Locale> _cacheCommonRealLocales =
        new MemoryTools.SimpleLRUMapAutoSizeForHitRate<Locale, Locale>(0.0001f, 1001, 1123456, 0.7f);
    
    /**Gets the common resource bundle for the given locale.
     * Assume that this is safe since the bundle should be immutable.
     * <p>
     * Canonicalises requested locale to a concrete available common bundle,
     * and looks that up.  Both parts of this lookup are cached.
     * <p>
     * Thread-safe and allows concurrency.
     */
    public ResourceBundle getCommonResourceBundle(final Locale l)
        {
        // Look up real locale.
        final Locale realLocale = _cacheCommonRealLocales.get(l);
        if(null != realLocale)
            {
            final ResourceBundle result = _cachedCommonBundles.get(realLocale);
            if(null != result)
                { return(result); } // Got result from cache.
            }

        // Fetch and permanently cache the resource bundle under its real locale.
        final ResourceBundle result = ResourceBundle.getBundle(I18NTools.BUNDLE_COMMON, l);
        final Locale realLocaleFound = result.getLocale();
        _cachedCommonBundles.putIfAbsent(realLocaleFound, result);
        _cacheCommonRealLocales.put(l, realLocaleFound); // Note the locale mapping from requested to real.
        return(result);
        }

    /**Gets the appropriate localised message from the common set.
     * In case of a missing message this returns the key itself.
     * <p>
     * If the name is null, the result is null.
     */
    public String getLocalisedMessage(final String msgName)
        {
        if(msgName == null) { return(null); }
        try
            {
            // Fetch the common resource bundle and look up our message.
            final ResourceBundle bundle = getCommonResourceBundle(userLocale);

            // Get the localised message, if it exists.
            final String localisedMsg = bundle.getString(msgName);

//            if(MONITOR_MISSING_I18N)
//                {
//                // Note a completely unsupported language (we had to fall back).
//                final Locale bundleLocale = bundle.getLocale();
//                final String bundleLang = bundleLocale.getLanguage();
//                if(!bundleLang.equals(userLocale.getLanguage()))
//                    { StatsLogger.captureDataPoint(LocaleBean.statsIDLang, "MISSING-LANG-" + userLocale.getLanguage()); }
//                // Note if we are missing specific translations
//                // for an otherwise-supported language.
//                // THIS MAY BE EXPENSIVE.
//                else if((COUNT_MISSING_TRANSLATIONS) &&
//                        (!SAFE_LOCALE.getLanguage().equals(bundleLang)))
//                    {
//                    // This assumes that the translation will NOT be the same as
//                    // the original fallback-language version!
//                    if(localisedMsg.equals(getCommonResourceBundle(SAFE_LOCALE).getString(msgName)))
//                        { StatsLogger.captureDataPoint(LocaleBean.statsIDLang, "MISSING-TRANS-" + bundleLang + '-' + msgName); }
//                    }
//                }

            return(localisedMsg);
            }
        catch(final MissingResourceException e)
            {
//            // Note missing message...
//            if(MONITOR_MISSING_I18N)
//                { StatsLogger.captureDataPoint(LocaleBean.statsIDLang, "MISSING-MSG-" + msgName); }
            return(msgName);
            }
        }

    /**Get localised message from the common set with embedded formatting to apply to its argument(s).
     * In case of a missing message this returns the key itself.
     * <p>
     * If the name is null, the result is null.
     */
    public String getLocalisedMessage(final String msgName, final Object ... args)
        {
        if(msgName == null) { return(null); }
        final String fmt = getLocalisedMessage(msgName);
        return((new MessageFormat(fmt, userLocale)).format(args)); // Use the correct locale...
//        return(java.text.MessageFormat.format(fmt, args)); // Use the server's locale.
        }

//    /**Gets the tree-description resource bundle for the given locale.
//     * Assume that the built-in caching is sufficient,
//     * and that this is safe since the bundle should be immutable.
//     */
//    public ResourceBundle getTreeDescResourceBundle(final Locale l)
//        {
//        return(ResourceBundle.getBundle(I18NTools.BUNDLE_TREEDESC, l));
//        }
//
//    /**Gets the appropriate localised tree-description message; never null unless the key is null.
//     * In case of a missing message this returns the key itself.
//     * <p>
//     * The key is converted to lower-case (if necessary) before lookup.
//     * <p>
//     * If the name is null, the result is null.
//     */
//    public String getLocalisedTreeDescMessage(final String msgName)
//        {
//        if(msgName == null) { return(null); }
//        try {
//            // Fetch the common resource bundle and look up our message.
//            final ResourceBundle b = getTreeDescResourceBundle(userLocale);
//            return(b.getString(msgName.toLowerCase()));
//            }
//        catch(final MissingResourceException e)
//            { return(msgName); }
//        }

    /**Deserialise. */
    private void readObject(final ObjectInputStream in)
        throws IOException, ClassNotFoundException
        {
        in.defaultReadObject();
        validateObject(); // Validate state immediately.
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        if(userLocale == null)
            { throw new InvalidObjectException("bad object: userLocale == null"); }
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -6180248508840300658L;
    }
