/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.MIME;

import org.hd.d.pg2k.svrCore.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataFormatImpl;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;

/**Base class with useful default behaviour for image exhibit media-handler classes.
 * By default this uses ImageIO (iio) routines to extract an image and metadata from
 * an exhibit stream,
 * and will allow the generation of thumbnails if an encoder and thumbnail parameters
 * can be found at run-time.
 * <p>
 * Various routines can be overridden in deriving classes to change this behaviour.
 * <p>
 * Potentially-memory-intensive routines may be synchronised in this class (on a private lock)
 * in order to reduce the chance of blowing up the VM by accident,
 * but not if this may cause deadlock elsewhere.
 */
public abstract class AbstractImageHandler extends AbstractHandler
    {
    /**Make thumbnails/samples for the specified exhibit; null if thumbnails cannot (currently) be made.
     * This may fail with an IOException or return null
     * to indicate that it is currently unable to create
     * thumbnails/samples (though this condition may be temporary).
     * <p>
     * If this wishes to indicate that it cannot ever make one or more
     * thumbnails/samples for a given exhibit then this should return a
     * ExhibitThumbnails object with one or both thumbnails set to null.
     * <p>
     * (If canMakeThumbnails() returns false, this should return null.)
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * This will only work correctly if the exhibit is of the correct type,
     * eg its magic number must already have been tested.
     * <p>
     * This assumes enough memory and other resource is available.
     * <p>
     * By default, returns null, ie thumbnails cannot be made now.
     * <p>
     * Sophisticated handlers may wish to override this.
     * <p>
     * Whatever thumbnails are returned,
     * any small thumbnail is smaller than any standard thumbnail, and
     * any standard thumbnail is smaller than the original.
     *
     * @param is  the whole raw image positioned at its start
     * @param originalLength  is the length of the encoded original in bytes;
     *     always positive and must reflect the stream input
     *     and will be constrained to Integer.MAX_VALUE if larger
     * @return thumbnails, else null if not possible now;
     *     ExhibitThumbnails.NO_THUMBNAILS may be returned if it looks
     *     like it never be possible/sensible to make thumbnails for
     *     this exhibit
     */
    public ExhibitThumbnails makeThumbnails(final InputStream is,
                                            final long originalLength)
        throws IOException
        {
        if(is == null)
            { throw new IllegalArgumentException(); }
        if(originalLength <= 0)
            { throw new IllegalArgumentException(); }

        if(!canMakeThumbnails())
            { return(null); }

        // Compute max size of thumbnails,
        // constrained by size of original.
        final int maxTNBytes =
            (int) Math.min(originalLength - 1, Integer.MAX_VALUE);

        // If we can't decode the image, give up for now.
        final BufferedImage bi;
        try { bi = decodeImage(is); }
        // Absorb IllegalArgumentException decoding some images, eg GIF/PNG.
        // Do not try again in this case.
        catch(final IllegalArgumentException e)
            {
            System.err.println("ERROR: could not decode image: will not try to build thumbnails again");
            e.printStackTrace();
            return(ExhibitThumbnails.NO_THUMBNAILS); /* Permanent error. */
            }
        if(bi == null) { return(null); /* Transient error; retry may be OK. */ }

        // Try first to make the small thumbnail.
        // We must always try to have this if possible,
        // and we will then have a lower bound for the size of the standard one.
        final byte sml[] = makeThumbnailImage(bi, false, 0, maxTNBytes);

        // Try next to make the large thumbnail.
        // It must be larger (in filesize) than the small thumbnail.
        final byte std[] = makeThumbnailImage(bi, true,
                            (sml == null) ? 0 : sml.length + 1,
                            maxTNBytes);

        // If neither small nor standard thumbnails could be made
        // (and no exception was thrown indicating a tranisent problem)
        // then assume the error is permanent to avoid wasting time later retrying.
        if((sml == null) && (std == null))
            { return(ExhibitThumbnails.NO_THUMBNAILS); }

        // Create and return the compound result with one/both thumbnails.
        return(ExhibitThumbnails.createExhibitThumbnails(
            ((sml == null) ? null :
                new ExhibitThumbnails.Thumbnail(sml,
                    get2DImageDimensions(new ByteArrayInputStream(sml)))),
            ((std == null) ? null :
                new ExhibitThumbnails.Thumbnail(std,
                    get2DImageDimensions(new ByteArrayInputStream(std))))));
        }

    /**Make thumbnails/samples for the specified exhibit; null if thumbnails cannot (currently) be made.
     * A data source for the exhibit must be supplied,
     * along with all available properties of that exhibit.
     * <p>
     * This may fail with an IOException or return null
     * to indicate that it is currently unable to create
     * thumbnails/samples (though this condition may be temporary).
     * <p>
     * If this wishes to indicate that it cannot make one or more
     * thumbnails/samples for a given exhibit then this should return a
     * ExhibitThumbnails object with one or both thumbnails set to null.
     * <p>
     * (If canMakeThumbnails() returns false, this should return null.)
     * <p>
     * By default, returns null, ie thumbnails cannot be made.
     * <p>
     * This routine regulates memory use, rejecting the attempt
     * to make the thumbnails if it cannot find/reserve sufficient memory
     * using MemoryTools.runMemoryIntensiveOperation().
     * <p>
     * Calls the InputStream version of makeThumbnails().
     *
     * @param unlimitedResources  if true, the generation routine is allowed
     *     to try to use unlimited resources (especially memory)
     * @return thumbnails, else null if not possible now
     *     or this format does not support thumbnails, eg a sound clip;
     *     ExhibitThumbnails.NO_THUMBNAILS may be returned if it looks
     *     like it never be possible/sensible to make thumbnails for
     *     this particular exhibit
     */
    public final ExhibitThumbnails makeThumbnails(
                        final ExhibitStaticAttr esa,
                        final AllExhibitProperties.ExhibitDataSource eds,
                        final AllExhibitProperties aep,
                        final boolean unlimitedResources)
        throws IOException
        {
        if((esa == null) ||
           (eds == null) ||
           (aep == null))
            { throw new IllegalArgumentException(); }

//if(ORG.hd.d.IsDebug.isDebug) { System.err.println("makeThumbnails(): " + esa.filePath); }

        if(!canMakeThumbnails())
            { return(null); }

        // If the exhibit is impossibly small, so we can't be smaller, give up permanently.
        if(esa.length < 2) { return(ExhibitThumbnails.NO_THUMBNAILS); }

        // Get the computed properties of the image.
        final ExhibitPropsComputable epc = aep.getExhibitPropsComputable(esa.filePath);
        // Can't/won't compute a thumbnail
        // if we can't get at the image's X,Y dimensions.
        //
        // This shouldn't really happen
        // so we'll treat this as a temporary failure (!)
        // and return null!
        //
        // (If the computable properties cannot be computed an IOException
        // will get thrown by getExhibitPropsComputable() in the same spirit.)
        if(epc == null)
            { return(null); }
        final Dimension imageXY = epc.getXyDimensions();
        if(imageXY == null)
            { return(null); }

        // The class in which we will build our results
        // if resources permit.
        final class MakeIt implements Runnable
            {
            /**Computed result: null unless successfully completed. */
            ExhibitThumbnails result;

            /**Set true if we failed to generate thumbnails but a retry may be OK. */
            boolean retry;

            public void run()
                {
                try
                    {
                    // Return our definitive answer.
                    // This might be ExhibitThumbnails.NO_THUMBNAILS,
                    // but that simply tells us that
                    // they definitely cannot be produced,
                    // and so we need not try again.
                    result = makeThumbnails(eds.getInputStream(esa),
                                            esa.length);
                    retry = true; // Retry later may well be OK!
                    }
                catch(final Error e)
                    {
                    // Print any error we encounter,
                    // but otherwise discard it.
                    System.err.println("ERROR generating thumbnails for: " + esa.filePath);
                    e.printStackTrace();
                    }
                catch(final IllegalArgumentException e)
                    {
                    // Print any error we encounter,
                    // but otherwise discard it.
                    // We don't encourage retries.
                    System.err.println("ERROR (IllegalArgumentException) generating thumbnails for: " + esa.filePath);
                    e.printStackTrace();
                    }
                catch(final Exception e)
                    {
                    // (IO)Exceptions are often transient,
                    // and we should be able to try again later.
                    System.err.println("Exception generating thumbnails for: " + esa.filePath);
                    e.printStackTrace();
                    retry = true;
                    }
                }
            }

        final MakeIt task = new MakeIt();
        if(!MemoryTools.runMemoryIntensiveOperation(task,
                        unlimitedResources,
                        estimateWorkingMemoryToCreateThumbnails(imageXY)))
            { return(null); } // Failed, maybe temporarily due to lack of resources.

        // Iff we had the resources but could not generate thumbnails
        // then mark this down as a permanent failure.
        if(!task.retry && (task.result == null))
            {
            System.err.println("WARNING: permanently failing generation of thumbnails for: " + esa.filePath);
            return(ExhibitThumbnails.NO_THUMBNAILS);
            }

        // Return the result.
        return(task.result);
        }


    /**Image size tolerance as a fraction of the target size; strictly positive.
     * The tolerance is targetSize/SIZE_TOL, possibly circumscribed by
     * other absolute limits.
     * <p>
     * The larger this number the tighter in conformance to the target size
     * the final image will have to be, which will make convergence
     * take longer or in extreme cases prevent generation of a thumbnail
     * altogether.
     * <p>
     * A reasonable value for this is probably in the region 3--50.
     */
    private static final int THUMBNAIL_SIZE_TOL = 5;

    /**Return individual thumbnail image as file-format byte array in same format as original; null if not possible.
     * This is used for still images and possible animations
     * or movies where the bounding rectangle and the contained data
     * is scaled to the given bounds.
     * <p>
     * This uses a binary chop algorithm on the "quality" parameter
     * calling the underlying makeScaledImage() routine
     * to try to get the output file size to reasonable bounds.
     * <strong>This may fail to work well, or at all, if the image size
     * is not fixed or monotonically increasing with quality,
     * or fails to produce any output at all some some values of
     * the quality parameter.</strong>  We treat the failure to generate
     * an image as an indication that the quality parameter needs to be
     * set higher.
     * <p>
     * The output image will, if possible, use the same colour scheme
     * and other characteristics as the input, though some optional
     * features that usually consume extra space, such as interlacing,
     * may be disabled.
     * <p>
     * If no output can be generated this returns null,
     * which is the default behaviour.
     * <p>
     * This should adjust the image "quality" with the detail
     * value (adjusted within the bounds supplied by the handler class)
     * to try to tune the output size.  For a lossy encoding format
     * such as JPEG this may be the "quality" factor or compression.
     * For a lossless format the may have to be the number of
     * bits-per-pixel that a colour map is reduced to, for example.
     * <p>
     * This base-class implementation can be overriden by sophisticated
     * handlers.
     * <p>
     * This first scales the image using a (fast) nearest-pixel method,
     * which should be good for true-colour and palette images.
     *
     * @param imageIn  source image; must not be null and must be suitable
     *     for the target format
     * @param std  if true we are making a standard thumbnail,
     *     else we are making a small thumbnail
     * @param minSize  minimum (floor) size of thumbnail in bytes;
     *     must be non-negative
     *     (can be zero if not required)
     * @param maxSize  maximum (ceiling) size of thumbnail in bytes;
     *     must be non-negative and no smaller than minSize
     *     (can me Integer.MAX_VALUE if not required)
     */
    protected byte[] makeThumbnailImage(final BufferedImage imageIn,
                                        final boolean std,
                                        final int minSize,
                                        final int maxSize)
        throws IOException
        {
        if((imageIn == null) || (minSize < 0) || (maxSize < minSize))
            { throw new IllegalArgumentException(); }

        // Get the thumbnail parameters.
        final ThumbnailParams tp = getThumbnailParams();
        if(tp == null) { return(null); } // Give up if no parameters.

        // Capture the source X,Y.
        final Dimension srcXyDim = new Dimension(imageIn.getWidth(),
                                              imageIn.getHeight());

        // Compute the thumbnail X,Y.
        final Dimension thumbnailXyDim =
            ExhibitThumbnails.computeThumbnailDimensions(srcXyDim, std);

        // We make the scaled image in memory once,
        // then call the handler routine to build the output image,
        // binary-chopping the "quality" to get as close as possible
        // to our target size reasonably quickly.

        // We use nearest-neighbour scaling because:
        //    * It is fast.
        //    * It should work for both true-colour and palette images.
        final AffineTransformOp op = new AffineTransformOp(
            AffineTransform.getScaleInstance(
                thumbnailXyDim.width  / (double) srcXyDim.width,
                thumbnailXyDim.height / (double) srcXyDim.height),
            AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
        final BufferedImage scaledImage = op.filter(imageIn, null);

        // Get the upper limit for this thumbnail size.
        final int absMaxSize = Math.min(maxSize, std ?
            ExhibitThumbnails.STD_ABS_MAX_BYTES :
            ExhibitThumbnails.SML_ABS_MAX_BYTES);

        // Target size in bytes for thumbnail; capped by the absolute limit.
        final int targetBytes = Math.min(absMaxSize,
                tp.approxMinOverheadBytes +
                Math.max(1,
                         (tp.thumbnailBppHint *
                          thumbnailXyDim.width *
                          thumbnailXyDim.height + 7) / 8));

        // Compute target max/min boundaries.
        // We'll stop immediately if we hit one of these.
        final int targetMax = Math.min(absMaxSize,
                targetBytes + targetBytes / THUMBNAIL_SIZE_TOL);
        final int targetMin = Math.max(minSize,
                targetBytes - targetBytes / THUMBNAIL_SIZE_TOL);

        return(makeSizeConstrainedEncodedImage(tp.minQuality,
                                           tp.initialQualityHint,
                                           tp.maxQuality,
                                           scaledImage,
                                           targetMin, targetMax,
                                           minSize, absMaxSize,
                                           targetBytes));
        }

    /**Make a byte[]-encoded image of this type constained above and below by size; null if not possible.
     * This uses a binary-chop algorithm to attempt to quickly
     * find the optimal "quality" to make the image at,
     * lower quality values asking the encoder to discard more information,
     * eg by doing colour reduction or other quantisation.
     * <p>
     * This is not possible for all exhibit types.
     *
     * @param upperQualityBound  maximum value of quality to use; non-negative
     * @param lowerQualityBound  minimum value of quality to use; non-negative
     *     and no greater than upperQualityBound
     * @param initialQualityHint  initial suggested quality hint; non-negative,
     *     no greater than upperQualityBound, no less than lowerQualityBound
     * @param inputImage  the input image to encode
     * @param targetMin  target lower bound, should be higher than absMinSize
     * @param targetMax  target upper bound, should be lower than absMaxSize
     * @param absMinSize  absolute minimum number of bytes
     * @param absMaxSize  absolute maximum number of bytes
     * @param targetBytes  target number of bytes
     *
     * @return encoded image, or null if the image cannot be generated
     *     with given constraints
     *
     * @throws IOException  in case of difficulty generating the image
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public byte[] makeSizeConstrainedEncodedImage(int lowerQualityBound,
                                               final int initialQualityHint,
                                               int upperQualityBound,
                                               final BufferedImage inputImage,
                                               final int targetMin,
                                               final int targetMax,
                                               final int absMinSize,
                                               final int absMaxSize,
                                               final int targetBytes)
        throws IOException,
               IllegalArgumentException
        {
        if((lowerQualityBound < 0) ||
           (upperQualityBound < 0))
            { throw new IllegalArgumentException("lower or upper quality bounds negative"); }
        if(lowerQualityBound > upperQualityBound)
            { throw new IllegalArgumentException("quality bounds reversed"); }
        if((initialQualityHint < lowerQualityBound) ||
           (initialQualityHint > upperQualityBound))
            { throw new IllegalArgumentException("initialQualityHint out of bounds"); }

        if((absMinSize < 0) ||
           (absMaxSize < 0))
            { throw new IllegalArgumentException("lower or upper size bounds negative"); }
        if(absMinSize > absMaxSize)
            { throw new IllegalArgumentException("size bounds reversed"); }

        // Note the best image so far, or null if none.
        // We will use this even if we don't find one near enough to
        // the target to stop early.
        //
        // Best means:
        //   * Within the absolute bounds on size.
        //   * Arithmetically closest in size to the target.
        byte bestImageSoFar[] = null;

        // Attempt to get reasonably close to the optimal value by binary chop.
        for(int qual = initialQualityHint;
            (qual >= lowerQualityBound) && (qual <= upperQualityBound);
            /* Re-init happens at loop end. */ )
            {
            final byte[] image = makeImageBinary(inputImage, qual);

            // Assume that the failure to generate an image
            // means that the quality parameter was unusably low.
            final int tfl = (image == null) ? -1 : image.length;

//System.err.println(" makeThumbnailImage(): image size = "+tfl+" at quality: " + qual);

            if(image != null)
                {
                // If image size is close to target, return image immediately.
                if((tfl >= targetMin) && (tfl <= targetMax))
                    {
                    // We have an acceptable size; return it!
//System.err.println(" makeThumbnailImage(): done: within tolerance at quality: " + qual);
                    return(image);
                    }

                // If target is within absolute limits,
                // then remember it iff it is the best so far.
                if((tfl >= absMinSize) && (tfl <= absMaxSize))
                    {
                    if((bestImageSoFar == null) ||
                        (Math.abs(image.length - targetBytes) <
                         Math.abs(bestImageSoFar.length - targetBytes)))
                        {
                        bestImageSoFar = image;
//System.err.println(" makeThumbnailImage(): best so far at quality: " + qual);
                        }
                    }
                }

            // Do we have to bump quality (and size) down?
            final boolean goDown = (tfl > absMaxSize);
//System.err.println(" makeThumbnailImage(): goDown = "+goDown+" with quality: " + qual);

            // Round again...
            // Adjust the bounds and the new value to test.
            int oldQual = qual;
            qual = goDown ? ((oldQual-1 + lowerQualityBound) / 2)
                          : ((oldQual+2 + upperQualityBound) / 2);
            if(goDown) { upperQualityBound = Math.min(upperQualityBound, oldQual - 1); }
            else       { lowerQualityBound = Math.max(lowerQualityBound, oldQual + 1); }
            }

//if(bestImageSoFar == null) { System.err.println(" makeThumbnailImage(): failed"); }
//else { System.err.println(" makeThumbnailImage(): done"); }

        // If we got something acceptable, return it, else null.
        return(bestImageSoFar);
        }


    /**Estimates bytes of working memory required to create thumbnails for an image.
     * This includes the space to load the old image,
     * plus one extra working copy of the original image,
     * make one of the thumbnails at a time,
     * and hold both the standard and small thumbnails in binary format.
     * <p>
     * This assumes still images amongst other things;
     * for movies or animations this may, depending on the processing
     * method, need to be multiplied by the number of frames, for example.
     * <p>
     * Fudge factors such as guessed overheads, are liberally sprinkled in.
     * <p>
     * This assumes that the JVM and the codecs are reasonably efficient
     * in their use of memory, but this is at best a wild guess.
     * <p>
     * A sophisticated handler may need to override this.
     * <p>
     * Should not be called if getThumbnailParams() returns null.
     *
     * @param xyDim  pixel dimensions of original image; must be non-null
     * @return estimated working memory in heap in bytes
     */
    protected int estimateWorkingMemoryToCreateThumbnails(final Dimension xyDim)
        {
        if((xyDim == null) || (xyDim.width < 1) || (xyDim.height < 1))
            { throw new IllegalArgumentException(); }

        final ThumbnailParams thumbnailParams = getThumbnailParams();
        // Make a conservative guess at bytes-per-pixel if need be.
        final int rawBytesPerPixel = (thumbnailParams == null) ? 4 :
            thumbnailParams.estimatedBytesPerImagePixelInMemory;
        if(rawBytesPerPixel <= 0)
            { throw new IllegalArgumentException(); }

        // This will give us the dimensions of the larger of the thumbnails
        // and thus the larger of the two images to be created in memory
        // before being converted to (compressed) file format.
        final Dimension stdTnXyDim =
            ExhibitThumbnails.computeThumbnailDimensions(xyDim, true);

        // Guess at overhead in memory be Image instance for
        // colour maps, encoding parameters, etc.
        final int fixedImageOverhead = 10240;

        // Estimated memory (bytes) used to load original image into memory,
        // plus an additional working copy.
        final int origImageMemory = fixedImageOverhead +
            2 * (1+rawBytesPerPixel) * xyDim.width * xyDim.height;

        // Estimated memory (bytes) used to make one of the thumbnails in memory
        // (assumed to be the larger, ie standard one).
        final int maxThumbnailImageMemory = fixedImageOverhead +
            (1+rawBytesPerPixel) * stdTnXyDim.width * stdTnXyDim.height;

        // Estimated memory (bytes) used to hold the final ExhibitThumbnails object,
        // including the two binary file-format thumbnail images.
        final int binaryThumbnailFormatMemory =
            ExhibitThumbnails.STD_ABS_MAX_BYTES * 2;

        return(origImageMemory +
               maxThumbnailImageMemory +
               binaryThumbnailFormatMemory);
        }


    /**Decode image using generic image IO routines; null if cannot be done.
     * It should already have been established that the image is of the correct type.
     * <p>
     * Individual handlers may override this default (JAI) implementation if desired.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * Calls to this implementation may be serialised
     * because they can be very memory-intensive.
     */
    public BufferedImage decodeImage(final InputStream is)
        throws IOException
        {
//        synchronized(_mem_lock)
            { return(javax.imageio.ImageIO.read(is)); }
        }

    /**Get dimensions X and Y of an image by decoding entire image, else null if dimensions cannot be computed.
     * This may be extremely slow and inefficient and should be overridden by handlers
     * if at all possible.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * Calls to this implementation may be serialised
     * because they may be very memory-intensive.
     *
     * @param is        the exhibit as a binary data stream
     *
     * @throws IOException  in case of problems with corrupt data
     *     (or a broken exhibit)
     */
    public Dimension get2DImageDimensions(final InputStream is)
        throws IOException
        {
//        synchronized(_mem_lock)
            {
            final BufferedImage bi = decodeImage(is);
            if(bi == null) { return(null); }
            return(new Dimension(bi.getWidth(), bi.getHeight()));
            }
        }


    /**If true, only get "standard" (lossy) metadata to save space. */
    private static final boolean ONLY_GET_STD_METADATA = true;

    /**Standard metadata format name.
     * This seems to be rather badly hidden in the API: yeuck!
     */
    private static final String STD_METADATA_FORMAT_NAME = IIOMetadataFormatImpl.standardMetadataFormatName;

    /**Get specific (image) metadata for one particular image type; null if none.
     * This does not have to be provided, but if so enables such things as
     * EXIF and IPTC extraction from JPEG images.
     * <p>
     * This is passed the metadata for the first available image.
     * <p>
     * This may be a mix of stream and image/frame metadata.
     * 
     * @param exhibitName  full exhibit name for diagnostics; never null
     */
    protected Node extractSpecificImageMetaData(final IIOMetadata imageMetadata, String exhibitName)
        { return(null); }

    /**If true, trim bulky and not-very-useful metadata (eg palette entries). */
    private static final boolean TRIM_METADATA = true;

    /**Maximum number of children of a single node before culling/trimming it; strictly positive. */
    private static final int TRIM_CHILDREN_MAX = 127;

    /**Removes bulky and not-very-useful metadata sub-trees from below the given node.
     * This removes the following:
     * <ul>
     * <li>Any "Palette" subtree.
     * <li>Any hugely-branched subtree (a very large number of children of one node).
     *     (This may be reported on System.out.)
     * </ul>
     * @param exhibitName non-null full exhibit name for diagnostics
     */
    protected void _trimMetadata(final Node n, final String exhibitName)
        {
        if(n == null) { return; } // Ignore missing node.

        for(Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
            {
            final String childName = child.getNodeName();
            final int size = child.getChildNodes().getLength();

            // Remove bulky palette sub-tree...
            if("Palette".equals(childName))
                {
//System.out.println("[INFO: AbstractImageHandler._trimMetadata(): removed Palette node of size "+size+" from "+exhibitName+".]");
                n.removeChild(child);
                continue;
                }

            // Trim very large nodes (with many children).
            if(size > TRIM_CHILDREN_MAX)
                {
//System.out.println("[INFO: AbstractImageHandler._trimMetadata(): removed (large) '"+childName+"' node of size "+size+" from "+exhibitName+".]");
                n.removeChild(child);
                continue;                
                }

            // Recursively trim child nodes as needed.
            _trimMetadata(child, exhibitName);
            }
        }

//    /**Lock to serialise assumed-memory-intensive operations. */
//    private static final Object _mem_lock = new Object();

    /**Gets all available exhibit metadata as a single XML DOM tree; null if none.
     * We do not (yet) have an XML schema for this and may never do so,
     * since its structure and content will depend on various external sources
     * and parts of the ImageIO and JAI subsystems and others.
     * <p>
     * This tries to fetch any stream data and any metadata from the first available image.
     * <p>
     * Note that this has to turn off the ImageIO file cache to avoid failing
     * since there seems to be no cache clearing.
     * <p>
     * By default we trim some of the most bulky and least useful data from the metadata,
     * mainly the palette information for indexed images.
     * <p>
     * This implementation is serialised as it may be very memory-intensive.
     *
     * @param is  input stream containing exhibit data; never null
     * @param exhibitName  full exhibit name (primarily to help with debugging); never null
     * @return top-level node "metadata" with captured metadata beneath, else null
     */
    public Node getMetadata(final InputStream is,
                            final String exhibitName)
        {
//        synchronized(_mem_lock)
            {
            Node result = null;

            // We have problems running out of space on disc, etc, without the ImageIO cache disabled.
            ImageIO.setUseCache(false);

            try
                {
                final ImageInputStream iis = ImageIO.createImageInputStream(is);
                if(iis == null) { return(null); }

                try
                    {
                    // Try and find a reader for this image type (and use the first available).
                    final Iterator<ImageReader> readerIt =
                            ImageIO.getImageReadersByMIMEType(getExhibitType().mimeType);
                    // If no reader, give up.
                    if(!readerIt.hasNext()) { return(null); }

                    final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    final DocumentBuilder db = dbf.newDocumentBuilder();
                    final Document doc = db.newDocument();

                    // Create a provisional root node,
                    // but only assign it to the result (ie to get a non-null result)
                    // when we know that there is some real data in there.
                    final Element root = doc.createElement(TAG_NAME_METADATA_TOP);
                    doc.appendChild(root);

                    final ImageReader ir = readerIt.next();
                    try
                        {
                        ir.setInput(iis);

                        // Try to extract (stream) metadata, if any.
                        final IIOMetadata mdStream = ir.getStreamMetadata();

                        if((mdStream != null) &&
                                (!ONLY_GET_STD_METADATA || mdStream.isStandardMetadataFormatSupported()))
                            {
                            final String[] mdFormatNames = !ONLY_GET_STD_METADATA ? mdStream.getMetadataFormatNames() :
                                                           new String[]{ STD_METADATA_FORMAT_NAME };
                            if(mdFormatNames != null)
                                {
                                final Element item = doc.createElement("stream");
                                root.appendChild(item);
                                for(final String formatName : mdFormatNames)
                                    {
                                    TextUtils.importCopy(item, mdStream.getAsTree(formatName));
                                    if(TRIM_METADATA) { _trimMetadata(item, exhibitName); }
                                    }
//                            { item.appendChild(doc.importNode(mdImage.getAsTree(formatName), true)); }
                                result = root; // We have some real metadata.
                                }
                            }

                        // Try to extract metadata, if any, from first image.
                        final IIOMetadata mdImage = ir.getImageMetadata(ir.getMinIndex());

                        if((mdImage != null) &&
                                (!ONLY_GET_STD_METADATA || mdImage.isStandardMetadataFormatSupported()))
                            {
                            final String[] mdFormatNames = !ONLY_GET_STD_METADATA ? mdImage.getMetadataFormatNames() :
                                                           new String[]{ STD_METADATA_FORMAT_NAME };
                            if(mdFormatNames != null)
                                {
                                final Element item = doc.createElement("image");
                                root.appendChild(item);
                                for(final String formatName : mdFormatNames)
                                    {
                                    TextUtils.importCopy(item, mdImage.getAsTree(formatName));
                                    if(TRIM_METADATA) { _trimMetadata(item, exhibitName); }
                                    }
//                            { item.appendChild(doc.importNode(mdImage.getAsTree(formatName), true)); }
                                result = root; // We have some real metadata.
                                }
                            }

                        // Try to extract additional format-specific metadata (eg EXIF from JPEG).
                        if(mdImage != null)
                            {
                            final Node n = extractSpecificImageMetaData(mdImage, exhibitName);
                            if(n != null)
                                { TextUtils.importCopy(root, n); }
                            }
                        }
                    finally
                        { ir.dispose(); } // Free (non-JVM) resources ASAP.
                    }
                finally
                    { iis.close(); } // Free (non-JVM) resources ASAP.
                }
            catch(final IOException e)
                {
                e.printStackTrace();
                return(null); // Cannot read stream...
                }
            catch(final ParserConfigurationException e)
                {
                e.printStackTrace();
                return(null); // Cannot build document...
                }

//System.err.println("getMetadata(): " + TextUtils.toXML(result, false, true));
            return(result);
            }
        }

    /**Gets an encoder/writer for this image type if one is available, else returns null.
     * Uses IIO facilities to locate an available writer.
     */
    protected ImageWriter _getEncoder()
        {
        final Iterator<ImageWriter> wit =
            ImageIO.getImageWritersByMIMEType(getExhibitType().mimeType);

        // Return null if no encoder available.
        if((wit == null) || !wit.hasNext())
            { return(null); }

        // Return first writer/encoder.
        final ImageWriter result = wit.next();
        return(result);
        }

    /**Claim that we can make thumbnails for this image type if we can find an encoder and parameters at run-time.
     */
    public boolean canMakeThumbnails()
        { return((getThumbnailParams() != null) && (_getEncoder() != null)); }

//    /**Encode the given BufferedImage for this image type in a simple way (using IIO); never null.
//     * @param imageIn  must have colour space (etc) compatable with image type; non-null
//     * @return the encoded file-format byte array
//     * @throws java.io.IOException  in case of difficulty with the encoding
//     */
//    protected byte[] _makeImageBinarySimple(final BufferedImage imageIn)
//        throws IOException
//        {
//        final ImageWriter wr = _getEncoder();
//
//        // If no writer available then give up immediately.
//        if(wr == null) { return(null); }
//        
//        final ImageWriteParam writeParam = wr.getDefaultWriteParam();
//        if(writeParam != null) { writeParam.setCompressionQuality(quality); }
//
//        try
//            {
//            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
//            ImageIO.setUseCache(false); // Avoid using the (disc) cache.
//            final ImageOutputStream ios = ImageIO.createImageOutputStream(baos);
//            if(ios == null) { return(null); }
//
//            try
//                {
//                wr.setOutput(ios);
//                wr.write(null, new IIOImage(imageIn, null, null), writeParam);
//                }
//            finally
//                { ios.close(); }
//
//            final byte[] result = baos.toByteArray();
//
//            // Do not allow zero-length return value.
//            if(result.length == 0)
//                { throw new IOException("generated zero-length result"); }
//
//            return result;
//            }
//        finally
//            { wr.dispose(); }
//        }
    
    /**Guide colour-reduction to reduce quality and size of lossless image formats.
     * Defaults to -1, implying that the image format is a lossy format
     * for which the ImageWriteParam.setCompressionQuality() mechanism
     * is enough to trade encoded image size for quality.
     * <p>
     * Else this is a positive value n, typically 8 or 24,
     * which sets a quality level at/below which the number of colours in the image
     * is reduced to (capped at) 2^n, and the image representation may be tweaked too
     * (eg converted to a palette/indexed format) as a strong hint to the encoder.
     */
    protected int _reduceColoursQualityThreshold() { return(-1); }

    /**Creates output file-formatted; null for a permanaent failure.
     * Can return null for a permanent error handling a particular image,
     * eg in the ImageIO GIF/PNG drivers for some palette sizes.
     *
     * @param imageIn  single image to encode, never null
     * @param quality  desired encoding quality in range 0--100 inclusive,
     *     with 0 interpreted as "high compression is important,"
     *     and 100 (or over) interpreted as "high image quality is important.";
     *     this parameter is ignored where it does not make sense for the format
     */
    public byte[] makeImageBinary(BufferedImage imageIn, int quality)
        throws IOException
        {
        if(imageIn == null)
            { throw new IllegalArgumentException(); }
        if(quality < 0)
            { throw new IllegalArgumentException("quality param must be in range [0--100]"); }

        /**Coerce "quality" into range if need be. */
        if(quality > 100)
            { quality = 100; }
        
        final ImageWriter wr = _getEncoder();

        // If no writer available then give up immediately.
        if(wr == null) { return(null); }
        
        final ImageWriteParam writeParam = wr.getDefaultWriteParam();
        try
            {
            if(writeParam != null)
                {
                try { writeParam.setCompressionQuality(quality / 100.0f); }
                catch(final Exception e) { /* Quietly ignore where this is not allowed/understood. */ }
                }

            // If appropriate then try colour reduction
            // and conversion to palette/indexed form.
            final int threshold = _reduceColoursQualityThreshold();
            if(quality <= threshold)
                {
                // Chose a maximum number of colours that is a power of two
                // to try to maximise the encoding efficiency.
                final int maxColours = (1 << (Math.min(30, Math.max(1, quality))));

//    System.out.println("quality-reduced encoding (quality = "+quality+"): maxColours = " + maxColours);

                // Force the output to a byte-indexed model to
                // maximise chance of compression
                // if less than 8 bits.
                final boolean forceToByteIndexModel = (quality <= 8);
                imageIn = ImageUtils.makeColourReducedBufferedImage(imageIn,
                                                              maxColours,
                                                              forceToByteIndexModel);
                }

            // Now generate the binary/file encoded image.
            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ImageIO.setUseCache(false); // Avoid using the (disc) cache.
            final ImageOutputStream ios = ImageIO.createImageOutputStream(baos);
            if(ios == null) { return(null); }

            try
                {
                wr.setOutput(ios);
                wr.write(null, new IIOImage(imageIn, null, null), writeParam);
                }
            finally
                { ios.close(); }

            final byte[] result = baos.toByteArray();

            // Do not allow zero-length return value.
            if(result.length == 0)
                { throw new IOException("generated zero-length result"); }

            return(result);
            }
        catch(final IllegalArgumentException e)
            {
            System.err.println("ERROR: could not makeImageBinary(): returning null to indicate permanent error for supplied parameters");
            e.printStackTrace();
            return(null); /* Permanent error. */
            }
        finally
            { wr.dispose(); }
        }
    }
