/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.MIME;

import org.hd.d.pg2k.svrCore.TextUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

/**Base class with useful default behaviour for (sampled) sound exhibit media-handler classes.
 * By default this uses javax.sound.* routines to extract sound and metadata from
 * an exhibit stream.
 * <p>
 * Various routines can be overridden in deriving classes to change this behaviour.
 */
public abstract class AbstractSampledSoundHandler extends AbstractHandler
    {
    /**Get format of audio data; null if none or format not usable/parseable.
     * This may require the stream to be seekable, at least with mark() and reset().
     */
    protected AudioFileFormat getAudioFileFormat(final InputStream is)
        {
        try { return(AudioSystem.getAudioFileFormat(is)); }

        // Dump any (expected) errors to the log/console and return null.
        catch(final UnsupportedAudioFileException e)
            {
//            e.printStackTrace();
            }
        catch(final IOException e)
            {
//            e.printStackTrace();
            }

        return(null); // Could not deduce file format.
        }

    /**Gets all available exhibit metadata as a single XML DOM tree; null if none.
     * We do not (yet) have a schema for this and may never do so,
     * since its structure and content will depend on various external sources
     * and parts of (for example) the ImageIO and JAI subsystems.
     * <p>
     * This implementation synthesises a metadata tree from the AudioFileFormat data,
     * if any.
     *
     * @param is input stream; never null
     * @return top-level node "metadata" with captured metadata beneath, else null
     */
    public Node getMetadata(final InputStream is, String exhibitName)
        {
        final AudioFileFormat aff = getAudioFileFormat(is);
        if(aff == null) { return(null); }

        final AudioFormat af = aff.getFormat();
        if(af == null) { return(null); }

        // Result: will be set non-null if we encounter any usable data.
        Node result = null;

        try
            {
            final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = dbf.newDocumentBuilder();
            final Document doc = db.newDocument();

            // Create a provisional root node,
            // but only assign it to the result (ie to get a non-null result)
            // when we know that there is some real data in there.
            final Element root = doc.createElement(TAG_NAME_METADATA_TOP);
            doc.appendChild(root);

            // The "stream" node contains data about the file format.
            final Element stream = doc.createElement("stream");
            root.appendChild(stream);

            final int nFrames = aff.getFrameLength();
            if(nFrames != AudioSystem.NOT_SPECIFIED)
                {
                final Element framesNode = doc.createElement("frames");
                framesNode.setAttribute("value", String.valueOf(nFrames));
                stream.appendChild(framesNode);
                result = root; // Got some real data...
                }

            final Map<String,Object> sPropMap = aff.properties();
            if((sPropMap != null) && (!sPropMap.isEmpty()))
                {
                final Element propsNode = doc.createElement("properties");
                // Insert all the properties.
                for(final String key : sPropMap.keySet())
                    {
                    final Element propNode = doc.createElement("property");
                    propNode.setAttribute("name", TextUtils.escapeHTMLMetaChars(key));
                    propNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(String.valueOf(sPropMap.get(key))));
                    propsNode.appendChild(propNode);
                    }
                stream.appendChild(propsNode);
                result = root; // Got some real data...
                }

            // The "audio" node contains data about the data encoding format such as rate.
            final Element audio = doc.createElement("audio");
            root.appendChild(audio);

            final AudioFormat.Encoding encoding = af.getEncoding();
            if(encoding != null)
                {
                final Element encodingNode = doc.createElement("encoding");
                encodingNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(encoding.toString()));
                audio.appendChild(encodingNode);
                result = root; // Got some real data...
                }

            final float sampleRate = af.getSampleRate();
            if(sampleRate != (float) AudioSystem.NOT_SPECIFIED)
                {
                final Element sampleRateNode = doc.createElement("sampleRate");
                sampleRateNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(String.valueOf(sampleRate)));
                sampleRateNode.setAttribute("unit", "Hz");
                audio.appendChild(sampleRateNode);
                result = root; // Got some real data...
                }

            final int sampleSizeInBits = af.getSampleSizeInBits();
            if(sampleSizeInBits != AudioSystem.NOT_SPECIFIED)
                {
                final Element sampleSizeNode = doc.createElement("sampleSize");
                sampleSizeNode.setAttribute("value", String.valueOf(sampleSizeInBits));
                sampleSizeNode.setAttribute("unit", "bits/sample");
                audio.appendChild(sampleSizeNode);
                result = root; // Got some real data...
                }

            final int channels = af.getChannels();
            if(channels != AudioSystem.NOT_SPECIFIED)
                {
                final Element channelsNode = doc.createElement("channels");
                channelsNode.setAttribute("value", String.valueOf(channels));
                audio.appendChild(channelsNode);
                result = root; // Got some real data...
                }

            final int frameSize = af.getFrameSize();
            if(frameSize != AudioSystem.NOT_SPECIFIED)
                {
                final Element frameSizeNode = doc.createElement("frameSize");
                frameSizeNode.setAttribute("value", String.valueOf(frameSize));
                frameSizeNode.setAttribute("unit", "bytes/frame");
                audio.appendChild(frameSizeNode);
                result = root; // Got some real data...
                }

            final float frameRate = af.getFrameRate();
            if((Math.abs(sampleRate - frameRate) > 0.00001) &&
               (frameRate != (float) AudioSystem.NOT_SPECIFIED))
                {
                final Element frameRateNode = doc.createElement("frameRate");
                frameRateNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(String.valueOf(frameRate)));
                frameRateNode.setAttribute("unit", "frames/second");
                audio.appendChild(frameRateNode);
                result = root; // Got some real data...
                }

            if((encoding.equals(AudioFormat.Encoding.PCM_SIGNED)
                    || encoding.equals(AudioFormat.Encoding.PCM_UNSIGNED)) &&
                 (sampleSizeInBits > 8)
                    || (sampleSizeInBits == AudioSystem.NOT_SPECIFIED))
                {
                final Element endianNode = doc.createElement("bigEndian");
                endianNode.setAttribute("value", String.valueOf(af.isBigEndian()));
                audio.appendChild(endianNode);
                result = root; // Got some real data...
                }

            final Map<String,Object> aPropMap = af.properties();
            if((aPropMap != null) && (!aPropMap.isEmpty()))
                {
                final Element propsNode = doc.createElement("properties");
                // Insert all the properties.
                for(final String key : aPropMap.keySet())
                    {
                    final Element propNode = doc.createElement("property");
                    propNode.setAttribute("name", TextUtils.escapeHTMLMetaChars(key));
                    propNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(String.valueOf(aPropMap.get(key))));
                    propsNode.appendChild(propNode);
                    }
                audio.appendChild(propsNode);
                result = root; // Got some real data...
                }
            }
        catch(final ParserConfigurationException e)
            {
            e.printStackTrace();
            return(null); // Cannot build document...
            }

        return(result);
        }
    }
