/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.MIME;

import org.hd.d.pg2k.svrCore.FileTools;

import java.io.*;
import java.util.*;

/**Routines to establish/check exhibit MIME types.
 * This is hardwired into the Gallery for now; we don't allow dynamic
 * loading of new types and handlers...
 * <p>
 * This is a rewrite and extension of the old (pre-PG2K) PG Attributes class.
 */
public final class ExhibitMIME
    {
    /**Package for default media handlers. */
    public static final String DEFAULT_HANDLER_PACKAGE =
        "org.hd.d.pg2k.svrCore.mediahandler";

    /**JavaBean outlining file details of one Gallery exhibit item type.
     * This object is immutable.
     * <p>
     * Not designed to be persisted, so not Serializable.
     * <p>
     * Note that this has no public no-arg constructor because
     * these objects are all created statically and we create
     * all the instances that should be created.
     * <p>
     * Note that this class is based on the assumption that we
     * recognise exactly one filename suffix for each type of
     * file archived in the Gallery, and when generating something
     * of that type (usually) use one, different, suffix.  Both
     * suffixes are chose to be generally acceptable to most tools
     * such as browsers.
     * <p>
     * These suffixes presented are intern()ed so that the number
     * of copies of them kicking around can be reduced.  Don't go
     * casually making up lots of random filetypes...
     */
    public static final class ExhibitTypeParameters implements Comparable
        {
        /**Make a new MIME type descriptor.
         * Private constructor and object not (de)serialisable
         * so as to ensure that only we can make instances here at run-time.
         *
         * @param _handlerClassName  the fully-qualified class name
         *     of the handler for this MIME type; if null a default
         *     of DEFAULT_HANDLER_PACKAGE package with the class name
         *     the same as the primary (all-lower-case) suffix
         *     is used instead
         */
        private ExhibitTypeParameters(final int _type,
                                      final String _suffixForInputFile,
                                      final String _mimeType,
                                      final String _handlerClassName,
                                      final String _magic,
                                      final String _description)
            {
            this(_type, _suffixForInputFile, null, _mimeType, _handlerClassName, _magic, null, _description);
            }

        /**Make a new MIME type descriptor.
         * Private constructor and object not (de)serialisable
         * so as to ensure that only we can make instances here
         * and at run-time.
         *
         * @param _allSuffixes  null or a set of (dotless) extensions
         *     (possibly including the primary one)
         *     by which we may recognise input files of this type
         * @param _allMagics  null or a set of magix numbers/strings
         *     (possibly including the primary one if not "")
         *     by which we may recognise input files of this type
         *
         * @param _handlerClassName  the fully-qualified class name
         *     of the handler for this MIME type; if null a default
         *     of DEFAULT_HANDLER_PACKAGE package with the class name
         *     the same as the primary (all-lower-case) suffix
         *     is used instead
         */
        private ExhibitTypeParameters(final int _type,
                                      final String _suffixForInputFile,
                                      final String _allSuffixes[],
                                      final String _mimeType,
                                      String _handlerClassName,
                                      final String _magic,
                                      final String _allMagics[],
                                      final String _description)
            {
            if((_type < ET__min) || (_type > ET__max) ||
               (_suffixForInputFile == null) ||
               (_suffixForInputFile.length() < 1) ||
               (_suffixForInputFile.startsWith(".")) ||
               (_mimeType == null) ||
               (_mimeType.length() < 3) || // Must be at least "a/b"...
               (_mimeType.indexOf('/') < 1) || // Must be at least "a/b"...
               (!_mimeType.equals(_mimeType.toLowerCase())) || // Must be lower-case.
               (_magic == null) ||
               ("".equals(_handlerClassName)) || // Must not be empty string.
               (_description == null))
                {
                throw new IllegalArgumentException();
                }
            type = _type;
            suffixForInputFile = _suffixForInputFile.intern();
            dotSuffixForInputFile = ("." + _suffixForInputFile).intern();
            mimeType = _mimeType;
            magic = _magic;
            description = _description;

            // Set up the suffixes list...
            if(_allSuffixes == null)
                {
                allSuffixesForFile = Collections.singleton(suffixForInputFile);
                }
            else
                {
                final Set<String> s = new HashSet<String>(2 + _allSuffixes.length*2);
                s.add(suffixForInputFile); // Ensure that the primary value is present.
                for(int i =  _allSuffixes.length; --i >= 0; )
                    {
                    final String t = _allSuffixes[i];
                    if((t == null) || (t.length() < 1) || (t.startsWith(".")))
                        { throw new IllegalArgumentException(); }
                    s.add(t);
                    }
                allSuffixesForFile = Collections.unmodifiableSet(s);
                }

            // Set up the magics list.
            if(_allMagics == null)
                {
                if(magic.length() == 0)
                    { allMagics = Collections.emptySet(); }
                else
                    { allMagics = Collections.singleton(magic); }
                }
            else
                {
                final Set<String> s = new HashSet<String>(2 + _allMagics.length*2);
                s.add(magic); // Ensure that the primary value is present.
                for(int i =  _allMagics.length; --i >= 0; )
                    {
                    final String t = _allMagics[i];
                    if((t == null) || (t.length() < 1))
                        { throw new IllegalArgumentException(); }
                    s.add(t);
                    }
                allMagics = Collections.unmodifiableSet(s);
                }

            // Iff no explicit handler has been supplied,
            // and the default one appears to at least exist,
            // use the default one.
            if(_handlerClassName == null)
                {
                // Compute the name for the default handler class for this
                // MIME type.
                final String defaultHandlerClassName =
                    DEFAULT_HANDLER_PACKAGE + dotSuffixForInputFile.toLowerCase();
                try {
                    // Just see if the default handler exists;
                    // report more subtle errors later.
                    Class.forName(defaultHandlerClassName);
                    // OK, seems to exist, so use it.
                    _handlerClassName = defaultHandlerClassName;
                    }
                // Ignore errors, but don't try to use the default handler.
                catch(Exception e) { }
                }

            // Create instance of handler class.
            Handler h = null;
            try {
                if(_handlerClassName != null)
                    {
                    h = (Handler) (Object) ((Class.forName(_handlerClassName)).newInstance());
                    }
                }
            // Note any errors we encounter.
            catch(Exception e)
                {
                System.err.println("ExhibitMIME: ERROR: could not create handler for MIME type "+mimeType+" with class ``"+_handlerClassName+"''.");
                e.printStackTrace();
                }
            // Keep the name and handler null if not successfully instantiated.
            handler = h;
            handlerClassName = (handler != null) ? _handlerClassName : null;
            }


        /**Make meaningful human-readable diagnostic String representation.
         * This is not intended to be machine-parsable,
         * but does not contain whitespace if possible so it might be.
         *
         * @return  human-readable indication of exhibit type.
         */
        public String toString()
            {
            final StringBuilder result = new StringBuilder(48);
            result.append("ExhibitTypeParameters:");
            result.append(suffixForInputFile).append('[');
                result.append(type).append("]:");
            result.append(mimeType);
            return(result.toString());
            }


        /**ET_XXX type of file. */
        public final int type;

        /**Suffix of file if this item is an exhibit in the Gallery; lower-case, never null.
         * This means it is an input item, and should (for example)
         * be read-only and never deleted.
         * <p>
         * This does not include the dot that comes before the extension.
         * <p>
         * This value is never null or zero-length, and is
         * intern()ed.
         */
        public final String suffixForInputFile;

        /**Suffix of input file including leading dot; lower-case, never null.
         * This is suffixForInputFile with a leading dot, and
         * is intern()ed, so is never null and is not less than length 2.
         */
        public final String dotSuffixForInputFile;

        /**All (dotless, lower-case) suffixes for this type, including the primary suffix; never null nor empty.
         * We can recognise all these suffixes on input files
         * as potentially indicating this MIME type.
         * <p>
         * No entries are null or "" or start with (or contain) a dot.
         * <p>
         * This is immutable.
         */
        public final Set<String> allSuffixesForFile;

        /**This is the primary `magic number' at the start of the file; never null though may be "".
         * This is really a byte string, in order, of the leading
         * bytes we must see at the start of a file to be happy
         * it is really the type the file suffix claims it is.
         * If zero-length, this indicates that there is no such
         * initial portion we can check.
         * <p>
         * This enables a simple form of file-type checking
         * like the UNIX ``file'' utility.
         * <p>
         * This string may be zero-length, but it is never
         * null.
         */
        public final String magic;

        /**All legitimate `magic number' values for this type; never null.
         * Contains the primary magic value unless it is "".
         * <p>
         * No entries are null or "".
         * <p>
         * Is empty if no magic number.
         * <p>
         * This is immutable.
         */
        public final Set<String> allMagics;

        /**The canonical MIME type for this exhibit format; never null nor "".
         * Of the form majorType/minorType eg "audio/basic" or "text/html",
         * with all text lower-case.
         */
        public final String mimeType;

        /**Fully-qualified name of handler class (implementing HandlerBase), or null if none.
         */
        public final String handlerClassName;

        /**Instance of handler class, or null if none.
         */
        public final Handler handler;

        /**Textual (English) description of the file type.
         * This describes in human-readable terms what we
         * expect the image to be (it may be that the file
         * extension and magic number would allow a wider range
         * than we describe here).
         * <p>
         * This is intended to be terse enough to go on a Web
         * page with an image file if need be.  It may be zero length,
         * though must not be null.
         */
        public final String description;


        /**Tested for equality and sorted on type. */
        public int hashCode() { return(type); }

        /**Equal if type is equal. */
        public boolean equals(final Object o)
            {
            if(!(o instanceof ExhibitTypeParameters)) { return(false); }
            return(type == ((ExhibitTypeParameters) o).type);
            }

        /**Sort by type. */
        public int compareTo(final Object o)
            { return(type - ((ExhibitTypeParameters) o).type); }


        /**Returns true if we may be able to create a thumbnail/sample with the same MIME type as this.
         * We might fail in practice, but if this returns false we know that
         * we definitely <em>cannot</em> make a thumbnail/sample for this
         * exhibit type in the current execution context, eg without some
         * loadable drivers.
         */
        public boolean canPossiblyCreateThumbnailOfSameMIMEType()
            {
            // Won't be able to do this if we don't have a handler.
            if(handler == null) { return(false); }

            // Won't be able to do this if the handler says it can't.
            if(!handler.canMakeThumbnails()) { return(false); }

            return(true); // OK, maybe we can!
            }
        }

    // All file types are strictly positive.
    // Exisiting values must not be changed as they may be persisted
    // to long-lasting media.
    /**Type of GIF file. */
    public static final int ET_GIF  = 1;
    /**Type of JPEG file. */
    public static final int ET_JPEG = 2;
    /**Type of Sun audio (usually 64kbps 8-bit u-law G.711) file. */
    public static final int ET_AU   = 3;
    /**Type of MPEG-1 Layer-II audio file. */
    public static final int ET_MP2  = 4;
    /**Type of MPEG-2 Layer-III audio file (8kbps and up). */
    public static final int ET_MP3  = 5;
    /**Type of MPEG-4 audio file (2kbps and up). */
    public static final int ET_MP4  = 6;
    /**Type of General MIDI audio file. */
    public static final int ET_MIDI = 7;
    /**Type of WAV audio file. */
    public static final int ET_WAV = 8;
    /**Type of HTML text fragment file. */
    public static final int ET_HTMLFRAG = 9;
    /**Type of MPEG video file. */
    public static final int ET_MPEG = 10;
    /**Type of TIFF image file. */
    public static final int ET_TIFF = 11;
    /**Type of BMP image file. */
    public static final int ET_BMP = 12;
    /**Type of PowerPoint presentation file. */
    public static final int ET_PPT = 13;
    /**Type of RealMedia file. */
    public static final int ET_RM = 14;
    /**Type of RealMedia file. */
    public static final int ET_AVI = 15;
    /**Type of PNG (GIF-replacement) file. */
    public static final int ET_PNG = 16;
    /**Type of MNG (GIF-replacement) file. */
    public static final int ET_MNG = 17;
    /**Type of JPEG-2000 image file. */
    public static final int ET_JP2 = 18;
    /**Type of Flash file. */
    public static final int ET_SWF = 19;
    /**Type of PDF file. */
    public static final int ET_PDF = 20;
    /**Type of 3GPP file. */
    public static final int ET_3GP = 21;
    /**Type of Windows Movie file. */
    public static final int ET_WMV = 22;
    /**Type of Windows Rich Text Format file. */
    public static final int ET_RTF = 23;

    /**Minimum valid ET_XXX value (these values may be sparse).
     * Guaranteed strictly positive.
     */
    public static final int ET__min = ET_GIF;

    /**Maximum valid ET_XXX value (these values may be sparse).
     * Guaranteed to be no less than ET__min.
     */
    public static final int ET__max = ET_RTF;

    /**File types accepted by (and generated by) the Gallery.
     * Not to be handed outside the gallery in this raw state.
     * Not necessarily in numeric order at construction here,
     * but sorted immedieately after in part of the static initialiser.
     */
    private static final ExhibitTypeParameters inputFileTypes[] =
        {
        new ExhibitTypeParameters(ET_GIF,      "gif",  "image/gif",   null, "GIF8", "GIF image"),
        // JPEG/JFIF would be ff d8 ff e0 XX XX 'J' 'F' 'I' 'F',
        // but Mavica images seem to omit this JFIF header.
        new ExhibitTypeParameters(ET_JPEG,     "jpg",  new String[]{"jpeg"}, "image/jpeg",  null, "\u00ff\u00d8\u00ff", null, "JPEG image"),
        new ExhibitTypeParameters(ET_AU,       "au",   "audio/basic", null, ".snd", "Sun audio"),
        // MP2 magic sequence may not be long enough to be unique.
        // See opening od -tx1 dump from an .mp2 file...
        // 0000000 00 00 01 ba 21 00 01 00 01 80 0e 3b 00 00 01 bb
        new ExhibitTypeParameters(ET_MP2,      "mp2",  "audio/mpeg",  null, "\u0000\u0000\u0001\u00ba", "MP2 [MPEG-1 Layer-II] audio"),
        // MP3 magic sequence may not be long enough to be unique.
        // See opening od -tx1 dump from two .mp3 files...
        // 0000000 ff f3 74 54 00 00 0c 01 2d 90 00 00 00 00 a8 02
        // 0000000 ff e3 18 c4 00 0a 28 69 bb 21 46 00 01 00 20 78
        new ExhibitTypeParameters(ET_MP3,      "mp3",  "audio/mpeg",  null, "\u00ff", "MP3 [MPEG-2 Layer-III ISO IS11172-3] audio"),
        //new ExhibitTypeParameters(ET_MP4,      "mp4", "MP4", "", "MPEG-4 audio"),
        // MIDI magic number guessed by looking at a few files...
        new ExhibitTypeParameters(ET_MIDI,     "mid",  "audio/midi",  null, "MThd", "General MIDI"),
        // WAV magic sequence may not be long enough to be unique.
        // See opening od -cv dump from one .wav file...
        // 0000000   R   I   F   F 204 033  \0  \0   W   A   V   E   f   m   t
        new ExhibitTypeParameters(ET_WAV,      "wav",  "audio/x-wav", null, "RIFF", "Wave audio"),
        new ExhibitTypeParameters(ET_HTMLFRAG, "htxt", "text/html",   null, "<", "HTML marked-up text fragment"),
        // Deduced magic from dump (od -tx1).
        // 0000000 00 00 01 ba 21 00 01 00 01 80 03 91 00 00 01 bb
        new ExhibitTypeParameters(ET_MPEG,     "mpg",  "video/mpeg",  null, "\u0000\u0000\u0001\u00ba", "MPEG video"),
        // Deduced magic from dump (od -tx1) and RFC3302...
        // 49 49 2a 00 (II: little endian)
        // 4d 4d 00 2a (MM: big endian)
        new ExhibitTypeParameters(ET_TIFF,     "tif",  new String[]{"tiff"}, "image/tiff",  null, "II*\u0000", new String[]{"MM\u0000*"}, "TIFF image"),
        // Deduced magic from dump (od -tx1).
        // 0000000 42 4d 76 b6 0f 00 00 00 00 00 36 00 00 00 28 00
        new ExhibitTypeParameters(ET_BMP,      "bmp",  "image/bmp",   null, "BM", "BMP image"),
        // Deduced magic from dump (od -tx1).
        // 0000000 d0 cf 11 e0 a1 b1 1a e1 00 00 00 00 00 00 00 00
        new ExhibitTypeParameters(ET_PPT,      "ppt",  "application/powerpoint",  null, "\u00d0\u00cf\u0011\u00e0\u00a1\u00b1\u001a\u00e1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000", "PowerPoint presentation"),
        // 0000000 2e 52 4d 46 00 00 00 12 00 01 00 00 00 00 00 00
        new ExhibitTypeParameters(ET_RM,       "rm",   "application/vnd.rn-realmedia",  null, ".RMF", "RealMedia audio/video"),
        // Magic number as suggest by http://www.jmcgowan.com/avi.html: "RIFF" <4-byte-length> "AVI ".
        // See also http://www.fourcc.org/.
        new ExhibitTypeParameters(ET_AVI,      "avi",  "video/avi",  null, "RIFF", "AVI audio/video"),
        // Magic number from formal tech documentation.
        // See http://www.libpng.org/.
        new ExhibitTypeParameters(ET_PNG,      "png",  "image/png",  null, "\u0089PNG\r\n\u001a\n", "PNG image"),
        // Magic number from formal tech documentation.
        // See http://www.libpng.org/.
        new ExhibitTypeParameters(ET_MNG,      "mng",  "video/x-mng",  null, "\u008aMNG\r\n\u001a\n", "MNG animation/video"),
        // JPEG-2000; 12-byte magic number from spec; also see RFC3745 and jpeg.org: 0000 000C 6A50 2020 0D0A 870A.
        new ExhibitTypeParameters(ET_JP2,      "jp2",  "image/jp2",  null, "\u0000\u0000\u0000\u000cjP  \r\n\u0087\n", "JPEG 2000 image"),
        // Macromedia Flash: magic number as per: http://www.garykessler.net/library/file_sigs.html
        new ExhibitTypeParameters(ET_SWF,      "swf",  null, "application/x-shockwave-flash",  null, "CWS", new String[]{"FWS"}, "Macromedia ShockWave Flash"),
        // Adobe PDF: magic number as per: http://www.garykessler.net/library/file_sigs.html
        new ExhibitTypeParameters(ET_PDF,      "pdf",  "application/pdf",  null, "%PDF", "Adobe Portable Document Format"),
        // See RCF3839 at http://www.faqs.org/rfcs/rfc3839.html
        new ExhibitTypeParameters(ET_3GP,      "3gp",  new String[]{"3gpp"}, "video/3gpp",  null, "", null, "3G mobile audio/video"),
        // Windows Movie.
        // Deduced magic from dump (od -tx1).
        // 0000000 30 26 b2 75 8e 66 cf 11 a6 d9 00 aa 00 62 ce 6c
        new ExhibitTypeParameters(ET_WMV,      "wmv",  null, "video/x-ms-wmv",  null, "0&", null, "Windows Movie"),
        // Windows Rich Text Format.
        new ExhibitTypeParameters(ET_RTF,      "rtf",  null, "application/rtf",  null, "{\\rtf1", null, "Windows Rich Text Format"),
        };

    /**Sort in situ by type as needed. */
    static { Arrays.sort(inputFileTypes); }

    /**Returns a (private) list of all valid exhibit types, in order by type. */
    public static ExhibitTypeParameters[] getAllValidExhibitTypes()
        { return(inputFileTypes.clone()); }

    /**Checks if the extension of the (putative) exhibit name is a recognised input extension.
     * The string should not be null, nor start with a dot.
     * (A null input gives a null return.)
     * <p>
     * This returns null if the extension is not recognised,
     * else the (immutable) ExhibitTypeParameters describing it.
     */
    public static ExhibitTypeParameters isValidInputExhibitNameExtension(
                                                        final String extension)
        {
        if(extension == null) { return(null); } // No extension at all!
        final ExhibitTypeParameters result =
            _iVIFE_cache.get(extension);
        if(result != null) { return(result); } // All is well.
        // Table might not have been initialised...
        synchronized(_iVIFE_cache)
            {
            if(_iVIFE_cache.size() == 0)
                {
                for(int i = inputFileTypes.length; --i >= 0; )
                    {
                    _iVIFE_cache.put(inputFileTypes[i].suffixForInputFile,
                                     inputFileTypes[i]);
                    }
                return(_iVIFE_cache.get(extension));
                }
            }
        return(null); // Failed; not a valid extension.
        }
    /**Private cache of mappings from extension to parameters.
     * Used by isValidInputExhibitNameExtension(), and filled in under
     * protection of its own lock on first use if zero-sized.
     */
    private static final Hashtable<String,ExhibitTypeParameters> _iVIFE_cache =
        new Hashtable<String, ExhibitTypeParameters>(inputFileTypes.length * 2 + 1);

    /**Gets type of an exhibit given its extension.
     * The string passed must not be null, nor start with a dot.
     * <p>
     * This returns null if the extension is not recognised,
     * else the (immutable) ExhibitTypeParameters describing the exhibit type.
     * <p>
     * This checks first by the primary suffix for each exhibit type,
     * but if that fails may check by secondary suffix.
     */
    public static ExhibitTypeParameters getExhibitType(final String extension)
        {
        final ExhibitTypeParameters result =
            _gET_cache.get(extension);
        if(result != null) { return(result); } // All is well.
        // Table might not have been initialised...
        synchronized(_gET_cache)
            {
            if(_gET_cache.size() == 0)
                {
                for(int i = inputFileTypes.length; --i >= 0; )
                    {
                    _gET_cache.put(inputFileTypes[i].suffixForInputFile,
                                   inputFileTypes[i]);
                    }
                return(_gET_cache.get(extension));
                }
            }
        return(null); // Failed; not a valid extension.
        }
    /**Private cache of mappings from extension to parameters.
     * Used by isValidInputExhibitNameExtension(), and filled in under
     * protection of its own lock on first use if zero-sized.
     */
    private static final Hashtable<String,ExhibitTypeParameters> _gET_cache =
        new Hashtable<String, ExhibitTypeParameters>(inputFileTypes.length * 2 + 1);

    /**Verifies the magic number for a given exhibit.
     * This will try to read enough data to check the magic number.
     * <p>
     * Returns false if the magic number check fails.
     * <p>
     * If the file type specified has no unique magic number to
     * be checked this always succeeds.
     * <p>
     * This function does not close the input stream.
     *
     * @param etp  is the exhibit type (must not be null)
     * @param exhibitByteStream  is a stream of bytes from the
     *     start of the exhibit at least long enough to contain the
     *     magic number
     */
    public static boolean magicOK(final ExhibitTypeParameters etp,
                                  final InputStream exhibitByteStream)
        {
        assert((etp != null) && (exhibitByteStream != null));

        // Now check the magic number if any...
        final int numMagics = etp.allMagics.size();
        if(numMagics == 0) { return(true); } // No magic to check...

        // Compute longest magic number/sequence for this file type.
        int magicLength = etp.magic.length();
        if(numMagics > 1)
            {
            for(final String s : etp.allMagics)
                { if(s.length() > magicLength) { magicLength = s.length(); } }
            }

        final DataInputStream dis; // = null;
        try
            {
            // OK, attempt to open the stream and read enough
            // bytes to check the magic...
            // (If the stream is not long enough we expect
            // to get an EOFException thrown.)
            dis = new DataInputStream(exhibitByteStream);
            final byte data[] = new byte[magicLength];
            dis.readFully(data);

            // If any magic matches then we are OK...
            nextMagic: for(final String s : etp.allMagics)
                {
                for(int i = s.length(); --i >= 0; )
                    {
                    if(s.charAt(i) != (0xff & data[i]))
                        { continue nextMagic; } // Match failed...
                    }
                return(true); // Match succeeded on entire magic sequence...
                }
            return(false); // Did not find a magic number to match.
            }
        catch(IOException e) { return(false); } // Failed...
        }

    /**Checks that file has expected magic number; throws IOException if not.
     * @throws InvalidObjectException  if given file does not have
     *     a correct magic number
     * @throws IOException in case of difficulty reading the file
     */
    public static void checkMagicOK(final ExhibitTypeParameters etp, final File file)
        throws IOException,
               InvalidObjectException
        {
        final FileInputStream fis = new FileInputStream(file);
        try {
            if(!ExhibitMIME.magicOK(etp, fis))
                { throw new InvalidObjectException("bad magic number: wrong file type"); }
            }
        finally { fis.close(); }
        }

    /**The longest known magic number in bytes of any exhibit type supported; strictly positive once computed.
     * Volatile to allow access without a lock.
     * <p>
     * Computed on first use.
     */
    private static volatile int longestMagicBytes;

    /**Get longest known magic number in bytes of any exhibit type supported; strictly positive. */
    public static int getLongestMagicBytes()
        {
        if(longestMagicBytes == 0)
            {
            int longest = 0;
            for(final ExhibitTypeParameters etp : inputFileTypes)
                {
                for(final String m : etp.allMagics)
                    {
                    final int mLen = m.length();
                    if(mLen > longest) { longest = mLen; }
                    }
                }
            longestMagicBytes = longest;
            assert(longestMagicBytes > 0);
            }
        return(longestMagicBytes);
        }


    /**Guesses the type of a file/exhibit from its magic number, or null if unrecognisable.
     * This may only be able to check for files that have a unique magic number.
     * <p>
     * The stream may need to be markable to allow various possibilties
     * to be tested.
     */
    public static ExhibitTypeParameters guessTypeFromMagic(final InputStream is)
        throws IOException
        {
        if(is == null) { throw new IllegalArgumentException(); }

        // Read in enough bytes for the longest-known magic number
        // if possible.
        final byte buf[] = new byte[getLongestMagicBytes()];
        int n = 0; // Bytes read so far.
        while(n < buf.length)
            {
            final int r = is.read(buf, n, buf.length - n);
            if(r < 1) { break; }
            n += r;
            }

        for(final ExhibitTypeParameters etp : inputFileTypes)
            {
            if(etp.allMagics.size() == 0)
                { continue; /* No magic number(s) to test for. */ }

            for(final String s : etp.allMagics)
                {
                final int mLen = s.length();
                if(mLen == 0)
                if(mLen > n)
                    { continue; /* File too short to be this type. */ }
                boolean isOK = true;
                for(int i = mLen; --i >= 0; )
                    {
                    // A single mismatched byte means not this type...
                    if(((byte) s.charAt(i)) != buf[i])
                        {
                        isOK = false;
                        break;
                        }
                    }

                // Return immediately if we found a match...
                if(isOK)
                    { return(etp); }
                }
            }

        return(null); // Couldn't recognise it.
        }

    /**Gets the type of a file from its name, or null if unrecognised here.
     * The string passed must not be null, nor start with a dot.
     * <p>
     * This returns null if the extension is not recognised for an exhibit,
     * else returns the (immutable) ExhibitTypeParameters describing it.
     * <p>
     * The extension should be for an exhibit input,
     * though this can be called to quickly check if a filename or URI
     * might be for an exhibit, returning null if not.
     *
     * @param name  name of file or URL/URI
     */
    public static ExhibitTypeParameters getInputFileType(final String name)
        {
        return(isValidInputExhibitNameExtension(FileTools.getExtension(name)));
        }

    /**Given an ET_XXX type, returns the ExhibitTypeParameters.
     * Throws IllegalArgumentException for an illegal type.
     */
    public static synchronized ExhibitTypeParameters getParamsByType(final int type)
        {
        if(_gFPBT_cache == null)
            {
            _gFPBT_cache = new ExhibitTypeParameters[ET__max+1];
            for(int i = inputFileTypes.length; --i >= 0; )
                {
                final ExhibitTypeParameters ifp = inputFileTypes[i];
                _gFPBT_cache[ifp.type] = ifp;
                }
            }
        if((type < ET__min) || (type > ET__max))
            { throw new IllegalArgumentException("bad type"); }
        final ExhibitTypeParameters result = _gFPBT_cache[type];
        if(result == null)
            { throw new IllegalArgumentException("bad type"); }
        return(result);
        }
    /**Private cache of mappings from type to parameters.
     * Constructed on first use by getParamsByType().
     */
    private static ExhibitTypeParameters _gFPBT_cache[];

    /**Get the MIME type of an exhibit (even at the end of filename or URI or URL) from its name; never null nor "".
     */
    public static String getMIMEType(final String name)
        throws IOException
        {
        final String extension = FileTools.getExtension(name);
        final ExhibitTypeParameters etp = getExhibitType(extension);
        if(etp == null) // Can only happen if we stop recognising something already in the db.
            { throw(new IOException("unknown MIME type for ``"+extension+"''")); }
        return(etp.mimeType);
        }
    }
