/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.MIME;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitStaticAttr;
import org.hd.d.pg2k.svrCore.ExhibitThumbnails;
import org.w3c.dom.Node;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 29-Aug-2003
 * Time: 19:36:01
 */

/**Interface for exhibit media-handler classes.
 * The media handlers do such things as make thumbnails/samples,
 * and are loaded by name to allow for run-time configuration
 * and presence/absence of such items as Sun's JPEG support.
 * <p>
 * Implementing classes must have an public non-args descriptor
 * so that we can use Class.forName() to manufacture instances.
 * <p>
 * Implementing classes must be immutable and stateless
 * and thread-safe, so that one instance can be shared between many
 * threads.
 * <p>
 * Both CPU time and memory will be severe run-time constraints on
 * implementing classes, and they must allow for concurrent operation
 * with many other instances or internally reduce concurrency
 * (eg with static locks).
 * <p>
 * This interface represents a ``fat'' API, ie implementing classes may not
 * implement all of the API at run-time but return default/failure
 * values instead.
 */
public interface Handler
    {
    /**Get internal type of ExhibitMIME type; never null. */
    ExhibitMIME.ExhibitTypeParameters getExhibitType();

    /**Get X,Y pixel dimensions of an image exhibit, else null if dimensions cannot be computed.
     * This will always return null for items which do not have a fixed
     * X,Y size in pixels such as for AU sound files for example.
     * <p>
     * For something like a movie in a fixed frame this may be able to
     * return its dimensions, eg to embed in a page.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * This should generally be efficient, by directly reading (opening) bytes
     * of the exhibit passed as an input stream, if at all possible,
     * but generic implementations may be very slow.
     * <p>
     * This will only work correctly if the exhibit is of the correct type,
     * eg its magic number must already have been tested.
     *
     * @param is  the exhibit as a binary data stream
     *
     * @throws java.io.IOException  in case of problems with corrupt data
     *     (or a broken exhibit)
     */
    java.awt.Dimension get2DImageDimensions(InputStream is)
        throws IOException;

    /**Make a byte[]-encoded image of this type constained above and below by size.
     * This uses a binary-chop algorithm to attempt to quickly
     * find the optimal "quality" to make the image at,
     * lower quality values asking the encoder to discard more information,
     * eg by doing colour reduction or other quantisation.
     * <p>
     * This is not possible for all exhibit types.
     *
     * @param upperQualityBound  maximum value of quality to use; non-negative
     * @param lowerQualityBound  minimum value of quality to use; non-negative
     *     and no greater than upperQualityBound
     * @param initialQualityHint  initial suggested quality hint; non-negative,
     *     no greater than upperQualityBound, no less than lowerQualityBound
     * @param scaledImage  the input image to encode
     * @param targetMin  target lower bound, should be higher than absMinSize
     * @param targetMax  target upper bound, should be lower than absMaxSize
     * @param absMinSize  absolute minimum number of bytes
     * @param absMaxSize  absolute maximum number of bytes
     * @param targetBytes  target number of bytes
     *
     * @return encoded image, or null if the image cannot be generated
     *     with given constraints
     *
     * @throws IOException  in case of difficulty generating the image
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public byte[] makeSizeConstrainedEncodedImage(final int lowerQualityBound,
                                               final int initialQualityHint,
                                               final int upperQualityBound,
                                               final BufferedImage scaledImage,
                                               final int targetMin,
                                               final int targetMax,
                                               final int absMinSize,
                                               final int absMaxSize,
                                               final int targetBytes)
        throws IOException,
               IllegalArgumentException;

    /**Get ThumbnailParameters for a particular handler.
     * A handler that does not build thumbnails
     * (ie canMakeThumbnails() returns false)
     * may return null for this, which is the default,
     * but otherwise this should return non-null.
     * <p>
     * This is assumed to be fast, ie to return a fixed static instance
     * for each handler.
     * <p>
     * This is protected, since only the support routines in this base
     * class need access this data.
     */
    public ThumbnailParams getThumbnailParams();

    /**Returns true if handler can make thumbnails for this type.
     * makeThumbnails() has to be able to succeed for some real exhibits of
     * this type,
     * producing thumbnails/samples of the same MIME type
     * (though possibly at much-reduced fidelity),
     * but makeThumbnails() need not succeed for all exhibits of this type.
     * <p>
     * By default, returns false, ie thumbnails cannot be made.
     */
    boolean canMakeThumbnails();

    /**Make thumbnails/samples for the specified exhibit.
     * This may fail with an IOException or return null
     * to indicate that it is currently unable to create
     * thumbnails/samples (though this condition may be temporary).
     * <p>
     * If this wishes to indicate that it cannot ever make one or more
     * thumbnails/samples for a given exhibit then this should return a
     * ExhibitThumbnails object with one or both thumbnails set to null.
     * <p>
     * (If canMakeThumbnails() returns false, this should return null.)
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * This will only work correctly if the exhibit is of the correct type,
     * eg its magic number must already have been tested.
     * <p>
     * This assumes enough memory and other resource is available.
     *
     * @param is  the whole raw image positioned at its start
     * @param originalLength  is the length of the encoded original in bytes;
     *     always positive and must reflect the stream input
     *     and will be constrained to Integer.MAX_VALUE if larger
     * @return thumbnails, else null if not possible now;
     *     ExhibitThumbnails.NO_THUMBNAILS may be returned if it looks
     *     like it never be possible/sensible to make thumbnails for
     *     this exhibit
     */
    ExhibitThumbnails makeThumbnails(InputStream is, long originalLength)
        throws IOException;

    /**Make thumbnails/samples for the specified exhibit.
     * A data source for the exhibit must be supplied,
     * along with all available properties of that exhibit.
     * <p>
     * This may fail with an IOException or return null
     * to indicate that it is currently unable to create
     * thumbnails/samples (though this condition may be temporary).
     * <p>
     * If this wishes to indicate that it cannot make one or more
     * thumbnails/samples for a given exhibit then this should return a
     * ExhibitThumbnails object with one or both thumbnails set to null.
     * <p>
     * (If canMakeThumbnails() returns false, this should return null.)
     * <p>
     * By default, returns null, ie thumbnails cannot be made.
     * <p>
     * This routine regulates memory use, rejecting the attempt
     * to make the thumbnails if it cannot find/reserve sufficient memory
     * using MemoryTools.runMemoryIntensiveOperation().
     * <p>
     * Calls the InputStream version of makeThumbnails().
     *
     * @param unlimitedResources  if true, the generation routine is allowed
     *     to try to use unlimited resources (especially memory)
     * @return thumbnails, else null if not possible now;
     *     ExhibitThumbnails.NO_THUMBNAILS may be returned if it looks
     *     like it never be possible/sensible to make thumbnails for
     *     this exhibit
     */
    ExhibitThumbnails makeThumbnails(
                        ExhibitStaticAttr esa,
                        AllExhibitProperties.ExhibitDataSource eds,
                        AllExhibitProperties aep,
                        boolean unlimitedResources)
        throws IOException;

//    /**Get any embedded text, eg comments, in the exhibit.
//     * This is to extract comments, creator details, etc,
//     * rather than to read the text by OCR, ie this is
//     * quick and unambigious.
//     * <p>
//     * This should avoid reading the entire image into memory
//     * or any other CPU- or memory- intensive operation
//     * if at all possible.
//     * <p>
//     * This <strong>does not close its input stream</strong> when done.
//     * <p>
//     * The file must be of the correct exhibit type or behaviour
//     * is undefined.
//     * <p>
//     * The default is to return an empty array, ie as if no such
//     * embedded text is available.
//     *
//     * @return an array of non-null embedded text strings within
//     *     the exhibit such as copyright notices and comments; never null
//     */
//    String[] getEmbeddedText(InputStream is)
//        throws IOException;

    /**Decode image as BufferedImage, or null if not possible.
     * The input stream must be non-null and of the correct exhibit type
     * else behaviour is undefined.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     */
    BufferedImage decodeImage(InputStream is)
        throws IOException;

    /**Return image as file-format byte array.
     * This should only be used for making an image that is compatable
     * with the output format, eg trying to build a GIF from a true-colour
     * image is unlikely to work.
     * <p>
     * This is used for still images and possible animations
     * or movies where the bounding rectangle and the contained data
     * is scaled to the given bounds.
     * <p>
     * The output image will, if possible, use the same colour scheme
     * and other characteristics as the input, though some optional
     * features that usually consume extra space, such as interlacing,
     * may be disabled.
     * <p>
     * If no output can be generated this returns null,
     * which is the default behaviour.
     * <p>
     * This should adjust the image "quality" with the detail
     * value (adjusted within the bounds supplied by the handler class)
     * to try to tune the output size.  For a lossy encoding format
     * such as JPEG this may be the "quality" factor or compression.
     * For a lossless format the may have to be the number of
     * bits-per-pixel that a colour map is reduced to, for example.
     * <p>
     * The quality parameter usage is format- and implementation-
     * dependent, but 0 will generally be the lowest-quality rendering
     * available and 100 will be the highest, with values in between
     * monotonically increasing, though that does not mean that there
     * will be any distinguishable different in the output with
     * any pair of adjacent (or indeed any) quality values.
     *
     * @param imageIn  source image; must not be null and must be suitable
     *     for the target format
     * @param quality  desired encoding quality in range 0--100 inclusive,
     *     with 0 interpreted as "high compression is important,"
     *     and 100 (or over) interpreted as "high image quality is important.";
     *     this parameter is ignored where it does not make sense for the format
     */
    byte[] makeImageBinary(BufferedImage imageIn, int quality)
        throws IOException;

    /**Some constant parameters and hints for thumbnail building.
     * Each handler that can build thumbnails should define one of these
     * statically and return it from getThumbnailParameters().
     * <p>
     * If canMakeThumbnails() returns true,
     * getThumbnailParameters() must return a non-null value.
     * <p>
     * This exists to help cluttering the base class API with things
     * relevant only to the internal mechanics of this system.
     */
    public static final class ThumbnailParams
        {
        /**Estimated bytes-per-pixel of image in memory; positive.
         * Defaults to 4 assuming RBG/ARGB int format of image.
         */
        public final int estimatedBytesPerImagePixelInMemory;

        /**Minimum usable value of "quality" parameter to makeScaledImage(); non-negative.
         */
        public final int minQuality;

        /**Maximum usable value of "quality" parameter to makeScaledImage().
         * Anything higher than this value is treated as equal to this value
         * and implies maximum possible: thus Integer.MAX_VALUE means
         * maximum possible too.
         * <p>
         * If the same as minQuality this indicates that this is not
         * parametrically variable.
         */
        public final int maxQuality;

        /**Suggested initial quality value.
         * This will lie between minQuality and maxQuality inclusive.
         * <p>
         * Careful choice of this will help reduce the number of
         * redundant image builds done.
         */
        public final int initialQualityHint;

        /**Suggested target bits-per-pixel for a thumbnail; positive.
         * This should ignore any typical fixed overheads.
         * <p>
         * This is a hint to generation routines.
         */
        public final int thumbnailBppHint;

        /**Approximate minimum overhead of format in bytes; non-negative.
         * Rough minimum size in this format including magic numbers,
         * cost of file structure, etc.
         * <p>
         * This is used to help make better estimates of target thumbnail
         * sizes, etc, and need not be (indeed, probably cannot be)
         * exact given that each format may have different representations.
         * <p>
         * Can be zero, though all reasonable implementations should at
         * least default to the size of any leading "magic" number.
         */
        public final int approxMinOverheadBytes;

        /**Construct an instance.
         * Argument values are range-checked as far as practical.
         */
        public ThumbnailParams(final int estimatedBytesPerImagePixelInMemory,
                               final int minQuality,
                               final int maxQuality,
                               final int initialQualityHint,
                               final int thumbnailBppHint,
                               final int approxMinOverheadBytes)
            throws IllegalArgumentException
            {
            if(estimatedBytesPerImagePixelInMemory < 1)
                { throw new IllegalArgumentException("estimatedBytesPerImagePixelInMemory must be positive"); }
            if(minQuality < 0)
                { throw new IllegalArgumentException("minQuality must be non-negative"); }
            if(maxQuality < minQuality)
                { throw new IllegalArgumentException("maxQuality must no less than minQuality"); }
            if((initialQualityHint < minQuality) || (initialQualityHint > maxQuality))
                { throw new IllegalArgumentException("initialQualityHint must no less than minQuality and no greater than maxQuality"); }
            if(thumbnailBppHint < 1)
                { throw new IllegalArgumentException("thumbnailBppHint must be positive"); }
            if(approxMinOverheadBytes < 0)
                { throw new IllegalArgumentException("approxMinOverheadBytes must be non-negative"); }


            this.estimatedBytesPerImagePixelInMemory =
                estimatedBytesPerImagePixelInMemory;
            this.minQuality = minQuality;
            this.maxQuality = maxQuality;
            this.initialQualityHint = initialQualityHint;
            this.thumbnailBppHint = thumbnailBppHint;
            this.approxMinOverheadBytes = approxMinOverheadBytes;
            }
        }


    /**Top-level node name for exhibit meta-data; never null or empty. */
    public static final String TAG_NAME_METADATA_TOP = "metadata";

    /**Gets all available exhibit metadata as a single XML DOM tree; null if none.
     * We do not (yet) have a schema for this and may never do so,
     * since its structure and content will depend on various external sources
     * and parts of (for example) the ImageIO and JAI subsystems.
     * <p>
     * Under extreme circumstances, eg with damaged exhibit files,
     * this may throw Error or other RuntimeException values,
     * though implementations should try to return null rather than do this
     * if the error appears to be permanent.
     *
     * @param is  input stream; never null
     * @param exhibitName TODO
     *
     * @return top-level node "metadata" with captured metadata beneath, else null
     */
    public Node getMetadata(final InputStream is, String exhibitName);
    }
