/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import ORG.hd.d.IsDebug;

import java.lang.ref.WeakReference;
import java.util.*;

/**This contains utilities for memory intensive/sensitive operations.
 * For example, this can veto or partly serialise memory-intensive
 * operations such as image resizing in order to prevent the JVM
 * running out of memory in a resource-constrained environment
 * such as a Web server.
 */
public final class MemoryTools
    {
    /**Prevent creation of instances of this class. */
    private MemoryTools() { }

    /**If true, only allows one memory-intensive operation to run at once.
     * This reduces the chance of blowing up the heap,
     * in return for reducing concurrency on multi-CPU machines.
     * <p>
     * We must not hold a lock to enforce this as doing so may deadlock;
     * we veto any concurrent call immediately.
     */
    private static final boolean NO_MEM_INTENSIVE_CONCURRENCY = true;

    /**Attempt to make sure that the specified amount of memory is available; return true if so.
     * If this cannot be done
     * (eg it is not available, or too soon since the last GC to try to force it)
     * then this returns false.
     * <p>
     * This does not hold any locks.
     *
     * @param  bytes non-negative approximate number of bytes free required
     *
     * @return  true if memory is available, false if not available
     */
    private static boolean _makeMemoryAvailable(final long bytes)
        {
        assert(bytes >= 0);

        final Runtime runtime = Runtime.getRuntime();

        // If the request is less than is currently free
        // then return true immediately.
        // We hope that this (short) path will be the usual one.
        if(bytes < runtime.freeMemory()) { return(true); }

        // If the request is more than could ever be available
        // then return false immediately.
        if(bytes > runtime.maxMemory()) { return(false); }

        // Time before we started any System.gc().
        final long beforeGC = System.currentTimeMillis();

        // Find out if it is long enough since our last call to System.gc().
        //
        // If not then return false immediately to avoid pointlessly thrashing the memory system.
        if(_rMIO_noGCBefore > beforeGC) { return(false); }

        // Loop to free up memory if necessary, if we can.
        // The number of iterations is limited.
        for(int i = _rMIO_MAX_GC_CALLS; --i >= 0; )
            {
            // Clear out any Internable items no longer in use.
            heldInternableCount();

            // OK, try to force some GC.
            System.gc();

            // Force finalize()rs to run while we are at it,
            // since it might release other resources.
            // This may also force us to see more of real GC costs.
            System.runFinalization();

            // Postpone the next use of GC by any other caller.
            final long now = System.currentTimeMillis();
            final long notBefore =
                now + Math.max(MIN_WAIT_AFTER_FAILED_GC,
                               ((now - beforeGC) * _rMIO_MAX_GC_FRAC));
            // Attempt to ensure that we only put the "not-before" time fowards
            // since there may be other concurrent updaters of this value,
            // though it is not vital to get it precisely right
            // (residual races are safe since _rMIO_noGCBefore is volatile and thus accesses are atomic).
            _rMIO_noGCBefore = Math.max(_rMIO_noGCBefore, notBefore);

            // If the request is less than is currently free
            // then return true immediately.
            if(bytes < runtime.freeMemory()) { return(true); }

            // If the request is less than *could* be allocated given maxMemory() and space already in use,
            // then return true immediately.
            // But, we only trust maxMemory() if something less/other than Long.MAX_VALUE.
            final long maxMemory = runtime.maxMemory();
            if((maxMemory != Long.MAX_VALUE) &&
               (bytes < (maxMemory - (runtime.totalMemory() - runtime.freeMemory()))))
                { return(true); }

System.out.println("[MemoryTools: FORCED GC FAILED (attempts left: "+i+") trying to make available "+bytes+" bytes: total GC time: " + (now - beforeGC) + "ms; free/total/max bytes = "+runtime.freeMemory()+'/'+runtime.totalMemory()+'/'+runtime.maxMemory()+".]");
            }

        return(false); // Failed to free up enough memory...
        }

    /**Run (or veto) memory-intensive operation.
     * This is passed a task to be run,
     * with an estimate of the amount of memory required.
     * <p>
     * The task is run, possibly after a delay,
     * if resources seem to be available for it,
     * and this routine returns true.
     * <p>
     * If sufficient resource (ie memory) does not seem to be available,
     * the task is not return and this routine returns false.
     * <p>
     * This routine may attempt to free memory by calling System.gc() one
     * or more times, but because this may be slow and may impact system
     * performance (eg concurrency) badly, this waits a significant
     * multiple of the time taken to perform GC before trying this again,
     * so this routine may refuse to try to free memory for a task even if
     * it would work.
     * <p>
     * This keeps a static variable to track memory use between separate
     * (concurrent) callers and is thus implicitly counting memory shared
     * in this JVM or class loader.  Theis routine tries to be robust enough
     * to work reasonably well even with multiple separate copies of this
     * class in separate class loaders
     * <p>
     * If the unlimitedMemory parameter is true then the task is
     * always run immediately, and can be used for testing and for
     * situations where heap space is freely available, eg the heap
     * may be easily expanded.
     * <p>
     * No checked exception is propagated to the caller,
     * but a non-checked exception may be.
     * <p>
     * This routine holds no lock for its duration to avoid deadlocks,
     * but may "baulk" at (ie veto) concurrent calls to reduce concurrency.
     *
     * @param task  the task to be run
     * @param estimatedMinBytesRequired  estimate of peak number of bytes
     *     of heap that the task will use (overestimate rather than
     *     underestimate this value); must be non-negative
     * @param unlimitedMemory  if true, the task is run immediately
     *     on the assumption that at the moment resources are effectively
     *     unlimited
     * @return true if the task was run,
     *     false if it was not run because of apparent memory shortage
     *     or because too many other memory-intensive tasks are running
     *
     * @throws IllegalArgumentException  if the task is null or
     *     a negative amount of memory is specified
     * @throws IllegalStateException  if too much aggregate memory
     *     is beging requested between all the concurrent tasks
     */
    public static boolean runMemoryIntensiveOperation(
                                        final Runnable task,
                                        final boolean unlimitedMemory,
                                        final int estimatedMinBytesRequired)
        throws IllegalArgumentException,
               IllegalStateException
        {
        if((task == null) ||
           (estimatedMinBytesRequired < 0))
            { throw new IllegalArgumentException(); }

        // Our (possibly inflated) view of the actual memory to set aside.
        // We have to do whole calculation as a long to avoid overflow
        // on extreme values.
        //
        // We ensure that this is positive,
        // which ensures that our view of committed memory
        // will always be positive when at least one
        // memory-intensive task is running.
        final long inflatedMemoryEstimate =
            1 + (estimatedMinBytesRequired * (long) _rMIO_FREE_MEM_MULT);

        // Adjust the committed memory on the way in
        // (and keep a copy of that new value).
        // We will release it at the end of the following try/finally.
        final long committedBytes;
        synchronized(_rMIO_lock)
            {
            if(NO_MEM_INTENSIVE_CONCURRENCY && !unlimitedMemory &&
                (_rMIO_commitedBytes > 0))
                { return(false); } // Fail immediately; another in progress.

            committedBytes =
                (_rMIO_commitedBytes += inflatedMemoryEstimate);
            }
        // Release commited memory at the end of this try block...
        try
            {
            // If committedBytes was so huge that our total has wrapped,
            // then refuse to run this task and undo the change.
            if(committedBytes <= 0)
                { throw new IllegalStateException(); }

            // If we have unlimited memory,
            // run the task immediately.
            if(unlimitedMemory)
                {
                task.run();
                return(true);  // Ran the task!
                }

            // If we can't make the memory available then give up and don't run the task.
            // We allow for memory already committed but possibly not yet allocated
            // (ie in outer calls where this call is nested).
            if(!_makeMemoryAvailable(committedBytes))
                { return(false); }

            // Got the memory: run task!
            // Yeah baby, yeah!
            task.run();
            return(true); // Ran the task.
            }
        finally
            {
            // Release the committed memory for this task on the way out...
            synchronized(_rMIO_lock)
                 { _rMIO_commitedBytes -= inflatedMemoryEstimate; }
            }
        }

//    /**See if there is enough free memory in the heap for runMemoryIntensiveOperation() to run its current task. */
//    private static boolean _rMIO_enoughFreeMemory(final long committedBytes,
//                                             final long inflatedMemoryEstimate)
//        {
//        final Runtime r = Runtime.getRuntime();
//
//        // Compute apparent remaining memory (from the maximum available).
//        // We take the smaller of apparently-available memory
//        // and total memory minus our committed memory
//        // (excluding this call to avoid double-counting).
//        // We do this to allow for huge amounts of memory that
//        // we expect to use but may not actually have consumed from
//        // the heap yet.
//        //
//        // If that yields a value as large as our memory estimate
//        // for the current task, we have enough memory.
//        return(inflatedMemoryEstimate <=
//            Math.min(r.freeMemory(),
//                r.maxMemory() - (committedBytes-inflatedMemoryEstimate)));
////                    r.totalMemory() - (committedBytes-inflatedMemoryEstimate)));
//        }


    /**Bytes of memory ``committed'' to runMemoryIntensiveOperation().
     * Private to runMemoryIntensiveOperation() and accessed by it under
     * the _rMIO_lock lock.
     * <p>
     * Initially zero.
     */
    private static long _rMIO_commitedBytes;

    /**Time before which System.gc() should not be retried; private to _makeMemoryAvailable().
     * Based on how long was spent running System.gc() on a previous
     * attempt, this prevents system resources being significantly
     * impacted by us continually running System.gc().
     * <p>
     * Initial (zero) value allows System.gc() to run immediately the
     * first time.
     * <p>
     * Is volatile to allow it to be read/written safely without a lock.
     */
    private static volatile long _rMIO_noGCBefore;

    /**Private lock for runMemoryIntensiveOperation(). */
    private static final Object _rMIO_lock = new Object();

    /**Maximum fraction of wall-clock time we will spend running GC; strictly positive.
     * If this is N we will spend at most 1/(1+N) of wall-clock time running
     * System.gc().
     * <p>
     * TODO: we might additionally abort if we've spent more than some
     * fixed time, eg 30s, trying to collect garbage already.
     * <p>
     * This is used to limit impact on global system performance.
     * <p>
     * A value between 10 and 100 is probably reasonable.
     */
    private static final int _rMIO_MAX_GC_FRAC = 37;

    /**Minimum time before attempting forced GC after previous attempt (ms); strictly positive.
     * This can save us from churning the heap too hard,
     * or wasting time when System.gc() does nothing at all.
     * <p>
     * This can be a few seconds through to a few minutes.
     */
    private static final int MIN_WAIT_AFTER_FAILED_GC = 50000 + Rnd.fastRnd.nextInt(31013);

    /**Maximum number of times that we will call System.gc() to free memory; strictly positive.
     * This is per call to runMemoryIntensiveOperation();
     * we may do less than this.
     * <p>
     * It may be worth calling System.gc() more than once because it may
     * in fact take several passes in some JVMs to liberate all the
     * heap from several different heap areas and across several threads,
     * but calling it lots of times is probably just a huge drain on system
     * resources or may simply be ignored by the JVM.
     * <p>
     * A value from 1 to 3 inclusive is probably good.
     */
    private static final int _rMIO_MAX_GC_CALLS = 2;

    /**Multiple of excess free memory there should be before attempting memory-intensive task; strictly positive.
     * Bigger numbers are more conservative and reduce the probability
     * of a crash due to running out of memory, but increase the
     * chance of not running such as task when we otherwise could.
     * <p>
     * If we eliminate concurrency between memory-intensive operations
     * then we can be less conservative than we otherwise would be,
     * and indeed can possibly drop this down to 1.
     * <p>
     * A value in the range 1--4 is probably good.
     */
    private static final int _rMIO_FREE_MEM_MULT =
        NO_MEM_INTENSIVE_CONCURRENCY ? 1 : 2;


    /**Request to do a pre-emptive garbage collection while the system is quiet.
     * We may ignore this request (eg if too soon after a previous one)
     * and the system may ignore our attempts to do a GC,
     * or may do it asynchronously, or may do a partial GC.
     * <p>
     * This should only be called when the system seems to be idle,
     * and is likely to idle for (say) a minute or so,
     * to allow us to get in a full GC run while no one is looking,
     * so as we can last as long as possible without an intrusive
     * GC when the system <em>is</em> in use later.
     *
     * @return true if we apparently managed to free up some memory, else false
     */
    public static boolean preemptiveGC()
        {
        // Give our intern() pool a chance to clean up.
        // Should be very quick and may free something.
        heldInternableCount();

        final Runtime runtime = Runtime.getRuntime();

        // Always run finalisers where possible
        // since this should be reasonably quick
        // and may free some orphaned resources.
        runtime.runFinalization();

        // Our target is to have about one quarter of the heap free,
        // which suggests an efficient and relatively-unstressed system.
        final long targetFree = runtime.totalMemory()/4;
        if(runtime.freeMemory() > targetFree) { return(false); }

        // If we did not have our target memory free
        // then push to get more freed up to get some hysteresis and less churn.
        if(!_makeMemoryAvailable(2*targetFree))
            { return(false); } // Didn't manage to do much useful...

        System.out.println("[MemoryTools.preemptiveGC(): free="+
                        TextUtils.sizeAsText(runtime.freeMemory(), true)+"/"+
                        TextUtils.sizeAsText(runtime.totalMemory(), true)+"/"+
                        TextUtils.sizeAsText(runtime.maxMemory(), true)+" "+
                                     "Internable#"+heldInternableCount()+".]");
        return(true);
        }





    /**Marker interface indicating class is suitable for deduping by intern().
     * Any class implementing this interface must be immutable,
     * and must have working equals() and hashCode() methods.
     */
    public static interface Internable { }

    /**Marker interface indicating class is suitable for compacting in memory.
     * Generally the single compact() method should not change the logical state,
     * nor should it consume lots of resources if avoidable
     * (eg lots of memory, or lots of CPUs)
     * and may call Thread.yield() periodically,
     * and by the time it returns all compact()ion activity has completed
     * (ie this will not launch background threads to do the work,
     * though may often be launched from one).
     */
    public static interface Compactable { void compact(); }

    /**Private to intern(): for holding Internable values.
     * Must be accessed only under a lock on the map itself to serialise access
     * as WeakHashMap is not itself thread-safe.
     * <p>
     * We start this off reasonably large,
     * and with a low load factor, for access performance.
     */
    private static final Map<Internable, WeakReference<Internable>> WHM_Internable =
            Collections.synchronizedMap(new WeakHashMap<Internable, WeakReference<Internable>>(65536, 0.5f));

    /**Get current number of Internable items held, and try to purge any dead ones. */
    private static int heldInternableCount()
        {
        synchronized(WHM_Internable)
            { return(WHM_Internable.size()); }
        }

    /**Private to intern() to help note WHM_Internable size increases and to help with housekeeping. */
    private static int lastWHMSizeReported /* = 0 */;

    /**If true, we trust String.intern() to be sensible and to allow GC of unused entries, and thus we delegate handling of String values to it. */
    private static final boolean _USE_STRING_INTERN = true;

    /**Specific overloading of intern() for String to avoid runtime instanceof cost where possible.
     */
    public static String intern(final String s)
        {
        if(s == null) { return(null); }

        // Assume that String.intern() is sensible and efficient and robust
        // and allows GC of no-longer-used items in modern JVMs (1.4+)...
        // Assume that String values are very frequent intern()ees,
        // so check first for these, and handle as a short-cut.
        if(_USE_STRING_INTERN) { return(s.intern()); }

        return(s); // Return unchanged.
        }

    /**Specific overloading of intern() for Internable to avoid runtime instanceof cost where possible.
     */
    @SuppressWarnings("unchecked")
    public static <E extends Internable> E intern(final E o)
        {
        if(o == null) { return(null); }

        // Serialise to protect the (non-thread-safe) Map and
        // to eliminate races so that we get complete canonicalisation.
        synchronized(WHM_Internable)
            {
            final E k = o;

            // If we have an extant entry then return it.
            final WeakReference<E> v = (WeakReference<E>) WHM_Internable.get(k);
            if(v != null)
                {
                final E kOld = v.get();
                if(kOld != null)
                    { return(kOld); } // Existing intern()ed value returned.
                }
            else
                {
                // Insert (and return) new (unique) value just passed in.
                WHM_Internable.put(k, new WeakReference<Internable>(k));

                if(ORG.hd.d.IsDebug.isDebug)
                    {
                    // If table has grown significantly since last report,
                    // then report again.
                    final int thresh = lastWHMSizeReported * 2;
                    final int sizeNow = WHM_Internable.size();
                    if(sizeNow > thresh)
                        {
                        System.out.println("[MemoryTools.intern(): WHM_Internable.size() = " + sizeNow + ".]");
                        lastWHMSizeReported = sizeNow;
                        }
                    }
                }
            }

        return(o); // Return original (unique) instance.
        }

    /**Specific overloading of intern() for Number to avoid runtime instanceof cost where possible.
     */
    @SuppressWarnings("unchecked")
    public static <E extends Number> E intern(final E o)
        {
        if(o == null) { return(null); }

        if(  (o instanceof java.lang.Long) ||
             (o instanceof java.lang.Integer) ||
             (o instanceof java.lang.Short) ||
             (o instanceof java.lang.Float) ||
             (o instanceof java.lang.Double))
            {
            final Number n = (Number) o;

            final class NumberWrapper implements Internable
                {
                /**Our encapsulated immutable Number; never null. */
                final Number num = n;

                /**Hash code as from underlying Number. */
                public final int hashCode() { return(num.hashCode()); }

                /**Equality based on underlying Number instances being equal. */
                public final boolean equals(final Object obj)
                    {
                    if(this == obj) { return(true); }
                    if(!(obj instanceof NumberWrapper)) { return(false); }
                    final NumberWrapper other = (NumberWrapper) obj;
                    return(num.equals(other.num));
                    }
                }

            // Recursively handle the wrapped value as an Internable.
            return((E) (intern(new NumberWrapper())).num);
            }
        // Eliminate duplicate Byte values with JDK1.5+ Byte.valueOf()
        // as an efficient alternative to our generic mechanism.
        else if(o instanceof java.lang.Byte)
            { return((E) Byte.valueOf(((Byte) o).byteValue())); }

        return(o); // Can't handle this unknown Number subtype.
        }

    /**Generic equivalent of String.intern(), for more immutable objects that we care about.
     * Any object which is immutable (and probably final) and has functional
     * equals() and hashcode() methods where only apparently-identical instances
     * are equal can potentially be cached/canonicalised/deduped
     * via this routine.
     * <p>
     * This generic routine is for where we do not know the type until run-time;
     * the compiler may select a better overloading at compile-time otherwise.
     * Thanks to Peter Ah&eacute; of Sun for the suggestion!
     * <p>
     * Note that lots of immutable/final objects are <em>not</em>
     * suitable for canonicalisation with this routine.
     * Caveat programmer.
     * <p>
     * A unique instance is returned of the same type.
     * <p>
     * When the intern method is invoked, if the pool already contains an
     * Object equal to this <code>Object</code> o as determined by
     * the equals(Object) method, then the object from the pool is
     * returned. Otherwise, this <code>Object</code> is added to the
     * pool and a reference to this <code>Object</code> is returned.
     * <p>
     * It follows that for any two Objects <code>s</code> and <code>t</code>,
     * <code>intern(s)&nbsp;==&nbsp;intern(t)</code> is <code>true</code>
     * if and only if <code>s.equals(t)</code> is <code>true</code>.
     * <p>
     * Entries are automatically dropped sometime after the last
     * strong/soft reference to them has gone,
     * possibly on the next call to this routine.
     * <p>
     * Unlike String.intern(), this applies to more than just String elements,
     * and memory is guaranteed to be recoverable from non-referenced items
     * (whereas String.intern() has not always been so and is still
     * not clear in the spec).
     * <p>
     * It is probably not especially useful to put String constants or other
     * static items in here routinely as their entries are unlikely
     * to become unreferenced.
     * <p>
     * This routine is likely to be quite slow and have significant overhead,
     * so only use it when not doing so would cause huge memory bloat or churn,
     * and when you expect the intern()ed values to live for a long time.
     * <p>
     * May not in fact be able to cache/canonicalise all or any objects;
     * mutable objects should never be intern()ed.
     * <p>
     * Calling this with null returns null.
     *
     * @return  an Object that has the same contents as this Object,
     *          but is guaranteed to be from a pool of unique Objects.
     */
//        public static final Object intern(final Object o)
    @SuppressWarnings("unchecked")
    public static <E> E intern(final E o)
        {
        if(o == null) { return(null); }

        // Default is to do nothing.
        E result = o;

        // Assume that String values are very frequent intern()ees,
        // so handle these first.
        if(o instanceof String)
            { result = (E) intern((String) o); }

        // Handle Internable objects...
        else if(o instanceof Internable)
            { result = (E) intern((Internable) o); }

        // Handle specific immutable instances of Number.
        // We only attempt to handle the common JDK-supplied types
        // from the java.lang.* API that we know to be immutable/safe.
        else if(o instanceof java.lang.Number)
            { result = (E) intern((Number) o); }

        // Canonicalise Boolean values in the obvious way...
        else if(o instanceof Boolean)
            { result = (E) (((Boolean) o).booleanValue() ? Boolean.TRUE : Boolean.FALSE); }

        // Else consider warning about pointless calls to intern().
        else if(IsDebug.isDebug)
            { System.err.println("WARNING: MemoryTools.intern() cannot handle objects of type: " + o.getClass().getName()); }

        // Check that the expected invariants hold when we're done.
        assert(result.getClass() == o.getClass());
        assert(result.equals(o) && o.equals(result) && (result.hashCode() == o.hashCode()));

        return(result);
        }


    /**Simplified thread-safe map with fixed capacity that discards excess items in LRU (Least Recently Used) order.
     * This provides a map with a fixed upper size.
     * Attempts to insert more items than the capacity
     * will result in old items being removed in LRU order,
     * thus leaving the newest items in the cache.
     * <p>
     * Each put() or get() makes its key/value pair the most-recently used
     * and thus the last to be removed from the map
     * of current key/value pairs
     * when a series of put()s forces the map to overflow.
     * <p>
     * Thread-safe.
     * <p>
     * A lock can be held on instances of this object to make compound operations atomic.
     * <p>
     * Does not support the full Map interface.
     */
    public static final class SimpleLRUMap<K,V>
        {
        /**Default load factor. */
        private static final float defaultLoadFactor = 0.7f;

        /**Underlying LinkedHashMap on which this is based; never null. */
        private final LinkedHashMap<K,V> lhm;

        /**Create an instance with given (positive) maximum size and the default load factor. */
        public SimpleLRUMap(final int maxCapacity) { this(maxCapacity, defaultLoadFactor); }

        /**Create an instance with given (positive) maximum size and load factor. */
        public SimpleLRUMap(final int maxCapacity, final float loadFactor)
            {
            if((maxCapacity < 0) || (loadFactor <= 0) || (loadFactor >= 1))
                { throw new IllegalArgumentException(); }
            // Build access-order map with override for removal of oldest entry when map too big.
            lhm = new LinkedHashMap<K,V>(maxCapacity, loadFactor, true){
                protected final boolean removeEldestEntry(final Map.Entry<K, V> entry)
                    { return(size() > maxCapacity); }
                /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
                private static final long serialVersionUID = 3013203100902112945L;
                };
            }

        /**Get an entry from the map, making it the Most Recently Used; null if no such element. */
        public synchronized V get(final K key) { return(lhm.get(key)); }

        /**Put an entry in the map, making it the Most Recently Used, returning previous value if any. */
        public synchronized V put(final K key, final V value) { return(lhm.put(key, value)); }

        /**Remove an entry from the map and return the removed value, if any, else null. */
        public synchronized V remove(final K key) { return(lhm.remove(key)); }

        /**Clear the map. */
        public synchronized void clear() { lhm.clear(); }

        /**Take a copy of the map contents, which does not alter LRU; never null. */
        public synchronized Map<K,V> getCopy() { return(new HashMap<K,V>(lhm)); }

        /**Return the number of entries in the map; non-negative. */
        public synchronized int size() { return(lhm.size()); }
        }


    /**Simplified map with fixed capacity that discards excess items in LRU (Least Recently Used) order to meet a hit/miss ratio goal.
     * This adjusts its size to maintain a specific hit/miss ratio in get(),
     * a miss being a get() that returns null,
     * trimming away older items at insert to maintain the ratio.
     * <p>
     * This provides optional fixed lower and upper sizes.
     * Attempts to insert more items than the ceiling capacity
     * will result in old items being removed in LRU order,
     * thus leaving the newest items in the cache.
     * <p>
     * Each put() or get() makes its key/value pair the most-recently used
     * and thus the last to be removed from the map
     * of current key/value pairs
     * when a series of put()s forces the map to overflow.
     * <p>
     * Thread-safe.
     * <p>
     * A lock can be held on instances of this object to make compound operations atomic.
     * <p>
     * Does not support the full Map interface.
     */
    public static final class SimpleLRUMapAutoSizeForHitRate<K,V>
        {
        /**Default maximum miss rate ]0.0,1.0[. */
        public static final float DEFAULT_MAX_MISS_RATE = 0.01f;

        /**Default load factor. */
        public static final float DEFAULT_LOAD_FACTOR = 0.7f;

        /**Default minimum size/capacity, also initial capacity; strictly positive. */
        public static final int DEFAULT_MIN_SIZE = 11;

        /**Underlying LinkedHashMap on which this is based; never null. */
        private final LinkedHashMap<K,V> lhm;

        /**The minimum size/capacity; strictly positive. */
        private final int minSize;

        /**The maximum miss rate ]0.0,1.0[. */
        private final float maxMissRate;

        /**Number of non-miss get()s after which we consider trimming the map if meeting out max-miss-rate targets; strictly positive.
         * This is inversely related to our maxMissRate so as to avoid
         * us shrinking the map too quickly and throwing away rarely-accessed
         * but still-useful items at the tail of our distribution.
         */
        private final int trimCheckRate;

        /**Create an instance with all defaults. */
        public SimpleLRUMapAutoSizeForHitRate()
            { this(DEFAULT_MAX_MISS_RATE, DEFAULT_MIN_SIZE, Integer.MAX_VALUE, DEFAULT_LOAD_FACTOR); }

        /**Create an instance with the given parameters.
         * The standard default is used for the load factor,
         * and the map size is effectively unlimited.
         *
         * @param maxMissRate  the target cache maximum miss rate;
         *     in the range 0.0f to 1.0f exclusive (typically &lt;&lt; 0.1)
         * @param minSize  the minimum (and initial) capacity;
         *     strictly positive
         */
        public SimpleLRUMapAutoSizeForHitRate(final float maxMissRate, final int minSize)
            { this(maxMissRate, minSize, Integer.MAX_VALUE, DEFAULT_LOAD_FACTOR); }

        /**Create an instance with the given parameters.
         *
         * @param maxMissRate  the target cache maximum miss rate;
         *     in the range 0.0f to 1.0f exclusive (typically &lt;&lt; 0.1)
         * @param minSize  the minimum (and initial) capacity;
         *     strictly positive
         * @param maxCapacity  the maximum capacity;
         *     greater than minCapacity (and this strictly positive)
         * @param loadFactor  the load factor of the underlying hash table;
         *     in the range 0.0f to 1.0f exclusive (typically ~0.7f)
         */
        public SimpleLRUMapAutoSizeForHitRate(final float maxMissRate, final int minSize, final int maxCapacity, final float loadFactor)
            {
            if((minSize <= 0) || (maxCapacity <= minSize) ||
               (maxMissRate <= 0) || (maxMissRate >= 1) ||
               (loadFactor <= 0) || (loadFactor >= 1))
                { throw new IllegalArgumentException(); }

            this.minSize = minSize;
            this.maxMissRate = maxMissRate;

            // Choose a sensible value significantly greater than 1
            // so as to keep the per-get() call cost down,
            // and high enough to avoid discarding useful long-tail entries...
            this.trimCheckRate = Math.max(minSize, Math.min(maxCapacity,
                Math.round(0.101f / maxMissRate)));

            // Build access-order map with override for removal of oldest entry when map too big.
            lhm = new LinkedHashMap<K,V>(minSize, loadFactor, true){
                protected final boolean removeEldestEntry(final Map.Entry<K, V> entry)
                    {
                    final int s = size();

                    // If over-size then zap the eldest entry.
                    if(s > maxCapacity) { return(true); }

                    // If under-size then definitely don't zap anything.
                    if(s <= minSize) { return(false); }

                    // If we're meeting the maximum-miss-rate target
                    // then we remove the oldest entry now
                    // so as to stop the map from growing any further...
                    return(meetingMissRateTarget());
                    }
                /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
                private static final long serialVersionUID = 3013203108802112945L;
                };
            }

        /**Count of get() calls; non-negative. */
        private int countGets;

        /**Count of get() calls that returned null, ie misses; non-negative. */
        private int countGetMisses;

        /**Count to trigger possible trim of cache size; non-negative. */
        private int cacheTrimTryCount;

        /**Returns true if we are meeting/bettering the miss-rate target.
         * Will not return true until we have a reasonable number of samples.
         */
        private boolean meetingMissRateTarget()
            {
            // If the number of actual misses is exceeded by
            // the number of misses allowed given
            // the get() count and target rate
            // then we are meeting the target.
            return(Math.round(countGets * maxMissRate) > countGetMisses);
            }

        /**Get an entry from the map, making it the Most Recently Used; null if no such element. */
        public synchronized V get(final K key)
            {
            // Note the get() operation,
            // and rescale all values if getting very large.
            if(++countGets >= Integer.MAX_VALUE/2)
                { countGets /= 2; countGetMisses /= 2; }

            final V result = lhm.get(key);

            // Note any cache "miss"...
            if(result == null)
                {
                // Rescale counts here if they represent too long a history,
                // ie could prevent us reacting to changing usage patterns.
                final int size = lhm.size();
                if(++countGetMisses > Math.max(size/2, trimCheckRate))
                    { countGets /= 2; countGetMisses /= 2; }
                }
            // Else throw away the oldest map entry
            // (on just a (small) sampling of successful calls
            // to reduce amortised cost and rate of map shrinkage)
            // but only if we have bettered the miss-rate target
            // and the map is above its minimum size.
            else if(++cacheTrimTryCount > trimCheckRate)
                {
                // Reset count ready for next trim sample/attempt.
                cacheTrimTryCount = 0;

                final int size = lhm.size();
                if((size > minSize) && meetingMissRateTarget())
                    {
                    // Discard the oldest entry so as to trim the map size.
                    final Iterator<Map.Entry<K,V>> it = lhm.entrySet().iterator();
                    assert(it.hasNext());
                    /* final Map.Entry<K,V> oldest = */ it.next();
//if(IsDebug.isDebug) { System.out.println("[SimpleLRUMapAutoSizeForHitRate.get(): miss rate = "+(countGetMisses/(float)Math.max(1,countGets))+": size = "+lhm.size()+": trimming oldest entry: "+oldest+".]"); }
                    it.remove();
                    }
                }

            return(result);
            }

        /**Put an entry in the map, making it the Most Recently Used, returning previous value if any. */
        public synchronized V put(final K key, final V value) { return(lhm.put(key, value)); }

        /**Remove an entry from the map and return the removed value, if any, else null. */
        public synchronized V remove(final K key) { return(lhm.remove(key)); }

        /**Clear the map and reset the hit/miss counters. */
        public synchronized void clear() { lhm.clear(); countGets = 0; countGetMisses = 0; }

        /**Return the number of entries in the map; non-negative. */
        public synchronized int size() { return(lhm.size()); }

        /**Take a copy of the map contents, which does not alter LRU; never null. */
        public synchronized Map<K,V> getCopy() { return(new HashMap<K,V>(lhm)); }

        /**Provide a human-readable summary of status.
         * Useful for debugging/tuning, for example.
         */
        public String toString()
            {
            return("[SimpleLRUMapAutoSizeForHitRate: miss rate = "+(countGetMisses/(float)Math.max(1,countGets))+": size = "+lhm.size()+".]");
            }
        }
    }
