/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.io.*;
import java.util.Arrays;

/**An immutable, Serializable, wrapper for a byte array.
 * This enables byte array data to be safely shared between multiple users,
 * thus avoid copies and duplicates.
 * <p>
 * We clone at entry and exit and deserialisation to protect our internal data.
 * <p>
 * This also has helper methods for storing/extracting (with zlib)
 * lossless-compressed String values.
 * (The internal form is deflated UTF-8 without headers or checksum.)
 * <p>
 * These instances are suitable for pooling with MemoryTools.intern().
 * <p>
 * Intended to be efficient on the wire and in memory.
 */
public final class ROByteArray implements Serializable, MemoryTools.Internable
    {
    /**Construct with (copy of) non-null data block.
     * @param data  non-null array of bytes
     */
    public ROByteArray(final byte[] data)
        {
        if(data == null) { throw new IllegalArgumentException(); }
        payload = data.clone();
        }

    /**The (non-null) payload reference. */
    private final byte[] payload;

    /**Get (a copy of) the entire data block.
     */
    public byte[] getData()
        {
        return(payload.clone());
        }

    /**Get (a copy of) a portion of the data block.
     */
    public byte[] getData(final int start, final int afterEnd)
        {
        final int length = afterEnd - start;
        final byte result[] = new byte[length];
        System.arraycopy(payload, start, result, 0, length);
        return(result);
        }

    /**Get length of the data; non-negative. */
    public int length()
        {
        return(payload.length);
        }

    /**Compute a hash based on the length and some of the payload.
     * This tries to fix the time taken to compute the hash
     * even for large payloads at the cost of possible/probable extra collisions
     * (as per the String experience).
     * <p>
     * This takes a fixed small number of sample points whose relative positions
     * partly depend on the length, and the length itself is part of the hash.
     */
    public int hashCode()
        {
        final int len = payload.length;
        final int dataHash = (len == 0) ? 907 :
            (
            (len * 0x305011) ^
            payload[0] ^
            (~payload[len/5] << 8) +
            ((payload[(23603 + (0xff & payload[len/3])) % len]) * 65053) -
            (payload[len-1] << 23)
            );
        return(dataHash);
        }

    /**Is equal to an instance with a payload of the same length and content; may be slow to compute. */
    public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof ROByteArray)) { return(false); }
        return(Arrays.equals(payload, ((ROByteArray)obj).payload));
        }

    /**Returns a human-readable representation of this object. */
    public String toString()
        { return(toHexString()); }



    /**Get internal data as InputStream (supporting mark()).
     * This gives read-only access to the data as a markable/rewindable stream.
     * <p>
     * Each call produces a completely independent stream.
     */
    public InputStream getInputStream()
        {
        // This relies for safety on ByteArrayInputStream
        // not allowing the payload to be altered.
        // This relies for efficiency on ByteArrayInputStream
        // not copying the data.
        return(new ByteArrayInputStream(payload));
        }


    /**The set of characters from which we produce our hex digits. */
    private static final String HEX_DIGITS = "0123456789abcdef";

    /**Returns a fixed-length unsigned lower-case hex representation of the data.
     * This returns exactly two (lower-case) hex digits for each byte of data,
     * with the high nybble before the low nybble,
     * and lower-index bytes before higher-index bytes.
     * <p>
     * For example, a data array containing { 13, 241 }
     * would produce "0df1".
     */
    public String toHexString()
        {
        final StringBuilder result = new StringBuilder(payload.length * 2);
        for(int i = 0; i < payload.length; ++i)
            {
            result.append(HEX_DIGITS.charAt((payload[i] >> 4) & 0xf));
            result.append(HEX_DIGITS.charAt(payload[i] & 0xf));
            }
        return(result.toString());
        }

    /**Creates instance from a hex representation as generated by toHexString().
     * Input must be non-null, even number of lower-case hex characters.
     *
     * @throws IllegalArgumentException  for invalid input
     */
    public static ROByteArray fromHexString(final String s)
        {
        if((s == null) || ((s.length() & 1) == 1))
            { throw new IllegalArgumentException(); }

        final byte payload[] = new byte[s.length() / 2];

        for(int i = payload.length; --i >= 0; )
            {
            final char hc = s.charAt(i*2);
            final char lc = s.charAt(i*2 + 1);

            final int hn = HEX_DIGITS.indexOf(hc);
            final int ln = HEX_DIGITS.indexOf(lc);

            if((hn == -1) || (ln == -1))
                { throw new IllegalArgumentException("invalid character"); }

            // Assemble the byte from the high and low nybbles.
            payload[i] = (byte) ((hn << 4) | ln);
            }

        return(new ROByteArray(payload));
        }



    /**Create an instance from the compressed UTF-8 form of the argumentString.
     * Unlikely to work well on short Strings.
     * <p>
     * The compression is zlib deflate without any checksums or other avoidable
     * redundant data, so error checking had better be done elsewhere.
     * <p>
     * The String value can be recovered with the uncompressToString() method.
     *
     * @param in  non-null String
     */
    public static final ROByteArray compressFromString(final String in)
        {
        if(in == null) { throw new IllegalArgumentException(); }
        try
            {
            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            final DataOutputStream dos = new DataOutputStream(baos);
            dos.writeUTF(in);
            dos.flush();
            final byte raw[] = baos.toByteArray();
            dos.close();
            final byte comp[] = FileTools.compressDeflatableData(raw, 0, raw.length);
            return(new ROByteArray(comp));
            }
        catch(final IOException e)
            {
            // Should never happen...
            throw new Error("Unexpected exception", e);
            }
        }

    /**Extract a String from an ROByteArray created by compressFromString(); never null.
     * Results are undefined on other inputs.
     *
     * @param in  non-null, non-empty value generated by compressFromString()
     *
     * @throws IOException  when having difficulty decoding the input
     */
    public static final String uncompressToString(final ROByteArray in)
        throws IOException
        {
        if((in == null) || (in.length() == 0)) { throw new IllegalArgumentException(); }

        final ByteArrayInputStream bais = new ByteArrayInputStream(
                FileTools.decompressDeflatedData(in.payload));
        final DataInputStream dis = new DataInputStream(bais);
        final String s = dis.readUTF();
        dis.close(); // Help GC...
        return(s);
        }


    /**Empty array. */
    public static final ROByteArray EMPTY = new ROByteArray(new byte[0]);

    /**Deserialise: use constructor for validation, defensive copying, etc.
     * Also resolve all empty instances to a single value as a minor optimisation.
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Avoid duplicates of empty case.
        if(payload.length == 0) { return(EMPTY); }

        // Construct new instance of object in normal defensive way.
        return(new ROByteArray(payload));
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = 3806652319473305560L;
    }
