/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
    package org.hd.d.pg2k.svrCore;

import java.util.Random;
import java.util.concurrent.atomic.AtomicInteger;

    /**Provides random number sources to be shared across the system.
     * The sources provided are static variables all extending
     * java.util.Random.
     * <p>
     * Reseeding is either not allowed (throws an exception),
     * or in the spirit of SecureRandom guarantees not to
     * reduce randomness, so these generators can be safely shared.
     * <p>
     * The sources are partially seeded from time of day or some
     * equivalent.
     * <p>
     * Using these shared sources helps:
     * <ul>
     * <li>Avoid rework and redundant resources.
     * <li>Avoid the cost of multiple initialisations.
     * <li>Break up correlations that might otherwise be present.
     * <li>Possibly add extra unpredictability given the multiple
     *     interleaved uses of each source.
     * </ul>
     *
     * @author Damon Hart-Davis
     */
    public final class Rnd
        {
        /**Don't allow creation of instances. */
        private Rnd() { }

        /**Our good-but-slow random number source.
         * We use a cryptographically-secure random number generator,
         * but because sequence-guessing may be possible this should
         * <em>not</em> be used as a source of numbers for highly
         * secure purposes.
         * <p>
         * We will take whatever default generator is available.
         * <p>
         * We allow the generator to seed itself on first use,
         * hopefully with some genuine entropy.
         * <p>
         * We make this generator available to all comers,
         * but they should probably not reseed it, just get random
         * values from it.  (However, a SecureRandom generator is not
         * completely reset by being reseeded, ie ``is guaranteed not
         * to have its randomness reduced'' by being reseeded, so even
         * reseeding probably does not matter too much.)
         * <p>
         * Note that this generator can be relatively slow, taking
         * circa 30s--60s to initialise on an SS5/Cycle5-155MHz
         * (oolong.exnet.com, 1999/12/28, no /dev/random).
         */
        public static final java.security.SecureRandom goodRnd =
            new java.security.SecureRandom();

        /**A fast random number generator (compared to goodRnd).
         * Less good than goodRnd, and possibly with unhelpful
         * correlations, but OK and seeded with at least the time.
         * <p>
         * We allow attempts to reseed this generator,
         * but the reseed value is combined with bits from goodRnd
         * which should make this safe to share
         * and prevents this sequence ever getting 'less' random.
         * <p>
         * This generator is initially seeded from the time
         * to ensure that we get off to a unique-ish start.
         * On subsequent reseeds will use goodRnd to ensure unpredictability.
         * This should allow for a fast start-up and overall decent behaviour.
         * <p>
         * We force reseeding from goodRnd often enough
         * (more than 1 once every sqrt(n) samples,
         * where n is the sequence length of the generator)
         * to avoid "drawing from an urn without replacement" behaviour.
         * For java.util.Random the sequence length is ~2^48,
         * so we reseed from goodRnd at most every 2^24 calls to the
         * internal next() generator routine.
         * We do this infrequently enough to keep the amortised cost-per-call
         * low, however.
         */
        public static final Random fastRnd = (new Random(System.currentTimeMillis() + (new Object()).hashCode()){

            /**Ensure that reseeding cannot "reduce randomness".
             * Setting a particular seed <em>does not</em> reset the generator
             * to a predictable value, but instead starts the sequence
             * in a new place using a value from goodRnd()
             * and the value supplied; this should be different every time
             * even if the same seed is used repeatedly,
             * ie the seed value supplied is relatively unimportant.
             */
            public final void setSeed(final long seed)
                { super.setSeed(seed ^ goodRnd.nextLong()); }

            /**Count-down to next reseed of this generator using goodRnd.
             * We reseed when this value goes negative.
             * (Starting it at zero would force a reseed on first use.)
             * <p>
             * We start this at a random-ish somewhat low count
             * to allow our initial seeding with time to give the system
             * a fast start-up, but then use more heavy-weight reseeding
             * from goodRnd as usual subsequently.
             * <p>
             * We use an AtomicInteger so that we can count down safely
             * without taking a lock which would reduce potential concurrency.
             */
            private final AtomicInteger untilReseed =
                new AtomicInteger(1151 + (int)((System.nanoTime()>>>1) % 9551));

            /**Generates random bits, and periodically reseeds from goodRnd to avoid long predictable runs.
             * The time until the next reseed is also randomly chosen
             * during each reseed, reducing predictability further.
             * <p>
             * But use goodRnd if you really do need good numbers.
             */
            protected final int next(final int bits)
                {
                // If we've drawn enough values from this sequence
                // then reseed it from a cryptographically-good source
                // to eliminate indefinitely-predictable sequences.
                if(untilReseed.decrementAndGet() < 0)
                    {
                    // Draw a long from our good generator.
                    // Use 16 of the bits for our count-down value,
                    // and the rest implicitly for our 48-bit seed.
                    final long r = goodRnd.nextLong();
                    // Reset count-down to next reseed with lower bound.
                    untilReseed.set(0xfff + (int) (r >>> 48));

                    // Now reseed with the new (secure) value,
                    // and some of the old state XORed over the
                    // "count-down-to-reseed" value bits we earlier extracted
                    // just for good measure!
                    super.setSeed(r ^ Long.rotateLeft(super.next(32), 37));

//System.out.println("***RESEEDED fastRnd()"); // Ideally will not happen until system up and running and stable...
//Thread.dumpStack();
                    }

                return(super.next(bits));
                }

            /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
            private static final long serialVersionUID = 3538275569093778199L;
            });
        }
