/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**Some simple common text utilities of scope throughout the application.
 */
public final class TextUtils
    {
    /**Prevent construction of an instance. */
    private TextUtils() { }

    /**Suffixes used by sizeAsText. */
    private static String sizeSuffixes[] = { "B", "kB", "MB", "GB", "TB", "EB" };
    /**Renders size in bytes as text, abbreviated if requested.
     * Can produce normal or abbreviated output.
     */
    public static final String sizeAsText(long size, boolean abbrev)
        {
        StringBuilder sb = new StringBuilder(abbrev ? 5 : 16);
        if(abbrev)
            {
            long divisor = 1;
            for(int i = 0; i < sizeSuffixes.length; ++i, divisor *= 1024)
                {
                if(size < divisor * 1000)
                    {
                    // Takes 3 or less digits with this suffix, so show it!
                    sb.append((size + divisor/2) / divisor);
                    sb.append(sizeSuffixes[i]);
                    return(sb.toString()); // Done.
                    }
                }
            // We may fall through to the default case...
            }
        sb.append(size);
        sb.append("Byte");
        return(sb.toString());
        }

    /**Returns true if given String is 7-bit clean, is is pure ASCII, or is null.
     */
    public static boolean isASCII7(String s)
        {
        if(s == null) { return(true); }
        for(int i = s.length(); --i >= 0; )
            { if((s.charAt(i) & ~0x7f) != 0) { return(false); } }
        return(true); // Seems OK.
        }

    /**Recursively copy second Node and contents into first node.
     * This works even when the second Node is from a different Document to the first
     * and importNode() does not work (eg due to JDK1.5 impl bug).
     */
    public static final void importCopy(final Node dest, final Node src)
        {
        if((dest == null) || (src == null))
            { throw new IllegalArgumentException(); }

        final Element cp = dest.getOwnerDocument().createElement(src.getNodeName());

        // Deal with attrs, etc.
        final boolean hasAttributes = (src.getNodeType() == Node.ELEMENT_NODE) &&
                (src.getAttributes().getLength() > 0);
        if(hasAttributes)
            {
            final NamedNodeMap attribs = src.getAttributes();
            for(int i = 0; i < attribs.getLength(); i++)
                { cp.setAttribute(attribs.item(i).getNodeName(), attribs.item(i).getNodeValue()); }
            }

        // Copy in any node value.
        cp.setNodeValue(src.getNodeValue());

        // Now add the new node to the destination tree.
        dest.appendChild(cp);

        // Recursively copy children.
        for(Node child = src.getFirstChild(); child != null; child = child.getNextSibling())
            { importCopy(cp, child); }
        }

    /**String to write as line terminator when converting to XML.
     * A simple LF would suit UNIX-like systems
     * though would probably be accepted by most.
     * A bulkier CRLF is more like the Internet "standard".
     */
    private static final String XML_LINE_TERM = "\r\n";

    /**Write DOM tree as XML/XHTML String; never "" nor null.
     * Can write as terse/efficient single-line form or longer more human-friendly format.
     * <p>
     * When writing XHTML we make text nodes XML/HTML-safe, though allow embedded entities.
     *
     * @param node  DOM tree root; never null
     * @param toXHTML  if true write with formatting suitable to include directly in HTML/XHTML for human consumption
     *     (eg using dl/ul/ol nested lists to produce formatted text representing the structure)
     *     rather than a pure XML representation of the Node tree itself
     * @param terse  if true write as compactly as possible, else make more human-readable if possible
     */
    public static final String toXML(final Node node, final boolean toXHTML, final boolean terse)
        {
        final StringBuilder result = new StringBuilder(64);
        toXML(result, node, toXHTML, terse);
        return(result.toString());
        }

    /**Write DOM tree as XML/XHTML String; appends to supplied StringBuilder.
     * Can write as terse/efficient single-line form or longer more human-friendly format.
     * <p>
     * When writing XHTML we make text nodes XML/HTML-safe, though allow embedded entities.
     *
     * @param node  DOM tree root; never null
     * @param toXHTML  if true write with formatting suitable to include directly in HTML/XHTML for human consumption
     *     (eg using dl/ul/ol nested lists to produce formatted text representing the structure)
     *     rather than a pure XML representation of the Node tree itself
     * @param terse  if true write as compactly as possible, else make more human-readable if possible
     */
    public static final void toXML(final StringBuilder result,
                                   final Node node, final boolean toXHTML, final boolean terse)
        {
        if((node == null) || (result == null))
            { throw new IllegalArgumentException(); }

        final String nodeValue = node.getNodeValue();

        final boolean hasValue = (node.getNodeValue() != null);

        final boolean hasChildren = (node.getFirstChild() != null);

        final boolean hasAttributes = (node.getNodeType() == Node.ELEMENT_NODE) &&
                (node.getAttributes().getLength() > 0);

        if(toXHTML) // XHTML/HTML formatted text.
            {
            result.append("<dl compact><dt>");
            result.append("<b>").append(node.getNodeName()).append("</b>");

            if(hasAttributes)
                {
                final NamedNodeMap attribs = node.getAttributes();
                for(int i = 0; i < attribs.getLength(); i++)
                    {
                    result.append(' ');
                    result.append(attribs.item(i).getNodeName());
                    result.append("=\"");
                    result.append(escapeHTMLMetaChars(attribs.item(i).getNodeValue()));
                    result.append('"');
                    }
                }

            // Write any value with < > & " ' escaped...
            if(hasValue) { result.append(" [").append(escapeHTMLMetaChars(nodeValue.trim())).append("] "); }

            // Recursively deal with any child nodes...
            if(hasChildren)
                {
                result.append("<dd>");
                for(Node child = node.getFirstChild(); child != null; child = child.getNextSibling())
                    { toXML(result, child, toXHTML, terse); }
                }

            result.append("</dl>");
            }
        else // Pure XML.
            {
            result.append('<').append(node.getNodeName());

            if(hasAttributes)
                {
                NamedNodeMap attribs = node.getAttributes();
                for(int i = 0; i < attribs.getLength(); i++)
                    {
                    result.append(' ');
                    result.append(attribs.item(i).getNodeName());
                    result.append("=\"");
                    result.append(escapeHTMLMetaChars(attribs.item(i).getNodeValue()));
                    result.append('"');
                    }
                }

            // If we have no value or children
            // then we can use the <.../> format with no separate closing tag.
            final boolean unitag = (!hasValue && !hasChildren);

            if(unitag) { result.append('/'); }

            result.append('>');

            if(!unitag)
                {
                if(hasValue) { result.append(escapeHTMLMetaChars(nodeValue.trim())); }

                for(Node child = node.getFirstChild(); child != null; child = child.getNextSibling())
                    { result.append(toXML(child, toXHTML, terse)); }

                result.append("</").append(node.getNodeName()).append('>');
                }
            }

        if(!terse) { result.append(XML_LINE_TERM); } // End the line for readability...
        }

    /**Rewrite HTML so that it displays as "raw" text and is safe to use in attribute values.
     * Meta-characters (&lt;, &gt;, &amp; &quot; ') are rewritten
     * to entity codes so that when shown on screen it looks like
     * the raw HTML text when displayed in-line in HTML,
     * and so as to avoid problems when embedded in XML and HTML.
     * <p>
     * If the input text does not contain these meta-characters
     * then it is returned unchanged.
     * <p>
     * Additionally, all characters &lt; 32 (ASCII control characters)
     * are converted to spaces.
     * <p>
     * If null is passed in, then this returns null,
     * to simplify its use in some cases where a null might be present.
     *
     * @param in  String possibly containing HTML/XML meta-characters
     */
    public static String escapeHTMLMetaChars(final String in)
        {
        if(in == null) { return(null); }

        // Check for "meta" characters in input.
        // If none found then we can return the input String as-is.
        boolean foundMetaChar = false;
        for(int i = in.length(); --i >= 0; )
            {
            final char c = in.charAt(i);
            if((c < 32) || ("<>\"'&".indexOf(c) != -1))
                {
                foundMetaChar = true;
                break;
                }
            }
        if(!foundMetaChar) { return(in); }

        // Take a working writable copy of the input.
        final StringBuilder work = new StringBuilder(in);

        // Work backwards through the string
        // (for efficiency and simplicity)
        // carefully escaping the meta-characters.
        for(int i = work.length(); --i >= 0; )
            {
            final char c = work.charAt(i);
            String replacement;
            switch(c)
                {
                case '<': { replacement =  "&lt;"; break; }
                case '>': { replacement =  "&gt;"; break; }
                case '&': { replacement =  "&amp;"; break; }
                case '"': { replacement =  "&quot;"; break; }
                case '\'': { replacement =  "&#039;"; break; }
                default:
                    {
                    // Replace ASCII control codes...
                    if(c < 32) { replacement = " "; break; }

                    // No replacement needed.
                    continue;
                    }
                }

            // Delete the old character.
            work.deleteCharAt(i);
            // Insert the new string
            // (we could merge these two ops for efficiency).
            work.insert(i, replacement);
            }

        return(work.toString());
        }
    
    /**Private to encode8To6()/decode8To6(); automagically created on first access. */
    private static final class Base64Cache
        {
        /**Prevent construction of any instance. */
        private Base64Cache() {}

        /**Private to encode8To6()/decode8To6(). */
        private static final char[] alphabet = {
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',  //  0 to  7
            'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',  //  8 to 15
            'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',  // 16 to 23
            'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',  // 24 to 31
            'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',  // 32 to 39
            'o', 'p', 'q', 'r', 's', 't', 'u', 'v',  // 40 to 47
            'w', 'x', 'y', 'z', '0', '1', '2', '3',  // 48 to 55
            '4', '5', '6', '7', '8', '9', '+', '/'}; // 56 to 63    

        /**Index/value of each legit 7-bit-ASCII character from the alphabet.
         * All other (ie illegal) values map to -1.
         * <p>
         * This is effectively the reverse mapping for alphabet[].
         */
        private static final int valueOf[] = new int[128];

        /**Initialise valueOf[]. */
        static
            {
            for(int i = valueOf.length; --i >= 0; ) { valueOf[i] = -1; }
            for(int i = alphabet.length; --i >= 0; ) { valueOf[alphabet[i]] = i; }
            }

        /**Gets valueOf given char; non-negative.
         * @throws IllegalArgumentException for invalid char (not from allowed alphabet)
         * @return value in range 0--63 inclusive
         */
        public static int getValueOf(final char c)
            {
            if(c > 128) { throw new IllegalArgumentException(); }
            final int v = valueOf[c];
            if(v == -1) { throw new IllegalArgumentException(); }
            return(v);
            }
        }

    /**Decode to a byte array (8 bit) from ASCII Base-64 (6 bit); never null.
     * The input data must always be a multiple of 4 characters,
     * and all character must come from the standard Base64 set [A-Za-z0-9+/=].
     *
     * @param base64Text  data in base-46, eg as encoded by encode8To6(); never null
     * @return binary data; never null
     */
    public static byte[] decode8To6(final String base64Text)
        {
        if(base64Text == null) { throw new IllegalArgumentException(); }
        final int inputLen = base64Text.length();
        // Handle empty input specially.
        if(inputLen == 0) { return(new byte[0]); }
        // Check input is legitimate length (eg appropriately padded with trailing '='s).
        if((inputLen & 3) != 0)
            { throw new IllegalArgumentException("input length must be multiple of 4: was "+base64Text.length()); }

        // Compute the maximum result length (absent trailing '='s).
        final int maxResultSize = 3 * (inputLen / 4);
        // Compute the actual result size.
        final boolean oneTrailingEquals = (base64Text.charAt(inputLen-1) == '=');
        final boolean twoTrailingEquals = oneTrailingEquals && (base64Text.charAt(inputLen-2) == '=');
        final int resultSize = twoTrailingEquals ? (maxResultSize-2) :
                                   (oneTrailingEquals ? (maxResultSize-1) :
                                       maxResultSize);
        // Make the array for the final result...
        final byte result[] = new byte[resultSize];

        // Decoding TO block of 3 bytes at index i in the result
        // FROM block of 4 characters at index j in the source.
        for(int i = 0, j = 0; i < resultSize; i += 3, j += 4)
            {
            // Special treatment for the last block if it is partial.
            final boolean partialLastBlock = ((i+3) > resultSize);

            // Always the first two chars of input block must be real encoded data.
            final int v0 = Base64Cache.getValueOf(base64Text.charAt(j+0));
            final int v1 = Base64Cache.getValueOf(base64Text.charAt(j+1));
            // Always at least one output byte.
            result[i+0] = (byte)((v0 << 2)         + (v1 >>> 4)); // b0 = v0v0v0v0 v0v0v1v1

            if(!partialLastBlock)
                {
                // Full block; two more bytes to decode from three more input chars.
                final int v2 = Base64Cache.getValueOf(base64Text.charAt(j+2));
                final int v3 = Base64Cache.getValueOf(base64Text.charAt(j+3));
                result[i+1] = (byte)(((v1 & 0xf) << 4) + (v2 >>> 2)); // b1 = v1v1v1v1 v2v2v2v2
                result[i+2] = (byte)(((v2 & 3) << 6)   +  v3       ); // b2 = v2v2v3v3 v3v3v3v3
                }
            else if(!twoTrailingEquals)
                {
                // Partial last block with more than one byte to decode from one more inout char.
                final int v2 = Base64Cache.getValueOf(base64Text.charAt(j+2));
                result[i+1] = (byte)(((v1 & 0xf) << 4) + (v2 >>> 2)); // b1 = v1v1v1v1 v2v2v2v2
                }
            }

        return(result);
        }

    /**Encode a byte array (8 bit) in ASCII Base-64 (6 bit); never null.
     * With thanks to Apache.
     * @param data8Bit  binary input data; never null
     * @return ASCII base-64 representation as String; never null
     */
    public static String encode8To6(final byte[] data8Bit)
        {
        final StringBuilder sb = new StringBuilder(data8Bit.length * 2);

        int i = 0; // Index in data8Bit[].
        while((i + 3) <= data8Bit.length)
            {
            int bits24 = (data8Bit[i++] & 0xFF) << 16;
            bits24 |= (data8Bit[i++] & 0xFF) << 8;
            bits24 |= (data8Bit[i++] & 0xFF);
            sb.append(Base64Cache.alphabet[(bits24 & 0x00FC0000) >> 18]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x0003F000) >> 12]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x00000FC0) >> 6]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x0000003F)]);
            }

        if(data8Bit.length - i == 2)
            {
            int bits24 = (data8Bit[i] & 0xFF) << 16;
            bits24 |= (data8Bit[i + 1] & 0xFF) << 8;
            sb.append(Base64Cache.alphabet[(bits24 & 0x00FC0000) >> 18]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x0003F000) >> 12]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x00000FC0) >> 6]);
            sb.append('=');
            }
        else if(data8Bit.length - i == 1)
            {
            final int bits24 = (data8Bit[i] & 0xFF) << 16;
            sb.append(Base64Cache.alphabet[(bits24 & 0x00FC0000) >> 18]);
            sb.append(Base64Cache.alphabet[(bits24 & 0x0003F000) >> 12]);
            sb.append('=');
            sb.append('=');
            }

        return(sb.toString());
        }

    /**Sanitise text for XHTML/HTML/WML/XML use.
         * The transformations are:
         * <ul>
         * <li>A null argument is converted to an empty string.
         * <li>Whitespace is trimmed from the ends.
         * <li>Converts any ``&lt;'' to ``{,'' ``&gt;'' to ``}''.
         * <li>Unless in "allowEntities" modes, converts any ``&amp;'' to ``+'',
         *     else leaves syntactically-correct entities intact and counts
         *     as a single character.
         * <li>Any non-ASCII characters are converted to a safe value;
         *     if entity codes are permitted then these values are converted
         *     to entity codes.
         * <li>The result is limited to at most the specified number of characters.
         * </ul>
         * <p>
         * Our simple definition of a syntactically-correct entity is that
         * it may contain any selection of ASCII digits and letters and optionally
         * a leading hash (`#'), but certainly no whitespace.  We may enforce a
         * length limit.
         * <p>
         * If the input string is OK it is returned untouched.
         * <p>
         * TODO: Needs version with extra argument to allow entity sequences
         *     (of the form '&amp;'xxx;) to be treated as single characters
         *     and to delete any '&lt;'...'&gt;' sequences,
         *     ie to do a more sophisticated job of sanitising XML/HTML
         *     that has some simple markup (primarily entity codes needed for i18n).
         * TODO: jUnit tests
         *
         * @param maxLen  if positive the output is limited to at most
         *     this number of characters;
         *     if &gt;= 3 truncated text is marked with a trailing ``...''
         *     in the last three prositions
         * @param allowEntities  if true, treats HTML/XML entity sequences
         *     as if single characters; the entities are vaguely tested for
         *     correct syntax and may not be allowed if invalid
         */
        public static String sanitiseForXML(final String input,
                                            final int maxLen,
                                            final boolean allowEntities)
            {
            if(maxLen < 0) { throw new IllegalArgumentException(); }
    
            if(input == null) { return(""); }
    
            // First trim whitespace from either end.
            String in = input.trim();
    
            // Always remove potentially-unsafe tag meta-characters.
            in = in.replace('<', '{').replace('>', '}');
    
            // If non-null, the input string has had all entities replaced with
            // just the leading '&' with the entity code
            // (not including the trailing ';')
            // at the matching index in this array,
            // so all entities will appear as a single character.
            String entities[] = null;
    
            // If not allowing entities zap all '&' chars...
            if(!allowEntities)
                { in = in.replace('&', '+'); }
            // If we are allowing entities, and there are any,
            // convert them all to placeholders.
            else if(in.indexOf('&') >= 0)
                {
                // in may get truncated but not longer while this array is in use.
                entities = new String[in.length()];
    
                // Break out the (valid) entities one at a time...
                for(int nextAmp = -1; (nextAmp = in.indexOf('&', nextAmp + 1)) != -1; )
                    {
                    // Look for the next semicolon...
                    final int nextSc = in.indexOf(';', nextAmp + 1);
                    if(nextSc == -1)
                        {
                        // Whoops; if no more semicolons, then
                        // all ampersands from current are bogus and we can stop.
                        in = ((nextAmp >= 0) ? in.substring(0, nextAmp) : "") +
                            (in.substring(nextAmp).replace('&', '+'));
                        break;
                        }
    
    
                    // TODO: do more validity checking of a putative entity:
                    // check that it letters and digits or a hash and all digits.
    
    
                    // Now assume that the entity is valid,
                    // so break it out into the entities array
                    // leaving the '&' behind
                    // and putting the body (not the trailing ';')
                    // in the entities[] array at the '&' index.
                    entities[nextAmp] = in.substring(nextAmp+1, nextSc);
                    in = in.substring(0, nextAmp+1) +
                         in.substring(nextSc+1);
                    }
                }
    
            // Truncate if necessary while
            // any entities that we care about are "broken out"...
            if((maxLen >= 0) && (in.length() > maxLen))
                {
                if(maxLen >= 3) { in = in.substring(0, maxLen-3) + "..."; }
                else { in = in.substring(0, maxLen); }
                }
    
            // If we broke out any entities,
            // put any remaining ones back in-line.
            if(entities != null)
                {
                final StringBuilder sb = new StringBuilder(in.length()); // At least this long.
    
                final int len = in.length();
                for(int i = 0; i < len; ++i)
                    {
                    final char c = in.charAt(i);
                    sb.append(c);
                    if(c != '&')
                        { continue; } // Usual case; no entity involved.
    
                    // Expand an entity.
                    assert(entities[i] != null);
                    sb.append(entities[i]).append(';');
                    }
    
                // Convert back to expanded string.
                in = sb.toString();
    
    //            entities = null; // Allow GC.
                }
    
            // Look for any char outwith range 32--126.
            // If one is found replace every occurence with ' ' and retrim at end,
            // unless allowing entities in which case replace with an entity code.
            for(int i = in.length(); --i >= 0; )
                {
                final char c = in.charAt(i);
                if((c < 32) || (c > 126))
                    {
                    // Whole string needs fixing...
                    final StringBuilder sb = new StringBuilder(in.length());
                    for(int j = 0; j < in.length(); ++j)
                        {
                        final char d = in.charAt(j);
                        if((d < 32) || (d > 126))
                            {
                            // If allowing entities,
                            // convert non-ASCII chars to entities.
                            if(!allowEntities) { sb.append(' '); }
                            else { sb.append("&#").append(0xffff & (int) d).append(';'); }
                            }
                        else { sb.append(d); }
                        }
                    in = sb.toString();
                    if(!allowEntities) { in = in.trim(); }
                    break;
                    }
                }
    
            if(input.equals(in)) { return(input); } // Avoid making copies!
            return(in);
            }
    }
