/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

import ORG.hd.d.IsDebug;

/**Utilities to assist with threading/concurrency.
 */
public final class ThreadUtils
    {
    /**Present creation of instances. */
    private ThreadUtils() { }


    /**A factory for creating daemon pool threads.
     * Adapted (purloined) from Doug Lea / Sun JVM code.
     */
    public static final class DaemonThreadFactory implements ThreadFactory
        {
        private final ThreadGroup group;
        private final AtomicInteger threadNumber = new AtomicInteger(1);
        private final String namePrefix;
        private final boolean lowPriority;

        public DaemonThreadFactory(final String poolName,
                                   final boolean lowPriority)
            {
            final SecurityManager s = System.getSecurityManager();
            group = (s != null)? s.getThreadGroup() :
                                 Thread.currentThread().getThreadGroup();
            namePrefix = "pool-" + poolName + "-thread-";
            this.lowPriority = lowPriority;
            }

        public Thread newThread(final Runnable r)
            {
            final Thread t = new Thread(group, r, 
                                  namePrefix + threadNumber.getAndIncrement(),
                                  0);
            t.setDaemon(true);
            if(lowPriority)
                { t.setPriority(Thread.MIN_PRIORITY); }
            else if(t.getPriority() != Thread.NORM_PRIORITY)
                { t.setPriority(Thread.NORM_PRIORITY); }
            return(t);
            }
        }

    /**Estimated number of CPUs available; strictly positive. */
    private static final int AVAILABLE_PROCESSORS = Runtime.getRuntime().availableProcessors();

    /**Maximum concurrent low-priority pool threads; strictly positive. */
    private static final int MAX_LO_PRI_THREADS = Math.max(1, (AVAILABLE_PROCESSORS/2));

    /**Maximum number of threads that may run in nonCPUThreadPool(); strictly positive.
     * We limit the amount of threading by:
     * <ul>
     * <li>The likely limit on back-end connectivity (eg HTTP-limited).
     * <li>The likely strain on other resources, such as disc and memory.
     * </ul>
     * <p>
     * Observation suggests that at least 8 threads should be deployed
     * to extract maximum throughput from a single disc spindle,
     * so we should always deploy at least that,
     * rising slowly as the number of CPUs goes up
     * simply to use some of them and because
     * I/O capacity probably grows somewhat with CPU count.
     */
    private static final int MAX_THREADS_NCTP = 8 + (AVAILABLE_PROCESSORS/2);

    /**Shared thread pool for I/O-bound activities.
     * Suitable for relatively short-lived and/or I/O-bound threads,
     * thus we have a fixed thread/resource limit.
     * <p>
     * Work cannot be queued for this pool
     * (ie tasks are always executed immediately),
     * and excess work (with no free pool threads available)
     * will be run directly in the caller's thread.
     * <p>
     * We always try to keep one or two worker threads alive for fast response.
     * <p>
     * The threads in the pool are daemon threads,
     * so should not prevent the JVM from exiting.
     */
    public static final ExecutorService nonCPUThreadPool =
        Executors.unconfigurableExecutorService(
            new ThreadPoolExecutor(2, MAX_THREADS_NCTP,
                300L, TimeUnit.SECONDS, // Keep worker threads alive for 5 minutes...
                new SynchronousQueue<Runnable>(), // Allow some work to be queued.
                new ThreadUtils.DaemonThreadFactory("ThreadUtils.nonCPUThreadPool", false),
                new ThreadPoolExecutor.CallerRunsPolicy()));

    /**Shared thread pool for compute-bound activities.
     * Suitable for almost-entirely CPU-bound threads
     * (though we allow for a little parallel slackness),
     * and so the pool size is bounded to a value roughly proportional to
     * the number of CPUs available to this JVM (at creation of this pool).
     * <p>
     * Work cannot be queued for this pool
     * (ie tasks are always executed immediately),
     * and excess work (with no free pool threads available)
     * will be run directly in the caller's thread.
     * <p>
     * This puts an upper bound on Thread resources (eg memory for stack)
     * consumed by threads in this pool.
     * <p>
     * The threads in the pool are daemon threads,
     * so should not prevent the JVM from exiting.
     */
    public static final ExecutorService computeIntensiveThreadPool =
        Executors.unconfigurableExecutorService(
            new ThreadPoolExecutor(1, 1 + ((5*AVAILABLE_PROCESSORS)/4),
                600L, TimeUnit.SECONDS, // Keep worker threads alive for 10 minutes...
                new SynchronousQueue<Runnable>(),
                new DaemonThreadFactory("ThreadUtils.computeIntensiveThreadPool", false),
                new ThreadPoolExecutor.CallerRunsPolicy()));

    /**Shared thread pool for low-priority compute-bound activities.
     * Suitable for almost-entirely CPU-bound threads
     * (though we allow for a little parallel slackness),
     * and so the pool size is bounded to a value roughly proportional to
     * the number of CPUs available to this JVM (at creation of this pool).
     * <p>
     * These threads run with minimum priority.
     * <p>
     * Tasks submitted to this pool <em>should not</em>
     * consume heavy resources (such as memory) other than CPU
     * so that we can run many of them at once,
     * though we try not to use more than half the CPUs at once
     * when there are multiple CPUs available.
     * <p>
     * A bounded (small) amount of work can be queued for this pool
     * (ie tasks are usually executed immediately or should not wait long),
     * and excess work (with no free pool threads or queue space available)
     * will be run directly in the caller's thread.
     * <p>
     * This puts an upper bound on Thread resources (eg memory for stack)
     * consumed by threads in this pool.
     * <p>
     * The threads in the pool are daemon threads,
     * so should not prevent the JVM from exiting.
     */
    public static final ExecutorService lowPriorityThreadPool =
        Executors.unconfigurableExecutorService(
            new ThreadPoolExecutor(MAX_LO_PRI_THREADS, MAX_LO_PRI_THREADS,
                600L, TimeUnit.SECONDS, // Keep worker threads alive for 10 minutes...
                new ArrayBlockingQueue<Runnable>(MAX_LO_PRI_THREADS), // Approx constant queue wait?
                new DaemonThreadFactory("ThreadUtils.lowPriorityThreadPool : ExecutorService", true),
                new ThreadPoolExecutor.CallerRunsPolicy()));

    /**Shared thread pool for discardable low-priority compute-bound activities.
     * Suitable for almost-entirely CPU-bound threads
     * (though we allow for a little parallel slackness),
     * and so the pool size is bounded to a value roughly proportional to
     * the number of CPUs available to this JVM (at creation of this pool).
     * <p>
     * When this pool and queue is already full
     * <em>it silently discard new entries</em>.
     * <p>
     * These threads run with minimum priority.
     * <p>
     * Tasks submitted to this pool <em>should not</em>
     * consume heavy resources (such as memory) other than CPU
     * so that we can run many of them at once,
     * though we try not to use more than half the CPUs at once
     * when there are multiple CPUs available.
     * <p>
     * A bounded (small) amount of work can be queued for this pool
     * (ie tasks are usually executed immediately or should not wait long),
     * and excess work (with no free pool threads or queue space available)
     * will be <em>silently discarded</em>.
     * <p>
     * This puts an upper bound on Thread resources (eg memory for stack)
     * consumed by threads in this pool.
     * <p>
     * The threads in the pool are daemon threads,
     * so should not prevent the JVM from exiting.
     */
    public static final ExecutorService lowPriorityThreadPoolDiscardable =
        Executors.unconfigurableExecutorService(
            new ThreadPoolExecutor(MAX_LO_PRI_THREADS, MAX_LO_PRI_THREADS,
                600L, TimeUnit.SECONDS, // Keep worker threads alive for 10 minutes...
                new ArrayBlockingQueue<Runnable>(2*MAX_LO_PRI_THREADS), // Approx constant queue wait?
                new DaemonThreadFactory("ThreadUtils.lowPriorityThreadPoolDiscardable : ExecutorService", true),
                new ThreadPoolExecutor.DiscardPolicy()));

    /**Dump some diagnostics during debug startup. */
    static
	    {
	    if(IsDebug.isDebug)
		    {
		    System.out.println("DEBUG: ThreadUtils: init: available processors="+AVAILABLE_PROCESSORS+
		    		                        		", MAX_LO_PRI_THREADS="+MAX_LO_PRI_THREADS+
		    		                        		", MAX_THREADS_NCTP="+MAX_THREADS_NCTP);
		    }
	    }
    }
