/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore;

import org.hd.d.pg2k.svrCore.vars.*;

    /**This class has tools handling common SystemVariables.
     */
    public final class VarTools
        {
        /**Prevent creation of instances of this class. */
        private VarTools() { }

        /**Minimum age of globalMap entry we will discard, ms; strictly positive. */
        public static final int MIN_AGE_MS = 2*SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS;





        /**Robustly computes simple stats on a named generic variable.
         * The stats instance produced is
         * @see #generateSimpleNumberStats
         *
         *
         * @param maxAge_ms  maximum age of globalMap entry we will use in ms;
         *     will be coerced upwards to a minimum value based on allowed
         *     variable-propagation latency
         *
         * Catches most exceptions and returns null in case of error.
         * <p>
         * Computes the system ID automatically, if available.
         */
        public static final SimpleVarStats generateSimpleStats(
                                final BasicVarMgrInterface dataSource,
                                final SimpleVariableDefinition def,
                                final int maxAge_ms)
            {
            if((dataSource == null) ||
               (def == null))
                { throw new IllegalArgumentException(); }

            try
                {
                final SimpleVariableValue sys =
                    dataSource.getVariable(SystemVariables.LOCAL_SYS_ID);
                // Build and return the requested stats...
                return(new SimpleVarStats(dataSource.getVariable(def),
                    System.currentTimeMillis() -
                        Math.max(maxAge_ms, MIN_AGE_MS),
                    (sys == null) ? null : (InstanceID) sys.getValue()));
                }
            catch(Exception e)
                { return(null); }
            }

        /**Robustly computes simple stats on a named TYPE_NUMBER variable.
         * @see #generateSimpleNumberStats
         *
         * @param maxAge_ms  maximum age of globalMap entry we will use in ms;
         *     will be coerced upwards to a minimum value based on allowed
         *     variable-propagation latency
         *
         * Catches most exceptions and returns null in case of error.
         * <p>
         * Computes the system ID automatically, if available.
         */
        public static final SimpleNumberStats generateSimpleNumberStats(
                                final BasicVarMgrInterface dataSource,
                                final SimpleVariableDefinition def,
                                final int maxAge_ms,
                                final String decimalFormatPattern)
            {
            if((dataSource == null) ||
               (def == null))
                { throw new IllegalArgumentException(); }

            try
                {
                final SimpleVariableValue sys =
                    dataSource.getVariable(SystemVariables.LOCAL_SYS_ID);
                return(generateSimpleNumberStats(dataSource.getVariable(def),
                                           maxAge_ms,
                                           decimalFormatPattern,
                                           (sys == null) ? null : (InstanceID) sys.getValue()));
                }
            catch(Exception e)
                { return(null); }
            }


        /**Robustly computes simple stats on a TYPE_NUMBER variable.
         * Intended to be used both in normal code and
         * directly in JSP pages
         * (eg is robust, never throws exceptions, HMTL-friendly toString()).
         * <p>
         * Returns an immutable object containing the stats
         * which has a toString() which renders a relatively
         * admin-human-friendly simple String containing all the stats,
         * or null if the variable presented is null or not numeric
         * or for some reason not suitable.
         * <p>
         * A local value is treated like a global with one entry
         * but flagged as local.
         * <p>
         * Only the young-enough members of a global are used;
         * if none are young enough then the main value is used
         * and flagged as old.
         *
         * @param sysID  identifies this system so we can select the local
         *     value of a global variable if available; can be null
         * @param maxAge  maximum age of globalMap entries to be used;
         *     coerced to be at least a little larger than the refresh
         *
         * @return null if variable is unsuitable, else encapsulated stats
         */
        public static final SimpleNumberStats generateSimpleNumberStats(
                                final SimpleVariableValue var,
                                final int maxAge,
                                final String decimalFormatPattern,
                                final InstanceID sysID)
            {
            // Return null if the variable is unsuitable.
            if((var == null) ||
               (var.getDef().getType() != SimpleVariableDefinition.TYPE_NUMBER) ||
               (var.getValue() == null))
                { return(null); }

            // Build and return the requested stats...
            return(new SimpleNumberStats(var,
                System.currentTimeMillis() -
                    Math.max(maxAge, MIN_AGE_MS),
                sysID, decimalFormatPattern
            ));
            }
        }
