/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.datasource;

import org.hd.d.par.ParTasks;
import org.hd.d.par.TaskProc;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.AllExhibitProperties.ExhibitDataSource;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.props.SecurityProps;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.*;
import java.nio.ByteBuffer;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;


/**Exhibit pipeline stage that fetches its data directly from a filesystem.
 * This emulates the way that the pre-PG2K Gallery accesses its data.
 * <p>
 * That this considers itself to be the definitive master data store
 * for persistent system variables (particularly event histories)
 * and will try to use the persistent data area to store them,
 * in a robust way with a history.
 * <p>
 * Past event values returned by this routine are considered to be
 * authoritative, and upstream consumers can cache such values
 * knowing them to be definitive.
 * <p>
 * All requests for valid events/periods should be responded to
 * as authoritative or not depending on the slot time as follows:
 * <ul>
 * <li>In the future: null are returned because no data *can* exist yet.
 * <li>For the current interval: a non-authoritative, non-null response.
 * <li>For any past interval: an authoritative response.
 * </ul>
 * Where the data for responses does not yet exist,
 * it must be invented with no events to indicate that
 * it definitely does not exist
 * so as to allow negative cacheing by upstream users.
 * <p>
 * We don't aim to be massively efficient with this since we hope the
 * cacheing stage that should normally be downstream of us will compensate
 * for our main inefficiencies.  We might, however, switch to some sort of
 * non-blocking I/O in future so that failures of networked filesystems,
 * (etc) internally won't stop Web service.
 * <p>
 * We locate configuration files and data using LocalProps.
 */
public final class ExhibitDataFileSource implements SimpleExhibitPipelineIF
    {
    /**Make uninitialised cache. */
    public ExhibitDataFileSource() { this(null);  }

    /**Make initialised cache.
     * This can be used to save some time by giving an initial
     * cache value restored from elsewhere, with computed properties
     * and so on.
     * <p>
     * Not necessarily safe, so private for now.
     * <p>
     * If the argument is null, this is like calling the no-arg constructor.
     * <p>
     * Loads any persisted system variable values.
     */
    private ExhibitDataFileSource(final AllExhibitProperties extantAep)
        {
        _gAEP_cache = extantAep;

        final File sysVarDir = _getEventHistoryStorageDir();
        if(!sysVarDir.isDirectory())
            { System.err.println("WARNING: event history storage dir not found: " + sysVarDir); }

        try { varMgr.loadEventHistories(sysVarDir); }
        catch(final IOException e)
            { System.err.println("WARNING: event history could not be restored from: " + sysVarDir); }
        }

    /**The cached exhibit-data file to read, or null if not set or not available.
     * This is usually relative to the data directory.
     * <p>
     * If this is non-null then we don't construct the exhibit static
     * data ourselves when called by any of the SimpleExhibitPipelineIF methods,
     * but instead read a GZIPed serialised AllExhibitProperties object at
     * the file given by this value, throwing an IOException if it cannot
     * be read.
     * <p>
     * The createStaticCacheFile() routine can be used to create this
     * file and do as much precomputation and preparation of thumbnails,
     * etc, as possible off-line.  Typically a command-line program
     * would be used to run this each time the exhibits are changed,
     * and the master will simply re-read the serialised file periodically.
     * It is important in this case that the file be replaced atomically.
     * The createStaticCacheFile() tries to create the file if it does not
     * exist.
     * <p>
     * File accesses always go directly to the filesystem.
     */
    private static final File _getStaticCacheFileName()
        {
        final File f = new File(LocalProps.getDataDir(),
                                CoreConsts.FS_DATA_CACHE_FILENAME);
        if(!f.canRead())
            { return(null); }

        return(f);
        }

    /**If true, load exhibit properties directly if staticCacheFile not available. */
    private static final boolean REVERT_IF_SCF_NOT_AVAILABLE = true;

    /**Get the static attributes for a given exhibit.
     */
    public ExhibitStaticAttr getStaticAttr(final String name)
        throws IOException
        {
        if(_getStaticCacheFileName() != null)
            { return(getAllExhibitProperties(-1L).aeid.getStaticAttr(name)); }
        return(_getStaticAttr(LocalProps.getDataDir(), name));
        }

    /**Get the static attributes for a given exhibit.
     * Returns null if the named exhibit does not exist.
     * We get very paranoid and reject anything that looks as if it
     * has a .xxx or .. component in it, then validate the name,
     * then look up the details.
     * <p>
     * This computes a new uncached value on each call.
     * <p>
     * TODO: prevent this from blocking indefinitely, even with a filesystem hang.
     */
    private static ExhibitStaticAttr _getStaticAttr(final String fileRoot,
                                                    final String name)
        throws IOException
        {
        if(fileRoot == null)
            { throw new IllegalArgumentException(); }

        if(name.startsWith(".") || (name.indexOf("/.") != -1)) /* Extra security check. */
            { throw new IOException("invalid name (unsafe)"); }
        if(!ExhibitName.validNameSyntax(name))
            { throw new IOException("invalid name (syntax)"); }
        final File file = new File(fileRoot, name);
        if(!file.exists() || !file.canRead())
            { throw new FileNotFoundException(name); }

        return(new ExhibitStaticAttr(name,
                                     file.length(),
                                     file.lastModified()));
        }

    /**Read a chunk of the raw exhibit binary into the supplied buffer.
     * The start position and an implied maximum read length are supplied.
     * The start must be non-negative and no larger than the exhibit data length.
     * <p>
     * If a zero-byte request is made then the file may not actually be accessed.
     * <p>
     * This goes directly to the filesystem for each call.
     * <p>
     * TODO: prevent this from blocking indefinitely, even with a filesystem hang.
     */
    public void getRawFile(final ByteBuffer buf, final String exhibitName, final int position, final boolean dontCache)
        throws IOException
        {
        if(!ExhibitName.validNameSyntax(exhibitName))
            { throw new IOException("invalid name (syntax)"); }

        // Do some extra security checks of requested file's name.
        if(exhibitName.startsWith(".") || (exhibitName.indexOf("/.") != -1))
            { throw new IOException("invalid name (unsafe)"); }

        final File file = new File(LocalProps.getDataDir(), exhibitName);
        if(!file.exists() || !file.canRead())
            { throw new FileNotFoundException(exhibitName); }

        // Do the read!
        RandomAccessFile raf = null;
        try
            {
            raf = new RandomAccessFile(file, "r");
//            raf.seek(position);
//            raf.readFully(buf, offset, len);
            raf.getChannel().read(buf, position);
            }
        catch(final IOException e)
            {
            System.err.println("IOException reading exhibit bytes start="+position+" name="+exhibitName+": " + e.getMessage());
            throw e; // Rethrow the error.
            }
        finally { if(raf != null) { raf.close(); } } // Release handle ASAP.
        }


    /**Gets all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned
     * unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     */
    public AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
        throws IOException
        {
        if(_getStaticCacheFileName() != null)
            {
            final AllExhibitImmutableData extant =  getAllExhibitProperties(-1L).aeid;
            // If caller has an up-to-date copy, return null.
            if((oldStamp >= 0) && (oldStamp == extant.timestamp))
                { return(null); }
            return(extant);
            }

        return(_getAllExhibitImmutableData(LocalProps.getDataDir(), oldStamp, true));
        }

    /**Number of least-significant bytes of AllExhibitImmutableData timestamp to use as hash of full collection.
     * Using a value of 1 would imply keeping about quarter-second timestamp
     * precision (which does not exist in UNIX systems, typically).
     * Using a value of 2 would imply about 1 minute precision.
     * Using a value of 3 would imply about 5 hour precision.
     * <p>
     * Given that the Gallery is usually updated at intervals of at most
     * hours to days normally, a value of 2 or 3 is probably optimal.
     * <p>
     * Must lie in the range 0 to 8 inclusive.
     */
    private static final int TS_LSbytes_HASH = 3;

    /**Gets all immutable exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned
     * unconditionally.
     * <p>
     * This does not use any cache, and computes afresh on each call,
     * and is thread-safe.
     * <p>
     * If no exhibits are currently installed then a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     * <p>
     * We make the timestamp we use for the whole collection be the
     * most significant bits of the latest-stamped exhibit and the
     * least significant bits of the collection timestamp XORed with a hash
     * over all the (sorted) ExhibitStaticAttr entries.  We don't have many
     * least significant bits, so we run a smallish risk of missing some
     * changes.  In fact, we are quite rough-and-ready about the hash too!
     * <p>
     * We know that many (typically 10) of the least significant bits of
     * the timestamp would not carry information anyway, eg with a
     * one-second granularity in the UNIX filesystem.
     * <p>
     * To avoid causing too much confusion with our somewhat-faked
     * timestamp, we limit it to between 1 and the current time of day
     * in the worst case, though because we ensure that the fake
     * timestamp calculated is no later than the timestamp of the newest
     * exhibit then unless there is lots of clock-skew between our host
     * and the file server we should not see this latter clamp actually used.
     * The latter limit, if needed, can make the system inefficient just after
     * a new exhibit has been added since our fake timestamp may seem
     * to change on every call.
     *
     * @param careful  if true, magic numbers of exhibits are checked and
     *     other extra-careful checking is done; this should be the default
     *     usage
     */
    public static AllExhibitImmutableData _getAllExhibitImmutableData(final String fileRoot,
                                                                      final long oldStamp,
                                                                      final boolean careful)
        throws IOException
        {
        if(fileRoot == null)
            { throw new IllegalArgumentException(); }
        final File baseDir = new File(fileRoot);

        // Load exhibit set carefully,
        // and store it in sorted order.
        // We access files in sorted order to improve filesystem performance.
        final SortedSet<String> names = new TreeSet<String>(ExhibitFile.getFilesystemBasedExhibitNames(
            baseDir, careful));

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[Found "+(names.size())+" exhibits based at "+baseDir+".]"); }

        // Now iterate over the set in order,
        // finding the newest exhibit,
        // and computing a hash on all the static attr data.
        // We also construct our map from names to attr at this point.
        final Map<String,ExhibitStaticAttr> m = new HashMap<String, ExhibitStaticAttr>(names.size() * 2, 0.7f);

        long newest = 0; // Newest exhibit's timestamp.

        final MessageDigest md = GenUtils.getStandardDigest();
        final DataOutputStream dos =
            new DataOutputStream(
                new DigestOutputStream(
                    (new OutputStream(){ // Null output stream...
                        public final void write(final int b) { }
                        public final void write(final byte[] b, final int off, final int len) { }
                        }),
                    md));
        for(final Iterator<String> it = names.iterator(); it.hasNext(); )
            {
            final String name = it.next();
            final ExhibitStaticAttr esa = _getStaticAttr(fileRoot, name);

            // Take care of the map...
            m.put(name, esa);

            // Take care of the timestamp...
            if(esa.timestamp > newest) { newest = esa.timestamp; }

            // Update the hash...
            dos.writeUTF(esa.filePath);
            dos.writeLong(esa.length);
            dos.writeLong(esa.timestamp);
            dos.writeByte(1); // Acts as a nearly-unique separator.
            }

        dos.flush();
        final byte[] hash = md.digest();

        // XOR lsbytes in `newest' until we run out of hash bytes...
        // ...first subtracting enough from the real timestamp to
        // make space for the hash
        // (ie so the computed value is no newer than the newest exhibit)...
        long fakeStamp = newest - (1 << (8 * TS_LSbytes_HASH));
        // XOR in LS bytes...
        for(int i = Math.min(hash.length, TS_LSbytes_HASH); --i >= 0; )
            { fakeStamp ^= ((long) (hash[i] & 0xff)) << (8 * i); }

        // Clamp the fake timestamp to within allowed limit values,
        // though this should not actually be necessary.
        // If there are no exhibits, make a zero timestamp.
        if(m.size() == 0) { fakeStamp = 0; }
        else { fakeStamp = Math.min(Math.max(1, newest), System.currentTimeMillis()); }

        // See if the client actually needs the value returned...
        if((oldStamp >= 0) && (oldStamp == fakeStamp))
            { return(null); } // No, client is up-to-date.

        // Now make the result!
        final AllExhibitImmutableData result = new AllExhibitImmutableData(m, fakeStamp);

        // Warn loudly if duplicate short names were found...
        final String dups[] = result.getFullNamesWithDuplicateShortNames();
        if(dups.length > 0)
            {
            for(int i = 0; i < dups.length; ++i)
                {
                System.err.println("WARNING: duplicate file component in: " + dups[i]);
                }
            }

        // Done!
        return(result);
        }

    /**Cache of most recent non-null answer to getAllExhibitProperties().
     * We avoid recomputing this if reasonable time has not passed,
     * so that, for example, multiple clients will not cause multiple
     * recomputations.
     * <p>
     * Accessed by getAllExhibitProperties() under the
     * _gAEP_cache_lock; may also be set in the constructor.
     * <p>
     * Since this object may be passed by reference it may be updated
     * by front-end or up-stream-precache activity, eg on the master.
     */
    private AllExhibitProperties _gAEP_cache;

    /**Timestamp of static cache file that _gAEP_cache was loaded from, or zero if not loaded from a file. */
    private long _gAEP_cache_file_ts;

    /**Time we should next recompute _gAEP_cache after.
     * Private to getAllExhibitProperties() and accessed under the
     * _gAEP_cache_lock.
     * <p>
     * Initially zero to force first recomputation.
     */
    private long _gAEP_cache_nextUpdate;

    /**Lock for _gAEP_cache. */
    private final Object _gAEP_cache_lock = new Object();

    /**Maximum number of parallel threads that we will use to load/compute properties; strictly positive.
     * Note that this work may be very memory- and CPU- intensive,
     * so we don't try for huge amounts of parallelism here;
     * maybe slightly more than the number of processors to to take up some
     * I/O slackness, though capped to avoid blowing up the heap.
     */
    private static final int CONCURRENCY_LIMIT =
        Math.min(4, Runtime.getRuntime().availableProcessors() + 1);

    /**If true, be eager about precomputing data for clients. */
    private static final boolean EAGER_PRECOMPUTE = true;

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed
     * then a default empty set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned) then null is returned.
     * <p>
     * This does not recompute its result more often than implied
     * by the global temporal-slackness value unless it loses its cache
     * due to memory shortage.
     */
    public AllExhibitProperties getAllExhibitProperties(final long oldHash)
        throws IOException
        { return(_getAllExhibitProperties(oldHash, true, false)); }

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed
     * then a default set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned)
     * then null is returned.
     * <p>
     * This does not recompute its result more often than implied
     * by the global temporal-slackness value unless it loses its cache
     * due to memory shortage.
     *
     * @param careful  if true, magic numbers of exhibits are checked and
     *     other extra-careful checking is done; this should be the default usage
     * @param forceLoadFromFilesystem  if true, ignore any cached file and
     *     refect any data from the filesystem,
     *     for example if rebuilding the cache file
     */
    private AllExhibitProperties _getAllExhibitProperties(final long oldHash,
                                                          final boolean careful,
                                                          final boolean forceLoadFromFilesystem)
        throws IOException
        {
        synchronized(_gAEP_cache_lock)
            {
            final long startTime = System.currentTimeMillis();

            // Cached copy is old iff we've waited long enough to check again without eating too much CPU
            // and the static cache file does not exist or has changed since when we last loaded it.
            final File staticCacheFileName = _getStaticCacheFileName();
            final boolean cachedCopyIsOld = (startTime > _gAEP_cache_nextUpdate) &&
                    ((_gAEP_cache_file_ts == 0) ||
                     (staticCacheFileName == null) ||
                     (_gAEP_cache_file_ts != staticCacheFileName.lastModified()));

            // Try to retrieve the cached copy...
            AllExhibitProperties cached = _gAEP_cache;

            // Note if we had retained a cached value.
            final boolean hadRetainedCache = (cached != null);

            // If it's not long enough since we last recomputed,
            // and if our cached copy hasn't been zapped due to memory shortage,
            // use the extant value without bothering the filesystem.
            // We do this because this activity is likely to be a very
            // heavy burden in I/O and CPU and we don't want all the
            // clients doing it individually!
            if((cached == null) || cachedCopyIsOld || forceLoadFromFilesystem) // RECOMPUTE!
                {
/* if(ORG.hd.d.IsDebug.isDebug) */ { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): starting: "+(new Date())+".]"); }

                // Turn back on any muted warnings.
                turnBackOnMutedWarnings();

                AllExhibitProperties newProps = null;

                // OK, if we're meant to be reading from a static
                // cache file, try it here...
                if(!forceLoadFromFilesystem && (staticCacheFileName != null))
                    {
                    try
                        {
                        try
                            {
                            newProps = (AllExhibitProperties)
                                FileTools.deserialiseFromFile(
                                    staticCacheFileName,
                                    true); // This file is compressed for speed...
                            }
                        catch(final IOException e)
                            {
                            // Retry once after a pause in case
                            // we happened to read the file while
                            // it was being updated (eg stale NFS file handle).
                            final int waitTime = 100000 + Rnd.fastRnd.nextInt((CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000) / 23);
                            System.err.println("ERROR reading static cache file: " + staticCacheFileName + "; will wait until "+(new Date(System.currentTimeMillis() + waitTime))+" and retry: " + e.getMessage());
                            // Wait at least long enough for cached (NFS) attributes to time out.
                            Thread.sleep(waitTime);
                            // Try again...
                            newProps = (AllExhibitProperties)
                                FileTools.deserialiseFromFile(
                                    staticCacheFileName,
                                    true); // This file is compressed for speed...
                            }
                        if(newProps != null)
                            {
                            // Note timestamp on cache file that we loaded from...
                            _gAEP_cache_file_ts = staticCacheFileName.lastModified();
                            // And note the internal date when exhibit data last changed.
                            System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): loaded static cache file marked unchanged since "+(new Date(newProps.hashNotChangedSince))+" at "+(new Date())+".]");
                            }
                        else
                            {
                            _gAEP_cache_file_ts = 0; // Note that we could not load from file.
                            }
                        }
                    catch(final Exception e)
                        {
                        _gAEP_cache_file_ts = 0; // Note that we could not load from file.

                        // If we don't revert in the absence of the static cache file,
                        // then fail now,
                        // else allow us to proceed to load in the normal way.
                        if(!REVERT_IF_SCF_NOT_AVAILABLE)
                            {
                            if(e instanceof IOException) { throw (IOException) e; }
                            throw new IOException("could not load static cache file: " + _getStaticCacheFileName());
                            }
                        // Complain, but plod on...
                        System.err.println("ExhibitDataFileSource: WARNING: could not load static cache file: " + _getStaticCacheFileName() + ": " + e.getMessage());
                        e.printStackTrace();
                        }
                    }
                else
                    {
                    // Note that we did not load from a static cache file...
                    _gAEP_cache_file_ts = 0; // Note that we could not load from file.
                    }

                // If we don't have the properties yet
                // then attempt to compute them.
                if(newProps == null)
                    {
                    final File baseDir = new File(LocalProps.getDataDir());

                    // Load the EPGI (should be quick: do it first).
                    final ExhibitPropsGlobalImmutable epgi = ExhibitPropsGlobalImmutable.loadFromDataDir(baseDir);
if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): loaded EPGI: "+epgi+"]"); }

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): recomputing AllExhibitImmutableData... " + ((cached == null) ? "[cached is null] " : "") + (new Date())+".]"); }
                    // First, get immutable core data unconditionally...
                    final AllExhibitImmutableData aeid =
                        _getAllExhibitImmutableData(baseDir.getPath(), -1, careful);

                    // Get sorted set of exhibit names.
                    // We hope that iterating over a sorted set will result in
                    // better filesystem performance than an unsorted set.
                    final String sortedNames[] = aeid.getAllExhibitNamesSorted();

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): recomputing loadedProperties over exhibit set of "+(sortedNames.length)+"... "+(new Date())+".]"); }

                    // Collect the loaded and computable properties.
                    // We do this in parallel to maximise I/O throughput,
                    // and use a ConcurrentHashMap as our working store as it is thread-safe.
                    // We get these using our non-CPU-bound thread pool
                    // and wait for them all to complete.
                    //
                    // We recover any EPC data that we can from any previous/cached AEP
                    // to save lots of time and effort.
                    final Map<String,ExhibitPropsLoadable> loadedProps = new ConcurrentHashMap<String, ExhibitPropsLoadable>(aeid.size() * 2);
                    final Map<String,ExhibitPropsComputable> computedProps = new ConcurrentHashMap<String, ExhibitPropsComputable>(aeid.size() * 2);
                    final List<Future<?>> tasks = new ArrayList<Future<?>>(sortedNames.length);
                    final ExhibitDataSource ds = makeExhibitDataSource();
                    final boolean recoverFromCache = (null != cached);
                    final AllExhibitProperties c = recoverFromCache ? cached : null;
                    for(final String name : sortedNames)
                        {
                        tasks.add(ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                            public final void run()
                                {
                                // We don't store null/EMPTY values.
                                try
                                    {
                                    // Always reload EPL data
                                    // else we will miss changes in description, etc.
                                    final ExhibitPropsLoadable epl = ExhibitPropsLoadable.getLoadableProperties(name, baseDir);
                                    if((epl != null) && !epl.equals(ExhibitPropsLoadable.EMPTY))
                                        { loadedProps.put(name, epl); }

                                    final ExhibitStaticAttr esa = aeid.getStaticAttr(name);

                                    ExhibitPropsComputable epc = null;
                                    if(recoverFromCache) 
                                        { epc = c.getExhibitPropsComputable(name); }
                                    if((null == epc) || epc.equals(ExhibitPropsComputable.EMPTY))
                                        { epc = ExhibitPropsComputable.createExhibitPropsComputable(esa, ds); }
                                    if((epc != null) && !epc.equals(ExhibitPropsComputable.EMPTY))
                                        { computedProps.put(name, epc); }
                                    }
                                catch(final Exception e) 
                                    { throw new RuntimeException("Could not create epl/epc for: " + name, e); }
                                }
                            }));
                        }
                    // Wait for all tasks to finish...
                    for(final Future<?> task : tasks)
                        {
                        try { task.get(); }
                        catch(final InterruptedException e)
                            {
                            final InterruptedIOException err = new InterruptedIOException(e.getMessage());
                            err.initCause(e);
                            throw err;
                            }
                        catch(final ExecutionException e)
                            {
                            final IOException err = new IOException(e.getMessage());
                            err.initCause(e);
                            throw err;
                            }
                        }

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): constructing AllExhibitProperties... "+(new Date())+"]"); }

                    // OK, construct and save the new value from the data gathered.
                    newProps = new AllExhibitProperties(cached,
                        epgi,
                        aeid,
                        loadedProps,
                        computedProps,
                        0);
                    }

                // An optimisation...
                // Only replace the old cache if there is no old value (ie it is null),
                // or the longHash or hashNotChangedSince values have changed.
                // This is so that if the master is also a Web server any cacheing work
                // done by the front-end may be effectively made available
                // for us to pass on precomputed to slaves, and we may thus reduce
                // front-end churn caused by apparent data changes.
                //
                // We run a small risk of missing some important change not
                // caught by the longHash, but the clients would miss it anyway!
                if((cached == null) ||
                   (cached.longHash != newProps.longHash) ||
                   (cached.hashNotChangedSince != newProps.hashNotChangedSince))
                    { _gAEP_cache = cached = newProps; }

                // Work out the minimum wait before starting the next computation.
                // Make sure that it will not happen within the
                // normal slackness interval and that we won't
                // spend more than 10% of real time doing it if possible.
                // We throw in a small random factor too, to help avoid collisions...
                final long endTime = System.currentTimeMillis();
                final long timeTaken = endTime - startTime;
                final long waitTime = Math.max(
                        CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * (1000 / 10), // Default wait time...
                        Math.min(10 * timeTaken, // ...limit time here to 10% of wall-clock time...
                        CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * (1000 * 10))) + // ...and limit by upper bound on wait time.
                    Rnd.fastRnd.nextInt(10001 + (CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 101));

                // Schedule next check.
                _gAEP_cache_nextUpdate = endTime + waitTime;
/* if(ORG.hd.d.IsDebug.isDebug) */ { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): done load: next check not before "+(waitTime/(1000 * 60))+"mins, at "+(new Date(_gAEP_cache_nextUpdate))+".]"); }
                }

            // If caller has an up-to-date copy, return null.
            if((oldHash >= 0) && (oldHash == cached.longHash))
                { return(null); }

            // Force update/computation of any computable properties
            // that we want to make available, precomputed, to slaves.
            // All possible values will have been salvaged from the
            // old object if we have replaced it.
            //
            // Only bother doing this if we had retained the cache
            // from the previous time, ie we are not likely to have
            // all our hard work GCed, and if we actually return
            // something other than null to a caller, ie they will
            // be using this copy.
            //
            // It probably makes sense to precompute attributes for
            // the largest exhibits first, since they are the ones
            // it is least good to have dragged across the network
            // for a client.
            //
            // We should attempt a bit of parallelism to maximise
            // what we get precached for the client in a given time,
            // given that the computations will probably be I/O bound.
            //
            // If we enocunter an I/O error we should probably simply
            // ignore it and let the client sort out any difficulties.
            // We should report the problem here though.
            //
            // We don't do any precomputation if we should have been
            // reading from a static cache file...
            if((_getStaticCacheFileName() == null) &&
               (EAGER_PRECOMPUTE || hadRetainedCache))
                {
                // We will spend no more than a small fraction of the time
                // taken so far in this routine on an eager precompute
                // else we may sigificantly prolong system a start-up, etc.
                final long now = System.currentTimeMillis();
                final long maxTimeToSpend =
                    Math.max(1001, // Spend a second at least...
                        (now - startTime) / 9);
                final long stopBy = now + maxTimeToSpend;

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): precomputing computable properties for at most "+maxTimeToSpend+"ms... "+(new Date())+"]"); }

                _doPrecompute(cached, stopBy);

                final long took = System.currentTimeMillis() - now;
if(ORG.hd.d.IsDebug.isDebug || (took > 2000)) { System.out.println("[ExhibitDataFileSource.getAllExhibitProperties(): done precompute at: "+(new Date())+" in "+(took/1000)+"s.]"); }

                // Defer next recompute time a little further.
                _gAEP_cache_nextUpdate += 2 * (System.currentTimeMillis() - now);
                }

            // Return current value!
            return(cached);
            }
        }

    /**Wrap this instance as an ExhibitDataSource.
     * Only really meant for internal use,
     * but may be useful for low-level use of the file-based data.
     */
    public AllExhibitProperties.ExhibitDataSource makeExhibitDataSource()
        {
        return(new AllExhibitProperties.ExhibitDataSource(){
                    public final void getRawFile(final ByteBuffer buf, final String exhibitName, final int position)
                        throws IOException
                        { ExhibitDataFileSource.this.getRawFile(buf, exhibitName, position, false); }
                    public final boolean isExhibitFullyLoaded(final ExhibitStaticAttr esa)
                        { return(true); }
                    });
        }

    /**Do precomputation of computable-and-immutable data.
     * A time-limit and copy to precompute values for are passed.
     */
    private static void _doPrecompute(final AllExhibitProperties aep,
                                      final long stopBy)
        {
        // Unsorted set of all ExhibitStaticAttr...
        final Set<ExhibitStaticAttr> asa = aep.aeid.getAllStaticAttrs();

        // Now sort to be largest first,
        // with ties broken first by timestamp
        // (we prefer older items as possibly more likely to be known
        // about and therefore viewed)
        // with remaining ties broken by the normal total sort order (by name).
        final ExhibitStaticAttr gettingSmaller[] =
            new ExhibitStaticAttr[asa.size()];
        asa.toArray(gettingSmaller);
        Arrays.sort(gettingSmaller, (new Comparator<ExhibitStaticAttr>(){
            public final int compare(final ExhibitStaticAttr e1, final ExhibitStaticAttr e2)
                {
                if(e1.length > e2.length) { return(-1); } // Correct.
                if(e1.length < e2.length) { return(+1); } // Wrong.
                // Break tie in favour of oldest first...
                if(e1.timestamp < e2.timestamp) { return(-1); } // Correct.
                if(e1.timestamp > e2.timestamp) { return(+1); } // Wrong.
                // Break tie unconditionally...
                return(e1.compareTo(e2));
                }
            }));

        // We make an ArrayList whose iterator() will
        // prematurely return false once the time of day
        // passes stopBy.
        final class AL extends ArrayList<ExhibitStaticAttr>
            {
            AL() { super(Arrays.asList(gettingSmaller)); }
            public final Iterator<ExhibitStaticAttr> iterator()
                {
                final Iterator<ExhibitStaticAttr> realIt = super.iterator();
                return(new Iterator<ExhibitStaticAttr>(){
                    /**Always return false if too much time has elapsed.
                     * Else return the real value.
                     */
                    public final boolean hasNext()
                        {
                        if(System.currentTimeMillis() >= stopBy)
                            { return(false); }
                        return(realIt.hasNext());
                        }
                    public final ExhibitStaticAttr next() { return(realIt.next()); }
                    public final void remove() { realIt.remove(); }
                    });
                }
            /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
            private static final long serialVersionUID = 2799693413319936828L;
            }

        // OK, attempt to bring computable data up to date in parallel...
        ParTasks.apply(new AL(),
            (new TaskProc(){
                /**Precompute the computable-and-immutable properties for one exhibit.
                 * The argument is the ExhibitStaticAttr.
                 */
                public final void fn(final Object o)
                    {
                    try {
                        final ExhibitStaticAttr esa = (ExhibitStaticAttr) o;
//if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[ExhibitDataFileSource: precomputing exhibit size "+esa.length+" name "+esa.filePath+".]"); }
                        aep.getExhibitPropsComputable(esa.filePath);
                        }
                    catch(final Throwable e) // Complain about any serious errors...
                        {
                        System.err.println("ERROR: ExhibitDataFileSource: serious error precomputing properties for exhibit ("+o+"): " + e.getMessage());
                        e.printStackTrace(); // Complain loudly.
                        }
                    }
            }),
            Math.min(CONCURRENCY_LIMIT, gettingSmaller.length),
            Math.max(ParTasks.UTILISATION_MAX / CONCURRENCY_LIMIT, ParTasks.UTILISATION_MIN));
        }

    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     * <p>
     * This computes a new uncached value on each call.
     */
    public GenProps getGenProps(final long oldStamp)
        throws IOException
        {
        final File file = new File(LocalProps.getConfDir(), CoreConsts.FS_CONF_SYSPROPS);

        // If the stamp presented is < 0 or different to that of the file,
        // reread the file now.
        final long newStamp = file.lastModified();
        if((oldStamp < 0) || (oldStamp != newStamp))
            {
            try
                {
                final FileInputStream fis = new FileInputStream(file);
                try
                    {
                    final Properties props = FileTools.loadProperties(new BufferedInputStream(fis));
                    return(new GenProps(props, newStamp));
                    }
                finally { fis.close(); }
                }
            catch(final IOException e)
                {
                e.printStackTrace(); // Dump a diagnostic...
                throw e; // Pass the error on.
                }
            }

        return(null); // Caller's copy seems to be up-to-date.
        }

    /**Gets the generic security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     * <p>
     * This computes a new uncached value on each call.
     * <p>
     * These generic properties are fetchable over the network, for example,
     * and need not be present locally at each host the JVM runs on.
     */
    public java.util.Properties getGenSecProps(final long oldStamp)
        throws IOException
        {
        final SecurityProps sp =
            SecurityProps.getSecurityPropsUncachedFromFilesystem(oldStamp);

        // If security props file has not changed, return null.
        if(sp == null)
            { return(null); }

        // Extract the relatively-safe "generic" component.
        return(sp.getGenSecProps());
        }

    /**Suffix appended to serialised thumbnail class data for each exhibit.
     * Such files are beside the original exhibit file,
     * are prefixed with FileTools.F_permPrefix,
     * and suffixed with this (which is chosen so as not to look like an exhibit).
     * <p>
     * The file is a GZIPped, serialised, ExhibitThumbnails class if present.
     */
    public static final String THUMBNAIL_SUFFIX = ".thumbnails";

    /**If true, the derialised thumbnail files are compressed (GZIPped) on disc.
     * This might cost a little time but might probably save a little space.
     * Note that these may well be GZIPped in transit to a slave server.
     */
    public static final boolean THUMBNAILS_ARE_GZIPPED = false;

    /**If true, we have warned about thumbnail directory absence.
     * Starts, false; may be cleared periodically.
     * Set true when we have warned.
     * <p>
     * Volatile to allow read/write without a lock.
     * <p>
     * Not critical to correct operation,
     * so we can tolerate races and sloppy handling.
     */
    private volatile boolean _haveWarnedNoTNDir;

    /**Turn back on warnings we have muted.
     * We do this occasionally so that persistent faults
     * will show up in the log and don't just vanish.
     * <p>
     * Avoids locking unless really necessary.
     */
    private void turnBackOnMutedWarnings()
        {
        _haveWarnedNoTNDir = false;
        }

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     * <p>
     * In this implementation we never try to create thumbnails
     * (we assume that the source directory is read only for example),
     * but we'll return one if we can find one.
     * <p>
     * Thumbnails are looked for by default in the exhibits directory
     * tree beside the source exhibits, but a LocalProps value
     * can override this so that source data can be separated
     * from derived data.
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible, else only return an existing one
     */
    public ExhibitThumbnails getThumbnails(final String name, final boolean create)
        throws IOException
        {
        // In the simple case where we know that thumbnails
        // definitely can't be created
        // (from the type, or because there is no exhibit)
        // return the NO_THUMBNAILS value immediately,
        // ignoring the create parameter.
        // We never cache this value.
        final ExhibitMIME.ExhibitTypeParameters type = (ExhibitMIME.getInputFileType(name));
        if((type == null) || (type.handler == null) ||
           !type.canPossiblyCreateThumbnailOfSameMIMEType())
            { return(ExhibitThumbnails.NO_THUMBNAILS); }

        final String dataDir = LocalProps.getDataDir();
        final ExhibitStaticAttr esa = _getStaticAttr(dataDir, name); // Avoid tickling cache badly...
        if(esa == null)
            { return(ExhibitThumbnails.NO_THUMBNAILS); }

        // Construct the name of the thumbnails file if it exists.
        final String relPath = thumbnailRelPath(name);

        // Compute full path from exhibit directory.
        final File topDir = (new File(dataDir, LocalProps.getThumbnailRelDir())).getCanonicalFile();

        // If the top directory for the thumbnails does not exist,
        // then we definitely will not be returning thumbnails.
        // We also warn when this directory does not exist.
        if(!topDir.isDirectory())
            {
            if(!_haveWarnedNoTNDir)
                {
                System.err.println("ExhibitDataFileSource: WARNING: top-level thumbnail directory absent: " + topDir);
                _haveWarnedNoTNDir = true;
                }

            // We don't return ExhibitThumbnails.NO_THUMBNAILS,
            // since thumbnails may be available/generated later.
            return(null);
            }

        // If the thumbnail file seems readable, etc, then try to load it.
        // Any problem, ie an IOException, will be propagated to the caller.
        final File fullPath = new File(topDir, relPath);
        if(fullPath.exists() && fullPath.canRead() && fullPath.isFile())
            {
            final ExhibitThumbnails result = (ExhibitThumbnails)
                FileTools.deserialiseFromFile(fullPath, THUMBNAILS_ARE_GZIPPED);
            return(result);
            }

        // Can't seem to find a thumbnail file,
        // so tell the caller no thumbnails at the moment.
        // We don't return ExhibitThumbnails.NO_THUMBNAILS,
        // since thumbnails may be available/generated later.
        return(null);
        }

    /**Poll periodically.
     * Periodically save the system variables.
     */
    public void poll(final GenProps gp)
        {
        _saveSystemVariables(varMgr, false);
        }

    /**Save the system variables.
     * Mainly save the histories of persistent events.
     * <p>
     * We save these with a history so that in effect we have backups
     * if something bad happens
     * (though this implies that some other housekeeping will
     * have to clear up excess copies, etc,
     * in order to avoid space requirements growing without bound).
     * <p>
     * The data is saved in the persistent data area.
     * <p>
     * Reports but does not propagate exceptions.
     *
     * @param vars  set of variables to save
     * @param force  if true, force an immediate complete save
     */
    private void _saveSystemVariables(final BasicVarMgr vars,
                                      final boolean force)
        {
        // Return immediately if too soon to save again
        // and a save is not being forced.
        if(!force && (_nextVarSave > System.currentTimeMillis())) { return; }

        // Keeps a history and attempts just to save changes...
        try
            {
            // First write out the persistent events as-is
            _appendToEventLogs();

            // Unless doing a forced save,
            // do an incremental save to be as quick as reasonably possible.
            vars.saveEventHistories(_getEventHistoryStorageDir(), true, true, !force);
            // Note that save seems to have completed fine.
            _nextVarSave = System.currentTimeMillis() + _VAR_SAVE_INTERVAL_MS +
                    Rnd.fastRnd.nextInt(1 + (_VAR_SAVE_INTERVAL_MS/3));
            }
        catch(final IOException e)
            { e.printStackTrace(); }
        }



    /**Our private static instance of a GMT calendar.
     * No one must change the ZONE or DST offset.
     */
    private static Calendar GMTCalendar = Calendar.getInstance();
    /**Initialise GMTCalendar. */
    static
        {
        // Set to GMT
        // i.e 0 Offset
        GMTCalendar.set(Calendar.ZONE_OFFSET, 0);
        // and 0 Daylight savings
        GMTCalendar.set(Calendar.DST_OFFSET, 0);
        }

    /**Format we use to insert into event history file name. */
    private static final SimpleDateFormat dateFmtEHFile =
        new SimpleDateFormat("yyyyMMdd");
    /**Set the Calendar for dateFmtEHFile to be the GMT calendar. */
    static { dateFmtEHFile.setCalendar(GMTCalendar); }

    /**Format we use for timestamp to insert into event log file name. */
    private static final SimpleDateFormat timestampFmtEHFile =
        new SimpleDateFormat("yyyyMMdd-HHmmss.SSS");
    /**Set the Calendar for dateFmtEHFile to be the GMT calendar. */
    static { timestampFmtEHFile.setCalendar(GMTCalendar); }

    /**Max log line length for event log. */
    private static final int MAX_EV_LOG_LINE_LENGTH = 4095;

    /**Private lock for _appendToEventLogs(). */
    private static final Object _aTWL_lock = new Object();

    /**Write out the accumulated event values to disc.
     * These are written to date-stamped files to allow automatic "rolling".
     * <p>
     * We keep a separate file for each named event,
     * and append events to each appropriate file in order
     * with a record of the form:
     * <pre>
     *     YYYYMMDD-HHmmss.SSS eventName eventValue
     * </pre>
     * with a null value being indicated by the value "null".
     * <p>
     * Note that the timestamp is that of the value,
     * not the time that the event is received or written.
     * <p>
     * We may cache file handles to reduce the cost of opening and closing
     * files for each event,
     * but in any case will have closed all log files opened by us
     * by the time this routine returns.
     * <p>
     * We synchronise on a private static lock to try to prevent two instances of this
     * running concurrently and leading to file corruption.
     */
    private void _appendToEventLogs()
        throws IOException
        {
        // Attempt to write out accumulated event values to the logs.
        final List<SimpleVariableValue> events = new ArrayList<SimpleVariableValue>(eventsToLog.size());
        // Move all events into working copy atomically to hold lock for short time.
        synchronized(eventsToLog) { events.addAll(eventsToLog); eventsToLog.clear(); }

        // If nothing to write then return immediately.
        final int nEvents = events.size();
        if(nEvents == 0) { return; }


//System.out.println("[EVENTS TO LOG: " + nEvents + "...]");

        // Separate the events into groups (preserving order) by event name
        // so that we can do one file operation for all instances of one event.
        final Map<String, List<SimpleVariableValue>> groupedEvents = new HashMap<String, List<SimpleVariableValue>>();
        for(final SimpleVariableValue svv : events)
            {
            final String name = svv.getDef().getName();
            List<SimpleVariableValue> evvs = groupedEvents.get(name);
            if(evvs == null)
                {
                evvs = new ArrayList<SimpleVariableValue>();
                groupedEvents.put(name, evvs);
                }
            evvs.add(svv);
            }

        // Check the log directory is writable.
        // Complain loudly and discard the events if not.
        final File logDir = _getEventLogDir();
        if((logDir == null) || !logDir.isDirectory() || !logDir.canWrite())
            {
            System.err.println("WARNING: persistent event log dir "+logDir+" not usable: discarding "+nEvents+" event(s).");
            return;
            }

        // Create/format the date string once, if we are going to use it.
        final String dateString = "." + dateFmtEHFile.format(new Date());

        // Any error accumulated so far...
        IOException err = null;

        synchronized(_aTWL_lock)
            {
            for(final String name : groupedEvents.keySet())
                {
                final StringBuilder sb = new StringBuilder(80);
                sb.append("eventLog.");
                sb.append(name);
                sb.append(BasicVarMgr.EVENT_STORE_NAMETERM);
                sb.append(dateString);
                sb.append(".log");
                final File f = new File(logDir, sb.toString());

                try
                    {
                    final PrintWriter pw = new PrintWriter(new FileWriter(f, true)); // Append to existing file.
                    try
                        {
                        for(final SimpleVariableValue svv : groupedEvents.get(name))
                            {
                            final StringBuilder logLine = new StringBuilder(80);

                            logLine.append(timestampFmtEHFile.format(new Date(svv.getTimestamp())));
                            logLine.append(' ');
                            logLine.append(name);
                            logLine.append(' ');
                            logLine.append(svv.getValue());

                            // Truncate VERY long entries...
                            if(logLine.length() > MAX_EV_LOG_LINE_LENGTH)
                                {
                                logLine.setLength(MAX_EV_LOG_LINE_LENGTH - 3);
                                logLine.append("..."); // Indicate truncation.
                                }

                            // We sanitise the log line to ensure that it
                            // contains no control codes especially CRs or LFs
                            // that would disrupt the record-per-line format.
                            // Replace such toxic chars with something innocuous.
                            for(int i = logLine.length(); --i >= 0; )
                                {
                                final char ch = logLine.charAt(i);
                                if((ch < 32) || (ch > 255) || (ch == 127))
                                    { logLine.setCharAt(i, ' '); }
                                }

                            // Write entry in own line to EOF...
                            pw.println(logLine.toString());
                            }

                        // Flush to disc...
                        pw.flush();
                        }
                    finally { pw.close(); }
                    }
                catch(IOException e)
                    {
                    e.printStackTrace();
                    err = e; // Absorb error for now.  Rethrow later.
                    }
                }
            }

        // If we encountered any error then (re)throw a sample one here.
        if(err != null) { throw err; }
        }

    /**Minimum interval between saves, ms; strictly positive.
     * Intended to avoid losing too much data at shut-down,
     * yet avoid wearing out the disc and consuming CPU time
     * with redundant work.
     * <p>
     * Something from tens of seconds to a few minutes is probably about right.
     */
    private static final int _VAR_SAVE_INTERVAL_MS = 3*60*1000;

    /**Time after which to do the next save (zero if no save yet completed).
     * Marked volatile to allow lock-less access.
     */
    private transient volatile long _nextVarSave;


    /**Get directory used for storing system-variable persistent event histories; never null.
     * Though never null, the indicated directory may not exist
     * or may otherwise not be usable.
     * <p>
     * We get derive this from the LocalProps value.
     */
    private static File _getEventHistoryStorageDir()
        {
        return(new File(LocalProps.getPersistentStateDir(), "history"));
        }

    /**Get directory used for storing system-variable persistent event logs; never null.
     * Though never null, the indicated directory may not exist
     * or may otherwise not be usable.
     * <p>
     * We get derive this from the LocalProps value.
     */
    private static File _getEventLogDir()
        {
        return(new File(LocalProps.getPersistentStateDir(), "log"));
        }

    /**Our set of system variables.
     * Persistent values are loaded from disc on first attempt to
     * get and variable(s) if possible.
     * <p>
     * We expect this to be at the top of the data pipe on the master server,
     * and thus handle all the system non-local variables,
     * plus all the local variables of the master.
     * <p>
     * If any persistent values are set then they all get written to disc
     * as compressed serialised data on the next poll().
     * (Doing saves on each poll() helps reduce expensive disc writes by
     * grouping multiple updates into one save if they are happening rapidly.)
     * <p>
     * The current set or variables (including locals) is held in memory,
     * which also allows us to merge globals if required.
     * <p>
     * The varMgr is marked as an end-point so that it will
     * generate and return a unique local system ID.
     */
    private final BasicVarMgr varMgr = new BasicVarMgr(true); // Endpoint.

    /**In-order List/Queue thread-safe list of persistent events to be logged; never null.
     * Remove items FIFO or lock the queue and remove all the items in one go.
     * <p>
     * We don't synchronously save these data.
     */
    private final Vector<SimpleVariableValue> eventsToLog = new Vector<SimpleVariableValue>();

    /**Set variable value; persistent values will eventually go to disc.
     *
     * @param newValue
     * @throws IOException
     */
    public void setVariable(final SimpleVariableValue newValue)
        throws IOException
        {
        varMgr.setVariable(newValue);

        // Capture any persistent event for later logging.
        if(newValue != null)
            {
            final SimpleVariableDefinition def = newValue.getDef();
            if(def.isEvent() && def.isPersistent() && SystemVariables.defs.contains(def))
                { eventsToLog.add(newValue); }
            }
        }

    /**Set variable values; persistent values will eventually go to disc.
     *
     * @param newValues
     * @throws IOException
     */
    public int setVariables(final SimpleVariableValue[] newValues)
        throws IOException
        {
        final int n = varMgr.setVariables(newValues);

        // Capture any persistent events for later logging, in order.
        for(final SimpleVariableValue svv : newValues)
            {
            if(svv != null)
                {
                final SimpleVariableDefinition def = svv.getDef();
                if(def.isEvent() && def.isPersistent() && SystemVariables.defs.contains(def))
                    { eventsToLog.add(svv); }
                }
            }

        return(n);
        }

    /**Get variable value; persistent values may have come from disc.
     */
    public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
        {
        return(varMgr.getVariable(var));
        }

    /**Get variable values; persistent values may have come from disc.
     */
    public SimpleVariableValue[] getVariables(final long changedSince)
        {
        return(varMgr.getVariables(changedSince));
        }

    /**Synchronise variable values.
     * @param force  if true, any changed persistent values will usually be
     *     synchronously written to persistent store though attempts to sync
     *     more than is sensible/safe/meaningful for the underlying medium
     *     (eg more than every minute or so for disc) may be ignored;
     *     changed persistent values will be periodically written to disc
     *     even if this is not called with the "true" argument
     */
    public void syncVariables(final boolean force)
//        throws IOException
        {
        // Nothing needed in this implementation.
        // In particular this does not flush values to disc.
        }

    /**Get the current partial, or previous full, event set at the specified interval; never null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     * <p>
     * If a "previous" value returned from our local store is not marked
     * as authoritative, then we convert it to an authoratitve value,
     * write that back to our local store and return it.
     * (The "current" value is generally not authoritative.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; may be empty but never null if requested set not available
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        final EventVariableValue eventValue = varMgr.getEventValue(def, intervalSelector, current);
        if(!current && !eventValue.isAuthoritative())
            {
            final EventVariableValue authValue = eventValue.makeAuthoritative();
            varMgr.setEventValue(authValue);
            return(authValue);
            }
        return(eventValue);
        }

    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * (An implementation that is not an end-point may go upstream to fetch
     * missing values and cache them to satisfy future requests.)
     * <p>
     * If a "previous" value (older than the current interval)
     * returned from our local store is null or not marked
     * as authoritative, then we convert it to an authoratitve value,
     * write that back to our local store and return it.
     * (The "current" value is generally not authoritative,
     * and entries in the future cannot be so.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param intervalNumber  a time (as from System.currentTimeMillis())
     *     which identifies the first interval for which data is potentially
     *     required; if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException
        {
        // Compute currentInterval before getting values
        // to avoid a race which might make wrongly capture
        // an incomplete "current" value as authoritative and complete.
        final long currentInterval = intervalSelector.getIntervalNumber(System.currentTimeMillis());

        synchronized(varMgr) // Avoid races with new events coming in...
            {
            EventVariableValue[] result = varMgr.getEventValues(def, intervalSelector, intervalNumber, whichValues);

            // Ensure result array is large enough,
            // else copy into large enough one.
            final int wvl = (whichValues == null) ? 1 : whichValues.length();
            if(result.length < wvl)
                {
                final EventVariableValue newResult[] = new EventVariableValue[wvl];
                System.arraycopy(result, 0, newResult, 0, result.length);
                result = newResult;
                }

            // For any values newer than the current interval,
            // and for which our local store had missing or non-authoritative values,
            // convert them to authoritative values
            // so that upstream users can cache them permanently.
            BitSet wv = whichValues;
            if(wv == null)
                {
                // Create synthetic BitSet for simplicity below.
                wv = new BitSet(1);
                wv.set(0);
                }
            for(int i = wv.nextSetBit(0); i >= 0; i = wv.nextSetBit(i+1))
                {
                final long slotIntervalNumber = intervalNumber - i;
                // We cannot make current/future events authoritative.
                if(slotIntervalNumber >= currentInterval)
                    { continue; }

                final EventVariableValue evv = result[i];
                if((evv == null) || !evv.isAuthoritative())
                    {
                    // Make value authoritative,
                    // creating a new empty one if need be.
                    final EventVariableValue authValue = (evv != null) ? evv.makeAuthoritative() :
                        new EventVariableValue(true,
                                               def,
                                               intervalSelector,
                                               slotIntervalNumber,
                                               0, null, null);

                    assert(authValue.isAuthoritative());
                    assert((evv == null) || (authValue.getIntervalNumber() == evv.getIntervalNumber()));
                    assert((evv == null) || (authValue.getDef().equals(evv.getDef())));

                    // Return authoritative value to caller.
                    result[i] = authValue;

                    // Save in our local store.
                    varMgr.setEventValue(authValue);

//if(IsDebug.isDebug) { System.out.println("Storing authoritative " + authValue); }
                    }
                }

            return(result);
            }
        }


//    /**Makes an exhibit backup list file.
//     * This is a plain-ASCII file, one file per line,
//     * of exhibits and per-exhibit property/attribute files,
//     * relative to the top of the exhibit data tree,
//     * that should be backed up.
//     * <p>
//     * These are collected in sorted order.
//     *
//     * @param careful  if true, magic numbers of exhibits are checked and
//     *     other extra-careful checking is done;
//     *     this should be the default usage
//     */
//    public static void makeExBackupList(File filename, boolean careful)
//        throws IOException
//        {
//        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
//        final PrintWriter pw = new PrintWriter(baos);
//
//        final String dataDir = LocalProps.getDataDir();
//        if(dataDir == null)
//            { throw new IOException("Data directory not set in LocalProps"); }
//        final File exhibitBaseDir = new File(dataDir);
//
//        System.out.println("[ExhibitDataFileSource.makeExBackupList(): started: " +(new Date())+ ".]");
//        final AllExhibitImmutableData aeid = ExhibitDataFileSource._getAllExhibitImmutableData(dataDir, -1L, careful);
//
//        final String[] names = aeid.getAllExhibitNamesSorted();
//        for(int i = 0; i < names.length; ++i)
//            {
//            final String exhibitName = names[i];
//            // Get and sort all attribute files, if any, for this exhibit.
//            final SortedSet<String> attrFiles = new TreeSet<String>(ExhibitPropsLoadable.listPropsFiles(exhibitName, exhibitBaseDir));
//
//            // List exhibit and then its attribute files, if any.
//            pw.println(exhibitName);
//            for(final Iterator<String> it = attrFiles.iterator(); it.hasNext(); )
//                {
//                final String name = it.next();
//                pw.println(name);
//                }
//            }
//
//        // OK, save the list to file.
//        pw.flush();
//        FileTools.replacePublishedFile(filename.getPath(), baos.toByteArray());
//        pw.close();
//
//        System.out.println("[ExhibitDataFileSource.makeExBackupList(): finished: " +(new Date())+ ".]");
//        }

    /**Creates the name of a thumbnail starting at the thumbnail base directory from a full exhibit name.
     */
    private static String thumbnailRelPath(final String exhibitName)
        {
        // Construct the name of the thumbnails file if it exists.
        return(ExhibitName.getDirComponent(exhibitName) + File.separatorChar +
                ExhibitName.getFileComponent(exhibitName) + THUMBNAIL_SUFFIX);
        }

    /**Creates thumbnails from all exhibits possible.
     * This is used to create exhibit thumbnails off-line,
     * taking as long and as much memory as required
     * and processing the resource-intensive thumbnail building sequentially.
     * <p>
     * These luxuries of non-real-time and lots of memory and CPU are denied
     * to the WAR/EAR servers when they try to compute thumbnails on the fly,
     * since there may be many simultaneous users and even attempted
     * Denial-of-Service attacks.
     * <p>
     * The thumbnails created here can be loaded by master servers,
     * and through it any slaves, thus making a much wider range of thunmbnails
     * available than could be computed on the fly.
     * <p>
     * We will protest but terminate without an exception if the
     * thumbnail directory does not exist; we will assume that precomputed
     * thumbnails are not required.
     * <p>
     * We will only attempt to (*re) compute a thumbnail if:
     * <ul>
     * <li>The exhibit type is one that we can compute thumbnails for.
     * <li>The thumbnail file does not exist or is older than the exhibit file.
     * </ul>
     * <p>
     * This will attempt to put thumbnail files where the getThumbnails()
     * routine will expect to find them.
     */
    private static void createThumbnails(final AllExhibitProperties aep,
                                         final long stopBy)
        throws IOException
        {
        // Compute full path from exhibit directory.
        final File topDir = (new File(LocalProps.getDataDir(), LocalProps.getThumbnailRelDir())).getCanonicalFile();

        // If the top directory for the thumbnails does not exist,
        // then we definitely will not be returning thunbnails.
        // We also warn when this directory does not exist.
        if(!topDir.isDirectory() || !topDir.canWrite())
            {
            // We might wish to limit frequency with which we report missing directory!
            System.err.println("Warning: top-level thumbnail directory absent: " + topDir);
            System.err.println("Aborting creation of thumbnails.");
            return;
            }

        // Make an instance of the data retriever for use by the thumbnail generator.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource(aep);
        final AllExhibitProperties.ExhibitDataSource eds = edfs.makeExhibitDataSource();

        // Now examine each exhibit in turn
        // (in sorted order to try to approximate a disc-friendly breadth-first search)
        // and if it is possible to build thumbnails
        // and they don't exist or are older than the exhibit,
        // try to build the thumbnails and save them.
        //
        // This tries to be robust and avoid blowing up if a single thumbnail build has problems.
        final String names[] = aep.aeid.getAllExhibitNamesSorted();
        
        // Queue of tasks in progress.
        final Queue<Future<?>> tasks = new LinkedBlockingQueue<Future<?>>();

        for(int i = names.length; (--i >= 0) && (System.currentTimeMillis() <= stopBy); )
            {
            final String name = names[i];

            // In the simple case where we know that thumbnails
            // definitely can't be created
            // (from the type, or because there is no exhibit)
            // return the NO_THUMBNAILS value immediately,
            // ignoring the create parameter.

            final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(name);
            if(esa == null)
                { continue; }

            final ExhibitMIME.ExhibitTypeParameters type = (ExhibitMIME.getInputFileType(esa.filePath));
            if((type == null) || (type.handler == null) ||
               !type.canPossiblyCreateThumbnailOfSameMIMEType())
                { continue; }

            // These tasks are assumed to be heavy on CPU, and not mainly I/O-bound.
            tasks.add(ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    // Construct the name of the thumbnails file if it exists.
                    final String relPath = thumbnailRelPath(name);

                    // If the thumbnail file seems to exist and is readable,
                    // and is newer than the exhibit,
                    // then assume that it is up-to-date.
                    final File fullPath = new File(topDir, relPath);
                    if(fullPath.exists() && fullPath.canRead() && fullPath.isFile() &&
                        (fullPath.lastModified() > esa.timestamp))
                        { return; }

                    // OK, looks like we need to (re)build the thumbnails...
                    try {
                        // Ensure that the destination directory exists.
                        final File targetDir = fullPath.getParentFile();
                        if(!targetDir.exists())
                            { targetDir.mkdirs(); }

                        // OK, attempt to generate thumbnails here, not under any lock.
                        final ExhibitThumbnails tns = type.handler.makeThumbnails(
                            esa, eds, aep,
                            true);  // Ignore resource limitations; this our best chance!

                        // Attempt to save thumbnails (atomically), without additional compression.
                        FileTools.serialiseToFile(tns, fullPath, THUMBNAILS_ARE_GZIPPED, false);
                        }
                    catch(final ThreadDeath e)
                        { throw e; } // Don't intercept this one...
                    catch(final Throwable e)
                        {
                        System.err.println("ERROR: unable to build thumbnails for: " + name);
                        e.printStackTrace();
                        return; // Attempt to continue.
                        }
                    }
                }));

            // Optimisation: release resources ASAP from any early completed tasks.
            Future<?> headTask;
            while(null != (headTask = tasks.peek()))
                { if(headTask.isDone()) { tasks.remove(); } }
            }

        // Wait for all remaining tasks.
        while(!tasks.isEmpty())
            {
            try { tasks.remove().get(); }
            catch(final Exception e)
                {
                final IOException err = new IOException("did not complete");
                err.initCause(e);
                throw err;
                }
            }
        }

    /**Computes an AllExhibitProperties object and saves it to outputFile.
     * The object is saved serialised and GZIPed.
     * <p>
     * The save is avoided if we are not forcing a complete recompute
     * and the exhibit hash has not changed,
     * ie we try to avoid unnecessarily churning the filesystem
     * and/or the master server.
     * <p>
     * It almost certainly does not make sense for the
     * WAR_SYSPROPNAME_WARONLY_STATICCACHEFILE property to be set;
     * you could get a similar result by copying the file.
     * <p>
     * In passing, this may update anciliary state such as computable properties
     * and thumbnails.
     *
     * @param outputFile desired location of cache file; not null
     * @param quick  if true, attempt to be as quick as possible,
     *     else if false, magic numbers of exhibits are checked and
     *     other extra-careful checking is done;
     *     this should be the default usage
     * @param recompute  if true, recompute all derived data from scratch,
     *     eg do not reload any extant cache file
     *
     * @return AllExhibitProperties (non-null)
     */
    private static AllExhibitProperties createStaticCacheFile(final File outputFile,
                                                              final boolean quick,
                                                              final boolean recompute,
                                                              final boolean fixaccessions)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();
        
        // Submit task to update accessions concurrently.
        final Future<?> accessionsUpdate = ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                try
                    {
                    // Get a simple listing of exhibits on disc.
                    final AllExhibitImmutableData aeid = _getAllExhibitImmutableData(
                        LocalProps.getDataDir(),
                        -1,
                        !quick);
                    assert(aeid != null);

                    // Create/fix accessions files first.
                    createAccessionFiles(aeid, recompute, fixaccessions);
                    }
                catch(final IOException e)
                    {
                    e.printStackTrace();
                    throw new Error("failed to create accessions", e);
                    }
                }
            });

        // Pre-loaded state to save time and preserve hash values...
        AllExhibitProperties extantAep = null;
        Long extantAepLongHash = null;

        // If we're being quick then we can't be too careful...
        final boolean careful = !quick;

        // If the old cache file exists and we are not forcing a full recompute
        // then try to preload it.
        if(!recompute && outputFile.isFile())
            {
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): reloading old static cache file: " +outputFile+ ".]");
            try {
                extantAep = (AllExhibitProperties) FileTools.deserialiseFromFile(
                        outputFile,
                        true); // This file is compressed for speed...
                if(extantAep != null)
                    {
                    extantAepLongHash = new Long(extantAep.longHash);
                    System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): preloaded static cache file marked unchanged since "+(new Date(extantAep.hashNotChangedSince))+".]");
                    }
                }
            catch(final Exception e)
                {
                System.err.println("WARNING: failed to preload from file ``" + outputFile + "''.");
                }
            }

        // Wait for accessions update to complete
        // so that we can incorporate any new files in the AEP.
        try { accessionsUpdate.get(); }
        catch(final Exception e) { throw new Error(e); }

        // Create an instance, possibly preloaded.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource(extantAep);

        // Now create and save the cache (if changed)...
        System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): started: " +(new Date())+ ".]");
        // Get the properties...
        // Force reconstruction of AEP from filesystem.
        final AllExhibitProperties aep = edfs._getAllExhibitProperties(-1L, careful, true);
        System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): *** exhibit count: " + aep.aeid.size() + ".]");
        // Save to file, GZIPed, complaining about anything nasty that we notice.
        // But only do the save if we are forcing a complete recompute or the hash has changed,
        // ie don't force the master server to reload it for no reason.
        if(recompute || !(new Long(aep.longHash)).equals(extantAepLongHash))
            {
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): longHash="+aep.longHash+", extantAepLongHash="+extantAepLongHash+ ".]");

            // Before saving, do some precomputing (if we're going to save to file).
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): precomputing: " +(new Date())+ ".]");
            // If we're going to save the new file then recompute everything needed if not quick,
            // else just run for a small multiple of the time we already spent to amortise overheads.
            // This will allow us to incrementally update some slow-to-compute state for the master.
            final long timeSoFar = Math.max(30000, 7 * (System.currentTimeMillis() - startTime));
            try { _doPrecompute(aep, (!quick) ? Long.MAX_VALUE : (System.currentTimeMillis() + timeSoFar)); }
            catch(final Exception e) { e.printStackTrace(); } // Print errors but continue if possible.

            // OK, do the same.
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): saving: " +(new Date())+ ".]");
            FileTools.serialiseToFile(aep, outputFile, true, false);
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): done: " +(new Date())+ ".]");
            }
        else
            {
            // Let user know that we decided NOT to save.
            System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): NOT SAVING: no changes found in exhibit data.]");
            }

        return(aep);
        }

    /**Create/fix accessions files.
     * @param recompute  if true, force all accessions files to be recomputed
     *     and tested against those on disc; may be very slow
     * @param fixaccessions  if true, attempt to fix existing accessions files
     *     if necessary
     *
     * TODO: need to be able to check and/or fix accessions data
     *
     * @return true if any accessions files were created or fixed
     */
    private static boolean createAccessionFiles(final AllExhibitImmutableData aeid,
                                                final boolean recompute,
                                                final boolean fixaccessions)
        throws IOException
        {
        if(recompute)
            {
            System.err.println("WARNING: cannot recompute accessions files yet");
            }
        if(fixaccessions)
            {
            System.err.println("WARNING: cannot fix accessions files yet");
            }

        final AtomicBoolean result = new AtomicBoolean(); // No updates/fixes/creations yet...
        final AtomicReference<String> couldNotMakeAccession = new AtomicReference<String>();

        final String dataDir = LocalProps.getDataDir();

        // Check accession files exist for all exhibits.

        // Queue of tasks yet to complete...
        final Queue<Future<?>> tasks = new LinkedBlockingQueue<Future<?>>();

        // Do this in approximately sorted order so as to improve disc throughput...
        for(final String exhibitName : aeid.getAllExhibitNamesSorted())
            {
            // These tasks are assumed to be light on CPU, and I/O-bound.
            tasks.add(ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    try
                        {
                        final String newAccessionFilename = ExhibitPropsLoadable.relPathToNewAccession(exhibitName);
                        final File newAccessionFile = new File(dataDir, newAccessionFilename);
                        if(!newAccessionFile.exists())
                            {
                            // Compute the accession data...
                            final AccessionData ad = AccessionData.fromExhibitFile(new File(dataDir, exhibitName));
                            // Convert to (terse) XML.
                            final String xml = TextUtils.toXML(ad.getAsDOM(), false, true);
                            // Convert to UTF-8 text bytes ready to save.
                            final byte utf8[] = xml.getBytes(CoreConsts.FILE_ENCODING_UTF_8);
                            // Save new accessions file as near atomically as possible.
                            if(FileTools.replacePublishedFile(newAccessionFile.toString(), utf8, false))
                                { result.set(true); }
                            }
                        }
                    catch(final IOException e)
                        {
                        // Whinge but continue if we encounter an error.
                        e.printStackTrace();
                        couldNotMakeAccession.set("IOException inspecting/making accession file for: " + exhibitName + ": " + e.getMessage());
                        }
                    }
                }));
            
            // Optimisation: release resources ASAP from any early completed tasks.
            Future<?> headTask;
            while(null != (headTask = tasks.peek()))
                { if(headTask.isDone()) { tasks.remove(); } }
            }
        
        // Wait for all remaining tasks.
        while(!tasks.isEmpty())
            {
            try { tasks.remove().get(); }
            catch(final Exception e)
                {
                final IOException err = new IOException("did not complete");
                err.initCause(e);
                throw err;
                }
            }

        if(couldNotMakeAccession.get() != null)
            {
            System.err.println("ERROR: could not make at least one missing (new) accession file: ");
            System.err.println("    REASON: " + couldNotMakeAccession);
            }

        return(result.get());
        }

    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(final PropsKey key,
                                              final long versionID)
        throws IOException
        {
        throw new IOException("NOT IMPLEMENTED");
        }


    /**Run from the command-line with a single argument (the output file name).
     * There is an optional <code>-quick</code> first argument,
     * which if present which attempts to cap the time spent in one run
     * (typically to no more than a few minutes)
     * so that any work to be done can be done incrementally
     * in several passes if need be.
     * <p>
     * When making a new static cache file we always try to reload the old one
     * (unless the <code>-recompute</code> flag is set, in which case we will not load it)
     * to save lots of time (especially for the computed properties)
     * and possibly preserve the hashNotChangedSince value.
     * <p>
     * If the <code>-fixaccessions</code> flag is passed then we try to fix existing
     * accessions files that are missing information, eg new checksum types,
     * though we will never try to "fix" an incorrect checksum for example
     * (that will always result in a warning on the console.)
     * <p>
     * If the <code>-checkexhibits</code> flag is true
     * then we try to verify that each exhibit's exhibit length, timestamp and hash
     * matches the values captured in the associated accession file.
     * We may check additional data (eg FEC data) in future.
     * Note that this option overrides any other options,
     * and avoids writing anything to disc.
     */
    public static void main(final String args[])
        {
        final long startTime = System.currentTimeMillis();

        if(args.length < 1)
            {
            System.err.println("Expects: " +
                "[-quick | -recompute | -fixaccessions | -checkexhibits ] " +
                "filename");
            System.exit(1);
            }

        boolean quick = false; // If true, use quicker exhibit-collection method.
        boolean recompute = false; // If true, force recompute from scratch; do not reload old cache file.
        boolean fixaccessions = false; // If true, fix existing accessions files.
        boolean checkexhibits = false; // If true, verify exhibit files are not corrupt.
        for(int i = args.length - 1; --i >= 0; )
            {
            if("-quick".equals(args[i])) { quick = true; }
            else if("-recompute".equals(args[i])) { recompute = true; }
            else if("-fixaccessions".equals(args[i])) { fixaccessions = true; }
            else if("-checkexhibits".equals(args[i])) { checkexhibits = true; }
            else
                {
            	System.err.println("Unknown flag '" + args[i] + "'.");
            	System.exit(1);
            	return;
            	}
            }

        System.out.println("[ExhibitDataFileSource: START: " +(new Date(startTime))+ ".]");

        // Get the filename to load/save from/to; always last.
        final File staticCacheFilename = new File(args[args.length - 1]);

        try {
            if(checkexhibits)
                {
            	checkExhibitData(quick, staticCacheFilename.getParentFile());
                }
            else
                {
                // Create static cache file and related state.
                final AllExhibitProperties aep = createStaticCacheFile(staticCacheFilename, quick, recompute, fixaccessions);
                assert(aep != null);

                // Create any thumbnails that are missing.
                System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): thumbnail building starting: " +(new Date())+ ".]");
                final long timeSoFar = Math.max(30000, 7 * (System.currentTimeMillis() - startTime));
                createThumbnails(aep, (!quick) ? Long.MAX_VALUE : (System.currentTimeMillis() + timeSoFar));
                System.out.println("[ExhibitDataFileSource.createStaticCacheFile(): thumbnail building finished: " +(new Date())+ ".]");
                }

            final long endTime = System.currentTimeMillis();
            final long runTime = endTime - startTime;
            System.out.println("[ExhibitDataFileSource: END: " +(new Date(endTime))+ ": "+runTime+"ms.]");
            }
        catch(final Exception e)
            {
            e.printStackTrace();
            System.exit(1); // Exit with error code.
            }
        }

	/**Check exhibit data for corruption.
	 * @param quick  take some optional speed-ups
	 * @throws IOException  in case of difficulty
	 */
	private static void checkExhibitData(final boolean quick,
                                         final File dataDir)
	    throws IOException
	    {
		System.out.println("Checking exhibit data... (cache file ignored)");
		final File baseDir = (dataDir != null) ? dataDir : new File(LocalProps.getDataDir());
		// Get a simple listing of exhibits on disc.
		final AllExhibitImmutableData aeid = _getAllExhibitImmutableData(
		    baseDir.getPath(),
		    -1,
		    !quick);
		assert(aeid != null);
		System.out.println("Exhibits found: " + aeid.length);

        // Collect existing accession data (as part of the loaded properties),
		// and recompute the hashes from scratch to compare,
		// and make sure that nothing hash changed.
        // We do this in parallel to maximise I/O throughput,
        // and use a ConcurrentHashMap as our working store as it is thread-safe.
        // We get these using our non-CPU-bound thread pool
        // and wait for them all to complete.
		final String sortedNames[] = aeid.getAllExhibitNamesSorted();
        final Map<String,AccessionData> accessionMap = new ConcurrentHashMap<String, AccessionData>(aeid.size() * 2);
        final List<Future<?>> tasks = new ArrayList<Future<?>>(sortedNames.length);
        for(final String name : sortedNames)
            {
            tasks.add(ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    // We don't store null/EMPTY values.
                    try
                        {
                        final ExhibitPropsLoadable epl = ExhibitPropsLoadable.getLoadableProperties(name, baseDir);
                        if((epl != null) && !epl.equals(ExhibitPropsLoadable.EMPTY) && (epl.getAccessionMetadata() != null))
                            {
                        	final AccessionData accessionMetadata = epl.getAccessionMetadata();
							accessionMap.put(name, accessionMetadata);

							// Compare the accession timestamp and length with the current file.
							final ExhibitStaticAttr esa = aeid.getStaticAttr(name);
							if(accessionMetadata.date == null)
							    { System.err.println("Accession timestamp missing for: " + name); }
							else if(accessionMetadata.date.longValue() != esa.timestamp)
							    { System.err.println("Accession timestamp does not match current file for: " + name + ": was/now " + new Date(accessionMetadata.date.longValue()) + '/' + new Date(esa.timestamp)); }
							if(accessionMetadata.size == null)
							    { System.err.println("Accession length missing for: " + name); }
							else if(accessionMetadata.size.longValue() != esa.length)
							    { System.err.println("Accession length does not match current file for: " + name + ": was/now " + accessionMetadata.size + '/' + esa.length); }

							// Now recompute and compare the accession data hashes.
							final InputStream is = new BufferedInputStream(new FileInputStream(new File(baseDir, name)));
							try
								{
								final Tuple.Pair<Integer,ROByteArray> hashes = AccessionData.computeFullFileHashes(is);
								if(!hashes.first.equals(accessionMetadata.hashCRC32))
							        { System.err.println("Accession hash CRC32 changed for: " + name); }
								if(!hashes.second.equals(accessionMetadata.hashMD5))
						            { System.err.println("Accession hash MD5 changed for: " + name); }
								}
							finally { is.close(); }
							}
                        else
                            { System.err.println("No accession data available for: " + name); }
                        }
                    catch(final Exception e) 
                        { throw new RuntimeException("Error loading accession data for: " + name, e); }
                    }
                }));
            }
        // Wait for all tasks to finish...
        for(final Future<?> task : tasks)
            {
            try { task.get(); }
            catch(final InterruptedException e)
                {
                final InterruptedIOException err = new InterruptedIOException(e.getMessage());
                err.initCause(e);
                throw err;
                }
            catch(final ExecutionException e)
                {
                final IOException err = new IOException(e.getMessage());
                err.initCause(e);
                throw err;
                }
            }
        System.out.println("Accession data loaded for exhibits: " + accessionMap.size());
        if(accessionMap.size() != aeid.length)
		    { throw new IOException("Missing (or changed) accession data."); }
	    }

    /**Shut down the data pipeline.
     * Saves any pending state (eg log entries and variables)
     * to disc if possible.
     * <p>
     * Has no upstream components to shut down
     * nor significant resources (memory) to release.
     */
    /* @Override */ public void destroy()
        {
        _saveSystemVariables(varMgr, true); // Force immediate save.
        }
    }