/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.datasource;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;

import ORG.hd.d.IsDebug;

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import java.util.concurrent.*;

import javax.crypto.SecretKey;


/**Exhibit pipeline stage that fetches its data across an HTTP[S] tunnel.
 * Derived from the abstract tunnel source base class.
 * <p>
 * This class does two important things:
 * <ul>
 * <li>Sets up an HTTP connection with the correct characteristics
 *     (such as with sensible timeouts on open and read).
 * <li>Controls the concurrency of the connection
 *     to avoid overwhelming the upstream server.
 * </ul>
 * <p>
 * A limited amount of concurrency (ie more than one connection)
 * may help overcome latency and improve throughput,
 * especially over a lossy network or during long operations
 * such as fetching the AEP or exhibit data blocks.
 * <p>
 * Concurrency should not be so large as to cause the master
 * to start rejecting connections,
 * eg no more than (say) the 8 connections that a normal
 * Web browser might open to overcome latency.
 * <p>
 * Note also that the master may itself have limited concurrency
 * available internally to service many types of operation.
 * <p>
 * Our concurrency limit may be by any type of RPC call,
 * or at most one of any particular operation type
 * (on the grounds that the master may only be able to service
 * limited numbers of calls by any particular operation type).
 * (At one point we limited concurrency using a simple Java monitor
 * that was selected by a hash on the RPC type.
 * Now we have one overall limit on the number of connections upstream
 * and abort rather than waiting indefinitely to get a connection.)
 * <p>
 * We may vary this value between instances to help prevent collisions
 * between different slaves, etc.
 * <p>
 * We may limit concurrency to 1 after any IOException
 * until one or more successful RPCs have completed.
 */
public final class ExhibitDataHTTPTunnelSource extends ExhibitDataTunnelSource
    {
    /**The full master-Web-server endpoint URL; never null. */
    private final URL serverURL;

    /**Get the endpoint URL of this tunnel; never null.
     * We assume that it is safe to return the URL directly,
     * ie that it behaves as if immutable.
     */
    public URL getServerURL()
        { return(serverURL); }

    /**User ID authentication info, never serialised; null if not set. */
    private transient volatile String userID;

    /**User ID password, never serialised; null if not set. */
    private transient volatile String passwd;

    /**Name of HTTP header for (optional) authentication user ID. */
    public static final String HEADER_AUTH_ID = "X-Tunnel-Auth-ID";

    /**Name of HTTP header for (optional) authentication password. */
    public static final String HEADER_AUTH_PW = "X-Tunnel-Auth-PW";

    /**Name of HTTP header for (optional) MAC and timestamp. */
    public static final String HEADER_MAC = "X-Tunnel-MAC";

    /**Set (or clear) additional authentication data in form of ID and password.
     * When the server requires explicit authentication from the client,
     * this form may be sufficient to allow full or partial (eg read-only)
     * access.
     * <p>
     * To set authentication arguments must be non-null and non-empty,
     * else all arguments must be null to clear authentication data.
     */
    public void setAuthenticationInfo(final String userID, final char[] passwd)
        {
        // Any null/empty argument implies unsetting the authentication data.
        if((userID == null) || (userID.length() == 0) ||
           (passwd == null) || (passwd.length == 0))
            {
            this.userID = null;
            this.passwd = null;
            return;
            }

        // The authentication data provided is syntactically OK.
        // Store a private copy of the authentication data.
        this.userID = userID;
        this.passwd = new String(passwd);
        }

    /**Maximum number of concurrent connections to allow back to the master; strictly positive.
     * We make this value public in case other modules
     * would like to base their load regulation on this connection limit.
     * <p>
     * Any value between 1 and 8 inclusive is probably reasonable.
     */
    public static final int MAX_CONCURRENT_CONNECTIONS = 8;

    /**If true, use "fair" lock access to try to reduce variance in connection service time.
     * May be slightly slower (and is fatally buggy under JDK 1.5.0_05 or earlier).
     */
    private static final boolean USE_FAIR_LOCKS = System.getProperty("java.version").compareTo("1.5.0_06") >= 0;

    static
        {
if(IsDebug.isDebug) { System.out.println("[ExhibitDataHTTPTunnelSource.USE_FAIR_LOCKS="+USE_FAIR_LOCKS+"; java.version="+System.getProperty("java.version")+".]"); }
        }

    /**Counting semaphore to limit RPC concurrency by total number of callers; never null */
    private final Semaphore countingSemaphore =
        new Semaphore(MAX_CONCURRENT_CONNECTIONS, USE_FAIR_LOCKS);

    /**Default HTTP connection timeout (ms); strictly positive.
     * This is set relatively low so that we don't block downstream/callers
     * too long if the Net is lossy or slow.
     * <p>
     * We assume that the default HTTP connection timeout would be longer than
     * the value chosen here.
     * <p>
     * We take failure to establish a connection quickly as indicative of
     * generally poor connectivity.
     * <p>
     * (We may need to allow extra time for HTTPS or tunnelled connections
     * to be set up.)
     * <p>
     * We may vary this value a little between instances to help prevent collisions
     * between different clients at one master.
     * <p>
     * A value of a few seconds to a few tens of seconds is probably suitable,
     * especially given the fact that connections that don't happen quickly
     * rarely happen at all.
     * <p>
     * (Note that in Windows NT the hard-wired initial TCP timeout was 3s,
     * which seems to have been a little too short for legitmate long-delay paths.)
     */
    private final int DEFAULT_HTTP_CONN_TIMEOUT_MS = 4000 +
            Rnd.fastRnd.nextInt(500);

    /**Default maximum HTTP read timeout (ms); strictly positive.
     * This is set high enough not to wrongly time-out long-running operations,
     * such as fetching the AEP, but short enough not to block clients indefinitely
     * if something goes seriously wrong,
     * eg a filesystem on the master hangs or dies.
     * <p>
     * This must probably be in the region of a few minutes to a few hours.
     */
    private static final int DEFAULT_HTTP_MAX_READ_TIMEOUT_MS = 20 * 60 * 1000 +
            Rnd.fastRnd.nextInt(5 * 60 * 1000);

    /**If true then have a watchdog thread ensure that no connection hangs indefinitely.
     * This is only needed as a fallback because the read and connect timeouts
     * on the HTTP URLConnection are not totally reliable as of JDK 1.5.0.
     */
    private static final boolean USE_WATCHDOG = true;

    /**Thread-safe Map of connections to times they should have terminated.
     * This may get checked in poll(),
     * or where another inbound thread might otherwise be blocked.
     */
    private final Map<HttpURLConnection, Tuple.Pair<Thread,Long>> openConnections =
        new Hashtable<HttpURLConnection, Tuple.Pair<Thread,Long>>();

    /**Closes any connections that have been alive too long.
     * Thread-safe.
     */
    private void killDeadConnections()
        {
        // Copy into another map (atomically).
        final Map<HttpURLConnection, Tuple.Pair<Thread,Long>> conns;
        synchronized(openConnections)
            { conns = new HashMap<HttpURLConnection, Tuple.Pair<Thread,Long>>(openConnections); }

        // Kill any connection that should have died by now.
        final long now = System.currentTimeMillis();
        for(final HttpURLConnection conn : conns.keySet())
            {
            final Tuple.Pair<Thread,Long> pair = conns.get(conn);
            final Long expires = pair.second;
            if(expires.longValue() < now)
                {
                logger.log("ERROR: ExhibitDataHTTPTunnelSource: attempting to kill hung connection to "+serverURL+": "+conn);
                try
                    {
                    // Try to wake/interrupt the jammed Thread.
                    pair.first.interrupt();

                    // Attempt to force the connection closed
                    // and try to disconnect it as it may be jammed.
                    conn.disconnect();
                    conn.getInputStream().close();
                    conn.getOutputStream().close();
                    }
                catch(final Throwable t)
                    {
                    t.printStackTrace();
                    }
                }
            }
        }

    /**Make an RPC call over HTTP[S] with the given outgoing packet.
     * This controls concurrency (number of connections to the master).
     * <p>
     * If we cannot get a connection, because:
     * <ul>
     * <li>we have reached the self-imposed concurrency limit, or
     * <li>there is a problem at the HTTP level (or in the upstream server),
     * </ul>
     * then we will try to time out and throw an IOException
     * rather than blocking indefinitely.
     *
     * @param  packetOut  request packet; never null
     * @return response packet; never null
     * @throws IOException  in case of I/O difficulties
     */
    @Override
    protected RawPacket doRPCRaw(final RawPacket packetOut)
        throws IOException
        {
        // Return the result as a complete packet,
        // having checked the trailer, etc.
        // We have a normal (short) read timeout.
        return((RawPacket) _doRPCRawInternal(packetOut, false, false));
        }

    /**If true then we enforce MACs when we have a local xfer key.
     * We may still try to send a request without a MAC (iff we don't have a key),
     * but we won't accept a response without a MAC unless we have no key.
     * <p>
     * All but a safe subset of ops is rejected with an OP__RUNTEX (RuntimeException)
     * on incoming requests without a verifiable MAC.
     */
    public static final boolean ENFORCE_MAC = true;

    /**Record of unique response packet IDs around our acceptance window; never null.
     * We use the final MAC segment ID, as that covers the entire frame.
     * <p>
     * We actually remember IDs for about twice the age implied by the skew
     * so that if our clock is wrong by, or slips by, the maximum skew,
     * we won't start admitting very old duplicate messages.
     * <p>
     * We should only ad IDs if possible when we are already fairly sure of the source
     * and when we have, for example, already checked for acceptable skew,
     * to make any sort of DoS attack against us harder.
     * <p>
     * We should reject otherwise-acceptable messages that have an ID
     * already present in this Map.
     * <p>
     * Thread-safe.
     */
    private final DuplicateIDChecker<ROByteArray> messageIDs =
        new DuplicateIDChecker<ROByteArray>(2*CoreConsts.MAX_PEER_CLOCK_SKEW_MS);

    /**If true allow excess connection attempts to queue for a while.
     * Setting this false reduces threads blocking.
     */
    private static final boolean ALLOW_CONNECTION_QUEUEING = false;

    /**Make an RPC call over HTTP[S] with the given outgoing packet.
     * This controls concurrency (number of connections to the master).
     * <p>
     * If we cannot get a connection, because:
     * <ul>
     * <li>we have reached the self-imposed concurrency limit, or
     * <li>there is a problem at the HTTP level (or in the upstream server),
     * </ul>
     * then we will eventually time out and throw an IOException
     * rather than blocking indefinitely.
     * <p>
     * If we have a data-protecting "XferKey" then we always send a MAC.
     * This enables the receiver to detect more errors in transmission.
     *
     * @param packetOut  request packet; never null
     * @param asStream  if true, return as streamed result rather than packet
     * @param allowBigReadTimeout if true then the RPC may take a lot of work at the master
     *     so we should be prepared to wait a long time before timing out
     *
     * @return response packet; never null
     * @throws IOException  in case of I/O difficulties
     */
    private Object _doRPCRawInternal(final RawPacket packetOut,
                                     final boolean asStream,
                                     final boolean allowBigReadTimeout)
        throws IOException
        {
        assert(packetOut != null);

        // Kill any dead connections that we find
        // if we seem to be short of resources.
        if(USE_WATCHDOG && (countingSemaphore.availablePermits() <= 0))
            { killDeadConnections(); }

        try
            {
            if(ALLOW_CONNECTION_QUEUEING)
                {
                // Only wait to obtain a permit for something like
                // the time that we allow for an HTTP connection.
                // This avoids unexpectedly blocking users for a long time,
                // though if there really is contention then it should probably be
                // handled properly at a higher level.
                if(!countingSemaphore.tryAcquire(DEFAULT_HTTP_CONN_TIMEOUT_MS/2,
                                                     TimeUnit.MILLISECONDS))
                    {
                    logger.log("doRPCRaw(): timeout while waiting to acquire HTTP tunnel permit to "+serverURL);
                    throw new IOException("timeout while waiting to acquire HTTP tunnel permit to "+serverURL+" ("+MAX_CONCURRENT_CONNECTIONS+" concurrent connections max) for opcode#"+packetOut.opCode+"; queue length = " + countingSemaphore.getQueueLength());
                    }
                }
            else
                {
                // Baulk/quit if a connection is not immediately available.
                if(!countingSemaphore.tryAcquire(0, TimeUnit.SECONDS))
                    {
                    logger.log("doRPCRaw(): HTTP tunnel busy to "+serverURL);
                    throw new IOException("HTTP tunnel busy to "+serverURL+" ("+MAX_CONCURRENT_CONNECTIONS+" concurrent connections max) for opcode#"+packetOut.opCode+"; queue length = " + countingSemaphore.getQueueLength());
                    }
                }
            }
        catch(final InterruptedException e)
            { throw new IOException("thread interrupted waiting to start HTTP tunnel connection"); }
        try
            {
if(_protocolDebug) { System.err.println("[doRPCRaw(): connecting upstream: " + serverURL + ".]"); }

            // Construct a HTTP connection to the master server.
            // Hopefully this will be transparently pooled by the JVM.
            final HttpURLConnection connection =
                (HttpURLConnection) serverURL.openConnection();

            // Set an explicit fairly short connection timeout
            // so as to be intolerant of ropey Net connectivity,
            // ie don't attempt to use a connection with very poor behaviour.
            // We don't set a very long read timeout as some calls
            // may take a lot of remote work before they can start to reply,
            // but we'll manage to escape a call that blocks completely.
            if(!CoreConsts.AVOID_UNSAFE_TCP_TIMEOUTS)
                {
                connection.setConnectTimeout(DEFAULT_HTTP_CONN_TIMEOUT_MS);
                // Unless we are doing a very slow and/or CPU-expensive RPC,
                // use something more like the connection timeout if appropriate.
                connection.setReadTimeout(allowBigReadTimeout ? DEFAULT_HTTP_MAX_READ_TIMEOUT_MS :
                    Math.min(DEFAULT_HTTP_MAX_READ_TIMEOUT_MS, 2*DEFAULT_HTTP_CONN_TIMEOUT_MS));
                }

            // Don't get any data from cache: must be a live connection!
            connection.setUseCaches(false);
            // Exact given URL must be valid...
            connection.setInstanceFollowRedirects(false);
            // Allow bi-directional communication.
            connection.setDoOutput(true);
            connection.setDoInput(true);
            // Don't allow user interaction.
            connection.setAllowUserInteraction(false);

            // We know what the content length will be
            // so set it here to allow some HTTP performance optimisations.
            final int contentLength = packetOut.getFrameLength();
            connection.setRequestProperty("Content-Length", String.valueOf(contentLength));
            connection.setFixedLengthStreamingMode(contentLength);

            // If authentication/check data is available,
            // then apply it to the request.
            final String uID = userID;
            final String pwd = passwd;
            if((uID != null) && (pwd != null))
                {
                connection.setRequestProperty(HEADER_AUTH_ID, uID);
                connection.setRequestProperty(HEADER_AUTH_PW, pwd);
                }
            // Apply a MAC to our outbound request for extra error-detection if we can.
            SecretKey xk = null;
            try
                {
                xk = LocalProps.getXferHMACKey();
                if(xk != null) // Need a key to create the MAC.
                    {
                    try
                        {
                        final PacketProtector pp = new PacketProtector(packetOut, xk);
                        final String checkString = pp.toCheckString();
if(IsDebug.isDebug) { logger.log("INFO: applying request MAC: " + checkString); }
                        connection.setRequestProperty(HEADER_MAC, checkString);
                        }
                    catch(final InvalidKeyException e)
                        {
                        // In case of error, don't apply the MAC header, but do complain loudly!
                        e.printStackTrace();
                        }
                    }
                }
            catch(final Exception e1)
                {
                // Absorb/log any other exception that we encounter,
                // but attempt to continue.
                e1.printStackTrace();
                }

            // Do simple HTTP security/sanity check...
            assert(!connection.usingProxy());

if(_protocolDebug) { System.err.println(" [doRPCRaw(): about to writePacket().]"); }

            if(USE_WATCHDOG)
                {
                final long stopBy = System.currentTimeMillis() +
                                    DEFAULT_HTTP_CONN_TIMEOUT_MS +
                                    DEFAULT_HTTP_MAX_READ_TIMEOUT_MS;
                openConnections.put(connection, new Tuple.Pair<Thread,Long>(Thread.currentThread(), new Long(stopBy)));
                }

            try
                {
                // Send the output (request) packet, compressed if appropriate...
                final OutputStream os = connection.getOutputStream();
                try { packetOut.writePacket(os); }
                finally { os.close(); }

                if(_protocolDebug) { System.err.println(" [doRPCRaw(): about to open input/response stream.]"); }

                final int responseCode = connection.getResponseCode();
                if(responseCode != 200)
                    { throw new IOException("ERROR: RPC failed: HTTP POST response code "+responseCode); }

                // Look for a MAC header on the response.
                // Veto obviously-broken communications, even if we don't have a key.
                PacketProtector pp = null;
                final String hdrMAC = connection.getHeaderField(HEADER_MAC);
                if(null != hdrMAC)
                    {
                    try { pp = PacketProtector.fromCheckString(hdrMAC); }
                    catch(final IllegalArgumentException e) { logger.log("garbled MAC on response"); }
if(IsDebug.isDebug) { logger.log("INFO: received response MAC: " + pp); }

                    // Veto reponses with the wrong length.
                    // (For now just warn about a missing Content-Length.)
                    final int contentLengthRcvd = connection.getContentLength();
                    if(contentLengthRcvd == -1)
                        { logger.log("WARNING: missing Content-Length on response in tunnel"); }
                    else if(pp.length != contentLengthRcvd)
                        { throw new IOException("ERROR: wrong-length (or missing Content-Length) response received"); }
                    // Check that the remote peer's clock is not vastly wrong
                    // (or that a very old packet is reappearing out of nowhere).
                    // We assume that the skew allowance is generous enough to cover
                    // actual transit time of messages in the network.
                    final long now = System.currentTimeMillis();
                    if((pp.timestamp < now - CoreConsts.MAX_PEER_CLOCK_SKEW_MS) ||
                       (pp.timestamp > now + CoreConsts.MAX_PEER_CLOCK_SKEW_MS))
                       { throw new IOException("ERROR: discarded response with bad timestamp: check local and peer clocks"); }
                    // Veto duplicate packets/requests in this tunnel.
                    if(messageIDs.add(pp.mac.get(pp.mac.size()-1)) != null)
                        { throw new IOException("ERROR: duplicate response received"); }
                    }

                if(ENFORCE_MAC && ((xk != null) && (pp == null)))
                    {
                    logger.log("ERROR: did not receive MAC on RPC "+packetOut.opCode+" response from "+serverURL+" so discarding because unsafe: MAC hdr="+hdrMAC);
                    throw new IOException("RPC "+packetOut.opCode+" response from "+serverURL+" not protected by MAC, thus not safe to use");
                    }

                // Get the response input stream.
                // If we have the XferKey and a MAC then always verify the input stream.
                // This protects all of the input data, including the RawPacket headers.
                // Using the stream mechanism helps spread the verification load,
                // as well as transparently permitting streamed handling of responses.
                final InputStream isRaw = connection.getInputStream();
                final InputStream is = ((pp == null) || (xk == null)) ? isRaw :
                    pp.protectInputStream(xk, isRaw);

                // Return the input stream for a streamed response.
                if(asStream)
                    { return(is); }

                try
                    {
                    if(_protocolDebug) { System.err.println(" [doRPCRaw(): about to readPacket().]"); }

                    final RawPacket result = RawPacket.readPacket(is);

                    if(_protocolDebug) { System.err.println(" [doRPCRaw(): about to return result.]"); }

                    // Return result successfully.
                    return(result);
                    }
                finally
                    {
                    // Explicitly close the input stream now that we've read the packet.
                    is.close();

                    // Try to discard the connection if that is our policy.
                    if(!KEEP_SERVER_CONNECTION_ALIVE) { connection.disconnect(); }
                    }
                }
            finally
                {
                // Record that the connection has been closed
                // (or is not being handled by the caller now).
                if(USE_WATCHDOG) { openConnections.remove(connection); }
                }
            }
        finally { countingSemaphore.release(); }
        }

    // Inherit javadoc comment...
    @Override
    protected Tuple.Pair<Integer, InputStream> doRPCRawWithStreamResponse(final RawPacket packetOut,
                                                                          final boolean allowBigReadTimeout)
        throws IOException
        {
        // Get the raw input/response stream...
        final InputStream is = (InputStream) _doRPCRawInternal(packetOut, true, allowBigReadTimeout);

        // Start reading/decoding the raw packet header.
        // Read the opcode (byte) header field.
        final int ch = is.read();
        if(ch < 0) { throw new EOFException(); }
        final byte opC = (byte) ch;

        // Check for mismatched response packets
        // or remote exceptions codes with special response op-codes.
        if(opC != packetOut.opCode)
            {
            // Deal specially with some fixed remote exception types.
            switch(opC)
                {
                case RawPacket.OP__PGMNISEX:
                    { throw new PGMasterNotInServiceException(); }
                case RawPacket.OP__RUNTEX:
                    { throw new RuntimeException("remote threw RuntimeException"); }
                case RawPacket.OP__REMEX:
                    { throw new IOException("remote threw RemoteException"); }
                case RawPacket.OP__INTEX:
                    { throw new InterruptedIOException("remote operation interrupted"); }
                }
            throw new IOException("ExhibitDataHTTPTunnelSource.doRPCRawWithStreamResponse(): packet response-type mismatch; got "+opC+" expected "+packetOut.opCode);
            }

        // Read the (int) length header field.
        final int ch1 = is.read();
        final int ch2 = is.read();
        final int ch3 = is.read();
        final int ch4 = is.read();
        if((ch1 | ch2 | ch3 | ch4) < 0) { throw new EOFException(); }
        final int rawLen = ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));

        // In the simplest (and fairly common) case of an empty response,
        // for efficiency and robustness we close the original response stream immediately
        // and return an empty immutable response to the caller.
        if(rawLen == 0)
            {
            // Double-check that we find our trailer (and nothing else)
            // after the empty payload.
            int ic;
            if(((byte)(ic = is.read())) != RawPacket.TRAILER) { throw new IOException("missing frame trailer byte, got "+ic+": op "+opC); }
            if((ic = is.read()) != -1) { throw new IOException("garbage ("+ic+") present after frame trailer rather than EOF"); }

            is.close(); // Immediate HTTP stream close to release resources ASAP.
            return(EMPTY_RESPONSE_STREAM);
            }

        // Check that the reported packet length is within bounds.
        if((rawLen > RawPacket.MAX_PAYLOAD_SIZE) ||
           (rawLen < -RawPacket.MAX_PAYLOAD_SIZE))
            {
            is.close();
            throw new IOException("Bad (huge) packet length");
            }

        final boolean isCompressed = (rawLen < 0);

        // In the simple (uncompressed-data) case we return the stream as-is.
        // This leaves the trailer byte on the stream visible to the caller,
        // but the true length of the data payload is returned too.
        if(!isCompressed)
            { return(new Tuple.Pair<Integer, InputStream>(Integer.valueOf(rawLen), is)); }

        // We wrap the underlying input in our custom deflater stream.
        // We do not know the final result length so return a null length value.
        // This leaves the trailer byte on the stream visible to the deflater,
        // but the deflater will not use its value though may read it.
        // We wrap the stream in a non-aggressive data pump
        // to improve throughput for large results.
        final InputStream cis = GenUtils.dataPump(new DefInputStream(is, true), 1<<16, false);
        return(new Tuple.Pair<Integer, InputStream>(null, cis));
        }
    
    /**Empty, immutable/unclosable InputStream that we can return to all callers. */
    private static final InputStream EMPTY_INPUTSTREAM = new ByteArrayInputStream(new byte[0]);

    /**Empty, immutable/unclosable response stream that we can return to all callers. */
    private static final Pair<Integer, InputStream> EMPTY_RESPONSE_STREAM = new Tuple.Pair<Integer, InputStream>(Integer.valueOf(0), EMPTY_INPUTSTREAM);


    /**Construct a new HTTP tunnel data source to the given URL.
     * Does not necessarily try to build the tunnel immediately.
     *
     * @throws IllegalArgumentException  if the URL is invalid or unusable.
     */
    public ExhibitDataHTTPTunnelSource(final String masterURL,
                                       final SimpleLoggerIF logger)
        {
        super(logger);

        if(masterURL == null)
            { throw new IllegalArgumentException("null masterURL not allowed"); }
        try { serverURL = new URL(masterURL); }
        catch(final MalformedURLException e)
            { throw new IllegalArgumentException("invalid masterURL \""+masterURL+"\": " + e.getMessage()); }

        if(_protocolDebug)
            {
            System.out.println("ExhibitDataHTTPTunnelSource("+masterURL+") instance created");
            }
        }

    /**Poll periodically.
     * We call super.poll(gp) to ensure that base-class polling work gets done,
     * including any upstream poll() calls.
     */
    @Override
	public void poll(final GenProps gp)
        {
        try
            {
            // Kill any dead connections that we find.
            if(USE_WATCHDOG) { killDeadConnections(); }

            final int permits = countingSemaphore.availablePermits();
            if(permits <= 0)
                {
                logger.log("[ExhibitDataHTTPTunnelSource: WARNING: busy HTTP tunnel to "+serverURL+": queue="+(countingSemaphore.getQueueLength())+".]");
                }
            }
        finally { super.poll(gp); }
        }
    }
