package org.hd.d.pg2k.svrCore.datasource;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.stats.StatsLogger;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.*;
import java.lang.ref.SoftReference;
import java.nio.ByteBuffer;
import java.security.*;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Pattern;

import javax.crypto.Mac;
import javax.crypto.SecretKey;

/**Exhibit pipeline stage that fetches its data across a master/slave tunnel.
 * Some of the visible types and values are used by the implementations
 * of both ends of the tunnel (eg the tunnel servlet)
 * to help implement the shared protocol.
 * <p>
 * This takes data from the server's own DataSourceBean by default.
 */
public abstract class ExhibitDataTunnelSource implements SimpleExhibitPipelineIF
    {
    /**Create an instance, passing in a (non-null) logger to use. */
    public ExhibitDataTunnelSource(final SimpleLoggerIF logger)
        {
        if(logger == null) { throw new IllegalArgumentException(); }
        this.logger = logger;
        statsIDTS =
            new StatsLogger.StatsConfig("TUNNELSOURCE",
                                        logger, // Use servlet log if poss.
                                        false, // Only dump summaries...
                                        12 * 3600, // About every 12 hours.
                                        true); // Adaptive.
        }

    /**If true, try GC and running finalisers after some connection failures.
     * If we think that connection failure may have been caused by
     * failure to release resources (elsewhere in the VM!)
     * then try to force a clean-up from time to time.
     * But, in case we are wrong, do not do this too often anyway.
     */
    private static final boolean FORCE_GC_AND_FINALISERS_AFTER_POLL_FAIL = true;

    /**Minimum wait after failed tunnel call before retry (ms); strictly positive.
     * A few tens of milliseconds to a second is probably best,
     * to recover from transient I/O problems quickly,
     * but without actually busy-waiting...
     */
    public static final int FAIL_RETRY_WAIT_MIN_MS = 101;

    /**Maximum wait after failed tunnel call before retry (ms).
     * Several minutes is probably good,
     * to cope with a crashed master server (and/or to wait for it to reboot),
     * or a failed connection,
     * but too long and our distributed variables will fail needlessly.
     * <p>
     * Most requests during the wait will be vetoed cleanly and quickly
     * allowing timely service to the slave's clients.
     * <p>
     * We vary this limit slightly between slaves/instances to help avoid them
     * colliding with one another especially if hitting this upper bound.
     */
    public static final int FAIL_RETRY_WAIT_MAX_MS = 2 * 60 * 1000 +
        (1 | Rnd.fastRnd.nextInt(1 * 60 * 1000));

    /**If true, allow connection/resource sharing between calls if possible.
     * We may also periodically check that we can still contact the master.
     * <p>
     * An open connection would probably improve TCP efficiency
     * and reduce latency for an underlying HTTP[S] connection, for example.
     * (We might allow HTTP/1.1 streaming and connection sharing to be
     * done for us by the JDK runtime if this is true.)
     */
    protected static final boolean KEEP_SERVER_CONNECTION_ALIVE = true;

    /**If true, extra instrumentation for protocol debug; definitely not for release code. */
    protected static final boolean _protocolDebug = false && ORG.hd.d.IsDebug.isDebug;

    /**Our unique client-side end-point identifier.
     * The ID goes into all outgoing variable updates (and possibly other data)
     * going upstream to the master to uniquely identify this client.
     * <p>
     * This is unique to each tunnel client end-point instance
     * and is created afresh on each run,
     * so a rebooted client gets a new ID.
     * <p>
     * Uses secure/good generator to try to ensure global uniqueness.
     * (This may be expensive to compute;
     * we might be able to defer this until first needed.)
     */
    public final SimpleVariableValue uniqueClientID =
        new SimpleVariableValue(SystemVariables.LOCAL_SYS_ID, InstanceID.createInstanceID());


    /**Our logger which falls back to System.out if servlet log not available; never null. */
    protected final SimpleLoggerIF logger;

    /**The stats set to which we log general tunnel source stats.
     * The unique codes are the constants TSNAME_XXX.
     */
    private final StatsLogger.StatsConfig statsIDTS;

    /**General stats event name: RPC request. */
    public static final String TSNAME_RPCREQUEST = "RPC-request";

    /**General stats event name: RPC failure due to an IOException. */
    public static final String TSNAME_RPCIOEX = "RPC-IOException";

    /**General stats event name prefix: RPC request packet type. */
    public static final String TSNAMEPR_RPCTYPE = "RPC-type-";

    /**General stats event name: "short" raw data read, less than bulk-transfer block size.. */
    public static final String TSNAME_SHORTREAD = "shortRead";

    /**Convenience value; (immutable) zero-length byte array for use as an empty packet payload.
     * Used by local classes and TunnelServlet; package visible.
     */
    public static final byte[] EMPTY_PAYLOAD = new byte[0];

    /**Time when last successful connection was made, or 0 if no connection..
     * Declared volatile so that it can be accessed without a lock.
     * <p>
     * Set by _noteResult() after a good exchange with the master;
     * never cleared and never null.
     */
    private volatile long lastSuccessfulConnectionTime;

    /**Get time of last successful connection, or 0 if none. */
    public long getLastSuccessfulConnectionTime()
        { return(lastSuccessfulConnectionTime); }

    /**Time before which we should not again try to contact the master; null if master is fine.
     * Declared volatile so that it can be accessed without a lock.
     * <p>
     * Cleared to null by _noteResult() after a good exchange with the master;
     * the wait to another attempt is approximately doubled after a failed exchange.
     */
    private volatile Long doNotTryMasterUntil;

    /**Returns true if we currently cannot talk to the master, false otherwise. */
    public boolean isBroken()
        {
        final Long notUntil = doNotTryMasterUntil;
        return((notUntil != null) &&
               (notUntil.longValue() > System.currentTimeMillis()));
        }

    /**Successive failure count; reset to zero upon success.
     * Private to _noteResult() and accessed under its private lock.
     * <p>
     * Cleared to zero on a successful packet exchange;
     * incremented after each failure.
     */
    private int successiveFailureCount;

    /**Private lock for _noteResult. */
    private final Object _nR_lock = new Object();

    /**Immutable empty result indicating no event values available. */
    private static final EventVariableValue[] NO_EVENT_VALUES = new EventVariableValue[0];

    /**Routine to note success or failure of RPC call and adjust control variables.
     * Has its own private lock to ensure atomic update of control variables.
     *
     * @param success  call with this true immediately after successful RPC exchange with master;
     *     call with false otherwise
     */
    private void _noteResult(final boolean success)
        {
        synchronized(_nR_lock)
            {
            final long now = System.currentTimeMillis();
            if(success)
                {
                // Successful RPC exchange with master.
                lastSuccessfulConnectionTime = now;
                doNotTryMasterUntil = null;
                successiveFailureCount = 0;
                return;
                }

            // Failed RPC exchange with the master...
            if(++successiveFailureCount < 0)
                { successiveFailureCount = Integer.MAX_VALUE; }

            // Compute first-cut back-off time.
            // Double for each successive failure.
            final long approxBackoff = ((long) ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MIN_MS) <<
                Math.min(32, successiveFailureCount);

            // Coerce to reasonable bounds...
            final int waitBounded = (int) Math.min(ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MAX_MS,
                Math.max(ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MIN_MS, approxBackoff));

            // Roughly double the pause before the next connection attempt,
            // with a random component to help avoid inter-slave collisions.
            final long newWaitTime = Math.min(ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MAX_MS,
                waitBounded/2 + Rnd.fastRnd.nextInt(waitBounded));

            // Store the new "do not try before" time...
            doNotTryMasterUntil = new Long(now + newWaitTime);
            }
        }

    /**Does a NO-OP on the server.
     * If this returns without throwing an exception,
     * then the connection to the back-end (master) server is probably OK.
     *
     * @param unguarded  if true, this ignores any recent connection problems
     *     and immediately tries to contact the master,
     *     else it behaves like other operations and is quivkly vetoed
     *     while the master is failing
     *
     * @throws java.io.IOException  in case of trouble communicating with master.
     */
    public void doNOOP(final boolean unguarded)
        throws IOException
        {
        final RawPacket response;
        if(unguarded)
            {
            response = doRPCUnguarded(
                new RawPacket(RawPacket.OP_NOOP, ExhibitDataTunnelSource.EMPTY_PAYLOAD));
            }
        else
            {
            response = doRPC(
                new RawPacket(RawPacket.OP_NOOP, ExhibitDataTunnelSource.EMPTY_PAYLOAD));
            }
        if(response.plLength != 0)
            { throw new IOException("invalid non-empty response to NOOP request"); }
        return;
        }

    /**Get the static attributes for a given exhibit.
     * Returns null if the named exhibit does not exist.
     * <p>
     * Sends name as a UTF string and allows compression.
     * (We use UTF since we know that the length is limited and
     * the all-ASCII nature of valid names should yield
     * one-byte-per-char encoding before compression.)
     * <p>
     * This will fail with a null name argument.
     * <p>
     * (Big secret: if this class is used behind, for example,
     * ExhibitDataSimpleCache, this method will never be called
     * because all such calls are answered direct from the cache.)
     */
    public ExhibitStaticAttr getStaticAttr(final String name)
        throws IOException
        {
        final ByteArrayOutputStream baos =
            new ByteArrayOutputStream(name.length());
        final DataOutputStream dos = new DataOutputStream(baos);
        dos.writeUTF(name);
        final byte data[] = baos.toByteArray();
        dos.close(); // Attempt to free some resources quickly.

        final RawPacket request = new RawPacket(
            RawPacket.OP_getStaticAttr,
            data);

        final RawPacket response = doRPC(request);
        return((ExhibitStaticAttr) response.getSerializedObjectPayload());
        }

    /**Get a chunk of the raw exhibit binary.
     * The start position must be non-negative.
     * <p>
     * Sends name as a UTF string and allows compression of the request.
     * (We use UTF since we know that the length is limited and
     * the all-ASCII nature of valid names should yield
     * one-byte-per-char encoding before compression.)
     * <p>
     * We send the start and (computed) afterEnd arguments as int values.
     * <p>
     * We return the data as-is.
     *
     *
     * @throws java.lang.IllegalArgumentException  for blatantly invalid name,
     *     or non-positive length
     */
    public void getRawFile(final ByteBuffer buf,
                           final String exhibitName, final int position,
                           final boolean dontCache)
        throws IOException
        {
        // We check that that the name is not blatantly illegal...
        if(!ExhibitName.validNameSyntaxBasic(exhibitName))
            { throw new IllegalArgumentException("invalid name"); }

        // We don't allow zero-byte requests over the expensive tunnel.
        int len = buf.limit() - buf.position();
        if(len < 1)
            { throw new IllegalArgumentException("request length must be positive"); }

        // Limit any request to a size that the back-end would allow.
        if(len > MAX_USER_READ_SIZE)
            { len = MAX_USER_READ_SIZE; }

        if(ExhibitDataTunnelSource._protocolDebug)
            {
            System.err.println("ExhibitDataTunnelSource.getRawFile(" +
                "len="+len+", dontCache="+dontCache+")");
            }

        // Right, actually do the transfer!
        final ByteArrayOutputStream baos =
            new ByteArrayOutputStream(exhibitName.length());
        final DataOutputStream dos = new DataOutputStream(baos);
        final int afterEnd = position + len;
        dos.writeUTF(exhibitName);
        dos.writeInt(position);
        dos.writeInt(afterEnd);
        dos.writeBoolean(dontCache);
        final byte data[] = baos.toByteArray();
        dos.close(); // Attempt to free some resources quickly.

        final RawPacket request = new RawPacket(
            RawPacket.OP_getRawFile,
            data);

        final RawPacket response = doRPC(request);
        response.getPayloadCopy(buf); // Straight into buffer for efficiency.

//        final byte[] payloadCopy = response.getPayloadCopy();
//
//        // If this is a "short" transfer then log it to help with tuning.
//        if(payloadCopy.length < CoreConsts.BULK_DATA_TRANSFER_SIZE)
//            { StatsLogger.captureDataPoint(statsIDTS, ExhibitDataTunnelSource.TSNAME_SHORTREAD); }
//
//        buf.put(payloadCopy);
        }

    /**Gets all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     */
    public AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
        throws IOException
        {
        final RawPacket response = doRPC(
            new RawPacket(RawPacket.OP_getAllExhibitImmutableData, ExhibitDataTunnelSource.longSer(oldStamp)));
        return((AllExhibitImmutableData) response.getSerializedObjectPayload());
        }

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed
     * then a default set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date
     * (eg the oldHash matches that that would have been returned)
     * then null is returned.
     * <p>
     * Because the AEP drives the entire Gallery,
     * we always let the request go over the tunnel even if recent calls have failed.
     * For example, we don't want mirrors to remain out of sync for too long.
     * <p>
     * This also typically takes a lot of bandwidth and CPU/wallclock time,
     * so is optimised and stripped down to a minimum,
     * and relies on the AEP deserialisation for serious error checking.
     * <p>
     * It is hoped that using the streamed form will overlap I/O and CPU work
     * and thus get the new AEP into the client quicker and smoother.
     * <p>
     * We attempt to be good citizens and note most RPC successes/failures
     * to help maintain the tunnel's view of the master's status.
     * <p>
     * (Because satisfying these calls may take a long time,
     * the upstream server may veto concurrent queries that return anything
     * other than null.)
     * <p>
     * As a further attempt to minimise time and bandwidth,
     * we may try the "diff" version of this call,
     * falling back to the normal version if this fails.
     */
    public AllExhibitProperties getAllExhibitProperties(final long oldHash)
        throws IOException
        {
        final RawPacket packetOut = new RawPacket(RawPacket.OP_getAllExhibitProperties, ExhibitDataTunnelSource.longSer(oldHash));

        // The RPC may take a long time to compute the response to, so have a big read timeout.
        final Pair<Integer, InputStream> stream = doRPCRawWithStreamResponse(packetOut, true);
        if((stream.first != null) && (stream.first.intValue() == 0))
            {
            // An empty response means a null AEP response.
            stream.second.close();
            _noteResult(true); /* Note RPC success. */
            return(null);
            }

        // Actually deserialise from the input stream.
        final ObjectInputStream ois = new ObjectInputStream(stream.second);
        // For robustness, treat a class mismatch problem like a data problem.
        // We rely on the AllExhibitProperties deserialisation for most error checking.
        try
            {
            final AllExhibitProperties result = (AllExhibitProperties) ois.readObject();
            _noteResult(true); /* Note RPC success. */
            return(result);
            }
        catch(final ClassNotFoundException e) { throw new IOException(e.getMessage()); }
        catch(final InterruptedIOException e)
            {
            // Not a failure: can retry later...
            throw e; /* Rethrow error as-is. */
            }
        catch(final IOException e)
            {
            _noteResult(false); /* Note RPC failure. */
            throw e; /* Rethrow error as-is. */
            }
        finally { ois.close(); }
        }
    
    /**Maximum level of compression in AEP "diff" RPC supported in this implementation, client or server side. */
    private static final CompressionLevel MAX_AEP_DIFF_COMP_LEVEL = CompressionLevel.BZIP2; // GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL;

    /**Gets set of all exhibit properties if not that specified, attempting minimise data transferred across the tunnel.
     * This is a tunnel-specific optimisation to minimise bandwidth.
     * For some clients this will save money and time.
     * <p>
     * If the AEP specified is null then the remote AEP will be fetched and returned unconditionally.
     * <p>
     * If no exhibits are currently installed
     * then a default set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date
     * (eg the oldHash matches that that would have been returned)
     * then null is returned.
     * <p>
     * This is an attempted optimisation of the getAllExhibitProperties(long oldHash),
     * returning a diff or extra-highly-compressed AEP representation.
     * <p>
     * This may not be supported at all by the remote end of the connection,
     * or may fail for lack of resources, etc,
     * so we can automatically fall back to the usual call in case of difficulty.
     * <p>
     * (We do not record failures as connection/tunnel problems,
     * though we do note successful calls in favour of the connection status.)
     * 
     * @param oldAEP  current latest AEP held by the caller
     * @param allowAutoRecovery  if true then allow fallback to generic AEP fetch method
     */
    public AllExhibitProperties getAllExhibitProperties(final AllExhibitProperties oldAEP,
                                                        final boolean allowAutoRecovery)
        throws IOException
        {
        final long oldHash = (oldAEP == null) ? -1 : oldAEP.longHash;

        try
            {
            // Construct the header.
            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            final DataOutputStream dos = new DataOutputStream(baos);
            // Write existing AEP longHash, or -1 if no extant AEP.
            dos.writeLong(oldHash);
            // Write current maximum supported AEP compression level as a UTF String.
            dos.writeUTF(MAX_AEP_DIFF_COMP_LEVEL.name());
            dos.close();
            final byte data[] = baos.toByteArray();
            final RawPacket packetOut = new RawPacket(RawPacket.OP_getAllExhibitPropertiesDiff, data);

            // An empty response (zero bytes) means that our AEP is up-to-date...
            // The RPC may take a long time to compute the response to, so have a big read timeout.
            final Pair<Integer, InputStream> stream = doRPCRawWithStreamResponse(packetOut, true);
            if((stream.first != null) && (stream.first.intValue() == 0))
                {
                // An empty response means a null AEP response.
                stream.second.close();
                _noteResult(true); /* Note RPC success. */
                return(null);
                }

            // Else read from the stream.
            //  1) A UTF string for the compression level/type (abort if unsupported).
            //  2) A compressed object stream (in the appropriate format) which may be:
            //  2a) AllExhibitProperties, in which case return as-is.
            //  2b) AllExhibitPropertiesDelta, in which case apply to the input AEP and return.
            try
                {
                final DataInputStream dis = new DataInputStream(stream.second);
                final CompressionLevel cl = CompressionLevel.valueOf(dis.readUTF());
                if(cl.getLevel() > MAX_AEP_DIFF_COMP_LEVEL.getLevel())
                    { throw new IOException("server returned data compressed at too high a level"); }
                final InputStream is = GenUtils.wrapForDecompression(stream.second, cl);
                final ObjectInputStream ois = new ObjectInputStream(is);
                // Read in the object from the stream on the fly.
                // TODO: we may want to add a data pump to keep data flowing during processing.
                final Object o = ois.readObject();

                // If the result is a full (and hopefully super-compressed) AEP
                // then unpack it and return it here.
                if(o instanceof AllExhibitProperties)
                    {
                    _noteResult(true); /* Note RPC success. */
/* if(IsDebug.isDebug) */ { logger.log("[ExhibitDataTunnelSource: successful getAEP(AEP) full: "+cl+".]"); }
                    return((AllExhibitProperties) o);
                    }

                // If the result is a diff then apply it to the input AEP.
                if(o instanceof AllExhibitPropertiesDelta)
                    {
                    final AllExhibitPropertiesDelta delta = (AllExhibitPropertiesDelta) o;
                    final AllExhibitProperties result = AllExhibitPropertiesDelta.applyDiff(oldAEP, delta);
/* if(IsDebug.isDebug) */ { logger.log("[ExhibitDataTunnelSource: successful getAEP(AEP) diff: #before/after="+delta.lengthAEIDBefore+"/"+delta.lengthAEIDAfter+", aep before/after="+oldAEP+"/"+result+", "+cl+".]"); }

//                    /* TEMPORARY SAFETY MEASURE: REJECT RESULT WITH FEWER EPL/EPC ENTRIES... */
//                    if(result.getExhibitPropsLoadableMap().size() < oldAEP.getExhibitPropsLoadableMap().size())
//                        { throw new IOException("WARNING: rejecting AEP diff result with fewer EPLs..."); }
//                    if(result.getExhibitPropsComputableMap().size() < oldAEP.getExhibitPropsComputableMap().size())
//                        { throw new IOException("WARNING: rejecting AEP diff result with fewer EPCs..."); }

                    _noteResult(true); /* Note RPC success. */
                    return(result);
                    }

                throw new IOException("unexpected return type");
                }
            finally { stream.second.close(); /* Free (network) resources. */ }            
            }
        catch(final InterruptedIOException e)
            { throw e; } // Rethrow as-is: don't attempt recovery yet...
        catch(final Exception e)
            {
            // Always log the problem/exception.
            e.printStackTrace();

            if(!allowAutoRecovery)
                {
                if(e instanceof IOException) { throw (IOException) e; }
                final IOException err = new IOException("aborting after failed AEP diff RPC");
                err.initCause(e);
                throw err;
                }

            logger.log("ERROR: ExhibitDataTunnelSource: automatically recovering from AEP diff error: "+e.getMessage()+".");

            // In case of error, automatically fall back to generic hash-arg call.
            return(getAllExhibitProperties(oldHash));
            }
        }

    /**Helper method to serialise a single free-standing long value.
     * The long value is serialised just as DataOutputStream would,
     * as 8 bytes with the high-byte first.
     */
    public static byte[] longSer(final long value)
        {
        final byte[] result = new byte[8];
        result[0] = (byte) (value >> 56);
        result[1] = (byte) (value >> 48);
        result[2] = (byte) (value >> 40);
        result[3] = (byte) (value >> 32);
        result[4] = (byte) (value >> 24);
        result[5] = (byte) (value >> 16);
        result[6] = (byte) (value >>  8);
        result[7] = (byte) (value      );
        return(result);
        }

    /**Helper method to serialise a single free-standing int value.
     * The int value is serialised just as DataOutputStream would,
     * as 4 bytes with the high-byte first.
     */
    public static byte[] intSer(final int value)
        {
        final byte[] result = new byte[4];
        result[0] = (byte) (value >> 24);
        result[1] = (byte) (value >> 16);
        result[2] = (byte) (value >>  8);
        result[3] = (byte) (value      );
        return(result);
        }

    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     * <p>
     * On the wire this is (outbound) a 8-byte timestamp and
     * by return should be an empty payload (corresponding to null)
     * or a serialised GenProps object (uncompressed for now).
     * <p>
     * Because GenProps are so important to the functioning of the entire Gallery,
     * we always let the request go over the tunnel even if recent calls have failed.
     */
    public org.hd.d.pg2k.svrCore.props.GenProps getGenProps(final long oldStamp)
        throws IOException
        {
//System.err.println("[Fetching GenProps over HTTP tunnel...]");

        final RawPacket response = doRPCUnguarded(
            new RawPacket(RawPacket.OP_getGenProps, longSer(oldStamp)));

        final GenProps result = (GenProps) response.getSerializedObjectPayload();
//System.err.println(" [Fetched GenProps over HTTP tunnel.]");
        return(result);
        }

    /**Gets the security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     */
    public java.util.Properties getGenSecProps(final long oldStamp)
        throws IOException
        {
//System.err.println("[Fetching GenSecProps over HTTP tunnel...]");

        final RawPacket response = doRPC(
            new RawPacket(RawPacket.OP_getGenSecProps, longSer(oldStamp)));

        final Properties result = (Properties) response.getSerializedObjectPayload();
//System.err.println(" [Fetched GenSecProps over HTTP tunnel.]");
        return(result);
        }

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible, else only return an existing one
     */
    public ExhibitThumbnails getThumbnails(final String name, final boolean create)
        throws IOException
        {
//System.err.println("[Fetching Thumbnails over HTTP tunnel...]");

        final ByteArrayOutputStream baos =
            new ByteArrayOutputStream(name.length());
        final DataOutputStream dos = new DataOutputStream(baos);
        dos.writeUTF(name);
        dos.writeBoolean(create);
        final byte data[] = baos.toByteArray();
        dos.close(); // Attempt to free some resources quickly.

        final RawPacket response = doRPC(
            new RawPacket(RawPacket.OP_getThumbnails, data));

        final ExhibitThumbnails result = (ExhibitThumbnails) response.getSerializedObjectPayload();
//System.err.println(" [Fetched Thumbnails over HTTP tunnel.]");
        return(result);
        }

    /**Set variable.
     * Only non-local values are propagated upstream;
     * others are explicitly rejected with an UnsupportedOperationException.
     *
     * @throws java.io.IOException in case of I/O problems
     * @throws java.lang.UnsupportedOperationException  when attempting to set locals
     */
    public void setVariable(final SimpleVariableValue newValue)
        throws IOException,
        UnsupportedOperationException
        {
        if(newValue == null) { throw new IllegalArgumentException(); }
        if(newValue.getDef().isLocal())
            { throw new UnsupportedOperationException("cannot send local variables over the tunnel"); }
        // Implement in terms of setVariables().
        setVariables(new SimpleVariableValue[]{newValue});
        }

    /**Adjust globalMap for outgoing (upstream) set operation.
     * This expects the value passed to be:
     * <ul>
     * <li>Non-null
     * <li>Non-local
     * <li>Read/write
     * <li>To have an empty globalMap or one with exactly one entry
     *     whose key is the system ID for this system;
     *     if the map is empty then this system is put in the global map
     * </ul>
     * <p>
     * If the argument is in the correct form, it is returned unchanged.
     * <p>
     * If the argument is illegal, ie not suitable to send upstream,
     * then an exception is thrown to veto the send.
     *
     * @throws java.lang.IllegalArgumentException  if an invalid argument is passed,
     *     eg unable to be converted to be sent upstream because it has
     *     an invalid globalMap
     */
    private SimpleVariableValue adjustGlobalMapForSet(final SimpleVariableValue svv)
        {
        if(svv == null)
            { throw new IllegalArgumentException(); }

        final SimpleVariableDefinition def = svv.getDef();
        if(def.isLocal() ||
           def.isReadOnly() ||
           !SystemVariables.defs.contains(def))
            { throw new IllegalArgumentException(); }

        final Map<InstanceID,SimpleVariableValue> globalMap = svv.getGlobalMap();
        if((globalMap != null) && (globalMap.size() > 1))
            { throw new IllegalArgumentException(); }

        // If no globalMap,
        // then make sure that we make an explicit map entry for this system.
        final InstanceID sysID = (InstanceID) uniqueClientID.getValue();
        if(globalMap == null)
            {
            // We're done.
            return(svv.put(sysID, svv, true));
            }

        if(!Collections.singleton(sysID).equals(globalMap.keySet()))
            {
            throw new IllegalArgumentException("globalMap contains ID that is not ours during set: " + sysID);
            }

        // Value is fine as it is, so return it untouched...
        return(svv);
        }

    /**Set variables; must not be null or contain nulls.
     * Only non-local variables are sent upstream;
     * locals are silently discarded.
     * <p>
     * Variables not defined in SystemVariables are rejected.
     * <p>
     * Duplicates may be discarded or rejected,
     * or sent as is in which case it is undefined in which order they
     * are applied.
     * <p>
     * We try to set all values that we can for
     * <p>
     * Format on the wire:
     * <ul>
     * <li>Outbound: serialised SimpleVariableValue[]
     * <li>Inbound: number of values set (int)
     * </ul>
     *
     * @throws java.io.IOException in case of I/O problems
     */
    public int setVariables(final SimpleVariableValue[] newValues)
        throws IOException,
        UnsupportedOperationException
        {
        if(newValues == null) { throw new IllegalArgumentException(); }

        // Make a set of all non-local (and writable) values to pass upstream.
        // Preserve order...
        final int nvLen = newValues.length;
        final List<SimpleVariableValue> toGo = new ArrayList<SimpleVariableValue>(nvLen);
        for(int i = 0; i < nvLen; ++i)
            {
            final SimpleVariableValue svv = newValues[i];

            // Reject null values...
            if(svv == null)
                { throw new IllegalArgumentException(); }

            final SimpleVariableDefinition def = svv.getDef();

            // Skip local values;
            // they should not be propagated out of the machine.
            // Silently ignore this so that any globals can be sent.
            if(def.isLocal())
                { continue; }

            // Adjust globalMap if necessary for sending upstream.
            // Veto invalid values with an IllegalArgumentException.
            toGo.add(adjustGlobalMapForSet(svv));
            }

        // If nothing to send, then return immediately.
        if(toGo.size() == 0)
            { return(0); }

        // TODO: eliminate duplicates
        // TODO: sort into name order to help compression (but preserving order for the same variable)

        // Send set of values upstream
        // by putting them into an array and serialising it.
        final SimpleVariableValue r[] = new SimpleVariableValue[toGo.size()];
        toGo.toArray(r);

        // Make the RPC.
        final RawPacket response = doRPC(new RawPacket(
            RawPacket.OP_setVariables,
            r,
            true)); // Compression may be useful...

        // Extract return value...
        if(response.plLength != 4)
            { throw new IOException("corrupt reply packet: plLength=" + response.plLength); }
        final DataInputStream dis = new DataInputStream(response.getPayloadAsInputStream());
        final int nSet = dis.readInt();
        // Validate result returned.
        if((nSet < 0) || (nSet > nvLen))
            { throw new IOException("corrupt reply packet: nSet=" + Integer.toHexString(nSet)); }
        return(nSet);
        }

    /**Get variable, or returns null if no such non-local variable.
     * We attempt only to fetch non-local variables; return null for others
     * (except for the system ID).
     * <p>
     * When asked for the the system ID,
     * we always answer the question locally.
     * <p>
     * Note that the local system ID can be retrieved whether or not
     * the master server is responding.
     * <p>
     * Format on the wire is:
     * <ul>
     * <li>Outbound: variable name in UTF format
     * <li>Inbound: serialised form of SimpleVariableValue (or empty for null)
     * </ul>
     *
     * @throws java.io.IOException  in case of I/O error
     * @throws UnsupportedOperationException  if a local variable
     *     other than the system ID is requested
     */
    public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
        throws IOException,
               UnsupportedOperationException
        {
        if(var == null) { throw new IllegalArgumentException(); }

        // If the request is for the system ID,
        // return it here.
        if(var.equals(SystemVariables.LOCAL_SYS_ID))
            { return(uniqueClientID); }

        // Always return null to requests for local data (other than system ID);
        // local variables should not be propagated over the tunnel.
        if(var.isLocal())
            { throw new UnsupportedOperationException("local variables not available at tunnel source"); }

        // Fetch individual variable value over tunnel...
        // We only send the name,
        // and get back a whole variable value,
        // but on return check for a compatible defintion,
        // returning null if the defintions don't match.
        final String name = var.getName();
        final ByteArrayOutputStream baos =
            new ByteArrayOutputStream(name.length());
        final DataOutputStream dos = new DataOutputStream(baos);
        dos.writeUTF(name);
        final byte data[] = baos.toByteArray();
        dos.close(); // Attempt to free some resources quickly.

        final RawPacket request = new RawPacket(
            RawPacket.OP_getVariable,
            data);

        final RawPacket response = doRPC(request);
        final SimpleVariableValue svv =
            (SimpleVariableValue) response.getSerializedObjectPayload();

        // If there was no variable,
        // then return null immediately.
        if(svv == null)
            { return(null); }

        // Discard incompatible types,
        // and local variables.
        if(!var.equals(svv.getDef()))
            { return(null); }

        return(svv);
        }

    /**Fetch variable values from the master.
     * We fetch values from the master
     * (from which we eliminate any illegal values such as local variables
     * or values which are not locally valid)
     * and insert our unique local system ID
     * if the request stamp is -1 or older than our creation.
     * <p>
     * Note that the local system ID can be retrieved whether or not
     * the master server is responding providing the stamp is -1,
     * though in that case it might be the only value returned.
     * The caller should periodically use -1 to ensure that they
     * have a full set of global variables, eg in case the server rebooted.
     * <p>
     * Format on the wire:
     * <ul>
     * <li>Outbound: serialised timestamp
     * <li>Inbound: SimpleVariableValue[]
     * </ul>
     *
     * @throws java.io.IOException
     */
    public SimpleVariableValue[] getVariables(final long changedSince)
        throws IOException
        {
        final List<SimpleVariableValue> incoming = new ArrayList<SimpleVariableValue>();

        // Fetch all (global) variables over the tunnel.
        try
            {
            final RawPacket response = doRPC(
                new RawPacket(RawPacket.OP_getVariables, longSer(changedSince)));
            // Copy incoming array to List...
            incoming.addAll(Arrays.asList(
                (SimpleVariableValue[]) response.getSerializedObjectPayload()));
            }
        catch(IOException e)
            {
            // If changedSince == -1 we guarantee to return the local ID,
            // else we rethrow the exception...
            if(changedSince != -1) { throw e; }
            }

        // Remove any extant system ID value,
        // and any local variable (should not have come up the wire),
        // and any value that does not have an identical local definition.
        // We do not expect to find/remove any of these in normal operation.
        final SimpleVariableDefinition uCIDDef = uniqueClientID.getDef();
        for(int i = incoming.size(); --i >= 0; )
            {
            final SimpleVariableValue svv =
                    incoming.get(i);
            final SimpleVariableDefinition def = svv.getDef();
            if(def.equals(uCIDDef) ||
               def.isLocal() ||
               !SystemVariables.defs.contains(def) ||
               !def.equals(SystemVariables.nameToDef.get(def.getName())))
                {
                incoming.remove(i);
                continue;
                }
            }

        // Insert the system ID variable value into the result if appropriate.
        if(changedSince <= uniqueClientID.getTimestamp()) // Handles -1 case...
            {
            incoming.add(uniqueClientID);
            }

        final SimpleVariableValue result[] =
            new SimpleVariableValue[incoming.size()];
        incoming.toArray(result);
        return(result);
        }

    /**Get the current partial, or previous full, event set at the specified interval; never null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     * <p>
     * Implemented in terms of the more general call
     * in the hope that batched calls for several values will be
     * more common and more efficient.
     * This also reduces the number of distinct RPC calls that
     * we have to implement.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; may be empty but never null if requested set not available
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        if((def == null) || (intervalSelector == null))
            { throw new IllegalArgumentException(); }

        final long now = System.currentTimeMillis();
        final long currentIntervalNumber = intervalSelector.getIntervalNumber(now);
        final long intervalNumber = (current ? currentIntervalNumber : (currentIntervalNumber-1));

        try
            {
            final EventVariableValue evv[] = getEventValues(def,
                                                            intervalSelector,
                                                            intervalNumber,
                                                            null);

            // If we got some real (non-null) data back, then return it...
            if((evv.length > 0) && (evv[0] != null))
                { return(evv[0]); }
            }
        catch(final IOException e)
            {
            // Fall through to return empty non-auth default result.
            }

        // Else invent empty non-authoritative return value (with no events)...
        // TODO: Cache these to save some redundant object creation?
        return(new EventVariableValue(false,
                                      def,
                                      intervalSelector,
                                      intervalNumber,
                                      0, null, null));
        }

    /**Get the specified global event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * We do various small optimisations to reduce fruitless traffic over
     * what may be a slow and/or expensive connection:
     * <ul>
     * <li>This will reject requests will huge sets that go back to far so as
     *     to avoid huge RPC arguments and since the request is unlikely to be
     *     satisfiable.  The upstream implementation may be even more restrictive.
     * <li>This ignores requests too far in the future or the past.
     * <li>Trim leading empty slots from request sent over the tunnel.
     * </ul>
     * <p>
     * The responses from the server may be large and slow,
     * so we allow streaming to give better incremental CPU/resource consumption
     * and better throughput.
     * <p>
     * TODO: Optimise RPC call by allowing for sparse requests with different request/response format.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param intervalNumber  a time (as from System.currentTimeMillis())
     *     which identifies the first interval for which data is potentially
     *     required; if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc;
     *     null is treated as the common case equivalent to just bit 0 set
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     * 
     * @throws EOFException  if a local value is requested
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               BitSet whichValues)
        throws IOException
        {
        if((def == null) ||
            (intervalSelector == null) ||
            (intervalNumber < 0))
            { throw new IllegalArgumentException(); }

        if(def.isLocal())
            { throw new EOFException("cannot fetch local event values over tunnel"); }

        // Don't bother trying this if the connection is currently known to be broken.
        if(isBroken()) { throw new PGMasterNotInServiceException(); }

        // See if we can optimise the whichValues argument by replacing it
        // with a null value...
        // This will reduce heap churn, data on the wire, etc.
        if((whichValues != null) && (whichValues.length() == 1) && whichValues.get(0))
            { whichValues = null; }

        // Various optimisations are possible for a null BitSet arg.
        final boolean nullWV = (whichValues == null);

        // Take copy of BitSet to ensure that it is minimal size,
        // and to reduce thread/race/safety issues.
        final BitSet wvCopy = nullWV ? null : new BitSet(whichValues.length());
        if(!nullWV) { wvCopy.or(whichValues); }

        final int wvl = nullWV ? 1 : wvCopy.length();

        // Avoid pointless call for no request bits.
        if(wvl < 1)
            { return(NO_EVENT_VALUES); }

        // Avoid a huge request/response packet.
        if(wvl > 2*SystemVariables.EVENT_SAMPLES_RETAINED)
            { throw new IOException("request set too large to send over the wire"); }

        // Avoid a call for values too far in the past to be likely to be available.
        // (This also ensures that the intervalNumber is large and positive for later.)
        // Although we allow a zero value for the "all" bucket.
        final long currentInterval = intervalSelector.getIntervalNumber(System.currentTimeMillis());
        if(intervalNumber == 0)
            {
            if(!nullWV && (!whichValues.get(0) || (wvl != 1)))
                { throw new IllegalArgumentException(); }
            }
        else if(intervalNumber < currentInterval - 2*SystemVariables.EVENT_SAMPLES_RETAINED - 1)
            { return(NO_EVENT_VALUES); }

        // Avoid a call for values far too far in the future
        // for any skew to account for, etc, so we're sure no data is available.
        if(intervalNumber > currentInterval + 4*SystemVariables.EVENT_SAMPLES_RETAINED + 1)
            { return(NO_EVENT_VALUES); }

        // Get the offset of the first item in the request...
        // If this is greater than zero
        // then we can in principle optimise the request/response
        // by omitting the leading empty slots.
        final int firstItemOffset = nullWV ? 0 : wvCopy.nextSetBit(0);
        assert(firstItemOffset >= 0);

        // Reduce the BitSet that we send if the request can be trimmed...
        final BitSet wvToSend = (firstItemOffset == 0) ? wvCopy :
            (wvCopy.get(firstItemOffset, wvl));

        // Request is:
        //   * A byte consisting of the ordinal of period/interval enum ordinal.
        //   * The UTF-8 representation of the name of the definition.
        //   * The interval number (8 bytes).
        //   * The serialised form of the request BitSet if non-null.
        // Result is:
        //   * The in-order array of EventVariableValues (never null).
        final String name = def.getName();
        final ByteArrayOutputStream baos = new ByteArrayOutputStream(32 + name.length() + wvl/2);
        final DataOutputStream dos = new DataOutputStream(baos);
        dos.write(intervalSelector.ordinal());
        dos.writeUTF(name);
        dos.writeLong(intervalNumber - firstItemOffset);
        if(wvToSend != null)
            {
            final ObjectOutputStream oos = new ObjectOutputStream(dos);
            oos.writeObject(wvToSend);
            oos.flush();
            }
        dos.flush();
        final byte data[] = baos.toByteArray();
        dos.close(); // Attempt to free some resources quickly.

        final RawPacket request = new RawPacket(
            RawPacket.OP_getEventValues,
            data);

//        final RawPacket response = doRPC(request);
//        if(IsDebug.isDebug) { logger.log("[ExhibitDataTunnelSource.getEventValues("+def+", "+intervalSelector+", "+intervalNumber+", "+whichValues+"): request/response bytes = " + request.plLength+"/"+response.plLength+".]"); }

        // Stream the result, since it may be large enough to be on the wire for a while.
        final EventVariableValue[] evvsRaw;
        try
            {
            final Pair<Integer, InputStream> stream = doRPCRawWithStreamResponse(request, false);
            try
                {
                final ObjectInputStream ois = new ObjectInputStream(stream.second);
                evvsRaw = (EventVariableValue[]) ois.readObject();
                ois.close();
                }
            finally { stream.second.close(); /* Ensure underlying resources are released. */ }

            // Check for unexpected return values.
            if(evvsRaw == null)
                { throw new IOException("unexpected null return value"); }

            // If we offset the request then make a copy set back.
            final EventVariableValue[] evvs;
            if(firstItemOffset != 0)
                {
                evvs = new EventVariableValue[evvsRaw.length + firstItemOffset];
                System.arraycopy(evvsRaw, 0, evvs, firstItemOffset, evvsRaw.length);
                }
            else
                {
                // Did not offset the request...
                evvs = evvsRaw;
                }

            // Check for unexpected/incorrect return values.
            for(int i = evvs.length; --i >= 0; )
                {
                final EventVariableValue evv = evvs[i];
                if(evv == null) { continue; }
                if(!def.equals(evv.getDef()))
                    { throw new IOException("unexpected return value with wrong def"); }
                if(!intervalSelector.equals(evv.getPeriod()))
                    { throw new IOException("unexpected return value with wrong period"); }
                if(evv.getIntervalNumber() != intervalNumber - i)
                    { throw new IOException("unexpected return value with wrong interval number"); }
                }

            _noteResult(true); // Seems to have succeeded.
            return(evvs);
            }
        catch(final IOException e) { _noteResult(false); throw e; }
        catch(final Exception e) { _noteResult(false); throw new IOException(e.getMessage()); }
        }

    /**Synchronise variables with upstream values.
     * Pushes updated values upstream to the source,
     * calls sync on the source if called with the "force" argument true,
     * and then retrieves changed values from upstream.
     * <p>
     * When called with force==true, this acts like a full "memory barrier",
     * flushing all write-cached items downstream immediately and afterwards
     * getting the value of all upstream values with getVariables(-1),
     * but may be expensive in terms of CPU or bandwidth, so use sparingly.
     * <p>
     * When called with force=false, this returns immediately
     * and the operation is not propagated across the tunnel.
     * <p>
     * In any case, it is rarely the right thing for a casual user
     * to vall this as it may be very expensive.
     *
     * @param force  if true, this will force a full write flush,
     *     a full sync upstream,
     *     then full read with getVariables(-1),
     *     to get the effect of a full "barrier";
     *     otherwise, do nothing
     *
     * @throws IOException if one is received from upstream
     */
    public void syncVariables(final boolean force)
        throws IOException
        {
        // Do not propagate across tunnel unless "force"d.
        if(!force)
            { return; }

        // Propagate the "force"d sync upstream,
        // and return when done...
        doRPC(new RawPacket(RawPacket.OP_syncVariables, EMPTY_PAYLOAD));
        }


    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(final PropsKey key,
                                              final long versionID)
        throws IOException
        {
        throw new IOException("NOT IMPLEMENTED");
        }


    /**Poll periodically.
     * We can use this to attempt to repair ailing connections, etc,
     * as well as poll for asynchronous messages being sent to us by the master.
     * <p>
     * However, our general policy is not to force traffic if we need not.
     * <p>
     * This can be overridden by a derived class,
     * though it is suggested that this be called with super.poll(gp) if so.
     */
    public void poll(final GenProps gp)
        {
        if(KEEP_SERVER_CONNECTION_ALIVE)
            {
            // If we have a broken connection to the master,
            // then a little while before other users can retry it,
            // we retry it to see if the master is reachable again.
            //
            // This enables us to repair a connection silently,
            // so that a normal caller is never blocked until
            // it is repaired (other calls are vetoed quickly).
            //
            // Our calls are still backed off if the master is really unwell.
            //
            // This does not try to force an initial connection,
            // just repair a failing one.
            final Long notUntil = doNotTryMasterUntil;
            if((notUntil != null) &&
               (notUntil.longValue() - _REPAIR_RETRY_LEAD_MS < System.currentTimeMillis()))
                {
                try { doNOOP(true); }
                catch(final Exception e)
                    {
                    // It is possible that our connection problem is caused
                    // by a resource leak (eg of unclosed descriptors) elsewhere
                    // so try and force come cleanup now.
                    //
                    // We rely on our normal pacing/backoff of retries
                    // to avoid us hurting the system too much if we are wrong!
                    if(FORCE_GC_AND_FINALISERS_AFTER_POLL_FAIL)
                        {
logger.log("[ExhibitDataTunnelSource: WARNING: failing to re-establish connection so attempting to free resources with gc(), etc...]");
                        System.runFinalization();
                        System.gc();
                        System.runFinalization();
                        }
                    }
                }
            }
        }

    /**May attempt to free up outbound connections and/or prevent new ones. */
    public void destroy()
        {
        // Prevent most new outgoing connections.
        doNotTryMasterUntil = new Long(Long.MAX_VALUE);
        }


    /**How soon ahead of other callers we try to silently repair a broken connection, ms; non-negative. */
    private final int _REPAIR_RETRY_LEAD_MS = 17001 + Rnd.fastRnd.nextInt(3007);


    /**Make an RPC call over the underlying medium with the given outgoing packet; never null.
     * This collects the response packet and will object
     * if it sees any IOException
     * or if the packets come back with the wrong op code.
     * <p>
     * This can be locked on the instance lock to serialise RPCs,
     * if the tunnel can only usefully handle one call at once.
     * <p>
     * Must be implemented by the deriving class
     * to suit its transmission medium.
     *
     * @param  packetOut  request packet; never null
     * @return response packet; never null
     * @throws java.io.IOException  in case of I/O difficulties
     */
    protected abstract RawPacket doRPCRaw(final RawPacket packetOut)
        throws IOException;

    /**Optimised RPC call with the given outgoing packet and returning packet body as an InputStream; never null nor with a null stream.
     * This is an <em>optimisation</em> for performance-critical cases <em>only</em>,
     * and foregoes some error checking/handling for speed
     * (and thus the caller should ensure that it performs integrity checks).
     * There is an onus on the streaming-related code to behave safely
     * even if fed bogus/corrupt data,
     * and it must be able to safely/easily undo any work done
     * if the message is found to be bogus as late as the final byte(s).
     * <p>
     * This streams the content of the response packet
     * and will object if it sees any IOException
     * or if the packets come back with the wrong op code.
     * <p>
     * A terminating trailer byte may or may not be visible on the returned stream
     * thus allowing the implementation to be as efficient as possible.
     * <p>
     * This may return after all the input data has been collected,
     * or while some or all is still to come,
     * and thus the returned stream may fail and throw an exception.
     * <p>
     * The first element of the result is the length of the response data
     * (not including any non-data trailer bytes from the packet even if present)
     * but may be null if this length is not known when the packet header is seen,
     * eg because the packet body was compressed.
     * <p>
     * This may be implemented/overridden by the deriving class
     * to suit its transmission medium,
     * and as an optimisation to reduce copying and allow streaming,
     * ie starting to process the input before it is all received.
     * <p>
     * The data stream is always of uncompressed data,
     * regardless of whether the data was sent compressed on the wire,
     * ie this routine will correctly decompress data on the fly as/when needed.
     * <p>
     * The caller <strong>must close</strong> the stream promptly
     * to release resources such as file handles and non-Java memory.
     *
     * @param  packetOut  request packet; never null
     * @param allowBigReadTimeout TODO
     * @return response  length and data stream; never null
     * @throws java.io.IOException  in case of I/O difficulties
     */
    protected Tuple.Pair<Integer, InputStream> doRPCRawWithStreamResponse(final RawPacket packetOut, boolean allowBigReadTimeout)
        throws IOException
        {
        // Simple implementation in terms of basic doRPCRaw().
        final RawPacket result = doRPCRaw(packetOut);

        // Check for mismatched response packets
        // or remote exceptions codes with special response op-codes.
        if(result.opCode != packetOut.opCode)
            {
            // Deal specially with some fixed remote exception types.
            switch(result.opCode)
                {
                case RawPacket.OP__PGMNISEX:
                    { throw new PGMasterNotInServiceException(); }
                case RawPacket.OP__RUNTEX:
                    { throw new RuntimeException("remote threw RuntimeException"); }
                case RawPacket.OP__REMEX:
                    { throw new IOException("remote threw RemoteException"); }
                case RawPacket.OP__INTEX:
                    { throw new InterruptedIOException("remote operation interrupted"); }
                }
            throw new IOException("packet response-type mismatch; got "+result.opCode+" expected "+packetOut.opCode);
            }

        final Tuple.Pair<Integer, InputStream> resultStream =
            new Tuple.Pair<Integer, InputStream>(result.plLength, result.getPayloadAsInputStream());
        return(resultStream);
        }


    /**Make an RPC call over HTTP[S] with the given outgoing packet.
     * This provides error recovery, back-off, stats, etc,
     * and serves as a wrapper for the medium-specific doRPCRaw().
     * <p>
     * Called by all the public data-pipeline methods to transport information
     * over the tunnel.
     *
     * @param  packetOut  request packet; never null
     * @return  response packet; never null
     * @throws java.io.IOException  in case of I/O difficulties
     */
    protected RawPacket doRPC(final RawPacket packetOut)
        throws IOException
        {
        // Note RPC request...
        StatsLogger.captureDataPoint(statsIDTS, TSNAME_RPCREQUEST);
        // Note its type...
        StatsLogger.captureDataPoint(statsIDTS, TSNAMEPR_RPCTYPE + ((int) packetOut.opCode));

        // If last communication with master failed too recently,
        // quickly veto a new connection attempt for now.
        if(isBroken())
            {
if(_protocolDebug) { System.err.println("[doRPC(): vetoing connection to broken master.]"); }
            throw new PGMasterNotInServiceException("no connection to master");
            }

        return(doRPCUnguarded(packetOut));
        }

    /**Just like doRPC() but does not back off in face of previous failures; never null.
     * Can be used for calls that must attempt to contact the master anyway,
     * or are probing for it to be alive.
     * <p>
     * Most callers should use the normal doRPC().
     *
     * @param packetOut  never null
     * @return  response to RPC call; never null
     */
    protected RawPacket doRPCUnguarded(final RawPacket packetOut)
        throws IOException
        {
        if(_protocolDebug) { System.err.println("[doRPC(): connecting to master.]"); }

        try {
            if(_protocolDebug) { System.err.println(" [doRPC(): about to do raw RPC.]"); }

            // Do the raw, medium-specific RPC.
            final RawPacket result = doRPCRaw(packetOut);

            if(_protocolDebug) { System.err.println(" [doRPC(): about to check result type.]"); }

            // Check for mismatched response packets
            // or remote exceptions codes with special response op-codes.
            if(result.opCode != packetOut.opCode)
                {
                // Deal specially with some fixed remote exception types.
                // Note that we don't count a OP__RUNTEX as a success OR failure
                // for the purposes of link monitoring;
                // this may be the master gently rebuffing a request for some reasons.
                switch(result.opCode)
                    {
                    case RawPacket.OP__PGMNISEX:
                        { throw new PGMasterNotInServiceException(); }
                    case RawPacket.OP__RUNTEX:
                        { throw new RuntimeException("remote threw RuntimeException"); }
                    case RawPacket.OP__REMEX:
                        { throw new IOException("remote threw RemoteException"); }
                    case RawPacket.OP__INTEX:
                        { throw new InterruptedIOException("remote operation interrupted"); }
                    }
                throw new IOException("packet response-type mismatch; got "+result.opCode+" expected "+packetOut.opCode);
                }

            // Note last successful use.
            _noteResult(true);

            if(_protocolDebug) { System.err.println(" [doRPC(): about to return result.]"); }

            // Return result successfully.
            return(result);
            }
        catch(final InterruptedIOException e)
            {
            // Not an error (nor a success): can retry later.
            throw e;
            }
        catch(final IOException e)
            {
            // Regard (only) IOException as indicator or some sort of connection problem.

            // Note RPC request failure with IOException...
            StatsLogger.captureDataPoint(statsIDTS, TSNAME_RPCIOEX);

            // Note failed connection to the master...
            // This will defer our next attempt...
            _noteResult(false);

            // Note failure in the logs:
            logger.log("[doRPC(): failed connection attempt upstream, failed packet: " + packetOut + "...]");
            final long lastS = lastSuccessfulConnectionTime;
            if(lastS > 0)
                { logger.log("[doRPC(): last success at "+(new Date(lastS))+".]"); }
            final Long notUntil2 = doNotTryMasterUntil;
            if(notUntil2 != null)
                { logger.log("[doRPC(): deferring next attempt until " + (new Date(notUntil2.longValue()))  + ".]"); }

            // Rethrow the error that we encountered...
            throw e;
            }
        }

    /**Cache/lock to improve performance of inbound RPC.
     * Meant to be opaque to all but handleInboundRPC()
     * and is used to provide a better overall responsiveness,
     * especially from expensive calls
     * by vetoing concurrent expensive calls
     * and by cacheing the responses from especially expensive calls.
     * <p>
     * Not all possibilities are exploited,
     * just those that in practice seem to be important.
     */
    public static final class HIRPCCache
        {
        /**Lock to prevent servicing more than one very expensive call at once.
         * We do this to reduce strain on the memory/GC and CPU of the server.
         */
        private final ReentrantLock slowResponseLock = new ReentrantLock();

        /**Cache of AEP as pair of longHash and response packet for simple AEP fetch RPC; mutable.
         * Initially null.
         * <p>
         * Elements of this pair are never null.
         * <p>
         * Marked volatile for safe multi-threaded access.
         * <p>
         * Currently held indefinitely,
         * until the AEP fetched has a new hash,
         * not for example held via a SoftReference
         * since a SoftReference would probably be cleared too soon
         * for this to be effective.
         * <p>
         * This is likely to be many MB in size.
         */
        private volatile Tuple.Pair<Long, RawPacket> _AEP_response;
        
        /**Cache of full extra-compressed AEP as tuple of longHash, compression format, and response packet for diff AEP fetch RPC; mutable.
         * Initially null.
         * <p>
         * Elements of this tuple are never null.
         * <p>
         * Marked volatile for safe multi-threaded access.
         * <p>
         * Currently held indefinitely,
         * until the AEP fetched has a new hash,
         * not for example held via a SoftReference
         * since a SoftReference would probably be cleared too soon
         * for this to be effective.
         * <p>
         * This is likely to be many MB in size.
         */
        private volatile Tuple.Pair<Long, Tuple.Pair<CompressionLevel, RawPacket>> _AEP_extracomp_response;

        /**Cache of extra-compressed AEP diff as tuple of longHash, compression format, and response packet for diff AEP fetch RPC; mutable.
         * Initially null.
         * <p>
         * Elements of this tuple are never null.
         * <p>
         * Marked volatile for safe multi-threaded access.
         * <p>
         * Currently held indefinitely,
         * until the AEP fetched has a new hash,
         * not for example held via a SoftReference
         * since a SoftReference would probably be cleared too soon
         * for this to be effective.
         * <p>
         * This is likely to be many MB in size.
         */
        private volatile Tuple.Pair<Tuple.Pair<AllExhibitProperties,AllExhibitPropertiesDelta>, Tuple.Pair<CompressionLevel, RawPacket>> _AEP_diff_response;

        /**Cache of previous AEP value to assist with AEP diffs; initially null.
         * Currently held indefinitely,
         * until the AEP fetched has a new hash,
         * not for example held via a SoftReference
         * since a SoftReference would probably be cleared too soon
         * for this to be effective.
         */
        private volatile AllExhibitProperties aepPrev;

        /**Cache of older AEP values to assist with AEP diffs; never null.
         * This is present to help with requests for diffs against AEPs
         * older than the previous value, if any such requests are made.
         * <p>
         * Thread-safe cache of SoftReferences to older AEPs (older values first).
         * <p>
         * It is possible to synchronise on this instance to exclude other activity.
         * <p>
         * We periodically remove entries whose references have been cleared.
         */
        private final List<SoftReference<AllExhibitProperties>> prevAEPs =
            new Vector<SoftReference<AllExhibitProperties>>();
        }

    /**Handles input request packet from slave across a tunnel.
     * Generates the response packet or throws an exception.
     *
     * @param reqPacket  the request packet; never null
     * @param clientAddr  the tunnel client's address as seen by us;
     *     may be null if not applicable or available
     * @param cache  if not null, then selected routines with
     *     slow/expensive responses will not be serviced concurrently
     *     but concurrent calls will be vetoed quickly instead
     *     and expensive-to-compute values may be (partially) cached
     * @return the response packet; never null
     *
     * @throws java.io.IOException  in case of difficulty upstream
     * @throws PGMasterNotInServiceException  if the upstream data source
     *     is down/unavailable
     */
    public static RawPacket handleInboundRPC(final SimpleExhibitPipelineIF source,
                                             final RawPacket reqPacket,
                                             final String clientAddr,
                                             final HIRPCCache cache)
        throws IOException,
               PGMasterNotInServiceException
        {
        if((source == null) || (reqPacket == null))
            { throw new IllegalArgumentException(); }

        switch(reqPacket.opCode)
            {
            // For NO-OP we expect an empty data section
            // (we simply ignore it even if not empty)
            // and reply with an empty data section.
            case RawPacket.OP_NOOP:
                {
                // Create and send the (empty) response packet.
                // We don't attempt to compress this.
                return(new RawPacket(
                    reqPacket.opCode,
                    EMPTY_PAYLOAD,
                    false)); // Don't attempt compression...
                }

            // getGenProps().
            case RawPacket.OP_getGenProps:
                {
                final long stamp = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readLong();
                final GenProps gp = source.getGenProps(stamp);
                // Create and send the response packet
                // (empty for null response,
                // else serialised form, compressed when helpful).
                return(new RawPacket(
                    reqPacket.opCode,
                    gp,
                    true)); // Attempt compression...
                }

            // getGenSecProps().
            case RawPacket.OP_getGenSecProps:
                {
                final long stamp = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readLong();
                final Properties gsp = source.getGenSecProps(stamp);
                // Create and send the response packet
                // (empty for null response,
                // else serialised form, compressed when helpful).
                return(new RawPacket(
                    reqPacket.opCode,
                    gsp,
                    true)); // Attempt compression...
                }

            // getAllExhibitImmutableData().
            case RawPacket.OP_getAllExhibitImmutableData:
                {
                final long stamp = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readLong();
                final AllExhibitImmutableData aeid =
                    source.getAllExhibitImmutableData(stamp);

                // Simple case; nothing to return since caller is up to date,
                // or we allow concurrent expensive calls.
                if((aeid == null) || (cache == null))
                    {
                    return(new RawPacket(
                        reqPacket.opCode,
                        aeid,
                        true)); // Allow compression.
                    }

                // If the response is other than null
                // and we reject concurrent expensive calls,
                // then veto it if another such call is already in progress.
                if(!cache.slowResponseLock.tryLock()) { throw new IOException("expensive call already in progress"); }
                // Create and send the response packet
                // (empty for null response,
                // else serialised form, compressed when helpful).
                try
                    {
                    return(new RawPacket(
                        reqPacket.opCode,
                        aeid,
                        true)); // Allow compression.
                    }
                finally { cache.slowResponseLock.unlock(); }
                }

                // getAllExhibitProperties().
                case RawPacket.OP_getAllExhibitProperties:
                    {
                    final long hash = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                            readLong();
                    final AllExhibitProperties aep =
                        source.getAllExhibitProperties(hash);

                    // Simple case; nothing to return since caller is up to date,
                    // or we allow concurrent expensive calls.
                    if((aep == null) || (cache == null))
                        {
                        return(new RawPacket(
                            reqPacket.opCode,
                            aep,
                            true)); // Allow compression.
                        }

                    // If we already have a suitable response cached
                    // ie for the very same AEP logHash
                    // then return it immediately.
                    // We assume therefore that the entire packet will
                    // be the same for the same response data
                    // regardless of client, etc.
                    //
                    // (This does mean that the caller may miss some
                    // internally cached updates within the AEP instance, etc,
                    // but nothing that it cannot recompute if needed.)
                    final Tuple.Pair<Long, RawPacket> cached = cache._AEP_response;
                    if((cached != null) && (aep.longHash == cached.first.longValue()))
                        {
                        // Good, can use cached packet...
                        assert(cached.second != null);
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): returning cached AEP response packet: "+cached.second+".]"); }
                        return(cached.second);
                        }

                    // If the response is other than null
                    // and we reject concurrent expensive calls,
                    // then veto it if another such call is already in progress.
                    if(!cache.slowResponseLock.tryLock()) { throw new InterruptedIOException("expensive call already in progress: please retry"); }
                    // Create and send the response packet
                    // (empty for null response,
                    // else serialised form, compressed when helpful).
                    try
                        {
                        // Compute the response packet...
                        RawPacket response = new RawPacket(
                                                reqPacket.opCode,
                                                aep,
                                                true); // Allow compression.
                        // Cache it if the AEP is not null.
                        if(aep != null)
                            {
                            // We intern() the (probably very large) response
                            // to try to avoid holding duplicates
                            // and to reduce old-generation heap churn.
                            response = MemoryTools.intern(response);
                            cache._AEP_response = new Tuple.Pair<Long,RawPacket>(new Long(aep.longHash), response);
                            }
                        // Return it...
                        return(response);
                        }
                    finally { cache.slowResponseLock.unlock(); }
                    }

            // getAllExhibitProperties(AEP).
            case RawPacket.OP_getAllExhibitPropertiesDiff:
                {
                // Prime a/new empty cache with an extant non-empty AEP
                // so that we are likely to be able to service
                // the first responses after an AEP change with deltas.
                if(cache != null)
                    {
                    if((cache.aepPrev == null) || (cache.aepPrev.aeid.length == 0))
                        {
                        final AllExhibitProperties first = source.getAllExhibitProperties(-1);
                        if((null != first) && (first.aeid.length != 0))
                            {
                            cache.aepPrev = first;
                            cache.prevAEPs.add(new SoftReference<AllExhibitProperties>(first));
                            }
                        }
                    }

                final DataInputStream dis = new DataInputStream(reqPacket.getPayloadAsInputStream());
                final long hash = dis.readLong();
                // We read the compression level, but need not parse it yet.
                final String compression = dis.readUTF();
                final AllExhibitProperties aep =
                    source.getAllExhibitProperties(hash);

                // Simple case; nothing to return since the caller is up to date.
                if(aep == null)
                    {
                    return(new RawPacket(
                                    reqPacket.opCode,
                                    ExhibitDataTunnelSource.EMPTY_PAYLOAD,
                                    false)); // No compression.
                    }

                // Full implementation is too expensive sans cache.
                if(cache == null)
                    { throw new IOException("operation not fully implemented due to resource constraints"); }

                // Decode the client's maximum supported compression level.
                // This will throw an IllegalArgumentException if unparseable.
                final CompressionLevel cl = CompressionLevel.valueOf(compression);


                // If we already have a suitable response cached
                // ie for the very same AEP diff with the same old/new hashes.
                // then return it immediately.
                // We assume therefore that the entire packet will
                // be the same for the same response data
                // regardless of client, etc.
                //
                // (This does mean that the caller may miss some
                // internally cached updates within the AEP instance, etc,
                // but nothing that it cannot recompute if needed.)
                //
                // Try first for a cached diff/delta.
                final Tuple.Pair<Tuple.Pair<AllExhibitProperties,AllExhibitPropertiesDelta>,
                    Tuple.Pair<CompressionLevel, RawPacket>> cachedDiff = cache._AEP_diff_response;
                if((hash != -1) &&
                    (cachedDiff != null) &&
                    (aep.longHash == cachedDiff.first.second.longHashAEPAfter) &&
                    (hash == cachedDiff.first.second.longHashAEPBefore))
                    {
                    // If result is compressed at level higher than caller can handle
                    // then we abort.
                    // (We could chose to try for the full AEP rather than abort now,
                    // but it's unlikely to work if this didn't.)
                    if(cachedDiff.second.first.getLevel() > cl.getLevel())
                        { throw new IOException("compression level too high for client"); }

                    // Check that correct client AEP is held.
                    assert(cachedDiff.first.first.longHash == hash);
                    // Check that we can construct the correct new AEP.
//                    assert(aep.equals(AllExhibitPropertiesDelta.applyDiff(cachedDiff.first.first, cachedDiff.first.second)));

                    // Good, can use cached diff packet...
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): returning cached AEP (diff) delta response packet: "+cachedDiff.second+".]"); }
                    return(cachedDiff.second.second);
                    }                
                
                // If the response is other than null
                // and we reject concurrent expensive calls,
                // then veto it if another such call is already in progress.
                // We allow a very limited wait for the lock.
                try
                    {
                    if(!cache.slowResponseLock.tryLock(CoreConsts.MAX_INTERACTIVE_DELAY_MS, TimeUnit.MILLISECONDS))
                        { throw new InterruptedIOException("expensive call already in progress: please retry later"); }
                    }
                catch(final InterruptedException e)
                    {
                    final InterruptedIOException err = new InterruptedIOException(e.getMessage());
                    err.initCause(e);
                    Thread.currentThread().interrupt(); // Don't lose "interrupt" status.
                    throw err;
                    }

                // Create and send the response packet
                // (empty for null response,
                // else serialised form, full or delta/diff, compressed when helpful).
                try
                    {
                    // Note whatever was the previous AEP that we had sequestered,
                    // partly to stop it being GCed until we finish handling this RPC.
                    final AllExhibitProperties oldAepPrev = cache.aepPrev;
                    
                    // If the AEP has changed then cache it for deltas, etc.
                    if((null == oldAepPrev) || (aep.longHash != oldAepPrev.longHash))
                        {
                        cache.aepPrev = aep;
                        cache.prevAEPs.add(new SoftReference<AllExhibitProperties>(aep));
                        }

                    // See if we have the correct cached extra-compressed full AEP.
                    final Tuple.Pair<Long, Tuple.Pair<CompressionLevel, RawPacket>> cachedFull = cache._AEP_extracomp_response;
                    final RawPacket cachedFullAEPResponse;
                    if((cachedFull != null) && (aep.longHash == cachedFull.first.longValue()))
                        { cachedFullAEPResponse = cachedFull.second.second; }
                    else
                        { cachedFullAEPResponse = null; }                  

                    // Look for extant AEP to diff against.
                    // Zap any dead cache entries that we find in passing.
                    if(hash != -1)
                        {
                        AllExhibitProperties callerAEP = null;
                        synchronized(cache.prevAEPs)
                        {
                        // Start from the most recent items first.
                        // Avoid using get() on old(er) items if possible,
                        // since that would artificially keep them from being GCed.
                        for(int i = cache.prevAEPs.size(); --i >= 0; )
                            {
                            final AllExhibitProperties caep = cache.prevAEPs.get(i).get();
                            if(null == caep)
                                {
                                cache.prevAEPs.remove(i);
                                continue;
                                }
                            
                            if(caep.longHash == hash)
                                {
                                // Good!
                                // We have the same AEP that the caller does!
                                callerAEP = caep;
                                break;
                                }
                            }
                        }
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): old AEP cache size:"+cache.prevAEPs.size()+", caep="+callerAEP+".]"); }
                        // Do not hold lock on cache.prevAEPs
                        // while trying to create and return diff/delta.
                        if(callerAEP != null)
                            {
                            try
                                {
                                final long diffStartTime = System.currentTimeMillis();
                                // We can do a diff/delta!
                                final AllExhibitPropertiesDelta diff = 
                                    AllExhibitPropertiesDelta.createDiff(callerAEP, aep, false);
                                final long diffEndTime = System.currentTimeMillis();
                                final Pair<CompressionLevel, byte[]> saepd =
                                    GenUtils.compressObject(diff, MAX_AEP_DIFF_COMP_LEVEL);
                                final long compEndTime = System.currentTimeMillis();

                                // Compute the response packet...
                                // Only allow packet-level compression if our compression level is null.
                                final ByteArrayOutputStream boas = new ByteArrayOutputStream(32 + saepd.second.length);
                                final DataOutputStream dos = new DataOutputStream(boas);
                                dos.writeUTF(saepd.first.name());
                                dos.write(saepd.second);
                                dos.close();
                                RawPacket response = new RawPacket(
                                                reqPacket.opCode,
                                                boas.toByteArray(),
                                                CompressionLevel.NONE == saepd.first);
                                
                                // Abort if we have a full (non-delta) response
                                // and the delta seems larger than the full responsw.
                                if((null != cachedFullAEPResponse) &&
                                   (cachedFullAEPResponse.plLength <= response.plLength))
                                    { throw new IOException("AEP delta larger than full response"); }

                                // Check that correct client AEP is held.
                                assert(callerAEP.longHash == hash);
                                // Check that we can construct the correct new AEP.
                                assert(aep.equals(AllExhibitPropertiesDelta.applyDiff(callerAEP, diff)));

                                // We intern() the (probably large) response
                                // to try to avoid holding duplicates
                                // and to reduce old-generation heap churn.
                                response = MemoryTools.intern(response);
                                final Tuple.Pair<CompressionLevel, RawPacket> newDiffValue = new Tuple.Pair<CompressionLevel,RawPacket>(saepd.first, response);
                                cache._AEP_diff_response = new Tuple.Pair<Tuple.Pair<AllExhibitProperties,AllExhibitPropertiesDelta>, Tuple.Pair<CompressionLevel,RawPacket>>(
                                                new Tuple.Pair<AllExhibitProperties,AllExhibitPropertiesDelta>(callerAEP, diff), newDiffValue);
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): computed and cached AEP (diff) delta response packet: "+newDiffValue+"; times: delta/comp="+(diffEndTime-diffStartTime)+"ms/"+(compEndTime-diffEndTime)+"ms; full: "+cachedFullAEPResponse+"; last full extra compressed response: "+cachedFull+".]"); }

                                // If result is compressed at level higher than caller can handle
                                // then we must abort.
                                if(saepd.first.getLevel() > cl.getLevel())
                                    { throw new IOException("compression level too high for client"); }

                                // Return it...
                                return(response);
                                }
                            catch(final InterruptedException e)
                                {
                                final InterruptedIOException err = new InterruptedIOException("interrupted RPC");
                                err.initCause(e);
                                throw err;
                                }
                            catch(final Exception e)
                                {
if(IsDebug.isDebug) { e.printStackTrace(); }
                                /* Fall through to send full AEP. */
                                }
                            }
                        }

                    // CANNOT COMPUTE A DIFF...
                    // So try to return/compute/cache a full response.
                    if(cachedFullAEPResponse != null)
                        {
                        // If result is compressed at level higher than caller can handle
                        // then we must abort.
                        if(cachedFull.second.first.getLevel() > cl.getLevel())
                            { throw new IOException("compression level too high for client"); }

                        // Good, can use cached packet...
/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): returning cached AEP (diff) response packet: "+cachedFull.second+".]"); }
                        return(cachedFull.second.second);
                        }

                    // Compute a new full AEP response packet...
                    // Only allow packet-level compression if our compression level is null.
                    // Super-compressed full AEP; null if not (yet) available/computed.
                    final Pair<CompressionLevel, byte[]> saep = GenUtils.compressObject(aep, MAX_AEP_DIFF_COMP_LEVEL);
                    ByteArrayOutputStream boas = new ByteArrayOutputStream(32 + saep.second.length);
                    DataOutputStream dos = new DataOutputStream(boas);
                    dos.writeUTF(saep.first.name());
                    dos.write(saep.second);
                    dos.close();
                    RawPacket response = new RawPacket(
                                    reqPacket.opCode,
                                    boas.toByteArray(),
                                    CompressionLevel.NONE == saep.first);
                    dos = null; // Help GC.
                    boas = null; // Help GC.

                    // We intern() the (probably very large) response
                    // to try to avoid holding duplicates
                    // and to reduce old-generation heap churn.
                    response = MemoryTools.intern(response);
                    final Tuple.Pair<CompressionLevel, RawPacket> newFull = new Tuple.Pair<CompressionLevel,RawPacket>(saep.first, response);
                    cache._AEP_extracomp_response = new Tuple.Pair<Long, Tuple.Pair<CompressionLevel,RawPacket>>(new Long(aep.longHash), newFull);

/* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataTunnelSource.handleInboundRPC(): computed and cached AEP (diff) response packet: "+newFull+".]"); }

                    // If result is compressed at level higher than caller can handle
                    // then we must abort.
                    if(saep.first.getLevel() > cl.getLevel())
                        { throw new IOException("compression level too high for client"); }
                    
                    // Return it...
                    return(response);
                    }
                catch(final InterruptedException e)
                    {
                    final InterruptedIOException err = new InterruptedIOException("interrupted RPC");
                    err.initCause(e);
                    throw err;
                    }
                finally { cache.slowResponseLock.unlock(); }
                }

            // getThumbnails().
            case RawPacket.OP_getThumbnails:
                {
                final DataInputStream dis = new DataInputStream(reqPacket.getPayloadAsInputStream());
                final String name = dis.readUTF();
                final boolean create = dis.readBoolean();
                dis.close(); // Hopefully free some resources quickly.
                final ExhibitThumbnails eth =
                    source.getThumbnails(name, create);
                // Create and send the response packet
                // (empty for null response, else serialised form).
                // We allow compression because it may save a little bandwidth.
                return(new RawPacket(
                    reqPacket.opCode,
                    eth,
                    true)); // Allow compression, though probably marginal.
                }

            // getStaticAttr().
            case RawPacket.OP_getStaticAttr:
                {
                final String name = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readUTF();
                final ExhibitStaticAttr esa =
                    source.getStaticAttr(name);
                // Create and send the response packet
                // (empty for null response,
                // else serialised form, compressed when helpful).
                return(new RawPacket(
                    reqPacket.opCode,
                    esa,
                    true)); // Allow compression.
                }

            // getRawFile().
            // We satisfy the initial portion of overlength requests.
            case RawPacket.OP_getRawFile:
                {
                final DataInputStream dis = new DataInputStream(reqPacket.getPayloadAsInputStream());
                final String name = dis.readUTF();
                final ExhibitStaticAttr esa = source.getStaticAttr(name);
                if(esa == null)
                    { throw new IOException("invalid argument: bad exhibit name"); }
                final int start = dis.readInt();
                // Coerce afterEnd into range, and allow zero-length requests.
                final int afterEnd = (int) Math.min(dis.readInt(), esa.length);
                // Disallow patently absurd values, eg due to request-packet corruption.
                if((start < 0) || (afterEnd < start))
                    { throw new IOException("invalid argument: bad range"); }
                final boolean dontCache = dis.readBoolean();
                dis.close(); // Hopefully free some resources quickly.
                final int len = afterEnd - start;
                if(len > MAX_USER_READ_SIZE)
                    { throw new IOException("invalid argument: request too large"); }

                // Fetch the data from upstream...
                final byte rawData[] = new byte[len];
                final ByteBuffer buf = ByteBuffer.wrap(rawData);
                // Fetch whatever we quickly can.
                source.getRawFile(buf, name, start, dontCache);
                // Flip the buffer to nominally switch to reading from it.
                buf.flip();

                // Create and send the response packet containing raw exhibit data.
                //
                // Useful compression may be possible for some exhibit fragments/types,
                // but here we are slightly uncomfortable with
                // the potential vulnerability to invisible corruption of binary data
                // on core (and non-re-compressable) exhibit type (JPEG)
                // with no application-level checksum on these frames,
                // so we simply avoid trying to compress *.jpg file data for now.
                return(new RawPacket(
                    reqPacket.opCode,
                    rawData, buf.limit(),
                    !esa.filePath.endsWith(".jpg")));
                }


            // getVariable().
            case RawPacket.OP_getVariable:
                {
                final String name = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readUTF();
                // Look up variable definition from name.
                final SimpleVariableDefinition def =
                    SystemVariables.nameToDef.get(name);
                // If no such variable,
                // or it is a local,
                // then immediately return an empty result.
                if((def == null) || def.isLocal())
                    {
                    return(new RawPacket(
                        reqPacket.opCode,
                        EMPTY_PAYLOAD,
                        false)); // No compression...
                    }
                // Get the variable value (if any)...
                // Note that this may include a big globalMap.
                final SimpleVariableValue svv =
                    source.getVariable(def);
                // Create and send the response packet
                // (empty for null response,
                // else serialised form, compressed when helpful).
                return(new RawPacket(
                    reqPacket.opCode,
                    svv,
                    true)); // Allow compression.
                }

            // setVariables().
            // NOTE: for security/sanity reasons,
            //     reject any attempt to set where the value does not
            //     have a credible globalMap of exactly size == 1.
            // We'll validate the entire set of values,
            // and reject the whole lot if any are dubious.
            case RawPacket.OP_setVariables:
                {
                // Decode array of variable values to apply.
                final SimpleVariableValue svvs[] = (SimpleVariableValue[])
                    reqPacket.getSerializedObjectPayload();

                // Get our local ID so that we can veto changes
                // claiming to be from us for sanity/security.
                // If we get a null (not defintion currently)
                // this will not cause an exception to be thrown,
                // we just won't be able to screen.
                final SimpleVariableValue ourID =
                    source.getVariable(SystemVariables.LOCAL_SYS_ID);
                final InstanceID iid = (ourID == null) ? null:
                    ((InstanceID) ourID.getValue());

                // Validate the requests.
                // Silently ignore:
                //   * null entries
                //   * locals
                //   * anything with a definition that we don't have
                //   * anything without a globalMap of size 1
                //   * TODO: anything claiming to be from our system ID
                //   * claims to be setting values from multiple client IDs
                //
                // We ignore invalid individual entries rather than
                // rejecting the entire request for robustness if talking to
                // a slightly different age client to eliminate any
                // variable value that has no local definition,
                // but still apply the rest of the values,
                // though we still reject obviously bogus values.
                //
                // Capture the client system ID in passing
                // (each variable value should have a globalMap with
                // exactly one entry, which is the client's ID).
                final InstanceID clientID = null;
                for(int i = svvs.length; --i >= 0; )
                    {
                    final SimpleVariableValue svv = svvs[i];

                    // Reject obviously bogus values.
                    if((svv == null) ||
                       svv.getDef().isLocal() ||
                       (svv.getGlobalMap() == null) ||
                            (svv.getGlobalMap().size() != 1))
                        { throw new IllegalArgumentException(); }

                    // Quietly ignore variables that we don't (currently)
                    // have definititions for locally.
                    if(!SystemVariables.defs.contains(svv.getDef()))
                        { continue; }

                    // Get the system that *is* in the globalMap...
                    final InstanceID putativeClientID =
                        svv.getGlobalMap().keySet().iterator().next();
                    // Abort if it is our ID!
                    if(clientID == null)
                        {
                        // OK, this must be from the first variable,
                        // so trigger some one-off processing...

                        // Reject obviously-forged / looped values.
                        if((iid != null) &&
                            iid.equals(putativeClientID))
                            { throw new IllegalArgumentException("client claiming our InstanceID!"); }

                        // Note the client's remote address
                        // (ie make an entry in an upstream global Map)...
                        final SimpleVariableValue newValue = new SimpleVariableValue(
                                                    SystemVariables.TunnelServlet_SLAVE_ADDRS,
                                                    clientAddr);
                        source.setVariable(newValue.put(putativeClientID,
                                                        newValue,
                                                        true));
                        }
                    else
                        {
                        if(!clientID.equals(putativeClientID))
                            { throw new IllegalArgumentException("client claiming multiple InstanceIDs!"); }
                        }
                    }

                // Apply the values, in the order received...
                final int nSet = source.setVariables(svvs);

                // Create and send the response packet.
                return(new RawPacket(
                    reqPacket.opCode,
                    intSer(nSet),
                    false)); // Don't try compression; none to be had...
                }

            // getVariables().
            case RawPacket.OP_getVariables:
                {
                final long changedSince = (new DataInputStream(reqPacket.getPayloadAsInputStream())).
                        readLong();
                // Get values.
                final List<SimpleVariableValue> l = new ArrayList<SimpleVariableValue>(Arrays.asList(
                    source.getVariables(changedSince)));
                // Return only non-local variables.
                for(final Iterator<SimpleVariableValue> it = l.iterator(); it.hasNext(); )
                    {
                    final SimpleVariableValue svv =
                            it.next();
                    if(svv.getDef().isLocal())
                        { it.remove(); }
                    }
                // Sort by variable name for better compression on the wire.
                Collections.sort(l, SimpleVariableValue.compByDef);
                // Convert to array.
                final SimpleVariableValue svvs[] =
                    new SimpleVariableValue[l.size()];
                l.toArray(svvs);
                return(new RawPacket(
                    reqPacket.opCode,
                    svvs,
                    true)); // Compression may be very effective.
                }

            // syncVariables().
            case RawPacket.OP_syncVariables:
                {
                // Now args, no return value.
                // The tunnel call is only made when "forced".
                source.syncVariables(true);
                return(new RawPacket(
                    reqPacket.opCode,
                    EMPTY_PAYLOAD,
                    false)); // Compression is not possible.
                }

            // getEventValues().
            // Request is:
            //   * A byte consisting of the ordinal of period/interval enum val.
            //   * The UTF-8 representation of the name of the definition.
            //   * The interval number (8 bytes).
            //   * The serialised form of the request BitSet.
            // Result is:
            //   * The in-order array of EventVariableValues (never null).
            case RawPacket.OP_getEventValues:
                {
                final DataInputStream dis = new DataInputStream(reqPacket.getPayloadAsInputStream());
                final int ordinal = dis.readUnsignedByte();
                final EventPeriod epv[] = EventPeriod.values();
                if(ordinal >= epv.length)
                    { throw new IOException("corrupt arguments: bad period ordinal"); }
                final EventPeriod intervalSelector = epv[ordinal];
                final String name = dis.readUTF();
                final SimpleVariableDefinition def = SystemVariables.nameToDef.get(name);
                // Deal kindly with possibly-new definition
                // (ie if there are newer clients than us)
                // by just returning an empty result.
                if(def == null)
                    {
                    return(new RawPacket(
                        reqPacket.opCode,
                        NO_EVENT_VALUES,
                        true)); // Compression may be effective.
                    }
                if(!def.isEvent() || def.isLocal())
                    { throw new IOException("corrupt arguments: bad definition type: "+def); }
                final long intervalNumber = dis.readLong();
                if(intervalNumber < 0)
                    { throw new IOException("corrupt arguments: bad intervalNumber"); }

                // If there is trailing data then it is the serialised BitSet
                // else this is a null BitSet equivalent to just bit-0 true.
                final BitSet whichValues;
                if(dis.available() != 0)
                    {
                    final ObjectInputStream ois = new ObjectInputStream(dis);
                    final Object o; // = null;
                    try { o = ois.readObject(); }
                    catch(final ClassNotFoundException e)
                        { throw new IOException("corrupt arguments: class not found: " + e.getMessage()); }
                    if((o != null) && !(o instanceof BitSet)) // Allow null.
                        { throw new IOException("corrupt arguments: bad whichValues class on decode"); }
                    whichValues = (BitSet) o;
                    }
                else
                    { whichValues = null; }
                assert(dis.available() == 0) : "There must be no excess/trailing data in getEventValues() RPC";

                // Call upstream to actually get any values that we have...
                final EventVariableValue result[] =
                    source.getEventValues(def, intervalSelector, intervalNumber, whichValues);

                // Return the serialised result array.
                return(new RawPacket(
                    reqPacket.opCode,
                    result,
                    true)); // Compression may be very effective.
                }


            // For unrecognised requests,
            // pretend to throw a RuntimeException.
            default:
                {
//                if(isDebug) { context.log("unrecognised opcode " +reqPacket.opCode+ " in request from tunnel user " + request.getRemoteAddr()); }
                System.err.println("unrecognised opcode " +reqPacket.opCode+ " in request");
                return(new RawPacket(
                        RawPacket.OP__RUNTEX,
                        EMPTY_PAYLOAD,
                        false));
                }
            }
        }


    /**Immutable raw packet to send in either direction over a byte stream connection.
     * This consists of an op-code byte,
     * a byte-array payload,
     * and a trailer byte (different to any op-code byte).
     * <p>
     * We informally attempt to keep a decent Hamming distance
     * between different op-codes (ie more than one bit at a time
     * should have to change to mutate one into another).
     * <p>
     * The byte array can be sent compressed (using the gzip
     * algorithm), and this compression is adaptive, ie if
     * compression does not save space then it is not used.
     * The assumption if compression is requested is that it
     * is worthwhile burning lots of CPU cycles to compress
     * the data as far as possible, and so we may attempt very
     * CPU-hungry settings.  Compressed data is sent on the
     * wire with the length actually -compressedDataLength.
     */
    public static final class RawPacket implements MemoryTools.Internable
        {
        /**Operation code: NO-OP (no operation). */
        public static final byte OP_NOOP = 0;

        /**Operation code: getGenProps(). */
        public static final byte OP_getGenProps = 3;

        /**Operation code: getAllExhibitImmutableData(). */
        public static final byte OP_getAllExhibitImmutableData = 5;

        /**Operation code: getStaticAttr(). */
        public static final byte OP_getStaticAttr = 9;

        /**Operation code: getRawFile(). */
        public static final byte OP_getRawFile = 10;

        /**Operation code: getGenSecProps(). */
        public static final byte OP_getGenSecProps = 13;

        /**Operation code: getAllExhibitProperties(). */
        public static final byte OP_getAllExhibitProperties = 17;
        /**Operation code: getAllExhibitPropertiesDiff(). */
        public static final byte OP_getAllExhibitPropertiesDiff = 18;

        /**Operation code: getThumbnails(). */
        public static final byte OP_getThumbnails = 23;

        /**Operation code: getVariable(). */
        public static final byte OP_getVariable = 26;

        /**Operation code: getVariables(). */
        public static final byte OP_getVariables = 29;

        /**Operation code: setVariables(). */
        public static final byte OP_setVariables = 30;

        /**Operation code: syncVariables(). */
        public static final byte OP_syncVariables = 33;

//        /**Operation code: getEventValue(). */
//        public static final byte OP_getEventValue = 38;

        /**Operation code: getEventValues(). */
        public static final byte OP_getEventValues = 41;


        /**Reserved op-code for use in response packets to indicate that the operation was interrupted and can be retried (not an error). */
        public static final byte OP__INTEX = 119;

        /**Reserved op-code for use in response packets to indicate a PGMasterNotInServiceException. */
        public static final byte OP__PGMNISEX = 121;

        /**Reserved op-code for use in response packets to indicate a RemoteException. */
        public static final byte OP__REMEX = 125;

        /**Reserved op-code for use in response packets to indicate a RuntimeException.
         * Receipt of one of these DOES NOT indicate a link failure
         * (and thus need to back off use of the link).
         */
        public static final byte OP__RUNTEX = 126;

        /**Trailer byte; different to all op-code byte values. */
        public static final byte TRAILER = (byte) 0xda;


        /**Minimum size of payload for which we will try heaviest compression modes; strictly positive.
         * This is for modes such as BZIP2 or LZMA with heavy memory, CPU and
         * start-up costs (ie much heavier than GZIP).
         * <p>
         * Below this we assume that the absolute bandwidth savings
         * are not worth the pain.
         * <p>
         * Probably bigger than a "bulk transfer" block too,
         * so that the heavy guns are wielded only for the biggest frames.
         */
        public static final int MIN_PLLENGTH_FOR_HEAVY_COMP_ALGS = 64 * 1024;


        /**Op code; sent first; not equal to TRAILER. */
        public final byte opCode;

        /**Frame overheader (header and trailer) in bytes; strictly positive. */
        public static final int FRAME_OVERHEAD_BYTES = 6;

        /**Maximum permitted total frame length including header/trailer; strictly positive power of two minus the frame overhead.
         * Set much larger than the user-level transfer-size limit
         * to allow for some control data overhead on top of the given data,
         * and for much larger non-user-controlled items such as the AEP.
         * <p>
         * This mainly exists to prevent DoS-style out-of-memory problems
         * arrising from corrupt (huge) length values.
         */
        public static final int MAX_FRAME_SIZE = 128*1024*1024;     

        /**Maximum permitted payload length; strictly positive power of two minus the frame overhead.
         * Set much larger than the user-level transfer-size limit
         * to allow for some control data overhead on top of the given data,
         * and for much larger non-user-controlled items such as the AEP.
         * <p>
         * This mainly exists to prevent DoS-style out-of-memory problems
         * arrising from corrupt (huge) length values.
         */
        public static final int MAX_PAYLOAD_SIZE = MAX_FRAME_SIZE - FRAME_OVERHEAD_BYTES;
        /**Check that we really do allow room for the maximum user-data transfer and then some. */
        static { assert(MAX_PAYLOAD_SIZE > 1024+SimpleExhibitPipelineIF.MAX_USER_READ_SIZE); }

        /**Payload length, send second; non-negative. */
        public final int plLength;

        /**Payload data, non-null unless the data is stored compressed.
         * Private in order to keep the whole packet immutable.
         */
        private final byte payload[];

        /**Compressed payload data, null unless the data is stored compressed.
         * Private in order to keep the whole packet immutable.
         */
        private final byte payloadCompressed[];

        /**Trailer byte; not a valid op-code. */
        public static final byte trailer = TRAILER;

        /**Depends on the whole packet being identical.
         * This depends on the header and data being identical,
         * including being compressed in the same way if at all.
         */
        @Override public boolean equals(final Object obj)
            {
            if(obj == this) { return(true); }
            if(!(obj instanceof RawPacket)) { return(false); }
            final RawPacket other = (RawPacket) obj;

            if(opCode != other.opCode) { return(false); }
            if(plLength != other.plLength) { return(false); }

            // Checking the data for equality may be very slow...
            if(!Arrays.equals(payloadCompressed, other.payloadCompressed)) { return(false); }
            if(!Arrays.equals(payload, other.payload)) { return(false); }

            return(true); // Yes, identical.
            }

        /**Checks for equivalent raw packets, ignoring internal representation details.
         * Mainly this means ignoring whether the data is held compressed or not,
         * as long as the values that the user can see are identical.
         */
        public boolean equivalentTo(final RawPacket other)
            {
            if(opCode != other.opCode) { return(false); }
            if(plLength != other.plLength) { return(false); }

            // Check that the data appears identical to the user.
            return(Arrays.equals(getPayloadCopy(), other.getPayloadCopy()));
            }

        /**Hash is based on just packet type and (uncompressed) packet length. */
        @Override public int hashCode()
            { return(((opCode * 69069) + plLength)); }

        /**Construct a new raw packet with the whole data array as the payload.
         * Note that the data is <em>not</em> copied (nor modified).
         * <p>
         * This assumes that it is worth trying to compress the payload,
         * if other conditions support it.
         *
         * @param op  the op-code of the RPC; one of the OP_XXX values
         * @param data  the payload bytes; never null
         */
        public RawPacket(final byte op, final byte data[])
            { this(op, data, true); }

        /**Construct a new raw packet with the whole data array as the payload.
         * Note that the data is <em>not</em> copied (nor modified).
         *
         * @param op  the op-code of the RPC; one of the OP_XXX values
         * @param data  the payload bytes; never null
         * @param attemptCompression  if true, it is probably worth attempting
         *     to compress the payload data unless very short
         */
        public RawPacket(final byte op, final byte data[],
                         final boolean attemptCompression)
            { this(op, data, data.length, attemptCompression); }

        /**Construct a new raw packet with the initial portion of the data array as the payload.
         * Note that the data is <em>not</em> copied (nor modified),
         * unless it is compressed or the data does not occupy the whole array.
         * <p>
         * If compression is requested and is possible
         * then the compression may be performed at construction time
         * and the compressed form of the data held to save space.
         *
         * @param op  the op-code of the RPC; one of the OP_XXX values
         * @param data  the payload bytes; never null
         * @param initialPortion  the initial portion of data[] containing
         *     payload data; never negative or larger than data.length or MAX_PAYLOAD_SIZE
         * @param attemptCompression  if true, it is probably worth attempting
         *     to compress the payload data unless very short
         */
        public RawPacket(final byte op, final byte data[], final int initialPortion,
                         final boolean attemptCompression)
            {
            if((op == TRAILER) ||
               (data == null) ||
               (initialPortion > MAX_PAYLOAD_SIZE) ||
               (initialPortion > data.length) || (initialPortion < 0))
                { throw new IllegalArgumentException(); }

            // Capture the opcode and the raw payload length immediately.
            opCode = op;
            plLength = initialPortion;

            // See if the data needs compressing and can be compressed.
            // If not, then keep it uncompressed.
            //
            // If compression is requested and the payload size is significant
            // (allowing for for data overheads of the compression scheme)
            // then check out the possibility of compression,
            // else ignore, and send the data uncompressed.
            //
            // We assume that trying to deflate anything less than
            // (say) 256 bytes is a waste of effort given the
            // deflate overhead and, indeed, the packet overhead.
            if(attemptCompression &&
               (plLength >= FileTools.TYPICAL_DEFLATE_MIN_TEXT_SIZE_COMPRESSABLE))
                {
                final byte compressed[] = FileTools.compressDeflatableData(data, 0, initialPortion);
                final boolean savedSpace = (compressed.length < plLength);

                // If successful then store the compressed data.
                if(savedSpace)
                    {
                    payload = null;
                    payloadCompressed = compressed;
if(ExhibitDataTunnelSource._protocolDebug) { System.err.println("[RawPacket: compressed payload from "+plLength+" to "+(payloadCompressed.length)+" bytes.]"); }
                    return;
                    }
                }

            // Store uncompressed data.
            // Store without copying if data is entire array.
            if(data.length == plLength)
                { payload = data; }
            else
                {
                // It is inefficient to copy,
                // so we want to avoid this where possible,
                // and in debug will warn where it is happening.
if(IsDebug.isDebug) { System.err.println("[RawPacket: WARNING: copying uncompressed payload from oversize buffer.]"); Thread.dumpStack(); }
                payload = new byte[plLength];
                System.arraycopy(data, 0, payload, 0, plLength);
                }
            payloadCompressed = null;
            }

        /**Construct a new raw packet with a serialised object as the payload.
         *
         * @param op  the op-code of the RPC; one of the OP_XXX values
         * @param obj  the Serializable object to send,
         *     or null to send empty payload
         * @param attemptCompression  if true, it is probably worth attempting
         *     to compress the payload data if not very short
         */
        public RawPacket(final byte op, final Serializable obj,
                         final boolean attemptCompression)
            throws IOException
            {
            this(op, _serObjForPayload(obj), attemptCompression);
            }

        /**Serialise object for payload.
         * Returns a zero-length array if the object is null.
         */
        private static byte[] _serObjForPayload(final Serializable obj)
            throws IOException
            {
            if(obj == null)
                { return(ExhibitDataTunnelSource.EMPTY_PAYLOAD); }

            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            final ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(obj);
            oos.flush();
            final byte[] responsePayload = baos.toByteArray();
            oos.close(); // Hopefully free resources quickly.
            return(responsePayload);
            }

        /**Retrieves a single serialised Object from the payload.
         * If the payload is empty (zero-length), we return null.
         * <p>
         * For robustness, treat a class mismatch problem like a data problem.
         */
        public Object getSerializedObjectPayload()
            throws IOException
            {
            if(plLength == 0) { return(null); }

            final InputStream is = getPayloadAsInputStream();
            final ObjectInputStream ois = new ObjectInputStream(is);
            // For robustness, treat a class mismatch problem like a data problem.
            try { return(ois.readObject()); }
            catch(ClassNotFoundException e) { throw new IOException(e.getMessage()); }
            }

        /**Computes the total number of bytes that would be written by writePacket(); strictly positive.
         */
        public int getFrameLength()
            {
            final boolean isCompressed = (payload == null);
            final byte[] dataOnTheWire =
                isCompressed ? payloadCompressed : payload;
            final int realPayloadLen = dataOnTheWire.length;

            // The total length on the wire of the entire packet.
            // An excess 6 bytes allows for 1 opCode, 4 length, 1 trailer.
            return(6 + realPayloadLen);
            }

        /**Send the packet down the given OutputStream with header and trailer.
         * We flush the data after writing it.
         * <p>
         * If the payload is small enough,
         * then the entire packet will be buffered
         * and sent in one write().
         * <p>
         * We do not close() the stream.
         */
        public void writePacket(final OutputStream os)
            throws IOException
            {
if(ExhibitDataTunnelSource._protocolDebug) { System.err.println("[writePacket(): op="+opCode+" len="+plLength+".]"); }

            final boolean isCompressed = (payload == null);
            final byte[] dataOnTheWire =
                isCompressed ? payloadCompressed : payload;
            final int realPayloadLen = dataOnTheWire.length;

            // The total length on the wire of the entire packet.
            final int totalPacketLen = getFrameLength();

            // Buffer the output (with a carefully-crafted buffer size)
            // so that we can write it all in one operating-system interaction
            // if small enough, else in efficient-ish disc/network-size chunks.
            // (Note that Tomcat 4.1.31 writes chunked data to the network
            // in blocks of 2kB on Solaris 8/9,
            // but also note that the write of a large payload will cause
            // the header data to be flushed through the buffer first,
            // so anything other than quite a small buffer may be a waste.)
            final DataOutputStream dos = new DataOutputStream(
                new BufferedOutputStream(os,
                    (realPayloadLen <= CoreConsts.BULK_DATA_TRANSFER_SIZE) ?
                        totalPacketLen : 512));

            // Send the opcode.
            dos.writeByte(opCode);
            // Send the payload length (negative indicating compressed).
            dos.writeInt(isCompressed ? -realPayloadLen : realPayloadLen);
            // Send the payload.
            dos.write(dataOnTheWire, 0, realPayloadLen);
            // Write the trailer.
            dos.writeByte(trailer);
            // Flush data though any buffers and down the wire if possible.
            dos.flush();
            }

        /**Reads a packet from the input stream, blocking until done.
         * Even if the data is received compressed it is stored in
         * the packet that this returns in uncompressed form, so the
         * presence or absence of compression on the wire is invisible
         * to the recipient.
         * This also makes (re)reading the packet data cheap.
         * <p>
         * Compressed data is signalled by a negative length on the wire,
         * where this is the negation of the compressed data size.
         * <p>
         * The ZLIB deflater with maximum compression and
         * no checksum is used for any (de)compression.
         * <p>
         * To prevent unpleasantness such a OutOfMemoryError from corrupt length value,
         * we veto lengths greater than fixed for this frame type.
         */
        public static RawPacket readPacket(final InputStream is)
            throws IOException
            {
            // While not wonderfully efficient,
            // to save a great deal of fiddly coding
            // the input stream is wrapped in a DataInputStream.
            final DataInputStream dis = new DataInputStream(is);
            // Start collecting the data...
            final byte opC = dis.readByte();
            final int rawLen = dis.readInt();
            final boolean isCompressed = (rawLen < 0);
            final int len = ((isCompressed) ? -rawLen : rawLen);
if(ExhibitDataTunnelSource._protocolDebug) { System.err.println("[readPacket(): op="+opC+" rawLen="+rawLen+"...]"); }
            if(len > MAX_PAYLOAD_SIZE) { throw new IOException("corrupt length"); }
            final byte data[] = new byte[len];
            dis.readFully(data);
            final byte tr = dis.readByte();
            if(tr != TRAILER)
                { throw new IOException("corrupt trailer"); }
            try {
                // Decompress the data if necessary...
                if(isCompressed)
                    {
                    // Uncompress from the input stream...
                    final byte uncompressedData[] = FileTools.decompressDeflatedData(data);

if(ExhibitDataTunnelSource._protocolDebug) { System.err.println("[readPacket(): compressed payload len="+data.length+", uncompressed = "+uncompressedData.length+" bytes.]"); }
                    return(new RawPacket(opC, uncompressedData, false));
                    }
                return(new RawPacket(opC, data, false));
                }
            catch(final IllegalArgumentException e)
                { throw new IOException("corrupt packet: " + e.getMessage()); }
            }

        /**Create a human-readable summary (not including the payload data). */
        public String toString()
            {
            final StringBuilder sb = new StringBuilder(79);
            sb.append("RawPacket");
            sb.append(":opCode=").append((int) opCode);
            sb.append(":plLength=").append(plLength);
            if(payloadCompressed != null)
                { sb.append(":compressedLength=").append(payloadCompressed.length); }
            return(sb.toString());
            }

        /**Get the payload data as a mark()able InputStream; never null.
         * If the internal data is stored compressed however,
         * then this will return the uncompressed version.
         * <p>
         * Each stream returned is independent.
         *
         * @return  the uncompressed payload data as a stream of length plLength
         */
        public InputStream getPayloadAsInputStream()
            {
            if(payload != null)
                {
                // Payload is uncompressed; wrap it as-is.
                assert(payload.length == plLength);
                return(new ByteArrayInputStream(payload));
                }

            // Payload is held compressed, so decompress it.
            try
                {
                final byte uncompressedData[] = FileTools.decompressDeflatedData(payloadCompressed);
                assert(uncompressedData.length == plLength);
                return(new ByteArrayInputStream(uncompressedData));
                }
            catch(final IOException e)
                {
                throw new IllegalStateException("corrupt compressed data");
                }
            }

        /**Get a copy of the payload data; never null.
         * We return a copy to keep this object immutable.
         * <p>
         * If the internal data is stored compressed
         * then this will return the original uncompressed version.
         *
         * @return  a copy of the payload data, of length plLength
         */
        public byte[] getPayloadCopy()
            {
            if(payload != null)
                {
                // Payload is uncompressed; clone it as-is.
                assert(payload.length == plLength);
                // Take copy to protect internal state.
                return(payload.clone());
                }

            // Payload is held compressed, so decompress it.
            try
                {
                final byte uncompressedData[] = FileTools.decompressDeflatedData(payloadCompressed);
                assert(uncompressedData.length == plLength);
                // No need to copy (private) data returned from decompression.
                return(uncompressedData);
                }
            catch(final IOException e)
                {
                throw new IllegalStateException("corrupt compressed data");
                }
            }

        /**Copy the payload data into the supplied (trusted) buffer.
         * We only make this package-visible to keep this object immutable.
         * <p>
         * If the internal data is stored compressed
         * then this will return the original uncompressed version.
         */
        void getPayloadCopy(final ByteBuffer buf)
            {
            if(payload != null)
                {
                // Payload is uncompressed; clone it as-is.
                assert(payload.length == plLength);
                // Copy uncompressed data directly.
                buf.put(payload);
                return;
                }

            // Payload is held compressed, so decompress it.
            try
                {
                final byte uncompressedData[] = FileTools.decompressDeflatedData(payloadCompressed);
                assert(uncompressedData.length == plLength);
                // No need to copy (private) data returned from decompression.
                buf.put(uncompressedData);
                return;
                }
            catch(final IOException e)
                {
                throw new IllegalStateException("corrupt compressed data");
                }
            }
        }
    
    /**The immutable adjunct for a RawPacket that includes the HMAC and other anti-attack data.
     * Note that this class is NOT directly serialisable,
     * with the data fields sent in some other way,
     * eg as HTTP header fields "out-of-band" from the actual HTTP message.
     */
    public static final class PacketProtector
        {
        /**The timestamp for the RawPacket, input to each MAC; strictly positive. */
        public final long timestamp;

        /**The length of the entire frame/datastream being protected, input to each MAC; non-negative. */
        public final int length;

        /**The immutable in-order list of MAC authenticator segments for the stream and fields herein; never null nor empty nor containing nulls.
         * This sequence of HMAC values together makes up the message MAC.
         * It is designed not to be possible to re-arrange these segments,
         * as each depends in part on the value of its predecessor
         * as well as the data in its segment.
         * <p>
         * Note that the <em>last</em> MAC, which depends on all the data being protected,
         * can be used as a unique message MAC on its own.
         */
        public final List<ROByteArray> mac;

        /**Maxium size in (ASCII) characters of output of toCheckString(); strictly positive.
         * Chosen to allow inclusion of the output of toCheckString() in an HTTP header.
         */
        public static final int MAX_CHECK_STRING_CHARS = 900;

        /**Generate (HTTP-header) check-string; never null nor empty.
         * This is a pure-ASCII single-line control-code-free check String
         * that contains the input fields and MAC from this object
         * to serve as the "checksum" of a RawPacket.
         * <p>
         * This object is suitable (short enough, avoidance of meta-characters)
         * to be used directly in an HTTP header.
         * <p>
         * This value is suitable to be decoded by fromCheckString().
         * <p>
         * The format is a space-separated list of fields:
         * <ol>
         * <li>The Java-style UTC timestamp in milliseconds as an unsigned decimal.
         * <li>The length of the protected stream in bytes as an unsigned decimal.
         * <li>The MAC values for each segment, in order, encoded Base64.
         * </ol>
         */
        public String toCheckString()
            {
            // Do a worst-case check given the size of the (first) MAC as encoded.
            assert((""+Long.MAX_VALUE+' '+Integer.MAX_VALUE+' ').length() + (MAX_SEGMENTS*(1+TextUtils.encode8To6(mac.get(0).getData()).length()))-1 < MAX_CHECK_STRING_CHARS) : "check string must always be short enough for HTTP header";

            final StringBuilder sb = new StringBuilder(32 + (45*mac.size()));
            sb.append(timestamp).append(' ');
            sb.append(length).append(' ');
            for(final ROByteArray r : mac)
                { sb.append(TextUtils.encode8To6(r.getData())).append(' '); }
            // Rip off trailing space...
            sb.setLength(sb.length()-1);
            assert(sb.length() <= MAX_CHECK_STRING_CHARS) : "check string must be short enough for HTTP header";
            return(sb.toString());
            }

        /**Field splitter regex pattern compiled once for efficiency; never null. */
        private static final Pattern fieldSplitPattern = Pattern.compile(" ");

        /**Parses a check-string as generated by toCheckString(); never null.
         * This is tolerant of some leading and trailing whitespace.
         * 
         * @throws IllegalArgumentException  if the input is unparsable
         */
        public static PacketProtector fromCheckString(final String check)
            {
            if((check == null) || (check.length() > 2*MAX_CHECK_STRING_CHARS))
                { throw new IllegalArgumentException("too big before trim"); }
            final String trimmed = check.trim();
            if(trimmed.length() > MAX_CHECK_STRING_CHARS)
                { throw new IllegalArgumentException("too big"); }
            final String fields[] = fieldSplitPattern.split(check, 0);
            if(fields.length < 3)
                { throw new IllegalArgumentException("not enough fields"); }
            final long timestamp = Long.parseLong(fields[0], 10);
            final int length = Integer.parseInt(fields[1], 10);
            if((length < 0) || (length > RawPacket.MAX_FRAME_SIZE))
                { throw new IllegalArgumentException("out-of-range length value"); }
            final ArrayList<ROByteArray> mac = new ArrayList<ROByteArray>(fields.length-2);
            for(int i = 2; i < fields.length; ++i)
                { mac.add(new ROByteArray(TextUtils.decode8To6(fields[i]))); }
            return(new PacketProtector(timestamp, length, mac));
            }

        /**Maximum number of segments protected stream may be broken into; strictly positive power of two.
         * This determines the maximum number of incremental portions
         * we can process while streaming a RawPacket.
         * <p>
         * This value is capped to limit the amount of MAC data that needs to be sent,
         * eg in an HTTP header of limited length.
         */
        public static final int MAX_SEGMENTS = 16;

        /**Minimum segment size (other than final segment); strictly positive power of two.
         * This is set to be somewhat larger than the largest typical frame
         * (ie the entire header, payload and trailer)
         * that we usually don't want to stream, eg bulk data transfer,
         * so that such frames/packets will get a single MAC for efficiency.
         * This is also set large enough to amortise the cost of each segment MAC.
         * <p>
         * This is small enough to represent a reasonable incremental amount of CPU
         * for streamed inputs.
         */
        public static final int MIN_SEGMENT_SIZE = (1<<16);

        /**Maximum segment size; strictly positive power of two.
         * This is determined from the maximum frame size
         * and the maximum number of segments.
         */
        public static final int MAX_SEGMENT_SIZE = RawPacket.MAX_FRAME_SIZE / MAX_SEGMENTS;
        /**Check some invariants. */
        static { assert(RawPacket.MAX_FRAME_SIZE == MAX_SEGMENT_SIZE * MAX_SEGMENTS); }
        static { assert(MAX_SEGMENT_SIZE > MIN_SEGMENT_SIZE); }

        /**Create an adjunct to protect a RawPacket, including a current timestamp.
         * This must be supplied with the (secret) Key shared between the servers.
         *
         * @param raw  the RawPacket to protect; never null
         * @param key  the (secret) key for the HMAC; never null
         */
        public PacketProtector(final RawPacket raw,
                               final SecretKey key)
            throws InvalidKeyException
            { this(raw, System.currentTimeMillis(), key); }

        /**Create an adjunct to protect a RawPacket, including the given timestamp.
         * This must be supplied with the (secret) Key shared between the servers.
         * 
         * @param raw  the RawPacket to protect; never null
         * @param timestamp  the timestamp for creation/send of the RawPacket; strictly positive
         * @param key  the (secret) key for the HMAC; never null
         */
        public PacketProtector(final RawPacket raw,
                               final long timestamp,
                               final SecretKey key)
            throws InvalidKeyException
            { this(timestamp, raw.getFrameLength(), computeMAC(timestamp, raw, key)); }

        /**Create an adjunct to protect a RawPacket.
         *
         * @param timestamp  the timestamp for creation/send of the RawPacket; strictly positive
         * @param mac  the HMAC for the RawPacket and other fields, not checked; never null
         */
        public PacketProtector(final long timestamp,
                               final int length,
                               final List<ROByteArray> mac)
            {
            this.timestamp = timestamp;
            this.length = length;
            // Take defensive immutable copy of MAC segment list.
            this.mac = Collections.unmodifiableList(new ArrayList<ROByteArray>(mac));

            try { validateObject(); } // Validate this instance.
            catch(final InvalidObjectException e) { throw new IllegalArgumentException(e); }
            }
        
        /**Compute segment size; strictly positive power of two.
         * Given the full message/frame length
         * this computes the segment size between/at the MIN/MAX limits
         * to maximise the number of segments
         * and thus maximise the the amount of incremental MAC processing
         * and incremental processing of streamed frames that can be done,
         * minimising buffer working space and time/bandwidth before corruption is found.
         * <p>
         * All segments are of the same size, except the last one which may be shorter.
         */
        private static final int computeSegmentSize(final int frameLength)
            {
            if((frameLength > RawPacket.MAX_FRAME_SIZE) ||
               (frameLength < 0))
                { throw new IllegalArgumentException(); }

            // Find minimum segment size that will cover the frame.
            for(int segSize = MIN_SEGMENT_SIZE; segSize <= MAX_SEGMENT_SIZE; segSize <<= 1)
                {
                if(segSize * MAX_SEGMENTS >= frameLength)
                    { return(segSize); }
                }

            throw new Error("should not get here");
            }

        /**Compute the MAC given the message and other fields to be included; never null.
         * The (secret) Key must also be supplied
         * and be suitable for the HMAC algorithm used.
         * <p>
         * The MAC is computed on a series of segments of the input of the same length
         * (except for a possibly-shorter final segment)
         * producing an HMAC on each segment.
         * <p>
         * Each HMAC is computed over a binary message consisting of:
         * <ol>
         * <li>the 4-byte length of the entire protected stream/frame.
         * <li>the 8-byte (Java-style millisecond UTC) timestamp in network order,
         * <li>for all segments other than the first, the HMAC of the previous segment,
         * <li>the full on-the-wire form of the RawPacket (including header/trailer).
         * </ol>
         * <p>
         * The chaining should make it impossible to reorder the segments.
         * <p>
         * The segmentation is so that the data in each segment is known-safe
         * and can be safely consumed by incremental/streaming CPU-heavy operations
         * before subsequent segments have been received and decoded.
         * This segmentation also means that we can abort a damaged message
         * as soon as we check the damaged segment: we do not need to wait
         * to receive and store and check the whole message.
         * 
         * @param raw  the RawPacket to protect; never null
         * @param timestamp  the timestamp for creation/send of the RawPacket; strictly positive
         * @param key  the (secret) key for the HMAC; never null
         *
         * @throws InvalidKeyException  if the Key supplied is inappropriate
         *
         * @throws IllegalArgumentException  if the timestamp is non-positive
         *     or any other argument is null
         * @throws IllegalStateException  if the HMAC algorithm is unavailable
         *
         * @return immutable in-order list of HMACs to protect each stream segment
         */
        public static List<ROByteArray> computeMAC(final long timestamp,
                                                   final RawPacket raw,
                                                   final SecretKey key)
            throws InvalidKeyException
            {
            if((timestamp <= 0) || (raw == null) || (key == null))
                { throw new IllegalArgumentException(); }

            // Compute the segment size that we will be using.
            final int frameLength = raw.getFrameLength();
            final int segmentSize = computeSegmentSize(frameLength);

            // Compute the number of segments that we will be using.
            final int nSegs = (frameLength + segmentSize - 1) / segmentSize;

            // Compose our result MAC list...
            final List<ROByteArray> mac = new ArrayList<ROByteArray>(nSegs);

            // Incrementally generate the MAC block by block.
            // We use an output stream sink to avoid having to copy large frames.
            // This OutputStream is not thread-safe and does not need to be.
            final OutputStream sink = new OutputStream(){
                /**We collect output in this segment-size buffer; never null.
                 * We accumulate output in this to create the next MAC value.
                 */
                private final byte[] segBuf = new byte[segmentSize];

                /**The previous MAC value, or null while gathering the first segment. */
                private byte[] prevMAC;

                /**Count of bytes seen so far; non-negative.
                 * Implicitly this also indicates how far into the current segment we are.
                 */
                private int bytesSeenSoFar;

                /**Process the next block to generate the next MAC.
                 * Must only been called when the sugBuf is full or at EOF/close().
                 */
                private void processBlock()
                    {                   
                    try
                        {
                        final Mac m = Mac.getInstance(CoreConsts.HMAC_ALG);
                        m.init(key);

                        // First, write 4-byte length of entire message.
                        m.update(intSer(frameLength));
                        // Second, write 8-byte timestamp.
                        m.update(longSer(timestamp));
                        // If there is a previous MAC, stir it in.
                        if(null != prevMAC) { m.update(prevMAC); }
                        // Finally, write frame bytes to protect.
                        final int bufLen = bytesSeenSoFar % segmentSize;
                        m.update(segBuf, 0, ((bufLen == 0) ? segmentSize : bufLen));

                        // Compute and record the MAC.
                        final byte[] hmac = m.doFinal();
                        prevMAC = hmac;
                        mac.add(new ROByteArray(hmac));
                        }
                    catch(final Exception e)
                        {
                        e.printStackTrace();
                        throw new Error("internal error", e);
                        }
                    }

                @Override
                public void write(final byte[] b, int off, int len) throws IOException
                    {
                    // Loop to generate multiple segment MACs if needed.
                    while(len > 0)
                        {
                        // Compute bytes needed to fill the buffer: always positive.
                        final int offsetInSegBuf = bytesSeenSoFar % segmentSize;
                        final int needed = segmentSize - (offsetInSegBuf);
                        // How many will we copy in...
                        final int toCopy = Math.min(needed, len);
                        // Do the copy...
                        System.arraycopy(b, off, segBuf, offsetInSegBuf, toCopy);
                        // Adjust all our indexes...
                        bytesSeenSoFar += toCopy;
                        off += toCopy;
                        len -= toCopy;

                        // If we copied exactly up to the end of the segBuf
                        // then it is time to generate the next MAC...
                        if(toCopy == needed) { processBlock(); }
                        }
                    }

                /**Horribly inefficient, but correct, and should never actually get called. */
                @Override public void write(int b) throws IOException
                    {
                    final byte buf[] = new byte[1];
                    buf[0] = (byte) b;
                    write(buf, 0, 1);
                    }

                /**Handle the tail of the input. */
                @Override public void close() throws IOException
                    {
                    assert(bytesSeenSoFar == frameLength); // Should have seen all the data.
                    processBlock();
                    }
                };
            // Compute the HMAC value(s).
            try { raw.writePacket(sink); sink.close(); }
            catch(final IOException e) { throw new Error("internal error", e); }
//System.out.println("frameLength/nSegs/mac.size() = " + frameLength + '/' + nSegs + '/' + mac.size());
            assert(nSegs == mac.size()); // Did we create the correct number of MACs values?

            return(Collections.unmodifiableList(mac));
            }
        
        /**Protect an input stream with our MAC; aborts with IOException in case of corruption.
         * This acts as a transparent pass-through filter
         * that will read a MAC-protected-segment at a time from the underlying stream
         * and either pass it through having verified it,
         * or abort with an IOException if the segment is corrupt.
         * <p>
         * This closes its input stream and vetoes any further operations
         * once any error has been encountered.
         * <p>
         * We use Key rather than SecretKey,
         * since the latter depends on Java extensions that may not be available,
         * eg when run in JWS.
         */
        public InputStream protectInputStream(final Key key, final InputStream is)
            {
            if(key == null) { throw new IllegalArgumentException(); }
            if(is == null) { throw new IllegalArgumentException(); }

            // Compute our segment size for buffering.
            final int segmentSize = computeSegmentSize(length);
            
            // Number of segments.
            final int nSegs = mac.size();

            // To avoid any confusion, we do not allow mark/reset.
            return(new FilterInputStream(is){
                /**Set true upon any error; subsequently we will not allow any further read()s. */
                private boolean hasErred;

                /**Our read buffer for one segment's data; never null. */
                private final byte[] segBuf = new byte[segmentSize];

                /**Data remaining in segBuf; non-negative. */
                private int bufLen;

                /**Start of remaining data in segBuf; non-negative. */
                private int bufPos;

                /**Which MAC segment is next to be loaded/verified. */
                private int nextSegNumber;

                /**Whatever is in out buffer plus from upstream is available without blocking. */
                @Override public synchronized int available() throws IOException
                    { return(in.available() + bufLen); }

                @Override public void mark(int readlimit)
                    { throw new UnsupportedOperationException("mark not supported"); }

                /**Mark/reset is not allowed to avoid confusion/complexity. */
                @Override public boolean markSupported() { return(false); }

                /**Private buffer for read(); never null. */
                private final byte[] buf1 = new byte[1];

                /**Read a single byte: we try to make this moderately efficient. */
                @Override public synchronized int read() throws IOException
                    {
                    if(hasErred) { throw new IOException("corrupt stream"); }

                    // Fast path to return data already in our buffer.
                    if(bufLen > 0)
                        {
                        --bufLen;
                        return(segBuf[bufPos++] & 0xff);
                        }

                    // Use full read() to do all the leg-work.
                    final int n = read(buf1, 0, 1);
                    if(n < 1) { return(-1); /* EOF */ }
                    assert(n == 1);
                    return(buf1[0] & 0xff);
                    }

                /**Bulk data read.
                 * This routine returns any data it has buffered (ie already verified)
                 * else it reads a segment of data from the underlying stream
                 * (possibly less than a full segment at EOF, but verifying the length)
                 * and validates it with the appropriate MAC.
                 * <p>
                 * If validation against the MAC fails
                 * then this and all subsequent calls to read()
                 * are vetoed with an IOException.
                 * <p>
                 * If validation succeeds then the requested data is returned.
                 * <p>
                 * We do not return data in one read across segment boundaries,
                 * as we are not required to and it would increase complexity.
                 */
                @Override
                public synchronized int read(byte[] b, int off, int len) throws IOException
                    {
                    if((b == null) ||
                       (off < 0) || (off >= b.length) ||
                       (len < 0) || (off + len > b.length))
                        { throw new IllegalArgumentException(); }

                    if(hasErred) { throw new IOException("corrupt stream"); }

                    if(len == 0) { return(0); }

                    // If we have no verified data left in our buffer
                    // then unless at EOF we need to load another segment and verify it.
                    if(bufLen == 0)
                        {
                        // We are at EOF if we have read the last segment.
                        if(nextSegNumber >= nSegs) { return(-1); }

                        // Allow for the final segment to be smaller.
                        final int readSize = (nextSegNumber < nSegs-1) ? segmentSize :
                            (length - (segmentSize * (nSegs-1)));
                        bufPos = 0; // Next caller will read from offset zero.
                        bufLen = 0; // Buffer is currently empty...
                        while(bufLen != readSize)
                            {
                            try
                                {
                                final int n = in.read(segBuf, bufLen, readSize - bufLen);
                                if(n < 1) { throw new IOException("premature EOF"); }
                                bufLen += n;
                                }
                            catch(final IOException e) { hasErred = true; throw e; }
                            }

                        // Verify the data.
                        try
                            {
                            final Mac m = Mac.getInstance(CoreConsts.HMAC_ALG);
                            m.init(key);

                            // First, write 4-byte length of entire message.
                            m.update(intSer(length));
                            // Second, write 8-byte timestamp.
                            m.update(longSer(timestamp));
                            // If there is a previous MAC, stir it in.
                            if(nextSegNumber > 0) { m.update(mac.get(nextSegNumber-1).getData()); }
                            // Finally, use the bytes read from upstream.
                            m.update(segBuf, 0, bufLen);

                            // Compute and record the MAC.
                            final byte[] hmac = m.doFinal();
                            if(!Arrays.equals(mac.get(nextSegNumber++).getData(), hmac))
                                { throw new IOException("corrupt input data"); }
                            }
                        catch(final IOException e)
                            {
                            hasErred = true; // Prevent further read()s from this stream.
                            throw e;
                            }
                        catch(final Exception e)
                            {
                            hasErred = true;
                            e.printStackTrace();
                            throw new Error("internal error", e);
                            }
                        }

                    // The buffer must now have something in it.
                    assert(bufLen > 0);
                    // Return verified data from our buffer to the caller.
                    final int toCopy = Math.min(bufLen, len);
                    System.arraycopy(segBuf, bufPos, b, off, toCopy);
                    bufLen -= toCopy;
                    bufPos += toCopy;
                    return(toCopy);
                    }

                /**Implemented as read(b, 0, b.length). */
                @Override public int read(byte[] b) throws IOException
                    { return(read(b, 0, b.length)); }

                @Override public void reset() throws IOException
                    { throw new IOException("reset not allowed"); }

                /**We have to read the input to skip it. */
                @Override public synchronized long skip(long n) throws IOException
                    {
                    long skipped = 0;
                    final byte b[] = new byte[(int) Math.min(1024, n)];
                    while(n > 0)
                        {
                        final int r = read(b, 0, (int) Math.min(n, b.length));
                        if(r < 1) { break; /* EOF */ }
                        n -= r;
                        skipped += r;
                        }
                    return(skipped);
                    }
                });
            }

        /**Checks that the object content is valid only.
         * This does NOT attempt to check the MAC matches the message and fields;
         * that must be done explicitly elsewhere.
         * <p>
         * Partly this does not check the MAC because it does not have access to the key.
         */
        public void validateObject() throws InvalidObjectException
            {
            if(timestamp <= 0) { throw new InvalidObjectException("bad object: bad timestamp"); }
            if(length < 0) { throw new InvalidObjectException("bad object: bad length"); }
            if(mac == null) { throw new InvalidObjectException("bad object: null MAC"); }
            if(mac.size() == 0) { throw new InvalidObjectException("bad object: empty MAC"); }
            for(final ROByteArray m : mac)
                {
                if(m == null) { throw new InvalidObjectException("bad object: null segement MAC"); }
                if(m.length() == 0) { throw new InvalidObjectException("bad object: empty segment MAC"); }                
                }
            }

        /**Equality depends on all the members being equal. */
        @Override
        public boolean equals(final Object obj)
            {
            if(this == obj) { return(true); }
            if(!(obj instanceof PacketProtector)) { return(false); }
            final PacketProtector other = (PacketProtector) obj;
            return((timestamp == other.timestamp) &&
                   (length == other.length) &&
                   mac.equals(other.mac));
            }

        /**We use the timestamp and length fields in the hash for the entire collection. */
        @Override public int hashCode() { return(length ^ (int) timestamp); }
        }
    
    
    /**Structure with possibly-fixed size to check for duplicates within a time window.
     * This grows as needed to accommodate everything up to a specific age,
     * and may be able to release resources when fewer items need to be stored.
     * <p>
     * This provides optional fixed lower and upper sizes.
     * Attempts to insert more items than the ceiling capacity
     * will result in old items being removed in LRU order regardless of age.
     * If the map is at or below the minimum size then entries never expire.
     * <p>
     * Thread-safe.
     * <p>
     * A lock can be held on instances of this object to make compound operations atomic.
     * <p>
     * Does not support the full Map interface.
     */
    public static final class DuplicateIDChecker<K>
        {
        /**Default load factor. */
        public static final float DEFAULT_LOAD_FACTOR = 0.7f;

        /**Default minimum size/capacity, also initial capacity; strictly positive. */
        public static final int DEFAULT_MIN_SIZE = 113;

        /**Underlying LinkedHashMap on which this is based; never null. */
        private final LinkedHashMap<K,Long> lhm;

        /**Maximum age that we will retain IDs for (ms); strictly positive. */
        private final int minRetentionMs;
        
        /**The minimum size/capacity; strictly positive. */
        private final int minSize;

        /**Create an instance with all defaults.
         * @param minRetentionMs  the minimum time that we will keep an ID for (ms);
         *    strictly positive
         */
        public DuplicateIDChecker(final int minRetentionMs)
            { this(minRetentionMs, DEFAULT_MIN_SIZE, Integer.MAX_VALUE, DEFAULT_LOAD_FACTOR); }

        /**Create an instance with the given parameters.
         * The standard default is used for the load factor,
         * and the map size is effectively unlimited.
         *
         * @param minRetentionMs  the minimum time that we will keep an ID for (ms);
         *    strictly positive
         * @param minSize  the minimum (and initial) capacity;
         *     strictly positive
         */
        public DuplicateIDChecker(final int minRetentionMs, final int minSize)
            { this(minRetentionMs, minSize, Integer.MAX_VALUE, DEFAULT_LOAD_FACTOR); }

        /**Create an instance with the given parameters.
         * @param minRetentionMs  the minimum time that we will keep an ID for (ms);
         *    strictly positive
         * @param minSize  the minimum (and initial) capacity;
         *     strictly positive
         * @param maxCapacity  the maximum capacity;
         *     greater than minCapacity (and this strictly positive)
         * @param loadFactor  the load factor of the underlying hash table;
         *     in the range 0.0f to 1.0f exclusive (typically ~0.7f)
         */
        public DuplicateIDChecker(final int minRetentionMs,
                                  final int minSize,
                                  final int maxCapacity,
                                  final float loadFactor)
            {
            if((minSize <= 0) || (maxCapacity <= minSize) ||
               (minRetentionMs <= 0) ||
               (loadFactor <= 0) || (loadFactor >= 1))
                { throw new IllegalArgumentException(); }

            this.minSize = minSize;
            this.minRetentionMs = minRetentionMs;

            // Build access-order map with override for removal of oldest entry when map too big
            // or when map is over our minimum size and the oldest entry has expired.
            lhm = new LinkedHashMap<K,Long>(minSize, loadFactor, true){
                protected final boolean removeEldestEntry(final Map.Entry<K, Long> entry)
                    {
                    final int s = size();

                    // If over-size then zap the eldest entry.
                    if(s > maxCapacity) { return(true); }

                    // If under-size then definitely don't zap anything.
                    if(s <= minSize) { return(false); }

                    // Zap an entry too old to matter.
                    return(hasExpired(minRetentionMs, entry));
                    }
                /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
                private static final long serialVersionUID = 3013203108802112945L;
                };
            }

        /**Returns true if we remember this ID, false otherwise. */
        public synchronized boolean isDup(final K key)
            {
            final Long v = lhm.get(key);
            final boolean result = ((v != null) &&
                (v.longValue() >= System.currentTimeMillis() - minRetentionMs));
            incExpire();
            return(result);
            }

        /**Remove the oldest entries if expired (and the map is above minimum size).
         * By calling this from out accessor/mutator methods
         * the map can gradually shrink (thus freeing resources)
         * when larger than the minimum size,
         * even while new items are being added.
         */
        private void incExpire()
            {
            // Incrementally delete a few old expired items here to free up resources.
            // We examine more than one so that the map can still shrink
            // if this routine is called just once before each put().
            for(int i = 3; --i >= 0; )
                {
                if(lhm.size() <= minSize) { return; }
                // Discard the oldest entry so as to trim the map size.
                final Iterator<Map.Entry<K,Long>> it = lhm.entrySet().iterator();
                if(hasExpired(minRetentionMs, it.next())) { it.remove(); }
                }
            }

        /**Record an ID, making it the Most Recently Used, returning the insertion time of the previous insertion of this ID if any. */
        public synchronized Long add(final K key) { incExpire(); return(lhm.put(key, new Long(System.currentTimeMillis()))); }

        /**Remove an ID and return the insertion time of the removed ID, if any, else null. */
        public synchronized Long remove(final K key) { return(lhm.remove(key)); }

        /**Clear the structure. */
        public synchronized void clear() { lhm.clear(); }

        /**Return the number of IDs held; non-negative. */
        public synchronized int size() { return(lhm.size()); }

        /**Provide a human-readable summary of status.
         * Useful for debugging/tuning, for example.
         */
        public String toString()
            { return("[DuplicateIDChecker: size = "+lhm.size()+".]"); }

        /**Returns true when an entry has expired and can be removed.
         * @param minRetentionMs  minimum time (ms) records are retained for; strictly positive
         * @param entry  full entry including the timestamp; never null
         */
        private boolean hasExpired(final int minRetentionMs, final Map.Entry<K, Long> entry)
            { return(entry.getValue().longValue() < System.currentTimeMillis() - minRetentionMs); }
        }
    }
