/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.datasource;

import org.hd.d.pg2k.svrCore.AllExhibitImmutableData;
import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitStaticAttr;
import org.hd.d.pg2k.svrCore.ExhibitThumbnails;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.EventPeriod;
import org.hd.d.pg2k.svrCore.vars.EventVariableValue;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableDefinition;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.BitSet;

/**Simple extendable base class which acts as a pipeline filter.
 * Analagous to a FilterInputStream, a SimpleExhibitPipelineFilter
 * wraps another data source/stage, which it uses as its basic source of data,
 * possibly transforming the data along the way or providing additional
 * functionality.
 * <p>
 * The class SimpleExhibitPipelineFilter itself simply implements all methods of
 * SimpleExhibitPipelineIF with versions that pass all requests to the wrapped
 * source.  Subclasses of SimpleExhibitPipelineFilter may further override
 * some of these methods and may also provide additional methods and fields.
 * <p>
 * A bare SimpleExhibitPipelineFilter is like a null/identity stage
 * which leaves the semantics of a data pipeline intact.
 * <p>
 * One possible use of this is to build a "Mock Object" for testing purposes.
 */
public class SimpleExhibitPipelineFilter implements SimpleExhibitPipelineIF
    {
    /**Construct a new instance of this filter.
     * @param source  the upstream data source; never null
     */
    public SimpleExhibitPipelineFilter(final SimpleExhibitPipelineIF source)
        {
        if(source == null)
            { throw new IllegalArgumentException(); }
        this.source = source;
        }

    /**The upstream source; never null. */
    protected final SimpleExhibitPipelineIF source;

    /**Get the static attributes for a given exhibit.
     * Returns null if the named exhibit does not exist.
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public ExhibitStaticAttr getStaticAttr(String name)
        throws IOException
        { return(source.getStaticAttr(name)); }

    /**Read a chunk of the raw exhibit binary into the given buffer.
     * The start index and the index after the last required
     * byte is supplied.  The start value must be non-negative
     * and the afterEnd value no smaller than start and no larger
     * than the exhibit data length.
     * <p>
     * This always reads all the bytes requested or throws an IOException.
     *
     * @param dontCache  if true, this is a hint not to attempt to cache this
     *     or displace anything from extant caches for this data
     *     as it maqy for example be precaching or random activity;
     *     by default callers should leave this false to allow cacheing
     *
     * @throws IllegalArgumentException  for invalid arguments such as
     *     a null or invalid name,
     *     a negative offset or start,
     *     a negative (or in some cases, non-positive) len
     * @throws IOException for requests that cannot be fulfilled because of
     *     I/O restrictions or problems, such as link failure or
     *     an upper bound on the length of a request
     */
    public void getRawFile(ByteBuffer buf,
                           String exhibitName, int position, boolean dontCache)
        throws IOException
        { source.getRawFile(buf, exhibitName, position, dontCache); }

    /**Gets set of all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public AllExhibitImmutableData getAllExhibitImmutableData(long oldStamp)
        throws IOException
        { return(source.getAllExhibitImmutableData(oldStamp)); }

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned) null is returned.
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public AllExhibitProperties getAllExhibitProperties(long oldHash)
        throws IOException
        { return(source.getAllExhibitProperties(oldHash)); }

    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public GenProps getGenProps(long oldStamp)
        throws IOException
        { return(source.getGenProps(oldStamp)); }

    /**Gets the generic security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public java.util.Properties getGenSecProps(long oldStamp)
        throws IOException
        { return(source.getGenSecProps(oldStamp)); }

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible, else if create is false
     *     only return an existing one and return null if none is to hand
     *
     * @throws IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public ExhibitThumbnails getThumbnails(String name, boolean create)
        throws IOException
        { return(source.getThumbnails(name, create)); }

    /**Poll periodically (of the order of seconds) to do bg work.
     * The current generic system properties are passed in...
     * <p>
     * @throws IOException in case of difficulty, but even if a sub-ordinate
     *     call throws IOException the poll() call should make some effort
     *     to do as much of its work as reasonably possible
     */
    public void poll(GenProps gp)
        throws IOException
        { source.poll(gp); }

    /**Synchronise variables with upstream values.
     * Pushes updated values upstream to the source,
     * calls sync on the source if called with the "force" argument true,
     * and then retrieves changed values from upstream.
     * <p>
     * When called with force==true, this acts like a full "memory barrier",
     * flushing all write-cached items downstream immediately and afterwards
     * getting the value of all upstream values with getVariables(-1),
     * but may be expensive in terms of CPU or bandwidth, so use sparingly.
     * <p>
     * When called with force=false, this incrementally flushes outstanding
     * writes and will then fetch all, or only new, values from upstream,
     * so is potentally much less resource-intensive.
     * In particular, this does not propagate the sync() upstream.
     * <p>
     * In any case, it is rarely the right thing for a casual user
     * to vall this as it may be very expensive.
     *
     * @param force  if true, this will force a full write flush,
     *     a full sync upstream,
     *     then full read with getVariables(-1),
     *     to get the effect of a full "barrier";
     *     otherwise, in general, a more incremental and non-propagating
     *     mode is used which still does a write flush but may chose
     *     to do a partial read of "new" upstream values
     *
     * @throws IOException if one is received from upstream
     */
    public void syncVariables(boolean force)
        throws IOException
        {
        source.syncVariables(force);
        }

    /**Get set of variable values altered on or after specified time, or get all values with -1; never null.
     * This may be slow if there are many live variables.
     *
     * @throws IOException  in case of I/O difficulty
     */
    public SimpleVariableValue[] getVariables(long changedSince)
        throws IOException
        {
        return(source.getVariables(changedSince));
        }

    /**Get a single variable value; returns null if no such value or wrong type.
     * Implementations should, where possible,
     * trim from the globalMap of returned values any apparently-stale data,
     * so as to return only reasonably-live data to the caller.
     *
     * @param var definition of variable to fetch; never null
     *
     * @throws IOException  in case of I/O difficulty
     * @throws UnsupportedOperationException  if a variable is requested that
     *     we could never supply, eg a non-System-ID local variable at
     *     the client end of a tunnel
     */
    public SimpleVariableValue getVariable(SimpleVariableDefinition var)
        throws IOException,
        UnsupportedOperationException
        {
        return(source.getVariable(var));
        }

    /**Update a number of variables at once for efficiency; returns the number of variables set.
     * Is passed a set of SimpleVariableValues and behaves as if it
     * operates on all of them by calling setVariable() for each item
     * in the set.
     * <p>
     * An implementation may "fail fast" on the first error/exception,
     * or may attempt to continue and do as much as possible.
     * <p>
     * An implementation may throw an IllegalArgumentException on attempt to:
     *     set a variable with value of wrong type or incompatible definition,
     *     set a non-existent or read-only variable (or these may be ignored)
     *
     * @return the number of variable values set;
     *     never negative, never more than the number passed in
     */
    public int setVariables(SimpleVariableValue[] newValues)
        throws IOException
        {
        return(source.setVariables(newValues));
        }

    /**Set variable to the given value (the variable name and definition are implicit).
     * The value set should be immediately readable with getVariable()
     * as the "main" value of the variable even for global variables.
     *
     * @throws IOException  in case of I/O difficulty
     * @throws UnsupportedOperationException  if a variable is set that
     *     we could never handle, eg a non-System-ID local variable at
     *     the client end of a tunnel
     */
    public void setVariable(SimpleVariableValue newValue)
        throws IOException,
        UnsupportedOperationException
        {
        source.setVariable(newValue);
        }

    /**Get the current partial, or previous full, event set at the specified interval.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; may be null if requested set not available
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        return(source.getEventValue(def, intervalSelector, current));
        }

    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * (An implementation that is not an end-point may go upstream to fetch
     * missing values and cache them to satisfy future requests.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param intervalNumber  a time (as from System.currentTimeMillis())
     *     which identifies the first interval for which data is potentially
     *     required; if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc;
     *     null is treated as the common case equivalent to just bit 0 set
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException
        {
        return(source.getEventValues(def, intervalSelector, intervalNumber, whichValues));
        }

    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(final PropsKey key,
                                              final long versionID)
        throws IOException
        {
        return(source.getProperties(key, versionID));
        }


    /**Shut down the data pipeline.
     * Simply calls source.destroy().
     */
    public void destroy()
        {
        source.destroy();
        }
    }
