/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.datasource;

import org.hd.d.pg2k.svrCore.AllExhibitImmutableData;
import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitStaticAttr;
import org.hd.d.pg2k.svrCore.ExhibitThumbnails;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.SimpleVariablePipelineIF;

import java.io.IOException;
import java.nio.ByteBuffer;

/**Simple interface implemented by several stages in the Web exhibit-handling pipeline.
 * In particular, the on-disc exhibit-data cache and backend datasource adaptors
 * implement this.
 * <p>
 * The pipeline stages that implement this may well implement other interfaces
 * in parallel to provide richer and more efficient cacheing, etc.
 * <p>
 * As this is in effect handling a stream of I/O on its way to a browser,
 * IOException seems the natural way to signal errors.
 * <p>
 * We may make implementers of this interface Serializable so as to allow
 * them to be persisted
 * in servlet context for robustness and distribution, but making
 * copies through this mechanism will upset the ability to fully track
 * usage and efficiently share caches, etc.
 * <p>
 * Note that this extends the variable-handling pipeline interface,
 * ie data pipelines also handle system variables.
 */
public interface SimpleExhibitPipelineIF extends SimpleVariablePipelineIF
    {
    /**Maximum exhibit user-driven I/O data transfer request in bytes; strictly positive.
     * This is the limit for filesystem access, tunnel connections, etc,
     * to avoid causing bottlenecks and out-of-memory conditions.
     * <p>
     * Some operations that request transfers larger than this
     * (and where the caller has control of the size of the request),
     * eg reading raw data from an exhibit,
     * may be rejected with an IOException.
     * <p>
     * Note that this does not limit the return of large data items
     * (eg AllExhibitProperties) whose size is not controlled by the caller.
     * <p>
     * This helps prevent denial-of-service by a faulty/malicious client.
     * <p>
     * A power of two for efficiency in various places.
     * <p>
     * A size of many times BULK_DATA_TRANSFER_SIZE to a few mBytes
     * is probably reasonable.
     */
    public static final int MAX_USER_READ_SIZE = 1 * 1024 * 1024;

    /**Get the static attributes for a given exhibit.
     * Returns null if the named exhibit does not exist.
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public ExhibitStaticAttr getStaticAttr(String name)
        throws IOException;

    /**Read a chunk of the raw exhibit binary into the given buffer.
     * The call may return less than the requested number of bytes;
     * this will be clear from the state of the buffer.
     *
     * @param buf  the buffer into which to read the data;
     *     must be non-null, in put()able state,
     *     and with remaining capacity of at least the requested number of bytes
     * @param exhibitName  the full name of the exhibit to read from;
     *     never null and must be syntactically valid
     * @param position  position/index of first byte in exhibitFile to read;
     *     non-negative
     * @param dontCache  if true, this is a hint not to attempt to cache this
     *     or displace anything from extant caches for this data
     *     as it may for example be precaching or random activity;
     *     by default callers should leave this false to allow cacheing
     *
     * @throws java.lang.IllegalArgumentException  for invalid arguments such as
     *     a null or syntactically-invalid name,
     *     a negative position or excessive number of requested bytes,
     *     a negative (or in some cases, non-positive) len
     * @throws java.io.IOException for requests that cannot be fulfilled because of
     *     I/O restrictions or problems, such as link failure or
     *     an upper bound on the length of a request
     */
    public void getRawFile(ByteBuffer buf,
                           String exhibitName, int position, boolean dontCache)
        throws IOException;

    /**Gets set of all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public AllExhibitImmutableData getAllExhibitImmutableData(long oldStamp)
        throws IOException;

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned) null is returned.
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public AllExhibitProperties getAllExhibitProperties(long oldHash)
        throws IOException;

    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public org.hd.d.pg2k.svrCore.props.GenProps getGenProps(long oldStamp)
        throws IOException;

    /**Gets the generic security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     *
     * @deprecated Use getProperties() for new code.
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    @Deprecated
    public java.util.Properties getGenSecProps(long oldStamp)
        throws IOException;

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible, else if create is false
     *     only return an existing one and return null if none is to hand
     *
     * @throws java.io.IOException if the operation cannot be completed due to I/O
     *     restrictions or failure
     */
    public ExhibitThumbnails getThumbnails(String name, boolean create)
        throws IOException;

    /**Class that holds key to Properties lookup/fetch.
     * Immutable, and can be used as a key in a hash table.
     */
    public enum PropsKey
        {
        /**Generic security properties. */
        GenSecProps,

        /**Properties map from IPv4 octet prefix to ccTLD/region. */
        IPToCCTld,

        /**Location map from exhibit name (usually prefix/suffix) to Location value. */
        NameToLocation,
        }

    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(PropsKey key, long versionID)
        throws IOException;


    /**Poll periodically (of the order of seconds) to do background work.
     * The current generic system properties are passed in...
     * <p>
     * @throws java.io.IOException in case of difficulty, but even if a sub-ordinate
     *     call throws IOException then poll() call should continue
     *     to do as much of its remaining work as reasonably possible
     */
    public void poll(GenProps gp)
        throws IOException;

    /**Shut down the data pipeline.
     * Free resources having saved any persistent state
     * and shut down any upstream components.
     */
    public void destroy();
    }
