/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.location;

/**
 * Created by IntelliJ IDEA.
 * User: Damon
 * Date: 10-Feb-2006
 * Time: 15:57:00
 */

/**Defines some broad-brush, ordered, geographical Internet proximity values.
 * Defines strictly-positive value for two points "close" to one another,
 * indicating likely relative connectivity (cf two random points in the Net)
 * including notions of bandwidth, packet loss, RTT, reliability, cost, etc.
 * <p>
 * I assume that international transit costs roughly 50x intra-country bandwidth,
 * and that cheap peering is not available for all intra-country traffic.
 * I also assume the performance of a long route from a user's point of view
 * is usually somewhat less bad than the cost would imply.
 * <p>
 * The enum tag is the most important; the numerical values are very approximate
 * and only to be used if nothing better is available.
 * <p>
 * The tag "NONE" has a value of exactly 1 to indicate no closeness
 * or unknown closeness.
 */
public enum GeoProximity
    {
    /**Typically MAN or LAN or closer; fast, reliable and almost free. */
    VCLOSE(50),

    /**Within one country; usually reasonably fast, reliable and cheap. */
    COUNTRY(7),

    /**Group of well-connected countries (eg uk and nl, or us and ca). */
    COUNTRYGROUP(3),

    /**Within one continent or (more-or-less equivalently) one address-registry's ambit (eg RIPE). */
    CONTINENT(2),

    /**No particular/known proximity, may well involve expensive/slow/unreliable transit. */
    NONE(1);


    private GeoProximity(final int closeness) { this.closeness = closeness; }

    /**The "closeness" factor higher meaning closer; strictly positive. */
    private final int closeness;

    /**Get the "closeness" factor higher meaning closer; strictly positive. */
    public final int getCloseness() { return(closeness); }
    }
