/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.location;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.AddrTools;
import org.hd.d.pg2k.svrCore.MemoryTools;

import java.net.InetAddress;
import java.util.*;


/**Geographical-related utility functions.
 * This contains utilities, for example, to guess the approximate geographical
 * location of an HTTP client from its IP address.
 */
public final class GeoUtils
    {
    /**Prevent construction of an instance. */
    private GeoUtils() { }


    /**The base name of the default geographic proximity bundle.
     * Does not include ".properties" or ".class" suffix,
     * and is assumed to be relative to the root of the package structure.
     */
    private static final String BUNDLE_NAME_DEFAULT_GEO_PROXIMITY = "defaultGeoProximity";

    /**The base name of the default IP-to-location bundle.
     * Does not include ".properties" or ".class" suffix,
     * and is assumed to be relative to the root of the package structure.
     */
    private static final String BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX = "ccTLDFromIPPrefix";

    /**Immutable Map for default proximity from CCTLD to Set of all neighbours.
     * The values (Set&lt;CCTLD&gt;) are also immutable,
     * so they can be returned to callers safely without copying.
     */
    private static final Map<GeoUtils.CCTLD,Set<GeoUtils.CCTLD>> defaultGeoMap;

    static
        {
        final ResourceBundle dGP; // = null;
        final Map<CCTLD,Set<CCTLD>> m = new HashMap<CCTLD, Set<CCTLD>>();
        try
            {
            // Get the bundle if present...
            dGP = ResourceBundle.getBundle(BUNDLE_NAME_DEFAULT_GEO_PROXIMITY);

            // Expand into the quick-lookup format.
            for(final Enumeration<String> en = dGP.getKeys(); en.hasMoreElements(); )
                {
                final String key = en.nextElement();
                final StringTokenizer st = new StringTokenizer(dGP.getString(key));
                final Set<CCTLD> group = new HashSet<CCTLD>();
                while(st.hasMoreTokens())
                    {
                    final String cc = st.nextToken();
                    final CCTLD cctld = new CCTLD(MemoryTools.intern(cc));
                    group.add(MemoryTools.intern(cctld));
                    }

                // Add whole group to neighbours of all members of that group.
                // Keep the value Sets immutable.
                for(final Iterator<CCTLD> it = group.iterator(); it.hasNext(); )
                    {
                    final CCTLD cctld = it.next();
                    final Set<CCTLD> s = m.get(cctld);
                    final Set<CCTLD> newSet = (s == null) ? new HashSet<CCTLD>() : new HashSet<CCTLD>(s);
                    newSet.addAll(group);
                    m.put(cctld, Collections.unmodifiableSet(newSet));
                    }
                }
            }
        catch(final Exception e)
            {
            System.err.println("ERROR: failed to load/parse geo-proximity bundle: " + BUNDLE_NAME_DEFAULT_GEO_PROXIMITY);
            e.printStackTrace();
            }
        finally
            {
//            defaultGeoProximity = dGP;
            defaultGeoMap = Collections.unmodifiableMap(m);
            }
        }

    /**Get geographically-close ccTLDs to the specified ccTLD; empty if none, but never null.
     * This logically returns the union of all the groups that
     * the specified ccTLD is present in,
     * from static, built-in data, and any loadable data.
     * <p>
     * If no near neighbours for the specified ccTLD are known
     * then an empty Set is returned,
     * otherwise the specified ccTLD will be present in the result
     * ie a country is always close to itself.
     */
    public static Set<GeoUtils.CCTLD> getCloseCCTLDs(final GeoUtils.CCTLD ccTLD)
        {
        final Set<GeoUtils.CCTLD> result = defaultGeoMap.get(ccTLD);

        // If this is not in group, return an empty set.
        if(result == null)
            {
            final Set<GeoUtils.CCTLD> empty = Collections.emptySet();
            return(empty);
            }

        return(result);
        }

    /**Get ccTLDs in the specified region; empty if none, but never null.
     * This should be a valid registry name
     * or a name from the "defaultGeoProximity" properties.
     * <p>
     * The region name should <em>not</em> be a ccTLD
     * nor a numeric partial/full address.
     * <p>
     * Currently this is computed on the fly each time;
     * we may need to precompute or cache this result.
     */
    public static Set<GeoUtils.CCTLD> getCountriesInRegion(final String region)
        {
        // Quick check on argument validity.
        if((region == null) || (region.length() < 1))
            { throw new IllegalArgumentException(); }

        final Set<GeoUtils.CCTLD> result = new HashSet<CCTLD>();

        try
            {
            final ResourceBundle dGP = ResourceBundle.getBundle(BUNDLE_NAME_DEFAULT_GEO_PROXIMITY);
            final String countries = dGP.getString(region);

            // Compute the countries list...
            final StringTokenizer st = new StringTokenizer(countries);
            while(st.hasMoreTokens())
                { result.add(new CCTLD(st.nextToken())); }
            }
        catch(final MissingResourceException e)
            {
            // Ignore these entirely...
            }
        catch(final Exception e)
            {
            // Absorb errors, though we don't expect any, so report this...
            e.printStackTrace();
            }

        // Save some heap by returning a fixed empty set if nothing found...
        if(result.isEmpty())
            {
            final Set<GeoUtils.CCTLD> empty = Collections.emptySet();
            return(empty);
            }

        return(result);
        }



    /**An immutable DNS ccTLD (two-letter) country code.
     * Lower-case, two-ASCII-letter code.
     * <p>
     * This is <em>not</em> the ISO3316-1 code;
     * eg the United Kingdom is "uk" not "gb"
     * though "gb" may be acceptable as an alias.
     */
    public static final class CCTLD implements MemoryTools.Internable
        {
        /**Construct instance.
         * Supplied code must not be null
         * and must consist of exactly two lower-case 7-bit-ASCII letters.
         * <p>
         * The code is only checked for syntax,
         * not that such a country code actually exists.
         */
        public CCTLD(final String code)
            {
            if(!isSyntaticallyValidCcTLD(code))
                { throw new IllegalArgumentException("code must be non-null length-two lower-case-ASCII"); }

            this.code = code;
            }

        /**The two-letter lower-case-ASCII ccTLD; never null. */
        public final String code;

        /**True if argument is a (non-null) syntatically-valid ccTLD code.
         * This only looks for syntactic validity.
         *
         * @return true iff two-letter lower-case-ASCII argument;
         *     false for null or otherwise-invalid arguments
         */
        public static boolean isSyntaticallyValidCcTLD(final String code)
            {
            if((code == null) ||
               (code.length() != 2))
                { return(false); }

            final char c0 = code.charAt(0);
            if((c0 < 'a') || (c0 > 'z'))
                { return(false); }
            final char c1 = code.charAt(1);
            if((c1 < 'a') || (c1 > 'z'))
                { return(false); }

            // Looks OK.
            return(true);
            }

        /**A fast hash reflecting approx ~5 bits of information per char. */
        public int hashCode()
            { return(code.charAt(0) + 29*code.charAt(1)); }

        /**Two ccTLDs are the same if their underlying codes are the same. */
        public boolean equals(final Object obj)
            {
            if(obj == this) { return(true); }
            if(!(obj instanceof CCTLD)) { return(false); }
            final CCTLD other = (CCTLD) obj;
            return((code.charAt(0) == other.code.charAt(0)) &&
                   (code.charAt(1) == other.code.charAt(1)));
            }

        /**This returns the code as its String representation. */
        public String toString()
            { return(code); }
        }


    /**Default maximum prefix octets to look up in registry; strictly positive.
     * This indicates the most specific lookup that we will attempt.
     * <p>
     * We will try the least-specific lookup first anyway,
     * and just remember the most specific we found, if any,
     * or failing any match an appropriate (short) numeric prefix instead.
     * <p>
     * This may be overridable from the data.
     */
    private static final int DEFAULT_MAX_OCTETS_LOOKUP = 3;

    /**Test if argument is a (non-null) syntactically-valid region/registry code.
     * This only checks syntactic validity.
     *
     * @return true iff three-or-more-ASCII-letter argument (1st upper-case);
     *     false for null or otherwise-invalid arguments
     */
    public static boolean isSyntaticallyValidRegistryName(final String code)
        {
        if(code == null) { return(false); }
        final int l = code.length();
        if(l < 3) { return(false); }

        // First letter upper-case; the rest any case:
        // eg RIPE and AfriNIC.
        final char c0 = code.charAt(0);
        if((c0 < 'A') || (c0 > 'Z'))
            { return(false); }

        for(int i = l; --i > 0; )
            {
            final char c = code.charAt(i);
            final boolean isLetter =
                ((c >= 'A') && (c <= 'Z')) ||
                ((c >= 'a') && (c <= 'z'));
            if(!isLetter)
                { return(false); }
            }

        // Looks OK.
        return(true);
        }

    /**Routine to validate a putative RHS value for the ccTLDFromIPPrefix table.
     * A valid RHS (right-hand-side) value is one of:
     * <ul>
     * <li>An empty string ("") meaning:
     *     "make sure that there is no mapping for this value".
     * <li>A valid country code (two-letter lower-case ASCII).
     * <li>A valid region code (three-or-more-letter mixed-case ASCII).
     * </ul>
     *
     * @return true  iff the value is valid on the right-hand-side of an entry in
     *     a ccTLDFromIPPrefix map
     */
    private static boolean isValidCcTldFromIPPrefixMapValue(final String v)
        {
        if("".equals(v)) { return(true); }
        if(CCTLD.isSyntaticallyValidCcTLD(v)) { return(true); }
        if(isSyntaticallyValidRegistryName(v)) { return(true); }
        return(false); // Not valid.
        }

    /**Cached immutable default ccTLD-from-IP-prefix SortedMap, or empty if none.
     * Loaded at class initialisation.
     * <p>
     * This may have non-lossy transformations performed on it
     * before it is stored.
     * <p>
     * The values will have been de-duped (intern()ed) for memory efficiency,
     * since there are relatively few distinct values.
     * <p>
     * The keys are essentially unique and are NOT intern()ed,
     * since attempting to do so would waste memory and time.
     */
    private static final SortedMap<AddrTools.AddrPrefix,String> ccTLDFromIPPrefix;

    /**Get built-in immutable IPv4-to-ccTLD map; never null. */
    public static SortedMap<AddrTools.AddrPrefix,String> getCcTLDFromIPPrefix()
        { return(ccTLDFromIPPrefix); }

    /**Longest key in octets in ccTLDFromIPPrefix; non-negative.
     * Can be used to restrict search effort at lookup.
     */
    private static final int ccTLDFromIPPrefixLongestKey;

    /**Initialise ccTLDFromIPPrefix. */
    static
        {
        // Create in a HashMap for lookup/insertion speed...
        // We guess a reasonable initial working size.
        final Map<AddrTools.AddrPrefix,String> m = new HashMap<AddrTools.AddrPrefix, String>(250123);
        int longestKey = 0;
        try
            {
            // Get the bundle if present...
            final ResourceBundle rb = ResourceBundle.getBundle(BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX);

            // Maximum number of times to warn about discarded (over-long) prefixes.
            int maxWarnOverlong = 10;

            // Set of unique RHS values that we've seen so far
            // (for de-duping with fewer calls to intern(), which is very slow).
            final HashMap<String,String> deDupedValues = new HashMap<String,String>(511);

            // Parse each key/value in turn and add to the map if acceptable.
            final Enumeration<String> keys = rb.getKeys();
            while(keys.hasMoreElements())
                {
                final String key = keys.nextElement();
                final AddrTools.AddrPrefix ap;
                try { ap = new AddrTools.AddrPrefix(key); }
                catch(final IllegalArgumentException e)
                    {
                    System.err.println("ERROR: skipped unparsable address prefix `"+key+"' in ccTLD-from-IP-prefix bundle: " + BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX);
                    continue;
                    }

                final int apLen = ap.length();
                if(apLen > DEFAULT_MAX_OCTETS_LOOKUP)
                    {
                    if(--maxWarnOverlong >= 0)
                        { System.err.println("WARNING: skipped too-long address prefix `"+key+"' (max "+DEFAULT_MAX_OCTETS_LOOKUP+") in ccTLD-from-IP-prefix bundle: " + BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX); }
                    continue;
                    }

                // Get RHS (ccTLD/region/etc) code.
                final String value = rb.getString(key);

                // Locally remove RHS duplicates,
                // intern()ing (and validating) new values once.
                String dedupedValue = deDupedValues.get(value);
                if(null == dedupedValue)
                    {
                    // First time we've seen this RHS value so validate/veto it.
                    if(!isValidCcTldFromIPPrefixMapValue(value))
                        {
                        System.err.println("ERROR: skipped unparsable/illegal ccTLD/region `"+value+"' for key `"+key+"' in ccTLD-from-IP-prefix bundle: " + BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX);
                        continue;
                        }

                    dedupedValue = MemoryTools.intern(value);
                    deDupedValues.put(dedupedValue, dedupedValue);
                    }

                // Note the longest key to help make lookups more efficient.
                if(apLen > longestKey)
                    { longestKey = apLen; }

                // Add valid entry to map...
                m.put(ap, dedupedValue);
//System.out.println("Map size now: " + m.size());
                }

//System.out.println("deDupedValues size: " + deDupedValues.size());
//System.out.println("Map size: " + m.size());
            }
        catch(final Exception e)
            {
            System.err.println("ERROR: failed to load/parse ccTLD-from-IP-prefix bundle: " + BUNDLE_NAME_DEFAULT_CCTLD_FROM_IP_PREFIX);
            e.printStackTrace();
            }
        finally
            {
            // Store as an unmodifiable SortedMap.
            ccTLDFromIPPrefix = Collections.unmodifiableSortedMap(new TreeMap<AddrTools.AddrPrefix,String>(m));
            // And record the longest key actually present in the map.
            ccTLDFromIPPrefixLongestKey = longestKey;
            }

if(IsDebug.isDebug) { System.out.println("[GeoUtils.ccTLDFromIPPrefix.size() = " + ccTLDFromIPPrefix.size() + ".]"); }
        }

    /**Guess geographical "location" of given IP address (usually of an HTTP client).
     * This attempts to guess the top-level country code corresponding to
     * the likely physical/geographic location of the machine/interface whose
     * IP address is passed.
     * <p>
     * This may try a number of methods, some of which may require live DNS
     * or other Internet connectivity, but if it fails, with return null.
     * The results are not guaranteed to be accurate.
     * <p>
     * The code returned, if any, is the lower-case, two-letter,
     * Internet country code, for the region that the IP is in.
     * (This is not the ISO3166-1 code; the United Kingdom is "uk" not "gb".)
     * <p>
     * (Our implementation here is to get the region,
     * and then return anything that looks like a country-code, else null.)
     *
     * @param addr  the (client) address to look up
     * @param quick  if true, this will try to minimise time taken
     *     and resources used, eg it may avoid going to DNS,
     *     so the answer may be less accurate or null more often
     */
    public static CCTLD getCCTLDByAddress(final InetAddress addr,
                                          final boolean quick)
        {
        final String guess = getRegionByAddress(addr, quick);

        // If this looks like a ccTLD,
        // then try to return it as one.
        if(CCTLD.isSyntaticallyValidCcTLD(guess))
            { return(new CCTLD(guess)); }

        // Don't know.
        return(null);
        }

    /**Guess country or geographic region of IP address; never returns null.
     * This attempts to guess the top-level country code (ccTLD) corresponding to
     * the likely physical/geographic location of the machine/interface whose
     * IP address is passed.  (A two-letter lower-case country code.)
     * <p>
     * Failing that this routine may attempt to return the regional IP registry
     * (eg RIPE or APNIC) that delegates the address,
     * which usually indicates the continent from which the IP address hails.
     * (Starting with an upper-case letter and being more than two characters,
     * ie guaranteed to look like neither a ccTLD nor an IP-address prefix.)
     * <p>
     * Failing that this will return one or more octets of the IP(v4) address.
     * (Eg looking like "127" or "127.0" or "127.0.0" or "127.0.0.1".)
     * <p>
     * The number of different values that this routine can return is finite,
     * certainly no more than thousands, and all values are human-readable,
     * so this should be directly usable as a key in stats info,
     * for example.
     * <p>
     * This may try a number of methods, some of which may require live DNS
     * or other Internet connectivity, but if they fail,
     * this will return a numeric value.
     * <p>
     * The results are not guaranteed to be accurate.
     *
     * @param addr  the (client) address to look up; never null
     * @param quick  if true, this will try to minimise time taken
     *     and resources used, eg it may avoid going to DNS,
     *     so the answer may be less accurate or numeric more often
     *
     * @return non-null, non-empty, short, human-readable, plain-ASCII String
     */
    public static String getRegionByAddress(final InetAddress addr,
                                            final boolean quick)
        {
        if(addr == null)
            { throw new IllegalArgumentException(); }

        // Look up the address in our map
        // for the most-specific information available.
        //
        // If the result is a null (and we're restricted to a quick lookup)
        // then we return the numeric representation of the first octet...
        String result = lookupAddrInIPToCcTLDMap(addr,
                                ccTLDFromIPPrefix, ccTLDFromIPPrefixLongestKey);
        if(result == null)
            { result = new AddrTools.AddrPrefix(addr.getAddress(), 1).toPaddedDottedPrefix(); }

        // If the caller is prepared for us to do a slow (ie more accurate or comprehensive) lookup, AND
        // if the result so far is NOT a country code, AND
        // if the address looks like a normal unicast address,
        // then try doing a reverse lookup on the IP address and examine any resulting domain name
        // and if the name ends with a ccTLD then we will return that ccTLD.
        if(!quick &&
                !GeoUtils.CCTLD.isSyntaticallyValidCcTLD(result) &&
                !addr.isMulticastAddress())
            {
            try
                {
                // Try to do a (reverse) lookup of the client host name
                // from its IP address...
                // We don't need to verify this name;
                // if the client lies then it is probably their problem!
                // Note that this may have a trailing "."
                // to show that it is an absolute FQFN.
                String name = AddrTools.USE_DNSJAVA ? AddrTools.doReverseLookup(addr, false) : addr.getHostName();

//System.out.println("[Lookup "+addr.getHostAddress()+" --> "+name+"]");

                if(name != null)
                    {
                    // Strip off any canonical trailing dot...
                    if(name.endsWith("."))
                        { name = name.substring(0, name.length()-1); }
                    assert !name.endsWith(".") : ("had more than one trailing dot in name: "+name+".");

                    // Allow for names as short as "x.cc"...
                    final int nameLen = name.length();
                    if(nameLen > 3)
                        {
                        // If the final component is exactly two characters...
                        if(name.lastIndexOf('.') == nameLen - 3)
                            {
                            // Take the final component and force to lower-case.
                            final String putativeCCTLD =
                                name.substring(nameLen - 2).toLowerCase();

                            // If it is a valid ccTLD,
                            // then return it immediately...
                            if(GeoUtils.CCTLD.isSyntaticallyValidCcTLD(putativeCCTLD))
                                { return(putativeCCTLD); }
                            }
                        }
                    }
                }
            catch(final Exception e)
                {
                e.printStackTrace();
                // Simply ignore any lookup errors...
                }
            }


        // Return special (but non-null) value to indicate a config error
        // if nothing was set...
        // This does not look like a valid CC nor region nor numeric prefix.
        if(result == null)
            { return("-"); }

        return(result);
        }

    /**Do a lookup in the supplied table for the most-specific available location data for the given key; null if none.
     * Never uses DNS or other external data.
     *
     * @param addr  the address to look up; never null
     * @param map  the prefix map; never null
     * @param maxOctetsLookup  the maximum number of octets of prefix to look up; non-negative
     *
     * @return  the most-specific value; null if nothing (or a "" value) found
     */
    public static String lookupAddrInIPToCcTLDMap(final InetAddress addr,
                                                  final Map<AddrTools.AddrPrefix,String> map,
                                                  final int maxOctetsLookup)
        {
        String result = null;

        final byte[] rawAddr = addr.getAddress();

        // Make an increasingly long nnn[.nnn[.nnn ...]] prefix.
        // Don't go longer than the longest key actually present in the map
        // since that will always fail to find anything.
        for(int octets = 1; octets <= maxOctetsLookup; ++octets)
            {
            // Stop extending the prefix when we run out of address!
            if(octets > rawAddr.length)
                { break; }

            final AddrTools.AddrPrefix addrPrefix = new AddrTools.AddrPrefix(rawAddr, octets);

            // Attempt to look up the region.
            // Absorb complaint if prefix not present in data.
            final String region = map.get(addrPrefix);

            // If we got an empty value then clear any previous result.
            if("".equals(region))
                { result = null; }
            // If we got a non-null non-empty result,
            // then it is our most-specific so far, so keep it.
            else if(region != null)
                { result = region; }
            }

        return(result);
        }

    /**Compute approximate proximity of one host to a given country.
     * This is most useful where we know the location of (for example)
     * a mirror host for sure.
     * <p>
     * Returns "NONE" if it cannot compute a proximity
     * or if the hosts appear to be a long way apart.
     *
     * @param host1  IP address of first host; never null
     * @param host2CC  country of second host; never null
     * @param quick  if true, do a quick estimate,
     *     else be prepared to spend some time and CPU cycles to be as accurate as possible
     *     (for example we might have to to reverse DNS lookups)
     */
    public static GeoProximity computeProximityByAddress(final InetAddress host1,
                                                         final CCTLD host2CC,
                                                         final boolean quick)
        {
        if((host1 == null) || (host2CC == null))
            { throw new IllegalArgumentException(); }

        // Look up region/country (if known) for the first host.
        final String region1 = getRegionByAddress(host1, quick);

        return(computeProximity(region1, host2CC));
        }

    /**Compute approximate proximity of a host in a given region to a given country; never null.
     * This is most useful where we know the location of (for example)
     * a mirror host for sure.
     * <p>
     * Returns "NONE" if it cannot compute a proximity
     * or if the hosts appear to be a long way apart.
     *
     * @param region1  region or country of first host; never null
     * @param host2CC  country of second host; never null
     */
    public static GeoProximity computeProximity(final String region1,
                                                final CCTLD host2CC)
        {
        if((host2CC == null) || (region1 == null) || (region1.length() < 1))
            { throw new IllegalArgumentException(); }

        // If the first host is known to be in the same country as the second
        // then we're done.
        if(region1.equals(host2CC.code))
            { return(GeoProximity.COUNTRY); }

        // If the first host is a country
        // then check if we're in the same group/region.
        if(CCTLD.isSyntaticallyValidCcTLD(region1))
            {
            if(getCloseCCTLDs(host2CC).contains(new CCTLD(region1)))
                { return(GeoProximity.COUNTRYGROUP); }

            return(GeoProximity.NONE); // Cannot determine proximity.
            }

        // If region seems not to be a valid region name
        // then give up now...
        if(!GeoUtils.isSyntaticallyValidRegistryName(region1))
            { return(GeoProximity.NONE); /* Cannot determine proximity. */ }

        // Deal with the same registry/region/continent...
        if(getCountriesInRegion(region1).contains(host2CC))
            { return(GeoProximity.CONTINENT); }

        return(GeoProximity.NONE); // Cannot determine proximity.
        }
    }
