/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.location;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.HostUtils;
import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.VarTools;
import org.hd.d.pg2k.svrCore.vars.InstanceID;
import org.hd.d.pg2k.svrCore.vars.SimpleVariablePipelineIF;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;


/**Load-balancing utility functions for stand-alone apps and Web apps.
 */
public final class LoadBalancingUtils
    {
    /**Prevent construction of an instance. */
    private LoadBalancingUtils() { }


    /**If true then log balancing decisions, etc. */
    public static final boolean LOG_BALANCING_DECISIONS = false;

    /**Private cache for getActiveMirrors().
     * A map from mirror tag (String) to available bandwidth
     * in bytes-per-second-per-new-client (Long).
     * <p>
     * A Hashtable is used for thread safety,
     * but also a lock on this value is used
     * as a lock by getActiveMirrors() where needed.
     */
    private static final Hashtable<String,Long> _cache_gAM = new Hashtable<String,Long>();

    /**Time _cache_gAM is valid until; initially zero marking the cache as invalid. */
    private static long _cache_gAM_valid_until;

    /**Roughly how stale getActiveMirrors() is prepared for its answers to be, ms. */
    private static final int _MAX_STALE_GET_ACTIVE_MIRRORS_MS = VarTools.MIN_AGE_MS;

    /**Get active mirrors listed by mirror tag; may be empty but never null.
     * Retrieves the tags (cc-xxxxx) of active mirrors from system variables,
     * filtering out stale entries and those without significant free bandwidth.
     * <p>
     * Returns a Map from the mirror tag to the per-client bandwidth
     * that the mirror claims to have available.
     * <p>
     * This does <em>not</em> check names in DNS,
     * or that mirrors are actually responding,
     * just which mirrors claim to be up as recorded in the system variables.
     * <p>
     * If this cannot retrieve the variable values that it needs,
     * then this routine returns an (immutable) empty result.
     * <p>
     * This does not re-compute its value on each call,
     * but caches it for a while since it may be expensive to compute.
     * <p>
     * It is possible to optionally leave in "stale" entries
     * so as to be able to "coast" for a little longer
     * after loss of contact with the master.
     *
     * @param vars  the system variables; never null
     * @param filterStale  if true, filter out stale entries;
     *     this result is not cached
     *
     * @return  non-null immutable empty,
     *     or new mutable non-empty Map that is private to the caller
     */
    public static final SortedMap<String,Long> getActiveMirrors(final SimpleVariablePipelineIF vars,
                                                                final boolean filterStale)
        {
        if(vars == null)
            { throw new IllegalArgumentException(); }

        synchronized(_cache_gAM)
            {
            // If the cached value is still valid,
            // then return a *copy* of it.
            final long now = System.currentTimeMillis();
            if(!filterStale)
                {
                // Not filtering, so we can use the cache.

                // Returned cached content if still valid.
                if(now <= _cache_gAM_valid_until)
                    { return(new TreeMap<String,Long>(_cache_gAM)); }
                else // clear and recompute stale values...
                    {
                    _cache_gAM.clear(); // Zap the now-invalid cache content...
                    // We'll set a valid (short) cache life
                    // so that we don't keep trying to recompute this
                    // even when we encounter errors.
                    _cache_gAM_valid_until = now + Math.min(17000, _MAX_STALE_GET_ACTIVE_MIRRORS_MS/4);
                    }  
                }

            try
                {
                // Get set of systems claiming to be alive...
                final SimpleVariableValue mirrorAvailableBandwidthRaw =
                        vars.getVariable(SystemVariables.ThroughputMonitorFilter_AVAIL_BPS_PER_CLIENT);
                if(mirrorAvailableBandwidthRaw == null)
                    { return(_NO_LIVE_MIRRORS); }
                final SimpleVariableValue mirrorAvailableBandwidth = (!filterStale) ? mirrorAvailableBandwidthRaw :
                        mirrorAvailableBandwidthRaw.removeAllKeysOlder(System.currentTimeMillis() - _MAX_STALE_GET_ACTIVE_MIRRORS_MS);
                final Map<InstanceID,SimpleVariableValue> mABMap = mirrorAvailableBandwidth.getGlobalMap();
                if(mABMap == null)
                    { return(_NO_LIVE_MIRRORS); }

                // Get the map of mirror tags...
                final SimpleVariableValue mirrorTagsRaw =
                        vars.getVariable(SystemVariables.ThroughputMonitorFilter_ACTIVE_MIRROR_NAME);
                if(mirrorTagsRaw == null)
                    { return(_NO_LIVE_MIRRORS); }
                final SimpleVariableValue mirrorTags = (!filterStale) ? mirrorTagsRaw :
                        mirrorTagsRaw.removeAllKeysOlder(System.currentTimeMillis() - _MAX_STALE_GET_ACTIVE_MIRRORS_MS);
                final Map<InstanceID,SimpleVariableValue> mTMap = mirrorTags.getGlobalMap();
                if(mTMap == null)
                    { return(_NO_LIVE_MIRRORS); }

                // First get a list of all active mirrors.
                // Create a map from system ID
                // to non-zero available per-client bandwidth.
                final SortedMap<String,Long> liveMirrors = new TreeMap<String,Long>();

                // Scan the global map of available bandwidth,
                // and for every positive value,
                // note it with the mirror tag (assuming that there is one)
                // of the system concerned.
                for(final InstanceID id : mABMap.keySet())
                    {
                    final SimpleVariableValue svvAvailBandwidth = mABMap.get(id);
                    if((svvAvailBandwidth != null) && (svvAvailBandwidth.getValue() != null) && (((Number) svvAvailBandwidth.getValue()).longValue() > 0))
                        {
                        final SimpleVariableValue svvTag = mTMap.get(id);

                        if((svvTag != null) && (svvTag.getValue() != null))
                            {
                            // Got a live mirror and its tag!
                            final String tag = (String) svvTag.getValue();
                            final Long bw = new Long(((Number) svvAvailBandwidth.getValue()).longValue());
                            liveMirrors.put(tag, bw);
                            }
                        }
                    }

                // Memory use (heap-churn) optimisation:
                // return the single "no mirrors" value if none found...
                if(liveMirrors.isEmpty())
                    { return(_NO_LIVE_MIRRORS); }

                // Return details (and cache a copy) of one or more mirrors...
                // Don't cache filtered data.
                if(!filterStale)
                    {
                    // Cache the unfiltered data.
                    _cache_gAM.putAll(liveMirrors);
                    // Set a reasonable life on the non-empty cache...
                    _cache_gAM_valid_until = now + _MAX_STALE_GET_ACTIVE_MIRRORS_MS/2;
                    }
                return(liveMirrors);
                }
            // If we have a problem getting the variable values that we need,
            // then return a value indicating no currently-active mirrors.
            catch(final IOException e)
                {
                e.printStackTrace(); // Whinge, but absorb the error.
                return(_NO_LIVE_MIRRORS);
                }
            }
        }

    /**Immutable empty Map to indicate no live mirrors. */
    private static final SortedMap<String,Long> _NO_LIVE_MIRRORS =
            Collections.unmodifiableSortedMap(new TreeMap<String,Long>());


    /**Make full hostname of site given a mirror tag; never null/empty.
     * The host name returned is relative to the given domain name.
     * <p>
     * No checking is done that the resulting name actually refers to a valid host.
     * <p>
     * Result is undefined if the tag is not a valid mirror tag.
     */
    public static String makeMirrorHostnameFromTag(final String mirrorTag,
                                                   final String domainName)
        {
        final String baseName = HostUtils.normaliseVirtualHostName(domainName);
        return("mirror-" + mirrorTag + "." + baseName);
        }


    /**Immutable info on interaction with one remote mirror.
     */
    private static final class MirrorInfo
        {
        /**Make a new instance with a suitable lifetime... */
        MirrorInfo(final Boolean isUp,
                   final Thread t)
            {
            // Set random cache expiry time with mean specified,
            // plus a but extra for positive results.
            validUntil = System.currentTimeMillis() +
                    (Boolean.TRUE.equals(isUp) ?
                        (MIRROR_RECHECK_MS) : (MIRROR_RECHECK_MS/2)) +
                    Rnd.fastRnd.nextInt(MIRROR_RECHECK_MS);
            this.isUp = isUp;
            this.checker = t;
            }

        /**Time that this info is valid until. */
        final long validUntil;

        /**TRUE if the mirror seems to be up and working and accessible.
         * FALSE if the mirror seems to be down.
         * null if we do not yet know, because we are checking asynchronously.
         */
        final Boolean isUp;

        /**Non-null if being checked in background by thread; null if not being checked.
         * If any referred-to thread is not alive then this entry is not being checked.
         */
        final Thread checker;

        /**If true then this entry can be expired.
         * This is true if the validUntil time has passed
         * and there is no thread working on this host/entry.
         */
        boolean canExpire()
            {
            if(System.currentTimeMillis() <= validUntil) { return(false); }
            if(checker == null) { return(true); }
            return(!checker.isAlive());
            }

        /**If true then this entry will expire soon.
         * This can be used to decide if we should start a new checker thread.
         * <p>
         * This is true if the validUntil time will pass soon
         * and there is no thread working on this host/entry.
         */
        boolean expiresSoon()
            {
            if(System.currentTimeMillis() <= validUntil - (MIRROR_RECHECK_MS/4))
                { return(false); }
            if(checker == null) { return(true); }
            return(!checker.isAlive());
            }
        }


    /**Maximum time before we (re)test that a mirror is working, in ms.
     * Should be of the order of minutes to tens of minutes,
     * (and the results may be effectively cached in pages' HTML for a while)
     * but not so long as to make the system insensitive to extended outages
     * or gross configuration errors.
     * <p>
     * Testing may be slow and/or consume significant network resources.
     * <p>
     * We may asymmetrically retain positive results longer than negative ones.
     */
    private static final int MIRROR_RECHECK_MS = 13 * 60 * 1000 +
        Rnd.fastRnd.nextInt(3 * 60 * 1000);


    /**Construct a full mirror host name from the mirror tag.
     * Result is undefined if the tag is invalid.
     */
    public static String makeMirrorNameFromTag(final String tag)
        {
        return("mirror-" + tag + "." + CoreConsts.MAIN_DATA_HOST);
        }

    /**Try the requested mirrors in turn, returning the full hostname of the first that is up, or null if none is available.
     * The mirrors are tried in the order in the list.
     * <p>
     * The cache of mirror state is used, and updated if need be.
     * <p>
     * This routine may take considerable time to complete
     * if this has to actually test the mirrors
     * and they are down or network connectivity is poor.
     *
     * @param orderedMirrorTags  valid mirror tags in the order the mirrors should be tried
     *
     * @return the full hostname of the first working mirror found,
     *     or null if none of the mirrors specified appear to be up
     */
    public static String findFirstWorkingMirror(final List<String> orderedMirrorTags)
        {
        // Try the tags in order, asynchronously.
        for(final String tag : orderedMirrorTags)
            {
            final String fullHostname = makeMirrorNameFromTag(tag);

            // If the server is up then our search is over.
            // Don't block to find out though.
            if(Boolean.TRUE.equals(testIfHTTPServerIsUp(fullHostname, true)))
                { return(fullHostname); }
            }

        return(null); // No available mirror.
        }

    /**Orders mirrors into "fastest first" order, returing a List of mirror tags and dropping any known unusable; never null.
     * This accepts as input a set of mirror details as returned from getActiveMirrors()
     * (a mapping from mirror tag to claimed available bandwidth per client).
     * <p>
     * The results will be ordered to be highest-bandwidth first, and then by name.
     * <p>
     * This will drop mirrors known from the cache to be dead/unreachable,
     * and any mirrors that are <en>much</em> slower than the best available.
     * <p>
     * The result will be smaller when inappropriate/slow/dead mirrors
     * are dropped from the result.
     * <p>
     * This does <em>not</em> synchronously attempt
     * to verify the accuracy of the data that it has been passed,
     * eg it does not do blocking DNS lookups nor make HTTP connections,
     * so should always be reasonably fast.
     * (It does however check the status of remote mirrors asynchrnously,
     * and won't return a candidiate unless verified to be working.)
     *
     * @param availableMirrors  mapping of mirror tags to available per-client bandwidth; never null
     * @return best-first list of mirrors, may be empty and/or smaller than the input set; never null
     */
    public static List<String> orderMirrorTagsBestBandwidthFirst(final Map<String, Long> availableMirrors)
        {
        // Assume that all candidates will usually make it to the result.
        final List<String> result = new ArrayList<String>(availableMirrors.size());

        // First filter out any unsuitable mirrors.
        // This may be any with non-positive bandwidth,
        // or bandwidth too small to sustain a good user experience,
        // or that a geographically inappropriate for the user.
        for(final String tag : availableMirrors.keySet())
            {
            final Long bwL = availableMirrors.get(tag);
            // Reject malformed or non-positive bandwidth entries...
            if(bwL == null) { continue; }
            if(bwL.longValue() <= 0) { continue; }

//            // Reject mirrors with far too little bandwidth to be likely to be of use.
//            // We filter for much lower than typical per-user bandwidth,
//            // in case our number is somewhat wrong.
//            if(_OMTBF_DROP_MIRRORS_BW_TOO_LOW &&
//                    (bwL.longValue() < WebConsts.DEFAULT_EXPECTED_USER_BW_BYTESPERSEC/10))
//                { continue; }

            final String fullHostname = makeMirrorNameFromTag(tag);

            // If we don't know the mirror to be up
            // then move on to the next possible mirror.
            if(!Boolean.TRUE.equals(testIfHTTPServerIsUp(fullHostname, true)))
                { continue; }

            result.add(tag);
            }

        // If there are no remaining results then return immutable empty List to save space.
        if(result.size() == 0)
            {
            final List<String> emptyList = Collections.emptyList();
            return(emptyList);
            }

        // If more than one remaining candidate
        // then sort the results into bandwidth order
        // and trim any VERY slow hosts...
        if(result.size() > 1)
            {
            Collections.sort(result, new BWOrder(availableMirrors));

            // Get the bandwidth of the best entry...
            final Long bestBW = availableMirrors.get(result.get(0));
            if(bestBW == null) { throw new IllegalStateException(); }

            // Compute minimum bandwidth to remain in the result.
            // In the first instance a fraction of the fastest mirror's bandwidth
            // that is about the same as the maximum as the maximum "closeness" benefit.
            // The reasoning is that even being very close does not compensate for being very slow.
            final long thresholdBW = bestBW.longValue() / (2*GeoProximity.VCLOSE.getCloseness());

            // Nibble away entries from the end of the list that don't make the grade.
            // Working backwards, as soon as we find one good enough then we can stop.
            for(int i = result.size(); --i >= 1; )
                {
                final Long bw = availableMirrors.get(result.get(i));
                if(bw == null) { throw new IllegalStateException(); }
                if(bw.longValue() >= thresholdBW) { break; }
                result.remove(i);
                }
            }

        // Return the ordered results.
        return(result);
        }

//    /**If true then we automatically filter out mirrors with too little bandwidth in orderMirrorTagsBestBandwidthFirst(). */
//    private static final boolean _OMTBF_DROP_MIRRORS_BW_TOO_LOW = true;

    /**Our simple sorting comparator for mirrors sorts by bandwidth (or other Long metric) then name. */
    public static final class BWOrder implements Comparator<String>
        {
        /**Create an instance with the tag info map behind it. */
        public BWOrder(final Map<String,Long> mirrorInfo)
            { this.mirrorInfo = mirrorInfo; }
        /**The map from tag name to bandwidth. */
        private final Map<String,Long> mirrorInfo;

        /**Sort in order by bandwidth (highest first), then by name to give a total ordering. */
        public final int compare(final String s1, final String s2)
            {
            final long b1 = mirrorInfo.get(s1).longValue();
            final long b2 = mirrorInfo.get(s2).longValue();

            // Sort for largest-bandwidth first...
            if(b1 > b2) { return(-1); }
            if(b1 < b2) { return(+1); }

            // Sort by name to break ties and thus to provide a total ordering.
            return(s1.compareTo(s2));
            }
        }

    /**Connection establishment timeout allowed by testIfHTTPServerIsUp(), in ms.
     * Fairly short to make sure that a server is responding reasonably quickly,
     * but long enough to allow for DNS lookup and the odd lost packet
     * and some intermittent Net congestion.
     * <p>
     * Should be of the order of a few seconds to a few tens of seconds.
     */
    private static final int _tIHSIU_CONN_TIMEOUT_MS = 9001 + Rnd.fastRnd.nextInt(3001);

    /**Number of connection attempts before deciding that remote host is dead; strictly positive. */
    private static final int _tIHSIU_CONN_ATTEMPTS = 13 + Rnd.fastRnd.nextInt(3);


    /**Private cache for testIfHTTPServerIsUp() for which mirrors are up.
     * Map from mirror full host name component of URL (not just the mirror tag)
     * to last time server at that URL was tested and found to respond properly
     * (eg fast enough and with no error) or not.
     * <p>
     * Uses a thread-safe Map implementation.
     */
    private static final Map<String,MirrorInfo> _cache_tIHSIU_mirrorStatus = new Hashtable<String,MirrorInfo>();

    /**Tests if an HTTP Gallery mirror server is up and running and responding quickly.
     * The hostname or the hostname:port must be supplied;
     * the "null" page will be polled
     * (which should always work unless the host is down or busy).
     * <p>
     * Tests that the web server responds quickly and without error
     * (we must get response code 200),
     * to a request for its top-level page "/null.jsp"
     * (redirections are allowed, eg to an index.html or index.jsp).
     * <p>
     * If an invalid (non-null) host name is supplied this will return false.
     * <p>
     * We should only do one of these at a time for a given host
     * in order to avoid overloading any one remote mirror.
     * <p>
     * Note that even if we check synchronously,
     * we may still record any result in the cache
     * so that another caller may benefit.
     * This also means that this routine can call itself recursively
     * (with the checkAsynchronously argument false)
     * via asynchronous threads.
     * <p>
     * Note that this may run as many threads as there are different hosts
     * whose status is being requested.
     * We could probably do better with non-blocking I/O
     * and a single worker thread if there are many hosts to check,
     * but the assumption is that there will not be.
     *
     * @param hostAndPort  hostname or hostname:port for mirror; never null
     * @param checkAsynchronously  if false we check synchronously
     *     and block until we can return TRUE or FALSE,
     *     else we return immediately and check in the background
     * @return TRUE if the specified server seems to be up and running,
     *         FALSE if the server seems to be down,
     *         and null if we do not know at the moment
     *         (never returns null for a synchronous check)
     */
    public static Boolean testIfHTTPServerIsUp(final String hostAndPort,
                                               final boolean checkAsynchronously)
        {
        // If we are not using the cache then block here
        // and update the cache if there is no thread updating
        // or if the updating thread is us.
        if(!checkAsynchronously)
            {
            boolean r; // = false;
            try { r = testIfHTTPServerIsUp(new URL("http://" + hostAndPort + CoreConsts.DO_NOTHING_PAGE_RRURL)); }
            catch(final MalformedURLException e)
                {
                e.printStackTrace(); // Unexpected error...
                r = false;
                }

            final Boolean result = r ? Boolean.TRUE : Boolean.FALSE;
            return(result);
            }

        // Removing expired entries...
        // See if this thread is going to be picked to expire stale entries.
        // We pick at random in inverse proportion to the size of the cache,
        // in order to make the average cost per call constant.
        // We do this under a lock on the cache to eliminate races,
        // but we can release this lock again as soon as we are done.
        synchronized(_cache_tIHSIU_mirrorStatus)
            {
            final int size = _cache_tIHSIU_mirrorStatus.size();
            if((size > 0) && (Rnd.fastRnd.nextInt(2 + size) == 0))
                {
                // Expire old entries.
                for(final Iterator<String> it = _cache_tIHSIU_mirrorStatus.keySet().iterator(); it.hasNext(); )
                    {
                    final String tag = it.next();
                    // Remove a stale entry as long as
                    // it does not still have a live thread updating it.
                    if(_cache_tIHSIU_mirrorStatus.get(tag).canExpire())
                        { it.remove(); }
                    }
                }
            }

        // See what we currently have in the cache for this host.
        // Make this, and the possible launch of a check thread, atomic.
        synchronized(_cache_tIHSIU_mirrorStatus)
            {
            final MirrorInfo extant = _cache_tIHSIU_mirrorStatus.get(hostAndPort);

            // If there is no entry or it has nearly expired
            // then launch a background thread to get it updated.
            // Create an entry that preserves the current value while we check;
            // this limits us to at most one checker thread per host.
            final Boolean result;
            if((extant == null) || extant.expiresSoon())
                {
                // Compute result to return now.
                result = (extant == null) ? null : extant.isUp;

                // Create a checker Thread.
                final Thread t = new Thread("checking liveness of server " + hostAndPort){
                    public final void run()
                        {
                        // Test synchronously...
                        final Boolean r = testIfHTTPServerIsUp(hostAndPort, false);
                        // This should always give a definite answer.
                        assert(r != null);

                        // If we can update the cache, then do it.
                        // We hold a lock to keep any test-and-update atomic.
                        synchronized(_cache_tIHSIU_mirrorStatus)
                            {
                            // If we can update the cache then do so.
                            // It can be updated if:
                            //   * There is no entry.
                            //   * There is no (live) checker thread.
                            //   * There is a current entry but we are the checker thread.
                            final MirrorInfo extant = _cache_tIHSIU_mirrorStatus.get(hostAndPort);
                            if((extant == null) ||
                               (extant.checker == null) || !extant.checker.isAlive() ||
                               (extant.checker == Thread.currentThread()))
                                {
                                final MirrorInfo newInfo = new MirrorInfo(r, null);
                                final MirrorInfo oldInfo = _cache_tIHSIU_mirrorStatus.put(hostAndPort, newInfo);
                                // Report any change in status.
                                // Assumes that the Boolean can only be
                                // one of the fixed TRUE/FALSE/null
                                // so !=/== can be used rather than equals().
                                if((oldInfo == null) || (r != oldInfo.isUp))
                                    {
                                    System.out.println("[testIfHTTPServerIsUp("+hostAndPort+"): status change: now "+r+"; recheck after: "+(new Date(newInfo.validUntil))+"]");
                                    }
                                }
                            }
                        }
                    };

                _cache_tIHSIU_mirrorStatus.put(hostAndPort,
                                               new MirrorInfo(result, t));
                // Mark the the thread as a daemon and start it off.
                t.setDaemon(true);
                t.start();
                }
            else
                {
                result = extant.isUp;
                }

            return(result);
            }
        }

    /**Tests if a Web server is up and running and responding quickly.
     * A full URL must be supplied.
     * <p>
     * Tests that the web server responds quickly and without error
     * (we must get response code 200),
     * to a HEAD request for a special light-weight page "/null.jsp"
     * (redirections are allowed, eg to an index.html or index.jsp).
     * <p>
     * If an invalid (non-null) host name is supplied this will return false.
     * <p>
     * We should only do one of these at a time for a given host
     * in order to avoid overloading any one remote mirror.
     *
     * @param serverURL  full URL for remote server (page) to test; never null
     * @return true if the specified server seems to be up and running, false otherwise
     */
    public static boolean testIfHTTPServerIsUp(final URL serverURL)
        {
        if(serverURL == null)
            { throw new IllegalArgumentException(); }

        for(int attemptsLeft = _tIHSIU_CONN_ATTEMPTS; --attemptsLeft >= 0; )
            {
            HttpURLConnection uc = null;
            try
                {
                uc = (HttpURLConnection) serverURL.openConnection();
                if(!CoreConsts.AVOID_UNSAFE_TCP_TIMEOUTS)
                    {
                    uc.setConnectTimeout(_tIHSIU_CONN_TIMEOUT_MS); // Allow DNS lookup, TCP handshake, etc...
                    uc.setReadTimeout(_tIHSIU_CONN_TIMEOUT_MS); // Page may be expensive/slow to generate.
                    }
                uc.setUseCaches(false); // Force a fresh fetch/connection.
                uc.setAllowUserInteraction(false);
                uc.setInstanceFollowRedirects(true);
                uc.setRequestMethod("HEAD"); // Only need to get the status, not the data.
                uc.connect(); // Try to force a connection...
                final int responseCode = uc.getResponseCode();
                if(200 == responseCode)
                    {
                    // All seems OK, so return immediately.
                    return(true);
                    }

//System.err.println("[testIfHTTPServerIsUp("+serverURL+") failed (attemptsLeft="+attemptsLeft+") with status code " + responseCode + ".]");
                }
            // Any IOException probably indicates that the server is NOT well.
            catch(final IOException e)
                {
//System.err.println("[testIfHTTPServerIsUp("+serverURL+") failed (attemptsLeft="+attemptsLeft+") with error " + e.getMessage() + ".]");
                }

            // Drop possibly-broken connection.
            if(uc != null) { uc.disconnect(); }

            // Pause a while before trying again...
            // This may give DNS and other temporary glitches time to pass.
            try { Thread.sleep(15101 + 2*_tIHSIU_CONN_TIMEOUT_MS); }
            catch(final InterruptedException e)
                {
                Thread.currentThread().interrupt();
                return(false); // Return immediately after interrupt.
                }
            }

        // Failed to connect...
        return(false);
        }
    }
