/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.svrCore.location;

import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.PGException;

import java.io.*;
import java.util.Properties; // For Hashtable, Collections, etc.

/**Classes to hold (time and space) location data on an item.
 * Such location data is immutable.
 * <p>
 * This is not considered an attribute of an item in the Gallery,
 * but instead data associated with such an item.  Many may not
 * have this data at all.
 * <p>
 * Location data is read from a Properties map using
 * Location.Base.buildFromProperties() to construct the
 * appropriate subtype of Base to contain the data (or
 * throw an exception if there is one).  See the routine's
 * description for the properties used.
 * <p>
 * Note that while location data read for a specific exhibit
 * is taken to be `specific' else not, this can be overridden
 * by use of the `specific' flag which takes a boolean ``true''
 * or ``false'' value.
 * <p>
 * TODO: needs object verification after deserialisation
 *     and mutable values need to be cloned.
 *
 * @author Damon Hart-Davis
 * @version 1.41 01/08/14
 */
public final class Location
    {
    /**Don't construct instances of this container class. */
    private Location() { }

    /**Base class of location value with error bounds.
     * The value is the mid-point of range of the value,
     * and the error is the width of the interval, ON EITHER SIDE OF
     * the value (the mid-point), encompassing both any actual `duration'
     * or `size' of the item and any error or uncertainty in
     * the value.  Thus the value can be shown as +/- the error.
     * <p>
     * The error values returned must always be non-negative
     * (zero is allowed).
     * <p>
     * This abstract class can be derived to store the units
     * in various forms.  All derived classes must be immutable,
     * so we will only allow it to be constructed by well-behaved
     * sub-classes that use guaranteed-immutable Number args,
     * and that also ensure that the error value is non-negative.
     * <p>
     * This and any deriving classes should be immutable,
     * serialisable, and support equals() and hashCode().
     * <p>
     * TODO: should validate on deserialisation.
     */
    public static abstract class ValueAndBounds implements Serializable
        {
        /**Construct a new value.
         * The constructor visibility limits deriving classes to this package.
         * This helps ensure that we can guarantee features sush as
         * immutability.
         */
        ValueAndBounds(Number v, Number e)
            {
            if((v == null) || (e == null))
                { throw new IllegalArgumentException("null arg"); }
            value = v; error = e;
            }

        /**Central value; never null. */
        public final Number value;
        /**Error bounds; never null. */
        public final Number error;

        /**Currently depends on only value for speed. */
        public int hashCode()
            { return(value.hashCode()); }

        /**Depends on equality of value and error. */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof ValueAndBounds)) { return(false); }
            final ValueAndBounds other = (ValueAndBounds) obj;
            return((value.equals(other.value)) &&
                   (error.equals(other.error)));
            }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 1802455018686196426L;
        }

    /**Value-and-error stored as longs, eg for time.
     * TODO: should do validation on deserialisation.
     */
    public static final class LongValueAndBounds extends ValueAndBounds
        {
        /**Make new bounded value. */
        public LongValueAndBounds(long _val, long _err)
            {
            super(new Long(_val), new Long(_err));
            if(_err < 0)
                { throw new IllegalArgumentException("-ve error"); }
            }
        /**Make new bounded value. */
        public LongValueAndBounds(Long _val, Long _err)
            {
            super(_val, _err);
            if(_err.longValue() < 0)
                { throw new IllegalArgumentException("-ve error"); }
            }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 4934329592740791248L;
        }

    /**Value-and-error stored as doubles, eg for place.
     * TODO: should do validation on deserialisation.
     */
    public static final class DoubleValueAndBounds extends ValueAndBounds
        {
        /**Make new bounded value. */
        public DoubleValueAndBounds(double _val, double _err)
            {
            super(new Double(_val), new Double(_err));
            if(_err < 0)
                { throw new IllegalArgumentException("-ve error"); }
            }
        /**Make new bounded value. */
        public DoubleValueAndBounds(Double _val, Double _err)
            {
            super(_val, _err);
            if(_err.doubleValue() < 0)
                { throw new IllegalArgumentException("-ve error"); }
            }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = -666396322028089442L;
        }

    /**Base of all location classes.
     * All such classes contain an optional time value
     * (with error bounds), and flag to indicate if this
     * location information was specific to this collection
     * item, or applied to a group of items.
     * <p>
     * A true value for `specific' says that the item had
     * its own time info rather than inheriting a general value.
     * <p>
     * If the time value is zero it indicates that none is known.
     * (If the time value is zero, so should its error bound be.)
     * <p>
     * All classes derived from this should be immutable
     * and serialisable and support equals() and hashCode().
     * <p>
     * TODO: should do validation on deserialisation.
     */
    public static abstract class Base implements java.io.Serializable, MemoryTools.Internable
        {
        /**Build new base from properties.
         * Looks for the type field, which must be present,
         * and the optional time fields.
         * <p>
         * Note that the _specific parameter is overridden
         * where a `specific' property is present.
         * <p>
         * Package-visible to restrict the range of derived classes.
         *
         * @param _specific  is true if the location info was specifically
         *                   for this item, else false to indicate that
         *                   this location info was inherited from some
         *                   more general source
         */
        protected Base(boolean _specific,
                       String prefix,
                       Properties p)
            throws PGException
            {
            type = p.getProperty(prefix + typeKey);
            if(type == null)
                { throw new PGException("missing type field: " + prefix + typeKey); }
            final Boolean pSpec = getBooleanProp(prefix, p, specificKey, false);
            specific = ((pSpec == null) ? _specific : pSpec.booleanValue());
            }
        /**The type of this location info. */
        public final String type;
        /**Did this item have its own specific location info? */
        public final boolean specific;

        /**Hash depends on type and "specific" flag. */
        public int hashCode()
            { return(type.hashCode() ^ (specific ? 0 : 5)); }

        /**Hash depends on equality of type and "specific" flags.
         * A derived class should probably make use of this recursively.
         */
        public boolean equals(Object obj)
            {
            if(obj == this) { return(true); }
            if(!(obj instanceof Base)) { return(false); }
            final Base other = (Base) obj;
            return((specific == other.specific) &&
                   (type.equals(other.type)));
            }

        /**Generate terse, human readable portion of location info.
         * This should be overridden by derived classes, but
         * with this used (via super.toString()) to generate the
         * prefix for the result.
         * <p>
         * The string generated by this routine contains no
         * white space to make the result handlable as one token
         * and is shell- and HTML- friendly, with portions separated
         * by colons, and deriving classes must abide by this
         * too.
         * <p>
         * This generates a string consisting of the type name
         * followed by a colon then `s' for a specific location or
         * `g' for a generic one.
         */
        public String toString()
            { return(type + ':' + (specific ? 's' : 'g')); }


        /**Factory method to build a class from a property set; never returns null.
         * The Properties map passed is only read (not updated),
         * and consists of a number of keys prefixed with
         * the given prefix (either the empty string, or
         * a fully-qualified name prefix, ending with ".").
         * <p>
         * Whatever type of map is being built, there must
         * always be the key given by the prefix + typeKey,
         * which must contain the type of location info, and
         * that type must be the name of the class in this
         * Location outer class, derived from Base, that holds
         * the location info.
         * <p>
         * For robustness, if we cannot use the location info
         * for some reason, we should probably pretend that the
         * item has none, though flag the problem to stderr.
         *
         * @exception PGException in case of inability to
         *     build a location value from the specified
         *     properties
         */
        public static Base buildFromProperties(boolean specific,
                                               String prefix,
                                               Properties p)
            throws PGException
            {
            // Find the type.
            final String type = p.getProperty(prefix + typeKey);
            if(type == null)
                { throw new PGException("missing type field: " + prefix + typeKey); }

            // Try to find the class indicated by the type...
            // Do this by hand rather than by pure reflection.
            Class c;
            if("Base".equals(type)) { c = Base.class; }
            else if("None".equals(type)) { c = None.class; }
            else if("Time".equals(type)) { c = Time.class; }
            else if("Estd".equals(type)) { c = Estd.class; }
            else
                { throw new PGException("unknown location type (class not found): " + type); }

            // Try to find the constructor we will need...
            final Class argTypes[] =
                { boolean.class, String.class, Properties.class };
            java.lang.reflect.Constructor cons;
            try { cons = c.getConstructor(argTypes); }
            catch(NoSuchMethodException e)
                { throw new PGException("unsuitable location type (no such method): " + type); }
            if(cons == null)
                { throw new PGException("unsuitable location type (no constructor): " + type); }

            // Now try and make an instance...
            // We try to instance-control the values we create to help conserve memory.
            final Object args[] =
                { specific ? Boolean.TRUE : Boolean.FALSE, prefix, p };
            final Object o;
            try { o = cons.newInstance(args); }
            catch(final ThreadDeath t) { throw t; } // Die on command.
            catch(final Throwable t)
                {
                t.printStackTrace();
                throw new PGException("uninstantiable location type (unexpected exception): " + type + ": " + t.getMessage());
                }
            if(!(o instanceof Base))
                { throw new PGException("unknown location type (bad class): " + type); }

            return(MemoryTools.intern((Base) o)); // All done!
            }
        /**Build from properties with empty prefix. */
        public static Base buildFromProperties(boolean specific, Properties p)
            throws PGException
            { return(buildFromProperties(specific, "", p)); }
        /**Build directly from a file, assuming empty prefix. */
        public static Base buildFromFile(boolean specific, File f)
            throws IOException, PGException
            {
            final InputStream is = new BufferedInputStream(new FileInputStream(f));
//            if(is == null) { throw new IOException("could not open location properties file: " + f); }
            try {
                final Properties p = new Properties();
                p.load(is);
                return(buildFromProperties(specific, p));
                }
            finally { is.close(); }
            }
        /**Attempts to build location from file.
         * Returns Location.NONE type in case of difficulty.
         */
        public static Base buildFromFileOrNONE(boolean specific, File f)
            {
            if(!f.exists()) { return(Location.NONE); }
            try {
                final Base result = buildFromFile(specific, f);
//System.err.println(" " + f + ": OK");
                return(result);
                }
            catch(PGException e)
                {
System.err.println(" " + f + ": pretending no location info: " + e);
                return(Location.NONE);
                }
            catch(IOException e)
                {
System.err.println(" " + f + ": IO problem: pretending no location info: " + e);
                return(Location.NONE);
                }
            }

        /**Name of key giving location type in Properties map. */
        public static final String typeKey = "type";

        /**Name of key giving specificity in Properties map. */
        public static final String specificKey = "specific";


        /**Routine to get a double property.
         * Throws a PGException if there is a problem.
         * <p>
         * Can be asked to veto negative values, eg for error margins.
         */
        public static final Double getDoubleProp(String prefix, Properties p,
                                                 String prop, boolean rejectNegative)
            throws PGException
            {
            final String v = p.getProperty(prefix + prop);
            if(v == null) { throw new PGException("missing property " + prop); }
            try {
                final Double result = new Double(v);
                if(!rejectNegative) { return(result); }
                if(result.doubleValue() < 0)
                    { throw new PGException("illegal negative value for property: " + prop); }
                return(result);
                }
            catch(NumberFormatException e)
                { throw new PGException("unparsable (double) property: " + prop); }
            }
        /**Get pair of double properties; requested one and its error interval.
         * If the specified property is <i>X</i>, the main property is
         * <i>X</i>Err.
         * <p>
         * If manditory and either property is missing, a PGException is
         * thrown.
         * <p>
         * If optional and the first property is missing, null is returned;
         * but if the first property is present and the second is absent
         * then a PGException is thrown.
         */
        public static final DoubleValueAndBounds getDoubleProps(String prefix, Properties p,
                                                                String prop,
                                                                boolean manditory)
            throws PGException
            {
            if(!manditory && (p.getProperty(prefix + prop) == null)) { return(null); }
            return(new DoubleValueAndBounds(
                getDoubleProp(prefix, p, prop, false),
                getDoubleProp(prefix, p, prop + "Err", true)));
            }


        /**Routine to get a boolean property.
         * Throws a PGException if there is a problem.
         */
        public static final Boolean getBooleanProp(String prefix, Properties p,
                                             String prop, boolean manditory)
            throws PGException
            {
            final String v = p.getProperty(prefix + prop);
            if(v == null)
                {
                if(!manditory) { return(null); }
                throw new PGException("missing property " + prop);
                }
            final Boolean result = Boolean.valueOf(v);
            return(result);
            }


        /**Routine to get a double property.
         * Throws a PGException if there is a problem.
         * <p>
         * Can be asked to veto negative values, eg for error margins.
         */
        public static final Long getLongProp(String prefix, Properties p,
                                             String prop, boolean rejectNegative)
            throws PGException
            {
            final String v = p.getProperty(prefix + prop);
            if(v == null) { throw new PGException("missing property " + prop); }
            try {
                final Long result = new Long(v);
                if(!rejectNegative) { return(result); }
                if(result.longValue() < 0)
                    { throw new PGException("illegal negative value for property: " + prop); }
                return(result);
                }
            catch(NumberFormatException e)
                { throw new PGException("unparsable (long) property: " + prop); }
            }

        /**Get pair of long properties; requested one and its error interval.
         * If the specified property is <i>X</i>, the main property is
         * <i>X</i>Err.
         * <p>
         * If manditory and either property is missing, a PGException is
         * thrown.
         * <p>
         * If optional and the first property is missing, null is returned;
         * but if the first property is present and the second absent a
         * PGException is thrown.
         */
        public static final LongValueAndBounds getLongProps(String prefix, Properties p,
                                                            String prop,
                                                            boolean manditory)
            throws PGException
            {
            if(!manditory && (p.getProperty(prefix + prop) == null)) { return(null); }
            return(new LongValueAndBounds(
                getLongProp(prefix, p, prop, false),
                getLongProp(prefix, p, prop + "Err", true)));
            }


        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = -7784138110080680826L;
        }

    /**Singleton: represents an absence of location information. */
    private static final class None extends Base
        {
        /**Special way to construct distinguished NONE value. */
        private None(Properties p)
            throws PGException
            { super(false, "", p); }

        /**Make distinguished NONE value; only call this once. */
        static None makeNone()
            {
            try {
                final Properties p = new Properties();
                p.put(typeKey, "None"); // Sneakily add a property...
                return(new None(p));
                }
            catch(PGException e)
                { throw new RuntimeException(e.getMessage()); }
            }

        /**Ensure that there is at most one instance of NONE. */
        private Object readResolve()
            { return(NONE); }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 1900944139123394635L;
        }

    /**Single distinguished value meaning `no location info'. */
    public static final Base NONE = None.makeNone();

    /**Just captures time; designed to be derived from.
     * TODO: needs to validate on deserialisation.
     */
    public static class Time extends Location.Base
        {
        public Time(boolean _specific,
                       String prefix,
                       Properties p)
            throws PGException
            {
            // Get the base class initialised.
            super(_specific, prefix, p);

            // Get optional Time prop.
            // TODO: Needs to accept YYYY[/MM[/DD [hh:[mm:[ss[.ddd]]]]]]
            // time format (UTC) plus nn[Y|M|D|h|m|s|ms] error, as
            // well as raw ms values for both.
            LongValueAndBounds t = null;
            try { t = getLongProps(prefix, p, "Time", false); }
            catch(PGException e)
                {
System.err.println("[WARNING: unable to parse Time field in location data; ignoring...]");
                }
            Time = t;
            }

        /**What time does the item record? */
        private final LongValueAndBounds Time;
        /**Get the time value; null means unknown. */
        public final LongValueAndBounds getTime() { return(Time); }

        /**Depends on fields from the base class and the time info, if any. */
        public int hashCode()
            {
            return((super.hashCode() * 19) +
                ((Time != null) ? Time.hashCode() : 1));
            }

        /**Depends on fields from the base class and the time info, if any. */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof Time)) { return(false); }
            final Time other = (Time) obj;

            // Base class fields must match.
            if(!super.equals(obj)) { return(false); }

            // Time fields, if any, must match.
            if(Time == null) { return(other.Time == null); }
            else { return(Time.equals(other.Time)); }
            }

        /**Get value as a String. */
        public String toString()
            {
            if(Time == null) { return(super.toString()); }
            final StringBuilder sb = new StringBuilder();
            sb.append(super.toString());
            sb.append(':');
            sb.append("TIME"); // Not implemented...
            return(sb.toString());
            }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 3902949361757128661L;
        }

    /**Earth-standard coordinates.
     * Eastings and Northings, with optional altitude
     * above mean sea level (in metres).
     * <p>
     * This class is immutable.
     * <p>
     * Time is handled by a base class if any.
     * <p>
     * TODO: needs to validate on deserialisation.
     */
    public static final class Estd extends Location.Time
        {
        /**Build Earth-standard coordinates in WGS84 space.
         */
        public Estd(boolean _specific,
                       String prefix,
                       Properties p)
            throws PGException
            {
            // Get the base class initialised.
            super(_specific, prefix, p);

            // Get mandatory E/N props.
            E = getDoubleProps(prefix, p, "E", true);
            N = getDoubleProps(prefix, p, "N", true);
            // Get optional Alt prop.
            Alt = getDoubleProps(prefix, p, "Alt", false);
            }

        /**Get maximum value for E (positive); minimum is negation of this. */
        public static final int MAX_E = 180;
        /**Get maximum value for N (positive); minimum is negation of this. */
        public static final int MAX_N = 90;

        /**How many degrees East are we? */
        private final DoubleValueAndBounds E;
        /**Get the Easting value; never null. */
        public final DoubleValueAndBounds getE() { return(E); }

        /**How many degrees North are we? */
        private final DoubleValueAndBounds N;
        /**Get the Northing value; never null. */
        public final DoubleValueAndBounds getN() { return(N); }

        /**How many metres above mean sea level are we? */
        private final DoubleValueAndBounds Alt;
        /**Get the altitude in metres; null indicates not known. */
        public final DoubleValueAndBounds getAlt() { return(Alt); }

        /**True if this Location is "bigger" than that of its argument.
         * True if:
         * <ul>
         * <li>This strictly contains the area of the argument,
         * <li>Or, this has a larger error area on the Mercator projection,
         *     ie NErr * EErr.
         * </ul>
         */
        public boolean isLargerThan(Location.Estd other)
            {
            // A completely contained area is smaller.
            if(strictlyContainsArea(other)) { return(true); }
            return(E.error.doubleValue() * N.error.doubleValue() >
                   other.E.error.doubleValue() * other.N.error.doubleValue());
            }

        /**True if this Location and bounding box contains the bounding box of the argument Location.
         * Note that this is reflexive, ie a Location contains its own bounding box.
         */
        public boolean containsArea(final Location.Estd other)
            {
            if(equals(other)) { return(true); }

            final double thisN = getN().value.doubleValue();
            final double thisNErr = getN().error.doubleValue();
            final double otherN = other.getN().value.doubleValue();
            final double otherNErr = other.getN().error.doubleValue();

            // If not in bounds for the North, not contained.
            // Structured to fail if any of the numbers is NaN.
            if(!((otherN - otherNErr >= thisN - thisNErr) && (otherN + otherNErr <= thisN + thisNErr)))
                { return(false); }

            final double thisE = getE().value.doubleValue();
            final double thisEErr = getE().error.doubleValue();
            final double otherE = other.getE().value.doubleValue();
            final double otherEErr = other.getE().error.doubleValue();

            // If not in bounds for the East, not contained.
            // Structured to fail if any of the numbers is NaN.
            if(!((otherE - otherEErr >= thisE - thisEErr) && (otherE + otherEErr <= thisE + thisEErr)))
                { return(false); }

            return(true); // Seems OK.
            }

        /**True if this Location and bounding box strictly contains the bounding box of the argument Location.
         * Note that this is not reflexive, ie a Location not not strictly contain its own bounding box,
         * since the contained item must be smaller,
         * or equivalently contained but not equal.
         */
        public boolean strictlyContainsArea(Location.Estd other)
            {
            if(equals(other)) { return(false); }
            return(containsArea(other));
            }

        /**True if this Location and bounding box contains the centre of the argument Location.
         * Note that this is reflexive, ie a Location contains its own centre.
         */
        public boolean containsCentre(Location.Estd other)
            {
            final double thisN = getN().value.doubleValue();
            final double thisNErr = getN().error.doubleValue();
            final double otherN = other.getN().value.doubleValue();

            // If not in bounds for the North, not contained.
            // Structured to fail if any of the numbers is NaN.
            if(!((otherN >= thisN - thisNErr) && (otherN <= thisN + thisNErr)))
                { return(false); }

            final double thisE = getE().value.doubleValue();
            final double thisEErr = getE().error.doubleValue();
            final double otherE = other.getE().value.doubleValue();

            // If not in bounds for the East, not contained.
            // Structured to fail if any of the numbers is NaN.
            if(!((otherE >= thisE - thisEErr) && (otherE <= thisE + thisEErr)))
                { return(false); }

            return (true); // Seems OK.
            }


        /**Depends on base hash plus E and N values. */
        public int hashCode()
            {
            return((super.hashCode() * 69069) ^
                   (N.hashCode() * 251) ^
                   (E.hashCode()));
            }

        /**Depends on all the elements of this class and base classes.
         */
        public boolean equals(Object obj)
            {
            if(obj == this) { return(true); }
            if(!(obj instanceof Estd)) { return(false); }
            final Estd other = (Estd) obj;

            // Base class fields must match.
            if(!super.equals(obj)) { return(false); }

            // Manditory fields must match.
            if(!N.equals(other.N)) { return(false); }
            if(!E.equals(other.E)) { return(false); }

            // Alt fields, if any, must match.
            if(Alt == null) { return(other.Alt == null); }
            else { return(Alt.equals(other.Alt)); }
            }

        /**Get value as a String. */
        public String toString()
            {
            final StringBuilder sb = new StringBuilder();
            sb.append(super.toString());
            sb.append(':');
            if(E.value.doubleValue() < 0)
                { sb.append(-E.value.doubleValue()); sb.append('W'); }
            else
                { sb.append( E.value.doubleValue()); sb.append('E'); }
            sb.append(',');
            if(N.value.doubleValue() < 0)
                { sb.append(-N.value.doubleValue()); sb.append('S'); }
            else
                { sb.append( N.value.doubleValue()); sb.append('N'); }
            if(Alt != null)
                {
                sb.append(',');
                sb.append(Alt.value.doubleValue());
                sb.append('m');
                }
            // Put in error bounds if needed.
            if((E.error.doubleValue() != 0) ||
               (N.error.doubleValue() != 0) ||
               ((Alt != null) && (Alt.error.doubleValue() != 0)))
                {
                sb.append("{+/-");
                sb.append(E.error.doubleValue());
                sb.append(',');
                sb.append(N.error.doubleValue());
                if(Alt != null) { sb.append(','); sb.append(Alt.error.doubleValue()); }
                sb.append('}');
                }
            return(sb.toString());
            }

        /**"Whole Earth" value; not null. */
        public static final Location.Estd WHOLE_EARTH;
        /**Initialise "Whole Earth" value. */
        static
            {
            final Properties p = new Properties();
            p.setProperty("type", "Estd");
            p.setProperty("E", "0");
            p.setProperty("EErr", String.valueOf(MAX_E));
            p.setProperty("N", "0");
            p.setProperty("NErr", String.valueOf(MAX_N));
            try { WHOLE_EARTH = new Estd(false, "", p); }
            catch(PGException e) { throw new Error("INTERNAL ERROR"); }
            }


        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 3108417206234772844L;
        }
    }
