/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.location;

import org.hd.d.pg2k.svrCore.*;

import java.util.*;

/* TODO: optimise serialised form on the wire (and in cache). */

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 28-Sep-2003
 * Time: 13:45:37
 *
 * Based on pre-PG2K code from ImageInfoCache.
 */

/**Holds a prefix-based lookup map from name to Location.
 * Is immutable.
 * <p>
 * FIXME: needs proper serialisation support such as read/write/validation.
 */
public final class LocationMap implements java.io.Serializable
    {
    /**Are location area prefixes case-sensitive? */
    private static final boolean LOCAREA_PREFIX_CASE_SENSITIVE = false;

    /**Location key prefix for properties (containing trailing dot). */
    private static final String locKeyPrefix = "location.";

    /**Location alias key prefix for properties (containing trailing dot). */
    private static final String locAliasKeyPrefix = "localias.";

    /**Timestamp passed in constructor; never negative.
     * Guaranteed zero if the map is empty.
     */
    public final long timestamp;

    /**Empty, immutable, zero-timestamp, lookup map. */
    public LocationMap() { this(new Properties(), 0); }


    /**Two LocationMaps are equal if the underlying maps are. */
    public boolean equals(final Object obj)
        {
        if(!(obj instanceof LocationMap)) { return(false); }
        return(mapFromNamePrefixToLocation.equals(
            ((LocationMap) obj).mapFromNamePrefixToLocation));
        }

    /**Hash code is that of the underlying map and zero if empty. */
    public int hashCode()
        { return(mapFromNamePrefixToLocation.hashCode()); }


    /**Constructed from non-null Properties containing auxInfo data. */
    public LocationMap(final Properties p, final long timestamp)
        {
        if((timestamp < 0) || (p == null))
            { throw new IllegalArgumentException(); }

        mapFromNamePrefixToLocation = Collections.unmodifiableMap(buildMap(p));
        this.timestamp = ((mapFromNamePrefixToLocation.isEmpty())) ? 0 : timestamp;

        // Size lookup cache in proportion to memory already used
        // and to allow for plenty of negative results to be cached too.
        lookupCache = new MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Location.Base>(0.01f, 1 + 3*mapFromNamePrefixToLocation.size());

        // Store reverse-lookup parameters if present and valid.
        final String rls = p.getProperty(PNAME_REVERSE_LOOKUP_SECTION);
        revLookupSection = (!ExhibitName.validNameInitialComponentSyntax(rls)) ?  null :
            (LOCAREA_PREFIX_CASE_SENSITIVE ? rls : rls.toLowerCase());
        int minWords = -1;
        try { minWords = Integer.parseInt(p.getProperty(PNAME_REVERSE_LOOKUP_MINWORDS), 10); }
        catch(final NumberFormatException e) { }
        revLookupMinWords = minWords;

        // Prepare the reverse map...
        final Map<Location.Base,String> revMap = new HashMap<Location.Base, String>(mapFromNamePrefixToLocation.size());
        final Set<Map.Entry<String,Location.Base>> entries = mapFromNamePrefixToLocation.entrySet();
        for(final Map.Entry<String,Location.Base> entry : entries)
            {
            final String fwdKey = entry.getKey();
            // If we have a preferred reverse-lookup section
            // then discard keys starting with anything else.
            if((revLookupSection != null) && !fwdKey.startsWith(revLookupSection))
                { continue; }
            final String extant = revMap.get(entry.getValue());
            if(extant == null)
                {
                revMap.put(entry.getValue(), fwdKey);
                continue;
                }
System.err.println("WARNING: LocationMap: duplicate keys: " + extant + " and " + fwdKey);
            if(extant.length() > fwdKey.length())
                {
                // Keep the shorter key.
                revMap.put(entry.getValue(), fwdKey);
                }
            }
        mapFromLocationToNamePrefix = Collections.unmodifiableMap(revMap);
        }

    /**Immutable Map from name prefix to Location.Base; never null.
     * Map from "virtual" exhibit name (prefix) to Location object
     * so that we can reuse previously generated objects.
     * The exhibit name key consists of the top-level section/directory
     * and some leading portion of the name;
     * no intermediate directories are stored.
     * <p>
     * A HashMap is preferred for speed.
     */
    private final Map<String,Location.Base> mapFromNamePrefixToLocation;

    /**Reverse section used for reverse matches, or null if none. */
    private final String revLookupSection;

    /**Minimum number of words matched for reverse matches, or non-positive if none. */
    private final int revLookupMinWords;

    /**Immutable reverse map from Location to prefix; may remove some duplicates.
     * Must be reconstructed after deserialisation.
     */
    private final transient Map<Location.Base,String> mapFromLocationToNamePrefix;

    /**Get immutable map from "virtual" prefix to Location.Base; never null.
     * We return the internal immutable map.
     */
    public Map<String,Location.Base> getMapFromNamePrefixToLocation()
        { return(mapFromNamePrefixToLocation); }

    /**Get immutable reverse map from Location.Base to "virtual" prefix; never null.
     * We return the internal immutable map.
     * <p>
     * Note that Location values have to be intern()ed or must compare
     * exactly for equality for this map to be of use.
     */
    public Map<Location.Base,String> getMapFromLocationToNamePrefix()
        { return(mapFromLocationToNamePrefix); }

    /**Property name for section to do reverse lookups in. */
    public static final String PNAME_REVERSE_LOOKUP_SECTION = "_reverse_match_.section";

    /**Property name for minimum words to match for a reverse lookup. */
    public static final String PNAME_REVERSE_LOOKUP_MINWORDS = "_reverse_match_.minWords";

    /**Build the basic map from prefix to Location; never null.
     * The result may be mutable.
     *
     * @param rawProperties  properties to extract the data from; never null
     */
    private static Map<String,Location.Base> buildMap(final Properties rawProperties)
        {
        if(rawProperties == null)
            { throw new IllegalArgumentException(); }

        // We use a HashMap for lookup speed.
        final Map<String,Location.Base> result = new HashMap<String,Location.Base>(1 + rawProperties.size());

        // Create a temporary lookup cache.
        final MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Location.Base> cache = new MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Location.Base>(0.01f, 1+rawProperties.size());

        // First deal with the basic location properties.
        final Enumeration en = rawProperties.propertyNames();
        while(en.hasMoreElements())
            {
            final String key = (String) en.nextElement();
            // If this is not a location key, ignore it.
            if(!key.startsWith(locKeyPrefix)) { continue; }
            // OK, parse it, skipping the prefix.
            final StringTokenizer st = new StringTokenizer(
                key.substring(locKeyPrefix.length()), ".");
            // Abort if not at least two tokens
            // (the location and a key for Location to interpret).
            if(st.countTokens() < 2)
                {
System.err.println("[WARNING: malformed auxInfo key ``" + key + "'': too few components to be valid.]");
                continue;
                }
            // Retrieve the prefix and stop if we already
            // have an entry for it in result.
            final String synthLocPrefix = st.nextToken();
            final String slKey = LOCAREA_PREFIX_CASE_SENSITIVE ?
                synthLocPrefix : synthLocPrefix.toLowerCase();
            if(result.get(slKey) != null) { continue; }
            // If the embedded key does not end with a "-", complain...
            if(!synthLocPrefix.endsWith(ExhibitName.WORD_SEPS))
                {
System.err.println("[WARNING: malformed auxInfo key ``" + key + "'': no trailing '"+ExhibitName.WORD_SEP+"' in "+synthLocPrefix+".]");
                continue;
                }
            // Now attempt to read the (generic) location and
            // put it in the result.
            try {
                final Location.Base lb = Location.Base.buildFromProperties(
                    false, locKeyPrefix + synthLocPrefix + '.', rawProperties);
                result.put(slKey, lb);
                }
            catch(final PGException e)
                {
System.err.println("[WARNING: malformed auxInfo location info for ``" + synthLocPrefix + "'': " + e.getMessage() + ".]");
                continue;
                }
            }


        // Now deal with any location aliases.
        final Enumeration en2 = rawProperties.propertyNames();
        while(en2.hasMoreElements())
            {
            final String key = (String) en2.nextElement();
            // If this is not a location key, ignore it.
            if(!key.startsWith(locAliasKeyPrefix)) { continue; }
            // OK, parse it, ignoring the prefix.
            // If the key does not end with a "-", complain...
            if(!key.endsWith(ExhibitName.WORD_SEPS))
                {
System.err.println("[WARNING: malformed auxInfo key ``" + key + "'': no trailing '"+ExhibitName.WORD_SEP+"'.]");
                continue;
                }
            final StringTokenizer st = new StringTokenizer(
                key.substring(locAliasKeyPrefix.length()), ".");
            // Abort if not exactly one token.
            // (the name to be aliased from).
            if(st.countTokens() != 1)
                {
System.err.println("[WARNING: malformed auxInfo key ``" + key + "'': too few components to be valid.]");
                continue;
                }

            // Retrieve the prefix and stop if we already
            // have an entry for it in result.
            final String synthLocPrefix = st.nextToken();
            final String slKey = LOCAREA_PREFIX_CASE_SENSITIVE ?
                synthLocPrefix : synthLocPrefix.toLowerCase();
            if(result.get(slKey) != null)
                {
System.err.println("[WARNING: duplicate location alias key ``" + key + "''.]");
                continue;
                }

            // Now attempt to look up the pointed-to location
            // and file our duplicate key/link to it.
            final String target = rawProperties.getProperty(key);
            // If the target does not end with a "-", complain...
            if(!target.endsWith(ExhibitName.WORD_SEPS))
                {
System.err.println("[WARNING: malformed auxInfo key target ``" + key + "'': no trailing '"+ExhibitName.WORD_SEP+"'.]");
                continue;
                }
            final String synthName;
            try {
                 synthName =
                    target + (target.endsWith("-") ? "" : "-") +
                    "DHD.jpg"; // Fake suffix to make legal name.
                }
            catch(final IllegalArgumentException e)
                {
e.printStackTrace();
System.err.println("[ERROR: problem with location alias: key ``" + key + "'': " + e.getMessage() + ".]");
                continue;
                }

//System.err.println("[Looking up Location alias key ``" + key + "'' as ``"+ synthName +"''.]");

            // Do lookup (ignoring effect of attribute words).
            final Set<String> noAttrWords = Collections.emptySet();
            final Location.Base lb = _locLookup(result,
                                                synthName,
                                                noAttrWords,
                                                null,
                                                -1,
                                                cache);
            if(lb == Location.NONE)
                {
System.err.println("[ERROR: location alias key ``" + key + "'' does not point to a valid location (is dangling).]");
                continue;
                }

//System.err.println("[Valid Location alias key ``" + key + "''.]");

            // Add duplicate.
            result.put(MemoryTools.intern(slKey), lb);
            }

        return(result);
        }



    /**Lookup of Location by prefix; never null.
     * Returns Location.NONE if nothing found, never null.
     *
     * @param exhibitName  name of exhibit to look up; not null
     * @param allAttrWords  attribute words; not null
     */
    public Location.Base locLookup(final String exhibitName,
                                   final Set<String> allAttrWords)
        {
        return(_locLookup(mapFromNamePrefixToLocation,
                           exhibitName,
                           allAttrWords,
                           revLookupSection,
                           revLookupMinWords,
                           lookupCache));
        }

    /**Private LRU lookup cache; never null.
     * Thread-safe and sized in proportion to the size of our main map,
     * allowing for lots of negative lookups too.
     * <p>
     * Not part of the serialised state of the object.
     * <p>
     * Results are filed against keys that are both:
     * <ul>
     * <li>Exhibit names, for speed ie without needing any string manipulation.
     * <li>Normalised full forward/reverse names stripped of attribute words
     *     and monocased as appropriate.
     * </ul>
     */
    private final transient MemoryTools.SimpleLRUMapAutoSizeForHitRate<String,Location.Base> lookupCache;

    /**Lookup of Location by prefix/suffix in given Map; never null.
     * Returns Location.NONE if nothing found, never null.
     *
     * @param exhibitName  name of exhibit to look up; not null
     * @param allAttrWords  set of all attribute words
     *     (if empty, canonical-key entries are not used); never null
     */
    private static Location.Base _locLookup(final Map<String,Location.Base> nameToLoc,
                                            final String exhibitName,
                                            final Set<String> allAttrWords,
                                            final String revLookupSection,
                                            final int revLookupMinWords,
                                            final MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Location.Base> cache)
        {
        // Try cache lookup on full name for positive and negative results.
        final Location.Base fromCache = cache.get(exhibitName);
        if(fromCache != null) { return(fromCache); }

        // Convert the final portion of the exhibit path
        // to lower-case for comparison.
        // Note that we are discarding attributes here.
        // Note the addition of a trailing '-' to ensure
        // that we match any prefix that should exactly
        // match our main text.
        final String mainWords =
            ExhibitName.getMainWordsComponent(exhibitName, allAttrWords);

        // We need to be careful to lowercase our
        // collection-name key if everything else
        // is being lowercased.
        final String collection = ExhibitName.getCategoryComponent(exhibitName);
        final String collNameKey = (LOCAREA_PREFIX_CASE_SENSITIVE ?
            collection : collection.toLowerCase());

        // Construct the canonical "virtual" name stem.
        final String searchKeyFwd = collNameKey + '/' +
            (LOCAREA_PREFIX_CASE_SENSITIVE ?
                mainWords : mainWords.toLowerCase()) + '-';

        final Location.Base fc1 = cache.get(searchKeyFwd);
        // If we have a negative result then we won't try a fwd lookup.
        final boolean dontTryFwd = Location.NONE.equals(fc1);
        if(!dontTryFwd)
            {
            // If we got a positive response then return it immediately.
            if(fc1 != null) { return(fc1); }

            // fullLocKey is the lookup key for the Location data,
            // or null if we found nothing suitable.
            String fullLocKey = null;

            // We can choose one of two algorithms to search
            // for a longest match in the locations map:
            //
            //  1) Look through all the keys in the database in
            //     turn (in random order).
            //
            //  2) Look for matches for successively shorter
            //     prefixes of our (main text) name.
            //
            // The advantages of (2) are its fixed upper bound
            // on search time (dependent on the name of the
            // item to be looked up) and its simplicity.
//System.err.println("[Searching directly for location by prefix of " + exhibitName.mainText + "; db.size() = " + mapFromNamePrefixToLocation.size() + ".]");

            // Search for successively shorter prefixes,
            // stripping off one tariling word each time round the loop.
            // The first found, if any, is taken to be the best.
            for(String lookupKeyFwd = searchKeyFwd; ; )
                {
                assert(lookupKeyFwd.endsWith(ExhibitName.WORD_SEPS));
                // Now try for the collection-specific.
                final Location.Base loc = nameToLoc.get(lookupKeyFwd);
                if(loc != null)
                    {
                    // Got it!
                    fullLocKey = lookupKeyFwd;
                    break;
                    }

                // Try to find the penultimate word separator
                // and retain it at the end of the new shorter key.
                final int pl = lookupKeyFwd.lastIndexOf(ExhibitName.WORD_SEP, lookupKeyFwd.length()-2);
                if(pl == -1) { break; }
                lookupKeyFwd = lookupKeyFwd.substring(0, pl+1);
                }

            // If we found a match then return it now.
            if(fullLocKey != null)
                {
//System.out.println("[Found location prefix match ``"+fullLocKey+"'' for ``"+exhibitName.mainText+"''.]");
                final Location.Base result = nameToLoc.get(fullLocKey);
                assert(result != null);
                assert(result != Location.NONE);
                // Cache result under full name.
                cache.put(exhibitName, result);
                // Record this positive result against the canonical key.
                cache.put(searchKeyFwd, result);
                return(result);
                }
            // Else cache a negative result for this canonical forward key.
            else
                { cache.put(searchKeyFwd, Location.NONE); }
            }


        // Try reverse lookup...
        // but not for exhibits already in the target section.
        // We will only use trailing capitalised words for reverse lookup.
        if((revLookupMinWords > 0) && (revLookupSection != null) &&
           !revLookupSection.equals(collNameKey))
            {
            // Build the new key in reverse order of main words.
            final StringBuilder sb = new StringBuilder(mainWords.length());
            int words = 0;
            for(final Enumeration wordsEn = ExhibitName.getMainWords(exhibitName, allAttrWords); wordsEn.hasMoreElements(); )
                {
                final String word = (String) wordsEn.nextElement();
                // If this word does not have an initial capital,
                // reset the collected component.
                // Only trailing capitalised words are used.
                final char firstLetter = word.charAt(0);
                if((firstLetter < 'A') || (firstLetter > 'Z'))
                    {
                    sb.setLength(0);
                    words = 0;
                    continue;
                    }

                sb.insert(0, '-');
                sb.insert(0, word);
                ++words;
                }
            final String searchKeyRev = revLookupSection + '/' +
                (LOCAREA_PREFIX_CASE_SENSITIVE ?
                    sb.toString() : sb.toString().toLowerCase());

            // Start off looking for the entire canonical key.
//System.out.println("Reverse key is "+searchKeyRev+" from exhibit " + exhibitName);

            final Location.Base fc2 = cache.get(searchKeyRev);
            // If we have a negative result then we won't try a fwd lookup.
            final boolean dontTryRev = Location.NONE.equals(fc2);
            if(!dontTryRev)
                {
                // If we got a positive response then return it immediately.
                if(fc2 != null) { return(fc2); }

                // fullLocKey is the lookup key for the Location data,
                // or null if we found nothing suitable.
                String fullLocKey = null;

                // Search for successively shorter prefixes,
                // removing one whole word at a time...
                // The first/longest found, if any, is taken to be the best.
                // Stop when there are too few words left
                // to meet our minimum-match-length restriction.
                for(String lookupKeyRev = searchKeyRev; words-- >= revLookupMinWords; )
                    {
                    assert(lookupKeyRev.endsWith(ExhibitName.WORD_SEPS)) : ("Key should end with dash: " + lookupKeyRev);
                    final Location.Base loc = nameToLoc.get(lookupKeyRev);
//System.out.println("Lookup (rev) of "+lookupKeyRev+" yields "+loc+ " (words="+(words+1)+")...");
                    if(loc != null)
                        {
                        // Got it!
                        fullLocKey = lookupKeyRev;
//System.out.println("Reverse lookup found "+lookupKeyRev+" for exhibit " + exhibitName);
                        break;
                        }

                    // Try to find the penultimate word separator
                    // and retain it at the end of the new shorter key.
                    final int pl = lookupKeyRev.lastIndexOf(ExhibitName.WORD_SEP, lookupKeyRev.length()-2);
                    if(pl == -1) { break; }
                    lookupKeyRev = lookupKeyRev.substring(0, pl+1);
                    }

                // If we found a match then return it now.
                if(fullLocKey != null)
                    {
                    final Location.Base result = nameToLoc.get(fullLocKey);
                    assert(result != null);
                    assert(result != Location.NONE);
                    // Cache result under full name.
                    cache.put(exhibitName, result);
                    // Record this positive result against the canonical key.
                    cache.put(searchKeyRev, result);
                    return(result);
                    }
                // Else cache a negative result for this canonical key.
                else
                    { cache.put(searchKeyRev, Location.NONE); }
                }
            }

        // Cache negative result under full name.
        cache.put(exhibitName, Location.NONE);
        return(Location.NONE);
        }

    /**Returns true if the map is completely empty, eg as built by the default constructor.
     */
    public boolean isEmpty()
        { return(mapFromNamePrefixToLocation.isEmpty()); }






//    /**Retrieves the built-in LocationMap from properties; never null.
//     * This is a backup/fallback, checked in with the code.
//     * @return  LocationMap, possibly empty but not null
//     */
//    private static LocationMap _getBuiltInLocationMap()
//        {
//        try
//            {
//            // Retrieve the location-map bundle...
//            final ResourceBundle lp = ResourceBundle.getBundle(CoreConsts.LOCDB_PROPS_NAME);
//
//            // Copy into a simple Peroperties object.
//            final Properties p = new Properties();
//            final Enumeration<String> en = lp.getKeys();
//            while(en.hasMoreElements())
//                {
//                final String key = en.nextElement();
//                p.setProperty(key, lp.getString(key));
//                }
//
//            return(new LocationMap(p, 0));
//            }
//        catch(final Exception e)
//            {
//            System.err.println("ERROR: could not load built-in location map (using empty one) from: " + CoreConsts.LOCDB_PROPS_NAME);
//            e.printStackTrace();
//            return(new LocationMap(new Properties(), 0));
//            }
//        }
//
//    /**Built-in, possibly vestigial/fallback-only, location map. */
//    private static final LocationMap builtInLocationMap = _getBuiltInLocationMap();
//
//    /**Get public view of built-in location map instance. */
//    public static LocationMap getBuiltInLocationMap()
//        { return(builtInLocationMap); }





    /**Serialisation UID. */
    private static final long serialVersionUID=-8356330012477285265L;
    }
