/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.mediahandler;

import org.hd.d.pg2k.svrCore.MIME.AbstractImageHandler;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;

import java.awt.*;
import java.io.*;

/**Default handler for gif file (MIME type image/gif).
 */
public final class gif extends AbstractImageHandler
    {
    /**Get internal type of ExhibitMIME type; never null. */
    public ExhibitMIME.ExhibitTypeParameters getExhibitType()
        { return(ExhibitMIME.getParamsByType(ExhibitMIME.ET_GIF)); }

    /**Get dimensions X and Y of a GIF exhibit, else null if dimensions cannot be computed.
     * This reads the opening bytes of the exhibit directly to extract the
     * dimensions, rather than loading and interpretting the whole image.
     * <p>
     * This input stream must be of the correct type, eg the magic number
     * already checked.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     *
     * @param is        the exhibit as a binary data stream
     *
     * @throws IOException  in case of problems with corrupt data
     *     (or a broken exhibit)
     */
    @Override public Dimension get2DImageDimensions(final InputStream is)
        throws IOException
        {
        final DataInputStream dis = new DataInputStream(is);
        byte b[] = new byte[10]; // Exactly enough for what we want.
        dis.readFully(b);
        final Dimension result = new Dimension();
        result.width  = (b[6]&0xff) + ((b[7]&0xff)<<8);
        result.height = (b[8]&0xff) + ((b[9]&0xff)<<8);
        return(result); // DONE!
        }

    /**Target bits-per-pixel for thumbnail images; strictly positive.
     * Does not need to leave all images perfect, but most should be OK.
     * Note that we'll be limited from above by the absolute size limits
     * of thumbnails in general.
     * <p>
     * A value in the range 4--8 is probably good.
     */
    private static final int TARGET_BPP = 6;

    /**Estimated file overhead constant, ie what would a 0x0 file size be; non-negative.
     * This factor helps avoid penalising small images by
     * having the overhead eat into the bits that should be being used
     * by the image.
     * <p>
     * One byte less than a 1x1 white image produced by PSP7.
     */
    private static final int FILE_OVERHEAD = 76;

    /**Estimated (maximum) bytes per pixel of a decoded image in memory.
     * GIFs use a max-256-entry palette,
     * so a reasonable in-memory representation will use one byte per pixel
     * or less.
     */
    private static final int EST_BYTES_PER_IMAGE_PIXEL_IN_MEMORY = 1;

    /**Maximum quality allowed in a thumbnail; taken to be colour depth.
     * Assume that this is the same as the maximum number of bits in memory
     * we expect an image to take.
     */
    private static final int maxThumbnailQuality =
        8 * EST_BYTES_PER_IMAGE_PIXEL_IN_MEMORY;

    /**Normal/initial quality for a thumbnail.
     * This is reasonable for most images, though erring slightly on
     * the side of retention of image quality at cost of download speed.
     */
    private static final int normalThumbnailQuality = TARGET_BPP;

    /**Maximum quality allowed in a thumbnail; taken to be colour depth.
     * Must be positive and can be as little as monochrome, ie one bit.
     */
    private static final int minThumbnailQuality = 1;

    /**Gather together thumbnail parameters. */
    private static final AbstractImageHandler.ThumbnailParams thumbnailParams =
        new ThumbnailParams(EST_BYTES_PER_IMAGE_PIXEL_IN_MEMORY,
            minThumbnailQuality,
            maxThumbnailQuality,
            normalThumbnailQuality,
            TARGET_BPP,
            FILE_OVERHEAD);

    /**Get ThumbnailParameters for GIF handler; non-null. */
    @Override public AbstractImageHandler.ThumbnailParams getThumbnailParams()
        { return(thumbnailParams); }

    // Inherit javadoc.
    @Override protected int _reduceColoursQualityThreshold() { return(8); }
    }
