/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.mediahandler;

import com.drew.metadata.Directory;
import com.drew.metadata.Tag;
import com.drew.metadata.exif.ExifDirectory;
import com.drew.metadata.exif.ExifReader;
import com.drew.metadata.iptc.IptcDirectory;
import com.drew.metadata.iptc.IptcReader;
import org.hd.d.pg2k.svrCore.MIME.AbstractImageHandler;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.TextUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.SortedMap;
import java.util.TreeMap;

/**Default handler for jpg file (MIME type image/jpeg).
 */
public final class jpg extends AbstractImageHandler
    {
    /** Get internal type of ExhibitMIME type; never null. */
    public ExhibitMIME.ExhibitTypeParameters getExhibitType()
        { return(ExhibitMIME.getParamsByType(ExhibitMIME.ET_JPEG)); }

    /**Get dimensions X and Y of a JPEG exhibit, else null if dimensions cannot be computed.
     * This reads the opening bytes of the exhibit directly to extract the
     * dimensions, rather than loading and interpretting the whole image.
     * <p>
     * This input stream must be of the correct type, eg the magic number
     * already checked.
     * <p>
     * This <strong>does not close its input stream</strong> when done.
     *
     * @param is        the exhibit as a binary data stream
     *
     * @throws IOException  in case of problems with corrupt data
     *     (or a broken exhibit)
     */
    public java.awt.Dimension get2DImageDimensions(final InputStream is)
        throws IOException
        {
        // Start on the JFIF header (offset 2),
        // and step through skipping each block/frame
        // by length.
        // Stop when we recognise one that we can use...
        // (Reading chucnks of 512 bytes seems to be able to cover
        // most JPEG files in one read for fetching the dimensions,
        // and is probably efficient for reading from underlying
        // 512-bytes-per-sector disc media.)
        final DataInputStream dis = new DataInputStream(new BufferedInputStream(is, 512));
        dis.skipBytes(2); // Skip SOI header...
        int nextByte;
        while((nextByte = dis.read()) != -1) // Until EOF.
            {
            if(nextByte != 0xff) { throw new IOException("expected JPEG frame marker byte 0xff"); }
            // Apparently any amount of 0xffs can be inserted
            // as padding, so quietly eat them...
            while((nextByte = dis.read()) == 0xff) { }
            final int frameType = nextByte;
//System.err.println(" [Got frame type 0x" +Integer.toHexString(frameType)+".]");
            final int frameLength = dis.readUnsignedShort();
            if(frameLength < 2) { throw new IOException("unexpected JPEG header frame length "+frameLength); }
            // OK, we are interested in frame types
            // 0xc0 to 0xc3 inclusive for dimensions...
            if((frameType >= 0xc0) && (frameType <= 0xc3))
                {
                // Skip sample precision...
                if(dis.read() == -1) { break; } // Failed.
                final Dimension result = new Dimension();
                result.height = dis.readUnsignedShort();
                result.width  = dis.readUnsignedShort();
//System.err.println(" [GOT DIMENSIONS! "+result.width+"x"+result.height+".]");
                return(result); // GOT IT!
                }
            // Now skip rest of frame;
            // 2 bytes from this frame have already been read...
            dis.skipBytes(frameLength - 2);
            }
        throw new EOFException("ran out of JPEG file without finding dimensions");
        }

    /**Returns true if this can make thumbnail/sample exhibits for this type.
     * Returns true because we can make JPEG thumbnails.
     */
    public boolean canMakeThumbnails()
        { return(true); }


    /**Target bits-per-pixel for thumbnail images; strictly positive.
     * Does not need to leave all images perfect, but most should be OK.
     * Note that we'll be limited from above by the absolute size limits
     * of thumbnails in general.
     * <p>
     * A value in the range 2--4 is probably good.
     */
    private static final int TARGET_BPP = 3;

    /**Estimated JPEG overhead constant, ie what would a 0x0 JPEG file size be; non-negative.
     * This factor helps avoid penalising small images by
     * having the overhead eat into the bits that should be being used
     * by the image.
     * <p>
     * One byte less than a 1x1 white image produced by PSP7.
     */
    private static final int FILE_OVERHEAD = 284;

    /**Max quality allowed in a JPEG montage or thumbnail.
     * We don't go to 100%, since that might inflate the image and
     * gives hardly any image improvement.
     */
    private static final int maxJPEGQuality = 95;

    /**Normal/initial quality for a JPEG thumbnail.
     * This is reasonable for most images, though erring slightly on
     * the side of retention of image quality at cost of download speed.
     * <p>
     * A quality of 75% is the default for the JPEG encoder.
     * <p>
     * A value from 50--90 is probably good; 85 has worked well in the past.
     */
    private static final int normalThumbnailJPEGQuality = 75;

    /**Min quality allowed in a JPEG montage or thumbnail.
     * We don't go to 0%, since the image would be unusable.
     */
    private static final int minJPEGQuality = 15;

    /**Estimated (maximum) bytes per pixel of a decoded image in memory.
     * This assumes that a 24-bit RGB or ARGB format is the most expensive
     * we will run across.  (The other main possibility is an 8-bit greyscale.)
     */
    private static final int EST_BYTES_PER_IMAGE_PIXEL_IN_MEMORY = 4;


    /**Gather together thumbnail parameters. */
    private static final AbstractImageHandler.ThumbnailParams thumbnailParams =
        new ThumbnailParams(EST_BYTES_PER_IMAGE_PIXEL_IN_MEMORY,
            minJPEGQuality,
            maxJPEGQuality,
            normalThumbnailJPEGQuality,
            TARGET_BPP,
            FILE_OVERHEAD);

    /**Get ThumbnailParameters for JPEG handler.
     */
    public AbstractImageHandler.ThumbnailParams getThumbnailParams()
        { return(thumbnailParams); }


    /**Type of EXIF segment in JPEG. */
    private static final int EXIF = 0xE1;

    /**Type of IPTC segment in JPEG. */
    private static final int IPTC = 0xED;

    /**Longest value (chars) we will accept without discarding; strictly positive. */
    private static final int MAX_VALUE_CHARS = 128;

    /**Get specific (image) metadata for one particular image type; null if none.
     * This does not have to be provided, but if so enables such things as
     * EXIF and IPTC extraction from JPEG images.
     * <p>
     * We trim and normalise values shown a little, and discard any huge values,
     * since this data is intended for a human to read.  If they need more detail
     * then they can download the image and open it in a decent Photoshop-like app.
     * <p>
     * This may be a mix of stream and image/frame metadata.
     * 
     * @param imageMetadata  imageio portable metadata for this exhibit; never null
     * @param exhibitName  full exhibit name (primarily to help with debugging); never null
     */
    protected Node extractSpecificImageMetaData(final IIOMetadata imageMetadata,
                                                final String exhibitName)
        {
        final String nativeMetadataFormatName = imageMetadata.getNativeMetadataFormatName();
        if(nativeMetadataFormatName == null)
            { return(null); /* Cannot get at native metadata. */ }

        Node result = null;

        try
            {
            // Create a new document with our metadata in...
            final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = dbf.newDocumentBuilder();
            final Document doc = db.newDocument();

            // Create a provisional root node,
            // but only assign it to the result (ie to get a non-null result)
            // when we know that there is some real data in there.
            final Element root = doc.createElement("native");
            doc.appendChild(root);

            // Walk the tree (breadth-first)
            // looking for the EXIF and IPTC "unknown" tags.
            // Upon finding one, convert it into a node hierarchy
            // and attach it to our top-level node.
            final IIOMetadataNode topNode = (IIOMetadataNode) imageMetadata.getAsTree(nativeMetadataFormatName);
            Queue<Node> nodesToCheck = new LinkedList<Node>();
            for(nodesToCheck.add(topNode); !nodesToCheck.isEmpty(); )
                {
                final Node n = nodesToCheck.remove();

                // If the current node has children
                // then add them to the queue for later processing.
                for(Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
                    { nodesToCheck.add(child); }

                // Check if this node represents one of the tags (eg EXIF) that we want.
                // This will be labelled as "unknown" in the native metadata extracted by IIO.
                if(n.getNodeName() == "unknown")
                    {
                    try
                        {
                        final int tagVal = Integer.parseInt(n.getAttributes().getNamedItem("MarkerTag").getNodeValue());
                        Directory d = null; // Collect metadata directory if possible.
                        String tagName = null;

                        switch(tagVal)
                            {
                            case EXIF:
                                {
                                final byte[] data = (byte[]) (((IIOMetadataNode) n).getUserObject());
                                d = (new ExifReader(data)).extract().getDirectory(ExifDirectory.class);
                                tagName = "EXIF";
                                break;
                                }

                            case IPTC:
                                {
                                final byte[] data = (byte[]) (((IIOMetadataNode) n).getUserObject());
                                d = (new IptcReader(data)).extract().getDirectory(IptcDirectory.class);
                                tagName = "IPTC";
                                break;
                                }

                            default:
                                // Unexpected tag/data; ignore.
                                break;
                            }

                        // If we've got a real directory
                        // then iterate through the tags,
                        // turning any non-empty values into nodes.
                        // We sort by the tag name.
                        if((d != null) && (d.getTagCount() > 0))
                            {
                            final Element item = doc.createElement(tagName);
                            root.appendChild(item);

                            // Map from name to value...
                            final SortedMap<String,String> tags = new TreeMap<String,String>();

                            // Turn tags and values into sub-nodes.
                            for(final Iterator tagIter = d.getTagIterator(); tagIter.hasNext(); )
                                {
                                try
                                    {
                                    final Tag tag = (Tag) tagIter.next();
                                    final String name = tag.getTagName().trim();

                                    // Ignore null (or empty after trimming) values...
                                    String desc = tag.getDescription();
                                    if(desc == null) { continue; }
                                    desc = desc.trim();
                                    if(desc.length() == 0) { continue; }

                                    // Discard excessively-long values,
                                    // but leave the tag to alert the reader
                                    // that the data exists.
                                    // Also log an informational message.
                                    if(desc.length() > MAX_VALUE_CHARS)
                                        {
                                        System.out.println("[INFO: long tag value ("+desc.length()+" chars) discarded (name="+name+") for exhibit: "+exhibitName+".]");
                                        desc = "...";
                                        }

                                    // Good tag, so retain, using trimmed description.
                                    tags.put(name, desc);
                                    }
                                catch(final RuntimeException e)
                                    {
                                    // Attempt to ignore isolated errors.
                                    System.err.println("Unexpected problem (attempting to ignore) while examining JPEG native metadata for exhibit "+exhibitName);
                                    e.printStackTrace();
                                    }
                                }

                            // If we got some valid tags then add them to the DOM tree.
                            if(!tags.isEmpty())
                                {
                                result = root; // We have found some real metadata to return...

                                for(final String key : tags.keySet())
                                    {
                                    // Create element with name and value attributes.
                                    final Element tv = doc.createElement("tag");
                                    tv.setAttribute("name", key); // Never contains metadata.
                                    tv.setAttribute("value", TextUtils.escapeHTMLMetaChars(tags.get(key)));
                                    item.appendChild(tv);            
                                    }
                                }
                            }
                        }
                    catch(final Throwable e)
                        {
                        // Unexpected problem examining metadata.
                        // Just note the problem and continue.
                        // We can get nasty Errors thrown for some broken exhibits.
                        System.err.println("Unexpected problem while examining JPEG native metadata for exhibit "+exhibitName);
                        e.printStackTrace();
                        }
                    }
                }
            }
        catch(final ParserConfigurationException e)
            {
            e.printStackTrace();
            return(null); // Give up if we cannot get a parser...
            }

        return(result);
        }
    }
