/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.mediahandler;

import org.hd.d.pg2k.svrCore.MIME.AbstractHandler;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.TextUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.InputStream;

/**Basic handler for MIDI file (MIME type audio/midi).
 */
public final class mid extends AbstractHandler
    {
    /**Get internal type of ExhibitMIME type; never null. */
    public ExhibitMIME.ExhibitTypeParameters getExhibitType()
        { return(ExhibitMIME.getParamsByType(ExhibitMIME.ET_MIDI)); }

    /**Get format of audio data; null if none or format not usable/parseable.
     * This may require the stream to be seekable, at least with mark() and reset().
     */
    protected Sequence getSequence(final InputStream is)
        {
        try { return(MidiSystem.getSequence(is)); }

        // Dump any (expected) errors to the log/console and return null.
        catch(final javax.sound.midi.InvalidMidiDataException e)
            {
            e.printStackTrace();
            }
        catch(IOException e)
            {
            e.printStackTrace();
            }

        return(null); // Could not deduce file format.
        }


    /**Gets all available exhibit metadata as a single XML DOM tree; null if none.
     * We do not (yet) have a schema for this and may never do so,
     * since its structure and content will depend on various external sources
     * and parts of (for example) the ImageIO and JAI subsystems.
     * <p>
     * This implementation synthesises a metadata tree from the AudioFileFormat data,
     * if any.
     *
     * @param is input stream; never null
     * @return top-level node "metadata" with captured metadata beneath, else null
     */
    public Node getMetadata(final InputStream is, String exhibitName)
        {
        final Sequence seq = getSequence(is);
        if(seq == null) { return(null); }

        // Result: will be set non-null if we encounter any usable data.
        Node result = null;

        try
            {
            final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = dbf.newDocumentBuilder();
            final Document doc = db.newDocument();

            // Create a provisional root node,
            // but only assign it to the result (ie to get a non-null result)
            // when we know that there is some real data in there.
            final Element root = doc.createElement(TAG_NAME_METADATA_TOP);
            doc.appendChild(root);

            // The "audio" node contains data about the data encoding format such as timebase.
            final Element audio = doc.createElement("audio");
            root.appendChild(audio);
            result = root; // Got some real data...

            final int nTracks = seq.getTracks().length;
            final Element trackNode = doc.createElement("tracks");
            trackNode.setAttribute("value", String.valueOf(nTracks));
            audio.appendChild(trackNode);

            final float divisionType = seq.getDivisionType();
            final Element divisionTypeNode = doc.createElement("divisionType");
            divisionTypeNode.setAttribute("value", TextUtils.escapeHTMLMetaChars(String.valueOf(divisionType)));
            audio.appendChild(divisionTypeNode);

            final long usLength = seq.getMicrosecondLength();
            final Element usLengthNode = doc.createElement("microsecondLength");
            usLengthNode.setAttribute("value", String.valueOf(usLength));
            usLengthNode.setAttribute("unit", "us");
            audio.appendChild(usLengthNode);

            final int resolution = seq.getResolution();
            final Element resolutionNode = doc.createElement("resolution");
            resolutionNode.setAttribute("value", String.valueOf(resolution));
            resolutionNode.setAttribute("unit", (divisionType == Sequence.PPQ) ? "ticks per beat" : "ticks per frame");
            audio.appendChild(resolutionNode);

            final long tickLength = seq.getTickLength();
            final Element tickLengthNode = doc.createElement("tickLength");
            tickLengthNode.setAttribute("value", String.valueOf(tickLength));
            tickLengthNode.setAttribute("unit", "ticks");
            audio.appendChild(tickLengthNode);
            }
        catch(final ParserConfigurationException e)
            {
            e.printStackTrace();
            return(null); // Cannot build document...
            }

        return(result);
        }
    }
