/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.mediahandler;

import org.hd.d.pg2k.svrCore.MIME.AbstractHandler;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.awt.*;
import java.io.DataInputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.InflaterInputStream;

/**Basic handler for SWF file (MacroMedia Shockwave Flash).
 */
public final class swf extends AbstractHandler
    {
    /**Get internal type of ExhibitMIME type; never null. */
    public ExhibitMIME.ExhibitTypeParameters getExhibitType()
        { return(ExhibitMIME.getParamsByType(ExhibitMIME.ET_SWF)); }


    /**TWIPS to pixels scaling factor assuming 72dpi screen resolution. */
    private static final int TWIPS_TO_PIXELS = 20;

    /**Efficiently get X,Y pixel dimensions of SWF exhibit, else null if dimensions cannot be computed.
     * This <strong>does not close its input stream</strong> when done.
     * <p>
     * This should work efficiently, by directly reading (opening) bytes
     * of the exhibit passed as an input stream.
     * <p>
     * This will only work correctly if the exhibit is of the correct type,
     * eg its magic number must already have been tested.
     *
     * @param is the exhibit as a binary data stream
     * @throws java.io.IOException in case of problems with corrupt data
     *                             (or a broken exhibit)
     */
    @Override public Dimension get2DImageDimensions(final InputStream is)
        throws IOException
        {
        // Don't allow the underlying stream to be closed.
        final InputStream isWrapped = new FilterInputStream(is){
            /**Ignore any attempt to close the stream. */
            @Override public void close() { }
            };
        final SWFMetaData swfMetaData = new SWFMetaData(isWrapped);
        final Dimension result = new Dimension(
            swfMetaData.xmax / TWIPS_TO_PIXELS,
            swfMetaData.ymax / TWIPS_TO_PIXELS);
        return(result);
        }

    /**Immutable meta-data for SWF exhibit. */
    public static final class SWFMetaData
        {
        /**SWF version. */
        public final int version;
        public final long length;
        /**True if file in compressed format. */
        public final boolean compressed;
        public final int xmin;
        public final int xmax;
        public final int ymin;
        public final int ymax;
        /**Frame rate. */
        public final float rate;
        /**Frame count. */
        public final int count;

        /**Parse meta-data from an input stream. */
        public SWFMetaData(final InputStream is)
            throws IOException
            {
            InflaterInputStream inflaterInputStream = null;
            try
                {
                // Starts with CWS for compressed format, FWS for uncompressed.
                DataInputStream dis = new DataInputStream(is);
                final int firstChar = dis.readUnsignedByte();
                dis.skip(2); // Skip the "WS"...

                version = dis.readUnsignedByte();
                length = 0x0ffffffffL & dis.readInt();

                compressed = (version >= 6) && ('C' == firstChar);
                if(compressed)
                    {
                    // Start reading from the decompressed stream instead.
                    inflaterInputStream = new InflaterInputStream(dis);
                    dis = new DataInputStream(inflaterInputStream);
                    }

                long sizeBits = dis.readLong();
                final int bitsPerLenValue = (int) (sizeBits >>> 59);
                sizeBits <<= 5;

                xmin = (int) (sizeBits >> (64 - bitsPerLenValue));
                sizeBits <<= bitsPerLenValue;
                xmax = (int) (sizeBits >> (64 - bitsPerLenValue));
                sizeBits <<= bitsPerLenValue;
                ymin = (int) (sizeBits >> (64 - bitsPerLenValue));
                sizeBits <<= bitsPerLenValue;
                ymax = (int) (sizeBits >> (64 - bitsPerLenValue));
                sizeBits <<= bitsPerLenValue;

                rate = dis.readUnsignedShort() / 256f;
                count = dis.readUnsignedShort();
                }
            finally
                {
                // Ensure that we free any ZIP resources.
                if(inflaterInputStream != null)
                    { inflaterInputStream.close(); }
                }
            }
        }

    /**Gets all available exhibit metadata as a single XML DOM tree; null if none.
     * We do not (yet) have a schema for this and may never do so,
     * since its structure and content will depend on various external sources
     * and parts of (for example) the ImageIO and JAI subsystems.
     * <p>
     * This implementation synthesises a metadata tree from the AudioFileFormat data,
     * if any.
     *
     * @param is  input stream; never null
     * @return top-level node "metadata" with captured metadata beneath, else null
     */
    public Node getMetadata(final InputStream is, String exhibitName)
        {
        // Result: will be set non-null if we encounter any usable data.
        Node result = null;

        try
            {
            final SWFMetaData md = new SWFMetaData(is);

            final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = dbf.newDocumentBuilder();
            final Document doc = db.newDocument();

            // Create a provisional root node,
            // but only assign it to the result (ie to get a non-null result)
            // when we know that there is some real data in there.
            final Element root = doc.createElement(TAG_NAME_METADATA_TOP);
            doc.appendChild(root);


            // The "misc" node contains misc items.
            final Element misc = doc.createElement("misc");
            root.appendChild(misc);
            result = root; // Got some real data...

            final Element compNode = doc.createElement("compressed");
            compNode.setAttribute("value", String.valueOf(md.compressed));
            misc.appendChild(compNode);

            final Element versionNode = doc.createElement("version");
            versionNode.setAttribute("value", String.valueOf(md.version));
            misc.appendChild(versionNode);

            final Element lengthNode = doc.createElement("length");
            lengthNode.setAttribute("value", String.valueOf(md.length));
            misc.appendChild(lengthNode);


            // The "audio" node contains data on any soundtrack.
            final Element audio = doc.createElement("audio");
            root.appendChild(audio);
            result = root; // Got some real data...


            // The "video" node contains data on any imagery.
            final Element video = doc.createElement("video");
            root.appendChild(video);
            result = root; // Got some real data...

            final Element xminNode = doc.createElement("xmin");
            xminNode.setAttribute("value", String.valueOf(md.xmin));
            video.appendChild(xminNode);

            final Element xmaxNode = doc.createElement("xmax");
            xmaxNode.setAttribute("value", String.valueOf(md.xmax));
            video.appendChild(xmaxNode);

            final Element yminNode = doc.createElement("ymin");
            yminNode.setAttribute("value", String.valueOf(md.ymin));
            video.appendChild(yminNode);

            final Element ymaxNode = doc.createElement("ymax");
            ymaxNode.setAttribute("value", String.valueOf(md.ymax));
            video.appendChild(ymaxNode);

            final Element rateNode = doc.createElement("fps");
            rateNode.setAttribute("value", String.valueOf(md.rate));
            video.appendChild(rateNode);

            final Element countNode = doc.createElement("count");
            countNode.setAttribute("value", String.valueOf(md.count));
            video.appendChild(countNode);
            }
        catch(final ParserConfigurationException e)
            {
            e.printStackTrace();
            return(null); // Cannot build document...
            }
        catch(final IOException e)
            {
            e.printStackTrace();
            return(null); // Cannot parse exhibit file...
            }

        return(result);
        }
    }
