/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.props;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.regex.Pattern;

/**This holds run-time settable properties true the master and all mirrors.
 * It is immutable, and can be persisted or sent over RMI-IIOP.
 * <p>
 * All values are accessed through getXXX() methods
 * (this object can be used as a bean)
 * to allow us to always constrain values to sensible limits
 * (and, for example, to patch up up any values that have `gone funny'
 * in serialization because (say) because the class version changed
 * and fields were added)
 * and to allow for local overrides from properties.
 * This means that some of the getXXXX() methods have to be synchronized.
 * <p>
 * We also extensively check object state at construction and
 * deserialisation; and we try to impose explicit or implicit limits
 * on the amount of space that an instance of this object can consume,
 * especially because at a transition between an old version and a new one
 * there may be multiple instances floating about in memory.
 * <p>
 * We don't retain the original raw properties internally, but parse them at
 * construction time, to do most of the expensive work once if possible,
 * and to keep the serialised form of the object small.
 * Defaults are set at parse/constuction time,
 * whereas limits are imposed at getXXX() time which makes for
 * a lot of robustness if the object is damaged in transit or if
 * the class details change slightly.
 * <p>
 * The version manufactured with a default contructor has a zero timestamp.
 */
public final class GenProps implements Serializable, ObjectInputValidation
    {
    /**Our serialisation version. */
    private static final long serialVersionUID = 8715644348905699663L;

    /**Deserialise. */
    private void readObject(final ObjectInputStream in)
        throws IOException, ClassNotFoundException
        {
        in.defaultReadObject();

        // Take defensive immutable copy of gen, if present, else make it empty.
        if(gen == null)
            { gen = Collections.emptyMap(); }
        else
            { gen = Collections.unmodifiableMap(new HashMap<String,String>(gen)); }

        // Take defensive copy of simpleAds[] if present, and re-sort.
        if(simpleAds != null)
            {
            simpleAds = (SimpleAd[]) simpleAds.clone();
            Arrays.sort(simpleAds);
            }

        // Take defensive copy of authDB if present.
        if(authDB != null)
            { authDB = Collections.unmodifiableSortedMap(new TreeMap<String, AuthData>(authDB)); }

        // Take defensive copy of popWeights if present.
        if(popWeights != null)
            { popWeights = Collections.unmodifiableSortedMap(new TreeMap<String, Byte>(popWeights)); }


        // Take defensive copies of hotlink hot/cold sets and DNSRBLs,
        // but renormalise hostnames and truncate to size if need be.
        if(hotLinkAllowHosts != null)
            { hotLinkAllowHosts = _normaliseHostList(hotLinkAllowHosts, MAX_ALLOW_DISALLOW_HOTLINK_HOSTS); }
        if(hotLinkDisallowHosts != null)
            { hotLinkDisallowHosts = _normaliseHostList(hotLinkDisallowHosts, MAX_ALLOW_DISALLOW_HOTLINK_HOSTS); }
        if(DNSRBLs != null)
            { DNSRBLs = _normaliseHostList(DNSRBLs, MAX_DNSRBLS); }


        validateObject(); // Validate state immediately.
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Check that all components are sane and safe.
        if(timestamp < 0)
            { throw new InvalidObjectException("bad object: timestamp < 0"); }

        // Verify generic properties.
        if(gen == null)
            { throw new InvalidObjectException("bad object: gen == null"); }
        if(gen.size() > MAX_GEN_PROPS)
            { throw new InvalidObjectException("bad object: gen too large"); }
        for(final Object kO : gen.keySet())
            {
            if(!(kO instanceof String))
                { throw new InvalidObjectException("bad object: gen key not a String"); }
            final String k = (String) kO;
            if(!isSafeGenPropValue(k))
                { throw new InvalidObjectException("bad object: gen key not valid"); }

            final Object vO = gen.get(k);
            if(!(vO instanceof String))
                { throw new InvalidObjectException("bad object: gen value not a String"); }
            final String v = (String) vO;
            if(!isSafeGenPropValue(v))
                { throw new InvalidObjectException("bad object: gen value not valid"); }
            }

        // Check that HTML meta-header text is still printable ASCII excluding ".
        if(!isSafeHTMLMetaHeaderString(WEBSVR_META_KEYWORDS))
            { throw new InvalidObjectException("bad object: unsafe " + PNAME_WEBSVR_META_KEYWORDS); }
        if(!isSafeHTMLMetaHeaderString(WEBSVR_META_DESCRIPTION))
            { throw new InvalidObjectException("bad object: unsafe " + PNAME_WEBSVR_META_DESCRIPTION); }
        // Check that background image name, if set, is syntactically valid.
        // We also check that it is a JPEG image.
        if(WEBSVR_BG_IMAGE != null)
            {
            if(!ExhibitName.validNameSyntax(WEBSVR_BG_IMAGE))
                { throw new InvalidObjectException("bad object: unsafe name " + PNAME_WEBSVR_BG_IMAGE); }
            final ExhibitMIME.ExhibitTypeParameters et =
                ExhibitMIME.getInputFileType(WEBSVR_BG_IMAGE);
            if((et == null) || (et.type != ExhibitMIME.ET_JPEG))
                { throw new InvalidObjectException("bad object: unsafe exhibit type " + PNAME_WEBSVR_BG_IMAGE); }
            }

        // Check authDB for consistency.
        if(authDB != null)
            {
            if((authDB.size() == 0) || (authDB.size() > MAX_AUTH_ENTRIES))
                { throw new InvalidObjectException("bad object: bad-length authDB"); }

            // Check all entries are of the correct type (AuthData).
            for(final Iterator it = authDB.keySet().iterator(); it.hasNext(); )
                {
                final Object key = it.next();
                if(!(key instanceof String))
                    { throw new InvalidObjectException("bad object: bad key in authDB"); }
                if(!(authDB.get(key) instanceof AuthData))
                    { throw new InvalidObjectException("bad object: bad value in authDB"); }
                }
            }

        // Check "goodness"/popularity weights for consistency.
        if(popWeights != null)
            {
            if((popWeights.size() == 0) || (popWeights.size() > MAX_POPWT_ENTRIES))
                { throw new InvalidObjectException("bad object: bad-length popWeights"); }

            // Check all entries are of the correct type (AuthData).
            for(final Iterator<String> it = popWeights.keySet().iterator(); it.hasNext(); )
                {
                final String key = it.next();
                if(!ExhibitName.validAuthorSyntax(key) &&
                   !ExhibitName.validAttributeWord(key) &&
                   (ExhibitMIME.isValidInputExhibitNameExtension(key) == null))
                    { throw new InvalidObjectException("bad object: bad key in popWeights"); }
                final Byte val = popWeights.get(key);
                if((val == null) ||
                   (val.byteValue() < MIN_POPWT_VAL) ||
                   (val.byteValue() > MAX_POPWT_VAL))
                    { throw new InvalidObjectException("bad object: bad value in popWeights"); }
                }
            }

        // Check ads for consistency.
        // Don't waste space with zero-length simpleAds.
        if((simpleAds != null) &&
            ((simpleAds.length == 0) || (simpleAds.length > MAX_SIMPLE_ADS)))
            { throw new InvalidObjectException("bad object: bad-length simpleAds[]"); }
        // Ensure no nulls in simpleAds,
        // and that ads are sorted.
        if(simpleAds != null)
            {
            for(int i = simpleAds.length; --i >= 0; )
                {
                if(simpleAds[i] == null)
                    { throw new InvalidObjectException("bad object: simpleAds[] contains null"); }
                // Check pair-wise sort order...
                // The ordering should be total,
                // but we'll live with monotonic.
                if(i > 0)
                    {
                    if(simpleAds[i-1].compareTo(simpleAds[i]) > 0)
                        { throw new InvalidObjectException("bad object: simpleAds[] not ordered"); }
                    }
                }
            }
        // Check correct total for simpleAds.
        if(totalSimpleAdWeight != _compute_totalSimpleAdWeight(simpleAds))
            { throw new InvalidObjectException("bad object: totalSimpleAdWeight inconsistent"); }
        if(totalSimpleAdWeight < 0)
            { throw new InvalidObjectException("bad object: totalSimpleAdWeight < 0"); }
        // There must be no ads held internally if simple ads are switched off.
        if(!WEBSVR_SIMPLEAD_ALLOW &&
            ((totalSimpleAdWeight != 0) || (simpleAds != null)))
            { throw new InvalidObjectException("bad object: simple ads loaded though disabled"); }

        // Verify hotlinker diversion URL is valid (http) URL.
        if(WEBSVR_EX_HOTLINK_DIVERT_URL != null)
            {
            try {
                if(!"http".equals((new URL(WEBSVR_EX_HOTLINK_DIVERT_URL)).getProtocol()))
                    { throw new InvalidObjectException("bad object: invalid non-HTTP hotlink divert URL"); }
                }
            catch(final MalformedURLException e)
                {
                throw new InvalidObjectException("bad object: invalid unparseable hotlink divert URL");
                }
            }

        // FIXME: Should check content to be normalised host names too...
        if(hotLinkAllowHosts != null)
            {
            if(hotLinkAllowHosts.size() > MAX_ALLOW_DISALLOW_HOTLINK_HOSTS)
                { throw new InvalidObjectException("bad object: too many allow hosts"); }
            }
        if(hotLinkDisallowHosts != null)
            {
            if(hotLinkDisallowHosts.size() > MAX_ALLOW_DISALLOW_HOTLINK_HOSTS)
                { throw new InvalidObjectException("bad object: too many disallow hosts"); }
            }
        if(DNSRBLs != null)
            {
            if(DNSRBLs.size() > MAX_DNSRBLS)
                { throw new InvalidObjectException("bad object: too many DNSRBLs"); }
            }

        }

    /**Check if String is save as generic property key or value.
     * Null and zero-length values are not permitted.
     * <p>
     * Only ASCII values in the range 32 to 126 are allowed.
     *
     * @param s  the putative generic property key or value
     * @return true iff the argument is a non-null, non-empty, not-too-long
     *     pure-ASCII value
     */
    private static final boolean isSafeGenPropValue(final String s)
        {
        // Check basic constraints.
        if(s == null) { return(false); }
        final int length = s.length();
        if(length == 0) { return(false); }
        if(length > MAX_GEN_LEN) { return(false); }

        for(int i = length; --i >= 0; )
            {
            final char c = s.charAt(i);
            if((c < 32) || (c > 126)) { return(false); }
            }

        return(true); // Benign string.
        }

    /**Check if String is safe to use as HTML meta-header; return false if not.
     * Null is treated as OK.
     */
    private static final boolean isSafeHTMLMetaHeaderString(String s)
        {
        if(s == null) { return(true); }
        for(int i = s.length(); --i >= 0; )
            {
            if(!isSafeHTMLMetaHeaderChar(s.charAt(i)))
                { return(false); }
            }
        return(true); // All seems OK!
        }

    /**Check that character is safe for an HTML meta-header; return false if not.
     * Basically such a character must be printable ASCII, and
     * one of:
     * <ul>
     * <li>a letter
     * <li>a digit
     * <li>a space
     * <li>one of the punctutation characters -+?!()[]{}/_.;:,@
     * </ul>
     */
    private static final boolean isSafeHTMLMetaHeaderChar(char c)
        {
        switch(c)
            {
            // Letters...
            case 'a': case 'A':
            case 'b': case 'B':
            case 'c': case 'C':
            case 'd': case 'D':
            case 'e': case 'E':
            case 'f': case 'F':
            case 'g': case 'G':
            case 'h': case 'H':
            case 'i': case 'I':
            case 'j': case 'J':
            case 'k': case 'K':
            case 'l': case 'L':
            case 'm': case 'M':
            case 'n': case 'N':
            case 'o': case 'O':
            case 'p': case 'P':
            case 'q': case 'Q':
            case 'r': case 'R':
            case 's': case 'S':
            case 't': case 'T':
            case 'u': case 'U':
            case 'v': case 'V':
            case 'w': case 'W':
            case 'x': case 'X':
            case 'y': case 'Y':
            case 'z': case 'Z':
            // Digits.
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            // Safe punctuation and space.
            case ' ':
            case ',': case '.': case ';': case ':':
            case '-': case '+':
            case '!': case '?':
            case '(': case ')':
            case '[': case ']':
            case '{': case '}':
            case '/': case '_': case '@':
                { return(true); } // OK!

            default:
                { return(false); } // Not safe.
            }
        }

    /**Construct a default, zero-timestamp set of generic properties.
     */
    public GenProps()
        { this(new Properties(), 0, null); }

    /**Construct a new, immutable, properties set.
     * The properties must be non-null
     * and the timestamp must be non-negative.
     * <p>
     * Minor problems with the properties themselves will
     * be silently ignored or logged, and defaults substituted
     * for broken or missing values.
     */
    public GenProps(final Properties props,
                    final long _timestamp)
        { this(props, _timestamp, GenUtils.systemErrLogger); }

    /**Construct a new, immutable, properties set.
     * The properties must be non-null
     * and the timestamp must be non-negative.
     * <p>
     * Minor problems with the properties themselves will
     * be silently ignored or logged, and defaults substituted
     * for broken or missing values.
     *
     * @param logger if non-null than non-fatal problems will be logged here
     */
    public GenProps(final Properties props,
                    final long _timestamp,
                    final SimpleLoggerIF logger)
        {
        if((props == null) || (_timestamp < 0))
            { throw new IllegalArgumentException(); }
        timestamp = _timestamp;

        // Parse properties.
        // Set up some working variables to help with parsing.
        int iTmp;
//        long lTmp;
        String sTmp;
        Boolean bTmp;

        // Extract the generic key/value properties, if any.
        final Map<String,String> g = new HashMap<String, String>();
        for(final Object keyO : props.keySet())
            {
            if(!(keyO instanceof String))
                { throw new IllegalArgumentException("bad properties key: not String"); }
            final String key = (String) keyO;

            if(!key.startsWith(GEN_PREFIX)) { continue; }

            final String gKey = key.substring(GEN_PREFIX.length());
            final String gVal = props.getProperty(key);

            // Validate the values later.
            g.put(gKey, gVal);
            }
        // If no generic keys then save space with shared empty Map.
        if(g.size() == 0)
            { gen = Collections.emptyMap(); }
        else
            { gen = Collections.unmodifiableMap(g); }

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_MIN_EX_IMATTR_RECHECK_MS, "120000"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_MIN_EX_IMATTR_RECHECK_MS = iTmp;

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_SYSPROPS_RECHECK_MS, "10013"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_SYSPROPS_RECHECK_MS = iTmp;

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_MAX_CACHEABLE_EX_BYTES, "2123456"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_MAX_CACHEABLE_EX_BYTES = iTmp;

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_BW_LIMITER, "10"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_BW_LIMITER = iTmp;

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_TN_CACHEPC, "10"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_TN_CACHEPC = iTmp;

        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_MAX_EX_BYTES, "1123456789"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_MAX_EX_BYTES = iTmp;

        // We canonicalise this keyword list to reduce the
        // space it consumes (especially after compression), by:
        //   * Converting to lower-case (helps compression; helps drop dups).
        //   * Tokenize with "," and trim whitespace around each token.
        //   * We could eliminate duplicate whitespace within a single token too.
        //   * Eliminate duplicates.
        //   * Sort (to help compression).
        //   * Glue back together again without whitespace.
        //   * ELiminate any quote (" or ') characters.
        // We object strongly to non-ASCII text (we remove it!).
        sTmp = null;
        try {
            sTmp = props.getProperty(PNAME_WEBSVR_META_KEYWORDS, "").trim().toLowerCase();
            final SortedSet<String> words = new TreeSet<String>();
            final StringTokenizer st = new StringTokenizer(sTmp, ",");
            while(st.hasMoreTokens())
                {
                // Take each token,
                // zap any non-ASCII printable (and unsafe) chars,
                // and trim of any obvious excess whitespace.
                String s = st.nextToken();
                for(int i = s.length(); --i >= 0; )
                    {
                    final char c = s.charAt(i);
                    if(!isSafeHTMLMetaHeaderChar(c)) { s = s.replace(c, ' '); }
                    }
                s = s.trim();
                words.add(s);
                }
            final StringBuilder sb = new StringBuilder(sTmp.length());
            for(final Iterator<String> it = words.iterator(); it.hasNext(); )
                {
                final String s = it.next();
                sb.append(s);
                if(it.hasNext()) { sb.append(','); }
                }
            sTmp = sb.toString();
            }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_META_KEYWORDS = sTmp;

        sTmp = null;
        try {
            sTmp = props.getProperty(PNAME_WEBSVR_META_DESCRIPTION, "");
            // Discard any unsafe characters for a meta header.
            for(int i = sTmp.length(); --i >= 0; )
                {
                final char c = sTmp.charAt(i);
                if(!isSafeHTMLMetaHeaderChar(c)) { sTmp = sTmp.replace(c, ' '); }
                }
            sTmp = sTmp.trim();
            }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_META_DESCRIPTION = sTmp;

        sTmp = null;
        try {
            sTmp = props.getProperty(PNAME_WEBSVR_BG_IMAGE);

            // Drop if not a valid-syntax JPEG exhibit name.
            // We *do not* attempt to check that it exists here.
            if(!ExhibitName.validNameSyntax(sTmp))
                { sTmp = null; }
            else
                {
                ExhibitMIME.ExhibitTypeParameters tp =
                    ExhibitMIME.getInputFileType(sTmp);
                if((tp == null) || (tp.type != ExhibitMIME.ET_JPEG))
                    { sTmp = null; }
                }

            }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_BG_IMAGE = sTmp;

        bTmp = null;
        try { bTmp = Boolean.valueOf(props.getProperty(PNAME_WEBSVR_SIMPLEAD_ALLOW)); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_SIMPLEAD_ALLOW = (bTmp == null) ? true : bTmp.booleanValue();
        SimpleAd[] simpleAdsTmp = null;
        if(WEBSVR_SIMPLEAD_ALLOW)
            {
            final Vector<SimpleAd> v = new Vector<SimpleAd>(); // Collect ads here...
            // Only even look for simple ads if they are allowed.
            for(int N = 1; N < Integer.MAX_VALUE; ++N)
                {
                final String prefix = PNAME_WEBSVR_SIMPLEAD_PREFIX + "." + N + ".";
                final String rawCode = props.getProperty(prefix + "HTML");
                if(rawCode == null) { break; } // End of ads.

                final String code = rawCode.trim();
                int weight = 100; // Default;
                try { weight = Integer.parseInt(props.getProperty(prefix + "wt"), 10); }
                catch(Exception e) { } // Ignore errors.
                if(weight <= 0) { continue; } // Temporarily disabled; skip.

                // Attempt to construct new ad item and save it,
                // but quietly skip any difficulties.
                try { v.add(MemoryTools.intern(new SimpleAd(weight, code))); }
                catch(Exception e) { } // Ignore errors.
                }
            // If we found some ads, extract and sort them, ready to store.
            if(v.size() > 0)
                {
                simpleAdsTmp = new SimpleAd[v.size()];
                v.toArray(simpleAdsTmp);
                Arrays.sort(simpleAdsTmp);
                }
            }
        simpleAds = simpleAdsTmp; // Store simple ads, if any.
        totalSimpleAdWeight = _compute_totalSimpleAdWeight(simpleAds);


        // Extract author details.
        // We iterate over all the properties, looking for suitable candidiates.
        final TreeMap<String,AuthData> putativeAuthDB = new TreeMap<String, AuthData>();
        for(final Iterator it = props.keySet().iterator(); it.hasNext(); )
            {
            try
                {
                final String key = (String) it.next();
                if(!key.startsWith(PPREFIX_AUTH_DETAILS))
                    { continue; /* Not the right prefix for authDB. */ }

                // We have a key that could be a valid author entry...
                final String auth = key.substring(PPREFIX_AUTH_DETAILS.length());
                // TODO: add warning for invalid entries?
                if(!ExhibitName.validAuthorSyntax(auth))
                    { continue; /* Not a valid author initials string. */ }

                // TODO: add warning for duplicate entries?
                putativeAuthDB.put(MemoryTools.intern(auth),
                    MemoryTools.intern(new AuthData(auth, props.getProperty(key))));
                }
            catch(Exception e)
                {
                // Absorb errors...
                e.printStackTrace();
                }
            }
        // Save immutable version of our authDB.
        // If empty, save space by storing a null instead.
        authDB = (putativeAuthDB.size() == 0) ? null :
            Collections.unmodifiableSortedMap(putativeAuthDB);


        // Extract supplied static "goodness"/popularity weightings, if any.
        // We iterate over all the properties, looking for suitable candidiates.
        final HashMap<String,Byte> putativePopWeights = new HashMap<String, Byte>();
        for(final Iterator it = props.keySet().iterator(); it.hasNext(); )
            {
            try
                {
                final String key = (String) it.next();
                if(!key.startsWith(PPREFIX_POPWT_DETAILS))
                    { continue; /* Not the right prefix for popWeight. */ }

                final String aWPrefix = PPREFIX_POPWT_DETAILS + PCOMP_POPWR_BYAUTH;
                if(key.startsWith(aWPrefix))
                    {
                    // We have a key that could be a valid author weighting entry...
                    final String auth = key.substring(aWPrefix.length());
                    // TODO: add warning for invalid entries?
                    if(!ExhibitName.validAuthorSyntax(auth))
                        { continue; /* Not a valid author initials string. */ }
                    final Byte v = Byte.decode(props.getProperty(key));
                    if((v < MIN_POPWT_VAL) || (v > MAX_POPWT_VAL))
                        { continue; /* Not a valid weight. */ }
                    putativePopWeights.put(MemoryTools.intern(auth),
                                           MemoryTools.intern(v));
                    continue;
                    }

                final String tWPrefix = PPREFIX_POPWT_DETAILS + PCOMP_POPWR_BYTYPE;
                if(key.startsWith(tWPrefix))
                    {
                    // We have a key that could be a valid type/extension weighting entry...
                    final String type = key.substring(tWPrefix.length());
                    // TODO: add warning for invalid entries?
                    if(ExhibitMIME.isValidInputExhibitNameExtension(type) == null)
                        { continue; /* Not a valid type string. */ }
                    final Byte v = Byte.decode(props.getProperty(key));
                    if((v < MIN_POPWT_VAL) || (v > MAX_POPWT_VAL))
                        { continue; /* Not a valid weight. */ }
                    putativePopWeights.put(MemoryTools.intern(type),
                                           MemoryTools.intern(v));
                    continue;
                    }

                final String attrWPrefix = PPREFIX_POPWT_DETAILS + PCOMP_POPWR_BYATTR;
                if(key.startsWith(attrWPrefix))
                    {
                    // We have a key that could be a valid attribute-word weighting entry...
                    final String attrWord = key.substring(attrWPrefix.length());
                    // TODO: add warning for invalid entries?
                    if(!ExhibitName.validAttributeWord(attrWord))
                        { continue; /* Not a valid attribute word string. */ }
                    final Byte v = Byte.decode(props.getProperty(key));
                    if((v < MIN_POPWT_VAL) || (v > MAX_POPWT_VAL))
                        { continue; /* Not a valid weight. */ }
                    putativePopWeights.put(MemoryTools.intern(attrWord),
                                           MemoryTools.intern(v));
                    continue;
                    }
                }
            catch(Exception e)
                {
                // Absorb errors...
                e.printStackTrace();
                }
            }
        // Save immutable version of our authDB.
        // If empty, save space by storing a null instead.
        popWeights = (putativePopWeights.size() == 0) ? null :
            Collections.unmodifiableMap(putativePopWeights);


        iTmp = -1;
        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_EX_HOTLINK_LIMITER, "0"), 10); }
        catch(Exception e) { } // Ignore errors.
        WEBSVR_EX_HOTLINK_LIMITER = (byte) Math.max(0, Math.min(255, iTmp));

        // Get alternative URL to divert unwanted exhibit hotlinks to, if any.
        // Invalid URLs are ignored and valid ones may be canonicalised.
        sTmp = null;
        try
            {
            final String p = props.getProperty(PNAME_WEBSVR_EX_HOTLINK_DIVERT_URL);
            if((p != null) && (p.startsWith("http:")))
                { sTmp = (new URL(p)).toExternalForm(); } // Parse and canonicalise.
            }
        catch(final MalformedURLException e) { } // Ignore malformed URLs.
        WEBSVR_EX_HOTLINK_DIVERT_URL = sTmp;

        // Parse lists of hotlinker allow/deny hosts, DNSRBLs...
        final Set<String> rawAllowNames = _parseHostList(
                        props.getProperty(PNAME_WEBSVR_EX_HOTLINK_LIMITER_ALLOW));
        hotLinkAllowHosts = _normaliseHostList(rawAllowNames, MAX_ALLOW_DISALLOW_HOTLINK_HOSTS);
        hotLinkAllowHostsRegex = _makeResidualRegexHostMatch(rawAllowNames, logger);
        final Set<String> rawDisallowNames = _parseHostList(
                        props.getProperty(PNAME_WEBSVR_EX_HOTLINK_LIMITER_DISALLOW));
        hotLinkDisallowHosts = _normaliseHostList(rawDisallowNames, MAX_ALLOW_DISALLOW_HOTLINK_HOSTS);
        hotLinkDisallowHostsRegex = _makeResidualRegexHostMatch(rawDisallowNames, logger);
        DNSRBLs = _normaliseHostList(_parseHostList(
                props.getProperty(PNAME_WEBSVR_DNSRBLs)), MAX_DNSRBLS);


        // Verify complete object state.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Timestamp of this properties set; zero for default/empty set. */
    public final long timestamp;


    /**Maximum number of general properties. */
    public static final int MAX_GEN_PROPS = 1024;

    /**Maximum length of general property key or value (chars). */
    public static final int MAX_GEN_LEN = 1024;

    /**Prefix for general properties in file (not needed on lookup). */
    public static final String GEN_PREFIX = "gen.";

    /**The (immutable) generic properties map; never null. */
    private Map<String,String> gen;

    /**Get the immutable generic properties; never null.
     * The keys in this Map here are stripped of the initial GEN_PREFIX
     * that appears in the properties file.
     */
    public Map<String,String> getGen()
        { return(gen); }


    /**Name of Web server minimum exhibit-immutable-attributes check interval (ms) property. */
    public static final String PNAME_WEBSVR_MIN_EX_IMATTR_RECHECK_MS = "pg2k.websvr.ex.imattr.minrecheckms";
    /**Web server minimum exhibit-immutable-attributes check interval (ms). */
    private final int WEBSVR_MIN_EX_IMATTR_RECHECK_MS;
    /**Get the Web server minimum exhibit-immutable-attributes check interval (ms).
     * Constrained to the range approx 1s to 6h, default approx 2m.
     */
    public final int getWEBSVR_MIN_EX_IMATTR_RECHECK_MS()
        { return(Math.max(1001, Math.min(6 * 3600 * 1000, WEBSVR_MIN_EX_IMATTR_RECHECK_MS))); }

    /**Name of Web server sysprops check interval (ms) property. */
    public static final String PNAME_WEBSVR_SYSPROPS_RECHECK_MS = "pg2k.websvr.props.sys.recheckms";
    /**Web server sysprops (ie most property files including GenProps) (re)check interval (ms). */
    private final int WEBSVR_SYSPROPS_RECHECK_MS;
    /**Get the Web server sysprops check interval (ms).
     * Constrained to the range approx 1s to 1h, default approx 10s (which applies until first props set is fetched).
     */
    public final int getWEBSVR_SYSPROPS_RECHECK_MS()
        { return(Math.max(1003, Math.min(1 * 3600 * 1000, WEBSVR_SYSPROPS_RECHECK_MS))); }

    /**Name of Web server maximum-cached-exhibit-prefix-size (bytes) property. */
    public static final String PNAME_WEBSVR_MAX_CACHEABLE_EX_BYTES = "pg2k.websvr.ex.maxcacheablesize";
    /**Web server maximum-cached-exhibit-prefix-size (bytes). */
    private final int WEBSVR_MAX_CACHEABLE_EX_BYTES;
    /**Get the Web server maximum-cached-exhibit-prefix-size (bytes).
     * Constrained to the range approx 16kB to 1GB, default approx 2MB.
     */
    public final int getWEBSVR_MAX_CACHEABLE_EX_BYTES()
        { return(Math.max(16384, Math.min(1123456789, WEBSVR_MAX_CACHEABLE_EX_BYTES))); }

    /**Name of Web server persistent thumbnail cache size limit as percentage of exhibit cache limit value. */
    public static final String PNAME_WEBSVR_TN_CACHEPC = "pg2k.websvr.ex.tncachepercent";
    /**Web server persistent thumbnail cache size limit as percentage of exhibit cache limit value. */
    private final int WEBSVR_TN_CACHEPC;
    /**Web server persistent thumbnail cache size limit as percentage of exhibit cache limit value.
     * Constrained to the range 0 to 100, default approx 10.
     * <p>
     * A value of zero disables the persistent cache
     * (though in-memory cacheing may continue).
     * <p>
     * Any additional cache of generated thumbnails (exhibit samples)
     * is limited to a size at most this many percent of the exhibit cache size.
     * This cache is managed separately from this exhibit cache, but yoking the
     * values together makes for simpler system management.
     */
    public final int getWEBSVR_TN_CACHEPC()
        { return(Math.max(0, Math.min(100, WEBSVR_TN_CACHEPC))); }

    /**Name of Web server precacheing bandwidth-limiter value (max fraction of available bandwidth used on precacheing). */
    public static final String PNAME_WEBSVR_BW_LIMITER = "pg2k.websvr.ex.bwlimiter";
    /**Web server precacheing bandwidth-limiter value (max fraction of available bandwidth used on precacheing). */
    private final int WEBSVR_BW_LIMITER;
    /**Web server precacheing bandwidth-limiter value (max fraction of available bandwidth used on precacheing).
     * Constrained to the range approx 1 to 1000, default approx 10.
     * <p>
     * If any precacheing operation takes N milliseconds the next one
     * will not be started until at least N times this value later.
     * <p>
     * The effect is to limit resource consumption to at most about
     * 1/N of the most constrained of slave and master resources
     * (typically CPU and/or bandwidth).
     */
    public final int getWEBSVR_BW_LIMITER()
        { return(Math.max(1, Math.min(1000, WEBSVR_BW_LIMITER))); }

    /**Name of Web server maximum exhibit size (32-bit value, bytes). */
    public static final String PNAME_WEBSVR_MAX_EX_BYTES = "pg2k.websvr.ex.maxexhibitsize";
    /**Web server maximum exhibit size (32-bit value, bytes). */
    private final int WEBSVR_MAX_EX_BYTES;
    /**Get the Web server maximum exhibit size (32-bit value, bytes).
     * Constrained to the range approx 1B to 1GB, default approx 1GB.
     */
    public final int getWEBSVR_MAX_EX_BYTES()
        { return(Math.max(1, Math.min(Integer.MAX_VALUE/2, WEBSVR_MAX_EX_BYTES))); }

    /**Name of keywords list for HTML meta header ("" if none; not null). */
    public static final String PNAME_WEBSVR_META_KEYWORDS = "pg2k.websvr.meta.keywords";
    /**Keywords list for HTML meta header ("" if none; not null). */
    private final String WEBSVR_META_KEYWORDS;
    /**Get the keywords list for HTML meta header ("" if none; never null).
     * May be trimmed of excess whitespace and canonicalised; "" if none.
     * <p>
     * Guaranteed only to contain characters safe to put directly into
     * an HTML description meta-header attribute value.
     * <p>
     * In particular this result contains no quote marks (" or ')
     * nor ampersand nor angle-brackets, though may contain whitespace,
     * so should be quoted.
     */
    public final String getWEBSVR_META_KEYWORDS()
        {
        final String result = WEBSVR_META_KEYWORDS;
        if(result == null) { return(""); }
        assert(result.indexOf('\'') == -1);
        assert(result.indexOf('"') == -1);
        assert(result.indexOf('&') == -1);
        assert(result.indexOf('<') == -1);
        assert(result.indexOf('>') == -1);
        return(result);
        }

    /**Name of description for HTML meta header ("" if none; not null). */
    public static final String PNAME_WEBSVR_META_DESCRIPTION = "pg2k.websvr.meta.description";
    /**Description for HTML meta header ("" if none; not null). */
    private final String WEBSVR_META_DESCRIPTION;
    /**Get the description list for HTML meta header ("" if none; not null).
     * May be trimmed of excess whitespace; "" if none.
     */
    public final String getWEBSVR_META_DESCRIPTION()
        {
        final String result = WEBSVR_META_DESCRIPTION;
        if(result == null) { return(""); }
        return(result);
        }

    /**Name of JPEG exhibit to use as a background image (null if none). */
    public static final String PNAME_WEBSVR_BG_IMAGE = "pg2k.websvr.livery.default.bg";
    /**JPEG exhibit to use as a background image (null if none). */
    private final String WEBSVR_BG_IMAGE;
    /**Get the JPEG exhibit to use as a background image (null if none).
     * Is a syntactically-valid JPEG exhibit name (or null).
     */
    public final String getWEBSVR_BG_IMAGE()
        {
        final String result = WEBSVR_BG_IMAGE;
        return(result);
        }

    /**Class holding details of one simple banner advertisement; immutable, serialisable.
     * Weighting must be strictly positive integer; text must be
     * non-zero-length, 7-bit printable (32--126) ASCII/HTML data,
     * though we also allow CRLF (\r\n) for tricky JavaScript/HTML issues...
     * <p>
     * We assume that the on-the-wire format for members is reasonably
     * efficient, especially if the transport layer includes compression.
     * <p>
     * This support equals(), hashCode() and compareTo(); two objects
     * are equal if all their fields are, and the sort order is to
     * mimimise ad-lookup and HTML-bandwidth (ie by highest weight
     * then smallest code, with ties being broken by code lexical ordering).
     * <p>
     * On deserialising we should possibly check that weight and code values
     * are valid.
     */
    private static final class SimpleAd implements Serializable,
                                                   ObjectInputValidation,
                                                   MemoryTools.Internable,
                                                   Comparable
        {
        /**Strictly positive weighting. */
        final int weight;
        /**HTML code as 7-bit ASCII; non-null, non-zero length. */
        final String code;

        /**Maximum HTML code length in bytes/characters. */
        public static final int MAX_AD_CHARACTERS = 2048;

        /**Equality depends on all fields. */
        public boolean equals(Object o)
            {
            if(!(o instanceof SimpleAd)) { return(false); }
            final SimpleAd other = (SimpleAd) o;
            return((weight == other.weight) && (code.equals(other.code)));
            }

        /**Hash depends on text as good source of variability. */
        public int hashCode() { return(code.hashCode()); }

        /**Provides total ordering, with primary sort by decreasing weight.
         * Full sort order is:
         * <ol>
         * <li>decreasing weight, then
         * <li>increasing code length, then
         * <li>code lexical order.
         * </ol>
         */
        public int compareTo(final Object o)
            {
            final SimpleAd other = (SimpleAd) o;

            if(weight > other.weight) { return(-1); } // Right order.
            if(weight < other.weight) { return( 1); } // Wrong order.

            if(code.length() < other.code.length()) { return(-1); } // Right order.
            if(code.length() > other.code.length()) { return( 1); } // Wrong order.

            // Break ties to make total ordering.
            return(code.compareTo(other.code));
            }

        /**Construct new simple ad details. */
        SimpleAd(final int adRelativeWeight, final String HTMLCode)
            {
            weight = adRelativeWeight;
            code = MemoryTools.intern(HTMLCode);
            // Verify object state and thus validate parameters...
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Our serialisation version. */
        private static final long serialVersionUID = 6649356978055850149L;

//        /**Deserialise. */
//        private void readObject(final ObjectInputStream in)
//            throws IOException, ClassNotFoundException
//            {
//            in.defaultReadObject();
//            validateObject(); // Validate state immediately.
//            }

        /**Deserialise: validate and eliminate duplicates coming off the wire.
         * @return identical, de-duped, non-null instance
         */
        protected Object readResolve()
            // throws ObjectStreamException
            {
            return(MemoryTools.intern(new SimpleAd(weight, code)));
            }

        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         */
        public void validateObject()
            throws InvalidObjectException
            {
            // Check that all components are sane and safe.
            // Weight must be strictly positive.
            if(weight <= 0)
                { throw new InvalidObjectException("bad object: weight <= 0"); }
            // HTML code must be non-null, non-zero-length,
            // 7-bit printable ASCII (and CRLF).
            if((code == null) ||
               (code.length() == 0) || (code.length() > MAX_AD_CHARACTERS))
                { throw new InvalidObjectException("bad object: HTML code length silly"); }
            for(int i = code.length(); --i >= 0; )
                {
                final char c = code.charAt(i);
                if(((c < 32) || (c > 126)) &&
                   ((c != '\r') && (c != '\n')))
                    { throw new InvalidObjectException("bad object: HTML code contains bad char " + ((int) c)); }
                }
            }
        }


    /**Immutable Set of String tokens to replace with random values in banner HTML.
     * These tokens, when encountered in HTML banner ad text,
     * are replaced with random (positive, decimal) integers.
     * These are to help with "cache-busting".
     * <p>
     * These tokens must be chosen so as to be very unlikely to appear
     * where they are <em>not</em> intended for replacement.
     * <p>
     * This set may become replaceable at run-time in future.
     */
    public static final Set<String> AD_BANNER_RND_TOKENS = Collections.unmodifiableSet(
        new HashSet<String>(Arrays.asList(new String[]{
            "$RND$", // TribalFusion cache-buster token.
        })));

    /**Name of prefix for simple banner ad code.
     * There is one sub-property ({prefix}.allow, either "true" or "false")
     * that allows or disallows all simple ads at once.
     * <p>
     * Then for N starting at 1, and stopping at the first gap,
     * there must be a non-empty {prefix}.{N}.HTML containing
     * non-empty, pure-printable 7-bit ASCII (32--126, +CRLF) HTML ad code.
     * There may be an optional {prefix}.{N}.wt strictly-positive
     * integer relative weighting (default 100) of this ad to others,
     * the higher the weighting the more of the available space taken.
     */
    public static final String PNAME_WEBSVR_SIMPLEAD_PREFIX = "pg2k.websvr.ad";
    /**Name of on/off switch for simple banner ad code, enabled by default. */
    public static final String PNAME_WEBSVR_SIMPLEAD_ALLOW = PNAME_WEBSVR_SIMPLEAD_PREFIX + ".allow";
    /**On/off switch for simple banner ad code, enabled by default. */
    private final boolean WEBSVR_SIMPLEAD_ALLOW;
    /**Get on/off switch for simple banner ad code, enabled by default.
     */
    public final boolean getWEBSVR_SIMPLEAD_ALLOW() { return(WEBSVR_SIMPLEAD_ALLOW); }
    /**Get HTML code for a simple ad at random; null if none available.
     * This returns ads with frequencies in proportion to their weightings.
     * <p>
     * This will always return null if getWEBSVR_SIMPLEAD_ALLOW()
     * returns false.
     * <p>
     * This algorithm will take O(n) time to return one of n available ads.
     * <p>
     * Will be faster when ads are unevenly weigted.
     * <p>
     * This will automatically do substitution of random "cache-buster" tokens
     * if required.
     * (Any of the recognised tokens will be substituted with a
     * random (positive decimal) integer value, different each time,
     * with all occurrences of any one token the same in any one call.)
     */
    public final String getSimpleAd()
        {
        if(/* !WEBSVR_SIMPLEAD_ALLOW || */ /* Test is redundant. */
           /* (simpleAds == null) || */ /* Test is redundant. */
           (totalSimpleAdWeight == 0)) { return(null); }

        // Pick uniformly-distributed selector value.
        // We primarily require speed from this generator,
        // not perfection nor security.
        final int r = Rnd.fastRnd.nextInt(totalSimpleAdWeight);

        // Total weight so far.
        int tw = 0;

        // Look until cumulative sum of weights passes selector,
        // then take the ad that passes that threshold.
        // If performance proves to be a problem we could
        // replace this with a binary chop by storing an auxillary
        // array containing the sum of the weights.
        // In the interim, if we discover that this array is not
        // sorted biggest weight first (which makes for fastest
        // linear search on average) then we can sort it when
        // we have finished selecting our ad this time.
        for(int i = 0; i < simpleAds.length; ++i)
            {
            tw += simpleAds[i].weight;
            if(r < tw)
                {
                String code = simpleAds[i].code;
                // Substitute cache-buster tokens as needed.
                if((code != null) && (AD_BANNER_RND_TOKENS.size() > 0))
                    {
                    for(final String token : AD_BANNER_RND_TOKENS)
                        {
                        // If the cache-buster token exists then replace all occurrences.
                        if(code.indexOf(token) != -1)
                            {
                            final int rnd = (Rnd.fastRnd.nextInt() >>> 1);
                            code = code.replace(token, String.valueOf(rnd));
                            }
                        }
                    }
                return(code);
                }
            }

        return(null); // No ad available.  (Probably should not be able to happen.)
        }

    /**Total weight of all simple ads; non-negative.
     * Strictly positive if there are some simple ads.
     */
    private final int totalSimpleAdWeight;

    /**Used to recompute totalSimpleAdWeight; zero if arg is null.
     * @throws IllegalArgumentException  if sum of weights would
     *     exceed Integer.MAX_VALUE.
     */
    private static int _compute_totalSimpleAdWeight(SimpleAd _simpleAds[])
        {
        if(_simpleAds == null) { return(0); }
        int result = 0;
        for(int i = _simpleAds.length; --i >= 0; )
            {
            final long r = result + (long) (_simpleAds[i].weight);
            if(r > Integer.MAX_VALUE)
                { throw new IllegalArgumentException("weights too large"); }
            result = (int) r;
            }
        return(result);
        }

    /**Sorted array of simple ads, no empty slots, non-zero length; or null if no simple ads. */
    private /* final */ SimpleAd simpleAds[];

    /**The maximum number of ads that we will entertain at once. */
    public static final int MAX_SIMPLE_ADS = 64;



    /**Get AuthData for specified author, or null if none.
     * @param auth  author initials of desired data.
     */
    public synchronized AuthData getAuthData(final String auth)
        {
        if(authDB == null) { return(null); }
        return(authDB.get(auth));
        }

    /**Property name prefix for author details. */
    public static final String PPREFIX_AUTH_DETAILS = "pg2k.authdb.";

    /**Immutable SortedMap of author details, no empty or duplicate slots; non-zero length, or null if none. */
    private /* final */ SortedMap<String,AuthData> authDB;

    /**The maximum number of author database entries. */
    public static final int MAX_AUTH_ENTRIES = 1024;

    /**Class holding details of one author; immutable, serialisable.
     * Author initial must be valid syntactically;
     * text must be 7-bit printable (32--126) ASCII/HTML data.
     * <p>
     * We assume that the on-the-wire format for members is reasonably
     * efficient, especially if the transport layer includes compression.
     * <p>
     * This supports equals(), hashCode() and compareTo(); two objects
     * are equal iff all the fields are, and the sort order is by author.
     */
    public static final class AuthData implements Serializable,
                                                  ObjectInputValidation,
                                                  Comparable,
                                                  MemoryTools.Internable
        {
        /**Author initials: must be syntactically valid; non-null. */
        public final String auth;

        /**Author name: must be pure 7-bit printable ASCII HTML with possible entities; not empty, non-null. */
        public final String name;

        /**Author home-page URL: should be syntactically valid URL or null. */
        public final String www;

        /**Author email: should be syntactically valid email address or null. */
        public final String email;

        /**Description HTML code as 7-bit ASCII; never empty but can be null. */
        public final String desc;

        /**Maximum HTML description length in bytes/characters. */
        public static final int MAX_DESC_CHARACTERS = 2048;

        /**Equality depends on all fields. */
        public boolean equals(final Object o)
            {
            if(!(o instanceof AuthData)) { return(false); }
            final AuthData other = (AuthData) o;
            if(!auth.equals(other.auth)) { return(false); }
            if(!name.equals(other.name)) { return(false); }
            if((www == null) ? (other.www != null) : !www.equals(other.www)) { return(false); }
            if((email == null) ? (other.email != null) : !email.equals(other.email)) { return(false); }
            if((desc == null) ? (other.desc != null) : !desc.equals(other.desc)) { return(false); }
            return(true);
            }

        /**Hash depends on author. */
        public int hashCode() { return(auth.hashCode()); }

        /**Provides total ordering, by author. */
        public int compareTo(final Object o)
            { return(auth.compareTo(((AuthData) o).auth)); }

        /**Construct new author details from single string.
         * The String should be of the form:
         * <samp>Real Name|HomePageURL|email|description</samp>
         * where any item can be blank except the name.
         */
        AuthData(final String authInitials,
                 final String pipeDelimitedString)
            {
            this(authInitials,
                 _getSection(pipeDelimitedString, 0),
                 _getSection(pipeDelimitedString, 1),
                 _getSection(pipeDelimitedString, 2),
                 _getSection(pipeDelimitedString, 3));
            }

        /**Extract given section from pipe-delimited String.
         * Skip given number of pipe symbols ("|")
         * then return next up to following one if present.
         * Returns null if section not present.
         */
        private static String _getSection(final String s, final int section)
            {
            int prevPipePos = -1;
            for(int i = section; --i >= 0; )
                {
                final int nextPipe = s.indexOf('|', prevPipePos+1);
                // Return null if requested section not present.
                if(nextPipe == -1) { return(null); }
                // Else adjust start marker.
                prevPipePos = nextPipe;
                }

            // Find the start of the following section, if any.
            final int nextPipePos = s.indexOf('|', prevPipePos+1);

            // No following section; return the rest of the string.
            if(nextPipePos == -1)
                { return(s.substring(prevPipePos+1)); }

            // Return the requested section.
            return(s.substring(prevPipePos+1, nextPipePos));
            }

        /**Construct new author details from individial components.
         * @param authInitials  the syntactally-valid author initials; non-null
         * @param authName  real name of author; not empty nor null
         * @param homePageURL  valid http URL, will be stripped,
         *     empty text is converted to null
         * @param emailAddress  valid email address, will be stripped,
         *     empty text is converted to null
         * @param descriptionText  the descriptive text, will be stripped,
         *     empty text is converted to null
         *
         * Not given public access as only GenProps should need to construct.
         */
        AuthData(final String authInitials,
                 final String authName,
                 String homePageURL,
                 String emailAddress,
                 String descriptionText)
            {
            // Canonicalise URL.
            if(homePageURL != null)
                {
                homePageURL = homePageURL.trim();
                if(homePageURL.length() == 0)
                    { homePageURL = null; }
                }

            // Canonicalise email address.
            if(emailAddress != null)
                {
                emailAddress = emailAddress.trim();
                if(emailAddress.length() == 0)
                    { emailAddress = null; }
                }

            // Canonicalise description text.
            if(descriptionText != null)
                {
                descriptionText = descriptionText.trim();
                if(descriptionText.length() == 0)
                    { descriptionText = null; }
                }

            // Use intern() mainly to reduce old-heap memory churn.
            auth = MemoryTools.intern(authInitials);
            name = MemoryTools.intern(authName.trim());
            www = MemoryTools.intern(homePageURL);
            email = MemoryTools.intern(emailAddress);
            desc = MemoryTools.intern(descriptionText);

            // Verify object state and thus validate parameters...
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Our serialisation version. */
        private static final long serialVersionUID = 2647433856891011944L;

        /**Deserialise. */
        private void readObject(final ObjectInputStream in)
            throws IOException, ClassNotFoundException
            {
            in.defaultReadObject();
            validateObject(); // Validate state immediately.
            }

        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         */
        public void validateObject()
            throws InvalidObjectException
            {
            // Validate author initials...
            if(!ExhibitName.validAuthorSyntax(auth))
                { throw new InvalidObjectException("bad object: author initials invalid"); }

            if((name == null) || (name.length() == 0) ||
                !name.equals(name.trim()))
                { throw new InvalidObjectException("bad object: author name missing or too short or non-canonical"); }
            for(int i = name.length(); --i >= 0; )
                {
                final char c = name.charAt(i);
                if((c < 32) || (c > 126))
                    { throw new InvalidObjectException("bad object: author name contains bad char " + ((int) c)); }
                }

            // Validate URL if not null.
            //   * Start with "http://".
            //   * Parseable...
            if(www != null)
                {
                if(!www.equals(www.trim()))
                    { throw new InvalidObjectException("bad object: WWW address non-canonical"); }
                if(!www.startsWith("http://"))
                    { throw new InvalidObjectException("bad object: WWW address seems invalid"); }
                try { new URL(www); }
                catch(MalformedURLException e)
                    { throw new InvalidObjectException("bad object: WWW address unparseable"); }
                }

            // Validate email address if not null.
            //   * At least 5 characters (eg "a@b.c").
            //   * Contains exactly one "@".
            if(email != null)
                {
                if(!email.equals(email.trim()))
                    { throw new InvalidObjectException("bad object: email address non-canonical"); }
                if(email.length() < 5)
                    { throw new InvalidObjectException("bad object: email address too short"); }
                final int firstAt = email.indexOf('@');
                if(firstAt < 1)
                    { throw new InvalidObjectException("bad object: email address user portion missing"); }
                if(firstAt != email.lastIndexOf('@'))
                    { throw new InvalidObjectException("bad object: email address contains more than one @"); }
                }

            // HTML code must be non-empty (but can be null),
            // 7-bit printable ASCII (and CRLF).
            if(desc != null)
                {
                if(!desc.equals(desc.trim()))
                    { throw new InvalidObjectException("bad object: HTML desc text non-canonical"); }
                if((desc.length() == 0) || (desc.length() > MAX_DESC_CHARACTERS))
                    { throw new InvalidObjectException("bad object: HTML desc text length silly"); }
                for(int i = desc.length(); --i >= 0; )
                    {
                    final char c = desc.charAt(i);
                    if(((c < 32) || (c > 126)) &&
                       ((c != '\r') && (c != '\n')))
                        { throw new InvalidObjectException("bad object: HTML desc text contains bad char " + ((int) c)); }
                    }
                }
            }
        }

    /**Prefix for static "goodness"/popularity weighting values. */
    public static final String PPREFIX_POPWT_DETAILS = "pg2k.popweight.";

    /**Goodness subcomponent by author; includes trailing dot. */
    public static final String PCOMP_POPWR_BYAUTH = "byAuth.";

    /**Goodness subcomponent by type/extension; includes trailing dot. */
    public static final String PCOMP_POPWR_BYTYPE = "byType.";

    /**Goodness subcomponent by attribute; includes trailing dot. */
    public static final String PCOMP_POPWR_BYATTR = "byAttr.";

    /**The maximum number of static "goodness"/popularity database entries. */
    public static final int MAX_POPWT_ENTRIES = 256;

    /**The maximum allowed "goodness"/popularity weight (maximally good); strictly negative. */
    public static final byte MAX_POPWT_VAL = 100;

    /**The minimum allowed "goodness"/popularity weight (maximally bad); strictly negative. */
    public static final byte MIN_POPWT_VAL = -MAX_POPWT_VAL;

    /**Map from String token to Byte ([-100,+100]) weighting.
     * We cheat and assume the that sets of author, attribute words and extensions
     * are disjoint, and so we keep all the mappings in a single table!
     */
    private /* final */ Map<String,Byte> popWeights;

    /**Get weighting by author in range ([-100,+100]); null if no weighting for specified author.
     * Author initials specified must be syntactically valid.
     * <p>
     * Note that this should be treated as if a value of +1.0
     * with a correlation of the returned value divided by MAX_POPWT_VAL.
     */
    public synchronized Byte getPopWeightForAuth(final String auth)
        {
        if(!ExhibitName.validAuthorSyntax(auth))
            { throw new IllegalArgumentException(); }
        final Map<String,Byte> p = popWeights;
        if(p == null) { return(null); }
        return(p.get(auth));
        }

    /**Get weighting by type/extension in range ([-100,+100]); null if no weighting for specified type.
     * Extension/type specified must be syntactically valid and known to the system.
     * <p>
     * Note that this should be treated as if a value of +1.0
     * with a correlation of the returned value divided by MAX_POPWT_VAL.
     */
    public synchronized Byte getPopWeightForType(final String type)
        {
        if(ExhibitMIME.isValidInputExhibitNameExtension(type) == null)
            { throw new IllegalArgumentException(); }
        final Map<String,Byte> p = popWeights;
        if(p == null) { return(null); }
        return(p.get(type));
        }

    /**Get weighting by attribute word in range ([-100,+100]); null if no weighting for specified attribute.
     * Attribute word specified must be syntactically valid and known to the system.
     * <p>
     * Note that this should be treated as if a value of +1.0
     * with a correlation of the returned value divided by MAX_POPWT_VAL.
     */
    public synchronized Byte getPopWeightForAttr(final String attrWord)
        {
        if(!ExhibitName.validAttributeWord(attrWord))
            { throw new IllegalArgumentException(); }
        final Map<String,Byte> p = popWeights;
        if(p == null) { return(null); }
        return(p.get(attrWord));
        }


    /**Name of Web server hotlinked-download-limiter value (as percentange of all exhibit downloads). */
    public static final String PNAME_WEBSVR_EX_HOTLINK_LIMITER = "pg2k.websvr.ex.hotlinklimiter";
    /**Web server hotlinked-download-limiter value (as percentange of all exhibit downloads). */
    private final byte WEBSVR_EX_HOTLINK_LIMITER;
    /**Web server hotlinked-download-limiter value (as percentange of all exhibit downloads).
     * Constrained to the range approx 0 to 255, default 0.
     * <p>
     * This limit is too prevent too much (lazy or bandwidth-theft) direct hot-linking
     * to exhibits (and/or thumbnails) by external Web sites.
     * <p>
     * There <em>are</em> bona fide reasons for some direct hotlinks, eg:
     * <ul>
     * <li>From some image search engines.
     * <li>From small Web sites and/or from contributors' sites.
     * </ul>
     */
    public final int getWEBSVR_EX_HOTLINK_LIMITER()
        { return(WEBSVR_EX_HOTLINK_LIMITER & 0xff); }

    /**Name of Web server hotlinked-download-limiter divert graphic URL value. */
    public static final String PNAME_WEBSVR_EX_HOTLINK_DIVERT_URL = "pg2k.websvr.ex.hotlinklimiter.altURL";
    /**Web server hotlinked-download-limiter divert graphic URL value. */
    private final String WEBSVR_EX_HOTLINK_DIVERT_URL;
    /**Web server hotlinked-download-limiter value divert graphic URL value, or null if none.
     * Absolute URL of an alternate graphic to divert a user's brower to
     * when they have been hotlinked to one of our exhibits,
     * or null if the user is simply to get a 5XX error.
     * <p>
     * Preferably a small image with a long cache time.
     */
    public final String getWEBSVR_EX_HOTLINK_DIVERT_URL()
        { return(WEBSVR_EX_HOTLINK_DIVERT_URL); }

    /**Maximum number of allow/disallow hotlink hosts that may be specified. */
    public static final int MAX_ALLOW_DISALLOW_HOTLINK_HOSTS = 64;

    /**Name of Web server hotlinked-download-host-allow set. */
    public static final String PNAME_WEBSVR_EX_HOTLINK_LIMITER_ALLOW = "pg2k.websvr.ex.hotlinklimiter.allowhosts";

    /**Name of Web server hotlinked-download-host-disallow set. */
    public static final String PNAME_WEBSVR_EX_HOTLINK_LIMITER_DISALLOW = "pg2k.websvr.ex.hotlinklimiter.disallowhosts";

    /**Immutable Set of "allow" hotlink hosts (normalised host names); can be null for no such hosts. */
    private /* final */ Set<String> hotLinkAllowHosts;

    /**Immutable compiled case-insensitive regex pattern for hotlink allow hosts not in literal list; can be null if no such "regex-match" hosts. */
    private final java.util.regex.Pattern hotLinkAllowHostsRegex;

    /**Immutable Set of "disallow" hotlink hosts (normalised/stripped/lower-cased host names); can be null for no such hosts. */
    private /* final */ Set<String> hotLinkDisallowHosts;

    /**Immutable compiled case-insensitive regex pattern for hotlink disallow hosts not in literal list; can be null if no such "regex-match" hosts. */
    private final java.util.regex.Pattern hotLinkDisallowHostsRegex;

    /**Get immutable Set of "allow" hotlink hosts (normalised host names); never null. */
    public final Set<String> getHotLinkAllowHosts()
        {
        final Set<String> r = hotLinkAllowHosts;
        if(r == null)
            {
            final Set<String> noHosts = Collections.emptySet();
            return(noHosts);
            }
        return(r);
        }

    /**Get immutable compiled case-insensitive regex pattern for hotlink allow hosts not in literal list; can be null if no such "regex-match" hosts. */
    public final Pattern getHotLinkAllowHostsRegex()
        {
        return(hotLinkAllowHostsRegex);
        }

    /**Get immutable Set of "disallow" hotlink hosts (normalised host names); never null. */
    public final Set<String> getHotLinkDisallowHosts()
        {
        final Set<String> r = hotLinkDisallowHosts;
        if(r == null)
            {
            final Set<String> noHosts = Collections.emptySet();
            return(noHosts);
            }
        return(r);
        }

    /**Get immutable compiled case-insensitive regex pattern for hotlink disallow hosts not in literal list; can be null if no such "regex-match" hosts. */
    public final Pattern getHotLinkDisallowHostsRegex()
        {
        return(hotLinkDisallowHostsRegex);
        }

    /**Returns true if this may be a regex (ie contains non-DNS-safe chars).
     * If a user-specified hostname contains characters
     * other than [-a-zA-Z0-9.]
     * then we assume that it may be intended as a regex.
     *
     * @param n  non-null non-empty user-specified non-normalised hostname
     */
    private static boolean _mayBeRegexHostname(final String n)
        {
        assert((n != null) && (n.length() != 0));
        for(int i = n.length(); --i >= 0; )
            {
            final char c = n.charAt(i);
            if((c >= 'a') && (c <= 'z')) { continue; }
            if((c >= 'A') && (c <= 'Z')) { continue; }
            if((c >= '0') && (c <= '9')) { continue; }
            if((c == '.') || (c == '-')) { continue; }
            return(true); // Not a plain DNS-safe name.
            }
        return(false); // Seems to be a plain name.
        }

    /**Make single compiled pattern from all regex-match hostname expressions; null if no such expressions.
     * The pattern will match if any of the sub-patterns match.
     * <p>
     * This will whinge about and drop any indiviual non-compilable patterns.
     * No whingeing is done, however, if the passed log is null!
     */
    private static Pattern _makeResidualRegexHostMatch(final Set<String> rawNames,
                                                       final SimpleLoggerIF logger)
        {
        if(rawNames == null) { return(null); }

        final StringBuilder compoundExpr = new StringBuilder();

        for(final String rawName : rawNames)
            {
            try
                {
                // Ignore plain host names that are not regexes...
                if(!_mayBeRegexHostname(rawName)) { continue; }

                // See if this expression can be compiled, ie is valid...
                try { Pattern.compile(rawName, Pattern.CASE_INSENSITIVE); }
                catch(final Exception e)
                    {
                    if(logger != null)
                        { logger.log("WARNING: GenProps: ignoring unusable/bad hostname regex: `"+rawName+"'"); }
                    continue; // Skip this unusable pattern.
                    }

                // Append this pattern to the final set...
                if(compoundExpr.length() != 0) { compoundExpr.append('|'); }
                compoundExpr.append('(').append(rawName).append(')');
                }
            catch(final IllegalArgumentException e) { } // Drop this name if it seems to be bad.
            }

        // If no regexes then return null.
        if(compoundExpr.length() == 0) { return(null); }

        // Return compiled expression...
        return(Pattern.compile(compoundExpr.toString(), Pattern.CASE_INSENSITIVE));
        }

    /**Normalise Set of normalised host names; result is immutable and not empty, or is null.
     * Normalises the set of names provided,
     * discarding any that seem to be invalid,
     * and stopping when/if we reach the size limit specified.
     * <p>
     * This skips any that seem to be regexes.
     *
     * @param rawNames  set of raw host names; if null then null is returned
     * @param maxSize  maximum size of result set; strictly positive
     */
    private static Set<String> _normaliseHostList(final Set<String> rawNames,
                                                  final int maxSize)
        {
        if((maxSize < 1))
            { throw new IllegalArgumentException(); }

        if(rawNames == null) { return(null); }

        final Set<String> rawResult = new HashSet<String>(1 + 2*Math.min(maxSize, rawNames.size()));

        for(final String rawName : rawNames)
            {
            try
                {
                if(_mayBeRegexHostname(rawName)) { continue; }
                final String normName = MemoryTools.intern(HostUtils.normaliseVirtualHostName(rawName));
                rawResult.add(normName); // Add to the result...
                if(rawResult.size() >= maxSize) { break; /* Stop when we have enough. */ }
                }
            catch(final IllegalArgumentException e) { } // Drop this name if it seems to be bad.
            }

        // Return null rather than an empty set.
        if(rawResult.size() == 0) { return(null); }

        // Make sure that the result is immutable.
        return(Collections.unmodifiableSet(rawResult));
        }

    /**Parse set of (space-separated) hostnames; returns null for an empty list.
     * This does not validate or normalise the names supplied.
     *
     * @param hostList  space-separate list of hostnames (or null for none)
     */
    private Set<String> _parseHostList(final String hostList)
        {
        if((hostList == null) || (hostList.length() == 0)) { return(null); }

        final StringTokenizer st = new StringTokenizer(hostList);
        if(st.countTokens() == 0) { return(null); }

        final Set<String> result = new HashSet<String>(1 + 2*st.countTokens());
        while(st.hasMoreTokens())
            { result.add(st.nextToken()); }

        return(result);
        }


    /**Maximum number of DNSRBLs that may be specified. */
    public static final int MAX_DNSRBLS = 8;

    /**Name of Web server DNSRBLs set. */
    public static final String PNAME_WEBSVR_DNSRBLs = "org.hd.d.pg2k.DNSRBLs";

    /**Immutable Set of DNS RBLs (normalised DNS zone names); can be null for no such hosts. */
    private /* final */ Set<String> DNSRBLs;

    /**Get immutable Set of DNS RBLs (normalised DNS zone names); never null. */
    public final Set<String> getDNSRBLs()
        {
        final Set<String> r = DNSRBLs;
        if(r == null)
            {
            // Return empty set if none.
            final Set<String> noHosts = Collections.emptySet();
            return(noHosts);
            }
        return(r);
        }

    }
