/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 27-May-02
 * Time: 09:32:06

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.props;

// We try to depend as little as possible on other classes in the
// system as we need to initialise early during bootstrapping.
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.location.GeoUtils;

import java.io.*;
import java.security.*;
import java.util.Properties;

import javax.crypto.SecretKey;

/**This is a set of properties local to the JVM.
 * This is the NEAR-minimum set of properties to do with security
 * or local system configuration (such as the local layout
 * of the file-system) that must not or cannot be sent over
 * the network and shared between master and slaves.
 * Much of this is bootstrapping data, such as the location of various
 * configuration and persistent-state directories.
 * Our aim is to minimise the need for other parts of the system
 * to get at random JVM system properties and files, etc, where possible.
 * <p>
 * Some of the more sensitive data are made available
 * via package-visible members so that the other props
 * classes can use them but nothing else can see the values
 * directly.
 * <p>
 * These properties are not Serializable and are never sent
 * across the network (unlike GenProps), and have to be
 * made available in a local file and/or system properties
 * for example.
 * (Do not include local properties file in the WAR or EAR file.)
 * <p>
 * These properties can be checked for being up-to-date
 * from time to time, and the in-memory values will be
 * refreshed if necessary.
 * <p>
 * Values set purely from system properties may be collected
 * once at class creation and not again.
 * <p>
 * All access to values is via static methods, and is
 * synchronized (on the class object) to ensure thread-safety.
 * <p>
 * This is necessarily a rag-tag of values used all over the system,
 * and will be loaded very early and should not depend on much of the
 * rest of the system if at all.
 */
public final class LocalProps
    {
    /**Minimum interval between rechecks of local properties file content (ms).
     * Since much of our data comes from the filesystem and many filesystems'
     * timestamp granularity is about one second, polling much faster than this
     * is pointless and may simply end up blocking the caller and wasting
     * CPU pointlessly.
     * <p>
     * (Arguably the only way that system properties can be changed is by
     * programmatic intervention want so we don't want to recheck such values
     * at all for fear of setting up a covert channel as well as just wasting
     * CPU time.)
     * <p>
     * We pick a primish retry interval to try to avoid collisions with other
     * regular events.
     * <p>
     * A value of a few seconds to a few tens of seconds is probably appropriate.
     */
    private static final int MIN_RECHECK_MS = 33007;

    /**The last time that we checked/refreshed the data that we hold.
     * We update this whenever we do a check whether we changed any values;
     * this avoids possibly expensive and unnecessary rechecking.
     * This value is private to us.
     * <p>
     * Initially zero to force immediate read on first use of these properties.
     */
    private static long lastRecheck;

    /**This routine rechecks and refreshes the data items if necessary.
     * This is called internally by all getter methods to lazily ensure that
     * data items are up to date.
     * <p>
     * When this has finished it sets lastRecheck and possibly timestamp;
     * if this fails to complete the update neither (but especially lastRecheck)
     * may be set.
     */
    private static synchronized void _recheck()
        {
        final long now = System.currentTimeMillis();

        if(now - lastRecheck <= MIN_RECHECK_MS)
            {
            // Not time for a recheck yet...
            return;
            }

        try
            {
            // Get location of local properties file, if extant.
            final File pf = new File(getLocalPropsFilename());

            final long fileTimestamp = pf.lastModified();

            // Warn user if file not found,
            // and suggest what settings they might use.
            if(fileTimestamp <= 0)
                {
System.err.println("WARNING: local properties file not found.");
System.err.println("  Set name in system property: " + PNAME_LOCAL_PROPS_FILENAME);
System.err.println("  or put file at: `" + DEFAULT_LOCAL_PROPS_FILENAME + "' .");
                }

            // Warn if the timestamp has gone backwards
            // (or to zero because the file has disappeared).
            if(fileTimestamp < timestamp)
                {
                System.err.println("WARNING: timestamp on local properties file has gone backwards or file has gone: " + pf);
                }

            // If the file has changed and is newer than our in-memory copy,
            // read it in.
            //
            // If we have problems reading the info,
            // we abort without updating any timestamps so that can try again
            // immediately.
            // We may have updated some of the properties...
            if(fileTimestamp > timestamp)
                {
                Properties props; // The properties that we load.
                try {
                    // Read the properties file in buffered chunks for efficiency.
                    final InputStream is = new BufferedInputStream(
                        new FileInputStream(pf));
                    try { props = FileTools.loadProperties(is); }
                    finally { is.close(); } // Release the handle ASAP...
                    }
                catch(final IOException e)
                    {
                    // If we encounter an I/O problem,
                    // log it to System.err,
                    // and return without updating our timestamps.
                    e.printStackTrace();
                    return;
                    }

                // Temporaries to help in parsing.
                int iTmp;
                long lTmp;
                String sTmp;

                // Get dir for read-only configuration files.
                confDir = MemoryTools.intern(props.getProperty(PNAME_CONF_DIR));

                // Get dir for data files.
                dataDir = MemoryTools.intern(props.getProperty(PNAME_DATA_DIR));

                // Get name of simple password file.
                simplePassfile = MemoryTools.intern(props.getProperty(PNAME_SIMPLEPASS_FILE));

                // Get name of upload directory.
                uploadDir = MemoryTools.intern(props.getProperty(PNAME_UPLOAD_DIR));

                // Get name of persistent-state dir.
                persistentStateDirFromSystemProperties = MemoryTools.intern(props.getProperty(PNAME_PERSISTENT_STATE_DIR));

                // Get mirror tag.
                // Check thoroughly for syntactic validity before assigning.
                // If invalid, then set to null instead.
                sTmp = props.getProperty(PNAME_MIRROR_TAG);
                // Reject invalid tags by returning a null.
                if((sTmp == null) || (sTmp.length() < 4) || (sTmp.length() > 255))
                    { sTmp = null; }
                // Check that the cc is valid; return null if not.
                else if(!GeoUtils.CCTLD.isSyntaticallyValidCcTLD(sTmp.substring(0, 2)))
                    { sTmp = null; }
                // Check that the dash is present; return null if not.
                else if(sTmp.charAt(2) != '-')
                    { sTmp = null; }
                // Check that the remainder of the tag is lower-case or digits; return null if not.
                else
                    {
                    for(int i = sTmp.length(); --i >= 3; )
                        {
                        final char c = sTmp.charAt(i);
                        if((c >= 'a') && (c <= 'z')) { continue; }
                        if((c >= '0') && (c <= '9')) { continue; }
                        sTmp = null;; // Bad character found so don't use this...
                        }
                    }
                mirrorTag = MemoryTools.intern(sTmp);


                iTmp = -1;
                try { iTmp = Integer.parseInt(props.getProperty(PNAME_MAX_UPLOAD_TOTAL_BYTES, "10123456"), 10); }
                catch(Exception e) { } // Ignore errors.
                uploadMaxBytesTotal = iTmp;

                iTmp = -1;
                try { iTmp = Integer.parseInt(props.getProperty(PNAME_MAX_UPLOAD_PER_USER_BYTES, "10123456"), 10); }
                catch(Exception e) { } // Ignore errors.
                uploadMaxBytesPerUser = iTmp;

                lTmp = -1;
                try { lTmp = Long.parseLong(props.getProperty(PNAME_WEBSVR_MAX_EX_CACHE_BYTES, "0"), 10); }
                catch(Exception e) { } // Ignore errors.
                WEBSVR_MAX_EX_CACHE_BYTES = lTmp;

                iTmp = -1;
                try { iTmp = Integer.parseInt(props.getProperty(PNAME_SERVER_SLOWDOWN_FACTOR, "1"), 10); }
                catch(Exception e) { } // Ignore errors.
                SERVER_SLOWDOWN_FACTOR = iTmp;

                // Get raw data protection key, if any.
                sTmp = props.getProperty(PNAME_XFER_HMAC_KEY);
                if(sTmp == null) { xferHMACKey = null; }
                else
                    {
                    sTmp = sTmp.trim();
                    if(sTmp.length() < 16)
                        {
                        xferHMACKey = null;
                        System.err.println("[WARNING: too-short simple data key ignored.]");
                        }
                    else
                        {
                        // OK, key looks superficially OK,
                        // so try to turn it into something usable for HMAC in our tunnel.
                        try
                            {
                            // Extract an entropy-dense form of the (trimmed) key supplied,
                            // ie with all of the goodness, but fewer bytes to process each time.
                            final byte hashed[] = MessageDigest.getInstance(CoreConsts.HASH_SHA1).
                                digest(sTmp.getBytes("UTF-8"));
                            // Home-brew immutable/tamper-resistent secret key.
                            xferHMACKey = new SecretKey(){
                                private static final long serialVersionUID = 3350400258615302664L;
                                public String getAlgorithm()    { return(""); }
                                public byte[] getEncoded()      { return(hashed.clone()); }
                                public String getFormat()       { return("RAW"); }
                                };
                            }
                        catch(final Exception e)
                            {
                            xferHMACKey = null;
                            System.err.println("[ERROR: LocalProps: unable to create/convert xferkey.]");
                            e.printStackTrace();
                            }
                        }
                    }
                }

            // Save the timestamp from the properties file
            // having successfully completed all that we wanted to.
            timestamp = fileTimestamp;

            // Update completed.
            lastRecheck = now;
            }
        catch(final SecurityException e)
            {
            if(!reportedSecurityException)
                {
                System.err.println("NOTE: no access to local properties file: " + e.getMessage());
                reportedSecurityException = true;
                }

            // Quietly absorb the security/permissions error,
            // and defer retrying for the normal length of time.
            lastRecheck = now;
            }
        }

    /**Set true if we've noted a SecurityException trying to read LocalProperties.
     * Marked volatile for lock-free thread-safe access.
     */
    private static volatile boolean reportedSecurityException;

    /**The last time our loaded values changed.
     * This can be used to cascade changes to parts of the system that depend
     * on these LocalProps.
     * <p>
     * Will never be newer than lastRecheck.
     * <p>
     * Initially zero, to indicate no (file-based) values loaded.
     * <p>
     * The timestamp is the latest timestamp of any of the files read to
     * make up the dataset.
     */
    private static long timestamp;

    /**Returns the timestamp of the current set of data values.
     * Forces a (re)load/refresh of data if necessary.
     * Zero indicates that no (file-based) values have yet been loaded.
     * <p>
     * The timestamp is the latest timestamp of any of the files read to
     * make up the dataset.
     */
    public static synchronized long getTimestamp()
        { _recheck(); return(timestamp); }

    /**Default suffix of local properties files, including the dot. */
    public static final String DEFAULT_PROPS_SUFFIX = ".properties";
    /**Default filename of local properties file. */
    public static final String DEFAULT_LOCAL_PROPS_FILENAME =
        "local" + DEFAULT_PROPS_SUFFIX;
    /**Name of props file for local config information property. */
    public static final String PNAME_LOCAL_PROPS_FILENAME = "org.hd.d.pg2k.localPropsFilename";
    /**Cache of name of props file for local config information; null if no explict system property value set. */
    private static final String localPropsFilenameFromSystemProperties;
    /**Initialise localPropsFilenameFromSystemProperties. */
    static
        {
        String s = null;
        try { s = System.getProperty(PNAME_LOCAL_PROPS_FILENAME); }
        catch(final SecurityException e) { /* Treat lack of permission as if lack of value. */ }
        localPropsFilenameFromSystemProperties = s;
        }
    /**Name of props file for local config information; never null.
     * It may be difficult for the system to run without this parameter
     * set correctly or if the corresponding file is not correctly set up.
     * <p>
     * This will specify a file in the current working directory if none
     * explicitly specified, and will deliver a warning if it cannot be found.
     * <p>
     * This might be bad news if the current working directory changes
     * or is inappropriate or unhelpful, so set the system property explicitly
     * if this is the case.
     * <p>
     * This does not check that the specified file actually exists.
     * <p>
     * We only use this internally; for example this may be inefficient to
     * continually recheck and we don't necessarily want to publicise the
     * actual location more than we must!
     */
    private static synchronized String getLocalPropsFilename()
        {
        // Use explicit value if set.
        String result = localPropsFilenameFromSystemProperties;
        // Else try to use caonicalised default name relative to pwd.
        if(result == null)
            {
            try {
                result = (new File(DEFAULT_LOCAL_PROPS_FILENAME)).getCanonicalPath();
                }
            catch(IOException e)
                {
                System.err.println("WARNING: cannot canonicalise local properties filename " + DEFAULT_LOCAL_PROPS_FILENAME);
                }
            }
        // Else try to use relative name directly.
        if(result == null)
            { result = DEFAULT_LOCAL_PROPS_FILENAME; }
        return(result);
        }

    /**Name of local properties parameter for name of subdirectory for (mainly read-only) data files. */
    public static final String PNAME_DATA_DIR = "pg2k.data.dir";
    /**Local cache of name of subdirectory for data files. */
    private static String dataDir;
    /**Get name of subdirectory for data files; never null.
     * If the PNAME_DATA_DIR local property is set then it is used,
     * else the CoreConsts.FS_DATA_ROOT value is used, canonicalised
     * if possible to give more meaningful error message at least.
     */
    public static synchronized String getDataDir()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use local property if set, ...
        String result = dataDir;
        // ... else use default from CoreConsts.
        if(result == null) { result = CoreConsts.FS_DATA_ROOT; }

        // If the directory name is not absolute,
        // make it relative to the directory the local properties are in.
        if(!(new File(result)).isAbsolute())
            {
            final File lpDir = new File(getLocalPropsFilename()).getParentFile();
            result = (new File(lpDir, result)).getAbsolutePath();
            }

        // In any case, try to canonicalise the path before returning it.
        try { result = (new File(result)).getCanonicalPath(); }
        catch(IOException e) { } // In case of error use non-canonical form...

        return(result);
        }

    /**Name of local properties parameter for name of subdirectory for read-only conf files. */
    public static final String PNAME_CONF_DIR = "pg2k.config.dir";
    /**Local cache of name of subdirectory for read-only conf files. */
    private static String confDir;
    /**Get name of subdirectory for read-only conf files; never null.
     * If the PNAME_CONF_DIR local property is set then it is used,
     * else the CoreConsts.FS_CONF_ROOT value is used, canonicalised
     * if possible to give more meaningful error message at least.
     */
    public static synchronized String getConfDir()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use local property if set, ...
        String result = confDir;
        // ... else use default from CoreConsts.
        if(result == null) { result = CoreConsts.FS_CONF_ROOT; }

        // If the directory name is not absolute,
        // make it relative to the directory the local properties are in.
        if(!(new File(result)).isAbsolute())
            {
            final File lpDir = new File(getLocalPropsFilename()).getParentFile();
            result = (new File(lpDir, result)).getAbsolutePath();
            }

        // In any case, try to canonicalise the path before returning it.
        try { result = (new File(result)).getCanonicalPath(); }
        catch(IOException e) { } // In case of error use non-canonical form...

        return(result);
        }


    /**Name of local properties parameter for name of simple password file. */
    public static final String PNAME_SIMPLEPASS_FILE =
        "pg2k.upload.simplepass.file";
    /**Local cache of name of simple password file; null if none. */
    private static String simplePassfile;
    /**Get name of simple password file; null if none.
     * There is no default for this; if it is not set
     * then there will be no simple passwords.
     * <p>
     * If the path supplied is not absolute,
     * it is made relative to the result of getConfDir().
     */
    public static synchronized String getSimplePassfile()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use local property if set, ...
        String result = simplePassfile;
        // ... else return null.
        if(result == null) { return(null); }

        // If the directory name is not absolute,
        // make it relative to the result of getConfDir().
        if(!(new File(result)).isAbsolute())
            {
            final File confDir = new File(getConfDir());
            result = (new File(confDir, result)).getAbsolutePath();
            }

        // In any case, try to canonicalise the path before returning it.
        try { result = (new File(result)).getCanonicalPath(); }
        catch(IOException e) { } // In case of error use non-canonical form...

        return(result);
        }


    /**Name of local properties parameter for name of upload dir. */
    public static final String PNAME_UPLOAD_DIR = "pg2k.upload.dir";
    /**Local cache of name of upload dir; null if none specified. */
    private static String uploadDir;
    /**Get name of upload dir or null if none was specified.
     * If the PNAME_UPLOAD_DIR local property is set then it is used.
     * <p>
     * If a directory is specified but is not absolute,
     * it is taken to be relative to the directory the local properties are in.
     * <p>
     * Warns if the specified directory is not present,
     * but returns the path anyway if a path was specifief;
     * the caller should check if it is usable.
     */
    public static synchronized String getUploadDir()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use local property if set.
        String result = uploadDir;

        if(result != null)
            {
            // If the directory name is not absolute,
            // make it relative to the directory the local properties are in.
            if(!(new File(result)).isAbsolute())
                {
                final File lpDir = new File(getLocalPropsFilename()).getParentFile();
                result = (new File(lpDir, result)).getAbsolutePath();
                }

            // In any case, try to canonicalise the path before returning it.
            try { result = (new File(result)).getCanonicalPath(); }
            catch(IOException e) { } // In case of error use non-canonical form...

            if(!(new File(result)).isDirectory())
                {
                System.err.println("WARNING: upload directory not present: " + result);
//                return(null); // Directory must be extant.
                }
            }

        return(result);
        }


    /**Name of Web server maximum-upload-space-per-user (bytes) property. */
    public static final String PNAME_MAX_UPLOAD_PER_USER_BYTES = "pg2k.upload.maxBytes.perUser";
    /**Web server maximum-upload-space-per-user (bytes). */
    private static int uploadMaxBytesPerUser;
    /**Get the Web server maximum-upload-space-per-user (bytes).
     * Constrained to the range 0 to 1GB, default approx 10MB.
     */
    public static synchronized int getUploadMaxBytesPerUser()
        { return(Math.max(0, Math.min(1123456789, uploadMaxBytesPerUser))); }

    /**Name of Web server maximum-upload-space-total (bytes) property. */
    public static final String PNAME_MAX_UPLOAD_TOTAL_BYTES = "pg2k.upload.maxBytes.total";
    /**Web server maximum-upload-space-total (bytes). */
    private static int uploadMaxBytesTotal;
    /**Get the Web server maximum-upload-space-total (bytes).
     * Constrained to the range 0 to 1GB, default approx 10MB.
     * Also constrained to be no smaller than uploadMaxBytesPerUser.
     */
    public static synchronized int getUploadMaxBytesTotal()
        { return(Math.max(getUploadMaxBytesPerUser(), Math.min(1123456789, uploadMaxBytesTotal))); }




    /**Name of local directory for store of persistent state property. */
    public static final String PNAME_PERSISTENT_STATE_DIR =
        "org.hd.d.pg2k.persistentStateDir";
    /**Local cache of name of local peristent-state directory; null if none. */
    private static String persistentStateDirFromSystemProperties;
    /**Relative path from local properties directory to persistent state dir. */
    public static final String DEFAULT_STATE_DIR = "persistentstate.dat";
    /**Get name of directory for storing local persistent state; never null.
     * This is used for storing (especially valuable) data that should survive
     * over many runs of the WAR/J2EE, but which is local to each WAR instance,
     * such as usage stats, though may also be used by the master instance to
     * persist user login data for example.
     * <p>
     * If the system property is not set, then we take this relative to the
     * directory in which we look for the local properties.
     * <p>
     * Warns if the specified directory is not present.
     */
    public final static synchronized String getPersistentStateDir()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use explicit value if set.
        String result = persistentStateDirFromSystemProperties;

        // Else try to use relative name directly.
        if(result == null) { result = DEFAULT_STATE_DIR; }


        // If the directory name is not absolute,
        // make it relative to the directory the local properties are in.
        if(!(new File(result)).isAbsolute())
            {
            final File lpDir = new File(getLocalPropsFilename()).getParentFile();
            result = (new File(lpDir, result)).getAbsolutePath();
            }

        // In any case, try to canonicalise the path before returning it.
        try { result = (new File(result)).getCanonicalPath(); }
        catch(IOException e) { } // In case of error use non-canonical form...


        if(!(new File(result)).isDirectory())
            {
            System.err.println("WARNING: persistent state directory not present: " + result);
            }
        return(result);
        }


    /**Name of mirror-tag property. */
    public static final String PNAME_MIRROR_TAG =
        "org.hd.d.pg2k.mirrorTag";
    /**Local cache of mirror tag; null if none. */
    private static String mirrorTag;
    /**Get valid "cc-xxxx" mirror-tag name for this host; null if none.
     * If this host is a mirror, it has a tag of the form <samp>cc-xxxxx</samp>
     * where "cc" is a two-letter country code
     * (as in a top-level ccTLD domain name, eg "uk" rather than "gb")
     * where the host resides for Internet connectivity purposes,
     * and "xxxx" is an alphanumeric tag indicating a particular host,
     * and the whole string is less than 256 characters.
     * <p>
     * The entire tag is a lower-case printable ASCII (no whitespace)
     * alphanumeric value.
     * <p>
     * This host should be reachable by the full name of
     * "mirror-cc-xxxx.gallery.hd.org"
     * for a tag of "cc-xxxx" and a main domain name of "gallery.hd.org".
     * <p>
     * This routine will not return an invalid/unsafe tag.
     * <p>
     * This routine does NOT attempt any DNS lookups or in other ways
     * attempt to verify that the full mirror host name exists or is valid.
     */
    public final static synchronized String getMirrorTag()
        {
        // Force properties reload if necessary.
        _recheck();

        return(mirrorTag);
        }



    /**Name of local directory for store of thumbnail data property. */
    public static final String PNAME_PERSISTENT_THUMBNAIL_DIR =
        "org.hd.d.pg2k.thumbnailDir";
    /**Local cache of name of local thumbnail directory; null if none. */
    private static final String thumbnailDirFromSystemProperties;
    /**Initialise thumbnailDirFromSystemProperties. */
    static
        {
        String s = null;
        try { s = System.getProperty(PNAME_PERSISTENT_THUMBNAIL_DIR); }
        catch(final SecurityException e) { /* Treat lack of permission as if lack of value. */ }
        thumbnailDirFromSystemProperties = s;
        }
    /**Relative path from exhibit directory to thumbnails dir. */
    public static final String DEFAULT_THUMBNAIL_DIR = "../_tn";
    /**Get name of directory for storing off-line master thumbnails; never null.
     * This is used for storing off-line computed thumbnails that should survive
     * over many runs of the WAR/J2EE, mainly intended for read-only
     * access by the master server, and write access by an off-line tool
     * to create thumbnails in a less restrictive environment than the WAS.
     * <p>
     * If the system property is not set, then we take this relative to the
     * directory in which we look for the exhibits.
     * <p>
     * If the property is set to "." then the exhibit directory is used.
     */
    public final static synchronized String getThumbnailRelDir()
        {
        // Use explicit value if set.
        String result = thumbnailDirFromSystemProperties;
        // Else try to use relative name directly.
        if(result == null)
            {
            result = DEFAULT_THUMBNAIL_DIR;
            }
        return(result);
        }


    /**Name of Web server maximum generic exhibit cache size (64-bit value, bytes). */
    public static final String PNAME_WEBSVR_MAX_EX_CACHE_BYTES = "pg2k.websvr.ex.maxcachesize";
    /**Web server maximum generic exhibit cache size (64-bit value, bytes) [can be overridden for a particular host]. */
    private static long WEBSVR_MAX_EX_CACHE_BYTES;
    /**Cached local property value (if any) as Long.
     * Set on first use (in the current VM); is transient so that
     * it is lost as this object moves from one VM to the next.
     * <p>
     * If null, then this has not yet been set (in the current VM).
     * <p>
     * If no suitable value exists in the properties then the local
     * value is stored here as a Long to be retrieved and used through
     * the same code-path.
     */
    private static Long _local_WEBSVR_MAX_EX_CACHE_BYTES;
    /**Get the Web server maximum generic exhibit cache size (64-bit value, bytes).
     * Constrained to the range approx 0 to 1TB, default approx 100MB.
     * <p>
     * This value is overridable with a system property of the same name
     * on the system <em>using</em> this value.
     */
    public static synchronized long getWEBSVR_MAX_EX_CACHE_BYTES()
        {
        _recheck();

        // Do we need to (re)compute the cached value?
        if(_local_WEBSVR_MAX_EX_CACHE_BYTES == null)
            {
            long computedValue = -1;

            try
                {
                // Fetch local override or cache portable value.
                final String lvS = System.getProperty(PNAME_WEBSVR_MAX_EX_CACHE_BYTES);
                if(lvS != null)
                    {
                    try { computedValue = Long.parseLong(lvS, 10); }
                    catch(final Exception e) { } // Ignore errors.
if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() local override from system property "+PNAME_WEBSVR_MAX_EX_CACHE_BYTES+" = " + computedValue + ".]"); }
                    }
                }
            catch(final SecurityException e)
                {
                // Silently ignore errors trying to read property.
                }

            // If there was no usable/valid system properties value,
            // use the value shipped with this instance.
            if(computedValue < 0)
                { computedValue = WEBSVR_MAX_EX_CACHE_BYTES; }

            // Take this opportunity to apply the limits exactly once
            // and atomically cache the constrained value.
            _local_WEBSVR_MAX_EX_CACHE_BYTES =
                new Long(Math.max(0L, Math.min(1123456123456L, computedValue)));
            }

        // Retrieve and use the already-constrained value.
        return(_local_WEBSVR_MAX_EX_CACHE_BYTES.longValue());
        }


    /**Name of server slowdown factor; strictly positive. */
    public static final String PNAME_SERVER_SLOWDOWN_FACTOR = "org.hd.d.pg2k.serverSlowdownFactor";
    /**Server slowdown factor [can be overridden for a particular host]. */
    private static int SERVER_SLOWDOWN_FACTOR;
    /**Get server-slowdown-factor; strictly positive.
     * Defaults to 1, and no more than about, say, 100.
     * <p>
     * Is a factor by which the server under-reports its capacity, etc.
     * <p>
     * If &gt; 1 then this has the effect of:
     * <ul>
     * <li>reducing aggressiveness of precaching,
     * <li>reducing reported available bandwidth,
     *     (and thus the chance of being picked as a user's "preferred" mirror)
     * </ul>
     * etc, for sites with constrained or expensive or non-dedicated bandwidth.
     */
    public static synchronized int getServerSlowdownFactor()
        {
        _recheck();

        // Coerce into reasonable range.
        return(Math.max(1, Math.min(100, SERVER_SLOWDOWN_FACTOR)));
        }
    

    /**Name of local properties parameter for HMAC key used to protect data in transit. */
    public static final String PNAME_XFER_HMAC_KEY =
        "org.hd.d.pg2k.xferkey";
    /**Local cache of immutable HMAC (secret) key used to protect data in transit; null if none. */
    private static SecretKey xferHMACKey;
    /**Get HMAC (secret) key used to protect data in transit; null if none.
     * There is no default for this; if it is not set
     * then there will be no data protection key.
     * <p>
     * The key returned is immutable.
     * <p>
     * The key should be handled with care, since it is sensitive data.
     */
    public static synchronized SecretKey getXferHMACKey()
        {
        // Force properties reload if necessary.
        _recheck();

        // Use local property if set.
        return(xferHMACKey);
        }
    }
