/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.props;

import java.io.*;
import java.util.*;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.AllExhibitPropertiesDelta.DiffException;
import org.hd.d.pg2k.svrCore.Tuple.Pair;

/**A diff/patch mechanism for a bundle of well-behaved simple String-->String Properties.
 * These diffs can be used to send just changes/deltas over the wire
 * for well-behaved bundles of closely-related pure-String Properties,
 * eg of the form of i18n ResourceBundle inputs.
 * <p>
 * This class itself is immutable and Serialisable,
 * and designed to be efficient on the wire.
 * <p>
 * This uses PropertiesDiff to hold each Properties instance immutably/efficiently.
 * 
 * @author DHD
 */
public class PropertiesBundleDiff implements Serializable, ObjectInputValidation
    {
    /**The immutable in-order set of Properties sets entirely removed; never null.
     * This field is NOT serialised directly.
     */
    private transient /*final*/ SortedSet<String> deletedNames;

    /**The immutable in-order map from items/names to the new/changed Properties sets; never null.
     * Note that we store the differences in each underlying Properties set
     * to efficiently represent small changes within each set.
     * <p>
     * This field is NOT serialised directly.
     */
    private transient /*final*/ SortedMap<String, PropertiesDiff> newValues;

    /**The input bundle count as a sanity-check; guaranteed non-negative. */
    public final int sizeBefore;

    /**The output bundle count  as a sanity-check; guaranteed non-negative. */
    public final int sizeAfter;

    /**Immutable empty diff. */
    public static final PropertiesBundleDiff EMPTY_DIFF = new PropertiesBundleDiff();

    /**Make new empty diff instance. */
    public PropertiesBundleDiff()
        {
        deletedNames = Collections.unmodifiableSortedSet(new TreeSet<String>());
        newValues = Collections.unmodifiableSortedMap(new TreeMap<String,PropertiesDiff>());
        sizeBefore = 0;
        sizeAfter = 0;
        // We assume that this does not need run-time validation with validateObject().
        }

    /**Create diff instance.
     * We defensively copy the potentially-mutable Collection argument.
     * @param sizeBefore  size of input Properties; non-negative
     * @param sizeAfter   size of output Properties; non-negative
     * @param deleted     named Properties sets present in input and not present in output; non-null
     * @param added       new/changed named Properties sets in output; non-null
     */
    public PropertiesBundleDiff(final int sizeBefore, final int sizeAfter,
                                final Set<String> deleted,
                                final Map<String,PropertiesDiff> added)
        {
        deletedNames = Collections.unmodifiableSortedSet(new TreeSet<String>(deleted));
        newValues = Collections.unmodifiableSortedMap(new TreeMap<String,PropertiesDiff>(added));
        this.sizeBefore = sizeBefore;
        this.sizeAfter = sizeAfter;
        try { validateObject(); }
        catch(final InvalidObjectException e) { throw new IllegalArgumentException(e); }
        }

    /**Get newValues; never null. */
    public SortedMap<String, PropertiesDiff> getNewValues()
        { return(newValues); }

    /**Create "change" values from a Properties bundle; never null.
     * This is the set of values to be added to an empty Properties value
     * to regenerate this one.
     * <p>
     * The first element in each Pair is the (non-null) Properties set name;
     * the second element is the (non-null) Properties set.
     * 
     * @param pb non-null pure String-->String properties set
     */
    private static Set<Tuple.Pair<String,Properties>> _computeChangeValues(final Map<String,Properties> pb)
        {
        final Set<Tuple.Pair<String,Properties>> result = new HashSet<Tuple.Pair<String,Properties>>(2*pb.size());
        for(final Map.Entry entry : pb.entrySet())
            {
            result.add(new Tuple.Pair<String,Properties>((String) entry.getKey(),
                                                         (Properties) entry.getValue()));
            }
        assert(result.size() == pb.size());
        return(result);
        }

    /**Private empty Properties map; never changed, never null. */
    private static final Map<String,Properties> EMPTY_BUNDLE = Collections.emptyMap();

    /**Create diff against empty Properties.
     * This is a safe, efficient, immutable, Serializable representation
     * for a single Properties instance,
     * where a raw Properties instance may not be safe/robust enough.
     * <p>
     * Forces intern()ing of the keys and values.
     */
    public static PropertiesBundleDiff createAsStandAloneDiff(final Map<String,Properties> p1)
        {
        try { return(createDiff(EMPTY_BUNDLE, p1, true, true)); }
        catch(final DiffException e) { throw new Error(e); } // Should never happen.
        }

    /**Create from diff against empty Properties.
     * Recovers Properties instance created by createAsDiff.
     */
    public static Map<String,Properties> createFromStandAloneDiff(final PropertiesBundleDiff pd)
        throws DiffException
        { return(applyDiff(EMPTY_BUNDLE, pd)); }

    /**Private empty Properties map; never changed, never null. */
    private static final Properties EMPTY_PROPERTIES = new Properties();

    /**Create diff between two Properties instances.
     * This creates a diff to be applied to an instance of the first argument
     * to recreate the second argument.
     * <p>
     * This will refuse to create a diff if it seems that
     * the diff is unlikely to be useful, for example:
     * <ul>
     * <li>If one properities set has more than twice the entries of the other
     *     (the diff may not be smaller than sending the second properties set whole).
     * </ul>
     *
     * @param force  if true then force a diff to be produced
     *     even if this routine would normally refuse to do so on efficiency grounds
     * @param intern  if true than force intern()ing of all String values
     *
     * @throws DiffException  if no diff can be generated
     *     or is unlikely to be worthwhile to use
     *     (eg would be more than a fraction of the size of the second AEP)
     */
    public static PropertiesBundleDiff createDiff(final Map<String,Properties> p1,
                                                  final Map<String,Properties> p2,
                                                  final boolean force,
                                                  final boolean intern)
        throws DiffException
        {
        if((null == p1) || (null == p2))
            { throw new IllegalArgumentException(); }

        if(!force)
            {
            if((p1.size() > 2*p2.size()) ||
               (p2.size() > 2*p1.size()))
                { throw new DiffException("sizes too different (force==false)"); }
            }

        final Set<Tuple.Pair<String,Properties>> p1c = _computeChangeValues(p1);
        final Set<Tuple.Pair<String,Properties>> p2c = _computeChangeValues(p2);

        // The items that we will need to have in our "added" list.
        final Set<Tuple.Pair<String,Properties>> added = new HashSet<Tuple.Pair<String,Properties>>(p2c);
        added.removeAll(p1c);

        // The minimal set of names that we must remove from the input properties set.
        final Set<Object> deleted = new HashSet<Object>(p1.keySet());
        deleted.removeAll(p2.keySet());

        // If there is probably not enough in common between the two instances
        // for a diff to be likely to be efficient, then abort.
        if(!force && (added.size() + deleted.size() > p2.size()/2))
            { throw new DiffException("exhibit sets too different (force==false): "+(added.size() + deleted.size())); }

        // Create the "deleted" list as a Set of the names.
        final Set<String> deletedNames = new HashSet<String>(2*deleted.size());
        if(intern)
            { for(final Object d : deleted) { deletedNames.add(MemoryTools.intern((String) d)); } }
        else
            { for(final Object d : deleted) { deletedNames.add((String) d); } }

        // We transform the "added"/changed names into a Map,
        // including the diff from the previous version of the specific named set
        // (if there was no previous version, we use an empty Properties set).
        final Map<String,PropertiesDiff> addedMap = new HashMap<String, PropertiesDiff>(2*added.size());
        for(final Tuple.Pair<String,Properties> entry : added)
            {
            final String name = entry.first;
            final Properties oldPS = p1.get(name);
            final PropertiesDiff pd = PropertiesDiff.createDiff(((null == oldPS) ? EMPTY_PROPERTIES : oldPS), entry.second, true, intern);
            addedMap.put((intern ? MemoryTools.intern(name) : name), pd);
            }
        assert(added.size() == addedMap.size());

        // Create the diff object.
        final PropertiesBundleDiff result = new PropertiesBundleDiff(
                        p1.size(),
                        p2.size(),
                        deletedNames,
                        addedMap);
        return(result);
        }

    /**Applies diff to an extant Properties instance to generate a new AEP instance; never null.
     * The diff must only be applied to the same Properties instance value
     * that the diff was generated from (ie an equal one).
     * <p>
     * We do very basic/quick sanity checks on the before/after sizes
     * to catch misapplication of a diff to a clearly-inappropriate Properties set.
     *
     * @throws DiffException  if the diff cannot be applied
     */
    public static Map<String,Properties> applyDiff(final Map<String,Properties> bp1,
                                                   final PropertiesBundleDiff diff)
        throws DiffException
        {
        if((null == bp1) || (null == diff))
            { throw new IllegalArgumentException(); }

        // Make sure that the diff "before" values match those from the input.
        if(diff.sizeBefore != bp1.size())
            { throw new DiffException("diff cannot be applied to these Properties"); }

        // Compute mutable Set of the Change values for the input AEP.
        // We will mutate this to become the full exhibit data set for the result.
        final Set<Tuple.Pair<String,Properties>> p1c = new HashSet<Tuple.Pair<String,Properties>>(_computeChangeValues(bp1));

        // Now synthesise a new Properties bundle.
        final Map<String,Properties> result = new HashMap<String,Properties>(2 * diff.sizeAfter);

        // Copy the existing items into the result,
        // except where they are marked as deleted.
        // If there were some deleted items,
        // then zap them here.
        // They must ALL be present in the input.
        final Set<String> del = diff.deletedNames;
        final int initialSize = p1c.size();
        for(final Iterator<Tuple.Pair<String,Properties>> cit = p1c.iterator(); cit.hasNext(); )
            {
            final Tuple.Pair<String,Properties> c = cit.next();
            // Remove this name if noted as deleted.
            if(del.contains(c.first))
                { cit.remove(); }
            else
                { result.put(c.first, c.second); }
            }
        // All "deleted" exhibits must exist in the initial AEP.
        if(initialSize - del.size() != p1c.size())
            { throw new DiffException("invalid disjunction/gap between initial and deleted items: initialSize="+initialSize+", delSize="+(del.size())+", trimmedSize="+p1c.size()); }

        // If there were some added/changed items,
        // then add them (or apply diffs) here.
        for(final Map.Entry<String,PropertiesDiff> entry : diff.newValues.entrySet())
            {
            final String name = entry.getKey();
            final Properties oldP = result.get(name);
            final PropertiesDiff pd = entry.getValue();
            final Properties newP = PropertiesDiff.applyDiff((oldP == null) ? EMPTY_PROPERTIES : oldP, pd);
            result.put(name, newP);
            }

        if(result.size() != diff.sizeAfter)
            { throw new DiffException("unable to reconstruct aeid correctly"); }

        return(result);
        }



    /**Returns true iff the diff is "empty", ie no deletions nor additions/changes. */
    public boolean isEmpty()
        { return(deletedNames.isEmpty() && newValues.isEmpty()); }

    /**Equal iff all fields are identical. */
    @Override public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof PropertiesBundleDiff)) { return(false); }
        final PropertiesBundleDiff other = (PropertiesBundleDiff) obj;
        if(sizeBefore != other.sizeBefore) { return(false); }
        if(sizeAfter != other.sizeAfter) { return(false); }
        return(deletedNames.equals(other.deletedNames) &&
                        newValues.equals(other.newValues));
        }

    /**Cached hash value; initially zero.
     * We cache this to ensure that hashCode() always remains quick.
     * <p>
     * Since access is atomic, this need not be marked volatile.
     * <p>
     * Never set to zero once non-zero.
     */
    private transient int _hash;

    /**Hash code depends on all deleted/changed elements; guaranteed zero for an empty diff, non-zero otherwise. */
    @Override public int hashCode()
        {
        // Try to use the cached value, if any.
        int h = _hash;
        if(h != 0) { return(h); }

        // Recompute as apparently not cached.
        // Should very rarely be genuinely zero (causing wasteful recompution)
        // unless the structures are empty and thus quick to rehash anyway.
        //
        // Note that hashCode for empty Set/Map is zero by definition,
        // so if deletedNames and newValues are empty our result is zero.
        // The "newValues" is often the more interesting component.
        h = newValues.hashCode() + (67021*deletedNames.hashCode());
        if((h == 0) && !isEmpty()) { h = 1; }
        return(_hash = h);
        }

    /**Human-readable summary of state. */
    @Override public String toString()
        {
        return("PropertiesBundleDiff:deletedNames#"+deletedNames.size()+":newValues#"+newValues.size());
        }


    /**Write out a less-redundant form of our internal information.
     * In particular, since deltas will not usually involve deletion of any values,
     * we make the no-deletions case take no space at all in the serialised form.
     */
    private void writeObject(final ObjectOutputStream oos)
        throws IOException
        {
        // Write the fields that we are not trying to optimise.
        // Note that this includes our length field.
        oos.defaultWriteObject();

        // Send the deleted names as bare Strings in sorted order
        // to help with compression on the wire.
        final int sizeD = deletedNames.size();
        final String outD[] = new String[sizeD];
        if(sizeD > 0) { deletedNames.toArray(outD); }
        // We don't even write a length.
        // The presence of the first non-String value is sufficient delimiter.
        for(final String dn : outD)
            { oos.writeObject(dn); }

        // Send the added/changed Properties sets in (name) sorted order
        // to help with compression on the wire.
        // We write the names and then the values.
        final int sizeA = newValues.size();
        final String outAK[] = new String[sizeA];
        final PropertiesDiff outAV[] = new PropertiesDiff[sizeA];
        if(sizeA > 0)
            {
            newValues.keySet().toArray(outAK);
//            newValues.values().toArray(outAV);
            // Ensure correct correspondence between key index and value index.
            for(int i = sizeA; --i >= 0; )
                { outAV[i] = newValues.get(outAK[i]); }
            }
        // We preserve the relative ordering so as not to scramble the mapping!
        oos.writeObject(outAK);
        oos.writeObject(outAV);
        }

    /**Deserialise.
     * This intern()s all the Properties set names.
     */
    private void readObject(final ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        // Read the fields that we are not trying to optimise.
        ois.defaultReadObject();

        // Get the deleted property names
        // reading until we hit the change/additions array.
        Object objectIn = null;
        final HashSet<String> dn = new HashSet<String>();
        while((objectIn = ois.readObject()) instanceof String)
            {
            if(!dn.add(MemoryTools.intern((String) objectIn)))
                { throw new InvalidObjectException("duplicate 'deleted' names"); }
            }
        deletedNames = Collections.unmodifiableSortedSet(
                        new TreeSet<String>(dn));

        // Get the added/changed name-value details.
        // Must be in the non-String item that we just read.
        final String inAK[] = (String[]) objectIn;
        final PropertiesDiff inAV[] = (PropertiesDiff[]) ois.readObject();
        if(inAK.length != inAV.length) // Implicitly checks non-null too...
            { throw new InvalidObjectException("damaged new/changed list"); }
        final SortedMap<String,PropertiesDiff> nv = new TreeMap<String,PropertiesDiff>();
        for(int i = inAK.length; --i >= 0; )
            { nv.put(MemoryTools.intern(inAK[i]), inAV[i]); }
        newValues = Collections.unmodifiableSortedMap(nv);

        // Validate the results.
        validateObject();
        }

    /**Deserialise: discard duplicate empty values. */
    protected Object readResolve()
        throws java.io.ObjectStreamException
        {
    	if(equals(EMPTY_DIFF)) { return(EMPTY_DIFF); }
        return(this);
        }

    /**Checks that the object is internally consistent. */
    public void validateObject() throws InvalidObjectException
        {
        if((sizeBefore < 0) || (sizeAfter < 0))
            { throw new InvalidObjectException("bad object: -ve size"); }

        if((deletedNames == null) || (newValues == null))
            { throw new InvalidObjectException("bad object: null deletedNames or newValues"); }

        final HashSet<String> overlap = new HashSet<String>(deletedNames);
        overlap.retainAll(newValues.keySet());
        if(!overlap.isEmpty())
            { throw new InvalidObjectException("bad object: deletedNames and newValues overlap: "+overlap.size()); }            

        try 
            {
            if(deletedNames.contains(null))
                { throw new InvalidObjectException("bad object: deletedNames contains null value"); }
            }
        catch(final NullPointerException e) { /* Set does not support null values; good. */ }

        try 
            {
            if(newValues.containsKey(null))
                { throw new InvalidObjectException("bad object: newValues contains null key"); }
            }
        catch(final NullPointerException e) { /* Map does not support null keys; good. */ }

        try 
            {
            if(newValues.containsValue(null))
                { throw new InvalidObjectException("bad object: newValues contains null value"); }
            }
        catch(final NullPointerException e) { /* Map does not support null values; good. */ }
        }

    /**Serial UID.*/
    private static final long serialVersionUID = 4307342238100184047L;


    /**Load a complete bundle of properties files with timestamp in one go; never null.
     * The timestamp is that of the newest (most-recently-modified) bundle member.
     * <p>
     * We load all files of form <code>basename</code>*<code>.properties</code>
     * in the given directory. 
     * 
     * @param dir  directory in which properties files are to be found; non-null
     * @param basename  base name of properties files; non-null
     * 
     * @throws FileNotFoundException  if directoiry not found or no bundle files found
     * @throws IOException  in case of difficulty loading bundle files
     */
    public static Tuple.Pair< Map<String, Properties>, Long> loadBundle(final File dir,
                                                                        final String basename)
        throws IOException
        {
        if((dir == null) || (basename == null))
            { throw new IllegalArgumentException(); }
        if(!dir.exists() || !dir.isDirectory())
            { throw new FileNotFoundException("cannot open bundle dir "+dir); }

        // Get all files in indicated directory.
        final String fileSuffix = ".properties";
        final String[] entries = dir.list(new FilenameFilter(){
            public boolean accept(final File d, final String name)
                {
                return(name.startsWith(basename) && name.endsWith(fileSuffix) &&
                                (new File(d, name)).isFile());
                }
            });

        long timestamp = -1; // No timestamp yet.
        final Map<String, Properties> bundle = new HashMap<String, Properties>();
        final int basenameLen = basename.length();
        final int fileSuffixLen = fileSuffix.length();
        for(final String f : entries)
            {
            final File path = new File(dir, f);
            final Pair<Properties, Long> props = PropertiesDiff.loadFromFile(path);
            final String shortName = f.substring(basenameLen, f.length() - fileSuffixLen);
            bundle.put(shortName, props.first);
            final long ts = props.second.longValue();
            if(ts > timestamp) { timestamp = ts; }
            }
 
        if((timestamp <= 0) || (bundle.isEmpty()))
            { throw new FileNotFoundException("no suitable files in bundle dir "+dir); }
        return(new Tuple.Pair< Map<String, Properties>, Long>(bundle, new Long(timestamp)));
        }
    }
