/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.props;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.FileTools;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;

/**This holds some security properties for the server.
 * These data are potentially sensitive, such as authentication keys for slaves,
 * but not massively secret (such as personal data), but we take care not to
 * ship them across any public network.
 * We also make it a little harder to get at some of
 * the security information even when in possession of
 * an instance of this object.
 * <p>
 * We are prepared to hand out a (read-only) copy of any properties
 * with prefix GENSECPROPS_PREFIX for pluggable security
 * components to examine.  This data can be shipped over
 * the network, but should be protected if possible.
 * <p>
 * FIXME: Should do validation on deserialisation.
 */
public final class SecurityProps implements java.io.Serializable
    {
    /**Construct a default, zero-timestamp set of generic properties.
     */
    public SecurityProps()
        { this(new Properties(), 0); }

    /**Construct a new, immutable, properties set.
     * The properties must be non-null
     * and the timestamp must be non-negative.
     * <p>
     * Minor problems with the properties themselves will
     * be silently ignored, and defaults substituted
     * for broken or missing values.
     */
    public SecurityProps(final Properties props, final long _timestamp)
        {
        if((props == null) || (_timestamp < 0))
            { throw new IllegalArgumentException(); }
        timestamp = _timestamp;

        // Parse properties.
        // Set up some working variable to help with parsing.
//        int iTmp;
        String sTmp;

//        iTmp = -1;
//        try { iTmp = Integer.parseInt(props.getProperty(PNAME_WEBSVR_MIN_EX_IMATTR_RECHECK_MS, "120000"), 10); }
//        catch(Exception e) { } // Ignore errors.
//        // Constrain to range approx 1s to 6h, default 2 mins.
//        WEBSVR_MIN_EX_IMATTR_RECHECK_MS = Math.max(1001, Math.min(6 * 3600 * 1000, iTmp));

        sTmp = null;
        try { sTmp = props.getProperty(PNAME_TUNNEL_CLIENT_LIST, "").trim(); }
        catch(Exception e) { } // Ignore errors.
        // Default is no clients accepted.
        TUNNEL_CLIENT_LIST = sTmp;


        // Set up GENSECPROPS_PREFIX set.
        final Properties gs = new Properties();
        for(final Enumeration en = props.propertyNames(); en.hasMoreElements(); )
            {
            final String name = (String) en.nextElement();
            if(name.startsWith(GENSECPROPS_PREFIX))
                { gs.put(name, props.getProperty(name)); }
            }
        gs.put(PNAME_GENSECPROPS_TIMESTAMP, String.valueOf(_timestamp));
        gensec = gs;
        }

    /**Timestamp of this properties set. */
    public final long timestamp;

    /**Get (private) security properties from filesystem.
     * This uses LocalProps to get the root of the file area.
     * <p>
     * This does not cache its result but will return null if the stamp
     * passed in is the same as the file and not -1.
     * <p>
     * The properties retrieved this way are not to be transmitted out of
     * this JVM across the network as they may be private/sensitive.
     * The most sensitive items may be marked transient to prevent them
     * being shipped out of a JVM, but that may make them impossible to
     * use throughout a multi-VM J2EE system, for example.
     *
     * @throws IOException  if security props cannot be read/parsed
     */
    public static SecurityProps getSecurityPropsUncachedFromFilesystem(final long oldStamp)
        throws IOException
        {
        final File file = new File(LocalProps.getConfDir(), CoreConsts.FS_CONF_SECPROPS);

        // If the stamp presented is < 0 or different to that of the file,
        // reread the file now.
        final long newStamp = file.lastModified();
        if((oldStamp < 0) || (oldStamp != newStamp))
            {
            try
                {
                final FileInputStream in = new FileInputStream(file);
                try
                    {
                    final Properties props = FileTools.loadProperties(
                        new BufferedInputStream(in));
                    return(new SecurityProps(props, newStamp));
                    }
                finally
                    {
                    in.close(); // Ensure that file is closed.
                    }
                }
            catch(final IOException e)
                {
                e.printStackTrace(); // Dump a diagnostic...
                throw e; // Pass the error on.
                }
            }

        return(null); // Caller's copy seems to be up-to-date.
        }

    /**Prefix of subset of properties we hand out with getGenSecProps. */
    public static final String GENSECPROPS_PREFIX = "pg2k.gensec.";
    /**Name of the timestamp property (a long; copied from the SecurityProps object) that we insert in the GENSECPROPS_PREFIX properties. */
    public static final String PNAME_GENSECPROPS_TIMESTAMP = GENSECPROPS_PREFIX + "timestamp";
    /**Internal private copy of GENSECPROPS_PREFIX properties. */
    private final Properties gensec;
    /**Get a private copy of the GENSECPROPS_PREFIX properties.
     * Wrapping an existing Properties object as the defaults
     * for an otherwise empty new Properties set is assumed to
     * make a private copy with copy-on-write style semantics.
     */
    public java.util.Properties getGenSecProps() { return(new Properties(gensec)); }

    /**Name of tunnel-client list property. */
    public static final String PNAME_TUNNEL_CLIENT_LIST = "pg2k.sec.tunnelclients";
    /**Web server permitted tunnel client list by IP address, space separated, never null. */
    private final String TUNNEL_CLIENT_LIST;

    /**Check if the IP of a given tunnel client is OK.
     * We <em>do not</em> just disclose the whole list.
     * <p>
     * We take each listed allowed address in turn by parsing from
     * the source string.  We convert it to an InetAddress,
     * skipping any that cause errors.  If any of the addresses
     * corresponding to an item on the list match any of the
     * addresses passed (assumed to be a raw IP address with no name),
     * then this returns OK.
     * <p>
     * TODO: possibly needs optimisation as may be very inefficient/slow.
     */
    public final boolean tunnelClientIsOK(final InetAddress addr)
        {
        final StringTokenizer st = new StringTokenizer(TUNNEL_CLIENT_LIST);
        while(st.hasMoreTokens())
            {
            try {
                final String name = st.nextToken();
                final InetAddress[] ia = InetAddress.getAllByName(name);
                if(ia == null) { continue; }
                for(int i = ia.length; --i >= 0; )
                    {
                    // If we have a match, we are done!
                    if(ia[i].equals(addr)) { return(true); }
                    }
                }
            // Silently ignore any errors.
            catch(final Exception e) { }
            }

        // No match, so rejected...
        return(false);
        }


    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -7930248582951598929L;
    }
