/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 27-May-02
 * Time: 23:23:08

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.props;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.FileTools;

import java.io.*;
import java.util.*;

/**Encapsulates the simplepass password file.
 * This can not be serialised and can only be constructed
 * from the file indicated by LocalProps.getSimplePassfile().
 */
public final class SimplepassProps
    {
    /**Returns true if the quoted password is correct for the given author.
     * This can only return true if simplepass passwords are enabled
     * and are set up correctly in LocalProps and in a simplepass file.
     * <p>
     * Will return false if either parameter is null or zero-length.
     *
     * @param auth  the author initials; must be correct syntactically
     * @param pass  the password; must be of a valid length
     * @return true if the quoted password is correct
     */
    public final static boolean isAuthorUploadPasswordCorrect(final String auth,
                                                              final String pass)
        {
        final SimplepassProps spp = getSimplepassProps();
        // If no simplepass, always fail.
        if(spp == null) { return(false); }
        // Check against actual passwords.
        return(spp.authorUploadPasswordCorrect(auth, pass));
        }

    /**The current singleton SimplepassProps object, if any.
     * This may be null if USE_SIMPLEPASS is null or if there
     * is no simplepass value set in LocalProps or there is no file.
     */
    private static SimplepassProps singleton;

    /**Get the current SimplepassProps object, if any.
     * This object normally exists as a singleton,
     * being replaced only if the simplepass file changes
     * and can be read.
     */
    private static synchronized SimplepassProps getSimplepassProps()
        {
        // If simple passwords are enabled, we don't create a singleton.
        if(!USE_SIMPLEPASS) { return(null); }

        final String passfileName = LocalProps.getSimplePassfile();
        if(passfileName != null)
            {
            final File pf = new File(passfileName);
            final long timestamp = pf.lastModified();

            if(pf.isFile())
                {
                // If the singleton object may be out of sync,
                // try to create a new one.
                if((singleton == null) || (singleton.timestamp != timestamp))
                    {
                    try { singleton = new SimplepassProps(pf); }
                    catch(IOException e)
                        {
                        // In case of error, dump the problem.
                        e.printStackTrace();
                        }
                    }
                }
            }

        return(singleton);
        }

    /**Construct a new, immutable, properties set.
     * The properties must be non-null
     * and the timestamp must be non-negative.
     * <p>
     * Minor problems with the properties themselves will
     * be silently ignored, and defaults substituted
     * for broken or missing values.
     * <p>
     * This is only designed to be fetched by the factory method.
     *
     * @param pf  is the passfile
     */
    private SimplepassProps(File pf)
        throws IOException
        {
        Properties props;

        // Read the properties file in buffered chunks for efficiency.
        final InputStream is = new BufferedInputStream(
            new FileInputStream(pf));
        try { props = FileTools.loadProperties(is); }
        finally { is.close(); } // Release the handle ASAP...

        // Capture the timestamp.
        timestamp = pf.lastModified();

        // Parse properties.

        // Set up simplepassMap.
        if(!USE_SIMPLEPASS) { simplepassMap = null; }
        else
            {
            // Use a HashMap for speed of lookup,
            // though speed is probably not critical.
            final HashMap<String,String> hm = new HashMap<String,String>();
            for(final Enumeration en = props.propertyNames(); en.hasMoreElements(); )
                {
                final String name = (String) en.nextElement();
                if(name.startsWith(SIMPLEPASS_PREFIX))
                    {
                    // Extract author initials and ensure that they are valid.
                    final String author =
                        name.substring(SIMPLEPASS_PREFIX.length());
                    if(!ExhibitName.validAuthorSyntax(author))
                        {
                        System.err.println("ERROR: invalid simplepass author in: "+name);
                        continue;
                        }
                    final String pass = props.getProperty(name);
                    // We could validate the password syntax/length here...
                    if((pass.length() < CoreConsts.MIN_PASSWORD_LEN) ||
                       (pass.length() > CoreConsts.MAX_PASSWORD_LEN))
                        {
                        System.err.println("ERROR: invalid simplepass password in: "+name);
                        continue;
                        }
                    if(hm.put(author, pass) != null)
                        {
                        System.err.println("ERROR: duplicate simplepass author in: "+name);
                        continue;
                        }
                    }
                }
            // Save immutable reference to map.
            simplepassMap = Collections.unmodifiableMap(hm);
            }
        }

    /**Timestamp of this properties set. */
    public final long timestamp;


    /**If true, allow use of pg2k.upload.simplepass.XXX simple passwords.
     * This is a primative password system for author exhibit uploads.
     * <p>
     * These will only be used in any case if a more advanced and safer
     * password store is not available.
     * <p>
     * Note that these passwords are:
     * <ul>
     * <li>plaintext
     * <li>not maintainable/changeable by the user
     * <li>only updatable slowly (when the props are re-read)
     * </ul>
     * so should be considered generally unacceptable for normal use.
     */
    private static final boolean USE_SIMPLEPASS = true;

    /**The "simplepass" prefix, including the trailing dot.
     * This is followed by the author initials with the value
     * being the password.  The password value must meet constraints
     * set elsewhere (in the code) such as maximum and minimum length.
     */
    private static final String SIMPLEPASS_PREFIX = "pass.";

    /**The immutable private map from author initials to plaintext password.
     * It is a shame that these will be floating around in the heap
     * unprotected.
     * <p>
     * This map is threadsafe because it is read-only.
     * <p>
     * The map is String to String.
     * <p>
     * Will be null if USE_SIMPLEPASS is false.
     * <p>
     * We don't directly disclose passwords, just check if they are valid
     * or not.  We hold a lock on the password while checking
     * a password, and if it is wrong we wait a while still holding
     * the lock, to make dictionary attacks harder at the risk of making
     * some denial of service attacks easier.
     */
    private final Map<String,String> simplepassMap;

    /**How long to wait (ms) on an unsuccessful lookup. */
    private static final int SIMPLEPASS_FAILURE_PAUSE = 2000;

    /**Returns true if presented author upload password is correct.
     * The author name must be valid and the password non-null and
     * non-zero-length to avoid trivial typo-related attacks.
     */
    public final boolean authorUploadPasswordCorrect(String auth, String pass)
        {
        if(!ExhibitName.validAuthorSyntax(auth)) { return(false); }
        if((pass == null) || (pass.length() == 0)) { return(false); }

        final String p = (String) simplepassMap.get(auth);

        try {
            // If author not present at all, just wait.
            if(p == null)
                { Thread.sleep(SIMPLEPASS_FAILURE_PAUSE); }
            // Author is present, so wait on its password.
            else synchronized(pass)
                {
                if(p.equals(pass)) { return(true); } // Correct password!
                Thread.sleep(SIMPLEPASS_FAILURE_PAUSE);
                }
            }
        catch(InterruptedException e) { } // Ignore.

        return(false); // Wrong password.
        }
    }
