/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 15-May-02
 * Time: 07:53:03

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.stats;

import org.hd.d.pg2k.svrCore.SimpleLoggerIF;

import java.io.PrintStream;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

/**Stats-gathering interface.
 * This is not designed for especially sensitive information.
 * <p>
 * The StatsLogger class is not instantiated.
 * <p>
 * The simplest interface requires the caller to hand a config object
 * in on every call to specify how to handle the stats.
 * <p>
 * On each stats-event call, the current event may be dumped immediately
 * and/or, if a stats summary is (over)due then it may be done now.
 */
public final class StatsLogger
    {
    /**No one may instantiate this class. */
    private StatsLogger() { }

    /**Map from StatsConfig item to collected stats.
     * The map is a WeakHashMap so that if no one has a reference
     * to a given StatsConfig key object then the stats data keyed from
     * it are automatically discarded.
     * <p>
     * This is held statically and all access to it should be
     * performed within a lock on the map object if multiple accesses
     * must be performed atomically.
     * <p>
     * This table is a synchronized WeakHashMap from StatsConfig to
     * a thread-safe map from some Object key to a thread-safe counter.
     * <p>
     * We don't perform non-essential operations on the per-set Hashtables
     * inside the dataStore lock so as to maximise concurrency.
     * <p>
     * Note that entries in a WeakHashMap can disappear even while
     * a lock is held on the map.
     */
    private static final Map<StatsConfig,ConcurrentHashMap<String,AtomicLong>> dataStore =
        Collections.synchronizedMap(new WeakHashMap<StatsConfig, ConcurrentHashMap<String,AtomicLong>>());

    /**Holds the configuration for one statistics set.
     * As this object uses == for equals() [from Object],
     * it is impossible to forge duplicates for any one stats source.
     * <p>
     * While immutable from outside the stats system,
     * this does note when the last stats summary was done.
     */
    public static final class StatsConfig
        {
        /**Time at which this stats set was created. */
        public final long creationTime = System.currentTimeMillis();

        /**Name/token to identify this stats set; never null or zero length. */
        public final String name;

        /**Logger to log to; null if none. */
        private final SimpleLoggerIF logger;

//        /**Stream to log to, or null if none. */
//        private final PrintStream logStream;

        /**If true we log every event, else only summary status. */
        private final boolean printEveryEvent;

        /**Interval in seconds between dumps of stats info; non-negative.
         * If zero then periodic dumps are disabled.
         */
        private final int dumpIntervalS;

        /**If true, we dump at shorter intervals to start with.
         * This helps yield useful info when the system is starting up;
         * until the first true dump interval has been reached
         * we dump at exponentially-increasing intervals
         * (driven by actual arrival of events).
         */
        private final boolean adaptiveTime;

        /**Time at which last summary dump was done.
         * Initally zero to indicate no previous dump.
         * <p>
         * Atomically read and/or updated by timeForSummary().
         */
        private long lastSummaryDumpTime;

        /**If stats summary dump is overdue, returns true and updates time.
         * This applies the adaptive algorithm if appropriate.
         */
        private synchronized boolean timeForSummary()
            {
            // If summary dumps are disabled
            // then return false immediately.
            if(dumpIntervalS == 0) { return(false); }

            final long now = System.currentTimeMillis();
            final long sinceLastDump = (now - lastSummaryDumpTime);

            // If adaptive dump-interval algorithm is on,
            // and this stats collector is still new,
            // then allow a dump depending on previous dump time...
            if(adaptiveTime)
                {
                final long timeSinceStart = now - creationTime;
                // Regard stats collection as "new"
                // if it has not yet reached its "normal" dump time.
                if(timeSinceStart < (dumpIntervalS*1000L))
                    {
                    // Allow dump if more than half life of this collection
                    // since the last one.
                    if(sinceLastDump > timeSinceStart / 2)
                        {
                        lastSummaryDumpTime = now;
                        return(true);
                        }
                    }
                }

            // If long enough since last summary,
            // update time and return true.
            if(sinceLastDump >= dumpIntervalS * 1000L)
                {
                lastSummaryDumpTime = now;
                return(true);
                }

            // Not yet time for a new summary dump.
            return(false);
            }

        /**Create a new stats sink description.
         * Uses an adaptive algorithm so as to dump stats more frequently
         * when a new stat sink has been created.
         *
         * @param setName  non-null token printed to identify this set
         * @param ps  if non-null, and while not closed, is a stream to
         *     print statistics information to
         * @param printEvery  if true, every event is printed, else
         *     only summary stats are recorded
         * @param dumpS  if positive, dump cumulative stats at intervals
         *     of this many seconds; if zero periodic dumps are disabled
         */
        public StatsConfig(final String setName,
                           final PrintStream ps,
                           final boolean printEvery,
                           final int dumpS)
            { this(setName, ps, printEvery, dumpS, true); }

        /**Create a new stats sink description.
         * @param setName  non-null token printed to identify this set
         * @param ps  if non-null, and while not closed, is a stream to
         *     print statistics information to
         * @param printEvery  if true, every event is printed, else
         *     only summary stats are recorded
         * @param dumpS  if positive, dump cumulative stats at intervals
         *     of this many seconds; if zero periodic dumps are disabled
         * @param adaptive  if true, print initial summaries at shorter
         *     delays so that even short runs of the system yield
         *     some interesting stats
         */
        public StatsConfig(final String setName,
                           final PrintStream ps,
                           final boolean printEvery,
                           final int dumpS,
                           final boolean adaptive)
            {
            this(setName,
                 (ps == null) ? null : (new SimpleLoggerIF() { public void log(final String message) { ps.println(message); } }),
                 printEvery,
                 dumpS,
                 adaptive);
            }

        /**Create a new stats sink description.
         * @param setName  non-null token printed to identify this set
         * @param logger  if non-null, is a logger to
         *     log statistics information to
         * @param printEvery  if true, every event is printed, else
         *     only summary stats are recorded
         * @param dumpS  if positive, dump cumulative stats at intervals
         *     of this many seconds; if zero periodic dumps are disabled
         * @param adaptive  if true, print initial summaries at shorter
         *     delays so that even short runs of the system yield
         *     some interesting stats
         */
        public StatsConfig(final String setName,
                           final SimpleLoggerIF logger,
                           final boolean printEvery,
                           final int dumpS,
                           final boolean adaptive)
            {
            if((setName == null) || (setName.length() == 0) ||
                (dumpS < 0))
                { throw new IllegalArgumentException(); }
            name = setName;
            this.logger = logger;
            printEveryEvent = printEvery;
            dumpIntervalS = dumpS;
            adaptiveTime = adaptive;
            }
        }

    /**Capture a data point.
     * Notes, in the statistics set uniquely identified by statsSet,
     * an occurrence of the event named by ev.
     * <p>
     * Neither statsSet nor ev may be null.
     * <p>
     * Essentially, instances of ev are counted (in a long counter).
     * <p>
     * This tries to be moderately efficient and quick, at least on average,
     * but may trigger/drive processing such as printing of the event.
     */
    public static void captureDataPoint(final StatsConfig statsSet,
                                        final String ev)
        {
        if((statsSet == null) || (ev == null))
            { throw new IllegalArgumentException(); }

        // Atmomically get this stats set,
        // creating it if necessary.
        ConcurrentHashMap<String,AtomicLong> h;
        synchronized(dataStore)
            {
            // If there is no entry for statsSet then make one.
            while((h = dataStore.get(statsSet)) == null)
                { dataStore.put(statsSet, new ConcurrentHashMap<String, AtomicLong>()); }
            }

        // Get and increment the counter (etc), or create it if necessary.
        AtomicLong item;
        while((item = h.get(ev)) == null) { h.putIfAbsent(ev, new AtomicLong()); }
        item.incrementAndGet();

        // If we are to log every event
        // then we do so now.
        final long now = System.currentTimeMillis();
        final long delta = now - statsSet.creationTime;
        if(statsSet.printEveryEvent)
            { printStats(false, statsSet, ev, item, delta); }

        // Check if we are due for a summary dump,
        // and do one if needed.
        if(statsSet.timeForSummary())
            {
            // Take a copy of the keys to work on at our leisure.
            // Nothing is going to be deleted while we are working,
            // so this guarantees no upsets while processing.
            final List<String> v = new ArrayList<String>(h.keySet());

            // We'll sort the keys into order for printing.
            Collections.sort(v);

            // Work though all our entries...
            for(final Iterator<String> it = v.iterator(); it.hasNext(); )
                {
                final String key = it.next();
                printStats(true, statsSet, key, h.get(key), delta);
                }
            }
        }

    /**Print stats item to its output sink.
     * This might be zero or more streams
     * and zero or more other sinks such as files.
     * <p>
     * This includes a delta time in milliseconds from collection creation.
     */
    private static void printStats(final boolean summary,
                                   final StatsConfig config,
                                   final String event,
                                   final AtomicLong item,
                                   final long delta)
        {
        // If we're logging to a stream, do so now.
        if(config.logger != null)
            {
            final StringBuilder toPrint = new StringBuilder(81);
            toPrint.append(config.name).append(": ");
            toPrint.append(summary ? "SUMMARY" : "EVENT").append(": ");
            toPrint.append(delta).append("ms: ");
            toPrint.append("event ").append(event).append(": ");
            toPrint.append("count ").append(item.get());

            config.logger.log(toPrint.toString());
            }
        }
    }
