/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 24-May-02
 * Time: 18:24:08

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.uploader;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitAttrUtils;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.LocaleBeanBase;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;

import java.io.Serializable;
import java.util.*;

/**Base for a JavaBean to handle exhibit data for search or upload.
 * Unset items are empty strings rather than null for simplest handling
 * in JSPs.
 * <p>
 * This is thread-safe and Serializable so as to be able to be stored in
 * a servlet session; nothing especially long-lived or sensitive.
 */
public class ExhibitHandlerBeanBase implements Serializable
    {
    /**If true, limit selections to existing values in the database.
     * If false, a wider range of legal values is allowed.
     * <p>
     * If true, file types and authors can only be selected from
     * ones already represented in the production database.
     */
    protected final boolean onlyAllowExtant;

    /**Choose whether new values of some fields are allowed. */
    protected ExhibitHandlerBeanBase(final boolean _onlyAllowExtant)
        { onlyAllowExtant = _onlyAllowExtant; }

    /**Only allow selections from extant values. */
    protected ExhibitHandlerBeanBase()
        { this(true); }

    /**Used in a selector value  or setter to mean all values or no item. */
    public static final String SETTER_ALL = "-";


    /**For some unset values, set the most common value.
     * Can be overridden for values unique to derived classes.
     * Any overriding method should hold the instance lock to avoid races.
     * <p>
     * Should only be called in selected environments where an empty value
     * is not useful anyway (eg not valid as a wildcard).
     */
    public synchronized void setCommonValuesForUnsetFields()
        {
        // Nothing to be done if no AEP info...
        if(aep.aeid.length == 0) { return; }

        // Try filling in with the most-popular category.
        if("".equals(getCategory()))
            {
            int bestCount = 0;
            String bestCat = "";
            final Map<String,Integer> categoryExhibitCounts = aep.getCategoryExhibitCounts();
            for(final String cat : categoryExhibitCounts.keySet())
                {
                final int count = categoryExhibitCounts.get(cat);
                if(count > bestCount)
                    {
                    bestCount = count;
                    bestCat = cat;
                    }
                }
            setCategory(bestCat);
            }

        // Try filling in with the most-popular suffix/type.
        if("".equals(getSuffix()))
            {
            int bestCount = 0;
            String bestSuf = "";
            final Map<String,Integer> suffixExhibitCounts = aep.getDottedExtensionExhibitCounts();
            for(final String suf : suffixExhibitCounts.keySet())
                {
                final int count = suffixExhibitCounts.get(suf);
                if(count > bestCount)
                    {
                    bestCount = count;
                    bestSuf = suf;
                    }
                }
            setSuffix(bestSuf);
            }
        }


    /**The AllExhibitProperties data; never null.
     * Must be set before other property values are set.
     */
    private AllExhibitProperties aep = new AllExhibitProperties();

    /**Suffix selected (starting with ``.''), or "" if none (ie all types OK). */
    private String suffix = "";

    /**Author (initials) selected, or "" if none. */
    private String author = "";

    /**Category selected, or "" if none. */
    private String category = "";

    /**List of attribute words (String), or empty if none; immutable.
     * We preserve order and repetition unless the dedupAttrs() call is made.
     * <p>
     * This item may be replaced but is not mutated in situ
     * and is immutable
     * and so can be safely handed out to callers.
     */
    private List<String> attributes = Collections.emptyList();

    /**Set the AllExhibitProperties; ignored if null and must happen before other properties are set. */
    public synchronized void setAep(final AllExhibitProperties _aep)
        { if(_aep != null) { aep = _aep; } }

    /**Get the AllExhibitProperties; never null. */
    public synchronized AllExhibitProperties getAep() { return(aep); }

    /**Get all file suffixes user is allowed to select, sorted.
     * If onlyAllowExtant is true,
     * only file types in the database can be selected,
     * else all acceptable types can be selected.
     */
    public synchronized String[] getAllSuffixes()
        {
        if(!onlyAllowExtant)
            { return(AllExhibitProperties.getLegalSuffixes()); }

        final Set<String> s = getAep().getDottedExtensionExhibitCounts().keySet();
        final Iterator<String> it = s.iterator();
        final String result[] = new String[s.size()];
        for(int i = result.length; --i >= 0; )
        { result[i] = (String) it.next(); }
        Arrays.sort(result);
        return(result);
        }

    /**Get suffix selected (starting with ``.''), or "" if none (ie all types OK). */
    public synchronized String getSuffix() { return(suffix); }

    /**Set suffix selected (starting with ``.''), or "" or SETTER_ALL if none (ie all types OK).
     * If the suffix is invalid it is ignored.
     */
    public synchronized void setSuffix(final String suf)
        {
        // Accept request to clear the value.
        if("".equals(suf) || SETTER_ALL.equals(suf)) { suffix = ""; return; }

        if((suf == null) || !suf.startsWith(".") || (suf.length() < 2))
            { return; } // Clearly invalid; ignore.

        // Quickly check if this is a legal, recognised suffix...
        final ExhibitMIME.ExhibitTypeParameters type =
            ExhibitMIME.isValidInputExhibitNameExtension(suf.substring(1));
        if(type == null)
            { return; } // Clearly invalid; ignore.

        // Accept any suffix we know how to handle.
        suffix = suf;

//        // Get valid/legal suffixes,
//        // and ignore input unless it matches one of them.
//        final String valid[] = getAllSuffixes();
//        for(int i = valid.length; --i >= 0; )
//            {
//            if(valid[i].equals(suf))
//                { suffix = suf; return; } // Yes, got it!
//            }
//        // No, invalid; so ignored.
        }

    /**Generates body of suffix HTML select statement (dependent on old value, if any).
     * Assumed not to require localisation or internationalisation.
     */
    public synchronized String makeSuffixSelectBody()
        {
        final String legit[] = getAllSuffixes();
        final String fullList[] = new String[legit.length + 1];
        fullList[0] = SETTER_ALL; // Meaning ``all''.
        System.arraycopy(legit, 0, fullList, 1, legit.length);

        // Now make the labels...
        final String labels[] = new String[fullList.length];
        labels[0] = ".*"; // Meaning all.

        // Make a meaningful set of labels.
        final Map<String,Integer> dottedExtensionCount = getAep().getDottedExtensionExhibitCounts();
        for(int i = labels.length; --i > 0; )
            {
            labels[i] = fullList[i];
            final ExhibitMIME.ExhibitTypeParameters type =
                ExhibitMIME.isValidInputExhibitNameExtension(fullList[i].substring(1));
            if(type != null)
                {
                labels[i] = labels[i] +
                    " [" + type.mimeType + ": " + type.description + "]";
                }
            final Integer count = (Integer) (dottedExtensionCount.get(fullList[i]));
            if(count != null) { labels[i] = labels[i] + " (" + count + ")"; }
            }

        return(UploaderUtils.makeSelectBody(fullList, labels, getSuffix()));
        }

    /**Get all authors user is allowed to accept, sorted. */
    public synchronized String[] getAllAuthors()
        {
        final Set<String> s = getAep().getAuthorExhibitCounts().keySet();
        final Iterator<String> it = s.iterator();
        final String result[] = new String[s.size()];
        for(int i = result.length; --i >= 0; )
            { result[i] = it.next(); }
        Arrays.sort(result);
        return(result);
        }

    /**Returns author (initials) selected, or "" if none. */
    public synchronized String getAuthor() { return(author); }

    /**Set author (initials) selected, or "" or SETTER_ALL if none (ie all types OK).
     * If the author is invalid it is ignored.
     */
    public synchronized void setAuthor(final String auth)
        {
        if((auth == null) || (auth.length() == 0))
            { return; } // Clearly invalid; ignore.

        if("".equals(auth) || SETTER_ALL.equals(auth)) { author = ""; return; }

        // Get valid/legal authors,
        // and ignore input unless it matches one of them.
        final String valid[] = getAllAuthors();
        for(int i = valid.length; --i >= 0; )
            {
            if(valid[i].equals(auth))
                { author = auth; return; } // Yes, got it!
            }
        // No, invalid; so ignored.
        }

    /**Generates body of author HTML select statement (dependent on old value, if any).
     * Assumed not to require localisation or internationalisation.
     */
    public synchronized String makeAuthorSelectBody()
        {
        final String legit[] = getAllAuthors();
        final String fullList[] = new String[legit.length + 1];
        fullList[0] = SETTER_ALL; // Meaning ``all''.
        System.arraycopy(legit, 0, fullList, 1, legit.length);

        // Now make the labels...
        final String labels[] = new String[fullList.length];
        labels[0] = "*"; // Meaning all.

        // Make a meaningful set of labels.
        final Map<String,Integer> authorCount = getAep().getAuthorExhibitCounts();
        for(int i = labels.length; --i > 0; )
            {
            labels[i] = fullList[i];
            final Integer count = (authorCount.get(fullList[i]));
            if(count != null) { labels[i] = labels[i] + " (" + count + ")"; }
            }

        return(UploaderUtils.makeSelectBody(fullList, labels, getAuthor()));
        }

    /**Get all categories user is allowed to accept, sorted. */
    public synchronized String[] getAllCategories()
        {
        final Set<String> s = getAep().getCategoryExhibitCounts().keySet();
        final Iterator<String> it = s.iterator();
        final String result[] = new String[s.size()];
        for(int i = result.length; --i >= 0; )
            { result[i] = it.next(); }
        Arrays.sort(result, String.CASE_INSENSITIVE_ORDER);
        return(result);
        }

    /**Returns category selected, or "" if none. */
    public synchronized String getCategory() { return(category); }

    /**Set category selected, or "" or SETTER_ALL if none (ie all types OK).
     * If the category is invalid it is ignored.
     */
    public synchronized void setCategory(final String cat)
        {
        // Accept request to clear the value.
        if("".equals(cat) || SETTER_ALL.equals(cat)) { category = ""; return; }

        if((cat == null) || (cat.length() == 0))
            { return; } // Clearly invalid; ignore.

        // Ignore putative category name with invalid syntax.
        if(!ExhibitName.validNameInitialComponentSyntax(cat)) { return; }

        // Get valid/legal categories,
        // (extant categories or ones we have description/title for)
        // and ignore input unless it matches one of them.
        if(getAep().getCategoryExhibitCounts().keySet().contains(cat))
            { category = cat; return; } // Yes, got it!
        // OK, if not already extant, then check if we have description...
        final String key = UploaderConsts.SECTION_TITLE_PROPNAME_PREFIX + cat;
        final String i18nedTitle = (new LocaleBeanBase()).getLocalisedMessage(key);
        if((i18nedTitle != null) &&
           (i18nedTitle.length() != 0) &&
           !i18nedTitle.equals(key)) // Usual graceful failure mode of LocaleBean.
            { category = cat; return; } // Title exists...

        // No, invalid; so ignored.
        return;
        }

    /**Generates body of category HTML select statement (dependent on old value, if any).
     * This version does not use a LocaleBean.
     * <p>
     * Assumed not to require localisation or internationalisation.
     * <p>
     * We munge the categories for display by stripping any non-alphanumerics
     * (actually just ``-'' and ``_'')
     * and converting everything to lower-case.
     */
    public String makeCategorySelectBody()
        { return(makeCategorySelectBody()); }

    /**Generates body of category HTML select statement (dependent on old value, if any).
     * Assumed not to require localisation or internationalisation.
     * <p>
     * We munge the categories for display by stripping any non-alphanumerics
     * (actually just ``-'' and ``_'')
     * and converting everything to lower-case.
     *
     * @param localeBean  if non-null, used to lookup and localise
     *     attribute descriptions
     */
    public synchronized String makeCategorySelectBody(final LocaleBeanBase localeBean)
        {
        final String legit[] = getAllCategories();
        final String fullList[] = new String[legit.length + 1];
        fullList[0] = SETTER_ALL; // Meaning ``all''.
        System.arraycopy(legit, 0, fullList, 1, legit.length);

        // Now make the labels...
        final String labels[] = new String[fullList.length];
        labels[0] = "*"; // Meaning all.

        // Make a meaningful set of labels.
        final Map<String,Integer> categoryCount = getAep().getCategoryExhibitCounts();
        for(int i = labels.length; --i > 0; )
            {
            // Munge into more acceptable form.
            // Might look this up (convert it) in some more
            // i18n way later.
            final String normalisedDirname = fullList[i].toLowerCase().replace('-', ' ').replace('_', ' ').trim();
            labels[i] = normalisedDirname;

            // Insert a localised description if possible.
            if(localeBean != null)
                {
                // We'll see if there is a hand-crafted title for this section.
                final String i18nName = UploaderConsts.SECTION_TITLE_PROPNAME_PREFIX + fullList[i];
                final String msg = localeBean.getLocalisedMessage(i18nName);
                // We show the localised version if it exists and isn't just in a different case.
                if(!msg.equals(i18nName) &&
                    !msg.equalsIgnoreCase(normalisedDirname))
                    {
                    labels[i] = labels[i] + " [" + msg + "]";
                    }
                }


            final Integer count = (categoryCount.get(fullList[i]));
            if(count != null) { labels[i] = labels[i] + " (" + count + ")"; }
            }

        return(UploaderUtils.makeSelectBody(fullList, labels, getCategory()));
        }

    /**Generates body of attribute HTML select statement (dependent on old value, if any).
     * This version does not take a localeBean.
     * <p>
     * Assumed not to require localisation or internationalisation.
     * <p>
     * We assume that all attribute words are safe for use as values.
     * <p>
     * We explicitly pass in the attribute word we want selected,
     * or SETTER_ALL (or null) to select the default.
     *
     * @param inUse  if true, only attributes currently in use in the Gallery
     *     will be includes, else all legal attributes will be included
     * @param selectedAttr  if non-null and not SETTER_ALL and
     *     and attribute in the final select list,
     *     causes that item to be selected by default
     */
    public String makeAttributeSelectBody(final boolean inUse,
                                          final String selectedAttr)
        { return(makeAttributeSelectBody(inUse, selectedAttr, null)); }

    /**Generates body of attribute HTML select statement (dependent on old value, if any).
     * Assumed not to require localisation or internationalisation.
     * <p>
     * We assume that all attribute words are safe for use as values.
     * <p>
     * We explicitly pass in the attribute word we want selected,
     * or SETTER_ALL (or null) to select the default.
     *
     * @param inUse  if true, only attributes currently in use in the Gallery
     *     will be includes, else all legal attributes will be included
     * @param selectedAttr  if non-null and not SETTER_ALL and
     *     and attribute in the final select list,
     *     causes that item to be selected by default
     * @param localeBean  if non-null, used to lookup and localise
     *     attribute descriptions
     */
    public synchronized String makeAttributeSelectBody(final boolean inUse,
                                                       final String selectedAttr,
                                                       final LocaleBeanBase localeBean)
        {
        final Map<String,Set<String>> eBA = getAep().getExhibitsByAttribute();
        final List<String> attrWords = new ArrayList<String>();
        // Iterate through the attribute words in sorted order.
        for(final Iterator it = (new TreeSet<String>(eBA.keySet())).iterator(); it.hasNext(); )
            {
            final String attrWord = (String) it.next();
            final Set<String> names = eBA.get(attrWord);

            // Skip attributes that are not in use if we care about that
            if(inUse && (names.size() == 0)) { continue; }

            attrWords.add(attrWord);
            }
        final String legit[] = new String[attrWords.size()];
        attrWords.toArray(legit);

        final String fullList[] = new String[legit.length + 1];
        fullList[0] = SETTER_ALL; // Meaning ``all''.
        System.arraycopy(legit, 0, fullList, 1, legit.length);

        // Now make the labels...
        final String labels[] = new String[fullList.length];
        labels[0] = "*"; // Meaning all.

        // Make a meaningful set of labels.
        final Map<String,Set<String>> attrCount = getAep().getExhibitsByAttribute();
        for(int i = labels.length; --i > 0; )
            {
            // Munge into more acceptable form.
            // Might look this up (convert it) in some more
            // i18n way later.
            labels[i] = fullList[i];

            // Insert a localised description if possible.
            if(localeBean != null)
                {
                // We'll see if there is descriptive text for this attribute.
                final String i18nName = UploaderConsts.ATTR_I18N_DESC_PREFIX + fullList[i];
                final String msg = localeBean.getLocalisedMessage(i18nName);
                if(!msg.equals(i18nName))
                    {
                    labels[i] = labels[i] + " [" + msg + "]";
                    }
                }

            final int count = ((attrCount.get(fullList[i]))).size();
            labels[i] = labels[i] + " (" + count + ")";
            }

        return(UploaderUtils.makeSelectBody(fullList, labels, selectedAttr));
        }

    /**Get all acceptable attribute words that are in use as a single String for display to the user.
     * This returns the list of attribute words in use in the Gallery exhibits if inUse is true,
     * else it returns the list of attribute words not currently in use though legal.
     * <p>
     * The result is a space-separated sorted list.
     * <p>
     * This returns the empty string if no list is available.
     */
    public synchronized String getInUseAttrWordListAsString(final boolean inUse)
        {
        final Map<String,Set<String>> eBA = getAep().getExhibitsByAttribute();
        final StringBuilder result = new StringBuilder();
        // Iterate through the attribute words in sorted order.
        for(final Iterator it = (new TreeSet<String>(eBA.keySet())).iterator(); it.hasNext(); )
            {
            final String attrWord = (String) it.next();
            final Set<String> names = eBA.get(attrWord);

            // Skip attributes that don't match our filter.
            if(inUse == (names.size() == 0)) { continue; }

            if(result.length() != 0) { result.append(' '); }
            result.append(attrWord);
            }
        return(result.toString());
        }

    /**Get list of all acceptable attribute words as single String for display to user.
     * This displays the legal attribuite word set in sorted order,
     * space-separated, for display to a user (eg in an HTML page).
     * <p>
     * This returns the empty string if no list is available.
     */
    public synchronized String getLegalAttrWordListAsString()
        {
        final Set<String> words = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();
        final StringBuilder result = new StringBuilder(words.size() * 8);
        for(final Iterator<String> it = words.iterator(); it.hasNext(); )
            {
            if(result.length() != 0) { result.append(' '); }
            result.append(it.next());
            }
        return(result.toString());
        }

    /**Eliminates duplicate attribute words and sorts into lexical order.
     * Must only be called once the attributes have been set.
     * <p>
     * Assumes all entries are valid ie are non-null, legal attribute words.
     * <p>
     * Idempotent.
     */
    public synchronized void dedupAttrs()
       {
       attributes = Collections.unmodifiableList(new ArrayList<String>(new TreeSet<String>(attributes)));
       }

    /**Get attribute words as immutable List; can be zero-length but never null.
     * All words are valid attribute words (at least when they were set)
     * and the trailing one will not be a number.
     */
    public synchronized List<String> getAttributeWordsAsList()
        {
        return(attributes); // Must be immutable.
        }

    /**Get attribute words as a valid string of hyphen-separated words; can be "" but never null.
     * All words are valid attribute words (at least when they were set)
     * and the trailing one will not be a number.
     */
    public synchronized String getAttributeWords()
        {
        final StringBuilder sb = new StringBuilder();
        for(final Iterator<String> it = attributes.iterator(); it.hasNext(); )
            {
            final String s = it.next();
            if(sb.length() != 0) { sb.append(ExhibitName.WORD_SEP); }
            sb.append(s);
            }
        return(sb.toString());
        }

    /**Set attribute words; any invalid items will be silently discarded.
     * Any existing attribute words held by this bean are removed first.
     * <p>
     * The argument may not be null nor contain nulls.
     */
    public synchronized void setAttributeWords(final String words[])
        {
        if(words == null) { throw new IllegalArgumentException(); }
        setAttributeWords(Arrays.asList(words));
        }

    /**Set attribute words; any invalid items will be silently discarded.
     * Any existing attribute words held by this bean are removed first.
     * <p>
     * The argument may not be null nor contain nulls.
     */
    public synchronized void setAttributeWords(final List<String>words)
        {
        if(words == null) { throw new IllegalArgumentException(); }

        // Deal with common case of no attributes.
        if(words.size() == 0)
            {
            attributes = Collections.emptyList();
            return;
            }

        final ArrayList<String> newAttrs = new ArrayList<String>(words.size());

        // Get set of valid attribute words.
        final Set<String> validAttrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

        for(final String s : words)
            {
            if(s == null) { continue; } // Definitely invalid!
            if(validAttrWords.contains(s)) { newAttrs.add(s); }
            }
        newAttrs.trimToSize(); // Conserve memory...

        attributes = Collections.unmodifiableList(newAttrs);
        }

    /**Set attribute words; valid string of hyphen-separated words; can be "" or null.
     * Use null or "" or SETTER_ALL to clear this field.
     * <p>
     * All words must be attribute words and the trailing one may not be
     * a number.
     * <p>
     * Any items that are not valid attribute words will just be silently
     * deleted.
     * <p>
     * Any non-valid word characters (eg including spaces) will be converted
     * to hyphens.
     * <p>
     * This tries very hard to keep as much of its input as possible.
     */
    public void setAttributeWords(final String attributeWords)
        {
        if((attributeWords == null) || (attributeWords.length() == 0))
            {
            attributes = Collections.emptyList();
            return;
            }

        // Work through the input treating space and dash as separators.
        final StringTokenizer st = new StringTokenizer(attributeWords,
                                                ExhibitName.WORD_SEPS + ' ');
        final String[] sa = new String[st.countTokens()];
        for(int i = 0; i < sa.length; ++i)
            { sa[i] = st.nextToken(); }
        setAttributeWords(sa);
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = 8025048065146981171L;
    }
