/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 21-May-02
 * Time: 20:49:34

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.uploader;

import org.hd.d.pg2k.svrCore.AllExhibitImmutableData;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitAttrUtils;
import org.hd.d.pg2k.svrCore.ExhibitName;

import java.util.Set;
import java.util.SortedSet;
import java.util.StringTokenizer;

/**Bean that collects information about an exhibit to upload.
 * Backs upload.jsp and, unusally, is stored in the session
 * as the author provides data about the exhibit.
 * <p>
 * This validates information its is presented with as far as
 * possible, syntactically and semantically.
 * For example, it tries to ensure that a correct and unique
 * exhibit name is created, and warns if new (possibly-misspelt)
 * words have been introduced.
 * <p>
 * This holds the filename and optional description, location,
 * etc data, and output can be dumped as a new-style properties
 * file.
 * <p>
 * This is thread-safe, all methods being synchronized where necessary.
 * <p>
 * In general, our upload process should observe the following:
    <ul>
    <li>Enforce maximum length for name.
    <li>Create full exhibit name and check unique.
    <li>Unique is:
        <ol>
        <li>amount production and upload set, and
        <li>only the final name component, and
        <li>ignoring case if possible
        </ol>
    <li>Don't allow upload unless name unique.
    <li>Don't allow upload unless within GenProps-limited:
        <ul>
        <li>uploads per user
        <li>total space per user
        <li>total space across all users
        <li>maximum individual upload size
        </ul>
    <li>Complete upload to transient area.
    <li>Complete upload passwords.
    <li>Put in link from upload.jsp to naming rules doc.
    <li>Collect location information.
    <li>Save exhibit and properties and accession files.
    <li>Add all-OK method to uib and check in upload.jsp and upload2.jsp.
    <li>Add system-wide attribute words,
        and check that all given attribute words are attribute words.
    <li>Add warning if user has specified completely new word.
    <li>Add listing of similar extant exhibit names a la search engine.
    <li>When checking for name clashes in upload,
        if possible warn of names which differ only in number, author, suffix
        or main words.
    <li>Save description data.
    <li>Check that exhibit is unique using (length then) hash.
    </ul>
 */
public final class UploadInfoBean extends ExhibitHandlerBeanBase
    {
    /**Create a new instance.
     * We allow some values to be selected beyond extant values in the
     * production database.
     */
    public UploadInfoBean()
        { super(false); }

    public synchronized void setUploadAeid(AllExhibitImmutableData uploadAeid)
        {
        this.uploadAeid = uploadAeid;
        }

    /**Aeid for upload area; should be set. */
    private AllExhibitImmutableData uploadAeid;

    /**Gets number in series of exhibit; never negative. */
    public synchronized int getNumber()
        {
        return(number);
        }

    /**Sets number in series for exhibit; if null or "" or unparsable or negative the value is ignored.
     * This is a convenience method to make setting from a JSP/form more robust.
     */
    public synchronized void setNumber(String number)
        {
        if((number == null) || (number.length() < 1)) { return; }
        try { setNumber(Integer.parseInt(number, 10)); }
        catch(NumberFormatException e) { } // Invalid; ignore.
        }

    /**Sets number in series for exhibit; if negative the value is ignored. */
    public synchronized void setNumber(final int number)
        {
        if(number < 0) { return; } // Invalid; ignore.
        this.number = number;
        }

    /**Increment number-in-series value iff already positive and returns new value; never negative. */
    public synchronized int incNumberIfNonZero()
        {
        if(number != 0)
            { if(++number < 0) { number = 0; /* Deal with overflow! */ } }
        return(number);
        }

    /**The number in the current series; never negative.
     * If zero, it will not be inserted into the filename
     * (ie is the default).
     */
    private int number;



    /**Forces string to a superficially valid word string.
     * Replaces all non-word chars with the word separator,
     * replaces all sequences of two or more separators with one,
     * and trims all leading and trailing separators.
     * <p>
     * The input must not be null.
     * <p>
     * This does not of itself make the string valid to use as
     * a main or attribute word string (it does not check the
     * contents of the words).
     * <p>
     * This limits the length of the value.
     * <p>
     * This returns the original string if valid.
     */
    private static String makeValidWordString(String in)
        {
        if(in == null) { in = ""; }
        else if(in.length() > ExhibitName.MAX_STEM_LENGTH)
            { in = in.substring(0, ExhibitName.MAX_STEM_LENGTH); }

        // Every time we encounter a non-word character,
        // replace all occurrences of it with the word separator.
        // We don't expect many invalid characters in any one go.
        for(int i = in.length(); --i >= 0; )
            {
            final char c = in.charAt(i);
            if(!ExhibitName.validWordCharacter(c) &&
               (c != ExhibitName.WORD_SEP))
                { in = in.replace(c, ExhibitName.WORD_SEP); }
            }

        // If there any repeated multiple seperators
        // then use a regex to replace them all with single separators.
        final String multipleSep = ExhibitName.WORD_SEPS + ExhibitName.WORD_SEP;
        if(in.lastIndexOf(multipleSep) != -1)
            { in = in.replaceAll(multipleSep + ExhibitName.WORD_SEP + '*', ExhibitName.WORD_SEPS); }

        // Trim any (by now single) leading/trailing separator.
        if((in.length() > 0) && (in.charAt(0) == ExhibitName.WORD_SEP))
            { in = in.substring(1); }
        if((in.length() > 0) && (in.charAt(in.length()-1) == ExhibitName.WORD_SEP))
            { in = in.substring(0, in.length()-1); }

        return(in);
        }

    /**Main body words; valid string of hyphen-separated words where; can be "" but never null.
     * Trailing words may not be attribute words and the final one may not be
     * a number.
     */
    private String mainWords = "";

    /**Set main body words; valid string of hyphen-separated words where; can be "" but never null.
     * Trailing words may not be attribute words and the final one may not be
     * a number.
     */
    public synchronized String getMainWords()
        {
        return mainWords;
        }

    /**Set main body words; valid string of hyphen-separated words where; can be "" but never null.
     * Use null or "" or "-" to clear this field.
     * <p>
     * Trailing words may not be attribute words and the final one may not be
     * a number.
     * <p>
     * Any non-valid word characters (eg including spaces) will be converted
     * to hyphens.
     * <p>
     * This tries very hard to keep as much of its input as possible.
     */
    public synchronized void setMainWords(String mainWords)
        {
        // Deal with inputs used to clear the value.
        if((mainWords == null) ||
            (mainWords.length() == 0) ||
            (mainWords.equals("-")))
            {
            this.mainWords = "";
            return;
            }

        // Crudely beat into shape.
        mainWords = makeValidWordString(mainWords);

        this.mainWords = mainWords;
        }

    /**Get main words that would be new to the Gallery: return the empty string "" if none.
     * This checks all the putative exhibit name main words against all main words of
     * all extant exhibits (in the main database), returning a space-separated list
     * of those new words that might have incorrect spelling or capitalisation.
     * This can be used to warn the User that they should possibly check their spelling, etc.
     */
    public synchronized String getNewMainWords()
        {
        final Set allMainWords = getAep().getMainWords();
        final StringBuilder sb = new StringBuilder();
        final StringTokenizer st = new StringTokenizer(getMainWords(), ExhibitName.WORD_SEPS);
        while(st.hasMoreTokens())
            {
            final String w = st.nextToken();
            if(!allMainWords.contains(w))
                {
                // New word...
                if(sb.length() > 0) { sb.append(' '); }
                sb.append(w);
                }
            }
        return(sb.toString());
        }


    /**Get description text or "" if none; never null.
     * Pure ASCII text, possibly with HTML mark-up.
     */
    public synchronized String getDescription()
        {
        return description;
        }

    /**Set description text or "" or "-" if none; never null.
     * Pure ASCII text, possibly with HTML mark-up.
     * <p>
     * This tries hard to preserve any input given to it,
     * replacing illegal characters with spaces, and truncating
     * to maximum length.
     * <p>
     * If the input string is OK it should be stored intact.
     */
    public synchronized void setDescription(String description)
        {
        // Deal with inputs used to explicitly clear the value (such as a single dash).
        // Deal also with simple misinterpretations of the instructions
        // which would not be that meaningful to keep such as one-character entries.
        if((description == null) ||
            (description.length() < 1) ||
            /* description.equals("-") || */ description.equals("\"-\""))
            {
            this.description = "";
            return;
            }

        // Convert all characters that are not printable ASCII to spaces.
        for(int i = description.length(); --i >= 0; )
            {
            final char c = description.charAt(i);
            if((c < 32) || (c >= 127))
                { description = description.replace(c, ' '); }
            }

        // Trim any whitespace from either end.
        description = description.trim();

        // We replace all runs of spaces with single spaces to save memory.
        // This is not the most efficient algorithm, but should be OK...
        int lastDupSp;
        while((lastDupSp = description.lastIndexOf("  ")) != -1)
            { description = new StringBuilder(description).deleteCharAt(lastDupSp).toString(); }

        // Truncate to length if need be.
        if(description.length() > CoreConsts.DESCRIPTION_MAX_CHARS)
            { description = description.substring(0, CoreConsts.DESCRIPTION_MAX_CHARS); }

        // Save value.
        this.description = description;
        }

    /**Description text or "" if none; never null.
     * Pure ASCII text, possibly with HTML mark-up.
     */
    private String description = "";

    /**When enough exhibit information has been collected, this returns true.
     * This returns true when there is enough information to unambiguously
     * store the exhibit under a complete and unique name.
     * <p>
     * Optional information need not be present for this to return true.
     */
    public synchronized boolean enoughInfo()
        {
        return(
            (getCategory().length() != 0) &&
            (getMainWords().length() != 0) &&
            (getAuthor().length() != 0) &&
            (getSuffix().length() != 0)
            );
        }

    /**Tries to reparse as the system will see the name.
     * Call after all available bean attributes from the user have been set.
     * <p>
     * If enoughInfo() is true, this will assemble the complete name and
     * the reset the component parts by reparsing it with ExhibitName.
     * <p>
     * This should make clear any ambiguities in, for example, putting
     * attribute words at the end of the main words section.
     * <p>
     * We only reset the parts about which there may be ambiguity,
     * ie main words, attribute words, and number-in-series.
     * <p>
     * If enoughInfo() is not true, this does nothing.
     */
    public synchronized void tryToSanitise()
        {
        if(!enoughInfo()) { return; }
        final String putativeName = getFullName();
        if(!ExhibitName.validNameSyntax(putativeName)) { return; }

        // Reconstruct the name from its parts.
        setNumber(ExhibitName.getNumberInSeriesComponent(putativeName));
        final int ptrs[] = ExhibitName.getMainAndAttrWordComponentBoundaries(
            putativeName, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
        setMainWords(putativeName.substring(ptrs[0]+1, ptrs[1]));
        setAttributeWords((ptrs[1] == ptrs[2]) ? "" :
                          putativeName.substring(ptrs[1]+1, ptrs[2]));
        }

    /**Assemble complete name.
     * The result is undefined if enoughInfo() is false,
     * but this will do the best it can/
     * <p>
     * This may not be valid because it may be too long.
     * <p>
     * The name starts with the category directory and the final name.
     */
    public synchronized String getName(final boolean fullName)
        {
        final StringBuilder sb = new StringBuilder(64);
        if(fullName) { sb.append(getCategory()).append('/'); }
        sb.append(getMainWords());
        if(getAttributeWords().length() > 0)
            { sb.append(ExhibitName.WORD_SEP).append(getAttributeWords()); }
        if(getNumber() > 0)
            { sb.append(ExhibitName.WORD_SEP).append(getNumber()); }
        sb.append(ExhibitName.WORD_SEP).append(getAuthor());
        sb.append(getSuffix());
        return(sb.toString());
        }

    /**Assemble complete name.
     * The result is undefined if enoughInfo() is false,
     * but this will do the best it can/
     * <p>
     * This may not be valid because it may be too long.
     * <p>
     * The name starts with the category directory and the final name.
     */
    public synchronized String getFullName()
        {
        return(getName(true));
        }

    /**Set all parameters from a supplied full name.
     * The result is undefined if the name is not valid,
     * but this will do the best that it can.
     */
    public synchronized void setFullName(final String name)
        {
        if(name == null) { return; } // Ignore.
        setCategory(ExhibitName.getCategoryComponent(name));
        final SortedSet<String> legalAttrs = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();
        setMainWords(ExhibitName.getMainWordsComponent(name, legalAttrs));
        setAttributeWords(ExhibitName.getAttributeWordsComponent(name, legalAttrs));
        setNumber(ExhibitName.getNumberInSeriesComponent(name));
        setAuthor(ExhibitName.getAuthorComponent(name));
        setSuffix("." + ExhibitName.getExtensionComponent(name));
        }


    /**Returns true if we can construct a valid name.
     * We explicitly make sure that we have enough information and
     * that when we construct a name it is acceptable.
     * <p>
     * We should make sure that we can insert a ``_moreXXXX/_moreYY/''
     * section and not get an overlong-name.
     */
    public synchronized boolean canMakeValidName()
        {
        return(enoughInfo() &&
               ExhibitName.validNameSyntax(getFullName()));
        }

    /**Check if the name is unique against the production data set.
     * The result is undefined if enoughInfo() is false.
     * <p>
     * We just check the short name.
     */
    public synchronized boolean nameUniqueInMainDatabase()
        {
        final String shortName = getName(false);
        return(getAep().aeid.getFullName(shortName) == null);
        }

    /**Check if the name is unique against the upload data set.
     * This returns false if the aux database data has not been set.
     * <p>
     * We just check the short name.
     */
    public synchronized boolean nameUniqueInAuxDatabase()
        {
        if(uploadAeid == null) { return(false); }
        final String shortName = getName(false);
        return(uploadAeid.getFullName(shortName) == null);
        }

    /**Returns true when we have enough, valid, unique info to save an exhibit.
     * The name must be unique in our production and temporary databases.
     * <p>
     * This is the "master" test of acceptability.
     */
    public synchronized boolean enoughValidUniqueInfo()
        {
        return(canMakeValidName() &&
               nameUniqueInMainDatabase() &&
               nameUniqueInAuxDatabase());
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -6989233428145950946L;
    }
