/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 21-May-02
 * Time: 21:35:07

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.svrCore.uploader;

import org.hd.d.pg2k.svrCore.*;

import java.io.*;
import java.util.concurrent.atomic.AtomicLong;


/**Utilities for exhibit uploader.
 */
public final class UploaderUtils
    {
    /**Prevent creation of instances. */
    private UploaderUtils() { }


    /**Size of (large) buffers used for I/O efficiency; strictly positive. */
    public static final int BATCH_UPLOAD_BUFSIZE = 65536;

    /**Make body of select HTML select statement, marking any extant value as selected.
     * The labels[] array can be null, and if so then the values will be used
     * as the text; if not null it must be the same length as values[] and must
     * not contain nulls.
     *
     * @param values  the values returned by the select item
     * @param labels  (optional) the text values shown for the select items
     * @param selected  if equal to one of the values[] entries then that
     *     item is marked as selected
     */
    public static String makeSelectBody(String values[], String labels[],
                                        String selected)
        {
        final StringBuilder sb = new StringBuilder(32 * values.length);
        for(int i = 0; i < values.length; ++i)
            {
            // Do we have different text for value and label?
            final boolean valueNELabel = (labels != null) &&
                !values[i].equals(labels[i]);

            sb.append("<option");
            if(valueNELabel)
                {
                sb.append(" value=");
                sb.append(quoteHTMLArg(values[i]));
                }
            if(values[i].equals(selected)) { sb.append(" selected"); }
            sb.append('>');
            if(valueNELabel) { sb.append(labels[i]); }
            else { sb.append(values[i]); }
            sb.append("</option>");
            }
        return(sb.toString());
        }

    /**Puts double quotes round any tag attribute value that needs them.
     * Any tag that does not just contain letters and digits will
     * get "..." wrapped round it.  If an embedded " is found, or
     * the tag is found to be too long, an exception is thrown.
     * We include '_' in this notion of letters.
     * <p>
     * If this does not need to change the input it is returned as is.
     * <p>
     * TODO: jUnit tests
     *
     * @param s     non-null, unquoted HTML tag attribute value.
     * @exception IllegalArgumentException    impossible to form tag
     */
    public static final String quoteHTMLArg(final String s)
        throws IllegalArgumentException
        {
        final int l = s.length();
        if(l > 1024)
            { throw new IllegalArgumentException("tag attribute too long"); }
        if(s.indexOf('"') != -1)
            { throw new IllegalArgumentException("tag attribute contains `\"'"); }

        boolean needToQuote = false;

        // If the string is of zero length, it will need quoting to
        // avoid ambiguity.
        if(l == 0) { needToQuote = true; }

        // Else search the String for characters that would force us to use quotes...
        else
            {
            // Get the data out of the String as chars to look for
            // anything that will need to be quoted.
            for(int i = l; --i >= 0; )
                {
                final char c = s.charAt(i);
                // Use the digit() routine as a hack to check if
                // character is [0-9A-Za-z].
                if((Character.digit(c, 36) == -1) && (c != '_'))
                    { needToQuote = true; break; }
                }
            }

        if(needToQuote)
            {
            StringBuilder sb = new StringBuilder(l + 2);
            sb.append('"');
            sb.append(s);
            sb.append('"');
            return(sb.toString());
            }

        return(s);
        }


    /**Terminator (7-bit ASCII) character byte for all messages. */
    public static final char BATCH_UPLOAD_MSG_TERMINATOR = '\n';

    /**Send response to batch upload client.
     * <p>
     * This routine does not close the output stream when done.
     * <p>
     * This routine does not flush the output stream when done;
     * the caller should possibly do so itself.
     *
     * @param logger
     * @param os  stream to write to; never null
     * @param isOK  if true then this is an "OK" response,
     *     else a "FAIL" response
     * @param exhibitName  the syntatically-correct name of the exhibit upload
     *     to which this response applies, or null for before the first upload
     * @param detail  the optional extra information on the failure;
     *     null/"" if to be omitted
     *
     * @throws IOException  in case of error writing to the stream
     */
    public static void sendResponseToBatchUploadClient(final SimpleLoggerIF logger,
                                                       final OutputStream os,
                                                       final boolean isOK,
                                                       final String exhibitName,
                                                       final String detail)
        throws IOException
        {
        if(os == null)
            { throw new IllegalArgumentException(); }
        if((exhibitName != null) && (!ExhibitName.validNameSyntax(exhibitName)))
            { throw new IllegalArgumentException(); }

        // Buffer for efficiency...
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final DataOutputStream dos = new DataOutputStream(baos);

        if(!isOK && (logger != null)) { logger.log(" [Batch client upload failed: "+detail+"]"); }

        // "O" for OK, "F" for fail.
        dos.write(isOK ? 'O' : 'F');

        // Write the exhibit name.
        dos.writeUTF((exhibitName == null) ? "" : exhibitName);

        // Write the optional detail.
        dos.writeUTF((detail == null) ? "" : detail);

        // Write the termination byte.
        dos.write(BATCH_UPLOAD_MSG_TERMINATOR);

        // Done; write to output stream as a block for efficiency.
        dos.flush();
        os.write(baos.toByteArray());
        }

    /**Decode response to batch upload client from server.
     * On a fail response the detail message
     * will be embedded in the detail of the IOException.
     * <p>
     * If this detects malformed input
     * then it will throw an IOException.
     *
     * @return  exhibitName or null on success/OK response
     * @throws IOException  on problem reading the input stream
     *     or on a fail response
     */
    public static String decodeResponseToBatchUploadClient(final InputStream is)
        throws IOException
        {
        if(is == null) { throw new IllegalArgumentException(); }

        // Read all the fields.
        // Validate the first field to reject random junk quickly.
        final DataInputStream dis = new DataInputStream(is);
        final int t = dis.readByte();
        if((t != 'O') && (t != 'F')) { throw new IOException("invalid response type code: " + t); }
        final String exhibitName = dis.readUTF();
        final String detail = dis.readUTF();
        final int term = dis.readByte();

        // Check terminator.
        if(term != BATCH_UPLOAD_MSG_TERMINATOR)
            {
            // Corrupt message (invalid terminator).
            throw new IOException("invalid message terminator: " + term);
            }

        // Check name.
        if((exhibitName.length() != 0) && !ExhibitName.validNameSyntax(exhibitName))
            {
            // Corrupt message (invalid name).
            throw new IOException("invalid exhibit name");
            }

        // Deal with the usual "OK" case...
        if(t == 'O')
            {
            // "OK" message...
            // Ignore the detail message.
            return((exhibitName.length() == 0) ? null : exhibitName);
            }

        assert(t == 'F');

        // Deal with the "FAIL" case.
        throw new IOException((detail.length() == 0) ? null : detail);
        }

    /**Compute length of upload request for given exhibit.
     * Can be used to set a content-length header (etc)
     * before calling sendRequestFromBatchUploadClient
     * to deliver the actual data.
     */
    public static int computeLengthOfRequestFromBatchUploadClient(
                                                    final ExhibitStaticAttr esa,
                                                    final ROByteArray hashMD5,
                                                    final String description)
        {
        if((esa == null) || (hashMD5 == null))
            { throw new IllegalArgumentException(); }

        // The easiest way to do this is
        // to write everything bar the data itself to a buffer
        // and count the bytes written.
        // Not marvelously efficient, but probably OK.
        //
        // An optimisation is to avoid writing some of the fixed-length parts
        // and allow for them in the returned length instead.
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final DataOutputStream dos = new DataOutputStream(baos);

        try
            {
            // Write the exhibit length.
//        dos.writeLong(esa.length); // Allowed for in result.

            // Write the exhibit name.
            dos.writeUTF(esa.filePath);

            // Write the hash (as length then raw bytes).
//        dos.writeInt(hashMD5.length()); // Allowed for in result.
//        dos.write(hashMD5.getData()); // Allowed for in result.
            // Write out the description (convert null to "").
            dos.writeUTF((description == null) ? "" : description);

            // Skip writing the data...

            // Write the termination byte.
//        dos.write(BATCH_UPLOAD_MSG_TERMINATOR); // Allowed for in result.

            // Force everything into the buffer...
            dos.flush();

            final long len =
                   8 + // Exhibit length.
                   4 + // Hash length.
                   hashMD5.length() + // Hash data.
                   baos.size() + // Variable-length parts...
                   esa.length + // Data length.
                   1; // Terminator byte.

            if(len > Integer.MAX_VALUE)
                { throw new IllegalArgumentException("exhibit too large"); }

            return((int) len);
            }
        catch(final IOException e)
            {
            throw new Error(e); // Should not happen.
            }
        }

    /**Send upload request and data from batch client to server.
     * We send up the wire, in order:
     * <ol>
     * <li>The putative exhibit length.
     * <li>The putative (full, syntactically-valid, unique) exhibit name.
     * <li>The putative exhibit MD5 hash.
     * <li>The putative exhibit description (or "" if none).
     * <li>The raw exhibit data.
     * </ol>
     * <p>
     * The exhibit hash serves two purposes:
     * <ul>
     * <li>To ensure that the exhibit is unique before starting to upload.
     * <li>To ensure that the exhibit arrives uncorrupted.
     * </ul>
     * <p>
     * This routine does not close the output stream when done.
     * <p>
     * This routine does not flush the output stream when done;
     * the caller should possibly do so itself.
     * The routine may flush the output from time to time,
     * eg after writing a large block.
     * <p>
     * If the caller supplies a null esa/hash/datastream
     * then we send a request up the wire for a zero-length exhibit upload,
     * which is treated as a request to drop the connection
     * (the client should drop it after sending this).
     * <p>
     * This can update a progress monitor value while it is working.
     *
     * @param progress  if non-null, is set to the amount of bytes uploaded
     *     so far,
     *     and -1 before any bytes have been uploaded
     *     (and Long.MAX_VALUE after we have finished while buffered data xfers)
     *
     * @throws IOException  in case of error sending data to the output stream;
     *     if this occurs then the output stream should be closed and discarded
     */
    public static void sendRequestFromBatchUploadClient(final OutputStream os,
                                                        final ExhibitStaticAttr esa,
                                                        final ROByteArray hashMD5,
                                                        final String description,
                                                        final InputStream exhibitData,
                                                        final AtomicLong progress)
        throws IOException
        {
        if(os == null)
            { throw new IllegalArgumentException(); }

        final DataOutputStream dos = new DataOutputStream(os);

        if((esa == null) && (hashMD5 == null) && (exhibitData == null))
            {
            dos.writeLong(0);
            dos.write(BATCH_UPLOAD_MSG_TERMINATOR);
            return;
            }

        if((esa == null) || (hashMD5 == null) || (exhibitData == null))
            { throw new IllegalArgumentException(); }

        if((description != null) &&
           (description.length() > CoreConsts.DESCRIPTION_MAX_CHARS))
            { throw new IllegalArgumentException("description too long"); }

        // If present, update the progress monitor.
        if(progress != null)
            {
            progress.set(-1); // Upload proper not started yet...
            }

        // Write the exhibit length.
        dos.writeLong(esa.length);

        // Write the exhibit name.
        dos.writeUTF(esa.filePath);

        // Write the hash (as length then raw bytes).
        dos.writeInt(hashMD5.length());
        dos.write(hashMD5.getData());

        // Write out the description (convert null to "").
        dos.writeUTF((description == null) ? "" : description);

        // If present then set progress to show upload just starting.
        if(progress != null)
            {
            progress.set(0); // Upload proper is just starting...
            }

        // Copy the raw exhibit data one (large, efficient) block at a time.
        final byte buf[] = new byte[(int) Math.min(esa.length, BATCH_UPLOAD_BUFSIZE)];
        for(long bytesLeft = esa.length; bytesLeft > 0; )
            {
            final int maxXfer = (int) Math.min(buf.length, bytesLeft);
            final int read = exhibitData.read(buf, 0, maxXfer);
            if(read < 1) { throw new IOException("error reading data for exhibit: " + esa); }
            dos.write(buf, 0, read);
            bytesLeft -= read;

            // If present, update the progress monitor with bytes sent so far.
            if(progress != null)
                {
                dos.flush(); /* Give truer view of progress... */
                progress.set(esa.length - bytesLeft); // Show bytes transferred.
                }
            }

        // If present then update the progress monitor.
        if(progress != null)
            {
            progress.set(Long.MAX_VALUE); // Buffered data may still need to be transferred, etc.
            }

        // Write the termination byte.
        dos.write(BATCH_UPLOAD_MSG_TERMINATOR);
        }

    /**Encapsulated immutable upload request from batch-upload client. */
    public static final class BatchUploadClientRequest
        {
        /**Exhibit full name; always valid, never null. */
        public final String exhibitName;
        /**Exhibit length; strictly positive. */
        public final long exhibitLength;
        /**Exhibit content MD5 hash; never null. */
        public final ROByteArray hashMD5;
        /**Exhibit description, never null but may be "". */
        public final String description;

        public BatchUploadClientRequest(final String name,
                                        final long length,
                                        final ROByteArray hash,
                                        final String desc)
            {
            if(!ExhibitName.validNameSyntax(name) ||
               (length <= 0) ||
               (hash == null) || (hash.length() != 16) ||
               (desc == null) || (desc.length() > CoreConsts.DESCRIPTION_MAX_CHARS))
                { throw new IllegalArgumentException(); }
            exhibitName = name;
            exhibitLength = length;
            hashMD5 = hash;
            description = desc;
            }
        }

    /**Decode an upload request from a batch client and return details.
     * In case of success returns the name/hash/length data,
     * with the input stream positioned directly before the exhibit data
     * (which in turn is followed by the message terminator).
     * <p>
     * (The timestamp in the returned ExhibitStaticAttr currently contains
     * no useful data.)
     * <p>
     * In case of failure this throws an IOException
     * and the input stream should be discarded as being in an unknown state.
     * <p>
     * A zero length is treated as a request from the client to drop the
     * connection and we return null in this case.
     *
     * @throws IOException  in case of error in input
     *
     * @return in case of successfuul decoding, a tuple of the exhibit data
     *     and the MD5 exhibit content hash
     */
    public static BatchUploadClientRequest decodeRequestFromBatchUploadClient(final InputStream is)
        throws IOException
        {
        if(is == null) { throw new IllegalArgumentException(); }

        // Read all the fields.
        final DataInputStream dis = new DataInputStream(is);
        final long exhibitLength = dis.readLong();
        if(exhibitLength < 0) { throw new IOException("corrupt exhibit length"); }
        if(exhibitLength == 0)
            {
            // Potentially the client asking to drop the connection.
            if(BATCH_UPLOAD_MSG_TERMINATOR != dis.readByte())
                { throw new IOException("corrupt exhibit length or input termination"); }

            return(null); // EOF...
            }
        final String exhibitName = dis.readUTF();
        // Check name.
        if((exhibitName.length() != 0) && !ExhibitName.validNameSyntax(exhibitName))
            {
            // Corrupt message (invalid name).
            throw new IOException("invalid exhibit name");
            }

        // Get the hash length/content.
        final int hashLen = dis.readInt();
        if(hashLen != 16) { throw new IOException("bad MD5 hash length ("+hashLen+"B vs 16B expected)"); }
        final byte hashBuf[] = new byte[hashLen];
        dis.readFully(hashBuf);
        final ROByteArray hash = new ROByteArray(hashBuf);

        // Read the description.
        final String description = dis.readUTF();
        if(description.length() > CoreConsts.DESCRIPTION_MAX_CHARS)
            {
            throw new IOException("over-length description");
            }

        return(new BatchUploadClientRequest(exhibitName, exhibitLength, hash,  description));
        }
    }
