package org.hd.d.pg2k.svrCore.vars;

import org.hd.d.pg2k.svrCore.*;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicReference;

/**Base class to "manage" the set of system variables.
 * This holds the actual values and does:
 * <ul>
 * <li>Type checking.
 * <li>Merging local values into globalmap values.
 * <li>Trimming entries in global maps for servers now dead/unreachable.
 * </ul>
 * <p>
 * Thread-safe; does not hold any externally-visible locks.
 * <p>
 * This implements the generic variable-manager interface.
 */
public class BasicVarMgr implements BasicVarMgrInterface
    {
    /**If true, this class (and possibly its derivatives) are in debug mode.
     * This will normally result in extra []-bracketted comments in System.err.
     */
    protected static final boolean isDebugMode = false /* && ORG.hd.d.IsDebug.isDebug */;

    /**Create a normal set of managed system variables, initially empty.
     * This instance is not marked as being at an end point.
     */
    public BasicVarMgr()
        { this(false); }

    /**Create a set of managed system variables.
     * If an instance is created as an end point,
     * that means that it is the master variable copy in the master server,
     * and so a unique, local, read-only instance ID is created
     * and inserted into the variable cache.
     *
     * @param  endPoint  if true, mark the system as an end-point
     */
    public BasicVarMgr(final boolean endPoint)
        {
//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": constructor.]"); }

        this.endPoint = endPoint;
        if(endPoint)
            {
            final InstanceID iid = InstanceID.createInstanceID();
            final SimpleVariableValue svv = new SimpleVariableValue(
                SystemVariables.LOCAL_SYS_ID, iid);

            // The instance ID variable must be local and read-only.
            assert(SystemVariables.LOCAL_SYS_ID.isReadOnly());
            assert(SystemVariables.LOCAL_SYS_ID.isLocal());

            variables.put(SystemVariables.LOCAL_SYS_ID, svv);

//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": isEndPoint, ID=" + iid + ".]"); }
            }
        }

    /**If true, this variable set is at an end-point. */
    private final boolean endPoint;

    /**Returns true if this is at an endpoint. */
    public boolean isEndPoint() { return(endPoint); }


    /**Cached variable values as map; never null.
     * Map from SimpleVariableDefinition to SimpleVariableValue.
     * <p>
     * Used by:
     * <ul>
     * <li>getVariable() and getVariables(),
     * <li>setVariable(), which can update it too.
     * </ul>
     * <p>
     * A lock must be held on this object to do multiple operations atomically,
     * but no other visible lock must be grabbed inside a lock on this.
     * <p>
     * Mappings are never remove()d from this table;
     * it is limited in size to the maximum number of variables
     * defined in SystemVariables.
     * <p>
     * Hashtable from SimpleVariableDefinition to SimpleVariableValue.
     * <p>
     * A Hashtable is used for thread-safety.
     * <p>
     * Never null.
     */
    private final Hashtable<SimpleVariableDefinition,SimpleVariableValue> variables =
            new Hashtable<SimpleVariableDefinition, SimpleVariableValue>();

    /**Map of event history; never null.
     * Map from definition to the (recent) history of an event,
     * to be updated when a setVariable() is done on an event-type variable.
     * <p>
     * A lock must be held on this object to do multiple operations atomically,
     * but no other visible lock must be grabbed inside a lock on this.
     * <p>
     * Mappings are never remove()d from this table;
     * it is limited in size to the maximum number of variables
     * defined in SystemVariables.
     * <p>
     * Hashtable from SimpleVariableDefinition to SimpleVariableValue.
     * <p>
     * A Hashtable is used for thread-safety.
     * <p>
     * Never null.
     */
    private final Hashtable<SimpleVariableDefinition,EventVariableValueSet> eventHistory =
            new Hashtable<SimpleVariableDefinition, EventVariableValueSet>();


    /**Set variable.
     * Set local cached value immediately;
     * store global values to periodically propagate upstream to master
     * but show last global values obtained from master on periodic poll.
     * <p>
     * We validate the the name/type of the variable being presented.
     * <p>
     * For local and globalMap variables,
     * the last value set is the one that can be retrieved.
     *
     * @throws java.lang.IllegalArgumentException  on attempt to:
     *     set variable with value of wrong type or incompatible defintion,
     *     set non-existent or read-only variable (or these can be ignored)
     */
    public void setVariable(final SimpleVariableValue newValue)
        throws IOException
        {
//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": setVariable("+newValue+").]"); }

        _setVariable(newValue, false);

//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": setVariable("+newValue+") FINISHED.]"); }
        }

    /**Set variable for upstream and downstream values.
     * Set local cached value immediately;
     * store global values to periodically propagate upstream to master
     * but show last global values obtained from master on periodic poll.
     * <p>
     * We validate the the name/type of the variable being presented.
     * <p>
     * Since this allows read-only values to be updated,
     * this is only visible to derived classes.
     * <p>
     * For local and globalMap variables,
     * the last value set is the one that can be retrieved.
     * <p>
     * When setting a globalMap variable,
     * we revisit the list of potentially-live systems and trim it if need be.
     *
     * @param fromUpstream  if true, we allow read-only variables to be set
     *     in our cache as they represent updated values from the source
     *     and we only allow updates with globalMaps of at least size 1
     *
     * @throws IllegalArgumentException  on attempt to:
     *     set variable with value of wrong type or incompatible defintion,
     *     set non-existent or read-only variable (or these can be ignored)
     */
    protected void _setVariable(final SimpleVariableValue newValue,
                               final boolean fromUpstream)
        {
        if(newValue == null) { throw new IllegalArgumentException(); }

//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": _setVariable("+newValue.getDef() +':'+ newValue+", "+fromUpstream+").]"); }

        final SimpleVariableDefinition def = newValue.getDef();
        // Reject if not a variable known on this system.
        if(!SystemVariables.defs.contains(def))
            { throw new IllegalArgumentException("bad variable name or type: " + def); }

        // Disallow updates from "upstream"
        // when at an upstream end-point.
        if(endPoint && fromUpstream)
            { throw new IllegalStateException("at upstream endpoint; cannot accept updates from upstream"); }

        if(!fromUpstream)
            {
            // If this is a read-only variable
            // then reject the attempted update if from a set operation.
            if(def.isReadOnly())
                { throw new IllegalArgumentException("variable is read-only and may not be set (from downstream): " + def); }

            // If it has a globalMap,
            // then reject updates with more than one entry.
            final Map globalMap = newValue.getGlobalMap();
            if((globalMap != null) && (globalMap.size() > 1))
                { throw new IllegalArgumentException("variable has globalMap with more than one entry and may not be set (from downstream): " + def); }
            }

        // Allow an incoming global with an empty globalMap
        // as indicative of an "expired" value,
        // but only if its timestamp is a long time in the past
        // and the variable really would have expired by now...
        if(fromUpstream &&
            !def.isLocal() &&
            (newValue.getTimestamp() + 2*SystemVariables.MAX_QUIET_SYSTEM_VAR_LIFE_MS > System.currentTimeMillis()))
            {
            // From upstream must have at least one entry if a global,
            // unless very old.
            final Map globalMap = newValue.getGlobalMap();
            if((globalMap == null) || (globalMap.size() < 1))
                {
//                System.err.println("WARNING: non-local, non-expired, upstream variable value has empty globalMap: " + newValue.getFullDescription());
                return; /* Quietly discard for now. */
                }
//                { throw new IllegalArgumentException("non-local, non-expired, upstream variable value has empty globalMap: " + newValue.getFullDescription()); }
            }

//        // Report when values propagate up or down...
//        if(isDebugMode)
//            {
//            if(!newValue.equals(variables.get(def)))
//                {
//                System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": UPDATING VALUE "+(fromUpstream?"(from upstream)":"")+": _setVariable("+newValue.getFullDescription()+", "+fromUpstream+").]");
//                }
//            }

        // Merge the new value into the cache.
        _mergeValues(newValue, fromUpstream);


        // If this is an event,
        // record it in the event history as appropriate.
        if(def.isEvent())
            {
            synchronized(eventHistory)
                {
                // Get any extant value from the cache.
                EventVariableValueSet extant =
                        eventHistory.get(def);

                // If there isn't one, create an empty one.
                if(extant == null)
                    {
                    extant = new EventVariableValueSet(def);
                    eventHistory.put(def, extant);
                    }

                // Record this event if it is not from upstream.
                if(!fromUpstream)
                    { extant.addEvent(newValue); }
                }
            }
        }

    /**Merge a new value into a variable in the cache.
     * This very carefully merges any extant cached value
     * with a set coming upstream or a new global value coming downstream,
     * trying to cope with the fact that a new value outbound may meet
     * an updated global value coming back from upstream
     * (from the master for example).
     * <p>
     * We always immeidately apply a local
     * or override the value and globalMap entry of a global,
     * though for a global it will probably require the value merged
     * at the master's end-point coming back to sort things out,
     * with some instability/flapping in the interim.
     * <p>
     * The update must already have been validated,
     * eg must not have more than one globalMap entry if a set.
     * <p>
     * This deals with events as well as plain variables.
     * <p>
     * Takes a lock on variables to make this atomic.
     *
     * @param newValue  a variable update/set coming upstream/downstream
     * @param fromUpstream  if true, this is an update coming downstream
     */
    private void _mergeValues(final SimpleVariableValue newValue,
                              final boolean fromUpstream)
        {
        if(newValue == null)
            { throw new IllegalArgumentException(); }

        final SimpleVariableDefinition def = newValue.getDef();

        // If local, apply the value unconditionally...
        if(def.isLocal())
            {
            variables.put(def, newValue); // Atomic update...
//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": _mergeValues("+newValue.getDef() +':'+ newValue+", "+fromUpstream+"): LOCAL UPDATE/SET; prev: "+extant+"]"); }
            return;
            }


        // WE ARE UPDATING A GLOBAL...

        // Get the local system ID, ready for a merge operation.
        final SimpleVariableValue iidVar = variables.get(SystemVariables.LOCAL_SYS_ID);
        if(iidVar == null)
            { throw new IllegalStateException("Cannot set global variable without our system ID; should this instance be an endpoint?"); }
        final InstanceID sysID = (InstanceID) iidVar.getValue();

        // Do the merge atomically...
        synchronized(variables)
            {
            // Get any extant value from the cache.
            final SimpleVariableValue extant = variables.get(def);

            // If the new value is a set (from downstream),
            // and has no global map entry,
            // then we assume it is locally generated and inject one for us,
            // preserving the existing timestamp.
            final SimpleVariableValue adjNewValue =
                (!fromUpstream && (newValue.getGlobalMap() == null)) ?
                newValue.put(sysID, newValue, true) : newValue;

            // By now, we should not see any value with an empty globalMap,
            // except possibly an "expired" value from upstream...
            assert(fromUpstream || (adjNewValue.getGlobalMap() != null));
            if(!fromUpstream && (adjNewValue.getGlobalMap() == null))
                { throw new IllegalStateException("ERROR: should NOT be set()ting global with empty globalMap"); }

            // If the extant value is null,
            // then save the new value as-is and return immediately.
            if(extant == null)
                {
                variables.put(def, adjNewValue);
//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": g1: _mergeValues("+adjNewValue.getFullDescription()+", "+fromUpstream+"): GLOBAL FIRST UPDATE/SET.]"); }
                return;
                }

            // Merge in new values from the update,
            // essentially taking something like the union
            // of the old and new values.
            SimpleVariableValue toSave = extant;
            final Map newGM = adjNewValue.getGlobalMap();
            if(newGM != null)
                {
                for(final Iterator it = newGM.keySet().iterator(); it.hasNext(); )
                    {
                    final InstanceID key = (InstanceID) it.next();
                    final SimpleVariableValue svv = (SimpleVariableValue) newGM.get(key);
                    toSave = toSave.put(key, svv, true);
                    }
                variables.put(def, toSave);
//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": gm: _mergeValues("+toSave.getFullDescription()+", "+fromUpstream+"): GLOBAL UPDATE/SET; prev: "+extant.getFullDescription()+".]"); }
                }
            }
        }

    /**Update a number of variables at once for efficiency; returns the number set.
     * Is passed a Set of SimpleVariableValues and behaves as if it
     * operates on all of them by calling setVariable() for each item
     * in the Set in order (from low to high index).
     * <p>
     * This implementation "fails fast" on the first error.
     *
     * @return the number of variable values set;
     *     never negative, never more than the number passed in
     *
     * @throws java.lang.IllegalArgumentException  on attempt to:
     *     set variable with value of wrong type or incompatible definition,
     *     set non-existent or read-only variable (or these can be ignored)
     */
    public int setVariables(final SimpleVariableValue newValues[])
        throws IOException
        {
        if(newValues == null) { throw new IllegalArgumentException(); }

//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": setVariables("+Arrays.asList(newValues)+").]"); }

        final int nvLen = newValues.length;
        for(int i = 0; i < nvLen; ++i)
            { _setVariable(newValues[i], false); }

//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": setVariables("+Arrays.asList(newValues)+") FINISHED.]"); }

        // Claims to have set all the values passed to it.
        return(nvLen);
        }

    /**Get a single variable value; returns null if no such value or wrong type.
     * Always get from local cache.
     * <p>
     * This trims from the globalMap any apparently-dead systems,
     * and stores any such trimmed variable back in the cache atomically
     * so as to incrementally eliminate stale data
     * and return only reasonably fresh data to the caller.
     *
     * @param var definition of variable to fetch; never null
     *
     * @throws IllegalArgumentException if var is null
     */
    public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
        {
//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": getVariable("+var+").]"); }

        if(var == null)
            { throw new IllegalArgumentException(); }

        // Do fetch/winnowing operation atomically.
        synchronized(variables)
            {
            // Get the variable if set/extant,
            // else return null.
            final SimpleVariableValue svv =
                    variables.get(var);

            // If null or no globalMap entries,
            // then return as-is immediately;
            // no winnowing to be done...
            if((svv == null) || (svv.getGlobalMap() == null))
                { return(svv); }

            // Attempt to trim out dead very old values from the map.
            final SimpleVariableValue svvWinnowed =
                svv.removeAllKeysOlder(System.currentTimeMillis() -
                    2*SystemVariables.MAX_QUIET_SYSTEM_VAR_LIFE_MS);

            // If this actually trimmed anything,
            // quietly replace the cached entry with the trimmed version.
            // This way we incrementally remove stale data.
            final Map gMapWinnowed = svvWinnowed.getGlobalMap();
            if((gMapWinnowed == null) ||
               (gMapWinnowed.size() < svv.getGlobalMap().size()))
                {
//if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": getVariable() trimmed globalMap from "+svv.getGlobalMap().size()+" to "+((gMapWinnowed==null)?0:gMapWinnowed.size())+" entries.]"); }
                // Store trimmed variable,
                // discarding stale state.
                variables.put(var, svvWinnowed);
                return(svvWinnowed);
                }

            // Original was OK,
            // so return that to avoid duplicates in the heap.
            return(svv);
            }
        }

    /**Get set of variable values altered on or after a given time, or all for -1; never null.
     * Always get from local cache
     * (the variable cache being periodically updated from the master).
     * <p>
     * This may be slow if there are many live variables.
     * <p>
     * This implementation never throws an IOException.
     * <p>
     * Note: we may limit the notion of "all" to be something a little greater
     * than the maximum live-variable time window (ie variable lifetime),
     * since beyond this the variable value has expired
     * and is not deemed trustworthy anyway.
     */
    public SimpleVariableValue[] getVariables(final long changedSince)
        {
//        if(isDebugMode) { System.err.println("[BasicVarMgr "+System.identityHashCode(this)+": getVariables("+changedSince+").]"); }

        // Get definitions of all extant variables atomically.
        // Note that we never remove variables once set.
        final Set<SimpleVariableDefinition> variableDefs;
        synchronized(variables)
            { variableDefs = new HashSet<SimpleVariableDefinition>(variables.keySet()); }

        // Approximately size the result...
        final List<SimpleVariableValue> result = new ArrayList<SimpleVariableValue>(variableDefs.size());

        // Compute a backstop based on the maximum variable life
        // and the maximum permitted clock skew between participating systems.
        // We use quite a conservative value to avoid over-filtering here.
        final long backStop = System.currentTimeMillis() - 2*SystemVariables.MAX_QUIET_SYSTEM_VAR_LIFE_MS
                                                         - 2*CoreConsts.MAX_PEER_CLOCK_SKEW_MS;
        final long cutoffTime = Math.max(backStop, changedSince);

        // Go through the set,
        // removing anything definitely older than the specified time,
        // and winnowing out stale data.
        for(final Iterator<SimpleVariableDefinition> it = variableDefs.iterator(); it.hasNext(); )
            {
            final SimpleVariableDefinition def = it.next();

            // Get value, trimmed of stale data if necessary.
            final SimpleVariableValue svv = getVariable(def);

            // If timestamp is new enough,
            // add the value to the result.
            if((svv.getTimestamp() >= cutoffTime)) // TODO: test this clause!
                { result.add(svv); }
            }

        final SimpleVariableValue r[] = new SimpleVariableValue[result.size()];
        result.toArray(r);
        return(r);
        }


    /**Get the current partial, or previous full, event set at the specified interval; never returns null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     * <p>
     * If the requested event set is not available,
     * an empty synthetic one is created and returned.
     * Thus, with this interface, it is not possible to distinguish between
     * there being no events in the given interval or simply not data.
     * <p>
     * This calls getEventValues().
     * <p>
     * This implementation will not throw an IOException.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  never null
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; never null
     *
     * @throws IllegalArgumentException  if the request arguments are invalid
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        if((def == null) || !def.isEvent() || (intervalSelector == null))
            { throw new IllegalArgumentException(); }

        final long now = System.currentTimeMillis();
        final long currentIntervalNumber = intervalSelector.getIntervalNumber(now);
        final long intervalNumber = (current ? currentIntervalNumber : (currentIntervalNumber-1));

        try
            {
            final EventVariableValue evv[] = getEventValues(def,
                                                            intervalSelector,
                                                            intervalNumber,
                                                            null);

            // If we got some real (non-null) data back, then return it...
            if((evv.length > 0) && (evv[0] != null))
                { return(evv[0]); }
            }
        catch(final IOException e)
            {
            // Fall through to return empty non-auth default result.
            }

        // Else invent empty non-authoritative return value (with no events)...
        // TODO: Cache these to save some redunant object creation?
        return(new EventVariableValue(false,
                                      def,
                                      intervalSelector,
                                      intervalNumber,
                                      0, null, null));
        }

    /**Get the specified event sets for the specified intervals; never null.
     * This implementation will not throw an IOException.
     *
     * @param intervalNumber  the number of the first interval
     *     for which data is potentially required;
     *     if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc;
     *     null is treated as the common case equivalent to just bit 0 set
     *
     * @return as many of the requested values as available,
     *     return may contain nulls or be zero-length but is never null
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException
        {
        if((def == null) || !def.isEvent() ||
                (intervalSelector == null) ||
                (intervalNumber < 0))
            { throw new IllegalArgumentException(); }
        final Set<EventPeriod> evPeriodSubset = def.getEvPeriodSubset();
        if((evPeriodSubset != null) && !evPeriodSubset.contains(intervalSelector))
            { throw new IllegalArgumentException("bad interval requested"); }

        final EventVariableValueSet evvs = eventHistory.get(def);

        // If no entry at all for this event, then return an empty result set.
        if(evvs == null) { return(EventVariableValuePeriodRow._NO_EVENT_VALUES); }

        assert(def.equals(evvs.getDef()));

        // Delegate handling of request to event set.
        return(evvs.getEventValues(intervalSelector, intervalNumber, whichValues));
        }


    /**Set/override the specified event sets for the specified intervals; never null.
     * This is used to set values that have been fetched from upstream
     * (ie closer to the master data store).
     * <p>
     * This may ignore the call,
     * and certainly will if the intervalNumber is outside the current time window it holds.
     * <p>
     * This will generally accept the value and store it if:
     * <ul>
     * <li>The intervalNumber is within the current window.
     * <li>The intervalSelector is a period for which this event has data.
     * <li>The supplied value is "better" than any current one held, ie:
     *     <ul>
     *     <li>We hold a null and evv is non-null.
     *     <li>We hold a non-authoritative value and evv is authoritative.
     *     <li>The authoritative-ness is the same but we hold an entry
     *         with a lower total count than evv.
     *     <ul>
     *     </ul>
     * </ul>
     * Howwever, we will <em>not</em> accept non-authoritative updates
     * if at the end-point (ie this is the master store),
     * since we are the authoritative source of data.
     *
     * @param evv  non-null event value
     */
    public synchronized void setEventValue(final EventVariableValue evv)
//        throws IOException
        {
        if(evv == null)
            { throw new IllegalArgumentException(); }
        final SimpleVariableDefinition def = evv.getDef();
        if(!def.isEvent())
            { throw new IllegalArgumentException(); }

        // If we are at an endpoint
        // then disregard attempts to set non-auth overrides.
        if(endPoint && !evv.isAuthoritative())
            { return; }

        // Find our cached event history for this event,
        // atomically creating it if need be.
        EventVariableValueSet extant;
        synchronized(eventHistory)
            {
            // Get any extant value from the cache.
            extant = eventHistory.get(def);

            // If there isn't one, create an empty one.
            if(extant == null)
                {
                extant = new EventVariableValueSet(def);
                eventHistory.put(def, extant);
                }
            }

        extant.setEventValue(evv);
        }




    /**Persist variables and events to disc.
     * This can either save everything applicable,
     * or just those things things that seem to be out-of-date on disc.
     * <p>
     * Only persistent event data is saved,
     * and generally each named event type has its data saved in a separate file,
     * in a form that should be reasonably robust
     * (eg a failure to load one item should not have all items unloadable).
     * <p>
     * This data may be stored in binary serialised or XML forms,
     * but is usually GZIPped to save space and speed up read/write.
     * <p>
     * If some saves fail then this will try to continue with others,
     * but throw one of the errors encountered (usually the last one)
     * at the end, so as to save as much as possible.
     * <p>
     * The incremental mode helps keeps the duration of any one call
     * as small as is reasonably possible, but will be somewhat less efficent
     * then non-incremental mode.
     * <p>
     * This locks out other activity while saving to ensure consitency;
     * it need not hold a lock continuously for whole period of the call
     * as it would usually be enough to do so while collecting candidate
     * histories to save, and while actually saving each one of them separately.
     * <p>
     * To avoid deadlocks, this takes a lock on this instance
     * before taking any other lock.
     *
     * @param dir  the directory to save to; never null
     * @param justChanges  just write histories changed since the last save if possible,
     *     else unconditionally save all persistent event histories
     * @param keepHistory  if true, try to roll the logs at least daily
     *     to help keep a trail of older snapshot though the storage space
     *     required may grow without bound,
     * @param incremental  if true, then at most one event is saved,
     *     chosen in such a way (eg by max time since last save or randomly)
     *     that eventually all eligible histories will be saved
     */
    public void saveEventHistories(final File dir,
                                   final boolean justChanges,
                                   final boolean keepHistory,
                                   final boolean incremental)
        throws IOException
        {
        if(dir == null)
            { throw new IllegalArgumentException(); }

        if(!dir.isDirectory())
            { throw new FileNotFoundException("not a directory: " + dir); }

        if(!dir.canWrite())
            { throw new IOException("cannot write to " + dir); }

        // Create/format the date string once, if we are going to use it.
        final String dateString = (!keepHistory) ? "" : ("." + dateFmtEHFile.format(new Date()));

        // Get potential events to be saved.
        // Collect this list under a lock on the eventHistory to avoid races.
        final List<SimpleVariableDefinition> events;
        synchronized(this)
            {
            synchronized(eventHistory)
                { events = new ArrayList<SimpleVariableDefinition>(eventHistory.keySet()); }
            }
        // Randomise the ordering for incremental mode to avoid starvation.
        // We trust the built-in random generator to be OK...
        if(incremental)
            { Collections.shuffle(events); }

        // Iterate over event data,
        // saving each persistent event's history to a separate file.
        // Include today's date in the name to provide backups/history where requested.
        IOException err = null; // Any error encountered.
        for(final SimpleVariableDefinition def : events)
            {
            assert(def.isEvent()); // Should not be any non-event variable data in here!

            // Only save persistent events' data.
            if(!def.isPersistent()) { continue; }

            final EventVariableValueSet history = eventHistory.get(def);
            if(history == null) { continue; }

            final StringBuilder sb = new StringBuilder(80);
            sb.append(EVENT_STORE_FILENAME_PREFIX);
            sb.append(def.getName());
            sb.append(EVENT_STORE_NAMETERM);
            sb.append(dateString); // Optional date...
            sb.append(EVENT_STORE_FILENAME_SUFFIX_SER_GZ);

            // Persist the data if necessary.
            final File file = new File(dir, sb.toString());
            try
                {
                // If only saving changes,
                // we see if any events arrived since the file was last saved.
                if(justChanges)
                    {
                    // If the history file appears to be present and correct
                    // and newer than the last added event
                    // then skip the save...
                    // We don't care much about minor race possibilites here.
                    if(file.exists() && file.isFile() && file.canRead() && (file.length() > 0) &&
                            (history.getLastEventTime() < file.lastModified()))
                        { continue; }
                    }

                // Save the data as a compressed serialised object...
                // Lock out mutators on this object during the save.
                synchronized(this)
                    {
                    synchronized(history)
                        { FileTools.serialiseToFile(history, file, SAVE_EVENT_HISTORIES_GZIPPED, !ORG.hd.d.IsDebug.isDebug); }
                    }

                // If incremental, bucket out after the first successful save.
                if(incremental)
                    { break; }
                }
            catch(final IOException e)
                {
                // Print and temporarily absorb any errors encountered during save.
                // Continue with other saves as far as possible.
                e.printStackTrace();
                err = e;
                }
            }

        // Throw the last error encountered, if any.
        if(err != null) { throw err; }
        }

    /**Our private static instance of a GMT calendar.
     * No one must change the ZONE or DST offset.
     */
    private static Calendar GMTCalendar = Calendar.getInstance();
    /**Initialise GMTCalendar. */
    static
        {
        // Set to GMT
        // i.e 0 Offset
        GMTCalendar.set(Calendar.ZONE_OFFSET, 0);
        // and 0 Daylight savings
        GMTCalendar.set(Calendar.DST_OFFSET, 0);
        }

    /**Format we use to insert into event history file name. */
    private static final SimpleDateFormat dateFmtEHFile =
        new SimpleDateFormat("yyyyMMdd");
    /**Set the Calendar for dateFmtEHFile to be the GMT calendar. */
    static { dateFmtEHFile.setCalendar(GMTCalendar); }

    /**If true, GZIP the saved histories to save space and read/write time. */
    private static final boolean SAVE_EVENT_HISTORIES_GZIPPED = true;

    /**Prefix to event history store file name. */
    public static final String EVENT_STORE_FILENAME_PREFIX = "eventStore.";

    /**Terminator to go after event name in file store.
     * Format of file is:
     * <pre>
     *     EVENT_STORE_FILENAME_PREFIX + name + EVENT_STORE_NAMETERM [ + date ] + legal suffix
     * </pre>
     * <p>
     * The nameterm includes at least one character that cannot appear in
     * an event name to avoid ambiguity where one name is a prefix of another.
     */
    public static final String EVENT_STORE_NAMETERM = "-T";

    /**Suffix for GZIPped serialised event history store file name. */
    public static final String EVENT_STORE_FILENAME_SUFFIX_SER_GZ = ".ser.gz";


    /**Load variables and events from disc.
     * This will load the "latest" saved state, if any,
     * for each persistent event.
     * "Latest" may be measured by file timestamp (potentially slow and fragile)
     * or lexically-latest (eg by the date embedded in the filename)
     * or by some combination of the two.
     * <p>
     * Only the event history will be loaded.
     * <p>
     * If some loads fail then this will try to continue with others,
     * but throw one of the errors encountered (usually the last one)
     * at the end, so as to load as much as possible.
     * <p>
     * This may lock out other activity while loading.
     *
     * @param dir  the directory to load from; never null
     */
    public synchronized void loadEventHistories(final File dir)
        throws IOException
        {
        if(dir == null)
            { throw new IllegalArgumentException(); }

        if(!dir.isDirectory())
            { throw new FileNotFoundException("not a directory: " + dir); }

        if(!dir.canRead())
            { throw new IOException("cannot read from " + dir); }

        final long startTime = System.currentTimeMillis();

        // Get listing of files in the directory...
        final File files[] = dir.listFiles();

        // Sort into natural order, so later-dated files should appear later.
//        Arrays.sort(files);

        // Construct a short-list of those that look like usable, ie are:
        //   * starts with correct prefix
        //   * contains the name terminator
        //   * ends with a legal suffix (may be more than one in future)
        //
        // We should  probably also only accept candidiates that are:
        //   * plain files
        //   * readable
        //   * non-zero-length
        // but retrieval of such stats from disc at start-up may be slow.
        // This shortlist is of short filenames, with no directory components.
        final List<String> candiates = new ArrayList<String>(files.length);
        for(final File f : files)
            {
            final String name = f.getName();
            if(!name.startsWith(EVENT_STORE_FILENAME_PREFIX)) { continue; }
            if(-1 == name.indexOf(EVENT_STORE_NAMETERM)) { continue; }
            if(!name.endsWith(EVENT_STORE_FILENAME_SUFFIX_SER_GZ)) { continue; }

//            final File fullPath = new File(dir, name);
//            if(!fullPath.isFile()) { continue; }
//            if(!fullPath.canRead()) { continue; }
//            if(fullPath.length() <= 0) { continue; }

            // Good candidate, so add it.
            candiates.add(name);
            }

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[BasicVarMgr.loadEventHistories(): found event history "+candiates.size()+" candidates in "+dir+".]"); }

        if(candiates.size() == 0)
            {
            System.err.println("WARNING: BasicVarMgr.loadEventHistories() found no event-history files in " + dir);
            return;
            }

        // Compute list of peristent events.
        final List<SimpleVariableDefinition> defs =
            new ArrayList<SimpleVariableDefinition>(SystemVariables.nameToDef.size());
        for(final SimpleVariableDefinition def : SystemVariables.nameToDef.values())
            {
            if(!def.isEvent() || !def.isPersistent()) { continue; }
            defs.add(def);
            }
        if(defs.isEmpty()) { return; /* Nothing to do... */ }

        // Tasks running.
        final List<Future<?>> tasks = new ArrayList<Future<?>>(defs.size());
        
        // Becomes non-null if any task fields an IOException.
        final AtomicReference<IOException> err = new AtomicReference<IOException>();

        // Start the tasks using the I/O-bound pool.
        for(final SimpleVariableDefinition def : defs)
            {
            tasks.add(ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    // For each peristent event in the SystemVariables list,
                    // look for the newest/best candidate file.
                    // (Double-check after deserialising
                    // that we have loaded data for the right def!)

                    // Best candidate so far...
                    File bestCandidate = null;

                    // Minimum non-ambiguous prefix for an event-store file
                    // for this event name.
                    final String minUniquePrefix = EVENT_STORE_FILENAME_PREFIX +
                            def.getName() + EVENT_STORE_NAMETERM;
                    for(final String name : candiates)
                        {
                        if(!name.startsWith(minUniquePrefix)) { continue; }

                        final File candiate = new File(dir, name);

                        // If there is no existing candidate,
                        // or this one is newer,
                        // then this becomes our best candidate.
                        if((bestCandidate == null) ||
                           (candiate.lastModified() > bestCandidate.lastModified()))
                            {
                            // Skip/veto obviously bogus/broken files.
                            if(!candiate.isFile()) { continue; }
                            if(!candiate.canRead()) { continue; }
                            if(candiate.length() <= 0) { continue; }

                            // Looks to be the best so far.
                            bestCandidate = candiate;
                            }
                        }

                    // If a suitable candidate has been found,
                    // then try to load it.
                    if(bestCandidate != null)
                        {
                        try
                            {
                            final Object o =
                                FileTools.deserialiseFromFile(bestCandidate, SAVE_EVENT_HISTORIES_GZIPPED);

                            if(!(o instanceof EventVariableValueSet))
                                { throw new InvalidPropertiesFormatException("unexpected event history data deserialised from " + bestCandidate); }

                            final EventVariableValueSet evvs = (EventVariableValueSet) o;
                            if(!def.equals(evvs.getDef()))
                                { throw new InvalidPropertiesFormatException("wrong event history data deserialised from " + bestCandidate); }

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[BasicVarMgr.loadEventHistories(): loaded history for "+def+".]"); }

                            // OK, got the data, so post it into our table atomically...
                            eventHistory.put(def, evvs);
                            }
                        catch(final IOException e)
                            {
                            err.set(e); // Note the error, but absorb it for now.
                            e.printStackTrace();
                            }
                        }
                    }
                }));
            }

        // Wait for worker tasks to finish.
        assert(tasks.size() == defs.size());
        for(final Future<?> t : tasks)
            {
            try { t.get(); }
            catch(final InterruptedException e) { Thread.currentThread().interrupt(); }
            catch(final ExecutionException e)
                {
                e.printStackTrace();
                final Throwable cause = e.getCause();
                // If underlying cause is an IOException then rethrow it directly.
                if(cause instanceof IOException) { throw (IOException) cause; }
                final IOException ioe = new IOException("failure in loader thread");
                ioe.initCause(e);
                throw ioe;
                }
            }

if(ORG.hd.d.IsDebug.isDebug) { System.out.println("[BasicVarMgr.loadEventHistories(): "+(System.currentTimeMillis()-startTime)+"ms.]"); }

        // Throw a sample error of those that we encountered, if any.
        final IOException sampleErr = err.get();
        if(sampleErr != null)
            { throw sampleErr; }
        }
    }
