package org.hd.d.pg2k.svrCore.vars;

import java.io.IOException;
import java.util.BitSet;

/**
 * Created by IntelliJ IDEA.
 * User: Damon
 * Date: 31-Aug-2004
 * Time: 20:23:59
 */

/**Basic variable manager interface.
 * The variable manager may or may not be part of a pipeline.
 */
public interface BasicVarMgrInterface
    {
    /**Set variable to the given value (the variable name and definition are implicit).
     * The value set should be immediately readable with getVariable()
     * as the "main" value of the variable even for global variables.
     * <p>
     * This is also used to send events,
     * the last event value being the variable's value.
     *
     * @throws IOException  in case of I/O difficulty
     * @throws UnsupportedOperationException  if a variable is set that
     *     we could never handle, eg a non-System-ID local variable at
     *     the client end of a tunnel
     */
    void setVariable(SimpleVariableValue newValue)
        throws IOException,
               UnsupportedOperationException;

    /**Update a number of variables at once for efficiency; returns the number of variables set.
     * Is passed a set of SimpleVariableValues and behaves as if it
     * operates on all of them by calling setVariable() for each item
     * in the set.
     * <p>
     * Furthermore it behaves as if the set operations are done in order,
     * from low index to high index,
     * unless variables are coalesced
     * (multiple sets of the same simple variable can keep just the last value).
     * Set operations of different variables may be re-ordered with respect to
     * one another in order that redundancy may reduced for transmission
     * and storage, so this implicitly allows, for example, events
     * to be re-ordered with respect to non-event variables.
     * <p>
     * An implementation may "fail fast" on the first error/exception,
     * or may attempt to continue and do as much as possible.
     * <p>
     * An implementation may throw an IllegalArgumentException on attempt to:
     *     set a variable with value of wrong type or incompatible definition,
     *     set a non-existent or read-only variable (or these may be ignored)
     *
     * @return the number of variable values set (the length of the array);
     *     never negative, never more than the number passed in
     */
    int setVariables(SimpleVariableValue[] newValues)
        throws IOException;

    /**Get a single variable value; returns null if no such value or wrong type.
     * Implementations should, where possible,
     * trim from the globalMap of returned values any apparently-stale data,
     * so as to return only reasonably-live data to the caller.
     * <p>
     * Only the last event value for a given event variable
     * is returned by this method.
     *
     * @param var definition of variable to fetch; never null
     *
     * @throws IOException  in case of I/O difficulty
     * @throws UnsupportedOperationException  if a variable is requested that
     *     we could never supply, eg a non-System-ID local variable at
     *     the client end of a tunnel
     */
    SimpleVariableValue getVariable(SimpleVariableDefinition var)
        throws IOException,
        UnsupportedOperationException;

    /**Get set of variable values altered on or after specified time, or get all values with -1; never null.
     * This may be slow if there are many live variables.
     * <p>
     * Only the last event value of each event variable
     * is returned by this method.
     *
     * @throws IOException  in case of I/O difficulty
     */
    SimpleVariableValue[] getVariables(long changedSince)
        throws IOException;


    /**Get the current partial, or previous full, event set at the specified interval; never null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     * <p>
     * If the requested event set is not available,
     * an empty non-auth synthetic one is created and returned.
     * Thus, with this interface, it is not possible to distinguish between
     * there being no events in the given interval or simply no data at all,
     * but this is relatively simple and safe to use.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  never null
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; never null
     *
     * @throws IllegalArgumentException  if the request argument are invalid
     */
    EventVariableValue getEventValue(SimpleVariableDefinition def,
                                     EventPeriod intervalSelector,
                                     boolean current);

    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * (An implementation that is not an end-point may go upstream to fetch
     * missing values and cache them to satisfy future requests.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  never null
     * @param intervalNumber  the number of the first interval
     *     for which data is potentially required;
     *     if too far in the past or future then possibly no data
     *     will be available
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc;
     *     null is treated as the common case equivalent to just bit 0 set
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    EventVariableValue[] getEventValues(SimpleVariableDefinition def,
                                        EventPeriod intervalSelector,
                                        long intervalNumber,
                                        BitSet whichValues)
        throws IOException;
    }
