/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.MemoryTools;

import java.io.*;
import java.util.*;


/**Immutable value of a local or global system event variable.
 * This contains the type/definition plus a total event count
 * and a Map of event values.
 * <p>
 * To retrieve the very latest value that was set a SimpleVariableValue
 * should be retrieved instead.
 * <p>
 * This is serialisable to enable it to be sent over network connections,
 * persisted, etc.
 * <p>
 * Note that the interval number is strictly positive,
 * the interval number multiplied by the period is the starting time of the interval in ms,
 * and the starting time is always representable as a positive long.
 */
public final class EventVariableValue implements Serializable,
                                                 ObjectInputValidation
    {
    /**Construct a single variable value.
     * This type-checks the value, vetoing construction if invalid.
     * <p>
     * Will attempt to effectively intern() key values to minimise memory consumption
     * at the cost of some performance at construction/deserialistion.
     *
     * @param isAuthoritative  if true, this is authoritative data
     *     from the master data store
     * @param def  the variable definition; never null and must be an event
     * @param period  the event period for this sample; never null
     * @param intervalNumber  the period number; strictly positive
     * @param totalEventCount  the total event count; non-negative
     * @param values  event values in decreasing order of count;
     *     null == empty, not containing duplicates, same length as counts[] argument
     * @param counts  event counts for each item in values[],
     *     in non-increasing order, with a sum no more than total-event count;
     *     null == empty, not containing negative values, same length as values[]
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public EventVariableValue(final boolean isAuthoritative,
                              final SimpleVariableDefinition def,
                              final EventPeriod period,
                              final long intervalNumber,
                              final int totalEventCount,
                              final Object values[],
                              final int counts[])
        throws IllegalArgumentException
        {
        // We will try to intern() suitable values
        // to save space in memory and when serialised.
        this(isAuthoritative, def, period, intervalNumber, totalEventCount, values, counts, true);
        }

    /**Construct a single variable value.
     * This type-checks the value, vetoing construction if invalid.
     * <p>
     * Will attempt to effectively intern() key values to minimise memory consumption
     * at the cost of some performance at construction/deserialistion.
     * <p>
     * Package-visible so that only classes in this package (that we trust)
     * can avoid having their values intern()ed.
     *
     * @param isAuthoritative  if true, this is authoritative data
     *     from the master data store
     * @param def  the variable definition; never null and must be an event
     * @param period  the event period for this sample; never null
     * @param intervalNumber  the period number; strictly positive
     * @param totalEventCount  the total event count; non-negative
     * @param values  event values in decreasing order of count;
     *     null == empty, not containing duplicates, same length as counts[] argument
     * @param counts  event counts for each item in values[],
     *     in non-increasing order, with a sum no more than total-event count;
     *     null == empty, not containing negative values, same length as values[]
     * @param tryIntern  if false we do not try to intern() any of the values
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    EventVariableValue(final boolean isAuthoritative,
                              final SimpleVariableDefinition def,
                              final EventPeriod period,
                              final long intervalNumber,
                              final int totalEventCount,
                              final Object values[],
                              final int counts[],
                              final boolean tryIntern)
        {
        this.def = def;
        this.period = period;
        this.intervalNumber = intervalNumber;
        this.totalEventCount = totalEventCount;
        this.authoratitive = isAuthoritative;

        // Canonicalise empty values to save memory.
        this.values = ((values == null) || (values.length == 0)) ? NO_VALUES : ((Object[]) values.clone());
        this.counts = ((counts == null) || (counts.length == 0)) ? NO_COUNTS : ((int[]) counts.clone());

        // Construct fast-lookup info map from constituents...
        // And intern() all the values it seems sensible to while we are at it.
        final int vLen = this.values.length;
        if(vLen != this.counts.length) // One early bit of sanity checking...
            { throw new IllegalArgumentException("counts and values array must be the same size"); }
        final Map<Object, ValueInfo> m = new HashMap<Object, ValueInfo>(2*vLen);
        final boolean shouldIntern = tryIntern && shouldIntern(def);
        for(int i = 0; i < vLen; ++i)
            {
            final int count = this.counts[i];

            // Optimisation/tweak...
            // If we have the same count as the preceeding item,
            // then we can share the same lower rank
            // (which is probably nicer for the exhibit (1))
            // and the same ValueInfo object (which saves memory (2)).
            final Object value = (!shouldIntern) ? this.values[i] :
                (this.values[i] = MemoryTools.intern(this.values[i])); // Eliminate dups.
            if((i > 0) && (count == this.counts[i-1]))
                {
                m.put(value, m.get(this.values[i-1]));
                continue;
                }

            // The general case is that this value's rank
            // is its position in the array.
            m.put(value, new ValueInfo(i, count));
            }
        // Store unmodifiable version.
        info = Collections.unmodifiableMap(m);


        try { validateObject(); }
        catch(final InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage(), e); }
        }

    /**Private shared (immutable) value meaning no values, to save memory. */
    private static final Object[] NO_VALUES = new Object[0];

    /**Private shared (immutable) value meaning no counts, to save memory. */
    private static final int[] NO_COUNTS = new int[0];


    /**Returns true if we should probably automatically intern() the value.
     * We will do this is the value is likely to be bulky (eg is a String),
     * or may be long-lived (eg peristent) or in multiple copies.
     */
    static boolean shouldIntern(final SimpleVariableDefinition def)
        {
        return((def.getType() == SimpleVariableDefinition.TYPE_STRING) ||
               def.isPersistent());
        }


    /**Return a (new) authoritative value with the same content as the current value; never null.
     * This instance is unaltered,
     * and it and the value returned remain immutable,
     * though they may or may not share state.
     * <p>
     * (If this instance is already authoritative,
     * then "this" is returned to avoid wasting memory.)
     *
     * @return instance with "authoritative" flag set
     */
    public EventVariableValue makeAuthoritative()
        {
        if(authoratitive) { return(this); }
        return(new EventVariableValue(true, def, period, intervalNumber, totalEventCount, values, counts));
        }


    /**The variable definition; never null.
     * @serial
     */
    private final SimpleVariableDefinition def;

    /**Get the variable definition; never null. */
    public SimpleVariableDefinition getDef() { return(def); }

    /**The event period; never null.
     * @serial
     */
    private final EventPeriod period;

    /**Get the event period; never null. */
    public EventPeriod getPeriod() { return(period); }

    /**The interval number; strictly positive.
     * @serial
     */
    private final long intervalNumber;

    /**Get the interval number; strictly positive. */
    public long getIntervalNumber() { return(intervalNumber); }

    /**If true, this variable has come from the authoritative end-point.
     * Due to an old typo now fossilised in old data,
     * the spelling of this should remain "authoratitive" even though wrong!
     * Thank goodness for encapsulation!
     * 
     * @serial
     */
    private final boolean authoratitive;

    /**Returns true if this variable has come from the authoritative end-point. */
    public boolean isAuthoritative() { return(authoratitive); }

    /**Total event count; non-negative.
     * @serial
     */
    private final int totalEventCount;

    /**Get the total event count; non-negative. */
    public int getTotalEventCount() { return(totalEventCount); }

    /**Counts of the different (non-null) values in non-increasing-count order (ie best first).
     * All counts are non-negative.
     * <p>
     * Same length as values[].
     *
     * @serial
     */
    private final int counts[];

    /**Values of the different (non-null) events in non-increasing-count order (ie best first).
     * All counts are non-null and valid values as determined by the def.
     * <p>
     * Same length as counts[].
     *
     * @serial
     */
    private final Object values[];


    /**Generate human-readable summary of important state. */
    public String toString()
        {
        return(def.getName() + ":auth="+authoratitive+ ":period="+period+ ":intNum="+intervalNumber+ ":count="+totalEventCount);
        }



    /**Immutable info collected about one particular value of an event.
     */
    public static final class ValueInfo implements Serializable, ObjectInputValidation
        {
        /**Construct an instance.
         * Only needs to be constructed locally, so this need not be public.
         */
        ValueInfo(final int pos, final int cnt)
            {
            rank = pos;
            count = cnt;
            try { validateObject(); } catch(final InvalidObjectException e) { throw new IllegalArgumentException(e.getMessage()); }
            }
        /**Rank: 0 means top, ie highest count; non-negative. */
        public final int rank;
        /**Count is count of this value; strictly positive. */
        public final int count;
        /**Deserialise: use constructor for validation, defensive copying, etc. */
        protected Object readResolve() // throws ObjectStreamException
            { return(new ValueInfo(rank, count)); }
        /**Check that the object state is consistent and legal. */
        public void validateObject() throws InvalidObjectException
            { if((rank < 0) || (count <= 0)) { throw new InvalidObjectException("bad object"); } }
        /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
        private static final long serialVersionUID = 887928932975973213L;
        };


    /**Map from event value (eg String, Number, etc) to ValueInfo item for that value; never null after construction.
     * Immutable (unmodifiable) so does not change after construction and fast to access.
     */
    private /* transient */ final Map<Object, ValueInfo> info;

    /**Get immutable Map of info on all events counted; never null.
     * Map from the event value to the rank and count for that event.
     */
    public Map<Object, ValueInfo> getInfo()
        { return(info); }

    /**Get total number of distinct event values recorded; non-negative. */
    public int getTotalDistinctValues()
        { return(values.length); }


    /**Get count of this event value; never negative.
     * A result of zero may indicate either that
     * no events with the specified value were seen,
     * or that they were seen but were later displaced due to capacity limits.
     *
     * @throws IllegalArgumentException  if the argument is null
     *     or not a legal type/value
     */
    public int getCount(final Object value)
        throws IllegalArgumentException
        {
        if((value == null) || !def.checkType(value))
            { throw new IllegalArgumentException(); }

        final ValueInfo vi = info.get(value);
        if(vi == null) { return(0); }
        return(vi.count);
        }

    /**Get rank/ranking of this event value; never negative.
     * Zero means the highest count of all values seen,
     * one means the next highest, etc.
     * So low values are good.
     * <p>
     * If we have no count recorded for this value, ie no ranking,
     * the routine returns Integer.MAX_VALUE.
     *
     * @throws IllegalArgumentException  if the argument is null
     *     or not a legal type/value
     */
    public int getRank(final Object value)
        throws IllegalArgumentException
        {
        if((value == null) || !def.checkType(value))
            { throw new IllegalArgumentException(); }

        final ValueInfo vi = info.get(value);
        if(vi == null) { return(Integer.MAX_VALUE); }
        return(vi.rank);
        }

    /**Get value by rank.
     * @throws ArrayIndexOutOfBoundsException  if the rank is out of order
     */
    public Object getValueByRank(final int rank) { return(values[rank]); }

    /**Get count by rank; strictly positive.
     * @throws ArrayIndexOutOfBoundsException  if the rank is out of order
     */
    public int getCountByRank(final int rank) { return(counts[rank]); }




    /**Equality is based on the definition, period and interval number, but not the event values.
     */
    public boolean equals(final Object obj)
        {
        if(!(obj instanceof EventVariableValue)) { return(false); }
        final EventVariableValue o = (EventVariableValue) obj;

        if(authoratitive != o.authoratitive) { return(false); }
        if(!def.equals(o.def)) { return(false); }
        if(!period.equals(period)) { return(false); }
        if(intervalNumber != o.intervalNumber) { return(false); }

        // All relevant fields seem to match, so objects are equal.
        return(true);
        }

    /**The hash is built on a subset of the fields.
     * This implementation uses the hash of the definition and the interval number and period values.
     */
    public int hashCode()
        {
        return(def.hashCode() ^ ((int)intervalNumber) ^ period.getIntervalMs());
        }


    /**Customise how we save state when serialising.
     * In particular, drop redundant fields that we cannot yet
     * mark as transient if we want to read old-format data.
     */
    private void writeObject(final ObjectOutputStream oos)
        throws IOException
        {
//        oos.defaultWriteObject();

        // Capture just the fields that we want to persist.
        final ObjectOutputStream.PutField fields = oos.putFields();
        fields.put("authoratitive", authoratitive); // NOTE: misspelling in historical serial data.
        fields.put("counts", counts);
        fields.put("def", def);
        fields.put("intervalNumber", intervalNumber);
        fields.put("period", period);
        fields.put("totalEventCount", totalEventCount);
        fields.put("values", values);
        oos.writeFields();
        }

    /**Deserialise: use constructor for validation, defensive copying, conversion from old formats, etc.
     * Also allows us to intern() keys, etc, to conserve memory.
     */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // If we have old format of object,
        // ie without the new values/counts but with the internal info map,
        // then convert to the new format now.
        if(values == null) // Assume that this implies counts == null and info != null.
            {
            // Convert info data to sortable form...
            final List<EventVariableValueBuffer.Count> cs = new ArrayList<EventVariableValueBuffer.Count>(info.size());
            for(final Object key : info.keySet())
                {
                final EventVariableValueBuffer.Count count = new EventVariableValueBuffer.Count(key);
                count.count = info.get(key).count;
                cs.add(count);
                }
            // Get counts ordered to get rankings...
            Collections.sort(cs);
            // Should be highest-count first...
            assert((cs.size() < 2) || (cs.get(0).count >= cs.get(1).count)) : "array out of order";

            // Create list of values and counts...
            final Object newValues[] = new Object[cs.size()];
            final int newCounts[] = new int[cs.size()];
            for(int i = cs.size(); --i >= 0; )
                {
                final EventVariableValueBuffer.Count count = cs.get(i);
                newValues[i] = count.value;
                newCounts[i] = count.count;
                }

            // Create instance from reconstituted data...
            return(new EventVariableValue(authoratitive, def, period, intervalNumber, totalEventCount,
                                          newValues, newCounts));
            }

        // Reconstruct using new format of object (using values/count).
        return(new EventVariableValue(authoratitive, def, period, intervalNumber, totalEventCount, values, counts));
        }

    /**Check that the object state is consistent and legal. */
    public void validateObject()
        throws InvalidObjectException
        {
        if(def == null)
            { throw new InvalidObjectException("bad object: def must not be null"); }
        if(!def.isEvent())
            { throw new InvalidObjectException("bad object: def not an event type"); }

        if(period == null)
            { throw new InvalidObjectException("bad object: period must not be null"); }

        if(totalEventCount < 0)
            { throw new InvalidObjectException("bad object: negative totalEventCount"); }

        if(intervalNumber < 0)
            { throw new InvalidObjectException("bad object: negative intervalNumber"); }
        if(Long.MAX_VALUE / period.getIntervalMs() < intervalNumber)
            { throw new InvalidObjectException("bad object: intervalNumber too large"); }


        // Check counts and values are valid.
        if((counts == null) || (values == null))
            { throw new InvalidObjectException("bad object: counts[] or values[] is null"); }
        final int diffVals = values.length;
        if(counts.length != diffVals)
            { throw new InvalidObjectException("bad object: counts.length != values.length"); }
        if((diffVals > totalEventCount) || (diffVals > def.getMaxDiffEventCount()))
            { throw new InvalidObjectException("bad object: too many distinct event values"); }
        // Check that counts are non-negative and non-increasing
        // and that the sum is not too large.
        int countsSum = 0;
        for(int i = diffVals; --i >= 0; )
            {
            final int c = counts[i];
            if(c < 0)
                { throw new InvalidObjectException("bad object: negative count"); }
            countsSum += c;
            if((i > 0) && (c > counts[i-1]))
                { throw new InvalidObjectException("bad object: counts increasing"); }
            }
        if(countsSum > totalEventCount)
            { throw new InvalidObjectException("bad object: counts[] sum > totalEventCount"); }
        // Check that values are non-null and type-check.
        for(final Object o : values)
            {
            if(o == null)
                { throw new InvalidObjectException("bad object: values[] entry is null"); }
            if(!def.checkType(o))
                { throw new InvalidObjectException("bad object: values[] entry has invalid type"); }
            }


        // Make very expensive test selectively enablable at run-time.
        assert(new TreeSet<Object>(Arrays.asList(values)).size() == diffVals) : "bad object: values[] contains duplicate(s)";


        // Check info Map contents are all legitimate (no null, type-unsafe keys or values).
        if(info == null)
            { throw new InvalidObjectException("bad object: info must not be null"); }
        // Check that we don't have too many distinct entries.
        if(info.size() > def.getMaxDiffEventCount())
            { throw new InvalidObjectException("bad object: info too large"); }
        if(info.size() != diffVals)
            { throw new InvalidObjectException("bad object: info.size() != values.length"); }

        // Double-check internally-constructed data only in debug mode.
        if(IsDebug.isDebug)
            {
            // Check keys and values are OK and consistent with counts[] and values[].
            for(final Object k : info.keySet())
                {
                if((k == null) || !def.checkType(k))
                    { throw new InvalidObjectException("bad object: info map key null or invalid type"); }
    //            if(!allVals.contains(k))
    //                { throw new InvalidObjectException("bad object: info map key not in values[]"); }
                if(!(info.get(k) instanceof ValueInfo))
                    { throw new InvalidObjectException("bad object: info map value null or invalid type"); }
                }
            // Check that info values are at least in range.
            for(final ValueInfo v : info.values())
                {
                if(v.count < 1)
                    { throw new InvalidObjectException("bad object: info map count invalid"); }
                if((v.rank < 0) || (v.rank >= info.size()))
                    { throw new InvalidObjectException("bad object: info map rank invalid"); }
                }
            // Check that totalCount is not smaller than the number of different entries.
            if(totalEventCount < info.size())
                { throw new InvalidObjectException("bad object: totalCount lower than number of different values"); }
            // Check that totalCount is not smaller than the total of all entries.
            int calcTotalCount = 0;
            for(final ValueInfo v : info.values())
                { calcTotalCount += v.count; }
            if(totalEventCount < calcTotalCount)
                { throw new InvalidObjectException("bad object: totalCount lower than sum of counts"); }
            // Check ranking order is correct/consistent.
            for(int i = diffVals; --i >= 0; )
                {
                // Look up value in map to check count is correct.
                final Object o = values[i];
                final int c = counts[i];
                final ValueInfo vi = info.get(o);
                if(vi.count != c)
                    { throw new InvalidObjectException("bad object: info/counts don't match"); }
                // Verify rank.
                if(vi.rank != i)
                    {
                    // If rank is not exactly equal to position in list
                    // then we have to check it.

                    // If the item at the claimed rank has the same count as us
                    // then we'll accept this.
                    if(counts[vi.rank] != c)
                        { throw new InvalidObjectException("bad object: invalid rank"); }
                    }
                }
            }
        }

    /**Unique Serialisation class ID. */
    private static final long serialVersionUID = 5094323684152972705L;
    }
