/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.Rnd;

import java.io.*;
import java.util.*;


/**Mutable value of a local or global system event variable.
 * This is the mutable version of EventVariableValue
 * as StringBuilder is to String.
 * <p>
 * The current stats should be collected in instances of this,
 * then converted to EventVariableValue for longer-term storage
 * and/or to return to users.
 * <p>
 * This is designed to be efficient and quick to mutate.
 * <p>
 * This class is thread-safe, and methods may be synchronized.
 * <p>
 * This attempts to conserve memory,
 * at the cost of some CPU time,
 * by attempting to effectively intern() event values.
 * This therefore assumes that all event values are sensibly intern()able,
 * eg immutable and with equals() and hashCode() methods that cover all visible state.
  */
public final class EventVariableValueBuffer implements Serializable,
                                                       ObjectInputValidation
    {
    /**Construct a single variable value.
     * This type-checks the value, vetoing construction if invalid.
     *
     * @param def  the variable definition; never null and must be an event
     * @param period  the event period for this sample; never null
     * @param intervalNumber  the period number, zero means "all";
     *     non-negative
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public EventVariableValueBuffer(final SimpleVariableDefinition def,
                                    final EventPeriod period,
                                    final long intervalNumber)
        throws IllegalArgumentException
        {
        this.def = def;
        this.period = period;
        this.intervalNumber = intervalNumber;

        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Private cache for toEventVariableValue().
     * Cleared on any change to the buffer, ie a new event or a merge.
     * <p>
     * Avoids having to expensively recreate a new evv when nothing has changed.
     * <p>
     * We do not serialise/persist this redundant state.
     */
    private transient EventVariableValue evvCache;

    /**Convert to (immutable) non-authoritative EventVariableValue.
     * If the cached value is non-null then we return that,
     * else we compute it, cache it, and return it.
     */
    public synchronized EventVariableValue toEventVariableValue()
        {
        if(evvCache != null) { return(evvCache); }

        // Get counts ordered to get rankings...
        final List<Count> cs = new ArrayList<Count>(counts.values());
        Collections.sort(cs);
        // Should be highest-count first...
        assert((cs.size() < 2) || (cs.get(0).count >= cs.get(1).count)) : "array out of order";

        // Create list of values and counts...
        final Object values[] = new Object[cs.size()];
        final int counts[] = new int[cs.size()];
        for(int i = cs.size(); --i >= 0; )
            {
            final Count count = cs.get(i);
            values[i] = count.value;
            counts[i] = count.count;
            }

        // Now make the non-authoritative immutable EventVariableValue...
        // We can skip expensively re-intern()ing all the values
        // since we have already done that in this class.
        final EventVariableValue evv = new EventVariableValue(false, def, period, intervalNumber, totalEventCount, values, counts, false);
        evvCache = evv; // Cache it...
        return(evv);
        }


    /**The (immutable) variable definition; never null. */
    private final SimpleVariableDefinition def;

    /**Get the variable definition; never null. */
    public SimpleVariableDefinition getDef() { return(def); }

    /**The (immutable) event period; never null. */
    private final EventPeriod period;

    /**Get the event period; never null. */
    public EventPeriod getPeriod() { return(period); }

    /**The (immutable) interval number; strictly positive. */
    private final long intervalNumber;

    /**Get the interval number; strictly positive. */
    public long getIntervalNumber() { return(intervalNumber); }


    /**Total event count; non-negative.
     * Initially zero.
     * <p>
     * Mutable; accessed under the instance lock.
     */
    private int totalEventCount;

    /**Get the total event count; non-negative. */
    public int getTotalEventCount() { return(totalEventCount); }

    /**Generate human-readable summary of state. */
    public String toString()
        {
        return(def.getName() + ':' + totalEventCount);
        }



    /**Private class to contain an int count value.
     * Contains the count of a paritcular value.
     * <p>
     * Is mutable to avoid unnecessary object churn in the heap.
     * <p>
     * When created has a count of 1.
     * <p>
     * Not thread-safe (access should be protected by other means).
     * <p>
     * Made available to EventVariableValue to help in canonicalising state.
     */
    static final class Count implements Comparable<Count>,
                                        Serializable,
                                        ObjectInputValidation
        {
        /**Construct an instance with the given (non-null) event value. */
        Count(final Object o)
            {
            value = o;
            try { validateObject(); } catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }
        /**Provide total ordering with largest count (lowest rank) first. */
        public int compareTo(final Count c)
            {
            if(count < c.count) { return(+1); }
            if(count > c.count) { return(-1); }
            return(0);
            }
        /**Value (should be immutable); non-null. */
        final Object value;
        /**Count value; should be strictly positive. */
        int count = 1;
        /**Check that the object state is consistent and legal. */
        public void validateObject() throws InvalidObjectException
            {
            if(value == null) { throw new InvalidObjectException("bad object: value is null"); }
            if(count <= 0) { throw new InvalidObjectException("bad object: count not positive"); }
            }
        /**Deserialise: validate and then intern() value to economise on memory.
         */
        protected Object readResolve()
            throws ObjectStreamException
            {
            validateObject();
            final Count c = new Count(MemoryTools.intern(value));
            c.count = count;
            return(c);
            }
        /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
        private static final long serialVersionUID = 187928939275973214L;
        };


    /**Map from event value to count of events of that value.
     * Initially empty.
     * <p>
     * Not thread-safe (access should be protected by other means).
     */
    private /* final */ Map<Object, Count> counts = new HashMap<Object, Count>();


    /**Record/add event to the tally.
     * Note that the event must match the definition that this buffer is constructed with
     * or an IllegalArgumentException will be thrown.
     * <p>
     * The total event count is incremented,
     * and if appropriate, the value and/or count of occurrences of that value,
     * will be incremented.
     * <p>
     * Note that null values are not recorded in the map,
     * but the totalCount is still incremented.
     * <p>
     * If the totalEventCount would be about to overflow,
     * then we halve it and halve the count of all extant entries
     * (removing any that become zero).
     *
     * @param svv  event to be recorded; must be non-null and of event type
     */
    public synchronized void addEvent(final SimpleVariableValue svv)
        {
        if((svv == null) || !def.equals(svv.getDef()))
            { throw new IllegalArgumentException(); }

        // Clear any evv cache since we are changing the buffer state.
        evvCache = null;

        // Avoid overflow by scaling all existing values (halving them).
        if(totalEventCount == Integer.MAX_VALUE)
            {
            for(final Iterator it = counts.keySet().iterator(); it.hasNext(); )
                {
                final Object key = it.next();
                final Count c = counts.get(key);
                // If the scaled count drops to zero then remove the entry.
                if((c.count /= 2) < 1)
                    { it.remove(); }
                }

            totalEventCount /= 2;
            }

        // Count the event.
        ++totalEventCount;

        final Object value = svv.getValue();

        // Void-valued events are not further counted.
        if(value == null)
            { return; }

        Count c = counts.get(value);
        // Create new entry (with count==1) if needed.
        if(c == null)
            {
            // Don't expand the table if already at/beyond the limit.
            // We are prepared to evict an existing item at random
            // and use its space.
            // The probability of eviction is inversely proportional
            // to its count to allow new values to get into a full table.
            final int maxDiffEventCount = def.getMaxDiffEventCount();
            if(counts.size() >= maxDiffEventCount)
                {
                // We amortise the cost of finding an evictee to be about
                // constant regardless of table size.
                // So we may give up and reject this putative new entry.
                // (If we don't keep value counts at all, we must give up too!)
                if((maxDiffEventCount < 1) || (Rnd.fastRnd.nextInt(maxDiffEventCount) != 0))
                    { return; }

                // OK, look for a victim...
                final List<Count> candiatesToEvict = new ArrayList<Count>();
                for(final Count cc : counts.values())
                    {
                    // Even high-count entries can ejected,
                    // but only at low priority,
                    // so that novel event values can enter a full table.
                    if((cc.count < 2) || (Rnd.fastRnd.nextInt(cc.count) == 0))
                        { candiatesToEvict.add(cc); }
                    }
                final int cs = candiatesToEvict.size();
                // No candidate victims to remove,
                // so we really will have to ignore the new entry!
                if(cs < 1) { return; }

                // There were some candidates,
                // so pick one at random,
                // and remove it from the table.
                // We now have space for our new event.
                final int toEvict = Rnd.fastRnd.nextInt(cs);
                final Object vToEvict = candiatesToEvict.get(toEvict).value;
                counts.remove(vToEvict);
                }

            // De-dup (new) values to economise on memory
            // if it is likely to be worthwhile long-term.
            c = new Count(EventVariableValue.shouldIntern(def) ? MemoryTools.intern(value) : value);
            counts.put(c.value, c);
            }
        // Else increment existing count.
        else
            { ++c.count; }
        }

    /**If true, replace our data rather than merge on update(). */
    private static final boolean REPLACE_ON_UPDATE = true;

    /**Update/merge the value with upstream data.
     * This is taken to be a set of (usually non-authoratative)
     * values from upstream which possible contains the events from
     * all (or more) participants in the distributed system.
     * <p>
     * We compute an approximate merge,
     * which is assumed to be OK to do at all but the master node since
     * downstream nodes will eventually get overridden with
     * correct authoritative values from upstream.
     * <p>
     * The approximation consists of increasing the event count to
     * the maximum found in this item and the supplied value,
     * and then adjusting the total event count suitably
     * (allowing for events not stored as values in the map)
     * This may result in a total event count higher than
     * either this or the upstream data.
     * <p>
     * This should <em>not</em> be used at the very top-level master store
     * to avoid double-counting values, etc.
     * <p>
     * This may be quite slow.
     */
    public synchronized void update(final EventVariableValue upstream)
        {
        // Clear any evv cache since we are changing the buffer state.
        evvCache = null;

        // If there is any danger at all of us exceeding the maximum
        // possible number of different entries on merge,
        // then simply replace our state with that from upstream.
        if(REPLACE_ON_UPDATE ||
           (upstream.getTotalDistinctValues() + counts.size() > def.getMaxDiffEventCount()))
            {
            counts.clear();
            totalEventCount = upstream.getTotalEventCount();
            final Map<Object, EventVariableValue.ValueInfo> m = upstream.getInfo();
            final boolean shouldIntern = EventVariableValue.shouldIntern(def);
            for(final Object value : m.keySet())
                {
                final EventVariableValue.ValueInfo vi = m.get(value);
                final Count c = new Count(shouldIntern ? MemoryTools.intern(value) : value);
                c.count = vi.count;
                counts.put(c.value, c);
                }
            }

        else // Do a merge...
            {
            // For all upstream values,
            // make sure that the count is adjusted upwards
            // to the higher of the upstream and extant values,
            // adjusting totalEventCount too if needed.
            final Map<Object, EventVariableValue.ValueInfo> m = upstream.getInfo();
            final boolean shouldIntern = EventVariableValue.shouldIntern(def);
            for(final Object value : m.keySet())
                {
                final EventVariableValue.ValueInfo vi = m.get(value);
                assert(vi.count > 0);
                Count c = counts.get(value);
                // If no record for this value then make a zero-count entry now.
                if(c == null)
                    {
                    c = new Count(shouldIntern ? MemoryTools.intern(value) : value);
                    c.count = vi.count;
                    totalEventCount += vi.count;
                    counts.put(c.value, c);
                    continue;
                    }
                // If the upstream count is greater then adjust our records.
                else if(vi.count > c.count)
                    {
                    totalEventCount += (vi.count - c.count);
                    c.count = vi.count; // Adjust extant record in place.
                    }
                }

            // Ensure that the total count reflects the highest available
            // to allow for events whose values have not been captured.
            final int upstreamTotalEventCount = upstream.getTotalEventCount();
            if(upstreamTotalEventCount > totalEventCount)
                { totalEventCount = upstreamTotalEventCount; }
            if(counts.size() > totalEventCount)
                { totalEventCount = counts.size(); }
            }

        // Double-check state...
        if(IsDebug.isDebug)
            {
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new Error(); }
            }
        }



    /**Equality is based on the definition, period and interval number, but not the event values.
     */
    public boolean equals(final Object obj)
        {
        if(!(obj instanceof EventVariableValueBuffer)) { return(false); }
        final EventVariableValueBuffer o = (EventVariableValueBuffer) obj;

        if(!def.equals(o.def)) { return(false); }
        if(period != o.period) { return(false); }
        if(intervalNumber != o.intervalNumber) { return(false); }

        // All relevant fields seem to match, so objects are equal.
        return(true);
        }

    /**The hash is built on a subset of the fields.
     * This implementation uses the hash of the definition and the interval number and period values.
     */
    public int hashCode()
        {
        return(def.hashCode() ^ ((int)intervalNumber) ^ period.getIntervalMs());
        }

    /**Deserialise.
     */
    private void readObject(final ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        ois.defaultReadObject();

        // Take a defensive copy of counts,
        // and if it is too large, trim it if at all possible,
        // else clear it completely.
        final List<Count> newCounts = new ArrayList<Count>(counts.values());
        final int diffValues = newCounts.size();
        counts = new HashMap<Object,Count>(1 + 2 * diffValues);
        final boolean tooLarge = diffValues > def.getMaxDiffEventCount();
if(tooLarge) { System.err.println("EventVariableValueBuffer.readObject(): WARNING: coercing value for "+def+" by trimming from "+diffValues+" different values."); }
        int total = 0; // Recompute totalCount.
        final boolean shouldIntern = EventVariableValue.shouldIntern(def);
        for(final Count rc : newCounts)
            {
            final int cc = rc.count;
            if(cc < 1) { throw new IOException(); }

            total += cc;

            // If currently too large, simply drop all single-count entries
            // to try to slim us down a little.
            if(tooLarge && (cc < 2)) { continue; }

            // Make a safe copy of the count data
            // with the (key) value intern()ed.
            final Count c = new Count(shouldIntern ? MemoryTools.intern(rc.value) : rc.value);
            c.count = cc;
            counts.put(c.value, c);
            }
        // If still too large then clear all values.
        if(counts.size() > def.getMaxDiffEventCount())
            { counts.clear(); }

        // Coerce totalEventCount to be sensible.
        if(total > totalEventCount)
            { totalEventCount = total; }

        // Check that object is now sensible.
        validateObject();
        }

    /**Check that the object state is consistent and legal. */
    public synchronized void validateObject()
        throws InvalidObjectException
        {
        if(def == null)
            { throw new InvalidObjectException("bad object: def must not be null"); }
        if(!def.isEvent())
            { throw new InvalidObjectException("bad object: def not an event type"); }

        if(period == null)
            { throw new InvalidObjectException("bad object: period must not be null"); }

        if(totalEventCount < 0)
            { throw new InvalidObjectException("bad object: negative totalEventCount"); }

        if(intervalNumber < 0)
            { throw new InvalidObjectException("bad object: negative intervalNumber"); }
        if(Long.MAX_VALUE / period.getIntervalMs() < intervalNumber)
            { throw new InvalidObjectException("bad object: intervalNumber too large"); }

        // Validate map.
        if(counts == null)
            { throw new InvalidObjectException("bad object: counts must not be null"); }
        // Check that we don't have too many distinct entries.
        if(counts.size() > def.getMaxDiffEventCount())
            { throw new InvalidObjectException("bad object: counts too large"); }
        // Check all keys are valid (correct type and not null).
        for(final Object k : counts.keySet())
            {
            if((k == null) || !def.checkType(k))
                { throw new InvalidObjectException("bad object: invalid key in counts"); }
            if(!k.equals(counts.get(k).value))
                { throw new InvalidObjectException("bad object: mismatched key and count value"); }
            }
        // Check all counts are valid (positive and not null).
        for(final Count c : counts.values())
            {
            if(c == null)
                { throw new InvalidObjectException("bad object: null count in counts"); }
            c.validateObject();
            }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -1721389511619513375L;
    }
