/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import ORG.hd.d.IsDebug;

import java.io.*;
import java.util.Arrays;
import java.util.BitSet;


/**Current and historical entries for one event forming one "row".
 * This is mutable and used to hold live event stats in memory.
 * <p>
 * This maintains state incrementally on each new event, for example,
 * shuffling out and discarding old historical even values if need be.
 * <p>
 * Thread-safe, with methods synchronized as necessary.
 * <p>
 * Serializable for direct persistence if required,
 * though an XML format rather than a binary one would probably be more robust.
 */
public final class EventVariableValuePeriodRow implements Serializable,
                                                          ObjectInputValidation
    {
    /**Construct a single event variable value period "row".
     *
     * @param def  the variable definition; never null and must be an event
     * @param period  the event period for this sample; never null
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public EventVariableValuePeriodRow(final SimpleVariableDefinition def,
                                       final EventPeriod period)
        throws IllegalArgumentException
        {
        this.def = def;
        this.period = period;

        current = new EventVariableValueBuffer(def, period, period.getIntervalNumber(System.currentTimeMillis()));
        all = new EventVariableValueBuffer(def, period, 0);

        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }



    /**The variable definition; never null. */
    private final SimpleVariableDefinition def;

    /**Get the variable definition; never null. */
    public SimpleVariableDefinition getDef() { return(def); }

    /**The event period; never null. */
    private final EventPeriod period;

    /**Get the event period; never null. */
    public EventPeriod getPeriod() { return(period); }



    /**All events being collected, never null.
     * Captures all events ever seen,
     * but may be rescaled to prevent overflow,
     * so should simply be regarded as a reasonably representative sub-sample.
     * <p>
     * Will never be authoritative.
     */
    private /* final */ EventVariableValueBuffer all;

    /**"Current" events being collected, never null.
     * Will be replaced with new instance as the current one stops being current.
     * <p>
     * Initially created with value suitable to receive
     * a new event immediately.
     * <p>
     * Will never be authoritative.
     */
    private EventVariableValueBuffer current;

    /**Set of "historical" periods, most-recent at index 0, possibly containing nulls; never null.
     * Initially all nulls.
     * <p>
     * The intervalNumber of any entry at index i
     * must be current.getIntervalNumber()-1-i.
     * <p>
     * FIXME: copy defensively during deserialisation.
     */
    private final EventVariableValue historical[] =
            new EventVariableValue[SystemVariables.EVENT_SAMPLES_RETAINED];



    /**Record/add event to the tally for this period.
     * Note that the event must match the definition that this instance is constructed with
     * or an IllegalArgumentException will be thrown.
     * <p>
     * If the "current" event buffer is too old to accept this incoming event,
     * we move all the historical data up and insert an immutable copy of
     * the buffered data as appropriate
     * (though we must cope with the case where we have to move data up several
     * slots due to an extended interval between updates).
     *
     * @param svv  event to be recorded; must be non-null and of event type
     */
    public synchronized void addEvent(final SimpleVariableValue svv)
        {
        if((svv == null) || !def.equals(svv.getDef()))
            { throw new IllegalArgumentException(); }

        // Get to normalised state if we need to...
        _canonicaliseState();

        // Add event to current list.
        current.addEvent(svv);

        // Add event to "all" event bucket.
        all.addEvent(svv);
        }

    /**Canonicalise our internal state, so that the current event collector covers "now".
     * If current event collector is too old,
     * then create a new one and age the historical data.
     * If the clock seems to have gone backwards for whatever reason,
     * we'll simply carry on collecting as we were until things get sane.
     */
    private void _canonicaliseState()
        {
//        // State validation...
//        try { validateObject();  } catch(InvalidObjectException e) { throw new IllegalStateException(e); }

        final long now = System.currentTimeMillis();
        final long calcCurrentIntervalNumber = period.getIntervalNumber(now);

        if(current.getIntervalNumber() < calcCurrentIntervalNumber)
            {
//            System.out.println("DOING SHIFT UP...");
            // Need to shift up data...
            final EventVariableValueBuffer oldCurrent = current;
            final EventVariableValueBuffer newCurrent = new EventVariableValueBuffer(def, period, calcCurrentIntervalNumber);
            current = newCurrent;

            // Compute the number of slots that we need to shift data up by,
            // usually 1, but may be more for rare events or after deserialisation.
            final long toShiftBy = newCurrent.getIntervalNumber() - oldCurrent.getIntervalNumber();
//            System.out.println("SHIFTING BY: " + toShiftBy);
            assert(toShiftBy > 0);

            // If a really huge shift,
            // eg because something has been restored after a long period on disc,
            // or a really rare event has just occurred,
            // then just throw away all the old data by clearing the history
            // and not even converting the oldCurrent to a historical value.
            if(toShiftBy > historical.length)
                { Arrays.fill(historical, null); }
            else
                {
                // Copy up all existing elements
                // to higher indexes (ie age them)...
                // This should correctly move nothing if toShiftBy == historical.length.
                System.arraycopy(historical, 0, historical, (int) toShiftBy, (int) (historical.length - toShiftBy));

                // Convert the oldCurrent to immutable form
                // and insert into the appropriate slot...
                final int oldest = (int) (toShiftBy - 1);
                historical[oldest] = oldCurrent.toEventVariableValue();

                // Null out any lower (newer) slots that we now have no data for.
                if(toShiftBy > 1)
                    { Arrays.fill(historical, 0, oldest, null); }

                assert(historical[oldest] != null); // We now have at least one historical entry...
                }

//System.out.println("EVVPR current/historical[] = " + current + "/" + Arrays.asList(historical));
            }

//        // State validation...
//        try { validateObject();  } catch(InvalidObjectException e) { throw new IllegalStateException(e); }
        }


    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     *
     * @param firstIntervalNumber  the number of the first interval
     *     for which data is potentially required;
     *     if too far in the past or future then possibly no data
     *     will be available;
     *     using zero accesses the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    synchronized EventVariableValue[] getEventValues(final long firstIntervalNumber,
                                                     final BitSet whichValues)
        {
        if(firstIntervalNumber < 0)
            { throw new IllegalArgumentException(); }

        final boolean nullWV = (whichValues == null);

        // If firstIntervalNumber == 0, then exactly slot zero must be requested.
        if(firstIntervalNumber == 0)
            {
            if(!nullWV && (!whichValues.get(0) || (whichValues.length() != 1)))
                { throw new IllegalArgumentException("request for 'all' bucket must be for exactly that bucket and no others"); }

            // Convert "all" bucket to correct form and return it...
            return(new EventVariableValue[]{ all.toEventVariableValue() });
            }

//System.out.println("[EventVariableValuePeriodRow.getEventValues("+firstIntervalNumber+", "+whichValues+")]");

//        // State validation...
//        try { validateObject();  } catch(InvalidObjectException e) { throw new IllegalStateException(e); }

        // If requested result size silly/huge, then give up now to save space, etc.
        final int resultSize = nullWV ? 1 : whichValues.length();
        final long currentIntervalNumber = current.getIntervalNumber();

        // Optimisations for where the request will definitely generate no results...
        if(resultSize < 1)
            { return(_NO_EVENT_VALUES); }
        if(resultSize > 1 + 2*SystemVariables.EVENT_SAMPLES_RETAINED)
            { return(_NO_EVENT_VALUES); }
        // If newest interval requested is too much newer than the "current" interval,
        // then give up now.
        if(firstIntervalNumber - resultSize > currentIntervalNumber)
            { return(_NO_EVENT_VALUES); }
        // If newest interval requested is too far in the past, then give up now.
        final long oldestPossibleEventNumber = currentIntervalNumber - historical.length;
        if(firstIntervalNumber < oldestPossibleEventNumber)
            { return(_NO_EVENT_VALUES); }

        // Make result...
        // Don't attempt to optimise its length here, yet...
        final EventVariableValue result[] = new EventVariableValue[resultSize];

        // Index in historical[index] corresponding to result[0].
        final int offset = (int) ((currentIntervalNumber-1) - firstIntervalNumber);

//System.out.println("EVVPR offset = " + offset);
//System.out.println("EVVPR current/historical[] = " + current + "/" + Arrays.asList(historical));

        // Walk the request list, filling in values where we have them.
        // The "current" item corresponds to index -1 in historical[],
        // and needs to be converted before being slotted into the result, if used.
        // Note that this value will never count as "authoritative" since not closed.
        // Remember: higher "i" means older (smaller intervalNumber) values.
        BitSet wv = whichValues;
        if(wv == null)
            {
            // Create synthetic BitSet for simplicity below.
            wv = new BitSet(1);
            wv.set(0);
            }
        for(int i = wv.nextSetBit(0); i >= 0; i = wv.nextSetBit(i+1))
            {
            final int hIndex = i + offset;

            if(hIndex == -1) // Current...
                { result[i] = current.toEventVariableValue(); }
            else if((hIndex >= 0) && (hIndex < historical.length))
                { result[i] = historical[hIndex]; }
            }

//        // Extra validation of output...
//        assert(result != null) : "unexpected null return value";
//        for(int i = result.length; --i >= 0; )
//            {
//            final EventVariableValue evv = result[i];
//            if(evv == null) { continue; }
//            assert(evv.getIntervalNumber() == firstIntervalNumber - i) : "unexpected return value with wrong interval number";
//            }

//        // State validation...
//        try { validateObject();  } catch(InvalidObjectException e) { throw new IllegalStateException(e); }

//System.out.println("EVVPR result[] = " + Arrays.asList(result));

        return(result);
        }

    /**Zero-length array corresponding to "no results", immutable so made available to the package; non-null. */
    static final EventVariableValue[] _NO_EVENT_VALUES = new EventVariableValue[0];


    /**Set/override the specified event sets for the specified intervals; never null.
     * This is used to set values that have been fetched from upstream
     * (ie closer to the master data store).
     * <p>
     * This may ignore the call,
     * and certainly will if the intervalNumber is outside the current time window it holds.
     * <p>
     * This will generally accept the value and store it if:
     * <ul>
     * <li>The intervalNumber is within the current window.
     * <li>The intervalSelector is a period for which this event has data.
     * <li>The supplied value is "better" than any current one held, ie:
     *     <ul>
     *     <li>We hold a null and evv is non-null.
     *     <li>We hold a non-authoritative value and evv is authoritative.
     *     <li>The authoritative-ness is the same bu we hold an entry
     *         with a lower total count than evv.
     *     <ul>
     *     </ul>
     * </ul>
     *
     * @param evv  non-null event value
     */
    public synchronized void setEventValue(final EventVariableValue evv)
        {
        if(evv == null)
            { throw new IllegalArgumentException(); }
        final SimpleVariableDefinition def = evv.getDef();
        if(!def.equals(this.def))
            { throw new IllegalArgumentException(); }
        if(!evv.getPeriod().equals(period))
            { throw new IllegalArgumentException(); }

        // Make sure that our state is canonicalised.
        _canonicaliseState();

        // Event of "current" set, newer (greater) than any we can set.
        final long currentIntervalNumber = current.getIntervalNumber();

        // Interval number from our new data.
        final long updateIntervalNumber = evv.getIntervalNumber();

        // Note that we may accept an override for the all/current periods,
        // (but we don't really expect it to be marked as authoritative)
        // by folding in events from upstream not seen here.
        if(updateIntervalNumber == 0)
            {
            // Accept an update if we seem not to have seen all events.
            if(all.getTotalEventCount() < evv.getTotalEventCount())
                {
if(IsDebug.isDebug) { System.out.println("[EVVPR.setEventValue("+evv+"): accepting all-period override (isAuth="+evv.isAuthoritative()+").]"); }
                all.update(evv);
                }
            return;
            }
        else if(updateIntervalNumber == currentIntervalNumber)
            {
            // Accept an update if we seem not to have seen all events.
            if(current.getTotalEventCount() < evv.getTotalEventCount())
                {
if(IsDebug.isDebug) { System.out.println("[EVVPR.setEventValue("+evv+"): accepting current-period override (isAuth="+evv.isAuthoritative()+") at period "+updateIntervalNumber+".]"); }
                current.update(evv);
                }
            return;
            }
        // Too new for us to accept the update.
        else if(updateIntervalNumber > currentIntervalNumber) { return; }

        final long oldestPossibleEventNumber = currentIntervalNumber - historical.length;
        // Too old for us to accept the update.
        if(updateIntervalNumber < oldestPossibleEventNumber) { return; }

        // Index in historical[index] corresponding to new value we have been passed.
        final int offset = (int) ((currentIntervalNumber-1) - updateIntervalNumber);

        // If the new data is "better" in some way
        // then accept it in place of what we have already.
        //
        // Better is any of:
        //   * We have now old value at all.
        //   * We have a non-auth value and the new data is authoritative.
        //   * We have a non-auth value and the new data has more events
        //     (and this is dealt with specially for the current interval).
        final EventVariableValue extant = historical[offset];
        if((extant == null) ||
           (!extant.isAuthoritative() && evv.isAuthoritative()) ||
           (extant.getTotalEventCount() < evv.getTotalEventCount()))
            {
//if(IsDebug.isDebug) { System.out.println("EVVPR.setEventValue("+evv+"): accepting override (isAuth="+evv.isAuthoritative()+") at period "+updateIntervalNumber+" (offset="+offset+")"); }
            historical[offset] = evv;
            }

        assert((extant == null) || (extant.getIntervalNumber() == updateIntervalNumber));

//        try { validateObject(); } catch(InvalidObjectException e) { e.printStackTrace();  }
        }

    /**Deserialise.
     */
    private void readObject(ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        ois.defaultReadObject();

        // If deserialising from an old object without an "all" field,
        // create it empty now.
        if(all == null) { all = new EventVariableValueBuffer(def, period, 0); }

        // Validate the state.
        validateObject();
        }



    /**Check that the object state is consistent and legal.
     * Is synchronized to prevent changes of state while this instance is examined.
     *
     * @throws InvalidObjectException
     */
    public synchronized void validateObject()
        throws InvalidObjectException
        {
        if(def == null)
            { throw new InvalidObjectException("bad object: def must not be null"); }
        if(!def.isEvent())
            { throw new InvalidObjectException("bad object: def not an event type"); }

        if(period == null)
            { throw new InvalidObjectException("bad object: period must not be null"); }

        if(current == null)
            { throw new InvalidObjectException("bad object: current must not be null"); }
        if((current.getPeriod() != period) || !current.getDef().equals(def))
            { throw new InvalidObjectException("bad object: current def/period incorrect"); }
        if(current.getIntervalNumber() <= 0)
            { throw new InvalidObjectException("bad object: current interval invalid"); }

        if(all == null)
            { throw new InvalidObjectException("bad object: all must not be null"); }
        if((all.getPeriod() != period) || !all.getDef().equals(def))
            { throw new InvalidObjectException("bad object: all def/period incorrect"); }
        if(all.getIntervalNumber() != 0)
            { throw new InvalidObjectException("bad object: all interval non-zero"); }

        if(all == current)
            { throw new InvalidObjectException("bad object: all == current"); }

        if(historical == null)
            { throw new InvalidObjectException("bad object: historical must not be null"); }
        if(historical.length != SystemVariables.EVENT_SAMPLES_RETAINED)
            { throw new InvalidObjectException("bad object: historical wrong length"); }
        for(int i = historical.length; --i >= 0; )
            {
            final EventVariableValue evv = historical[i];
            if(evv != null)
                {
                if((evv.getPeriod() != period) || !evv.getDef().equals(def))
                    { throw new InvalidObjectException("bad object: historical def/period incorrect"); }
                final long expectedIntervalNumber = current.getIntervalNumber() - 1 - i;
                if(evv.getIntervalNumber() != expectedIntervalNumber)
                    { throw new InvalidObjectException("bad object: historical intervalNumber incorrect"); }
                }
            }
        }

    /**Unique Serialisation class ID. */
    private static final long serialVersionUID = -614477516323489097L;
    }
