/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.Serializable;
import java.util.BitSet;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Set;


/**Current and historical entries for all periods/intervals one event forming one "set".
 * This is mutable and used to hold live event stats in memory.
 * <p>
 * This maintains state incrementally on each new event, for example,
 * shuffling out and discarding old historical event values if need be.
 * <p>
 * Thread-safe.
 * <p>
 * Mutator calls are synchronized,
 * so that serialisation can take place on an instance
 * providing that a lock is held on that instance.
 * <p>
 * Serializable for direct persistence if required,
 * though an XML format rather than a binary one would probably be more robust.
 * <p>
 * Notes the last time an event was add()ed or a whole slot value was "set"
 * so as to allow an observer to tell quickly when this instance may need a save.
 */
public final class EventVariableValueSet implements Serializable,
                                                    ObjectInputValidation
    {
    /**Construct a single event variable value set (for all periods/intervals).
     *
     * @param def  the variable definition; never null and must be an event
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    public EventVariableValueSet(final SimpleVariableDefinition def)
        throws IllegalArgumentException
        {
        this.def = def;

        // Create an entry for each available period
        // to a "period row" for that period,
        // unless a subset has been requested in which case just do those specified.
        final Set<EventPeriod> evPeriodSubset = def.getEvPeriodSubset();
        // If no subset is specified, then use the full set of periods/intervals.
        final Set<EventPeriod> _evPeriodSubset = (evPeriodSubset != null) ? evPeriodSubset :
                                                     EnumSet.allOf(EventPeriod.class);
        byPeriod = new EnumMap<EventPeriod, EventVariableValuePeriodRow>(EventPeriod.class);
        for(final EventPeriod ep : _evPeriodSubset)
            { byPeriod.put(ep, new EventVariableValuePeriodRow(def, ep)); }

        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }



    /**The variable definition; never null. */
    private final SimpleVariableDefinition def;

    /**Get the variable definition; never null. */
    public SimpleVariableDefinition getDef() { return(def); }


    /**Non-empty Map from period to period row; never null.
     * Logically immutable once created (and thus thread-safe),
     * though the values remain mutable.
     */
    private final EnumMap<EventPeriod, EventVariableValuePeriodRow> byPeriod;


    /**Time last event was added, or zero if none recorded or instance just deserialised.
     * Is transient (not retained though serialisation).
     * <p>
     * Is volatile to allow access without a lock.
     */
    private transient volatile long lastEventTime;

    /**Get time last event was added, or zero if none recorded or instance just deserialised. */
    public long getLastEventTime() { return(lastEventTime); }


    /**Record/add event to the tallies for each period.
     * Note that the event must match the definition that this instance is constructed with
     * or an IllegalArgumentException will be thrown.
     * <p>
     * Note that for some types, not all periods/intervals are selected,
     * ie we record data for some periods and not others.
     * <p>
     * Updates the lastEventTime to the current time.
     *
     * @param svv  event to be recorded; must be non-null and of event type
     */
    public synchronized void addEvent(final SimpleVariableValue svv)
        {
        if((svv == null) || !def.equals(svv.getDef()))
            { throw new IllegalArgumentException(); }

        // Add event to each period row that we have.
        for(final EventVariableValuePeriodRow pr : byPeriod.values())
            { pr.addEvent(svv); }

        // Take the latest possible view of when all the event setting was finished...
        lastEventTime = System.currentTimeMillis();
        }


    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * If data is requested for an interval/period excluded in the definition,
     * an empty array is returned.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     *
     * @param intervalSelector  never null
     * @param intervalNumber  the number of the first interval
     *     for which data is potentially required;
     *     if too far in the past or future then possibly no data
     *     will be available
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    EventVariableValue[] getEventValues(final EventPeriod intervalSelector,
                                        final long intervalNumber,
                                        final BitSet whichValues)
        {
        final EventVariableValuePeriodRow eventVariableValuePeriodRow = byPeriod.get(intervalSelector);
        if(eventVariableValuePeriodRow == null) { return(EventVariableValuePeriodRow._NO_EVENT_VALUES); }
        return(eventVariableValuePeriodRow.getEventValues(intervalNumber, whichValues));
        }

    /**Set/override the specified event sets for the specified intervals; never null.
     * This is used to set values that have been fetched from upstream
     * (ie closer to the master data store).
     * <p>
     * This may ignore the call,
     * and certainly will if the intervalNumber is outside the current time window it holds.
     * <p>
     * This will generally accept the value and store it if:
     * <ul>
     * <li>The intervalNumber is within the current window.
     * <li>The intervalSelector is a period for which this event has data.
     * <li>The supplied value is "better" than any current one held, ie:
     *     <ul>
     *     <li>We hold a null and evv is non-null.
     *     <li>We hold a non-authoritative value and evv is authoritative.
     *     <li>The authoritative-ness is the same bu we hold an entry
     *         with a lower total count than evv.
     *     <ul>
     *     </ul>
     * </ul>
     *
     * @param evv  non-null event value
     */
    public synchronized void setEventValue(final EventVariableValue evv)
        {
        if(evv == null)
            { throw new IllegalArgumentException(); }
        final SimpleVariableDefinition def = evv.getDef();
        if(!def.equals(this.def))
            { throw new IllegalArgumentException("mismatched def: "+def+" vs "+this.def); }

        final EventVariableValuePeriodRow eventVariableValuePeriodRow =
                byPeriod.get(evv.getPeriod());
        // If data is not maintained for the specified period
        // then return immediately...
        if(eventVariableValuePeriodRow == null) { return; }

        eventVariableValuePeriodRow.setEventValue(evv);

        // Mark the set as needing a save since we probably updated it...
        lastEventTime = System.currentTimeMillis();
        }




    /**Constructor for defensive copying on deserialisation.
     *
     * @param def  the variable definition; never null and must be an event
     * @param byPeriod  the extant mapping from period to "period row" entries; never null
     *
     * @throws IllegalArgumentException  if the arguments are invalid
     */
    private EventVariableValueSet(final SimpleVariableDefinition def,
                                  final EnumMap<EventPeriod, EventVariableValuePeriodRow> byPeriod)
        throws IllegalArgumentException
        {
        this.def = def;

        // Defensive copy for safety.
        // FIXME: probably this should be a deep copy, since the map values are mutable...
        this.byPeriod = new EnumMap<EventPeriod, EventVariableValuePeriodRow>(byPeriod);

        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Deserialise: use constructor for validation, defensive copying, etc. */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        return(new EventVariableValueSet(def, byPeriod));
        }


    /**Check that the object state is consistent and legal. */
    public void validateObject()
        throws InvalidObjectException
        {
        if(def == null)
            { throw new InvalidObjectException("bad object: def must not be null"); }
        if(!def.isEvent())
            { throw new InvalidObjectException("bad object: def not an event type"); }

        if(byPeriod == null)
            { throw new InvalidObjectException("bad object: byPeriod map must not be null"); }
        if(byPeriod.isEmpty())
            { throw new InvalidObjectException("bad object: byPeriod map must not be empty"); }
        final Set<EventPeriod> evPeriodSubset = def.getEvPeriodSubset();
        // If no subset is specified, then use the full set of periods/intervals.
        final Set<EventPeriod> _evPeriodSubset = (evPeriodSubset != null) ? evPeriodSubset :
                                                     EnumSet.allOf(EventPeriod.class);
        if(_evPeriodSubset.size() != byPeriod.size())
            { throw new InvalidObjectException("bad object: byPeriod map size is wrong"); }
        if(!_evPeriodSubset.containsAll(byPeriod.keySet()))
            { throw new InvalidObjectException("bad object: byPeriod contains extra keys"); }
        if(!byPeriod.keySet().containsAll(_evPeriodSubset))
            { throw new InvalidObjectException("bad object: byPeriod missing at least one key"); }
        for(final Object v : byPeriod.values())
            {
            if(!(v instanceof EventVariableValuePeriodRow))
                { throw new InvalidObjectException("bad object: byPeriod map has missing/invalid value"); }
            }
        }

    /**Unique Serialisation class ID. */
    private static final long serialVersionUID = -1932690647089333134L;
    }
