/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.Rnd;

import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.ObjectStreamException;
import java.io.Serializable;


/**Immutable globally-unique instance ID for a JVM participating in a set of SystemVariables.
 * This is the key to a globalmap variable,
 * and each participating system should have just one such value,
 * probably generated by a data pipeline endpoint.
 * <p>
 * Internally, this is a randomly-generated (non-negative) long value.
 */
public final class InstanceID implements Serializable,
                                         ObjectInputValidation,
                                         MemoryTools.Internable,
                                         Comparable
    {
    /**Create a new, unique instance ID; never null.
     * This is immediately intern()ed so that there will never be more than
     * one copy of each unique InstanceID in memory at any one time
     * (except transiently during deserialisation).
     */
    public static InstanceID createInstanceID()
        {
        return(MemoryTools.intern(new InstanceID()));
        }

    /**Create a new instance with a new/unique ID from a cryptographically-good random-number source. */
    private InstanceID()
        {
        // Guaranteed non-negative new ID.
        ID = (Rnd.goodRnd.nextLong() >>> 1);
        }

    /**Unique (positive) long ID from a cryptographically-good random-number source. */
    private final long ID;

    /**Long representation of ID; non-negative. */
    public long longValue()
        {
        return(ID);
        }

    /**String representation of ID as a non-negative base-36 number. */
    public String toString()
        {
        return(Long.toString(ID, Character.MAX_RADIX));
        }

    /**Comparable ordering is on ID value alone in numeric ID order. */
    public int compareTo(final Object o)
        {
        final InstanceID other = (InstanceID) o;
        if(ID < other.ID) { return(-1); }
        if(ID > other.ID) { return(+1); }
        return(0);
        }

    /**Equality is based on the whole ID.
     * But in fact due to instance control we can in fact use ==
     * at least as an optimisation!
     */
    public boolean equals(final Object obj)
        {
        if(this == obj) { return(true); }
        if(!(obj instanceof InstanceID)) { return(false); }
        final InstanceID other = (InstanceID) obj;

        // Compare the ID values.
        return(ID == other.ID);
        }

    /**The hash is derived from the ID.
     * This implementation just uses the hash of the name.
     */
    public int hashCode()
        {
        return((int) ID);
        }

    /**Deserialise: validate and eliminate duplicates coming off the wire.
     * Ensures that there is never more than one copy of any given
     * InstanceID value in memory (other than transiently).
     *
     * @return identical, de-duped, non-null instance
     */
    protected Object readResolve()
        throws ObjectStreamException
        {
        validateObject();
        return(MemoryTools.intern(this));
        }

    /**Check for legal state. */
    public void validateObject()
        throws InvalidObjectException
        {
        if(ID < 0)
            { throw new InvalidObjectException("bad object: negative ID"); }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -2320309863687936784L;
    }
