/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.Tuple;
import org.hd.d.pg2k.svrCore.props.LocalProps;

import java.io.EOFException;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;


/**Class to "manage" variables for an element of a SimpleExhibitPipelineIF.
 * Caches values to minimise read (and optionally write) activity.
 * <p>
 * Always mantains a local store of variables,
 * and normally sync() needs to be called periodically
 * to merge with upstream values.
 * <p>
 * Typical places to inject one of these is in a caching stage,
 * or at pipeline or tunnel end-points.
 * <p>
 * A non-write-though instance has the effect of grouping multiple
 * setVariable() and setVariables() operations into a single
 * setVariables() upstream action when sync() is invoked.
 * This may still write-though in some circumstances.
 * <p>
 * In no case does a getVariable() or getVariables() operation
 * directly cause a request to go upstream,
 * (thus helping to ensure that get operations are fast even if there
 * are I/O problems (etc) upstream),
 * but rather the set of values is
 * updated periodically with a single getVariables() operation.
 * <p>
 * Thread-safe; does not hold any externally-visible locks.
 * <p>
 * This does not count as an "end-point", ie is never the master data store.
 */
public class PipelineVarMgr extends BasicVarMgr
                            implements BasicVarMgrInterface,
                                       SimpleVariablePipelineIF
    {
    /**If true, always use getVariables(-1) to get all values from upstream.
     * Inefficient but very robust; generally not the right thing to do though.
     */
    private static final boolean ALWAYS_FETCH_ALL_UPSTREAM_VALUES = false;

    /**Create a default PipelineVarMgr instance.
     * Defaults to write-through (ie read caching only),
     * because that is "safer" behaviour.
     *
     * @param upstream  upstream source of data and variable values;
     *     never null
     */
    public PipelineVarMgr(final SimpleVariablePipelineIF upstream)
        {
        this(upstream, true); // Write-through.
        }

    /**Creates a customised VarMgr instance.
     * May try to get the local system ID immediately from upstream,
     * so that it is available for creating globals.
     *
     * @param upstream  upstream source of data and variable values;
     *     never null
     * @param writeThrough  if true, all set operations are propagated
     *     downstream immediately
     */
    public PipelineVarMgr(final SimpleVariablePipelineIF upstream,
                          final boolean writeThrough)
        {
        if(upstream == null)
            { throw new IllegalArgumentException(); }

        // If this cache is write-through then we don't need an outbound store.
        outboundUpdates = writeThrough ? null : new Hashtable<SimpleVariableDefinition, SimpleVariableValue>();
        earlyUpdates = writeThrough ? null : new ArrayList<SimpleVariableValue>();
        source = upstream;

        if(GET_SYSID_IN_CONS)
            {
            // Get the local system ID from upstream (if we can)
            // and save it since we shall need it for globals.
            try
                {
                final SimpleVariableValue svv =
                    source.getVariable(SystemVariables.LOCAL_SYS_ID);
                if(svv != null)
                    super._setVariable(svv, true);
                }
            catch (IOException e)
                {
                // Ignore error; hope to get value later...
                }
            }
        }

    /**Time after which items pending to go upstream are dropped after an error.
     * This is to keep a finite bound on the size of any outbound queue
     * in case of some long-lasting error (eg the master server down).
     * <p>
     * This is taken to be somewhat longer than the longest time the data
     * could reasonably be useful at first glance,
     * to allow for clock skew and other insults.
     * <p>
     * Everything can probably be discarded safely after this time;
     * some values can probably be discarded much sooner.
     */
    private final static long _MAX_UPSTREAM_RETENTION_MS = 1001 +
        ((3 * Math.max(SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS,
                       SystemVariables.EVENT_INTERVAL_VLONG_TERM_MS)) / 2);

    /**If true, true in the constructor to get the local system ID from upstream.
     * This helps us build globalMaps,
     * and also ensures that the variable is available to clients downstream.
     */
    private final static boolean GET_SYSID_IN_CONS = true;

    /**Upstream source of variable values; never null. */
    private final SimpleVariablePipelineIF source;

    /**Returns true if this is a write-through cache. */
    public boolean isWriteThrough()
        { return(outboundUpdates == null); }

    /**Cached outbound variable value updates to be sent upstream.
     * Will not be used (and will be null) if in write-through mode.
     * <p>
     * Used by:
     * <ul>
     * <li>setVariable() (to add values)
     * <li>sync() to push accumulated values upstream
     * </ul>
     * <p>
     * Variable updates are held in this table and periodically flushed
     * upstream in a block.  If multiple updates to a single variable
     * occur between flushes, only the last one will be retained,
     * (the others overwritten) thus reducing traffic.
     * These may be re-ordered with respect to one another when flushed.
     * <p>
     * Note however that one value cannot be overwritten by another if
     * they have different globalMap keys.
     * This can only really happen on the master when simultaneous updates
     * arrive from two different slaves for the same global value.
     * In this case, we force the previous value upstream immediately.
     * <p>
     * A lock must be held on this object to do multiple operations atomically,
     * but no other visible lock must be grabbed inside a lock on this.
     * <p>
     * Hashtable from SimpleVariableDefinition to SimpleVariableValue.
     * <p>
     * A Hashtable is used for thread-safety.
     * <p>
     * Never null.
     */
    private final Hashtable<SimpleVariableDefinition,SimpleVariableValue> outboundUpdates;

    /**A List of non-coalesced variable set operations to be done first.
     * These are variables that could not be coalesced for whatever reason,
     * and must be passed upstream ahead of anything in outboundUpdates.
     * <p>
     * This will be non-null when outboundUpdates is.
     * <p>
     * The content of this List must only be accessed in the scope of a lock
     * on outboundUpdates, so this list implementation need not be inherently
     * thread-safe (eg ArrayList could be used instead of Vector).
     */
    private final List<SimpleVariableValue> earlyUpdates;


    /**Last time we successfully fetched variable values from upstream; initially null.
     * Declared as volatile so allow safe access without a lock.
     */
    private volatile Long lastFetch;

    /**Set a single variable value.
     * If writing though,
     * then set the value upstream immediately,
     * else simply put it in our "pending" list waiting for sync().
     *
     * @param newValue  new variable value to set; never null
     * @throws java.io.IOException in case of difficulty setting the value upstream
     */
    public void setVariable(final SimpleVariableValue newValue)
        throws IOException
        {
        // Store in our local cache before doing anything else...
        super._setVariable(newValue, false);

        // Since we seem to have set the value OK
        // (so it type-checked, for example),
        // add to outgoing update list immediately,
        // or write-though if appropriate.

        // Write-though.
        if(outboundUpdates == null)
            {
            source.setVariable(newValue);
            return;
            }

        // Accumulate and wait for sync().
        // Attempt to merge redundant (non-event) operations for efficiency.
        synchronized(outboundUpdates)
            {
            // But for now we never coalesce events.
            final SimpleVariableDefinition def = newValue.getDef();
            if(def.isEvent())
                {
                earlyUpdates.add(newValue);
                return;
                }

            // Note however that one value cannot be overwritten by another
            // iff they have different globalMap keys (or null-ness).
            // This happens on the master when simultaneous updates
            // arrive from two different slaves for the same global value.
            // In this case, we force the previous value upstream now.
            final SimpleVariableValue prevHeld =
                    outboundUpdates.put(def, newValue);

            // Have to handle coalescing if there was a previous value
            // for this variable.
            if(prevHeld != null)
                {
                // Whoops, may conflict...
                // If global maps have different keys, or null-ness,
                // then they do conflict,
                // and we must flush the previous value
                // into the earlyUpdates list to go first,
                // else we can discard the previous version.
                final Map nGM = newValue.getGlobalMap();
                final Map pGM = prevHeld.getGlobalMap();
                final boolean differentMapKeys = (nGM == null) ?
                        (pGM != null) :
                        ((pGM == null) || !nGM.keySet().equals(pGM.keySet()));
                if(differentMapKeys)
                    { earlyUpdates.add(prevHeld); }
                }
            }
        }

    /**Set several variable values.
     * This is equivalent to set series of calls to setVariables()
     * unless in write-through mode where a single setVariables()
     * is done upstream for efficiency.
     *
     * @return number of values successfully set, at least locally...
     * @param newValues  new variables values to set; never null
     * @throws java.io.IOException in case of difficulty setting the values upstream
     */
    public int setVariables(final SimpleVariableValue newValues[])
        throws IOException
        {
        if(newValues == null)
            { throw new IllegalArgumentException(); }

        if(outboundUpdates != null)
            {
            // If not writing-through,
            // then put in "outbound" list,
            // being careful to preserve order,
            // but eliminating duplicates.
            final int nvLen = newValues.length;
            for(int i = 0; i < nvLen; ++i)
                { setVariable(newValues[i]); }
            return(newValues.length);
            }
        else
            {
            // If writing through,
            // stick to single setVariables() call for efficiency.
            super.setVariables(newValues);
            return(source.setVariables(newValues));
            }
        }


    /**Private lock to serialise calls on sync(). */
    private final Object _sync_lock = new Object();

    /**Synchronise variables with upstream values.
     * We push updated values upstream to the source,
     * call sync on the source if this sync is "forced",
     * and then retrieve changed values from upstream.
     * <p>
     * We remember when we last did this and request any changes since then,
     * except on the first call or if called with force==true,
     * in which case all values are retrieved.
     * <p>
     * To allow for (varying) clock skew between different server instances,
     * we actually ask for somewhat older values than our last poll,
     * which will result in some redundant traffic.
     * <p>
     * Holds no externally-visible locks,
     * and does not hold locks while communicating upstream
     * (so can handle set/get operations even if upstream is slow/blocking),
     * but if called by multiple threads this will safely serialise the calls.
     *
     * @param force  if true, this will force a full sync on the read side
     *     by using getVariables(-1) rather than attempting to choose a
     *     nearer timestamp for efficiency;
     *     the implementation is at liberty to use getVariables(-1)
     *     at any time whatever the argument value,
     *     and almost certainly should use it on the first call
     *
     * @throws java.io.IOException if one is received from upstream
     */
    public void syncVariables(final boolean force)
        throws IOException
        {
//if(isDebugMode) { System.err.println("[PipelineVarMgr "+System.identityHashCode(this)+": sync("+force+").]"); }

        // Serialise calls to sync() to avoid out-of-order updates...
        synchronized(_sync_lock)
            {
            // Push locally-cached updates upstream,
            // if there are any pending,
            // and clear those pending values.
            if(outboundUpdates != null)
                {
                // Get any earlyUpdates (first)
                // then any other outbountUpdates,
                // and assemble them into a single update to send upstream.
                // Upon failure, put them at the start of the list
                // in case new items have arrived.
                //
                // We *do not* hold the lock while communicating upstream.
                SimpleVariableValue toGoUpstreamAll[] = null;
                synchronized(outboundUpdates)
                    {
                    // Atomically retrieve current pending updates, if any,
                    // and clear the pending updates.
                    // We can put them back afterwards if we have to, sort of.
                    final int size =
                        earlyUpdates.size() + outboundUpdates.size();
//if(isDebugMode && (size > 0)) { System.err.println("[PipelineVarMgr "+System.identityHashCode(this)+": sync("+force+"): outboundUpdates.size()="+size+".]"); }
                    if(size > 0)
                        {
                        // Assemble updates in order, earlyUpdates first.
                        final List<SimpleVariableValue> allU = new ArrayList<SimpleVariableValue>(size);
                        allU.addAll(earlyUpdates);
                        allU.addAll(outboundUpdates.values());

                        // Get into one array, in order,
                        // ready to send upstream.
                        toGoUpstreamAll = new SimpleVariableValue[size];
                        allU.toArray(toGoUpstreamAll);

                        // Clear the originals...
                        earlyUpdates.clear();
                        outboundUpdates.clear();
                        }
                    }

                if(toGoUpstreamAll != null)
                    {
                    // Attempt to send the outstanding values upstream...
                    // However, if this fails, due to an I/O problem,
                    // then try to reinsert any not-yet-overwritten values
                    // at the start of the earlyUpdates
                    // so that they will go before any later arrivals.
                    //
                    // Discard anything too old to be useful upstream,
                    // which also helps ensure that this list does not grow without bound.
                    // We can probably safely discard events sure to have missed the right "slot"
                    // and non-event passing the normal latency limits by a comfortable margin.
                    try { source.setVariables(toGoUpstreamAll); }
                    catch(IOException e)
                        {
                        // Filter out anything too old.
                        final long now = System.currentTimeMillis();
                        final long tooOld = now - _MAX_UPSTREAM_RETENTION_MS;
                        // Ordinary non-event values can be dumped sooner...
                        final long tooOldNonEvent = now - 501 - 2*SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS;

                        final List<SimpleVariableValue> toGoUpstreamFiltered =
                                new ArrayList<SimpleVariableValue>(toGoUpstreamAll.length);
                        for(int i = 0; i < toGoUpstreamAll.length; ++i)
                            {
                            final SimpleVariableValue svv = toGoUpstreamAll[i];
                            if(svv == null) { continue; }
                            if(svv.getTimestamp() < tooOld) { continue; }
                            if(!svv.getDef().isEvent() && (svv.getTimestamp() < tooOldNonEvent)) { continue; }
                            toGoUpstreamFiltered.add(svv);
                            }

                        // Protect access to earlyUpdates...
                        synchronized(outboundUpdates)
                            {
                            // Insert the not-too-old items in order at the start...
                            earlyUpdates.addAll(0, toGoUpstreamFiltered);
                            }
//if(isDebugMode) { System.err.println("[PipelineVarMgr "+System.identityHashCode(this)+": sync("+force+") got IOException.]"); }
//if(isDebugMode) { e.printStackTrace(); }
                        throw e; // Rethrow the error...
                        }
                    }
                }

            // Now force any changes all the way to the master's end-point
            // iff this is a "forced" sync().
            if(force) { source.syncVariables(true); }

//if(isDebugMode) { System.err.println("[PipelineVarMgr "+System.identityHashCode(this)+": sync("+force+") about to get...]"); }

            // Now retrieve any values from downstream.
            // For the moment we'll assume that our local clock,
            // plus the cache hold time used as a skew corrector,
            // will ensure that we don't miss anything
            // at the cost of a few wasted/redundant transfers,
            // though:
            //   * on the first run, or
            //   * if "force"d, or
            //   * we got nothing on our last poll,
            // then we fetch everything with a getVariables(-1).
            final Long lastTime = lastFetch; // Fetch value just once.
            final boolean firstRun = (lastTime == null);
            final boolean getAll = (ALWAYS_FETCH_ALL_UPSTREAM_VALUES || force || firstRun);
            final SimpleVariableValue[] newVals =
                source.getVariables(getAll ? -1 :
                    (lastTime.longValue() - CoreConsts.MAX_PEER_CLOCK_SKEW_MS - 1));
            // Apply the values fetched...
            // (For safety, we discard any that appear to be invalid.)
            for(int i = newVals.length; --i >= 0; )
                {
                final SimpleVariableValue svv = newVals[i];
                final SimpleVariableDefinition def = svv.getDef();
                if(!SystemVariables.defs.contains(def))
                    {
                    System.err.println("PipelineVarMgr: ERROR: upstream variable value rejected: " + svv.getDef());
                    continue;
                    }

                // Store the new value in our local cache,
                // but don't have it propagate back upstream!
                // Note that it is from upstream,
                // so read-only values *can* be set in our cache.
                super._setVariable(svv, true);
                }


            // If we have any deferred requests for current/all interval events,
            // then attempt to fetch and store/cache one at random now
            // in the hope of being able to satisfy the next request for it.
            // If something goes wrong and we "lose" one of these
            // deferred requests, that is OK, since we made no promises
            // about the timeliness of current-interval data.
            for(final boolean current : new boolean[]{ false, true } )
                {
                final Tuple.Pair<SimpleVariableDefinition, EventPeriod> deferred =
                    _getAndClearDeferredRequestForInterval(current);
                if(deferred != null)
                    {
                    // Current interval, which we never expect to be authoritative...
                    final long currentInterval =
                        deferred.second.getIntervalNumber(System.currentTimeMillis());

                    final BitSet justTheOne = new BitSet(1);
                    justTheOne.set(0); // Just get the one current value set.
                    final EventVariableValue upstreamValues[] =
                        source.getEventValues(deferred.first, deferred.second,
                                              current ? currentInterval : 0,
                                              justTheOne);

                    // Save the fetched value for the next request!
                    if(upstreamValues.length > 0)
                        {
                        final EventVariableValue evv = upstreamValues[0];
                        if(evv != null)
                            {
if(IsDebug.isDebug) { System.out.println("[Post-fetched "+(current?"current":"all")+"-interval data for " + deferred + ".]"); }
                            super.setEventValue(evv);
                            }
                        }
                    }
                }

            // Note when we successfully finished our work.
            // However, if we got no upstream updates at all,
            // then VERY RARELY revert to a full update next time
            // to ensure that we are not missing something important.
            // Do this less often for more bandwidth-constrained systems
            // to reduce unnecessary "chatter".
            final int serverSlowdownFactor = LocalProps.getServerSlowdownFactor();
            lastFetch = ((newVals.length == 0) && (0 == Rnd.fastRnd.nextInt(3 + 101*serverSlowdownFactor))) ?
                null : new Long(System.currentTimeMillis());
            }

//if(isDebugMode) { System.err.println("[PipelineVarMgr "+System.identityHashCode(this)+": sync("+force+") FINISHED]"); }
        }


    /**If true, don't go upstream (synchronously) just to fetch values for the current (or all) interval; respond from local store.
     * We may nonetheless ask for current-interval data from upstream:
     * <ul>
     * <li>when asking for something else in the same request,
     * <li>asynchronously when a particular event/period combination
     *     has been asked for recently.
     * </ul>
     * thus only making for a marginal extra cost.
     * <p>
     * If true, there is no guarantee that the "current" data reflects more than
     * locally-originated events or will be hugely up-to-date,
     * though the master's main copy will be.
     */
    private static final boolean DONT_FETCH_SPECIAL_INTERVALS_SYNCHRONOUSLY = true;

    /**Set of "current" interval event-values requested that we declined to fetch synchronously.
     * We clear these as we asynchronously fetch the requested values.
     * <p>
     * A map from event variable definitions to
     * a set of any deferred current-interval requests for that set
     * (or to null if none have ever been deferred for that event type).
     * <p>
     * A Hashtable is used to guarantee basic thread safety.
     * All manipulation of the contents of the table,
     * and in particular of the non-threadsafe EnumSet values,
     * must be done holding a lock on the table instance.
     */
    private final Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>> _deferredCIRequests =
        new Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>>();

    /**Set of "all" interval event-values requested that we declined to fetch synchronously.
     * We clear these as we asynchronously fetch the requested values.
     * <p>
     * A map from event variable definitions to
     * a set of any deferred all-interval requests for that set
     * (or to null if none have ever been deferred for that event type).
     * <p>
     * A Hashtable is used to guarantee basic thread safety.
     * All manipulation of the contents of the table,
     * and in particular of the non-threadsafe EnumSet values,
     * must be done holding a lock on the table instance.
     */
    private final Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>> _deferredAIRequests =
        new Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>>();

    /**Class definining all details of a deferred request.
     * This comprises:
     * <ul>
     * <li>Event name/definition,
     * <li>EventPeriod,
     * <li>The all/current flag.
     * </ul>
     * <p>
     * Equality depends on all fields.
     */
    private final class NextRequestKey
        {
        /**Event definition; never null. */
        final SimpleVariableDefinition def;

        /**Event period; never null. */
        final EventPeriod p;

        /**True if "all", else false for "current". */
        final boolean isAll;

        public NextRequestKey(final SimpleVariableDefinition def,
                              final EventPeriod p,
                              final boolean isAll)
            {
            this.def = def;
            this.p = p;
            this.isAll = isAll;

            assert(p != null);
            assert((def != null) && def.isEvent());
            assert((def.getEvPeriodSubset() == null) ||
                   def.getEvPeriodSubset().contains(p));
            }

        /**The hash depends on all fields. */
        public int hashCode()
            { return(def.hashCode() ^ (isAll ? p.ordinal() : p.name().hashCode())); }

        /**Equality depends on all fields. */
        public boolean equals(final Object obj)
            {
            if(!(obj instanceof NextRequestKey)) { return(false); }
            final NextRequestKey other = (NextRequestKey) obj;
            if(isAll != other.isAll) { return(false); }
            if(!p.equals(other.p)) { return(false); }
            if(!def.equals(other.def)) { return(false); }
            return(true);
            }
        }

    /**Next permitted (deferred) request for a given all/current period data set.
     * We defer repeated requests for the same item until either:
     * <ul>
     * <li>The start of the next interval for that item.
     * <li>A reasonable fraction of the period for that item
     *     has elapsed since the last request for it.
     * </ul>
     * <p>
     * This means that we limit the cost of repeated requests for the same item
     * in such a way that we still have a reasonable chance of seeing
     * intra-interval updates.
     * <p>
     * The table is thread-safe and parameterised for quick, concurrent access.
     */
    private final Map<NextRequestKey,Long> _nextRequestNotBefore =
        new ConcurrentHashMap<NextRequestKey, Long>(41, 0.7f, 4);

    /**Note that we have deferred a request for the given event and period.
     *
     * @param currentInterval  true for "current" interval,
     *     false for "all" interval
     * @param def  non-null event defintion
     * @param intervalSelector  non-null interval/period value
     */
    private final void _noteDeferredRequestForInterval(final boolean currentInterval,
                                                       final SimpleVariableDefinition def,
                                                       final EventPeriod intervalSelector)
        {
        if((def == null) || !def.isEvent() ||
           (intervalSelector == null))
            { throw new IllegalArgumentException(); }

        // Caller must be asking for valid period for this event type.
        assert((def.getEvPeriodSubset() == null) ||
               def.getEvPeriodSubset().contains(intervalSelector));


        // We may ignore rapid repeated requests for the same item,
        // if the repeat is within a small fraction of the period.
        final NextRequestKey key = new NextRequestKey(def, intervalSelector, !currentInterval);
        final Long notBefore = _nextRequestNotBefore.get(key);
        // Return if too early for another request for this value yet.
        final long now = System.currentTimeMillis();
        if((notBefore != null) && (notBefore.longValue() > now))
            { return; }

        // Compute/record when it would next be OK to ask for an update.
        // At most we put this off by a fraction of our temporal slackness,
        // and as a minimum a substantial fraction of our sysvar latency.
        final long nextNotBefore = now +
            Math.min(CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 249,
                intervalSelector.getIntervalMs() >>> 4) +
            Rnd.fastRnd.nextInt(SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS/2);
        _nextRequestNotBefore.put(key, new Long(nextNotBefore));


        // Select which deferred set we are working with...
        final Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>> diSet =
            currentInterval ? _deferredCIRequests : _deferredAIRequests;

        // Hold a lock briefly while we do our work.
        synchronized(diSet)
            {
            // Get any existing Set of values.
            final EnumSet<EventPeriod> s = diSet.get(def);
            // If none then create and store a new Set with just this value.
            if(s == null)
                {
                diSet.put(def, EnumSet.of(intervalSelector));
                return;
                }
            // If non-null then just add this bit to the Set already held.
            s.add(intervalSelector);
            }
        }

    /**Select, remove and return at random one of the deferred current/all interval requests; null if none.
     * This is an atomic fetch-and-clear operation,
     * designed to be used in association with
     * _noteDeferredRequestForInterval().
     *
     * @param currentInterval  true for "current" interval,
     *     false for "all" interval
     * @return null or a pair with both elements valid and non-null
     */
    private Tuple.Pair<SimpleVariableDefinition, EventPeriod> _getAndClearDeferredRequestForInterval(
                                            final boolean currentInterval)
        {
        // Select which deferred set we are working with...
        final Hashtable<SimpleVariableDefinition,EnumSet<EventPeriod>> diSet =
            currentInterval ? _deferredCIRequests : _deferredAIRequests;

        // Hold a lock briefly while we do our work.
        synchronized(diSet)
            {
            // Return null immediately if no deferred requests pending.
            final int n = diSet.size();
            if(n == 0) { return(null); }

            // Pick one of the event types at random.
            final SimpleVariableDefinition defs[] = new SimpleVariableDefinition[n];
            diSet.keySet().toArray(defs);
            final SimpleVariableDefinition def =
                defs[(n == 1) ? 0 : Rnd.fastRnd.nextInt(n)];

            // The def should be a valid event type.
            assert(def != null);
            assert(def.isEvent());

            // Get the Set of deferred requests for this event type.
            final EnumSet<EventPeriod> s = diSet.get(def);

            // The Set should be non-null and non-empty.
            assert(s != null);
            assert(!s.isEmpty());

            // Pick one of the pending iterval/period values at random.
            final int m = s.size();
            final EventPeriod eps[] = new EventPeriod[m];
            s.toArray(eps);
            final EventPeriod ep =
                eps[(m == 1) ? 0 : Rnd.fastRnd.nextInt(m)];

            // We should have chosen a valid interval.
            assert((def.getEvPeriodSubset() == null) ||
                   def.getEvPeriodSubset().contains(ep));

            // Compose our result.
            final Tuple.Pair<SimpleVariableDefinition, EventPeriod> result =
                new Tuple.Pair<SimpleVariableDefinition, EventPeriod>(def, ep);

            // Now remove our pending value,
            // and indeed the whole Set if now empty.
            if(m == 1) { diSet.remove(def); }
            else { s.remove(ep); }

            return(result);
            }
        }
    
    /**Locks to prevent redundant concurrent updates of the same event type from upstream in getEventValues().
     * Locking at the per-def level probably admits almost all potential concurrency
     * and is relatively simple.
     * <p>
     * We allow for a limited amount of concurrency in writing
     * but we mainly use ConcurrentHashMap for thread-safety and putIfAbsent().
     */
    private final ConcurrentHashMap<SimpleVariableDefinition, ReentrantLock> _getEventValues_locks =
        new ConcurrentHashMap<SimpleVariableDefinition, ReentrantLock>(2*SystemVariables.defs.size(), 0.5f, 4);

    /**Get event values.
     * We get what we can from our local store,
     * then go upstream to fill in any unfilled requested slots
     * for which we did not have authoritative data
     * (and then attempt to cache them locally for next time).
     * <p>
     * We treat the 'current' and 'all' intervals specially
     * since upstream will never offer us an authoritative value
     * for the current slot, only for older completed ones.
     * We get requested values asynchronously,
     * to be reasonably up-to-date for the <em>next</em> request.
     * <p>
     * If a request for a "local" event upstream fails with an IllegalArgumentException
     * then we assume that upstream of us is a tunnel which local values don't cross
     * so we simply report whatever value we have to hand.
     *
     * @param def  non-null event-variable definition
     * @param intervalSelector  non-null valid interval for specified event
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc;
     *     null is treated as the common case equivalent to just bit 0 set
     *
     * @throws IOException  if we have difficulty getting values from upstream
     * @throws InterruptedIOException  if interrupted or if blocking for too long.
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException, InterruptedIOException
        {
//System.out.println("PipelineVarMgr.getEventValues("+def+", "+intervalSelector+", "+intervalNumber+", "+whichValues+")");

        // Current interval number (never authoritative).
        final long currentInterval = intervalSelector.getIntervalNumber(System.currentTimeMillis());

        // Optimisation fast-path for common/simple case:
        // the single requested slot/value is available locally and is authoritative.
        // If available from local store then we don't need to use our locks,
        // else we fall through for our more general case.
        final boolean nullWV = (whichValues == null);
        if(nullWV ||
           ((whichValues.cardinality() == 1) && whichValues.get(0)))
            {
            // Don't try to satisfy requests for the current/all slots here...
            if((intervalNumber != 0) && (intervalNumber != currentInterval))
                {
                // Get what we can from local store.
                final EventVariableValue[] result =
                    super.getEventValues(def, intervalSelector, intervalNumber, null);

                if((result.length != 0) &&
                    (result[0] != null) &&
                    (result[0].isAuthoritative()))
                    { return(result); /* Good, local cached value. */ }
                }
            }

        // Verify that this this def is an event and is genuine!
        assert(def.isEvent());
        assert(SystemVariables.defs.contains(def));

        // First get what we can from our local store/cache.
        // This call will validate our arguments for us too.
        EventVariableValue[] result =
            super.getEventValues(def, intervalSelector, intervalNumber, whichValues);

        // If any requested values are null or non-authoritative,
        // then try to go upstream for them (and then cache them locally).
        BitSet wv = whichValues;
        if(wv == null)
            {
            // Create synthetic BitSet for simplicity below.
            wv = new BitSet(1);
            wv.set(0);
            }
        // Note any "missing" values.
        final BitSet nonAuthValues = new BitSet(); 
        for(int i = wv.nextSetBit(0); i >= 0; i = wv.nextSetBit(i+1))
            {
            if((i >= result.length) ||
                    (result[i] == null) ||
                    !result[i].isAuthoritative())
                { nonAuthValues.set(i); }
            }

        // If the entire request has been satisfied from local cache
        // then we can return immediately!
        if(nonAuthValues.length() == 0)
            { return(result); }

        // If the only non-authoritative answer is
        // in the "current" or "all" slot
        // (it never would be authoritative anyway)
        // then return now what we have locally cached/available.
        // Don't bother with the expense of going upstream synchronously,
        // but just note that the value was requested and deferred.
        // We will go and fetch a fresh value soon, in the background.
        if(DONT_FETCH_SPECIAL_INTERVALS_SYNCHRONOUSLY &&
                (nonAuthValues.cardinality() == 1))
            {
            final int which = nonAuthValues.length()-1; // Index of the one bit that is set...

//        assert(which == nonAuthValues.nextSetBit(0));
//        assert(nonAuthValues.get(which));

            if(intervalNumber == 0)
                {
                _noteDeferredRequestForInterval(false, def, intervalSelector);
                return(result);
                }
            else if(intervalNumber - which == currentInterval)
                {
                _noteDeferredRequestForInterval(true, def, intervalSelector);
                return(result);
                }
            }

        // WILL NEED TO GO UPSTREAM: ALLOW AT MOST ONE PER DEF AT ONCE.
        // Get the unique lock for this event/def.
        ReentrantLock _getEventValues_lock;
        while(null == (_getEventValues_lock = _getEventValues_locks.get(def)))
            {
            // Ensure no races for a lock creation,
            // ie only one lock will ever be current for one def.
            _getEventValues_locks.putIfAbsent(def, new ReentrantLock());
//if(IsDebug.isDebug) { System.out.println("[_getEventValues_locks.size()="+_getEventValues_locks.size()+".]"); }
            }

        // Now try to grab the lock associated with this def,
        // but we are only prepared to block for a short while.
        try
            {
            if(!_getEventValues_lock.tryLock(CoreConsts.MAX_INTERACTIVE_DELAY_MS/2, TimeUnit.MILLISECONDS))
                {
                System.err.println("INFO: could not quickly acquire lock in PipelineVarMgr.getEventValues() for "+def);
                final InterruptedIOException e = new InterruptedIOException("could not quickly acquire lock in PipelineVarMgr.getEventValues("+def+", "+
                                               intervalSelector+", "+
                                               intervalNumber+"[currentInterval="+currentInterval+"], "+
                                               whichValues+"[missing="+nonAuthValues+"])");
if(IsDebug.isDebug) { e.printStackTrace(); }
                throw e;
                }
            }
        catch(final InterruptedException e)
            {
            Thread.currentThread().interrupt(); // Don't lose "interruptedness".
            throw new InterruptedIOException("interrupted trying to grab lock");
            }
        try
            {
            // Try again to get what we can from our local store
            // in case another caller has filled in some values
            // while we were waiting for the lock.
            // We assume that this refetch from local store is cheap
            // (pointer copies) for auth values.
            result = super.getEventValues(def, intervalSelector, intervalNumber, whichValues);

            for(int i = wv.nextSetBit(0); i >= 0; i = wv.nextSetBit(i+1))
                {
                nonAuthValues.set(i,
                                    ((i >= result.length) ||
                                     (result[i] == null) ||
                                     !result[i].isAuthoritative()));
                }

            // If the entire request has now been satisfied from local cache
            // then we can return immediately!
            if(nonAuthValues.length() == 0)
                { return(result); }

            // Make sure that the result value is large enough to accommodate
            // all the results requested.
            // If not,
            // then replace it with a large enough one and copy existing results in.
            final int wvl = nullWV ? 1 : whichValues.length();
            if(result.length < wvl)
                {
                final EventVariableValue newResult[] = new EventVariableValue[wvl];
                System.arraycopy(result, 0, newResult, 0, result.length);
                result = newResult;
                }

//System.out.println("PipelineVarMgr.getEventValues() upstream request: " + nonAuthValues);

            // Make an upstream request for the slots/intervals
            // where our local store did not have authoritative answers.
            final EventVariableValue upstreamValues[];
            try
                {
	            upstreamValues =
	                source.getEventValues(def, intervalSelector, intervalNumber, nonAuthValues);
                }
            catch(final IOException e)
	            {
            	// Take any IOException in response to a request for a local def
            	// as possibly indicative that upstream of us is a tunnel,
            	// so we should (at least for now) assume that the requested values
            	// do not exist or are otherwise unavailable.
            	if(!def.isLocal())
	                { throw e; /* Probably a genuine error. */ }
            	// Just return what we already had to hand.
            	return(result);
	            }

            // Offer the values we got from upstream to our local store
            // and then merge them into our initial response.
            for(int i = nonAuthValues.nextSetBit(0); i >= 0; i = nonAuthValues.nextSetBit(i+1))
                {
                if(i >= upstreamValues.length) { break; }
                final EventVariableValue evv = upstreamValues[i];
                // Skip missing responses.
                if(evv == null) { continue; }
                // Skip dubious claims to auth answers for the current interval or the future (!)
                // which might be the result of a race or clock skew, etc...
                if((evv.getIntervalNumber() >= currentInterval) && evv.isAuthoritative()) { continue; }

//System.out.println("PipelineVarMgr.getEventValues() upstream response "+i+" isAuth="+evv.isAuthoritative());
//            assert((intervalSelector.getIntervalNumber(System.currentTimeMillis()) == intervalNumber-i) || super.getEventValues(def, intervalSelector, intervalNumber-i, _FIRST_BIT)[0].isAuthoritative());

                // Put upstream value into our response.
                result[i] = evv;

                // Offer upstream value to the local store, ie cache it...
                super.setEventValue(evv);
                }

            // OK, respond as well as we can...
            return(result);
            }
        finally { _getEventValues_lock.unlock(); }
        }
    }