package org.hd.d.pg2k.svrCore.vars;

import java.text.DecimalFormat;
import java.util.Iterator;
import java.util.Map;

/**Immutable simple stats derived from a NUMBER globalMap variable.
 * Has a toString() method directly usable on a JSP page.
 * <p>
 * Captures a reference to the value it was constructed from.
 * <p>
 * For most combination operators such as getSum(),
 * values are manipulated as double (as risk of precision loss) to:
 * <ul>
 * <li>Have the maximum-possible dynamic range.
 * <li>Represent integer and floating values.
 * <li>Preserve special IEEE floating-point behaviour, such as for NaN.
 * </ul>
 */
public final class SimpleNumberStats extends SimpleVarStats
    {
    /**Construct a new instance, filtering variable globalMap members by age.
     *
     * @param originalVar  the variable to compute; must be non-null
     *     TYPE_NUMBER (and getValue() must return non-null value)
     * @param oldestAllowed  any globalMap entry than this will
     *     be ignored, which allows only up-to-date information
     *     to be included
     * @param decimalFormatPattern  numeric formatting string to
     *     be used for display of most quantities; never null
     * @param sysID  identifies this system so we can select the local
     *     value of a global variable if available; can be null
     */
    public SimpleNumberStats(final SimpleVariableValue originalVar,
                             final long oldestAllowed,
                             final InstanceID sysID,
                             final String decimalFormatPattern)
        {
        super(originalVar, oldestAllowed, sysID);

        // Post-hoc veto non-numeric or null values.
        if((originalVar == null) ||
           (originalVar.getDef().getType() != SimpleVariableDefinition.TYPE_NUMBER) ||
           (originalVar.getValue() == null))
            { throw new IllegalArgumentException("bad NUMBER argument"); }

        if(decimalFormatPattern == null)
            { throw new IllegalArgumentException("bad pattern string"); }

        pattern = decimalFormatPattern;

        // Create the formatting object once to save time.
        format = new DecimalFormat(decimalFormatPattern);
        }

    /**The pattern String; never null. */
    private final String pattern;

    /**Get the pattern String; never null. */
    public String getPattern()
        {
        return(pattern);
        }

    /**Our computed formatting object; never null.
     * Not immutable, so must nto be passed outside the class.
     */
    private final transient DecimalFormat format;

    /**Gets the total of the non-null values.
     * If there are no usable globalMap values,
     * then this returns the main value, which may be null.
     */
    public Number getSum()
        {
        assert(filteredVar != null);
        final Map globalMap = filteredVar.getGlobalMap();
        if((globalMap == null) || (getNonNullValueCount() < 1))
            { return((Number) filteredVar.getValue()); }

        double result = 0;
        for(final Iterator it = globalMap.values().iterator(); it.hasNext(); )
            {
            final SimpleVariableValue svv = (SimpleVariableValue) it.next();
            assert(svv != null);
            if(svv.getValue() != null)
                { result += ((Number) svv.getValue()).doubleValue(); }
            }
        return(new Double(result));
        }

    /**Gets the maximum of the non-null values.
     * If there are no usable globalMap values,
     * then this returns the main value, which may be null.
     */
    public Number getMax()
        {
        final Map globalMap = filteredVar.getGlobalMap();
        if((globalMap == null) || (getNonNullValueCount() < 1))
            { return((Number) filteredVar.getValue()); }

        Number best = null;
        for(final Iterator it = globalMap.values().iterator(); it.hasNext(); )
            {
            final SimpleVariableValue svv = (SimpleVariableValue) it.next();
            final Number value = (Number) svv.getValue();
            if(value != null)
                {
                if((best == null) ||
                   (value.doubleValue() > best.doubleValue()))
                    { best = value; }
                }
            }
        return(best);
        }

    /**Gets the minimum of the non-null values.
     * If there are no usable globalMap values,
     * then this returns the main value, which may be null.
     */
    public Number getMin()
        {
        final Map globalMap = filteredVar.getGlobalMap();
        if((globalMap == null) || (getNonNullValueCount() < 1))
            { return((Number) filteredVar.getValue()); }

        Number best = null;
        for(final Iterator it = globalMap.values().iterator(); it.hasNext(); )
            {
            final SimpleVariableValue svv = (SimpleVariableValue) it.next();
            final Number value = (Number) svv.getValue();
            if(value != null)
                {
                if((best == null) ||
                   (value.doubleValue() < best.doubleValue()))
                    { best = value; }
                }
            }
        return(best);
        }


    /**Returns HTML- and human- friendly summary of numerical stats.
     * This does <em>not</em> include the variable name.
     * <p>
     * If the count is zero,
     * just returns a formatted version of the value.
     */
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(71);

        sb.append(format.format(getMainValue()));

        final int n = getNonNullValueCount();
        if(n > 0)
            {
            sb.append(", n=").append(n);
            }
        if(n > 1)
            {
            // Order stats values from smallest to largest...
            sb.append(", min=").append(format.format(getMin()));
            sb.append(", max=").append(format.format(getMax()));
            sb.append(", sum=").append(format.format(getSum()));
            }

        return(sb.toString());
        }
    }
