package org.hd.d.pg2k.svrCore.vars;

import java.util.Iterator;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 * User: Damon
 * Date: 14-Aug-2004
 * Time: 18:07:57
 */

/**Generic stats on a variable.
 * More specific instances may be useful in many cases,
 * but this will do for simple cases,
 * and just shows a mapping from participating system to value
 * without interpretting those values.
 */
public class SimpleVarStats
    {
    /**Construct a new instance, filtering variable globalMap members by age.
     *
     * @param originalVar  the variable to compute; must be non-null
     *     but getValue() is permitted to return null
     *     so that we can handle TYPE_NONE values for example
     * @param oldestAllowed  any globalMap entry than this will
     *     be ignored, which allows only up-to-date information
     *     to be included
     * @param sysID  identifies this system so we can select the local
     *     value of a global variable if available; can be null
     */
    public SimpleVarStats(final SimpleVariableValue originalVar,
                          final long oldestAllowed,
                          final InstanceID sysID)
        {
        if(originalVar == null)
            { throw new IllegalArgumentException("null variable not allowed"); }

        filteredVar = originalVar.removeAllKeysOlder(oldestAllowed);
        thisSys = sysID;
        }

    /**This system ID, or null if none selected. */
    private final InstanceID thisSys;

    /**The TYPE_NUMBER variable we were constructed from after filtering; never null. */
    protected final SimpleVariableValue filteredVar;

    /**Get the variable stripped of old globalMap entries. */
    public SimpleVariableValue getFilteredVar()
        {
        return(filteredVar);
        }

    /**Get "main" value; may be null.
     * For a local value it is simply the result of getValue().
     * <p>
     * For a global, if a system ID is supplied and has an entry
     * in the globalMap, its value is used, else the normal getValue()
     * is used as for a local.
     */
    public Object getMainValue()
        {
        final Map globalMap = filteredVar.getGlobalMap();
        if((thisSys != null) &&
           (globalMap != null) &&
           (globalMap.get(thisSys) != null))
            {
            return(((SimpleVariableValue) globalMap.get(thisSys)).getValue());
            }
        else
            { return(filteredVar.getValue()); }
        }

    /**Get count of participating systems (ie globalMap entries); 0 for local variable or no globalMap.
     * A local value will never have globalMap entries;
     * a global value will not if all the entries were too old to use.
     */
    public int getSystemCount()
        {
        final Map globalMap = filteredVar.getGlobalMap();
        if(globalMap == null)
            { return(0); }
        return(globalMap.size());
        }

    /**Get count of globalMap entries with non-null values.
     * This can never be greater than the result of getSystemCount()
     * and never negative.
     */
    public int getNonNullValueCount()
        {
        assert(filteredVar != null);
        final Map globalMap = filteredVar.getGlobalMap();
        if(globalMap == null)
            { return(0); }

        int count = 0;
        for(final Iterator it = globalMap.values().iterator(); it.hasNext(); )
            {
            final SimpleVariableValue svv = (SimpleVariableValue) it.next();
            assert(svv != null);
            if(svv.getValue() != null) { ++count; }
            }
        return(count);
        }

    /**Force deriving class to provide a toString() implementation. */

    /**Returns HTML- and human- friendly summary of stats.
     * This does <em>not</em> include the variable name.
     * <p>
     * If the count is zero,
     * just returns a formatted version of the value.
     * <p>
     * Note that this simply relies on the toString() method
     * of the value type, and shows a direct map of all the values,
     * so may not work well for large numbers of systems or unfriendly values.
     * <p>
     * It may be best to override this for many/most variable types.
     */
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(71);

        sb.append(getMainValue());

        final Map globalMap = filteredVar.getGlobalMap();
        if(globalMap != null)
            {
            final int n = globalMap.size();
            sb.append(", n=").append(n);

            for(final Iterator it = globalMap.entrySet().iterator(); it.hasNext(); )
                {
                final Map.Entry e = (Map.Entry) it.next();
                sb.append(", ");
                sb.append(e.getKey());
                sb.append("-->");
                sb.append(e.getValue());
                }
            }

        return(sb.toString());
        }
    }
