/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import org.hd.d.pg2k.svrCore.MemoryTools;

import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.Collections;
import java.util.EnumSet;
import java.util.Map;
import java.util.Set;


/**Immutable definition of a simple local or global system variable.
 * This defines the type, default value, scope, writability, etc
 * of a single system variable.
 * <p>
 * Some variables may also be events,
 * and gain some extra functionality such as a record of the values over time.
 * <p>
 * Names of variables are in the same scheme as Java properties and class names,
 * ie structured hierarchical alphanumeric ASCII strings,
 * with sections separated by "." dots,
 * and with the more significant components on the left.
 * <p>
 * These variables may be limited in scope to a single JVM,
 * or may be global,
 * or may have local values that are automatically merged into a global version.
 * <p>
 * (We may in future allow these definitions to be provided centrally
 * and distributed via GenProps, for example,
 * possibly merged with or filtered by a built-in set,
 * if we have the need and can avoid security issues for example.)
 * <p>
 * The Comparable ordering is on the name;
 * behaviour is undefined if two different variable definitions
 * with the same name are compared.
 */
public final class SimpleVariableDefinition implements Serializable,
                                                       ObjectInputValidation,
                                                       MemoryTools.Internable,
                                                       Comparable
    {
    /**Absolute maximum different event values that can be held; strictly positive.
     * Individual event variables may impose a smaller limit;
     * this is here to impose a bearable upper limit on system resources.
     * <p>
     * This should probably not be lots smaller than the number of exhibits,
     * for example to allow accurate retention of sparse votes.
     * <p>
     * Preferably a power of two to possibly improve efficiency.
     */
    public static final int EVENT_ABS_MAX_DIFF_VALUES = 65536;

    /**Construct a single variable definition.
     *
     * @param name  the variable name; never null or empty
     * @param type  the variable type; one of the TYPE_XXX values
     * @param persistent  if true then the value persists through a
     *     system/JVM restart if possible; usually false
     * @param readOnly  if true then the variable can never be explicitly set;
     *     it might be synthesised by specialised getter methods, for example
     * @param event  true if this is an event rather than just a simple value
     * @param maxDiffEventCount  maximum number of different event values recorded;
     *     never negative, must be zero for non-event variables
     * @param evPeriodSubset  for an event non-empty set of periods/intervals
     *     for which events are records (null means all periods/intervals);
     *     must be null if not an event
     *
     * @throws java.lang.IllegalArgumentException  if the arguments are invalid
     */
    public SimpleVariableDefinition(final String name,
                                    final int type,
                                    final boolean local,
                                    final boolean persistent,
                                    final boolean readOnly,
                                    final boolean event,
                                    final int maxDiffEventCount,
                                    final EnumSet<EventPeriod> evPeriodSubset)
        throws IllegalArgumentException
        {
        this.name = MemoryTools.intern(name);
        this.type = type;
        this.local = local;
        this.persistent = persistent;
        this.readOnly = readOnly;
        this.event = event;
        this.maxDiffEventCount = maxDiffEventCount;
        // Take defensive (and efficient) read-only copy of set if not null.
        // Also convert full set to null to canonicalise and to save memory.
        this.evPeriodSubset = ((evPeriodSubset == null) || (EnumSet.complementOf(evPeriodSubset).isEmpty())) ? null :
                              EnumSet.copyOf(evPeriodSubset);

        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Simplified constructor setting many of the typical defaults.
     * Essentially a local, read-write, non-persistent value.
     *
     * @param name  the variable name; never null or empty
     * @param type  the variable type; one of the TYPE_XXX values
     *
     * @throws java.lang.IllegalArgumentException  if the arguments are invalid
     */
    public SimpleVariableDefinition(final String name,
                                    final int type)
        throws IllegalArgumentException
        {
        this(name,
            type,
            true,  // Local.
            false, // Non-persistent.
            false, // Read-write.
            false, 0, null); // Not an event.
        }

    /**The variable name; never null or empty. */
    private final String name;

    /**Get the variable name; never null or empty. */
    public String getName()
        {
        return(name);
        }

    /**The variable type; one of the TYPE_XXX values. */
    private final int type;

    /**Get the variable type; one of the TYPE_XXX values. */
    public int getType()
        {
        return(type);
        }

    /**True if the variable value is to persist through restarts. */
    private final boolean persistent;

    /**Get the veriable persistence; defaults to false. */
    public boolean isPersistent()
        {
        return(persistent);
        }

    /**True if the variable is read-only; defaults to false. */
    private final boolean readOnly;

    /**Get the variable writablity; true if read-only. */
    public boolean isReadOnly()
        {
        return(readOnly);
        }

    /**True if the variable is read-only; defaults to false. */
    private final boolean event;

    /**True if this reprents an event rather than a simple value. */
    public boolean isEvent()
        {
        return(event);
        }

    /**Event period/interval subset (immutable) to be collected, or null if not an event or all to be collected.
     * If this is non-null then it is non-empty.
     */
    private final EnumSet<EventPeriod> evPeriodSubset;

    /**Get unmodifiable view of the event period/interval subset to be collected, or null if not an event or all event periods to be collected.
     * Returns a read-only view to protect our internal state.
     */
    public Set<EventPeriod> getEvPeriodSubset()
        {
        if(evPeriodSubset == null) { return(null); }
        return(Collections.unmodifiableSet(evPeriodSubset));
        }

    /**True if the variable is local; defaults to true. */
    private final boolean local;

    /**Get the variable scope; true if local.
     * Local values never get propagated out of their local system (often JVM),
     * and don't get merged into a map,
     * and are thus cheaper and simpler to manage than globalmap values.
     *
     * @return true if this is a local variable
     */
    public boolean isLocal()
        {
        return(local);
        }

    /**Maximum number of different event values retained; defaults to 0. */
    private final int maxDiffEventCount;

    /**Get maximum number of different event values retained; non-negative, zero if not an event */
    public int getMaxDiffEventCount()
        {
        return(maxDiffEventCount);
        }


    /**Put a (high) upper bound on name length in character; strictly positive.
     * This is much longer than we ever expect to use,
     * and is mainly to reduce the possibility of accidents,
     * for example during deserialisation.
     * <p>
     * Limiting this also makes it more likely that we can use an event name
     * as part of a filename on a range of systems, eg *nix and Windows.
     * <p>
     * We will apply this to (String) event values too.
     */
    public static final int MAX_NAME_LEN = 128;


    /**Variable type: "none"; can only be null. */
    public static final int TYPE_NONE = 0;

    /**Variable type: InstanceID; can only be of type InstanceID. */
    public static final int TYPE_IID = 1;

    /**Variable type: String; can only be of type java.lang.String. */
    public static final int TYPE_STRING = 2;

    /**Variable type: Number; can only be of type java.lang.Number.
     * In practice, for security reasons,
     * only java.lang.* implementation types are allowed.
     */
    public static final int TYPE_NUMBER = 3;

    /**Minimum legal TYPE_XXX value. */
    public static final int TYPE__MIN = TYPE_NONE;

    /**Maxiumum legal TYPE_XXX value. */
    public static final int TYPE__MAX = TYPE_NUMBER;

    /**Checks the type of the value passed is valid for this definition.
     * Note that null is always a valid value for any var type.
     *
     * @return true if value is of valid type for this definition,
     *     false otherwise
     */
    public boolean checkType(final Object value)
        {
        // Null is always valid.
        if(value == null)
            { return(true); }

        switch(type)
            {
            case TYPE_NONE: // Can only be null.
                {
                return(false);
                }

            case TYPE_IID: // Can only be InstanceID.
                {
                return(value instanceof InstanceID);
                }

            case TYPE_STRING: // Can only be java.lang.String.
                {
                return(value instanceof java.lang.String);
                }

            case TYPE_NUMBER: // Can only be java.lang impl of java.lang.Number.
                {
                if(value instanceof java.lang.Number)
                    {
                    // For extra security (eg to ensure immutability)
                    // check that this is an immutable JDK impl.
                    return(value.getClass().getName().startsWith("java.lang."));
                    }
                break;
                }
            }

        // Failed to type-check.
        return(false);
        }

    /**Human-readable summary of variable definition. */
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(77);
        sb.append("name=").append(name);
        sb.append(":type=").append(type);
        if(local) { sb.append(":local"); }
        if(persistent) { sb.append(":persistent"); }
        if(readOnly) { sb.append(":readOnly"); }
        if(event) { sb.append(":event"); }
        if(event) { sb.append(":maxDiffEventCount=").append(maxDiffEventCount); }
        return(sb.toString());
        }


    /**Comparable ordering is on name alone in default String order. */
    public int compareTo(Object o)
        {
        return(name.compareTo(((SimpleVariableDefinition) o).name));
        }

    /**Equality is based on the name and all the other parameters; ie equal definitions are identical.
     * This defintion of equals() allows use of MemoryTools.intern().
     * <p>
     * This tests fields in an order likely to give the best performance.
     */
    public boolean equals(Object o)
        {
        if(this == o)
            {
            return true;
            }
        if(!(o instanceof SimpleVariableDefinition))
            {
            return false;
            }

        final SimpleVariableDefinition simpleVariableDefinition = (SimpleVariableDefinition) o;

        if(type != simpleVariableDefinition.type)
            {
            return false;
            }
        if(local != simpleVariableDefinition.local)
            {
            return false;
            }
        if(event != simpleVariableDefinition.event)
            {
            return false;
            }
        if(maxDiffEventCount != simpleVariableDefinition.maxDiffEventCount)
            {
            return false;
            }
        if(persistent != simpleVariableDefinition.persistent)
            {
            return false;
            }
        if(readOnly != simpleVariableDefinition.readOnly)
            {
            return false;
            }
        if(!name.equals(simpleVariableDefinition.name))
            {
            return false;
            }

        if((evPeriodSubset == null))
            {
            if(simpleVariableDefinition.evPeriodSubset != null) { return(false); }
            }
        else if(!evPeriodSubset.equals(simpleVariableDefinition.evPeriodSubset))
            {
            return false;
            }

        return true;
        }

    /**The hash is built on a subset of the fields.
     * This implementation just uses the hash of the name.
     */
    public int hashCode()
        {
        return(name.hashCode());
        }

    /**Deserialise: use constructor for validation, defensive copying, etc.
     * We also use this to eliminate duplicate copies and thus conserve memory.
     * We attempt to effectively intern() the entire definition and also the name.
     *
     * @return identical, de-duped, defensively-copied, non-null instance
     */
    protected Object readResolve()
        throws ObjectStreamException
        {
        // We first try to look up this definition in the
        // canonical set of definitions.
        // If we find it there and the canonical definition is identical
        // (and the canonical set actually exists; during bootstrapping it may be null!)
        // then we return that.
        // However, we do NOT allow deserialisation of a definition with the same name
        // but a different type/flags/etc since that could break type-safety.
        final Map<String,SimpleVariableDefinition> m = SystemVariables.nameToDef;
        if(m != null) // Static initialisation is finished.
            {
            final SimpleVariableDefinition d = m.get(name);

            // Replace with canonical copy if possible.
            if(this.equals(d)) { return(d); }

            // Disallow clashing definitions for type-safety's sake.
            if(d != null)
                { throw new InvalidObjectException("Attempting to deserialise SimpleVariableDefinition with name and type that clash with that of a system variable: " + d + " vs " + this); }
            }

        // Else we resort to more general intern() methods.
        return(MemoryTools.intern(new SimpleVariableDefinition(name,
            type, local, persistent, readOnly, event, maxDiffEventCount, evPeriodSubset)));
        }

//    /**Deserialise. */
//    private void readObject(ObjectInputStream in)
//        throws IOException, ClassNotFoundException
//        {
//        in.defaultReadObject();
//        validateObject(); // Validate state immediately.
//        }

    public void validateObject()
        throws InvalidObjectException
        {
        if((name == null) ||
           (name.length() < 1) || (name.length() > MAX_NAME_LEN))
            { throw new InvalidObjectException("bad object: name must not be null nor empty nor too long"); }
        // Check for basic validity of name.
        // Should look like a fully-qualified plain-ASCII Java class/member name (using dots).
        // Should thus also be safe as a filename component (no slashes for example).
        for(int i = name.length(); --i >=0; )
            {
            final char c = name.charAt(i);
            if((c < 32) || (c > 126))
                { throw new InvalidObjectException("bad object: name contains non-printable ASCII"); }
            if((c != '.') && !Character.isLetterOrDigit(c))
                { throw new InvalidObjectException("bad object: name contains something other than dot or alphanumeric"); }
            }

        if((type < TYPE__MIN) || (type > TYPE__MAX))
            { throw new InvalidObjectException("bad object: invalid type"); }

        if((maxDiffEventCount < 0) || (maxDiffEventCount > EVENT_ABS_MAX_DIFF_VALUES))
            { throw new InvalidObjectException("bad object: invalid maxDiffEventCount"); }
        if(!isEvent() && (maxDiffEventCount != 0))
            { throw new InvalidObjectException("bad object: not event but non-zero maxDiffEventCount"); }
        if(!isEvent() && (evPeriodSubset != null))
            { throw new InvalidObjectException("bad object: not event but non-null evPeriodSubset"); }

        // The set should be non-empty and non-full (a proper sub-set) if non-null.
        if((evPeriodSubset != null) &&
                (evPeriodSubset.isEmpty() || EnumSet.complementOf(evPeriodSubset).isEmpty()))
            { throw new InvalidObjectException("bad object: event set non-null and empty or full"); }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 2361192791747313629L;
    }
