/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import org.hd.d.pg2k.svrCore.MemoryTools;

import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.io.Serializable;
import java.util.*;


/**Immutable value of a simple local or global system variable.
 * This contains the type and value and date-of-change of a system variable.
 * <p>
 * This is Serializable to enable it to be sent over network connections,
 * persisted, etc.
 * <p>
 * The types of all legitimate values are immutable,
 * thus preserving the immutability of instances of this class.
 * <p>
 * This is not Internable because some fields are ignored by equals().
 */
public final class SimpleVariableValue implements Serializable,
                                                  ObjectInputValidation
    {
    /**Construct a single variable value.
     * This type-checks the value, vetoing construction if invalid.
     * <p>
     * This is private for now because though all parameters are checked
     * so it is safe to make this public,
     * we generally only want to add or remove map entries one at a time,
     * and we want to handle timestamp values carefully.
     * <p>
     * We attempt to intern() the value to economise on memory
     * if we think that it will be worthwhile.
     *
     * @param def the variable definition; never null
     * @param value the variable value; null or correct type
     * @param globalMap  if non-null the Map from
     *     InstanceID to SimpleVariableValue;
     *     a private copy is taken in the constructor
     *     and an empty Map is converted to a null
     *
     * @throws java.lang.IllegalArgumentException  if the arguments are invalid
     */
    private SimpleVariableValue(final SimpleVariableDefinition def,
                                final Object value,
                                final Map<InstanceID,SimpleVariableValue> globalMap,
                                final long timeStamp)
        throws IllegalArgumentException
        {
        this.def = def;
        // If worthwhile then intern() the value.
        this.value = ((value != null) && shouldIntern(def)) ? MemoryTools.intern(value) : value;
        // Make a defensive immutable copy of any non-empty Map,
        // else use a null (convert an empty Map to null to save space).
        this.globalMap = ((globalMap == null) || (globalMap.size() == 0)) ? null :
            Collections.unmodifiableMap(new HashMap<InstanceID,SimpleVariableValue>(globalMap));
        this.timestamp = timeStamp;

        try { validateObject(); }
        catch(final InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**Construct a single variable value.
     * This type-checks the value, vetoing construction if invalid.
     *
     * @param def the variable definition; never null
     * @param value the variable value; null or correct type
     *
     * @throws java.lang.IllegalArgumentException  if the arguments are invalid
     */
    public SimpleVariableValue(final SimpleVariableDefinition def,
                               final Object value)
        throws IllegalArgumentException
        {
        this(def, value, null, System.currentTimeMillis());
        }

    /**The variable definition; never null. */
    private final SimpleVariableDefinition def;

    /**Get the variable definition; never null. */
    public SimpleVariableDefinition getDef()
        {
        return(def);
        }

    /**The variable value; null or a type allowed by the definition. */
    private final Object value;

    /**Get the variable definition; null or a type allowed by the definition. */
    public Object getValue()
        {
        return(value);
        }

    /**The time at which the value object was constructed.
     * This value comes with a deserialised object.
     */
    private final long timestamp;

    /**Get the time at which the value object was constructed. */
    public long getTimestamp()
        {
        return(timestamp);
        }

    /**An immutable Map from InstanceID keys to SimpleVariableValues each of the same definition and with no global map.
     * This contains an encapsulation of the constiuent values from each
     * particpating system.
     * <p>
     * It must be null for local variables.
     * <p>
     * For global variables where this is non-null it must contain
     * only keys that are distinct InstanceID values,
     * and only values that are SimpleVariableValues with exactly the same
     * definition and a null globalMap.
     * <p>
     * Is immutable (if non-null) so that it can be returned directly.
     * <p>
     * Never an empty Map; null is used in this case.
     */
    private final Map<InstanceID,SimpleVariableValue> globalMap;

    /**Gets immutable Map from InstanceID to SimpleVariableValue; may be null or non-empty.
     * Maps from the ID of the participating system that provided a value
     * of this global to the value.
     * <p>
     * Never an empty Map.
     */
    public Map<InstanceID,SimpleVariableValue> getGlobalMap()
        {
        return(globalMap);
        }


    /**Returns true if we should probably automatically intern() the value.
     * We will do this is the value is likely to be bulky (eg a String),
     * or may exist in many copies and be long-lived (eg an event).
     */
    private static boolean shouldIntern(final SimpleVariableDefinition def)
        {
        return((def.isEvent() ||
               (def.getType() == SimpleVariableDefinition.TYPE_STRING)));
        }


    /**Human-readable summary of variable value.
     * Simply the raw value rendered as a String.
     * <p>
     * This also makes it easy to use directly in (say) a JSP
     * where if the variable is missing entirely we'll get a benign "null",
     * and to render it simply we don't have to risk calling a method
     * on a possibly null value.
     */
    public String toString()
        {
        return(String.valueOf(value));
        }

    /**Routine to print a full techie-readable description of a variable.
     * This includes the definition, current value, timestamp,
     * and all the globalMap values.
     * <p>
     * This will be a long (single line) of text with values embedded.
     */
    public String getFullDescription()
        {
        final StringBuilder result = new StringBuilder(189);
        result.append("SimpleVariableValue");
        result.append(":def=").append(def);
        result.append(":value=").append(value);
        result.append(":timestamp=").append(timestamp);

        result.append(":globalMap=");
        if(globalMap == null)
            {
            result.append("null");
            }
        else
            {
            final int size = globalMap.size();
            result.append("[size=").append(size);
            for(final Iterator<InstanceID> it = globalMap.keySet().iterator(); it.hasNext(); )
                {
                final InstanceID sys = it.next();
                final SimpleVariableValue svv = globalMap.get(sys);

                result.append(", ");
                result.append("sysID=").append(sys);
                result.append("->value=").append(svv.getFullDescription());
                }
            result.append("]");
            }

        return(result.toString());
        }


    /**Equality is based on the definition (name, type, etc) and value.
     * The timestamp and globalMap values are ignored for the purposes of equality.
     */
    public boolean equals(final Object obj)
        {
        if(!(obj instanceof SimpleVariableValue)) { return(false); }
        final SimpleVariableValue o = (SimpleVariableValue) obj;

        // Compare all fields except the timestamp and globalMap.
        if(value == null) { if(o.value != null) { return(false); } }
        else if(!value.equals(o.value)) { return(false); }
        if(!def.equals(o.def)) { return(false); }

        // All relevant fields seem to match, so objects are equal.
        return(true);
        }

    /**The hash is built on a subset of the fields.
     * This implementation uses the hash of the definition and of the value.
     */
    public int hashCode()
        {
        return(def.hashCode() ^
            ((value != null) ? value.hashCode() : 0));
        }

    /**Make a new value identical except for a new/updated globalMap mapping.
     * The new instance has a globalMap which contains the mapping from
     * client to value, creating the Map if need be,
     * and adding or replacing a mapping for client as required.
     * <p>
     * Both the client and value must be non-null,
     * and the value:
     * <ul>
     * <li>must not have a globalMap
     * <li>must have the same definition (or an equal one).
     * <ul>
     */
    public SimpleVariableValue put(final InstanceID client,
                                   final SimpleVariableValue mapEntryValue)
        {
        return(put(client, mapEntryValue, false));
        }

    /**Make a new value identical except for a new/updated globalMap mapping.
     * The new instance has a globalMap which contains the mapping from
     * client to value, creating the Map if need be,
     * and adding or replacing a mapping for client as required.
     * <p>
     * Both the client and value must be non-null,
     * and the value:
     * <ul>
     * <li>must not have a globalMap
     * <li>must have the same definition (or an equal one).
     * <ul>
     *
     * @param useNewValue  if true, the basic value and timestamp
     *     are taken from the new map entry, rather than from this
     */
    public SimpleVariableValue put(final InstanceID client,
                                   SimpleVariableValue mapEntryValue,
                                   final boolean useNewValue)
        {
        if((client == null) || (mapEntryValue == null))
            { throw new IllegalArgumentException(); }

        if(!def.equals(mapEntryValue.def))
            { throw new IllegalArgumentException("value must have same definition"); }

        // If the new map entry has a globalMap,
        // create a version with it stripped out
        // but with the value and timestamp preserved.
        if(mapEntryValue.globalMap != null)
            {
            mapEntryValue = new SimpleVariableValue(def,
                mapEntryValue.value,
                null,
                mapEntryValue.timestamp);
            }

        // Make a map with any previous values
        // and the new mapping...
        final Map<InstanceID,SimpleVariableValue> newMap = (globalMap == null) ? (new HashMap<InstanceID, SimpleVariableValue>(3)) :
            (new HashMap<InstanceID, SimpleVariableValue>(globalMap));
        newMap.put(client, mapEntryValue);

        return(new SimpleVariableValue(def,
            (useNewValue ? mapEntryValue.value : value),
            newMap,
            (useNewValue ? mapEntryValue.timestamp : timestamp)));
        }

    /**Return instance with all mappings older than specified limit removed.
     * The argument is a time, presumably in the past.
     * <p>
     * Used for trimming away data from dead systems.
     * <p>
     * Returns unchanged any instance with a null or empty globalMap,
     * or where all of the keys in the globalMap are in the specified Set.
     * Else returns a new, modified value.
     *
     * @param oldestAllowed  any key older than this excluded from the result
     */
    public SimpleVariableValue removeAllKeysOlder(final long oldestAllowed)
        {
        // Nothing to remove, so return the original.
        // (Includes local values.)
        if(globalMap == null)
            { return(this); }

        // Make a new map,
        // copying only values whose keys are new enough.
        final Map<InstanceID,SimpleVariableValue> newMap = new HashMap<InstanceID,SimpleVariableValue>(1 + 2 * globalMap.size());
        for(final Iterator<InstanceID> it = globalMap.keySet().iterator(); it.hasNext(); )
            {
            final InstanceID k = it.next();
            final SimpleVariableValue svv = globalMap.get(k);
            if(svv.getTimestamp() >= oldestAllowed)
                { newMap.put(k, svv); }
            }

        // If we didn't actually remove anything then return the original.
        if(newMap.size() == globalMap.size())
            { return(this); }

        // Return new instance.
        return(new SimpleVariableValue(def, value, newMap, timestamp));
        }

    /**A simple comparator that sorts by definition (and thus name) alone.
     * This is useful for ordering values in lists when at most one of each
     * type is present, eg for display or for compression sending over the wire.
     */
    public static final Comparator<SimpleVariableValue> compByDef =
        new Comparator<SimpleVariableValue>()
            {
            public final int compare(final SimpleVariableValue svv1,
                                     final SimpleVariableValue svv2)
                { return(svv1.def.compareTo(svv2.def)); }
            };


    /**Deserialise: use constructor for validation, defensive copying, etc. */
    protected Object readResolve()
        // throws ObjectStreamException
        {
        // Construct new instance of object in the normal defensive way.
        // We preserve the timestamp.
        return(new SimpleVariableValue(def, value, globalMap, timestamp));
        }

    /**Check that the object state is consistent and legal. */
    public void validateObject()
        throws InvalidObjectException
        {
        if(def == null)
            { throw new InvalidObjectException("bad object: def must not be null"); }
        if(!def.checkType(value))
            { throw new InvalidObjectException("bad object: value type is wrong"); }

        // If there is any globalMap data then check it thoroughly.
        // We assume that it has already been defensively copied,
        // so as not to be mutable from outside this instance.
        if(globalMap != null)
            {
            // Ensure no global data in local variable...
            if(def.isLocal())
                { throw new InvalidObjectException("bad object: globalMap data in local variable"); }

            // If non-null,
            // then check Map content.
            for(final Iterator it = globalMap.keySet().iterator(); it.hasNext(); )
                {
                final Object k = it.next();
                if(!(k instanceof InstanceID))
                    { throw new InvalidObjectException("bad object: globalMap key not InstanceID"); }

                final Object d = globalMap.get(k);
                if(!(d instanceof SimpleVariableValue))
                    { throw new InvalidObjectException("bad object: globalMap data not SimpleVariableValue"); }

                final SimpleVariableValue svv = (SimpleVariableValue) d;
                if(!def.equals(svv.def))
                    { throw new InvalidObjectException("bad object: globalMap data of wrong variable type"); }

                if(svv.globalMap != null)
                    { throw new InvalidObjectException("bad object: globalMap data contains other data"); }
                }
            }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -5094323683052972705L;
    }
