/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.vars;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.MemoryTools;

import java.util.*;


/**Immutable enumeration of distributed system variables.
 * To avoid too much recursive definition,
 * variable names (etc) should be defined here and used by the setters, etc.
 */
public final class SystemVariables
    {
    /**Maximum latency we allow in distributing variable values (ms); positive.
     * Downstream and upstream caches should probably be refreshed at
     * a fraction of this delay (say one quarter) to ensure that the
     * round-trip from one slave VM, to the master,
     * and back to the same or another VM is not much more than this.
     * <p>
     * This value should be identical for all participants in a system.
     * <p>
     * A value of a few seconds to a few minutes is probably reasonable
     * to balance network traffic against prompt response to events that affect
     * the whole distributed system.
     * <p>
     * We base this on the maximum permitted inter-peer clock skew.
     */
    public static final int MAX_VALUE_DISTRIBUTION_LATENCY_MS = 2 * CoreConsts.MAX_PEER_CLOCK_SKEW_MS;

    /**Maximum interval between variable updates for system to be considered live (ms).
     * This is between any two variable updates, not simply of the same value.
     * <p>
     * If it longer than this between any variable updates for a given server instance,
     * then other participants may consider that insatnce to be "dead"/resting,
     * and discard knowledge of it and/or variable values automatically.
     * <p>
     * There are other ways of finding out how old data from a particular
     * system is for filtering, but this is a backstop.
     * <p>
     * Should be long enough to survive a temporaily-disconnected system,
     * eg due to a short network outage or a down or disconnected master,
     * but short enough to avoid filling up variable stores/caches
     * with useless stale data from dead subsystems.
     * <p>
     * Must be much longer than MAX_VALUE_DISTRIBUTION_LATENCY_MS.
     * <p>
     * A value of the order of minutes to hours is probably good;
     * and we might add in some small multiple of the normal inter-system skew
     * that we are normally prepared to tolerate.
     */
    public static final int MAX_QUIET_SYSTEM_VAR_LIFE_MS =
        Math.max(50 * MAX_VALUE_DISTRIBUTION_LATENCY_MS,
                 30 * CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000);


    /**Maximum number of samples retained at each interval; strictly positive.
     * A power of two for efficiency,
     * and enough so that at the longest interval we have enough samples
     * to cover at least a year.
     * <p>
     * Most periods/intervals are chosen so that the full span of samples at one interval
     * covers a useful time, ie from over an hour to over a year from VSHORT to VLONG.
     * Indeed, the span is designed to overlap a useful cycle time by a little
     * so that this time last hour/day/year etc can be easily compared with the same place
     * last time round the cycle.
     */
    public static final int EVENT_SAMPLES_RETAINED = 512;

    /**Very-short-term event interval in milliseconds.
     * Of the order of a few seconds.
     * <p>
     * (One use is to make real-time testing of the event system possible.)
     */
    public static final int EVENT_INTERVAL_VSHORT_TERM_MS = 10 * 1000; // 10s

    /**Short-term event interval in milliseconds.
     * Of the order of a few minutes.
     */
    public static final int EVENT_INTERVAL_SHORT_TERM_MS = 5 * 60 * 1000; // 5m

    /**Medium-term event interval in milliseconds.
     * An integer multiple of the next-shorter interval.
     * <p>
     * Of the order of a few tens of minutes.
     */
    public static final int EVENT_INTERVAL_MEDIUM_TERM_MS = 30 * 60 * 1000; // 30m

    /**Long-term event interval in milliseconds.
     * An integer multiple of the next-shorter interval.
     * <p>
     * Of the order of a few hours.
     */
    public static final int EVENT_INTERVAL_LONG_TERM_MS = 2 * 3600 * 1000; // 2h

    /**Very-long-term event interval in milliseconds.
     * An integer multiple of the next-shorter interval.
     * <p>
     * Of the order of tens of hours to days.
     */
    public static final int EVENT_INTERVAL_VLONG_TERM_MS = 24 * 3600 * 1000; // 1D


    /**A read-only, local, "null" variable used to do "keep alives" in variable stores.
     * This contains no data,
     * and may be sent automatically by a slave system
     * to tell the master and other slaves that it is still alive.
     */
    public static final SimpleVariableDefinition KEEP_ALIVE =
        new SimpleVariableDefinition("keepAlive",
                                     SimpleVariableDefinition.TYPE_NONE,
            false, // global
            false, // non-persistent (we want a new one on each restart)
            false, // read/write nominally, though there is no data
            false, 0, null);


    /**The read-only, local, String variable for the local system ID.
     * This value of this (local) variable should be unique to each system,
     * and not propagated out of each system directly.
     */
    public static final SimpleVariableDefinition LOCAL_SYS_ID =
        new SimpleVariableDefinition("localSysID",
                                     SimpleVariableDefinition.TYPE_IID,
            true,  // local (purely for the local machine)
            false, // non-persistent (we want a new one on each restart)
            true, // read-only (the tunnel end-point should inject this)
            false, 0, null);


    /**A read-write global Number variable for testing. */
    public static final SimpleVariableDefinition TEST_NUMBER_GLOBAL =
        new SimpleVariableDefinition("test.NUMBER.global",
                                     SimpleVariableDefinition.TYPE_NUMBER,
            false, // global
            false, // non-persistent
            false, // read/write
            false, 0, null);

    /**A second read-write global Number variable for testing. */
    public static final SimpleVariableDefinition TEST_NUMBER_GLOBAL2 =
        new SimpleVariableDefinition("test.NUMBER.global2",
                                     SimpleVariableDefinition.TYPE_NUMBER,
            false, // global
            false, // non-persistent
            false, // read/write
            false, 0, null);

//    /**A special read-write global Number variable for testing (for the wait routine only). */
//    public static final SimpleVariableDefinition TEST_NUMBER_GLOBAL_WAIT =
//        new SimpleVariableDefinition("test.NUMBER.globalwait",
//                                     SimpleVariableDefinition.TYPE_NUMBER,
//            false, // global
//            false, // non-persistent
//            false  // read/write
//            );

    /**A read-write local Number variable for testing. */
    public static final SimpleVariableDefinition TEST_NUMBER_LOCAL =
        new SimpleVariableDefinition("test.NUMBER.local",
                                     SimpleVariableDefinition.TYPE_NUMBER,
            true,  // local
            false, // non-persistent
            false, // read/write
            false, 0, null);

    /**A second read-write local Number variable for testing. */
    public static final SimpleVariableDefinition TEST_NUMBER_LOCAL2 =
        new SimpleVariableDefinition("test.NUMBER.local2",
                                     SimpleVariableDefinition.TYPE_NUMBER,
            true,  // local
            false, // non-persistent
            false, // read/write
            false, 0, null);

    /**A read-write local String variable for testing. */
    public static final SimpleVariableDefinition TEST_STRING_LOCAL =
        new SimpleVariableDefinition("test.STRING.local",
                                     SimpleVariableDefinition.TYPE_STRING,
            true,  // local
            false, // non-persistent
            false, // read/write
            false, 0, null);

    /**A read-write global String variable for testing. */
    public static final SimpleVariableDefinition TEST_STRING_GLOBAL =
        new SimpleVariableDefinition("test.STRING.global",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            false, // non-persistent
            false, // read/write
            false, 0, null);

    /**A read-write persistent global String event variable for testing. */
    public static final SimpleVariableDefinition TEST_STRING_GLOBAL_EVENT =
        new SimpleVariableDefinition("test.STRING.global.event",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, 100, null); // event with max 100x different values...






    /**Maximum number of different peformance-monitoring (String) stats recordable.
     * This should be high enough to see a reasonably long tail of data,
     * but not so huge as to cripple the system upon transfer or for space.
     */
    public static final int PERFMON_MAX_EVENTS = 203;

    /**Generic read-write non-persistent global String event variable for performance monitoring.
     * Note that this is valid for longer event periods.
     * <p>
     * This is non-persistent since the data is ephemeral
     * and has little value once fixes have been made, etc.
     * <p>
     * We expect this to be very quiet most of the time,
     * except during heavy load or when background sampling is enabled.
     */
    public static final SimpleVariableDefinition PERFMON_STRING_GLOBAL_EVENT =
        new SimpleVariableDefinition("perfmon",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            false, // non-persistent
            false, // read/write
            true, // event
            PERFMON_MAX_EVENTS,
            EnumSet.of(EventPeriod.VLONG, EventPeriod.LONG)); // Long periods.



    /**Maximum number of different generic (String) stats recordable.
     * This should be high enough to accommodate many disjoint sets of values,
     * but not so huge as to cripple the system upon transfer or for space.
     */
    public static final int GENSTATS_MAX_EVENTS = 2003;

    /**Generic read-write persistent global String event variable for general stats.
     * This is used for stats events that do not otherwise have their own event variable.
     * <p>
     * Note that this is valid for all event periods.
     */
    public static final SimpleVariableDefinition GENSTATS_STRING_GLOBAL_EVENT =
        new SimpleVariableDefinition("pg2k.genstats.STRING.global.event",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, // event
            GENSTATS_MAX_EVENTS,
            null); // All event periods.

    /**Generic read-write persistent local String event variable for general stats.
     * This is used for stats events that do not otherwise have their own event variable.
     * <p>
     * Note that this is valid for all event periods.
     * <p>
     * Likely to receive overlap of events with GENSTATS_STRING_GLOBAL_EVENT.
     */
    public static final SimpleVariableDefinition GENSTATS_STRING_LOCAL_EVENT =
        new SimpleVariableDefinition("pg2k.genstats.STRING.local.event",
                                     SimpleVariableDefinition.TYPE_STRING,
            true, // local
            true, // persistent
            false, // read/write
            true, // event
            GENSTATS_MAX_EVENTS,
            null); // All event periods.



    /**ExhibitDataSimpleCache pseudo-variable name prefix. */
    public static final String ExhibitDataSimpleCache_PSEUDOVAR_PREFIX =
        "pg2k.ExhibitDataSimpleCache.";

    /**ExhibitDataSimpleCache cache-percent-full pseudo-variable definition.
     * In the range [0, 100] indicating approximately
     * what percentage of the available cache space is currently in use.
     * <p>
     * If there is more than one ExhibitDataSimpleCache in a local pipeline
     * then behaviour is undefined as all instances will update this value.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Local.
     * <li>Of type Number.
     * <li>Non-peristent.
     * </ul>
     */
    public static final SimpleVariableDefinition ExhibitDataSimpleCache_CACHE_AVAIL_SPACE_PERCENT_USED =
        new SimpleVariableDefinition(ExhibitDataSimpleCache_PSEUDOVAR_PREFIX + "percentUsed",
                                     SimpleVariableDefinition.TYPE_NUMBER);

    /**ExhibitDataSimpleCache cached-exhibit count.
     * In the range [0, +inf[ indicating how many exhibit
     * are fully or partially cached.
     * <p>
     * If there is more than one ExhibitDataSimpleCache in a local pipeline
     * then behaviour is undefined as all instances will update this value.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Local.
     * <li>Of type Number.
     * <li>Non-peristent.
     * </ul>
     */
    public static final SimpleVariableDefinition ExhibitDataSimpleCache_CACHED_EXHIBIT_COUNT =
        new SimpleVariableDefinition(ExhibitDataSimpleCache_PSEUDOVAR_PREFIX + "cachedCount",
                                     SimpleVariableDefinition.TYPE_NUMBER);






    /**ThroughputMonitorFilter pseudo-variable name prefix. */
    public static final String ThroughputMonitorFilter_PSEUDOVAR_PREFIX =
        "pg2k.ThroughputMonitorFilter.";

    /**ThroughputMonitorFilter user-count value.
     * In the range [0,+inf[ indicating approximately
     * how many users are on the system including spiders, hotlinkers, etc.
     * <p>
     * The global map of values can be summed to give the total on the
     * entire distributed system.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-peristent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_CLIENT_COUNT =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "clientCount",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.

    /**ThroughputMonitorFilter "sticky" user-count value.
     * In the range [0,+inf[ indicating approximately
     * how many distinct "sticky" human users are on the system.
     * <p>
     * A "sticky" user is typically one that views at least a couple of pages
     * and stays around for a few minutes.
     * <p>
     * The global map of values can be summed to give the total on the
     * entire distributed system.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-peristent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_STICKY_CLIENT_COUNT =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "stickyClientCount",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.

    /**ThroughputMonitorFilter long-term (smoothed) bytes-per-second value.
     * In the range [0,+inf[ indicating approximately
     * the total download bandwidth on the entire system.
     * <p>
     * This is expressed as an integral quantity for simplicity/robustness,
     * as an Integer if possible for compactness on the wire.
     * <p>
     * The global map of values can be summed to give the total on the
     * entire distributed system.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-peristent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_ltBps =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "ltBps",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.

    /**ThroughputMonitorFilter yearly-unique-visitors value.
     * In the range [0,+inf[ indicating approximately
     * the computed number of unique visitors per year system-wide
     * based on current traffic.
     * <p>
     * The global map of values can be summed to give the total on the
     * entire distributed system.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_YEARLY_UNIQUE_VISITORS =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "yearlyUniqueVisitors",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.

    /**ThroughputMonitorFilter busy fraction (0.0 means idle, 1.0 flat-out busy).
     * In the range [0,1] indicating approximately
     * the fraction of available capacity in use.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_BUSY_FRACTION =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "busyFraction",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false,  // Read/write.
                                     false, 0, null); // Not an event.



    /**ThroughputMonitorFilter available-bandwidth-per-client value in Bps (byte/second).
     * In the range [0,+inf[ indicating approximately
     * the the available bandwidth to a new client
     * based on current traffic.
     * <p>
     * Will be forced to zero if the system is busy or otherwise unwilling
     * to accept new clients.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number.
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_AVAIL_BPS_PER_CLIENT =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "availableBpsPerClient",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, // Read/write.
                                     false, 0, null);  // Not an event.


    /**ThroughputMonitorFilter current AEP hash (a Long).
     * Set by each TMF to indicate which AEP version its server is using.
     * <p>
     * Can be used in load balancing to avoid sending traffic to a mirror
     * which is not on the same version of the AEP.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type Number (ie a Long).
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_AEP_LONGHASH =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "AEPLHash",
                                     SimpleVariableDefinition.TYPE_NUMBER,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, // Read/write.
                                     false, 0, null);  // Not an event.



    /**ThroughputMonitorFilter mirror name value.
     * The string short name of this mirror (CC-suffix) if this machine is a mirror.
     * <p>
     * The global map of values can be used to find which mirrors are active.
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type String.
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition ThroughputMonitorFilter_ACTIVE_MIRROR_NAME =
        new SimpleVariableDefinition(ThroughputMonitorFilter_PSEUDOVAR_PREFIX + "activeMirrorName",
                                     SimpleVariableDefinition.TYPE_STRING,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.





    /**ThroughputMonitorFilter pseudo-variable name prefix. */
    public static final String TunnelServlet_PSEUDOVAR_PREFIX =
        "pg2k.TunnelServlet.";

    /**TunnelServlet slave/client/mirror remote-addresses value (as Strings).
     * String value of each tunnel client (slave) addr as seen (and injected by)
     * the tunnel servlet.
     * <p>
     * These addresses are one of:
     * <ul>
     * <li>null for local tunnels
     * <li>dotted-quad IPv4 addresses for HTTP-tunnelled clients over IPv4
     * </ul>
     * <p>
     * Main attributes:
     * <ul>
     * <li>Globalmap.
     * <li>Of type String.
     * <li>Non-persistent.
     * </ul>
     */
    public static final SimpleVariableDefinition TunnelServlet_SLAVE_ADDRS =
        new SimpleVariableDefinition(TunnelServlet_PSEUDOVAR_PREFIX + "slaveIPs",
                                     SimpleVariableDefinition.TYPE_STRING,
                                     false,  // Global map.
                                     false,  // Not persistent.
                                     false, false, 0, null); // Read/write.



    /**Maximum number of different exhibit name events recorded by most access-pattern stats.
     * Should be large enough not to miss too many one-off events
     * (else we may never see the second-and-subsequent hits for the same value)
     * yet small enough to reasonably limit memory consumption for the stats.
     * <p>
     * We pick a prime-ish value for a larf.
     */
    public static final int MAX_DIFF_EXHIBIT_NAME_VALUES = 1001;

    /**A read-write persistent global String event variable for recording catalogue page views by exhibit.
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_CAT_PAGE_VIEW =
        new SimpleVariableDefinition("AccessPattern.catPageView",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));

    /**A read-write persistent global String event variable for recording completed downloads by exhibit.
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_COMPLETED_DOWNLOAD =
        new SimpleVariableDefinition("AccessPattern.download.completed",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));

    /**A read-write persistent local String event variable for recording completed downloads by exhibit.
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_COMPLETED_DOWNLOAD_LOCAL =
        new SimpleVariableDefinition("AccessPattern.download.completed.local",
                                     SimpleVariableDefinition.TYPE_STRING,
            true, // local
            true, // persistent
            false, // read/write
            true, MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));


    /**A read-write persistent global String event variable for recording ad click-throughs by exhibit.
     * Notes a click through to a paid ad from a catalogue page,
     * where we can capture this information.
     * <p>
     * At best this is likely to be a sampling.
     * <p>
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_CLICKTHROUGH =
        new SimpleVariableDefinition("AccessPattern.clickthru",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));


    /**A read-write persistent global String event variable for recording ad click-throughs by canonicalised URI.
     * Notes a click through to a paid ad from any (valid) page.
     * <p>
     * At best this is likely to be a sampling.
     * <p>
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The URIs are lower-cased (since some are case-insensitive)
     * and limited in length.
     * <p>
     * We allow quite a number of unique values as information is likely
     * to be quite sparse.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_CLICKTHROUGH_BY_URI =
        new SimpleVariableDefinition("AccessPattern.clickthru.byURI",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, 3*MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));

    /**A read-write persistent global String event variable for recording page-views by canonicalised URI.
     * Notes a view of an page (not images, etc).
     * <p>
     * At best this is likely to be a sampling.
     * <p>
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The URIs are lower-cased (since some are case-insensitive)
     * and limited in length.
     * <p>
     * We allow quite a number of unique values as information is likely
     * to be quite sparse.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_PAGEVIEW_BY_URI =
        new SimpleVariableDefinition("AccessPattern.pageview.byURI",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, 3*MAX_DIFF_EXHIBIT_NAME_VALUES, EnumSet.of(EventPeriod.VLONG));


    /**A read-write persistent global String event variable for recording hotlinking to our exhibits by referrer host.
     * This notes where an external site links directly to one of our exhibits (or thumbnails)
     * rather than to one of our HTML pages.
     * <p>
     * We only really need to record the top offenders, not every minor infraction!
     * <p>
     * This is sometimes OK, eg from a bona-fide image search engine,
     * but is often bad, eg bandwidth theft by idle Web site builders/bloggers,
     * so we need to keep an eye on it.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition ACCESSPATTERN_EX_HOTLINK_REF_HOST =
        new SimpleVariableDefinition("AccessPattern.hotlinker",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, 256, EnumSet.of(EventPeriod.VLONG));




    /**Maximum number of different exhibit name events recorded for votes.
     * We regard votes as quite valuable,
     * so will accommodate a larger set of these than for many other events.
     */
    public static final int MAX_DIFF_VOTE_VALUES = 1001;
        /* 1 | (3 * MAX_DIFF_EXHIBIT_NAME_VALUES); */

    /**A read-write persistent global String event variable for recording positive votes by exhibit.
     * Notes a "thumbs-up" vote for the names exhibit.
     * <p>
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition VOTE_PRO =
        new SimpleVariableDefinition("Vote.pro",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true,  // persistent
            false, // read/write
            true,  // event
            MAX_DIFF_VOTE_VALUES, EnumSet.of(EventPeriod.VLONG));

    /**A read-write persistent global String event variable for recording negative votes by exhibit.
     * Notes a "thumbs-down" vote for the names exhibit.
     * <p>
     * Should ideally reflect access by (sane) humans,
     * thus excluding obvious spiders, repeat clickers, etc, if possible!
     * This may well record a representative sampling rather than all events,
     * so rank is more reliable than absolute count.
     * <p>
     * The event values are exhibit short names (eg no directory path),
     * since they save a little memory, should be unique,
     * and may even be a bit more long-lasting than the full name.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition VOTE_CON =
        new SimpleVariableDefinition("Vote.con",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true,  // persistent
            false, // read/write
            true,  // event
            MAX_DIFF_VOTE_VALUES, EnumSet.of(EventPeriod.VLONG));

    /**A read-write persistent global String event variable for recording vote comments by exhibit.
     * Contains the (short) exhibit name
     * followed (after a space) by a sanitised version of a user's comment.
     * <p>
     * This has a maximum of "0" different values since all we are interested in
     * is the audit trail, from which we may hand-pick various items.
     * <p>
     * We are only interested in the longest-term stats.
     */
    public static final SimpleVariableDefinition VOTE_COMMENT =
        new SimpleVariableDefinition("Vote.comment",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, 0, EnumSet.of(EventPeriod.VLONG));


    /**Maximum number of Scorer global "events"; strictly positive.
     * The global event store exists for interchange of "best" Scorers between servers,
     * whereas the local event store additionally acts as persistent storage,
     * and thus the local store should be (much) larger than the global one.
     * (A large global store also implies significant network transmission costs.)
     */
    public static final int AI_SCORER_GLOBAL_MAX_EVENTS = 2003;

    /**Interchange between separate Gallery servers of the "best" Scorers/parameters.
     * Note that this is valid for all event periods.
     */
    public static final SimpleVariableDefinition AI_SCORER_STRING_GLOBAL_EVENT =
        new SimpleVariableDefinition("pg2k.ai.scorer.global.event",
                                     SimpleVariableDefinition.TYPE_STRING,
            false, // global
            true, // persistent
            false, // read/write
            true, // event
            AI_SCORER_GLOBAL_MAX_EVENTS,
            null); // All event periods.

    /**Persistent storage of a large selection of the "best" Scorers/parameters.
     * This is larger than the global set since we persist a larger set of Scorers
     * for a deep local "gene pool", and we don't pay to send this over a network.
     * <p>
     * Note that this is valid for all event periods.
     * <p>
     * Likely to receive overlap of events with AI_SCORER_STRING_GLOBAL_EVENT.
     */
    public static final SimpleVariableDefinition AI_SCORER_STRING_LOCAL_EVENT =
        new SimpleVariableDefinition("pg2k.ai.scorer.local.event",
                                     SimpleVariableDefinition.TYPE_STRING,
            true, // local
            true, // persistent
            false, // read/write
            true, // event
            4*AI_SCORER_GLOBAL_MAX_EVENTS,
            null); // All event periods.




    /**Private array of system variable definitions; non-null. */
    private static final SimpleVariableDefinition vars[] =
        {
        // "System" variables...
        KEEP_ALIVE,
        LOCAL_SYS_ID,

        // Generic stats.
        GENSTATS_STRING_GLOBAL_EVENT,
        GENSTATS_STRING_LOCAL_EVENT,

        // Sub-system-specific stats...
        ThroughputMonitorFilter_CLIENT_COUNT,
        ThroughputMonitorFilter_STICKY_CLIENT_COUNT,
        ThroughputMonitorFilter_ltBps,
        ThroughputMonitorFilter_YEARLY_UNIQUE_VISITORS,
        ThroughputMonitorFilter_ACTIVE_MIRROR_NAME,
        ThroughputMonitorFilter_AVAIL_BPS_PER_CLIENT,
        ThroughputMonitorFilter_BUSY_FRACTION,
        ThroughputMonitorFilter_AEP_LONGHASH,

        // System performance monitoring.
        PERFMON_STRING_GLOBAL_EVENT,

        ExhibitDataSimpleCache_CACHE_AVAIL_SPACE_PERCENT_USED,
        ExhibitDataSimpleCache_CACHED_EXHIBIT_COUNT,

        TunnelServlet_SLAVE_ADDRS,

        ACCESSPATTERN_CAT_PAGE_VIEW,
        ACCESSPATTERN_COMPLETED_DOWNLOAD,
        ACCESSPATTERN_COMPLETED_DOWNLOAD_LOCAL,
        ACCESSPATTERN_CLICKTHROUGH,
        ACCESSPATTERN_CLICKTHROUGH_BY_URI,
        ACCESSPATTERN_PAGEVIEW_BY_URI,
        ACCESSPATTERN_EX_HOTLINK_REF_HOST,

        VOTE_PRO,
        VOTE_CON,
        VOTE_COMMENT,

        AI_SCORER_STRING_LOCAL_EVENT,
        AI_SCORER_STRING_GLOBAL_EVENT,

        // Values used for unit testing.
        // In normal use these should never get set
        // and never be transmitted across tunnels, etc,
        // so should not consume significant resources.
        TEST_NUMBER_GLOBAL, TEST_NUMBER_GLOBAL2,
        TEST_NUMBER_LOCAL,  TEST_NUMBER_LOCAL2,
        TEST_STRING_GLOBAL,
        TEST_STRING_LOCAL,
        TEST_STRING_GLOBAL_EVENT,
        //TEST_NUMBER_GLOBAL_WAIT,
        };
    /**Internalise/canonicalise all the legal system variables.
     * This means that we should be able to eliminate duplicates
     * easily when, for example, deserialing over the wire,
     * thus saving memory at the cost of a small amount of time.
     */
    static
        {
        for(final SimpleVariableDefinition def : vars)
            { MemoryTools.intern(def); }
        }

    /**Immutable SortedSet of system SimpleVariableDefinition items. */
    public static final SortedSet<SimpleVariableDefinition> defs =
        Collections.unmodifiableSortedSet(new TreeSet<SimpleVariableDefinition>(Arrays.asList(vars)));

    /**Immutable Map from String variable name to SimpleVariableDefinition. */
    public static final Map<String,SimpleVariableDefinition> nameToDef;
    /**Initialise nameToDef. */
    static
        {
        final HashMap<String,SimpleVariableDefinition> m = new HashMap<String,SimpleVariableDefinition>(1 + 2 * vars.length);
        for(int i = vars.length; --i >= 0; )
            {
            final SimpleVariableDefinition def = vars[i];
            m.put(def.getName(), def);
            }
        nameToDef = Collections.unmodifiableMap(m);
        }

    /**Immutable Set of String names of variables with slightly sensitive content.
     * Values in these variables should possibly not be published as freely
     * as other system variables, maybe because it contains unedited user input.
     * <p>
     * However, no data in the system is secret or sensitive in a legal sense,
     * so this is only a weak hint!
     */
    public static final SortedSet<String> sensitiveVars =
        Collections.unmodifiableSortedSet(new TreeSet<String>(Arrays.asList(new String[]{
            ACCESSPATTERN_CLICKTHROUGH.getName(),
            ACCESSPATTERN_CLICKTHROUGH_BY_URI.getName(),
            ACCESSPATTERN_PAGEVIEW_BY_URI.getName(),
            VOTE_COMMENT.getName(),
        })));
    }