/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;

import org.hd.d.pg2k.ai.scorer.ScoreAndConf;
import org.hd.d.pg2k.ai.scorer.ScorerIF;
import org.hd.d.pg2k.ai.scorer.fixed.*;
import org.hd.d.pg2k.ai.scorer.parameterised.LocalSampler;
import org.hd.d.pg2k.ai.scorer.parameterised.SimpleExposure;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME.ExhibitTypeParameters;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;

import java.awt.image.*;
import java.io.ByteArrayInputStream;
import java.util.*;

/**Tests of AI-based image scoring.
 */
public final class AIImgScoreTest extends TestCase
    {
    public AIImgScoreTest(String name)
        {
        super(name);
        }

//    /**Do any setup needed for the tests. */
//    protected void setUp()
//        {
//        }

//    /**Do any clearup needed after the tests. */
//    protected void tearDown()
//        {
//        // cleanup code
//        }

    /**Test that simplest "fixed" image scorers return sensible/expected (fixed) values. */
    public static final void testFixedImgScorers()
        throws Exception
        {
        // Construct simple instances of specific Scorers here.
        final NoConfidence noConfidence = new NoConfidence();
        final SimpleExposure simpleExposure = (new SimpleExposure());
        final LocalSampler localSampler = new LocalSampler();
        final ScorerIF scorers[] = { noConfidence, simpleExposure, localSampler };

        final ScoreAndConf sc00 = new ScoreAndConf(0, 0);
        // Should always return a fixed (0-value, 0-confidence) pair.
        assertEquals(noConfidence.computeScoreAndConfidence(null, (String)null), sc00);

        // Attempt to do simple exposure score on null image should throw IAE.
        try
            {
            simpleExposure.computeScoreAndConfidence(null);
            fail("A null image should have been rejected");
            }
        catch(final IllegalArgumentException e) { /* Correctly rejected null/bad image. */ }

        if(!Main.isAccessToFilesystem())
            {
            System.err.println("WARNING: no access to filesystem: skipping some image-scorer tests...");
            return;
            }

        // Try to load the exhibit data, including the specific test samples that we use.
        final ExhibitDataFileSource efds = new ExhibitDataFileSource();
        final AllExhibitProperties aep = efds.getAllExhibitProperties(-1);
        assert(aep.aeid.length > 0);

        // Short names of specific image exhibits that we use.
        final String snWD1 = "water-drops-1-AJHD.jpg"; // Good: very popular water-drop image...
        final String snWD7 = "water-drops-7-AJHD.jpg"; // Good: very popular water-drop image...
        final String sn1pxTransparent = "transparent-single-pixel-ANON.gif"; // Poor: "info-free" image.

        // Check that all test images are present,
        // and make a map from short name to RenderedImage.
        final AllExhibitProperties.ExhibitDataSource eds = efds.makeExhibitDataSource();
        final String allTestImages[] = { snWD1, snWD7, sn1pxTransparent };
        final Map<String, BufferedImage> images = new HashMap<String, BufferedImage>(1+2*allTestImages.length);
        final Map<String, BufferedImage> stdTns = new HashMap<String, BufferedImage>(1+2*allTestImages.length);
        final List<String> allExhibitsSorted = Arrays.asList(aep.aeid.getAllExhibitNamesSorted());
        final String aesAsString = allExhibitsSorted.toString();
        for(final String sn : allTestImages)
            {
            final String fullName = aep.aeid.getFullName(sn);
            assertNotNull("Specific test images must be present: " + aesAsString, fullName);
            final ExhibitTypeParameters exhibitType = ExhibitMIME.getInputFileType(fullName);
            final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(fullName);
            final BufferedImage bi = exhibitType.handler.decodeImage(eds.getInputStream(esa));
            assertNotNull("Decoded exhibit image must not be null", bi);
            images.put(sn, bi);
            if(exhibitType.handler.canMakeThumbnails())
                {
                final ExhibitThumbnails tns = exhibitType.handler.makeThumbnails(esa, eds, aep, true);
                if((tns != null) && (tns.getStandard() != null))
                    {
                    final BufferedImage tnbi = exhibitType.handler.decodeImage(new ByteArrayInputStream(tns.getStandard().toByteArrray()));
                    if(tnbi != null) { stdTns.put(sn, tnbi); }
                    }
                }
            }

        // Output should be (0, 0) for 1-pixel transparent (source) image
        // for simple exposure measure.
        assertEquals(sc00, simpleExposure.computeScoreAndConfidence(images.get(sn1pxTransparent)));
        // Output should be "good" for selected example "good" images.
        assertEquals(Boolean.TRUE, simpleExposure.computeScoreAndConfidence(images.get(snWD1)).isGood());
        assertEquals(Boolean.TRUE, simpleExposure.computeScoreAndConfidence(images.get(snWD7)).isGood());

        // Output should be (0, 0) for 1-pixel transparent (source) image
        // for default local sampler measure.
        assertEquals(sc00, localSampler.computeScoreAndConfidence(images.get(sn1pxTransparent)));
        // Output should be "good" for selected example "good" images.
        assertEquals(Boolean.TRUE, localSampler.computeScoreAndConfidence(images.get(snWD1)).isGood());
//        assertEquals(Boolean.TRUE, localSampler.computeScoreAndConfidence(images.get(snWD7)).isGood());

        // We can try running all scorers on the underlying images
        // and the (larger) thumbnails where available,
        // at least to make sure that they don't go bang.
        for(final String sn : allTestImages)
            {
            // Output should always be the same (0, 0) for the "NoConfidence" scorer.
//            assertEquals(noConfidence.computeScoreAndConfidence(images.get(sn)), sc00);
            final BufferedImage tn = stdTns.get(sn);
//            assertEquals(noConfidence.computeScoreAndConfidence(tn), sc00);

            // Check that nothing bad happens with "SimpleExposure" scorer.
            final ScoreAndConf seSC = simpleExposure.computeScoreAndConfidence(images.get(sn));
            Main.getOut().println("SimpleExposure score for "+sn+" "+seSC);
            assertNotNull(seSC);
            if(tn != null)
                {
                final ScoreAndConf seSCtn = simpleExposure.computeScoreAndConfidence(tn);
                Main.getOut().println("SimpleExposure score for "+sn+" thumbnail "+seSCtn);
                assertNotNull(seSCtn);
                }

            // Check that nothing bad happens with "LocalSampler" scorer.
            final ScoreAndConf lsSC = localSampler.computeScoreAndConfidence(images.get(sn));
            Main.getOut().println("LocalSampler score for "+sn+" "+lsSC);
            assertNotNull(lsSC);
            if(tn != null)
                {
                final ScoreAndConf lsSCtn = localSampler.computeScoreAndConfidence(tn);
                Main.getOut().println("LocalSampler score for "+sn+" thumbnail "+lsSCtn);
                assertNotNull(lsSCtn);
                }
            }

        // Check that we can (safely) perturb all our scorers.
        for(final ScorerIF sc : scorers)
            {
            final String perturbed = sc.getPerturbedNameAndParameters();
            assertTrue("Perturbed Scorer name-and-parameter always starts with base name",
                    perturbed.startsWith(sc.getName()));
            Main.getOut().println("[Sample Scorer name: "+perturbed+".]");
            }

        // Now run all tests on all exhibits for which we can generate a large thumbnail
        // to ensure that we get no crashes.
        // This is optional, and usually disabled for speed.
        if(false)
            {
            Main.getOut().println("Testing on full exhibit set...");
            for(final ExhibitStaticAttr esa : new TreeSet<ExhibitStaticAttr>(aep.aeid.getAllStaticAttrs()))
                {
                final String fullName = esa.filePath;
                final ExhibitTypeParameters exhibitType = ExhibitMIME.getInputFileType(fullName);
                if(!exhibitType.canPossiblyCreateThumbnailOfSameMIMEType()) { continue; }
                final BufferedImage bi;
                try { bi = exhibitType.handler.decodeImage(eds.getInputStream(esa)); }
                catch(final Exception e) { Main.getErr().println("WARNING: error decoding image "+fullName+": " + e.getMessage()); continue; } // Absorb decoding errors.
                if(bi == null) { continue; }
                final ExhibitThumbnails tns = exhibitType.handler.makeThumbnails(esa, eds, aep, true);
                if((tns == null) || (tns.getStandard() == null)) { continue; }
                final BufferedImage tnbi = exhibitType.handler.decodeImage(new ByteArrayInputStream(tns.getStandard().toByteArrray()));
                if(tnbi == null) { continue; }
    
                // Output should always be the same (0, 0) for the "NoConfidence" scorer.
//                assertEquals(noConfidence.computeScoreAndConfidence(tnbi), sc00);
    
                // Output should be non-null and there should be no failures for "SimpleExposure".
                final ScoreAndConf seSCtn = simpleExposure.computeScoreAndConfidence(tnbi);
                Main.getOut().println("SimpleExposure score for "+fullName+" thumbnail "+seSCtn);
                assertNotNull(seSCtn);
                }
            }
        }

//    /**Private source of OK pseudo-random numbers. */
//    private static final Random rnd = new Random();
    }
