/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;

import org.hd.d.pg2k.svrCore.ExhibitPropsGlobalImmutable;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth.AEParams;
import org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth.AEUtils;

import java.awt.*;
import java.io.File;
import java.util.HashMap;
import java.util.Random;

// TODO: test HTTP request path/param parsing/handling.

/**Test of various Aloha Earth components.
 * Includes some tests of Location which could maybe go elsewhere.
 */
public final class AlohaEarthTest extends TestCase
    {
    public AlohaEarthTest(String name)
        {
        super(name);
        }

    /**Do some basic checks on AEParams object.
     * Test such as that we can set parameters away from and to their
     * defaults.
     */
    public void testAlohaEarthParams()
        {
        // Sanity check on some constants.
        assertTrue("INITIAL_ZOOM should be in allowed range",
                   (AEUtils.MIN_ZOOM <=
                        AEUtils.INITIAL_ZOOM) &&
                   (AEUtils.INITIAL_ZOOM <=
                        AEUtils.MAX_ZOOM));

        // Create a new object and ensure that it starts with default values.
        final AEParams aep1 = new AEParams();
        assertTrue("New bean must start off in default state",
                   aep1.inDefaultState());

        // We should start with initial zoom.
        assertTrue("Should start off with initial zoom",
                   aep1.getZoomFactor() == AEUtils.INITIAL_ZOOM);
        // We should start with zero initial offsets.
        assertTrue("Should start off with zero initial offsets",
                   (aep1.getEastOffset(false) == 0) &&
                   (aep1.getNorthOffset(false) == 0));
        // Location should be OE, ON.
        final Location.Estd l1 = aep1.getLocation(true);
        assertTrue("Location should have OE ON centre",
                   (l1.getE().value.intValue() == 0) &&
                   (l1.getN().value.intValue() == 0));
        // Target display pixel of own location should be the centre pixel.
        final Point p = aep1.getDisplayPixelForEstdLocationCentre(l1);
        // We'll be slightly kind about what the centre pixel is.
        assertTrue("Target pixel for display of own location should be display centre: " + p,
                   (Math.abs(p.x - AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH/2) < 2) &&
                   (Math.abs(p.y - AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT/2) < 2));



        // Check that if we set the zoom to something non-initial,
        // we can retrieve it and the parameters are seen as non-default.
        // We'll then try resetting it to its default value again.
        if(AEUtils.INITIAL_ZOOM !=
                        AEUtils.MAX_ZOOM)
            {
            // Do the set via the String-parameter method...
            aep1.setZoomFactor(
                String.valueOf(AEUtils.MAX_ZOOM));
            assertTrue("Must be able to retrieve new zoomFactor",
                       aep1.getZoomFactor() == AEUtils.MAX_ZOOM);
            assertTrue("Must see parameters as non-default",
                       !aep1.inDefaultState());

            // Reset to default zoom via the int parameter method...
            aep1.setZoomFactor(AEUtils.INITIAL_ZOOM);
            assertTrue("Must be able to reset zoom parameter to default",
                       aep1.inDefaultState());
            }


        // FIXME: Do some brief checking of the east/north offsets...


System.err.println("WHOLE_EARTH: " + Location.Estd.WHOLE_EARTH);

        // Check that no zoomed view is no bigger than WHOLE_EARTH.
        // FIXME: rounding errors make this fail at MIN_ZOOM for the moment.
        for(int zf = AEUtils.MIN_ZOOM + 1; zf <= AEUtils.MAX_ZOOM; ++zf)
            {
            final AEParams aeps = (new AEParams()).setZoomFactor(zf);
            assertFalse("No zoomed-out view must be no bigger than WHOLE_EARTH: "+aeps,
                        aeps.getLocation(true).
                            isLargerThan(Location.Estd.WHOLE_EARTH));
            }
        }

    /**Ensure that rectangle and other readable parameters are always in bounds where requested.
     * This is particularly important on edge cases,
     * and were different sets of bounds interact
     * so although individual values might be in range,
     * combinations may not be.
     * <p>
     * We systematically run through a full range of zooms
     * (indeed, nominally beyond the allowed bounds)
     * and try extracting derived values for a range of input offsets.
     * <p>
     * This test is about making sure that things remain within bounds and
     * is not about making sure that they can be usefully set, for example.
     */
    public void testBoundsEnforcement()
        {
        final AEParams aep1 = new AEParams();

        // Pick any old source of random-ish numbers for this.
        final Random rnd = new Random();

        // Set of East offsets we try as test cases,
        // including out-of-bounds and edge cases,
        // and at least one random "joker" value.
        final int eo[] =
            {
            -AEUtils.BASE_2D_EARTH_MAP_WIDTH,
            -AEUtils.BASE_2D_EARTH_MAP_WIDTH/2 - 1,
            -AEUtils.BASE_2D_EARTH_MAP_WIDTH/2,
            -AEUtils.BASE_2D_EARTH_MAP_WIDTH/2 + 1,
            -1, 0, 1,
            AEUtils.BASE_2D_EARTH_MAP_WIDTH/2 - 1,
            AEUtils.BASE_2D_EARTH_MAP_WIDTH/2,
            AEUtils.BASE_2D_EARTH_MAP_WIDTH/2 + 1,
            AEUtils.BASE_2D_EARTH_MAP_WIDTH,
            (rnd.nextInt(4 * AEUtils.BASE_2D_EARTH_MAP_WIDTH) - // Joker...
                2 * AEUtils.BASE_2D_EARTH_MAP_WIDTH)
            };

        // Set of North offsets we try as test cases,
        // including out-of-bounds and edge cases,
        // and at least one random "joker" value.
        final int no[] =
            {
            -AEUtils.BASE_2D_EARTH_MAP_HEIGHT,
            -AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2 - 1,
            -AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2,
            -AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2 + 1,
            -1, 0, 1,
            AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2 - 1,
            AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2,
            AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2 + 1,
            AEUtils.BASE_2D_EARTH_MAP_HEIGHT,
            (rnd.nextInt(4 * AEUtils.BASE_2D_EARTH_MAP_HEIGHT) - // Joker...
                2 * AEUtils.BASE_2D_EARTH_MAP_HEIGHT)
            };

        // Run from beyond the valid bounds of zoom...
        for(int zf = AEUtils.MIN_ZOOM - 1; zf <= AEUtils.MAX_ZOOM + 1; ++zf)
            {
            // Set the zoom.
            aep1.setZoomFactor(zf);

            // Try full range of East Offset values.
            for(int eoi = eo.length; --eoi >= 0; )
                {
                // Set the East Offset.
                aep1.setEastOffset(eo[eoi]);

                // Try full range of North Offset values.
                for(int noi = no.length; --noi >= 0; )
                    {
                    // Set the North Offset.
                    aep1.setNorthOffset(no[noi]);

// Dump assembled parameters.
//System.out.println("Inputs: zf="+zf+", eo="+eo[eoi]+", no=" + no[noi]);
//System.out.println("Parameters computed: " + aep1);

                    // Check that all extractable parameters are in bounds
                    // where so requested.

                    // Check that the zoom behaves correctly.
                    assertTrue("Zoom factor must be in range",
                               (aep1.getZoomFactor() >= AEUtils.MIN_ZOOM) &&
                               (aep1.getZoomFactor() <= AEUtils.MAX_ZOOM));

                    // Location must be in range.
                    final Location.Estd l1 = aep1.getLocation(true);
                    assertTrue("Location must not be null",
                               l1 != null);
                    assertTrue("Location East degrees must be within bounds at all times",
                               (l1.getE().value.doubleValue() >= -Location.Estd.MAX_E) &&
                               (l1.getE().value.doubleValue() <=  Location.Estd.MAX_E));
                    assertTrue("Location North degrees must be within bounds at all times",
                               (l1.getN().value.doubleValue() >= -Location.Estd.MAX_N) &&
                               (l1.getN().value.doubleValue() <=  Location.Estd.MAX_N));

                    assertTrue("East offset must be in bounds when so requested",
                               (aep1.getEastOffset(true) >= -AEParams.OFFSET_LIMIT_EAST) &&
                               (aep1.getEastOffset(true) <=  AEParams.OFFSET_LIMIT_EAST));
                    assertTrue("East offset must be in bounds at all times",
                               (aep1.getEastOffset(false) >= -AEParams.OFFSET_LIMIT_EAST) &&
                               (aep1.getEastOffset(false) <=  AEParams.OFFSET_LIMIT_EAST));

                    assertTrue("North offset must be in bounds when so requested",
                               (aep1.getNorthOffset(true) >= -AEParams.OFFSET_LIMIT_NORTH) &&
                               (aep1.getNorthOffset(true) <=  AEParams.OFFSET_LIMIT_NORTH));
                    assertTrue("North offset must be in bounds at all times",
                               (aep1.getNorthOffset(false) >= -AEParams.OFFSET_LIMIT_NORTH) &&
                               (aep1.getNorthOffset(false) <=  AEParams.OFFSET_LIMIT_NORTH));

                    assertTrue("Source pixels width must be in bounds when so requested",
                               (aep1.getBaseSrcPixelsWidth(true) > 0) &&
                               (aep1.getBaseSrcPixelsWidth(true) <= AEUtils.BASE_2D_EARTH_MAP_WIDTH));
                    assertTrue("Source pixels height must be in bounds when so requested",
                               (aep1.getBaseSrcPixelsHeight(true) > 0) &&
                               (aep1.getBaseSrcPixelsHeight(true) <= AEUtils.BASE_2D_EARTH_MAP_HEIGHT));

                    assertTrue("East degrees must be within bounds when so requested",
                               (aep1.getEastDegrees(true) >= -Location.Estd.MAX_E) &&
                               (aep1.getEastDegrees(true) <=  Location.Estd.MAX_E));
                    assertTrue("East degrees must be within bounds at all times",
                               (aep1.getEastDegrees(false) >= -Location.Estd.MAX_E) &&
                               (aep1.getEastDegrees(false) <=  Location.Estd.MAX_E));

                    assertTrue("North degrees must be within bounds when so requested",
                               (aep1.getNorthDegrees(true) >= -Location.Estd.MAX_N) &&
                               (aep1.getNorthDegrees(true) <=  Location.Estd.MAX_N));
                    assertTrue("North degrees must be within bounds at all times",
                               (aep1.getNorthDegrees(false) >= -Location.Estd.MAX_N) &&
                               (aep1.getNorthDegrees(false) <=  Location.Estd.MAX_N));

                    // The source-image Rectangle must always be in bounds...
                    final Rectangle rect = aep1.getSourceRectangleToDisplay();
                    assertTrue("Source-pixel rectangle must always be in bounds: " + rect + " from: " + aep1,
                               (rect.x >= 0) &&
                               (rect.y >= 0) &&
                               (rect.width <= AEUtils.BASE_2D_EARTH_MAP_WIDTH) &&
                               (rect.height <= AEUtils.BASE_2D_EARTH_MAP_HEIGHT) &&
                               (rect.x + rect.width <= AEUtils.BASE_2D_EARTH_MAP_WIDTH) &&
                               (rect.y + rect.height <= AEUtils.BASE_2D_EARTH_MAP_HEIGHT));

                    // Check that when tiling offset is always multiple of tile.
                    assertTrue("Tile width must be positive",
                               aep1.getTileSrcPixelsWidth() > 0);
                    assertTrue("Tile height must be positive",
                               aep1.getTileSrcPixelsHeight() > 0);
                    if(AEParams.FORCE_TILE_POSITIONING)
                        {
                        assertTrue("Constrained East Offset must be multiple of tile width when tiling: " + aep1.getEastOffset(true),
                            (aep1.getEastOffset(true) % aep1.getTileSrcPixelsWidth()) == 0);
                        assertTrue("Constrained North Offset must be multiple of tile height when tiling: " + aep1.getNorthOffset(true),
                            (aep1.getNorthOffset(true) % aep1.getTileSrcPixelsHeight()) == 0);
                        }
                    }
                }
            }
        }

    /**Test lookup-key generation.
     * Tests that a key is always generated,
     * and that hashCode and equals work sufficiently well that a HashSet
     * of the values correctly distinguishes all different pairs of
     * AEParams values as seen from their zoom and constrained offsets.
     * <p>
     * We also check that these unique keys are one-to-one with unique views
     * as indicated by the source rectangle a view (map fragment)
     * is geenrated from.
     * <p>
     * To avoid running out of memory an patience this does not attempt to
     * generate *all* possible valid values, but rather a finite subset
     * randomly, maybe with a few edge cases thrown in.
     * <p>
     * TODO: Generalise this to test unconstrained keys too
     */
    public void testLookupKey()
        {
        final HashMap keys = new HashMap();   // From key to params...
        final HashMap values = new HashMap(); // From slow-but-right key to params.
        final HashMap views = new HashMap();  // From rect-based-key to params.

        // Any old randomish source will do...
        final Random rnd = new Random();

        // Maximum number of random values to try...
        // A sample set of 100 only seemed to catch it about
        // one time in 2 as at 2003/09/23.
        final int maxRndValues = 1000;

        // We'll create and insert random values.
        for(int i = maxRndValues; --i >= 0; )
            {
            final AEParams aep1 = new AEParams();
            // Set random zoom.
            aep1.setZoomFactor(rnd.nextInt(AEUtils.MAX_ZOOM - AEUtils.MIN_ZOOM + 1) + AEUtils.MIN_ZOOM);
            // Set random offsets.
            aep1.setEastOffset(rnd.nextInt(AEUtils.BASE_2D_EARTH_MAP_WIDTH) - AEParams.OFFSET_LIMIT_EAST);
            aep1.setNorthOffset(rnd.nextInt(AEUtils.BASE_2D_EARTH_MAP_HEIGHT) - AEParams.OFFSET_LIMIT_NORTH);

            // Make a (constrained) key.
            final Object key = aep1.makeKey(true);
            assertTrue("Lookup key must never be null",
                       key != null);
            final Object oldKey = keys.put(key, aep1);

            // Now synthesise a key (String) that should distinguish
            // all correctly-distinguishable pairs.
            // This is inefficient but should be clearly correct.
            final String fakeKey =
                aep1.getZoomFactor() + "#" +
                aep1.getEastOffset(true) + "#" +
                aep1.getNorthOffset(true);
            final Object oldValue = values.put(fakeKey, aep1);

            if((oldKey == null) != (oldValue == null))
                {
                System.err.println("Keys not replaced at same time: about to fail assertion");
                System.err.println("New AEParam: " + aep1);
                System.err.println("Displaced key from: " + oldKey);
                System.err.println("Displaced true duplicate from: "+ oldValue);
                }

            assertTrue("keys must be replaced at the same time (failed after "+(maxRndValues-i)+" goes): " + aep1,
                       (oldKey == null) == (oldValue == null));

            // Make a key from the rectangle parameters
            // to ensure that unique keys map to unique views and vice versa.
            final Rectangle rect = aep1.getSourceRectangleToDisplay();
            final Object oldRect = views.put(rect, aep1);
            assertTrue("views must map one-to-one with keys: " + aep1,
                       (oldKey == null) == (oldRect == null));
            }

        assertTrue("count of distinct values should match",
                   keys.size() == values.size());

        assertTrue("counts of distinct keys and views should match",
                   keys.size() == views.size());

//System.out.println("[Unique keys out of set of random values: " + keys.size() + "/" + maxRndValues + "]");
        }

    /**Test that title lookup works.
     * Test generation of titles is correct.
     */
    public void testTitleGeneration()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        final ExhibitPropsGlobalImmutable epgi = ExhibitPropsGlobalImmutable.loadFromDataDir(new File(LocalProps.getDataDir()));
        final LocationMap lm = epgi.getLocationMap();
        assertFalse("Loaded LocationMap must not be empty", lm.isEmpty());

        // Check that no title ("") is generated at the most-zoomed-out view.
        final AEParams aeps = new AEParams();
        aeps.setZoomFactor(AEUtils.MIN_ZOOM);
        final String topLevelTitle = AEUtils.getViewLocationMapVirtualPrefix(aeps, lm);
        assertEquals("Title at top level (most zoomed out) must be empty: \"" + topLevelTitle + "\"",
                     "", topLevelTitle);
        }


    // FIXME: check that bean results are sorted and correct.
    }
