/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import ORG.hd.d.jIndexer.server.JIndexBean;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**Simple tests of backward compatibility.
 * This tests for things that we explicitly want to keep compatible,
 * or alternatively do not expect nor want to needlessly break.
 * <p>
 * This is also used for testing the handling of accession data,
 * since that is stuff that we need to be able to handle for a long time.
 */
public final class BackCompatTest extends TestCase
    {
    public BackCompatTest(final String name)
        {
        super(name);
        }

    /**If true, print out some compression stats to help with tuning.
     * These stats DO NOT form part of the official test suite.
     */
    private static final boolean CALC_COMPRESSION_STATS = false;
    
    /**Immutable set of usuable (GZIPed) serialised AEP files in date order.
     * Manually kept up-to-date.
     */
    public static final List<String> frozenAEPs;
    /**Initialise frozenAEPs. */
    static
        {
        // GZIPped, serialised AllExhibitProperties exhibit meta-data.
        // Paths relative to the source root.
        frozenAEPs = Collections.unmodifiableList(Arrays.asList(new String[]        {
            "misc/testdata/20041108._ExProps.dat",  // Captured from local server's cache.
            "misc/testdata/20041108.2._expropsCache.dat", // From live server after fixes for more robust serialisation.
            "misc/testdata/20050620._expropsCache.dat", // From live server before accession data expansion.
            "misc/testdata/20051113._ExProps.dat", // Captured from local server's cache.
            "misc/testdata/20060619._ExProps.dat", // Captured from local server's cache.
            "misc/testdata/20060726._ExProps.dat", // Captured from local server's cache.
            "misc/testdata/20060729._ExProps.dat", // Captured from local server's cache after EPC semantics change.
            "misc/testdata/20060731._ExProps.dat", // Captured from local server's cache after compression efficiency changes.
            "misc/testdata/20060818._ExProps.dat", // Captured from local server's cache to match compressed JIndexBean.
            "misc/testdata/20060819._ExProps.dat", // Captured from local server's cache after trouble with AEP diff RPC.
            "misc/testdata/20061122._ExProps.dat", // Static cache file after including treedesc and locationDB.
            }));
        }
    

    /**Check that we can deserialise older snapshots of our exhibit data.
     * This includes snapshots of real Gallery meta-data,
     * and snapshots of test set meta-data.
     * <p>
     * TODO: we could take advantage of multiple CPUs to load these in parallel, which should be safe.
     */
    public static void testAEPDeserialise()
        throws Exception
        {
        final String lastOne = frozenAEPs.get(frozenAEPs.size()-1);

        // List of tasks/loads/deserialisations pending completion.
        final List<Future<?>> tasks = new ArrayList<Future<?>>(frozenAEPs.size());

        // Run the tasks as concurrently as we have CPUs for (and a little more)...
        for(final String filename : frozenAEPs)
            {
            tasks.add(Main.computeIntensiveTestThreadPool.submit(new Runnable(){
                final public void run()
                    {
                    Main.getOut().println("[Loading frozen AEP: "+filename+"...]");
                    final File f = new File(filename);
                    if(!f.isFile() || !f.canRead())
                        {
                        Main.getErr().println("WARNING: skipping test as cannot open AEP snapshot: " + f);
                        return;
                        }

                    try
                        {
                        final Object o = FileTools.deserialiseFromFile(f, true);
                        assertNotNull("Deserialised data from "+f+" must not be null", o);
                        assertTrue("Deserialised data from "+f+" must have correct type",
                                   o instanceof AllExhibitProperties);
                        final AllExhibitProperties aep = (AllExhibitProperties) o;

                        // Do extra-thorough checks on the last (most recent) AEP.
                        if(lastOne.equals(filename))
                            {
                            // Checks that the AEP can be locally serialised and deserialised
                            // without getting broken.
                            SerializationTest.checkSerialisationPreservesEquality(o);

                            // Checks that we can correctly serialise/deserialise
                            // via compression at our highest-available rate
                            // (eg as if for the AEP diff RPC call),
                            // and that compact() doesn't break anything either!
                            final Pair<CompressionLevel, byte[]> squished = GenUtils.compressObject(aep, GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL);
                            final AllExhibitProperties unsquished = (AllExhibitProperties)
                                (new ObjectInputStream(GenUtils.wrapForDecompression(new ByteArrayInputStream(squished.second), squished.first))).readObject();
                            assertEquals("Must be able to (de)serialise and be equal after max (de)compression",
                                            aep, unsquished);
                            final long t1 = System.currentTimeMillis();
                            unsquished.compact();
                            final long t2 = System.currentTimeMillis();
                            Main.getOut().println("[AEP.compact() took "+(t2-t1)+"ms.]");
                            assertEquals("Must be able to compact() and be equal afterwards",
                                            aep, unsquished);
                            }

                        // We can piggy-back some compression/tuning calculations here.
                        if(CALC_COMPRESSION_STATS)
                            {
                            System.out.println("Computing stats for sample: " + f + "; gzipped sample length = " + f.length());
                            System.out.println("  aep.aeid.length (number of exhibits) = " + aep.aeid.length);
                            final ByteArrayOutputStream baos = new ByteArrayOutputStream((int) f.length() * 3);
                            // Write raw form of AEID.
                            (new ObjectOutputStream(baos)).writeObject(aep.aeid);
                            final byte[] rawAEID = baos.toByteArray();
                            baos.reset();
                            System.out.println("  raw AEID serialised length = " + rawAEID.length + "; " + (rawAEID.length / (float) aep.aeid.length) + "bytes/exhibit.");
                            showCompression(rawAEID, System.out);
                            // Write raw form of full AEP.
                            (new ObjectOutputStream(baos)).writeObject(aep);
                            final byte[] rawAEP = baos.toByteArray();
                            baos.reset();
                            System.out.println("  raw AEP serialised length = " + rawAEP.length + "; " + (rawAEP.length / (float) aep.aeid.length) + "bytes/exhibit.");
                            //baos = null; // Help GC.
                            showCompression(rawAEP, System.out);
                            }
                        }
                    catch(final Exception e)
                        {
                        // Report the problem immediately.
                        e.printStackTrace();
                        // Force the Executor mechanism to forward the exception.
                        throw new Error(e);
                        }
                    }
                }));
            }

        // Wait for all tasks to complete,
        // propagating any exception thrown therein.
        for(final Future<?> t : tasks)
            { t.get(); }
        }

    /**Give some indication of compression available in supplied raw data.
     * Produces a map from compression level to compressed data,
     * and can optionally write what it is doing on System.out.
     * <p>
     * The input data is not altered, and is returned in the result
     * under the "NONE" tag.
     * <p>
     * Note that not all compression levels may be implemented/tested,
     * eg due to resource constraints.
     */
    public static Map<CompressionLevel, byte[]> showCompression(final byte raw[],
                                                                final PrintStream out)
        // throws IOException
        {
        final Map<CompressionLevel, byte[]> result = new HashMap<CompressionLevel, byte[]>();
        for(final CompressionLevel level : CompressionLevel.values())
            {
            final long start = System.currentTimeMillis();
            try
                {
                switch(level)
                    {
                    case NONE:
                        {
                        result.put(CompressionLevel.NONE, raw);
                        break;
                        }

                    case ZIP:
                        {
                        final byte zipped[] = FileTools.compressDeflatableData(raw, 0, raw.length);
                        result.put(CompressionLevel.ZIP, zipped);
                        break;
                        }

                    case BZIP2:
                        {
                        final ByteArrayOutputStream baos = new ByteArrayOutputStream(raw.length / 2);
                        final OutputStream os = new org.apache.tools.bzip2.CBZip2OutputStream(baos, 9);
                        os.write(raw);
                        os.close(); // Finish compression.
                        final byte bzipped[] = baos.toByteArray();
                        result.put(CompressionLevel.BZIP2, bzipped);
                        break;
                        }

                    default: // Unsupported method, at least here.
                        break;
                    }
                }
            catch(final Throwable t)
                {
                // Unexpected problem...
                t.printStackTrace();
                System.err.println("HEAP SIZE: " + Runtime.getRuntime().totalMemory());
                }

            if(out != null)
                {
                final byte[] compData = result.get(level);
                if(compData == null)
                    { out.println("    [Compression not supported at level: "+level+".]"); }
                else
                    {
                    final long end = System.currentTimeMillis();
                    final float ratio = compData.length / (float) raw.length;
                    out.println("    Compressed with method "+level+" to "+compData.length+" bytes; ratio=" + ratio + "; time=" + (end-start) + "ms.");
                    }
                }
            }
        return(result);
        }

    /**Test generation of hashes for AccessionData.
     * We test simple streams against some known or previously-computed values.
     */
    @SuppressWarnings("unchecked")
    public static void testAccessionHashes()
        throws Exception
        {
        // Input byte sequences vs expected full-file hashes.
        final Tuple.Pair/*<byte[], Tuple.Pair<Integer,ROByteArray>>*/ testData[] =
            {
            // Zero-length sequence...
            new Tuple.Pair<byte[], Tuple.Pair<Integer,ROByteArray>>(new byte[0],
                new Tuple.Pair<Integer,ROByteArray>(new Integer(0), ROByteArray.fromHexString("d41d8cd98f00b204e9800998ecf8427e"))),

            // Simple non-zero-length sequence... "abc\n"
            new Tuple.Pair<byte[], Tuple.Pair<Integer,ROByteArray>>(new byte[]{0x61, 0x62, 0x63, 0x0a},
                new Tuple.Pair<Integer,ROByteArray>(new Integer(0x4788814e), ROByteArray.fromHexString("0bee89b07a248e27c83fc3d5951213c1"))),

            // Simple non-zero-length sequence... "jim"
            new Tuple.Pair<byte[], Tuple.Pair<Integer,ROByteArray>>(new byte[]{'J', 'i', 'm'},
                new Tuple.Pair<Integer,ROByteArray>(new Integer(91233295), ROByteArray.fromHexString("d54b3c8fcd5ba07e47b400e69a287966"))),

            // Long sequence of zeros...
            new Tuple.Pair<byte[], Tuple.Pair<Integer,ROByteArray>>(new byte[262144],
                new Tuple.Pair<Integer,ROByteArray>(new Integer(0xe20eea22), ROByteArray.fromHexString("ec87a838931d4d5d2e94a04644788a55"))),
            };

        for(int i = testData.length; --i >= 0; )
            {
            final Tuple.Pair<byte[], Tuple.Pair<Integer,ROByteArray>> d = testData[i];

            final Tuple.Pair<Integer,ROByteArray> result =
                AccessionData.computeFullFileHashes(new ByteArrayInputStream(d.first));

            assertEquals("Must compute correct hashes", d.second, result);
            }
        }

    /**Check that we can compute diffs between AEPs successfully.
     * Potentially for every pair of frozen AEPs that we have available,
     * test that we can generate a diff between them
     * and that we can recreate a new AEP using the old AEP and a diff.
     * <p>
     * It is acceptable for the diff routine to refuse to create a diff/delta
     * for whatever reason (unless the force flag is true).
     * 
     * @throws Exception
     */
    public static final void testAEPDiffs()
        throws Exception
        {
        // Limit the amount of time that we spend on this test.
    	// Note that there are a lot of subtle issues that this needs to check for.
        final long stopBy = System.currentTimeMillis() + 1000;

        boolean doneADiff = false; // Set true when a diff has been created/applied.
        boolean testedFinalAEP = false; // Set true when the final AEP has been tested in a diff.
        boolean overrunning; // Set true when we've passed our target deadline and need to finish up ASAP.
        while((!(overrunning = (System.currentTimeMillis() > stopBy))) || !doneADiff || !testedFinalAEP)
            {
//        	if(overrunning) { Main.getOut().println("[Overrunning...]"); }

            final int numFrozenAEPs = frozenAEPs.size();
			// Pick first frozen AEP at random (unless overrunning!).
            final int firstAEPIndex = (!overrunning) ? rnd.nextInt(numFrozenAEPs-1) : (numFrozenAEPs-2);
            final AllExhibitProperties aep1 =
                (AllExhibitProperties) FileTools.deserialiseFromFile(new File(frozenAEPs.get(firstAEPIndex)), true);

            // Pick second frozen AEP to be following instance.
            final int secondAEPIndex = firstAEPIndex + 1;
            // Note that we consider a diff to the final item to be the most intesting.
            final boolean secondIsLatestAEP = (secondAEPIndex == numFrozenAEPs-1);
            final AllExhibitProperties aep2 =
                (AllExhibitProperties) FileTools.deserialiseFromFile(new File(frozenAEPs.get(secondAEPIndex)), true);

            Main.getOut().println("aep1="+aep1+", aep2="+aep2+".");

            final boolean force = overrunning || rnd.nextBoolean();
            final AllExhibitPropertiesDelta aepDiff;
            try { aepDiff = AllExhibitPropertiesDelta.createDiff(aep1, aep2, force); }
            catch(final AllExhibitPropertiesDelta.DiffException e)
                {
                // Refusing to create a diff is OK if the force flag is false.
                System.out.println("Did not create diff: " + e.getMessage());
                if(force) { fail("must create diff if force==true"); }
                continue;
                }

            // Test that the computed diff works correctly.
            final AllExhibitProperties aep2Synth = AllExhibitPropertiesDelta.applyDiff(aep1, aepDiff);
            assertTrue("Result of applying a diff must be non-null",
                            (null != aep2Synth));

            // Detailed look at some of the reconstructed object.
            assertEquals("The epgi must be correctly reconstructed",
                            aep2.epgi, aep2Synth.epgi);
            assertEquals("The aeid must be correctly reconstructed",
                            aep2.aeid, aep2Synth.aeid);
            assertEquals("The longHash must be correctly reconstructed",
                            aep2.longHash, aep2Synth.longHash);

            assertEquals("Result of applying a diff must be equal to the original second AEP",
                            aep2, aep2Synth);
            assertEquals("The 'unchanged-since' timestamp must be preserved",
                            aep2.hashNotChangedSince, aep2Synth.hashNotChangedSince);

            // Show that the diff is not damaged by (de)serialisation.
            final AllExhibitPropertiesDelta aepDiff2 = (AllExhibitPropertiesDelta)
                SerializationTest.checkObjectCanBeSerialisedAndDeserialised(aepDiff);
            final AllExhibitProperties aep2Synth2 = AllExhibitPropertiesDelta.applyDiff(aep1, aepDiff2);
            assertTrue("Result of applying a diff must be non-null",
                            (null != aep2Synth2));
            assertEquals("Result of applying a diff must be equal to the original second AEP",
                            aep2, aep2Synth2);
            assertEquals("The 'unchanged-since' timestamp must be preserved",
                            aep2.hashNotChangedSince, aep2Synth2.hashNotChangedSince);

            // If this was an unforced (realistic) diff
            // (or any diff if we were running out of time),
            // then ensure that the serialised diff is (much) smaller than the AEP.
            if((!force || overrunning) && !aep1.equals(aep2))
                {
                final ByteArrayOutputStream baosDiff = new ByteArrayOutputStream();
                final ObjectOutputStream oosDiff = new ObjectOutputStream(baosDiff);
                oosDiff.writeObject(aepDiff);
                oosDiff.close();
Main.getOut().println("oosDiff size = " + baosDiff.size());
                final ByteArrayOutputStream baosAEP = new ByteArrayOutputStream();
                final ObjectOutputStream oosAEP = new ObjectOutputStream(baosAEP);
                oosAEP.writeObject(aep2);
                oosAEP.close();
Main.getOut().println("oosAEP size = " + baosAEP.size());
                assertTrue("Serialised AEP diff must be smaller than new AEP", baosDiff.size() < baosAEP.size());

                // Created and tested at least one unforced diff.
                doneADiff = true;
                if(secondIsLatestAEP) { testedFinalAEP = true; }
                }
            }
        }
    
    /**Check that the current AEP serialisation is not less compact (when compressed) than the last one persisted.
     * This is a useful test for tuning the serialised format
     * against a realistic/large AEP instance.
     * <p>
     * This is quiet-ish unless the AEP changes in size significantly
     * when re-serialised and GZIPped.  A big enough bloating is an error.
     * <p>
     * Often serialisation formats will compress slightly better
     * when less compact at the micro level.
     * <p>
     * We only test with the last captured AEP, assumed to be the nearest to current.
     * <p>
     * This also measures code performance in the (de)serialisation path.
     */
    public static void testAEPSerialisationSize()
        throws Exception
        {
        final File lastAEPFile = new File(frozenAEPs.get(frozenAEPs.size()-1));
        final long lastAEPSize = lastAEPFile.length();

        Main.getOut().println("Using captured AEP " + lastAEPFile);
        Main.getOut().println("  Size (compressed) on disc (bytes): " + lastAEPSize);
        final AllExhibitProperties lastAEP =
        (AllExhibitProperties) FileTools.deserialiseFromFile(lastAEPFile, true);

////FIXME: TEMPORARY
//    // Extract and dump the largest current single XML and description values.
//    // We use these in other tests.
//    String biggestDesc = "";
//    String biggestMetaData = "";
//    for(final String e : lastAEP.aeid.getAllExhibitNamesSet())
//        {
//        final ExhibitPropsComputable epc = lastAEP.getExhibitPropsComputable(e);
//        if(epc != null)
//            {
//            final String md = epc.getMetadataAsXML();
//            if((null != md) && (md.length() > biggestMetaData.length()))
//                { biggestMetaData = md; }
//            }
//        final ExhibitPropsLoadable epl = lastAEP.getExhibitPropsLoadable(e);
//        if(epl != null)
//            {
//            final String desc = epl.getDescription();
//            if((null != desc) && (desc.length() > biggestDesc.length()))
//                { biggestDesc = desc; }
//            }
//        }
//    System.out.println("Biggest desc length="+biggestDesc.length());
//    System.out.println(biggestDesc);
//    System.out.println("Biggest metadata length="+biggestMetaData.length());
//    System.out.println(biggestMetaData);

        final AllExhibitProperties newAEP = AllExhibitProperties.canonicalise(lastAEP);
        // Serialise to a byte[].
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        // We may gzip the stream to remove some of the redundancy.
        final GZIPOutputStream gos = new GZIPOutputStream(baos);
        final DataOutputStream dos = new DataOutputStream(gos);
        final ObjectOutputStream oos = new ObjectOutputStream(dos);
        oos.writeObject(newAEP);
        oos.flush();
        gos.finish();
        oos.close();
        final byte data[] = baos.toByteArray();

        Main.getOut().println("  SIZE NOW: " + data.length + ", uncompressed: " + dos.size());
        if(data.length > lastAEPSize)
            {
            Main.getErr().println("  FILE NOW LARGER, WHOOPS!");
            assertTrue("compressed serialised AEP size has grown significantly",
                            data.length <= 8192 + ((11*lastAEPSize)/10));
            }

        // Deserialise to try to verify that we really didn't break anything.
        final ByteArrayInputStream bais = new ByteArrayInputStream(data);
        final GZIPInputStream gis = new GZIPInputStream(bais);
        final ObjectInputStream ois = new ObjectInputStream(gis);
        final AllExhibitProperties newAEPSerDeser = (AllExhibitProperties) ois.readObject();
        assertEquals("AEP must survive (de)serialisation in new format", newAEP, newAEPSerDeser);

        // Compute optimal static dictionaries for EPC and EPL (etc)
        // given the number of *unique* instances of each text held in memory,
        // ie assuming that we intern() duplicates
        // and so only save memory for the first instance compressed.
        if(ExhibitPropsComputable.sDict.isStatsEnabled() || ExhibitPropsLoadable.sDict.isStatsEnabled())
            {
            // Find all the unique EPL and EPC texts.
            final Set<String> uniqueEPLTexts = new HashSet<String>(newAEP.aeid.length);
            final Set<String> uniqueEPCTexts = new HashSet<String>(newAEP.aeid.length);

            for(final String name : newAEP.aeid.getAllExhibitNamesSet())
                {
                final ExhibitPropsComputable epc = newAEP.getExhibitPropsComputable(name);
                if(epc != null)
                    {
                    // Capture what the class is internally trying to store/compress.
                    final String s = ExhibitPropsComputable._trimStringFormOfOuterMetadataTag(epc.getMetadataAsXML());
                    if(s != null) { uniqueEPCTexts.add(s); }
                    }
                final ExhibitPropsLoadable epl = newAEP.getExhibitPropsLoadable(name);
                if(epl != null)
                    {
                    final String s = epl.getDescription();
                    if(s != null) { uniqueEPLTexts.add(s); }
                    }
                }

            // Zero out any stats for the static (de)compression dictionaries.
            ExhibitPropsComputable.sDict.resetStats();
            ExhibitPropsLoadable.sDict.resetStats();
            WebUtils.sDictMD.resetStats();

            // Collect/dump stats to help generate good static dictionaries, etc.
            // Note that with a non-empty dictionary in place this shows residual "noise".
            // Skip/ignore items that cannot be compressed with Compact7BitString.

            for(final String s : uniqueEPCTexts)
                { try { Compact7BitString.convertToCompact7BitString(s, ExhibitPropsComputable.sDict); } catch(final IllegalArgumentException e) { } }
            ExhibitPropsComputable.sDict.dumpStats(System.out);

            for(final String s : uniqueEPLTexts)
                { try { Compact7BitString.convertToCompact7BitString(s, ExhibitPropsLoadable.sDict); } catch(final IllegalArgumentException e) { } }
            ExhibitPropsLoadable.sDict.dumpStats(System.out);

            for(final String name : newAEP.aeid.getAllExhibitNamesSet())
                { WebUtils.getCatPageExhibitMetaDataHTMLRaw(name, newAEP); }
            WebUtils.sDictMD.dumpStats(System.out);
            }

        // See how well we can compress a serialised AEP.
        final Pair<CompressionLevel, byte[]> squished = GenUtils.compressObject(newAEP, GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL);
Main.getOut().println("Maximal compression of AEP: scheme="+squished.first+", size="+squished.second.length);


        // Profile total cost of maximal decompression and deserialisation.
        // This reflects the costs at the (more common) client end.
        final ConcurrentHashMap<StackTraceElement, AtomicInteger> perfCounts = new ConcurrentHashMap<StackTraceElement, AtomicInteger>(1001);
        final Thread perfMonitorThread = GenUtils.startThreadPerfMonitor(
            Thread.currentThread(),
            perfCounts,
            "org.", // Capture our code and Apache (BZIP2) code.
            System.currentTimeMillis() + 1000000, // Monitor thread for at most 1000s.
            50); // Sample relatively slowly, esp for WinTel development machine.
        try
            {
            final ObjectInputStream ois2 = new ObjectInputStream(GenUtils.wrapForDecompression(new ByteArrayInputStream(squished.second), squished.first));
            assertEquals("AEP must survive (de)serialisation after maximal compression", newAEP, ois2.readObject());
            }
        finally
            {
            GenUtils.stopPerfMonitorandDumpSamples(
                perfMonitorThread,
                "AEP (de)serialisation profile",
                perfCounts,
                20,
                GenUtils.systemOutLogger);
            }
        }

    /**Historical GZIPped serialised JIndexBean instances, newest last, for checking backward compatibility. */
    public static final List<String> JIB_SER_GZ_FILES = Collections.unmodifiableList(Arrays.asList(new String[]{
        "misc/testdata/_cached_byWord_Index.20060818.ser.gz" /* Matches AEP dump for cross-testing. */,
        }));

    /**Test that deserialisation of old JIB frozen instances is still possible.
     * While we don't promise that the serialised form of a JIndexBean will be long-lived
     * (and it may always be relatively easy to regenerate from current input data)
     * checking that we don't accidentally break something is useful,
     * and this also allows us to check how fast and memory-efficient new versions are.
     */
    public static void testJIBLegacyDeser()
        throws Exception
        {
        // Try to read in each old serialised JIB in turn...
        for(final String fn : JIB_SER_GZ_FILES)
            {
            final File f = new File(fn);
            final long len = f.length();
System.out.println("[Loading JIB from "+f+" length "+len+"...]");
            final AtomicInteger bytesIn = new AtomicInteger();
            final ObjectInputStream ois = new ObjectInputStream(new FilterInputStream(
                new GZIPInputStream(new BufferedInputStream(new FileInputStream(f)))){
                    @Override public boolean markSupported() { return(false); /* Prevent rewinding. */ }
                    @Override public int read() throws IOException
                        {
                        final int result = in.read();
                        if(result >= 0) { bytesIn.incrementAndGet(); }
                        return(result);
                        }
                    @Override public int read(byte[] b, int off, int len) throws IOException
                        {
                        final int result = in.read(b, off, len);
                        if(result > 0) { bytesIn.addAndGet(result); }
                        return(result);
                        }
                    @Override public int read(byte[] b) throws IOException
                        {
                        final int result = in.read(b);
                        if(result > 0) { bytesIn.addAndGet(result); }
                        return(result);
                        }
                    });
            final JIndexBean jib;
            try { jib = (JIndexBean) ois.readObject(); }
            finally { ois.close(); }
System.out.println("  [JIB key count = "+jib.keyCount()+", doc count = "+jib.docCount()+".]");
            jib.compact(); // Force a re-compaction of the JIB.
            final ByteArrayOutputStream baos = new ByteArrayOutputStream(bytesIn.get());
            final ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(jib);
            oos.close();
            System.out.println("  [Input/old JIB raw serialised bytes: in = "+bytesIn.get()+", out = "+baos.size()+".]");
            }
        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
