/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitAttrUtils;
import org.hd.d.pg2k.svrCore.GenUtils;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.webSvr.ads.AdUtils;

import java.util.ArrayList;
import java.util.Map;

/**Simple tests of access to local configuration data.
 */
public final class ConfigFilesTest extends TestCase
    {
    public ConfigFilesTest(String name)
        {
        super(name);
        }

    /**Test that we can find some of the principal files.
     * We may also check that we can parse/use them.
     * <p>
     * We skip these tests if we do not have full/normal access
     * to the filesystem.
     */
    public void testConfigFileAccess()
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        // Check that local properties are loaded.
        assertTrue("Must have access to LocalProps",
                   LocalProps.getTimestamp() > 0);

        // Ensure that we can find key global properties
        // such as the set of exhibit attribute/discardable words....
        assertTrue("Must be able to load exhibit attribute words from global properties",
                   ExhibitAttrUtils.getAttrWords().getAttrWords().length > 0);
//        // ...and the built-in location database.
//        assertTrue("Must be able to load the built-in location DB",
//                   !LocationMap.getBuiltInLocationMap().isEmpty());

        // Check that we've correctly loaded the LocalProps
        // by validating a specific value or two.
        assertEquals("Could not retrieve correct server-slowdown factor from LocalProps",
                     2, LocalProps.getServerSlowdownFactor());
        }

    /**Test access to hostname matching in GenProps.
     * This checks for some data that we know to be there.
     */
    public void testGPHostMatching()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        // Check that we can load test data and that it is non-empty.
        final ExhibitDataFileSource efds = new ExhibitDataFileSource();
        final GenProps gp = efds.getGenProps(-1);
        assertNotNull("Must not retrieve null GenProps", gp);
        assertTrue("Must not retrieve empty GenProps",
                   gp.timestamp != 0);

        // Check that we have hotlink-allow host lists working...
        assertNotNull("We expect the allow-hotlink hosts list to be non-null", gp.getHotLinkAllowHosts());
        assertFalse("We expect the allow-hotlink hosts list to be non-empty", gp.getHotLinkAllowHosts().isEmpty());
        assertEquals("We expect the allow-hotlink hosts list to have right length after conflating normalised entries", 3, gp.getHotLinkAllowHosts().size());
        assertNull("We expect the allow-hotlink regex pattern to be null", gp.getHotLinkAllowHostsRegex());

        // Check that we have hotlink-disallow host lists working...
        assertNotNull("We expect the disallow-hotlink hosts list to be non-null", gp.getHotLinkDisallowHosts());
        assertFalse("We expect the disallow-hotlink hosts list to be non-empty", gp.getHotLinkDisallowHosts().isEmpty());
        assertEquals("We expect the disallow-hotlink hosts list to have right length after conflating normalised entries", 4, gp.getHotLinkDisallowHosts().size());
        assertNotNull("We expect the disallow-hotlink regex pattern to be non-null", gp.getHotLinkDisallowHostsRegex());

        // Check that one paticular set of cases is covered by the regex.
        assertTrue("Must be able to match simple disallow case", gp.getHotLinkDisallowHostsRegex().matcher("profile.myspace.com").matches());
        assertTrue("Must be able to match arbitrary disallow case", gp.getHotLinkDisallowHostsRegex().matcher("new.subDOMAIN-MixEd-caSE9.myspace.COM").matches());


//TODO

        }

    /**Test access to author database in GenProps.
     * This checks for some data that we know to be there.
     */
    public void testAuthDB()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        // Check that we have a standard regex match for disallow hosts working.

        // Check that we can load test data and that it is non-empty.
        final ExhibitDataFileSource efds = new ExhibitDataFileSource();
        final GenProps gp = efds.getGenProps(-1);
        assertNotNull("Must not retrieve null GenProps", gp);
        assertTrue("Must not retrieve empty GenProps",
                   gp.timestamp != 0);

        final GenProps.AuthData adDHD = gp.getAuthData("DHD");
        assertNotNull("Cannot retrieve author data for DHD", adDHD);
        assertEquals("Author initials wrong for DHD",
                     "DHD", adDHD.auth);
        assertEquals("Author name is wrong for DHD",
                     "Damon Hart-Davis", adDHD.name);
        assertEquals("Author home-page URL is wrong for DHD",
                     "http://d.hd.org/", adDHD.www);
        assertEquals("Author email is wrong for DHD",
                     "d@hd.org", adDHD.email);
        assertEquals("Author description is wrong for DHD",
                     "Gallery maintainer.", adDHD.desc);

        // Check that this seems to survive (de)serialisation...
        final GenProps sgp = (GenProps)
            SerializationTest.checkObjectCanBeSerialisedAndDeserialised(gp);
        final GenProps.AuthData adDHD2 = sgp.getAuthData("DHD");
        assertNotNull("Cannot retrieve author data for DHD after (de)serialisation", adDHD2);
        assertEquals("Author initials wrong for DHD after (de)serialisation",
                     "DHD", adDHD2.auth);
        assertEquals("Author name is wrong for DHD after (de)serialisation",
                     "Damon Hart-Davis", adDHD2.name);
        assertEquals("Author home-page URL is wrong for DHD after (de)serialisation",
                     "http://d.hd.org/", adDHD2.www);
        assertEquals("Author email is wrong for DHD after (de)serialisation",
                     "d@hd.org", adDHD2.email);
        assertEquals("Author description is wrong for DHD after (de)serialisation",
                     "Gallery maintainer.", adDHD2.desc);

        assertTrue("Must contain expected hotlink-allowed host",
                   gp.getHotLinkAllowHosts().contains("www.hd.org"));
        assertTrue("Must contain expected hotlink-disallowed host",
                   gp.getHotLinkAllowHosts().contains("bad.host1"));

        // Check that equality works too.
        assertEquals("AuthData should compare equals after (de)serialisation",
                     adDHD, adDHD2);
        }

    /**Test access to generic key/value pairs in GenProps.
     * This checks for some data that we know to be there.
     */
    public void testGen()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        // Check that we can load test data and that it is non-empty.
        final ExhibitDataFileSource efds = new ExhibitDataFileSource();
        final GenProps gp = efds.getGenProps(-1);
        assertNotNull("Must not retrieve null GenProps", gp);
        assertTrue("Must not retrieve empty GenProps",
                   gp.timestamp != 0);

        // Check that generic map is not null.
        final Map<String,String> gen = gp.getGen();
        assertNotNull("Gen map must not be null", gen);

        if(gen.isEmpty())
            {
            System.err.println("gen = " + (new ArrayList(gen.entrySet())));
            fail("Gen map must not be empty (should contain test value(s))");
            }

        // Check that the generic map is not (trivially) mutable.
        try
            {
            gen.put("a", "b");
            fail("Gen map is apparently mutable");
            }
        catch(final UnsupportedOperationException e) { /* Good; rejected change. */ }

        // Get generic test value...
        final String testKey = "testGen.key";
        final String testValue = "testGenValue";
        assertEquals("Must be able to retrieve expected test value", testValue, gen.get(testKey));

        // Look for ECPM value...
        assertEquals("Must be able to look up ECPM value", 10, AdUtils.computeECPM(gp, "kPMPILV=MRTF"));

        // Check for "sensitive" names.
        final AllExhibitProperties emptyAEP = new AllExhibitProperties();
        assertTrue("Must be able to detect sensitive names correctly [1]", GenUtils.isSensitive("a/an-obituary-ANON.tif", emptyAEP, gp));
        assertTrue("Must be able to detect sensitive names correctly [2]", GenUtils.isSensitive("a/some-death-notice-mono-ANON.gif", emptyAEP, gp));
        assertFalse("Must be able to detect sensitive names correctly [3]", GenUtils.isSensitive("a/my-cat-ANON.gif", emptyAEP, gp));

        // Check that this seems to survive (de)serialisation...
        final GenProps sgp = (GenProps)
            SerializationTest.checkObjectCanBeSerialisedAndDeserialised(gp);
        final Map<String,String> gen2 = sgp.getGen();
        assertEquals("Gen maps must be equal after (de)serialisation", gen, gen2);
        assertEquals("Must be able to retrieve expected test value after (de)serialisation", testValue, gen2.get(testKey));
        assertEquals("Must be able to look up ECPM value after (de)serialisation", 10, AdUtils.computeECPM(sgp, "kPMPILV=MRTF"));
        }

    /**Ensure that we can load the exhibits and meta data from the filesystem.
     * This uses the ExhibitDataFileSource.
     * <p>
     * We skip these tests if we do not have full/normal access
     * to the filesystem.
     */
    public void testLoadOfExhibitData()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        // Check that we can load test data and that it is non-empty.
        final ExhibitDataFileSource efds = new ExhibitDataFileSource();
        final GenProps gp = efds.getGenProps(-1);
        assertNotNull("Must not retrieve null GenProps", gp);
        assertTrue("Must not retrieve empty GenProps",
                   gp.timestamp != 0);
        final AllExhibitProperties aep =
            efds.getAllExhibitProperties(-1);
        assertNotNull("Must not retrieve null AllExhibitProperties", aep);
        assertTrue("Must not retrieve empty AllExhibitProperties",
                   aep.aeid.size() > 0);
        assertTrue("Must retrieve non-empty global-immutable properties",
                   aep.epgi.longHash != 0);
//        assertTrue("Must retrieve non-empty LocationMap",
//                   aep.lm.hashCode() != 0);
        }
    }
