/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.GenUtils;
import org.hd.d.pg2k.svrCore.props.GenProps;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 27-Jul-2003
 * Time: 17:37:14
 */

/**Tests various features of exhibit name syntax, parsing, etc.
 * This should be extended to test all the various conversions and parses
 * and comparators from ExhibitName.
 */
public final class ExhibitNameTest extends TestCase
    {
    /**Test the "valid syntax" methods.
     * Created after enabling assert()s revealed internal inconsistencies.
     */
    public void testValidSyntaxMethods()
        {
        final String minFinalNameComponent = "b-C.d";
        final String minName ="a/" + minFinalNameComponent;

        // Check minimal valid name is regarded as valid.
        assertTrue("Minimal valid final name component must be seen as valid",
                   ExhibitName.validNameFinalComponentSyntax(minFinalNameComponent));
        assertTrue("Minimal valid name must be seen as valid by ExhibitName.validNameSyntaxBasic()",
                   ExhibitName.validNameSyntaxBasic(minName));
        assertTrue("Minimal valid name must be seen as valid by ExhibitName.validNameSyntax()",
                   ExhibitName.validNameSyntax(minName));
        }

    /**Test generation of "virtual" names.
     * Simple tests that names are "virtualised" correctly.
     */
    public void testVirtualNameGeneration()
        {
        assertSame("Already-virtual name must be returned intact",
                   "a/b-C.d",
                   ExhibitName.getVirtualExhibitName("a/b-C.d"));
        assertEquals("Name must be correctly virtualised",
                     "a/B-C.def",
                     ExhibitName.getVirtualExhibitName("a/_more1999/_more01/B-C.def"));
        }

    /**Test extraction of main-words component of a name.
     * Does some very basic testing of the ability to extract the main words.
     * <p>
     * These tests are entirely syntactic,
     * ie do not require particular exhibit names or meanings.
     */
    public void testMainWordExtraction()
        {
        // Test set of attribute words, were used.
        final Set<String> attrWords = new HashSet<String>(Arrays.asList(
            new String[]{ "mono", "false", "colour", "rotated" }));

        // Set of values to test:
        //   * Column 1 is the exhibit names.
        //   * Column 2 is the main words fragment expected with no attribute words.
        //   * Column 3 is the main words fragment expected with attrWords.
        final String testData[][] =
            {
                { "a/b-C.d", "b", "b" },
                { "a/b-3-C.d", "b", "b" },
                { "a/b-mono-3-C.d", "b-mono", "b" },
                { "a/b-false-colour-05-C.d", "b-false-colour", "b" },
                { "a/b-dog-Cat-zoo-apple-false-colour-05-C.d", "b-dog-Cat-zoo-apple-false-colour", "b-dog-Cat-zoo-apple" },
            };

        // Work through all the test cases.
        for(int i = testData.length; --i >= 0; )
            {
            final String exhibitName = testData[i][0];
            final String noAWParse = testData[i][1];
            final String aWParse = testData[i][2];
            final Set<String> noAttrWords = Collections.emptySet();
            final String actualNoAWParse = ExhibitName.getMainWordsComponent(exhibitName, noAttrWords);
            assertTrue("unexpected results without attr words: got `"+actualNoAWParse+"', expected `"+noAWParse+"'",
                       noAWParse.equals(actualNoAWParse));
            final String acutalAWParse = ExhibitName.getMainWordsComponent(exhibitName, attrWords);
            assertTrue("unexpected results with attr words: got `"+acutalAWParse+"', expected `"+aWParse+"'",
                       aWParse.equals(acutalAWParse));
            }
        }

    /**Test isSensitive() exhibit-flagging mechanism.
     */
    public void testIsSensitive()
        {
        final String SENSITIVE_PHRASE = "-Monty-Python-";

        // Synthetic sensitive name.
        final String sensName = "memes/some-prefix" + SENSITIVE_PHRASE + "some-suffix-ANON.jpg";

        // Set up empty AEP.
        final AllExhibitProperties dummyAEP = new AllExhibitProperties();

        // Set up minimal GP with a tset "sensitive" value.
        final Properties p = new Properties();
        p.setProperty(GenProps.GEN_PREFIX + GenUtils.GP_SENSITIVE_NAME_SUBSTRS_KEY, SENSITIVE_PHRASE);
        final GenProps gp = new GenProps(p, System.currentTimeMillis());

        // Check sensitive values against default GP.
        assertTrue("Sensitive exhibit name must be flagged as such: " + sensName, GenUtils.isSensitive(sensName, dummyAEP, gp));
        assertFalse("Non-sensitive exhibit name must not be flagged as sensitive", GenUtils.isSensitive("a/a-A.a", dummyAEP, gp));
        }
    }
