/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.uploader.UploadInfoBean;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;
import org.hd.d.pg2k.webSvr.upload.HTTPUploaderUtils;

import javax.servlet.ServletInputStream;
import java.io.*;
import java.util.Random;

/**Tests the exhibit upload mechanisms.
 * May need to be running in a WAR container and in master mode
 * for all tests to be operable
 * (unless, exceptionally, the loopbackURL is pointing at another instance).
 */
public final class ExhibitUploadTest extends TestCase
    {
    public ExhibitUploadTest(String name)
        {
        super(name);
        }

    /**The upload directory, or null if none set. */
    private File uploadDir;

    /**Do any setup needed for the tests. */
    protected void setUp()
        throws Exception
        {
        // Empty and recreate the upload directory if defined.
        final String uploadDirS = LocalProps.getUploadDir();
        uploadDir = (uploadDirS == null) ? null : new File(uploadDirS);
        if(uploadDir != null)
            {
            if(uploadDir.exists())
                { FileTools.rmRecursively(uploadDir); }

            // Try to ensure that it exists...
            uploadDir.mkdirs();
            }
        }

    /**Do any clearup needed after the tests. */
    protected void tearDown()
        throws Exception
        {
        // Zap the upload directory when done.
        if(uploadDir != null)
            {
            if(uploadDir.exists())
                { FileTools.rmRecursively(uploadDir); }
            }
        }


    /**Test simple upload (of non-binary data) directly, not over HTTP.
     * Generate a new random HTML text-fragment file,
     * with a new, valid name,
     * and upload it simulating a browser POST transaction,
     * checking that it arrives intact in the upload directory.
     */
    public void testDirectUploadSimple()
        throws Exception
        {
        // If no upload dir, we can't do the test.
        if(uploadDir == null)
            {
//            System.err.println("WARNING: cannot use/create/empty upload dir so skipping test");
            fail("WARNING: cannot use/create/empty upload dir so skipping test");
            return;
            }

        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        final UploadInfoBean uib = _makeTestUploadUIB(edfs, ".htxt");

        // Create the full name...
        final String newExhibitName = uib.getFullName();

        // Full name at which we expect upload to appear...
        final File targetFileName = new File(uploadDir, newExhibitName);

        // Ensure that it does not currently exist in the upload area...
        assertFalse("File to upload must not exist in upload area before we start!",
                    targetFileName.exists());

        // Generate a random HTML text fragment, legal as an exhibit.
        final String exhibitText =
            "<b>The cat sat on the mat " + rnd.nextInt() + " times.</b>";

        // Chose a boundary string; would normally be chosen randomly...
        final String boundary = "------------bdYG0aDRc9XTBFOYCIHkaq";

        // Input lines from to servlet expected with the above posted.
        final String POSTlines[] =
            {
boundary,
"Content-Disposition: form-data; name=\"" + UploaderConsts.exhibitFile + "\"; filename=\"" + newExhibitName + "\"",
"Content-Type: application/octet-stream",
"",
exhibitText,
boundary + "--"
            };

/*
------------CZwvTEJ33GJISQKwfRxy4g
Content-Disposition: form-data; name="exhibitFile"; filename="sample.htxt"
Content-Type: application/octet-stream

<pre>
The cat sat on the mat.
</pre>
------------CZwvTEJ33GJISQKwfRxy4g--
*/

        // This stream version allows *only* the line-read, not byte reads.
        final ServletInputStream is = new ServletInputStream(){
            int l;

            public int readLine(byte[] buf, int off, int len)
                // throws IOException
                {
                if(l >= POSTlines.length)
                    { return(-1); } // EOF

                // Object if buffer not big enough to read line in one go,
                // or if offset is not zero.
                // May refine this to be more forgiving later...
                final String t = POSTlines[l++];
                final int tl = t.length();
                if((off != 0) || (tl+2 > len))
                    { throw new IllegalArgumentException("only simple uses permitted"); }

                // Else copy lower bytes into place.
                // We only use ASCII text in this test.
                for(int i = tl; --i >= 0; )
                    { buf[i] = (byte) t.charAt(i); }
                buf[tl] = '\r';
                buf[tl+1] = '\n';

                return(tl+2);
                }

            public int read() // throws IOException
                { throw new IllegalStateException(); }
            };

        // Do upload...
        HTTPUploaderUtils.doUpload(uib,
                             uploadDir,
                             is,
                             boundary,
                             Integer.MAX_VALUE,
                             GenUtils.systemOutLogger);

        assertTrue("Upload destination directory must exist (or have been created)",
                   targetFileName.getParentFile().isDirectory());
        assertTrue("File has not been uploaded and it should have been, to: " + targetFileName,
                   targetFileName.exists());
        assertTrue("Target of upload must be a readable file",
                   targetFileName.isFile() && targetFileName.canRead());
        assertEquals("Uploaded file must have the correct length",
                     exhibitText.length(), targetFileName.length());
        // Ignore whitespace (such as line-end details) for simple content check.
        assertEquals("Uploaded file must have the correct content",
                     exhibitText.trim(), FileTools.readTextFile(targetFileName).trim());
        }

    /**Makes a new random, legal UploadInfoBean for test purposes; never null.
     * Creates a unique new random exhibit name for uploading to,
     * picking one of the existing categories and authors from the current live
     * data set.
     * <p>
     * May cause a jUnit assertion failure if it cannot do this successfully.
     *
     * @param p  data pipeline for test data set;
     *     must have at least one author and one category we can use
     * @param suffix  must be legal non-null suffix for the file type
     *     to be uploaded, eg ".htxt" for HTML text, ".jpg" for a JPEG image
     * @return correctly initialised UploadInfoBean; never null
     * @throws IOException  in case of difficulty with the data source
     */
    private static UploadInfoBean _makeTestUploadUIB(final SimpleExhibitPipelineIF p,
                                                     final String suffix)
        throws IOException
        {
        // Collect the raw exhibit info from the underlying file system.
        final AllExhibitProperties aep = p.getAllExhibitProperties(-1);

        // Pretend that nothing else has been uploaded.
        final AllExhibitProperties aepUpload = new AllExhibitProperties();

        // Create a new, random but-legal exhibit name to upload as.
        // Pick first extant author name.
        final String author =
                aep.getAuthorExhibitCounts().keySet().iterator().next();
        // Pick first extant category.
        final String category =
            aep.getCategoryExhibitCounts().keySet().iterator().next();

        // Synthesise information loaded into form.
        final UploadInfoBean uib = new UploadInfoBean();
        uib.setAep(aep);
        uib.setUploadAeid(aepUpload.aeid);
        uib.setCategory(category);
        uib.setMainWords("new-test-text");
        uib.setNumber(rnd.nextInt() >>> 1); // Non-negative number-in-series.
        uib.setAuthor(author);
        uib.setSuffix(suffix);

        assertTrue("Must have enough info to make a full name",
                   uib.enoughInfo());
        assertTrue("Must have enough valid info to make a full unique name",
                   uib.enoughValidUniqueInfo());

        return(uib);
        }

    /**Minimum binary upload speed in Bps (bytes per second); strictly positive.
     * The pre-200401020 algorithm could usually manage about 10kBps
     * when uploading into a slow, remote (NFS-mounted) filesystem
     * (with JDK 1.3.1 -server so that the JVM was not the bottleneck).
     * So we enforce a minimum of at least 10kBps.
     */
    public static final int MIN_UPLOAD_SPEED_BYTESPERSEC = 10240;

    /**Test direct binary upload (of a JPEG image) for correctness.
     * We also measure and print the overall speed of upload which can be tuned,
     * though the performance threshold enforced is low enough
     * to allow even slow development machines to meet it.
     */
    public void testDirectUploadJPEG()
        throws Exception
        {
        // If no upload dir, we can't do the test.
        if(uploadDir == null)
            {
//            System.err.println("WARNING: cannot use/create/empty upload dir so skipping test");
            fail("WARNING: cannot use/create/empty upload dir so skipping test");
            return;
            }

        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        final UploadInfoBean uib = _makeTestUploadUIB(edfs, ".jpg");

        // Create the full name...
        final String newExhibitName = uib.getFullName();

        // Full name at which we expect upload to appear...
        final File targetFileName = new File(uploadDir, newExhibitName);

        // Ensure that it does not currently exist in the upload area...
        assertFalse("File to upload must not exist in upload area before we start!",
                    targetFileName.exists());

        // Create a new, random file-encoded JPEG.
        // Large enough to provide a reasonable performance measure too.
        final byte[] image = MediaHandlerTest.makeTestRGBTrueColourJPEGImage();
        assertNotNull("Must be able to create a JPEG image", image);
        assertTrue("Must be able to create non-zero-length JPEG image", 0 != image.length);
        assertTrue("JPEG image must be reasonable size to test upload performance (>1sec at minimum acceptable speed): " +
                        image.length+" vs "+MIN_UPLOAD_SPEED_BYTESPERSEC,
                    image.length > MIN_UPLOAD_SPEED_BYTESPERSEC);
        System.out.println("Test JPEG upload image size (bytes): " + image.length);

        // Gather up output stream for POST command...
        final ByteArrayOutputStream baos = new ByteArrayOutputStream(image.length + 1024);

        // Choose suitable random boundary string.
        final String boundary = "------------" + Long.toString(rnd.nextLong() >>> 1, 36);

        // Actually generate the POST body bytes...
        HTTPUploaderUtils.generateUploadPOSTBody(uib, baos, image, boundary);
        final byte body[] = baos.toByteArray();
        assertTrue("Extracted body cannot be zero length", body.length > 0);

        final ServletInputStream is = new ServletInputStream(){
            /**Position in body[]. */
            private int pos;

            /**Reads the next byte of data from the input stream, or -1 at EOF.
             */
            final public int read() // throws IOException
                {
                if(pos >= body.length) { return(-1); }
                return(body[pos ++] & 0xff);
                }
            };

        // Do upload (and time it)...
        final long uploadStart = System.currentTimeMillis();
        HTTPUploaderUtils.doUpload(uib,
                             uploadDir,
                             is,
                             boundary,
                             Integer.MAX_VALUE,
                             GenUtils.systemOutLogger);
        final long uploadStop = System.currentTimeMillis();
        final long uploadTime = uploadStop - uploadStart;

        System.out.println("Upload time (ms): " + uploadTime);
        System.out.println("Upload speed (kByte/s): " + (((float) image.length)/uploadTime));

        assertTrue("Upload destination directory must exist (or have been created)",
                   targetFileName.getParentFile().isDirectory());
        assertTrue("File has not been uploaded and it should have been, to: " + targetFileName,
                   targetFileName.exists());
        assertTrue("Target of upload must be a readable file",
                   targetFileName.isFile() && targetFileName.canRead());
        assertEquals("Uploaded file must have the correct length",
                     image.length, targetFileName.length());
        // Check that the file has arrived intact.
        final InputStream fis =
            new BufferedInputStream(new FileInputStream(targetFileName));
        try
            {
            for(int i = 0; i < image.length; ++i)
                {
                assertEquals("Uploaded file must match byte-by-byte (failed@"+i+")",
                             image[i] & 0xff, fis.read());
                }
            }
        finally { fis.close(); } // Free resources...
        }

    /**Verify that bad uploads are rejected.
     * In particular, test that:
     * <ul>
     * <li>The magic number is checked (relatively quickly if possible).
     * <li>The entire file is verified, eg the image is loaded, if possible.
     * </ul>
     */
    public void testDirectUploadBad()
        throws Exception
        {
        // If no upload dir, we can't do the test.
        if(uploadDir == null)
            {
//            System.err.println("WARNING: cannot use/create/empty upload dir so skipping test");
            fail("WARNING: cannot use/create/empty upload dir so skipping test");
            return;
            }

        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        final UploadInfoBean uib = _makeTestUploadUIB(edfs, ".jpg");

        // Create the full name...
        final String newExhibitName = uib.getFullName();

        // Full name at which we expect upload to appear...
        final File targetFileName = new File(uploadDir, newExhibitName);

        // Ensure that it does not currently exist in the upload area...
        assertFalse("File to upload must not exist in upload area before we start!",
                    targetFileName.exists());

        // Create a random, bad JPEG (ie just random numbers).
        final byte[] badImage = new byte[rnd.nextInt(987654)];
        rnd.nextBytes(badImage);

        // Gather up output stream for POST command...
        final ByteArrayOutputStream baos = new ByteArrayOutputStream(badImage.length + 1024);

        // Choose suitable random boundary string.
        final String boundary = "------------" + Long.toString(rnd.nextLong() >>> 1, 36);

        // Actually generate the POST body bytes...
        HTTPUploaderUtils.generateUploadPOSTBody(uib, baos, badImage, boundary);
        final byte body[] = baos.toByteArray();
        assertTrue("Extracted body cannot be zero length", body.length > 0);

        final ServletInputStream is = new ServletInputStream(){
            /**Position in body[]. */
            private int pos;

            /**Reads the next byte of data from the input stream, or -1 at EOF.
             */
            final public int read() // throws IOException
                {
                if(pos >= body.length) { return(-1); }
                return(body[pos ++] & 0xff);
                }
            };

        // Make sure upload is vetoed because of bad magic number.
        try
            {
            // Do upload (and time it)...
            HTTPUploaderUtils.doUpload(uib,
                                 uploadDir,
                                 is,
                                 boundary,
                                 Integer.MAX_VALUE,
                                 GenUtils.systemOutLogger);

            // Various checks should have vetoed upload
            fail("Allowed bad exhibit to be uploaded");
            }
        catch(InvalidObjectException e)
            {
            // Good, want exhibit upload to be vetoed for bad magic number
            }

        // Ensure that bad file does not get uploaded.
        assertFalse("Bad file must nto get uploaded",
                    targetFileName.exists());
        }


    /**Test send/receive of server response to batch-upload client.
     */
    public void testBatchUploaderResponseHandling()
        throws Exception
        {
        // Resuable output buffer...
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();

        // A random unique valid exhibit name.
        final String exhibitName = "a/" + (Rnd.fastRnd.nextLong() >>> 1) + "-A.a";

        // A random unique non-empty detail message.
        final String detail = "detail: " + Rnd.fastRnd.nextLong();

        final SimpleLoggerIF logger = null;

        // Test a simple "OK" case.
        baos.reset();
        UploaderUtils.sendResponseToBatchUploadClient(logger, baos, true, exhibitName, detail);
        assertEquals("Must be able to recover exhibit name",
                     exhibitName, UploaderUtils.decodeResponseToBatchUploadClient(new ByteArrayInputStream(baos.toByteArray())));
        // Show that damaging the trailer causes rejection.
        try
            {
            final byte[] buf = baos.toByteArray();
            buf[buf.length-1] ^= (Rnd.fastRnd.nextInt() | 1); // Trailer byte definitely changed from correct value.
            UploaderUtils.decodeResponseToBatchUploadClient(new ByteArrayInputStream(buf));
            fail("Did not throw IOException from damaged OK message");
            }
        catch(final IOException e)
            {
            // Good, expected an error.
            }

        // Test a simple "FAIL" case.
        baos.reset();
        UploaderUtils.sendResponseToBatchUploadClient(logger, baos, false, exhibitName, detail);
        try
            {
            UploaderUtils.decodeResponseToBatchUploadClient(new ByteArrayInputStream(baos.toByteArray()));
            fail("Did not throw IOException from FAIL message");
            }
        catch(final IOException e)
            {
            // Good, expected an error.
            assertTrue("Detail message not included in exception", e.getMessage().contains(detail));
            }

        // Make sure that some unpleasant random junk is discarded safely!
        try
            {
            UploaderUtils.decodeResponseToBatchUploadClient(new InputStream(){
                /**Returns random byte that could often be a valid start-of-message or text, ie semi-pathological! */
                public final int read() // throws IOException
                    { return(rnd.nextBoolean() ? 'O' : (rnd.nextInt() & 0xff)); }
                });
            fail("Did not throw IOException from random garbage input");
            }
        catch(final IOException e)
            {
            // Good, expected an error.
            }

        // Make sure that EOF is rejected correctly.
        try
            {
            UploaderUtils.decodeResponseToBatchUploadClient(new ByteArrayInputStream(new byte[0]));
            fail("Did not throw IOException EOF");
            }
        catch(final IOException e)
            {
            // Good, expected an error.
            }
        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
