/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataHTTPTunnelSource;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource.PacketProtector;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;

import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.servlet.ServletContext;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.nio.ByteBuffer;
import java.security.Key;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Queue;
import java.util.Random;
import java.util.concurrent.ArrayBlockingQueue;

/**Test the HTTP communications tunnel.
 * May need to be running in a WAR container and in master mode
 * for this to work
 * (unless, exceptionally, the loopbackURL is pointing at another instance).
 */
public final class HTTPTunnelTest extends TestCase
    {
    public HTTPTunnelTest(final String name)
        {
        super(name);
        }

    private final SimpleLoggerIF logger = new SimpleLoggerIF(){
        public void log(final String message) { Main.getOut().println(message); }
        };

    /**Test that our hand-crafted serialisation routines work.
     * This primarily means that we can decode with DataInputStream
     * what we hand-encoded.
     */
    public static void testIntSerRoutines()
        throws Exception
        {
        // Test for a few values...
        for(int i = 10; --i >= 0; )
            {
            final int i1 = rnd.nextInt();
            final DataInputStream dis1 = new DataInputStream(
                new ByteArrayInputStream(
                    ExhibitDataTunnelSource.intSer(i1)));
            assertEquals("DataInputStream.readInt() must be able to decode output of intSer()",
                i1, dis1.readInt());

            final long l1 = rnd.nextLong();
            final DataInputStream dis2 = new DataInputStream(
                new ByteArrayInputStream(
                    ExhibitDataTunnelSource.longSer(l1)));
            assertEquals("DataInputStream.readInt() must be able to decode output of longSer()",
                l1, dis2.readLong());
            }
        }

    /**Check that we can open a basic HTTP connection OK to the loopbackURL.
     * Returns immediately if no loopbackURL is set.
     */
    public static boolean canConnectToLoopbackURL()
        {
        final URL lb = Main.getLoopbackURL();
        if(lb == null) { return(false); }
        try
            {
            final URLConnection conn = lb.openConnection();
            conn.setUseCaches(false); // Force new connection each time.
            if(conn instanceof HttpURLConnection)
                {
                final HttpURLConnection hconn = (HttpURLConnection) conn;
                hconn.setInstanceFollowRedirects(true); // Allow redirects.

                // Check that the response is a 200 code.
                final int respCode = hconn.getResponseCode();
                if(respCode == HttpURLConnection.HTTP_OK)
                    { return(true); } // HTTP "OK" means loopback is OK.
                }
            }
        catch(IOException e)
            { return(false); } // Regard this as a failure.

        // Assume failure if we get here.
        return(false);
        }

    /**If ServletContext exists and we are in master mode, return DataSource.
     * Also, we check that basic loop-back is OK.
     * <p>
     * Else we return null.
     */
    public static DataSourceBean getDataSourceBeanIfMaster()
        {
        if(!canConnectToLoopbackURL()) { return(null); }
        final ServletContext sc = (ServletContext) Main.getServletContext();
        if(sc == null) { return(null); }
        final DataSourceBean dsb = DataSourceBean.getApplicationInstance(sc);
        if(!dsb.isSlave().equals(Boolean.FALSE)) { return(null); }
        return(dsb);
        }

    /**Does basic loopback test.
     * If not, does not cause a failure but warns that this is not possible.
     * Mainly useful in case other tests are silently disabled by this.
     */
    public static void testLoopback()
        {
        if(!canConnectToLoopbackURL())
            {
            Main.getErr().println("WARNING: HTTPTunnelTest: cannot connect to loopback HTTP[S] server, some tests may be disabled: " +
                                  Main.getLoopbackURL());
            }

        final DataSourceBean dsb = getDataSourceBeanIfMaster();
        if(dsb == null)
            {
            Main.getErr().println("WARNING: HTTPTunnelTest: no servlet context, test skipped");
            return;
            }
        }

    /**Attempts basic test of tunnel by doing a NO-OP (NOOP).
     * May require us to be in "master" mode.
     */
    public void testTunnelNOOP()
        throws Exception
        {
        final DataSourceBean dsb = getDataSourceBeanIfMaster();
        if(dsb == null)
            {
            Main.getErr().println("WARNING: HTTPTunnelTest: no servlet context, test skipped");
            return;
            }

        final ServletContext ctxt = (ServletContext) Main.getServletContext();
        // Get the usual URL for connecting to the master server tunnel...
        final URL masterURL = new URL(ctxt.getInitParameter(CoreConsts.WAR_CTXTPARAM_BOOTURL));
        // Then take the loopback URL and add the server tunnel URI...

        final String tunnelLoopbackURL = Main.getLoopbackURL() +
            masterURL.getPath();

        // Now try to set up a tunnel connection.
        final ExhibitDataTunnelSource ts =
            new ExhibitDataHTTPTunnelSource(tunnelLoopbackURL, logger);

        // Check that we can successfully send a NOOP request to the server.
        // (Throws an IOException if not.)
        ts.doNOOP(rnd.nextBoolean());
        }

    /**Attempts to test basic use of the tunnel for normal traffic.
     * May require us to be in "master" mode.
     */
    public void testTunnelBasic()
        throws Exception
        {
        final DataSourceBean dsb = getDataSourceBeanIfMaster();
        if(dsb == null)
            {
            Main.getErr().println("WARNING: HTTPTunnelTest: no servlet context, test skipped");
            return;
            }

        final ServletContext ctxt = (ServletContext) Main.getServletContext();

        // Get the usual URL for connecting to the master server tunnel...
        final URL masterURL = new URL(ctxt.getInitParameter(CoreConsts.WAR_CTXTPARAM_BOOTURL));

        // Then take the loopback URL and add the server tunnel URI...
        final String tunnelLoopbackURL = Main.getLoopbackURL() +
            masterURL.getPath();

        // Now try to set up a tunnel connection.
        final ExhibitDataTunnelSource ts =
            new ExhibitDataHTTPTunnelSource(tunnelLoopbackURL, logger);

        // Check that we can successfully send a NOOP request to the server.
        // (Throws an IOException if not.)
        ts.doNOOP(rnd.nextBoolean());

        // Now try to get the entire set of exhibit properties.
        final AllExhibitProperties aepRemote = ts.getAllExhibitProperties(-1);
        final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
        // Must not be null.
        assertNotNull("AEP retrieved over tunnel must not be null", aepRemote);
        // Must be equal to the set I have locally.
        assertEquals("AEP retrieved over tunnel must match local version",
                     aep, aepRemote);
        // But must not be the same object,
        // it it must really have been sent over the "wire".
        assertNotSame("AEP retrieved over tunnel must have been (de)serialised",
                      aep, aepRemote);

        // Now try to get the entire set of exhibit immutable properties.
        final AllExhibitImmutableData aeidRemote = ts.getAllExhibitImmutableData(-1);
        final AllExhibitImmutableData aeid = dsb.getAllExhibitImmutableData(-1);
        // Must not be null.
        assertNotNull("AEID retrieved over tunnel must not be null", aeidRemote);
        // Must be equal to the set I have locally.
        assertEquals("AEID retrieved over tunnel must match local version",
                     aeid, aeidRemote);
        // But must not be the same object,
        // it it must really have been sent over the "wire".
        assertNotSame("AEID retrieved over tunnel must have been (de)serialised",
                      aeid, aeidRemote);

        // Now try to get the entire set of exhibit properties.
        final GenProps gpRemote = ts.getGenProps(-1);
        final GenProps gp = dsb.getGenProps(-1);
        // Must not be null.
        assertNotNull("GP retrieved over tunnel must not be null", gpRemote);
        // NOTE: cannot test for equality because of local overrides of gp.
        // Must not be the same object,
        // it it must really have been sent over the "wire".
        assertNotSame("GP retrieved over tunnel must have been (de)serialised",
                      gp, gpRemote);

        // Now try to get the system ID.
        // It should be different via the tunnel.
        final SimpleVariableValue id1 = dsb.getVariable(SystemVariables.LOCAL_SYS_ID);
        final SimpleVariableValue id2 = ts.getVariable(SystemVariables.LOCAL_SYS_ID);
        assertNotNull("Should be able to retrieve system ID from DataSourceBean", id1);
        assertNotNull("System ID from DataSourceBean should have non-null value", id1.getValue());
        assertNotNull("Should be able to retrieve system ID from tunnel", id2);
        assertNotNull("System ID from tunnel should have non-null value", id2.getValue());
        assertFalse("System ID from DataSourceBean and tunnel should be different",
            id1.getValue().equals(id2.getValue()));


        // Test that we can set a global on either side of the tunnel
        // and read it immediately on either side.
        // We can't test for it being initially unset
        // since the server may hold a value from a previous run.
//        assertNull("Must see null for unset global on master",
//            dsb.getVariable(SystemVariables.TEST_NUMBER_GLOBAL));
//        assertNull("Must see null for unset global on slave",
//            ts.getVariable(SystemVariables.TEST_NUMBER_GLOBAL));
        // Set from global side, test on both.
        final Number v1 = new Integer(1);
        final SimpleVariableValue svv1 = new SimpleVariableValue(
            SystemVariables.TEST_NUMBER_GLOBAL, v1);
        dsb.setVariable(svv1);
        assertEquals("Must be able to see on master variable set on master",
            v1, dsb.getVariable(SystemVariables.TEST_NUMBER_GLOBAL).getValue());
        assertEquals("Must be able to see on slave variable set on master",
            v1, ts.getVariable(SystemVariables.TEST_NUMBER_GLOBAL).getValue());
        // Now check that getVariables() is working on both sides...
        assertTrue("Must be able to retrieve value set on master with getVariables(-1) on master",
            SystemVariablesTest.checkValuePresent(dsb.getVariables(-1), svv1));
        assertTrue("Must be able to retrieve value set on master with getVariables(-1) on slave",
            SystemVariablesTest.checkValuePresent(ts.getVariables(-1), svv1));
        // Set from slave side, test on both.
        final Number v2 = new Short((short) 42);
        final SimpleVariableValue svv2 = new SimpleVariableValue(
            SystemVariables.TEST_NUMBER_GLOBAL, v2);
        ts.setVariable(svv2);
        assertEquals("Must be able to see on master variable set on slave",
            v2, dsb.getVariable(SystemVariables.TEST_NUMBER_GLOBAL).getValue());
        assertEquals("Must be able to see on slave variable set on slave",
            v2, ts.getVariable(SystemVariables.TEST_NUMBER_GLOBAL).getValue());
        // Now check that getVariables() is working on both sides...
        assertTrue("Must be able to retrieve value set on slave with getVariables(-1) on master",
            SystemVariablesTest.checkValuePresent(dsb.getVariables(-1), svv2));
        assertTrue("Must be able to retrieve value set on slave with getVariables(-1) on slave",
            SystemVariablesTest.checkValuePresent(ts.getVariables(-1), svv2));

        // Do general pipeline and tunnel tests...
        SystemVariablesTest.pipelineVariableTest(
            new SimpleExhibitPipelineIF[]{ts}, // Simple tunnel pipeline.
            null,
            new SimpleExhibitPipelineIF[]{dsb}, // Top of master pipeline.
            System.currentTimeMillis() + 10000L); // Run for up to 10s.

        // Check that data is seen consistently on either side of the tunnel.
        checkExhibitDataConsistency(ts, dsb,
                                    true, // On two sides of a tunnel.
            System.currentTimeMillis() + 10000L);  // Run for up to 10s.
        }

    /**Check that exhibit properties and data from two points match.
     * The two points can be either side of a tunnel,
     * or either side of a cache,
     * for example.
     * <p>
     * Any cacheing stages are assumed to have been running long enough
     * (usually at least one poll() if they are asynchronous)
     * to have brought meta-data (eg properties) up-to-date.
     * <p>
     * Package-visible to be available to other test cases.
     *
     * @param stageA  one sample stage in a pipeline
     *     possibly on slave side of tunnel); never null
     * @param stageB  another sample stage in a pipeline in the same system;
     *     never null
     * @param overTunnel  if true, the two stages are on either side of
     *     a tunnel and objects should have been serialised
     *     between the sample stages
     * @param stopBy  a target time to stop by
     *     when all distinct tests have been run at least once
     *
     * @throws IOException  in case of unexpected I/O problems
     */
    static void checkExhibitDataConsistency(
                                    final SimpleExhibitPipelineIF stageA,
                                    final SimpleExhibitPipelineIF stageB,
                                    final boolean overTunnel,
                                    final long stopBy)
        throws IOException
        {
        // Now try to get the entire set of exhibit properties.
        final AllExhibitProperties aepA = stageA.getAllExhibitProperties(-1);
        final AllExhibitProperties aepB = stageB.getAllExhibitProperties(-1);
        // Must not be null.
        assertNotNull("AEP retrieved from stageA must not be null", aepA);
        // Must be equal to the set I have locally.
        assertEquals("AEP retrieved from two sample points must be equal",
                     aepB, aepA);

        final AllExhibitImmutableData aeid = aepB.aeid;

        // Test exhibits randomly for attributes and data...
        if(aeid.length == 0)
            {
            Main.getErr().println("WARNING: HTTPTunnelTest: no exhibits, content test skipped");
            return;
            }

        final String names[] = aeid.getAllExhibitNamesSorted();
        final int nTests = 50 + rnd.nextInt(5 + names.length/100); // Do a few tests...
//        Main.getOut().println("Doing up to " + nTests + " tests on exhibits...");
        for(int i = nTests; --i >= 0; )
            {
            final String exhibit = names[rnd.nextInt(names.length)];

            // Check that static attributes can be retrieved correctly.
            final ExhibitStaticAttr esaRemote = stageA.getStaticAttr(exhibit);
            final ExhibitStaticAttr esa = stageB.getStaticAttr(exhibit);
            // Must not be null.
            assertNotNull("ESA retrieved from stageA must not be null", esaRemote);
            // Must be equal to the set I have locally.
            assertEquals("ESA retrieved from two sample points must be equal",
                         esa, esaRemote);
            if(overTunnel)
                {
                // But must not be the same object,
                // so it must really have been sent over the "wire".
                assertNotSame("ESA retrieved over tunnel must have been (de)serialised",
                              esa, esaRemote);
                }

            if(esa.length < 1)
                {
                Main.getOut().println("Exhibit too short to use: " + esa);
                continue;
                }

            // Check that exhibit data can be retrieved correctly.

            // Pick a random start.
            // Make sure that we start from zero in at least some cases
            // as this emulates the start of an exhibit download.
            int start = rnd.nextBoolean() ?
                0 : rnd.nextInt((int) esa.length - 1);

            // Read a smallish number of successive blocks.
            for(int j = 1 + rnd.nextInt(3); (--j >= 0) && (start < esa.length); )
                {
                // Limit the maximum read size to avoid rejection.
                final int len = Math.min(SimpleExhibitPipelineIF.MAX_USER_READ_SIZE,
                                         1 + rnd.nextInt((int) (esa.length - start)));
                final ByteBuffer rawDataRemote = ByteBuffer.allocate(len);
                stageA.getRawFile(rawDataRemote, exhibit, start, rnd.nextBoolean());
                rawDataRemote.flip();
                final ByteBuffer rawData = ByteBuffer.allocate(len);
                stageB.getRawFile(rawData, exhibit, start, rnd.nextBoolean());
                rawData.flip();

                // We'll compare the common prefixes of the fetched items.
                final int commonLen = Math.min(rawDataRemote.remaining(), rawData.remaining());
                assertTrue("Each read should read at least one byte", commonLen > 0);

//{ System.out.println("***First bytes for "+esa+"@"+start+"... remote/local="+rawDataRemote.get()+'/'+rawData.get()); rawDataRemote.rewind(); rawData.rewind(); }
                // Must be identical to the data fetched locally.
                for(int k = commonLen; --k >= 0; )
                    { assertEquals("raw data retrieved over tunnel must match local version", rawData.get(), rawDataRemote.get()); }

                start += commonLen;
                }

            // If we're out of time, then stop having done at least one test.
            if(System.currentTimeMillis() >= stopBy) { break; }
            }
        }

    /**Test behaviour against dead master.
     * We do this by connecting to a bogus loopback address.
     * <p>
     * We do not have to be master for this,
     * nor even running as a servlet...
     * <p>
     * We simply check that we recover from a first dead connection,
     * then subsequent connection attempts are refused for at least
     * the minimum failure time and not much more than the maximum...
     * These refusals should be with a "PGMasterNotInService" exception.
     */
    public void testAgainstDeadMaster()
        throws Exception
        {
        Main.getOut().println("Checking for correct connection back-off against failed master...");

        // Bogus tunnerl URL that should fail quickly (and locally/cheaply)...
        final String bogusTunnelURL = "http://localhost:12345/bogus.noway.Hose";

        // Now try to set up a tunnel connection.
        final ExhibitDataTunnelSource ts =
            new ExhibitDataHTTPTunnelSource(bogusTunnelURL, logger);

        // Check that we cannot successfully send a NOOP request to the server.
        // If this works, something is wrong!
        try
            {
            ts.doNOOP(rnd.nextBoolean());
            fail("This connection should not have succeeded!");
            }
        catch(IOException e) /* We expect this to fail... */ { }

        final SimpleVariableValue sysID = ts.getVariable(SystemVariables.LOCAL_SYS_ID);
        assertNotNull("Should be able to get system ID even with master down with getVariable()",
            sysID);
        // Just make sure that at least one value is returned...
        final SimpleVariableValue[] variables = ts.getVariables(-1);
        assertTrue("Should be able to get system ID even with master down with getVariables()",
            SystemVariablesTest.checkValuePresent(variables, sysID));


        // Now all subsequent attempts should be deferred by a time
        // approximately bounded by the min/max retry times.
        // We test this out a few times...
        for(int i = 3; --i >= 0; )
            {
            final long start = System.currentTimeMillis();

            // Intensively try to talk to the master...
            // We care about when we get the first "real" IOException
            // indicating that a real connection attempt was made.
            for( ; ; )
                {
                // Sleep a fairly short time before trying again...
                try { Thread.sleep(1 + rnd.nextInt(131)); }
                catch(InterruptedException e) {}

                try
                    {
                    // Should not be able to do NO-OP with master unreachable.
                    ts.doNOOP(false); // This should show normal deferral behaviour...
                    fail("This connection should not have succeeded!");
                    }
                catch(PGMasterNotInServiceException e) /* Good, deferring... */ { }
                catch(IOException e)
                    {
                    // We take an IOException to indicate that real I/O was attempted again...
                    final long realIOTime = System.currentTimeMillis();

                    // We compute how long "real" I/O was deferred...
                    final long deferralTime = realIOTime - start;

                    assertTrue("We must wait at least about the minimum retry time before real I/O",
                        deferralTime > ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MIN_MS / 2);
                    assertTrue("We must wait at most about the maximum retry time before real I/O",
                        deferralTime < ExhibitDataTunnelSource.FAIL_RETRY_WAIT_MAX_MS * 2);

                    // OK, timing was reasonable.
                    break;
                    }
                }
            }

        // Check that tunnel still behaves rationally with a dead master.
        // Note that it is not required to be able to hold any values,
        // though it must be able to return the system ID...
        SystemVariablesTest.pipelineVariableTest(
            new SimpleExhibitPipelineIF[]{ts}, null, null,
            System.currentTimeMillis() + 10000L);
        }

    /**Test getAEP 'diff' routine across the tunnel.
     * Checks that it always generates correct results.
     */
    public static void testAEPDiffTunnel()
        throws Exception
        {
        final SimpleLoggerIF logger = GenUtils.systemOutLogger;
        
        // Set of frozen AEPs to test with.
        final List<String> frozenAEPNamesAll = BackCompatTest.frozenAEPs;
        // Only use last few frozen AEPs to represent recent state.
        final int MAX_AEPs = 3;
        final int sizeAll = frozenAEPNamesAll.size();
        final List<String> frozenAEPNames = (sizeAll <= MAX_AEPs) ? frozenAEPNamesAll :
            frozenAEPNamesAll.subList(sizeAll-MAX_AEPs, sizeAll);
        final List<AllExhibitProperties> frozenAEPValues = new ArrayList<AllExhibitProperties>(frozenAEPNames.size());
        for(final String filename: frozenAEPNames)
            {
            Main.getOut().println("[Loading "+filename+"...]");
            frozenAEPValues.add(AllExhibitProperties.canonicalise((AllExhibitProperties)
                FileTools.deserialiseFromFile(new File(filename), true)));
            }

        // Full test set.
        // Test set: empty AEP, then "in-order" set or random-order set.
        final List<AllExhibitProperties> fts = new ArrayList<AllExhibitProperties>(1+frozenAEPValues.size());
        fts.addAll(frozenAEPValues); // Start with roughly in-order AEPs.
        if(rnd.nextBoolean() && Rnd.goodRnd.nextBoolean())
            {
            // Occasionally try in random (atypical) order to shake out bugs...
            Main.getOut().println("[Shuffling into random order to test robustness...]");
            Collections.shuffle(fts);
            }
        fts.add(0, new AllExhibitProperties()); // Start with empty AEP!

        // Randomly chose to force serialisation or not.
        final boolean forceSerialisation = Rnd.goodRnd.nextBoolean();
        Main.getOut().println("[Testing with serialisation forced: "+forceSerialisation+"...]");
        final Queue<AllExhibitProperties> testValues = new ArrayBlockingQueue<AllExhibitProperties>(fts.size());
        final SimpleExhibitPipelineIF source = new SimpleCacheTest.DummyDataSource(){
            /**Return the head AEP test value (or null if hash matches), or throw exception if none. */
            @Override public AllExhibitProperties getAllExhibitProperties(final long oldHash) throws IOException
                {
                final AllExhibitProperties result = testValues.peek();
                if(result == null) { throw new EOFException(); }
                if(result.longHash == oldHash) { return(null); }
                return(result);
                }
            };
        final SystemVariablesTest.LocalTunnelServer server = new SystemVariablesTest.LocalTunnelServer(source, true);
        final ExhibitDataTunnelSource tunnelSource = new SystemVariablesTest.LocalTunnelSource(server, forceSerialisation, logger);

        // Process test AEP values in turn.
        AllExhibitProperties prevAEP = null;
        for(testValues.addAll(fts); !testValues.isEmpty(); prevAEP = testValues.remove())
            {
            // Repeat fetch a few times to represent
            // multiple clients updating from one server.
            for(int i = 3; --i >= 0; )
                {
                final AllExhibitProperties aep = testValues.peek();
                final AllExhibitProperties fetchedAEP = tunnelSource.getAllExhibitProperties(prevAEP, false);

                assertEquals("fetched value should be null iff prev and current AEP are the same",
                                aep.equals(prevAEP), null == fetchedAEP);
                if(null != fetchedAEP)
                    { assertEquals("AEP retrieved via tunnel diff call must equal what went in", aep, fetchedAEP); }
                }
            }
        }

    /**Test that our HMAC does detect corruption in, or tampering with, messages.
     * Not even a single-bit change, nor any deletion or insertion of bytes,
     * should go unobserved.
     * <p>
     * Also, the timestamp and unique ID should prevent replay attacks
     * (and should keep potential participants with bad clocks from communicating).
     * <p>
     * Note that whatever works on complete messages should also work on streamed messages,
     * though there is an onus on the streaming-related code to behave safely
     * even if fed bogus data.
     */
    public static void testHMACEfficacy()
        throws Exception
        {
        // VERY BASIC TESTING OF MAC COMPUTATION.
        // Make up a new secret key dynamically.
        final KeyGenerator keyGen = javax.crypto.KeyGenerator.getInstance(CoreConsts.HMAC_ALG);
        keyGen.init(Rnd.goodRnd);
        final SecretKey key1 = keyGen.generateKey();
        // Capture a timestamp.
        final long stamp1 = System.currentTimeMillis();
        // Create an empty NOP RawPacket.
        final ExhibitDataTunnelSource.RawPacket emptyRP1 =
            new ExhibitDataTunnelSource.RawPacket(ExhibitDataTunnelSource.RawPacket.OP_NOOP, new byte[0]);
        // Compute MAC for the message "by hand"...
        final List<ROByteArray> mac1 = ExhibitDataTunnelSource.PacketProtector.
            computeMAC(stamp1, emptyRP1, key1);
        assertNotNull("Computed MAC must not be null", mac1);
        assertNotSame("Computed MAC must not be empty", 0, mac1.size());
        for(final ROByteArray r : mac1)
            {
            assertNotNull("Computed segment MAC must not be null", r);
            assertTrue("Computed segment MAC must be at least 256 bits for safety", r.length() >= 32);
            }

        // Check that we can do the same for any XferKey available from LocalProps.
        final SecretKey xk = LocalProps.getXferHMACKey();
        if(xk == null) { Main.getOut().println("No local key for testing: LocalProps.getXferHMACKey() returns null"); }
        else
            {
            Main.getOut().println("Testing (extant) local xfer-protection key...");
            final List<ROByteArray> mac2 = ExhibitDataTunnelSource.PacketProtector.
                computeMAC(stamp1, emptyRP1, xk);
            assertNotNull("Computed MAC must not be null", mac2);
            assertNotSame("Computed MAC must not be empty", 0, mac2.size());
            // TODO: check for all segment MACs.
            assertTrue("Computed MAC must be at least 256 bits for safety", mac2.get(0).length() >= 32);            
            assertTrue("Computed MAC from local key must not match that from generated key!",
                            !mac1.equals(mac2));
            }

        // Create a PacketProtector and ensure the MAC is the same as for direct computation.
        final ExhibitDataTunnelSource.PacketProtector prot1 =
            new ExhibitDataTunnelSource.PacketProtector(emptyRP1, stamp1, key1);
        assertEquals("Must get equal MAC on construction of PacketProtector", prot1.mac, mac1);

        // Generate and parse the check-string.
        final String prot1CheckString = prot1.toCheckString();
        Main.getOut().println("Example check-string: "+prot1CheckString);
        // Check that we can re-parse the check string and get the same MAC, etc.
        assertEquals("Must be able to regenerate MAC (etc) from check string",
            prot1, PacketProtector.fromCheckString(prot1CheckString));

        // Estimate MAC-computation time on large (random-data) packet,
        // similar to a (compressed) serialised AEP.
        // This is likely to be much worse than for most frames because:
        //   * Most packets will be very much shorter.
        //   * The HotSpot will have had a chance to optimise the code after repeated uses.
        final byte rawPayload0[] = new byte[2123123];
        rnd.nextBytes(rawPayload0);
        final ExhibitDataTunnelSource.RawPacket hugeRP = /* Do not allow compression. */
            new ExhibitDataTunnelSource.RawPacket(ExhibitDataTunnelSource.RawPacket.OP_NOOP, rawPayload0, false);
        final long macCompStart = System.currentTimeMillis();
        assertEquals("MAC must recompute to same value each time",
            (new ExhibitDataTunnelSource.PacketProtector(hugeRP, stamp1, key1)).mac,
                (new ExhibitDataTunnelSource.PacketProtector(hugeRP, stamp1, key1)).mac);
        final long macCompEnd = System.currentTimeMillis();
        Main.getOut().println("Large MAC compute time ~"+((macCompEnd-macCompStart)/2)+"ms."); // Slightly conservative measure.

        // Try several times to sneak undetected corruptions into a packet with a raw-binary payload.
        // We expect this to succeed some of the time for a RawPacket (especially uncompressed),
        // but it should essentially never work for a PacketProtector.
        // (This test only checks the anti-corruption features.)
        for(int i = 5; --i >= 0; )
            {
            // Make our RawPacket with a random payload and size similar to a bulk transfer.
            // Bigger than the minimum MAC segment size to sometimes force more than one segment.
            // Randomly allow compression to be attempted on the payload.
            final byte rawPayload[] = new byte[rnd.nextInt(1024 + 3*ExhibitDataTunnelSource.PacketProtector.MIN_SEGMENT_SIZE)];
            // Half the time have the payload a very-compressable all-zeros.
            if((i & 1) == 0) { rnd.nextBytes(rawPayload); }
            final ExhibitDataTunnelSource.RawPacket inputRP =
                new ExhibitDataTunnelSource.RawPacket(ExhibitDataTunnelSource.RawPacket.OP_NOOP, rawPayload, rnd.nextBoolean());
            // Store a copy of the on-the-wire state of the (uncorrupt) RawPacket.
            final ByteArrayOutputStream baos1 = new ByteArrayOutputStream(inputRP.getFrameLength());
            inputRP.writePacket(baos1);
            final byte goodRPBytes[] = baos1.toByteArray();
            baos1.reset();
            // Null hypothesis?
            // Now check that we can always successfully read back from the uncorrupted state.
            assertTrue("RawPacket read from uncorrupted data must be equal",
                            inputRP.equivalentTo(ExhibitDataTunnelSource.RawPacket.readPacket(new ByteArrayInputStream(goodRPBytes))));

            // Make a PacketProtector wrapper for the RawPacket.
            // We're happy to use the key generated earlier.
            final ExhibitDataTunnelSource.PacketProtector inputPP =
                new ExhibitDataTunnelSource.PacketProtector(inputRP, key1);
            // And show that we can regenerate this on demand, getting the same MAC.
            assertEquals("Regenerating the MAC must work for the same timestamp and RawPacket",
                            inputPP.mac, (new ExhibitDataTunnelSource.PacketProtector(inputRP, inputPP.timestamp, key1)).mac);

            // Generate and parse the check-string.
            final String inputPPCheckString = inputPP.toCheckString();
            //Main.getOut().println("Example check-string: "+inputPPCheckString);
            // Check that we can re-parse the check string and get the same MAC, etc.
            assertEquals("Must be able to regenerate MAC (etc) from check string",
                            inputPP, PacketProtector.fromCheckString(inputPPCheckString));

            // Make sure that the streaming version of MAC validation/protection
            // allows reading of a good (non-corrupt) segment/stream.
            final DataInputStream dis = new DataInputStream(
                    inputPP.protectInputStream(key1, new ByteArrayInputStream(goodRPBytes)));
            final byte[] bs = new byte[inputRP.getFrameLength()];
            dis.readFully(bs); // Attempt to read entire frame...
            dis.close();
            assertTrue("Data must not be harmed by being read through protector wrapper",
                            Arrays.equals(goodRPBytes, bs));

            // Now make a corrupt version of the on-the-wire frame data.
            // This should include changes/insertions/deletions.
            // TODO: Test corruption types other than a single-bit flip.
            // CORRUPTION: induce a single-bit change/error.
            final byte badRPBytes[] = goodRPBytes.clone();
            final int corruptionPos = rnd.nextInt(badRPBytes.length);
            badRPBytes[corruptionPos] ^= (1 << rnd.nextInt(8));
            // New and old data will now definitely be different.
            assertTrue(!Arrays.equals(goodRPBytes, badRPBytes));

            // Attempt to unpack the corrupt data.
            // An early failure (to decode the byte stream) here is fine:
            // we probably damaged the header or made compressed data unusable.
            final ExhibitDataTunnelSource.RawPacket badRP;
            try { badRP = ExhibitDataTunnelSource.RawPacket.readPacket(new ByteArrayInputStream(badRPBytes)); }
            catch(final IOException e)
                {
//                Main.getOut().println("Could not decode RawPacket, which is fine.");
                continue;
                }
            // We assume that there is not much redundancy in the RawPacket form,
            // at least if it was decodable from the byte-stream format,
            // so therefore the flipped bit should usually show up somewhere that we can see.
            if(inputRP.equivalentTo(badRP))
                { Main.getOut().println("[RawPacket was not harmed by the corruption (corruptionPos="+corruptionPos+"/"+badRPBytes.length+")!]"); }
            // But computing the MAC for the new packet (with the previous timestamp)
            // must give a different value, showing that the (on-the-wire) data is corrupt.
            final ExhibitDataTunnelSource.PacketProtector badPP =
                new ExhibitDataTunnelSource.PacketProtector(badRP, inputPP.timestamp, key1);
            assertTrue("MAC for corrupt data must be different to original MAC",
                            !inputPP.mac.equals(badPP.mac));
//            Main.getOut().println("MAC old/new: " + inputPP.mac + "/" + badPP.mac);

            // Make sure that the streaming version of MAC validation/protection
            // vetoes reading of a corrupt segment/stream with an IOException.
            final DataInputStream disB = new DataInputStream(
                    inputPP.protectInputStream(key1, new ByteArrayInputStream(badRPBytes)));
            try
                {
                disB.readFully(bs); // Attempt to read entire frame...
                fail("Should have detected corruption (using MAC) and thrown IOException");
                }
            catch(final IOException e) { /* Good: stream corruption was detected. */ }
            disB.close();
            }
        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
