/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.AddrTools;
import org.hd.d.pg2k.svrCore.ExhibitPropsGlobalImmutable;
import org.hd.d.pg2k.svrCore.location.GeoProximity;
import org.hd.d.pg2k.svrCore.location.GeoUtils;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth.AEParams;

import java.io.File;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

/**Test of various aspects of Location and GeoUtils, etc.
 * This includes geographical lookup by client IP address,
 * though maybe these should be moved to a separate GeoUtilsTests class.
 * <p>
 *
 * TODO: test that exhibits correctly get specific/generic flag
 * depending on details being with exhibit or looked up in generic map
 */
public final class LocationTest extends TestCase
    {
    public LocationTest(final String name)
        {
        super(name);
        }

    /**Basic tests of LocationMap behaviour.
     * These tests are not dependent on external data.
     */
    public static void testLocationMapBasics()
        {
        // Default constructor should yield empty map with zero hashcode.
        final LocationMap elm = new LocationMap();
        assertTrue("Default LocationMap must be empty", elm.isEmpty());
        assertTrue("Default LocationMap must have zero timestamp",
                   elm.hashCode() == 0);

        // TODO: some real tests!
        }

    /**Tests use of contains() and related operations on Location.Estd.
     * We use some randomness to stochastically test the entire space
     * of parameters that should be unimportant (ie the "specific" flag).
     */
    public static void testLocationEstdContains()
        throws Exception
        {
        final Properties propsAllZeros = new Properties();
        propsAllZeros.setProperty("type", "Estd");
        propsAllZeros.setProperty("E", "0");
        propsAllZeros.setProperty("EErr", "0");
        propsAllZeros.setProperty("N", "0");
        propsAllZeros.setProperty("NErr", "0");
        final Location.Estd allZeros =
            new Location.Estd(rnd.nextBoolean(), "", propsAllZeros);

        final Properties propsZeroCentre = new Properties();
        propsZeroCentre.setProperty("type", "Estd");
        propsZeroCentre.setProperty("E", "0");
        propsZeroCentre.setProperty("EErr", "1");
        propsZeroCentre.setProperty("N", "0");
        propsZeroCentre.setProperty("NErr", "1");
        final Location.Estd zeroCentre =
            new Location.Estd(rnd.nextBoolean(), "", propsZeroCentre);

        final Properties propsAllOnes = new Properties();
        propsAllOnes.setProperty("type", "Estd");
        propsAllOnes.setProperty("E", "1");
        propsAllOnes.setProperty("EErr", "1");
        propsAllOnes.setProperty("N", "1");
        propsAllOnes.setProperty("NErr", "1");
        final Location.Estd allOnes =
            new Location.Estd(rnd.nextBoolean(), "", propsAllOnes);


        // All items should contain their own centre.
        // All items should be contained by WHOLE_EARTH and not contain it.
        final Location.Estd allEstdItems[] =
            {
                allZeros, zeroCentre, allOnes
            };
        for(int i = allEstdItems.length; --i >= 0; )
            {
            final Location.Estd l = allEstdItems[i];
            assertTrue("Location.Estd item must always contain its own centre",
                       l.containsCentre(l));
            assertTrue("Location.Estd item must always contain its own area",
                       l.containsArea(l));
            assertFalse("Location.Estd item must not strictly contain its own area",
                       l.strictlyContainsArea(l));
            assertTrue("All item centres must be contained by WHOLE_EARTH",
                       Location.Estd.WHOLE_EARTH.containsCentre(l));
            assertTrue("All item areas must be contained by WHOLE_EARTH",
                       Location.Estd.WHOLE_EARTH.containsArea(l));
            assertFalse("No item area may strictly contain WHOLE_EARTH",
                        l.strictlyContainsArea(Location.Estd.WHOLE_EARTH));
            assertFalse("No item area may be larger than WHOLE_EARTH",
                        l.isLargerThan(Location.Estd.WHOLE_EARTH));
            }

        // Test (some) expected pair-wise relationships.
        assertTrue("0E0N,+/-0 centre should be contained by bigger zero-centred item",
                   zeroCentre.containsCentre(allZeros));
        assertTrue("0E0N,+/-0 area should be contained by bigger zero-centred item",
                   zeroCentre.containsArea(allZeros));
        assertTrue("0E0N,+/-0 area should be strictly contained by bigger zero-centred item",
                   zeroCentre.strictlyContainsArea(allZeros));
        assertTrue("0E0N,+/-0 should contain centre of (bigger) zero-centred item",
                   allZeros.containsCentre(zeroCentre));
        assertFalse("0E0N,+/-0 should not contain area of (bigger) zero-centred item",
                    allZeros.containsArea(zeroCentre));
        assertFalse("0E0N,+/-0 should not strictly contain area of (bigger) zero-centred item",
                    allZeros.strictlyContainsArea(zeroCentre));
        assertFalse("0E0N,+/-0 should not contain centre of non-zero-centred item",
                    allZeros.containsCentre(allOnes));
        assertFalse("0E0N,+/-0 should not contain area of non-zero-centred item",
                    allZeros.containsArea(allOnes));
        assertTrue("1E1N,+/-1 should contain centre of (big) zero-centred item",
                   allOnes.containsCentre(zeroCentre));
        assertFalse("1E1N,+/-1 should not contain area of (big) zero-centred item",
                    allOnes.containsArea(zeroCentre));
        assertTrue("1E1N,+/-1 should contain centre of (point) zero-centred item",
                   allOnes.containsCentre(allZeros));
        assertTrue("1E1N,+/-1 should contain area of (point) zero-centred item",
                   allOnes.containsArea(allZeros));
        assertTrue("1E1N,+/-1 should strictly contain area of (point) zero-centred item",
                   allOnes.strictlyContainsArea(allZeros));
        }




    /**Test that we can look up exhibits by name in the sample auxInfo file.
     * FIXME: Also needs stand-alone tests that do not depend on the
     * vast config file.
     * <p>
     * We skip these tests if we do not have full/normal access
     * to the filesystem.
     */
    public void testLocationLookup()
        throws Exception
        {
        if(!Main.isAccessToFilesystem())
            {
            System.err.println("No access to filesystem: skipping tests: " + getName());
            return;
            }

        final ExhibitPropsGlobalImmutable epgi = ExhibitPropsGlobalImmutable.loadFromDataDir(new File(LocalProps.getDataDir()));
        final LocationMap lm = epgi.getLocationMap();
        assertFalse("Loaded LocationMap must not be empty", lm.isEmpty());


        // Looking up this data item...
//    location.places-and-sights/New-Zealand-.type=Estd
//    location.places-and-sights/New-Zealand-.E=172.5
//    location.places-and-sights/New-Zealand-.EErr=6.5
//    location.places-and-sights/New-Zealand-.N=-41
//    location.places-and-sights/New-Zealand-.NErr=6.5
        // List of exhibits all of which we ought to be able to
        // get general (non-specific) location info for,
        // either forwards or reverse.
        // (We assume no attribute words for this test.)
        final String[] testExhibits =
            {
               "places-and-sights/New-Zealand-random-DHD.gif",
               "places-and-sights/_more2003/_more04/New-Zealand-South-Island-west-coast-Milford-Sound-fjord-taken-20001213-waterfall-tumbling-down-steep-vegetated-bank-rotated-MB.jpg",
               "places-and-sights/Earth-DHD.gif",
//               "places-and-sights/Earth-whole-planet-DHD.gif", // Is specific!
               "places-and-sights/England-London-made-up-name-Z.wav",
               "natural-science/_more2005/_more07/bee-bumblebee-bumble-bee-pollinating-pollination-of-lavender-purple-flowers-from-bush-flowering-in-July-in-London-England-8-DHD.jpg",
               "natural-science/_more2005/_more07/bee-bumblebee-bumble-bee-pollinating-pollination-of-lavender-purple-flowers-from-bush-flowering-in-July-in-London-England-closeup-9-DHD.jpg",
            };
        // Create enough attribute words to cover our examples.
        final Set<String> attrWords = new HashSet<String>(Arrays.asList(new String[]{
            "closeup", "rotated"
            }));
        for(int i = testExhibits.length; --i >= 0; )
            {
            final String name = testExhibits[i];
            final Location.Base te1 = lm.locLookup(name, attrWords);
            assertNotNull("Result of LocationMap lookup (te1) must not be null: " + name,
                          te1);
            assertFalse("Result of te1 lookup should not be specific for " + name + ": " + te1,
                        te1.specific);
            assertTrue("Result of te1 lookup should be Estd for " + name + ": " + te1,
                       te1 instanceof Location.Estd);
            final Location.Estd te1E = (Location.Estd) te1;
            assertTrue("Result of (te1) lookup should be on full Aloha Earth map",
                       (new AEParams()).getLocation(true).containsCentre(te1E));
            assertFalse("No looked-up item should be larger than WHOLE_EARTH",
                        te1E.isLargerThan(Location.Estd.WHOLE_EARTH));
            }
        }



    /**Test that we can use the geographical-proximity-by-ccTLD data.
     * This tests that we can find close-by countries (in connectivity terms).
     */
    public static void testGeoProximity()
        {
        // Look for "gb"/"uk" to be in the same group.
        // This also establishes that we can load at least the default set.
        final GeoUtils.CCTLD uk = new GeoUtils.CCTLD("uk");
        final GeoUtils.CCTLD us = new GeoUtils.CCTLD("us");
        final List<GeoUtils.CCTLD> group = Arrays.asList(new GeoUtils.CCTLD[]{
            uk, new GeoUtils.CCTLD("gb")});
        final Set<GeoUtils.CCTLD> groupS = new HashSet<GeoUtils.CCTLD>(group);

        for(final Iterator<GeoUtils.CCTLD> it = groupS.iterator(); it.hasNext(); )
            {
            final GeoUtils.CCTLD cc = it.next();
            assertTrue("Must be able to find all of our group "+group+" when we look up " + cc,
                       GeoUtils.getCloseCCTLDs(cc).containsAll(groupS));
            }

        // Check that "uk" is in the "RIPE" region.
        assertTrue("Expect to find uk in the RIPE region",
                   GeoUtils.getCountriesInRegion("RIPE").contains(uk));

        // Check a few expected proximities.
        assertEquals("The same country should have COUNTRY proximity to itself",
                     GeoProximity.COUNTRY, GeoUtils.computeProximity("uk", uk));
        assertEquals("The same country should have COUNTRY proximity to itself",
                     GeoProximity.COUNTRY, GeoUtils.computeProximity("us", us));
        assertEquals("Well-connected countries should have COUNTRY proximity",
                     GeoProximity.COUNTRYGROUP, GeoUtils.computeProximity("uk", new GeoUtils.CCTLD("fr")));
        assertEquals("Well-connected countries should have COUNTRY proximity",
                     GeoProximity.COUNTRYGROUP, GeoUtils.computeProximity("nl", new GeoUtils.CCTLD("de")));
        assertEquals("Well-connected countries should have COUNTRY proximity",
                     GeoProximity.COUNTRYGROUP, GeoUtils.computeProximity("ca", us));
        assertEquals("A region/registry should have CONTINENT proximity to a country in it",
                     GeoProximity.CONTINENT, GeoUtils.computeProximity("RIPE", uk));
        assertEquals("A region/registry should have CONTINENT proximity to a country in it",
                     GeoProximity.CONTINENT, GeoUtils.computeProximity("ARIN", us));
        assertEquals("A region/registry should have NONE proximity to a country not in it",
                     GeoProximity.NONE, GeoUtils.computeProximity("ARIN", uk));
        assertEquals("A region/registry should have NONE proximity to a country not in it",
                     GeoProximity.NONE, GeoUtils.computeProximity("RIPE", us));
        assertEquals("A region/registry should have NONE proximity to a (non-existant) country not in it",
                     GeoProximity.NONE, GeoUtils.computeProximity("RIPE", new GeoUtils.CCTLD("xx")));
        assertEquals("A (non-existant) region/registry should have NONE proximity to a country not in it",
                     GeoProximity.NONE, GeoUtils.computeProximity("NONESUCH", us));
        assertEquals("A (non-existant) region/registry should have NONE proximity to a (non-existant) country not in it",
                     GeoProximity.NONE, GeoUtils.computeProximity("NONESUCH", new GeoUtils.CCTLD("xx")));
        }

    /**Check lookup of geographic location of IP address.
     * This checks that basic lookup of location by IP address
     * (ie in which country is the machine/interface with that address)
     * works for some simple basic cases.
     * <p>
     * TODO: whitebox check that mappings in source data are actually visible.
     */
    public static void testRegionLookup()
        throws Exception
        {
        final long t0 = System.currentTimeMillis();

        // Do quick region/ccTLD lookups for a couple of addresses whose location is known
        // and should be pretty stable...
        final String reg1 = GeoUtils.getRegionByAddress(InetAddress.getByName("195.137.27.135"), true);
        assertTrue("Must be able to look up known-UK address quickly",
                   ("RIPE".equals(reg1) || ("uk".equals(reg1))));
        final String reg2 = GeoUtils.getRegionByAddress(InetAddress.getByName("72.9.239.218"), true);
        assertTrue("Must be able to look up known-US address quickly",
                   ("ARIN".equals(reg2) || ("us".equals(reg2))));
//        final String reg3 = GeoUtils.getRegionByAddress(InetAddress.getByName("193.178.223.10"), true);
//        assertTrue("Must be able to look up known-UK address quickly",
//                   ("RIPE".equals(reg3) || ("uk".equals(reg3))));

        final long t1 = System.currentTimeMillis();
        System.out.println("[Initial lookups: "+(t1-t0)+"ms.]"); // Can help with performance tuning.


        // Do region/ccTLD lookups for the
        // loop-back and other "private" addresses
        // which should never have a geographical association.

        // Handle the loop-back on its own for now.
        final InetAddress loopback = InetAddress.getByName("127.0.0.1");

        // Try to look up the region.
        // This should simply return some numeric prefix of the address
        // since the whole of the 127/8 block is non-geographic private space.
        // We try the quick flag randomly either way as it should not matter.
        final boolean quickLoopbackLookup = rnd.nextBoolean();
        final String region = GeoUtils.getRegionByAddress(loopback, quickLoopbackLookup);
        assertTrue("Region lookup should be 127 or start 127.",
                   "127".equals(region) || region.startsWith("127."));

        // Try to look up the ccTLD; the result should be null.
        // We try the quick flag randomly either way as it should not matter.
        final GeoUtils.CCTLD ccTLD = GeoUtils.getCCTLDByAddress(loopback, !quickLoopbackLookup);
        assertNull("We should not find a country code associated with the loopback address",
                   ccTLD);

        // Handle generic private/non-routable/non-geographic addresses.
        // These should in principle include all the RFC1597 addresses,
        // though in practice we may not be precise enough to handle
        // all of them correctly...
        final String nonGeoAddrs[] =
            { "127.0.0.1", "10.0.0.0", "172.16.0.0", "192.168.0.0" };
        final int indexOfFullLookup = rnd.nextInt(nonGeoAddrs.length);
        for(int i = nonGeoAddrs.length; --i >= 0; )
            {
            final InetAddress addr = InetAddress.getByName(nonGeoAddrs[i]);
            assertNull("ccTLD (quick) lookup for private address ("+addr+") should be null",
                       GeoUtils.getCCTLDByAddress(addr, true));
            // Do one of these at random as a full lookup to make sure it does not matter.
            final String regionByAddress = GeoUtils.getRegionByAddress(addr,
                                       (i != indexOfFullLookup));
            assertTrue("Region lookup for private address ("+addr+") should not be ccTLD nor registry/region name: " + regionByAddress,
                       !GeoUtils.CCTLD.isSyntaticallyValidCcTLD(regionByAddress) &&
                       !GeoUtils.isSyntaticallyValidRegistryName(regionByAddress));
            }

        // Check that at least one single-octet IPv4 prefix
        // returns a non-numeric region and/or a non-null ccTLD
        // even when doing a "quick" lookup with internal data only
        // (or at least not going to the Net for further data).
        boolean foundCcTLD = false;
        boolean foundRegion = false;
        for(int i = 256; --i >= 0; )
            {
            final String dottedQuad = String.valueOf(i) + ".0.0.1";
            final InetAddress ia = InetAddress.getByName(dottedQuad);

            // We do a quick lookup of ccTLD, ie by internal data only.
            final GeoUtils.CCTLD cc = GeoUtils.getCCTLDByAddress(ia, true);
            if(null != cc)
                { foundCcTLD = true; }

            // We do a quick lookup of region, ie by internal data only.
            final String reg = GeoUtils.getRegionByAddress(ia, true);
            assertNotNull("region must not be null", reg);
            assertTrue("region must not be empty String", reg.length() > 0);
            for(int r = reg.length(); --r >= 0; )
                {
                final char rc = reg.charAt(r);
                assertTrue("all region characters must be printable ASCII",
                           (rc > 32) && (rc < 127));
                }
            final char regFirstChar = reg.charAt(0);
            assertTrue("region must start with a digit or letter",
                       Character.isDigit(regFirstChar) ||
                       Character.isLetter(regFirstChar));
            if(!Character.isDigit(regFirstChar))
                { foundRegion = true; }
//Main.getOut().println("[cc/reg="+cc+"/"+reg+"]");
            }
        assertTrue("Must find at least one non-null ccTLD by IP address", foundCcTLD);
        assertTrue("Must find at least one non-numeric region by IP address", foundRegion);

//        // Try doing sample lookups on the "local" address...
//        // This is no more of a test than to ensure we don't blow up!
//        Main.getOut().println("Quick lookup on local address gives: "
//            + GeoUtils.getRegionByAddress(InetAddress.getLocalHost(), true));
//        Main.getOut().println("Full lookup on local address gives: "
//            + GeoUtils.getRegionByAddress(InetAddress.getLocalHost(), false));
        }

    /**Check DNS lookup.
     * If DNS lookup is working at all,
     * check that forward and reverse lookup is working as expected.
     */
    public static void testDNSLookup()
        {
        // Set of names to look up addresses for,
        // then lookup in reverse to get back the names for.
        // We pick addresses that are pretty likely to work correctly for this case!
        final String testNames[] = {
            "www.ibm.com.",
            "mirror-cn-bj1.gallery.hd.org.",
            "www.parliament.uk.",
            };
        try
            {
            for(final String n : testNames)
                {
//                final InetAddress addrs[] = InetAddress.getAllByName(n);
                final InetAddress addr = InetAddress.getByName(n);
                final String rlName = AddrTools.doReverseLookup(addr, false);
                if(rlName == null)
	                {
	                Main.getOut().println("Reverse lookup currently not possible...");
	                continue;
	                }
                assertEquals("Reverse lookup should return orginal name", n, rlName);
                final String rlName2 = AddrTools.doReverseLookup(addr, true);
                if(rlName2 == null)
	                {
	                Main.getOut().println("Reverse lookup currently not possible (2)...");
	                continue;
	                }
                assertEquals("Verifying reverse lookup should return orginal name", n, rlName2);
                }
            }
        catch(final UnknownHostException e)
            {
            e.printStackTrace(Main.getErr());
            Main.getErr().println("WARNING: lookup failed: aborting all DNS lookup tests since DNS may be unavailable...");
            }
        }

    /**Test AddrPrefix class briefly.
     */
    public static void testAddrPrefix()
        {
        assertEquals("Should create correct padded dotted form", "001", new AddrTools.AddrPrefix(new byte[]{1}).toPaddedDottedPrefix());
        assertEquals("Should create correct padded dotted form", "213.092", new AddrTools.AddrPrefix(new byte[]{(byte)213, 92}).toPaddedDottedPrefix());
        assertEquals("Should create correct padded dotted form", "004.000.255", new AddrTools.AddrPrefix(new byte[]{4, 0, (byte)255}).toPaddedDottedPrefix());
        }

    /**Test some simple cases for AddrPrefix dotted-quad parsing. */
    public static void testAddrPrefixParsing()
        {
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127")).toByteArray(), new byte[]{127}));
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127.000")).toByteArray(), new byte[]{127,0}));
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127.00")).toByteArray(), new byte[]{127,0}));
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127.0")).toByteArray(), new byte[]{127,0}));
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127.0.0")).toByteArray(), new byte[]{127,0,0}));
        assertTrue("Should be able to parse prefix correctly", Arrays.equals((new AddrTools.AddrPrefix("127.0.0.1")).toByteArray(), new byte[]{127,0,0,1}));

        // Test a few random prefixes/addresses of various length,
        // with sub-3-digit octets randomly padded or not.
        final StringBuilder sb = new StringBuilder();
        final StringBuilder o = new StringBuilder();
        for(int i = 20000; --i >= 0; )
            {
            final int length = 1 + rnd.nextInt(4);
            final byte[] data = new byte[length];
            rnd.nextBytes(data); // Fill in random bytes.

            sb.setLength(0);
            for(int j = 0; j < length; ++j)
                {
                o.setLength(0);
                o.append(data[j] & 0xff);
                while((o.length() < 3) && rnd.nextBoolean())
                    { o.insert(0, '0'); }
                if(sb.length() != 0) { sb.append('.'); }
                sb.append(o);
                }

            final String s = sb.toString();
            assertTrue("Should be able to parse prefix correctly: "+s, Arrays.equals((new AddrTools.AddrPrefix(s)).toByteArray(), data));
            }
        }

    /**Check AddrPrefix.hashCode() effectiveness, as this may be important for performance.
     * We test this against any real data we have to hand
     * plus a few choice manually-selected values.
     * <p>
     * The aim is to ensure that there are not excessive collisions
     * for real data values, as that would make HashSet and related containers
     * behave badly.
     */
    public static void testAddrPrefixHashCode()
        {
        // Maximum collisions that we will permit on any one hash value.
        final int MAX_COLLISIONS = 3;

        // Get addresses from built-in IP-->ccTLD map first.
        final SortedSet<AddrTools.AddrPrefix> testPrefixes =
            new TreeSet<AddrTools.AddrPrefix>(GeoUtils.getCcTLDFromIPPrefix().keySet());
        // Add a few more potentially-important values by hand.
        testPrefixes.add(new AddrTools.AddrPrefix("127.0.0.1"));
        testPrefixes.add(new AddrTools.AddrPrefix("127.0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("127.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("127"));
        testPrefixes.add(new AddrTools.AddrPrefix("0.0.0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("0.0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("0"));
        testPrefixes.add(new AddrTools.AddrPrefix("10.0.0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("10.0.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("10.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("10"));
        testPrefixes.add(new AddrTools.AddrPrefix("192.168"));
        testPrefixes.add(new AddrTools.AddrPrefix("192.168.0"));
        testPrefixes.add(new AddrTools.AddrPrefix("192.168.1"));
        testPrefixes.add(new AddrTools.AddrPrefix("192.168.100"));
        testPrefixes.add(new AddrTools.AddrPrefix("192.168.101"));
        testPrefixes.add(new AddrTools.AddrPrefix("172.16"));
        final int tps = testPrefixes.size();
        Main.getOut().println("[AddrPrefix sample size: " + tps + ".]");

        // All hash values seen so far, and what they are for.
        // Test hash quality across full Integer range.
        final HashMap<Integer,List<AddrTools.AddrPrefix>> hashes = new HashMap<Integer, List<AddrTools.AddrPrefix>>(2*tps);
        int collisionCount = 0;
        // Test hash quality across minimum-possible range (ie the set size).
        final BitSet constrainedHashes = new BitSet(tps);
        int collisionCountConstrained = 0;
        for(final AddrTools.AddrPrefix ap : testPrefixes)
            {
            final Integer h = new Integer(ap.hashCode());

            List<AddrTools.AddrPrefix> collisions = hashes.get(h);
            if(collisions == null)
                {
                collisions = new ArrayList<AddrTools.AddrPrefix>(1);
                hashes.put(h, collisions);
                }
            else
                {
                ++collisionCount;
                Main.getOut().println("Hash collision between "+ap+" and "+collisions);
                if(collisions.size() >= MAX_COLLISIONS-1)
                    { fail("TOO MANY COLLISIONS"); }
                }
            collisions.add(ap);

            // Test very harsh constrained conditions.
            final int hc = Math.abs(ap.hashCode() % tps);
            if(constrainedHashes.get(hc)) { ++collisionCountConstrained; }
            else { constrainedHashes.set(hc); }
            }

        Main.getOut().println("[Hash collisions: " + collisionCount + " = " + ((100.0f * collisionCount) / tps) + "%.]");
        Main.getOut().println("[Maximally constrained hash collisions: " + collisionCountConstrained + " = " + ((100.0f * collisionCountConstrained) / tps) + "%.]");
        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
