/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.test.dev;

import junit.framework.*;

import org.hd.d.pg2k.svrCore.ThreadUtils.DaemonThreadFactory;
import org.hd.d.pg2k.svrCore.props.LocalProps;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.Writer;
import java.net.URL;
import java.util.Enumeration;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**This is the main entry point for development-time tests.
 * This is also a valid top-level JUnit test class.
 * <p>
 * They can be run from the command-line or IDE to test most self-contained features,
 * or from a WAR to allow self-test in an HTTP environment.
 * <p>
 * (In a WAR environment the out/err values and exitWhenDone/inWAR flags
 * should be set appropriately.)
 */
public final class Main extends TestCase
    {
    /**If true, this indicates that we have access to the filesystem for tests that need it. */
    private static final boolean _accessToFilesystem =
        (LocalProps.getTimestamp() > 0);

    /**If true, this indicates that we have access to the filesystem for tests that need it.
     * We deduce this by testing if we can load the local proprties file,
     * which we only test at most once per run to avoid lots of log messages.
     * Access to the file can be arranged by having the tests run in the correct
     * working directory, or by setting a system property to indicate its
     * location.
     * <p>
     * Tests which expect access to configuration files, test data, etc,
     * should avoid running when this does not return true.
     */
    public static boolean isAccessToFilesystem()
        {
        return(_accessToFilesystem);
        }

    /**Note availablity of filesystem test data. */
    static
        {
        System.out.println("isAccessToFilesystem for tests: " + _accessToFilesystem);
        }

    /**Get URL for our mount-point if running as a WAR; or null if none.
     * This is used for loop-back testing when running in a WAR.
     * <p>
     * Usually a top-level URL such as "http://localhost:8080/".
     */
    public synchronized static URL getLoopbackURL()
        {
        return(loopbackURL);
        }

    /**Set URL for our mount-point if running as a WAR; or null if none.
     * This is used for loop-back testing when running in a WAR.
     */
    public synchronized static void setLoopbackURL(URL loopbackURL)
        {
        Main.loopbackURL = loopbackURL;
        }

    /**URL for our mount-point if running as a WAR; null if none.
     * This is used for loop-back testing when running in a WAR.
     */
    private static URL loopbackURL;

    /**Get current ServletContext if running in a WAR; null if none.
     * Declared as an Object to avoid making Main depend on Servlet API.
     */
    public synchronized static Object getServletContext()
        {
        return(servletContext);
        }

    /**Set current ServletContext if running in a WAR; null if none.
     * Declared as an Object to avoid making Main depend on Servlet API.
     */
    public synchronized static void setServletContext(Object servletContext)
        {
        Main.servletContext = servletContext;
        }

    /**Current ServletContext if running in a WAR; null if none.
     * Declared as an Object to avoid making Main depend on Servlet API.
     */
    private static Object servletContext;

    /**Get stream to use in lieu of System.out for jUnit tests; never null. */
    public synchronized static PrintStream getOut()
        {
        return out;
        }

    /**Set the out stream; never null. */
    private synchronized static void setOut(PrintStream o)
        {
        if(o == null) { throw new IllegalArgumentException(); }
        out = o;
        }

    /**Get stream to use in lieu of System.err for jUnit tests; never null. */
    public synchronized static PrintStream getErr()
        {
        return err;
        }

    /**Set the err stream; never null. */
    private synchronized static void setErr(PrintStream e)
        {
        if(e == null) { throw new IllegalArgumentException(); }
        err = e;
        }

    /**The <code>PrintStream</code> to direct standard output to.
     * Defaults to <code>System.out</code>.
     * Could be over-ridden by a servlet etc to send output back to the client.
     * <p>
     * This is package-visible only for safety.
     */
    private static PrintStream out = System.out;

    /**The <code>PrintStream</code> to direct standard output to.
     * Defaults to <code>System.out</code>.
     * Could be over-ridden by a servlet etc to send output back to the client.
     * <p>
     * This is package-visible only for safety.
     */
    private static PrintStream err = System.err;


    /**Main entry point from command-line.
     * Normally we try to run all tests and see how many produced errors.
     * <p>
     * If the first arg is -forever then the tests are run continuously
     * until the process is killed.
     * <p>
     * If one or more test classes are given on the command-line then
     * those tests alone are run, else all tests are run.
     * <p>
     * Only one top-level test routine may run at any one time.
     */
    public static final void main(String args[])
        {
        Main.getOut().println("STARTED DEV TESTS FROM COMMAND-LINE...");

        // Do we run the tests forever?
        final boolean forever = (args.length > 0) &&
                                "-forever".equals(args[0]);

        boolean allOK = runTests(forever, System.out, System.err, null, null);

        // Exit with an error return value if the error count is not zero.
        System.exit((allOK) ? 0 : 1);
        }

    /**Run the unit tests from a servlet.
     * Takes a Writer argument (assumed to be the servlet output stream)
     * and wraps it to take output from both err and out streams.
     */
    public static void runTestsFromServlet(final Writer writer,
                                           final URL loopbackURL,
                                           final Object servletContext)
        {
        if((writer == null) ||
           (loopbackURL == null) ||
           (servletContext == null))
            { throw new IllegalArgumentException(); }

        // Make a minimal (and fairly inefficient) wrapper for the Writer.
        final PrintStream ps = new PrintStream(new OutputStream()
            {
            /**Writes the specified byte to this output stream.
             * @param      b   the <code>byte</code>.
             * @exception  IOException  if an I/O error occurs
             */
            public void write(int b) throws IOException
                { writer.write(b); }

            /**Flushes this output stream.
             * @exception  IOException  if an I/O error occurs.
             */
            public void flush() throws IOException
                { writer.flush(); }

            }, true); // Force a flush on newline...

        runTests(false, ps, ps, loopbackURL, servletContext);
        }

    /**Flattens test case structure from src to dest.
     */
    private static void flattenSuite(final TestSuite dest,
                                     final TestSuite src)
        {
        final Enumeration en = src.tests();
        while(en.hasMoreElements())
            {
            final Test t = (Test) en.nextElement();
            if(t instanceof TestSuite)
                {
                flattenSuite(dest, (TestSuite) t);
                }
            else
                {
                dest.addTest(t);
                }
            }
        }

    /**Actually runs the unit tests.
     * This is synchronized to only also one set of tests to be run at once
     * to ensure consistent behaviour and limit resource usage if run from
     * a servlet.
     * <p>
     * Output and error streams supplied replace the normal ones for the
     * duration of the run; they must not be null.  The previous values
     * are restored upon completion.
     *
     * @param forever  if true, tests are run repeatedly until an error is found
     * @return true if all tests ran OK
     */
    public static synchronized boolean runTests(final boolean forever,
                                                final PrintStream o,
                                                final PrintStream e,
                                                final URL loopbackURL,
                                                final Object servletContext)
        {
        final PrintStream originalOut = getOut();
        final PrintStream originalErr = getErr();
        final URL originalLoopbackURL = getLoopbackURL();
        final Object originalServletContext = getServletContext();

        try
            {
            // Set the requested streams to run the tests.
            setOut(o);
            setErr(e);
            setLoopbackURL(loopbackURL);
            setServletContext(servletContext);

            boolean allOK = false; // Assume failed.
            int errorCount = 0;

            do
                {
                // Run all the sub-tests.
                final TestSuite s = suite();

                getOut().println("Tests to run: " + s.countTestCases());

                final Enumeration te = s.tests();
                for(int i =  1; te.hasMoreElements(); ++i)
                    {
                    final TestResult r = new TestResult();
                    final TestCase test = (TestCase) te.nextElement();

                    getOut().println();
                    getOut().println("RUNNING TEST "+i+"; " + test.getName());
                    // Flush any output pending...
                    getOut().flush();
                    getErr().flush();

                    // Run the test...
                    test.run(r);

                    // Report any errors and failures.
                    if(r.wasSuccessful())
                        {
                        allOK = true;
                        getOut().println("SUCCEEDED");
                        }
                    if(r.errorCount() > 0)
                        {
                        allOK = false;
                        errorCount += r.errorCount();
                        Main.getErr().println("Errors: " + r.errorCount());
                        for(final Enumeration en = r.errors(); en.hasMoreElements(); )
                            {
                            final TestFailure tf = (TestFailure) en.nextElement();
                            Main.getErr().println(tf);
                            final Throwable t = tf.thrownException();
                            if(t != null)
                                { t.printStackTrace(Main.getErr()); }
                            }
                        }
                    if(r.failureCount() > 0)
                        {
                        allOK = false;
                        errorCount += r.failureCount();
                        Main.getErr().println("Failures: " + r.failureCount());
                        for(final Enumeration en = r.failures(); en.hasMoreElements(); )
                            {
                            Main.getErr().println(en.nextElement());
                            }
                        }
                    }

                // Flush any output pending...
                getOut().flush();
                getErr().flush();

                } while(forever && allOK); // Repeat tests forever if necessary.

            if(errorCount > 0)
                {
                getErr().println();
                getErr().println("TOTAL ERRORS: " + errorCount);
                }

            return(allOK);
            }
        finally
            {
            // Restore previous stream values.
            setOut(originalOut);
            setErr(originalErr);
            setLoopbackURL(originalLoopbackURL);
            setServletContext(originalServletContext);
            }
        }


    /**A factory for creating daemon pool threads.
     * Adapted (purloined) from Doug Lea / Sun JVM code.
     */
    public static final class TestThreadFactory implements ThreadFactory
        {
        private final ThreadGroup group;
        private final AtomicInteger threadNumber = new AtomicInteger(1);
        private final String namePrefix;
        private final boolean lowPriority;

        public TestThreadFactory(final String poolName,
                                 final boolean lowPriority)
            {
            final SecurityManager s = System.getSecurityManager();
            group = (s != null)? s.getThreadGroup() :
                                 Thread.currentThread().getThreadGroup();
            namePrefix = "pool-" + poolName + "-thread-";
            this.lowPriority = lowPriority;
            }

        public Thread newThread(final Runnable r)
            {
            final Thread t = new Thread(group, r, 
                                  namePrefix + threadNumber.getAndIncrement(),
                                  0);
            t.setDaemon(true);
            if(lowPriority)
                { t.setPriority(Thread.MIN_PRIORITY); }
            else if(t.getPriority() != Thread.NORM_PRIORITY)
                { t.setPriority(Thread.NORM_PRIORITY); }
            return(t);
            }
        }

    /**Number of available CPUs. */
    private static final int AVAILABLE_PROCESSORS = Runtime.getRuntime().availableProcessors();

    /**Shared thread pool for compute-bound tests.
     * Suitable for almost-entirely CPU-bound threads
     * with the pool size exactly the number of CPUs available to this JVM
     * (at creation of this pool).
     * <p>
     * Some work can be queued for this pool to maximise throughput
     * and excess work (with no free pool threads or queue space available)
     * will be run directly in the caller's thread.
     * In simple uses this will allow for a little parallel slackness.
     * <p>
     * This puts an upper bound on Thread resources (eg memory for stack)
     * consumed by threads in this pool.
     * <p>
     * The threads in the pool are daemon threads,
     * so should not prevent the JVM from exiting.
     * <p>
     * Only available to test routines in this package.
     */
    static final ExecutorService computeIntensiveTestThreadPool =
        Executors.unconfigurableExecutorService(
            new ThreadPoolExecutor(AVAILABLE_PROCESSORS, AVAILABLE_PROCESSORS,
                600L, TimeUnit.SECONDS, // Keep worker threads alive for 10 minutes...
                new ArrayBlockingQueue<Runnable>(AVAILABLE_PROCESSORS),
                new TestThreadFactory("Test Main.computeIntensiveThreadPool", false),
                new ThreadPoolExecutor.CallerRunsPolicy()));

    
    /**List of classes containing test suites to be run.
     * They should be run in the order specified by default.
     */
    private static final Class testClasses[] =
        {
            // Put the internal sanity tests first...
            TinyTest.class,

            // Bulk of the tests (alphabetic order by class)...
            AIImgScoreTest.class,
            AlohaEarthTest.class,
            BackCompatTest.class,
            ExhibitFilterTest.class,
            ExhibitNameTest.class,
            ExhibitUploadTest.class,
            HTTPTunnelTest.class,
            I18NTest.class,
            IndexTest.class,
            LocationTest.class,
            PaginationBeanTest.class,
            SearchPageJavaBeanTest.class,
            SerializationTest.class,
            SimpleCacheTest.class,
            SystemVariablesTest.class,
            TreeFilterBeanTest.class,
            VirtualHostTest.class,
            WebUtilsTest.class,

            // Put the slowest tests towards the end...
            MemoryToolsTest.class,
            MediaHandlerTest.class,

            // Leave this MiscTest set until last to encourage creation of specific test sets.
            MiscTest.class
        };

    /**This returns a suite of all the top-level tests that we want to perform.
     * This returns a flattened suite of tests.
     */
    public static TestSuite suite()
        {
        final TestSuite suite = new TestSuite();

        for(int i = 0; i < testClasses.length; ++i)
            { flattenSuite(suite, new TestSuite(testClasses[i])); }

        return(suite);
        }
    }
