/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.MIME.Handler;
import org.hd.d.pg2k.svrCore.MIME.Quantize;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.w3c.dom.Node;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.IndexColorModel;
import java.io.*;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

/**Tests of media-generation classes, eg for making thumbnails.
 * This will test features such as transcoding, etc.
 * <p>
 * Also tests some relevant features of the ExhibitMIME class.
 */
public final class MediaHandlerTest extends TestCase
    {
    public MediaHandlerTest(String name)
        {
        super(name);
        }

    /**Test some basic features of ExhibitMIME.
     * Including all exhibit type values being in range, in order, and unique.
     */
    public void testExhibitMIMEBasics()
        {
        // Make sure that all types within the max/min limits.
        final ExhibitMIME.ExhibitTypeParameters[] allTypes = ExhibitMIME.getAllValidExhibitTypes();
        assertTrue("ExhibitMIME.getAllValidExhibitTypes() must not return null", allTypes != null);

        int prevType = ExhibitMIME.ET__max + 1; // Previous type value.
        for(int i = allTypes.length; --i >= 0; )
            {
            final ExhibitMIME.ExhibitTypeParameters etp = allTypes[i];
            final int t = etp.type;
            assertTrue("All ExhibitTypeParameters values must be in max/min range",
                       ((t >= ExhibitMIME.ET__min) && (t <= ExhibitMIME.ET__max)));
            assertTrue("All ExhibitTypeParameters values must unique and in order",
                       t < prevType);
            prevType = t;
//            assertTrue("Minimum format overhead should be no smaller than magic",
//                       (etp.handler == null) ||
//                       (etp.handler.get() >= etp.magic.length()));
            }
        }

    /**Does some basic tests on the JPG (JPEG) handler.
     * For example, make sure that it always exists.
     */
    public void testJPGBasics()
        throws Exception
        {
        // Check that we have a .jpg handler.
        final ExhibitMIME.ExhibitTypeParameters type = ExhibitMIME.getParamsByType(ExhibitMIME.ET_JPEG);
        final Handler h = type.handler;
        assertTrue("Must always have a valid JPEG (.jpg) handler", h != null);
        // Check that .jpg handler is prepared to try to make thumbnails.
        assertTrue("JPEG (.jpg) handler should be prepared to make thumbnails", h.canMakeThumbnails());

        // Test that we can create a big image.
        final BufferedImage bigImage = makeTestRGBTrueColourBufferedImage(
            bigImageWidth, bigImageHeight);
        final byte[] bigJPEGImage =
            h.makeImageBinary(bigImage, 95);
            //makeJPEGImage(bigImageWidth, bigImageHeight);
        assertTrue("Must be able to create JPEG image from BufferedImage",
                   bigJPEGImage != null);
        assertTrue("Must be able to create valid JPEG image from BufferedImage",
                   ExhibitMIME.magicOK(type, new ByteArrayInputStream(bigJPEGImage)));
        assertTrue("Encoded image should be much bigger than 1kByte",
                   bigJPEGImage.length > 1024);

        final ExhibitThumbnails tn = h.makeThumbnails(
            new ByteArrayInputStream(bigJPEGImage),
            bigJPEGImage.length
            );
        assertTrue("Must be able to make JPEG thumbnails of some sort",
                   tn != null);
        assertTrue("Must be able to make small JPEG thumbnail",
                   tn.getSmall() != null);
        assertTrue("Must be able to make standard JPEG thumbnail",
                   tn.getStandard() != null);
        assertTrue("Small thumbnail must be smaller than standard thumbnail",
                   tn.getSmall().size() < tn.getStandard().size());
        assertTrue("Standard thumbnail must be smaller than original",
                   tn.getStandard().size() < bigJPEGImage.length);
        assertTrue("Small thumbnail must be no larger than size limit",
                   tn.getSmall().size() < ExhibitThumbnails.SML_ABS_MAX_BYTES);
        assertTrue("Standard thumbnail must be no larger than size limit",
                   tn.getStandard().size() < ExhibitThumbnails.STD_ABS_MAX_BYTES);

        // Fetch the metadata.
        // We expect this to be non-null for JPEG.
        final Node metadata = h.getMetadata(new ByteArrayInputStream(bigJPEGImage), "dummyName.jpg");
        assertNotNull("Standard meta-data should always exist for JPEG", metadata);
        System.out.println(TextUtils.toXML(metadata, false, false));
        }

    /**Does some basic tests on the JP2 (JPEG 2000) handler.
     * For example, make sure that it always exists.
     */
    public void testJP2Basics()
        throws Exception
        {
        // Check that we have a .jp2 handler.
        final ExhibitMIME.ExhibitTypeParameters type = ExhibitMIME.getParamsByType(ExhibitMIME.ET_JP2);
        final Handler h = type.handler;
        assertTrue("Must always have a valid JPEG 2000 (.jp2) handler", h != null);

        // Test that we can create a big image.
        final BufferedImage bigImage = makeTestRGBTrueColourBufferedImage(
            bigImageWidth, bigImageHeight);
        final byte[] bigJPEG2000Image =
            h.makeImageBinary(bigImage, 95);
        assertTrue("Must be able to create JPEG 2000 image from BufferedImage",
                   bigJPEG2000Image != null);
        assertTrue("Must be able to create valid JPEG 2000 image from BufferedImage",
                   ExhibitMIME.magicOK(type, new ByteArrayInputStream(bigJPEG2000Image)));
        assertTrue("Encoded image should be much bigger than 1kByte",
                   bigJPEG2000Image.length > 1024);

        // Fetch the metadata.
        // We expect this to be non-null for JPEG.
        final Node metadata = h.getMetadata(new ByteArrayInputStream(bigJPEG2000Image), "dummyName.jp2");
        assertNotNull("Standard meta-data should always exist for JPEG 2000", metadata);
        System.out.println(TextUtils.toXML(metadata, false, false));
        }

    /**Does some basic tests on the GIF handler.
     * Because of the Unisys LZW patent issues until 2004 this may not initially
     * be prepared to make thumbnails.
     */
    public void testGIFBasics()
        {
        // Check that we have a .gif handler.
        final Handler h = ExhibitMIME.getParamsByType(ExhibitMIME.ET_GIF).handler;
        assertTrue("Should have a valid GIF (.gif) handler", h != null);
        // Check that handler is prepared to try to make thumbnails.
        assertTrue("GIF handler should be prepared to make thumbnails", h.canMakeThumbnails());
        }

    /**Does some basic tests on the PNG handler.
     */
    public void testPNGBasics()
        {
        // Check that we have a .png handler.
        final Handler h = ExhibitMIME.getParamsByType(ExhibitMIME.ET_PNG).handler;
        assertTrue("Should have a valid PNG (.png) handler", h != null);
        // Check that handler is prepared to try to make thumbnails.
        assertTrue("PNG handler should be prepared to make thumbnails", h.canMakeThumbnails());
        }

    /**Test thumbnail generation of a given set of exhibit types.
     * In passing, tests that images can be generated at all.
     */
    public void testThumbnailGeneration()
        throws Exception
        {
        final int typesToTest[] =
            { ExhibitMIME.ET_PNG, ExhibitMIME.ET_JPEG, ExhibitMIME.ET_GIF };

        for(int i = typesToTest.length; --i >= 0; )
            {
            final int t = typesToTest[i];

            // Check that he have a handler.
            final ExhibitMIME.ExhibitTypeParameters type = ExhibitMIME.getParamsByType(t);
            final Handler h = type.handler;
            assertTrue("Must always have a valid handler for t = " + t, h != null);
            // Check that handler is prepared to try to make thumbnails.
            assertTrue("Handler should be prepared to make thumbnails: " + type.suffixForInputFile, h.canMakeThumbnails());

            // Test that we can create a big image.
            // Note that we use an indexed image for GIF tests.
            final BufferedImage bigImage = (t == ExhibitMIME.ET_GIF) ?
                    makeTestRGBIndexedColourBufferedImage(bigImageWidth, bigImageHeight) :
                    makeTestRGBTrueColourBufferedImage(bigImageWidth, bigImageHeight);
            final int bigImageLongestEdge = Math.max(bigImageWidth, bigImageHeight);
            final byte[] bigEncodedImage =
                h.makeImageBinary(bigImage, 90);
                //makeJPEGImage(bigImageWidth, bigImageHeight);
            assertTrue("Must be able to create image from BufferedImage",
                       bigEncodedImage != null);
            assertTrue("Must be able to create valid image from BufferedImage",
                       ExhibitMIME.magicOK(type, new ByteArrayInputStream(bigEncodedImage)));
            assertTrue("Encoded image should be much bigger than 1kByte",
                       bigEncodedImage.length > 1024);

            final ExhibitThumbnails tn = h.makeThumbnails(
                new ByteArrayInputStream(bigEncodedImage),
                bigEncodedImage.length);
            assertTrue("Must be able to make thumbnails of some sort",
                       tn != null);
            assertTrue("Must be able to make small thumbnail",
                       tn.getSmall() != null);
            assertTrue("Must be able to make standard thumbnail",
                       tn.getStandard() != null);
            assertTrue("Small thumbnail must be smaller than standard thumbnail",
                       tn.getSmall().size() < tn.getStandard().size());
            assertTrue("Standard thumbnail should be smaller than original",
                       tn.getStandard().size() < bigEncodedImage.length);
            assertTrue("Small thumbnail must be no larger than size limit",
                       tn.getSmall().size() < ExhibitThumbnails.SML_ABS_MAX_BYTES);
            assertTrue("Standard thumbnail must be no larger than size limit",
                       tn.getStandard().size() < ExhibitThumbnails.STD_ABS_MAX_BYTES);
            final Dimension stdTNDim = h.get2DImageDimensions(
                new ByteArrayInputStream(tn.getStandard().toByteArrray()));
            final int stdTNLongestEdge = Math.max(stdTNDim.width, stdTNDim.height);
            assertTrue("Standard thumbnail should have correct longest edge",
                       stdTNLongestEdge == Math.min(ExhibitThumbnails.STD_STATIC_IMAGE_TN_LDIM_PX,
                                                    bigImageLongestEdge));
            final Dimension smlTNDim = h.get2DImageDimensions(
                new ByteArrayInputStream(tn.getSmall().toByteArrray()));
            final int smlTNLongestEdge = Math.max(smlTNDim.width, smlTNDim.height);
            assertTrue("Small thumbnail should have correct longest edge",
                       smlTNLongestEdge == Math.min(ExhibitThumbnails.SML_STATIC_IMAGE_TN_LDIM_PX,
                                                    bigImageLongestEdge));            }
        }

    /**Test ability to accurately extract image dimensions from raw binaries.
     * In passing, we will make sure that the overhead for each format is
     * no larger than the (presumed valid, and very small) test images.
     */
    public void testDimensionExtraction()
        throws IOException
        {
        // Dimensions for random-sized JPEG
        // with the dimensions sometimes needing two bytes to hold.
        final int jpegX = 64 + rnd.nextInt(512);
        final int jpegY = 64 + rnd.nextInt(512);

        // Test data: exhibit type, XxY dimensions, raw data as byte array.
        final class DataRow
            {
            public final int exhibitType;
            public final Dimension xyPixels;
            /**Raw exhibit data. */
            public final byte data[];

            /**Construct a test case.
             * Does not defensively copy data as assumed to be in
             * benign environment.
             */
            DataRow(int type, Dimension xy, byte rawData[])
                {
                exhibitType = type;
                xyPixels = xy;
                data = rawData;
                }
            }

        final DataRow testData[] =
            {
            // Simple 2x1 GIF image.
            new DataRow(ExhibitMIME.ET_GIF, new Dimension(2, 1), new byte[]{
0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x2, 0x0,
0x1, 0x0, (byte) 0xb3, 0x0, 0x0, (byte) 0xff, (byte) 0x99, 0x66,
(byte) 0xff, (byte) 0x99, (byte) 0x99, (byte) 0xff, (byte) 0x99, (byte) 0xcc, (byte) 0xff, (byte) 0x99,
(byte) 0xff, (byte) 0xff, (byte) 0xcc, 0x0, (byte) 0xff, (byte) 0xcc, 0x33, (byte) 0xff,
(byte) 0xcc, 0x66, (byte) 0xff, (byte) 0xcc, (byte) 0x99, (byte) 0xff, (byte) 0xcc, (byte) 0xcc,
(byte) 0xff, (byte) 0xcc, (byte) 0xff, (byte) 0xff, (byte) 0xff, 0x0, (byte) 0xff, (byte) 0xff,
0x33, (byte) 0xff, (byte) 0xff, 0x66, (byte) 0xff, (byte) 0xff, (byte) 0x99, (byte) 0xff,
(byte) 0xff, (byte) 0xcc, (byte) 0xff, (byte) 0xff, (byte) 0xff, 0x2c, 0x0, 0x0,
0x0, 0x0, 0x2, 0x0, 0x1, 0x0, 0x0, 0x4,
0x3, (byte) 0xf0, (byte) 0xbd, 0x8, 0x0, 0x3b
                }),

            // Simple 2x1 PNG image.
            new DataRow(ExhibitMIME.ET_PNG, new Dimension(2, 1), new byte[]{
(byte) 0x89, 0x50, 0x4e, 0x47, 0xd, 0xa, 0x1a, 0xa,
0x0, 0x0, 0x0, 0xd, 0x49, 0x48, 0x44, 0x52,
0x0, 0x0, 0x0, 0x2, 0x0, 0x0, 0x0, 0x1,
0x4, 0x3, 0x0, 0x0, 0x0, 0x6, 0xc, 0x62,
(byte) 0xb9, 0x0, 0x0, 0x0, 0x2c, 0x74, 0x45, 0x58,
0x74, 0x43, 0x72, 0x65, 0x61, 0x74, 0x69, 0x6f,
0x6e, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54,
0x75, 0x65, 0x20, 0x32, 0x36, 0x20, 0x41, 0x75,
0x67, 0x20, 0x32, 0x30, 0x30, 0x33, 0x20, 0x32,
0x30, 0x3a, 0x31, 0x34, 0x3a, 0x35, 0x34, 0x20,
0x2d, 0x30, 0x30, 0x30, 0x30, (byte) 0xe5, (byte) 0xc3, 0x1c,
0x21, 0x0, 0x0, 0x0, 0x7, 0x74, 0x49, 0x4d,
0x45, 0x7, (byte) 0xd3, 0x8, 0x1a, 0x13, 0x14, 0x36,
(byte) 0xa8, (byte) 0xde, 0x67, (byte) 0xfe, 0x0, 0x0, 0x0, 0x9,
0x70, 0x48, 0x59, 0x73, 0x0, 0x0, 0xb, 0x12,
0x0, 0x0, 0xb, 0x12, 0x1, (byte) 0xd2, (byte) 0xdd, 0x7e,
(byte) 0xfc, 0x0, 0x0, 0x0, 0x4, 0x67, 0x41, 0x4d,
0x41, 0x0, 0x0, (byte) 0xb1, (byte) 0x8f, 0xb, (byte) 0xfc, 0x61,
0x5, 0x0, 0x0, 0x0, 0x3, 0x50, 0x4c, 0x54,
0x45, (byte) 0xff, (byte) 0xff, (byte) 0xff, (byte) 0xa7, (byte) 0xc4, 0x1b, (byte) 0xc8,
0x0, 0x0, 0x0, 0xa, 0x49, 0x44, 0x41, 0x54,
0x78, (byte) 0xda, 0x63, 0x60, 0x0, 0x0, 0x0, 0x2,
0x0, 0x1, (byte) 0xe5, 0x27, (byte) 0xde, (byte) 0xfc, 0x0, 0x0,
0x0, 0x0, 0x49, 0x45, 0x4e, 0x44, (byte) 0xae, 0x42,
0x60, (byte) 0x82
                }),

            // Simple 2x1 JPG image.
            new DataRow(ExhibitMIME.ET_JPEG, new Dimension(2, 1), new byte[]{
(byte) 0xff, (byte) 0xd8, (byte) 0xff, (byte) 0xe0, 0x0, 0x10, 0x4a, 0x46,
0x49, 0x46, 0x0, 0x1, 0x1, 0x0, 0x1, 0x2c,
0x1, 0x2c, 0x0, 0x0, (byte) 0xff, (byte) 0xdb, 0x0, 0x43,
0x0, 0x6, 0x4, 0x5, 0x6, 0x5, 0x4, 0x6,
0x6, 0x5, 0x6, 0x7, 0x7, 0x6, 0x8, 0xa,
0x10, 0xa, 0xa, 0x9, 0x9, 0xa, 0x14, 0xe,
0xf, 0xc, 0x10, 0x17, 0x14, 0x18, 0x18, 0x17,
0x14, 0x16, 0x16, 0x1a, 0x1d, 0x25, 0x1f, 0x1a,
0x1b, 0x23, 0x1c, 0x16, 0x16, 0x20, 0x2c, 0x20,
0x23, 0x26, 0x27, 0x29, 0x2a, 0x29, 0x19, 0x1f,
0x2d, 0x30, 0x2d, 0x28, 0x30, 0x25, 0x28, 0x29,
0x28, (byte) 0xff, (byte) 0xdb, 0x0, 0x43, 0x1, 0x7, 0x7,
0x7, 0xa, 0x8, 0xa, 0x13, 0xa, 0xa, 0x13,
0x28, 0x1a, 0x16, 0x1a, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28,
0x28, 0x28, 0x28, 0x28, 0x28, 0x28, (byte) 0xff, (byte) 0xc0,
0x0, 0x11, 0x8, 0x0, 0x1, 0x0, 0x2, 0x3,
0x1, 0x22, 0x0, 0x2, 0x11, 0x1, 0x3, 0x11,
0x1, (byte) 0xff, (byte) 0xc4, 0x0, 0x15, 0x0, 0x1, 0x1,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8,
(byte) 0xff, (byte) 0xc4, 0x0, 0x14, 0x10, 0x1, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, (byte) 0xff, (byte) 0xc4,
0x0, 0x14, 0x1, 0x1, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, (byte) 0xff, (byte) 0xc4, 0x0, 0x14,
0x11, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
0x0, 0x0, (byte) 0xff, (byte) 0xda, 0x0, 0xc, 0x3, 0x1,
0x0, 0x2, 0x11, 0x3, 0x11, 0x0, 0x3f, 0x0,
(byte) 0xaa, 0x40, 0x7, (byte) 0xff, (byte) 0xd9
                }),

            // Random-sized JPG image.
            new DataRow(ExhibitMIME.ET_JPEG, new Dimension(jpegX, jpegY),
                        ExhibitMIME.getParamsByType(ExhibitMIME.ET_JPEG).handler.
                            makeImageBinary(
                                makeTestRGBTrueColourBufferedImage(jpegX, jpegY), 50)),

            // Random-sized PNG image.
            new DataRow(ExhibitMIME.ET_PNG, new Dimension(jpegX, jpegY),
                        ExhibitMIME.getParamsByType(ExhibitMIME.ET_PNG).handler.
                            makeImageBinary(
                                makeTestRGBTrueColourBufferedImage(jpegX, jpegY), 50)),
            };

        // Passes data to given handler (which must exist)
        // to test that it can correctly extract (X,Y) pixel dimensions.
        for(int i = testData.length; --i >= 0; )
            {
            final DataRow dr = testData[i];
            final ExhibitMIME.ExhibitTypeParameters etp = ExhibitMIME.getParamsByType(dr.exhibitType);
            assertTrue("ExhibitMIME.getParamsByType() must find parameters for type: " + dr.exhibitType,
                       etp != null);
            assertTrue("ExhibitMIME.getParamsByType() must find parameters for correct type",
                       etp.type == dr.exhibitType);
            assertTrue("HandlerBase must exist for type: " + etp, etp.handler != null);
            final Dimension dim = etp.handler.get2DImageDimensions(new ByteArrayInputStream(dr.data));
            assertTrue("Wrong Dimension extracted",
                       dr.xyPixels.equals(dim));
//            assertTrue("Minimum overhead is smaller than magic for: " + etp,
//                       etp.handler.approxMinOverheadBytes() >= etp.magic.length());
//            assertTrue("Minimum overhead is larger than test exhibit for: " + etp,
//                       etp.handler.approxMinOverheadBytes() <= dr.data.length);
            }
        }


    /**Width of image big enough to make both thumbnails of; several times longest large thumbnail dimension.
     * Package-visible to be available to other tests.
     */
    static final int bigImageWidth = ExhibitThumbnails.STD_STATIC_IMAGE_TN_LDIM_PX * 2;

    /**Height of image big enough to make both thumbnails of; several times longest large thumbnail dimension.
     * Package-visible to be available to other tests.
     */
    static final int bigImageHeight = ExhibitThumbnails.STD_STATIC_IMAGE_TN_LDIM_PX * 3;


    /**Create a test RGB fully-opaque true-colour image in memory.
     * This image will be suitable to make thumbnails from, for example,
     * if large enough.
     * <p>
     * This is an RGB image with a mixture of slowly-varying
     * colours and noise.
     * <p>
     * Package-visible to be available to other tests.
     */
    static BufferedImage makeTestRGBTrueColourBufferedImage(final int width,
                                                            final int height)
        {
        if((width < 1) || (height < 1))
            { throw new IllegalArgumentException(); }

        // Create a blank RGB (buffered) image.
        final BufferedImage bi = new BufferedImage(width,
                                                   height,
                                                   BufferedImage.TYPE_INT_ARGB);
        // Set this image up to have a relatively smoothly-changing backdrop
        // with some noise imposed on it.
        // Efficiency is not a huge issue here...
        for(int y = height; --y >= 0; )
            {
            final int r = Math.round((y / (float) height) * 128)
                + rnd.nextInt(16);
            for(int x = width; --x >= 0; )
                {
                final int g = Math.round((x / (float) width) * 128)
                    + rnd.nextInt(16);
                final int b = ((x+y) & 0xff);
                final int rgb = 0xff000000 | (r << 16) | (g << 8) | b;
                bi.setRGB(x, y, rgb);
                }
            }

        return(bi);
        }

    /**Create a byte-indexed fully-opaque RGB test image in memory.
     * This image will be suitable to make thumbnails from, for example,
     * if large enough.
     * <p>
     * This is an RGB image some geometric patterns
     * in a small number of colours (but more than one colour),#
     * but that should be colour-reducable.
     * <p>
     * In passing this also tests a technique for generating indexed-colour
     * images given RGB values.
     */
    private static BufferedImage makeTestRGBIndexedColourBufferedImage(final int width, final int height)
        {
        if((width < 1) || (height < 1))
            { throw new IllegalArgumentException(); }

        // Create a small palette of colours to use in a byte-indexed model.
        // These are ARGB values with no transparency.
        final int palette[] =
            {
            0xfffffff0, 0xff0f0f00, 0xffff000f, 0xff0f0f0f,
            0xffffffff, 0xffffff00, 0xffff00ff, rnd.nextInt() | 0xff000000
            };

        final IndexColorModel icm = new IndexColorModel(8, // 8-bit index.
                                                        palette.length,
                                                        palette,
                                                        0, // Offset into palette[].
                                                        false, // No alpha.
                                                        -1, // No transparent colour.
                                                        DataBuffer.TYPE_BYTE);

        // Create a blank RGB (buffered) image.
        final BufferedImage bi = new BufferedImage(width,
                                                   height,
                                                   BufferedImage.TYPE_BYTE_INDEXED,
                                                   icm);

        // Set this image up to have a regular pattern with some noise.
        // We ensure that even a 2-pixel image will have two colours.
        for(int y = height; --y >= 0; )
            {
            // Have vertical stripes with an non-uniform selection of
            // colours from the palette (biased towards low-numbered entries).
            final int yindex = rnd.nextInt(rnd.nextInt(palette.length) + 1);
            for(int x = width; --x >= 0; )
                {
                // Have an alternating pixel pattern horizontally.
                final int index = (yindex + ((x^y)&1)) % palette.length;
                bi.setRGB(x, y, palette[index]);
                }
            }


        // Check that result is a multi-colour image if more than 1 pixel.
        if(width * height > 1)
            {
            // Checks that the generated image has more than one colour
            // (ignoring alpha).
            final int firstPixelRGB = bi.getRGB(0, 0) & 0xffffff;
            boolean oneColour = true; // Only one colour.
            for(int y = height; --y >= 0; )
                {
                for(int x = width; --x >= 0; )
                    {
                    final int thisRGB = bi.getRGB(x, y) & 0xffffff;
                    if(thisRGB != firstPixelRGB)
                        {
                        oneColour = false;
                        break;
                        }
                    }
                }
            assertTrue("Generated test image should have more than one colour if more than one pixel",
                       !oneColour);
            }

        return(bi);
        }


    /**Make test JPEG encoded image of reasonably typical size; never null.
     * The image is bigger than any thumbnail,
     * and encoded at a typical quality.
     * <p>
     * Will throw a jUnit exception if something unexpected happens.
     * <p>
     * Package-visible to be available to other tests.
     *
     * @throws IOException if the JPEG encoder routines have difficulty
     */
    static byte[] makeTestRGBTrueColourJPEGImage()
        throws IOException
        {
        // Choose JPEG type...
        final int t = ExhibitMIME.ET_JPEG;

        // Check that he have a handler.
        final ExhibitMIME.ExhibitTypeParameters type = ExhibitMIME.getParamsByType(t);
        final Handler h = type.handler;
        assertTrue("Must always have a valid handler for t = " + t, h != null);
        // Check that handler is prepared to try to make thumbnails.
        assertTrue("Handler should be prepared to make thumbnails", h.canMakeThumbnails());

        // Test that we can create a big image.
        final BufferedImage bigImage = makeTestRGBTrueColourBufferedImage(
            bigImageWidth, bigImageHeight);
//        final int bigImageLongestEdge = Math.max(bigImageWidth, bigImageHeight);
        final byte[] bigEncodedImage =
            h.makeImageBinary(bigImage, 75);
            //makeJPEGImage(bigImageWidth, bigImageHeight);
        assertTrue("Must be able to create image from BufferedImage",
                   bigEncodedImage != null);
        assertTrue("Must be able to create valid image from BufferedImage",
                   ExhibitMIME.magicOK(type, new ByteArrayInputStream(bigEncodedImage)));
        assertTrue("Encoded image should be much bigger than 1kByte",
                   bigEncodedImage.length > 1024);

        return(bigEncodedImage);
        }




    /**Test colour reduction of RGB and indexed images using ImageUtils routine.
     * This tests that we don't get back more colours than we start with,
     * and that this works regardless of source image type.
     */
    public void testColourReduction()
        {
        // Try alternately with indexed and non-indexed images
        // at realistic thumbnail size.
        for(int srcFormat = 2; --srcFormat >= 0; )
            {
            // Create a typical thumbnail-sized image.
            final BufferedImage srcImage = (srcFormat == 0) ?
                makeTestRGBTrueColourBufferedImage(64, 48) :
                makeTestRGBIndexedColourBufferedImage(48, 64);

            // Try alternately forcing the output to indexed format and not.
            for(int dstFormat = 2; --dstFormat >= 0; )
                {
                final boolean forceByteIndexModel = (dstFormat == 0);

                // Try making successively smaller colour spaces.
                // This may not be able to do successful colour reduction
                // to very small numbers of colours.
                for(int colorBits = 17; --colorBits > 0; )
                    {
                    final int maxColours = (1 << colorBits);

                    // Extract copy of pixels from source image in TYPE_INT_ARGB format.
                    // Set scansize to be width to efficiently pack the result.
                    final int width = srcImage.getWidth();
                    final int height = srcImage.getHeight();
                    final int pixels[] =
                        srcImage.getRGB(0, 0, width, height, null, 0, width);

                    // Count the colours in the original image (ignoring alpha).
                    final Set<Integer> originalDistinctColours = new TreeSet<Integer>();
                    for(int i = pixels.length; --i >= 0; )
                        { originalDistinctColours.add(new Integer(pixels[i] & 0xffffff)); }
//System.out.println("testColourReduction(): original colour count: " + originalDistinctColours.size());

                    // Create a colour-reduced image.
                    final BufferedImage reducedColImage =
                        ImageUtils.makeColourReducedBufferedImage(srcImage,
                                                                  maxColours,
                                                                  forceByteIndexModel);
                    // Count the colours in the new image (ignoring alpha).
                    final int newColours =
                        ImageUtils.countDistinct24BitRGBColours(reducedColImage);
//System.out.println("testColourReduction(): new colour count (maxColours:"+maxColours+"): " + newColours);

                    assertTrue("Must produce an output image", reducedColImage != null);
                    assertTrue("Must produce an output image the same dimensions as the input",
                               (reducedColImage.getWidth() == width) &&
                               (reducedColImage.getHeight() == height));
                    assertTrue("Must not increase number of colours",
                               newColours <= originalDistinctColours.size());
                    assertTrue("Must leave no more than requested colours in result",
                               newColours <= maxColours);
//System.out.println("output colour model: " + reducedColImage.getColorModel());
                    assertTrue("Must have at least two colours in palette if original did: new colour map is: " + reducedColImage.getColorModel(),
                               (originalDistinctColours.size() < 2) || (newColours >= 2));
                    }
                }
            }
        }



    /**Test that we can do colour reduction with the ImageMagik-derived Quantizer.
     */
    public void testQuantizer()
        {
        // Create a typical thumbnail-sized image.
        final BufferedImage bigImage = makeTestRGBTrueColourBufferedImage(
            256, 192);

        // Try making successively smaller colour spaces.
        for(int colorBits = 9; --colorBits > 0; )
            {
            final int maxColours = (1 << colorBits);

            // Extract copy of pixels from source image in TYPE_INT_ARGB format.
            // Set scansize to be width to efficiently pack the result.
            final int pixels[] =
                bigImage.getRGB(0, 0, bigImage.getWidth(), bigImage.getHeight(),
                            null, 0, bigImage.getWidth());

            // Count the colours in the original image (ignoring alpha).
            final Set<Integer> originalDistinctColours = new TreeSet<Integer>();
            for(int i = pixels.length; --i >= 0; )
                { originalDistinctColours.add(new Integer(pixels[i] & 0xffffff)); }
//System.out.println("testQuantizer(): original colour count: " + originalDistinctColours.size());

            // Now do the colour reduction...
            final int[] palette = Quantize.quantizeImage(
                new int[][]{pixels}, maxColours);
//System.out.println("testQuantizer(): new colour count: " + palette.length);
            assertTrue("Must leave at least two colours in palette if orgininal had at least two",
                       (originalDistinctColours.size() < 2) || (palette.length >= 2));

            // Create a new ARGB image and copy the new pixels in there.
            final BufferedImage reducedColImage = new BufferedImage(
                bigImage.getWidth(), bigImage.getHeight(),
                BufferedImage.TYPE_INT_ARGB);
            reducedColImage.setRGB(0, 0, bigImage.getWidth(), bigImage.getHeight(),
                            pixels, 0, bigImage.getWidth());

            assertTrue("Must produce an output image", reducedColImage != null);
            assertTrue("Must produce an output image the same dimensions as the input",
                       (reducedColImage.getWidth() == bigImage.getWidth()) &&
                       (reducedColImage.getHeight() == bigImage.getHeight()));
            assertTrue("Must not increase number of colours",
                       palette.length <= originalDistinctColours.size());
            }
        }

    /**Simple tests of exhibit metadata.
     * For example, check that it is safe to attempt to extract metadata from all exhibits with a handler,
     * even though there legitimately be no metadata to extract
     * though we in fact expect to always extract some metadata from every object where we have a handler.
     * <p/>
     * We check that we can extract EXIF data from at least one (JPEG) exhibit.
     */
    public void testMetadataBasics()
        throws Exception
        {
        // Set up a simple instance of a file data source using our sample data.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        // Collect the raw exhibit info from the underlying file system.
        final AllExhibitProperties aepRaw = edfs.getAllExhibitProperties(-1);

        // Note if we found EXIF data.
        boolean foundEXIF = false;
        // We look for the existence of the following path.
        final String EXIFPath[] = { "metadata", "native", "EXIF", "tag" };

        // For every exhibit that has a handler,
        // try to extract available metadata,
        // ensuring that no failure (eg exception) is caused,
        // and that metadata is definitely available where it should be.
        for(final ExhibitStaticAttr esa : aepRaw.aeid.getAllStaticAttrs())
            {
            ExhibitMIME.ExhibitTypeParameters et = (ExhibitMIME.getInputFileType(esa.filePath));
            assertNotNull("All exhibits should have a type: " + esa, et);
            final Handler h = et.handler;
            if(h == null) { continue; }

            // Read the data into memory in one go for simplicity.
            // None of our sample exhibits is vast.
            final byte data[] = new byte[(int) esa.length];
            final File rawDataLoc = new File(LocalProps.getDataDir(), esa.filePath);
            final DataInputStream dis = new DataInputStream(new FileInputStream(rawDataLoc));
            try { dis.readFully(data); }
            finally { dis.close(); }

            // Fetch the metadata directly; we expect there to be some for every type that has a handler...
            final Node metadata = h.getMetadata(new ByteArrayInputStream(data), esa.filePath);
            assertNotNull("Standard meta-data should be extractable where we have a handler: " + et.suffixForInputFile, metadata);
//            System.out.println(TextUtils.toXML(metadata, false, false));

            if(metadata != null)
                {
                // Check that top-level metadata node name is correct.
                assertSame("Top-level metadata node incorrect", "metadata", metadata.getNodeName());

                // Fetch the metadata from the EPC; should be equivalent to getting it directly.
                final Node metadataEPC = aepRaw.getExhibitPropsComputable(esa.filePath).getMetadata();
                assertNotNull("Metadata should be available via the EPC if available directly", metadataEPC);
                // FIXME: test that we get an equivalent tree...

                // Check if this exhibit has EXIF metadata...
                Node n = metadata;
                for(int i = 0; ++i < EXIFPath.length; )
                    {
                    final String name = EXIFPath[i];
                    boolean foundCorrectChild = false;
                    for(Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
                        {
                        if(name.equals(child.getNodeName()))
                            {
                            n = child;
                            foundCorrectChild = true;
                            if(i + 1 == EXIFPath.length)
                                {
                                System.out.println("Found EXIF metadata for: " + esa.filePath);
                                foundEXIF = true;
                                }
                            break;
                            }
                        }
                    if(!foundCorrectChild) { break; }
                    }
                }
            }

        assertTrue("At least one (JPEG) exhibit must have EXIF metadata", foundEXIF);
        }


    /**Reasonable shared random-number generator. */
    private static final Random rnd = new Random();
    }
