/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.svrCore.ROByteArray;

import java.awt.*;
import java.util.Random;

/**Tests of the memory tools.
 */
public final class MemoryToolsTest extends TestCase
    {
    public MemoryToolsTest(final String name)
        {
        super(name);
        }

    /**Simple test of runMemoryIntensiveOperation().
     * Tests that:
     * <ul>
     * <li>Tasks are always run if we claim to have unlimited memory.
     * <li>Tasks are always run if we require very limited memory.
     * <li>Tasks are never run if we demand enormous memory and memory is
     *     not unlimited.
     * </ul>
     * <p>
     * TODO: add test that we can run these concurrently without difficulty.
     */
    public void test_runMemoryIntensiveOperation()
        {
        // Our "small" memory requirement.
        final int smallMem = 1;

        // Our "huge" memory requirement.
        final int hugeMem = Integer.MAX_VALUE;

        // Small task that sets a flag when invoked.
        final class SmallTask implements Runnable
            {
            public boolean done; // Initially false.
            public void run() { done = true; }
            }

        // Run the tasks a few times to make sure that behaviour
        // remains correct even during intensive use.
        for(int i = 1000; --i >= 0; )
            {
            final SmallTask t1 = new SmallTask();
            assertTrue("With unlimited memory, runMemoryIntensiveOperation() should return true",
                       MemoryTools.runMemoryIntensiveOperation(t1, true, hugeMem));
            assertTrue("With unlimited memory, runMemoryIntensiveOperation() should run the task",
                       t1.done);

            final SmallTask t2 = new SmallTask();
            assertTrue("With limited memory and a huge task, runMemoryIntensiveOperation() should return false",
                       !MemoryTools.runMemoryIntensiveOperation(t2, false, hugeMem));
            assertTrue("With limited memory and a huge task, runMemoryIntensiveOperation() should not run the task",
                       !t2.done);

            final SmallTask t3 = new SmallTask();
            assertTrue("With limited memory and a small task, runMemoryIntensiveOperation() should return true",
                       MemoryTools.runMemoryIntensiveOperation(t3, false, smallMem));
            assertTrue("With limited memory and a small task, runMemoryIntensiveOperation() should run the task",
                       t3.done);
            }
        }

    /**Test our intern() for various supported data types.
     * This also tests refusal to intern various mutable data types.
     */
    public void testIntern()
        {
        // Check that MemoryTools.intern() works correctly for String.
        final String s1 = "abc";
        final String s2 = new String("abc");
        assertEquals("s1.equals(s2) should be true", s1, s2);
        assertNotSame("Should be same value but different instances.", s1, s2);
        assertSame("intern()ing String should give same instance.", MemoryTools.intern(s1), MemoryTools.intern(s2));

        // Check that MemoryTools.intern() works correctly for ROByteArray.
        final ROByteArray roba1 = ROByteArray.compressFromString("the quick brown fox");
        final ROByteArray roba2 = ROByteArray.compressFromString("the quick brown fox");
        assertEquals("roba1.equals(roba2) should be true", roba1, roba2);
        assertNotSame("Should be same value but different instances.", roba1, roba2);
        assertSame("intern()ing ROByteArray should give same instance.", MemoryTools.intern(roba1), MemoryTools.intern(roba2));




        // TODO: Test for Boolean and other supported types/ranges.




        // Check that we don't attempt to intern() obvious things that we should not!
        assertNotSame("Should not intern mutable objects (eg arrays)",
                      MemoryTools.intern(new byte[100]), MemoryTools.intern(new byte[100]));
        assertNotSame("Should not intern mutable objects (eg Dimension)",
                      MemoryTools.intern(new Dimension()), MemoryTools.intern(new Dimension()));


        // Simple stress-test of intern() that it does not obviously fail to free its unused values.
        // We spend a few seconds churning as much as we can
        // with items that we might actually intern() in real life.
        final long stopBy = System.currentTimeMillis() + 9999;
        final byte buf[] = new byte[16];
        int count = 0;
        while(System.currentTimeMillis() <= stopBy)
            {
            MemoryTools.intern(String.valueOf(rnd.nextLong()));
            MemoryTools.intern(new Long(rnd.nextLong()));
            rnd.nextBytes(buf);
            assertSame("intern(ROByteArray) failed",
                       MemoryTools.intern(new ROByteArray(buf)),
                       MemoryTools.intern(new ROByteArray(buf)));
            ++count;
            }
        System.out.println("Survived allocation and intern() of "+(4*count)+" new items.");
        }



    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
