/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.GenUtils;
import org.hd.d.pg2k.svrCore.TextUtils;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Random;

/**Miscellaneous tests not filed elsewhere.
 * This should be kept as small as possible with tests being moved to better homes ASAP.
 */
public final class MiscTest extends TestCase
    {
    public MiscTest(String name)
        {
        super(name);
        }

//    /**Do any setup needed for the tests. */
//    protected void setUp()
//        {
//        }

//    /**Do any clearup needed after the tests. */
//    protected void tearDown()
//        {
//        // cleanup code
//        }

    /**Test our quick log-2 computation. */
    public static final void testLog2Approx()
        {
        // Test some important fixed/boundary cases.
        assertEquals(-1, GenUtils.log2Approx(0));
        assertEquals(0, GenUtils.log2Approx(1));
        assertEquals(1, GenUtils.log2Approx(2));
        assertEquals(1, GenUtils.log2Approx(3));
        assertEquals(2, GenUtils.log2Approx(4));
        assertEquals(2, GenUtils.log2Approx(5));
        assertEquals(2, GenUtils.log2Approx(6));
        assertEquals(2, GenUtils.log2Approx(7));
        assertEquals(3, GenUtils.log2Approx(8));
        assertEquals(30, GenUtils.log2Approx(Integer.MAX_VALUE));
        assertEquals(31, GenUtils.log2Approx(1L + Integer.MAX_VALUE));
        assertEquals(31, GenUtils.log2Approx(2L + Integer.MAX_VALUE));
        assertEquals(62, GenUtils.log2Approx(Long.MAX_VALUE));

        // Try on a variety of random non-negative values.
        for(int i = 100; --i >= 0; )
            {
            final long v = (rnd.nextLong() >>> 1);
            final long l = GenUtils.log2Approx(v);
            assertTrue("GenUtils.log2Approx(v) answer out of range", (l >= -1) && (l < 64));
            }
        }
    
    /**Test if the data pump.
     * Makes sure that at the very least it does not corrupt data
     * passing through it.
     * <p>
     * We set up and test a number of pump instances
     * to try to ensure that we not
     * eating up (and failing to release) vast system resources
     * even if the caller fails to explicitly close() all the stream.
     */
    public static void testDataPump()
        throws Exception
        {
        final long startTime = System.currentTimeMillis();
        final long stopBy = startTime + 1000; // 1s max on this test.
        for(int i = 1000; --i >= 0; )
            {
            // Random data size up to several megabytes...
            final byte[] data = new byte[rnd.nextInt(8123123)];
            rnd.nextBytes(data);
//System.out.println("[dataPump() test on length "+data.length);
            // Wrap the input data, hiding the available() value (sometimes)
            // to prevent over-optimisation by dataPump() undermining this test,
            // ie we "dumb down" the InputStream here a little.
            final ByteArrayInputStream bais = new ByteArrayInputStream(data.clone()){
                @Override public int available()
                    { return(rnd.nextBoolean() ? 0 : super.available()); }
                @Override public boolean markSupported()
                    { return(false); }                
                };
            // Try data pumps with varying buffering settings.
            final InputStream is = GenUtils.dataPump(GenUtils.dataPump(bais, 1 + rnd.nextInt(1123123), rnd.nextBoolean()), 1 + rnd.nextInt(1123123), rnd.nextBoolean());
            final DataInputStream dis = new DataInputStream(is);

            final byte buf[] = new byte[data.length];
            dis.readFully(buf); // All data must be available through the pump.
            assertTrue("Data must not be corrupted on pumped stream", Arrays.equals(data, buf));
            assertEquals("Must be no spurious trailing data on pumped stream", -1, dis.read());

            // Only sometimes remember to close() the stream.
            // This simulates some programming errors of omission
            // that we'd like to protect against.
            if(rnd.nextBoolean()) { dis.close(); }

            // Stop if we've taken a long time on this test.
            if(System.currentTimeMillis() > stopBy) { break; }
            }
        }

    /**Test encoding to Base64 (as used in HTTP headers) and decoding thereof.
     */
    public static final void test8To6()
        {
        // Standard simple test cases.
        // Trivial empty case.
        assertEquals("Must be able to encode empty value correctly", "", TextUtils.encode8To6(new byte[0]));
        assertTrue("Must be able to decode empty value correctly", TextUtils.decode8To6("").length == 0);
        // Aligned (multiple of 3 bytes, thus no partial last block).
        final byte[] test0 = { 'h', 'e', 'l' };
        final String encoded0 = "aGVs";
        assertEquals("Must be able to encode simple test0 correctly", encoded0, TextUtils.encode8To6(test0));
        assertTrue("Must be able to decode simple test0 correctly", Arrays.equals(test0, TextUtils.decode8To6(encoded0)));
        // Not aligned, so partial last block.
        final byte[] test1 = { 'h', 'e', 'l', 'l', 'o' };
        final String encoded1 = "aGVsbG8=";
        assertEquals("Must be able to encode simple test1 correctly", encoded1, TextUtils.encode8To6(test1));
        assertTrue("Must be able to decode simple test1 correctly", Arrays.equals(test1, TextUtils.decode8To6(encoded1)));

        // Now try encode/decode on some lumps of random data...
        for(int i = 100; --i >= 0; )
            {
            final byte[] data = new byte[rnd.nextInt(1000)];
            rnd.nextBytes(data);
            assertTrue("Must be able to encode/decode arbitrary binary data",
                Arrays.equals(data, TextUtils.decode8To6(TextUtils.encode8To6(data.clone()))));
            }
        }
    
    

    /**Dummy test to keep jUnit framework happy when there are not other tests in this class. */
    public static final void testDummy()
        { }

    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
