/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.ExhibitStaticAttr;
import org.hd.d.pg2k.webSvr.catalogue.PaginationBeanBase;
import org.hd.d.pg2k.webSvr.catalogue.PaginationBeanNumeric;
import org.hd.d.pg2k.webSvr.catalogue.PaginationBeanTree;
import org.hd.d.pg2k.webSvr.exhibit.TreeFilterBean;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 03-Jun-2003
 * Time: 17:55:04
 */

/**Tests the variants and derivatives of PaginationBeanNumeric.
 * These beans can be used stand-alone (outside a JSP page) at least in part,
 * so we can test the core algorithms.
 */
public final class PaginationBeanTest extends TestCase
    {
    /**Performs test of the base bean.
     * This ensures that the basic underlying algorithms are correct.
     */
    public void testPaginationBeanBase()
        {
        // First just create an instance of the bean.
        final PaginationBeanNumeric pbn = new PaginationBeanNumeric();

        // Test initial values.
        assertTrue("default number of items should be zero", pbn.getNumberOfItems() == 0);
        assertTrue("default page number should be 1", pbn.getPg() == 1);
        // Do the generic tests with current setup.
        testPBContract(pbn);

        // Check that any attmept to set the page with zero items to paginate is quetly ignored.
        pbn.setPg(-1);
        assertTrue("page number should be 1", pbn.getPg() == 1);
        pbn.setPg(9);
        assertTrue("page number should be 1", pbn.getPg() == 1);
        pbn.setPg("rubbish");
        assertTrue("page number should be 1", pbn.getPg() == 1);

        // Check number-of-pages algorithm.
        pbn.setNumberOfItems(0);
        assertTrue("0 items needs zero pages", pbn.getNumberOfPages() == 0);
        pbn.setNumberOfItems(1);
        assertTrue("1 item needs one page", pbn.getNumberOfPages() == 1);
        pbn.setNumberOfItems(WebConsts.MAX_RESULTS_PER_PAGE - 1);
        assertTrue("1 item less than the max per page needs one page", pbn.getNumberOfPages() == 1);
        pbn.setNumberOfItems(WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("the max per page needs one page", pbn.getNumberOfPages() == 1);
        pbn.setNumberOfItems(WebConsts.MAX_RESULTS_PER_PAGE + 1);
        assertTrue("1 item more than the max per page needs two pages", pbn.getNumberOfPages() == 2);
        pbn.setNumberOfItems(99 * WebConsts.MAX_RESULTS_PER_PAGE + 1);
        assertTrue("wrong result for many items", pbn.getNumberOfPages() == 100);
        // Do the generic tests with current setup.
        testPBContract(pbn);

        // Test the (plain-old-int) page-setting algorithm.
        pbn.setNumberOfItems(10 * WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("wrong result for ten-pages' worth of items", pbn.getNumberOfPages() == 10);
        pbn.setPg(3);
        assertTrue("did not set page number correctly to 3", pbn.getPg() == 3);
        pbn.setPg(-999);
        assertTrue("did not force page number correctly to 1 (always safe)", pbn.getPg() == 1);
        pbn.setPg(7);
        assertTrue("did not set page number correctly to 7", pbn.getPg() == 7);
        pbn.setPg(999);
        assertTrue("did not force page number correctly to 1 (always safe)", pbn.getPg() == 1);
        pbn.setPg(10);
        assertTrue("did not set page number correctly to 10 (upper limit)", pbn.getPg() == 10);
        pbn.setPg(1);
        assertTrue("did not set page number correctly to 1 (lower limit)", pbn.getPg() == 1);
        // Test the String argument version.
        pbn.setPg("4");
        assertTrue("did not set page number correctly to 4", pbn.getPg() == 4);
        pbn.setPg("broken");
        assertTrue("did not force page number correctly to 1 (always safe)", pbn.getPg() == 1);
        // Do the generic tests with current setup.
        testPBContract(pbn);

        // Test item upper/lower bound algorithm
        pbn.setNumberOfItems(2 + 3 * WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("wrong result for four pages' worth of items", pbn.getNumberOfPages() == 4);
        pbn.setPg(1);
        assertTrue("lower bound for first page wrong", pbn.getStartIndex() == 0);
        assertTrue("upper bound for first page wrong", pbn.getEndIndex() == WebConsts.MAX_RESULTS_PER_PAGE);
        pbn.setPg(2);
        assertTrue("lower bound for second page wrong", pbn.getStartIndex() == WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("upper bound for second page wrong", pbn.getEndIndex() == 2 * WebConsts.MAX_RESULTS_PER_PAGE);
        pbn.setPg(4);
        assertTrue("lower bound for last (partial) page wrong", pbn.getStartIndex() == 3 * WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("upper bound for last (partial) page wrong", pbn.getEndIndex() == pbn.getNumberOfItems());
        // Do the generic tests with current setup.
        testPBContract(pbn);

        // Test label generation for smallish number of pages
        // (less than the maximum we can do buttons for in one go).
        pbn.setNumberOfItems(1 + 2 * WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("Check label generation for page 2", "2".equals(pbn.getPageLabel(2)));
        assertTrue("Check label generation for page 1", "1".equals(pbn.getPageLabel(1)));
        assertTrue("Check label generation for page 3", "3".equals(pbn.getPageLabel(3)));
        final List<String> l1 = pbn.getPageLabels();
        assertTrue("Check length of label list (3)", l1.size() == 3);
        assertTrue("Check labels are right for the pages on small list",
                   "1".equals(l1.get(0)) &&
                   "2".equals(l1.get(1)) &&
                   "3".equals(l1.get(2)));
        // Do the generic tests with current setup.
        testPBContract(pbn);

        // Test label generation for large number of items
        // (too many to generate page buttons for all at once).
        pbn.setNumberOfItems(1 + 2 * WebConsts.MAX_RESULTS_PAGES * WebConsts.MAX_RESULTS_PER_PAGE);
        assertTrue("Check label generation for page 2", "2".equals(pbn.getPageLabel(2)));
        assertTrue("Check label generation for page 1", "1".equals(pbn.getPageLabel(1)));
        assertTrue("Check label generation for page 3", "3".equals(pbn.getPageLabel(3)));
        assertTrue("Check label generation for page 17: got "+pbn.getPageLabel(17), "17".equals(pbn.getPageLabel(17)));
        final List<String> l2 = pbn.getPageLabels();
        assertTrue("Check length of label list for large set", l2.size() <= WebConsts.MAX_RESULTS_PAGES);
        // Check first and last pages have labels,
        // and rest are monotonically increasing.
        assertTrue("Check first and last labels are right for the pages on large list",
                   "1".equals(l1.get(0)) &&
                   String.valueOf(pbn.getNumberOfPages()).equals(l2.get(l2.size() - 1)));
        int lastP1 = 0;
        for(final Iterator<String> it = l2.iterator(); it.hasNext(); )
            {
            final int n = Integer.parseInt(it.next());
            assertTrue("Checking for monotonically-increasing list", n > lastP1);
            lastP1 = n;
            }
        // Do the generic tests with current setup.
        testPBContract(pbn);
        }

    /**Test the PaginationBeanTree.
     *
     */
    public void testTreeFilterBean()
        {
        // First just create an instance of the bean.
        final PaginationBeanTree pbt = new PaginationBeanTree();

        // Test initial values.
        assertTrue("default number of items should be zero", pbt.getNumberOfItems() == 0);
        assertTrue("default page number should be 1", pbt.getPg() == 1);
        // Do the generic tests with current setup.
        testPBContract(pbt);

        // Make sure that we can't call setNumberOfItems() directly.
        try {
            pbt.setNumberOfItems(43);
            assertTrue("Should not have been able to call setNumberOfItems()", false);
            }
        catch(Error e) { } // OK: wanted this call vetoed.

        // Make sure that things work with an empty List.
        final List<String> emptyList = Collections.emptyList();
        pbt.setTreeNodeDetails("", emptyList);
        assertTrue("An empty list should mean no items", pbt.getNumberOfItems() == 0);


        // NOW TEST ON SOME REAL DATA.
        // Short names of the exhibits.
        final String SHNAME1 = "cat-sat-on-mat-ANON.htxt";
        final String SHNAME2 = "CAT-fat-DHD.jpg"; // Note different case of main word.
        final String SHNAME3 = "dog-in-manger-1-ANON.gif";
        final String SHNAME4 = "dog-in-manger-2-ANON.gif";

        // Full names of exhibits.
        final String EXNAME1 = "memes/" + SHNAME1;
        final String EXNAME2 = "memes/" + SHNAME2;
        final String EXNAME3 = "other/" + SHNAME3;
        final String EXNAME4 = "another/" + SHNAME4;

        // For now, ignore the effects of attribute words...
        final long now = System.currentTimeMillis();
        final Map<String,ExhibitStaticAttr> exhibits1 = new HashMap<String, ExhibitStaticAttr>();
        exhibits1.put(EXNAME1, new ExhibitStaticAttr(EXNAME1, 100, now));
        exhibits1.put(EXNAME2, new ExhibitStaticAttr(EXNAME2, 100, now));
        exhibits1.put(EXNAME3, new ExhibitStaticAttr(EXNAME3, 100, now));
        exhibits1.put(EXNAME4, new ExhibitStaticAttr(EXNAME4, 100, now));
        // Add lots of very similar exhibits for same node that will need lots of pages. */
        for(int i = 3 + 2 * WebConsts.MAX_RESULTS_PAGES * WebConsts.MAX_RESULTS_PER_PAGE; --i >= 0; )
            {
            final String name = "yetanother/dog-in-mangerette-" + i + "-ANON.jpg";
            exhibits1.put(name, new ExhibitStaticAttr(name, 101, now));
            }
//        final AllExhibitProperties aep1 = new AllExhibitProperties(
//                                new ExhibitPropsGlobalImmutable(),
//                                new AllExhibitImmutableData(exhibits1, now),
//                                Collections.EMPTY_MAP /* Map _loadedProps */);

        // Build this Map non-optimised for simplicity.
        final boolean optimised = false;
        // Build this map with List nodes.
        final boolean asSet = false;
        final Map<String, Collection<String>> m1 = TreeFilterBean.computePrefixMap(new ArrayList<String>(exhibits1.keySet()), asSet, optimised);

        // Extract the root node (which is small enough to get on one page).
        final String rootPrefix = "";
        final Collection<String> prefixesAndShortExhibitName = m1.get(rootPrefix);
        assertTrue(prefixesAndShortExhibitName instanceof List);
        pbt.setTreeNodeDetails(rootPrefix, (List<String>) prefixesAndShortExhibitName);
        assertTrue("Expect small root node (2 items)", pbt.getNumberOfItems() == 2);
        assertTrue("Expect small root node (1 page)", pbt.getNumberOfPages() == 1);
        // Do the generic tests with current setup.
        testPBContract(pbt);

        // Extract the "dog-in-mangerette-" node, which should be huge.
        final String dimPrefix = "dog-in-mangerette-";
        final Collection<String> prefixesAndShortExhibitName2 = m1.get(dimPrefix);
        assertTrue(prefixesAndShortExhibitName2 instanceof List);
        pbt.setTreeNodeDetails(dimPrefix, (List<String>) prefixesAndShortExhibitName2);
        assertTrue("Expect huge root node",
                   pbt.getNumberOfItems() > WebConsts.MAX_RESULTS_PAGES * WebConsts.MAX_RESULTS_PER_PAGE);
        // Do the generic tests with current setup.
        testPBContract(pbt);
        }

    /**If true, we are very strict and insist that we must use all available buttons.
     * This is to give users maximum number of useful choices.
     */
    private static final boolean MUST_USE_ALL_AVAILABLE_BUTTONS = true;

    /**Do some tests of generic contract that should work on any PaginationBaseBean or sub-class.
     * The bean must be passed in already set up.
     * <p>
     * This may alter some or all of the bean state.
     */
    private static void testPBContract(PaginationBeanBase pbb)
        {
        // Test labels of bean in whatever state it arrived...
        testLabels(pbb);

        // For every possible page value, test label generation and parsing...
        for(int pg = pbb.getNumberOfPages() + 1; --pg > 0; )
            {
            // Test that we can always correctly parse every page label and set the page.
            for(int j =  pbb.getNumberOfPages() + 1; --j > 0; )
                {
                assertTrue("must always be able to generate and parse every page label", j == pbb.getPageNumber(pbb.getPageLabel(j)));
                pbb.setPg(j);
                assertTrue("must always be able to set any page from a label", j == pbb.getPg());
                }

            pbb.setPg(pg);
            assertTrue("Must be able to set current page number", pg == pbb.getPg());

//System.out.println("labels..."); for(final Iterator it = pbb.getPageLabels().iterator(); it.hasNext(); ) { System.out.println(it.next()); }

            // Test that all the page-button labels represent pages in monotonically-increasing order,
            // and produce the same labels as getPageLabel() for each individual page.
            int last = 0;
            final int pageNumbers[] = pbb.getPageNumbers();
            final List<String> theseLabels = pbb.getPageLabels();
            final int lastPageIndex = Math.max(1, pbb.getNumberOfPages());
            assertTrue("labels and numbers must be same length", pageNumbers.length == theseLabels.size());
//System.out.print("Page sequence (current page = "+pg+"):");
            for(int i = 0; i < pageNumbers.length; ++i)
                {
                // Test labels of bean in every possible bean state...
                testLabels(pbb);

                final int pgNum = pageNumbers[i];
//System.out.print(" " + pgNum);
                assertTrue("all displayed-page numbers must be in range", (pgNum >= 1) && (pgNum <= lastPageIndex));
                assertTrue("labels must represent monotonically-ascending pages", pgNum > last);
                final String label = pbb.getPageLabel(pgNum);
                assertTrue("getPageNumber() must be the inverse of getPageLabel()", pgNum == pbb.getPageNumber(label));
                final String s = theseLabels.get(i);
                assertTrue("label set must match individual labels (pg "+pgNum+"): "+s+"/"+label, s.equals(label));
                last = pgNum;
                }
//System.out.println(" ...end ("+theseLabels.size()+" buttons).");
            }
        }

    /**Test some characteristics of the labels obtained from the bean.
     */
    private static void testLabels(PaginationBeanBase pbb)
        {
        final List<String> labels = pbb.getPageLabels();

        // Under all circumstances, number of pages and items must be non-negative.
        assertTrue("Number of items and pages must be non-negative",
                   (pbb.getNumberOfItems() >= 0) && (pbb.getNumberOfPages() >= 0));
        // The page number must be suitably constrained.
        assertTrue("Page number must be [1, numberOfPages] if more than zero items: " + pbb.getPg() + " [" + pbb.getNumberOfPages() + "]",
                   (pbb.getPg() >= 1) &&
                        ((pbb.getNumberOfItems() == 0) || (pbb.getPg() <= pbb.getNumberOfPages())));
        // Under all circumstances, the number of page buttons must be no more than WebConsts.MAX_RESULTS_PAGES.
        assertTrue("Must not produce too many pages", labels.size() <= WebConsts.MAX_RESULTS_PAGES);

        // We also ensure (though this is in principle less critical)
        // that at least a substantial fraction of the possible labels are used,
        // else the user may have to click too many buttons.
        assertTrue("A reasonable fraction of available buttons must be used",
                   labels.size() >= Math.min(WebConsts.MAX_RESULTS_PAGES, pbb.getNumberOfPages()) / 2);

        // If being strict we insist that all buttons are used.
        if(MUST_USE_ALL_AVAILABLE_BUTTONS)
            {
            assertTrue("All available buttons must be used so as to give the user the shortest path to their target page",
                       labels.size() == Math.min(WebConsts.MAX_RESULTS_PAGES, pbb.getNumberOfPages()));
            }

        // Test that all the page-button labels are unique.
        assertTrue("All labels must be unique", (new HashSet(labels)).size() == labels.size());
        assertTrue("labels and numbers must be same length", pbb.getPageNumbers().length == labels.size());
        }
    }
