/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.uploader.ExhibitHandlerBeanBase;
import org.hd.d.pg2k.webSvr.catalogue.SearchPageJavaBean;
import org.hd.d.pg2k.webSvr.util.LocaleBean;

import java.util.Iterator;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 03-Jun-2003
 * Time: 17:55:04
 */

/**Tests ExhibitHandlerBeanBase and SearchPageJavaBean.
 * This bean can be used stand-alone (outside a JSP page) at least in part,
 * so we can test the core algorithms.
 */
public final class SearchPageJavaBeanTest extends TestCase
    {
    /**Tests getters and setters of SearchPageJavaBeanTest bean.
     *
     */
    public static void testSPJBBasics()
        {
        final SearchPageJavaBean spjb = new SearchPageJavaBean();

        // Tests some initial state.
        assertTrue("query should initially be empty and non-null", "".equals(spjb.getQ()));
        assertTrue("recent-days value should default to zero", spjb.getRecentDaysFilter() == 0);
        }

    /**Test recentDaysFilter.
     * We had particular problems with this on the Web server,
     * (2003/08/10, though it worked fine in development)
     * so we give it a special grilling here.
     */
    public static void testRecentDaysFilter()
        {
        final SearchPageJavaBean spjb = new SearchPageJavaBean();

        // Tests initial state.
        assertTrue("recent-days value should default to zero", spjb.getRecentDaysFilter() == 0);

        // Setting to any half-reasonable non-negative integer value should work.
        // We test with some random values and some fixed values.
        final int fixedValues[] = { 0, 1, 3, 6,7,8, 27,28,29,30,31,32, 364,365,366,367 };
        for(int i = 99; --i >= 0; )
            {
            final int d1 = Rnd.fastRnd.nextInt(99999);
            spjb.setRecentDaysFilter(String.valueOf(d1));
            assertTrue("Should be able to set any non-negative recent-days value [1]: " + d1, d1 == spjb.getRecentDaysFilter());
            final int d2 = fixedValues[i % fixedValues.length];
            spjb.setRecentDaysFilter(String.valueOf(d2));
            assertTrue("Should be able to set any non-negative recent-days value [2]: " + d2, d2 == spjb.getRecentDaysFilter());
            }

        // Ensure that setting to any synonym for 'none'
        // ("0" or "-" or null or "")
        // resets to zero.
        final String noneSyn[] = { "0", ExhibitHandlerBeanBase.SETTER_ALL, "", null };
        for(int i = noneSyn.length; --i >= 0; )
            {
            spjb.setRecentDaysFilter("1");
            assertTrue("setting to non-zero value did not work", 0 != spjb.getRecentDaysFilter());
            spjb.setRecentDaysFilter(noneSyn[i]);
            assertTrue("setting with " + noneSyn[i] + " should reset to zero", 0 == spjb.getRecentDaysFilter());
            }

        // Check that setting to one of the symbolic values gives
        // the right number of days,
        // and that an appropriate HTML item is marked "selected".
        for(final Iterator it = SearchPageJavaBean.symFilterDays.keySet().iterator(); it.hasNext(); )
            {
            final String symbolicName = (String) it.next();
            final int d = (SearchPageJavaBean.symFilterDays.get(symbolicName)).intValue();
            spjb.setRecentDaysFilter(symbolicName);
            assertTrue("setting with " + symbolicName + " should set to " + d, d == spjb.getRecentDaysFilter());

            // We don't use locale in first instance, so we use the symbolic names.
            final String selectBodySimple = spjb.makeRecentDaysFilterBody(null);
            final String selectedOptionSimple = "<option selected>" + symbolicName + "</option>";
//System.out.println("RDF simple: " + symbolicName + " in " + selectBody);
            assertTrue("appropriate option should be selected: " + selectedOptionSimple + " FROM " + selectBodySimple,
                       selectBodySimple.indexOf(selectedOptionSimple) >= 0);

            // Now we use the default locale,
            // for which some basic non-symbolic text must be present,
            // else we won't get the form of option tag that we expect.
            final String selectBody = spjb.makeRecentDaysFilterBody(new LocaleBean());
            final String selectedOption = "<option value="+symbolicName+" selected>";
//System.out.println("RDF: " + symbolicName + " in " + selectBody);
            assertTrue("appropriate option should be selected: " + selectedOptionSimple + " FROM " + selectBody,
                       selectBody.indexOf(selectedOption) >= 0);
            }
        }
    }
