/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;

import org.apache.tools.bzip2.*;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.location.*;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.props.PropertiesBundleDiff;
import org.hd.d.pg2k.svrCore.props.PropertiesDiff;
import org.hd.d.pg2k.svrCore.vars.*;
import org.hd.d.pg2k.webSvr.util.FlushableGZIPOutputStream;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import java.awt.Dimension;
import java.io.*;
import java.util.*;
import java.util.zip.*;

/**Test that critical classes can be correctly serialised and deserialised.
 * This is useful for items that will be:
 * <ul>
 * <li>Cached/persisted to disc, especially those that must last a long time and are valuable.
 * <li>Sent over the wire.
 * </ul>
 */
public final class SerializationTest extends TestCase
    {
    public SerializationTest(String name)
        {
        super(name);
        }

    /**If true, print out serialised value in form that can be cut-n-paste back into test code. */
    private static final boolean PRINT_SER_BYTES = false;

    /**Maximum bytes to print on each line of serialised form. */
    private static final int MAX_BYTES_PER_LINE = 8;

    /**Check that the object passed can be serialised and deserialised.
     *
     * @param obj  must implement java.io.Serializable; never null
     * @return  the deserialised version of the object
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public static Object checkObjectCanBeSerialisedAndDeserialised(final Object obj)
        throws IOException,
               ClassNotFoundException
        {
        assertTrue("Original object does not implement Serializable", obj instanceof Serializable);

        // Serialise and deserialise to make a deep copy as if sent over the wire...
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(obj);
        oos.flush();
        final byte data[] = baos.toByteArray();

        if(PRINT_SER_BYTES)
            { dumpSerData(obj, data); }

        final Object obj2 = deserialiseFromByteArray(data);

        assertTrue("Deserialisation of a class " + obj.getClass() + " resulted in a null object", obj2 != null);
        return(obj2);
        }

    /**Dump serialised data in a form suitable to copy-and-paste into code. */
    private static void dumpSerData(final Object obj, final byte[] data)
        {
        System.out.println("***DUMPING DATA FOR " + obj);
        System.out.println("private static final byte serData[/*"+data.length+"*/] = {");
        int count = 0;
        final byte values[] = new byte[MAX_BYTES_PER_LINE];
        final StringBuilder sbTmp = new StringBuilder(); // Used to build one value to print.
        for(final byte b : data)
            {
            sbTmp.setLength(0);
            sbTmp.append(Integer.toString(b, 10));
            while(sbTmp.length() < 4) { sbTmp.insert(0, ' '); }
            sbTmp.append(", ");
            System.out.print(sbTmp.toString());
            values[count] = b;
            if(++count >= MAX_BYTES_PER_LINE)
                {
                dumpAsciiComment(sbTmp, count, values);
                count = 0;
                }
            }
        dumpAsciiComment(sbTmp, count, values);
        System.out.println("};");
        }

    /**Dump to System.out "count" characters from values, substituting non-printable-ASCII. */
    private static void dumpAsciiComment(final StringBuilder sbTmp, final int count, final byte[] values)
        {
        sbTmp.setLength(0);
        sbTmp.append("// ");
        for(int i = 0; i < count; ++i)
            {
            final int bc = values[i];
            if((bc > 32) && (bc < 127)) { sbTmp.append((char) bc); }
            else { sbTmp.append('.'); /* Substitute "safe" character. */ }
            }
        System.out.println(sbTmp.toString());
        }

    /**Deserialise an Object from the supplied serialised form. */
    static Object deserialiseFromByteArray(final byte[] data)
        throws IOException, ClassNotFoundException
        {
        final ByteArrayInputStream bais = new ByteArrayInputStream(data);
        final ObjectInputStream ois = new ObjectInputStream(bais);
        final Object obj2 = ois.readObject();
        return obj2;
        }

    /**Check that the passed (non-null) object compares equals after being serialised/deserialised.
     * This looks at equals() and hashCode().
     * <p>
     * Tests in passing that the object <em>can</em> be (de)serialised.
     *
     * @throws Exception  in case of difficulty.
     * @return deserialised object in case we want to perform extra tests on it
     */
    public static Object checkSerialisationPreservesEquality(final Object obj)
        throws Exception
        {
        final Object obj2 = checkObjectCanBeSerialisedAndDeserialised(obj);

        assertTrue("Equality test fails for orig.equals(copy)", obj.equals(obj2));
        assertTrue("Equality test fails for copy.equals(orig)", obj2.equals(obj));
        assertTrue("Hash codes must be equal", obj.hashCode() == obj2.hashCode());

        return(obj2);
        }

    /**Basic tests for those objects whose empty instances should have zero hash and compare equal.
     * (Empty instances are those made with the default constructor.)
     * <p>
     * Most of our persistable and transmitable-across-the-wire objects
     * should show this property so that we can easily tell empty objects
     * from ones containing useful data for example.
     * <p>
     * These objects should also be Serializable ((de)serializable) and still
     * compare equal.
     *
     * @throws Exception  in case of difficulty.
     */
    public static void checkEmptyInstancesAreZeroAndEqualAndSerializable(final Class c)
        throws Exception
        {
        final Object empty = c.newInstance();
        assertTrue(empty.hashCode() == 0);
        assertTrue(empty.equals(c.newInstance()));
        checkSerialisationPreservesEquality(empty);
        }

    /**Test AllExhibitImmutableData for (de)serialisability.
     * We check that:
     * <ul>
     * <li>Empty instances are consutructed correctly, with a zero timestamp
     *     (and a zero hash, and compare equal).
     * <li>Non-empty items can be serialised and deserialised correctly,
     *     including compute-on-demand cached values.
     * </ul>
     */
    public static void testAllExhibitImmutableData()
        throws Exception
        {
        checkEmptyInstancesAreZeroAndEqualAndSerializable(AllExhibitImmutableData.class);
        final AllExhibitImmutableData empty = new AllExhibitImmutableData();
        assertTrue(empty.timestamp == 0);
        assertTrue(empty.getAllExhibitNamesSet().isEmpty());
        assertTrue(empty.getAllStaticAttrs().isEmpty());

        // TODO: finish!
        }

    /**Test AllExhibitProperties for (de)serialisability.
     * We check that:
     * <ul>
     * <li>Empty instances are constructed correctly, with a zero timestamp
     *     (and a zero hash, and compare equal).
     * <li>Non-empty items can be serialised and deserialised correctly,
     *     including compute-on-demand cached values.
     * </ul>
     */
    public static void testAllExhibitProperties()
        throws Exception
        {
        checkEmptyInstancesAreZeroAndEqualAndSerializable(AllExhibitProperties.class);
        final AllExhibitProperties empty = new AllExhibitProperties();
        assertTrue(empty.longHash == 0);

        // Load a real instance from the filesystem...
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();
        final AllExhibitProperties aep = edfs.getAllExhibitProperties(-1);
        // Check that the AEP can be (de)serialised correctly.
        checkObjectCanBeSerialisedAndDeserialised(aep);
        checkSerialisationPreservesEquality(aep);


        // TODO: finish!
        }

    /**Test GenProps for (de)serialisability.
     * We check that:
     * <ul>
     * <li>Empty instances are constructed correctly, with a zero timestamp
     *     (and a zero hash, and compare equal).
     * <li>Non-empty items can be serialised and deserialised correctly,
     *     including compute-on-demand cached values.
     * </ul>
     */
    public static void testGenProps()
        throws Exception
        {
        final GenProps empty = new GenProps();
        assertTrue(empty.timestamp == 0);
        final GenProps empty2 =
            (GenProps) checkObjectCanBeSerialisedAndDeserialised(empty);
        assertTrue(empty2.timestamp == 0);

        // TODO: finish!
        }

    /**Test ExhibitThumbnails.
     * We want to check that object resolving works and we discard
     * duplicate values that equals NO_THUMBNAILS on deserialisation.
     */
    public static void testExhibitThumbnails()
        throws Exception
        {
        final Object obj2 = checkSerialisationPreservesEquality(ExhibitThumbnails.NO_THUMBNAILS);
        assertTrue("ExhibitThumbnails.NO_THUMBNAILS should resolve to itself (ie discard duplicates)",
                   obj2 == ExhibitThumbnails.NO_THUMBNAILS);

        // TODO: some tests on more realistic thumbnail objects.
        }

    /**Test Location data (esp Estd) can be (de)serialised.
     * We also test the LocationMap.
     */
    public void testLocation()
        throws Exception
        {
        // Set up some properties as if from a file:
        //    type=Estd
        //    # The error in each value is +/- the Err value.
        //    # Degrees East/North.
        //    E=-0.29
        //    EErr=0.01
        //    N=51.41
        //    NErr=0.01
        final Properties p1 = new Properties();
        p1.setProperty("type", "Estd");
        p1.setProperty("E", "-0.29");
        p1.setProperty("EErr", "0.01");
        p1.setProperty("N", "51.41");
        p1.setProperty("NErr", "0.01");
        // Construct a simple example Estd item.
        final Location.Estd e1 = new Location.Estd(
            true, // Claim that this data is "specific" to a particular exhibit.
            "", // No prefix on the property names.
            p1 // The properties.
            );
        // Check that we can (de)serialise this (and preserve equality).
        checkSerialisationPreservesEquality(e1);

        // Test behaviour of entire LocationMap.
        checkEmptyInstancesAreZeroAndEqualAndSerializable(LocationMap.class);
        }

    /**Raw value of String variable value for "hills are alive" value. */
    private static final String hillsAreAlive = " the hills are alive! 87239473075!^%#^";

    /**Serialised String variable value for "hills are alive" value. */
    private static final byte serData_hillsAreAlive_20050130[/*482*/] = {
        -84,  -19,    0,    5,  115,  114,    0,   46, // ....sr..
        111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
         46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
        114,   67,  111,  114,  101,   46,  118,   97, // rCore.va
        114,  115,   46,   83,  105,  109,  112,  108, // rs.Simpl
        101,   86,   97,  114,  105,   97,   98,  108, // eVariabl
        101,   86,   97,  108,  117,  101,  -71,   77, // eValue.M
         83, -103,  -28,  -60, -107,   95,    2,    0, // S...._..
          4,   74,    0,    9,  116,  105,  109,  101, // .J..time
        115,  116,   97,  109,  112,   76,    0,    3, // stampL..
        100,  101,  102,  116,    0,   53,   76,  111, // deft.5Lo
        114,  103,   47,  104,  100,   47,  100,   47, // rg/hd/d/
        112,  103,   50,  107,   47,  115,  118,  114, // pg2k/svr
         67,  111,  114,  101,   47,  118,   97,  114, // Core/var
        115,   47,   83,  105,  109,  112,  108,  101, // s/Simple
         86,   97,  114,  105,   97,   98,  108,  101, // Variable
         68,  101,  102,  105,  110,  105,  116,  105, // Definiti
        111,  110,   59,   76,    0,    9,  103,  108, // on;L..gl
        111,   98,   97,  108,   77,   97,  112,  116, // obalMapt
          0,   15,   76,  106,   97,  118,   97,   47, // ..Ljava/
        117,  116,  105,  108,   47,   77,   97,  112, // util/Map
         59,   76,    0,    5,  118,   97,  108,  117, // ;L..valu
        101,  116,    0,   18,   76,  106,   97,  118, // et..Ljav
         97,   47,  108,   97,  110,  103,   47,   79, // a/lang/O
         98,  106,  101,   99,  116,   59,  120,  112, // bject;xp
          0,    0,    1,    1,  -60, -115,   49,   29, // ......1.
        115,  114,    0,   51,  111,  114,  103,   46, // sr.3org.
        104,  100,   46,  100,   46,  112,  103,   50, // hd.d.pg2
        107,   46,  115,  118,  114,   67,  111,  114, // k.svrCor
        101,   46,  118,   97,  114,  115,   46,   83, // e.vars.S
        105,  109,  112,  108,  101,   86,   97,  114, // impleVar
        105,   97,   98,  108,  101,   68,  101,  102, // iableDef
        105,  110,  105,  116,  105,  111,  110,   32, // inition.
        -60,  -92,   85,  126,   79,  -65,  -35,    2, // ..U~O...
          0,    8,   90,    0,    5,  101,  118,  101, // ..Z..eve
        110,  116,   90,    0,    5,  108,  111,   99, // ntZ..loc
         97,  108,   73,    0,   17,  109,   97,  120, // alI..max
         68,  105,  102,  102,   69,  118,  101,  110, // DiffEven
        116,   67,  111,  117,  110,  116,   90,    0, // tCountZ.
         10,  112,  101,  114,  115,  105,  115,  116, // .persist
        101,  110,  116,   90,    0,    8,  114,  101, // entZ..re
         97,  100,   79,  110,  108,  121,   73,    0, // adOnlyI.
          4,  116,  121,  112,  101,   76,    0,   14, // .typeL..
        101,  118,   80,  101,  114,  105,  111,  100, // evPeriod
         83,  117,   98,  115,  101,  116,  116,    0, // Subsett.
         19,   76,  106,   97,  118,   97,   47,  117, // .Ljava/u
        116,  105,  108,   47,   69,  110,  117,  109, // til/Enum
         83,  101,  116,   59,   76,    0,    4,  110, // Set;L..n
         97,  109,  101,  116,    0,   18,   76,  106, // amet..Lj
         97,  118,   97,   47,  108,   97,  110,  103, // ava/lang
         47,   83,  116,  114,  105,  110,  103,   59, // /String;
        120,  112,    0,    1,    0,    0,    0,    0, // xp......
          0,    0,    0,    0,    0,    2,  112,  116, // ......pt
          0,   14,  116,  101,  115,  116,   83,  101, // ..testSe
        114,   46,  118,   97,  108,  117,  101,   51, // r.value3
        112,  116,    0,   38,   32,  116,  104,  101, // pt.&.the
         32,  104,  105,  108,  108,  115,   32,   97, // .hills.a
        114,  101,   32,   97,  108,  105,  118,  101, // re.alive
         33,   32,   56,   55,   50,   51,   57,   52, // !.872394
         55,   51,   48,   55,   53,   33,   94,   37, // 73075!^%
         35,   94,                                     // #^
    };

    /**Test that variables can be correctly serialised/deserialised.
     * We check all the built-in definitions and
     * instances/values of several types.
     */
    public static void testSimpleVariables()
        throws Exception
        {
        // Test that every definition can be correctly serialised and deserialised.
        for(final Iterator it = SystemVariables.defs.iterator(); it.hasNext(); )
            { checkSerialisationPreservesEquality(it.next()); }

        // Test an instance of TYPE_NONE...
        final SimpleVariableValue svv1 =
            new SimpleVariableValue(SystemVariables.KEEP_ALIVE,
                                    null);
        checkSerialisationPreservesEquality(svv1);

        // Test a simple instance of TYPE_NUMBER...
        final SimpleVariableValue svv2 =
            new SimpleVariableValue(SystemVariables.ThroughputMonitorFilter_CLIENT_COUNT,
                                    new Integer(rnd.nextInt(9999)));
        checkSerialisationPreservesEquality(svv2);

        // Try this on a few random variable definitions
        // plus all extant system variable values.
        final List<SimpleVariableDefinition> testDefSet = new ArrayList<SimpleVariableDefinition>();
        testDefSet.addAll(SystemVariables.defs);
        for(int i = 37; --i >= 0; )
            { testDefSet.add(SystemVariablesTest.makeRandomSimpleVariableDefinition()); }

        // Try out one possible value for each test definition.
        for(final Iterator it = testDefSet.iterator(); it.hasNext(); )
            {
            final SimpleVariableDefinition def =
                (SimpleVariableDefinition) it.next();
            final SimpleVariableValue svv = SystemVariablesTest.makeRandomSimpleVariableValue(def);
            checkSerialisationPreservesEquality(svv);
            }

        // Check that we can deserialise at least one frozen serialised (String) non-standard value.
        // This is a test of robustness of both the definition and the value objects.
        final SimpleVariableDefinition newDef = new SimpleVariableDefinition("testSer.value3",
                                                                             SimpleVariableDefinition.TYPE_STRING);
        final SimpleVariableValue svv3 =
            new SimpleVariableValue(newDef, hillsAreAlive);
        checkSerialisationPreservesEquality(svv3);
        // Now check that deserialising the static data gives an equal value.
        assertTrue("Deserialised frozen value must be equal to newly constructed one",
                   svv3.equals(deserialiseFromByteArray(serData_hillsAreAlive_20050130)));
        }


    /**First test string value for EVV; has a count of 1. */
    private static final String evvTestString1 = "45dgjytref";

    /**Second test string value for EVV; has a count of 2. */
    private static final String evvTestString2 = "dfgtr67kjhfh";

    /**Total event count in map. */
    private static final int evv_TOTAL_EVENT_COUNT = 5;

    /**Old serialised format of a single EVV with embedded Map&lt;Object, ValueInfo&gt;. */
    private static final byte serData_evv1_20050130_1748[/*812*/] = {
     -84,  -19,    0,    5,  115,  114,    0,   45, // ....sr.-
     111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
      46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
     114,   67,  111,  114,  101,   46,  118,   97, // rCore.va
     114,  115,   46,   69,  118,  101,  110,  116, // rs.Event
      86,   97,  114,  105,   97,   98,  108,  101, // Variable
      86,   97,  108,  117,  101,   70,  -78,  -84, // ValueF..
     102,   92,  -52,   21,  -95,    2,    0,    6, // f\......
      90,    0,   13,   97,  117,  116,  104,  111, // Z..autho
     114,   97,  116,  105,  116,  105,  118,  101, // ratitive
      74,    0,   14,  105,  110,  116,  101,  114, // J..inter
     118,   97,  108,   78,  117,  109,   98,  101, // valNumbe
     114,   73,    0,   15,  116,  111,  116,   97, // rI..tota
     108,   69,  118,  101,  110,  116,   67,  111, // lEventCo
     117,  110,  116,   76,    0,    3,  100,  101, // untL..de
     102,  116,    0,   53,   76,  111,  114,  103, // ft.5Lorg
      47,  104,  100,   47,  100,   47,  112,  103, // /hd/d/pg
      50,  107,   47,  115,  118,  114,   67,  111, // 2k/svrCo
     114,  101,   47,  118,   97,  114,  115,   47, // re/vars/
      83,  105,  109,  112,  108,  101,   86,   97, // SimpleVa
     114,  105,   97,   98,  108,  101,   68,  101, // riableDe
     102,  105,  110,  105,  116,  105,  111,  110, // finition
      59,   76,    0,    4,  105,  110,  102,  111, // ;L..info
     116,    0,   15,   76,  106,   97,  118,   97, // t..Ljava
      47,  117,  116,  105,  108,   47,   77,   97, // /util/Ma
     112,   59,   76,    0,    6,  112,  101,  114, // p;L..per
     105,  111,  100,  116,    0,   40,   76,  111, // iodt.(Lo
     114,  103,   47,  104,  100,   47,  100,   47, // rg/hd/d/
     112,  103,   50,  107,   47,  115,  118,  114, // pg2k/svr
      67,  111,  114,  101,   47,  118,   97,  114, // Core/var
     115,   47,   69,  118,  101,  110,  116,   80, // s/EventP
     101,  114,  105,  111,  100,   59,  120,  112, // eriod;xp
       1,    0,    0,    0,    0,    0,    0,    0, // ........
       1,    0,    0,    0,    5,  115,  114,    0, // .....sr.
      51,  111,  114,  103,   46,  104,  100,   46, // 3org.hd.
     100,   46,  112,  103,   50,  107,   46,  115, // d.pg2k.s
     118,  114,   67,  111,  114,  101,   46,  118, // vrCore.v
      97,  114,  115,   46,   83,  105,  109,  112, // ars.Simp
     108,  101,   86,   97,  114,  105,   97,   98, // leVariab
     108,  101,   68,  101,  102,  105,  110,  105, // leDefini
     116,  105,  111,  110,   32,  -60,  -92,   85, // tion...U
     126,   79,  -65,  -35,    2,    0,    8,   90, // ~O.....Z
       0,    5,  101,  118,  101,  110,  116,   90, // ..eventZ
       0,    5,  108,  111,   99,   97,  108,   73, // ..localI
       0,   17,  109,   97,  120,   68,  105,  102, // ..maxDif
     102,   69,  118,  101,  110,  116,   67,  111, // fEventCo
     117,  110,  116,   90,    0,   10,  112,  101, // untZ..pe
     114,  115,  105,  115,  116,  101,  110,  116, // rsistent
      90,    0,    8,  114,  101,   97,  100,   79, // Z..readO
     110,  108,  121,   73,    0,    4,  116,  121, // nlyI..ty
     112,  101,   76,    0,   14,  101,  118,   80, // peL..evP
     101,  114,  105,  111,  100,   83,  117,   98, // eriodSub
     115,  101,  116,  116,    0,   19,   76,  106, // sett..Lj
      97,  118,   97,   47,  117,  116,  105,  108, // ava/util
      47,   69,  110,  117,  109,   83,  101,  116, // /EnumSet
      59,   76,    0,    4,  110,   97,  109,  101, // ;L..name
     116,    0,   18,   76,  106,   97,  118,   97, // t..Ljava
      47,  108,   97,  110,  103,   47,   83,  116, // /lang/St
     114,  105,  110,  103,   59,  120,  112,    1, // ring;xp.
       0,    0,    0,    0,  101,    1,    0,    0, // ....e...
       0,    0,    2,  112,  116,    0,   18,  116, // ...pt..t
     101,  115,  116,   83,  101,  114,   46,  115, // estSer.s
     116,  114,  105,  110,  103,   46,  118,   97, // tring.va
     108,  115,  114,    0,   17,  106,   97,  118, // lsr..jav
      97,   46,  117,  116,  105,  108,   46,   72, // a.util.H
      97,  115,  104,   77,   97,  112,    5,    7, // ashMap..
     -38,  -63,  -61,   22,   96,  -47,    3,    0, // ....`...
       2,   70,    0,   10,  108,  111,   97,  100, // .F..load
      70,   97,   99,  116,  111,  114,   73,    0, // FactorI.
       9,  116,  104,  114,  101,  115,  104,  111, // .thresho
     108,  100,  120,  112,   63,   64,    0,    0, // ldxp?@..
       0,    0,    0,    3,  119,    8,    0,    0, // ....w...
       0,    4,    0,    0,    0,    2,  116,    0, // ......t.
      10,   52,   53,  100,  103,  106,  121,  116, // .45dgjyt
     114,  101,  102,  115,  114,    0,   55,  111, // refsr.7o
     114,  103,   46,  104,  100,   46,  100,   46, // rg.hd.d.
     112,  103,   50,  107,   46,  115,  118,  114, // pg2k.svr
      67,  111,  114,  101,   46,  118,   97,  114, // Core.var
     115,   46,   69,  118,  101,  110,  116,   86, // s.EventV
      97,  114,  105,   97,   98,  108,  101,   86, // ariableV
      97,  108,  117,  101,   36,   86,   97,  108, // alue$Val
     117,  101,   73,  110,  102,  111,   12,   82, // ueInfo.R
    -114,  -88,   -5,  -45,   19,   93,    2,    0, // .....]..
       2,   73,    0,    5,   99,  111,  117,  110, // .I..coun
     116,   73,    0,    4,  114,   97,  110,  107, // tI..rank
     120,  112,    0,    0,    0,    1,    0,    0, // xp......
       0,    1,  116,    0,   12,  100,  102,  103, // ..t..dfg
     116,  114,   54,   55,  107,  106,  104,  102, // tr67kjhf
     104,  115,  113,    0,  126,    0,   13,    0, // hsq.~...
       0,    0,    2,    0,    0,    0,    0,  120, // .......x
     126,  114,    0,   38,  111,  114,  103,   46, // ~r.&org.
     104,  100,   46,  100,   46,  112,  103,   50, // hd.d.pg2
     107,   46,  115,  118,  114,   67,  111,  114, // k.svrCor
     101,   46,  118,   97,  114,  115,   46,   69, // e.vars.E
     118,  101,  110,  116,   80,  101,  114,  105, // ventPeri
     111,  100,    0,    0,    0,    0,    0,    0, // od......
       0,    0,   18,    0,    0,  120,  114,    0, // .....xr.
      14,  106,   97,  118,   97,   46,  108,   97, // .java.la
     110,  103,   46,   69,  110,  117,  109,    0, // ng.Enum.
       0,    0,    0,    0,    0,    0,    0,   18, // ........
       0,    0,  120,  112,  116,    0,    5,   86, // ..xpt..V
      76,   79,   78,   71,                         // LONG
    };

    /**New (more compact) serialised format of a single EVV with counts[]/values[]. */
    private static final byte serData_evv1_20050131_1512[/*743*/] = {
     -84,  -19,    0,    5,  115,  114,    0,   45, // ....sr.-
     111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
      46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
     114,   67,  111,  114,  101,   46,  118,   97, // rCore.va
     114,  115,   46,   69,  118,  101,  110,  116, // rs.Event
      86,   97,  114,  105,   97,   98,  108,  101, // Variable
      86,   97,  108,  117,  101,   70,  -78,  -84, // ValueF..
     102,   92,  -52,   21,  -95,    3,    0,    8, // f\......
      90,    0,   13,   97,  117,  116,  104,  111, // Z..autho
     114,   97,  116,  105,  116,  105,  118,  101, // ratitive
      74,    0,   14,  105,  110,  116,  101,  114, // J..inter
     118,   97,  108,   78,  117,  109,   98,  101, // valNumbe
     114,   73,    0,   15,  116,  111,  116,   97, // rI..tota
     108,   69,  118,  101,  110,  116,   67,  111, // lEventCo
     117,  110,  116,   91,    0,    6,   99,  111, // unt[..co
     117,  110,  116,  115,  116,    0,    2,   91, // untst..[
      73,   76,    0,    3,  100,  101,  102,  116, // IL..deft
       0,   53,   76,  111,  114,  103,   47,  104, // .5Lorg/h
     100,   47,  100,   47,  112,  103,   50,  107, // d/d/pg2k
      47,  115,  118,  114,   67,  111,  114,  101, // /svrCore
      47,  118,   97,  114,  115,   47,   83,  105, // /vars/Si
     109,  112,  108,  101,   86,   97,  114,  105, // mpleVari
      97,   98,  108,  101,   68,  101,  102,  105, // ableDefi
     110,  105,  116,  105,  111,  110,   59,   76, // nition;L
       0,    4,  105,  110,  102,  111,  116,    0, // ..infot.
      15,   76,  106,   97,  118,   97,   47,  117, // .Ljava/u
     116,  105,  108,   47,   77,   97,  112,   59, // til/Map;
      76,    0,    6,  112,  101,  114,  105,  111, // L..perio
     100,  116,    0,   40,   76,  111,  114,  103, // dt.(Lorg
      47,  104,  100,   47,  100,   47,  112,  103, // /hd/d/pg
      50,  107,   47,  115,  118,  114,   67,  111, // 2k/svrCo
     114,  101,   47,  118,   97,  114,  115,   47, // re/vars/
      69,  118,  101,  110,  116,   80,  101,  114, // EventPer
     105,  111,  100,   59,   91,    0,    6,  118, // iod;[..v
      97,  108,  117,  101,  115,  116,    0,   19, // aluest..
      91,   76,  106,   97,  118,   97,   47,  108, // [Ljava/l
      97,  110,  103,   47,   79,   98,  106,  101, // ang/Obje
      99,  116,   59,  120,  112,    1,    0,    0, // ct;xp...
       0,    0,    0,    0,    0,    1,    0,    0, // ........
       0,    5,  117,  114,    0,    2,   91,   73, // ..ur..[I
      77,  -70,   96,   38,  118,  -22,  -78,  -91, // M.`&v...
       2,    0,    0,  120,  112,    0,    0,    0, // ...xp...
       2,    0,    0,    0,    2,    0,    0,    0, // ........
       1,  115,  114,    0,   51,  111,  114,  103, // .sr.3org
      46,  104,  100,   46,  100,   46,  112,  103, // .hd.d.pg
      50,  107,   46,  115,  118,  114,   67,  111, // 2k.svrCo
     114,  101,   46,  118,   97,  114,  115,   46, // re.vars.
      83,  105,  109,  112,  108,  101,   86,   97, // SimpleVa
     114,  105,   97,   98,  108,  101,   68,  101, // riableDe
     102,  105,  110,  105,  116,  105,  111,  110, // finition
      32,  -60,  -92,   85,  126,   79,  -65,  -35, // ...U~O..
       2,    0,    8,   90,    0,    5,  101,  118, // ...Z..ev
     101,  110,  116,   90,    0,    5,  108,  111, // entZ..lo
      99,   97,  108,   73,    0,   17,  109,   97, // calI..ma
     120,   68,  105,  102,  102,   69,  118,  101, // xDiffEve
     110,  116,   67,  111,  117,  110,  116,   90, // ntCountZ
       0,   10,  112,  101,  114,  115,  105,  115, // ..persis
     116,  101,  110,  116,   90,    0,    8,  114, // tentZ..r
     101,   97,  100,   79,  110,  108,  121,   73, // eadOnlyI
       0,    4,  116,  121,  112,  101,   76,    0, // ..typeL.
      14,  101,  118,   80,  101,  114,  105,  111, // .evPerio
     100,   83,  117,   98,  115,  101,  116,  116, // dSubsett
       0,   19,   76,  106,   97,  118,   97,   47, // ..Ljava/
     117,  116,  105,  108,   47,   69,  110,  117, // util/Enu
     109,   83,  101,  116,   59,   76,    0,    4, // mSet;L..
     110,   97,  109,  101,  116,    0,   18,   76, // namet..L
     106,   97,  118,   97,   47,  108,   97,  110, // java/lan
     103,   47,   83,  116,  114,  105,  110,  103, // g/String
      59,  120,  112,    1,    0,    0,    0,    0, // ;xp.....
     101,    1,    0,    0,    0,    0,    2,  112, // e......p
     116,    0,   18,  116,  101,  115,  116,   83, // t..testS
     101,  114,   46,  115,  116,  114,  105,  110, // er.strin
     103,   46,  118,   97,  108,  112,  126,  114, // g.valp~r
       0,   38,  111,  114,  103,   46,  104,  100, // .&org.hd
      46,  100,   46,  112,  103,   50,  107,   46, // .d.pg2k.
     115,  118,  114,   67,  111,  114,  101,   46, // svrCore.
     118,   97,  114,  115,   46,   69,  118,  101, // vars.Eve
     110,  116,   80,  101,  114,  105,  111,  100, // ntPeriod
       0,    0,    0,    0,    0,    0,    0,    0, // ........
      18,    0,    0,  120,  114,    0,   14,  106, // ...xr..j
      97,  118,   97,   46,  108,   97,  110,  103, // ava.lang
      46,   69,  110,  117,  109,    0,    0,    0, // .Enum...
       0,    0,    0,    0,    0,   18,    0,    0, // ........
     120,  112,  116,    0,    5,   86,   76,   79, // xpt..VLO
      78,   71,  117,  114,    0,   19,   91,   76, // NGur..[L
     106,   97,  118,   97,   46,  108,   97,  110, // java.lan
     103,   46,   79,   98,  106,  101,   99,  116, // g.Object
      59, -112,  -50,   88,  -97,   16,  115,   41, // ;..X..s)
     108,    2,    0,    0,  120,  112,    0,    0, // l...xp..
       0,    2,  116,    0,   12,  100,  102,  103, // ..t..dfg
     116,  114,   54,   55,  107,  106,  104,  102, // tr67kjhf
     104,  116,    0,   10,   52,   53,  100,  103, // ht..45dg
     106,  121,  116,  114,  101,  102,  120,       // jytrefx
    };

    /**Test that we can succesfully create and (de)serialise event container objects...
     * In particular we ensure that we can recover data in older formats where necessary
     * so that we don't lose hard-to-collect historic data.
     */
    public static void testEventVariableValues()
        throws Exception
        {
        // Check that we can deserialise at least one frozen serialised (String) non-standard value.
        // This is a test of robustness of both the definition and the value objects.
        final SimpleVariableDefinition newDef = new SimpleVariableDefinition("testSer.string.val",
                                                                             SimpleVariableDefinition.TYPE_STRING,
                                                                             false, // Global.
                                                                             true,  // Persistent.
                                                                             false, // Read-write.
                                                                             true, 101, null);
        final Object values[] = { evvTestString2, evvTestString1 };
        final int counts[] = { 2, 1 };
//        final Map<Object, EventVariableValue.ValueInfo> map = new HashMap<Object, EventVariableValue.ValueInfo>();
//        map.put(evvTestString1, new EventVariableValue.ValueInfo(1, 1));
//        map.put(evvTestString2, new EventVariableValue.ValueInfo(0, 2));
        final EventVariableValue evv = new EventVariableValue(true, // Authoritative...
                                                              newDef,
                                                              EventPeriod.VLONG,
                                                              1,
                                                              evv_TOTAL_EVENT_COUNT,
                                                              values, counts);
        // Make a deep copy via deserialisation...
        final EventVariableValue evvDS = (EventVariableValue)
                checkSerialisationPreservesEquality(evv);
        // Reconstruct from byte[] format before the internal format change...
        final EventVariableValue evvFrozen20050130 = (EventVariableValue)
            deserialiseFromByteArray(serData_evv1_20050130_1748);
        // Reconstruct from byte[] more compact format change...
        final EventVariableValue evvFrozen20050131 = (EventVariableValue)
            deserialiseFromByteArray(serData_evv1_20050131_1512);

        // Perform the same tests on all copies of the value.
        final EventVariableValue evvs[] = { evv, evvDS, evvFrozen20050130, evvFrozen20050131 };
        for(final EventVariableValue e: evvs)
            {
            assertTrue("EVV must equal itself", e.equals(e));
            assertTrue("EVV must equal main copy", e.equals(evv));
            assertEquals("Total event count must be correct", evv_TOTAL_EVENT_COUNT, e.getTotalEventCount());
            assertEquals("Total different events must be correct", 2, e.getTotalDistinctValues());
            assertEquals("Total count of first value must be correct", 1, e.getCount(evvTestString1));
            assertEquals("Total count of second value must be correct", 2, e.getCount(evvTestString2));
            assertEquals("Rank of first value must be correct", 1, e.getRank(evvTestString1));
            assertEquals("Rank of second value must be correct", 0, e.getRank(evvTestString2));
            assertEquals("Rank of non-existant value must be correct", Integer.MAX_VALUE, e.getRank("not there"));
            }
        }

    /**Empty ROByteArray (ie wrapping a byte[0]). */
    private static final byte serData_ROByteArray_EMPTY_20050522[/*92*/] = {
     -84,  -19,    0,    5,  115,  114,    0,   33, // ....sr.!
     111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
      46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
     114,   67,  111,  114,  101,   46,   82,   79, // rCore.RO
      66,  121,  116,  101,   65,  114,  114,   97, // ByteArra
     121,   52,  -45,  -14,   29,  -75,  -97, -121, // y4......
     -40,    2,    0,    1,   91,    0,    7,  112, // ....[..p
      97,  121,  108,  111,   97,  100,  116,    0, // ayloadt.
       2,   91,   66,  120,  112,  117,  114,    0, // .[Bxpur.
       2,   91,   66,  -84,  -13,   23,   -8,    6, // .[B.....
       8,   84,  -32,    2,    0,    0,  120,  112, // .T....xp
       0,    0,    0,    0, // ....
    };

    /**Value to check compression is working and can be correctly undone even on non-ASCII/non-8-bit text. */
    private static final String ROBA_compS1 = "The quick Brown fox jumped over the PM Blair [satire over, Ed] " +
        "over and over and over and over and over again until he got bored... " +
        "(And now here is the news in English: \u7777\u9231\u0000\r\n\t...)";

    /**On-the-wire format on ROBA_compS1. */
    private static final byte serData_ROBA_compS1_20050522[/*233*/] = {
     -84,  -19,    0,    5,  115,  114,    0,   33, // ....sr.!
     111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
      46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
     114,   67,  111,  114,  101,   46,   82,   79, // rCore.RO
      66,  121,  116,  101,   65,  114,  114,   97, // ByteArra
     121,   52,  -45,  -14,   29,  -75,  -97, -121, // y4......
     -40,    2,    0,    1,   91,    0,    7,  112, // ....[..p
      97,  121,  108,  111,   97,  100,  116,    0, // ayloadt.
       2,   91,   66,  120,  112,  117,  114,    0, // .[Bxpur.
       2,   91,   66,  -84,  -13,   23,   -8,    6, // .[B.....
       8,   84,  -32,    2,    0,    0,  120,  112, // .T....xp
       0,    0,    0, -115,   99,  -40,   25, -110, // ....c...
    -111,  -86,   80,   88, -102, -103, -100,  -83, // ..PX....
     -32,   84, -108,   95,  -98,  -89, -112, -106, // .T._....
      95,  -95, -112,   85, -102,   91, -112, -102, // _..U.[..
     -94, -112,   95, -106,   90,  -92,   80,    2, // .._.Z.P.
    -108,   15,  -16,   85,  112,  -54,   73,  -52, // ...Up.I.
      44,   82, -120,   46,   78,   44,  -55,   44, // ,R..N,.,
      74,    5,  -53,  -24,   40,  -72,  -90,  -60, // J...(...
      66,  -44,   36,  -26,  -91,   16,  102,  -92, // B.$...f.
      39,  102,  -26,   41, -108,  -26, -107,  100, // 'f.)...d
     -26,   40,    0, -115,   76,  -49,   47,   81, // .(..L./Q
      72,  -54,   47,   74,   77,  -47,  -45,  -45, // H./JM...
      83,  -48,  112,    4,  -86,  -54,  -53,   47, // S.p..../
       7,   74,    0,   13,  -49,   44,    6,   91, // .J...,.[
    -102, -105,   90,   94,  -84,    0,  -44,  -30, // ..Z^....
    -102, -105,  -98, -109,   89, -100,   97,  -91, // ....Y.a.
     -16,  124,  -18,  -10, -105,   29,   27,   15, // .|......
      52,  -16,  114,  113,    2,   53,  105,    2, // 4.rq.5i.
       0, // .
    };

    /**Test handling of ROByteArray.
     */
    public void testROByteArray()
        throws Exception
        {
        // Check empty instance can be deserialised and is instance controlled!
        assertEquals("EMPTY ROByteArray must be deserialised correctly.",
                     ROByteArray.EMPTY, checkSerialisationPreservesEquality(ROByteArray.EMPTY));
        assertSame("EMPTY ROByteArray must be deserialised as itself!",
                     ROByteArray.EMPTY, checkSerialisationPreservesEquality(ROByteArray.EMPTY));
        // Reconstruct from byte[] format.
        final ROByteArray emptyDeSer = (ROByteArray)
            deserialiseFromByteArray(serData_ROByteArray_EMPTY_20050522);
        assertEquals("Must be able to deserialise original on-the-wire format (EMPTY)",
                     ROByteArray.EMPTY, emptyDeSer);
        assertEquals("Must get back empty array on deserialisation", 0, emptyDeSer.length());

        // Check non-empty instance can be (de)serialised and (de)compressed.
        final ROByteArray c1 = ROByteArray.compressFromString(ROBA_compS1);
        assertEquals("Compressable String in ROByteArray must be deserialised correctly.",
                     c1, checkSerialisationPreservesEquality(c1));
        final ROByteArray c1DeSer = (ROByteArray)
            deserialiseFromByteArray(serData_ROBA_compS1_20050522);
        assertEquals("Must be able to deserialise original on-the-wire format (non-empty, compressed String)",
                     c1, c1DeSer);
        assertEquals("Must be able to extract String from serialised form",
                     ROBA_compS1, ROByteArray.uncompressToString(c1DeSer));
        }

    /**Tests of accessions of serialisation of AccessionData.
     * This is both for Java serialisation and XML format.
     */
    public static void testAccessionDataSer()
        throws Exception
        {
        assertSame("deserialising EMPTY should give object == EMPTY",
                   checkSerialisationPreservesEquality(AccessionData.EMPTY),
                   AccessionData.EMPTY);

        // Check that we can (de)serialise via Java native and XML formats.
        final int NUM_TESTS = 10;
        for(int i = NUM_TESTS; --i >= 0; )
            {
            final byte[] dummyMD5Data = new byte[16];
            rnd.nextBytes(dummyMD5Data);
            final boolean b1 = rnd.nextBoolean();
            final boolean b2 = rnd.nextBoolean();
            final boolean b3 = rnd.nextBoolean();
            final boolean b4 = rnd.nextBoolean();
            final AccessionData ad = new AccessionData(
                b1 ? null : (System.currentTimeMillis()),
                b2 ? null : (new Long((rnd.nextLong() >>> 1) | 1)),
                b3 ? null : (new Integer(rnd.nextInt())),
                b4 ? null : (new ROByteArray(dummyMD5Data)));

            // Test native-format (de)serialisation...
            checkSerialisationPreservesEquality(ad);

            // Test serialisation via XML format.
            final String xml = TextUtils.toXML(ad.getAsDOM(), false, true);
if(i == 0) { System.out.println("[AccessionData XML example: "+xml+".]"); }
            assertEquals("conversion to XML and back should preserve equality",
                         ad,
                         AccessionData.parseFromXML(xml));
            }
        }

    /**On-the-wire format of Thumbnail. */
    private static final byte serData_Thumbnail_20060704[/*199*/] = {
        -84,  -19,    0,    5,  115,  114,    0,   49, // ....sr.1
        111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
         46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
        114,   67,  111,  114,  101,   46,   69,  120, // rCore.Ex
        104,  105,   98,  105,  116,   84,  104,  117, // hibitThu
        109,   98,  110,   97,  105,  108,  115,   36, // mbnails$
         84,  104,  117,  109,   98,  110,   97,  105, // Thumbnai
        108,  -18,   63,  -36,   92,   71,   75,   76, // l.?.\GKL
        -46,    2,    0,    2,   91,    0,    4,  100, // ....[..d
         97,  116,   97,  116,    0,    2,   91,   66, // atat..[B
         76,    0,    5,  120,  121,   68,  105,  109, // L..xyDim
        116,    0,   20,   76,  106,   97,  118,   97, // t..Ljava
         47,   97,  119,  116,   47,   68,  105,  109, // /awt/Dim
        101,  110,  115,  105,  111,  110,   59,  120, // ension;x
        112,  117,  114,    0,    2,   91,   66,  -84, // pur..[B.
        -13,   23,   -8,    6,    8,   84,  -32,    2, // .....T..
          0,    0,  120,  112,    0,    0,    0,    3, // ..xp....
          1,    2,    3,  115,  114,    0,   18,  106, // ...sr..j
         97,  118,   97,   46,   97,  119,  116,   46, // ava.awt.
         68,  105,  109,  101,  110,  115,  105,  111, // Dimensio
        110,   65, -114,  -39,  -41,  -84,   95,   68, // nA...._D
         20,    2,    0,    2,   73,    0,    6,  104, // ....I..h
        101,  105,  103,  104,  116,   73,    0,    5, // eightI..
        119,  105,  100,  116,  104,  120,  112,    0, // widthxp.
          0,    0,    4,    0,    0,    0,    3, // .......
        };

    /**Test that we can (de)serialise Thumbnail data OK. */
    public static void testThumbnailSer()
        throws Exception
        {
        final Dimension dim1 = new Dimension(3, 4);
        final ExhibitThumbnails.Thumbnail tn1 =
            new ExhibitThumbnails.Thumbnail(new byte[]{1,2,3}, dim1);
        checkSerialisationPreservesEquality(tn1);
        
        final ExhibitThumbnails.Thumbnail tn1Frozen20060704 = (ExhibitThumbnails.Thumbnail)
            deserialiseFromByteArray(serData_Thumbnail_20060704);
        assertEquals("Must preserve equality when deserialised", tn1, tn1Frozen20060704);
        }
    
    /**Test the robustness of our decompression routines in conjunction with deserialization.
     * We expect clean failures rather than hangs or messy crashes,
     * whatever junk is thrown at the decompressors.
     */
    public void testDecompDeserRobustness()
        {
        // We construct various badly-broken samples of input data.
        final byte randomRubbish[] = new byte[rnd.nextInt(1<<18)];
        rnd.nextBytes(randomRubbish);
        final byte[][] testData = {
                        new byte[0] /* Empty. */ ,            
                        new byte[1] /* Impossible/invalid single zero byte for compressed results. */ ,
                        randomRubbish
                                 };
        
        for(final byte[] data : testData)
            {
            for(final CompressionLevel cl : CompressionLevel.values())
                {               
                if(cl.getLevel() > GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL.getLevel())
                    {
                    Main.getOut().println("[Skipping test on unsupported compression level "+cl+"...]");
                    continue;
                    }
                try
                    {
                    final ByteArrayInputStream bais = new ByteArrayInputStream(data);
                    final InputStream wis = GenUtils.wrapForDecompression(bais, cl);
                    final ObjectInputStream ois = new ObjectInputStream(wis);
                    ois.readObject();
                    }
                catch(final IOException e) 
                    {
                    // IOException is OK in face of garbage input.
                    }
                catch(final ClassNotFoundException e)
                    {
                    // ClassNotFoundException is OK though rare in face of garbage input.
                    }
                }
            }
        }

    /**Empty AEP delta instance. */
    private static final byte serData_AllExhibitPropertiesDelta_empty[/*408*/] = {
            -84,  -19,    0,    5,  115,  114,    0,   47, // ....sr./
            111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
             46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
            114,   67,  111,  114,  101,   46,   65,  108, // rCore.Al
            108,   69,  120,  104,  105,   98,  105,  116, // lExhibit
             80,  114,  111,  112,  101,  114,  116,  105, // Properti
            101,  115,   68,  101,  108,  116,   97,  119, // esDeltaw
            118, -113,   43,   63,  120,   56,   79,    3, // v.+?x8O.
              0,    8,   74,    0,   24,  104,   97,  115, // ..J..has
            104,   78,  111,  116,   67,  104,   97,  110, // hNotChan
            103,  101,  100,   83,  105,  110,   99,  101, // gedSince
             65,  102,  116,  101,  114,   73,    0,   15, // AfterI..
            108,  101,  110,  103,  116,  104,   65,   69, // lengthAE
             73,   68,   65,  102,  116,  101,  114,   73, // IDAfterI
              0,   16,  108,  101,  110,  103,  116,  104, // ..length
             65,   69,   73,   68,   66,  101,  102,  111, // AEIDBefo
            114,  101,   74,    0,   16,  108,  111,  110, // reJ..lon
            103,   72,   97,  115,  104,   65,   69,   80, // gHashAEP
             65,  102,  116,  101,  114,   74,    0,   17, // AfterJ..
            108,  111,  110,  103,   72,   97,  115,  104, // longHash
             65,   69,   80,   66,  101,  102,  111,  114, // AEPBefor
            101,   74,    0,   18,  116,  105,  109,  101, // eJ..time
            115,  116,   97,  109,  112,   65,   69,   73, // stampAEI
             68,   65,  102,  116,  101,  114,   74,    0, // DAfterJ.
             19,  116,  105,  109,  101,  115,  116,   97, // .timesta
            109,  112,   65,   69,   73,   68,   66,  101, // mpAEIDBe
            102,  111,  114,  101,   76,    0,    9,  101, // foreL..e
            112,  103,  105,   65,  102,  116,  101,  114, // pgiAfter
            116,    0,   51,   76,  111,  114,  103,   47, // t.3Lorg/
            104,  100,   47,  100,   47,  112,  103,   50, // hd/d/pg2
            107,   47,  115,  118,  114,   67,  111,  114, // k/svrCor
            101,   47,   69,  120,  104,  105,   98,  105, // e/Exhibi
            116,   80,  114,  111,  112,  115,   71,  108, // tPropsGl
            111,   98,   97,  108,   73,  109,  109,  117, // obalImmu
            116,   97,   98,  108,  101,   59,  120,  112, // table;xp
              0,    0,    0,    0,    0,    0,    0,    0, // ........
              0,    0,    0,    0,    0,    0,    0,    0, // ........
              0,    0,    0,    0,    0,    0,    0,    0, // ........
              0,    0,    0,    0,    0,    0,    0,    0, // ........
              0,    0,    0,    0,    0,    0,    0,    0, // ........
              0,    0,    0,    0,    0,    0,    0,    0, // ........
            112,  117,  114,    0,   57,   91,   76,  111, // pur.9[Lo
            114,  103,   46,  104,  100,   46,  100,   46, // rg.hd.d.
            112,  103,   50,  107,   46,  115,  118,  114, // pg2k.svr
             67,  111,  114,  101,   46,   65,  108,  108, // Core.All
             69,  120,  104,  105,   98,  105,  116,   80, // ExhibitP
            114,  111,  112,  101,  114,  116,  105,  101, // ropertie
            115,   68,  101,  108,  116,   97,   36,   67, // sDelta$C
            104,   97,  110,  103,  101,   59,  -39,  -39, // hange;..
             14,  -36,   97,   70,  -99,  -85,    2,    0, // ..aF....
              0,  120,  112,    0,    0,    0,    0,  120, // .xp....x
           };

    /**Simple AllExhibitPropertiesDelta (de)serialisation tests.
     * While it is not promised that the serialised format for
     * AllExhibitPropertiesDelta will be stable indefinitely,
     * we want to make sure that we don't change it inadvertently,
     * which would prevent clients and servers exchanging it for example.
     */
    public static void testSerAllExhibitPropertiesDelta()
        throws Exception
        {
        final AllExhibitPropertiesDelta empty = new AllExhibitPropertiesDelta();
        checkObjectCanBeSerialisedAndDeserialised(empty);

        // Check that we can deserialise a frozen instance.
        final AllExhibitPropertiesDelta aepd1 = (AllExhibitPropertiesDelta)
            deserialiseFromByteArray(serData_AllExhibitPropertiesDelta_empty);
        // Test sample field(s).
        assertEquals("Fields must have correct value", 0, aepd1.hashNotChangedSinceAfter);
        assertEquals("Fields must have correct value", 0, aepd1.lengthAEIDBefore);
        assertEquals("Fields must have correct value", 0, aepd1.lengthAEIDAfter);
        }

    /**Sample input String values for Compact7BitString tests. 
     * These values include hard/edge cases and realistic sample data.
     * <p>
     * These are tested individually and all together in one stream.
     */
    private static final String[] TEST_C7BS_DATA =
        {
        "prefix1 prefix1suffix prefix2 prefix2suffix prefix2suffix" /* Frozen in serialsed test cases.  Hard case for partial token matching. */,
        "" /* The empty string is one edge case. */,
        "\0" /* A single null/control character is potentially tricky. */,
        "abc abc" /* Short text with one repeat. */,
        "abc abc abc abc  abc  abc" /* Short text with some repeats. */,
        "Repeated internal    sequences\r\n\r\nof various sorts.\r\nBanana, banadana?\r\n------\r\n              0,    0,    0,    0,    0,    0,    0,    0, // ........",
        "compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression compression" /* Highly compressable text. */,
        "Some simple English text.",
        "Some simple English text\r\ncontaining line-\r\nbreaks and\t\ttabs but no easy repeats.",
        "<sample><tag name=\"X Resolution\" value=\"72 dots per inch\"/><tag name=\"t2\" value=\"3.141\"/><tag name=\"Y Resolution\" value=\"72 dots per inch\"/><tag name=\"Exposure Program\" value=\"Program normal\"></sample>",
        "<sample><tag name=\"X Resolution\" value=\"72 dots per inch\"/><tag name=\"t2\" value=\"2.1828\"/><tag name=\"Y Resolution\" value=\"72 dots per inch\"/><tag name=\"Exposure Program\" value=\"Program normal\"></sample>",
        "<p class=MsoNormal><span lang=EN-US style='mso-ansi-language:EN-US'>Simp<o:p></o:p></span></p><p class=MsoNormal><span style='font-family:Arial'>Asia Pac &#20122; &#22826; &#21306;<o:p></o:p></span></p><p class=MsoNormal><span style='font-family:Arial'><![if !supportEmptyParas]>&nbsp;<![endif]><o:p></o:p></span></p>",
        "<h2>Metadata</h2><dl compact><dt><b>metadata</b><dd><dl compact><dt><b>image</b><dd><dl compact><dt><b>javax_imageio_1.0</b><dd><dl compact><dt><b>Chroma</b><dd><dl compact><dt><b>ColorSpaceType</b> name=\"YCbCr\"</dl><dl compact><dt><b>NumChannels</b> value=\"3\"</dl></dl><dl compact><dt><b>Compression</b><dd><dl compact><dt><b>CompressionTypeName</b> value=\"JPEG\"</dl><dl compact><dt><b>Lossless</b> value=\"false\"</dl><dl compact><dt><b>NumProgressiveScans</b> value=\"1\"</dl></dl><dl compact><dt><b>Dimension</b><dd><dl compact><dt><b>ImageOrientation</b> value=\"normal\"</dl></dl></dl></dl><dl compact><dt><b>native</b><dd><dl compact><dt><b>EXIF</b><dd><dl compact><dt><b>tag</b> name=\"Make\" value=\"SONY\"</dl><dl compact><dt><b>tag</b> name=\"Model\" value=\"DSC-F828\"</dl><dl compact><dt><b>tag</b> name=\"Orientation\" value=\"top, left side\"</dl><dl compact><dt><b>tag</b> name=\"X Resolution\" value=\"72 dots per inch\"</dl><dl compact><dt><b>tag</b> name=\"Y Resolution\" value=\"72 dots per inch\"</dl><dl compact><dt><b>tag</b> name=\"Resolution Unit\" value=\"Inch\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time\" value=\"2005:11:22 13:35:58\"</dl><dl compact><dt><b>tag</b> name=\"YCbCr Positioning\" value=\"Datum point\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Time\" value=\"1/60 sec\"</dl><dl compact><dt><b>tag</b> name=\"F-Number\" value=\"F3.2\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Program\" value=\"Program normal\"</dl><dl compact><dt><b>tag</b> name=\"ISO Speed Ratings\" value=\"64\"</dl><dl compact><dt><b>tag</b> name=\"Exif Version\" value=\"2.20\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time Original\" value=\"2005:11:22 13:35:58\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time Digitized\" value=\"2005:11:22 13:35:58\"</dl><dl compact><dt><b>tag</b> name=\"Components Configuration\" value=\"YCbCr\"</dl><dl compact><dt><b>tag</b> name=\"Compressed Bits Per Pixel\" value=\"8 bits/pixel\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Bias Value\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Max Aperture Value\" value=\"F2\"</dl><dl compact><dt><b>tag</b> name=\"Metering Mode\" value=\"Center weighted average\"</dl><dl compact><dt><b>tag</b> name=\"Light Source\" value=\"Unknown\"</dl><dl compact><dt><b>tag</b> name=\"Flash\" value=\"Unknown (16)\"</dl><dl compact><dt><b>tag</b> name=\"Focal Length\" value=\"9.5 mm\"</dl><dl compact><dt><b>tag</b> name=\"FlashPix Version\" value=\"1.00\"</dl><dl compact><dt><b>tag</b> name=\"Color Space\" value=\"sRGB\"</dl><dl compact><dt><b>tag</b> name=\"Exif Image Width\" value=\"3264 pixels\"</dl><dl compact><dt><b>tag</b> name=\"Exif Image Height\" value=\"2448 pixels\"</dl><dl compact><dt><b>tag</b> name=\"File Source\" value=\"Digital Still Camera (DSC)\"</dl><dl compact><dt><b>tag</b> name=\"Scene Type\" value=\"Directly photographed image\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa401)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa402)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa403)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa406)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa408)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa409)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xa40a)\" value=\"0\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xc4a5)\" value=\"PrintIM\"</dl><dl compact><dt><b>tag</b> name=\"Compression\" value=\"JPEG compression\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Offset\" value=\"2486 bytes\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Length\" value=\"9657 bytes\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Data\" value=\"[9657 bytes of thumbnail data]\"</dl></dl></dl></dl><dl compact><dt><b>accessionData</b><dd><dl compact><dt><b>date</b> value=\"1132678851000\"</dl><dl compact><dt><b>size</b> value=\"3464887\"</dl><dl compact><dt><b>hash-CRC32</b> value=\"e697175\"</dl><dl compact><dt><b>hash-MD5</b> value=\"2fc4e2c7fd8ea3047d6bf9ae22df8f44\"</dl></dl>",
        "The Indian way apparently takes years of practice, whereas for the Irish version one may only need strong booze... Arrived over the Net by email.",
        "Even the drawstrings [draw-strings] are quite interesting in close-up [closeup].",
        "Hac&#305; Bekta&#351;-&#305 Veli is a famous 13th century Islamic mystic, philosopher and dervish.&nbsp;&nbsp; His tomb and monestary are in the town of Hacibekta&#351;, which is named after him.&nbsp;&nbsp; Hac&#305; Bekta&#351;-&#305 Veli was born and educated in Khorasan (north-east Iran); he was the grandson of Sheik Ahmet.&nbsp;&nbsp; After his pilgrimage to Mecca, he travelled around the middle east, finally coming to Anatolia in a period of political and economic turmoil.&nbsp;&nbsp; Wishing to help restore unity he settled here and developed and taught his own approach to being a pure Moslem.&nbsp;&nbsp; This approach was based on four tenets: love of God (the only way to attain maturity and peace); focus on substance rather than on the superficial (e.g. God does not see your actions, he sees your intentions); love and unity among people (the only things that illuminate your way are devine and human love); and the power of sanctity (you should seek God within, not externally).&nbsp;&nbsp; He particularly taught tolerance and equality for all through love, and at it's heart his philosophy incorporates the same substance as the Universal Declaration of Human Rights of 1948.&nbsp;&nbsp; Through the work of Hac&#305; Bekta&#351;-&#305 Veli unity was restored to the Turks of Anatolia, and the Turkish language became the language of scholarship and law (instead of Arabic or Persian).<BR> The interior of Hac&#305; Bekta&#351;-&#305 Veli's tomb chamber is an interesting shape.&nbsp;&nbsp; At the bottom the room is square; higher up it is octagonal; and at the top there is a circular dome.&nbsp;&nbsp; This is essentially the same shape as the Pharos, the ancient lighthouse at Alexandria, Egypt.&nbsp;&nbsp; I've seen this shape in other places too, but I don't know whether the Pharos was the original use of this shape, or whether the shape is older than that.&nbsp;&nbsp; I'm certain that there must be some meaning behind the shape, but I'm afraid I don't know it.",
        "<image><javax_imageio_1.0><Chroma><ColorSpaceType name=\"RGB\"/><NumChannels value=\"4\"/><Gamma value=\"0.45453998\"/><BlackIsZero value=\"true\"/><Palette><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"0\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"1\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"2\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"3\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"4\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"5\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"6\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"7\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"8\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"9\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"10\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"11\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"12\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"13\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"14\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"15\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"16\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"17\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"18\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"19\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"20\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"21\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"22\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"23\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"24\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"25\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"26\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"27\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"28\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"29\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"30\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"31\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"32\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"33\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"34\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"35\" red=\"255\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"36\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"37\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"38\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"39\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"40\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"41\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"42\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"43\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"44\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"45\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"46\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"47\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"48\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"49\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"50\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"51\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"52\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"53\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"54\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"55\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"56\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"57\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"58\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"59\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"60\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"61\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"62\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"63\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"64\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"65\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"66\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"67\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"68\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"69\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"70\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"71\" red=\"204\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"72\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"73\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"74\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"75\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"76\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"77\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"78\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"79\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"80\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"81\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"82\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"83\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"84\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"85\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"86\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"87\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"88\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"89\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"90\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"91\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"92\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"93\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"94\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"95\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"96\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"97\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"98\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"99\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"100\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"101\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"102\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"103\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"104\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"105\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"106\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"107\" red=\"153\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"108\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"109\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"110\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"111\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"112\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"113\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"114\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"115\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"116\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"117\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"118\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"119\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"120\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"121\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"122\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"123\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"124\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"125\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"126\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"127\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"128\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"129\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"130\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"131\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"132\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"133\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"134\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"135\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"136\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"137\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"138\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"139\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"140\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"141\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"142\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"143\" red=\"102\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"144\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"145\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"146\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"147\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"148\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"149\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"150\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"151\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"152\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"153\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"154\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"155\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"156\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"157\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"158\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"159\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"160\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"161\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"162\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"163\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"164\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"165\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"166\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"167\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"168\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"169\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"170\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"171\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"172\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"173\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"174\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"175\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"176\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"177\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"178\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"179\" red=\"51\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"255\" index=\"180\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"255\" index=\"181\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"255\" index=\"182\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"255\" index=\"183\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"255\" index=\"184\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"255\" index=\"185\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"204\" index=\"186\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"204\" index=\"187\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"204\" index=\"188\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"204\" index=\"189\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"204\" index=\"190\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"204\" index=\"191\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"153\" index=\"192\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"153\" index=\"193\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"153\" index=\"194\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"153\" index=\"195\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"153\" index=\"196\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"153\" index=\"197\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"102\" index=\"198\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"102\" index=\"199\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"102\" index=\"200\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"102\" index=\"201\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"102\" index=\"202\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"102\" index=\"203\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"51\" index=\"204\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"51\" index=\"205\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"51\" index=\"206\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"51\" index=\"207\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"51\" index=\"208\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"51\" index=\"209\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"0\" index=\"210\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"204\" green=\"0\" index=\"211\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"153\" green=\"0\" index=\"212\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"102\" green=\"0\" index=\"213\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"51\" green=\"0\" index=\"214\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"0\" green=\"0\" index=\"215\" red=\"0\"/><PaletteEntry alpha=\"255\" blue=\"255\" green=\"254\" index=\"216\" red=\"254\"/><PaletteEntry alpha=\"255\" blue=\"7\" green=\"6\" index=\"217\" red=\"6\"/><PaletteEntry alpha=\"255\" blue=\"1\" green=\"6\" index=\"218\" red=\"6\"/><PaletteEntry alpha=\"255\" blue=\"1\" green=\"14\" index=\"219\" red=\"2\"/><PaletteEntry alpha=\"255\" blue=\"46\" green=\"15\" index=\"220\" red=\"46\"/><PaletteEntry alpha=\"255\" blue=\"201\" green=\"184\" index=\"221\" red=\"201\"/><PaletteEntry alpha=\"255\" blue=\"72\" green=\"51\" index=\"222\" red=\"69\"/><PaletteEntry alpha=\"255\" blue=\"232\" green=\"219\" index=\"223\" red=\"221\"/><PaletteEntry alpha=\"255\" blue=\"78\" green=\"16\" index=\"224\" red=\"11\"/><PaletteEntry alpha=\"255\" blue=\"46\" green=\"13\" index=\"225\" red=\"10\"/><PaletteEntry alpha=\"255\" blue=\"78\" green=\"39\" index=\"226\" red=\"36\"/><PaletteEntry alpha=\"255\" blue=\"91\" green=\"81\" index=\"227\" red=\"79\"/><PaletteEntry alpha=\"255\" blue=\"201\" green=\"183\" index=\"228\" red=\"179\"/><PaletteEntry alpha=\"255\" blue=\"24\" green=\"13\" index=\"229\" red=\"10\"/><PaletteEntry alpha=\"255\" blue=\"127\" green=\"83\" index=\"230\" red=\"71\"/><PaletteEntry alpha=\"255\" blue=\"217\" green=\"210\" index=\"231\" red=\"208\"/><PaletteEntry alpha=\"255\" blue=\"141\" green=\"136\" index=\"232\" red=\"132\"/><PaletteEntry alpha=\"255\" blue=\"178\" green=\"171\" index=\"233\" red=\"165\"/><PaletteEntry alpha=\"255\" blue=\"248\" green=\"247\" index=\"234\" red=\"246\"/><PaletteEntry alpha=\"255\" blue=\"173\" green=\"153\" index=\"235\" red=\"132\"/><PaletteEntry alpha=\"255\" blue=\"92\" green=\"70\" index=\"236\" red=\"33\"/><PaletteEntry alpha=\"255\" blue=\"207\" green=\"200\" index=\"237\" red=\"184\"/><PaletteEntry alpha=\"255\" blue=\"47\" green=\"43\" index=\"238\" red=\"14\"/><PaletteEntry alpha=\"255\" blue=\"42\" green=\"74\" index=\"239\" red=\"38\"/><PaletteEntry alpha=\"255\" blue=\"13\" green=\"42\" index=\"240\" red=\"12\"/><PaletteEntry alpha=\"255\" blue=\"8\" green=\"24\" index=\"241\" red=\"9\"/><PaletteEntry alpha=\"255\" blue=\"113\" green=\"129\" index=\"242\" red=\"125\"/><PaletteEntry alpha=\"255\" blue=\"184\" green=\"201\" index=\"243\" red=\"200\"/><PaletteEntry alpha=\"255\" blue=\"15\" green=\"45\" index=\"244\" red=\"45\"/><PaletteEntry alpha=\"255\" blue=\"38\" green=\"69\" index=\"245\" red=\"82\"/><PaletteEntry alpha=\"255\" blue=\"117\" green=\"151\" index=\"246\" red=\"175\"/><PaletteEntry alpha=\"255\" blue=\"148\" green=\"162\" index=\"247\" red=\"172\"/><PaletteEntry alpha=\"255\" blue=\"178\" green=\"181\" index=\"248\" red=\"184\"/><PaletteEntry alpha=\"255\" blue=\"19\" green=\"29\" index=\"249\" red=\"77\"/><PaletteEntry alpha=\"255\" blue=\"181\" green=\"184\" index=\"250\" red=\"201\"/><PaletteEntry alpha=\"255\" blue=\"8\" green=\"9\" index=\"251\" red=\"229\"/><PaletteEntry alpha=\"255\" blue=\"11\" green=\"11\" index=\"252\" red=\"45\"/><PaletteEntry alpha=\"255\" blue=\"8\" green=\"8\" index=\"253\" red=\"24\"/><PaletteEntry alpha=\"255\" blue=\"8\" green=\"8\" index=\"254\" red=\"8\"/><PaletteEntry alpha=\"0\" blue=\"255\" green=\"255\" index=\"255\" red=\"255\"/></Palette></Chroma><Compression><CompressionTypeName value=\"deflate\"/><Lossless value=\"true\"/><NumProgressiveScans value=\"1\"/></Compression><Data><PlanarConfiguration value=\"PixelInterleaved\"/><SampleFormat value=\"Index\"/><BitsPerSample value=\"8 8 8 8\"/></Data><Dimension><PixelAspectRatio value=\"1.0\"/><ImageOrientation value=\"Normal\"/><HorizontalPixelSize value=\"0.35273367\"/><VerticalPixelSize value=\"0.35273367\"/></Dimension><Transparency><Alpha value=\"nonpremultipled\"/><TransparentColor/></Transparency></javax_imageio_1.0></image>",
        "<image><javax_imageio_1.0><Chroma><ColorSpaceType name=\"YCbCr\"/><NumChannels value=\"3\"/></Chroma><Compression><CompressionTypeName value=\"JPEG\"/><Lossless value=\"false\"/><NumProgressiveScans value=\"1\"/></Compression><Dimension><PixelAspectRatio value=\"1.0\"/><ImageOrientation value=\"normal\"/><HorizontalPixelSize value=\"0.254\"/><VerticalPixelSize value=\"0.254\"/></Dimension></javax_imageio_1.0></image>",
        "<image><javax_imageio_1.0><Chroma><ColorSpaceType name=\"YCbCr\"/><NumChannels value=\"3\"/></Chroma><Compression><CompressionTypeName value=\"JPEG\"/><Lossless value=\"false\"/><NumProgressiveScans value=\"1\"/></Compression><Dimension><ImageOrientation value=\"normal\"/></Dimension></javax_imageio_1.0></image><native><EXIF><tag name=\"Color Space\" value=\"sRGB\"/><tag name=\"Components Configuration\" value=\"YCbCr\"/><tag name=\"Compressed Bits Per Pixel\" value=\"8 bits/pixel\"/><tag name=\"Compression\" value=\"JPEG (old-style)\"/><tag name=\"Contrast\" value=\"None\"/><tag name=\"Custom Rendered\" value=\"Normal process\"/><tag name=\"Date/Time\" value=\"2005:04:18 17:19:32\"/><tag name=\"Date/Time Digitized\" value=\"2005:04:18 17:19:32\"/><tag name=\"Date/Time Original\" value=\"2005:04:18 17:19:32\"/><tag name=\"Exif Image Height\" value=\"480 pixels\"/><tag name=\"Exif Image Width\" value=\"640 pixels\"/><tag name=\"Exif Version\" value=\"2.20\"/><tag name=\"Exposure Bias Value\" value=\"0 EV\"/><tag name=\"Exposure Mode\" value=\"Auto exposure\"/><tag name=\"Exposure Program\" value=\"Program normal\"/><tag name=\"Exposure Time\" value=\"1/60 sec\"/><tag name=\"F-Number\" value=\"F4\"/><tag name=\"File Source\" value=\"Digital Still Camera (DSC)\"/><tag name=\"Flash\" value=\"Flash did not fire, auto\"/><tag name=\"FlashPix Version\" value=\"1.00\"/><tag name=\"Focal Length\" value=\"7.1 mm\"/><tag name=\"ISO Speed Ratings\" value=\"64\"/><tag name=\"Light Source\" value=\"Unknown\"/><tag name=\"Make\" value=\"SONY\"/><tag name=\"Max Aperture Value\" value=\"F2\"/><tag name=\"Metering Mode\" value=\"Center weighted average\"/><tag name=\"Model\" value=\"DSC-F828\"/><tag name=\"Orientation\" value=\"Top, left side (Horizontal / normal)\"/><tag name=\"Resolution Unit\" value=\"Inch\"/><tag name=\"Saturation\" value=\"None\"/><tag name=\"Scene Capture Type\" value=\"Standard\"/><tag name=\"Scene Type\" value=\"Directly photographed image\"/><tag name=\"Sharpness\" value=\"None\"/><tag name=\"Thumbnail Data\" value=\"[3071 bytes of thumbnail data]\"/><tag name=\"Thumbnail Length\" value=\"3071 bytes\"/><tag name=\"Thumbnail Offset\" value=\"2486 bytes\"/><tag name=\"Unknown tag (0xc4a5)\" value=\"80 114 105 110 116 73 77 0 48 50 53 48 0 0 2 0 2 0 1 0 0 0 1 1 0 0 0 0\"/><tag name=\"White Balance\" value=\"Auto white balance\"/><tag name=\"X Resolution\" value=\"72 dots per inch\"/><tag name=\"Y Resolution\" value=\"72 dots per inch\"/><tag name=\"YCbCr Positioning\" value=\"Datum point\"/></EXIF></native>",
        "<image><javax_imageio_1.0><Chroma><ColorSpaceType name=\"YCbCr\"/><NumChannels value=\"3\"/></Chroma><Compression><CompressionTypeName value=\"JPEG\"/><Lossless value=\"false\"/><NumProgressiveScans value=\"1\"/></Compression><Dimension><ImageOrientation value=\"normal\"/></Dimension></javax_imageio_1.0></image><native><EXIF><tag name=\"Color Space\" value=\"sRGB\"/><tag name=\"Components Configuration\" value=\"YCbCr\"/><tag name=\"Compression\" value=\"JPEG (old-style)\"/><tag name=\"Contrast\" value=\"None\"/><tag name=\"Custom Rendered\" value=\"Normal process\"/><tag name=\"Date/Time\" value=\"0000:00:00 00:00:00\"/><tag name=\"Date/Time Digitized\" value=\"0000:00:00 00:00:00\"/><tag name=\"Date/Time Original\" value=\"0000:00:00 00:00:00\"/><tag name=\"Digital Zoom Ratio\" value=\"1\"/><tag name=\"Exif Image Height\" value=\"1920 pixels\"/><tag name=\"Exif Image Width\" value=\"2560 pixels\"/><tag name=\"Exif Version\" value=\"2.20\"/><tag name=\"Exposure Bias Value\" value=\"0.3 EV\"/><tag name=\"Exposure Mode\" value=\"Manual exposure\"/><tag name=\"Exposure Program\" value=\"Program normal\"/><tag name=\"Exposure Time\" value=\"1/60 sec\"/><tag name=\"F-Number\" value=\"F3.2\"/><tag name=\"File Source\" value=\"Digital Still Camera (DSC)\"/><tag name=\"Flash\" value=\"Flash did not fire, auto\"/><tag name=\"FlashPix Version\" value=\"1.00\"/><tag name=\"Focal Length\" value=\"9.42 mm\"/><tag name=\"Gain Control\" value=\"Low gain down\"/><tag name=\"ISO Speed Ratings\" value=\"320\"/><tag name=\"Image Description\" value=\"OLYMPUS DIGITAL CAMERA         \"/><tag name=\"Light Source\" value=\"Unknown\"/><tag name=\"Make\" value=\"OLYMPUS OPTICAL CO.,LTD\"/><tag name=\"Max Aperture Value\" value=\"F2.8\"/><tag name=\"Metering Mode\" value=\"Multi-segment\"/><tag name=\"Model\" value=\"X-2,C-50Z       \"/><tag name=\"Orientation\" value=\"Top, left side (Horizontal / normal)\"/><tag name=\"Resolution Unit\" value=\"Inch\"/><tag name=\"Saturation\" value=\"None\"/><tag name=\"Scene Capture Type\" value=\"Standard\"/><tag name=\"Sharpness\" value=\"None\"/><tag name=\"Software\" value=\"28-1008                        \"/><tag name=\"Thumbnail Data\" value=\"[6228 bytes of thumbnail data]\"/><tag name=\"Thumbnail Length\" value=\"6228 bytes\"/><tag name=\"Thumbnail Offset\" value=\"7206 bytes\"/><tag name=\"Unknown tag (0xc4a5)\" value=\"80 114 105 110 116 73 77 0 48 50 53 48 0 0 20 0 1 0 20 0 20 0 2 0 1 0 0 0 3 0 -120 0 0 0 7 0 0 0 0 0 8 0 0 0 0 0 9 0 0 0 0 0 10 0 0 0 0 0 11 0 -48 0 0 0 12 0 0 0 0 0 13 0 0 0 0 0 14 0 -24 0 0 0 0 1 1 0 0 0 1 1 -1 0 0 0 2 1 -125 0 0 0 3 1 -128 0 0 0 4 1 -125 0 0 0 5 1 -125 0 0 0 6 1 -125 0 0 0 7 1 -128 -128 -128 0 16 1 -126 0 0 0 9 17 0 0 16 39 0 0 11 15 0 0 16 39 0 0 -105 5 0 0 16 39 0 0 -80 8 0 0 16 39 0 0 1 28 0 0 16 39 0 0 94 2 0 0 16 39 0 0 -117 0 0 0 16 39 0 0 -53 3 0 0 16 39 0 0 -27 27 0 0 16 39 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\"/><tag name=\"White Balance\" value=\"Auto white balance\"/><tag name=\"X Resolution\" value=\"72 dots per inch\"/><tag name=\"Y Resolution\" value=\"72 dots per inch\"/><tag name=\"YCbCr Positioning\" value=\"Datum point\"/></EXIF></native>",
        "I.A.N.A. is as I.A.N.A. does..." /* Repeated 1-char token sequences. */,
        "<dl compact><dt><b>metadata</b><dd><dl compact><dt><b>image</b><dd><dl compact><dt><b>javax_imageio_1.0</b><dd><dl compact><dt><b>Chroma</b><dd><dl compact><dt><b>ColorSpaceType</b> name=\"YCbCr\"</dl><dl compact><dt><b>NumChannels</b> value=\"3\"</dl></dl><dl compact><dt><b>Compression</b><dd><dl compact><dt><b>CompressionTypeName</b> value=\"JPEG\"</dl><dl compact><dt><b>Lossless</b> value=\"false\"</dl><dl compact><dt><b>NumProgressiveScans</b> value=\"1\"</dl></dl><dl compact><dt><b>Dimension</b><dd><dl compact><dt><b>PixelAspectRatio</b> value=\"1.0\"</dl><dl compact><dt><b>ImageOrientation</b> value=\"normal\"</dl><dl compact><dt><b>HorizontalPixelSize</b> value=\"0.26458332\"</dl><dl compact><dt><b>VerticalPixelSize</b> value=\"0.26458332\"</dl></dl></dl></dl><dl compact><dt><b>native</b><dd><dl compact><dt><b>EXIF</b><dd><dl compact><dt><b>tag</b> name=\"Color Space\" value=\"sRGB\"</dl><dl compact><dt><b>tag</b> name=\"Components Configuration\" value=\"YCbCr\"</dl><dl compact><dt><b>tag</b> name=\"Compressed Bits Per Pixel\" value=\"4 bits/pixel\"</dl><dl compact><dt><b>tag</b> name=\"Compression\" value=\"JPEG (old-style)\"</dl><dl compact><dt><b>tag</b> name=\"Custom Rendered\" value=\"Normal process\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time\" value=\"2006:01:15 15:38:56\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time Digitized\" value=\"2006:01:15 15:38:56\"</dl><dl compact><dt><b>tag</b> name=\"Date/Time Original\" value=\"2006:01:15 15:38:56\"</dl><dl compact><dt><b>tag</b> name=\"Exif Image Height\" value=\"960 pixels\"</dl><dl compact><dt><b>tag</b> name=\"Exif Image Width\" value=\"1280 pixels\"</dl><dl compact><dt><b>tag</b> name=\"Exif Version\" value=\"2.20\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Bias Value\" value=\"0 EV\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Mode\" value=\"Auto exposure\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Program\" value=\"Program normal\"</dl><dl compact><dt><b>tag</b> name=\"Exposure Time\" value=\"1/200 sec\"</dl><dl compact><dt><b>tag</b> name=\"F-Number\" value=\"F5.2\"</dl><dl compact><dt><b>tag</b> name=\"File Source\" value=\"Digital Still Camera (DSC)\"</dl><dl compact><dt><b>tag</b> name=\"Flash\" value=\"Flash did not fire\"</dl><dl compact><dt><b>tag</b> name=\"FlashPix Version\" value=\"1.00\"</dl><dl compact><dt><b>tag</b> name=\"Focal Length\" value=\"23.7 mm\"</dl><dl compact><dt><b>tag</b> name=\"ISO Speed Ratings\" value=\"100\"</dl><dl compact><dt><b>tag</b> name=\"Light Source\" value=\"Unknown\"</dl><dl compact><dt><b>tag</b> name=\"Make\" value=\"SONY\"</dl><dl compact><dt><b>tag</b> name=\"Max Aperture Value\" value=\"F2.8\"</dl><dl compact><dt><b>tag</b> name=\"Metering Mode\" value=\"Multi-segment\"</dl><dl compact><dt><b>tag</b> name=\"Model\" value=\"DSC-P10\"</dl><dl compact><dt><b>tag</b> name=\"Orientation\" value=\"Top, left side (Horizontal / normal)\"</dl><dl compact><dt><b>tag</b> name=\"Resolution Unit\" value=\"Inch\"</dl><dl compact><dt><b>tag</b> name=\"Scene Capture Type\" value=\"Standard\"</dl><dl compact><dt><b>tag</b> name=\"Scene Type\" value=\"Directly photographed image\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Data\" value=\"[4198 bytes of thumbnail data]\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Length\" value=\"4198 bytes\"</dl><dl compact><dt><b>tag</b> name=\"Thumbnail Offset\" value=\"2342 bytes\"</dl><dl compact><dt><b>tag</b> name=\"Unknown tag (0xc4a5)\" value=\"80 114 105 110 116 73 77 0 48 50 53 48 0 0 2 0 2 0 1 0 0 0 1 1 0 0 0 0\"</dl><dl compact><dt><b>tag</b> name=\"White Balance\" value=\"Auto white balance\"</dl><dl compact><dt><b>tag</b> name=\"X Resolution\" value=\"72 dots per inch\"</dl><dl compact><dt><b>tag</b> name=\"Y Resolution\" value=\"72 dots per inch\"</dl><dl compact><dt><b>tag</b> name=\"YCbCr Positioning\" value=\"Datum point\"</dl></dl></dl></dl><dl compact><dt><b>accessionData</b><dd><dl compact><dt><b>date</b> value=\"1155636545000\"</dl><dl compact><dt><b>size</b> value=\"122672\"</dl><dl compact><dt><b>hash-CRC32</b> value=\"6cb4f395\"</dl><dl compact><dt><b>hash-MD5</b> value=\"1a8a965611c8171dca007b13937cd0d8\"</dl></dl>" /* Example HTML form of metadata. */ ,
        "<table cellpadding=0 border=1 cols=6 width=396><tr height=64 align=center><td width=64><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-1-BG.jpg.html\"><img src=\"/_tn/sml/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-1-BG.jpg\" border=0 width=43 height=64 alt=\"Norway ~ -1-BG.jpg\"></a></td><td width=64><small>JPEG image<br /><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-tweaked-1-BG.jpg.html\"><i>Norway ~  tweaked 1 BG.jpg</i></a></small></td><td width=64><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-2-BG.jpg.html\"><img src=\"/_tn/sml/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-2-BG.jpg\" border=0 width=64 height=43 alt=\"Norway ~ -2-BG.jpg\"></a></td><td width=64><small>JPEG image<br /><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-tweaked-2-BG.jpg.html\"><i>Norway ~  tweaked 2 BG.jpg</i></a></small></td><td width=64><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-3-BG.jpg.html\"><img src=\"/_tn/sml/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-3-BG.jpg\" border=0 width=43 height=64 alt=\"Norway ~ -3-BG.jpg\"></a></td><td width=64><a href=\"/_c/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-tweaked-3-BG.jpg.html\"><img src=\"/_tn/sml/places-and-sights/_more2002/_more04/Norway-Flam-Myrdal-railway-Flamsbana-Kjosfossen-waterfall-tweaked-3-BG.jpg\" border=0 width=43 height=64 alt=\"Norway ~ -tweaked-3-BG.jpg\"></a></td></tr></table>",
        "<table cellpadding=0 border=1 cols=10 width=660><tr height=64 align=center><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-1-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-1-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -1-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-2-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-2-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -2-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-3-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-3-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -3-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-4-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-4-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -4-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-5-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-5-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -5-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-6-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-6-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -6-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-7-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-7-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -7-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-8-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-8-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -8-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-9-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-9-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -9-DHD.jpg\"></a></td><td width=64><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-10-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-10-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -10-DHD.jpg\"></a></td></tr><tr height=64 align=center><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-11-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-11-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -11-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-12-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-12-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -12-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-13-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-13-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -13-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-14-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-14-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -14-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-15-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-15-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -15-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-16-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-16-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -16-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-17-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-17-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -17-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-18-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-18-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -18-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-19-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-19-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -19-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-20-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-20-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -20-DHD.jpg\"></a></td></tr><tr height=64 align=center><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-21-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-21-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -21-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-22-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-22-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -22-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-23-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-23-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -23-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-24-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-24-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -24-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-25-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-25-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -25-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-26-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-26-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -26-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-27-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-27-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -27-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-28-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-28-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -28-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-29-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-29-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -29-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-30-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-30-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -30-DHD.jpg\"></a></td></tr><tr height=64 align=center><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-31-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-31-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -31-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-32-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-32-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -32-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-33-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-33-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -33-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-34-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-34-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -34-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-35-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-35-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -35-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-36-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-36-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -36-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-37-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-37-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -37-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-38-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-38-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -38-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-39-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-39-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -39-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-40-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-40-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -40-DHD.jpg\"></a></td></tr><tr height=64 align=center><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-41-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-41-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -41-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-42-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-42-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -42-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-43-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-43-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -43-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-44-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-44-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -44-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-45-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-45-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -45-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-46-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-46-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -46-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-47-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-47-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -47-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-48-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-48-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -48-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-49-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-49-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -49-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-50-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-50-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -50-DHD.jpg\"></a></td></tr><tr height=64 align=center><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-51-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-51-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -51-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-52-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-52-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -52-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-53-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-53-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -53-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-54-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-54-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -54-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-55-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-55-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -55-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-56-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-56-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -56-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-57-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-57-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -57-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-58-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-58-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -58-DHD.jpg\"></a></td><td><a href=\"/_c/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-59-DHD.jpg.html\"><img src=\"/_tn/sml/money/_more2004/_more12/sale-sales-posters-in-shop-windows-in-Kingston-upon-Thames-just-after-Christmas-2004-after-a-poor-retail-month-59-DHD.jpg\" border=0 width=64 height=48 alt=\"sale ~ -59-DHD.jpg\"></a></td></tr></table>",
        };

    /**Forced serialised form to check deserialisation is not broken.
     * Encoded value: "prefix1 prefix1suffix prefix2 prefix2suffix prefix2suffix".
     */
    private static final byte serData_C7BS_20060803[/*109*/] = {
        -84,  -19,    0,    5,  115,  114,    0,   39, // ....sr.'
        111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
         46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
        114,   67,  111,  114,  101,   46,   67,  111, // rCore.Co
        109,  112,   97,   99,  116,   55,   66,  105, // mpact7Bi
        116,   83,  116,  114,  105,  110,  103,  -81, // tString.
          2,   23,   50,  -72, -120,   34,   44,    3, // ..2..",.
          0,    0,  120,  112,  119,   46,   45,  112, // ..xpw.-p
        114,  101,  102,  105,  120,   49,   32,  112, // refix1.p
        114,  101,  102,  105,  120,   49,  115,  117, // refix1su
        102,  102,  105,  120,   32,  112,  114,  101, // ffix.pre
        102,  105,  120,   50,   32,  112,  114,  101, // fix2.pre
        102,  105,  120,   50,  115,  117,  102,  102, // fix2suff
        105,  120,   32,   -2,  120,                   // ix..x
       };

    /**Forced serialised form after allowing for partial token (prefix) matching.
     * Encoded value: "prefix1 prefix1suffix prefix2 prefix2suffix prefix2suffix".
     */
    private static final byte serData_C7BS_20060804[/*103*/] = {
        -84,  -19,    0,    5,  115,  114,    0,   39, // ....sr.'
        111,  114,  103,   46,  104,  100,   46,  100, // org.hd.d
         46,  112,  103,   50,  107,   46,  115,  118, // .pg2k.sv
        114,   67,  111,  114,  101,   46,   67,  111, // rCore.Co
        109,  112,   97,   99,  116,   55,   66,  105, // mpact7Bi
        116,   83,  116,  114,  105,  110,  103,  -81, // tString.
          2,   23,   50,  -72, -120,   34,   44,    3, // ..2..",.
          0,    0,  120,  112,  119,   40,   39,  112, // ..xpw('p
        114,  101,  102,  105,  120,   49,   32,   -2, // refix1..
        115,  117,  102,  102,  105,  120,   32,  112, // suffix.p
        114,  101,  102,  105,  120,   50,   32,  112, // refix2.p
        114,  101,  102,  105,  120,   50,  115,  117, // refix2su
        102,  102,  105,  120,   32,   -2,  120,       // ffix..x
       };
    
    /**Test of serialisation and other behaviour of Compact7BitString.
     * The aim is that Compact7BitString should usually be significant smaller than String
     * in memory and serialised for 7-bit ASCII text,
     * especially when multiple similar values are serialised to a single stream.
     * <p>
     * We take the (maximally) compressed serialised stream to be our target usage.
     */
    public static void testCompact7BitString()
        throws Exception
        {
        // Check deserialistion of our frozen values.
        // The encoded value should be our first test text.
        final Compact7BitString frozen1 = (Compact7BitString) (new ObjectInputStream(new ByteArrayInputStream(serData_C7BS_20060803))).readObject();
        assertEquals("must be able to extract correct value from frozen instance",
                        TEST_C7BS_DATA[0], frozen1.toString());
        final Compact7BitString frozen2 = (Compact7BitString) (new ObjectInputStream(new ByteArrayInputStream(serData_C7BS_20060804))).readObject();
        assertEquals("must be able to extract correct value from frozen instance",
                        TEST_C7BS_DATA[0], frozen2.toString());

        // Use all the fixed cases in the tests.
        final ArrayList<String> testData = new ArrayList<String>(Arrays.asList(TEST_C7BS_DATA));

        // Create one nasty-but-legal block of random 7-bit ASCII values.
        // This (eventually!) covers all possible legal input values.
        final byte[] r = new byte[rnd.nextInt(Short.MAX_VALUE+1)];
        rnd.nextBytes(r);
        final StringBuffer rsb = new StringBuffer(r.length);
        for(final byte b : r) { rsb.append((char) (b & 0x7f)); }
        // Add it to the test set.
        testData.add(rsb.toString());

        // Set of static compression dictionaries that we will test against.
        final Compact7BitString.StaticDictionary dicts[] = new Compact7BitString.StaticDictionary[]{
                 ExhibitPropsComputable.sDict,
                 ExhibitPropsLoadable.sDict,
                 WebUtils.sDictMD,
            };

        final ByteArrayOutputStream allStrings = new ByteArrayOutputStream();
        final ObjectOutputStream aSoos = new ObjectOutputStream(allStrings);
        final ByteArrayOutputStream allC7BSs = new ByteArrayOutputStream();
        final ObjectOutputStream aCoos = new ObjectOutputStream(allC7BSs);

        boolean didCompress = false; // True if any input was compressed.
        boolean didBetterWithDict = false; // True if any input was better compressed against any dictionary.
        for(final String t : testData)
            {
            final Compact7BitString c7bs = Compact7BitString.convertToCompact7BitString(t, null);
            final long nt1 = System.nanoTime();
            final String asString = c7bs.toString();
            final long nt2 = System.nanoTime();
            final long nt = nt2-nt1;
            // Estimate nanoseconds per char to decompress.
            final long nspc = nt / Math.max(1, asString.length());
            assertEquals("must be able to recover String value correctly",
                            t, asString);
            assertEquals("must be able to recover String value after (de)serialisation",
                            t, checkObjectCanBeSerialisedAndDeserialised(c7bs).toString());

            // Internal format must be no more bytes than the String has chars,
            // ie half the memory space or less for most JVMs.
            final byte cd[] = c7bs.getInternalBytes();
            final int cl = ((null == cd) ? 0 : cd.length);
            Main.getOut().println("[chars => bytes (toString() ~"+nt+"ns, ~"+nspc+"ns/char): "+(t.length())+" => "+cl+": "+TextUtils.sanitiseForXML(t, 40, true)+"]");
            assertTrue("internal format must be no more bytes than the String has chars",
                            cl <= t.length());
            // We expect at least one of the sample inputs to compress further...
            if(cl < t.length()) { didCompress = true; }

            // Now compress sample input with static dictionaries we use.
            for(final Compact7BitString.StaticDictionary d : dicts)
                {
                final Compact7BitString c7bsA = Compact7BitString.convertToCompact7BitString(t, d);
                assertEquals("must be able to recover String value correctly: dictionary="+d,
                                t, c7bsA.toString());
                assertEquals("must be able to recover String value after (de)serialisation: dictionary="+d,
                                t, checkObjectCanBeSerialisedAndDeserialised(c7bsA).toString());
                final byte[] cdAlt = c7bsA.getInternalBytes();
                final int lenWithDict = (null == cdAlt) ? 0 : cdAlt.length;
                assertTrue("compression with static dictionary must be no worse than with no dictionary: dictionary="+d,
                                lenWithDict <= cl);
                // Note if we do better with this compression dictionary.
                if(lenWithDict < cl)
                    {
                    didBetterWithDict = true;
                    Main.getOut().println("  [with dictionary "+d.name+" => bytes: "+lenWithDict+".]");
                    }
                }

            aSoos.writeObject(t);
            aCoos.writeObject(c7bs);
            final Pair<CompressionLevel, byte[]> uS = GenUtils.compressObject(new Object[]{Compact7BitString.EMPTY, t}, CompressionLevel.NONE);

            final Pair<CompressionLevel, byte[]> uC7BS= GenUtils.compressObject(new Object[]{Compact7BitString.EMPTY, c7bs}, CompressionLevel.NONE);
//            Main.getOut().println("[String/C7BS adjusted uncompressed serialised size: "+uS.second.length+"/"+uC7BS.second.length+".]");
            assertTrue("uncompressed form must be similar in size",
                            uS.second.length >= uC7BS.second.length);

            // Compute moderately-compressed serialised forms...
            final Pair<CompressionLevel, byte[]> cS = GenUtils.compressObject(new Object[]{Compact7BitString.EMPTY, t}, CompressionLevel.ZIP);
            final Pair<CompressionLevel, byte[]> cC7BS = GenUtils.compressObject(new Object[]{Compact7BitString.EMPTY, c7bs}, CompressionLevel.ZIP);
//            Main.getOut().println("[String/C7BS adjusted compressed serialised size: "+cS.second.length+"/"+cC7BS.second.length+".]");
            assertTrue("compressed form must be similar in size: S/C7BS="+cS.second.length+"/"+cC7BS.second.length,
                            cS.second.length >= cC7BS.second.length);
            }
        aSoos.close();
        aCoos.close();

        assertTrue("at least one input should have compressed (fewer bytes than chars)", didCompress);

        // Serialising all the test objects as C7BS should beat serialising them as plain Strings when compressed.
        final Pair<CompressionLevel, byte[]> aScomp = GenUtils.compressObject(allStrings.toByteArray(), GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL);
        final Pair<CompressionLevel, byte[]> aCcomp = GenUtils.compressObject(allC7BSs.toByteArray(), GenUtils.MAX_SUPPORTED_COMPRESSION_LEVEL);
//        Main.getOut().println("[Serialised String/C7BS compressed sequences: "+aScomp.second.length+"/"+aCcomp.second.length+".]");
        assertTrue("serialised sequence of (similar) C7BS instances should be no larger than serialised plain Strings",
                        aScomp.second.length >=
                        aCcomp.second.length);

        assertTrue("at least one input should have compressed better with at least one of the static dictionaries!",
                        didBetterWithDict);
        }
    
    /**Test (de)serialisation of LocationMap.
     */
    public static void testLocationMap()
        throws Exception
        {
        checkEmptyInstancesAreZeroAndEqualAndSerializable(LocationMap.class);
        checkSerialisationPreservesEquality(new LocationMap());
        
        final ExhibitPropsGlobalImmutable epgi = ExhibitPropsGlobalImmutable.loadFromDataDir(new File(LocalProps.getDataDir()));
//        final LocationMap lm = epgi.getLocationMap();

        // Note size of simple ZIPped (quickly-compressed) serialised form
        // of realistic map (wrapped in EPGI).
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final GZIPOutputStream gos = new GZIPOutputStream(baos);
        final ObjectOutputStream oos = new ObjectOutputStream(gos);
        oos.writeObject(epgi);
        oos.close();
        Main.getOut().println("Compressed serialised ExhibitPropsGlobalImmutable size (bytes) ~ " + baos.size());
        }
    
    /**Test correctness of some of our (de)compressors often used with serialisation.
     * This ensures that they correctly handle some sample data streams
     * and that they work correctly in the face of a mid-stream flush().
     */
    public static void testCompressors()
        throws Exception
        {
        // We'll use a variety of existing and new test data.
        final List<byte[]> testData = new ArrayList<byte[]>(64);

        // Some synthetic edge/hard cases.
        testData.add(new byte[0]);
        testData.add(new byte[101]);
        final byte randomRubbish[] = new byte[rnd.nextInt(1<<16)];
        rnd.nextBytes(randomRubbish);
        testData.add(randomRubbish);

        // Test some of the frozen serialised data streams as realistic samples.
        testData.add(serData_AllExhibitPropertiesDelta_empty);
        testData.add(serData_C7BS_20060803);
        testData.add(serData_ROBA_compS1_20050522);
        testData.add(serData_evv1_20050131_1512);

        // Compressors to test (represented by their "levels" here).
        final CompressionLevel compressors[] =
            { CompressionLevel.BZIP2, CompressionLevel.ZIP };

        // Try all the compressors in turn...
        for(final CompressionLevel comp : compressors)
            {
            // Try each data sample in turn...
            for(final byte[] data : testData)
                {
                // Compress once with no flush()es and once with at least one.
                for(final boolean doFlush : new boolean[]{ false, true } )
                    {
                    final ByteArrayOutputStream baos1 = new ByteArrayOutputStream();
                    final OutputStream cos1;
                    switch(comp)
                        {
                        /** Note that we actually test our flushable GZIP here... */
                        case ZIP: cos1 = new FlushableGZIPOutputStream(baos1); break;
                        case BZIP2: cos1 = new CBZip2OutputStream(baos1); break;
                        default: throw new Error();
                        }
                    if(doFlush && rnd.nextBoolean()) { cos1.flush(); }
                    if(!doFlush || (data.length < 2))
                        { cos1.write(data.clone()); }
                    else
                        {
                        // Break the write into two pieces with a flush() between.
                        final int flushAfter = rnd.nextInt(data.length - 1);
                        final byte d1[] = new byte[flushAfter];
                        final byte d2[] = new byte[data.length - flushAfter];
                        System.arraycopy(data, 0, d1, 0, flushAfter);
                        System.arraycopy(data, flushAfter, d2, 0, d2.length);
                        cos1.write(d1);
                        cos1.flush();
                        cos1.write(d2);
                        }
                    if(doFlush && rnd.nextBoolean()) { cos1.flush(); }
                    cos1.close();
                    final byte compressed[] = baos1.toByteArray();
//Main.getOut().println("[Input length / method / doFlush / output length = "+data.length+" / "+comp+" / "+doFlush+" / "+compressed.length+".]");
                    
                    // Check that decompressing reproduces the original input.
                    final ByteArrayInputStream bais = new ByteArrayInputStream(compressed);
                    final InputStream dcis;
                    switch(comp)
                        {
                        case ZIP: dcis = new GZIPInputStream(bais); break;
                        case BZIP2: dcis = new CBZip2InputStream(bais); break;
                        default: throw new Error();
                        }
                    final DataInputStream dis = new DataInputStream(dcis);
                    final byte decompData[] = new byte[data.length];
                    dis.readFully(decompData);
                    assertTrue("must be able to recover data after decompression", Arrays.equals(data, decompData));
                    assertEquals("must be no trailing garbage in decompressed input", -1, dis.read());
                    }
                }
            }
        }

    /**Test behaviour of PropertiesDiff object.
     * This is primarily used to allow efficient transmission of
     * properties values/updates over the wire.
     */
    public static void testPropertiesDiff()
        throws Exception
        {
        // Test that empty object is valid.
        final PropertiesDiff emptyDiff = PropertiesDiff.EMPTY_DIFF;
        (emptyDiff).validateObject();

        // Test that an (empty) object can survive (de)serialisation intact.
        checkEmptyInstancesAreZeroAndEqualAndSerializable(PropertiesDiff.class);

        // Check that the diff from one empty Properties set to another is empty.
        final Properties emptyProps = new Properties();
        assertEquals("Diff between empty Properties sets must be empty",
            emptyDiff, PropertiesDiff.createDiff(emptyProps, emptyProps, true, rnd.nextBoolean()));

        // Check that the diff between an empty property set and a non-empty one is NOT empty.
        final Properties oneEntry = new Properties();
        oneEntry.setProperty("hello", "world");
        assertFalse("Diff between empty Properties and non-empty sets must not be empty",
            emptyDiff.equals(PropertiesDiff.createDiff(oneEntry, emptyProps, true, rnd.nextBoolean())));
        assertFalse("Diff between empty Properties and non-empty sets must not be empty",
            emptyDiff.equals(PropertiesDiff.createDiff(emptyProps, oneEntry, true, rnd.nextBoolean())));

        // Construct simple properties set with one of each pertinent type:
        //   * unchanged
        //   * deleted
        //   * new/added
        //   * changed
        final Properties p1 = new Properties();
        p1.setProperty("deleted", "byebye");
        p1.setProperty("changed", "oldvalue");
        final Properties p2 = new Properties();
        p1.setProperty("unchanged", "sameold");
        p2.setProperty("unchanged", "sameold");
        p2.setProperty("changed", "newvalue");
        p2.setProperty("new", "wotcha!");
        final PropertiesDiff diff1 = PropertiesDiff.createDiff(p1, p2, true, rnd.nextBoolean());
        assertFalse("This diff should not be empty", diff1.isEmpty());
        assertFalse("This diff should not be empty (==emptyDiff)", emptyDiff.equals(diff1));
        assertNotSame("This diff's hash should not be zero", 0, diff1.hashCode());
        final Properties synthP2 = PropertiesDiff.applyDiff(p1, diff1);
        assertEquals("Must be able to reconstruct 'p2' Properties using diff",
                        p2, synthP2);
        // Check that the diff survives (de)serialisation.
        checkSerialisationPreservesEquality(diff1);
if(PRINT_SER_BYTES) { Main.getErr().println("WARNING: stopping tests early because of volume of output..."); return; } // TOO MUCH GARBAGE PRINTED OTHERWISE...

        // Create some random-ish Properties sets, diff them, etc.
        // Generally check for sensible behaviour.
        // We have one unique key in each the before and after set
        // to help check for correct behaviour.
        final String uniqueBeforeKey = "_unique_before_";
        final String uniqueAfterKey = "_unique_after_";
        // Test (exponentially) increasing set sizes,
        // up to something reasonably representative.
        for(int i = 0; i <= 12; i += 2)
            {
            final Properties before = new Properties();
            before.setProperty(uniqueBeforeKey, "U" + rnd.nextLong());
            final Properties after = new Properties();
            after.setProperty(uniqueAfterKey, "U" + rnd.nextLong());
            for(int j = rnd.nextInt(1 + (1<<i)); --j >= 0; )
                {
                // We aim to have the "before" and "after" values very similar,
                // with a few deletions and additions/changes,
                // with the "after" value on average very slightly larger
                // to represent realistic slow growth/change of a Properties set.
                final String key = "key." + (rnd.nextLong() >>> 1);
                final String beforeValue = "val" + rnd.nextLong();
                // The "before" value is usually set.
                if(rnd.nextInt(17) != 0) { before.setProperty(key, beforeValue); }
                // The "after" value is usually the same as the before value.
                final String afterValue = (rnd.nextInt(101) != 0) ? beforeValue : "val" + rnd.nextLong();
                // The "after" value is usually set.
                if(rnd.nextInt(19) != 0) { after.setProperty(key, afterValue); }
                }

            // Simplest and most basic test.
            // Force reconstruction of "after" vaue from "before" + diff.
            final PropertiesDiff basicDiff = PropertiesDiff.createDiff(before, after, true, rnd.nextBoolean());
            assertFalse("This diff should not be empty", basicDiff.isEmpty());
            final Properties synthAfter = PropertiesDiff.applyDiff(before, basicDiff);
            assertEquals("Must be able to reconstruct 'after' Properties with correct size using diff",
                            after.size(), synthAfter.size());
            assertEquals("Must be able to reconstruct 'after' Properties with correct key set using diff",
                            after.keySet(), synthAfter.keySet());
            assertEquals("Must be able to reconstruct 'after' Properties with correct value set using diff",
                            new HashSet<Object>(after.values()), new HashSet<Object>(synthAfter.values()));
            assertEquals("Must be able to reconstruct 'after' Properties using diff",
                            after, synthAfter);

            // Force reconstruction of "after" from empty.
            final PropertiesDiff eDiff = PropertiesDiff.createDiff(emptyProps, after, true, rnd.nextBoolean());
            final Properties synthAfterE = PropertiesDiff.applyDiff(emptyProps, eDiff);
            assertEquals("Must be able to reconstruct 'after' Properties from diff from empty Properties",
                            after, synthAfterE);

            // Force reverse diff of "before" from "after".
            final PropertiesDiff rDiff = PropertiesDiff.createDiff(after, before, true, rnd.nextBoolean());
            final Properties synthBefore = PropertiesDiff.applyDiff(after, rDiff);
            assertEquals("Must be able to reconstruct 'before' Properties from (reverse) diff from 'after'",
                            before, synthBefore);

            // Check that the diff survives (de)serialisation.
            checkSerialisationPreservesEquality(basicDiff);
            checkSerialisationPreservesEquality(eDiff);
            checkSerialisationPreservesEquality(rDiff);
            }
        }

    /**Test behaviour of PropertiesBundleDiff object.
     * This is primarily used to allow efficient transmission of
     * properties values/updates over the wire.
     */
    public static void testPropertiesBundleDiff()
        throws Exception
        {
        // Test that empty object is valid.
        final PropertiesBundleDiff emptyBundleDiff = PropertiesBundleDiff.EMPTY_DIFF;
        (emptyBundleDiff).validateObject();

        // Test that an (empty) object can survive (de)serialisation intact.
        checkEmptyInstancesAreZeroAndEqualAndSerializable(PropertiesBundleDiff.class);

        // Hand-crafted small properties sets representative of small system.
        // Version 1.
        final Properties v1_treedec_properties = new Properties();
        final Properties v1_treedec_fr_properties = new Properties();
        final Properties v1_treedec_ERROR_properties = new Properties();
        final Map<String,Properties> v1 = new TreeMap<String,Properties>();
        v1.put("", v1_treedec_properties);
        v1_treedec_properties.setProperty("deleted", "byebye");
        v1_treedec_properties.setProperty("changed", "oldvalue");
        v1_treedec_properties.setProperty("unchanged", "sameold");
        v1.put("_fr", v1_treedec_fr_properties);
        v1_treedec_fr_properties.setProperty("world", "monde");
        v1.put("_ERROR", v1_treedec_ERROR_properties);

        // Version 2, after update.
        final Map<String,Properties> v2 = new TreeMap<String,Properties>();
        final Properties v2_treedec_properties = new Properties();
        final Properties v2_treedec_fr_properties = new Properties();
        final Properties v2_treedec_zh_properties = new Properties();
        v2.put("", v2_treedec_properties);
        v2_treedec_properties.setProperty("unchanged", "sameold");
        v2_treedec_properties.setProperty("changed", "newvalue");
        v2_treedec_properties.setProperty("new", "wotcha!");
        v2.put("_fr", v2_treedec_fr_properties);
        v2_treedec_fr_properties.setProperty("world", "monde");
        v2.put("_zh", v2_treedec_zh_properties);

        assertFalse("Modified properties set should not compare equal", v1.equals(v2));

        final PropertiesBundleDiff pbDiff = PropertiesBundleDiff.createDiff(v1, v2, true, rnd.nextBoolean());
        assertFalse("Bundle diff should not be empty", pbDiff.isEmpty());
        assertNotSame("Bundle diff hash should not be zero", 0, pbDiff.hashCode());
        checkSerialisationPreservesEquality(pbDiff);

        // Check that we can correctly recreate v2 from v1 + diff.
        final Map<String,Properties> v2Synth = PropertiesBundleDiff.applyDiff(v1, pbDiff);
        assertEquals("Reconstricted bundle should equal original", v2, v2Synth);
        }

    /**Basic serialisation tests of EPGI and its diff.
     */
    public void testEPGIBasic()
        throws Exception
        {
        // Check (de)serialisation of empty EPGI.
        checkObjectCanBeSerialisedAndDeserialised(new ExhibitPropsGlobalImmutable());

        // Test that an (empty) EPGL object can survive (de)serialisation intact.
        checkEmptyInstancesAreZeroAndEqualAndSerializable(ExhibitPropsGlobalImmutable.class);

        // Check (de)serialisation of realistic EPGI.
        final ExhibitPropsGlobalImmutable epgi = ExhibitPropsGlobalImmutable.loadFromDataDir(new File(LocalProps.getDataDir()));
        checkSerialisationPreservesEquality(epgi);

        // Very basic EPGIDiff tests.
        checkObjectCanBeSerialisedAndDeserialised(new ExhibitPropsGlobalImmutable.EPGIDiff());
        }

//    /**Generic simple serialisation tests.
//     * This is for very basic tests that do not warrant their own harness.
//     */
//    public void testGenericSer()
//        throws Exception
//        {
//        checkObjectCanBeSerialisedAndDeserialised(new AllExhibitPropertiesDelta());
//        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
