/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataSimpleCache;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineFilter;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.util.*;

/**Test the ExhibitSimpleDataCache.
 */
public final class SimpleCacheTest extends TestCase
    {
    public SimpleCacheTest(final String name)
        {
        super(name);
        }

    private final SimpleLoggerIF logger = new SimpleLoggerIF(){
        public void log(final String message) { Main.getOut().println(message); }
        };

    /**Temporary directory in which we build a cache; null if not needed.
     * We only chose a directory name
     * (which we store in absolute form in case we change working directory)
     * and delete any existing content before and after the tests run.
     * The test routines must create the directory if they want it.
     */
    private File tempDir;

    /**Timer that we can use to run pipelines; never null during tests. */
    private Timer t;

    /**Do any setup needed for the tests. */
    protected void setUp()
        {
        // Work in a temporary subdirectory of the current directory.
        tempDir = (new File("CacheTst.tmp")).getAbsoluteFile();

        // Make sure that we obliterate any extant content.
        final File f = tempDir;
        try { FileTools.rmRecursively(f); }
        catch(IOException e) { e.printStackTrace(Main.getErr()); }

        t = new Timer();
        }

    /**Do any clearup needed after the tests. */
    protected void tearDown()
        {
        final File f = tempDir;
        if(f != null)
            {
            try { FileTools.rmRecursively(f); }
            catch(IOException e) { e.printStackTrace(Main.getErr()); }
            }

        // Kill the timer and free resources...
        t.cancel();
        t = null;
        }

    /**Dummy data source.
     * Has one or more fake "virtual" exhibits,
     * each of which contain some synthetic virtual data.
     * <p>
     * This can be wrapped to make a sort of "mock object"
     * that can flag certain undesirable access patterns as test failures.
     * <p>
     * This returns data based on the exhibit data supplied to the constructor.
     * <p>
     * This never returns thumbnails.
     */
    public static class DummyDataSource implements SimpleExhibitPipelineIF
        {
        /**Construct a new instance with the given properties.
         * @param aep  set of exhibit properties; never null
         * @param gp  set of GenProps; may be null
         */
        public DummyDataSource(final AllExhibitProperties aep,
                               final GenProps gp)
            {
            if(aep == null)
                { throw new IllegalArgumentException(); }

            this.aep = aep;
            this.gp = gp;
            }

        /**Construct a new instance with empty properties. */
        public DummyDataSource()
            { this(new AllExhibitProperties(), new GenProps()); }

        /**The set of exhibit properties; never null. */
        private final AllExhibitProperties aep;

        /**Test set of GenProps; may be null. */
        private final GenProps gp;

        public ExhibitStaticAttr getStaticAttr(final String name)
            throws IOException
            {
            return(aep.aeid.getStaticAttr(name));
            }

        /**Minimalist implementation; not really the spirit of the contract.
         * May or may not bother to check arguments
         * or even fill in the data buffer.
         */
        public void getRawFile(final ByteBuffer buf,
                               final String exhibitName, final int position, final boolean dontCache)
            throws IOException
            {
            }

        public AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
            throws IOException
            {
            if(oldStamp != aep.aeid.timestamp)
                { return(aep.aeid); }

            // Caller has up-to-date copy.
            return(null);
            }

        public AllExhibitProperties getAllExhibitProperties(final long oldHash)
            throws IOException
            {
            if(oldHash != aep.longHash)
                { return(aep); }

            // Caller has up-to-date copy.
            return(null);
            }

        /**Returns fixed (supplied) set of GenProps. */
        public GenProps getGenProps(final long oldStamp)
            throws IOException
            {
            return(gp);
            }

        /**Always returns empty Properties set. */
        public Properties getGenSecProps(final long oldStamp)
            throws IOException
            {
            return(new Properties());
            }

        /**Always returns null, ie we have no thumbnails. */
        public ExhibitThumbnails getThumbnails(final String name, final boolean create)
            throws IOException
            {
            return(null);
            }

        /**Basic "endpoint" set of variables. */
        private final BasicVarMgr vars = new BasicVarMgr(true);

        public void setVariable(final SimpleVariableValue newValue)
            throws IOException
            {
            vars.setVariable(newValue);
            }

        public int setVariables(final SimpleVariableValue[] newValues)
            throws IOException
            {
            return(vars.setVariables(newValues));
            }

        public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
            throws IOException
            {
            return(vars.getVariable(var));
            }

        public SimpleVariableValue[] getVariables(final long changedSince)
            throws IOException
            {
            return(vars.getVariables(changedSince));
            }

        public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                                final EventPeriod intervalSelector,
                                                final boolean current)
            {
            return(vars.getEventValue(def, intervalSelector, current));
            }

        public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                                   final EventPeriod intervalSelector,
                                                   final long intervalNumber,
                                                   final BitSet whichValues)
            throws IOException
            {
            return(vars.getEventValues(def, intervalSelector, intervalNumber, whichValues));
            }

        /**Synchronise with upstream values.
         * Does nothing in this implementation.
         */
        public void syncVariables(final boolean force)
            throws IOException
            {
            }

        /**Get requested Properties selected by key and versionID.
         * Fetches a Properties set unconditionally (versionID == -1)
         * else if the versionID presented is not current.
         *
         * @param key  selector (with possible embedded sub-key)
         *     for desired properties set; never null
         * @param versionID  if -1 then map is always returned if available,
         *     else must be non-negative and null is returned if the versionID
         *     presented matches that of the current version
         *     (ie if the caller has presumably got the up-to-date version);
         *     may be a timestamp or a hash or other value,
         *     and by convention is zero only for an empty properties set
         *
         * @return null, or Properties map guaranteed to contain only
         *     String keys and values
         */
        public java.util.Properties getProperties(final PropsKey key,
                                                  final long versionID)
            throws IOException
            {
            throw new IOException("NOT IMPLEMENTED");
            }


        /**Polling does nothing in this implementation. */
        public void poll(final GenProps gp)
            throws IOException
            {
            }

        /**Does nothing in this implementation. */
        public void destroy()
            {
            }
        }

    /**Internal test that DummyDataSource behaves as expected.
     * Essentially we ensure that basic routines return
     * legitimate non-null values.
     */
    public static void testDummyDataSource()
        throws Exception
        {
        // Ensure that we can create a dummy source...
        final DummyDataSource dds = new DummyDataSource();

        assertNotNull("Fetched GenProps should not be null",
                      dds.getGenProps(-1));
        assertNotNull("Fetched getAllExhibitProperties should not be null",
                      dds.getAllExhibitProperties(-1));
        assertNotNull("Should be able to get system ID for variables",
                      dds.getVariable(SystemVariables.LOCAL_SYS_ID));
        }

    /**Check that we can construct a cache wrapped round a (dummy) source.
     * And check that we can do some basic superficial operations on it.
     * <p>
     * Look elsewhere for a thorough test to behaviour.
     */
    public void testBasicCacheBehaviour()
        throws Exception
        {
        assertTrue("Must be no extant temporary/test cache dir.", !tempDir.exists());

        // Check that trying to remove a cache from a non-existant directory
        // does not fail in an ugly way.
        ExhibitDataSimpleCache.rmCache(tempDir);

        // Check that a null argument to rmCache() is rejected.
        try
            {
            ExhibitDataSimpleCache.rmCache(null);
            fail("rmCache() should reject null argument");
            }
        catch(final IllegalArgumentException e) { /* Correctly rejected. */ }

        // Set up new, empty, test cache directory.
        tempDir.mkdirs();
        assertTrue("Must be able to create temporary/test cache dir.", tempDir.isDirectory());

        // Check that removing an empty cache does not fail.
        ExhibitDataSimpleCache.rmCache(tempDir);
        // Check that removing an empty cache directory does not
        // remove the top-level (containing) directory.
        assertTrue("Removing cache should not remove containing temporary/test cache dir.", tempDir.isDirectory());

        // Check that we disallow creating a cache with a null up-stream source.
        try
            {
            ExhibitDataSimpleCache.cacheFactory(null, tempDir, logger);
            fail("should not allow creation of cache with null data source");
            }
        catch(IllegalArgumentException e) { /* Rejected as expected. */ }

        // Create a dummy data source...
        final DummyDataSource dds = new DummyDataSource();

        // Check that we disallow creating a cache with a null cache dir.
        try
            {
            ExhibitDataSimpleCache.cacheFactory(dds, null, logger);
            fail("should not allow creation of cache with null cache dir");
            }
        catch(final IllegalArgumentException e) { /* Rejected as expected. */ }

        // Attempt to create a new cache
        // wrapped around the dummy source...
        final ExhibitDataSimpleCache c1 = ExhibitDataSimpleCache.cacheFactory(
            dds, tempDir, logger);
        assertNotNull("Failed to create a cache instance", c1);

        // Check that some basic superficial operations work.
        assertNotNull("Fetched GenProps should not be null",
                      c1.getGenProps(-1));
        assertNotNull("Fetched getAllExhibitProperties should not be null",
                      c1.getAllExhibitProperties(-1));
        assertNotNull("Should be able to get system ID for variables",
                      c1.getVariable(SystemVariables.LOCAL_SYS_ID));

        // Check that system ID has come properly from upstream.
        assertSame("System ID should have come from upstream",
                   dds.getVariable(SystemVariables.LOCAL_SYS_ID),
                   c1.getVariable(SystemVariables.LOCAL_SYS_ID));
        }

    /**Test that variables work properly with the cache.
     * This is just to make sure that its semantics are correct.
     */
    public void testCacheVarHandling()
        throws Exception
        {
        // Set up new, empty, test cache directory.
        tempDir.mkdirs();

        // Create a dummy data source...
        final DummyDataSource dds1 = new DummyDataSource();

        // Attempt to create a new cache (c1)
        // wrapped around the dummy source...
        final ExhibitDataSimpleCache c1 = ExhibitDataSimpleCache.cacheFactory(
            dds1, tempDir, logger);
        assertNotNull("Failed to create a cache instance", c1);

        // Start polling at a typical rate...
        final TimerTask tt1 = new TimerTask(){
            public void run()
                {
                final GenProps gp = c1.getGenProps(-1);
                try { c1.poll(gp); }
                catch(final Throwable t) { t.printStackTrace(Main.getErr()); }
                }
            };
        t.schedule(tt1, 0L, 100 + rnd.nextInt(700));

        // Spend at most a few seconds testing behaviour...
        final SimpleExhibitPipelineIF pipeline[] = { c1, dds1 };
        SystemVariablesTest.singlePipelineTest(pipeline,
            10000 + System.currentTimeMillis()); // 10s test.

        // Stop poll activity for c1; it effectively dies now.
        tt1.cancel();
        }

    /**Test cacheing and resilience after restart.
     * This test attempts to test that precacheing works
     * and that the cache remains intact after a restart.
     * <p>
     * This relies on the test data set being present
     * and being easily small enough to fit into cache.
     */
    public void testCachePersistence()
        throws Exception
        {
        // Set up new, empty, test cache directory.
        tempDir.mkdirs();

        // Combined "trace" of all relevant events so far, in order.
        final List<RawDataAccessMonitor.GetRawDataEvent> events = new ArrayList<RawDataAccessMonitor.GetRawDataEvent>();

        // Target time to stop first round of tests.
        final long stopBy1 = System.currentTimeMillis() + 15000L; // 15s...

        // Running count of cached exhibits...
        // Must never go down
        // (on the assumption that our test data set is
        // MUCH smaller than our configured cache size).
        int exhibitsCached = 0;

        // TEST FIRST INSTANCE OF CACHE...
        // Starting with an absolutely fresh, empty cache area...
        Main.getOut().println("Creating/testing new/empty cache...");
        exhibitsCached = checkCacheInstance(true, exhibitsCached, events,
                                            stopBy1);

        // Target time for remaining (re)tests...
        final long stopBy = System.currentTimeMillis() + 30000L;

        // At least once test that any new cache sees all the previously-cached
        // data, ie does not lose it on shut-down or can reconstruct it.
        for(int i = 5; --i >= 0; ) // Try at most a few times...
            {
            Main.getOut().println("[Exhibits partly/fully cached: "+exhibitsCached+"...]");

            // TEST NEW INSTANCE OF CACHE OBJECT IN SAME DISC AREA...
            // Make sure that it does not lose data from the previous run.
            Main.getOut().println("Re-testing exiting cache...");
            exhibitsCached = checkCacheInstance(false, exhibitsCached, events,
                                                stopBy);

            if(System.currentTimeMillis() >= stopBy) { break; }
            }
        }

    /**Thread-safe monitor of data-read activity.
     * An instance of this class can be inserted into a pipeline
     * to record data-read activity.
     * <p>
     * This filter stage does not generally impede normal operation,
     * but simply records the data access pattern for processing.
     */
    static final class RawDataAccessMonitor extends SimpleExhibitPipelineFilter
        {
        /**Construct instance wrapped round upstream source.
         * @param upstream  datasource; never null
         */
        RawDataAccessMonitor(final SimpleExhibitPipelineIF upstream)
            { super(upstream); }

        /**A List of GetRawDataEvent values in the order they occurred.
         * A Vector is used for thread-safety.
         */
        private final List<GetRawDataEvent> events = new Vector<GetRawDataEvent>();

        /**A "getRawData()" event.
         * Immutable.
         * <p>
         * Records the time, stack trace, exhibit name, start and length.
         */
        static final class GetRawDataEvent
            {
            public GetRawDataEvent(final String name, final int start, final int len)
                {
                this.name = name;
                this.start = start;
                this.len = len;
                }

            /**Time at which event happened. */
            public final long timestamp = System.currentTimeMillis();

            /**Stack trace when event happened. */
            private final Throwable th = new Throwable();
            /**Print the stack trace to the specified stream. */
            public void printStackTrace(final PrintStream ps)
                { th.printStackTrace(ps); }

            /**Name of exhibit concerned. */
            public final String name;
            /**Start offset of read within exhibit. */
            public final int start;
            /**Length of read within exhibit. */
            public final int len;
            }

        /**Read a chunk of the raw exhibit binary into the given buffer.
         * This passes the request upstream without modification,
         * but records it in a trace buffer <em>before</em> passing it upstream.
         * The trace is taken before in case the fulfillment of the request
         * happens to be vetoed by the data source; it is the requests that
         * we are interested in here.
         */
        public void getRawFile(final ByteBuffer buf,
                               final String exhibitName, final int position, final boolean dontCache)
            throws IOException
            {
            events.add(new GetRawDataEvent(exhibitName, position, buf.limit() - buf.position()));
            super.getRawFile(buf, exhibitName, position, dontCache);
            }

        /**Retrieve the List of events in order; never null though possibly empty.
         * This returns a private snapshot of the event list.
         */
        public List<GetRawDataEvent> getEventList()
            {
            // Hold a lock on events while we take the snapshot...
            synchronized(events)
                {
                return(new ArrayList<GetRawDataEvent>(events));
                }
            }
        }

    /**Check behaviour of single cache instance, mainly for persistence.
     * This checks for different things on the first and subsequent runs,
     * but in all cases makes sure that the cache behaves properly
     * and does not lose data upon restart,
     * whether it gets shut down properly or not.
     * <p>
     * The number of exhibits partially or fully cached must not go down
     * and must be greater than zero when we return if there are exhibits.
     * <p>
     * This tests the cache's use of its upstream data source
     * knowing that this should read and cache exhibit prefixes
     * (ie it will/may only cache activity that reads starting at the
     * beginning of an exhibit, as a normal download should),
     * and that having read such a prefix it should never have to go
     * back to the underlying source for anything within that prefix again,
     * even when the cache is shut down an restarted,
     * assuming that our cache area is big enough to hold all the exhibits.
     *
     * @param firstRun  if true, this is the first run on a new cache area
     * @param prevCachedCount  number of exhibits previously cached;
     *     non-negative
     * @param eventsFromTestStart  List of events collected so far in order;
     *     we will append to this (the content is opaque to the caller);
     *     must be non-null and appendable-to
     * @param stopBy  target time to stop by
     *
     * @return number of exhibits partially or fully cached; non-negative
     *
     * @throws IOException
     * @throws InterruptedException
     */
    private int checkCacheInstance(final boolean firstRun,
                                   final int prevCachedCount,
                                   final List<RawDataAccessMonitor.GetRawDataEvent> eventsFromTestStart,
                                   final long stopBy)
        throws IOException,
               InterruptedException
        {
        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        // Set up a data-access monitoring stage.
        // We use this to monitor the effectiveness of the cache.
        final RawDataAccessMonitor rdam = new RawDataAccessMonitor(edfs);

        // Set up a simple cache instance.
        final ExhibitDataSimpleCache cache = ExhibitDataSimpleCache.cacheFactory(
            rdam, tempDir, logger);
        assertNotNull("Failed to create"+(firstRun?" first":"")+" cache instance", cache);


        // Check some features of the cache before the first poll()...

        // Collect the raw exhibit info from the underlying file system.
        final AllExhibitProperties aepRaw = edfs.getAllExhibitProperties(-1);
        // When re-using an existing cache directory,
        // we assume that the exhibit properties are immediately available
        // before even a single poll.
        if(!firstRun)
            {
            assertEquals("Exhibit properties via cache must match those fetched directly even before first poll() is cache is not new",
                         aepRaw, cache.getAllExhibitProperties(-1));
            }

        // Set up data pipeline.
        final SimpleExhibitPipelineIF pipeline[] = { cache, rdam, edfs };
        final SimpleExhibitPipelineIF top = pipeline[0];


        // Start poll()ing the pipeline at a typical rate...
        // Vary the poll rate randomly a little
        // in order to demonstate that we are not very sensitive to it.
        // Note that cache implementation does not necessarily
        // use all poll() calls for real processing...
        final TimerTask tt = new TimerTask(){
            public void run()
                {
                try
                    {
                    final GenProps gp = top.getGenProps(-1);
                    top.poll(gp);
                    }
                catch(Exception e) { e.printStackTrace(Main.getErr()); }
                }
            };
        t.schedule(tt, 0L, 100 + rnd.nextInt(700));


        // Make sure that we can see all exhibit data via the cache
        // just like when we get it directly...
        // We have forced one poll(),
        // and that one poll() should be enough even on the first run
        // to ensure that all data is up-to-date
        // unless there is some sort of I/O error...
        AllExhibitProperties aepViaCache = cache.getAllExhibitProperties(-1);
        assertNotNull("Exhibit properties (aep) via cache must not be null",
                      aepViaCache);
        // Wait until aep (and GenProps) are not at their default values,
        // ie the cache has had time to fetch them.
        // This must not take more than a few seconds however...
        if(aepRaw.aeid.size() > 0)
            {
            // Wait a few seconds for the data to get loaded...
            for(int i = 30; --i >= 0; )
                {
                aepViaCache = cache.getAllExhibitProperties(-1);
                if(aepViaCache.aeid.size() > 0) { break; }
                System.out.println("Waiting for cache to be reloaded...");
                Thread.sleep(2000L);
                }
            assertTrue("Exhibit properties (aep) via cache must not be empty if underlying ones are not",
                       aepViaCache.aeid.size() > 0);
            }
        assertEquals("Exhibit properties (aep) via cache must match those fetched directly",
                     aepRaw, aepViaCache);
        assertEquals("Exhibit properties (aeid) via cache must match those fetched directly",
                     edfs.getAllExhibitImmutableData(-1),
                     cache.getAllExhibitImmutableData(-1));

        // Check that data is seen consistently on either side of the cache.
        HTTPTunnelTest.checkExhibitDataConsistency(cache, edfs,
                                                   false, // Not a tunnel...
                                                   stopBy);

        // Read the first byte of each exhibit
        // to ensure that something should be cached for each.
        // Have the "dontCache" flag false to try to ensure we cache the data.
        final ByteBuffer buf = ByteBuffer.allocate(1);
        for(final Iterator<String> it = aepViaCache.aeid.getAllExhibitNamesSet().iterator(); it.hasNext(); )
            {
            final String name = it.next();
            buf.clear();
            cache.getRawFile(buf, name, 0, false);
            }

        // Now check the cached exhibit count...
        cache.syncVariables(true); // Force our view to be up-to-date...
        final SimpleVariableValue cc =
            cache.getVariable(SystemVariables.ExhibitDataSimpleCache_CACHED_EXHIBIT_COUNT);
        assertNotNull("Must be able to retrieve cached exhibit count",
                      cc);
        assertNotNull("Must be able to retrieve non-null cached exhibit count",
                      cc.getValue());
        final int cachedCountVar = ((Number) cc.getValue()).intValue();
        assertTrue("Must be able to retrieve non-negative cached exhibit count",
                   cachedCountVar >= 0);

        // The number of (partly/fully) cached exhibits must not go down;
        // at worst we must be able to retrieve the previous
        // possibly-slightly-stale meta-data from disc
        // and/or reconstruct it from the data we actually find in place.
        final int cachedCount = cache.getLiveCachedExhibitCount();
        assertTrue("Number of cached exhibits must not go down",
                   cachedCount >= prevCachedCount);

        // Get the live raw count of exhibits partly/fully cached.
        if(aepRaw.aeid.size() > 0)
            {
            assertTrue("The number of exhibits cached after some data has been retrieved must be greater than zero if there are exhibits",
                       cachedCount > 0);
            }


//        if(firstRun)
//            {
//            // Test variable-handling behaviour...
//            // (We must be calling poll() by now.)
//            SystemVariablesTest.singlePipelineTest(pipeline, stopBy);
//            }


        // Kill off this instance.
        // It will get no more poll() calls so is effectively dead.
        tt.cancel();

        // If we are feeling nice then we'll properly destroy() it
        // to allow data to be flushed to disc.
        /* if(rnd.nextBoolean()) */  { cache.destroy(); }

        // Now examine the upstream activity trace from the current run...
        // We make sure that the following bad/marginal things do not happen:
        //   * No requests for non-existant exhibits.
        //   * No requests from beyond exhibit bounds.
        //   * No zero-length (or other badly-formed) requests...
        // Event count before this run.
        final int oldEventCount = eventsFromTestStart.size();
        Main.getOut().println("[getRawData() events before this run: "+oldEventCount+".]");
        final List<RawDataAccessMonitor.GetRawDataEvent> eventsThisTime = rdam.getEventList();
        Main.getOut().println("[getRawData() events counted this run: "+eventsThisTime.size()+".]");
        for(final Iterator<RawDataAccessMonitor.GetRawDataEvent> it = eventsThisTime.iterator(); it.hasNext(); )
            {
            final RawDataAccessMonitor.GetRawDataEvent e =
                    it.next();

            // Look for obviously badly-formed requests.
            if((e.name == null) || (e.start < 0) || (e.len < 0))
                { fail("Badly formed getRawData() request detected (null/-ve args)."); }
            if(e.len == 0)
                { fail("Marginal/deprecated getRawData() request detected (len==0)."); }
            final ExhibitStaticAttr staticAttr = aepRaw.aeid.getStaticAttr(e.name);
            if(staticAttr == null)
                { fail("Bad getRawData() request detected (no such exhibit)."); }
            if((e.start >= staticAttr.length))
                { fail("Bad getRawData() request detected (start >= length)."); }
            if((e.start + e.len > staticAttr.length))
                { fail("Bad getRawData() request detected (start+len > length)."); }
            }

        // Append this run's events to the running tally.
        eventsFromTestStart.addAll(eventsThisTime);

        // Check that the cache has not "forgotten" anything after a restart.
        // Ensure that exhibit prefixes that have already been fetched
        // are not refetched...
        // But we do it on all runs in case we are generally amnesiac...
//        if(!firstRun)
//        if(true) { System.err.println("WARNING: skipping refetch tests"); } else
            {
            // Map from exhibit name (String)
            // to high-water mark that should have been cached (Long).
            final Map<String,Long> marks = new HashMap<String, Long>();

            int eventNum = 1;
            for(final Iterator<RawDataAccessMonitor.GetRawDataEvent> it = eventsFromTestStart.iterator(); it.hasNext(); ++eventNum)
                {
                final RawDataAccessMonitor.GetRawDataEvent e =
                        it.next();

                // Get any extant entry for the exhibit
                // that is the subject of this event...
                // If no mark so far then we have cached zero bytes so far.
                final Long cachedSoFarL = marks.get(e.name);
                final long cachedSoFar = (cachedSoFarL == null) ?
                    0 : cachedSoFarL.longValue();

                // Create a description of the current event,
                // just in case we need to print it...
                final StringBuilder sb = new StringBuilder(32);
                if(eventNum >= oldEventCount)
                    { sb.append("NEW,"); }
                sb.append("event#=").append(eventNum);
                sb.append(",cachedSoFar=").append(cachedSoFar);
                sb.append(",e.start=").append(e.start);
                sb.append(",e.len=").append(e.len);
                sb.append(",e.name=").append(e.name);
                final String description = sb.toString();

                // If any part of this request should have been
                // locally satisfied from cache,
                // it is an error.
                if(e.start < cachedSoFar)
                    {
                    Main.getErr().println("ERROR: invalid refetch: " + description);
                    e.printStackTrace(Main.getErr());
                    fail("Re-fetched data from source that should still be in cache: ("+description+")");
                    }

                // If this request starts with a gap from the
                // currently-cached prefix then it is a random access
                // and will not be cached...
                if(e.start > cachedSoFar)
                    {
//                    Main.getOut().println("Random read assumed not added to cache... ("+description+")");
                    continue;
                    }

                // If this read extends the cached prefix,
                // update the mark entry for this exhibit.
                final int readTo = e.start + e.len;
                if(readTo > cachedSoFar)
                    {
//                    Main.getOut().println("Extended cache... ("+description+")");
                    marks.put(e.name, new Long(readTo));
                    }
                }
            }

        // Return cached exhibit count.
        return(cachedCount);
        }



    /**Thread-safe monitor of data-read activity.
     * An instance of this class can be inserted into a pipeline
     * to record or modify data-read activity.
     */
    static final class DataConcurrencyAccessMonitor extends SimpleExhibitPipelineFilter
        {
        DataConcurrencyAccessMonitor(final SimpleExhibitPipelineIF upstream)
            { super(upstream); }

        /**Once true, threads do not block in getRawFile(); volatile so as not to need the mutex. */
        private volatile boolean dontBlock;

        /**Count of blocked thread in getRawData().
         * Must only be accessed under protection of the notifier mutex.
         */
        private int blockedThreadsCount;

        /**Get count of blocked thread in getRawData(); non-negative. */
        private int getBlockedThreadsCount() { synchronized(notifier) { return(blockedThreadsCount); } }

        /**Used to notify/wake blocked threads. */
        private final Object notifier = new Object();

        /**Call this to release any waiting/blocked threads in getRawFile(). */
        void release()
            {
            dontBlock = true;
            synchronized(notifier)
                {
                notifier.notifyAll(); // Wake everyone up...
                }
            }

        /**Intercept and block calls as if blocked waiting for disc access. */
        @Override
        public void getRawFile(final ByteBuffer buf, final String exhibitName, final int position, final boolean dontCache) throws IOException
            {
            // Block until released.
            try
                {
                synchronized(notifier)
                    {
//(new Throwable("BLOCKING: blockedThreadsCount = " + blockedThreadsCount)).printStackTrace(System.out);
                    ++blockedThreadsCount;
                    while(!dontBlock)
                        { notifier.wait(501 + rnd.nextInt(501)); }
                    --blockedThreadsCount;
//(new Throwable("UNBLOCKING: blockedThreadsCount = " + blockedThreadsCount)).printStackTrace(System.out);
                    }
                }
            catch(final InterruptedException e)
                {
                throw new IllegalMonitorStateException("Unexpected wakeup");
                }

            // Really do the call...
            super.getRawFile(buf, exhibitName, position, dontCache);
            }
        }


    /**Test that cache has indefinite (file read) concurrency.
     * The aim is to demonstrate that we can scale well to serve exhibit data,
     * eg on a highly-threadable system such as Niagara/CoolThreads.
     */
    public void testCacheFileReadConcurrency()
        throws Exception
        {
        // Pick how many threads that we will try to push through the cache.
        final int nThreads = 30 + rnd.nextInt(50);

        // Set up new, empty, test cache directory.
        tempDir.mkdirs();

        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        // Set up a data-access monitoring stage.
        // We use this to monitor the effectiveness of the cache.
        final DataConcurrencyAccessMonitor dcam = new DataConcurrencyAccessMonitor(edfs);

        // Set up a simple cache instance.
        final ExhibitDataSimpleCache cache = ExhibitDataSimpleCache.cacheFactory(
            dcam, tempDir, logger);
        assertNotNull("Failed to create cache instance", cache);

        // Collect the raw exhibit info from the underlying file system.
        final AllExhibitProperties aepRaw = edfs.getAllExhibitProperties(-1);

        // Start poll()ing the pipeline at a typical rate...
        // Vary the poll rate randomly a little
        // in order to demonstrate that we are not very sensitive to it.
        // Note that cache implementation does not necessarily
        // use all poll() calls for real processing...
        final TimerTask tt = new TimerTask(){
            public void run()
                {
                try
                    {
                    final GenProps gp = cache.getGenProps(-1);
                    cache.poll(gp);
                    }
                catch(final Exception e) { e.printStackTrace(Main.getErr()); }
                }
            };
        t.schedule(tt, 0L, 100 + rnd.nextInt(700));

        // Wait until the cache has loaded the exhibit set.
        final long waitStartTime = System.currentTimeMillis();
        do
            {
            Main.getOut().println("Waiting for cache to load AEP... ("+(System.currentTimeMillis()-waitStartTime)+"ms)");
            Thread.sleep(1000);
            } while(cache.getAllExhibitProperties(-1).aeid.size() != aepRaw.aeid.size());

        // A worker class to read a chunk of data via the cache
        // any verify it against the underlying data.
        final class ReaderThread extends Thread
            {
            /**Read some data from the named exhibit. */
            ReaderThread(final String exhibitName)
                {
                super("ReaderThread for " + exhibitName);
                exhibit = exhibitName;
                }

            /**Full name of exhibit to read from; never null. */
            private final String exhibit;

            /**Set true iff completed OK. */
            volatile boolean completedOK;

            public final void run()
                {
                final ExhibitStaticAttr esa = aepRaw.aeid.getStaticAttr(exhibit);
                assertNotNull(esa);

                // Often start read at zero, to get the data cached.
                final int start = rnd.nextBoolean() ? 0 : rnd.nextInt((int) esa.length);
                // Limit length of read to reasonable block size.
                final int length = rnd.nextInt(Math.min(16384, (int) esa.length - start));

                try
                    {
                    final ByteBuffer buf1 = ByteBuffer.allocate(length);
                    cache.getRawFile(buf1, exhibit, 0, rnd.nextBoolean());

                    final ByteBuffer buf2 = ByteBuffer.allocate(length);
                    edfs.getRawFile(buf2, exhibit, 0, rnd.nextBoolean());

                    // If the data was the same through the cache as raw
                    // then we succeeded.
                    if(buf1.equals(buf2))
                        { completedOK = true; }
                    else
                        { Main.getErr().println("Failed for: " + exhibit); }
                    }
                catch(final Exception e)
                    { e.printStackTrace(); /* Whoops! */ }
                }
            }

        // Manage a group of workers to use the cache.
        final List<ReaderThread> workers = new ArrayList<ReaderThread>(nThreads);

        try
            {
            // Start one reader/worker thread per exhibit.
            // All these should be able to run concurrently,
            // reaching through the cache to the underlying file source,
            // as each will be the first read of this exhibit
            // with no data in the cache for it.
            for(final String exhibitName : aepRaw.aeid.getAllExhibitNamesSet())
                {
                if(workers.size() >= nThreads) { break; }

                final ReaderThread rt = new ReaderThread(exhibitName);
                workers.add(rt);
                rt.setPriority(Thread.MAX_PRIORITY); // Should run before us...
                rt.start();
                Thread.yield();
                }

            // Be prepared to sleep a little for all threads to block...
            mainLoop: for(int i = 9+(5*workers.size()); (workers.size() > dcam.getBlockedThreadsCount()) && (--i >= 0); )
                {
                Main.getOut().println("Waiting for threads to block/settle... ("+(System.currentTimeMillis()-waitStartTime)+"ms)");

                // Look for any prematurely-finished workers, and zap them.
                for(final Iterator<ReaderThread> rtIt = workers.iterator(); rtIt.hasNext(); )
                    {
                    final ReaderThread rt = rtIt.next();
                    if(!rt.isAlive())
                        {
                        Main.getErr().println("WARNING: worker exited early.");
                        rtIt.remove();
                        continue mainLoop;
                        }
                    }

                Thread.sleep(1000);
                }

            // Check that new threads hit the back-end
            // even if we have blocked all previous threads in the back-end,
            // ie that the cache is not a concurrency bottleneck.
            final int initialWorkerCount = workers.size();
            assertEquals("All threads must be blocked in the getRawFile() interceptor",
                         initialWorkerCount, dcam.getBlockedThreadsCount());

            if(workers.size() < nThreads)
                {
                Main.getOut().println("Starting extra reader threads... ("+(System.currentTimeMillis()-waitStartTime)+"ms)");

                // Set off remaining threads for random exhibits...
                // These may not all immediately be allowed through by the cache,
                // but they should all work correctly.
                final String[] names = aepRaw.aeid.getAllExhibitNamesSorted();
                while(workers.size() < nThreads)
                    {
                    final ReaderThread rt = new ReaderThread(names[rnd.nextInt(names.length)]);
                    workers.add(rt);
                    rt.start();
                    }

                Thread.yield(); // Give the new threads a chance to start...

                // Check that the blocked-threads count is still reasonable.
                assertTrue("All initial worker threads must (still) be blocked in the getRawFile() interceptor",
                            initialWorkerCount <= dcam.getBlockedThreadsCount());
                }
            }
        finally
            {
            final long releaseStart = System.currentTimeMillis();
            final long timeSoFar = (releaseStart-waitStartTime);
            Main.getOut().println("Releasing threads... ("+timeSoFar+"ms)");

            dcam.release(); // Let the threads go...
            Thread.yield(); // Give them a sporting chance to execute...
            Main.getOut().println("Released threads...");

            // Wait for all the worker/reader threads and test their status.
            // Allow a fair amount of time for thread thrashing, etc.
            // We make this slightly adaptive to the speed of this machine/JVM.
            final long maxWaitTime = 60001 + (timeSoFar/2);
            final long targetFinishTime = releaseStart + maxWaitTime;
            final long hopedForFinishTime = releaseStart + maxWaitTime/4;
            int failCount = 0;
            for(int i = workers.size(); --i >= 0; )
                {
                final ReaderThread readerThread = workers.get(i);
                // If we've already been waiting a while, print progress...
                final long now = System.currentTimeMillis();
                if(now > hopedForFinishTime)
                    { Main.getOut().println("Waiting for worker thread "+i+"... ("+(now-waitStartTime)+"ms)"); }
                // We will not wait indefinitely...
                readerThread.join(Math.max(1001, targetFinishTime - now));
                if(!readerThread.completedOK)
                    { ++failCount; }
                }
            if(failCount > 0) { fail("Workers failed: " + failCount + " out of " + workers.size()); }
            }
        }


    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
