/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;

import org.hd.d.pg2k.ai.scorer.ScorerCacheIF;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineFilter;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.*;

/**Test the behaviour of system variables.
 * This includes as stand-alone objects,
 * in variable-manager caches,
 * in pipelines and tunnels, etc.
 */
public final class SystemVariablesTest extends TestCase
    {
    public SystemVariablesTest(final String name)
        {
        super(name);
        }

    private final SimpleLoggerIF logger = new SimpleLoggerIF(){
        public void log(final String message) { Main.getOut().println(message); }
        };

    /**Timer that we can use to run pipelines; never null during tests. */
    private Timer t;

    /**Do any setup needed for the tests. */
    protected void setUp()
        {
        t = new Timer();
        }

    /**Do any clearup needed after the tests. */
    protected void tearDown()
        {
        t.cancel();
        t = null;
        }


    /**Check some simple (de)serialistion works as expected.
     * In particular, deserialising one of the canonical system variable definition
     * should yield the canonical instance, ie that we our providing instance control.
     * <p>
     * However, it should still be possible to (de)serialise definitions
     * (and compound values containing them)
     * that are no longer legitimate system variable (and do not conflict with them)
     * so that we can examine historical data for example.
     */
    public void testSimpleSerialisation()
        throws Exception
        {
        // Test on all system definitions...
        for(final SimpleVariableDefinition def : SystemVariables.nameToDef.values())
            {
            assertTrue("SimpleVariableDefinition.equals() should work for itself!", def.equals(def));
            final SimpleVariableDefinition defDeser = (SimpleVariableDefinition)
                    SerializationTest.checkSerialisationPreservesEquality(def);
            assertTrue("Deserialised canonical object should == original",
                       defDeser == def);
            }

        // Check that we can (de)serialise arbitrary (non-conflicting) definitions.
        final SimpleVariableDefinition newDefs[] =
                {
                    new SimpleVariableDefinition("newvar." + (rnd.nextLong() >>> 1),
                                         SimpleVariableDefinition.TYPE_NUMBER,
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(), 0, null),
                    new SimpleVariableDefinition("newvar." + (rnd.nextLong() >>> 1),
                                         SimpleVariableDefinition.TYPE_STRING,
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(), 0, null),
                    new SimpleVariableDefinition("newvar." + (rnd.nextLong() >>> 1),
                                         SimpleVariableDefinition.TYPE_NONE,
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(), 0, null),
                    new SimpleVariableDefinition("newvar." + (rnd.nextLong() >>> 1),
                                         SimpleVariableDefinition.TYPE_IID,
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(),
                        rnd.nextBoolean(), 0, null),
                };
        for(final SimpleVariableDefinition def : newDefs)
            {
            assertTrue("SimpleVariableDefinition.equals() should work for itself!", def.equals(def));
            final SimpleVariableDefinition defDeser = (SimpleVariableDefinition)
                    SerializationTest.checkSerialisationPreservesEquality(def);
            assertTrue("Deserialised non-canonical object should != original",
                       defDeser != def);

            if(def.isEvent())
                {
                // Check that we can (de)serialise event collection objects with non-std defs.
                final EventVariableValueBuffer evvb =
                        new EventVariableValueBuffer(def,
                                                     EventPeriod.values()[rnd.nextInt(EventPeriod.values().length)],
                                                     1);
                assertTrue("EventVariableValueBuffer.equals() should work for itself!", evvb.equals(evvb));
                SerializationTest.checkSerialisationPreservesEquality(evvb);

                final EventVariableValue evv = evvb.toEventVariableValue();
                assertTrue("EventVariableValue.equals() should work for itself!", evv.equals(evv));
                SerializationTest.checkSerialisationPreservesEquality(evv);
                }
            }
        }


    /**Do simple tests on InstanceID type...
     * Such as making sure that all instance are unique, sort correctly, etc.
     */
    public void testInstanceID()
        throws Exception
        {
        // Check equality works for equal values.
        final InstanceID iid  = InstanceID.createInstanceID();
        assertTrue("InstanceID.equals() must work correctly for equal/same value",
            iid.equals(iid));
        assertTrue("InstanceID.compareTo() must work correctly for equal/same value",
            iid.compareTo(iid) == 0);

        // Just FYI...
        Main.getOut().println("[Sample InstanceID: " + iid + ".]");

        // Check equals() returns false for worng object tyupe.
        assertFalse("equals() must fail for non-InstanceID arg",
            iid.equals(rnd));

        final int NUM_TO_TEST = 100;
        final List<InstanceID> allValues = new ArrayList<InstanceID>(NUM_TO_TEST);
        for(int i = NUM_TO_TEST; --i >= 0; )
            { allValues.add(InstanceID.createInstanceID()); }
        // Check that they are all distinct
        // as far as Set is concerned using equals()/hashCode().
        assertEquals("All InstanceID values generated must in practice be distinct for equals()/hashCode()",
            NUM_TO_TEST, (new HashSet<InstanceID>(allValues)).size());

        // Now sort our values to see if Comparable implementation OK...
        Collections.sort(allValues);
        for(int i = NUM_TO_TEST; --i > 0; )
            {
            final InstanceID vn = allValues.get(i-1);
            final InstanceID vm = allValues.get(i);
            assertTrue("Values must be correctly totally ordered",
                vn.longValue() < vm.longValue());
            assertTrue("longValue()s must be non-negative",
                vn.longValue() >= 0);
            assertTrue("String representation must be reasonable size",
                (vn.toString().length() > 0) && (vn.toString().length() < 80));
            }

        // Check that we can (de)serialise...
        SerializationTest.checkSerialisationPreservesEquality(
            InstanceID.createInstanceID());

        // Check that we don't make extra copies in memory when deserialising.
        assertTrue("Deserialisation must not make extra copies of InstanceID",
            iid == SerializationTest.checkSerialisationPreservesEquality(iid));
        }

    /**Test that variables of fairly arbitrary names/types can be constructed.
     * This means, for example, that we can ship any variables between systems
     * across HTTP tunnels, etc, providing that any stored values type-check,
     * (even if we are then not allowed to set the values in a local variable
     * store).
     */
    public void testArbitraryVarConstruction()
        throws Exception
        {
        // Run a few test constructions.
        for(int i = 100; --i >= 0; )
            {
            final int type = SimpleVariableDefinition.TYPE__MIN +
                rnd.nextInt(SimpleVariableDefinition.TYPE__MAX + 1 -
                            SimpleVariableDefinition.TYPE__MIN);

            final StringBuilder name = new StringBuilder(99);
            do  // Make name of random length...
                {
                if(name.length() > 0) { name.append('.'); }
                name.append(Long.toHexString(rnd.nextLong() >>> 1));
                } while((rnd.nextInt(3) > 0) &&
                        (name.length() < SimpleVariableDefinition.MAX_NAME_LEN / 2));
            final boolean isLocal = rnd.nextBoolean();
            final boolean isPersistent = rnd.nextBoolean();
            final boolean isReadOnly = rnd.nextBoolean();
            final boolean isEvent = rnd.nextBoolean();
            // Ensure that it is possible to construct arbitrary definition.
            final SimpleVariableDefinition def = new SimpleVariableDefinition(
                name.toString(),
                type,
                isLocal,
                isPersistent,
                isReadOnly,
                isEvent, 0, null);
            // ...and check that we can (de)serialise it.
            SerializationTest.checkSerialisationPreservesEquality(def);
            // Check that all the getters work.
            assertEquals("Broken isLocal()", isLocal, def.isLocal());
            assertEquals("Broken isPersistent()", isPersistent, def.isPersistent());
            assertEquals("Broken isReadOnly()", isReadOnly, def.isReadOnly());
            assertEquals("Broken isEvent()", isEvent, def.isEvent());

            // Select a legal value for the type;
            // including null (which is valid for all types).
            Object value = null;
            if(rnd.nextBoolean())
                {
                switch(type)
                    {
                    case SimpleVariableDefinition.TYPE_NUMBER:
                        {
                        value = new Double(rnd.nextGaussian());
                        break;
                        }

                    case SimpleVariableDefinition.TYPE_IID:
                        {
                        value = InstanceID.createInstanceID();
                        break;
                        }

                    case SimpleVariableDefinition.TYPE_STRING:
                        {
                        value = "silly text " + rnd.nextLong();
                        break;
                        }
                    }
                }
            // Ensure that we can construct an instance...
            final SimpleVariableValue svv = new SimpleVariableValue(def, value);
            // Check that the type getter method works OK.
            assertEquals("Must be able to getType() correctly",
                type, def.getType());
            // ...and check that we can (de)serialise it.
            SerializationTest.checkSerialisationPreservesEquality(svv);

            // Now try building a value of each type
            // with totally bogus values that should fail to type-check.
            final Object badValues[] = { rnd, new Object(), Void.class };
            for(int b = badValues.length; --b >= 0; )
                {
                try
                    {
                    new SimpleVariableValue(def, badValues[b]);
                    fail("Type-checking failed with bogus construction, def=" + def);
                    }
                catch(IllegalArgumentException e) { } // Should reject.
                }
            }
        }

    /**Test basic variable manager behaviour for the endpoint.
     * Instances not marked as an endpoint should not have a system ID defined,
     * whereas an endpoint should have one defined.
     */
    public void testBasicVarMgrAsEndPoint()
        {
        final BasicVarMgr bvm1 = new BasicVarMgr();
        assertFalse("BasicVarMgr should not be end-point unless marked as such",
            bvm1.isEndPoint());
        assertNull("BasicVarMgr should not contain an IID unless an endpoint",
            bvm1.getVariable(SystemVariables.LOCAL_SYS_ID));

        final BasicVarMgr bvm2 = new BasicVarMgr(true);
        assertTrue("BasicVarMgr should be end-point when marked as such",
            bvm2.isEndPoint());
        assertNotNull("BasicVarMgr should contain an IID when an end-point",
            bvm2.getVariable(SystemVariables.LOCAL_SYS_ID));

        // Make sure that we cannot override system ID in any BasicVarMgr...
        final SimpleVariableValue svvBogusSysID = new SimpleVariableValue(
            SystemVariables.LOCAL_SYS_ID, InstanceID.createInstanceID());
        // Nor set variable not listed in SystemVariables.defs.
        final SimpleVariableDefinition bogusDef = new SimpleVariableDefinition(
            "bogus.random.name." + (rnd.nextInt() >>> 1),
            SimpleVariableDefinition.TYPE__MIN +
                rnd.nextInt(1 + SimpleVariableDefinition.TYPE__MAX -
                                SimpleVariableDefinition.TYPE__MIN)
            );
        final SimpleVariableValue svvBogus = new SimpleVariableValue(
            bogusDef, null);
        final BasicVarMgr bvms[] = { bvm1, bvm2 };
        for(int i = bvms.length; --i >= 0; )
            {
            try
                {
                bvms[i].setVariable(svvBogusSysID);
                fail("Apparently allowed system ID to be set with setVariable() (from downstream)");
                }
            catch(IOException e) { }
            catch(UnsupportedOperationException e) { }
            catch(IllegalArgumentException e) { }

            try
                {
                if(bvms[i].setVariables(new SimpleVariableValue[]{svvBogusSysID}) != 0)
                    { fail("Apparently allowed system ID to be set with setVariables() (from downstream)"); }
                }
            catch(IOException e) { }
            catch(UnsupportedOperationException e) { }
            catch(IllegalArgumentException e) { }

            try
                {
                bvms[i].setVariable(svvBogus);
                fail("Apparently allowed non-extant variable to be set with setVariable() (from downstream)");
                }
            catch(IOException e) { }
            catch(UnsupportedOperationException e) { }
            catch(IllegalArgumentException e) { }

            try
                {
                if(bvms[i].setVariables(new SimpleVariableValue[]{svvBogus}) != 0)
                    { fail("Apparently allowed non-extant variable to be set with setVariables() (from downstream)"); }
                }
            catch(IOException e) { }
            catch(UnsupportedOperationException e) { }
            catch(IllegalArgumentException e) { }
            }
        }

    /**Test basic setting/getting of local variables.
     * It is assumed that the endPoint setting should be irrelevant.
     */
    public void testLocalVariableOps()
        throws Exception
        {
        final BasicVarMgr bvm = new BasicVarMgr(rnd.nextBoolean());

        // Use a local, TYPE_NUMBER, read/write variable...
        final SimpleVariableDefinition def =
            SystemVariables.ExhibitDataSimpleCache_CACHE_AVAIL_SPACE_PERCENT_USED;

        // Initial attempt to retrieve this should return null.
        assertNull("getVariable(X) for local X should return null until X has been set",
            bvm.getVariable(def));

        // Now try setting it to a few values at random.
        // It should always be possible to retrieve the value set.
        final Number testValues[] =
            {
                new Integer(rnd.nextInt()),
                new Long(rnd.nextLong()),
                new Float(rnd.nextFloat()),
                new Double(rnd.nextDouble()),
            };
        for(int i = testValues.length; --i >= 0; )
            {
            final Number n = testValues[i];
            final SimpleVariableValue svv = new SimpleVariableValue(def, n);
            bvm.setVariable(svv);
            assertEquals("Must be able to retrieve the local value that was set",
                n, bvm.getVariable(def).getValue());
            assertTrue("Must be able to retrieve the local value with getVariables()",
                checkValuePresent(bvm.getVariables(-1), svv));
            }
        }

    /**Test basic setting/getting of global variables.
     * This will have to be an end-point to set global variables.
     */
    public void testGlobalVariableOps()
        throws Exception
        {
        final BasicVarMgr bvm = new BasicVarMgr(true);

        // Use a global, TYPE_NUMBER, read/write variable...
        final SimpleVariableDefinition def =
            SystemVariables.ThroughputMonitorFilter_CLIENT_COUNT;

        // Just FYI...
        Main.getOut().println("[Sample SimpleVariableDefinition: " + def + ".]");

        // Initial attempt to retrieve this should return null.
        assertNull("getVariable(X) for global X should return null until X has been set",
            bvm.getVariable(def));

        // Create two system IDs.
        final InstanceID system1 = InstanceID.createInstanceID();
        final InstanceID system2 = InstanceID.createInstanceID();

        // Now try setting it to a few values at random.
        // It should always be possible to retrieve the value set.
        final Number testValues[] =
            {
                new Integer(rnd.nextInt()),
                new Long(rnd.nextLong()),
                new Float(rnd.nextFloat()),
                new Double(rnd.nextDouble()),
                new Integer(rnd.nextInt()),
            };
        for(int i = testValues.length; --i >= 0; )
            {
            final Number n = testValues[i];
            final SimpleVariableValue svvL = new SimpleVariableValue(def, n);
            final SimpleVariableValue svvG = svvL.put(
                (rnd.nextBoolean() ? system1 : system2), svvL);
            bvm.setVariable(svvG);
            assertEquals("Must be able to retrieve the global value that was set",
                n, bvm.getVariable(def).getValue());
            assertTrue("Must be able to retrieve the global value with getVariables()",
                checkValuePresent(bvm.getVariables(-1), svvL));
            }

        // Now explicitly set one value for each system,
        // and check that we can see both in the globalMap.
        final Integer v1 = new Integer(1);
        final SimpleVariableValue svvL1 = new SimpleVariableValue(def, v1);
        final SimpleVariableValue svvG1 = svvL1.put(system1, svvL1);
        final Integer v2 = new Integer(2);
        final SimpleVariableValue svvL2 = new SimpleVariableValue(def, v2);
        final SimpleVariableValue svvG2 = svvL2.put(system2, svvL2);
        bvm.setVariable(svvG1);
        bvm.setVariable(svvG2);
        final SimpleVariableValue svv12 = bvm.getVariable(def);
        assertEquals("globalMap must contain exactly two entries",
            2, svv12.getGlobalMap().size());
        assertEquals("system1 entry must appear in global map and be right",
            v1, svv12.getGlobalMap().get(system1).getValue());
        assertEquals("system2 entry must appear in global map and be right",
            v2, svv12.getGlobalMap().get(system2).getValue());

        // Just FYI...
        Main.getOut().println("[Sample SimpleVariableValue: " + svvG1 + ".]");
        Main.getOut().println("[Sample SimpleVariableValue full description: " + svvG2.getFullDescription() + ".]");
        }

    /**Test handling of TYPE_NUMBER values.
     * We want to ensure that all java.lang.XXX Number extensions are allowed,
     * but not any user-defined ones.
     */
    public void testTypeNumber()
        throws Exception
        {
        final BasicVarMgr bvm = new BasicVarMgr(rnd.nextBoolean());

        // Use a local, TYPE_NUMBER, read/write variable...
        final SimpleVariableDefinition def =
            SystemVariables.ExhibitDataSimpleCache_CACHE_AVAIL_SPACE_PERCENT_USED;

        // Good values which should all be settable...

        // Now try setting it to a few values at random.
        // It should always be possible to retrieve the value set.
        final Number goodValues[] =
            {
                new Integer(rnd.nextInt()),
                new Long(rnd.nextLong()),
                new Float(rnd.nextFloat()),
                new Double(rnd.nextDouble()),
                Byte.valueOf((byte) rnd.nextInt()),
                new Short((short) rnd.nextInt()),
            };
        for(int i = goodValues.length; --i >= 0; )
            {
            final Number n = goodValues[i];
            final SimpleVariableValue svv = new SimpleVariableValue(def, n);
            bvm.setVariable(svv);
            assertEquals("Must be able to set and retrieve JVM Number impl",
                n, bvm.getVariable(def).getValue());
            }

        // Make sure that we can set it to null too.
        final SimpleVariableValue svvNull = new SimpleVariableValue(def, null);
        bvm.setVariable(svvNull);
        assertNull("Must be able to set NUMBER value to null",
            bvm.getVariable(def).getValue());

        // Invent a bad implementation of Number,
        // which should be rejected.
        final Number bad1 = new Number(){
            public int intValue()
                {
                return 0;
                }

            public long longValue()
                {
                return 0;
                }

            public float floatValue()
                {
                return 0;
                }

            public double doubleValue()
                {
                return 0;
                }

            /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
            private static final long serialVersionUID = -8016942945220968907L;
            };
        // Take several values which TYPE_NUMBER variable should not allow
        // even to be constructed.
        final Object badVals[] = { bad1, "hello", new Object(), rnd };
        for(int i = badVals.length; --i >= 0; )
            {
            try
                {
                new SimpleVariableValue(def, badVals[i]);
                fail("Should not allow user-defined Number instance as value");
                }
            catch(IllegalArgumentException e)
                {
                // Good: rejected dodgy value...
                }
            }
        }



    /**Simple final pipeline stage (an endpoint) variable values.
     * Note that all data operations return null;
     * this class is for var testing purposes only (thus is not public).
     */
    static final class BasicVarMgrEndpoint implements SimpleExhibitPipelineIF
        {
        public ExhibitStaticAttr getStaticAttr(final String name)
//            throws IOException
            {
            return null;
            }

        public void getRawFile(final ByteBuffer buf,
                               final String exhibitName, final int position, final boolean dontCache)
//            throws IOException
            {
            }

        public AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
//            throws IOException
            {
            return null;
            }

        public AllExhibitProperties getAllExhibitProperties(final long oldHash)
//            throws IOException
            {
            return null;
            }

        public GenProps getGenProps(final long oldStamp)
//            throws IOException
            {
            return null;
            }

        public Properties getGenSecProps(final long oldStamp)
//            throws IOException
            {
            return null;
            }

        public ExhibitThumbnails getThumbnails(final String name, final boolean create)
//            throws IOException
            {
            return null;
            }

        public void setVariable(final SimpleVariableValue newValue)
            throws IOException,
            UnsupportedOperationException
            {
            vars.setVariable(newValue);
            }

        public int setVariables(final SimpleVariableValue[] newValues)
            throws IOException
            {
            return(vars.setVariables(newValues));
            }

        public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
//            throws IOException
            {
            return(vars.getVariable(var));
            }

        public SimpleVariableValue[] getVariables(final long changedSince)
//            throws IOException
            {
            return(vars.getVariables(changedSince));
            }

        public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                                final EventPeriod intervalSelector,
                                                final boolean current)
            {
            return(vars.getEventValue(def, intervalSelector, current));
            }

        public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                                   final EventPeriod intervalSelector,
                                                   final long intervalNumber,
                                                   final BitSet whichValues)
            throws IOException
            {
            return(vars.getEventValues(def, intervalSelector, intervalNumber, whichValues));
            }

        public void syncVariables(final boolean force)
//            throws IOException
            {
            }

        /**Get requested Properties selected by key and versionID.
         * Fetches a Properties set unconditionally (versionID == -1)
         * else if the versionID presented is not current.
         *
         * @param key  selector (with possible embedded sub-key)
         *     for desired properties set; never null
         * @param versionID  if -1 then map is always returned if available,
         *     else must be non-negative and null is returned if the versionID
         *     presented matches that of the current version
         *     (ie if the caller has presumably got the up-to-date version);
         *     may be a timestamp or a hash or other value,
         *     and by convention is zero only for an empty properties set
         *
         * @return null, or Properties map guaranteed to contain only
         *     String keys and values
         */
        public java.util.Properties getProperties(final PropsKey key,
                                                  final long versionID)
            throws IOException
            {
            throw new IOException("NOT IMPLEMENTED");
            }

        public void poll(final GenProps gp)
//            throws IOException
            {
            }

        /**Does nothing in this implementation. */
        public void destroy()
            {
            }

        /**Variable store; must be an end-point */
        private final BasicVarMgr vars = new BasicVarMgr(true);
        }


    /**Simple pipeline stage that caches variable values.
     * This can be optionally write-through.
     * <p>
     * Note that all data operations return null;
     * this class is for var testing purposes only (thus is not public).
     */
    static final class TestVarCacheStage extends SimpleExhibitPipelineFilter
        {
        /**Construct test variable-caching stage.
         *
         * @param upstream  upstream stage of pipeline; non-null
         * @param writeThrough  if true, all set operations "write-through"
         *     this cahce stage immediately...
         */
        TestVarCacheStage(final SimpleExhibitPipelineIF upstream,
                   final boolean writeThrough)
            {
            super(upstream);
            vars = new PipelineVarMgr(upstream, writeThrough);
            }

        public void setVariable(final SimpleVariableValue newValue)
            throws IOException,
            UnsupportedOperationException
            {
            vars.setVariable(newValue);
            }

        public int setVariables(final SimpleVariableValue[] newValues)
            throws IOException
            {
            return(vars.setVariables(newValues));
            }

        public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
//            throws IOException
            {
            return(vars.getVariable(var));
            }

        public SimpleVariableValue[] getVariables(final long changedSince)
//            throws IOException
            {
            return(vars.getVariables(changedSince));
            }

        public void syncVariables(final boolean force)
            throws IOException
            {
            vars.syncVariables(force);
            }

        public void poll(final GenProps gp)
            throws IOException
            {
            // Never explicitly force a full refresh...
            try { vars.syncVariables(false); }
            finally { source.poll(gp); }
            }

        /**Variable manager. */
        private final PipelineVarMgr vars;
        };

    /**Check BasicVarMgr embedded at the end of a pipeline.
     * We're prepared to spend some time running this
     * (though in fact it should be fast)
     * since it is so basic to correct behaviour.
     */
    public void testBasicVarMgrAsPipelineEndpoint()
        throws Exception
        {
        // Tunnel endpoint (wraps a BasicVarMar).
        final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

        // Now run our battery of pipeline tests...
        final SimpleVariablePipelineIF[] pipeline = new SimpleVariablePipelineIF[]{ep};
        Main.getOut().println("Starting tests on simple pipeline...");
        singlePipelineTest(pipeline,
            System.currentTimeMillis() + 120000L); // Test for 2m max.
        }

    /**Check vanilla SimpleExhibitPipelineFilter embedded in a pipeline.
     * This is only checking the variable-handling aspects.
     */
    public void testSimpleExhibitPipelineFilterVariableHandling()
        throws Exception
        {
        // Tunnel endpoint (wraps a BasicVarMar).
        final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

        // Simple vanilla SimpleExhibitPipelineFilter instance.
        final SimpleExhibitPipelineFilter sepf =
            new SimpleExhibitPipelineFilter(ep);

        // Now run our battery of pipeline tests...
        final SimpleVariablePipelineIF[] pipeline =
            new SimpleVariablePipelineIF[]{sepf, ep};
        singlePipelineTest(pipeline,
            System.currentTimeMillis() + 5000L); // Test for 5s max.
        }

    /**Test PipelineVarMgr in pipeline in write-through mode.
     * Checks that we can build a simple pipeline with:
     * <ul>
     * <li>a BasicVarMgr endpoint.
     * <li>a PipelineVarMgr (itself an extended BasicVarMgr) upstream.
     * <ul>
     * <p>
     * Write-though mode is faster and "safer"
     * but may incur higher costs, eg sending updates over a tunnel.
     */
    public void testPipelineVarMgrSimpleWriteThrough()
        throws Exception
        {
        checkShortPipelineVarMgr(true);
        }

    /**Test PipelineVarMgr in pipeline in non-write-through mode.
     * Checks that we can build a simple pipeline with:
     * <ul>
     * <li>a BasicVarMgr endpoint.
     * <li>a PipelineVarMgr (itself an extended BasicVarMgr) upstream.
     * <ul>
     * <p>
     * Non-write-though mode groups writes together
     * and eliminates redundant writes for efficiency,
     * but delays writes and thus information propagation.
     */
    public void testPipelineVarMgrSimpleNonWriteThrough()
        throws Exception
        {
        checkShortPipelineVarMgr(false);
        }

    /**Check a short pipeline with a single PipelineVarMgr stage.
     * This can be run in write-through or non-write-through mode.
     * <p>
     * We run this stand-alonw because as of 20040808
     * the non-write-though implementation was buggy/lossy.
     *
     * @param writeThrough
     * @throws InterruptedException
     * @throws IOException
     */
    private void checkShortPipelineVarMgr(final boolean writeThrough)
        throws InterruptedException, IOException
        {
        // Tunnel endpoint (wraps a BasicVarMar).
        final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

        // Make several cacheing stages...
        // Randomly make at least one stage write-through and another other not.
        final TestVarCacheStage c1 = new TestVarCacheStage(ep, writeThrough);

        // Get run poll() for cacheStage so it can sync() its vars.
        t.scheduleAtFixedRate(new TimerTask(){
            public void run()
                {
                try { c1.poll(new GenProps()); }
                catch(Exception e) { e.printStackTrace(Main.getErr()); }
                }
            },
            rnd.nextInt(2000), 100L); // Run poll() every 0.1s, after a delay.

        // Now run our battery of pipeline tests...
        final SimpleExhibitPipelineIF[] pipeline =
            new SimpleExhibitPipelineIF[]{c1, ep};

        // Do extra tests of variable set/get (covered in main tests too)
        // looking for race condition we have seen in past.
        Main.getOut().println("Starting pre-tests on short cache pipeline, writeThrough="+writeThrough+"...");
        checkSimpleSetGet(pipeline,
            System.currentTimeMillis() + 10000L); // Test for 10s max.

        Main.getOut().println("Starting tests on short cache pipeline, writeThrough="+writeThrough+"...");
        singlePipelineTest(pipeline,
            System.currentTimeMillis() + 60000L); // Test for 1m max.
        }

    /**Test several PipelineVarMgr stages in a long pipeline.
     * Checks that we can build a simple pipeline with:
     * <ul>
     * <li>a BasicVarMgr endpoint.
     * <li>Several PipelineVarMgr stages upstream,
     *     randomly write-through or not.
     * <ul>
     * <p>
     * Normally it would not be sensible to stack several cache stages
     * like this because of latency problems, so this is somewhat harsh.
     */
    public void testPipelineVarMgrLong()
        throws Exception
        {
        // Tunnel endpoint (wraps a BasicVarMar).
        final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

        final boolean wt = rnd.nextBoolean();

        // Make several cacheing stages...
        // Randomly make at least one stage write-through and another other not.
        final TestVarCacheStage c1 = new TestVarCacheStage(ep, wt);
        final TestVarCacheStage c2 = new TestVarCacheStage(c1, !wt);
        final TestVarCacheStage c3 = new TestVarCacheStage(c2, rnd.nextBoolean());
        final TestVarCacheStage c4 = new TestVarCacheStage(c3, rnd.nextBoolean());
        final TestVarCacheStage c5 = new TestVarCacheStage(c4, rnd.nextBoolean());

        // Get run poll() for cacheStage so it can sync() its vars.
        t.scheduleAtFixedRate(new TimerTask(){
            public void run()
                {
                try { c5.poll(new GenProps()); }
                catch(Exception e) { e.printStackTrace(Main.getErr()); }
                }
            },
            rnd.nextInt(2000), 100L); // Run poll() every 0.1s, after a delay.

        // Now run our battery of pipeline tests...
        final SimpleExhibitPipelineIF[] pipeline =
            new SimpleExhibitPipelineIF[]{c5, c4, c3, c2, c1, ep};
        Main.getOut().println("Starting tests on long pipeline...");
        singlePipelineTest(pipeline,
            System.currentTimeMillis() + 120000L); // Test for 2m max.
        }


    /**Returns true iff the specified variable is present in the array with the specified value.
     * Checks name, type and value.
     *
     * @param svv  variable/value to check for; must not be null
     */
    public static boolean checkValuePresent(final SimpleVariableValue[] vars,
                                            final SimpleVariableValue svv)
        {
        if(svv == null)
            { throw new IllegalArgumentException(); }

        for(int i = vars.length; --i >= 0; )
            {
            if(svv.equals(vars[i]))
                { return(true); }
            }
        return(false);
        }

    /**Driver routine to test variable handling in pipelines.
     * This makes sure that values are correctly propagated,
     * globals correctly merged, etc.
     * <p>
     * The pipeline order is downstream-first, end-point-last,
     * eg the persistent store or tunnel for a pipeline would
     * be at the highest-numbered index,
     * and the normal access point for the tunnel
     * (eg as exposed by the DataSourceBean) is at index 0.
     * <p>
     * Takes in-order lists of taps/entrypoints on a data pipeline,
     * the first (two) being separate slaves (before a tunnel)
     * and the last being in a master (after a tunnel).
     * All but one of the lists can be null (but never empty).
     * <p>
     * Internal correctness tests will be run on each of the constituent
     * slave/master pipelines.
     * <p>
     * The parameter arrays must not be altered while the tests are running.
     * <p>
     * This checks that:
     * <ul>
     * <li>System IDs are unique to each side of each tunnel.
     * <li>System IDs are available everywhere in the tunnel.
     * <li>System IDs are not writable in the tunnel.
     * <li>Non-existent variables cannot be set in the tunnel.
     * <li>Cache stages propagate values within allowed latencies.
     * <li>Locals never a have a (non-null) globalMap.
     * <li>Locals are not propagated across a tunnel.
     * <li>Globals are propagated both ways across a tunnel.
     * <li>Globals are propagated between slaves.
     * <li>Globals are correctly merged.
     * <li>Multiple settings of the same variable happen in order.
     * <li>TODO: That at least the top of each pipeline is thread-safe
     * </ul>
     *
     * @param slave1 pipeline for slave 1; null or non-empty
     * @param slave2 pipeline for slave 2; null or non-empty
     * @param master pipeline for master; null or non-empty
     *
     * @throws IOException  if an inappropriate IOException is thrown by
     *     the pipeline under test
     */
    public static void pipelineVariableTest(final SimpleVariablePipelineIF slave1[],
                                            final SimpleVariablePipelineIF slave2[],
                                            final SimpleVariablePipelineIF master[],
                                            final long stopBy)
        throws IOException,
               InterruptedException
        {
        // Make a list of just the non-null pipelines...
        final List<BasicVarMgrInterface[]> livePipes = new ArrayList<BasicVarMgrInterface[]>(3);
        if(slave1 != null) { livePipes.add(slave1); }
        if(slave2 != null) { livePipes.add(slave2); }
        if(master != null) { livePipes.add(master); }

        // Start with independent tests on each pipeline.
        final SimpleVariablePipelineIF pipelines[][] =
            { slave1, slave2, master };
        final Map<InstanceID,SimpleVariablePipelineIF[]> systemIDs = new HashMap<InstanceID, SimpleVariablePipelineIF[]>(); // From InstanceID --> pipeline.
        final int nPipes = pipelines.length;
//        for(int i = nPipes; --i >= 0; )
//            {
//            if(pipelines[i] != null)
//                {
//                Main.getOut().println("[pipelineVariableTest(): testing pipeline "+i+"...]");
//                if(null != systemIDs.put(singlePipelineTest(pipelines[i], stopBy), pipelines[i]))
//                    {
//                    fail("pipelines had duplicate system IDs");
//                    }
//                }
//            }

        // We can explicitly test propagation of globals between systems
        // as long as we have the master pipeline and at least one slave.
        // TODO: all tests to do with propagation of globals, etc...
        // TODO: demonstrate that locals are never propagated...
        if((master != null) &&
           ((slave1 != null) || (slave2 != null)))
            {
            Main.getOut().println("[pipelineVariableTest(): testing global propagation and merge...]");

            // Take our set of distinct (writable) local variables,
            // for each, pick one non-null pipeline
            // and set it at the top of the pipeline,
            // then check that none of those local values
            // appears in any of the other pipelines.
            // We wait for up to the maximum propagation latency
            // to make sure that none of the values have propagated.
            final SimpleVariableDefinition localWritableTestVars[] =
                {
                    SystemVariables.TEST_NUMBER_LOCAL,
                    SystemVariables.TEST_NUMBER_LOCAL2,
                    SystemVariables.TEST_STRING_LOCAL,
                };
            // The values that we have set for each local variable.
            final SimpleVariableValue values[] =
                new SimpleVariableValue[localWritableTestVars.length];
            // Record which test variable(s) we wrote where...
            // We only set values for non-null pipeline stages.
            final List<SimpleVariableDefinition> what[] =
                new List[nPipes];
            // Now set up the local variables before we do our other tests...
            // This gives values time to (erroneously) propagate in error.
            _setUpLocalVars(localWritableTestVars, pipelines, nPipes, what, values);

            // TEST GLOBALS!
            // Check that if I set a test global in each pipeline,
            // then in all pipelines I can see a merged globalMap
            // within the allowed latency.
            _checkGlobalMerge(livePipes, stopBy);

            // Now make sure that the locals did not propagate...
            _testLocalNonPropagation(values, nPipes, pipelines, what);
            }

        // Now re-run most of the per-pipeline tests
        // to make sure that nothing has broken,
        // TODO: but maybe use a random ordering this time.
        systemIDs.clear();
        for(int i = nPipes; --i >= 0; )
            {
            if(pipelines[i] != null)
                {
                Main.getOut().println("[pipelineVariableTest(): retesting pipeline "+i+"...]");
                if(null != systemIDs.put(singlePipelineTest(pipelines[i], stopBy), pipelines[i]))
                    {
                    fail("pipelines had duplicate system IDs");
                    }
                }
            }
        }

    /**Check global variable propagation and merge.
     * Check that if I set a test global in each pipeline,
     * then in all pipelines I can see a merged globalMap
     * within the allowed latency.
     * <p>
     * This does not alter its parameters, though sets/gets values in them.
     *
     * @param livePipes  list of pipelines of participating systems;
     *     no pipelines are null or empty
     * @param stopBy  time to attempt to stop by
     */
    private static void _checkGlobalMerge(final List<BasicVarMgrInterface[]>/*<SimpleExhibitPipelineIF[]>*/ livePipes,
                                         final long stopBy)
        throws InterruptedException
        {
        // Test set of global variables to use.
        final SimpleVariableDefinition testGlobals[] =
            {
                SystemVariables.TEST_NUMBER_GLOBAL,
                SystemVariables.TEST_NUMBER_GLOBAL2,
                SystemVariables.TEST_STRING_GLOBAL,
            };
        if(RANDOMISE_VARIABLE_TEST_ORDER)
            {
            // Randomise the ordering...
            Collections.shuffle(Arrays.asList(testGlobals));
            }

        for(int g = testGlobals.length; --g >= 0; )
            {
            final SimpleVariableDefinition def = testGlobals[g];

            // Values that we successfully assigned for this variable,
            // at most one per pipeline.
            // All of these should appear in each globalMap...
            final List<SimpleVariableValue> values = new ArrayList<SimpleVariableValue>();

            // For each live pipeline, try to set a value for this variable.
            for(final Iterator<BasicVarMgrInterface[]> it = livePipes.iterator(); it.hasNext(); )
                {
                final BasicVarMgrInterface[] pipeline =
                        it.next();
                final BasicVarMgrInterface pipelineTop = pipeline[0];

                // Chose a random value for each variable
                // (of an appropriate type),
                // record it,
                // and set it,
                // (and check that it got set here).
                // It is acceptable for the set/get to be vetoed with
                // IOException or UnsupportedOperationException.
                final SimpleVariableValue svv = makeRandomSimpleVariableValue(def);
                try
                    {
                    // Set the value at the top of the pipeline...
                    pipelineTop.setVariable(svv);
                    // Once the set is allowed, note what we set...
                    values.add(svv);
                    // Verify that it really did get set...
                    assertEquals("Global variable set accepted but did not work",
                        svv, pipelineTop.getVariable(def));
                    }
                catch(IOException e)
                    {
                    // Not capable of local set/get right now...
                    }
                catch(UnsupportedOperationException e)
                    {
                    // Not capable of local set/get...
                    }
                }

            // Now try and retrieve the merged globalMap for this variable,
            // from all live pipelines,
            // waiting up to the maximum allowed latency.
            final long now = System.currentTimeMillis();
            final long warnAfter = now + 10 * 1000L;
            final long retrieveBy = now +
                SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS;

            // For each live pipeline, try to set a value for this variable.
            for(final Iterator<BasicVarMgrInterface[]> it = livePipes.iterator(); it.hasNext(); )
                {
                final BasicVarMgrInterface[] pipeline =
                        it.next();
                final BasicVarMgrInterface pipelineTop = pipeline[0];

                try
                    {
                    // Loop until we get the globalMap we expect,
                    // or we run out of time...
                    int waitTime = 10 + rnd.nextInt(100);
                    for( ; ; )
                        {
                        final SimpleVariableValue svv =
                            pipelineTop.getVariable(def);

//Main.getErr().println("  For def="+def+", svv=" + svv);
//Main.getErr().println("    Values being waited for: " + new ArrayList(values));

                        // We can get some sort of value...
                        if(svv != null)
                            {
                            // And it has a globalMap...
                            final Map<InstanceID,SimpleVariableValue> globalMap = svv.getGlobalMap();
                            if(globalMap != null)
                                {
// Main.getErr().println("    Value retrieved="+svv.getFullDescription());

                                // Then make sure that they are all there...
                                if(globalMap.values().containsAll(values))
                                    {
                                    // Yes!  We found all our values!
                                    break;
                                    }
                                }
                            }

                        final long n = System.currentTimeMillis();
                        if(n > retrieveBy)
                            { fail("Could not see global propagated to remote pipeline for " + def); }
                        if(n > warnAfter)
                            {
                            Main.getOut().println("[Waiting to see global propagated to remote pipeline for "+def+"...]");
                            Main.getOut().println("    [Values being waited for: " + new ArrayList<SimpleVariableValue>(values)+"]");
                            if(svv != null) { Main.getOut().println("    [Value retrieved="+svv.getFullDescription()+"]"); }
                            }
                        Thread.sleep(waitTime *= 2); // Wait a while...
                        }
                    }
                catch(IOException e)
                    {
                    // Cannot get this value right now...
                    }
                catch(UnsupportedOperationException e)
                    {
                    // Cannot get this value...
                    }

                // IFF we are randomising variable-test order
                // then we can bucket out when we run out of time
                // since we will eventually test all possible values.
                if(RANDOMISE_VARIABLE_TEST_ORDER)
                    {
                    if(!notAfter(stopBy))
                        { return; }
                    }
                }
            }
        }

    /**Given a variable definition, make a random value for it; never null.
     * This attempts to return values distributed across the entire
     * legitimate range for a given type;
     * where that type has a large range then the values generated should
     * be unique within a given test run
     * (unless the random number source behaves poorly).
     * <p>
     * Conversely, for restricted types such as TYPE_NONE,
     * little or no practical variation is possible.
     * <p>
     * Non-local values are generated without a globalMap;
     * the results of several calls to this function can be merged
     * with put() to generate a non-empty globalMap if desired.
     * <p>
     * Package-visible so as to be available to other test routines.
     *
     * @param def  defintion for which value is wanted; never null
     * @return  random value for that definition.
     */
    static SimpleVariableValue makeRandomSimpleVariableValue(
        final SimpleVariableDefinition def)
        {
        if(def == null)
            { throw new IllegalArgumentException(); }

        // Some of the time just return a null value...
        if(rnd.nextInt(3) == 0)
            { return(new SimpleVariableValue(def, null)); }

        // Try to generate a type-specific, non-null value if possible.
        switch(def.getType())
            {
            case SimpleVariableDefinition.TYPE_NONE:
                return(new SimpleVariableValue(def, null));

            case SimpleVariableDefinition.TYPE_IID:
                return(new SimpleVariableValue(def, InstanceID.createInstanceID()));

            case SimpleVariableDefinition.TYPE_NUMBER:
                // Randomly pick a Number data type.
                // Fall back to a good/large/simple Number type.
                switch(rnd.nextInt(6))
                    {
                    case 1:
                        return(new SimpleVariableValue(def,
                            new Integer(rnd.nextInt())));
                    case 2:
                        return(new SimpleVariableValue(def,
                            new Short((short) rnd.nextInt())));
                    case 3:
                        return(new SimpleVariableValue(def,
                            new Byte((byte) rnd.nextInt())));
                    case 4:
                        return(new SimpleVariableValue(def,
                            new Float((float) rnd.nextDouble())));
                    case 5:
                        return(new SimpleVariableValue(def,
                            new Double((short) rnd.nextGaussian())));
                    }
                // Default is to return a Long.
                return(new SimpleVariableValue(def,
                    new Long(rnd.nextLong())));

            case SimpleVariableDefinition.TYPE_STRING:
                return(new SimpleVariableValue(def,
                    "random test string ... " + String.valueOf(rnd.nextLong())));

            default: // Should not happen: we need to fix the code!
                throw new IllegalStateException("Cannot generate variable of this type: " + def);
            }
        }

    /**Run after _setUpLocalVars() to ensure that locals did NOT propagate between pipelines.
     *
     * @param values
     * @param nPipes
     * @param pipelines
     * @param what
     */
    private static void _testLocalNonPropagation(final SimpleVariableValue[] values,
                                                 final int nPipes,
                                                 final BasicVarMgrInterface[][] pipelines,
                                                 final List[] what)
        {
        // Now, after we know that variables have had time to propagate,
        // we make sure that local variables have not done so,
        // but that any original values we set are still intact.
        // We'll check the state of each of the variables
        // on each of the pipelines...
        for(int i = values.length; --i >= 0; )
            {
            final SimpleVariableValue val = values[i];
            final SimpleVariableDefinition def = val.getDef();
            for(int j = nPipes; --j >= 0; )
                {
                // Skip over null (absent) pipelines...
                if(pipelines[j] == null)
                    { continue; }

                final BasicVarMgrInterface pipelineTop =
                    pipelines[j][0];

                // The given variable should be set to our value
                // ONLY on the pipeline that we set it on...
                final boolean shouldBeSetHere =
                    what[j].contains(def);
                try
                    {
                    final SimpleVariableValue svv = pipelineTop.getVariable(def);
                    if(shouldBeSetHere)
                        {
                        assertEquals("Local variable should still be set on the pipeline we set it",
                            val, svv);
                        }
                    else
                        {
                        assertFalse("Local variable should NOT have propagated to another pipeline",
                            val.equals(svv));
                        }
                    }
                catch(IOException e)
                    {
                    // Not capable of local get right now...
                    }
                catch(UnsupportedOperationException e)
                    {
                    // Not capable of local get...
                    }
                }
            }
        }

    /**Set random variables on random pipes for testing later.
     * This records which variables and values were set on each pipeline
     * for later testing by _testLocalNonPropagation() later.
     * <p>
     * There must be at least a master and one slave for this to be
     * meaningful.
     *
     * @param localWritableTestVars
     * @param pipelines
     * @param nPipes
     * @param what
     * @param values
     */
    private static void _setUpLocalVars(final SimpleVariableDefinition[] localWritableTestVars,
                                        final BasicVarMgrInterface[][] pipelines,
                                        final int nPipes,
                                        final List<SimpleVariableDefinition>[] what,
                                        final SimpleVariableValue[] values)
        {
        // Set empty lists for each pipeline of what variables were set for it.
        for(int i = nPipes; --i >= 0; )
            { what[i] = new ArrayList<SimpleVariableDefinition>(localWritableTestVars.length); }

        // Now randomly choose one of the pipelines
        // to set each test variable on...
        for(int i = localWritableTestVars.length; --i >= 0; )
            {
            final SimpleVariableDefinition def =
                localWritableTestVars[i];

            // Chose a pipeline to set the variable in.
            int which;
            // Spin until we find a non-null pipeline...
            while(pipelines[which = rnd.nextInt(nPipes)] == null)
                { }
            // Note which variable is set in which pipeline.
            what[which].add(def);

            // Chose a random value for each variable
            // (of an appropriate type),
            // record it,
            // and set it,
            // (and check that it got set here).
            // It is acceptable for the set/get to be vetoed with
            // IOException or UnsupportedOperationException.
            switch(def.getType())
                {
                case SimpleVariableDefinition.TYPE_NUMBER:
                    values[i] = new SimpleVariableValue(def,
                        new Long(rnd.nextLong()));
                    break;
                case SimpleVariableDefinition.TYPE_STRING:
                    values[i] = new SimpleVariableValue(def,
                        String.valueOf(rnd.nextGaussian()));
                    break;
                default:
                    throw new Error("Cannot support local variable of this type in this test: " + def);
                }
            try
                {
                final BasicVarMgrInterface pipelineTop = pipelines[which][0];
                pipelineTop.setVariable(values[i]);
                assertEquals("Local variable set accepted but did not work",
                    values[i], pipelineTop.getVariable(def));
                }
            catch(IOException e)
                {
                // Not capable of local set/get right now...
                }
            catch(UnsupportedOperationException e)
                {
                // Not capable of local set/get...
                }
            }
        }

    /**Run variable tests on one pipeline ans returns the InstanceID; never null.
     * Runs a whole suite of tests on one individual pipeline within one system.
     * <p>
     * Used by pipelineVariableTest().
     * <p>
     * Throws an exception, assertion or fail()s upon discovering a problem.
     * <p>
     * Tests that:
     * <ul>
     * <li>The system ID can be fetched from any tap on the pipeline,
     *     at any time, in any order (a random sampling will be done),
     *     and that that ID is consistent.  No IOException can be thrown.
     * <li>Local and global variables can be set at any stage of a pipeline
     *     with the value immediately retrievable there or the set rejected with
     *     UnsupportedOperationException;
     *     check with setVariable() and setVariables().
     * <li>Non-existent variables cannot be set in the tunnel.
     * <li>Values successfully written anywhere in the pipeline are available
     *     everywhere else within the specified maximum latency.
     * <li>Variable set operations are not visibly reordered,
     *     eg that cacheing and tunnel (etc) stages do not cause re-ordering.
     * </ul>
     * <p>
     * The pipeline order is downstream-first, end-point-last,
     * eg the persistent store or tunnel for a pipeline would
     * be at the highest-numbered index,
     * and the normal access point for the tunnel
     * (eg as exposed by the DataSourceBean) is at index 0.
     *
     * @return  the system ID for the pipeline.
     *
     * @param pipeline  one or more stages/taps on a pipeline in order,
     *     downstream first,
     *     never null nor empty nor containing null or duplicate entries
     * @param stopBy  stop testing by or soon after this time,
     *     having done at least one of each type of test
     *
     * @throws IOException  if an inappropriate IOException is thrown by
     *     the pipeline under test
     */
    public static InstanceID singlePipelineTest(
                                        final SimpleVariablePipelineIF pipeline[],
                                        final long stopBy)
        throws IOException,
               InterruptedException
        {
        if((pipeline == null) || (pipeline.length == 0))
            { throw new IllegalArgumentException(); }

//Main.getOut().println("[singlePipelineTest(): starting...]");

        // Force the pipeline to be quiet.
        try{ pipeline[0].syncVariables(true); }
        catch(IOException e) { /* Ignore legitimate I/O issues. */ }

//Main.getOut().println("[singlePipelineTest(): pipeline quietened...]");

        // Check handling of system ID in pipeline
        // (and get the value for this pipeline).
        final InstanceID iid = checkSystemIDHandling(pipeline, stopBy);

//Main.getOut().println("[singlePipelineTest(): SystemID handling checked...]");

        // Test simple get and set operations on locals and globals.
        checkSimpleSetGet(pipeline, stopBy);

//Main.getOut().println("[singlePipelineTest(): simple set/get checked...]");

        // Ensure that illegal ops are vetoed, and in the right way.
        checkIllegalGetSetOpsVetoedCorrectly(pipeline, stopBy);

//Main.getOut().println("[singlePipelineTest(): vetoing of illegal ops checked...]");

        // Simply check that we can retrieve the "all" bucket
        // (which should not be authoritative)
        // or at least not blow up when trying to do so.
        final BitSet whichValues = new BitSet(1);
        whichValues.set(0);
        for(int i = pipeline.length; --i >= 0; )
            {
            try
                {
                // Try getting first slot with real/null BitSet,
                // which should be equivalent.
                final EventVariableValue evvs[] =
                    pipeline[0].getEventValues(SystemVariables.TEST_STRING_GLOBAL_EVENT,
                                               EventPeriod.VLONG,
                                               0,
                                               rnd.nextBoolean() ? null : whichValues);
                assert(evvs != null);
                if(evvs.length > 0)
                    {
                    if(evvs[0] != null)
                        { assert(!evvs[0].isAuthoritative()); }
                    }
                }
            // Ignore problems if master is not available...
            catch(final PGMasterNotInServiceException e) { }
            }

        // Force the pipeline to be quiet.
        try { pipeline[0].syncVariables(true); }
        catch(IOException e) { /* Ignore legitimate I/O issues. */ }

//Main.getOut().println("[singlePipelineTest(): done; pipeline quietened.]");

        return(iid);
        }

    /**If true, randomise order of variable tests (and skip some if we are short of time).
     * This can only be done if we quieten the pipelines
     * between tests so that tests don't conflict with one another
     * by having updates from one test still bouncing around during the next.
     * <p>
     * This allows us to do as few as one test each time,
     * on the grounds that we will test the entire range over time.
     */
    private static final boolean RANDOMISE_VARIABLE_TEST_ORDER = true;

    /**Test set and get of globals and locals.
     * Tests simple set/get operations work and preserve timestamps,
     * and that values do not get "lost" due to races, cache errors, etc.
     * <p>
     * Also tests that variable set operations are not visibly reordered,
     * eg that cacheing and tunnel (etc) staged do not cause re-ordering.
     * <p>
     * Tests that locals (that we set) never get a (non-null) globalMap.
     *
     * @param pipeline  pipeline; non-null, non-zero-length
     * @param stopBy  time to stop running tests by, if possible
     */
    private static void checkSimpleSetGet(final SimpleVariablePipelineIF[] pipeline,
                                          final long stopBy)
        throws InterruptedException
        {
        if((pipeline == null) || (pipeline.length < 1))
            { throw new IllegalArgumentException(); }
        final int plen = pipeline.length;

        // Test that we can set and get globals and locals anywhere,
        // or else have the updates rejected with an
        // UnsupportedOperationException (or possibly an IOException).
        //
        // Also test that we can read back any values set from everywhere
        // else in the pipeline within the agreed maximum latency...
        //
        // Unless we force the pipeline to go quiet between tests,
        // we cannot test the same variable twice in succession,
        // since old updates may still be richoceting around the system,
        // so we must interleave tests of different variables.
        // Further, we cannot test one global variable twice in succession
        // in case we are not operating on the master's pipeline
        // and thus interleaving a local-variable test
        // does not guarantee that updates from a previous global
        // have died down.
        //
        // So, we can only randomise the test order
        // if we wait for the pipeline to be quiet before and after...
        //
        // TODO: add non-Number test variables
        final SimpleVariableDefinition[] writeTestVars =
            {
                SystemVariables.TEST_NUMBER_GLOBAL,
                SystemVariables.TEST_NUMBER_GLOBAL2,
                SystemVariables.TEST_NUMBER_LOCAL,
                SystemVariables.TEST_NUMBER_LOCAL2,
            };
        if(RANDOMISE_VARIABLE_TEST_ORDER)
            { Collections.shuffle(Arrays.asList(writeTestVars)); }
        try
            {
            // Try a few times for each stage...
            // Run quite a few tests to catch intermittent problems.
            final int maxTests = 5 + plen/5;
            for(int j = 0; j < maxTests; ++j)
                {
                // Make sure we never try to use the same variable twice
                // in succession else previous changes may still be
                // reverberating around the system...
                for(int i = 0; i < writeTestVars.length; ++i)
                    {
                    // Once we've tested all the variables at least once,
                    // we can force syncVariables(true) to speed things up...
                    // If we are randomising the variable test order,
                    // then we can get away with syncing all but the first test.
                    final boolean doSync = (j > 0) ||
                        (RANDOMISE_VARIABLE_TEST_ORDER && (i > 0));

                    final SimpleVariableDefinition def = writeTestVars[i];
                    // Pick a pipeline stage at random;
                    // all should behave reasonably
                    // even if not all functionality is available.
                    final int stageN = rnd.nextInt(plen);
                    final BasicVarMgrInterface whichStage = pipeline[stageN];

                    // Use setVariable()...
                    // Create new value...
                    final SimpleVariableValue svv1 = new SimpleVariableValue(
                        def, new Integer(rnd.nextInt()));
                    try
                        {
                        // Check that we can do zero-or-more
                        // previous and distinct sets of the variable,
                        // and that these values "lose"
                        // and that the final set operation prevails.
                        final int preSets = rnd.nextInt(3);
                        for(int p = preSets; --p >= 0; )
                            {
                            whichStage.setVariable(new SimpleVariableValue(def,
                                new Integer(rnd.nextInt())));
                            }

                        // Now do the set that we care about...
                        // Randomly check that setVariable() or setVariables()
                        // work equivalently.
                        if(rnd.nextBoolean())
                            { whichStage.setVariable(svv1); }
                        else
                            {
                            final List<SimpleVariableValue> vars = new ArrayList<SimpleVariableValue>();
                            for(int p = preSets; --p >= 0; )
                                {
                                vars.add(new SimpleVariableValue(def,
                                    new Integer(rnd.nextInt())));
                                }
                            // Put our wanted value last so that it should win...
                            vars.add(svv1);
                            final SimpleVariableValue vs[] = new SimpleVariableValue[vars.size()];
                            vars.toArray(vs);
                            if(0 == whichStage.setVariables(vs))
                                { throw new UnsupportedOperationException("no variable were set; probably OK"); }
                            }

                        // Make sure that we can get our value back straight away.
                        assertEquals("Must be able to immediately retrieve from same stage ("+stageN+") value set with setVariable() for " + def,
                            svv1, whichStage.getVariable(def));
                        if(def.isLocal())
                            {
                            assertEquals("Timestamp must be preserved (at least for locals) at same stage value set with setVariable() for " + def,
                                svv1.getTimestamp(), whichStage.getVariable(def).getTimestamp());
                            }

                        // Check that local variable never has non-null globalMap...
                        if(def.isLocal())
                            {
                            assertNull("Local variable should always have null globalMap",
                                whichStage.getVariable(def).getGlobalMap());
                            }

                        // Now test that value can be seen at all stages
                        // within the allowed maximum latency...
                        // Sometimes allow a syncVariables(true)
                        // at the TOP of the pipeline
                        // to stochastically speed up the tests
                        // and to check that the semantics are correct.
                        final long waitStart = System.currentTimeMillis();
                        final long warnAfter = waitStart + 10 * 1000;
                        final long retrieveBy = waitStart +
                            SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS;
                        for(int r = plen; --r >= 0; )
                            {
                            try
                                {
                                // Keep trying to retrieve the variable's value
                                // until we see the value we expect,
                                // or we exceed the maximum allowed latency.
                                // If it is a global, we allow the value
                                // to be in the globalMap instead.
                                int waitTime = 10 + rnd.nextInt(100);
                                if(doSync) { pipeline[0].syncVariables(true); }
                                while(!svv1.equals(pipeline[r].getVariable(def)))
                                    {
                                    final SimpleVariableValue svvActual =
                                        pipeline[r].getVariable(def);
                                    if(svvActual != null)
                                        {
                                        final Map<InstanceID,SimpleVariableValue> sAGM = svvActual.getGlobalMap();
                                        if(sAGM != null)
                                            {
                                            if(sAGM.values().contains(svv1))
                                                {
                                                // OK, found it in the map instead!
                                                break;
                                                }
                                            }
                                        }

                                    // If we did a syncVariables(true),
                                    // we should have seen the value immediately.
                                    if(doSync)
                                        { fail("Could not see value propagated to stage "+r+" immediately after syncVariables(true) (written at stage "+stageN+") for " + def); }

                                    final long now = System.currentTimeMillis();
                                    if(now > retrieveBy)
                                        { fail("Could not see value propagated to stage "+r+" in time (written at stage "+stageN+") for " + def); }
                                    if(now > warnAfter)
                                        { Main.getOut().println("[Waiting at stage "+r+" to see value written at stage "+stageN+" for "+def+" ("+svv1.getFullDescription()+", got "+((svvActual==null)?"null":svvActual.getFullDescription())+")"+"...]"); }
                                    Thread.sleep(waitTime *= 2); // Wait a while...
                                    }

                                // Check timestamp is preserved at least for
                                // at least for all locals.
                                if(def.isLocal())
                                    {
                                    assertEquals("Timestamp must be preserved at same stage value set with setVariable() for " + def,
                                        svv1.getTimestamp(), pipeline[r].getVariable(def).getTimestamp());
                                    }

                                // Check that local variable never has non-null globalMap...
                                if(def.isLocal())
                                    {
                                    assertNull("Local variable should always have null globalMap",
                                        pipeline[r].getVariable(def).getGlobalMap());
                                    }
                                }
                            catch(IOException e)
                                {
                                // A stage that throws an IOException
                                // is excused from this test...
                                Main.getOut().println("[IOException at stage "+r+" fetching value written at stage "+stageN+", so skipping.]");
                                continue;
                                }
                            catch(UnsupportedOperationException e)
                                {
                                // A stage that throws an UnsupportedOperationException
                                // is excused from this test...
                                Main.getOut().println("[UnsupportedOperationException at stage "+r+" fetching value written at stage "+stageN+", so skipping.]");
                                continue;
                                }
                            }
                        }
                    // It is acceptable for the operation to be vetoed.
                    catch(IOException e)
                        {
                        Main.getOut().println("[setVariable() was rejected with IOException for " + def + ".]");
                        }
                    catch(UnsupportedOperationException e)
                        {
                        Main.getOut().println("[setVariable() was rejected with UnsupportedOperationException for " + def + ".]");
                        // But if the operation was vetoed should we have done it?
//                        assertNotSame("Value seems to have been set with setVariable() even though apparently vetoed " + def,
//                            svv1, whichStage.getVariable(def));
                        }

                    // Use setVariables()...
                    // Create new value...
                    final SimpleVariableValue svv2 = new SimpleVariableValue(
                        def, new Short((short) rnd.nextInt()));
                    try
                        {
                        final int nSet =
                            whichStage.setVariables(new SimpleVariableValue[]{svv2});
                        assertTrue("setVariables() result must not be negative",
                            nSet >= 0);
                        assertTrue("setVariables() result must not exceed number of values passed",
                            nSet <= 1);
                        if(nSet > 0)
                            {
                            assertEquals("Must be able to retrieve value set with setVariables() for " + def,
                                svv2, whichStage.getVariable(def));
                            }
                        }
                    // It is acceptable for the operation to be vetoed.
                    catch (IOException e)
                        {
                        Main.getOut().println("[setVariables() was rejected with IOException for " + def + ".]");
                        }
                    catch (UnsupportedOperationException e)
                        {
                        Main.getOut().println("[setVariables() was rejected with UnsupportedOperationException for " + def + ".]");
                        // But if the operation was vetoed should we have done it?
//                        assertNotSame("Value seems to have been set with setVariables() even though apparently vetoed " + def,
//                            svv2, whichStage.getVariable(def));
                        }

                    // Stop once we've done at least one test
                    // if doing tests in random order
                    // so that over time we do test all the variables.
                    if(RANDOMISE_VARIABLE_TEST_ORDER)
                        { if(!notAfter(stopBy)) { return; } } // Outta time...
                    }

                // We've tested all the variables at least once,
                // so if we're out of time we quit now.
                if(!notAfter(stopBy)) { return; } // Outta time...
                }
            }
        catch(IllegalArgumentException e)
            {
            e.printStackTrace(Main.getErr());
            fail("Operation failed inappropriately with IllegalArgumentException");
            }
//        catch(IOException e)
//            {
//            e.printStackTrace(Main.getErr());
//            fail("Operation failed inappropriately with IOException");
//            }
        }

    /**Check that the pipeline handles system ID correctly.
     * Mainly that the value can be fetched from any stage at any time,
     * and that it is always the same for a given pipeline.
     *
     * @param pipeline  pipeline; non-null, non-zero-length
     * @param stopBy  time to stop running tests by, if possible
     */
    private static InstanceID checkSystemIDHandling(
                                    final BasicVarMgrInterface[] pipeline,
                                    final long stopBy)
        throws IOException
        {
        if((pipeline == null) || (pipeline.length < 1))
            { throw new IllegalArgumentException(); }
        final int plen = pipeline.length;

        try
            {
            // Check that ID can be fetched from any stage at any time,
            // and is consistent.
            final SimpleVariableValue iidVar = pipeline[rnd.nextInt(plen)].
                        getVariable(SystemVariables.LOCAL_SYS_ID);
            assertNotNull("Must always be able to fetch IID variable", iidVar);
            final InstanceID iid = (InstanceID) iidVar.getValue();
            assertNotNull("IID variable value must always be non null", iid);
            // Test a reasonable number of samples.
            for(int i = 2 + plen / 10; --i >= 0; )
                {
                final InstanceID iid2 = (InstanceID) pipeline[rnd.nextInt(plen)].
                    getVariable(SystemVariables.LOCAL_SYS_ID).getValue();
                assertEquals("System ID must be the same on each fetch", iid, iid2);
                if(!notAfter(stopBy)) { break; } // Outta time...
                }

            // Check that we cannot overwrite the ID value...
            // Try at every stage of the pipeline,
            // and make sure that we can read it back.
            final InstanceID bogusIID = InstanceID.createInstanceID();
            assertFalse("Got duplicate InstanceID", iid.equals(bogusIID));
            for(int j = 3; --j >= 0; ) // Try several times to overwrite.
                {
                for(int i = plen; --i >= 0; )
                    {
                    try
                        {
                        pipeline[i].setVariable(new SimpleVariableValue(
                            SystemVariables.LOCAL_SYS_ID, bogusIID));
                        }
                    catch(IllegalArgumentException e) { } // Absorb attempts to veto.
                    catch(UnsupportedOperationException e) { } // Absorb attempts to veto.
                    final InstanceID iid2 = (InstanceID) pipeline[i].
                        getVariable(SystemVariables.LOCAL_SYS_ID).getValue();
                    assertEquals("System ID must not be overwritable", iid, iid2);
                    }
                if(!notAfter(stopBy)) { break; } // Outta time...
                }

            return(iid);
            }
        catch(IOException e)
            {
            e.printStackTrace(Main.getErr());
            fail("IOException inappropriately thrown while testing system ID access");
            throw e; // Should not actually get here...
            }
        }

    /**This checks that operations are vetoed when they should be.
     * This also checks that an appriopiate mechanism is used,
     * eg the correct exception type so that the caller can understand
     * the propblem.
     *
     * @param pipeline  pipeline; non-null, non-zero-length
     * @param stopBy  time to stop running tests by, if possible
     */
    private static void checkIllegalGetSetOpsVetoedCorrectly(final BasicVarMgrInterface[] pipeline,
                                                             final long stopBy)
        {
        if((pipeline == null) || (pipeline.length < 1))
            { throw new IllegalArgumentException(); }
        final int plen = pipeline.length;

        // Try a few times for each stage...
        for(int j = 3 + plen/10; --j >= 0; )
            {
            final SimpleVariableDefinition bogusDef = makeRandomSimpleVariableDefinition();
            final SimpleVariableValue svvBogus = new SimpleVariableValue(
                bogusDef, null);
            for(int i = plen; --i >= 0; )
                {
                final BasicVarMgrInterface stage = pipeline[i];

                try
                    {
                    // Ensure that attempt to set null variable is disallowed.
                    try
                        {
                        stage.setVariable(null);
                        fail("Apparently allowed null variable to be set with setVariable() (from downstream)");
                        }
                    catch(IllegalArgumentException e) { }

                    // Ensure that attempt to get null variable is disallowed.
                    try
                        {
                        stage.getVariable(null);
                        fail("Apparently allowed null variable fetch with getVariable() (from downstream)");
                        }
                    catch(IllegalArgumentException e) { }

                    // Ensure that attempt to set null variable list is disallowed.
                    try
                        {
                        stage.setVariables(null);
                        fail("Apparently allowed null variable array to be set with setVariables() (from downstream)");
                        }
                    catch(IllegalArgumentException e) { }

                    // Ensure that attempt to set null variable in non-empty list is disallowed.
                    try
                        {
                        stage.setVariables(new SimpleVariableValue[]{null});
                        fail("Apparently allowed null variable to be set with setVariables() (from downstream)");
                        }
                    catch(IllegalArgumentException e) { }

                    // But ensure that a legal set of zero veriuables is allowed.
                    assertEquals("Harmless set of empty array of variables should be allowed",
                        0, stage.setVariables(new SimpleVariableValue[0]));
                    }
                catch(IOException e)
                    {
                    e.printStackTrace(Main.getErr());
                    fail("Illegal operation inappropriately vetoed with IOException");
                    }
                catch(UnsupportedOperationException e)
                    {
                    e.printStackTrace(Main.getErr());
                    fail("Illegal operation inappropriately vetoed with UnsupportedOperationException");
                    }

                try
                    {
                    stage.setVariable(svvBogus);
                    fail("Apparently allowed non-extant variable to be set with setVariable() (from downstream)");
                    }
                catch(IOException e) { }
                catch(UnsupportedOperationException e) { }
                catch(IllegalArgumentException e) { }

                try
                    {
                    if(stage.setVariables(new SimpleVariableValue[]{svvBogus}) != 0)
                        { fail("Apparently allowed non-extant variable to be set with setVariables() (from downstream)"); }
                    }
                catch(IOException e) { }
                catch(UnsupportedOperationException e) { }
                catch(IllegalArgumentException e) { }
                }

            if(!notAfter(stopBy)) { break; } // Outta time...
            }
        }

    /**Makes a random definition; never returns null.
     * This randomly generates a variable definition with:
     * <ul>
     * <li>a unique random name
     * <li>a random type (out of all the possible variable types)
     * <li>random properties such as local/global, read-only/read-write
     * </ul>
     * <p>
     * Package-visible so as to be available to other test routines.
     */
    static SimpleVariableDefinition makeRandomSimpleVariableDefinition()
        {
        // Check that we cannot set null variable
        // or variable not listed in SystemVariables.defs.
        // We try this with a random type and flags...
        final SimpleVariableDefinition bogusDef = new SimpleVariableDefinition(
            "bogus.random.name." + (rnd.nextLong() >>> 1),
            SimpleVariableDefinition.TYPE__MIN +
                rnd.nextInt(1 + SimpleVariableDefinition.TYPE__MAX -
                                SimpleVariableDefinition.TYPE__MIN),
            rnd.nextBoolean(),
            rnd.nextBoolean(),
            rnd.nextBoolean(),
            rnd.nextBoolean(), 0, null);
        return(bogusDef);
        }

    /**Class to encapsulate "local" tunnel "server".
     * Attaches to top of data pipeline and allows
     * raw RPC to interact with that pipeline.
     */
    static class LocalTunnelServer
        {
        /**Construct this with the data pipeline.
         * @param source  data source; never null.
         */
        LocalTunnelServer(final SimpleExhibitPipelineIF source)
            {
            this(source, false);
            }

        /**Construct this with the data pipeline.
         * @param source  data source; never null.
         * @param alwaysUseCache  if true then always enable the RPC cache,
         *     else switch it off and on at random
         */
        LocalTunnelServer(final SimpleExhibitPipelineIF source,
                          final boolean alwaysUseCache)
            {
            if(source == null) { throw new IllegalArgumentException(); }
            this.source = source;
            this.alwaysUseCache = alwaysUseCache;
            }

        /**Upstream data source; never null. */
        private final SimpleExhibitPipelineIF source;
        
        /**If true then always enable the RPC cache, else switch it off and on at random. */
        private final boolean alwaysUseCache;

        /**An RPC cache-handler instance. */
        private final ExhibitDataTunnelSource.HIRPCCache _RPC_cache = new ExhibitDataTunnelSource.HIRPCCache();

        /**Perform RPC calls on the wrapped data source.
         *
         * @param request  request packet; never null
         * @return response packet; never null
         * @throws IOException  in case of I/O difficulties
         */
        ExhibitDataTunnelSource.RawPacket doRPCRaw(
                final ExhibitDataTunnelSource.RawPacket request)
            throws IOException
            {
            // Randomly use the cache or not unless alwaysUseCache.
            // Overall behaviour should be correct either way.
            final ExhibitDataTunnelSource.HIRPCCache cache =
                (alwaysUseCache || rnd.nextBoolean()) ? _RPC_cache : null;
            return(ExhibitDataTunnelSource.handleInboundRPC(source, request, null, cache));
            }
        }

    /**Local tunnel client/source, for connecting to LocalTunnelServer.
     * This simulates a real tunnel, but all done locally.
     * <p>
     * Can be configured to force (de)serialisation of the RPC data
     * as if sent on the wire, which will test the serialisation code
     * and ensure no accidental data sharing "by reference" is going on.
     */
    static class LocalTunnelSource extends ExhibitDataTunnelSource
        {
        /**Construct local tunnel source.
         * Sequesters reference to the LocalTunnelServer (the master),
         * and notes whether all RPC packets should be (de)serialised.
         *
         */
        LocalTunnelSource(final LocalTunnelServer master,
                          final boolean forceSerialization,
                          final SimpleLoggerIF logger)
            {
            super(logger);
            if(master == null) { throw new IllegalArgumentException(); }
            this.master = master;
            this.forceSerialization = forceSerialization;
            }

        /**Master server connection; never null. */
        private final LocalTunnelServer master;

        /**If true, force (de)serialisation as if on the wire. */
        private final boolean forceSerialization;

        /**Provide the implementation of the data transport.
         */
        protected ExhibitDataTunnelSource.RawPacket doRPCRaw(
                            final ExhibitDataTunnelSource.RawPacket request)
            throws IOException
            {
            // If not forcing serialisation,
            // do the RPC call directly.
            if(!forceSerialization)
                {
                return(master.doRPCRaw(request));
                }

            // Do it the hard way, serialised via byte streams.
            // We allow caller-selected attempts to compress the payload.
            final ByteArrayOutputStream baos1 =
                new ByteArrayOutputStream(32 + request.plLength);
            request.writePacket(baos1);
            final ByteArrayInputStream bais1 =
                new ByteArrayInputStream(baos1.toByteArray());
            final ExhibitDataTunnelSource.RawPacket response =
                master.doRPCRaw(ExhibitDataTunnelSource.RawPacket.readPacket(bais1));
            final ByteArrayOutputStream baos2 =
                new ByteArrayOutputStream(32 + response.plLength);
            response.writePacket(baos2);
            final ByteArrayInputStream bais2 =
                new ByteArrayInputStream(baos2.toByteArray());
            return(ExhibitDataTunnelSource.RawPacket.readPacket(bais2));
            }

        /**Does nothing in this implementation. */
        public void destroy()
            {
            }
        }

    /**Do simple test of tunnel connectivity.
     * We do this with and without real serialisation,
     * so in passing we are testing the on-the-wire protocol too.
     * <p>
     * We explicitly check that getting/setting a novel variable on the slave
     * (ie not known to the master)
     * is quietly ignored and does not cause any exception/failure.
     * <p>
     * This is a <em>very</em> simple test;
     * look elsewhere for comprehensive tests!
     */
    public void testLocalTunnelSimple()
        throws Exception
        {
        // Do alternate tests with/out forced (de)serialisation...
        for(int i = 2; --i >= 0; )
            {
            // Chose whether to force serialisation...
            final boolean forceSerialisation = (i & 1) == 0;

            // Tunnel endpoint at master (wraps a BasicVarMar).
            final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

            // Tunnel client end on slave.
            final SimpleExhibitPipelineIF ts = new LocalTunnelSource(
                new LocalTunnelServer(ep), forceSerialisation, logger);

            // Check that we can set a value on the slave
            // and see it on the master,
            // and vice versa.
            // This should be immediate,
            // and show up as the main value of the variable,
            // because there is no cacheing involved.
            final SimpleVariableDefinition def =
                SystemVariables.TEST_NUMBER_GLOBAL;
            final Number v1 = new Long(rnd.nextLong());
            final SimpleVariableValue svv1 = new SimpleVariableValue(def, v1);
            ts.setVariable(svv1);
            assertEquals("Must be able to see on the master the value set on the slave",
                svv1, ep.getVariable(def));
            final Number v2 = new Integer(rnd.nextInt());
            final SimpleVariableValue svv2 = new SimpleVariableValue(def, v2);
            ep.setVariable(svv2);
            assertEquals("Must be able to see on the slave the value set on the master",
                svv2, ts.getVariable(def));


            // Check that get of variable unknown to master does no harm.
            for( ; ; )
                {
                final SimpleVariableDefinition novelDef = makeRandomSimpleVariableDefinition();
//                final SimpleVariableValue rndSvv = makeRandomSimpleVariableValue(novelDef);
                if(!novelDef.isLocal())
                    {
                	assertNull("Getting novel (non-local) value with getVariable() must return null", ts.getVariable(novelDef));
                    if(novelDef.isEvent())
	                    {
	                    assertNotNull(ts.getEventValues(novelDef, EventPeriod.VLONG, 0, null)); // Must not throw exception...
	                    }
                	}

                // Stop when we've tested at least one global event value...
                if(!novelDef.isLocal() && novelDef.isEvent()) { break; }
                }


            // Now try normal pipeline tests in the very simple setup.
            final SimpleExhibitPipelineIF slave1[] = { ts };
            final SimpleExhibitPipelineIF master[] = { ep };
            pipelineVariableTest(slave1, null, master,
                System.currentTimeMillis() + 30000L); // Run test for <= 30s.
            }
        }

    /**Run full tunnel variable test.
     * Emulates a reasonably-realistic setup with a master and two slaves,
     * all of which hold a caching stage.
     * <p>
     * One of the tunnels forces full (de)serialisation of its RPC traffic,
     * thus exercising much of the HTTP tunnel code in passing.
     * <p>
     * We poll() all the pipelines at different speeds to ensure that
     * rate synchronisation is not critical to correct functioning.
     */
    public void testLocalTunnelFull()
        throws Exception
        {
        // Tunnel endpoint at master (wraps a BasicVarMar).
        final SimpleExhibitPipelineIF epm = new BasicVarMgrEndpoint();
        // Caching stage in master (no write-through).
        final SimpleExhibitPipelineIF csm = new TestVarCacheStage(epm, false);
        // Master pipeline.
        final SimpleExhibitPipelineIF master[] = { csm, epm };
        // Run poll() so pipeline can sync() its vars.
        final SimpleExhibitPipelineIF masterTop = master[0];
        t.scheduleAtFixedRate(new TimerTask(){
            public void run()
                {
                try { masterTop.poll(new GenProps()); }
                catch(final Exception e) { e.printStackTrace(Main.getErr()); }
                }
            },
            0L, 100L); // Run master poll() every 0.1s.

        // Master stage to which tunnel attaches on master.
        final SimpleExhibitPipelineIF masterTunnelServer = csm;

        // Tunnel client end on slave1; serialisation is forced.
        final SimpleExhibitPipelineIF ts1 = new LocalTunnelSource(
            new LocalTunnelServer(masterTunnelServer), true, logger);
        // Tunnel caching stage on slave1.
        final SimpleExhibitPipelineIF cs1 = new TestVarCacheStage(ts1, false);
        // Slave1 pipeline.
        final SimpleExhibitPipelineIF slave1[] = { cs1, ts1 };
        // Run poll() so pipeline can sync() its vars.
        t.scheduleAtFixedRate(new TimerTask(){
            public void run()
                {
                try { slave1[0].poll(new GenProps()); }
                catch(Exception e) { e.printStackTrace(Main.getErr()); }
                }
            },
            0L, 203L); // Run master poll() every 0.2s.

        // Tunnel client end on slave2; serialisation is not done.
        final SimpleExhibitPipelineIF ts2 = new LocalTunnelSource(
            new LocalTunnelServer(masterTunnelServer), false, logger);
        // Tunnel caching stage on slave2.
        final SimpleExhibitPipelineIF cs2 = new TestVarCacheStage(ts2, false);
        // Slave2 pipeline.
        final SimpleExhibitPipelineIF slave2[] = { cs2, ts2 };
        // Run poll() so pipeline can sync() its vars.
        t.scheduleAtFixedRate(new TimerTask(){
            public void run()
                {
                try { slave2[0].poll(new GenProps()); }
                catch(Exception e) { e.printStackTrace(Main.getErr()); }
                }
            },
            1000L, 151L); // Run master poll() every 0.15s after a delay.

        pipelineVariableTest(slave1, slave2, master,
            System.currentTimeMillis() + 120000L); // Run test for <= 2m.

        // Check that the slaves' IDs appear in a map by address...
        // We could check this more or less anywhere,
        // but we'll use the top of the master pipeline for simplicity...
        final SimpleVariableValue slaves =
            masterTop.getVariable(SystemVariables.TunnelServlet_SLAVE_ADDRS);
        assertNotNull("Must have String globalMap value of participating slaves", slaves);
        final Map<InstanceID,SimpleVariableValue> slavMap = slaves.getGlobalMap();
        assertEquals("Must see exactly two slaves in globalMap",
                     2, slavMap.size());
        assertTrue("Must find our two slaves as keys in globalMap",
                   slavMap.keySet().contains(ts1.getVariable(SystemVariables.LOCAL_SYS_ID).getValue()) &&
                   slavMap.keySet().contains(ts2.getVariable(SystemVariables.LOCAL_SYS_ID).getValue()));
        Main.getOut().println("[Sample slave list: " +VarTools.generateSimpleStats(masterTop, SystemVariables.TunnelServlet_SLAVE_ADDRS, 0)+ ".]");
        }

    /**Test that we can correctly generate stats safely.
     * We test that is possible to try to generate a SimpleNumberStats instance,
     * even with weird inputs.
     * <p>
     * TODO: Test with local value
     */
    public void testSimpleNumberStats()
        {
        // Check that we correctly and politely handle null variables.
        assertNull("A null arg to generateSimpleNumberStats() should result in null",
            VarTools.generateSimpleNumberStats(null, -1, "", null));

        // Check that we correctly and politely reject wrongly-typed variables.
        assertNull("A TYPE_STRING arg to generateSimpleNumberStats() should result in null",
            VarTools.generateSimpleNumberStats(new SimpleVariableValue(
                SystemVariables.TEST_STRING_GLOBAL, "oh happy day!"),
                -1, "", null));
        assertNull("A TYPE_NONE arg to generateSimpleNumberStats() should result in null",
            VarTools.generateSimpleNumberStats(new SimpleVariableValue(
                SystemVariables.KEEP_ALIVE, null),
                -1, "", null));

        // Check that we correctly and politely reject null-values variables.
        assertNull("A null TYPE_NUMBER arg to generateSimpleNumberStats() should result in null",
            VarTools.generateSimpleNumberStats(new SimpleVariableValue(
                SystemVariables.TEST_NUMBER_LOCAL, null),
                -1, "", null));


        // Create three system IDs.
        final InstanceID system1 = InstanceID.createInstanceID();
        final InstanceID system2 = InstanceID.createInstanceID();
        final InstanceID system3 = InstanceID.createInstanceID();

        // Create two values, with different sub-types of Number.
        // Pick values that should be computable without loss of accuracy
        // for all the operators in SimpleNumberStats.
        final Number v1 = new Integer(rnd.nextInt());
        final Number v2 = new Float((float) rnd.nextGaussian());

        // Pick yet a different "main" value.
        final Number vMain = new Short((short) rnd.nextInt());

        // Our variable definition.
        final SimpleVariableDefinition def = SystemVariables.TEST_NUMBER_GLOBAL;

        // Now synthesise our variable with a globalMap...
        final SimpleVariableValue svvGlobal =
            (new SimpleVariableValue(def, vMain)).
            put(system1, new SimpleVariableValue(def, v1), false).
            put(system2, new SimpleVariableValue(def, v2), false).
            put(system3, new SimpleVariableValue(def, null), false);

        // Now make the stats object...
        // We claim to be system1...
        final SimpleNumberStats statsGlobal =
            new SimpleNumberStats(svvGlobal, -1, system1, "#.##");

        // FYI...
        Main.getOut().println("[Sample SimpleNumberStats: " + statsGlobal + ".]");

        // Check the accessor methods.
        assertEquals("Number of systems should be 3", 3, statsGlobal.getSystemCount());
        assertEquals("Number of non-null values should be 2", 2, statsGlobal.getNonNullValueCount());
        assertEquals("Max computed incorrectly",
            new Double(Math.max(v1.doubleValue(), v2.doubleValue())),
            new Double(statsGlobal.getMax().doubleValue()));
        assertEquals("Min computed incorrectly",
            new Double(Math.min(v1.doubleValue(), v2.doubleValue())),
            new Double(statsGlobal.getMin().doubleValue()));
        assertEquals("Sum computed incorrectly",
            new Double(v1.doubleValue() + v2.doubleValue()),
            new Double(statsGlobal.getSum().doubleValue()));
        }


    /**Test simple variable stats.
     * Check that the basic, generic, SimpleVarStats works correctly
     * over a wide range of inputs...
     */
    public void testSimpleVarStats()
        {
        // Try this on a few random variable definitions
        // plus all extant system variable values.
        final List<SimpleVariableDefinition>/*<SimpleVariableDefinition>*/ testDefSet = new ArrayList<SimpleVariableDefinition>();
        testDefSet.addAll(SystemVariables.defs);
        for(int i = 37; --i >= 0; )
            { testDefSet.add(makeRandomSimpleVariableDefinition()); }

        // Try out one possible value for each test definition.
        for(final Iterator<SimpleVariableDefinition> it = testDefSet.iterator(); it.hasNext(); )
            {
            final SimpleVariableDefinition def =
                    it.next();
            final SimpleVariableValue svv = makeRandomSimpleVariableValue(def);

            // Wrap this up in a simple stats object.
            // TODO: check the globalMap filtering behaviour.
            // TODO: check the selection by sysID.
            final SimpleVarStats svs = new SimpleVarStats(svv, -1, null);

            // Check features of this basic value,
            // and its handling by the stats value.
            assertNull("Basic value must always have empty globalMap",
                       svv.getGlobalMap());
            assertEquals("Stats must correctly report non-null entries as zero for basic value",
                         0, svs.getNonNullValueCount());
            assertEquals("Stats must correctly report system count as zero for basic value",
                         0, svs.getSystemCount());
            assertSame("Stats must return main value correctly for basic value",
                       svv.getValue(), svs.getMainValue());

            // Simply make sure that we can convert it to a String
            // without the roof crashing on our heads...
            final String readableString = svs.toString();
            assertNotNull("SimpleVarStats.toString() must not return null", readableString);
            // And the result can only possibly be an empty String
            // if there is no globalMap
            // AND the String representation of the main value is "".
            if(!"".equals(String.valueOf(svv.getValue())) ||
                (svv.getGlobalMap() != null))
                {
                assertTrue("SimpleVarStats.toString() can only be empty if there is no global map and the main value renders as an empty string",
                           !"".equals(svs.toString()));
                }
            }

        // TODO: implement tests for non-empty globalMap...
        }


    /**Do very basic event-handling tests.
     * This works on a stand-alone BasicVarMgr end-point for simplicity.
     */
    public void testBasicEventHandling()
        throws Exception
        {
        // Set of test event names.
        // All String types for now.
        final SimpleVariableDefinition[] testEvents =
                {
                    SystemVariables.TEST_STRING_GLOBAL_EVENT,
                };

        // Check that behaviour is correct for each event name above.
        for(final SimpleVariableDefinition def : testEvents)
            {
            // Create a new (end-point) var manager.
            final BasicVarMgr vars = new BasicVarMgr(true);

            final String sval1 = "test value " + rnd.nextInt();
            final SimpleVariableValue svv1 = new SimpleVariableValue(def, sval1);

            // Set the value a variable number of times...
            final int nEvents = rnd.nextInt(100);
            for(int i = nEvents; --i >= 0; ) { vars.setVariable(svv1); }

            // Check that the total count is correct.
            // (This may be split between current and previous slots if we crossed a boundary.)
            // This should be true at all event scales,
            // though this may be fragile on the shortest (VSHORT) timescale.
            for(final EventPeriod ep : EventPeriod.values())
                {
                final EventVariableValue currentEventSet = vars.getEventValue(def, ep, false);
                final EventVariableValue prevEventSet = vars.getEventValue(def, ep, true);
                assertNotNull("current event set returned by getEventValue() must not be null", currentEventSet);
                assertNotNull("previous event set returned by getEventValue() must not be null", prevEventSet);
                final int countedEvents = currentEventSet.getTotalEventCount() +
                                          prevEventSet.getTotalEventCount();
                assertEquals("Total number of events must be correctly recorded",
                             nEvents, countedEvents);
                assertEquals("Total number of events reported by getCount() must be correct",
                             nEvents, currentEventSet.getCount(sval1) + prevEventSet.getCount(sval1));
                assertTrue("Rank for prev must be 0 (or MAX_VALUE if count is 0)",
                             (prevEventSet.getRank(sval1) == 0) ||
                             (prevEventSet.getRank(sval1) == Integer.MAX_VALUE) && (prevEventSet.getCount(sval1) == 0));
                assertTrue("Rank for current must be 0 (or MAX_VALUE if count is 0)",
                             (currentEventSet.getRank(sval1) == 0) ||
                             (currentEventSet.getRank(sval1) == Integer.MAX_VALUE) && (currentEventSet.getCount(sval1) == 0));
                assertFalse("All event histories returned from bare BasicVarMgr must be non-authoritative",
                            currentEventSet.isAuthoritative() || prevEventSet.isAuthoritative());
                }
            }
        }

    /**Check that events are correctly split into multiple periods.
     * We test this on just the very shortest period/interval here
     * in order to keep test times reasonable,
     * on the assumption that processing is uniform for all periods/intervals.
     */
    public void testEventBucketting()
        throws Exception
        {
        // Create a new (end-point) var manager.
        final BasicVarMgr vars = new BasicVarMgr(true);

        // The test event that we wish to use.
        final SimpleVariableDefinition def = SystemVariables.TEST_STRING_GLOBAL_EVENT;

        // The test value that we use.
        final String value = "MyLittleTestValue(MT)" + Rnd.fastRnd.nextLong();

        // Total event count we believe that we have delivered so far.
        int totalCountSoFar = 0;

        // Now that we have done our set-up, start timing...
        final long startTime = System.currentTimeMillis();
        // Run this test up to long enough to span more than two of the shortest buckets.
        final long stopTime = startTime + Rnd.fastRnd.nextInt(1+(5*SystemVariables.EVENT_INTERVAL_VSHORT_TERM_MS)/2);

        do
            {
            final long now = System.currentTimeMillis();
            // What is our current interval number?
            final long currentIntervalNumber = EventPeriod.VSHORT.getIntervalNumber(now);

            // The maximum number of intervals that can have passed since we started.
            // We allow some leeway here.
            final int maxIntervals = 3 + (int) ((now - startTime) / SystemVariables.EVENT_INTERVAL_VSHORT_TERM_MS);

            // Record an event.
            final SimpleVariableValue svv = new SimpleVariableValue(def, value);
            vars.setVariable(svv);
            ++totalCountSoFar;

            // Set up a bit vector of values to fetch.
            // We always include enough to catch enough recent slots
            // to be able to see all our events,
            // but sometimes we throw in other requests
            // to make sure that that does not itself break anything.
            final BitSet whichValues = new BitSet(maxIntervals);
            // Enough bits to cover any slots that we may have events in.
            whichValues.set(0, maxIntervals);
            // Sometimes add other bogus requests.
            if(Rnd.fastRnd.nextBoolean())
                { for(int i = 10; --i >= 0; ) { whichValues.set(Rnd.fastRnd.nextInt(SystemVariables.EVENT_SAMPLES_RETAINED)); } }

            // Retrieve results, and check that we find the right count across all of the entries...
            // Allow for a little clock skew and look forward one extra slot...
            final EventVariableValue evvs[] =
                    vars.getEventValues(def, EventPeriod.VSHORT, currentIntervalNumber+1, whichValues);
            assertNotNull("Result of getEventValues() must never be null", evvs);
            assertTrue("Result of getEventValues() must never be zero-length once we have recorded an event",
                       evvs.length > 0);

            // Tot up total count found in results.
            // And make sure that we only get non-null results at most where we asked for them.
            int totalCountFound = 0;
            int resultsReturned = 0;
            for(int i = evvs.length; --i >= 0; )
                {
                final EventVariableValue evv = evvs[i];
                if(evv == null) { continue; }
                assertTrue("Must get a result only where one was asked for", whichValues.get(i));
                ++resultsReturned;
                totalCountFound += evv.getTotalEventCount();
                }
            assertEquals("Total count across all results must be correct (evvs.length="+evvs.length+", resultsReturned="+resultsReturned+")", totalCountSoFar, totalCountFound);

            // Sleep so as to aim for at least 100-or-so events in total.
            Thread.sleep(1 + Rnd.fastRnd.nextInt(1 | SystemVariables.EVENT_INTERVAL_VSHORT_TERM_MS/101));
            } while(System.currentTimeMillis() < stopTime);
        }



    /**Do simple test of tunnel connectivity for set()ting and get()ting events.
     * We do this with and without real serialisation,
     * so in passing we are testing the on-the-wire protocol too.
     * <p>
     * This is a <em>very</em> simple test;
     * look elsewhere for comprehensive tests!
     */
    public void testLocalTunnelEventSimple()
        throws Exception
        {
        // Do alternate tests with/out forced (de)serialisation...
        for(int s = 2; --s >= 0; )
            {
            // Chose whether to force serialisation...
            final boolean forceSerialisation = (s & 1) == 0;

            // Set of test event names.
            // All String types for now.
            final SimpleVariableDefinition[] testEvents =
                    {
                        SystemVariables.TEST_STRING_GLOBAL_EVENT,
                    };

            // Check that behaviour is correct for each event name above.
            for(final SimpleVariableDefinition def : testEvents)
                {
                // Tunnel endpoint at master (wraps a BasicVarMar).
                final SimpleExhibitPipelineIF ep = new BasicVarMgrEndpoint();

                // Tunnel client end on slave.
                final SimpleExhibitPipelineIF ts = new LocalTunnelSource(
                    new LocalTunnelServer(ep), forceSerialisation, logger);

                final String sval1 = "test value " + rnd.nextInt();
                final SimpleVariableValue svv1 = new SimpleVariableValue(def, sval1);

                // Set the value a variable number of times...
                final int nEvents = rnd.nextInt(100);
                for(int i = nEvents; --i >= 0; ) { ts.setVariable(svv1); }

                // Check that the total count is correct.
                // (This may be split between current and previous slots if we crossed a boundary.)
                // This should be true at all event scales,
                // though this may be fragile on the shortest (VSHORT) timescale.
                for(final EventPeriod p : EventPeriod.values())
                    {
                    final EventVariableValue currentEventSet = ts.getEventValue(def, p, false);
                    final EventVariableValue prevEventSet = ts.getEventValue(def, p, true);
                    assertNotNull("current event set returned by getEventValue() must not be null", currentEventSet);
                    assertNotNull("previous event set returned by getEventValue() must not be null", prevEventSet);
                    final int countedEvents = currentEventSet.getTotalEventCount() +
                                              prevEventSet.getTotalEventCount();
                    assertEquals("Total number of events must be correctly recorded",
                                 nEvents, countedEvents);
                    assertEquals("Total number of events reported by getCount() must be correct",
                                 nEvents, currentEventSet.getCount(sval1) + prevEventSet.getCount(sval1));
                    assertTrue("Rank for prev must be 0 (or MAX_VALUE if count is 0)",
                                 (prevEventSet.getRank(sval1) == 0) ||
                                 (prevEventSet.getRank(sval1) == Integer.MAX_VALUE) && (prevEventSet.getCount(sval1) == 0));
                    assertTrue("Rank for current must be 0 (or MAX_VALUE if count is 0)",
                                 (currentEventSet.getRank(sval1) == 0) ||
                                 (currentEventSet.getRank(sval1) == Integer.MAX_VALUE) && (currentEventSet.getCount(sval1) == 0));

                    // Get a value as far in the past as we might expect to be able to,
                    // just to make sure that it all works, or at least does not blow up.
                    // Base it on the previous interval to exacerbate "wastage".
                    final BitSet wastefulReq = new BitSet();
                    wastefulReq.set(SystemVariables.EVENT_SAMPLES_RETAINED-1);
                    final long wastefulReqIN = p.getIntervalNumber(System.currentTimeMillis());
                    final EventVariableValue evvs[] =
                            ts.getEventValues(def, p, wastefulReqIN, wastefulReq);
                    assertNotNull("Result of getEventRequests() must not be null", evvs);
                    // Check that at most the request slot is not null.
                    for(int i = evvs.length; --i >= 0; )
                        {
                        final EventVariableValue evv = evvs[i];
                        assertFalse("Must not get data not requested", (evv != null) && !wastefulReq.get(i));
                        // If I get a result then it must be what I asked for.
                        if(evv != null)
                            {
                            assertEquals("Def of result must be right", def, evv.getDef());
                            assertEquals("Period of result must be right", p, evv.getPeriod());
                            assertEquals("Interval number of result must be right", wastefulReqIN, evv.getIntervalNumber());
                            }
                        }
                    }
                }
            }
        }



    /**Check that the various event objects/collections can be (de)serialised.
     * Basic sanity checks; nothing very clever or extensive.
     */
    public void testEventCollectionSerialisation()
        throws Exception
        {
        final SimpleVariableDefinition def = SystemVariables.TEST_STRING_GLOBAL_EVENT;
        SerializationTest.checkObjectCanBeSerialisedAndDeserialised(
                new EventVariableValueSet(def));
        for(final EventPeriod ep : EventPeriod.values())
            {
            SerializationTest.checkObjectCanBeSerialisedAndDeserialised(
                    new EventVariableValuePeriodRow(def, ep));

            final long intervalNumber = ep.getIntervalNumber(System.currentTimeMillis());
            SerializationTest.checkSerialisationPreservesEquality(
                    new EventVariableValue(false, def, ep, intervalNumber, (Rnd.fastRnd.nextInt() >>> 1), null, null));
            SerializationTest.checkSerialisationPreservesEquality(
                    new EventVariableValue(true, def, ep, intervalNumber, (Rnd.fastRnd.nextInt() >>> 1), null, null));

            SerializationTest.checkSerialisationPreservesEquality(
                    new EventVariableValueBuffer(def, ep, intervalNumber));
            }
        }

    /**Test that we can persist the event histories.
     */
    public void testEventHistoryPeristence()
        throws Exception
        {
        // Work in a temporary subdirectory of the current directory.
        final File tempDir = (new File("CacheTst.tmp")).getAbsoluteFile();

        try { FileTools.rmRecursively(tempDir); }
        catch(IOException e) { e.printStackTrace(Main.getErr()); }
        if(tempDir.isDirectory())
            { fail("unable to remove extant temp test directory"); }

        if(!tempDir.mkdirs() || !tempDir.isDirectory())
            { fail("unable to create temp test directory"); }
        try
            {
            final SimpleVariableDefinition def = SystemVariables.TEST_STRING_GLOBAL_EVENT;
            assertTrue("test event must be persistent", def.isPersistent());

            // Create our main variable manager...
            final BasicVarMgr vars = new BasicVarMgr(true);

            // Check that we cannot see events in the new var manager.
            for(final EventPeriod period : EventPeriod.values())
                { assertEquals("must not yet be able to see any events", 0, vars.getEventValue(def, period, rnd.nextBoolean()).getTotalEventCount()); }

            // Check that loading from our new empty directory still shows zero events.
            vars.loadEventHistories(tempDir);
            for(final EventPeriod period : EventPeriod.values())
                { assertEquals("must not yet be able to see any events", 0, vars.getEventValue(def, period, rnd.nextBoolean()).getTotalEventCount()); }

            // Insert a random number of events and check that we can see them.
            final String value = "dummyValue" + Rnd.fastRnd.nextInt();
            final SimpleVariableValue svv = new SimpleVariableValue(def, value);
            final int eventCount = Rnd.fastRnd.nextInt(101);
            for(int i = eventCount; --i >= 0; )
                { vars.setVariable(svv); }

            // Check that we can see the new events in the variable manager.
            for(final EventPeriod period : EventPeriod.values())
                {
                assertEquals("must be able to see events", eventCount,
                             vars.getEventValue(def, period, false).getTotalEventCount() +
                             vars.getEventValue(def, period, true).getTotalEventCount());
                }

            // Save the history data...
            // For the options that should not matter, try them at random.
            vars.saveEventHistories(tempDir, Rnd.fastRnd.nextBoolean(), Rnd.fastRnd.nextBoolean(), false);

            // Check that we can still see the new events in the variable manager.
            for(final EventPeriod period : EventPeriod.values())
                {
                assertEquals("must be able to see events after save", eventCount,
                             vars.getEventValue(def, period, false).getTotalEventCount() +
                             vars.getEventValue(def, period, true).getTotalEventCount());
                }


            // Create a new variable manager...
            final BasicVarMgr vars2 = new BasicVarMgr(true);

            // Check that we cannot see events in the new var manager.
            for(final EventPeriod period : EventPeriod.values())
                { assertEquals("must not yet be able to see any events in new vars", 0, vars2.getEventValue(def, period, rnd.nextBoolean()).getTotalEventCount()); }

            // Now make sure that we can load the state into the new manager,
            // and see the correct state in there.
            vars2.loadEventHistories(tempDir);
            // Check that we can still see the new events in the variable manager.
            for(final EventPeriod period : EventPeriod.values())
                {
                assertEquals("must be able to see events after load", eventCount,
                             vars2.getEventValue(def, period, false).getTotalEventCount() +
                             vars2.getEventValue(def, period, true).getTotalEventCount());
                }
            }
        finally
            {
            // Zap any files that we wrote...
            FileTools.rmRecursively(tempDir);
            }
        }


    /**Observe upstream accesses to event values.
     * This can optionally veto redundant reads that indicate that a cacheing stage
     * is not working properly.
     * <p>
     * The EventVariableValue.equals() and hashCode() methods allow
     * us to simply keep a Set of authorised values returned;
     * any duplicate is an error.
     */
    static final class EventValueRequestMonitor extends SimpleExhibitPipelineFilter
        {
        /**Construct instance wrapped round upstream source.
         * @param upstream  datasource; never null
         * @param vetoRedundantReads  if true, veto attempts to read authoritative values
         *     more than once, ie redunantly;
         * should be safe to use only behind a cache stage
         */
        EventValueRequestMonitor(final SimpleExhibitPipelineIF upstream,
                                 final boolean vetoRedundantReads)
            {
            super(upstream);
            this.vetoRedundantReads = vetoRedundantReads;
            }
        /**If true, veto attempts to read authoritative values more than once, ie redunantly. */
        private final boolean vetoRedundantReads;

        /**Set of authoritative responses so far passed downstream; never null. */
        private final Set<EventVariableValue> authValueSeen = new HashSet<EventVariableValue>();

        /**Intercept event requests and look for incorrect redundancy. */
        public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                                final EventPeriod intervalSelector,
                                                final boolean current)
            {
            if((def == null) || !def.isEvent() || (intervalSelector == null))
                { throw new IllegalArgumentException(); }

            final long now = System.currentTimeMillis();
            final long currentIntervalNumber = intervalSelector.getIntervalNumber(now);
            final long intervalNumber = (current ? currentIntervalNumber : (currentIntervalNumber-1));

            final BitSet justTheFirst = new BitSet(1);
            justTheFirst.set(0); // Get just the first interval's data.
            try
                {
                final EventVariableValue evv[] = getEventValues(def,
                                                                intervalSelector,
                                                                intervalNumber,
                                                                null);

                // If we got some real (non-null) data back, then return it...
                if((evv.length > 0) && (evv[0] != null))
                    { return(evv[0]); }
                }
            catch(final IOException e)
                {
                // Fall through to return empty non-auth default result.
                }

            // Else invent empty non-authoritative return value (with no events)...
            return(new EventVariableValue(false,
                                          def,
                                          intervalSelector,
                                          intervalNumber,
                                          0, null, null));
            }

        /**Intercept event requests and look for incorrect redundancy. */
        public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                                   final EventPeriod intervalSelector,
                                                   final long intervalNumber,
                                                   final BitSet whichValues)
            throws IOException
            {
//            System.out.println("***Upstream request: " + whichValues);

            final EventVariableValue[] evvs = source.getEventValues(def, intervalSelector, intervalNumber, whichValues);

            // Look out for duplicate reads if they are not to be allowed...
            if(vetoRedundantReads)
                {
                final long now = System.currentTimeMillis();
                final long currentIntervalNumber = intervalSelector.getIntervalNumber(now);
                // A cacheing stage need not hold values further in the past
                // than our holding window.
                final long oldestCacheable = currentIntervalNumber - SystemVariables.EVENT_SAMPLES_RETAINED;

                for(int i = whichValues.nextSetBit(0); i >= 0; i = whichValues.nextSetBit(i+1))
                    {
                    // Stop when we've gone too far back to be cacheable...
                    if(intervalNumber - i < oldestCacheable) { break; }

                    final EventVariableValue evv = evvs[i];
                    if(evv == null) { continue; }
                    if(!evv.isAuthoritative()) { continue; }

                    if(!authValueSeen.add(evv))
                        {
                        // This is a forbidden duplicate...
                        fail("Duplicate auth EVV retrieved: " + evv.toString());
                        }
                    }
                }

            return(evvs);
            }
        }



    /**Test correct generation and handling of authoritative event values.
     * Check that values (other than the current interval) retrieved from the master store
     * are correctly marked as authoritative.
     * Ensure that absent values are returned as authoritative empty values
     * to help upstream caches avoid re-requesting them.
     * <p>
     * Check that these are cached correctly by a pipeline cache stage
     * and not redundantly requested from the source.
     * This is important to ensure that slaves do not waste time and bandwidth
     * in redundantly re-requesting past values from the master.
     */
    public void testAuthEventValueGenerationAndCacheing()
        throws IOException
        {
        // Set up a simple instance of a file data source.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();

        // Test that we get appropriate data from the ExhibitDataFileSource
        // which will be used in the master to be a persistent data repository.
        //
        // All requests for valid events/periods should be responded to
        // as authoritative or not depending on the slot time as follows:
        //   * In the future: null are returned because no data *can* exist yet.
        //   * For the current interval: a non-authoritative, non-null response.
        //   * For any past interval: an authoritative response.
        // Where the data for responses does not yet exist,
        // it must be invented with no events to indicate that
        // it definitely does not exist
        // so as to allow negative cacheing by upstream users.

        // Test event; must support all periods.
        final SimpleVariableDefinition def = SystemVariables.TEST_STRING_GLOBAL_EVENT;

        // Run the tests a few times for different values.
        for(int testNumber = 1000 + rnd.nextInt(10); --testNumber >= 0; )
            {
            // On alternate tests, introduce a cache stage.
            final boolean insertCache = ((testNumber & 1) != 0);

            // Wrap the raw source in such a way as to veto cacheing errors,
            // ie where the cache wrongly passes redundant reads upstream.
            final SimpleExhibitPipelineFilter dataSource =
                    new EventValueRequestMonitor(edfs, insertCache);

            // Now wrap the data source as needed.
            final SimpleVariablePipelineIF source = (!insertCache) ? (SimpleVariablePipelineIF) dataSource :
                                                    new PipelineVarMgr(dataSource, rnd.nextBoolean());

            // Compute an offset of the start of the request interval
            // from the current interval,
            // since this must work correctly whatever the interval is.
            final int offset = rnd.nextInt(2*SystemVariables.EVENT_SAMPLES_RETAINED) -
                    (SystemVariables.EVENT_SAMPLES_RETAINED);

            // Pick a period at random;
            // the test event is assumed to support all periods.
            final EventPeriod p = (EventPeriod.values())[rnd.nextInt(EventPeriod.values().length)];

            // Get "current" interval for the given period.
            final long currentInterval = p.getIntervalNumber(System.currentTimeMillis());

            // Compute start interval for our test.
            final long startInterval = currentInterval + offset;

            // Compute random (small, possibly zero) set of random test interval slots.
            // We pick a range unlikey to upset bounds checking anywhere.
            final BitSet whichValues = new BitSet();
            for(int sn = rnd.nextInt(31); --sn >= 0; )
                { whichValues.set(rnd.nextInt(SystemVariables.EVENT_SAMPLES_RETAINED)); }

            // Show that repeated requests are idempotent,
            // and that past values are correctly handled by any cache stage.
            for(int requestNumber = 2 + rnd.nextInt(3); --requestNumber >= 0; )
                {
//System.out.println("[Requesting (@offset="+offset+"): " + whichValues + ".]");
                // Make the request...
                final EventVariableValue result[] = source.getEventValues(def,
                                                                        p,
                                                                        startInterval,
                                                                        whichValues);

                assertNotNull("Result of getEventValues() must not be null", result);
                assertTrue("Result of getEventValues() must be at least long enough to cover all requests",
                           result.length >= whichValues.length());

                // We examine every slot in the returned array...
                for(int i = result.length; --i >= 0; )
                    {
                    final long thisInterval = startInterval-i;

                    final EventVariableValue evv = result[i];
                    assertEquals("Slot must be full iff request bit set and vice versa, but all null in the future",
                                 whichValues.get(i) && (thisInterval <= currentInterval),
                                 (evv != null));

                    // Verify correct null/auth/nonauth based on slot number.
                    // Future slots MUST be null...
                    if(thisInterval > currentInterval)
                        { assertNull("future interval slots must be null", evv); }
                    else if(thisInterval == currentInterval)
                        {
                        if(evv == null)
                            { assertFalse("Must have non-null current slot value if requested", whichValues.get(i)); }
                        else
                            { assertFalse("Current value returned must not be authoritative", evv.isAuthoritative()); }
                        }
                    else
                        {
                        if(evv == null)
                            { assertFalse("Must have non-null past slot value if requested", whichValues.get(i)); }
                        else
                            { assertTrue("Past value returned must be authoritative", evv.isAuthoritative()); }
                        }

                    if(evv == null) { continue; }

                    // For non-null values check def, etc, is as requested.
//System.out.println("[evv="+evv+".]");
                    assertEquals("Returned value must be for correct def", def, evv.getDef());
                    assertEquals("Returned value must be for period", p, evv.getPeriod());
                    assertEquals("Returned value must be for correct interval", thisInterval, evv.getIntervalNumber());
                    }

                // TODO: check that rerequesting the results through a cache stage does not fetch them from the source again.
                }
            }
        }

    /**Test that (re)computing ExhibitPropsComputableMutable data does not cause redundant event data fetches.
     * We care about this because event data is quite bulky
     * and therefore quite expensive and slow to fetch,
     * so we want to make sure that we fetch each event history at most once.
     */
    public void testEPCMComputationDoesNotRefetchEventHistory()
        throws Exception
        {
        // Set up a simple instance of a file data source,
        // wrapped in a filter to discover redundant event-data fetches,
        // and wrapped in a simple pipeline cache stage.
        final ExhibitDataFileSource edfs = new ExhibitDataFileSource();
        final SimpleVariablePipelineIF dataSource = new PipelineVarMgr(
                new EventValueRequestMonitor(edfs, true),
                true); // Make this a write-through variable cache for simplicity...

        final AllExhibitProperties aep = edfs.getAllExhibitProperties(-1);
        assertNotSame("Must be real exhibit data for us to work on", 0, aep.aeid.length);

        final GenProps gp = edfs.getGenProps(-1);
        assertNotSame("Must be real GenProps for us to work with", 0, gp.timestamp);

        // Force explicit full computation of all EPCM several times...
        for(int i = 10; --i >= 0; )
            {
            for(final ExhibitStaticAttr esa : aep.aeid.getAllStaticAttrs())
                {
                // Force the computation...
                ExhibitPropsComputableMutable.compute(esa, gp, aep, dataSource,
                    ExhibitPropsComputableMutableVoteCacheIF.TRIVIAL,
                    ScorerCacheIF.TRIVIAL);
                }
            }
        }




    /**Returns true if time is not after that stated. */
    private static boolean notAfter(final long stopBy)
        { return(System.currentTimeMillis() <= stopBy); }

    /**Private source of OK pseudo-random numbers. */
    private static final Random rnd = new Random();
    }
