/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.ExhibitStaticAttr;
import org.hd.d.pg2k.webSvr.exhibit.TreeFilterBean;

import java.util.*;

/**Tests aspects of the the TreeFilterBean and related functionality.
 */
public final class TreeFilterBeanTest extends TestCase
    {
    public TreeFilterBeanTest(String name)
        {
        super(name);
        }

    /**Test conversion of partial URIs to prefix word sequences.
     */
    public void testConvertTrailingURIToWordSequence()
        {
        assertEquals("Simple conversion 1a: one-word partial URI with leading /", "word1-", TreeFilterBean.convertTrailingURIToWordSequence("/word1/index.html"));
        assertEquals("Simple conversion 1b: one-word partial URI without leading /", "word1-", TreeFilterBean.convertTrailingURIToWordSequence("word1/index.html"));
        assertEquals("Simple conversion 1c: one-word partial URI with repeated /", "word1-", TreeFilterBean.convertTrailingURIToWordSequence("////word1/////index.html"));

        assertEquals("Simple conversion 2: multiple words", "the-cat-sat-", TreeFilterBean.convertTrailingURIToWordSequence("/the/cat/sat/index.jsp?page=9"));

        assertEquals("Simple conversion 3: conversion to lowercase", "the-cat-sat-", TreeFilterBean.convertTrailingURIToWordSequence("/The/caT/sat/index.jsp?"));

        assertEquals("Simple conversion 4a: final component with dot is discarded", "word-second-", TreeFilterBean.convertTrailingURIToWordSequence("/word/second/x.y"));
        assertEquals("Simple conversion 4b: final component before trailing slash is kept", "word-second-", TreeFilterBean.convertTrailingURIToWordSequence("/word/second/"));
        assertEquals("Simple conversion 4c: final component without dot is treated as dir/word, not discarded", "word-second-", TreeFilterBean.convertTrailingURIToWordSequence("/word/second"));

        try {
            TreeFilterBean.convertTrailingURIToWordSequence("/the-cat/index.html");
            assertTrue(false); // Embedded word separator in URI should be rejected...
            }
        catch(IllegalArgumentException e) { } // Good: we want an exception to have been thrown...

        try {
            TreeFilterBean.convertTrailingURIToWordSequence("/the*cat/index.html");
            assertTrue(false); // Invalid word character in URI should be rejected...
            }
        catch(IllegalArgumentException e) { } // Good: we want an exception to have been thrown...
        }

    /**Test the generation of a browsable tree of exhibits from an input set...
     * This tests if the tree construction (and generated data set)
     * behaves as expected.
     */
    public void testBrowsableTree()
        {
        // Check that looking for uninteresting entries with a root node
        // unconditionally returns an empty list.
        final Map<String,Collection<String>> emptyMap = Collections.emptyMap();
        assertTrue("Inspection at root node unconditionally should return an empty list",
                   TreeFilterBean.getUninterestingNodesFromMap(emptyMap, "").length == 0);

        // Short names of the exhibits.
        final String SHNAME1 = "cat-sat-on-mat-ANON.htxt";
        final String SHNAME2 = "CAT-fat-DHD.jpg"; // Note different case of main word.
        final String SHNAME3 = "dog-in-manger-1-ANON.gif";
        final String SHNAME4 = "dog-in-manger-2-ANON.gif";
        final String SHNAME5 = "dog-in-manger-11-ANON.gif";

        // Full names of exhibits.
        final String EXNAME1 = "memes/" + SHNAME1;
        final String EXNAME2 = "memes/" + SHNAME2;
        final String EXNAME3 = "other/" + SHNAME3;
        final String EXNAME4 = "another/" + SHNAME4;
        final String EXNAME5 = "another/" + SHNAME5;

        // For now, ignore the effects of attribute words...
        final long now = System.currentTimeMillis();
        final Map<String,ExhibitStaticAttr> exhibits1 = new HashMap<String, ExhibitStaticAttr>();
        exhibits1.put(EXNAME1, new ExhibitStaticAttr(EXNAME1, 100, now));
        exhibits1.put(EXNAME2, new ExhibitStaticAttr(EXNAME2, 100, now));
        exhibits1.put(EXNAME3, new ExhibitStaticAttr(EXNAME3, 100, now));
        exhibits1.put(EXNAME4, new ExhibitStaticAttr(EXNAME4, 100, now));
        exhibits1.put(EXNAME5, new ExhibitStaticAttr(EXNAME5, 100, now));
//        final AllExhibitProperties aep1 = new AllExhibitProperties(
//                                new ExhibitPropsGlobalImmutable(),
//                                new AllExhibitImmutableData(exhibits1, now),
//                                Collections.emptyMap() /* Map _loadedProps */);

        // Build this Map non-optimised and then optimised,
        // checking that it gives the right results in both cases.
        for(final boolean optimised : new boolean[]{ false, true} )
            {
            // Build this map first with SortedSet nodes then with List nodes.
            for(final boolean asSet : new boolean[]{ false, true} )
                {
                final Map<String,Collection<String>> m1 = TreeFilterBean.computePrefixMap(new ArrayList<String>(exhibits1.keySet()), asSet, optimised);

                // Check that all node entries are in correct sorted order.
                for(final Iterator<Collection<String>> it = m1.values().iterator(); it.hasNext(); )
                    {
                    final Collection<String> nodeEntries = (it.next());
                    if(nodeEntries.size() > 1)
                        {
                        final Iterator<String> eit = nodeEntries.iterator();
                        String prev = eit.next();
                        while(eit.hasNext())
                            {
                            String cur = eit.next();
                            assertTrue("all node elements must be correctly ordered",
                                       String.CASE_INSENSITIVE_ORDER.compare(prev, cur) <= 0);
                            }
                        }
                    }

                // Root node should contain just the entire flattened tree
                // if it is small enough when we are optimising.
                // Otherwise we expect to find the "cat-" and "dog-...-" subtrees.
                // (And should be the right type.)
                final Collection<String> rootNode1 = m1.get("");
                assertTrue("expect nodes to be of right type (List/SortedSet)",
                           (asSet ? (rootNode1 instanceof SortedSet) : (rootNode1 instanceof List)));
                if(!optimised)
                    {
                    assertEquals("expected exactly two subtrees (for cat and dog) at root", 2, rootNode1.size());
                    assertTrue("expected to find the cat and dog subtrees at root",
                                    (rootNode1.contains("cat-")) &&
                                        (optimised ? rootNode1.contains("dog-in-manger-") : rootNode1.contains("dog-")));
                    }
                else
                    {
                    assertEquals("expect full exhibit set (flattened tree) at root", exhibits1.size(), rootNode1.size());
                    }

                // Look at "cat-" sub-tree.
                // We expect to find the entries merged together ignoring case,
                // and exactly the two subtrees.
                final Collection<String> catNode1 = m1.get("cat-");
                assertTrue("expect exactly two subtrees [or exhibits in optimised case] (for sat and fat) below cat", catNode1.size() == 2);
                if(!optimised)
                    {
                    assertTrue("expected to find the sat and fat subtrees below cat",
                        (catNode1.contains("cat-fat-")) && (catNode1.contains("cat-sat-")));
                    }
                else
                    {
                    assertTrue("expected to find the sat and fat exhibits below cat",
                        (catNode1.contains(SHNAME1)) && (catNode1.contains(SHNAME2)));
                    }

                // Look at "cat-fat-" sub-tree.
                // We expect to find one actual exhibit entry..
                final Collection<String> catFatNode1 = m1.get("cat-fat-");
                assertTrue("expect one exhibit entry in beneath cat-fat", catFatNode1.size() == 1);
                assertTrue("expected to find the one entry for " + SHNAME2, catFatNode1.contains(SHNAME2));

                // Look at "dog-in-manger-" sub-tree.
                // We expect to find the two exhibit entries.
                final Collection<String> dogInMangerNode1 = m1.get("dog-in-manger-");
                assertTrue("expect three exhibit entries below dog-in-manger", dogInMangerNode1.size() == 3);
                assertTrue("expected to find the three dog-in-manger entries",
                    (dogInMangerNode1.contains(SHNAME3)) &&
                    (dogInMangerNode1.contains(SHNAME4)) &&
                    (dogInMangerNode1.contains(SHNAME5)));
                // Check uninteresting parents.
                final boolean dogInMangerBoring[] = TreeFilterBean.getUninterestingNodesFromMap(m1, "dog-in-manger-");
                assertTrue("dog-in-manger- should have two parents noted",
                           dogInMangerBoring.length == 2);
                assertTrue("dog-in-manger- parent nodes should both be marked uninteresting if not optimised",
                           dogInMangerBoring[0] && dogInMangerBoring[1]);

                // Now in the non-optimised case the "dog-" node should point to the "dog-in-" node,
                // but in the optimised case it should point straight at the less dull "dog-in-manger-" node.
                final Collection<String> dogNode1 = m1.get("dog-");
                if(!optimised)
                    {
                    assertTrue("expect one sub-tree below dog", dogNode1.size() == 1);
                    assertTrue("expected one subnode: " + dogNode1.iterator().next(), dogNode1.contains("dog-in-"));
                    }

                // Check that looking for uninteresting entries with a single-word node
                // returns an empty list.
                assertTrue("Inspection at single-word node unconditionally should an empty list",
                           TreeFilterBean.getUninterestingNodesFromMap(m1, "cat-").length == 0);
                assertTrue("Inspection at single-word node unconditionally should an empty list",
                           TreeFilterBean.getUninterestingNodesFromMap(m1, "dog-").length == 0);

                final boolean catSatOnMatBoring[] = TreeFilterBean.getUninterestingNodesFromMap(m1, "cat-sat-on-mat-");
                assertTrue("cat-sat-on-mat- should have three (un)interesting parents noted: " + catSatOnMatBoring.length,
                           catSatOnMatBoring.length == 3);
                assertTrue("cat-sat-on-mat- parent nodes should both be marked uninteresting except first: " +
                           catSatOnMatBoring[0]+','+catSatOnMatBoring[1]+','+catSatOnMatBoring[2],
                           (!catSatOnMatBoring[0]) && catSatOnMatBoring[1] && catSatOnMatBoring[2]);

                }
            }
        }
    }
