/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: 28-Dec-02
 * Time: 22:24:51
 */
package org.hd.d.pg2k.test.dev;

import junit.framework.TestCase;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.HostUtils;
import org.hd.d.pg2k.webSvr.virtualHosts.VirtualHosts;

/**Test of various virtual host features.
 */
public final class VirtualHostTest extends TestCase
    {
    public VirtualHostTest(String name)
        {
        super(name);
        }

    /**Test canonicalisation of HTTP host names.
     * This checks for:
     * <ul>
     * <li>Conversion to a single case.
     * <li>Stripping of superfluous trailing dots.
     * <li>Stripping of prefixes such as "www.", "master." and "mirror-*.".
     * </ul>
     */
    public void testHostnameCanonicalisation()
        {
        // canonical form we are aiming for.
        final String canonical = "gallery.hd.org";

        // Items that should reduce to the canonical form.
        final String good[] =
            {
                canonical,
                canonical.toUpperCase(),
                "www." + canonical + ".",
                "master." + canonical + ".",
                "mirror-uk-th9." + canonical + ".",
                "mirror-us-ga1." + canonical + ".",
                "www." + canonical + "....",
                canonical + "...",
                "wWw.GaLLERy.HD.ORG."
            };
        for(int i = good.length; --i >= 0; )
            {
            assertEquals("Should have reduced to canonical form",
                         canonical,
                         HostUtils.normaliseVirtualHostName(good[i]));
            }

        // Items that should NOT reduce to the canonical form...
        final String bad[] =
            {
            "", ".",
            "random.hd.org",
            "someother.gallery.hd.org",
            "wwwgallery.hd.org",
            "ibm.com"
            };
        for(int i = bad.length; --i >= 0; )
            {
            try
                {
                assertNotSame("Should NOT have reduced to canonical form",
                             canonical,
                             HostUtils.normaliseVirtualHostName(bad[i]));
                }
            catch(IllegalArgumentException e) { } // Allowed to reject them outright...
            }
        }

    /**Check that lookups by name and/or URI yield the correct VirtualHost details.
     */
    public void testVirtualHostLookupDetails()
        {
        // Ensure that the host map is not empty.
        assertTrue("There must be at least one good host name",
                   VirtualHosts.hostDetails.size() > 0);

        // Make sure that if we look up a bogus hostname
        // we get a null.
        assertNull("Looking up a bogus hostname should give null",
                   VirtualHosts.getVirtualHostDetails("bogus.name", null));

        // Make sure that if we look up the default/main hostname
        // we get the standard default livery.
        assertSame("Looking up the main hostname should give standard livery details",
                   VirtualHosts.VIRTUAL_HOST_DEFAULT,
                   VirtualHosts.getVirtualHostDetails(CoreConsts.MAIN_DATA_HOST, null));

        // Make sure that if we look up non-canonicalised default/main hostname
        // we get the standard default livery.
        assertSame("Looking up the main hostname in non-canonical form should give standard livery details",
                   VirtualHosts.VIRTUAL_HOST_DEFAULT,
                   VirtualHosts.getVirtualHostDetails(
                       "www." + CoreConsts.MAIN_DATA_HOST.toUpperCase() + ".",
                       null));

        // Make sure that if we look up "localhost"
        // we get correct non-null "localhost" VirtualHost details
        // so that we won't get redirected and can do sensible testing!
        assertSame("Looking up localhost should give correct non-null livery details",
                   VirtualHosts.VIRTUAL_HOST_LOCALHOST,
                   VirtualHosts.getVirtualHostDetails("localhost", null));

        // Make sure that some critical aliases *can* be looked up with
        // non-null VirtualHost results...
        final String testNames[] =
            {
                "aloha-earth.com", "alohaearth.com", "alohaearth.org",
                "pix.org.za", "photo.org.za", "photos.org.za",
                "londonpix.org.uk",
                "localhost"
            };
        for(int i = testNames.length; --i >= 0; )
            {
            final String tn = testNames[i];
            assertNotNull("Must be able to look up main/alias name: " + tn,
                          VirtualHosts.getVirtualHostDetails(tn, null));
            }

//        // Makes sure that all the canonical VirtualHost entries
//        // TODO: and all their aliases
//        // correctly map to themselves, even with de-canonicalisation.
//        for(final Iterator it = VirtualHosts.canonicalVirtualHosts.iterator(); it.hasNext(); )
//            {
//            final VirtualHosts.VirtualHost vh = (VirtualHosts.VirtualHost) it.next();
//            assertNotNull("No canonicalVirtualHosts entry may be null", vh);
//
//            }
        }
    }
