/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 10-Jul-02
 * Time: 11:17:04

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.webSvr.ads;

import java.util.Locale;
import java.util.Random;

/**This is the interface of all AdBean advertising beans.
 * Beans to support advertising in the body of a page
 * are put in the classpath and loaded by name, and must
 * implement this interface.
 * <p>
 * An AdBean can produce several sorts of output, the usual
 * one being some well-behaved HTML to insert into a page as
 * a banner or elsewhere.
 * <p>
 * AdBean entities should be completely thread-safe and should
 * probably not carry significant state from call to call.
 * <p>
 * AdBean items can signal their affinity to certain Gallery
 * topic areas or keywords or even exhibit prefixes and may
 * be used preferentially around such exhibits.
 * <p>
 * When asked to produce output, this is passed a random-number
 * source to help generate its output.  A bean should assume that
 * the source is a shared resource (eg that is should not be locked
 * for long periods) and that it <em>may</em> be expensive to use
 * and so using it sparingly is a good thing.
 * <p>
 * Many of the routines take an immutable Context argument, which details
 * which Gallery section and/or exhibit is featured near the ad
 * insert, if appropriate, and what is the preferred user locale.
 * <p>
 * <strong>This interface and its subcomponents must not depend on any PG2K
 * types/classes, because it must be possible to take this file/class
 * on its own and compile against it.</strong>
 */
public interface AdBeanInterface
    {
    /**Details context in which advertisement is to be shown.
     * Is immutable for thread-safety, etc.
     * <p>
     * Context items may be null, indicating that they are not
     * applicable to this particular context,
     * but any items that are non-null must be mutually consistent.
     * <p>
     * A generation routine is generally free to ignore all of the
     * supplied data if it wishes to.
     */
    public static final class Context
        {
        /**Construct a complete context.
         * Items that do not apply can be null.
         * Any that are non-null must be consistent with one another
         *
         * @param loc  if non-null, the preferred locale
         * @param section  if non-null, the appropriate top-level Gallery section
         * @param exhibitName  if non-null, is the full name of the specific exhibit associated with this page
         */
        public Context(final Locale loc, final String section, final String exhibitName)
            {
            locale = loc;
            gallerySection = section;
            exhibit = exhibitName;

            if((gallerySection != null) && (exhibit != null))
                {
                if(!exhibitName.startsWith(gallerySection + '/'))
                    { throw new IllegalArgumentException("inconsistent arguments"); }
                }
            }

        /**The Locale of the visitor, or null.
         * Assumed immutable, so handed out directly without a copy being
         * taken (nor a copy made on construction).
         */
        private final Locale locale; // Assumed immutable.

        /**The gallery section (top-level directory, or null.
         * If not null and exhibit is non-null, must be the directory in exhibit.
         */
        private final String gallerySection;

        /**The immutable details of the exhibit close to this ad, or null. */
        private final String exhibit;

        /**Get the locale of the visitor, or null if none specified or relevant. */
        public Locale getLocale() { return(locale); }

        /**Get the gallery section (top-level directory) for this ad, or null if none relevant.
         * If non-null and the exhibit is non-null, these two will be consistent.
         */
        public String getGallerySection() { return(gallerySection); }

        /**Get the full name of the exhibit connected with this ad, or null if none relevant.
         * If non-null and the exhibit is non-null, these two will be consistent.
         */
        public String getExhibit() { return(exhibit); }
        }

    /**Can be used to set general parameters for the bean, or can be null.
     * It is up to the bean how or if to interpret this value at all.
     *
     * @return instance so as to allow pure expression constructions of form
     *     (new ...Bean()).setParams().getHTMLBodyInsert(xxx, rnd)
     */
    public AdBeanInterface setParams(String rawParams);

    /**Can make an HTML body-text insert in the given context.
     */
    public boolean canMakeHTMLBodyInsert(Context ctxt);

    /**Make HTML body text in the given context, or null if none can be made.
     * This takes a random-number source and the user's locale
     * and generates a segment of HTML to insert into a page.
     * <p>
     * Returns the empty string if not possible.
     */
    public String getHTMLBodyInsert(Context ctxt, Random source);
    }
