/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.ads;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.Tuple.Pair;
import org.hd.d.pg2k.svrCore.location.GeoUtils;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.*;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.util.*;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

/**Advertising-related utility functions.
 * One advantage of having code here rather than in-line in a JSP
 * is that it is pre-compiled off-line for speed and robustness;
 * code here is also potentially easier to test.
 */
public final class AdUtils
    {
	/**Prevent construction of an instance. */
    private AdUtils() { }


    /**Threshold (in thousands) when we switch from logging individual hits/clicks to thousands; strictly positive.
     * Raising this value improves accuracy when stats come in slowly
     * and/or servers are frequently restarted,
     * but increases stats traffic and log size.
     * <p>
     * Ideally this should be large enough that loss of a nearly full "k-click"
     * should not be hugely significant, which implies a value of maybe 10 or more.
     * <p>
     * We should set this as low as we can bear in terms of
     * potential statistical inaccuracy to keep log traffic reasonable.
     */
    private static final int KHIT_THRESHOLD = 4;

    /**Extra weighting in ad layout/format decisions given to local hits/stats; non-negative.
     * Although all ad layout and click-through stats are recorded globally,
     * some of them are also recorded locally, in a somewhat volatile form.
     * (Such local stats might be lost at an upgrade for example,
     * or be less robust in the face of local disc errors.)
     * <p>
     * A weighting of zero means to not treat local stats specially,
     * and just use the global stats (which includes the local stats).
     * <p>
     * A weighting of 1 gives local stats equal weight to global stats,
     * though clicks (etc) are recorded once in the global stats already,
     * so that means that local stats are effectively given somewhat higher weight.
     * <p>
     * Note that the weighting is usually in terms of the entire eCPM/eCTR/etc
     * (providing that there are enough local pageviews to keep noise acceptable,
     * else the local stats may be ignored or used in a more-noise-tolerant way),
     * and thus even low-traffic mirrors can show significant favour to local mores.
     */
    public static final int LOCAL_AD_STATS_WEIGHTING = 4;

    /**Minimum number of pageviews to accept one of the local stats values; strictly positive.
     * Because the stats system switches from recording individual page views and clicks
     * to recording thousands of page views and clicks at a round 1000 mark,
     * this value should be either 1000 or less (though that may be too small)
     * or a multiple of 1000 (but not so high as to exclude local stats entirely),
     * or enough below the KHIT_THRESHOLD to be immune to the large-count lumpiness.
     * <p>
     * NB: Google AdWords seems to regard 1000 as a minimum sensible sample size.
     * <p>
     * A value of 1000 up to some fraction of our KHIT_THRESHOLD may be good.
     */
    private static final int MIN_LOCAL_STATS_PAGEVIEWS = Math.max(1000, 334*KHIT_THRESHOLD);


    /**Ranking threshold assumed to indicate "well" ranked amongst many peers; strictly positive. */
    private static final int TOP_N_RANK_AD_LIMIT = 31 + Rnd.fastRnd.nextInt(19);

    /**What fraction of ad slots are shown in "lite" mode; strictly positive. */
    private static final int LITE_AD_SHOW_FRACTION = 5 + Rnd.fastRnd.nextInt(3);

    /**Show ads or not on the current page?
     * Returns false if ads are generally disabled, etc,
     * though even in the case that the user session has been set to "lite" mode,
     * will return true a small fraction of the time so as to show occasional ads.
     * <p>
     * We don't show ads if they've been explicitly disabled system-wide,
     * or if the URI seems to indicate possibly-sensitive content...
     *
     * @return true if it is OK to show an ad
     */
    public static final boolean showAds(final DataSourceBean dsb,
                                        final HttpServletRequest request)
        {
        if(!AdConsts.AD_SUPPORT)
            { return(false); }

        try
            {
            final GenProps gp = dsb.getGenProps(-1);
            if(!gp.getWEBSVR_SIMPLEAD_ALLOW())
                { return(false); }

            // Don't show ads against potentially-sensitive content...
            final Object fro = request.getAttribute(ThroughputMonitorFilterPG2K.PNAME_SAVED_REQ_URI);
            final String canonReqURI = (fro instanceof String) ? canonicaliseURI((String) fro) : request.getRequestURI();
            if(GenUtils.isSensitive(canonReqURI, dsb.getAllExhibitProperties(-1), gp))
                {
if(IsDebug.isDebug) { System.err.println("Page may be sensitive: "+request.getPathInfo()); }
                return(false);
                }

            // Show ads anyway about 10% of the time in "lite" mode.
            final SessionVarBean svb = SessionVarBean.getExtantSessionVars(request, false);
            if((svb != null) && svb.isSessionVarLiteUI() && (Rnd.fastRnd.nextInt(LITE_AD_SHOW_FRACTION) != 0))
                { return(false); }
            }
        catch(final Exception e)
            {
            e.printStackTrace(); // Note unexpected error.
            return(false); // And avoid showing any ad...
            }

        return(true); // OK to show an ad...
        }

    /**Decides if an underperforming ad slot should be skipped (ie not filled).
     * This uses the system variable info indicated to decide how well
     * the ad slot is performing, and then choses randomly with an appropriate
     * probability whether it should be filled or skipped.
     * <p>
     * Reducing the number of (non-performing) ads shown to users
     * probably improves the user experience in various ways
     * and probably also increases the effective revenue from other slots.
     * <p>
     * The system variable may represent something such as page views or CTR.
     * <p>
     * This must not throw any exceptions so as to avoid killing any JSP
     * that it happens to be called from.
     * <p>
     * TODO: May want to factor in local stats specially at some point.
     *
     * @param vars  handle on the system variables; never null
     * @param history  event history (definition) to be examined; never null
     * @param eventPeriod  period we are interested in from event history;
     *     must be valid for the given event and non-null
     * @param slotIdentifier  the unique ID for advertising "events" for this slot;
     *     must be valid value and type for specified history and non-null
     *
     * @return true if the slot should be skipped this time,
     *     false if it should be filled
     */
    public static boolean skipUnderperformingAdSlot(final SimpleVariablePipelineIF vars,
                                                    final SimpleVariableDefinition history,
                                                    final EventPeriod eventPeriod,
                                                    final Object slotIdentifier)
        {
        // First-pass sanity check on arguments.
        if((vars == null) || (eventPeriod == null) ||
           (history == null) || !history.isEvent())
            { throw new IllegalArgumentException(); }

        // For a null or inappropriate ID we play safe and skip the ad.
        if((slotIdentifier == null) || !history.checkType(slotIdentifier))
            { return(true); }

        // If we seem to have sufficient/meaningful click-through data
        // then we can use it to help us decide about showing an ad here.
        // Show the ad if the exhibit is any of:
        //   * well-ranked recently by click-through
        //   * in the all-time top-N-ranked clicked-thru slots
        //   * reasonably-highly ranked from all distinct values recorded
        //   * randomly about 1 time in N (to continue testing the slot slowly).
        // The end result is that only a small number of relatively-high CTR
        // URIs should get ads shown in this slot,
        // but all exhibits will get an ad shown in this slot occasionally.
 
        try
            {
            // If the slot has ranked well "recently"
            // then fill the slot again now.
            // (We check the "prev" first for efficiency,
            // since it should always be able to come from local cache.)
            // TODO: consider normalising to CTR.
            final EventVariableValue eventsYesterday = vars.getEventValue(history, eventPeriod, false);
            if(eventsYesterday.getRank(slotIdentifier) < (eventsYesterday.getTotalDistinctValues()/3))
                { return(false); /* Show the ad! */ }
            final EventVariableValue eventsToday = vars.getEventValue(history, eventPeriod, true);
            if(eventsToday.getRank(slotIdentifier) < (eventsToday.getTotalDistinctValues()/2))
                { return(false); /* Show the ad! */ }

            // Now we look into the full collected history for this value.
            final EventVariableValue[] all = vars.getEventValues(history, eventPeriod, 0, null);

            // If no "all" history at all or not enough to be significant
            // then fill this ad slot until there is more data available.
            if(all.length < 1) { return(false); /* Show the ad! */ }
            final EventVariableValue allEVV = all[0];
            if(allEVV == null) { return(false); /* Show the ad! */ }
            final int totalDistinctValues = allEVV.getTotalDistinctValues();
            if(totalDistinctValues < TOP_N_RANK_AD_LIMIT*2) { return(false); /* Show the ad! */ }

            // The better that this slot has performed over time,
            // then the lower the chance of us "skipping" this slot.
            // We leave a roughly 1-in-N chance of showing the ad anyway
            // so that we can continue to gather stats, for example.
            // We ensure that the top-ranked slot is always filled.
            return((Rnd.fastRnd.nextInt(1 + Math.min(allEVV.getRank(slotIdentifier), totalDistinctValues)) >
                    Math.max(TOP_N_RANK_AD_LIMIT, totalDistinctValues/TOP_N_RANK_AD_LIMIT)));
            }
        catch(final IOException e)
            {
            // Something went wrong: show an ad to pay for repairs!
            System.err.println("WARNING: skipUnderperformingAdSlot() caught an IOException.");
            return(false);
            }
        catch(final Throwable e)
            {
            // Something went wrong: show an ad to pay for repairs!
            System.err.println("WARNING: skipUnderperformingAdSlot() caught an Exception.");
            return(false);
            }
        }

    /**Maximum leading portion of request URI that we will consider significant.
     * This is based on the maximum legitimate exhibit name length
     * (for when all or part of an exhibit name is embedded in a URI)
     * plus an allowance for overhead.
     */
    private static final int MAX_SIGNIFICANT_URI_PREFIX = ExhibitName.MAX_NAME_LENGTH + 32;

    /**Time we wait for a user to click through on ads (of the order of minutes), in ms. */
    private static final int CLICKTHRU_MIN_WAIT_TIME_MS = 10 * 60 * 1000 +
                                        Rnd.fastRnd.nextInt(9 * 60 * 1000);

    /**Create and post the handler for a click-through if possible and returns the unique listenerID.
     * This replaces any existing listener for this users,
     * eg if the user goes back and clicks then we won't capture it.
     * <p>
     * This returns null if we could not create a listener.
     *
     * @param request  client's HTTP request; never null
     * @param vars  where stats updates are posted; never null
     * @param otherTags  any non-null (and non-empty) values
     *     (which should be distinctive but short)
     *     are recorded to GENSTATS_STRING_GLOBAL_EVENT upon a click-through
     *
     * @return listenerID, or null if one could not be generated
     */
    public static String createAndFileClickThruListener(final HttpServletRequest request,
                                                        final SimpleVariablePipelineIF vars,
                                                        final String ... otherTags)
        {
        if(vars == null) { return(null); }

        // Take a copy of any tags passed in to fire upon click-through.
        final String otherTagsCopy[] = (otherTags == null) ? null : otherTags.clone();

        // Treat the exhibitName as null unless valid at least syntactically.
        final Object eno = request.getAttribute("exhibitName");
        final String exhibitName =((eno instanceof String) &&
                                ExhibitName.validNameSyntax((String) eno)) ?
            ((String) eno) : null;

        final String dpID = WebUtils.generateUserDataPointID("ClickThru-", request);
        if(dpID == null) { return(null); }


        // Compute the unique/canonical URI for the page displaying the ad,
        // or null if not available.
        final Object fro = request.getAttribute(ThroughputMonitorFilterPG2K.PNAME_SAVED_REQ_URI);
        final String canonReqURI = (fro instanceof String) ? canonicaliseURI((String) fro) : null;


        // Compute expiry time (with a random element).
        final long expireBy = System.currentTimeMillis() + CLICKTHRU_MIN_WAIT_TIME_MS +
                Rnd.fastRnd.nextInt(3*CLICKTHRU_MIN_WAIT_TIME_MS);

        // Capture the visitor's IP address for later (in case they vote).
        InetAddress ca = null;
        try { ca = InetAddress.getByName(request.getRemoteAddr()); }
        catch(final UnknownHostException e) { } // Ignore errors...
        final InetAddress clickerIPAddr = ca;

        // Create new listener.
        final StatsSink.AbstractStatsListener newAsl =
                new StatsSink.AbstractStatsListener(dpID, expireBy)
            {
            public final String handle(final Map<String, String[]> parameters)
                {
                if(parameters != null)
                    {
                    // We must see a non-empty url parameter (for AdSense-type click-throughs).
                    final String[] values = parameters.get("url");
                    if((values != null) && (values.length == 1) &&
                            (values[0] != null) && (values[0].length() > 0))
                        {
                        try
                            {
                            // If this was for a bona-fide catalogue page,
                            // then log the click-through for that exhibit,
                            // else just log as a null (valueless) event.

                            // Post vote event value to correct event stream...
if(IsDebug.isDebug) { System.out.println("Click-through registered for: " + exhibitName + " by " + dpID + " to " + values[0]); }
                            vars.setVariable(new SimpleVariableValue(
                                    SystemVariables.ACCESSPATTERN_CLICKTHROUGH,
                                    (exhibitName == null) ? null : ExhibitName.getFileComponent(exhibitName)));

                            if(clickerIPAddr != null)
                                {
                                // Note location of clicker (as ccTLD or region).
                                // We are prepared to spend some time on this as clicking-through is valuable
                                // and we've possibly already collected and cached any needed data.
                                final String location = GeoUtils.getRegionByAddress(clickerIPAddr, false);
                                vars.setVariable(new SimpleVariableValue(
                                        SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                                        "clickerLocation=" + location));
                                vars.setVariable(new SimpleVariableValue(
                                        SystemVariables.GENSTATS_STRING_LOCAL_EVENT,
                                        "clickerLocation=" + location));
                                }

                            // Note page URI that generated the click.
                            if(canonReqURI != null)
                                {
if(IsDebug.isDebug) { System.out.println("Click-through URI: " + canonReqURI + " by " + dpID + " to " + values[0]); }
                                vars.setVariable(new SimpleVariableValue(
                                        SystemVariables.ACCESSPATTERN_CLICKTHROUGH_BY_URI,
                                        canonReqURI));
                                }

                            // Note any other (non-null, non-empty)
                            // tags for click-through.
                            if(otherTagsCopy != null)
                                {
                                for(final String tag : otherTagsCopy)
                                    {
                                    if((tag == null) || (tag.length() == 0)) { continue; }
                                    // Log it globally and locally.
                                    vars.setVariable(new SimpleVariableValue(
                                            SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                                            tag));
                                    vars.setVariable(new SimpleVariableValue(
                                            SystemVariables.GENSTATS_STRING_LOCAL_EVENT,
                                            tag));
                                    }
                                }
                            }
                        catch(final IOException e)
                            { e.printStackTrace();  /* Just absorb errors. */ }
                        }
                    }

                // No redirection; no user should be visiting this page...
                return(null);
                }
            };

        // Get the listener ID...
        final String listnerID = newAsl.uniqueListenerID;

        // File the request!
        StatsSink.addListenerForDataPoint(newAsl);

        return(listnerID);
        }

    /**Canonicalise URI (lower-case and truncate if need be; null is returned as null. */
    public static String canonicaliseURI(final String uri)
        {
        if(uri == null) { return(null); }

        // Compute the unique/canonical URI for the page displaying the ad.
        // Note that because some parts of our URIs are case-insensitive
        // we lower-case the recorded URI to collapse the state-space a little.
        // We truncate the URI length again to trim the state space.
        final String lcReqURI = uri.toLowerCase();
        return((lcReqURI.length() <= MAX_SIGNIFICANT_URI_PREFIX) ?
                lcReqURI : lcReqURI.substring(0, MAX_SIGNIFICANT_URI_PREFIX));
        }


    /**Compute value (in arbitrary units) of named thousand CPM-ad-display impressions; non-negative.
     * This will apply a default value if no specific one applies.
     */
    public static int computeECPM(final GenProps gp, final String statName)
        {
        assert(gp != null);

        // Look for generic key value for this stat and parse any value as int.
        final String s = gp.getGen().get("eCPM." + statName);
        if(s != null)
            {
            try { return(Integer.parseInt(s, 10)); }
            catch(final NumberFormatException e)
                { e.printStackTrace(); /* Whinge but fall through to default. */ }
            }

        // Return default value of 0.
        return(0);
        }

    /**Compute value (in arbitrary units) of named click-though; non-negative.
     * This will apply a default value if no specific one applies.
     */
    public static int computeECPC(final GenProps gp, final String statName)
        {
        assert(gp != null);

        // Look for generic key value for this stat and parse any value as int.
        final String s = gp.getGen().get("eCPC." + statName);
        if(s != null)
            {
            try { return(Integer.parseInt(s, 10)); }
            catch(final NumberFormatException e)
                { e.printStackTrace(); /* Whinge but fall through to default. */ }
            }

        // Return default value of 1.
        return(1);
        }

    /**Minimum sample period to decide which is best-peforming layout, ms; strictly positive.
     * This should be long enough to cover:
     * <ul>
     * <li>At least one rotation of all possible major ad-layout styles.
     * <li>At least one rotation of all major/minor ad-layout styles together.
     * <li>At least one day to allow for daily traffic cycles/patterns.
     * <li>At least one week/month to allow for traffic cycles/patterns.
     * <li>Long enough to survive glitches such as frequent restarts that
     *     "lose" low-frequency events.
     * </ul>
     * but not so long as to incur a huge calculation penalty
     * nor make it very slow to respond to changes.
     * <p>
     * (It may be wise to factor in a small component of the all-time numbers.)
     * <p>
     * A length of just over a week-multiple (to give some extra emphasis to
     * patterns pertient to the day just coming up) may be good.
     */
    private static final long BEST_AD_LAYOUT_SAMPLE_PERIOD_MIN_MS = 35 * 24 * 3600 * 1000L;

    /**Ad-layout selection period expressed in VLONG units; strictly positive. */
    private static final int BEST_AD_LAYOUT_SAMPLE_PERIOD_MIN_VLONG_UNITS = 1 +
        (int) (BEST_AD_LAYOUT_SAMPLE_PERIOD_MIN_MS / SystemVariables.EVENT_INTERVAL_VLONG_TERM_MS);

    /**Extra weighting given to "within last week" clickthrough stats (normal==1); strictly positive.
     * This is based on the assumption that recent behaviour is the most telling of the neat future.
     */
    private static final int WT_LAST_WEEK = 2;

    /**Extra weighting given to "this day last week" clickthrough stats (normal==1); strictly positive.
     * This is based on the assumption that surfers' behaviour shows a weekly cycle.
     */
    private static final int WT_THIS_DAY_LAST_WEEK = 3;

    /**Recent-stats weighting period (1 week) expressed in VLONG units; strictly positive. */
    private static final int WT__VLONG_UNITS = Math.max(1,
        (int) ((7 * 24 * 3600 * 1000L) / SystemVariables.EVENT_INTERVAL_VLONG_TERM_MS));

    /**Return optimal value(s) best-first of parameter based on past data, recomputing/cacheing as needed; never null.
     * Any recomputation is done asynchronously.
     * We may wait a very short time for the computation to finish
     * but if the result isn't very quickly available then we return a default
     * (and cache it for a short while to help avoid concurrent recomputations).
     *
     * @param defaultValue  the default value to use
     *     when we cannot (immediately) compute the optimum; never null
     *
     * @return immutable computed optimum values best first,
     *     or the (non-null) default as a singleton
     *     if there is no (valid) cached optimium value
     *     and we are still (re)computing it asynchronously;
     *     never null, never empty, never containing nulls nor duplicates
     */
    @SuppressWarnings("unchecked")
    private static <E extends Enum<E>> List<E> _computeOptimalLayout(
                        final String callerName,
                        final DataSourceBean dsb,
                        final String layoutKHitsEventPrefix,
                        final String ctEventPrefix,
                        final DataSourceBean.UnlinkedKey cacheKey,
                        final E defaultValue)
        throws InterruptedException
        {
        assert(callerName != null);
        assert(dsb != null);
        assert(layoutKHitsEventPrefix != null);
        assert(ctEventPrefix != null);
        assert(cacheKey != null);
        assert(defaultValue != null);

        // Retrieve any currently-cached value.
        // or atomically create a temporary default one.
        Tuple.Pair<List<E>, Long> cachedResult;
        while(((null == (cachedResult = (Tuple.Pair<List<E>, Long>) dsb.getUnlinkedValue(cacheKey)))) ||
              (cachedResult.second.longValue()) < System.currentTimeMillis())
            {
            // Store "cheat" guessed-best value immediately
            // (providing that we can avoid overwriting a new non-stale value):
            //  1) To prevent/minimise re-computation races.
            //  2) In case something stops us doing the computation.
            // We store the defaultValue supplied whilst calculating,
            // and we allow plenty of time to do the recomputation,
            // but not so long as to prevent retry reasonably quickly
            // after a temporary problem (eg connectivity).
            // We throw in the random time component to help avoid collisions/races
            // with the activities of other parts of the system.
            // We carefully avoid overwriting any non-null/non-stale value
            // that materialises just after we get into the loop
            // by use of the atomic primitives available for the DSB store.
            final Tuple.Pair<List<E>, Long> holdingValue = new Tuple.Pair<List<E>, Long>(Collections.singletonList(defaultValue),
                System.currentTimeMillis() + 300123 + Rnd.fastRnd.nextInt(600123));
            final boolean anotherThreadWorking;
            if(null == cachedResult)
                {
                // If a value has been stored against this key
                // then another thread has started computing this value.
                anotherThreadWorking = (null != dsb.putIfAbsentUnlinkedValue(cacheKey, holdingValue));
                }
            else
                {
                // If the stored value has changed under our feet
                // then another thread has started computing this value.
                anotherThreadWorking = !dsb.replaceUnlinkedValue(cacheKey, cachedResult, holdingValue);
                }

            // If we detect that another thread has started working on this value
            // then abort now (returning the default) to avoid duplicate computations.
            if(anotherThreadWorking) { return(holdingValue.first); }

            final long recalcStart = System.currentTimeMillis();
            // Treat this task as I/O-bound, which early ones will probably be.
            /* final Future<E> f = */ ThreadUtils.nonCPUThreadPool.submit(new Callable<List<E>>(){
                /**Do the recomputation. */
                public final List<E> call()
                    {
                    if(IsDebug.isDebug) { System.out.println("["+callerName+"(): recalculating optimal format/layout...]"); }

                    // Recompute...
                    try
                        {
                        final org.hd.d.pg2k.svrCore.props.GenProps gp = dsb.getGenProps(-1);

                        final long prevInterval = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis()) - 1;
                        final BitSet whichIntervals = new BitSet(BEST_AD_LAYOUT_SAMPLE_PERIOD_MIN_VLONG_UNITS);
                        whichIntervals.set(0, BEST_AD_LAYOUT_SAMPLE_PERIOD_MIN_VLONG_UNITS);
                        final SimpleVariableDefinition def = SystemVariables.GENSTATS_STRING_GLOBAL_EVENT;
                        assert(def.isEvent() && (def.getType() == SimpleVariableDefinition.TYPE_STRING));

                        // Get global stats for enough historical periods, plus current.
                        // We put the "current" value at the end for efficiency
                        // and to conserve any recent-values weighting for the initial whole days/intervals.
                        final List<EventVariableValue> evvs = new ArrayList<EventVariableValue>(Arrays.asList(dsb.getEventValues(def,
                            EventPeriod.VLONG, prevInterval, whichIntervals)));
                        evvs.add(dsb.getEventValue(def, EventPeriod.VLONG, true));

                        // Optionally get local stats for the period matching the global stats.
                        final List<EventVariableValue> evvsLocal;
                        if(LOCAL_AD_STATS_WEIGHTING > 0)
                            {
                            final SimpleVariableDefinition defLocal = SystemVariables.GENSTATS_STRING_LOCAL_EVENT;
                            assert(defLocal.isEvent() && (defLocal.getType() == SimpleVariableDefinition.TYPE_STRING));
                            // We put the "current" value at the end for efficiency
                            // and to conserve any recent-values weighting for the initial whole days/intervals.
                            evvsLocal = new ArrayList<EventVariableValue>(Arrays.asList(dsb.getEventValues(defLocal,
                                    EventPeriod.VLONG, prevInterval, whichIntervals)));
                            evvsLocal.add(dsb.getEventValue(defLocal, EventPeriod.VLONG, true));
                            }
                        else { evvsLocal = Collections.emptyList(); }

                        // Estimate value of a thousand impressions (VPM)
                        // of a given format (in arbitrary units)
                        // to try to chose the relatively-optimal format.
                        //
                        // Total up click-throughs for and pages in each layout
                        // (ie generate properly normalised measure of CTR)
                        // and add in any "CPM" external (non-clickthrough) factors
                        // and record the "best" layout found so far.
                        // Complain (gently) if no useful data is found.

                        // In-order (highest/best first) VPM value for each format.
                        final SortedMap<Double, E> best = new TreeMap<Double, E>(Collections.reverseOrder());
                        for(final E pos : (E[]) defaultValue.getClass().getEnumConstants())
                            {
                        	// Global stats.
                            int clicks = 0;
                            int pageViews = 0;
                            int realPageViews = 0;

                            // Name of the event used to record (each thousand)
                            // page hits in the given format/layout.
                            final String nameKHits = layoutKHitsEventPrefix + pos;
                            final String oneHitName = make1HitName(nameKHits);
                            // Get eCPM value.
                            final int eCPM = computeECPM(gp, nameKHits);

                            // Name of event to record ad click-through
                            // in the given format/layout.
                            final String nameCT = ctEventPrefix + pos;
                            // Get eCPC value.
                            final int eCPC = computeECPC(gp, nameCT);

                            // Tot up global stats.
                            for(int i = evvs.size(); --i >= 0; )
                                {
                                final EventVariableValue evv = evvs.get(i);
                                if(evv == null) { continue; }
                                final int multiplier = (i > WT__VLONG_UNITS) ? 1 :
                                    ((i == WT__VLONG_UNITS) ? WT_THIS_DAY_LAST_WEEK : WT_LAST_WEEK); 
                                final int count = evv.getCount(nameCT);
                                clicks += (multiplier > 1) ? (multiplier * count) : count;
                                int pv = 1000 * evv.getCount(nameKHits);
                                if(oneHitName != null)
                                    { pv += evv.getCount(oneHitName); }
                                pageViews += (multiplier > 1) ? (multiplier * pv) : pv;
                                realPageViews += pv; // Always unweighted.
                                }

                        	// Tot up local stats, if used.
                            int clicksLocal = 0;
                            int pageViewsLocal = 0;
                            int realPageViewsLocal = 0;
                            for(int i = evvsLocal.size(); --i >= 0; )
                                {
                                final EventVariableValue evvLocal = evvsLocal.get(i);
                                if(evvLocal == null) { continue; }
                                final int multiplier = (i > WT__VLONG_UNITS) ? 1 :
                                    ((i == WT__VLONG_UNITS) ? WT_THIS_DAY_LAST_WEEK : WT_LAST_WEEK); 
                                final int count = evvLocal.getCount(nameCT);
                                clicksLocal += (multiplier > 1) ? (multiplier * count) : count;
                                int pv = 1000 * evvLocal.getCount(nameKHits);
                                if(oneHitName != null)
                                    { pv += evvLocal.getCount(oneHitName); }
                                pageViewsLocal += (multiplier > 1) ? (multiplier * pv) : pv;
                                realPageViewsLocal += pv; // Always unweighted.
                                }

                            if((pageViews > 0) || (pageViewsLocal > 0))
                                {
                                final double ctr = clicks / (double) Math.max(1, pageViews);
                                final double ctrLocal = clicksLocal / (double) Math.max(1, pageViewsLocal);

								// We avoid fully factoring in the local CTR if not enough data collected yet,
                                // but we'll still fold in the local data in a less noisy/dramatic way.
                                // We generally insist on a minimum sample size to use local stats,
                                // unless the global stats aren't that much better (by sample count).
                                final boolean notEnoughLocalStats =
                                	(realPageViewsLocal < MIN_LOCAL_STATS_PAGEVIEWS) &&
                                    (realPageViewsLocal < realPageViews/2);
                                final double ctrWeighted = notEnoughLocalStats ?
                                    ((clicks + LOCAL_AD_STATS_WEIGHTING*clicksLocal) / (double) Math.max(1, pageViews + LOCAL_AD_STATS_WEIGHTING*pageViewsLocal)) :
                                	(ctr + (LOCAL_AD_STATS_WEIGHTING*ctrLocal)) / (1+LOCAL_AD_STATS_WEIGHTING);
                                final double vpmWeighted = eCPM + (eCPC * ctrWeighted * 1000);

dsb.log("["+callerName+"(): calc VPM for "+pos+" is "+((float)vpmWeighted)+" at a weighted CTR of "+(100*(float)ctrWeighted)+"% (local="+(100*(float)ctrLocal)+"%"+(notEnoughLocalStats?" (noisy)":"")+", global="+(100*(float)ctr)+"%).]");

                                // Record this VPM (if positive).
                                if(vpmWeighted > 0)
                                    { best.put(new Double(vpmWeighted), pos); }
                                }
                            }

                        if(!best.isEmpty())
                            {
dsb.log("["+callerName+"(): optimal ad format/layout calculated to be: "+best.get(best.firstKey())+" with a VPM of "+(best.firstKey().floatValue())+": took "+(System.currentTimeMillis()-recalcStart)+"ms.]");

                            // Put off next computation until shortly after
                            // the start of the next interval
                            // (when we will have new data available)
                            // but not immediately after so as to avoid
                            // colliding with other recomputation work,
                            // keeping different sites/mirrors in reasonable sync.
                            final long nextInterval = prevInterval + 2;
                            final long recompTime = EventPeriod.VLONG.getIntervalStartTime(nextInterval) +
                                30123 + Rnd.fastRnd.nextInt(300123); // Keep site sync within ~5min.
                            // Extract the best-first formats as an immutable List.
                            final List<E> bestFirst = Collections.unmodifiableList(new ArrayList<E>(best.values()));
                            // Cache the result against this DataSourceBean unconditionally.
                            final Tuple.Pair<List<E>, Long> result = new Tuple.Pair<List<E>, Long>(
                                            bestFirst, recompTime);
                            dsb.putUnlinkedValue(cacheKey, result);
if(IsDebug.isDebug) { dsb.log("["+callerName+"(): next recalc not before "+(new Date(recompTime))+".]"); }

                            return(bestFirst);
                            }
                        else // Complain if we could not find the data needed...
                            {
System.err.println("WARNING: "+callerName+"(): recalc did not find enough data to compute optimal format/layout, using: " + defaultValue);
                            }
                        }
                    catch(final IOException e)
                        {
                        // Note that we were blown out of the water this time...
System.err.println("WARNING: "+callerName+"(): recalc stopped by IOException: " + e.getMessage());
                        }
                    catch(final Throwable t)
                        {
                        // Note that we were blown out of the water this time...
System.err.println("WARNING: "+callerName+"(): recalc stopped by unexpected Throwable: " + t.getMessage());
                        t.printStackTrace();
                        }

                    return(Collections.singletonList(defaultValue)); // No result could be computed.
                    }
                });

// FIXME: This Future.get(wait) call is a problem under JDK 1.5.0_03...
//            // Wait a very short while in case computation finishes quickly.
//            // Otherwise the caller always gets to use the default instead.
//            try { return(f.get(1 + CoreConsts.MAX_INTERACTIVE_DELAY_MS/5, TimeUnit.MILLISECONDS)); }
//            catch(final Exception e) { e.printStackTrace(); }

            // Extract the result.
            // (Guaranteed never to be null since defaultValue was not.)
            cachedResult = (Pair<List<E>, Long>) dsb.getUnlinkedValue(cacheKey); // This *may* be the new result.
            }

        // Check some quick invariants of the returned value.
        assert(cachedResult != null);
        assert(cachedResult.first != null);
        assert(cachedResult.first.size() > 0);

        return(cachedResult.first);
        }

    /**Counters of page/ad basic tower format used, indexed by AdTowerInternalLayout; never null.
     * All elements are initialised to 0 at start-up,
     * so no element is null and the map itself is never changed,
     * and is implicitly thread-safe when used read-only this way.
     * <p>
     * All values will be non-negative, and usually less than 1000.
     * <p>
     * FIXME: should be keyed from DataSourceBean to allow for multiple sites in one VM.
     */
    private static final EnumMap<AdTowerInternalLayout, AtomicInteger> pvTowerInternalFormatHits = new EnumMap<AdTowerInternalLayout, AtomicInteger>(AdTowerInternalLayout.class);

    /**Initialise pvTowerInternalFormatHits to all zero counts. */
    static
        {
        for(final AdTowerInternalLayout atp : AdTowerInternalLayout.values())
            { pvTowerInternalFormatHits.put(atp, new AtomicInteger()); }
        }

    /**Key for thread-safe cache from Web site (DataSourceBean) to computed results for selectAdMidPageInternalLayout(); never null. */
    private static final DataSourceBean.UnlinkedKey _cache_sAMPIL_key = new DataSourceBean.UnlinkedKey("selectAdMidPageInternalLayout");

    /**Preferred mid-page ad format. */
    private static final AdMidPageInternalLayout PREF_MP_FORMAT = AdMidPageInternalLayout.SQAS;

    /**Chose internal layout of ad mid-age slot to be shown on page; never null.
     * Uses algorithm to ensure that all servers should generally choose
     * the same layout at any one time so that, for example,
     * switching between servers should not be disorienting.
     * <p>
     * May collect stats on layout to help monitor/select the most effective.
     *
     * @param request  the RQPNAME_MIDPAGEINTLAYOUT attribute is set
     *     to indicate which format has been selected
     */
    public static final AdMidPageInternalLayout selectAdMidPageInternalLayout(
                                            final DataSourceBean dsb,
                                            final HttpServletRequest request)
        throws InterruptedException
        {
        final AdMidPageInternalLayout result;

        // Pick using prime-ish value much larger than number of available formats
        // so that a large chunk of ads will be in the computed "best" format.
        final Random r = (new Random(System.currentTimeMillis() / (17 * 60 * 1000)));
        final int format = r.nextInt(17);
        switch(format)
            {
            case 0: // Extra showing for preferred format.
                result = PREF_MP_FORMAT;
                break;

            case 1:
                result = AdMidPageInternalLayout.LBAS;
                break;

            case 2:
                result = AdMidPageInternalLayout.MRAS;
                break;

            case 3:
                result = AdMidPageInternalLayout.MRTF;
                break;

            case 4:
                result = AdMidPageInternalLayout.MRCM;
                break;

            case 5:
                result = AdMidPageInternalLayout.SQAS;
                break;

            case 6:
                result = AdMidPageInternalLayout.sqAS;
                break;

            default: // Use computed "best" style most of the time.
                {
                final List<AdMidPageInternalLayout> bestLayouts =
                    _computeOptimalLayout("selectAdMidPageInternalLayout",
                                dsb,
                                AdConsts.MPINTLAYOUT_KHITS_EVENT_PREFIX,
                                AdConsts.MPINTLAYOUT_EVENT_PREFIX,
                                _cache_sAMPIL_key,
                                PREF_MP_FORMAT);
                final int s = bestLayouts.size();
                assert(s > 0);
                // Have smoothish distribution between best formats,
                // but the top format always gets at least half the slots.
                result = r.nextBoolean() ? bestLayouts.get(0) : bestLayouts.get(r.nextInt(1+r.nextInt(s)));
                break;
                }
            }

        assert(result != null);


        // Note which mid-page internal format is being used,
        // and on each 1000 uses, log to the central stats.
        // We do this to reduce the load on the stats system.
        // Ignore hits from probable spiders.
        if(!WebUtils.requestProbablyFromSpider(request))
            {
            final AtomicInteger count = pvMPInternalFormatHits.get(result);
            final String kHitName = AdConsts.MPINTLAYOUT_KHITS_EVENT_PREFIX + result;
            countPageImpressions(count, dsb, kHitName);
            }


        // Automatically note with the request which tower format is being used.
        // This enables click-tracking and CTR measurement later.
        request.setAttribute(AdConsts.RQPNAME_MPINTLAYOUT, AdConsts.MPINTLAYOUT_EVENT_PREFIX + result);

        assert(result != null);
        return(result);
        }

    /**Count page impressions in this layout/ad-style.
     * Logs thousands of impressions (kHits) to reduce event traffic,
     * but can log single hits initially for good resolution
     * and robustness for short-running servers and/or sparse events.
     */
    private static void countPageImpressions(final AtomicInteger count,
                                             final DataSourceBean dsb,
                                             final String kHitName)
        {
        final String oneHitName = make1HitName(kHitName);
        final int countNow = count.getAndIncrement();
        try
            {
        	// Record a k-hit (one thousand clicks or hits).
        	// If recording individual clicks initially
        	// then we click down from 1000 + the higher threshold
        	// rather than down from 1000.
            if(countNow >= 1000 * (1 + ((oneHitName == null) ? 0 : KHIT_THRESHOLD)))
                {
                count.addAndGet(-1000);

                // Log (another) 1000 uses of this page/ad layout style...
                dsb.setVariable(new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                        kHitName));
                dsb.setVariable(new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_LOCAL_EVENT,
                        kHitName));
                }
            else if((oneHitName != null) && (countNow < 1000*KHIT_THRESHOLD))
                {
                // We record the first few hits individually.
                dsb.setVariable(new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                        oneHitName));
                dsb.setVariable(new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_LOCAL_EVENT,
                        oneHitName));
                }
            }
        catch(final IOException e)
            { e.printStackTrace(); /* Log but absorb any error. */ }
        }

    /**Make single-hit event name from k-hit name, or return null if not possible.
     * This is done by replacing the initial "k" with a "1".
     */
    private static String make1HitName(final String kHitName)
        {
        if(!kHitName.startsWith("k")) { return(null); }
        return("1" + kHitName.substring(1));
        }

    /**Counters of page/ad basic tower format used, indexed by AdMInternalLayout; never null.
     * All elements are initialised to 0 at start-up,
     * so no element is null and the map itself is never changed,
     * and is implicitly thread-safe when used read-only this way.
     * <p>
     * All values will be non-negative, and usually less than 1000.
     * <p>
     * FIXME: should be keyed from DataSourceBean to allow for multiple sites in one VM.
     */
    private static final EnumMap<AdMidPageInternalLayout, AtomicInteger> pvMPInternalFormatHits = new EnumMap<AdMidPageInternalLayout, AtomicInteger>(AdMidPageInternalLayout.class);

    /**Initialise pvMPInternalFormatHits to all zero counts. */
    static
        {
        for(final AdMidPageInternalLayout atp : AdMidPageInternalLayout.values())
            { pvMPInternalFormatHits.put(atp, new AtomicInteger()); }
        }


    /**Key for thread-safe cache from Web site (DataSourceBean) to computed results for selectAdTowerInternalLayout(); never null. */
    private static final DataSourceBean.UnlinkedKey _cache_sATIL_key = new DataSourceBean.UnlinkedKey("selectAdTowerInternalLayout");

    /**Chose internal layout of ad tower(s) to be shown on page; never null.
     * Uses algorithm to ensure that all servers should generally choose
     * the same layout at any one time so that, for example,
     * switching between servers should not be disorienting.
     * <p>
     * May collect stats on layout to help monitor/select the most effective.
     *
     * @param request  the RQPNAME_TOWERINTLAYOUT attribute is set
     *     to indicate which format has been selected
     */
    public static final AdTowerInternalLayout selectAdTowerInternalLayout(
                                            final DataSourceBean dsb,
                                            final HttpServletRequest request)
        throws InterruptedException
        {
        final AdTowerInternalLayout result;

        // Pick using prime-ish value larger than number of available formats
        // so that most ads will be in the computed "best" format.
        final Random r = (new Random(System.currentTimeMillis() / (1 * 3600 * 1111)));
        final int format = r.nextInt(11);
        switch(format)
            {
            case 0:
                result = AdTowerInternalLayout.WideMixed;
                break;

            case 1:
                result = AdTowerInternalLayout.WideAS;
                break;

            case 2:
                result = AdTowerInternalLayout.NarrowMixed;
                break;

            case 3:
                result = AdTowerInternalLayout.NarrowShort;
                break;

            default: // Use computed "best" style most of the time.
                {
                final List<AdTowerInternalLayout> bestLayouts =
                    _computeOptimalLayout("selectAdTowerInternalLayout",
                                            dsb,
                                            AdConsts.TOWERINTLAYOUT_KHITS_EVENT_PREFIX,
                                            AdConsts.TOWERINTLAYOUT_EVENT_PREFIX,
                                            _cache_sATIL_key,
                                            AdTowerInternalLayout.WideMixed);
                final int s = bestLayouts.size();
                assert(s > 0);
                // Have smoothish distribution between best formats,
                // but the top format always gets at least half the slots.
                result = r.nextBoolean() ? bestLayouts.get(0) : bestLayouts.get(r.nextInt(1+r.nextInt(s)));
                break;
                }
            }

        assert(result != null);


        // Note which tower internal format is being used,
        // and on each 1000 uses, log to the central stats.
        // We do this to reduce the load on the stats system.
        // Ignore hits from probable spiders.
        if(!WebUtils.requestProbablyFromSpider(request))
            {
            final AtomicInteger count = pvTowerInternalFormatHits.get(result);
            final String kHitName = AdConsts.TOWERINTLAYOUT_KHITS_EVENT_PREFIX + result;
            countPageImpressions(count, dsb, kHitName);
            }


        // Automatically note with the request which tower format is being used.
        // This enables click-tracking and CTR measurement later.
        request.setAttribute(AdConsts.RQPNAME_TOWERINTLAYOUT, AdConsts.TOWERINTLAYOUT_EVENT_PREFIX + result);

        assert(result != null);
        return(result);
        }

    /**Key for thread-safe cache from Web site (DataSourceBean) to computed results for selectPageAdLayout(); never null. */
    private static final DataSourceBean.UnlinkedKey _cache_sPAL_key = new DataSourceBean.UnlinkedKey("selectPageAdLayout");

    /**Choose general ad/page layout; never null.
     * Uses algorithm to ensure that all servers should generally choose
     * the same layout at any one time so that, for example,
     * switching between servers should not be disorienting.
     * <p>
     * May collect stats on layout to help monitor/select the most effective.
     * <p>
     * Note that most effective combo seems to be:
     * <ul>
     * <li>Wide tower on left.
     * <li>Horizontal banner at top.
     * </ul>
     * <p>
     * Computed values may be cached wrt the DataSourceBean variable,
     * so should consistently be called with the same instance for one site.
     *
     * @param probeOnly  this call is to find out what
     *     the layout <em>should be</em>,
     *     and should not count as a page view
     */
    public static final AdTowerPos selectPageAdLayout(final DataSourceBean dsb,
                                                      final HttpServletRequest request,
                                                      final boolean siteIsBusy,
                                                      final boolean probeOnly)
        throws InterruptedException
        {
        // What placement of tower ad will we use, if any?
        // The most useful placement moneywise is probably LEFT or NOWHERE,
        // but rotation to avoid ad-blindness is probably useful too.
        AdTowerPos towerPos;

        // Chose tower-ad style using "hash" such that
        // all servers/mirrors will choose same style at once.
        // We choose a new layout several times per day
        // with a period that rotates relative to a 24-hour cycle
        // so that we don't always make the same choice (or switch)
        // at the same hour of the day.
        final long basicChoice = System.currentTimeMillis() / (3 * 3600 * 1301);
        // Force every style to be used at least once in each rotation.
        // The best-performing style should get a clear majority of slots.
        final Random r = new Random(basicChoice);
        switch(probeOnly ? -1 : r.nextInt(19)) // Prime much larger than number of layouts.
            {
            case 0:
                towerPos = AdTowerPos.RIGHT;
                break;

            case 1:
                towerPos = AdTowerPos.RIGHT_DROPPED;
                break;

            case 2:
                towerPos = AdTowerPos.NOWHERE;
                break;

            case 3:
                towerPos = AdTowerPos.LEFT;
                break;

            default:
                // Use computed "best" style most of the time.
                // Forced in "probe-only" mode.
                {
                final List<AdTowerPos> bestLayouts =
                    _computeOptimalLayout("selectPageAdLayout",
                                          dsb,
                                          AdConsts.PAGE_LAYOUT_KHITS_EVENT_PREFIX,
                                          AdConsts.LAYOUT_STYLE_EVENT_PREFIX,
                                          _cache_sPAL_key,
                                          AdTowerPos.NOWHERE);
                final int s = bestLayouts.size();
                assert(s > 0);
                // When probing we only want to know what the "best" format is.
                if(probeOnly || (s == 1)) { return(bestLayouts.get(0)); }
                // Have smoothish distribution between best formats,
                // but the top format always gets at least 67% of the slots.
                towerPos = (r.nextInt(3) != 0) ? bestLayouts.get(0) : bestLayouts.get(r.nextInt(1+r.nextInt(1+r.nextInt(s))));
                break;
                }
            }

        // Should definitely be set by here...
        assert(towerPos != null);

        // If the ads are not on the left
        // then we avoid ads on top-level pages (such as the home page)
        // or if the site is busy.
        // But if the ads are on the left then we don't do this
        // as it could be very disturbing to move the page content around.
        if(towerPos != AdTowerPos.LEFT)
            {
            // Drop right-hand towers when site is busy to reduce load a little.
            if(siteIsBusy)
                { towerPos = AdTowerPos.NOWHERE; }
            // Drop non-left (LHS) towers for top-level pages.
            else if(request != null)
                {
                final String reqURI = request.getRequestURI();
                final boolean topLevelURI = (reqURI == null) || (reqURI.lastIndexOf('/') <= 0);
                if(topLevelURI)
                    { towerPos = AdTowerPos.NOWHERE; }
                }
            }


        // Note which layout style is being used,
        // and on each 1000 uses, log to the central stats.
        // We do this to reduce the load on the stats system.
        // Ignore hits from probable spiders.
        if(!probeOnly && !WebUtils.requestProbablyFromSpider(request))
            {
            final AtomicInteger count = pvLayoutHits.get(towerPos);
            final String kHitName = AdConsts.PAGE_LAYOUT_KHITS_EVENT_PREFIX + towerPos;
            countPageImpressions(count, dsb, kHitName);
            }


        return(towerPos);
        }

    /**Compute the full event name for recording a clickthrough with a particular layout style. */
    public static String computeCTEventName(final AdTowerPos towerPos)
        {
        return(AdConsts.LAYOUT_STYLE_EVENT_PREFIX + towerPos);
        }

    /**Counters of page/ad basic layout-style used, indexed by AdTowerPos; never null.
     * All elements are initialised to 0 at start-up,
     * so no element is null and the map itself is never changed,
     * and is implicitly thread-safe when used read-only this way.
     * <p>
     * All values will be non-negative, and usually less than 1000.
     * <p>
     * FIXME: should be keyed from DataSourceBean to allow for multiple sites in one VM.
     */
    private static final EnumMap<AdTowerPos, AtomicInteger> pvLayoutHits = new EnumMap<AdTowerPos, AtomicInteger>(AdTowerPos.class);

    /**Initialise pvLayoutHits to all zero counts. */
    static
        {
        for(final AdTowerPos atp : AdTowerPos.values())
            { pvLayoutHits.put(atp, new AtomicInteger()); }
        }

    /**Maximum exhibits to show in drop-in ad-tower replacement text; strictly positive. */
    private static final int DA_MAX_EXHIBITS_TO_SHOW = 4;

    /**Get the HTML content for a fall-back/fill-in ad-tower filler; never null.
     * Returns "" if nothing appropriate to show (or in case of error).
     */
    public static String generateAdTowerFallbackHTML(final ServletContext application,
                                                     final LocaleBean localeBean,
                                                     final int pxWidth,
                                                     final int pxHeight)
        {
        if((application == null) ||
           (localeBean == null) ||
           (pxWidth <= 0) || (pxHeight <= 0))
            { throw new IllegalArgumentException(); }

        try
            {
            final DataSourceBean dataSource = DataSourceBean.getApplicationInstance(application);

            // Title to display (short and sweet).
            // Filled in in case statement.
            final String DA_title;
            final String DA_exhibits[];

            switch(Rnd.fastRnd.nextInt(3))
                {
                // Show the most recent additions...
                case 1:
                    {
                    DA_title = "<a href=\"" + WebConsts.VIRTUAL_COLLECTIONS_NEW_PAGE  + "\" target=\"_top\">" +
                            localeBean.getLocalisedMessage("common.selection.newExhibits.title") +
                            "</a>";
                    DA_exhibits = HTMLThumbnailInsertGenerators.getNewExhibitSelection(application,
                                                               DA_MAX_EXHIBITS_TO_SHOW,
                                                               (Rnd.fastRnd.nextBoolean() ? null : Rnd.fastRnd));
                    break;
                    }

                // Show the "best" DA_exhibits
                // but only if we are relatively idle
                // since this can be expensive.
                // (Else fall through to cheaper default.)
                case 2: if(WebUtils.isLightlyLoaded(application))
                    {
                    DA_title = "<a href=\"" + WebConsts.VIRTUAL_COLLECTIONS_BEST_PAGE  + "\" target=\"_top\">" +
                            localeBean.getLocalisedMessage("common.selection.bestExhibits.title") +
                            "</a>";
                    DA_exhibits = HTMLThumbnailInsertGenerators.getBestExhibitSelection(application,
                                                               DA_MAX_EXHIBITS_TO_SHOW,
                                                               (Rnd.fastRnd.nextBoolean() ? null : Rnd.fastRnd));
                    break;
                    }

                // Default to showing a random JPEG selection...
                default:
                    {
                    DA_title = localeBean.getLocalisedMessage("common.selection.randomJPEGs.title");
                    DA_exhibits = HTMLThumbnailInsertGenerators.getRandomJPEGSelection(application,
                                                               DA_MAX_EXHIBITS_TO_SHOW);
                    break;
                    }
                }

            if(DA_exhibits.length < 1) { return(""); /* Nothing to show. */ }

            final StringBuilder result = new StringBuilder(128 + (64*DA_exhibits.length));

            result.append("<p align=center>");
            result.append("<font size=\"+1\" color=").
                append(PageSkinUtils.GENERIC_AH_COLOUR_QUOTED).
                append('>').append(DA_title).
                append("</font>");
            result.append("<br />");
            result.append(HTMLThumbnailInsertGenerators.makeHTMLExhibitColumnDisplay(DA_exhibits,
                                                          pxWidth,
                                                          pxHeight - 20, // Allow for a header...
                                                          dataSource,
                                                          localeBean,
                                                          WebUtils.isOverloaded(application)));
            result.append("</p>");
            return(result.toString());
            }
        catch(final IOException e)
            {
            return(""); // Absorb the error quietly as this is non-essential.
            }
        }
    }
