/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 22-Jul-02
 * Time: 18:13:44

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.webSvr.ads;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import java.io.IOException;
import java.net.URLEncoder;
import java.util.Random;

/**This implements the e-shirt advertisement bean.
 * This is mainly here as an example of how an AdBean might work;
 * in general AdBean classes should be made available at run-time
 * to the server(s) by being placed in their class-path or by
 * some other mechanism.
 */
public final class EShirtAdBean implements AdBeanInterface
    {
    /**If non-null, overrides normal body text.
     * Accessed under the instance lock.
     */
    private String overrideBodyText;

    /**Can be used to set general parameters for the bean, or can be null.
     * If supplied and is non-null, this overrides the body copy supplied
     * in the bean within the link, though we feel at liberty to trim
     * whitespace.
     */
    public synchronized AdBeanInterface setParams(String rawParams)
        {
        if(rawParams != null) { rawParams = rawParams.trim(); }

        overrideBodyText = rawParams;

        return(this);
        }

    /**Can make an HTML body-text insert in the given context.
     * We can only do this if the full exhibit name is passed to us
     * and if the file is of a suitable type (a .jpg or a .gif)
     * (so we will perform various simple checks).
     */
    public boolean canMakeHTMLBodyInsert(final AdBeanInterface.Context ctxt)
        {
        final String exhibit = ctxt.getExhibit();
        if(exhibit == null) { return(false); } // No exhibit supplied.

        // A full exhibit name must contain at least a '/'.
        if(exhibit.indexOf('/') == -1) { return(false); } // Bad name...

        // We only support certain exhibit types for e-shirt,
        // so just check for the right extensions here.
        if(!exhibit.endsWith(".jpg") &&
           !exhibit.endsWith(".gif"))
            { return(false); }

        // Because we can (we are being compiled in with PG2K itself),
        // we check that the name syntax is valid,
        // but an externally-compiled bean would not do this
        // (it would not be able to).
        if(!ExhibitName.validNameSyntax(exhibit))
            { return(false); }

        // Everything seems OK.
        return(true);
        }

    /**Make HTML body text in the given context, or null if none can be made.
     * This takes a random-number source and the user's locale
     * and generates a segment of HTML to insert into a page.
     * <p>
     * We should not be called if the context is wrong, but we double-check
     * and return "" if we know we can't generate a sensible HTML insert,
     * else we blindly accept the exhibit name we are given in the context.
     */
    public synchronized String getHTMLBodyInsert(final AdBeanInterface.Context ctxt,
                                                 final Random source)
        {
        if(!canMakeHTMLBodyInsert(ctxt)) { return(""); } // Definitely cannot.

        // Insert the exhibit link as an (UTF-8) encoded URL.
        // We reach into PG2K utility classes to help with this,
        // but in general an AdBean could not and need not do this.
        //
        // If we encounter an error then we return ""...
        try
            {
            return((new StringBuilder(512)).
                append("<img src=\"/_static/_3rdPartyImg/EShirt.gif\" width=150 height=58 align=middle> ").
                append("<a target=\"_top\" ").
                append("href=\"http://www.eshirt.it/carrello/gt_urlimg.php?urluserfile=").
                append(URLEncoder.encode(WebUtils.makeExhibitURL(ctxt.getExhibit(), null, null).toString(), CoreConsts.FILE_ENCODING_UTF_8)).
                append("&max=1&newusrid=3d2308ed663ce&").
                append("newfrom=3d2308ed663ce&").
                append("prj_source=0207246D40C5\">").
                append((overrideBodyText == null) ? "Print this on a T-shirt!" : overrideBodyText).
                append("</a>").toString());
            }
        catch(final IOException e)
            {
            e.printStackTrace();
            return(""); // Whoops!
            }
        }
    }
