/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.bgImg;

import org.hd.d.pg2k.svrCore.Rnd;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Random;


/**Servlet to serve dynamic and partially-random HTML-page background image.
 * This serves a randomly-generated image which we would like to be
 * slightly different for each visitor.
 * <p>
 * This is intended to be pleasing on the eye, and possible, if required,
 * to read normal black text over.  So the image is nearly white with
 * a slowly changing 'watermark' image.
 * <p>
 * To minimise bandwidth fetching this, and CPU time generating it,
 * this is designed to be served with the "caching" filter in front of it,
 * so that any one user should typically only fetch a new image once per week
 * or thereabouts, though each user should get a different one.
 * Also, the image dimensions and quality are adjusted to those of a
 * "standard" thumbnail, which does not take you lots of space in memory
 * or on the wire.
 * <p>
 * We may cache the last image generated, so that if the system is busy
 * the last image may be returned again rather than a new one being
 * generated.
 */
public final class BgImageServlet extends HttpServlet
    {
    /**Respond to a GET request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doGet(HttpServletRequest request,
                      HttpServletResponse response)
        throws IOException // , ServletException
        {
        doAction(request, response);
        }

    /**Respond to a HEAD request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doHead(HttpServletRequest request,
                       HttpServletResponse response)
        throws IOException // , ServletException
        {
        doAction(request, response);
        }

    /**Respond to a GET/HEAD request for the content served by this servlet.
     * Returns the exhibit data, with a correct MIME type.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doAction(final HttpServletRequest request,
                         final HttpServletResponse response)
        throws IOException // , ServletException
        {
        // Compute a seed or image selector based on the client IP address.
        // This is made to select from the full range of bg images
        // as uniformly as is reasonably possible.
        final long seed = creationTime ^ request.getRemoteAddr().hashCode();

        // Repeatable pseudo-random numbers seeded by client IP address.
        // We assume that it is fast to get a repeatable sequence out of this,
        // but that the relationship of successive values is non-obvious.
        final Random pRnd = new Random(seed);

        // Chose cache slot based on client IP address.
        final int slot = pRnd.nextInt(_imageCache.length);

        // Serialise access to the cache content...
        // And ensure that we don't use more than one thread at once
        // for generating background images
        // so as to conserve CPU/memory resources.
        byte encodedImage[];
        synchronized(_imageCacheLock)
            {
            encodedImage = _imageCache[slot];
            // Need to compute the image...
            if(encodedImage == null)
                {
                // Chose background image type based on client.
                final int imgType = pRnd.nextInt(BgImageUtils.allDynGens.size());
                final BgImageUtils.DynImageGen dig = BgImageUtils.allDynGens.get(imgType);

                // Build image based on client address.
                encodedImage = BgImageUtils.generateEncodedImage(
                    dig.createImage(pRnd.nextInt()));
                _imageCache[slot] = encodedImage;
                }
            }

        // Set the correct MIME type and result length.
        response.setContentType(BgImageUtils.ETP.mimeType);
        response.setContentLength(encodedImage.length);

        // If this is a HEAD, then return without providing the body.
        if("HEAD".equalsIgnoreCase(request.getMethod()))
            { return; }

        try
            {
            // Write entire bg image in one go for efficiency.
            final OutputStream os = response.getOutputStream();
            os.write(encodedImage);
            os.flush();
            }
        catch(final Throwable t)
            {
            getServletContext().log("BgImageServlet caught Throwable", t);
            }
        }

    /**Construction time for this class.
     * Used to modify the relationship between client IP address
     * and the background selected.
     */
    private static final long creationTime = System.currentTimeMillis();

    /**Minimum number of different backgrounds seen; strictly positive.
     * Possibly good if prime to scatter client's more uniformly across
     * the image set.
     * <p>
     * As this value gets larger, memory consumption will go up.
     * <p>
     * We set this not to use more than about 1%--2% of total memory,
     * although we may find an abnormally-small heap size during startup
     * when this will get computed.
     */
    private static final int MIN_BGS = Math.max(11,
        3 | (int) (Runtime.getRuntime().totalMemory() / (100 * BgImageUtils.dynBgImgMaxBytes)));

    /**Simple cache by "seed" value of background images as raw byte[]; each entry is computed on demand.
     * This is indexed by a hash on the client's IP address,
     * and caches a background image.
     * <p>
     * The size of this array determines how many different images we hold,
     * and more will use more memory.
     * <p>
     * Access the array contents only under the _imageCacheLock.
     * <p>
     * We vary the size of this array a little on each run to vary the mix.
     * The array size stays constant while the class remains loaded.
     */
    private static final byte[][] _imageCache =
        new byte[MIN_BGS + Rnd.fastRnd.nextInt(MIN_BGS)][];

    /**Lock for access to _imageCache. */
    private static final Object _imageCacheLock = new Object();

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 2040799493050847367L;
    }
