/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.bgImg;

import org.hd.d.pg2k.svrCore.ExhibitThumbnails;
import org.hd.d.pg2k.svrCore.MIME.AbstractImageHandler;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.Rnd;

import java.awt.*;
import java.awt.color.ColorSpace;
import java.awt.image.*;
import java.io.IOException;
import java.util.*;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 29-Nov-2003
 * Time: 18:05:45
 */

/**Supporting constants and classes for dynamic background images.
 */
public final class BgImageUtils
    {
    /**Maximum brightness (assuming 0 is black, 255 is white). */
    static final int MAX_Y = 255;

    /**Minimum brightness (assuming 0 is black, 255 is white).
     * Should be less than MAX_Y by at least 0x33 (51)
     * to (almost) force browsers with a 6x6x6 colour cube to show the
     * image in at least two colours.
     */
    static final int MIN_Y = 204;

    /**Type of background image that we dynamically generate.
     * Since we want to be able to read text against this,
     * and not consume too much bandwidth,
     * both implying smoothly varying images,
     * JPEG is probably the best format,
     * especially as it is widely supported.
     */
    static final int DYN_MAP_IMG_TYPE = ExhibitMIME.ET_JPEG;

    /**ExhibitMIME entry for the chosen output type. */
    static final ExhibitMIME.ExhibitTypeParameters ETP =
        ExhibitMIME.getParamsByType(DYN_MAP_IMG_TYPE);

    /**URI of dynamic background-image JPEG; suffix should match MIME type. */
    public static final String dynBgImgRRUL = "/_static/dynImg/bgImg.jpg";

    /**We chose the background image dimensions to match that of a standard thumbnail. */
    public static final int dynBgImgEdgePixels = ExhibitThumbnails.STD_STATIC_IMAGE_TN_LDIM_PX;

    /**The maximum size of image produced (in bytes) is that of a standard thumbnail. */
    public static final int dynBgImgMaxBytes = ExhibitThumbnails.STD_ABS_MAX_BYTES;

    /**Functor/interface for dynamic background image generator.
     * Generates a BufferedImage with the dimensions of a large thumbnail,
     * suitable to use as a background image.
     * <p>
     * The simage will be relatively small (in bytes) and suitable to read
     * black text again (will be fairly light colour).
     */
    public interface DynImageGen
        {
        /**Generate simple random monochrome BufferedImage with values uniformly distributed between max and min Y.
         * This routine may alter the nature of the output based on the seed,
         * which should ideally be distributed evenly across all possible
         * values, like a hashCode.
         */
        public BufferedImage createImage(final int seed);
        }

    /**Simple monochrome random-pixel image.
     * Each pixel is independently chosen.
     */
    public static final class SimpleUniformNoiseMonoImageGen implements DynImageGen
        {
        /**Generate simple random monochrome BufferedImage with values uniformly distributed between max and min Y.
         * This routine may alter the nature of the output based on the seed,
         * which should ideally be distributed evenly across all possible
         * values, like a hashCode.
         * <p>
         * In this implementation, the lowest-order bit selects between
         * "fast" and "good" random number sources.
         */
        public BufferedImage createImage(int seed)
            {
            // Select between random-number generators based on the seed.
            final Random rnd = ((seed & 1) == 0) ? Rnd.fastRnd : Rnd.goodRnd;

            // Create a blank RGB (buffered) image.
            final BufferedImage bi = new BufferedImage(dynBgImgEdgePixels,
                                                       dynBgImgEdgePixels,
                                                       BufferedImage.TYPE_BYTE_GRAY);

            final int range = MAX_Y - MIN_Y + 1;
            for(int y = dynBgImgEdgePixels; --y >= 0; )
                {
                for(int x = dynBgImgEdgePixels; --x >= 0; )
                    {
                    // Compute grey-scale value.
                    final int grey = rnd.nextInt(range) + MIN_Y;
                    // Convert to opaque rgb.
                    final int rgb = (grey << 16) | (grey << 8) | (grey) | 0xff000000;
                    bi.setRGB(x, y, rgb);
                    }
                }

            return(bi);
            }
        }

    /**Simple monochrome uniform colour image.
     * Each pixel is the same colour
     * and will be suitable to read black/dark text against.
     */
    public static final class SimpleUniformMonoImageGen implements DynImageGen
        {
        /**Generate simple random monochrome BufferedImage with all pixels on colour between max and min Y.
         * The colour is derived from the seed.
         */
        public BufferedImage createImage(int seed)
            {
            // Create a blank RGB (buffered) image.
            final BufferedImage bi = new BufferedImage(dynBgImgEdgePixels,
                                                       dynBgImgEdgePixels,
                                                       BufferedImage.TYPE_BYTE_GRAY);

            final int range = MAX_Y - MIN_Y + 1;
            // Compute grey-scale value.
            final int grey = (new Random(seed)).nextInt(range) + MIN_Y;
            // Convert to opaque rgb.
            final int rgb = (grey << 16) | (grey << 8) | (grey) | 0xff000000;
            // Set entire array to one value.
            // TODO: must be more efficient way of doing this...
            for(int y = dynBgImgEdgePixels; --y >= 0; )
                {
                for(int x = dynBgImgEdgePixels; --x >= 0; )
                    {
                    bi.setRGB(x, y, rgb);
                    }
                }

            return(bi);
            }
        }

    /**Makes 2D sine-plus-noise colour image.
     * This is slowly varying, and light,
     * so should be easy to read dark text against.
     * <p>
     * This may revert to mono dependending on the seed.
     */
    public static final class Sine2DColourImageGen implements DynImageGen
        {
        /**Generate 2D sine-based colour with each colour distributed between max and min Y.
         * The image parameters are mostly entirely derived from the seed.
         */
        public BufferedImage createImage(int seed)
            {
            // Select between random-number generators based on the seed.
            final Random pRnd = new Random(seed);

            // Choose mono or colour.
            // FIXME: Fix this so it works in colour...
            final boolean isMono = true; // pRnd.nextBoolean();

            // Extract basic metadata.
            final ColorModel cm = isMono ?
                new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_GRAY),
                                        (new int[]{8}), false, true,
                                        Transparency.OPAQUE,
                                        DataBuffer.TYPE_BYTE) :
                ColorModel.getRGBdefault();
            final int width = dynBgImgEdgePixels;
            final int height = dynBgImgEdgePixels;
            final boolean isAlphaPremultiplied = cm.isAlphaPremultiplied();

            // Construct BufferedImage and copy (raster) data in.
            final WritableRaster raster = cm.createCompatibleWritableRaster(width, height);
            final BufferedImage bi = new BufferedImage(cm, raster, isAlphaPremultiplied, null);

            // Chromance (Y) range, and range of any one colour component.
            final int range = MAX_Y - MIN_Y + 1;

            // Chose the frequency in the x and y directions
            // and a number of whole cycles per edge length.
            // Make this a whole number of cycles so that
            // the image will tile properly.
            final int xFreq = getRndCycles(pRnd);
            final int yFreq = getRndCycles(pRnd);

            // Compute the sine values for the x and y dimensions.
            final int xVals[] = new int[dynBgImgEdgePixels];
            final int yVals[] = new int[dynBgImgEdgePixels];
            final int rd2 = range / 2;
            final int centre = MIN_Y + rd2;
            final double f = (2.0 * Math.PI / dynBgImgEdgePixels);
            for(int i = dynBgImgEdgePixels; --i >= 0; )
                {
                final double ixf = i * f;
                xVals[i] = Math.max(MIN_Y, Math.min(MAX_Y,
                    centre + (int) Math.round(Math.sin(ixf * xFreq) * rd2)));
                yVals[i] = Math.max(MIN_Y, Math.min(MAX_Y,
                    centre + (int) Math.round(Math.sin(ixf * yFreq) * rd2)));
                }

            for(int y = dynBgImgEdgePixels; --y >= 0; )
                {
                for(int x = dynBgImgEdgePixels; --x >= 0; )
                    {
                    // Compute least-significant (blue) value as noise.
                    final int blue = pRnd.nextInt(range) + MIN_Y;
                    // Convert to opaque rgb.
                    int rgb = (xVals[x] << 16) | (yVals[y] << 8) | (blue) | 0xff000000;
                    bi.setRGB(x, y, rgb);
                    }
                }

            return(bi);
            }

        /**Chose a number of cycles for the background pattern across the image; strictly positive.
         * The result is always (much) less than the number of pixels,
         * and always at least 1.
         *
         * @param pRnd  the pseudo-random (repeatable) source; never null
         */
        private static int getRndCycles(final Random pRnd)
            {
            return(1 + pRnd.nextInt(1 | pRnd.nextInt(dynBgImgEdgePixels / 16)));
            }
        }











    /**Immutable List of instances all available basic DynImageGen; never null or empty. */
    public static final List<DynImageGen> allDynGens =
        Collections.unmodifiableList(new ArrayList<DynImageGen>(Arrays.asList(
            new DynImageGen[]{
                new SimpleUniformMonoImageGen(),
                new SimpleUniformNoiseMonoImageGen(),
                new Sine2DColourImageGen(),
            })));


    /**Convert BufferedImage to byte[]-encoded JPEG no larger than dynBgImgMaxBytes bytes.
     */
    static byte[] generateEncodedImage(final BufferedImage bi)
        throws IOException
        {
        final AbstractImageHandler.ThumbnailParams tp = ETP.handler.getThumbnailParams();
        if(tp == null) { throw new IllegalStateException("cannot get thumbnail parameters"); }
        return(ETP.handler.makeSizeConstrainedEncodedImage(
            tp.minQuality, tp.initialQualityHint, tp.maxQuality,
            bi,
            dynBgImgMaxBytes/4, dynBgImgMaxBytes/2, // Target limits.
            1, dynBgImgMaxBytes, // Absolute limits...
            dynBgImgMaxBytes/2)); // Target size...
        }
    }
