/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.catalogue;

import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 07-Jun-2003
 * Time: 08:58:31
 */

/**Base bean to help with the pagination for otherwise large pages.
 * This can parse an input argument indicating the desired page,
 * and work out which page is meant and return it as an int for the page logic to use.
 * <p>
 * This can also generate the labels/values for one or more HTML form submit buttons
 * to help generate the HTML page.
 * <p>
 * This implementation generates sequential page labels for buttons in its simplest form,
 * but if there are too many buttons to show this then generates a sparser set
 * that ensures that a user can navigate in few clicks to any page and especially so
 * to nearby pages as they home in on a target.
 * <p>
 * This class is designed to be extended by classes that can put arbitrary labels on
 * the buttons, for example words to help the user go directly to the desired page.
 * <p>
 * If given an unparseable page number as input, this and deriving implementations
 * should always return a valid page number,
 * returning the "safe" page number of 1 if they cannot deduce the correct page or a near one.
 * <p>
 * This bases its pagination on the properties in WebConsts dictating the maximum number
 * of items to show on a page and the maximum number of page-choice buttons to show to a user.
 * <p>
 * This needs to be told the number of entries to be paginated;
 * deriving classes may accept other or extra information in lieu.
 * <p>
 * This does its basic pagination based on the number of items to be paginated
 * and the WebConsts that dictate maximum items per page and maximum pages
 * or choices of pages to go to at once.
 * In principle these might be user-settable rather than constants.
 * <p>
 * This is designed to be used as a bean at page or request scope in a JSP,
 * though may be usable outside that environment.
 * It may not be usefully serialisable.
 * <p>
 * <strong>THIS CLASS IS NOT THREAD SAFE</strong> since
 * its usual enviroment and indeed lifetime
 * is within the thread of a single servlet/JSP page service
 * and thread-safety is an unnecessary overhead.
 * Its methods are not synchronized,
 * and if it is necessary to share one instance between threads
 * they must hold a lock on the instance.
 * <p>
 * This Serializable so as to be able to be stored in
 * a servlet session; nothing especially long-lived or sensitive.
 * <p>
 * TODO: Should do validation on deserialisation.
 */
public abstract class PaginationBeanBase implements java.io.Serializable
    {
    /**Number of items to be paginated; non-negative.
     * Defaults to zero meaning no items to paginate;
     * any attempt to set it to a negative number sets it to zero.
     */
    private int numberOfItems;

    /**The page number of the current page; strictly positive.
     * Defaults to 1; any attempt to set it to an invalid value will set it to 1, which is always safe.
     * <p>
     * The page number must be positive and less than numberOfPages() unless numberOfPages is 0,
     * in which case the page number must be exactly 1.
     * <p>
     * This property matches the property named by WebConsts.PAGE_NUMBER_PARAMETER
     * to allow for easy setting of the property from a JSP.
     */
    private int pg = 1;

    /**Get number of items to be paginated; non-negative.
     * Defaults to zero meaning no items to paginate.
     */
    final public int getNumberOfItems()
        {
        return(numberOfItems);
        }

    /**Set number of items to be paginated; non-negative.
     * Defaults to zero meaning no items to paginate;
     * any attempt to set it to a negative number sets it to zero.
     */
    public void setNumberOfItems(int numberOfItems)
        {
        if(numberOfItems < 0) { numberOfItems = 0; } // Defeast attempts to make negative.
        this.numberOfItems = numberOfItems;

        // Force pg to be constrained by the number of items and thus pages.
        if(pg > getNumberOfPages()) { pg = Math.max(1, getNumberOfPages()); }
        }

    /**Compute the number of pages based on the numberOfItems to be paginated.
     * Is zero if there are no items, else strictly positive.
     * <p>
     * Uses the WebConsts.MAX_RESULTS_PER_PAGE to determine how many pages will be needed.
     */
    final public int getNumberOfPages()
        {
        return((numberOfItems + WebConsts.MAX_RESULTS_PER_PAGE - 1) / WebConsts.MAX_RESULTS_PER_PAGE);
        }

    /**Set the page number of the current page; strictly positive.
     * Defaults to 1; any attempt to set it to an invalid value will set it to 1, which is always safe.
     * <p>
     * The page number must be positive and less than numberOfPages() unless numberOfPages is 0,
     * in which case the page number must be exactly 1.
     * <p>
     * This property name matches the property named by WebConsts.PAGE_NUMBER_PARAMETER
     * to allow for easy setting of the property from a JSP.
     */
    public void setPg(int pg)
        {
        if((pg < 1) || (pg > getNumberOfPages())) { pg = 1; } // Avoid setting an invalid value.
        this.pg = pg;
        }

    /**Set the page number as a String; if invalid (eg badly-formatted input or null) set page number to 1.
     * It must always be possible to parse the label for any page whatever set
     * getPageLabels() returns.
     * <p>
     * A deriving class must override this for specialised parses.
     */
    public void setPg(String s)
        {
        setPg(getPageNumber(s));
        }

    /**Find out what page number a given label indicates.
     * This should be the reverse of getPageLabel().
     * <p>
     * On badly-formed or otherwise-unparsable input this should return 1.
     */
    public abstract int getPageNumber(String s);

    /**Get the page number of the current page; strictly positive.
     * Defaults to 1.
     */
    final public int getPg()
        {
        return(pg);
        }

    /**Get (index of) lowest-numbered item to display; non-negative.
     * If numberOfItems() is zero this returns zero,
     * else this returns less than numberOfItems().
     * <p>
     * Computed from the page number and the WebConsts.MAX_RESULTS_PER_PAGE value.
     */
    final public int getStartIndex()
        {
        return((pg - 1) * WebConsts.MAX_RESULTS_PER_PAGE);
        }

    /**Get (index of) one beyond highest-numbered item to display; non-negative.
     * Exactly numberOfItems() for the last page,
     * else exactly one page more than getStartIndex().
     * <p>
     * Computed from the page number and the WebConsts.MAX_RESULTS_PER_PAGE value.
     */
    final public int getEndIndex()
        {
        return(Math.min(numberOfItems, pg * WebConsts.MAX_RESULTS_PER_PAGE));
        }

    /**Return the page label for the given page (1 upwards), never null; the page must be in range else the result is undefined.
     * This will commonly be overridden in deriving classes.
     * <p>
     * The label generated must be deterministic and depend only on the attributes set for the bean,
     * and the same as getPageLabels() would have generated for this given page.
     * The value for a label must not depend on the current page value,
     * and it must always be possible to parse the label for any page whatever set
     * getPageLabels() returns.
     * <p>
     * In many cases getPageLabels() can simply be implemented in terms of getPageLabel()
     * though sometimes efficiency considerations (for example) may preclude that.
     */
    public abstract String getPageLabel(int pageNum);

    /**Get an ordered array of page numbers that correspond index-by-index to the page labels returned by getPageLabels().
     * If there are no more than MAX_RESULTS_PAGES in total than a contigious set of pages
     * from 1 up to the getNumberOfPages() is returned,
     * else some subset of pages is returned usually including the first and last items and
     * the current page and its neighbours, in ascending order.
     * <p>
     * The array returned is private to the caller.
     * <p>
     * This can be overridden to vary the page distribution details though the general contract
     * must be maintained.
     */
    public int[] getPageNumbers()
        {
        final int pageNum = pg;
        final int totalPages = this.getNumberOfPages();

        // If we can show all the page labels, do so as a contigious list.
        if(totalPages <= WebConsts.MAX_RESULTS_PAGES)
            {
            final int ra[] = new int[totalPages];
            for(int i = ra.length; --i >= 0; )
                { ra[i] = i+1; }
            return(ra);
            }
        // We have too many buttons to show, do a sub-set.
        // Create a SortedSet of them to iterate over.
        final SortedSet<Integer> pagesToShow = new TreeSet<Integer>();

        // Always include:
        //   * first and last,
        //   * the current page,
        //   * the prev/next and onwards at exponentially greater steps.
        // Then divide up the rest evenly.
        pagesToShow.add(Integer.valueOf(1));
        pagesToShow.add(Integer.valueOf(totalPages));
        pagesToShow.add(Integer.valueOf(pageNum));

        // Insert pages with exponential spacing from current page...
        for(int interval = 1; interval < totalPages; interval *= 2)
            {
            int remainingPages = WebConsts.MAX_RESULTS_PAGES - pagesToShow.size();
            if(remainingPages <= 0) { break; }
            final int pDown = pageNum - interval;
            if(pDown > 0) { pagesToShow.add(Integer.valueOf(pDown)); }

            remainingPages = WebConsts.MAX_RESULTS_PAGES - pagesToShow.size();
            if(remainingPages <= 0) { break; }
            final int pUp = pageNum + interval;
            if(pUp <= totalPages) { pagesToShow.add(Integer.valueOf(pUp)); }
            }

        // Insert remaining pages linearly between first and last pages,
        // assuming them to have already been inserted.
        int remainingPages = WebConsts.MAX_RESULTS_PAGES - pagesToShow.size();
        if(remainingPages > 0)
            {
            final float interval = (totalPages-1) / (float) (remainingPages+1);
            while(--remainingPages >= 0)
                {
                final int index = 1 +
                    Math.round((remainingPages + 1) * interval);

                // Weed out any invalid page numbers we produce.
                if((index < 1) || (index >= totalPages)) { continue; }

                pagesToShow.add(Integer.valueOf(index));
                }

            // Mop up any remaining unused buttons/pages with items working out
            // linearly from current page.
            // We assume that the current page has always been included.
            for(int gap = 1; gap < totalPages; ++gap)
                {
                // Do "up" first to slightly redress balance with
                // exponential stepping above.
                if((remainingPages = WebConsts.MAX_RESULTS_PAGES - pagesToShow.size()) <= 0) { break; }
                final int pUp = pageNum + gap;
                if(pUp <= totalPages) { pagesToShow.add(Integer.valueOf(pUp)); }

                if((remainingPages = WebConsts.MAX_RESULTS_PAGES - pagesToShow.size()) <= 0) { break; }
                final int pDown = pageNum - gap;
                if(pDown > 0) { pagesToShow.add(Integer.valueOf(pDown)); }
                }
            }


        // Convert our SortedSet to an ordered array of int page numbers.
        final int ra[] = new int[pagesToShow.size()];
        final Iterator<Integer> it = pagesToShow.iterator();
        for(int i = 0; i < ra.length; ++i)
            { ra[i] = (it.next()).intValue(); }
        return(ra);
        }

    /**Gets ordered (unmodifiable) List of String labels/tokens for page buttons.
     * Computes a set of pages and returns String values to use as the values (and legends)
     * for submit buttons for this class to parse with the setPg(String) routine.
     * <p>
     * These labels have to be accepted by setPg(String).
     * <p>
     * The labels are unique and identify pages in monotonically-increasing order
     * as you iterate through the List.
     * <p>
     * The labels generated must be deterministic and depend only on the attributes set for the bean,
     * <strong>possibly including the current page</strong>.
     * <p>
     * Usually the first and last items in the List a labels for the first and last pages in the range.
     * <p>
     * Usually labels for the current page and those immediately before and after it are in the
     * result list.
     * <p>
     * The base implementation calls getPageLabel() to generate the individual labels,
     * so it may often be sufficient to override getPageLabel() only in a derived class.
     * <p>
     * This will commonly be overridden in deriving classes but the general contract
     * must be maintained.
     */
    public List<String> getPageLabels()
        {
        final int ra[] = getPageNumbers();
        final String sarr[] = new String[ra.length];
        for(int i = 0; i < ra.length; ++i)
            { sarr[i] = getPageLabel(ra[i]); }
        return(Collections.unmodifiableList(Arrays.asList(sarr)));
        }


    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -3756427234954341532L;
    }
