/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.catalogue;

import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 03-Jun-2003
 * Time: 17:40:27
 */

/**Bean to help with the pagination for otherwise large pages; this generates numeric labels.
 * This can parse an input argument indicating the desired page,
 * or find it directly in an incoming HttpRequest,
 * and work out which page is meant and return it as an int for the page logic to use.
 * <p>
 * This can also generate the labels/values for one or more HTML form submit buttons
 * to help generate the HTML page.
 * <p>
 * This implementation generates sequential page-numbering for buttons in its simplest form,
 * but if there are too many buttons to show this then generates a sparser set
 * that ensures that a user can navigate in few clicks to any page and especially so
 * to nearby pages as they home in on a target.
 * <p>
 * This class is designed to be extended by classes that can put arbitraty labels on
 * the buttons, for example words to help the user go directly to the desired page.
 * <p>
 * If given an unparseable page number as input, this and deriving implementations
 * should always return a valid page number,
 * returning the "safe" page number of 1 if they cannot deduce the correct page or a near one.
 * <p>
 * This bases its pagination on the properties in WebConsts dictating the maximum number
 * of items to show on a page and the maximum number of page-choice buttons to show to a user.
 * <p>
 * This needs to be told the number of entries to be paginated;
 * deriving classes may accept other or extra information in lieu.
 * <p>
 * This does its basic pagination based on the number of items to be paginated
 * and the WebConsts that dictate maximum items per page and maximum pages
 * or choices of pages to go to at once.
 * In principle these might be user-settable rather than constants.
 * <p>
 * This is designed to be used as a bean at page or request scope in a JSP,
 * though may be usable outside that environment.
 * It may not be usefully serialisable.
 * <p>
 * <strong>THIS CLASS IS NOT THREAD SAFE</strong> since
 * its usual enviroment and indeed lifetime
 * is within the thread of a single servlet/JSP page service
 * and thread-safety is an unnecessary overhead.
 * Its methods are not synchronized,
 * and if it is necessary to share one instance between threads
 * they must hold a lock on the instance.
 */
public final class PaginationBeanNumeric extends PaginationBeanBase
    {
    /**Get the page number as a from the String label; in case of bad input return 1.
     * This parses using Integer.parseInt("string", 10).
     * <p>
     * A deriving class may override this for more sophisticated or specialised parses.
     */
    public int getPageNumber(final String s)
        {
        try { return(Integer.parseInt(s, 10)); }
        catch(final NumberFormatException e) { return(1); }
        }

    /**Gets ordered (unmodifiable) List of String labels/tokens for page buttons.
     * This implementation makes the labels numeric
     * and optimises the case of returning contigious labels starting at "1"
     * to minimise object churn.
     * <p>
     * Computes a set of pages and returns String values to use as the values (and legends)
     * for submit buttons for this class to parse with the setPg(String) routine.
     * <p>
     * These labels have to be accepted by setPg(String).
     * <p>
     * The labels are unique and identify pages in monotonically-increasing order
     * as you iterate through the List.
     * <p>
     * The labels generated must be deterministic and depend only on the attributes set for the bean.
     * <p>
     * Usually the first and last items in the List a labels for the first and last pages in the range.
     * <p>
     * Usually labels for the current page and those immediately before and after it are in the
     * result list.
     * <p>
     * The base implementation returns numeric labels,
     * but this calls getPageLabel() to generate the individual labels,
     * so it may often be sufficient to override getPageLabel() only in a derived class.
     * <p>
     * This will commonly be overridden in deriving classes but the general contract
     * must be maintained.
     */
    public List<String> getPageLabels()
        {
        final int _totalPages_ = this.getNumberOfPages();

        // As an optimisation...
        // If there few enough pages to show all of them as a contigious sequence from "1",
        // do so in a way which requires minimal object creation/churn,
        // since it a very common case.
        if(_totalPages_ <= WebConsts.MAX_RESULTS_PAGES)
            { return(generateNumericPageLabels1Upwards(_totalPages_)); }

        return(super.getPageLabels());
        }

    /**Return the page label for the given page (1 upwards), never null; the page must be in range else the result is undefined.
     * This will commonly be overridden in deriving classes.
     * <p>
     * The label generated must be deterministic and depend only on the attributes set for the bean,
     * and the same as getPageLabels() would have generated for this given page.
     */
    public String getPageLabel(int pageNum)
        {
        if((pageNum < 1) || (pageNum > getNumberOfPages())) { return("1"); } // Safe response.
        return(String.valueOf(pageNum));
        }

    /**Return an immutable List of contigious String (int-valued) labels for page buttons from 1 upwards.
     * In the case where the top value is no more than WebConsts.MAX_RESULTS_PAGES
     * we return a sub-list of a static singleton immutable List covering this range
     * without synchronsiation or (significant) object creation for efficiency.
     *
     * @param nPages  requests list of String labels for buttons for the decimal integers
     *     from "1" upwards so a value of 3 would return the list "1", "2", "3" for example;
     *     the argument must be non-negative
     */
    private static List<String> generateNumericPageLabels1Upwards(int nPages)
        {
        if(nPages <= WebConsts.MAX_RESULTS_PAGES)
            {
            // Return a completely unwrapped reference where we want the whole lot
            // as this is faster and saves a object allocation.
            // The case is quite common, when a search returns lots of results
            // and is limited to the maximum number of pages.
            if(nPages == pageLabels1Upwards.size())
                { return(pageLabels1Upwards); }

            // Return sub-list view of early portion.
            return(pageLabels1Upwards.subList(0, nPages));
            }

        // OK, list is too long, so manufacture it from scratch.
        return(generateNewNumericPageLabels1Upwards(nPages));
        }

    /**Generate a new immutable List of pageLabels from "1" upwards each time; argument must be non-negative.
     * We String.intern() the first few labels for maximum memory efficiency!
     */
    private static final List<String> generateNewNumericPageLabels1Upwards(int nPages)
        {
        final String s[] = new String[nPages];
        for(int i = nPages; --i >= 0; )
            {
            String t = String.valueOf(i + 1);
            if(i < WebConsts.MAX_RESULTS_PAGES) { t = t.intern(); } // Intern commonly-used values.
            s[i] = t;
            }
        return(Collections.unmodifiableList(Arrays.asList(s)));
        }

    /**Immutable List of String labels containing decimal integers from 1 to WebConsts.MAX_RESULTS_PAGES.
     * Note that the label at index i is the value of i+1, eg at index 0 is "1".
     * <p>
     * Used to cover the most common cases without needing to generate lots of new objects each time.
     */
    private static final List<String> pageLabels1Upwards =
        generateNewNumericPageLabels1Upwards(WebConsts.MAX_RESULTS_PAGES);

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -5376824526954786928L;
    }
