/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.catalogue;

import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.util.AbstractList;
import java.util.Collections;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 06-Jun-2003
 * Time: 19:13:26
 */

/**Provides pagination to be used in conjunction with (tree-structured) browsable exhibit collections.
 * A specialisation of PaginationBeanBase that provides pagination for nodes
 * created with the TreeFilterBean that creates alphanumeric labels
 * based on distinguishing prefixes of the leaves of the node,
 * indended to be as short as possible but give the user a large hint as to the
 * full set of available choices even when only a tiny fraction of the choices
 * (page buttons) can be shown.
 *
 * @see org.hd.d.pg2k.webSvr.exhibit.TreeFilterBean
 * @see org.hd.d.pg2k.webSvr.catalogue.PaginationBeanNumeric
 */
public final class PaginationBeanTree extends PaginationBeanBase
    {
    /**If true, convert all page labels to lower-case rather than allowing mixed case. */
    public static final boolean LABEL_ALL_LC = true;

    /**Minimum letters in label, but only up to first word boundary. */
    public static final int LABEL_MIN_LEN = 3;

    /**Given a List of prefixes and short exhibit names as produced by TreeFilterBean sets enables computation of the page labels.
     * This replaces direct use of the setNumberOfItems() call.
     * <p>
     * This expects an (immutable) List of word-prefix lists and short exhibit names (as String entries),
     * sorted in case-insensitive order, as produced by TreeFilterBean.
     * Results are undefined if this is not the case.
     * <p>
     * All entries in the list must start with (case-insensitively) the supplied prefix for the node.
     * <p>
     * This caches the List that it is passed.
     *
     * @see org.hd.d.pg2k.webSvr.exhibit.TreeFilterBean
     */
    public void setTreeNodeDetails(final String prefix,
                                   final List<String> prefixesAndShortExhibitName)
        {
        if((prefix == null) || (prefixesAndShortExhibitName == null))
            { throw new IllegalArgumentException("arguments cannot be null"); }

        // Trivial quick test that prefix is right at least for first entry.
        if((prefixesAndShortExhibitName.size() > 0) &&
            !(prefixesAndShortExhibitName.get(0)).toLowerCase().startsWith(prefix))
        { throw new IllegalArgumentException("all list entries must start (ignoring case) with prefix"); }

        // First, we can set the number of items.
        super.setNumberOfItems(prefixesAndShortExhibitName.size());

        // Store prefix list...
        nodePrefix = prefix;
        prefixList = prefixesAndShortExhibitName;
        }

    /**Store of prefix for the current node.
     * Default value is safe ("").
     */
    private String nodePrefix = "";

    /**Store of prefix list.
     * Default value is safe (the empty list).
     */
    private List<String> prefixList = Collections.emptyList();


    /**Don't call this directly; use setTreeNodeDetails() instead. */
    public final void setNumberOfItems() { throw new Error("Do not call this directly"); }

    /**Safe page value that will be interpretted as page 1.
     * Should only be called on in case of difficulty.
     * <p>
     * Is not a valid prefix to avoid ambiguity.
     */
    public static final String SAFE_LABEL = "!";

    /**Return the page label for the given page (1 upwards), never null; the page must be in range else the result is undefined.
     * This will usually be the shortest prefix that will distinguish
     * the first entry in the desired page from the first entry in the pages
     * before and after.
     * <p>
     * (This may either be the shortest word prefix
     * or the shortest character prefix.)
     */
    public String getPageLabel(final int pageNum)
        {
        if((pageNum < 1) || (pageNum > getNumberOfPages()))
            { return(SAFE_LABEL); }

        // We can rip the common prefix (to get to this node)
        // off the front of all labels.
        final int npLen = nodePrefix.length();

        final int itemIndex = (pageNum - 1) * WebConsts.MAX_RESULTS_PER_PAGE;
        assert(itemIndex < prefixList.size());
        final String rawItemEntry = prefixList.get(itemIndex);
        final String itemEntry = (!LABEL_ALL_LC) ? rawItemEntry : rawItemEntry.toLowerCase();

        // If the prev/next items are out of range,
        // then use null to indicate that the item is no longer in the running.
        final int prevItemIndex = itemIndex - WebConsts.MAX_RESULTS_PER_PAGE;
        final int nextItemIndex = itemIndex + WebConsts.MAX_RESULTS_PER_PAGE;
        String prevItemEntry = (prevItemIndex >= 0) ?
                prefixList.get(prevItemIndex) : null;
        String nextItemEntry = (nextItemIndex < prefixList.size()) ?
                prefixList.get(nextItemIndex) : null;

        // Build this new label a character at a time from this page's first entry
        // (with a string length of at least 1, or LABEL_MIN_LEN if possible)
        // until it is different to both the adjacent labels (ignoring case).
        final int startLen = Math.max(1, // Always at least one character...
                                      Math.min(LABEL_MIN_LEN, // up to minimum prefix length...
                                               itemEntry.indexOf(ExhibitName.WORD_SEP, npLen+1)-npLen)); // or word end.
        for(int i = npLen + startLen; i <= itemEntry.length(); ++i)
            {
            final String putativeLabel = itemEntry.substring(npLen, i);

            // Check that putativeLabel does not look like previous label.
            if(prevItemEntry != null)
                {
                if(i > prevItemEntry.length())
                    { prevItemEntry = null; } // Need not worry about it.
                else if(!putativeLabel.equalsIgnoreCase(prevItemEntry.substring(npLen, i)))
                    { prevItemEntry = null; } // No longer a contender.
                else
                    { continue; } // Whoops, does look like previous label, so skip.
                }

            // Check that putativeLabel does not look like next label.
            if(nextItemEntry != null)
                {
                if(i > nextItemEntry.length())
                    { nextItemEntry = null; } // Need not worry about it.
                else if(!putativeLabel.equalsIgnoreCase(nextItemEntry.substring(npLen, i)))
                    { nextItemEntry = null; } // No longer a contender.
                else
                    { continue; } // Whoops, does look like next label, so skip.
                }

            return(putativeLabel); // Done it!
            }

        // Shouldn't really get here: something is wrong.
        return(SAFE_LABEL);
        }

    /**Get the page number as a from the String label; in case of bad input return 1.
     * We reconstruct the label set we generated and look up in that.
     */
    public int getPageNumber(final String s)
        {
        final int totalPages = getNumberOfPages();
        if((s == null) || (totalPages < 2))
            { return(1); }

        // Make a virtual unmodifiable list with one entry for every page,
        // that we will binary search in.
        // We compute each value on demand, so we only compute O(log(n)) values.
        final List<String> l = (new AbstractList<String>(){
            /**Returns the element at the specified position in this list.
             * Computes the page label for page index + 1 on the fly.
             */
            public final String get(final int index)
                { return(getPageLabel(index + 1)); }

            /**Returns the number of elements in this collection (the number of pages). */
            public final int size() { return(totalPages); }
            });

        final int insertPos = Collections.binarySearch(l, s, String.CASE_INSENSITIVE_ORDER);
        if(insertPos >= 0) { return(insertPos + 1); }

        // Unrecognised label, so go to first page.
        return(1);
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 8393863768211031724L;
    }
