/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.catalogue;

import ORG.hd.d.jIndexer.server.JIndexBean;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.uploader.ExhibitHandlerBeanBase;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.*;

/**This is the JavaBean that backs the (HTML) catalogue search page.
 * This should be held at page scope so that its state is
 * set from fresh on each page load.
 * <p>
 * This attempts to sanitise values as they are set.
 * <p>
 * Most methods are synchronized to provide thread-safety.
 */
public final class SearchPageJavaBean extends ExhibitHandlerBeanBase
    {
    /**The sanitised simple text query, or "" if none. */
    private String q = "";

    /**Sets the simple query, or "" if none.
     * This truncates and canonicalises and sanitises
     * the query text as necessary.
     */
    public synchronized void setQ(final String rawQueryText)
        { q = JIndexBean.canonicaliseSimpleByWordQuery(rawQueryText, WebConsts.MAX_SIMPLE_QUERY_LEN); }

    /**Get the sanitised query text, or "" if none (never null, never longer than WebConsts.MAX_SIMPLE_QUERY_LEN). */
    public synchronized String getQ() { return(q); }


    /**How long in milliseconds we consider a day to be. */
    public static final int DAY_MS = 24 * 3600 * 1000;

    /**The number of recent days to filter for; 0 means no filter. */
    private int recentDaysFilter = 0;

    /**Map from symbolic filter period to number of days (String to Integer); immutable.
     * Keys in here can also be suffixed to sommon i18n resource "common.cat.filtersince.recent."
     * to get descriptive text.
     */
    public static final Map<String,Integer> symFilterDays;
    /**Sorted map from number of days back to symbolic name if any (Integer to String). */
    public static final SortedMap<Integer,String> daysFilterSym;
    /**Initialise symFilterDays and daysFilterSym. */
    static
        {
        final Map<String,Integer> m = new HashMap<String, Integer>(11);
        m.put("d", Integer.valueOf(1));
        m.put("w", Integer.valueOf(7));
        m.put("m", Integer.valueOf(30));
        m.put("q", Integer.valueOf(90));
        m.put("y", Integer.valueOf(365));
        symFilterDays = Collections.unmodifiableMap(m);

        // Reverse mapping of m.
        final SortedMap<Integer,String> r = new TreeMap<Integer, String>();
        for(final Iterator<String> it = m.keySet().iterator(); it.hasNext(); )
            {
            final String s = it.next();
            r.put(m.get(s), s);
            }
        daysFilterSym = Collections.unmodifiableSortedMap(r);
        }

    /**Set the recent period (to today) for which exhibits are wanted.
     * The parameter can either be one of the single letter keys in symFilterDays,
     * or null or "" or "-" (SETTER_ALL) or "0" for no filter,
     * or a positive Integer number of days.
     */
    public synchronized void setRecentDaysFilter(String sRecentDaysFilter)
        {
        // Look for all cases meaning "no filter".
        if((sRecentDaysFilter == null) ||
           (sRecentDaysFilter.length() == 0) ||
           sRecentDaysFilter.equals(ExhibitHandlerBeanBase.SETTER_ALL) ||
           sRecentDaysFilter.equals("0"))
            {
            recentDaysFilter = 0;
            return;
            }

        // If this is a symbolic value, convert it to the actual number of days.
        final Integer ir = symFilterDays.get(sRecentDaysFilter);
        if(ir != null)
            {
            recentDaysFilter = ir.intValue();
            return;
            }

        // Try to interpret as integer number of days...
        try {
            final int i = Integer.parseInt(sRecentDaysFilter, 10);
            // Convert negative/zero values to "no filter".
            if(i <= 0) { recentDaysFilter = 0; }
            // Else store the number of days specified.
            else { recentDaysFilter = i; }
            }
        catch(NumberFormatException e)
            {
            // Treat malformed input as "no filter".
            recentDaysFilter = 0;
            }
        }

    /**Get the "recent days" filter period; 0 for none else (positive) number of days. */
    public synchronized int getRecentDaysFilter()
        {
        return(recentDaysFilter);
        }

    /**Generates body of recent-days filter select statement (dependent on old value, if any).
     * Assumed not to require localisation or internationalisation.
     * <p>
     * This takes the symbolic values in increasing number of days.
     * If the locale bean is passed in this makes full labels else it makes none.
     */
    public synchronized String makeRecentDaysFilterBody(LocaleBeanBase l)
        {
        final String legit[] = new String[daysFilterSym.size()];
        daysFilterSym.values().toArray(legit);
        final String fullList[] = new String[legit.length + 1];
        fullList[0] = SETTER_ALL; // Meaning ``all''.
        System.arraycopy(legit, 0, fullList, 1, legit.length);

        // Now make the labels...
        final String labels[] = new String[fullList.length];
        labels[0] = "*"; // Meaning all.

        // Make a meaningful set of labels.
        for(int i = labels.length; --i > 0; )
            {
            labels[i] = fullList[i]; // Default is to use symbolic name unaltered.

            if(l != null)
                {
                final String name = UploaderConsts.SYMTIME_I18N_PREFIX + fullList[i];
                final String m = l.getLocalisedMessage(name);
                if(!name.equals(m))
                    { labels[i] = m; }
                }
            }

        return(UploaderUtils.makeSelectBody(fullList, labels,
                                       daysFilterSym.get(Integer.valueOf(getRecentDaysFilter()))));
        }



    /**Gets the search filter, or null if none.
     * Note that we expect short exhibit names to be presented
     * to these filters, as that is what is stored in the index.
     */
    public synchronized JIndexBean.SearchFilterByName getSearchFilter()
        {
        JIndexBean.SearchFilterByName result = null;

        // If an attribute-word list is set, filter with it.
        // Relatively slow, so should be towards the tail of the filter chain,
        // ie towards the start of this function.
        // This will be more efficient if the attribute list is deduped first.
        final List attributes = getAttributeWordsAsList();
        if(!attributes.isEmpty())
            {
            final JIndexBean.SearchFilterByName upstream = result;
            result = (new JIndexBean.SearchFilterByName(){
                final public boolean accept(final String name)
                    {
                    final String fullName = getAep().aeid.getFullName(name);
                    final Map<String,Set<String>> ebam = getAep().getExhibitsByAttribute();
                    for(final Iterator it = attributes.iterator(); it.hasNext(); )
                        {
                        final Set<String> s = ebam.get(it.next());
                        if(s == null) { continue; } // Ignore invalid attribute word!
                        if(!s.contains(fullName)) { return(false); } // Does not have this attribute.
                        }

                    if(upstream != null) { return(upstream.accept(name)); } // Chain previous filter.
                    return(true); // OK.
                    }
                });
            }

        // If the recent-addition value is set, filter with it.
        // Relatively slow, so should be towards the tail of the filter chain,
        // ie towards the start of this function.
        if(getRecentDaysFilter() > 0)
            {
            final long now = System.currentTimeMillis();
            final long since = now - (getRecentDaysFilter() * (long)DAY_MS);
            assert(since < now);
            final AllExhibitImmutableData aeid = getAep().aeid;
            final JIndexBean.SearchFilterByName upstream = result;
            result = (new JIndexBean.SearchFilterByName(){
                final public boolean accept(final String name)
                    {
                    // Note that we have to extend name to full form to check its age...
                    final ExhibitStaticAttr esa = aeid.getStaticAttr(aeid.getFullName(name));
                    if(esa.timestamp < since) { return(false); } // Definitely not OK.
                    if(upstream != null) { return(upstream.accept(name)); } // Chain previous filter.
                    return(true); // OK.
                    }
                });
            }

        // If the category is set, filter with it.
        // Relatively slow, so should be towards the tail of the filter chain,
        // ie towards the start of this function.
        if(!"".equals(getCategory()))
            {
            final String cat = getCategory();
            final AllExhibitImmutableData aeid = getAep().aeid;
            final JIndexBean.SearchFilterByName upstream = result;
            result = (new JIndexBean.SearchFilterByName(){
                final public boolean accept(final String name)
                    {
                    // Note that we have to extend name to full form to check its category...
                    if(!cat.equals(ExhibitName.getCategoryComponent(aeid.getFullName(name)))) { return(false); } // Definitely not OK.
                    if(upstream != null) { return(upstream.accept(name)); } // Chain previous filter.
                    return(true); // OK.
                    }
                });
            }

        // If the author is set, filter with it.
        // Relatively fast, so should be towards the head of the filter chain,
        // ie towards the end of this function.
        if(!"".equals(getAuthor()))
            {
            final String auth = getAuthor();
            final JIndexBean.SearchFilterByName upstream = result;
            result = (new JIndexBean.SearchFilterByName(){
                final public boolean accept(final String name)
                    {
                    if(!auth.equals(ExhibitName.getAuthorComponent(name))) { return(false); } // Definitely not OK.
                    if(upstream != null) { return(upstream.accept(name)); } // Chain previous filter.
                    return(true); // OK.
                    }
                });
            }

        // If the suffix is set, filter with it.
        // Always fast, so should be towards the head of the filter chain,
        // ie towards the end of this function.
        if(!"".equals(getSuffix()))
            {
            final String suf = getSuffix();
            final JIndexBean.SearchFilterByName upstream = result;
            result = (new JIndexBean.SearchFilterByName(){
                final public boolean accept(final String name)
                    {
                    if(!name.endsWith(suf)) { return(false); } // Definitely not OK.
                    if(upstream != null) { return(upstream.accept(name)); } // Chain previous filter.
                    return(true); // OK.
                    }
                });
            }

        return(result);
        }

    /**Fixed immutable "no results" value. */
    private static final String[] NO_RESULTS = new String[0];

    /**Get search results as list of (short) exhibit names; may be zero-length but never null.
     * Using the query string and other details that the user has supplied
     * this does the search and returns
     * a (possibly-zero-length) list of exhibit names.
     * <p>
     * As a side-effect this sets the time the search takes in milliseconds
     * accessible by getLastSearchTimeMs().
     *
     * @param maxResults  maximum length list of results to return;
     *     should be positive
     */
    public synchronized String[] doLookup(final DataSourceBean dataSource,
                                          final int maxResults)
        throws IOException
        {
        final long startTime = System.currentTimeMillis();
        try {
            if(maxResults < 1) { return(NO_RESULTS); }

            final String q = getQ();
            final JIndexBean.SearchFilterByName sfbn = getSearchFilter();

            // Special cases if the query string is empty...
            if(q.length() == 0)
                {
                // If there is no search string and no filter,
                // return an empty list immediately.
                if(sfbn == null) { return(NO_RESULTS); }

                // If there is no search string but there is a filter,
                // apply that filter to the whole exhibit list...
                // This returns full names, which we convert to short names
                // before returning...
                final String result[] = getAep().select((new AllExhibitProperties.AEPFilter(){
                    public final boolean accept(AllExhibitProperties _aep, String fullName)
                        { return(sfbn.accept(ExhibitName.getFileComponent(fullName))); }
                    }), null, maxResults);
                for(int i = result.length; --i >= 0; )
                    { result[i] = ExhibitName.getFileComponent(result[i]); }
                return(result);
                }

            // Try doing a strict match on "all" terms first, which may be quick and gives high relevance.
            final String result[] =
                dataSource.findExhibitsByWord(q, DataSourceBean.FEBY_MATCH_TYPE_ALL, maxResults, sfbn);
            if(result.length > 0) { return(result); }

            // Default lookup by most words, as fallback to try to find some results.
            return(dataSource.findExhibitsByWord(q, DataSourceBean.FEBY_MATCH_TYPE_MOST, maxResults, sfbn));
            }
        finally
            {
            // Record the time taken to do the search.
            lastSearchTimeMs = System.currentTimeMillis() - startTime;
            }
        }
    
    /**Maximum (UTF-8 URL-encoded) GET-style query string that we will generate; strictly positive.
     * We base this on the maximum simple query text length that we support.
     */
    public static final int MAX_GET_QUERY_CHARS = 128 + WebConsts.MAX_SIMPLE_QUERY_LEN;
    
    /**Make simple (length-limited) GET query string.
     * Can be used where POST is not possible or desirable.
     * <p>
     * Takes an optional (positive) page number argument.
     * <p>
     * This is UTF-8 encoded.
     * <p>
     * Parameter are separated with <code>&amp;amp;</code>
     * ready for direct inclusion at the end of a URL.
     * <p>
     * If the result is getting too long we stop adding fields,
     * so we encode the more critical ones first.
     * 
     * @param pg  if positive then the pg (page) parameter is included
     * 
     * @return  non-empty result starting with '?' suitably encoded
     *     to drop into an [X]HTML attribute value at the end of a URL
     */
    public String computeGETQueryStringURLTail(final int pg)
        {
        final StringBuffer sb = new StringBuffer(MAX_GET_QUERY_CHARS);

        // Parameter separator to use: &amp; for [X][HT]ML attribute values.
        final String paramSep = "&amp;";

        // Hand-encode (as necessary) the main values for speed and reliability.
        // Always include the query text.
        sb.append("?q=").append(getQ().replace(' ', '+'));
        // Always include the page number if supplied.
        if(pg > 0)
            { sb.append(paramSep).append(WebConsts.PAGE_NUMBER_PARAMETER).append('=').append(pg); }

        try
            {
            final String encFormat = "UTF-8";

            // Handle optional parameters, most important first.
            // Use dummy loop to enable break-out when result getting too big.
            do
                {
                // Category.
                final String category = getCategory();
                if(category.length() > 0)
                    {
                    final StringBuffer tmp = new StringBuffer();
                    tmp.append(paramSep).append("category=").append(URLEncoder.encode(category, encFormat));
                    if(sb.length() + tmp.length() > MAX_GET_QUERY_CHARS) { break; }
                    sb.append(tmp);
                    }

                // Suffix.
                final String suffix = getSuffix();
                if(suffix.length() > 0)
                    {
                    final StringBuffer tmp = new StringBuffer();
                    tmp.append(paramSep).append("suffix=").append(URLEncoder.encode(suffix, encFormat));
                    if(sb.length() + tmp.length() > MAX_GET_QUERY_CHARS) { break; }
                    sb.append(tmp);
                    }

                // Author.
                final String author = getAuthor();
                if(author.length() > 0)
                    {
                    final StringBuffer tmp = new StringBuffer();
                    tmp.append(paramSep).append("author=").append(author); // Needs no encoding.
                    if(sb.length() + tmp.length() > MAX_GET_QUERY_CHARS) { break; }
                    sb.append(tmp);
                    }
                } while(false);
            }
        catch(final UnsupportedEncodingException e)
            {
            // TODO Auto-generated catch block
            e.printStackTrace();
            }

        return(sb.toString());
        }

    /**Last search time in ms, -1 if no search yet done. */
    private long lastSearchTimeMs = -1;

    /**Gets the last search time in ms, -1 if no search yet done.
     */
    public synchronized long getLastSearchTimeMs()
        { return(lastSearchTimeMs); }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -8656439331186143583L;
    }
