package org.hd.d.pg2k.webSvr.catalogue;

import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.LocaleBeanBase;
import org.hd.d.pg2k.webSvr.exhibit.Expr;
import org.hd.d.pg2k.webSvr.exhibit.TreeFilterBean;
import org.hd.d.pg2k.webSvr.util.Breadcrumbs;
import org.hd.d.pg2k.webSvr.util.LocaleBean;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.List;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 04-Jul-2003
 * Time: 20:23:55
 * To change this template use Options | File Templates.
 */

/**Servlet to do filtering and selection of one tree of exhibits for display.
 * Sub-classes must implement getTitleCommonCatalogueName() and getFilter().
 */
public abstract class TreeFilterServletBase extends HttpServlet
    {
    /**Bean that holds cache of selected data in tree-form for us.
     * Only created at init().
     */
    private final TreeFilterBean tfb = new TreeFilterBean();

    /**Handle a GET request.
     * Ultimately we will internally redirect to the tree-display JSP, unless:
     * <ul>
     * <li>There is an error, eg the request URI is malformed or impossible.
     * <li>We canonoicalise the request with a redirection.
     * </ul>
     */
    protected void doGet(final HttpServletRequest request,
                         final HttpServletResponse response)
        throws ServletException, IOException
        {
        // Get the base URI to pass to the display JSP/servlet.
        final String baseURI = request.getServletPath();

        // Extract tail of request containing embedded word sequence prefix,
        // ensuring that it is always non-null and
        // ends with some file component (to make parsing easier)...
        String tailURI = request.getPathInfo();
        if(tailURI == null) { tailURI = "/"; }

        // Convert this into a word-sequence prefix.
        String wsPrefix = "";
        try { wsPrefix = TreeFilterBean.convertTrailingURIToWordSequence(tailURI); }
        catch(final IllegalArgumentException e) // User entered malformed input.
            {
            // Mark malformed input as "not found".
            //response.sendError(response.SC_NOT_FOUND);
            request.setAttribute("javax.servlet.error.request_uri", tailURI);
            request.getRequestDispatcher(WebConsts.NOT_FOUND_PAGE).forward(request, response);
            return;
            }

//System.err.println("baseURI: " + baseURI);
//System.err.println("tailURI: " + tailURI);
//System.err.println("wsPrefix: " + wsPrefix);

        // Look up given prefix; if null it is an invalid URL.
        // Try for the root of this index instead.
        final ServletContext context = getServletContext();
        List<String> nodeContent = tfb.selectPrefixesAndShortNames(context, wsPrefix, null);
        boolean[] uninterestingNodes;
        Map<String,Integer> exhibitCounts;
        if(nodeContent == null)
            {
            // Try using the top node of the tree instead.
            // If this is still null, the display page will understand that it has no entries.
            // Note that the exhibit names are in short form to sort properly with the prefixes.
            nodeContent = tfb.selectPrefixesAndShortNames(context, "", null);
            uninterestingNodes = tfb.getUninterestingNodes(context, "", null);
            exhibitCounts = tfb.getExhibitCount(context, "", null);
            // Reset the prefix to match.
            wsPrefix = "";
            }
        else
            {
            uninterestingNodes = tfb.getUninterestingNodes(context, wsPrefix, null);
            exhibitCounts = tfb.getExhibitCount(context, wsPrefix, null);
            }

        // 302 TEMP REDIRECTS FOR NODES THAT CURRENTLY DUPLICATE THEIR PARENT NODE.
        // If we can redirect to a parent node (ie shorter URI) than this
        // that (currently) contains the same info, then we do,
        // to show shorter URLs to users and to reduce duplicate pages in SEs/caches.
        // Since the tree structure may change this must only be a temporary redirect.
        final boolean redirected = redirectIfCanonicalisationNeeded(tfb, response, baseURI, wsPrefix, context, nodeContent);
        if(redirected) { return; }

        // Allow for i18n.
        final LocaleBean localeBean = new LocaleBean();
        localeBean.setRequest(request, getServletContext());

        // Set breadcrumbs back to top of virtual collections...
        final Breadcrumbs crumbs = getBreadcrumbs(localeBean);

        // Redirect to display JSP, having stored the data we need in attributes at request level.
        request.setAttribute(WebConsts.STDHDR_ATTR_NAME_BREADCRUMBS, crumbs);
        request.setAttribute(WebConsts.BETDJ_BASE_URI_ATTR_NAME, baseURI);
        request.setAttribute(WebConsts.BETDJ_LIST_ATTR_NAME, nodeContent);
        request.setAttribute(WebConsts.BETDJ_LIST_UNIN_NAME, uninterestingNodes);
        request.setAttribute(WebConsts.BETDJ_LIST_SNEXCOUNT_NAME, exhibitCounts);
        request.setAttribute(WebConsts.BETDJ_PRWSEQ_ATTR_NAME, wsPrefix);
        request.setAttribute(WebConsts.BETDJ_TITLE_ATTR_NAME,
            localeBean.getLocalisedMessage(getTitleCommonCatalogueName()));
        request.setAttribute(WebConsts.BETDJ_HTML_GENERAL_DESC_ATTR_NAME,
            localeBean.getLocalisedMessage(getDescriptionCommonCatalogueName()));

        request.getRequestDispatcher(WebConsts.BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP).forward(request, response);
        }

    /**302 TEMP REDIRECTS FOR NODES THAT CURRENTLY DUPLICATE THEIR PARENT NODE.
     * @return true if a redirect has been done,
     *     so the caller should return immediately
     *     and not try to generate any page content.
     */
    public static boolean redirectIfCanonicalisationNeeded(final TreeFilterBean tfb,
                                                           final HttpServletResponse response,
                                                           final String baseURI,
                                                           final String wsPrefix,
                                                           final ServletContext context,
                                                           final List<String> nodeContent)
        throws IOException, IllegalStateException
        {
        // 302 TEMP REDIRECTS FOR NODES THAT CURRENTLY DUPLICATE THEIR PARENT NODE.
        // If we can redirect to a parent node (ie shorter URI) than this
        // that (currently) contains the same info, then we do,
        // to show shorter URLs to users and to reduce duplicate pages in SEs/caches.
        // Since the tree structure may change this must only be a temporary redirect.
        // We must go as far as we can in one step/redirection
        // since clients will object to multiple redirections one step at a time.
        // We do not (yet) attempt to preserve any query strings/parameters, etc...
        boolean redirected = false;
        if(WebConsts.CANON_TREE_URIS_WITH_REDIR && (wsPrefix.length() > 1))
            {
            // The best (shortest) prefix we can find with the same info; never null.
            String bestParentNodePrefix = wsPrefix;
            while(bestParentNodePrefix.length() > 1)
                {
                // Include everything up to and including
                // the *penultimate* word separator,
                // ie avoiding the very final character.
                final String parentPrefix = bestParentNodePrefix.substring(0,
                    1 + bestParentNodePrefix.lastIndexOf(ExhibitName.WORD_SEP, bestParentNodePrefix.length()-2));
                assert((parentPrefix.length() == 0) || (parentPrefix.endsWith(ExhibitName.WORD_SEPS))) : parentPrefix;
                assert(parentPrefix.length() < bestParentNodePrefix.length());
                assert(bestParentNodePrefix.startsWith(parentPrefix));

                // If the parent node contains the same info as the current node
                // then it becomes our redirection target but we continue looking,
                // else we stop since if this node doesn't match then higher ones won't either.
                if(nodeContent.equals(tfb.selectPrefixesAndShortNames(context, parentPrefix, null)))
                    { bestParentNodePrefix = parentPrefix; }
                else
                    { break; }
                }

            // Did we manage to find an equivalent parent node?
            // Do a temp redirect to it if so.
            if(bestParentNodePrefix.length() < wsPrefix.length())
                {
                response.sendRedirect(makeOfficialURI(baseURI, bestParentNodePrefix));
                redirected = true;
                }
            }
        return redirected;
        }

    /**This clears the cache in the filter bean.
     * We chain to super.destroy() to tidy up the underlying generic servlet.
     */
    public void destroy()
        {
        try
            {
            // Free up any now-redundant memory resources quickly.
            tfb.clearCache();
            }
        finally { super.destroy(); }
        }

    /**Initialises the servlet the filter bean.
     *
     * @throws javax.servlet.ServletException
     */
    public void init() throws ServletException
        {
        // By default retain cache of computed data as this is assumed to be expensive.
        tfb.setMemorySensitiveCache(false);

        // By default, set exactly once, at initialisation.
        tfb.setExpr(getFilter());
        }

    /**Get the filter expression to apply.
     * Override in derived classes.
     */
    protected abstract Expr getFilter();

    /**Get (short) title text common catalogue name; null if none.
     * The expanded text should be 7-bit printable ASCII,
     * possibly with embedded entities and mark-up.
     * <p>
     * Override in derived classes.
     */
    protected abstract String getTitleCommonCatalogueName();

    /**Get descriptive text (i18n-ed if possible); null if none.
     * The expanded text should be 7-bit printable ASCII,
     * possibly with embedded entities and mark-up.
     * <p>
     * Defaults to null (no descriptive text).
     * <p>
     * Override as needed in derived classes.
     */
    protected String getDescriptionCommonCatalogueName()
        { return(null); }

    /**Get Breadcrumbs; null if none.
     * By default, just a link to the top of the virtual collections,
     * but can either be added to or replaced by an overriding class.
     */
    protected Breadcrumbs getBreadcrumbs(final LocaleBeanBase localeBean)
        {
        final Breadcrumbs crumbs = new Breadcrumbs();
        crumbs.append(new Breadcrumbs.Breadcrumb(
            localeBean.getLocalisedMessage("common.main.VirtualCollections"),
            WebConsts.VIRTUAL_COLLECTIONS_ROOT));
        return(crumbs);
        }

    /**Function to provide full "official" URI for page given base URI and word prefix.
     * Note that the URI that this returns always ends with a slash
     * to unambiguiously mark the whole thing as a directory;
     * no trailing "welcome" file component is added nor should be.
     * <p>
     * The suffix is also forced to lower-case to ensure canonicalisation.
     *
     * @param baseURI  base URI for this tree; never null
     * @param prwseq  prefix word sequence, "" or lower-case word sequence
     *     ending in a hyphen.
     */
    public static String makeOfficialURI(final String baseURI, final String prwseq)
        {
        final String result = baseURI + (baseURI.endsWith("/") ? "" : "/") +
            prwseq.toLowerCase().replace(ExhibitName.WORD_SEP, '/');
        assert(result.endsWith("/")); // Always a directory.
        return(result);
        }
    }
