/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.AllExhibitImmutableData;
import org.hd.d.pg2k.svrCore.Rnd;

import java.io.IOException;
import java.io.Serializable;
import java.lang.ref.SoftReference;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 24-May-2003
 * Time: 17:42:56
 */

/**JavaBean that can filter/sort and cache selections of exhibits.
 * Designed to be used as a bean in a JSP or servlet, and can cache
 * filter/sort results, for example, at application level or in a
 * user's session.
 * <p>
 * This applies its filter to the input List (or entire exhibit set)
 * that it is provided with; deriving classes may make all or part
 * of the result directly or indirectly available.
 * <p>
 * This silently recomputes its results if they may have been affected by
 * any update in the exhibit data (or when select() is called differently
 * from the last call to the same instance).  The cached state is not persisted
 * if the bean is serialised, but the filter description is.  This is
 * only meant to be persisted, for example, to be passed between
 * multiple instances of a distributable WAR and not to be stored for lengthy
 * periods on disc, so does not use some of the fancier support for
 * serialisation.
 * <p>
 * This uses but does not cache a DataSourceBean which can be supplied
 * explicitly or recovered implicitly from a servlet context
 * (at application scope), and can take the whole set of exhibits as
 * input or can use the output of another filter.  In all cases the output
 * is only recomputed if the DataSourceBean.exhibitDataVersionHash()
 * method result changes (so any randomised or conditional element had better be the
 * last stage of the chain and only have (say) page or request scope),
 * or if select() is called with different args to the last call on this
 * instance (so, if for example, a different List argument is passed;
 * select does a test on the hash code of the select() List arg).
 * <p>
 * This is thread-safe and as concurrent as possible, with the intention
 * that it may be involved in many users' activity simultaneously.
 * <p>
 * The expression must be set (ideally exactly once) before any attempt
 * is made to fetch results.  For JSPs this set can be done in
 * the body of the JSP so that it is done only on creation of the bean
 * as a small efficiency gain.
 * <p>
 * If the expression is changed and is not equals() to the previous value,
 * the cache is cleared.
 * <p>
 * The result returned by any select() operation is an immutable, thread-safe,
 * RandomAccess List (eg like ArrayList).  (This cannot be enforced on
 * derived classes, but is part of the general contract.)
 * <p>
 * The expression is of the form:
 * <pre>
 *     expr = expr sortExpr
 *          | expr filterExpr
 *          | ALL
 *     sortExpr   = sort(args)
 *     filterExpr = filter(args)
 *                | filterExpr filterExpr &amp;&amp;
 *                | filterExpr filterExpr ||
 *                | filterExpr !
 * </pre>
 * ie a filter on its own is applied to all the available exhibits,
 * or can be negated or combined (ANDed or ORed) with other filter
 * expressions to make more complex filter expressions.
 * A complete expression is a series of filter expressions and/or
 * sorts applied in a pipeline to the result of some filtering.
 * <p>
 * The expression syntax is reverse Polish amd dataflow is from left
 * to right.
 * <p>
 * This class implements the Observer interface and when it receives
 * a notification its immediately clears its cache by default.
 * This is in addition to the normal automatic cache discard in select()
 * when the imput has changed, and is an optimisation; behvaiour is correct
 * without it.
 * This mechanism helps discard stale data quickly and thus make better use of memory.
 * <p>
 * This can also be set to expire automatically from time to time,
 * for example to allow for slowly mutating/changing data
 * such as ExhibitPropsComputableMutable items.
 * The default is not to expire periodically.
 * <p>
 * This Serializable so as to be able to be stored in
 * a servlet session; nothing especially long-lived or sensitive.
 * <p>
 * TODO: Should do validation on deserialisation.
 * <p>
 * FIXME: Observing the DataSourceBean may cause a deadlock.
 */
public abstract class AbstractFilterBean implements Serializable
    {
    /**If true, become an Observer of the DataSourceBean so that we can discard a stale cache quickly.
     * FIXME: Observing the DataSourceBean may cause a deadlock.
     */
    private static final boolean OBSERVE_DATASOURCEBEAN = true;

    /**The (filter/sorter) expression, or null if not yet set.
     * Accessed under the instance lock.
     * <p>
     * This is persisted if the object is serialised.
     */
    private Expr expr;

    /**Do we let any cached value go if we run short of memory?
     * This is useful for infrequently-accessed and not-too-expensive
     * filters.
     * <p>
     * This can be set or reset at any time but is only effective when
     * the cache is filled, so is probably done once, before the first
     * set of values is retrieved.
     * <p>
     * By default this value is true and the cache is memory sensitive,
     * ie won't cause the system to run out of memory.
     * <p>
     * Accessed under the instance lock.
     * <p>
     * This is persisted if the object is serialised.
     */
    private boolean memorySensitiveCache = true;



    /**Expiry interval in ms, zero (the default) means no automatic expiry; non-negative.
     * The data may expire in a little less than this.
     * <p>
     * Is volatile so as to allow access without a lock.
     */
    private int expiryInterval;

    /**Get expiry interval in ms, zero (the default) means no automatic expiry; non-negative.
     * The data may expire in a little less than this.
     */
    public int getExpiryInterval()
        { return(expiryInterval); }

    /**Set expiry interval in ms, zero means no automatic expiry; non-negative.
     * The data may expire in a little less than this
     * at random to spread out expiries and recalculations.
     */
    public void setExpiryInterval(final int expiryInterval)
        {
        if(expiryInterval < 0) { throw new IllegalArgumentException(); }
        this.expiryInterval = expiryInterval;
        }





    /**Set exhibitDataHash. */
    private int getExhibitDataHash()
        { return(exhibitDataHash); }

    /**Get exhibitDataHash. */
    private void setExhibitDataHash(final int exhibitDataHash)
        { this.exhibitDataHash = exhibitDataHash; }

    /**The exhibit data hash when the cache was computed.
     * If this changes we should regard the cache as invalid.
     * <p>
     * Is volatile so that it can be accessed from multiple
     * threads without taking a lock.
     * <p>
     * Should be accessed only by getExhibitDataHash() and
     * setExhibitDataHash().
     * <p>
     * Zero when the cache is clear (the initial state).
     */
    private transient volatile int exhibitDataHash;

    /**The hashCode of the last List arg to select or 0 if none.
     * This is used to try to cheaply ensure that we discard our
     * cache if our calling pattern changes (which should not
     * happen, but this may save us from poor usage).
     * <p>
     * Accessed under the instance lock and is private to
     * _select() and clearCache().
     * <p>
     * Zero when the cache is clear.
     */
    private transient int lastListArgHashCode;

    /**The cached results (or null if not yet computed).
     * This consists of an immutable RandomAccess List,
     * not sorted unless the expression sorts it, of the accepted
     * results; no slots in the List are duplicates or nulls.
     * <p>
     * If the exhibit data hash (or select()-arg hash)
     * changes then we should regard the cache as invalid.
     * <p>
     * This is null only if not yet computed; it is zero length
     * if the result set was empty.
     * <p>
     * Accessed under the instance lock and is private to
     * _select() and clearCache().
     * <p>
     * Is either a List or a SoftReference wrapper around one.
     */
    private transient Object _basicCache;

    /**Embedded observer object.
     * Designed to break link between DataSourceBean and our class
     * to allow us to be GCed even if we are recorded as an observer.
     * <p>
     * This inner object has only a SoftReference to us.
     */
    private final MyObserver observer = new MyObserver(this);

    /**Set the `memory-sensitive' flag for the cache.
     * This is useful for infrequently-accessed and not-too-expensive
     * filters; if set true then the cache is automatically discarded
     * if we run short of heap space.
     * <p>
     * This can be set or reset at any time but is only effective when
     * the cache is filled, so is probably best done once, before the first
     * set of values is retrieved.
     * <p>
     * By default this value is false and the cache is not memory sensitive.
     * <p>
     * This is persisted if the object is serialised.
     */
    public synchronized void setMemorySensitiveCache(final boolean f)
        { memorySensitiveCache = f; }

    /**Get the `memory-sensitive' flag for the cache.
     */
    public synchronized boolean getMemorySensitiveCache()
        { return(memorySensitiveCache); }

    /**Set the simple single (input) filter/sort expression directly.
     * If the expression is set to a new value not equals() to the old,
     * then the cache is cleared.
     *
     * @throws java.lang.IllegalArgumentException  if the argument is null
     */
    public synchronized void setExpr(final Expr _expr)
        throws IllegalArgumentException, IllegalStateException
        {
        if(_expr == null) { throw new IllegalArgumentException(); }
        if(!_expr.equals(expr)) { clearCache(); }
        expr = _expr;
        }

    /**Get the simple single (input) filter/sort expression directly.
     * Has protected scope to prevent every Tom Dick and Harry fiddling with
     * a possibly-mutable expression.
     * <p>
     * This can be overridden to provide some extra filtering or sorting, for example,
     * in derived classes.
     *
     * @return the expression or null if none yet set
     */
    protected synchronized Expr getExpr()
        {
        return(expr);
        }

    /**Set the filter/sorter expression as a textual expression.
     * Will clear the cache if the expression changes.
     *
     * @throws java.lang.IllegalArgumentException  if the argument is null or unparseable
     */
    public synchronized void setExpr(final String _expr)
        throws IllegalArgumentException, IllegalStateException
        {
        if(_expr == null) { throw new IllegalArgumentException(); }

        // PARSE THE EXPESSION.
        throw new RuntimeException("NOT IMPLEMENTED");
        }

    /**Clears the cache of the basic set of filtered/sorted items.
     * Replaces any cached item with null and
     * resets the internal hashes of the data and argument to null.
     * <p>
     * A deriving class can override this to clear any extra state it maintains,
     * but must chain to this before in order to clear the underlying cache.
     * <p>
     * This routine is synchronized and overriding methods
     * almost certainly should be so too.
     */
    protected synchronized void clearCache()
        {
        _basicCache = null;
        setExhibitDataHash(0);
        lastListArgHashCode = 0;
        }

    /**Invalidate the cache.
     * Can be called by anyone;
     * may force the cache to be invalidated immediately or before next use.
     * <p>
     * May need to grab locks.
     */
    public void invalidate()
        { clearCache(); }


    /**Expiry time.
     * Private to _isValid and accessed in the scope of its lock.
     * <p>
     * Transient to force an immediate recomputation if deserialised.
     */
    private transient long expiryTime;

    /**Returns cache reference if the cache still seems to be valid.
     * If null, then the cache needs to be rebuilt.
     * As a side-effect, this will clear an invalid cache.
     * <p>
     * Parameters as for _select(),
     * return value as for _select() or null if invalid.
     * <p>
     * Cheaper than calling _select() for force the cache to be invalidated
     * since it does not attempt to rebuild the cache.
     */
    protected synchronized List<String> _isValid(final DataSourceBean ds,
                                                 final List<String> inputSet)
        throws IOException
        {
        // Cache is (1) null or (2) a List or (3) a SoftReference to a List.
        final List<String> currentCache = (_basicCache instanceof SoftReference)
          ? ((List<String>) (((SoftReference) _basicCache).get()))
          : ((List<String>) _basicCache);

        final boolean isInvalid =
           (currentCache == null) ||
           ((expiryInterval != 0) && (expiryTime < System.currentTimeMillis())) ||
           (lastListArgHashCode != _computeListArgHashCode(inputSet)) ||
           (getExhibitDataHash() != ds.exhibitDataVersionHash());

        if(isInvalid)
            {
            // Abolish old cached values.
            clearCache();
            return(null);
            }

        return(currentCache);
        }

    /**Compute the hash over our List argument, used by _isValid() and _select().
     * Note that the identityHashCode() is used.
     */
    private static int _computeListArgHashCode(final List<String> inputSet)
        { return((inputSet == null) ? 0 : System.identityHashCode(inputSet)); }

    /**Internal method that computes the result of accepted exhibits.
     * The result is a set of full exhibit names as a private unsorted array.
     * <p>
     * The ds argument must not be null.
     * <p>
     * If the inputSet argument is null the whole set of
     * exhibits is used, else the supplied set is used.
     * This must be a List of String items with no nulls nor
     * duplicates.
     * <p>
     * This will (re)compute its results if necessary, ie:
     * <ul>
     * <li>there are no cached results,
     * <li>the exhibit hash has changed,
     * <li>the List argument hashCode has changed.
     * </ul>
     * otherwise this routine is relatively fast and returns the cached List.
     * <p>
     * This never returns null.
     * <p>
     * Null names and names that do not represent items currently
     * in the exhibit set are silently dropped.
     * <p>
     * If this bean is not marked as memory sensitive,
     * then this routine will register this instance with the data source
     * as an Observer of exhibit set changes in order to be able to quickly
     * invalidate our (then-stale) cache upon any changes.
     *
     * @param inputSet  null, or a list of full exhibit names
     *
     * @throws java.lang.IllegalStateException  if the filter has not been set
     */
    protected synchronized List<String> _select(final DataSourceBean ds,
                                                final List<String> inputSet)
        throws IOException, IllegalStateException
        {
        if(expr == null) { throw new IllegalStateException(); }

        // Cache is (1) null or (2) a List or (3) a SoftReference to a List.
        List<String> currentCache = _isValid(ds, inputSet);

        // Hash over all exhibits...
        final int currentHash = ds.exhibitDataVersionHash();

        // If the cache is invalid, (re)build it.
        // The _isValid() call will have cleared the cache already if invalid.
        if(currentCache == null)
            {
            // Warn against possible misuse of a filter, dumping a stack trace.
            if(lastListArgHashCode != 0)
                {
                final Error e = new Error("WARNING: FilterBean List argument changed between non-null input List items which is probably a programming error; please check");
                e.printStackTrace();
                }

            // Recompute the results...
//if(ORG.hd.d.IsDebug.isDebug) { System.err.println("[Recomputing AbstractFilterBean result for: " + this + ".]"); }

            final AllExhibitImmutableData aeid = ds.getAllExhibitImmutableData(-1);

            // We can optionally Observe the DataSourceBean so as to
            // discard our cache when it changes.
            // FIXME: At the moment this presents a potential deadlock hazard.
            if(OBSERVE_DATASOURCEBEAN)
                {
                // If not marked as memory-sensitive, then
                // register to observe changes (to clear our cache).
                // No firm promises are made to deliver these,
                // and many or all may be missed.
                // Receiving one just helps us clear our cache ASAP,
                // helping recycle memory quickly.
                if(!memorySensitiveCache)
                    { ds.addObserver(observer); }
                }

            // Use the input set if supplied,
            // else examine the whole exhibit set...
            String inputNames[];
            if(inputSet == null) // No input set; start with all names.
                { inputNames = aeid.getAllExhibitNamesSorted(); }
            else
                {
                // Extract String[] copy of input List.
                //
                // We should consider verifying no duplicate inputs
                // in debug mode...
                inputNames = new String[inputSet.size()];
                inputSet.toArray(inputNames);
                // Quick check for gross problems in copy...
                if((inputNames.length > 0) && (inputNames[0] == null))
                    { throw new IllegalArgumentException("problem copying inputSet"); }
                }

            // Apply the expression to the input set...
            final String[] rawResult = getExpr().eval(ds.getAllExhibitProperties(-1L), inputNames);
            // ...and prepare to cache an unmodifiable (and thread-safe) copy.
            currentCache = Collections.unmodifiableList(Arrays.asList(rawResult));

            // Cache the data, with a strong or soft reference as appropriate.
            if(memorySensitiveCache)
                { _basicCache = new SoftReference<List<String>>(currentCache); }
            else
                { _basicCache = currentCache; }

            // We use the hash we got before recomputing the results
            // so that we recompute again later if the cached
            // data changed while we were in this function, ie
            // we err on the side of caution.
            setExhibitDataHash(currentHash);

            // We store the hash of the current arg list,
            // so if it changes we are likely to notice and
            // will recompute the cached value.
            lastListArgHashCode = _computeListArgHashCode(inputSet);

            // Set new expiry time if on automatic expiry.
            // No more than specified expiry interval;
            // may be as little as half that.
            if(expiryInterval > 0)
                {
                expiryTime = System.currentTimeMillis() +
                        ((expiryInterval/2)|1) + Rnd.fastRnd.nextInt((expiryInterval/2)|1);
                }
            }

        // Return immutable list...
        return(currentCache);
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -4409526925504499328L;




    /**This object has only a SoftReference to the outer class to allow GC.
     * This way it remains fairly safe for us to register with a DataSourceBean
     * even if the outer class may be short-lived and churned in huge numbers.
     * <p>
     * FIXME: should find way to stop ourself from observing any DataSourceBean
     *     (there can be at most one)
     *     when the outer object is GCed or maybe when the cache is cleared.
     *     Else, lots of redundant (if small) MyObserver objects may
     *     accumulate in the DataSourceBean until the exhibit set changes.
     */
    private static final class MyObserver implements Observer
        {
        /**Capture a SoftReference to the outer class.
         * @param outer  non-null reference to outer class.
         */
        MyObserver(AbstractFilterBean outer)
            { srOuter = new SoftReference<AbstractFilterBean>(outer); }

        /**SoftReference to outer class; never null. */
        private final SoftReference<AbstractFilterBean> srOuter;

        /**This causes the cache to be cleared immediately.
         * If [Object] arg is an Integer whose value is same as the value that
         * we hold for the DataSourceBean.ds.exhibitDataVersionHash()
         * then we return immediately on the assumption that this indicates
         * that we still hold an up-to-date cache.
         * This avoids extra builds and clears of the cache when the FilterBean
         * is accessed frequently, and in particular between the underlying
         * data changing and a call being made sometime later to this routine
         * to clear all potentially-expired caches.
         * <p>
         * We use [Observable] o to deregister ourself from the Observable
         * on the grounds that our next select() operation will (re-)register
         * us with whatever DataSourceBean we are then using.
         * This removal avoids leaving references to us
         * from multiple DataSourceBeans or after we have otherwise expired,
         * which would otherwise make GC harder.
         * <p>
         * The cache-clearing code is synchronized so the caller may be briefly
         * blocked and there may be a small risk of deadlocks if clearCache()
         * does something foolish.
         */
        public final void update(final Observable o, final Object arg)
            {
            final AbstractFilterBean outer = srOuter.get();
            if(outer == null)
                {
                // Outer class has been GCed,
                // so we don't want to be an Observer any more!
                o.deleteObserver(this);
                return;
                }

            // We take a lock only to perform the clearCache()
            // at a slight risk of a race with a resulting
            // extra or missed cache clear.

            // If the argument indicates that we are in fact still up to date,
            // return immediately.
            if((arg instanceof Integer) &&
                (((Integer) arg).intValue() == outer.getExhibitDataHash()))
                { return; }

            o.deleteObserver(this); // No more updates until the next select().
            outer.invalidate(); // OK, cache probably stale, so allow GC...
            return;
            }
        }
    }
