/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.props.GenProps;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Random;

/**Built-in filters and sorters that can be looked up by name.
 * The filters are inner classes and not publicly visible
 * so that by-and-large they do have to be looked up by name.
 * There is an internal registry so that reflection doesn't
 * have to be used and packagers such as dashO should have
 * any easier time working out what's going on...
 * <p>
 * All the filters/sorters are constructed with a String array or arguments.
 * This argument list is never null, though may be zero length if there are
 * no arguments for a particular filter/sorter.  None of the arguments
 * itself will be null.
 * <p>
 * The exhibit names are full names, ie including the leading directory component.
 */
public final class BuiltInFilters
    {
    /**No one can instantiate this. */
    private BuiltInFilters() { }

    /**Accepts all exhibits.
     */
    public static final class filtAll implements FilterIF
        {
        /**Ignores any arguments.
         */
        public filtAll(String args[]) { }
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            { return(true); }

        /**Hash is constant. */
        public int hashCode() { return(0); }
        /**All instances are equal (could be a singleton). */
        public boolean equals(Object obj) { return(obj instanceof filtAll); }        }

    /**Rejects all exhibits.
     */
    public static final class filtNothing implements FilterIF
        {
        /**Ignores any arguments.
         */
        public filtNothing(String args[]) { }
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            { return(false); }

        /**Hash is constant. */
        public int hashCode() { return(0); }
        /**All instances are equal (could be a singleton). */
        public boolean equals(Object obj) { return(obj instanceof filtNothing); }
        }

    /**Filter by author.
     * FIXME: add equals() and hashCode().
     */
    public static final class filtByAuthor implements FilterIF
        {
        private final String acceptableAuthors[];
        /**Accepts zero or more author (initials); an exhibit matching any of them is accepted.
         */
        public filtByAuthor(String args[]) { acceptableAuthors = (String[]) args.clone(); }
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            {
            final String initials = ExhibitName.getAuthorComponent(exhibitName);
            for(int i = acceptableAuthors.length; --i >= 0; )
                { if(initials.equals(acceptableAuthors[i])) { return(true); } }
            return(false); // No match.
            }
        }



    /**Filter by category.
     * FIXME: add equals() and hashCode().
     */
    public static final class filtByCategory implements FilterIF
        {
        private final String acceptableCategories[];
        /**Accepts zero or more categories; an exhibit matching any of them is accepted.
         */
        public filtByCategory(String args[]) { acceptableCategories = (String[]) args.clone(); }
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            {
            final String category = ExhibitName.getCategoryComponent(exhibitName);
            for(int i = acceptableCategories.length; --i >= 0; )
                { if(category.equals(acceptableCategories[i])) { return(true); } }
            return(false); // No match.
            }
        }


    /**Accepts an exhibit with the given case-sensitive substring in its full pathname.
     * This is no respector of syntax, or aliases.
     * <p>
     * Very fast and does not look at the data or meta-data.
     * <p>
     * As a special case this is publicly visible until expression
     * parsing is implemented.
     * <p>
     * FIXME: add equals() and hashCode().
     */
    public   static final class filtSimpleSubstringMatch implements FilterIF
        {
        /**Takes exactly one argument; the case-sensisitve substring to match.
         */
        public   filtSimpleSubstringMatch(String args[])
            {
            if((args.length != 1) /* || (args[0] == null) */)
                { throw new IllegalArgumentException(); }
            subString = args[0];
            }
        /**The substring to match. */
        private final String subString;
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            { return(exhibitName.indexOf(subString) != -1); }
        }

    /**Accepts an that has an Estd location within the target area; possibly within the specified area.
     */
    public   static final class filtByEstdLocationCentre implements FilterIF
        {
        /**Particular location to filter by.
         * If non-null, item must fall within the area delimited by this
         * location to be considered.
         */
        private final Location.Estd area;

        /**Takes zero or one arguments; the (optional) area and bounds to filter else any Estd location is acceptable.
         */
        public   filtByEstdLocationCentre(String args[])
            {
            if((args != null) || (args.length >= 0))
                { throw new Error("NOT IMPLEMENTED"); }
            area = null;
            }

        /**Takes area and bounds to filter for; if null any Estd location is acceptable.
         */
        public   filtByEstdLocationCentre(Location.Estd filterArea)
            {
            area = filterArea;
            }

        public boolean accept(AllExhibitProperties aep, String exhibitName)
            {
            final Location.Base l = aep.getLocation(exhibitName);
            if(!(l instanceof Location.Estd))
                { return(false); } // Wrong type or no location info.

            // Filter on area if so requested.
            if(area != null)
                {
                final Location.Estd el = (Location.Estd) l;
                if(!area.containsCentre(el))
                    { return(false); } // Centre not in specified area.
                }

            return(true); // Seems OK.
            }

        /**Filters are equal if area objects are (or are both null). */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof filtByEstdLocationCentre)) { return(false); }
            final filtByEstdLocationCentre other = (filtByEstdLocationCentre) obj;
            if(area == null) { return(other.area == null); }
            return(area.equals(other.area));
            }

        /**Filter has hashcode of its area; 0 if none. */
        public int hashCode()
            {
            if(area == null) { return(0); }
            return(area.hashCode());
            }
        }



    /**Accepts an exhibit with the given case-sensitive substring starting its final component.
     * This is not much of a respector of syntax.
     * <p>
     * Very fast and does not look at the data or meta-data.
     * <p>
     * As a special case this is publicly visible until expression
     * parsing is implemented.
     * <p>
     * FIXME: add equals() and hashCode().
     */
    public   static final class filtPrefixMatch implements FilterIF
        {
        /**Takes exactly one argument; the case-sensisitve substring to match.
         */
        public   filtPrefixMatch(String args[])
            {
            if((args.length != 1) /* || (args[0] == null) */)
                { throw new IllegalArgumentException(); }
            prefix = args[0];
            }
        /**The substring to match. */
        private final String prefix;
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            { return(ExhibitName.getFileComponent(exhibitName).startsWith(prefix)); }
        }

    /**Accepts an exhibit with the given case-sensitive substring ending its final component.
     * This is not much of a respector of syntax.
     * <p>
     * Very fast and does not look at the data or meta-data.
     * <p>
     * As a special case this is publicly visible until expression
     * parsing is implemented.
     * <p>
     * FIXME: add equals() and hashCode().
     */
    public   static final class filtSimpleSuffixMatch implements FilterIF
        {
        /**Takes exactly one argument; the case-sensisitve substring to match.
         */
        public   filtSimpleSuffixMatch(String args[])
            {
            if((args.length != 1) /* || (args[0] == null) */)
                { throw new IllegalArgumentException(); }
            suffix = args[0];
            }
        /**The substring to match. */
        private final String suffix;
        public boolean accept(AllExhibitProperties aep, String exhibitName)
            { return(ExhibitName.getFileComponent(exhibitName).endsWith(suffix)); }
        }


    /**Selects N exhibits at random and randomises their order.
     * If the number of exhibits is less than N then this returns
     * all the exhibits.
     * <p>
     * The first value is the unsigned, positive, base-ten integer value
     * of N.
     * <p>
     * If the second argument is missing or null then the random number
     * sequence used is taken differently each time we run, else the
     * hash of the second string is taken as the random number seed.
     */
    public   static final class sortRandomN implements SortIF
        {
        /**Takes exactly one or two arguments.
         * The first argument is the strictly-positive N.
         * <p>
         * The second argument, if absent or null, indicates that
         * a different random number sequence is used on each run.
         * If present and non-null then its hash is used to seed a
         * random number generator to give repeatable results.
         */
        public   sortRandomN(String args[])
            {
            this(Integer.parseInt(args[0]),
                 ((args.length > 1) ? args[1] : null));
//            if((args.length < 1) || (args.length > 2))
//                { throw new IllegalArgumentException(); }
            }

        /**More efficient programmatic constructor.
         * Takes count as the first (int) argument and
         * a String to drive the hash or null for a totally random selection.
         */
        public sortRandomN(int count, String hashSrc)
            {
            if(count <= 0) { throw new IllegalArgumentException(); }
            n = count;
            seed = hashSrc;
            }

        /**The upper bound on result size. */
        private final int n;
        /**The seed; if null we use a non-repeatable source. */
        private String seed;

        public String[] sort(AllExhibitProperties aep, String exhibitNames[])
            {
            // Set up our random-number source.
            final Random rnd = (seed == null) ?
                Rnd.fastRnd : (new Random(seed.hashCode()));

            String result[];
            if(exhibitNames.length <= n)
                {
                // We don't need to discard any of our input set...
                result = exhibitNames;
                }
            else
                {
                // We need to pick n items at random
                // from our input set, with no duplicates.
                // We'll make a boolean array logically beside
                // our input array, and set a slot true when
                // we've used the input value at the same index...
                final boolean used[] = new boolean[exhibitNames.length];

                // Output array...
                result = new String[n];

                // This will work badly where n is large and just
                // smaller than the number of available exhibits.
                // We should possibly reverse the mechanism in
                // this case, and mark values not to use...
                for(int i = n; --i >= 0; )
                    {
                    // Go on looking for an input slot that we have not yet found...
                    for( ; ; )
                        {
                        // Pick an input slot at random...
                        final int slot = rnd.nextInt(exhibitNames.length);
                        // If we've already used the value in this slot, try again...
                        if(used[slot]) { continue; }
                        // Whopee!  A new one; capture this one and break out...
                        result[i] = exhibitNames[slot];
                        used[slot] = true;
                        break;
                        }
                    }
                }

            // Randomise the result set...
            Collections.shuffle(Arrays.asList(result), rnd);

            return(result);
            }

        /**Equality depends on identical seeds and result size.
         * In other words this checks equivalence rather than the
         * results being identical each time.
         */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof sortRandomN)) { return(false); }
            final sortRandomN other = (sortRandomN) obj;

            if(n != other.n) { return(false); }

            // If exactly the same seed (eg a static final), then equal.
            if(seed == other.seed) { return(true); }

            // If we are null (and therefore the other is not), then not equal.
            if(seed == null) { return(false); }

            // Check seeds for equality.
            return(seed.equals(other.seed));
            }

        /**Hash depends on the seed and the result size limit. */
        public int hashCode()
            {
            return(n + ((seed == null) ? 0 : seed.hashCode()));
            }
        }


    /**Sorts exhibits by (ascending) timestamp, ie oldest first.
     * Takes no parameters, so all instances of this class are equal.
     */
    public   static final class sortByTimestamp implements SortIF
        {
        /**Ignores its arguments.
         */
        public   sortByTimestamp(final String args[])
            {  }

        /**Programmatic constructor.
         */
        public   sortByTimestamp()
            {  }

        /**Sorts its arguments in situ by ascending timestamp (oldest first).
         * Ties are broken by full name.
         */
        public String[] sort(final AllExhibitProperties aep,
                             final String exhibitNames[])
            {
            Collections.sort(
                Arrays.asList(exhibitNames),
                new Comparator(){
                    public final int compare(final Object o1, final Object o2)
                        {
                        final String n1 = (String) o1;
                        final String n2 = (String) o2;
                        final ExhibitStaticAttr e1 = aep.aeid.getStaticAttr(n1);
                        final ExhibitStaticAttr e2 = aep.aeid.getStaticAttr(n2);
                        if(e1.timestamp < e2.timestamp) { return(-1); } // Correct.
                        if(e1.timestamp > e2.timestamp) { return(+1); } // Wrong.
                        // Break tie unconditionally...
                        return(e1.compareTo(e2));
                        }
                    });

            return(exhibitNames);
            }

        /**All instances of this class are identical.
         */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof sortByTimestamp)) { return(false); }
            return(true);
            }

        /**Hash is constant. */
        public int hashCode()
            {
            return(0);
            }
        }


    /**Sorts exhibits by (descending) goodness, ie best first.
     * Takes no parameters, so all instances of this class are equal.
     * <p>
     * Cannot supply all the data needed to recompute a full version,
     * so works with what has already been computed or may force an
     * approximate ("fast") computation.
     */
    public   static final class sortByGoodness implements SortIF
        {
        /**Ignores its arguments.
         */
        public   sortByGoodness(final String args[])
            {
            this.nMax = 0;
            }

        /**Programmatic constructor.
         * @param nMax  if positive, is the maximum number of exhibits returned
         */
        public   sortByGoodness(final int nMax)
            {
            this.nMax = nMax;
            }

        /**If positive, is the maximum number of exhibits returned. */
        private final int nMax;

        /**Sorts its arguments in situ by ascending timestamp.
         * Ties are broken by full name.
         */
        public String[] sort(final AllExhibitProperties aep,
                             final String exhibitNames[])
            {
            final GenProps gp = new GenProps();
            final boolean allowStale = true; // Make do with what is already computed.

            Collections.sort(
                Arrays.asList(exhibitNames),
                new Comparator<String>(){
                    public final int compare(final String n1, final String n2)
                        {
                        final ExhibitPropsComputableMutable epcm1 = aep.getExhibitPropsComputableMutable(n1, allowStale, gp, null, null);
                        final ExhibitPropsComputableMutable epcm2 = aep.getExhibitPropsComputableMutable(n2, allowStale, gp, null, null);

                        // Sort best items first.
                        // Use the int goodness value for speed and precision.
                        if(epcm1.getGoodness() > epcm2.getGoodness()) { return(-1); } // Correct.
                        if(epcm1.getGoodness() < epcm2.getGoodness()) { return(+1); } // Wrong.

                        // Break ties (not resolved by goodness).
                        // Sort newer items first
                        // as potentially more interesting
                        // and so that they will get looked at!
                        final ExhibitStaticAttr e1 = aep.aeid.getStaticAttr(n1);
                        final ExhibitStaticAttr e2 = aep.aeid.getStaticAttr(n2);
                        if(e1.timestamp > e2.timestamp) { return(-1); }
                        if(e1.timestamp < e2.timestamp) { return(+1); }

                        // Break tie unconditionally by name.
                        // Should almost never be used.
                        // Only here to guarantee a total ordering.
                        return(e1.filePath.compareTo(e2.filePath));
                        }
                    });

            // If we are limiting the result length,
            // then return the "creme de la creme" here!
            if((nMax > 0) && (nMax < exhibitNames.length))
                {
                // Return a subset of the items; the very best ones in fact...
                final String result[] = new String[nMax];
                System.arraycopy(exhibitNames, 0, result, 0, nMax);
                return(result);
                }

            return(exhibitNames);
            }

        /**All instances of this class are identical where their nMax is.
         */
        public boolean equals(Object obj)
            {
            if(!(obj instanceof sortByGoodness)) { return(false); }
            final sortByGoodness other = (sortByGoodness) obj;
            return(nMax == other.nMax);
            }

        /**Hash is constant. */
        public int hashCode()
            {
            return(nMax);
            }
        }

    /**Simple sort of exhibits by name.
     * By default sorts in natural String order,
     * but can take a Comparator argument to use
     * (eg to use a 'smart' ordering(s) for human benefit).
     */
    public   static final class sortByName implements SortIF
        {
        /**Simple sort; default String lexical ordering; can use null instead. */
        public static final String SIMPLE = "simple";

        /**Simple smart sort; essentially case-insensitive ordering on file component. */
        public static final String SIMPLE_SMART = "simpleSmart";

        /**Full smart sort; gets current version of comparator during sort().
         * This uses the data-dependent sort
         * (dependent on the set of attribute words)
         * that comes with the data so that it is always current.
         */
        public static final String SMART = "smart";

        /**Sort Comparator; null for default (simple) sort. */
        private final String sortType;

        /**Call with argument giving sort type (or null for natural ordering).
         * Recognised values are:
         * <ul>
         * <li>simple (default String lexical ordering)
         * <li>simpleSmart (essentially case-insensitive ordering on file component)
         * <li>smart (fully-smart case-insensitive sort using attribute words)
         * </ul>
         */
        public sortByName(String sortType)
            {
            this.sortType = sortType;
            }


        /**Call with no args or single argument giving sort type.
         * Recognised values are:
         * <ul>
         * <li>simple (default String lexical ordering)
         * <li>simpleSmart (essentially case-insensitive ordering on file component)
         * <li>smart (fully-smart case-insensitive sort using attribute words)
         * </ul>
         */
        public sortByName(String args[])
            {
            if((args != null) && (args.length > 0))
                {
                sortType = args[0];
                }
            else
                { sortType = null; } // Default sort.
            }

        public String[] sort(AllExhibitProperties aep, String exhibitNames[])
            {
            // Simple smart order: not dependent on attribute words.
            if(sortType.equals(SIMPLE_SMART))
                { Arrays.sort(exhibitNames, ExhibitName.SIMPLE_SMART_ORDER); }

            // Fully-smart sort: dependent on attribute words.
            else if(sortType.equals(SMART))
                { Arrays.sort(exhibitNames, ExhibitAttrUtils.getAttrWords().SMART_ORDER); }

            // Default case (eg if comp is SIMPLE or null or unknown).
            // Simple sort in String order: fast but not human-friendly.
            // Not dependent on attribute words.
            else
                { Arrays.sort(exhibitNames); }
            return(exhibitNames);
            }

        /**The hashcode of the sort, or null for the default sort. */
        public int hashCode()
            {
            return(sortType == null ? 91 : sortType.hashCode());
            }

        /**Tests equality between two sortByName objects.
         * For this to be the case the two comparator objects must
         * be == (eg both be null), or compare equals().
         */
        public boolean equals(Object obj)
            {
//System.err.println("sortByName.equals() ...");

            // Always equal to myself.
            if(obj == this) { return(true); }

            if(!(obj instanceof sortByName)) { return(false); }
            final sortByName other = (sortByName) obj;

            // If exactly the same sort type (eg a static final), then equal.
            if(other.sortType == sortType) { return(true); }

            // If type is null (and therefore the other is not), then not equal.
            if(sortType == null) { return(false); }

            // Compare sort types for equality.
            if(!sortType.equals(other.sortType)) { return(false); }

//System.err.println("sortByName.equals() = true");

            return(true); // Equivalent sorts.
            }
        }

    }
