/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import ORG.hd.d.jIndexer.server.*;

import org.hd.d.pg2k.ai.scorer.*;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.datasource.*;
import org.hd.d.pg2k.svrCore.location.*;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.*;
import org.hd.d.pg2k.webSvr.util.*;
import org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth.*;

import javax.servlet.ServletContext;
import java.io.*;
import java.lang.ref.SoftReference;
import java.net.*;
import java.nio.ByteBuffer;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;

/**JavaBean encapsulating access to exhibit data and meta-data.
 * This is expected to exist in one copy for the exhibit servlet,
 * enabling (cached) access exhibit data and meta data.
 * <p>
 * This also provides access to shared data-pipeline and cache
 * facilities, and is expected to be created on demand at application
 * level.  This, therefore, is designed to be used as a bean, and
 * is suitable for direct instantiation from a JSP.
 * <p>
 * We have a public no-arg constructor to allow use from JSPs.
 * <p>
 * This object is thread-safe, and maximises concurrency while
 * synchronizing where necessary.  (Quick calls are synchronized
 * on this instance; longer-running calls may be synchronized on a
 * private lock.)
 * <p>
 * This creates an internal thread to maintain the cache, and the
 * thread is destroyed when the object is GCed.  We only hold a
 * SoftReference back to this instance from the background thread.
 * <p>
 * External callers can call the poll() method, though it is not
 * recommended.  They can also call the destroy method to close
 * down the cache and stop the background thread, though this
 * object must not be used thereafter.
 * <p>
 * When initialised for first use a number of parameters need to be set.
 * (For JSP, these should only be done upon creation, eg between the
 * useBean ... /useBean tags, to save cycles.)
 * Many of the actual values passed are not stored but just used
 * to extract other necessary parameters.
 * Some parameters, notably contextPath and servletContext, should be
 * set just before each use after retrieving the data from
 * attributes, in case this instance has been passivated and restored.
 * <p>
 * (The servlet context parameter should be set every time the bean
 * is used so that we always have a valid value to hand; we don't
 * persist this if the bean is serialised.)
 * <p>
 * If an attempt to made to use the bean without initialising it, or
 * to reinitialise it with conflicting values, an exception is thrown.
 * <p>
 * If an attempt is made to serialise/deserialise is made,
 * the internal cache is reconstructed; careless use may result in
 * multiple cache copies per servlet context which is wasteful,
 * and may be unsafe.
 * <p>
 * One item created here is the by-word inverted index of the exhibits,
 * created on first use.
 * <p>
 * This supports simple logging; by preference to the servlet context log,
 * but failing that to System.out.
 * <p>
 * This Serializable so as to be able to be stored in a servlet session;
 * nothing especially long-lived or sensitive.
 */
public final class DataSourceBean implements SimpleExhibitPipelineIF,
                                             SimpleLoggerIF,
                                             Serializable
    {
    /**Name that should be used for unique application-level instance. */
    public static final String BEAN_NAME = "dataSource";

    /**Factory method to create/return the unique application-level instance; never returns null.
     * Makes an instance at application level, or returns an extant one
     * if present.
     * <p>
     * Grabs appropriate locks to ensure that there are no data races
     * and that there is at most one instance per servlet context.
     * <p>
     * The instance has the name ``dataSource'' and can be accessed
     * from JSPs by that name.
     *
     * @param ctxt  current servlet context; must not be null
     */
    public static DataSourceBean getApplicationInstance(final ServletContext ctxt)
        {
        DataSourceBean ds;

        boolean created = false;
        synchronized(ctxt)
            {
            ds = (DataSourceBean) ctxt.getAttribute(BEAN_NAME);
            if(ds == null)
                {
                // Create and store a new instance if necessary.
                ds = new DataSourceBean();
                ds.setSlave(ctxt.getInitParameter(CoreConsts.WAR_CTXTPARAM_DISPOSITION));
                ctxt.setAttribute(BEAN_NAME, ds);
                created = true;
                }
            }

        // We take this opportunity to set the servlet context,
        // possibly starting a background thread.
        //
        // We avoid doing this inside the lock on the context itself
        // to avoid any possiblity of deadlock.
        ds.setServletContext(ctxt);

        // When we have newly created the DSB instance
        // do some warm-up asynchronously
        // to try to minimise the delay seen by the first visitor(s).
        if(created)
            {
            System.out.println("[DataSourceBean created.]");

            // We also ensure that a couple of other things are warmed up
            // on a separate thread, partially because they may be I/O bound.
            // We also help take advantage of any extra CPUs available.
            ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    // Get the IP-->location data loaded.
                    try { GeoUtils.getRegionByAddress(InetAddress.getLocalHost(), false); }
                    catch(final UnknownHostException e) { /* Ignore... */ }

                    // Prime the DNS cache/lookup.
                    AddrTools.hasARecordQuick(CoreConsts.MAIN_DATA_HOST);

                    // Warm up and (re)seed the random number generators.
                    Rnd.fastRnd.setSeed(System.nanoTime() + Rnd.goodRnd.nextInt());
                    }
                });
            }

        return(ds);
        }


    /**Public no-arg constructor for ease of use as a JavaBean.
     * This defers as much work as it reasonably can.
     * <p>
     * However, if the system property
     * CoreConsts.WAR_SYSPROPNAME_DISPOSITION_PRESET
     * is set to a usable value, this may allow us to start the
     * cache sooner.
     */
    public DataSourceBean()
        {
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[DataSourceBean: instance created: " + this + ".]"); }
        // Set the master/slave disposition
        // explicitly if the use has tweaked the system properties
        // with an explicit ``master'' or ``slave'' or ``waronly'' value.
        setSlave(System.getProperty(CoreConsts.WAR_SYSPROPNAME_DISPOSITION_PRESET));
if(ORG.hd.d.IsDebug.isDebug && (isSlave() != null)) { logger.log("[DataSourceBean: preset disposition ``"+ isSlave() +"'': " + this + ".]"); }

        // Verify what we've been given.
        try { validateObject(); }
        catch(InvalidObjectException e)
            { throw new IllegalArgumentException(e.getMessage()); }
        }

    /**If true, ensure that index is up-to-date on each poll() call.
     * If true, this minimises delay for users,
     * but can seriously impact start-up time and may perform significant
     * unnecessary work when new items are being loaded into the database.
     * <p>
     * We may try to avoid a forced rebuild if the system is busy
     * or else do each (re)build in a low-priority thread.
     */
    private static final boolean FORCE_INDEX_REBUILD_IN_POLL = true;


    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**Log the given message.
     * Allow external users to use our managed logger.
     */
    public void log(final String message)
        { logger.log(message); }


    /**Poll periodically (of the order of a second) to do background tasks.
     * One job we do is post some properties where they can be found
     * by other system components...
     */
    public final void poll(final GenProps gp)
        throws IOException
        {
        // Last error encountered...
        IOException lastErr = null;

        final long startTime = System.currentTimeMillis();

        // Get the pipeline.
        final SimpleExhibitPipelineIF pipeline = _getPipeline();

        // Set the cache aggressiveness depending on how busy we are.
        if(pipeline instanceof ExhibitDataSimpleCache)
            {
            // Make the cache aggressive iff we are not too busy...
            final boolean overloaded = WebUtils.isOverloaded(_servletContext);
            ((ExhibitDataSimpleCache) pipeline).setAggressive(
                isAggressive() && !overloaded);
            }

        // Do any upstream work...
        try { pipeline.poll(gp); }
        catch(final IOException e) { lastErr = e; }


        // Now do our work...

        // Keep any posted properties up to date.
        try { _postProps(); }
        catch(final IOException e) { lastErr = e; }

        if(FORCE_INDEX_REBUILD_IN_POLL && !byWordIndexIsAvailableAndUpToDate())
            {
            // Make sure that the by-word index is always up to date
            // since it is heavily used (eg in the catalogue pages).
            try { _rebuildJIB(); }
            catch(final IOException e) { lastErr = e; }
            }

        try
            {
            // Notify any Observer(s) registered with us
            // and clear the AEP-linked store
            // when the AEP content hash has changed.
            // This is typically used to help clear caches quickly
            // and we do not promise to update them every time or immediately,
            // ie this mechanism is lossy and only "best efforts".
            //
            // TO AVOID DEADLOCKS, DON'T DO THIS IN THE SCOPE OF OUR LOCKS.
            final EDVHObservable obs = _observable;

            final int evdh = exhibitDataVersionHash();
            obs.setCurrentEDVH(evdh);

            // Iff there has been a change, spin off a thread to deal with it
            // to try to avoid deadlocks.
            //
            // Pass the current hash value to help Observers avoid a redundant
            // cache clear if they have already discarded/rebuilt their cache.
            if(obs.hasChanged())
                {
                final Map<AEPLinkedKey, Object> sl = storeLinked;
logger.log("[DataSourceBean notifying observers of change in exhibit data set, count: " + obs.countObservers() + ", map entries: "+((sl==null)?-1:sl.size())+".]");

                // First clear any AEP-linked store.
                storeLinked = null; // Fast and helps GC!

                // This will be spun off in a separate thread,
                // if none already running.
                obs.notifyObserversInNewThread(new Integer(evdh));
                }
            }
        catch(final IOException e) { lastErr = e; }

        // Chose to do some of the work only if not too busy...
        final ServletContext servletContext = _servletContext;
        final boolean lightlyLoaded = (_servletContext != null) &&
                       WebUtils.isLightlyLoaded(_servletContext);

        // Keep the "best" exhibit set up to date
        // as long as we are not too busy and have not taken too long so far.
        // We don't do this on every round.
        try
            {
            if(lightlyLoaded &&
               ((System.currentTimeMillis() - startTime) < MEAN_POLL_INTERVAL_MS/2) &&
               Rnd.fastRnd.nextBoolean())
                {
                final int toGet = 10 + Rnd.fastRnd.nextInt(10); // Number to force loading of.
                final String exhibits[];
                if(Rnd.fastRnd.nextBoolean())
                    {
                    // Keep "best" exhibit set and thumbnails up to date.
                    // Get the "creme de la creme" as shown on the "best" page,
                    // and sometimes a selection from further down the "best" set,
                    // to represent the sampling done on on the front page.
                    // Also fetch a little more than either of those would
                    // to try to do any just-out-of-sight updates pre-emptively.
                    exhibits = HTMLThumbnailInsertGenerators.getBestExhibitSelection(servletContext,
                        1 + WebConsts.SINGLE_PAGE_CONTACT_SHEET_TN_COUNT,
                        Rnd.fastRnd.nextBoolean() ? null : Rnd.goodRnd); // Help churn good rnd generator a little!
                    }
                else
                    {
                    // Make sure that we have fetched and have fresh
                    // a selection of the very newest exhibit thumbnails,
                    // but not so many as to displace other things from caches.
                    // This should help right after new exhibits are added.
                    exhibits = HTMLThumbnailInsertGenerators.getNewExhibitSelection(servletContext,
                                    toGet,
                                    null);                    
                    }
                
                // Try to keep the selected exhibits' thumbnails "live".
                for(int i = Math.min(toGet, exhibits.length); --i >= 0; )
                    {
                    final String exhibitName = exhibits[i];
                    // Get thumbnail fetched and into cache if possible.
                    if(WebUtils.TN_AVAIL_CACHE &&
                       (pipeline.getThumbnails(exhibitName, true) != null))
                        {
                        // Cache status for UI.
                        WebUtils.exhibitHasThumbnail(this, exhibitName, false, false);
                        }
                    
                    // Stop if we've run out of time...
                    if((System.currentTimeMillis() - startTime) > MEAN_POLL_INTERVAL_MS/2)
                        { break; }
                    }
                }
            }
        catch(final IOException e) { /* IGNORE BENIGN ERRORS */ /* lastErr = e; */ }

        // Ensure that the basic AlohaEarth data is prepared
        // as long as we are not too busy and have not taken too long so far.
        // We don't do this on every round.
        try
            {
            if(lightlyLoaded &&
               ((System.currentTimeMillis() - startTime) < MEAN_POLL_INTERVAL_MS/2) &&
               Rnd.fastRnd.nextBoolean())
                {
                final AlohaEarthMapCache aemfb = AEUtils.getAemfb(this);
                if((servletContext != null) && Rnd.fastRnd.nextBoolean())
                    {
                    // Compute most data for basic/default view.
                    final AEParams aeps = new AEParams();
                    aemfb.selectViewLocationLabels(this, aeps);
                    // Don't load/compute image(s) until actually needed
                    // as they can take up quite a bit of memory
                    // and load time should not be that terrible,
                    // and we won't let the most expensive go again once loaded.
//                    aemfb.selectMapEncodedImage(aeps, servletContext);
                    }
                }
            }
        catch(final IOException e) { lastErr = e; }

        // Incrementally pre-compute parts of very popular catalogue pages.
        try
            {
            if(lightlyLoaded &&
               Rnd.fastRnd.nextBoolean())
                {
                final AllExhibitProperties aep = getAllExhibitProperties(-1);

                // Get previous days' most visited pages...
                final EventVariableValue evv = getEventValue(SystemVariables.ACCESSPATTERN_CAT_PAGE_VIEW, EventPeriod.VLONG, false);
                final int totalDistinctValues = evv.getTotalDistinctValues();

                // Precompute parts of popular pages viewed yesterday,
                // with more popular ones done sooner...
                // Any IOException will stop this round of precomputation.
                if(totalDistinctValues > 0)
                    {
                    // Max pages to systematically precompute parts of.
                    // Limit this effort to the same sort of size as elsewhere...
                    final int pp = Math.min(totalDistinctValues, 11 + 3*WebConsts.SINGLE_PAGE_CONTACT_SHEET_TN_COUNT);

                    // Precompute the search results, most-popular-page first.
                    final Map<String,String> shortToFullMap = aep.aeid.getShortToFullMap();
                    for(int i = 0; (i < pp) && ((System.currentTimeMillis() - startTime) < MEAN_POLL_INTERVAL_MS/2); ++i)
                        {
                        // Don't bother for pages with lowish (noise?) visit counts...
                        if(evv.getCountByRank(i) < 2) { break; }
                        final String exhibitShortName = (String) evv.getValueByRank(i);
                        final String fullName = shortToFullMap.get(exhibitShortName);
                        if(fullName != null)
                            {
                            // Do the search so as to cache the results...
                            final HashSet<String> r = new HashSet<String>(SearchResultSimpleCache.doCachedCatPageSimilarItems(this,
                                fullName,
                                SearchResultSimpleCache.ABS_MAX_RESULTS));
                            // Compute and cache any metadata...
                            WebUtils.getCatPageExhibitMetaDataHTML(this, fullName);
                            // Compute and cache thumbnail availability...
                            // Try to force tn creation for top-visited pages...
                            if(WebUtils.TN_AVAIL_CACHE)
                                { WebUtils.exhibitHasThumbnail(this, fullName, true, true); }
                            // Now ensure that the EPCM is up-to-date for this exhibit
                            // and many/most related exhibits for which we will show ratings.
                            r.add(exhibitShortName);
                            for(final String esn : r)
                                {
                                final String fn = shortToFullMap.get(esn);
                                if(fn == null) { continue; }
                                aep.getExhibitPropsComputableMutable(fn, false, gp, this, scorers);
                                }
                            }
                        }

//                    // Pick one exhibit at random to precompute for...
//                    // This picks more-popular pages more often.
//                    final int n = Rnd.fastRnd.nextInt(Rnd.fastRnd.nextInt(totalDistinctValues)+1);
//                    final String nameS = (String) evv.getValueByRank(n);
//                    final String nameF = shortToFullMap.get(nameS);
//                    if(nameF != null)
//                        {
//                        // Do the search so as to cache the results...
//                        SearchResultSimpleCache.doCachedCatPageSimilarItems(this,
//                            nameF,
//                            SearchResultSimpleCache.ABS_MAX_RESULTS);
//                        // Compute and cache any metadata...
//                        WebUtils.getCatPageExhibitMetaDataHTML(this, nameF);
//                        // Compute and cache thumbnail availability...
//                        if(WebUtils.TN_AVAIL_CACHE)
//                            { WebUtils.exhibitHasThumbnail(this, nameF, true, true); }
//                        }
                    }
                }
             }
        catch(final IOException e) { /* IGNORE BENIGN ERRORS */ /* lastErr = e; */ }
        catch(final Exception e) { e.printStackTrace(); } // Not expected...

        // Keep the AI exhibit Scorers stats up to date (if we are lightly loaded).
        // Because each call may take a very long time to complete,
        // this call uses a low-priority (discardable) background thread to do the work.
        if(lightlyLoaded)
            {
            try { scorers.poll(); }
            catch(final IOException e) { /* IGNORE BENIGN ERRORS */ /* lastErr = e; */ }
            }

        // Prime the DNS cache/lookup and/or keep it live and primed.
        // This may block for a little while, so do it last.
        if(lightlyLoaded) { AddrTools.hasARecordQuick(CoreConsts.MAIN_DATA_HOST); }

        // Rethrow any subordinate error.
        if(lastErr != null) { throw lastErr; }
        }

    /**How often do we repost properties (ms, approx); strictly positive.
     * This should be at least once per minute.
     */
    private final int PROPS_REPOST_INTERVAL_MS = 15001 + Rnd.fastRnd.nextInt(5001);

    /**Note of the last time that we successfully ran raw properties repost without error.
     * Private to _postProps().
     */
    private long _lastGoodPostProps;

    /**Update and post properties to application level as required.
     * Operates under the instance lock for thread safety.
     */
    private synchronized void _postProps()
        throws IOException
        {
        final long now = System.currentTimeMillis();

        // Exit quickly if nothing to do yet.
        if((now - _lastGoodPostProps) < PROPS_REPOST_INTERVAL_MS)
            { return; }

        // Get the servlet context to be able to set app-level attributes.
        final ServletContext ctx = _servletContext;
        if(ctx == null) { return; } // No context set yet; give up.

        // Post the GenSec general security props.
        // Maybe in future it would be worth checking the
        // timestamp for an updated value before reposting.
        ctx.setAttribute("org.hd.d.pg2k.props.gensec", getGenSecProps(-1));

        // Done!  Mark successful completion...
        _lastGoodPostProps = now;
        }


    /**Set true when the servlet is destroyed. */
    private volatile boolean destroyed;

    /**Free up some system resources and make our poller thread go away.
     * This instance is unusable once this methid has been called.
     */
    public final void destroy()
        {
        // Ask our threads to die ASAP.
        destroyed = true;

        try
            {
            // Save Scorer work, if any.
            if(scorers != null)
                {
                try { scorers.destroy(); }
                finally { scorers = null; }
                }
            }
        finally
            {
            // Shut down pipeline.
            final SimpleExhibitPipelineIF p = _dataPipeline;
            try { if(p != null) { p.destroy(); } }
            finally
                {
                // Free up resources.
                _slave = null; // Prevent (re)construction of data pipeline.
    
                _dataPipeline = null; // Let any cache (etc) go...
    
                // Discard any observers; don't do this in the scope of any lock.
                _observable.deleteObservers();
    
                // Discard any data we manage.
                storeLinked = null;
                storeUnlinked = null;
    
                // Discard AI Scorers.
                scorers = null;
    
                logger.setContext(null); // Drop reference to the servlet logger.
    
    if(ORG.hd.d.IsDebug.isDebug) { logger.log("[DataSourceBean: instance destroyed: " + this + ".]"); }
                }
            }
        }

    /**Check if destroyed... */
    private boolean isDestroyed()
        { return(destroyed); }

    /**An Observable that changes when its "currentEDVH" value changes.
     * This currentEDVH is the hash over the exhibit data.
     * <p>
     * When we call notifyObserversInNewThread() we do so where possible with a single
     * Integer argument carrying the current hash value.
     * An Observer may ignore a call when the value
     * matches its current hash in case it has already rebuilt its cache.
     * <p>
     * Note that we build this on top of the default Observable implementation
     * which holds no locks while delivering notifications,
     * which should help us avoid deadlocks.
     */
    private static final class EDVHObservable extends Observable
        {
        /**The current hash of the exhibits: when this changes then the whole Observable hasChanged(). */
        private int currentEVDH;
        /**Set the currentEVDH; when changed hasChanged() returns true. */
        synchronized final void setCurrentEDVH(final int currentEVDH)
            {
            if(this.currentEVDH != currentEVDH)
                {
                setChanged();
                this.currentEVDH = currentEVDH;
                }
            }

        /**Thread, if any, being used to deliver notifications.
         * Private to notifyObserversInNewThread() and accessed under the instance lock.
         */
        private Thread deliveryThread;

        /**Spins off background thread to call super.notifyObserversInNewThread().
         * Will not attempt to start a new thread to do this
         * if the last one appears not to have died,
         * so will in that case just not send any notifications.
         * <p>
         * No lock is held while notifications are delivered;
         * the instance lock may be briefly held while the thread
         * field is being examined and a thread launched.
         *
         * @param arg  any object
         */
        final synchronized void notifyObserversInNewThread(final Object arg)
            {
            // If last thread still running, do not attempt to start another.
            if((deliveryThread != null) &&
                deliveryThread.isAlive())
                {
System.err.println("DataSourceBean: previous notification thread still running, new one aborted.");
                return;
                }

            final Observable obs = this;
            final Thread notifierThread =
                (new Thread("DataSourceBean: background Observer notifier thread")
                    { public final void run() { obs.notifyObservers(arg); } }
                );
            deliveryThread = notifierThread;
            notifierThread.setDaemon(true);
            notifierThread.start(); // Ensures thread isAlive().
            }
        }

    /**Can be observed by an Observer that wants to know when the underlying data has changed; never null.
     * Observers are notified when our hash changes,
     * for example so that they can clear any now-stale cache and help GC.
     * <p>
     * Volatile so that access to the reference itself does not need a lock.
     * <p>
     * May be cleared once a notification has been delivered during a poll(),
     * ie may be at most one-shot.
     * <p>
     * Missed notifications are not considered very important.
     */
    private volatile transient EDVHObservable _observable = new EDVHObservable();

    /**Register Observer to be told when exhibit data changes.
     * This will probably be poll driven.
     * <p>
     * We may discard Observers at will, and will certainly do so
     * if serialised and deserialised.
     */
    public void addObserver(final Observer obs)
//        throws IOException
        {
        _observable.addObserver(obs);

if(ORG.hd.d.IsDebug.isDebug) { logger.log("[DataSourceBean observers: " + _observable.countObservers() + ".]"); }
        }

    /**Remove an Observer.
     * An Observer may have no further need to keep observing,
     * eg because its job is done and it wants to be GCed.
     */
    public void deleteObserver(final Observer o)
        {
        _observable.deleteObserver(o);

if(ORG.hd.d.IsDebug.isDebug) { logger.log("[DataSourceBean observers: " + _observable.countObservers() + ".]"); }
        }



    /**Indicates if this is a master or a slave Web server, or not known.
     * "Not known" is indicated by a null value.
     * <p>
     * This value may be persisted.
     * <p>
     * Volatile to be accessed without a lock by isSlave() and setSlave().
     * <p>
     * Must be set non-null to initialise the bean for use.
     * <p>
     * May be able to be set immediately from the context.
     */
    private volatile Boolean _slave;

    /**Get status of this server; master, slave or unknown.
     * Once initialised for use the status is known and
     * is either TRUE or FALSE; until then it is null indicating unknown.
     * Once set non-null its value cannot be changed.
     */
    public Boolean isSlave()
        { return(_slave); }

//    /**Mirrors _slave and provides extra information.
//     * In the case that the disposition has been set with a string,
//     * that is stored here, and in particular if set with the
//     * ``waronly'' string, that can be retrieve from here.
//     */
//    private String _dispositionToken;

    /**Set disposition.
     * If the parameter is true, the disposition is set to slave mode.
     * If the parameter is false, the disposition is set to master mode.
     * <p>
     * If value is passed that conflicts with a non-null value
     * already set, an IllegalStateException is thrown, ie the disposition
     * cannot be changed once set.
     */
    public void setSlave(final boolean isASlave)
        throws IllegalStateException
        {
        final Boolean currentState = _slave;

        // Do not allow changes in disposition.
        if((currentState != null) && (currentState.booleanValue() != isASlave))
            { throw new IllegalStateException("master/slave disposition conflict"); }

        // Set disposition.
        _slave = (isASlave ? Boolean.TRUE : Boolean.FALSE);
        }



    /**Set slave from context or property value.
     * The value passed is ignored if null,
     * and is not case-insensitively-equal to ``master'' or ``slave'' or ``waronly'',
     * else it is set to slave or master mode as appropriate.
     */
    public synchronized void setSlave(String masterSlave)
        throws IllegalStateException
        {
        if(masterSlave == null) { return; }

        masterSlave = masterSlave.toLowerCase();

        // Only look for the exact words, albeit case-insensitive.
        if(masterSlave.equals(CoreConsts.WAR_CTXTPARAM_DISPOSITION_MASTER))
            { setSlave(false); /* _dispositionToken = masterSlave; */ }
        else if(masterSlave.equals(CoreConsts.WAR_CTXTPARAM_DISPOSITION_SLAVE))
            { setSlave(true);  /* _dispositionToken = masterSlave; */ }
        else if(masterSlave.equals(CoreConsts.WAR_CTXTPARAM_DISPOSITION_WARONLY))
            { setSlave(false); /* _dispositionToken = masterSlave; */ }
        }

    /**We examine the context path to see if we can tell if we are slave or master.
     * We are definitely a master server if our context path is not the empty string.
     * All slaves definitely have an empty-string context path, though a
     * ``waronly'' master does too.
     * <p>
     * This must be set before calling any of the get...() methods of
     * the bean (since we may need this to set up our cache, for example)
     * and is idempotent so long as we call it with the same value.
     * If we call it with a different value the call may be aborted.
     */
    public void setContextPath(final String cPath)
        throws IllegalStateException
        {
        if(cPath.length() != 0) { setSlave(false); }
        }


    /**A valid servlet context, or null.
     * We don't persist the value if the bean is serialised.
     * <p>
     * Declared volatile to allow safe lock-free access.
     * <p>
     * Should only be accessed by setServletContext() for write and
     * generally by _getPipeline() and _postProps() for read.
     */
    private transient volatile ServletContext _servletContext;

    /**The aggressiveness flag; defaulting to false.
     * Is set when we set the context; we do not bother to persist it.
     * <p>
     * Is volatile so that no lock is needed to access it.
     * <p>
     * Private to setServletContext() and isAggressive().
     */
    private transient volatile boolean aggressive;

    /**Check if we are going to be aggressive in cacheing.
     * Defaults to false; is only valid once setServletContext() has been called.
     */
    public boolean isAggressive()
        { return(aggressive); }

    /**Set a valid servlet context.
     * We don't really care which it is so long as is valid
     * (eg not null).
     * <p>
     * We don't persist the value if the bean is serialised.
     * <p>
     * Note that if we are going to be aggressive <em>and</em>
     * we already know if we are master or slave then we automatically
     * start the data pipeline and thread if necessary, but not inside
     * the instance lock.
     * <p>
     * If the context path is set before the servlet context
     * and we are in aggressive mode then that will force an
     * immediate pipeline start.
     */
    public void setServletContext(final ServletContext context)
        {
        if(context == null)
            { throw new IllegalArgumentException("invalid null servlet context"); }

        logger.setContext(context);

        synchronized(this)
            {
            // Only have the side-effects purely to do with setting the
            // context if it has changed.
            // We save a bit of redundant parsing, etc, this way.
            if(_servletContext != context)
                {
                _servletContext = context;

                // Find out how aggressive we should be...
                final String aggressiveS =
                    context.getInitParameter(CoreConsts.WAR_CTXTPARAM_AGGRESSIVE_CACHE);
                if(aggressiveS != null)
                    {
                    aggressive = Boolean.valueOf(aggressiveS).booleanValue();
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[DataSourceBean: isAggressive()=="+aggressive+" ("+aggressiveS+") for: " + this + ".]"); }
                    }
                else // Default to not aggressive.
                    { aggressive = false; }
                }
            }

        if(isAggressive() && (isSlave() != null))
            {
            // Force pipeline start.
            _getPipeline();
            }
        }


    /**Private lock to ensure only one thread tries to build _dataPipeline. */
    private static final Object _dataPipeline_build_lock = new Object();

    /**Non-persistable data pipeline.
     * This is reconstructed under the pipeline lock
     * if found to be null and if slave is non-null.
     * <p>
     * Should be accessed only by _getPipeline(),
     * and (without a lock) nulled out in destroy().
     */
    private transient volatile SimpleExhibitPipelineIF _dataPipeline;


    /**Approximate mean interval between poll() calls (ms); strictly positive. */
    private static final int MEAN_POLL_INTERVAL_MS = 900 + Rnd.fastRnd.nextInt(200);

    /**Start background thread...
     */
    private static void _startBackgroundThread(final DataSourceBean dsb)
        {
        // Now that we have just created a cache,
        // we'll create a thread to manage it,
        // but not under our instance lock.
        // We do this to avoid any deadlock on the object,
        // and we assume that this cannot cause any significant race hazard.
        //
        // The beanInstance is the only reference the thread
        // should have back to us...
        final SoftReference<DataSourceBean> beanInstance = new SoftReference<DataSourceBean>(dsb);

        // Start background thread after pipeline is set up...
        final Thread th = (new Thread("DataSourceBean: poll() thread"){
            /**Approx average interval between calls to doBG() in ms; maximum interval no more than twice this.
             * Of the order of a second to allow processing of resource-usage monitoring,
             * which requires the fastest processing.
             * <p>
             * Different per instance to make collisions, etc, less likely.
             */
            private final int basicBgInterval = 1 + (MEAN_POLL_INTERVAL_MS/2) + Rnd.fastRnd.nextInt(MEAN_POLL_INTERVAL_MS);

            /**We'll back off to an upper limit if we encounter errors in polling.
             * We do this mainly to avoid flooding logs if things are badly broken,
             * but we usually expect to stay at or close to the minimum bg interval.
             * <p>
             * This is short enough not to completely wreck
             * performance/interaction, or (eg) break the variable pipeline.
             */
            private final int maxBgInterval = 15017 + Rnd.fastRnd.nextInt(1001 + basicBgInterval);

            /**Call poll(). */
            @Override public final void run()
                {
dsb.logger.log("[Started cache thread: " + dsb + ".]");

                try {
                    // Private copy of generic properties.
                    // Start with default/empty set of properties...
                    GenProps genProps = new GenProps();

                    // Last GenProps poll time.
                    long _lGpPoll = 0;

                    // Start polling slowly since beans (etc) may not even
                    // be deployed yet and the system in general may be `warming up'.
                    // This will probably cause some uglyness in the data pipeline
                    // to start with, but a single successful poll cycle will
                    // restore us to maximum clip anyway...
                    int basicSleepInterval = (2 * basicBgInterval) + Rnd.fastRnd.nextInt(5011 + basicBgInterval);

                    for( ; ; )
                        {
                        try {
                            // Interval in ms to sleep before next doBg(); different each time.
                            final int bgInterval = 1 + (basicSleepInterval/2 + Rnd.fastRnd.nextInt( 1 | basicSleepInterval/2));

                            // Default to lengthening the sleep interval
                            // (assuming an error is going to be encountered)
                            // but reset it if all is well post hoc.
                            basicSleepInterval = Math.min(basicSleepInterval * 2, maxBgInterval);

                            // Now sleep...
                            Thread.sleep(bgInterval);

                            final DataSourceBean ds = beanInstance.get();

                            // If the bean has been GCed, quit the thread.
                            if(ds == null) { return; }

                            // If the bean has had destroy() called, quit the thread.
                            if(ds.isDestroyed()) { return; }

                            // If not yet initialised, just sleep until ready...
                            if(ds.isSlave() == null)
                                { continue; }

                            final long start = System.currentTimeMillis();

                            try {
                                // We get the new system properties, using the
                                // specified recheck interval (the default being quite short).
                                if((start - _lGpPoll > genProps.getWEBSVR_SYSPROPS_RECHECK_MS()) ||
                                   (start < _lGpPoll)) // Clock wobble?
                                    {
                                    final GenProps newGp = ds.getGenProps(genProps.timestamp);
                                    if(newGp != null) { genProps = newGp; }
                                    _lGpPoll = start;
                                    }

                                // Now do normal background processing...
                                ds.poll(genProps);
                                }

                            // Silently discard `master not in service' errors.
                            // We don't want these to lengthen our sleep time.
                            catch(final PGMasterNotInServiceException e) { }

                            // If we get here without error,
                            // we can reset polling to the maximum rate.
                            basicSleepInterval = basicBgInterval;
                            }
                        catch(IOException e)
                            {
                            // Quietly ignore IOExceptions as we expect some in the
                            // normal course of business.  They may be very regular and
                            // we don't want to bloat any log files...
//System.err.println("In bg thread run()...");
//e.printStackTrace();
                            continue; // Allow sleep time to increase...
                            }
                        catch(Throwable t)
                            {
                            // Report unusual errors...
dsb.logger.log("Unexpected exception in bg thread run(): " + dsb);
t.printStackTrace();
                            continue; // Allow sleep time to increase...
                            }
                        }
                    }
                finally
                    {
dsb.logger.log("[Quit cache thread: " + dsb + ".]");
                    }
                }
            });
        th.setDaemon(true);
        th.start();
        }


    /**Get data pipeline; never null.
     * Will return any extant pipeline, or try to create one
     * providing that the bean has been initialised
     * and knows if this is a master (getting its data directly from EJB)
     * or a slave (getting data from the master server).
     * <p>
     * Operates under the instance lock except to start any background thread.
     * <p>
     * Should be small enough to inline, statically or with JIT compiler.
     *
     * @throws IllegalStateException  if the cache cannot be created
     *     due to incorrect initialisation
     */
    private SimpleExhibitPipelineIF _getPipeline()
        throws IllegalStateException
        {
        // If the pipeline is not null then we can use it as-is.
        final SimpleExhibitPipelineIF dp = _dataPipeline;
        if(dp != null) { return(dp); }

        // Pipeline needs to be created.
        // Should be called (about) once during the life of any DSB instance.
        return(_createPipeline());
        }

    /**Do the heavy-lifting of creating the data pipeline; never null.
     * Should be called (about) once during the life of any DSB instance.
     * @return  the non-null pipeline
     * @throws IllegalStateException  if the cache cannot be created
     *     due to incorrect initialisation
     */
    private SimpleExhibitPipelineIF _createPipeline()
        throws IllegalStateException
        {
        boolean createdCache = false;
        try {
            // Build the cache under the pipeline build lock
            // to eliminate any possibility of doing it multiple times.
            // It is vital that we don't build multiple cache instances.
            synchronized(_dataPipeline_build_lock)
                {
                // Return any extant pipeline immediately.
                if(_dataPipeline != null) { return(_dataPipeline); }

                // If destroyed, do not create a new pipeline!
                if(destroyed) { throw new IllegalStateException("Cannot create pipeline: shutting down"); }

                // Pipeline needs creating but we don't know where
                // to find its data: abort the call.
                final Boolean iS = isSlave();
                if(iS == null)
                    { throw new IllegalStateException("bean not initialised for master/slave"); }

                // We need a valid servlet context; abort if we don't have one.
                final ServletContext sCtxt = _servletContext;
                if(sCtxt == null)
                    { throw new IllegalStateException("DataSourceBean needs servlet context"); }

                // Find the temporary dir we can use for a file-based cache.
                final File cacheDir =
                    (File) (sCtxt.getAttribute("javax.servlet.context.tempdir"));

                // Select our data source (HTTP tunnel or EJB).
                final SimpleExhibitPipelineIF dataSource = iS.booleanValue() ?
                    // This is a slave, so get our data over HTTP...
                    TunnelServlet.createFromContext(sCtxt) :
                    // This is a `master'...
                    ((SimpleExhibitPipelineIF) new ExhibitDataFileSource());

                // Wrap a cache round the data source.
                final ExhibitDataSimpleCache cache =
                    ExhibitDataSimpleCache.cacheFactory(dataSource, cacheDir, logger);

                // Set the cache aggressiveness flag.
                cache.setAggressive(isAggressive());

                // Now save the new pipeline.
                _dataPipeline = cache;

                // We created a new cache
                // so start a new poll() thread in a moment...
                createdCache = true;
                return(cache);
                }
            }
        catch(final IOException e)
            {
            throw new IllegalStateException("could not create pipeline due to IOException: " + e.getMessage());
            }
        finally
            {
            if(createdCache) { _startBackgroundThread(this); }
            }
        }


    /**Get the static attributes for a given exhibit.
     * Returns null if the named exhibit does not exist.
     */
    public ExhibitStaticAttr getStaticAttr(final String name)
        throws IOException
        {
        return(_getPipeline().getStaticAttr(name));
        }

    /**Get a chunk of the raw exhibit binary.
     * The start index and the index after the last required
     * byte is supplied.  The start value must be non-negative
     * and the afterEnd value no smaller than start and no larger
     * than the exhibit data length.
     */
    public void getRawFile(final ByteBuffer buf,
                           final String exhibitName, final int position, final boolean dontCache)
        throws IOException
        {
        _getPipeline().getRawFile(buf, exhibitName, position, dontCache);
        }

    /**Gets all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     */
    public AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
        throws IOException
        {
        return(_getPipeline().getAllExhibitImmutableData(oldStamp));
        }

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned) null is returned.
     */
    public AllExhibitProperties getAllExhibitProperties(final long oldHash)
        throws IOException
        {
        return(_getPipeline().getAllExhibitProperties(oldHash));
        }


    /**Computes a hash on the versions of immutable and mutable exhibit data currently held.
     * This value is computed on the timestamps/hashes of the exhibit data
     * and properties available through this bean.  If this hash changes
     * then some of the underlying exhibit data has changed
     * and users of the bean should recompute values based on this bean.
     */
    public final int exhibitDataVersionHash()
        throws IOException
        {
        // This is derived from all the exhibit properties.
        final AllExhibitProperties aep = getAllExhibitProperties(-1);
        return(((int) (aep.longHash >>> 32)) + ((int) aep.longHash));
        }

    /**Encoding we use to convert name or name+description to byte stream for indexing.
     * We can assume 7-bit ASCII or 8-bit ISO-8859-1.
     */
    private static final String NAMEDESC_ENCODING = CoreConsts.FILE_ENCODING_8859_1;

    /**Gets the by-word-index, possibly recreating it if not cached or if out of date; may return null.
     * This is created from the full exhibit names and any extant description.
     * <p>
     * We know all names to be 7-bit (printable) ASCII,
     * and assume all descriptions to be either 7-bit ASCII or
     * at most 8-bit ISO-8859-1 (Latin-1) text,
     * and we use that when converting to a byte stream for indexing.
     * <p>
     * This may not force a synchronous rebuild since this may take too long,
     * and therefore may return a stale or null index
     * while a new one is being computed.
     */
    private JIndexBean _getJIB()
        throws IOException
        {
        // If the index is not getting rebuilt asynchronously in the background
        // (or if it is currently null/stale)
        // then start a (single-threaded) asynchronous rebuild here if needed.
        if(!FORCE_INDEX_REBUILD_IN_POLL || !byWordIndexIsAvailableAndUpToDate())
            { _rebuildJIB(); }

        // Asynchronously return whatever index we have to hand, or null if none.
        return(_getJIB_cache);
        }

    /**Returns true if there is a by-word index available.
     * Returns false until one is built or loaded.
     * <p>
     * If this returns true it does not guarantee that the index is up-to-date,
     * simply that one exists and can be used.
     */
    public boolean byWordIndexIsAvailable()
        { return(_getJIB_cache != null); }

    /**Returns true if there is a by-word index available and it is up-to-date.
     * Returns false until one is (re)built or loaded.
     */
    public boolean byWordIndexIsAvailableAndUpToDate()
        {
        final JIndexBean jib = _getJIB_cache;
        if(jib == null) { return(false); /* Faster exit at (busy?) startup! */ }
        try
            {
            final int eDVH = exhibitDataVersionHash();
            return(jib.getRefValue() == eDVH);
            }
        catch(final IOException e)
            {
            return(false); /* Treat failure as implying busy... */
            }
        }

    /**Private lock for _rebuildJIB.
     * Attempts by another thread to enter the lock are vetoed immediately,
     * preventing other threads from blocking pointlessly.
     */
    private final ReentrantLock _rebuildJIB_lock = new ReentrantLock();

    /**If true then attempt to cache index in a file for potentially-faster system restart. */
    private static final boolean CACHE_INDEX_AS_FILE = false;

    /**Name under which we attempt to cache the by-word index on disc (relative to servlet temp directory); never null.
     * The format is implementation-dependent,
     * and could be a serialised object or a saved index file.
     */
    private static final String CACHED_BYWORD_INDEX_FNAME = "_cached_byWord_Index.ser.gz";

    /**Rebuild the JIB (index bean) if necessary.
     * Thread safe, and synchronized on a private lock
     * in order to avoid wasting CPU cycles with multiple update attempts.
     * <p>
     * Directly updates the cache used by _getJIB()
     * if the index needs recomputing,
     * ie if it is null or out of date wrt the AEP.
     * <p>
     * If a second thread tries to enter while one is already rebuilding
     * the index, then the second thread returns immediately (does not block),
     * but meaning that the cache will not be guaranteed rebuilt
     * by the time that the second thread returns,
     * so any caller has to be prepared to deal with the cache still null
     * when this routine returns.
     *
     * @throws IOException
     */
    private void _rebuildJIB()
        throws IOException
        {
        // Return immediately if index exists and is up-to-date.
        final JIndexBean jib = _getJIB_cache;
        final int eDVH = exhibitDataVersionHash();
        if((jib != null) && (jib.getRefValue() == eDVH))
            { return; }

        // Return immediately (thus avoiding starting a(nother) useless thread)
        // if a rebuild appears already to be in progress.
        if(_rebuildJIB_lock.isLocked())
            { return; }

        // Measure start time so as to capture thread start-up overhead.
        final long startTime = System.currentTimeMillis();

        // The index needs (re)building,
        // so spin off a background thread to do it,
        // but have any such thread abort quickly
        // if one is already working on the task.
        // This wastes a small amount of effort
        // spinning off and killing some redundant background threads.
        final AllExhibitProperties aep = getAllExhibitProperties(-1);
        // We refuse to construct an index from an empty AEP.
        if(aep.aeid.length == 0) { return; }
        // We use completion of this maximally-low-priority thread
        // as one measure of system load.
        final Thread th = new Thread("index (re)build in background: started at " + (new Date(startTime))){
            /**Rebuild the index if still necessary. */
            public void run()
                {
                // Return immediately if index already being (re)built.
                if(!_rebuildJIB_lock.tryLock()) { return; }
                try
                    {
                    final JIndexBean jib2 = _getJIB_cache;
                    // If index now up-to-date then exit quickly...
                    if((jib2 != null) && (jib2.getRefValue() == eDVH))
                        { return; }

                    // If we have a servlet context
                    // and we have no index in memory at all
                    // for example because we have just started up,
                    // then try to reload a cached index first
                    // even though it may be somewhat stale,
                    // and then (re)compute an up-to-date index.
                    final ServletContext sCtxt = _servletContext;
                    final File cacheDir = (sCtxt == null) ? null : (File) (sCtxt.getAttribute("javax.servlet.context.tempdir"));
                    final File cacheFile = (cacheDir == null) ? null : new File(cacheDir, CACHED_BYWORD_INDEX_FNAME);
// TODO
//                    if((jib2 == null) && (cacheFile != null) && cacheFile.canRead())
//                        {
//logger.log("[Reloading cached by-word index...]");
//
//                        // Don't let an error during reload de-rail us...
//                        try
//                            {
//                            final JIndexBean jib4 = _getJIB_cache =
//                                new JIndexBean();
//
//logger.log("[Reload of by-word index complete: " + ((System.currentTimeMillis() - startTime + 500) / 1000) + "s.]");
//
//                            // If index in fact up-to-date then exit quickly...
//                            if((jib4 != null) && (jib4.getRefValue() == eDVH))
//                                { return; }
//                            }
//                        catch(final Exception e)
//                            {
//                            e.printStackTrace(); /* Absorb any error. */
//                            }
//                        }

                    // Either could not load cached index or it is stale...

logger.log("[Building by-word index...]");

                    final JIndexBean jib3 = computeByWordIndex(aep);
                    jib3.setRefValue(eDVH); // Store up-to-date hash as the ref value of the index.
                    _getJIB_cache = jib3; // Store ref to the the new index atomically.

                    final long endTime = System.currentTimeMillis();
logger.log("[Build of by-word index complete: " + ((endTime - startTime + 500) / 1000) + "s.]");

                    // Attempt to cache our newly-computed index if possible.
                    // Don't attempt to save an empty index...
                    if(CACHE_INDEX_AS_FILE && (cacheFile != null) && (aep.aeid.length > 0))
                        {
//                        try
                            {
                            FileTools.serialiseToFile(jib3, cacheFile, true, true);
logger.log("[Cached by-word index: " + ((System.currentTimeMillis() - endTime + 500) / 1000) + "s.]");
                            }
//                        catch(final InvertedIndexException e)
//                            { e.printStackTrace();  /* Absorb any error. */ }
                        }
                    }
                catch(final InterruptedException e) { e.printStackTrace(); /* Retry later. */ }
                catch(final IOException e) { e.printStackTrace(); /* Retry later. */ }
                finally { _rebuildJIB_lock.unlock(); }
                }
            };

        // Run the index (re)build in a low(er)-priority daemon thread.
        th.setPriority(Math.max(Thread.currentThread().getPriority() - 1, Thread.MIN_PRIORITY));
        th.setDaemon(true);
        th.start();
        }

    /**Computes a (compact) by-word index of the current exhibits; never null.
     * This is based on the exhibit names, descriptions,
     * and any "AKA" (Also-Known-As) keyword data.
     *
     * @return compacted by-word index
     *
     * @throws IOException
     */
    public static JIndexBean computeByWordIndex(final AllExhibitProperties aep)
        throws IOException, InterruptedException
        {
        if(aep == null) { throw new IllegalArgumentException(); }

        // In the trival case of no exhibits, return a new empty index.
        if(aep.aeid.length == 0) { return(new JIndexBean()); }

        // Queue of (shortname,indexablePairs) values.
        // We can pick an implementation to allow a reasonable queue of work,
        // but without excessive memory consumption for queued data.
        // This is the main communication between our worker threads.
        final BlockingQueue<Tuple.Pair<String,String>> workQueue =
            new ArrayBlockingQueue<Tuple.Pair<String, String>>(Math.min(1024, aep.aeid.length));

        // Exception to be rethrown from the text-generation thread, if any.
        final AtomicReference<InterruptedException> bgException = new AtomicReference<InterruptedException>();

        // In one thread generate the text to be indexed.
        // We don't launch this in one of the shared thread pools
        // because we need to guarantee that it always gets launched immediately
        // and in a separate thread so as to avoid deadlock.
        final Thread t1 = new Thread("by-word-index text-generation thread"){
            @Override public final void run()
                {
                // Get a default-locale LocaleBean
                // to extract the descriptive/AKA text via.
                final LocaleBeanBase lb = new LocaleBean();

                // Push the data in, indexed by short name.
                // The text data always contains the virtual or short name,
                // with some additional descriptive text where available.
                // We use the virtual or short name since it omits the "noise"
                // of any intermediate directory names.
                //
                // We work through the names in sorted (short name) order
                // in case this helps performance and the resulting data structures
                // through locality.
                final List<String> shortNames = new ArrayList<String>(aep.aeid.getAllExhibitShortNamesSet());
                Collections.sort(shortNames, String.CASE_INSENSITIVE_ORDER);
                final StringBuilder scratch = new StringBuilder(2048); // Used (and reused!) to build text to be indexed.
                try
                    {
                    final Map<String, String> shortToFullMap = aep.aeid.getShortToFullMap();
                    for(final String shortName : shortNames)
                        {
                        final String exhibitName = shortToFullMap.get(shortName);
                        if(exhibitName == null) { continue; }
                        String mType = ""; // MIME type/name, "eg video/mpeg".
                        try { mType = ExhibitMIME.getMIMEType(exhibitName); }
                        catch(final IOException e) { e.printStackTrace(); /* Shouldn't happen; whinge but ignore if it does. */ }
                        final ExhibitPropsLoadable epl = aep.getExhibitPropsLoadable(exhibitName);
                        final String description = epl.getDescription();
                        final String akaText = GenUtils.getLocalisedTreeDesc(
                            aep, exhibitName, lb, false, true, false, false); // Bare AKA text only.
                        // Index the short name, MIME type,
                        // and whichever of the AKA and descriptions are present.
                        // We'll auto-detect before indexing if these might contain HTML.
                        scratch.setLength(0);
                        scratch.append(shortName);
                        scratch.append(' ').append(mType); 
                        if(akaText != null)
                            { scratch.append(' ').append(akaText); }
                        if(description != null)
                            { scratch.append(' ').append(description); }
                        workQueue.put(new Tuple.Pair<String, String>(shortName, scratch.toString()));
                        }
                    }
                catch(final InterruptedException e)
                    {
                    // Note any explicit interruption and quit immediately...
                    bgException.set(e);
                    return;
                    }
                }
            };
        t1.setDaemon(true);
        t1.start(); // Start generating the text to be indexed.

        // This will be our new index.
        final JIndexBean jib = new JIndexBean();
        jib.setIndexFragments(false); // No fragments.

        try
            {
            // Keep running until the text-generation thread is dead
            // and the work queue is empty.
            while(!workQueue.isEmpty() || t1.isAlive())
                {
                // Get the next doc, waiting a short while for the next entry.
                final Tuple.Pair<String,String> doc = workQueue.poll(100, TimeUnit.MILLISECONDS);
                if(doc == null) { continue; } // Try again...

                final String shortName = doc.first;
                final String text = doc.second;

                // Index as HTML if we find a '<' or '&' HTML meta character.
                final boolean asHTML =
                    (text.indexOf('<') != -1) ||
                    (text.indexOf('&') != -1);

                jib.setDocumentSimple(asHTML,
                    shortName,
                    new ByteArrayInputStream(text.getBytes(NAMEDESC_ENCODING)));
                }
            }
        catch(final InvertedIndexException e)
            {
            final IOException err = new IOException("index-generation error: " + e.getMessage());
            err.initCause(e);
            throw err;
            }

        // If the text-generation thread was interrupted
        // then (re)throw the exception now
        // rather than returning a partial index.
        final InterruptedException ie = bgException.get();
        if(ie != null) { throw ie; }

        // Now tidy up and save some memory in a background thread,
        // since compact() allows other operations to be interleaved with its own.
        // This is CPU-intensive, but not very memory-intensive.
        // If the system is potentially short of CPU, this will run synchronously,
        // ie this thread will block until compaction is complete.
        ThreadUtils.lowPriorityThreadPool.submit(new Runnable(){
            public final void run() { jib.compact(); }
            });

        // We return the new index for use,
        // though compaction may be continuing safely in the background.
        return(jib);
        }

    /**Cache for, and private to, _getJIB(); initially null.
     * We do not attempt to persist this data, so it is transient.
     * <p>
     * Is volatile so that access without a lock is safe.
     */
    private volatile transient JIndexBean _getJIB_cache;

    /**Canonicalises a simple by-word query string and sanitises it; useful for potentially-unsafe user input.
     * This not only potentially makes the query string small(er),
     * it also shows the query that is actually being performed,
     * whatever the user thinks that they are doing!
     * <p>
     * This also converts the query string to pure ASCII,
     * having first converted any non-ISO-Latin-1 characters
     * to spaces, trimmed any obvious whitespace, and truncated
     * to the given length (if non-negative).
     * <p>
     * This can be directly handed the result of an HTTP request
     * parameter, for example.  The output is ideal to hand to
     * findExhibitsByWord().
     * <p>
     * This returns "" if the input is null.
     * <p>
     * If the input string does not need transforming then it is returned
     * intact.
     */
    public static String canonicaliseSimpleByWordQuery(final String s, final int maxLength)
        {
        return(JIndexBean.canonicaliseSimpleByWordQuery(s, maxLength));
        }

    /**Query type or findExhibitsByWord() to return entries matching any word. */
    public static final int FEBY_MATCH_TYPE_ANY = -1;
    /**Query type or findExhibitsByWord() to return entries matching most words. */
    public static final int FEBY_MATCH_TYPE_MOST = 0;
    /**Query type or findExhibitsByWord() to return entries matching all words. */
    public static final int FEBY_MATCH_TYPE_ALL = 1;

    /**Perform simple query on the exhibit data without filtering.
     * The query should be a simple whitespace-separated string
     * of query words.
     * <p>
     * This returns a ranked list of (short) exhibit names;
     * possibly zero-length if no matches (but never null).
     * <p>
     * The underlying index is (re)built on first use or
     * when the underlying data has changed.
     * <p>
     * This may take longer than we'd like and use more memory than we'd like.
     *
     * @param query  a plain-text query of search words, space-separated
     * @param matchType  the number of input words to match;
     *          -1 means any word,
     *           0 means most words,
     *           1 means all words.
     * @param maxResults  is the maximum number of results returned; should
     *          be at least twice the maximum shown to reduce the risk
     *          of having abandoned good results prematurely
     */
    public String[] findExhibitsByWord(final String query,
                                       final int matchType,
                                       final int maxResults)
        throws IOException
        {
        return(findExhibitsByWord(query, matchType, maxResults, null));
        }

    /**Perform simple query on the exhibit data with candidate filtering; never null.
     * The query should be a simple whitespace-separated string
     * of query words.
     * <p>
     * This returns a ranked list of (short) exhibit names;
     * possibly zero-length if no matches (but never null).
     * <p>
     * The underlying index is (re)built on first use or
     * when the underlying data has changed.
     * <p>
     * This may take longer than we'd like and use more memory
     * than we'd like.
     *
     * @param query  a plain-text query of search words, space-separated
     * @param matchType  the number of input words to match;
     *          -1 means any word,
     *           0 means most words,
     *           1 means all words.
     * @param maxResults  is the maximum number of results returned; should
     *          be at least twice the maximum shown to reduce the risk
     *          of having abandoned good results prematurely
     * @param docFilter  if non-null, any documents for which the
     *     accept method returns false are excluded from the search
     */
    public String[] findExhibitsByWord(final String query,
                                       final int matchType,
                                       final int maxResults,
                                       final JIndexBean.SearchFilterByName docFilter)
        throws IOException
        {
        if(LOG_QUERIES)
            { logger.log("QUERY" + ((docFilter == null) ? "" : " (filtered)") + ": " + query); }

        final JIndexBean jIndexBean = _getJIB();

        // If there is no index available yet
        // then return an empty answer (not null).
        if(jIndexBean == null) { return(NO_RESULTS); }

        // Do the lookup in the index.
        try { return(jIndexBean.simpleSearchToDocNames(query, matchType, maxResults, docFilter)); }
        // Handle unexpected errors/exceptions gracefully, and log them.
        catch(final Exception e)
            {
            e.printStackTrace();
            return(NO_RESULTS);
            }
        }

    /**Empty results list, eg after an error. */
    private static final String[] NO_RESULTS = new String[0];

    /**If true, log queries made on the system for analysis. */
    public static final boolean LOG_QUERIES = false;

    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available
     * then a default set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) then null is returned.
     */
    public org.hd.d.pg2k.svrCore.props.GenProps getGenProps(final long oldStamp)
        throws IOException
        {
        return(_getPipeline().getGenProps(oldStamp));
        }

    /**Gets the security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) then null is returned.
     */
    public java.util.Properties getGenSecProps(final long oldStamp)
        throws IOException
        {
        return(_getPipeline().getGenSecProps(oldStamp));
        }

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     * <p>
     * As a kindness to callers, we convert any IOException thrown
     * to returning a null instead, to fold the two cases together.
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible, else only return an existing one
     */
    public ExhibitThumbnails getThumbnails(final String name, final boolean create)
        {
        try { return(_getPipeline().getThumbnails(name, create)); }
        catch(IOException e)
            {
            return(null); // Just one failure case to deal with.
            }
        }

    public void setVariable(final SimpleVariableValue newValue)
        throws IOException
        {
        // Delegate.
        _getPipeline().setVariable(newValue);
        }

    public int setVariables(final SimpleVariableValue[] newValues)
        throws IOException
        {
        // Delegate.
        return(_getPipeline().setVariables(newValues));
        }

    public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
        throws IOException
        {
        // Delegate.
        return(_getPipeline().getVariable(var));
        }

    public SimpleVariableValue[] getVariables(final long changedSince)
        throws IOException
        {
        // Delegate.
        return(_getPipeline().getVariables(-1));
        }

    /**Get the current partial, or previous full, event set at the specified interval; never null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; may be empty but not null if requested set not available
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        return(_getPipeline().getEventValue(def, intervalSelector, current));
        }

    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * (An implementation that is not an end-point may go upstream to fetch
     * missing values and cache them to satisfy future requests.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  one of EVENT_INTERVAL_SELECTOR_xxx values
     * @param intervalNumber  a time (as from System.currentTimeMillis())
     *     which identifies the first interval for which data is potentially
     *     required; if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException
        {
        return(_getPipeline().getEventValues(def, intervalSelector, intervalNumber, whichValues));
        }

    /**Synchronise variables with upstream values.
     * Pushes updated values upstream to the source,
     * calls sync on the source if called with the "force" argument true,
     * and then retrieves changed values from upstream.
     * <p>
     * When called with force==true, this acts like a full "memory barrier",
     * flushing all write-cached items downstream immediately and afterwards
     * getting the value of all upstream values with getVariables(-1),
     * but may be expensive in terms of CPU or bandwidth, so use sparingly.
     * <p>
     * When called with force=false, this incrementally flushes outstanding
     * writes and will then fetch all, or only new, values from upstream,
     * so is potentally much less resource-intensive.
     * In particular, this does not propagate the sync() upstream.
     * <p>
     * In any case, it is rarely the right thing for a casual user
     * to vall this as it may be very expensive.
     *
     * @param force  if true, this will force a full write flush,
     *     a full sync upstream,
     *     then full read with getVariables(-1),
     *     to get the effect of a full "barrier";
     *     otherwise, in general, a more incremental and non-propagating
     *     mode is used which still does a write flush but may chose
     *     to do a partial read of "new" upstream values
     *
     * @throws IOException if one is received from upstream
     */
    public void syncVariables(final boolean force)
        throws IOException
        {
        // Delegate.
        _getPipeline().syncVariables(force);
        }

    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(final PropsKey key,
                                              final long versionID)
        throws IOException
        {
        return(_getPipeline().getProperties(key, versionID));
        }


    /**Base of (immutable) lookup key for AEP-linked and non-AEP-linked tables.
     * Each object instance is unique,
     * and thus also can be statically created and kept private by its user.
     * <p>
     * Contains optional key useful for debugging/tracing.
     * <p>
     * This is not meaningfully Serializable or Cloneable.
     */
    private static class KeyBase
        {
        /**Default constructor; results in a null comment key. */
        public KeyBase() { this(null); }

        /**Optional comment key (may be null). */
        public KeyBase(final String optComment) { comment = optComment; }

        /**Hashcode depends only on object instance; not overridable. */
        public final int hashCode()
            { return(System.identityHashCode(this)); }

        /**Depends only on == and is not overridable. */
        public final boolean equals(final Object obj)
            { return(this == obj); }

        /**Comment; may be null. */
        public final String comment;

        /**Render as a String; never null. */
        public final String toString()
            { return((comment != null) ? comment : "(null-comment)"); }
        }

    /**Key for AEP-linked store. */
    public static final class AEPLinkedKey extends KeyBase
        { public AEPLinkedKey(final String comment) { super(comment); } }

    /**Key for non-AEP-linked store. */
    public static final class UnlinkedKey extends KeyBase
        { public UnlinkedKey(final String comment) { super(comment); } }

    /**Thread-safe (highly-concurrent) AEP-linked store; initialised on demand, not serialised. */
    private transient volatile ConcurrentHashMap<AEPLinkedKey, Object> storeLinked;

    /**Thread-safe (highly-concurrent) non-AEP-linked store; initialised on demand, not serialised. */
    private transient volatile ConcurrentHashMap<UnlinkedKey, Object> storeUnlinked;

    /**Get AEP-linked store, creating if necessary; never null.
     * Race conditions are possible that might allow two calls
     * to return two different values concurrently,
     * but this is assumed to at worst result in some wasted work.
     */
    private ConcurrentHashMap<AEPLinkedKey, Object> getStoreLinked()
        {
        ConcurrentHashMap<AEPLinkedKey, Object> sl;
        // Create new store if needed.
        while((sl = storeLinked) == null)
            { storeLinked = new ConcurrentHashMap<AEPLinkedKey, Object>(); }
        return(sl);
        }

    /**Get value against supplied key in AEP-linked store; result may be null.
     * May be cleared at any time, and will be cleared when the AEP changes.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     */
    public Object getAEPLinkedValue(final AEPLinkedKey key)
        { return(getStoreLinked().get(key)); }

    /**Store value against supplied key in AEP-linked store.
     * A null value is used to clear any extant entry.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return any previous value, or null if none
     */
    public Object putAEPLinkedValue(final AEPLinkedKey key, final Object value)
        {
        final ConcurrentHashMap<AEPLinkedKey, Object> sl = getStoreLinked();
        if(value == null)
            { return(sl.remove(key)); }
        else
            { return(sl.put(key, value)); }
        }
    
    /**Replace value for supplied key in non-AEP-linked store.
     * This (atomically) replaces the value for the given key
     * only if the extant value is that supplied.
     * <p>
     * Null values are not allowed.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return true if the value was replaced
     */
    public boolean replaceAEPLinkedValue(final AEPLinkedKey key, final Object oldValue, final Object newValue)
        {
        final ConcurrentHashMap<AEPLinkedKey, Object> sl = getStoreLinked();
        return(sl.replace(key, oldValue, newValue));
        }

    /**Store value against supplied key in AEP-linked store only if no value already mapped for that key.
     * A null value is used to clear any extant entry.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return any previous value, or null if none
     */
    public Object putIfAbsentAEPLinkedValue(final AEPLinkedKey key, final Object value)
        {
        final ConcurrentHashMap<AEPLinkedKey, Object> sl = getStoreLinked();
        if(value == null)
            { return(sl.remove(key)); }
        else
            { return(sl.putIfAbsent(key, value)); }
        }

    /**Get non-AEP-linked store, creating if necessary; never null.
     * Race conditions are possible that might allow two calls
     * to return two different values concurrently,
     * but this is assumed to at worst result in some wasted work.
     */
    private ConcurrentHashMap<UnlinkedKey, Object> getStoreUnlinked()
        {
        ConcurrentHashMap<UnlinkedKey, Object> su;
        // Create new store if needed.
        while((su = storeUnlinked) == null)
            { storeUnlinked = new ConcurrentHashMap<UnlinkedKey, Object>(); }
        return(su);
        }

    /**Get value against supplied key in non-AEP-linked store; result may be null.
     * This is designed for good concurrency and as little locking as possible.
     */
    public Object getUnlinkedValue(final UnlinkedKey key)
        { return(getStoreUnlinked().get(key)); }

    /**Store value against supplied key in non-AEP-linked store.
     * A null value is used to clear any extant entry.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return any previous value, or null if none
     */
    public Object putUnlinkedValue(final UnlinkedKey key, final Object value)
        {
        final ConcurrentHashMap<UnlinkedKey, Object> su = getStoreUnlinked();
        if(value == null)
            { return(su.remove(key)); }
        else
            { return(su.put(key, value)); }
        }
    
    /**Replace value for supplied key in non-AEP-linked store.
     * This (atomically) replaces the value for the given key
     * only if the extant value is that supplied.
     * <p>
     * Null values are not allowed.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return true if the value was replaced
     */
    public boolean replaceUnlinkedValue(final UnlinkedKey key, final Object oldValue, final Object newValue)
        {
        final ConcurrentHashMap<UnlinkedKey, Object> su = getStoreUnlinked();
        return(su.replace(key, oldValue, newValue));
        }

    /**Store value against supplied key in non-AEP-linked store only if no value already mapped for that key.
     * A null value is used to clear any extant entry.
     * <p>
     * This is designed for good concurrency and as little locking as possible.
     *
     * @return any previous value, or null if none
     */
    public Object putIfAbsentUnlinkedValue(final UnlinkedKey key, final Object value)
        {
        final ConcurrentHashMap<UnlinkedKey, Object> su = getStoreUnlinked();
        if(value == null)
            { return(su.remove(key)); }
        else
            { return(su.putIfAbsent(key, value)); }
        }

    /**The Scorer cache, currently not persistable/serialisable; never null. */
    private transient ScorerCacheImpl scorers = new ScorerCacheImpl(this, logger);

    /**Get the Scorer cache; never null.
     * We may restrict access in future.
     */
    public ScorerCacheIF getScorerCache() { return(scorers); }


    /**Deserialise.
     */
    private void readObject(final ObjectInputStream ois)
        throws IOException, ClassNotFoundException
        {
        // Default read of object.
        ois.defaultReadObject();

        // Ensure that a (new) embedded Observable is in place.
        _observable = new EDVHObservable();

        // Ensure that a (new) Scorer cache is in place.
        scorers = new ScorerCacheImpl(this, logger);

        // Validate object stat immediately.
        validateObject();
        }

    /**Validate fields/state.
     * Called in the constructor and possibly after de-serialising.
     * <p>
     * Barf if something bad is found.
     * (Maybe allow some extra info in debug version.)
     */
    public void validateObject()
        throws InvalidObjectException
        {
        // Ensure that Observable is in place.
        if(_observable == null)
            { throw new InvalidObjectException("invalid null Observable"); }
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -8932316265828567742L;
    }
