/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.location.GeoProximity;
import org.hd.d.pg2k.svrCore.location.GeoUtils;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;
import org.hd.d.pg2k.webSvr.util.WebConsts;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.net.InetAddress;
import java.nio.ByteBuffer;
import java.util.Enumeration;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.atomic.AtomicReference;

/**This is the servlet that serves exhibit data.
 * This is optimised to serve large quantities of binary data
 * efficiently, and to get headers right to optimise perfomance
 * of caches, spiders, etc, while not losing site of usage logging.
 * <p>
 * This expects to be loaded when the whole Web application
 * starts and remain there until the Web app shuts down,
 * and so creates and destroys the data pipeline as a side-effect.
 */
public final class ExhibitServlet extends HttpServlet
    {
    /**Default transfer buffer size; chosen to be efficient but not huge. */
    private static final int DEFAULT_BUF_SIZE = CoreConsts.BULK_DATA_TRANSFER_SIZE;

    /**True if we know that this servlet is only unloaded when the site is brought down.
     * We can use this to quickly destroy an old cache, for example.
     */
    public static final boolean ONLY_DESTROYED_ON_SHUTDOWN = true;

    /**Shut down gracefully.
     */
    public void destroy()
        {
System.out.println("ExhibitServlet shutting down...");

        // We can safely shut down the cache if we know
        // we are the last one out...
        if(ONLY_DESTROYED_ON_SHUTDOWN)
            {
            getDataSource(getServletConfig(), null).destroy();
            }
        }

    /**Get singleton (per-servlet-context) data pipeline/cache instance.
     * The config param must not be null, but for some operations
     * (such as calling destroy()) request can be null.
     */
    private static DataSourceBean getDataSource(
                final ServletConfig config,
                final HttpServletRequest request)
        {
        // Fetches/creates the data source...
        final DataSourceBean dataSource =
            DataSourceBean.getApplicationInstance(config.getServletContext());

        // Ensure that the essential details are set up.
        dataSource.setServletContext(config.getServletContext());
        if(request != null)
            { dataSource.setContextPath(request.getContextPath()); }

        return(dataSource);
        }

    /**Private to init(); ensure exactly-once semantics for init(). */
    private boolean initialised;

    /**Retrieve some important configuration data and cache it away.
     * A bug in Tomcat V3.1--V3.2.1 can cause this init() method to be
     * called twice (concurrently) on the same instance of this servlet,
     * so this routine must be safe in that scenario.
     */
    public void init(final ServletConfig config)
        throws javax.servlet.ServletException
        {
        super.init(config);

        // Work round Tomcat V3.x bug that produces repeated calls to init().
        synchronized(this)
            {
            if(initialised) { return; }
            initialised = true; // Don't try to reinitialise once started.

            // I create the DataSourceBean here if possible,
            // to have everything up and running before the first
            // user visits, since else they may see a huge start-up delay
            // and/or an empty site!
            //
            // These quick-start features will probably only be
            // enabled if the cache is in aggressive mode.
            getDataSource(config, null);
            }
        }


    /**Respond to a GET request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doGet(final HttpServletRequest request,
                      final HttpServletResponse response)
        throws IOException //, ServletException
        {
        doAction(request, response, false);
        }

    /**Respond to a HEAD request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doHead(final HttpServletRequest request,
                       final HttpServletResponse response)
        throws IOException // , ServletException
        {
        doAction(request, response, true);
        }

    /**If true then enable async read-ahead for large exhibits where throughput is low. */
    private static final boolean ENABLE_READAHEAD = true;

    /**Minimum number of blocks to consider using async read-ahead; strictly positive.
     * This reflects the fact that starting an additional thread
     * may have a significant overhead.
     */
    private static final int MIN_BLOCKS_FOR_READAHEAD = 2;

    /**Respond to a GET/HEAD request for the content served by this servlet.
     * Returns the exhibit data, with a correct MIME type.
     * <p>
     * If an exhibit is requested by a syntactically-invalid name,
     * we produce an HTTP `forbidden' (403) response.
     * <p>
     * If we request an exhibit with a valid name, but that does not
     * exist, return return a `non found' (404) response (unless
     * there are no exhibits available, perhaps because of a database
     * problem, in which case we may return `service unavailable' in the
     * hope that that will prevent search engines from purging our
     * exhibits if they happen to visit when we are reloading the
     * database).
     * <p>
     * We refuse to deal with some dubious requests
     * (eg apparently links directly to our exhibits from external sites)
     * if we are too busy, or if this server is marked as slow/expensive,
     * so as to conserve resources for bona fide users.
     * Dubious requests are rejected with a temporary failure (5xx) code
     * or a redirection to an alternate URL.
     * <p>
     * Were we observe poor throughput due to excessive time in data fetches
     * then we may switch to an asynchronous read-ahead mode.
     * <p>
     * TODO:
     * <ul>
     * <li>Deal with range requests
     *     in case container does not.
     * </ul>
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     */
    public void doAction(final HttpServletRequest request,
                         final HttpServletResponse response,
                         final boolean isHEAD)
        throws IOException //, ServletException
        {
        final long startTime = System.currentTimeMillis();

        // Reject any non-human fetch by an over-keen "precaching" client.
        // We can't spare the bandwidth for spurious cache-ahead behaviour.
        final ServletContext servletContext = getServletContext();
        if(WebUtils.isPrecacheRequest(request))
            {
            servletContext.log("ExhibitServlet: blocked attempt to 'precache' exhibit");
            response.sendError(HttpServletResponse.SC_FORBIDDEN, "Please do not attempt to 'precache' exhibits!");
            return;
            }

        // Get the relative request path...
        final String rawPathInfo = request.getPathInfo();
//System.err.println("rawPathInfo = " + rawPathInfo);
        // Adjust for any leading slash.
        final String pathInfo = ((rawPathInfo != null) && rawPathInfo.startsWith("/")) ?
            (rawPathInfo.substring(1)) : rawPathInfo;

        // Attempt to get all the info on the exhibit, which
        // also implies that it exists!
        // If anything fails, report `not found'.
        try
            {
            if(!ExhibitName.validNameSyntax(pathInfo))
                {
                if(!response.isCommitted()) // Don't bother with error if too late...
                    { response.sendError(HttpServletResponse.SC_NOT_FOUND); }
//                request.getRequestDispatcher(WebConsts.NOT_FOUND_PAGE).forward(request, response);
                return;
                }

            final DataSourceBean ds = getDataSource(getServletConfig(), request);
            final AllExhibitProperties aep = ds.getAllExhibitProperties(-1);

            if(aep.aeid.length < 1)
                {
                servletContext.log("WARNING: request for exhibit with none currently available.");
                if(!response.isCommitted()) // Don't bother with error if too late...
                    { response.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE); }
                return;
                }

            final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(pathInfo);
            if(esa == null)
                {
                if(!response.isCommitted()) // Don't bother with error if too late...
                    { response.sendError(HttpServletResponse.SC_NOT_FOUND); }
                return;
                }

            final org.hd.d.pg2k.svrCore.props.GenProps gp = ds.getGenProps(-1);

            // Veto dubious requests with a temporary fail if we are too busy.
            final boolean isBusy = WebUtils.isOverloaded(servletContext);
            final String hotlinkHost = WebUtils.requestProbablyReferredFromExternalSite(request);
            // If this looks like an unauthorised hotlinked referral,
            // then we may reject/deflect it to conserve bandwidth.
            if(hotlinkHost != null)
                {
                // If busy then reject request quickly with a 5xx error code.
                // Likewise if this server has scarce/expensive bandwidth.
                if(isBusy || (LocalProps.getServerSlowdownFactor() > 1))
                    {
                    if(!response.isCommitted()) // Don't bother with error if too late...
                        { response.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE); }
                    return;
                    }

                // Looks like a dodgy hotlink from someone unauthorised...
                if(ServletUtils.noteAndOrBlockHotlinker(servletContext, request, hotlinkHost, ds, gp))
                    {
                    // Note the blocked hotlink attempt.
                    final SimpleVariableValue svvLoc = new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                        "blockedHotlinkFrom=" + hotlinkHost);
                    ds.setVariable(svvLoc);

                    // Just give up if already too late to signal an error, etc.
                    if(!response.isCommitted())
                        {
                        // If an image URL has been supplied then divert to it...
                        final String altURL = gp.getWEBSVR_EX_HOTLINK_DIVERT_URL();
                        if(/* !isBusy && */ (altURL != null))
                            { response.sendRedirect(altURL); }
                        // For hotlinkers that we explicitly hate,
                        // just sent a 403 to save time and effort.
                        else if(gp.getHotLinkDisallowHosts().contains(hotlinkHost))
                            { response.sendError(HttpServletResponse.SC_FORBIDDEN); }
                        // Else redirect to the HTML catalogue page for the exhibit,
                        // which should work reasonably nicely
                        // for bona-fide search engine referrals, for example.
                        else
                            { response.sendRedirect(WebUtils.makeCatPageURL(esa.filePath, WebConsts.F_secondary_generated_HTML_suffix).toExternalForm()); }

                        final Enumeration rEn = request.getHeaders("Referer");
                        final boolean noRef = ((rEn == null) || !(rEn.hasMoreElements())); // No referer...
                        final String referer = noRef ? "" : (String) rEn.nextElement();
servletContext.log("WARNING: rejecting external exhibit hotlink from `"+hotlinkHost+"' for: " +esa.filePath+ " with Referer: " + TextUtils.sanitiseForXML(referer, 1024, true));
                        }

                    return; // Terminate response without sending any exhibit data.
                    }
                }


            final long llength = esa.length;
            assert(llength > 0);
            assert(llength <= Integer.MAX_VALUE);
            final int length = (int) llength;
            final String type = ExhibitMIME.getMIMEType(pathInfo);

            // If we get this far, everything is probably OK.

            // Set the main type and length headers for browsers.
            response.setContentLength(length);
            response.setContentType(type);

            // Set some cacheing headers too.
            // Make the cache time a small multiple of
            // the interval between rechecks of exhibit immutable data
            // (or our default slackness, if greater).
            // Increase it if we are busy so as to reduce future server load.
            final int basicCacheLifetime = 5 * Math.max(CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000,
                gp.getWEBSVR_MIN_EX_IMATTR_RECHECK_MS());
            final int cacheLifetime = isBusy ? (7 * basicCacheLifetime) : basicCacheLifetime;
            final boolean isFromSpider = WebUtils.requestProbablyFromSpider(request);

            // Set the HTTP/1.1 cache-control header to reflect
            // the maximum time that any proxy/end-user should cache the exhibit.
            response.setHeader("Cache-Control", "public,max-age="+(cacheLifetime/1000));

            // If this might be an HTML request from a search engine
            // then we set a long expiry so that search engines will
            // index and retain the page; of the order of weeks or even months.
            // This *may* force HTTP/1.0-browser users to hit RELOAD
            // more often than we'd like...
            final long expiryTime = isFromSpider
                ? Math.max(cacheLifetime, WebConsts.SPIDER_PAGE_EXPIRY_MS)
                : cacheLifetime;
            response.setDateHeader("Expires", System.currentTimeMillis() + expiryTime);

            // Handle any If-Modified-Since if not already done by the container.
            final long ifModifiedSince = request.getDateHeader("If-Modified-Since");
            if((ifModifiedSince > -1) && (esa.timestamp/1000 <= ifModifiedSince/1000))
                {
                // Abort if exhibit has not changed...
                response.sendError(HttpServletResponse.SC_NOT_MODIFIED);
                return;
                }
            // Set the true last-modified date.
            response.setDateHeader("Last-Modified", esa.timestamp);

            // If we have the exhibit MD5 hash available
            // then we can add the appropriate HTTP header as a bonus.
            final ExhibitPropsLoadable epl = aep.getExhibitPropsLoadable(esa.filePath);
            final AccessionData ad = epl.getAccessionMetadata();
            if((ad != null) && (ad.hashMD5 != null))
                {
                final String base64Hash = TextUtils.encode8To6(ad.hashMD5.getData());
//System.out.println("Generating extra HTTP header...  Content-MD5: " + base64Hash);
                response.setHeader("Content-MD5", base64Hash);
                }

            // Create (strong) ETag header as "size-date[-CRC32]" in hex.
            final StringBuilder etsb = new StringBuilder(36);
            etsb.append('"');
            etsb.append(Integer.toHexString(length));
            etsb.append('-').append(Long.toHexString(esa.timestamp));
            if((ad != null) && (ad.hashCRC32 != null))
                { etsb.append('-').append(Integer.toHexString(ad.hashCRC32.intValue())); }
            etsb.append('"');
            response.setHeader("ETag", etsb.toString());


            // If this is a HEAD request then return without providing the body.
            if(isHEAD) { return; }


            // Generate the body (ie send the raw exhibit).

            // Get access to the output stream...
            final ServletOutputStream os = response.getOutputStream();

            // Monitor relative read and write performance.
            long totalReadTime = 0;
            long totalWriteTime = 0;

            // Real humans fetching big exhibits (not hotlinked!)
            // are eligible for async read-ahead to improve throughput.
            // We use this to measure throughput
            // even when a read-ahead thread is not allowed.
            final int thresholdBytes = DEFAULT_BUF_SIZE * MIN_BLOCKS_FOR_READAHEAD;
            final boolean mayNeedReadAhead = !isFromSpider && (hotlinkHost == null) &&
                (esa.length >= 3*thresholdBytes);

            // We copy exhibit data a chunk at a time, followed by a flush().
            //
            // We attempt to make this robust in the face of
            // transient data-fetch errors,
            // retrying for a time proportional to that already spent
            // in order to reflect the cost of restarting from the beginning
            // to the end user.
            //
            // In general we assume that we buffer enough data
            // in the HTTP/TCP stack to avoid significant loss of throughput,
            // but when we wait a significant fraction of service time for reads
            // then we can start a potentially-resource-hungry read-ahead.
            //
            // We use a direct buffer for maximum read performance.
            // We'll use this same buffer in the async and sync read modes.
            final ByteBuffer buffer = ByteBuffer.allocateDirect(Math.min(length, DEFAULT_BUF_SIZE));
            // Queued data, for when we do read-ahead.
            final ArrayBlockingQueue<byte[]> data = (!ENABLE_READAHEAD || !mayNeedReadAhead) ? null :
                new ArrayBlockingQueue<byte[]>(1+2*MIN_BLOCKS_FOR_READAHEAD);
            Thread reader = null; // Reader thread, if we use it.
            // Signal thrown in thread, if any.
            final AtomicReference<IOException> bgThreadErr = new AtomicReference<IOException>();
            try
                {
                for(int start = 0; start < length; )
                    {
                    // Be prepared to retry to get the data if need be.
                    final long timeBeforeRead = System.currentTimeMillis();

                    // If read-ahead is enabled and has not been started yet
                    // (reads taking a significant fraction of write/wallclock time)
                    // and we're far enough through the transfer to have reasonable stats
                    // and we're far enough from the end to be worth starting a thread
                    // and we've been running slow (reads taking too long)
                    // and the write time is enough to make read-ahead/overlap useful
                    // and the download time so far has actually been significant
                    // then start the asynch thread to read ahead for us.
                    if(ENABLE_READAHEAD && mayNeedReadAhead && (reader == null) &&
                       (start >= DEFAULT_BUF_SIZE) && (length-start > thresholdBytes) &&
                       (totalReadTime > (totalWriteTime>>>4)) /* Reads are slowing throughput >~10%. */ &&
                       (totalReadTime < (totalWriteTime<<4)) /* >~10% time to be saved by overlapping reads and writes. */ &&
                       ((timeBeforeRead - startTime) > WebConsts.MAX_PG_DOWNLOAD_MS/2))
                        {
                        // Actually create/start the read-ahead thread.
                        final int startPos = start;

/* if(IsDebug.isDebug) */ { servletContext.log("[ExhibitServlet: INFO: starting async read-ahead @"+startPos+" for slow download (fetch "+totalReadTime+"ms, send "+totalWriteTime+"ms) for "+esa+".]"); }

                        reader = new Thread("ExhibitServlet read-ahead for "+esa){
                            @Override public final void run()
                                {
                                try
                                    {
                                    try
                                        {
                                        for(int i = startPos; i < length; )
                                            {
//if(IsDebug.isDebug) { System.out.println("[ExhibitServlet: async read-ahead for "+esa+".]"); }
                                            // Read the data...
                                            final byte b[] = readNextBlock(startTime, buffer, ds, esa, i, servletContext, null);
                                            // Queue it up to be read.
                                            data.put(b);
                                            // Adjust for the data read.
                                            i += b.length;
                                            }
                                        }
                                    catch(final IOException e)
                                        {
                                        bgThreadErr.set(e);
                                        data.put(QUEUE_POISON); // Poison the queue.
servletContext.log("ExhibitServlet: caught IOException in read-ahead thread: " + e.getClass().getName() + ": " + e.getMessage());
                                        }
                                    catch(final InterruptedException e)
                                        {
                                        data.put(QUEUE_POISON); // Poison the queue.
                                        throw e;
                                        }
                                    catch(final Throwable t)
                                        {
                                        final IOException e = new IOException("unexpected error");
                                        e.initCause(t); // Note the original exception.
                                        bgThreadErr.set(e);
                                        data.put(QUEUE_POISON); // Poison the queue.
                                        e.printStackTrace();
                                        }
                                    }
                                catch(final InterruptedException e)
                                    {
                                    // We were presumably killed off by the servlet thread,
                                    // So log the fact and quit immediately.
                                    servletContext.log("read-ahead thread interrupted");
                                    }
                                }
                            };
                        reader.setDaemon(true);
                        // Try to raise the reader priority a notch,
                        // but gracefully handle refusal!
                        try { reader.setPriority(Math.min(Thread.MAX_PRIORITY, 1+Thread.currentThread().getPriority()));}
                        catch(final Exception e) { e.printStackTrace(); }
                        reader.start();
                        }

                    // Read the next chunk of data.
                    final byte[] tmpBuf;
                    if(ENABLE_READAHEAD && (reader != null))
                        {
                        // Get data from async read-ahead thread.
                        //
                        // If the background thread got an error
                        // then rethrow it here to stop abruptly.
                        final IOException err = bgThreadErr.get();
                        if(err != null) { throw err; }
                        tmpBuf = data.take();
                        // Quit if the read-ahead thread deliberately poisoned the queue.
                        if(tmpBuf.length == 0) { throw new IOException("problem with read-ahead thread"); }
                        }
                    else
                        {
                        // Fetch the next block of exhibit data synchronously,
                        // ie using the current servlet thread.
                        tmpBuf = readNextBlock(startTime, buffer, ds, esa, start, servletContext, os);
                        }

                    // Note the amount of data actually read...
                    start += tmpBuf.length;

                    final long timeBeforeWrite = System.currentTimeMillis();
                    // Send the exhibit data to the user/client.
                    os.write(tmpBuf);
                    final long timeAfterWrite = System.currentTimeMillis();
                    totalReadTime += (timeBeforeWrite - timeBeforeRead);
                    totalWriteTime += (timeAfterWrite - timeBeforeWrite);
                    }

                // If the background thread got an error not already caught
                // then rethrow it here to stop abruptly.
                final IOException err = bgThreadErr.get();
                if(err != null) { throw err; }
                }
            finally
                {
                // Clean up the async read-ahead thread if (still) running,
                // interrupt()ing it if need be to wake it up or unblock it.
                if(ENABLE_READAHEAD && (reader != null))
                    {
                    while(reader.isAlive())
                        { reader.interrupt(); reader.join(CoreConsts.MAX_INTERACTIVE_DELAY_MS); }
                    }
                }
            os.flush(); // Done!
            os.close(); // Try to catch any broken/aborted connection now.

            assert((reader == null) || (!reader.isAlive())); // No reader thread running now.

            // Note slow downloads to help us tune the system.
            // We can only really hope to significantly speed up downloads
            // of exhibits that were eligible for read-ahead.
            if(mayNeedReadAhead)
                {
                // If the exhibit took significant time to download and
                // if we spent longer fetching the data than sending it
                // then we probably gave the user a slow download.
                final long wallclockTime = (System.currentTimeMillis() - startTime);
                if((totalReadTime > wallclockTime/4) &&
                   (wallclockTime > WebConsts.MAX_PG_DOWNLOAD_MS))
                    { servletContext.log("[ExhibitServlet: WARNING: slow download (fetch "+totalReadTime+"ms, send "+totalWriteTime+"ms, wallclock "+wallclockTime+"ms) for "+esa+", "+(float)(length/(0.001f*Math.max(1,wallclockTime)))+"Bps.]"); }
                }

            // Rate/note only deliberate downloads by humans on our site(s),
            // thus ignore:
            //   * Hits that appear to be from spiders.
            //   * Hits that appear to be direct hotlinks from external sites.
            //   * Hits for any (small) image/exhibit potentially
            //     being used as its own thumbnail.
            //   * Hits for any current GenProps-selected background image.
            // Note proximity of user also...
            if(!isFromSpider &&
               (hotlinkHost == null) &&
               !ImageUtils.canBeOwnThumbnail(esa) &&
               !esa.filePath.equals(gp.getWEBSVR_BG_IMAGE()))
                {
                try
                    {
                    final String fileComponent = ExhibitName.getFileComponent(esa.filePath);
					// Record the short exhibit name as an event...
                    final SimpleVariableValue svvDL = new SimpleVariableValue(
                        SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD,
                        fileComponent);
                    ds.setVariable(svvDL);
                    final SimpleVariableValue svvDLl = new SimpleVariableValue(
                        SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD_LOCAL,
                        fileComponent);
                    ds.setVariable(svvDLl);


                    final InetAddress clientIP =
                        InetAddress.getByName(request.getRemoteAddr());

                    // Note the download by approximate geographical region...
                    // Do so only AFTER the response completes to minimise latency,
                    // and only for requests that did not abort with an exception.
                    // We may be prepared to spend some time trying to look this up
                    // dynamically, hoping that DNS cacheing will keep the cost down.
                    final String regionByAddress = GeoUtils.getRegionByAddress(clientIP, false);
                    final SimpleVariableValue svvLoc = new SimpleVariableValue(
                        SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                        "download.loc=" + regionByAddress);
                    ds.setVariable(svvLoc);

                    // Note the rough proximity of the client so as to monitor
                    // how well geo-sensitive load balancing is doing.
                    // We can only do this if we know our physical location
                    // (eg if this is a mirror and thus has an embedded ccTLD).
                    // In particular we note downloads from very remote clients.
                    final String mirrorTag = LocalProps.getMirrorTag();
                    if(mirrorTag != null)
                        {
                        final GeoUtils.CCTLD cc = new GeoUtils.CCTLD(mirrorTag.substring(0, 2));
                        final GeoProximity proximity = GeoUtils.computeProximityByAddress(clientIP, cc, false);
                        final SimpleVariableValue svvProx = new SimpleVariableValue(
                            SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                            "download.prox=" + proximity);
                        ds.setVariable(svvProx);

                        // Log the location of users with no proximity
                        // since we might need to provide a server locally!
                        if(proximity == GeoProximity.NONE)
                            {
                            final SimpleVariableValue svvNoProx = new SimpleVariableValue(
                                SystemVariables.GENSTATS_STRING_GLOBAL_EVENT,
                                "download.noprox=" + regionByAddress);
                            ds.setVariable(svvNoProx);
                            }
                        }
                    }
                catch(final Exception e) // Absorb (but briefly log) any errors...
                    {
                    servletContext.log("ExhibitServlet: caught Exception: " + e.getClass().getName() + ": " + e.getMessage());
                    }
                }

            return; // Finished OK!
            }
        catch(final Throwable t) // Absorb (but briefly log) any error...
            {
            servletContext.log("ExhibitServlet: caught Throwable: " + t.getClass().getName() + ": " + t.getMessage());
if(IsDebug.isDebug) { t.printStackTrace(); }
            }

        // Some problem handling the request...
        if(!response.isCommitted()) // Don't bother with error if too late...
            { response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR); }
        return;
        }

    /**Can be inserted into the data queue by the async thread to force the foreground thread to quit. */
    private static final byte[] QUEUE_POISON = new byte[0];

    /**Read the next block of data from the exhibit using the given buffer returning it as a byte[]; never null nor zero-length.
     * Retries failed reads up to a time proportional to that already spent
     * on the entire transfer (to reflect the cost of restarting).
     *
     * @param os  if non-null may be used to flush() data down to the user;
     *     not safe to call from any thread other than the servlet thread
     */
    private static byte[] readNextBlock(final long startTime,
                                        final ByteBuffer buffer,
                                        final DataSourceBean ds,
                                        final ExhibitStaticAttr esa,
                                        final int start,
                                        final ServletContext servletContext,
                                        final ServletOutputStream os)
        throws IOException, InterruptedException
        {
        assert(buffer != null);

        final long timeBeforeRead = System.currentTimeMillis();
        final long retryUntil = timeBeforeRead + ((timeBeforeRead - startTime) / 4) +
                                    WebConsts.MAX_PG_DOWNLOAD_MS + 30000;

        // High-level recovery from temporary I/O errors.
        for( ; ; )
            {
            try
                {
                buffer.clear();
                assert(buffer.remaining() > 0); // Must not be zero-sized buffer.
                // Get the data, and cache it locally if possible.
                ds.getRawFile(buffer, esa.filePath, start, false);
                buffer.flip();
                final int bytesRead = buffer.limit();
                if(bytesRead > 0)
                    { break; /* OK, got some data! (Normal case.) */ }
                // Didn't get any data,
                // so fall through to sleep and retry.
                }
            catch(final IOException e)
                {
                if(System.currentTimeMillis() > retryUntil)
                    {
                    servletContext.log("ERROR: ExhibitServlet: retry FAILED at offset "+start+" for "+esa+": " + e.getMessage() + ".");
                    throw e; // Abort.
                    }
                }

            // Flush what we already read/wrote down to the user if possible,
            // thus aborting (SIGPIPE) if the user has given up and disconnected.
            if(null != os) { os.flush(); }

            // Wait a little while before trying again,
            // with some randomisation to help avoid collisions.
            // Not so long (1s+) as to force the TMF
            // into thinking that we are "busy".
            Thread.sleep(101 + Rnd.fastRnd.nextInt(107));
            servletContext.log("[ExhibitServlet: retrying read at offset "+start+" for "+esa+".]");
            }

        // FIXME: for now, with the current servlet API, we have to copy data into a byte[].
        final int bytesToWrite = buffer.limit();
        assert(buffer.position() == 0);
        assert(bytesToWrite > 0);
        final byte[] tmpBuf = new byte[bytesToWrite];
        buffer.get(tmpBuf);
        return(tmpBuf);
        }


    /**Get `last-modified' time for exhibit.
     * If we can't find out, eg because the exhibit is not present,
     * we return -1L, the default value.
     * <p>
     * If returning a last-modified value causes difficulty
     * (eg with Tomat 4.0.1) then we return -1.
     *
     * @param request The servlet request we are processing
     */
    public final long getLastModified(final HttpServletRequest request)
        {
//        if(WebConsts.AVOID_LAST_MODIFIED) { return(-1L); }

        // Get the relative request path...
        final String rawPathInfo = request.getPathInfo();
//System.err.println("rawPathInfo = " + rawPathInfo);
        // Adjust for any leading slash.
        final String pathInfo = ((rawPathInfo != null) && rawPathInfo.startsWith("/")) ?
            (rawPathInfo.substring(1)) : rawPathInfo;

        try
            {
            if(!ExhibitName.validNameSyntax(pathInfo))
                {
                // Invalid name; return default `not-known' value.
                return(-1L);
                }

            final ExhibitStaticAttr esa = getDataSource(getServletConfig(), request).
                            getStaticAttr(pathInfo);

            // No such exhibit; return default `not-known' value.
            if(esa == null) { return(-1); }

            // Return the exhibit's timestamp...
            return(esa.timestamp);
            }
        catch(final Exception e)
            {
            // In case of error return default `not-known' value.
            return(-1L);
            }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -1774514190468389217L;
    }
