/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;

/**Abstract base for an expression (for a FilterBean).
 * Expressions must be Serializable so that they can
 * be persisted.  They must not attempt to cache any reference,
 * direct or indirect, to a DataSourceBean.
 * <p>
 * Expression objects must be immutable and their accept()/sort()
 * methods must be idempotent.
 * <p>
 * Expressions may be nested like this:
 * <pre>
 * new SortExpr(new FilterExpr(null, new BuiltInFilters.filtAll(null)),
 *              new BuiltInFilters.sortByName(BuiltInFilters.sortByName.SMART))
 * </pre>
 * <p>
 * FIXME: this and all derived classes and expressions must support equals()
 *     and hashCode() where possible.
 */
public abstract class Expr implements java.io.Serializable
    {
    /**The upstream expression pipeline. */
    private final Expr upstream;

    /**Construct us with the upstream expression.
     * If the upstream expression is null then we are
     * the head of the expression chain and process the input first.
     */
    protected Expr(Expr _upstream)
        {
        upstream = _upstream;
        }

    /**Evaluate this expression.
     * Given the upstream expression, compute its results and
     * then compute ours on it.
     * <p>
     * We return a set of names, possibly empty.
     * <p>
     * The result array contains no duplicates nor nulls
     * and is never null itself.
     * <p>
     * Designed to be overridden and have super() called to
     * collect upstream results...  By itself returns all
     * the exhibits untouched, ie is a null filter/sorter,
     * the identity expression.
     * <p>
     * The head of the expression chain takes a copy of the inputs
     * and returns them; other items just pass the data up untouched.
     */
    public String[] eval(AllExhibitProperties aep, String in[])
        {
        // We are the head of the chain, we use the input...
        if(upstream == null) { return((String[]) (in.clone())); }

        // He are not the head of the chain, pass the input up.
        return(upstream.eval(aep, in));
        }

    /**Hash depends on upstream object, or is zero if none. */
    public int hashCode()
        {
        final Expr e = upstream;
        if(e == null) { return(0); }
        return(e.hashCode() + 1);
        }

    /**Equality depends on the upstream expressions being equal (or both null).
     * This does not return true unless the structure is the same and
     * all elements are the same.
     */
    public boolean equals(Object obj)
        {
        // We are identical to ourself.
        if(obj == this) { return(true); }

        if(!(obj instanceof Expr)) { return(false); }
        final Expr other = (Expr) obj;

        // If identical upstream (eg both null or both same object)
        // then these are the same.
        if(other.upstream == upstream) { return(true); }

        // If we are null (and the other is not), them we are not equal.
        if(upstream == null) { return(false); }

        // Compare upstream objects; can only work if equals is symmetrical.
        // This may be hard to achieve in derived classes without care,
        // so, for a start, we make sure that the implementation classes are
        // exactly the same.
        return((upstream.getClass() == other.upstream.getClass()) &&
               upstream.equals(other.upstream));
        }


    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -881322470744786046L;
    }
