/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import javax.servlet.ServletContext;
import java.io.IOException;
import java.util.List;

/**Version of the filter/sorter (cache) JavaBean that returns selected/sorted items as a List.
 * Other classes may find the simple select() methods useful
 * and thus this is not final.
 */
public class FilterBean extends AbstractFilterBean
    {
    /**Get the results of applying this to the whole exhibit set given the context (the data source is found automagically).
     * The data source must already have been initialised for this to
     * work.
     * <p>
     * The result is a set of full exhibit names as a private unsorted array.
     * <p>
     * This will only (re)compute its results if:
     * <ul>
     * <li>there are no cached results, or,
     * <li>the exhibit hash has changed, or,
     * <li>this is a memory-sensitive cache that has been automatically cleared.
     * </ul>
     * <p>
     * This never returns null.
     *
     * @throws IllegalStateException  if the filter has not been set
     */
    public List<String> select(final ServletContext ctxt)
        throws IOException, IllegalStateException
        {
        return(_select(DataSourceBean.getApplicationInstance(ctxt), null));
        }

    /**Get the results of applying this to the given exhibit set given the context (the data source is found automagically).
     * The data source must already have been initialised for this to
     * work.
     * <p>
     * The result is a set of full exhibit names as an immutable
     * RandomAccess List of String values.
     * <p>
     * If the inputSet argument is null the whole set of
     * exhibits is used, else the supplied set is used.
     * <p>
     * This will only (re)compute its results if:
     * <ul>
     * <li>there are no cached results, or,
     * <li>the exhibit hash has changed, or,
     * <li>the argument List is not the same one presented before (its hash has changed), or,
     * <li>this is a memory-sensitive cache that has been automatically cleared.
     * </ul>
     * <p>
     * This never returns null.
     * <p>
     * Null names and names that do not represent items currently
     * in the exhibit set are silently dropped.
     *
     * @param inputSet  null, or a list of full exhibit names
     *
     * @throws IllegalStateException  if the filter has not been set
     */
    public List<String> select(final ServletContext ctxt, final List<String> inputSet)
        throws IOException, IllegalStateException
        {
        return(_select(DataSourceBean.getApplicationInstance(ctxt), inputSet));
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -7508449420464606209L;
    }
