/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;

import java.util.HashSet;
import java.util.Set;

/**Wraps a filter to make an expression (for a FilterBean).
 * Filters must be Serializable so that filter expressions can
 * be persisted.  They must not attempt to cache any reference,
 * direct or indirect, to the DataSourceBean.
 * <p>
 * Filter objects must be immutable and the accept() method must be
 * idempotent and not depend on any state held within the object.
 * <p>
 * TODO: do object validation on deserialisation
 */
public final class FilterExpr extends Expr
    {
    /**The filter we apply here; never null. */
    private final FilterIF filter;

    /**Construct us with the upstream expression.
     * If the upstream expression is null this is the first
     * element in the expression pipeline.
     */
    public FilterExpr(Expr _upstream, FilterIF _filter)
        {
        super(_upstream);
        if(_filter == null) { throw new IllegalArgumentException(); }
        filter = _filter;
        }

    /**Evaluate this expression applying our filter.
     * Given the upstream expression, compute its results and
     * then compute ours on it.
     * <p>
     * We return a set of names, possibly empty.
     * <p>
     * The result array contains no duplicates nor nulls
     * and is never null itself.
     * <p>
     * Designed to be overridden and have super() called to
     * collect upstream results...  By itself returns all
     * the exhibits untouched, ie is a null filter/sorter.
     */
    public String[] eval(AllExhibitProperties aep, String in[])
        {
        // Chose our input.
        final String input[] = super.eval(aep, in);

        // Small optimisation to save lots of trouble...
        if(input.length == 0) { return(input); }

        // Ensure results non-duplicated.
        // Use a HashSet for insertion speed.
        final Set result = new HashSet(input.length);

        for(int i = input.length; --i >= 0; )
            {
            final String name = input[i];
            // Apply the user filter.
            if(!filter.accept(aep, name)) { continue; }
            // Hurrah!  We have a hit!
            result.add(name);
            }

        final String output[] = new String[result.size()];
        result.toArray(output);

        return(output);
        }

    /**Hash depends on underlying expression and this filter. */
    public int hashCode()
        {
        return((super.hashCode() << 5) - filter.hashCode());
        }

    /**Equality depends on underlying expression and this filter. */
    public boolean equals(Object obj)
        {
        if(!(obj instanceof FilterExpr)) { return(false); }
        final FilterExpr other = (FilterExpr) obj;
        if(!filter.equals(other.filter)) { return(false); }
        return(super.equals(obj));
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -7059115865126107312L;
    }
