/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.TextUtils;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.vars.EventPeriod;
import org.hd.d.pg2k.svrCore.vars.SimpleVariablePipelineIF;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.util.regex.Pattern;

/**Servlet-related utility functions.
 * This mainly consists of small support routines for servlets in this package.
 */
public final class ServletUtils
    {
    /**Prevent construction of an instance. */
    private ServletUtils() { }


    /**Default maximum percent of all exhibit requests that may come from one external source without us blocking them.
     * Normally in the range [1,100].
     * <p>
     * A value of a few percent should allow legitimate referrals from bona fide image search engines
     * and low volumes of traffic from small sites that we don't object to,
     * but should veto egregious heavy misuse of bandwidth.
     * <p>
     * At 20050205 a threshold of about 25% would just about let in all legitimate referrals
     * from the highest-volume search engines (images.search.yahoo.com, images.google.com)
     * and disallow most of the unwanted hotlinking from third-party sights.
     * This should prove itself to be reasonably conservative,
     * and is used if no explicit GenProps values is supplied.
     */
    private static final int DEFAULT_MAX_HOTLINK_REQ_PERCENT = 25;
    /**If true, allow Thread.yield() to delay the response to a hotlinker.
     * The downside of doing this is it ties up response threads for longer
     * and may prevent the server answering legitimate requests, ie appear busy.
     */
    private static final boolean DELAY_HOTLINKER_WITH_YIELD = false;

    /**Check if an apparent hotlinked request for an (exhibit or thumbnail) should be blocked.
     * Makes a note of the request (even "" referrals that we will never block),
     * and if the referrer is especially egregious will return false to suggest rejecting the request.
     * <p>
     * Absent or unparsable referrer URLs are not blocked by this mechanism
     * (as these may be spiders or users running with "Referer" turned off for security reasons).
     * They are dealt with in other ways.
     * <p>
     * Only call this where the referring URL seems dubious.
     * <p>
     * Algorithm for blocking requests by referrer:
     * <ol>
     * <li>If the host is on the "allow" list, then allow the request.
     * <li>Else if the host is on the "block" list, then block the request.
     * <li>If there were no hotlinks from this host yesterday,
     *     then block the request.
     *     The aim of this is to discourage people setting up hotlinks
     *     to start with from their own pages since they just won't work.
     * <li>Else if the requests from this referrer today/yesterday have amounted to
     *     (say) 10% or more of all exhibit requests in that period, then block the request.
     *     (Note that thumbnail requests should normally be much less common than exhibit requests
     *     because they can be cached for a very long time.)
     * </ol>
     * <p>
     * Only package visible since only needed by servlets in this package.
     *
     * @param request  the incoming request; never null
     * @param normalisedReferringHost  the normalised hotlinking host's name; never null
     * @param vars  the DataSource; never null
     *
     * @return true  if this request should probably be blocked
     */
    static boolean noteAndOrBlockHotlinker(final ServletContext context,
                                           final HttpServletRequest request,
                                           final String normalisedReferringHost,
                                           final SimpleVariablePipelineIF vars,
                                           final GenProps gp)
        {
        assert((context != null) &&
               (request != null) &&
               (normalisedReferringHost != null) &&
               (vars != null) &&
               (gp != null));

        // Immediately mark the connection as low-priority to allow us to conserve bandwidth, etc.
        request.setAttribute("LowPriConn", Boolean.TRUE);

//System.err.println("WARNING: seeing possible hotlink to exhibit/thumbnail from host/site: " + normalisedReferringHost);


        // Give up the CPU to potentially more-worthy users,
        // and possibly allow misuse stats to propagate
        // through more of the system,
        // but at the risk of tieing up connection resources.
        if(DELAY_HOTLINKER_WITH_YIELD && WebUtils.isLightlyLoaded(context))
            { Thread.yield(); }


        // We don't log or block requests with no referring URL or an unparsable one...
        // (Otherwise we'd never let spiders in, for example.)
        if("".equals(normalisedReferringHost))
            { return(false);}


        // Now record the dubious referring URL to keep tabs on such activity...
        // Note that the attempt is recorded even if we reject it below.
        try
            {
            // Record the short exhibit name as an event...
            final SimpleVariableValue svv = new SimpleVariableValue(SystemVariables.ACCESSPATTERN_EX_HOTLINK_REF_HOST,
                                                                    normalisedReferringHost);
            vars.setVariable(svv);
            }
        catch(final Exception e) // Absorb any errors and continue, but whinge...
            { e.printStackTrace(); }


        // Give up the CPU to potentially more-worthy users,
        // and possibly allow misuse stats to propagate through the system...
        if(DELAY_HOTLINKER_WITH_YIELD) { Thread.yield(); }


        // Decide whether this request should be rejected...

        // If the referring host is on our allow list then allow it immediately.
        // This can be used to quickly deal with well-known friends.
        if(gp.getHotLinkAllowHosts().contains(normalisedReferringHost))
            { return(false); /* Allow. */}
        final Pattern hotLinkAllowHostsRegex = gp.getHotLinkAllowHostsRegex();
        if((hotLinkAllowHostsRegex != null) &&
           hotLinkAllowHostsRegex.matcher(normalisedReferringHost).matches())
            { return(false); /* Allow. */}

        // If the referring host is on disallow list then block it immediately.
        // This can be used to quickly deal with well-known transgressors.
        if(gp.getHotLinkDisallowHosts().contains(normalisedReferringHost))
            { return(true); /* Block. */ }
        final Pattern hotLinkDisallowHostsRegex = gp.getHotLinkDisallowHostsRegex();
        if((hotLinkDisallowHostsRegex != null) &&
           hotLinkDisallowHostsRegex.matcher(normalisedReferringHost).matches())
            { return(true); /* Block. */ }

        try
            {
            // If there were no hotlink referrals from this host
            // in the previous period,
            // then we guess that this might be someone
            // designing a new page with the hotlink to us in it,
            // so refuse the hotlink outright so as to try to discourage them!
            // Genuine search engines, etc, that refer every day,
            // should stay clear of this block.
            // This should be relatively fast to check,
            // especially as the "previous period" events should be cache locally.
            final int prevCount =
                vars.getEventValue(SystemVariables.ACCESSPATTERN_EX_HOTLINK_REF_HOST, EventPeriod.VLONG, false).
                                getCount(normalisedReferringHost);
            if(prevCount < 1)
                {
                context.log("WARNING: reject/divert of new NEW hotlink Referer: " + TextUtils.sanitiseForXML(request.getHeader("Referer"), 256, true));
                return(true); /* Block. */
                }

            // If more than the allowable percentage of all exhibit downloads
            // have come from this referring URL in the current or previous period,
            // then disallow this request.

            // We include the current period for a fast response to new loads even though partial/noisy.
            // We include the previous period to give us some memory and a stable system-wide sample.

            // Total completed exhibit downloads in current/previous periods.
            final int totalExDownloads =
                    vars.getEventValue(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD, EventPeriod.VLONG, true).
                        getTotalEventCount() +
                    vars.getEventValue(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD, EventPeriod.VLONG, false).
                        getTotalEventCount();

            // Total attempted exhibit and thumbnail direct referrals from same host/site as this request.
            final int totalAttemptedReferrals =
                    vars.getEventValue(SystemVariables.ACCESSPATTERN_EX_HOTLINK_REF_HOST, EventPeriod.VLONG, true).
                        getCount(normalisedReferringHost) +
                    prevCount;

            // If too many referrals from this one host/site, then block this request.
            final int rawThresholdPC = gp.getWEBSVR_EX_HOTLINK_LIMITER();
            // A zero GenProps threshold percentage means "use the hard-coded (conservative) default".
            final int thresholdPC = (rawThresholdPC > 0) ? rawThresholdPC : DEFAULT_MAX_HOTLINK_REQ_PERCENT;
            final int threshold = ((totalExDownloads * thresholdPC) / 100);
            if(totalAttemptedReferrals > threshold)
                { return(true); /* Block the request... */ }
            // When close to the threshold, "brown-out" and reject some connections randomly as a warning...
            if(totalAttemptedReferrals >= ((4*threshold)/5))
                {
                context.log("WARNING: starting to dynamically reject/divert direct exhibit hotlinks (threshold="+threshold+", "+thresholdPC+"%) with Referer: " + TextUtils.sanitiseForXML(request.getHeader("Referer"), 256, true));
                if(Rnd.fastRnd.nextBoolean())
                    {
                    return(true); /* Block the request... */
                    }
                }
            }
        catch(final Exception e)
            {
            // Absorb unexpected error and let request the user's go through...
            e.printStackTrace(); // Log the unexpected problem...
            }

        return(false); // Let the request go through...
        }
    }
