/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 14-Apr-2003
 * Time: 18:00:51
 */
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.ExhibitName;

import java.util.*;

/**A simple fast mechanism to collect "similar" exhibits to a named one.
 * Can be used in a catalogue page to find similar exhibits quickly (in a CPU-efficient way),
 * possibly (when the site is busy) as an alternative to the more general default search.
 */
public final class SimpleSimilarExhibitFinder
    {
    /**Quickly returns a set of exhibits (short names) similar to the named exhibit (passed by full name); never null.
     * This should be constructed with a sorted List of exhibits (full names)
     * and an exhibit that exists in that list.
     * <p>
     * The sort order should probably be SMART_ORDER or something similar,
     * eg case-insensitively by the first "main" word in the name
     * (ignoring the directory component).
     * <p>
     * This will then return a list of exhibits adjacent in the original named
     * exhibit that has the first main word (matched case-insensitively)
     * but not including any variant of the original exhibit
     * and no more than one variant of any of the returned names.
     * <p>
     * A maximum number of results is specified,
     * and a list no longer than this will be returned.
     * <p>
     * This is designed to be CPU efficient yet return a useful set of results.
     * <p>
     * Results in same order as original list.
     * <p>
     * This is designed to be efficient generating a small number of results
     * (eg less than ten).
     * <p>
     * This attempts to be reasonably i=unbiased in terms of direction.
     *
     * @param exhibits  list of full exhibit names sorted with the exhibitComparator;
     *     not modified by the routine
     * @param exhibitComparator  exhibits must be sorted with this
     * @param exhibitName  full name of main exhibit; must be in exhibits list
     *     and neither it nor any variant will appear in the output
     * @param attrWords  set of all valid attribute words
     * @param maxResults  maximum number if results to return
     */
    public static String[] getSimpleSimilarExhibitList(final List<String> exhibits,
                                                       final Comparator<String> exhibitComparator,
                                                       final String exhibitName,
                                                       final Set<String> attrWords,
                                                       final int maxResults)
        {
        if((exhibits == null) ||
            (exhibitComparator == null) ||
            (exhibitName == null) ||
            (attrWords == null))
            { throw new IllegalArgumentException(); }

        // If no results required, don't compute any...
        if(maxResults <= 0) { return(NO_RESULTS); }

        // Find our own location in the array using same order as sort used...
        final int ourPos = Collections.binarySearch(exhibits, exhibitName, exhibitComparator);
        // If exhibit is not in list, return no results!  SHOULD NOT HAPPEN IF INPUTS ARE WELL-FORMED...
        if(ourPos == -1) { return(NO_RESULTS); }

        // Save recomputing the upper bound frequently.
        final int exhibitsSize = exhibits.size();
        // If there are no exhibits,
        // or only one which shoul dby rights be exhibitName,
        // we can return an empty result set.
        if(exhibitsSize < 2) { return(NO_RESULTS); }

        // Get our main word...
        final String mainWord = (String) (ExhibitName.getMainWords(exhibitName, attrWords).nextElement());

        // Case-insensitive store of main stems; primed with the initial exhibit.
        final Set<String> mainStems = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
        mainStems.add(ExhibitName.getMainWordsComponent(exhibitName, attrWords));

        // Gather results; keep them sorted as we go, though this might be inefficient.
        // Set our working result size to the largest it could possibly be
        // as we assume (a) that's not huge (b) we will often fill it.
        final Vector<String> result = new Vector<String>(Math.min(maxResults, exhibitsSize - 1));

        int lPos = ourPos;
        int hPos = ourPos;
        while((result.size() < maxResults))
            {
            // If we've fallen off the ends of the array,
            // we have to stop...
            if((lPos <= 0) && (hPos >= exhibitsSize - 1))
                { break; }

            // If the lower bound is still on the right word,
            // check out the entry.
            if(lPos > 0)
                {
                final String name = (String) exhibits.get(--lPos);

                // If it's got the wrong main word stop searching
                // in this direction immediately...
                final String lMainWord = (String) (ExhibitName.getMainWords(name, attrWords).nextElement());
                if(!mainWord.equalsIgnoreCase(lMainWord))
                    { lPos = -1; } // Put index out of bounds.
                else
                    {
                    // If we've not already seen something with this main stem,
                    // add it at the right end of our results!
                    if(mainStems.add(ExhibitName.getMainWordsComponent(name, attrWords)))
                        { result.insertElementAt(name, 0); }
                    }
                }

            // If the upper bound is still on the right word,
            // check out the entry.
            if(hPos < exhibitsSize - 1)
                {
                final String name = (String) exhibits.get(++hPos);

                // If it's got the wrong main word stop searching
                // in this direction immediately...
                final String hMainWord = (String) (ExhibitName.getMainWords(name, attrWords).nextElement());
                if(!mainWord.equalsIgnoreCase(hMainWord))
                    { hPos = exhibitsSize; } // Put index out of bounds.
                else
                    {
                    // If we've not already seen something with this main stem,
                    // add it at the right end of our results!
                    if(mainStems.add(ExhibitName.getMainWordsComponent(name, attrWords)))
                        { result.addElement(name); }
                    }
                }
            }

        // Truncate to size if we overshot a little...
        if(result.size() > maxResults)
            { result.setSize(maxResults); } // Truncate to requested size...

        // Convert to array of short names.
        if(result.size() == 0) { return(NO_RESULTS); } // Saves an allocation...
        final String r[] = new String[result.size()];
        result.toArray(r);
        // Finally, convert to short names...
        for(int i = r.length; --i >= 0; )
            { r[i] = ExhibitName.getFileComponent(r[i]); }
        return(r);
        }

    /**Fixed empty result object. */
    private static final String[] NO_RESULTS = new String[0];
    }
