/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;

/**Wraps a sorter to make an expression (for a FilterBean).
 * Sorters must be Serializable so that expressions can
 * be persisted.  They must not attempt to cache any reference,
 * direct or indirect, to a DataSourceBean.
 * <p>
 * Sorter objects must be immutable and the sort() method must be
 * idempotent.
 * <p>
 * TODO: do object validation on deserialisation
 */
public final class SortExpr extends Expr
                            implements java.io.Serializable
    {
    /**The sort we apply here; never null. */
    private final SortIF sorter;

    /**Construct us with the upstream expression.
     * If the upstream expression is null this is the first
     * element in the expression pipeline.
     */
    public SortExpr(Expr _upstream, SortIF _sorter)
        {
        super(_upstream);
        if(_sorter == null) { throw new IllegalArgumentException(); }
        sorter = _sorter;
        }

    /**Evaluate this expression applying our sort.
     * Given the upstream expression, compute its results and
     * then compute ours on it.
     * <p>
     * We return a set of names, possibly empty.
     * <p>
     * The result array contains no duplicates nor nulls
     * and is never null itself.
     * <p>
     * Designed to be overridden and have super() called to
     * collect upstream results...  By itself returns all
     * the exhibits untouched, ie is a null filter/sorter.
     * <p>
     * The result by be the input array sorted in place to avoid
     * copying, or may be a new array possibly smaller than the
     * input array; this may perform filtering or other operations
     * that needs to operate on the whole (current) set of exhibits
     * at once.
     */
    public String[] eval(AllExhibitProperties aep, String in[])
        {
        // Chose our input.
        final String input[] = super.eval(aep, in);
        // Sort in situ or into a new array.
        return(sorter.sort(aep, input));
        }

    /**Hash depends on underlying expression and this sorter. */
    public int hashCode()
        {
        return((super.hashCode() << 7) + sorter.hashCode());
        }

    /**Equality depends on underlying expression and this sorter. */
    public boolean equals(Object obj)
        {
        if(!(obj instanceof SortExpr)) { return(false); }
        final SortExpr other = (SortExpr) obj;
        if(!sorter.equals(other.sorter)) { return(false); }
        return(super.equals(obj));
        }


    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -7947761366290611963L;
    }
