/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.webSvr.util.*;

import javax.servlet.ServletConfig;
import javax.servlet.http.*;

import java.io.*;


/**This is the servlet that serves thumbnail data.
 * This is optimised to serve medium quantities of binary data
 * efficiently, and to get headers right to optimise perfomance
 * of caches, spiders, etc, while not losing site of usage logging.
 */
public final class ThumbnailServlet extends HttpServlet
    {
    /**Get singleton (per-servlet-context) data pipeline/cache instance.
     * The config param must not be null, but for some operations
     * (such as calling destroy()) request can be null.
     */
    private static DataSourceBean getDataSource(
            final ServletConfig config,
            final HttpServletRequest request)
        {
        // Fetches/creates the data source...
        final DataSourceBean dataSource =
            DataSourceBean.getApplicationInstance(config.getServletContext());

        // Ensure that the essential details are set up.
        dataSource.setServletContext(config.getServletContext());
        if(request != null)
            { dataSource.setContextPath(request.getContextPath()); }

        return(dataSource);
        }

    /**Respond to a GET request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     */
    public void doGet(final HttpServletRequest request,
                      final HttpServletResponse response)
//        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**Respond to a HEAD request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     */
    public void doHead(final HttpServletRequest request,
                       final HttpServletResponse response)
//        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**If true then we may generate/send the MD5 content hash for some/all thumbnails. */
    private static final boolean SEND_TN_MD5 = true;

    /**Respond to a GET/HEAD request for the content served by this servlet.
     * Returns the exhibit data, with a correct MIME type.
     * <p>
     * The request path is of the form /XXX/exhibitname where XXX
     * identifies the required thumbnail/sample size (small or standard).
     * <p>
     * If an exhibit is requested by a syntactically-invalid name,
     * we produce an HTTP `forbidden' (403) response.
     * <p>
     * If we request an exhibit with a valid name, but that does not
     * exist, return return a `not-found' (404) response (unless
     * there are no exhibits available, perhaps because of a database
     * problem, in which case we may return `service unavailable' in the
     * hope that that will prevent search engines from purging our
     * exhibits if they happen to visit when we are reloading the
     * database).
     * <p>
     * We refuse to deal with some dubious requests
     * (eg apparently hotlinks directly to our exhibits from external sites)
     * if we are too busy,
     * so as to conserve resources for bona fide users.
     * Dubious requests are rejected with a temporary failure (5xx) code.
     * <p>
     * TODO:
     * <ul>
     * <li>Deal with range requests
     *     in case container does not.
     * </ul>
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     */
    @SuppressWarnings("unchecked")
    public void doAction(final HttpServletRequest request,
                         final HttpServletResponse response)
//        throws IOException, ServletException
        {
        // Get the relative request path...
        final String rawPathInfo = request.getPathInfo();
//System.err.println("rawPathInfo = " + rawPathInfo);
        // Adjust for any leading slash.
        final String fullPathInfo = (rawPathInfo.startsWith("/")) ?
            (rawPathInfo.substring(1)) : rawPathInfo;

        final int firstSlash = fullPathInfo.indexOf('/');
        if(firstSlash == -1) // Whoops, syntax error...
            {
            response.setStatus(HttpServletResponse.SC_FORBIDDEN);
            return;
            }

        // Extract the putative exhibit name.
        final String exhibitName = fullPathInfo.substring(firstSlash + 1);

        // Is the request for standard or small?
        // Default to small to save bandwidth in case of mistakes!
        final boolean smallTn = !fullPathInfo.substring(0, firstSlash).equals(WebConsts.PATH_TN_STD);

        // Attempt to get all the thumbnails for the underlying exhibit,
        // which also implies that it exists!
        // If anything fails, report `not found'.
        try
            {
            if(!ExhibitName.validNameSyntax(exhibitName))
                {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return;
                }

            final DataSourceBean ds = getDataSource(getServletConfig(), request);

            // Veto dubious requests with a temporary fail if we are too busy.
            // Likewise if bandwidth is scarce/expensive for this server.
            final String hotlinkHost = WebUtils.requestProbablyReferredFromExternalSite(request);
            if((hotlinkHost != null) &&
               (WebUtils.isOverloaded(getServletContext()) ||
               (LocalProps.getServerSlowdownFactor() > 1) ||
               ServletUtils.noteAndOrBlockHotlinker(getServletContext(), request, hotlinkHost, ds, ds.getGenProps(-1))))
                {
                if(!response.isCommitted()) // Don't bother with error if too late...
                    { response.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE); }
//System.err.println("WARNING: rejecting external thumbnail referral from `"+hotlinkHost+"' for: " +pathInfo);
                return;
                }


            // Now try and retrieve the specified thumbnail(s);
            // create them if need be.
            final ExhibitThumbnails tns = ds.getThumbnails(exhibitName, true);
            // No thumbnails at all and never will be for this exhibit?
            // Stop immediately with a 404 Not Found.
            if(ExhibitThumbnails.NO_THUMBNAILS.equals(tns))
                { response.setStatus(HttpServletResponse.SC_NOT_FOUND); return; }
            // Might be possible to get/return them in the future...
            // So just note that the thumbnails are *currently* unavailable,
            //ie this may be a temporary problem worth a retry later.
            if(tns == null)
                { response.setStatus(HttpServletResponse.SC_SERVICE_UNAVAILABLE); return; }

            final ExhibitThumbnails.Thumbnail tn = smallTn ?
                tns.getSmall() : tns.getStandard();
            // No thumbnail of requested size?
            // Stop immediately with a 404 Not Found.
            if(tn == null) { response.setStatus(HttpServletResponse.SC_NOT_FOUND); return; }

            final int length = tn.size();
            final String type = ExhibitMIME.getMIMEType(exhibitName);

            // If we get this far, everything is probably OK.
            response.setContentLength(length);
            response.setContentType(type);

            // Set some cacheing headers too.
            // We let thumbnails stay in the user's broswer cache for a LONG time,
            // as long as static items of Web-site furniture,
            // because we assume that they almost never change once generated.
            // This should help improve system performance and reduce bandwidth requirements.
//            final int cacheLifetime = CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000; // Standard slackness.
            final long cacheLifetime = WebConsts.DEFAULT_STATIC_WEBITEMS_CACHE_MS;
            response.setHeader("Cache-Control", "public,max-age="+(cacheLifetime/1000));
            response.setDateHeader("Expires", System.currentTimeMillis() + cacheLifetime);

            // Set the last-modified date if available.
//            if(!WebConsts.AVOID_LAST_MODIFIED)
                {
                final long lastModified = getLastModified(request);
                if(lastModified != -1)
                    {
                    // Handle any If-Modified-Since if not already done by the container.
                    final long ifModifiedSince = request.getDateHeader("If-Modified-Since");
                    if((ifModifiedSince > -1) && (lastModified/1000 <= ifModifiedSince/1000))
                        {
                        // Abort if thumbnail has not changed...
                        response.sendError(HttpServletResponse.SC_NOT_MODIFIED);
                        return;
                        }
                    response.setDateHeader("Last-Modified", lastModified);
                    }
                }

            // We send the MD5 content hash header for a standard thumbnail
            // mainly to try to deter transparent proxies from transcoding
            // to "optimise" the image which may unacceptably degrade it;
            // some end users will use a standard thumbnail as-is.
            // However, we only attempt this for standard (larger) thumbnails,
            // and we let the small ones, of which more are transmitted, get munged.
            // We also don't want to add the bandwidth overhead
            // of an extra HTTP header for small thumbnails.
            if(SEND_TN_MD5 && !smallTn)
                {
                final String base64Hash = TextUtils.encode8To6(tn.getMD5Hash());
                response.setHeader("Content-MD5", base64Hash);
                }

            // If this is a HEAD request then return without providing the body.
            if("HEAD".equalsIgnoreCase(request.getMethod()))
                { return; }

// TODO: Deal with If-Modified-Since (and maybe ETAG) here...

            // Generate the body (ie send the raw thumbnail data).

            // Write the thumbnail in one go to the output stream...
            final OutputStream os = response.getOutputStream();
            tn.writeData(os);
            os.flush(); // Done!
            os.close(); // Try to catch any broken/aborted connection now.
            return; // Completed successfully.
            }
        catch(final Exception e) { } // Fall through to handle unexpected exception.

        // Failure while handling the result;
        // ask the user to try again if possible...
        if(!response.isCommitted())
            { response.setStatus(HttpServletResponse.SC_SERVICE_UNAVAILABLE); }
        }


    /**Get `last-modified' time for a thumbnail, or -1 if not available.
     * If we can find any thumbnails for the exhibit implied,
     * and those exhibits have a creation timestamp (older ones may not)
     * then we return that as our last-modified date.
     * <p>
     * In case of any difficulty we return the -1 "not-known" value.
     *
     * @param request The servlet request being processed
     */
    public final long getLastModified(final HttpServletRequest request)
        {
        // Get the relative request path...
        final String rawPathInfo = request.getPathInfo();
        // Adjust for any leading slash.
        final String fullPathInfo = (rawPathInfo.startsWith("/")) ?
            (rawPathInfo.substring(1)) : rawPathInfo;

        final int firstSlash = fullPathInfo.indexOf('/');
        if(firstSlash == -1) // Whoops, syntax error...
            { return(-1); }

        // Extract the putative exhibit name.
        final String exhibitName = fullPathInfo.substring(firstSlash + 1);

        // Attempt to get the thumbnails for the underlying exhibit.
        // If anything fails, return -1 to indicate no last-modified-date available.
        try
            {
            if(!ExhibitName.validNameSyntax(exhibitName))
                { return(-1); }

            final DataSourceBean ds = getDataSource(getServletConfig(), request);

            // Now try and retrieve the specified thumbnail(s) but don't create them.
            final ExhibitThumbnails tns = ds.getThumbnails(exhibitName, false);

            // If no tns available, stop now.
            if(tns == null)
                { return(-1); }

            // If this tns instance has no creation date
            // (eg because we have an old-style thumbnails object cached)
            // then use the AEID exhibit-set timestamp as a plausible proxy.
            if(tns.created <= 0)
                { return(ds.getAllExhibitImmutableData(-1).timestamp); }

            // We have the thumbnails' creation date, so return it!
            return(tns.created);
            }
        catch(final Exception e)
            { return(-1); }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -2497705561191352995L;
    }
