/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.exhibit;

import org.hd.d.pg2k.svrCore.ExhibitAttrUtils;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.MemoryTools;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import javax.servlet.ServletContext;
import java.io.IOException;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 24-May-2003
 * Time: 18:06:42
 */

/**This bean is designed to be used to support browsable sets of exhibits (ordered by name) as trees.
 * Only the main portion of the name (not the category, nor attributes, etc)
 * is relevant to the construction of the tree.
 * <p>
 * In general this allows a specified set of exhibits to be anchored at some partial URL in a WAR
 * and the tail URL be used to select some sub-tree or a leaf of exhibits to display in a page.
 * <p>
 * This, like its base class, silently forces internal data structures to be recomputed if
 * either the underlying data changes or the List input changes.
 * <p>
 * This logically maintains a tree organised by monocased prefix word sequences from the main stem
 * (ie exluding attribute words, number-in-series, author and exhibit suffix)
 * of the exhibits in the selected set.
 * <p>
 * This is thread-safe.
 * <p>
 * (This class contains a static convenience method for converting a partial trailing URL
 * to a monocased (lowercased) prefix word sequence sequence.)
 */
public final class TreeFilterBean extends AbstractFilterBean
    {
    /**Gets List of exhibits from filtered set with given lowercased prefix (from the main-word stem); null if prefix is invalid.
     * Returns a sorted (case-insensitive) unmodifiable List of:
     * <ul>
     * <li>lowercase prefixes (ending in "-"; links to nodes further from the root and starting with this prefix),
     * <li>short exhibit names (just the file component, not ending in "-").
     * </ul>
     * <p>
     * The data source is found automagically.
     * <p>
     * The data source must already have been initialised for this to work.
     * <p>
     * The result is a List of (short) exhibit names at this level
     * (ie that contain no more main words) interleaved with
     * the main-word prefixes of any longer exhibit names
     * (that can be used as the basis of a hyperlink for example)
     * as an immutable RandomAccess List of String values.
     * <p>
     * If the inputSet argument is null the whole set of
     * exhibits is used, else the supplied set is used.
     * <p>
     * This will only (re)compute its result if:
     * <ul>
     * <li>there are no cached results, or,
     * <li>the exhibit hash has changed, or,
     * <li>the argument List is not the same one presented before (its hash has changed), or,
     * <li>this is a memory-sensitive cache that has been automatically cleared.
     * </ul>
     * <p>
     * This returns null if the prefix is invalid syntactically
     * or if no item with that prefix exists.
     * This result may lead the Web server to return a 404 error code.
     * <p>
     * A prefix of "" returns the root node of the tree.
     * <p>
     * Null names and names that do not represent items currently
     * in the exhibit set are silently dropped.
     *
     * @throws IllegalStateException  if the filter has not been set
     */
    public synchronized List<String> selectPrefixesAndShortNames(final ServletContext ctxt,
                                                         final String prefixWordList,
                                                         final List<String> inputSet)
        throws IOException, IllegalStateException
        {
        // Force cached basic list to be (re-)computed if need be.
        final DataSourceBean dsb = DataSourceBean.getApplicationInstance(ctxt);
        final List<String> all = _select(dsb, inputSet); // This may clear our cached state.

        if(_cachedMap == null)
            {
            // (Re-)compute optimised map from stem/prefix to List nodes.
            _cachedMap = computePrefixMap(all, false, true);
            }

        // Look up prefix in map.
        // If nothing present (no such node) return null immediately.
        return((List<String>) _cachedMap.get(prefixWordList));
        }

    /**Given arguments as for selectPrefixesAndShortNames(), returns set of uninteresting parent nodes.
     * See getUninterestingNodesFromMap() for the definition of the returned value.
     *
     * @see TreeFilterBean#getUninterestingNodesFromMap(Map, String)
     *
     * @throws IllegalArgumentException  if prefix does not end with "-" or is otherwise ill-formed,
     *     or if prefix or any of its parents are not present in the map
     * @throws IllegalStateException  if the filter has not been set
     */
    public synchronized boolean[] getUninterestingNodes(final ServletContext ctxt,
                                                        final String prefixWordList,
                                                        final List<String> inputSet)
        throws IOException, IllegalStateException
        {
        // Do a select() to ensure that the cached map is up-to-date.
        // The whole routine is synchronized so that once this is done,
        // the Map cannot go away until we have had the chance to use it.
        selectPrefixesAndShortNames(ctxt, prefixWordList, inputSet);

        // Do the lookup in the cached map.
        return(getUninterestingNodesFromMap(_cachedMap, prefixWordList));
        }

    /**Given a Map as computed by computePrefixMap() and a prefix as passed to selectPrefixesAndShortNames(), returns set of uninteresting parent nodes.
     * Given an prefix with n words, this returns a length n-1 boolean array
     * for each parent node/word between (but excluding) the root and the specified node
     * (where this is passed the root prefix, in which case a zero-length array),
     * with a true element corresponding to each parent node/word in left to right (0 to high index) set
     * for each "uninteresting" node.  Other nodes are considered interesting.
     * <p>
     * The node immediately below the root is represented at element 0.
     * <p>
     * The returned array is private to the caller.
     * <p>
     * A node is considered "uninteresting" if it has exactly one child element which is either a prefix
     * (so presumably the prefix of a node below it and thus a wasted click for the user),
     * or a copy of the node below it (and thus not worth visiting in its own right).
     * <p>
     * A node is also considered uninteresting if it contains exactly the same information
     * as its parent node.
     * <p>
     * If the input is not a well-formed word-prefix string the result is undefined.
     * <p>
     * Thus this should work with optimised and non-optimised maps.
     * <p>
     * This version of the method is public primarily to allow testing of the algorithm in jUnit,
     * bit is used internally.  It does not alter any of its inputs and relies on them remaining unchanged
     * during its use of them.
     *
     * @throws IllegalArgumentException  if prefix does not end with "-" or is otherwise ill-formed,
     *     or if prefix or any of its parents are not present in the map
     */
    public static boolean[] getUninterestingNodesFromMap(final Map<String, Collection<String>> m,
                                                         final String prefixWordList)
        {
        // Nothing for the root node, since it has no parent nodes.
        if("".equals(prefixWordList)) { return(NO_ANSWER); }

        // Object like crazy if supposed prefix is ill-formed in a way we care about, ie:
        //  1) Does not end with a word separator.
        //  2) Starts with a word separator.
        //  3) Contains a run of two or more word separators.
        if(!prefixWordList.endsWith(ExhibitName.WORD_SEPS) ||
           prefixWordList.startsWith(ExhibitName.WORD_SEPS) ||
           (prefixWordList.indexOf(ExhibitName.WORD_SEPS + ExhibitName.WORD_SEPS) != -1))
            { throw new IllegalArgumentException("ill-formed prefix"); }

        final int prefixLen = prefixWordList.length();

        // Count the words in the input prefix.
        // We take this to be the same as the number of separators,
        // (of which the last is always in the last position and the first is at least index 1).
        int nSeps = 1;
        for(int i = prefixLen - 1; --i > 0; )
            { if(prefixWordList.charAt(i) == ExhibitName.WORD_SEP) { ++nSeps; } }

        // If there are less than two words in the prefix then the answer will always be an empty array.
        final int nTokens = nSeps;
        if(nTokens < 2) { return(NO_ANSWER); }

        // Create the result: mark the uninteresting ones.
        final boolean result[] = new boolean[nTokens - 1];

        String child = prefixWordList; // Prefix of child node (away from the root; the parent is a prefix if this).
        String parent; // Will be immediate parent of child...
        for(int resultIndex = result.length; --resultIndex >= 0; child = parent)
            {
            // Compute the prefix of the parent to this child node.
            final int lastSep = child.lastIndexOf(ExhibitName.WORD_SEP, child.length() - 2); // Last char is sep...
            parent = child.substring(0, lastSep + 1);

            final Collection<String> parentNode = m.get(parent);
            if(parentNode == null)
                { throw new IllegalArgumentException("prefix missing from map: " + parent); }

            // Is uninteresting if same content as child node.
            final Collection<String> childNode = m.get(child);
            if(childNode == null)
                { throw new IllegalArgumentException("prefix missing from map: " + child); }

            // This (child) node is considered boring
            // if it has the same content as its parent,
            // ie nothing new to see here at this child node.
            if((parentNode.size() == childNode.size()) &&
               ((new HashSet<String>(parentNode)).equals(new HashSet<String>(childNode))))
                {
                result[resultIndex] = true;
                continue;
                }

            // Is a contender to be uninteresting if there is only one entry at the node.
            if(parentNode.size() != 1)
                { continue; } // No, node is interesting!

            final String soleParentEntry = parentNode.iterator().next();

            // Is uninteresting if the entry is a prefix.
            if(soleParentEntry.endsWith(ExhibitName.WORD_SEPS))
                {
                result[resultIndex] = true;
                continue;
                }

            // If child is not same size, parent isn't uninteresting.
            if(childNode.size() != 1)
                { continue; }

            // If the child node content isn't the same as the child, parent isn't uninteresting.
            if(!soleParentEntry.equals(childNode.iterator().next()))
                { continue; }

            // This (parent) node is dull as ditchwater...
            result[resultIndex] = true;
            }

        return(result);
        }

    /**Fixed response for getUninterestingNodesFromMap() when the prefix is too short to answer. */
    private static final boolean NO_ANSWER[] = new boolean[0];


    /**Used to indicate a single exhibit at or below the given node. */
    private static final Integer JUST_ONE = Integer.valueOf(1);

    /**Given arguments as for selectPrefixesAndShortNames(), returns Map from name to count for all nodes at the given level; null if no nodes woth given prefix.
     * Returns an immutable Map from the given name (String)
     * to the count (Integer) of exhibits at or below that subnode
     * (or just 1 for an exhibit rather than subnode).
     * <p>
     * Caches its results for speed.
     *
     * @see TreeFilterBean#getUninterestingNodesFromMap(Map, String)
     *
     * @throws IllegalArgumentException  if prefix does not end with "-" or is otherwise ill-formed,
     *     or if prefix or any of its parents are not present in the map
     * @throws IllegalStateException  if the filter has not been set
     *
     * @return Map from String (name) to Integer count; null if no such prefix
     */
    public synchronized Map<String,Integer> getExhibitCount(
                                                    final ServletContext ctxt,
                                                    final String prefixWordList,
                                                    final List<String> inputSet)
        throws IOException,
               IllegalStateException
        {
        // Do a select() to ensure that the cached map is up-to-date.
        // The whole routine is synchronized so that once this is done,
        // the Map cannot go away until we have had the chance to use it.
        final List<String> l = selectPrefixesAndShortNames(ctxt, prefixWordList, inputSet);
        if(null == l)
            {
            return(null); // No such prefix.
            }

        // If no cache, create it.
        if(_prefixToCount == null) { _prefixToCount = new HashMap<String, Map<String,Integer>>(l.size()); }
        // Else if we have a cached value, return it...
        else
            {
            final Map<String,Integer> r = _prefixToCount.get(prefixWordList);
            if(r != null)
                { return(r); }
            }

        // Must compute (and cache) the result...
        final Map<String,Integer> result = getExhibitCountFromMap(_cachedMap, prefixWordList);
        _prefixToCount.put(prefixWordList, result);
        return(result);
        }

    /**Given arguments as for selectPrefixesAndShortNames(), returns Map from name to count for all nodes at the given level; null if no nodes woth given prefix.
     * Returns an immutable Map from the given name (String)
     * to the count (Integer) of exhibits at or below that subnode
     * (or just 1 for an exhibit rather than subnode).
     * <p>
     * This computes the value recursively from the source Map,
     * so may be slow on large exhibit sets, especially near the base.
     *
     * @throws IllegalArgumentException  if prefix does not end with "-"
     *     or is otherwise ill-formed,
     *     or if prefix or any of its parents are not present in the map
     * @throws IllegalStateException  if the filter has not been set
     *
     * @return Map from String (name) to Integer count; null if no such prefix
     */
    public synchronized Map<String,Integer> getExhibitCountFromMap(
                                                    final Map<String,Collection<String>> m,
                                                    final String prefixWordList)
        throws IOException,
               IllegalStateException
        {
        final Map<String,Integer> result = new HashMap<String, Integer>();

        final Collection<String> nodesAtThisLevel = m.get(prefixWordList);
        if((nodesAtThisLevel == null) || (nodesAtThisLevel.size() < 1))
            { return(null); }

        for(final Iterator<String> it = nodesAtThisLevel.iterator(); it.hasNext(); )
            {
            final String name = it.next();

            final boolean isSubTree = name.endsWith(ExhibitName.WORD_SEPS);

            // Ordinary exhibits have a count of exactly 1.
            if(!isSubTree)
                {
                result.put(name, JUST_ONE);
                continue;
                }

            // Sub-nodes have the sum/count of all the exhibits beneath them...
            // So we do a recursive call to compute this.
            final Map<String,Integer> snm = getExhibitCountFromMap(m, name);
            int sum = 0;
            final Collection<Integer> counts = snm.values();
            for(final Iterator<Integer> itsn = counts.iterator(); itsn.hasNext(); )
                {
                sum += (itsn.next()).intValue();
                }

            result.put(name, Integer.valueOf(sum));
            }

        return(Collections.unmodifiableMap(result));
        }


    /**Compute Map from prefix word sequence to SortedSet or List of exhibits at this node and any sub-nodes; never null.
     * This computes a Map from each possible (lower-cased) legal input prefix to
     * a SortedSet or immutable List of any sub-nodes
     * (as prefix word sequences ending with "-" and starting with the key)
     * and the short names of any exhibits at that level; all String values.
     * The sort of the SortedSet or List is case-insensitive.
     * <p>
     * The argument nodeAsSet determines whether nodes are SortedSet or List objects;
     * they both provide the same iteration order and both support operations such as contains().
     * A SortedSet will be faster for lookups (and is faster for this routine to construct),
     * whereas the (immutable) List form is more memory efficient and can be handed out
     * directly by selectPrefixesAndShortNames() for example.
     * <p>
     * (As an optimisation sub-nodes that contain only links may be omitted.)
     * <p>
     * The Map returned is mutable as are its keys and values.
     * To make access by multiple threads safe it must not be changed
     * and/or other activity must be otherwise serialised.
     * <p>
     * This computes the Map based on the exhibit list supplied rather than the
     * current AllExhibitProperties,
     * though the results are undefined unless all the exhibits
     * in the input list are part of the AllExhibitProperties.
     * <p>
     * This uses the attribute word set to help decide
     * how to parse the exhibit names.
     * <p>
     * This routine is uncached and stateless.
     * (It is exposed mainly to make testing easy.)
     *
     * @param inputExhibitList  null for all exhibits else
     *     a set of exhibits (as full exhibit names) for this tree;
     *     non-null, String, non-duplicated, valid full exhibit names
     *     (that should be a subset of those in aep)
     * @param nodeAsSet  if true, the values in the Map are of type SortedSet
     *     else they of type List and are sorted and unmodifiable;
     *     in both cases the values are the same type (String)
     *     and the iterator order is the same (case-insensitive)
     *     so for many uses they are algorithmically interchangable
     *     though with different performance considerations for different use
     *     patterns
     * @param optimise  if true, the routine will remove links down the tree
     *     into nodes that contain exactly one sub-node link and no exhibits
     *     (and thus save a user some redundant clicking);
     *     and such link will instead point to the the first multi-entry node
     *     en route to the leaf exhibts or become a direct reference to the
     *     leaf exhibit if there is only one below this point
     *     or the link will be replaced with the expanded set of exhibits
     *     at its destination if small enough to display on one page
     *
     * @throws IllegalArgumentException  if any input is invalid
     *     and in particular if any name in the input exhibit list
     *     is not a valid exhibit name
     *
     * @return Map from each possible (lower-cased) legal input prefix to
     *     a SortedSet or immutable List of any sub-nodes; never null
     */
    public static Map<String, Collection<String>> computePrefixMap(
                                       final List<String> inputExhibitList,
                                       final boolean nodeAsSet,
                                       final boolean optimise)
        {
        // Temp build map: completely mutable.
        final HashMap<String, Collection<String>> workingMap = new HashMap<String, Collection<String>>();

        // Get the set of attribute words.
        final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

        // Work through each exhibit in turn...
        for(final Iterator<String> it = inputExhibitList.iterator(); it.hasNext(); )
            {
            final String name = it.next();

            if(!ExhibitName.validNameSyntax(name))
                { throw new IllegalArgumentException("invalid exhibit name: " + name); }

            // Compute the lower-cased word stem, ignoring attribute words...
            // We carefully extract the main stem to get the trailing separator almost for free...
            final int[] offsets = ExhibitName.getMainAndAttrWordComponentBoundaries(name, attrWords);
            final String stem = name.substring(offsets[0]+1, offsets[1]+1).toLowerCase();

            // Create all the links to sub-nodes and create or add to our final node.
            final StringTokenizer st = new StringTokenizer(stem, ExhibitName.WORD_SEPS);
            // Full prefix so far, ie potentially partial.
            final StringBuilder partialPrefix = new StringBuilder(stem.length());
            // Partial prefix minus last component.
            final StringBuilder parentPartialPrefix = new StringBuilder(stem.length() - 2);
            while(st.hasMoreTokens())
                {
                final String thisWord = st.nextToken();
                partialPrefix.append(thisWord).append(ExhibitName.WORD_SEP);

                // First, if there are more main words below this point,
                // ensure that the intermediate nodes exist.
                // The root node is "".
                //
                // We may later optimise this to skip any intermediate nodes with no local exhibits.
                _insertRefToSubNode(workingMap, parentPartialPrefix.toString(), partialPrefix.toString());

                parentPartialPrefix.append(thisWord).append(ExhibitName.WORD_SEP);
                }

            // Now insert exhibit in its child node.
            _insertExhibit(workingMap, stem, name);
            }

        // Remove redundant internal links if requested.
        if(optimise)
            { optimiseMap(workingMap); }

        // If caller would prefer unmodifiable (sorted) List nodes,
        // do the conversion as the last step.
        if(!nodeAsSet)
            { convertNodesToList(workingMap); }

        // NOTE: this returned value is mutable.
        return(workingMap);
        }

    /**Convert the Map values to unmodifiable sorted List objects.
     * These can be directly handed out by select(), for example.
     * <p>
     * We may also take this opportunity to fold duplicates together
     * to trim memory use.
     */
    private static void convertNodesToList(final HashMap<String, Collection<String>> workingMap)
        {
        // Keep track of all the List values that we have already used,
        // so that we can easily eliminate duplicates.
        final Map<List<String>, List<String>> itemsCreated = new HashMap<List<String>, List<String>>(workingMap.size());

        // Replace each SortedSet value with a List item.
        for(final Iterator<String> it = workingMap.keySet().iterator(); it.hasNext(); )
            {
            final String key = it.next();
            final Collection<String> oldValue = workingMap.get(key);

            // When converting to a List, as an optimisation,
            // if the node has exactly one entry,
            // then directly use the (immutable) Collections.singletonList().
            List<String> newValue;
            if(oldValue.size() == 1)
                { newValue = Collections.singletonList(oldValue.iterator().next()); }
            else
                { newValue = Collections.unmodifiableList(new ArrayList<String>(oldValue)); }

            // Check if this is a duplicate node value...
            // This relies on equals()/hashCode() from the List implementations.
            final List<String> v = itemsCreated.get(newValue);
            if(v != null) { newValue = v; }
            else { itemsCreated.put(newValue, newValue); }

            workingMap.put(key, newValue); // Replace value.
            }
        }

    /**Optimise the map to eliminate redundant internal link traversals towards the leaves.
     * We do this so that a user can make fewer clicks to reach exhibits or a point where
     * they have to chose between two or more items.
     * <p>
     * This optimises the Map in situ.
     * <p>
     * This does not delete any nodes; arriving at an intermediate node will still work fine,
     * linking downwards to the final target node
     * (but also skipping any redundant intermediaries).
     * <p>
     * This routine is idempotent.
     * <p>
     * This routine must be run when the Map is complete
     * and while the nodes are still of SortedSet type.
     *
     * @param workingMap
     */
    private static void optimiseMap(final HashMap<String, Collection<String>> workingMap)
        {
        // If we iterate through the keys in reverse order
        // (that is, reverse case-insensitive)
        // then we will find that all sub-nodes (further from the root)
        // come before parent nodes.
        //
        // So, in one pass we can simply find any node where a link points
        // to a node with exactly one entry
        // (or where the node <em>is</em> excatly one link entry);
        // replace that link with (a copy of) whatever the sub-node contained.
        // Since we will be working up towards the root by virtue
        // of the sort order, that will eliminate all internal
        // downward links to uninteresting internal nodes.
        //
        // Also attempt to propagate up as far as possible entire subtrees
        // that are small enough to display on one page (including links and exhibits),
        // thus saving the user clicking one or more levels to see one pageful.
        final List<String> keys = (new ArrayList<String>(workingMap.keySet()));
        Collections.sort(keys, String.CASE_INSENSITIVE_ORDER);
        Collections.reverse(keys);

        for(final Iterator<String> it = keys.iterator(); it.hasNext(); )
            {
            final String key = it.next();

            // Expensive test; do only in debug mode or with assert().
            if(ORG.hd.d.IsDebug.isDebug && (((String) key).indexOf('/') != -1))
                { throw new IllegalArgumentException("non-short-name or prefix in map keys"); }

            final SortedSet<String> ss = (SortedSet<String>) workingMap.get(key);
            final boolean singleEntry = (ss.size() == 1); // Is single-entry node.
            // Putative new node(s) are copy of old one.
            final SortedSet<String> newSs = new TreeSet<String>(ss); // Conservative.
            final SortedSet<String> newSsEager = new TreeSet<String>(ss); // Eager.
            for(final Iterator<String> nit = ss.iterator(); nit.hasNext(); )
                {
                final String link = nit.next();
                // Only applicable where the entry we wish to replace is a link.
                if(!link.endsWith(ExhibitName.WORD_SEPS)) { continue; }

                // Now look up what the linked-to node contains.
                final SortedSet<String> linkedSs = (SortedSet<String>) workingMap.get(link);

                // Always expand with the linked-to target in the eager case.
                newSsEager.remove(link);
                newSsEager.addAll(linkedSs);

                // Conservatively, if this is a single-entry node
                // or the target is a single item
                // then we can replace our link with that of the linked-to target.
                if(singleEntry || (linkedSs.size() == 1))
                    {
                    // Replace our old link with the new single target.
                    newSs.remove(link);
                    newSs.addAll(linkedSs);
                    }
                }

            // Can we use an "optimised" node version?
            // (This assumes that we managed to optimise at all...)
            // If the "eager" version will all fit on one page then use it.
            if(!newSsEager.equals(ss) && (newSsEager.size() <= WebConsts.MAX_RESULTS_PER_PAGE))
                { workingMap.put(key, newSsEager); }
            // Else take advantage of any link-chasing, etc.
            else if(!newSs.equals(ss))
                { workingMap.put(key, newSs); }
            }
        }

    /**Helper method: inserts link from parent node to current node, creating parent node if need be.
     * This may correspond to a hyperlink down the tree in a realised HTML page for example,
     * allowing a user to browse sub-nodes with longer exhibit name stems.
     * <p>
     * Note the following conditions:
     * <ul>
     * <li>The arguments must not be null.
     * <li>The Map and its values must be modifiable.
     * <li>The sub-node must be a valid exhibit stem word-prefix string ending in a dash,
     *     and the parent must be a prefix of it.
     * <li>The parent must be a valid exhibit stem word-prefix string ending in a dash,
     *     or "" if the root node.
     * <li>The map must not be in use by another thread while this routine is running.
     * <ul>
     * A subset of these conditions may be (explicitly) checked at run-time.
     * <p>
     * Any new set created is sorted keyed on a String case-insensitively.
     * <p>
     * This intern()s the String values to eliminate duplicates.
     */
    private static void _insertRefToSubNode(final Map<String,Collection<String>> work,
                                            final String parentNode,
                                            final String subNode)
        {
        if(!subNode.startsWith(parentNode))
            { throw new IllegalArgumentException("parent not prefix of sub-node"); }
        if(!subNode.endsWith(ExhibitName.WORD_SEPS))
            { throw new IllegalArgumentException("sub-node does net end with " + ExhibitName.WORD_SEP); }
        if((parentNode.length() != 0) && !parentNode.endsWith(ExhibitName.WORD_SEPS))
            { throw new IllegalArgumentException("parent is not root and does net end with " + ExhibitName.WORD_SEP); }

        SortedSet<String> s = (SortedSet<String>) work.get(parentNode);

        // Create (and insert) new set for parent node if need be;
        if(s == null)
            {
            s = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
            work.put(MemoryTools.intern(parentNode), s);
            }

        // Create link to sub-node in situ if not already present.
        if(!s.contains(subNode))
            { s.add(MemoryTools.intern(subNode)); }
        }

    /**Helper routine to insert exhibit at its final node.
     * Note that the stem must be lower-cased and (ignoring case) a prefix of the exhibit name,
     * which in turn must be valid.
     * <p>
     * This will create the child node if need be.
     * <p>
     * The map must not be in use by another thread while this routine is running.
     * <p>
     * Any new set created is sorted keyed on a String case-insensitively.
     * <p>
     * Any subset of the conditions may be checked at runtime.
     * <p>
     * The exhibit name argument must be a full name,
     * but only the short version is inserted.
     */
    private static void _insertExhibit(Map<String,Collection<String>> work, String stem, String exhibit)
        {
        if(!stem.endsWith(ExhibitName.WORD_SEPS))
            { throw new IllegalArgumentException("stem (node) must end with " + ExhibitName.WORD_SEP + ": " + stem); }

        final String shortName = ExhibitName.getFileComponent(exhibit);

        // Expensive tests: do debug only or with assert().
        assert(ExhibitName.validNameSyntax(exhibit));
        assert(shortName.toLowerCase().startsWith(stem));

        SortedSet<String> s = (SortedSet<String>) work.get(stem);

        // Create (and insert) new set for exhibit node if need be;
        if(s == null)
            {
            s = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
            work.put(stem, s);
            }

        // Create entry for this exhibit in situ.
        // WE INSERT JUST THE SHORT NAME.
        s.add(shortName);
        }

    /**Cached Map from String prefix word list to Integer count of exhibits with that prefix.
     * Built on the fly as required.
     * <p>
     * Accessed under the instance lock.
     * <p>
     * Not stored if the bean is serialized.
     */
    private transient Map<String,Map<String,Integer>> _prefixToCount;

    /**Cached map in format returned by computePrefixMap with List nodes.
     * Not stored if the bean is serialized.
     */
    private transient Map<String, Collection<String>> _cachedMap;

    /**Clear cached state.
     * We have to chain to the super method first.
     * <p>
     * This method is available to a caller to help free memory,
     * eg if data is known to have changed and so any cache would be invalid anyway,
     * or if the user of this object has no further use for it.
     */
    public synchronized void clearCache()
        {
        super.clearCache();

        // Discard cached data.
        _cachedMap = null;
        _prefixToCount = null;
        }

    /**Convert trailing URL to monocased word sequence.
     * Converts a partial trailing URI of the form:
     *     <pre>
     *     <i>[/]</i>word1/word2/<i>...</i>/wordn/final.component
     *     </pre>
     * (containing no dashes or dots in the word components,
     * and ignoring any leading slash and final (dot-but-not-slash-containing) component)
     * into a monocased (lowercased) prefix word sequence sequence of the form:
     *     <pre>
     *     word1-word2-<i>...</i>wordn-
     *     </pre>
     * (note the trailing dash) which would be the prefix of any valid exhibit
     * (and exhibit main stem) in the requested subtree, ignoring case.
     * <p>
     * Note in particular that these three are all parsed the same, ie give the same result:
     * <pre>
     *     word1/word2/index.html
     *     word1/word2/
     *     word1/word2
     * </pre>
     * <p>
     * If trailing words are attributes or number-in-series then the resulting
     * sequence will not match any exhibits.
     * <p>
     * (This treats any runs of slashes as a single slash.)
     *
     * @throws IllegalArgumentException if a dash or non-word character is
     *    found in any but the final component of the URI,
     *    or the components are in any other way not legal words,
     *    or the result prefix would be too long to form part of a valid exhibit name
     */
    public static String convertTrailingURIToWordSequence(final String trailingURI)
        throws IllegalArgumentException
        {
        // Quick rejection of unlikely (far too long) names (ignoring final component)
        // which may be someone trying to break the server...
        if(trailingURI.lastIndexOf('/') > ExhibitName.MAX_NAME_LENGTH)
            { throw new IllegalArgumentException("partial URI too long"); }

        final StringBuilder result = new StringBuilder(trailingURI.length()); // Slight over-estimate...

        final StringTokenizer st = new StringTokenizer(trailingURI, "/");
        while(st.hasMoreTokens())
            {
            final String putativeWord = st.nextToken();

            // Ignore final component, unless:
            //   * URL ends in slash, or
            //   * Final component contains no dot (ie last dot is no later than last slash).
            if(!st.hasMoreTokens() &&
               !trailingURI.endsWith("/") &&
                (trailingURI.indexOf('.') > trailingURI.indexOf('/')))
                { break; }

            if(!ExhibitName.validWord(putativeWord))
                { throw new IllegalArgumentException("invalid word"); }

            // Append the word.
            result.append(putativeWord).append(ExhibitName.WORD_SEP);
            }

        // If too long (excluding final dash) to be a valid stem, then complain.
        if(result.length() > ExhibitName.MAX_STEM_LENGTH + 1)
            { throw new IllegalArgumentException("too long to be likely to be valid stem"); }

        return(result.toString().toLowerCase()); // Return converted to single case.
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 8710974049842992459L;
    }
