/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.exhibit;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.GenUtils;
import org.hd.d.pg2k.svrCore.PGMasterNotInServiceException;
import org.hd.d.pg2k.svrCore.ROByteArray;
import org.hd.d.pg2k.svrCore.SimpleLoggerIF;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataHTTPTunnelSource;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource;
import org.hd.d.pg2k.svrCore.datasource.SimpleExhibitPipelineIF;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource.DuplicateIDChecker;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataTunnelSource.PacketProtector;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.props.SecurityProps;
import org.hd.d.pg2k.svrCore.props.SimplepassProps;
import org.hd.d.pg2k.svrCore.stats.StatsLogger;
import org.hd.d.pg2k.svrCore.vars.InstanceID;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableValue;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.crypto.SecretKey;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.net.InetAddress;
import java.security.InvalidKeyException;
import java.security.Key;
import java.util.Map;


/**This serves data through an HTTP tunnel from master Web-app to authorised slaves, ie is an RPC end-point.
 * We are communicating on the client side with ExhibitDataHTTPTunnelSource
 * and may have (at least) one long-running open connection per slave.
 * <p>
 * Takes data from whatever the DataSourceBean is using.
 * <p>
 * This allows concurrency between different slaves,
 * and for multiple connections from each slave,
 * though this concurrency may be limited by resource constraints.
 * <p>
 * Notice that all operations require some degree of authentication:
 * <ul>
 * <li>A core set of safe read-only non-sensitive operations
 *     are available to clients that can provide a suitable authenticator
 *     (username/password) as HTTP fields,
 *     such as the JWS-based uploader.
 * <li>The full set of operations is available to clients (eg slave servers)
 *     that can authenticate by address,
 *     though "write" operations may be silently discarded/ignored
 *     when a second authenticator (eg HMAC using a shared pricate key)
 *     is available and required locally but is not provided by the client.
 *     (If the second authenticator is provided but the message does not validate,
 *     eg because the message was corrupted in transit,
 *     then the client will get an error so that they can retry.)
 * </ul>
 * <p>
 * Note that for messages that have a MAC:
 * <ul>
 * <li>We screen the message stamp for validity
 *    (ie within a certain window of our current time to allow for
 *    transmission delays and clock skew).
 * <li>For those message whose MAC is successfully validated (to stop DoS attacks)
 *     we store the MAC until some time after its stamp would become invalid
 *     and reject any further messages with the same MAC,
 *     ie we reject duplicate/replayed messages.
 * </ul>
 * <p>
 * TODO: enhanced security (prevent tampering) for plain HTTP connection.
 */
public final class TunnelServlet extends HttpServlet
    {
    /**If debugging locally, note some activity at protocol level. */
    private final static boolean isDebug = false && ORG.hd.d.IsDebug.isDebug;


    /**Unhook from servlet logger. */
    public void destroy()
        {
        logger.setContext(null); // Stop using servlet logger.
        super.destroy();
        }

    /**Hook into servlet logger. */
    public void init(final ServletConfig servletConfig)
        throws ServletException
        {
        logger.setContext(servletConfig.getServletContext()); // Use servlet.log().
        super.init(servletConfig);
        }

    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**The stats set to which we log general tunnel servlet stats.
     * The unique codes are the constants TSNAME_XXX.
     */
    private final StatsLogger.StatsConfig statsIDTSV =
        new StatsLogger.StatsConfig("TUNNELSERVLET",
                                    logger, // Use servlet log if poss.
                                    false, // Only dump summaries...
                                    12 * 3600, // About every 12 hours.
                                    true); // Adaptive.

    /**General stats event name: inbound HTTP request. */
    public static final String TSVNAME_HTTPREQUEST = "HTTP-request";

    /**General stats event name prefix: inbound HTTP request service time (ms exponent). */
    public static final String TSVNAMEP_HTTPSERVICETIME = "HTTP-serviceTime=";

    /**General stats event name prefix: slow inbound HTTP request op code. */
    public static final String TSVNAMEP_HTTPSLOWOP = "HTTP-slowOp=";

    /**General stats event name: failures handling input HTTP requests. */
    public static final String TSVNAME_HTTPFAIL = "HTTP-failure";

    /**General stats event name: inbound HTTP request rejected by client IP address. */
    public static final String TSVNAME_HTTPREJADDR = "HTTP-reqRejectedByAddr";

    /**General stats event name: inbound HTTP request rejected by ID/passwd. */
    public static final String TSVNAME_HTTPREJID = "HTTP-reqRejectedByID";

    /**General stats event name: inbound HTTP request rejected by request op-code. */
    public static final String TSVNAME_HTTPREJOP = "HTTP-reqRejectedByOp";

    /**General stats event name: inbound HTTP request rejected by failed MAC. */
    private static final String TSVNAME_HTTPREJMAC = "HTTP-reqRejectedByMAC";

    /**General stats event name: inbound HTTP request rejected as a duplicate. */
    private static final String TSVNAME_HTTPREJDUP = "HTTP-reqRejectedAsDup";

    /**General stats event name prefix: RPC request packet type. */
    public static final String TSVNAMEPR_RPCTYPE = ExhibitDataTunnelSource.TSNAMEPR_RPCTYPE;


    /**Maximum expected typical service time, ms.
     * Should generally be well under 1000ms.
     */
    private static final int MAX_TYPICAL_SERVICE_TIME_MS = 8000;

    /**Get singleton (per-servlet-context) data pipeline/cache instance.
     * The config param must not be null, but for some operations
     * (such as calling destroy()) request can be null.
     * <p>
     * This does not cache its return value.
     */
    private static DataSourceBean getDataSourceBean(
            final ServletContext ctxt,
            final HttpServletRequest request)
        {
        // Fetches/creates the data source...
        final DataSourceBean dataSource =
            DataSourceBean.getApplicationInstance(ctxt);

        // Ensure that the essential details are set up.
        dataSource.setServletContext(ctxt);
        if(request != null)
            { dataSource.setContextPath(request.getContextPath()); }

        return(dataSource);
        }

    /**Get data source/pipeline; never returns null.
     * This routine may or may not cache its result,
     * but the caller should not cache it.
     */
    private SimpleExhibitPipelineIF getDataSource()
        {
        final DataSourceBean dataSource =
            getDataSourceBean(getServletContext(), null);
        return(dataSource);
        }

    /**Cached SecurityProps; private to getSecurityProps(); never null.
     * Volatile so as to allow access without a lock.
     */
    private volatile SecurityProps _gSP_cache = new SecurityProps();

    /**Get SecurityProps (including sensitives values); never returns null.
     * May cache results for performance.
     *
     * @throws IOException  if SecurityProps cannot be retrived
     *     (eg because no local file exists)
     */
    private SecurityProps getSecurityProps()
        throws IOException
        {
        // Get handle on cached properties.
        final SecurityProps cached = _gSP_cache;
        assert(cached != null);

        // Get timestamp of cached properties.
        final long ts = cached.timestamp;
        // Get updated set if need be...
        final SecurityProps sp =
            SecurityProps.getSecurityPropsUncachedFromFilesystem(ts);

        // If new then cache and return the new set.
        if(sp != null)
            {
            _gSP_cache = sp;
            return(sp);
            }

        // Return the still-current cached set.
        return(cached);
        }

    /**Cache/lock object to optimise handling of incoming RPCs; never null. */
    private final ExhibitDataTunnelSource.HIRPCCache _RPC_cache =
        new ExhibitDataTunnelSource.HIRPCCache();
    
    /**Record of unique request packet IDs around our acceptance window; never null.
     * We use the final MAC segment ID, as that covers the entire frame.
     * <p>
     * We actually remember IDs for about twice the age implied by the skew
     * so that if our clock is wrong by, or slips by, the maximum skew,
     * we won't start admitting very old duplicate messages.
     * <p>
     * We should only ad IDs if possible when we are already fairly sure of the source
     * and when we have, for example, already checked for acceptable skew,
     * to make any sort of DoS attack against us harder.
     * <p>
     * We should reject otherwise-acceptable messages that have an ID
     * already present in this Map.
     * <p>
     * Thread-safe.
     */
    private final DuplicateIDChecker<ROByteArray> messageIDs =
        new DuplicateIDChecker<ROByteArray>(2*CoreConsts.MAX_PEER_CLOCK_SKEW_MS);

    /**Respond to a POST request to open and use an HTTP tunnel (for a single transfer).
     * Slave servers should refuse to respond to these requests;
     * ie only respond if seemingly a master.
     * <p>
     * Normal slave servers are primarily authenticated by IP address
     * and have full real/write access.
     * <p>
     * Uploader JWS clients are authenticated by username/password,
     * and have partial, read-only access.
     *
     * @exception IOException if an input/output error occurs
     */
    public void doPost(final HttpServletRequest request,
                       final HttpServletResponse response)
        throws IOException // , ServletException
        {
        // Record when we started to service the request...
        final long startTime = System.currentTimeMillis();

        final ServletContext context = getServletConfig().getServletContext();

        // HTTP/1.0 headers.
        response.setHeader("Expires", "0");
        response.setHeader("Pragma", "no-cache");
        // HTTP/1.1 headers.
        response.setHeader("Cache-Control", "no-cache,no-store,must-revalidate");

if(isDebug) { logger.log("[Receiving tunnel request...]"); }

        // Note HTTP request...
        StatsLogger.captureDataPoint(statsIDTSV, TSVNAME_HTTPREQUEST);

        final DataSourceBean ds = getDataSourceBean(getServletContext(), request);

        // It is now possible for slaves to accept a limited range of requests.
//        final Boolean isS = ds.isSlave();
//        // If we don't know or we are definitely a slave, reject the request.
//        if(!Boolean.FALSE.equals(isS))
//            {
//if(isDebug) { logger.log("[ERROR: slave rejecting tunnel request.]"); }
//            response.setStatus(HttpServletResponse.SC_FORBIDDEN);
//            return;
//            }

        // Initialisation time for call and response...
        final long initTime = System.currentTimeMillis();

        // Values we'll set later if we get that far.
        long validationTime = 0; // When we finish validation...
        long requestReadTime = 0; // When we finish reading the request...
        long responseGenerationTime = 0; // When we finish generating the response...
        Byte opcode = null; // The request's opcode.
        long requestSize = 0; // Request payload size.
        long responseSize = 0; // Response payload size.

        // By default allow only partial, read-only access.
        boolean partialAccess = true;

        final String remoteAddr = request.getRemoteAddr();

        try
            {
            // If the client does not seem to be authorised,
            // tell them to go away!
            // By default reject callers...

            // Get the user's IP address.
            final java.net.InetAddress addr = java.net.InetAddress.getByName(remoteAddr);

            // If the client has ID/passwd, then authenticate with that.
            final String userID = request.getHeader(ExhibitDataHTTPTunnelSource.HEADER_AUTH_ID);
            if(userID != null)
                {
                // Authenticate by ID/passwd rather than by header.
                if(!SimplepassProps.isAuthorUploadPasswordCorrect(userID,
                    request.getHeader(ExhibitDataHTTPTunnelSource.HEADER_AUTH_PW)))
                    {
                    context.log("rejecting putative tunnel user by ID/pass: " + userID + " @ " + remoteAddr);

                    // Note rejected HTTP request (by ID/password).
                    StatsLogger.captureDataPoint(statsIDTSV, TSVNAME_HTTPREJID);

                    // Set an appropriate HTTP error response code if possible.
                    response.setStatus(HttpServletResponse.SC_FORBIDDEN);
                    return;
                    }

                // Authenticated (for partial read-only access) by password.
                partialAccess = true;
                }
            // Try validating as a slave, ie a tunnel client, by the remote IP.
            else if(getSecurityProps().tunnelClientIsOK(addr))
                {
                // Authenticated (for full tunnel access) by IP address.
                partialAccess = false;
                }
            // Try validating the client as a peer mirror/slave.
            // We'll only allow this access if bandwidth is not expensive
            // and if we are not too busy.
            else if((LocalProps.getServerSlowdownFactor() < 2) &&
                    isPeerMirror(ds, addr) &&
                    !WebUtils.isOverloaded(context))
                {
                // Authenticated (for partial tunnel access) by IP address.
                partialAccess = true;
                }
            else // Unauthorised client...
                {
                context.log("rejecting putative tunnel user by address: " + remoteAddr);

                // Note rejected HTTP request (by client address).
                StatsLogger.captureDataPoint(statsIDTSV, TSVNAME_HTTPREJADDR);

                // Set an appropriate HTTP error response code if possible.
                response.setStatus(HttpServletResponse.SC_FORBIDDEN);
                return;
                }

            // Once we have done preliminary authentication/screening of the request,
            // look for a MAC header.
            // Veto obviously-broken communications, even if we don't have a key.
            PacketProtector pp = null;
            final String hdrMAC = request.getHeader(ExhibitDataHTTPTunnelSource.HEADER_MAC);
            if(null != hdrMAC)
                {
                try { pp = PacketProtector.fromCheckString(hdrMAC); }
                catch(final IllegalArgumentException e) { logger.log("garbled MAC on request from: "+remoteAddr); }
//if(IsDebug.isDebug) { logger.log("INFO: received request MAC: " + pp); }

                // Veto reponses with the wrong length (or a missing Content-Length).
                final int contentLengthRcvd = request.getContentLength();
                if(contentLengthRcvd == -1)
                    { logger.log("WARNING: missing Content-Length on tunnel request from: "+remoteAddr); }
                else if(pp.length != request.getContentLength())
                    { response.setStatus(HttpServletResponse.SC_FORBIDDEN); return; }
                // Check that the remote peer's clock is not vastly wrong
                // (or that a very old packet is reappearing out of nowhere).
                // We assume that the skew allowance is generous enough to cover
                // transit time of messages in the network.
                final long now = System.currentTimeMillis();
                if((pp.timestamp < now - CoreConsts.MAX_PEER_CLOCK_SKEW_MS) ||
                   (pp.timestamp > now + CoreConsts.MAX_PEER_CLOCK_SKEW_MS))
                    {
                    // This can happen if the remote machine's system clock is badly wrong.
                    System.err.println("ERROR: rejecting message with bad timestamp from: "+remoteAddr);
                    logger.log("ERROR: rejecting message with bad timestamp from: "+remoteAddr);
                    response.setStatus(HttpServletResponse.SC_FORBIDDEN); return;
                    }
                // Veto duplicate packets/requests in this tunnel.
                if(messageIDs.add(pp.mac.get(pp.mac.size()-1)) != null)
                    {
                    logger.log("ERROR: rejecting duplicate message from: "+remoteAddr);
                    response.setStatus(HttpServletResponse.SC_FORBIDDEN); return;
                    }
                }

            // Get the Xfer key if we have one.
            final SecretKey xk = LocalProps.getXferHMACKey();

            // If enforcing the MAC, and one is missing,
            // then enforce at most partial (read-only) access.
            if(ExhibitDataHTTPTunnelSource.ENFORCE_MAC &&
                !partialAccess &&
                ((xk != null) && (pp == null)))
                {
                logger.log("INFO: did not receive MAC from client "+remoteAddr+": restricting to partial access");
                partialAccess = true;
                }

            // Note when initial validation/authorisation complete.
            validationTime = System.currentTimeMillis();

            // Indicate that we are sending binary data down the line...
            response.setContentType("application/octet-stream");

            // Get access to the input and output streams...
            // If we have (and can check) a MAC then we wrap the input for verification
            // (which also buffers it),
            // else we buffer the input stream for efficiency.
//            final InputStream is = new BufferedInputStream(request.getInputStream(), 1024);
            final InputStream isRaw = request.getInputStream();
            final InputStream is = ((pp == null) || (xk == null)) ? new BufferedInputStream(isRaw, 1024) :
                pp.protectInputStream(xk, isRaw);

if(isDebug) { logger.log("ready to process request for tunnel user " + remoteAddr); }

            // Get the request from the slave...
            final ExhibitDataHTTPTunnelSource.RawPacket reqPacket =
                ExhibitDataHTTPTunnelSource.RawPacket.readPacket(is);
            // Note when we have finished reading the raw request off the wire.
            requestReadTime = System.currentTimeMillis();

            // For partial access reject all but specified op-codes.
            // The allowed operations are all safe "read-only" activities,
            // and allow the fetching of generic and exhibit properties,
            // and the fetching of exhibit data and thumbnails,
            // but no access at all to (possibly-slightly-sensitive) event data.
            if(partialAccess)
                {
                switch(reqPacket.opCode)
                    {
                    // A small number of "safe" operations are allowed.
                    // This excludes all (read and write) operations on event data.
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_NOOP:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_getGenProps:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_getAllExhibitProperties:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_getAllExhibitPropertiesDiff:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_getRawFile:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_getThumbnails:
                        break;

                    // Everything else is not allowed by default...
                    // Explicitly, write access to vars/events is forbidden.
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_setVariables:
                    case ExhibitDataHTTPTunnelSource.RawPacket.OP_syncVariables:
                    default:
                        {
                        context.log("rejecting partial-access tunnel user for op-code: " + (reqPacket.opCode) + " @ " + remoteAddr);

                        // Note rejected HTTP request (by client address).
                        StatsLogger.captureDataPoint(statsIDTSV, TSVNAME_HTTPREJOP);

                        // Quickly reject non-safe request with a proper packet...
                        // This OP__REMEX is not interpretted by the other end of the tunnel
                        // as indicating that the tunnel is broken,
                        // so thus this does not deter other, allowed, ops.
                        //throw new RuntimeException("illegal op-code");
                        writeResponsePacketToWire(response, xk, response.getOutputStream(),
                                        new ExhibitDataHTTPTunnelSource.RawPacket(
                                                ExhibitDataHTTPTunnelSource.RawPacket.OP__REMEX,
                                                ExhibitDataTunnelSource.EMPTY_PAYLOAD));
                        return;
                        }
                    }
                }

            // Note op type...
            opcode = Byte.valueOf(reqPacket.opCode);
            requestSize = reqPacket.plLength;
            StatsLogger.captureDataPoint(statsIDTSV, TSVNAMEPR_RPCTYPE + ((int) reqPacket.opCode));

            // Get the output stream (as late as possible).
            final ServletOutputStream os = response.getOutputStream();

            try
                {
                // Handle the request,
                // and serialise the result back onto the wire to the caller.
                final ExhibitDataTunnelSource.RawPacket rawResponse = ExhibitDataTunnelSource.handleInboundRPC(getDataSource(),
                                                                         reqPacket,
                                                                         remoteAddr,
                                                                         _RPC_cache);

                // Note when we've generated the response...
                responseGenerationTime = System.currentTimeMillis();
                responseSize = rawResponse.plLength;

                // Write the response onto the wire...
                writeResponsePacketToWire(response, xk, os, rawResponse);
                }
            // Generically handle common exception types with appropriate error responses.
            catch(final PGMasterNotInServiceException e)
                {
                writeResponsePacketToWire(response, xk, os,
                    new ExhibitDataHTTPTunnelSource.RawPacket(
                            ExhibitDataHTTPTunnelSource.RawPacket.OP__PGMNISEX,
                            ExhibitDataTunnelSource.EMPTY_PAYLOAD));
                }
            catch(final java.rmi.RemoteException e)
                {
                writeResponsePacketToWire(response, xk, os,
                    new ExhibitDataHTTPTunnelSource.RawPacket(
                            ExhibitDataHTTPTunnelSource.RawPacket.OP__REMEX,
                            ExhibitDataTunnelSource.EMPTY_PAYLOAD));
                }
            catch(final InterruptedIOException e)
                {
                writeResponsePacketToWire(response, xk, os,
                    new ExhibitDataHTTPTunnelSource.RawPacket(
                            ExhibitDataHTTPTunnelSource.RawPacket.OP__INTEX,
                            ExhibitDataTunnelSource.EMPTY_PAYLOAD));
                }
            catch(final RuntimeException e)
                {
if(IsDebug.isDebug) { e.printStackTrace(); }
                writeResponsePacketToWire(response, xk, os,
                    new ExhibitDataHTTPTunnelSource.RawPacket(
                            ExhibitDataHTTPTunnelSource.RawPacket.OP__RUNTEX,
                            ExhibitDataTunnelSource.EMPTY_PAYLOAD));
                }
            // A generic IOException is a "remote" exception to the caller...
            catch(final IOException e)
	            {
	            writeResponsePacketToWire(response, xk, os,
	                new ExhibitDataHTTPTunnelSource.RawPacket(
	                        ExhibitDataHTTPTunnelSource.RawPacket.OP__REMEX,
	                        ExhibitDataTunnelSource.EMPTY_PAYLOAD));
	            }

            // Make sure that all output has been sent back to the caller.
            os.flush();
            os.close();
            }
        catch(final Exception e)
            {
            // Note failure...
            StatsLogger.captureDataPoint(statsIDTSV, TSVNAME_HTTPFAIL);

            context.log("error handling tunnel user " + remoteAddr, e);
            e.printStackTrace();
            if(e instanceof IOException) { throw ((IOException) e); }
            throw new IOException("ERROR: tunnel failure: " + e.getMessage());
            }
        finally
            {
            final long endTime = System.currentTimeMillis();
            final long serviceTime = endTime - startTime;
            final int sTE = GenUtils.log2Approx(serviceTime);
            // Note service time bucket...
            StatsLogger.captureDataPoint(statsIDTSV, TSVNAMEP_HTTPSERVICETIME + sTE);

            // If the service time was long, note some details...
            if(serviceTime > MAX_TYPICAL_SERVICE_TIME_MS)
                {
                context.log("WARNING: TunnelServlet: long service time: "+serviceTime+"ms...");
                context.log("    Response set-up time: "+(initTime - startTime)+"ms.");
                if(opcode != null)
                    { context.log("    Request op-code: "+(opcode.byteValue() & 0xff)+"."); }
                if(opcode != null) // Note slow op...
                    { StatsLogger.captureDataPoint(statsIDTSV, TSVNAMEP_HTTPSLOWOP + (opcode.byteValue() & 0xff)); }
                if(validationTime != 0)
                    { context.log("    Validiation/auth time: "+(validationTime - initTime)+"ms."); }
                if(requestSize != 0)
                    { context.log("    Request payload size: "+(requestSize)+" bytes."); }
                if(requestReadTime != 0)
                    { context.log("    Request read time: "+(requestReadTime - validationTime)+"ms."); }
                if(responseGenerationTime != 0)
                    { context.log("    Response generation time: "+(responseGenerationTime - requestReadTime)+"ms."); }
                if(requestSize != 0)
                    { context.log("    Response payload size: "+(responseSize)+" bytes."); }
                if(responseGenerationTime != 0)
                    { context.log("    Response write time: "+(endTime - responseGenerationTime)+"ms."); }
                }
            }
        }

    /**Efficiently write the raw response frame/packet to the write, with a MAC if possible.
     * TODO: This may also log traffic volumes by type and destination.
     */
    private void writeResponsePacketToWire(final HttpServletResponse response,
                                           final SecretKey xk,
                                           final ServletOutputStream os,
                                           final ExhibitDataTunnelSource.RawPacket rawResponse)
        throws InvalidKeyException, IOException
        {
        response.setContentLength(rawResponse.getFrameLength());
        if(xk != null) // Create the MAC if we have the Xfer key.
            {
            final PacketProtector ppResp = new PacketProtector(rawResponse, xk);
            response.setHeader(ExhibitDataHTTPTunnelSource.HEADER_MAC, ppResp.toCheckString());
            assert(ppResp.length == rawResponse.getFrameLength());
            }
        rawResponse.writePacket(os);
        }

    /**Returns true iff the address is that of a (peer) mirror/slave.
     * Does this by looking at the IP addresses of inbound slave connections
     * as seen by the master, so as to capture the <em>outgoing</em>
     * IP addresses of each mirror.
     * <p>
     * No local/loopback addresses are considered valid.
     *
     * @param ds  current data source; never null
     * @param addr  client address; never null
     */
    private static boolean isPeerMirror(final DataSourceBean ds, final InetAddress addr)
        throws IOException
        {
        if(addr.isAnyLocalAddress() || addr.isLoopbackAddress() || addr.isLinkLocalAddress())
            { return(false); /* Local addresses not valid for this. */ }

        final SimpleVariableValue var = ds.getVariable(SystemVariables.TunnelServlet_SLAVE_ADDRS);
        if(var == null)
            { return(false); /* No extant slaves/mirrors at all. */ }

        final String addrS = addr.getHostAddress();
        if(addrS.equals(var.getValue()))
            { return(true); }

        final Map<InstanceID,SimpleVariableValue> globalMap = var.getGlobalMap();
        if(globalMap == null)
            { return(false); /* No extant slaves/mirrors at all. */ }

        for(final SimpleVariableValue v : globalMap.values())
            {
            // If we find this address as a String,
            // then this is a legitimate peer/mirror.
            if(addrS.equals(v.getValue()))
                { return(true); }
            }

        // Not a peer...
        return(false);
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -1276945887652232880L;

    /**Factory method to create an instance given the ServletContext.
     * Will fail if the CoreConsts.WAR_CTXTPARAM_BOOTURL param is not set.
     */
    public static ExhibitDataTunnelSource createFromContext(
                                                   final ServletContext ctxt)
        {
        if(ctxt == null)
            { throw new IllegalArgumentException("null ServletContext not allowed"); }

        final String masterURL = ctxt.getInitParameter(CoreConsts.WAR_CTXTPARAM_BOOTURL);
        if(masterURL == null)
            { throw new IllegalArgumentException("missing context-param " + CoreConsts.WAR_CTXTPARAM_BOOTURL); }

        final SimpleLoggerIF logger = new WebUtils.ServletLogger(ctxt);
        return(new ExhibitDataHTTPTunnelSource(masterURL, logger));
        }
    }
