/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.location;

import javax.servlet.ServletException;
import javax.servlet.http.*;

import java.io.IOException;
import java.io.OutputStream;

/**Servlet to serve dynamic location thumbnail icon (ie with crosshairs).
 * This expects two parameters, x and y, specifying the pixel coordinates
 * of the centre of the crosshairs, ie where the exhibit is.
 * <p>
 * If the parameters are missing or out of bounds then the thumbnail is
 * produced without crosshairs, possibly by redirection to the original
 * (source) raw thumbnail image.
 * <p>
 * The image produced is with the type indicated by DYN_MAP_IMG_TYPE,
 * in an efficient (small), indexed format.
 * <p>
 * Altered thumbnails may be cached, possibly with SoftReferences.
 */
public final class LocThumbnailServlet extends HttpServlet
    {
    /**Respond to a GET request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception java.io.IOException if an input/output error occurs
     * @exception javax.servlet.ServletException if a servlet error occurs
     */
    public void doGet(final HttpServletRequest request,
                      final HttpServletResponse response)
        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**Respond to a HEAD request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doHead(final HttpServletRequest request,
                       final HttpServletResponse response)
        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**Respond to a GET/HEAD request for the content served by this servlet.
     * Returns the exhibit data, with a correct MIME type.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doAction(final HttpServletRequest request,
                         final HttpServletResponse response)
        throws IOException, ServletException
        {
        try
            {
            // Attempt to retrieve x and y parameters.
            // If missing or unparsable or out of range throw an Exception.
            final int x = Integer.parseInt(request.getParameter("x"), 10);
            final int y = Integer.parseInt(request.getParameter("y"), 10);
            final byte[] encodedTN =
                LocationUtils.getLocTNWithCrosshairs(getServletContext(), x, y);

            // Set the correct MIME type.
            response.setContentType(LocationUtils.ETP.mimeType);
            response.setContentLength(encodedTN.length);

            // We expect this servlet to be mounted under /_static
            // and thus to have the Expires/Cache-Control headers set automagically.
            // If we attempt to do it here, then we may have confusion and dischord.

            // If this is a HEAD, then return without providing the body.
            if("HEAD".equalsIgnoreCase(request.getMethod()))
                { return; }

            // Write entire thumbnail in one go.
            final OutputStream os = response.getOutputStream();
            os.write(encodedTN);
            os.flush();
            }
        catch(final Exception e)
            {
            // Redirect back to original raw thumbnail in case of error.
            // If our result is already committed this may be messy...
            response.reset();
            response.resetBuffer();
            request.getRequestDispatcher(LocationUtils.BASE_ESTD_LOC_TN_URL).
                forward(request, response);
            return;
            }
        }

    /**Creation/load time of this class. */
    private static final long classLoadTime = System.currentTimeMillis();

    /**Have the last-modified date be the creation time of this class. */
    @Override protected long getLastModified(final HttpServletRequest request)
        { return(classLoadTime); }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -8253998074827856644L;
    }
