/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.location;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.ImageUtils;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletContext;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.lang.ref.SoftReference;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 01-Sep-2003
 * Time: 18:40:11
 */

/**Utilities to support Location display in Web pages.
 *
 */
public class LocationUtils
    {
    /**Width of (raw and modified) Estd thumbnail in pixels as returned by getEstdLocationThumbnailURL(). */
    public static final int LOC_TN_WIDTH = 32;

    /**Height of (raw and modified) Estd thumbnail in pixels as returned by getEstdLocationThumbnailURL(). */
    public static final int LOC_TN_HEIGHT = 22;

    /**Type of thumbnail image that we dynamically generate.
     * Need not be the same as the raw thumbnail image,
     * as long as it can (efficiently) encode the modified thumbnail
     * in a byte-indexed format.
     * <p>
     * Good types to use are PNG and GIF (when the LZW patent expires).
     */
    static final int DYN_LOC_TN_IMG_TYPE = ExhibitMIME.ET_PNG;

    /**ExhibitMIME entry for the chosen output type. */
    static final ExhibitMIME.ExhibitTypeParameters ETP =
        ExhibitMIME.getParamsByType(DYN_LOC_TN_IMG_TYPE);

    /**Colour of cross-hairs to show location of item in RGB format.
     * Since the icon is usually rendered in green (land) and blue (sea)
     * this should probably be red or some other high-contrast colour.
     * This should probably also be fully opaque to minimise encoded image size
     * (eg 0xffff0000 for opaque red).
     * <p>
     * Make this available to (for example) Aloha Earth,
     * so that we can use a common colour/style throughout.
     */
    public static final int DYN_LOC_TN_CROSSHAIR_RGB_COLOUR = 0xffff0000; // Opaque red.

    /**Basic root-relative URL for base thumbnail icon.
     * This has to be an image type that we can decode, modify, and encode
     * on the fly.
     * <p>
     * Assumed to be low-colour suitable for an index/palette representation
     * such as GIF or PNG.
     * <p>
     * By default rendered n green (land) and blue (sea).
     * <p>
     * ON OE is assumed to be centered in the image,
     * nominally between pixels for an even-pixel dimension.
     */
    static final String BASE_ESTD_LOC_TN_URL = "/_static/icon/Earth.png";

    /**Mount point for servlet, including usual suffix for given image type.
     * May mount this in the "/_static/" area to help clients cache the
     * modified thumbnails and thus reducing redundant requests for them.
     * <p>
     * This must correspond with where the servlet is actually mounted.
     * (The servlet can be mounted to serve a tree that includes this.)
     */
    private static final String DYN_LOC_TN_MOUNT_POINT =
        "/_static/locEstdTN/Earth" + ETP.dotSuffixForInputFile;


    /**Get cross-hair X,Y pixel centre on thumbnail given Estd location.
     * Since the dimension values are integral,
     * this rounds to the nearest correct pixel value.
     * <p>
     * The result is contrained to valid pixel coordinates
     * [0, width[ and [0, height[.
     */
    private static Dimension getTNCrosshairCentre(Location.Estd loc)
        {
        if(loc == null) { throw new IllegalArgumentException(); }

        final float ePos = loc.getE().value.floatValue() / Location.Estd.MAX_E;
        final float nPos = loc.getN().value.floatValue() / Location.Estd.MAX_N;

        final int x = Math.max(0, Math.min(LOC_TN_WIDTH - 1,
                               Math.round(((1 + ePos) / 2.0f) * LOC_TN_WIDTH)));
        // Note that y axis polarity is reverse of N dimension.
        final int y = Math.max(0, Math.min(LOC_TN_HEIGHT - 1,
                               Math.round(((1 - nPos) / 2.0f) * LOC_TN_HEIGHT)));

        return(new Dimension(x, y));
        }

    /**Return root-relative (starting with "/") URL for Estd location thumbnail.
     * May be fixed, or may depend on actual location.
     *
     * @param loc the location on Earth to link to the map for; never null
     * @param embedInHTML  if true, result will be embedded in HTML
     *     so we will separate parameters with '&amp;amp;' rather than
     *     an unprotected '&amp;'
     */
    public static String getEstdLocationThumbnailURL(final Location.Estd loc,
                                                     final boolean embedInHTML)
        {
        final Dimension dim = getTNCrosshairCentre(loc);
        final StringBuilder result =
            new StringBuilder(DYN_LOC_TN_MOUNT_POINT.length() + 16);
        result.append(DYN_LOC_TN_MOUNT_POINT);
        result.append("?x=").append(dim.width);
        if(embedInHTML) { result.append("&amp;"); } else { result.append('&'); }
        result.append("y=").append(dim.height);
        return(result.toString());
        }

    /**Returns a modified location thumbnail with crosshairs centred on x,y; never null.
     * This will create an image on first use,
     * and cache it (via a SoftReference).
     * <p>
     * The byte[] returned is not copied and so must not be altered.
     * This routine is only package-visible.
     * <p>
     * This routine is synchronized primarily to make the cache thread-safe,
     * but has the side-effect of ensuring that only one thumbnail
     * can be generated at once, thus limiting peak CPU and memory demands.
     *
     * @param context  context of current WAR; must not be null
     * @param x  x coordinate of crosshair centre within thumbnail
     * @param y  x coordinate of crosshair centre within thumbnail
     * @return encoded image for modeified thumbnail;
     *     this is a shared copy and must not be altered by the caller
     *     which is one reason that the routine is only package-visible
     * @throws java.lang.IllegalArgumentException  if context is null
     * @throws java.lang.RuntimeException  if x or y not valid pixel coords
     *     within the thumbnail, ie where the crosshair could be centred.
     * @throws java.io.IOException  if difficulty generating modified thumbnail
     *     for example because the site is too busy
     */
    static synchronized byte[] getLocTNWithCrosshairs(
                                                final ServletContext context,
                                                final int x, final int y)
        throws IllegalArgumentException,
        RuntimeException,
        IOException
        {
        byte result[];
        if((_gLTWC_cache[x][y] == null) ||
            ((result = (byte[]) _gLTWC_cache[x][y].get()) == null))
            {
            // Refuse to actually generate new thumbnail if overloaded.
            if(WebUtils.isOverloaded(context))
                { throw new IOException("too busy to generate thumbnail"); }

            final BufferedImage in =
                WebUtils.getAndCacheStaticImage(true, // Get a copy.
                                                BASE_ESTD_LOC_TN_URL,
                                                true, // Force to true colour.
                                                context,
                                                true); // Let this be GCed.

            // Draw in the crosshairs...
            // Do not overwrite at focus so that target pixel can be seen.
            for(int yc = LOC_TN_HEIGHT; --yc >= 0; )
                { if(yc != y) { in.setRGB(x, yc, DYN_LOC_TN_CROSSHAIR_RGB_COLOUR); } }
            for(int xc = LOC_TN_WIDTH; --xc >= 0; )
                { if(xc != x) { in.setRGB(xc, y, DYN_LOC_TN_CROSSHAIR_RGB_COLOUR); } }

            // Force to indexed image; four colours should be enough/efficient,
            // and may allow packing of at least 4 pixels to the byte.
            final BufferedImage out = ImageUtils.makeColourReducedBufferedImage(
                in, 4, true);

            // Generate and cache encoded image.
            // Don't attempt to fudge the quality.
            result = ETP.handler.makeImageBinary(out, Integer.MAX_VALUE);
            if(result == null) { throw new IOException(); }
            _gLTWC_cache[x][y] = new SoftReference<byte[]>(result);
            }
        return(result);
        }

    /**Private cache for getLocTNWithCrosshairs() accessed under the class lock.
     * Arranged as (x,y)-crosshair-centre array of
     * SoftReference to byte[] encoded images.
     */
    private static final SoftReference<byte[]> _gLTWC_cache[][] =
       new SoftReference[LOC_TN_WIDTH][LOC_TN_HEIGHT];


    /**Make HTML img for correct Estd location thumbnail with E/N crosshairs.
     * If there is no location then return the empty string.
     * <p>
     * The returned img has the alignment "top" and alt text "Aloha Earth!"
     * and no spaces round it.
     */
    public static String makeLocTNImgHTML(final Location.Estd location)
        {
        if(location == null) { return(""); }

        final int capacity = 140;
        final StringBuilder result = new StringBuilder(capacity);
        result.append("<img src=\"").
            append(LocationUtils.getEstdLocationThumbnailURL(location, true)).
            append('"');
        result.append(" width=").append(LOC_TN_WIDTH);
        result.append(" height=").append(LOC_TN_HEIGHT);
        result.append(" align=top alt=\"Aloha Earth!\" title=\"Aloha Earth!\">");
if(IsDebug.isDebug && ((result.length() > capacity) || (result.length() < capacity/2))) { System.err.println("WARNING: makeLocTNImgHTML(): len="+result.length()+" vs capacity="+capacity+"."); }
        return(result.toString());
        }
    }
