/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.threeD;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitAttrUtils;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.Tuple;
import org.hd.d.pg2k.svrCore.stats.StatsLogger;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;


/**3D Walkthrough data servlet.
 * Supports 3D walkthrough app by servicing its simple, small, incremental
 * data requests.
 * <p>
 * All requests are GET requests.
 * <p>
 * Also contains static client-side send/receive code
 * to keep everything in one place.
 */
public final class DataServlet extends HttpServlet
    {
    /**Unhook from servlet logger. */
    public void destroy()
        {
        logger.setContext(null); // Stop using servlet logger.
        super.destroy();
        }

    /**Hook into servlet logger. */
    public void init(final ServletConfig servletConfig)
        throws ServletException
        {
        logger.setContext(servletConfig.getServletContext()); // Use servlet.log().
        super.init(servletConfig);
        }

    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**The stats set to which we log general tunnel servlet stats.
     * The unique codes are the constants DSVNAME_XXX.
     */
    private final StatsLogger.StatsConfig statsIDDSV =
        new StatsLogger.StatsConfig("3DWTSERVLET",
                                    logger, // Use servlet log if poss.
                                    false, // Only dump summaries...
                                    12 * 3600, // About every 12 hours.
                                    true); // Adaptive.

    /**General stats event name: inbound HTTP request. */
    public static final String DSVNAME_HTTPREQUEST = "HTTP-request";

    /**General stats event name: failures handling input HTTP requests. */
    public static final String DSVNAME_HTTPFAIL = "HTTP-failure";

    /**General stats event name: inbound HTTP request rejected by request op-code. */
    public static final String DSVNAME_HTTPREJOP = "HTTP-reqRejectedByOp";





    /**Get singleton (per-servlet-context) data pipeline/cache instance.
     * The config param must not be null, but for some operations
     * (such as calling destroy()) request can be null.
     * <p>
     * This does not cache its return value.
     */
    private static DataSourceBean getDataSourceBean(
            final ServletContext ctxt,
            final HttpServletRequest request)
        {
        // Fetches/creates the data source...
        final DataSourceBean dataSource =
            DataSourceBean.getApplicationInstance(ctxt);

        // Ensure that the essential details are set up.
        dataSource.setServletContext(ctxt);
        if(request != null)
            { dataSource.setContextPath(request.getContextPath()); }

        return(dataSource);
        }


    /**Respond to a GET request from 3DWT clients.
     *
     * @exception IOException if an input/output error occurs
     */
    public void doGet(final HttpServletRequest request,
                      final HttpServletResponse response)
        throws IOException
        {
        // Don't allow this response to be cached.
        // HTTP/1.0 headers.
        response.setHeader("Expires", "0");
        response.setHeader("Pragma", "no-cache");
        // HTTP/1.1 headers.
        response.setHeader("Cache-Control", "no-cache");

//if(IsDebug.isDebug) { System.out.println("[Received 3DWT request...]"); }
//if(IsDebug.isDebug) { logger.log("[Received 3DWT request...]"); }

        // Note HTTP the request...
        StatsLogger.captureDataPoint(statsIDDSV, DSVNAME_HTTPREQUEST);

        final DataSourceBean ds = getDataSourceBean(getServletContext(), request);

        // Get the current set of 3DWT exhibit data...
        final Tuple.Pair<LightweightMetaDataFetchInterface.GalleryBasicMetaData, List<String>> data = getData(ds);

        // The (printable 7-bit ASCII) result string...
        final String result;

        // Get operation type.
        final String opType = request.getParameter(Utils.HP_OPCODE);
        if(Utils.OP_LUS.equals(opType)) // Most common request, so do first.
            {
            // Request for name lookup.
            try
                {
                final List<String> names = data.second;
                final String ordinals = request.getParameter(Utils.HP_ORD);
                final String[] nums = ordinals.split(",");
                // Ignore requested items beyond the maximum allowed per call.
                final int nResults = Math.min(nums.length, LightweightMetaDataFetchInterface.MAX_NAMES_REQUEST);
                final String results[] = new String[nResults];
                int len = nResults-1; // For the separators...
                for(int i = nResults; --i >= 0; )
                    {
                    final String n = names.get(Integer.parseInt(nums[i], 10));
                    if(n == null) { throw new IOException("bad index"); }
                    results[i] = n;
                    len += n.length();
                    }
                // Separate the results with commas, like the inputs.
                // Null items are not allowed.
                final StringBuilder sb = new StringBuilder(len);
                for(int i = 0; i < nResults; ++i)
                    {
                    if(i != 0) { sb.append(','); }
                    sb.append(results[i]);
                    }
                result = sb.toString();
                }
            catch(final Exception e)
                {
                // Any parameter problem generates a tidy error...
                response.sendError(HttpServletResponse.SC_BAD_REQUEST);
                return;
                }
            }
//        else if(Utils.OP_LU.equals(opType)) // Next most common request...
//            {
//            // Request for name lookup.
//            try
//                {
//                final List<String> names = data.second;
//                result = names.get(Integer.parseInt(request.getParameter(Utils.HP_ORD), 10));
//                }
//            catch(final Exception e)
//                {
//                // Any parameter problem generates a tidy error...
//                response.sendError(HttpServletResponse.SC_BAD_REQUEST);
//                return;
//                }
//            }
        else if(Utils.OP_MD.equals(opType))
            {
            // Request for metadata.
            result = data.first.toWireString();
            }
        else
            {
            // Missing/wrong op-type generates an error.
            response.sendError(HttpServletResponse.SC_BAD_REQUEST);
            return;
            }

        // Write (printable 7-bit ASCII) output in one operation.
        final int length = result.length();
        response.setContentLength(length);
        final byte buf[] = new byte[length];
        for(int i = length; --i >= 0; ) { buf[i] = (byte) result.charAt(i); }
        final ServletOutputStream os = response.getOutputStream();
        os.write(buf);
        os.flush();

//if(IsDebug.isDebug) { System.out.println("[Replied to 3DWT request.]"); }
//if(IsDebug.isDebug) { logger.log("[Replied to 3DWT request.]"); }
        }


    /**Get the filtered data associated with the current DataSourceBean; never null.
     * A new value is created on first use and when the AEP changes.
     * <p>
     * This call should always be very fast
     * except when doing the initial filtering/construction of the exhibit list
     * when there would be a danger of races and redundant resource usage,
     * so we serialise calls to eliminate this possibility.
     */
    private static synchronized Tuple.Pair<LightweightMetaDataFetchInterface.GalleryBasicMetaData, List<String>> getData(final DataSourceBean dsb)
        throws IOException
        {
        assert(dsb != null);

        Tuple.Pair<LightweightMetaDataFetchInterface.GalleryBasicMetaData, List<String>> result;
        while((result = (Tuple.Pair<LightweightMetaDataFetchInterface.GalleryBasicMetaData, List<String>>) dsb.getAEPLinkedValue(WTKey)) == null)
            {
            final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
            final String[] rawNames = aep.select(new AllExhibitProperties.AEPFilter()
                {
                /**Only accept results that the 3DWT app can handle.
                 * @return  true if the exhibit is suitable
                 */
                public final boolean accept(final AllExhibitProperties aep,
                                            final String exhibitName)
                    {
                    final ExhibitMIME.ExhibitTypeParameters exhibitType = ExhibitMIME.getInputFileType(exhibitName);
                    if(!Utils.supportedTypeFor3DWT(exhibitType)) { return(false); }
                    return(true);
                    }
                }, null, 0);
            Arrays.sort(rawNames, ExhibitAttrUtils.getAttrWords().SMART_ORDER);
            // Create read-only smart-sorted exhibit name List...
            final List<String> exhibits = Collections.unmodifiableList(Arrays.asList(rawNames));
            dsb.putIfAbsentAEPLinkedValue(WTKey,
                new Tuple.Pair<LightweightMetaDataFetchInterface.GalleryBasicMetaData, List<String>>(
                    new LightweightMetaDataFetchInterface.GalleryBasicMetaData(aep.aeid.timestamp ^ exhibits.size(), exhibits.size(), CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 500),
                    exhibits));
            }

        return(result);
        }

    /**Key for 3DWT-linked data in DataSourceBean. */
    private static final DataSourceBean.AEPLinkedKey WTKey = new DataSourceBean.AEPLinkedKey("3DWT");

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -1276945667652232880L;
    }
