/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.threeD;

import java.io.IOException;
import java.util.StringTokenizer;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 11-Mar-2006
 * Time: 13:22:06
 */

/**This interface is used for light-weight, incremental fetching of Gallery metadata, eg for an interactive 3D walkthrough.
 */
interface LightweightMetaDataFetchInterface
    {
    /**Immutable store of the most basic Gallery meta-data.
     * This contains a hash over the Gallery contents
     * that should change whenever the set of exhibits in the Gallery changes,
     * and a (non-negative) count of exhibits in the Gallery.
     * <p>
     * This also contains an expected maximum validity period
     * for this and related mata-data fetched through this interface,
     * ie within what short of time it should be revalidated or discarded.
     */
    public static final class GalleryBasicMetaData
        {
        /**Hash over exhibit set. */
        public final long exhibitSetHash;

        /**Count of exhibits in the Gallery; non-negative. */
        public final int exhibitCount;

        /**Validity period (maximum cache time) in milliseconds; non-negative. */
        public final int metaDataMaxCacheLifeMs;

        /**Create an instance.
         * The exhibit count and cache time must be non-negative.
         */
        public GalleryBasicMetaData(final long hash,
                                    final int count,
                                    final int cacheMs)
            {
            exhibitSetHash = hash;
            exhibitCount = count;
            metaDataMaxCacheLifeMs = cacheMs;

            if(exhibitCount < 0) { throw new IllegalArgumentException(); }
            if(metaDataMaxCacheLifeMs < 0) { throw new IllegalArgumentException(); }
            }

        /**Create an instance parsed from on-the-wire format; never null.
         * @param s  value as generated by toWireString(); never null
         * @return  instance; never null
         * @throws IllegalArgumentException  if input String is invalid
         */
        public static GalleryBasicMetaData fromWireString(final String s)
            throws IllegalArgumentException
            {
            final StringTokenizer st = new StringTokenizer(s, ":");
            final long hash = Long.parseLong(st.nextToken(), 10);
            final int count = Integer.parseInt(st.nextToken(), 10);
            final int cacheTime = Integer.parseInt(st.nextToken(), 10);
            return(new GalleryBasicMetaData(hash, count, cacheTime));
            }

        /**Convert to String for suitable to use on the wire; pure printable ASCII. */
        public String toWireString()
            {
            final StringBuilder sb = new StringBuilder(64);
            sb.append(exhibitSetHash).
                append(':').append(exhibitCount).
                append(':').append(metaDataMaxCacheLifeMs);
            return(sb.toString());
            }

        /**Human-readable string format. */
        public String toString() { return(toWireString()); }

        /**Default/empty instance. */
        public static final GalleryBasicMetaData EMPTY = new GalleryBasicMetaData(0, 0, 0);
        }

    /**Fetch the current exhibit-set hash and exhibit count; never null.
     * A change in the hash is a signal that cached data may be invalid
     * and need to be purged or revalidated.
     */
    public GalleryBasicMetaData getGalleryBasicMetaData() throws IOException;

    /**Fetch the full name of the given numbered exhibit; null if no such exhibit or not currently available.
     * The range is 0 to numberOfExhibits-1.
     * <p>
     * This assumes that there is a sensible stable ordering of exhibits,
     * probably in "smart-sorted" order or similar.
     * <p>
     * This ordering is fixed from one exhibit set to the next,
     * ie is stable while the exhibit set hash remains unchanged.
     */
    public String getExhibitName(final int ordinal);

    /**Maximum number of names that can be requested at once with getExhibitNames(); strictly positive.
     * High enough to help overcome network latency,
     * but not large enough to take significant CPU time or response size.
     */
    public static final int MAX_NAMES_REQUEST = 32;

    /**Fetch the full name of the given numbered exhibits; each entry null if no such exhibit or not currently available, overall result never null.
     * The range for each item is 0 to numberOfExhibits-1.
     * <p>
     * This assumes that there is a sensible stable ordering of exhibits,
     * probably in "smart-sorted" order or similar.
     * <p>
     * This ordering is fixed from one exhibit set to the next,
     * ie is stable while the exhibit set hash remains unchanged.
     */
    public String[] getExhibitNames(final int ordinals[]);
    }
