/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.threeD;

import ORG.hd.d.IsDebug;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.pickfast.PickCanvas;
import com.sun.j3d.utils.universe.*;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import javax.jnlp.SingleInstanceListener;
import javax.media.j3d.*;
import javax.swing.*;
import javax.vecmath.*;
import java.awt.*;
import java.awt.event.*;
import java.net.URL;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.concurrent.atomic.AtomicInteger;


/**Main (UI) class of JWS-based 3D walkthrough.
 * Runs as a Swing App.
 * <p>
 * This class/file contains as little non-UI code as is reasonably practical,
 * so that if we change the UI details then other classes should be unaffected.
 * <p>
 * We don't really want/need this class to be Serializable,
 * but this class inherits Serializable from JFrame.
 */
public final class ThreeDMain extends JFrame // implements ActionListener
    {
    /**Central logger instance for uploader; never null.
     * This instance may log to the status bar and elsewhere.
     */
    private final SimpleLoggerIF logger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            {
            // Log to the Java console.
            System.out.println(message);

            // Schedule a job for the event-dispatching thread:
            // logging this message in the status bar.
            // We hope that these events do not get re-ordered too often.
            SwingUtilities.invokeLater(new Runnable(){
                public final void run()
                    { status.setText(message); }
                });
            }
        };

    /**Our companion "business-logic" class; never null. */
    private final ThreeDLogic logic = new ThreeDLogic(logger);

//    /**The action performed by the "About" menu entry. */
//    private final AboutAction aboutAction = new AboutAction();
//    /**The action performed by the "Exit" menu entry. */
//    private final ExitAction exitAction = new ExitAction();

//    /**Select our locale; never null. */
//    private final LocaleBeanBase lbb = new LocaleBeanBase();

    /**Status bar; never null. */
    private final JLabel status = createStatusBar();

    /**Single listener instance. */
    private final SISListener sisListener = new SISListener();

//    /**Handles Mouse over messages on toolbar buttons and menu items; never null. */
//    private final MouseHandler mouseHandler = new MouseHandler(status);

    /**Title shown for application. */
    private static final String APPLICATION_WINDOW_TITLE = "Gallery 3D Walkthrough";

    /**Our simple universe; only null if Java3D not available or set-up not yet complete.
     * Marked volatile for thread-safe lock-free access.
     */
    private volatile SimpleUniverse simpleUniverse;

    /**3D canvas for Java3D to draw on; only null if Java3D not available. */
    private final Canvas3D canvas3D;

    /**BranchGroup used to hold the exhibits; never null.
     * This is set up to be writable/extendable at runtime
     * so that we can update the exhibits scene as needed.
     */
    private final BranchGroup exhibitsBranchGroup = new BranchGroup();

    /**Allow exhibit children to be added/removed at run-time. */
    {
//    exhibitsBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    exhibitsBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    exhibitsBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    }

    /**X-axis slider; never null. */
    private final JSlider sliderX = new JSlider(JSlider.HORIZONTAL);

    /**Y-axis slider; never null. */
    private final JSlider sliderY = new JSlider(JSlider.VERTICAL);

    /**Z-axis slider; never null. */
    private final JSlider sliderZ = new JSlider(JSlider.VERTICAL);


    /**Create an instance of the Uploader app main window.
     * Designed to be called by main().
     */
    private ThreeDMain()
        {
        super(APPLICATION_WINDOW_TITLE);

//        // Set up user actions.
//        initActions();

        // ASAP, try to avoid multiple instances being started.
        if(logic.sis != null)
            { logic.sis.addSingleInstanceListener(sisListener); }

        // Try to create 3D canvas.
        Canvas3D c3D = null;
        try
            {
            final GraphicsConfiguration config = SimpleUniverse.getPreferredConfiguration();
            c3D = new Canvas3D(config);
            }
        catch(final Throwable t)
            {
            logger.log("Unable to create Canvas3D...");
            t.printStackTrace();
            }
        canvas3D = c3D;

        // Create the contents of this frame...
//        setJMenuBar(createMenu());
        final JComponent mainPane = create3DPane();
        getContentPane().add(mainPane, BorderLayout.CENTER);
        getContentPane().add(status, BorderLayout.SOUTH);

        // Arrange to exit the VM if the window is closed...
        addWindowListener(new WindowAdapter()
            {
            public final void windowClosing(final WindowEvent evt)
                {
                // Shut down gracefully (and exit()) if allowed to,
                // else attempt to veto with an exception.
                try { shutdown(); }
                catch(final UnsupportedOperationException e) { }
                }
            });
        }

    /**Perform any activity required to shut down cleanly, eg save state, then exit.
     * This should try to avoid taking a long time.
     * <p>
     * This may throw an UnsupportedOperationException to try to veto
     * an exit that the user changes their mind about,
     * eg because we have work in progress!
     *
     * @throws UnsupportedOperationException  if the user vetoes the shut-down
     */
    private void shutdown()
        throws UnsupportedOperationException
        {
        // Do any shutdown required by the non-GUI components...
        logic.shutdown();

        // Now allow other instances to be started...
        if(logic.sis != null)
            { logic.sis.removeSingleInstanceListener(sisListener); }

        // Now perform a gracefull exit!
        System.exit(0);
        }

//    /**This method should be called before creating the UI to create all the Actions. */
//    private void initActions()
//        {
//        registerAction(aboutAction);
//        registerAction(exitAction);
//        }
//
//    private void registerAction(final JLFAbstractAction action)
//        {
//        action.addActionListener(this);
////        actions.addElement(action);
//        }

//    /**Create the application menu bar. */
//    private JMenuBar createMenu()
//        {
//        final JMenuBar menuBar = new JMenuBar();
//
//        // Build the File menu
//        final JMenu fileMenu = new JMenu("File");
//        fileMenu.setMnemonic('F');
//        fileMenu.add(exitAction).addMouseListener(mouseHandler);
//
//        // Build the help menu
//        final JMenu helpMenu = new JMenu("Help");
//        helpMenu.setMnemonic('H');
//        helpMenu.add(aboutAction).addMouseListener(mouseHandler);
//
//        menuBar.add(fileMenu);
//        menuBar.add(helpMenu);
//
//        return menuBar;
//        }

    /**Canvas width in pixels. */
    private static final int CANVAS3D_WIDTH = 512;

    /**Canvas height in pixels. */
    private static final int CANVAS3D_HEIGHT = 512;

    /**Create the (main) 3D component, called during construction; never null. */
    private JPanel create3DPane()
        {
        final JPanel result = new JPanel();
        result.setLayout(new BorderLayout());
        if(canvas3D != null)
            {
            canvas3D.setSize(CANVAS3D_WIDTH, CANVAS3D_HEIGHT);
            result.add(canvas3D, BorderLayout.CENTER);
            result.add(sliderX, BorderLayout.SOUTH);
            result.add(sliderY, BorderLayout.EAST);
            result.add(sliderZ, BorderLayout.WEST);

            // Some minimal decoration...
            sliderX.setToolTipText("X-axis");
            sliderY.setToolTipText("Y-axis");
            sliderZ.setToolTipText("Z-axis");
            }
        else
            {
            result.add(new JLabel("Unable to start Java3D"), BorderLayout.CENTER);
            result.setPreferredSize(new Dimension(CANVAS3D_WIDTH, CANVAS3D_HEIGHT));
            logger.log("java.library.path=" + System.getProperty("java.library.path"));
            }
        return(result);
        }
    /**If true, animate the name banner (at the cost of CPU time, etc). */
    private static final boolean ANIMATE_BANNER = false;

    /**Maximum exhibit dimension (eg height or width) in metres; strictly positive.
     * A value of around 1.0m, eg less than the viewer's nominal height, may be good.
     */
    public static final float MAX_EXHIBIT_DIM_M = 1.0f;

    /**Exhibit (centres) spacing in metres; strictly positive and no smaller than MAX_EXHIBIT_DIM_M.
     * A value of around twice MAX_EXHIBIT_DIM_M, may be good.
     */
    public static final float EXHIBIT_CSPACING_M = 3.0f * MAX_EXHIBIT_DIM_M;

    /**Exhibit visibility spacing in metres; strictly positive and no smaller than MAX_EXHIBIT_DIM_M.
     * When out of this radius we can save CPU and memory
     * by reverting to a fall-back view.
     * <p>
     * We ensure that the nearest exhibit to the user
     * is visible in their initial view.
     */
    public static final float EXHIBIT_VISIBLE_M = Math.max(15, // At least 15m view...
                   Math.max(3*MAX_EXHIBIT_DIM_M, 4.5f*EXHIBIT_CSPACING_M)); // See ~4 exhibits away.

    /**Distance at which fog starts to have an effect, in open range ]0, EXHIBIT_VISIBLE_M[. */
    private static final float FOG_START = Math.min(2.5f * EXHIBIT_CSPACING_M,
                                                    0.8f * EXHIBIT_VISIBLE_M);

//    /**Distance at which objects are considered barely visible, in open range ]FOG_START, EXHIBIT_VISIBLE_M[. */
//    private static final float BARELY_VISIBLE = (FOG_START + 3*EXHIBIT_VISIBLE_M) / 4;

    /**Colour of text for captions, etc, logically immutable; never null. */
    private static final Color3f TEXT_COLOUR = new Color3f(1,1,0);

    /**Alternate/highlight/not-good colour of text/othe3r for captions, etc, logically immutable; never null. */
    private static final Color3f TEXT_ALT_COLOUR = new Color3f(1,0.25f,0.25f);

    /**Alternate/highlight/not-good colour of text/othe3r for captions, etc, logically immutable; never null. */
    private static final Color3f TEXT_ALT2_COLOUR = new Color3f(1,0.5f,0.5f);

    /**Alternate/highlight/not-good colour of text/othe3r for captions, etc, logically immutable; never null. */
    private static final Color3f TEXT_ALT3_COLOUR = new Color3f(1,0.8f,0.8f);

    /**Alternate/highlight/OK colour of text/other for captions, etc, logically immutable; never null. */
    private static final Color3f TEXT_OK_COLOUR = new Color3f(0.25f,1,0.25f);

//    /**Background colour of text for captions, etc, logically immutable; never null. */
//    private static final Color3f TEXT_BG_COLOUR = new Color3f(0,0,0.25f);

    /**Standard text colour/appearance, logically immutable; never null. */
    private static final Appearance TEXT_APPEARANCE = new Appearance();
    /**Initialise textAppearance. */
    static
        {
        final ColoringAttributes textColouring = new ColoringAttributes(TEXT_COLOUR, ColoringAttributes.SHADE_FLAT);
        TEXT_APPEARANCE.setColoringAttributes(textColouring);
        }

//    /**Standard text background colour/appearance, logically immutable; never null. */
//    private static final Appearance TEXT_BG_APPEARANCE = new Appearance();
//    /**Initialise textAppearance. */
//    static
//        {
//        final ColoringAttributes textBgColouring = new ColoringAttributes(TEXT_BG_COLOUR, ColoringAttributes.SHADE_FLAT);
//        TEXT_BG_APPEARANCE.setColoringAttributes(textBgColouring);
//        }

    /**Creates the PG2K banner branch graph; never null.
     * This sits just above (0,0,0), is 3D text, is about human height,
     * and may have some animation effects...
     * <p/>
     * Derived from the "HelloUniverse" example!
     * <p>
     * We add some other misc bits of the scene here, such as fog.
     */
    private static BranchGroup createMainBannerSceneGraph(final Point3f bottomRight)
        {
        // Create the root of the branch graph
        final BranchGroup objRoot = new BranchGroup();

        // Create the transform group node and initialize it to the
        // identity.  Enable the TRANSFORM_WRITE capability so that
        // our behavior code can modify it at runtime.  Add it to the
        // root of the subgraph.
        final TransformGroup objTrans = new TransformGroup();
        if(ANIMATE_BANNER)
            { objTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE); }
        objRoot.addChild(objTrans);


        // Create a text node...
        final int textHeightMetres = 1;
        final Text3D bannerText = new Text3D(new Font3D(new Font("Serif", Font.PLAIN, textHeightMetres), new FontExtrusion()),
                                            CoreConsts.MAIN_DATA_HOST,
                                            bottomRight,
                                            Text3D.ALIGN_LAST,
                                            Text3D.PATH_RIGHT);
        final Appearance textAppearance = new Appearance();
        final ColoringAttributes textColouring = new ColoringAttributes(1,0,0, ColoringAttributes.SHADE_GOURAUD);
        textAppearance.setColoringAttributes(textColouring);
        final Shape3D textShape = new Shape3D(bannerText, textAppearance);
        objTrans.addChild(textShape);

        if(ANIMATE_BANNER)
            {
            // Create a new Behavior object that will perform the desired
            // operation on the specified transform object and add it into
            // the scene graph.
            final Transform3D yAxis = new Transform3D();
            final Alpha rotationAlpha = new Alpha(-1, Alpha.INCREASING_ENABLE,
                                            0, 0,
                                            8000, 0, 0,
                                            0, 0, 0);

            final RotationInterpolator rotator =
                new RotationInterpolator(rotationAlpha, objTrans, yAxis,
                                         0.0f, (float) Math.PI * 2.0f)
                /*  {
                    @Override public final void processStimulus(final Enumeration enumeration)
                        {
                        super.processStimulus(enumeration);
if(IsDebug.isDebug) { System.err.println("Rotator got stimulus..."); }
                        }
                    } */ ;
            // Don't waste CPU time animating banner when out of sight...
            // Keep this to a minimal area to intersect with the view platform.
            final BoundingSphere bounds =
                new BoundingSphere(new Point3d(), EXHIBIT_VISIBLE_M);
            rotator.setSchedulingBounds(bounds);
            objTrans.addChild(rotator);
            }


        // Create influencing bounds for fog.
        final Bounds fogBounds = new BoundingSphere(new Point3d(), Double.POSITIVE_INFINITY);
        // Set the fog colour, density, and its influencing bounds
        final Color3f FOG_COLOUR = new Color3f(); // Black, to match backdrop.
        final Fog fog = new LinearFog(FOG_COLOUR, FOG_START, EXHIBIT_VISIBLE_M);
        // front and back distances set below
//        fog.setCapability(Fog.ALLOW_INFLUENCING_BOUNDS_WRITE);
        fog.setInfluencingBounds(fogBounds);
        objRoot.addChild(fog);


        // Have Java 3D perform optimisations on this scene graph.
        objRoot.compile();

        return(objRoot);
        }


    /**Create one exhibit's "space" and "sphere of influence"; never null.
     * This centres the exhibit on the point given,
     * and surrounds it with concentric bounds:
     * <ol>
     * <li>An inner bound at which the exhibit can be seen/heard/etc in hi-fi.
     * <li>A bound at which the exhibit can be seen, maybe at lo-fi,
     *     and may have the hi-fi version (pre-)fetched as needed.
     * <li>An outer bound at which the exhibit is pre-fetched at lo-fi
     *     and possibly seen.
     * </ol>
     * <p>
     * When outside the lo-red bounds the exhibit may be invisible
     * and indeed only the scheduling bounds may exist to conserve memory.
     *
     * @param centre  the centre of the exhibit and exhibit's space; never null
     * @param index  the index in the server's ordered list of exhibits;
     *     non-negative
     */
    private Node makeExhibitSpace(final Point3f centre,
                                  final int index)
        {
        assert((centre != null) && (index >= 0));

        final Point3d centre3d = new Point3d(centre);

        // BranchGroup of reconstructable/discardable items
        // that exist only when in view
        // to save memory and CPU time.
        final BranchGroup discardable = new BranchGroup();
        discardable.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
        discardable.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);

        final Behavior appearanceBehaviour = new Behavior()
            {
            /**Bounds within which exhibit thumbnail texture is displayed at all.
             * Outside this area the exhibit cannot nominally be seen at all.
             * <p>
             * This is also the scheduling bounds for each exhibit.
             */
            private final BoundingSphere BOUNDS_LORES = new BoundingSphere(centre3d, EXHIBIT_VISIBLE_M);

            /**Bounds within which exhibit thumbnail can be seen in high-res.
             * This is mainly the unique space around the exhibit
             * though there may be some overlap.
             */
            private final BoundingSphere BOUNDS_HIRES = new BoundingSphere(centre3d, EXHIBIT_CSPACING_M*0.8f);

            /**Normal low-res visibility entry criterion. */
            private final WakeupOnViewPlatformEntry loResWakeupOnViewPlatformEntry = new WakeupOnViewPlatformEntry(BOUNDS_LORES);

            /**Normal high-res visibility entry criterion. */
            private final WakeupOnViewPlatformEntry hiResWakeupOnViewPlatformEntry = new WakeupOnViewPlatformEntry(BOUNDS_HIRES);

            /**Normal low-res visibility exit criterion. */
            private final WakeupOnViewPlatformExit loResWakeupOnViewPlatformExit = new WakeupOnViewPlatformExit(loResWakeupOnViewPlatformEntry.getBounds());

            /**Normal high-res visibility exit criterion. */
//            private final WakeupOnViewPlatformExit hiResWakeupOnViewPlatformExit = new WakeupOnViewPlatformExit(BOUNDS_HIRES);

            /**Timer wakeup for when we are waiting to set the hi-res texture; we try to do this quickly since it is close to the user. */
            private final WakeupOnElapsedTime hiResWakeupOnElapsedTime = new WakeupOnElapsedTime(CoreConsts.MAX_INTERACTIVE_DELAY_MS/2 + Rnd.fastRnd.nextInt(CoreConsts.MAX_INTERACTIVE_DELAY_MS));

            /**Timer wakeup for when we are waiting to set the lo-res texture; we can wait longer for this and induce less "polling" CPU load. */
            private final WakeupOnElapsedTime loResWakeupOnElapsedTime = new WakeupOnElapsedTime(CoreConsts.MAX_INTERACTIVE_DELAY_MS*4 + Rnd.fastRnd.nextInt(CoreConsts.MAX_INTERACTIVE_DELAY_MS*2));

            /**At initialisation. */
            public final void initialize()
                {
                wakeupOn(loResWakeupOnViewPlatformEntry);
                }

            /**Return TRUE if object is definitely NOT visible, for optimisation purposes.
             * Return FALSE if not sure, null is partially/peripherally visible.
             * <p>
             * This is used to optimise behaviours, texture fetches, etc.
             *
             * @return TRUE if object is definitely NOT visible,
             *         null if probably not fully visible,
             *         FALSE if definitely visible (or if not sure)
             */
            private Boolean isNotVisibleToUser()
                {
                final VirtualUniverse virtualUniverse = simpleUniverse.getLocale().getVirtualUniverse();
                if((virtualUniverse instanceof SimpleUniverse) && (front != null))
                    {
                    // We can get the view-platform this way...
                    final SimpleUniverse su = (SimpleUniverse) virtualUniverse;
                    final Vector3d v3dUser = computeViewPlatformVWorldXYZ(su);
//System.out.println("USER Z = " + v3dUser.z);

                    // Get exhibit central Z coordinate...
                    final Transform3D t3dMe = new Transform3D(); // My transform...
                    front.getLocalToVworld(t3dMe);
                    final Vector3d v3dMe = new Vector3d(); // My centre...
                    t3dMe.get(v3dMe);
//System.out.println("FRONT CENTRAL Z = " + v3dMe.z);

                    // If the user is in front of (more -ve z) than box centre
                    // then regard this box as not visible to the user.
                    // We can only do this because the user is not permitted
                    // to turn their view sideways/backwards.
                    //
                    // This should actually be slightly conservative
                    // given that the back of each box is not interesting,
                    // and the near view-clipping plane is ahead of the user.
                    if(v3dUser.z < v3dMe.z - EXHIBIT_CSPACING_M)
                        {
//System.out.println("BOX NOT VISIBLE (behind user): "+index);
                        return(Boolean.TRUE);
                        }

                    // Establishing if exhibit is "peripherially" visible,
                    // ie should have less computational effort spent on it.

                    // If the distance is such that the exhibit (centre)
                    // is not fully visible
                    // then we can return null
                    // to reduce effort spent on peripheral items.
                    final Vector3d diff = new Vector3d(v3dUser);
                    diff.sub(v3dMe);
                    final double lenSq = diff.lengthSquared();
                    if(lenSq > FOG_START*FOG_START)
                        { return(null); }

                    // Crudely, if out of the centre of vision and not v close
                    // computed as x or y distance approaching z distance,
                    // ie about 45 degrees from view platform normal,
                    // then we can regard this as "peripheral".
                    if((lenSq > 2*EXHIBIT_CSPACING_M*EXHIBIT_CSPACING_M) &&
                       (Math.max(Math.abs(diff.x), Math.abs(diff.y)) > Math.abs(diff.z)/2))
                        { return(null); }

                    // Definitely fully visible, so far as we can tell.
                    return(Boolean.FALSE);
                    }

                return(null); // Don't know, so assume may be partly visible.
                }

            /**Front face of exhibit box to which exhibit is applied; null initially and when not in use (not visible).
             * Marked volatile for thread-safe lock-free access.
             */
            private volatile Shape3D front;

            /**Load state indicator; null initially and when not in use (not visible).
             * Marked volatile for thread-safe lock-free access.
             */
            private volatile Sphere indicatorSphere;

            /**Caption text Node; null initially and when not in use (not visible).
             * Marked volatile for thread-safe lock-free access.
             */
            private volatile Group tgCaption;

            /**Accept stimuli.
             * When we are far enough from the viewer to be invisible
             * then we set the texture to null to save rendering and memory.
             */
            public final void processStimulus(final Enumeration stimuli)
                {
                boolean hasEnteredHiResArea = false;

                while(stimuli.hasMoreElements())
                    {
                    final WakeupCriterion w = (WakeupCriterion) stimuli.nextElement();

                    // Regard any activity, other than timer events,
                    // as evidence of user activity.
                    if(!(w instanceof WakeupOnElapsedTime))
                        { logic.setUserLastActive("Received stimulus"); }

                    // If exiting the viewable area altogether
                    // then discard all other easily-reconstucatble state
                    // to save memory and rendering CPU cycles,
                    // and only wake up on the next entry.
                    // This should be present as a possible condition
                    // in all other cases so that we can release resources.
                    //
                    // This idea to drastically reduce memory consumption
                    // was suggested by the first of the famous Punch limericks:
                    //
                    //     There once was a man who said, "God
                    //     Must think it exceedingly odd
                    //     If he finds that this tree
                    //     Continues to be
                    //     When there's no one about in the Quad."
                    //
                    // Here only the vestigial Behavior of the tree
                    // exists when no one is looking!
                    if(w == loResWakeupOnViewPlatformExit)
                        {
                        // Discard all visible elements.
                        discardable.removeAllChildren();
                        // Drop references to items in the discardable group
                        // to enable GC to do its stuff...
                        indicatorSphere = null;
                        front = null;
                        tgCaption = null;

//if(IsDebug.isDebug) { System.err.println("ZAPPED VIEWABLES FOR #"+index); }

                        // Avoid processing any more events this round,
                        // such as timed wakeups which could refetch textures.
if(false && IsDebug.isDebug && stimuli.hasMoreElements())
    {
    logger.log("WARNING discarding excess stimuli after exiting lo-res bounds.");
    while(stimuli.hasMoreElements())
        { logger.log("  DISCARDING: " + stimuli.nextElement()); }
    }
                        // Sleep until next entry into viewing space,
                        // even skipping straight into hi-res space.
                        wakeupOn(new WakeupOr(new WakeupCriterion[]{
                            hiResWakeupOnViewPlatformEntry,
                            loResWakeupOnViewPlatformEntry
                            }));

                        return;
                        }

                    // Create viewable scene elements on demand.
                    if(discardable.numChildren() == 0)
                        {
                        final BranchGroup bg = new BranchGroup();
                        bg.setCapability(BranchGroup.ALLOW_DETACH);

                        // Place exhibit at the centre of the specified space...
                        Texture texture = null;
//                        try { texture = logic.getThumbnailImageAsTexture(-1, false, false); }
//                        catch(final Throwable t) { t.printStackTrace(); }
                        final Appearance appearance = new Appearance();
                        appearance.setCapability(Appearance.ALLOW_TEXTURE_WRITE);
                        appearance.setTexture(texture);
                        final int textureWidth = (texture == null) ? 1 : texture.getWidth();
                        final TexCoordGeneration texCoordGeneration = _createTexGen(textureWidth);
                        appearance.setTexCoordGeneration(texCoordGeneration);
                        appearance.setCapability(Appearance.ALLOW_TEXGEN_WRITE);
                        final ColoringAttributes colouring = new ColoringAttributes(1,1,1, ColoringAttributes.SHADE_FLAT);
                        appearance.setColoringAttributes(colouring);

                        final float boxDim = MAX_EXHIBIT_DIM_M/2;
                        final float boxZDepth = MAX_EXHIBIT_DIM_M/5;
                        final com.sun.j3d.utils.geometry.Box box =
                            new com.sun.j3d.utils.geometry.Box(boxDim, boxDim, boxZDepth,
                                                               appearance);
                        box.setPickable(true); // Pickable to pop up exhibit details/page.
                        final String exhibitIndex = MemoryTools.intern(String.valueOf(index));
                        box.setUserData(exhibitIndex); // Set the name to be the index.
                        final Transform3D translate = new Transform3D();
                        translate.set(new Vector3f(centre3d));
                        final TransformGroup tg = new TransformGroup(translate);
                        front = box.getShape(com.sun.j3d.utils.geometry.Box.FRONT);
                        front.setUserData(exhibitIndex); // Set the name to be the index.
                        front.getAppearance().setTexture(texture);
                        front.getAppearance().setTexCoordGeneration(texCoordGeneration);
                        front.getAppearance().setCapability(Appearance.ALLOW_TEXTURE_WRITE);
                        front.getAppearance().setCapability(Appearance.ALLOW_TEXGEN_WRITE);
                        tg.addChild(box);
                        bg.addChild(tg);

                        // Create transform group for "indicator".
                        final Transform3D translateiS = new Transform3D();
                        translateiS.set(new Vector3f(centre.x - (0.55f*MAX_EXHIBIT_DIM_M), centre.y, centre.z + boxZDepth));
                        final TransformGroup tgiS = new TransformGroup(translateiS);
//                        tgiS.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
//                        tgiS.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
                        bg.addChild(tgiS);
                        // Create and save ref to indicator.
                        indicatorSphere = new Sphere(0.04f*MAX_EXHIBIT_DIM_M);
                        final Appearance iSApp = new Appearance();
                        final ColoringAttributes iSC = new ColoringAttributes(TEXT_ALT_COLOUR, ColoringAttributes.SHADE_FLAT);
                        iSC.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
                        iSApp.setColoringAttributes(iSC);
                        indicatorSphere.setAppearance(iSApp);
                        tgiS.addChild(indicatorSphere);

                        // Add caption's TransformGroup.
                        // Set text node ASAP when exhibit is visible.
                        final Transform3D translateCaption = new Transform3D();
                        final float captionYOffset = (0.6f*MAX_EXHIBIT_DIM_M);
                        translateCaption.set(new Vector3f(centre.x - (0.5f*MAX_EXHIBIT_DIM_M), centre.y - captionYOffset, centre.z + boxZDepth));
                        tgCaption = new TransformGroup(translateCaption);
                        tgCaption.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
                        tgCaption.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
                        bg.addChild(tgCaption);

                        bg.compile();
                        discardable.addChild(bg);
                        }

                    final Boolean notVisible = isNotVisibleToUser();
                    final boolean notAtAllVisible = Boolean.TRUE.equals(notVisible);
                    final boolean notFullyVisible = !Boolean.FALSE.equals(notVisible);

//if(IsDebug.isDebug && !(w instanceof WakeupOnElapsedTime)) { System.err.println("NOTVISIBILE="+notVisible+" for #"+index); }

                    // Short-cuts for later...
                    final Appearance app = front.getAppearance();
                    final ColoringAttributes iSC = indicatorSphere.getAppearance().getColoringAttributes();

                    // Try to create caption if not yet done,
                    // perhaps because the name was not immediately available.
                    if(tgCaption.numChildren() == 0)
                        {
                        final String exhibitName = logic.getExhibitName(index);
                        if(exhibitName != null)
                            {
                            // Make caption.
                            // Always start with at least the main component...
                            final String captionFullText = ExhibitName.getFileComponent(exhibitName);
                            final Text2D caption = _makeCaptionTexture(captionFullText);
                            final String exhibitIndex = MemoryTools.intern(String.valueOf(index));
                            caption.setUserData(exhibitIndex); // Set the name to be the exhibit index.
                            caption.setPickable(true);

                            // Make (monocased) index/initial letter.
                            final char indexLetter = Character.toLowerCase(captionFullText.charAt(0));
                            // Create a text node with the index value...
                            final Shape3D indexTextShape = _makeIndexLetter(indexLetter);

                            final BranchGroup cbg = new BranchGroup();
                            // Allow this entire group to be discarded.
//                            cbg.setCapability(BranchGroup.ALLOW_DETACH);
                            cbg.addChild(caption);
                            cbg.addChild(indexTextShape);
                            cbg.compile();
                            tgCaption.addChild(cbg);
                            }
                        }

                    // Flicker the indicator while loading the image...
                    iSC.setColor(Rnd.fastRnd.nextBoolean() ? TEXT_ALT_COLOUR : (!notFullyVisible ? TEXT_ALT3_COLOUR : TEXT_ALT2_COLOUR));

                    // If entering lo-res area
                    // or receiving a wakeup to try to refetch lo-res texture
                    // (or receiving initial wakeup in reduced lo-res area)
                    // then set the texture to the lo-res texture
                    // to save memory and CPU cycles,
                    // and only wake up on the next entry
                    // or periodically to try to (re)fetch the texture...
                    if((w == loResWakeupOnViewPlatformEntry) ||
                       (w == loResWakeupOnElapsedTime))
                        {
                        // Once started on hi-res texture fetch,
                        // don't interrupt it except by an exit.
                        if(hasEnteredHiResArea) { continue; }

                        Texture tnTexture = null;
                        // Keep postponing texture fetch if out of sight....
                        if(!notAtAllVisible)
                            {
                            try {
                            tnTexture = logic.getThumbnailImageAsTexture(index, false, notFullyVisible); }
                            catch(final Throwable t) { t.printStackTrace(); }
                            }

//if(IsDebug.isDebug && (w == loResWakeupOnViewPlatformEntry)) { System.err.println("LO-RES ENTRY FOR " + logic.getExhibitName(index) + ", tnTexture:" + tnTexture + ", notVisible:"+notVisible); }
//if(IsDebug.isDebug) { System.err.println("LO-RES TICK FOR " + logic.getExhibitName(index) + ", tnTexture:" + tnTexture + ", notVisible:"+notVisible); }

                        // Set the texture to the lo-res version
                        // if immediately available...
                        if(tnTexture != null)
                            {
//                            app.setTexCoordGeneration(_createTexGen(tnTexture.getWidth()));
                            app.setTexture(tnTexture);

                            // Show that we have loaded the texture fully.
                            iSC.setColor(TEXT_OK_COLOUR);

                            // Sleep until next entry into hi-res viewing space
                            // or exit even from lo-res viewing space...
                            wakeupOn(new WakeupOr(new WakeupCriterion[]{
                                hiResWakeupOnViewPlatformEntry,
                                loResWakeupOnViewPlatformExit
                                }));
                            }
                        // Else also set a timer to try again to get texture.
                        else
                            {
                            wakeupOn(new WakeupOr(new WakeupCriterion[]{
                                hiResWakeupOnViewPlatformEntry,
                                loResWakeupOnViewPlatformExit,
                                loResWakeupOnElapsedTime
                                }));
                            }
                        continue;
                        }

                    // If entering hi-res area
                    // or on receiving a wakeup to try to refetch hi-res texture
                    // then set the texture to the hi-res texture
                    // and only wake up on the next entry/exit of lo-res space
                    // or periodically to (re)try to fetch hi-res texture...
                    if((w == hiResWakeupOnViewPlatformEntry) ||
                       (w == hiResWakeupOnElapsedTime))
                        {
                        // Indicate that we're processing hi-res texture now.
                        hasEnteredHiResArea = true;

                        Texture tnTexture = null;
                        // Keep postponing texture fetch if out of sight....
                        if(!notAtAllVisible)
                            {
                            try { tnTexture = logic.getThumbnailImageAsTexture(index, true, notFullyVisible); }
                            catch(final Throwable t) { t.printStackTrace(); }
                            }

//if(IsDebug.isDebug && (w == hiResWakeupOnViewPlatformEntry)) { System.err.println("HI-RES ENTRY FOR " + logic.getExhibitName(index) + ", tnTexture:" + tnTexture + ", notVisible:"+notVisible); }
//if(IsDebug.isDebug && (w != hiResWakeupOnViewPlatformEntry)) { System.err.println("HI-RES TICK FOR " + logic.getExhibitName(index) + ", tnTexture:" + tnTexture + ", notVisible:"+notVisible); }

                        // Set the texture to the hi-res version
                        // if immediately available...
                        if(tnTexture != null)
                            {
//                            app.setTexCoordGeneration(_createTexGen(tnTexture.getWidth()));
                            app.setTexture(tnTexture);

                            // Show that we have loaded the texture fully.
                            iSC.setColor(TEXT_OK_COLOUR);

                            // Sleep until next entry into lo-res viewing space
                            // or exit even from lo-res viewing space...
                            wakeupOn(/*new WakeupOr(new WakeupCriterion[]{
                                loResWakeupOnViewPlatformEntry, */
                                loResWakeupOnViewPlatformExit /* ,
                                }) */ );
                            }
                        // Else also set a timer to try again to get texture.
                        else
                            {
                            // If we don't have the hi-res texture
                            // then (re)use the lo-res one in the interim
                            // if none is currently set.
                            // Keep postponing texture fetch if out of sight....
                            if((app.getTexture() == null) && !notAtAllVisible)
                                {
                                try { tnTexture = logic.getThumbnailImageAsTexture(index, false, notFullyVisible); }
                                catch(final Throwable t) { t.printStackTrace(); }
    //                            app.setTexCoordGeneration(_createTexGen(tnTexture.getWidth()));
                                app.setTexture(tnTexture);
                                }

                            wakeupOn(new WakeupOr(new WakeupCriterion[]{
//                                loResWakeupOnViewPlatformEntry,
                                loResWakeupOnViewPlatformExit,
                                hiResWakeupOnElapsedTime
                                }));
                            }
                        continue;
                        }

                    logger.log("WARNING: unrecognised stimulus: " + w);
                    }
                }
            };

        // This result will always directly contain
        // the behaviour and scheduling bounds.
        final BranchGroup result = new BranchGroup();
        result.addChild(discardable);
        appearanceBehaviour.setSchedulingBounds(new BoundingSphere(centre3d, EXHIBIT_VISIBLE_M));
        result.addChild(appearanceBehaviour);

        if(false && IsDebug.isDebug)
            {
            // Create a text node with the index value...
            final Text3D bannerText = new Text3D(new Font3D(new Font("Serif", Font.PLAIN, 1), new FontExtrusion()),
                                                 String.valueOf(index),
                                                 centre,
                                                 Text3D.ALIGN_CENTER,
                                                 Text3D.PATH_RIGHT);
            final Appearance textAppearance = new Appearance();
            final ColoringAttributes textColouring = new ColoringAttributes(1,1,0, ColoringAttributes.SHADE_FLAT);
            textAppearance.setColoringAttributes(textColouring);
//            textAppearance.setTexture(texture);
//            textAppearance.setTexCoordGeneration(new TexCoordGeneration());
            final Shape3D textShape = new Shape3D(bannerText, textAppearance);
            result.addChild(textShape);
            }

        result.compile();
        return(result);
        }

    /**Minimum number of characters that we will show in a caption for comprehensibility; non-negative. */
    private static final int MIN_CAPTION_CHARS = 10;

    /**Maximum number of chars (excluding any trailer) so far found to fit in a caption; no less than MIN_CAPTION_CHARS.
     * Suitable for thread-safe lockless access.
     */
    private static final AtomicInteger maxCaptionCharsSoFar = new AtomicInteger(MIN_CAPTION_CHARS);

    /**Suffix to add to any trimmed caption. */
    private static final String TRIMMED_CAPTION_SUFFIX = "...";

    /**Makes the caption for an exhibit as a texture; never null.
     * Tries to reuse textures, etc, to save time and space.
     *
     * @param captionFullText  printable-ASCII string; never null
     */
    private static Text2D _makeCaptionTexture(final String captionFullText)
        {
//final long captionBuildStart = System.currentTimeMillis();

        String workingText = captionFullText;
        Text2D caption = null;
        // Consider trimming caption text only if text longer than minimum.
        if(workingText.length() > MIN_CAPTION_CHARS)
            {
            // Note if we have had to shorten the caption to fit...
            boolean shortened; // = false;

            // Intersection point to check for caption being too long.
            final Point3d lengthCheckPoint = new Point3d(2*MAX_EXHIBIT_DIM_M, 0, 0);

            // If text is longer than max chars yet found to fit
            // then assume that we will need to truncate to at most that length.
            // (If that is too short then bump up max chars notion.)
            // Go on increasing the length until the caption is just too long...
            final int mc1 = maxCaptionCharsSoFar.get();
            if(workingText.length() > mc1)
                {
                final int fullLength = captionFullText.length();
                for(int len = mc1; len <= workingText.length(); ++len)
                    {
                    shortened = len < fullLength;
                    final String s = workingText.substring(0, len);
                    String trimmed = s;
                    if(shortened) { trimmed += TRIMMED_CAPTION_SUFFIX; }
                    final Text2D trialCaption = _makeRawCaptionText2D(trimmed);
                    final boolean tooLong = trialCaption.getBounds().intersect(lengthCheckPoint);

                    // We stop as soon as the trial caption is too long
                    // and make that our working text...
                    if(tooLong)
                        {
                        workingText = s; // Will need at least one char trimmed...
                        break;
                        }

                    // If this trial caption was unexpectedly NOT too long
                    // then ensure that we bump up the max limit to at least
                    // the current length (+1) without race problems.
                    for( ; ; )
                        {
                        final int current = maxCaptionCharsSoFar.get();
                        final int next = Math.max(current, len+1);
                        if(maxCaptionCharsSoFar.compareAndSet(current, next))
                            { break; }
                        }
//System.err.println("*** maxCaptionCharsSoFar = " + maxCaptionCharsSoFar);
                    }

                // Now work down towards the minimum acceptable length
                // stopping when we get a trimmed version that just fits...
                final StringBuilder sb = new StringBuilder(workingText);
                for(int len = workingText.length(); len >= MIN_CAPTION_CHARS; --len)
                    {
                    shortened = len < fullLength;
                    sb.setLength(len); // Trim off any surplus...
                    if(shortened) { sb.append(TRIMMED_CAPTION_SUFFIX); }
                    // Eliminate duplicates from memory...
                    caption = _makeRawCaptionText2D(MemoryTools.intern(sb.toString()));
                    final boolean tooLong = caption.getBounds().intersect(lengthCheckPoint);
                    // When not too long then we can stop!
                    if(!tooLong) { break; }
                    }
                }
            }
        // (Re)build caption if necessary...
        if(caption == null)
            { caption = _makeRawCaptionText2D(MemoryTools.intern(workingText)); }

//        caption.setAppearance(TEXT_BG_APPEARANCE);

//final long captionBuildEnd = System.currentTimeMillis();
//System.err.println("Caption text build/trim time (ms): " + (captionBuildEnd - captionBuildStart) + " for " + captionFullText);
        return(caption);
        }

    /**Make raw caption Text2D from the given text; never null. */
    private static Text2D _makeRawCaptionText2D(final String sb)
        {
        return new Text2D(sb,
                          TEXT_COLOUR,
                          "Serif",
                          Math.round(20 * MAX_EXHIBIT_DIM_M),
                          Font.PLAIN);
        }

    /**Private geometry cache for _makeIndexLetter; never null.
     * Big enough to hold geometry for all ASCII chars up to 'z' by index.
     * <p>
     * Accessed under a lock on this array for thread-safety.
     */
    private static final Text3D _makeIndexLetter_geomCache[] = new Text3D[1 + 'z'];

    /**Makes an index letter bottom-right aligned to (0,0,0); never null.
     * We try to cache the geometry of the text to save time and memory.
     *
     * @param indexLetter  printable ASCII character (up to 'z')
     */
    private static Shape3D _makeIndexLetter(final char indexLetter)
        {
        assert((indexLetter >= ' ') && (indexLetter < _makeIndexLetter_geomCache.length));

        synchronized(_makeIndexLetter_geomCache)
            {
            Text3D cached = _makeIndexLetter_geomCache[indexLetter];
            if(cached == null)
                {
                // Compute geometry since not yet done for this letter.
                _makeIndexLetter_geomCache[indexLetter] = cached =
                new Text3D(new Font3D(new Font("Serif", Font.PLAIN, 1), new FontExtrusion()),
                                                    MemoryTools.intern("" + indexLetter),
                                                    new Point3f(),
                                                    Text3D.ALIGN_LAST,
                                                    Text3D.PATH_RIGHT);
                }
            final Shape3D indexTextShape = new Shape3D(cached, TEXT_APPEARANCE);
            return(indexTextShape);
            }
        }

    /**Compute the view-platform location (x,y,z) in virtual-world coordinates; never null. */
    private static Vector3d computeViewPlatformVWorldXYZ(final SimpleUniverse su)
        {
        final Transform3D t3dUser = new Transform3D();
        su.getViewer().getView().getUserHeadToVworld(t3dUser);
        final Vector3d v3dUser = new Vector3d();
        t3dUser.get(v3dUser);
        return(v3dUser);
        }

    /**Cache for _createTexGen; never null.
     * Caches values for up to the standard thumbnail dimensions,
     * created on first use.
     * <p>
     * This is likely to remain largely empty
     * since we are only likely to use the slots corresponding to
     * textures that are powers of two.
     */
    private static final TexCoordGeneration _createTexGen_cache[] = new TexCoordGeneration[ThreeDLogic.TN_STD_IMAGE_DIM];

    /**Build a tex-gen that maps a texture of the given width across our exhibit box front face; never null.
     * This assumes that the caller will never modify the result,
     * and thus instances can be safely cached and reused.
     *
     * @param textureWidth  positive power of two
     */
    private synchronized static TexCoordGeneration _createTexGen(final int textureWidth)
        {
        assert(textureWidth > 0);

        TexCoordGeneration result; // = null;
        // Check the cache.
        final boolean canUseCache = textureWidth <= ThreeDLogic.TN_STD_IMAGE_DIM;
        if(canUseCache)
            {
            result = _createTexGen_cache[textureWidth - 1];
            if(result != null) { return(result); }
            }

        result = new TexCoordGeneration();
        result.setPlaneS(new Vector4f(1.0f / textureWidth, 0, 0, 0.5f));
        result.setPlaneT(new Vector4f(0, 1.0f / textureWidth, 0, 0.5f));

        if(canUseCache)
            {
            // Cache the computed value if possible.
            _createTexGen_cache[textureWidth - 1] = result;
            }

        return(result);
        }

    /**If true then lay out exhibits in 2D grid, else use 3D layout. */
    private static final boolean USE_2D_LAYOUT = false;

    /**Compute width/height (X/Y dimension) of exhibit grid given exhibit count. */
    private static int computeGridXYDim(final int exhibitCount)
        {
        if(USE_2D_LAYOUT)
            { return((int) Math.ceil(Math.sqrt(exhibitCount))); }
        else
            { return((int) Math.ceil(Math.cbrt(exhibitCount))); }
        }

    /**Compute depth (Z dimension) of exhibit grid given exhibit count. */
    private static int computeGridZDim(final int exhibitCount)
        {
        if(USE_2D_LAYOUT) /* Of fixed (1) depth for 2D layout. */
            { return(1); }
        else // For a 3D layout this is thre same as the X/Y dimension. */
            { return(computeGridXYDim(exhibitCount)); }
        }

    /**Maximum (most positive) permitted view-platform-centre X value to avoid losing sight of exhibits. */
    private static final float MAX_VPC_X = 0;

    /**Maximum (most positive) permitted view-platform-centre Y value to avoid losing sight of exhibits. */
    private static final float MAX_VPC_Y = MAX_VPC_X;

    /**Maximum (most positive) permitted view-platform-centre Z value to avoid losing sight of exhibits. */
    private static final float MAX_VPC_Z = EXHIBIT_VISIBLE_M/2;

    /**Compute minimum (most negative) permitted view-platform-centre X value to avoid losing sight of exhibits. */
    private float computeMinVPCX()
        {
        return(Math.min(MAX_VPC_X - EXHIBIT_CSPACING_M * (2 + computeGridXYDim(currentSet.exhibitCount)), 0));
        }

    /**Compute minimum (most negative) permitted view-platform-centre Y value to avoid losing sight of exhibits. */
    private float computeMinVPCY()
        {
        return(Math.min(MAX_VPC_Y - EXHIBIT_CSPACING_M * (1 + computeGridXYDim(currentSet.exhibitCount)), 0));
        }

    /**Compute minimum (most negative) permitted view-platform-centre Z value to avoid losing sight of exhibits. */
    private float computeMinVPCZ()
        {
        return(Math.min(MAX_VPC_Z - EXHIBIT_CSPACING_M * (2 + computeGridZDim(currentSet.exhibitCount)), -EXHIBIT_CSPACING_M));
        }

    /**Make the BranchGroup containing the current set of exhibits; never null.
     * This should <em>replace</em> any previous exhibit BranchGroup.
     *
     * @param galleryBasicMetaData  new exhibit set meta-data
     * @param topRightFront  origin/offset of this exhibit set; never null
     *
     * @return  non-null compiled BranchGroup containing the exhibits,
     *     with BranchGroup.ALLOW_DETACH capability set
     */
    private BranchGroup makeExhibitsScene(final LightweightMetaDataFetchInterface.GalleryBasicMetaData galleryBasicMetaData,
                                          final Point3f topRightFront)
        {
        final BranchGroup result = new BranchGroup();

        // Allow for removal (and replacement) of this exhibit-based scene
        // when the exhibit set changes.
        result.setCapability(BranchGroup.ALLOW_DETACH);

        final float offsetX = topRightFront.x;
        final float offsetY = topRightFront.y - EXHIBIT_CSPACING_M / 2;
        final float offsetZ = topRightFront.z - EXHIBIT_CSPACING_M / 2;

        if(USE_2D_LAYOUT)
            {
            // Compute length of one X/Y grid edge; Z depth is 1.
            final int gridSize = computeGridXYDim(galleryBasicMetaData.exhibitCount);

            // Make a solid grid: compute the exhibit index as we go.
            // Have the index weave from left to right to left, top to bottom.
            int index = 0;
            for(int y = 0; y < gridSize; ++y)
                {
                final float yo = offsetY - (y * EXHIBIT_CSPACING_M);
                final boolean reverseX = ((y & 1) != 0);
                for(int x = (reverseX ? gridSize : -1); reverseX ? (--x >= 0) : (++x < gridSize); )
                    {
                    final float xo = offsetX - (x * EXHIBIT_CSPACING_M);
                    if(index >= galleryBasicMetaData.exhibitCount) { break; }
                    final Node node = makeExhibitSpace(new Point3f(xo, yo, offsetZ), index++);
                    result.addChild(node);
                    }

                // Be nice to the rendering thread, etc...
                Thread.yield();
                }
            }
        else
            {
            // Compute length of one (Z) grid edge; others assumed to be same.
            final int gridSize = computeGridZDim(galleryBasicMetaData.exhibitCount);

            // Make a solid grid: compute the exhibit index as we go.
            // Have index weave right-left-right, top-bottom-top.
            int index = 0;
            boolean yUp = false; // Go upwards.
            boolean reverseX = false; // Go right-to-left.
            for(int z = 0; z < gridSize; ++z)
                {
                final float zo = offsetZ - (z * EXHIBIT_CSPACING_M);
                // Offset adjacent planes a little.
                final float skewX = ((float)Math.sin(z)) * (MAX_EXHIBIT_DIM_M/2);
                final float skewY = ((float)Math.cos(z)) * (MAX_EXHIBIT_DIM_M/2);
                for(int y = (yUp ? gridSize : -1); yUp ? (--y >= 0) : (++y < gridSize); )
                    {
                    final float yo = offsetY - (y * EXHIBIT_CSPACING_M) - skewY;
                    for(int x = (reverseX ? gridSize : -1); reverseX ? (--x >= 0) : (++x < gridSize); )
                        {
                        final float xo = offsetX - (x * EXHIBIT_CSPACING_M) - skewX;
                        if(index >= galleryBasicMetaData.exhibitCount) { break; }
                        final Node node = makeExhibitSpace(new Point3f(xo, yo, zo), index++);
                        result.addChild(node);
                        }
                    reverseX = !reverseX;

                    // Be nice to the rendering thread, etc...
                    Thread.yield();
                    }
                yUp = !yUp;
                }
            }

        result.compile();
        return(result);
        }

    /**Initialise user's location, etc. */
    private void initUserControls(final SimpleUniverse u,
                                  final BranchGroup exhibitsBranchGroup)
        {
        // Cap the frame rate to save a CPU cycle or two...
        final View view = u.getViewer().getView();
        view.setMinimumFrameCycleTime(10); // 100Hz

        // This will move the ViewPlatform back a bit so that
        // the objects in the scene can be viewed.
        final ViewingPlatform vp = u.getViewingPlatform();
        vp.setNominalViewingTransform();

        // Allow us to get the location of the user/viewer at run-time.
        view.setUserHeadToVworldEnable(true);

        // Harmonise the view frustrum clipping with our use of fog, etc.
        view.setBackClipDistance(EXHIBIT_VISIBLE_M * 1.1);
        // View exhibits almost with your nose pressed up agains them (~10cm).
        view.setFrontClipDistance(EXHIBIT_VISIBLE_M * 0.01);
if(IsDebug.isDebug) { System.out.println("[Field of view (radians) = " + view.getFieldOfView() + ".]"); }

        // Now set up mouse-based control of the viewing platform...
        final TransformGroup viewPlatformTransform = vp.getViewPlatformTransform();
        final Behavior navigatorBehavior1 = new MouseTranslate(viewPlatformTransform);
        final BoundingSphere boundingSphere = new BoundingSphere(new Point3d(), Double.POSITIVE_INFINITY);
        navigatorBehavior1.setSchedulingBounds(boundingSphere);
        final Behavior navigatorBehavior2 = new MouseZoom(viewPlatformTransform);
        navigatorBehavior2.setSchedulingBounds(boundingSphere);
        final BranchGroup scene = new BranchGroup();
        scene.addChild(navigatorBehavior1);
        scene.addChild(navigatorBehavior2);
        scene.compile();
        vp.addChild(scene);

        // Set up picking...
        final PickCanvas pickCanvas = new PickCanvas(canvas3D, exhibitsBranchGroup);
        pickCanvas.setMode(PickInfo.PICK_BOUNDS);
        pickCanvas.setFlags(PickInfo.NODE);
        canvas3D.addMouseListener(new MouseListener()
            {
            public void mouseClicked(final MouseEvent e)
                {
                // Regards user as active.
                logic.setUserLastActive("Mouse click");

                // Search all matching selected visible nodes, closest first.
                pickCanvas.setShapeLocation(e);
                final PickInfo pickInfos[] = pickCanvas.pickAllSorted();
                if(pickInfos == null) { return; /* Nothing to see here; move along please. */ }
                for(final PickInfo pickInfo : pickInfos)
                    {
//System.out.println("pickInfo = " + pickInfo);
                    if(pickInfo == null) { continue; }

                    // Filter out out-of-sight objects.
                    final double closestDistance = pickInfo.getClosestDistance();
                    if(Double.isNaN(closestDistance)) { continue; }
                    if(!(closestDistance <= EXHIBIT_VISIBLE_M)) { continue; }

                    final Node node = pickInfo.getNode();
//System.out.println("pickInfo.getNode() = " + node);
                    if(node == null) { continue; }

                    // Try to retrieve the exhibit number from the node name.
                    final String name = (String) node.getUserData();
//System.out.println("node.getName() = " + name);
                    if(name == null) { continue; }

                    // Reject unparseable or out-of-bounds exhibit index.
                    int exhibitNumber = -1;
                    try { exhibitNumber = Integer.parseInt(name, 10); }
                    catch(final NumberFormatException nfe) { /* Ignore. */ }
                    if(exhibitNumber < 0) { continue; }
                    if(exhibitNumber >= currentSet.exhibitCount) { continue; }

                    logger.log("Selected exhibit #" + exhibitNumber);
                    final String exhibitName = logic.getExhibitName(exhibitNumber);
                    // If we can't get the exhibit name (yet)
                    // then ignore this pick for now...
                    if(!ExhibitName.validNameSyntax(exhibitName)) { return; }

                    logger.log("Selected exhibit " + ExhibitName.getFileComponent(exhibitName));

                    // If we can, then show exhibit in browser.
                    if((logic.bs != null) &&
                       logic.bs.isWebBrowserSupported())
                        {
                        // Try to show exhibit page in browser...
                        try
                            {
//                            final String rrURL = WebUtils.makeCatPageRRURL(exhibitName, WebConsts.F_secondary_generated_HTML_suffix);
                            final String rrURL = ("/_c/" + exhibitName + WebConsts.F_secondary_generated_HTML_suffix);
                            if(logic.bs.showDocument(new URL(logic.bs.getCodeBase(), rrURL)))
                                { return; /* Success! */ }
                            }
                        catch(final Exception be)
                            {
                            logger.log("Problem showing exhibit page: " + be.getMessage());
                            return; // Failed.
                            }
                        }

                    logger.log("Could not show selected exhibit in browser: " + ExhibitName.getFileComponent(exhibitName));
                    return; // Failed.
                    }
                }

            public void mousePressed(MouseEvent e)
                { /* Ignore this event. */ }

            public void mouseReleased(MouseEvent e)
                { /* Ignore this event. */ }

            public void mouseEntered(MouseEvent e)
                { /* Ignore this event. */ }

            public void mouseExited(MouseEvent e)
                { /* Ignore this event. */ }
            });
        }


//    /**Make the standard insets to wrap round most components.
//     * Note that the returned Insets() object is mutable.
//     */
//    private static Insets makeStandardInsets()
//        { return(new Insets(2, 2, 2, 2)); }

    /**Creates and initialises a status bar. */
    static private JLabel createStatusBar()
        {
        final JLabel sbar = new JLabel("Ready");
        sbar.setBorder(BorderFactory.createEtchedBorder());
        return(sbar);
        }

    /**This method acts as the Action handler delegate for all the actions. */
    public void actionPerformed(final ActionEvent evt)
        {
        final String command = evt.getActionCommand();

//        // Compare the action command to the known actions,
//        // most-frequent first for efficiency.
//        if(command.equals(aboutAction.getActionCommand()))
//            {
//            // The "About" action was invoked...
//            JOptionPane.showMessageDialog(this, aboutAction.getLongDescription(), aboutAction.getShortDescription(), JOptionPane.INFORMATION_MESSAGE);
//            }
//        else if(command.equals(exitAction.getActionCommand()))
//            {
//            // The "Exit" action was invoked...
//            // Shut down unless vetoed...
//            try { shutdown(); }
//            catch(final UnsupportedOperationException e) { }
//            }
//        else
            {
            logger.log("Unexpected command: " + command);
            }

        // Note user activity...
        logic.setUserLastActive("actionPeformed: " + command);
        }

//    /**This adapter is constructed to handle mouse-over component events. */
//    private static final class MouseHandler extends MouseAdapter
//        {
//        private JLabel label;
//        private String oldMsg;
//
//        /**Adaptor constructor.
//         * @param label the JLabel which will recieve value of the
//         *              Action.LONG_DESCRIPTION key
//         */
//        public MouseHandler(final JLabel label)
//            {
//            setLabel(label);
//            oldMsg = label.getText();
//            }
//
//        public void setLabel(final JLabel label)
//            {
//            this.label = label;
//            }
//
//        public void mouseEntered(final MouseEvent evt)
//            {
//            if(evt.getSource() instanceof AbstractButton)
//                {
//                AbstractButton button = (AbstractButton) evt.getSource();
//                Action action = button.getAction(); // getAction is new in JDK 1.3
//                if(action != null)
//                    {
//                    oldMsg = label.getText();
//                    String message = (String) action.getValue(Action.LONG_DESCRIPTION);
//                    label.setText(message);
//                    }
//                }
//            }
//
//        public void mouseExited(final MouseEvent evt)
//            {
//            label.setText(oldMsg);
//            }
//        }


//    /**ActionListener invoked by pollUI; never null. */
//    private final ActionListener pollAL;

    /**Poll UI (in AWT/Swing thread, ie Swing-safe).
     * This routine notices when the exhibit set changes
     * and rebuilds the exhibit part of the scene graph to suit.
     */
    private void pollUI(/*final ActionEvent e*/)
        {
        // If the interface is not active
        // then save resources here and at the server...
        if(logic.userInactive())
            {
            logger.log("Sleeping...");
            return;
            }

//        pollAL.actionPerformed(e);

        final LightweightMetaDataFetchInterface.GalleryBasicMetaData galleryBasicMetaData = logic.getGalleryBasicMetaData();
        if(galleryBasicMetaData.exhibitSetHash != currentSet.exhibitSetHash)
            {
            logger.log("Gallery has "+galleryBasicMetaData.exhibitCount+" exhibits.");

            // Construct new screen graph for exhibits and replace old one.
            // Top/right/front at approx (0,0,0).
            final BranchGroup newExhitbitBranchGroup = makeExhibitsScene(galleryBasicMetaData,
                                                                         new Point3f());
            // Remove old exhibit scene graph.
            exhibitsBranchGroup.removeAllChildren();
            // Add new (pre-compiled) graph.
            exhibitsBranchGroup.addChild(newExhitbitBranchGroup);

            // Remember new metadata.
            currentSet = galleryBasicMetaData;

            // Reset slider boundaries...
            // Round these boundaries inwards
            // so that extreme slider values are just within the viewing bounds.
            sliderX.setMinimum((int) Math.ceil(computeMinVPCX()));
            sliderX.setMaximum((int) Math.floor(MAX_VPC_X));
            sliderY.setMinimum((int) Math.ceil(computeMinVPCY()));
            sliderY.setMaximum((int) Math.floor(MAX_VPC_Y));
            sliderZ.setMinimum((int) Math.ceil(computeMinVPCZ()));
            sliderZ.setMaximum((int) Math.floor(MAX_VPC_Z));
            }

        // Coerce view-platform centre back into allowed range
        // if straying out of sight of the exhibits.
        final SimpleUniverse su = simpleUniverse;
        if(su != null)
            {
            final Vector3d vpcLocation = computeViewPlatformVWorldXYZ(su);
//if(IsDebug.isDebug) { logger.log("ViewPlatform location is: " + vpcLocation); }
            boolean resetVP = false;
            if((vpcLocation.x > MAX_VPC_X) || (vpcLocation.x < computeMinVPCX()))
                {
                resetVP = true;
                }
            else if((vpcLocation.y > MAX_VPC_Y) || (vpcLocation.y < computeMinVPCY()))
                {
                resetVP = true;
                }
            else if((vpcLocation.z > MAX_VPC_Z) || (vpcLocation.z < computeMinVPCZ()))
                {
                resetVP = true;
                }

            // OK, take account of out-of-bounds view platform..
            if(resetVP)
                {
                // Bring user back to where they can see things...
if(IsDebug.isDebug) { logger.log("Forcing out-of-bounds ViewPlatform from: " + vpcLocation); }
                su.getViewingPlatform().setNominalViewingTransform();

//                final TransformGroup viewPlatformTransform = su.getViewingPlatform().getViewPlatformTransform();
//                Transform3D View_Transform3D = new Transform3D();
//                viewPlatformTransform.getTransform(View_Transform3D);
//                View_Transform3D.setTranslation(vpcLocation);
//                viewPlatformTransform.setTransform(View_Transform3D);
                }

            // Update position on sliders if out of sync.
            final int xRound = (int) Math.round(vpcLocation.x);
            if(sliderX.getValue() != xRound) { sliderX.setValue(xRound); }
            final int yRound = (int) Math.round(vpcLocation.y);
            if(sliderY.getValue() != yRound) { sliderY.setValue(yRound); }
            final int zRound = (int) Math.round(vpcLocation.z);
            if(sliderZ.getValue() != zRound) { sliderZ.setValue(zRound); }
            }
        }

    /**What exhibit set are we currently displaying; never null.
     * When this changes we reconstruct our scene graph.
     * <p>
     * Marked volatile for thread-safe lock-free access.
     * <p>
     * Initially an empty set of exhibits.
     */
    private volatile LightweightMetaDataFetchInterface.GalleryBasicMetaData currentSet = LightweightMetaDataFetchInterface.GalleryBasicMetaData.EMPTY;


    /**Listener class used to veto attempts to start another app instance.
     * A (modal) dialog[ue] is shown instead.
     */
    private static final class SISListener implements SingleInstanceListener
        {
        public final void newActivation(final String[] params)
            {
            System.err.println("Attempted to launch another instance, args: " + Arrays.asList(params));

            // Schedule a job for the event-dispatching thread:
            // creating and a blocking dialogue.
            SwingUtilities.invokeLater(new Runnable(){
                public final void run()
                    {
                    JOptionPane.showMessageDialog(null, "Gallery 3D Walkthrough already running...", "Already running", JOptionPane.ERROR_MESSAGE);
                    }
                });
            }
        }

    /**Main method invoked from JWS.
     */
    public static void main(final String[] args)
        {
        // Try to set a local look-and-feel...
        try { UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); }
        catch(final Exception e) { }

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        // Also start a daemon thread to drive async non-GUI activity.
        SwingUtilities.invokeLater(new Runnable(){
            public final void run()
                {
                final ThreeDMain mainFrame = new ThreeDMain();
                mainFrame.pack();
                mainFrame.setVisible(true);

                // Try to "warm up" first exhibit that user will see.
                mainFrame.logic.getThumbnailImageAsTexture(0, false, false);
                mainFrame.logic.getThumbnailImageAsTexture(0, true, false);

                // After displaying the UI,
                // but before allowing much else to happen,
                // reload any persisted data
                // and start any background processing...
                mainFrame.logic.startup();

                // Start a (daemon) poller thread for UI async activity.
                // This is called from a Swing timerUI, so is Swing-safe.
                // This is created and started AFTER the UI object construction
                // is complete so as to try to avoid any unpleasant races.
                // We call this every 100ms or so in order to feel responsive.
                final Timer timerUI = new Timer(70 + Rnd.fastRnd.nextInt(30),
                                                        (new ActionListener(){
                    /**Poll the UI logic. */
                    public void actionPerformed(final ActionEvent e)
                        { mainFrame.pollUI(/* e */); }
                    }));
                // Delay first poll a little to allow the system to warm up...
                timerUI.setInitialDelay(301);
                timerUI.start();


                // Create and attach and animate the scene.
                if(mainFrame.canvas3D != null)
                    {
                    // Create our Universe...
                    final SimpleUniverse u = new SimpleUniverse(mainFrame.canvas3D);
                    mainFrame.simpleUniverse = u;

                    // Create a simple banner and attach it to the universe.
                    final BranchGroup banner = createMainBannerSceneGraph(new Point3f());
                    u.addBranchGraph(banner);

                    // Attach main exhibits branch to the universe.
                    // We will add/remove visible exhibits to/from this group.
                    final BranchGroup exhibitsBranchGroup = mainFrame.exhibitsBranchGroup;
                    exhibitsBranchGroup.compile();
                    u.addBranchGraph(exhibitsBranchGroup);

                    // Set up user controls of view platform...
                    mainFrame.initUserControls(u, exhibitsBranchGroup);

                    // Try to "warm up" first exhibit that user will see.
                    mainFrame.logic.getThumbnailImageAsTexture(0, false, false);
                    mainFrame.logic.getThumbnailImageAsTexture(0, true, false);
                    }
                }
            });
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 8885229711913005041L;
    }
