/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.threeD;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.Rnd;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;


/**3D Walkthrough utilities.
 * Shared by the JWS code and the supporting server-side code.
 */
final class Utils
    {
    /**Prevent construction of an instance. */
    private Utils() { }


    /**Returns true for exhibit types supported in the walkthrough, false otherwise. */
    public static boolean supportedTypeFor3DWT(final ExhibitMIME.ExhibitTypeParameters exhibitType)
        {
        switch(exhibitType.type)
            {
            case ExhibitMIME.ET_JPEG:
            case ExhibitMIME.ET_GIF:
            case ExhibitMIME.ET_PNG:
                return(true); // These types are potentially OK...

            default:
                return(false);
            }
        }



    /**Expected mount point of this servlet; starts with "/". */
    public static final String SERVLET_MOUNT_POINT = "/_3D/data";

    /**HTTP parameter used to select the operation type; URL safe. */
    public static final String HP_OPCODE = "op";
    /**HTTP parameter used to indicate the exhibit ordinal(s); URL safe. */
    public static final String HP_ORD = "n";
    /**Operation type to request basic metadata; URL safe. */
    public static final String OP_MD = "md";
//    /**Operation type to request exhibit name lookup by ordinal; URL safe. */
//    public static final String OP_LU = "lu";
    /**Operation type to request multiple exhibit name lookups by ordinal; URL safe. */
    public static final String OP_LUS = "lus";

    /**"File" portion of meta-data request. */
    private static final String MDREQ = SERVLET_MOUNT_POINT + '?' + HP_OPCODE + '=' + OP_MD;

    /**Basic connect timeout for HTTP/RPCs for reasonable interactivity, in ms; strictly positive.
     * Generally HTTP/TCP connects very quickly if it is going to connect at all
     * and anything longer usually implies a lost handshake packet
     * or a dead network connection, so waiting a longer time rarely helps.
     * <p>
     * The random factor helps to prevent different clients from colliding.
     */
    static final int RPC_HTTP_CONNECT_TIMEOUT_MS = 2011 + Rnd.fastRnd.nextInt(511);

    /**Basic read timeout for HTTP/RPCs for reasonable interactivity, in ms; strictly positive.
     * Once we have connected then the network is probably OK,
     * so it is worth waiting a while in case the server needs time to
     * compute the results; so the read timeout should usually be a little more
     * than the connect timeout.
     * <p>
     * The random factor helps to prevent separate clients from colliding.
     */
    static final int RPC_HTTP_READ_TIMEOUT_MS = 2*RPC_HTTP_CONNECT_TIMEOUT_MS + 1003 + Rnd.fastRnd.nextInt(1003);

    /**Client-side call to fetch metadata from the server; never null.
     *
     * @param hostAndPort  server host with optional port,
     *     or null to use the default/main server name
     *
     * @throws java.io.IOException in case of error
     */
    static final LightweightMetaDataFetchInterface.GalleryBasicMetaData getMetaData(URL hostAndPort)
        throws IOException
        {
        if(hostAndPort == null) { hostAndPort = new URL("http", CoreConsts.MAIN_DATA_HOST, "/"); }

        final HttpURLConnection urlConnection = _setUpConnection(hostAndPort, MDREQ);
        final int contentLength = urlConnection.getContentLength();
        if((contentLength < 5) || (contentLength > 128))
            { throw new IOException("corrupt response: Content-Length " + contentLength); }

        // Read and parse response.
        final InputStream inputStream = urlConnection.getInputStream();
        try
            {
            final DataInputStream dis = new DataInputStream(inputStream);
            final byte buf[] = new byte[contentLength];
            dis.readFully(buf);
            final StringBuilder sbi = new StringBuilder(contentLength);
            for(int i = 0; i < contentLength; ++i)
                { sbi.append((char) buf[i]); }
            try { return(LightweightMetaDataFetchInterface.GalleryBasicMetaData.fromWireString(sbi.toString())); }
            catch(final Exception e) { throw new IOException("unparsable response"); }
            }
        finally
            { inputStream.close(); /* Free resources. */ }
        }

    /**Set up and open HTTP connection to server for RPC; never null. */
    private static HttpURLConnection _setUpConnection(final URL hostAndPort,
                                                      final String rrURL)
        throws IOException
        {
        final HttpURLConnection urlConnection =
            (HttpURLConnection) (new URL(hostAndPort.getProtocol(), hostAndPort.getHost(), hostAndPort.getPort(), rrURL)).openConnection();
        urlConnection.setAllowUserInteraction(false);
        urlConnection.setConnectTimeout(RPC_HTTP_CONNECT_TIMEOUT_MS);
        urlConnection.setReadTimeout(RPC_HTTP_READ_TIMEOUT_MS);
        urlConnection.setRequestProperty("Referer", "http://" + CoreConsts.MAIN_DATA_HOST + SERVLET_MOUNT_POINT);
        urlConnection.setUseCaches(false);
        final int responseCode = urlConnection.getResponseCode();
        if(200 != responseCode)
            { throw new IOException("unexpected response code " + responseCode + " from " + urlConnection.getURL()); }
        return(urlConnection);
        }

//    /**Client-side call to get exhibit name (by index) from the server; never null.
//     *
//     * @param hostAndPort  server host with optional port,
//     *     or null to use the default/main server name
//     *
//     * @throws java.io.IOException in case of error
//     */
//    static final String getExhibitName(URL hostAndPort, final int ordinal)
//        throws IOException
//        {
//        if(ordinal < 0)
//            { throw new IllegalArgumentException(); }
//
//        if(hostAndPort == null) { hostAndPort = new URL("http", CoreConsts.MAIN_DATA_HOST, "/"); }
//
//        final HttpURLConnection urlConnection = _setUpConnection(hostAndPort, SERVLET_MOUNT_POINT + '?' + HP_OPCODE + '=' + OP_LU + '&' + HP_ORD + '=' + ordinal);
//        final int contentLength = urlConnection.getContentLength();
//        if((contentLength < ExhibitName.MIN_NAME_LENGTH) || (contentLength > ExhibitName.MAX_NAME_LENGTH))
//            { throw new IOException("corrupt response: Content-Length " + contentLength); }
//
//        // Read and parse response.
//        final InputStream inputStream = urlConnection.getInputStream();
//        try
//            {
//            final DataInputStream dis = new DataInputStream(inputStream);
//            final byte buf[] = new byte[contentLength];
//            dis.readFully(buf);
//            final StringBuilder sbi = new StringBuilder(contentLength);
//            for(int i = 0; i < contentLength; ++i)
//                { sbi.append((char) buf[i]); }
//            final String s = sbi.toString();
//            if(!ExhibitName.validNameSyntax(s))
//                { throw new IOException("corrupt exhibit name received"); }
//            return(s);
//            }
//        finally
//            { inputStream.close(); /* Free resources. */ }
//        }

    /**Client-side call to get exhibit names (by index) from the server; never null.
     *
     * @param hostAndPort  server host with optional port,
     *     or null to use the default/main server name
     *
     * @throws java.io.IOException in case of error
     */
    static final String[] getExhibitNames(URL hostAndPort, final int ordinals[])
        throws IOException
        {
        if((ordinals == null) || (ordinals.length > LightweightMetaDataFetchInterface.MAX_NAMES_REQUEST))
            { throw new IllegalArgumentException(); }

        if(hostAndPort == null) { hostAndPort = new URL("http", CoreConsts.MAIN_DATA_HOST, "/"); }

        final StringBuilder req = new StringBuilder(64);
        req.append(SERVLET_MOUNT_POINT + '?' + HP_OPCODE + '=' + OP_LUS + '&' + HP_ORD + '=');
        for(int i = 0; i < ordinals.length; ++i)
            {
            final int ordinal = ordinals[i];
            if(ordinal < 0)
                { throw new IllegalArgumentException(); }
            if(i != 0) { req.append(','); }
            req.append(ordinal);
            }
        final HttpURLConnection urlConnection = _setUpConnection(hostAndPort, req.toString());
        final int contentLength = urlConnection.getContentLength();
        if((contentLength < ExhibitName.MIN_NAME_LENGTH) || (contentLength >= (1+ExhibitName.MAX_NAME_LENGTH)*LightweightMetaDataFetchInterface.MAX_NAMES_REQUEST))
            { throw new IOException("corrupt response: Content-Length " + contentLength); }

        // Read and parse response.
        final InputStream inputStream = urlConnection.getInputStream();
        try
            {
            final DataInputStream dis = new DataInputStream(inputStream);
            final byte buf[] = new byte[contentLength];
            dis.readFully(buf);
            final StringBuilder sbi = new StringBuilder(contentLength);
            for(int i = 0; i < contentLength; ++i)
                { sbi.append((char) buf[i]); }
            final String results[] = new String[ordinals.length];
            final String s = sbi.toString();
            final String names[] = s.split(",");
            for(int i = Math.min(names.length, results.length); --i >= 0; )
                {
                final String n = names[i];
                if(!ExhibitName.validNameSyntax(n))
                    { throw new IOException("corrupt exhibit name received"); }
                results[i] = n;
                }
            return(results);
            }
        finally
            { inputStream.close(); /* Free resources. */ }
        }
    }
