/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 06-Jun-02
 * Time: 17:09:06

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

package org.hd.d.pg2k.webSvr.upload;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.datasource.ExhibitDataFileSource;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.uploader.UploadInfoBean;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;

import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.StringTokenizer;

/**Code snippets to support exhibit upload.
 * Some of these are here just to keep the JSP files shorter, simpler,
 * clearer, and faster to compile.
 * <p>
 * Thanks to <a href="mailto:bmarchal@pineapplesoft.com">Beno&icirc;t Marchal</a>
 * for example code to handle multipart/form-data HTTP POST data.
 */
public final class HTTPUploaderUtils
    {
    /**Prevent instantiation. */
    private HTTPUploaderUtils() { }


    /**Returns ID if user is logged in for uploaded, else null.
     * Can only be logged in if there is a session in place.
     *
     * @param request  current HTTP request; never null
     */
    public static final String isLoggedInForUploads(final HttpServletRequest request)
        {
        if(request == null) { throw new IllegalArgumentException(); }

        final HttpSession session = request.getSession(false); // Don't force a session.
        final Object userIDo = (session == null) ? null :
                               session.getAttribute(UploaderConsts.USERID_FIELD_NAME);
        // Do some very faint checking of validity here,
        // but really non-null would do since validation should be done elsewhere.
        final boolean authed = (userIDo instanceof String) &&
                           ExhibitName.validAuthorSyntax((String) userIDo);
        if(authed) { return((String) userIDo); }

        return(null); // Not logged in.
        }

    /**Accepts the body of a multipart/form-data POST request.
     * This does not handle lots of complicated possible cases
     * from complicated forms or strange/broken browsers.
     * <p>
     * In particular this only expects one file upload at a time.
     * <p>
     * This streams the data more-or-less directly to disc,
     * so that huge uploads can be handled with fixed, modest memory.
     * <p>
     * This will verify the magic number as the file is uploaded,
     * and abort the upload if it is too long.
     * <p>
     * The upload is to a temporary file beside (in the same directory as)
     * the final destination (to avoid problems moving across filesystems)
     * and is moved into place atomically if the upload is successful,
     * and removed if unsuccessful.
     * (Any that are left in place after a crash are easily identifiable
     * for manual removal.)
     * <p>
     * This puts the uploaded file below destinationDir,
     * creating any subdirectories needed,
     * in a path of the form categoryDir/name eg <tt>art/cat-1-ANON.jpg</tt>.
     * <p>
     * This will not allow overwrite any existing file as a security
     * precaution.
     * <p>
     * This does not serialise access to the upload area;
     * this must be done externally if required.
     * <p>
     * We rely on UploadInfoBean only creating sensible and safe pathnames.
     * <p>
     * FIXME: we should probably validate the uploaded file content
     * by trying whatever tests we have to hand before finally accepting it.
     *
     * @param request  the HTTP POST request
     * @param destinationDir  the root directory to upload to
     * @param maxUploadFileSize  the maximum space that the uploaded file
     *     may consume (bytes); must be positive
     * @param uib  information on the bean to upload, must be non-null and
     *     be able to generate a valid unique name
     * @throws IOException in case of I/O problems
     * @throws InvalidObjectException  if the upload has a bad magic number
     */
    public static void doUpload(final HttpServletRequest request,
                                final UploadInfoBean uib,
                                final int maxUploadFileSize,
                                final File destinationDir,
                                final SimpleLoggerIF logger)
        throws IOException,
               InvalidObjectException
        {
        if((maxUploadFileSize <= 0) ||
            (request == null) ||
            (uib == null) || !uib.enoughValidUniqueInfo() ||
            (destinationDir == null) ||
            (logger == null))
            { throw new IllegalArgumentException(); }

        if(!request.getContentType().startsWith("multipart/form-data"))
            { throw new IOException("wrong request type"); }

        String boundary =
            request.getHeader("Content-Type");
        boundary = boundary.substring(boundary.indexOf('=') + 1);
        boundary = "--" + boundary;
//System.out.println("boundary='" + boundary + "'");
        final ServletInputStream in = request.getInputStream();
        doUpload(uib, destinationDir, in, boundary, maxUploadFileSize, logger);
        }

    /**Accepts the body of a multipart/form-data POST request.
     * This does not handle lots of complicated possible cases
     * from complicated forms or strange/broken browsers.
     * <p>
     * In particular this only expects one file upload at a time.
     * <p>
     * This streams the data more-or-less directly to disc,
     * so that huge uploads can be handled with fixed, modest memory.
     * <p>
     * This will verify the magic number as the file is uploaded,
     * and abort the upload if it is too long.
     * <p>
     * The upload is to a temporary file beside (in the same directory as)
     * the final destination (to avoid problems moving across filesystems)
     * and is moved into place atomically if the upload is successful,
     * and removed if unsuccessful.
     * (Any that are left in place after a crash are easily identifiable
     * for manual removal.)
     * <p>
     * This puts the uploaded file below destinationDir,
     * creating any subdirectories needed,
     * in a path of the form categoryDir/name eg <tt>art/cat-1-ANON.jpg</tt>.
     * <p>
     * This will not allow overwrite any existing file as a security
     * precaution.
     * <p>
     * This does not serialise access to the upload area;
     * this must be done externally if required.
     * <p>
     * We rely on UploadInfoBean only creating sensible and safe pathnames.
     * <p>
     * FIXME: we should probably validate the uploaded file content
     * by trying whatever tests we have to hand before finally accepting it.
     *
     * @param is  the input stream from the HTTP POST request
     * @param destinationDir  the root directory to upload to
     * @param maxUploadFileSize  the maximum space that the uploaded file
     *     may consume (bytes); must be positive
     * @param uib  information on the bean to upload, must be non-null and
     *     be able to generate a valid unique name
     * @param boundary the MIME multi-part boundary string
     *     (prefixed with lots of "-"s, exactly as will start a line)
     * @throws IOException in case of I/O problems
     * @throws InvalidObjectException  if the upload has a bad magic number
     */
    public static void doUpload(final UploadInfoBean uib,
                                 final File destinationDir,
                                 final ServletInputStream is,
                                 final String boundary,
                                 final int maxUploadFileSize,
                                 final SimpleLoggerIF logger)
        throws IOException,
               InvalidObjectException
        {
        if(logger == null)
            { throw new IllegalArgumentException(); }

        // Transcribe the upload stream if requested.
        final ServletInputStream in = (transcriptName != null) ?
            transcribeInputStream(is) : is;

        final byte[] ibuf = new byte[RW_BLOCK_SIZE];
        int state = DU_INITIAL_STATE;
        String name = null;
        String value = null;
        String filename = null;

//        // Collect any fields...
//        final Map fields = new Hashtable();

        // Get the type of the exhibit we are expecting to upload.
        final String extn = uib.getSuffix();
        assert("".equals(extn) || extn.startsWith("."));
        final ExhibitMIME.ExhibitTypeParameters etp =
            ExhibitMIME.getExhibitType(extn.substring(1));
        assert(etp != null);

        // We assume that UploadInfoBean always makes sensible and safe names.
        assert ExhibitName.validNameSyntax(uib.getFullName());

        // Generate the final target filename.
        final File targetName = new File(destinationDir, uib.getFullName());
        // Name of final directory in which exhibit file will go.
        final File targetDir = targetName.getParentFile();
        // Generate temporary name; the final component has the
        // standard "temporary" component prefixed
        // which is easy to recognise and ensures that the file does
        // not have a valid exhibit file name.
        final File tempFile = new File(targetDir,
            FileTools.F_tmpPrefix + targetName.getName());

        logger.log("Upload starting for file: dest = " + targetName);
        logger.log(" [Temporary file " + tempFile + ".]");

        // Make any parent directories needed.
        targetDir.mkdirs();

        // Abort if the target file or the temporary target already exist.
        if(targetName.exists() || !tempFile.createNewFile())
            { throw new IOException("target file/temporary already exists"); }
        //FileOutputStream os = null;
        int writeCount = 0; // Number of bytes.
        int totalBytes = 0; // Bytes of exhibit written.

        // Create a buffer to accumulate output into big chunks...
        // We add a few hundred bytes on to the end to allow for
        // the average interface between \n characters
        // that will force readLine() to return,
        // ie we aim to accommodate the RW_BLOCK_SIZE plus
        // an full read without forcing obuf to be resized, mostly.
        final ByteArrayOutputStream obuf =
            new ByteArrayOutputStream(RW_BLOCK_SIZE + 1024);

        RandomAccessFile os = null;
        try {
            // Open file for output...
            // We only actually need pure write access...
            os = new RandomAccessFile(tempFile, "rw");

            // Read a block of bytes from the HTTP request...
            int n; // = in.readLine(buf, 0, RW_BLOCK_SIZE);
            while(-1 != (n = in.readLine(ibuf, 0, RW_BLOCK_SIZE))) // Until end of stream...
                {

//System.out.print("Bytes read: ");
//for(int i = 0; i < n; ++i)
//    {
//    final int ib = buf[i] & 0xff;
//    if((ib < 33) || (ib > 126)) { System.out.print(" 0x" + Integer.toHexString(ib)); }
//    else { System.out.print(" " + (char)(ib)); }
//    }
//System.out.println();

                final String st = new String(ibuf, 0, n);
                if(st.startsWith(boundary))
                    {
                    // Reset to fetch next field/value.
                    state = DU_INITIAL_STATE;
                    if(null != name)
                        {
                        // If it looks like we collected a parameter
                        // previously...
                        if(value != null)
                            {
//                            fields.put(name,
//                                       value.substring(0,
//                                                       // -2 to remove CR/LF
//                                                       value.length() - 2));
                            }

                        // If it looks like we collected an exhibit...
                        else if(totalBytes >= 2)
                            {
                            // Trim off trailing CRLF.
                            // Adjust note of bytes written...
                            totalBytes -= 2;

                            // Flush out anything in the memory buffer.
                            // But don't bother with any of the trailing CRLF.
                            // This should usually avoid need to truncate file
                            // to explicitly remove the CRLF.
                            if(obuf.size() > 2)
                                {
                                final byte[] bn = obuf.toByteArray();
                                os.write(bn, 0, bn.length - 2); // Drop CRLF...
                                obuf.reset(); // Clear memory buffer...
                                ++writeCount; // Disc write done...
                                }

                            // Truncate to zap the trailing CRLF if need be.
                            // Should generally not be needed.
                            if(tempFile.length() > totalBytes)
                                {
                                os.setLength(totalBytes);
                                ++writeCount; // Disc write done...
                                }

                            // Prevent any further writing to the temp file...
                            os.close();
                            os = null;

                            // Validate file content...
                            // (Abort with InvalidObjectException if invalid.)
                            if(tempFile.length() < 1)
                                { throw new InvalidObjectException("exhibit file too short"); }
                            ExhibitMIME.checkMagicOK(etp, tempFile);
                            // FIXME: validate file content more thoroughly...

                            // And move it into position...
                            tempFile.renameTo(targetName);

logger.log("Uploaded file: source name = " + filename + ": dest = " + targetName + ": size = " + totalBytes);
logger.log(" Write count = " + writeCount + "; bytes/write = " + (totalBytes / Math.max(1, writeCount)));
                            }
                        // Done!

                        name = null;
                        value = null;
                        filename = null;
//                        contentType = null;
                        return;
                        }
                    }
                else if(st.startsWith(
                    "Content-Disposition: form-data") &&
                    state == DU_INITIAL_STATE)
                    {
                    final StringTokenizer tokenizer =
                        new StringTokenizer(st, ";=\"");
                    while(tokenizer.hasMoreTokens())
                        {
                        String token = tokenizer.nextToken();
                        if(token.startsWith(" name"))
                            {
                            name = tokenizer.nextToken();
                            state = DU_GOT_PARAM_NAME;
                            }
                        else if(token.startsWith(" filename"))
                            {
                            filename = tokenizer.nextToken();
                            StringTokenizer ftokenizer =
                                new StringTokenizer(filename, "\\/:");
                            filename = ftokenizer.nextToken();
                            while(ftokenizer.hasMoreTokens())
                                filename = ftokenizer.nextToken();
                            state = DU_GOT_FILE_NAME;
                            break;
                            }
                        }
                    }
                else if(st.startsWith("Content-Type") &&
                       (state == DU_GOT_FILE_NAME))
                    {
//                    pos = st.indexOf(":");
                    // + 2 to remove the space
                    // - 2 to remove CR/LF
//                    contentType =
//                        st.substring(pos + 2, st.length() - 2);
                    }
                // When we see end of filename parameter header
                // start collecting actual file data.
                else if(st.equals("\r\n") && (state == DU_GOT_FILE_NAME))
                    { state = DU_COLLECTING_FILE; }
                else if(st.equals("\r\n") && (state == DU_GOT_PARAM_NAME))
                    { state = DU_COLLECTING_PARAM; }
                else if(state == DU_COLLECTING_PARAM)
                    { value = ((value == null) ? st : value + st); }

                // Append a block of uploaded bytes to the
                // exhibit that we are collecting.
                else if(state == DU_COLLECTING_FILE)
                    {
                    // If we would exceed the available space,
                    // abort...
                    if(os.getFilePointer() + n > maxUploadFileSize)
                        { throw new EOFException(); }

//System.out.println("Writing upload file block (bytes): " + n);

                    // Append next block to our in-memory buffer.
                    obuf.write(ibuf, 0, n);

                    // Logically extended the file...
                    totalBytes += n;

                    // If more than half the block size, flush out...
                    if(obuf.size() >= RW_BLOCK_SIZE)
                        {
                        final byte[] bn = obuf.toByteArray();
                        os.write(bn);
                        obuf.reset(); // Clear memory buffer...
                        ++writeCount; // Disc write done...

                        // As soon as we can check the magic number,
                        // do so, exactly once.
                        // (Abort with Exception if invalid.)
                        final int magicLength = etp.magic.length();
                        final long filePointer = os.getFilePointer();
                        if((filePointer >= magicLength) && // Now enough...
                           !(filePointer - bn.length >=  magicLength)) // Not before.
                            { ExhibitMIME.checkMagicOK(etp, tempFile); }
                        }

                    }
                else
                    {
                    throw new IOException("Unexpected state during upload; bytes just read: " + n);
                    }
                }
            }
        finally
            {
            // Make sure that file is closed...
            if(os != null) { try { os.close(); } catch(IOException e) { } }

            // Make sure that we remove any temporary file that we created
            // (and didn't move to the real file).
            tempFile.delete();
            }
        }

    /**Writes the POST body for a multi-part MIME stream to upload an exhibit.
     * This assumes that the other end has already been conveyed the
     * content of the UploadInfoBean,
     * and that the appropriate headers will be set for the body.
     * <p>
     * The non-null, non-empty, unique, not-appearing-in-the-data
     * boundary String must be supplied.
     * <p>
     * We insist that the boundary marker is pure printable ASCII
     * and at least 8 characters; it should usually be much longer.
     * This should have a random component for safety.
     * <p>
     * The OutputStream should be buffered if at all possible, for efficiency.
     *
     * @param uib  description of the the exhibit; never null
     * @param os  stream to write body to; never null
     * @param data  raw exhibit to upload, not altered by the routine; never null
     * @param boundary  unique boundary string; never empty or null
     *
     * @throws IOException
     */
    public static final void generateUploadPOSTBody(final UploadInfoBean uib,
                                                    final OutputStream os,
                                                    final byte data[],
                                                    final String boundary)
        throws IOException
        {
        if((uib == null) || (os == null) ||
           (boundary == null) || (boundary.length() == 0))
            { throw new IllegalArgumentException(); }

        // Check that the boundary marker looks superficially safe.
        if(boundary.length() < 8)
            { throw new IllegalArgumentException("boundary marker too short to be safe"); }
        for(int i = boundary.length(); --i >= 0; )
            {
            final char c = boundary.charAt(i);
            if((c < 33) || (c > 126))
                { throw new IllegalArgumentException("boundary marker must be printable ASCII"); }
            }

        // Write prefix (pure printable ASCII).
        final String prefix =
            boundary + "\r\n" +
            "Content-Disposition: form-data; name=\"" + UploaderConsts.exhibitFile + "\"; filename=\"autoupload\"\r\n" +
            "Content-Type: application/octet-stream\r\n" +
            "\r\n";
        for(int i = 0; i < prefix.length(); ++i)
            { os.write(prefix.charAt(i) & 0xff); }

        // Write the actual exhibit data.
        // Must be followed CRLF (which we prepend to the suffix)
        // to reduce the number of separate writes.
        os.write(data);

        // Write trailer (pure printable ASCII).
        final String suffix = "\r\n" + boundary + "--\r\n";
        for(int i = 0; i < suffix.length(); ++i)
            { os.write(suffix.charAt(i) & 0xff); }
        }


    /**If not null, relative pathname used to record transcript of exhibit upload.
     * Generally only used for debugging.
     * <p>
     * Multiple uploads' transcripts will overwrite/corrupt one another.
     * This is not intended for anything other than helping with testing.
     * <p>
     * Is null by default for release builds.
     */
    private static final File transcriptName = null;
//        new File("uploadTranscript.log");

    /**Take a transcript of the input stream.
     * The transcriptName must be non-null (else no transcript will be made)
     * and it must be a file that we can open write.
     *
     * @param is  input stream to be transcribed
     * @throws FileNotFoundException
     */
    private static ServletInputStream transcribeInputStream(final ServletInputStream is)
        throws FileNotFoundException
        {
        if(transcriptName == null) { return(is); }

        System.err.println("Writing upload transcript to: " + transcriptName);

        final BufferedOutputStream bos = new BufferedOutputStream(
            new FileOutputStream(transcriptName));

        final ServletInputStream in = new ServletInputStream(){
            /**Read a single byte; hideously inefficient. */
            public int read()
                throws IOException
                {
                final int b = is.read();
                if(b >= 0) { bos.write(b); }
                bos.flush();
                return(b);
                }

            /**Read a block of bytes. */
            public int read(byte b[], int off, int len)
                throws IOException
                {
                final int n = is.read(b, off, len);
                if(n > 0) { bos.write(b, off, n); }
                bos.flush();
                return(n);
                }

            /**Read a whole array. */
            public int read(byte b[])
                throws IOException
                {
                return(read(b, 0, b.length));
                }

            /**Read a whole line. */
            public int readLine(byte[] b, int off, int len)
                throws IOException
                {
                final int n = is.readLine(b, off, len);
                if(n > 0) { bos.write(b, off, n); }
                bos.flush();
                return(n);
                }

            /**Closes this input stream.
             * Releases any system resources associated
             * with the stream and the transcript.
             * @exception  IOException  if an I/O error occurs.
             */
            public void close() throws IOException
                {
                bos.close();
                super.close();
                }
            };

        return(in);
        }

    // States for state machine in doUpload().
    /**doUpload() initial state. */
    private static final int DU_INITIAL_STATE = 0;
    /**doUpload() got filename. */
    private static final int DU_GOT_FILE_NAME = 1;
    /**doUpload() got parameter name. */
    private static final int DU_GOT_PARAM_NAME = 2;
    /**doUpload() collecting file. */
    private static final int DU_COLLECTING_FILE = 3;
    /**doUpload() collecting parameter value. */
    private static final int DU_COLLECTING_PARAM = 4;

    /**Chunk size for reads/writes in upload; bigger is probably more efficient.
     * Ideally (much) bigger than maximum magic number size so that we
     * can read that from the first uploaded block.
     * <p>
     * Should probably be at least 512 to exceed minimum
     * guaranteed Internet MTU and disc sector size,
     * and probably a power of 2.
     * <p>
     * A good value is probably from 512 to 65536;
     */
    private final static int RW_BLOCK_SIZE = 64 * 1024;


    /**Attempts to programmatically login at the given login URL with the user ID and password supplied.
     * Assumed that any cookies/session required
     * will automatically be held onto by the JVM.
     *
     * @param justPoll  see if we remain logged in from previous attempt
     * @param loginURL  non-null URL (within our codebase)
     * @param userID  valid userID; never null
     * @param passS  valid password; never null
     * @return  true if successful, false if not
     * @throws IOException  in case of I/O problems
     */
    public static boolean doUploadLogin(final boolean justPoll,
                                        final URL loginURL,
                                        final String userID, final String passS)
        throws IOException
        {
        // Regardless of expected status, try to force (idempotent) logout.
        final URLConnection uc = loginURL.openConnection();

        uc.setConnectTimeout(20000); // Don't wait indefinitely...
        uc.setUseCaches(false);
        uc.setDoOutput(true);
        uc.setDoInput(true);
        uc.setAllowUserInteraction(false);
        uc.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
        if(uc instanceof HttpURLConnection)
            {
            final HttpURLConnection hc = (HttpURLConnection) uc;
            hc.setRequestMethod("POST");
            }

        // Do the login action if possible.
        final OutputStreamWriter out = new OutputStreamWriter(
                new BufferedOutputStream(uc.getOutputStream(), 512));
        if(!justPoll)
            {
            out.write(UploaderConsts.USERID_FIELD_NAME);
            out.write("=");
            out.write(userID); // Should not need any encoding.
            out.write("&");
            out.write(UploaderConsts.PASSWD_FIELD_NAME);
            out.write("=");
            out.write(java.net.URLEncoder.encode(passS, "UTF-8")); // May need encoding...
            out.flush();
            out.close();
            }

        // Can we find the "logged-in-OK" header.
        // (And are we logged in with the right ID?!)
        final String header = uc.getHeaderField(UploaderConsts.LOGGED_IN_HEADER);
        return(userID.equals(header));
        }

    /**Logout from the upload server if possible.
     * @param loginURL  non-null URL (within our codebase)
     * @return  true if successful, false if not
     * @throws IOException  in case of I/O problems
     */
    public static boolean doLogout(final URL loginURL)
        throws IOException
        {
        // Regardless of expected status, try to force (idempotent) logout.
        final URLConnection uc = loginURL.openConnection();

        uc.setConnectTimeout(10000); // Don't wait too long...
        uc.setUseCaches(false);
        uc.setDoOutput(false);
        uc.setDoInput(false);
        uc.setAllowUserInteraction(false);
        uc.setRequestProperty(UploaderConsts.LOGOUT_FIELD_NAME, "byebye");

        // Do the logout action if possible.
        uc.connect();

        final HttpURLConnection hc = (HttpURLConnection) uc;
        return(hc.getResponseCode() == 200);
        }


    /**Minimum space that must be left before we will allow an upload; strictly positive.
     * Should be enough for an average-to-large typical exhibit, eg several MB.
     */
    public static final int MIN_BATCH_UPLOAD_SPACE = 10123456;


    /**Run the server side of the batch upload protocol, after authentication.
     * Can be called directly from servlet doPOST() method.
     * <p>
     * (Can also be used stand-alone for unit testing.)
     * <p>
     * This waits for a request from the client,
     * then sends a FAIL response in case of immediate trouble,
     * or collects the uploaded file,
     * then sends an OK or FAIL response as appropriate.
     *
     * @param rawIS  input stream from client; never null
     * @param rawOS  output stream to client; never null
     * @param uploadDirS  upload directory; valid and never null
     * @param userID  user ID authenticated with; valid syntax and never null
     * @param userPass  password authenticated with; valid syntax and never null
     *
     * @throws IOException  in case of error
     */
    public static void runBatchUploadProtocolServerSide(final SimpleLoggerIF logger,
                                                        final AllExhibitProperties aep,
                                                        final GenProps gp,
                                                        final ServletInputStream rawIS,
                                                        final ServletOutputStream rawOS,
                                                        final String uploadDirS,
                                                        final String userID,
                                                        final String userPass)
        throws IOException
        {
        if((aep == null) || (gp == null) ||
           (rawIS == null) || (rawOS == null) ||
           (uploadDirS == null) ||
           (userID == null) || (userPass == null))
            { throw new IllegalArgumentException(); }

        final InputStream is = new BufferedInputStream(rawIS, UploaderUtils.BATCH_UPLOAD_BUFSIZE);
        final OutputStream os = new BufferedOutputStream(rawOS, 1024);

        try
            {
            // Collect list of all files currently in upload area.
            final AllExhibitImmutableData aeidUploaded = ExhibitDataFileSource._getAllExhibitImmutableData(uploadDirS, -1L, false);

            // Compute size consumed in total and by each user.
            final long uploadedTotal = aeidUploaded.computeFileSpaceBytes(null);
            final long uploadedByThisUser = aeidUploaded.computeFileSpaceBytes(userID);
            final long uploadSpaceLeft = Math.max(0,
                Math.min(LocalProps.getUploadMaxBytesPerUser() - uploadedByThisUser,
                         LocalProps.getUploadMaxBytesTotal() - uploadedTotal));

//            // Don't allow upload if a reasonable amount of space not free.
//            if(uploadSpaceLeft < MIN_BATCH_UPLOAD_SPACE)
//                {
//                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Not enough space to start upload ("+uploadSpaceLeft+"B vs "+MIN_BATCH_UPLOAD_SPACE+"B minimum).");
//                return;
//                }

//            // Double-check credentials still OK.
//            if(!SimplepassProps.isAuthorUploadPasswordCorrect(userID, userPass))
//                {
//                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Credentials expired.");
//                return;
//                }

//                // Everything seems OK; tell the client so.
//                UploaderUtils.sendResponseToBatchUploadClient(os, true, lastExhibitName, "Ready");
//                os.flush(); // Force response back to client.

            // Fetch next request from client.
            final UploaderUtils.BatchUploadClientRequest clientRequest =
                UploaderUtils.decodeRequestFromBatchUploadClient(is);

            if(clientRequest == null)
                {
                // Client-requested connection termination; so quit now.
                return;
                }

            // Note the requested exhibit for subsequent responses.
            final String lastExhibitName = clientRequest.exhibitName;

            // Check requested upload size, name, hash are OK.
            final int exhibitFileSizeLimit = gp.getWEBSVR_MAX_EX_BYTES();
            if(clientRequest.exhibitLength > exhibitFileSizeLimit)
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit too big for Gallery ("+exhibitFileSizeLimit+"B vs "+clientRequest.exhibitLength+"B requested).");
                return;
                }
            if(clientRequest.exhibitLength > uploadSpaceLeft)
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Not enough space to start upload ("+uploadSpaceLeft+"B vs "+clientRequest.exhibitLength+"B requested).");
                return;
                }
            final String fullNameInUploadArea = aeidUploaded.getFullName(ExhibitName.getFileComponent(clientRequest.exhibitName));
            if(null != fullNameInUploadArea)
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Chosen name clashes with one already uploaded ("+fullNameInUploadArea+").");
                return;
                }
            final String fullNameInGallery = aep.aeid.getFullName(ExhibitName.getFileComponent(clientRequest.exhibitName));
            if(null != fullNameInGallery)
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Chosen name clashes with one already in the Gallery ("+fullNameInGallery+").");
                return;
                }
            final String nameInGalleryDopplehash = aep.getHashMD5ToName().get(clientRequest.hashMD5);
            if(null != nameInGalleryDopplehash)
                 {
                 UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit appears to be duplicate (MD5 hash same) of one already in the Gallery ("+nameInGalleryDopplehash+").");
                 return;
                 }

            // Get the exhibit type.
            // (Check that it is one that we recognise.)
            final ExhibitMIME.ExhibitTypeParameters etp = ExhibitMIME.getInputFileType(clientRequest.exhibitName);
            if(null == etp)
                 {
                 UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit type unrecognised.");
                 return;
                 }

            // Get the exhibit section/category.
            // Reject the request if the section does not already exist.
            final String category = ExhibitName.getCategoryComponent(clientRequest.exhibitName);
            if(aep.getCategoryExhibitCounts().get(category) == null)
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit category unrecognised.");
                return;
                }

            final File destinationDir = new File(uploadDirS);

            // Make sure the category directory exists
            // else try to create it...
            final File categoryDir = new File(destinationDir, category);
            if(!categoryDir.isDirectory() || !categoryDir.canWrite())
                {
                logger.log(" [Making upload directory: " + categoryDir + ".]");
                if(!categoryDir.mkdir() || !categoryDir.isDirectory() || !categoryDir.canWrite())
                    {
                    UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Category upload directory cannot be made or is unusable: contact the maintainer.");
                    return;
                    }
                }

            // OK try to create the exhibit description and file
            // if not already present,
            // copy the data in as fast as possible,
            // and at the end check that the hash is OK,
            // else abort and tidy up.

            final long uploadStart = System.currentTimeMillis();

            // *** Actually do the upload! ***

            // Generate the final target filename.
            // Do not use any intermediate path supplied.
            final File targetName = new File(categoryDir, ExhibitName.getFileComponent(clientRequest.exhibitName));
//            // Name of final directory in which exhibit file will go.
//            final File targetDir = targetName.getParentFile();
            // Generate temporary name; the final component has the
            // standard "temporary" component prefixed
            // which is easy to recognise and ensures that the file does
            // not have a valid exhibit file name.
            final File tempFile = new File(categoryDir,
                FileTools.F_tmpPrefix + targetName.getName());

            if(tempFile.exists())
                {
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit upload apparently in progress already.");
                return;
                }

            // Open file in non-append mode.
            final MessageDigest digestMD5;
            try { digestMD5 = MessageDigest.getInstance(CoreConsts.HASH_MD5); }
            catch(final NoSuchAlgorithmException e) { throw new Error(e); } // Should not happen...
            DigestOutputStream efos = new DigestOutputStream(
                new FileOutputStream(tempFile, false), digestMD5);
            try
                {
                logger.log("Batch upload starting for file: dest = " + targetName);
                logger.log(" [Temporary file " + tempFile + ".]");

                // Copy the raw exhibit data one (large, efficient) block at a time.
                final byte buf[] = new byte[(int) Math.min(clientRequest.exhibitLength, RW_BLOCK_SIZE)];
                for(long bytesLeft = clientRequest.exhibitLength; bytesLeft > 0; )
                    {
                    final int maxXfer = (int) Math.min(buf.length, bytesLeft);
                    final int read = is.read(buf, 0, maxXfer);
                    if(read < 1) { throw new IOException("error reading data for exhibit: " + clientRequest.exhibitName); }
                    efos.write(buf, 0, read);
                    bytesLeft -= read;
                    }

                // Flush out all the data.
                efos.flush();

                // Capture the hash.
                final byte[] digestBytes = digestMD5.digest();

                efos.close();
                efos = null;

                if(UploaderUtils.BATCH_UPLOAD_MSG_TERMINATOR != is.read())
                    {
                    UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Missing terminator after exhibit data.");
                    return;
                    }

                if(tempFile.length() != clientRequest.exhibitLength)
                    {
                    UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit upload failed: write to local disc failed.");
                    return;
                    }

                // Check the exhibit survived intact.
                if(!Arrays.equals(digestBytes, clientRequest.hashMD5.getData()))
                    {
                    UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit hash wrong after upload.");
                    return;
                    }

                // Check the magic number of the uploaded file.
                final InputStream tfis = new BufferedInputStream(
                    new FileInputStream(tempFile), ExhibitMIME.getLongestMagicBytes());
                try
                    {
                    if(!ExhibitMIME.magicOK(etp, tfis))
                        {
                        UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit magic number wrong after upload.");
                        return;
                        }
                    }
                finally { tfis.close(); } // Free resources...

                // Move upload file into place.
                if(!tempFile.renameTo(targetName))
                    {
                    UploaderUtils.sendResponseToBatchUploadClient(logger, os, false, lastExhibitName, "Exhibit upload failed: failed to move into final place.");
                    return;
                    }

                // Generate the final description target filename.
                final File targetDescName = new File(categoryDir,
                    ExhibitName.getFileComponent(clientRequest.exhibitName) +
                    CoreConsts.DESCRIPTION_FILE_SUFFIX);

                // If there was a description, attempt to save it now.
                final String desc = clientRequest.description;
                if((desc != null) && (desc.length() > 0))
                    {
                    FileTools.replacePublishedFile(targetDescName.getPath(),
                                                   desc.getBytes(CoreConsts.FILE_ENCODING_8859_1),
                                                   false, // Be noisy for now...
                                                   true); // Use pure-Java mechanisms.
                    }

                final long uploadEnd = System.currentTimeMillis();
                final long uploadTime = uploadEnd - uploadStart;

                // Compute and log effective upload bandwidth.
                final long uploadBps = ((clientRequest.exhibitLength * 1000) / Math.max(1, uploadTime));
                logger.log(" Exhibit upload completed, "+clientRequest.exhibitLength+"B, "+uploadTime+"ms: " + uploadBps + "Bps.");

                // If we get here then it's probably all worked!
                // Tell the client...
                UploaderUtils.sendResponseToBatchUploadClient(logger, os, true, lastExhibitName, "Exhibit upload OK.");
                os.flush(); // Flush down the wire immediately...
                }
            finally
                {
                // Ensure that the temporary file is closed and removed.
                if(efos != null) { efos.close(); }
                tempFile.delete();
                }
            }
        finally
            {
            // Flush the output buffer...
            os.flush();
            }
        }
    }
