/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.upload;

import org.hd.d.pg2k.svrCore.AllExhibitProperties;
import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ExhibitName;
import org.hd.d.pg2k.svrCore.Rnd;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.props.SimplepassProps;
import org.hd.d.pg2k.svrCore.uploader.UploaderConsts;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletConfig;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.IOException;

/**This is the servlet that accepts uploads from non-browser sources.
 * This is used by the JWS-based uploader.
 * <p>
 * This should co-exist with the HTML/JSP/browser-based uploader,
 * and respect at least the same security and safety restrictions.
 * <p>
 * This will respond only to POST requests for efficiency and security.
 * <p>
 * The general form of the protocol for upload is:
 * <ol>
 * <li>The client sends auth credentials.
 * <li>The server sends a reject and drops the connection,
 *     else an OK
 *     [and the current space available]
 *     [and the hashes of all items in the upload area].
 * <li>The client then sends, one per exhibit:
 *     <ul>
 *     <li>The full proposed exhibit name.
 *     <li>The exhibit hash.
 *     <li>The exhibit description, if any.
 *     <li>The exhibit data.
 *     </ul>
 * <li>The server responds with an OK
 *    (naming the accepted exhibit to avoid ambibuity)
 *    or reject, dropping the connection.
 * <li>If the connection remains idle too long,
 *     or for any other reason,
 *     the server may asynchronously send a "reject" and drop the connection.
 * <li>The client, when it has no more to upload,
 *     may send a "close" request, wait for a response,
 *     and then drop the connection.
 * </ol>
 */
public final class UploadServlet extends HttpServlet
    {
    /**Get singleton (per-servlet-context) data pipeline/cache instance.
     * The config param must not be null, but for some operations
     * (such as calling destroy()) request can be null.
     */
    private static DataSourceBean getDataSource(
                final ServletConfig config,
                final HttpServletRequest request)
        {
        // Fetches/creates the data source...
        final DataSourceBean dataSource =
            DataSourceBean.getApplicationInstance(config.getServletContext());

        // Ensure that the essential details are set up.
        dataSource.setServletContext(config.getServletContext());
        if(request != null)
            { dataSource.setContextPath(request.getContextPath()); }

        return(dataSource);
        }


    /**Respond to a POST request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception java.io.IOException if an input/output error occurs
     */
    public void doPost(final HttpServletRequest request,
                       final HttpServletResponse response)
        throws IOException // , ServletException
        {
        // Make this page uncacheable.
        response.setHeader("Cache-Control", "no-cache");
        response.setHeader("Pragma", "no-cache");
        response.setHeader("Expires", "0");

        // Reject requests immediately if not prepared to accept any uploads.
        final String uploadDirS = LocalProps.getUploadDir();
        if((uploadDirS == null) || !(new File(uploadDirS)).isDirectory() ||
            (LocalProps.getUploadMaxBytesTotal() <= 0))
            {
            // If no upload directory or space, uploads are disabled.
            response.sendError(HttpServletResponse.SC_FORBIDDEN, "Uploads are disabled for this system, sorry.");
            return;
            }
        final DataSourceBean dataSource = getDataSource(getServletConfig(), request);
        if(!Boolean.FALSE.equals(dataSource.isSlave()))
            {
            // If not the master then we can't do an upload.
            response.sendError(HttpServletResponse.SC_FORBIDDEN, "You are not connected to the master system, so you cannot do an upload, sorry.");
            return;
            }

        // Check the user credentials (supplied in the headers).
        //
        // We insist on:
        //  1) A syntactically-valid author name.
        //  2) The author being already extant in the Gallery.
        //  3) A valid-length password being supplied.
        //  4) The password actually being correct!
        final String userID = request.getHeader("X-" + UploaderConsts.USERID_FIELD_NAME);
        final String userPass = request.getHeader("X-" + UploaderConsts.PASSWD_FIELD_NAME);

/* if(IsDebug.isDebug) */ { System.out.println("[***userID = "+userID+" uploading with batch client.]"); }

        final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);
        if(!ExhibitName.validAuthorSyntax(userID) || // Valid author syntax.
           (aep.getAuthorExhibitCounts().get(userID) == null) || // Extant.
           (userPass == null) || // Must be a password and a suitable length.
               (userPass.length() < CoreConsts.MIN_PASSWORD_LEN) ||
               (userPass.length() > CoreConsts.MAX_PASSWORD_LEN) ||
            !SimplepassProps.isAuthorUploadPasswordCorrect(userID, userPass))
            {
            // Bad credentials...
            // Wait a while before responding
            // so as to make brute-force attacks hard(er),
            // but not so long as to absorb excess resources.
            try { Thread.sleep(5000 + Rnd.fastRnd.nextInt(15000)); }
            catch(final InterruptedException e) { }
            response.sendError(HttpServletResponse.SC_FORBIDDEN, "Bad credentials.");
            return;
            }

        final GenProps gp = dataSource.getGenProps(-1);
        final WebUtils.ServletLogger logger = new WebUtils.ServletLogger(getServletContext());

        // Get access to the input and output streams.
        final ServletInputStream rawIS = request.getInputStream();
        final ServletOutputStream rawOS = response.getOutputStream();

        // Do the batch upload protocol..
        HTTPUploaderUtils.runBatchUploadProtocolServerSide(logger, aep, gp, rawIS, rawOS, uploadDirS, userID, userPass);
        }


    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 4321077856636574117L;
    }
