/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 18-May-02
 * Time: 15:46:38

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.webSvr.util;

import java.io.Serializable;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

/**Holds the data corresponding to a breadcrumbs navigation aid on a Web page.
 * This consists of a list of text/hyperlink pairs,
 * running from the largest/outermost part of the Website first,
 * to the innermost part (corresponding to that nearest the current page) last.
 * <p>
 * It is possible to insert items at start or end,
 * though most efficient to insert at the end.
 * <p>
 * It is possible to insert items with no link, but not with no text.
 * <p>
 * This can export a read-only view of its content as a List.
 * <p>
 * This is thread-safe and Serializable so as to be able to be stored in
 * a servlet session; nothing especially long-lived or sensitive.
 */
public final class Breadcrumbs implements java.io.Serializable
    {
    /**This is our list of breadcrumbs, in order.
     * We use a Vector so as to be inherently thread-safe.
     * <p>
     * We expect that usually we will have no more than about
     * four crumbs, and so default to that size.
     */
    private final List<Breadcrumb> crumbs = new Vector<Breadcrumb>(4);

    /**Returns read-only view of the crumbs as a List.
     */
    public List<Breadcrumb> viewAsList() { return(Collections.unmodifiableList(crumbs)); }

    /**Returns <tt>true</tt> if there are no breadcrumbs.
     */
    public boolean isEmpty() { return(crumbs.isEmpty()); }

    /**Returns the number of crumbs.
     */
    public int size() { return(crumbs.size()); }

    /**Removes all of the crumbs.
     */
    public void clear() { crumbs.clear(); }

    /**Appends crumb (to end of list).
     * Crumb must not be null.
     */
    public void append(final Breadcrumb crumbForEnd)
        {
        if(crumbForEnd == null) { throw new IllegalArgumentException(); }
        crumbs.add(crumbForEnd);
        }

    /**Prepends crumb (to start of list).
     * Crumb must not be null.
     */
    public void prepend(final Breadcrumb crumbForStart)
        {
        if(crumbForStart == null) { throw new IllegalArgumentException(); }
        crumbs.add(0, crumbForStart);
        }

    /**Get specified crumb by its index.
     * @throws IndexOutOfBoundsException if the index is invalid
     */
    public Breadcrumb getCrumb(final int index)
        throws IndexOutOfBoundsException
        { return((Breadcrumb) crumbs.get(index)); }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -7990918038988551161L;


    /**One item in the breadcrumbs. */
    public static final class Breadcrumb implements Serializable
        {
        /**The text of the link; never null nor zero-length. */
        public final String linkText;
        /**The URL (possibly relative); never zero-length but may be null. */
        public final String link;

        /**Make a breadcrumb.
         *
         * @param text  the text for the breadcrumb link; must not be null nor
         *     zero-length
         * @param url   the absolute or relative URL; can be null if not needed
         */
        public Breadcrumb(final String text, final String url)
            {
            if((text == null) || (text.length() == 0))
                { throw new IllegalArgumentException("link text cannot be empty"); }
            if(url == null)
                { throw new IllegalArgumentException("link cannot be zero-length"); }
            linkText = text;
            link = url;
            }

        /**Unique Serialisation class ID generated by http://random.hd.org/. */
        private static final long serialVersionUID = 6755765230040158278L;
        }
    }
