/* DERIVED FROM APACHE CODE, modifications (c) Damon Hart-Davis 2001-2006, under same terms:
 *
 * CompressionFilter.java
 * $Header: /rw/CVS-repository.s0.l/general/PG2Ksrc/javasrc/webSvr/org/hd/d/pg2k/webSvr/util/CompressionFilter.java,v 1.8 2005/08/18 09:30:46 dhd Exp $
 * $Revision: 1.8 $
 * $Date: 2005/08/18 09:30:46 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.hd.d.pg2k.webSvr.util;

import org.hd.d.pg2k.svrCore.stats.StatsLogger;

import ORG.hd.d.IsDebug;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Enumeration;
import java.util.StringTokenizer;

/**
 * Implementation of <code>javax.servlet.Filter</code> used to compress
 * the ServletResponse if it is bigger than a threshold.
 * <p>
 * We might need to modify this to suppress caching for HTTP/1.0
 * clients (eg by setting Pragma: no-cache and Expires: 0).
 *
 * @author Amy Roh (original)
 * @author Damon Hart-Davis
 * @version $Revision: 1.8 $, $Date: 2005/08/18 09:30:46 $
 */

public final class CompressionFilter implements Filter
    {
    /**Encoding token indicating that GZIP compression is accepted.
     * The test should be case-insensitive and work if the
     * client-supplied token starts with "x-".
     */
    static final String GZIP_TOKEN = "gzip";


    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**The stats set to which we log HTTP gzip-(un)compressed-transport counts.
     * This is one of three codes for:
     * <ul>
     * <li>The compression filter is invoked (may only be available in debug modes).
     * <li>Not compressable (eg explicitly disabled).
     * <li>Compressable (including client browser support,
     *                   though possibly not compressed if not large enough, etc).
     * </ul>
     * <p>
     * We may enable or disable this logging (etc) from system parameters,
     * but this is nonetheless our unique identifier.
     */
    private final StatsLogger.StatsConfig statsIDCompression =
        new StatsLogger.StatsConfig("HTTP-GZIP-COMPRESSION",
                                    logger, // Dump in general log.
                                    false, // Only dump summaries...
                                    12 * 3600, // About every 24 hours.
                                    true); // Adaptive.

    /**The filter configuration object we are associated with.
     * If this value is null, this filter instance is not currently configured.
     */
    private FilterConfig config;

    public FilterConfig getFilterConfig()
        {
        return (config);
        }

    public void setFilterConfig(final FilterConfig c)
        {
        config = c;
        logger.setContext((c == null) ? null : c.getServletContext());
        }

    /**The name of the application-level attribute to disable compression if Boolean.FALSE, or null. */
    private String overrideAttribute;

    /**The threshold result size below which not to try to compress. */
    protected int compressionThreshold;

    /**Debug level for this filter. */
    private int debug = 0;

    /**
     * Place this filter into service.
     *
     * @param filterConfig The filter configuration object
     */
    public void init(final FilterConfig filterConfig)
        {
        config = filterConfig;
        logger.setContext((filterConfig == null) ? null : filterConfig.getServletContext());
        if(filterConfig != null)
            {
            overrideAttribute = filterConfig.getInitParameter("overrideAttribute");

            final String str = filterConfig.getInitParameter("compressionThreshold");
            if(str != null)
                {
                compressionThreshold = Integer.parseInt(str);
                }
            else
                {
                compressionThreshold = 0;
                }
            final String value = filterConfig.getInitParameter("debug");
            if(value != null)
                {
                debug = Integer.parseInt(value);
                }
            else
                {
                debug = IsDebug.isDebug ? 1 : 0;
                }
            }
        else
            {
            compressionThreshold = 0;
            }

        }

    /**Take this filter out of service. */
    public void destroy()
        {
        this.config = null;
        logger.setContext(null);
        }

    /**Name of attribute that we use to avoid duplicate application of this filter on one filter chain. */
    private static final String ANTI_DUP_ATTR_NAME = "org.hd.d.pg2k.webSvr.util.CompressionFilter.DUPFLAG";

    /**Filters the server operation.
     * The <code>doFilter</code> method of the Filter is called by the container
     * each time a request/response pair is passed through the chain due
     * to a client request for a resource at the end of the chain.
     * The FilterChain passed into this method allows the Filter to pass on the
     * request and response to the next entity in the chain.
     * <p>
     * This method first examines the request to check whether the client supports
     * compression.
     * <p>
     * It passes the request and response unaltered if there is no support for
     * compression.
     * <p>
     * If compression support is available, this filter creates a
     * CompressionServletResponseWrapper object which compresses the content and
     * modifies the header if the content length is big enough.
     * It then invokes the next entity in the chain using the FilterChain object
     * (<code>chain.doFilter()</code>).
     **/
    public void doFilter(ServletRequest request, ServletResponse response,
                         FilterChain chain) throws IOException, ServletException
        {
        if(debug > 0)
            {
            System.out.println("@doFilter");
            }

        StatsLogger.captureDataPoint(statsIDCompression, "COMPFILT"); // Note comp filter invokation...

        boolean supportCompression = false;

        // Fake loop to break out of as soon as we determine
        // that we are *not* going to do compression...
        do
            {
            // Don't compress if not HTTP.
            if(!(request instanceof HttpServletRequest) ||
                !(response instanceof HttpServletResponse))
                {
                supportCompression = false; // No compression...
                break;
                }

            final HttpServletRequest httpServletRequest = ((HttpServletRequest) request);
            final HttpServletResponse httpServletResponse = ((HttpServletResponse) response);

            // Avoid duplicate application of a compression filter,
            // eg during servlet redirection.
            if(null != httpServletRequest.getAttribute(ANTI_DUP_ATTR_NAME))
                {
                supportCompression = false; // Avoid duplicate compression...
                StatsLogger.captureDataPoint(statsIDCompression, "NOCOMPDUP");
                break;
                }
            httpServletRequest.setAttribute(ANTI_DUP_ATTR_NAME, Boolean.TRUE);

            // Don't compress if application-level attribute forbids it.
            final ServletContext ctxt = config.getServletContext();
            if((overrideAttribute != null) &&
                Boolean.FALSE.equals(ctxt.getAttribute(overrideAttribute)))
                {
                supportCompression = false; // Compression is explicitly disabled.
                break;
                }

            // Don't compress if this output is already encoded in some way.
            // Amongst other things, this prevents back-to-back applications
            // of two compression streams which just results in garbage for the user.
            if(httpServletResponse.containsHeader("Content-Encoding"))
                {
                StatsLogger.captureDataPoint(statsIDCompression, "CE-ALREADY-SET"); // No compression possible...
if(IsDebug.isDebug || (debug > 1)) { System.err.println("WARNING: CompressionFilter saw Content-Encoding: aborting..."); }
                supportCompression = false;
                break;
                } // Already encoded.

            // We cannot apply any encoding/compression if there is a Content-MD5 header,
            // since it should be applied/recomputed after this compression,
            // and the GZIP checksum will probably do to protect the content in transit.
            if(httpServletResponse.containsHeader("Content-MD5"))
                {
                //StatsLogger.captureDataPoint(statsIDCompression, "CE-ALREADY-SET"); // No compression possible...
if(IsDebug.isDebug || (debug > 1)) { System.err.println("WARNING: CompressionFilter saw Content-MD5: aborting..."); }
                supportCompression = false;
                break;
                } // Already encoded.

            // Only compress if client can accept gzip encoding.
            final Enumeration e = httpServletRequest.getHeaders("Accept-Encoding");
            gzipSearch: while(e.hasMoreElements())
                {
                final String name = (String) e.nextElement();
                if(name.indexOf(GZIP_TOKEN) != -1)
                    {
                    final StringTokenizer st = new StringTokenizer(name, ", ");
                    while(st.hasMoreTokens())
                        {
                        final String tok = st.nextToken();
                        if(tok.equalsIgnoreCase(GZIP_TOKEN) ||
                            tok.equalsIgnoreCase("x-" + GZIP_TOKEN))
                            {
                            if(debug > 0)
                                {
                                System.out.println("supports compression");
                                }
                            supportCompression = true;
                            break gzipSearch;
                            }
                        }
                    }
                }
            } while(false);


        if(!supportCompression)
            {
            if(debug > 0)
                { System.out.println("doFilter gets called WITHOUT compression"); }
            chain.doFilter(request, response);
            StatsLogger.captureDataPoint(statsIDCompression, "NOCOMP"); // No compression possible...
            return;
            }
        else
            {
            if(response instanceof HttpServletResponse)
                {
                CompressionServletResponseWrapper wrappedResponse =
                    new CompressionServletResponseWrapper((HttpServletResponse) response);
                wrappedResponse.setCompressionThreshold(compressionThreshold);
                if(debug > 0)
                    {
                    System.out.println("doFilter gets called WITH compression");
                    }

                try
                    {
                    chain.doFilter(request, wrappedResponse);
                    }
                finally
                    {
                    wrappedResponse.finishResponse();
                    }
                StatsLogger.captureDataPoint(statsIDCompression, "CANCOMP"); // Compression possible...
                return;
                }
            }
        }
    }

