/* DERIVED FROM APACHE CODE, modifications (c) Damon Hart-Davis 2001-2006, under same terms:
 *
 * CompressionResponseStream.java
 * $Header: /rw/CVS-repository.s0.l/general/PG2Ksrc/javasrc/webSvr/org/hd/d/pg2k/webSvr/util/CompressionResponseStream.java,v 1.8 2004/12/21 19:03:34 dhd Exp $
 * $Revision: 1.8 $
 * $Date: 2004/12/21 19:03:34 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.hd.d.pg2k.webSvr.util;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;

import ORG.hd.d.IsDebug;

import java.io.*;


/**Implementation of <b>ServletOutputStream</b> that works with CompressionServletResponseWrapper implementation.
 * @author Amy Roh (original)
 * @author Damon Hart-Davis
 * @version $Revision: 1.8 $, $Date: 2004/12/21 19:03:34 $
 */

public final class CompressionResponseStream extends ServletOutputStream
    {
    // ----------------------------------------------------------- Constructors


    /**Construct a servlet output stream associated with the specified Response.
     * @param response  the associated response
     */
    public CompressionResponseStream(final CompressionServletResponseWrapper wrapper,
                                     final HttpServletResponse response)
        throws IOException
        {
        if((wrapper == null) || (response == null)) { throw new IllegalArgumentException(); }
        this.response = response;
        this.wrapper = wrapper;

        output = response.getOutputStream();
        if(output == null) { throw new IllegalStateException(); }
        }


    // ----------------------------------------------------- Instance Variables


    /**The threshold number which decides to compress or not; non-negative.
     * Users can configure in web.xml to set it to fit their needs.
     * <p>
     * If positive, size of output below which we will not attempt
     * to compress but will leave the output is it arrived.
     */
    private int compressionThreshold;

    /**Buffer in which we accumulate output before starting compressed stream.
     * Initially null,
     * created on demand if the the compressionThreshold is positive,
     * flushed into the compression stream and destroyed
     * if it overflows during a write() or if still extant and close().
     */
    private ByteArrayOutputStream buffer;

    /**The underlying gzip output stream to which we should write data.
     * If this is non-null then we are committed to writing the
     * stream compressed.
     * <p>
     * As an emergency swerve if we discover that someelse in the filter chain
     * decided late to add a content encoding,
     * this may in fact be a pass-through stream.
     */
    protected /*Deflater*/OutputStream compressedStream;

    /**Has this stream been closed?
     */
    protected boolean closed;

    /**The response with which this servlet output stream is associated; never null. */
    private final HttpServletResponse response;
    
    /**The wrapper that created this; never null. */
    private final CompressionServletResponseWrapper wrapper;

    /**The underlying servlet output stream to which we should write data; never null. */
    private final ServletOutputStream output;


    /**Minimum output buffer size to mop up output from GZIPOutputStream.
     * Without this, the leading and trailing bytes are liable to
     * end up as separate length==1 (HTTP/1.1) chunks because of the
     * implementation of GZIPOutputStream up to JDK1.3.0.
     * <p>
     * Using a BufferedOutputStream helps ensure that the chunks are
     * large and efficent both to represent and to transmit.
     */
    private static final int BUFSIZE = 8192;
    
    /**If true, use JazzLib's flush()able Deflator rather than the JDK's. */
    private static final boolean USE_JAZZLIB = true;

    /**If false, suppress flushes of the GZIP output stream itself until the stream is closed.
     * This is because a flush() generally cannot push out anything useful
     * with Sun's GZIPOutputStream up to an including JDK 5,
     * ie that the client could actually uncompress immediately,
     * and so just represents a waste of network packets and bandwidth.
     * <p>
     * If we use JazzLib then flush() is useful.
     */
    private static final boolean FLUSH_GZIP = USE_JAZZLIB;

    // --------------------------------------------------------- Public Methods


    /**Set the compressionThreshold; must not be negative.
     * Zero means always compress.
     * <p>
     * This value may be ignored, ie treated as if zero.
     */
    protected synchronized void setBuffer(final int threshold)
        {
        if(threshold < 0) { throw new IllegalArgumentException(); }
        compressionThreshold = threshold;
        }

    /**Close this output stream.
     * Cause any buffered data to be flushed and
     * any further output data to throw an IOException.
     */
    public synchronized void close()
        throws IOException
        {
        //System.out.println("close() @ CompressionResponseStream");
        if(closed)
            { throw new IOException("This output stream has already been closed"); }

        // In the case we didn't get to start compressing (too little or no data),
        // then write out any initial data we collected directly downstream.
        // Presumably we have just avoided wasted compression overhead.
        // We can also re-apply some Content-XXX headers to help transfer integrity.
        // Where we were compressing, correctly terminate the compression.
        if(buffer != null)
            {
            assert(compressedStream == null);
            reapplySuppledContentHeaders();
//if(IsDebug.isDebug) { System.out.println("INFO: close() pushing out initial buffer of size: "+buffer.size()); }
            output.write(buffer.toByteArray());
            buffer = null; // Allow GC.
            output.close();
            }
        // If we actually create a compression stream, close it and downstream.
        else if(compressedStream != null)
            {
            compressedStream.close();
            }
        // We seem to have written nothing at all.  Close downstream.
        else
            {
            reapplySuppledContentHeaders();
            output.close();
            }

        closed = true;
        }

    /**Flush any buffered data for this output stream, and force the response to be committed.
     * This may resist attempting the flush() the compressed stream
     * as it will have no effect other than to wastefully push out some
     * header bytes at most.
     * <p>
     * This will always ensure that the underlying stream
     * is directly or indirectly flushed so as to commit the response.
     */
    public synchronized void flush()
        throws IOException
        {
        if(closed) { throw new IOException("Cannot flush a closed output stream"); }

//if(WATCH_FLUSHES) { System.out.println("servlet.out.flush() called..."); }
//(new Throwable()).printStackTrace();

        if(FLUSH_GZIP)
            {
            // If we are not blocking all flush() operations entirely...
            makeSureGZOSCreated(); // Commit to use compression or not right now...
            if(buffer != null)
                {
//if(IsDebug.isDebug) { System.out.println("INFO: flush() pushing out initial buffer of size: "+buffer.size()); }
                writeToGZip(buffer.toByteArray(), 0, buffer.size());
                buffer = null; // Discard buffer and allow GC.
                }
            compressedStream.flush(); // Flush through compressor and thus commit downstream too.
            }
        else
            {
            output.flush(); // Commit downstream.
            }
        }

    /**Create the GZIPOutputStream if necessary and advertise the Content-Encoding in the HTTP headers.
     * It does nothing if the compressedStream is no longer null.
     * <p>
     * After this completes compressedStream will not be null.
     * <p>
     * Calling this commits us to whether we will use compression at all
     * for this HTTP request.
     * We <em>will</em> apply compression
     * <em>unless</em> we find another Content-Encoding is being applied.
     * <p>
     * We wrap the output in a BufferedOutputStream first to
     * soak up the single-byte leading and trailing writes that
     * GZIPOutputStream up to JDK 1.4.0 does, and to avoid sending
     * redundant blocks of data downstream (such as the gzip header)
     * before we have to, since these will otherwise get sent as
     * separate chunks/packets in some cases.
     * <p>
     * We also take this opportunity to add the "Content-Encoding" header
     * to indicate that the output is GZIPed, and to add a "Vary" header
     * to indicate that this content encoding was negotiated with the client
     * so caches/proxies can handle differently-able clients properly.
     * <p>
     * We double-check that the a Content-Encoding has not now been applied
     * since this might happen lazily.
     * If it has then we duck and dive by using the raw underlying stream.
     */
    private synchronized void makeSureGZOSCreated()
        throws IOException
        {
        if(compressedStream != null) { return; }

        // We cannot apply any encoding/compression if there is a Content-MD5 header,
        // since it should be applied/recomputed after this compression,
        // and the GZIP checksum will probably do to protect the content in transit.
        // Last-ditch check that this chain/page/etc is not getting multiply ZIPped...
        // We DO NOT re-apply deferred Content-XXX header values.
        if(((HttpServletResponse) response).containsHeader("Content-MD5") ||
           ((HttpServletResponse) response).containsHeader("Content-Encoding"))
            {
if(IsDebug.isDebug) { System.err.println("WARNING: CompresionFilter saw Content-Encoding/Content-MD5: LATE abort/swerve..."); }
            compressedStream = output;
            return;
            }

        // We buffer behind the compressor stream for network efficiency,
        // eg to ensure that its headers and trailers don't get sent separately.
        final int bufsize = Math.max(BUFSIZE, compressionThreshold);

        // Set our compressor.
        compressedStream = new FlushableGZIPOutputStream(output, bufsize);

        // Note that we now have a content encoding in place.
        response.addHeader("Content-Encoding", CompressionFilter.GZIP_TOKEN);

        // OK, compresion depends on content negotiation with the client,
        // then we'd better add a Vary header to avoid mucking up any proxy/cache...
        response.addHeader("Vary", "Accept-Encoding");

//if(IsDebug.isDebug) { System.out.println("INFO: compression filter started..."); }
        }
    
    /**Reapply user-supplied header values that would have conflicted with compression.
     * We can reapply these Content-XXX headers to improve on-the-wire integrity.
     * <p>
     * We carefully avoid overwriting any values that have been set elsewhere.
     */
    private void reapplySuppledContentHeaders()
        {
        final HttpServletResponse httpServletResponse = ((HttpServletResponse) response);
        // Cannot do anything if response already committed.
        if(httpServletResponse.isCommitted())
            {
if(IsDebug.isDebug) { System.out.println("CompressionResponseStream.reapplySuppledContentHeaders(): response already committed."); }
            return;
            }

        // Content-Length
        if((wrapper.getContentLength() > 0) &&
           !httpServletResponse.containsHeader("Content-Length"))
            {
if(IsDebug.isDebug) { System.out.println("CompressionResponseStream.reapplySuppledContentHeaders(): re-applying Content-Length."); }
            httpServletResponse.setContentLength(wrapper.getContentLength());
            }

        // Content-MD5
        if((wrapper.getContentMD5() != null) &&
           !httpServletResponse.containsHeader("Content-MD5"))
            {
if(IsDebug.isDebug) { System.out.println("CompressionResponseStream.reapplySuppledContentHeaders(): re-applying Content-MD5."); }
            httpServletResponse.setHeader("Content-MD5", wrapper.getContentMD5());
            }
        }

    /**Private per-instance buffer for write(int); never null. */
    private final byte buf1[] = new byte[1];

    /**Write the specified byte to our output stream.
     * This is likely to be hideously inefficient compared to the multi-byte write() calls.
     *
     * @param b The byte to be written
     * @exception IOException if an input/output error occurs
     */
    public synchronized void write(final int b) throws IOException
        {
        buf1[0] = (byte) b;
        write(buf1, 0, 1);
        }

    /**Write the specified byte array to our output stream.
     * @param b The byte array to be written; never null
     * @exception IOException if an input/output error occurs
     */
    public void write(final byte b[]) throws IOException
        { write(b, 0, b.length); }

    /**Write the specified array portion to our output stream.
     * All write()s to this output stream ultimately pass through this method.
     *
     * @param b  the byte array containing the bytes to be written
     * @param off  zero-relative starting offset of the bytes to be written
     * @param len  the number of bytes to be written
     *
     * @exception IOException if an input/output error occurs
     */
    public synchronized void write(final byte b[], final int off, final int len)
        throws IOException
        {
        if(closed)
            throw new IOException("Cannot write to a closed output stream");

        if(len == 0) { return; } // Nothing to do...
        
//System.out.println("servlet.out.write(b, off, "+len+") called...");

        // If the compressing output stream does not yet exist,
        // and the threshold is positive,
        // we buffer internally (creating the buffer if need be).
        // Calling flush() may push this buffered data down the wire.
        if((compressedStream == null) && (compressionThreshold > 0))
            {
            final int bufferedBytes = (buffer == null) ? 0 : buffer.size();

            // If data would not overflow buffer (ie not pass threshold)
            // then create buffer if need be,
            // append new data,
            // and return.
            if(len + bufferedBytes <= compressionThreshold)
                {
                if(buffer == null)
                    { buffer = new ByteArrayOutputStream(compressionThreshold); }
                buffer.write(b, off, len);
                return;
                }

            // If buffer would overflow because we would pass our threshold;
            // then write buffer content to the compressed stream (and discard buffer),
            // then write the new data to the compressed stream.
            if(buffer != null)
                {
                writeToGZip(buffer.toByteArray(), 0, buffer.size());
                buffer = null; // Allow GC.
                }

            // Fall through to write new data directly
            // to compressed stream.
            //
            // This will ensure that compressedStream!= null,
            // so we will not try to use the buffer again.
            }

        // Write to the compressedStream.
        writeToGZip(b, off, len);
        }

    /**Write the specified array portion to our compressed GZIP output stream.
     * @exception IOException if an input/output error occurs
     */
    private synchronized void writeToGZip(byte b[], int off, int len) throws IOException
        {
        makeSureGZOSCreated();
//System.out.println("writeToGZip(b, off, "+len+")...");
        compressedStream.write(b, off, len);
        }
    }
