/*
 * Created by IntelliJ IDEA.
 * User: d@hd.org
 * Date: 25-Aug-02
 * Time: 19:35:35

Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package org.hd.d.pg2k.webSvr.util;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;

import ORG.hd.d.IsDebug;

import java.io.IOException;

/**Explicitly allows browser to cache filtered items.
 * This sets values that explicitly
 * allow the client to cache the item for a longish time
 * (of the order of hours to weeks) on the assumption that
 * the filtered item changes slowly.
 * <p>
 * This should work even when the items are delivered over HTTPS,
 * which would normally prevent them being cached.
 * <p>
 * This should work for HTTP/1.0 clients (we set Expires), and
 * HTTP/1.1 client (we set Cache-Control).
 * <p>
 * This filter is stateless and as lightweight and simple as possible.
 */
public final class ExplicitCacheTimeFilter implements Filter
    {
    /**Bring filter into service.
     */
    public void init(FilterConfig filterConfig)
        {
        }

    /**Take this filter out of service.
     */
    public void destroy()
        {
        }

    /**Headers that we set to make cacheing work.
     * We reject attempts to override any of these (and warn on System.err).
     * <p>
     * These should be matched case-insensitively.
     */
    private static final String DANGER_HEADERS[] =
        { "Expires", "Pragma", "Cache-Control" };

    /**Name of attribute that we use to avoid duplicate application of this filter on one filter chain. */
    private static final String ANTI_DUP_ATTR_NAME = "org.hd.d.pg2k.webSvr.util.ExplicitCacheTimeFilter.DUPFLAG";

    /**Filter one item.
     * If none of the cache-controlling headers for HTTP/1.0 or 1.1 are
     * already set, set them ourselves to a reasonably lengthy time.
     */
    public void doFilter(final ServletRequest request,
                         final ServletResponse response,
                         final FilterChain chain)
        throws IOException, ServletException
        {
        // Avoid duplicate application of this filter,
        // eg during servlet redirection.
        if(null != request.getAttribute(ANTI_DUP_ATTR_NAME))
            {
            // Avoid counting the same traffic more than once.
if(IsDebug.isDebug) { System.err.println("WARNING: avoiding duplicate ExplicitCacheTimeFilter application to " + request); }
            chain.doFilter(request, response);
            return;
            }
        request.setAttribute(ANTI_DUP_ATTR_NAME, Boolean.TRUE);

        // If this is not an HTTP response, do not fiddle with it.
        if(!(response instanceof HttpServletResponse) ||
           !(request instanceof HttpServletRequest))
            {
            chain.doFilter(request, response);
            return;
            }

        final HttpServletResponse hsr = (HttpServletResponse) response;

        // If any of the `danger' headers is already set,
        // indicating that the cache behaviour is already being
        // explicitly handled/manipulated,
        // back out immediately!
        for(int i = DANGER_HEADERS.length; --i >= 0; )
            {
            if(hsr.containsHeader(DANGER_HEADERS[i]))
                {
if(ORG.hd.d.IsDebug.isDebug) { System.err.println("ExplicitCacheTimeFilter: WARNING: backing off setting cache headers (having found header "+DANGER_HEADERS[i]+") for: " + ((HttpServletRequest) request).getRequestURI()); }
                chain.doFilter(request, response);
                return;
                }
            }

        // Make wrapper to force headers to end up set the way we want...
        final class Wrapper extends HttpServletResponseWrapper
            {
            /**Pass the original response to the base class.
             * We also immediately set the values we want set.
             * We complain later if there is any attempt to set any of them.
             */
            Wrapper(HttpServletResponse hsResponse)
                {
                // Pass in original response...
                super(hsResponse);

                // Time to allow cacheing, in ms.
                final long cacheMs = WebConsts.DEFAULT_STATIC_WEBITEMS_CACHE_MS;

//System.out.println("Setting cache headers (for "+cacheMs+"ms) for: " + ((HttpServletRequest) request).getRequestURI());

                // Add Expires for HTTP/1.0 clients and
                // add Cache-Control for HTTP/1.1 client.
                // Maybe we should wrap the response with a modified one,
                // rather than modifying the existing one.
                hsResponse.setHeader("Cache-Control", "public,max-age="+(cacheMs/1000));
                hsResponse.setDateHeader("Expires", System.currentTimeMillis() + cacheMs);
                }

            public void setHeader(String s, String s1)
                {
                for(int i = DANGER_HEADERS.length; --i >= 0; )
                    {
                    if(DANGER_HEADERS[i].equals(s))
                        {
System.err.println("ExplicitCacheTimeFilter: WARNING: vetoing attempt to set header "+DANGER_HEADERS[i]+" for: " + ((HttpServletRequest) request).getRequestURI());
                        return;
                        }
                    }
                super.setHeader(s, s1);
                }

            public void addHeader(String s, String s1)
                {
                for(int i = DANGER_HEADERS.length; --i >= 0; )
                    {
                    if(DANGER_HEADERS[i].equals(s))
                        {
System.err.println("ExplicitCacheTimeFilter: WARNING: vetoing attempt to add header "+DANGER_HEADERS[i]+" for: " + ((HttpServletRequest) request).getRequestURI());
                        return;
                        }
                    }
                 super.addHeader(s, s1);
                }
            }

        // Continue, with new explicit cacheing headers in place.
        chain.doFilter(request, new Wrapper(hsr));
        }
    }

