/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

import java.io.BufferedOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.CRC32;
import java.util.zip.Deflater;

/**Substitute for GZIPOutputStream that maximises compression and has a usable flush().
 * This is also more careful about its output writes for efficiency,
 * and indeed buffers them to minimise the number of write()s downstream
 * which is especially useful where each write() has a cost
 * such as an OS call, a disc write, or a network packet.
 */
public class FlushableGZIPOutputStream extends net.sf.jazzlib.DeflaterOutputStream
    {
    private final CRC32 crc = new CRC32();
    private final static int GZIP_MAGIC = 0x8b1f;
    private final OutputStream os;

    /**Note when input has arrived and not yet been compressed and flushed downstream. */
    private boolean somethingWritten;

    /**Creates flush()able equivalent of GZIPOutputStream, with maximum/best compression and default output buffer size.
     * Useful for, for example, HTTP compression.
     * @param os  underlying output stream; must not be null
     */
    public FlushableGZIPOutputStream(final OutputStream os)
        throws IOException
        { this(os, 8192); }

    /**Creates flush()able equivalent of GZIPOutputStream, with maximum/best compression.
     * Useful for, for example, HTTP compression.
     * @param os  underlying output stream; must not be null
     * @param bufsize  output buffer size; must be positive
     */
    public FlushableGZIPOutputStream(final OutputStream os, final int bufsize)
        throws IOException
        {
        super(new FilterOutputStream(new BufferedOutputStream(os, bufsize)){
            /**Suppress inappropriate/inefficient flush()es by DeflaterOutputStream. */
            @Override public void flush() { }
            },
            new net.sf.jazzlib.Deflater(net.sf.jazzlib.Deflater.BEST_COMPRESSION, true));
        this.os = os;
        writeHeader();
        crc.reset();
        }

    /**All output is written via this routine. */
    public synchronized void write(byte[] buf, int off, int len)
        throws IOException
        {
//if(WATCH_FLUSHES) { System.out.println("compressedStream.out.write(b, off, "+len+")"); }
        somethingWritten = true;
        super.write(buf, off, len);
        crc.update(buf, off, len);
        }

    /**Flush any accumulated input downstream in compressed form.
     * We overcome some bugs/misfeatures here so that:
     * <ul>
     * <li>We won't allow the GZIP header to be flushed on its own
     *     without real compressed data in the same write downstream.
     * <li>We ensure that any accumulated uncompressed data really is forced
     *     through the compressor.
     * <li>We prevent spurious empty compressed blocks being produced
     *     from successive flush()es with no intervening new data.
     * </ul>
     */
    @Override
    public synchronized void flush() throws IOException
        {
        // If no uncompressed input data has been written
        // (since our last flush or since creation)
        // then avoid trying to flush anything out now,
        // which also avoids forcing out the GZIP header on its own.
        if(!somethingWritten) { return; }

        // We call this to get def.flush() called,
        // but suppress the (usually premature) out.flush() called internally.
        super.flush();

        // Since super.flush() seems to fail to reliably force output, 
        // possibly due to over-cautious def.needsInput() guard following def.flush(),
        // we try to force the issue here by bypassing the guard.
        int len;
        while((len = def.deflate(buf, 0, buf.length)) > 0)
            { out.write(buf, 0, len); }

        // Really flush the stream below us...
//if(WATCH_FLUSHES) { System.out.println("downstream.out.flush()"); }
        os.flush();

        // Further flush()es ignored until more input data data written.
        somethingWritten = false;
        }

    /**Flush out any remaining compressed data, add the trailer, and close the stream. */
    public synchronized void close()
        throws IOException
        {
//        boolean writtenTrailer = false;

        if(!def.finished())
            {
            def.finish();
            do
                {
                int len = def.deflate(buf, 0, buf.length);
                if(len <= 0) { break; }

//                // If there is room for the trailer at the end of the buffer
//                // and we've just finished deflating everything
//                // then sneak in the trailer here to save one write.
//                if((len + TRAILER_BYTES <= buf.length) && def.finished())
//                    {
//                    System.arraycopy(generateTrailer(), 0, buf, len, TRAILER_BYTES);
//                    writtenTrailer = true;
//                    len += TRAILER_BYTES;
//                    };

                out.write(buf, 0, len);
                } while(!def.finished());
            }

        // Write trailer iff only not already done.
//        if(!writtenTrailer)
            { out.write(generateTrailer()); }

        out.close();
        }

    /**Constant (10-byte) header array.
     * All zeros except magic number and algorithm.
     */
    private static final byte[] header = {
        (byte) GZIP_MAGIC, // Intel order, LSB first.
        (byte) (GZIP_MAGIC >> 8),
        (byte) Deflater.DEFLATED,
        0, 0, 0, 0, 0, 0, 0
        };

    /**Write header directly downstream. */
    private void writeHeader()
        throws IOException
        {
        // Single write for efficiency.
        out.write(header);
        }

    /**Length of the trailer in bytes. */
    private static final int TRAILER_BYTES = 8;

    /**Return header to write downstream. */
    private byte[] generateTrailer()
//            throws IOException
        {
        final byte b[] = new byte[TRAILER_BYTES];
        final int crcValue = (int) crc.getValue();
        b[0] = (byte) (crcValue); // Intel order, LSB first.
        b[1] = (byte) (crcValue >> 8);
        b[2] = (byte) (crcValue >> 16);
        b[3] = (byte) (crcValue >> 24);
        final int totalIn = def.getTotalIn();
        b[4] = (byte) (totalIn); // Intel order, LSB first.
        b[5] = (byte) (totalIn >> 8);
        b[6] = (byte) (totalIn >> 16);
        b[7] = (byte) (totalIn >> 24);

        assert(b.length == TRAILER_BYTES);
        return(b);
        }
    }