/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;
import org.hd.d.pg2k.svrCore.vars.EventPeriod;
import org.hd.d.pg2k.svrCore.vars.EventVariableValue;
import org.hd.d.pg2k.svrCore.vars.SimpleVariableDefinition;
import org.hd.d.pg2k.svrCore.vars.SystemVariables;
import org.hd.d.pg2k.webSvr.exhibit.*;

import javax.servlet.ServletContext;
import java.io.IOException;
import java.text.DateFormat;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.ReentrantLock;

/**Utility functions to generate HTML inserts for showing batches of thumbnails/links.
 * One advantage of having a method here rather than in-line in a JSP
 * is that this code is pre-compiled off-line for speed and robustness.
 */
public final class HTMLThumbnailInsertGenerators
    {
    /**Prevent construction of an instance. */
    private HTMLThumbnailInsertGenerators() { }


    /**Minimum width for column display in pixels; strictly positive. */
    public static final int COLDISP_MIN_WIDTH_PIXELS =
        ExhibitThumbnails.SML_STATIC_IMAGE_TN_LDIM_PX +
        2 * PageSkinUtils.GENERIC_TN_BORDER_WIDTH_PIXELS +
        10; // Fudge factor.

    /**Minimum height for column display in pixels; strictly positive.
     * Taken to be the height of display for one exhibit
     * with a small thumbnail.
     */
    public static final int COLDISP_MIN_HEIGHT_PIXELS =
        ExhibitThumbnails.SML_STATIC_IMAGE_TN_LDIM_PX +
        2 * PageSkinUtils.GENERIC_TN_BORDER_WIDTH_PIXELS +
        90; // Fudge factor.

    /**Assumed approximate mean pixels wide per (small) character; strictly positive. */
    public static final int COLDISP_MEAN_CHAR_WIDTH_PIXELS = 7;

    /**Assumed approximate mean pixels wide per (small) character; strictly positive. */
    public static final int COLDISP_MEAN_CHAR_HEIGHT_PIXELS = 12;

    /**Fraction of "newest" exhibits getNewExhibitSelection() draws from; strictly positive.
     * Typically in the range 10--50 (ie 10% to 2%).
     */
    private static final int NEW_FRACTION = 19;

    /**Cache of new exhibits; private to getNewExhibitSelection(). */
    private static final FilterBean _newE =
       new FilterBean();

    /**Initialise _newE. */
    static
        {
        _newE.setMemorySensitiveCache(false); // Hold onto this data.

        // We filter this (on creation) to be:
        //   * Sorted, oldest first...
        HTMLThumbnailInsertGenerators._newE.setExpr(new SortExpr(null,
            new BuiltInFilters.sortByTimestamp()
            ));
        }

    /**Cache of new exhibits; private to getRandomJPEGSelection(). */
    private static final FilterBean _rndJPEGE =
       new FilterBean();

    /**Initialise _rndJPEGE. */
    static
        {
        HTMLThumbnailInsertGenerators._rndJPEGE.setMemorySensitiveCache(false); // Hold onto this data.

        // We filter this (on creation) to be:
        //   * JPEG images.
        //   * Too large to be thumbnails themselves (ie not very small).
        HTMLThumbnailInsertGenerators._rndJPEGE.setExpr(new FilterExpr(null, (new FilterIF()
            {
            final public boolean accept(final AllExhibitProperties aep, final String exhibitName)
                {
                final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);

                // Quick-n-dirty filter to exclude very small images...
                if(esa.length <= ExhibitThumbnails.STD_ABS_MAX_BYTES)
                    { return (false); }

                // Quick filter to exclude non-JPEG exhibits.
                final ExhibitMIME.ExhibitTypeParameters et = (ExhibitMIME.getInputFileType(esa.filePath));
                if((et == null) || (et.type != ExhibitMIME.ET_JPEG))
                    { return (false); }

                return (true); // Looks OK...
                }
            /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
            private static final long serialVersionUID = -5173930708308579233L;
            })));
        }

    /**Time before which we will not attempt incremental update of _bestE set.
     * Set so as to limit fraction of system CPU spent recomputing/updating.
     * <p>
     * Is volatile so as to allow access without a lock.
     * <p>
     * Private to getBestExhibitsSelection().
     */
    private static volatile long _bestE_dontUpdateBefore;

    /**Cache of new exhibits; private to getBestExhibitsSelection(). */
    private static final FilterBean _bestE = new FilterBean();

    /**Lock to prevent more than one update run on _bestE at once. */
    private static final ReentrantLock _bestE_update_lock = new ReentrantLock();

    /**Initialise _bestE. */
    static
        {
        // Use a dummy/empty GenProps for our filtering.
        final GenProps gp = new GenProps();

        // Hold onto this data since it is expensive to recompute.
        HTMLThumbnailInsertGenerators._bestE.setMemorySensitiveCache(false);

        // We filter this (on creation) to be:
        //   * Exhibits better than average/neutral "goodness".
        // We sort this to be best (highest-goodness) first.
        HTMLThumbnailInsertGenerators._bestE.setExpr(new SortExpr(new FilterExpr(null, (new FilterIF()
            {
            final public boolean accept(final AllExhibitProperties aep, final String exhibitName)
                {
                // Make do with whatever value is already computed...
                final ExhibitPropsComputableMutable epcm =
                        aep.getExhibitPropsComputableMutable(exhibitName, true, gp, null, null);

                // True only if better than neutral.
                if((epcm != null) && (epcm.getGoodness() > 0))
                    { return(true); }

                // Not good enough.
                return(false);
                }

            /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
            private static final long serialVersionUID = 671635642661710967L;
            })),
            // Limit number of items to be considered "best" for efficiency,
            // to somewhat more than just the "first page" of results
            // so that we can show some exhibits beyond it.
            new BuiltInFilters.sortByGoodness(5*WebConsts.SINGLE_PAGE_CONTACT_SHEET_TN_COUNT + 1)));

        // Set this to automatically timeout and recompute periodically.
        // Aim to be at about the default system temporal slackness
        // or the computable-mutable recalculation time if less.
        // If very small then there may be a performance impact.
        // If too high then we will not reflect changes in popularity quickly.
        HTMLThumbnailInsertGenerators._bestE.setExpiryInterval(Math.min(2*CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S,
                                                 ExhibitPropsComputableMutable.MAX_AGE_BEFORE_STALE_MS/2503) *
                                        (500 + Rnd.fastRnd.nextInt(1000)));
        }


    /**Makes HTML for columnar (ie taller and thin) display of exhibits; never null.
     * Given the width and height for display,
     * and a list of exhibits,
     * will format and show as many as is reasonably possible in the space.
     * Exhibits from the start of the list will be shown first.
     * <p>
     * Text is horizontally centred.
     * <p>
     * This may not show any exhibits at all if the area is too small.
     * <p>
     * This will not work well if the space is very narrow or shallow,
     * much less than about 100 pixels.
     * <p>
     * Note that this may not get things exactly right if the user
     * selects different font sizes in their browser.
     *
     * @param exhibits  non-null list of full, valid exhibit names,
     *     more important first (later items may not be displayed)
     * @param dataSource  non-null source of exhibit data and thumbnails
     * @param pixelsWidth  non-negative width of display space in pixels
     * @param pixelsHeight  non-negative height of display space in pixels
     * @param reduceEffort  if true, reduce CPU and bandwidth load
     *     in executing this routine and implied by the output
     *
     * @throws java.io.IOException  if there is difficulty retrieving meta-data
     */
    public static String makeHTMLExhibitColumnDisplay(final String exhibits[],
                                                      final int pixelsWidth,
                                                      final int pixelsHeight,
                                                      final DataSourceBean dataSource,
                                                      final LocaleBeanBase localeBean,
                                                      final boolean reduceEffort)
        throws IOException
        {
        if((exhibits == null) ||
           (pixelsWidth < 0) ||
           (pixelsHeight < 0) ||
           (dataSource == null) ||
           (localeBean == null))
            { throw new IllegalArgumentException(); }

        // If too narrow then return an empty string.
        if((pixelsWidth < COLDISP_MIN_WIDTH_PIXELS) ||
            (pixelsHeight < COLDISP_MIN_HEIGHT_PIXELS))
            { return(""); }

        // Compute maximum number of exhibits we could display vertically.
        final int maxVertical = pixelsHeight / COLDISP_MIN_HEIGHT_PIXELS;

        // Compute the actual number of exhibits we will show.
        final int toShow = Math.min(maxVertical, exhibits.length);
        assert(toShow > 0);

        // Try to get all thumbnails pre-loaded/computed up-front, concurrently.
        final BitSet exists = WebUtils.exhibitsHaveThumbnail(dataSource,
                        Arrays.asList(exhibits).subList(0, toShow),
                        false, // Small thumbnails.
                        !reduceEffort);

        // Get the exhibit properties.
        final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);

        final StringBuilder result = new StringBuilder(exhibits.length * 101);
        result.append("<table align=center cellpadding=0 cellspacing=0>");

        for(int i = 0; i < toShow; ++i)
            {
            final String exhibitName = exhibits[i];

            // Check for a bad exhibit name...
            if(exhibitName == null)
                { throw new IllegalArgumentException(); }

            // Get this exhibit's properties; never null.
            final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);

            // Skip entry for a no-longer-extant exhibit,
            // eg after an AEP change.
            if(esa == null)
                { continue; }

            // Start new row for new exhibit.
            result.append("<tr><td align=center height=").
                append(COLDISP_MIN_HEIGHT_PIXELS).
                append('>');

            // Get the exhibit type...
            final ExhibitMIME.ExhibitTypeParameters exhibitType = (ExhibitMIME.getInputFileType(esa.filePath));

            // Compute the relative URL to the catalogue page; never null.
            final String catPage = WebUtils.makeCatPageRRURL(exhibitName,
                                    WebConsts.F_secondary_generated_HTML_suffix);

            // Try and get a thumbnail, which must be an inlineable image.
            final java.awt.Dimension thumbnailXyDim = new java.awt.Dimension();
            final String thumbnailRRURL = (!exists.get(i)) ? null :
                WebUtils.makeHTMLInlineImageThumbnailURL(
                    dataSource,
                    exhibitName,
                    false, // Get a small thumbnail.
                    true, // Get relative URL.
                    thumbnailXyDim,
                    reduceEffort);

            // If we have a thumbnail, insert it.
            if(thumbnailRRURL != null)
                {
                // AKA for title...
                final String specificAKA = GenUtils.getLocalisedTreeDesc(aep, exhibitName, localeBean, false, true, false, true);
                final boolean hasSpecificAKA = (specificAKA.length() != 0);

                result.append("<table align=center border=").
                    append(PageSkinUtils.GENERIC_TN_BORDER_WIDTH_PIXELS).
                    append("><tr><td>").
                    append("<a href=\"").append(catPage).append("\" target=_top>").
                    append("<img src=\"").append(thumbnailRRURL).append('\"').
                        append(" width=").append(thumbnailXyDim.width).
                        append(" height=").append(thumbnailXyDim.height).
                        append(" alt=").append(UploaderUtils.quoteHTMLArg(localeBean.getLocalisedMessage("common.thumbnail")));
                        if(hasSpecificAKA) { result.append(" title=\"").append(specificAKA.replace('"', '\'')).append('"'); }
                        result.append(WebConsts.TN_IMG_TAG_EXTRA_ATTR).append('>').
                    append("</a></td></tr></table>");
                }
            // Put in some summary info in lieu of a thumbnail.
            else
                {
                if(exhibitType != null)
                    { result.append(exhibitType.description); }
                }

            // Force line-break after thumbnail.
            result.append("<br />");

            // Link to the catalogue page.
            result.append("<a href=\"").
                append(catPage).
                append("\" target=_top>").
                append(TextUtils.sanitiseForXML(ExhibitName.getFileComponent(exhibitName),
                                               pixelsWidth / COLDISP_MEAN_CHAR_WIDTH_PIXELS,
                                               false)).
                append("</a>");

            // Force line-break after link.
            result.append("<br />");

            // Show size and date.
            result.append(TextUtils.sizeAsText(esa.length, true)).append("; ");
            result.append("<nobr>").
                append(DateFormat.getDateInstance(DateFormat.MEDIUM, localeBean.getLocale()).format(new Date(esa.timestamp))).
                append("</nobr>");

            result.append("</td></tr>");
            }

        result.append("</table>");
        return(result.toString());
        }

    /**Empty (immutable) array to return to indicate no results. */
    private static final String[] NO_RESULTS = new String[0];

    /**Get random selection of new-ish exhibit full names; never null.
     * Selects from newest few percent of available exhibits,
     * and within those the very newest are most likely to be returned.
     * Note that no absolute age/time limits are imposed.
     * <p>
     * If there are any exhibits at all,
     * and maxExhibits is positive,
     * this routine ensures that at least one exhibit is a candidate to show.
     * <p>
     * These results are cached statically,
     * and automatically discarded/recomputed when the exhibit set changes.
     * <p>
     * May return empty list, but no nulls nor duplicates.
     * <p>
     * Should only be used with one servlet context.
     *
     * @param application  the servlet context; never null
     * @param maxExhibits  the maximum number of exhibits to return;
     *     non-negative
     * @param rnd          the random number source or null to pick top items
     */
    public static String[] getNewExhibitSelection(final ServletContext application,
                                                  final int maxExhibits,
                                                  final Random rnd)
        throws IOException
        {
        if((application == null) ||
           (maxExhibits < 0))
            { throw new IllegalArgumentException(); }

        // Get full list of exhibits newest first.
        final List<String> exhibits = _newE.select(application);

        final int ne = exhibits.size();

        // If no suitable exhibits at all,
        // then return immediately.
        if(ne < 1) { return(NO_RESULTS); }

        // Heavily skew selection towards very newest exhibits...
        final int candidates = Math.min(ne, 11+(ne/NEW_FRACTION));
        final String result[] = new String[Math.min(candidates, maxExhibits)];

        // Pick a starting exhibit.
        // (Forced to be zero if no more exhibits than requested, or no random source.)
        int which = ((rnd == null) || (candidates <= maxExhibits)) ?
            0 : rnd.nextInt(1 + rnd.nextInt(1 + candidates - maxExhibits));

        for(int i = 0; i < result.length; ++i)
            {
            result[i] = exhibits.get(ne - ++which); // Select from the HIGH index items.
            }

        return(result);
        }

    /**Get random selection of JPEG exhibit full names; never null.
     * These results are cached statically,
     * and automatically discarded/recomputed when the exhibit set changes.
     * <p>
     * May return empty list, but no nulls nor duplicates.
     * <p>
     * Should only be used with one servlet context.
     *
     * @param application  the servlet context; never null
     * @param maxExhibits  the maximum number of exhibits to return;
     *     non-negative
     */
    public static String[] getRandomJPEGSelection(final ServletContext application,
                                                  final int maxExhibits)
        throws IOException
        {
        if((application == null) ||
           (maxExhibits < 0))
            { throw new IllegalArgumentException(); }

        // Get full list of exhibits in possibly-strange order.
        final List<String> exhibits = _rndJPEGE.select(application);

        final int ne = exhibits.size();

        // If no suitable exhibits at all,
        // then return immediately.
        if(ne < 1) { return(NO_RESULTS); }

        final String result[] = new String[Math.min(maxExhibits, ne)];
        // Pick a starting exhibit...
        int which = Rnd.fastRnd.nextInt(ne);
        int loc = 0;

        // ...and wrap around from it...
        for(int i = result.length; --i >= 0; )
            {
            final String exhibitName = exhibits.get(ne - 1 - which);
            result[loc++] = exhibitName;
            if(++which >= ne) { which = 0; }
            }

        return(result);
        }


    /**If true then invalidate "best" collection if any stale value is found.
     * Otherwise, only invalidate if there is a significant value change.
     * (If not then we can survive until the next periodic re-sort.)
     */
    private static final boolean CHECK_FOR_STALE = false;

    /**Out-of-order tolerance we will allow before forcing "best" selection to be recomputed.
     * This corresponds to an insignificantly small difference
     * that will rarely be noticed by a (non-obsessive!) human,
     * so need not force us to invalidate and re-sort the entire collection.
     */
    private static final int BEST_EPSILON = Integer.MAX_VALUE / (1 + WebConsts.SINGLE_PAGE_CONTACT_SHEET_TN_COUNT);

    /**Get names of the the best exhibits ("goodness" greater than zero); never null.
     * These results are cached statically,
     * and automatically discarded/recomputed when the exhibit set changes,
     * and also periodically to allow for the values changing over time.
     * <p>
     * May return an empty list, but no nulls nor duplicates.
     * <p>
     * Should only be used with one servlet context.
     * <p>
     * If a random number source is supplied then it is used to chose
     * a random (though weighted towarded best) selection of exhibits.
     * <p>
     * If no random number source is supplied then the very best exhibits
     * are returned.
     * <p>
     * (May invalidate its cache and spend a short time updating entries
     * if any of the items that it is about to return is stale
     * or the results are no longer in order (best first)
     * in the hope that non-stale ones will have been computed in the mean time.
     * This incremental cache validation and recomputation should help
     * keep the results reasonable without penalising any one caller too much,
     * and returns reasonable results in a short time on each call.)
     *
     * @param application  the servlet context; never null
     * @param maxExhibits  the maximum number of exhibits to return;
     *     non-negative
     * @param rnd          random number source, else null to pick top items
     */
    public static String[] getBestExhibitSelection(final ServletContext application,
                                                   final int maxExhibits,
                                                   final Random rnd)
        throws IOException
        {
        if((application == null) ||
           (maxExhibits < 0))
            { throw new IllegalArgumentException(); }

        // Note start time of routine,
        // as just fetching/updating the results may take some time.
        final long startTime = System.currentTimeMillis();

        // Get full list of "good" exhibits best first.
        final List<String> exhibits = _bestE.select(application);

        final int ne = exhibits.size();

        // If no suitable exhibits at all,
        // then return immediately.
        if(ne < 1) { return(NO_RESULTS); }

        final String result[] = new String[Math.min(maxExhibits, ne)];
        // Pick a starting exhibit.
        // (Forced to be zero if no more exhibits than requested, or no random source.)
        int which = ((rnd == null) || (ne <= maxExhibits)) ?
            0 : rnd.nextInt(1 + rnd.nextInt(1 + ne - maxExhibits));
//        // Note if we are selecting the "best-of-best" exhibits.
//        final boolean isFromTop = (which == 0);

        // Copy the exhibits in order to the result array.
        for(int i = 0; i < result.length; ++i)
            { result[i] = exhibits.get(which++); }

        // If it is too soon to consider doing an update check
        // then return our result immediately.
        if(startTime <= _bestE_dontUpdateBefore) { return(result); }

        // If an update is already in progress
        // then return our result immediately.
        if(!_bestE_update_lock.tryLock()) { return(result); }
        try
            {
            // Recompute any stale (or now-not-good) items found
            // in the list that we are about to return
            // (plus at least one other at random if possible).

            // Try to keep run-time of routine below page-generation time limit.
            // The more time that we can spend in one go, however,
            // the more efficient any incremental updating will be.
            final long stopBy = startTime + WebConsts.MAX_PG_DOWNLOAD_MS/2;

            // We look at (and try to bring up-to-date)
            // at least one entry before giving up
            // so as to try to ensure some progress
            // if stale entries do need recomputing.
            // We try to ensure that at least whatever we rated as
            // the "best" item in this result set
            // (first in the array) is non-stale by the time we have finished.
            final DataSourceBean dsb = DataSourceBean.getApplicationInstance(application);
            final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
            final GenProps gp = dsb.getGenProps(-1);

            // Flag to be set non-null if we need to invalidate the cache.
            String invalidate = null;

            // Compute goodness of current "best" item.
            final int bestGoodness = aep.getExhibitPropsComputableMutable(exhibits.get(0), false, gp, dsb, dsb.getScorerCache()).getGoodness();
            // Compute goodness of current "worst-best" item.
            final int worstGoodness = aep.getExhibitPropsComputableMutable(exhibits.get(ne-1), false, gp, dsb, dsb.getScorerCache()).getGoodness();

            // Some early funnies to check...
            if((bestGoodness <= 0) || (worstGoodness <= 0) || (worstGoodness > bestGoodness))
                { invalidate = "best/worst entries invalid or mis-ordered"; }

            // Sample/compute goodness of one item entirely at random.
            // This forces incremental update one way or another,
            // at least of items already in the putative "best" set.
            // We won't invalidate the whole set just because this was stale.
            final String sampleItem = exhibits.get(Rnd.fastRnd.nextInt(ne));
            final ExhibitPropsComputableMutable sampleEPCM = aep.getExhibitPropsComputableMutable(sampleItem, false, gp, dsb, dsb.getScorerCache());
            if(sampleEPCM == null) { invalidate = "sample entry invalid: " + sampleItem; }
            else
                {
                final int sampleGoodness = sampleEPCM.getGoodness();
                if((sampleGoodness < worstGoodness) || (sampleGoodness > bestGoodness))
                    { invalidate = "sample entry out of range wrt best/worst: " + sampleGoodness; }
                }

            // Try to ensure all values for results are non-stale and in order.
            // We don't care if we had to re-compute an entry
            // as long as it is still roughly in place.
            for(int i = 0; i < result.length; ++i) /* Check "best" items first. */
                {
                // See if we have a stale, missing, or out-of-order item...
                final String exhibitName = result[i];
                final ExhibitPropsComputableMutable epcm = aep.getExhibitPropsComputableMutable(
                        exhibitName, CHECK_FOR_STALE, gp, dsb, dsb.getScorerCache());
                final int goodness = (epcm == null) ? Integer.MIN_VALUE : epcm.getGoodness();
                if(goodness <= 0)
                    {
                    // Something weird happened and this entry
                    // is missing or now inadmissable,
                    // so we must re-filter and re-sort the cache next time.
                    invalidate = "returned entry is missing/invalid: " + exhibitName;
                    if(epcm != null) { aep.getExhibitPropsComputableMutable(exhibitName, false, gp, dsb, dsb.getScorerCache()); }
                    else { break; /* Avoid blowing up later tests with NPEs! */ }
                    }
                else if(CHECK_FOR_STALE && epcm.isStale())
                    {
                    // Force this item up-to-date...
                    aep.getExhibitPropsComputableMutable(exhibitName, false, gp, dsb, dsb.getScorerCache());
                    // ...and make sure that we recompute our cache next time.
                    invalidate = "returned entry had stale EPCM: " + exhibitName;
                    }
                else if((i < result.length) && (i > 0) &&
                        (goodness > BEST_EPSILON + aep.getExhibitPropsComputableMutable(result[i-1], false, gp, dsb, dsb.getScorerCache()).getGoodness()))
                    {
                    // Each item in the result array must be
                    // no more "good" than its predecessor
                    // (allowing an "epsilon" tolerance).
                    // If not so, then entries are mis-ordered,
                    // (so must have been recomputed)
                    // then we must re-sort the cache next time.
                    invalidate = "returned entry is out of order wrt neighbour: " + exhibitName + ", index: "+i;
                    }
                else if((goodness > bestGoodness) || (goodness < worstGoodness))
                    {
                    // If we've found one better than the current "best"
                    // or one worse than the current "worst best"
                    // then we must re-sort the cache next time.
                    invalidate = "returned entry is out of range wrt best/worst: " + exhibitName;
                    }

                // Give up now if we've taken a noticable time...
                if(System.currentTimeMillis() > stopBy) { break; }
                }

            // If we are going to invalidate the whole cache,
            // then spend some time bringing entries up to date first
            // to make the recomputation process as efficient as possible.
            // We try to make sure that even if we run out of time that
            // we still get good coverage by systematic and random sampling.
            if(invalidate != null)
                {
                for(int i = ne; (--i >= 0) && (System.currentTimeMillis() <= stopBy); )
                    {
                    // Employ all available threads to update these EPCM values.
                    // We assume these to be CPU-intensive.
                    // We assume that when all other compute threads are busy
                    // then we will end up computing whatever is left on our thread,
                    // so we will not use resources or queue tasks indefinitely.
                    final int next = i;
                    ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
                        public final void run()
                            {
                            aep.getExhibitPropsComputableMutable(exhibits.get(next), false, gp, dsb, dsb.getScorerCache());
                            aep.getExhibitPropsComputableMutable(exhibits.get(Rnd.fastRnd.nextInt(ne)), false, gp, dsb, dsb.getScorerCache());                            
                            }
                    });
                    }
                }

            // Some feedback on the effort being invested!
//if(IsDebug.isDebug && (invalidate != null)) { System.out.println("[getBestExhibitSelection(): invalidating 'best' exhibit cache; reason='"+invalidate+"', size="+ne+", best="+bestGoodness+", worst="+worstGoodness+", out of "+aep.aeid.length+" exhibits; recompute took "+(System.currentTimeMillis()-startTime)+"ms (maxExhibits="+maxExhibits+").]"); }
//            if(invalidate != null) { application.log("[getBestExhibitSelection(): invalidating 'best' exhibit cache; reason='"+invalidate+"', size="+ne+", best="+bestGoodness+", worst="+worstGoodness+", out of "+aep.aeid.length+" exhibits; recompute took "+(System.currentTimeMillis()-startTime)+"ms (maxExhibits="+maxExhibits+").]"); }

            // Do cache invalidation at most once,
            // so as to minimise cost to concurrent threads for example.
            if(invalidate != null) { _bestE.invalidate(); }
            }
        catch(final Exception e)
            {
            // Absorb any transient errors from data set shifting beneath us
            // eg a now-invalid exhibit name may cause an NPE.
            application.log("WARNING: exception in getBestExhibitSelection(): " + e.getMessage());
            // Force recomputation of data in case it is now boobytrapped.
            _bestE.invalidate();
            }
        finally
            {
            // Put off next update to limit overall CPU time used.
            // Note that we only really care about keeping up-to-date
            // the entries that are "good".
            final long doneAt = System.currentTimeMillis();
            final long computeTime = doneAt - startTime;
//System.err.println("[getBestExhibitsSelection() done check() [invalidate="+invalidate+"] in "+computeTime+"ms.]");
            final long putOffBy = Math.min(13*computeTime, // I can bear ~10% on updating...
                69000 + Rnd.fastRnd.nextInt(39503)); // Cap the delay time...
            _bestE_dontUpdateBefore = doneAt + putOffBy; /* Volatile update is atomic. */

            // Let the lock go now that we're done...
            _bestE_update_lock.unlock();
            }

        return(result);
        }


    /**If true then always omit from the contact sheet any items that do not immediately/ever have a thumbnail. */
    private static final boolean _CS_ALWAYS_ONLY_SHOW_THUMBNAILS = false;

    /**If true then try to cache cat page contact sheets; this may require minor cosmetic changes in generated HTML for efficiency. */
    public static final boolean CS_CACHE_FOR_CAT_PAGES = true;

    /**Private key used by getContactSheetHTML(); never null. */
    private static final DataSourceBean.AEPLinkedKey _CSCacheKey = new DataSourceBean.AEPLinkedKey("CSCacheKey");

    /**Make "contact print" / "preview" HTML fragment to show a user all similar exhibits, "" if none possible; never null.
     * This routine returns a table of (small thumbnails of) all the exhibits
     * with the same main words prefix as the supplied word, in order,
     * in as compact and efficient form as reasonable.
     * <p>
     * This may choose to omit or show substitute text for exhibits with
     * no immediately-available thumbnail.
     * <p>
     * This does not generate any headings.
     * <p>
     * If this encounters an error (eg an exception, or inconsistent arguments)
     * then it will attempt to return gracefully a "no-contact-sheet-available"
     * value of "".
     * <p>
     * This may spend a small amount of time trying to create thumbnails,
     * bring up-to-date computable data, etc, but in any case tried to be quick.
     * <p>
     * Has special HTML tag at end if incomplete due to missing thumbnails
     * or if we ran out of time.
     * <p>
     * This may cache completed results against all of the exhibit names given,
     * to be discarded if the AEP for the supplied DataSourceBean changes.
     *
     * @param exhibitName  this name of the main exhibit for which a contact sheet
     *     is to be generated; must be a valid, non-null exhibit name
     * @param specialColour  HTML colour name (or #RRGGBB value) to distinguish
     *     the main exhibit in the contact sheet somehow,
     *     quoted if it need be for an attribute value,
     *     may be ignored for efficiency/consistency;
     *     null if no such marking to be done
     * @param dataSource  data source for thumbnails, etc; never null
     * @param smartSortedNames  list of all exhibit names,
     *     sorted with the SMART_ORDER Comparator,
     *     and should be random-access else this will probably be horribly slow;
     *     never null
     *
     * @return "" if a contact sheet is not applicable or cannot be generated
     *     (so result is safe to drop into HTML without checking if need be)
     *     else a well-formed HTML fragment suitable to drop into a catalogue page;
     *     never null
     */
    @SuppressWarnings("unchecked")
    public static String getContactSheetHTML(final String exhibitName,
                                             final String specialColour,
                                             final DataSourceBean dataSource,
                                             final List<String> smartSortedNames)
        {
        if(!ExhibitName.validNameSyntax(exhibitName) ||
           (dataSource == null) ||
           (smartSortedNames == null))
            { throw new IllegalArgumentException(); }
        
        final int maxPixelsWidth = (PageSkinUtils.GENERIC_WIDTH_PX*9)/10; // About 90% to match other tables...

//        try
            {
            final Comparator<String> exhibitComparator = ExhibitAttrUtils.getAttrWords().SMART_ORDER;

            // Compute the main-words component of the main exhibit.
            // We'll potentially display in the contact sheet all exhibits
            // with the same main-words component.
            final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();
            final String mainWords =
                ExhibitName.getMainWordsComponent(exhibitName, attrWords);

            // If cacheing then first ensure that our cache map exists...
            // The cache, if used, maps from main-words component to HTML.
            // The HTML text is stored as a String or a Compact7BitString
            // (or more generally, something that can recover the tex using toString()).
            // We use a thread-safe map allowing full read concurrency
            // and a small amount of write concurrency.
            ConcurrentMap<String,Object> cache;
            if(CS_CACHE_FOR_CAT_PAGES)
                {
                while(null == (cache = (ConcurrentMap<String,Object>) dataSource.getAEPLinkedValue(_CSCacheKey)))
                    { dataSource.putIfAbsentAEPLinkedValue(_CSCacheKey, new ConcurrentHashMap<String,Object>(128+2*smartSortedNames.size(), 0.7f, 4)); }
                }

            // See if we have the value cached already.
            // (A "" value indicates an empty result that need not be recomputed.)
            // If not then compute it.
            final Object cachedValue;
            String result;
            if(!CS_CACHE_FOR_CAT_PAGES || (null == (cachedValue = cache.get(mainWords))))
                {
                final int ssnLen = smartSortedNames.size();
                final int ourPos = Collections.binarySearch(smartSortedNames, exhibitName, exhibitComparator);
                if((ourPos < 0) || (ourPos >= ssnLen))
                    { return(""); /* The main exhibit cannot be found; exit gracefully and never cache this result. */ }

                // Find the first exhibit that matches our main words component.
                // (We'll ignore case differences here regardless for how the smart-sort works.)
                int firstMatch = ourPos;
                for(int i = ourPos; --i >= 0; )
                    {
                    final String ex = smartSortedNames.get(i);
                    final String mwc = ExhibitName.getMainWordsComponent(ex, attrWords);
                    if(!mwc.equalsIgnoreCase(mainWords))
                        { break; }
                    firstMatch = i; // This matched, so move the limit down...
                    }

                // Now find the last matching exhibit.
                int lastPos;
                for(lastPos = ourPos+1 ; lastPos < ssnLen; ++lastPos)
                    {
                    final String ex = smartSortedNames.get(lastPos);
                    final String mwc = ExhibitName.getMainWordsComponent(ex, attrWords);
                    if(!mwc.equalsIgnoreCase(mainWords))
                        { break; }
                    }

                // Note sublist of matching entries.
                // Take copy if we might have to erase entries...
                final List<String> matchingNames = _CS_ALWAYS_ONLY_SHOW_THUMBNAILS ?
                    new ArrayList<String>(smartSortedNames.subList(firstMatch, lastPos)) :
                    smartSortedNames.subList(firstMatch, lastPos);

                // If there is only one matching name,
                // then it is our main exhibit,
                // and there is no contact sheet to show.
                if(matchingNames.size() <= 1)
                    {
                    if(CS_CACHE_FOR_CAT_PAGES) { cache.put(mainWords, ""); }
                    return(""); /* No sheet to show. */
                    }

                // If we only want to show in our sheet
                // thumbnails available here and now,
                // then weed any others out of our list.
                if(_CS_ALWAYS_ONLY_SHOW_THUMBNAILS)
                    {
                    for(final Iterator<String> it = matchingNames.iterator(); it.hasNext(); )
                        {
                        final String ex = it.next();
                        // Zap items without an extant thumbnail...
                        if(!WebUtils.exhibitHasThumbnail(dataSource, ex, false, false))
                            { it.remove(); }
                        }

                    // If nothing is left in our list
                    // then return an empty sheet immediately.
                    if(matchingNames.size() < 1)
                        {
                        if(CS_CACHE_FOR_CAT_PAGES) { cache.put(mainWords, ""); }
                        return(""); /* No sheet to show. */
                        }
                    // If the only thing left in our list is our main exhibit
                    // then return an empty sheet immediately.
                    if((matchingNames.size() == 1) && exhibitName.equals(matchingNames.get(0)))
                        {
                        if(CS_CACHE_FOR_CAT_PAGES) { cache.put(mainWords, ""); }
                        return(""); /* No sheet to show. */
                        }
                    }

                // Target time to stop doing optional computations by;
                // probably short enough to avoid annoying interactive users
                // while giving us the opportunity to get most/all thumbnails
                // in the contact sheet.
                final long startTime = System.currentTimeMillis();
                final long stopBy = startTime + WebConsts.MAX_PG_DOWNLOAD_MS/2;

                // Return our generated HTML fragment...
                result = generateGenericContactSheetHTML(maxPixelsWidth, matchingNames, CS_CACHE_FOR_CAT_PAGES ? null : specialColour, exhibitName, stopBy, dataSource);

                // If we are cacheing and the result is complete,
                // then capture the result, in a space-efficient form if possible.
                if(CS_CACHE_FOR_CAT_PAGES && !result.endsWith(CS_INCOMPLETE))
                    {
                    try { cache.putIfAbsent(mainWords, Compact7BitString.convertToCompact7BitString(result, null)); }
                    catch(final IllegalArgumentException e) { cache.putIfAbsent(mainWords, result); }
                    }
                }
            // Use the (non-null) cached value.
            else { result = cachedValue.toString(); }

            return(result);
            }
//        catch(final IOException e)
//            {
//            // Error encountered, so give up gracefully.
//            return("");
//            }
        }

    /**Comment tag at end of HTML for contact sheet if we ran out of time or it is otherwise incomplete. */
    public static final String CS_INCOMPLETE = "<!--INCOMPLETE-->";

    /**Generate contact sheet of the exhibits given, in order; never null.
     * Has special HTML tag at end if incomplete due to missing thumbnails
     * or if we ran out of time.
     *
     * @param exhibitNames  in-order list of exhibits to display; never null
     * @param specialExhibitName  name of "special" exhibit to make stand out;
     *     null if none
     * @param stopBy  approximate target time to finish by;
     *     can be set in the past to generate HTML as quickly as possible
     * @param specialColour  HTML colour name (or #RRGGBB value) to distinguish
     *     the main exhibit in the contact sheet somehow,
     *     quoted if it need be for an attribute value;
     *     null if no such marking to be done
     * @param maxPixelsWidth  maximum pixels available for display (approx);
     *     strictly positive
     * @param dataSource  data source for thumbnails, etc; never null
     * @return  HTML fragment for contact sheet, or ""; never null
     */
    public static String generateGenericContactSheetHTML(final int maxPixelsWidth,
                                                         final List<String> exhibitNames,
                                                         final String specialColour,
                                                         final String specialExhibitName,
                                                         final long stopBy,
                                                         final DataSourceBean dataSource)
        {
        if((maxPixelsWidth <= 0) ||
                (exhibitNames == null) ||
                (dataSource == null))
            { throw new IllegalArgumentException(); }

        if(exhibitNames.size() == 0)
            { return(""); }

        try
            {
            final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);

            // Calculate how many columns of thumbnails we may be able to show on this page.
            // We assume that each thumbnail may be its maximum in both directions.
            // We always show at least one column.
            final int borderRoundEachThumbnailPx = 1; // Allow this many pixels' border.
            final int tnSize = ExhibitThumbnails.SML_STATIC_IMAGE_TN_LDIM_PX;
            final int colWidth = (tnSize + 2*borderRoundEachThumbnailPx);
            final int maxCols = Math.max(1, maxPixelsWidth / colWidth);
            final int actualCols = Math.min(maxCols, exhibitNames.size());

            // Generate the HTML as a table.
            final StringBuilder result = new StringBuilder(64 + 256*exhibitNames.size());
            boolean firstRow = true; // Are we producing the first row?
            int columnNumber = 0; // Which column are we about to fill?

            final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

            // Start the table...
            result.append("<table cellpadding=0 border=").append(borderRoundEachThumbnailPx).
                    // Try to help the brower to lay out the table quickly...
                    append(" cols=").append(actualCols).
                    append(" width=").append(actualCols * colWidth).
                    append('>');

            // Start the first row.
            final String rowStartHTML = "<tr height="+tnSize+" align=center>";
            result.append(rowStartHTML);

            // Set true if we ran out of time
            // or couldn't get a thumbnail that should be available.
            boolean incomplete = false;

            boolean reduceEffort = false;

            // Find out what thumbnails are to hand locally right now,
            // possibly forcing them.
            // This may be able to check 
            final BitSet thumbnails = WebUtils.exhibitsHaveThumbnail(dataSource, exhibitNames, false, true);

            final int n = exhibitNames.size();
            for(int i = 0; i < n; ++i)
                {
                final String ex = exhibitNames.get(i);
                
                // If we have run off the end of this row,
                // then finish this row and start a new one,
                // and note that we are not on the first row any more.
                if(columnNumber++ >= maxCols)
                    {
                    result.append("</tr>").append(rowStartHTML);
                    columnNumber = 1; // Note that we are just about to fill column 0 of the new row below...
                    firstRow = false;
                    }

                // Make this cell with a link to the target catalogue page.
                // On the first row only, force the width too.
                // Mark the main exhibit with a special background colour if so requested.
                result.append("<td");
                if(firstRow)
                    { result.append(" width=").append(tnSize); }
                if((specialColour != null) && (ex.equals(specialExhibitName)))
                    { result.append(" bgcolor=").append(specialColour); }
                result.append('>');

                // Extract part of name that distinguishes this exhibit from its peers.
                final int bp[] = ExhibitName.getMainAndAttrWordComponentBoundaries(ex, attrWords);
                // Make unique-ish summary with first word and tail attributes...
                final String uniquePart = ExhibitName.getMainWords(ex, attrWords).nextElement() + " ~ " + ex.substring(bp[1]+1);
                // Make unique-ish bit that is easier to word-wrap.
                final String uWW = uniquePart.replace(ExhibitName.WORD_SEP, ' ');

                // Have we run out of time and need to reduce effort spent?
                // Once true, stays true for this call.
                if(!reduceEffort)
                    { reduceEffort = System.currentTimeMillis() >= stopBy; }

                // Compute the relative URL to the catalogue page; never null.
                final String catPage = WebUtils.makeCatPageRRURL(ex,
                                        WebConsts.F_secondary_generated_HTML_suffix);

                // Try and create HTML for an extant thumbnail,
                // which must be an HTML inlineable image/object.
                final java.awt.Dimension thumbnailXyDim = new java.awt.Dimension();
                final String thumbnailRRURL = (!thumbnails.get(i)) ? null :
                    WebUtils.makeHTMLInlineImageThumbnailURL(
                        dataSource,
                        ex,
                        false, // Get a small thumbnail.
                        true, // Get relative URL.
                        thumbnailXyDim,
                        reduceEffort);

                // If we have a thumbnail then insert it.
                if(thumbnailRRURL != null)
                    {
                    result.append("<a href=\"").append(catPage).append("\">").
                        append("<img src=\"").append(thumbnailRRURL).append('\"').
                            append(" border=0").
                            append(" width=").append(thumbnailXyDim.width).
                            append(" height=").append(thumbnailXyDim.height).
                            append(" alt=").append(UploaderUtils.quoteHTMLArg(uWW)).append('>').
                        append("</a>");
                    }
                // Put in some summary info in lieu of a thumbnail.
                else
                    {
                    // Get this exhibit's properties; never null.
                    final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(ex);

                    // Get the exhibit type...
                    final ExhibitMIME.ExhibitTypeParameters exhibitType =
                            (esa == null) ? null : (ExhibitMIME.getInputFileType(esa.filePath));

                    result.append("<small>");

                    if(exhibitType != null)
                        { result.append(exhibitType.description).append("<br />"); }

                    result.append("<a href=\"").append(catPage).append("\"><i>");
                    result.append(uWW); // Make word-wrap possible on (eg) FF1.
                    result.append("</i></a>");
                    result.append("</small>");

                    // If this apparently *should* have had a thumbnail
                    // then mark as our result as incomplete.
                    if((exhibitType != null) && (exhibitType.canPossiblyCreateThumbnailOfSameMIMEType()))
                        { incomplete = true; }
                    }

                result.append("</td>");
                }

            // End the last row...
            result.append("</tr>");

            // End the table...
            result.append("</table>");

            // If we had to reduce effort
            // then mark the result as incomplete.
            if(incomplete)
                { result.append(CS_INCOMPLETE); }

            return(result.toString());
            }
        catch(final Exception e)
            {
            return(""); // In case of error return an empty result.
            }
        }

    /**Generate a narrow column of popular/useful internal links for HTML/XHTML, "" if none; never null.
     * @param dsb  data source; never null
     * @param pixelsHigh  approx number of pixels high; strictly positive
     * @param pixelsWide  approx number of pixels wide; strictly positive
     */
    public static String popularUsefulLinksNarrowCol(final DataSourceBean dsb,
                                                     final int pixelsHigh,
                                                     final int pixelsWide,
                                                     final boolean toXHTML)
        {
        if((dsb == null) ||
           (pixelsHigh <= 0) || (pixelsWide <= 0))
            { throw new IllegalArgumentException(); }

        try
            {
            final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
            final GenProps gp = dsb.getGenProps(-1);

            final int maxLinks = pixelsHigh / COLDISP_MEAN_CHAR_HEIGHT_PIXELS;
            final int maxChars = pixelsWide / COLDISP_MEAN_CHAR_WIDTH_PIXELS;

            // If no space to show anything then return...
            if((maxLinks < 1) || (maxChars < 1)) { return(""); }

            // Collect map from sorted clipped link text to (full) names of exhibits
            // to link to catalogue pages for.
            // These are kept sorted in a reasonable order.
            final SortedMap<String, String> exhibits = new TreeMap<String, String>(String.CASE_INSENSITIVE_ORDER);

            // Periods/intervals' data that we will use, in order.
            final long currentInterval = EventPeriod.VLONG.getIntervalNumber(System.currentTimeMillis());
            final long periods[] =
                {
                currentInterval-1, // Previous period (cheap, stable, reasonably topical).
                currentInterval, // Current period (more topical, more expensive, more noisy).
                0, // "All" period (less topical, more expensive, more comprehensive).
                };

            // System events that we will use, in order.
            // Randomise these a little each time.
            final SimpleVariableDefinition defs[] =
                {
        		(Rnd.fastRnd.nextBoolean() ?
	                SystemVariables.ACCESSPATTERN_CLICKTHROUGH : // Usefulness to users and possible profit to us!
	                SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD_LOCAL), // Local popularity.
                (Rnd.fastRnd.nextBoolean() ?
                    SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD : // Pure popularity.
                    SystemVariables.ACCESSPATTERN_CAT_PAGE_VIEW), // Pure popularity.
                };

            final Map<String,String> shortToFullMap = aep.aeid.getShortToFullMap();

            // Pick items from further down the lists as we go...
            // Only to show the best item for a given displayed text/link...
            topLoop: for(int rank = 0; rank < maxLinks; ++rank)
                {
                for(int p = 0; p < periods.length; ++p)
                    {
                    for(final SimpleVariableDefinition def : defs)
                        {
                        // If we have enough links then break out of all loops!
                        if(exhibits.size() >= maxLinks) { break topLoop; }

                        final EventVariableValue[] eventValues = dsb.getEventValues(def, EventPeriod.VLONG, periods[p], null);
                        if((eventValues.length == 0) || (eventValues[0] == null)) { continue; }
                        final EventVariableValue eventValue = eventValues[0];
                        if(rank >= eventValue.getTotalDistinctValues()) { continue; }

                        // Only use the exhibit if it seems (still) to be valid.
                        final Object exhibitShortName = eventValue.getValueByRank(rank);
                        if(!(exhibitShortName instanceof String)) { continue; }
                        final String sn = (String) exhibitShortName;  
                        // Skip potentially-sensitive subjects.
                        if(GenUtils.isSensitive(sn, aep, gp)) { continue; }
                        // Extract the full name.
                        final String fullName = shortToFullMap.get(sn);
                        if(fullName != null)
                            {
                            final String linkText = TextUtils.sanitiseForXML(sn, maxChars, false);

                            // If we have an extant item with this link text
                            // then keep that with the better score.
                            // Do NOT force an expensive recomputation of either score.
                            final String extant = exhibits.get(linkText);
                            if(extant != null)
                                {
                                try
                                    {
                                    if(aep.getExhibitPropsComputableMutable(extant, true, gp, dsb, dsb.getScorerCache()).getGoodness() >=
                                       aep.getExhibitPropsComputableMutable(fullName, true, gp, dsb, dsb.getScorerCache()).getGoodness())
                                        { continue; /* New exhibit is no better than extant. */ }
                                    }
                                // Allow for transients/races causing NPEs, etc.
                                catch(final Exception e)
                                    { continue; /* In case of error keep extant value. */ }
                                }

                            // Use the new value.
                            exhibits.put(linkText, fullName);
                            }
                        }
                    }
                }

            if(exhibits.size() == 0) { return(""); } // Nothing to show...

            // Generate the HTML...
            final StringBuilder sb = new StringBuilder(32 + exhibits.size()*4*maxChars);
            sb.append("<small>");
            for(final String linkText : exhibits.keySet())
                {
                final String fullName = exhibits.get(linkText);

                // Should always be a syntactically-valid full exhibit name.
                assert(ExhibitName.validNameSyntax(fullName));
                // Should be a currently-extant exhibit.
                assert(null != aep.aeid.getStaticAttr(fullName));

                // Compute the relative URL to the catalogue page; never null.
                final String catPage = WebUtils.makeCatPageRRURL(fullName,
                                toXHTML ? WebConsts.F_secondary_generated_XHTML_suffix : WebConsts.F_secondary_generated_HTML_suffix);

                // Link to the catalogue page.
                sb.append("<a href=\"").
                    append(catPage).
                    append("\">").
                    append(linkText).
                    append("</a><br />");
                }
            sb.append("</small>");
            return(sb.toString());
            }
        catch(final Exception e)
            {
            e.printStackTrace();
            return(""); // Give up quietly in case of error.
            }
        }
    }
