/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

import org.hd.d.pg2k.svrCore.HostUtils;
import org.hd.d.pg2k.svrCore.I18NTools;
import org.hd.d.pg2k.svrCore.LocaleBeanBase;
import org.hd.d.pg2k.svrCore.location.GeoUtils;
import org.hd.d.pg2k.svrCore.props.LocalProps;
import org.hd.d.pg2k.svrCore.stats.StatsLogger;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.net.InetAddress;
import java.util.Locale;

/**JavaBean encapsulating notions of locale for JSP/servlet pages.
 * This should be created with request scope.
 * <p>
 * This puts off loading the properties bundle(s) until necessary,
 * since it can suck CPU cycles.
 * <p>
 * If the user has a session in progress an explicit locale is looked for there,
 * else the user's locale is obtained from the HTTP request object
 * if the Accept-Languages header is present
 * (which resorts to the server's locale if the user has not specified one),
 * else we attempt to guess a locale given the user's IP address.
 * <p>
 * Access to this object is thread-safe, though since this object
 * should appear at request (or page) level we do not actually expect
 * threading to be an issue.
 * <p>
 * This is Serializable so as to be able to be stored in
 * a servlet session; nothing especially long-lived or sensitive.
 */
public final class LocaleBean extends LocaleBeanBase
    {
    /**Name that should be used for the unique servlet-contect application-level instance. */
    public static final String BEAN_NAME = "localeBean";

    /**If true, we may default the locale to that of the server location rather than the client, if no explicit locale is specified.
     * This may help SEs distinguish between different mirrors
     * which may have some different i18n-ed content to offer.
     * <p>
     * This is only in cases where no locale is specified, usually by SEs,
     * whose spiders are predominantly in the US.
     */
    private static final boolean SHOW_SERVER_ACCENT = true;

    /**If true then note when we seem to be visited by a client with multiple locale preferences set.
     * Probably very unusual, but if it becomes common we should handle it well.
     */
    private static final boolean LOG_MULTI_LOCALE_CLIENTS = false;


    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final static WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**The stats set to which we log HTTP-client preferred language code counts.
     * This is a set of two-letter ISO 639 lower-case codes,
     * with the default being the server's locale.
     * <p>
     * We attempt to ignore requests from spiders,
     * ie we try to log requests only from live humans.
     * <p>
     * We may enable or disable this logging (etc) from system parameters,
     * but this is nonetheless our unique identifier.
     */
    private final static StatsLogger.StatsConfig statsIDLang =
        new StatsLogger.StatsConfig("WEB-USER-LOCALE",
                                    logger, // Use servlet log if poss.
                                    false, // Only dump summaries...
                                    12 * 3600, // About every 12 hours.
                                    true); // Adaptive.

    /**The stats set to which we log HTTP-client explicitly-set language code counts.
     * This is a set of two-letter ISO 639 lower-case codes.
     * <p>
     * We attempt to ignore requests from spiders,
     * ie we try to log requests only from live humans.
     * <p>
     * We may enable or disable this logging (etc) from system parameters,
     * but this is nonetheless our unique identifier.
     */
    private final static StatsLogger.StatsConfig statsIDLangExplicit =
        new StatsLogger.StatsConfig("WEB-USER-LOCALE-EXPLICIT",
                                    logger, // Use servlet log if poss.
                                    false, // Only dump summaries...
                                    12 * 3600, // About every 12 hours.
                                    true); // Adaptive.

    /**If true then try to guess locale from user's IP address if no explicit locale indicated in HTTP request. */
    private static final boolean ALLOW_LOCALE_GUESS_FROM_IP_ADDR = true;

    /**Cached request object.
     * Marked volatile so as to be safe to access without a lock.
     */
    private transient volatile HttpServletRequest cachedRequest;

    /**Set the current HTTP request to initialise the bean.
     * This allows us to retrieve any locale information that we need.
     * <p>
     * This first looks to see if there is a session
     * containing a locale-override value;
     * if so then it is used.
     * No session is created if none exists.
     * <p>
     * Else, this looks for a locale set by the user's browser,
     * and uses that if present.
     * <p>
     * Else, this tried to guess a suitable locale from the user's IP address,
     * using the IP-to-location tools.
     * <p>
     * Else a `safe' locale is used, usually the server's.
     * <p>
     * If the request is null or some other invalid condition is
     * encountered then this resets the local to a safe value
     * (usually the server's locale).
     * <p>
     * If the new cached request is identical to the old one,
     * then nothing is done.
     * <p>
     * @param request  if non-null is used to try to determine the user's locale
     * @param ctxt  if non-null selects the servlet log to write to
     *     (if null, output will go to System.out)
     */
    public void setRequest(final HttpServletRequest request,
                           final ServletContext ctxt)
        {
        // If we've already done the work for this request then return now!
        if(cachedRequest == request) { return; }

        try
            {
            // Whoops, no request at all, so use a safe locale.
            if(request == null)
                {
                userLocale = SAFE_LOCALE;
                return;
                }

            // Divert any output to the correct servlet log...
            logger.setContext(ctxt);

            // Try first for session-based override.
            final SessionVarBean svb = SessionVarBean.getExtantSessionVars(request, false);
            if(svb != null)
                {
                final Locale l = svb.getSessionVarLocale();
                if((l != null) && !l.equals(userLocale))
                    {
                    userLocale = l;

                    // Log locale (well, language actually) from session,
                    // but only for apparent real humans...
                    if(!WebUtils.requestProbablyFromSpider(request))
                        { StatsLogger.captureDataPoint(statsIDLangExplicit, userLocale.toString()); }

                    return;
                    }
                }

            // With no explicit client-specified locale(s),
            // possibly try giving the server a local accent
            // (especially for spiders when an explicit mirror is requested),
            // or speak with the client's probable locale given their location.
            final String headerAcceptLang = request.getHeader("Accept-Language");
            if(headerAcceptLang == null)
                {
                final String serverName = request.getServerName();
                final String mirrorTag;
                final Locale mirrorLocale;
                if(SHOW_SERVER_ACCENT &&
                   (serverName != null) && HostUtils.isMirrorName(serverName) &&
                   WebUtils.requestProbablyFromSpider(request) &&
                   ((mirrorTag = LocalProps.getMirrorTag()) != null) &&
                   ((mirrorLocale = I18NTools.LOCALE_BY_CCTLD.get(new GeoUtils.CCTLD(mirrorTag.substring(0, 2)))) != null))
                    {
                    // Will speak with my local "accent"...
                    userLocale = mirrorLocale;

                    // Note in stats.
                    StatsLogger.captureDataPoint(statsIDLang, "LOCAL-" + mirrorLocale);
                    return;
                    }
                else if(ALLOW_LOCALE_GUESS_FROM_IP_ADDR)
                    {
                    // Try to guess user's locale
                    // from client IP and thus user's country/location.
//                    final long resStart = System.currentTimeMillis();
                    try
                        {
                        final InetAddress clientAddr = InetAddress.getByName(request.getRemoteAddr());
                        // Try quick lookup first
                        // since all this delays the user...
                        GeoUtils.CCTLD cctld = GeoUtils.getCCTLDByAddress(clientAddr, true);
//                        // Try slow lookup if quick one does not help us.
//                        // We hope that second and subsequent lookups
//                        // for a given user will be cached and thus fast.
//                        if(cctld == null)
//                            { cctld = GeoUtils.getCCTLDByAddress(clientAddr, false); }

                        // If we got a ccTLD then look for a locale for it.
                        if(cctld != null)
                            {
                            // See if we know a default locale for this country.
                            final Locale l = I18NTools.LOCALE_BY_CCTLD.get(cctld);
                            if(l != null)
                                {
                                // OK, guessed locale from IP address.
                                userLocale = l;

                                // Note in stats.
                                StatsLogger.captureDataPoint(statsIDLang, "NONE-" + l);
                                return;
                                }
                            }
                        }
                    catch(final Exception e)
                        {
                        e.printStackTrace();  // Whinge but drop through to continue.
                        }
//                    finally
//                        {
//                        final long resEnd = System.currentTimeMillis();
//                        final long resTime = resEnd - resStart;
//                        // If resolve/lookup time was significant then log it...
//                        if(resTime > 1)
//                            { StatsLogger.captureDataPoint(statsIDLang, "TIME-RES-ccTLD-" + GenUtils.log2Approx(resTime)); }
//                        }

                    // No browser-set nor geo-guessable-from-IP-address locale.
                    StatsLogger.captureDataPoint(statsIDLang, "NONE");

                    // Fall through to try other methods anyway.
                    }
                }

            else if(LOG_MULTI_LOCALE_CLIENTS)
                {
                if(headerAcceptLang.indexOf(',') != -1)
                    {
                    // Log potential multi-locale client.
                    StatsLogger.captureDataPoint(statsIDLang, "MULTIPLE");
                    }
                }

            // Finally try for the browser-set locale-preference value(s)
            // (which falls back to our Web-server hard-coded default).
            // We validate this a little before accepting it,
            // and normalise too (eg by dropping any "variant" portion).
            // The language component must be present (and of length 2)
            // and the country must be absent or (in lower case) a valid ccTLD.
            // TODO: be prepared to try user 2nd-and-later-choice locales.
            final Locale l = request.getLocale();
            if((l != null) && (l.getLanguage().length() == 2) &&
               (I18NTools.LOCALES.contains(l) ||
                "".equals(l.getCountry()) ||
                GeoUtils.CCTLD.isSyntaticallyValidCcTLD(l.getCountry().toLowerCase())))
                {
                userLocale = new Locale(l.getLanguage().toLowerCase(), l.getCountry().toUpperCase(), "");

                // Log locale from HTTP request
                // iff this does not look like a spider.
                if(!WebUtils.requestProbablyFromSpider(request))
                    {
                    if(headerAcceptLang != null)
                        { StatsLogger.captureDataPoint(statsIDLang, l.toString()); }
                    }

                return;
                }

            // Give up and use a safe value.
            userLocale = SAFE_LOCALE;
            }
        finally
            {
            // Note that we've finished the work for this request.
            cachedRequest = request;
            }
        }


    /**Public no-arg constructor for ease of use as a JavaBean.
     * Sets to a safe locale.
     */
    public LocaleBean()
        { }

    /**Public no-arg constructor for ease of use as a JavaBean.
     * This defers as much work as it reasonably can.
     *
     * @param l  initial locale; never null
     */
    public LocaleBean(final Locale l)
        {
        super(l);
        }

    /**Unique Serialisation class ID generated by http://random.hd.org/. */
    private static final long serialVersionUID = -6180248508840300659L;
    }
