/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import javax.servlet.jsp.PageContext;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 02-Nov-2003
 * Time: 20:52:36
 */

/**A utilities class to support page skins.
 * Many skin-specific overrides from the default values are accessed
 * from the page-level (PageContext) properties via constants
 * and methods here.
 * <p>
 * Default skin/UI methods are also defined here.
 */
public final class PageSkinUtils
    {
    /**Prevent creation of an instance. */
    private PageSkinUtils() { }

    /**Main background colour for catalogue and other generic pages. */
    public static final String GENERIC_BG_COLOUR = "white";
    /**Quoted (if necessary) version of GENERIC_BG_COLOUR for HTML attributes; computed once per run. */
    public static final String GENERIC_BG_COLOUR_QUOTED = GENERIC_BG_COLOUR;
    /**Main text colour for architrave headings for catalogue and other generic pages. */
    public static final String GENERIC_AH_COLOUR = "navy";
    /**Quoted (if necessary) version of GENERIC_AH_COLOUR for HTML attributes; computed once per run. */
    public static final String GENERIC_AH_COLOUR_QUOTED = GENERIC_AH_COLOUR;
    /**Main background colour for architrave headings for catalogue and other generic pages. */
    public static final String GENERIC_ABG_COLOUR = "#ffffcc"; // "yellow";
    /**Quoted (if necessary) version of GENERIC_ABG_COLOUR for HTML attributes; computed once per run. */
    public static final String GENERIC_ABG_COLOUR_QUOTED = "\"" + GENERIC_ABG_COLOUR + "\"";
    /**Columns in table for "generic"/default skin. */
    public static final int GENERIC_TABLE_COLS = 4;
    /**Width of chisled border for thumbnails in generic layout. */
    public static final int GENERIC_TN_BORDER_WIDTH_PIXELS = 5; // Decorative border width for thumbnail tables.
    /**Nominal generic page width (pixels).
     * Up to 2003/02/19 was 600 as per http://news.bbc.co.uk/
     * but since then, we too have increased to 800.
     * <p>
     * I presume that the BBC has discovered that people are using
     * larger (more pixels) screens than before.
     */
    public static final int GENERIC_WIDTH_PX = 800;


    /**Attribute in PageContext used to hold non-default alternate background colour.
     * Essentially private to getAltBgColour() and setAltBgColour().
     */
    private static final String _gABC_PROP_NAME = "_gABC";

    /**Sets the alternate background colour for the current skin.
     * Must be a value unquoted HTML colour name or value,
     * eg white or #fffff.
     * <p>
     * This can be null to restore the default value,
     * ie remove any override.
     */
    public static void setAltBgColour(final PageContext pc, final String colour)
        {
        if(colour == null)
            {
            pc.removeAttribute(_gABC_PROP_NAME);
            return;
            }
        pc.setAttribute(_gABC_PROP_NAME, colour);
        }

    /**Gets the alternate background colour for the current skin; never null.
     * This gets an unquoted HTML colour name or value,
     * eg white or #fffff.
     * <p>
     * This may need to be quoted to be used as an attribute value.
     * <p>
     * This defaults to the normal value,
     * but can be overridden for the current JSP page (in the PageContext)
     * with a call to setAltBgColour().
     */
    public static String getAltBgColour(final PageContext pc)
        {
        final Object o = pc.getAttribute(_gABC_PROP_NAME);
        if(o instanceof String) { return((String) o); }
        return(GENERIC_ABG_COLOUR);
        }
    }
