/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

import javax.servlet.*;
import java.io.IOException;

/**Filter to dynamically load and wrap a pluggable filter.
 * This enables filters to be loaded entirely at runtime.
 * <p>
 * The param pluginClassName is used to give the class name of a
 * Filter class to load and pass filter calls on to.
 * If this class cannot be found or loaded, or is not specified,
 * this PluginFilter performs no transformations at all.
 */
public final class PluginFilter implements Filter
    {
    /**The filter configuration object we are associated with.
     * If this value is null, this filter instance is not currently configured.
     */
    private FilterConfig config;

//    public FilterConfig getFilterConfig() { return(config); }
//    public void setFilterConfig(FilterConfig c)
//        {
//        config = c;
//        if(plugin != null) { plugin.setFilterConfig(c); }
//        }

    /**Config parameter; name of flag to record overload, or null if none chosen. */
    private String pluginClassName;

    /**Our plug-in class; or null if none loaded. */
    private Filter plugin;


    /**Place this filter into service.
     * Defer starting the timer until the first activity that we see...
     *
     * @param filterConfig  the filter configuration object
     */
    public void init(FilterConfig filterConfig)
        {
        config = filterConfig;

        pluginClassName = null;
        if(plugin != null) { plugin.destroy(); plugin = null; }

        if(filterConfig != null)
            {
            pluginClassName = filterConfig.getInitParameter("pluginClassName");
            if(pluginClassName != null)
                {
                try {
                    plugin = (Filter) (Class.forName(pluginClassName).newInstance());
                    // Successful!  Initialise the plugin.
                    plugin.init(filterConfig);
                    }
                catch(Exception e)
                    {
                    plugin = null;
                    if(!(e instanceof ClassNotFoundException)) { e.printStackTrace(); }
                    else { System.err.println("could not find plugin class " + pluginClassName); }
                    System.err.println("could not init plugin for class " + pluginClassName + ": " + e.getClass() + ": " + e.getMessage());
                    config.getServletContext().log("could not init plugin for class " + pluginClassName, e);
                    }
                }
            }
        }

    /**Take this filter out of service.
     */
    public void destroy()
        {
        if(plugin != null)
            {
            plugin.destroy();
            plugin = null;
            }

        // Help GC.
        config = null;
        }

    /**Called to handle each request.
     */
    public void doFilter(ServletRequest request,
                         ServletResponse response,
                         FilterChain chain)
        throws IOException, ServletException
        {
        if(plugin == null)
            {
            // Pass-though because no plugin.
            chain.doFilter(request, response);
            return;
            }

        // Hand the request to the plugin to process!
        plugin.doFilter(request, response, chain);
        }
    }

