/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import ORG.hd.d.jIndexer.server.JIndexBean;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.*;

/**Bean for cached search results.
 * Uses SearchUtils.doRelatedExhibitsSearch() to compute results lazily
 * and caches the results.
 * <p>
 * Excludes the current exhibit and variants
 * (assuming that they will be displayed elsewhere).
 */
public final class SearchResultSimpleCache
    {
    /**Map from full exhibit name (and canonicalised search term) to immutable ordered list of short-name results, best match first.
     * Thread-safe, with efficient and concurrent get() access.
     * <p>
     * A multi-CPU host may have some genuine write concurrency
     * immediately after start-up or an AEP change,
     * so we allow for some write concurrency here, but not much.
     */
    private final ConcurrentMap<String,List<String>> searchResults = new ConcurrentHashMap<String, List<String>>(4013, 0.7f, 4);
    
    /**Absolute maximum number of results to store/cache per exhibit.
     * Limits storage requirements,
     * as large as we should ever want to show on one cat page.
     * <p>
     * If the cat page logic actually uses this value as the maximum result size
     * then we should get optimal CPU and memory use, and minimal heap churn.
     */
    public static final int ABS_MAX_RESULTS = WebConsts.MAX_RESULTS_PER_PAGE/2;

    /**Do cached search lookup on given full exhibit name; never null.
     * If fewer items are requested than computed/cached,
     * then the best (first) cached items are returned.
     * <p>
     * The (immutable) results are limited to at most ABS_MAX_RESULTS items.
     * <p>
     * Excludes exhibit and variants of same exhibit
     * (same main words stem, etc, giving same canonicalised search term)
     * on the assumption that they may already have been shown.
     * <p>
     * This allows potentially-redundant requests to execute in parallel
     * at the risk of a little wasted work
     * so as to maximise potential concurrency.
     * <p>
     * We do NOT cache any results if we don't believe the index
     * to be up-to-date (eg to avoid cacheing stale/empty values).
     *
     * @return  private String[] of result short exhibit names, best first;
     *     possibly zero-length but never null
     */
    public List<String> getRelatedExhibits(final DataSourceBean dataSource,
                                           final String exhibitFullName,
                                           final int maxResults)
        throws IOException
        {
        if((dataSource == null) ||
           !ExhibitName.validNameSyntax(exhibitFullName) ||
           (maxResults < 0))
            { throw new IllegalArgumentException(); }

        // Check if we have anything cached under the full exhibit name...
        List<String> r = searchResults.get(exhibitFullName);
        if(r == null)
            {
            // Nothing cached under the full exhibit name,
            // so compute our search term which is the canonical cache key.

            // By default cache the results that we compute.
            boolean cacheResults = true;

            // The search term is all the main and attribute words,
            // canonicalised in the standard way for all searches.
            final String searchTerm = computeSearchTerm(exhibitFullName);

            // See if cached under the canonical search term.
            r = searchResults.get(searchTerm);
            // No, not cached under the canonical search term either,
            // so we'll really have to use the main index!
            if(r == null)
                {
                final Map<String, String> shortToFullMap = dataSource.getAllExhibitProperties(-1).aeid.getShortToFullMap();

                final SortedSet<String> attrWords = ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet();

                // Set up the exhibit filter for the index bean.
                // We arrange to reject the original exhibit,
                // and all close variants of it
                // (those that produce the same canonical search term).
                // We reject multiple close variants of any other exhibit
                // so as to give maximum variety in the results.
                final JIndexBean.SearchFilterByName filter =
                    new JIndexBean.SearchFilterByName(){
                        /**Thread-safe set of exhibit name stems already accepted; never null. */
                        private final Set<String> alreadyAccepted =
                            Collections.synchronizedSet(new HashSet<String>(16+128*ABS_MAX_RESULTS));

                        /**Accept or reject based on (short) name; no dir component. */
                        public final boolean accept(final String shortName)
                            {
                            // Convert short name to
                            // syntactically-valid full name.
                            // Doing a lookup may be cheaper than synthesising
                            // a fake full name.
                            final String fn = shortToFullMap.get(shortName);
                            if(fn == null) { return(false); }
                            final String st = computeSearchTerm(fn);

                            // Reject if the canonicalised search term is the same.
                            if(st.equals(searchTerm)) { return(false); }

                            // Our stem excludes attribute words here.
                            final String stem = ExhibitName.getMainWordsComponent(fn, attrWords);
                            // If we already accepted same stem earlier, reject this.
                            if(!alreadyAccepted.add(stem)) { return(false); }

                            // OK, accept this!
                            return(true);
                            }
                        };

                // If we find the index is not up-to-date right now
                // then don't cache the results after computing them.
                if(!dataSource.byWordIndexIsAvailableAndUpToDate())
                    { cacheResults = false; }

                // Now really do the lookup/search.
                // Don't allow fallback to wide searches
                // that may litter a page with confusing spurious links.
                // If nothing else, such spurious links and text
                // may confuse classification engines
                // such as that in Google's AdSense "Mediabot".
                final String[] raw = SearchUtils.doRelatedExhibitsSearch(searchTerm,
                                            dataSource,
                                            ABS_MAX_RESULTS,
                                            filter,
                                            false,
                                            true);

                // Since we may hold onto this data for a while,
                // we eliminate globally-duplicate values with intern().
                if(raw.length > 0)
                    {
                    for(int i = raw.length; --i >= 0; )
                        { raw[i] = MemoryTools.intern(raw[i]); }
                    // Wrap in immutable List...
                    r = Collections.unmodifiableList(Arrays.asList(raw));
                    }
                // Optimise the no-answers case.
                else { r = Collections.emptyList(); }

                // Cache under the canonical term.
                if(cacheResults) { searchResults.putIfAbsent(searchTerm, r); }

//if(IsDebug.isDebug) { System.out.println("SearchResultSimpleCache: found/cached "+(raw.length)+" result(s) for exhibit "+exhibitName+" with canonical search term: " + searchTerm); }
                }

            // Cache the results (including empty results) under the full name.
            if(cacheResults) { searchResults.putIfAbsent(exhibitFullName, r); }
            }

        // If the request was for less results than we happen to have cached,
        // return a sub-list view (still immutable)...
        if(maxResults < r.size())
            { return(r.subList(0, maxResults)); }

        // Return the entire computed/cached result.
        return(r);
        }

    /**Map from full exhibit name to canonicalised search term.
     * Thread-safe, with efficient and concurrent get() access.
     * <p>
     * We keep this to avoid doing all O(m.(n^2)) potential canonicalisations
     * for n exhibits with m changes of AEP.
     * <p>
     * Note that this need not be linked to the AEP
     * since the search term does not depend on anything but the name itself,
     * though keeping it AEP-linked makes for easier/faster storage management.
     */
    private final ConcurrentMap<String,String> fullNameToSearchTerm = new ConcurrentHashMap<String,String>(16384);

    /**Compute canonicalised search term for a full exhibit name.
     * This is all the main and attribute words, canonicalised
     * (mono-cased and with duplicates removed, and potentially length-limited).
     */
    private /* static */ String computeSearchTerm(final String exhibitFullName)
        {
        String result = fullNameToSearchTerm.get(exhibitFullName);
        if(null == result)
            {
            final Set<String> noAttrs = Collections.emptySet();
            result = DataSourceBean.canonicaliseSimpleByWordQuery(
                            ExhibitName.getMainWordsComponent(exhibitFullName, noAttrs),
                            ExhibitName.MAX_STEM_LENGTH);
            result = MemoryTools.intern(result); // Will probably be many the same.
            // Cache (first) result.
            fullNameToSearchTerm.putIfAbsent(exhibitFullName, result);
            }
        return(result);
        }

    /**Private lookup key for full searches of similar items; never null. */
    private static final DataSourceBean.AEPLinkedKey fullSearchCacheKey = new DataSourceBean.AEPLinkedKey("fullSearchCacheKey");

    /**Do cached lookup for cat page "similar items"; never null.
     * The (immutable) result is keyed to the exhibit name and dsb.
     * <p>
     * All results assosciated with a dsb are discarded when its AEP changes.
     */
    public static List<String> doCachedCatPageSimilarItems(final DataSourceBean dsb,
                                                           final String exhibitName,
                                                           final int maxToShow)
        {
        // Get existing cache, or create unique new one (eg after AEP change).
        SearchResultSimpleCache cachedResults;
        while((cachedResults = (SearchResultSimpleCache) dsb.getAEPLinkedValue(fullSearchCacheKey)) == null)
            { dsb.putIfAbsentAEPLinkedValue(fullSearchCacheKey, new SearchResultSimpleCache()); }

        try
            {
            return(cachedResults.getRelatedExhibits(dsb,
                                                    exhibitName,
                                                    maxToShow));
            }
        catch(final Exception e)
            {
            e.printStackTrace();  // Whinge but absorb any error.
            final List<String> emptyResult = Collections.emptyList();
            return(emptyResult);
            }
        }
    
    /**Do cached search lookup as for doCachedCatPageSimilarItems() but asynchronously; never null.
     */
    public static Future<List<String>> doCachedCatPageSimilarItemsFuture(
                                           final DataSourceBean dsb,
                                           final String exhibitFullName,
                                           final int maxResults)
        {
        // Get existing cache, is any.
        final SearchResultSimpleCache cachedResults = (SearchResultSimpleCache) dsb.getAEPLinkedValue(fullSearchCacheKey);
        if(null != cachedResults)
            {
            // Check if we have anything cached under the full exhibit name...
            final List<String> r = cachedResults.searchResults.get(exhibitFullName);
            if(null != r)
                {
                // Have cached result, so return already-finished value...
                return(new Future<List<String>>(){
                    public boolean cancel(boolean mayInterruptIfRunning)
                        { return(false); }
                    public List<String> get() // throws InterruptedException, ExecutionException
                        { return(r); }
                    public List<String> get(long timeout, TimeUnit unit) // throws InterruptedException, ExecutionException, TimeoutException
                        { return(r); }
                    public boolean isCancelled()
                        { return(false); }
                    public boolean isDone()
                        { return(true); }
                    });
                }
            }

        // If we did not find a cached value,
        // set up the task to compute the value asynchronously.        
        final Callable<List<String>> callable = new Callable<List<String>>(){
            public final List<String> call() throws Exception
                { return(doCachedCatPageSimilarItems(dsb, exhibitFullName, maxResults)); }
            };
//        final FutureTask<List<String>> result = new FutureTask<List<String>>(callable);
//        final Thread t = new Thread(result, "cat page serach results async computation thread");
//        t.setDaemon(true);
//        t.start();
        final Future<List<String>> result = ThreadUtils.computeIntensiveThreadPool.submit(callable);

        // Return the handle for the caller to retrieve the value...
        return(result);        
        }
    }
