/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import ORG.hd.d.jIndexer.server.JIndexBean;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;

import java.io.IOException;
import java.util.regex.Pattern;

/**Search related utility functions.
 * This is for algorithms only of interest to Web apps, often JSPs.
 * <p>
 * One advantage of having a method here rather than in-line in a JSP
 * is that is is pre-compiled off-line for speed and robustness.
 */
public final class SearchUtils
    {
    /**Prevent construction of an instance. */
    private SearchUtils() { }

    /**Do look-up of "similar" items, eg in the foot of each catalogue page; never null.
     * The search term is assumed to be some leading part of the
     * (short) exhibit name, ie excluding the directory components.
     * <p>
     * This may try a narrow search,
     * insiting that the first search word must appear in candidate result
     * exhibit names and preferrably most search terms should be matched,
     * optionally falling back to a wider search if the narrow one fails.
     * <p>
     * This does not cache its results.
     *
     * @param searchTerm non-null, non-empty search term(s);
     *     earlier words are possibly more significant
     * @param dataSource  data source; never null
     * @param maxResults maximum number of results; never null
     * @param filter  non-null filter of acceptable exhibits
     * @param fallback if true and a narrow search fails to generate answers,
     *     fall back to a wider search
     * @param alreadyCanon if true then the searchTerm is already canonicalised
     * @throws IOException in case of error
     *
     * @return in-order list of matches as short exhibit names, best first,
     *     up to requested maximum number of results;
     *     may be zero-length but never null
     */
    public static String[] doRelatedExhibitsSearch(final String searchTerm,
                                                   final DataSourceBean dataSource,
                                                   final int maxResults,
                                                   final JIndexBean.SearchFilterByName filter,
                                                   final boolean fallback,
                                                   final boolean alreadyCanon)
        throws IOException
        {
        if((searchTerm == null) || (searchTerm.length() == 0) ||
           (dataSource == null) ||
           (maxResults < 1) ||
           (filter == null))
            { throw new IllegalArgumentException(); }

        // Ensure that we have the search terms canonicalised in a simple way.
        // This means order preserved,
        // but all converted to lower-case
        // and all dashes converted to spaces.
        final String canonTerms = alreadyCanon ? searchTerm :
            DataSourceBean.canonicaliseSimpleByWordQuery(searchTerm, ExhibitName.MAX_STEM_LENGTH);

        // Compute the first word from the search terms.
        final int firstSpace = canonTerms.indexOf(' ');
        final String firstWord = (firstSpace < 1) ?
            canonTerms : canonTerms.substring(0, firstSpace);
        
        // Compile a regex for a case-insensitive whole-word match.
        // By insisting on the explicit trailing separator after the word
        // we implicitly exclude the author initials and extension from any match.
        final Pattern p = Pattern.compile("\\b" + firstWord + ExhibitName.WORD_SEP, Pattern.CASE_INSENSITIVE);
        
        // Try a narrow-ish search,
        // asking for most of the terms to be matched if possible,
        // and insisting that the main (first) search word is in the result name.
        // (We'll accept a crude case-insensitive substring match first,
        // then an exact case-insensitive match.)
        final String resultNarrow[] = dataSource.findExhibitsByWord(canonTerms,
                                      DataSourceBean.FEBY_MATCH_TYPE_MOST,
                                      maxResults,
            new JIndexBean.SearchFilterByName(){
                /**Accept or reject based on (short) name; no dir component. */
                public final boolean accept(final String s)
                    {
                    // Reject unless we can find entire first/main word
                    // as an entire word within the specified name.
                    // For speed we will allow a match against any part
                    // except the extension and author initials
                    // (ie including any attribute words).
                    if(!p.matcher(s).find())
                        { return(false); }

                    // Reject if supplied filter rejects.
                    return(filter.accept(s));
                    }
                });
        if(resultNarrow.length > 0)
            { return(resultNarrow); }

        // Possibly allow fall back to much wider seraches.
        if(fallback)
            {
            // Try doing a match insisting on most terms being present.
            final String resultWide[] = dataSource.findExhibitsByWord(canonTerms,
                                          DataSourceBean.FEBY_MATCH_TYPE_MOST,
                                          maxResults,
                                          filter);
            if(resultWide.length > 0)
                { return(resultWide); }

            // Try doing a match on any terms.
            final String resultWider[] = dataSource.findExhibitsByWord(canonTerms,
                                          DataSourceBean.FEBY_MATCH_TYPE_ANY,
                                          maxResults,
                                          filter);
            if(resultWider.length > 0)
                { return(resultWider); }
            }

        return(NO_ANSWER);
        }

    /**No answer to search. */
    private static final String[] NO_ANSWER = new String[0];
    }
