/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;

import ORG.hd.d.IsDebug;

import javax.servlet.http.*;
import java.io.Serializable;
import java.net.*;
import java.util.*;

/**JavaBean to hold explicit session variables such as user-selected-locale.
 * Error tolerant on being set (should be able to be done by a filter using reflection)
 * and can spit out a set of non-default values to adjust or pass to another server instance.
 * <p>
 * Names of all (visible) properties start "sessionVar" to avoid collision with other
 * HTML (GET/POST) parameter users.
 * <p>
 * Not thread-safe; access to an instance in a session should be synchronised on the session object.
 */
public final class SessionVarBean implements Serializable, Cloneable
    {
    /**HTTP tag/parameter we use to indicate that session variable should be set to defaults.
     * The parameter with this name can be given a random value as a "cache buster".
     * <p>
     * When this is set, session variables should be set to all defaults and then
     * set to whatever other actual values are present,
     * meaning that only changes from defaults need be sent.
     * <p>
     * This parameter is set to a fixed value for spiders to avoid presenting them
     * with a spurious vast URI space to index.
     */
    public static final String KEY_sessionVar_CLEAR = "sessionVar";

    /**Immutable random token (shared with any clones/copies to save time) for cache-busting in URLs. */
    private final int rndToken = Rnd.fastRnd.nextInt();

    /**Lite UI; if true we should run a "lite" UI for users on slow modems, etc, else full UI.
     * Full UI (flag off) is the default.
     */
    private boolean sessionVarLiteUI;
    public static final String KEY_sessionVarLiteUI = "sessionVarLiteUI";
    public boolean isSessionVarLiteUI() { return(sessionVarLiteUI); }
    public SessionVarBean setSessionVarLiteUI(final boolean sessionVarLiteUI)
        { this.sessionVarLiteUI = sessionVarLiteUI; return(this); }

    /**Session locale: if non-null this overrides any locale information from elsewhere.
     * No override (null) is the default.
     * <p>
     * Note that this field can only be set to Locales with a valid language code
     * (ie two-letter lowercase)
     * and an optional valid country code (ie two-letter uppercase) and no variant
     * and that are found in our full supported-locales list.
     * Note thae we do this validation since the Locale object does not (SHAME)!
     * <p>
     * We <em>ignore</em> attempts to set to an unsupported/unsafe/broken locale.
     * <p>
     * We rely on immutability of the Locale object here.
     */
    private Locale sessionVarLocale;
    public static final String KEY_sessionVarLocale = "sessionVarLocale";
    public Locale getSessionVarLocale() { return(sessionVarLocale); }
    public SessionVarBean setSessionVarLocale(final Locale sessionVarLocale)
        {
        if((sessionVarLocale != null) && !I18NTools.LOCALES.contains(sessionVarLocale))
            { return(this); /* Refuse/ignore the invalid "set" request. */ }
        this.sessionVarLocale = sessionVarLocale;
        return(this);
        }


    /**Make an independent clone of this object. */
    public Object clone()
        throws CloneNotSupportedException
        {
        // No shared-mutable state, so basic Object.clone() does the trick.
        return(super.clone());
        }

    /**Make an independent clone of this object, returning the correct type; never null.
     * Typically used to get a modified version of the current state
     * to form a URL to set the new value.
     *
     * @return  a non-null independent clone/copy of this instance
     */
    public SessionVarBean copy()
        {
        // No shared-mutable state, so basic Object.clone() does the trick.
        try { return((SessionVarBean) clone()); }
        // Should never throw an exception.
        catch(final CloneNotSupportedException e) { throw new Error(e); }
        }


    /**Update session vars from request, saving result back to session if need be.
     * This should be called early enough (before output is committed)
     * that a session can be created if need be.
     * <p>
     * This should probably only be routinely applied to GET operations
     * where it cannot accidentally gobble data such as in a file-upload POST.
     * <p>
     * If a "clear" tag is present then we ignore extant values and start from defaults.
     */
    public static void updateSessionVarsFromRequest(final HttpServletRequest request)
        {
        final String clParam = request.getParameter(KEY_sessionVar_CLEAR);
        final boolean clearVars = (clParam != null) && !"".equals(clParam);
        final SessionVarBean vars = clearVars ?
                new SessionVarBean() /* All defaults. */ :
                getExtantSessionVars(request, true) /* Adjust extant values. */ ;

        // Set the individual values.
        final String liteUI = request.getParameter(KEY_sessionVarLiteUI);
        if((liteUI != null) && !"".equals(liteUI)) { vars.setSessionVarLiteUI(Boolean.parseBoolean(liteUI)); }
        final String locale = request.getParameter(KEY_sessionVarLocale);
        // We will accept any explicitly-supported locales of form ll or ll_CC.
        // We will NOT accept random/broken locale values for security/robustness.
        if((locale != null) && !"".equals(locale) &&
                        ((locale.length() == 2) || (locale.length() == 5)))
            {
            final Locale ul = new Locale(locale.substring(0, 2),
                                         (locale.length() < 5) ? "" : locale.substring(3, 5));
            if(I18NTools.LOCALES.contains(ul))
                { vars.setSessionVarLocale(ul); }
            }

        // If any values are non-default and we don't have a session yet, then save.
        // If we do have a session then save regardless (or if we can detect a change).
        // This avoids us creating a session just to store default values
        // which is unnecessary and may be intrusive or irritate the user.
        if((request.getSession(false) != null) || !vars.getNonDefaultValues().isEmpty())
            {
            // Create session if need be...
            final HttpSession session = request.getSession(true); // Force session creation.
            synchronized(session)
                { session.setAttribute(ATTR_NAME_SVB, vars); }
            }
        }

    /**Generate new URL to set the desired state; never null.
     * This generates a URL based on the current one
     * with GET parameters added to set session values to those in this instance.
     * <p>
     * For brevity/efficiency of result,
     * this returns the shortest unambiguous relative or full URL
     * that can be used as an href in the current page to get to the desired page
     * with the desired session values set.
     * <p>
     * This may drop any query component of the original URL.
     * <p>
     * This may drop or preserve any fragment/ref component of the original URL.
     * <p>
     * This involves resetting all implicit session values to default and then
     * explicitly setting all non-default values with query parameters.
     * <p>
     * This also adds a "cache-busting" random component to help ensure that
     * the browser or a proxy does not just replay an old value from cache.
     * This is given a fixed value if the request appears to be from a spider
     * to avoid creating a huge virtual page space.
     * <p>
     * This can optionally force a diversion to an optimal mirror for the current site,
     * helping to ensure that the user's browser is correctly "pinned" to a server
     * and thus will not "lose" the state
     * (and will probably give better performance for the user and be cheaper to serve too).
     *
     * @return relative or full URL based on the current request; never null
     */
    public String makeSessionVarSetURL(final HttpServletRequest request,
                                       final boolean attemptToPinToMirror,
                                       final DataSourceBean vars)
        {
        if((request == null) || (vars == null))
            { throw new IllegalArgumentException(); }

        final URL origURL;
        try { origURL = new URL(request.getRequestURL().toString()); }
        catch(final MalformedURLException e) { throw new IllegalArgumentException(e); }

        final int initialCapacity = 64; // Resulting URLs can be BIG!
        final StringBuilder sb = new StringBuilder(initialCapacity);

        // Insert mirror host name if requested to pin to (any) mirror
        // and on a normalised (non-master, non-mirror) URL
        // and on a standard port (etc) so that this appears to be a "normal" access.
        // Don't do this if the "generic" host is selected
        // or the user is apparently already on the right mirror.
        // All this reduces the size of the HTML generated as much as possible.
        if(attemptToPinToMirror &&
                !HostUtils.isMirrorName(origURL.getHost()) &&
                !HostUtils.isMasterName(origURL.getHost()) &&
                (origURL.getPort() == -1))
            {
            final String selectedMirror = WebUtils.chooseMirrorHostToBalanceLoad(request, vars);
            if(!CoreConsts.MAIN_DATA_HOST.equals(selectedMirror) &&
               !origURL.getHost().equals(selectedMirror))
                {
                sb.append(origURL.getProtocol()).append("://");
                sb.append(selectedMirror);
                // Only put in full path when creating an absolute/full URL.
                sb.append(origURL.getFile());
                }
            }

        // Now append the query string,
        // starting by resetting all var values to defaults and "cache busting",
        // and then setting all non-default values.
        sb.append('?').append(KEY_sessionVar_CLEAR).append('=');
        if(WebUtils.requestProbablyFromSpider(request))
            { sb.append("spider"); /* Fixed token to avid trapping spider in huge page space. */ }
        else
            { sb.append(Integer.toString(rndToken >>> 1, Character.MAX_RADIX)); /* Random (non-negative high-radix integer) token to "cache bust". */ }

        final Map<String,String> vals = getNonDefaultValues();
        for(final String key : vals.keySet())
            {
            sb.append('&').append(key).append('=');
            // Assume value is non-null and does not requre URL encoding for now.
            final String value = vals.get(key);
            sb.append(value);
            }

        // Append any fragment component is any.
        final String ref = origURL.getRef();
        if((ref != null) && (ref.length() > 0)) { sb.append('#').append(ref); }

if(IsDebug.isDebug && (sb.length() > initialCapacity)) { System.err.println("[WARNING: makeSessionVarSetURL result needed copy/resize: "+(sb.length())+" chars]"); }

        return(sb.toString());
        }


    /**Get (copy of) any existing value from the HTTP session; null if no session values and create is false.
     * This is suitable to then override with user-specified values as required.
     * <p>
     * Note that this returns a private copy of any value held in the session.
     *
     * @param request  HTTP request object; never null
     * @param create  if true and session or session variable is missing
     *                then createand return a new all-default value,
     *     else if false and session or session variable is missing then return null
     */
    public static SessionVarBean getExtantSessionVars(final HttpServletRequest request,
                                                      final boolean create)
        {
        if(request == null)
            { throw new IllegalArgumentException(); }

        final HttpSession s = request.getSession(false);

        // If we have a session AND the session attr exists AND is the right type
        // then return a copy of it.
        if(s != null)
            {
            final Object attr;
            synchronized(s) { attr = s.getAttribute(ATTR_NAME_SVB); }
            if(attr instanceof SessionVarBean)
                {
                final SessionVarBean svb = (SessionVarBean) attr;
                if(svb != null)
                    { return(svb.copy()); }
                }
            }

        // Could not find attr and not creating
        // so return null.
        if(!create) { return(null); }

        // Return new all-defaults instance.
        return(new SessionVarBean());
        }

    /**Get sorted map of attribute names to values for all non-default values; never null.
     * Values are given as an HTML/HTTP-safe String.
     * <p>
     * If the result is not empty then at least one value is non-default
     * and we should probably save this in the session (creating a session if need be).
     * <p>
     * The map is sorted to so that it is easier to dump keys into HTML in sorted order
     * and thus help compression if we are doing any.
     *
     * @return non-null sorted map from attribute name to String representation of value
     */
    public SortedMap<String,String> getNonDefaultValues()
        {
        final SortedMap<String,String> result = new TreeMap<String, String>();
        if(sessionVarLiteUI) { result.put(KEY_sessionVarLiteUI, "true"); }
        if(sessionVarLocale != null) { result.put(KEY_sessionVarLocale, String.valueOf(sessionVarLocale)); }
        return(result);
        }

    /**Name of SessionVarBean instance in HttpSession. */
    public static final String ATTR_NAME_SVB = "org.hd.d.pg2k.SessionVarBean";

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -200996533296387845L;
    }
