/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.AddrTools;
import org.hd.d.pg2k.svrCore.Rnd;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.io.InvalidObjectException;
import java.io.ObjectInputValidation;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**Web-server dynamic (authenticated) stats-collection support.
 * This enables a number of types of stats to be collected
 * in such a way as to be unlikely to be accidentally triggered by a spider
 * or off a cached Web page/URL somewhere.
 * <p>
 * This also facilitates sampling stats so as to not pester any one user too much
 * or gather more than one data point at a time from them.
 * <p>
 * Most operations on this class are contant time on average,
 * although mutators in particular may perform maintenance work
 * that does not raise the O() cost when amortised across all calls.
 * <p>
 * Access to all significant data structures is under the class lock.
 */
public final class StatsSink
    {
    /**Prevent construction of an instance. */
    private StatsSink() { }

    /**The name of the HTTP parameter carrying our manditory unique event ID value. */
    public static final String HTTP_ID_PARAM_NAME = "ID";

    /**Map from uniqueListenerID to Listener; never null.
     * Incoming stats events are looked up in here by ID.
     */
    private static final Map<String, AbstractStatsListener> byListenerID = new HashMap<String, AbstractStatsListener>();

    /**Map from uniqueDataPointID to Listener; never null.
     * The normal programmatic interface makes use of this.
     */
    private static final Map<String, AbstractStatsListener> byDataPointID = new HashMap<String, AbstractStatsListener>();

    /**Find (and remove) listener identified by listener ID, or null if none live by that ID.
     * Run under the class lock.
     * <p>
     * Has removed the listener from the database, if present beforehand,
     * by the time it returns.
     */
    private static synchronized AbstractStatsListener _getAndRemoveListenerByListenerID(final String listenerID)
        {
        // Get the extant listener, if any, by that listener ID.
        final AbstractStatsListener asl = byListenerID.get(listenerID);
        if(asl == null) { return(null); }

        // Zap the (one-shot) listener, whether expired or not...
        byListenerID.remove(listenerID);
        byDataPointID.remove(asl.uniqueDataPointID);

        // If the listener has expired then pretend that it does not even exist.
        if(asl.isExpired()) { return(null); }

        return(asl); // OK, listener is live and well so return it.
        }

    /**Accept an inbound HTTP stats event.
     * We will filter it and decide whether to pass it on to an agent,
     * and if so, which agent.
     * <p>
     * No lock is held while the agent/handler is called.
     * <p>
     * We must see an "ID" parameter with the unique allocated ID,
     * or this is rejected out of hand (ignored).
     * <p>
     * This may return a non-null value that is treated as a URL to redirect to.
     * The returned URL may be root-relative.
     *
     * @param request  valid non-null HTTP request object; never null
     *
     * @return URL, possibly root-relative, to redirect to, or null if none
     */
    public static String acceptEvent(final HttpServletRequest request,
                                     final ServletContext context)
        {
        // Spend on average O(1) time clearing expired entries.
        _sometimesClearExpiredEntries();

        if(request == null) { return(null); /* Invalid request, so ignore. */ }

        String redirectURL = null; // Do not redirect at all by default.

        // Get the ID, if any.
        final String ID = request.getParameter(HTTP_ID_PARAM_NAME);
        if(ID == null)
            {
if(IsDebug.isDebug) { context.log("ERROR: StatsSink: invalid request missing ID from: " + request.getRemoteAddr() + '(' + AddrTools.doReverseLookup(request.getRemoteAddr(), true)+ ')'); }
            return(redirectURL); /* Invalid request, so ignore. */
            }

        // If a listener is extant then call it with the HTTP parameters.
        final AbstractStatsListener asl = _getAndRemoveListenerByListenerID(ID);
        if(asl != null)
            {
            final HashMap<String,String[]> params = new HashMap<String, String[]>(request.getParameterMap());
            // Log interesting parameters such as click-through ad URL.
            if((params != null) && !params.isEmpty()) { context.log("INFO: StatsSink: request URL with parameters: " + request.getQueryString()); }
            redirectURL = asl.handle(params);
            }
        else
            {
if(IsDebug.isDebug) { context.log("WARNING: StatsSink: invalid/expired request ID from: " + request.getRemoteAddr() + '(' + AddrTools.doReverseLookup(request.getRemoteAddr(), true)+ ')'); }
            }

        return(redirectURL);
        }

    /**Returns true if there is a live (non-expired) listener for a given data-point ID.
     * Useful for helping ensure that we don't try to collect more than one
     * value for a data point at once.
     * <p>
     * This is fast and read-only.
     *
     * @param uniqueDataPointID  the unique ID; must not be null
     */
    public static synchronized boolean isListenerLiveForDataPoint(final String uniqueDataPointID)
        {
        // May in fact not cause a problem if null,
        // but does indicate a programming error by the caller.
        assert(uniqueDataPointID != null) : "uniqueDataPointID must not be null";

        final AbstractStatsListener asl = byDataPointID.get(uniqueDataPointID);
        return((asl != null) && !asl.isExpired());
        }

    /**Add new listener for named unique data point.
     * Not added if already expired, though an extant listener will be removed.
     * <p>
     * Any extant listener for the same data point is removed.
     * <p>
     * This may do clean-up work on some calls.
     * <p>
     * This call has an average constant (ie O(1)) call time.
     */
    public static synchronized void addListenerForDataPoint(final AbstractStatsListener newAsl)
        {
        if(newAsl == null)
            { throw new IllegalArgumentException(); }

        // Spend on average O(1) time clearing expired entries.
        _sometimesClearExpiredEntries();

        // See if there is an extant listener for this uniqueDataPointID.
        // If so then zap it from both maps first.
        final AbstractStatsListener oldAsl = byDataPointID.get(newAsl.uniqueDataPointID);
        if(oldAsl != null)
            {
            byDataPointID.remove(oldAsl.uniqueDataPointID);
            byListenerID.remove(oldAsl.uniqueListenerID);
            }

        // Don't add an already-expired listener...
        if(newAsl.isExpired())
            { return; }

        // Now add the new listener to both maps.
        byDataPointID.put(newAsl.uniqueDataPointID, newAsl);
        byListenerID.put(newAsl.uniqueListenerID, newAsl);
        }

    /**Spend on average O(1) (ie constant) time expiring old entries. */
    private static synchronized void _sometimesClearExpiredEntries()
        {
        // See if this caller pulls the short straw to expire old entries.
        // Amortise cost to be approximately O(1), ie constant cost per call.
        if(0 == Rnd.fastRnd.nextInt(1 + byDataPointID.size() + byListenerID.size()))
            {
            // Remove dead entries from both maps.
            for(final Iterator<String> it = byDataPointID.keySet().iterator(); it.hasNext();)
                {
                if(byDataPointID.get(it.next()).isExpired())
                    { it.remove(); }
                }
            for(final Iterator<String> it = byListenerID.keySet().iterator(); it.hasNext();)
                {
                if(byListenerID.get(it.next()).isExpired())
                    { it.remove(); }
                }
            }
        }


    /**Abstract/base listener object waiting for stats to arrive.
     * This has an embedded unique listener ID,
     * cryptographically generated,
     * and pure-alphanumeric-ASCII suitable for embedding in URLs, etc.
     */
    public static abstract class AbstractStatsListener implements ObjectInputValidation
        {
        /**Construct an instance.
         *
         * @param uniqueDataPointID  unique data-point identifier for a data point to collect,
         *     only one data point with this name can be requested at any one time
         *     any previous one being ignored for example; never null nor empty
         * @param expireBy  time by which this listener should expire
         *     regardless of whether the listened-for data point has arrived;
         *     strictly positive
         */
        public AbstractStatsListener(final String uniqueDataPointID,
                                     final long expireBy)
            {
            this.uniqueDataPointID = uniqueDataPointID;
            this.expireBy = expireBy;

            try { validateObject(); }
            catch(final InvalidObjectException e) { throw new IllegalArgumentException(e.getMessage(), e); }
            }

        /**Unique listener ID: unique-for-all-time pure-alphanumeric-ASCII identifier for this listener.
         * This ID ends up in external URLs and identifies this listener
         * until it expires or removes itself.
         * <p>
         * Generated from a cryptographically-secure source,
         * so should not be possible to guess in advance.
         * <p>
         * The use of this may make creation of instances of this class expensive.
         */
        public final String uniqueListenerID = Long.toString(Rnd.goodRnd.nextLong() >>> 1, Character.MAX_RADIX);

        /**Unique data-point identifier: not more than one listener with given ID can exist at once; never null nor empty.
         * Can be used to contain, for example, something like VOTE-214.21.39,
         * eg a combination of a stats purpose and a unique identifier for a user such as part of their IP address,
         * such that we never ask them (or accept responses) for more than one vote per user.
         * <p>
         * Note therefore that this is not a unique-for-all-time identifier,
         * but rather a unique-data-point-to-collect identifier.
         */
        public final String uniqueDataPointID;

        /**Date/time that we expire by (ms).
         * Once this time has passed this listener will disappear.
         */
        private final long expireBy;

        /**Returns true if this listener has expired. */
        public boolean isExpired() { return(expireBy <= System.currentTimeMillis()); }


        /**Handle a (non-null) Map of input parameters (String name to String[] value array).
         * Return a redirection URL (or null for no redirection).
         * <p>
         * The ID string will generally be amongst the properties in its external form.
         * <p>
         * As a side-effect, this should so what ever processing it needs
         * in order to record the data point.
         * <p>
         * By default, this listener instance is removed once this routine has been called.
         * <p>
         * The ID (and thus this data point) has already been validated
         * by the time that this call is made.
         */
        public abstract String handle(final Map<String,String[]> parameters);


        /**The hash is based on the data point ID. */
        public int hashCode() { return(uniqueDataPointID.hashCode()); }

        /**Equality is on just the data point ID. */
        public boolean equals(final Object obj)
            {
            if(!(obj instanceof AbstractStatsListener)) { return(false); }
            final AbstractStatsListener other = (AbstractStatsListener) obj;
            return(uniqueDataPointID.equals(other.uniqueDataPointID));
            }


        /**Check state is valid; complain if not. */
        public void validateObject()
                throws InvalidObjectException
            {
            if((uniqueDataPointID == null) || (uniqueDataPointID.length() == 0))
                { throw new InvalidObjectException("bad object: null or empty uniqueDataPointID"); }
            if(expireBy <= 0)
                { throw new InvalidObjectException("bad object: non-positive expireBy time"); }
            }
        }
    }
