/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.Rnd;

/**Web-server-related constants.
 * This is for constants only of interest to Web apps.
 */
public final class WebConsts
    {
    /**Prevent construction of an instance. */
    private WebConsts() { }

    /**Web/Net standard line-end string; CRLF. */
    public static final String WEB_LINE_END = "\r\n";

    /**If true then try very hard to push out the first page a user sees on our site(s) ASAP.
     * <a href="http://news.bbc.co.uk/2/hi/technology/4616700.stm" target="_top">20050116: BBC News suggests</a>
     * that visitors make up their minds about a new Web site within 50ms,
     * so we will try to push out the first page
     * as fast as possible, preferably within this time.
     */
    public static final boolean FIRST_PAGE_AFAP = true;

    /**Mount point for whole WAR tree on J2EE master servers.
     * Note the leading ``/''.
     * <p>
     * On deployed front-ends the root should always be ``/'';
     * this applies to a WAR deployed as part of a J2EE app,
     * which cannot currently be deployed at the root of the Web
     * server (due to bugs in, for example, Sun's 1.3 RI J2EE SDK.
     */
    public static final String MASTER_CONTEXT_ROOT = "/master";
    /**Name of base of exhibits area below the WAR root; no leading or trailing `/'. */
    public static final String BASE_PATH_EXHIBITS = "_exhibits";
    /**Name of base of admin area below the WAR root; no leading or trailing `/'. */
    public static final String BASE_PATH_ADMIN = "_admin";
    /**Name of base of catalogue-page area below the WAR root; no leading or trailing `/'. */
    public static final String BASE_PATH_CATPAGE = "_c";
    /**Name of base of thumbnail/sample area below the WAR root; no leading or trailing `/'. */
    public static final String BASE_PATH_TN = "_tn";
    /**Name of base of static/cacheable content below the WAR root; no leading or trailing `/'. */
    public static final String BASE_PATH_STATIC = "_static";
    /**Name of standard thumbnail/sample area below BASE_PATH_TN; no leading or trailing `/'. */
    public static final String PATH_TN_STD = "std";
    /**Name of small thumbnail/sample area below BASE_PATH_TN; no leading or trailing `/'. */
    public static final String PATH_TN_SML = "sml";


    /**Suffix for secondary, generated HTML objects, including ``.''.*/
    public static final String F_secondary_generated_HTML_suffix = ".html";

    /**Suffix for secondary, generated WML objects, including ``.''. */
    public static final String F_secondary_generated_WML_suffix = ".wml";

    /**Suffix for secondary, generated XHTML objects, including ``.''. */
    public static final String F_secondary_generated_XHTML_suffix = ".xhtml";

    /**Suffix for secondary, generated XML objects, including ``.''. */
    public static final String F_secondary_generated_XML_suffix = ".xml";

    /**HTML welcome page for a directory. */
    public static final String F_HTML_welcome_page = "index" + F_secondary_generated_HTML_suffix;

    /**Target limit for page-download time in ms; strictly positive. */
    public static final int MAX_PG_DOWNLOAD_MS = 1500;

    /**General (client/proxy) page-cache time, including catalogue pages (ms).
     * This can be used for positive and negative cacheing, and can be of
     * the order of a few minutes (much, much shorter than the timeframe
     * of changes in the exhibit collection [typically days]), and long
     * enough to cover typical use of the BACK button.
     * <p>
     * This should be much shorter than (eg less than half) the system's
     * default parallel slackness so as to ensure that the a gets rechecked
     * often enough to catch important changes automatically.
     * <p>
     * A big-enough value will help caches/proxies and "Web accelerators",
     * and generally improve perceived performance.
     */
    public static final int DEFAULT_PAGE_CACHE_MS = CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 456;

    /**Cache/expiry time for pages we want captured in search engines (ms).
     * We set a long expiry to encourage search engines to index the page;
     * of the order of days to months is probably right.
     * <p>
     * This may force HTTP/1.0-browser users (esp IE5/IE6) to hit RELOAD
     * more often than we'd like...
     * <p>
     * We may use this value only when the page/data looks like it is
     * being retrieved by a search engine spider/robot.
     * <p>
     * A value of several days to several months is probably reasonable.
     */
    public static final long SPIDER_PAGE_EXPIRY_MS = 183 * (24 * 60 * 60 * 1000L); // ~6M.

    /**Spider revisit time in days; if zero it means not to attempt to use the "revisit-after" metatag.
     * This tag is reported to be obsolete.
     */
    public static final int SPIDER_REVISIT_DAYS = 0; // Math.max(1, (int) (SPIDER_PAGE_EXPIRY_MS / (24 * 60 * 60 * 1000L)));

    /**Default time to allow client to cache objects under /_static.
     * These are usually static items of "page furniture",
     * and allowing them to be served from cache for a long time
     * improved perceived performance and reduces server load.
     * <p>
     * There is no point having this less than the default respidering time.
     * <p>
     * Something of the order of many days to many months is probably good.
     */
    public static final long DEFAULT_STATIC_WEBITEMS_CACHE_MS = Math.max(SPIDER_PAGE_EXPIRY_MS,
        367 * 24 * 3600 * 1000L); // ~1Y.


    /**Name of application-scope Boolean attribute, if any, to indicate bandwidth/CPU overload.
     * If present, we expect this to be set when the server is very busy by
     * a throughput filter/monitor to FALSE when not busy and TRUE when busy.
     */
    public static final String BANDWIDTH_OVERLOAD_ATTR_NAME = "org.hd.d.pg2k.highTraffic";

    /**Name of application-scope Boolean attribute, if any, to indicate bandwidth/CPU light load.
     * If present, we expect this to be set when the server is fairly quiet by
     * a throughput filter/monitor to TRUE when quiet and FALSE otherwise.
     */
    public static final String BANDWIDTH_LIGHTLOAD_ATTR_NAME = "org.hd.d.pg2k.lowTraffic";

    /**Name of request-scope Long attribute, if any, to indicate If-Modified-Since date to check against for spiders.
     * If present and the client for a page (or other downloadable) is a spider/bot,
     * and the client presents an If-Modified-Since value,
     * then we check the If-Modified-Since value against this instead,
     * though we limit it to no more than SPIDER_PAGE_EXPIRY_MS in the past.
     * <p>
     * We may use this with ordinary clients (humans) too if under very high load.
     * <p>
     * This can be used to have spiders refetch less useful pages less often.
     */
    public static final String SPIDER_IF_MOD_SINCE_THRESHOLD_ATTR_NAME = "org.hd.d.pg2k.spider.IfModSince";


    /**Default name of bandwidth-limit system property (bytes per second). */
    public static final String BANDWITDH_LIMIT_BPS_DEFAULT_PNAME = "org.hd.d.pg2k.maxBytesPerSec";

    /**Default name of bandwidth-limit system property (gigabytes per 30 days). */
    public static final String BANDWITDH_LIMIT_GBP30D_DEFAULT_PNAME = "org.hd.d.pg2k.maxGBytesPer30D";


    /**Minumum percent-saved before we will consider sending an exhibit compressed.
     * If (say) 5 then we must know a particular exhibit is compressable to
     * no more than 95% of its original size before we will consider
     * sending it compressed over the wire (eg when the site is busy).
     * <p>
     * Normally we don't bother trying to compress exhibits for download
     * because:
     * <ul>
     * <li>Most of them are already very efficiently encoded,
     *     ie hardly compressable,
     *     and the extra CPU time at our end is probably unjustified.
     * <li>We risk tripping up browsers with the non-standard encoding.
     * <li>Any resend or byte-range fetch may be more difficult to handle.
     * </ul>
     */
    public static final int MIN_PC_COMPRESSABLE_BEFORE_ENCODE = 5;


    /**Maximum length of simple-text-query in bytes (pure-ASCII characters). */
    public static final int MAX_SIMPLE_QUERY_LEN = 128;


    /**In catalogue entries display spaces instead of hyphens.
     * This is mainly needed because Netscape 4.x and 6.x refuses to
     * correctly hyphenate inside nbr section seven when heavily
     * prompted with wbr tags!
     */
    public static final boolean DISPLAY_SPACE_IN_EXHIBIT_NAMES = true;


    /**If true, allow browsers to scale JPEG/GIF images when it saves us significant work. */
    public static final boolean ALLOW_BROWSER_IMAGE_SCALE = true;

    /**If true, we force thumbnails to be made if we can use them in (catalogue) pages. */
    public static final boolean EAGER_TN_USE = true;


    /**Optional attributes for insertion at end of thumbnail img tags.
     * Can be "", so that the img tag will have the basic minimal/default
     * behavior, or something like " border=0" to turn off
     * a visible border if the image is clickable (note the leading space).
     */
    public static final String TN_IMG_TAG_EXTRA_ATTR = " border=0";


    /**If true, avoid using last-modified timestamps on HTTP pages.
     * This is because they provoke browsers into doing If-Modified-Since
     * requests to try to improve peformance, and those headers trip up
     * Tomcat 4.0.1.
     */
    public static final boolean AVOID_LAST_MODIFIED = true;

    /**Method by which search form is posted; can be "post" or "get". */
    public static final String SEARCH_FORM_METHOD = "get";

    /**URL of page/JSP to handle 404 (not found) errors.
     * Given as an absolute URI to work from any context (current URI).
     */
    public static final String NOT_FOUND_PAGE = "/notFound.jsp";

    /**URL of URI requested by spiders to check if spidering is allowed..
     * Given as an absolute URI to work from any context (current URI).
     */
    public static final String ROBOTS_PAGE = "/robots.txt";


    /**Name of (page-level) attribute in which last search time (ms) is stored (as a Long).
     * If null or a negative value (or not a Long) then no search has been done
     * through the standard mechanisms.
     */
    public static final String SEARCH_TIME_MS_ATTR_NAME = "SEARCH_TIME_MS";

    /**Name of (request-level) attribute in which JSP page start time is stored (as a Date). */
    public static final String REQUEST_START_DATETIME_ATTR_NAME = "JSP_REQUEST_START_TIMEDATE";


    /**When generating large chunks of paginatable output, how many to limit each page to where possible.
     * A value of 10 to 20 is typical.
     */
    public static final int MAX_RESULTS_PER_PAGE = 10;

    /**Maximum number of pages to show, or maximum number of page buttons if there have to be more pages.
     * Should maybe be a fraction of page width?
     * <p>
     * A value of 10 to 20 is typical.
     */
    public static final int MAX_RESULTS_PAGES = 15;


    /**Request-level attribute name used to communicate breadcrumbs (if any) to standard page-header routine. */
    public static final String STDHDR_ATTR_NAME_BREADCRUMBS = "breadcrumbs";
    /**Request-level attribute name used to communicate title String to standard page-header routine.
     * Title text must be short and 7-bit printable ASCII possibly containing HTML entity codes.
     */
    public static final String STDHDR_ATTR_NAME_TITLE = "title";
    /**Request-level attribute name used to communicate sub-title String to standard page-header routine.
     * Sub-title text must be short and 7-bit printable ASCII possibly containing HTML entity codes.
     * <p>
     * May be ignored if title is not set.
     */
    public static final String STDHDR_ATTR_NAME_SUBTITLE = "subTitle";
    /**Request-level attribute name used to communicate keywords String to standard page-header routine.
     * Sub-title text must be short and 7-bit printable ASCII possibly containing HTML entity codes.
     * <p>
     * May be ignored if title is not set.
     */
    public static final String STDHDR_ATTR_NAME_KEYWORDS = "keywords";

    /**Request-level attribute name used to indicate that this page should not be SE food to standard page-header routine.
     * If set non-null then we may wish to insert a robots noindex meta tag,
     * and we may also want to supress some or all ads on the page.
     * <p>
     * This may be the case for low-content pages, such as navigation pages.
     * <p>
     * The "safe" default is to let the SEs index everything
     * and make the quality/relevance judgements!
     */
    public static final String STDHDR_ATTR_NAME_NOINDEX = "pg2k.page.noindex";

    /**If true then we make use of the STDHDR_ATTR_NAME_NOINDEX mechanism. */
    public static final boolean USE_NOINDEX_ON_LOW_CONTENT_PAGES = true;


    /**When paginating, name of parameter (HTML fieldname and GET/POST field name) to pass page number in.
     * We keep this short because it gets used lots of times, sometimes many times per page.
     */
    public static final String PAGE_NUMBER_PARAMETER = "pg";


    /**URI of JSP page that handles display of browsable exhibit trees. */
    public static final String BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP = "/_cat/doHTMLTree.jsp";
    /**Name of the (request-level) attribute used to pass the (unmodifiable, sorted) List of exhibits to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_LIST_ATTR_NAME = "org.hd.d.pg2k.BETDJ_LIST_ATTR";
    /**Name of the (request-level) attribute used to pass the boolean array of uninteresting parent nodes to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_LIST_UNIN_NAME = "org.hd.d.pg2k.BETDJ_UNIN_ATTR";
    /**Name of the (request-level) attribute used to pass (unmodifiable, String-to-Integer) Map of sub-nodes to exhibit counts BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_LIST_SNEXCOUNT_NAME = "org.hd.d.pg2k.BETDJ_SNEXCOUNT_ATTR";
    /**Name of the (request-level) attribute used to pass the String base absolute URI of the tree to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_BASE_URI_ATTR_NAME = "org.hd.d.pg2k.BETDJ_BASE_URI_ATTR";
    /**Name of the (request-level) attribute used to pass the String title of the tree to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_TITLE_ATTR_NAME = "org.hd.d.pg2k.BETDJ_TITLE_ATTR";
    /**Name of the (request-level) attribute used to pass the String prefix word sequence of the tree to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP. */
    public static final String BETDJ_PRWSEQ_ATTR_NAME = "org.hd.d.pg2k.BETDJ_PRWSEQ_ATTR";
    /**Name of the (request-level) attribute used to pass the String HTML general tree description to BROWSABLE_EXHIBIT_TREE_DISPLAY_JSP.
     * This is allowed to contain structural elements such as headings for now,
     * though in future might not be allowed to.
     */
    public static final String BETDJ_HTML_GENERAL_DESC_ATTR_NAME = "org.hd.d.pg2k.BETDJ_HTML_GENERAL_DESC_ATTR";
    /**Name of common section/category dir (String) for all exhibits in the view, or null if none or not known. */
    public static final String BETDJ_COMMON_SECTION = "org.hd.d.pg2k.BETDJ_COMMON_SECTION";

    /**Root of "virtual" collections tree, starting and ending with "/". */
    public static final String VIRTUAL_COLLECTIONS_ROOT = "/_virtual/";

    /**Root of "ByAuthor" collections forest, ending with "/". */
    public static final String VIRTUAL_COLLECTIONS_BYAUTHOR_ROOT = VIRTUAL_COLLECTIONS_ROOT + "ByAuthor/";

    /**Root of "ByCategory" collections forest, ending with "/". */
    public static final String VIRTUAL_COLLECTIONS_BYCATEGORY_ROOT = VIRTUAL_COLLECTIONS_ROOT + "ByCategory/";

    /**Root of custom icons collection, ending with "/". */
    public static final String VIRTUAL_COLLECTIONS_CUST_ICONS_ROOT = VIRTUAL_COLLECTIONS_ROOT + "Cust_icons/";

    /**Best-exhibits collection page. */
    public static final String VIRTUAL_COLLECTIONS_BEST_PAGE = VIRTUAL_COLLECTIONS_ROOT + "best.jsp";

    /**New-exhibits collection page. */
    public static final String VIRTUAL_COLLECTIONS_NEW_PAGE = VIRTUAL_COLLECTIONS_ROOT + "new.jsp";

    /**Most-downloaded-exhibits collection page. */
    public static final String VIRTUAL_COLLECTIONS_MOST_DOWNLOADED_PAGE = VIRTUAL_COLLECTIONS_ROOT + "downloads.jsp";




    /**Base of custom sites tree, starting and ending with "/". */
    public static final String VIRTUAL_SITES_ROOT = "/_site/";

    /**Root of custom LondonPix site, starting and ending with "/". */
    public static final String VIRTUAL_SITE_CUST_LONDONPIX_ROOT = VIRTUAL_SITES_ROOT + "LondonPix/";

    /**Root of custom LondonPix site, starting and ending with "/". */
    public static final String VIRTUAL_SITE_CUST_ZAPIX_ROOT = VIRTUAL_SITES_ROOT + "ZAPix/";

    /**Root of custom AlohaEarth site, starting and ending with "/". */
    public static final String VIRTUAL_SITE_CUST_ALOHAEARTH_ROOT = VIRTUAL_SITES_ROOT + "AlohaEarth/";

    /**Root of custom SmlPx site, starting and ending with "/". */
    public static final String VIRTUAL_SITE_CUST_SMLPX_ROOT = VIRTUAL_SITES_ROOT + "SmlPx/";


    /**Stats-sink receiver page root-relative URL. */
    public static final String STATS_SINK_PAGE_RRURL = "/statsSink.jsp";


    /**Most thumbnails to show on one of the best/new/etc collections pages. */
    public static final int SINGLE_PAGE_CONTACT_SHEET_TN_COUNT = 100;


    /**If true have the catalogue page generic handler use jsp:forward where possible to get to specific handler. */
    public static final boolean CAT_HANDLER_USE_JSP_FORWARD = false;


    /**Expected typical mean bandwidth per active/sticky user, bytes per second; strictly positive.
     * We may override this default value from (say) GenProps
     * or from collected stats.
     * <p>
     * Not expected to be terribly precise,
     * but should err a little on the high side.
     * <p>
     * Used, amongst other things, for a server to estimate its own capacity.
     * <p>
     * As of 20050510:
     * <ul>
     * <li>about 4kBps per visitor that views more than one page;
     * <li>about 1300Bps per distinct class-C IP block of all users of the site including hotlinkers.
     * </ul>
     * <p>
     * As of 20060116: revising per-sticky-user estimate downwards,
     * but all spiders and hotlinkers (etc) add significant per-server load,
     * so a simple overall estimate of 2kBps is possibly more accurate.
     */
    public static final int DEFAULT_EXPECTED_USER_BW_BYTESPERSEC = 2048;


    /**If true, attempt to break out of third-party HTML frames.
     * We may want to do this to stop people accidentally
     * (or deliberately, to steal revenue, for example)
     * wrapping our pages,
     * though this may break Google search's cacheing features for example.
     */
    public static final boolean DO_FRAME_BREAKOUT = true;

    /**If true, enable Urchin (ie Google) site stats tracking. */
    public static final boolean ENABLE_URCHIN_STATS = true;

    /**Request-level attribute set non-null when Urchin code has already been set in the HEAD or page text. */
    public static final String URCHIN_REQ_ATTR_NAME = "urchinOn";


    /**Unique prefix for data-point name for vote StatsSink IDs. */
    public static final String VOTER_DATA_POINT_PREFIX = "VOTER-";

    /**Minimum time between requests to one user to vote, and also how long they get to vote, ms; non-negative.
     * Small values may help shut out spiders and collect more votes,
     * but at an increased risk of annoying users
     * and of discarding votes from thoughtful/slow users.
     * <p>
     * A value of the order of minutes is usually suitable.
     * <p>
     * Once a user has voted we should generally wait many times as long
     * before requesting another vote
     * eg in order to suppress false votes from spiders
     * that we otherwise fail to screen out.
     */
    public static final int VOTE_MIN_REQUEST_GAP_MS = 7 * 60 * 1000 +
            Rnd.fastRnd.nextInt(3 * 60 * 1000);

    /**If true, use JavaScript to attempt to auto-hide vote buttons (roughly) when the vote permit expires. */
    public static final boolean VOTE_BUTTONS_AUTO_HIDE_ON_EXPIRY = true;


    /**Name of 15x15 "smile"/"pro"/"for" image within static-icon area. */
    public static final String PRO_ICON_15x15_NAME = "smile.gif";
    /**Name of 15x15 "frown"/"con"/"against" image within static-icon area. */
    public static final String CON_ICON_15x15_NAME = "frown.gif";

    /**If true then allow users to be pushed to a local country mirror.
     * To avoiding bouncing people around
     * (eg in case of inconsistent algorithm versions)
     * or making wrong guesses about which is the "best" machine for a user,
     * and overly polluting the users' minds with the non-canonical name,
     * any such "bouncing" has to be very conservative.
     * <p>
     * Typically only done if:
     * <ul>
     * <li>The user has no extant session (so no state to lose when bounced).
     * <li>The user is on the generic URL, ie no explicit mirror URL.
     * <li>The user is not already on a server "close" to them (eg same continent).
     * <li>There is a same-country mirror "better" than the current server
     *     for this user, and it is not too busy.
     * <li>Possibly Only do this for text pages to avoid bouncing
     *     (for example) thumbnail requests that must go to the same host.
     * <li>Possibly done for only a fraction of visitors
     *     and possibly only on selected pages so that
     *     if the user cannot reach the "better" host
     *     then they can "stick" to the generic one by doing a reload/refresh.
     *     Users should always be able to get to the home page without bouncing.
     * </ul>
     * <p>
     * Any such redirection would be temporary (302) and not done to spiders/bots.
     */
    public static final boolean ALLOW_BOUNCE_TO_LOCAL_MIRROR = true;


    /**Recommended/widely-supported screen-width for .mobi devices/sites, in pixels. */
    public static final int MOBI_REC_DEFAULT_SCREEN_WIDTH_PX = 120;

    /**Recommended maximum page weight for .mobi devices/sites, in bytes. */
    public static final int MOBO_REC_MAX_PG_WEIGHT_BYTES = 20 * 1000;
    

    /**GenProps generic property name prefix used to pull in per-URL custom HTML head text/tags.
     * This is the lookup key, ie excluding the "gen." prefix.
     * The literal full (case-sensitive) request URL should be appended.
     */
    public static final String GP_HEAD_TAG_INSERT_KEY_PREFIX = "head-insert.";


    /**If true then we canonicalise tree-display URIs to try to reduce state-space with 301/302 redirects.
     * We do this to reduce the number of different pages (and URI lengths)
     * that SEs and end-users may need/try to fetch and cache,
     * without reducing the actual information available.
     * <p>
     * This should reduce noise and server load!
     * <p>
     * We use 301 redirects to deal with non-standard URL forms,
     * eg missing a trailing '/' or not monocased,
     * since the duplicates that we might otherwise accept never have independent value.
     * <p>
     * We use 302 redirects to send the client to a higher node (shorter URI)
     * that contains the same information.  This can only be a temporary rediect
     * because the tree shape (and thus the/any redirection)
     * may change at any time in principle.
     */
    public static final boolean CANON_TREE_URIS_WITH_REDIR = true;



    /**URL of (external, BETA) message board. */
//    public static final String EXT_MSG_BOARD_URL = "http://www.activeboard.com/forum.spark?forumID=56026";
    }
