/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.util;

import java.awt.image.*;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.SoftReference;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;

import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME.ExhibitTypeParameters;
import org.hd.d.pg2k.svrCore.location.*;
import org.hd.d.pg2k.svrCore.props.*;
import org.hd.d.pg2k.svrCore.vars.*;
import org.hd.d.pg2k.webSvr.exhibit.*;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean.AEPLinkedKey;
import org.hd.d.pg2k.webSvr.virtualHosts.VirtualHosts;
import org.w3c.dom.Node;

import ORG.hd.d.IsDebug;

/**Web-server-related utility functions.
 * This is for algorithms only of interest to Web apps, often JSPs.
 * <p>
 * One advantage of having code here rather than in-line in a JSP
 * is that is is pre-compiled off-line for speed and robustness;
 * code here is also easier to test.
 */
public final class WebUtils
    {
    /**Prevent construction of an instance. */
    private WebUtils() { }


    /**Name of event/variable to which we post a voter's approximate geo location. */
    public static final SimpleVariableDefinition VOTER_LOCATION_STATS_EVENT_DEF = SystemVariables.GENSTATS_STRING_GLOBAL_EVENT;

    /**Prefix of event value for voter's approximate geo location. */
    public static final String VOTER_LOCATION_STATS_EVENT_PREFIX = "voterLocation=";


    /**Private limited-size cache for chooseMirrorHostToBalanceLoad() for best mirror for given client IP.
     * Map from client IP address to full mirror host name and time mirror was computed.
     * <p>
     * A lock can be held on this object to make multiple operations atomic.
     * <p>
     * Entries in this map go stale very quickly
     * (in a matter of tens of seconds at most)
     * and are mainly here to avoid repeated recalculations while preparing
     * one HTML page to serve to the client for example.
     * <p>
     * This has a fixed maximum size to limit memory consumption
     * and to avoid the need for explicit cleaning of old entries
     * (they can just hang around until displaced).
     * We vary this maximum size a little to try to avoid deadly embraces.
     * <p>
     * A size limit of a few thousand to allow for a few mega-proxies,
     * and a time limit of some tens of seconds,
     * is probably about right.
     */
    private static final MemoryTools.SimpleLRUMap<InetAddress,Tuple.Pair<String,Long>> _cache_cMHTBL_byClient =
            new MemoryTools.SimpleLRUMap<InetAddress, Tuple.Pair<String,Long>>(10001 + Rnd.fastRnd.nextInt(4001));

    /**Maximum time an entry in _cache_cMHTBL_byClient is valid in ms; strictly positive.
     * Should be no more than the maximum staleness of the mirror liveness variables,
     * and is mainly intended to avoid repeated recalculations during the preparation
     * of a single HTML page for a user giving multiple different values on that one page...
     */
    private static final int MAX_cache_cMHTBL_byClient_AGE_MS = 1001 +
            VarTools.MIN_AGE_MS/3;

    /**Lock to prevent multiple simultaneous polls of the main or "loopback" URLs at once.
     * Otherwise multiple incoming requests may trigger multiple outgoing polls
     * which actually overwhelm (or trigger throttling by) the remote machine(s).
     */
    private static final Object _lock_fallback_URL_poll = new Object();

    /**Fraction of time we redistribute load to non-optimal mirrors; strictly positive.
     * A higher value means that we redistribute traffic
     * to non-optimal mirrors less often.
     * <p>
     * A little bit of redistributed traffic makes the system more robust
     * and primed for action, but may give some unlucky users poor performance.
     * <p>
     * We make this probability lower (this value higher)
     * as the potential (maximum) cost of a wrong choice goes up.
     * The biggest error we can make for most user is to send them to
     * the furthest corner of the planet when there is a same-COUNTRY server.
     */
    private static final int REDIST_FRAC = GeoProximity.COUNTRY.getCloseness() + Rnd.fastRnd.nextInt(GeoProximity.COUNTRYGROUP.getCloseness());

    /**If true, always select the fastest available mirror; don't use any stochastic factor. */
    private static final boolean ALWAYS_SELECT_FASTEST_MIRROR = false;

    /**Get name (and optional port) of server for best user experience; never null/empty.
     * This is a hostname[:port] suitable to insert into the URL
     * for retrieving an exhibit or another page, etc.
     * <p>
     * This will attempt to pick one based on a number of factors
     * such as current load, client proximity, which mirrors are up,
     * which mirrors are one the same AEP version as us, etc.
     * <p>
     * If this cannot chose a suitable server/mirror name
     * it returns the generic "main" name, CoreConsts.MAIN_DATA_HOST.
     * <p>
     * Though either or both arguments can be null,
     * this is most likely to give useful non-default answers
     * when both values are non-null.
     * <p>
     * A couple of items of hardwired policy need to be pointed out:
     * <ul>
     * <li>If the client for this request looks as if it is from a spider
     *     (and therefore may cache the results for a long time)
     *     then we always return the generic main hostname for speed
     *     and for long-term robustness.
     * <li>If there is no request/IP address for the client
     *     (so we could do no more than statistical load balancing at best)
     *     then we always return the generic main hostname for speed
     *     and in the hope that we will be doing
     *     DNS-based load/proximity stuff sometime.
     * </ul>
     *
     * @param request  the incoming request or null if none available
     * @param vars  access to the system variables or null if not available
     *
     * @return a "mirror" subdomain of CoreConsts.MAIN_DATA_HOST,
     *     or by default CoreConsts.MAIN_DATA_HOST
     *     thus relying on DNS-based balancing
     */
    public static String chooseMirrorHostToBalanceLoad(final HttpServletRequest request,
                                                       final DataSourceBean vars)
        {
        // The default host name to fall back to
        // in the absense of any available/suitable mirrors.
        // We must not cache the fallbackHost by IP address
        // because multiple clients may share a single IP address (eg at a large proxy)
        // and the fallbackHost may be computed depending on client-specific features.
        String fallbackHost = CoreConsts.MAIN_DATA_HOST;

        // Get the client's IP address if possible.
        // If we can then see if we have a valid cached result for this exact client IP
        // and return it immediately!
        // This is the fastest-possible path through the routine.
        InetAddress clientIP = null;
        if(request != null)
            {
            try
                {
                clientIP = InetAddress.getByName(request.getRemoteAddr());
                assert(clientIP != null);

                // Look in cache for non-stale entry for this exact client IP...
                // (Fall through to do normal computation if not present.)
                synchronized(_cache_cMHTBL_byClient)
                    {
                    final Tuple.Pair<String,Long> entry = _cache_cMHTBL_byClient.get(clientIP);
                    if(entry != null)
                        {
                        if(entry.second.longValue() >= System.currentTimeMillis())
                            { return(entry.first); }
                        // Remove the useless stale entry that will otherwise keep being held onto LRU!
                        _cache_cMHTBL_byClient.remove(clientIP);
                        }
                    }
                }
            catch(final UnknownHostException e) { /* Cannot parse client address... */ }
            }
        // No request and thus no client IP address,
        // so always return the generic/fallback hostname.
        else
            { return(fallbackHost); }


        // If the client looks like it might cache the result for a long time
        // then we give it the generic domain for robustness (and speed).
        // The most obvious case is spidering of pages to go in a search engine
        // for recall years later when specific mirrors may have come and gone.
        if((request != null) && requestProbablyFromSpider(request))
            { return(fallbackHost); }


        // If we can extract the host from the original request URL,
        // and it is a legitimate name/alias for the site,
        // then make it (and any non-standard port) the fallback.
        if(request != null)
            {
            try
                {
                final URL requestURL = new URL(request.getRequestURL().toString());
                final String ourName = requestURL.getHost();
                final int ourPort = requestURL.getPort();
                final String fullName = ((ourPort == 80) || (ourPort == -1)) ? ourName :
                                         (ourName + ':' + ourPort);
                if(!fullName.equals(fallbackHost) &&
                        (null != VirtualHosts.getVirtualHostDetails(ourName, null)))
                    { fallbackHost = fullName; }
                }
            catch(MalformedURLException e)
                {
                e.printStackTrace(); // Shouldn't really happen...
                }
            }

//System.out.println("chooseMirrorHostToBalanceLoad(): *** START CALC *** : clientIP = " + clientIP);


        // In the case that we have the client's request available
        // and the default/canonical domain/host seems to be unavailable,
        // then we'll use the hostname the client specified for us
        // in its request URL as the fallback.
        // (This also allows off-line testing of the Gallery!)
        if((request != null) && (fallbackHost.equals(CoreConsts.MAIN_DATA_HOST)))
            {
            // Prevent multiple redundant polls on the main/own URLs from occuring at once,
            // which may have the effect of ovewhelming the remote site.
            synchronized(_lock_fallback_URL_poll)
                {
                final Boolean mainDomainIsUp =
                    LoadBalancingUtils.testIfHTTPServerIsUp(CoreConsts.MAIN_DATA_HOST, true);

                // Try the client's suggested hostname
                // if the main host is definitely down
                // (because otherwise we are stuck!).
                // Security note: we must not trust this client-supplied
                // information nor cache it.
                if(Boolean.FALSE.equals(mainDomainIsUp))
                    {
                    try
                        {
                        // If the main host is not available,
                        // but the name under which the client arrived at this server *is*,
                        // then make the client's name for us the fallback hostname.
                        // We don't cache these client-supplied names and their liveness results,
                        // since we cannot necessarily trust them and they may be unbounded in number.
                        final URL requestURL = new URL(request.getRequestURL().toString());
                        final String ourName = requestURL.getHost();
                        final int ourPort = requestURL.getPort();
                        final String fullName = ((ourPort == 80) || (ourPort == -1)) ? ourName :
                                                 (ourName + ':' + ourPort);
                        // Don't bother testing liveness
                        // if the client's request hostname is already the fallback host...
                        // Nor if this does not look like one of our valid virtual names
                        // (to avoid pestering third-party sites by mistake/proxy).
                        if(!fullName.equals(fallbackHost) &&
                                (null != VirtualHosts.getVirtualHostDetails(ourName, null)))
                            {
                            // Do not look these up asynchronously
                            // nor cache the results.
                            if(Boolean.TRUE.equals(LoadBalancingUtils.testIfHTTPServerIsUp(fullName, false)))
                                {
System.err.println("WARNING: chooseMirrorHostToBalanceLoad(): having to fall back to our host as " + fullName);
                                fallbackHost = fullName;
                                }
                            }
                        }
                    catch(IOException e) { } // Ignore/absorb.
                    }
                }
            }


        // If we don't have access to variables to check for live mirrors
        // then return the fallback host immediately...
        // Don't cache this result,
        // as we didn't have all information to hand that we should have had.
        if(vars == null)
            { return(fallbackHost); /* Cannot find a working mirror, and can't cache this result. */ }


        // See what mirrors *claim* to be available...
        // We will check in a moment if they actually are...
        final Map<String,Long> liveMirrors = LoadBalancingUtils.getActiveMirrors(vars, true);
        // If there are no mirrors claiming to be live
        // then return immediately with the generic name.
        if(liveMirrors.size() < 1)
            { return(fallbackHost); /* Cannot find a working mirror, and can't cache this result. */ }

if(LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): mirror bandwidths: " + liveMirrors); }


        // Filter out any mirrors not using the same AEP version as us.
        // Else the user may, for example, get a 404 on a new exhibit!
        try
            {
            final AllExhibitProperties aep = vars.getAllExhibitProperties(-1);
            final Long ourHash = new Long(aep.longHash);

            // First collect all those mirrors whose hash is correct...
            // Usually it will be all of them.
            final Set<InstanceID> goodMirrorIDs = new HashSet<InstanceID>(liveMirrors.size());
            final SimpleVariableValue hashes =
                    vars.getVariable(SystemVariables.ThroughputMonitorFilter_AEP_LONGHASH);
            if(hashes != null)
                {
                // Zap any stale values...
                final SimpleVariableValue h2 = hashes.removeAllKeysOlder(System.currentTimeMillis() - VarTools.MIN_AGE_MS);
                final Map<InstanceID,SimpleVariableValue> globalMap = h2.getGlobalMap();
                if(globalMap != null)
                    {
                    for(final InstanceID id : globalMap.keySet())
                        {
                        if(ourHash.equals(globalMap.get(id).getValue()))
                            { goodMirrorIDs.add(id); /* Found a good mirror... */ }
                        }
                    }
                }

            // If none of the candidate mirrors has the right AEP hash,
            // then return immediately but don't cache this result.
            if(goodMirrorIDs.size() < 1)
                {
if(IsDebug.isDebug || LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): no mirror on same AEP hash as us"); }
                return(fallbackHost); /* Cannot find a suitable mirror, and can't cache this result. */
                }

if(IsDebug.isDebug || LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): mirrors on same AEP hash as us: " + new ArrayList<InstanceID>(goodMirrorIDs)); }

            // Convert list of good IDs to list of good mirrors...
            final SimpleVariableValue activeMirrors =
                    vars.getVariable(SystemVariables.ThroughputMonitorFilter_ACTIVE_MIRROR_NAME);
            final Set<String> goodMirrorTags = new HashSet<String>(goodMirrorIDs.size());
            if(activeMirrors != null)
                {
                final Map<InstanceID,SimpleVariableValue> globalMap = activeMirrors.getGlobalMap();
                if(globalMap != null)
                    {
                    for(final InstanceID id : goodMirrorIDs)
                        {
                        final SimpleVariableValue svv = globalMap.get(id);
                        if(svv != null)
                            {
                            final Object value = svv.getValue();
                            if(value instanceof String)
                                { goodMirrorTags.add((String) value); }
                            }
                        }
                    }
                }

            // Iterate over live mirrors, zapping any unsuitable ones.
            for(final Iterator<String> it = liveMirrors.keySet().iterator(); it.hasNext(); )
                {
                final String mirror = it.next();
                if(!goodMirrorTags.contains(mirror)) { it.remove(); }
                }

            // If none of the candidate mirrors has the right AEP hash,
            // then return immediately but don't cache this result.
            if(liveMirrors.size() < 1)
                {
if(IsDebug.isDebug || LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): no mirror on same AEP hash as us from candidate set"); }
                return(fallbackHost); /* Cannot find a suitable mirror, and can't cache this result. */
                }
            }
        catch(final Exception e)
            {
            // Could not filter by hash for some reason...
            // Whinge, but continue, as this is not essential...
            e.printStackTrace();
            }


        // Sort the mirrors into bandwidth order and drop any known to be dead
        // (or unreachable by us).
        final List<String> orderedList = LoadBalancingUtils.orderMirrorTagsBestBandwidthFirst(liveMirrors);
        // If they've all been filtered out
        // then return immediately with the fallback name.
        if(orderedList.size() < 1)
            { return(fallbackHost); /* Cannot find a working mirror, and can't cache this result. */ }


        // If this host is a mirror
        // and this mirror appears at all in the list of surviving candidates
        // and the user has explicitly named this mirror alias as the hostname
        // then return this mirror name as the preferred mirror
        // so that the user's explicit mirror choice is respected.
        if(request != null)
            {
            final String mirrorTag = LocalProps.getMirrorTag();
            if((mirrorTag != null) && orderedList.contains(mirrorTag))
                {
                try
                    {
                    final URL requestURL = new URL(request.getRequestURL().toString());
                    final String hostName = requestURL.getHost();
                    final String mirrorName = LoadBalancingUtils.makeMirrorNameFromTag(mirrorTag);
                    if(mirrorName.equalsIgnoreCase(hostName))
                        { return(mirrorName); }
                    }
                catch(final MalformedURLException e)
                    { /* Ignore/absorb this error... */ }
                }
            }


        // If we always select the fastest mirror,
        // or if we only have one candidate,
        // then work through the list fastest-first for speed/simplicity...
        if(ALWAYS_SELECT_FASTEST_MIRROR || (orderedList.size() == 1))
            {
//System.out.println("chooseMirrorHostToBalanceLoad(): mirrors to try in order: " + orderedList);

            // One at a time,
            // from first (best) to worst (last)
            // check if we know the mirror to be alive or dead,
            // of if we have an expired cache entry,
            // then check it explicitly and update the cache accordingly.
            // Return the first (fastest) live mirror found.
            final String result = LoadBalancingUtils.findFirstWorkingMirror(orderedList);
            if(result != null)
                {
                // File the result and return it...
                if(clientIP != null)
                    { _cache_cMHTBL_byClient.put(clientIP, new Tuple.Pair<String, Long>(result, new Long(System.currentTimeMillis() + MAX_cache_cMHTBL_byClient_AGE_MS))); }
                return(result);
                }

            // Cannot choose a specific mirror so return the fallback host name.
            // If nothing else, and if this is the generic name,
            // then this may achieve some DNS-based load balancing.
            return(fallbackHost); // Cannot find a working mirror.
            }

        // Compute the client's location as a country code or
        // (less good) a region/registry
        // if there is actually a choice of mirrors...
        // Compute a rough-and-ready quick value,
        // and sometimes additionally try to compute a more accurate value,
        // but more slowly.
        String clientLocationQuick = null;
        String clientLocationSlow = null;
        if(clientIP != null)
            {
            // Get approximate location first with a quick lookup...
            // We'll be happy if this gives a country or a region/registry name.
            // We don't need complete accuracy on this,
            // just a rough indication for routing.
            clientLocationQuick = GeoUtils.getRegionByAddress(clientIP, true);

            // If we haven't got a country OR region code
            // then do a full (slow, expensive) lookup.
            // We could skip this if all available mirrors are in one country/region.
            if(!GeoUtils.CCTLD.isSyntaticallyValidCcTLD(clientLocationQuick) &&
               !GeoUtils.isSyntaticallyValidRegistryName(clientLocationQuick))
                { clientLocationSlow = GeoUtils.getRegionByAddress(clientIP, false); }

//System.out.println("chooseMirrorHostToBalanceLoad(): client location: " + clientLocation);
            }


        // If we appear not to have any sort of valid country code or region
        // then we're not going to be able to do better
        // than just basing our choice on the available bandwidth.
        final boolean bandwidthOnly =
                (!GeoUtils.isSyntaticallyValidRegistryName(clientLocationQuick) &&
                 !GeoUtils.CCTLD.isSyntaticallyValidCcTLD(clientLocationQuick)) &&
                (!GeoUtils.isSyntaticallyValidRegistryName(clientLocationSlow) &&
                 !GeoUtils.CCTLD.isSyntaticallyValidCcTLD(clientLocationSlow));

        // Choose from available mirrors based on more factors than just bandwidth if possible.
        // We include a notion of proximity (ie RTT, packet loss, reliability, etc),
        // and stochastically spread load amongst servers in proportion to their capacity,
        // though having discarded the worst entries so as never to give too poor a performance.
        final Map<String,Long> weightings = new HashMap<String, Long>(1 + 2*orderedList.size());
        // By default, retain just the entries for items still in the ordered list of candidiates.
        for(final String tag : orderedList)
            { weightings.put(tag, liveMirrors.get(tag)); }

        if(!bandwidthOnly)
            {
            // Create a new map of available bandwidth multiplied by proximity
            // for all the remaining candidiates
            // and use it to re-sort our ordered list.
            // Compute two versions: one with our quick result and one with our slow result,
            // and use whichever gives the highest "top" weighting.

            // The best "top" weighting that we found so far.
            // Intially this is the bandwidth of the fastest mirror.
            final long bestWeight = liveMirrors.get(orderedList.get(0)).longValue();

            // Compute weightings with quick lookup...
            if(clientLocationQuick != null)
                {
                final Map<String,Long> weightingsQuick = new HashMap<String, Long>(1 + 2*orderedList.size());
                for(final String tag : orderedList)
                    {
                    final GeoUtils.CCTLD mirrorCC = new GeoUtils.CCTLD(tag.substring(0, 2)); // Extract country of mirror.
                    final int proximity = GeoUtils.computeProximity(clientLocationQuick, mirrorCC).getCloseness();
                    weightingsQuick.put(tag, new Long(liveMirrors.get(tag).longValue() * proximity));
                    }
                final ArrayList<String> tmp = new ArrayList<String>(orderedList);
                Collections.sort(tmp, new LoadBalancingUtils.BWOrder(weightingsQuick));
                final long topWeight = weightingsQuick.get(tmp.get(0)).longValue();
                // If this is better than before, reorder the main list and save the weightings.
                if(topWeight > bestWeight)
                    {
                    Collections.sort(orderedList, new LoadBalancingUtils.BWOrder(weightingsQuick));
                    weightings.clear();
                    weightings.putAll(weightingsQuick);
                    }
                }

            // Compute weightings with slow/accurate lookup...
            if(clientLocationSlow != null)
                {
                final Map<String,Long> weightingsSlow = new HashMap<String, Long>(1 + 2*orderedList.size());
                for(final String tag : orderedList)
                    {
                    final GeoUtils.CCTLD mirrorCC = new GeoUtils.CCTLD(tag.substring(0, 2)); // Extract country of mirror.
                    final int proximity = GeoUtils.computeProximity(clientLocationSlow, mirrorCC).getCloseness();
                    weightingsSlow.put(tag, new Long(liveMirrors.get(tag).longValue() * proximity));
                    }
                final ArrayList<String> tmp = new ArrayList<String>(orderedList);
                Collections.sort(tmp, new LoadBalancingUtils.BWOrder(weightingsSlow));
                final long topWeight = weightingsSlow.get(tmp.get(0)).longValue();
                // If this is better than before, reorder the main list and save the weightings.
                if(topWeight > bestWeight)
                    {
                    Collections.sort(orderedList, new LoadBalancingUtils.BWOrder(weightingsSlow));
                    weightings.clear();
                    weightings.putAll(weightingsSlow);
                    }
                }

//System.out.println("chooseMirrorHostToBalanceLoad(): mirror weightings: " + weightings);
            }


        // Sometimes, perturb our list so that the first item in the list
        // will be picked in proportion to its weighting
        // (it will be swapped from its usual rank with the top item)
        // and all other entries will be left as they are.
        // This should distribute some load to all mirrors.
        // We don't do this all the time because generally
        // we want to direct requests to the best (closest*fastest) mirror.
        long totalWeight = 0;
        for(final Long w : weightings.values())
            { totalWeight += w.longValue(); }
        if((totalWeight > Integer.MAX_VALUE) || (totalWeight < weightings.size()))
            { System.err.println("WARNING: weightings overflowed/underflowed in chooseMirrorHostToBalanceLoad(): " + totalWeight); }
        else if(Rnd.fastRnd.nextInt(REDIST_FRAC) == 0) // Sometimes perturb...
            {
            // Pick the mirror to promote...
            int r = Rnd.fastRnd.nextInt((int) totalWeight);
            for(int i = 0; i < orderedList.size(); ++i)
                {
                final String tag = orderedList.get(i);
                final int w = weightings.get(tag).intValue();
                if(r < w)
                    {
                    if(i != 0)
                        {
                        // This is the one we have selected to promote,
                        // and it isn't the top slot anyway,
                        // so swap it into position 0,
                        // and stop.
                        final String tmp = orderedList.get(0);
                        orderedList.set(0, tag);
                        orderedList.set(i, tmp);
if(LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): perturbed mirror list to: " + orderedList + " at probability " + (w / (float) totalWeight)); }
                        }
                    break;
                    }

                r -= w;
                }
            }


        // Go through the (possibly-reordered) mirror tag list in order,
        // using the first available working mirror we find
        // else drop back to the fallback host.
        final String result = LoadBalancingUtils.findFirstWorkingMirror(orderedList);
        if(result != null)
            {
if(LoadBalancingUtils.LOG_BALANCING_DECISIONS) { System.out.println("chooseMirrorHostToBalanceLoad(): *** BALANCING CHOICE MADE *** |clientIP|locationQuick|locationSlow|mirror| = |" + clientIP + "|" + clientLocationQuick + "|" + clientLocationSlow + "|" + result + "|"); }

            // File the result and return it...
            if(clientIP != null)
                { _cache_cMHTBL_byClient.put(clientIP, new Tuple.Pair<String, Long>(result, new Long(System.currentTimeMillis() + MAX_cache_cMHTBL_byClient_AGE_MS))); }
            return(result);
            }

        // Cannot choose a specific mirror so return the fallback host name.
        // If nothing else, and if this is the generic name,
        // then this may achieve some DNS-based load balancing.
        return(fallbackHost);
        }


    /**Simple class to allow logging to the given servlet's log(). */
    public static final class ServletLogger implements SimpleLoggerIF
        {
        public ServletLogger(final ServletContext ctxt)
            {
            if(ctxt == null) { throw new IllegalArgumentException(); }
            this.ctxt = ctxt;
            }

        /**Servlet context; never null. */
        private final ServletContext ctxt;

        /**Log the given message. */
        public void log(final String message)
            {
            ctxt.log(message);
            }
        }


    /**Simple class to allow logging to a given servlet's log() or System.out if none available.
     * This allows a logger to be created at instance scope
     * for (say) a Filter, and set with a context when the config is set
     * and cleared when one is not available,
     * all the while remaining a valid logger.
     */
    public static final class ServletLoggerWithFallback implements SimpleLoggerIF
        {
        /**Servlet context. */
        private volatile ServletContext ctxt;

        /**Set context, or remove it if null. */
        public void setContext(final ServletContext context) { ctxt = context; }

        /**Log the given message.
         * Logs to the servlet context logger if available,
         * else logs to System.out.
         */
        public void log(final String message)
            {
            final ServletContext context = ctxt;
            if(context != null) { context.log(message); }
            else { System.out.println(message); }
            }
        }


    /**System variables tried, in order, for a "popular" exhibit; private to getPopularExhibit().
     * We put the download var first to get a decent rate of update
     * since this value changes quite frequently.
     * <p />
     * We don't put the vote var first so as to reduce the temptation
     * to "throw" a vote to get an exhibit shown on the front page.
     */
    private static final SimpleVariableDefinition _gPE_vars[] = {
        SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD_LOCAL,
        SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD,
        SystemVariables.VOTE_PRO,
        SystemVariables.ACCESSPATTERN_CLICKTHROUGH,
        SystemVariables.ACCESSPATTERN_CAT_PAGE_VIEW,
        };

    /**Gets "popular" exhibit, possibly filtered by type; null if none available.
     * Tries to pick a "popular" exhibit by looking at one recently voted for,
     * or downloaded, etc, in the system variables,
     * and that has both thumbnails available where they are possible.
     * <p>
     * This rejects exhibits with a below-par (negative) rating.
     * <p>
     * Using the system variables should mean that this can pick up
     * values set from any mirror, etc, fairly quickly.
     * <p>
     * This cannot guarantee to return a non-null value,
     * but any value that it does return is a current, valid exhibit.
     *
     * @param dsb  handle on the system variables and data; never null
     * @param type  if not null only exhibits of this type are candidates
     * @deprecated Use {@link #getPopularExhibit(DataSourceBean,ExhibitMIME.ExhibitTypeParameters,Collection)} instead
     */
    public static String getPopularExhibit(final DataSourceBean dsb,
                                           final ExhibitMIME.ExhibitTypeParameters type)
        { return getPopularExhibit(dsb, type, null); }

    /**Gets "popular" exhibit, possibly filtered by type; null if none available.
     * Tries to pick a "popular" exhibit by looking at one recently voted for,
     * or downloaded, etc, in the system variables,
     * and that has both thumbnails available where they are possible.
     * <p>
     * This rejects exhibits with a below-par (negative) rating.
     * <p>
     * Using the system variables should mean that this can pick up
     * values set from any mirror, etc, fairly quickly.
     * <p>
     * This cannot guarantee to return a non-null value,
     * but any value that it does return is a current, valid exhibit.
     *
     * @param dsb  handle on the system variables and data; never null
     * @param type  if not null only exhibits of this type are candidates
     * @param exclude if non-null, any exhibits included by full name
     *     are not candidates to be returned
     */
    public static String getPopularExhibit(final DataSourceBean dsb,
                                           final ExhibitMIME.ExhibitTypeParameters type,
                                           final Collection<String> excludeFullNames)
        {
        if(dsb == null) { throw new IllegalArgumentException(); }

        for(final SimpleVariableDefinition def : _gPE_vars)
            {
            assert(def != null);

            // Half the time skip a "local" definition
            // so as to get to see global popular items in the mix.
            if(def.isLocal() && Rnd.fastRnd.nextBoolean())
                { continue; }

            try
                {
                final SimpleVariableValue svv = dsb.getVariable(def);
                if(svv == null) { continue; }

                // Specified variable must be of String type.
                assert(def.getType() == SimpleVariableDefinition.TYPE_STRING);

                final String s = (String) svv.getValue();
                if(s == null) { continue; }

                // Skip any null values.
                if(null == s) { continue; }

                final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
                final String fullName = aep.aeid.getFullName(s);

                // If this is in the exclusion list then skip it...
                if((excludeFullNames != null) && excludeFullNames.contains(fullName)) { continue; }

                // Seems not to be a valid/extant exhibit, so give up...
                if(fullName == null) { continue; }
                final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(fullName);
                if(esa == null) { continue; }

                // If the type was specified and does not match, give up!
                final ExhibitTypeParameters actualType = (ExhibitMIME.getInputFileType(esa.filePath));
                if(actualType == null)
                    { continue; /* Reject untyped exhibit. */ }
                if((type != null) && !type.equals(actualType))
                    { continue; /* Reject wrong-type exhibit. */ }

                // Skip if this exhibit may be "sensitive" somehow.
                final GenProps gp = dsb.getGenProps(-1);
                if(GenUtils.isSensitive(fullName, aep, gp)) { continue; }

                // If this exhibit type supports thumbnails
                // then reject anything without both immediately available.
                if(actualType.canPossiblyCreateThumbnailOfSameMIMEType())
                    {
                    final ExhibitThumbnails thumbnails = dsb.getThumbnails(fullName, false);
                    if((thumbnails == null) ||
                       (thumbnails.getSmall() == null) ||
                       (thumbnails.getStandard() == null))
                        { continue; /* Reject this. */ }
                    }

                // Reject anything without an above-par (positive) rating.
                // We allow a stale result so as to be quick.
                final ExhibitPropsComputableMutable ePCM =
                    aep.getExhibitPropsComputableMutable(fullName, true, gp, dsb, dsb.getScorerCache());
                if((ePCM != null) && (ePCM.getGoodness() <= 0)) { continue; }

                return(fullName); // Got it!
                }
            catch(final IOException e)
                {
                // Silently ignore this probably-transient problem...
                }
            }

        return(null); // Nothing found...
        }

    /**If true then check for bots/spiders by UA (User-Agent); note that clients can forge their UA. */
    private static final boolean CHECK_FOR_SPIDERS_BY_UA = true;

    /**Immutable Set of known spider/bot UA strings; should probably be moved to a text/properties file.
     * This is a set of lower-cased first (space/tab/bracket-delimited) words
     * from the UA strings.
     * <p>
     * The names consist only of non-regex-metacharacters in the set [a-z'_-],
     * so are safe to embed in a regex.
     */
    @SuppressWarnings("unchecked")
    private static final Set<String> spiderUAName1stWordsLC = (!CHECK_FOR_SPIDERS_BY_UA) ? ((Set<String>)(Object)Collections.emptySet()) :
        Collections.unmodifiableSet(new HashSet<String>(Arrays.asList(new String[]{
        "", /* Empty UA string... */
        "-",
        "alexibot",
        "appie",
        "aqua_products",
        "asterias",
        "b2w",
        "baiduspider",
        "backdoorbot",
        "becomebot",
        "blowfish",
        "bookmark",
        "botalot",
        "builtbottough",
        "bullseye",
        "bunnyslippers",
        "cheesebot",
        "cherrypicker",
        "cherrypickerelite",
        "cherrypickerse",
        "copernic",
        "copyrightcheck",
        "cosmos",
        "crescent",
        "curl",
        "dittospyder",
        "dumbot",
        "emailcollector",
        "emailsiphon",
        "emailwolf",
        "enterprise_search",
        "erocrawler",
        "extractorpro",
        "fairad",
        "faxobot",
        "findlinks",
        "flaming",
        "foobot",
        "freefind",
        "gaisbot",
        "getright",
        "gigabot",
        "googlebot-image",
        "grub",
        "grub-client",
        "harvest",
        "hatena",
        "hloader",
        "http",
        "httplib",
        "humanlinks",
        "ia_archiver",
        "infonavirobot",
        "iron33",
        "jennybot",
        "jetbot",
        "kenjin",
        "keyword",
        "larbin",
        "lexibot",
        "libweb",
        "linkextractorpro",
        "linkscan",
        "linkwalker",
        "lnspiderguy",
        "looksmart",
        "lwp-trivial",
        "lynx",
        "mata",
        "miixpc",
        "mister",
        "moget",
        "msiecrawler",
        "msnbot",
        "naver",
        "netants",
        "netmechanic",
        "nicerspro",
        "nutch",
        "offline",
        "omniexplorer_bot",
        "openbot",
        "openfind",
        "oracle",
        "perman",
        "port",
        "propowerbot",
        "prowebwalker",
        "psbot",
        "python-urllib",
        "queryn",
        "radiation",
        "repomonkey",
        "rma",
        "searchpreview",
        "sitesnagger",
        "sootle",
        "spankbot",
        "spanner",
        "stanford",
        "suzuran",
        "szukacz",
        "teleport",
        "teleportpro",
        "telesoft",
        "thenomad",
        "tocrawl",
        "true_robot",
        "turingos",
        "url",
        "url_spider_pro",
        "urly",
        "vci",
        "wbdbot",
        "webauto",
        "webbandit",
        "webcopier",
        "webenhancer",
        "websauger",
        "website",
        "webster",
        "webstripper",
        "webvac",
        "webzip",
        "wget",
        "www-collector-e",
        "xenu's",
        "yahooseeker",
        "zeus",
        })));

    /**Set of characters taken as a main-part terminator in a User-Agent header, including whitespace.
     * This should be usable as the separator arg to StringTokenizer,
     * and in a regex when wrapped in "[]" square brackets.
     */
    private static final String UA_TERMINATOR_CHARS = " \t/(:";

    /**Extra case-insensitive patterns matched in UA names, "|"-separated, or null if none. */
    private static final String UA_BOT_PATTERNS = "([a-z0-9]*bot)|([a-z0-9]*spider[a-z0-9]*)";

    /**Case-insenstive regex match for all non-empty UA names from spiderUAName1stWordsLC; null if not checking UA names.
     * Made public to enable some unit testing.
     */
    public static final Pattern UA_REGEX;
    /**Initialise UA_REGEX. */
    static
        {
        if(!CHECK_FOR_SPIDERS_BY_UA) { UA_REGEX = null; }
        else
            {
            final StringBuilder sb = new StringBuilder(32 + (spiderUAName1stWordsLC.size()<<4));
            // First build the simple compound regex name1|name2|...
            // possibly with some additional common patterns.
            if(null != UA_BOT_PATTERNS) { sb.append(UA_BOT_PATTERNS); }
            for(final String s : spiderUAName1stWordsLC)
                {
                if((s == null) || (s.length() == 0)) { continue; }
                if(sb.length() > 0) { sb.append('|'); }
                sb.append(s);
                }
            // Now wrap and add terminator...
            sb.insert(0, "^(");
            sb.append(")[").append(UA_TERMINATOR_CHARS).append("]?.*$");
            // Now compile the regex...
            UA_REGEX = Pattern.compile(sb.toString(), Pattern.CASE_INSENSITIVE);
            }
        }

    /**Name of Boolean attribute in request we cache result of requestProbablyFromSpider() by. */
    private static final String _rPFS_CACHE_PNAME = "org.hd.pg2k._rPFS_CACHE";

    /**LRU cache from (common) whole UAs to "bot"ness to save some repeated/slow String parsing; never null when checking for bots by UA. */
    private static final MemoryTools.SimpleLRUMap<String,Boolean> _isBot_cache = !CHECK_FOR_SPIDERS_BY_UA ? null :
        new MemoryTools.SimpleLRUMap<String,Boolean>(10001);

    /**Attempt to determine quickly if the current request is probably from a spider/bot (ie not a human).
     * NB: This does not atempt to distinguish between good spiders (ie bond fide SEs)
     * and bad/broken bots/scrapers.
     * <p>
     * This tries to distinguish human from non-human,
     * at least in part to conserve (interactive) resources for humans,
     * and does not claim to be perfect.
     * <p>
     * This returns true if there is no (valid) referring page,
     * though some browsers and firewalls may routinely block this info
     * for live humans too.
     * <p>
     * This should be quick as most or all requests may be tested with this.
     * <p>
     * TODO: Should expand this to test against well-known IP addresses.
     * <p>
     * TODO: Should include a behavioural element, eg mindless, pauseless browsing.
     *
     * @param request  the incoming request; never null
     */
    public static boolean requestProbablyFromSpider(final HttpServletRequest request)
        {
        // If we already worked this out, return result already computed/cached!
        final Object cachedResult = request.getAttribute(_rPFS_CACHE_PNAME);
        if(cachedResult instanceof Boolean)
            { return(((Boolean) cachedResult).booleanValue()); }

        // First check for lack of a valid "Referer" header...
        if("".equals(requestProbablyReferredFromExternalSite(request)))
            {
if(IsDebug.isDebug) { System.out.println("[Client has no/unparsable Referer: probably a spider/robot.]"); }
            request.setAttribute(_rPFS_CACHE_PNAME, Boolean.TRUE);
            return(true);
            }

        // Now check for a known bot User-Agent...
        if(CHECK_FOR_SPIDERS_BY_UA)
            {
            final Enumeration aeEn = request.getHeaders("User-Agent");
            final boolean hasNoUA = (aeEn == null) || !aeEn.hasMoreElements();
            if(hasNoUA)
                {
if(IsDebug.isDebug) { System.out.println("[Client has no UA: assumed to be a spider/robot.]"); }
                request.setAttribute(_rPFS_CACHE_PNAME, Boolean.TRUE);
                return(true); /* Rude/suspicious. */
                }
            else
                {
                // Check all such UA headers if more than one for some reason...
                while(aeEn.hasMoreElements())
                    {
                    final String wholeUA = (String) aeEn.nextElement();
   
                    final int lenUA = wholeUA.length();
                    if((lenUA < 2) || (lenUA > 512))
                        {
                        // Strange-sized UA is suspicious...
                        // And we don't want our cache/memory full of huge UA strings.
if(IsDebug.isDebug) { System.out.println("[Client has empty/tiny/huge (ie rude/broken/silly/dangerous) UA: assumed to be a spider/robot.]"); }
                        request.setAttribute(_rPFS_CACHE_PNAME, Boolean.TRUE);
                        return(true);
                        }

//final long t1 = System.nanoTime();
                    final boolean isBot;
                    final Boolean b = _isBot_cache.get(wholeUA);
                    if(b != null) { isBot = b.booleanValue(); }
                    else
                        {
                        isBot = UA_REGEX.matcher(wholeUA).matches();
                        // Cache match result for this whole User-Agent value.
                        _isBot_cache.put(wholeUA, isBot ? Boolean.TRUE : Boolean.FALSE);

                        // When we see a (new-ish) bot UA then log it and where it came from...
                        if(isBot) { System.out.println("[INFO: Probable bot/spider UA seen (full UA='"+TextUtils.sanitiseForXML(wholeUA, 256, false)+"'); client IP: "+request.getRemoteAddr()+".]"); }
                        }
//final long t2 = System.nanoTime(); System.out.println("[UA match time: "+(t2-t1)+"ns]");

                    if(isBot)
                        {
                        request.setAttribute(_rPFS_CACHE_PNAME, Boolean.TRUE);
                        return(true);
                        }
                    }
                }
            }

        // TODO: check known spider/bot IP addresses...

        request.setAttribute(_rPFS_CACHE_PNAME, Boolean.FALSE);
        return(false); // Probably human...
        }

    /**Attempt to detect if the current request has been referred from an external Web site.
     * This simply checks if the "Referer" is apparently from any of our
     * sites or their aliases.
     * <p>
     * Note that since this info can be forged,
     * or disabled/knobbled for security reasons,
     * this is not completely reliable.
     * <p>
     * A missing "Referer" will be taken as indicating an "external" reference,
     * and will catch most well-behaved spiders as a result.
     * <p>
     * Since we may have to do some string parsing this may not be very fast,
     * but it should not be very slow either.
     * <p>
     * Even if there is more than one "Referer" header we will only look at one.
     *
     * @param request  the incoming request; never null
     *
     * @return  null if referred from one of our sites/hosts,
     *          "" if unparsable or no referring URL,
     *          else normalised (lowercase, stripped of common prefixes, etc)
     *          referring host's name
     */
    public static String requestProbablyReferredFromExternalSite(final HttpServletRequest request)
        {
        final Enumeration rEn = request.getHeaders("Referer");
        final boolean noRef = ((rEn == null) || !(rEn.hasMoreElements())); // No referer...
        if(noRef) { return(""); }

        // Get the referring URL...
        final String ref = (String) rEn.nextElement();

        try
            {
            // Attempt to parse it for the hostname.
            // (If we fail (ie ref is ill-formed) then count this as "external".)
            final URL url = new URL(ref);

            // If we can't find virtual host details for the host
            // then assume that this is an external host.
            final String rawHostName = url.getHost();
            final String normalisedHostName = HostUtils.normaliseVirtualHostName(rawHostName);
            if(null == VirtualHosts.getVirtualHostDetails(normalisedHostName, null))
                { return(normalisedHostName); } // OK, got the referring host, and it's not us.
            return(null); // This was apparently referred by us, so is OK.
            }
        catch(final Exception e)
            {
            return(""); /* Cannot parse referrer, so treat as if external. */
            }
        }

    /**Generate full URL for exhibit given exhibit name; never null.
     * This may take account of such factors as which servers
     * are up, how heavily loaded, etc,
     * in order to perform automatic load balancing
     * and give the best possible user experience.
     * <p>
     * In order to do this well we may need the request details,
     * so they should be supplied if possible.
     * These should be the client's request to a Gallery site,
     * else null.
     */
    public static URL makeExhibitURL(final String exhibitName,
                                     final HttpServletRequest request,
                                     final DataSourceBean vars)
        throws MalformedURLException
        {
        return(new URL("http://" +
                       chooseMirrorHostToBalanceLoad(request, vars) +
                       makeExhibitRRURL(exhibitName)));
        }

    /**Generate root-relative URL for exhibit given exhibit name; never null.
     */
    public static String makeExhibitRRURL(final String exhibitName)
        //throws MalformedURLException
        {
        return("/" + WebConsts.BASE_PATH_EXHIBITS + "/" + exhibitName);
        }

    /**Generate full URL for thumbnail/sample given exhibit name and standard/small selector; never null.
     */
    public static URL makeThumbnailURL(final String exhibitName, final boolean std)
        throws MalformedURLException
        {
        return(new URL("http", CoreConsts.MAIN_DATA_HOST, makeThumbnailRRURL(exhibitName, std)));
        }

    /**Generate root-relative URL for thumbnail/sample given exhibit name and standard/small selector; never null.
     */
    public static String makeThumbnailRRURL(final String exhibitName, final boolean std)
        //throws MalformedURLException
        {
        return("/" + WebConsts.BASE_PATH_TN + "/" + (std ? WebConsts.PATH_TN_STD : WebConsts.PATH_TN_SML) + "/" + exhibitName);
        }

    /**Generate full URL for catalogue page given exhibit name; never null.
     * This should always generate the canonical/"official" form of the URL,
     * eg not including mirrors or aliases.
     * <p>
     * The media-type suffix (eg ".html" or ".wml") must be supplied.
     */
    public static URL makeCatPageURL(final String exhibitName, final String mediaTypeSuffix)
        throws MalformedURLException
        {
        return(new URL("http", CoreConsts.MAIN_DATA_HOST, makeCatPageRRURL(exhibitName, mediaTypeSuffix)));
        }

    /**Generate root-relative URL for catalogue page given exhibit name; never null.
     * This should take account of such factors as which servers
     * are up, how heavily loaded, etc.
     * <p>
     * The media-type suffix (eg ".html" or ".wml") must be supplied.
     */
    public static String makeCatPageRRURL(final String exhibitName, final String mediaTypeSuffix)
        //throws MalformedURLException
        {
        return("/" + WebConsts.BASE_PATH_CATPAGE + "/" + exhibitName + mediaTypeSuffix);
        }

    /**Returns true if this Web server is overloaded (eg for bandwidth or CPU).
     * If this server is overloaded then this routine returns true.
     * <p>
     * If not overloaded or not possible to tell, this returns false.
     * <p>
     * If the context is null then this routine may have to return false.
     * <p>
     * This may gather information from a number of sources,
     * but in the main the ThroughputFilter's data will be used.
     * We may explicitly set the overload flag here
     * if we detect the system to be overloaded
     * to make it quicker for us and others on a subsequent call;
     * this will get overwritten by TMF when it next gets to assess load.
     * <p>
     * If true then the UI and application should attempt to use less bandwidth
     * and CPU time than normal, perhaps by using cheaper algorithms than
     * usual (eg less comprehensive searches) or a less-graphics-intensive UI.
     * <p>
     * This may default to true while the server is starting up to ensure that
     * as little extra CPU (for example) as possible is consumed while
     * the server is compiling JSPs, loading classes, etc.
     * <p>
     * This is intended to be relatively cheap to call,
     * since it may get called frequently,
     * for example especially when the system is busy,
     * and/or at start-up before JIT compilation (eg HotSpot) has kicked in.
     */
    public static boolean isOverloaded(final ServletContext ctxt)
        {
        if(ctxt != null)
            {
            // If the ThroughputFilter thinks that we are overloaded,
            // then we report the system as overloaded.
            final Object overloadFlag =
                    ctxt.getAttribute(WebConsts.BANDWIDTH_OVERLOAD_ATTR_NAME);
            // We take the absense of any value as indication of a problem,
            // eg that the system may be too busy even to clear the flag!
            if((!(overloadFlag instanceof Boolean)) ||
                   ((Boolean) overloadFlag).booleanValue())
                { return(true); }

            // If the by-word index is not built/available/up-to-date
            // then regard the site as overloaded
            // and set the flag here for a faster check next time.
            final DataSourceBean dsb = DataSourceBean.getApplicationInstance(ctxt);
            if(!dsb.byWordIndexIsAvailableAndUpToDate())
                {
                // Potentially make the next call quicker
                // by cacheing the discovered busy status for a while
                // until TMF overwrites it on its next tick
                // by removing the "FALSE" value that we don't believe.
                ctxt.removeAttribute(WebConsts.BANDWIDTH_OVERLOAD_ATTR_NAME);
                return(true);
                }
            }

        // Site seems OK for CPU and bandwidth.
        return(false);
        }

    /**Returns true if the site seems to be lightly loaded (CPU, bandwidth, etc.
     * In case of difficulties/confusion this reports false.
     */
    public static boolean isLightlyLoaded(final ServletContext ctxt)
        {
        if(ctxt != null)
            {
            // If the ThroughputFilter doesn't think that we are lightly loaded,
            // then we report the system not lightly loaded immediately.
            final Object underloadFlag =
                    ctxt.getAttribute(WebConsts.BANDWIDTH_LIGHTLOAD_ATTR_NAME);
            // We take the absense of any value as indication of a problem,
            // eg that the system may be too busy even to set the flag!
            if((!(underloadFlag instanceof Boolean)) ||
                   (!((Boolean) underloadFlag).booleanValue()))
                { return(false); }

            // We double-check that some problem is not making us seem
            // overloaded at the same time!
            return(!isOverloaded(ctxt));
            }

        // Site status is unclear, so assume not quiet for now.
        return(false);
        }

    /**If true then exhibitHasThumbnail() will try to cache its results.
     * This should avoid us having to load thumbnails into memory
     * just to see if they exist or not.
     */
    public static final boolean TN_AVAIL_CACHE = true;

    /**The TN availability cache size, strictly positive.
     * Should be large enough that most thumbnail availablity requests
     * will be caught by it regardless of exhibit set size,
     * but small enough to limit memory requirements to something reasonable.
     * Should be *much* larger than the full set of thumbnails ever seen
     * on one page and/or that might be in our "popular" page set
     * so as to avoid thrashing fruitlessly.
     */
    private static final int TN_AVAIL_MIN_CACHE_SIZE = 131 +
        5*WebConsts.SINGLE_PAGE_CONTACT_SHEET_TN_COUNT +
        SystemVariables.MAX_DIFF_EXHIBIT_NAME_VALUES;

    /**Private key used by exhibitHasThumbnail(); never null. */
    private static final DataSourceBean.UnlinkedKey tnHTMLCacheKey = new DataSourceBean.UnlinkedKey("tnHTMLCacheKey");

    /**Test if the given exhibits have thumbnails (locally) available; never null.
     * This is suitable for a bulk check, eg before rendering a table,
     * and may be internally parallelised to overcome I/O latency, etc.
     * 
     * @param exhibitNames  non-null list of (String) valid full exhibit names;
     *     not altered by this routine and must not be changed by the caller
     *     while this routine is running
     * 
     * @return  a bit in the result set for each thumbnail that definitely
     *     exists in the requested size; an unset bit may represent "not known"
     */
    @SuppressWarnings("unchecked")
    public static BitSet exhibitsHaveThumbnail(final DataSourceBean dataSource,
                    final List<String> exhibitNames,
                    final boolean standard,
                    final boolean forceCreation)
        {
        final int n = exhibitNames.size();

        // Only try the parallel fetch if:
        //   * we have an extant cache of tn state.
        //   * there is more than one item to check.
        final MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Byte> tnExistenceMap;
        if(TN_AVAIL_CACHE && (n > 1) &&
           (null != (tnExistenceMap = (MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Byte>) dataSource.getUnlinkedValue(tnHTMLCacheKey))))
            {
            // Find all the exhibits whose thumbnail state is not cached at all,
            // and try to start one or more background thread(s) to check their status.
            // We expect this activity to be largely I/O bound.

            // Tasks to wait for the completion of... 
            final List<Future<?>> tasks = new LinkedList<Future<?>>();

            for(final String en : exhibitNames)
                {
                // If we have any tn state already cached then skip this entry...
                if(null != tnExistenceMap.get(en)) { continue; }
                
                // Try to fetch this missing entry into cache concurrently...
                tasks.add(ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                    public final void run()
                        { exhibitHasThumbnail(dataSource, en, standard, forceCreation); }
                    }));
                }
            
            // Wait for any tasks to complete.
            for(final Future<?> task : tasks)
                {
                try { task.get(); }
                catch(final Exception e) { e.printStackTrace(); /* Absorb but log any error. */ }
                }   
            }

        // Fetch all values through the cache using the caller's thread.
        final BitSet result = new BitSet(n);
        for(int i = n; --i >= 0; )
            {
            if(exhibitHasThumbnail(dataSource, exhibitNames.get(i), standard, forceCreation))
                { result.set(i); }
            }

        return(result);
        }
    
    /**Test if the given exhibit has a thumbnail (locally) available.
     * Does not try to force one to be made if one is not extant (or in cache)
     * unless the forceCreation argument is true.
     * <p>
     * Since this is likely to be testing what is in (local) cache,
     * its results can be considered at best a (good) hint.
     * <p>
     * We do not cache entirely negative answers
     * (ie that an exhibit has no thumbnails)
     * as this may change if we overcome (eg) a temporary resource issue.
     * But we assume the converse,
     * ie that once we have a thumbnail then it will always be available.
     * <p>
     * Always returns false for exhibit types that do not support thumbnails.
     *
     * @param dataSource  the data source; never null
     * @param exhibitName  the full exhibit name; must be valid (eg non-null)
     * @param standard  if true, tests for the presence of a standard-size
     * @param forceCreation  if true, we can try to force creation/fetch
     *     of the underlying thumbnail if not already available locally
     *
     * @return true if exhibit definitely has/had one/both thumbnails,
     *     false if unknown or thumbnails are not currently available
     */
    @SuppressWarnings("unchecked")
    public static boolean exhibitHasThumbnail(final DataSourceBean dataSource,
                                              final String exhibitName,
                                              final boolean standard,
                                              final boolean forceCreation)
        {
        if((dataSource == null) || (exhibitName == null))
            { throw new IllegalArgumentException(); }

        // Find any egregious misuse of this routine early when debugging!
        assert(ExhibitName.validNameSyntax(exhibitName));

        // If we are cacheing the availability of thumbnails
        // then ensure the cache map exists here.
        // This is a thread-safe size-limited LRU map.
        // Note that this is *not* linked to the AEP
        // so that we may retain a little stale information indefinitely,
        // especially though AEP changes when much has to be recomputed.
        MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Byte> tnExistenceMap;
        if(TN_AVAIL_CACHE)
            {
            while((tnExistenceMap = (MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Byte>) dataSource.getUnlinkedValue(tnHTMLCacheKey)) == null)
                {
                // Size the table to suit the exhibit set and allow growth...
                int nExhibits = 0;
                try { nExhibits = dataSource.getAllExhibitProperties(-1).aeid.length; } catch(final IOException e) { /* Ignore. */ }
                dataSource.putIfAbsentUnlinkedValue(tnHTMLCacheKey, new MemoryTools.SimpleLRUMapAutoSizeForHitRate<String, Byte>(0.0001f, TN_AVAIL_MIN_CACHE_SIZE, 100001+Math.max(109*TN_AVAIL_MIN_CACHE_SIZE, 2*nExhibits), 0.7f));
                }
            }

        // We store availability as a bit-mask
        //   * bit 0 is 1 iff the small thumbnail is available
        //   * bit 1 is 1 iff the std thumbnail is available
        byte availability = 0; // Assume none available by default.

        // Only use cached positive values.
        // A negative cached result has us check again...
        // We cache negative results mainly to maintain correct hit-rate stats.
        final Byte cachedAvailability;
        if(!TN_AVAIL_CACHE || (null == (cachedAvailability = tnExistenceMap.get(exhibitName))) || (cachedAvailability.byteValue() <= 0))
            {
            // Hauling the thumbnail pair into view can be very expensive,
            // eg may involve disc fetches or image decoding or worse,
            // and may simply displace other items better kept in cache/memory.
            final ExhibitThumbnails tns = dataSource.getThumbnails(exhibitName, forceCreation);

            if(tns != null)
                {
                if(tns.getSmall() != null)
                    { availability |= 1; }
                if(tns.getStandard() != null)
                    { availability |= 2; }
                }

            if(TN_AVAIL_CACHE)
                {
                // Byte.valueOf() avoids creating distinct instances.
                tnExistenceMap.put(exhibitName, Byte.valueOf(availability));
if(IsDebug.isDebug && ((tnExistenceMap.size() & 0x3ff) == 0)) { dataSource.log("    tnExistenceMap: " + tnExistenceMap.toString()); }
                }
            }
        // Use the cached (positive) value...
        else
            { availability = cachedAvailability.byteValue(); }

        final boolean tnAvailable = (0 != (availability & (standard ? 2 : 1)));
        return(tnAvailable);
        }

    /**Get thumbnail image URL to embed in HTML page (usually JPEG/GIF/PNG); null if none available.
     * By preference uses purpose-built thumbnail, else tries to
     * use image itself if small enough.
     * <p>
     * Has to be passed a dataSource and a full exhibit name.
     * <p>
     * Returns null if no suitable thumbnail URL can be generated.
     * <p>
     * This may cache its results,
     * in particular assuming that once a particular thumbnail becomes available
     * that it does not go away again.
     *
     * @param tnDim  (output argument) if not null and result is not null,
     *     is filled in with the thumbnail dimensions
     * @param dontCreateTn  if true, don't force creation of a thumbnail if
     *     one is not already readily available
     */
    public static String makeHTMLInlineImageThumbnailURL(final DataSourceBean dataSource,
                                                         final String exhibitName,
                                                         final boolean std,
                                                         final boolean relativeURL,
                                                         final java.awt.Dimension tnDim,
                                                         final boolean dontCreateTn)
        throws IOException
        {
        final AllExhibitProperties aep = dataSource.getAllExhibitProperties(-1);

        // Get full exhibit details...
        final ExhibitStaticAttr esa = aep.aeid.getStaticAttr(exhibitName);

        if(WebUtils.canInlineInHTMLPage((ExhibitMIME.getInputFileType(esa.filePath))))
            {
            // Get the exhibit immutable computable properties if possible.
            final ExhibitPropsComputable epc = aep.getExhibitPropsComputable(exhibitName);
            final java.awt.Dimension xyDim = (epc == null) ? null : epc.getXyDimensions();
            // Compute what thumbnail dimensions should be...
            final java.awt.Dimension thumbnailXyDim = (xyDim == null) ? null : ExhibitThumbnails.computeThumbnailDimensions(xyDim, std);

            // Is this in fact a 2D image?
            if(thumbnailXyDim != null)
                {
                // Pass dimensions back to caller.
                if(tnDim != null)
                    {
                    tnDim.width  = thumbnailXyDim.width;
                    tnDim.height = thumbnailXyDim.height;
                    }

                // Is this a small enough (in bytes) exhibit to be its own thumbnail?
                final boolean smallExhibit = (esa.length <=
                    (std ? ExhibitThumbnails.STD_ABS_MAX_BYTES : ExhibitThumbnails.SML_ABS_MAX_BYTES));
                // Can this be its own thumbnail?
                // It must be small enough in bytes,
                // and possibly small enough in XxY pixels too.
                final boolean canBeOwnThumbnail =
                    (smallExhibit && (WebConsts.ALLOW_BROWSER_IMAGE_SCALE || thumbnailXyDim.equals(xyDim)));

                // See if we have thumbnails available
                // (if eager, force one to be made,
                // else see if one already exists if we can't inline directly).
                // If we force them to be created
                // then we expect them to remain cached!
                // Shall we force creation?
                final boolean forceCreation = !dontCreateTn &&
                    (WebConsts.EAGER_TN_USE || !canBeOwnThumbnail);

                // Check for availability of real thumbnail.
                final boolean tnAvailable = exhibitHasThumbnail(dataSource,
                                                                exhibitName,
                                                                std,
                                                                forceCreation);

                final boolean canShowThumbnail = tnAvailable || canBeOwnThumbnail;

//System.err.println("[sE/cBOT/cST: "+smallExhibit+"/"+canBeOwnThumbnail+"/"+canShowThumbnail+".]");

                if(canShowThumbnail)
                    {
                    // Chose the (relative) URL to use
                    // (always serve from same host as this page).
                    // Show true thumbnail by preference...
                    if(relativeURL)
                        { return(tnAvailable ? WebUtils.makeThumbnailRRURL(exhibitName, std) : WebUtils.makeExhibitRRURL(exhibitName)); }
                    else
                        { return((tnAvailable ? WebUtils.makeThumbnailURL(exhibitName, std) : WebUtils.makeExhibitURL(exhibitName, null, dataSource)).toString()); }
                    }
                }
            }

        return(null); // Can't make a thumbnail.
        }

    /**Find the set of exhibits with the same subject as the indicated one.
     * Given a List of String exhibit names sorted by
     * ExhibitPropsGlobalImmutable.SMART_ORDER
     * (or possibly just by ExhibitPropsGlobalImmutable.SUBJECT_ORDER),
     * the SUBJECT_ORDER comparator,
     * and the index to a given item in that List,
     * finds all the items surrounding that have the same SUBJECT_ORDER,
     * ie that differ only in attribute words and are variants on the same
     * exhibit.  (The result will always contain the input item.)
     * <p>
     * The List must not contain nulls, all entries must be valid exhibit
     * names, and the List must be sorted implicitly the the comparator
     * passed in.  The input index must be within bounds.
     * <p>
     * This does not alter the List passed in.  The return value is
     * an unmodifiable sublist of the input.
     * <p>
     * This returns the sublist of variants on the indicated exhibit;
     * this will degenerate to just containing the input exhibit name if there
     * are no other exhibit names for the same exhibit.
     * <p>
     * This assumes that the number of variants of any given exhibit is
     * small, and so a linear search is used.
     * <p>
     * If the List does not support efficient bidirectional movement and seeks
     * then this routine will be very inefficient.
     */
    public static List<String> getExhibitVariantRange(final List<String> allExhibitNames,
                                                      final Comparator<String> comparator,
                                                      final int thisExhibitIndex)
        {
        final String thisExhibit = allExhibitNames.get(thisExhibitIndex);
        final int listSize = allExhibitNames.size();

        int first, last;

        // Search backwards for first matching item.
        for(first = thisExhibitIndex; first > 0; --first)
            {
            if(comparator.compare(thisExhibit, allExhibitNames.get(first-1)) != 0)
                { break; } // Stop when we hit an item with a different subject.
            }

        // Search forwards for last matching item.
        for(last = thisExhibitIndex; last < listSize-1; ++last)
            {
            if(comparator.compare(thisExhibit, allExhibitNames.get(last+1)) != 0)
                { break; } // Stop when we hit an item with a different subject.
            }

        return(Collections.unmodifiableList(allExhibitNames.subList(first, last+1)));
        }

    /**Computes a minimal human-readable unique prefix of name file component to distinguish given exhibit from most others.
     * Uniqueness is not guaranteed, but this is meant to give a reasonably
     * short result that a human can read in the title of a page, for example.
     * <p>
     * If inputs are bad then this will try to fail gently with an empty-string
     * result rather than an exception to make it robust if called directly from
     * JSP code, for example.
     * <p>
     * This is passed a List of exhibits sorted in a "smart" order that
     * sorts mainly on the file component of the name,
     * probably in a case-insensitive way.
     * <p>
     * This routine finds a short word prefix that (case-insensitively)
     * is different from neighbouring exhibits
     * and is thus (depending on the overall sort order)
     * probably the shortest unique prefix amongst the entire collection.
     * <p>
     * (If no unique value is possible, the entire prefix is returned.)
     * <p>
     * If the List of size 0 we return the empty string and do not use
     * the index parameter at all.  This simplifies use where the list
     * may transiently be empty during start-up.
     * <p>
     * This routine may run very slowly if the List argument does not support
     * efficient random seeking.
     * <p>
     * The result is designed to be used in headings and other display text.
     * <p>
     * TODO: fix inefficient double-parsing of main words...
     *
     * @param exhibits  sorted (smart-ish sorted) list of exhibit names
     *     (String value); must not change while routine is running
     * @param ourIndex  index (strictly positive) of the exhibit whose
     *     abbreviated name we wish to produce; must be within the List
     * @return "" in case of invalid aruguments,
     *     else some whole-word prefix of the name
     */
    public static String minimalUniqueENTitlePrefix(final List<String> exhibits,
                                                    final int ourIndex)
        {
        // Treat bad arguments gently as the outputs are likely to
        // be displayed in HTML directly.
        if(exhibits.size() == 0) { return(""); }
        if((ourIndex < 0) || (ourIndex >= exhibits.size())) { return(""); }

        // Set of attribute words we use for passing names.
        // In this case, deliberately empty.
        final Set<String> noAttrWords = Collections.emptySet();

        // "empty" enumeration we use for missing neighbours.
        final Enumeration emptyEn = Collections.enumeration(Collections.emptyList());

        // Get our name and previous/next names
        // as word enumerations (or empty enumerations if not present).
        // for this purpose we pretend that there are no attribute words...
        final ListIterator<String> liFwd = exhibits.listIterator(ourIndex);
        final String ourNameFull = liFwd.next();
        final String ourNameMainWords = ExhibitName.getMainWordsComponent(
                                        ourNameFull, noAttrWords);
        final Enumeration ourNameWords = ExhibitName.getMainWords(
                                        ourNameFull, noAttrWords);

        // Default to empty enumerations.
        Enumeration nextNameWords = emptyEn;
        Enumeration prevNameWords = emptyEn;

        // Now search forwards for different name to compare against...
        // Note that we implicitly start from just after ourName above.
        while(liFwd.hasNext())
            {
            final String n = liFwd.next();
            final String nextNameMainWords =
                ExhibitName.getMainWordsComponent(n, noAttrWords);
            if(!nextNameMainWords.equalsIgnoreCase(ourNameMainWords))
                {
                nextNameWords = ExhibitName.getMainWords(n, noAttrWords);
                break;
                }
            }

        // Now search backwards for different name to compare against...
        final ListIterator<String> liBack = exhibits.listIterator(ourIndex);
        while(liBack.hasPrevious())
            {
            final String n = liBack.previous();
            final String prevNameMainWords =
                ExhibitName.getMainWordsComponent(n, noAttrWords);
            if(!prevNameMainWords.equalsIgnoreCase(ourNameMainWords))
                {
                prevNameWords = ExhibitName.getMainWords(n, noAttrWords);
                break;
                }
            }

        // Result word prefix (preserves case, ends with word separator).
        final StringBuilder result = new StringBuilder();

        // Iterate until we run out of words or both neighbours have
        // (or have different words).
        while(ourNameWords.hasMoreElements())
            {
            final String ourNameWord = (String) ourNameWords.nextElement();
            result.append(ourNameWord).append(ExhibitName.WORD_SEP);

            // Take next word, if extant, on both prev and next sides.
            final String nextNameWord = nextNameWords.hasMoreElements() ?
                (String) nextNameWords.nextElement() : "";
            final String prevNameWord = prevNameWords.hasMoreElements() ?
                (String) prevNameWords.nextElement() : "";

            // If neighbour has another word but it is different to ours,
            // pretend neighbour was truncated at previous word.
            // We ignore case, since users probably will.
            if(nextNameWord.equalsIgnoreCase(ourNameWord))
                { continue; } // Looks the same so cannot stop yet.
            else
                { nextNameWords = emptyEn; } // No longer a contender.

            if(prevNameWord.equalsIgnoreCase(ourNameWord))
                { continue; } // Looks the same so cannot stop yet.
//            else
//                { prevNameWords = emptyEn; } // No longer a contender.

            break; // Done!
            }

        // TODO: To save heap space,
        // we could return a prefix from the original name.
        return(result.toString());
        }

   /**Get BufferedImage containing expanded image loaded as static resource from WAR; never null.
    * Loaded on first use and cached statically,
    * optionally via a SoftReference to allow the system to reclaim memory.
    * <p>
    * There may be an adverse effect on system behaviour if many large images
    * are cached in memory; this may be mitigated by storing them via
    * a SoftReference so that the memory can be recycled automatically.
    * <p>
    * This method is internally synchronized to allow only one image load/decode
    * to happen at once to conserve CPU and memory (and other) resources.
    * <p>
    * If the image is indexed and forceToARGBTrueColour is true then
    * we expand it to true-colour to make it possible to add markings.
    * <p>
    * This uses our built-in mediahandler classes to decode the image,
    * so the image type must be one that we have a decoder for.
    * <p>
    * This may not handle alpha correctly in all cases.
    * <p>
    * <strong>Beware:</strong> since a reference is returned rather than
    * a copy, be careful not to adjust the returned image unintentionally.
    *
    * @param context  gives context for WAR from which to load the raw
    *     base clickable-map image
    * @param resourceRRURL  root-relative URL (eg "/my/image.gif") of
    *     source image in WAR; must not be null and must be a type
    *     (and with a file extension) that we have a MIME mediahandler for
    * @param forceToARGBTrueColour  if true, and the source image is not
    *     ARGB true-colour, then we force conversion to ARGB true-colour
    *     before cacheing to make it easier to draw on the image
    * @param cacheViaSoftReference  if true then we attempt to cache the
    *     image via a SoftReference; if at any point this is true when the
    *     image needs to be (re)cached, the image remains permanently cached
    *     via a strong reference thereafter
    * @param copyResult  if true, we force the result to be a copy of
    *     the cached value to avoid accidentally changing the cached copy;
    *     this may force a change in colour model and/or discarding properties
    *
    * @return image, possibly a copy, possibly with a converted colour model
    *
    * @throws java.lang.IllegalStateException if the image is not loadable
    */
    public static BufferedImage getAndCacheStaticImage(
                                            final boolean copyResult,
                                            final String resourceRRURL,
                                            final boolean forceToARGBTrueColour,
                                            final ServletContext context,
                                            final boolean cacheViaSoftReference)
        throws IllegalStateException
        {
        if((context == null) ||
           (resourceRRURL == null))
            { throw new IllegalArgumentException(); }

        synchronized(_gACSI_cache)
            {
            // An item in the cache is one of:
            //   * null (ie completely absent)
            //   * BufferedImage
            //   * SoftReference (possibly cleared) to BufferedImage
            final Object rawItem = _gACSI_cache.get(resourceRRURL);
            BufferedImage result = (!(rawItem instanceof SoftReference))
                ? (BufferedImage) rawItem
                : (BufferedImage) (((SoftReference) rawItem).get());

            if(result == null) // Needs fetching.
                {
                final InputStream is =
                    context.getResourceAsStream(resourceRRURL);
                if(is == null)
                    { throw new IllegalStateException("cannot get InputSteam for image to cache from " + resourceRRURL); }

                // Find correct handler given name of file.
                final ExhibitMIME.ExhibitTypeParameters etp =
                    ExhibitMIME.getInputFileType(resourceRRURL);
                if(etp == null)
                    { throw new IllegalStateException("cannot get MIME type for image to cache from " + resourceRRURL); }
                if(etp.handler == null)
                    { throw new IllegalStateException("cannot get handler for image to cache from " + resourceRRURL); }

                try {
                    // Decode the image.
                    result = etp.handler.decodeImage(is);
                    if(result == null)
                        { throw new IllegalStateException("cannot get decode image to cache from " + resourceRRURL); }

                    // Force to ARGB if required.
                    if(forceToARGBTrueColour)
                        { result = ImageUtils.convertToTrueColourARGB(result, false); }

                    // Cache the result.
                    if(cacheViaSoftReference)
                        { _gACSI_cache.put(resourceRRURL, new SoftReference<BufferedImage>(result)); }
                    else
                        { _gACSI_cache.put(resourceRRURL, result); }
                    }
                catch(final IOException e)
                    { throw new IllegalStateException("cannot decode image to cache from " + resourceRRURL + ": IOException: " + e.getMessage()); }
                }

            // If forced to copy the result, do so.
            // TODO: Maybe try to preserve relevant image properties?
            if(copyResult)
                {
                // Coerce data into original colour model.
                // Discard any properties of the original.
                final ColorModel cm = ImageUtils.extractColorModelOrRGB(result);
                final int width = result.getWidth();
                final int height = result.getHeight();
                final WritableRaster raster = cm.createCompatibleWritableRaster(
                    width, height);
                final BufferedImage copiedResult =
                    new BufferedImage(cm, raster, false, null);

                // Actually copy the pixels...
                copiedResult.setRGB(0, 0, width, height,
                    result.getRGB(0, 0, width, height, null, 0, width),
                                    0, width);

                return(copiedResult);
                }

            return(result);
            }
        }

    /**Private cache for getAndCacheStaticImage().
     * Is a map from root-relative URL to a BufferedImage
     * or SoftReference to a BufferedImage.
     * <p>
     * A Hashtable is used for its inherent thread safety,
     * and all getAndCacheStaticImage() activity is synchronized on this too.
     */
    private static final Hashtable<String,Object> _gACSI_cache = new Hashtable<String,Object>();


    /**Generate a unique key for the given HTTP request; returns null if not possible to generate.
     * For example, we use this to help limit the number of times that
     * a given user is asked to vote.
     * <p>
     * Generates a string starting with the given prefix and some
     * leading portion of the client's IP address...
     * <p>
     * It is not fatal if this conflates users,
     * but it is more useful that this never thinks one user is more than one
     * to avoid pestering them too much
     * (or letting a spider inject too much noise for example).
     *
     * @param prefix  unique prefix to the generated key; non-empty, non-null
     * @param request  the user's request
     */
    public static String generateUserDataPointID(final String prefix,
                                                 final HttpServletRequest request)
        {
        if((prefix == null) || (prefix.length() == 0)) { return(null); }
        if(request == null) { return(null); }

        final StringBuilder sb = new StringBuilder(prefix.length() + 21);
        sb.append(prefix);

        try
            {
            // Get the IP address
            final InetAddress ia = InetAddress.getByName(request.getRemoteAddr());

            // Use all but the final octet to construct the data-point ID.
            // Assume that this will distinguish most genuine users.
            final byte[] addr = ia.getAddress();
            for(int i = 0; i < addr.length-1; ++i)
                { sb.append(addr[i] & 0xff).append('.'); }

            return(sb.toString());
            }
        catch(UnknownHostException e)
            {
            return(null); // Could not understand the client's address.
            }
        }

    /**Name of additional parameter to record if this is a vote "for" ("pro").
     * Value of the parameter is to be "true" or "false".
     */
    public static final String VOTE_PRO_PARAM_NAME = "pro";

    /**Name of additional parameter to record if this is a vote "against" ("con").
     * Value of the parameter is to be "true" or "false".
     */
    public static final String VOTE_CON_PARAM_NAME = "con";

    /**Create and post the handler for a vote if possible and returns the unique listenerID.
     * This replaces any existing listener for this voter.
     * <p>
     * This returns null if we could not create a listener.
     * <p>
     * If the handler is invoked,
     * then this registers a dummy (inactive) handler
     * to delay the next request to the user to vote.
     *
     * @param exhibitName  valid exhibit name; never null
     * @param request  client's HTTP request; never null
     * @param vars  where stats updates are posted; never null
     *
     * @return listenerID, or null if one could not be generated
     */
    public static String createAndFileVoteListener(final String exhibitName,
                                                   final HttpServletRequest request,
                                                   final SimpleVariablePipelineIF vars)
        {
        if(!ExhibitName.validNameSyntax(exhibitName)) { return(null); }
        if(vars == null) { return(null); }

        final String dpID = generateUserDataPointID(WebConsts.VOTER_DATA_POINT_PREFIX, request);
        if(dpID == null) { return(null); }

        // Compute expiry time (with a random element).
        final long expireBy = System.currentTimeMillis() + WebConsts.VOTE_MIN_REQUEST_GAP_MS +
                Rnd.fastRnd.nextInt(WebConsts.VOTE_MIN_REQUEST_GAP_MS);

        // Capture the visitor's IP address for later (in case they vote).
        InetAddress va = null;
        try { va = InetAddress.getByName(request.getRemoteAddr()); }
        catch(final UnknownHostException e) { } // Ignore errors...
        final InetAddress voterIPAddr = va;

        // Create new listener.
        final StatsSink.AbstractStatsListener newAsl =
                new StatsSink.AbstractStatsListener(dpID, expireBy)
            {
            public final String handle(final Map<String, String[]> parameters)
                {
                if(parameters != null)
                    {
                    // Allow for GET or POST plain/image style (.x, .y) values.
                    final boolean votePro =
                        (null != parameters.get(VOTE_PRO_PARAM_NAME)) ||
                        (null != parameters.get(VOTE_PRO_PARAM_NAME + ".x"));
                    final boolean voteCon =
                        (null != parameters.get(VOTE_CON_PARAM_NAME)) ||
                        (null != parameters.get(VOTE_CON_PARAM_NAME + ".x"));
                    if(votePro != voteCon) // Exactly one selected...
                        {
                        try
                            {
                            // Post vote event value to correct event stream...
if(IsDebug.isDebug) { System.out.println("***VOTE registered (for="+votePro+"): " + exhibitName + " by " + dpID); }
                            vars.setVariable(new SimpleVariableValue(
                                    (votePro ? SystemVariables.VOTE_PRO : SystemVariables.VOTE_CON),
                                    ExhibitName.getFileComponent(exhibitName)));

                            if(voterIPAddr != null)
                                {
                                // Note location of voter (as ccTLD or region).
                                // We are prepared to spend some time on this
                                // as voting is rare and significant
                                // and we've probably already collected/cached
                                // any required data.
                                final String location = GeoUtils.getRegionByAddress(voterIPAddr, false);
                                vars.setVariable(new SimpleVariableValue(
                                    VOTER_LOCATION_STATS_EVENT_DEF,
                                        VOTER_LOCATION_STATS_EVENT_PREFIX + location));
                                }
                            }
                        catch(final IOException e)
                            { e.printStackTrace();  /* Just absorb errors. */ }
                        }
                    }

                // Compute time before next vote will be requested
                // if a user does take this opportunity to vote.
                // A relatively long time after we expect the original voting opportunity to expire
                // to act as a second-level screen against spiders
                // and to avoid pestering a human voter too often.
                // (We also make this time more unpredictable with a good random source.)
                final long nextVote = expireBy + 11*WebConsts.VOTE_MIN_REQUEST_GAP_MS +
                        Rnd.goodRnd.nextInt(7*WebConsts.VOTE_MIN_REQUEST_GAP_MS);

                // Install a dummy handler to postpone the next time that this user gets asked to vote.
                // Assumes that the extant listener is removed *before* a call to handle()
                // so that this new listener will not be removed on return.
                StatsSink.addListenerForDataPoint(new StatsSink.AbstractStatsListener(dpID, nextVote){
                    /**Dummy handler that should never in fact be invoked. */
                    public final String handle(final Map<String, String[]> parameters)
                        { return(WebUtils.makeCatPageRRURL(exhibitName, WebConsts.F_secondary_generated_HTML_suffix)); }
                    });

                // Redirect back to the exhibit page...
                // Stick a random parameter value on the end to ensure that
                // all common/broken browsers (eg IE6, FF1) reload the page.
                return(WebUtils.makeCatPageRRURL(exhibitName, WebConsts.F_secondary_generated_HTML_suffix) + "?rnd="+(Rnd.fastRnd.nextLong() >>> 1));
                }
            };

        // Get the listener ID...
        final String listnerID = newAsl.uniqueListenerID;

        // File the request!
        StatsSink.addListenerForDataPoint(newAsl);

        return(listnerID);
        }


    /**Get sorted, hyperlinked HTML i18n text list of exhibit categories; never null.
     * The list is sorted:
     * <ol>
     * <li>By dictionary-order i18n localised title.
     * </ol>
     * <p>
     * We may flag categories as good or bad if significantly so.
     *
     * @param asList  if true, entries are preceded by &lt;li&gt;
     *     else they are followed by &lt;br /&gt;
     */
    public static final String getCategoryListSortedAsHTML(final DataSourceBean dsb,
                                                           final LocaleBeanBase localeBean,
                                                           final boolean asList)
        throws IOException
        {
        if((dsb == null) || (localeBean == null))
            { throw new IllegalArgumentException(); }

        // Private data on each category, for sorting.
        final class PerCat implements Comparable<PerCat>
            {
            PerCat(final String cat,
                   final String i18nTitle,
                   final Boolean isGood,
                   final int entriesInCat)
                {
                category = cat;
                title = i18nTitle;
                good = isGood;
                numEntries = entriesInCat;
                }

            final String category;
            final String title;
            final Boolean good;
            final int numEntries;

            /**Compares this object with the specified object for order.
             * Order is:
             * <ol>
             * <li>By dictionary-order i18n localised title.
             * </ol>
             */
            public int compareTo(final PerCat other)
                {
                // Sort by localised title, case insensitively.
                final int tComp = String.CASE_INSENSITIVE_ORDER.compare(title, other.title);
                if(tComp != 0) { return(tComp); }

                // Officially break ties by underlying canonical category name.
                // This should never really be needed.
                return(category.compareTo(other.category));
                }
            }

        final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);

        // Get the list of categories.
        final Map<String,Integer> categories = aep.getCategoryExhibitCounts();

        // Create sorted set of details.
        // We don't expend huge effort on getting the category good/bad status.
        final SortedSet<PerCat> cs = new TreeSet<PerCat>();
        for(final String cat : categories.keySet())
            {
            cs.add(new PerCat(
                cat,
                GenUtils.computeSectionTitle(cat, localeBean),
                aep.isCategoryGood(cat, dsb, false),
                categories.get(cat)
                ));
            }

        // Convert to nicely-formatted HTML.
        final StringBuilder sb = new StringBuilder(79 * categories.size());
        for(final PerCat pc : cs)
            {
            if(asList) { sb.append("<li>"); }

            // If category significantly good/bad then mark it so.
            if(pc.good != null)
                {
                final String proIcon = "smile.gif";
                final String conIcon = "frown.gif";
                sb.append("<img src=\"/_static/icon/").
                    append(pc.good.booleanValue() ? proIcon : conIcon).
                    append("\" width=15 height=15> ");
                }

            // Open link to category RRURL.
            sb.append("<a href=\"").
                append(WebConsts.VIRTUAL_COLLECTIONS_BYCATEGORY_ROOT).
                append(pc.category).
                append("/\">");
            sb.append(pc.title);
            // Close link to category.
            sb.append("</a>");

            // Insert exhibit count.
            sb.append(" <i>(").append(pc.numEntries).append(")</i>");

            // Finish the point
            if(!asList) { sb.append("<br />"); }

            // Finish the line for readability.
            sb.append("\r\n");
            }

        return(sb.toString());
        }

    /**Returns true if this seems to be a slave disconnected from the master.
     * We may, for example, not wish to collect votes from users
     * if they may get discarded without getting to the master.
     */
    public static final boolean isDisconnectedSlave(final DataSourceBean dsb)
        {
        if(dsb == null) { throw new IllegalArgumentException(); }

        // If definitely a master then we can return false immediately.
        if(Boolean.FALSE.equals(dsb.isSlave())) { return(false); }

        // We decide that we are NOT disconnected
        // if we can see at least one other system via the system variables,
        // ie at least two systems in total.
        // We use the availablity of client-count as a measure of connectivity.
        final SimpleVarStats stats = VarTools.generateSimpleStats(dsb,
                                SystemVariables.ThroughputMonitorFilter_CLIENT_COUNT,
                                0); /* Minimum possible life. */
        return((stats == null) || (stats.getSystemCount() < 2));
        }

    /**Returns true if this request is apparently a precacheing request, eg from a "Web accelerator".
     * This is true if a client (such as FireFox) is "reading ahead"
     * but it may be the case that no real human gets to see the content.
     */
    public static boolean isPrecacheRequest(final HttpServletRequest request)
        {
        // "X-Moz: precache" header covers FF1 and Google's Web Accelerator.
        return("precache".equalsIgnoreCase(request.getHeader("X-Moz")));
        }


    /**Private key used by getCatPageExhibitMetaDataHTML(); never null. */
    private static final AEPLinkedKey metadataCacheKey = new AEPLinkedKey("metadataCacheKey");

    /**Static dictionary used by getCatPageExhibitMetaDataHTML() for compression of in-memory data; never null.
     * The static dictionary content should be reviewed after any major format changes,
     * though this is not a correctness issue, only a matter of compression efficiency.
     */
    public static final Compact7BitString.StaticDictionary sDictMD = new Compact7BitString.StaticDictionary("getCatPageExhibitMetaDataHTML",
        Arrays.asList(new String[]{
            "JPEG",    /* MANUALLY ADDED: count=16817, saving=50451, meanFirstPos=126 */
            "javax_imageio_1", /* count=17761, saving=248654, meanFirstPos=37 */
            "ColorSpaceType", /* count=17752, saving=230776, meanFirstPos=67 */
            "Compression", /* count=17684, saving=176840, meanFirstPos=106 */
            "NumChannels", /* count=16942, saving=169420, meanFirstPos=86 */
            "metadata", /* count=17939, saving=125573, meanFirstPos=9 */
            "TypeName", /* count=16826, saving=117782, meanFirstPos=120 */
            "compact", /* count=18684, saving=112104, meanFirstPos=3 */
            "Chroma", /* count=17755, saving=88775, meanFirstPos=53 */
            "value", /* count=18684, saving=74736, meanFirstPos=87 */
            "image", /* count=16951, saving=67804, meanFirstPos=23 */
            "YCbCr", /* count=15240, saving=60960, meanFirstPos=73 */
            "name", /* count=17751, saving=53253, meanFirstPos=71 */
            "\"</", /* count=18684, saving=37368, meanFirstPos=72 */
            "></", /* count=17805, saving=35610, meanFirstPos=95 */
            "</", /* count=18684, saving=18684, meanFirstPos=10 */
            "=\"", /* count=18684, saving=18684, meanFirstPos=70 */
            "> ", /* count=18684, saving=18684, meanFirstPos=68 */
            "><", /* count=18684, saving=18684, meanFirstPos=4 */
            "dd", /* count=18684, saving=18684, meanFirstPos=13 */
            "dl", /* count=18684, saving=18684, meanFirstPos=1 */
            "dt", /* count=18684, saving=18684, meanFirstPos=5 */
            "BackgroundIndex", /* count=808, saving=11312, meanFirstPos=103 */
            "BlackIsZero", /* count=934, saving=9340, meanFirstPos=89 */
            "accessionData", /* count=745, saving=8940, meanFirstPos=9 */
            "stream", /* count=984, saving=4920, meanFirstPos=23 */
            "GRAY", /* count=1157, saving=3471, meanFirstPos=73 */
            "CRC32", /* count=745, saving=2980, meanFirstPos=61 */
            "RGB", /* count=1353, saving=2706, meanFirstPos=73 */
            "TRUE", /* count=837, saving=2511, meanFirstPos=91 */
            "date", /* count=745, saving=2235, meanFirstPos=23 */
            "hash", /* count=745, saving=2235, meanFirstPos=59 */
            "size", /* count=745, saving=2235, meanFirstPos=41 */
            "sampleRate", /* count=174, saving=1566, meanFirstPos=88 */
            "MD5", /* count=745, saving=1490, meanFirstPos=82 */
            "encoding", /* count=174, saving=1218, meanFirstPos=70 */
            "frames", /* count=162, saving=810, meanFirstPos=37 */
            "audio", /* count=178, saving=712, meanFirstPos=56 */
            "57024", /* count=173, saving=692, meanFirstPos=47 */
            "PCM_SIGNED", /* count=55, saving=495, meanFirstPos=78 */
            "ULAW", /* count=112, saving=336, meanFirstPos=75 */
            "PCM_UNSIGNED", /* count=7, saving=77, meanFirstPos=77 */
            "BI_RGB", /* count=6, saving=30, meanFirstPos=74 */
            "1136915", /* MANUALLY ADDED: count ~ 10 as prefix. */
            }));

    /**Generates HTML form of exhibit metadata, "" if no such metadata for the specified exhibit; never null.
     * The result is keyed to both the DataSourceBean and the exhibitName.
     * <p>
     * Cached values are discarded when the AEP changes.
     * <p>
     * The computed HTML is locale-invariant, which is why cacheing is viable.
     * <p>
     * The toString() method should be called on the result to get the String HTML text,
     * which may be a String or some other internal representation.
     *
     * @param dsb  valid non-null DataSourceBean
     * @param exhibitName  valid non-null exhibit
     */
    @SuppressWarnings("unchecked")
    public static Object getCatPageExhibitMetaDataHTML(final DataSourceBean dsb,
                                                       final String exhibitName)
        {
        assert((dsb != null) && ExhibitName.validNameSyntax(exhibitName));

        // Get existing cache map, or create new one.
        // The cache is a thread-safe Map (chosen for efficient get() access)
        // from full exhibit name to HTML formatted metadata ("" if none).
        // Races here may result in some wasted work but no errors.
        Map<String,Object> cachedMetaDataMap;
        while((cachedMetaDataMap = (Map<String,Object>) dsb.getAEPLinkedValue(metadataCacheKey)) == null)
            { dsb.putIfAbsentAEPLinkedValue(metadataCacheKey, new ConcurrentHashMap<String, Object>(1931, 0.7f, 4)); }
        // Get/compute the metadata as expanded HTML form.
        Object cachedMetaData = cachedMetaDataMap.get(exhibitName);
        if(cachedMetaData == null)
            {
            // Need to compute metadata for this exhibit
            // for the first time and/or after an AEP load/change.
            try
                {
                final AllExhibitProperties aep = dsb.getAllExhibitProperties(-1);
                cachedMetaData = getCatPageExhibitMetaDataHTMLRaw(exhibitName, aep);
                }
            catch(final IOException e)
                {
                // Give up without cacheing anything in case of error.
                // e.printStackTrace(); // Absorb the error and don't whinge...
                return(""); // Cannot compute the value right now...
                }

            // Cache the expanded HTML metadata.
            cachedMetaDataMap.put(exhibitName, cachedMetaData);
            }
        assert(cachedMetaData != null);

        return(cachedMetaData);
        }

    /**Computes the raw data for getCatPageExhibitMetaDataHTML() without cacheing; never null.
     * @return "" if there is no metadata for this exhibit
     */
    public static Object getCatPageExhibitMetaDataHTMLRaw(final String exhibitName, final AllExhibitProperties aep)
        {
        final ExhibitPropsLoadable epl = aep.getExhibitPropsLoadable(exhibitName);
        final ExhibitPropsComputable epc = aep.getExhibitPropsComputable(exhibitName);

        final Node metadata = (epc == null) ? null : epc.getMetadata();
        final AccessionData accessionMetadata = epl.getAccessionMetadata();
        if((metadata != null) || (accessionMetadata != null))
            {
            final StringBuilder sb = new StringBuilder(2048);
            if(metadata != null)
                { sb.append(TextUtils.toXML(metadata, true, true)); }
            if(accessionMetadata != null)
                { sb.append(TextUtils.toXML(accessionMetadata.getAsDOM(), true, true)); }
            // We do not intern() the result
            // since we expect each non-"" metadata value to be unique.
            // We do try to hold in an a compact form if possible,
            // but return as a plain String if not.
            try { return(Compact7BitString.convertToCompact7BitString(sb, sDictMD)); }
            catch(final IllegalArgumentException e) { return(sb.toString()); }
            }

        // No metadata at all.
        return("");
        }
    
    /**As for getCatPageExhibitMetaDataHTML(), but will compute a missing value asynchronously; never null.
     * If the value is already computed then it is available immediately,
     * else a thread is started to compute the value,
     * and get() will block until the value is ready/computed.
     * <p>
     * The toString() method should be called on the result to get the String text.
     */
    @SuppressWarnings("unchecked")
    public static Future<?> getCatPageExhibitMetaDataHTMLFuture(
                                             final DataSourceBean dsb,
                                             final String exhibitName)
        {
        // Try first to return any extant cached value without blocking.
        // If present, we need not start any thread at all.
        final Map<String,Object> cachedMetaDataMap = (Map<String,Object>) dsb.getAEPLinkedValue(metadataCacheKey);
        if(null != cachedMetaDataMap)
            {
            final Object cachedMetaData = cachedMetaDataMap.get(exhibitName);
            if(null != cachedMetaData)
                {
                // Found cached value!
                // Return already-finished Future with value.
                return(new Future<Object>(){
                    public boolean cancel(boolean mayInterruptIfRunning)
                        { return(false); }
                    public Object get() // throws InterruptedException, ExecutionException
                        { return(cachedMetaData); }
                    public Object get(long timeout, TimeUnit unit) // throws InterruptedException, ExecutionException, TimeoutException
                        { return(cachedMetaData); }
                    public boolean isCancelled()
                        { return(false); }
                    public boolean isDone()
                        { return(true); }
                    });
                }
            }

        // If we did not find a cached value,
        // then set up the task to compute the value asynchronously.        
        final Callable<?> callable = new Callable<Object>(){
            public final Object call() throws Exception
                { return(getCatPageExhibitMetaDataHTML(dsb, exhibitName)); }
            };
        // Start the (CPU-intensive) thread immediately.
        final Future<?> result = ThreadUtils.computeIntensiveThreadPool.submit(callable);

        // Return the handle for the caller to retrieve the value...
        return(result);
        }

    /**Minimum length of string we will consider hyphenating.
     * Takes into account typical viewing width and the
     * inefficiency of introducing all the control tags to
     * do the hyphenation properly.
     */
    private static final int minHypenationLength = 30;

    /**Minimum distance from start of string (possibly from the
     * end too at some point) that a hyphen will be inserted.
     * If comparable in length to the ``<wbr>'' tag then the
     * text will not be expanded by more than a factor of two.
     */
    private static final int hyphenAfter = 6;

    /**Hyphenate long HTML text (that contains zero or more `-' characters and little or no whitespace).
     * Basic algorithm: takes text (without embedded tags), and wraps it in
     * &lt;nobr&gt;...&lt;/nobr&gt; tags to prevent line breaks in those browsers
     * that understand it.  If the hyphenate argument is true, and if
     * the input string is long enough to justify the overhead, insert
     * a <WBR> word-break tag after each ``-'', which on both the
     * new browsers that understand <NOBR>, and old browsers that break
     * at all tags, will allow a line-break after the dashes.
     * <P>
     * The <nobr>...</nobr> tags are only wrapped in if there is at least
     * one whitespace character (taken to be space, TAB, newline or
     * return) and/or hyphenation is turned on and is possible.
     * <P>
     * This `hyphenation' is full of wired-in heuristics.
     * <p>
     * If DISPLAY_SPACE_IN_EXHIBIT_NAMES is true, then to `hyphenate'
     * simply replace all the hyphens with spaces to let the browser
     * wrap the text more easily, and do not wrap in nbr tags.
     */
    public static final String hyphenateHTMLText(final String s)
        {
        if(WebConsts.DISPLAY_SPACE_IN_EXHIBIT_NAMES)
            { return(s.replace('-', ' ')); }

        final int l = s.length();

        // First possible hyphenation position considered.
        final int firstHyphenPos = Math.max(
            hyphenAfter, minHypenationLength/2);

        // Crude measure if hypehenation is possible is that
        // the string is long enough to be worth hyphenating and
        // contains at least one non-initial ``-''.
        final boolean hyphenationIsPossible = (
            (l >= minHypenationLength) &&
                (s.lastIndexOf('-') >= firstHyphenPos));

        // See if the string contains any whitespace actually requiring
        // the <NOBR>...</NOBR> tags to prevent breaking in...
        final boolean containsWhitespace = (
            (s.indexOf(' ')  != -1) || (s.indexOf('\t') != -1) ||
            (s.indexOf('\n') != -1) || (s.indexOf('\r') != -1));

        // OK, so wrap in NOBRs only if they are needed.
        final boolean needsNOBR =
            (hyphenationIsPossible || containsWhitespace);

        // Allow space for the various tags we will fold in.
        final StringBuilder sb = new StringBuilder(l +
                            (needsNOBR ? 13 : 0) + // Space for <NOBR>s.
                            (hyphenationIsPossible ? l : 0));

        if(needsNOBR) { sb.append("<nobr>"); }

        // If:
        //  1) hyphenation is requested, and
        //  2) the string is long enough to be worth hyphenating, and
        //  3) the string contains at least one dash not at the very start,
        // then try to hyphenate.
        if(hyphenationIsPossible)
            {
            final char ca[] = s.toCharArray();
            int prevhyphen = 0;
            // Don't bother to hyphenate within the first
            // few or last few characters (the normal min
            // interhyphen gap), nor within the initial
            // few characters anyway (half the minimum
            // hyphenation length).
            final int end = l - hyphenAfter;
            for(int i = 0; i < l; ++i)
                {
                final char c = ca[i];
                sb.append(c);
                if((i < firstHyphenPos) || (i >= end)) { continue; }
                // Avoid hyphenating in the first few chars.
                if((c == '-') && (i >= prevhyphen + hyphenAfter))
                    { sb.append("<wbr>"); prevhyphen = i; }
                }
            }
        else { sb.append(s); } // Too small to hyphenate...

        if(needsNOBR) { sb.append("</nobr>"); }
        return(sb.toString());
        }

    /**LRU thread-safe private cache mapping from exhibit type and Accept header to acceptablity of that MIME type for inlining in XHTML mobile text.
     * We assume that the Accept headers will be more or less constant for a given device,
     * and probably constant between instances of the device,
     * so we have enough entries to cover the likely different <em>types</em> of device.
     * <p>
     * We take care to avoid huge using huge Accept values in keys to avoid DoS-style issues.
     */
    private static final MemoryTools.SimpleLRUMap<Tuple.Pair<ExhibitMIME.ExhibitTypeParameters,String>, Boolean> _inlineableInXHTML = new MemoryTools.SimpleLRUMap<Tuple.Pair<ExhibitMIME.ExhibitTypeParameters,String>, Boolean>(1001);

    /**Returns true if the given MIME-type can always be inlined in an XHTML (mobile) page.
     * If the type argument is null, this returns false.
     * <p>
     * This always allows JPEG and GIF,
     * but may also allow other (image) types listed in an incoming "Accept" header.
     * <p>
     * TODO: create LRU cache of mapping from type and (common) Accept header to result to save time on the String processing.
     */
    public static boolean canInlineInXHTMLPage(final ExhibitMIME.ExhibitTypeParameters exhibitType,
                                               final HttpServletRequest request)
        {
        if(exhibitType == null) { return(false); }

        // Allow commonly-supported GIF (89a) and JPEG types always.
        switch(exhibitType.type)
            {
            case ExhibitMIME.ET_JPEG:
            case ExhibitMIME.ET_GIF: // Should really check GIF version.
                return(true);
            }
        // No header to analyse, so cannot allow more than the basic types.
        if(request == null) { return(false); }

        // We are prepared to test for a limited selection of other inlineable types.
        switch(exhibitType.type)
            {
            case ExhibitMIME.ET_PNG:
            case ExhibitMIME.ET_SWF:
            case ExhibitMIME.ET_BMP:
            // TODO: WBMP, SVG?
                {
                // Look for any Accept headers.
                final Enumeration headers = request.getHeaders("Accept");
                // No Accept headers to analyse, so cannot allow more than the basic types.        
                if(headers == null) { return(false); }

                // Allow for possibility of multiple Accept headers (probably rare though).
                while(headers.hasMoreElements())
                    {
                    final String h = (String) headers.nextElement();
                    // We refuse to process gigantic header values at all
                    // to preserve performance and avoid DoS-style problems.
                    if(h.length() >= 1024) { continue; }

                    // Make cache lookup key and attempt cache lookup.
                    final Tuple.Pair<ExhibitMIME.ExhibitTypeParameters,String> key =
                        new Tuple.Pair<ExhibitMIME.ExhibitTypeParameters,String>(exhibitType, h);
                    final Boolean cachedResult = _inlineableInXHTML.get(key);
                    if(Boolean.TRUE.equals(cachedResult)) { return(true); }
                    if(Boolean.FALSE.equals(cachedResult)) { continue; /* Try other headers. */ }
System.out.println("INFO: new XHTML Accept header '"+TextUtils.sanitiseForXML(h, 256, true)+"' from User-Agent: "+TextUtils.sanitiseForXML(request.getHeader("User-Agent"), 256, true));

                    // Do explicit search though header for specified MIME type.
                    for(final String t : h.split(","))
                        {                            
                        // Allow for direct MIME-type match,
                        // or with trailing q factor, eg "image/png;q=0.9",
                        // or just a generic catch-all */*.
                        final String trimmed = t.trim(); // Remove padding whitespace.
                        if(trimmed.equals("*/*") ||
                           trimmed.equals(exhibitType.mimeType) ||
                           (trimmed.startsWith(exhibitType.mimeType) && trimmed.substring(exhibitType.mimeType.length()).startsWith(";")))
                            {
                            _inlineableInXHTML.put(key, Boolean.TRUE); // Cache success.
                            return(true);
                            }                        
                        }

                    // Cache failure (against this header).
                    _inlineableInXHTML.put(key, Boolean.FALSE);
                    }
                }
            }
        return(false);
        }

    /**Returns true if the given MIME-type can always be inlined in an HTML page.
     * If the argument is null, this returns false.
     */
    public static boolean canInlineInHTMLPage(final ExhibitMIME.ExhibitTypeParameters exhibitType)
        {
        if(exhibitType == null) { return(false); }
        switch(exhibitType.type)
            {
            case ExhibitMIME.ET_JPEG:
            case ExhibitMIME.ET_GIF:
            case ExhibitMIME.ET_PNG: // Most HTML browsers will accept PNG now.
            case ExhibitMIME.ET_SWF: // Most browsers have a Flash plug-in now.
            case ExhibitMIME.ET_BMP: // Embedded BMP should be widely supported.
            case ExhibitMIME.ET_HTMLFRAG:
                return(true);
            }
        return(false);
        }

    /**Private key used by getCategoryTreeFilterBean(); never null. */
    private static final DataSourceBean.AEPLinkedKey _getCategoryTreeFilterBeanKey = new DataSourceBean.AEPLinkedKey("_getCategoryTreeFilterBeanKey");

    /**Get selected by-category TreeFilterBean from entire exhibit set; never null.
     * Used for the "by category" exhibit tree view and elsewhere.
     * <p>
     * This data is cached linked to the DSB.
     * <p>
     * The category name is only checked for syntactic validity,
     * not for actual presence in the AEP.
     * <p>
     * Designed to be passed a null inputList parameter for select, eg:
     * <code>List nodeContent = tfb.selectPrefixesAndShortNames(application, wsPrefix, null)</code>.
     */
    @SuppressWarnings("unchecked")
    public static TreeFilterBean getCategoryTreeFilterBean(final DataSourceBean dsb,
                                                           final String category)
        {
        if((dsb == null) || !ExhibitName.validNameInitialComponentSyntax(category))
            { throw new IllegalArgumentException(); }

        // Map from category name to TreeFilerBean.
        // Created as necessary on first use (after AEP change).
        ConcurrentMap<String,TreeFilterBean> trees;
        while(null == (trees = (ConcurrentMap<String,TreeFilterBean>) dsb.getAEPLinkedValue(_getCategoryTreeFilterBeanKey)))
            { dsb.putIfAbsentAEPLinkedValue(_getCategoryTreeFilterBeanKey, new ConcurrentHashMap<String,TreeFilterBean>()); }

        // Atomically ensure right tree filter bean for category exists in Hashtable, else create it.
        TreeFilterBean tfb;
        while(null == (tfb = (TreeFilterBean) trees.get(category)))
            {
            // Create the filter bean.
            tfb = new TreeFilterBean();
            // By default, hold cache of computed data as this is assumed to be expensive.
            tfb.setMemorySensitiveCache(false);
            // Set the filter for the right category.
            tfb.setExpr(new FilterExpr(null, new BuiltInFilters.filtByCategory(new String[]{category})));
            // Store the new bean in the cache, iff no one else got there first.
            trees.putIfAbsent(category, tfb);
            }

        return(tfb);
        }


    /**If true then allow some "sneaky" browser concurrency.
     * For browsers that do not always pipeline by default
     * but do allow fetches from different-named hosts in parallel
     * (eg the biggies such as IE6/IE7, FF1/1.5/2 as of 2006Q4),
     * then we can possibly help throughput by fetching some images (etc)
     * from a different name for this same host.
     * <p>
     * The trick is that where the user has arrived at the site
     * with a name other than the local mirror name,
     * and the URI that we were going to use was a rrURL (root-relative URL),
     * then we can prepend the local mirror name to make new absolute URL
     * that the browser may be prepared to fetch from concurrently.
     * <p>
     * This has to be done completely consistently for any given item,
     * eg a static "page-furniture" image, to avoid defeating cacheing.
     */
    public static final boolean ALLOW_SNEAKY_HTTP_CONCURRENCY = true;

    /**Get optional prefix for rrURL for extra "sneaky" browser concurrency.
     * This can only apply if:
     * <ul>
     * <li>ALLOW_SNEAKY_HTTP_CONCURRENCY is true.
     * <li>This host has a mirror prefix
     *     (ie so that there is an "alternate" name for this host.
     * <li>The incoming request hostname is not this local mirror hostname.
     * </ul>
     * else this routine always returns "".
     * <p>
     * This basically only works if the user is using a "generic" URL,
     * but that is more likely to be a machine far away (ie with large RTT),
     * so extra concurrency to try to overcome latency is especially helpful.
     * <p>
     * This slightly inflates the HTML that the client will see
     * but only if using a generic URL.
     * <p>
     * May inflate the number of concurrent connections back to this host,
     * but usually only by 1 or 2 at most.
     * <p>
     * Note that this scheme <em>does not</em> rely on any other host being up,
     * nor having exactly the same content as us.
     * <p>
     * This <em>is not</em> a technique for distributing load.
     *
     * @return "" or the http://mirror-... name (with no trailing slash) for this host
     */
    public static String getOptionalSneakyConcurrencyRRURLPrefix(final HttpServletRequest request)
        {
        if(!ALLOW_SNEAKY_HTTP_CONCURRENCY)
            { return(""); /* The default, ie no prefix. */}

        final String mirrorTag = LocalProps.getMirrorTag();
        if(null == mirrorTag)
            { return(""); /* No reliable client-visible alternate hostname for us. */ }

        final String localMirrorName = HostUtils.makeMirrorNameGeneric(mirrorTag);
        if(localMirrorName.equals(request.getServerName()))
            { return(""); /* Client is already using the alternate hostname for us. */ }

        // Return suitable prefix to make rrURL into absolute URL.
        return("http://" + localMirrorName);
        }

    /**Name of generic GenProps property for news-flash HTML segment. */
    private static final String GEN_NEWSFLASH_HTML = "newsflash.HTML";

    /**Get "newsflash" HTML for the main site front page, or "" if none; never null.
     * This is retrieved from the GenProps.
     */
    public static String getNewsflashHTML(final GenProps gp)
        {
        if(gp == null) { return(""); /* Be kind to the caller. */ }
        final String result = gp.getGen().get(GEN_NEWSFLASH_HTML);
        if(result == null) { return(""); }
        return(result);
        }
    }
