/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

//import ORG.hd.d.IsDebug;

import javax.servlet.*;
import javax.servlet.http.HttpServletResponse;

import org.hd.d.pg2k.svrCore.stats.StatsLogger;
//import javax.servlet.http.*;
import java.io.IOException;
//import java.util.*;

/**Compresses XML by removing redundant whitespace (etc).
 * Should quickly reduce the size of many XML documents on the fly
 * without changing their semantics.
 * <p>
 * This will allow document streaming (ie incremental processing)
 * and will support flush() to allow the push of partial documents,
 * although this may not be able to push partial XML atoms
 * (eg partial elements and entities).
 * <p>
 * To avoid upsetting some older consumers of XML,
 * such as old mobile phones and PDAs,
 * this may avoid some transformations such as
 * <code>&lt;x&gt;&lt;/x&gt;</code> to <code>&lt;x/&gt;</code>.
 * <p>
 * To avoid upsetting older HTML consumers of XHTML
 * this may avoid some transformations such as
 * <code>&lt;br /&gt;</code> to <code>&lt;br/&gt;</code>.
 */
public final class XMLCompressionFilter implements Filter
    {
    /**Our logger which falls back to System.out if servlet log not available; never null. */
    private final WebUtils.ServletLoggerWithFallback logger = new WebUtils.ServletLoggerWithFallback();

    /**The filter configuration object we are associated with.
     * If this value is null, this filter instance is not currently configured.
     */
    private FilterConfig config;

    public FilterConfig getFilterConfig()
        {
        return (config);
        }

    public void setFilterConfig(final FilterConfig c)
        {
        config = c;
        logger.setContext((c == null) ? null : c.getServletContext());
        }

    /**Place this filter into service.
     * @param filterConfig  filter configuration object; null to unset
     */
    public void init(final FilterConfig filterConfig)
        {
        config = filterConfig;
        logger.setContext((filterConfig == null) ? null : filterConfig.getServletContext());
        }

    /**Take this filter out of service. */
    public void destroy()
        {
        this.config = null;
        logger.setContext(null);
        }

    /**Compacts the (XML) reponse text when possible.
     */
    public void doFilter(final ServletRequest request,
                         final ServletResponse response,
                         final FilterChain chain)
        throws IOException, ServletException
        {
        if(response instanceof HttpServletResponse)
            {
            final XMLCompressionServletResponseWrapper wrappedResponse =
                new XMLCompressionServletResponseWrapper((HttpServletResponse) response);
            try { chain.doFilter(request, wrappedResponse); }
            finally { wrappedResponse.finishResponse(); }
            return;
            }

        // Pass on unchanged...
        chain.doFilter(request, response);
        }
    }

