/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.webSvr.util;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;

import ORG.hd.d.IsDebug;

import java.io.*;
import java.util.concurrent.atomic.AtomicBoolean;

/**Allows conservative compaction of (removal of redundant characters from) an XML response.
 * This implementation is only prepared to try to compact the text
 * when it can be parse unambiguously and in a light-weight way.
 * <p>
 * Thus it will only attempt to filter a (char-based) Writer response,
 * not a (byte-based) Stream response, for now.
 * <p>
 * We call the Writer of the underlying response
 * to preserve the char-based nature of the text as we are processing it.
 * <p>
 * This implementation is not thread-safe.
 * <p>
 * This class does not need to be public nor derived from.
 */
final class XMLCompressionServletResponseWrapper extends HttpServletResponseWrapper
    {
    public XMLCompressionServletResponseWrapper(final HttpServletResponse response)
        {
        super(response);
        if(null == response) { throw new IllegalArgumentException(); }
        origResponse = response;
        }

    /**Original response; never null. */
    private final HttpServletResponse origResponse;

    /**Any ServletOutputStream that has been returned, else null. */
    private ServletOutputStream stream;

    /**Any PrintWriter that has been returned, else null. */
    private PrintWriter writer;

    /**The PrintWriter from the underlying response, else null.
     * Note that underlyingWriter is only non-null if writer is non-null,
     * and the writers and stream cannot not be non-null.
     */
    private PrintWriter underlyingWriter;

    /**Get a ServletOutputStream to write the content associated for this response.
     * This will prevent us from filtering the content.
     * <p>
     * We get the stream from the underlying response directly.
     */
    @Override public ServletOutputStream getOutputStream() throws IOException
        {
        if(writer != null)
            { throw new IllegalStateException("Writer already created for this response"); }
        if(stream != null)
            { return(stream); }
if(IsDebug.isDebug) { System.out.println("WARNING: XML compression disabled because stream requested."); }
        stream = origResponse.getOutputStream();
        return(stream);
        }

//    // Inherit javadoc.
//    @Override public void flushBuffer() throws IOException
//        {
//        // Flush everything from the top down...
//        if(writer != null) { writer.flush(); }
//        if(underlyingWriter != null) { underlyingWriter.flush(); }
//        if(stream != null) { stream.flush(); }
//        super.flushBuffer();
//        }

    /**If true then we are removing whitespace.
     * We start off (after creation or reset)
     * by removing any leading whitespace in the output,
     * which is technically not allowed anyway in XML.
     */
    private final AtomicBoolean removingWhitespace = new AtomicBoolean(true);

    // Inherit javadoc.
    @Override public PrintWriter getWriter() throws IOException
        {
        if(writer != null)
            { return(writer); }
        if(stream != null)
            { throw new IllegalStateException("Stream already created for this response"); }

        // We cannot apply any encoding/compression if there is a Content-MD5 header,
        // since it should be applied/recomputed after this compression.
        if(origResponse.containsHeader("Content-MD5"))
            {
            writer = origResponse.getWriter();
            return(writer);
            } // Already encoded.

        // Record the underlying writer.
        underlyingWriter = origResponse.getWriter();

        // We don't provide any internal buffering,
        // but assume that it is provided by the wrapped response object.
        final Writer out = underlyingWriter;

//if(IsDebug.isDebug) { System.out.println("INFO: XML compression on."); }

        // Wrap the output in a compacting and buffering [Print]Writer instance.
        // We choose to override methods at the Writer level for robustness
        // (we cannot control how the internal implmentation of PrintWriter may change).
        // We buffer our output for efficiency on the wire.
        // The basic aim is to preserve XML semantics and linebreaks,
        // and be fast, while removing most common redundant whitespace,
        // so we remove whitespace after a \n until a char > 32.
        // This also removes blank lines.
        //
        // Since \n should never appear in the middle of attribute values, etc,
        // this should preserve essentially all significant XML whitespace.
        //
        // TODO: MAY HAVE TO MAKE SURE THAT WE HANDLE reset[Buffer]() correctly.
        writer = new PrintWriter(new Writer(){

            @Override
            public void close() throws IOException
                { out.close(); }

            @Override
            public void flush() throws IOException
                { out.flush(); }

            /**Write character data to the underlying stream. */
            @Override
            public void write(final char[] cbuf, int off, int len) throws IOException
                {
                synchronized(lock) // Have lock, so will use it, even though this code never gets multi-threaded.
                    {
                    // Esnure only one read and one write of volatile value is needed.
                    boolean rWS = removingWhitespace.get();
                    try
                        {
                        // We process the output characters one at a time.
                        outer: while(len > 0)
                            {
                            final char c = cbuf[off];
                            // Skip any ASCII chars <= 32 while removing whitespace.
                            if(rWS && (c <= ' ')) { --len; ++off; continue; }

                            // Assume that we now have a non-whitespace character
                            // or are in a context where we are not stripping whitespace,
                            // so write the next character(s) as-is.
                            rWS = false;

                            // For efficiency, try to write chars up to next \n in one go.
                            for(int o = off, l = len; l > 0; --l)
                                {
                                if(cbuf[o++] == '\n')
                                    {
                                    final int nChars = o - off;
                                    out.write(cbuf, off, nChars);
                                    off += nChars;
                                    len -= nChars;
                                    rWS = true; // Back to stripping whitespace after \n.
                                    continue outer;
                                    }
                                }

                            // No \n in tail of buffer,
                            // so write the rest of the chars out in one go.
                            out.write(cbuf, off, len);
                            break;
                            }
                        }
                    finally
                        { removingWhitespace.set(rWS); }
                    }
                }
            });
        return(writer);
        }

    @Override
    public void reset()
        {
        super.reset();
        removingWhitespace.set(true); // If reset was allowed then be prepared to zap whitespace again.
        }

    @Override
    public void resetBuffer()
        {
        super.resetBuffer();
        removingWhitespace.set(true); // If reset was allowed then be prepared to zap whitespace again.
        }

    /**Prevent the Content-Length header from being set.
     * Since we aim to compact the text,
     * then any content length that the caller is aware of
     * is hopefully wrong (too large).
     */
    @Override public void setContentLength(int length)
        {
        // Do not allow this value/header to be set.
        }
    
    /**Finish up. */
    public void finishResponse()
        {
        try
            {
            if(writer != null) { writer.close(); }
            else if(stream != null) { stream.close(); }
            }
        catch(final IOException e)
            {
            // Absorb error but log it.
            e.printStackTrace();
            }
        }
    }
