/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth;

import org.hd.d.pg2k.svrCore.PGException;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.uploader.UploaderUtils;

import javax.servlet.http.HttpServletRequest;
import java.awt.*;
import java.io.Serializable;
import java.util.Properties;
import java.util.StringTokenizer;

/**JavaBean used to parse/validate/adjust/save view parameters.
 * This can have the zoom and (x,y)-offset parameters of a particular
 * view set in various ways, and will coerce them to valid values.
 * This can also be used to generate suitably-encoded parameters for
 * a visitor to navigate in the Aloha Earth space.
 * <p>
 * So that this is a valid JavaBean we make it Serializable,
 * but in practice never expect it to be persisted,
 * even temporarily in a session,
 * so we don't bother with a unique serialisation ID, validation, etc,
 * but we do make any derived internal values transient.
 * <p>
 * This is <strong>NOT</strong> currently designed to be thread-safe.
 * <p>
 * This implements Clonable to as to make it possible to easily
 * take a copy of the parameters, alter them, and turn the new values
 * into a form's parameter set, for example.
 * <p>
 * The main parameters held in this bean are:
 * <ul>
 * <li>The zoom factor: the amount the original image is zoomed in/out
 *     for display to the user.  Bigger numbers mean more zoom in,
 *     ie greater expansion.
 * <li>The offset of the center of the (zoomed, shifted) image shown to the user
 *     north and east from the original image (which is centered on 0N OE).
 * </ul>
 * These parameters are constrained internally to legal range,
 * independently of one another.  On extraction of the values with
 * accessor methods the values may optionally be contrained to mutually-legal
 * combinations.
 * <p>
 * Note that for the East Offset and North Offset,
 * the offset is measured from the centre of the current image
 * to the centre of the original underlying image, where either
 * centre may potentially lie within a pixel or between pixels.
 * <p>
 * The unit of measure is pixels of the original image,
 * making the units invariant wrt the zoom factor of the current image,
 * and thus this value is constrained to lie in the range
 * [-BASE_2D_EARTH_MAP_WIDTH/2, +BASE_2D_EARTH_MAP_WIDTH/2] for EO, and
 * [-BASE_2D_EARTH_MAP_HEIGHT/2, +BASE_2D_EARTH_MAP_HEIGHT/2] for NO
 * though usually within a narrower range than that.
 * <p>
 * Note that the fact that when we zoom in one pixel in the original
 * takes several pixels in the current/displayed image means that
 * careful handling is required to get image alignment correct.
 */
public final class AEParams implements Serializable, Cloneable
    {
    /**If true, round offset values to "tile" multiples.
     * This ensures that fewer unique views can be seen,
     * which should make cacheing feasable.
     * <p>
     * Across each view in each direction (E and N)
     * are ZOOM_RATIO tiles.
     * A user may move laterally in either dimension one tile width,
     * and may zoom in to one of the tiles of the current image.
     * Plainly, tiles from adjacent "windows" can in principle be shared,
     * which we might make use of explicitly with a tiled display in HTML,
     * or implicitly by say having JAI draw the image on demand and
     * have a Renderable in cache.
     */
    public static final boolean FORCE_TILE_POSITIONING = true;

    /**Field name for map image as mouse-sensitive submit button.
     * If the user clicks on the map image, we should receive a
     * pair of [name].x and [name].y parameters of where the user clicked.
     */
    public static final String NAME_MAP_FIELD = "map";

    /**Amount to add to zoomFactor on clicking on the map image; 0 means no zoom. */
    public static final int MAP_CLICK_ZF_ADJ = +1;

    /**Name of zoom-control field (submit button).
     * This, if set, carries changes to be applied after the
     * zoom factor has been set from the NAME_ZOOM_FACTOR parameter.
     */
    public static final String NAME_ZOOM_CONTROL_FIELD = "zoomCtl";

    /**External name of zoomFactor parameter, eg in HTTP GET/POST. */
    public static final String NAME_ZOOM_FACTOR = "zf";

    /**The current zoom factor.
     * Constrained to lie between
     * MIN_ZOOM and MAX_ZOOM inclusive,
     * and INITIAL_ZOOM unless otherwise set (ie the initial value).
     */
    private int zoomFactor = AEUtils.INITIAL_ZOOM;

    /**The absolute maximum limit (+/-) of the East Offset in pixels; positive and rounded down. */
    public static final int OFFSET_LIMIT_EAST = AEUtils.BASE_2D_EARTH_MAP_WIDTH/2;

    /**The absolute maximum limit (+/-) of the NORTH Offset in pixels; positive and rounded down. */
    public static final int OFFSET_LIMIT_NORTH = AEUtils.BASE_2D_EARTH_MAP_HEIGHT/2;

    /**Get the current zoom factor.
     * Constrained to lie between
     * MIN_ZOOM and MAX_ZOOM inclusive,
     * and INITIAL_ZOOM unless otherwise set.
     */
    public int getZoomFactor() { return(zoomFactor); }

    /**Set the current zoom factor as an integer; will be coerced to range.
     * Constrained to lie between
     * MIN_ZOOM and MAX_ZOOM inclusive,
     * and the value will be coerced to those limits.
     *
     * @return reference to this modified object
     */
    public AEParams setZoomFactor(int zf)
        {
        if(zf < AEUtils.MIN_ZOOM)
            { zf = AEUtils.MIN_ZOOM; }
        else if(zf > AEUtils.MAX_ZOOM)
            { zf = AEUtils.MAX_ZOOM; }
        zoomFactor = zf;
        return(this);
        }

    /**Set the current value as a String; parse and coerced as necessary.
     * Parsed as a decimal (possibly-signed) integer
     * and constrained to lie between
     * MIN_ZOOM and MAX_ZOOM inclusive,
     * and the value will be coerced to those limits.
     * <p>
     * In case of a parse error (including a null argument)
     * no value is set.
     *
     * @return reference to this modified object
     */
    public AEParams setZoomFactor(final String zf)
        {
        try { setZoomFactor(Integer.parseInt(zf, 10)); }
        // Simply ignore set attempt in case of null or bad argument.
        catch(final NumberFormatException e) { }
        return(this);
        }

    /**External name of North Offset parameter, eg in HTTP GET/POST. */
    public static final String NAME_NORTH_OFFSET = "no";

    /**External name of East Offset parameter, eg in HTTP GET/POST. */
    public static final String NAME_EAST_OFFSET = "eo";

    /**The East Offset of of the current view from original image.
     * The offset is measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_WIDTH/2, +BASE_2D_EARTH_MAP_WIDTH/2].
     * <p>
     * The initial value is zero.
     */
    private int eastOffset; // Initially zero.

    /**The North Offset of of the current view from original image.
     * As the centre of the user's view goes more north,
     * this becomes more positive.
     * <p>
     * The offset of measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_HEIGHT/2, +BASE_2D_EARTH_MAP_HEIGHT/2].
     * <p>
     * Note that is is the negative of the change in the y axis
     * which in computer graphics usually has higher values lower down.
     * <p>
     * The initial value is zero.
     */
    private int northOffset; // Initially zero.


    /**Set the East Offset of the current view from the base/original image.
     * As the centre of the user's view goes more east,
     * this becomes more positive.
     * <p>
     * The offset of measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_WIDTH/2, +BASE_2D_EARTH_MAP_WIDTH/2];
     * any value outside these bounds is coerced to range.
     * <p>
     * When read back "constrained" this is coerced to try to keep
     * the current view within the original source image.
     * <p>
     * The initial (unset/default) value is zero.
     * <p>
     *
     * @return reference to this modified object
     */
    public AEParams setEastOffset(int pixelsOffset)
        {
        // Coerce into range if need be.
        if(pixelsOffset < -OFFSET_LIMIT_EAST)
            { eastOffset = -OFFSET_LIMIT_EAST; }
        else if (pixelsOffset > OFFSET_LIMIT_EAST)
            { eastOffset = OFFSET_LIMIT_EAST; }
        else
            { eastOffset = pixelsOffset; }

        return(this);
        }

    /**Set East offset as String.
     * Unparseable (or null) values are ignored,
     * parseable values are coerced to fit the valid range.
     *
     * @return reference to this modified object
     */
    public AEParams setEastOffset(String pixelsOffset)
        {
        try { setEastOffset(Integer.parseInt(pixelsOffset, 10)); }
        // Simply ignore set attempt in case of null or bad argument.
        catch(NumberFormatException e) { }
        return(this);
        }

    /**Set the North Offset of the current view from the base/original image.
     * As the centre of the user's view goes more north,
     * this becomes more positive.
     * <p>
     * The offset of measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_HEIGHT/2, +BASE_2D_EARTH_MAP_HEIGHT/2];
     * any value outside these bounds is coerced to range.
     * <p>
     * When read back "constrained" this is coerced to try to keep
     * the current view within the original source image.
     * <p>
     * The initial (unset/default) value is zero.
     * <p>
     *
     * @return reference to this modified object
     */
    public AEParams setNorthOffset(int pixelsOffset)
        {
        // Coerce into range if need be.
        if(pixelsOffset < -OFFSET_LIMIT_NORTH)
            { northOffset = -OFFSET_LIMIT_NORTH; }
        else if (pixelsOffset > OFFSET_LIMIT_NORTH)
            { northOffset = OFFSET_LIMIT_NORTH; }
        else
            { northOffset = pixelsOffset; }

        return(this);
        }

    /**Set North offset as String.
     * Unparseable (or null) values are ignored,
     * parseable values are coerced to fit the valid range.
     *
     * @return reference to this modified object
     */
    public AEParams setNorthOffset(String pixelsOffset)
        {
        try { setNorthOffset(Integer.parseInt(pixelsOffset, 10)); }
        // Simply ignore set attempt in case of null or bad argument.
        catch(NumberFormatException e) { }
        return(this);
        }



    /**Get East Offset of the current view from the base/original image.
     * As the centre of the user's view goes more east,
     * this becomes more positive.
     * <p>
     * The offset of measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_WIDTH/2, +BASE_2D_EARTH_MAP_WIDTH/2]
     * at most.
     * <p>
     * If called for a "constrained" value, the return is clamped
     * so as to ensure that all of the current view lies within
     * the base image if possible.  In all cases the result, if adjusted,
     * is adjusted towards zero.
     * <p>
     * If we are rounding to tile widths, we round down.
     * <p>
     * The initial value is zero.
     *
     * @param constrained  if true, return is if necessary adjusted towards
     *     zero to try to force the current zoomed view to be from entirely
     *     within the source image and possibly rounded to a tile width,
     *     else the raw (but still clamped) value is returned
     */
    public int getEastOffset(boolean constrained)
        {
        if(!constrained) { return(eastOffset); }

        // Shift the apparent centre to keep the current viewport within
        // the base image.
        int eo = eastOffset;

        // Zero is a special case and is never adjusted.
        if(eo == 0)
            { return(0); }

        // Compute a limit on the offset which keep the rectangle of pixels
        // extracted from the source image within the source image.
        final int offsetLimit = OFFSET_LIMIT_EAST -
            ((getBaseSrcPixelsWidth(true) + 1) / 2); // Rounded-up measure.
        if(eo < -offsetLimit) { eo = -offsetLimit; }
        else if(eo > offsetLimit) { eo = offsetLimit; }

        // Round inwards to tile boundary towards (0,0) if needed.
        if(FORCE_TILE_POSITIONING)
            {
            final int tw = getTileSrcPixelsWidth();
            eo = (eo / tw) * tw;
            }

        // Return possibly-adjusted value.
        return(eo);
        }

    /**Get source image left (x) coordinate.
     * As the centre of the user's view goes more east,
     * this becomes more positive.
     * <p>
     * This calls getEastOffset().
     *
     * @param constrained  if true, the result is always within
     *     the source image bounds [0,BASE_2D_EARTH_MAP_WIDTH-1] inclusive
     *     and if FORCE_TILE_POSITIONING is true will be rounded to a tile size
     */
    private int getEastOffsetSrcPixelsX(boolean constrained)
        {
        final int width = getBaseSrcPixelsWidth(constrained);
        final int x = ((AEUtils.BASE_2D_EARTH_MAP_WIDTH - width + 1) / 2)
            + getEastOffset(constrained);
//System.out.println("getEastOffsetSrcPixelsX(): constrained="+constrained+", width="+width+", x="+x+", eo="+getEastOffset(false));
        if(constrained)
            {
            if(x < 0)
                { return(0); }
            assert(width > 0);
            if(x + width > AEUtils.BASE_2D_EARTH_MAP_WIDTH)
                { return(Math.max(0, AEUtils.BASE_2D_EARTH_MAP_WIDTH - width)); }
            }
        return(x);
        }

    /**Get North Offset of the current view from original image.
     * As the centre of the user's view goes more north,
     * this becomes more positive.
     * <p>
     * The offset of measured from the centre of the current image
     * to the centre of the original underlying image, where either
     * centre may potentially lie within a pixel or between pixels.
     * <p>
     * This value is constrained to lie in the range
     * [-BASE_2D_EARTH_MAP_HEIGHT/2, +BASE_2D_EARTH_MAP_HEIGHT/2]
     * at most.
     * <p>
     * Note that is is the negative of the change in the y axis
     * which in computer graphics usually has higher values lower down.
     * <p>
     * The initial value is zero.
     *
     * @param constrained  if true, return is if necessary adjusted towards
     *     zero to try to force the current zoomed view to be from entirely
     *     within the source image,
     *     else the raw (but still clamped) value is returned
     */
    public int getNorthOffset(boolean constrained)
        {
        if(!constrained) { return(northOffset); }

        // Shift the apparent centre to keep the current viewport within
        // the base image.
        int no = northOffset;

        // Zero is a special case and is never adjusted.
        if(no == 0)
            { return(0); }

        // Compute a limit on the offset which keep the rectangle of pixels
        // extracted from the source image within the source image.
        final int offsetLimit = OFFSET_LIMIT_NORTH -
            ((getBaseSrcPixelsHeight(true) + 1) / 2); // Rounded-up measure.
        if(no < -offsetLimit) { no = -offsetLimit; }
        else if(no > offsetLimit) { no = offsetLimit; }

        // Round inwards to tile boundary towards (0,0) if needed.
        if(FORCE_TILE_POSITIONING)
            {
            final int th = getTileSrcPixelsHeight();
            no = (no / th) * th;
            }

        // Return possibly-adjusted value.
        return(no);
        }

    /**Get source image top (y) coordinate.
     * As the centre of the user's view goes more north,
     * this becomes more <strong>negative</strong>.
     * <p>
     * This calls getNorthOffset().
     *
     * @param constrained  if true, the result is always within
     *     the source image bounds [0,BASE_2D_EARTH_MAP_HEIGHT-1] inclusive
     *     and if FORCE_TILE_POSITIONING is true will be rounded to a tile size
     */
    private int getNorthOffsetSrcPixelsY(boolean constrained)
        {
        final int height = getBaseSrcPixelsHeight(constrained);
        final int y = ((AEUtils.BASE_2D_EARTH_MAP_HEIGHT - height + 1) / 2)
            - getNorthOffset(constrained);
//System.out.println("getNorthOffsetSrcPixelsY(): constrained="+constrained+", height="+height+", y="+y+", no="+getNorthOffset(false));
        if(constrained)
            {
            if(y < 0)
                { return(0); }
            assert(height > 0);
            if(y + height > AEUtils.BASE_2D_EARTH_MAP_HEIGHT)
                { return(Math.max(0, AEUtils.BASE_2D_EARTH_MAP_HEIGHT - height)); }
            }
        return(y);
        }

    /**Get East Offset as degrees East; range approximately [-180f, +180f].
     * As the centre of the user's view goes more east,
     * this becomes more positive.
     * <p>
     * The conversion will be a bit lumpy and inexact,
     * especially as the underlying step in the offset is likely to be similar
     * but not identical to a degree.  The caller could usefully round this.
     * <p>
     * Clamped to +/- Location.Estd.MAX_E.
     *
     * @param constrained  when true, if the source image area would lie
     *     partly or completely outside the base image, this is adjusted
     *     in line with getEastOffset() towards zero to try
     *     to bring it within the base image
     */
    public float getEastDegrees(boolean constrained)
        {
        return(Math.min(Location.Estd.MAX_E, Math.max(-Location.Estd.MAX_E,
            (getEastOffset(constrained) * 360.0f) / AEUtils.BASE_2D_EARTH_MAP_WIDTH)));
        }

    /**Get North Offset as degrees North; range approximately [-90f, +90f].
     * As the centre of the user's view goes more north,
     * this becomes more positive.
     * <p>
     * The conversion will be a bit lumpy and inexact,
     * especially as the underlying step in the offset is likely to be similar
     * but not identical to a degree.  The caller could usefully round this.
     * <p>
     * Clamped to +/- Location.Estd.MAX_N.
     */
    public float getNorthDegrees(boolean constrained)
        {
        return(Math.min(Location.Estd.MAX_N, Math.max(-Location.Estd.MAX_N,
            (getNorthOffset(constrained) * 180.0f) / AEUtils.BASE_2D_EARTH_MAP_HEIGHT)));
        }

    /**Get current offset as a Location.Estd; never null.
     * The error bounds indicate the area covered by the current view.
     */
    public Location.Estd getLocation(final boolean constrained)
        {
        // Compute the Error bounds as half the number of degrees
        // implied by the width and height of the constained display
        // source relative to the full source image.
        final float NErr =
            (90.0f * getBaseSrcPixelsHeight(constrained)) / AEUtils.BASE_2D_EARTH_MAP_HEIGHT;
        final float EErr =
            (180.0f * getBaseSrcPixelsWidth(constrained)) / AEUtils.BASE_2D_EARTH_MAP_WIDTH;

        // TODO: eliminate wasteful conversion to/from String if possible.
        final Properties props = new Properties();
        props.setProperty("type", "Estd");
        props.setProperty("E", String.valueOf(getEastDegrees(constrained)));
        props.setProperty("EErr", String.valueOf(EErr));
        props.setProperty("N", String.valueOf(getNorthDegrees(constrained)));
        props.setProperty("NErr", String.valueOf(NErr));
        // Construct a simple example Estd item.
        final Location.Estd e1;
        try {
            e1 = new Location.Estd(
                        true, // Claim that this data is "specific" to a particular exhibit.
                        "", // No prefix on the property names.
                        props // The properties.
                        );
            }
        catch(final PGException e) // Should not happen.
            { throw new Error("internal error" + e.getMessage()); }
        return(e1);
        }

    /**Gets pixel in current view corresponding to centre of given location; null if not in current view.
     *
     * @param targetLoc  location to find pixel for centre of; null if not visible
     */
    public Point getDisplayPixelForEstdLocationCentre(final Location.Estd targetLoc)
        {
        // If not even in the current view, return null.
        final Location.Estd viewArea = getLocation(true);
        if(!viewArea.containsCentre(targetLoc))
            { return(null); }

        final float targetN = targetLoc.getN().value.floatValue();
        int y = computeYfromN(targetN);

        final float targetE = targetLoc.getE().value.floatValue();
        int x = computeXfromE(targetE);

        return(new Point(x, y));
        }

    /**Compute X pixel from E degrees given current view; may not be visible. */
    int computeXfromE(final float targetE)
        {
        final Location.Estd viewArea = getLocation(true);
        final float offE = targetE - viewArea.getE().value.floatValue();
        final int maxWidth = AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH-1;
        final int x = Math.max(0, Math.min(maxWidth,
            Math.round((maxWidth+1) * 0.5f * ((offE / viewArea.getE().error.floatValue()) + 1))));
        return x;
        }

    /**Compute Y pixel from N degrees given current view; may not be visible. */
    int computeYfromN(final float targetN)
        {
        final Location.Estd viewArea = getLocation(true);
        final float offN = targetN - viewArea.getN().value.floatValue();
        final int maxHeight = AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT-1;
        final int y = maxHeight - Math.max(0, Math.min(maxHeight,
            Math.round((maxHeight+1) * 0.5f *((offN / viewArea.getN().error.floatValue()) + 1))));
        return y;
        }

    /**Get bounding rectangle of area to be zoomed to make display image.
     * Specified in pixels of the original image.
     * <p>
     * Depends on zoom and offsets.
     * <p>
     * Clamped so that no part of this will lie outside the original image,
     * but this may change the aspect ratio, for example.
     * <p>
     * The width and height are always strictly positive.
     * <p>
     * Note that we adjust the zoom we use so that the minimum allowable
     * zoom shows the whole image, but shunk or expanded from the raw image
     * if the MIN_ZOOM is non-zero.
     */
    public Rectangle getSourceRectangleToDisplay()
        {
        final int clX = getEastOffsetSrcPixelsX(true);
        final int clY = getNorthOffsetSrcPixelsY(true);
        final int clWidth = getBaseSrcPixelsWidth(true);
        final int clHeight = getBaseSrcPixelsHeight(true);
        return(new Rectangle(clX, clY, clWidth, clHeight));
        }

    /**Get height of current zoomed view in terms of original base-image pixels; always positive.
     * Depends only on the zoom factor and does not (mutually) recurse with
     * any of the the other accessor functions.
     *
     * @param constrained  if true, the result is constrained to be no more
     *     than the height of the base image
     */
    public int getBaseSrcPixelsHeight(boolean constrained)
        {
        final int nzf = _negativeZoomFactor();
        final int height = Math.max(1,
            AEUtils.applyZoom(AEUtils.BASE_2D_EARTH_MAP_HEIGHT, nzf));
        if(constrained)
            {
            if(height > AEUtils.BASE_2D_EARTH_MAP_HEIGHT)
                { return(AEUtils.BASE_2D_EARTH_MAP_HEIGHT); }
            }
        return(height);
        }

    /**Get width of current zoomed view in terms of original base-image pixels; always positive.
     * Depends only on the zoom factor and does not (mutually) recurse with
     * any of the the other accessor functions.
     *
     * @param constrained  if true, the result is constrained to be no more
     *     than the width of the base image
     */
    public int getBaseSrcPixelsWidth(boolean constrained)
        {
        final int nzf = _negativeZoomFactor();
        final int width = Math.max(1,
            AEUtils.applyZoom(AEUtils.BASE_2D_EARTH_MAP_WIDTH, nzf));
        if(constrained)
            {
            if(width > AEUtils.BASE_2D_EARTH_MAP_WIDTH)
                { return(AEUtils.BASE_2D_EARTH_MAP_WIDTH); }
            }
        return(width);
        }

    /**Get tile width in source-image pixels; strictly positive.
     * Is the rounded-down current image width in source-image pixels
     * divided by ZOOM_FACTOR, ie the number of steps left or right
     * a user view is the same as the ratio by which we zoom in or out.
     */
    public int getTileSrcPixelsWidth()
        {
        return(getBaseSrcPixelsWidth(false) / AEUtils.ZOOM_RATIO);
        }

    /**Get tile height in source-image pixels; strictly positive.
     * Is the rounded-down current image height in source-image pixels
     * divided by ZOOM_FACTOR, ie the number of steps up or down
     * a user view is the same as the ratio by which we zoom in or out.
     */
    public int getTileSrcPixelsHeight()
        {
        return(getBaseSrcPixelsHeight(false) / AEUtils.ZOOM_RATIO);
        }


    /**Adjusts East/North offsets given x,y coordinates of mouse click on displayed area.
     * The x is allowed to range from [0,DISPLAY_2D_EARTH_MAP_WIDTH[
     * and y from [0,DISPLAY_2D_EARTH_MAP_HEIGHT[
     * with the effect of moving the centre offset of the image
     * to be as close as possible to the source image pixel
     * representing the displayed map-fragment pixel clicked on.
     * <p>
     * This assumes that the displayed map fragment has a constrained center
     * and width/height.
     * <p>
     * Coerces values to allowed range if need be.
     */
    public void adjustCentreWithXYClick(int x, int y)
        {
        if(x < 0) { x = 0; }
        else if(x >= AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH)
            { x = AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH - 1; }
        if(y < 0) { y = 0; }
        else if(y >= AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT)
            { y = AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT - 1; }

        final Rectangle rect = getSourceRectangleToDisplay();

        setEastOffset(getEastOffset(true) + Math.round(rect.width *
            ((((float) x) / (AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH - 1)) - .5f)));
        setNorthOffset(getNorthOffset(true) - Math.round(rect.height *
            ((((float) y) / (AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT - 1)) - .5f)));
        }

    /**Computes a negated zoom factor rebased to the minimum zoom.
     * This is used to compute the actual source image width and height
     * from which the displayed image is extracted and zoomed to size.
     *
     * @return non-positive factor
     */
    private int _negativeZoomFactor()
        {
        return(-(getZoomFactor() - AEUtils.INITIAL_ZOOM));
        }

    /**Prefix for (first) path component for embedded parameters, in the order prefix/e/n/z. */
    private static final String PARAMS_PREFIX = "coords";
    
    /**If true then the PARAMS_PREFIX will not be visible in the path info. */
    private static final boolean PARAMS_PREFIX_NOT_IN_PATHINFO = true;

    /**Sets all parameters possible from a servlet request.
     * Arguments are parsed and coerced to correct values
     * or ignored if missing or badly broken,
     * leaving previous values in place.
     * <p>
     * Where necessary, arguments are coerced to form a consistent set.
     * <p>
     * This coercion may depend on data gathered from the current
     * exhibit data set and base clickable-map image can cached statically
     * where we can get at it.
     * <p>
     * This first sets the basic values from their simple parameter values,
     * the applies any adjustment implied by such things as the zoom control
     * and/or the map (x,y) click location.
     *
     * @return reference to this modified object
     */
    public AEParams setRequest(final HttpServletRequest request)
        {
        // Properties are tried after parsing components of the path,
        // so properties win in case of conflict,
        // and in particular to allow new params to override extant path.
        final String pathInfo = request.getPathInfo();
        if(pathInfo != null)
            {
//System.out.println("AEParams.setRequest(): pathInfo="+pathInfo+", query="+request.getQueryString());
            final StringTokenizer st = new StringTokenizer(pathInfo, "/");
            if(PARAMS_PREFIX_NOT_IN_PATHINFO ||
               (st.hasMoreTokens() && PARAMS_PREFIX.equals(st.nextToken())))
                {
                if(st.hasMoreTokens()) { setEastOffset(st.nextToken()); }
                if(st.hasMoreTokens()) { setNorthOffset(st.nextToken()); }
                if(st.hasMoreTokens()) { setZoomFactor(st.nextToken()); }
                }
            }
        // Override with form params if present.
        setEastOffset(request.getParameter(NAME_EAST_OFFSET));
        setNorthOffset(request.getParameter(NAME_NORTH_OFFSET));
        setZoomFactor(request.getParameter(NAME_ZOOM_FACTOR));

        // If there is a zoom-control parameter (makes no sense in path),
        // apply it after basic settings have been applied.
        final String zc = request.getParameter(NAME_ZOOM_CONTROL_FIELD);
        if(zc != null)
            {
            // Look for +/- adjustment...
            if("-".equals(zc)) { setZoomFactor(getZoomFactor() - 1); }
            else if("+".equals(zc)) { setZoomFactor(getZoomFactor() + 1); }
            // Else try to interpret as rebased zoom factor (1 == MIN).
            else
                {
                try { setZoomFactor(Integer.parseInt(zc, 10) + AEUtils.MIN_ZOOM - 1); }
                catch(NumberFormatException e) { } // Ignore unparseable value.
                }
            }

        // If there are (both) map-click (x,y) parameters are set and usable,
        // apply after basic settings have been applied.
        try
            {
            final int x = Integer.parseInt(request.getParameter(NAME_MAP_FIELD + ".x"), 10);
            final int y = Integer.parseInt(request.getParameter(NAME_MAP_FIELD + ".y"), 10);

            // Only try to use the values if in range...
            if((x >= 0) && (y >= 0) &&
                (x < AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH) &&
                (y < AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT))
                { adjustCentreWithXYClick(x, y); }

            // Now adjust the zoom.
            setZoomFactor(getZoomFactor() + MAP_CLICK_ZF_ADJ);
            }
        catch(final NumberFormatException e) { } // No usable map-click parameters.

//System.out.println("AEParams="+this);
        return(this);
        }

    /**Set the best view to see the specified Location.
     * Used to choose the view for a given area,
     * eg in a link from a catalogue page.
     * <p>
     * This sets the (unconstrained) centre to be as close as possible
     * to the specified location's centre.
     * <p>
     * This zooms in as far as it can to still totally enclose the target area
     * in the putative map vewport.
     */
    public AEParams setLocation(final Location.Estd loc)
        {
        if(loc == null) { return(this); }

        // Set the centroid as close as we can.
        setEastOffset(Math.round((loc.getE().value.floatValue() / 360.0f) * AEUtils.BASE_2D_EARTH_MAP_WIDTH));
        setNorthOffset(Math.round((loc.getN().value.floatValue() / 180.0f) * AEUtils.BASE_2D_EARTH_MAP_HEIGHT));

        // Zoom in as far as possible to (roughly) contain the target area.
        setZoomFactor(AEUtils.MIN_ZOOM);
        do
            {
            // Make sure that putative zoomed-in (constrained) viewport
            // would still completely enclose target location.
            final int newZF = getZoomFactor() + 1;
            if(!makeClone().setZoomFactor(newZF).getLocation(true).containsArea(loc))
                { break; }

            // OK, new zoom would be fine, it seems, so do it...
            setZoomFactor(newZF);
            } while(getZoomFactor() < AEUtils.MAX_ZOOM);

        return(this);
        }

    /**Gets parameters suitable to tag on to a URL after a '?', eg for GET operation.
     * This is an easy way to build a link back to the page.
     * <p>
     * Ampersands ("&amp;") separating parameters are expanded to entities,
     * eg "x=3&amp;amp;y=2" for safety embedded in HTML/XML text.
     * <p>
     * Note that this presents unconstrained parameters to preserve visitor's
     * chosen offset as they zoom in and out.
     *
     * @param constrained  if true, values returned are constrained to legal
     *    (and possibly tile) boundaries, which may help browser caching
     *    for example
     */
    public String getAsGetURLParameters(final boolean constrained)
        {
        final StringBuilder result = new StringBuilder(32);
        result.append(NAME_NORTH_OFFSET).append('=').append(getNorthOffset(constrained));
            result.append("&amp;");
        result.append(NAME_EAST_OFFSET).append('=').append(getEastOffset(constrained));
            result.append("&amp;");
        result.append(NAME_ZOOM_FACTOR).append('=').append(getZoomFactor());
        return(result.toString());
        }

    /**Gets parameters suitable to tag on to end of servlet URI to make new path.
     * This is an easy way to build a link back to the page.
     * <p>
     * This is a URI suffix of the form "PREFIX/e/n/z"
     * with no leading or trailing '/'.
     * <p>
     * Note that this presents unconstrained parameters to preserve visitor's
     * chosen offset as they zoom in and out.
     *
     * @param constrained  if true, values returned are constrained to legal
     *    (and possibly tile) boundaries, which may help browser caching
     *    for example
     */
    public String getAsGetPathInfo(final boolean constrained)
        {
        final StringBuilder result = new StringBuilder(32);
        result.append(PARAMS_PREFIX);
        result.append('/').append(getEastOffset(constrained));
        result.append('/').append(getNorthOffset(constrained));
        result.append('/').append(getZoomFactor());
        return(result.toString());
        }

    /**Gets parameters as hidden fields to include in a form.
     * Note that this presents unconstrained parameters to preserve visitor's
     * chosen offset as they zoom in and out.
     */
    public String getAsHiddenFormElements()
        {
        final StringBuilder result = new StringBuilder(128);
        result.append("<input type=hidden name=").
                append(NAME_ZOOM_FACTOR).
            append(" value=").
                append(UploaderUtils.quoteHTMLArg(String.valueOf(getZoomFactor()))).
            append(">");
        result.append("<input type=hidden name=").
                append(NAME_EAST_OFFSET).
            append(" value=").
                append(UploaderUtils.quoteHTMLArg(String.valueOf(getEastOffset(false)))).
            append(">");
        result.append("<input type=hidden name=").
                append(NAME_NORTH_OFFSET).
            append(" value=").
                append(UploaderUtils.quoteHTMLArg(String.valueOf(getNorthOffset(false)))).
            append(">");
        return(result.toString());
        }

    /**Gets parameters as human-readable (albeit terse) String.
     * Do not attempt to automatically parse this:
     * the format may change without notice.
     * <p>
     * This includes key derived values too.
     */
    public String toString()
        {
        final StringBuilder sb = new StringBuilder(64);
        sb.append("AEParams:");
        sb.append("zf=").append(getZoomFactor()).append(',');
        sb.append("eo=").append(getEastOffset(false)).append(',');
        sb.append("no=").append(getNorthOffset(false)).append(';');
        // Add derived values.
//        sb.append("sx=").append(getEastOffsetSrcPixelsX(false)).append(',');
//        sb.append("sy=").append(getNorthOffsetSrcPixelsY(false)).append(',');
//        sb.append("sw=").append(getBaseSrcPixelsWidth(false)).append(',');
//        sb.append("sh=").append(getBaseSrcPixelsHeight(false));
        sb.append(getLocation(true));
        return(sb.toString());
        }

    /**Returns true if all values are as for initial state.
     * This can be used to check that it is as if a visitor has newly
     * arrived at the page with no parameters set, so the zoom, etc,
     * are left at their default values.
     */
    public boolean inDefaultState()
        {
        // Check zoom, etc.
        return((getZoomFactor() == AEUtils.INITIAL_ZOOM) &&
               (getEastOffset(false) == 0) &&
               (getNorthOffset(false) == 0));
        }

    /**Clone this object; all state is separated from the original.
     */
    public Object clone()
        {
        // All fields can be copied as is; no deep copy is needed.
        try {
            final Object result = super.clone();
            return(result);
            }
        catch(CloneNotSupportedException e)
            { throw new Error("internal error"); }
        }

    /**Make copy of this object; like clone() except no need to cast result.
     * This is intended to be used with an idiom such as:
     * <pre>
     *     makeCopy().setZoomFactor(newValue).getAsGetURLParameters()
     * </pre>
     * to get a URL fragment for a link to change the zoom level.
     */
    public AEParams makeClone()
        { return((AEParams) clone()); }

    /**Makes a clone and steps it by the requested number of tile units in each direction (East and North).
     * Useful utility for JSP/servlet.
     * <p>
     * This coerces offsets to constrained values,
     * which may help, for example, with cacheing by browsers and in the server,
     * especially when tiling is being enforced.
     */
    public AEParams cloneAndTranslate(final int eTiles, final int nTiles)
        {
        final AEParams result = makeClone();
        result.setEastOffset(result.getEastOffset(true) + eTiles * result.getTileSrcPixelsWidth());
        result.setEastOffset(result.getEastOffset(true));
        result.setNorthOffset(result.getNorthOffset(true) + nTiles * result.getTileSrcPixelsHeight());
        result.setNorthOffset(result.getNorthOffset(true));
        return(result);
        }

    /**Makes an immutable canonical key for use in a hashtable; never null.
     * This is most valuable when FORCE_TILE_POSITIONING is true.
     * <p>
     * The result is produced from the zoom and unconstrained East and North
     * offset values.  This is unconstrained to allow for different target
     * centroids even within a single view/tile.
     * <p>
     * The result supports equals() and hashCode but is otherwise
     * to be considered opaque.
     *
     * @param constrained  if true then the key is constrained to a
     *     coarser-grained grid of viewport coordinates
     */
    public Object makeKey(final boolean constrained)
        {
        // First we normalise the values to be canonical and non-negative.
        final int normZF = getZoomFactor() - AEUtils.MIN_ZOOM;
        final int normEO = getEastOffset(constrained) + OFFSET_LIMIT_EAST;
        final int normNO = getNorthOffset(constrained) + OFFSET_LIMIT_NORTH;

        // Then work out the multipliers given storage order
        // from most to least significant:
        //     zf, eo, no
        // and need to accommodate full range of value in each "column".
        final int multNO = 1;
        final int multEO = multNO * (1 + 2*OFFSET_LIMIT_NORTH);
        final int multZF = multEO * (1 + 2*OFFSET_LIMIT_EAST);

        final int result =
            (normZF * multZF) +
            (normEO * multEO) +
            (normNO * multNO);

        // Return fast, compact, immutable result supporting equals/hashCode.
        return(new Integer(result));
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = 3079086832160645496L;
    }
