/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.svrCore.props.GenProps;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.location.LocationUtils;
import org.hd.d.pg2k.webSvr.util.WebConsts;
import org.hd.d.pg2k.webSvr.util.WebUtils;

import javax.servlet.ServletContext;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.*;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 13-Sep-2003
 * Time: 21:46:23
 */

/**Support for Aloha Earth virtual site.
 */
public final class AEUtils
    {
//    /**Site-relative URL to "page" at which Aloha Earth map HTML is shown.
//     * This is used to allow the Aloha Earth page to easily refer
//     * to itself by name, for example as the target of a form submission.
//     */
//    public static final String AE_SELF_LRURL = "coords/";

    /**Root-relative URL to "page" at which Aloha Earth is shown.
     * This is used to allow the Aloha Earth page to easily refer
     * to itself by name, for example as the target of a form submission.
     */
    public static final String AE_SELF_RRURL =
        WebConsts.VIRTUAL_SITE_CUST_ALOHAEARTH_ROOT /* + AE_SELF_LRURL */;

    /**Root-relative URL or source/base 2D Earth image for clickable map.
     * This is assumed to be a smallish low-colour image
     * on which we can write red and white markings that will be visible
     * and which we convert back to a clear and small palette-based image
     * in a popular and widely-supported format such as GIF or PNG.
     */
    public static final String BASE_2D_EARTH_MAP_RRURL =
        WebConsts.VIRTUAL_SITE_CUST_ALOHAEARTH_ROOT + "Earth-whole-planet.gif";

    /**Width of base image (pixels). */
    public static final int BASE_2D_EARTH_MAP_WIDTH = 4320;

    /**Height of base image (pixels). */
    public static final int BASE_2D_EARTH_MAP_HEIGHT = 2160;


    /**Root-relative URL to base directory on which map image servlet is mounted.
     * This must match entry in web.xml.
     * <p>
     * Note that this contains a leading and trailing '/'.
     * <p>
     * Any URI under this point will be directed to the servlet,
     * so we can pick the "expected" suffix to avoid surprising broswers
     * with the MIME type of the downloaded image.
     * <p>
     * Because these generated images may have to change the exhibit set
     * or meta data may change it it not appropriate for this to be under
     * /_static.
     */
    public static final String MAPIMG_SERVLET_MOUNT_DIR =
        WebConsts.VIRTUAL_SITE_CUST_ALOHAEARTH_ROOT + "mapimg/";

   /**Type of clickable-map image that we dynamically generate.
     * Need not be the same as the raw base image,
     * as long as it can (efficiently) encode the modified image
     * in a byte-indexed format.
     * <p>
     * Good types to use are PNG and GIF (when the LZW patent expires).
     */
    static final int DYN_MAP_IMG_TYPE = ExhibitMIME.ET_PNG;

    /**ExhibitMIME entry for the chosen outout type. */
    static final ExhibitMIME.ExhibitTypeParameters ETP =
        ExhibitMIME.getParamsByType(DYN_MAP_IMG_TYPE);

    /**Full root-relative URL given for map image servlet.
     * The extension is chosen to match the MIME type.
     * <p>
     * We will append GET-style parameters to allow the appropriate
     * zoom and shift to be applied.
     */
    public static final String MAP_SERVLET_RRURL =
        MAPIMG_SERVLET_MOUNT_DIR + "Earth" + ETP.dotSuffixForInputFile;


    /**Zoom ratio: what factor we zoom in/out each time; at least 2.
     * Also determines size of lateral (up/down/left/right) step
     * as fraction of width of whole image as shown to the user.
     * <p>
     * A value of 2 or 3 is probably good; up to 8 is probably OK.
     */
    public static final int ZOOM_RATIO = 2;

    /**Minimum zoom ratio from initial base image.
     * A negative value means the image can be zoomed out (shrunk)
     * relative to the base image.
     * <p>
     * This is in terms of zooms in or out
     * (changes in each dimension) by ZOOM_RATIO.
     */
    public static int MIN_ZOOM = -3;

    /**Maximum zoom ratio from initial base image.
     * A positive value means the image can be zoomed in (expanded)
     * relative to the base image.  This will determine how many
     * pixels on a side a single source pixel can be expanded to.
     * <p>
     * This is in terms of zooms in or out
     * (changes in each dimension) by ZOOM_RATIO.
     * <p>
     * This must be no less than MIN_ZOOM,
     * and is usually positive.
     * <p>
     * Since this effectively limits the precision with which a user can
     * select an area to view on the one hand, and yet increases the
     * 'chunkiness' of the maximally zoomed image, is is difficult to chose
     * an optimal value, but one that expand source pixels to no more
     * than about 8--16 pixels on a side is probably reasonable.
     */
    public static int MAX_ZOOM = +5;

    /**The initial zoom ratio as seen by a visitor.
     * The raw image is zoomed by this amount and this determines
     * the size of the derived clickable map to be shown to the visitor
     * at all levels of zoom.
     * <p>
     * This is usually but not always the same as the MIN_ZOOM,
     * but must be in the inclusive range
     * [MIN_ZOOM, MAX_ZOOM].
     */
    public static int INITIAL_ZOOM = MIN_ZOOM;

    /**Apply zoom to value in pixels.
     * A negative value reduces the value (brings it towards zero)
     * and a positive one increases it;
     * zero returns the value unchanged.
     * <p>
     * This computes input * (ZOOM_RATIO ^ zoomFactor).
     */
    public static int applyZoom(int pixels, int zoomFactor)
        {
        // A zoom factor of zero is the identity function.
        if(zoomFactor == 0) { return(pixels); }

        if(zoomFactor > 0)
            {
            // Special case for power of two.
            if(ZOOM_RATIO == 2)
                { return(pixels << zoomFactor); }

            // General case.
            while(--zoomFactor >= 0) { pixels *= ZOOM_RATIO; }
            return(pixels);
            }


        // Else zoomFactor < 0.

        // Special case for power of two.
        if(ZOOM_RATIO == 2)
            { return(pixels >> -zoomFactor); }

        // Avoid repeated division for speed
        // and to try to avoid unnecessary loss of information/accuracy.
        int divisor = 1;
        while(++zoomFactor <= 0) { divisor *= ZOOM_RATIO; }
        pixels /= divisor;
        return(pixels);
        }

    /**Display width of image to show visitor.
     * Determined by original image width and INITIAL_ZOOM.
     */
    public static final int DISPLAY_2D_EARTH_MAP_WIDTH =
        applyZoom(BASE_2D_EARTH_MAP_WIDTH, INITIAL_ZOOM);

    /**Display height of image to show visitor.
     * Determined by original image height and INITIAL_ZOOM.
     */
    public static final int DISPLAY_2D_EARTH_MAP_HEIGHT =
        applyZoom(BASE_2D_EARTH_MAP_HEIGHT, INITIAL_ZOOM);


    /**Make scaled base image fragment given Rectangle bounding source pixels.
     * Returns writable ARGB true colour image fragment suitable to draw on.
     * <p>
     * May consume significant resources (CPU and memory),
     * so access and concurrency should be controlled.
     *
     * @param srcRect  as returned by AEParams.getSourceRectangleToDisplay
     * @param context  servlet context to retrieve base image from
     */
    static BufferedImage makeScaledBaseMapFragment(final Rectangle srcRect,
                                                   final ServletContext context)
        {
        final BufferedImage imageIn =
            WebUtils.getAndCacheStaticImage(false, // We'll copy below.
                BASE_2D_EARTH_MAP_RRURL,
                false, // Don't force to memory-hogging ARGB true-colour here!
                context,
                false); // Cache source image indefinitely.

//System.out.println("makeScaledBaseMapFragment("+srcRect+", ctxt)");

        // Extract the map image fragment that we will be copying from.
        // Avoid copying any data...
        final BufferedImage srcFragment = imageIn.getSubimage(
            srcRect.x, srcRect.y, srcRect.width, srcRect.height);

        // Make a scaled copy of the original image.
        // We use nearest-neighbour scaling because:
        //    * It is fast.
        //    * It should work for indexed-colour (and other) images
        //      (eg this does not create pixels with new colours).
        final AffineTransformOp op = new AffineTransformOp(
            AffineTransform.getScaleInstance(
                DISPLAY_2D_EARTH_MAP_WIDTH / (double) srcRect.width,
                DISPLAY_2D_EARTH_MAP_HEIGHT / (double) srcRect.height),
            AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
        final BufferedImage scaledImage = op.filter(srcFragment, null);

        // Convert to ARGB true-colour to be easily drawn on, if necessary.
        final BufferedImage result = ImageUtils.convertToTrueColourARGB(scaledImage, false);

        return(result);
        }


    /**Colour of marker to show location of exhibit on map, in RGB format.
     * Since the map is usually rendered in green (land) and blue (sea)
     * this should probably be red or some other high-contrast colour.
     * This should probably also be fully opaque to minimise encoded image size
     * (eg 0xffff0000 for opaque red).
     */
    static final int EXHIBIT_MARKER_RGB_COLOUR =
        LocationUtils.DYN_LOC_TN_CROSSHAIR_RGB_COLOUR;

    /**Colour of marker to show area indicated by title, in RGB format.
     * Since the map is usually rendered in green (land) and blue (sea)
     * and individual exhibits in red
     * this should probably be orange or some other nearish-red colour.
     * This should probably also be fully opaque to minimise encoded image size
     * (eg 0xffff0000 for opaque red).
     */
    static final int TITLE_AREA_MARKER_RGB_COLOUR = 0xffE47833;



    /**Category whose LocationMap entries are used by getViewLocationMapVirtualPrefix(). */
    private static final String _GVT_CATEGORY = "places-and-sights";
    // Lookup prefix.
    private static final String _GVT_CATEGORY_LU_PREFIX = _GVT_CATEGORY + '/';

    /**This gets a virtual prefix from the LocationMap for the current view or empty String if none suitable; never null.
     * This returns the prefix the best possible match for the current view,
     * or returns "" if it cannot find anything suitable.
     * The prefix returned does contain the category,
     * and is of the form dir/word-{word-}*,
     * eg "places-and-sights/England-London-".
     * <p>
     * This should always return "" at the outermost level.
     * <p>
     * This looks up places-and-sights prefixes in the given LocationMap,
     * assuming that each such prefix is a reasonable region.
     * This tries a number of methods to find the "best" match,
     * in this order (excluding any "whole Earth" match:
     * <ol>
     * <li>The smallest location prefix wholly contained in the current view
     *     <em>that contains the user's selected target centroid</a>.
     * <li>The largest location prefix wholly contained in the current view.
     * <li>The smallest such Estd entry that wholly contains the current view.
     * <li>The smallest such Estd entry whose centre is in the current view.
     * <ol>
     *
     * @param aeps  parameters with which map to display is chosen; not null
     * @param lm  map to do lookup in; not null
     *
     * @return virtual prefix lookup into
     *     LocationMap.getMapFromNamePrefixToLocation() or "" if none
     */
    public static String getViewLocationMapVirtualPrefix(final AEParams aeps,
                                                         final LocationMap lm)
        {
        // Optimisation: at max zoom just return empty title immediately.
        if(aeps.getZoomFactor() == MIN_ZOOM)
            { return(""); }

        // Get prefix map.
        final Map<String,Location.Base> prefixMap = lm.getMapFromNamePrefixToLocation();
        // Get current view area as seen on the screen and as requested...
        final Location.Estd viewArea = aeps.getLocation(true);
        final Location.Estd viewAreaUnconstrained = aeps.getLocation(false);

        // Current best result.
        String result = null;
        // Area of current best result.
        Location.Estd resultArea = null;


        // Look for smallest contained prefix/area, if any,
        // CONTAINING THE CURRENT (UNCONSTRAINED) TARGET CENTROID.
        for(final Iterator it = prefixMap.keySet().iterator(); it.hasNext(); )
            {
            final String prefix = (String) it.next();
            final Location.Estd area = (Location.Estd) prefixMap.get(prefix);
            // Screen for usability.
            if(!_screenForGetViewTitle(prefix, area)) { continue; }

            // If area not completely contained by view, skip it.
            if(!viewArea.containsArea(area)) { continue; }

            // If this area does not contain the target centroid, skip it.
            if(!area.containsCentre(viewAreaUnconstrained)) { continue; }

            // If this is smaller than the previous best,
            // or there is no previous best,
            // then take it.
            if((resultArea == null) || resultArea.isLargerThan(area))
                {
                result = prefix;
                resultArea = area;
                }
            }
        // If we found an enclosed area, return prefix.
        if(result != null) { return(result); }


        // Look for largest contained prefix/area, if any.
        for(final Iterator it = prefixMap.keySet().iterator(); it.hasNext(); )
            {
            final String prefix = (String) it.next();
            final Location.Estd area = (Location.Estd) prefixMap.get(prefix);
            // Screen for usability.
            if(!_screenForGetViewTitle(prefix, area)) { continue; }

            // If area not completely contained by view, skip it.
            if(!viewArea.containsArea(area)) { continue; }

            // If this is larger than the previous best,
            // or there is no previous best,
            // then take it.
            if((resultArea == null) || area.isLargerThan(resultArea))
                {
                result = prefix;
                resultArea = area;
                }
            }
        // If we found an enclosed area, return prefix.
        if(result != null) { return(result); }


        // Look for smallest enclosing prefix/area, if any.
        for(final Iterator it = prefixMap.keySet().iterator(); it.hasNext(); )
            {
            final String prefix = (String) it.next();
            final Location.Estd area = (Location.Estd) prefixMap.get(prefix);
            // Screen for usability.
            if(!_screenForGetViewTitle(prefix, area)) { continue; }

            // If area does not completely contain view, skip it.
            if(!area.containsArea(viewArea)) { continue; }

            // If this is smaller than the previous best,
            // or there is no previous best,
            // then take it.
            if((resultArea == null) || resultArea.isLargerThan(area))
                {
                result = prefix;
                resultArea = area;
                }
            }
        // If we found an enclosing area, return prefix.
        if(result != null) { return(result); }


        // Look for smallest prefix/area whose centre is in view, if any.
        for(final Iterator it = prefixMap.keySet().iterator(); it.hasNext(); )
            {
            final String prefix = (String) it.next();
            final Location.Estd area = (Location.Estd) prefixMap.get(prefix);
            // Screen for usability.
            if(!_screenForGetViewTitle(prefix, area)) { continue; }

            // If view does not contain centre of area, skip it.
            if(!viewArea.containsCentre(area)) { continue; }

            // If this is smaller than the previous best,
            // or there is no previous best,
            // then take it.
            if((resultArea == null) || resultArea.isLargerThan(area))
                {
                result = prefix;
                resultArea = area;
                }
            }
        // If we found an enclosing area, return prefix.
        if(result != null) { return(result); }


        return(""); // Nothing found.
        }

    /**Screens entry from Location map; returns true if usable.
     * Returns true if:
     * <ul>
     * <li>Name indicates a places-and-sights entry.
     * <li>The area is smaller than whole Earth area.
     * </ul>
     */
    private static boolean _screenForGetViewTitle(String prefix,
                                                  Location.Estd area)
        {
        // If not a places-and-sights prefix, ignore it.
        if(!prefix.startsWith(_GVT_CATEGORY_LU_PREFIX)) { return(false); }

        // Explicitly reject items that have area of whole Earth,
        // ie don't carry much information as labels.
        if((area.getN().error.intValue() >= Location.Estd.MAX_N) &&
           (area.getE().error.intValue() >= Location.Estd.MAX_E))
            { return(false); }

        return(true); // Seems OK.
        }


    /**True if labels run horizontally, false if vertical. */
    public static final boolean LABEL_HORIZ = true;

    /**Label clearance (ie approximately the typical text height or 1em) in pixels; strictly positive. */
    public static final int LABEL_CLEARANCE_PX = 20;

    /**Overlay thumbnail clearance is longest (small) thumbnail dimension. */
    public static final int LABEL_TN_CLEARANCE_PX = ExhibitThumbnails.SML_STATIC_IMAGE_TN_LDIM_PX;

    /**Maximum number of labels to show; strictly positive.
     * This could be partially determined by the map image size,
     * ie how many labels can be confortably shown.
     */
    public static final int LABEL_COUNT_MAX = Math.max(5, Math.min(20,
        (LABEL_HORIZ ? DISPLAY_2D_EARTH_MAP_HEIGHT : DISPLAY_2D_EARTH_MAP_WIDTH) /
                                                   (2 * LABEL_CLEARANCE_PX)));

    /**Gets locationMap keys for labels to add to the map; result may be empty but is never null.
     * More important labels are earlier in the list.
     * <p>
     * We always try to include the title label, if available.
     * <p>
     * We then count how many uses of each prefix there are in this view,
     * and essentially return the top LABEL_COUNT_MAX of them,
     * but exclude any that are too close to any already selected
     * (and maybe those that are too near an edge to be read properly).
     * <p>
     * We are only interested in items wholly contained in the current view.
     *
     * @return best-first list of prefixes thumbnails; never null
     */
    public static List<String> getViewLocationMapLabelPrefixes(
                                final AEParams aeps,
                                final LocationMap lm,
                                final AlohaEarthMapCache cache)
        {
        if((aeps == null) || (lm == null) || (cache == null))
            { throw new IllegalArgumentException(); }

        final long startTime = System.currentTimeMillis();

        final ArrayList<String> result = new ArrayList<String>(LABEL_COUNT_MAX);

        // Return the heading prefix/label first if available
        // and regardless of its position on the map.
        final String headingPrefix = getViewLocationMapVirtualPrefix(aeps, lm);
        if(!"".equals(headingPrefix)) { result.add(headingPrefix); }

        final Location.Estd currentView = aeps.getLocation(true);

        // Count the number of exhibits in this view with each location.
        // The keys are Location.Estd instances which, for this to work,
        // we assume to have useful equality and/or have been intern()ed.
        // The values are instances of int[1] as counters.
        // When done we try to map back from Location to prefix.
        final Map<Location.Estd, int[]> counts = new HashMap<Location.Estd, int[]>();
        final List<String> exhibits = cache.selectContainedExhibits(aeps);
        for(final String ex : exhibits)
            {
            final Location.Base location = cache.aep.getLocation(ex);
            if(!(location instanceof Location.Estd)) { continue; }
            final Location.Estd locE = (Location.Estd) location;

            // Ignore items not enclosed in the current viewport.
            if(!currentView.containsArea(locE)) { continue; }

            int[] count = counts.get(locE);
            if(count == null)
                {
                count = new int[1];
                counts.put(locE, count);
                }
            ++count[0];
            }

if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): counts.size(): "+(counts.size())+".]"); }

        // Get list of all the Locations found.
        final ArrayList<Location.Estd> locations = new ArrayList<Location.Estd>(counts.keySet());
        // Sort into order by count, highest count first.
        Collections.sort(locations, new Comparator<Location.Estd>()
            {
            public final int compare(final Location.Estd estd1, final Location.Estd estd2)
                {
                final int count1 = counts.get(estd1)[0];
                final int count2 = counts.get(estd2)[0];
                if(count1 > count2) { return(-1); } // Correct order.
                if(count1 < count2) { return(+1); } // Wrong order.
                return(0);
                }
            });

        // We will use the location lookup maps both ways...
        final Map<String,Location.Base> mapFromNamePrefixToLocation = lm.getMapFromNamePrefixToLocation();
        final Map<Location.Base,String> mapFromLocationToNamePrefix = lm.getMapFromLocationToNamePrefix();

        // Go through locations in descending-count order,
        // picking those that do not clash with existing items,
        // until we hit our label-count limit.
        vetLocations: for(final Location.Estd l : locations)
            {
            if(result.size() >= LABEL_COUNT_MAX) { break; }

            final String prefix = mapFromLocationToNamePrefix.get(l);
            if(prefix == null)
                {
if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): no prefix for location "+l+", COUNT="+(counts.get(l)[0])+".]"); }
                continue;
                }

            // Avoid duplicates, eg of the title prefix.
            if(result.contains(prefix))
                {
//if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): already have prefix "+prefix+".]"); }
                continue;
                }

            final Point displayPixel = aeps.getDisplayPixelForEstdLocationCentre(l);

//            // Avoid anything too close to any edge...
//            if(displayPixel == null)
//                {
//if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): cannot compute display pixel for location "+l+".]"); }
//                continue;
//                }
//            if((displayPixel.x < LABEL_CLEARANCE_PX) ||
//               (displayPixel.y < LABEL_CLEARANCE_PX) ||
//               (displayPixel.x >= DISPLAY_2D_EARTH_MAP_WIDTH - LABEL_CLEARANCE_PX) ||
//               (displayPixel.y >= DISPLAY_2D_EARTH_MAP_HEIGHT - LABEL_CLEARANCE_PX))
//                {
////if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): too close to edge: "+prefix+".]"); }
//                continue;
//                }

            // Avoid too close vertically or horizontally to any extant result.
            // This avoids the labels colliding with whatever length/alignment,
            // whether written vertically or horizontally.
            for(final String r : result)
                {
                final Location.Base loc = mapFromNamePrefixToLocation.get(r);
                assert(loc instanceof Location.Estd);
                final Location.Estd locE = (Location.Estd) loc;
                final Point dp = aeps.getDisplayPixelForEstdLocationCentre(locE);
                assert(dp != null);
                if((Math.abs(dp.x - displayPixel.x) < LABEL_CLEARANCE_PX) ||
                   (Math.abs(dp.y - displayPixel.y) < LABEL_CLEARANCE_PX))
                    {
//if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): too close to exiting result: "+prefix+".]"); }
                    continue vetLocations;
                    }
                }

            // Looks OK to add this prefix to the results...
            result.add(prefix);
            }

if(IsDebug.isDebug) { System.out.println("[AEUtils.getViewLocationMapLabelPrefixes(): result.size(): "+(result.size())+", time="+(System.currentTimeMillis() - startTime)+"ms.]"); }

        return(result);
        }

    /**Preamble for label overlay HTML. */
    private static final String LABEL_HTML_PREAMBLE = "<ul id=\"labels\" style=\"list-style-type:none; position:absolute; left:0; top:0; width:0; height:0; margin:0; padding:0;\">";

    /**Postamble for label overlay HTML. */
    private static final String LABEL_HTML_POSTAMBLE = "</ul>";

    /**If true then do a "poor-man's drop-shadow" to help improve contrast. */
    private static final boolean LABEL_PMDS = true;

    /**Create HTML label text to overlay the map; returns "" if none, never null.
     * This creates an unnumbered list with one or more absolutely-positioned
     * (within the encompassing relatively positioned parent element).
     * If there are no labels to show then this does not create the UL element
     * and the result is "".
     * <p>
     * The first view labels may be rendered more strongly.
     */
    public static String getViewLocationLabels(final AEParams aeps,
                                               final DataSourceBean dsb,
                                               final AlohaEarthMapCache aemfb,
                                               final boolean doThumbnails)
        {
        final long startTime = System.currentTimeMillis();

        final LocationMap lm = aemfb.aep.epgi.getLocationMap();

        final List<String> prefixes = getViewLocationMapLabelPrefixes(aeps, lm, aemfb);
        if(prefixes.isEmpty()) { return(""); }

        final Map<String,Location.Base> mapFromNamePrefixToLocation = lm.getMapFromNamePrefixToLocation();
        final int capacity = LABEL_HTML_PREAMBLE.length() + LABEL_HTML_POSTAMBLE.length() +
                    (LABEL_PMDS?512:256)*(doThumbnails?2:1)*prefixes.size();
        final StringBuilder sb = new StringBuilder(capacity);

        final List<String> exhibitsForTNs = doThumbnails ? aemfb.selectContainedExhibits(aeps) : null;

        // Allow a little local computation time,
        // but not so much as to drive users mad, we hope...
        final long localStart = System.currentTimeMillis();
        final long stopBy = localStart + CoreConsts.MAX_INTERACTIVE_DELAY_MS +
                            (localStart - startTime); // Spend at least as long as our subroutines did...

        // True until we've gathered at least one requested thumbnail...
        boolean stillToGetThumbnail = doThumbnails;

        // Generate HTML for each label...
        int ordinal = 0;
        for(final String prefix : prefixes)
            {
            final Location.Base loc = mapFromNamePrefixToLocation.get(prefix);
            if(!(loc instanceof Location.Estd)) { continue; }
            final Location.Estd eloc = ((Location.Estd) loc);
            final Point p = aeps.getDisplayPixelForEstdLocationCentre(eloc);
            if(p == null) { continue; }

            // Emphasise first few labels...
            final boolean strong = (++ordinal <= 3);

            // Compute label link target.
            final String labelHref = AEUtils.AE_SELF_RRURL + (new AEParams()).setLocation(eloc).getAsGetPathInfo(false);

            // Offset label very slightly so as not to obscure the target.
            final int leftPX = p.x + 1;
            final int topPX = p.y + 1;
            // Use part after the '/' as the label,
            // and strip off the trailing '-' too.
            final String labelText = prefix.substring(1 + prefix.indexOf('/'),
                                                      prefix.length()-1);

            String imgHTML = null;
            if(doThumbnails &&
               (stillToGetThumbnail || (System.currentTimeMillis() <= stopBy)))
                {
                GenProps gp = new GenProps();
                try { gp = dsb.getGenProps(-1); } catch(final IOException e) { }

                // Best thumbnail candidiate found so far; null if none.
                String bestSoFar = null;
                // Score of best so far; initially neutral.
                int bestScore = 0;

                // Find (first) good (small) thumbnail image for each label.
                for(final String exhibitName : exhibitsForTNs)
                    {
                    // If this exhibit can't have a thumbnail then skip it.
                    if(!(ExhibitMIME.getInputFileType(aemfb.aep.aeid.getStaticAttr(exhibitName).filePath)).canPossiblyCreateThumbnailOfSameMIMEType())
                        { continue; }
//System.out.println("CAN HAVE TN: " + exhibitName);
                    // Skip if this exhibit does not have the right location.
                    if(!eloc.equals(aemfb.aep.getLocation(exhibitName)))
                        { continue; }
//System.out.println("CORRECT LOC: " + exhibitName);
                    // If this exhibit is "sensitive" then skip it.
                    if(GenUtils.isSensitive(exhibitName, aemfb.aep, gp))
                        { continue; }
                    // Skip if this exhibit is not notably good,
                    // or at least the best non-bad one so fdar!
                    // Don't necessarily force goodness value to be computed.
                    final boolean runningLate = System.currentTimeMillis() >= stopBy;
                    final ExhibitPropsComputableMutable epcm = aemfb.aep.getExhibitPropsComputableMutable(exhibitName, runningLate, gp, dsb, dsb.getScorerCache());
                    final int goodness = epcm.getGoodness();
                    // Skip anything not at least as that we already have...
                    if(goodness <= bestScore)
                        { continue; }
if(IsDebug.isDebug) { System.out.println("BEST SO FAR: " + exhibitName); }
                    final boolean isGood = Boolean.TRUE.equals(epcm.isGood());
                    // Only accept less-than-good if we have nothing at all yet.
                    if((bestSoFar != null) && !isGood)
                        { continue; }
if(IsDebug.isDebug) { System.out.println("IS GOOD: " + exhibitName); }
if(IsDebug.isDebug) { System.out.println("  RUNNING LATE: " + runningLate); }
                    // Skip if this exhibit does not have a small thumbnail.
                    // If we get here then it may well be worth trying hard,
                    // at least once, to create/fetch a missing thumbnail...
                    // Keep going until we have at least one.
                    final boolean createTN = stillToGetThumbnail || !runningLate;
                    final ExhibitThumbnails thumbnails = dsb.getThumbnails(exhibitName, createTN);
                    if((thumbnails == null) || (thumbnails.getSmall() == null))
                        { continue; }
if(IsDebug.isDebug) { System.out.println("HAS SMALL TN: " + exhibitName); }

                    // OK, this is the best so far!
                    bestSoFar = exhibitName;
                    bestScore = goodness;

                    // We seem to have a candidate...
                    // Create img tag for thumbnail...
                    final ExhibitPropsComputable epc = aemfb.aep.getExhibitPropsComputable(exhibitName);
                    final java.awt.Dimension xyDim = (epc == null) ? null : epc.getXyDimensions();
                    final java.awt.Dimension thumbnailXyDim = ExhibitThumbnails.computeThumbnailDimensions(xyDim, false);
                    final String rrURL = WebUtils.makeThumbnailRRURL(exhibitName, false);
                    final String mainWordsComponent = ExhibitName.getMainWordsComponent(exhibitName, ExhibitAttrUtils.getAttrWords().getAttrWordsSortedSet());
                    imgHTML = "<img src=\"" + rrURL + "\" border=\"0\" " +
                                ((thumbnailXyDim == null) ? "" :
                                    "width=\""+thumbnailXyDim.width+"\" height=\""+thumbnailXyDim.height+"\" " +
                                    "style=\"position:relative; left:-"+(thumbnailXyDim.width/2)+"px; top:-"+(thumbnailXyDim.height/2)+"px;\" ") +
                              "alt=\"" + mainWordsComponent + "\" " +
                              "title=\"" + mainWordsComponent + "\">";

if(IsDebug.isDebug) { System.out.println("FIRST TN FOR OVERLAY: " + stillToGetThumbnail); }
                    // We've done at least one thumbnail as requested.
                    stillToGetThumbnail = false;

                    // We can stop immediately if this is positively good,
                    // and if we're short of time,
                    // else we'll keep hunting for the "best" image.
                    if(isGood && runningLate)
                        { break; }
                    }
                }

            // If there's an image then put it in first, partly transparent,
            // so that we can overlay the label text on top of it.
            if(imgHTML != null)
                {
                sb.append("<li>");
                if(labelHref != null) { sb.append("<a href=\"").append(labelHref).append("\">"); }
                sb.append("<span style=\"opacity:").append(strong ? "0.9" : "0.7").append("; ").
                          append("position:absolute; ").
                          append("left:").append(leftPX).append("px; ").
                          append("top:").append(topPX).append("px; ");
                    sb.append("\">");
                sb.append(imgHTML);
                sb.append("</span>");
                if(labelHref != null) { sb.append("</a>"); }
                sb.append("</li>");
                }

            if(LABEL_PMDS)
                {
                // Create crude drop-shadow to help improve contrast.
                appendLabelHTML(sb, leftPX+1, topPX+1, strong, labelText, "black", labelHref);
                }

            // Main label.
            // Allow label text to use extant page colour scheme.
            appendLabelHTML(sb, leftPX, topPX, strong, labelText, null, labelHref);
            }

        if(sb.length() == 0)
            { return(""); }
        sb.insert(0, LABEL_HTML_PREAMBLE);
        sb.append(LABEL_HTML_POSTAMBLE);

if(IsDebug.isDebug && (sb.length() > capacity)) { System.err.println("WARNING: getViewLocationLabels(): length="+sb.length()+", capacity="+capacity); }

if(IsDebug.isDebug) { System.out.println("[getViewLocationLabels(): "+(System.currentTimeMillis()-startTime)+"ms.]"); }

        return(sb.toString());
        }

    /**Generate HTML for one overlay label and append to given StringBuilder.
     * The text label's content, colour, position and strength is chose.
     *
     * @param labelColour  colour for label text; null if default
     * @param labelHref  href for label text; null if none
     */
    private static void appendLabelHTML(final StringBuilder sb,
                                        final int leftPX,
                                        final int topPX,
                                        final boolean strong,
                                        final String labelText,
                                        final String labelColour,
                                        final String labelHref)
        {
        sb.append("<li>");
        if(labelHref != null) { sb.append("<a href=\"").append(labelHref).append("\">"); }
        if(strong) { sb.append("<big>"); }
        sb.append("<span style=\"white-space:nowrap; position:absolute; ").
                  append("left:").append(leftPX).append("px; ").
                  append("top:").append(topPX).append("px; ");
            if(labelColour != null) { sb.append("color:").append(labelColour).append(";"); }
            sb.append("\">");
        sb.append(labelText);
        sb.append("</span>");
        if(strong) { sb.append("</big>"); }
        if(labelHref != null) { sb.append("</a>"); }
        sb.append("</li>");
        }


    /**Key for AE-linked data in DataSourceBean.
     * Package-only to keep scope as narrow as possible!
     */
    static final DataSourceBean.AEPLinkedKey AEKey = new DataSourceBean.AEPLinkedKey("AlohaEarth");

    /**Get the filter bean associated with the current DataSourceBean; never null.
     * A new empty value is created on first use and when the AEP changes.
     */
    public static AlohaEarthMapCache getAemfb(final DataSourceBean dsb)
        throws IOException
        {
        if(dsb == null) { throw new IllegalArgumentException(); }

        AlohaEarthMapCache aemfb;
        while((aemfb = (AlohaEarthMapCache) dsb.getAEPLinkedValue(AEKey)) == null)
            { dsb.putIfAbsentAEPLinkedValue(AEKey, new AlohaEarthMapCache(dsb)); }

        return(aemfb);
        }
    }
