/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth;

import ORG.hd.d.IsDebug;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.location.Location;
import org.hd.d.pg2k.svrCore.location.LocationMap;
import org.hd.d.pg2k.webSvr.exhibit.BuiltInFilters;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;
import org.hd.d.pg2k.webSvr.exhibit.Expr;
import org.hd.d.pg2k.webSvr.exhibit.FilterExpr;

import javax.servlet.ServletContext;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.lang.ref.SoftReference;
import java.util.*;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 23-Sep-2003
 * Time: 21:39:33
 */

/**This class caches Estd-located exhibits and map fragments for Aloha Earth.
 */
public final class AlohaEarthMapCache
    {
    /**If true, use SoftReference memory-sensitive cache; use if memory-starved.
     * We will always cache the top couple of levels which are likely
     * to be small in number but relatively expensive to recompute.
     */
    private static final boolean MEMORY_SENSITIVE_CACHE = false;

    /**Empty exhibits list. */
    private static final List<String> NO_EXHIBITS = Collections.emptyList();

    /**Default constructor.
     * Sets the filter to be one on location (exhibits with Estd location),
     * and sets this not to be a memory-sensitive cache
     * (since we expect its lazy computation to take lots of CPU in total).
     * <p>
     * Only package-visible since only AlohaEarth utility methods need
     * to construct this.
     */
    AlohaEarthMapCache(final DataSourceBean dsb)
        throws IOException
        {
        if(dsb == null) { throw new IllegalArgumentException(); }

        aep = dsb.getAllExhibitProperties(-1);
        final String[] result = aep.select(new AllExhibitProperties.AEPFilter(){
            public final boolean accept(final AllExhibitProperties aep, final String exhibitName)
                {
                final Location.Base l = aep.getLocation(exhibitName);
                return(l instanceof Location.Estd);
                }
            }, null, 0);
        allEstdExhibits = Collections.unmodifiableList(Arrays.asList(result));
        }

    /**Underlying AEP; never null. */
    public final AllExhibitProperties aep;

    /**Unmodifiable smart-sorted list of all exhibits with Estd location; never null. */
    private final List<String> allEstdExhibits;

    /**If true, we attempt to cache images and exhibit sub-sets by area.
     * If false, everything is created on demand (which may be slow).
     * <p>
     * We only attempt to cache where the constrained offsets of AEParams
     * are forced to tile increments, ensuring that only a relatively
     * small number of distinct areas can be seen by the viewer,
     * reducing the maximum number of distinct cached items and increasing
     * the potential hit rate.
     */
    private static final boolean CACHEING = AEParams.FORCE_TILE_POSITIONING;

    /**If true, we attempt to select exhibits from the zoom-out set.
     * This is an optimisation to attempt to filter exhibits of an area
     * from the same offset but zoomed out one step rather than from the
     * base set of exhibits, which will probably save a lot of time
     * on all but the outermost view.
     * <p>
     * This only make sense if we are cacheing results of lookups,
     * and if we are constraining lookups to tile boundaries.
     */
    private static final boolean RECURSIVE_SELECT = CACHEING &&
        AEParams.FORCE_TILE_POSITIONING;

    /**If false, the entire set of exhibits selected by the master filter appears at top level; we may sort but we will not further filter.
     * This should save time for the critical top-level view by avoiding
     * performing a redundant second check that all values are Estd values,
     * at least with the default filter in place.
     */
    public static final boolean TOP_LEVEL_ESTD_FILTER = false;

    /**Routine to extract cache entry from given cache; null if no cached entry.
     * If not CACHEING then always return null,
     * else if raw cache entry is a SoftReference then it is de-referenced,
     * else the raw entry is returned (null if no entry).
     */
    private static Object _extractCacheEntry(final Map cache,
                                             final Object key)
        {
        // If not cacheing, don't even attempt a lookup.
        if(!CACHEING) { return(null); }

        // Extract the entry if any.
        final Object rawEntry = cache.get(key);

        // If a SoftReference (can only only true if memory-sensitive cache),
        // return de-referenced.
        if(MEMORY_SENSITIVE_CACHE && (rawEntry instanceof SoftReference))
            { return(((SoftReference) rawEntry).get()); }

        // Return the unvarnished entry.
        return(rawEntry);
        }

    /**Insert item in given cache, forcing non-SoftReference if need be.
     * If this not a memory-sensitive cache or if strongRef is true,
     * the item will be cached with a strong reference,
     * else a SoftReference will be used.
     * <p>
     * To avoid ambiguity,
     * the item to be cached must not itself be a SoftReference.
     * <p>
     * If not CACHEING, this does nothing.
     */
    @SuppressWarnings("unchecked")
    private static void _insertCacheEntry(final Map cache,
                                          final Object key,
                                          final Object item,
                                          final boolean strongRef)
        {
        if(!CACHEING) { return; }

        if(MEMORY_SENSITIVE_CACHE && !strongRef)
            { cache.put(key, new SoftReference(item)); }
        else
            { cache.put(key, item); }
        }

    /**Get selectViewLocationVirtualPrefix() value suitable for use as title; "" if none available. */
    public String selectViewLocationVirtualPrefixAsTitle(final AEParams aeps,
                                                         final ServletContext context)
        {
        final LocationMap lm = aep.epgi.getLocationMap();
        final String result = selectViewLocationVirtualPrefix(aeps, lm);
        if("".equals(result)) { return(""); }
        return(_extractMainWordPrefixFromVirtualPrefix(result));
        }

    /**Extract main-word-prefix portion from LocationMap virtual prefix.
     * Eg, from "x/y-z-" return "y-z-".
     */
    private static String _extractMainWordPrefixFromVirtualPrefix(final String result)
        { return(result.substring(result.indexOf('/') + 1)); }

    /**Private cache for selectViewLocationVirtualPrefix(); never null.
     * A map from AEParams-synthesised key to (possibly SoftReference to)
     * LocationMap lookup key for area for which title is generated.
     */
    private final Hashtable<String,Object> _sVLVP_cache = new Hashtable<String,Object>(103);

    /**Get best-match prefix of form "section/main-words-" from LocationMap for current view, or "" if none; never null.
     */
    public String selectViewLocationVirtualPrefix(final AEParams aeps,
                                                  final LocationMap lm)
        {
        // Create a lookup key (non-constrained).
        final Object key = aeps.makeKey(false);

        synchronized(_sVLVP_cache)
            {
            String result = (String) _extractCacheEntry(_sVLVP_cache, key);

            // If no cache entry or a cleared one,
            // compute and cache the filtered exhibit set.
            if(!CACHEING || (result == null))
                {
                // The result value is assumed already to be intern()ed.
                result = AEUtils.getViewLocationMapVirtualPrefix(aeps, lm);

                // Cache computed prefix.
                // Always keep top few levels which may be expensive to
                // compute and should consume little memory.
                _insertCacheEntry(_sVLVP_cache, key, result,
                      aeps.getZoomFactor() <= AEUtils.INITIAL_ZOOM + 3);
                }

            return(result);
            }
        }


    /**Private cache for selectViewLocationVirtualPrefix(); never null.
     * A map from AEParams-synthesised key to (possibly SoftReference to)
     * LocationMap lookup key for area for which title is generated.
     */
    private final Hashtable<String,Object> _sVLL_cache = new Hashtable<String,Object>(103);

    /**Create HTML label text to overlay the map; returns "" if none, never null.
     * This creates an unnumbered list with one or more absolutely-positioned
     * (within the encompassing relatively positioned parent element).
     * If there are no labels to show then this does not create the UL element
     * and the result is "".
     * <p>
     * The first view labels may be rendered more strongly.
     */
    public String selectViewLocationLabels(final DataSourceBean dsb,
                                           final AEParams aeps)
        {
        // Create a lookup key (unconstrained).
        final Object key = aeps.makeKey(false);

        synchronized(_sVLL_cache)
            {
            String result = (String) _extractCacheEntry(_sVLL_cache, key);

            // If no cache entry or a cleared one,
            // compute and cache the filtered exhibit set.
            if(!CACHEING || (result == null))
                {
                // Ask for overlay thumbnails when the zoom is positive
                // ie when the map is starting to be visibly pixelated.
                // The result value is probably worth intern()ing
                // since this value may be slow to compute and far from unique.
                result = MemoryTools.intern(AEUtils.getViewLocationLabels(aeps,
                                                                          dsb,
                                                                          this,
                                                    aeps.getZoomFactor() > 0));

                // Cache computed prefix.
                // Always keep top few levels
                // which may be expensive to compute
                // but which should consume relatively little memory.
                _insertCacheEntry(_sVLL_cache, key, result,
                      aeps.getZoomFactor() <= AEUtils.MIN_ZOOM + 2);
                }

            return(result);
            }
        }


    /**Private cache for selectContainedExhibits(); never null.
     * A map from AEParams-synthesised key to (possibly SoftReference to)
     * immutable String List of exhibits with centres in the current map view.
     * <p>
     * All access is synchronised on this object.
     * <p>
     * A Hashtable is used as inherently thread-safe.
     * <p>
     * Start this small; let it grow if needed.
     */
    private final Hashtable<String,Object> _sCE_cache = new Hashtable<String,Object>(137);

    /**Selects an immutable List of exhibits whose centres lie in the current map view; never null.
     * If we are using tile encoding,
     * then this caches its results via a SoftReference
     * based on the constrained/canonicalised AEParams value
     * set from the current request.
     * The cache is cleared when the exhibit set changes.
     * The very top result or results at the outermost zooms
     * (that a typical visitor is likely to encounter first)
     * may be cached with strong references to ensure that they
     * remain in cache for fast access.
     * <p>
     * If we are not using a tile encoding then this does not cache at all,
     * but just generates everything on demand.
     * <p>
     * This routine tries to avoid drawing each distinct point on the map
     * more than once to save time on the assumptions that:
     * <ul>
     * <li>The marker is idempotent, so redrawing does not add anything.
     * <li>Drawing is relatively expensive.
     * </ul>
     *
     * @param aeps  parameters with which map to display is chosen; not null
     * @return immutable List of full names of exhibits in current map view; not null
     */
    public List<String> selectContainedExhibits(final AEParams aeps)
        {
        // Create a lookup key.
        // This can be coarse-grained since it depends only on the viewport.
        final Object key = aeps.makeKey(true);

        synchronized(_sCE_cache)
            {
            List<String> result = (List<String>) _extractCacheEntry(_sCE_cache, key);

            // If no cache entry or a cleared one
            // then compute and cache the filtered exhibit set.
            if(!CACHEING || (result == null))
                {
                // Get String[] set of exhibits to filter.
                // Nominally the entire set filtered by the base filter bean.
                final String[] inputNames =
                    _getRawInputSetForFilter(aeps);

                // Filter for exhibits in the current map bounds.
                final Expr expr = _chooseLocationFilter(aeps);
                final String[] rawResult = (expr == null) ? inputNames :
                    expr.eval(aep, inputNames);
                // ...sort into human-friendly order...
                Arrays.sort(rawResult, ExhibitAttrUtils.getAttrWords().SMART_ORDER);
                // ...and prepare to cache an unmodifiable (and thread-safe) copy.
                // Eliminate duplicate empty lists (use single instance).
                result = (rawResult.length == 0) ?
                         NO_EXHIBITS :
                         Collections.unmodifiableList(Arrays.asList(rawResult));

                // Cache new List if appropriate.
                // Always keep top couple of levels which are expensive to
                // compute and should basically share a few Strings.
                _insertCacheEntry(_sCE_cache, key, result,
                      aeps.getZoomFactor() <= AEUtils.INITIAL_ZOOM + 1);
                }

            return(result);
            }
        }

    /**Returns the String[] set of exhibits to filter; never null.
     * This is nominally the entire set filtered by the base filter bean,
     * but may instead be a subset thereof from an area enclosing the
     * current view, eg from a zoomed-out view.
     *
     * @param aeps  parameters for current view of map; never null
     */
    private String[] _getRawInputSetForFilter(final AEParams aeps)
        {
//System.out.println("[_getRawInputSetForFilter("+aeps+")..]");

        List<String> inputSet = null;

        // Attempt to optimise by filtering from a smaller (super) set,
        // trying successively-more zoomed-out views,
        // though don't bother to use top-most, assumed-full, set.
        // If not possible, revert to filtering from entire top-level set.
        if(RECURSIVE_SELECT)
            {
            for(int zozf = aeps.getZoomFactor(); --zozf > AEUtils.MIN_ZOOM; )
                {
                final AEParams zoomedOut = aeps.makeClone().setZoomFactor(zozf);
                // Use of strictlyContainsArea() ensures:
                //   * The zoomed-out view is a superset of our current view.
                //   * That the recursion will terminate.
                final boolean canUseZoomedOutSuperset =
                    zoomedOut.getLocation(true).strictlyContainsArea(aeps.getLocation(true));
//System.out.println("  [_getRawInputSetForFilter("+aeps+"): able to use zoomed-out set: "+canUseZoomedOutSuperset+" at "+zoomedOut+".]");
                if(canUseZoomedOutSuperset)
                    {
                    inputSet = selectContainedExhibits(zoomedOut);
                    break; // OK, can stop looking.
                    }
                }
            }

        // If we dind't find a suitable sub-set to filter from
        // then use the full set.
        if(inputSet == null) { inputSet = allEstdExhibits; }

        final String[] inputNames = new String[inputSet.size()];
        inputSet.toArray(inputNames);

//System.out.println("[_getRawInputSetForFilter() returned length = "+inputNames.length+"]");
        return(inputNames);
        }

    /**Choose the filter to apply to the source data set of null if not filter is to be applied.
     * This may return null on the outermost zoom level to return all the
     * input data, else it will return a filter to select exhibits
     * within the current view area only.
     *
     * @param aeps  current properties (zoom, etc); never null
     * @return  filter expression
     */
    private static FilterExpr _chooseLocationFilter(final AEParams aeps)
        {
        if(!TOP_LEVEL_ESTD_FILTER &&
           (aeps.getZoomFactor() == AEUtils.MIN_ZOOM))
            { return(null); }

        return(new FilterExpr(null,
            new BuiltInFilters.filtByEstdLocationCentre(aeps.getLocation(true))));
        }


    /**Private cache for selectMapEncodedImage(); never null.
     * A map from AEParams-synthesised key to (possibly SoftReference to)
     * byte[] image.
     * <p>
     * All access is synchronised on this object.
     * <p>
     * This can be clear()ed by clearCache().
     * <p>
     * A Hashtable is used as inherently thread-safe.
     * <p>
     * Start this small; let it grow if needed.
     */
    private final Hashtable<String,Object> _sMEI_cache = new Hashtable<String,Object>(119);

    /**Create an encoded map fragment for display; never null.
     * If we are using tile encoding,
     * then this caches its images via a SoftReference
     * based on the constrained/canonicalised AEParams value
     * set from the current request.
     * The cache is cleared when the exhibit set changes.
     * The very top result or results at the outermost zooms
     * (that a typical visitor is likely to encounter first)
     * may be cached with strong references to ensure that they
     * remain in cache for fast access.
     * <p>
     * If we are not using a tile encoding then this does not cache at all,
     * but just generates everything on demand.
     * <p>
     * Access is synchronized (on a private lock) so as to:
     * <ul>
     * <li>Protect the cache if one is used.
     * <li>Limit peak resource consumption (CPU and memory).
     * </ul>
     * However, a different lock may be held while a new image is being
     * created to the normal cache-access lock in order that the creation
     * of new images does not block retrieval of already-generated ones.
     * <p>
     * Draws a (red) marker cross at the centre of every exhibit
     * within the current area,
     * and zero-or-more titles/labels to highlight the selected area.
     *
     * @param aeps  parameters with which map to display is chosen; not null
     */
    public ROByteArray selectMapEncodedImage(final AEParams aeps,
                                             final ServletContext context)
        throws IOException
        {
        // Create a lookup key.
        // Treat minimum zoom specially to get a single unique image
        // for memory efficiency and speed.
        final boolean zoomedOut = (aeps.getZoomFactor() == AEUtils.MIN_ZOOM);
        final Object key = aeps.makeKey(zoomedOut);

        synchronized(_sMEI_cache)
            {
            ROByteArray result = (ROByteArray) _extractCacheEntry(_sMEI_cache, key);

            // If no cache entry or a cleared one,
            // compute and cache the image.
            if(!CACHEING || (result == null))
                {
//System.out.println("Making map fragment for: " + aeps);

                // Get a raw zoomed fragment of the underlying map...
                final BufferedImage bi = AEUtils.makeScaledBaseMapFragment(
                    aeps.getSourceRectangleToDisplay(),
                    context);

                // Get the set of exhibits in this map area.
                final List<String> exhibits = selectContainedExhibits(aeps);

                // For each exhibit, compute its centre on this fragment,
                // and mark it on the image.
                _drawExhibitMarkers(exhibits, aep, aeps, bi);

                // Draw on the title marker, if any.
                // None at the outermost zoom.
                if(!zoomedOut)
                    { _drawTitleMarkers(aep.epgi.getLocationMap(), aeps, bi); }

                // Force to an indexed image with small colour palette
                // for on-the-wire efficiency.
                final BufferedImage crbi =
                    ImageUtils.makeColourReducedBufferedImage(bi,
                                                              32,
                                                              true);

                // Finally create the (compact) encoded binary form.
                // We throw away anything that is a duplicate,
                // since most views of one tile will be identical for example.
                result = MemoryTools.intern(
                    new ROByteArray(AEUtils.ETP.handler.makeImageBinary(crbi, Integer.MAX_VALUE)));

                // Cache new image if appropiate.
                // Always keep the initial zoom so user's first view is fast.
                _insertCacheEntry(_sMEI_cache, key, result, zoomedOut);
                }

            return(result);
            }
        }

    /**Draw the marker(s) corresponding to the title location, if any.
     * If there is no title then this may draw nothing.
     * <p>
     * This may draw crosshairs for the centroid.
     */
    private void _drawTitleMarkers(final LocationMap lm,
                                   final AEParams aeps,
                                   final BufferedImage bi)
        {
        assert(lm != null);
        final String prefix = selectViewLocationVirtualPrefix(aeps, lm);
        if("".equals(prefix)) { return; } // Nothing to draw.

if(IsDebug.isDebug) { System.out.println("[_drawTitleMarkers(): using prefix="+prefix+".]"); }


        // TODO: Draw crosshairs for unconstrained view centroid.
        final Location.Estd viewAreaUnconstrained = aeps.getLocation(false);
        final Point centroid =
            aeps.getDisplayPixelForEstdLocationCentre(viewAreaUnconstrained);
        if(centroid != null)
            {
            final int cx = centroid.x;
            final int cy = centroid.y;
            _drawVLine(bi, AEUtils.EXHIBIT_MARKER_RGB_COLOUR, cx, 0, AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT-1);
            _drawHLine(bi, AEUtils.EXHIBIT_MARKER_RGB_COLOUR, cy, 0, AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH-1);
            }


        // Get area used for the title;
        // if null, nothing to draw.
        final Location.Estd titleArea = (Location.Estd)
            lm.getMapFromNamePrefixToLocation().get(prefix);
        if(titleArea == null) { return; }

        final Point p =
            aeps.getDisplayPixelForEstdLocationCentre(titleArea);
        if(p != null)
            { bi.setRGB(p.x, p.y, AEUtils.TITLE_AREA_MARKER_RGB_COLOUR); }

        final float ev = titleArea.getE().value.floatValue();
        final float ee = titleArea.getE().error.floatValue();
        final int minX = aeps.computeXfromE(ev - ee);
        final int maxX = aeps.computeXfromE(ev + ee);

        final float nv = titleArea.getN().value.floatValue();
        final float ne = titleArea.getN().error.floatValue();
        final int minY = aeps.computeYfromN(nv + ne);
        final int maxY = aeps.computeYfromN(nv - ne);

        _drawVLine(bi, AEUtils.TITLE_AREA_MARKER_RGB_COLOUR, minX, minY-TITLE_AREA_EG, maxY+TITLE_AREA_EG);
        _drawVLine(bi, AEUtils.TITLE_AREA_MARKER_RGB_COLOUR, maxX, minY-TITLE_AREA_EG, maxY+TITLE_AREA_EG);

        _drawHLine(bi, AEUtils.TITLE_AREA_MARKER_RGB_COLOUR, minY, minX-TITLE_AREA_EG, maxX+TITLE_AREA_EG);
        _drawHLine(bi, AEUtils.TITLE_AREA_MARKER_RGB_COLOUR, maxY, minX-TITLE_AREA_EG, maxX+TITLE_AREA_EG);
        }

    /**Number of extra pixels to draw the marker in each direction; non-negative. */
    private static final int MARKER_EXTRA_PIXELS = 2;

    /**Pixel step for title area boundaries, 1 means solid; strictly positive. */
    private static final int TITLE_AREA_BOUNDARY_STEP = 2;

    /**Extra size of corner guides on marker title area in pixels, larger than exhibit marker for emphasis; strictly positive. */
    private static final int TITLE_AREA_EG = 2 * (MARKER_EXTRA_PIXELS + TITLE_AREA_BOUNDARY_STEP);

    /**Draw vertical line of given colour at x from minY to maxY.
     * This is careful not to attempt to draw pixels out of the image.
     */
    private static void _drawVLine(final BufferedImage bi,
                                   final int colour,
                                   final int x,
                                   int minY,
                                   int maxY)
        {
        if(minY > maxY)
            { throw new IllegalArgumentException(); }

        if((x < 0) || (x >= bi.getWidth()))
            { return; } // Entirely off the visible image; nothing to do.

        if(minY < 0) { minY = 0; }
        if(maxY >= bi.getHeight()) { maxY = bi.getHeight() - 1; }

        for(int y = minY; y <= maxY; y += TITLE_AREA_BOUNDARY_STEP)
            { bi.setRGB(x, y, colour); }
        }

    /**Draw horizontal line of given colour at y from minX to maxX.
     * This is careful not to attempt to draw pixels out of the image.
     */
    private static void _drawHLine(final BufferedImage bi,
                                   final int colour,
                                   final int y,
                                   int minX,
                                   int maxX)
        {
        if(minX > maxX)
            { throw new IllegalArgumentException(); }

        if((y < 0) || (y >= bi.getHeight()))
            { return; } // Entirely off the visible image; nothing to do.

        if(minX < 0) { minX = 0; }
        if(maxX >= bi.getWidth()) { maxX = bi.getWidth() - 1; }

        for(int x = minX; x <= maxX; x += TITLE_AREA_BOUNDARY_STEP)
            { bi.setRGB(x, y, colour); }
        }

    /**Draw the markers for individual exhibits.
     *
     * @param exhibits  List of exhibits; never null
     * @param aep  exhibit properties; nevr null
     * @param aeps  current view; never null
     * @param bi  BufferedImage to draw markers on; never null
     */
    static private void _drawExhibitMarkers(final List<String> exhibits,
                                            final AllExhibitProperties aep,
                                            final AEParams aeps,
                                            final BufferedImage bi)
        {
        /**Keep a collection of points already drawn to avoid redrawing them redundantly. */
        final HashSet<Point> pointsDrawn = new HashSet<Point>(exhibits.size() | 1);
        // Work through all the putative exhibtis.
        for(final Iterator<String> it = exhibits.iterator(); it.hasNext(); )
            {
            final String exhibitName = it.next();
            final Location.Base lb = aep.getLocation(exhibitName);
            if(!(lb instanceof Location.Estd)) { continue; }
            final Location.Estd le = (Location.Estd) lb;

            final Point p =
                aeps.getDisplayPixelForEstdLocationCentre(le);
//System.out.println("Exhibit marker at "+p+" for exhibit " + exhibitName);

            // Don't draw if already drawn.
            if(!pointsDrawn.add(p))
                {
//System.out.println("  [Avoided drawing redundant exhibit marker at "+p+" for exhibit " + exhibitName + " (" + pointsDrawn.size() + '/' + exhibits.size() + ").]");
                continue;
                }

            // Attempt to absorb (but report) any unexpected snafus.
            try {
                // Draw marker for exhibit.
                _drawExhibitMarker(p, bi);
                }
            catch(Exception e)
                {
                System.err.println("Problem drawing map for " + aeps);
                System.err.println("Target point " + p);
                System.err.println("Display W,H = " + AEUtils.DISPLAY_2D_EARTH_MAP_WIDTH + "," + AEUtils.DISPLAY_2D_EARTH_MAP_HEIGHT);
                System.err.println("Actual image W,H = " + bi.getWidth() + "," + bi.getHeight());
                e.printStackTrace();  //To change body of catch statement use Options | File Templates.
                }
            }
        }

    /**Draw marker on map image fragment to indicate location of an exhibit.
     * The point argument is itself guaranteed to be within the supplied
     * image area, but not all pixels adjacent to it are, ie it might be
     * on an edge.
     *
     * @param p  location to centre the exhibit market
     * @param bi  image to draw the marker on
     */
    private static void _drawExhibitMarker(final Point p, final BufferedImage bi)
        {
        if(p != null)
            {
            // Centre pixel is always within the image bounds.
            bi.setRGB(p.x, p.y, AEUtils.EXHIBIT_MARKER_RGB_COLOUR);
//System.out.println("Drawn exhibit marker centre at ("+p.x+","+p.y+").");

            // If not just drawing a point glyph, draw it on.
            if(MARKER_EXTRA_PIXELS > 0)
                {
                // Draw crosshairs.
                for(int yoff = MARKER_EXTRA_PIXELS + 1; --yoff > 0; )
                    {
                    final int yn = p.y - yoff;
                    if(yn >= 0) { bi.setRGB(p.x, yn, AEUtils.EXHIBIT_MARKER_RGB_COLOUR); }
                    final int yp = p.y + yoff;
                    if(yp < bi.getHeight()) { bi.setRGB(p.x, yp, AEUtils.EXHIBIT_MARKER_RGB_COLOUR); }
                    }
                for(int xoff = MARKER_EXTRA_PIXELS + 1; --xoff > 0; )
                    {
                    final int xn = p.x - xoff;
                    if(xn >= 0) { bi.setRGB(xn, p.y, AEUtils.EXHIBIT_MARKER_RGB_COLOUR); }
                    final int xp = p.x + xoff;
                    if(xp < bi.getWidth()) { bi.setRGB(xp, p.y, AEUtils.EXHIBIT_MARKER_RGB_COLOUR); }
                    }
                }
            }
        }
    }
