/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.virtualHosts.AlohaEarth;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.ROByteArray;
import org.hd.d.pg2k.webSvr.exhibit.DataSourceBean;

import com.sun.enterprise.deployment.Application;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStream;

/**Servlet to serve dynamic Aloha Earth map image.
 * This expects to be able to collect parameters with a AEParams object.
 * <p>
 * If the parameters are missing or out of bounds then a default image
 * is produced.
 * <p>
 * The image produced is with the type indicated by DYN_MAP_IMG_TYPE,
 * in an efficient (small), indexed format.
 * <p>
 * Images may be cached, possibly with SoftReference,
 * though any such cache may need to be cleared when the
 * exhibit set or meta data changes.
 */
public final class AlohaEarthMapServlet extends HttpServlet
    {
    /**Respond to a GET request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doGet(final HttpServletRequest request,
                      final HttpServletResponse response)
        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**Respond to a HEAD request for the content served by this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doHead(final HttpServletRequest request,
                       final HttpServletResponse response)
        throws IOException, ServletException
        {
        doAction(request, response);
        }

    /**Respond to a GET/HEAD request for the content served by this servlet.
     * Returns the exhibit data, with a correct MIME type.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doAction(final HttpServletRequest request,
                         final HttpServletResponse response)
        throws IOException, ServletException
        {
        try
            {
            // Get/create encoded image ready to send to the browser.
            final AEParams aeps = (new AEParams()).setRequest(request);
            final ServletContext ctxt = getServletContext();
            final DataSourceBean dsb = DataSourceBean.getApplicationInstance(ctxt);
            final ROByteArray encodedImage = AEUtils.getAemfb(dsb).
                selectMapEncodedImage(aeps, ctxt);

            // Set the correct MIME type and result length.
            response.setContentType(AEUtils.ETP.mimeType);
            response.setContentLength(encodedImage.length());

            // Set some cacheing headers too.
            // We let the image stay in cache for our standard `slackness'.
            // These images may change with the exhibit set.
            final int cacheLifetime = CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000;
            response.setHeader("Cache-Control", "max-age="+(cacheLifetime/1000));
            response.setDateHeader("Expires", System.currentTimeMillis() + cacheLifetime);

            // If this is a HEAD, then return without providing the body.
            if("HEAD".equalsIgnoreCase(request.getMethod()))
                { return; }

            // Write entire map-fragment image in one go for efficiency.
            final OutputStream os = response.getOutputStream();
            os.write(encodedImage.getData());
            os.flush();
            }
        catch(final Exception e)
            {
            // Log the problem!
            getServletContext().log("failed to generate Aloha Earth map image", e);
            // Redirect back to original raw image in case of error.
            // (This may force the browser to rescale the image)
            // If our result is already committed this may be messy...
            response.reset();
            response.resetBuffer();
            request.getRequestDispatcher(AEUtils.BASE_2D_EARTH_MAP_RRURL).
                forward(request, response);
            return;
            }
        }

    /**Have the last-modified date be the AEP unchanged-since timestamp. */
    @Override protected long getLastModified(final HttpServletRequest request)
        {
        try
            {
            final ServletContext ctxt = getServletContext();
            final DataSourceBean dsb = DataSourceBean.getApplicationInstance(ctxt);
            return(dsb.getAllExhibitProperties(-1).hashNotChangedSince);
            }
        catch(final IOException e)
            { return(-1); /* "Not known/available" value. */ }
        }

    /**Unique Serialisation class ID generated by http://random&#46;hd&#46;org/. */
    private static final long serialVersionUID = -6552601635962150780L;
    }
