/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.pg2k.webSvr.virtualHosts;

import org.hd.d.pg2k.svrCore.CoreConsts;
import org.hd.d.pg2k.svrCore.HostUtils;
import org.hd.d.pg2k.webSvr.util.WebConsts;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: Damon Hart-Davis
 * Date: 27-Jun-2003
 * Time: 18:08:17
 */

/**Contains details of the virtual hosts supported.
 * This also contains the current canonical list of skin names.
 */
public final class VirtualHosts
    {
    /**Skin name for black background, minimalist page style. */
    public static final String SKIN_BLACK_MINIMAL = "blackMin";

    /**Skin name for basic .mobi compliant site. */
    public static final String SKIN_MOBI_BASIC = "mobiBasic";

    /**Default virtual host description (ie keyed from main site name) for default livery, etc.
     * Default livery has normal navigation, advertising, layout, etc.
     */
    public static final VirtualHost VIRTUAL_HOST_DEFAULT =
        new VirtualHost(CoreConsts.MAIN_DATA_HOST, null, null,
                        new String[]{ // Deprecated aliases...
                            "data.pg2k.hd.org", "hd.org", "d.hd.org",
                        });

    /**Virtual host description and main name for SmlPx.
     * This is a basic .mobi-compliant version of the Gallery.
     */
    public static final VirtualHost VIRTUAL_HOST_SMLPX =
        new VirtualHost("smlpx.mobi",
                        WebConsts.VIRTUAL_SITE_CUST_SMLPX_ROOT,
                        SKIN_MOBI_BASIC,
                        new String[]{ // Aliases...
                            "smallpix.mobi"
                        });

    /**Virtual host description and main name for Aloha Earth.
     * Aloha Earth has a black background and minimalist furniture.
     */
    public static final VirtualHost VIRTUAL_HOST_ALOHAEARTH =
        new VirtualHost("aloha-earth.com",
                        WebConsts.VIRTUAL_SITE_CUST_ALOHAEARTH_ROOT,
                        SKIN_BLACK_MINIMAL,
                        new String[]{ // Deprecated aliases...
                            "aloha-earth.org", "alohaearth.org", "alohaearth.com"
                        });

    /**Virtual host description and main name for LondonPix.
     * LondonPix has a black background and minimalist furniture.
     */
    public static final VirtualHost VIRTUAL_HOST_LONDONPIX =
        new VirtualHost("londonpix.org.uk",
                        WebConsts.VIRTUAL_SITE_CUST_LONDONPIX_ROOT,
                        SKIN_BLACK_MINIMAL, null);

    /**Virtual host description and main name for ZAPix.
     * ZAPix has a black background and minimalist furniture.
     */
    public static final VirtualHost VIRTUAL_HOST_ZAPIX =
        new VirtualHost("pix.org.za",
                        WebConsts.VIRTUAL_SITE_CUST_ZAPIX_ROOT,
                        SKIN_BLACK_MINIMAL,
                        new String[]{ // Deprecated aliases.
                            "photo.org.za", "photos.org.za"
                        });

    /**Virtual host description and main name for localhost; primarily for testing.
     * Usually standard livery, but need not be so.
     */
    public static final VirtualHost VIRTUAL_HOST_LOCALHOST =
        new VirtualHost("localhost", null, null, null);

    /**Public (immutable) List of canonical "main" VirtualHost descriptions.
     * In principle all host names or site URIs should map to one of these,
     * and we may even force redirects and so on so that search engines
     * and users learn and use these canonical versions.
     * <p>
     * This is not intended for fast lookup,
     * and nothing is implied by the order of entries.
     * <p>
     * This should help:
     * <ul>
     * <li>Reduce complexity of PG2K management, eg in deployment descriptors.
     * <li>Reduce redundant Web spider activity to deprecated host names.
     * <li>Improve load balancing by making better use of "core" aliases.
     * </ul>
     */
    public static final List<VirtualHost>/*<VirtualHost>*/ canonicalVirtualHosts =
        Collections.unmodifiableList(Arrays.asList(new VirtualHost[]{
            VIRTUAL_HOST_DEFAULT,
            VIRTUAL_HOST_LONDONPIX,
            VIRTUAL_HOST_ZAPIX,
            VIRTUAL_HOST_SMLPX,
            VIRTUAL_HOST_ALOHAEARTH,
            VIRTUAL_HOST_LOCALHOST,
        }));

    /**Public (immutable) Map from String normalised host name to VirtualHost.
     * This is built from the main and alias host names.
     * <p>
     * For the moment this is hard-wired in;
     * it may become loadable in GenProps in future.
     */
    public static final Map<String,VirtualHost> hostDetails;
    /**Initialise hostDetails. */
    static
        {
        // Create the underlying host map as a HashMap for speed.
        final Map<String,VirtualHost> hm = new HashMap<String, VirtualHost>(1 + 3 * canonicalVirtualHosts.size());

        // Make entries for all the canonical and alias names...
        for(final Iterator<VirtualHost> it = canonicalVirtualHosts.iterator(); it.hasNext(); )
            {
            final VirtualHost vh = it.next();
            if(hm.put(vh.normalisedName, vh) != null)
                { System.err.println("WARNING: duplicate virtual host name: " + vh.normalisedName); }
            for(final Iterator<String> it2 = vh.normalisedAliasNames.iterator(); it2.hasNext(); )
                {
                final String an = it2.next();
                if(hm.put(an, vh) != null)
                    { System.err.println("WARNING: duplicate virtual host name: " + an); }
                }
            }

        hostDetails = Collections.unmodifiableMap(hm);
        }

    /**Lookup VirtualHost details by hostname and request URL; null if no match found.
     * The hostname supplied is normalised before lookup.
     * <p>
     * Before a lookup by hostname is attempted,
     * a match is attempted on the request URI
     * iff the URI starts with the normal prefix for virtual sites.
     * If a virtual host root is a prefix of the request URI
     * then we force usage of the the virtual host livery.
     *
     * @param serverHostname  the host name with which the page was fetched,
     *     as from request.getServerName()
     * @param requestURI  the within-host URI as from request.getRequestURI(),
     *     or null for a faster lookup just by normalised name
     */
    public static VirtualHost getVirtualHostDetails(final String serverHostname,
                                                    final String requestURI)
        {
        // Try lookup on the URI first iff it looks like a virtual site.
        if((requestURI != null) &&
           (requestURI.startsWith(WebConsts.VIRTUAL_SITES_ROOT)))
            {
            // This is potentially a little slow,
            // and does not necessarily return the primary virtual alias.
            for(final Iterator<VirtualHost> it = hostDetails.values().iterator(); it.hasNext(); )
                {
                final VirtualHost vh = it.next();
                final String dir = vh.virtualRootDir;
                if((dir != null) && requestURI.startsWith(dir))
                    { return(vh); }
                }
            }

        // Try lookup on the virtual hostname.
        final VirtualHost result = hostDetails.get(
            HostUtils.normaliseVirtualHostName(serverHostname));
        return(result);
        }


    /**Details of one virtual host.
     * This supports hashCode() and equals() which work on the
     * (normalised) name alone.
     */
    public static final class VirtualHost
        {
        /**Construct the record of one virtual host.
         *
         * @param hostname  non-null name of the host; will be stored normalised
         * @param root  root URI of home directory starting and ending with '/';
         *     if null or "/" (stored as null) then normal home dir is to be used
         * @param skinName  name of livery or skin to be use on each page;
         *     if null or unrecognised then standard livery may be used
         * @param aliases  alias host names for the main host specified
         *     all values supplied will be normalised with duplicates removed;
         *     may be null or zero-length
         *     but entries may not be null or zero-length
         *
         * @throws IllegalArgumentException  if arguments are malformed
         */
        public VirtualHost(final String hostname,
                           final String root,
                           final String skinName,
                           final String[] aliases)
            {
            normalisedName = HostUtils.normaliseVirtualHostName(hostname);

            if((root == null) || root.equals("/"))
                { virtualRootDir = null; }
            else if(!root.startsWith("/") || !root.endsWith("/"))
                { throw new IllegalArgumentException(); }
            else
                { virtualRootDir = root; }

            final Set<String> normAliases = new HashSet<String>();
            if(aliases != null)
                {
                for(int i = aliases.length; --i >= 0; )
                    {
                    final String a = aliases[i];
                    if(a == null)
                        { throw new IllegalArgumentException("null alias not allowed"); }
                    final String na = HostUtils.normaliseVirtualHostName(a);
                    if(na.length() == 0)
                        { throw new IllegalArgumentException("zero-length (normalised) alias not allowed"); }

                    // Only add this alias if not the same as the main name.
                    if(!na.equals(normalisedName))
                        { normAliases.add(na); }
                    }
                }
            final Set<String> noAliases = Collections.emptySet();
            normalisedAliasNames = (normAliases.size() == 0) ? noAliases :
                Collections.unmodifiableSet(normAliases);

            skin = skinName;
            }

        /**Normalised (prefix-less, all-lower-case, no-final-dot) virtual host name; non-null. */
        public final String normalisedName;

        /**Root of virtual host home page in main hierarchy, or null.
         * Null if normal root (and normal home page) is used,
         * else starts and ends with '/' but is not "/" and gives position
         * of opening directory for virtual host.
         * <p>
         * The fact that this cannot be "/" makes it safe to do a redirect
         * without checking the value.
         */
        public final String virtualRootDir;

        /**Skin/livery name to be used; null if default is to be used. */
        public final String skin;

        /**Immutable Set of normalised String hostname aliases for this virtual host; never null.
         * May be empty.
         * <p>
         * Entries are assumed to be deprecated and visitors (and spiders)
         * arriving at these may be diverted to the normalisedName
         * to concentrate activity on it.
         */
        public final Set<String> normalisedAliasNames;

        /**Return the normalisedName for the String representation. */
        public String toString() { return(normalisedName); }
        }
    }
