/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.pg2k.svrCore.datasource;

import ORG.hd.d.IsDebug;

import org.hd.d.pg2k.ai.scorer.ScorerCacheIF;
import org.hd.d.pg2k.svrCore.*;
import org.hd.d.pg2k.svrCore.MIME.ExhibitMIME;
import org.hd.d.pg2k.svrCore.location.LoadBalancingUtils;
import org.hd.d.pg2k.svrCore.props.*;
import org.hd.d.pg2k.svrCore.stats.StatsLogger;
import org.hd.d.pg2k.svrCore.vars.*;

import java.io.*;
import java.lang.ref.SoftReference;
import java.nio.ByteBuffer;
import java.security.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.*;

/**Exhibit pipeline cache stage.
 * This performs transparent, persistent cacheing of exhibit data
 * and variables.
 * <p>
 * The presence of an instance of this stage upstream of a tunnel or
 * other potentially slow/expensive data source should,
 * for normal data access (eg sequential download of exhibits)
 * significantly reduce upstream bandwidth requirements
 * and reduce downstream latency by answering requests from local cache.
 * <p>
 * This also attempts to shelter its downstream callers/users from I/O errors
 * upstream, by fulfilling requests from cache,
 * or, when synchronous calls upstream have to be made,
 * transforming some requests and replies into less toxic forms where possible.
 * <p>
 * When caching exhibit data this class only does so as a continuous prefix
 * from offset zero; other (random) accesses may have to read-through the cache.
 * <p>
 * This cache is also able to pre-cache data likely to be valuable,
 * such as thumbnails and the initial portions of exhibits,
 * though this will only be attempted if the cache appears to be in use.
 * Bandwidth/resource consumption used by pre-cacheing are regulated.
 * This cache regards thumbnails and meta-data as precious,
 * and tries not to let them go once collected and cached
 * because reasonable application performance will often depend on
 * fast access to these data.
 * <p>
 * This attempts to cache data well enough that, especially if aggressive
 * (pre)cacheing is available and the cache area is large enough,
 * almost no reference should be needed to the backend server in response to
 * a data request on the pipeline except in response to the tail end
 * of very long exhibits; all requests are answered from the local
 * cache where possible.  Ideally, it should be possible for the backend
 * server to go down altogether and have the front end still provide a
 * high degree of functionality.  The front-end and back-end are almost
 * completely decoupled in this cache design.
 * <p>
 * This includes limited in-memory caching, in many cases using
 * soft references to allow peaceful coexistence with other (potentially
 * heavy) users of memory.
 * <p>
 * This class relies mainly on the asynchronous calls to poll() to fetch
 * meta-data updates such as GenProps and AllExhibitImmutableData.  These
 * happen under a private lock and do not block cache access much or at all.
 * <p>
 * The full lock ordering (where multiple locks need to be taken at once) is:
 * <ol>
 * <li> ExhibitDataSimpleCache.rwl
 * <li> ExhibitDataSimpleCache.metaData
 * </ol>
 * <p>
 * This cache may serialise access to raw exhibit data (and may serialise
 * accesses to back-end resources too).  No two live instances of this
 * class should refer to the same cache directory at once
 * else madness and corruption will almost certainly break out.
 * <p>
 * On disc, the files are some prefix of the full exhibit, retrieved if possible
 * in MAX_TRANSFER_CHUNK_SIZE chunks.  They are touched every time
 * accessed or updated and the timestamps can therefore be used as
 * the basis of an LRU cache.  We expect almost all access to be
 * sequential, starting at the beginning.
 * <p>
 * We statisfy requests wholy within already-cached data immediately,
 * and will extend (up to the limit) by up to one
 * chunk each time by downloading from the server to statisfy
 * requests just beyond the current end.  Requests starting well beyond
 * the current end of cache are punted directly to the server,
 * which is rather ugly and slow, but there we go.
 * <p>
 * This cache only considers general precaching until the low-water
 * mark is reached, and by default only deletes existing entries if it has to
 * in order to statisfy an incoming cacheable request.
 * This means that stale entries for deleted/renamed exhibits
 * may persist for a while, but this is mainly harmless.
 * <p>
 * This maintains a bi-directional cache of variable values and updates,
 * and also some running parameters of the cache may be read as variables.
 * <p>
 * This cache supports a limited amount of peer-to-peer (P2P) data transfers
 * to reduce load on the master.  The general policy is that any
 * synchronous (and thus presumably time-sensitive) data request from an end user
 * that cannot be satisfied from local cache is satisfied upstream from the master.
 * Asynchronous data fetches, such as read-ahead and pre-cache activity,
 * can be fetched P2P.  Also if the master fails or is unavailable
 * then it may be acceptable to use P2P if loops/cycles can be avoided.
 * <p>
 * TODO: re-analyse/reduce locking and possibly avoid lock on metaData object.
 */
public final class ExhibitDataSimpleCache implements SimpleExhibitPipelineIF
    {
    /**If true then we may purge cached data for exhibits that appear to have been deleted (or renamed).
     * These are exhibits that have been "orphaned" in the cache,
     * ie they are not accessible because they do not logically exist
     * (though they do potentially serve as a backup in case of disaster).
     * <p>
     * We will generally only do this if:
     * <ul>
     * <li>the exhibit data has not been used for a long time, and
     * <li>if cache size is above the low-water mark that would prevent us
     *     from doing useful precacheing.
     * </ul>
     * <p>
     * We do not rush to delete exhibits' data in case a transient problem
     * has made an exhibit disappear temporarily.
     * (Data for deleted exhibits will in any case eventually be deleted LRU
     * (Least-Recently-Used) if the cache becomes full.)
     */
    private static final boolean ORPHANED_EXHIBIT_EXPIRY_ALLOWED = true;

    /**The minimim time before we will pre-emptively purge orphaned cache entries (ms); strictly positive.
     * We don't expect to delete or rename exhibits very often,
     * and the only harm in NOT purging them may be to prevent precaching of new exhibits,
     * ie a minor performance issue rather than a correctness issue.
     * <p>
     * Add a random component so that all clients do not purge orphans at once!
     * <p>
     * A value of the order of a few days to a few months is probably reasonable.
     */
    private static final long ORPHANED_EXHIBIT_MIN_UNUSED_TIME_MS = 14 * 24 * 3600 * 1000L + // A fortnight.
        Rnd.fastRnd.nextInt(4 * 24 * 3600 * 1000);

    /**Base dir within cache dir for all our exhibit data. */
    private static final String CACHE_BASE_DIR = "_exhibitsCache";

    /**Name of file in which to persist immutable exhibit data. */
    private static final String CACHE_EXPROPS_FILENAME = CACHE_BASE_DIR + "/_ExProps.dat";

    /**Name of file in which to persist cache meta data. */
    private static final String CACHE_METADATA_FILENAME = CACHE_BASE_DIR + "/_metadata.dat";

    /**If true, store exprops data (and cache metadata) GZIPed to possibly save space and I/O time. */
    private static final boolean STORE_EXPROPS_GZIPED = true;

    /**Base dir within cache dir for all our raw exhibit content data. */
    private static final String CACHE_EXDATA_DIR = CACHE_BASE_DIR + "/exd";

    /**Base dir within cache dir for all our event history data. */
    private static final String EVENT_HISTORY_DIR = CACHE_BASE_DIR + "/history";

    /**The prefix for all aux files associated with an exhibit file. */
    private static final String CACHE_EXAUX_PREFIX = ".aux.";

    /**The keyword for the file containing the (decimal) exhibit timestamp. */
    private static final String CACHE_EXAUX_TIMESTAMP_KW = "ts";

    /**The keyword for the file containing the serialised thumbnails object. */
    private static final String CACHE_EXAUX_TN_KW = "tn";

    /**Maximum (normal) exhibit data transfer chunk size (bytes); strictly positive.
     * Maximum chunk transferred in one call (bytes),
     * to avoid creating huge gaps in other activity by
     * jamming up transactions and/or hogging all I/O bandwidth.
     * <p>
     * Should probably be of the order of a few tens of kBytes
     * to allow efficient transfers on the wire, and a power of
     * two to be more likely to interact efficiently with other
     * caches.
     * <p>
     * If we use (a small multiple of) the bulk data transfer chunk size
     * this will be reasonably efficient in terms of disc/network traffic,
     * and if we can also keep it aligned to whole chunk boundaries
     * then we may get maximally efficient accesses into disc (etc) data.
     */
    private static final int MAX_TRANSFER_CHUNK_SIZE = Math.max(1<<15,
        2 * CoreConsts.BULK_DATA_TRANSFER_SIZE);

    /**Maximum extended exhibit data transfer chunk size (bytes); strictly positive.
     * When asked for data but we have a small gap in our cache before the request start
     * normally we would have to pass the result upstream and not cache any.
     * This is potentially very wasteful in P2P sharing of new exhibits.
     * <p>
     * So in this case we may allow the upstream request window to be moved back
     * to patch up the hole from the end of our cached data to the start of the request,
     * thus allowing us to capture the result.
     * <p>
     * This value should be a small multiple of the MAX_TRANSFER_CHUNK_SIZE.
     */
    private static final int MAX_EXTD_TRANSFER_CHUNK_SIZE = MAX_TRANSFER_CHUNK_SIZE << 2;

    /**The maximum number of bytes we will force a transfer of to force an immediate thumbnail generation.
     * This might as well be at least one block,
     * and might as well be a little bigger than we anticipate the
     * resulting thumbnail size to be (limited to),
     * but no larger than the maximum single transfer that we can make.
     */
    private static final int MAX_REMOTE_FETCH_TO_MAKE_THUMBNAIL =
        Math.min(MAX_USER_READ_SIZE,
            Math.max(MAX_TRANSFER_CHUNK_SIZE,
                     ExhibitThumbnails.MAX_BYTES_EST));

    /**Minimum cache size to allow if GenProps is not set (bytes).
     * This prevents thrashing on an empty cache and should be enough
     * to store a few thumbnails and a few reasonable image chunks.
     */
    private static final int FALLBACK_MIN_CACHE_SIZE =
        Math.max(1234567, 8 * MAX_TRANSFER_CHUNK_SIZE) +
        (50 * ExhibitThumbnails.MAX_BYTES_EST);

    /**Maximum leading portion of exhibit cached in memory.
     * Because of the crude way we retrieve data for this cache
     * in one lump from the source, we keep this limit quite low.
     * <p>
     * We also want to avoid busting the heap with very large byte
     * arrays being copied around, etc.
     * <p>
     * For simplicity we limit this to to maximum data chunk that
     * we are happy to transfer in one go.
     */
//    private static final int MAX_INMEM_EXHIBIT_PREFIX_BYTES =
//        MAX_TRANSFER_CHUNK_SIZE;

    /**If true, accessing a thumbnail marks its exhibit as accessed.
     * This means that for the purposes of avoiding eviction from
     * cache, accessing a thumbnail is taken as being as significant
     * as downloading (part of) the exhibit itself, if true.
     * <p>
     * If false, then only actually reading part of the exhibit itself
     * helps keep the exhibit ``fresh'' in the cache.
     */
    private static final boolean THUMBNAIL_ACCESS_UPDATES_ACCESS_TIMESTAMP = false;

    /**Approximate minimum interval between rechecks of on-disc cache.
     * When a check is done, the in-memory record of disc cache status
     * is reloaded from disc, any debris is removed, etc.
     * <p>
     * A period of the order of at least a day is probably about right;
     * not exactly a day multiple helps ensure that we do not hit the same time
     * every day, which might otherwise collide with other regular activity.
     * Note that this recheck may take at least several minutes,
     * so we don't want to do it too often!
     * <p>
     * The interval selected here is ~2D + a random fudge factor,
     * which seems plently in practice.
     */
    private static final int DISC_RECHECK_INTERVAL_MS =
        (36 * 3600 * 1000) + Rnd.fastRnd.nextInt(12 * 3600 * 1000);

    /**Approximate minimum interval between saves of the metadata; strictly positive.
     * Since access to the exhibit/thumbnail data causes
     * this to be updated (along with more significant changes to
     * the cache), and saving can take significant time, we do not
     * want to save this immediately we enocunter a change.
     * <p>
     * We can postpone a save for a while at the risk that if the
     * system crashes or shuts down during that time and there
     * was a structural change to the cache, we might have to abandon
     * the old metadata and start again, which could be slow and
     * a bit messy (losing some useful though not vital information).
     * <p>
     * Taking our cue from the UNIX sync interval of 30s, a value in
     * the range 30s to a few minutes is probably reasonable.
     */
    private static final int METADATA_MIN_SAVE_INTERVAL_MS =
        (3 * 60 * 1000) + Rnd.fastRnd.nextInt(60 * 1000);


    /**Our local logger; never null. */
    private final SimpleLoggerIF logger;

    /**The stats set to which we log general cache behaviour.
     * The unique codes are the constants SCGNAME_XXX.
     */
    private final StatsLogger.StatsConfig statsIDSCGEN;

    /**General stats event name: an exhibit was evicted from the cache in LRU order. */
    public static final String SCGNAME_MDSAVE = "metaDataSave";

    /**General stats event name: an exhibit was evicted from the cache in LRU order. */
    public static final String SCGNAME_CACHEEVICTLRU = "exhibitEvictedFromCacheLRU";

    /**General stats event name: an exhibit was removed from the cache. */
    public static final String SCGNAME_CACHEREM = "exhibitRemovedFromCache";

    /**General stats event name: an exhibit was added to the cache. */
    public static final String SCGNAME_CACHEADD = "exhibitAddedToCache";

    /**General stats event name: a corrupt exhibit was removed from the cache. */
    public static final String SCGNAME_CACHEREM_CORRUPT = "exhibitRemovedFromCacheCORRUPT";

    /**General stats event name: an exhibit in the cache was fully validated against checksums, etc. */
    public static final String SCGNAME_CACHE_VALIDATION = "exhibitValidatedInCache";

    /**General stats event name: an exhibit in the cache was partially validated against checksums, etc. */
    public static final String SCGNAME_CACHE_VALIDATION_PART = "exhibitValidatedInCachePartial";

    /**General stats event name: cache raw data read miss.
     * We had to go upstream for at least part of the data.
     * <p>
     * (It is possible to have a hit and a miss on the same read
     * if part is satisfied from cache and part not.)
     */
    public static final String SCGNAME_CACHERAWDATAMISS = "exhibitCacheRawDataMiss";

    /**General stats event name: cache raw data read hit.
     * We satisfied at least part of the read from cache.
     * <p>
     * (It is possible to have a hit and a miss on the same read
     * if part is satisfied from cache and part not.)
     */
    public static final String SCGNAME_CACHERAWDATAHIT = "exhibitCacheRawDataHit";

    /**General stats event name: on-disc cache hit for thumbnail. */
    public static final String SCGNAME_CACHETNHIT = "exhibitCacheThumbnailHit";

    /**General stats event name: in-memory cache hit for thumbnail. */
    public static final String SCGNAME_CACHETNMEMHIT = "exhibitCacheThumbnailMemoryHit";

    /**General stats event name: in-memory cache hit for thumbnail. */
    public static final String SCGNAME_CACHETNMISS = "exhibitCacheThumbnailMiss";

    /**General stats event name: fetched a data block from a peer. */
    public static final String SCGNAME_DATAFETCHFROMPEER_PREFIX = "exhibitPeerDataFetch-";

    /**General stats event name: restarted scanning all exhibits for precaching. */
    public static final String SCGNAME_PRECACHERESTART = "exhibitPrecacheRestart";

    /**General stats event name: exhibits examined for precaching. */
    public static final String SCGNAME_PRECACHEEXAMINED = "exhibitPrecacheExhibitsExamined";

    /**General stats event name: exhibit data block precached. */
    public static final String SCGNAME_PRECACHEEXDATABLOCK = "exhibitPrecacheExhibitBlock";

    /**General stats event name prefix: exhibit data block precache source (if not from master/upstream). */
    public static final String SCGPREF_PRECACHEEXDATABLOCKSRC = "exhibitPrecacheExhibitBlockSrc=";

    /**General stats event name prefix: exhibit data block precache source for error (if not from master/upstream). */
    public static final String SCGPREF_PRECACHEEXDATABLOCKSRCERR = "exhibitPrecacheExhibitBlockSrcErr=";

    /**General stats event name prefix: exhibit data block precache (succesful) fetch time (log2 ms). */
    public static final String SCGPREF_PRECACHEEXDATABLOCKFETCHTIME = "exhibitPrecacheExhibitBlockFetchTime=";

    /**General stats event name: errors encountered during precaching. */
    public static final String SCGNAME_PRECACHEERROR = "exhibitPrecacheErrors";

    /**General stats event name: incoming request for exhibit data. */
    public static final String SCGNAME_EXDATAREQIN = "exhibitRawDataRequestIn";

    /**General stats event name: incoming request for exhibit data with "dontCache" flag set. */
    public static final String SCGNAME_EXDATAREQINDC = "exhibitRawDataRequestIn_dontCache";

    /**General stats event name: incoming request for exhibit thumbnails. */
    public static final String SCGNAME_EXTHUREQIN = "exhibitThumbnailRequestIn";

    /**General stats event name: incoming request for exhibit thumbnails with "dontCreate" flag set. */
    public static final String SCGNAME_EXTHUREQINDC = "exhibitThumbnailRequestIn_dontCreate";

    /**General stats event name: created thumbnails locally from cached data. */
    public static final String SCGNAME_EXTHUCREATED = "exhibitThumbnailCreated";

    /**If true, assume that newly-loaded meta-data at is OK at start-up until proven otherwise. */
    private static final boolean ASSUME_LOADED_METADATA_OK = true;

    /**If true, synchronously force a save of meta-data each time we add a new entry at least.
     * May be slow (O(n^2) for n exhibits),
     * especially where the cache is not large enough to hold all exhibits
     * so items are continually coming and going,
     * but potentially makes the system more robust against loss of data.
     */
    private static final boolean FORCE_IMMEDIATE_SAVE_ON_EXPANDED_METADATA = false;


    /**Key in generic props of P2P-profiling flag. */
    private static final String KEY_debugFlag_P2P_BLOCKXFER = CoreConsts.GEN_PREFIX_debugFlag + "P2P.profile.blockXfer";


    /**Maximum number of "best" exhibits to get enhanced precaching; non-negative. */
    private static final int MAX_BEST_EX_PRECACHED = 301;

    /**Contains the full exhibit names of the "best" few exhibits for enhanced precacheing; never null.
     * Maintained/updated by _doPrecache().
     * <p>
     * This is a snapshot of what the "best" exhibits are estimated to be
     * as each precacheing round starts, maybe on "quick approximation" data.
     * <p>
     * The implementation is optimised for fast lookup with "contains()".
     * <p>
     * Thread-safe.
     */
    private final Set<String> _bestExhibits =
        Collections.synchronizedSet(new HashSet<String>(2 * MAX_BEST_EX_PRECACHED));

    /**Cache meta-data class.
     * Holds information about what exhibit data is cached on disc.
     * <p>
     * Is serialisable to be able to persist to disc and
     * allow for a fast start-up.
     * <p>
     * Can be marked read-only to hold the cache in its current state.
     * <p>
     * This object is completely thread-safe and holds its instance
     * lock to synchronise activity.
     * <p>
     * The names of all members affecting the state of cached exhibit data
     * and meta-data start with `exhibit'.
     * <p>
     * All activity that actually affects disc operations must be
     * passed the rwl and it will be held while disc
     * I/O takes place, and most of these operations will throw
     * and IOException.
     * (Note that if the ExhibitDataSimpleCache instance lock
     * must be held as well then it and the rwl must be
     * grabbed in the appropriate order before one of these
     * routines is called.)
     */
    private final static class MetaData implements Serializable,
                                                   ObjectInputValidation
        {
        /**Build an empty, read-only set of meta-data. */
        MetaData()
            {
            // Update derived data.
            _recomputeDerivedValues();

            // Verify object state.
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Build a new meta-data set by examining the given cache directory.
         * This is understood not to be perfect, because, for example,
         * the last-accessed time may not be available on disc and
         * we may have to approximate with an available last-modified-time;
         * generally this is used when:
         * <ol>
         * <li>no persisted MetaData value is available, and we have
         *     to do the best we can, or,
         * <li>we want to load a copy of what is on disc to check that
         *     what we have in memory is not badly wrong.
         * </ol>
         * <p>
         * No locks (ie expecially the rwl or cache's instance lock)
         * should be held while this is running.
         * <p>
         * This does not modify the disc.
         * <p>
         * The constructed object is marked as needing to be saved.
         * <p>
         * This relies in part on being able to find the data files for
         * the cached (prefixes of) exhibits by their valid full exhibit names.
         * <p>
         * We do also explicitly attempt to recover any plausible entries
         * corresponding to current AEP entries.
         * <p>
         * Thus we may miss entries which have no current data file,
         * eg because we had zero bytes cached for that file,
         * and may have to recover such data later as we run across it.
         * <p>
         * We do also explicitly attempt to recover any plausible entries
         * corresponding to current AEP entries.
         * 
         * @param aepCurrent  reasonably current AEP as a hint for entries to look for;
         *     never null
         * @param cacheDir  the cache top-level directory; never null
         * @param logger  to log to; never null
         *
         * @throws java.io.IOException  in case of major problems in reconstructing
         *     the cache state (minor problems will be ignored)
         */
        MetaData(final AllExhibitProperties aepCurrent,
                 final File cacheDir,
                 final SimpleLoggerIF logger)
            throws IOException
            {
            assert(aepCurrent != null);
            assert(cacheDir != null);
            assert(logger != null);

            // Compute the base dir of the exhibit data cache.
            final File dataBaseDir = new File(cacheDir, CACHE_EXDATA_DIR);

            if(!dataBaseDir.isDirectory() ||
               !dataBaseDir.canRead() ||
               !dataBaseDir.canWrite())
                { throw new IOException("invalid or unusable cache directory ``"+dataBaseDir+"''"); }

            // Get the set of cached files,
            // but not checking validity (eg magic numbers)
            // because:
            //  a) we may not have cached enough for the item to be strictly valid,
            //  b) we want to be quick.
            final HashSet<String> cachedFileNames = new HashSet(ExhibitFile.getFilesystemBasedExhibitNames(
                dataBaseDir, false));

            // We also throw in all names from the supplied AEP as candidates.
            cachedFileNames.addAll(aepCurrent.aeid.getAllExhibitNamesSet());
            
            // Now load all cached file stats into in-memory records.
            final Iterator<String> it = cachedFileNames.iterator();
            while(it.hasNext())
                {
                final String name = it.next();
                try {
                    final CachedFile cf = CachedFile.recoverExtantCachedFileDetails(cacheDir, name);

                    // Verify that the candidiate has something worth recovering,
                    // (or at least inspecting and discarding to free up space)
                    // ie exhibit data and/or thumbnails.
                    // TODO: this may include other metadata (eg FEC) in future.
                    if((cf.cachedLength > 0) || cf.hasThumbnails() ||
                        (cf.lastAccessed > 0))
                        { cachedExhibits.put(name, cf); }
                    }
                catch(final Exception e)
                    {
                    // Since we seemed to encounter a problem with this putative
                    // item, we ignore it.
                    // FIXME: Possibly we should forcefully purge it.
logger.log("[ExhibitDataSimpleCache.MetaData: ERROR: trouble loading cache item ``"+name+"''...]");
                    }
                }

            // We should save this data that we have captured.
            setNeedsSave();

            // Update derived data.
            _recomputeDerivedValues();

            // Verify object state.
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Set of cached exhibits, from String exhibit name to CachedFile entry; never null after construction/deserialisation.
         * We may like to optimise memory use by sharing the String values (etc)
         * with those from AllExhibitImmutableData held elsewhere.
         * <p>
         * This is the primary store of data; other (transient)
         * values are derived from it;
         * this is saved to and restored from the serialised form
         * in an efficient and defensive way.
         * <p>
         * This is a Hashtable for thread-safety.
         * <p>
         * Compound operations may be made atomic by a lock on this object,
         * but no other lock may be taken while this happens.
         * <p>
         * Any access that updates this table should be under the instance lock.
         */
        private /*final*/ transient Hashtable<String,CachedFile> cachedExhibits = new Hashtable<String, CachedFile>();

        /**Get Set of exhibits on which metadata is currently held in this cache.
         * This takes a private atomic copy of the Set of exhibits
         * for which we have some data, thumbnails, etc, in the cache.
         */
        /*synchronized*/ Set<String> getKnownExhibits()
            { return(new HashSet<String>(cachedExhibits.keySet())); }

        /**LRU ordered set of the CachedFile values in cachedExhibits; never null after construction/deserialisation.
         * This should be exactly the same set as cachedExhibits.valueSet().
         * <p>
         * The least-recently-accessed items are first in this list,
         * and are the first to be discarded if the cache becomes over-full.
         * <p>
         * This is not of itself thread-safe and all access must be under
         * the instance lock.
         */
        private /*final*/ transient SortedSet<CachedFile> exhibitsLRU;

        /**Estimated total bytes of disc used by cached entries.
         * Computed as sum of estimates from CachedFile values.
         * <p>
         * Write access restricted to _setTotalBytes();
         * marked volatile to allow unlocked read access.
         */
        private volatile transient long totalBytes;

        /**Get estimated total bytes of disc used by cached entries; never negative. */
        long getTotalBytes() { return(totalBytes); }

        /**Set estimated total bytes of disc used by cached entries; never negative. */
        private synchronized void _setTotalBytes(final long newTB)
            {
            assert(newTB >= 0);
            totalBytes = newTB;
            if(newTB > totalBytesHighWaterMark) { totalBytesHighWaterMark = newTB; }
            }

        /**The highest value of totalBytes, ie the cache actual high-water mark; never negative.
         * Computed as sum of estimates from CachedFile values.
         * <p>
         * Write access restricted to _setTotalBytes() and by _setTotalBytesHighWaterMark()
         * and must be under the instance lock.
         * <p>
         * Marked volatile to allow lock-free read access.
         */
        private volatile transient long totalBytesHighWaterMark;

        /**Get the highest value of totalBytes, ie the cache actual high-water mark; never negative. */
        long getTotalBytesHighWaterMark() { return(totalBytesHighWaterMark); }

        /**Resets totalBytesHighWaterMark to the current value of totalBytes. */
//        private synchronized void _setTotalBytesHighWaterMark() { totalBytesHighWaterMark = getTotalBytes(); }

        /**This private routine updates all (transient) data derived from the primary cachedExhibits map.
         * This is used after deserialisation, construction or insertion where
         * it is easier to recompute such derived data from stratch
         * than change it incrementally as usual.
         */
        private synchronized void _recomputeDerivedValues()
            {
            final SortedSet<CachedFile> _new_exhibitsLRU = new TreeSet<CachedFile>();
            long _new_totalBytes = 0;

            for(final Iterator<String> it = cachedExhibits.keySet().iterator(); it.hasNext(); )
                {
                final String name = it.next();
                final CachedFile cf = cachedExhibits.get(name);

                _new_exhibitsLRU.add(cf);
                _new_totalBytes += cf.calcDiscSpace();
                }

            // Save new computed values.
            exhibitsLRU = _new_exhibitsLRU;
            _setTotalBytes(_new_totalBytes);
            }

        /**If true, we need a save to disc (ie our state has changed).
         * Is transient so always false when we recover from disc.
         * <p>
         * Is volatile (so accessed with no lock) only by
         * setNeedsSave() (to set it true) and
         * saveToDisc() (to clear it after a successful save to disc).
         */
        private transient volatile boolean needsSave;

        /**Get the `needs to be saved to disc' value.
         * Always false after deserialisation and construction of a new instance.
         */
        boolean getNeedsSave() { return(needsSave); }

        /**Set the `needs to be saved to disc' value to be true. */
        void setNeedsSave() { needsSave = true; }

        /**Save to disc; throws IOException in case of difficulty.
         * Holds a lock on the metaData object while the save takes place;
         * clears needsSave if successful.
         * <p>
         * @param cacheDir  specified the top-level cache directory; never null
         */
        void saveToDisc(final File cacheDir,
                        final SimpleLoggerIF logger,
                        final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            //assert...
            if(cacheDir == null)
                { throw new IllegalArgumentException(); }

            final File toSaveTo = new File(cacheDir, CACHE_METADATA_FILENAME);

            // Do as little as possible with the lock held...
            synchronized(this)
                {
                final long start = System.currentTimeMillis();
                FileTools.serialiseToFile(this, toSaveTo, STORE_EXPROPS_GZIPED, !IsDebug.isDebug);
if(IsDebug.isDebug) { logger.log("[MetaData.saveToDisc() done: "+ (System.currentTimeMillis()-start) +"ms.]"); }
                needsSave = false; // OK, seems to have worked.
                }

            // Note save of meta-data from poll()...
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_MDSAVE);
            }

        /**If true, this cache meta-data is read/write.
         * This can be used to pin the state of the cache
         * while we compare it to on-disc state and/or
         * while we are unsure of its veracity.
         * <p>
         * Is transient so thst it is deserialised false,
         * and is also false on construction of a new empty MetaData instance;
         * should only be set true after comparing with
         * or recreating from, real disc state.
         * <p>
         * Is volatile so that it can be read without taking a lock.
         */
        private volatile transient boolean readWrite;

        /**Get the `cache-is-read/write' value.
         * Always false after deserialisation and after construction of a new instance.
         */
        boolean isReadWrite() { return(readWrite); }

        /**Set the `cache-is-read/write' value.
         * Only the cache-checking routines should call this at all.
         * <p>
         * Insists on grabbing the instance lock to block changes to this
         * without the instance lock held.
         */
        synchronized void setReadWrite(final boolean rw) { readWrite = rw; }


        /**Get count of cached exhibits, including partially-cached ones; never negative. */
        /*synchronized*/ int size() { return(cachedExhibits.size()); }

        /**Is our meta-data store empty, ie no exhibits cached at all? */
        /*synchronized*/ boolean isEmpty() { return(cachedExhibits.isEmpty()); }

        /**Get cached-file details for a given exhibit; null if exhibit not cached.
         * This may incrementally verify cache data against the disc copy,
         * though for performance reasons this may not happen very often.
         * <p>
         * This is thread-safe and does not need nor grab any cache locks,
         * but to ensure this data remains valid during any compound operation,
         * at least a read lock should usually be obtained.
         *
         * @param name  the exhibit whose meta-data is required; never null
         */
        CachedFile exhibitGetInfo(final String name)
            {
            return(cachedExhibits.get(name));
            }

        /**Check if the exhibit is fully loaded.
         * <p>
         * This does not grab any cache locks,
         * but to ensure this data remains valid during any compound operation,
         * at least a read lock should generally be obtained.
         *
         * @param esa  the exhibit whose meta-data is required; never null
         *
         * @return false if no such exhibit or it is not fully loaded
         */
        boolean exhibitIsFullyLoaded(final ExhibitStaticAttr esa)
            //throws IOException
            {
            if(esa == null) { return(false); }
            final CachedFile icf = exhibitGetInfo(esa.filePath);
            if((icf == null) || (icf.cachedLength != esa.length))
                { return(false); }
            return(true); // Looks OK.
            }

        /**Remove cached-file details, adjusting cache data in situ; does nothing if not present.
         * The argument must not be null.
         * Private to this instance.
         * <p>
         * Does not check the read/write status; should be checked
         * before this is called.
         * <p>
         * Cache write lock must be held by current thread.
         */
        private synchronized void _remove(final ReentrantReadWriteLock rwl,
                                          final String name)
            {
            assert (rwl != null) || (name != null);

            // Write lock must be held by current thread.
            assert(rwl.isWriteLockedByCurrentThread());

            // Data should be consistent when we start.
            assert (exhibitsLRU != null) && (exhibitsLRU.size() == cachedExhibits.size()) && (totalBytes >= 0);

            final CachedFile oldCf = exhibitGetInfo(name);
            if(oldCf != null)
                {
                // If old data exists, carefully remove it
                // from all the places it affects.
                _setTotalBytes(getTotalBytes() - oldCf.cachedLength);
                exhibitsLRU.remove(oldCf);
                cachedExhibits.remove(oldCf.name);

                // Note need for meta-data save.
                setNeedsSave();

                // Data should be consistent after removing old metadata.
                assert (exhibitsLRU != null) && (exhibitsLRU.size() == cachedExhibits.size()) && (totalBytes >= 0);
                }
            }

        /**Update or add new cached-file details, adjusting cache data in situ.
         * The argument must not be null.
         * Private to this instance.
         * <p>
         * Does not check the read/write status; should be checked
         * before this is called.
         * <p>
         * Marks the meta-data as needing to be saved.
         * <p>
         * Cache write lock must be held by current thread.
         */
        private synchronized void _update(final ReentrantReadWriteLock rwl,
                                          final CachedFile newCf,
                                          final SimpleLoggerIF logger)
            {
            assert (rwl != null) && (newCf != null);

            // Write lock must be held by current thread.
            assert(rwl.isWriteLockedByCurrentThread());

            // Data should be consistent when we start.
            assert (exhibitsLRU != null) && (exhibitsLRU.size() == cachedExhibits.size()) && (totalBytes >= 0);

            // First, zap any extant entry.
            _remove(rwl, newCf.name);

            // Now add new data incrementally.
            cachedExhibits.put(newCf.name, newCf);
            exhibitsLRU.add(newCf);
            _setTotalBytes(getTotalBytes() + newCf.cachedLength);

            // Check for internal consistency...
            if(cachedExhibits.size() != exhibitsLRU.size())
                {
                logger.log("ExhibitDataSimpleCache.MetaData._update(): ERROR: internal data structures incosistent: repairing...");

                // Chuck away old values; insert new ones...
                exhibitsLRU.clear();
                exhibitsLRU.addAll(cachedExhibits.values());
                _recomputeDerivedValues();
                }

            // Note need for meta-data save.
            setNeedsSave();

            // Data should be consistent when we finish.
            assert (exhibitsLRU != null) && (exhibitsLRU.size() == cachedExhibits.size()) && (totalBytes >= 0);
            }

        /**Returns true if this metadata is essentially equivalent to another one.
         * This has a sufficiently loose notion to allow for discrepancies
         * in (for example) last-accessed timestamps which cannot be completely
         * reconstructed from the disc copy, but vital differences are noted.
         * <p>
         * The result is false if the argument is null.
         */
        synchronized boolean isEquivalent(final MetaData other)
            {
            if(other == null) { return(false); }

            // Clearly not equivalent if number of cached exhibits differ...
            if(size() != other.size()) { return(false); }

            // Check that each cached entry matches,
            // except in the last-accessed timestamp values.
            for(final Iterator<String> it = cachedExhibits.keySet().iterator(); it.hasNext(); )
                {
                final String name = it.next();
                final CachedFile cf1 = exhibitGetInfo(name);
                final CachedFile cf2 = other.exhibitGetInfo(name);

                // If we find a significant discrepancy,
                // these MetaData sets are not equivalent.
                if(!cf1.isEquivalent(cf2)) { return(false); }
                }

            return(true); // Seem to be essentially the same.
            }

        /**Install info from new cache object in this one.
         * This would allow us to incorporate the results
         * of a disc scan to see the actual state of the disc.
         * (Usually the argument is the best-efforts reconstruction
         * from the disc of the cache state, with the timestamps
         * arbitrarily too old.)
         * <p>
         * This essentially throws away the old state and
         * replaces it with that passed in,
         * though this may do things such as retain the best
         * notion of last-accessed timestamp from both.
         * <p>
         * Both this and the newData objects are locked for
         * the duration of this operation.
         * <p>
         * The readWrite status is not altered or updated
         * by this operation but needsSave may be;
         * therefore both objects must be read-only before this is called.
         * <p>
         * Only the cache-checking routines should call this at all.
         */
        synchronized void mergeWithNewMetaData(final MetaData newData,
                                               final SimpleLoggerIF logger)
            throws IllegalStateException
            {
            synchronized(newData)
                {
                if(readWrite || newData.readWrite)
                    { throw new IllegalStateException("must both be read-only"); }

                // By default, assume that this data will need saving.
                setNeedsSave();

                // Save the old data away.
                final Map<String,CachedFile> oldCachedExhibits = new HashMap<String, CachedFile>(cachedExhibits);

                // Replace old data with new.
                // Don't preserve any data from the old set.
                cachedExhibits.clear();
                cachedExhibits.putAll(newData.cachedExhibits);

                // Now for every new entry,
                // if we have a usable old CachedFile entry
                // use it but fix it up from disc.
                for(final Iterator<String> it = cachedExhibits.keySet().iterator(); it.hasNext(); )
                    {
                    final String name = it.next();
                    final CachedFile newCf = cachedExhibits.get(name);
                    final CachedFile oldCf = oldCachedExhibits.get(name);
                    if(!newCf.isEquivalent(oldCf)) { continue; } // Old value not usable.

                    // Restore old value, fixed-up.
                    cachedExhibits.put(name, oldCf.fixup(newCf, logger));
                    }

                // Update derived data.
                _recomputeDerivedValues();
                }
            }

        /**Mark exhibit as accessed (`touch' it) and returns the possibly-modified CachedFile record.
         * Returns null if the named exhibit does not exist in the cache at all,
         * else never returns null.
         * <p>
         * This call does not access disc (ie affects in-memory status only),
         * and silently does nothing if there is no entry for the named
         * exhibit.
         * <p>
         * If this metadata not read-write, this action is silently vetoed
         * as it is assumed not to be of vital importance.
         * <p>
         * This marks the last-accessed time in the MetaData as now,
         * to be preserved on disc at some point in the future.
         * <p>
         * This MetaData object is marked as needing to be saved if
         * an exhibit was `touched' successfully.
         * <p>
         * This is preferrably used to:
         * <ul>
         * <li>Mark end-user access to an exhibit's data.
         * <li>Mark some expensive operation (such as computing thumbnails)
         *     which we would like to avoid throwing away too soon.
         * </ul>
         * <p>
         * This holds a write lock on the cache while updating state.
         *
         * @param cacheDir  if not null
         *     then we force an update of a timestamp on disc
         *     as a backup in case the serialised metadata is lost;
         *     this causes extra disc traffic though not necessarily synchronous
         */
        CachedFile exhibitMarkAsAccessed(final ReentrantReadWriteLock rwl,
                                         final String name,
                                         final SimpleLoggerIF logger,
                                         final File cacheDir)
            throws IOException
            {
            _getWriteLock(rwl, "exhibitMarkAsAccessed()", logger);
            try
                {
                final CachedFile cf = exhibitGetInfo(name);
                if(cf == null) { return(null); }

                // Ignore `touch' if read-only.
                if(!isReadWrite()) { return(cf); }

                // Make new touched entry.
                final long now = System.currentTimeMillis();
                final CachedFile touched = cf.touchedEntry(now);

                // Update with touched entry.
                _update(rwl, touched, logger);

                // Force secondary timestamp backup to disc if requested.
                if(cacheDir != null)
                    {
                    final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
                    dataFile.setLastModified(now); // Enforce our clock's `now'.
                    }

                // Return new entry.
                return(touched);
                }
            finally { rwl.writeLock().unlock(); }
            }

        /**If true, when creating cache entry if data found on disc use it rather than refuse to build meta-data entry.
         * This is useful for co-existing with old cache mechanism.
         */
        private static final boolean USE_EXTANT_DISC_ENTRY_ON_CREATE = true;

        /**Remove the least-recently-used exhibit cache item, if any.
         * We won't remove the given named item if non-null;
         * this may prevent us removing anything if it is the only cached item.
         * <p>
         * If the named item is the oldest then we'll remove the next oldest,
         * if there is one.
         *
         * @param justData  if true then we try to retain metadata/thumbnails
         *
         * @return true if we removed something
         */
        private boolean _exhibitRemoveLRUCacheEntry(final ReentrantReadWriteLock rwl,
                                                    final File cacheDir,
                                                    final String dontRemoveName,
                                                    final boolean justData,
                                                    final SimpleLoggerIF logger,
                                                    final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            // Grab both locks in correct order.
            _getWriteLock(rwl, "_exhibitRemoveLRUCacheEntry()", logger);
            try
                {
                synchronized(this)
                    {
                    final Iterator<CachedFile> it = exhibitsLRU.iterator();

                    // Get oldest entry.
                    if(!it.hasNext()) { return(false); } // Cache is empty.
                    final CachedFile cf = it.next();
                    if(!cf.name.equals(dontRemoveName))
                        {
                        // Good, we can try this one...
                        if(!exhibitRemoveCacheEntry(rwl, cacheDir, cf.name, justData, logger, statsIDSCGEN))
                            { return(false); }
                        }
                    else if(it.hasNext())
                        {
                        // Get second-oldest entry;
                        // this can't be the named exhibit,
                        // since we only get here if the oldest entry was
                        // the special named exhibit...
                        final CachedFile cf2 = it.next();
                        // Good, we can try this one...
                        if(!exhibitRemoveCacheEntry(rwl, cacheDir, cf2.name, justData, logger, statsIDSCGEN))
                            { return(false); }
                        }
                    else
                        { return(false); /* Nothing removed from the cache. */ }
                    }
                }
            finally { rwl.writeLock().unlock(); }

            // Note eviction of old exhibit from cache...
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHEEVICTLRU);
            return(true);
            }

        /**Remove an exhibit entry from cache, possibly including all its data and metadata.
         * Returns true if successful, false if not.
         * <p>
         * This will fail if the metaData is read only,
         * or if an entry for the exhibit does not exist
         * in the metadata and on disc.
         * <p>
         * Since these failures will be by returning false
         * rather than by causing an exception then this
         * can be used as a gentle unconditional way to make sure that
         * an entry does not exist whether or not one did before.
         * <p>
         * This needs to access disc and so will need to
         * hold the main write lock while it does so, grabbing
         * the metaData instance lock inside the main lock
         * where both need to be held simultaneously.
         * <p>
         * Can optionally attempt to just remove the exhibit data,
         * leaving metadata and any thumbnails intact.
         *
         * @param rwl  the main cache lock object; must not be null
         * @param justData  if true then we try to retain metadata/thumbnails
         */
        boolean exhibitRemoveCacheEntry(final ReentrantReadWriteLock rwl,
                                        final File cacheDir,
                                        final String name,
                                        final boolean justData,
                                        final SimpleLoggerIF logger,
                                        final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert (rwl != null) && (name != null);

            // Fail gently if read-only.
            if(!isReadWrite()) { return(false); }

            _getWriteLock(rwl, "MetaData.exhibitRemoveCacheEntry()", logger);
            try
                {
                synchronized(this)
                    {
                    // Double-check that we haven't gone read-only...
                    if(!isReadWrite()) { return(false); }

                    // Does not exist in metaData; fail gently.
                    final CachedFile cf = exhibitGetInfo(name);
                    if(cf == null) { return(false); }

/* if(ORG.hd.d.IsDebug.isDebug) */ { logger.log("[MetaData.exhibitRemoveCacheEntry(): INFO: request to remove cache "+(justData?"data":"data and metadata")+" (with extant meta-data): " + name + ", last access: "+(new Date(cf.getLastAccessed()))+", cached size: "+cf.cachedLength+".]"); /* Thread.dumpStack(); */ }

                    // Only attempt removal of any extant data components.
                    //
                    // Update the exhibit timestamp regardless
                    // to try to ensure progress during purge() etc...
                    //
                    // This is "clock daemon" working-set management.
                    if(justData)
                        {
                        // To avoid leaving this exhibit in the LRU slot,
                        // the timestamp is updated
                        // not quite to "now" (less good than a real access)
                        // but enough to avoid being seen again immediately.
                        final long newStamp = (cf.lastAccessed + System.currentTimeMillis())/2 + 1;

                        // Explicitly get the file deleted if we have some data cached
                        // or if the file seems to exist anyway (eg is zero length).
                        if((cf.cachedLength > 0) ||
                           (new File(new File(cacheDir, CACHE_EXDATA_DIR), name).exists()))
                            {
                            final CachedFile newCf = cf.zapData(cacheDir);
                            _update(rwl, newCf.touchedEntry(newStamp), logger);
                            return(true);
                            }

                        _update(rwl, cf.touchedEntry(newStamp), logger);
                        return(false); // No exhibit data to remove.
                        }

                    // Does exist: remove metadata...
                    _remove(rwl, name);
                    // Does exist: remove file(s)/data...
                    cf.zapMe(cacheDir);
                    }
                }
            finally { rwl.writeLock().unlock(); }

            // Note removal of exhibit from cache...
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHEREM);

            return(true); // Done!
            }

        /**Create a new exhibit entry.
         * Returns true if successful, false if not.
         * <p>
         * This will fail if the metaData is read only,
         * or if an entry for the exhibit already exists
         * in the metaData or on disc.
         * <p>
         * Since these failures will be by returning false
         * rather than by causing an exception then this
         * can be used as a gentle unconditional way to make sure that
         * an entry does exist whether or not one did before.
         * <p>
         * This needs to access disc and so will need to
         * grab a write lock on the main cache lock while it does so, grabbing
         * the metaData instance lock inside the main lock
         * where both need to be held simultaneously.
         * <p>
         * Marks the meta-data as needing to be saved
         * if a new meta-data entry had to be created...
         *
         * @param rwl  the main lock object; must not be null
         * @param esa  the basic info on the exhibit to have an entry created
         *
         * @throws java.io.IOException  in case of serious problems
         *     accessing the on-disc cache state
         */
        boolean exhibitCreateNewCacheEntry(final ReentrantReadWriteLock rwl,
                                           final File cacheDir,
                                           final ExhibitStaticAttr esa,
                                           final SimpleLoggerIF logger,
                                           final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert (rwl != null) && (esa != null);

            // Fail gently if read-only.
            if(!isReadWrite()) { return(false); }

//if(ORG.hd.d.IsDebug.isDebug) { System.out.println("MetaData.exhibitCreateNewCacheEntry(): requested to create cache file: " + esa.filePath); }

            // Path to (possibly partially-) cached exhibit file.
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), esa.filePath);

            // Grab both locks in correct order.
            _getWriteLock(rwl, "exhibitCreateNewCacheEntry()", logger);
            try
                {
                synchronized(this)
                    {
                    // Already exists in metaData; fail gently.
                    if(exhibitGetInfo(esa.filePath) != null) { return(false); }

                    CachedFile newCf; // = null;
                    if(dataFile.exists())
                        {
                        // Already exists on disc; fail gently.
                        if(!USE_EXTANT_DISC_ENTRY_ON_CREATE) { return(false); }

                        // MORE PERMISSIVE STYLE...
                        // Already exists on disc; try to use it!
                        try
                            {
                            newCf = CachedFile.recoverExtantCachedFileDetails(
                                cacheDir, esa.filePath);
                            }
                        catch(final IOException e)
                            {
                            newCf = CachedFile.makeNewDiscCacheFile(
                                cacheDir, esa.filePath, esa.timestamp);

                            // Data on disc may be damaged,
                            // so ignore it and start again...
                            e.printStackTrace();
                            }
                        }
                    else
                        {
                        // OK, create new entry on disc ready for meta-data.
                        newCf = CachedFile.makeNewDiscCacheFile(
                            cacheDir, esa.filePath, esa.timestamp);
                        }

                    // Post updated entry to metaData.
                    assert(newCf != null);
                    _update(rwl, newCf, logger);

                    if(FORCE_IMMEDIATE_SAVE_ON_EXPANDED_METADATA)
                        { saveToDisc(cacheDir, logger, statsIDSCGEN); }
                    }
                }
            finally { rwl.writeLock().unlock(); }

            // Note addition of exhibit to cache...
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHEADD);

            return(true); // Done!
            }

        /**Computes the target high-water mark.
         * If we currently only have a default (zero-timestamp) genProps
         * then we use the initial size of the cache as our high-water
         * mark, ie do not allow it to become any larger.
         */
        long computeTargetHighWaterMark()
            {
            final long lpWMECB = LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES();

            // We don't necessarily trust the props value, eg if zero.
            final boolean weTrustProps = (lpWMECB > 0);

            // If we trust genProps use its high-water mark
            // (which may have a local override),
            // else we use the actual high-water mark so far
            // (or the fallback if larger).
            final long highWaterMark = weTrustProps ? lpWMECB :
                Math.max(FALLBACK_MIN_CACHE_SIZE, getTotalBytesHighWaterMark());

            return(highWaterMark);
            }

        /**Compute current definite free-space in cache.
         * This returns the amount of space that we have within the low-water
         * boundary; if this returns a positive number then the cache is within
         * that low water boundary and can be considered nowhere near full.
         * <p>
         * If we don't trust the GenProps value then the low-water mark
         * is the current actual high-water mark for the cache, ie
         * we try to maintain its size approximately.
         * <p>
         * Is synchronized to maintain consistency while working.
         */
        synchronized long computeFreeSpaceBelowLowWaterMark()
            {
            // We don't necessarily trust the props value, eg if zero.
            final boolean weTrustProps = (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0);

            // Conservative fall-back; maintain cache size for now.
            if(!weTrustProps)
                { return(getTotalBytesHighWaterMark() - getTotalBytes()); }

            final long highWaterMark = computeTargetHighWaterMark();
            final long lowWaterMark = Math.max(0, (long) (highWaterMark * LOW_WATER_FRACTION));
            return(lowWaterMark - getTotalBytes());
            }

        /**Returns true if we have lots of disc space free.
         * This means that usage is well below the low-water mark;
         * in this case less than half the low-water mark.
         * <p>
         * If we don't trust GenProps this always returns false.
         */
        boolean lotsFree()
            {
            // We don't necessarily trust the props value, eg if zero.
            final boolean weTrustProps = (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0);

            // Conservative fall-back; maintain cache size for now.
            if(!weTrustProps) { return(false); } // Can't be sure.

            final long highWaterMark = computeTargetHighWaterMark();
            final long lowWaterMark = Math.max(0, (long) (highWaterMark * LOW_WATER_FRACTION));

            // Lots of empty room!
            return(getTotalBytes() < (lowWaterMark/2 - 3*MAX_EXTD_TRANSFER_CHUNK_SIZE));
            }

        /**Returns true if we have some (not lots of) disc space free.
         * This means that usage is below the low-water mark.
         * <p>
         * If we don't trust GenProps this always returns false.
         */
        synchronized boolean someFree()
            {
            // We don't necessarily trust the props value, eg if zero.
            final boolean weTrustProps = (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0);

            // Conservative fall-back; maintain cache size for now.
            if(!weTrustProps) { return(false); } // Can't be sure.

            return(computeFreeSpaceBelowLowWaterMark() >= MAX_TRANSFER_CHUNK_SIZE);
            }

        /**Compute available free space in cache (bytes); zero or negative if none.
         * This ignores the purge state of the system and simply reports what
         * space is available before the cache overflows.
         */
        synchronized long computeFreeSpaceBelowHighWaterMark()
            {
            // Compute current free space.
            final long freeSpace =
                (computeTargetHighWaterMark() - getTotalBytes());

            return(freeSpace);
            }

        /**If true, we can pre-cache data blocks for (new or existing) exhibits.
         * POLICY: only true if we have (lots of) free space
         * and the cache is read/write,
         * and the server-slowdown factor is no greater than 1.
         */
        boolean canPrecacheExhibitData()
            {
            if(!isReadWrite()) { return(false); }
            if(LocalProps.getServerSlowdownFactor() > 1) { return(false); }
            return(lotsFree());
            }

        /**If true, we can pre-cache thumbnails for (existing) exhibits.
         * POLICY: only true if
         *         the free space is enough to store another thumbnail
         *         and fetch the data to do so first if necessary
         *         and the cache is read/write.
         * <p>
         * Basically we're prepared to go right up to the wire to generate
         * thumbnails because they are so useful to the user.
         * <p>
         * We leave a little bit of wiggle room on space.
         */
        boolean canPrecacheThumbnails()
            { return(isReadWrite() &&
                     (computeFreeSpaceBelowHighWaterMark() >
                      2 * (MAX_REMOTE_FETCH_TO_MAKE_THUMBNAIL + MAX_TRANSFER_CHUNK_SIZE))); }

        /**If true, some precaching may be possible, going by our generally-least-strict limit.
         * We will only allow precaching if we have a valid LocalProps value.
         */
        boolean canPrecache()
            {
            // We don't necessarily trust the props value, eg if zero.
            final boolean weTrustProps = (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0);
            if(!weTrustProps) { return(false); }
            return(canPrecacheThumbnails());
            }


        /**If true then try to retain metadata and thumbnails of exhibits we purge.
         * Thumbnails in particular may be expensive to regenerate/fetch,
         * and do not take much space.
         */
        private static final boolean TRY_TO_RETAIN_TN_IN_PURGE = true;

        /**Purges the cache of old entries to make space for new, if needed.
         * This will do nothing unless the space used by the cache is
         * above the low-water mark.  We compute the low-water mark
         * from GenProps, or from the actual high-water mark in
         * the current use of the cache if the passed GenProps is
         * the default (has a zero timestamp).  (We attempt to
         * roughly maintain the cache or possibly run it down slightly
         * while waiting for GenProps to arrive.)
         * <p>
         * This will avoid purging the cache of the item passed to it
         * by name in case it was a candidate for removal.
         * <p>
         * If the argument is zero this tries to ensure that space
         * for at least one new (empty) exhibit cache entry to be created.
         * <p>
         * If greater than zero this tries to ensure that an existing cache
         * entry can be extended by at least the given number of bytes
         * (either for the exhibit data or for thumbnail data).
         * <p>
         * This grabs the cache write lock and instance lock.
         * <p>
         * This will fail if the cache is marked read-only.
         *
         * @param howMuch  howMuch space we need to ensure is free; positive
         */
        void purge(final ReentrantReadWriteLock rwl,
                     final GenProps gp,
                     final File cacheDir,
                     long howMuch,
                     final String dontPurgeName,
                     final SimpleLoggerIF logger,
                     final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert (rwl != null) && (gp != null) && (cacheDir != null);

            if(howMuch < 0)
                { throw new IllegalArgumentException(); }

            // If lots of space free then return immediately.
            if(lotsFree()) { return; }

            // We convert a zero-length request
            // (which is a request to make space for a new empty entry)
            // into the estimated space required for such.
            if(howMuch == 0)
                { howMuch = CachedFile.MAX_EMPTY_ENTRY_BYTES_ON_DISC; }

            assert howMuch > 0;

            // We round up the (positive) request
            // to a disc block multiple.
            howMuch = FileTools.roundUpToFSBlockSize(howMuch);

            // Grab both locks in correct order.
            _getWriteLock(rwl, "MetaData.purge()", logger);
            try
                {
                synchronized(this)
                    {
                    if(!isReadWrite())
                        { throw new IllegalStateException("Cannot purge() read-only cache"); }

                    // We don't want to spend very long on this
                    // as we are probably blocking a user.
                    // However, as it probably involved real disc I/O,
                    // then we should not make this too short.
                    final long stopBy = System.currentTimeMillis() + 1001 +
                            Math.max(Rnd.fastRnd.nextInt(3000),
                                     3*CoreConsts.MAX_INTERACTIVE_DELAY_MS);

                    // When this gets set true
                    // we're getting desperate, eg due to lack of time.
                    boolean gettingDesperate = false;

                    // Zap files until we have made enough space
                    // (or we can't make any more because there are no files).
                    // We'll allow that the last entry might be the
                    // exhibit that we are trying to make (more) space for
                    // so we would have to exit the loop with (possibly us)
                    // left and rely on other mechanisms to zap this last item.
                    for(int itemsExamined = 0; size() > 1; ++itemsExamined)
                        {
                        // Recompute each time to see how far we've gotten...
                        final long freeSpace = computeFreeSpaceBelowHighWaterMark();

                        // Quit immediately if we have enough space...
                        if(freeSpace >= howMuch)
                            { break; }

                        // We may attempt to retain thumbnail/metadata
                        // unless running out of time.
                        final boolean justPurgeData =
                            TRY_TO_RETAIN_TN_IN_PURGE &&
                            !gettingDesperate;

                        // Purge the oldest extant cache entry if we can...
                        if(!_exhibitRemoveLRUCacheEntry(rwl,
                                          cacheDir,
                                          dontPurgeName,
                                          justPurgeData,
                                          logger, statsIDSCGEN))
                            { logger.log("INFO: ExhibitDataSimpleCache.purge() failed to remove exhibit entry on iteration #"+(1+itemsExamined)); }

                        // We were not desperate, but maybe we are now, due to:
                        //   * Having examined all the exhibits at least once.
                        //   * Having taken too long searching for a victim.
                        //   * Other threads queueing for the cache.
                        if(!gettingDesperate)
                            {
                            final int cacheSize = size();
                            if((itemsExamined > cacheSize) ||
                               (System.currentTimeMillis() > stopBy) ||
                               ((itemsExamined > cacheSize/8) && rwl.hasQueuedThreads()))
                                {
                                // We've spent quite long enough
                                // trying to gently free up space,
                                // so it's time to put our big boots on...
                                gettingDesperate = true;
                                }
                            }
                        }
                    }
                }
            finally { rwl.writeLock().unlock(); }
            }

        /**If true, we NEVER ask upstream to cache for us.
         * If true, we try to avoid churning the master's cache
         * even if it has lots of space,
         * and we are assuming that our local caching is as good as it gets.
         * <p>
         * If false, we just ask it not to cache our precache requests
         * which do not reflect explicit user requests,
         * and assume that there will still be shared locality
         * between different slaves or that it might be expensive
         * for the master to fetch data.
         * <p>
         * Note that in either case, if data can be satisfied from the
         * local cache then it will be, and if we have lots of space we might
         * cache locally regardless.
         */
        private static final boolean NEVER_REQUEST_UPSTREAM_CACHEING = false;

        /**Approximate maximum time to wait for another thread to extend an exhibit cache entry (ms); strictly positive. */
        private static final int MAX_CONC_EXT_WAIT_TIME_MS = 25001 + Rnd.fastRnd.nextInt(10101);

        /**Reads data for an exhibit into the given buffer.
         * This will complain with an IOException if an
         * attempt is made to read beyond the bounds of the
         * exhibit, or if data is unavailable for any reason,
         * or if the read is far too long to be sensibly handled,
         * else the requested chunk of data is read into the
         * caller's buffer.
         * <p>
         * This will extend the underlying cache entry if necessary
         * (and possible) to satisfy the read, and may also
         * apply read-ahead to maximise user-perceived performance.
         * <p>
         * If the metaData is read-only, then any (portion of the) read that
         * cannot be satisfied from the cache is passed back up the pipeline.
         * <p>
         * This needs to access disc and so will need to
         * hold the rwl while it does so, grabbing
         * the metaData instance lock inside the rwl
         * where both need to be held simultaneously.
         * <p>
         * Note that we control caching to suit our precaching and any
         * downstream precache/random activity,
         * so at this level we have a 3-way choice.
         * <p>
         * This may only mark a cache entry as accessed
         * if dontCache is FALSE
         * and we actually read from or add data to the cache,
         * though we may mark it as updated at other times too.
         * <p>
         * Whenever we (easily) know that we can satisfy the request from cache,
         * we only take a read lock to improve concurrency.
         * <p>
         * We return as soon as we have satisfied some part of the request
         * so as minimise internal copying of data, etc.
         *
         * @param rwl  the main cache lock object; must not be null
         * @param fetchFromPeer  if non-null and we need to extend the cache
         *     then we will try to do so from the mirror with the given tag
         * @param dontCache  if FALSE, then cache here if possible and
         *     ask upstream to cache;
         *     if null then cache locally but not upstream
         *     (this is our precaching activity);
         *     if TRUE then don't cache here or upstream
         *     (this is downstream precaching or random activity);
         *     note therefore that we ask upstream to cache
         *     unless dontCache is FALSE
         *
         * @throws java.io.IOException  in case of serious problems
         *     accessing the on-disc cache state
         *     or if request is out of bounds of underlying exhibit
         *     or too big to handle
         */
        void exhibitRead(final ReentrantReadWriteLock rwl,
                         final String fetchFromPeer,
                         final File cacheDir,
                         final String name,
                         final SimpleExhibitPipelineIF upstream,
                         final AllExhibitImmutableData aeid,
                         final GenProps gp,
                         final int dataStart,
                         final ByteBuffer buf,
                         final Boolean dontCache,
                         final SimpleLoggerIF logger,
                         final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert (rwl != null) && (cacheDir != null) && (name != null);
            assert (upstream != null) && (aeid != null) && (gp != null);
            assert (buf != null);

            // Veto out-of-bounds request.
            if(dataStart < 0) // Current practical limit.
                { throw new IllegalArgumentException(); }

            final ExhibitStaticAttr esa = aeid.getStaticAttr(name);
            // Immediately deflect requests for bogus exhibits.
            if(esa == null)
                { throw new FileNotFoundException("exhibit " + name + " does not exist"); }

            // Maximum length of request implied by buffer size.
            final int maxLen = buf.remaining();
            assert(maxLen >= 0);

            // Compute actual implicit (max) request length; non-negative.
            final int len = Math.min(maxLen,
                                (int) Math.min(MAX_USER_READ_SIZE,
                                               esa.length - dataStart));
            assert(dataStart + len <= esa.length);

            // Immediately deal with trivial zero-length reads.
            // They don't even mark the cache entry as used.
            if(len == 0) { return; }

            // Decide whether we will ask upstream to cache for us.
            final boolean cacheEverywhere = Boolean.FALSE.equals(dontCache);
            final boolean dontCacheUpstream = NEVER_REQUEST_UPSTREAM_CACHEING ||
                !cacheEverywhere;
            final boolean dontCacheLocally = Boolean.TRUE.equals(dontCache);

            // Is this a read of the last block/byte of the exhibit?
            final boolean finalByteRead = (dataStart + len == esa.length);


            // FAST PATH... (read-lock only)
            // Optimistically try for a read lock only to achieve good concurrency,
            // hoping that we don't have to go upstream.
            // Note that we'll have to take a write lock
            // for at least one read/access of the data if we want
            // to update timestamps to maintain the LRU cache;
            // we use the read of the last byte(s) to do this.
            //
            // We can use this read-only if ANY of the following holds:
            //   * we are not reading the last block/byte of the exhibit
            //   * we don't care about cacheing the file locally anyway
            //   * the metadata is currently read-only
            //   * enough of the exhibit is loaded to satifsy the request
            //   * the cache looks busy so it is worth taking this short-cut
            //   * this same exhibit is being extended by another thread
            //
            // If any "fast-path" condition is not met,
            // then simply fall through to the usual handler.
            //
            // Note: if the cache looks to be busy
            // (ie it might take us a long time to acquire a write lock)
            // then we'll not bother trying to update timestamps.
            // Our goal is to give fast responses to users.
            // This should particularly help performance for small exhibits
            // that may be used as their own thumbnails.
            do
                {
                if(!finalByteRead ||
                   dontCacheLocally ||
                   !isReadWrite() ||
                   rwl.hasQueuedThreads() || (rwl.getReadLockCount() > 0) || // Cache seems busy...
                   (_beingExtended.get(esa.filePath) != null)) // Being extended by another thread.
                    {
                    // Satisfy as much of the read from the cache as possible.
                    final int n = _readRawDataStartFromCache(rwl, name, dataStart, cacheDir, buf, logger, statsIDSCGEN);
                    // If we read anything then return it immediately.
                    if(n > 0) { return; }
                    }
                else { break; /* No conditions met so drop out of "fast path"... */ }

//                // Return immediately if request now completely satisfied...
//                if(len == 0) { return; }

                // If another thread is (now) extending this exhibit
                // then we read/sleep again until they are done,
                // or until we have been waiting a long time.
                final Long extendedSince = _beingExtended.get(esa.filePath);
                if(extendedSince == null) { break; }

                final long waitTime = Math.max(0,
                    System.currentTimeMillis() - extendedSince.longValue());
                if(waitTime > MAX_CONC_EXT_WAIT_TIME_MS/2) { logger.log("[ExhibitDataSimpleCache: waiting (@"+dataStart+", for "+len+"bytes) while another thread extends ("+waitTime+"ms so far) exhibit "+esa+".]"); }
                // Ensure that we don't block indefinitely...
                if(waitTime > MAX_CONC_EXT_WAIT_TIME_MS)
                    {
                    logger.log("[ExhibitDataSimpleCache: WARNING: waited too long for another thread to finish, so dropping though to extend data concurrently...]");
                    break;
                    }
                synchronized(extendedSince)
                    {
                    // Wait a little while (bearable for interactive users)
                    // or until we are signalled that the extending is done.
                    // Use a random wait to avoid multiple threads colliding.
                    // Wait in approximate proportion to time already waited
                    // to avoid wasting too much time spinning in this loop.
                    try { extendedSince.wait(CoreConsts.MAX_INTERACTIVE_DELAY_MS/2 + Rnd.fastRnd.nextInt(((int)(waitTime/4)) + CoreConsts.MAX_INTERACTIVE_DELAY_MS)); }
                    catch(final InterruptedException e) { throw new InterruptedIOException(e.getMessage()); }
                    }
                } while(true);

//            // Return if request now completely satisfied...
//            if(len == 0) { return; }

            // Create the cache entry or update the timestamp on the extant one,
            // and purge enough stale data to make room for the new data.
            _getWriteLock(rwl, "exhibitRead()", logger); // Grab both locks in correct order.
            try
                {
                synchronized(this)
                    {
                    do
                        {
                        // Do we know about this exhibit?
                        // If so mark then it as accessed too,
                        // providing that this was not a "dontCache" request,
                        // ie probably not from a local client.
                        CachedFile cf = dontCacheLocally ? exhibitGetInfo(name) :
                                                exhibitMarkAsAccessed(rwl, name, logger, finalByteRead ? cacheDir : null);
                        // If we don't, then we want to create it if possible.
                        if(cf == null)
                            {
                            // If the cache is not read/write,
                            // then prepare to duck through to upstream source.
                            if(!isReadWrite()) // Can't alter local cache.
                                { break; }

                            // Make space for new cache entry if necessary.
                            purge(rwl, gp, cacheDir, 0, name, logger, statsIDSCGEN);

                            // Cache is read-write, so create a new entry.
                            if(!exhibitCreateNewCacheEntry(rwl, cacheDir, esa, logger, statsIDSCGEN))
                                { throw new IOException("cannot create cache entry"); }
                            cf = exhibitGetInfo(name);
                            assert cf != null;
                            }

                        // Verify that the in-memory record of the cached length is correct.
                        // Fix-up the in-memory copy if wrong.
                        final long actualLength = CachedFile.getCachedDataLength(
                            cacheDir, name);
                        if(actualLength != cf.cachedLength)
                            {
                            final CachedFile rec =
                                CachedFile.recoverExtantCachedFileDetails(
                                    cacheDir, name);
                            final CachedFile fixedUp = cf.fixup(rec, logger);

                            // If we actually had to fix-up,
                            // then replace the in-memory copy.
                            if(fixedUp != cf)
                                {
                                cf = fixedUp;
                                _update(rwl, cf, logger);
                                }
                            }

                        } while(false);
                    } // Drop out of metaData lock...
                } // Drop cache write lock.
            finally { rwl.writeLock().unlock(); }

//            // Return if request completely satisfied...
//            if(len == 0) { return; }

            // Can we (now) satisfy any initial portion of the
            // request with data that we have in our cache?
            // This might be the case now because we just did a fix-up
            // or because the cache was extended while we were waiting
            // to acquire a write lock.
            // (This could happen with the cache read-only.)
            final int canDo = _readRawDataStartFromCache(rwl,
                                                     name,
                                                     dataStart,
                                                     cacheDir,
                                                     buf,
                                                     logger,
                                                     statsIDSCGEN);
            // If we managed to read from the "found" data
            // then we can return what we have immediately!
            // This avoids an upstream fetch for now.
            if(canDo > 0) { return; }

            // As long as:
            //   * the cache is not read-only,
            //   * we have an entry for this file in the cache,
            //   * the current request follows on from the cached data
            //     (or we can extend the request backwards a little to follow on from cached data),
            //   * we can cache locally (or there is lots of space free),
            // then we can fetch the next portion of the data from upstream
            // (NOT within the scope of any lock, ie NOT blocking other access)
            // and then write it into our cache (and return it to the caller too).
            CachedFile cf;
            if(isReadWrite() &&
               ((cf = exhibitGetInfo(name)) != null) &&
               (dataStart <= cf.cachedLength+(MAX_EXTD_TRANSFER_CHUNK_SIZE-MAX_TRANSFER_CHUNK_SIZE)) &&
               (!dontCacheLocally || lotsFree()))
                {
                final long fetchStart = cf.cachedLength;
                assert(fetchStart <= Integer.MAX_VALUE);

                // Find out what initial portion we are prepared to cache.
                final int MAX_CACHEABLE_EX_BYTES = gp.getWEBSVR_MAX_CACHEABLE_EX_BYTES();

                // We will extend the cache entry
                // by at most MAX_USER_READ_SIZE bytes at a time
                // and we will round up small residues to
                // MAX_TRANSFER_CHUNK_SIZE for efficiency if possible.

                // We allow a small hole just before the current request
                // to be backfilled so that we can cache the data fetched.

                // Now impose the relevant limits...
                // We impose the lowest of several potential limits.
                final long maxExtendTo =
                    // No more than larger of a max-transfer-size block or the request length,
                    // but capped by the maximum user transfer size,
                    // starting from the caller's requested start.
                    Math.min(dataStart + Math.min(Math.max(len, MAX_TRANSFER_CHUNK_SIZE),
                                                  ExhibitDataTunnelSource.MAX_USER_READ_SIZE),

                    // No more than the max-extended transfer from the current cache start.
                    Math.min(fetchStart + MAX_EXTD_TRANSFER_CHUNK_SIZE,

                    // Limit to a valid +ve int size.
                    Math.min(Integer.MAX_VALUE,

                    // Limit to no more than the actual exhibit length
                    // or the maximum initial portion of any exhibit that can be cached.
                    Math.min(esa.length,
                             MAX_CACHEABLE_EX_BYTES))));
                // How much would we extend the cache entry by?
                final int extendBy = (int) (maxExtendTo - fetchStart);
                assert(extendBy >= 0);

                // Fetch data from upstream to cache locally
                // if we have enough space for the data...
                // Indicate to other threads that we are extending this exhibit.
                final String exhibitExtendKey = new String(esa.filePath); // Create new String instance as a unique key.
                if(_beingExtended.put(exhibitExtendKey, new Long(System.currentTimeMillis())) != null)
                    {
                    logger.log("[ExhibitDataSimpleCache: WARNING: exhibit already being extended by another thread: "+esa+".]"); // Probably should not happen often.
                    }
                try
                    {
                    // Briefly grab a cache lock and purge some old data
                    // if we don't seem to have enough space to grow the exhibit.
                    if((extendBy > 0) && (computeFreeSpaceBelowHighWaterMark() < extendBy))
                        {
                        _getWriteLock(rwl, "exhibitRead()-purge", logger); // Grab both locks in right order...
                        try
                            {
                            synchronized(this)
                                { purge(rwl, gp, cacheDir, extendBy, esa.filePath, logger, statsIDSCGEN); }
                            }
                        finally { rwl.writeLock().unlock(); }
                        }

                    if((extendBy > 0) && (computeFreeSpaceBelowHighWaterMark() >= extendBy))
                        {
                        // Note raw data cache miss...
                        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHERAWDATAMISS);

                        // Get the data from upstream
                        // WITHOUT ANY CACHE LOCK HELD.

                        // Create a short-lived buffer (as non-direct).
                        final ByteBuffer tmpbuf = ByteBuffer.allocate(extendBy);

                        // We may allow a direct fetch from our peers
                        // rather than from upstream
                        // so as to lighten the load on the master.
                        if(ALLOW_DATA_FETCH_FROM_PEERS &&
                           (fetchFromPeer != null))
                            {
                            // Create temporary tunnel to the selected peer.
                            final String url = "http://" +
                                LoadBalancingUtils.makeMirrorNameFromTag(fetchFromPeer) +
                                CoreConsts.TUNNEL_URI;
                            // Suppress logging by these short-lived tunnels...
                            final SimpleLoggerIF fakeLogger = GenUtils.nullLogger;
                            final ExhibitDataHTTPTunnelSource tempTunnel =
                                new ExhibitDataHTTPTunnelSource(url, fakeLogger);
                            // Get the data or abort with an IOException...
                            try
                                {
                                tempTunnel.getRawFile(tmpbuf,
                                                      esa.filePath,
                                                      (int) fetchStart,
                                                      dontCacheUpstream);
                                // Log the (non-aborted) fetch.
                                StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_DATAFETCHFROMPEER_PREFIX+fetchFromPeer);
                                }
                            finally
                                { tempTunnel.destroy(); }
                            }
                        else // Fetch the data from our upstream source...
                            { upstream.getRawFile(tmpbuf, esa.filePath, (int) fetchStart, dontCacheUpstream); }

                        if(tmpbuf.position() == 0)
                            {
                            // We failed to fetch anything...
                            return;
                            }

                        // Prepare to read the data fetched from upstream.
                        tmpbuf.flip();

                        // Grab a write lock and cache the data fetched from upstream.
                        // Re-acquire both locks in the right order.
                        _getWriteLock(rwl, "exhibitRead()-data-save", logger);
                        try
                            {
                            synchronized(this)
                                {
                                // Check that the cache has not changed
                                // too radically for this fetched data to be usable.
                                final CachedFile newEntry = exhibitGetInfo(name);
                                if((newEntry != null) &&
                                   (newEntry.cachedLength >= fetchStart) &&
                                   (newEntry.cachedLength < fetchStart + tmpbuf.remaining()) &&
                                   isReadWrite())
                                    {
                                    cf = newEntry.extendCacheFile(cacheDir,
                                                fetchStart,
                                                tmpbuf);
                                    // Update our records.
                                    _update(rwl, cf, logger);
                                    }
                                }
                            }
                        finally { rwl.writeLock().unlock(); }

                        // Now read what we can from the extended on-disc cache.
                        // We do this under the read lock.
                        _readRawDataStartFromCache(rwl,
                                                     name,
                                                     dataStart,
                                                     cacheDir,
                                                     buf,
                                                     logger,
                                                     statsIDSCGEN);

                        // Return with whatever we got (if anything).
                        return;
                        }
                    }
                finally
                    {
                    synchronized(exhibitExtendKey)
                        {
                        // Wake up at any thread waiting for us to finish
                        // extending this exhibit.
                        // If this notification is lost then any waiting threads
                        // will soon wake/continue anyway.
                        exhibitExtendKey.notifyAll();

                        // Allow someone else to extend this exhibit...
                        _beingExtended.remove(exhibitExtendKey);
                        }
                    }
                }

//            // Return if request completely satisfied...
//            if(len == 0) { return; }

            // Fall-through catch-all case:
            // If we have not managed to satisfy the request,
            // then note a raw-data cache miss,
            // and out of the scope of any lock,
            // directly satisfy the request from upstream.
            //
            // This should only happen for the tail of huge exhibits,
            // or if the cache is read-only and we can't extend/expand it,
            // or if requests have been made in a non-sequential order from 0,
            // or when a client needs to (re)fetch the tail of something we new don't have cached.
            // This is more likely to happen on the master than on slaves.
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHERAWDATAMISS);
//if(cacheEverywhere && isReadWrite()) { logger.log("[ExhibitDataCache: WARNING: having to go upstream to complete request: dataStart/len/name=" + dataStart+"/"+len+"/"+name+"]"); }
            assert(buf.remaining() == maxLen); // No data read yet...
            upstream.getRawFile(buf, name, dataStart, dontCacheUpstream);
            assert(buf.remaining() <= maxLen); // Maybe we have fetched some data...
            }

        /**Fill as much as possible of the read request from disc cache, returning the number of bytes read; non-negative.
         * This is done grabbing a read lock on the entire cache,
         * but no exclusive lock so that we can do multiple concurrent reads,
         * ie so that we can queue concurrent reads at the OS/disc level
         * for maximum throughput.
         * <p>
         * Likely to be heavily used, so efficiency is important here.
         */
        private int _readRawDataStartFromCache(final ReentrantReadWriteLock rwl,
                                               final String name,
                                               final int dataStart,
                                               final File cacheDir,
                                               final ByteBuffer buf,
                                               final SimpleLoggerIF logger,
                                               final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert(dataStart >= 0);

            _getReadLock(rwl, "_readRawDataStartFromCache()", logger);
            try
                {
                // With cache locked read-only with main rwl lock
                // then we don't need the MetaData instance lock,
                // and not taking this allows concurrent disc reads.
                final CachedFile cf = exhibitGetInfo(name);

                // Can we satisfy some of the request from cache?
                // This can happen with the cache read-only.
                if((cf != null) && (dataStart < cf.cachedLength))
                    {
//System.out.println("Satisfying leading part of exhibitRead() from cache r/o ["+dataStart+","+canDo+"] for " + name);
                    // Note raw data cache hit...
                    StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHERAWDATAHIT);
                    // Do the read as fast as possible; no timestamps...
                    final int posBefore = buf.position();
                    cf.getCachedData(cacheDir, dataStart, buf, true);
                    return(buf.position() - posBefore); // # bytes read.
                    }
                }
            finally { rwl.readLock().unlock(); }
            return(0); // Couldn't read anything.
            }

        /**Thread-safe Map of which exhibits we are currently extending and when we started extending.
         * Used to help reduce the probability of redundant concurrent fetches
         * of the same data by multiple threads.
         * <p>
         * The key is a unique (new) String of the full exhibit name.
         * <p>
         * We notify on this key object instance when removing an entry.
         */
        private transient /*final*/ Map<String, Long> _beingExtended =
            new Hashtable<String, Long>();


        /**Get the cached thumbnails for an exhibit, or null if none cached.
         * This needs to access disc and so will need to
         * hold the main cache while it does so, grabbing
         * the metaData instance lock inside the main lock
         * where both need to be held simultaneously.
         * <p>
         * If a problem is encountered deserialising thumbnails,
         * the cache is not-read-only, and AUTO_REPAIR_DURING_IO is true,
         * we may zap the offending serialised file to try
         * to recreate it later.
         * <p>
         * We will need to hold a write lock if updating timestamps
         * when accessing the thumbnails; this may greatly reduce concurrency.
         *
         * @param rwl  the main cache lock object; must not be null
         *
         * @throws java.io.IOException  in case of serious problems
         *     accessing the on-disc cache state
         */
        ExhibitThumbnails exhibitGetThumbnails(final ReentrantReadWriteLock rwl,
                                               final File cacheDir,
                                               final String name,
                                               final SimpleLoggerIF logger)
            throws IOException
            {
//System.out.println("MetaData.exhibitGetThumbnails(): requested to get thumbnail for: " + name);

            // Start by grabbing the right flavour of main cache lock.
            // We don't need a lock on this metadata instance,
            // and not taking one allows concurrent disc reads.
            if(THUMBNAIL_ACCESS_UPDATES_ACCESS_TIMESTAMP) { _getWriteLock(rwl, "exhibitGetThumbnails()", logger); }
            else { _getReadLock(rwl, "exhibitGetThumbnails()", logger); }
            try
                {
                // Do we know about this exhibit?
                // Also mark exhibit entry as accessed
                // if that is our policy.
                final CachedFile cf = THUMBNAIL_ACCESS_UPDATES_ACCESS_TIMESTAMP ?
                    exhibitMarkAsAccessed(rwl, name, logger, cacheDir) : exhibitGetInfo(name);

                if((cf == null) || (cf.tnBytes == 0))
                    { return(null); } // None cached.

                // OK, try to get extant thumbnail,
                return(cf.getThumbnails(cacheDir));
                }
            finally
                {
                if(THUMBNAIL_ACCESS_UPDATES_ACCESS_TIMESTAMP) { rwl.writeLock().unlock(); }
                else { rwl.readLock().unlock(); }
                }
            }

        /**Saves the thumbnails for an exhibit.
         * Returns true if successful, false if not.
         * <p>
         * This will fail if the metaData is read only.
         * <p>
         * If the thumbnails argument is null, this will purge
         * any extant thumbnails for this exhibit.
         * <p>
         * This will create a new zero-length exhibit entry if
         * necessary.
         * <p>
         * Since these failures will be by returning false
         * rather than by causing an exception then this
         * can be used as a gentle unconditional way to make sure that
         * an entry does exist whether or not one did before.
         * <p>
         * This needs to access disc and so will need to
         * hold the main lock while it does so, grabbing
         * the metaData instance lock inside the main lock
         * where both need to be held simultaneously.
         * <p>
         * Note that this will have to grab a write lock.
         *
         * @param rwl  the main lock object; must not be null
         * @param esa  the basic info on the exhibit to have an entry created
         *
         * @throws java.io.IOException  in case of serious problems
         *     accessing the on-disc cache state
         */
        boolean exhibitSaveThumbnails(final ReentrantReadWriteLock rwl,
                                      final GenProps gp,
                                      final File cacheDir,
                                      final ExhibitStaticAttr esa,
                                      final ExhibitThumbnails tns,
                                      final SimpleLoggerIF logger,
                                      final StatsLogger.StatsConfig statsIDSCGEN)
            throws IOException
            {
            assert (rwl != null) && (gp != null) && (cacheDir != null) && (esa != null);

            // Fail gently if read-only.
            if(!isReadWrite()) { return(false); }

//System.out.println("MetaData.exhibitSaveThumbnails(): requested to "+(tns==null?"remove":"save")+" thumbnail for: " + esa.filePath);

            // Grab both locks in correct order.
            _getWriteLock(rwl, "exhibitSaveThumbnails()", logger);
            try
                {
                synchronized(this)
                    {
                    // Do we know about this exhibit?
                    CachedFile cf = exhibitGetInfo(esa.filePath);

                    if(cf == null)
                        {
                        // Apparently would be a new exhibit...

                        // If asked to purge thumbnails, nothing to do!
                        if(tns == null) { return(true); } // Easy!

                        // Make space for new cache entry...
                        purge(rwl,
                             gp,
                             cacheDir,
                             0,
                             esa.filePath,
                             logger, statsIDSCGEN);

                        // If asked to save thumbnails
                        // then we have to make a new cache entry.
                        if(!exhibitCreateNewCacheEntry(rwl,
                                                       cacheDir,
                                                       esa,
                                                       logger,
                                                       statsIDSCGEN))
                            { return(false); } // Could not make new entry.
                        }

                    // Mark the cache entry as accessed
                    // (since we want to avoid discarding what we just saved!)
                    cf = exhibitMarkAsAccessed(rwl, esa.filePath, logger, THUMBNAIL_ACCESS_UPDATES_ACCESS_TIMESTAMP ? cacheDir : null);
                    assert(cf != null);

                    // If zapping thumbnails,
                    // or if they already exist,
                    // zap them now.
                    if((tns == null) || cf.hasThumbnails())
                        { cf = cf.zapThumbnails(cacheDir); }

                    // If saving new thumbnails...
                    if(tns != null)
                        {
                        // Make space for new thumbnails,
                        // using worst-case size estimate.
                        purge(rwl,
                             gp,
                             cacheDir,
                             Math.max(1, ExhibitThumbnails.MAX_BYTES_EST),
                             esa.filePath,
                             logger, statsIDSCGEN);

                        // Now actually do the save.
                        cf = cf.saveThumbnails(cacheDir, tns);
                        }

                    // Save updated metadata...
                    _update(rwl, cf, logger);

                    return(true); // Done!
                    }
                }
            finally { rwl.writeLock().unlock(); }
            }




        /**My serial ID. */
        private static final long serialVersionUID = -7596195262594968694L;

        /**Write out a less-redundant form of our internal information.
         * Prevent (lazy-evaluation-caused) state changes while serialising
         * by being synchronized.
         * <p>
         * The more-efficient on-the-wire format also makes defensive
         * reading easier.
         */
        private synchronized void writeObject(final ObjectOutputStream oos)
            throws IOException
            {
            // Write the fields that we are not trying to optimise.
            // Note that this includes our length field.
            oos.defaultWriteObject();

            // Send the primary map as a count and then the CachedFile objects
            // in name-sorted order to attempt to enhance compression
            // (by improving locality, with similar items close to one another).
            final int size = cachedExhibits.size();
            final CachedFile outValues[] = new CachedFile[size];
            cachedExhibits.values().toArray(outValues);
            Arrays.sort(outValues, (new Comparator<CachedFile>(){
                public final int compare(final CachedFile o1, final CachedFile o2)
                    { return(o1.name.compareTo(o2.name)); }
                }));
            oos.writeInt(size);
            for(int i = size; --i >= 0; )
                { oos.writeObject(outValues[i]); }
            }

        /**Deserialise. */
        private synchronized void readObject(final ObjectInputStream in)
            throws IOException, ClassNotFoundException
            {
            in.defaultReadObject();

            // Defensively (and efficiently) read our primary map.
            final int size = in.readInt();
            if(size < 0)
                { throw new InvalidObjectException("bad stream: cachedExhibits.size() < 0"); }
            // Make suitably-sized new map.
            cachedExhibits = new Hashtable<String, CachedFile>(size * 2 + 1);
            // Read in entries from stream.
            for(int i = size; --i >= 0; )
                {
                final CachedFile cf = (CachedFile) in.readObject();
                cachedExhibits.put(cf.name, cf);
                }
            assert(size == cachedExhibits.size()); // Should be no duplicates.

            // Update derived data.
            _recomputeDerivedValues();

            // Nothing currently being extended.
            _beingExtended = new Hashtable<String, Long>();

            validateObject(); // Validate state immediately.
            }

        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         * <p>
         * Assumes _recomputeDerivedValues() has been called
         * before this is called since deserialisation or construction.
         */
        public synchronized void validateObject()
            throws InvalidObjectException
            {
            // Check that all components are sane and safe.

            if(_beingExtended == null)
                { throw new InvalidObjectException("bad object: _beingExtended == null"); }

            // cachedExhibits must be a Map from String to CachedFile.
            if(cachedExhibits == null)
                { throw new InvalidObjectException("bad object: cachedExhibits == null"); }
            // Check that the Map entries are correct and consistent.
            for(final Iterator<String> it = cachedExhibits.keySet().iterator(); it.hasNext(); )
                {
                final String name = it.next();
                if(!ExhibitName.validNameSyntax(name))
                    { throw new InvalidObjectException("bad object: cachedExhibits has invalid name as key"); }
                final CachedFile cf = cachedExhibits.get(name);
                if((cf == null) || !name.equals(cf.name))
                    { throw new InvalidObjectException("bad object: cachedExhibits has invalid value"); }
                }

            // Do some simple validation of derived (transient) values.
            if(totalBytes < 0)
                { throw new InvalidObjectException("bad object: totalBytes < 0"); }
            if(totalBytesHighWaterMark < 0)
                { throw new InvalidObjectException("bad object: totalBytesHighWaterMark < 0"); }
            if((exhibitsLRU == null) || (exhibitsLRU.size() != cachedExhibits.size()))
                { throw new InvalidObjectException("bad object: exhibitsLRU is invalid"); }
            }

        }

    /**In-memory copy of whole-cache meta-data; never null.
     * Note that the read/write status may change at any time.
     * <p>
     * The instance is never replaced; the state is replaced
     * in-situ if need be to ensure that we never have two
     * instances of this that believe they control the disc cache.
     */
    private final MetaData metaData;

    /**Return directly the number of partly- or fully- cached exhibits; never negative.
     * This may be more than the number of exhibits,
     * for example before deleted/renamed exhibits are removed.
     */
    public int getLiveCachedExhibitCount()
        {
        return(metaData.size());
        }


    /**Get the an instance copy of this class; may be a singleton.
     * If operating as a singleton then this
     * creates an instance on the first call; all subsequent requests/calls
     * are vetoed (at least in this servlet context and thus namespace)
     * unless the cacheDir matches that for the extant instance
     * in which case the new dataSource is ignored and the extant cache
     * instance is returned.
     *
     * @throws java.lang.IllegalStateException  if this is a singleton,
     *     and a request to create with a different cache dir to an
     *     extant instance is made
     *
     * @throws IOException  if cache directory does not exist
     *     and/or cannot be created
     *     (the containing directory passed in must always exist)
     */
    public static ExhibitDataSimpleCache cacheFactory(
                                    final SimpleExhibitPipelineIF dataSource,
                                    final File cacheDir,
                                    final SimpleLoggerIF logger)
        throws IOException,
               IllegalStateException
        { return(new ExhibitDataSimpleCache(dataSource, cacheDir, logger)); }


    /**Remove a persistent cache.
     * Pass in the same _cacheDir as for a call to cacheFactory(),
     * but this must not be called if a cache instance is using the cache.
     * <p>
     * Will not remove entries in cacheDir unrelated to the cache.
     * <p>
     * Do not use this lightly; it may discard gigabytes of useful state.
     *
     * @throws IOException in case of I/O error
     */
    public static void rmCache(final File cacheDir)
        throws IOException
        {
        if(cacheDir == null)
            { throw new IllegalArgumentException(); }

        // Potentially very expensive/significant event,
        // so always report it.
        System.out.println("ExhibitDataSimpleCache: removing cache below: " + cacheDir);

        final File baseDir = new File(cacheDir, CACHE_BASE_DIR);
        FileTools.rmRecursively(baseDir);
        }



    /**The upstream data source; never null. */
    private final SimpleExhibitPipelineIF source;

    /**The cache dir, else null.
     * If this is not a valid dir at class creation time we ensure
     * that we save a null here.
     */
    private final File cacheDir;

    /**The read/write lock for the whole cache except system variables; never null.
     * Any access that may update the cache state in memory or on disc
     * must be protected by a write lock.
     * <p>
     * Any access that may just read the cache need only have a read lock.
     * <p>
     * Most accesses may have to start by taking a write lock
     * (for example because they may have to fetch data from upstream
     * and insert it into the cache)
     * but can downgrade it to a read lock as soon as they know that they
     * will not be altering the cache at all or any further.
     * <p>
     * The main exception is any state component held in volatile fields.
     * <p>
     * Note that the variable store is internally thread-safe and does
     * not require protection by this lock.
     * <p>
     * We want performance (ie best throughput) rather than fairness.
     * <p>
     * We generally do not want lock attempts to block forever,
     * which means that we give up attempting to obtain a lock after
     * a given number of attempts (with a maximum time per attempt).
     * While this is intended to prevent to limit delays in the face of I/O problems,
     * this <em>may</em> rescue us from logic errors in extremis.
     */
    private final ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();

    /**Maximum number of consecutive attempts to obtain lock for read or write; strictly positive. */
    private static final int MAX_LOCK_ATTEMPTS = 10;

    /**Stack offset for _getXXXXLock() to find caller's stack frame. */
    private static final int _gXL_offset = 3;

    /**Get the cache write lock, complaining/aborting if we have to wait for a long time.
     * @param rwl  cache lock; never null
     * @throws InterruptedIOException  if the thread is interrupted or locking is aborted
     */
    private static void _getWriteLock(final ReentrantReadWriteLock rwl,
                                      final String detail,
                                      final SimpleLoggerIF logger)
        throws InterruptedIOException
        {
        int lockAttemptsLeft = MAX_LOCK_ATTEMPTS;
        try
            {
            while(!rwl.writeLock().tryLock(30, TimeUnit.SECONDS))
                {
                final StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
                final String caller =
                    ((stackTrace == null) || (stackTrace.length <= _gXL_offset) || (stackTrace[_gXL_offset] == null)) ? "[no trace]" :
                    (stackTrace[_gXL_offset].toString());
                logger.log("[ExhibitDataSimpleCache: waiting to obtain write lock: "+rwl.writeLock()+" "+detail+": "+caller+" isWriteLocked()="+rwl.isWriteLocked()+" getReadLockCount()="+rwl.getReadLockCount()+"]");
                if(--lockAttemptsLeft < 0) { throw new InterruptedIOException("too write many lock attempts"); }
                }
            }
        catch(final InterruptedException x)
            {
            x.printStackTrace();
            throw new InterruptedIOException("ExhibitDataSimpleCache: interrupted while waiting to obtain write lock..." +detail+": " + x.getMessage());
            }
        }

    /**Get a cache read lock, complaining/aborting if we have to wait for a long time.
     * We complain sooner waiting for a read lock rather than the write lock,
     * since read locks are expected to be easier/quicker to obtain.
     *
     * @param rwl  cache lock; never null
     * @throws InterruptedIOException  if the thread is interrupted or locking is aborted
     */
    private static void _getReadLock(final ReentrantReadWriteLock rwl,
                                     final String detail,
                                     final SimpleLoggerIF logger)
        throws InterruptedIOException
        {
        int lockAttemptsLeft = MAX_LOCK_ATTEMPTS;
        try
            {
            while(!rwl.readLock().tryLock(11, TimeUnit.SECONDS))
                {
                final StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
                final String caller =
                    ((stackTrace == null) || (stackTrace.length <= _gXL_offset) || (stackTrace[_gXL_offset] == null)) ? "[no trace]" :
                    (stackTrace[_gXL_offset].toString());
                logger.log("[ExhibitDataSimpleCache: waiting to obtain read lock: "+rwl.writeLock()+" "+detail+": "+caller+"]");
                if(--lockAttemptsLeft < 0) { throw new InterruptedIOException("too many read lock attempts"); }
                }
            }
        catch(final InterruptedException x)
            {
            x.printStackTrace();
            throw new InterruptedIOException("ExhibitDataSimpleCache: interrupted while waiting to obtain read lock..." +detail+": " + x.getMessage());
            }
        }


    /**Wrap a new cache instance around a data source.
     * This is private so that we can enforce a singleton pattern
     * and avoid multiple simultaneous users of the underlying file-based
     * cache.
     * <p>
     * We try to load the cache meta-data and exhibit properties
     * from persisted copies.  We can survive without the exhibit
     * properties, but if we can't load our meta data the default
     * value we use is read-only so that we don't trust it until
     * it's been checked against disc, presumably in the background.
     *
     * @throws IOException  if cache directory does not exist
     *     and/or cannot be created
     *     (the containing directory passed in must always exist)
     */
    private ExhibitDataSimpleCache(final SimpleExhibitPipelineIF dataSource,
                                   final File cacheDir,
                                   final SimpleLoggerIF logger)
        throws IOException
        {
        if(dataSource == null)
            { throw new IllegalArgumentException("null source"); }
        if(cacheDir == null)
            { throw new IllegalArgumentException("null cacheDir"); }
        if(logger == null)
            { throw new IllegalArgumentException("null logger"); }

        final long startTime = System.currentTimeMillis();

        source = dataSource;

        this.logger = logger;
        statsIDSCGEN =
            new StatsLogger.StatsConfig("SIMPLECACHE-GENERAL",
                                        logger,
                                        false, // Only dump summaries...
                                        8 * 3600, // About every 8 hours.
                                        true); // Adaptive.

        // Set up our variable manager.
        varMgr = new PipelineVarMgr(dataSource, false);
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons [1] "+(System.currentTimeMillis()-startTime)+".]"); }

        // Use the cache dir only if it looks reasonably plausable.
if(IsDebug.isDebug) { logger.log("[ExhibitDataSimpleCache: cache dir = " + cacheDir + "]"); }
        if(!cacheDir.isDirectory())
            { throw new IOException("missing cache directory " + cacheDir); }
        this.cacheDir = cacheDir;

if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons [2] "+(System.currentTimeMillis()-startTime)+".]"); }

        // If things are still looking rosy then try to load the
        // exhibit properties and cache meta-data immediately...
        // We quietly give up if we have any problems with loading,
        // eg due to a corrupt file or a changed class definition.

        // Attempt to make our base cache directory if necessary.
        final File baseDir = new File(this.cacheDir, CACHE_BASE_DIR);
        if(!baseDir.isDirectory()) { baseDir.mkdirs(); }
        if(!baseDir.isDirectory())
            { throw new IOException("cannot make cache data store directory: " + baseDir); }


        // Load any persisted/cached event history data that we can find.
        // It may be too stale to be usable,
        // but if so then that will be dealt with automatically.
        // Absorb but report any errors encountered...
        // We may have to make the history dir, eg if this is the first run.
        final File evhd = new File(cacheDir, EVENT_HISTORY_DIR);
        if(!evhd.isDirectory()) { evhd.mkdirs(); }
        // We need this done before we can do much else...
        varMgr.loadEventHistories(evhd);
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons [3] "+(System.currentTimeMillis()-startTime)+".]"); }


        // Initialise cache metadata state.
        MetaData md = new MetaData(); // Read-only empty cache info.
        // Try to load cached meta-data...
        try {
            final File f = new File(this.cacheDir, CACHE_METADATA_FILENAME);
            if(f.isFile() && f.canRead())
                {
                final MetaData tmpMd = (MetaData) FileTools.deserialiseFromFile(f, STORE_EXPROPS_GZIPED);
                if(tmpMd != null)
                    {
                    md = tmpMd;
                    logger.log("[ExhibitDataSimpleCache: loaded metaData, exhibit count: "+tmpMd.size()+".]");
                    }
                }
            }
        catch(final Exception e) { e.printStackTrace(); } // Complain!
        metaData = md; // Capture cache meta-data.
        if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons [4] "+(System.currentTimeMillis()-startTime)+".]"); }

        
        // Try to load cached exhibit properties
        // (after everything else is set up and ready).
        // This may take significant time so is worth doing asynchronously.
        // We allow this to continue asynchronously after the constructor returns
        // to overlap with other work.
        // We assume that this is significantly I/O-bound.
        _AEPPoller = ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                try {
                    final File f = new File(cacheDir, CACHE_EXPROPS_FILENAME);
                    if(f.isFile() && f.canRead())
                        {
                        logger.log("[ExhibitDataSimpleCache: reloading cached AEP from disc...]");
                        _gAEP_lock.lock();
                        try
                            {
                            final AllExhibitProperties aep =
                                (AllExhibitProperties) FileTools.deserialiseFromFile(f, STORE_EXPROPS_GZIPED);
                            if(aep != null)
                                {
                                // Now cache in memory.
                                _getAllExhibitProperties_postUpdate(aep, true);
                                logger.log("[ExhibitDataSimpleCache: loaded AEP, exhibit count: "+aep.aeid.size()+".]");
                                }
                            }
                        finally { _gAEP_lock.unlock(); }
                        }
                    }
                catch(final Exception e) { e.printStackTrace(); } // Complain!
                }
            });
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons [5] "+(System.currentTimeMillis()-startTime)+".]"); }

        
        // If we loaded some exhibit properties but no metadata,
        // then this may indicate a nasty problem...
        // We may not be able to tell if the AEP has not yet completed loading,
        // so we do this test as late as possible.
        if(md.isEmpty() && (_AEP.aeid.size() > 0))
            {
            logger.log("WARNING: ExhibitDataSimpleCache: NOT loaded metaData.");
            // Leave metadata read-only, and check it ASAP...
            }
        // Else, if we being optimisitic, mark meta-data read/write immediately.
        else if(ASSUME_LOADED_METADATA_OK)
            {
            md.setReadWrite(true);
            // Put off first metadata check as it is probably OK.
            _checkMetaData_notBefore = System.currentTimeMillis() +
                (15 * 60 * 1000) + Rnd.fastRnd.nextInt(30 * 60 * 1000);
            }

        final long endTime = System.currentTimeMillis();
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: cons time "+(endTime-startTime)+".]"); }
        }

    /**An AllExhibitProperties.ExhibitDataSource wrapping ourselves; never null. */
    private final AllExhibitProperties.ExhibitDataSource exhibitDataSource =
        (new AllExhibitProperties.ExhibitDataSource(){
                public final void getRawFile(final ByteBuffer buf, final String exhibitName, final int position)
                    throws IOException
                    { ExhibitDataSimpleCache.this.getRawFile(buf, exhibitName, position, false); }
                /**Fully loaded if all in cache.
                 * False if exhibit not present or not full length.
                 */
                public final boolean isExhibitFullyLoaded(final ExhibitStaticAttr esa)
                    //throws IOException
                    {
                    assert(ExhibitDataSimpleCache.this != null);
                    assert(metaData != null);
                    return(ExhibitDataSimpleCache.this.
                        metaData.exhibitIsFullyLoaded(esa));
                    }
                });


    /**Maximum number of threads that may run in _asyncTNFetch() and other local discardable data read-ahead tasks; strictly positive.
     * We limit the amount of threading by:
     * <ul>
     * <li>The number of available CPUs (as sometimes this work may be CPU intensive).
     * <li>The likely limit on back-end connectivity (eg HTTP-connection-count limited)
     *     (given that this cache will often by upstream of an HTTP tunnel).
     * <li>Likely limits on local and upstream disc throughput.
     * <li>The likely strain on other resources, such as memory.
     * </ul>
     * This limit/count/cap should generally be &gt;1 since the work is mainly I/O bound,
     * but should generally be not much more than (say)
     * half the maximum simultaneous outbound tunnel HTTP connection count.
     */
    private static final int MAX_THREADS_aTWQ = Math.max(2, Math.min((ExhibitDataHTTPTunnelSource.MAX_CONCURRENT_CONNECTIONS/2)+1, Runtime.getRuntime().availableProcessors()));

    /**Shared thread pool for I/O-bound activities for Web-driven requests.
     * Suitable for mainly-I/O-bound threads, thus we have a fixed thread limit.
     * This limit also protects upstream servers from excess load.
     * <p>
     * A limited amount of work can be queued,
     * but excess is handled by discarding the oldest queued items silently.
     * <p>
     * The threads in the pool are daemon threads,
     * so will not prevent the JVM from exiting.
     */
    private final ExecutorService discardableReadAheadTaskThreadPool =
        new ThreadPoolExecutor(0, MAX_THREADS_aTWQ,
            300L, TimeUnit.SECONDS, // Keep worker threads alive for 5 minutes...
            new ArrayBlockingQueue<Runnable>(128), // Allow some work to be queued.
            new ThreadUtils.DaemonThreadFactory("ExhibitDataSimpleCache.discardableReadAheadTaskThreadPool", false),
            new ThreadPoolExecutor.DiscardOldestPolicy());

    /**Get a chunk of the raw exhibit binary.
     * The name, start byte offset/position and a buffer to fill are supplied.
     * 
     * @param position  must be non-negative and less than the exhibit size in bytes
     * @param dontCache  if true do not cache locally, unless we have lots of free space
     */
    public void getRawFile(final ByteBuffer buf,
                           final String exhibitName, final int position,
                           final boolean dontCache)
        throws IOException
        {
        // Note inbound request for raw exhibit data.
        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_EXDATAREQIN);
        if(dontCache)
            { StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_EXDATAREQINDC); }

        try
            {
            final AllExhibitImmutableData aeid = _AEP.aeid;
            boolean succeeded = false;
            try
                {
                metaData.exhibitRead(rwl, null, cacheDir, exhibitName, source,
                    aeid,
                    genProps,
                    position,
                    buf,
                    dontCache ? Boolean.TRUE : Boolean.FALSE,
                    logger, statsIDSCGEN);
                succeeded = true;

                // Any activity against this cache instance spurs us into life.
                if(!_userRequestedDataFromCache)
                    {
                    _userRequestedDataFromCache = true;
                    logger.log("INFO: user requested data from cache: precacheing may start");
                    }
                }
            finally
                {
                // Be prepared to spin off a read-ahead thread for any uncached tail portion.
                // Only do this if precacheing is currently allowed,
                // and if this appears to be a normal data request from an end user,
                // eg via the exhibit servlet.
                // This prevents precache requests (ie not directly human-requested)
                // from being wastefully routed through many intermediate servers
                // and/or causing a huge avalanche of unnecessary activity.
                //
                // IF THE NORMAL FETCH FAILED (eg disc trouble, or master down)
                // THEN RETRY THIS FETCH FROM A PEER TO ATTEMPT RECOVERY OF THE DATA.
                // (This recovery attempt happens in the background, ie asynchronously.)
                if(ALLOW_DATA_FETCH_FROM_PEERS && (!dontCache) &&
                   ((!succeeded) || metaData.canPrecacheExhibitData()))
                    {
                    final boolean peerFetchForRecovery = !succeeded;

                    final ExhibitStaticAttr esa = aeid.getStaticAttr(exhibitName);
                    final CachedFile cf = metaData.exhibitGetInfo(exhibitName);

                    // Only proceed if it looks like there is something useful to be done.
                    if((esa != null) && (cf != null))
                        {
                        // Compute maximum number of tail bytes of exhibit we could logically cache.
                        final int tailBytesStillCacheable =
                            (int) Math.max(0, Math.min(esa.length, genProps.getWEBSVR_MAX_CACHEABLE_EX_BYTES()) - cf.cachedLength);
                        // Only actually do the precacheing if the end of user's request
                        // (and thus where we anticipate that the next one may start)
                        // is close to (or beyond) the end of what we have already cached.
                        // (Or if we are recovering from an error.)
                        if((tailBytesStillCacheable > 0) &&
                           (!succeeded || (position + 4*MAX_TRANSFER_CHUNK_SIZE >= cf.cachedLength)))
                            {
                            discardableReadAheadTaskThreadPool.submit(new Runnable(){
                                public final void run()
                                    {
                                    try
                                        {
                                        // Recheck just before running that this is still appropriate.
                                        if(!peerFetchForRecovery && !metaData.canPrecacheExhibitData()) { return; /* Abort. */ }
                                        final CachedFile cf2 = metaData.exhibitGetInfo(exhibitName);
                                        if((cf2 == null) || (cf2.cachedLength != cf.cachedLength)) { return; /* Abort. */ }

                                        // Cache from the end of what we have already cached,
                                        // regardless of the caller's start position.
                                        _getExhibitDataFromUpstreamToPrecache(esa,
                                                    aeid,
                                                    genProps,
                                                    cf.cachedLength,
                                                    Math.min(tailBytesStillCacheable, MAX_TRANSFER_CHUNK_SIZE),
                                                    peerFetchForRecovery); // Is this an error-recovery attempt?
                                        }
                                    catch(final IOException e)
                                        { /* Silently discard errors from this speculative work. */ }
                                   }
                                });
                            }
                        }
                    }
                }
            }
        finally
            {
            // We must not be holding a (write) lock on exit.
            // We log any held lock and break it and propagate an Error.
            // Of course, this should NEVER happen,
            // but the cost of checking for it is low, so we'll always check.
            if(rwl.isWriteLockedByCurrentThread())
                {
                rwl.writeLock().unlock();
                System.err.println("SEVERE ERROR: getRawFile(): holding write lock on exit from cache");
                throw new Error("SEVERE ERROR: getRawFile(): holding write lock on exit from cache");
                }
            }
        }

    /**Get the static attributes for a given exhibit.
     * We get this from our cache of the immutable data rather than
     * going to the source directly.
     * We don't block or hold any locks to fetch this.
     * <p>
     * Returns null if the named exhibit does not exist.
     */
    public final ExhibitStaticAttr getStaticAttr(final String name)
        //throws IOException
        {
        // Return value from cache without blocking...
        return(_AEP.aeid.getStaticAttr(name));
        }

    /**Gets all static exhibit data if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed then a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     * <p>
     * We get this from our cache of the immutable data rather than
     * going to the source directly.
     * We don't block or hold any locks to fetch this.
     */
    public final AllExhibitImmutableData getAllExhibitImmutableData(final long oldStamp)
        //throws IOException
        {
        final AllExhibitImmutableData cached = _AEP.aeid;

        if((oldStamp < 0) || (oldStamp != cached.timestamp))
            { return(cached); }

        // Caller seems to have up-to-date copy already.
        return(null);
        }

    /**Gets set of all exhibit properties if its hash is not that specified.
     * If the hash specified is negative the object will be returned unconditionally.
     * <p>
     * If no exhibits are currently installed a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldHash
     * matches that that would have been returned) null is returned.
     * <p>
     * We get this from our cache rather than
     * going to the source directly.
     * We don't block or hold any locks to fetch this.
     */
    public AllExhibitProperties getAllExhibitProperties(final long oldHash)
        //throws IOException
        {
        final AllExhibitProperties cached = _AEP;

        if((oldHash < 0) || (oldHash != cached.longHash))
            { return(cached); }

        // Caller seems to have up-to-date copy already.
        return(null);
        }

    /**Private lock for _getAllExhibitProperties()/constructor to prevent re-entry and multiple concurrent AEP fetches. */
    private final ReentrantLock _gAEP_lock = new ReentrantLock();

    /**Minimum time before attempting to poll again for AEP while we don't have a real one loaded (ms). */
    private static final int MIN_AEP_POLL_TIME_UNTIL_LOADED_MS = 10291;

    /**Attempts to get all exhibit properties if our cached copy may be stale.
     * Because fetching/computing this value can take a very long time
     * (upwards of several tens of minutes)
     * we attempt to split the activity into two parts, and
     * have the actual computation/fetch done in the background,
     * and then an atomic post of the results back to the cache proper.
     * <p>
     * We also adaptively attempt to use an AEP diff fetch if one is available
     * (ie if the underlying connection is a tunnel).
     */
    private void _getAllExhibitProperties()
        {
        // Normal we try polling for the AEP much more often
        // if we don't yet have a valid set in place.
        final boolean aepNotYetLoaded = (_AEP.aeid.length == 0);

        // We return immediately if no need to poll,
        // because last poll was recent enough
        // and we have a valid (non-empty) AEP in place.
        final int serverSlowdownFactor = LocalProps.getServerSlowdownFactor();
        if(System.currentTimeMillis() - _lastPollAEP <= serverSlowdownFactor *
               (aepNotYetLoaded ? MIN_AEP_POLL_TIME_UNTIL_LOADED_MS
                                : genProps.getWEBSVR_MIN_EX_IMATTR_RECHECK_MS()))
            { return; }

        // If we can't get the lock immediately
        // then give up immediately and try again later.
        if(!_gAEP_lock.tryLock()) { return; }
        try
            {
            // Postpone next fetch now just in case we fail to start new thread
            // or the old one is still running, ie very slow.
            _lastPollAEP = System.currentTimeMillis();

            // Quit immediately if we're still busy completing a fetch...
            if(!_AEPPoller.isDone())
                { return; } // Still busy...

            // Capture current AEP.
            final AllExhibitProperties curAEP = _AEP;

            // Create new background thread to get AEP...
            _AEPPoller = ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    final long startTime = System.currentTimeMillis();
                    try
                        {
                        final AllExhibitProperties new_AEP;
                        // Try to fetch new value; this might take a while or even fail...
                        if(source instanceof ExhibitDataTunnelSource)
                            {
                            // New, potentially faster way to fetch AEP.
                            new_AEP = ((ExhibitDataTunnelSource) source).getAllExhibitProperties(curAEP, true);
                            }
                        else // Use generic mechanism.
                            { new_AEP = source.getAllExhibitProperties(curAEP.longHash); }
                        // Post the update back to the cache...
                        // even if the value is null
                        // so as to let us the save current values to disc.
                         _getAllExhibitProperties_postUpdate(new_AEP, false);
                        }
                    // If an error occurs then we won't post an update
                    // of the AEP value and will have to try again later.
                    catch(final Throwable e) // We'll note that an error occurred, but otherwise ignore it.
                        {
                        logger.log("[ExhibitDataSimpleCache: error trying to fetch and update AEP in background: " + e.getMessage() + ".]");
                        e.printStackTrace();
                        }
                    finally
                        {
                        final long endTime = System.currentTimeMillis();
                        _lastPollAEP = endTime; // Schedule next poll.

                        // For the record note how long we took to get AEP
                        // unless it took very little time at all.
                        final long ms = endTime - startTime;
                        if(ms >= 10001) { logger.log("[ExhibitDataSimpleCache: fetch/update of AEP in background took "+ms+"ms.]"); }
                        }
                    }
                });
            }
        finally { _gAEP_lock.unlock(); }
        }

    /**Accepts the value posted from the background thread.
     * Must grab a write lock to (potentially) update/change the cache.
     * <p>
     * If the static exhibit data was stale then we also clear our in-memory
     * raw-exhibit data cache entirely, to be refilled by slower means.
     *
     * @param inCons  if true then this was called from the constructor
     *     so we don't save the AEP nor do some other expensive things
     *     that may rely on external mechanisms not yet set up
     */
    private void _getAllExhibitProperties_postUpdate(final AllExhibitProperties new_AEP,
                                                     final boolean inCons)
        throws IOException
        {
        // If the size of the new exhibit list is zero
        // then discard it on the grounds that it is almost certainly an error.
        if((new_AEP != null) && (new_AEP.aeid.size() == 0))
            {
            logger.log("[ExhibitDataSimpleCache: rejecting null/zero-size AEP update at "+(new Date())+".]");
            return;
            }

        // We only need to consider doing an update if we got
        // a non-null response.
        if(new_AEP != null)
            {
            // OK, we got an updated value.

            // Log the receipt of a new exhibit set.
            final String message = "[ExhibitDataSimpleCache: new AEP: " +
                "old/new timestamp and exhibit count: " +
                (new Date(_AEP.aeid.timestamp)) + " " + _AEP.aeid.length + " / " +
                (new Date(new_AEP.aeid.timestamp)) + " " + new_AEP.aeid.length +
                " at " +(new Date()) + ".]";
            logger.log(message);

            // We attempt to recover some expensive-to-recompute data...
            // This has to be done before any other AEP activity
            // else it may be blocked for security reasons
            // and any extant expensive-to-compute old EPCM may be lost.
            new_AEP.recoverOldExhibitPropsComputableMutableData(_AEP, logger);

            // Try to bring the vote cache up-to-date
            // using a pool thread to allow other activities to proceed.
            // This is likely to be I/O bound and may take seconds-->minutes.
            ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
                public final void run()
                    { try { new_AEP.updateVoteCache(varMgr, true); } catch(final Throwable t) { t.printStackTrace(); } }
                });

            // In a pool thread precompute/cache some often-used values.
            // This should result in better performance seen by users.
            ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
                public final void run()
                    {
                    new_AEP.getTotalExhibitBytes();
                    new_AEP.getCategoryExhibitCounts();
                    new_AEP.aeid.getAllExhibitNamesSorted();
                    }
                });

            // OK, update the new cache-visible AEP value (atomically)...
            _AEP = new_AEP;

            logger.log("[ExhibitDataSimpleCache: warmed up and posted AEP.]");

            // Save new AEP to disc.
            if(!inCons)
                {
                _save_AEP();
                logger.log("[ExhibitDataSimpleCache: saved new AEP to disc.]");
                }

            // Start any compaction at low priority when we've finished...
            // If the system is potentially short of CPU, this will run synchronously,
            // ie this thread will block until compaction is complete.
            ThreadUtils.lowPriorityThreadPool.submit(new Runnable(){
                public final void run() { new_AEP.compact(); }
                });
            }
        }

    /**Save AllExhibitProperties to disc.
     * We should do this when we receive a new set from downstream
     * so that we can restart with the appropriate set,
     * and periodically to save any cached state that we have
     * accumulated.
     * <p>
     * In principle this needs a write lock to alter state on disc.
     * In practice at most one thread at once will ever try to call this,
     * and the serialiseToFile() routine attempts to atomicly replace the file,
     * and this can take a long time and thus needlessly block cache activity,
     * so we do not take a cache lock here.
     */
    private void _save_AEP()
        throws IOException
        {
//        _getWriteLock(rwl, "_save_AEP()");
//        try
            {
            // Replace any extant file atomically if possible.
            final AllExhibitProperties aep = _AEP;
            if((aep != null) && (aep.aeid.length > 0) &&
               (cacheDir != null) &&
               (cacheDir.isDirectory()) &&
               (cacheDir.canWrite()))
                {
                // Abort at last moment if cache instance shut down.
                if(destroyed) { return; }

                final File f = new File(cacheDir, CACHE_EXPROPS_FILENAME);
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[_save_AEP() start: "+ System.currentTimeMillis() +".]"); }
                FileTools.serialiseToFile(aep, f, STORE_EXPROPS_GZIPED, true);
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[_save_AEP() end:   "+ System.currentTimeMillis() +".]"); }
                }
            }
//        finally { rwl.writeLock().unlock(); }
        }

    /**Cached AllExhibitProperties; never null.
     * Volatile so that it can be safely accessed without a lock.
     */
    private volatile AllExhibitProperties _AEP = new AllExhibitProperties();

    /**Last time we polled for AllExhibitImmutableData; initially zero.
     * Private to _getAllExhibitProperties().
     */
    private volatile transient long _lastPollAEP;

    /**Future for task fetching _AEP in the background; never null after construction.
     * If this is done/cancelled then a collection is in progress.
     * <p>
     * Private to _getAllExhibitProperties() and the constructor.
     * <p>
     * Marked volatile for thread-safe lock-free read/write access.
     */
    private volatile Future<?> _AEPPoller;


    /**Time before which next _checkMetaData() call should not be initiated.
     * The initial check is usually put off a few minutes
     * since the system is often very busy on start-up,
     * and we don't expect significant problems anyway most of the time.
     * <p>
     * Volatile for thread-safe access without a lock.
     * <p>
     * Private to _checkMetaData() and _checkMetaData_postResults().
     */
    private volatile transient long _checkMetaData_notBefore;

    /**Private lock for _checkMetaData() to avoid starting more than one thread. */
    private final ReentrantLock _cMD_lock_ = new ReentrantLock();

    /**Initiates, in the background, a check of the in-memory cache meta data against disc.
     * Grabs a write lock on the cache and makes the cache read-only
     * while it works.
     * <p>
     * Refuses to do the check if it is too soon since the last one
     * or if the cache seems busy.
     */
    private void _checkMetaData()
        {
        // Don't try to check the cache if there isn't one!
        if(cacheDir == null) { return; }

        // Check that it is not too soon for another check;
        // if so, quit for another try later.
        if(_checkMetaData_notBefore >= System.currentTimeMillis())
            { return; } // Too soon; try again later!

        // Don't do the check now if the cache seems busy.
        if((rwl.getReadLockCount() > 0) ||
           (rwl.isWriteLocked()))
            { return; }

        // If a run seems to be in progess
        // then don't even try to start a new one...
        if(_cMD_lock_.isLocked())
            { return; }
        
        // Start background thread for immediate execution,
        // itself unencumbered by locks.
        ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                if(!_cMD_lock_.tryLock()) { return; }
                try
                    {
                    // Note when we start...
                    final long startTime = System.currentTimeMillis();
                    logger.log("[ExhibitDataSimpleCache: cache metadata check starting at "+(new Date(startTime))+".]");

                    try
                        {
                        // Compute the base dir of the exhibit data cache.
                        final File dataBaseDir = new File(cacheDir, CACHE_EXDATA_DIR);
                        // Try to make sure that the cache directory actually exists
                        // and is a directory and is writable;
                        // else zap it and try to remake it.
                        if(!dataBaseDir.isDirectory() || !dataBaseDir.canWrite())
                            {
                            dataBaseDir.delete();
                            dataBaseDir.mkdirs();
                            }

                        // Go and compute meta-data from what is on disc
                        // (and knowing what current exhibits to look for),
                        final MetaData newMd = new MetaData(_AEP, cacheDir, logger);

                        // Abort if cache is being shut down.
                        if(destroyed) { return; }

                        // Lock out other cache activity while we make sure
                        // that the meta-data is correct,
                        // or fix it if not.
                        // Grab both locks in correct order...
                        _getWriteLock(rwl, "_checkMetaData() thread", logger);
                        try
                            {
                            synchronized(metaData)
                                {
                                try
                                    {
                                    // Set working meta-data read-only just before we before
                                    // we cross-check in-memory and reconstructed versions.
                                    metaData.setReadWrite(false);

                                    // If the in-memory meta-data is empty,
                                    // mergeWithNewMetaData the new one if it is not empty.
                                    if(metaData.isEmpty())
                                        {
                                        if(!newMd.isEmpty())
                                            {
                                            logger.log("[ExhibitDataSimpleCache: loading in-memory cache metadata: exhibits/bytes = "+newMd.size()+"/"+newMd.getTotalBytes()+".]");
                                            metaData.mergeWithNewMetaData(newMd, logger);
                                            }
                                        }
                                    // ELSE if there is a serious inconsistency,
                                    // (and it's not just that the in-memory copy is empty)
                                    // complain and try to fix the in-memory copy.
                                    else if(!newMd.isEquivalent(metaData))
                                        {
                                        logger.log("[ExhibitDataSimpleCache: WARNING: on-disc and in-memory cache metadata inconsistent: attempting to fix.]");
                                        metaData.mergeWithNewMetaData(newMd, logger);
                                        }
                                    }
                                finally
                                    {
                                    // OK, the cache should be sane now,
                                    // so make it read/write,
                                    // and allow normal cache operations to resume.
                                    metaData.setReadWrite(true);
                                    }
                                }
                            }
                        finally { rwl.writeLock().unlock(); }

                        // Note when we finish main work...
                        final long endTimeMainWork = System.currentTimeMillis();
                        final long mainWorkTime = endTimeMainWork - startTime;

                        logger.log("[ExhibitDataSimpleCache: cache metadata check ended at "+(new Date(endTimeMainWork))+".]");

// NOW NORMAL CACHE OPERATION CAN RESUME WHILE WE DO SOME OTHER TIDY-UP...

                        // Spend a proportion of the time that we spent
                        // checking the metadata actually checking the data.
                        // Give up if we find the cache to have become
                        // write locked, ie if we may not be able to get in.
                        final long finishDataCheckBy = System.currentTimeMillis() + mainWorkTime*3;
                        do
                            {
                            // Check one cache entry for data corruption.
                            _incrCheckMRUExhibitEntries();
                            } while((System.currentTimeMillis() < finishDataCheckBy) &&
                                    !rwl.isWriteLocked());

                        // Remove orphaned temporary files
                        // and empty directories.
                        // Anything not modified for over a month is fair game.
                        // TODO: Zap any other crud that builds up.
                        final long old = System.currentTimeMillis() - (32 * 24 * 3600 * 1000L);
                        FileTools.rmRecursively(new File(cacheDir, CACHE_EXDATA_DIR),
                            new FileFilter(){
                            /**Tests whether or not the file with the specified abstract pathname should be removed.
                             * If a directory then everything within it
                             * is dealt with (eg possibly removed) first.
                             * <p>
                             * Only remove plain files that appear to be old
                             * temporary files (starting with the temp prefix),
                             * and empty dirs (can be recreated on demand).
                             *
                             * @param pathname The abstract pathname to be tested
                             * @return true iff <code>pathname</code>
                             *         should be deleted
                             */
                            public final boolean accept(final File pathname)
                                {
                                // Must be "old" (not modified for ~1M).
                                // This avoids races with other components
                                // actively creating/using temporary files
                                // and allows some grace time to allow for
                                // glitches in the exhibit set, etc.
                                if(pathname.lastModified() >= old)
                                    { return(false); }

                                // Empty directories can be zapped.
                                if(pathname.isDirectory())
                                    {
                                    final String[] files = pathname.list();
                                    if((files != null) && (files.length == 0))
                                        {
                                        logger.log("INFO: ExhibitDataSimpleCache: found empty cache directory for removal: " + pathname);
                                        return(true);
                                        }
                                    return(false); // Cannot zap this dir.
                                    }

                                // Anything not a dir and not a plain file
                                // is skipped/rejected.
                                if(!pathname.isFile())
                                    { return(false); }

                                // Temporary files start with a known prefix.
                                if(pathname.getName().startsWith(FileTools.F_tmpPrefix))
                                    {
                                    logger.log("WARNING: ExhibitDataSimpleCache: found orphaned/foreign cached temporary file for removal: " + pathname);
                                    return(true);
                                    }

                                // Whatever this is, we can't zap it...
                                return(false);
                                }
                            });

                        // Note when we finish all the work...
                            final long endTime = System.currentTimeMillis();

if(IsDebug.isDebug) { logger.log("[ExhibitDataSimpleCache._checkMetaData(): took "+(endTime - startTime)+"ms.]"); }

                        // Compute a postponement long enough
                        // that we don't spend more than about
                        // 1% of system (CPU/IO) time in the scan,
                        // though bounded from above and below
                        // in case of unusual events.
                        final long nextDiscCacheRecheckDue =
                            endTime +
                            Math.min(DISC_RECHECK_INTERVAL_MS * 3L,
                                Math.max(((endTime - startTime) * 100), // Max 1% of wall-clock time.
                                         DISC_RECHECK_INTERVAL_MS / 3L)) +
                            Rnd.fastRnd.nextInt(0xfff | (DISC_RECHECK_INTERVAL_MS / 8));

                        // Set the (volatile)`do-not-try-again-before' value...
                        _checkMetaData_notBefore = nextDiscCacheRecheckDue;

                        logger.log("[ExhibitDataSimpleCache: next cache metadata check due not before "+(new Date(nextDiscCacheRecheckDue))+".]");
                        }
                    catch(final IOException e)
                        {
                        // Complain lots if we could not even reload the cache.
                        e.printStackTrace();
                        }
                    }
//                catch(final IllegalStateException e)
//                    {
//                    e.printStackTrace();
//                    }
                finally { _cMD_lock_.unlock(); }
                }
            });
        }


    /**Time before which not to to save metaData again; private to _cleanAndSaveMetaData(). */
    private transient long _saveMetaData_notBefore; // Initially zero for immediate save.

    /**Saves the cache metadata if needed.
     * Grabs a write lock to update disc (and memory) state.
     * <p>
     * Aims to avoid saving the metaData more than once every
     * METADATA_MIN_SAVE_INTERVAL_MS, though if no save has taken
     * place for a while then the next save will happen on the
     * next call.
     * <p>
     * If saving the meta-data is taking a long time this aims
     * to postpone the next save at least a reasonable multiple of that
     * to avoid wasting too much system/CPU/disc bandwidth,
     * though we do put a cap on the maximum delay in case of weirdness...
     * <p>
     * This may also incrementally purge stale meta-data and data
     * just before the save to avoid the need for an extra meta-data save
     * to account for the purge-induced changes themselves.
     * <p>
     * This may also pick one or more exhibits at random to spot-check
     * for consistency with the master copy (eg looking for data corruption).
     *
     * @param force  if true, force an immediate save to disc
     */
    private void _cleanAndSaveMetaData(final boolean force)
        throws IOException
        {
        // If no save is needed then return immediately.
        if(!metaData.getNeedsSave()) { return; }

        // Too soon to consider another save, so return.
        if(!force && (_saveMetaData_notBefore >= System.currentTimeMillis())) { return; }

        // OK, attempt to clean the metadata before saving.
        // But if this produced an error,
        // don't let it stop us doing the same anyway.
        IOException lastErr = null;
        final long start = System.currentTimeMillis();
        try
            {
            // Incrementally purge any orphaned entries...
            _incrPurgeOrphanedExhibits();
            // Incrementally test integrity of most-recently-used entries...
            _incrCheckMRUExhibitEntries();
            }
        catch(final IOException e) { lastErr = e; }

        // Abort at last moment if cache instance shut down.
        if(destroyed) { return; }

        // Now actually save the cleaned-up metadata to disc.
        metaData.saveToDisc(cacheDir, logger, statsIDSCGEN);
        final long end = System.currentTimeMillis();
        final long took = end - start;

        // Postpone next save.
        // Limit the wait time from above and below.
        _saveMetaData_notBefore = System.currentTimeMillis() +
            METADATA_MIN_SAVE_INTERVAL_MS +
            Rnd.fastRnd.nextInt(1 + METADATA_MIN_SAVE_INTERVAL_MS/3) +
            Math.min(79*took, DISC_RECHECK_INTERVAL_MS/11);

        // Rethrow any error encountered in cleaning the metadata.
        if(lastErr != null) { throw lastErr; }
        }

    /**Incrementally check the most-recently-used cached exhibits for integrity.
     * This will attempt to remove any entry it finds that is corrupt.
     * <p>
     * This concentrates on the most recent entries as data corruption
     * in these would probably have the most serious effects.
     * <p>
     * This may examine any cached entry.
     * <p>
     * This may not examine any entry at all if the cache seems to be busy.
     * <p>
     * This will stop after removing at most one corrupt entry.
     *
     * @throws IOException
     */
    private void _incrCheckMRUExhibitEntries()
        throws IOException
        {
        final AllExhibitProperties aep = _AEP;
        final ExhibitStaticAttr esa;

        // We need only a cache read lock to find an exhibit to test.
        // We may then later need to grab a different lock to test data
        // or to remove corrupt data.

        // Grab both locks in correct order.
        // But give up if we can't get a cache lock immediately...
        if(!rwl.readLock().tryLock()) { return; }
        try
            {
            synchronized(metaData)
                {
                // We directly access the LRU list to get the most-recently-used
                // (and thus, probably, the most-frequently-used)
                // cached exhibits.
                //
                // We must hold the metaData instance lock
                // while reading this LRU data,
                // but we do *not* directly update it.
                //
                // The LRU data may change if we update/remove a corrupt entry.
                final SortedSet<CachedFile> lru = metaData.exhibitsLRU;
                final int size = lru.size();
                // If nothing to do then return immediately.
                if(size < 1) { return; }

//                // Half the time just get the most-recently-used exhibit...
//                if(Rnd.fastRnd.nextBoolean())
//                    {
//                    final CachedFile cf = lru.last();
//                    esa = aep.aeid.getStaticAttr(cf.name);
//                    }
                // Pick amongst all exhibits, weighted towards MRU.
//                else
                    {
                    // Get as list of all cached exhibits, MRU last.
                    final List<CachedFile> cfLRU = new ArrayList<CachedFile>(lru);

                    // Pick an entry, heavily weighted towards the MRU/popular,
                    // since the most recently/frequently used entries
                    // are assumed to be the most important to keep healthy.
                    //
                    // We work hard to distribute these checks well.
                    final int entry = size - 1 -
                        Rnd.goodRnd.nextInt(
                            Rnd.goodRnd.nextInt(
                                Rnd.goodRnd.nextInt(size) + 1) + 1);
                    final CachedFile cf = cfLRU.get(entry);
                    esa = aep.aeid.getStaticAttr(cf.name);
                    }
                }
            }
        finally
            { rwl.readLock().unlock(); }

        // If we have been handed a legitimate exhibit to inspect
        // then test it for validity
        // outside any locks as far as possible.
        if(esa != null)
            {
            // Test validity of entry.
            _doCacheDataValidityTest(aep, esa);
            }
        }

    /**Do incremental purge of orphaned cache entries conditions are right.
     * Tries to grab a write lock to do its work;
     * if it can't get one immediately (ie the cache is busy)
     * then it returns immediately.
     * <p>
     * Don't attempt any clear if we have an empty exhibit set
     * or if there is no cache size currently set
     * since we may not even be properly initialised...
     * <p>
     * We clear at most one orphaned entry on each call.
     */
    private void _incrPurgeOrphanedExhibits()
        throws IOException
        {
        if(ORPHANED_EXHIBIT_EXPIRY_ALLOWED &&
            (_AEP.aeid.length > 0) &&
            (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0))
            {
            if(!rwl.writeLock().tryLock()) { return; }
            try
                {
                // We have run out of space to freely pre-cache (new) exhibits,
                // so a pre-emptive orphaned-exhibit purge may well be worthwhile.
                if(!metaData.someFree())
                    {
                    // Now look for exhibits in the metadata not in the current exhibit set.
                    final Set<String> orphaned = metaData.getKnownExhibits();
                    orphaned.removeAll(_AEP.aeid.getAllExhibitNamesSet());

                    // There are some candidate orphaned exhibits.
                    if(orphaned.size() > 0)
                        {
logger.log("[ExhibitDataSimpleCache: WARNING: orphaned exhibits in cache: " + orphaned.size() + ".]");

                        // Work out minimum time since last use that we will purge.
                        // Note that for a small cache with a fast turnover of exhibits,
                        // we may have to purge sooner than we'd consider ideal.
                        final long thresholdAccessTime;
                        synchronized(metaData)
                            {
                            final CachedFile leastRecentlyUsed = metaData.exhibitsLRU.first();
                            final CachedFile mostRecentlyUsed = metaData.exhibitsLRU.last();
                            // Make threshold for purge much closer to now than to LRU exhibit.
                            thresholdAccessTime = (leastRecentlyUsed.timestamp + 3*mostRecentlyUsed.timestamp) / 4;
                            }
                        final long purgeThreshold = Math.max(thresholdAccessTime,
                            System.currentTimeMillis() - ORPHANED_EXHIBIT_MIN_UNUSED_TIME_MS);

                        // Only consider removing any that have not been used/updated in a while.
                        for(final String name : orphaned)
                            {
                            // Get info on cache entry, if any.
                            final CachedFile info = metaData.exhibitGetInfo(name);

                            // Found unloved orphan, so purge it.
                            if((info != null) && (info.getLastAccessed() < purgeThreshold) &&
                                (info.getLastAccessed() != 0)) // Ensure that this is not broken/partial meta-data.
                                {
                                // Only remove one orphaned exhibit each time
                                // so as to avoid taking too long,
                                // thus stop if successful in removing this one.
                                if(metaData.exhibitRemoveCacheEntry(rwl, cacheDir, name, false, logger, statsIDSCGEN))
                                    {
logger.log("[ExhibitDataSimpleCache: completely removed orphaned exhibit from cache: " + name + ".]");
                                    break;
                                    }
                                }
                            }
                        }
                    }
                }
            finally { rwl.writeLock().unlock(); }
            }
        }


    /**Gets the general properties as a GenProps object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no fresh props have yet been fetched
     * then a default set with a zero timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that we would have been returned) null is returned.
     * <p>
     * We get this from our cache of the immutable data rather than
     * going to the source directly.
     * So we don't block or grab any lock to fetch this value.
     * <p>
     * We do not attempt to persist this data since carrying old GenProps values
     * across a restart may be a very poor idea.
     */
    public org.hd.d.pg2k.svrCore.props.GenProps getGenProps(final long oldStamp)
        /* throws IOException */
        {
        final GenProps cached = genProps;

        if((oldStamp < 0) || (oldStamp != cached.timestamp))
            { return(cached); }

        // Caller seems to have up-to-date copy already.
        return(null);
        }

    /**Last time we polled for genProps.
     * Private to _getGenProps() and is volatile to avoid needing locked access.
     */
    private transient volatile long _lastPollGp;

    /**Attempts to get sysprops if our cached copy may be stale.
     * Slightly strange is that we use our cached sys props value
     * to determine the frequency at which we recheck the cache; the
     * default value is short so we should initially poll quickly until
     * we get a kosher value.
     * <p>
     * A special case here: if we have a GenProps object with a non-zero
     * timestamp (presumably pulled over from a running master) and then
     * we get one with a zero timestamp, we ignore the new, zero, instance
     * since it probably means that the master has just been restarted
     * and has not yet loaded new GenProps.
     * <p>
     * This does not need to hold any locks since all the values it touches
     * are volatile.
     */
    private void _getGenProps()
        throws IOException
        {
        final GenProps currentGenProps = genProps;

        // We return immediately if no need to poll,
        // because last poll was recent enough.
        if(System.currentTimeMillis() - _lastPollGp <=
           (currentGenProps.getWEBSVR_SYSPROPS_RECHECK_MS() * LocalProps.getServerSlowdownFactor()))
            { return; }
        
        // Postpone next poll, whether this succeeds or not.
        _lastPollGp = System.currentTimeMillis();

        // Spin off work into new thread if possible to avoid delaying poll() unduly.
        ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                // Capture old props timestamp.
                final long currentGPTimestamp = currentGenProps.timestamp;

                // Try to fetch new value; this might take a while or even fail...
                final GenProps newGp;
                try { newGp = source.getGenProps(currentGPTimestamp); }
                catch(final IOException e) { e.printStackTrace(); return; }
                // ...but don't let an uninitialised master GenProps displace
                // an initialised one that we already hold.
                if((newGp != null) &&
                   ((currentGPTimestamp == 0) || (newGp.timestamp != 0)))
                    {
                    // OK, we got an updated value.
                    // No lock required to update this volatile value.
                    genProps = newGp;

                    // Log the receipt of a changed props set.
                    logger.log("[ExhibitDataSimpleCache: new GenProps: " +
                        "old/new timestamp: " +
                        (new Date(currentGPTimestamp)) + " / " +
                        (new Date(newGp.timestamp)) +
                        " at " +(new Date()) + ".]");
                    }
                _lastPollGp = System.currentTimeMillis(); // Schedule next poll.                
                }
            });
        }











    /**Gets the security properties as a Properties object if its timestamp is not that specified.
     * If the time specified is negative the object will be returned unconditionally.
     * <p>
     * If no props are currently installed/available a default set with a zero
     * timestamp is returned.
     * <p>
     * If the caller's copy appears to be up-to-date (eg the oldStamp
     * matches that that would have been returned) null is returned.
     * <p>
     * We get this from our cache of the immutable data rather than
     * going to the source directly.
     * We don't block or grab any locks to fetch this.
     * <p>
     * We do not attempt to persist this data since carrying old values
     * across a restart may be a very poor idea.
     * <p>
     * We wrap this as the defaults to a new Properties object
     * to protect our copy against accidental alteration.
     */
    public java.util.Properties getGenSecProps(final long oldStamp)
        /* throws IOException */
        {
        final Properties cached = genSecProps;

        if((oldStamp < 0) || (oldStamp != Long.parseLong(cached.getProperty(SecurityProps.PNAME_GENSECPROPS_TIMESTAMP))))
            { return(new Properties(cached)); }

        // Caller seems to have up-to-date copy already.
        return(null);
        }

    /**Last time we polled for genSecProps.
     * Private to _getGenSecProps(); is volatile to avoid the need for locking.
     */
    private volatile transient long _lastPollGSp;

    /**Our record of the current GenProps; never null.
     * Maintained by poll(); is volatile to avoid the need for locking.
     */
    private volatile Properties genSecProps = (new SecurityProps()).getGenSecProps(); // Start with defaults.

    /**Attempts to get gensecprops if our cached copy may be stale.
     * This does not need any locks since the state is mainatined
     * in volatile values.
     */
    private void _getGenSecProps()
        throws IOException
        {
        // We return immediately if no need to poll,
        // because last poll was very recent.
        if(System.currentTimeMillis() - _lastPollGSp <=
               (genProps.getWEBSVR_SYSPROPS_RECHECK_MS() * LocalProps.getServerSlowdownFactor()))
            { return; }
        
        // Put off next poll for a while, even if this one fails...
        _lastPollGSp = System.currentTimeMillis(); // Schedule/postpone next poll.
        
        ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                try
                    {
                    // Capture old props timestamp.
                    final Properties currentGSP = genSecProps;
                    final String timestamp = currentGSP.getProperty(SecurityProps.PNAME_GENSECPROPS_TIMESTAMP, "0");
                    final long currentGSPTimestamp = Long.parseLong(timestamp);

                    // Try to fetch new value; this might take a while or even fail...
                    final Properties newGSp = source.getGenSecProps(currentGSPTimestamp);
                    // ...but don't let an uninitialised master GenSecProps displace
                    // any initialised one that we already hold.
                    if((newGSp != null) &&
                       ((currentGSPTimestamp == 0) || (Long.parseLong(newGSp.getProperty(SecurityProps.PNAME_GENSECPROPS_TIMESTAMP, "0")) != 0)))
                        {
                        // OK, we got an updated value.
                        // No lock is needed to update this volatile value.
                        genSecProps = newGSp;
                        }
                    _lastPollGSp = System.currentTimeMillis(); // Schedule/postpone next poll.
                    }
                catch(final Exception e)
                    {
                    e.printStackTrace(); // Whinge and stop in case of error.
                    }
                }
            });
        }

    /**Wait time between attempts to fetch or generate thumbnails (ms).
     * We have this in order to avoid pestering a master server unnecessarily
     * or wasting CPU cycles attempting to build a thumbnail.
     * <p>
     * A value of several times the allowed system latency/skew
     * up to of the order of a day in the expectation of a daily
     * exhibit-accession and thumbnail-build cycle on the server
     * is probably reasonable.
     * <p>
     * We randomise the value so that different clients will not conflict
     * with one another unduly.
     */
    private static final long WAIT_BETWEEN_THUMBNAIL_REPEAT_FETCHES_MS =
        Math.min(25 * 3600 * 1000L, // Retries no more than ~1 day apart...
            1013L * (3*CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S +
                Rnd.fastRnd.nextInt(4*CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S)));

    /**Private in-memory cache of deserialised thumbnails; never null.
     * All compound accesses to this object must with lock held on this object,
     * but other accesses are already thread-safe.
     * <p>
     * Is a Hashtable mapping:
     * <ul>
     * <li>from String name to SoftReference to ExhibitThumbnails, or
     * <li>from String name to Long last time we fail to generate a thumbnail
     * </ul>
     */
    private final Map<String,Object> _thumbnailsInMemory =
        new Hashtable<String, Object>(101);

    /**If true then allow missing thumbnails to be fetched synchronously at the risk of blocking for extened periods. */
    private static final boolean ALLOW_SYNC_TN_FETCH = false;

    /**Gets the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     * <p>
     * This retains a private in-memory cache of
     * deserialised thumbnails held by SoftReference,
     * and it tries first to recover them from there.
     * <p>
     * This tries to retrieve thumbnails from the cache,
     * and returns them if they are there.
     * <p>
     * Else, if create is true, this tries to create the thumbnails,
     * cache them, and return the value.  But we won't bother unless
     * the main image is fully loaded.
     * <p>
     * Note that only the read and write of cache is done under lock;
     * the thumbnail generation is unlocked and concurrency is restricted,
     * if at all, by the handler routine(s).
     * <p>
     * Partly because this routine is called by our own
     * precache routines, we do not regard reading a thumbnail
     * as proving user access to the cache
     * (exhibit data has to be read for that).
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible; else if create is false
     *     only return an existing one and return null if none is to hand
     *     (or possibly allow fetch of pre-built remote one)
     */
    public ExhibitThumbnails getThumbnails(final String name,
                                           final boolean create)
//        throws IOException
        {
        // Note inbound request for exhibit thumbnails.
        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_EXTHUREQIN);
        if(!create)
            { StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_EXTHUREQINDC); }

        try
            {
            // Call our internal fetcher/generator.
            final ExhibitThumbnails exhibitThumbnails = _getThumbnails(name,
                    create, ALLOW_SYNC_TN_FETCH);

            // If we could not get the requested thumbnails
            // (or any firm indication that they do not exist)
            // then try again in the background as they may still be useful.
            if(exhibitThumbnails == null)
                { _asyncTNFetch(name); }

            return(exhibitThumbnails);
            }
        finally
            {
            // We must not be holding a (write) lock on exit.
            // We log any held lock and break it and propagate an Error.
            // Of course, this should NEVER happen,
            // but the cost of checking for it is low, so we'll always check.
            if(rwl.isWriteLockedByCurrentThread())
                {
                rwl.writeLock().unlock();
                System.err.println("SEVERE ERROR: getThumbnails(): holding write lock on exit from cache");
                throw new Error("SEVERE ERROR: getThumbnails(): holding write lock on exit from cache");
                }
            }
        }

    /**Incrementally purge stale entries from the in-memory thumbnail cache.
     * On average has O(1) cost regardless of cache size.
     */
    private void _thumbnailsInMemoryTidy()
        {
        // Only spend roughly constant time per poll on cleaning
        // regardless of cache size.
        if(Rnd.fastRnd.nextInt(3 + _thumbnailsInMemory.size()) != 0)
            { return; }

        // Do cleanup in a separate thread if possible.
        ThreadUtils.computeIntensiveThreadPool.submit(new Runnable(){
            public final void run()
                {
                // Atomically take a copy of the set of keys.
                final Set<String> keys;
                synchronized(_thumbnailsInMemory)
                    { keys = new HashSet<String>(_thumbnailsInMemory.keySet()); }

                // Now remove stale entries.
                for(final String key : keys)
                    {
                    assert(ExhibitName.validNameSyntax(key));

                    // Avoid read/update races on the thumbnail cache
                    // but only while examining each individual entry.
                    synchronized(_thumbnailsInMemory)
                        {
                        final Object val = _thumbnailsInMemory.get(key);
                        if(val == null)
                            {
                            // Value has now gone away...
                            continue;
                            }

                        // Deal with cached thumbnails
                        // that have been kicked out of memory...
                        if(val instanceof SoftReference)
                            {
                            final SoftReference sr = (SoftReference) val;
                            if(sr.get() == null)
                                {
                                // Stale entry: zap it!
                                _thumbnailsInMemory.remove(val);
                                continue;
                                }
                            }

                        // Remove orphaned entries quickly.
                        if(_AEP.aeid.getStaticAttr(key) == null)
                            {
                            // Exhibit has gone away, so zap entry!
                            _thumbnailsInMemory.remove(val);
                            continue;
                            }
                        }
                    }
                }
            });
        }

    /**Get the thumbnails for an exhibit.
     * A data source is at liberty to refuse to compute thumbnails
     * in which case it may return null, else it returns a
     * non-null value which may include the `could-not-compute'
     * value to indicate that a thumbnail/sample cannot be made
     * for this exhibit and no attempt need be made in future.
     * <p>
     * This retains a private in-memory cache of
     * deserialised thumbnails held by SoftReference,
     * and it tries first to recover them from there.
     * This is very important to fast delivery of thumbnails
     * for building pages referencing many thumbnails.
     * <p>
     * This tries to retrieve thumbnails from the cache,
     * and returns them if they are there.
     * <p>
     * Else, if create is true, this tries to create the thumbnails,
     * cache them, and return the value.  But we won't bother unless
     * the main image is fully loaded.
     * <p>
     * Note that only the read and write of our tn cache is done under lock;
     * the thumbnail generation is unlocked and concurrency is restricted,
     * if at all, by the handler routine(s).
     * <p>
     * Partly because this routine is called by our own
     * precache routines, we do not regard reading a thumbnail
     * as proving user access to the cache
     * (exhibit data has to be read for that).
     *
     * @param create  if true, and no thumbnail yet exists, try to
     *     create one if possible; else if create is false
     *     only return an existing one and return null if none is to hand
     *     (or possibly allow fetch of pre-built remote one)
     */
    private ExhibitThumbnails _getThumbnails(final String name,
                                             final boolean create,
                                             final boolean allowSyncFetch)
        {
        // In the simple case where we know that thumbnails
        // definitely can't be created
        // (from the type, or because there is no exhibit)
        // return the NO_THUMBNAILS value immediately,
        // ignoring the create parameter.
        // We never cache this value.
        final ExhibitStaticAttr esa = getStaticAttr(name);
        if(esa == null)
            { return(ExhibitThumbnails.NO_THUMBNAILS); }

        final ExhibitMIME.ExhibitTypeParameters type = (ExhibitMIME.getInputFileType(esa.filePath));
        if((type == null) || (type.handler == null) ||
           !type.canPossiblyCreateThumbnailOfSameMIMEType())
            { return(ExhibitThumbnails.NO_THUMBNAILS); }

        // Get some of our important system state for later.
        final GenProps gp = getGenProps(-1L);
        final AllExhibitProperties aep = getAllExhibitProperties(-1L);

        // Get a handle on the private in-memory thumbnail cache.
        // Access to this could in practice be highly concurrent.
        final Map<String,Object> _tIM = _thumbnailsInMemory;

        // Check the memory cache to see if it is still valid
        // and see if we can find our thumbnails in there.
        // Do with a lock held on the table for atomicity.
        //
        // There are these sorts of things in this table:
        //   * A mapping from an exhibit name to a
        //     SoftReference to ExhibitThumbnails
        //     (with the reference possibly null if we ran out of memory).
        //   * Long with the time that we last tried to fetch the thumbnail
        //     from the datasource (or generate it locally);
        //     we should postpone trying again until reasonable time elapses
        //     to avoid wasting bandwidth and CPU cycles.
        //
        // Don't attempt to fetch from upstream if:
        //   * We were told not to fetch it.
        //   * The caller said not to bother creating the thumbnail.
        //   * It isn't long enough since we last tried to create/fetch it.
        boolean dontFetchFromSource = !allowSyncFetch || !create;
        synchronized(_tIM)
            {
            // OK, try from in-memory cache.
            Object o = _tIM.get(name);

            if(o instanceof Long)
                {
                // This was the time we last tried to do a fetch/generate.
                final long lastTime = ((Long) o).longValue();

                // If we have not yet waited long enough
                // before another fetch/generate attempt,
                // then veto any fetch attempt at least.
                if(lastTime + WAIT_BETWEEN_THUMBNAIL_REPEAT_FETCHES_MS <
                    System.currentTimeMillis())
                    {
                    dontFetchFromSource = true;
                    }
                else
                    {
                    // OK, passed the wait time,
                    // so purge this value to save space.
                    o = null;
                    _tIM.remove(name);
                    }
                }

            // See if we have a cached copy of the thumbnail.
            else if(o instanceof SoftReference)
                {
                final SoftReference sr = (SoftReference) o;
                final ExhibitThumbnails tns = (ExhibitThumbnails) sr.get();
                if(tns != null) // Got it!  Excellent!
                    {
                    // Note hit from in-memory thumbnail cache...
                    StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHETNMEMHIT);

                    return(tns);
                    }
                else // Expunge expired reference to save some heap space.
                    { _tIM.remove(name); }
                }
            }

        // If we've got it in the disc cache
        // then return the cached value
        // having saved it to the in-memory cache.
        // Fetching from disc cache grabs a read lock on the cache.
        try {
            final ExhibitThumbnails tns = // _getThumbnailsFromPersistentCache(esa);
                metaData.exhibitGetThumbnails(rwl, cacheDir, esa.filePath, logger);
            if(tns != null)
                {
                // Save thumbnails to the in-memory cache and return it.
                _tIM.put(esa.filePath, new SoftReference<ExhibitThumbnails>(tns));

                // Note hit from on-disc thumbnail cache...
                StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHETNHIT);

                return(tns);
                }
            }
        catch(final IOException e) { } // Ignore.

        // Note cache miss if we have gotten this far...
        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHETNMISS);

        // If we can get it from the underlying data source,
        // then return it having saved it to the disc and in-memory caches.
        // We *don't* hold any locks in the cache while fetching from upstream.
        if(!dontFetchFromSource)
            {
            try {
                final ExhibitThumbnails tns =
                    source.getThumbnails(esa.filePath, create);
                if(tns != null)
                    {
                    // Save in the in-memory cache ASAP.
                    _tIM.put(esa.filePath, new SoftReference<ExhibitThumbnails>(tns));

                    // Now save in the disc cache...
                    // Writing to disc cache grabs a write lock on the cache.
                    try { metaData.exhibitSaveThumbnails(rwl, gp, cacheDir, esa, tns, logger, statsIDSCGEN); }
                    catch(final IOException e) { } // Ignore.

if(IsDebug.isDebug) { logger.log("[ExhibitDataSimpleCase: retrieved thumbnail from source (and cached it): " + esa.filePath + ".]"); }

                    // Return the thumbnails to the caller.
                    return(tns);
                    }

                // Note failure to fetch by putting off any further tries
                // by noting at what time we experienced the failure,
                // unless one of the following mechanisms
                // manages to overwrite our Long with a real set of thumbnails.
                // THUS THE FOLLOWING METHODS MUST UPDATE THE MEMORY CACHE.
                _tIM.put(esa.filePath, new Long(System.currentTimeMillis()));
                }
            catch(final IOException e) { } // Ignore.
            }
        // If not allowed to fetch the thumbnail immediately,
        // then at least make space in the cache for it (and other missing ones)
        // for later (pre)caching if enough space is not currently available...
        else if(metaData.isReadWrite() && !metaData.canPrecacheThumbnails() &&
                (LocalProps.getWEBSVR_MAX_EX_CACHE_BYTES() > 0)) // We trust the cache-sizing properties?
            {
            try
                {
                // Make space for thumbnail(s) to be (pre)cached later.
                // Make space for several in one go here, for efficiency,
                // on the grounds that if we are missing one thumbnail
                // then we may well be missing more than one.
                metaData.purge(rwl,
                         gp,
                         cacheDir,
                         // Very generous notion of elbow room needed...
                         7 * (ExhibitThumbnails.MAX_BYTES_EST +
                              CachedFile.MAX_EMPTY_ENTRY_BYTES_ON_DISC +
                              MAX_REMOTE_FETCH_TO_MAKE_THUMBNAIL +
                              MAX_TRANSFER_CHUNK_SIZE),
                         esa.filePath,
                         logger, statsIDSCGEN);

                logger.log("[ExhibitDataSimpleCache: made space for missing thumbnail(s).]");
                }
            catch(final IOException e) { e.printStackTrace(); } // Absorb error but whinge...
            }


        // Not asked to create thumbnails here,
        // so return a null to indicate that
        // the requested thumbnails were not immediately to hand.
        if(!create) { return(null); }


        ExhibitThumbnails result = null;
        try
            {
            // Don't try to make the thumbnail unless the exhibit is fully loaded
            // (or if we only have at most about a block or
            // the maximum size of the thumbnails themselves to fetch)
            // because it might take an indefinitely long time
            // to load a big exhibit across the Net.
            final CachedFile exhibitMetadata = metaData.exhibitGetInfo(name);
            if(/* !exhibitDataSource.isExhibitFullyLoaded(esa) || */
               (dontFetchFromSource ||
                (null == exhibitMetadata) ||
                (esa.length > MAX_REMOTE_FETCH_TO_MAKE_THUMBNAIL + exhibitMetadata.cachedLength)))
                { return(null); }

            // OK, attempt to generate thumbnails here, not under any lock.
            // Claim that we can take as much memory/time (etc) as we need,
            // since we expect a strictly limited number of these at once,
            // and no user should be waiting for us to finish.
            result = type.handler.makeThumbnails(
                        esa,
                        exhibitDataSource,
                        aep,
                        true);
if(IsDebug.isDebug && (result == null)) { System.err.println("[Failed to generate thumbnail during precache for "+esa+"...]"); }
            }
        catch(final Exception e)
            { e.printStackTrace(); } // Absorb errors but whinge...

        // Having generated it, cache it if not null.
        if(result != null)
            {
            // Save to the in-memory cache.
            _tIM.put(esa.filePath, new SoftReference<ExhibitThumbnails>(result));

            // Save to disc cache.
            try { metaData.exhibitSaveThumbnails(rwl, gp, cacheDir, esa, result, logger, statsIDSCGEN); }
            catch(final IOException e) { } // Ignore.

            // Note local creation of thumbnails...
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_EXTHUCREATED);

            if(IsDebug.isDebug) { logger.log("[ExhibitDataSimpleCase: generated thumbnail (and cached it): " + esa.filePath + ".]"); }
            }

        return(result);
        }

    /**Attempt to asynchronously fetch/create thumbnail that we have failed to return to the user.
     * We so this to attempt to fetch soon any thumbnail that was recently
     * requested but was not immediately available on the grounds that
     * it may be needed again soon.
     * <p>
     * This only uses a strictly limited number of threads,
     * but avoids a wait for general precacheing, which may never happen.
     *
     * @param exhibitName  full exhibit name; non-null valid exhibit name
     */
    private void _asyncTNFetch(final String exhibitName)
        {
        // Don't bother trying to fetch thumbnails in the background
        // for exhibits that cannot have them,
        // ie don't waste time queueing requests that we cannot ever satisfy.
        final ExhibitMIME.ExhibitTypeParameters type = (ExhibitMIME.getInputFileType(exhibitName));
        if((type == null) || (type.handler == null) ||
           !type.canPossiblyCreateThumbnailOfSameMIMEType())
            { return; }

        // Put this in the queue of things to be done.
        discardableReadAheadTaskThreadPool.submit(new Runnable(){
            public final void run()
                {
                // Always allow a sync fetch from server.
                // Always allow local thumbnail creation attempt.
                _getThumbnails(exhibitName, true, true);
                }
            });
        }

    /**Manages our local cache of variables, etc; never null.
     * We set this up to be:
     * <ul>
     * <li><em>Not</em> write-through,
     *     ie we reduce "set" (write) traffic to the upstream source.
     * </ul>
     * <p>
     *
     */
    private final PipelineVarMgr varMgr;

    /**Variable flush/retrieval interval (ms); strictly positive.
     * Based on the allowed distribution latency as centrally defined.
     * <p>
     * We randomise this a little to help keep slaves bothering the master
     * simuktaneously.
     */
    private static int VAR_CACHE_HOLD_TIME_MS = Math.min(30001 + Rnd.fastRnd.nextInt(16384),
        (SystemVariables.MAX_VALUE_DISTRIBUTION_LATENCY_MS / 4) + 1);

    /**Last time we flushed/fetched variables, private to _handleSysVars().
     * We do another flush/fetch when we find this to be null (eg initially)
     * or more than VAR_CACHE_HOLD_TIME_MS in the past.
     * <p>
     * This is volatile so that we do not need to hold a lock to access it.
     */
    private transient volatile Long _handleSysVars_lastFetch;

    /**Last time we saved event historied, private to _handleSysVars(); initially zero.
     * This is volatile so that we do not need to hold a lock to access it.
     */
    private transient volatile long _handleSysVars_evSave;

    /**Handle (update, sync, persist) system variables as required.
     * Recompute any local values generated by the cache itself,
     * flush any outbound values,
     * and retrieve any upstream values periodically.
     * <p>
     * We recompute local variables when we would be prepared to
     * flush/fetch system variables.
     * <p>
     * Timing is handled with volatile values,
     * so we do not need to take out any other locks while working.
     * <p>
     * We rely on the locking within varMgr to ensure consistency,
     * including during the save.  Bad things may happen if trying to
     * remove a cache while we are trying to save event histories!
     * <p>
     * We also use this to recompute any vote/correlation factors.
     *
     * @param force  if true, force an immediate complete save of state upstream
     *     and to disc
     */
    private void _handleSysVars(final boolean force)
        throws IOException
        {
        final long now = System.currentTimeMillis();
        final Long lastTime = _handleSysVars_lastFetch;
        // If still not time for another round of variable updates,
        // then return immediately.
        if(!force &&
           (lastTime != null) &&
           (lastTime.longValue() + VAR_CACHE_HOLD_TIME_MS > now))
            { return; }

        // OK, recompute and set locally-generated variable values.
        // Percentage-of-cache-space used, to float precision...
        final Float percentUsed = new Float((100.0f * metaData.getTotalBytes()) /
            metaData.computeTargetHighWaterMark());
        final SimpleVariableValue pU = new SimpleVariableValue(
            SystemVariables.ExhibitDataSimpleCache_CACHE_AVAIL_SPACE_PERCENT_USED,
            percentUsed);
        varMgr.setVariable(pU);
        // Exhibit (partially or fully) cached count...
        varMgr.setVariable(new SimpleVariableValue(
            SystemVariables.ExhibitDataSimpleCache_CACHED_EXHIBIT_COUNT,
            new Integer(metaData.size())));

        // Push locally-cached updates upstream,
        // and bring upstream values into our local cache.
        // Always request an incremental update.
        try { varMgr.syncVariables(false); }
        catch(final IOException e) { /* Ignore the error so we get to save events... */ }

        // Save updated event histories from time-to-time.
        // Use the form which has bounded space requirements,
        // and persist to the cache area, if extant.
        // We can probably risk saving at most
        // a reasonable fraction of our temporal slackness...
        if((cacheDir != null) &&
           (force || (_handleSysVars_evSave < now)))
            {
            final File evhd = new File(cacheDir, EVENT_HISTORY_DIR);
            if((evhd != null) &&
               (evhd.isDirectory()) &&
               (evhd.canWrite()))
                {
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[event history save start: "+ System.currentTimeMillis() +".]"); }
                // Usually save incrementally to be quick,
                // Only recording histories with updates since the last save
                // and don't leave a trail (eg use bounded disc space).
                varMgr.saveEventHistories(evhd, true, false, !force);
if(ORG.hd.d.IsDebug.isDebug) { logger.log("[event history save end:   "+ System.currentTimeMillis() +".]"); }
                }
            else
                {
                logger.log("[ExhibitDataSimpleCache: WARNING: could not save event history in: " + evhd);
                }

            // Postpone the next event save.
            // Make it a small fraction of the slackness time.
            _handleSysVars_evSave = System.currentTimeMillis() +
                    ((CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S * 1000) / 11) +
                    (Rnd.fastRnd.nextInt(1 + CoreConsts.DEFAULT_TEMPORAL_SLACKNESS_S/13) * 1007);
            }

        // Note when we successfully finished our work.
        _handleSysVars_lastFetch = new Long(System.currentTimeMillis());
        }

    /**Set variable.
     * Set local cached value immediately;
     * store global values to periodically propagate upstream to master
     * but show last global values obtained from master on periodic poll.
     *
     * @throws java.lang.IllegalArgumentException  on attempt to:
     *     set variable with value of wrong type or incompatible defintion,
     *     set non-existent or read-only variable (or these can be ignored)
     */
    public void setVariable(final SimpleVariableValue newValue)
        throws IOException
        {
        varMgr.setVariable(newValue);
        }

    /**Update number of variables at once for efficiency.
     * Is passed a Set of SimpleVariableValues and behaves as if it
     * operates on all of them by calling setVariable() for each item
     * in the Set.
     * <p>
     * This implementation "fails fast" on the first error.
     * <p>
     * This implementation never throws an IOException.
     *
     * @throws java.lang.IllegalArgumentException  on attempt to:
     *     set variable with value of wrong type or incompatible defintion,
     *     set non-existent or read-only variable (or these can be ignored)
     */
    public int setVariables(final SimpleVariableValue[] newValues)
        throws IOException
        {
        return(varMgr.setVariables(newValues));
        }

    /**Get a single variable value; returns null if no such value or wrong type.
     * Always get from local cache.
     * <p>
     * This implementation never throws an IOException.
     */
    public SimpleVariableValue getVariable(final SimpleVariableDefinition var)
        {
        return(varMgr.getVariable(var));
        }

    /**Get immutable Set of variable values altered on or after a given time, or all for -1.
     * Always get from local cache
     * (the variable cache being periodically updated from the master).
     * <p>
     * This may be slow if there are many live variables.
     * <p>
     * This implementation never throws an IOException.
     */
    public SimpleVariableValue[] getVariables(final long changedSince)
        {
        return(varMgr.getVariables(changedSince));
        }

    /**Get the current partial, or previous full, event set at the specified interval; never returns null.
     * This is a simplified interface to return either the current event set
     * that is being collected, or the previous completed set.
     * <p>
     * The current set is the most timely, but may not contain enough data
     * to be meaningful if the new interval has just started.
     * <p>
     * The previous set is complete and thus most likely to have enough samples
     * to be useful, but is not completely current.
     * <p>
     * If the requested event set is not available,
     * an empty synthetic one is created and returned.
     * Thus, with this interface, it is not possible to distinguish between
     * there being no events in the given interval or simply not data.
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  never null
     * @param current  if true the current event set is returned,
     *     else the previous complete set is returned
     *
     * @return  requested event set; never null
     *
     * @throws IllegalArgumentException  if the request arguments are invalid
     */
    public EventVariableValue getEventValue(final SimpleVariableDefinition def,
                                            final EventPeriod intervalSelector,
                                            final boolean current)
        {
        return(varMgr.getEventValue(def, intervalSelector, current));
        }

    /**Get the specified event sets for the specified intervals; never null.
     * This allows retrieval of zero or more event sets for the specified
     * interval size.
     * <p>
     * Requests for more than SystemVariables.EVENT_SAMPLES_RETAINED in the
     * past (or for the future!) cannot be satisfied and data will not be
     * returned for them.
     * <p>
     * Usually not more than SystemVariables.EVENT_SAMPLES_RETAINED samples
     * will be returned in response to any one request as a safety measure.
     * <p>
     * (An implementation that is not an end-point may go upstream to fetch
     * missing values and cache them to satisfy future requests.)
     *
     * @param def  event definition (must be for an event); never null
     * @param intervalSelector  never null
     * @param intervalNumber  a time (as from System.currentTimeMillis())
     *     which identifies the first interval for which data is potentially
     *     required; if too far in the past or future then possibly no data
     *     will be available,
     *     zero is used to access the "all" bucket
     * @param whichValues  each true bit represents a slot for which data is
     *     required, bit 0 indicating data from the slot within which
     *     firstIntervalTime is located, bit 1 the previous slot, etc
     *
     * @return as many of the requested values as available,
     *     at least long enough to return all the available values,
     *     with [0] corresponding to bit 0 in the BitSet;
     *     may contain nulls or be zero-length but is never null
     */
    public EventVariableValue[] getEventValues(final SimpleVariableDefinition def,
                                               final EventPeriod intervalSelector,
                                               final long intervalNumber,
                                               final BitSet whichValues)
        throws IOException
        {
        return(varMgr.getEventValues(def, intervalSelector, intervalNumber, whichValues));
        }

    /**Synchronise with upstream values.
     * Pushes updated values upstream to the source,
     * calls sync on the source with the same "force" argument,
     * and then retrieves changed values from upstream.
     * <p>
     * Holds no externally-visible locks,
     * but if called by multiple threads this will serialise the calls.
     *
     * @param force  if true, this will force a full sync on the read side
     *     by using getVariables(-1) rather than attempting to choose a
     *     nearer timestamp for efficiency;
     *     the implementation is at liberty to use getVariables(-1)
     *     at any time whatever the argument value,
     *     and almost certainly should use it on the first call
     *
     * @throws java.io.IOException if one is received from upstream
     */
    public void syncVariables(final boolean force)
        throws IOException
        { varMgr.syncVariables(force); }


    /**Flag to adjust the aggressiveness of the cache; by default not aggressive.
     * Aggressive cacheing may include read-ahead, and fetching exhibits
     * or at least some leading portion of them to keep the cache full.
     * <p>
     * This can be set false when the system is busy to eliminate most
     * effort not strictly necessary.
     * <p>
     * Volatile to eliminate the need for locking.
     */
    private volatile boolean _aggressive;

    /**Set the aggressiveness of the cache; by default not aggressive.
     */
    public void setAggressive(final boolean isAggressive)
        { _aggressive = isAggressive; }


    /**Get requested Properties selected by key and versionID.
     * Fetches a Properties set unconditionally (versionID == -1)
     * else if the versionID presented is not current.
     *
     * @param key  selector (with possible embedded sub-key)
     *     for desired properties set; never null
     * @param versionID  if -1 then map is always returned if available,
     *     else must be non-negative and null is returned if the versionID
     *     presented matches that of the current version
     *     (ie if the caller has presumably got the up-to-date version);
     *     may be a timestamp or a hash or other value,
     *     and by convention is zero only for an empty properties set
     *
     * @return null, or Properties map guaranteed to contain only
     *     String keys and values
     */
    public java.util.Properties getProperties(final PropsKey key,
                                              final long versionID)
        throws IOException
        {
        return(source.getProperties(key, versionID)); // FIXME: needs cacheing.
        }


    /**Poll periodically (of the order of a second) to do cache maintenance.
     * We keep the poll up to date to keep the work out of a servlet
     * response; the data retrieval might easily take a long time...
     * <p>
     * This routine takes care of calling the upstream poll().
     * <p>
     * We have to be careful about not restricting servlet callers'
     * concurrency here...  We try not to do fetches from the back-end,
     * which may be very slow, with the instance lock, which would shut
     * out all foreground users needlessly.
     * <p>
     * We ignore the caller's GenProps and fetch and cache our own...
     */
    public void poll(final GenProps _gp)
        throws IOException
        {
        // Last error encountered...
        IOException lastErr = null;

        // Do any upstream work...
        try { source.poll(genProps); }
        catch(final IOException e) { lastErr = e; }

        // Make sure that the sysprops cache is not stale.
        try { _getGenProps(); }
        catch(final IOException e) { lastErr = e; }

        // Make sure that the generic-security-props cache is not stale.
        try { _getGenSecProps(); }
        catch(final IOException e) { lastErr = e; }

        // Make sure that the immutable-attr cache is not stale.
        /* try */ { _getAllExhibitProperties(); }
        // catch(final IOException e) { lastErr = e; }

        // Do any incremental precacheing.
        /* try */ { _doPreCache(genProps); }
//        catch(final IOException e) { lastErr = e; }

        // Check our cache meta-data against disc.
        /* try */ { _checkMetaData(); }
        // catch(final IOException e) { lastErr = e; }

        // Save our cache meta-data if we need to.
        try { _cleanAndSaveMetaData(false); }
        catch(final IOException e) { lastErr = e; }

        // Handle system variables if necessary...
        try { _handleSysVars(false); }
        catch(final IOException e) { lastErr = e; }

        // Incrementally purge the thumbnail image cache.
        _thumbnailsInMemoryTidy();

        // Update any vote/correlation values as required.
        // Entirely ignore IOExceptions that are usually entirely benign.
        ThreadUtils.nonCPUThreadPool.submit(new Runnable(){
            public final void run()
                {
                try { _AEP.updateVoteCache(varMgr, false); }
                catch(final IOException e) { }
                }
            });

        // Log if the cache seems to be busy...
        final int ql = rwl.getQueueLength();
        if(ql > 0)
            { logger.log("[ExhibitDataSimpleCache: busy: threads queued for lock: "+ql+".]"); }

        // Rethrow any subordinate error.
        // However, trim simple master-not-in-service errors.
        if(lastErr != null)
            {
if(ORG.hd.d.IsDebug.isDebug) { lastErr.printStackTrace(); }

            if(!(lastErr instanceof PGMasterNotInServiceException))
                { throw lastErr; }
            else
                { logger.log("PGMasterNotInServiceException: " + lastErr.getMessage()); }
            }
        }

    /**Our record of the current GenProps; never null.
     * Maintained by poll() under the instance lock.
     * <p>
     * Is volatile so can be accessed without a lock.
     */
    private volatile GenProps genProps = new GenProps(); // Start with defaults.




    /**If true, we are in purge mode, zapping files before answering each query.
     * Set true when we hit or exceed the cache size limit,
     * set false when we drop below the cache low-water mark.
     * <p>
     * This must be accessed only under the instance lock,
     * and only by _getRawFileViaPersistentCache().
     */
    //private boolean discCachePurgeMode; // Initially false.

    /**Object representing one (partially) cached file on disc.
     * This is immutable, serialisable, and totally ordered.
     * <p>
     * The natural sort order is LRU (Least-Recently Used)
     * first, with ties (which should be very rare) broken
     * by increasing cached length (ignoring thumbnails, etc)
     * (to represent the cost expended getting the data to cache),
     * and then exhibit name.
     * <p>
     * This carries the minimum that need be known about each
     * file on disc.
     * <p>
     * The item is immutable to avoid causing difficulties when
     * already inserted in a sorted collection.
     * <p>
     * We store the timestamp of the original source exhibit file;
     * if it doesn't match the current
     * exhibit timestamp then this cache entry and its cached data
     * has to be regarded as invalidated and discarded.
     */
    private static final class CachedFile implements Comparable,
        Serializable, ObjectInputValidation,
        Cloneable
        {
        /**The official name of the exhibit; never null and always a valid exhibit name. */
        final String name;

        /**The master-exhibit timestamp; strictly positive. */
        final long timestamp;

        /**The initial portion cached on disc (bytes); never negative. */
        final long cachedLength;

        /**The time last written or read; non-negative.
         * May be zero if no exhibit data is cached.
         */
        public long getLastAccessed()
            {
            return(lastAccessed);
            }

        /**The time last written or read; non-negative.
         * We will use setLastModified(System.currentTimeMillis())
         * to enforce our notion of `now', since the filesystem
         * may be mounted from a remote server with a different
         * clock to ours.
         * <p>
         * We only do the setLastModified() when writing to the
         * filesystem anyway, or in the special case of
         * reading the first byte of an exhibit.  We do this to
         * save (synchronous) disc traffic updating timestamps
         * when in fact this should all be in the metadata
         * and is only needed if that is lost for some reason.
         * <p>
         * Logically final, but made mutable for touchedEntry()
         * to be able to overwrite in a cloned copy.
         * <p>
         * May be zero if only metadata but no exhibit data is present.
         */
        /* final */ long lastAccessed;

        /**The size of the serialised thumbnail object, zero if none; never negative. */
        final int tnBytes;


        /**Creates a raw entry given all the data about the entry.
         * The name must be a syntactically-correct exhibit name.
         * <p>
         * Should not be called directly from outside the class.
         */
        private CachedFile(final String _name,
                           final long _ts,
                           final long _len,
                           final long _lastA,
                           final int _tnBytes)
            throws IllegalArgumentException
            {
            name = _name;
            timestamp = _ts;
            cachedLength = _len;
            lastAccessed = _lastA;
            tnBytes = _tnBytes;

            // Verify object state.
            try { validateObject(); }
            catch(InvalidObjectException e)
                { throw new IllegalArgumentException(e.getMessage()); }
            }

        /**Returns true if this exhibit's metadata is essentially equivalent to another one.
         * This insists that everything but the last-accessed timestamp is correct.
         * <p>
         * The result is false if the argument is null.
         */
        boolean isEquivalent(final CachedFile other)
            {
            if(other == null) { return(false); }
            return(name.equals(other.name) &&
                   (timestamp == other.timestamp) &&
                   (cachedLength == other.cachedLength) &&
                   (tnBytes == other.tnBytes));
            }

        /**Do we have thumbnails?
         * True if the thumbnails file is non-zero length.
         * Does not prove that we can load and/or use the thumbnail
         * file even if this returns true.
         */
        boolean hasThumbnails()
            { return(tnBytes != 0); }

        /**Returns thumbnails; never null.
         * In case of difficulty, eg if thumbnails do not exist
         * or cannot be deserialised, throws IOException.
         * <p>
         * We're fairly dumb about this, assuming everything has
         * been checked elsewhere.
         */
        final ExhibitThumbnails getThumbnails(final File cacheDir)
            throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);
            return((ExhibitThumbnails) FileTools.deserialiseFromFile(tnFile, false));
            }

        /**Return data chunk from cached file in the result argument.
         * This file must live in the usual place, in CACHE_EXDATA_DIR,
         * and the name must be a syntactically-correct exhibit name.
         * <p>
         * We touch the file (set its stamp to `now') to show that it
         * has been used iff we retrieve the first byte otherwise
         * we rely on the in-memory metadata to hold the last access time;
         * this is a belt-and-braces measure in case we have to reconstruct
         * the metadata from the disc cache for some reason.
         * <p>
         * This will throw an IOException or IllegalArgumentException
         * if it cannot find the file or there is some other problem.
         * <p>
         * We're fairly dumb about this, assuming that everything has
         * been checked elsewhere.
         * <p>
         * If the read length is zero the buffer may be null.
         *
         * @param quick  be as fast as possible, eg don't update timestamps
         */
        CachedFile getCachedData(final File cacheDir,
                                 final int start,
                                 final ByteBuffer buf,
                                 final boolean quick)
            throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);

            RandomAccessFile raf = null;
            try
                {
                raf = new RandomAccessFile(dataFile, "r");
                raf.getChannel().read(buf, start);
                if(quick) { return(this); }
                final long now = System.currentTimeMillis();
                if(start == 0) { dataFile.setLastModified(now); } // Enforce our clock's `now'.
                return(touchedEntry(now));
                }
            finally
                {
                if(raf != null) { raf.close(); }
                }
            }


        /**Recovers the actual on-disc cached exhibit prefix length, or 0 if not present.
         * Can be used to double-check the on-disc cached data length before
         * adjusting it, but may be slow since it requires a real disc access.
         * Use sparingly.
         * <p>
         * All arguments must be validated and safe before calling this routine;
         * for speed it does no further validation.
         *
         * @param cacheDir  non-null directory containing cache
         * @param exhibitName  name of exhibit; syntactically valid, non-null
         *
         * @return  length cached on disc; 0 if no data cached for this exhibit
         */
        static long getCachedDataLength(final File cacheDir,
                                        final String exhibitName)
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), exhibitName);
            return(dataFile.length());
            }

        /**Fixes-up in-memory cached meta-data for exhibit with that from disc.
         * Reconstructs a set of details for the exhibit from disc,
         * makes sure that the record of the cached length and timestamps is OK,
         * and returns a new record if not (else the original is returned).
         * <p>
         * In particular this returns a new item if:
         * <ul>
         * <li>this record of the cached prefix length is wrong, or
         * <li>the timestamp in this record is older than that recovered, or
         * <li>the record of cached thumbnail data is wrong.
         * </ul>
         * <p>
         * This instance is not altered, nor is the disc-based data.
         * <p>
         * TODO: Possibly discard all cached data for an exhibit with a changed timestamp.
         *
         * @param rec  the meta-data as reconstructed from disc; never null
         */
        CachedFile fixup(final CachedFile rec,
                         final SimpleLoggerIF logger)
            {
            if(!name.equals(rec.name))
                { throw new IllegalArgumentException(); }

            // If something is wrong with this meta-data
            // then we must do a fixup.
            final boolean lengthWrong = (rec.cachedLength != cachedLength);
            if(lengthWrong ||
               (rec.lastAccessed > lastAccessed) || /* old timestamp. */
               (rec.tnBytes != tnBytes))
                {
                // Warn if the problem was cached-data-length,
                // since it might indicate significant lost data.
                if(lengthWrong) { logger.log("[ExhibitDataSimpleCache: WARNING: doing fix-up because in-memory ("+cachedLength+") and on-disc ("+rec.cachedLength+") cached length meta-data differ for "+name+".]"); }

                // Return fixed-up value...
                return(new CachedFile(name, timestamp,
                                      rec.cachedLength,
                                      Math.max(rec.lastAccessed, lastAccessed),
                                      rec.tnBytes));
                }

            // This instance does not need fixing up, so return as-is.
            return(this);
            }

        /**Recovers an approximate entry for an extant cached file; never null.
         * This file must live in the usual place, in CACHE_EXDATA_DIR,
         * and the name must be a syntactically-correct exhibit name.
         * <p>
         * This will throw an IOException or IllegalArgumentException
         * if it cannot find the file or other data that it needs.
         * <p>
         * This does not adjust the in-memory records,
         * nor alter anything on disc.
         * 
         * @return  synthesised cache record; never null
         */
        static CachedFile recoverExtantCachedFileDetails(final File cacheDir,
                                                         final String exhibitName)
            throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), exhibitName);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File stampFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TIMESTAMP_KW + '.' + fileComponent);
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);

            return(new CachedFile(exhibitName,
                // Read the contents of the timestamp file
                // as a decimal signed long.
                // (Note that this file need not be read again since it
                // is assumed not to change and so the in-memory value
                // can be used instead.)
                // We take a missing timestamp file to indicate
                // potential data corruption.
                Long.parseLong(FileTools.readTextFile(stampFile).trim(), 10),

                dataFile.length(), // Must be present, so must be >= 0.

                // We take the newest of the last-modified times of
                // the exhibit data and the thumbnails (if present)
                // as our best approximation to the last-accessed time.
                // (Since java.io.File does not have a lastAccessed() member
                // or equivalent.)
                Math.max(dataFile.lastModified(),
                         tnFile.lastModified()),

                // A thumbnail file may be present,
                // in which case we need its size,
                // else we record 0.
                Math.max(0, (int) tnFile.length()) // Thumbnail file, if extant.
                ));
            }

        /**If true, touchedEntry() used clone() rather than a constructor. */
        private static final boolean _TE_USE_CLONE = true;


        /**Make touched cache file entry.
         * This returns a new object identical to the old (this)
         * one except that the last-accessed stamp is set to the current time of day.
         * <p>
         * This does not alter anything on disc.
         */
        CachedFile touchedEntry()
            { return(touchedEntry(System.currentTimeMillis())); }
        
        /**Make touched cache file entry.
         * This returns a new object identical to the old (this)
         * one except that the last-accessed stamp is set to the
         * time of day passed in (usually the current time of day).
         * <p>
         * This does not alter anything on disc.
         * <p>
         * This could construct a new object,
         * but to save time and avoid the heavy lifting in the constructor,
         * this uses clone().
         *
         * @param newLastAccessedTimestamp  time to make touched version of descriptor with,
         *     strictly positive
         */
        CachedFile touchedEntry(final long newLastAccessedTimestamp)
            {
            assert(newLastAccessedTimestamp > 0);

            if(!_TE_USE_CLONE)
                {
                // Validate arguments via the constructor.
                return(new CachedFile(name,
                    timestamp, cachedLength,
                    newLastAccessedTimestamp,
                    tnBytes));
                }

            // To try to keep this frequent operation as quick as possible,
            // knowing that we are overriding just one value,
            // we use clone() rather than constructing a new object.
            // This avoids redundantly revalidating all the fields.
            else
                {
                assert(newLastAccessedTimestamp > 0);

                final CachedFile result = cloneMe();
                result.lastAccessed = newLastAccessedTimestamp;
                return(result);
                }
            }

        /**Clone the object; does not throw CloneNotSupportedException. */
        private CachedFile cloneMe()
            {
            try { return((CachedFile) super.clone()); }
            catch(CloneNotSupportedException e) { throw new Error(); } // Cannot happen.
            }

        /**Make a new file on disc and return its CachedFile.
         * This is passed the case base dir, the name of the exhibit
         * to have a cache entry created, and the timestamp of the
         * master exhibit itself.
         * <p>
         * No zero-length cache entry is created.
         * <p>
         * No thumbnail is created; any extant one is expunged.
         * <p>
         * This does not adjust the in-memory records.
         */
        static CachedFile makeNewDiscCacheFile(final File cacheDir,
                                               final String name,
                                               final long exhibitStamp)
            throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File stampFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TIMESTAMP_KW + '.' + fileComponent);
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);

            // Expunge any extant files.
            dataFile.delete();
            tnFile.delete();
            stampFile.delete();

            // Make sure that any parent directories exist...
            dataFileDir.mkdirs();

            // Attempt to write the (signed decimal) timestamp file.
            PrintWriter pw = null;
            try {
                pw = new PrintWriter(new FileWriter(stampFile));
                pw.println(exhibitStamp);
                }
            finally
                {
                if(pw != null) { pw.close(); }
                }

            final long now = System.currentTimeMillis();

//            // Write a zero-length data file, and set its timestamp
//            // to our notion of `now' rather than the fileserver's.
//            // Ignore any error report for now...
//            dataFile.createNewFile();
//            dataFile.setLastModified(now);

            return(new CachedFile(name,
                exhibitStamp,
                0, // No data cached...
                now,
                0)); // No thumbnail.
            }


        /**Extend given cache file on disc and return new CachedFile.
         * This is passed the cache base dir, the name of the exhibit
         * to have a cache entry created, the start point of the new data
         * and the data itself, which must be more than zero length.
         * <p>
         * Note that though the new data should normally be exactly at the end
         * of the existing data,
         * it is not an error, though probably inefficient,
         * to start writing before the end of the existing data,
         * since that indicates wasted effort re-cacheing data we already have.
         * <p>
         * It <em>is</em> an error to start writing at a point beyond the end
         * of the existing cached data since we don't support "sparse" data
         * and on some operating systems (eg UNIX) the gaps would be
         * filled with zeros which we probably don't want.
         * <p>
         * If there is an overlap with the existing data,
         * the old (overlapped) data is overwritten silently with the new.
         * This may allow us to update parts in-situ to fix errors,
         * and to silently handle partly-overlapping concurrent updates.
         * <p>
         * The cache file is extended with (the non-overlapping part of)
         * the given data.
         * <p>
         * This does not adjust the in-memory records.
         * <p>
         * This must only be called when other cache write activity
         * (such as removing entries) is locked out
         * to avoid possible file corruption.
         */
        CachedFile extendCacheFile(final File cacheDir,
                                   final long dataStart,
                                   final ByteBuffer data)
            throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);

            if(dataStart < 0)
                { throw new IllegalArgumentException("dataStart must be non-negative"); }
            if(dataStart > cachedLength)
                { throw new IllegalArgumentException("dataStart must be no later than existing cachedLength"); }
            if(data == null)
                { throw new IllegalArgumentException("data buffer must not be null"); }
            final int len = data.remaining();
            if(len < 1)
                { throw new IllegalArgumentException("data length must be greater than zero"); }

            // Double-check that data on disc is as long as we think that it is.
            final long actualLenBefore = dataFile.length();
            if(dataStart > actualLenBefore)
                { throw new IOException("cannot leave gaps in cached data; actual cached data is shorter than expected"); }

            RandomAccessFile raf = null;
            try
                {
                raf = new RandomAccessFile(dataFile, "rw");
//                raf.seek(dataStart);
//                raf.write(data);
                raf.getChannel().write(data, dataStart);
                if(data.remaining() != 0) { throw new IOException("did not write all requested data"); }
                final long now = System.currentTimeMillis();
                dataFile.setLastModified(now); // Enforce our clock's `now'.
                return(new CachedFile(name, timestamp, Math.max(actualLenBefore, dataStart + len), now, tnBytes));
                }
            finally
                {
                if(raf != null) { raf.close(); }
                }
            }

        /**Remove any thumbnail file and return a new in-memory cache entry.
         * This just zaps any thumbnail file and sets the thumbnail length to zero.
         * <p>
         * This does not mark the entry as updated on disc nor in the new
         * record.
         * <p>
         * This does not adjust the in-memory records itself.
         */
        CachedFile zapThumbnails(final File cacheDir)
            //throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);
            tnFile.delete();
            return(new CachedFile(name, timestamp, cachedLength, lastAccessed, 0));
            }

        /**Remove any exhibit data and return a new in-memory cache entry.
         * This just zaps any exhibit data file
         * and records the cached length as zero in the returned value.
         * <p>
         * This attempts to force a deletion even if the file is not obviously present,
         * just to make best efforts to purge it,
         * and we whinge (on System.err) if we cannot make it go away.
         * <p>
         * This does not mark the entry as updated on disc nor in the new record.
         * <p>
         * This does not adjust the in-memory records itself.
         */
        CachedFile zapData(final File cacheDir)
            //throws IOException
            {
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final boolean dfExists = dataFile.exists();
            final boolean dfDeleted = dataFile.delete();
if(dfExists && !dfDeleted && dataFile.exists()) { System.err.println("ERROR: ExhibitDataSimpleCache: unable to zapData() for: "+name); }
            return(new CachedFile(name, timestamp, 0, lastAccessed, tnBytes));
            }

        /**Add a thumbnail file and return a new in-memory cache entry.
         * This is not allowed if we already think we have a
         * thumbnail or if there is a thumbnail file already on disc.
         * <p>
         * The thumbnail argument must be non-null.
         * <p>
         * This does not mark the entry as updated on disc or in the new
         * record.
         * <p>
         * This does not adjust the in-memory records itself.
         */
        CachedFile saveThumbnails(final File cacheDir,
                                  final ExhibitThumbnails tns)
            throws IOException
            {
            if(tns == null)
                { throw new IllegalArgumentException(); }

            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);

            if(hasThumbnails() || tnFile.exists())
                { throw new IOException("thumbnails already exist"); }

            FileTools.serialiseToFile(tns, tnFile, false, true);
            final long now = System.currentTimeMillis();
            dataFile.setLastModified(now); // Enforce our clock's `now'.
            final int tnBytesNew = (int) tnFile.length();
            if(tnBytesNew <= 0)
                {
                tnFile.delete();
                throw new IOException("error saving thumbnail");
                }
            return(new CachedFile(name, timestamp, cachedLength, now, tnBytesNew));
            }




        /**Equality relies on comparison, ie on last-access time, cached length and name.
         */
        public final boolean equals(final Object o)
            {
            if(!(o instanceof CachedFile)) { return(false); }
            return(compareTo(o) == 0);
            }

        /**The hash depends on the timestamp.
         * Two objects can only be equal if their timestamps and cached lengths are.
         */
        public final int hashCode()
            {
            return(((int) lastAccessed) ^ ((int) cachedLength));
            }

        /**Total ordering: oldest first, then smallest first, then by name.
         */
        public final int compareTo(final Object o)
            {
            final CachedFile other = (CachedFile) o;

            // Oldest first, primarily.
            if(lastAccessed < other.lastAccessed) { return(-1); }
            if(lastAccessed > other.lastAccessed) { return( 1); }

            // Then smallest first.
            if(cachedLength < other.cachedLength) { return(-1); }
            if(cachedLength > other.cachedLength) { return( 1); }

            // Then sort by name to break ties, ie provide a total ordering.
            return(name.compareTo(other.name));
            }

        /**Rough estimate of maximum space required for new empty cache entry.
         * This assumes a maximum-length exhibit name and no data nor thumbnail
         * for the exhibit.
         * <p>
         * This should approximately match what calcDiscSpace() should
         * produce for such an exhibit, but the values may not be
         * exactly the same.
         */
        static final int MAX_EMPTY_ENTRY_BYTES_ON_DISC = (int) (
            // Basic file name representation...
            FileTools.roundUpToFSBlockSize(ExhibitName.MAX_NAME_LENGTH) +
            // Extra directory space for timestamp file...
            FileTools.roundUpToFSBlockSize(ExhibitName.MAX_NAME_LENGTH-2+32));

        /**Calculate the disc space taken up by this exhibit cache entry all told.
         * This guesses a little for overheads such as
         * the directory entry, rounding for allocation space on disc,
         * etc, aiming to be slightly conservative.
         * <p>
         * This assumes a roughly UFS-like (UNIX File System) pattern of disc usage.
         */
        final long calcDiscSpace()
            {
            // Estimate of space taken by exhibit,
            // including full cost of its name from top of tree
            // presenting there is no sharing of directories
            // nor per-path-component directory overhead.
            final long exhibitDataSpace =
                FileTools.roundUpToFSBlockSize(cachedLength) + // For exhibit data.
                FileTools.roundUpToFSBlockSize(name.length()); // For dir entry.

            // Estimated storage consumed as a result of
            // length of final portion of each aux file name
            // plus some overhead for prefixes,
            // the directory entry itself, etc.
            final int dirEntry = 32 + name.length() - name.lastIndexOf('/');

            // Extra storage cost of timestamp file.
            final long stampFileSpace =
                FileTools.FS_EST_BLOCK_SIZE_BYTES + // For `timestamp' file block.
                dirEntry; // For extra directory entry.

            // Extra storage cost of thumbnails file,
            // iff the thumbnails exist.
            final long tnFileSpace = (tnBytes == 0) ? 0 :
                (FileTools.roundUpToFSBlockSize(tnBytes) + // For `thumbnails' file block.
                dirEntry); // For extra directory entry.

            return(exhibitDataSpace + stampFileSpace + tnFileSpace);
            }

        /**Used to zap my disc files, including all auxiliary files for this exhibit.
         * Quietly ignores any errors.
         * <p>
         * This does not adjust the in-memory records.
         */
        final void zapMe(final File cacheDir)
            {
            assert(cacheDir != null);
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), name);
            final File dataFileDir = new File(dataFile.getParent());
            final String fileComponent = dataFile.getName();
            final File stampFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TIMESTAMP_KW + '.' + fileComponent);
            final File tnFile = new File(dataFileDir,
                CACHE_EXAUX_PREFIX + CACHE_EXAUX_TN_KW + '.' + fileComponent);
            dataFile.delete();
            stampFile.delete();
            tnFile.delete();
            }


        /**My serialisation version number. */
        private static final long serialVersionUID = -5246653973669373340L;

        /**Deserialise. */
        private void readObject(final ObjectInputStream in)
            throws IOException, ClassNotFoundException
            {
            in.defaultReadObject();
            validateObject(); // Validate state immediately.
            }

        /**Validate fields/state.
         * Called in the constructor and possibly after de-serialising.
         * <p>
         * Barf if something bad is found.
         * (Maybe allow some extra info in debug version.)
         */
        public void validateObject()
            throws InvalidObjectException
            {
            // Check that all components are sane and safe.
            if(!ExhibitName.validNameSyntax(name))
                { throw new InvalidObjectException("bad object: invalid name: " + name); }

            if(timestamp <= 0) // Could maybe have tighter constraint...
                { throw new InvalidObjectException("bad object: invalid timestamp"); }

            if(cachedLength < 0)
                { throw new InvalidObjectException("bad object: invalid cached length"); }

            if(lastAccessed < 0) // Could maybe have tighter constraint...
                { throw new InvalidObjectException("bad object: bad access time: " + lastAccessed); }

            if(tnBytes < 0)
                { throw new InvalidObjectException("bad object: invalid thumbnail size"); }
            }
        }


    /**Fraction of max cache size that is the low-water mark.
     * In the range ]0.0f, 0.1f[ excluding both end points.
     * <p>
     * We will only do precacheing when the cache size is below
     * the low-water mark.
     * <p>
     * Don't get this too close to 1 to avoid churning the cache
     * when loading large single blocks of data
     * or upon other minor disturbances.
     */
    private static final float LOW_WATER_FRACTION = 0.9f;

    /**Routine to do incremental pre-cacheing.
     * Exits as soon as it detects other threads queueing for the cache lock.
     * <p>
     * This relies on the following items being fetched and entirely
     * maintained by other activity/methods, probably under poll():
     * <ul>
     * <li>GenProps
     * <li>AllExhibitProperties
     * </ul>
     * <p>
     * This will only do any data prefetching while the cache is
     * below the low-water mark, and will limit the amount
     * of client and master resource used.
     * <p>
     * We won't actually start any precacheing until
     * we see some evidence of user/upstream activity that
     * might eventually benefit from it.
     * <p>
     * This will also only precache while the "aggressive" flag is set,
     * and this should be set true only when the system is not busy.
     * <p>
     * This may also involve precomputation/preloading of optional data,
     * though should not be relied in lieu of other methods to keep this fresh,
     * so we may have work to do even if there is no space for prefetching.
     * <p>
     * This also incrementally checks the cache for consistency with the
     * current exhibit properties, ie timestamp, size, hashes.
     * <p>
     * This does not hold a cache lock for its duration,
     * but does hold a private lock to protect its internal state
     * because it must not be multi-threaded;
     * any attempt to run this in a second thread is quietly vetoed.
     */
    private void _doPreCache(final GenProps gp)
        {
        // If no evidence of continuing downstream use of this cache instance
        // then return immediately.
        if(!_userRequestedDataFromCache) { return; }

        // If not in an aggressive mode
        // then return immediately.
        if(!_aggressive) { return; }

        // If it is too soon to do any more precacehing
        // then return immediately.
        if(System.currentTimeMillis() < _noMorePrecacheUntil) { return; }

        // If cache is too full to allow any precacheing
        // then return immediately.
        if(!metaData.canPrecache()) { return; }

        // If there is anyone waiting to acquire a cache lock,
        // then quit immediately so as not to cause extra contention and so as
        // to keep precacheing as unintrusive as possible.
        if(rwl.hasQueuedThreads()) { return; }

        // If we're already running a precache thread
        // then don't start another one.
        if(_preCacheLock.isLocked()) { return; }

        // Spin off a thread for the next bit of precacheing...
        // We do this to avoid blocking poll() for long periods.
        // This is executed immediately, in this thread if the pool is full.
        discardableReadAheadTaskThreadPool.submit(new Runnable(){
            public final void run()
                {
                // Don't attempt to precache if already in progress...
                if(!_preCacheLock.tryLock()) { return; }
                try
                    {
                    final long startTime = System.currentTimeMillis();

                    // If it is too soon to do any more precacehing
                    // then return immediately.
                    if(startTime < _noMorePrecacheUntil) { return; }

                    // If we have been asked to consume < 1% resources
                    // then just don't even try precaching.
                    final int websvr_bw_limiter =
                        gp.getWEBSVR_BW_LIMITER() * LocalProps.getServerSlowdownFactor();
                    if(websvr_bw_limiter > 100) { return; }

                    // OK, do some precache work...
                    final AllExhibitProperties aep = _AEP;
                    final AllExhibitImmutableData aeid = aep.aeid;

                    // Force expiry of the precache iterator
                    // if we can tell that the exhibit set has changed
                    // since the iterator was created.
                    final Long exhibitHash = new Long(aep.longHash);

                    try
                        {
                        if((_precacheExhibitHash != null) && !exhibitHash.equals(_precacheExhibitHash))
                            { _precacheIterator = null; }

                        // If the precache iterator has `expired'
                        // then reset it.
                        if((_precacheIterator == null) ||
                           !_precacheIterator.hasNext())
                            {
                            // If the hash is (non-null and) equal to the exhibit hash,
                            // it means that there is no further precaching work to do
                            // for now.
                            if(exhibitHash.equals(_precacheExhibitHash))
                                {
                                // Postpone precacheing until the exhibit set changes
                                // OR data is requested from the cache by a visitor.
                                _userRequestedDataFromCache = false;
                                _precacheExhibitHash = null;
                                logger.log("[ExhibitDataSimpleCache: stopped precaching until more user activity seen.]");
                                return;
                                }

                            ScorerCacheIF scorers = null; // FIXME: should get access to RealThing(TM).

                            // We (re)set it to the full set of
                            // currently valid (full) exhibit names,
                            // since those are the only things that we should be
                            // aggressively precaching (there may be defunct
                            // items in the metadata that have been deleted
                            // from the exhibit set, for example).
                            //
                            // Sort the list by recent download popularity,
                            // then sort the list "best"-first,
                            // at least to a quick approximation,
                            // since we usually will not have the
                            // fully-computed "goodness" data available
                            // when an exhibit set changes for example.
                            //
                            // We give priority to exhibits with
                            // no metadata or thumbnails cached yet.
                            //
                            // For speed we'll accept stale/approximate data,
                            // as this only alters the order of processing.
                            final EventVariableValue downloads = varMgr.getEventValue(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD, EventPeriod.VLONG, false);
                            final ArrayList<String> al = new ArrayList<String>(aeid.getAllExhibitNamesSet());
                            Collections.sort(al, (new Comparator<String>(){
                                /**Sort "best"-first, being prepared to accept stale data if need be. */
                                public final int compare(final String s1, final String s2)
                                    {
                                    // First give priority to exhibits with
                                    // no metadata or thumbnails cached yet.
                                    final CachedFile cf1 = metaData.exhibitGetInfo(s1);
                                    final boolean hasMetaData1 = (cf1 != null);
                                    final CachedFile cf2 = metaData.exhibitGetInfo(s2);
                                    final boolean hasMetaData2 = (cf2 != null);
                                    // Item without metadata should come first.
                                    if(hasMetaData1 != hasMetaData2)
                                        { return(hasMetaData1 ? +1 : -1); }

                                    final ExhibitMIME.ExhibitTypeParameters exhibitType1 = ExhibitMIME.getExhibitType(ExhibitName.getExtensionComponent(s1));
                                    final boolean needsTN1 = (cf1 != null) && !cf1.hasThumbnails() &&
                                                             (exhibitType1 != null) && exhibitType1.canPossiblyCreateThumbnailOfSameMIMEType();
                                    final ExhibitMIME.ExhibitTypeParameters exhibitType2 = ExhibitMIME.getExhibitType(ExhibitName.getExtensionComponent(s2));
                                    final boolean needsTN2 = (cf2 != null) && !cf2.hasThumbnails() &&
                                                             (exhibitType2 != null) && exhibitType2.canPossiblyCreateThumbnailOfSameMIMEType();
                                    // Item with missing thumbnail should come first.
                                    if(needsTN1 != needsTN2)
                                        { return(needsTN2 ? +1 : -1); }

                                    // Now sort with more-downloads (more-popular) item first.
                                    final int dl1 = downloads.getCount(s1);
                                    final int dl2 = downloads.getCount(s2);
                                    if(dl1 > dl2) { return(-1); /* Correct order. */ }
                                    if(dl1 < dl2) { return(+1); /* Wrong order. */ }

                                    // If we've not yet taken much time
                                    // then we can potentially force entries up-to-date as we go.
                                    final boolean runningSlow = (System.currentTimeMillis() - startTime) > MAX_dPC_SPIN_TIME_MS/2;

                                    ScorerCacheIF scorers = null; // FIXME: should get access to RealThing(TM).

                                    // Then sort by (integer) goodness value.
                                    final int g1 = aep.getExhibitPropsComputableMutable(s1, runningSlow, gp, ExhibitDataSimpleCache.this, scorers).getGoodness();
                                    final int g2 = aep.getExhibitPropsComputableMutable(s2, runningSlow, gp, ExhibitDataSimpleCache.this, scorers).getGoodness();
                                    if(g1 > g2) { return(-1); /* Correct order. */ }
                                    if(g1 < g2) { return(+1); /* Wrong order. */ }
                                    return(0); // Identical goodness.
                                    }
                                }));

                            // Compute the "best" exhibits for enhanced precaching.
                            final int lastIndex = al.size() / 10;
                            // Filter to a list of "good", non-stale-EPCM items.
                            final List<String> bestFiltered = new ArrayList<String>(al.subList(0, Math.min(lastIndex, MAX_BEST_EX_PRECACHED)));
                            // Since items to be trimmed will more often be at the end
                            // work backwards through the list to minimise copying.
                            for(int i = bestFiltered.size(); --i >= 0; )
                                {
                                final String ex = bestFiltered.get(i);
                                final ExhibitPropsComputableMutable epcm =
                                    aep.getExhibitPropsComputableMutable(ex, true, gp, ExhibitDataSimpleCache.this, scorers);
                                // Remove stale/bad entry.
                                if((epcm == null) || (epcm.getGoodness() <= 0) || epcm.isStale())
                                    { bestFiltered.remove(i); }
                                }
if(IsDebug.isDebug) { logger.log("[ExhibitDataSimpleCache: bestFiltered.size()="+bestFiltered.size()+".]"); }

                            // Update the best-exhibit list atomically...
                            synchronized(_bestExhibits)
                                {
                                _bestExhibits.clear();
                                _bestExhibits.addAll(bestFiltered);
                                }

                            // Reinsert the "bestFiltered" entries at the start
                            // so that they get two bites of the cherry each pass.
                            al.addAll(0, bestFiltered);

                            // Help avoid wasted memory.
                            bestFiltered.clear();
                            al.trimToSize();

                            // Create/save the iterator.
                            _precacheIterator = al.iterator();

                            // Note the hash of the current exhibit set.
                            _precacheExhibitHash = exhibitHash;

                            // Note restart of precaching scan of all exhibits.
                            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_PRECACHERESTART);
                            logger.log("[ExhibitDataSimpleCache: started precaching round for "+aeid.length+" exhibits.]");
                            }

                        // Do the precacheing work.
                        // Stop when we have used up all the pending names to examine,
                        // or we've used a just-GUI-significant amount of time,
                        // or we see other users queueing to get into the cache...
                        do
                            {
                            // If cache is too full to allow any precacheing
                            // then return immediately.
                            if(!metaData.canPrecache()) { return; }

                            // If there is anyone waiting to acquire a cache lock,
                            // then quit immediately so as not to cause extra contention and so as
                            // to keep precacheing as unintrusive as possible.
                            if(rwl.hasQueuedThreads()) { return; }

                            // Stop if we have run out of exhibits to precache...
                            if(!_precacheIterator.hasNext())
                                {
                                if(_precacheExhibitHash != null)
                                    {
                                    logger.log("[ExhibitDataSimpleCache: finished precaching round with no work to do.]");
                                    return;
                                    }

                                logger.log("[ExhibitDataSimpleCache: finished precaching round with work still pending.]");

                                // Force the meta-data out to disc
                                // since we did do some work this time around.
                                metaData.setNeedsSave();
                                return;
                                }

                            // Note precache examination of next exhibit.
                            StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_PRECACHEEXAMINED);

                            // Get the next item to examine and make sure that
                            // it is still a valid exhibit.
                            final String nextToExamine = _precacheIterator.next();
                            final ExhibitStaticAttr esa = aeid.getStaticAttr(nextToExamine);
                            if(esa == null)
                                {
                                // If we find that an exhibit name from the iterator
                                // is no longer valid
                                // then force creation of a new iterator from
                                // the new exhibit set.
                                // This is to stop us wasting time on deleted exhibits.
                                // This should restart this precacheing round
                                // rather than terminating it.
                                final Set<String> empty = Collections.emptySet();
                                _precacheIterator = empty.iterator();
                                _precacheExhibitHash = null;
                                break;
                                }

                            // Update one exhibit.
                            // Keep trying to bring it up-to-date if:
                            //   * This exhibit is not completely cached.
                            //   * We have not run out of time.
                            //   * We draw a Gaussian random number smaller than
                            //     the exhibit's goodness.
                            // The aim is that better exhibits should have relatively
                            // more time spent on bringing them up to date.
                            //
                            // An IOException terminates this burst of precaching.
                            //
                            // Note that we are prepared to try to force
                            // the EPCM to be (re)computed if necessary
                            // as a service to the rest of the system.
                            ScorerCacheIF scorers = null; // FIXME: should get access to RealThing(TM).
                            while(_updateOneExhibit(esa, gp, aep, false) &&
                                  (Rnd.fastRnd.nextGaussian() <
                                        1.5*aep.getExhibitPropsComputableMutable(esa.filePath, Rnd.fastRnd.nextBoolean(), gp, ExhibitDataSimpleCache.this, scorers).getGoodnessAsFloat()))
	                            {
	                            // If there is anyone waiting to acquire a lock,
	                            // then quit immediately to let them in
	                            // to keep precacheing as unintrusive as possible.
	                            if(rwl.hasQueuedThreads()) { return; }

	                            // If we run out of time,
	                            // then quit immediately.
	                            if((System.currentTimeMillis() - startTime) > MAX_dPC_SPIN_TIME_MS) { return; }
	                            }

                        // If we still haven't taken much time,
                        // then examine another exhibit.
                        } while((System.currentTimeMillis() - startTime) < MAX_dPC_SPIN_TIME_MS);
                    }
                catch(final IOException e)
                    {
                    // Absorb any IOException quietly.
                    }
                finally
                    {
                    // See how long all that took,
                    // and put off the next bit of precacheing appropriately.
                    // We limit our sleep time in case of freak behaviour!
                    //
                    // If we finish a whole precaching round
                    // then put off more precache work for a reasonable time.
                    final long now = System.currentTimeMillis();
                    final long timeTaken = now - startTime;

                    _noMorePrecacheUntil = now +
                        Rnd.fastRnd.nextInt(31) +
                            (((_precacheIterator != null) && _precacheIterator.hasNext()) ? 11 : 3*MAX_dPC_BACKOFF_TIME) +
                        (Math.min(MAX_dPC_BACKOFF_TIME, timeTaken) * websvr_bw_limiter);
                    }
                }
            finally { _preCacheLock.unlock(); }
            }
        });
        }

    /**Partly check the cache data (including metadata, tns, etc) for validity.
     * This picks one or more aspects (at random) of the currently cached data
     * for the specified exhibit for validity.
     * <p>
     * Checks include length, timestamp, and hashes
     * dependent on the data available.
     * <p>
     * This is designed to complete reasonably quickly in most cases,
     * to perform an incremental check,
     * removing or in some cases repairing damaged/corrupt/invalid data.
     * <p>
     * If this finds the cache entry to be broken somehow
     * then this routine may delete the cache entry entirely,
     * or repair the data, or prune back to some valid prefix of the data held,
     * or just remove the corrupt underlying data to leave the metadata
     * to be fixed at a later date.
     * <p>
     * This will grab locks only as it needs them
     * in order to be as unintrusive as possible.
     * <p>
     * This is not expected to be needed very often,
     * but is mainly designed to avoid silent disc corruption,
     * and to provide for automatic repair.
     *
     * @param aep the current exhibit properties; never null
     * @param esa the exhibit whose cache data is to be verified; never null
     */
    private final void _doCacheDataValidityTest(final AllExhibitProperties aep,
                                                final ExhibitStaticAttr esa)
        throws IOException
        {
        assert((aep != null) && (esa != null));

        // Get a snapshot of the cache data, if any, for this exhibit.
        CachedFile cf = metaData.exhibitGetInfo(esa.filePath);

        // If nothing in the cache for this exhibit then return immediately;
        // there is nothing to do!
        if(cf == null)
            { return; }

//if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache._doCacheDataValidityTest(): validating exhibit "+esa+"...]"); }

        // If there is a gross change,
        // ie in length or timestamp,
        // then remove the current cache entry entirely and return.
        if((cf.timestamp != esa.timestamp) ||
           (cf.cachedLength > esa.length))
            {
logger.log("ERROR: cached data timestamp/length wrong for exhibit: " + esa);
            _removeCorruptData(cf);
            return;
            }

        // If thumbnail data exists but is corrupt/unusable then remove it.
        if(cf.hasThumbnails())
            {
            ExhibitThumbnails tns = null;

            // Grab cache read lock just long enough to read the thumbnails.
            // This is assumed to force a validation check during deserialisation.
            _getReadLock(rwl, "ExhibitDataSimpleCache._doCacheDataValidityTest() reading thumbnails", logger);
            try { tns = cf.getThumbnails(cacheDir); }
            catch(final Exception e) { e.printStackTrace(); }
            finally { rwl.readLock().unlock(); }

            // If we couldn't load/validate the thumbnails
            // or they are of the less less-safe form
            // without built-in timestamp/checksums
            // then get a write lock and zap them
            // and use the updated metadata subsequently.
            // Iff this cache entry is removed entirely while we are working
            // then we will end up harmlessly causing and absorbing an NPE.
            if((tns == null) || (tns.created <= 0) ||
               ((tns.getSmall() != null) && !tns.getSmall().hasMD5Hash()) ||
               ((tns.getStandard() != null) && !tns.getStandard().hasMD5Hash()))
                {
                logger.log("WARNING: purging damaged or old-style thumbnail for "+esa+": "+tns);
                _getWriteLock(rwl, "ExhibitDataSimpleCache._doCacheDataValidityTest() purging damaged/old thumbnail file", logger);
                try { metaData._update(rwl, cf = metaData.exhibitGetInfo(esa.filePath).zapThumbnails(cacheDir).touchedEntry(), logger); }
                catch(final Exception e) { e.printStackTrace(); }
                finally { rwl.writeLock().unlock(); }
                }
            }

        // If we don't have the accession data then we have no hashes to check.
        final ExhibitPropsLoadable exhibitPropsLoadable = aep.getExhibitPropsLoadable(esa.filePath);
        final AccessionData ad = exhibitPropsLoadable.getAccessionMetadata();
        if(ad == null) { return; }

        // OK, note when we are about to start the validation.
        final long startTime = System.currentTimeMillis();

        // We can check the whole-exhibit hashes
        // if we have the whole exhibit in cache
        // AND we have the hashes available.
        final boolean wholeExhibitInCache = (cf.cachedLength == esa.length);
        final boolean canCheckWholeExhibit = wholeExhibitInCache &&
            ((ad.hashCRC32 != null) || (ad.hashMD5 != null));

        if(canCheckWholeExhibit)
            {
            // Set up our hashes...
            final java.util.zip.Checksum hCRC32 = new java.util.zip.CRC32();
            final MessageDigest hMD5;
            try { hMD5 = MessageDigest.getInstance(CoreConsts.HASH_MD5); }
            catch(final NoSuchAlgorithmException e) // Should never happen...
                { throw new Error("could not find "+CoreConsts.HASH_MD5+" digester!"); }

            // Create our buffer...
            // We can economise on memory for very short exhibits...
            final byte buf[] = new byte[Math.min(AccessionData.HASH_BLOCK_SIZE_BYTES, (int) cf.cachedLength)];

            // To examine the whole file without locking out cache activity
            // we'll read the exhibit file a block at a time,
            // reacquiring the cache read lock each time.
            // We ignore any data added to the cached file while we work.
            // If the file gets *shorter* while we're working
            // (this *might* be legit if the exhibit was tossed out of cache)
            // or we encounter any other unexpected event
            // (eg an IOException caused by an unreadable disc sector),
            // then we just trash the whole cached entry.
            final File dataFile = new File(new File(cacheDir, CACHE_EXDATA_DIR), esa.filePath);
            final DataInputStream dis = new DataInputStream(new FileInputStream(dataFile));
            try
                {
                for(int block = 0; ; ++block)
                    {
                    // Compute start offset of this block.
                    final int start = block * AccessionData.HASH_BLOCK_SIZE_BYTES;

                    // Stop when we run out of file!
                    if(start >= cf.cachedLength) { break; /* Finished. */ }

                    // Compute size of block to read.
                    final int len = Math.min(AccessionData.HASH_BLOCK_SIZE_BYTES,
                                                (int) (cf.cachedLength - start));

                    // Grab cache read lock just long enough to read one block.
                    _getReadLock(rwl, "ExhibitDataSimpleCache._doCacheDataValidityTest() reading one block of full file", logger);
                    try { dis.readFully(buf, 0, len); }
                    finally { rwl.readLock().unlock(); }

                    // Update the hashes.
                    hCRC32.update(buf, 0, len);
                    hMD5.update(buf, 0, len);
                    }

                // Now extract the final hashes and compare them
                // against the accession data...
                final Integer hashCRC32 = new Integer((int) hCRC32.getValue());
                final ROByteArray hashMD5 = new ROByteArray(hMD5.digest());

                // Check the CRC32 hash if we have it in the accession data.
                if((ad.hashCRC32 != null) && !hashCRC32.equals(ad.hashCRC32))
                    {
                    logger.log("ERROR: cached data failed CRC32 check ("+hashCRC32+"/"+ad.hashCRC32+") for exhibit: " + esa);
                    _removeCorruptData(cf);
                    return;
                    }

                // Check the MD5 hash if we have it in the accession data.
                if((ad.hashMD5 != null) && !hashMD5.equals(ad.hashMD5))
                    {
                    logger.log("ERROR: cached data failed MD5 check ("+hashMD5+"/"+ad.hashMD5+") for exhibit: " + esa);
                    _removeCorruptData(cf);
                    return;
                    }

                // OK, exhibit was fully validated!
                StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHE_VALIDATION);
//if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache._doCacheDataValidityTest(): whole-file hashes verified for exhibit: "+esa+".]"); }

                final long endTime = System.currentTimeMillis();
                final long dur = endTime - startTime;
                if(dur > 60000) // Warn about relatively slow validations...
                    { logger.log("[ExhibitDataSimpleCache: validation took "+dur+"ms for "+esa+".]"); }
                return;
                }
            catch(final IOException e)
                {
                _removeCorruptData(cf); // Zap dodgy data ASAP.
                logger.log("ERROR: unexpected IOException when checking cached data (so removed ALL data) for exhibit: " + esa + ": " + e.getMessage());
                e.printStackTrace();
                return;
                }
            finally
                { dis.close(); }
            }


        // TODO: write the per-block hash checks...

        }

    /**Remove cached exhibit data identified as corrupt.
     * If possible then this clears up the metadata and raw data, etc,
     * but if not then this just removes the suspect raw data, thumbnails, etc,
     * from the disc cache.
     * <p>
     * The metadata will have to get fixed later.
     * <p>
     * Since this will have to grab a cache write lock
     * and then the metadata lock if it is to remove the corrupt data,
     * the caller must not have a read lock in place for example.
     *
     * @param cf   exhibit cache metadata; never null
     */
    private void _removeCorruptData(final CachedFile cf)
        throws IOException
        {
        // Note removal of corrupt exhibit from cache...
        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_CACHEREM_CORRUPT);

        logger.log("WARNING: ExhibitDataSimpleCache: cached exhibit corrupt so being completely purged from cache: " + cf.name);
        if(!metaData.exhibitRemoveCacheEntry(rwl, cacheDir, cf.name, false, logger, statsIDSCGEN))
            {
            // If we could not remove the entry neatly,
            // eg because the metadata was read-only,
            // then just remove the corrupt underlying data brutally!
            logger.log("WARNING: ExhibitDataSimpleCache: cannot update metadata; just removing raw data: " + cf.name);
            // Grab both locks in correct order...
            _getWriteLock(rwl, "ExhibitDataSimpleCache._removeCorruptData() raw data removal", logger);
            try
                {
                synchronized(this)
                    { cf.zapMe(cacheDir); /* Zap raw cache data. */ }
                }
            finally
                { rwl.writeLock().unlock(); }
            }
        return;
        }

    /**Update one exhibit incrementally during precaching.
     * Must not be called within a lock; will grab any locks it needs.
     *
     * @param forceDataFetch  if true, strongly encourage data fetch,
     *     ie extension of exhibit data if at all possible
     *
     * @return true  iff this exhibit is not completely cached,
     *     ie may benefit from another call to this routine
     */
    private boolean _updateOneExhibit(final ExhibitStaticAttr esa,
                                      final GenProps gp,
                                      final AllExhibitProperties aep,
                                      final boolean forceDataFetch)
        throws IOException
        {
        boolean notCompletelyCached = false;

        // If possible, attempt to fetch first/next block of an exhibit.
        // We are very keen to do this for "popular"/tiny/"best" exhibits,
        // even when the cache is quite full and we otherwise might not.
        // We also like to bring in the tail of very-nearly-completely-cached
        // exhibits so that we can promptly check hashes, etc.
        CachedFile icf = metaData.exhibitGetInfo(esa.filePath);
        final long cachedLen = (icf == null) ? 0 : icf.cachedLength;
        final boolean dataNotFullyCached = cachedLen < esa.length;
        final boolean canPrecacheAtAll = metaData.canPrecache();
        if(metaData.canPrecacheExhibitData() ||
           (canPrecacheAtAll && forceDataFetch /* Asked to extend if at all possible. */ ) ||
           (canPrecacheAtAll && metaData.someFree() &&
                ((esa.length <= MAX_TRANSFER_CHUNK_SIZE) /* Small exhibit. */ ||
                 ImageUtils.canBeOwnThumbnail(esa) /* Tiny exhibit. */ ||
                 _bestExhibits.contains(esa.filePath) /* Good exhibit. */ ||
                 ((icf != null) && dataNotFullyCached &&
                    ((esa.length - cachedLen) <= MAX_TRANSFER_CHUNK_SIZE))  /* Very-nearly-completely cached exhibit. */ ||
                 isPopularDownload(esa) /* Popular exhibit. */ )))
            {
            // Attempt to extend the on-disc copy of the exhibit
            // by one block, if not yet fully cached.
            if(dataNotFullyCached)
                {
                // Exhibit data is not fully cached locally...
                // Read one whole chunk more for efficiency,
                // up to the lower of the file end or cache limit.
                // For now, we don't bother to align this read.

                // Find out what initial portion we are prepared to cache.
                final int MAX_CACHEABLE_EX_BYTES = gp.getWEBSVR_MAX_CACHEABLE_EX_BYTES();

                // Max position to aspire to read up to...
                final long readLimit =
                        Math.min(esa.length, MAX_CACHEABLE_EX_BYTES);
                // Maximum number of bytes that we will actually read...
                // We read the biggest normally-allowed chunk for efficiency.
                final int toRead = Math.min(MAX_TRANSFER_CHUNK_SIZE,
                                            (int) (readLimit - cachedLen));
                if(toRead > 0)
                    {
                    try
                        {
                        // Extend our cache if possible.
                        _getExhibitDataFromUpstreamToPrecache(esa, aep.aeid, gp, cachedLen, toRead, false);

                        // Note successful fetch/cache of data block.
                        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_PRECACHEEXDATABLOCK);

                        // Read data successfully,
                        // but would it be useful to read some more immediately?
                        final CachedFile icf2 = metaData.exhibitGetInfo(esa.filePath);
                        if((icf2 == null) || (icf2.cachedLength < readLimit))
                            {
                            notCompletelyCached = true; // This exhibit could do with more precaching work.
                            }
                        else
                            {
                            logger.log("[ExhibitDataSimpleCache: completely precached data for "+esa+"; will now test data integrity...]");

                            // Now test the exhibit's hash(es), etc,
                            // to ensure that everything arrived intact...
                            _doCacheDataValidityTest(aep, esa);
                            }
                        }
                    catch(final IOException e)
                        {
                        // Note error during precaching...
                        StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_PRECACHEERROR);
                        throw e;
                        }
                    finally
                        {
                        // If file still not read as far as possible/allowed,
                        // then note that there is still work to do...
                        final CachedFile icf2 = metaData.exhibitGetInfo(esa.filePath);
                        if((icf2 == null) || (icf2.cachedLength < readLimit))
                            {
                            _precacheExhibitHash = null; // Work to be done...
                            }
                        }
                    }
                }
            }

        // If possible and necessary,
        // then attempt to get/make thumbnails...
        //
        // We'll create a cache entry for this exhibit first if need be.
        // This allows us to pre-cache just metadata
        // when the server is in a "slow" (less aggressive) mode.
        if(metaData.canPrecacheThumbnails())
            {
//            CachedFile icf = metaData.exhibitGetInfo(esa.filePath);
            if(icf == null)
                {
                // Try to create new entry...
                // Ignore the return code...
                metaData.exhibitCreateNewCacheEntry(rwl,
                                                    cacheDir,
                                                    esa,
                                                    logger,
                                                    statsIDSCGEN);
                // ...but try again to get the cache entry.
                icf = metaData.exhibitGetInfo(esa.filePath);
                }

            // Is it worth precaching a thumbnail?
            // Only do it if:
            //  0) There is a cache entry for this exhibit.
            //  1) There is not yet a thumbnail.
            // AND
            //  2) A thumbnail can potentially be made.
            // THEN
            //  3) We try to fetch the thumbnail directly from the
            //     back end without forcing it to be created...
            // AND IF THAT DOESN'T GET US THE THUMBNAIL, THEN
            //  4) If we have the full exhibit data immediately available
            //     (note that this might not be in local cache,
            //     eg on the master this might be coming from the
            //     filesystem still), and
            //     if it looks worth doing, then we call the normal
            //     front-end getThumbnails() routine to MAKE the
            //     thumbnail from the exhibit data.
            if((icf != null) &&
               !icf.hasThumbnails() &&
               (((ExhibitMIME.getInputFileType(esa.filePath)) != null) &&
               (ExhibitMIME.getInputFileType(esa.filePath)).canPossiblyCreateThumbnailOfSameMIMEType()))
                {
                // Try to fetch pre-made thumbnails from upstream...
                // (Don't "create" them.)
//                try
                    {
                    // Try to force thumbnail to be fetched/made.
                    if(_getThumbnails(esa.filePath, true, true) == null)
                        // If we can't get the thumbnail made this time
                        // then clear the _precacheExhibitHash
                        // to indicate there is work still to be done,
                        // eg on another precache pass if need be.
                        { _precacheExhibitHash = null; } // Work to be done...
                    }
//                catch(final IOException e)
//                    {
//                    // Note error during precaching...
//                    StatsLogger.captureDataPoint(statsIDSCGEN, SCGNAME_PRECACHEERROR);
//                    throw e;
//                    }
                }
            }

        // Try to get mutable properties computed.
        ScorerCacheIF scorers = null; // FIXME: should get access to RealThing(TM).
        final ExhibitPropsComputableMutable epcmS =
                aep.getExhibitPropsComputableMutable(esa.filePath, true, gp, this, scorers);
        if(epcmS.isStale())
            {
            // If we find that the current properties are stale
            // then we take that as an indication
            // that others (already passed over) may need recomputing too.
            _precacheExhibitHash = null; // Quite aggressive...

            // Now attempt to force recomputation for this value...
            aep.getExhibitPropsComputableMutable(esa.filePath, false, gp, this, scorers);
            }

// DO OTHER PRE-CACHE OPERATIONS FOR THIS EXHIBIT...

        return(notCompletelyCached);
        }


    /**If true, allow us to try fetching some exhibit data from peers rather than master. */
    private static final boolean ALLOW_DATA_FETCH_FROM_PEERS = true;

    /**Thread-safe Map from mirror ID to strictly-positive rating with "" for master; never null.
     * The rating is a synthetic (milliseconds) time to fetch a big data block.
     * Lower values represent peers that have the data available quickly
     * more of the time.
     * <p>
     * All values are strictly positive.
     * <p>
     * This is periodically purged of stale data (ie inactive peers)
     * to keep it from growing without bound.
     */
    private final Map<String,Long> altDataSourceRating = new Hashtable<String, Long>();

    /**If true then use a cautious strategy to select a peer to talk to. */
    private static final boolean PEER_SELECTION_CAUTIOUS = false;

    /**Default rating/time (ms) for "unknown" data source/mirror/peer; strictly positive.
     * The value chose corresponds to 10000ms RTT and 1kBps throughput,
     * ie much slower than most reasonable peers.
     */
    private static final int PEER_STATS_UNKNOWN_MS = 10000 + MAX_TRANSFER_CHUNK_SIZE;

    /**Time-constant for updating peer fetch time value; strictly positive.
     * Higher values means a lower-pass filter,
     * and more robustness in the face of temporary glitches.
     * <p>
     * A value that is a power of two may result in more efficient code.
     * <p>
     * A value from 8 to 256 is probably reasonable.
     */
    private static final int PEER_STATS_TC = PEER_SELECTION_CAUTIOUS ? 64 : 32;

    /**Fake tag we use to indicate a fetch from the master via the pipeline. */
    private static final String MASTER_FAKE_TAG = "";

    /**Try to extend cached data for given exhibit.
     * By default we try to get this from the master,
     * but we may try to fetch it from a peer/mirror instead (P2P)
     * to reduce the load on the master.
     * <p>
     * Also, this may be used to attempt data/error recovery
     * if data cannot be fetched from the master for some reason.
     *
     * @param start  byte offset in exhibit to start reading/fetching for; non-negative
     * @param len  number of bytes to read; strictly positive.
     * @param forceIt  if true then we try very hard to get the data from a peer,
     *     for example to help with master/server error recovery
     */
    private void _getExhibitDataFromUpstreamToPrecache(final ExhibitStaticAttr esa,
                                                       final AllExhibitImmutableData aeid,
                                                       final GenProps gp,
                                                       final long start,
                                                       final int len,
                                                       final boolean forceIt)
        throws IOException
        {
        assert(esa != null);
        assert(aeid != null);
        assert(gp != null);
        assert(start <= Integer.MAX_VALUE);
        assert(len > 0);

        // Dummy buffer to read the data into.
        // The data is discarded; we only want the side-effect of it beign cached.
        final ByteBuffer buf = ByteBuffer.allocate(len);

        // Attempt to adaptively fetch data from peers if allowed.
        // This may significantly reduce load on the master for exhibit data.
        if(ALLOW_DATA_FETCH_FROM_PEERS || forceIt)
            {
            final String mirrorTag = LocalProps.getMirrorTag();
            // This host has to be a mirror itself to be allowed to use P2P...
            if(mirrorTag != null)
                {
                // Get candidate mirrors...
                // We use whatever (even stale) entries that we can find
                // in order to allow P2P to work longer after
                // losing contact with the master
                // (when it may prove to be especially useful).
                final Set<String> activeMirrors = new HashSet<String>(
                    LoadBalancingUtils.getActiveMirrors(varMgr, false).keySet());
                // Remove this instance/mirror as a candidate...
                activeMirrors.remove(mirrorTag);

                // Fall back to using the master as usual if no other candidates...
                if(activeMirrors.size() > 0)
                    {
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache: potential peers for data fetch: "+(new ArrayList<String>(activeMirrors))+".]"); }

                    // Add the master as a synthetic candidiate...
                    activeMirrors.add(MASTER_FAKE_TAG);

                    // Choose which peer to contact...
                    final String peerToTry = _pickPeer(activeMirrors);

                    // If we selected the master
                    // then just fall through to the default handler.
                    if(!MASTER_FAKE_TAG.equals(peerToTry))
                        {
                        final long startFetch = System.currentTimeMillis();
                        boolean successful = false;
                        try
                            {
                            metaData.exhibitRead(rwl,
                                                 peerToTry,
                                                 cacheDir, esa.filePath, source,
                                                 aeid,
                                                 gp,
                                                 (int) start,
                                                 buf,
                                                 null, // Cache locally but not upstream...
                                                 logger,
                                                 statsIDSCGEN);

                            // Got the data OK!
                            successful = true;

                            // Don't need to go to the master!
                            return;
                            }
                        catch(final IOException e)
                            {
                            logger.log("INFO: data block fetch from peer for exhibit "+esa+" failed with IOException: " + peerToTry + ": " + e.getMessage());
                            /* Fall through to get data from master. */
                            }
                        finally
                            {
                            final long endFetch = System.currentTimeMillis();
                            _updatePeerStats(peerToTry,
                                             successful,
                                             (endFetch - startFetch));

                            // Remove some stats logging from the timing...
                            if(successful)
                                {
                                // Note successful fetch/cache of data block.
                                StatsLogger.captureDataPoint(statsIDSCGEN, SCGPREF_PRECACHEEXDATABLOCKSRC + peerToTry);
                                logger.log("INFO: successful data block fetch from peer: "+peerToTry+"; stats: " + (new ArrayList<Map.Entry<String,Long>>(altDataSourceRating.entrySet())));
                                }
                            else
                                {
                                // Note unsuccessful fetch/cache of data block.
                                final String ev = SCGPREF_PRECACHEEXDATABLOCKSRCERR + peerToTry;
                                StatsLogger.captureDataPoint(statsIDSCGEN, ev);
                                logger.log("INFO: FAILED data block fetch from peer "+peerToTry+"; stats: " + (new ArrayList<Map.Entry<String,Long>>(altDataSourceRating.entrySet())));
                                }
                            }
                        }
                    }
else /* if(IsDebug.isDebug) */ { System.out.println("[ExhibitDataSimpleCache: NO potential peers for data fetch...]"); }
                }
            }

        // Fall through to use the master/upstream by default.

        // Get the data from the master...
        // Note stats...
        final long startFetchFromMaster = System.currentTimeMillis();
        boolean fetchFromMasterSuccessful = false;
        try
            {
            metaData.exhibitRead(rwl,
                                 null, // Fetch from upstream/master.
                                 cacheDir, esa.filePath, source,
                                 aeid,
                                 gp,
                                 (int) start,
                                 buf,
                                 null, // Cache locally but not upstream...
                                 logger,
                                 statsIDSCGEN);
            fetchFromMasterSuccessful = true;
            }
        finally
            {
//            // Possibly note/update stats for large fetches only.
//            final boolean largeFetch = (len >= CoreConsts.BULK_DATA_TRANSFER_SIZE);
//            if(largeFetch)
                {
                final long endFetchFromMaster = System.currentTimeMillis();
                _updatePeerStats(MASTER_FAKE_TAG,
                                 fetchFromMasterSuccessful,
                                 (endFetchFromMaster - startFetchFromMaster));

                // Remove some stats logging from the timing...
                if(fetchFromMasterSuccessful)
                    {
                    // Note successful fetch/cache of data block.
                    StatsLogger.captureDataPoint(statsIDSCGEN, SCGPREF_PRECACHEEXDATABLOCKSRC + "master");
                    }
                else
                    {
                    // Note unsuccessful fetch/cache of data block.
                    final String ev = SCGPREF_PRECACHEEXDATABLOCKSRCERR + "master";
                    StatsLogger.captureDataPoint(statsIDSCGEN, ev);

                    logger.log("INFO: data block fetch from peer stats: " + (new ArrayList<Map.Entry<String,Long>>(altDataSourceRating.entrySet())));
                    }
                }
            }
        }

    /**Fraction of the time to pick a peer completely at random; strictly positive. */
    private final int P2P_RND_FRAC = (PEER_SELECTION_CAUTIOUS ? 37 : 17) +
                                    Rnd.fastRnd.nextInt(11);

    /**Fraction of the time to choose a 2nd-tier peer rather than the best peer; strictly positive. */
    private final int P2P_NEXT_BEST_FRAC = 3 + Rnd.fastRnd.nextInt(3);

    /**Factor/multiplier of peers worse than top that will be considered for fetches routinely; strictly positive. */
    private static final int NEARLY_TOP_FACTOR = 2;

    /**If true, deter use of master where peers are available. */
    private static final boolean PREFER_PEERS_TO_MASTER_FOR_DATA_PREFETCH = true;

    /**Pick one of the supplied peers to attempt to fetch the exhibit data from; never null.
     * Usually this returns the "best" peer;
     * sometimes this will return an apparently sub-optimal peer so as to:
     * <ul>
     * <li>to test the waters (ie keep sampling all peers occasionally),
     * <li>spread load around a little, and
     * <li>rehabilitate once-poor peers (eg bad guys may come good).
     * </ul>
     * In particular we are allowing for peer/server load
     * and inter-peer network conditions to change continually.
     * <p>
     * Occasionally this will purge the cached peer stats of anything
     * not in the argument set.
     *
     * @param activeMirrors  set of mirror tags
     *     (and possible "" for the master) of size two or greater; never null
     * @return selected peer; never null
     */
    private String _pickPeer(final Set<String> activeMirrors)
        {
        assert(activeMirrors != null);
        assert(activeMirrors.size() > 1);

        // A small fraction of the time pick a peer completely at random
        // (providing it is the master or seems to be "up").
        // This allows us to rehabilitate peers that fall out of favour,
        // and spreads data load a little.
        // This "hot potato" selection method is also fairly quick.
        if(Rnd.fastRnd.nextInt(P2P_RND_FRAC) == 0)
            {
            final String putativePeer = (new ArrayList<String>(activeMirrors)).get(
		        Rnd.goodRnd.nextInt(activeMirrors.size()));
            if((MASTER_FAKE_TAG.equals(putativePeer) ||
                (Boolean.TRUE == LoadBalancingUtils.testIfHTTPServerIsUp(HostUtils.makeMirrorNameGeneric(putativePeer), true))))
			    { return(putativePeer); }
            }

        // Pick a peer in relation to its weighting,
        // ie pick "better" peers (with lower service times) more often.
        // For the moment aim to pick the peer with the lowest score.
        long bestScore = Long.MAX_VALUE;
        String selectedPeer = MASTER_FAKE_TAG; // Default to the master.
        for(final String peer : activeMirrors)
            {
            final Long score = altDataSourceRating.get(peer);

            if(!PEER_SELECTION_CAUTIOUS)
                {
                // If not feeling cautious,
                // then try any new peer immediately that it becomes available,
            	// providing that it is the master or already seems to be "up".
                if((score == null) &&
                		(MASTER_FAKE_TAG.equals(peer) ||
                		    (Boolean.TRUE == LoadBalancingUtils.testIfHTTPServerIsUp(HostUtils.makeMirrorNameGeneric(peer), true))))
                    { return(peer); }
                }

            // Treat unknown peers as having a high "unknown" fallback time, a la XNTPD.
            // Possibly mark the master down to prefer peers, so as to reduce strain on the master.
            final long sl = (score == null) ? PEER_STATS_UNKNOWN_MS :
                ((PREFER_PEERS_TO_MASTER_FOR_DATA_PREFETCH && MASTER_FAKE_TAG.equals(peer)) ? (score.longValue() * 2*NEARLY_TOP_FACTOR) :
                    score.longValue());
            if(sl < bestScore)
                {
                bestScore = sl;
                selectedPeer = peer;
                }
            }

        // Some of the time, when scores are sufficiently good,
        // select randomly any peer with a score close enough to the best.
        // This helps spread the load amongst, and keep an eye on,
        // the top few candidates.
        //
        // We deliberately exclude any peer whose status is unknown (or worse).
        if((bestScore < PEER_STATS_UNKNOWN_MS / (2*NEARLY_TOP_FACTOR)) &&
            (Rnd.fastRnd.nextInt(P2P_NEXT_BEST_FRAC) == 0))
            {
            final List<String> candidatePeers = new ArrayList<String>(1 + activeMirrors.size()/2);
            for(final String peer : activeMirrors)
                {
                final Long score = altDataSourceRating.get(peer);

                // Skip unknown peers altogether...
                if(score == null) { continue; }

                // Possibly mark the master down to prefer peers.
                final long sl = ((PREFER_PEERS_TO_MASTER_FOR_DATA_PREFETCH && MASTER_FAKE_TAG.equals(peer)) ? (score.longValue() * 2*NEARLY_TOP_FACTOR) :
                    score.longValue());
                if(sl < NEARLY_TOP_FACTOR*bestScore) // Consider peer if relatively good...
                    { candidatePeers.add(peer); }
                }
            // Override best peer if we have a real choice...
            // Use goodRnd to make as independent a selection as possible.
            final int nCands = candidatePeers.size();
            if(nCands > 1)
                { selectedPeer = candidatePeers.get(Rnd.goodRnd.nextInt(nCands)); }
            }

        // Take this opportunity to clear out stale data,
        // amortising the mean cost-per-call to be roughly independent of peer-set size.
        if(Rnd.fastRnd.nextInt(altDataSourceRating.size() + 2) == 0)
            {
            // Atomically remove any entries for peers not in the argument set.
            synchronized(altDataSourceRating)
                {
                for(final Iterator<String> it = altDataSourceRating.keySet().iterator(); it.hasNext(); )
                    {
                    final String peer = it.next();
                    if(!activeMirrors.contains(peer))
                        { it.remove(); }
                    }
                }
            }

        // Return the selected peer...
        return(selectedPeer);
        }

    /**Update data-transfer stats for the given peer.
     * A failed fetch is treated as a very slow access,
     * so that a failing peer compares numerically unfavourably with reliable peers.
     */
    private void _updatePeerStats(final String peer,
                                  final boolean fetchSuccessful,
                                  final long timeTaken)
        {
        assert(peer != null);

        // Treat a fail as worse numerically than "unknown"
        // and as worse than the actual time taken
        // so that consistent failures are liked less than unknown/slow peers.
        final long l = fetchSuccessful ? timeTaken :
            (1 + (timeTaken + PEER_STATS_UNKNOWN_MS));

        // Atomic wrt the cache...
        synchronized(altDataSourceRating)
            {
            // If there is no entry for this peer then create one now.
            Long t = altDataSourceRating.get(peer);
            if(t == null)
                {
                if(PEER_SELECTION_CAUTIOUS)
                    {
                    // Start with a (very) cautious estimate of performance...
                    t = new Long(PEER_STATS_UNKNOWN_MS);
                    }
                else
                    {
                    // Start with the value/time that we see for this new peer.
                    t = new Long(l);
                    }
                altDataSourceRating.put(peer, t);
                }

            // Compute new value for the cache.
            // Back off fast in case of failure.
            final int timeConst = fetchSuccessful ? PEER_STATS_TC :
                                    (Math.max(1, PEER_STATS_TC/2));
            final Long v = new Long(Math.max(1, ((t.longValue() * (timeConst-1)) + l + (timeConst/2)) / timeConst));
            altDataSourceRating.put(peer, v);

            // Record some stats...
if(IsDebug.isDebug) { System.out.println("[ExhibitDataSimpleCache._updatePeerStats(\""+peer+"\", "+fetchSuccessful+", "+timeTaken+") new value = "+v+"...]"); }
            }

        if(fetchSuccessful)
            {
            StatsLogger.captureDataPoint(statsIDSCGEN, SCGPREF_PRECACHEEXDATABLOCKFETCHTIME + GenUtils.log2Approx(Math.max(0, timeTaken)));
            }

        // Should we be recording performance/usage stats centrally?
        final boolean recordStats = "true".equals(genProps.getGen().get(KEY_debugFlag_P2P_BLOCKXFER));
        // If so, do it now.
        if(recordStats)
            {
            final StringBuilder sb = new StringBuilder(32);
            sb.append("P2P:bx:");
            sb.append(LocalProps.getMirrorTag());
            sb.append(':');
            sb.append("".equals(peer) ? "M" : peer);
            sb.append('=');
            if(fetchSuccessful)
                {
                // Note time taken for successful transfer.
                sb.append(GenUtils.log2Approx(timeTaken));
                }
            else
                {
                // Note failure.
                sb.append("FAIL");
                }

            try
                {
                varMgr.setVariable(new SimpleVariableValue(
                    SystemVariables.PERFMON_STRING_GLOBAL_EVENT,
                    sb.toString()));
                }
            catch(final IOException e)
                {
                e.printStackTrace(); // Whinge but absorb error...
                }
            }
        }

    /**In the top-N (global) downloads recently, with at least 2 downloads.
     * The "more than one download" filter is to trim off a noisy "long tail"
     * (eg on a quiet day).
     * <p>
     * This uses a combination of hits from the previous and current days.
     *
     * @param esa  the exhibit details; never null
     * @return true if this exhibit is a "popular" download.
     */
    private boolean isPopularDownload(final ExhibitStaticAttr esa)
        throws IOException
        {
        final EventVariableValue downloadsYesterday = varMgr.getEventValue(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD, EventPeriod.VLONG, false);
        final EventVariableValue downloadsToday;
        final String shortName = ExhibitName.getFileComponent(esa.filePath);
        return(((downloadsYesterday.getRank(shortName) < 100) &&
                (downloadsYesterday.getCount(shortName) > 1)) ||
               // Note that fetching current period incurs more cost.
               (((downloadsToday = varMgr.getEventValue(SystemVariables.ACCESSPATTERN_COMPLETED_DOWNLOAD, EventPeriod.VLONG, true)).getRank(shortName) < Integer.MAX_VALUE) &&
                (downloadsToday.getCount(shortName) > 1)));
        }

    /**Precache lock to prevent multi-threaded precache runs. */
    private final ReentrantLock _preCacheLock = new ReentrantLock();

    /**Time before which we will not do more precacheing.
     * Private to _doPreCache().
     * <p>
     * Initial value of 0 allows precaching to start immediately.
     */
    private volatile transient long _noMorePrecacheUntil;

    /**Set true when a user requests data from the cache.
     * This is as a result of user activity, and without this
     * being true we still won't indulge in any precacheing.
     * This means that we can safely have more than one context
     * set up (for example), in a servlet runner as long as
     * only one is ever actually used.
     * <p>
     * Accessed without locking; read by _doPreCache()
     * and may be set by a routine that is sure that it has received
     * a user request to fetch exhibit/thumbnail data.
     * <p>
     * We can reset this if we believe we have finished precaching
     * (or at least a reasonable chunk of precacheing work)
     * for the current exhibit set.
     */
    private transient volatile boolean _userRequestedDataFromCache;

    /**An iterator over a snapshot of all exhibit names.
     * This is initially null, and when null or we reach the
     * end of the underlying set of values it is reset to
     * be a new snapshot of the exhibit names.  This avoids
     * starvation of some exhibits.
     * <p>
     * We access this only from _doPreCache(), so it need not
     * be thread safe.
     * <p>
     * We may order the iteration in some way as to try to
     * precache as efficiently as possible, eg smallest or
     * `best' first, or we might store the exhibits in, for
     * example, a shuffled order.
     * <p>
     * When we get a name from this iterator we must make sure
     * that it still represents a valid exhibit, since the exhibit
     * might have been deleted, for example.
     * <p>
     * Accessed only by _doPreCache().
     */
    private transient Iterator<String> _precacheIterator;

    /**Indicator for which image set we are working on.
     * When we start a new round of precaching we set this
     * to the hash of the current exhibit set.
     * <p>
     * If we come across an exhibit that we do some precaching work on,
     * we set this to null.
     * <p>
     * When we are about to start a new round of precaching
     * and discover this is set to the hash of the current exhibit set,
     * we assume that there was no work to be done and we skip precaching.
     * When the exhibit set changes we will then resume.
     * <p>
     * Accessed under the rwl by _doPreCache().
     */
    private transient Long _precacheExhibitHash;

    /**Maximum time that _doPreCache() can spend in one go (ms).
     * Designed to be short enough to avoid causing massively irritating
     * interruptions to user interactivity if we lock other activity
     * out for this long,
     * though long enough to be relatively efficient if possible.
     * <p>
     * Precacheing will not generally interfere with interactive operations
     * so we try to make this time large enough
     * to allow the fetch of a block or three of wxhibit data over a slow Net link,
     * allowing for RTT and connection setup and bandwidth, etc.
     * <p>
     * Something of the order of a few seconds may be good.
     */
    private final int MAX_dPC_SPIN_TIME_MS = 3607 + Rnd.fastRnd.nextInt(1703);

    /**Max time _doPreCache() has to sleep for (ms).
     * This is basically if some freak event happens beyond
     * _doPreCache()'s reasonable control.
     * <p>
     * A few minutes is probably reasonable.
     */
    private final int MAX_dPC_BACKOFF_TIME = 5 * 60 * 1000 +
        Rnd.fastRnd.nextInt(2 * 60 * 1000);


    /**Set true once destroy() is called; never set false again. */
    private volatile boolean destroyed;

    /**Shut down the data pipeline.
     * Flush state, variables and logs to disc,
     * then make sure that upstream of us is destroyed.
     */
    public void destroy()
        {
        // Quickly discourage precaching and free some memory...
        _aggressive = false;
        _userRequestedDataFromCache = false;
        _thumbnailsInMemory.clear();

        // Stop the thread pool.
        discardableReadAheadTaskThreadPool.shutdownNow();

        // Try to save any useful persistable state.
        try
            {
            _cleanAndSaveMetaData(true);
            _handleSysVars(true);
            }
        catch(final IOException e) { e.printStackTrace(); }
        finally { source.destroy(); }

        // Mark this instance as doomed once we've completed any saves.
        destroyed = true;

        // Check for background threads...
        if(_gAEP_lock.isLocked())
            { System.err.println("WARNING: AEP fetch thread still running..."); }
        // Grab read lock on cache to prevent further changes.
        try { _getReadLock(rwl, "shutting down cache in destroy()", logger); }
        catch(final InterruptedIOException e) { e.printStackTrace(); }

        // Prevent further updates to metadata
        // and thus most cache updates.
        metaData.setReadWrite(false);
        }
    }
